Package: Luminescence
Type: Package
Title: Comprehensive Luminescence Dating Data Analysis
Version: 1.2.1
Date: 2026-03-25
Authors@R: c(
    person("Sebastian", "Kreutzer", role = c("aut", "trl", "cre", "dtc"), email = "maintainer_luminescence@r-luminescence.org", comment = c(ORCID = "0000-0002-0734-2199")),
    person("Christoph", "Burow", role = c("aut", "trl", "dtc"), comment = c(ORCID = "0000-0002-5023-4046")),
    person("Michael", "Dietze", role = c("aut"), comment = c(ORCID = "0000-0001-6063-1726")),
    person("Margret C.", "Fuchs", role = c("aut"), comment = c(ORCID = "0000-0001-7210-1132")),
    person("Christoph", "Schmidt", role = c("aut"), comment = c(ORCID = "0000-0002-2309-3209")),
    person("Manfred", "Fischer", role = c("aut", "trl")),
    person("Johannes", "Friedrich", role = c("aut"), comment = c(ORCID = "0000-0002-0805-9547")),
    person("Norbert", "Mercier", role = c("aut"), comment = c(ORCID = "0000-0002-6375-9108")),
    person("Rachel K.", "Smedley", role = c("ctb"), comment = c(ORCID = "0000-0001-7773-5193")),
    person("Claire", "Christophe", role = c("ctb")),
    person("Antoine", "Zink", role = c("ctb"), comment = c(ORCID = "0000-0001-7146-1101")),
    person("Julie", "Durcan", role = c("ctb"), comment = c(ORCID = "0000-0001-8724-8022")),
    person("Georgina E.", "King", role = c("ctb", "dtc"), comment = c(ORCID = "0000-0003-1059-8192")),
    person("Anne", "Philippe", role = c("aut"), comment = c(ORCID = "0000-0002-5331-5087")),
    person("Guillaume", "Guérin", role = c("ctb"), comment = c(ORCID = "0000-0001-6298-5579")),
    person("Svenja", "Riedesel", role = c("aut"), comment = c(ORCID = "0000-0003-2936-8776")),
    person("Martin", "Autzen", role = c("aut"), comment = c(ORCID = "0000-0001-6249-426X")),
    person("Pierre", "Guibert", role = c("ctb"), comment = c(ORCID = "0000-0001-8969-8684")),
    person("Dirk", "Mittelstrass", role = c("aut"), comment = c(ORCID = "0000-0002-9567-8791")),
    person("Harrison J.", "Gray", role = c("aut"), comment = c(ORCID = "0000-0002-4555-7473")),
    person("Jean-Michel", "Galharret", role = c("aut"), comment = c(ORCID = "0000-0003-2219-8727")),
    person("Marco", "Colombo", role = c("aut"), comment = c(ORCID = "0000-0001-6672-0623")),
    person("Luc", "Steinbuch", role = c("aut"), comment = c(ORCID = "0000-0001-6484-0920")),
    person("Anna-Maartje", "de Boer", role = c("aut"), comment = c(ORCID = "0000-0002-7359-6939")),
    person("Andrzej", "Bluszcz", role = c("aut"), comment = c(ORCID = "0000-0002-8027-6941")),
    person("Markus", "Fuchs", role = c("ths"), comment = c(ORCID = "0000-0003-4669-6528")))
Maintainer: Sebastian Kreutzer <maintainer_luminescence@r-luminescence.org>
Description: A collection of various R functions for the purpose of Luminescence
    dating data analysis. This includes, amongst others, data import, export,
    application of age models, curve deconvolution, sequence analysis and
    plotting of equivalent dose distributions.
Contact: Package Developers <developers@r-luminescence.org>
License: GPL-3
URL: https://r-lum.github.io/Luminescence/
BugReports: https://github.com/R-Lum/Luminescence/issues
Depends: R (>= 4.4), utils
LinkingTo: Rcpp (>= 1.1.0)
Imports: bbmle (>= 1.0.25.1), data.table (>= 1.18.2.1), DEoptim (>=
        2.2-8), httr (>= 1.4.7), interp (>= 1.1-6), lamW (>= 2.2.3),
        matrixStats (>= 1.3.0), methods, minpack.lm (>= 1.2-4), mclust
        (>= 6.1), Rcpp (>= 1.0.12), shape (>= 1.4.6), parallel, XML (>=
        3.99-0.16),
Suggests: spelling (>= 2.3.0), plotly (>= 4.10.4), rmarkdown (>= 2.27),
        rstudioapi (>= 0.16.0), rjags (>= 4-15), coda (>= 0.19-4),
        knitr, pander (>= 0.6.5), testthat (>= 3.2.1), vdiffr (>=
        1.0.0), tiff (>= 0.1-12), devtools (>= 2.4.5), R.rsp (>=
        0.46.0)
VignetteBuilder: R.rsp, knitr
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
NeedsCompilation: yes
Packaged: 2026-03-25 08:55:44 UTC; marco
Author: Sebastian Kreutzer [aut, trl, cre, dtc] (ORCID:
    <https://orcid.org/0000-0002-0734-2199>),
  Christoph Burow [aut, trl, dtc] (ORCID:
    <https://orcid.org/0000-0002-5023-4046>),
  Michael Dietze [aut] (ORCID: <https://orcid.org/0000-0001-6063-1726>),
  Margret C. Fuchs [aut] (ORCID: <https://orcid.org/0000-0001-7210-1132>),
  Christoph Schmidt [aut] (ORCID:
    <https://orcid.org/0000-0002-2309-3209>),
  Manfred Fischer [aut, trl],
  Johannes Friedrich [aut] (ORCID:
    <https://orcid.org/0000-0002-0805-9547>),
  Norbert Mercier [aut] (ORCID: <https://orcid.org/0000-0002-6375-9108>),
  Rachel K. Smedley [ctb] (ORCID:
    <https://orcid.org/0000-0001-7773-5193>),
  Claire Christophe [ctb],
  Antoine Zink [ctb] (ORCID: <https://orcid.org/0000-0001-7146-1101>),
  Julie Durcan [ctb] (ORCID: <https://orcid.org/0000-0001-8724-8022>),
  Georgina E. King [ctb, dtc] (ORCID:
    <https://orcid.org/0000-0003-1059-8192>),
  Anne Philippe [aut] (ORCID: <https://orcid.org/0000-0002-5331-5087>),
  Guillaume Guérin [ctb] (ORCID: <https://orcid.org/0000-0001-6298-5579>),
  Svenja Riedesel [aut] (ORCID: <https://orcid.org/0000-0003-2936-8776>),
  Martin Autzen [aut] (ORCID: <https://orcid.org/0000-0001-6249-426X>),
  Pierre Guibert [ctb] (ORCID: <https://orcid.org/0000-0001-8969-8684>),
  Dirk Mittelstrass [aut] (ORCID:
    <https://orcid.org/0000-0002-9567-8791>),
  Harrison J. Gray [aut] (ORCID: <https://orcid.org/0000-0002-4555-7473>),
  Jean-Michel Galharret [aut] (ORCID:
    <https://orcid.org/0000-0003-2219-8727>),
  Marco Colombo [aut] (ORCID: <https://orcid.org/0000-0001-6672-0623>),
  Luc Steinbuch [aut] (ORCID: <https://orcid.org/0000-0001-6484-0920>),
  Anna-Maartje de Boer [aut] (ORCID:
    <https://orcid.org/0000-0002-7359-6939>),
  Andrzej Bluszcz [aut] (ORCID: <https://orcid.org/0000-0002-8027-6941>),
  Markus Fuchs [ths] (ORCID: <https://orcid.org/0000-0003-4669-6528>)
Repository: CRAN
Date/Publication: 2026-03-25 10:40:08 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 05:57:37 UTC; unix
Archs: Luminescence.so.dSYM

Package: Luminescence
Type: Package
Title: Comprehensive Luminescence Dating Data Analysis
Version: 1.2.1
Date: 2026-03-25
Authors@R: c(
    person("Sebastian", "Kreutzer", role = c("aut", "trl", "cre", "dtc"), email = "maintainer_luminescence@r-luminescence.org", comment = c(ORCID = "0000-0002-0734-2199")),
    person("Christoph", "Burow", role = c("aut", "trl", "dtc"), comment = c(ORCID = "0000-0002-5023-4046")),
    person("Michael", "Dietze", role = c("aut"), comment = c(ORCID = "0000-0001-6063-1726")),
    person("Margret C.", "Fuchs", role = c("aut"), comment = c(ORCID = "0000-0001-7210-1132")),
    person("Christoph", "Schmidt", role = c("aut"), comment = c(ORCID = "0000-0002-2309-3209")),
    person("Manfred", "Fischer", role = c("aut", "trl")),
    person("Johannes", "Friedrich", role = c("aut"), comment = c(ORCID = "0000-0002-0805-9547")),
    person("Norbert", "Mercier", role = c("aut"), comment = c(ORCID = "0000-0002-6375-9108")),
    person("Rachel K.", "Smedley", role = c("ctb"), comment = c(ORCID = "0000-0001-7773-5193")),
    person("Claire", "Christophe", role = c("ctb")),
    person("Antoine", "Zink", role = c("ctb"), comment = c(ORCID = "0000-0001-7146-1101")),
    person("Julie", "Durcan", role = c("ctb"), comment = c(ORCID = "0000-0001-8724-8022")),
    person("Georgina E.", "King", role = c("ctb", "dtc"), comment = c(ORCID = "0000-0003-1059-8192")),
    person("Anne", "Philippe", role = c("aut"), comment = c(ORCID = "0000-0002-5331-5087")),
    person("Guillaume", "Guérin", role = c("ctb"), comment = c(ORCID = "0000-0001-6298-5579")),
    person("Svenja", "Riedesel", role = c("aut"), comment = c(ORCID = "0000-0003-2936-8776")),
    person("Martin", "Autzen", role = c("aut"), comment = c(ORCID = "0000-0001-6249-426X")),
    person("Pierre", "Guibert", role = c("ctb"), comment = c(ORCID = "0000-0001-8969-8684")),
    person("Dirk", "Mittelstrass", role = c("aut"), comment = c(ORCID = "0000-0002-9567-8791")),
    person("Harrison J.", "Gray", role = c("aut"), comment = c(ORCID = "0000-0002-4555-7473")),
    person("Jean-Michel", "Galharret", role = c("aut"), comment = c(ORCID = "0000-0003-2219-8727")),
    person("Marco", "Colombo", role = c("aut"), comment = c(ORCID = "0000-0001-6672-0623")),
    person("Luc", "Steinbuch", role = c("aut"), comment = c(ORCID = "0000-0001-6484-0920")),
    person("Anna-Maartje", "de Boer", role = c("aut"), comment = c(ORCID = "0000-0002-7359-6939")),
    person("Andrzej", "Bluszcz", role = c("aut"), comment = c(ORCID = "0000-0002-8027-6941")),
    person("Markus", "Fuchs", role = c("ths"), comment = c(ORCID = "0000-0003-4669-6528")))
Maintainer: Sebastian Kreutzer <maintainer_luminescence@r-luminescence.org>
Description: A collection of various R functions for the purpose of Luminescence
    dating data analysis. This includes, amongst others, data import, export,
    application of age models, curve deconvolution, sequence analysis and
    plotting of equivalent dose distributions.
Contact: Package Developers <developers@r-luminescence.org>
License: GPL-3
URL: https://r-lum.github.io/Luminescence/
BugReports: https://github.com/R-Lum/Luminescence/issues
Depends: R (>= 4.4), utils
LinkingTo: Rcpp (>= 1.1.0)
Imports: bbmle (>= 1.0.25.1), data.table (>= 1.18.2.1), DEoptim (>=
        2.2-8), httr (>= 1.4.7), interp (>= 1.1-6), lamW (>= 2.2.3),
        matrixStats (>= 1.3.0), methods, minpack.lm (>= 1.2-4), mclust
        (>= 6.1), Rcpp (>= 1.0.12), shape (>= 1.4.6), parallel, XML (>=
        3.99-0.16),
Suggests: spelling (>= 2.3.0), plotly (>= 4.10.4), rmarkdown (>= 2.27),
        rstudioapi (>= 0.16.0), rjags (>= 4-15), coda (>= 0.19-4),
        knitr, pander (>= 0.6.5), testthat (>= 3.2.1), vdiffr (>=
        1.0.0), tiff (>= 0.1-12), devtools (>= 2.4.5), R.rsp (>=
        0.46.0)
VignetteBuilder: R.rsp, knitr
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
NeedsCompilation: yes
Packaged: 2026-03-25 08:55:44 UTC; marco
Author: Sebastian Kreutzer [aut, trl, cre, dtc] (ORCID:
    <https://orcid.org/0000-0002-0734-2199>),
  Christoph Burow [aut, trl, dtc] (ORCID:
    <https://orcid.org/0000-0002-5023-4046>),
  Michael Dietze [aut] (ORCID: <https://orcid.org/0000-0001-6063-1726>),
  Margret C. Fuchs [aut] (ORCID: <https://orcid.org/0000-0001-7210-1132>),
  Christoph Schmidt [aut] (ORCID:
    <https://orcid.org/0000-0002-2309-3209>),
  Manfred Fischer [aut, trl],
  Johannes Friedrich [aut] (ORCID:
    <https://orcid.org/0000-0002-0805-9547>),
  Norbert Mercier [aut] (ORCID: <https://orcid.org/0000-0002-6375-9108>),
  Rachel K. Smedley [ctb] (ORCID:
    <https://orcid.org/0000-0001-7773-5193>),
  Claire Christophe [ctb],
  Antoine Zink [ctb] (ORCID: <https://orcid.org/0000-0001-7146-1101>),
  Julie Durcan [ctb] (ORCID: <https://orcid.org/0000-0001-8724-8022>),
  Georgina E. King [ctb, dtc] (ORCID:
    <https://orcid.org/0000-0003-1059-8192>),
  Anne Philippe [aut] (ORCID: <https://orcid.org/0000-0002-5331-5087>),
  Guillaume Guérin [ctb] (ORCID: <https://orcid.org/0000-0001-6298-5579>),
  Svenja Riedesel [aut] (ORCID: <https://orcid.org/0000-0003-2936-8776>),
  Martin Autzen [aut] (ORCID: <https://orcid.org/0000-0001-6249-426X>),
  Pierre Guibert [ctb] (ORCID: <https://orcid.org/0000-0001-8969-8684>),
  Dirk Mittelstrass [aut] (ORCID:
    <https://orcid.org/0000-0002-9567-8791>),
  Harrison J. Gray [aut] (ORCID: <https://orcid.org/0000-0002-4555-7473>),
  Jean-Michel Galharret [aut] (ORCID:
    <https://orcid.org/0000-0003-2219-8727>),
  Marco Colombo [aut] (ORCID: <https://orcid.org/0000-0001-6672-0623>),
  Luc Steinbuch [aut] (ORCID: <https://orcid.org/0000-0001-6484-0920>),
  Anna-Maartje de Boer [aut] (ORCID:
    <https://orcid.org/0000-0002-7359-6939>),
  Andrzej Bluszcz [aut] (ORCID: <https://orcid.org/0000-0002-8027-6941>),
  Markus Fuchs [ths] (ORCID: <https://orcid.org/0000-0003-4669-6528>)
Repository: CRAN
Date/Publication: 2026-03-25 10:40:08 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",RLum.Analysis)
S3method("$",RLum.Data.Curve)
S3method("$",RLum.Results)
S3method("$<-",DRAC.list)
S3method("*",RLum.Data.Curve)
S3method("*",RLum.Data.Spectrum)
S3method("+",RLum.Data.Curve)
S3method("+",RLum.Data.Spectrum)
S3method("-",RLum.Data.Curve)
S3method("-",RLum.Data.Spectrum)
S3method("/",RLum.Data.Curve)
S3method("/",RLum.Data.Spectrum)
S3method("[",RLum.Analysis)
S3method("[",RLum.Data.Curve)
S3method("[",RLum.Data.Image)
S3method("[",RLum.Data.Spectrum)
S3method("[",RLum.Results)
S3method("[<-",DRAC.list)
S3method("[<-",RLum.Data.Curve)
S3method("[[",RLum.Analysis)
S3method("[[",RLum.Results)
S3method("[[<-",DRAC.list)
S3method(as.data.frame,DRAC.list)
S3method(as.data.frame,RLum.Data.Curve)
S3method(as.data.frame,RLum.Data.Spectrum)
S3method(as.data.frame,Risoe.BINfileData)
S3method(as.list,RLum.Analysis)
S3method(as.list,RLum.Data.Curve)
S3method(as.list,RLum.Data.Image)
S3method(as.list,RLum.Data.Spectrum)
S3method(as.list,RLum.Results)
S3method(as.matrix,RLum.Data.Curve)
S3method(as.matrix,RLum.Data.Image)
S3method(as.matrix,RLum.Data.Spectrum)
S3method(bin,RLum.Data.Curve)
S3method(bin,RLum.Data.Spectrum)
S3method(dim,RLum.Data.Curve)
S3method(dim,RLum.Data.Spectrum)
S3method(hist,RLum.Analysis)
S3method(hist,RLum.Data.Curve)
S3method(hist,RLum.Data.Image)
S3method(hist,RLum.Results)
S3method(length,RLum.Analysis)
S3method(length,RLum.Data.Curve)
S3method(length,RLum.Results)
S3method(length,Risoe.BINfileData)
S3method(merge,RLum)
S3method(names,RLum.Analysis)
S3method(names,RLum.Data.Curve)
S3method(names,RLum.Data.Image)
S3method(names,RLum.Data.Spectrum)
S3method(names,RLum.Results)
S3method(names,Risoe.BINfileData)
S3method(plot,RLum.Analysis)
S3method(plot,RLum.Data.Curve)
S3method(plot,RLum.Data.Image)
S3method(plot,RLum.Data.Spectrum)
S3method(plot,RLum.Results)
S3method(plot,Risoe.BINfileData)
S3method(plot,list)
S3method(print,DRAC.highlights)
S3method(print,DRAC.list)
S3method(rep,RLum)
S3method(row.names,RLum.Data.Spectrum)
S3method(subset,RLum.Analysis)
S3method(subset,Risoe.BINfileData)
S3method(summary,RLum.Analysis)
S3method(summary,RLum.Data.Curve)
S3method(summary,RLum.Data.Image)
S3method(summary,RLum.Results)
S3method(unlist,RLum.Analysis)
export("add_metadata<-")
export("rename_metadata<-")
export("replace_metadata<-")
export(.as.latex.table)
export(Risoe.BINfileData2RLum.Analysis)
export(analyse_Al2O3C_CrossTalk)
export(analyse_Al2O3C_ITC)
export(analyse_Al2O3C_Measurement)
export(analyse_FadingMeasurement)
export(analyse_IRSAR.RF)
export(analyse_SAR.CWOSL)
export(analyse_SAR.TL)
export(analyse_baSAR)
export(analyse_pIRIRSequence)
export(analyse_portableOSL)
export(apply_CosmicRayRemoval)
export(apply_Crosstalk)
export(apply_EfficiencyCorrection)
export(bin)
export(bin_RLum.Data)
export(calc_AliquotSize)
export(calc_AverageDose)
export(calc_CentralDose)
export(calc_CobbleDoseRate)
export(calc_CommonDose)
export(calc_CosmicDoseRate)
export(calc_EED_Model)
export(calc_FadingCorr)
export(calc_FastRatio)
export(calc_FiniteMixture)
export(calc_FuchsLang2001)
export(calc_HomogeneityTest)
export(calc_Huntley2006)
export(calc_IEU)
export(calc_Lamothe2003)
export(calc_MaxDose)
export(calc_MinDose)
export(calc_MoransI)
export(calc_OSLLxTxDecomposed)
export(calc_OSLLxTxRatio)
export(calc_SourceDoseRate)
export(calc_Statistics)
export(calc_TLLxTxRatio)
export(calc_ThermalLifetime)
export(calc_WodaFuchs2008)
export(calc_gSGC)
export(calc_gSGC_feldspar)
export(combine_De_Dr)
export(convert_Activity2Concentration)
export(convert_BIN2CSV)
export(convert_CW2pHMi)
export(convert_CW2pLM)
export(convert_CW2pLMi)
export(convert_CW2pPMi)
export(convert_Concentration2DoseRate)
export(convert_Daybreak2CSV)
export(convert_PSL2CSV)
export(convert_PSL2Risoe.BINfileData)
export(convert_RLum2Risoe.BINfileData)
export(convert_SG2MG)
export(convert_Second2Gray)
export(convert_Wavelength2Energy)
export(convert_XSYG2CSV)
export(correct_PMTLinearity)
export(extract_IrradiationTimes)
export(extract_ROI)
export(fit_CWCurve)
export(fit_DoseResponseCurve)
export(fit_EmissionSpectra)
export(fit_LMCurve)
export(fit_OSLLifeTimes)
export(fit_SurfaceExposure)
export(fit_ThermalQuenching)
export(get_Layout)
export(get_RLum)
export(get_rightAnswer)
export(import_Data)
export(length_RLum)
export(melt_RLum)
export(merge_RLum)
export(merge_RLum.Analysis)
export(merge_RLum.Data.Curve)
export(merge_RLum.Data.Spectrum)
export(merge_RLum.Results)
export(merge_Risoe.BINfileData)
export(names_RLum)
export(normalise_RLum)
export(plot_AbanicoPlot)
export(plot_DRCSummary)
export(plot_DRTResults)
export(plot_DetPlot)
export(plot_DoseResponseCurve)
export(plot_FilterCombinations)
export(plot_GrowthCurve)
export(plot_Histogram)
export(plot_KDE)
export(plot_MoranScatterplot)
export(plot_NRt)
export(plot_OSLAgeSummary)
export(plot_RLum)
export(plot_RLum.Analysis)
export(plot_RLum.Data.Curve)
export(plot_RLum.Data.Image)
export(plot_RLum.Data.Spectrum)
export(plot_RLum.Results)
export(plot_ROI)
export(plot_RadialPlot)
export(plot_Risoe.BINfileData)
export(plot_SingleGrainDisc)
export(plot_ViolinPlot)
export(read_BIN2R)
export(read_BINXLOG2R)
export(read_Daybreak2R)
export(read_HeliosOSL2R)
export(read_PSL2R)
export(read_RF2R)
export(read_SPE2R)
export(read_TIFF2R)
export(read_XSYG2R)
export(remove_RLum)
export(remove_SignalBackground)
export(replicate_RLum)
export(report_RLum)
export(sTeve)
export(scale_GammaDose)
export(set_RLum)
export(set_Risoe.BINfileData)
export(smooth_RLum)
export(sort_RLum)
export(structure_RLum)
export(subset_SingleGrainData)
export(template_DRAC)
export(trim_RLum.Data)
export(tune_Data)
export(use_DRAC)
export(verify_SingleGrainData)
export(view)
export(write_R2BIN)
export(write_R2TIFF)
export(write_RLum2CSV)
exportClasses(RLum)
exportClasses(RLum.Analysis)
exportClasses(RLum.Data)
exportClasses(RLum.Data.Curve)
exportClasses(RLum.Data.Image)
exportClasses(RLum.Data.Spectrum)
exportClasses(RLum.Results)
exportClasses(Risoe.BINfileData)
exportMethods("add_metadata<-")
exportMethods("rename_metadata<-")
exportMethods("replace_metadata<-")
exportMethods(bin_RLum.Data)
exportMethods(get_RLum)
exportMethods(length_RLum)
exportMethods(melt_RLum)
exportMethods(names_RLum)
exportMethods(normalise_RLum)
exportMethods(remove_RLum)
exportMethods(replicate_RLum)
exportMethods(set_RLum)
exportMethods(set_Risoe.BINfileData)
exportMethods(show)
exportMethods(smooth_RLum)
exportMethods(sort_RLum)
exportMethods(structure_RLum)
exportMethods(view)
import(data.table)
import(methods)
import(utils)
importFrom(Rcpp,evalCpp)
importFrom(grDevices,adjustcolor)
importFrom(grDevices,rgb)
importFrom(graphics,abline)
importFrom(graphics,axTicks)
importFrom(graphics,axis)
importFrom(graphics,curve)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,segments)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(stats,approx)
importFrom(stats,coef)
importFrom(stats,density)
importFrom(stats,dnorm)
importFrom(stats,median)
importFrom(stats,na.exclude)
importFrom(stats,nls)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,uniroot)
useDynLib(Luminescence, .registration = TRUE)

<!-- README.md was auto-generated by README.Rmd. Please DO NOT edit by hand!-->

# Luminescence <img width=120px src="man/figures/Luminescence_logo.png" align="right" />

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN](https://www.r-pkg.org/badges/version/Luminescence)](https://CRAN.R-project.org/package=Luminescence)
[![R-CMD-check](https://github.com/R-Lum/Luminescence/actions/workflows/R-CMD-check.yml/badge.svg)](https://github.com/R-Lum/Luminescence/actions)
[![Coverage
Status](https://img.shields.io/codecov/c/github/R-Lum/Luminescence.svg)](https://app.codecov.io/github/R-Lum/Luminescence?branch=master)

[![CRAN
DOI](https://img.shields.io/badge/DOI-10.32614/CRAN.package.Luminescence-1f57b6?style=flat&link=https://doi.org/10.32614/CRAN.package.Luminescence)](https://doi.org/10.32614/CRAN.package.Luminescence)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.596252.svg)](https://doi.org/10.5281/zenodo.596252)

[![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/Luminescence)](https://www.r-pkg.org/pkg/Luminescence)
[![Downloads](https://cranlogs.r-pkg.org/badges/Luminescence)](https://www.r-pkg.org/pkg/Luminescence)
[![Downloads](https://cranlogs.r-pkg.org/badges/last-week/Luminescence)](https://www.r-pkg.org/pkg/Luminescence)
[![Downloads](https://cranlogs.r-pkg.org/badges/last-day/Luminescence)](https://www.r-pkg.org/pkg/Luminescence)

The R package `'Luminescence'` by the R-Luminescence Group provides a
collection of various R functions for luminescence dating data analysis.

## Social media and other resources

Visit our [R-Luminescence homepage](https://r-luminescence.org). News
and updates on `'Luminescence'` are also available from the [REPLAY
website](https://replay.geog.uni-heidelberg.de/).

## Installation

#### i. Requirements

- *Windows (32/64bit)*:
  [Rtools](https://cran.r-project.org/bin/windows/Rtools/) (provided by
  CRAN)
- *macOS*: [Xcode](https://developer.apple.com/) (provided by Apple)
- *Linux*: [gcc](https://gcc.gnu.org) often comes pre-installed in most
  distributions.

#### ii. Install the package

Install any development versions using our [RStudio](https://posit.co)
add-in

![](man/figures/README-Screenshot_AddIn.png)

Alternatively, to install the latest development builds directly from
GitHub, run:

``` r
if (!require("pak"))
  install.packages("pak")
pak::pkg_install("R-Lum/Luminescence")
```

Please note that the version on the GitHub `master` branch is a
development version and may change day by day. To install the stable
version from CRAN, simply run the following from an R console:

``` r
install.packages("Luminescence")
```

For stable branches please visit the [Luminescence package page on
CRAN](https://CRAN.R-project.org/package=Luminescence).

## Contribute

The R-Luminescence project is based on and evolves from ideas,
contributions and constructive criticism of its users. Help us maintain
and improve the package by reporting bugs, requesting features,
suggesting design improvements or contributing new functions.

Open an [issue](https://github.com/R-Lum/Luminescence/issues) or write
us an [e-mail](mailto:developers@r-luminescence.org) if anything crosses
your mind or if you want your new self-written function to be to
included in the package.

You are kindly invited to bring forward the package with us!

## License

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or any later
version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the [GNU
General Public
License](https://github.com/R-Lum/Luminescence/blob/master/LICENSE) for
more details.

## Funding

- 2011-2013: The initial version of the package was developed in the
  framework of the PhD thesis by Sebastian Kreutzer, while he was funded
  through the DFG programme “Rekonstruktion der Umweltbedingungen des
  Spätpleistozäns in Mittelsachsen anhand von Löss-Paläobodensequenzen”
  ([GEPRIS id: 46526743](https://gepris.dfg.de/gepris/projekt/46526743))

- 2014-2018: Cooperation and personal exchange between the developers is
  gratefully funded by the DFG in the framework of the program
  “Scientific Networks”. Project title: “RLum.Network: Ein
  Wissenschaftsnetzwerk zur Analyse von Lumineszenzdaten mit R” ([GEPRIS
  id: 250974974](https://gepris.dfg.de/gepris/projekt/250974974))

- 05/2014-12/2019: The work of Sebastian Kreutzer as maintainer of the
  package was supported by LabEx LaScArBx (ANR - n. ANR-10-LABX-52).

- 01/2020-04/2022: Sebastian Kreutzer as maintainer of the package has
  received funding from the European Union’s Horizon 2020 research and
  innovation programme under the Marie Skłodowska-Curie grant agreement
  [No 844457 (CREDit)](https://cordis.europa.eu/project/id/844457), and
  could continue maintaining the package.

- since 03/2023: Sebastian Kreutzer as maintainer of the package
  receives funding from the DFG Heisenberg programme [No
  505822867](https://gepris.dfg.de/gepris/projekt/505822867).

- since 08/2024 the future and sustainable development of
  `'Luminescence'` towards better reproducibility and usability is
  supported through the DFG programme “REPLAY: REProducible Luminescence
  Data AnalYses” [No
  528704761](https://gepris.dfg.de/gepris/projekt/528704761?language=en)
  led by Dr Sebastian Kreutzer (PI at Heidelberg University, DE (until
  2025), LIAG - Institute for Applied Geophysics, Hannover, DE (from
  2026)) and Dr Thomas Kolb (PI at Justus-Liebig-University Giessen,
  DE).

- All other authors gratefully received additional funding from various
  public funding bodies.

## Related projects

- [RLumModel](https://github.com/R-Lum/RLumModel)
- [RLumShiny](https://github.com/R-Lum/RLumShiny)
- [BayLum](https://github.com/crp2a/BayLum)
- [RCarb](https://github.com/R-Lum/RCarb)
- [RLumCarlo](https://github.com/R-Lum/RLumCarlo)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----init---------------------------------------------------------------------
library(Luminescence)

## ----simulate_signal, fig.dim = c(8, 6), out.width="800px"--------------------

vn_simulated <- sample(x = c(rnorm(n = 30, mean = 2000, sd = 500),
                             rnorm(n = 70, mean = 20, sd = 1)),
                       size = 100)

vn_simulated <- round(vn_simulated) # Because photons are discrete

head(vn_simulated, n = 25)

hist(vn_simulated,
     main = "Simulated signal (histogram)",
     xlab = "Photon counts",
     ylab = "Frequency",
     breaks = 30
     )

## ----show_disc_basic, fig.dim = c(7, 7), out.width="800px"--------------------

par(mar = c(1, 4, 6, 4))

plot_SingleGrainDisc(object = vn_simulated,
          main = "Simulated signal (measurement disc)"
          )


## ----calculate_moransI--------------------------------------------------------

calc_MoransI(object = vn_simulated)

calc_MoransI(object = vn_simulated, compute_pseudo_p = TRUE)



## ----add_crosstalk, fig.dim = c(8, 6), out.width="800px"----------------------

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                              n_crosstalk = 0.10)

vn_simulated_with_crosstalk <- round(vn_simulated_with_crosstalk)

hist(vn_simulated_with_crosstalk,
     main = "Simulated signal with crosstalk (histogram)",
     xlab = "Photon counts",
     ylab = "Frequency",
     breaks = 30
     )

plot_SingleGrainDisc(object = vn_simulated_with_crosstalk,
          main = "Simulated signal with crosstalk (measurement disc)")

calc_MoransI(object = vn_simulated_with_crosstalk)

calc_MoransI(object = vn_simulated_with_crosstalk, compute_pseudo_p = TRUE)



## ----test_different_amounts, fig.dim = c(8, 6), out.width="800px"-------------


df_MoransI <- data.frame(crosstalk = seq(from = 0,
                    to = 0.30,
                    length.out=50))

df_MoransI$MoransI <- NA
df_MoransI$pseudo_p <- NA

old.opts <- options(warn = -1) # silence warnings from compute_pseudo_p
for (i in 1:nrow(df_MoransI))
{

  vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                              n_crosstalk = df_MoransI$crosstalk[i])

  df_MoransI$MoransI[i]  <- calc_MoransI(object = vn_simulated_with_crosstalk)
  df_MoransI$pseudo_p[i] <-
    calc_MoransI(object = vn_simulated_with_crosstalk, compute_pseudo_p = TRUE)
}
options(old.opts) # restore the default options

n_expected_I_no_spatial_autocorr <- calc_MoransI(1:100,
                                                 spatial_autocorrelation = FALSE)

##

plot(x = df_MoransI$crosstalk,
     y = df_MoransI$MoransI,
     ylim = range(
          pretty(x = c(df_MoransI$MoransI, n_expected_I_no_spatial_autocorr))
          ),
     ## Set ylim manually to make sure the value for I for no crosstalk is visible
     xlab = "Amount of added crosstalk",
     ylab = "Calculated Moran's I"
     )
graphics::grid()
abline(h = n_expected_I_no_spatial_autocorr,
       col = "purple")

legend(x = "topleft",
       legend = "Expected I if no spatial autocorrelation",
       lty = "solid",
       col = "purple",
       cex = 0.8)

plot(x = df_MoransI$crosstalk,
     y = df_MoransI$pseudo_p,
     xlab = "Amount of added crosstalk",
     ylab = "Generated pseudo-p of related Moran's I")
graphics::grid()

## ----show_`disc_extended_moran_scatterplot, fig.dim = c(8, 5), out.width="800px"----


plot_MoranScatterplot(object = vn_simulated,
           main = "Moran scatterplot, simulated signal without crosstalk")

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                              n_crosstalk = 0.25)
vn_simulated_with_crosstalk <- round(vn_simulated_with_crosstalk)

plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
           main = "Moran scatterplot, simulated signal with added crosstalk")

## ----adjacent_grain_locations, fig.dim = c(8, 5), out.width="800px"-----------

    vn_simulated_with_holes <- c(rnorm(30, mean = 10, sd = 5),
                              rnorm(30, mean = 500, sd = 50),
                              rep(NA, times = 40)
                              )

df_Neighbours <- Luminescence:::.get_Neighbours(object = vn_simulated_with_holes)

head(df_Neighbours)

## ----adjacent_grain_locations_show_disc, fig.dim = c(8, 5), out.width="800px"----

    plot_SingleGrainDisc(object = vn_simulated_with_holes,
              show_neighbours = TRUE,
              show_location_ids = TRUE)

## ----adjacent_grain_locations_add, fig.dim = c(8, 5), out.width="800px"-------

df_Neighbours_with_diag <- Luminescence:::.get_Neighbours(object = vn_simulated)


for (i in c(1:9, 11:19, 21:29) )
{
  df_Neighbours_with_diag <- rbind(df_Neighbours_with_diag,
                                            c(i, i+11, 1/sqrt(2))
                                          )
}

tail(df_Neighbours_with_diag)


plot_SingleGrainDisc(object = vn_simulated,
          df_neighbours = df_Neighbours_with_diag,
          show_neighbours = TRUE,
          show_location_ids = TRUE)

## ----restrict_to_inner_8x8, fig.dim = c(8, 5), out.width="800px"--------------

vn_values_to_show <-
      sample(x = c(rnorm(n = 30, mean = 2000, sd = 500),
                   rnorm(n = 70, mean = 20, sd = 1)),
             size = 100)

## Set the outer rows to NA before adding crosstalk
vn_disc_border_locations <- c(1:10,
                              91:100,
                              seq(from = 11, to = 81, by = 10),
                              seq(from = 20, to = 90, by = 10)
                             )
vn_values_to_show[vn_disc_border_locations] <- NA

vn_values_to_show <- apply_Crosstalk(object = vn_values_to_show,
                                     n_crosstalk = 0.15)

calc_MoransI(object = vn_values_to_show)

plot_SingleGrainDisc(object = vn_values_to_show,
                     show_neighbours = TRUE,
                     ignore_borders = TRUE)

calc_MoransI(object = vn_values_to_show,
             ignore_borders = TRUE)


## ----show_disc_extended, fig.dim = c(8, 5), out.width="800px"-----------------

plot_SingleGrainDisc(object = vn_simulated,
          main = "",
          legend = TRUE,
          show_coordinates = TRUE,
          show_location_ids = TRUE,
          show_positioning_holes = FALSE)


## ----show_disc_scales, fig.dim = c(8, 5), out.width="800px"-------------------

plot_SingleGrainDisc(object = vn_simulated,
          main = "Linear scale",
          legend = TRUE,
          show_coordinates = FALSE,
          show_location_ids = FALSE,
          show_positioning_holes = TRUE,
          str_transform = "lin")


plot_SingleGrainDisc(object = vn_simulated,
          main = "Logarithmic scale",
          legend = TRUE,
          show_coordinates = FALSE,
          show_location_ids = FALSE,
          show_positioning_holes = TRUE,
          str_transform = "log")


## ----show_scatterplot_options, fig.dim = c(8, 5), out.width="800px"-----------


vn_simulated <- c(rnorm(75, mean = 10, sd = 5),
                  rnorm(25, mean = 500, sd = 50)  )
vn_simulated <- sample(size = 100, vn_simulated)

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                               n_crosstalk = 0.15)

## Base use
plot_MoranScatterplot(object = vn_simulated,
                      main = "Without crosstalk")


plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
                      main = "With crosstalk")


## Layout options
plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
           pch = "show_location_ids",
           legend = FALSE,
           log = "xy",
           main = "With location ID's; and with log scales"
           )

plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
           pch = "show_n_neighbours",
           legend = FALSE,
           str_y_def = "weighted_sum",
           main = "With number of neighbours, and other y calculation"
           )

## ----moransI_intermediate_numbers---------------------------------------------


calc_MoransI(object = 1:100,
             return_intermediate_values = TRUE)


## ----moransI_intermediate_numbers_extended------------------------------------

vn_simulated <- sample(x = c(rnorm(n = 30, mean = 2000, sd = 500),
                             rnorm(n = 70, mean = 20, sd = 1)),
                       size = 100)
vn_simulated <- round(vn_simulated)

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                               n_crosstalk = 0.20)
vn_simulated_with_crosstalk <- round(vn_simulated_with_crosstalk)


df_compare <-
  data.frame(Case = c("Without crosstalk", "With crosstalk"),
             MoransI =  c(calc_MoransI(object = vn_simulated),
                          calc_MoransI(object = vn_simulated_with_crosstalk)
                          ),
             PopulationVar =
               c(calc_MoransI(object = vn_simulated,
                 return_intermediate_values = TRUE)$n_population_variance,
                 calc_MoransI(object = vn_simulated_with_crosstalk,
                 return_intermediate_values = TRUE)$n_population_variance),
             SpatialAutoCor =
               c(calc_MoransI(object = vn_simulated,
                 return_intermediate_values = TRUE)$n_average_auto_correlation,
                 calc_MoransI(object = vn_simulated_with_crosstalk,
                 return_intermediate_values = TRUE)$n_average_auto_correlation)
      )

df_compare[,2] <-  round(df_compare[,2],2)

(df_compare)

#' @title Comprehensive Luminescence Dating Data Analysis\cr
#' \if{html}{
#' \figure{Luminescence_logo.png}{options: width="75" alt="r-luminescence.org"}
#' }
#'
#' @description A collection of various R functions for the purpose of luminescence dating
#' data analysis. This includes, amongst others, data import, export,
#' application of age models, curve deconvolution, sequence analysis and
#' plotting of equivalent dose distributions.
#'
#' @name Luminescence-package
#' @aliases Luminescence-package Luminescence
#'
#' @details
#'
#' **Supervisor of the initial version in 2012**
#'
#' Markus Fuchs, Justus-Liebig-University Giessen, Germany
#'
#' **Support contact**
#'
#' * \email{developers@@r-luminescence.org}
#' * [https://github.com/R-Lum/Luminescence/discussions]()
#'
#' **Bug reporting**
#'
#' * \email{developers@@r-luminescence.org} or
#' * [https://github.com/R-Lum/Luminescence/issues]()
#'
#' **Project website**
#'
#' * [https://r-luminescence.org]()
#'
#' **Project source code repository**
#'
#' * [https://github.com/R-Lum/Luminescence]()
#'
#' **Related package projects**
#'
#' * [https://cran.r-project.org/package=RLumShiny]()
#' * [https://cran.r-project.org/package=RLumModel]()
#' * [https://cran.r-project.org/package=RLumCarlo]()
#' * [https://cran.r-project.org/package=RCarb]()
#'
#' **Funding**
#'
#' * 2011-2013:  The initial version of the package was developed, while Sebastian Kreutzer
#' was funded through the DFG programme "Rekonstruktion der Umweltbedingungen
#' des Spätpleistozäns in Mittelsachsen anhand von Löss-Paläobodensequenzen"
#' (DFG id: 46526743)
#'
#' * 2014-2018: Cooperation and personal exchange between the developers is gratefully
#' funded by the DFG (SCHM 3051/3-1) in the framework of the program
#' "Scientific Networks". Project title: "RLum.Network: Ein
#' Wissenschaftsnetzwerk zur Analyse von Lumineszenzdaten mit R" (2014-2018)
#'
#' * 05/2014-12/2019: The work of Sebastian Kreutzer as maintainer of the package was supported
#' by LabEx LaScArBx (ANR - n. ANR-10-LABX-52).
#'
#' * 01/2020-04/2022: Sebastian Kreutzer as maintainer of the package has received funding
#' from the European Union’s Horizon 2020 research and innovation programme under
#' the Marie Skłodowska-Curie grant agreement No 844457 (CREDit), and could continue
#' maintaining the package.
#'
#' * since 03/2023: Sebastian Kreutzer as maintainer of the package receives funding from the
#' DFG Heisenberg programme No 505822867.
#'
#' * since 08/2024: The future and sustainable development of 'Luminescence'
#' towards better reproducibility and usability is supported through the DFG
#' programme "REPLAY: REProducible Luminescence Data AnalYses" No 528704761,
#' led by Dr Sebastian Kreutzer (PI at LIAG - Institute for Applied Geophysics, Hannover, DE) and Dr Thomas
#' Kolb (PI at Justus-Liebig-University Giessen, DE).
#'
#' * All other authors gratefully received additional funding from various public funding bodies.
#'
#' @references
#' Dietze, M., Kreutzer, S., Fuchs, M.C., Burow, C., Fischer, M.,
#' Schmidt, C., 2013. A practical guide to the R package Luminescence.
#' Ancient TL 31 (1), 11-18.
#'
#' Dietze, M., Kreutzer, S., Burow, C., Fuchs, M.C., Fischer, M., Schmidt, C., 2016. The abanico plot:
#' visualising chronometric data with individual standard errors.
#' Quaternary Geochronology 31, 1-7. \doi{10.1016/j.quageo.2015.09.003}
#'
#' Fuchs, M.C., Kreutzer, S., Burow, C., Dietze, M., Fischer, M., Schmidt, C.,
#' Fuchs, M., 2015. Data processing in luminescence dating analysis: An
#' exemplary workflow using the R package 'Luminescence'. Quaternary
#' International 362, 8-13. \doi{10.1016/j.quaint.2014.06.034}
#'
#' Kreutzer, S., Schmidt, C., Fuchs, M.C., Dietze, M., Fischer, M., Fuchs, M.,
#' 2012. Introducing an R package for luminescence dating analysis. Ancient TL
#' 30 (1), 1-8.
#'
#' Mercier, N., Kreutzer, S., Christophe, C., Guérin, G., Guibert, P., Lahaye, C., Lanos, P., Philippe, A.,
#' Tribolo, C., 2016. Bayesian statistics in luminescence dating: The 'baSAR'-model and its
#' implementation in the R package 'Luminescence'. Ancient TL 34 (2), 14-21.
#'
#' Mercier, N., Galharret, J.-M., Tribolo, C., Kreutzer, S., Philippe, A., 2022.
#' Luminescence age calculation through Bayesian convolution of equivalent dose
#' and dose-rate distributions: the De_Dr model.
#' Geochronology 4, 297–310. \doi{10.5194/gchron-4-297-2022}
#'
#' Smedley, R.K., 2015. A new R function for the Internal External Uncertainty (IEU) model.
#' Ancient TL 33 (1), 16-21.
#'
#' King, E.G., Burow, C., Roberts, H., Pearce, N.J.G., 2018. Age determination
#' using feldspar: evaluating fading-correction model performance.
#' Radiation Measurements 119, 58-73. \doi{10.1016/j.radmeas.2018.07.013}
#'
#' @keywords package
#'
#' @import utils methods data.table
#'
#' @importFrom  Rcpp evalCpp
#' @importFrom graphics plot abline mtext text lines par lines axTicks axis curve hist legend points polygon segments title
#' @importFrom grDevices adjustcolor rgb
#' @importFrom stats approx density dnorm median na.exclude nls quantile rnorm runif sd uniroot setNames coef predict residuals
#'
#' @useDynLib Luminescence, .registration = TRUE
#'
"_PACKAGE"

#' @title Base datasets
#'
#' @description
#' Collection of datasets with published and unpublished data used within the
#' package.
#'
#' \describe{
#' \item{BaseData.ConversionFactors}{Collection of published dose-rate
#' conversion factors to convert concentrations of radioactive isotopes to
#' dose rate values}
#' \item{BaseData.GrainSizeAttenuation}{Grain size attenuation data by Guérin
#' et al. (2012)}
#' \item{BaseData.FractionalGammaDose}{Collection of (un-)published fractional
#' gamma dose-rate values to scale the gamma-dose rate considering layer-to-layer
#' variations in soil radioactivity}
#' }
#'
#' @section Version: 0.2.0
#'
#' @keywords datasets
#'
#' @aliases
#' BaseDataSet.ConversionFactors
#' BaseDataSet.GrainSizeAttenuation
#' BaseDataSet.FractionalGammaDose
#'
#' @name BaseDataSet
NULL

#' @format
#' **Dose-rate conversion factors**
#'
#' A [`list`] with three elements with dose-rate conversion factors
#' sorted by article and radiation type (alpha, beta, gamma):
#'
#' \tabular{ll}{
#'
#' `AdamiecAitken1998`: \tab
#' Conversion factors from Tables 5 and 6 \cr
#'
#' `Cresswelletal2018`: \tab
#' Conversion factors from Tables 5 and 6 \cr
#'
#' `Guerinetal2011`: \tab
#' Conversion factors from Tables 1, 2 and 3 \cr
#'
#' `Liritzisetal2013`: \tab
#' Conversion factors from Tables 1, 2 and 3 \cr
#' }
#'
#' @source
#' **Dose-rate conversion factors**
#'
#' All gamma conversion factors were carefully read from the tables given in:
#'
#' Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update.
#' Ancient TL 16, 37-46.
#'
#' Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018.
#' Dose rate conversion parameters: Assessment of nuclear data.
#' Radiation Measurements 120, 195-201.
#'
#' Guérin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion
#' factors: update. Ancient TL 29, 5-8.
#'
#' Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
#' A re-evaluation of radiation dose-rate conversion factors. Mediterranean
#' Archaeology and Archaeometry 13, 1-15.
#'
#' @examples
#' ## conversion factors
#' data("BaseDataSet.ConversionFactors", envir = environment())
#'
#' @name BaseDataSet
NULL

#' @format
#' **Grain size attenuation data**
#'
#' A [`data.frame`] seven columns and sixteen rows. Column headers
#' are `GrainSize`, `Q_K`, `FS_K`, `Q_Th`, `FS_Th`, `Q_U`, `FS_U`.
#' Grain sizes are quoted in µm (e.g., 20, 40, 60 etc.)
#'
#' @source
#' **Grain size attenuation data**
#'
#' Guérin, G., Mercier, N., Nathan, R., Adamiec, G., Lefrais, Y., 2012.
#' On the use of the infinite matrix assumption and associated concepts:
#' A critical review. Radiation Measurements 47, 778-785.
#'
#' @examples
#' ## grain size attenuation
#' data("BaseDataSet.GrainSizeAttenuation", envir = environment())
#'
#' @name BaseDataSet
NULL

#' @format
#' **Fractional gamma dose-rate values**
#'
#' A [`list`] with fractional gamma dose-rate values sorted by article:
#'
#' \tabular{ll}{
#'
#' `Aitken1985`: \tab
#' Fractional gamma-dose values from table H.1
#' }
#'
#' @source
#' **Fractional gamma dose-rate values**
#'
#' Fractional gamma dose values were carefully read from the tables given in:
#'
#' Aitken, M.J., 1985. Thermoluminescence Dating. Academic Press, London.
#'
#' @examples
#' ## fractional gamma dose
#' data("BaseDataSet.FractionalGammaDose", envir = environment())
#'
#' @name BaseDataSet
NULL

#' Example data for scale_GammaDose()
#'
#' An example data set for the function `scale_GammaDose()` containing layer
#' specific information to scale the gamma dose rate considering variations in
#' soil radioactivity.
#'
#' @format
#'
#' A [`data.frame`]. Please see `?scale_GammaDose()` for a detailed description
#' of its structure.
#'
#' @section Version: 0.1
#'
#' @keywords datasets
#'
#' @examples
#'
#' ## Load data
#' data("ExampleData.ScaleGammaDose", envir = environment())
#'
#' @name ExampleData.ScaleGammaDose
NULL

#' Example data for calc_CobbleDoseRate()
#'
#' An example data set for the function [Luminescence::calc_CobbleDoseRate] containing layer
#' specific information for the cobble to be used in the function.
#'
#' @format
#'
#' A [`data.frame`]. Please see [Luminescence::calc_CobbleDoseRate] for detailed information
#' on the structure of the [data.frame].
#'
#' @section Version: 0.1.0
#'
#' @keywords datasets
#'
#' @examples
#'
#' ## Load data
#' data("ExampleData.CobbleData", envir = environment())
#'
#' @name ExampleData.CobbleData
NULL

#' Base data set for cosmic dose rate calculation
#'
#' Collection of data from various sources needed for cosmic dose rate
#' calculation
#'
#' @format
#'
#' \tabular{ll}{
#'
#' `values.cosmic.Softcomp`: \tab
#' data frame containing cosmic dose rates
#' for shallow depths (< 167 g cm^-2) obtained using the "AGE" program by
#' Rainer Gruen (cf. Gruen 2009). These data essentially reproduce the graph
#' shown in Fig. 1 of Prescott & Hutton (1988). \cr
#'
#' `values.factor.Altitude`: \tab
#' data frame containing altitude factors
#' for adjusting geomagnetic field-change factors. Values were read from Fig. 1
#' in Prescott & Hutton (1994). \cr
#'
#' `values.par.FJH`: \tab
#' data frame containing values for parameters F, J
#' and H (read from Fig. 2 in Prescott & Hutton 1994) used in the expression \cr
#' }
#'
#' \deqn{Dc = D0*(F + J \exp((altitude/1000)/H))}
#'
#' @section Version: 0.1
#'
#' @references
#' Gruen, R., 2009. The "AGE" program for the calculation of luminescence age estimates.
#' Ancient TL 27, pp. 45-46.
#'
#' Prescott, J.R., Hutton, J.T., 1988. Cosmic ray and gamma ray dosimetry for
#' TL and ESR. Nuclear Tracks and Radiation Measurements 14, pp. 223-227.
#'
#' Prescott, J.R., Hutton, J.T., 1994. Cosmic ray contributions to dose rates
#' for luminescence and ESR dating: large depths and long-term time variations.
#' Radiation Measurements 23, pp. 497-500.
#'
#' @source
#' The following data were carefully read from figures in mentioned
#' sources and used for fitting procedures. The derived expressions are used in
#' the function `calc_CosmicDoseRate`.
#'
#' **values.cosmic.Softcomp**
#'
#' \tabular{ll}{
#' Program: \tab "AGE"\cr
#' Reference: \tab Gruen (2009) \cr
#' Fit: \tab Polynomials in the form of
#' }
#'
#' For depths between 40-167 g cm^-2:
#'
#' \deqn{y = 2*10^-6*x^2-0.0008*x+0.2535}
#'
#' (For depths <40 g cm^-2)
#'
#' \deqn{y = -6*10^-8*x^3+2*10^-5*x^2-0.0025*x+0.2969}
#'
#' **`values.factor.Altitude`**
#'
#' \tabular{ll}{
#' Reference: \tab Prescott & Hutton (1994) \cr
#' Page: \tab 499 \cr
#' Figure: \tab 1 \cr
#' Fit: \tab 2-degree polynomial in the form of
#' }
#'
#' \deqn{y = -0.026*x^2 + 0.6628*x + 1.0435}
#'
#'
#' **`values.par.FJH`**
#'
#' \tabular{ll}{
#' Reference: \tab Prescott & Hutton (1994) \cr
#' Page: \tab 500 \cr
#' Figure: \tab 2 \cr
#' Fits: \tab 3-degree polynomials and linear fits
#' }
#'
#' F (non-linear part, \eqn{\lambda} < 36.5 deg.):
#'
#' \deqn{y = -7*10^-7*x^3-8*10^-5*x^2-0.0009*x+0.3988}
#'
#' F (linear part, \eqn{\lambda} > 36.5 deg.):
#'
#' \deqn{y = -0.0001*x + 0.2347}
#'
#' J (non-linear part, \eqn{\lambda} < 34 deg.):
#'
#' \deqn{y = 5*10^-6*x^3-5*10^-5*x^2+0.0026*x+0.5177}
#'
#' J (linear part, \eqn{\lambda} > 34 deg.):
#'
#' \deqn{y = 0.0005*x + 0.7388}
#'
#' H (non-linear part, \eqn{\lambda} < 36 deg.):
#'
#' \deqn{y = -3*10^-6*x^3-5*10^-5*x^2-0.0031*x+4.398}
#'
#' H (linear part, \eqn{\lambda} > 36 deg.):
#'
#' \deqn{y = 0.0002*x + 4.0914}
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##load data
#' data(BaseDataSet.CosmicDoseRate)
#'
#' @docType data
#' @name BaseDataSet.CosmicDoseRate
#' @aliases values.cosmic.Softcomp values.factor.Altitude values.par.FJH
NULL

#' @title All examples
#' @description
#' These examples are fully documented:
#'
#' \itemize{
#' \item [Luminescence::ExampleData.CobbleData]
#' \item [Luminescence::ExampleData.DeValues]
#' \item [Luminescence::ExampleData.Fading]
#' \item [Luminescence::ExampleData.portableOSL]
#' \item [Luminescence::ExampleData.RLum.Data.Image]
#' \item [Luminescence::ExampleData.ScaleGammaDose]
#' \item [Luminescence::ExampleData.SurfaceExposure]
#' \item [Luminescence::ExampleData.TR_OSL]
#' }
#'
#' The following ones are also available, but are visible only via the
#' internal documentation, as they correspond to unpublished or synthetic data:
#'
#' \itemize{
#' \item [Luminescence::ExampleData.Al2O3C]
#' \item [Luminescence::ExampleData.BINfileData]
#' \item [Luminescence::ExampleData.CW_OSL_Curve]
#' \item [Luminescence::ExampleData.FittingLM]
#' \item [Luminescence::ExampleData.LxTxData]
#' \item [Luminescence::ExampleData.LxTxOSLData]
#' \item [Luminescence::ExampleData.MortarData]
#' \item [Luminescence::ExampleData.RF70Curves]
#' \item [Luminescence::ExampleData.RLum.Analysis]
#' \item [Luminescence::ExampleData.XSYG]
#' }
#'
#' @keywords datasets
#'
#' @docType data
#' @name ExampleData
NULL

#' @title Example data from a SAR OSL and SAR TL measurement for the package
#' Luminescence
#'
#' @description Example data from a SAR OSL and TL measurement for package Luminescence
#' directly extracted from a Risoe BIN-file and provided in an object of type
#' [Luminescence::Risoe.BINfileData-class]
#'
#' @format
#'
#' `CWOSL.SAR.Data`: SAR OSL measurement data
#'
#' `TL.SAR.Data`: SAR TL measurement data
#'
#' Each class object contains two slots: (a) `METADATA` is a [data.frame] with
#' all metadata stored in the BIN file of the measurements and (b) `DATA`
#' contains a list of vectors of the measured data (usually count values).
#'
#' @section Version: 0.1
#'
#' @references
#' **CWOSL.SAR.Data**: unpublished data
#'
#' **TL.SAR.Data**: unpublished data
#'
#' @source **CWOSL.SAR.Data**
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Bayreuth \cr
#' Lab-Code: \tab BT607 \cr
#' Location: \tab Saxony/Germany \cr
#' Material: \tab Middle grain quartz measured on aluminium cups on a Risø TL/OSL DA-15 reader\cr
#' Reference: \tab unpublished
#' }
#'
#' **TL.SAR.Data**
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory of Cologne\cr
#' Lab-Code: \tab LP1_5\cr
#' Location: \tab Spain\cr
#' Material: \tab Flint \cr
#' Setup: \tab Risoe TL/OSL DA-20 reader (Filter: Semrock Brightline, HC475/50, N2, unpolished steel discs) \cr
#' Reference: \tab unpublished \cr
#' Remarks: \tab dataset limited to one position
#' }
#'
#' @note
#' Please note that this example data cannot be exported to a BIN-file using the function
#' `writeR2BIN` as it was generated and implemented in the package long time ago. In the meantime
#' the BIN-file format changed.
#'
#' @docType data
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ## show first 5 elements of the METADATA and DATA elements in the terminal
#' data(ExampleData.BINfileData, envir = environment())
#' CWOSL.SAR.Data@@METADATA[1:5,]
#' CWOSL.SAR.Data@@DATA[1:5]
#'
#' @name ExampleData.BINfileData
#' @aliases CWOSL.SAR.Data TL.SAR.Data
NULL

#' Example CW-OSL curve data for the package Luminescence
#'
#' `data.frame` containing CW-OSL curve data (time, counts)
#'
#' @docType data
#'
#' @format Data frame with 1000 observations on the following 2 variables:
#'
#' - `list("x")`: a numeric vector, time
#' - `list("y")`: a numeric vector, counts
#'
#' @references
#' Baartman, J.E.M., Veldkamp, A., Schoorl, J.M., Wallinga, J.,
#' Cammeraat, L.H., 2011. Unravelling Late Pleistocene and Holocene landscape
#' dynamics: The Upper Guadalentin Basin, SE Spain. Geomorphology 125,
#' 172-185.
#'
#' Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL signal
#' components. Radiation Measurements 47, 752-758.
#'
#' @source **ExampleData.CW_OSL_Curve**
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Bayreuth\cr
#' Lab-Code: \tab BT607\cr
#' Location: \tab Saxony/Germany\cr
#' Material: \tab Middle grain quartz measured on aluminium cups on a Risø TL/OSL DA-15 reader.\cr
#' Reference: \tab unpublished data }
#'
#' **CW_Curve.BosWallinga2012**
#'
#' \tabular{ll}{
#' Lab: \tab Netherlands Centre for Luminescence Dating (NCL)\cr
#' Lab-Code: \tab NCL-2108077\cr
#' Location: \tab Guadalentin Basin, Spain\cr
#' Material: \tab Coarse grain quartz\cr
#' Reference: \tab Bos & Wallinga (2012) and Baartman et al. (2011)
#' }
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#' plot(ExampleData.CW_OSL_Curve)
#'
#' @name ExampleData.CW_OSL_Curve
#' @aliases CW_Curve.BosWallinga2012 ExampleData.CW_OSL_Curve
NULL

#' Example portable OSL curve data for the package Luminescence
#'
#' A `list` of [Luminescence::RLum.Analysis-class] objects, each containing
#' the same number of [Luminescence::RLum.Data.Curve-class] objects representing
#' individual OSL, IRSL and dark count measurements of a sample.
#'
#' @name ExampleData.portableOSL
#'
#' @docType data
#'
#' @source
#'
#' **ExampleData.portableOSL**
#'
#' \tabular{ll}{
#' Lab: \tab Cologne Luminescence Laboratory\cr
#' Lab-Code: \tab `none` \cr
#' Location: \tab Nievenheim/Germany\cr
#' Material: \tab Fine grain quartz \cr
#' Reference: \tab unpublished data
#' }
#'
#' @keywords datasets
#'
#' @examples
#'
#' data(ExampleData.portableOSL, envir = environment())
#' plot_RLum(ExampleData.portableOSL)
#'
NULL

#' Example data for fit_LMCurve() in the package Luminescence
#'
#' Linearly modulated (LM) measurement data from a quartz sample from Norway
#' including background measurement. Measurements carried out in the
#' luminescence laboratory at the University of Bayreuth.
#'
#' @format Two objects (data.frames) with two columns (time and counts).
#'
#' @references
#' Fuchs, M., Kreutzer, S., Fischer, M., Sauer, D., Soerensen, R., 2012. OSL and IRSL
#' dating of raised beach sand deposits along the south-eastern coast of Norway.
#' Quaternary Geochronology 10, 195-200.
#'
#' @source
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Bayreuth\cr
#' Lab-Code: \tab BT900\cr
#' Location: \tab Norway\cr
#' Material: \tab Beach deposit, coarse grain quartz measured on aluminium discs on a Risø TL/OSL DA-15 reader\cr
#' }
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##show LM data
#' data(ExampleData.FittingLM, envir = environment())
#' plot(values.curve,log="x")
#'
#' @name ExampleData.FittingLM
#' @aliases values.curve values.curveBG
NULL

#' Example Lx/Tx data from CW-OSL SAR measurement
#'
#' LxTx data from a SAR measurement for the package Luminescence.
#'
#' @format A [`data.frame`] with 4 columns (Dose, LxTx, LxTx.Error, TnTx).
#'
#' @references unpublished data
#'
#' @source
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Bayreuth\cr
#' Lab-Code: \tab BT607\cr
#' Location: \tab Ostrau (Saxony-Anhalt/Germany)\cr
#' Material: \tab Middle grain (38-63 \eqn{\mu}m) quartz measured on a Risoe TL/OSL DA-15 reader.
#' }
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ## plot Lx/Tx data vs dose [s]
#' data(ExampleData.LxTxData, envir = environment())
#' plot(LxTxData$Dose,LxTxData$LxTx)
#'
#' @name ExampleData.LxTxData
#' @aliases LxTxData
NULL

#' Example Lx and Tx curve data from an artificial OSL measurement
#'
#' `Lx` and `Tx` data of continuous wave (CW-) OSL signal curves.
#'
#' @format Two [`data.frame`]s containing time and count values.
#'
#' @references unpublished data
#'
#' @source
#' Arbitrary OSL measurement.
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.LxTxOSLData, envir = environment())
#'
#' ##plot data
#' plot(Lx.data)
#' plot(Tx.data)
#'
#' @name ExampleData.LxTxOSLData
#' @aliases Lx.data Tx.data
NULL

#' Example equivalent dose data from mortar samples
#'
#' Arbitrary data to test the function `calc_EED_Model`
#'
#' @format Two [`data.frame`]s containing De and De error
#'
#' @references unpublished data
#'
#' @source
#' Arbitrary measurements.
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.MortarData, envir = environment())
#'
#' ##plot data
#' plot(MortarData)
#'
#' @name ExampleData.MortarData
#' @aliases MortarData
NULL

#' Example data as [Luminescence::RLum.Analysis-class] objects
#'
#' Collection of different [Luminescence::RLum.Analysis-class] objects for
#' protocol analysis.
#'
#' @format
#'
#' `IRSAR.RF.Data`: IRSAR.RF.Data on coarse grain feldspar
#'
#' Each object contains data needed for the given protocol analysis.
#'
#' @section Version: 0.1
#'
#' @references
#' **IRSAR.RF.Data**
#'
#' Kreutzer, S., Lauer, T., Meszner, S., Krbetschek, M.R., Faust, D., Fuchs,
#' M., 2014. Chronology of the Quaternary profile Zeuchfeld in Saxony-Anhalt /
#' Germany - a preliminary luminescence dating study. Zeitschrift fuer
#' Geomorphologie 58, 5-26. doi: 10.1127/0372-8854/2012/S-00112
#'
#' @source **IRSAR.RF.Data**
#'
#' These data were kindly provided by Tobias Lauer and Matthias Krbetschek.
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory TU Bergakademie Freiberg\cr
#' Lab-Code: \tab ZEU/SA1\cr
#' Location: \tab Zeuchfeld (Zeuchfeld Sandur; Saxony-Anhalt/Germany)\cr
#' Material: \tab K-feldspar (130-200 \eqn{\mu}m)\cr
#' Reference: \tab Kreutzer et al. (2014)
#' }
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.RLum.Analysis, envir = environment())
#'
#' ##plot data
#' plot_RLum(IRSAR.RF.Data)
#'
#' @name ExampleData.RLum.Analysis
#' @aliases IRSAR.RF.Data
NULL

#' RF70 Example data as [Luminescence::RLum.Analysis-class] objects
#'
#' Collection two curves from Kreutzer et al. (2018) measured using the RF70 protocol
#' (Frouin et al., 2017) as [Luminescence::RLum.Analysis-class] objects.
#'
#' @format
#'
#' `BDX16648`: [Luminescence::RLum.Analysis-class] with two IR-RF curves as [Luminescence::RLum.Data.Curve-class] objects
#'
#' @section Version: 0.1.0
#'
#' @references
#'
#' Frouin, M., Huot, S., Kreutzer, S., Lahaye, C., Lamothe, M., Philippe, A., Mercier, N., 2017. An improved
#' radiofluorescence single-aliquot regenerative dose protocol for K-feldspars. Quaternary Geochronology 38, 13–24.
#' \doi{10.1016/j.quageo.2016.11.004}
#'
#' Kreutzer, S., Duval, M., Bartz, M., Bertran, P., Bosq, M., Eynaud, F., Verdin, F., Mercier, N., 2018.
#' Deciphering long-term coastal dynamics using IR-RF and ESR dating: A case study from Médoc,
#' south-West France. Quaternary Geochronology 48, 108–120. \doi{10.1016/j.quageo.2018.09.005}
#'
#' @source **RF70Curves**
#'
#' \tabular{ll}{
#' Lab: \tab IRAMAT-CRP2A, CNRS-Université Bordeaux Montaigne, Bordeaux, France\cr
#' Lab-Code: \tab BDX16648\cr
#' Location: \tab Médoc, France (details see reference)\cr
#' Material: \tab K-feldspar (100-200 \eqn{\mu}m)\cr
#' Reference: \tab Kreutzer et al. (2018)
#' }
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.RF70Curves, envir = environment())
#'
#' ##plot data
#' plot_RLum(RF70Curves)
#'
#' @name ExampleData.RF70Curves
#' @aliases RF70Curves
NULL

#' Example data as [Luminescence::RLum.Data.Image-class] objects
#'
#' Measurement of Princton Instruments camera imported with the function
#' [Luminescence::read_SPE2R] to R to produce an
#' [Luminescence::RLum.Data.Image-class] object.
#'
#' @format Object of class [Luminescence::RLum.Data.Image-class]
#'
#' @section Version: 0.1
#'
#' @source
#' **ExampleData.RLum.Data.Image**
#'
#' These data were kindly provided by Regina DeWitt.
#'
#' \tabular{ll}{
#' Lab.: \tab Department of Physics, East-Carolina University, NC, USA\cr
#' Lab-Code: \tab - \cr
#' Location: \tab - \cr
#' Material: \tab - \cr
#' Reference: \tab - \cr
#' }
#'
#' Image data is a measurement of fluorescent ceiling lights with a cooled
#' Princeton Instruments (TM) camera fitted on Risø DA-20 TL/OSL reader.
#'
#' @keywords datasets
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.RLum.Data.Image, envir = environment())
#'
#' ##plot data
#' plot_RLum(ExampleData.RLum.Data.Image)
#'
#' @name ExampleData.RLum.Data.Image
NULL

#' Example data for a SAR OSL measurement and a TL spectrum using a lexsyg
#' reader
#'
#' Example data from a SAR OSL measurement and a TL spectrum for package
#' Luminescence imported from a Freiberg Instruments XSYG file using the
#' function [Luminescence::read_XSYG2R].
#'
#' @format
#'
#' `OSL.SARMeasurement`: SAR OSL measurement data
#'
#' The data contain two elements: (a) `$Sequence.Header` is a
#' [data.frame] with metadata from the measurement,(b)
#' `Sequence.Object` contains an [Luminescence::RLum.Analysis-class] object
#' for further analysis.
#'
#' `TL.Spectrum`: TL spectrum data
#'
#' [Luminescence::RLum.Data.Spectrum-class] object for further analysis. The
#' spectrum was cleaned from cosmic-rays using the function
#'
#' `apply_CosmicRayRemoval`. Note that no quantum efficiency calibration
#' was performed.
#'
#' @section Version: 0.1
#'
#' @seealso [Luminescence::read_XSYG2R], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Data.Spectrum-class],
#' [Luminescence::plot_RLum], [Luminescence::plot_RLum.Analysis], [Luminescence::plot_RLum.Data.Spectrum]
#'
#' @references
#' Unpublished data measured to serve as example data for that
#' package. Location origin of sample BT753 is given here:
#'
#' Fuchs, M., Kreutzer, S., Rousseau, D.D., Antoine, P., Hatte, C., Lagroix,
#' F., Moine, O., Gauthier, C., Svoboda, J., Lisa, L., 2013. The loess sequence
#' of Dolni Vestonice, Czech Republic: A new OSL-based chronology of the Last
#' Climatic Cycle. Boreas, 42, 664--677.
#'
#' @source
#' **OSL.SARMeasurement**
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Giessen\cr
#' Lab-Code: \tab no code\cr
#' Location: \tab not specified\cr
#' Material: \tab Coarse grain quartz  on steel cups on lexsyg research reader\cr
#' Reference: \tab unpublished
#' }
#'
#' **TL.Spectrum**
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Giessen\cr
#' Lab-Code: \tab BT753\cr
#' Location: \tab Dolni Vestonice/Czech Republic\cr
#' Material: \tab Fine grain polymineral on steel cups on lexsyg research reader\cr
#' Reference: \tab Fuchs et al., 2013 \cr
#' Spectrum: \tab Integration time 19 s, channel time 20 s\cr
#' Heating: \tab 1 K/s, up to 500 deg. C
#' }
#'
#' @keywords datasets internal
#'
#' @examples
#' ##show data
#' data(ExampleData.XSYG, envir = environment())
#'
#' ## =========================================
#' ##(1) OSL.SARMeasurement
#' OSL.SARMeasurement
#'
#' ##show $Sequence.Object
#' OSL.SARMeasurement$Sequence.Object
#'
#' ##grep OSL curves and plot the first curve
#' OSLcurve <- get_RLum(OSL.SARMeasurement$Sequence.Object,
#' recordType="OSL")[[1]]
#' plot_RLum(OSLcurve)
#'
#' ## =========================================
#' ##(2) TL.Spectrum
#' TL.Spectrum
#'
#' ##plot simple spectrum (2D)
#' plot_RLum.Data.Spectrum(TL.Spectrum,
#'                         plot.type="contour",
#'                         xlim = c(310,750),
#'                         ylim = c(0,300),
#'                         bin.rows=10,
#'                         bin.cols = 1)
#'
#' ##plot 3d spectrum (uncomment for usage)
#' # plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="persp",
#' # xlim = c(310,750), ylim = c(0,300), bin.rows=10,
#' # bin.cols = 1)
#'
#' @name ExampleData.XSYG
#' @aliases OSL.SARMeasurement TL.Spectrum
NULL

#' Example De data sets for the package Luminescence
#'
#' Equivalent dose (De) values measured for a fine grain quartz sample from a
#' loess section in Rottewitz (Saxony/Germany) and for a coarse grain quartz
#' sample from a fluvial deposit in the rock shelter of Cueva Anton
#' (Murcia/Spain).
#'
#' @format A [list] with two elements, each containing a two column [data.frame]:
#'
#' \describe{
#' `$BT998`: De and De error values for a fine grain quartz
#' sample from a loess section in Rottewitz.
#'
#' `$CA1`: Single grain De
#' and De error values for a coarse grain quartz sample from a fluvial deposit
#' in the rock shelter of Cueva Anton
#' }
#'
#' @references
#'  **BT998**
#'
#' Unpublished data
#'
#' **CA1**
#'
#' Burow, C., Kehl, M., Hilgers, A., Weniger, G.-C., Angelucci, D., Villaverde,
#' V., Zapata, J. and Zilhao, J.  (2015). Luminescence dating of fluvial
#' deposits in the rock shelter of Cueva Anton, Spain. Geochronometria 52, 107-125.
#'
#' **BT998**
#'
#' \tabular{ll}{
#' Lab: \tab Luminescence Laboratory Bayreuth\cr
#' Lab-Code: \tab BT998\cr
#' Location: \tab Rottewitz (Saxony/Germany)\cr
#' Material: \tab Fine grain quartz measured on aluminium discs on a Risø TL/OSL DA-15 reader\cr
#' Units: \tab Values are given in seconds \cr
#' Dose Rate: \tab Dose rate of the beta-source at measurement ca. 0.0438 Gy/s +/- 0.0019 Gy/s\cr
#' Measurement Date: \tab 2012-01-27
#' }
#'
#' **CA1**
#'
#' \tabular{ll}{
#' Lab: \tab Cologne Luminescence Laboratory (CLL)\cr
#' Lab-Code: \tab C-L2941\cr
#' Location: \tab Cueva Anton (Murcia/Spain)\cr
#' Material: \tab Coarse grain quartz (200-250 microns) measured on single grain discs on a Risoe TL/OSL DA-20 reader\cr
#' Units: \tab Values are given in Gray \cr
#' Measurement Date: \tab 2012
#' }
#'
#' @keywords datasets
#'
#' @examples
#'
#' ##(1) plot values as histogram
#' data(ExampleData.DeValues, envir = environment())
#' plot_Histogram(ExampleData.DeValues$BT998, xlab = "De [s]")
#'
#' ##(2) plot values as histogram (with second to gray conversion)
#' data(ExampleData.DeValues, envir = environment())
#'
#' De.values <- convert_Second2Gray(ExampleData.DeValues$BT998,
#'                          dose.rate = c(0.0438, 0.0019))
#'
#' plot_Histogram(De.values, xlab = "De [Gy]")
#'
#' @name ExampleData.DeValues
NULL

#' Example data for feldspar fading measurements
#'
#' Example data set for fading measurements of the IR50, IR100, IR150 and
#' IR225 feldspar signals of sample UNIL/NB123. It further contains regular equivalent dose
#' measurement data of the same sample, which can be used to apply a
#' fading correction to.
#'
#' @format A [list] with two elements, each containing a further [list] of
#' [data.frame]s containing the data on the fading and equivalent dose measurements:
#'
#' \describe{
#' `$fading.data`: A named [list] of [data.frame]s,
#' each having three named columns (`LxTx, LxTx.error, timeSinceIrradiation`).\cr
#' `..$IR50`: Fading data of the IR50 signal.\cr
#' `..$IR100`: Fading data of the IR100 signal.\cr
#' `..$IR150`: Fading data of the IR150 signal.\cr
#' `..$IR225`: Fading data of the IR225 signal.\cr
#'
#' `$equivalentDose.data`: A named of [data.frame]s,
#' each having three named columns (`dose, LxTx, LxTx.error`).\cr
#' `..$IR50`: Equivalent dose measurement data of the IR50 signal.\cr
#' `..$IR100`: Equivalent dose measurement data of the IR100 signal.\cr
#' `..$IR150`: Equivalent dose measurement data of the IR150 signal.\cr
#' `..$IR225`: Equivalent dose measurement data of the IR225 signal.\cr
#' }
#'
#' @source
#'
#' These data were kindly provided by Georgina E. King. Detailed information
#' on the sample UNIL/NB123 can be found in the reference given below. The raw
#' data can be found in the accompanying supplementary information.
#'
#' @references
#'
#' King, G.E., Herman, F., Lambert, R., Valla, P.G., Guralnik, B., 2016.
#' Multi-OSL-thermochronometry of feldspar. Quaternary Geochronology 33, 76-87.
#' doi:10.1016/j.quageo.2016.01.004
#'
#' **Details**
#'
#' \tabular{ll}{
#' Lab: \tab University of Lausanne \cr
#' Lab-Code: \tab UNIL/NB123 \cr
#' Location: \tab Namche Barwa (eastern Himalayas)\cr
#' Material: \tab Coarse grained (180-212 microns) potassium feldspar \cr
#' Units: \tab Values are given in seconds \cr
#' Lab Dose Rate: \tab Dose rate of the beta-source at measurement ca. 0.1335 +/- 0.004 Gy/s \cr
#' Environmental Dose Rate: \tab 7.00 +/- 0.92 Gy/ka (includes internal dose rate)
#' }
#'
#' @keywords datasets
#'
#' @examples
#'
#' ## Load example data
#' data("ExampleData.Fading", envir = environment())
#'
#' ## Get fading measurement data of the IR50 signal
#' IR50_fading <- ExampleData.Fading$fading.data$IR50
#' head(IR50_fading)
#'
#' ## Determine g-value and rho' for the IR50 signal
#' IR50_fading.res <- analyse_FadingMeasurement(IR50_fading)
#'
#' ## Show g-value and rho' results
#' gval <- get_RLum(IR50_fading.res)
#' rhop <- get_RLum(IR50_fading.res, "rho_prime")
#'
#' gval
#' rhop
#'
#' ## Get LxTx values of the IR50 DE measurement
#' IR50_De.LxTx <- ExampleData.Fading$equivalentDose.data$IR50
#'
#' ## Calculate the De of the IR50 signal
#' IR50_De <- fit_DoseResponseCurve(IR50_De.LxTx,
#'                                 mode = "interpolation",
#'                                 fit.method = "EXP")
#'
#' ## Extract the calculated De and its error
#' IR50_De.res <- get_RLum(IR50_De)
#' De <- c(IR50_De.res$De, IR50_De.res$De.Error)
#'
#' ## Apply fading correction (age conversion greatly simplified)
#' IR50_Age <- De / 7.00
#' IR50_Age.corr <- calc_FadingCorr(IR50_Age, g_value = IR50_fading.res)
#'
#' @name ExampleData.Fading
NULL

#' Example OSL surface exposure dating data
#'
#' A set of synthetic OSL surface exposure dating data to demonstrate the
#' [Luminescence::fit_SurfaceExposure] functionality. See examples to reproduce the data
#' interactively.
#'
#' @details
#'
#' **`$sample_1`**
#'
#' \tabular{ccc}{
#' **mu** \tab **`sigmaphi`** \tab **age** \cr
#'   0.9 \tab 5e-10 \tab 10000 \cr
#' }
#'
#' **`$sample_2`**
#'
#' \tabular{ccccc}{
#' **mu** \tab **`sigmaphi`** \tab **age** \tab **Dose rate** \tab **D0** \cr
#'   0.9 \tab 5e-10 \tab 10000 \tab 2.5 \tab 40 \cr
#' }
#'
#' **`$set_1`**
#'
#' \tabular{ccc}{
#' **mu** \tab **`sigmaphi`** \tab **ages** \cr
#'   0.9 \tab 5e-10 \tab 1e3, 1e4, 1e5, 1e6 \cr
#' }
#'
#' **`$set_2`**
#'
#' \tabular{ccccc}{
#' **mu** \tab **`sigmaphi`** \tab **ages** \tab **Dose rate** \tab **D0** \cr
#'   0.9 \tab 5e-10 \tab 1e2, 1e3, 1e4, 1e5, 1e6 \tab 1.0 \tab 40 \cr
#' }
#'
#' @format A [list] with 4 elements:
#'
#' \tabular{ll}{
#' **Element** \tab **Content** \cr
#' `$sample_1` \tab A [data.frame] with 3 columns (depth, intensity, error) \cr
#' `$sample_2` \tab A [data.frame] with 3 columns (depth, intensity, error) \cr
#' `$set_1` \tab A [list] of 4 [data.frame]s, each representing a sample with different ages \cr
#' `$set_2` \tab A [list] of 5 [data.frame]s, each representing a sample with different ages \cr
#' }
#'
#' @references Unpublished synthetic data
#'
#' @source
#'
#' See examples for the code used to create the data sets.
#'
#' @examples
#'
#' ## ExampleData.SurfaceExposure$sample_1
#' sigmaphi <- 5e-10
#' age <- 10000
#' mu <- 0.9
#' x <- seq(0, 10, 0.1)
#' fun <- exp(-sigmaphi * age * 365.25*24*3600 * exp(-mu * x))
#'
#' set.seed(666)
#' synth_1 <- data.frame(depth = x,
#'                       intensity = jitter(fun, 1, 0.1),
#'                       error = runif(length(x), 0.01, 0.2))
#'
#' ## VALIDATE sample_1
#' fit_SurfaceExposure(synth_1, mu = mu, sigmaphi = sigmaphi)
#'
#'
#' ## ExampleData.SurfaceExposure$sample_2
#' sigmaphi <- 5e-10
#' age <- 10000
#' mu <- 0.9
#' x <- seq(0, 10, 0.1)
#' Ddot <- 2.5   / 1000 / 365.25 / 24 / 60 / 60 # 2.5 Gy/ka in Seconds
#' D0 <- 40
#' fun <- (sigmaphi * exp(-mu * x) *
#'           exp(-(age * 365.25*24*3600) *
#'                 (sigmaphi * exp(-mu * x) + Ddot/D0)) + Ddot/D0) /
#'   (sigmaphi * exp(-mu * x) + Ddot/D0)
#'
#' set.seed(666)
#' synth_2 <- data.frame(depth = x,
#'                       intensity = jitter(fun, 1, 0.1),
#'                       error = runif(length(x), 0.01, 0.2))
#'
#' ## VALIDATE sample_2
#' fit_SurfaceExposure(synth_2, mu = mu, sigmaphi = sigmaphi, Ddot = 2.5, D0 = D0)
#'
#'
#' ## ExampleData.SurfaceExposure$set_1
#' sigmaphi <- 5e-10
#' mu <- 0.9
#' x <- seq(0, 15, 0.2)
#' age <- c(1e3, 1e4, 1e5, 1e6)
#' set.seed(666)
#'
#' synth_3 <- vector("list", length = length(age))
#'
#' for (i in 1:length(age)) {
#'   fun <- exp(-sigmaphi * age[i] * 365.25*24*3600 * exp(-mu * x))
#'   synth_3[[i]] <- data.frame(depth = x,
#'                              intensity = jitter(fun, 1, 0.05))
#' }
#'
#'
#' ## VALIDATE set_1
#' fit_SurfaceExposure(synth_3, age = age, sigmaphi = sigmaphi)
#'
#'
#' ## ExampleData.SurfaceExposure$set_2
#' sigmaphi <- 5e-10
#' mu <- 0.9
#' x <- seq(0, 15, 0.2)
#' age <- c(1e2, 1e3, 1e4, 1e5, 1e6)
#' Ddot <- 1.0 / 1000 / 365.25 / 24 / 60 / 60 # 2.0 Gy/ka in Seconds
#' D0 <- 40
#' set.seed(666)
#'
#' synth_4 <- vector("list", length = length(age))
#'
#' for (i in 1:length(age)) {
#'   fun <- (sigmaphi * exp(-mu * x) *
#'             exp(-(age[i] * 365.25*24*3600) *
#'                   (sigmaphi * exp(-mu * x) + Ddot/D0)) + Ddot/D0) /
#'     (sigmaphi * exp(-mu * x) + Ddot/D0)
#'
#'   synth_4[[i]] <- data.frame(depth = x,
#'                              intensity = jitter(fun, 1, 0.05))
#' }
#'
#'
#' ## VALIDATE set_2
#' fit_SurfaceExposure(synth_4, age = age, sigmaphi = sigmaphi, D0 = D0, Ddot = 1.0)
#'
#' \dontrun{
#' ExampleData.SurfaceExposure <- list(
#'   sample_1 = synth_1,
#'   sample_2 = synth_2,
#'   set_1 = synth_3,
#'   set_2 = synth_4
#' )
#' }
#'
#' @keywords datasets
#' @name ExampleData.SurfaceExposure
NULL

#' Example Al2O3:C Measurement Data
#'
#' Measurement data obtained from measuring Al2O3:C chips at the IRAMAT-CRP2A, Université Bordeaux
#' Montaigne in 2017 on a Freiberg Instruments lexsyg SMART reader.
#' The example data used in particular to allow test of the functions
#' developed in framework of the work by Kreutzer et al., 2018.
#'
#' @format Two datasets comprising [Luminescence::RLum.Analysis-class] data imported using the function
#' [Luminescence::read_XSYG2R]
#'
#' \describe{
#' `data_ITC`: Measurement data to determine the irradiation time correction, the data can
#' be analysed with the function [Luminescence::analyse_Al2O3C_ITC]
#'
#' `data_CrossTalk`: Measurement data obtained while estimating the irradiation cross-talk of the
#' reader used for the experiments. The data can be analysed either with the function
#' [Luminescence::analyse_Al2O3C_CrossTalk] or [Luminescence::analyse_Al2O3C_Measurement]
#' }
#'
#' @note From both datasets unneeded curves have been removed and
#' the number of aliquots have been reduced to a required minimum to keep the file size small, but
#' still being able to run the corresponding functions.
#'
#' @seealso [Luminescence::analyse_Al2O3C_ITC], [Luminescence::analyse_Al2O3C_CrossTalk],
#' [Luminescence::analyse_Al2O3C_Measurement]
#'
#' @references Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018.
#' Environmental Dose Rate Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
#' Geochronometria 45, 56–67.
#'
#' @keywords datasets internal
#'
#' @examples
#'
#' ##(1) curves
#' data(ExampleData.Al2O3C, envir = environment())
#' plot_RLum(data_ITC[1:2])
#'
#' @name ExampleData.Al2O3C
#' @aliases data_CrossTalk data_ITC
NULL

#' Example TR-OSL data
#'
#' Single TR-OSL curve obtained by Schmidt et al. (2019) for quartz sample BT729
#' (origin: Trebgast Valley, Germany, quartz, 90-200 µm, unpublished data).
#'
#' @format One [Luminescence::RLum.Data.Curve-class] dataset imported using the function [Luminescence::read_XSYG2R]
#'
#' \describe{
#' `ExampleData.TR_OSL`: A single [Luminescence::RLum.Data.Curve-class] object with the TR-OSL data
#' }
#'
#' @seealso [Luminescence::fit_OSLLifeTimes]
#'
#' @references Schmidt, C., Simmank, O., Kreutzer, S., 2019.
#' Time-Resolved Optically Stimulated Luminescence of Quartz in the Nanosecond Time Domain.
#' Journal of Luminescence 213, 376-387.
#'
#' @keywords datasets
#'
#' @examples
#'
#' ##(1) curves
#' data(ExampleData.TR_OSL, envir = environment())
#' plot_RLum(ExampleData.TR_OSL)
#'
#' @name ExampleData.TR_OSL
NULL

#' Collection of External Data
#'
#' @description Description and listing of data provided in the folder `data/extdata`
#'
#' @details
#' The **R** package `Luminescence` includes a number of raw data files, which are mostly used in
#' the example sections of appropriate functions. They are also used internally for testing corresponding
#' functions using the `testthat` package (see files in `tests/testthat/`) to ensure their operational
#' reliability.
#'
#' **Accessibility**
#'
#' If the **R** package `Luminescence` is installed correctly the preferred way to access and use these
#' data from within **R** is as follows:
#'
#'  `system.file("extdata/<FILENAME>", package = "Luminescence")`
#'
#' **Individual file descriptions**
#'
#' *>>Daybreak_TestFile.DAT.txt<<*
#'
#' **Type:** raw measurement data \cr
#' **Device:** Daybreak OSL/TL reader\cr
#' **Measurement date:** unknown\cr
#' **Location:** unknown\cr
#' **Provided by:** unknown\cr
#' **Related R function(s):** `read_Daybreak2R()`\cr
#' **Reference:** unknown
#'
#' *>>DorNie_0016.psl<<*
#'
#' **Type:** raw measurement data \cr
#' **Device:** SUERC portable OSL reader \cr
#' **Measurement date:** 19/05/2016 \cr
#' **Location:** Dormagen-Nievenheim, Germany \cr
#' **Provided by:** Christoph Burow (University of Cologne) \cr
#' **Related R function(s):** `read_PSL2R()` \cr
#' **Reference:** unpublished \cr
#' **Additional information:** Sample measured at an archaeological site near \cr
#'  Dormagen-Nievenheim (Germany) during a practical course on Luminescence dating in 2016. \cr
#'
#' *>>QNL84_2_bleached.txt*, *QNL84_2_unbleached.txt<<*
#'
#' **Type:** Test data for exponential fits \cr
#' **Reference:** Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46. \doi{10.26034/la.atl.1989.150}\cr
#'
#'
#' *>>STRB87_1_bleached.txt*, *STRB87_1_unbleached.txt<<*
#'
#' **Type:** Test data for exponential fits \cr
#' **Reference:** Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46. \doi{10.26034/la.atl.1989.150}
#'
#' *>>XSYG_file.xsyg*
#'
#' **Type:** XSYG-file stump \cr
#' **Info:** XSYG-file with some basic curves to test functions \cr
#' **Reference:** no reference available
#'
#' @keywords datasets
#' @name extdata
NULL

#' @title Analyse Thermochronometry Data
#'
#' @description Analyse thermochronometry data based on MatLab code written
#' by XXX and XXX and XXX
#'
#' @details ##TODO
#'
#' @param object [character] (**required**):
#' path to a CSV file ##TODO;
#' alternatively a [vector] of paths
#'
#' @param ITL_model [character] (*with default*):
#' type of model to fit, either `"GOK"` (default) or `"BTS"`.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... further parameters passed to [Luminescence::fit_IsothermalHolding]
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Svenja Riedesel, DTU Risø (Denmark)\cr
#'
#' @keywords datagen
#'
#' @return
#' An [Luminescence::RLum.Results-class] object is returned:
#'
#' Slot: **@data**
#'
#' \tabular{lll}{
#'  **OBJECT** \tab **TYPE** \tab **COMMENT**\cr
#' `FAD` \tab `RLum.Results` \tab results of the fading measurement \cr
#' `ITL` \tab `RLum.Results` \tab results of the isothermal holding \cr
#' `DRC` \tab `RLum.Results` \tab results of the dose response curve \cr
#' }
#'
#' Slot: **@info**
#'
#' \tabular{lll}{
#' **OBJECT** \tab `TYPE` \tab `COMMENT`\cr
#' `call` \tab `call` \tab the original function call\cr
#' }
#'
#' @references
#'
#' Bouscary, C., King, G.E, Kranz-Bartz, M., Bernard, M., Biswas, R.H., Bossin,
#' L., Duverger, A., Guralnik, B., Herman, F., Nanni, U., Stalder, N., Valla,
#' P.G., Visnjevic, V., Wen, X., 2025.
#' OSLThermo and ESRThermo: Libraries of code for trapped-charge thermochronometry.
#'
#' @seealso [Luminescence::analyse_FadingMeasurement], [Luminescence::fit_IsothermalHolding],
#' [Luminescence::fit_DoseResponseCurve]
#'
#' @examples
#' # example code ##TODO
#'
#' @noRd
analyse_ThermochronometryData <- function(
  object,
  ITL_model = c("GOK", "BTS"),
  plot = TRUE,
  verbose = TRUE,
  ...
) {
  .set_function_name("analyse_ThermochronometryData")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO --- general
  ## - output of this function follows the suggestions from the MatLab code by Benny Gurlanik; however
  ## the implementation is R based
  ## - Excepted output
  ## - proposed a new format for a CSV file with data ... or even better ... in the long-term
  ## run the entire analysis in R only
  ## - later we have to think about two different modes of analysis
  ## - we want to pass certain parameters over to functions

  ## Integrity checks -------------------------------------------------------
  ## for a start we only allow data in the format proposed by the MatLab script
  .validate_class(object, "character")
  ITL_model <- .validate_args(ITL_model, c("GOK", "BTS"))
  .validate_logical_scalar(plot)
  .validate_logical_scalar(verbose)

  object <- .import_ThermochronometryData(object, output_type = "RLum.Results")
  sample_names <- object@info$sample_names

  ## prepare plot and reset to default on exit
  if(plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)
    par(mfrow = c(1,3))
  }

  ## Reminder: We have n samples in one Excel sheet ... each set will be analysed
  ## separately
  results_combined <- lapply(seq_along(sample_names), function(i) {
    # (1) Fading data ------------------------------------------------------------
    ## get fading object for this particular sample
    df_FAD <- object@data$FAD[object@data$FAD$SAMPLE == sample_names[i],
                              c("LxTx", "LxTx_ERROR", "TIME")]

    ## set NA values in LxTx error to 0
    df_FAD$LxTx_ERROR[is.na(df_FAD$LxTx_ERROR)] <- 0

    results_FAD <- analyse_FadingMeasurement(
      object = df_FAD,
      verbose = FALSE,
      plot = plot,
      plot_singlePanels = 3,
      plot.trend = FALSE)

    # (2) ITL Data ------------------------------------------------------------
    df_ITL <- object@data$ITL[object@data$ITL$SAMPLE == sample_names[i],]
    results_ITL <- fit_IsothermalHolding(
      data = df_ITL,
      ITL_model = ITL_model,
      rhop = results_FAD,
      plot = plot,
      verbose = verbose,
      ...)

    # (3) DRC Data ------------------------------------------------------------
    ## get DRC data
    df_DRC <- object@data$DRC[object@data$DRC$SAMPLE == sample_names[i],]

    ## get DRC data in correct form
    df_DRC <- df_DRC[,c("TIME", "LxTx", "LxTx_ERROR", "ALQ")]
    df_DRC$LxTx_ERROR[is.na(df_DRC$LxTx_ERROR)] <- 0

    ## add dose rate if available
    if (!is.null(object@info$Ddot_DRC)) {
      Ddot_DRC <- object@info$Ddot_DRC[[i]]
      ## we have to do this aliquot wise
      for (d in unique(df_DRC[["ALQ"]])) {
        aliquot <- df_DRC$ALQ == d
        df_DRC$TIME[aliquot] <- df_DRC$TIME[aliquot] * Ddot_DRC[d]
      }

      ## adjust column names
      colnames(df_DRC) <- c("DOSE", colnames(df_DRC[-1]))
    }

    ## DRC fitting
    results_DRC <- suppressWarnings(
        fit_DoseResponseCurve(
            object = df_DRC,
            mode = "alternate",
            verbose = verbose))

    ## DRC plotting
    plot_DoseResponseCurve(
      object = results_DRC,
      xlab = if ("DOSE" %in% colnames(df_DRC)) "Dose [Gy]" else "Dose [s]",
      log = "x",
      plot_extended = FALSE,
      mar = c(5.1, 4.1, 4.1, 2.1),
      mgp = c(3, 1, 0),
      tcl = -0.5,
      main = sample_names[i])

    ## return single lists
    return(list(results_FAD, results_ITL, results_DRC))
  })

  ## merge results for each
    ## flatten list
    results_combined <- unlist(results_combined, recursive = TRUE)

  ## get originators (we will merge accordingly)
  originator <- vapply(results_combined, function(x) x@originator, character(1))

  ## get list with merged results
  results_combined <- lapply(unique(originator), function(x) {
    merge_RLum(results_combined[originator == x])
  })
  names(results_combined) <- unique(originator)

# Results -----------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      FAD = results_combined[["analyse_FadingMeasurement"]],
      ITL = results_combined[["fit_IsothermalHolding"]],
      DRC = results_combined[["fit_DoseResponseCurve"]]),
    info = list(
      call = sys.call()
    ))
}

## load data
input.csv <- file.path(test_path("_data"),
                       paste0("CLBR_IR", c(50, 100, 150, 225), ".csv"))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(analyse_ThermochronometryData(list()),
               "'object' should be of class 'character'")
  expect_error(analyse_ThermochronometryData("error"),
               "File 'error' does not exist")
  expect_error(analyse_ThermochronometryData(input.csv[1], verbose = "error"),
               "'verbose' should be a single logical value")
  expect_error(analyse_ThermochronometryData(test_path("_data/CLBR.xlsx")),
               "File extension 'xlsx' is not supported, only 'csv' is valid")
  expect_error(analyse_ThermochronometryData(input.csv[1], ITL_model = "error"),
               "'ITL_model' should be one of 'GOK' or 'BTS'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_s4_class(analyse_ThermochronometryData(input.csv[1],
                                                verbose = FALSE),
                  "RLum.Results")
  SW({
  expect_s4_class(analyse_ThermochronometryData(input.csv[1], ITL_model = "BTS",
                                                verbose = TRUE,
                                                num_s_values_bts = 10),
                  "RLum.Results")
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              analyse_ThermochronometryData(input.csv[1]))
  })
})

#' @title Calculate Lx/Tx ratio for decomposed CW-OSL signal components
#'
#' @description Calculate `Lx/Tx` ratios from a given set of decomposed
#' CW-OSL curves decomposed by `OSLdecomposition::RLum.OSL_decomposition`.
#'
#' @param OSL.component [integer] or [character] (*optional*):
#' a single index or a name describing which OSL signal component shall be evaluated.
#' This argument can either be the name of the OSL component assigned by
#' `OSLdecomposition::RLum.OSL_global_fitting` or the index of component.
#' Then `'1'` selects the fastest decaying component, `'2'` the
#' second fastest and so on. If not defined, the fastest decaying component is selected.
#'
#' @param Lx.data [data.frame] (**required**): Component table created by
#' `OSLdecomposition::RLum.OSL_decomposition` and per default located
#' at `object@records[[...]]@info$COMPONENTS`.The value of `$n[OSL.component]`
#' is set as `LnLx`. The value of `$n.error[OSL.component]` is set as `LnLx.error`
#'
#' @param Tx.data [data.frame] (*optional*): Component table created by
#' `OSLdecomposition::RLum.OSL_decomposition` and per default located at
#' `object@records[[...]]@info$COMPONENTS`. The value of `$n[OSL.component]`
#' is set as `TnTx`. The value of `$n.error[OSL.component]` is set as `TnTx.error`
#'
#' @param sig0 [numeric] (*with default*): allows adding an extra error component
#' to the final `Lx/Tx` error value (e.g., instrumental error).
#'
#' @param digits [integer] (*with default*): round numbers to the specified digits.
#' If digits is set to `NULL` nothing is rounded.
#'
#' @return Returns an S4 object of type [Luminescence::RLum.Results-class].
#'
#' Slot `data` contains a [list] with the following structure:
#'
#' **@data**
#' ```
#' $LxTx.table (data.frame)
#' .. $ Net_LnLx
#' .. $ Net_LnLx.Error
#' .. $ Net_TnTx
#' .. $ Net_TnTx.Error
#' .. $ SN_RATIO_LnLx = NA,
#' .. $ SN_RATIO_TnTx = NA
#' .. $ LxTx
#' .. $ LxTx.Error
#' ```
#'
#' @section Function version: 0.1.1
#'
#' @author Dirk Mittelstrass
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::fit_DoseResponseCurve],
#' [Luminescence::analyse_SAR.CWOSL]
#'
#' @references Mittelstrass D., Schmidt C., Beyer J., Straessner A., 2019.
#' Automated identification and separation of quartz CW-OSL signal components with R.
#' talk presented at DLED 2019, Bingen, Germany
#' [http://luminescence.de/OSLdecomp_talk.pdf]()\cr
#'
#' @keywords datagen
#' @export
calc_OSLLxTxDecomposed <- function(
  Lx.data,
  Tx.data = NULL,
  OSL.component = 1L,
  sig0 = 0,
  digits = NULL
) {
  .set_function_name("calc_OSLLxTxDecomposed")
  on.exit(.unset_function_name(), add = TRUE)

  # ToDo:
  # - Integrity checks for the component table
  # - Handle background-signal-component if present
  # - add Tx.data integrity checks
  # - add previous-residual-subtraction functionality
  # - add list with decomposition algorithm parameters to return object
  # - add example in documentation

  ##--------------------------------------------------------------------------##
  ## (1) - integrity checks
  .validate_class(Lx.data, "data.frame")
  if (nrow(Lx.data) < 2)
    .throw_error("No valid component data.frame for Lx value")
  if (!all(c("n", "n.error") %in% colnames(Lx.data)))
    .throw_error("'Lx.data' should contain the following columns: 'n', 'n.error'")

  .validate_class(Tx.data, "data.frame", null.ok = TRUE)
  if (!is.null(Tx.data)) {
    if (nrow(Tx.data) < 2)
      .throw_error("No valid component data.frame for Tx value")
    if (!all(c("n", "n.error") %in% colnames(Tx.data)))
      .throw_error("'Tx.data' should contain the following columns: 'n', 'n.error'")
  }

  # define the component
  component_index <- NA

  #select only the first element; we do this silently because it is clearly
  #written in the documentation
  OSL.component <- OSL.component[1]
  .validate_class(OSL.component, c("integer", "numeric", "character"))

  # get component index from component name
  if (is.character(OSL.component)) {
    if (tolower(OSL.component) %in% tolower(Lx.data$name)) {
      component_index <- which(tolower(OSL.component) == tolower(Lx.data$name))
    } else {
      .throw_error("Invalid OSL component name, valid names are: ",
                   .collapse(Lx.data$name))
    }
  }

  # if a numeric is given, check if it matches with any component index
  if (is.numeric(OSL.component)) {
    OSL.component <- as.integer(OSL.component)
    if (OSL.component %in% 1:nrow(Lx.data)) {
      component_index <- OSL.component

      # insert background-signal-component check here

    } else {
      .throw_error("Invalid OSL component index, ",
                   "component table has ", nrow(Lx.data), " rows")
    }
  }

  .validate_positive_scalar(digits, int = TRUE, null.ok = TRUE)

  ##--------------------------------------------------------------------------##
  ## (2) - extract Lx and Tx values

  LnLx <- Lx.data$n[component_index]
  LnLx.Error <- Lx.data$n.error[component_index]

  TnTx <- 1
  TnTx.Error <- 0
  if (!is.null(Tx.data)) {
    TnTx <- Tx.data$n[component_index]
    TnTx.Error <- Tx.data$n.error[component_index]
  }

  ##--------------------------------------------------------------------------##
  ##(4) Calculate LxTx error according Galbraith (2014)

  ## transform results to a data.frame
  LnLxTnTx <- data.frame(Net_LnLx = LnLx,
                         Net_LnLx.Error = LnLx.Error,
                         Net_TnTx = TnTx,
                         Net_TnTx.Error = TnTx.Error,
                         SN_RATIO_LnLx = NA,
                         SN_RATIO_TnTx = NA)

  temp <- .calculate_LxTx_error(LnLxTnTx, sig0, digits)

  # ToDo: Add decomposition algorithm parameters here
  # calc.parameters <- list(...)

  ##set results object
  set_RLum(
      class = "RLum.Results",
      data = list(
        LxTx.table = temp),
      #  calc.parameters = calc.parameters),
      info = list(call = sys.call())
  )
}

## load data
data(ExampleData.LxTxOSLData, envir = environment())
colnames(Lx.data) <- colnames(Tx.data) <- c("n", "n.error")

## object from OSLdecomposed
Lx.decomp <- data.frame(name = c("Component 1", "Component 2"),
                        lambda = c(2.03, 0.02),
                        lambda.error = c(0.009, 0.002),
                        bleaching.grade = c(1, 0.3816),
                        t.start = c(0, 1.7),
                        t.end = c(1.7, 20),
                        ch.start = c(1, 18),
                        ch.end = c(17, 200),
                        n = c(4899.3, 7507.4),
                        bin = c(5046, 2701),
                        bin.RSS = c(5166.9, 5004.9),
                        bin.error = c(74.09, 70.94),
                        n.error = c(77.28, 208.61),
                        n.residual = c(0, 4643),
                        initial.signal = c(0.9804, 0.0196))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_OSLLxTxDecomposed(),
               "'Lx.data' should be of class 'data.frame")
  expect_error(calc_OSLLxTxDecomposed("test"),
               "'Lx.data' should be of class 'data.frame")
  expect_error(calc_OSLLxTxDecomposed(data.frame(col = integer(0))),
               "No valid component data.frame for Lx value")
  expect_error(calc_OSLLxTxDecomposed(iris),
               "'Lx.data' should contain the following columns:")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, "test"),
               "'Tx.data' should be of class 'data.frame")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, data.frame(col = integer(0))),
               "No valid component data.frame for Tx value")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, iris),
               "'Tx.data' should contain the following columns:")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, Tx.data,
                                      OSL.component = NA),
               "'OSL.component' should be of class 'integer', 'numeric' or")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, Tx.data,
                                      OSL.component = "error"),
               "Invalid OSL component name, valid names are:")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, Tx.data,
                                      OSL.component = 1000),
               "Invalid OSL component index, component table has 100 rows")
  expect_error(calc_OSLLxTxDecomposed(Lx.data, digits = NA),
               "'digits' should be a single positive integer value")
  expect_error(calc_OSLLxTxDecomposed(Lx.decomp, OSL.component = "1"),
               "Invalid OSL component name, valid names are: 'Component 1', 'Component 2'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_RLum(calc_OSLLxTxDecomposed(Lx.data, Tx.data, digits = 2))
  expect_snapshot_RLum(calc_OSLLxTxDecomposed(Lx.data, Tx.data, digits = 2,
                                              OSL.component = 2))
  expect_snapshot_RLum(calc_OSLLxTxDecomposed(Lx.data, Tx.data, digits = 2,
                                              OSL.component = 3, sig0 = 1000))
  expect_snapshot_RLum(calc_OSLLxTxDecomposed(Lx.decomp,
                                              OSL.component = "Component 1"))
})

#' @title Calculate `Lx/Tx` ratio for CW-OSL curves
#'
#' @description
#' Calculate `Lx/Tx` ratios from a given set of CW-OSL curves assuming late light
#' background subtraction.
#'
#' @details
#' The function checks the integrity of the values chosen for the signal and
#' background integrals; the signal integral limits have to be lower than
#' the background integral limits. If a [vector] is given as input instead
#' of a [data.frame], an artificial [data.frame] is produced. The
#' error calculation is done according to Galbraith (2002).
#'
#' **Please note:** In cases where the calculation results in `NaN` values (for
#' example due to zero-signal, and therefore a division of 0 by 0), these `NaN`
#' values are replaced by 0.
#'
#' **`sigmab`**
#'
#' The default value of `sigmab` is calculated assuming the background is
#' constant and **would not** applicable when the background varies as,
#' e.g., as observed for the early light subtraction method.
#'
#' **`sig0`**
#'
#' This argument allows to add an extra component of error to the final `Lx/Tx`
#' error value. The input will be treated as factor that is multiplied with
#' the already calculated `LxTx` and the result is add up by:
#'
#' \deqn{se(LxTx) = \sqrt{se(LxTx)^2 + (LxTx * sig0)^2}}
#'
#' **`SN_RATIO_LnLx` and `SN_RATIO_TnTx`**
#'
#' For convenience, the function returns the signal-to-noise ratio (`SN_RATIO`)
#' for the `LnLx` and the `TnTx` curves. This is simply the signal divided
#' by the background signal counts normalised to the `k` value (see below).
#'
#' **`background.count.distribution`**
#'
#' This argument allows selecting the distribution assumption that is used for
#' the error calculation. According to Galbraith (2002, 2014) the background
#' counts may be overdispersed (i.e. do not follow a Poisson distribution,
#' which is assumed for the photomultiplier counts). In that case (might be the
#' normal case) it has to be accounted for the overdispersion by estimating
#' \eqn{\sigma^2} (i.e. the overdispersion value). Therefore the relative
#' standard error is calculated as:
#'
#' - `poisson`
#' \deqn{rse(\mu_{S}) \approx \sqrt{Y_{0} + Y_{1}/k^2) / (Y_{0} - Y_{1}/k)} }
#'
#' - `non-poisson`
#' \deqn{rse(\mu_{S}) \approx \sqrt{Y_{0} + Y_{1}/k^2 + \sigma^2(1+1/k)) / (Y_{0} - Y_{1}/k)} }
#'
#' If `background_integral = NA`, then in both cases the relative standard
#' error simplifies to:
#'
#' \deqn{rse(\mu_{S}) \approx \sqrt{Y_{0}} / Y_{0}}
#'
#' **Please note** that when using the early background subtraction method in
#' combination with the 'non-poisson' distribution argument, the corresponding `Lx/Tx` error
#' may considerably increase due to a high `sigmab` value.
#' Please check whether this is valid for your data set and if necessary
#' consider to provide an own `sigmab` value using the corresponding argument `sigmab`.
#'
#' @param Lx.data [Luminescence::RLum.Data.Curve-class], [data.frame], [list] (**required**):
#' requires a CW-OSL shine down curve (x = time, y = counts). Data can also be
#' provided as a list.
#'
#' @param Tx.data [Luminescence::RLum.Data.Curve-class] or [data.frame] (*optional*):
#' requires a CW-OSL shine down curve (x = time, y = counts). If no
#' input is given the `Tx.data` will be treated as `NA` and no `Lx/Tx` ratio
#' is calculated. When `Lx.data` is a list, it must be provided as a list of
#' the same length.
#'
#' @param signal_integral [integer] (**required**):
#' vector of channels for the signal integral. If set to `NA` (alternate mode),
#' no integrals are taken into account and their settings are ignored.
#'
#' @param background_integral [integer] (**required**):
#' vector of channels for the background integral. If set to `NA`, no
#' background integral is subtracted; in this case, `sigmab.LnLx` (unless
#' manually set) and the signal-to-noise ratio for Ln/Lx` will be `NA`.
#'
#' @param signal_integral_Tx [integer] (*optional*):
#' vector of channels for the signal integral for the `Tx` curve.
#' If `NULL`, the `signal_integral` vector is used.
#'
#' @param background_integral_Tx [integer] (*optional*):
#' vector of channels for the background integral for the `Tx` curve.
#' If `NULL`, the `background_integral` vector is used. If set to `NA`, no
#' background integral for the `Tx` curve is subtracted; in this case,
#' `sigmab.TxTx` (unless manually set) and the signal-to-noise ratio for
#' `Tn/Tx` will be `NA`.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time/temperature range are selected.
#'
#' @param background.count.distribution [character] (*with default*):
#' sets the count distribution assumed for the error calculation.
#' Possible arguments are `"poisson"` or `"non-poisson"` (default). See
#' details for further information.
#'
#' @param use_previousBG [logical] (*with default*):
#' If set to `TRUE` the background of the `Lx`-signal is subtracted also
#' from the `Tx`-signal. Please note that in this case separate
#' signal integral limits for the `Tx`-signal are not allowed and will be reset.
#'
#' @param sigmab [numeric] (*optional*):
#' option to set a manual value for the overdispersion (for `LnTx` and `TnTx`),
#' used for the `Lx/Tx` error calculation. The value should be provided as
#' absolute squared count values, e.g. `sigmab = c(300,300)`.
#' **Note:** If only one value is provided this value is taken for both (`LnTx` and `TnTx`) signals.
#'
#' @param sig0 [numeric] (*with default*):
#' allow adding an extra component of error to the final `Lx/Tx` error value
#' (e.g., instrumental error, see details).
#'
#' @param digits [integer] (*with default*):
#' round numbers to the specified digits. If set to `NULL` no rounding occurs.
#'
#' @param ... currently not used.
#'
#' @return
#' Returns an S4 object of type [Luminescence::RLum.Results-class].
#'
#' Slot `data` contains a [list] with the following structure:
#'
#' **@data**
#' ```
#' $LxTx.table (data.frame)
#' .. $ LnLx
#' .. $ LnLx.BG
#' .. $ TnTx
#' .. $ TnTx.BG
#' .. $ Net_LnLx
#' .. $ Net_LnLx.Error
#' .. $ Net_TnTx
#' .. $ Net_TnTx.Error
#' .. $ SN_RATIO_LnLx,
#' .. $ SN_RATIO_TnTx,
#' .. $ LxTx
#' .. $ LxTx.Error
#' $calc.parameters (list)
#' .. $ sigmab.LnTx
#' .. $ sigmab.TnTx
#' .. $ k
#' ```
#'
#' **@info**
#' ```
#' $ call (original function call)
#' ```
#'
#' @note
#' The results of this function have been cross-checked with the Analyst
#' (version 3.24b). Access to the results object via [Luminescence::get_RLum].
#'
#' **Caution:** If you are using early light subtraction (EBG), please either provide your
#' own `sigmab` value or use `background.count.distribution = "poisson"`.
#'
#' @section Function version: 0.9.8
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::fit_DoseResponseCurve],
#' [Luminescence::analyse_SAR.CWOSL]
#'
#' @references Duller, G., 2018. Analyst v4.57 - User Manual.
#' `https://users.aber.ac.uk/ggd`\cr
#'
#' Galbraith, R.F., 2002. A note on the variance of a background-corrected OSL
#' count. Ancient TL, 20 (2), 49-51. \doi{10.26034/la.atl.2002.348}
#'
#' Galbraith, R.F., 2014. A further note on the variance of a
#' background-corrected OSL count. Ancient TL, 31 (2), 1-3. \doi{10.26034/la.atl.2014.477}
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.LxTxOSLData, envir = environment())
#'
#' ##calculate Lx/Tx ratio
#' results <- calc_OSLLxTxRatio(
#'  Lx.data = Lx.data,
#'  Tx.data = Tx.data,
#'  signal_integral = 1:2,
#'  background_integral = 85:100)
#'
#' ##get results object
#' get_RLum(results)
#'
#' @export
calc_OSLLxTxRatio <- function(
  Lx.data,
  Tx.data = NULL,
  signal_integral = NULL,
  background_integral = NULL,
  signal_integral_Tx = NULL,
  background_integral_Tx = NULL,
  integral_input = c("channel", "measurement"),
  background.count.distribution = "non-poisson",
  use_previousBG = FALSE,
  sigmab = NULL,
  sig0 = 0,
  digits = NULL,
  ...
) {
  .set_function_name("calc_OSLLxTxRatio")
  on.exit(.unset_function_name(), add = TRUE)

  ## Self-call --------------------------------------------------------------
  if (inherits(Lx.data, "list")) {
    if (!is.null(Tx.data) &&
        (!inherits(Tx.data, "list") || length(Lx.data) != length(Tx.data))) {
      .throw_error("When 'Lx.data' is a list, 'Tx.data' should either be ",
                   "a list of the same length or NULL")
    }
    ret <- lapply(seq_along(Lx.data), function(x) {
      calc_OSLLxTxRatio(Lx.data[[x]],
                        Tx.data[[x]],
                        signal_integral = signal_integral,
                        background_integral = background_integral,
                        signal_integral_Tx = signal_integral_Tx,
                        background_integral_Tx = background_integral_Tx,
                        integral_input = integral_input,
                        background.count.distribution = background.count.distribution,
                        use_previousBG = use_previousBG,
                        sigmab = sigmab,
                        sig0 = sig0,
                        digits = digits,
                        ...)
    })

    return(ret)
  }

  ## Integrity checks -------------------------------------------------------
  valid.classes <- c("RLum.Data.Curve", "data.frame", "matrix", "numeric")
  .validate_class(Lx.data, valid.classes, extra = "a list of such objects")
  .validate_not_empty(Lx.data)
  .validate_class(Tx.data, valid.classes, null.ok = TRUE,
                  extra = "a list of such objects")
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  .validate_class(background.count.distribution, "character", length = 1)
  .validate_logical_scalar(use_previousBG)
  .validate_class(sigmab, "numeric", null.ok = TRUE, length = 1:2)
  .validate_nonnegative_scalar(sig0)
  .validate_nonnegative_scalar(digits, int = TRUE, null.ok = TRUE)

  .coerce <- function(data) {
    data <- switch(
        class(data)[1],
        RLum.Data.Curve = as(data, "data.frame"),
        matrix = as.data.frame(data),
        data.frame = data,
        data.frame(x = 1:length(data), y = data)
    )
    if (ncol(data) < 2)
      .throw_error(.first_argument(), " should have 2 columns")
    data
  }

  ## Lx - coerce if required
  Lx.data <- .coerce(Lx.data)
  len.Lx <- nrow(Lx.data)

  ## Tx - coerce if required
  if(!is.null(Tx.data)){
    .validate_not_empty(Tx.data)
    Tx.data <- .coerce(Tx.data)
    len.Tx <- nrow(Tx.data)

    ## check channel number
    if (len.Lx != len.Tx) {
      .throw_error("Different number of channels for Lx (", len.Lx, ") ",
                   "and Tx (", len.Tx, ")")
    }
  } else {
    Tx.data <- data.frame(x = NA,y = NA)
  }

  ## deprecated arguments
  extraArgs <- list(...)
  if (any(grepl("[signal|background]\\.integral\\.?[.Tx]?", names(extraArgs))) &&
      !is.null(c(extraArgs$signal.integral, extraArgs$signal.integral.Tx,
                 extraArgs$background.integral, extraArgs$background.integral.Tx))) {
    .deprecated(old = c("signal.integral", "background.integral",
                        "signal.integral.Tx", "background.integral.Tx"),
                new = c("signal_integral", "background_integral",
                        "signal_integral_Tx", "background_integral_Tx"),
                since = "1.2.0")
    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal_integral <- extraArgs$signal.integral
    background_integral <- extraArgs$background.integral
    signal_integral_Tx <- extraArgs$signal.integral.Tx
    background_integral_Tx <- extraArgs$background.integral.Tx
  }

  # Alternate mode ----------------------------------------------------------
  if (.strict_na(signal_integral)) {
    LnLx <- sum(Lx.data[,2])
    TnTx <- sum(Tx.data[,2])

    LnLxTnTx <- data.frame(
      LnLx = LnLx,
      LnLx.BG = 0,
      TnTx = TnTx,
      TnTx.BG = 0,
      Net_LnLx = LnLx,
      Net_LnLx.Error = 0,
      Net_TnTx = TnTx,
      Net_TnTx.Error = 0,
      SN_RATIO_LnLx = 0,
      SN_RATIO_TnTx = 0,
      LxTx = LnLx/TnTx,
      LxTx.Error = 0)

    return(set_RLum(
      class = "RLum.Results",
      data = list(
        LxTx.table = LnLxTnTx,
        calc.parameters = NULL,
      info = list(call = sys.call())
    )))
  }

  # Continue checks ---------------------------------------------------------
  if (integral_input == "measurement") {
    unit <- "time"
    signal_integral <- .convert_to_channels(Lx.data[, 1], signal_integral, unit,
                                            na.ok = TRUE)
    background_integral <- .convert_to_channels(Lx.data[, 1], background_integral, unit,
                                                na.ok = TRUE)
    signal_integral_Tx <- .convert_to_channels(Lx.data[, 1], signal_integral_Tx, unit,
                                               null.ok = TRUE)
    background_integral_Tx <- .convert_to_channels(Lx.data[, 1], background_integral_Tx, unit,
                                                   na.ok = TRUE, null.ok = TRUE)
  }

  signal_integral <- .validate_integral(signal_integral, na.ok = TRUE,
                                        max = len.Lx)
  background_integral <- .validate_integral(background_integral, na.ok = TRUE,
                                            min = max(signal_integral) + 1,
                                            max = len.Lx)

  if (!anyNA(Tx.data) && xor(is.null(signal_integral_Tx),
                             is.null(background_integral_Tx))) {
    .throw_error("When 'Tx.data' is provided, either both 'signal_integral_Tx' ",
                 "and 'background_integral_Tx' must be provided, or neither")
  }

  any_given <- !is.null(signal_integral_Tx) || !is.null(background_integral_Tx)
  if (!anyNA(Tx.data) && use_previousBG && any_given) {
    .throw_warning("When 'use_previousBG = TRUE', independent Tx integral ",
                   "limits are not allowed, Lx limits will be used for Tx")
  }

  both_given <- !is.null(signal_integral_Tx) && !is.null(background_integral_Tx)
  if (!anyNA(Tx.data) && !use_previousBG && both_given) {
      signal_integral_Tx <- .validate_integral(signal_integral_Tx, max = len.Tx,
                                               null.ok = TRUE)
      background_integral_Tx <- .validate_integral(background_integral_Tx, na.ok = TRUE,
                                                   min = max(signal_integral_Tx) + 1,
                                                   max = len.Tx,
                                                   null.ok = TRUE)
  } else {
    signal_integral_Tx <- signal_integral
    background_integral_Tx <- background_integral
  }

  ##--------------------------------------------------------------------------##
  ##(2) - read data and produce background subtracted values

  ## calculate k value - express the background as multiple value from the number
  ## of signal integral channels, however, it can be < 1 also
  n <- length(signal_integral)
  m <- length(background_integral)
  k <- m/n
  n.Tx <- length(signal_integral_Tx)

  ##use previous BG and account for the option to set different integral limits
  m.Tx <- if (use_previousBG) m else length(background_integral_Tx)
  k.Tx <- m.Tx/n.Tx

  ##LnLx (comments are corresponding variables to Galbraith, 2002)
  Lx.curve <- Lx.data[,2]
  Lx.signal <- sum(Lx.curve[signal_integral])                #Y.0
  Lx.background <- sum(Lx.curve[background_integral]) / k    #Y.1, mu.B
  SN.ratio.LnLx <- Lx.signal / Lx.background
  if (.strict_na(background_integral))
    Lx.background <- 0

  LnLx <- Lx.signal - Lx.background

  ##TnTx
  Tx.curve <- ifelse(is.na(Tx.data[, 1]), NA, Tx.data[, 2])
  Tx.signal <- sum(Tx.curve[signal_integral_Tx])

  ##use previous BG
  if(use_previousBG)
    Tx.background <- Lx.background
  else
    Tx.background <- sum(Tx.curve[background_integral_Tx]) / k.Tx
  SN.ratio.TnTx <- Tx.signal / Tx.background
  if (.strict_na(background_integral_Tx))
    Tx.background <- 0

  TnTx <- (Tx.signal-Tx.background)

  ##--------------------------------------------------------------------------##
  ##(3)
  ## calculate Lx/Tx Errors according Galbraith (2002) and the personal
  ## communication of Galbraith (2014) via e-mail
  ## Nomenclature as stated in the articles

  ##(a)
  ## set Y.0 (sum OSL signal including the background) and
  ## Y.1 (total counts over m later channels)
  Y.0 <- Lx.signal
  Y.0_TnTx <- Tx.signal
  Y.1 <- if (.strict_na(background_integral)) 0 else sum(Lx.curve[background_integral])
  Y.1_TnTx <- if (.strict_na(background_integral_Tx)) 0 else sum(Tx.curve[background_integral_Tx])

  ##(b) estimate overdispersion (here called sigmab), see equation (4) in
  ## Galbraith (2002), Galbraith (2014)
  .calc_sigmab <- function(curve, signal_integral, background_integral,
                           m, k, what) {
    len.sg.integral <- length(signal_integral)
    min.bg.integral <- min(background_integral)

    if (round(k, digits = 1) >= 2 &&
        min.bg.integral + len.sg.integral * (2 + 1) <= length(curve)) {

      ## note that m = n*k = multiple of background_integral from signal_integral
      Y.i <- vapply(0:(floor(k) - 1), function(i) {
        sum(curve[min.bg.integral +
                  (len.sg.integral * i):(len.sg.integral * (i + 1) - 1)])
      }, FUN.VALUE = numeric(1))
      n <- 1
    } else {
      ## warn if m is < 25, as suggested by Rex Galbraith (low number of
      ## degrees of freedom)
      if (m < 25 && !.strict_na(background_integral)) {
        .throw_warning("Number of background channels for ", what, " < 25, ",
                       "error estimation might not be reliable")
      }

      Y.i <- curve[background_integral]
      n <- len.sg.integral
    }

    ## sigmab is denoted as sigma^2 = s.Y^2 - Y.mean, provided it's positive
    max(stats::var(Y.i) - mean(Y.i), 0) * n
  }

  ##account for a manually set sigmab value
  if (!is.null(sigmab)) {
    sigmab.LnLx <- sigmab[1]
    sigmab.TnTx <- sigmab[length(sigmab)]
  } else {
    sigmab.LnLx <- .calc_sigmab(Lx.curve, signal_integral, background_integral,
                                m, k, "Lx")
    sigmab.TnTx <- .calc_sigmab(Tx.curve, signal_integral_Tx, background_integral_Tx,
                                m.Tx, k.Tx, "Tx")
  }

  ##(c)
  ## Calculate relative error of the background subtracted signal
  ## according to Galbraith (2002), equation (6) with changes
  ## from Galbraith (2014), equation 6
  ## Discussion with Rex Galbraith via e-mail (2014-02-27):
  ## Equation 6 is appropriate to be implemented as standard

  ## relative standard error from equation (6)
  ## when sigmab = 0, this reduces to equation (3), valid for poisson
  ## when background_integral = NA, we have Y1 = 0 and sigmab = 0, and this
  ## reduces to sqrt(Y0) / Y0
  rse <- function(Y0, Y1, k, sigmab) {
    sqrt(Y0 + Y1 / k^2 + sigmab * (1 + 1 / k)) / (Y0 - Y1 / k)
  }

  if(background.count.distribution == "poisson"){
    used.sigmab.LnLx <- used.sigmab.TnTx <- 0
  }else{
    if(background.count.distribution != "non-poisson"){
      .throw_warning("Unknown method for 'background.count.distribution', ",
                     "a non-poisson distribution is assumed")
    }
    used.sigmab.LnLx <- if (.strict_na(background_integral)) 0 else sigmab.LnLx
    used.sigmab.TnTx <- if (.strict_na(background_integral_Tx)) 0 else sigmab.TnTx
  }

  LnLx.relError <- rse(Y.0, Y.1, k, used.sigmab.LnLx)
  TnTx.relError <- rse(Y.0_TnTx, Y.1_TnTx, k, used.sigmab.TnTx)

  ##(d)
  ##calculate absolute standard error
  LnLx.Error <- abs(LnLx*LnLx.relError)
  TnTx.Error <- abs(TnTx*TnTx.relError)

    ##we do not want to have NaN values, as they are mathematically correct, but make
    ##no sense and would result in aliquots that become rejected later
    if(is.nan(LnLx.Error)) LnLx.Error <- 0
    if(is.nan(TnTx.Error)) TnTx.Error <- 0

  ##combine results
  LnLxTnTx <- data.frame(
    LnLx = Lx.signal,
    LnLx.BG = Lx.background,
    TnTx = Tx.signal,
    TnTx.BG = Tx.background,
    Net_LnLx = LnLx,
    Net_LnLx.Error = LnLx.Error,
    Net_TnTx = TnTx,
    Net_TnTx.Error = TnTx.Error,
    SN_RATIO_LnLx = SN.ratio.LnLx,
    SN_RATIO_TnTx = SN.ratio.TnTx)

  ## ------------------------------------------------------------------------
  ## (4) Calculate LxTx error according Galbraith (2014)
  temp <- .calculate_LxTx_error(LnLxTnTx, sig0, digits)

  calc.parameters <- list(
    sigmab.LnLx = sigmab.LnLx,
    sigmab.TnTx = sigmab.TnTx,
    k = k)

  ##set results object
  set_RLum(
      class = "RLum.Results",
      data = list(
        LxTx.table = temp,
        calc.parameters = calc.parameters),
      info = list(call = sys.call())
  )
}

## load data
data(ExampleData.LxTxOSLData, envir = environment())

test_that("test input", {
  testthat::skip_on_cran()

  ##RLum.Curve
  expect_silent(calc_OSLLxTxRatio(
    set_RLum(class = "RLum.Data.Curve", data = as.matrix(Lx.data)),
    set_RLum(class = "RLum.Data.Curve", data = as.matrix(Tx.data)),
    signal_integral = 1:2,
    background_integral = 70:100))

  ##matrix
  expect_silent(calc_OSLLxTxRatio(
    as.matrix(Lx.data),
    as.matrix(Tx.data),
    signal_integral = 1:2,
    background_integral = 70:100))

  ##numeric
  expect_silent(calc_OSLLxTxRatio(
    as.numeric(Lx.data[,2]),
    as.numeric(Tx.data[,2]),
    signal_integral = 1:2,
    background_integral = 70:100))

  ##RLum.Curve
  expect_silent(calc_OSLLxTxRatio(
    set_RLum(class = "RLum.Data.Curve", data = as.matrix(Lx.data)),
    Tx.data = NULL,
    signal_integral = 1:2,
    background_integral = 70:100))

  ##matrix
  expect_silent(calc_OSLLxTxRatio(
    as.matrix(Lx.data),
    Tx.data = NULL,
    signal_integral = 1:2,
    background_integral = 70:100))

  ##numeric
  expect_silent(calc_OSLLxTxRatio(
    as.numeric(Lx.data[,2]),
    Tx.data = NULL,
    signal_integral = 1:2,
    background_integral = 70:100))

  ## list
  expect_type(calc_OSLLxTxRatio(
      list(Lx.data),
      Tx.data = NULL,
      signal_integral = 1:2,
      background_integral = 70:100),
      "list")

  ## integral_input
  expect_equal(calc_OSLLxTxRatio(Lx.data, Tx.data,
                                 signal_integral = c(0.0, 0.4),
                                 background_integral = c(13.8, 15.8, 19.8),
                                 integral_input = "measurement")@data,
               calc_OSLLxTxRatio(Lx.data, Tx.data,
                                 signal_integral = 1:3,
                                 background_integral = 70:100,
                                 integral_input = "channel")@data)
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_OSLLxTxRatio(numeric()),
               "'Lx.data' cannot be an empty numeric")
  expect_error(calc_OSLLxTxRatio(Lx.data, "error"),
               "'Tx.data' should be of class 'RLum.Data.Curve', 'data.frame', 'matrix', 'numeric', a list")
  expect_error(calc_OSLLxTxRatio("error", "error"),
               "'Lx.data' should be of class 'RLum.Data.Curve', 'data.frame'")
  expect_error(calc_OSLLxTxRatio(Lx.data[, 1, drop = FALSE]),
               "'Lx.data' should have 2 columns")
  expect_error(calc_OSLLxTxRatio(Lx.data, Tx.data[, 1, drop = FALSE]),
               "'Tx.data' should have 2 columns")
  expect_error(calc_OSLLxTxRatio(Lx.data[1:10, ], Tx.data),
               "Different number of channels for Lx (10) and Tx (100)",
               fixed = TRUE)
  expect_error(calc_OSLLxTxRatio(Lx.data[1:10, ], signal_integral = NULL),
               "'signal_integral' should be of class 'integer', 'numeric' or NA")
  expect_error(calc_OSLLxTxRatio(Lx.data[1:10, ], signal_integral = list(1, 2)),
               "'signal_integral' should be of class 'integer', 'numeric' or NA")
  expect_warning(expect_error(
      calc_OSLLxTxRatio(Lx.data[1:10, ], signal_integral = -1:4),
      "'background_integral' should be of class 'integer', 'numeric' or NA"),
      "'signal_integral' out of bounds, reset to be between 1 and 4")
  expect_error(calc_OSLLxTxRatio(Lx.data[1:10, ], signal_integral = 1:2,
                                 background_integral = matrix(1:4, ncol =2)),
               "'background_integral' should be of class 'integer', 'numeric' or NA")
  expect_error(calc_OSLLxTxRatio(list(Lx.data, Lx.data), Tx.data),
               "'Tx.data' should either be a list of the same length or NULL")

  expect_error(calc_OSLLxTxRatio(Lx.data, Tx.data,
                                 signal_integral = 1:20, background_integral = NA,
                                 signal_integral_Tx = 1:20,
                                 background_integral_Tx = NULL),
               "When 'Tx.data' is provided, either both 'signal_integral_Tx' and")

  expect_warning(expect_error(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:2000,
    background_integral = 85:100
  ), "'background_integral' is expected to be at least 101, but the maximum allowed is 100"),
  "'signal_integral' out of bounds, reset to be between 1 and 100")

  SW({
  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:90,
    background_integral = 85:100
  ), "'background_integral' out of bounds, reset to be between 91 and 100")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:10,
    signal_integral_Tx = 1:90,
    background_integral = 85:100,
    background_integral_Tx = 85:100
  ), "'background_integral_Tx' out of bounds, reset to be between 91 and 100")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    background_integral = 85:1000
  ), "'background_integral' out of bounds, reset to be between 85 and 100")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:10,
    signal_integral_Tx = 1:10,
    background_integral = 85:100,
    background_integral_Tx = 85:10000
  ), "'background_integral_Tx' out of bounds, reset to be between 85 and 100")
  })

  expect_warning(expect_error(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:10,
    signal_integral_Tx = 1:1000,
    background_integral = 85:100,
    background_integral_Tx = 85:100
  ), "background_integral_Tx' is expected to be at least 101, but the maximum"),
  "'signal_integral_Tx' out of bounds, reset to be between 1 and 100")

  expect_error(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    signal_integral_Tx = 1:20,
    background_integral = 80:100,
    background_integral_Tx = NULL
  ), "When 'Tx.data' is provided, either both 'signal_integral_Tx' and")

  expect_error(calc_OSLLxTxRatio(Lx.data, Tx.data, background.count.distribution = NA),
               "'background.count.distribution' should be of class 'character' and")
  expect_error(calc_OSLLxTxRatio(Lx.data, Tx.data, use_previousBG = NA),
               "'use_previousBG' should be a single logical value")

  expect_error(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    background_integral = 80:100,
    sigmab = "test"
  ), "'sigmab' should be of class 'numeric'")

  expect_error(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    background_integral = 80:100,
    sigmab = c(1, 2, 3)
  ), "'sigmab' should be of class 'numeric' or NULL and have length 1 or 2")

  expect_error(calc_OSLLxTxRatio(Lx.data, Tx.data, sig0 = -1),
               "'sig0' should be a single non-negative value")
  expect_error(calc_OSLLxTxRatio(Lx.data, Tx.data, digits = -1),
               "'digits' should be a single non-negative integer value or NULL")
})

test_that("create warnings", {
  testthat::skip_on_cran()

  ## deprecated arguments
  expect_warning(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal.integral = 1:2,
      background.integral = 85:100,
      digits = 1),
      "were deprecated in v1.2.0, use 'signal_integral', 'background_integral'")
  expect_warning(expect_error(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal.integral = 1:2,
      background.integral = 85:100,
      integral_input = "measurement"),
      "'integral_input' is not supported with old argument names"),
      "were deprecated in v1.2.0, use 'signal_integral', 'background_integral'")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    signal_integral_Tx = 1:20,
    background_integral = 80:100,
    background_integral_Tx = 60:100
  ), "Number of background channels for Lx < 25, error estimation might not be reliable")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    signal_integral_Tx = 1:20,
    background_integral = 60:100,
    background_integral_Tx = 80:100
  ), "Number of background channels for Tx < 25, error estimation might not be reliable")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    background_integral = 60:100,
    background.count.distribution = "hallo"
  ), "Unknown method for 'background.count.distribution', a non-poisson")

  expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    signal_integral_Tx = 2:20,
    background_integral = 60:100,
    background_integral_Tx = 40:100,
    use_previousBG = TRUE
  ), "When 'use_previousBG = TRUE', independent Tx integral limits are not")

  ## integral_input = "measurement"
  expect_warning(expect_warning(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:10,
    signal_integral_Tx = 1:7,
    background_integral = 10:20,
    background_integral_Tx = 7:100,
    integral_input = "measurement"),
    "'background_integral' out of bounds, reset to be between 52 and 100"),
    "'background_integral_Tx' out of bounds, reset to be between 37 and 100")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()
  snapshot.tolerance <- 1.5e-6

  ## digits
  expect_snapshot_RLum(expect_silent(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal_integral = 1:2,
      background_integral = 85:100,
      digits = 1)
  ), tolerance = snapshot.tolerance)

  ## sigmab
  expect_snapshot_RLum(expect_silent(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal_integral = 1:2,
      background_integral = 85:100,
      sigmab = c(1000,100))
  ), tolerance = snapshot.tolerance)

  ## poisson
  expect_snapshot_RLum(expect_silent(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal_integral = 1:2,
      background_integral = 85:100,
      background.count.distribution = "poisson")
  ), tolerance = snapshot.tolerance)

  ## use_previousBG
  expect_snapshot_RLum(calc_OSLLxTxRatio(
      Lx.data = Lx.data,
      Tx.data = Tx.data,
      signal_integral = 1:2,
      background_integral = 85:100,
      use_previousBG = TRUE
  ), tolerance = snapshot.tolerance)

  ## check weird circumstances
  ##(1) - Lx curve 0
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    data.frame(Lx.data[,1],0),
    Tx.data,
    signal_integral = 1:2,
    background_integral = 85:100
  ))

  ##(2) - Tx curve 0
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    Lx.data,
    data.frame(Tx.data[,1],0),
    signal_integral = 1:2,
    background_integral = 85:100
  ))

  ##(3) - Lx and Tx curve 0
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    data.frame(Lx.data[,1],0),
    data.frame(Tx.data[,1],0),
    signal_integral = 1:2,
    background_integral = 85:100
  ))

  ##(4) - Lx < 0
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    data.frame(Lx.data[,1],-1000),
    data.frame(Tx.data[,1],0),
    signal_integral = 1:2,
    background_integral = 85:100
  ))

  ##(5) - Tx < 0
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    Lx.data,
    data.frame(Lx.data[,1],-1000),
    signal_integral = 1:2,
    background_integral = 85:100
  ))

  ##(6) - Lx & Tx < 0
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    data.frame(Lx.data[,1],-1000),
    data.frame(Tx.data[,1],-1000),
    signal_integral = 1:2,
    background_integral = 85:100
  ))

  ## alternate mode with no signal integrals
  expect_snapshot_RLum(calc_OSLLxTxRatio(
    Lx.data = Lx.data,
    Tx.data = Tx.data,
    signal_integral = NA,
    background_integral = NA
  ))

  ## ------------------------------------------------------------------------
  ## background_integral = NA

  expect_snapshot_RLum(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data = NULL,
      signal_integral = 1:2,
      background_integral = NA))

  expect_snapshot_RLum(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal_integral = 1:2,
      background_integral = NA))

  expect_snapshot_RLum(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    background_integral = NA,
    signal_integral_Tx = 1:10,
    background_integral_Tx = NA),
    tolerance = snapshot.tolerance)

  expect_snapshot_RLum(calc_OSLLxTxRatio(
    Lx.data,
    Tx.data,
    signal_integral = 1:20,
    background_integral = NA,
    signal_integral_Tx = 1:10,
    background_integral_Tx = 70:100))

  expect_snapshot_RLum(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal_integral = 1:2,
      background_integral = NA,
      use_previousBG = TRUE))

  expect_snapshot_RLum(calc_OSLLxTxRatio(
      Lx.data,
      Tx.data,
      signal_integral = 1:20,
      background_integral = 70:100,
      signal_integral_Tx = 1:10,
      background_integral_Tx = NA),
      tolerance = snapshot.tolerance)
})

#' General structure function for RLum S4 class objects
#'
#' Function calls object-specific get functions for RLum S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum-class] objects.\cr
#' Depending on the input object, the corresponding structure function will
#' be selected. Allowed arguments can be found in the documentations of the
#' corresponding [RLum-class] class.
#'
#' @param object [RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @param ... further arguments that one might want to pass to the specific
#' structure method
#'
#' @return
#' Returns a [data.frame] with structure of the object.
#'
#' @section Function version: 0.2.0
#'
#' @author
#' Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [RLum.Data.Curve-class], [RLum.Data.Image-class],
#' [RLum.Data.Spectrum-class], [RLum.Analysis-class], [RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.XSYG, envir = environment())
#'
#' ##show structure
#' structure_RLum(OSL.SARMeasurement$Sequence.Object)
#'
#' @md
#' @export
setGeneric("structure_RLum", function(object, ...) {
  standardGeneric("structure_RLum")
})

# Method for structure_RLum method for RLum objects in a list for a list of objects  -------------
#' @describeIn structure_RLum
#' Returns a list of [RLum-class] objects that had been passed to [structure_RLum]
#'
#'
#' @md
#' @export
setMethod("structure_RLum",
          signature = "list",
          function(object, ...) {
            ##apply method in the objects and return the same
            lapply(object, function(x) {
              if (inherits(x, "RLum")) {
                return(structure_RLum(x, ...))
              } else{
                return(x)
              }

            })

          })

#' @title Obtain the equivalent dose using the approach by Woda and Fuchs 2008
#'
#' @description
#' The function generates a histogram-like reorganisation of the data, to
#' assess counts per bin. The log-transformed counts per bin are used to
#' calculate the second derivative of the data (i.e., the curvature of the
#' curve) and to find the central value of the bin hosting the distribution
#' maximum. A normal distribution model is fitted to the counts per bin
#' data to estimate the dose distribution parameters. The uncertainty of the
#' model is estimated based on all input equivalent doses smaller that of the
#' modelled central value.
#'
#' @param data [data.frame] [vector], or [Luminescence::RLum.Results-class] object (**required**):
#' for [data.frame]: either two columns: De (`values[,1]`) and De error
#' (`values[,2]`), or one: De (`values[,1]`). If a numeric vector or a
#' single-column data frame is provided, De error is set to `NA`.
#' For plotting multiple data sets, these must be provided as `list`
#' (e.g. `list(dataset1, dataset2)`).
#'
#' @param breaks [numeric]:
#' Either number or locations of breaks. See `[hist]` for details.
#' If missing, the number of breaks will be estimated based on the bin width
#' (as function of median error).
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... Further plot arguments passed to the function.
#'
#' @section Function version: 0.2.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany)
#'
#' @seealso [Luminescence::calc_FuchsLang2001], [Luminescence::calc_CentralDose]
#'
#' @references
#' Woda, C., Fuchs, M., 2008. On the applicability of the leading edge method to
#' obtain equivalent doses in OSL dating and dosimetry. Radiation Measurements 43, 26-37.
#'
#' @examples
#'
#' ## read example data set
#' data(ExampleData.DeValues, envir = environment())
#'
#' results <- calc_WodaFuchs2008(
#'   data = ExampleData.DeValues$CA1,
#'    xlab = expression(paste(D[e], " [Gy]"))
#'  )
#'
#' @export
calc_WodaFuchs2008 <- function(
  data,
  breaks = NULL,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_WodaFuchs2008")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO
  # - complete manual
  # - add statistics to the plot
  # - check whether this makes sense at all ...

  ## Integrity checks -------------------------------------------------------

  if (!.validate_class(data, c("data.frame", "RLum.Results", "numeric"),
                       throw.error = FALSE)) {
    return(NULL)
  }
  .validate_not_empty(data)
  .validate_positive_scalar(breaks, null.ok = TRUE)
  .validate_logical_scalar(plot)

  if (inherits(data, "RLum.Results")) {
        data <- tryCatch(get_RLum(data, "data"),
                         error = function(e) get_RLum(data))
  }

  if (!is.numeric(data[[1]])) {
    .throw_error("'data' should have only numeric fields")
  }

      ## if data is a numeric vector or a single-column data frame,
      ## append a second column of NAs
      if (NCOL(data) < 2) {
        data <- cbind(data, NA)
      }

      ## with just one data point, it's possible to cause nls() to hang
      if (nrow(data) < 2) {
        .throw_error("Insufficient number of data points")
      }

  if (any(is.infinite(unlist(data)))) {
    .throw_warning("Inf values found in 'data', replaced by NA")
    data[is.infinite(data[, 1]), 1] <- NA
    data[is.infinite(data[, 2]), 2] <- NA
  }

  ## read additional arguments
  trace <- isTRUE(list(...)$trace)

  ## calculations -------------------------------------------------------------

  ## estimate bin width based on Woda and Fuchs (2008)
  if (all(is.na(data[, 2]))) {
    .throw_message("No errors provided, bin width set by 10 percent ",
                   "of input data", error = FALSE)
    bin_width <- median(data[,1] / 10,
                        na.rm = TRUE)
  } else {
    bin_width <- median(data[,2],
                        na.rm = TRUE)
  }

  if (bin_width <= 0)
    .throw_error("The estimated bin width is not positive, check your data")

  ## optionally estimate class breaks based on bin width
  n_breaks <- breaks %||% (diff(range(data[, 1], na.rm = TRUE)) / bin_width)

  if (n_breaks <= 3) {
    .throw_warning("Fewer than 4 bins produced, 'breaks' set to 4")
    n_breaks <- 4
  }

  ## calculate histogram
  H <- hist(x = data[,1],
            breaks = n_breaks,
            plot = FALSE)

  ## extract values from histogram object
  H_c <- H$counts
  H_m <- H$mids

  ## log counts
  counts_log <- log(H_c)

  ## commented out for issue 223
  ## ## estimate curvature
  ## curvature <- (counts_log[1] - counts_log[2]) /
  ##   (counts_log[1] - counts_log[3])

  ## do some other black magic
  class_center <- H$mids[H_c == max(H_c)]

  ## optionally print warning
  if(length(class_center) != 1) {
    .throw_warning("More than one maximum, fit may be invalid")
    class_center <- class_center[1]
  }

  ## fit normal distribution to data
  fit <-	nls(H_c ~ (A / sqrt(2 * pi * sigma^2)) *
               exp(-(H_m - class_center)^2 / (2 * sigma^2)),
             start = c(A = mean(H_m),
                       sigma = bin_width),
             control = c(maxiter = 5000),
             algorithm = "port",
             trace = trace)

  ## extract fitted parameters
  A <- coef(fit)["A"]
  sigma <- coef(fit)["sigma"]

  ## estimate dose
  D_estimate <- as.numeric(x = class_center - sigma)

  ## count number of values smaller than center class
  count_ID <- length(which(H_m <= class_center))

  ## extract H_m values smaller than center class
  H_m_smaller <- H_m[1:count_ID]

  ## calculate uncertainty according to Woda and Fuchs (2008)
  s <- round(sqrt(sum((H_m_smaller - D_estimate)^2) / (count_ID - 1)),
             digits = 2)

  ## plot output --------------------------------------------------------------
  if(plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    ##define plot settings
    plot_settings <- list(
      xlab = expression(paste(D[e], " [s]")),
      ylab = "Frequency",
      xlim = range(data[,1], na.rm = TRUE) + c(-10, 20),
      ylim = NULL,
      main = expression(paste(D[e]," applying Woda and Fuchs (2008)")),
      sub = NULL
    )

    plot_settings <- modifyList(x = plot_settings, val = list(...), keep.null = TRUE)

    plot(
      x = H,
      xlab = plot_settings$xlab,
      ylab = plot_settings$ylab,
      xlim = plot_settings$xlim,
      ylim = plot_settings$ylim,
      main = plot_settings$main,
      sub = plot_settings$sub
    )

    ## add curve with normal distribution
    x <- 0
    rm(x)
    curve((A / sqrt(2 * pi * sigma^2)) * exp(-(x- class_center)^2 / (2 * sigma^2)),
          add = TRUE,
          to = class_center,
          col = "red"
          )
  }

  ## return output ------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      D_estimate = data.frame(
        DP = D_estimate,
        DP.ERROR = s,
        CLASS_CENTER = class_center,
        BIN_WIDTH = bin_width,
        SIGMA = sigma,
        A = A,
        row.names = NULL
      ),
      breaks = H$breaks
    ),
    info = list(call = sys.call())
  )
}

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_warning(expect_null(calc_WodaFuchs2008("error")),
                 "'data' should be of class 'data.frame', 'RLum.Results' or")
  expect_error(calc_WodaFuchs2008(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(calc_WodaFuchs2008(data.frame(a = letters)),
               "'data' should have only numeric fields")
  expect_error(calc_WodaFuchs2008(ExampleData.DeValues$CA1, breaks = 0),
               "'breaks' should be a single positive value")
  expect_error(calc_WodaFuchs2008(ExampleData.DeValues$CA1, plot = 0),
               "'plot' should be a single logical value")
  res <- calc_WodaFuchs2008(ExampleData.DeValues$CA1)
  expect_error(calc_WodaFuchs2008(res, breaks = 4),
               "Insufficient number of data points")
  expect_error(expect_message(
      calc_WodaFuchs2008(c(-1, 0, 1)),
      "No errors provided, bin width set by 10 percent of input data"),
      "The estimated bin width is not positive, check your data")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##test arguments
  expect_s4_class(calc_WodaFuchs2008(ExampleData.DeValues$CA1, plot = FALSE),
                  "RLum.Results")
  expect_warning(calc_WodaFuchs2008(ExampleData.DeValues$CA1[1:40, ]),
                 "More than one maximum, fit may be invalid")

  ## issue 197
  set.seed(1)
  df <- data.frame(rnorm(20, 10), rnorm(20, 0.5))
  expect_silent(calc_WodaFuchs2008(df))

  ## more coverage
  expect_warning(calc_WodaFuchs2008(df, breaks = 3),
                 "Fewer than 4 bins produced, 'breaks' set to 4")
  expect_output(calc_WodaFuchs2008(df, trace = TRUE))

  ## numeric vector
  expect_message(calc_WodaFuchs2008(df[, 1]),
                 "No errors provided")

  ## single-column data.frame
  expect_message(calc_WodaFuchs2008(df[, 1, drop = FALSE]),
                 "No errors provided")

  ## RLum.Results object
  obj <- calc_CommonDose(ExampleData.DeValues$BT998, verbose = FALSE)
  expect_silent(calc_WodaFuchs2008(obj))

  ## Inf values
  expect_warning(expect_warning(
      calc_WodaFuchs2008(data.frame(c(Inf, -0.5, 1234, 5), rep(1, 4))),
      "Inf values found in 'data', replaced by NA"),
      "More than one maximum, fit may be invalid")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("defaults",
                              calc_WodaFuchs2008(ExampleData.DeValues$CA1))
  vdiffr::expect_doppelganger("main breaks",
                              calc_WodaFuchs2008(ExampleData.DeValues$CA1,
                                                 main = "Woda-Fuchs (2008)",
                                                 breaks = 20))
  })
})

#' @title Calculate the Average Dose and the dose rate dispersion
#'
#' @description
#' This functions calculates the Average Dose and its extrinsic dispersion,
#' estimating the standard errors by bootstrapping based on the Average
#' Dose Model by Guérin et al., 2017.
#'
#' **`sigma_m`**\cr
#'
#'The program requires the input of a known value of `sigma_m`,
#'which corresponds to the intrinsic overdispersion, as determined
#'by a dose recovery experiment. Then the dispersion in doses (`sigma_d`)
#'will be that over and above `sigma_m` (and individual uncertainties `sigma_wi`).
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with `De` `(data[,1])` and `De error` `(values[,2])`
#'
#' @param sigma_m [numeric] (**required**):
#' the overdispersion resulting from a dose recovery
#' experiment, i.e. when all grains have  received the same dose. Indeed in such a case, any
#' overdispersion (i.e. dispersion on top of analytical uncertainties) is, by definition, an
#' unrecognised measurement uncertainty.
#'
#' @param Nb_BE [integer] (*with default*):
#' sample size used for the bootstrapping
#'
#' @param na.rm [logical] (*with default*):
#' exclude `NA` values from the data set prior to any further operation.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... further arguments that can be passed to [graphics::hist]. As three plots
#' are returned all arguments need to be provided as [list],
#' e.g., `main = list("Plot 1", "Plot 2", "Plot 3")`.
#' Note: not all arguments of `hist` are
#' supported, but the output of `hist` is returned and can be used of own plots. \cr
#'
#' Further supported arguments: `mtext` ([character]), `rug` (`TRUE/FALSE`).
#'
#' @section Function version: 0.1.7
#'
#' @author Claire Christophe, IRAMAT-CRP2A, Université de Nantes (France)\cr
#' Anne Philippe, Université de Nantes, (France)\cr
#' Guillaume Guérin, IRAMAT-CRP2A, Université Bordeaux Montaigne, (France)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [read.table], [graphics::hist]
#'
#' @return
#' The function produces numerical output and an optional plot. In addition,
#' an [Luminescence::RLum.Results-class] object containing the following
#' elements is returned:
#'
#' \item{$summary}{[data.frame] summary of all relevant model results.}
#' \item{$data}{[data.frame] with original input data.}
#' \item{$dstar}{[matrix] with bootstrap values.}
#' \item{$hist}{[list] object as produced by the [hist] function.}
#'
#' The `$summary` data.frame contains the following columns:
#' \tabular{lll}{
#'  **Column** \tab **Type** \tab **Description**\cr
#'  AVERAGE_DOSE \tab [numeric] \tab the obtained average dose\cr
#'  AVERAGE_DOSE.SE \tab [numeric] \tab the average dose error \cr
#'  SIGMA_D \tab [numeric]\tab sigma \cr
#'  SIGMA_D.SE \tab [numeric]\tab standard error of the sigma  \cr
#'  IC_AVERAGE_DOSE.LEVEL  \tab [character]\tab confidence level average dose\cr
#'  IC_AVERAGE_DOSE.LOWER  \tab [character]\tab lower quantile of average dose \cr
#'  IC_AVERAGE_DOSE.UPPER \tab [character]\tab upper quantile of average dose\cr
#'  IC_SIGMA_D.LEVEL \tab [integer]\tab confidence level sigma\cr
#'  IC_SIGMA_D.LOWER \tab [character]\tab lower sigma quantile\cr
#'  IC_SIGMA_D.UPPER \tab [character]\tab upper sigma quantile\cr
#'  L_MAX \tab [character]\tab maximum likelihood value \cr
#'  de \tab [numeric] \tab same as `AVERAGE_DOSE` (internal use)\cr
#'  de_err \tab [numeric] \tab same as `AVERAGE_DOSE.SE` (internal use)\cr
#' }
#'
#' The function returns two different plot panels.
#'
#' (1) An abanico plot with the dose values
#'
#' (2) A histogram panel comprising 3 histograms with the equivalent dose and the bootstrapped average
#' dose and the sigma values.
#'
#' @references
#' Guérin, G., Christophe, C., Philippe, A., Murray, A.S., Thomsen, K.J., Tribolo, C., Urbanova, P.,
#' Jain, M., Guibert, P., Mercier, N., Kreutzer, S., Lahaye, C., 2017. Absorbed dose, equivalent dose,
#' measured dose rates, and implications for OSL age estimates: Introducing the Average Dose Model.
#' Quaternary Geochronology 1-32. doi:10.1016/j.quageo.2017.04.002
#'
#' **Further reading**\cr
#'
#' Efron, B., Tibshirani, R., 1986. Bootstrap Methods for Standard Errors, Confidence Intervals,
#' and Other Measures of Statistical Accuracy. Statistical Science 1, 54-75.
#'
#' @note This function has beta status!
#'
#' @keywords datagen
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## calculate Average dose (using only the first 56 values here)
#' AD <- calc_AverageDose(ExampleData.DeValues$CA1[1:56, ], sigma_m = 0.1)
#'
#' ## plot De and set Average dose as central value
#' plot_AbanicoPlot(AD, z.0 = AD$summary$de)
#'
#' @export
calc_AverageDose <- function(
  data,
  sigma_m,
  Nb_BE = 500,
  na.rm = TRUE,
  plot = TRUE,
  verbose = TRUE,
  ...
) {
  .set_function_name("calc_AverageDose")
  on.exit(.unset_function_name(), add = TRUE)

  # Define internal functions ------------------------------------------------------------------

  # function which compute mle's for data (yu,su)
  .mle <- function(yu , su, wu.start, sigma_d.start, delta.start){

    ##set start parameters, otherwise the function will try to get them
    ##from the parent environment, which is not wanted ...
    delta.temp <- 0
    sigma_d.temp <- 0

    sigma_d <- sigma_d.start
    delta <- delta.start
    wu <- wu.start

    j <- 0
    iteration_limit <- 10000

    ##loop until convergence or the iteration limit is reached
    while(j < iteration_limit) {

      ##code by Claire; in the 2nd and 3rd line delta and sigma_d are replaced by delta.temp and
      ##sigma_d.temp; otherwise the iteration and its test for convergence will not work
      delta.temp <- exp( sum(wu*(yu+(0.5*(sigma_d^2)))) / sum(wu) )
      sigma_d.temp <- sigma_d*sum( (wu^2) * (yu-log(delta.temp)+0.5*sigma_d^2)^2) / (sum( wu*(1+yu-log(delta.temp)+0.5*sigma_d^2)))
      wu <- 1/(sigma_d.temp^2 + su^2)

      ##break loop if convergence is reached ... if not update values
      if(is.infinite(delta.temp) | is.infinite(sigma_d.temp)){
        .throw_warning("Inf/NaN values produced by .mle(), NA returned")
        return(c(NA,NA))

      }else if (
        ##compare values ... if they are equal we have convergence
        all(
          c(round(c(delta, sigma_d), 4)) == c(round(c(delta.temp, sigma_d.temp), 4))
        )
      ) {
        break()

      } else{
        ##update input values
        delta <- delta.temp
        sigma_d <- sigma_d.temp
        j <- j + 1
      }
    }

    ##if no convergence was reached stop entire function; no stop as this may happen during the
    ##bootstraping procedure
    if(j == iteration_limit){
      .throw_warning("No convergence reached by .mle() after ",
                     iteration_limit, " iterations, NA returned")
      return(c(NA,NA))
    }else{
      return(c(round(c(delta, sigma_d),4)))
    }
  }

  .CredibleInterval <- function(a_chain, level = 0.95) {
    ## Aim : estimation of the shortest credible interval of the sample of parameter a
    # A level % credible interval is an interval that keeps N*(1-level) elements of the sample
    # The level % credible interval is the shortest of all those intervals.
    ## Parameters :
    # a_chain : the name of the values of the parameter a
    # level : the level of the credible interval expected
    ## Returns : the level and the endpoints

    sorted_sample <- sort(a_chain)
    N <- length(a_chain)
    OutSample <- N * (1 - level)

    I <- cbind(sorted_sample[1:(OutSample + 1)] , sorted_sample[(N - OutSample):N])

    l <-  I[, 2] - I[, 1] # length of intervals
    i <- which.min(l) # look for the shortest interval

    return(c(
      level = level,
      CredibleIntervalInf = I[i, 1],
      CredibleIntervalSup = I[i, 2]
    ))
  }

  ##////////////////////////////////////////////////////////////////////////////////////////////////
  ##HERE THE MAIN FUNCTION STARTS
  ##////////////////////////////////////////////////////////////////////////////////////////////////

  # Integrity checks ----------------------------------------------------------------------------

  .validate_class(data, c("RLum.Results", "data.frame"))
  .validate_positive_scalar(sigma_m)
  .validate_positive_scalar(Nb_BE, int = TRUE)

  # Data preparation -----------------------------------------------------------------------------

  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data)
  }

  ## check that we actually have data
  if (length(data) == 0 || nrow(data) == 0) {
    .throw_message("'data' contains no data, NULL returned")
    return(NULL)
  }

  ##problem: the entire code refers to column names the user may not provide...
  ##  >> to avoid changing the entire code, the data will shape to a format that
  ##  >> fits to the code

  ##check for number of columns
  if(ncol(data)<2){
    .throw_message("'data' contains < 2 columns, NULL returned")
    return(NULL)
  }

  ##used only the first two colums
  if(ncol(data)>2){
    data <- data[,1:2]
    .throw_warning("'data' contains > 2 columns, ",
                   "only the first 2 columns were used")
  }

  ##exclude NA values
  if (anyNA(data)) {
    data <- na.exclude(data)
    .throw_warning("NA values in 'data' detected, rows with NA values removed")
  }

  if (any(data[, 1] <= 0)) {
    data <- data[data[, 1] > 0, ]
    .throw_warning("Non-positive values in 'data' detected, rows removed")
  }

  ##check data set
  if(nrow(data) == 0){
    .throw_message("After NA removal, nothing is left from the data set, ",
                   "NULL returned")
    return(NULL)
  }

  ##data becomes to dat (thus, make the code compatible with the code by Claire and Anne)
  dat <- data

  ##preset column names, as the code refers to it
  colnames(dat) <- c("cd", "se")


  # Pre calculation -----------------------------------------------------------------------------

  ##calculate  yu = log(CD) and su = se(logCD)
  yu <- log(dat$cd)

  su <- sqrt((dat$se / dat$cd) ^ 2 + sigma_m ^ 2)

  # calculate starting values and weights
  sigma_d <- sd(dat$cd) / mean(dat$cd)

  ## sigma_d may be NA if there is only one measurement (NA values in the
  ## data have already been excluded prior to this)
  if (is.na(sigma_d))
    sigma_d <- 0

  wu <- 1 / (sigma_d ^ 2 + su ^ 2)

  delta <- mean(dat$cd)
  n <- length(yu)

  ##terminal output
  if (verbose) {
    cat("\n[calc_AverageDose()]")
    cat("\n\n>> Initialisation <<")
    cat("\nn:\t\t", n)
    cat("\ndelta:\t\t", delta)
    cat("\nsigma_m:\t", sigma_m)
    cat("\nsigma_d:\t", sigma_d)
  }


  # mle's computation
  dhat <- .mle(yu, su, wu.start = wu, sigma_d.start = sigma_d, delta.start = delta)
  delta <- dhat[1]
  sigma_d <- dhat[2]
  wu <- 1 / (sigma_d ^ 2 + su ^ 2)

  # maximum log likelihood
  llik <- sum(-log(sqrt(2 * pi / wu)) - (wu / 2) * ((yu - log(delta) + 0.5 * (sigma_d ^ 2)) ^ 2))

  ##terminal output
  if(verbose){
    cat("\n\n>> Calculation <<\n")
    cat("log likelihood:\t", round(llik, 4), "\n")
  }

  if (is.na(llik)) {
    .throw_error("Maximum likelihood estimation failed")
  }

  # standard errors obtained by bootstrap, we refer to Efron B. and Tibshirani R. (1986)
  n <- length(yu)

  ##calculate dstar
  ##set matrix for I
  I <- matrix(data = sample(x = 1:n, size = n * Nb_BE, replace = TRUE), ncol = Nb_BE)

  ##iterate over the matrix and produce dstar
  ##(this looks a little bit complicated, but is far more efficient)
  dstar <- t(vapply(
    X = 1:Nb_BE,
    FUN = function(x) {
      .mle(yu[I[, x]], su[I[, x]], sigma_d.start = sigma_d, delta.start = delta, wu.start = wu)
    },
    FUN.VALUE = numeric(2)))

  ##exclude NA values
  dstar <- na.exclude(dstar)

  ## calculate confidence intervals
  IC_delta <- .CredibleInterval(dstar[,1],0.95)
  IC_sigma_d <- .CredibleInterval(dstar[,2],0.95)
  IC <- rbind(IC_delta, IC_sigma_d)

  # standard errors
  sedelta <- sqrt((1 / (Nb_BE - 1)) * sum((dstar[, 1] - mean(dstar[, 1]))^2))
  sesigma_d <- sqrt((1 / (Nb_BE - 1)) * sum((dstar[, 2] - mean(dstar[, 2]))^2))

  ##Terminal output
  if (verbose) {
    cat("\nconfidence intervals\n")
    cat("--------------------------------------------------\n")
    print(t(IC), print.gap = 6, digits = 4)
    cat("--------------------------------------------------\n")

    cat("\n>> Results <<\n")
    cat("----------------------------------------------------------\n")
    cat(
      "Average dose:\t ",
      round(delta, 4),
      "\tse(Aver. dose):\t",
      round(sedelta, 4)
    )
    cat("\nsigma_d:\t ",
        round(sigma_d, 4),
        if (sigma_d == 0) "\t",
        "\tse(sigma_d):\t",
        round(sesigma_d, 4)
        )
    cat("\n----------------------------------------------------------\n")
  }

  ##compile final results data frame
  results_df <- data.frame(
    AVERAGE_DOSE = delta,
    AVERAGE_DOSE.SE = sedelta,
    SIGMA_D = sigma_d,
    SIGMA_D.SE = sesigma_d,
    IC_AVERAGE_DOSE.LEVEL = IC_delta[1],
    IC_AVERAGE_DOSE.LOWER = IC_delta[2],
    IC_AVERAGE_DOSE.UPPER = IC_delta[3],
    IC_SIGMA_D.LEVEL = IC_sigma_d[1],
    IC_SIGMA_D.LOWER = IC_sigma_d[2],
    IC_SIGMA_D.UPPER = IC_sigma_d[3],
    L_MAX = llik,
    de = delta,
    de_err = sedelta,
    row.names = NULL
  )

  # Plotting ------------------------------------------------------------------------------------

  ##the plotting (enable/disable) is controlled below, as with this
  ##we always get a histogram object

  ##set data list
  data_list <- list(dat$cd, dstar[,1], dstar[,2])

  ##preset plot arguments
  plot_settings <- list(
    breaks = list("FD", "FD", "FD"),
    probability = list(FALSE, TRUE, TRUE),
    main = list(
      "Observed: Equivalent dose",
      "Bootstrapping: Average Dose",
      "Bootstrapping: Sigma_d"),
    xlab = list(
      "Equivalent dose [a.u.]",
      "Average dose [a.u.]",
      "Sigma_d"),
    axes = list(TRUE, TRUE, TRUE),
    col = NULL,
    border = NULL,
    density = NULL,
    freq = NULL,
    mtext = list(
      paste("n = ", length(data_list[[1]])),
      paste("n = ", length(data_list[[2]])),
      paste("n = ", length(data_list[[3]]))),
    rug = list(TRUE, TRUE, TRUE)
  )

  ##modify this list by values the user provides

  ##expand all elements in the list
  ##problem: the user might provid only one item, then the code will break
  plot_settings.user <- lapply(list(...), rep, length = 3)

  ##modify
  plot_settings <- modifyList(plot_settings, plot_settings.user)

  ##get change par setting and reset on exit
  if(plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)
    par(mfrow = c(1,3))
  }

  ##Produce plots
  ##(1) - histogram of the observed equivalent dose
  ##(2) - histogram of the bootstrapped De
  ##(3) - histogram of the bootstrapped sigma_d

  ##with lapply we get fetch also the return of hist, they user might want to use this later
  hist <- lapply(1:length(data_list), function(x){
    temp <- suppressWarnings(hist(
      x = data_list[[x]],
      breaks = if (NROW(data_list[[x]]) > 1) plot_settings$breaks[[x]] else 1,
      probability = plot_settings$probability[[x]],
      main = plot_settings$main[[x]],
      xlab = plot_settings$xlab[[x]],
      axes = plot_settings$axes[[x]],
      freq = plot_settings$freq[[x]],
      plot = plot,
      col = plot_settings$col[[x]],
      border = plot_settings$border[[x]],
      density = plot_settings$density[[x]]
    ))

    if (plot) {
      ##add rug
      if (plot_settings$rug[[x]]) {
        graphics::rug(data_list[[x]])
      }

      ##plot mtext
      mtext(side = 3,
            text = plot_settings$mtext[[x]],
            cex = par()$cex)
    }

    return(temp)
  })

  # Return --------------------------------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      summary = results_df,
      data = data,
      dstar = as.data.frame(dstar),
      hist = hist
    ),
    info = list(call = sys.call())
  )
}

## load data
data(ExampleData.DeValues, envir = environment())
set.seed(1)
ADM <- calc_AverageDose(ExampleData.DeValues$CA1[1:56, ], sigma_m = 0.1,
                        plot = FALSE, verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  data <- ExampleData.DeValues$CA1
  expect_error(calc_AverageDose(),
               "'data' should be of class 'RLum.Results' or 'data.frame'")
  expect_error(calc_AverageDose("test"),
               "'data' should be of class 'RLum.Results' or 'data.frame'")
  expect_error(calc_AverageDose(data),
               "'sigma_m' should be a single positive value")
  expect_error(calc_AverageDose(data, sigma_m = NULL),
               "'sigma_m' should be a single positive value")
  expect_error(calc_AverageDose(data, sigma_m = 0.1, Nb_BE = NULL),
               "'Nb_BE' should be a single positive integer value")
  expect_message(expect_null(
      calc_AverageDose(data[, 1, drop = FALSE], sigma_m = 0.1)),
      "Error: 'data' contains < 2 columns")
  expect_message(expect_null(
      calc_AverageDose(data[0, ], sigma_m = 0.1)),
      "Error: 'data' contains no data, NULL returned")

  SW({
  expect_warning(calc_AverageDose(cbind(data, data), sigma_m = 0.1),
                 "'data' contains > 2 columns")
  expect_warning(calc_AverageDose(rbind(data, NA), sigma_m = 0.1),
                 "NA values in 'data' detected")
  expect_message(expect_null(
      calc_AverageDose(data.frame(NA, NA), sigma_m = 0.1)),
      "Error: After NA removal, nothing is left from the data set")
  expect_warning(expect_error(
      calc_AverageDose(data * c(1e-12, 1e12), sigma_m = 0.9),
      "Maximum likelihood estimation failed"),
      "Inf/NaN values produced by .mle(), NA returned", fixed = TRUE)
  })
})

test_that("check class and length of output", {
  testthat::skip_on_cran()

  set.seed(1)
  expect_snapshot_RLum(ADM,
                       tolerance = 1.5e-6)
  expect_output(
      calc_AverageDose(ExampleData.DeValues$CA1[1:56, ],
                       sigma_m = 0.9,
                       plot = FALSE,
                       verbose = TRUE)
  )

  results <- get_RLum(ADM)

  expect_equal(round(results$AVERAGE_DOSE, digits = 4), 65.3597)
  expect_equal(round(results$SIGMA_D, digits = 4), 0.3092)
  expect_equal(round(results$L_MAX, digits = 5), -19.25096)

  ## RLum.Results
  expect_warning(calc_AverageDose(ADM, sigma_m = 0.1, verbose = FALSE, col = 1),
                 "'data' contains > 2 columns, only the first 2 columns were used")

  ## non-positive values
  data.zero <- ExampleData.DeValues$CA1
  data.zero[1, 1] <- 0
  expect_warning(calc_AverageDose(data.zero, sigma_m = 0.1, verbose = FALSE),
                 "Non-positive values in 'data' detected, rows removed")

  ## iteration limit reached
  SW({
  set.seed(1)
  expect_warning(calc_AverageDose(ExampleData.DeValues$CA1 * c(1e-2, 1e2),
                                  sigma_m = 0.9),
                 "No convergence reached by .mle() after 10000 iterations",
                 fixed = TRUE)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("abanico",
                              plot_AbanicoPlot(data = ADM, z.0 = ADM$summary$de))
  })
})

#' @title Function to calculate statistic measures
#'
#' @description This function calculates a number of descriptive statistics for estimates
#' with a given standard error (SE), most fundamentally using error-weighted approaches.
#'
#' @details The option to use Monte Carlo Methods (`n.MCM`) allows calculating
#' all descriptive statistics based on random values. The distribution of these
#' random values is based on the Normal distribution with `De` values as
#' means and `De_error` values as one standard deviation. Increasing the
#' number of MCM-samples linearly increases computation time. On a Lenovo X230
#' machine evaluation of 25 Aliquots with n.MCM = 1000 takes 0.01 s, with
#' n = 100000, ca. 1.65 s. It might be useful to work with logarithms of these
#' values. See Dietze et al. (2016, Quaternary Geochronology) and the function
#' [Luminescence::plot_AbanicoPlot] for details.
#'
#' @param data [data.frame] or [Luminescence::RLum.Results-class] object (**required**):
#' for [data.frame] two columns: De (`data[, 1]`) and De error (`data[, 2]`).
#'
#' @param weight.calc [character] (*with default*):
#' type of weight calculation. One out of `"reciprocal"` (weight is 1/error),
#' `"square"` (weight is 1/error^2). Default is `"square"`.
#'
#' @param digits [integer] (*with default*):
#' number of decimal places to be used when rounding numbers. If set to `NULL`
#' (default), no rounding occurs.
#'
#' @param n.MCM [numeric] (*with default*):
#' number of samples drawn for Monte Carlo-based statistics.
#' `NULL` (the default) disables MC runs.
#'
#' @param na.rm [logical] (*with default*):
#' indicating whether `NA` values should be stripped before the computation proceeds.
#'
#' @return Returns a list with weighted and unweighted statistic measures.
#'
#' @section Function version: 0.1.7
#'
#' @keywords datagen
#'
#' @author Michael Dietze, GFZ Potsdam (Germany)
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## show a rough plot of the data to illustrate the non-normal distribution
#' plot_KDE(ExampleData.DeValues$BT998)
#'
#' ## calculate statistics and show output
#' str(calc_Statistics(ExampleData.DeValues$BT998))
#'
#' \dontrun{
#' ## now the same for 10000 normal distributed random numbers with equal errors
#' x <- as.data.frame(cbind(rnorm(n = 10^5, mean = 0, sd = 1),
#'                          rep(0.001, 10^5)))
#'
#' ## note the congruent results for weighted and unweighted measures
#' str(calc_Statistics(x))
#' }
#'
#' @export
calc_Statistics <- function(
  data,
  weight.calc = c("square", "reciprocal"),
  digits = NULL,
  n.MCM = NULL,
  na.rm = TRUE
) {
  .set_function_name("calc_Statistics")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("RLum.Results", "data.frame"))
  .validate_not_empty(data)
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")[, 1:2]
  }

  ##strip na values
  .validate_logical_scalar(na.rm)
  if(na.rm){
    data <- na.exclude(data)
    if (nrow(data) == 0) {
      .throw_error("'data' contains only NA values")
    }
  }

  ## handle error-free data sets
  if(ncol(data) == 1) {
    data <- cbind(data, rep(NA, length(data)))
  }

  ## replace Na values in error by 0
  data[is.na(data[,2]),2] <- 0

  ## replace zeros by a small value
  if (any(data[, 2] == 0)) {
    if (sum(data[, 2]) == 0) {
      .throw_warning("All errors are NA or zero, automatically set to 10^-9")
    }
    data[,2] <- rep(x = 10^-9, length(data[,2]))
  }

  weight.calc <- .validate_args(weight.calc, c("square", "reciprocal"))
  if(weight.calc == "reciprocal") {
    S.weights <- 1 / data[,2]
  } else if(weight.calc == "square") {
    S.weights <- 1 / data[,2]^2
  }

  S.weights <- S.weights / sum(S.weights)

  .validate_positive_scalar(digits, int = TRUE, null.ok = TRUE)
  .validate_positive_scalar(n.MCM, int = TRUE, null.ok = TRUE)

  ## create MCM data
  if (is.null(n.MCM)) {
    data.MCM <- cbind(data[, 1])
  } else {
    data.MCM <-
      matrix(data = rnorm(
        n = n.MCM * nrow(data),
        mean = data[, 1],
        sd = data[, 2]
      ),
      ncol = n.MCM)
  }

  ## calculate n
  S.n <- nrow(data)
  S.m.n <- S.n * ncol(data.MCM)

  ## calculate mean
  S.mean <- mean(x = data[,1],
                 na.rm = na.rm)

  S.wg.mean <- stats::weighted.mean(x = data[,1],
                                    w = S.weights,
                                    n.rm = na.rm)

  S.m.mean <- mean(x = data.MCM,
                   na.rm = na.rm)

  ## calculate median
  S.median <- median(x = data[,1],
                     na.rm = na.rm)

  S.wg.median <- .weighted.median(data[, 1],
                                  w = S.weights,
                                  na.rm = na.rm)

  S.m.median <- median(x = data.MCM,
                       na.rm = na.rm)

  ## calculate absolute standard deviation
  S.sd.abs <- sd(x = data[,1],
                 na.rm = na.rm)

  S.wg.sd.abs <- sqrt(sum(S.weights * (data[,1] - S.wg.mean)^2) /
                        (((S.n - 1) * sum(S.weights)) / S.n))

  S.m.sd.abs <- sd(x = data.MCM,
                   na.rm = na.rm)

  ## calculate relative standard deviation
  S.sd.rel <- S.sd.abs / S.mean * 100

  S.wg.sd.rel <- S.wg.sd.abs / S.wg.mean * 100

  S.m.sd.rel <- S.m.sd.abs / S.m.mean * 100

  ## calculate absolute standard error of the mean
  S.se.abs <- S.sd.abs / sqrt(S.n)

  S.wg.se.abs <- S.wg.sd.abs / sqrt(S.n)

  S.m.se.abs <- S.m.sd.abs / sqrt(S.n)

  ## calculate relative standard error of the mean
  S.se.rel <- S.se.abs / S.mean * 100

  S.wg.se.rel <- S.wg.se.abs / S.wg.mean * 100

  S.m.se.rel <- S.m.se.abs / S.m.mean * 100

  ## calculate skewness
  S.skewness <- 1 / S.n * sum(((data[,1] - S.mean) / S.sd.abs)^3)

  S.m.skewness <- 1 / S.m.n * sum(((data.MCM - S.m.mean) / S.m.sd.abs)^3)

  ## calculate kurtosis
  S.kurtosis <- 1 / S.n * sum(((data[,1] - S.mean) / S.sd.abs)^4)

  S.m.kurtosis <- 1 / S.m.n * sum(((data.MCM - S.m.mean) / S.m.sd.abs)^4)

  ## create list objects of calculation output
  S.weighted <- list(n = S.n,
                     mean = S.wg.mean,
                     median = S.wg.median,
                     sd.abs = S.wg.sd.abs,
                     sd.rel = S.wg.sd.rel,
                     se.abs = S.wg.se.abs,
                     se.rel = S.wg.se.rel,
                     skewness = S.skewness,
                     kurtosis = S.kurtosis)


  if(!is.null(digits)) {

     S.weighted <- sapply(names(S.weighted),
                          simplify = FALSE,
                          USE.NAMES = TRUE,
                          function(x) {
                            round(S.weighted[[x]],
                                  digits = digits)})
  }

  S.unweighted <- list(n = S.n,
                       mean = S.mean,
                       median = S.median,
                       sd.abs = S.sd.abs,
                       sd.rel = S.sd.rel,
                       se.abs = S.se.abs,
                       se.rel = S.se.rel,
                       skewness = S.skewness,
                       kurtosis = S.kurtosis)

  if(!is.null(digits)){
    S.unweighted  <- sapply(names(S.unweighted),
                            simplify = FALSE,
                            USE.NAMES = TRUE,
                            function(x) {
                              round(S.unweighted [[x]],
                                    digits = digits)})
  }

  S.MCM <- list(n = S.n,
                mean = S.m.mean,
                median = S.m.median,
                sd.abs = S.m.sd.abs,
                sd.rel = S.m.sd.rel,
                se.abs = S.m.se.abs,
                se.rel = S.m.se.rel,
                skewness = S.m.skewness,
                kurtosis = S.m.kurtosis)

  if(!is.null(digits)){

    S.MCM  <- sapply(names(S.MCM),
                     simplify = FALSE,
                     USE.NAMES = TRUE,
                     function(x) {
                       round(S.MCM [[x]],
                             digits = digits)})
  }

  list(weighted = S.weighted,
       unweighted = S.unweighted,
       MCM = S.MCM)
}

## load data
data(ExampleData.BINfileData, envir = environment())
CWOSL.sub <- subset(CWOSL.SAR.Data,
                    subset = POSITION %in% c(1:3) & LTYPE == "OSL")

test_that("input validation", {
  skip_on_cran()

  expect_error(analyse_baSAR("error",
                             signal_integral = 1:2, background_integral = 80:100,
                             verbose = FALSE),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_output(expect_null(analyse_baSAR(list("error"),
                             signal_integral = 1:2, background_integral = 80:100,
                             verbose = FALSE)),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_error(analyse_baSAR(data.frame(), verbose = FALSE),
               "'object' should be of class 'Risoe.BINfileData', 'RLum.Results'")
  expect_error(analyse_baSAR(character(0), verbose = FALSE),
               "'object' cannot be an empty character")
  expect_error(analyse_baSAR(list(data.frame()),
                             signal_integral = 1:2, background_integral = 80:100,
                             verbose = FALSE),
               "All elements of 'object' should be of class 'Risoe.BINfileData'")
  expect_error(analyse_baSAR(list(CWOSL.sub, "error"),
                             signal_integral = 1:2, background_integral = 80:100,
                             verbose = FALSE),
               "'object' only accepts a list of objects of the same type")
  expect_error(analyse_baSAR(CWOSL.sub,
                             signal_integral = 1:2, background_integral = 80:100,
                             n.MCMC = NULL),
               "'n.MCMC' should be a single positive integer value")
  expect_error(analyse_baSAR(CWOSL.sub,
                             signal_integral = 1:2, background_integral = 80:100,
                             verbose = FALSE),
               "'source_doserate' is missing, but the current implementation")
  expect_error(analyse_baSAR(CWOSL.sub,
                             signal_integral = 1:2, background_integral = 80:100,
                             fit.method = "error"),
               "'fit.method' should be one of 'EXP', 'EXP+LIN' or 'LIN'",
               fixed = TRUE)
  expect_error(analyse_baSAR(CWOSL.sub,
                             background_integral = 80:100,
                             verbose = FALSE,
                             source_doserate = c(0.04, 0.001)),
               "'signal_integral' should be of class 'integer', 'numeric' or 'list'")
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2),
               "'background_integral' should be of class 'integer', 'numeric' or 'list'")

  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             signal_integral = 1:2, background_integral = 70:90,
                             method_control = list(lower_centralD = -2)),
               "'lower_centralD' in 'method_control' should be a single non-negative")
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             signal_integral = 1:2, background_integral = 70:90,
                             method_control = list(upper_centralD = list())),
               "'upper_centralD' in 'method_control' should be a single non-negative")
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             signal_integral = 1:2,
                             background_integral = 70:90,
                             distribution = "user_defined",
                             method_control = list(lower_centralD = 2000)),
               "'upper_centralD' in 'method_control' must be greater")

  ## CSV_file
  csv.file <- tempfile()
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             CSV_file = list()),
               "'CSV_file' should be of class 'data.frame', 'character' or NULL")
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             CSV_file = "error"),
               "does not exist or is non-readable")
  data.table::fwrite(data.frame(BIN_file = "a", DISC = 1), file = csv.file)
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             CSV_file = csv.file),
               "'CSV_file' should have at least 3 columns for the name of the")
  data.table::fwrite(data.frame(a = "error", b = 1, c = 2), file = csv.file)
  expect_warning(expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             CSV_file = csv.file),
                             "The BIN-file names provided via 'CSV_file' do not"),
               "File 'error' doesn't match, skipped")
  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             CSV_file = data.frame(a = NA, b = NA)),
               "'CSV_file' should have at least 3 columns for the name of the")
  expect_warning(expect_error(
                 analyse_baSAR(CWOSL.sub, verbose = FALSE,
                               sigmab = list(0.23), sig0 = list(0.02),
                               source_doserate = list(0.04, 0.001),
                               signal_integral = list(1, 2),
                               signal_integral_Tx = 2:4,
                               background_integral = list(80, 100),
                               background_integral_Tx = 80:100,
                               CSV_file = data.frame(a = "error", b = 1, c = 2)),
                 "BIN-file names provided via 'CSV_file' do not match the loaded BIN-files",
                 fixed = TRUE),
  "File 'error' doesn't match, skipped")

  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             source_doserate = c(0.04, 0.001),
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             CSV_file = data.frame(a = NA, b = 1, c = 2)),
               "Number of discs/grains = 0")

  SW({
  obj <- Risoe.BINfileData2RLum.Analysis(CWOSL.sub)
  expect_error(suppressWarnings(
      analyse_baSAR(obj, signal_integral = 1:2, background_integral = 80:100,
                    verbose = TRUE)),
      "No records of the appropriate type found")
  expect_error(suppressWarnings(
      analyse_baSAR(obj, signal_integral = 1:2, background_integral = 80:100,
                    recordType = "NONE", verbose = TRUE)),
    "No records of the appropriate type found")

  empty <- CWOSL.sub
  empty@METADATA <- empty@METADATA[integer(0), ]
  expect_error(expect_warning(
      analyse_baSAR(list(empty),
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100),
      "No data selected from BIN-file 1, BIN-file removed from input"),
      "All provided objects were removed")
  })

  expect_error(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                             signal_integral = 1:2,
                             background_integral = 80:100,
                             distribution = "error"),
               "'distribution' should be one of 'cauchy', 'normal', 'log_normal' or")

  expect_warning(expect_output(
      analyse_baSAR(CWOSL.sub, verbose = FALSE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    distribution = "user_defined"),
      "[analyse_baSAR()] You specified a 'user_defined' distribution",
      fixed = TRUE),
      "Only multiple grain data provided, automatic selection skipped")

  expect_message(expect_output(suppressWarnings(
      analyse_baSAR(CWOSL.sub, verbose = FALSE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    distribution = "cauchy",
                    baSAR_model = "error")),
      "Error parsing model file"),
      "'baSAR_model' provided, setting distribution to 'user_defined'")

  expect_error(suppressWarnings(
      analyse_baSAR(CWOSL.SAR.Data, verbose = FALSE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    distribution = "user_defined")),
      "Different number of channels for Lx (250) and Tx (1000)",
      fixed = TRUE)

  SW({
  data(ExampleData.RLum.Analysis, envir = environment())
  expect_error(analyse_baSAR(list(IRSAR.RF.Data),
                             signal_integral = 1:2, background_integral = 80:100,
                             verbose = TRUE),
               "At least two aliquots are needed for the calculation")

  expect_warning(expect_output(
      analyse_baSAR(list(CWOSL.sub, CWOSL.sub), verbose = TRUE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    fit.method = "LIN", fit.force_through_origin = FALSE,
                    distribution = "normal",
                    n.MCMC = 75)),
      "'ExampleData.BINfileData' is a duplicate and therefore removed from the input")

  CWOSL.min <- subset(CWOSL.sub, subset = ID < 20)
  expect_warning(expect_error(
      analyse_baSAR(CWOSL.min, source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 5:15,
                    method_control = list(n.chains = 1),
                    n.MCMC = 10),
      "In input 1 the number of data points (19) is not a multiple of the",
      fixed = TRUE),
      "Only multiple grain data provided, automatic selection skipped")

  CWOSL.sub2 <- CWOSL.sub
  CWOSL.sub2@METADATA$FNAME <- "Other file"
  expect_warning(analyse_baSAR(list(CWOSL.sub, CWOSL.sub2),
                               source_doserate = list(c(0.04, 0.001), c(0.05, 0.02)),
                               signal_integral = 1:2,
                               background_integral = 5:15,
                               method_control = list(n.chains = 1),
                               n.MCMC = 10),
                 "Provided source dose rate errors differ")
  })

  CWOSL.mod <- CWOSL.sub
  CWOSL.mod@METADATA$POSITION[c(1:nrow(CWOSL.mod@METADATA))] <- 2.5
  expect_warning(expect_message(expect_null(
      analyse_baSAR(CWOSL.mod, verbose = FALSE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100)),
      "position number 2 does not exist, NULL returned"),
      "Only multiple grain data provided, automatic selection skipped")

  CWOSL.mod <- CWOSL.sub
  CWOSL.mod@METADATA$GRAIN[c(1:nrow(CWOSL.mod@METADATA))] <- 2
  expect_warning(expect_message(expect_null(
      analyse_baSAR(CWOSL.mod, verbose = FALSE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100)),
      "grain number 0 does not exist, NULL returned"),
      "Only multiple grain data provided, automatic selection skipped")
})

test_that("Full check of analyse_baSAR function", {
  skip_on_cran()

    set.seed(1)
    ##(3) run analysis
    ##please not that the here selected parameters are
    ##chosen for performance, not for reliability
    SW({
    expect_snapshot_RLum(results <- suppressWarnings(analyse_baSAR(
      object = CWOSL.sub,
      source_doserate = c(0.04, 0.001),
      signal_integral = 1:2,
      background_integral = 80:100,
      fit.method = "EXP",
      method_control = list(inits = list(
        list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 1),
        list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 2),
        list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 3)
      )),
      plot = TRUE,
      verbose = TRUE,
      n.MCMC = 100,
      txtProgressBar = TRUE
      )),
      tolerance = 1.5e-6)
    })

    ## source dose rate only scalar
    SW({
      expect_snapshot_RLum(results <- suppressWarnings(analyse_baSAR(
        object = CWOSL.sub,
        source_doserate = c(0.04),
        signal_integral = 1:2,
        background_integral = 80:100,
        fit.method = "EXP",
        method_control = list(inits = list(
          list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 1),
          list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 2),
          list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 3)
        )),
        plot = TRUE,
        verbose = TRUE,
        n.MCMC = 100,
        txtProgressBar = TRUE
      )),
      tolerance = 1.5e-6)
    })

  ## rerun with previous results as input
  SW({
  expect_output(results2 <- analyse_baSAR(
     object = results,
      plot = FALSE,
      verbose = TRUE,
      txtProgressBar = FALSE,
      n.MCMC = 100),
     "Number of aliquots used:.+3/3")

  expect_output(analyse_baSAR(
      object = results,
      plot = FALSE,
      verbose = FALSE,
      txtProgressBar = FALSE,
      baSAR_model = "empty",
      n.MCMC = 10),
      "Error parsing model file")

  expect_warning(analyse_baSAR(
      object = results,
      plot = FALSE,
      verbose = FALSE,
      txtProgressBar = FALSE,
      method_control = list(lower_centralD = 0),
      n.MCMC = 100),
      "You have modified the lower central_D boundary")

  suppressWarnings(expect_warning(analyse_baSAR(
      object = results,
      plot = FALSE,
      verbose = FALSE,
      txtProgressBar = FALSE,
      method_control = list(upper_centralD = 2),
      n.MCMC = 100),
      "Your lower_centralD and/or upper_centralD values seem not to fit",
      fixed = TRUE))

  expect_warning(expect_warning(
    analyse_baSAR(
      object = results,
      plot = TRUE,
      verbose = TRUE,
      txtProgressBar = FALSE,
      fit.method = "EXP+LIN",
      fit.includingRepeatedRegPoints = FALSE,
      fit.force_through_origin = FALSE,
      distribution = "log_normal",
      aliquot_range = 1:3,
      distribution_plot = "abanico",
      method_control = list(n.chains = 2, thin = 25),
      plot.single = TRUE,
      output.plot = FALSE,
      output.plotExtended = FALSE,
      n.MCMC = 100),
    "'output.plotExtended' was deprecated in v1.2.0, use 'plot_extended' instead"),
    "'plot.single' was deprecated in v1.0.0, use 'plot_singlePanels' instead")

  expect_message(
      analyse_baSAR(
          object = results,
          plot = FALSE,
          verbose = TRUE,
          txtProgressBar = FALSE,
          fit.method = "EXP",
          fit.force_through_origin = TRUE,
          distribution = "cauchy",
          aliquot_range = 100:300,
          distribution_plot = NULL,
          method_control = list(n.chains = 1, thin = 25),
          n.MCMC = 100),
      "Error: 'aliquot_range' out of bounds, input ignored")
  expect_message(
      analyse_baSAR(
          object = results,
          plot = FALSE,
          verbose = TRUE,
          txtProgressBar = FALSE,
          fit.method = "EXP",
          fit.force_through_origin = TRUE,
          distribution = "cauchy",
          aliquot_range = -1:3,
          distribution_plot = NULL,
          method_control = list(n.chains = 1, thin = 25),
          n.MCMC = 100),
      "Error: 'aliquot_range' out of bounds, input ignored")
  expect_error(analyse_baSAR(results, aliquot_range = NA),
               "'aliquot_range' should be of class 'integer', 'numeric' or NULL")

  expect_warning(expect_message(
      analyse_baSAR(
          object = results,
          plot = TRUE,
          verbose = TRUE,
          txtProgressBar = FALSE,
          source_doserate = 2,
          fit.method = "LIN",
          distribution = "log_normal",
          method_control = list(n.chains = 2, variable.names = "Q"),
          n.MCMC = 100),
      "Plots for 'central_D' and 'sigma_D' could not be produced",
      fixed = TRUE),
      "'source_doserate' is ignored in this mode as it was already set")

  ## filename as input
  expect_warning(expect_message(expect_null(
      analyse_baSAR(test_path("_data/BINfile_V3.bin"),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    source_doserate = 2,
                    fit.method = "LIN",
                    plot = FALSE,
                    verbose = FALSE,
                    method_control = list(n.chains = 1),
                    n.MCMC = 100)),
      "Error: All irradiation times are identical, NULL returned"),
      "Only multiple grain data provided, automatic selection skipped")

  ## more coverage
  SW({
  expect_warning(analyse_baSAR(
      object = CWOSL.sub,
      source_doserate = c(NaN, 0.001),
      signal_integral = 1:2,
      background_integral = 80:100,
      plot = FALSE,
      verbose = TRUE,
      n.MCMC = 100),
      "'Nb_aliquots' corrected due to NaN or Inf values")

  data(ExampleData.Al2O3C, envir = environment())
  expect_output(expect_warning(analyse_baSAR(
      object = list(data_CrossTalk[[1]][1:4], data_ITC),
      source_doserate = c(0.04, 0.001),
      signal_integral = 1:2,
      background_integral = 80:99,
      fit.method = "EXP",
      n.MCMC = 100, irradiation_times = c(1, 2, 3)),
      "The number of dose points differs across your data set"),
      "Error : [calc_Statistics()] 'data' cannot be an empty data.frame",
      fixed = TRUE)

  CWOSL.large <- subset(CWOSL.SAR.Data,
                        subset = POSITION %in% 1:18 & LTYPE == "OSL")
  analyse_baSAR(
        object = CWOSL.large,
        source_doserate = 0.04,
        signal_integral = 1:2,
        background_integral = 80:100,
        fit.method = "LIN",
        method_control = list(n.chains = 1),
        n.MCMC = 80,
        n.MC = 30, # for fit_DoseResponseCurve()
        plot = TRUE,
        plot_drc = FALSE,
        plot_extended = FALSE,
        verbose = FALSE)
  })

  CWOSL.mod <- CWOSL.sub
  CWOSL.mod@METADATA$SEL <- FALSE
  expect_warning(expect_null(analyse_baSAR(CWOSL.mod, verbose = TRUE,
                                           source_doserate = c(0.04, 0.001),
                                           signal_integral = 1:2,
                                           background_integral = 80:100,
                                           method_control = list(n.chains = 1),
                                           n.MCMC = 100)),
                 "No records selected, NULL returned")

  CWOSL.mod <- CWOSL.sub
  CWOSL.mod@METADATA$SEL[19:24] <- FALSE
  expect_message(analyse_baSAR(CWOSL.mod, verbose = TRUE,
                               source_doserate = c(0.04, 0.001),
                               signal_integral = 1:2,
                               background_integral = 80:100,
                               method_control = list(n.chains = 1),
                               n.MCMC = 100),
                 "Record pre-selection in BIN-file detected, record reduced to")

  CWOSL.mod <- CWOSL.sub
  CWOSL.mod@METADATA$GRAIN[-c(19:24)] <- 2
  expect_warning(analyse_baSAR(CWOSL.mod, verbose = TRUE,
                               source_doserate = c(0.04, 0.001),
                               signal_integral = 1:2,
                               background_integral = 80:100,
                               method_control = list(n.chains = 1),
                               n.MCMC = 100),
                 "Automatic grain selection: 3 curves with grain index 0 have been removed")
  })

  results2@originator <- "unknown"
  expect_error(analyse_baSAR(object = results2),
               "'object' has an unsupported originator (expected 'analyse_baSAR'",
               fixed = TRUE)

  results2@originator <- "analyse_baSAR"
  results2@data$input_object <- results2$input_object[1:2, ]
  expect_message(expect_null(analyse_baSAR(object = results2)),
                 "Error: Number of aliquots < 3, NULL returned")

  SW({
  expect_no_warning(analyse_baSAR(CWOSL.sub, source_doserate = c(0.04, 0.001),
                               signal_integral = 1:2,
                               background_integral = 8:10,
                               method_control = list(n.chains = 1),
                               n.MCMC = 10),
                 message = "Number of background channels for Tx < 25")

  df <- CWOSL.sub@METADATA[, c("FNAME", "POSITION", "GRAIN")]
  analyse_baSAR(CWOSL.sub,
                CSV_file = df,
                source_doserate = c(0.04, 0.001),
                signal_integral = 1:2,
                background_integral = 8:10,
                method_control = list(n.chains = 1),
                aliquot_range = 1:2,
                n.MCMC = 10)

  df$FNAME[3] <- NA
  analyse_baSAR(CWOSL.sub,
                CSV_file = df,
                source_doserate = c(0.04, 0.001),
                signal_integral = 1:2,
                background_integral = 8:40,
                method_control = list(n.chains = 1),
                aliquot_range = 1:4,
                n.MCMC = 10)

  vnames <- c("central_D", "sigma_D", "")
  expect_message(analyse_baSAR(CWOSL.sub,
                               source_doserate = c(0.04, 0.001),
                               signal_integral = 1:2,
                               background_integral = 60:100,
                               method_control = list(n.chains = 1,
                                                     variable.names = vnames),
                               plot_reduced = FALSE,
                               n.MCMC = 10, verbose = FALSE),
                 "Dose-response curves could not be plotted as 'variable.names'")

  vnames <- c("D", "Q", "a", "b", "c", "g")
  expect_message(analyse_baSAR(CWOSL.sub,
                               source_doserate = c(0.04, 0.001),
                               signal_integral = 1:2,
                               background_integral = 60:100,
                               method_control = list(n.chains = 1,
                                                     variable.names = vnames),
                               n.MCMC = 10, verbose = FALSE),
                 "Plots for 'central_D' and 'sigma_D' could not be produced")

  ## user model just for coverage
  expect_output(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                              source_doserate = c(0.04, 0.001),
                              signal_integral = 1:2,
                              background_integral = 80:100,
                              distribution = "cauchy",
                              baSAR_model =
                                "model {
                                  central_D ~ dunif(lower_centralD, upper_centralD)
                                  sigma_D ~ dunif(0.01, 1 * central_D)
                                  for (i in 1:Nb_aliquots) {
                                    a[i] ~ dnorm(6.5 , 1/(9.2^2))
                                    b[i] ~ dnorm(50 , 1/(1000^2))
                                    c[i] ~ dnorm(1.002 , 1/(0.9^2))
                                    g[i] ~ dnorm(0.5 , 1/(2.5^2))
                                    sigma_f[i] ~ dexp(20)
                                    D[i] ~ dnorm( central_D  + Dose[1, i], 1/(sigma_D^2))
                                    S_y[1,i] <- 1 / (sLum[1,i]^2 + sigma_f[i]^2)
                                    Lum[1,i] ~ dnorm( Q[1,i] , S_y[1,i])
                                    Q[1,i] <- GC_Origin * LinGC * ExpoGC * Limited_cycles[1]
                                  }
                                }",
                              method_control = list(n.chains = 1),
                              n.MCMC = 10))

  ## deprecated arguments
  expect_warning(analyse_baSAR(CWOSL.sub, verbose = FALSE,
                               source_doserate = c(0.04, 0.001),
                               signal.integral = 1:2,
                               background.integral = 80:100,
                               method_control = list(n.chains = 1, thin = 10),
                               n.MCMC = 60),
                 "deprecated in v1.2.0, use 'signal_integral', 'background_integral'")
  })
})

test_that("regression tests", {
  skip_on_cran()

  ## issue 407
  SW({
  expect_warning(expect_s4_class(
      analyse_baSAR(CWOSL.sub, verbose = FALSE, plot = FALSE,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    method_control = list(n.chains = 1, thin = 60),
                    n.MCMC = 60),
      "RLum.Results"),
      "'thin = 60' is too high for 'n.MCMC = 60', reset to 30")

  ## issue 834
  df <- CWOSL.sub@METADATA[, c("FNAME", "POSITION", "GRAIN")]
  expect_warning(expect_s4_class(
      analyse_baSAR(CWOSL.sub,
                    CSV_file = df,
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 8:40,
                    method_control = list(n.chains = 1),
                    aliquot_range = 1,
                    n.MCMC = 10),
      "RLum.Results"),
      "Adaptation incomplete")
  })

  ## issue 1426
  suppressMessages( # Error: All points have the same dose, NULL returned
  expect_warning(expect_warning(
      analyse_baSAR(CWOSL.sub, verbose = FALSE, plot = FALSE,
                    source_doserate = 0,
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    method_control = list(n.chains = 1),
                    n.MCMC = 60),
      "Only multiple grain data provided, automatic selection skipped"),
      "which may indicate an incorrect 'source_doserate'")
  )

  ## check parameters irradiation times
  SW({
  expect_s4_class(suppressWarnings(analyse_baSAR(CWOSL.sub,
                verbose = FALSE,
                plot = FALSE,
                source_doserate = c(0.04, 0.001),
                signal_integral = 1:2,
                irradiation_times = c(0, 0, 0, 0, 0, 0, 450, 450, 450, 0, 0, 0,
                                      1050, 1050, 1050, 0, 0, 0, 2000, 2000, 2000, 0, 0, 0,
                                      2550, 2550, 2550, 0, 0, 0, 450, 450,
                                      450, 0, 0, 0, 0, 0, 0, 0, 0, 0),
                background_integral = 80:100,
                n.MCMC = 10)), "RLum.Results")
  })

  ## have all irradiation times identical
  expect_message(
    suppressWarnings(analyse_baSAR(CWOSL.sub,
                verbose = FALSE,
                plot = FALSE,
                source_doserate = c(0.04, 0.001),
                signal_integral = 1:2,
                irradiation_times = c(0),
                background_integral = 80:100,
                n.MCMC = 10)),
     regexp = "Error: All irradiation times are identical, NULL returned")

  ## test removal of unwanted curves
  ## check irradiation times assignment and non-OSL curve removal
  tmp_object <- Risoe.BINfileData2RLum.Analysis(CWOSL.sub)
  tmp_curves <- tmp_object[[3]]@records[[1]]
  tmp_curves@recordType <- "NONE"
  tmp_object[[3]]@records <- c(tmp_object[[3]]@records, rep(tmp_curves,2))

  SW({
  expect_s4_class(
    suppressWarnings(
      analyse_baSAR(tmp_object,
                verbose = TRUE,
                plot = FALSE,
                recordType = c("OSL", "NONE"),
                source_doserate = c(0.04, 0.001),
                signal_integral = 1:2,
                background_integral = 80:100,
                n.MCMC = 10)), "RLum.Results")
  })

  tmp_object[[3]] <- set_RLum("RLum.Analysis")
  SW({
  expect_message(expect_null(
      analyse_baSAR(tmp_object,
                    verbose = FALSE,
                    recordType = c("OSL", "NONE"),
                    source_doserate = c(0.04, 0.001),
                    signal_integral = 1:2,
                    background_integral = 80:100,
                    n.MCMC = 10)),
      "Object conversion failed, NULL returned")
  })
})

## load example data
data(ExampleData.DeValues, envir = environment())

test_that("Test certain input scenarios", {
  testthat::skip_on_cran()

  temp_RLum <- set_RLum(class = "RLum.Results",
                        data = list(data = ExampleData.DeValues$BT998))
  expect_type(calc_Statistics(temp_RLum), "list")

  df <- ExampleData.DeValues$BT998
  df[, 2] <- NULL
  expect_warning(calc_Statistics(df),
                 "All errors are NA or zero, automatically set to")

  df <- ExampleData.DeValues$BT998
  df[,2] <- 0
  expect_warning(calc_Statistics(df),
                 "All errors are NA or zero, automatically set to")

  df <- ExampleData.DeValues$BT998
  expect_silent(calc_Statistics(df, weight.calc = "reciprocal"))
})

test_that("input validation", {
  testthat::skip_on_cran()

  df <- ExampleData.DeValues$BT998

  expect_error(calc_Statistics(data = matrix(0,2)),
               "[calc_Statistics()] 'data' should be of class 'RLum.Results' or 'data.frame'",
               fixed = TRUE)
  expect_error(calc_Statistics(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(calc_Statistics(data.frame(a = NA)),
               "'data' contains only NA values")
  expect_error(calc_Statistics(data = df, weight.calc = "error"),
               "'weight.calc' should be one of 'square' or 'reciprocal'")
  expect_error(calc_Statistics(df, digits = 2.4),
               "'digits' should be a single positive integer value")
  expect_error(calc_Statistics(df, n.MCM = "error"),
               "'n.MCM' should be a single positive integer value")
  expect_error(calc_Statistics(df, na.rm = iris),
               "'na.rm' should be a single logical value")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.5e-6

  expect_snapshot_plain(calc_Statistics(
      ExampleData.DeValues$BT998, n.MCM = 1000),
      tolerance = snapshot.tolerance)

  expect_snapshot_plain(calc_Statistics(
      ExampleData.DeValues$BT998, n.MCM = 1000, digits = 2),
      tolerance = snapshot.tolerance)

  expect_snapshot_plain(calc_Statistics(data.frame(1:10, 0:9)),
                        tolerance = snapshot.tolerance)
})

## bin_RLum.Data ------------------------------------------------------------
#' @title Channel binning for RLum.Data-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum.Data-class] objects. Depending on the input object, the corresponding
#' function will be selected.
#'
#' @param object [Luminescence::RLum.Data-class] (**required**):
#' S4 object of class `RLum.Data`
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' An object of the same type as the input provided after binning is applied.
#'
#' @section Function version: 0.2.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @note Currently only `RLum.Data` objects of class [Luminescence::RLum.Data.Curve-class]
#' and [Luminescence::RLum.Data.Spectrum-class] are supported.
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Data.Spectrum-class]
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ## create RLum.Data.Curve object from this example
#' curve <-
#'   set_RLum(
#'       class = "RLum.Data.Curve",
#'       recordType = "OSL",
#'       data = as.matrix(ExampleData.CW_OSL_Curve)
#'   )
#'
#' ## plot data without and with 2 and 4 channel binning
#' plot_RLum(curve)
#' plot_RLum(bin_RLum.Data(curve, bin_size = 2))
#' plot_RLum(bin_RLum.Data(curve, bin_size = 4))
#'
#' @keywords utilities
#'
#' @export
setGeneric("bin_RLum.Data", function(object, ...) {
  standardGeneric("bin_RLum.Data")
})


## get_RLum() ---------------------------------------------------------------
#' @title General accessor function for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the corresponding
#' function will be selected.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum` or an object of type [list] containing only objects
#' of type [Luminescence::RLum-class]
#'
#' @param ... further arguments passed to the specific class method.
#'
#' @return
#' An object of the same type as the input object provided.
#'
#' @section Function version: 0.3.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ## Example based using data and from the calc_CentralDose() function
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## apply the central dose model 1st time
#' temp1 <- calc_CentralDose(ExampleData.DeValues$CA1)
#'
#' ## get results and store them in a new object
#' temp.get <- get_RLum(object = temp1)
#'
#' @export
setGeneric("get_RLum", function(object, ...) {
  standardGeneric("get_RLum")
})

#' @describeIn get_RLum
#' Returns a list of [Luminescence::RLum-class] objects that had been passed to
#' [Luminescence::get_RLum]
#'
#' @param class [character] (*optional*):
#' define which class gets selected if applied to a list, e.g., if a list
#' consists of different type of [Luminescence::RLum-class] objects, this arguments allows
#' to make a selection. If nothing is provided, all [Luminescence::RLum-class] objects are
#' treated.
#'
#' @param null.rm [logical] (*with default*):
#' whether empty and `NULL` objects should be removed.
#'
#' @export
setMethod("get_RLum", signature = "list",
    function(object, class = NULL, null.rm = FALSE, ...) {
      ## input validation
      .validate_class(class, "character", null.ok = TRUE, length = 1)
      .validate_logical_scalar(null.rm)

      ## take care of the class argument
      if (!is.null(class)) {
        sel <- class[1] == vapply(object, function(x) class(x)[1], character(1))
        if (any(sel))
          object <- object[sel]
      }

      ## make remove all non-RLum objects
      selection <- lapply(seq_along(object), function(x) {
        ## get rid of all objects that are not of type RLum, this is better
        ## than leaving that to the user
        if (!inherits(object[[x]], what = "RLum")) {
          warning("[get_RLum()] object #", x, " in the list is not of class ",
                  "'RLum' and has been removed", call. = FALSE)
          return(NULL)
        }

          ## it might be the case the object already comes with empty objects,
          ## this would cause a crash
          if (inherits(object[[x]], "RLum.Analysis") &&
              length(object[[x]]@records) == 0)
            return(NULL)

          get_RLum(object[[x]], ...)
      })

      ## remove empty or NULL objects after the selection ... if wanted
      if (null.rm) {
        ## first set all empty objects to NULL ... for RLum.Analysis objects
        selection <- lapply(selection, function(x) {
          if (length(x) == 0 ||
              (inherits(x, "RLum.Analysis") && length(x@records) == 0))
            return(NULL)
          return(x)
        })
        ## get rid of all NULL objects
        selection <- selection[!vapply(selection, is.null, logical(1))]
      }
      return(selection)
    })

#' @describeIn get_RLum
#' Returns `NULL`.
#'
#' @export
setMethod("get_RLum", signature = "NULL",
    function(object, ...) {
      NULL
    })


## remove_RLum() ------------------------------------------------------------
#' @title Strips records from RLum-class objects
#'
#' @description
#' Remove records from an RLum-class object in a convenient way using
#' [Luminescence::get_RLum] for the selection.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' object with records to be removed.
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' An object of the same type as the input provided with records removed; it
#' can result in an empty object.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class]
#'
#' @examples
#' ## load example data
#' data(ExampleData.XSYG, envir = environment())
#' sar <- OSL.SARMeasurement$Sequence.Object[1:9]
#'
#' ## strop only OSL curves
#' sar <- remove_RLum(sar, recordType = "OSL")
#' sar
#'
#' @keywords utilities
#'
#' @export
setGeneric("remove_RLum", function(object, ...) {
  standardGeneric("remove_RLum")
})

#' @describeIn remove_RLum
#' Returns a list of [Luminescence::RLum-class] objects where the selected records are stripped
#'
#' @export
setMethod("remove_RLum", signature = "list",
          function(object, ...) {
            ## apply method in the objects and return the same
            tmp <- lapply(object, function(x) {
              if (inherits(x, "RLum.Analysis")) {
                return(remove_RLum(x,...))
              } else {
                return(x)
              }
            })

            ## remove empty elements
            tmp[lengths(tmp) > 0]
})

## length_RLum() ------------------------------------------------------------
#' @title Length retrieval function for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the corresponding
#' function will be selected.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @return
#' An [integer] indicating the length of the object.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso
#' [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Data.Spectrum-class],
#' [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#' @keywords utilities
#'
#' @export
setGeneric("length_RLum", function(object) {
  standardGeneric("length_RLum")
})


## melt_RLum() --------------------------------------------------------------
#' @title Melt RLum-class objects into a flat data.frame
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' A flat [data.frame].
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Analysis-class]
#'
#' @examples
#' data(ExampleData.XSYG, envir = environment())
#' melt_RLum(OSL.SARMeasurement[[2]][[1]])
#'
#' @keywords utilities
#'
#' @export
setGeneric("melt_RLum", function(object, ...) {
  standardGeneric("melt_RLum")
})

#' @describeIn melt_RLum
#' Returns a list of melted [Luminescence::RLum-class] objects; non-RLum objects are silently
#' removed.
#'
#' @export
setMethod("melt_RLum", signature = "list",
    function(object, ...) {
      ## silently remove non-RLum objects
      l <- .rm_nonRLum(object)

      ## just return NULL
      if (length(l) == 0)
        return(NULL)

      ## apply method in the objects and return the same
      l <- lapply(object, function(x) {
        t <- try(melt_RLum(x), silent = TRUE)

        if (inherits(t, "try-error"))
          return(NULL)
        else
          t
      })

      ## remove NULL
      l <- l[!vapply(l, is.null, logical(1))]

      ## now bind the data.frame
      as.data.frame(data.table::rbindlist(l))
    })


## add_metadata<-() ---------------------------------------------------------
#' @title Safe manipulation of object metadata
#'
#' @description
#' Generic functions for manipulation of metadata in [Luminescence::Risoe.BINfileData-class],
#' [Luminescence::RLum.Analysis-class] and [Luminescence::RLum.Data-class] objects.
#'
#' @param object [Luminescence::RLum.Analysis-class],
#' [Luminescence::Risoe.BINfileData-class] (**required**):
#' object of class `RLum.Analysis` or `Risoe.BINfileData` to manipulate.
#'
#' @param ... further arguments passed to the specific class method
#'
#' @param value (**required**):
#' value to be assigned to the selected metadata entry. A `NULL` value is
#' acceptable only for `replace_metadata`, in which case the elements named
#' in `info_element` will be removed.
#'
#' @author
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [Luminescence::RLum.Data-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::Risoe.BINfileData-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ## show data
#' CWOSL.SAR.Data
#'
#' ## add a new field
#' add_metadata(CWOSL.SAR.Data,
#'              info_element = "INSTITUTE") <- "Heidelberg University"
#'
#' ## rename a field
#' rename_metadata(CWOSL.SAR.Data,
#'                 info_element = "INSTITUTE") <- "INSTITUTION"
#'
#' ## replace all LTYPE to RSL
#' ## but only for the first position
#' replace_metadata(
#'  object = CWOSL.SAR.Data,
#'  info_element = "LTYPE",
#'  subset = (POSITION == 1)) <- "RSL"
#'
#' ## replacing a field with NULL allows to remove that field
#' replace_metadata(CWOSL.SAR.Data,
#'                  info_element = "PREVIOUS") <- NULL
#'
#' ## show the modified data
#' CWOSL.SAR.Data
#'
#' @rdname metadata
#' @export
setGeneric("add_metadata<-", function(object, ..., value) {
  standardGeneric("add_metadata<-")
})

#' @rdname metadata
#' @export
setGeneric("rename_metadata<-", function(object, ..., value) {
  standardGeneric("rename_metadata<-")
})

#' @rdname metadata
#' @export
setGeneric("replace_metadata<-", function(object, ..., value) {
  standardGeneric("replace_metadata<-")
})


## names_RLum() -------------------------------------------------------------
#' @title Name retrieval function for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the corresponding
#' function will be selected.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @return
#' A [character] vector.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#'
#' @keywords utilities
#'
#' @export
setGeneric("names_RLum", function(object) {
  standardGeneric("names_RLum")
})

#' @describeIn names_RLum
#' Returns a list of names of the [Luminescence::RLum-class] objects that had been passed to
#' it.
#'
#' @export
setMethod("names_RLum", signature = "list",
    function(object) {
      ## apply method in the objects and return the same
      lapply(object, function(x) {
        if (inherits(x, "RLum")) {
          return(names_RLum(x))
        } else {
          return(x)
        }
      })
    })


## replicate_RLum() ---------------------------------------------------------
#' @title General replication function for RLum-class objects
#'
#' @description
#' The function replicates [Luminescence::RLum-class] objects and returns a list of such
#' objects.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' an [Luminescence::RLum-class] object
#'
#' @param times [integer] (*optional*):
#' number for times each element should be repeated.
#'
#' @return
#' A [list] with the object repeated.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum-class]
#'
#' @keywords utilities
#'
#' @export
setGeneric("replicate_RLum", function(object, times = 1) {
   standardGeneric("replicate_RLum")
})


## set_Risoe.BINfileData() --------------------------------------------------
#' @title General setter function for Risoe.BINfileData objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::Risoe.BINfileData-class] objects. Depending on the input object, the
#' corresponding function will be selected.
#' Allowed arguments can be found in the documentations of the corresponding
#' [Luminescence::Risoe.BINfileData-class] class.
#'
#' @param METADATA x
#'
#' @param DATA x
#'
#' @param .RESERVED x
#'
#' @return
#' A [Luminescence::Risoe.BINfileData-class] object.
#'
#' @section Function version: 0.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData-class]
#'
#' @keywords utilities
#'
#' @export
setGeneric("set_Risoe.BINfileData", function(METADATA = data.frame(),
                                             DATA = list(), .RESERVED = list()) {
  standardGeneric("set_Risoe.BINfileData")
})


## set_RLum() ---------------------------------------------------------------
#' @title General setter function for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the given class, the corresponding
#' method to create an object from this class will be selected.
#'
#' @param class [character] (**required**):
#' name of the S4 class to create, must correspond to one of the [Luminescence::RLum-class]
#' classes.
#'
#' @param originator [character] (*automatic*):
#' contains the name of the calling function (the function that produces this object);
#' can be set manually.
#'
#' @param .uid [character] (*automatic*):
#' unique ID for this object, by default set using the internal C++ function
#' `create_UID`.
#'
#' @param .pid [character] (*with default*):
#' option to provide a parent id for nesting at will.
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' An object of the specified [Luminescence::RLum-class] class.
#'
#' @section Function version: 0.3.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ## produce empty objects from each class
#' set_RLum(class = "RLum.Data.Curve")
#' set_RLum(class = "RLum.Data.Spectrum")
#' set_RLum(class = "RLum.Data.Spectrum")
#' set_RLum(class = "RLum.Analysis")
#' set_RLum(class = "RLum.Results")
#'
#' ## produce a curve object with arbitrary curve values
#' object <- set_RLum(
#' class = "RLum.Data.Curve",
#' curveType = "arbitrary",
#' recordType = "OSL",
#' data = matrix(c(1:100,exp(-c(1:100))),ncol = 2))
#'
#' ## plot this curve object
#' plot_RLum(object)
#'
#' @export
setGeneric("set_RLum", function(class, originator, .uid = create_UID(),
                                .pid = NA_character_, ... ) {
  .set_function_name("set_RLum")
  on.exit(.unset_function_name(), add = TRUE)

  .validate_class(class, "character", length = 1)
  if (!nzchar(class)) {
    .throw_error("'class' cannot be an empty character")
  }
  class(class) <- as.character(class)

  if (missing(originator)) {
    if (is.language(sys.call(which = -1)[[1]])) {
      originator <- as.character(sys.call(which = -1)[[1]])

      ## account for calls using the double colons, in this case the vector
      ## is of length 3, not only 1
      if (length(originator) == 3) {
        originator <- originator[3]
      }

    } else {
      originator <- NA_character_
    }
  }

  standardGeneric("set_RLum")
})


## show() -------------------------------------------------------------------
#' @name show
#' @title Show the structure of RLum-class and Risoe.BINfileData-class objects
#'
#' @param object [Luminescence::RLum-class], [Luminescence::Risoe.BINfileData-class] (**required**):
#' object of class `RLum` or `Risoe.BINfileData`.
#'
NULL

## smooth_RLum() ------------------------------------------------------------
#' @title Smoothing of data for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the corresponding
#' function will be selected. The smoothing is performed in the internal
#' function `.smoothing()`.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' An object of the same type as the input object provided.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @note
#' Currently only `RLum` objects of class `RLum.Data.Curve` and `RLum.Analysis`
#' (with curve data) are supported.
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Analysis-class]
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ## create RLum.Data.Curve object from this example
#' curve <-
#'   set_RLum(
#'       class = "RLum.Data.Curve",
#'       recordType = "OSL",
#'       data = as.matrix(ExampleData.CW_OSL_Curve)
#'   )
#'
#' ## plot data without and with smoothing
#' plot_RLum(curve)
#' plot_RLum(smooth_RLum(curve))
#'
#' @keywords utilities
#'
#' @export
setGeneric("smooth_RLum", function(object, ...) {
  standardGeneric("smooth_RLum")
})

#' @describeIn smooth_RLum
#' Returns a list of [Luminescence::RLum-class] objects that had been passed to
#' [Luminescence::smooth_RLum]
#'
#' @export
setMethod("smooth_RLum", signature = "list",
    function(object, ...) {
      ## apply method in the objects and return the same
      lapply(object, function(x) {
        if (inherits(x, "RLum")) {
          return(smooth_RLum(x,...))
        } else {
          return(x)
        }
      })
    })

## normalise_RLum() ------------------------------------------------------------
#' @title Normalisation of RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the corresponding
#' function will be selected. The normalisation is performed in the internal
#' function `.normalise_curve()`.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @param norm [logical] [character] (**required**):
#' if logical, whether curve normalisation should occur; alternatively, one
#' of `"max"` (used with `TRUE`), `"min"`, `"first"`, `"last"`, `"huot"`,
#' `"intensity"` or a positive number (e.g., 2.2).
#'
#' @param ... further arguments passed to the specific class method
#'
#' @details
#' The `norm` argument normalises all count values. The following options are
#' supported:
#'
#' `norm = TRUE` or `norm = "max"`: Curve values are normalised to the highest
#' count value in the curve
#'
#' `norm = "min"`: Curve values are normalised to the smallest count value
#' in the curve
#'
#' `norm = "first"`: Curve values are normalised to the first count value.
#'
#' `norm = "last"`: Curve values are normalised to the last count value
#' (this can be useful in particular for radiofluorescence curves)
#'
#' `norm = "huot"`: Curve values are normalised as suggested by Sébastien Huot
#'  via GitHub:
#' \deqn{
#' y = (observed - median(background)) / (\max(observed) - median(background))
#' }
#'
#' The background of the curve is defined as the last 20% of the count values
#' of a curve.
#'
#' `norm = "intensity"`: Curve values are normalised to the channel length.
#'
#' `norm = 2.2`: Curve values are normalised to a positive number (e.g., 2.2).
#'
#' @return
#' An object of the same type as the input object provided.
#'
#' @section Function version: 0.1.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Data.Spectrum-class],  [Luminescence::RLum.Data.Image-class]
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ## create RLum.Data.Curve object from this example
#' curve <-
#'   set_RLum(
#'       class = "RLum.Data.Curve",
#'       recordType = "OSL",
#'       data = as.matrix(ExampleData.CW_OSL_Curve)
#'   )
#'
#' ## plot data without and with smoothing
#' plot_RLum(curve)
#' plot_RLum(normalise_RLum(curve))
#'
#' @keywords utilities
#'
#' @export
setGeneric("normalise_RLum", function(object, norm = TRUE, ...) {
  .set_function_name("normalise_RLum")
  on.exit(.unset_function_name(), add = TRUE)

  ## validation
  .validate_class(norm, c("logical", "character", "numeric"), length = 1)
  valid.norms <- c("max", "min", "first", "last", "huot", "intensity")
  if(inherits(norm, "character"))
    .validate_args(norm, valid.norms)
  else if (inherits(norm, "logical"))
    .validate_logical_scalar(norm, extra = paste("one of", .collapse(valid.norms)))
  else if (inherits(norm, "numeric"))
    .validate_positive_scalar(norm, extra = paste("one of", .collapse(valid.norms)))

  ## set generic
  standardGeneric("normalise_RLum")
})

#' @describeIn normalise_RLum
#' Returns a list of [Luminescence::RLum.Data-class] objects that had been passed to
#' [Luminescence::normalise_RLum]
#'
#' @export
setMethod("normalise_RLum", signature = "list",
          function(object, norm, ...) {
            ## apply method in the objects and return the same
            lapply(object, function(x) {
              if (inherits(x, c("RLum.Analysis", "RLum.Data"))) {
                return(normalise_RLum(x, norm = norm, ...))
              } else {
                return(x)
              }
            })
          })

## sort_RLum() --------------------------------------------------------------
#' @title Sort data for RLum-class and Risoe.BINfileData-class objects
#'
#' @description
#' The function provides a generalised access point for specific [Luminescence::RLum-class]
#' and [Luminescence::Risoe.BINfileData-class] objects. Depending on the input object,
#' the corresponding function will be selected.
#'
#' @param object [Luminescence::RLum-class] or
#' [Luminescence::Risoe.BINfileData-class] (**required**):
#' S4 object of class `RLum.Analysis` or `Risoe.BINfileData`.
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' An object of the same type as the input object provided.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::Risoe.BINfileData-class]
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.XSYG, envir = environment())
#' obj <- OSL.SARMeasurement$Sequence.Object[1:9]
#'
#' sort_RLum(obj, slot = "recordType")
#' sort_RLum(obj, info_element = "curveDescripter")
#'
#' @keywords utilities
#'
#' @export
setGeneric("sort_RLum", function(object, ...) {
  standardGeneric("sort_RLum")
})

#' @describeIn sort_RLum
#' Returns a list of sorted [Luminescence::RLum-class] objects.
#'
#' @export
setMethod("sort_RLum", signature = "list",
          function(object, ...) {
            ## apply method in the objects and return the same
            lapply(object, function(x) {
              if (inherits(x, "RLum.Analysis")) {
                return(sort_RLum(x,...))
              } else {
                return(x)
              }
            })
          })


## structure_RLum() ---------------------------------------------------------
#' @title General structure function for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the corresponding
#' function will be selected.
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' Returns a [data.frame] with structure of the object.
#'
#' @section Function version: 0.2.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.XSYG, envir = environment())
#'
#' ## show structure
#' structure_RLum(OSL.SARMeasurement$Sequence.Object)
#'
#' @export
setGeneric("structure_RLum", function(object, ...) {
  standardGeneric("structure_RLum")
})

#' @describeIn structure_RLum
#' Returns a list of data frames containing the structure of each [Luminescence::RLum-class]
#' object in the input list.
#'
#' @export
setMethod("structure_RLum", signature = "list",
    function(object, ...) {
      ## apply method in the objects and return the same
      lapply(object, function(x) {
        if (inherits(x, "RLum")) {
          return(structure_RLum(x, ...))
        } else {
          return(x)
        }
      })
    })


## view() -------------------------------------------------------------------
#' @title Convenience data visualisation function
#'
#' @description
#' Invokes the [utils::View] function tailored to objects in the package.
#' If started from RStudio, it uses the RStudio viewer.
#'
#' @param object (**required**) object to view
#'
#' @param ... further arguments passed to the specific class method
#'
#' @seealso [utils::View()]
#'
#' @returns
#' `NULL` and opens the data viewer.
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @keywords utilities
#'
#' @export
setGeneric("view", function(object, ... ) {
  standardGeneric("view")
})

## ensure that we can use the internal RStudio view function
## https://stackoverflow.com/questions/48234850/how-to-use-r-studio-view-function-programatically-in-a-package
#' @noRd
.view <- function(x, title) {
  get("View", envir = as.environment("package:utils"))(x, title) # nocov
}

#  ------------------------------------------------------------------------
# Author: Christoph Burow <christoph.burow@gmx.net>
# Affiliation: University of Cologne
# Date: 15/01/2019
# API version: v3
# Reference: https://docs.github.com/v3/
#  ------------------------------------------------------------------------

#' GitHub API - Deprecated
#'
#' R Interface to the GitHub API v3.
#'
#' These functions can be used to query a specific repository hosted on GitHub. \cr
#'
#'
#' @param user [character] (*with default*):
#' GitHub user name (defaults to `'r-lum'`).
#'
#' @param repo [character] (*with default*):
#' name of a GitHub repository (defaults to `'luminescence'`).
#'
#' @param branch [character] (*with default*):
#' branch of a GitHub repository (defaults to `'master'`).
#'
#' @param n [integer] (*with default*):
#' number of commits returned (defaults to 5).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @author Christoph Burow, University of Cologne (Germany)
#'
#' @section Function version: 0.1.0
#'
#' @references
#'
#' GitHub Developer API v3. [https://docs.github.com/v3/](), last accessed: 10/01/2017.
#'
#' @examples
#'
#' \dontrun{
#' github_branches(user = "r-lum", repo = "luminescence")
#' github_issues(user = "r-lum", repo = "luminescence")
#' github_commits(user = "r-lum", repo = "luminescence", branch = "master", n = 10)
#' }
#'
#' @name GitHub-API
NULL

# COMMITS -----------------------------------------------------------------
#' @rdname GitHub-API
#'
#' @details
#' `github_commits` lists the most recent `n` commits of a specific branch of a repository.
#'
#' @return
#' `github_commits`: [data.frame] with columns:
#'
#' \tabular{ll}{
#'  `[ ,1]` \tab SHA \cr
#'  `[ ,2]` \tab AUTHOR \cr
#'  `[ ,3]` \tab DATE \cr
#'  `[ ,4]` \tab MESSAGE \cr
#' }
#'
#' @export
github_commits <- function(user = "r-lum", repo = "luminescence",
                           branch = "master", n = 5) {
  .set_function_name("github_commits")
  on.exit(.unset_function_name(), add = TRUE)
  .Deprecated(msg = "This function is deprecated, it will be removed in a future release")

  ## input validation
  .validate_class(user, "character", length = 1)
  .validate_class(repo, "character", length = 1)
  .validate_class(branch, "character", length = 1)
  .validate_positive_scalar(n, int = TRUE)

  # fetch available branches and check if provided branch exists
  branches <- github_branches(user, repo)
  # nocov start
  if (!any(grepl(branch, branches$BRANCH)))
    .throw_error("Branch '", branch, "' does not exist")
  # nocov end

  # build URL and retrieve content
  sha <- branches$SHA[grep(paste0("^", branch, "$"), branches$BRANCH)]
  url <- paste0("https://api.github.com/repos/", user, "/", repo, "/commits?",
                "per_page=", n, "&sha=", sha)
  content <- .github_getContent(url)

  # format output as data.frame
  output <- do.call(rbind, lapply(content, function(x) {
    data.frame(SHA = x$sha,
               AUTHOR = x$commit$author$name,
               DATE = x$commit$author$date,
               MESSAGE = x$commit$message,
               stringsAsFactors = FALSE)
  }))

  return(output)
}


# BRANCHES ----------------------------------------------------------------
#' @rdname GitHub-API
#'
#' @details
#' `github_branches` can be used to list all current branches of a
#' repository and returns the corresponding SHA hash as well as an installation
#' command to install the branch in R via the 'devtools' package.
#'
#' @return
#' `github_branches`: [data.frame] with columns:
#'
#' \tabular{ll}{
#'  `[ ,1]` \tab BRANCH \cr
#'  `[ ,2]` \tab SHA \cr
#'  `[ ,3]` \tab INSTALL \cr
#' }
#'
#' @export
github_branches <- function(user = "r-lum", repo = "luminescence") {
  .set_function_name("github_branches")
  on.exit(.unset_function_name(), add = TRUE)
  .Deprecated(msg = "This function is deprecated, it will be removed in a future release")

  .validate_class(user, "character", length = 1)
  .validate_class(repo, "character", length = 1)

  # build URL and retrieve content
  url <- paste0("https://api.github.com/repos/", user, "/", repo, "/branches")
  content <- .github_getContent(url)

  # extract relevant information from server response
  branches <- sapply(content, function(x) x$name)
  sha <- sapply(content, function(x) x$commit$sha)

  # format output as data.frame
  data.frame(
    BRANCH = branches,
    SHA = sha,
    INSTALL = paste0("devtools::install_github('r-lum/luminescence@", branches, "')"),
    stringsAsFactors = FALSE
  )
}


# ISSUES ------------------------------------------------------------------
#' @rdname GitHub-API
#'
#' @details
#' `github_issues` lists all open issues for a repository in valid YAML.
#'
#' @return
#' `github_commits`: Nested [list] with `n` elements.
#' Each commit element is a list with elements:
#'
#' \tabular{ll}{
#'  `[[1]]` \tab NUMBER \cr
#'  `[[2]]` \tab TITLE \cr
#'  `[[3]]` \tab BODY \cr
#'  `[[4]]` \tab CREATED \cr
#'  `[[5]]` \tab UPDATED \cr
#'  `[[6]]` \tab CREATOR \cr
#'  `[[7]]` \tab URL \cr
#'  `[[8]]` \tab STATUS \cr
#' }
#'
#' @export
github_issues <- function(user = "r-lum", repo = "luminescence", verbose = TRUE) {
  .set_function_name("github_issues")
  on.exit(.unset_function_name(), add = TRUE)
  .Deprecated(msg = "This function is deprecated, it will be removed in a future release")

  .validate_class(user, "character", length = 1)
  .validate_class(repo, "character", length = 1)

  # build URL and retrieve content
  url <- paste0("https://api.github.com/repos/", user,"/", repo, "/issues")
  content <- .github_getContent(url)

  # format output as nested list
  issues <- lapply(content, function(x) {
    list(
      NUMBER = x$number,
      TITLE = x$title,
      BODY = gsub("\n", "", x$body),
      CREATED = x$created_at,
      UPDATED = x$updated_at,
      CREATOR = x$user$login,
      URL = x$url,
      STATUS = x$state,
      MILESTONE = x$milestone$title)
  })

  # custom printing of the issues list, as print.list produces unreadable
  # console output
  if (verbose) {
    lapply(issues, function(x) {

      # limit width of description text
      DESCRIPTION <- ""
      for (i in seq_len(ceiling(nchar(x$BODY) / 100)))
        DESCRIPTION <- paste(DESCRIPTION, "  ",
                             substr(x$BODY, i*100-99, i*100), "\n")

      # print to console in valid YAML
      cat(paste0("---\n",
                 'title: "', x$TITLE, '"', "\n",
                 "number: ", x$NUMBER, "\n",
                 'url: "', x$URL, '"', "\n",
                 "created: ", x$CREATED, "\n",
                 "updated: ", x$UPDATED, "\n",
                 "creator: ", x$CREATOR, "\n",
                 "status: ", x$STATUS, "\n",
                 'milestone: "', x$MILESTONE, '"', "\n",
                 "description: >\n", DESCRIPTION,
                 "\n\n\n"))
    })
  }
  # return invisible as we explicitly print the output
  invisible(issues)
}


# HELPER ------------------------------------------------------------------

# This function queries the URL, checks the server response and returns
# the content.
.github_getContent <- function(url) {
  response <- tryCatch(httr::GET(url, httr::accept_json()),
                       error = function(e) .throw_error(e$message))
  # nocov start
  if (httr::status_code(response) != 200)
    .throw_error("Contacting ", url, " returned status code ",
                 httr::status_code(response))
  # nocov end
  httr::content(response)
}

test_that("Deprecated github functions", {
  testthat::skip_on_cran()

  SW({
  expect_warning(github_commits(),
                 "This function is deprecated")
  expect_warning(github_branches(),
                 "This function is deprecated")
  expect_warning(github_issues(),
                 "This function is deprecated")
  })
})

#' Class `"RLum.Data.Spectrum"`
#'
#' Class for representing luminescence spectra data (TL/OSL/RF).
#'
#' @name RLum.Data.Spectrum-class
#'
#' @docType class
#'
#' @slot recordType
#' Object of class [character] containing the type of the curve (e.g. "TL" or "OSL")
#'
#' @slot curveType
#' Object of class [character] containing curve type, allowed values are
#' "measured" or "predefined"
#'
#' @slot data
#' Object of class [matrix] containing spectrum (count) values.
#' Row labels indicate wavelength/pixel values, column labels are temperature or time values.
#'
#' @slot info
#' Object of class [list] containing further meta information objects
#'
#' @note
#' The class should only contain data for a single spectra data set. For
#' additional elements the slot `info` can be used. Objects from this class are automatically
#' created by, e.g., [Luminescence::read_XSYG2R]
#'
#' @section Objects from the Class:
#' Objects can be created by calls of the form `set_RLum("RLum.Data.Spectrum", ...)`.
#'
#' @section Class version: 0.5.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum-class], [Luminescence::RLum.Data-class], [Luminescence::plot_RLum]
#'
#' @keywords classes
#'
#' @examples
#'
#' showClass("RLum.Data.Spectrum")
#'
#' ##show example data
#' data(ExampleData.XSYG, envir = environment())
#' TL.Spectrum
#'
#' ##show data matrix
#' get_RLum(TL.Spectrum)
#'
#' ##plot spectrum
#' \dontrun{
#' plot_RLum(TL.Spectrum)
#' }
#'
#' @export
setClass(
  "RLum.Data.Spectrum",
  slots = list(
    recordType = "character",
    curveType = "character",
    data = "matrix",
    info = "list"
  ),
  contains = "RLum.Data",
  prototype = list(
    recordType = NA_character_,
    curveType = NA_character_,
    data = matrix(),
    info = list()
  )
)


## as() ---------------------------------------------------------------------
##data.frame
##COERCE RLum.Data.Spectrum >> data.frame AND data.frame >> RLum.Data.Spectrum
#' as()
#'
#' for `[RLum.Data.Spectrum-class]`
#'
#' **[Luminescence::RLum.Data.Spectrum-class]**
#'
#' \tabular{ll}{
#'   **from** \tab **to**\cr
#'   `data.frame` \tab `data.frame`\cr
#'   `matrix` \tab `matrix` \cr
#'   `list` \tab `list`
#' }
#'
#' @name as
setAs("data.frame", "RLum.Data.Spectrum",
      function(from,to){

        new(to,
            recordType = NA_character_,
            curveType = NA_character_,
            data = as.matrix(from),
            info = list())
      })

setAs("RLum.Data.Spectrum", "data.frame",
      function(from){
        as.data.frame(from@data)
      })

##MATRIX
##COERCE RLum.Data.Spectrum >> matrix AND matrix >> RLum.Data.Spectrum
setAs("matrix", "RLum.Data.Spectrum",
      function(from,to){
        new(to,
            recordType = NA_character_,
            curveType = NA_character_,
            data = from,
            info = list())
      })

setAs("RLum.Data.Spectrum", "matrix",
      function(from){
        from@data
      })

## LIST
## COERCE RLum.Data.Spectrum >> list AND list >> RLum.Data.Spectrum
setAs("list", "RLum.Data.Spectrum",
      function(from, to){
        if (length(from) == 0)
          return(set_RLum("RLum.Data.Spectrum"))
        new(to,
            recordType = NA_character_,
            curveType = NA_character_,
            data = matrix(unlist(from)),
            info = list())
      })

setAs("RLum.Data.Spectrum", "list",
      function(from){
        apply(from@data, 2, list)
      })

## show() -------------------------------------------------------------------
#' @describeIn show
#' Show the structure of `RLum.Data.Spectrum` objects.
#'
#' @export
setMethod("show",
          signature(object = "RLum.Data.Spectrum"),
          function(object){

            x.range <- suppressWarnings(range(as.numeric(rownames(object@data))))
            y.range <- suppressWarnings(range(as.numeric(colnames(object@data))))
            z.range <- range(object@data)

            ##print information

            cat("\n [RLum.Data.Spectrum-class]")
            cat("\n\t recordType:", object@recordType)
            cat("\n\t curveType:",  object@curveType)
            cat("\n\t .. recorded frames:", length(object@data[1,]))
            cat("\n\t .. .. measured values per frame:", length(object@data[,1]))
            cat("\n\t .. .. range wavelength/pixel:", x.range)
            cat("\n\t .. .. range time/temp.:", y.range)
            cat("\n\t .. .. range count values:", z.range)
            cat("\n\t additional info elements:", length(object@info))
            #cat("\n\t\t >> names:", names(object@info))
            cat("\n")
          }
)


## set_RLum() ---------------------------------------------------------------
#' @describeIn set_RLum
#' Construction method for [Luminescence::RLum.Data.Spectrum-class] objects.
#'
#' @export
setMethod(
  "set_RLum",
  signature = signature("RLum.Data.Spectrum"),
  definition = function(
    class,
    originator,
    .uid,
    .pid,
    recordType = "Spectrum",
    curveType = NA_character_,
    data = matrix(),
    info = list()) {

    ##The case where an RLum.Data.Spectrum object can be provided
    ##with this RLum.Data.Spectrum objects can be provided to be reconstructed

    if (inherits(data, "RLum.Data.Spectrum")) {
      ##check for missing curveType
      if (missing(curveType))
        curveType <- data@curveType

      ##check for missing recordType
      if (missing(recordType))
        recordType <- data@recordType

      ##check for missing info
      if (missing(info))
        info <- data@info

      originator <- data@originator
      .uid <- data@.uid
      .pid <- data@.pid
      data <- data@data
    }

    ## set empty class from object
    newRLumDataSpectrum <- new("RLum.Data.Spectrum")

    ## fill - this is the faster way, filling in new() costs ...
    newRLumDataSpectrum@originator <- originator
    newRLumDataSpectrum@recordType <- recordType
    newRLumDataSpectrum@curveType <- curveType
    newRLumDataSpectrum@data <- data
    newRLumDataSpectrum@info <- info
    newRLumDataSpectrum@.uid <- .uid
    newRLumDataSpectrum@.pid <- .pid

    return(newRLumDataSpectrum)
  }
)


## get_RLum() ---------------------------------------------------------------
#' @describeIn get_RLum
#' Accessor method for [Luminescence::RLum.Data.Spectrum-class] objects.
#' The argument `info.object` is optional to directly access the info elements.
#' If no info element name is provided, the raw curve data (matrix) will be
#' returned.
#'
#' @export
setMethod("get_RLum",
          signature("RLum.Data.Spectrum"),
          definition = function(object, info.object = NULL) {
              .set_function_name("get_RLum")
              on.exit(.unset_function_name(), add = TRUE)

              .validate_class(info.object, "character", null.ok = TRUE, length = 1)
              if (!is.null(info.object)) {
                if (!info.object %in% names(object@info)) {
                  .throw_error("Invalid element name, valid names are: ",
                               .collapse(names(object@info)))
                }
                unlist(object@info[info.object])
              } else {
                ## info.object not provided, return the curve data
                object@data
              }
          })


## names_RLum ---------------------------------------------------------------
#' @describeIn names_RLum
#' Returns the names of the info elements stored in the object.
#'
#' @export
setMethod("names_RLum",
          "RLum.Data.Spectrum",
          function(object){
            names(object@info)
          })
## bin_RLum() ---------------------------------------------------------------
#' @describeIn bin_RLum.Data
#' Allows binning of RLum.Data.Spectrum data. Count values and values on the
#' x-axis are summed up; for wavelength/energy values, the mean is calculated.
#'
#' @param bin_size.col [integer] (*with default*):
#' number of channels used for each bin, e.g. `bin_size.col = 2` means that
#' two channels are binned. Note: The function does not check the input, very large values
#' mean a full column binning (a single sum)
#'
#' @param bin_size.row [integer] (*with default*):
#' number of channels used for each bin, e.g. `bin_size.row = 2` means that
#' two channels are binned. Note: The function does not check the input, very large values
#' mean a full row binning (a single sum)
#'
#' @export
setMethod(f = "bin_RLum.Data",
          signature = "RLum.Data.Spectrum",
          function(object, bin_size.col = 1, bin_size.row = 1) {
            .set_function_name("bin_RLum.Data.Spectrum")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            if (length(object@data) < 2) {
              .throw_error("'object' contains no data")
            }
            .validate_positive_scalar(bin_size.row, int = TRUE)
            .validate_positive_scalar(bin_size.col, int = TRUE)

            ##make sure that we do not get in trouble with negative values
            bin_size.col <- abs(bin_size.col)
            bin_size.row <- abs(bin_size.row)

            ##perform binning
            ##we want to be efficient, so we start
            ##with the larger object
            if(bin_size.row > bin_size.col){
              ##row binning first
              m <- .matrix_binning(object@data, bin_size = bin_size.row, bin_col = FALSE, names = "mean")
              m <- .matrix_binning(m, bin_size = bin_size.col, bin_col = TRUE, names = "groups")

            } else {
              ##column binning first
              m <- .matrix_binning(object@data, bin_size = bin_size.col, bin_col = TRUE, names = "groups")
              m <- .matrix_binning(m, bin_size = bin_size.row, bin_col = FALSE, names = "mean")

            }

            ##write back to object
            object@data <- m

            ##return object
            return(object)
          })

## normalise_RLum() --------------------------------------------------------------
#' @describeIn normalise_RLum
#' Normalise [Luminescence::RLum.Data.Spectrum-class] objects to value set via
#' the argument `norm`
#'
#' @export
setMethod(
  f = "normalise_RLum",
  signature = "RLum.Data.Spectrum",
  function(object, norm = TRUE) {
    if (norm == "intensity") {
      ## compute the channel length and assign it to `norm`, so that we can
      ## piggy-back on the code for normalisation by a scalar
      time.temp <- as.numeric(colnames(object@data))
      norm <- rep(diff(c(0, time.temp)), each = nrow(object@data))
    }
    object@data[] <- .normalise_curve(object@data[], norm = norm)
    object
  }
)

#' General accessors function for RLum S4 class objects
#'
#' Function calls object-specific get functions for RLum S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum-class] objects.\cr
#' Depending on the input object, the corresponding get function will be selected.
#' Allowed arguments can be found in the documentations of the corresponding
#' [RLum-class] class.
#'
#' @param object [RLum-class] (**required**):
#' S4 object of class `RLum` or an object of type [list] containing only objects
#' of type [RLum-class]
#'
#' @param ... further arguments that will be passed to the object specific methods. For
#' further details on the supported arguments please see the class
#' documentation: [RLum.Data.Curve-class], [RLum.Data.Spectrum-class],
#' [RLum.Data.Image-class], [RLum.Analysis-class] and [RLum.Results-class]
#'
#' @return Return is the same as input objects as provided in the list.
#'
#' @section Function version: 0.3.3
#'
#' @author
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#'
#' @seealso [RLum.Data.Curve-class], [RLum.Data.Image-class],
#'  [RLum.Data.Spectrum-class], [RLum.Analysis-class], [RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ##Example based using data and from the calc_CentralDose() function
#'
#' ##load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ##apply the central dose model 1st time
#' temp1 <- calc_CentralDose(ExampleData.DeValues$CA1)
#'
#' ##get results and store them in a new object
#' temp.get <- get_RLum(object = temp1)
#'
#' @md
#' @export
setGeneric("get_RLum", function (object, ...) {standardGeneric("get_RLum") })

# Method for get_RLum method for RLum objects in a list for a list of objects  -------------------
#' @describeIn get_RLum
#' Returns a list of [RLum-class] objects that had been passed to [get_RLum]
#'
#' @param class [character] (*optional*): allows to define the class that gets selected if
#' applied to a list, e.g., if a list consists of different type of RLum-class objects, this
#' arguments allows to make selection. If nothing is provided, all RLum-objects are treated.
#'
#' @param null.rm [logical] (*with default*): option to get rid of empty and NULL objects
#'
#' @md
#' @export
setMethod("get_RLum",
          signature = "list",
          function(object, class = NULL, null.rm = FALSE, ...){

            ##take care of the class argument
            if(!is.null(class)){
              sel <- class[1] == vapply(object, function(x) class(x), character(1))
              if(any(sel))
                object <- object[sel]

              rm(sel)
            }


            ##make remove all non-RLum objects
            selection <- lapply(1:length(object), function(x){

              ##get rid of all objects that are not of type RLum, this is better than leaving that
              ##to the user
              if(inherits(object[[x]], what = "RLum")){

                ##it might be the case the object already comes with empty objects, this would
                ##cause a crash
                if(class(object[[x]]) == "RLum.Analysis" && length(object[[x]]@records) == 0)
                  return(NULL)

                get_RLum(object[[x]], ...)


              } else {

                warning(paste0("[get_RLum()] object #",x," in the list was not of type 'RLum' and has been removed!"),
                        call. = FALSE)
                return(NULL)

              }

            })

            ##remove empty or NULL objects after the selection ... if wanted
            if(null.rm){


              ##first set all empty objects to NULL ... for RLum.Analysis objects
              selection <- lapply(1:length(selection), function(x){
                if(class(selection[[x]]) == "RLum.Analysis" && length(selection[[x]]@records) == 0){
                  return(NULL)

                }else{
                  return(selection[[x]])

                }

              })

              ##get rid of all NULL objects
              selection <- selection[!sapply(selection, is.null)]


            }

            return(selection)

          })


#' Method to handle NULL if the user calls get_RLum
#'
#' @describeIn get_RLum
#'
#' Returns NULL
#'
#' @md
#' @export
setMethod("get_RLum",
          signature = "NULL",
          function(object, ...){NULL})

#' @title Nonlinear Least Squares Fit for OSL surface exposure data
#'
#' @description
#' This function determines the (weighted) least-squares estimates of the
#' parameters of either equation 1 in *Sohbati et al. (2012a)* or equation 12 in
#' *Sohbati et al. (2012b)* for a given OSL surface exposure data set (**BETA**).
#'
#' @details
#' **Weighted fitting**
#'
#' If `weights = TRUE` the function will use the inverse square of the error (\eqn{1/\sigma^2})
#' as weights during fitting using [minpack.lm::nlsLM]. Naturally, for this to
#' take effect individual errors must be provided in the third column of the
#' `data.frame` for `data`. Weighted fitting is **not** supported if `data`
#' is a list of multiple `data.frame`s, i.e., it is not available for global
#' fitting.
#'
#' **Dose rate**
#' If any of the arguments `Ddot` or `D0` is at its default value (`NULL`),
#' this function will fit equation 1 in Sohbati et al. (2012a) to the data. If
#' the effect of dose rate (i.e., signal saturation) needs to be considered,
#' numeric values for the dose rate (`Ddot`) (in Gy/ka) and the characteristic
#' saturation dose (`D0`) (in Gy) must be provided. The function will then fit
#' equation 12 in Sohbati et al. (2012b) to the data.
#'
#' **Note:** Currently, this function does **not** consider the variability
#' of the dose rate with sample depth (`x`)! In the original equation the dose
#' rate `D` is an arbitrary function of `x` (term `D(x)`), but here `D` is assumed
#' constant.
#'
#' **Global fitting**
#' If `data` is [list] of multiple `data.frame`s, each representing a separate
#' sample, the function automatically performs a global fit to the data. This
#' may be useful to better constrain the parameters `sigmaphi` or `mu` and
#' **requires** that a known age for each sample
#' (e.g., `age = c(100, 1000)` if `data` is a list with two samples).
#'
#' @param data [data.frame] or [list] (**required**):
#' Measured OSL surface exposure data with the following structure:
#'
#' ```
#'                                (optional)
#'      | depth (a.u.)| intensity | error |
#'      |     [ ,1]   |    [ ,2]  | [ ,3] |
#'      |-------------|-----------|-------|
#' [1, ]|    ~~~~     |    ~~~~   | ~~~~  |
#' [2, ]|    ~~~~     |    ~~~~   | ~~~~  |
#'  ... |     ...     |     ...   |  ...  |
#' [x, ]|    ~~~~     |    ~~~~   | ~~~~  |
#'
#' ```
#'
#' Alternatively, a [list] of `data.frames` can be provided, where each
#' `data.frame` has the same structure as shown above, with the exception that
#' they must **not** include the optional error column. Providing a [list] as
#' input automatically activates the global fitting procedure (see details).
#'
#' @param sigmaphi [numeric] (*optional*):
#' A numeric value for `sigmaphi`, i.e. the charge detrapping rate.
#' Example: `sigmaphi = 5e-10`
#'
#' @param mu [numeric] (*optional*):
#' A numeric value for `mu`, i.e. the light attenuation coefficient. If `data`
#' is a list of *n* samples, then `mu` must be a vector of length 1 or *n*.
#' Example: `mu = 0.9`, or `mu = c(0.9, 0.8, 0.9)`.
#'
#' @param age [numeric] (*optional*):
#' The age (a) of the sample, if known. If `data` is a [list] of *n* samples,
#' then `age` must be a numeric vector of length *n*.
#' Example: `age = 10000`, or `age = c(1e4, 1e5, 1e6)`.
#'
#' @param Ddot [numeric] (*optional*):
#' A numeric value for the environmental dose rate (Gy/ka). For this argument
#' to be considered a value for `D0` must also be provided; otherwise it will be
#' ignored.
#'
#' @param D0 [numeric] (*optional*):
#' A numeric value for the characteristic saturation dose (Gy). For this argument
#' to be considered a value for `Ddot` must also be provided; otherwise it will be
#' ignored.
#'
#' @param weights [logical] (*with default*):
#' If `TRUE` the fit will be weighted by the inverse square of the error.
#' Requires `data` to be a [data.frame] with three columns.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param legend [logical] (*with default*):
#' Show or hide the equation inside the plot.
#'
#' @param error_bars [logical] (*with default*):
#' Show or hide error bars (only applies if errors were provided).
#'
#' @param coord_flip [logical] (*with default*):
#' Flip the coordinate system.
#'
#' @param ... Further parameters passed to [plot].
#' Custom parameters include:
#' - `verbose` ([logical]): enable/disable output to the terminal
#' - `line_col`: Colour of the fitted line
#' - `line_lty`: Type of the fitted line (see `lty` in `?par`)
#' - `line_lwd`: Line width of the fitted line (see `lwd` in `?par`)
#'
#' @return
#'
#' Function returns results numerically and graphically:
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$summary` \tab `data.frame` \tab summary of the fitting results \cr
#'  `$data` \tab `data.frame` \tab the original input data \cr
#'  `$fit` \tab `nls` \tab the fitting object produced by [minpack.lm::nlsLM] \cr
#'  `$args` \tab `character` \tab arguments of the call \cr
#'  `$call` \tab `call` \tab the original function call \cr
#' }
#'
#'**slot:** **`@info`**
#'
#' Currently unused.
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' A scatter plot of the provided depth-intensity OSL surface exposure data
#' with the fitted model.
#'
#' @section Function version: 0.1.0
#'
#' @note
#' **This function has BETA status. If possible, results should be**
#' **cross-checked.**
#'
#' @author Christoph Burow, University of Cologne (Germany)
#'
#' @seealso  [Luminescence::ExampleData.SurfaceExposure], [minpack.lm::nlsLM]
#'
#' @references
#'
#' Sohbati, R., Murray, A.S., Chapot, M.S., Jain, M., Pederson, J., 2012a.
#' Optically stimulated luminescence (OSL) as a chronometer for surface exposure
#' dating. Journal of Geophysical Research 117, B09202. doi:
#' \doi{10.1029/2012JB009383}
#'
#' Sohbati, R., Jain, M., Murray, A.S., 2012b. Surface exposure dating of
#' non-terrestrial bodies using optically stimulated luminescence: A new method.
#' Icarus 221, 160-166.
#'
#' @keywords datagen
#'
#' @examples
#'
#' ## Load example data
#' data("ExampleData.SurfaceExposure")
#'
#' ## Example 1 - Single sample
#' # Known parameters: 10000 a, mu = 0.9, sigmaphi = 5e-10
#' sample_1 <- ExampleData.SurfaceExposure$sample_1
#' head(sample_1)
#' results <- fit_SurfaceExposure(
#'  data = sample_1,
#'  mu = 0.9,
#'  sigmaphi = 5e-10)
#' get_RLum(results)
#'
#' ## Example 2 - Single sample and considering dose rate
#' # Known parameters: 10000 a, mu = 0.9, sigmaphi = 5e-10,
#' # dose rate = 2.5 Gy/ka, D0 = 40 Gy
#' sample_2 <- ExampleData.SurfaceExposure$sample_2
#' head(sample_2)
#' results <- fit_SurfaceExposure(
#'  data = sample_2,
#'  mu = 0.9,
#'  sigmaphi = 5e-10,
#'  Ddot = 2.5,
#'  D0 = 40)
#' get_RLum(results)
#'
#' ## Example 3 - Multiple samples (global fit) to better constrain 'mu'
#' # Known parameters: ages = 1e3, 1e4, 1e5, 1e6 a, mu = 0.9, sigmaphi = 5e-10
#' set_1 <- ExampleData.SurfaceExposure$set_1
#' str(set_1, max.level = 2)
#' results <- fit_SurfaceExposure(
#'   data = set_1,
#'   age = c(1e3, 1e4, 1e5, 1e6),
#'   sigmaphi = 5e-10)
#' get_RLum(results)
#'
#' ## Example 4 - Multiple samples (global fit) and considering dose rate
#' # Known parameters: ages = 1e2, 1e3, 1e4, 1e5, 1e6 a, mu = 0.9, sigmaphi = 5e-10,
#' # dose rate = 1.0 Ga/ka, D0 = 40 Gy
#' set_2 <- ExampleData.SurfaceExposure$set_2
#' str(set_2, max.level = 2)
#' results <- fit_SurfaceExposure(
#'  data = set_2,
#'  age = c(1e2, 1e3, 1e4, 1e5, 1e6),
#'  sigmaphi = 5e-10,
#'  Ddot = 1,
#'  D0 = 40)
#' get_RLum(results)
#'
#' @export
fit_SurfaceExposure <- function(
    data,
    sigmaphi = NULL,
    mu = NULL,
    age = NULL,
    Ddot = NULL,
    D0 = NULL,
    weights = FALSE,
    plot = TRUE,
    legend = TRUE,
    error_bars = TRUE,
    coord_flip = FALSE,
    ...
) {
  .set_function_name("fit_SurfaceExposure")
  on.exit(.unset_function_name(), add = TRUE)

  ## SETTINGS ----
  settings <- list(
    verbose = TRUE,
    info = list()
  )
  settings <- modifyList(settings, list(...))

  ## Integrity checks -------------------------------------------------------

  .validate_not_empty(data)
  .validate_logical_scalar(weights)
  if (!is.null(mu))
    .validate_not_empty(mu)

  ## Data type validation
  if (inherits(data, "RLum.Results"))
    data <- get_RLum(data, "data")

  if (inherits(data, "matrix"))
    data <- as.data.frame(data)

  if (inherits(data, "data.table"))
    data <- as.data.frame(data)

  ## For global fitting of multiple data sets 'data' must be a list
  if (inherits(data, "list")) {
    lapply(data, function(x) .validate_class(x, "data.frame",
                                             name = "All elements of 'data'"))

    # Global fitting requires and equal amount of ages to be provided
    if (length(data) != length(age))
      .throw_error("If 'data' is a list of data sets for global fitting, ",
                   "'age' must be of the same length.")

    ## if mu is specified, it must have length 1 or match the number of data sets
    if (!is.null(mu) && length(mu) != 1 && length(data) != length(mu))
      .throw_error("If 'data' is a list of data sets for global fitting, ",
                   "'mu' must either be of the same length or of length 1.")

    # TODO: Support weighted fitting for global fit
    if (weights) {
      if (settings$verbose)
        .throw_warning("'weights' is not supported when multiple data sets ",
                       "are provided for global fitting")
      weights <- FALSE
    }

    # collapse list into a data.frame with a $group column to distinguish
    # between individual samples
    data_list <- data

    for (i in seq_along(data))
      data[[i]]$group <- LETTERS[[i]]

    data <- do.call(rbind, data)
    data$group <- as.factor(data$group)
    global_fit <- TRUE
  } else {
    # ignore 'global_fit' if 'data' is a data.frame
    global_fit <- FALSE
  }

  # Exit if data type is invalid
  .validate_class(data, "data.frame")
  if (ncol(data) < 2) {
    .throw_error("'data' should have at least two columns")
  }

  # Check which parameters have been provided
  if (!is.null(age) && anyNA(age)) age <- NULL
  if (!is.null(sigmaphi) && anyNA(sigmaphi)) sigmaphi <- NULL
  if (!is.null(mu) && anyNA(mu)) mu <- NULL
  .validate_class(mu, "numeric", null.ok = TRUE)
  .validate_positive_scalar(sigmaphi, null.ok = TRUE)
  .validate_logical_scalar(settings$verbose, name = "'verbose'")
  .validate_logical_scalar(plot)
  .validate_logical_scalar(legend)
  .validate_logical_scalar(error_bars)
  .validate_logical_scalar(coord_flip)

  ## Weighting options (only available for global fitting)
  if (ncol(data) >= 3 && weights && !global_fit)
    wi <- (1 / data[ ,3]^2) / sum(1 / data[ ,3]^2)
  else
    wi <- rep(1, times = nrow(data))

  ## remove rows with NA
  if (anyNA(data)) {
    data <- data[stats::complete.cases(data), ]
    if (nrow(data) == 0)
      .throw_error("After NA removal, nothing is left from the data set")
    if (settings$verbose)
      .throw_message("NA values in 'data' were removed", error = FALSE)
  }

  ## extract errors into separate variable
  if (ncol(data) >= 3 && !global_fit)
    error <- data[ ,3]
  else
    error <- NULL

  ## Take only the first to columns (depth, signal)
  if (ncol(data) > 2 && !global_fit)
    data <- data[ ,1:2]

  ## Data preprocessing ----

  # set column names
  if (!global_fit)
    colnames(data) <- c("x", "y")
  else
    colnames(data) <- c("x", "y", "group")

  ## FITTING ----

  ## Functions
  # w/o dose rate
  year_s <- .const$year_s
  fun <- y ~ exp(-sigmaphi * age * year_s * exp(-mu * x))
  fun_global <- y ~ exp(-sigmaphi * age[group] * year_s * exp(-mu * x))

  # w/ dose rate (Sohbati et al. 2012, eq 12)
  if (!is.null(Ddot))
    Ddot <- Ddot / (1000 * year_s)

  fun_w_dr <- y ~ (sigmaphi * exp(-mu * x) * exp(-(age * year_s) * (sigmaphi * exp(-mu * x) + Ddot/D0)) + Ddot/D0) /
                         (sigmaphi * exp(-mu * x) + Ddot/D0)
  fun_global_w_dr <- y ~ (sigmaphi * exp(-mu * x) * exp(-(age[group] * year_s) * (sigmaphi * exp(-mu * x) + Ddot/D0)) + Ddot/D0) /
                                (sigmaphi * exp(-mu * x) + Ddot/D0)

  ## start parameter
  start <- list(sigmaphi = if (is.null(sigmaphi)) 5.890e-09 else NULL,
                mu = if (is.null(mu)) 1 else NULL,
                age = if (is.null(age)) 2 else NULL)

  start <- .rm_NULL_elements(start)

  ## fitting boundaries
  lower <- list(sigmaphi = if (is.null(sigmaphi)) -Inf else NULL,
                mu = if (is.null(mu)) 0 else NULL,
                age = if (is.null(age)) 0 else NULL)
  upper <- list(sigmaphi = if (is.null(sigmaphi)) Inf else NULL,
                mu = if (is.null(mu)) Inf else NULL,
                age = if (is.null(age)) Inf else NULL)

  ## Decision tree which of the functions to use
  if (!is.null(Ddot) && !is.null(D0)) {
    if (global_fit)
      use_fun <- fun_global_w_dr
    else
      use_fun <- fun_w_dr
  } else if (global_fit) {
      use_fun <- fun_global
  } else {
      use_fun <- fun
  }

  # (un)constrained fitting
  fit <- tryCatch({
    suppressWarnings(minpack.lm::nlsLM(formula = use_fun,
                      data = data,
                      start = start,
                      lower = unlist(lower),
                      upper = unlist(upper),
                      weights = wi))
  },
  error = function(e) { e }
  )

  # return NULL if fitting failed
  if (!inherits(fit, "simpleError") && !inherits(try(summary(fit), silent = TRUE), "try-error")) {
    # Extract coefficients
    coef <- as.data.frame(coef(summary(fit)))
  } else {
    if (settings$verbose)
      .throw_message("Unable to fit the data. ",
              "Original error from minpack.lm::nlsLM(): ", fit$message)

    # Fill with NA values
    coef <- data.frame(
      Estimate = rep(NA, 3),
      "Std. Error" = rep(NA, 3),
      row.names = c("age", "sigmaphi", "mu"), check.names = FALSE
    )
  }

  ## RESULTS ----
  summary <- data.frame(
    age = age %||% coef["age", "Estimate"],
    age_error = coef["age", "Std. Error"],
    sigmaphi = sigmaphi %||% coef["sigmaphi", "Estimate"],
    sigmaphi_error = coef["sigmaphi", "Std. Error"],
    mu = mu %||% coef["mu", "Estimate"],
    mu_error = coef["mu", "Std. Error"]
  )

  ## Create RLum.Results object
  results <- set_RLum(class = "RLum.Results",
                      originator = "fit_SurfaceExposure",
                      data = list(summary = summary,
                                  data = data,
                                  fit = fit,
                                  args = as.list(sys.call()[-1]),
                                  call = sys.call()),
                      info = settings$info
  )

  ## PLOT ----
  if (plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    # remove $group column for easier data handling
    if (global_fit)
      data$group <- NULL

    # re-order x,y columns
    if (coord_flip)
      data <- data.frame(data$y, data$x)

    # set default plot settings
    plot_settings <- list(
      x = data,
      main = "",
      pch = 21,
      col = "black",
      bg = "red",
      xlab = if (coord_flip) "OSL intensity (Ln/Tn)" else "Depth (mm)",
      ylab = if (coord_flip) "Depth (mm)" else "OSL intensity (Ln/Tn)",
      cex = 1.0,
      lty = 1,
      lwd = 1,
      log = "",
      ylim = if (coord_flip) rev(range(pretty(data[, 2]))) else range(pretty(data[, 2])),
      xlim = range(pretty(data[ ,1]))
    )

    # override default settings with valid arguments in ...
    plot_settings <- modifyList(plot_settings, list(...))
    valid_settings <- c(names(par()), formalArgs("title"), formalArgs("plot.default"), "cex")
    plot_settings <- plot_settings[names(plot_settings) %in% valid_settings]

    # set global plot settings
    par(cex = plot_settings$cex)

    if (grepl("y", plot_settings$log)) {
      plot_settings$ylim[1] <- 0.01
      pos.idx <- which(data[, 2] > 0)
      error <- error[pos.idx]
      plot_settings$x <- data[pos.idx, ]
    }

    ## create main plot
    do.call("plot", modifyList(plot_settings, list(x = NA)))


    ## add data points
    if (!global_fit) {
      points(data, type = "p", pch = plot_settings$pch, bg = plot_settings$bg, col = plot_settings$col)
    } else {
      Map(function(d, i) {
        points(d, type = "p", pch = plot_settings$pch, bg = i, col = plot_settings$col)
      }, split(results$data, results$data$group), 1:length(unique(results$data$group)))
    }

    ## add fitted curve
    if (!inherits(fit, c("error", "simpleError"))) {

      if (coord_flip) {
        oldx <- data[ ,2]
      } else {
        oldx <- data[ ,1]
      }

      newx <- seq(range(oldx)[1], range(oldx)[2], length.out = 10000)
      newy <- suppressWarnings(predict(fit, newdata = list(x = newx)))

      if (coord_flip) {
        tmp <- newx
        newx <- newy
        newy <- tmp
      }

      if (!global_fit) {
        points(newx, newy,
               type = "l",
               col = list(...)$line_col %||% "blue",
               lty = list(...)$line_lty %||% 1,
               lwd = list(...)$line_lwd %||% 1)
      } else {
        for (i in 1:length(data_list)) {
          seg <- seq(i * 101 - 100, 10000, nrow(data))
          points(newx[seg], newy[seg],
                 type = "l",
                 col = list(...)$line_col %||% i,
                 lty = list(...)$line_lty %||% 1,
                 lwd = list(...)$line_lwd %||% 1)
        }
      }

    } else {
      legend("center", legend = "Unable to fit the data", adj = 0.05)
    }

    # add error bars (if weighted fit)
    if (!is.null(error) && error_bars) {
      segments(plot_settings$x[ ,1], plot_settings$x[ ,2] - error,
               plot_settings$x[ ,1], plot_settings$x[ ,2] + error)

    }

    ## show the formula in the legend
    if (legend && !inherits(fit, "simpleError")) {
      formula_text <- paste0("y = ", as.character(fit$m$formula())[3])

      if (!is.null(age)) {
        if (!global_fit) {
          formula_text <- gsub("age", age, formula_text)
        } else {
          formula_text <- gsub("age", paste0("[", paste(age, collapse = "|"), "]"), formula_text)
          formula_text <- gsub("\\[group\\]", "", formula_text)
        }
      }
      if (!is.null(sigmaphi))
        formula_text <- gsub("sigmaphi", sigmaphi, formula_text)
      if (!is.null(mu)) {
        if (!global_fit) {
          formula_text <- gsub("mu", mu, formula_text)
        } else {
          formula_text <- gsub("mu", paste0("[", paste(mu, collapse = "|"), "]"), formula_text)
        }
      }

      legend(ifelse(coord_flip, "bottomleft", "bottomright"), legend = formula_text, cex = 0.8, bty = "n")
    }
  }

  ## CONSOLE ----
  if (settings$verbose) {
    cat("\n [fit_SurfaceExposure()] \n\n")

    if (global_fit) {
      cat(" Shared estimated parameter(s):\n")
    } else {
      cat(" Estimated parameter(s):\n")
    }
    cat(" ---------------------\n")
    if (is.null(age))
      cat(" age (a):\t", signif(results$summary$age, 3), "\u00B1",
          signif(results$summary$age_error, 3), "\n")
    if (is.null(sigmaphi))
      cat(" sigmaphi:\t", signif(unique(results$summary$sigmaphi), 3), "\u00B1",
          signif(unique(results$summary$sigmaphi_error), 3), "\n")
    if (is.null(mu))
      cat(" mu:\t\t", signif(unique(results$summary$mu), 3), "\u00B1",
          signif(unique(results$summary$mu_error), 3), "\n")
    cat("\n")

    ## STANDARD OUTPUT
    cat(" Fixed parameter(s):\n",
        "-------------------\n")
    if (!is.null(age))
      cat(" age (a):\t", .collapse(age, quote = FALSE), "\n")
    if (!is.null(sigmaphi))
      cat(" sigmaphi:\t", sigmaphi, "\n")
    if (!is.null(mu))
      cat(" mu:\t\t", .collapse(mu, quote = FALSE), "\n")
    cat("\n")

    if (!is.null(age)) {
      message("To apply the estimated parameters to a sample of unknown age run:\n\n",
              "fit_SurfaceExposure(data = ", capture.output(results$args[[1]]),
              ", sigmaphi = ", signif(unique(results$summary$sigmaphi), 3),
              ", mu = c(", .collapse(signif(results$summary$mu, 3), quote = FALSE),
              "))\n")
    }
  }

  ## EXIT ----
  return(results)
}

## load data
data("ExampleData.SurfaceExposure", envir = environment())
d1 <- ExampleData.SurfaceExposure$sample_1
d2 <- ExampleData.SurfaceExposure$sample_2
d3 <- ExampleData.SurfaceExposure$set_1
d4 <- ExampleData.SurfaceExposure$set_2

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fit_SurfaceExposure("test"),
               "'data' should be of class 'data.frame'")
  expect_error(fit_SurfaceExposure(list()),
               "'data' cannot be an empty list")
  expect_error(fit_SurfaceExposure(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(fit_SurfaceExposure(matrix()),
               "'data' should have at least two columns")
  expect_error(fit_SurfaceExposure(list(set_RLum("RLum.Data.Curve"))),
               "All elements of 'data' should be of class 'data.frame'")
  expect_error(fit_SurfaceExposure(d3),
               "'age' must be of the same length")
  expect_error(fit_SurfaceExposure(d4, age = 1e4),
               "'age' must be of the same length")
  expect_error(fit_SurfaceExposure(d1, age = 1e4, mu = numeric()),
               "'mu' cannot be an empty numeric")
  expect_error(fit_SurfaceExposure(list(d1), age = 1e4, mu = numeric()),
               "'mu' cannot be an empty numeric")
  expect_error(fit_SurfaceExposure(d4, age = rep(1e4, 5), mu = c(0.8, 0.9)),
               "'mu' must either be of the same length or of length 1")
  expect_error(fit_SurfaceExposure(d1, age = 1e4, plot = NA),
               "'plot' should be a single logical value")
  expect_error(fit_SurfaceExposure(data.frame(NA, 1:5)),
               "After NA removal, nothing is left from the data set")
  expect_error(fit_SurfaceExposure(data = d1, sigmaphi = 5e-10, mu = 0.9,
                                   verbose = NULL),
               "'verbose' should be a single logical value")
  SW({
  expect_message(fit_SurfaceExposure(rbind(d1, NA), mu = 0.9),
                 "\\[fit\\_SurfaceExposure\\(\\)\\] NA values in 'data' were removed")
  })
})

test_that("check values from example 1", {
  testthat::skip_on_cran()

  fit <- fit_SurfaceExposure(data = d1, sigmaphi = 5e-10, mu = 0.9,
                             plot = TRUE, verbose = FALSE)

  ## Example data 1
  expect_s4_class(fit, "RLum.Results")
  expect_equal(is(fit), c("RLum.Results", "RLum"))
  expect_equal(length(fit), 5)
  expect_equal(is(fit$fit), "nls")

  expect_equal(round(fit$summary$age), 9893)
  expect_equal(round(fit$summary$age_error), 369)
})

test_that("check values from example 1 - weighted fitting", {
  testthat::skip_on_cran()

  fit <- fit_SurfaceExposure(data = d1, sigmaphi = 5e-10, mu = 0.9, weights = TRUE,
                             plot = FALSE, verbose = FALSE)

  expect_equal(round(fit$summary$age), 9624)
  expect_equal(round(fit$summary$age_error), 273)
})

test_that("check values from example 2", {
  testthat::skip_on_cran()

  fit <- fit_SurfaceExposure(data = data.table(d2), age = 1e4,
                             sigmaphi = 5e-10, Ddot = 2.5, D0 = 40,
                             plot = FALSE, verbose = FALSE)

  expect_equal(is(fit), c("RLum.Results", "RLum"))
  expect_equal(length(fit), 5)
  expect_equal(is(fit$fit), "nls")

  expect_equal(round(fit$summary$mu, 3), 0.904)
  expect_equal(round(fit$summary$mu_error, 3), 0.007)
})

test_that("check values from example 3", {
  testthat::skip_on_cran()

  fit <- fit_SurfaceExposure(data = d3, age = c(1e3, 1e4, 1e5, 1e6), sigmaphi = 5e-10,
                             plot = FALSE, verbose = FALSE)

  expect_equal(is(fit), c("RLum.Results", "RLum"))
  expect_equal(nrow(fit$summary), 4)
  expect_equal(length(fit), 5)
  expect_equal(is(fit$fit), "nls")

  expect_equal(round(unique(fit$summary$mu), 3), 0.901)
  expect_equal(round(unique(fit$summary$mu_error), 3), 0.002)

  ## multiple mus given
  expect_s4_class(fit_SurfaceExposure(data = d3, age = c(1e3, 1e4, 1e5, 1e6),
                                      mu = 0.1 * 7:10, verbose = FALSE),
                  "RLum.Results")

  SW({
  ## all paramers specified, nothing to fit
  expect_message(fit_SurfaceExposure(data = d3, age = c(1e3, 1e4, 1e5, 1e6),
                                     mu = 0.1 * 7:10, sigmaphi = 5e-10),
                 "Error: Unable to fit the data")
  })
})

test_that("check values from example 4", {
  testthat::skip_on_cran()

  fit <- fit_SurfaceExposure(data = d4, age = c(1e2, 1e3, 1e4, 1e5, 1e6), sigmaphi = 5e-10,
                           Ddot = 1.0, D0 = 40,
                           plot = FALSE, verbose = FALSE)

  expect_equal(is(fit), c("RLum.Results", "RLum"))
  expect_equal(nrow(fit$summary), 5)
  expect_equal(length(fit), 5)
  expect_equal(is(fit$fit), "nls")

  expect_equal(round(unique(fit$summary$mu), 3), 0.899)
  expect_equal(round(unique(fit$summary$mu_error), 3), 0.002)
})

#### WARNINGS & FAILURES
test_that("not enough parameters provided", {
  testthat::skip_on_cran()

  SW({
  expect_message(
    fit_SurfaceExposure(data = d1, plot = FALSE, verbose = TRUE),
    "Unable to fit the data"
  )
  expect_equal(
    is(fit_SurfaceExposure(data = d2, plot = FALSE, verbose = FALSE)$fit),
    "simpleError"
  )
  expect_warning(
    fit_SurfaceExposure(data = d4, age = c(1e2, 1e3, 1e4, 1e5, 1e6), sigmaphi = 5e-10,
                        Ddot = 1.0, D0 = 40, weights = TRUE,
                        plot = TRUE, coord_flip = TRUE, verbose = TRUE),
    "is not supported when"
  )

  expect_message(res <- fit_SurfaceExposure(list(d1, d1), age = c(NA, 1e4),
                                            sigmaphi = NA, mu = NA,
                                            log = "y", plot = TRUE),
                 "Original error from minpack.lm::nlsLM(): evaluation of fn",
                 fixed = TRUE)
  expect_message(fit_SurfaceExposure(res),
                 "Original error from minpack.lm::nlsLM(): singular gradient",
                 fixed = TRUE)
  expect_message(fit_SurfaceExposure(as.matrix(d1), log = "y"),
                 "Original error from minpack.lm::nlsLM(): singular gradient",
                 fixed = TRUE)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("single",
                              fit_SurfaceExposure(d1, mu = 0.9,
                                                  age = 12000, sigmaphi = 5e-10))

  vdiffr::expect_doppelganger("multiple",
                              fit_SurfaceExposure(d3, mu = 0.1 * 7:10,
                                                  age = c(1e3, 1e4, 1e5, 1e6)))
  })
})

#' @title Export RLum.Data.Image and RLum.Data.Spectrum objects to TIFF Images
#'
#' @description
#' Simple wrapper around [tiff::writeTIFF] to export suitable [Luminescence::RLum-class]
#' objects to TIFF images. Per default 16-bit TIFF files are exported.
#'
#' @param object [Luminescence::RLum.Data.Image-class] or [Luminescence::RLum.Data.Spectrum-class] object (**required**):
#' input object, can be a [list] of such objects.
#'
#' @param file [character] (**required**):
#' name of the output file.
#'
#' @param norm [numeric] (*with default*):
#' normalisation value. Usually, in imaging applications the pixel values are
#' integer count values, but values in TIFF files must be in the 0-1 range.
#' Normalising to the to the highest 16-bit integer values - 1 ensures that
#' the numerical values are retained in the exported image. If `1` nothing is
#' normalised.
#'
#' @param ... further arguments to be passed to [tiff::writeTIFF].
#'
#' @return A TIFF file
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @section Function version: 0.1.2
#'
#' @seealso [tiff::writeTIFF], [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Data.Spectrum-class]
#'
#' @keywords IO
#'
#' @examples
#' data(ExampleData.RLum.Data.Image, envir = environment())
#' write_R2TIFF(ExampleData.RLum.Data.Image, file = tempfile(fileext = ".tiff"))
#'
#' @export
write_R2TIFF <- function(
  object,
  file = tempfile(),
  norm = 65535,
  ...
) {
  .set_function_name("write_R2TIFF")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  ## most of the users don't need this import, no need to bother them
  ## with required libraries
  .require_suggested_package("tiff", "Exporting objects to TIFF files")

  ## make a list ... it is just easier
  if (!inherits(object, "list"))
    object <- list(object)

  ## check list input
  sapply(object, function(x) {
    .validate_class(x, c("RLum.Data.Image", "RLum.Data.Spectrum"),
                    extra = "a 'list' of such objects",
                    name = "'object'")
    if (inherits(x, "RLum.Data.Image") && length(dim(x@data)) != 3)
      .throw_error("Empty RLum.Data.Image object detected")
  })

  .validate_class(file, "character", length = 1)
  .validate_positive_scalar(norm)

  ## Prepare filenames ------------------------------------------------------

  file <- normalizePath(file, mustWork = FALSE)
  file_dir <- dirname(file)
  if (!dir.exists(file_dir))
    .throw_error("Path '", file_dir, "' does not exist")

  ## expand if longer than 1
  if(length(object) > 1)
    file <- paste0(tools::file_path_sans_ext(file), "_", 1:length(object), ".tiff")

  ## Export to TIFF ---------------------------------------------------------

  ## modify arguments
  args <- modifyList(x = list(
    bits.per.sample = 16L
  ), list(...))

  ## remove arguments we already use
  args[c("what", "where")] <- NULL

  for (i in seq_along(object)) {
    data <- object[[i]]@data

    ## consider the case that we have already an image stack
    if (length(dim(data)) > 2 && dim(data)[3] > 1) {
      img_list <- lapply(1:dim(data)[3], function(x) {
        m <- data[, , x]
        storage.mode(m) <- "numeric"
        m / norm
      })
    } else {
      img_list <- data / norm
    }

    ## write file
    do.call(what = tiff::writeTIFF, args = c(list(img_list, where = file[i]), args))
  }
}

## load data
data(ExampleData.RLum.Data.Image, envir = environment())
data(ExampleData.XSYG, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(write_R2TIFF(object = "test"),
               "[write_R2TIFF()] 'object' should be of class 'RLum.Data.Image'",
               fixed = TRUE)
  expect_error(write_R2TIFF(ExampleData.RLum.Data.Image, file = NULL),
               "'file' should be of class 'character'")
  expect_error(write_R2TIFF(ExampleData.RLum.Data.Image, file = character()),
               "'file' should be of class 'character' and have length 1")
  expect_error(write_R2TIFF(ExampleData.RLum.Data.Image, norm = 0),
               "'norm' should be a single positive value")
  expect_error(write_R2TIFF(ExampleData.RLum.Data.Image, file = "error/error"),
               "Path '.*error' does not exist") # windows CI needs the regexp
  expect_error(write_R2TIFF(set_RLum("RLum.Data.Image")),
               "Empty RLum.Data.Image object detected")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## export RLum.Data.Image
  expect_null(write_R2TIFF(ExampleData.RLum.Data.Image, file = tempfile(fileext = "tiff")))

  ## export image stack
  t_stack <- set_RLum(
    "RLum.Data.Image",
    data = array(c(ExampleData.RLum.Data.Image@data, ExampleData.RLum.Data.Image@data),
                 dim = c(1340, 101, 2)))
  expect_null(write_R2TIFF(t_stack, file = tempfile(fileext = "tiff")))

  ## export RLum.Data.Spectrum
  expect_null(write_R2TIFF(TL.Spectrum, file = tempfile(fileext = "tiff")))

  ## a list
  expect_null(write_R2TIFF(list(ExampleData.RLum.Data.Image, TL.Spectrum), file = tempfile(fileext = "tiff")))
  expect_null(write_R2TIFF(list()))
})

#' @title General merge function for RLum-class objects
#'
#' @description
#' The function provides a generalised access point for merging specific
#' [Luminescence::RLum-class] objects. Depending on the input object, the
#' corresponding merge function will be selected.  Allowed arguments can be
#' found in the documentation of each merge function.
#' Empty list elements (`NULL`) are automatically removed from the input list.
#'
#' \tabular{lll}{
#' **object** \tab \tab **corresponding merge function** \cr
#' [Luminescence::RLum.Data.Curve-class] \tab -> \tab [Luminescence::merge_RLum.Data.Curve] \cr
#' [Luminescence::RLum.Data.Spectrum-class] \tab -> \tab [Luminescence::merge_RLum.Data.Spectrum] \cr
#' [Luminescence::RLum.Analysis-class] \tab -> \tab [Luminescence::merge_RLum.Analysis] \cr
#' [Luminescence::RLum.Results-class] \tab -> \tab [Luminescence::merge_RLum.Results]
#' }
#'
#' @param objects [list] of [Luminescence::RLum-class] (**required**):
#' list of S4 object of class `RLum`.
#'
#' @param ... further arguments that one might want to pass to the specific merge function
#'
#' @return
#' Returns an [Luminescence::RLum.Analysis-class] object of class if any of the inputs is
#' of that class. Otherwise, it returns an object of the same type as the
#' input.
#'
#' @note
#' So far merging of [Luminescence::RLum.Data.Image-class] objects is not supported.
#'
#' @section Function version: 0.1.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ##Example based using data and from the calc_CentralDose() function
#'
#' ##load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ##apply the central dose model 1st time
#' temp1 <- calc_CentralDose(ExampleData.DeValues$CA1)
#'
#' ##apply the central dose model 2nd time
#' temp2 <- calc_CentralDose(ExampleData.DeValues$CA1)
#'
#' ##merge the results and store them in a new object
#' temp.merged <- get_RLum(merge_RLum(objects = list(temp1, temp2)))
#'
#' @export
merge_RLum<- function(
  objects,
  ...
) {
  .set_function_name("merge_RLum")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(objects, "list")

    ##we are friendly and remove all empty list elements, this helps a lot if we place things
    ##we DO NOT provide a warning as this lowers the computation speed in particular cases.
    objects <- .rm_NULL_elements(objects)

  ## if the list is empty we do nothing
  if (length(objects) < 1) {
    .throw_warning("Nothing was merged as the object list was found ",
                   "to be empty or contains only one object")
    return(NULL)
  }

  ## check if objects are of class RLum
  temp.class.test <- unique(sapply(objects, function(x) {
       .validate_class(x, "RLum",
                       name = "All elements of 'objects'")
        is(x)[1]
      }))

  ## objects must be consistent, unless there's an RLum.Analysis object
  if (length(temp.class.test) > 1 && !"RLum.Analysis" %in% temp.class.test) {
    .throw_error("Objects cannot be merged, different classes found: ",
                 .collapse(temp.class.test))
  }

  ## determine the output class
  objects.class <-
        ifelse("RLum.Analysis" %in% temp.class.test, "RLum.Analysis", temp.class.test)

  ## select which merge function should be used
  switch(
        objects.class,
        RLum.Data.Image = .throw_error("Merging of 'RLum.Data.Image' objects is currently not supported"),
        RLum.Data.Spectrum = merge_RLum.Data.Spectrum(objects, ...),
        RLum.Data.Curve = merge_RLum.Data.Curve(objects, ...),
        RLum.Analysis = merge_RLum.Analysis(objects, ...),
        RLum.Results = merge_RLum.Results(objects, ...)
      )
}

## load data
data(ExampleData.RLum.Analysis, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(analyse_IRSAR.RF(),
               "is missing, with no default")
  expect_error(analyse_IRSAR.RF("test"),
               "'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_IRSAR.RF(iris),
               "'object' should be of class 'RLum.Analysis' or a 'list'")
  expect_error(analyse_IRSAR.RF(list(iris)),
               "All elements of 'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_IRSAR.RF(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, n.MC = 0),
               "'n.MC' should be a single positive integer value")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, method = "error"),
               "'method' should be one of 'FIT', 'SLIDE', 'VSLIDE' or 'NONE'")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, method_control = 3),
               "'method_control' should be of class 'list'")
  expect_warning(expect_null(analyse_IRSAR.RF(list())),
                 "Nothing was merged as the object list was found to be empty")

  ## sequence_structure
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, sequence_structure = FALSE),
               "'sequence_structure' should be of class 'character'")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, sequence_structure = ""),
               "'sequence_structure' should contain at least two elements")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data,
                                sequence_struct = c("NATURAL", "NATURAL")),
               "'sequence_structure' must contain one each of 'NATURAL' and")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data,
                                sequence_struct = c("REGENERATED", "REGENERATED")),
               "'sequence_structure' must contain one each of 'NATURAL' and")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data,
                                sequence_struct = c("NATURAL", "NATURAL", "REGENERATED")),
               "'sequence_structure' is missing one of 'NATURAL' or")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data,
                                sequence_struct = c("REGENERATED", "NATURAL")),
               "The number of data channels for the natural (524) cannot exceed",
               fixed = TRUE)

  ## RF_nat.lim
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, RF_nat.lim = "error"),
               "'RF_nat.lim' should be of class 'numeric', 'integer' or NULL")
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, RF_nat.lim = 6),
                 "'RF_nat.lim' out of bounds, reset to c(1, 5)", fixed = TRUE)

  ## RF_reg.lim
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, RF_reg.lim = "error"),
               "'RF_reg.lim' should be of class 'numeric', 'integer' or NULL")
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, RF_reg.lim = 2000),
                 "'RF_reg.lim' out of bounds, reset to c(1, 524)", fixed = TRUE)
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, RF_reg.lim = 521),
               "'RF_reg.lim' defines too short an interval and it's not")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, RF_reg.lim = 520),
               "The range of regenerated channels should be larger than")
  suppressWarnings( # FIXME(mcol): lmdif: info = -1. Number of iterations has reached `maxiter' == 50.
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, RF_reg.lim = c(3, 6)),
                 "'RF_reg.lim' defines too short an interval, reset to c(3, 8)",
                 fixed = TRUE)
  )

  ## curves of the same length
  expect_error(analyse_IRSAR.RF(set_RLum("RLum.Analysis",
                                         records = list(IRSAR.RF.Data[[1]],
                                                        IRSAR.RF.Data[[1]])),
                                method = "VSLIDE"),
               "There is no further sliding space left")

  SW({
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data,
                                  method_control = list(unknown = "test")),
                 "'unknown' not supported for 'method_control'")

  ## disable test that produces this error on CI:
  ## Error in `.check_ncores(length(names))`: 4 simultaneous processes spawned
  if (FALSE) {
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                  method_control = list(cores = 10000)),
                 "Number of cores limited to the maximum available")
  }

  ## vslide_range
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                method_control = list(vslide_range = FALSE)),
                 "'vslide_range' in 'method_control' should be of class")

  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                method_control = list(vslide_range = "error")),
                 "'vslide_range' in 'method_control' should be either 'auto'")
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                  method_control = list(vslide_range = 1:4)),
                 "'vslide_range' in 'method_control' has more than 2 elements")
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, method = "SLIDE",
                                method_control = list(vslide_range = c(0, 1e7))),
               "[:::src_analyse_IRSAR_SRS()] 'vslide_range' exceeded maximum size (1e+07)",
               fixed = TRUE)

  ## num_slide_windows
  expect_error(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                method_control = list(num_slide_windows = NA)),
                 "'num_slide_windows' in 'method_control' should be a single positive")
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                  method_control = list(num_slide_windows = 20)),
                 "should be between 1 and 10, reset to 10")

  expect_message(analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE",
                                  method_control = list(cores = "4")),
                 "Invalid value for control argument 'cores'")
  })

  ## deprecated option
  expect_warning(analyse_IRSAR.RF(IRSAR.RF.Data, verbose = FALSE,
                                  method.control = list(cores = 1)),
                 "'method.control' was deprecated in v1.0.0, use 'method_control'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  expect_snapshot_RLum(analyse_IRSAR.RF(IRSAR.RF.Data, method = "FIT",
                                        plot = FALSE))
  SW({
  expect_snapshot_RLum(
    analyse_IRSAR.RF(
      object = IRSAR.RF.Data,
      plot = FALSE,
      method = "SLIDE",
      n.MC = 10,
      method_control = list(vslide_range = 'auto', trace_vslide = TRUE,
                            num_slide_windows = 10),
      txtProgressBar = FALSE
    )
  )

  expect_snapshot_RLum(
    analyse_IRSAR.RF(
      object = IRSAR.RF.Data,
      plot = FALSE,
      method = "VSLIDE",
      n.MC = 10,
      method_control = list(vslide_range = 'auto', trace_vslide = FALSE,
                            num_slide_windows = 10),
      txtProgressBar = FALSE
    )
  )
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("fit",
                              analyse_IRSAR.RF(IRSAR.RF.Data,
                                               method = "FIT",
                                               n.MC = NULL))
  vdiffr::expect_doppelganger("slide",
                              analyse_IRSAR.RF(IRSAR.RF.Data,
                                               method = "SLIDE",
                                               n.MC = NULL))
  vdiffr::expect_doppelganger("vslide",
                              analyse_IRSAR.RF(IRSAR.RF.Data,
                                               method = "VSLIDE",
                                               n.MC = NULL))
  vdiffr::expect_doppelganger("none subtitle log",
                              analyse_IRSAR.RF(IRSAR.RF.Data,
                                               method = "None",
                                               mtext = "Subtitle",
                                               log = "xy",
                                               n.MC = 10,
                                               txtProgressBar = FALSE))
  })
})

test_that("test support for IR-RF data", {
  testthat::skip_on_cran()

  ## get needed data
  file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
  temp <- read_RF2R(file, verbose = FALSE)

  SW({
  expect_warning(expect_s4_class(
      analyse_IRSAR.RF(object = temp[1:3], method = "SLIDE",
                       cex = 1.1, xlim = c(750, 9000), ylim = c(640, 655),
                       method_control = list(show_fit = TRUE),
                       plot_reduced = TRUE, n.MC = 1),
      "RLum.Results"),
      "Narrow density distribution, no density distribution plotted")
  })
})

test_that("test edge cases", {
  testthat::skip_on_cran()

  data(ExampleData.RLum.Analysis, envir = environment())
  RF_nat <- RF_reg <- IRSAR.RF.Data[[2]]
  RF_reg@data[,2] <- runif(length(RF_reg@data[,2]), 0.007557956, 0.05377426 )
  RF_nat@data[,2] <- runif(length(RF_nat@data[,2]), 65.4, 76.7)
  RF_nat@data <- RF_nat@data[1:50,]
  RF_nat@info <- list(startDate = "20210101150845")
  object <- set_RLum("RLum.Analysis", records = list(RF_nat, RF_reg))

  SW({
  expect_warning(analyse_IRSAR.RF(
    IRSAR.RF.Data,
    method = "FIT",
    plot = TRUE,
    RF_reg = c(1, 400),
    txtProgressBar = FALSE),
    "Threshold exceeded for: 'curves_bounds'")

  expect_warning(expect_s4_class(analyse_IRSAR.RF(
    list(object),
    method = "SLIDE",
    method_control = list(vslide_range = 'auto', correct_onset = FALSE,
                          show_fit = TRUE, trace = TRUE, n.MC = 2,
                          cores = 2),
    RF_nat.lim = 2,
    RF_reg.lim = 2,
    plot = TRUE,
    main = "Title",
    mtext = "Subtitle",
    log = "x",
    txtProgressBar = FALSE),
    "RLum.Results"),
    "Threshold exceeded for: 'curves_ratio'")
  })

  ## this RF_nat.lim after
  ##  'length = 2' in coercion to 'logical(1)' error
  expect_s4_class(suppressWarnings(analyse_IRSAR.RF(
    object,
    method = "SLIDE",
    method_control = list(vslide_range = 'auto', correct_onset = FALSE),
    RF_nat.lim = c(10,100),
    #RF_reg.lim = c(),
    plot = TRUE,
    txtProgressBar = FALSE
  )), "RLum.Results")

  expect_s4_class(suppressWarnings(analyse_IRSAR.RF(
    object,
    method = "FIT",
    mtext = "FIT method",
    plot = TRUE,
    txtProgressBar = FALSE
  )), "RLum.Results")

  ## test parameters values only set for coverage
  SW({
  expect_s4_class(analyse_IRSAR.RF(
    object,
    method = "SLIDE",
    method_control = list(vslide_range = 'auto', correct_onset = FALSE),
    #RF_nat.lim = c(10,100),
    RF_reg.lim = c(10,100),
    plot = TRUE,
    test_parameters = list(curves_ratio = NULL,
                           intersection_ratio = 5,
                           residuals_slope = 1,
                           dynamic_ratio = 1,
                           lambda = 0,
                           beta = 1e-4,
                           delta.phi = 1e-4),
    txtProgressBar = FALSE
  ), "RLum.Results")
  })

  ## more coverage
  expect_s4_class(analyse_IRSAR.RF(
      IRSAR.RF.Data,
      method = "SLIDE",
      RF_reg.lim = c(1, 7),
      verbose = FALSE),
      "RLum.Results")

  tmp <- IRSAR.RF.Data
  tmp@records[[2]]@data <- tmp@records[[2]]@data[1:460, ]
  expect_s4_class(analyse_IRSAR.RF(
      tmp,
      method = "SLIDE",
      verbose = FALSE),
      "RLum.Results")
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 372
  expect_silent(analyse_IRSAR.RF(list(IRSAR.RF.Data), plot = FALSE,
                                 method_control = list(maxiter = 10)))

  ## issue 382
  expect_silent(analyse_IRSAR.RF(list(IRSAR.RF.Data), plot = FALSE))

  ## issue 816
  data(ExampleData.portableOSL, envir = environment())
  expect_silent(analyse_IRSAR.RF(merge_RLum(ExampleData.portableOSL)[1:4],
                                 plot = FALSE))
  expect_silent(analyse_IRSAR.RF(merge_RLum(ExampleData.portableOSL)[1:9],
                                 sequence_structure = c("NATURAL", rep("REGENERATED", 2)),
                                 plot = FALSE))

  ## issue 1055
  expect_silent(analyse_IRSAR.RF(IRSAR.RF.Data, method = "FIT", n.MC = NULL))
})

## load example data (sample UNIL/NB123, see ?ExampleData.Fading)
data("ExampleData.Fading", envir = environment())
fading_data <- ExampleData.Fading$fading.data$IR50

## create artificial object
set.seed(0)
time <- as.POSIXct("2025-09-26 10:26:20 CEST")
ll <- list()
for (x in runif(3, 120, 130)) {
  irr <- set_RLum("RLum.Data.Curve",
                  data = matrix(c(1:x, rep(1, x)), ncol = 2),
                  originator = "read_XSYG2R",
                  recordType = "irradiation (NA)",
                  curveType = "simulated",
                  info = list(startDate = format(time, "%Y%m%d%H%M%S"),
                              position = 1))
  lum <- set_RLum("RLum.Data.Curve",
                  data = matrix(c(1:40, exp(-c(1:40) / x * 10)), ncol = 2),
                  originator = "read_XSYG2R",
                  recordType = "IRSL",
                  curveType = "measured",
                  info = list(startDate = format(time + x + 30, "%Y%m%d%H%M%S"),
                              position = 1))
  time <- time + x + 60
  ll <- c(ll, irr, lum, lum)
}
object <- set_RLum("RLum.Analysis", records = ll, originator = "read_XSYG2R")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(analyse_FadingMeasurement(object = "test"),
               "'object' should be of class 'RLum.Analysis', 'data.frame' or")
  expect_error(analyse_FadingMeasurement(object = iris[0, ]),
               "'object' cannot be an empty data.frame")
  expect_error(analyse_FadingMeasurement(object = iris[, 0]),
               "'object' cannot be an empty data.frame")
  expect_warning(expect_error(
      analyse_FadingMeasurement(list(fading_data, "test")),
      "No valid records in 'object' left"),
      "2 unsupported records removed")
  expect_error(analyse_FadingMeasurement(cbind(fading_data, fading_data[, 1])),
               "When 'object' is a data.frame, the number of columns must be")
  expect_error(analyse_FadingMeasurement(data.frame(NA, 1:5, 1:5)),
               "After NA removal, nothing is left from the data set")
  expect_error(analyse_FadingMeasurement(data.frame(1:3, 1:3, -1:1 + 0.1)),
               "'object' cannot contain negative irradiation times")

  expect_error(analyse_FadingMeasurement(object, signal_integral = 1:2,
                                         background_integral = 2),
               "'background_integral' is of length 0 after removing values smaller than 3")
  expect_error(analyse_FadingMeasurement(object[-3], signal_integral = 1:2,
                                         background_integral = 3:40,
                                         structure = c("Lx", "Tx")),
               "The number of Lx curves (3) differs from the number of Tx curves (2)",
               fixed = TRUE)

  ## check various for t_star
  ## stop t_star
  expect_error(analyse_FadingMeasurement(fading_data, t_star = "error"),
               "'t_star' should be one of 'half', 'half_complex', 'end' or a function")

  ## wrong originator
  psl.file <- system.file("extdata/DorNie_0016.psl", package = "Luminescence")
  SW({
  expect_message(expect_null(analyse_FadingMeasurement(read_PSL2R(psl.file))),
                 "Error: Unknown or unsupported originator, NULL returned")

  ## missing originator
  data(ExampleData.RLum.Analysis, envir = environment())
  expect_message(expect_null(analyse_FadingMeasurement(IRSAR.RF.Data)),
                 "Error: Unknown or unsupported originator, NULL returned")
  expect_message(expect_null(analyse_FadingMeasurement(list())),
                 "Error: Unknown or unsupported originator, NULL returned")

  ## no irradiation steps
  xsyg.file <- system.file("extdata/XSYG_file.xsyg", package = "Luminescence")
  expect_error(analyse_FadingMeasurement(read_XSYG2R(xsyg.file, fastForward = TRUE)),
               "No irradiation times could be retrieved, check that 'object'")
  })
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## run routine analysis
  SW({
  expect_s4_class(analyse_FadingMeasurement(
    fading_data,
    plot = TRUE,
    verbose = TRUE,
    n.MC = 10), class = "RLum.Results")
  })

  ## test merging of objects if combined in a list
  ## this crashed before and was fixed
  expect_s4_class(merge_RLum(
    list(analyse_FadingMeasurement(
      fading_data[1,],
      plot = FALSE,
      verbose = FALSE,
      n.MC = 10),
    analyse_FadingMeasurement(
      fading_data[1:10,],
      plot = FALSE,
      verbose = FALSE,
      n.MC = 10))), "RLum.Results")

  expect_warning(analyse_FadingMeasurement(fading_data,
                                           plot = TRUE, plot.single = TRUE,
                                           verbose = FALSE,
                                           n.MC = 10),
                 "'plot.single' was deprecated in v1.0.0, use 'plot_singlePanels'")

  ## more coverage
  data.inf <- fading_data
  data.inf$LxTx[24] <- Inf
  expect_s4_class(analyse_FadingMeasurement(
    data.inf,
    plot = FALSE,
    verbose = FALSE,
    n.MC = 10), class = "RLum.Results")

  ## deprecated arguments
  expect_warning(analyse_FadingMeasurement(
      object,
      signal.integral = 1:2,
      background.integral = 10:40,
      verbose = FALSE,
      n.MC = 10),
      "deprecated in v1.2.0, use 'signal_integral' and 'background_integral'")
  expect_warning(expect_error(analyse_FadingMeasurement(
      object,
      signal.integral = 1:2,
      background.integral = 10:40,
      integral_input = "measurement",
      verbose = FALSE,
      n.MC = 10),
      "'integral_input' is not supported with old argument names"),
      "deprecated in v1.2.0, use 'signal_integral' and 'background_integral'")
})

test_that("test XSYG file fading data", {
  testthat::skip_on_cran()

  SW({
  expect_s4_class(analyse_FadingMeasurement(
    object,
    signal_integral = 1:2,
    background_integral = 10:40,
    structure = "Lx"
  ), "RLum.Results")
  })

  SW({
  expect_s4_class(analyse_FadingMeasurement(
    object,
    signal_integral = 1:2,
    t_star = "end",
    background_integral = 10:40,
    structure = "Lx",
    plot = FALSE
  ), "RLum.Results")
  })

  obj.sub <- get_RLum(object, record.id = 1:2, drop = FALSE)
  expect_message(expect_null(analyse_FadingMeasurement(obj.sub)),
                 "After irradiation step removal not enough curves are left")

  SW({
  expect_warning(analyse_FadingMeasurement(object[-c(3,6,9)], signal_integral = 1:2,
                                           background_integral = 3,
                                           structure = "Lx"),
                 "Number of background channels for Lx < 25")

  object@records[[3]]@data <- object@records[[3]]@data[1:10, ]
  expect_warning(analyse_FadingMeasurement(object, signal_integral = 1:2,
                                           background_integral = 3:40),
                 "Skipped the following samples because Lx and Tx have different sizes: 1")
  })

  obj.mod <- object
  obj.mod@records[[6]]@data <- obj.mod@records[[6]]@data[1:10, ]
  obj.mod@records[[9]]@data <- obj.mod@records[[9]]@data[1:10, ]
  expect_error(analyse_FadingMeasurement(obj.mod, signal_integral = 1:2,
                                         background_integral = 3:40),
               "No curves left after removing those with different Lx and Tx sizes")
})

test_that("test BIN file while fading data", {
  testthat::skip_on_cran()

  data(ExampleData.BINfileData, envir = environment())
  d1 <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
  expect_error(analyse_FadingMeasurement(d1),
               "BIN-file has version 03, but only versions from 05 on are supported")

  SW({
  d2 <- read_BIN2R(test_path("_data/BINfile_V5.binx"), verbose = FALSE,
                   fastForward = TRUE)
  })
  expect_output(analyse_FadingMeasurement(d2, signal_integral = 1:2,
                                          background_integral = 10:30,
                                          plot = TRUE))
  expect_message(expect_null(
      analyse_FadingMeasurement(d2, signal_integral = 1:2,
                                background_integral = 10:30,
                                structure = c("Lx", "Tx", "Lx"))),
      "Error: 'structure' can only be 'Lx' or c('Lx', 'Tx'), NULL returned",
      fixed = TRUE)

  ## integral_input
  set.seed(1)
  res1 <- analyse_FadingMeasurement(d2,
                                    signal_integral = c(220.004, 220.012),
                                    background_integral = c(220.760, 221),
                                    integral_input = "measurement",
                                    t_star = "end",
                                    structure = "Lx",
                                    plot = FALSE,
                                    verbose = FALSE)
  set.seed(1)
  res2 <- analyse_FadingMeasurement(d2,
                                    signal_integral = 1:3,
                                    background_integral = 190:250,
                                    integral_input = "channel",
                                    t_star = "end",
                                    structure = "Lx",
                                    plot = FALSE,
                                    verbose = FALSE)
  res1@info <- res2@info <- list() # remove $call
  res1@.uid <- res2@.uid <- NA_character_
  res1@data$fit <- res2@data$fit <- NULL
  expect_equal(res1, res2)

  ## more coverage
  analyse_FadingMeasurement(d2, signal_integral = 1:2,
                            background_integral = 10:30,
                            signal_integral_Tx = 2,
                            background_integral_Tx = 5:30,
                            plot_singlePanels = 2:3, ylim = c(0.1, 1.1),
                            background.count.distribution = "poisson",
                            sig0 = 2, verbose = FALSE, plot = TRUE)

  d2[[2]]@records[[1]]@info$TIMESINCEIRR <- -1
  expect_message(expect_warning(expect_null(
      analyse_FadingMeasurement(d2, signal_integral = 1:2,
                                background_integral = 10:30,
                                verbose = TRUE)),
      "removed 2 records with negative 'time since irradiation'"),
      "After record removal nothing is left from the data set, NULL returned")
  suppressWarnings( # repeated warning about negative time since irradiation
  expect_warning(expect_s4_class(
      analyse_FadingMeasurement(d2, signal_integral = 1:2,
                                background_integral = 10:40,
                                structure = "Lx", verbose = FALSE),
      "RLum.Results"),
      "removed 1 records with negative 'time since irradiation'")
  )

  fd.mod <- fading_data
  fd.mod[6, 1] <- 0.712
  expect_output(analyse_FadingMeasurement(fd.mod, n.MC = 10))
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 558
  df <- data.frame(LxTx = c(1, 0.9879, 0.9865, 0.9872, 0.9759, 0.9978, 0.9879, 0.98372),
                   LxTxError = rep(0, 8),
                   timeSinceIrr = c(2516, 41353, 50357, 140342, 1040044, 2516, 41360, 50360))
  res <- analyse_FadingMeasurement(df, n.MC = 10, plot = FALSE, verbose = FALSE)
  expect_false(is.nan(res$rho_prime$MEAN))

  ## issue 616
  expect_output(analyse_FadingMeasurement(df[1:2, ]))

  ## issue 1132
  class(iris) <- c("data.frame", "extra")
  expect_error(expect_warning(analyse_FadingMeasurement(list(iris)),
                              "1 unsupported records removed"),
               "No valid records in 'object' left")

  ## issue 1144
  df[1, 1] <-  0
  expect_error(analyse_FadingMeasurement(df, n.MC = 10, plot = TRUE),
               "Normalisation term is not positive, check your input")
  df <- data.frame(Dose = c(1, 450, 1050, 2000, 2550, 450, 0),
                   LxTx = c(4.269, 1.532, 3.007, 4.571, 5.259, 1.609, 0.018),
                   LxTx.Error = c(0.076, 0.029, 0.054, 0.081, 0.09, 0.029, 0.002))
  expect_error(analyse_FadingMeasurement(df, n.MC = 10, plot = TRUE),
               "Normalisation term is not positive, check your input")

  ## issue 1236
  input <- template_DRAC(preset = "DRAC-example_quartz", notification = FALSE)
  expect_warning(expect_error(analyse_FadingMeasurement(input),
                              "No valid records in 'object' left"),
                 "53 unsupported records removed")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  set.seed(1)
  vdiffr::expect_doppelganger("default",
                              analyse_FadingMeasurement(fading_data,
                                                        n.MC = 10))
  vdiffr::expect_doppelganger("panel 3",
                              analyse_FadingMeasurement(fading_data,
                                                        plot_singlePanels = 3,
                                                        n.MC = 10))
  vdiffr::expect_doppelganger("panel 4",
                              analyse_FadingMeasurement(fading_data,
                                                        plot_singlePanels = 4,
                                                        n.MC = 10))
  vdiffr::expect_doppelganger("lx half_complex",
                              analyse_FadingMeasurement(object,
                                                        signal_integral = 1:2,
                                                        background_integral = 10:40,
                                                        structure = "Lx",
                                                        t_star = "half_complex",
                                                        n.MC = 10))
  vdiffr::expect_doppelganger("lx tx identity",
                              analyse_FadingMeasurement(object,
                                                        signal_integral = 1:2,
                                                        background_integral = 10:40,
                                                        structure = c("Lx", "Tx"),
                                                        t_star = identity,
                                                        n.MC = 10))
  vdiffr::expect_doppelganger("integral tx",
                              analyse_FadingMeasurement(object,
                                                        signal_integral = 1:2,
                                                        background_integral = 10:40,
                                                        signal_integral_Tx = 4:8,
                                                        background_integral_Tx = 30:40,
                                                        structure = c("Lx", "Tx"),
                                                        n.MC = 10))
  })
})

## load example data
data(ExampleData.XSYG, envir = environment())
TL.Spectrum_types <- TL.Spectrum
TL.Spectrum_types@recordType <- "OSL (Spectrum)"
TL.Spectrum_short <- TL.Spectrum
TL.Spectrum_short@data <- TL.Spectrum_short@data[- 1, ]
TL.Spectrum_zeros <- TL.Spectrum
TL.Spectrum_zeros@data[10:12, 2] <- 0
TL.curve.1 <- get_RLum(OSL.SARMeasurement$Sequence.Object,
                       recordType = "TL (UVVIS)")[[1]]

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(merge_RLum.Data.Spectrum("error", merge.method = "/"),
               "'object' should be of class 'list'")
  expect_error(merge_RLum.Data.Spectrum(list("error")),
               "All elements of 'object' should be of class 'RLum.Data.Spectrum'")
  expect_error(merge_RLum.Data.Spectrum(list(), merge.method = "-"),
               "'object' contains no data")

  expect_error(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                               merge.method = "error"),
               "'merge.method' should be one of 'mean', 'median', 'sum', 'sd'")
  expect_error(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                               method.info = "error"),
               "'method.info' should be a single positive integer value or NULL")
  expect_error(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum_types)),
               "Objects cannot be merged, different record types found")

  expect_error(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum_short)),
               "'RLum.Data.Spectrum' objects of different size cannot be merged")
  TL.Spectrum_other <- TL.Spectrum
  rownames(TL.Spectrum_other@data) <- 1:nrow(TL.Spectrum_other@data)
  expect_error(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum_other)),
               "'RLum.Data.Spectrum' objects with different channels cannot")
  TL.Spectrum_other <- TL.Spectrum
  TL.Spectrum_other@info$cameraType <- "other"
  expect_error(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum_other)),
               "'RLum.Data.Spectrum' objects from different camera types")

  ## time/temperature differences
  TL.Spectrum_other <- TL.Spectrum
  colnames(TL.Spectrum_other@data) <- as.numeric(colnames(TL.Spectrum@data)) + 1
  expect_warning(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum_other)),
                 "The time/temperatures recorded are too different")
  expect_silent(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum_other),
                                         max.temp.diff = 1))
  spectrum <- set_RLum("RLum.Data.Spectrum", data = matrix(1:10, ncol = 2))
  expect_no_warning(merge_RLum(list(spectrum, spectrum)))
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## only one spectrum
  expect_s4_class(new <- merge_RLum.Data.Spectrum(list(TL.Spectrum)),
                  "RLum.Data.Spectrum")
  expect_equal(TL.Spectrum@data,
               new@data)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                method.info = 1))
  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "sum"))
  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "median"))

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "sd"))

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "var"))

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "max"))

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "min"))

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "-"))

  expect_s4_class(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum, TL.Spectrum),
                                           merge.method = "-"), "RLum.Data.Spectrum")

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "append"))

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "*"))

  expect_s4_class(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum, TL.Spectrum),
                                                merge.method = "*"), "RLum.Data.Spectrum")

  expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                                merge.method = "/"))

  expect_s4_class(merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum, TL.Spectrum),
                                           merge.method = "/"), "RLum.Data.Spectrum")

  expect_warning(
      expect_snapshot_RLum(merge_RLum.Data.Spectrum(list(TL.Spectrum,
                                                         TL.Spectrum_zeros),
                                                    merge.method = "/")),
      "3 'inf' values replaced by 0 in the matrix")
})

## load data
data(ExampleData.portableOSL, envir = environment())

## generate test data set for profile
merged <- merge_RLum(ExampleData.portableOSL)

test_that("check functionality", {
    testthat::skip_on_cran()

    ## standard run profile
    results <- expect_s4_class(
      analyse_portableOSL(
        merged,
        signal_integral = 1:5,
        invert = FALSE,
        mode = "profile",
        normalise = TRUE,
        plot = TRUE
      ), "RLum.Results")

    ## check standard with coordinates
    coord <- as.matrix(results$summary[,c("COORD_X", "COORD_Y")])

    ## verify output
      expect_equal(length(results), 3)
      expect_s3_class(results$summary, "data.frame")
      expect_s4_class(results$data, "RLum.Analysis")
    expect_equal(round(sum(results$summary[,c(-1, -2, -10, -11)]), 2),
                 82.1)

    ## standard surface
    results <- expect_s4_class(
        analyse_portableOSL(
          merged,
          signal_integral = 1:5,
          invert = FALSE,
          mode = "surface",
          normalise = TRUE,
          plot = TRUE
        ), "RLum.Results")

    ## surface with options
    set.seed(1234)
    results <- expect_s4_class(
      analyse_portableOSL(
        merged,
        signal_integral = 1:5,
        invert = TRUE,
        mode = "surface",
        xlim = c(0.1, 0.6),
        ylim = c(0.1, 0.8),
        zlim = c(0.1, 2),
        zlim_image = c(1, 2),
        col_ramp = grDevices::topo.colors(20),
        surface_value = c("BSL", "IRSL"),
        normalise = TRUE,
        contour = TRUE,
        contour_nlevels = 20,
        plot = TRUE
      ), "RLum.Results")

    ## check list input
    expect_s4_class(
      suppressWarnings(analyse_portableOSL(list(merged), mode = "surface")),
      "RLum.Results")
    expect_warning(expect_null(analyse_portableOSL(list())),
                   "Nothing was merged as the object list was found to be empty")

    ## check additional argument sample
    expect_s4_class(analyse_portableOSL(
      merged,
      signal_integral = 1:5,
      invert = FALSE,
      normalise = TRUE,
      ylim = c(1,2),
      zlim = list(BSL = c(0,1.1), IRSL = c(0,1)),
      plot = TRUE,
      sample = "test"
    ),
    "RLum.Results")

    ## integral_input
    res1 <- analyse_portableOSL(merged,
                                signal_integral = 1:5,
                                integral_input = "measurement",
                                invert = FALSE,
                                mode = "profile",
                                normalise = TRUE,
                                plot = FALSE)
    res2 <- analyse_portableOSL(merged,
                                signal_integral = 1:5,
                                integral_input = "channel",
                                invert = FALSE,
                                mode = "profile",
                                normalise = TRUE,
                                plot = FALSE)
    res1@info <- res2@info <- list() # remove $call
    res1@.uid <- res2@.uid <- NA_character_
    res1@data$args$integral_input <- res2@data$args$integral_input <- NULL
    expect_equal(res1, res2)

    ## more coverage
    expect_s4_class(analyse_portableOSL(
        merged,
        signal_integral = 1:5,
        mode = "surface",
        bg_img = as.raster(matrix(0:1, ncol = 4, nrow = 3))
    ), "RLum.Results")

    ## deprecated arguments
    expect_warning(analyse_portableOSL(merged,
                                       signal.integral = 1:5),
                   "was deprecated in v1.2.0, use 'signal_integral'")
    expect_warning(expect_error(analyse_portableOSL(merged,
                                                    signal.integral = 1:5,
                                                    integral_input = "measurement"),
                   "'integral_input' is not supported with old argument names"),
                   "was deprecated in v1.2.0, use 'signal_integral'")
})

test_that("input validation", {
    testthat::skip_on_cran()

    expect_error(analyse_portableOSL("error"),
                 "[analyse_portableOSL()] 'object' should be of class 'RLum.Analysis'",
                 fixed = TRUE)
    expect_error(analyse_portableOSL(set_RLum("RLum.Analysis")),
                 "'object' cannot be an empty RLum.Analysis")
    expect_error(analyse_portableOSL(merged, signal_integral = 1:5,
                                     mode = "error"),
                 "'mode' should be one of 'profile' or 'surface'")

    ## Only RLum.Data.Curves
    tmp <- merged
    tmp@records <- list(tmp@records, "error")
    expect_error(analyse_portableOSL(tmp),
                 regexp = "\\[analyse\\_portableOSL\\(\\)\\] The 'RLum.Analysis' object must contain only.+")

    ## Check originator
    tmp <- merged
    tmp@records[[1]]@originator <- "error"
    expect_error(analyse_portableOSL(tmp),
                 "At least one element of 'object' has an unsupported originator")

    ## Sequence pattern
    expect_error(analyse_portableOSL(merged[-7], signal_integral = 1:5),
                 "Sequence pattern not supported: see the manual for details")
    expect_error(analyse_portableOSL(merged[1:3], signal_integral = 1:3),
                 "Sequence pattern not supported: see the manual for details")
    expect_error(analyse_portableOSL(merged[c(1:5, rep(7, 5))],
                                     signal_integral = 1:5),
                 "Sequence pattern not supported: the number of OSL records")
    expect_error(analyse_portableOSL(merged[c(1:5, rep(6, 5))],
                                     signal_integral = 1:5),
                 "Sequence pattern not supported: expected 3 DARK_COUNT records")
    expect_error(analyse_portableOSL(merged[-c(7:11)], signal_integral = 1:5),
                 "'object' references 14 sample names, but 13 IRSL/OSL pairs")

    ## coordinates not list or matrix
    expect_error(analyse_portableOSL(merged, signal_integral = 1:5,
                                     coord = "error"),
      "'coord' should be of class 'matrix' or 'list'")

    ## coordinates are not of the correct size
    expect_error(analyse_portableOSL(merged, signal_integral = 1:5,
                                     coord = list(COORD_X = c(0, 0),
                                                  COORD_Y = c(1, 2))),
                 "The number of coordinates in 'coord' should match the number",
                 fixed = TRUE)
    expect_error(analyse_portableOSL(merged, signal_integral = 1:5,
                                     coord = as.list(1:14)),
                 "'coord' should specify two coordinates per sample")

    ## trigger message if sample names don't contain coordinates
    mod <- merged
    mod@records <- lapply(mod@records, function(x) {
      x@info$settings$Sample <- substr(x@info$settings$Sample, 1, 4)
      x
    })
    expect_message(
      analyse_portableOSL(
        mod,
        signal_integral = 1:5,
        invert = FALSE,
        normalise = TRUE,
        mode = "surface",
        surface_value = c("BSL"),
        plot = TRUE,
        sample = "test"),
      "Surface plot is not available when all x-coordinates are 0")

    expect_error(
      analyse_portableOSL(
        merged,
        signal_integral = 1:5,
        mode = "surface",
        surface_value = "error"),
      "Unknown value to plot, valid values are:")

    expect_warning(
      analyse_portableOSL(
        merged,
        signal_integral = 1:5,
        invert = FALSE,
        normalise = TRUE,
        mode = "profile",
        zlim = c(1,2),
        plot = TRUE,
        sample = "test"),
      regexp = "\\[analyse\\_portableOSL\\(\\)\\] In profile mode, zlim.+")

    expect_warning(analyse_portableOSL(merged[1:5],
                                       signal_integral = c(1, 102)),
                   "'signal_integral' out of bounds, reset to be between 1 and 1")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("profile",
                              analyse_portableOSL(merged, mode = "profile",
                                                  signal_integral = 1:5))
  vdiffr::expect_doppelganger("surface",
                              analyse_portableOSL(merged, mode = "surface",
                                                  signal_integral = 1:5,
                                                  nx = 40,
                                                  ny = 40))
  vdiffr::expect_doppelganger("surface contour cex",
                              analyse_portableOSL(merged, mode = "surface",
                                                  signal_integral = 1:5,
                                                  nx = 40,
                                                  ny = 40,
                                                  contour = TRUE,
                                                  contour_col = "gray50",
                                                  cex = 2))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  SW({
  ## issue 675
  expect_warning(analyse_portableOSL(ExampleData.portableOSL[[1]],
                                     coord = list(c(1, 1)),
                                     signal_integral = 1:5, mode = "surface"),
                 "Surface interpolation failed: this happens when all points")

  ## issue 680
  expect_warning(analyse_portableOSL(ExampleData.portableOSL[[1]],
                                     signal_integral = -3:200),
                 "'signal_integral' out of bounds, reset to be between 1 and 100")
  })
})

test_that("simple test", {
  testthat::skip_on_cran()
  local_edition(3)
  data("ExampleData.portableOSL", envir = environment())
  merged <- merge_RLum(ExampleData.portableOSL)
  bin <- PSL2Risoe.BINfileData(merged)

  ##checks
  expect_s4_class(bin, "Risoe.BINfileData")
  expect_equal(length(bin), 70)

})

## load data
data(ExampleData.BINfileData, envir = environment())
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
results <- analyse_SAR.CWOSL(
    object = object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    plot = FALSE,
    verbose = FALSE
)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_DRCSummary("test"),
               "'object' should be of class 'RLum.Results'")
  expect_error(plot_DRCSummary(set_RLum("RLum.Results")),
               "'object' has an unsupported originator")

  ## different fit
  err <- merge_RLum(list(results, results))
  err@data$data$Fit[2] <- "err"
  expect_error(plot_DRCSummary(err),
               "Dose-response curves fitted using different equations: 'EXP', 'err'")
})

test_that("Test plotting", {
  testthat::skip_on_cran()

  ## create OTOR DRC
  results_OTOR <- analyse_SAR.CWOSL(
    object = object,
    fit.method = "OTOR",
    signal_integral = 1:2,
    background_integral = 900:1000,
    n.MC = 2,
    plot = FALSE,
    verbose = FALSE
  )

  ## create OTOR DRC
  results_OTORX <- analyse_SAR.CWOSL(
    object = object,
    fit.method = "OTORX",
    dose.points.test = 10,
    signal_integral = 1:2,
    background_integral = 900:1000,
    n.MC = 2,
    plot = FALSE,
    verbose = FALSE
  )

  ##simple
  expect_silent(plot_DRCSummary(results))

  ##simple with OTOR
  expect_silent(plot_DRCSummary(results_OTOR))

  ##simple with OTORX
  expect_silent(plot_DRCSummary(results_OTORX))

  ## list
  expect_silent(plot_DRCSummary(list(results, results_OTOR),
                                main = "Title"))
  expect_silent(plot_DRCSummary(list(results, results_OTOR),
                                source_dose_rate = 1))
  l <- expect_silent(plot_DRCSummary(list()))
  expect_length(l, 0)

  ##expect warning
  expect_warning(plot_DRCSummary(results, show_dose_points = TRUE,
                                 show_natural = TRUE, sel_curves = 1000),
                 "'sel_curves' out of bounds, reset to full dataset")
  expect_warning(plot_DRCSummary(results_OTOR, xlim = c(-1e12, 1e12)),
                 "Dose response curve 1 contains NA/NaN values, curve removed")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("dose points natural",
                              plot_DRCSummary(results,
                                              show_dose_points = TRUE,
                                              show_natural = TRUE))
  vdiffr::expect_doppelganger("dose_rate",
                              plot_DRCSummary(results,
                                              col.lty = "red",
                                              source_dose_rate = 0.5))
  })
})

## load data
bin.v8 <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
risoe <- read_BIN2R(bin.v8, verbose = FALSE)
SW({
analysis <- merge_RLum(Risoe.BINfileData2RLum.Analysis(risoe))
})
curve <- analysis@records[[1]]

test_that("input validation", {
  testthat::skip_on_cran()

  ## add_metadata
  expect_error(add_metadata(risoe, list()) <- 1,
               "'info_element' should be of class 'character'")
  expect_error(add_metadata(curve, list()) <- 1,
               "'info_element' should be of class 'character'")
  expect_error(add_metadata(risoe, c("VAL1", "VAL2")) <- 1,
               "'info_element' should be of class 'character' and have length 1")
  expect_error(add_metadata(curve, c("VAL1", "VAL2")) <- 1,
               "'info_element' should be of class 'character' and have length 1")
  expect_error(add_metadata(risoe, "POSITION") <- 1,
               "'info_element' already present, to modify it you should use")
  expect_error(add_metadata(curve, "POSITION") <- 1,
               "'info_element' already present, to modify it you should use")
  expect_error(add_metadata(risoe, "NEW") <- NULL,
               "Cannot store a metadata entry with NULL value")
  expect_error(add_metadata(curve, "NEW") <- NULL,
               "Cannot store a metadata entry with NULL value")

  ## rename_metadata
  expect_error(rename_metadata(risoe, list()) <- 1,
               "'info_element' should be of class 'character'")
  expect_error(rename_metadata(curve, list()) <- 1,
               "'info_element' should be of class 'character'")
  expect_error(rename_metadata(risoe, c("VAL1", "VAL2")) <- 1,
               "'info_element' should be of class 'character' and have length 1")
  expect_error(rename_metadata(curve, c("VAL1", "VAL2")) <- 1,
               "'info_element' should be of class 'character' and have length 1")
  expect_error(rename_metadata(risoe, "error") <- 1,
               "'info_element' not recognised ('error'), valid terms are",
               fixed = TRUE)
  expect_error(rename_metadata(curve, "error") <- 1,
               "'info_element' not recognised ('error'), valid terms are",
               fixed = TRUE)
  expect_error(rename_metadata(risoe, "POSITION") <- NULL,
               "Cannot rename a metadata entry to NULL, to remove it you should")
  expect_error(rename_metadata(curve, "POSITION") <- NULL,
               "Cannot rename a metadata entry to NULL, to remove it you should")

  ## replace_metadata
  expect_error(replace_metadata(risoe, list()) <- 1,
               "'info_element' should be of class 'character'")
  expect_error(replace_metadata(curve, list()) <- 1,
               "'info_element' should be of class 'character'")
  expect_error(replace_metadata(risoe, "error") <- 1,
               "'info_element' not recognised ('error'), valid terms are",
               fixed = TRUE)
  expect_error(replace_metadata(curve, "error") <- 1,
               "'info_element' not recognised ('error'), valid terms are",
               fixed = TRUE)
  expect_error(replace_metadata(risoe, "SEL", subset = POSITION == 2) <- NULL,
               "'subset' is incompatible with assigning NULL")
  expect_error(replace_metadata(curve, "SEL", subset = POSITION == 2) <- NULL,
               "'subset' is incompatible with assigning NULL")
  expect_error(replace_metadata(risoe, "SEL", subset = error == 99) <- 0,
               "Invalid 'subset' expression, valid terms are")
  expect_error(replace_metadata(curve, "SEL", subset = error == 99) <- 0,
               "Invalid 'subset' expression, valid terms are")
  expect_error(replace_metadata(risoe, "SEL", subset = ID + 99) <- 0,
               "'subset' should contain a logical expression")
  expect_error(replace_metadata(curve, "SEL", subset = ID + 99) <- 0,
               "'subset' should contain a logical expression")
  expect_message(replace_metadata(risoe, "SEL", subset = ID == 99) <- 0,
                 "'subset' expression produced an empty selection, nothing done")
  expect_message(replace_metadata(risoe, "SEL", subset = NA) <- 0,
                 "'subset' expression produced an empty selection, nothing done")
  expect_message(replace_metadata(curve, "SEL", subset = SET == 99) <- 0,
                 "'subset' expression produced an empty selection, nothing done")
  expect_message(replace_metadata(curve, "SEL", subset = NA) <- 0,
                 "'subset' expression produced an empty selection, nothing done")
})

test_that("check functionality for Risoe.BINfileData", {
  testthat::skip_on_cran()

  res <- risoe

  ## add_metadata
  add_metadata(res, "NEW") <- 123
  expect_equal(res@METADATA$NEW,
               rep(123, nrow(res@METADATA)))

  ## rename_metadata
  rename_metadata(res, "NEW") <- "NEWER"
  expect_equal(res@METADATA$NEWER,
               rep(123, nrow(res@METADATA)))

  ## replace_metadata
  replace_metadata(res, "SEL") <- FALSE
  expect_equal(res@METADATA$SEL,
               rep(FALSE, nrow(res@METADATA)))
  replace_metadata(res, "LTYPE", subset = SET == 2 & POSITION == 1) <- "OSL"
  expect_equal(res@METADATA$LTYPE,
               c("OSL", "TL"))
  replace_metadata(res, c("PTENABLED", "DTENABLED")) <- NULL
  expect_null(res@METADATA$PTENABLED)
  expect_null(res@METADATA$DTENABLED)

  ## the original object is unchanged
  expect_null(risoe@METADATA$NEW)
  expect_null(risoe@METADATA$NEWER)
  expect_equal(risoe@METADATA$SEL,
               rep(TRUE, nrow(res@METADATA)))
  expect_equal(risoe@METADATA$LTYPE,
               rep("TL", nrow(res@METADATA)))
  expect_equal(risoe@METADATA$PTENABLED,
               c(0, 0))
  expect_equal(risoe@METADATA$DTENABLED,
               c(0, 0))
})

test_that("check functionality for RLum.Analysis", {
  testthat::skip_on_cran()

  res <- analysis
  num.records <- length(analysis@records)

  ## add_metadata
  add_metadata(res, "NEW") <- 123
  expect_equal(sapply(res@records, function(x) x@info[["NEW"]]),
               c(123, 123))

  ## rename_metadata
  rename_metadata(res, "NEW") <- "NEWER"
  expect_equal(sapply(res@records, function(x) x@info[["NEWER"]]),
               c(123, 123))

  ## replace_metadata
  replace_metadata(res, "SEL") <- FALSE
  expect_equal(sapply(res@records, function(x) x@info[["SEL"]]),
               c(FALSE, FALSE))
  replace_metadata(res, "LTYPE", subset = SET == 2 & POSITION == 1) <- "OSL"
  expect_equal(sapply(res@records, function(x) x@info[["LTYPE"]]),
               c("OSL", "TL"))
  replace_metadata(res, "SEQUENCE") <- NULL
  expect_null(unlist(sapply(res@records, function(x) x@info[["SEQUENCE"]])))

  ## the original object is unchanged
  expect_null(unlist(sapply(analysis@records, function(x) x@info[["NEW"]])))
  expect_null(unlist(sapply(analysis@records, function(x) x@info[["NEWER"]])))
  expect_equal(sapply(analysis@records, function(x) x@info[["SEL"]]),
               rep(TRUE, num.records))
  expect_equal(sapply(analysis@records, function(x) x@info[["LTYPE"]]),
               rep("TL", num.records))
  expect_equal(sapply(analysis@records, function(x) x@info[["SEQUENCE"]]),
               rep("", num.records))
})

test_that("check functionality for RLum.Data", {
  testthat::skip_on_cran()

  res <- curve

  ## add_metadata
  add_metadata(res, "NEW") <- 123
  expect_equal(res@info$NEW, 123)

  ## rename_metadata
  rename_metadata(res, "NEW") <- "NEWER"
  expect_equal(res@info$NEWER, 123)

  ## replace_metadata
  replace_metadata(res, "SEL") <- FALSE
  expect_equal(res@info$SEL, FALSE)
  replace_metadata(res, "LTYPE", subset = SET == 2) <- "OSL"
  expect_equal(res@info$LTYPE, "OSL")
  replace_metadata(res, c("AN_TEMP", "AN_TIME")) <- NULL
  expect_null(res@info$AN_TEMP)
  expect_null(res@info$AN_TIME)

  ## the original object is unchanged
  expect_null(curve@info$NEW)
  expect_null(curve@info$NEWER)
  expect_equal(curve@info$SEL, TRUE)
  expect_equal(curve@info$LTYPE, "TL")
  expect_equal(curve@info$AN_TEMP, 220)
  expect_equal(curve@info$AN_TIME, 10)
})

## load data
data(ExampleData.RLum.Analysis, envir = environment())
data(ExampleData.RLum.Data.Image, envir = environment())
o1 <- IRSAR.RF.Data
o2 <- IRSAR.RF.Data
c1 <- as(object = list(1:10), Class = "RLum.Data.Curve")
s1 <- as(object = matrix(1:20, 4, 5, dimnames = list(1:4, 1:5)),
         Class = "RLum.Data.Spectrum")
r1 <- as(object = list(1:10), Class = "RLum.Results")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(merge_RLum("error"),
               "[merge_RLum()] 'objects' should be of class 'list'",
               fixed = TRUE)
  expect_error(merge_RLum(list(o1, o2, "test")),
               "All elements of 'objects' should be of class 'RLum'")
  expect_error(merge_RLum(list(r1, c1)),
               "Objects cannot be merged, different classes found: 'RLum.Results'")
  expect_error(merge_RLum(list(ExampleData.RLum.Data.Image)),
               "Merging of 'RLum.Data.Image' objects is currently not supported")

  expect_warning(merge_RLum(list(NULL)),
                 "Nothing was merged as the object list was found to be empty")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## simple test
  expect_s4_class(merge_RLum(list(o1,o2)), "RLum.Analysis")
  expect_s4_class(merge_RLum(list(c1,c1)), "RLum.Data.Curve")
  expect_s4_class(merge_RLum(list(s1,s1)), "RLum.Data.Spectrum")
  expect_s4_class(merge_RLum(list(r1,r1)), "RLum.Results")

  ## with null objects
  expect_s4_class(merge_RLum(list(o1,o2, NULL)), "RLum.Analysis")

  ## single object
  expect_s4_class(
    merge_RLum(list(o1)), "RLum.Analysis")
})

## load data
data("ExampleData.portableOSL", envir = environment())
merged <- merge_RLum(ExampleData.portableOSL)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_PSL2Risoe.BINfileData("error"),
               "'object' should be of class 'RLum.Analysis'")
  expect_error(convert_PSL2Risoe.BINfileData(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  bin <- convert_PSL2Risoe.BINfileData(merged)
  expect_s4_class(bin, "Risoe.BINfileData")
  expect_equal(length(bin), 70)

  ## manipulate the object to trigger other errors
  fake <- merged
  fake@records[20] <- "unexpected-subclass"
  expect_error(convert_PSL2Risoe.BINfileData(fake),
               "All elements of 'object' should be of class 'RLum.Data.Curve'")

  fake <- merged
  fake@records[10][[1]]@originator <- "unexpected-originator"
  expect_error(convert_PSL2Risoe.BINfileData(fake),
               "At least one element of 'object' has an unsupported originator")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  SW({
  expect_snapshot_Risoe(convert_PSL2Risoe.BINfileData(merged))
  })
})

#' @title Al2O3:C Passive Dosimeter Measurement Analysis
#'
#' @description The function provides the analysis routines for measurements on a
#' FI lexsyg SMART reader using Al2O3:C chips according to Kreutzer et al., 2018
#'
#' @details
#'
#' **Working with a travel dosimeter**
#'
#' The function allows to define particular position numbers as travel
#' dosimeters. For example: `travel_dosimeter = c(1,3,5)` sets positions 1,
#' 3 and 5 as travel dosimeters. These dose values #' of this dosimeters
#' are combined and automatically subtracted from the obtained dose values
#' of the other dosimeters.
#'
#' **Calculate TL dose**
#'
#' The argument `calculate_TL_dose` provides the possibility to experimentally calculate a TL-dose,
#' i.e. an apparent dose value derived from the TL curve ratio. However, it should be noted that
#' this value is only a fall back in case something went wrong during the measurement of the optical
#' stimulation. The TL derived dose value is corrected for cross-talk and for the irradiation time,
#' but not considered if a travel dosimeter is defined.
#'
#' Calculating the palaeodose is possible without **any TL** curve in the sequence!
#'
#' **Test parameters**
#'
#' `TL_peak_shift` [numeric] (default: `15`):
#'
#' Checks whether the TL peak shift is bigger > 15 K, indicating a problem with the
#' thermal contact of the chip.
#'
#' `stimulation_power` [numeric] (default: `0.05`):
#'
#' So far available, information on the delivered optical stimulation are compared. Compared are
#' the information from the first curves with all others. If the ratio differs more from
#' unity than the defined by the threshold, a warning is returned.
#'
#' @param object [Luminescence::RLum.Analysis-class] (**required**):  measurement input
#'
#' @param signal_integral [numeric] (*optional*): signal integral, used for the signal
#' and the background. Example: `c(1:10)` for the first 10 channels.
#' If nothing is provided the full range is used
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param dose_points [numeric] (*with default*):
#' vector with dose points, if dose points are repeated, only the general
#' pattern needs to be provided. Default values follow the suggestions
#' made by Kreutzer et al., 2018
#'
#' @param recordType [character] (*with default*): input curve selection, which is passed to
#' function [Luminescence::get_RLum]. To deactivate the automatic selection set the argument to `NULL`
#'
#' @param irradiation_time_correction [numeric] or [Luminescence::RLum.Results-class] (*optional*):
#' information on the used irradiation time correction obtained by another experiments.
#' If a `numeric` is provided it has to be of length two: mean, standard error
#'
#' @param calculate_TL_dose [logical] (*with default*):
#' enables/disable experimental dose estimation based on the TL curves.
#' It is computed as the ratio of the peak sums of each curves +/- 5 channels.
#'
#' @param cross_talk_correction [numeric] or [Luminescence::RLum.Results-class] (*optional*):
#' information on the used irradiation time correction obtained by another experiments.
#' If a `numeric` vector is provided it has to be of length three:
#' mean, 2.5 % quantile, 97.5 % quantile.
#'
#' @param travel_dosimeter [numeric] (*optional*): specify the position of
#' the travel dosimeter (so far measured at the same time). The dose of
#' travel dosimeter will be subtracted from all other values.
#'
#' @param test_parameters [list] (*with default*):
#' set test parameters. Supported parameters are: `TL_peak_shift` All input: [numeric]
#' values, `NA` and `NULL` (see details).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#' If `object` is of type [list], a numeric vector can be provided to limit
#' the plot output to certain aliquots.
#'
#' @param ... further arguments that can be passed to the plot output, supported are `norm`, `main`, `mtext`,
#' `title` (for self-call mode to specify, e.g., sample names)
#'
#' @return Function returns results numerically and graphically:
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$data` \tab `data.frame` \tab the estimated equivalent dose \cr
#'  `$data_table` \tab `data.frame` \tab full dose and signal table \cr
#'  `test_parameters` \tab `data.frame` \tab results with test parameters \cr
#'  `data_TDcorrected` \tab `data.frame` \tab travel dosimeter corrected results (only if TD was provided)\cr
#' }
#'
#' **Note:** If correction the irradiation time and the cross-talk correction
#' method is used, the De values in the table `data` table are already
#' corrected, i.e. if you want to get an uncorrected value, you can use the
#' column `CT_CORRECTION` remove the correction.
#'
#'**slot:** **`@info`**
#'
#' The original function call
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' - OSL and TL curves, combined on two plots.
#'
#'
#' @section Function version: 0.2.7
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::analyse_Al2O3C_ITC]
#'
#' @references
#'
#' Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018. Environmental Dose Rate
#' Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
#' Geochronometria 45, 56-67.
#'
#' @keywords datagen
#'
#' @examples
#' ##load data
#' data(ExampleData.Al2O3C, envir = environment())
#'
#' ##run analysis
#' analyse_Al2O3C_Measurement(data_CrossTalk)
#'
#' @export
analyse_Al2O3C_Measurement <- function(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  dose_points = c(0,4),
  recordType = c("OSL (UVVIS)", "TL (UVVIS)"),
  calculate_TL_dose = FALSE,
  irradiation_time_correction = NULL,
  cross_talk_correction = NULL,
  travel_dosimeter = NULL,
  test_parameters = NULL,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("analyse_Al2O3C_Measurement")
  on.exit(.unset_function_name(), add = TRUE)

  ## Self call --------------------------------------------------------------
  if (inherits(object, "list")) {
    .validate_not_empty(object)
    lapply(object, .validate_class, "RLum.Analysis",
           name = "All elements of 'object'")

    ## expand input arguments
    rep.length <- length(object)

    if(!is.null(signal_integral)){
      signal_integral <- .listify(list(signal_integral), rep.length)
    }

    dose_points <- .listify(dose_points, rep.length)
    irradiation_time_correction <- .listify(irradiation_time_correction,
                                            rep.length)
    cross_talk_correction <- .listify(cross_talk_correction, rep.length)

    ## since `test_parameters` is always expected to be a list, we can't
    ## use .listify() as it will not generate sublists
    test_parameters <- rep(list(test_parameters), rep.length)

    ##plot
    if (is.logical(plot)) {
      .validate_logical_scalar(plot)
      plot <- rep(x = plot, length(object))
    }else{
      plot <- 1:length(object)%in%plot
    }

    ## run analysis
    results <- lapply(seq_along(object), function(x) {
      temp <- analyse_Al2O3C_Measurement(
        object = object[[x]],
        signal_integral = signal_integral[[x]],
        integral_input = integral_input,
        dose_points = dose_points[[x]],
        irradiation_time_correction = irradiation_time_correction[[x]],
        cross_talk_correction = cross_talk_correction[[x]],
        test_parameters = test_parameters[[x]],
        calculate_TL_dose = calculate_TL_dose,
        verbose = verbose,
        plot = plot[x],
        ...
      )

     ##adjusting the terminal output, to avoid confusions
     if(verbose)
       cat(" ... (#",x, " | ALQ POS: ", temp$data$POSITION,")\n", sep = "")

     ##add running number to the plot, but only of we had a plot here...
     if(plot[x]){
       title(main = paste0(list(...)$title[x], " ","#", x), adj = 1, line = 3)
     }

     return(temp)
    })

    ##merge results
    results <- merge_RLum(results)

    ##correct sys.call, otherwise it gets a little bit strange
    ##why this is not implemented in the merge_RLum() method ... because here it would be wrong!
    results@info[names(results@info) == "call"] <- NULL
    results@info$call <- sys.call()

    ##travel dosimeter
    ##check for travel dosimeter and subtract the values so far this is meaningful at all
    .validate_class(travel_dosimeter, c("numeric", "integer"), null.ok = TRUE)
    if(!is.null(travel_dosimeter)){
      ##check whether everything is subtracted from everything ... you never know, users do weird stuff
      if(length(travel_dosimeter) == nrow(results$data))
        .throw_message("'travel_dosimeter' specifies every position, nothing corrected")

      ##check if the position is valid
      if (!all(travel_dosimeter %in% results$data$POSITION))
        .throw_message("Invalid position in 'travel_dosimeter', nothing corrected")

      ##correct for the travel dosimeter calculating the weighted mean and the sd (as new error)
      ##if only one value is given just take it
      if(length(travel_dosimeter) == 1 && nrow(results$data[travel_dosimeter==results$data$POSITION,c(1,2)]) == 1){
        correction <- as.numeric(results$data[travel_dosimeter==results$data$POSITION,c(1,2)])

      }else{
        temp.correction <- results$data[results$data$POSITION%in%travel_dosimeter,c(1,2)]
        correction <- c(
          stats::weighted.mean(
            x = temp.correction[[1]],
            w = if(all(temp.correction[[2]]==0)){rep(1, length(temp.correction[[2]]))} else {temp.correction[[2]]}),
          sd(temp.correction[,1]))
        rm(temp.correction)
      }

      ##subtract all the values, in a new data frame, we do not touch the original data
      data_TDcorrected <- data.frame(
        DE = results@data$data[!results$data$POSITION%in%travel_dosimeter,1] - correction[1],
        DE_ERROR = sqrt(results@data$data[!results$data$POSITION%in%travel_dosimeter,2]^2 + correction[2]^2),
        POSITION = results@data$data[!results$data$POSITION%in%travel_dosimeter, "POSITION"]
      )

      ##however, we set information on the travel dosimeter in the corresponding column
      results@data$data$TRAVEL_DOSIMETER <- results$data$POSITION%in%travel_dosimeter

      ##attach the new element to the results output
      results@data <- c(results@data,  list(data_TDcorrected = data_TDcorrected))

      ##return message
      if(verbose)
        cat("\n ...+ travel dosimeter correction applied.\n ...+ results stored in object $data_TDcorrected.\n\n")
    } ##end travel dosimeter

    ##return results
    return(results)
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis",
                  extra = "a 'list' of such objects")
  .validate_not_empty(object)
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  .validate_class(irradiation_time_correction, c("RLum.Results", "numeric"),
                  null.ok = TRUE)
  .validate_class(cross_talk_correction, c("numeric", "RLum.Results"),
                  null.ok = TRUE)
  .validate_class(travel_dosimeter, c("numeric", "integer"), null.ok = TRUE)
  .validate_class(test_parameters, "list", null.ok = TRUE)

  ## Preparation ------------------------------------------------------------

  ##select curves based on the recordType selection; if not NULL
  if(!is.null(recordType)){
    object_raw <- object
    object <- get_RLum(object, recordType = recordType, drop = FALSE)
    if (length(object) == 0) {
      .throw_error("No records with recordType ", .collapse(recordType), " found")
    }
  }

  ## signal integral
  x.range <- object[[1]][, 1]
  if (integral_input == "measurement") {
    signal_integral <- .convert_to_channels(x.range, signal_integral,
                                            "time", null.ok = TRUE)
  }
  signal_integral <- .validate_integral(signal_integral,
                                        max = length(x.range), null.ok = TRUE)
  if (is.null(signal_integral)) {
    signal_integral <- seq_along(x.range)
  }

  ## Set Irradiation Time Correction ---------------
  if (!is.null(irradiation_time_correction)) {
    if (is.numeric(irradiation_time_correction)) {
      .validate_length(irradiation_time_correction, 2)
    } else {
      ## RLum.Results case
      .validate_originator(irradiation_time_correction, "analyse_Al2O3C_ITC")
      irradiation_time_correction <- get_RLum(irradiation_time_correction)

      ## consider the case for more than one observation ...
      irradiation_time_correction <-
        c(mean(irradiation_time_correction[[1]]),
          if (nrow(irradiation_time_correction) > 1)
            sd(irradiation_time_correction[[1]])
          else
            irradiation_time_correction[[2]])
    }
  }

  ## Set Cross-Talk Correction ----------------------------------------------
  ## check whether information on the position was stored in the input object
  POSITION <- get_RLum(object = object[[1]], info.object = "position")
  if (is.null(POSITION)) {
    .throw_message("Aliquot position not found, no cross-talk correction applied")
    cross_talk_correction <- NULL
    POSITION <- NA
  }

  if(is.null(cross_talk_correction)){
    cross_talk_correction <- c(0,0,0)
  } else if (is.numeric(cross_talk_correction)) {
    .validate_length(cross_talk_correction, 3)
  } else {
    ## RLum.Results case
    .validate_originator(cross_talk_correction, "analyse_Al2O3C_CrossTalk")

      ## calculate the cross-talk correction values for this particular
      ## carousel position
      interval <- ifelse(length(POSITION) > 1, "confidence", "none")
      cross_talk_correction <-
        as.numeric(predict(cross_talk_correction$fit,
                           newdata = data.frame(x = POSITION),
                           interval = interval))

      ## we could not compute a confidence interval, so we create it manually
      if (length(POSITION) == 1) {
        cross_talk_correction <- matrix(cross_talk_correction, 1, 3,
                                        dimnames = list("position",
                                                        c("fit", "lwr", "upr")))
      }
  }

  # Calculation -------------------------------------------------------------
  ##we have two dose points, and one background curve, we do know only the 2nd dose

  ##set test parameters
  test_parameters.default <- list(
    TL_peak_shift = 15,
    stimulation_power = 0.05
  )

  ##modify default values by given input
  if(!is.null(test_parameters)){
    test_parameters <- modifyList(test_parameters.default, test_parameters)

    ##remove NULL elements from list
    test_parameters <- .rm_NULL_elements(test_parameters)
  }else{
    test_parameters <- test_parameters.default
  }

  ##calculate integrated light values
  NATURAL <- sum(get_RLum(object, recordType = "OSL")[[1]]@data[signal_integral, 2])
  REGENERATED <- sum(get_RLum(object, recordType = "OSL")[[2]]@data[signal_integral, 2])
  BACKGROUND <- sum(get_RLum(object, recordType = "OSL")[[3]]@data[signal_integral, 2])

  ##do the same for the TL
  if (calculate_TL_dose[1] && any(grepl("TL", names(object)))){
    nat_integr.idx <- which.max(object@records[[2]]@data[, 2]) + c(-5, 5)
    NATURAL_TL <- tryCatch(
        sum(object@records[[2]]@data[nat_integr.idx, 2]),
        error = function(e) {
          .throw_warning("Natural TL signal out of bounds, NA returned")
          NA
        })
    reg_integr.idx <- which.max(object@records[[4]]@data[, 2]) + c(-5, 5)
    REGENERATED_TL <- tryCatch(
        sum(object@records[[4]]@data[reg_integr.idx, 2]),
        error = function(e) {
          .throw_warning("Regenerated TL signal out of bounds, NA returned")
          NA
        })
  }else{
    NATURAL_TL <- NA
    REGENERATED_TL <- NA
  }

  ##combine into data.frame
  temp_df <- data.frame(
      POSITION = POSITION,
      DOSE = if(!is.null(irradiation_time_correction)){
        dose_points + irradiation_time_correction[1]
        }else{
         dose_points
        },
      DOSE_ERROR = if(!is.null(irradiation_time_correction)){
        dose_points * irradiation_time_correction[2]/irradiation_time_correction[1]
        }else{
        0
        },
      STEP = c("NATURAL", "REGENERATED"),
      INTEGRAL = c(NATURAL, REGENERATED),
      BACKGROUND = c(BACKGROUND, BACKGROUND),
      NET_INTEGRAL = c(NATURAL - BACKGROUND, REGENERATED - BACKGROUND),
      NATURAL_TL = NATURAL_TL,
      REGENERATED_TL = REGENERATED_TL,
      row.names = NULL
    )

     ##0 dose points should not be biased by the correction ..
     ##Note: it does not mean that 0 s beneath the source has a dose of 0, however, in the certain
     ##case aliquot was never moved under the source
     id_zero <- which(dose_points == 0)
     temp_df$DOSE[id_zero] <- 0
     temp_df$DOSE_ERROR[id_zero] <- 0


   ##calculate DE by using the irradiation time correction AND the cross talk correction

   ##(1) sample dose point values with irradiation time corrections (random)
   if(!is.null(irradiation_time_correction)){
    DOSE_MC <- rnorm(1000, mean = temp_df$DOSE[2], sd = temp_df$DOSE_ERROR[2])

   }else{
    DOSE_MC <- temp_df$DOSE[2]
  }

   ##(2) random sampling from cross-irradiation
   CT <- runif(1000, min = cross_talk_correction[2], max = cross_talk_correction[3])

   ##(3) signal ratio
   INTEGRAL_RATIO <- temp_df$NET_INTEGRAL[1]/temp_df$NET_INTEGRAL[2]

   ##(4) calculate DE
   temp_DE <- (DOSE_MC * INTEGRAL_RATIO)

   ##(5) substract cross-talk value from DE
   temp_DE  <- temp_DE  - CT

     ##(5.1) calculate TL based DE
     ##calculate a dose based on TL
     ##Note: we use irradiation time correction and CT correction based on GSL measurements
     if(calculate_TL_dose){
       TL_Ratio <- NATURAL_TL/REGENERATED_TL
       temp_TL_DE <- (DOSE_MC * TL_Ratio) - CT
       TL_DE <- mean(temp_TL_DE)
       TL_DE.ERROR <- sd(temp_TL_DE)

     }else{
       TL_DE <- NA
       TL_DE.ERROR <- NA
     }

   ##(6) create final data.frame
   data <- data.frame(
     DE = mean(temp_DE),
     DE_ERROR = sd(temp_DE),
     POSITION,
     INTEGRAL_RATIO,
     TRAVEL_DOSIMETER = NA,
     CT_CORRECTION = cross_talk_correction[1],
     CT_CORRECTION_Q2.5 = cross_talk_correction[2],
     CT_CORRECTION_Q97.5 = cross_talk_correction[3],
     TL_DE = TL_DE,
     TL_DE.ERROR = TL_DE.ERROR,
     row.names = NULL
   )

  ##calculate test parameters
  ##TL_peak_shift
  ##check TL peak positions, if it differers more than the threshold, return a message
  ##can be done better, but should be enough here.
  if (any("TL_peak_shift"%in%names(test_parameters)) && any(grepl("TL", names(object)))){
    ##calculate value
    TP_TL_peak_shift.value <- abs((object[[2]][which.max(object[[2]][,2]),1] -
             object[[4]][which.max(object[[4]][,2]),1]))


    ##compare
    TP_TL_peak_shift.status <-  TP_TL_peak_shift.value > test_parameters$TL_peak_shift

    ##return warning
    if(TP_TL_peak_shift.status)
      .throw_warning("TL peak shift detected for aliquot position ", POSITION,
                     ", check the curves")

    ##set data.frame
    TP_TL_peak_shift <- data.frame(
      CRITERIA = "TL_peak_shift",
      THRESHOLD = test_parameters$TL_peak_shift,
      VALUE = TP_TL_peak_shift.value,
      STATUS = TP_TL_peak_shift.status,
      stringsAsFactors = FALSE)

  }else{
    TP_TL_peak_shift <- data.frame(stringsAsFactors = FALSE)
  }

  ##stimulation_power
  TP_stimulation_power <- data.frame(stringsAsFactors = FALSE)
  if(any("stimulation_power"%in%names(test_parameters))){
     ##get curves ids holding the information on the stimulation power
     temp_curves_OSL <- get_RLum(object_raw, recordType = "OSL", curveType = "measured")
     temp_curves_OSL <- lapply(temp_curves_OSL, function(o){
       if ("stimulator" %in% names(o@info) &&
           grepl(o@info$stimulator, pattern = "LED", fixed = TRUE)) {
           return(o)
       }
      return(NULL)
     })

     ##remove NULL
     temp_curves_OSL <- .rm_NULL_elements(temp_curves_OSL)

     ##check whether something is left
     if(length(temp_curves_OSL) < 2){
       TP_stimulation_power.value <- NA
       TP_stimulation_power.status <- FALSE

     }else{
       ##calculate sum of the power
       TP_stimulation_power.value <-  vapply(temp_curves_OSL, function(x){
         sum(x@data[,2])

       }, numeric(1))

       ##estimate a theoretical value based on the first value ... it does not
       ##matter which value is correct or not
       TP_stimulation_power.value <- abs(1 -
         sum(TP_stimulation_power.value)/(TP_stimulation_power.value[1] * length(TP_stimulation_power.value)))

       TP_stimulation_power.status <- TP_stimulation_power.value > test_parameters$stimulation_power

       if(TP_stimulation_power.status)
         .throw_warning("Stimulation power was not stable for ALQ ",
                        POSITION, ", results are likely to be wrong")
     }

     ##remove object
     rm(temp_curves_OSL)

     ##set data.frame
     TP_stimulation_power <- data.frame(
       CRITERIA = "stimulation_power",
       THRESHOLD = test_parameters$stimulation_power,
       VALUE = TP_stimulation_power.value,
       STATUS = TP_stimulation_power.status,
       stringsAsFactors = FALSE)
  }

   ##compile all test parameter df
   df_test_parameters <- rbind(
     TP_TL_peak_shift,
     TP_stimulation_power)


  # Terminal output -----------------------------------------------------------------------------
  if(verbose){
    cat(" [analyse_Al2O3_Measurement()] #",POSITION, " ", "DE: ",
               round(data$DE, 2), " \u00B1 ", round(data$DE_ERROR,2), "\n", sep = "")
  }

  # Plotting ------------------------------------------------------------------------------------
    ##enable or disable plot ... we cannot put the condition higher, because we here
    ##calculate something we are going to need later
    if (plot) {
      par.default <- .par_defaults()
      on.exit(par(par.default), add = TRUE)

      ##settings
      plot_settings <- list(
        main = c(paste("ALQ POS:", POSITION, "| OSL"), paste("ALQ POS:", POSITION, "| TL")),
        norm = TRUE,
        mtext = ""
      )

     ##modify on request
     plot_settings <- modifyList(plot_settings, val = list(...))

     ##plot curves
     if(any(grepl("TL", names(object))))
       par(mfrow = c(1,2))

     plot_RLum(
       object,
       plot_singlePanels = TRUE,
       combine = TRUE,
       mtext = list(paste0("DE: ", round(data$DE,2), " \u00b1 ", round(data$DE_ERROR,2)), ""),
       xlab = list("Stimulation [s]", "Temperature [\u00B0C]"),
       legend.text = list(list("#1 NAT", "#3 REG", "#5 BG"), list("#2 NAT", "#4 REG")),
       legend.pos = list("topright", "topleft"),
       main = as.list(plot_settings$main),
       norm = plot_settings$norm
     )
    }

  # Output --------------------------------------------------------------------------------------
  UID <- create_UID()

  set_RLum(
    class = "RLum.Results",
    data = list(
      data = cbind(data, UID),
      data_table = cbind(temp_df, UID),
      test_parameters = cbind(df_test_parameters, UID)
      ),
    info = list(
      call = sys.call()
    )
  )
}

## load data
data(ExampleData.Al2O3C, envir = environment())
osl <- get_RLum(data_CrossTalk, recordType = "OSL", drop = FALSE)

test_that("input validation", {
  skip_on_cran()

  SW({
  expect_error(analyse_Al2O3C_Measurement(),
               "is missing, with no default")
  expect_error(analyse_Al2O3C_Measurement("error"),
               "'object' should be of class 'RLum.Analysis' or a 'list' of such objects")
  expect_error(analyse_Al2O3C_Measurement(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_error(analyse_Al2O3C_Measurement(list()),
               "'object' cannot be an empty list")
  expect_error(analyse_Al2O3C_Measurement(list(data_CrossTalk, "error")),
               "All elements of 'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk, plot = c(TRUE, FALSE)),
               "'plot' should be a single logical value")
  suppressWarnings(
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          travel_dosimeter = "error"),
               "'travel_dosimeter' should be of class 'numeric', 'integer' or")
  )
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          irradiation_time_correction = 7),
               "'irradiation_time_correction' should have length 2")
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          irradiation = set_RLum("RLum.Results")),
               "'irradiation_time_correction' has an unsupported originator")
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          irradiation_time_correction = "a"),
               "should be of class 'RLum.Results', 'numeric' or NULL")
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          cross_talk_correction = "a"),
               "'cross_talk_correction' should be of class 'numeric', 'RLum.Results'")
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          cross_talk_correction = c(1, 0.4)),
               "'cross_talk_correction' should have length 3")
  expect_error(analyse_Al2O3C_Measurement(data_CrossTalk,
                                          cross_talk_correction = set_RLum("RLum.Results")),
               "'cross_talk_correction' has an unsupported originator")

  expect_error(analyse_Al2O3C_Measurement(osl, signal_integral = 1000),
               "'signal_integral' is of length 0 after removing values smaller")
  })

  data(ExampleData.RF70Curves, envir = environment())
  expect_warning(expect_error(analyse_Al2O3C_Measurement(RF70Curves),
                              "No records with recordType 'OSL (UVVIS)'", fixed = TRUE),
                 "This request produced an empty list of records")
})

test_that("check functionality", {
  skip_on_cran()

  ## run analysis
  SW({ # warning: TL peak shift detected for aliquot position 1
  expect_s4_class(analyse_Al2O3C_Measurement(data_CrossTalk),
                  "RLum.Results")
  expect_s4_class(analyse_Al2O3C_Measurement(data_CrossTalk,
                                             calculate_TL_dose = TRUE),
                  "RLum.Results")
  expect_output(analyse_Al2O3C_Measurement(data_CrossTalk,
                        test_parameter = list(stimulation_power = 0.03)))
  expect_output(analyse_Al2O3C_Measurement(data_CrossTalk,
                        test_parameter = list(list(stimulation_power = 0.1),
                                              list(stimulation_power = 0.2))))
  })
  expect_output(analyse_Al2O3C_Measurement(data_CrossTalk[[2]],
                        test_parameter = list(stimulation_power = 0.01)))
  expect_output(analyse_Al2O3C_Measurement(data_CrossTalk[[2]],
                                           dose_points = list(3)))

  ## tests without TL curves
  SW({
  expect_output(analyse_Al2O3C_Measurement(osl, travel_dosimeter = 2),
                "travel dosimeter correction applied")
  expect_message(analyse_Al2O3C_Measurement(osl, travel_dosimeter = 1:2),
                 "'travel_dosimeter' specifies every position")
  expect_message(analyse_Al2O3C_Measurement(osl, travel_dosimeter = 2000),
                 "Invalid position in 'travel_dosimeter', nothing corrected")
  })

  ## irradiation_time_correction
  it.corr <- analyse_Al2O3C_ITC(data_ITC, verbose = FALSE)
  analyse_Al2O3C_Measurement(osl, irradiation_time_correction = list(it.corr),
                             plot = 1, verbose = FALSE)

  ## cross_talk_correction
  ct.corr <- analyse_Al2O3C_CrossTalk(data_CrossTalk)
  analyse_Al2O3C_Measurement(osl, cross_talk_correction = list(ct.corr),
                             plot = FALSE, verbose = FALSE)
  expect_message(analyse_Al2O3C_Measurement(osl, travel_dosimeter = 1:2,
                                            cross_talk_correction = c(-2e-4, -3e-4, 0),
                                            plot = FALSE, verbose = FALSE),
                 "'travel_dosimeter' specifies every position")

  ## integral_input
  set.seed(1)
  res1 <- analyse_Al2O3C_Measurement(osl, signal_integral = c(42.1, 42.3),
                                     integral_input = "measurement", verbose = FALSE)
  set.seed(1)
  res2 <- analyse_Al2O3C_Measurement(osl, signal_integral = 1:3,
                                     integral_input = "channel", verbose = FALSE)
  res1@info <- res2@info <- list() # remove $call
  expect_equal(res1, res2)

  expect_warning(analyse_Al2O3C_Measurement(osl[[1]], signal_integral = 1:4,
                                            integral_input = "measurement",
                                            verbose = FALSE),
                 "from time to channels failed: expected values in 42.1:51.9")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("defaults",
                              analyse_Al2O3C_Measurement(osl[2]))
  })
})

test_that("more coverage", {
  skip_on_cran()

  ## failed integration
  CT.mod <- data_CrossTalk
  CT.mod[[1]]@records[[2]]@data <- CT.mod[[1]]@records[[2]]@data[1:5, ]
  CT.mod[[1]]@records[[4]]@data <- CT.mod[[1]]@records[[4]]@data[1:5, ]
  warnings <- capture_warnings(
      analyse_Al2O3C_Measurement(CT.mod, verbose = FALSE,
                                 calculate_TL_dose = TRUE))
  expect_match(warnings, all = FALSE,
               "Natural TL signal out of bounds, NA returned")
  expect_match(warnings, all = FALSE,
               "Regenerated TL signal out of bounds, NA returned")

  ## missing position
  CT.mod <- data_CrossTalk
  CT.mod[[1]]@records[[1]]@info["position"] <- list(NULL)
  expect_message(expect_warning(
      analyse_Al2O3C_Measurement(CT.mod, verbose = FALSE),
      "TL peak shift detected for aliquot position NA"),
      "Error: Aliquot position not found, no cross-talk correction applied")

  CT.mod <- data_CrossTalk
  for (i in seq_along(CT.mod[[1]]@records)) {
    CT.mod[[1]]@records[[i]]@info$stimulator <- "LED"
  }
  suppressWarnings( # TL peak shift detected for aliquot position 1
  expect_warning(
      analyse_Al2O3C_Measurement(CT.mod, verbose = FALSE,
                                 test_parameter = list(stimulation_power = 0.03)),
      "Stimulation power was not stable for ALQ 1, results are likely to be wrong"))
})

#' Apply the Kars et al. (2008) model (deprecated)
#'
#' A function to calculate the expected sample specific fraction of saturation
#' following Kars et al. (2008) and Huntley (2006). This function is deprecated
#' and will eventually be removed. Please use `calc_Huntley2006()` instead.
#'
#' This function applies the approach described in Kars et al. (2008),
#' developed from the model of Huntley (2006) to calculate the expected sample
#' specific fraction of saturation of a feldspar and also to calculate fading
#' corrected age using this model. \eqn{\rho}' (`rhop`), the density of recombination
#' centres, is a crucial parameter of this model and must be determined
#' separately from a fading measurement. The function [analyse_FadingMeasurement]
#' can be used to calculate the sample specific \eqn{\rho}' value.
#'
#' @param fit.method [character] (*with default*):
#' Fit function of the dose response curve. Can either be `EXP` (the default)
#' or `GOK`. Note that `EXP` (single saturating exponential) is the original
#' function the model after Huntley (2006) and Kars et al. (2008) was
#' designed to use. The use of a general-order kinetics function (`GOK`)
#' is an experimental adaption of the model and should only be used
#' with great care.
#'
#' @param ... Parameters passed to [calc_Huntley2006].
#'
#' @return An [RLum.Results-class] object is returned:
#'
#' @section Function version: 0.4.0
#'
#' @author
#' Georgina E. King, University of Bern (Switzerland) \cr
#' Christoph Burow, University of Cologne (Germany)
#'
#' @note **This function is deprecated and will eventually be removed from the package.**
#' **Please use the function [calc_Huntley2006()] instead** 
#' **(use `fit.method = "EXP"` to apply the model after Kars et al., 2008).**
#'
#' @keywords datagen
#'
#' @references
#'
#' Kars, R.H., Wallinga, J., Cohen, K.M., 2008. A new approach towards anomalous fading correction for feldspar
#' IRSL dating-tests on samples in field saturation. Radiation Measurements 43, 786-790. doi:10.1016/j.radmeas.2008.01.021
#'
#' Huntley, D.J., 2006. An explanation of the power-law decay of luminescence.
#' Journal of Physics: Condensed Matter 18, 1359-1365. doi:10.1088/0953-8984/18/4/020
#'
#' King, G.E., Herman, F., Lambert, R., Valla, P.G., Guralnik, B., 2016.
#' Multi-OSL-thermochronometry of feldspar. Quaternary Geochronology 33, 76-87. doi:10.1016/j.quageo.2016.01.004
#'
#'
#' **Further reading**
#'
#' Morthekai, P., Jain, M., Cunha, P.P., Azevedo, J.M., Singhvi, A.K., 2011. An attempt to correct
#' for the fading in million year old basaltic rocks. Geochronometria 38(3), 223-230.
#'
#' @examples
#'
#' ## Load example data (sample UNIL/NB123, see ?ExampleData.Fading)
#' data("ExampleData.Fading", envir = environment())
#'
#' ## (1) Set all relevant parameters
#' # a. fading measurement data (IR50)
#' fading_data <- ExampleData.Fading$fading.data$IR50
#'
#' # b. Dose response curve data
#' data <- ExampleData.Fading$equivalentDose.data$IR50
#'
#' ## (2) Define required function parameters
#' ddot <- c(7.00, 0.004)
#' readerDdot <- c(0.134, 0.0067)
#'
#' # Analyse fading measurement and get an estimate of rho'.
#' # Note that the RLum.Results object can be directly used for further processing.
#' # The number of MC runs is reduced for this example
#' rhop <- analyse_FadingMeasurement(fading_data, plot = TRUE, verbose = FALSE, n.MC = 10)
#'
#' ## (3) Apply the Kars et al. (2008) model to the data
#' kars <- suppressWarnings(
#'   calc_Kars2008(data = data,
#'                 rhop = rhop,
#'                 ddot = ddot,
#'                 readerDdot = readerDdot,
#'                 n.MC = 25)
#' )
#'
#' @md
#' @export
calc_Kars2008 <- function(fit.method = "EXP", ...) {
  .Deprecated("calc_Huntley2006")
  calc_Huntley2006(fit.method = fit.method, ...)
}

set.seed(1)
data("ExampleData.Fading", envir = environment())
fading_data <- ExampleData.Fading$fading.data$IR50
data <- ExampleData.Fading$equivalentDose.data$IR50
ddot <- c(7.00, 0.004)
readerDdot <- c(0.134, 0.0067)

rhop <-
  analyse_FadingMeasurement(fading_data,
                            plot = FALSE,
                            verbose = FALSE,
                            n.MC = 10)
huntley <- calc_Huntley2006(
  data = data,
  rhop = rhop,
  ddot = ddot,
  readerDdot = readerDdot,
  n.MC = 50,
  plot = FALSE,
  verbose = FALSE
)

test_that("check class and length of output", {
  testthat::skip_on_cran()
  local_edition(3)

  ##rhop
  expect_s4_class(rhop, class = "RLum.Results")
  expect_s3_class(rhop$fading_results, "data.frame")
  expect_s3_class(rhop$fit, "lm")
  expect_s3_class(rhop$rho_prime, "data.frame")

  ##kars
  expect_s4_class(huntley, class = "RLum.Results")
  expect_s3_class(huntley$results, class = "data.frame")
  expect_s3_class(huntley$data, class = "data.frame")
  expect_type(huntley$Ln, "double")
  expect_type(huntley$fits, "list")

})

test_that("check values from analyse_FadingMeasurement()", {
    expect_equal(round(sum(rhop$fading_results[,1:9]),0),415)
    expect_equal(round(sum(rhop$rho_prime),5),2e-05)
    expect_equal(round(sum(rhop$irr.times)), 2673108)

})

test_that("check values from calc_Huntley2008()", {
  testthat::skip_on_cran()

  ##fix for different R versions
  if(R.version$major == "3" && as.numeric(R.version$minor) < 6){
    expect_equal(round(huntley$results$Sim_Age, 1), 41.3)
    expect_equal(round(huntley$results$Sim_Age_2D0, 0), 164)
    expect_equal(round(sum(huntley$Ln),4), 0.1585)

  }else{
    expect_equal(round(huntley$results$Sim_Age, 1), 42.3)
    expect_equal(round(huntley$results$Sim_Age_2D0, 0), 163)
    expect_equal(round(sum(huntley$Ln),4), 0.1621)

  }


  expect_equal(round(sum(huntley$data),0), 191530)
  expect_equal(round(sum(residuals(huntley$fits$simulated)),1),  0.3)
  expect_equal(round(sum(residuals(huntley$fits$measured)),4),  0.1894)
  expect_equal(round(sum(residuals(huntley$fits$unfaded)),4),  1.6293)

## COMPARE calc_Kars2008 (deprecated) vs. re-named calc_Huntley2006
test_that("compare deprecated calc_Kars2008 and calc_Huntley2006", {
  testthat::skip_on_cran()

  expect_identical({
    set.seed(1)
    calc_Huntley2006(
      data = data,
      rhop = rhop,
      ddot = ddot,
      readerDdot = readerDdot,
      n.MC = 50,
      fit.method = "EXP",
      plot = FALSE, verbose = FALSE)$results
  }, {
    set.seed(1)
    suppressWarnings(
    calc_Kars2008(
      data = data,
      rhop = rhop,
      ddot = ddot,
      readerDdot = readerDdot,
      n.MC = 50,
      fit.method = "EXP",
      plot = FALSE, verbose = FALSE)$results
    )
  })#EndOf::expect_identical

  expect_identical({
    set.seed(1)
    calc_Huntley2006(
      data = data,
      rhop = rhop,
      ddot = ddot,
      readerDdot = readerDdot,
      n.MC = 500,
      fit.method = "GOK",
      plot = FALSE, verbose = FALSE)$results
  }, {
    set.seed(1)
    suppressWarnings(
      calc_Kars2008(
        data = data,
        rhop = rhop,
        ddot = ddot,
        readerDdot = readerDdot,
        n.MC = 500,
        fit.method = "GOK",
        plot = FALSE, verbose = FALSE)$results
    )
  })#EndOf::expect_identical

})

})

#' Function to return essential quotes
#'
#' This function returns one of the collected essential quotes in the
#' growing library. If called without any parameters, a random quote is
#' returned.
#'
#' @param ID [character] (*optional*): quote ID to be returned.
#'
#' @param separated [logical] (*with default*): return result in separated form.
#'
#' @return Returns a character with quote and respective (false) author.
#'
#' @section Function version: 0.1.5
#'
#' @author Quote credits: Michael Dietze, GFZ Potsdam (Germany), Sebastian Kreutzer, Geography & Earth Science, Aberystwyth University (United Kingdom), Dirk Mittelstraß, TU Dresden (Germany), Jakob Wallinga (Wageningen University, Netherlands)
#'
#' @examples
#'
#' ## ask for an arbitrary quote
#' get_Quote()
#'
#' @noRd
get_Quote <- function(
  ID,
  separated = FALSE
) {

  ## definition of the ever growing quote data set
  quotes <- rbind(
    c("Anonymous student hotel employee", "Let me double check this."),
    c("The ordinary reviewer", "I love it when a plan comes together."),
    c("A tunnelling electron", "God does not play dice."),
    c("Goldfinger", "You cannot get this machine better and cheaper than from us."),
    c("A PhD supervisor", "Live long and in prosper."),
    c("A PhD supervisor", "You are not depressive, you simply have a crappy life."),
    c("A trapped charge", "I want to break free."),
    c("The R-package Luminescence manual", "Call unto me, and I will answer thee, and will shew thee great things, and difficult, which thou knowest not."),
    c("A stimulated feldspar grain", "I'm so excited and I just can't hide it."),
    c("The true age", "How many roads..."),
    c("The undecided OSL component", "Should I stay or should I go?"),
    c("A fluvially transported quartz grain at night", "Always look at the bright side of life."),
    c("An arctic sediment outcrop", "Marmor, Stein und Eisen bricht..."),
    c("A common luminescence reader customer", "If anything can go wrong, it will."),
    c("A blue LED to a trapped electron", "Resistance is futile."),
    c("A trapped electron to a yellow LED", "Well, that's all you've got?"),
    c("A weathering rock", "Who wants to live forever?"),
    c("A new pIRIR derivative", "20,000 miles below the sea."),
    c("Robert Oppenheimer", "I want this thing to work by just pressing one button."),
    c("An arbitrary member of the CRAN team", "No shirt, no shoes, no service!"),
    c("Rubber mallet to steel cylinder", "Let's rock and roll."),
    c("A data import function", "Better late than never."),
    c("A luminescence lab staff member to its customer", "Tell me the age, I tell you the price."),
    c("The NSA", "O'zapft is."),
    c("The natural dose", "You only live once."),
    c("A Windows user", "An apple a day keeps the doctor away."),
    c("The authors of sTeve", "We love to entertain you."),
    c("Any arbitrary independent OSL device manufacturer", "Sure it will work, it was me who built it!"),
    c("Response to the reviewer", "You are right, it was just a guess."),
    c("An aliquot disc", "The answer [...] is: 48"),
    c("Push Pin", "Made of used sample carriers"),
    c("A motivated R-Team member", "We are doing this not just for statistical reasons, there is real science behind it!"),
    c("An unbiased reviewer", "The data is too poor to be published in QG, try a higher ranked journal."),
    c("R Team member, asked about statistical details", "No idea, I'm just here for the visualisation."),
    c("An arbitrary unexperienced RLum-user", "Little by little, the bird builds its nest."),
    c("The answer to life, the universe and everything", "get_rightAnswer()"),
    c("Der Tatortreiniger", "Dreck ist nur Materie am falschen Ort."),
    c("Die Ex vom Tatortreiniger", "Das Ziel ist im Weg."),
    c("Bright grain to dim grains", "I'm so shiny!"),
    c("Fast component to slow component", "Life is short!"),
    c("Fast component to slow component", "What are you waiting for?"),
    c("Violet photon to deep trap electron", "Today I'm kicking you out of your comfort zone!"),
    c("Deep trap electron to infrared photon", "Don't bother me, I need to rest."),
    c("A single grain", "I feel so lonley."),
    c("Luminescence data to Bayesian process", "Don't you ever touch me again."),
    c("Quartz grain to heating plate", "Go ahead, I need a phase change."),
    c("Photon to electron", "I am in charge!"),
    c("You cannot spell 'data analysis' without 'daily satan'"),
    c("Bright grain to dim grain", "Don't you get it?")
    )

  ## Check input data
  if (missing(ID)) {
    ID <- sample(x = seq(from = 1,
                         to = nrow(quotes)),
                 size = 1)
  }

  ## check for correct ID and generate qoute
  if (length(ID) < 1 || ID > nrow(quotes)) {
    "Sorry, but this was an impossible task!"
  } else if (separated) {
    quotes[ID, ]
  } else {
    paste0(quotes[ID, 1], ": '", quotes[ID, 2], "'")
  }
}

test_that("Test zzz functions ... they should still work", {
  testthat::skip_on_cran()

  ##get right answer
  expect_equal(get_rightAnswer(), 46)
  expect_equal(get_rightAnswer("test"), 46)

  ##get quote
  expect_silent(get_Quote())
  expect_silent(get_Quote(ID = 1))
  expect_silent(get_Quote(ID = 10, separated = TRUE))
  expect_silent(get_Quote(ID = 1e06))

  ##tune data
  expect_warning(tune_Data(1:10))
  expect_warning(tune_Data(data.frame(runif(n = 10, 8,12),runif(n = 10, 0.1,0.3) ), decrease.error = TRUE))
  expect_warning(tune_Data(data.frame(runif(n = 10, 8,12),runif(n = 10, 0.1,0.3) ), increase.data = TRUE))

  ##sTeve
  ## read example data set
  data(ExampleData.DeValues, envir = environment())
  ExampleData.DeValues <-
    convert_Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))

  ## create plot straightforward
  expect_error(sTeve("error"),
               "'n_frames' should be a single positive integer value")
  expect_error(sTeve(c(1, NaN, 0)),
               "'n_frames' should be a single positive integer value")
  expect_error(sTeve(t_animation = 0),
               "'t_animation' should be a single positive integer value")
  expect_error(sTeve(n.tree = 0.12),
               "'n.tree' should be a single positive integer value")
  expect_silent(plot_KDE(data = ExampleData.DeValues))
  expect_silent(sTeve(type = 1))
  expect_silent(sTeve(type = 2, t_animation = 1))
  expect_silent(sTeve(type = 3, t_animation = 1, n.tree = 2))
  expect_silent(sTeve(n_frames = 1, n.tree = 1))
})

#' @title Apply spectral efficiency correction to RLum.Data.Spectrum objects
#'
#' @description
#' The function applies spectral efficiency correction to
#' [Luminescence::RLum.Data.Spectrum-class] objects.
#'
#' @details
#' The efficiency correction is based on a spectral response dataset provided
#' by the user. Usually the data set for the quantum efficiency is of lower
#' resolution and values are interpolated for the required spectral resolution using
#' function [stats::approx].
#'
#' If the energy calibration differs for both data set `NA` values are produces that
#' will be removed from the matrix.
#'
#' @param object [Luminescence::RLum.Data.Spectrum-class] or [Luminescence::RLum.Analysis-class] (**required**):
#' object of class `RLum.Data.Spectrum`, `RLum.Analysis`or a [list] of such
#' objects. Other objects in the list are skipped.
#'
#' @param spectral.efficiency [data.frame] (**required**):
#' data frame with 2 columns containing wavelengths and relative spectral
#' response values (values between 0 and 1). The provided data will be used to
#' correct all spectra if `object` is a [list].
#'
#' @return Returns same object as provided as input
#'
#' @note
#' Please note that the spectral efficiency data from the camera alone may not
#' sufficiently correct for spectral efficiency of the entire optical system
#' (e.g., spectrometer, camera ...).
#'
#' @section Function version: 0.2.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Johannes Friedrich, University of Bayreuth (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class]
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(1) - use with your own data (uncomment for usage)
#' ## spectral.efficiency <- read.csv("your data")
#' ##
#' ## your.spectrum <- apply_EfficiencyCorrection(your.spectrum, )
#'
#' @export
apply_EfficiencyCorrection <- function(
  object,
  spectral.efficiency
) {
  .set_function_name("apply_EfficiencyCorrection")
  on.exit(.unset_function_name(), add = TRUE)

  # self-call -----------------------------------------------------------------------------------

  ##case we have a list
  if(inherits(object, "list")){
    output_list <- lapply(object, function(o){
      if (!inherits(o, c("RLum.Data.Spectrum", "RLum.Analysis"))) {
        .throw_warning("Skipping '", class(o), "' object in input list")
        return(o)
      }
      apply_EfficiencyCorrection(o, spectral.efficiency)
    })

    return(output_list)
  }

  ##the case of an RLum.Analysis object
  if(inherits(object, "RLum.Analysis")){
    object@records <- lapply(object@records, function(o){
      if (!inherits(o, "RLum.Data.Spectrum")) {
        .throw_warning("Skipping '", class(o), "' object in input list")
        return(o)
      }
      apply_EfficiencyCorrection(o, spectral.efficiency)
    })

    return(object)
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Data.Spectrum")
  .validate_class(spectral.efficiency, "data.frame")
  .validate_not_empty(spectral.efficiency)
  if (ncol(spectral.efficiency) < 2)
    .throw_error("'spectral.efficiency' should have 2 columns")

  ## grep data matrix from the input object
  temp.matrix <- as(object, "matrix")

  ## remove missing values
  spectral.efficiency <- na.exclude(spectral.efficiency)
  if (nrow(spectral.efficiency) == 0)
    .throw_error("No valid data remains in 'spectral.efficiency' after ",
                 "removing NA values")

  ## grep efficency values
  temp.efficiency <- as.matrix(spectral.efficiency[,1:2])

  ##test max
  if(max(temp.efficiency[,2]) > 1)
    .throw_error("Relative quantum efficiency values > 1 are not allowed")

  # Apply method ------------------------------------------------------------

  ##the interpolation is needed to align the resolution
  #set data for interpolation
  temp.efficiency.x <- as.numeric(row.names(temp.matrix))

  interpolated  <- approx(
    x = temp.efficiency[,1],
    y = temp.efficiency[,2],
    xout = temp.efficiency.x,
    ties = mean)

  if (all(is.na(interpolated$y))) {
    .throw_error("Interpolation failed: this happens when the x-values in ",
                 "'spectral.efficiency' are outside the range of the x-values ",
                 "in 'object'")
  }

  ##correct for quantum efficiency
  temp.matrix <- vapply(1:ncol(temp.matrix), FUN = function(x) {
    temp.matrix[, x] / interpolated$y * max(interpolated$y, na.rm = TRUE)
  }, FUN.VALUE = numeric(nrow(temp.matrix)))

  ##remove NA values
  temp.matrix <- na.exclude(temp.matrix)

  ##correct colnames
  colnames(temp.matrix) <- colnames(get_RLum(object))

  # Return Output------------------------------------------------------------
  set_RLum(
    class = "RLum.Data.Spectrum",
    recordType = object@recordType,
    curveType = object@curveType,
    data = temp.matrix,
    info = object@info)
}

## load data
set.seed(1)
data(ExampleData.XSYG, envir = environment())
eff_data <- data.frame(WAVELENGTH = 1:1000, runif(1000))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(apply_EfficiencyCorrection(object = "error"),
               "'object' should be of class 'RLum.Data.Spectrum'")
  expect_error(apply_EfficiencyCorrection(TL.Spectrum,
                                          spectral.efficiency = "error"),
               "'spectral.efficiency' should be of class 'data.frame'")
  expect_error(apply_EfficiencyCorrection(TL.Spectrum,
                                          spectral.efficiency = data.frame()),
               "'spectral.efficiency' cannot be an empty data.frame")
  expect_error(apply_EfficiencyCorrection(TL.Spectrum,
                                          spectral.efficiency = iris[, 1, drop = FALSE]),
               "'spectral.efficiency' should have 2 columns")
  expect_error(apply_EfficiencyCorrection(TL.Spectrum,
                                          spectral.efficiency = data.frame(1:10, NA)),
               "No valid data remains in 'spectral.efficiency' after removing")
  expect_error(apply_EfficiencyCorrection(TL.Spectrum,
                                          spectral.efficiency = data.frame(1:10,
                                                                           runif(10))),
               "Interpolation failed: this happens when the x-values in")
  eff_data[1, 2] <- 2
  expect_error(apply_EfficiencyCorrection(TL.Spectrum,
                                          spectral.efficiency = eff_data),
               "Relative quantum efficiency values > 1 are not allowed")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##run list test
  expect_warning(
    apply_EfficiencyCorrection(list(a = "test", TL.Spectrum), spectral.efficiency = eff_data),
    "Skipping 'character' object in input list")

  ##run test with RLum.Analysis objects
  expect_s4_class(
    apply_EfficiencyCorrection(set_RLum("RLum.Analysis",
                                        records = list(TL.Spectrum)), spectral.efficiency = eff_data),
    "RLum.Analysis")
  expect_warning(
      apply_EfficiencyCorrection(set_RLum("RLum.Analysis",
                                          records = list(TL.Spectrum, "test")),
                                 spectral.efficiency = eff_data),
      "Skipping 'character' object in input list")

  ##run test with everything combined
  input <- list(a = "test", TL.Spectrum,set_RLum("RLum.Analysis", records = list(TL.Spectrum)))
  expect_warning(apply_EfficiencyCorrection(input, eff_data),
                 "Skipping 'character' object in input list")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-5

  expect_snapshot_RLum(apply_EfficiencyCorrection(TL.Spectrum,
                                                  spectral.efficiency = eff_data),
                       tolerance = snapshot.tolerance)
})

#' @title Plot a disc with its values
#'
#' @description Shows a schematic representation of the physical appearance
#' of one disc (one position in the reader)
#' and illustrates the measured or calculated values per grain location.
#'
#' @details Depending of the available plotting space, some optional elements might have not enough room
#' to be displayed. As this function is wrapped around the base plot function, one can also choose to add elements
#' manually.
#'
#' @param object [Luminescence::RLum.Results-class] or [numeric] (**required**):
#' the values to show, should have length 100.
#'
#' @param show_coordinates [logical] (*with default*): Show coordinates (1..10)
#' in x and in y direction. Defaults to `FALSE`.
#'
#' @param show_location_ids [logical] (*with default*): Show id with every
#' grain location (1..100). Defaults to `FALSE`.
#'
#' @param show_neighbours [logical] (*with default*): Show which
#' neighbour connections are taken into account if calculating Moran's I.
#' This makes sense when there are `NA` observations, or when a non-standard
#' neighbour setting is defined.
#'
#' @param show_positioning_holes [logical] (*with default*): Show the 3
#' positioning holes for orientation. Defaults to `TRUE`.
#'
#' @param df_neighbours [data.frame] (*with default*): only relevant if
#' `show_neighbours` is `TRUE`. Data frame indicating which borders to
#' consider, and their respective weights (see the description provided for
#' [Luminescence::calc_MoransI]). If `NULL` (default), this is constructed automatically by
#' the internal function `.get_Neighbours`.
#'
#' @param ignore_borders [logical] (*with default*): whether only grain
#' locations that do not lie on the border of the disc should be considered
#' (`FALSE` by default). Thus if `TRUE`, only the inner 8x8 grain locations
#' rather than the full 10x10 are considered. Ignored if `df_neighbours` is
#' not `NULL` or if `show_neighbours = FALSE`.
#'
#' @param str_transform [character] (*with default*): The observed value of each individual grain is
#' reflected in the size of a triangle (or other dot-like element). To account for large value differences,
#' the transformation from value to triangle size can be `"lin"` (linear), `"log"` (logarithmic) and `"sqrt"`
#' (square root). Defaults to `"sqrt"`, so that the surface is linear to the value. Note that
#' the log and sqrt transformations can come with an addition to avoid negative values. When the legend
#' is shown, the actual lower, middle and upper values are printed.
#'
#' @param ... other arguments to be given to the base R plot function, such
#' as `main`, `col` and `pch`. `legend` can be used to enable/disable the
#' legend (`FALSE` by default).
#'
#' @author
#' Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
#' Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
#'
#' @references
#' de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
#' A novel method to assess crosstalk in single-grain luminescence detection.
#' Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
#'
#' @examples
#'
#' plot_SingleGrainDisc(1:100)
#'
#' @export
plot_SingleGrainDisc <- function(
    object,
    show_coordinates = FALSE,
    show_location_ids = FALSE,
    show_neighbours = FALSE,
    show_positioning_holes = TRUE,
    df_neighbours = NULL,
    ignore_borders = FALSE,
    str_transform = c("sqrt", "lin", "log"),
    ...
) {
  .set_function_name("plot_SingleGrainDisc")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("RLum.Results", "numeric", "integer"))
  if (is.numeric(object)) {
    vn_values_to_show <- object
  } else {
    vn_values_to_show <- get_RLum(object)
  }
  .validate_length(vn_values_to_show, 100, name = "'object'")
  if (all(is.na(vn_values_to_show))) {
    .throw_error("'object' should contain at least one non-NA value")
  }

  .validate_logical_scalar(show_coordinates)
  .validate_logical_scalar(show_location_ids)
  .validate_logical_scalar(show_neighbours)
  .validate_logical_scalar(show_positioning_holes)

  .validate_class(df_neighbours, "data.frame", null.ok = TRUE)
  .validate_logical_scalar(ignore_borders)
  if (is.null(df_neighbours)) {
    df_neighbours <- .get_Neighbours(object, ignore_borders)
  } else if (ncol(df_neighbours) != 3) {
    .throw_error("'df_neighbours' should be a data frame with 3 columns")
  }

  str_transform <- .validate_args(str_transform, c("sqrt", "lin", "log"))

  ## get ... arguments
  plot_settings <- modifyList(
      x = list(
          main = "",
          legend = FALSE,
          col =  "darkolivegreen",
          pch = 17),
      val = list(...))

  ## Settings  -----------------------

  n_lb_cex <- 0.8  # lower bound of plotted grain value size
  n_ub_cex <- 3.2  # upper bound of plotted grain value size

  str_title <- plot_settings$main

  ## Scaling from value to point size -----------------------

  if (sum(!is.na(vn_values_to_show)) == 1) {
    ## Scale if there is one observation
    vn_range_values <- range(c(vn_values_to_show, 0), na.rm = TRUE)
    n_mean_value <- mean(vn_range_values)
    vn_values_to_show[101] <- n_mean_value  ## Temporarily add mean value to vector for scaling
    vn_values_to_show <- (vn_values_to_show-vn_range_values[1]) / diff(vn_range_values)

  } else {
    vn_range_values <- range(vn_values_to_show, na.rm = TRUE)
    n_mean_value <- mean(vn_range_values)
    vn_values_to_show[101] <- n_mean_value ## Temporarily add mean value to vector for scaling
    ## Scale to from 0 to inf
    vn_values_to_show <- vn_values_to_show - min(vn_values_to_show, na.rm = TRUE)
    if (str_transform == "log") {
      ## Scale to from 1 to inf, also to avoid problems with non-positive values
      vn_values_to_show <- log(vn_values_to_show + 1)
    }
    else if (str_transform == "sqrt") {
      ## Transform
      vn_values_to_show <- sqrt(vn_values_to_show)
    }

    ## Scale to from n_ub_cex to n_lb_cex)
    vn_values_to_show <- n_lb_cex + vn_values_to_show /
      (max(vn_values_to_show, na.rm = TRUE) / (n_ub_cex - n_lb_cex))
  }

  n_mv_cex <- vn_values_to_show[101] ## Extract mean value point size
  vn_values_to_show <- vn_values_to_show[-101]

  ## Plotting  -----------------------
  df_disc <- data.frame(x = rep_len(1:10, 100),
                        y = rep(1:10, each=10)
  )

  ## remove used arguments to avoid collisions
  args <- list(...)
  args <- args[!...names() %in% c("main", "legend", "col", "pch")]

  do.call(
      what = plot,
      args = c(list(
          x = df_disc$x,
          y = df_disc$y,
          cex = vn_values_to_show,
          asp = 1,
          col = plot_settings$col,
          pch = plot_settings$pch,
          xlab = "",
          ylab = "",
          xlim = c(1,10),
          ylim = c(10,1),
          axes = FALSE,
          mar = c(4, 4, 6, 0.5)),
          args))

  ## Show title (if any)
  if (show_positioning_holes || show_coordinates) {
    ## move title up
    mtext(str_title, side = 3, line = 2.5, adj = 0.5, cex=1.3)
  } else {
    mtext(str_title, side = 3, line = 1, adj = 0.5, cex=1.5)
  }

  ## Show NA values
  if (anyNA(vn_values_to_show)) {
    points(x = df_disc$x,
           y = df_disc$y,
           cex = 0.5,
           pch = ifelse(is.na(vn_values_to_show), 4, NA),
           col = "orange"
    )
  }

  ## Show coordinates along axes
  if (show_coordinates) {
    axis(1, pos=-0.4, at=1:10, lwd=1, lwd.ticks=1)
    axis(4, pos=-0.3, at=1:10, las=2, lwd=1, lwd.ticks=1)
  }

  ## Show all grain location id's
  if (show_location_ids) {
    text(x = df_disc$x-0.2,
         y = df_disc$y-0.2,
         labels = 1:100,
         cex = 0.5,
         col = "black"
    )
  }

  ## Show the 3 big positioning holes
  if (show_positioning_holes) {
    ##            left middle  right
    points(x = c(-0.25,  5.5,   11.2),
           y = c(5.5,     0,    5.5),
           pch = 1,
           cex = 5,
           xpd=NA
    )
  }

  ## Indicate neighbouring positions used for Moran's I calculations
  if (show_neighbours) {
    n_lines <- nrow(df_neighbours)
    if (n_lines == 0)
      .throw_error("'show_neighbours' is TRUE but 'df_neighbours' is empty")


    ## From location ID to x y coordinates; put into lists for use with mapply
    list_x <- split((df_neighbours[,1:2] - 1) %% 10 + 1,
                    1:n_lines )
    list_y <- split((df_neighbours[,1:2] - 1) %/% 10 + 1,
                    1:n_lines )

    ## Scale weights for nice plotting
    li_weight <- split((2 * df_neighbours$weight / max(df_neighbours$weight)),
                       1:n_lines)

    mapply(lines, x = list_x, y = list_y, col = "purple", lwd = li_weight, lty = "dotted")
  }

  if (plot_settings$legend) {
    pch <- plot_settings$pch
    col <- plot_settings$col
    n_low <-  prettyNum(x = vn_range_values[1], digits = 2, format = "fg")
    n_mean <- prettyNum(x = n_mean_value, digits = 2, format = "fg")
    n_high <- prettyNum(x = vn_range_values[2], digits = 2, format = "fg")

    vs_legend <- c( n_low,    n_mean,    n_high )
    vn_pch    <- c( pch,      pch,       pch)
    vs_lty    <- c( NA,       NA,        NA)
    vs_cols   <- c( col,      col,       col)
    vn_pt_cex <- c( n_lb_cex, n_mv_cex,  n_ub_cex)

    ## if also NA values
    if (anyNA(vn_values_to_show)) {
      vs_legend <- c(vs_legend, "No obs.")
      vn_pch    <- c(vn_pch,     4 )
      vs_lty    <- c(vs_lty,    NA)
      vs_cols   <- c(vs_cols,   "orange")
      vn_pt_cex <- c(vn_pt_cex, 0.5)
    }

    if (show_neighbours) {
      vs_legend <- c(vs_legend, "Neighb.")
      vn_pch    <- c(vn_pch,     NA )
      vs_lty    <- c(vs_lty,    'dotted')
      vs_cols   <- c(vs_cols,   "purple")
      vn_pt_cex <- c(vn_pt_cex, NA)
    }

    legend(x = -0.0,
           y = 10.5,
           legend = vs_legend,
           pch = vn_pch,
           lty= vs_lty,
           lwd = 1,
           col = vs_cols,
           pt.cex = vn_pt_cex,
           cex = 0.8,
           bg = "gray98",
           horiz = TRUE,
           adj = c(0, 0.75),
           y.intersp=1.4,
           xpd=TRUE
    )
  }
}

## load data
obj <- set_RLum(class = "RLum.Results",
                data = list(vn_values = rep(x = c(1,2), each = 50)))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_SingleGrainDisc("error"),
               "'object' should be of class 'RLum.Results', 'numeric' or")
  expect_error(plot_SingleGrainDisc(1:50),
               "'object' should have length 100")
  expect_error(plot_SingleGrainDisc(set_RLum(class = "RLum.Results",
                                             data = list(vn_values = rep(NA, 100)))),
               "'object' should contain at least one non-NA value")
  expect_error(plot_SingleGrainDisc(obj, show_coordinates = "error"),
               "'show_coordinates' should be a single logical value")
  expect_error(plot_SingleGrainDisc(obj, show_location_ids = "error"),
               "'show_location_ids' should be a single logical value")
  expect_error(plot_SingleGrainDisc(obj, show_neighbours = "error"),
               "'show_neighbours' should be a single logical value")
  expect_error(plot_SingleGrainDisc(obj, df_neighbours = "error"),
               "'df_neighbours' should be of class 'data.frame'")
  expect_error(plot_SingleGrainDisc(obj, df_neighbours = data.frame()),
               "'df_neighbours' should be a data frame with 3 columns")
  expect_error(plot_SingleGrainDisc(obj, ignore_borders = "error"),
               "'ignore_borders' should be a single logical value")
  expect_error(plot_SingleGrainDisc(obj, show_positioning_holes = "error"),
               "'show_positioning_holes' should be a single logical value")
  expect_error(plot_SingleGrainDisc(obj, str_transform = "error"),
               "'str_transform' should be one of 'sqrt', 'lin' or 'log'")

  expect_error(plot_SingleGrainDisc(obj, df_neighbours = iris[0, 1:3],
                         show_neighbour = TRUE),
               "'show_neighbours' is TRUE but 'df_neighbours' is empty")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_silent(plot_SingleGrainDisc(obj))
  expect_silent(plot_SingleGrainDisc(1:100,
                                     show_location_ids = TRUE,
                                     show_neighbours = TRUE,
                                     show_positioning_holes = FALSE,
                                     legend = TRUE))
  expect_silent(plot_SingleGrainDisc(1:100,
                                     show_neighbours = TRUE,
                                     ignore_borders = TRUE))
  expect_silent(plot_SingleGrainDisc(c(rep(NA, 99), 100),
                                     show_coordinates = TRUE))

  expect_silent(plot_SingleGrainDisc(obj, str_transform = "lin"))
  expect_silent(plot_SingleGrainDisc(obj, str_transform = "log",
                                     col = "cornflowerblue", pch = 16))

  obj.na <- obj
  obj.na@data$vn_values[c(24, 73)] <- NA
  expect_silent(plot_SingleGrainDisc(obj.na, legend = TRUE))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("SingleGrainDisc defaults",
                              plot_SingleGrainDisc(obj))
  vdiffr::expect_doppelganger("SingleGrainDisc options",
                              plot_SingleGrainDisc(1:100,
                                                   show_location_ids = TRUE,
                                                   show_neighbours = TRUE,
                                                   show_positioning_holes = FALSE,
                                                   ignore_borders = TRUE,
                                                   show_coordinates = TRUE))
  })
})

#' @title Analyse post-IR IRSL measurement sequences
#'
#' @description
#' The function performs an analysis of post-IR IRSL sequences including curve
#' fitting on [Luminescence::RLum.Analysis-class] objects.
#'
#' @details To allow post-IR IRSL protocol (Thomsen et al., 2008) measurement
#' analyses, this function has been written as extended wrapper for function
#' [Luminescence::analyse_SAR.CWOSL], thus facilitating an entire sequence analysis in
#' one run. With this, its functionality is strictly limited by the
#' functionality provided by  [Luminescence::analyse_SAR.CWOSL].
#'
#' **Defining the sequence structure**
#'
#' The argument `sequence.structure` expects a shortened pattern of your
#' sequence structure and was mainly introduced to ease the use of the function.
#' For example: If your measurement data contains the following curves: `TL`,
#' `IRSL`, `IRSL`, `TL`, `IRSL`, `IRSL`, the pattern in `sequence.structure`
#' becomes `c('TL', 'IRSL', 'IRSL')`. The second part of your sequence for one
#' cycle should be similar and can be discarded. If this is not the case (e.g.,
#' additional hotbleach) such curves must be removed before using the function.
#'
#' **If the input is a `list`**
#'
#' If the input is a list of [Luminescence::RLum.Analysis-class] objects, every argument
#' can be provided as list to allow
#' for different sets of parameters for every single input element.
#' For further information see [Luminescence::analyse_SAR.CWOSL].
#'
#' @param object [Luminescence::RLum.Analysis-class] or [list] of
#' [Luminescence::RLum.Analysis-class] objects (**required**):
#' input object containing data for analysis.
#' If a [list] is provided the functions tries to iterate over each element
#' in the list.
#'
#' @param signal_integral [integer] (**required**):
#' vector of channels for the signal integral. Provide this value as a list
#' for different integration limits for the different IRSL curves.
#'
#' @param background_integral [integer] (**required**):
#' vector of channels for the background integral. Provide this value as a list
#' for different integration limits for the different IRSL curves.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param dose.points [numeric] (*optional*):
#' a numeric vector containing the dose points values. Using this argument overwrites dose point
#' values in the signal curves.
#'
#' @param sequence.structure [vector] [character] (*with default*):
#' specifies the general sequence structure. Allowed values are `"TL"` and
#' any `"IR"` combination (e.g., `"IR50"`,`"pIRIR225"`).
#' Additionally a parameter `"EXCLUDE"` is allowed to exclude curves from
#' the analysis (Note: If a preheat without PMT measurement is used, i.e.
#' preheat as none TL, remove the TL step.)
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param plot_singlePanels [logical] (*with default*):
#' enable/disable plotting of the results in a single windows for each plot.
#' Ignored if `plot = FALSE`.
#'
#' @param ... further arguments that will be passed to
#' [Luminescence::analyse_SAR.CWOSL] and [Luminescence::plot_DoseResponseCurve]. Furthermore, the
#' arguments `main` (headers), `log` (IRSL curves), `cex` (control
#' the size) and `mtext.outer` (additional text on the plot area) can be passed to influence the plotting. If the input
#' is a list, `main` can be passed as [vector] or [list].
#'
#' @return
#' Plots (*optional*) and an [Luminescence::RLum.Results-class] object is
#' returned containing the following elements:
#'
#' \tabular{lll}{
#' **DATA.OBJECT** \tab **TYPE** \tab **DESCRIPTION** \cr
#' `..$data` : \tab  `data.frame` \tab Table with De values \cr
#' `..$LnLxTnTx.table` : \tab `data.frame` \tab with the `LnLxTnTx` values \cr
#' `..$rejection.criteria` : \tab [data.frame] \tab rejection criteria \cr
#' `..$Formula` : \tab [list] \tab Function used for fitting of the dose response curve \cr
#' `..$call` : \tab [call] \tab the original function call
#' }
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @note
#' Best graphical output can be achieved by using the function `pdf`
#' with the following options:
#'
#' `pdf(file = "<YOUR FILENAME>", height = 18, width = 18)`
#'
#' @section Function version: 0.2.8
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::analyse_SAR.CWOSL], [Luminescence::calc_OSLLxTxRatio],
#' [Luminescence::plot_DoseResponseCurve], [Luminescence::fit_DoseResponseCurve]
#' [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Results-class],
#' [Luminescence::get_RLum]
#'
#' @references
#' Murray, A.S., Wintle, A.G., 2000. Luminescence dating of quartz
#' using an improved single-aliquot regenerative-dose protocol. Radiation
#' Measurements 32, 57-73. \doi{10.1016/S1350-4487(99)00253-X}
#'
#' Thomsen, K.J., Murray, A.S., Jain, M., Boetter-Jensen, L., 2008. Laboratory
#' fading rates of various luminescence signals from feldspar-rich sediment
#' extracts. Radiation Measurements 43, 1474-1486.
#' \doi{10.1016/j.radmeas.2008.06.002}
#'
#' @keywords datagen plot
#'
#' @examples
#'
#' ### NOTE: For this example existing example data are used. These data are non pIRIR data.
#' ###
#' ##(1) Compile example data set based on existing example data (SAR quartz measurement)
#' ##(a) Load example data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##(b) Transform the values from the first position in a RLum.Analysis object
#' object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#'
#' ##(c) Grep curves and exclude the last two (one TL and one IRSL)
#' object <- get_RLum(object, record.id = c(-29,-30))
#'
#' ##(d) Define new sequence structure and set new RLum.Analysis object
#' sequence.structure  <- c(1,2,2,3,4,4)
#' sequence.structure <- as.vector(sapply(seq(0,length(object)-1,by = 4),
#'                                        function(x){sequence.structure + x}))
#'
#' object <-  sapply(1:length(sequence.structure), function(x){
#'   object[[sequence.structure[x]]]
#' })
#'
#' object <- set_RLum(class = "RLum.Analysis", records = object, protocol = "pIRIR")
#'
#' ##(2) Perform pIRIR analysis (for this example with quartz OSL data!)
#' ## Note: output as single plots to avoid problems with this example
#' results <- analyse_pIRIRSequence(object,
#'      signal_integral = 1:2,
#'      background_integral = 900:1000,
#'      fit.method = "EXP",
#'      sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
#'      main = "Pseudo pIRIR data set based on quartz OSL",
#'      plot_singlePanels = TRUE)
#'
#'
#' ##(3) Perform pIRIR analysis (for this example with quartz OSL data!)
#' ## Alternative for PDF output, uncomment and complete for usage
#' \dontrun{
#' tempfile <- tempfile(fileext = ".pdf")
#' pdf(file = tempfile, height = 18, width = 18)
#'   results <- analyse_pIRIRSequence(object,
#'          signal_integral = 1:2,
#'          background_integral = 900:1000,
#'          fit.method = "EXP",
#'          main = "Pseudo pIRIR data set based on quartz OSL")
#'
#'   dev.off()
#' }
#'
#' @export
analyse_pIRIRSequence <- function(
  object,
  signal_integral,
  background_integral,
  integral_input = c("channel", "measurement"),
  dose.points = NULL,
  sequence.structure = c("TL", "IR50", "pIRIR225"),
  plot = TRUE,
  plot_singlePanels = FALSE,
  ...
) {
  .set_function_name("analyse_pIRIRSequence")
  on.exit(.unset_function_name(), add = TRUE)

  ## store the graphical parameters: this is done here because the function
  ## self calls, and this should ensure that we reset the global state to
  ## what it was when we were first called
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  integral_input <- .validate_args(integral_input, c("channel", "measurement"))

  ## deprecated argument
  if (any(grepl("[signal|background]\\.integral", ...names()))) {
    .deprecated(old = c("signal.integral", "background.integral"),
                new = c("signal_integral", "background_integral"),
                since = "1.2.0")
    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal_integral <- list(...)$signal.integral
    background_integral <- list(...)$background.integral
  }

  ## Self-call --------------------------------------------------------------
  if (inherits(object, "list")) {
    .validate_not_empty(object)
    lapply(object, .validate_class, "RLum.Analysis",
           name = "All elements of 'object'")

    ## make life easy
    if (missing("signal_integral")){
      signal_integral <- 1:2
      .throw_warning("'signal_integral' missing, set to 1:2")
    }

    .validate_class(background_integral, c("integer", "numeric"))

   ## expand input arguments
   rep.length <- length(object)

   signal_integral <- .listify(signal_integral, rep.length)
   background_integral <- .listify(background_integral, rep.length)
   sequence.structure <- .listify(sequence.structure, rep.length)
   dose.points <- .listify(dose.points, rep.length)

   if ("main" %in% ...names()) {
     main_list <- .listify(list(...)$main, rep.length)
   }

   ## run analysis
   temp <- .warningCatcher(lapply(1:length(object), function(x) {
      analyse_pIRIRSequence(object[[x]],
                        signal_integral = signal_integral[[x]],
                        background_integral = background_integral[[x]],
                        integral_input = integral_input,
                        dose.points = dose.points[[x]],
                        sequence.structure = sequence.structure[[x]],
                        plot = plot,
                        plot_singlePanels = plot_singlePanels,
                        main = ifelse("main"%in% ...names(), main_list[[x]], paste0("ALQ #",x)),
                        ...)
    }))

    ##combine everything to one RLum.Results object as this as what was written ... only
    ##one object

    ##merge results and check if the output became NULL
    results <- merge_RLum(temp)

    ##DO NOT use invisible here, this will stop the function from stopping
    if(length(results) == 0)
      return(NULL)
    return(results)
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis", extra = "'list'")
  .validate_logical_scalar(plot)
  .validate_logical_scalar(plot_singlePanels)

  ## there must be at least an IR step
  if (!any(grepl("IR", sequence.structure))) {
    .throw_error("'sequence.structure' should contain at least one IR step")
  }

  ## check allowed values in sequence structure
  invalid.terms <- .collapse(unique(grep("TL|IR|OSL|EXCLUDE", sequence.structure,
                                         invert = TRUE, value = TRUE)),
                             last_sep = " and ")
  if (invalid.terms != "") {
    .throw_error(invalid.terms, " not allowed in 'sequence.structure'")
  }

  ## deprecated argument
  if ("plot.single" %in% ...names()) {
    plot_singlePanels <- list(...)$plot.single
    .deprecated("plot.single", "plot_singlePanels", since = "1.0.0")
  }

  ## Deal with extra arguments
  extraArgs <- list(...)
  main <- extraArgs$main %||% "MEASUREMENT INFO"
  cex <- extraArgs$cex %||% 0.7

  ## Enforce a minimum plot device size: this is necessary as otherwise users
  ## may experience "figure margins too large" errors when trying to draw all
  ## plots on a single page. We need to round the device size values because
  ## often they are values such as 15.99999999999 which would incorrectly
  ## trigger our check
  min.size <- ceiling(11 * cex)
  dev.size <- round(grDevices::dev.size("in"), 5)
  if (plot && !plot_singlePanels && any(dev.size < min.size)) {
    plot <- FALSE
    msg <- paste0("Argument 'plot' reset to 'FALSE': the smallest plot ",
                  "size required is IN x IN in (at cex = ", cex, "). ",
                  "Consider plotting via `pdf(..., width = IN, height = IN)` ",
                  "or setting `plot_singlePanels = TRUE`. ")
    .throw_warning(gsub(x = msg, "IN", min.size))
  }

# Protocol Integrity Checks --------------------------------------------------
  ##(1) Check structure and remove curves that fit not the recordType criteria

  ##get sequence structure
  temp.sequence.structure  <- structure_RLum(object)

  ##remove data types that fit not to the allowed values
  temp.sequence.rm.id <- temp.sequence.structure[["id"]][
      !grepl("(T|OS|IRS)L", temp.sequence.structure[["recordType"]])]

  if(length(temp.sequence.rm.id)>0){
  ##removed record from data set
  object <- get_RLum(
    object = object,
    record.id = -temp.sequence.rm.id,
    drop = FALSE)

  .throw_warning("The following unrecognised record types have been removed: ",
                 .collapse(temp.sequence.structure[temp.sequence.rm.id,
                                                   "recordType"]))
  }

  ##(2) Apply user sequence structure
  temp.sequence.structure  <- structure_RLum(object)

    ##try to account for a very common mistake
    if(any(grepl(sequence.structure, pattern = "TL", fixed = TRUE)) && !any(grepl(temp.sequence.structure[["recordType"]], pattern = "TL", fixed = TRUE))){
      .throw_warning("Your sequence does not contain 'TL' curves, trying ",
                     "to adapt 'sequence.structure' for you ...")
      sequence.structure <- sequence.structure[!grepl(sequence.structure, pattern = "TL", fixed = TRUE)]
    }

  ##set values to structure data.frame
  ##but check first
  if(2 * length(
    rep(sequence.structure, nrow(temp.sequence.structure)/2/length(sequence.structure))) == length(temp.sequence.structure[["protocol.step"]])){
    temp.sequence.structure[["protocol.step"]] <- rep(
      sequence.structure, nrow(temp.sequence.structure)/2/length(sequence.structure))

  }else{
    .throw_message("The number of records is not a multiple of the defined ",
                   "sequence structure, NULL returned")
    return(NULL)
  }

  ##remove values that have been excluded
  rm.id <- temp.sequence.structure$id[temp.sequence.structure$protocol.step == "EXCLUDE"]

  if (length(rm.id) > 0) {
    ##remove from object
    object  <- get_RLum(object, record.id = -rm.id, drop = FALSE)

    ##remove from sequence structure
    sequence.structure  <- sequence.structure[sequence.structure != "EXCLUDE"]

    ##set new structure
    temp.sequence.structure  <- structure_RLum(object)

    temp.sequence.structure[, "protocol.step"] <- rep(
      sequence.structure, nrow(temp.sequence.structure)/2/length(temp.sequence.structure))

    .throw_warning(length(rm.id), " records have been removed due to EXCLUDE")
  }

##============================================================================##
# Analyse data and plotting ----------------------------------------------------
##============================================================================##

  ##(1) find out how many runs are needed for the analysis by checking for "IR"
  ##    now should by every signal except the TL curves
  n.TL <- sum(grepl("TL", sequence.structure))
  n.loops <- as.numeric(length(sequence.structure) - n.TL)

  ##grep ids of TL curves (we need them later on)
  TL.curves.id <- temp.sequence.structure[
    temp.sequence.structure[,"protocol.step"] == "TL","id"]

  ##grep ids of all OSL curves (we need them later on)
  IRSL.curves.id <- temp.sequence.structure[
    grepl("IR", temp.sequence.structure[,"protocol.step"]),"id"]

  ##grep information on the names of the IR curves, we need them later on
  pIRIR.curve.names  <- unique(temp.sequence.structure[
    temp.sequence.structure[IRSL.curves.id,"id"],"protocol.step"])

  ## Integrals checks -------------------------------------------------------

  if (integral_input == "measurement") {
    x.range <- object@records[[IRSL.curves.id[1]]][, 1]
    unit <- "time"
    signal_integral <- .convert_to_channels(x.range, signal_integral, unit,
                                            list.ok = TRUE)
    background_integral <- .convert_to_channels(x.range, background_integral, unit,
                                                list.ok = TRUE)
  }
  signal_integral <- .validate_integral(signal_integral, list.ok = TRUE)
  background_integral <- .validate_integral(background_integral, list.ok = TRUE)


  ##===========================================================================#
  ## set graphic layout using the layout option
  ## unfortunately a little bit more complicated then expected previously due
  ## the order of the produced plots by the previous functions

  if (plot && !plot_singlePanels) {
    ##first (Tx,Tn, Lx,Ln)
    temp.IRSL.layout.vector.first <- c(3, 5, 6, 7, 3, 5, 6, 8)

    ## middle (any other Lx,Ln)
    temp.IRSL.layout.vector.middle <- NULL
    if (n.loops > 2) {
    temp.IRSL.layout.vector.middle <-
      vapply(2:(n.loops - 1),
        FUN = function(x) 5 * x - 1 + c(0:3, 0:2, 4),
        FUN.VALUE = numeric(8)
      )
    }

    ## last (Lx, Ln and legend)
    temp.IRSL.layout.vector.last <- c(1, 2, 4, 5, 1, 2, 4, 6) +
      (if (n.loops > 2) max(temp.IRSL.layout.vector.middle)
       else max(temp.IRSL.layout.vector.first))

    temp.IRSL.layout.vector <- c(
      temp.IRSL.layout.vector.first,
      temp.IRSL.layout.vector.middle,
      temp.IRSL.layout.vector.last)

  ##set up layout matrix linked to the number of plot areas needed
  layout.matrix  <- c(
    rep(c(2,4,1,1),2), #header row with TL curves and info window
    temp.IRSL.layout.vector, #IRSL curves,
    rep((max(temp.IRSL.layout.vector)-3),8), #legend,
    rep((max(temp.IRSL.layout.vector)+1),1), #GC
    rep((max(temp.IRSL.layout.vector)+2),1), #TnTx
    rep((max(temp.IRSL.layout.vector)+3),2), #Rejection criteria
    rep((max(temp.IRSL.layout.vector)+1),1), #GC
    rep((max(temp.IRSL.layout.vector)+2),1), #TnTx
    rep((max(temp.IRSL.layout.vector)+3),2)) #Rejection criteria

  ##set layout
  nrows <- max(layout.matrix) / 2 + ifelse(n.loops > 2, 0, 2)
  mat <- matrix(layout.matrix, nrow = nrows, ncol = 4, byrow = TRUE)
  graphics::layout(
    mat = mat,
    widths = c(rep(c(1, 1, 1, 0.75), 6), 1, 1, 1, 1),
    heights = c(rep(1, 2 + 2 * n.loops), 0.20, 0.20))
  }

  ##(1) INFO PLOT
  if (plot) {
    plot(NA,NA,
         ylim = c(0,1), xlab = "",
         xlim = c(0,1), ylab = "",
         axes = FALSE,
         main = main)

    text(0.5,0.5, paste(sequence.structure, collapse = "\n"), cex = cex *2)
  }

  ##(2) set loop
  for(i in 1:n.loops){
    ##compile record ids
    temp.id.sel <-
      sort(c(TL.curves.id, IRSL.curves.id[seq(i,length(IRSL.curves.id),by=n.loops)]))

    ##(a) select data set (TL curves has to be considered for the data set)
    temp.curves <- get_RLum(object, record.id = temp.id.sel, drop = FALSE)

    ##(b) grep integral limits as they might be different for different curves
    if (inherits(signal_integral, "list") && length(signal_integral) > 1) {
      temp.signal_integral <- signal_integral[[i]]
      temp.background_integral <- background_integral[[i]]

    }else{
      temp.signal_integral <- signal_integral
      temp.background_integral <- background_integral
    }

    ##(c) call analysis sequence and plot
    ## call single plots
     temp.plot.single <-
    if(i == 1) {
      c(1,2,3,4,6)
    } else if(i == n.loops) {
      c(2,4,5,6)
    } else {
      c(2,4,6)
    }

    ##start analysis
    par(cex = cex)
    temp.results <- analyse_SAR.CWOSL(
      temp.curves,
      signal_integral = temp.signal_integral,
      background_integral = temp.background_integral,
      plot = plot,
      dose.points = dose.points,
      plot_singlePanels = temp.plot.single,
      cex = cex,
      ...
    ) ##TODO should be replaced with useful explicit arguments

      ##check whether NULL was return
      if (is.null(temp.results)) {
        .throw_message("Analysis skipped: check your sequence, NULL returned")
        return(NULL)
      }

      ##add signal information to the protocol step
      temp.results.pIRIR.De <- as.data.frame(c(
        get_RLum(temp.results, "data"),
        data.frame(Signal = pIRIR.curve.names[i])))

      temp.results.pIRIR.LnLxTnTx <- as.data.frame(c(
        get_RLum(temp.results, "LnLxTnTx.table"),
        data.frame(Signal = pIRIR.curve.names[i])))

      temp.results.pIRIR.rejection.criteria <- as.data.frame(c(
        get_RLum(temp.results, "rejection.criteria"),
        data.frame(Signal = pIRIR.curve.names[i])))

      temp.results.pIRIR.formula <- list(get_RLum(temp.results,
                                                  "Formula"))
      names(temp.results.pIRIR.formula)  <- pIRIR.curve.names[i]

      ##create now object
      temp.results  <- set_RLum(
        class = "RLum.Results",
        data = list(
          data = temp.results.pIRIR.De,
          LnLxTnTx.table = temp.results.pIRIR.LnLxTnTx,
          rejection.criteria = temp.results.pIRIR.rejection.criteria,
          Formula = temp.results.pIRIR.formula
        ),
        info = list(
          call = sys.call()
        )
      )

      ##merge results
      if (exists("temp.results.final")) {
        temp.results.final <- merge_RLum(list(temp.results.final, temp.results))
      } else{
        temp.results.final <- temp.results
      }
  }

##============================================================================##
# Plotting additional --------------------------------------------------------
##============================================================================##
if(plot){
  ##extract LnLnxTnTx.table
  LnLxTnTx.table <- get_RLum(temp.results.final, "LnLxTnTx.table")

  ## remove Inf
  LnLxTnTx.table$LxTx[is.infinite(LnLxTnTx.table$LxTx)] <- NA
  LnLxTnTx.table$LxTx.Error[is.infinite(LnLxTnTx.table$LxTx.Error)] <- NA

  ## plot dose response curves
  min.LxTx <- min(LnLxTnTx.table$LxTx, na.rm = TRUE)
  max.LxTx <- max(LnLxTnTx.table$LxTx, na.rm = TRUE)
  max.LxTx.Error <- max(LnLxTnTx.table$LxTx.Error, na.rm = TRUE)
  plot(NA, NA,
       xlim = range(get_RLum(temp.results.final, "LnLxTnTx.table")$Dose),
       ylim = c(min(min.LxTx - max.LxTx.Error, 0), max.LxTx + max.LxTx.Error),
       xlab = "Dose [s]",
       ylab = expression(L[x]/T[x]),
       mgp = c(2, 0.7, 0),
       tcl = -0.4,
       xpd = NA,
       main = "Summarised Dose Response Curves")

    ##set x for expression evaluation
    x <- seq(0,max(LnLxTnTx.table$Dose)*1.05, length.out = 100)

  for (j in 1:length(pIRIR.curve.names)) {
    ## dose points
    curve.points <- LnLxTnTx.table[LnLxTnTx.table$Signal == pIRIR.curve.names[j],
                                   c("Dose", "LxTx", "LxTx.Error")]

    points(curve.points[-1, c("Dose", "LxTx")], col = j, pch = j)
    segments(x0 = curve.points$Dose[-1],
             y0 = curve.points$LxTx[-1] - curve.points$LxTx.Error[-1],
             x1 = curve.points$Dose[-1],
             y1 = curve.points$LxTx[-1] + curve.points$LxTx.Error[-1],
             col = j)

    ## De values
    De <- get_RLum(temp.results.final, "data")[j, 1]
    Lx <- curve.points$LxTx[1]
    lines(c(-100, De, De, De),
          c(Lx, Lx, Lx, -100),
           col = j,
           lty = 2)

     ##curve
     temp.curve.formula  <- get_RLum(
        temp.results.final, "Formula")[[pIRIR.curve.names[j]]]

     try(lines(x, eval(temp.curve.formula), col = j), silent = TRUE)
  }

    rm(x)

    ##plot legend
    legend("bottomright", legend = pIRIR.curve.names,
           lty = 1, col = 1:length(pIRIR.curve.names),
           bty = "n",
           pch = 1:length(pIRIR.curve.names))

    ##plot Tn/Tx curves
    ##select signal
    temp.curve.TnTx <- LnLxTnTx.table[, c("TnTx", "Signal")]

    temp.curve.TnTx.matrix <- matrix(NA,
                                    nrow = nrow(temp.curve.TnTx)/
                                      length(pIRIR.curve.names),
                                    ncol =  length(pIRIR.curve.names))

    ##calculate normalised values
    for(j in 1:length(pIRIR.curve.names)){
      temp.curve.TnTx.sel <- temp.curve.TnTx[
        temp.curve.TnTx[,"Signal"] == pIRIR.curve.names[j]
        , "TnTx"]

      temp.curve.TnTx.matrix[,j] <- temp.curve.TnTx.sel/temp.curve.TnTx.sel[1]
    }

    plot(NA, NA,
       xlim = c(0, nrow(LnLxTnTx.table) / n.loops),
       ylim = if (anyNA(range(temp.curve.TnTx.matrix))) c(0,1) else range(temp.curve.TnTx.matrix),
       xlab = "# Cycle",
       ylab = expression(T[x]/T[n]),
       mgp = c(2, 0.7, 0),
       tcl = -0.4,
       main = "Sensitivity change")

    ##zero line
    abline(h = 1:nrow(temp.curve.TnTx.matrix), col = "gray")

    for(j in 1:length(pIRIR.curve.names)){
     lines(1:nrow(temp.curve.TnTx.matrix),
           temp.curve.TnTx.matrix[,j],
           type = "b",
           col = j,
           pch = j)
    }

   ##plot legend
   legend("bottomleft", legend = pIRIR.curve.names,
         lty = 1, col = 1:length(pIRIR.curve.names),
         bty = "n",
         pch = 1:length(pIRIR.curve.names))

   ##Rejection criteria
   temp.rejection.criteria <- get_RLum(temp.results.final,
                                               data.object = "rejection.criteria")

   temp.rc.reycling.ratio <- temp.rejection.criteria[
     grep("Recycling ratio",temp.rejection.criteria[,"Criteria"]),]

   temp.rc.recuperation.rate <- temp.rejection.criteria[
     grep("Recuperation rate",temp.rejection.criteria[,"Criteria"]),]

   temp.rc.palaedose.error <- temp.rejection.criteria[
     grep("Palaeodose error",temp.rejection.criteria[,"Criteria"]),]

   plot(NA,NA,
        xlim = c(-0.5,0.5),
        ylim = c(0,30),
        yaxt = "n", ylab = "",
        xaxt = "n", xlab = "",
        bty = "n",
        main = "Rejection criteria")

   axis(side = 1, at = c(-0.2,-0.1,0,0.1,0.2), labels = c("- 0.2", "- 0.1","0/1","+ 0.1", "+ 0.2"))
   ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++##
   ##polygon for recycling ratio
   text(x = -0.3, y = 25, "Recycling ratio", pos = 2, offset = 0.5, xpd = TRUE)
   thres_range <- as.numeric(as.character(temp.rc.reycling.ratio$Threshold))[1] - 1
   polygon(
     x = c(-thres_range, -thres_range, thres_range, thres_range),
     y = c(21, 29, 29, 21),
     col = "gray",
     border = NA)
    polygon(x = c(-0.3,-0.3,0.3,0.3) , y = c(21,29,29,21))

   ##consider possibility of multiple pIRIR signals and multiple recycling ratios
   col.id  <- 1

   ##the conditional case might valid if no rejection criteria could be calculated
   if(nrow(temp.rc.recuperation.rate)>0){

   for(i in seq(1,nrow(temp.rc.recuperation.rate),
                  length(unique(temp.rc.recuperation.rate[,"Criteria"])))){

        for(j in 0:length(unique(temp.rc.recuperation.rate[,"Criteria"]))){
         points(temp.rc.reycling.ratio[i+j, "Value"]-1,
               y = 25,
               pch = col.id,
               col = col.id)
        }
        col.id <- col.id + 1
   }
   }#endif

    rm(col.id)

   ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++##
   ##polygon for recuperation rate
   text(x = -0.3, y = 15, "Recuperation rate", pos = 2, offset = 0.5, xpd = TRUE)

   if(length(as.character(temp.rc.recuperation.rate$Threshold))>0){
   polygon(x = c(0,
                0,
                as.numeric(as.character(temp.rc.recuperation.rate$Threshold))[1],
                as.numeric(as.character(temp.rc.recuperation.rate$Threshold))[1]),
          y = c(11,19,19,11), col = "gray", border = NA)

   polygon(x = c(-0.3,-0.3,0.3,0.3) , y = c(11,19,19,11))
   polygon(x = c(-0.3,-0.3,0,0) , y = c(11,19,19,11), border = NA, density = 10, angle = 45)

  for(i in 1:nrow(temp.rc.recuperation.rate)){

    points(temp.rc.palaedose.error[i, "Value"],
           y = 15,
           pch = i,
           col = i)
  }
  }#endif

   ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++##
   ##polygon for palaeodose error
   text(x = -0.3, y = 5, "Palaeodose error", pos = 2, offset = 0.5, xpd = TRUE)
   polygon(x = c(0,
                0,
                as.numeric(as.character(temp.rc.palaedose.error$Threshold))[1],
                as.numeric(as.character(temp.rc.palaedose.error$Threshold))[1]),
          y = c(1,9,9,1), col = "gray", border = NA)
   polygon(x = c(-0.3,-0.3,0.3,0.3) , y = c(1,9,9,1))
   polygon(x = c(-0.3,-0.3,0,0) , y = c(1,9,9,1), border = NA, density = 10, angle = 45)

   for(i in 1:nrow(temp.rc.palaedose.error)){
     if(length(temp.rc.palaedose.error[i, "Value"]) > 0 && !is.na(temp.rc.palaedose.error[i, "Value"]))
       points(temp.rc.palaedose.error[i, "Value"],
              y = 5,
              pch = i,
              col = i)
   }

   ##add 0 value
   lines(x = c(0, 0), y = c(21, 29), lwd = 2)
   lines(x = c(0, 0), y = c(11, 19), lwd = 2)
   lines(x = c(0, 0), y = c(1, 9), lwd = 2)

  ##plot legend
  legend(0.3, 15,
         legend = pIRIR.curve.names,
         col = 1:length(pIRIR.curve.names),
         pch = 1:length(pIRIR.curve.names),
         yjust = 0.5,
         bty = "n",
         xpd = TRUE)
}##end plot == TRUE


##============================================================================##
# Return Values -----------------------------------------------------------
##============================================================================##

  temp.results.final
}

## load data
set.seed(1)
data(ExampleData.BINfileData, envir = environment())
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
object <- get_RLum(object, record.id = c(-29, -30))
sequence.structure  <- c(1, 2, 2, 3, 4, 4)
sequence.structure <-
  as.vector(sapply(seq(0, length(object) - 1, by = 4),
                   function(x) {
                     sequence.structure + x
                   }))

object <-  sapply(1:length(sequence.structure), function(x) {
  object[[sequence.structure[x]]]
})

object <-
  set_RLum(class = "RLum.Analysis",
           records = object,
           protocol = "pIRIR")

SW({
results <- analyse_pIRIRSequence(
  object,
  signal_integral = 1:2,
  background_integral = 900:1000,
  fit.method = "EXP",
  sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
  main = "Pseudo pIRIR data set based on quartz OSL",
  plot = TRUE,
  plot_singlePanels = TRUE,
  verbose = FALSE
)
})

test_that("check plot stuff", {
  ## check plot_RLum.Results
  expect_silent(plot_RLum.Results(results))

  ## it should throw a warning about the plot size
  pdf.out <- tempfile(fileext = ".pdf")
  pdf(pdf.out, width = 6, height = 8)
  expect_warning(analyse_pIRIRSequence(
    object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    fit.method = "EXP",
    sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
    main = "Pseudo pIRIR data set based on quartz OSL",
    plot = TRUE,
    plot_singlePanels = FALSE,
    verbose = FALSE),
    "[analyse_pIRIRSequence()] Argument 'plot' reset to 'FALSE'",
    fixed = TRUE)
  dev.off()

  ## this should not throw any warning because we print to a large enough pdf
  pdf(pdf.out, width = 16, height = 16)
  expect_silent(analyse_pIRIRSequence(
    object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    fit.method = "EXP",
    sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
    main = "Pseudo pIRIR data set based on quartz OSL",
    plot = TRUE,
    plot_singlePanels = FALSE,
    verbose = FALSE))
  dev.off()

  ## more coverage: get n.loops > 2
  pdf(pdf.out, width = 25, height = 25)
  suppressWarnings(analyse_pIRIRSequence(
      object,
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "EXP",
      sequence.structure = c("TL", paste0("pseudoIRSL", 1:6)),
      main = "Pseudo pIRIR data set based on quartz OSL",
      cex = 0.5,
      plot = TRUE,
      verbose = FALSE))
  dev.off()
  unlink(pdf.out)

  ## this should not throw any warning with plot_singlePanels = TRUE
  expect_silent(analyse_pIRIRSequence(
    object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    fit.method = "EXP",
    sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
    main = "Pseudo pIRIR data set based on quartz OSL",
    plot = TRUE,
    plot_singlePanels = TRUE,
    verbose = FALSE))

  suppressWarnings( # duplicated plot.single warnings from sanalyse_SAR.CWOSL()
  expect_warning(analyse_pIRIRSequence(
    object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    fit.method = "EXP",
    plot = TRUE,
    plot.single = TRUE,
    verbose = FALSE),
    "'plot.single' was deprecated in v1.0.0, use 'plot_singlePanels' instead")
  )

  ## integral_input
  set.seed(1)
  res1 <- analyse_pIRIRSequence(object,
                                signal_integral = c(0.04, 0.16),
                                background_integral = 34:40,
                                integral_input = "measurement",
                                fit.method = "EXP",
                                plot = FALSE,
                                verbose = FALSE)
  set.seed(1)
  res2 <- analyse_pIRIRSequence(object,
                                signal_integral = 1:4,
                                background_integral = 850:1000,
                                integral_input = "channel",
                                fit.method = "EXP",
                                plot = FALSE,
                                verbose = FALSE)
  res1@info <- res2@info <- list() # remove $call
  res1@.uid <- res1@.pid <- res2@.uid <- res2@.pid <- NA_character_
  res1@data$data$UID <- res2@data$data$UID <- NULL
  res1@data$LnLxTnTx.table$UID <- res2@data$LnLxTnTx.table$UID <- NULL
  res1@data$rejection.criteria$UID <- res2@data$rejection.criteria$UID <- NULL
  expect_equal(res1, res2)

  ## deprecated arguments
  suppressWarnings( # Argument 'plot' reset to 'FALSE'
  expect_warning(analyse_pIRIRSequence(object,
                                       signal.integral = 1:2,
                                       background.integral = 900:1000,
                                       verbose = FALSE),
                 "were deprecated in v1.2.0, use 'signal_integral' and")
  )
  expect_warning(expect_error(analyse_pIRIRSequence(object,
                                                    signal.integral = 1:2,
                                                    background.integral = 90:100,
                                                    integral_input = "measurement"),
                              "'integral_input' is not supported with old argument names"),
                 "were deprecated in v1.2.0, use 'signal_integral' and")
})

test_that("input validation", {
  expect_error(analyse_pIRIRSequence("test"),
               "'object' should be of class 'RLum.Analysis' or 'list'")
  expect_error(analyse_pIRIRSequence(list()),
               "'object' cannot be an empty list")
  expect_error(analyse_pIRIRSequence(list(data.frame())),
               "All elements of 'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_pIRIRSequence(list("test"),
                                     signal_integral = 1:2,
                                     background_integral = 900:1000),
               "'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_pIRIRSequence(object, plot = FALSE),
               "'signal_integral' should be of class 'integer', 'numeric' or 'list'")
  expect_error(analyse_pIRIRSequence(object,
                                     signal_integral = 1:2,
                                     background_integral = 900:1000,
                                     sequence.structure = "TL"),
               "'sequence.structure' should contain at least one IR step")
  expect_error(analyse_pIRIRSequence(object,
                                     signal_integral = 1:2,
                                     background_integral = 900:1000,
                                     sequence.structure = c("IR50", "error",
                                                            "error2", "error2")),
               "'error' and 'error2' not allowed in 'sequence.structure'")

  SW({
  expect_warning(expect_error(analyse_pIRIRSequence(list(object)),
                              "'background_integral' should be of class"),
                 "'signal_integral' missing, set to 1")
  expect_warning(analyse_pIRIRSequence(list(object),
                                       signal_integral = 1:2,
                                       background_integral = 900:1000,
                                       sequence.structure = c("TL", "IRSL",
                                                              "EXCLUDE"),
                                       plot = FALSE),
                 "14 records have been removed due to EXCLUDE")

  expect_message(expect_null(
      analyse_pIRIRSequence(list(object),
                            signal_integral = 1:2,
                            background_integral = 900:1000,
                            sequence.structure = c("TL", "pseudoIRSL1"),
                            plot = FALSE)),
      "The number of records is not a multiple of the defined sequence structure")

  object.warn <- object
  for (i in 1:6) object.warn@records[[i]]@recordType <- "error"
  expect_warning(analyse_pIRIRSequence(object.warn,
                                       main = "Title",
                                       signal_integral = 1:2,
                                       background_integral = 900:1000,
                                       plot = FALSE),
                 "The following unrecognised record types have been removed:")

  object.noTL <- subset(object, recordType != "TL (PMT)")
  expect_warning(analyse_pIRIRSequence(list(object.noTL),
                                       main = "Title",
                                       signal_integral = 1:2,
                                       background_integral = 900:1000,
                                       plot = FALSE),
                 "Your sequence does not contain 'TL' curves")
  })
})

test_that("check class and length of output", {
    testthat::skip_on_cran()

    expect_s4_class(results, "RLum.Results")
    expect_equal(length(results), 4)
    expect_s3_class(results$LnLxTnTx.table, "data.frame")
    expect_s3_class(results$rejection.criteria, "data.frame")

   expect_equal(round(sum(results$data[1:2, 1:4]), 0), 7583)
   expect_equal(round(sum(results$rejection.criteria$Value), 2), 4016.66)
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 838
  object.mod <- object
  object.mod@records <- lapply(object.mod@records, function(x) {
    x@recordType <- paste(x@recordType, "(UVVIS)")
    x
  })
  expect_s4_class(analyse_pIRIRSequence(object.mod,
                                        signal_integral = 1:10,
                                        background_integral = 900:1000,
                                        plot = FALSE,
                                        verbose = FALSE),
               "RLum.Results")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)
  SW({
    vdiffr::expect_doppelganger("default",
                                analyse_pIRIRSequence(
                                  object,
                                  signal_integral = 1:2,
                                  background_integral = 900:1000,
                                  fit.method = "EXP",
                                  sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
                                  main = "Pseudo pIRIR data set based on quartz OSL",
                                  plot = TRUE,
                                  plot_singlePanels = FALSE,
                                  verbose = FALSE))
    vdiffr::expect_doppelganger("different limits",
                                analyse_pIRIRSequence(
                                    object,
                                    signal_integral = list(1:2, 2:3),
                                    background_integral = list(900:1000, 800:1000),
                                    fit.method = "EXP",
                                    sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
                                    plot = TRUE,
                                    plot_singlePanels = FALSE,
                                    verbose = FALSE))
  })
})

#' Function to get the right answer
#'
#' This function returns just the right answer
#'
#' @param ... you can pass an infinite number of further arguments
#'
#' @return Returns the right answer
#'
#' @section Function version: 0.1.0
#'
#' @author inspired by R.G.
#'
#' @examples
#'
#' ## you really want to know?
#' get_rightAnswer()
#'
#' @export
get_rightAnswer <- function(...) {
  return(46)
}

#' @title Export RLum-objects to CSV
#'
#' @description This function exports [Luminescence::RLum-class]-objects to CSV-files using the R function
#' [utils::write.table]. All [Luminescence::RLum-class]-objects are supported, but the
#' export is lossy, i.e. the pure numerical values are exported only. Information
#' that cannot be coerced to a [data.frame] or a [matrix] are discarded as well as
#' metadata.
#'
#' @details However, in combination with the implemented import functions, nearly every
#' supported import data format can be exported to CSV-files, this gives a great
#' deal of freedom in terms of compatibility with other tools.
#'
#' **Input is a list of objects**
#'
#' If the input is a [list] of objects all explicit function arguments can be provided
#' as [list].
#'
#' @param object [Luminescence::RLum-class] or a [list] of `RLum` objects (**required**):
#' objects to be written. Can be a [data.frame] if needed internally.
#'
#' @param path [character] (*optional*):
#' character string naming folder for the output to be written. If nothing
#' is provided `path` will be set to the working directory.
#' **Note:** this argument is ignored when `export = FALSE`.
#'
#' @param prefix [character] (*with default*):
#' optional prefix to name the files. This prefix is valid for all written files
#'
#' @param export [logical] (*with default*):
#' enable/disable the file export. If set to `FALSE` nothing is written to
#' the file connection, but a list comprising objects of type [data.frame] and [matrix]
#' is returned instead.
#'
#' @param compact [logical] (*with default*): if `TRUE` (default) the output
#' will be simpler but less comprehensive, that is not all elements in the
#' objects will be fully broken down. This is in particular useful for writing
#' [Luminescence::RLum.Results-class] objects to CSV files, as such objects can be rather complex
#' and not all information are needed in a CSV file or can be meaningfully
#' translated to CSV format.
#'
#' @param ... further arguments that will be passed to the function
#' [utils::write.table]. All arguments except the argument `file` are supported
#'
#' @return
#' The function returns either a CSV-file (or many of them) or for the
#' option `export == FALSE` a list comprising objects of type [data.frame] and [matrix]
#'
#' @section Function version: 0.2.2
#'
#' @author
#' Sebastian Kreutzer, Geography & Earth Science, Aberystwyth University (United Kingdom)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data-class], [Luminescence::RLum.Results-class],
#' [utils::write.table]
#'
#' @keywords IO
#'
#' @examples
#'
#' ## transform values to a list (and do not write)
#' data(ExampleData.BINfileData, envir = environment())
#' object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)[[1]]
#' write_RLum2CSV(object, export = FALSE)
#'
#' \dontrun{
#' ## write to a temporary file
#' temp_file <- tempfile(pattern = "output", fileext = ".csv")
#' write_RLum2CSV(object, temp_file)
#'
#' ## write to the working directory
#' write_RLum2CSV(object)
#' }
#'
#' @export
write_RLum2CSV <- function(
  object,
  path = NULL,
  prefix = "",
  export = TRUE,
  compact = TRUE,
  ...
) {
  .set_function_name("write_RLum2CSV")
  on.exit(.unset_function_name(), add = TRUE)

  ## support using `file` for the output path to help RLumShiny
  if ("file" %in% ...names() && is.null(path)) {
    path <- list(...)$file
  }

  # Self-call -----------------------------------------------------------------------------------
  ##this option allows to work on a list of RLum-objects
  if (inherits(object, "list")) {
    ## expand input arguments
    rep.length <- length(object)
    path <- .listify(path, rep.length)
    export <- .listify(export, rep.length)

      ##prefix ... create automatic prefix if nothing is provided
      prefix <- as.list(paste0(prefix[1], "[[",1:length(object),"]]_"))

      ## write list name to object
      for (i in seq_along(object))
        attr(object[[i]], "list_name") <- names(object)[i]

    ##execute the self-call function
      temp <- lapply(seq_along(object), function(x) {
        write_RLum2CSV(
          object = object[[x]],
          path = path[[x]],
          prefix = prefix[[x]],
          export = export[[x]],
          ...
        )
      })

    ## this prevents that we get a list of NULL
    if (is.null(unlist(temp)))
      return(NULL)
    return(temp)
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("RLum", "data.frame"),
                  extra = "a 'list' of such objects")
  .validate_not_empty(object)
  .validate_class(prefix, "character")
  .validate_logical_scalar(export)
  .validate_logical_scalar(compact)

  ## check export path
  if (export) {
    .validate_class(path, "character", null.ok = TRUE, length = 1)
    if (is.null(path) || !nzchar(path)) {
      path <- getwd()
      .throw_message("Path automatically set to: ", path, error = FALSE)
    } else if (!dir.exists(path)) {
      .throw_error("Path '", path, "' does not exist")
    }
  }

  ## What do we need at the end of the day is a named list of data.frames or matrices we can export
  ## using the function write.table; the name of the list elements will become the file names

  ## extract all elements ... depending on the input
  if (inherits(object, "RLum.Analysis")) {
        object_list <- lapply(object, get_RLum)

        ##change names of the list and produce the right format straight away
        names(object_list) <- paste0(1:length(object_list),"_",names(object))

  } else if (inherits(object, "RLum.Data")) {
        ##get object and make list
        object_list <- list(get_RLum(object))

        ##set new name
        names(object_list) <- paste0("1_",object@recordType)

  } else if (inherits(object, "RLum.Results")) {
      ##unlist what ever comes, but do not break structures like matrices, numerics and
      names <- names(object@data)

      ##get elements
      object_list <- lapply(object@data, function(e){
        ##only run something on the list of it is worth it and pack it in the list
        if (inherits(e, c("matrix", "numeric", "data.frame")))
          return(list(e))

        ##unlist the rest until the end
        if(!compact)
          e <- unlist(e)

        ##now we return whatever we have
        return(e)
      })

      ##now unlist again one level
      object_list <- unlist(object_list, recursive = FALSE)

      ##sort out objects we do not like and we cannot procede ...
      object_list_rm <- vapply(object_list, function(x) {
         inherits(x, c("matrix", "numeric", "data.frame"))
      }, FUN.VALUE = logical(1))

      ##remove unwanted objects
      object_list <- object_list[object_list_rm]

      ##set warning
      if (!all(object_list_rm))
        .throw_warning(length(which(!object_list_rm)),
                       " elements could not be converted to CSV")

      if (length(object_list) == 0) {
        .throw_error("No valid records in 'object'")
      }

      ##adjust the names
      names(object_list) <- paste0(1:length(object_list),"_",names(object_list))

  } else if (inherits(object, "data.frame")) {
    object_list <- list(object)
    filename <- attr(object, "filename") %||% ""
    names(object_list) <- paste0("conv_", attr(object, "list_name"), filename)
  }

  # Export --------------------------------------------------------------------------------------
  if (!export) {
    return(object_list)
  }

    ##set export settings for write.table
    export_settings.default <- list(
      append = FALSE,
      quote = TRUE,
      sep = ";",
      eol = "\n",
      na = "NA",
      dec = ".",
      row.names = FALSE,
      col.names = FALSE,
      qmethod = c("escape", "double"),
      fileEncoding = ""
    )

    ##modify on demand
    export_settings <- modifyList(x = export_settings.default, val = list(...))

    ##write files to file system
    for(i in 1:length(object_list)){
      utils::write.table(
        x = object_list[[i]],
        file = paste0(path,"/", prefix, names(object_list)[i],".csv"),
        append = export_settings$append,
        quote =  export_settings$quote,
        sep =  export_settings$sep,
        eol =  export_settings$eol,
        na =  export_settings$na,
        dec =  export_settings$dec,
        row.names =  export_settings$row.names,
        col.names =  export_settings$col.names,
        qmethod =  export_settings$qmethod,
        fileEncoding =  export_settings$fileEncoding)
    }
}

## load data
data(ExampleData.portableOSL, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(write_RLum2CSV(object = "", export = FALSE),
               "'object' should be of class 'RLum', 'data.frame' or a 'list'")
  expect_error(write_RLum2CSV(ExampleData.portableOSL[[1]], export = TRUE,
                              path = "non-existing"),
               "Path 'non-existing' does not exist")
  expect_error(write_RLum2CSV(ExampleData.portableOSL[[1]], export = TRUE,
                              file = "non-existing/filename.csv"),
               "Path 'non-existing/filename.csv' does not exist")
  expect_error(write_RLum2CSV(set_RLum("RLum.Results"), verbose = FALSE),
               "'object' cannot be an empty RLum.Results")
  expect_error(write_RLum2CSV(list(set_RLum("RLum.Data.Spectrum")), path = character()),
               "'path' should be of class 'character' or NULL and have length 1")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## move temporarily to avoid polluting the working directory
  cwd <- setwd(tempdir())
  expect_message(expect_null(
      write_RLum2CSV(ExampleData.portableOSL[[1]], export = TRUE, path = "")),
      "Path automatically set to")
  expect_message(
      expect_null(write_RLum2CSV(ExampleData.portableOSL[[1]], export = TRUE)),
      "Path automatically set to")
  expect_type(write_RLum2CSV(list(set_RLum("RLum.Data.Spectrum")),
                             path = character(), export = FALSE),
              "list")
  expect_type(write_RLum2CSV(ExampleData.portableOSL, export = FALSE), "list")
  setwd(cwd)

  data("ExampleData.RLum.Data.Image", envir = environment())
  write_RLum2CSV(ExampleData.RLum.Data.Image, path = tempdir())

  ##test RLum.Results objects
  ## load example data
  data(ExampleData.DeValues, envir = environment())
  SW({
  results <-  calc_CommonDose(ExampleData.DeValues$CA1)
  })

  ##using option compact
  expect_warning(write_RLum2CSV(object = results,export = FALSE),
                 "elements could not be converted to CSV")

  ##using option compact = FALSE
  expect_warning(write_RLum2CSV(object = results, export = FALSE,
                                compact = FALSE),
                 "elements could not be converted to CSV")
  expect_warning(write_RLum2CSV(object = results,export = FALSE, compact = TRUE),
                 "elements could not be converted to CSV")

  ## no valid records
  res.invalid <- results
  res.invalid@data$summary <- res.invalid@data$data <- NULL
  res.invalid@data$args$sigmab <- NULL
  expect_warning(expect_error(write_RLum2CSV(res.invalid, export = FALSE)),
                 "elements could not be converted to CSV")

  ##real export
  expect_warning(
    write_RLum2CSV(object = results, path = tempdir(), compact = TRUE),
    "elements could not be converted to CSV")

  ## data.frame
  df <- results@data$data
  expect_null(write_RLum2CSV(object = df, path = tempdir()))
  attr(df, "filename") <- "test"
  expect_null(write_RLum2CSV(object = df, path = tempdir()))

  ## empty list
  expect_null(write_RLum2CSV(list()))
})

#' @title Visualise natural/regenerated signal ratios
#'
#' @description
#' This function creates a Natural/Regenerated signal vs. time (NR(t)) plot
#' as shown in Steffen et al. 2009.
#'
#' This function accepts the individual curve data in many different formats. If
#' `data` is a `list`, each element of the list must contain a two
#' column `data.frame` or `matrix` containing the `XY` data of the curves
#' (time and counts). Alternatively, the elements can be objects of class
#' [Luminescence::RLum.Data.Curve-class].
#'
#' Input values can also be provided as a `data.frame` or `matrix` where
#' the first column contains the time values and each following column contains
#' the counts of each curve.
#'
#' @param data [list], [data.frame], [matrix] or [Luminescence::RLum.Analysis-class] (**required**):
#' X,Y data of measured values (time and counts). See details on individual data structure.
#'
#' @param log [character] (*with default*):
#' logarithmic axes (`"x"`, `"y"`, `"xy"`) or `""` (default) for linear axes.
#'
#' @param smooth [character] (*with default*):
#' apply data smoothing. If `"none"` (default), no data smoothing is applied.
#' Use `"rmean"` to calculate the rolling mean, where `k` determines the
#' width of the rolling window (see [data.table::frollmean]). `"spline"`
#' applies a smoothing spline to each curve (see [stats::smooth.spline])
#'
#' @param k [integer] (*with default*):
#' integer width of the rolling window.
#'
#' @param legend [logical] (*with default*):
#' enable/disable the plot legend.
#'
#' @param legend.pos [character] (*with default*):
#' keyword specifying the position of the legend (see [legend]).
#'
#' @param ... further parameters passed to [plot] (also see [par]).
#'
#'
#' @author Christoph Burow, University of Cologne (Germany)
#'
#' @seealso [plot]
#'
#' @return Returns a plot and [Luminescence::RLum.Analysis-class] object.
#'
#' @references
#' Steffen, D., Preusser, F., Schlunegger, F., 2009. OSL quartz underestimation due to
#' unstable signal components. Quaternary Geochronology, 4, 353-362.
#'
#' @examples
#'
#' ## load example data
#' data("ExampleData.BINfileData", envir = environment())
#'
#' ## EXAMPLE 1
#'
#' ## convert Risoe.BINfileData object to RLum.Analysis object
#' data <- Risoe.BINfileData2RLum.Analysis(object = CWOSL.SAR.Data, pos = 8, ltype = "OSL")
#'
#' ## extract all OSL curves
#' allCurves <- get_RLum(data)
#'
#' ## keep only the natural and regenerated signal curves
#' pos <- seq(1, 9, 2)
#' curves <- allCurves[pos]
#'
#' ## plot a standard NR(t) plot
#' plot_NRt(curves)
#'
#' ## re-plot with rolling mean data smoothing
#' plot_NRt(curves, smooth = "rmean", k = 10)
#'
#' ## re-plot with a logarithmic x-axis
#' plot_NRt(curves, log = "x", smooth = "rmean", k = 5)
#'
#' ## re-plot with custom axes ranges
#' plot_NRt(curves, smooth = "rmean", k = 5,
#'          xlim = c(0.1, 5), ylim = c(0.4, 1.6),
#'          legend.pos = "bottomleft")
#'
#' ## re-plot with smoothing spline on log scale
#' plot_NRt(curves, smooth = "spline", log = "x",
#'          legend.pos = "top")
#'
#' ## EXAMPLE 2
#'
#' # you may also use this function to check whether all
#' # TD curves follow the same shape (making it a TnTx(t) plot).
#' posTD <- seq(2, 14, 2)
#' curves <- allCurves[posTD]
#'
#' plot_NRt(curves, main = "TnTx(t) Plot",
#'          smooth = "rmean", k = 20,
#'          ylab = "TD natural / TD regenerated",
#'          xlim = c(0, 20), legend = FALSE)
#'
#' ## EXAMPLE 3
#'
#' # extract data from all positions
#' data <- lapply(1:24, FUN = function(pos) {
#'    Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = pos, ltype = "OSL")
#' })
#'
#' # get individual curve data from each aliquot
#' aliquot <- lapply(data, get_RLum)
#'
#' # set graphical parameters
#' par(mfrow = c(2, 2))
#'
#' # create NR(t) plots for all aliquots
#' for (i in 1:length(aliquot)) {
#'    plot_NRt(aliquot[[i]][pos],
#'             main = paste0("Aliquot #", i),
#'             smooth = "rmean", k = 20,
#'             xlim = c(0, 10),
#'             cex = 0.6, legend.pos = "bottomleft")
#' }
#'
#' # reset graphical parameters
#' par(mfrow = c(1, 1))
#'
#' @export
plot_NRt <- function(data, log = "", smooth = c("none", "spline", "rmean"), k = 3,
                     legend = TRUE, legend.pos = "topright", ...) {
  .set_function_name("plot_NRt")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("list", "data.frame", "matrix", "RLum.Analysis"))
  .validate_not_empty(data)
  if (inherits(data, "list")) {
    if (length(data) < 2)
      .throw_error("'data' contains only curve data for the natural signal")
    class.data <- unique(sapply(data, function(x) class(x)[1]))
    if (length(class.data) > 1) {
      .throw_error("'data' contains elements of different types: ",
                   .collapse(class.data))
    }
    if (class.data %in% c("RLum.Data.Curve", "RLum.Analysis"))
      curves <- lapply(data, get_RLum)
    else if (class.data %in% c("data.frame", "matrix"))
      curves <- data
    else
      .throw_error("'data' doesn't contain the expected type of elements")
  }
  else if (inherits(data, c("data.frame", "matrix"))) {
    if (ncol(data) < 3)
      .throw_error("'data' contains only curve data for the natural signal")
    if (is.matrix(data))
      data <- as.data.frame(data)
    if (!all(sapply(data, class) %in% c("numeric", "integer"))) {
      .throw_error("'data' contains non-numerical columns")
    }
    curves <- apply(data[2:ncol(data)], MARGIN = 2, function(curve) {
      data.frame(data[ ,1], curve)
    })
  }
  else if (inherits(data, "RLum.Analysis")) {
    RLum.objects <- get_RLum(data, drop = FALSE)
    if (any(sapply(RLum.objects, class) != "RLum.Data.Curve"))
      .throw_error("The provided 'RLum.Analysis' object ",
                   "must exclusively contain 'RLum.Data.Curve' objects")
    curves <- lapply(RLum.objects, get_RLum)
    if (length(curves) < 2)
      .throw_error("'data' contains only curve data for the natural signal")
  }
  .validate_class(log, "character", length = 1)

  smooth <- .validate_args(smooth, c("none", "spline", "rmean"))

  ## BASIC SETTINGS ------
  natural <- curves[[1]]
  regCurves <- curves[2:length(curves)]
  time <- curves[[1]][ ,1]

  if (anyNA(time)) {
    .throw_error("'data' contains missing values in the time column, ",
                 "check your data")
  }
  if (anyNA(natural[, 2])) {
    .throw_error("'data' contains missing values in the natural signal, ",
                 "check your data")
  }
  if (anyNA(regCurves[[1]][, 2])) {
    .throw_error("'data' contains missing values in the regenerated signal, ",
                 "check your data")
  }
  if (any(sapply(regCurves, nrow) != nrow(natural))) {
    .throw_error("The time values for the natural signal don't match ",
                 "those for the regenerated signal")
  }

  ## DATA TRANSFORMATION -----

  # calculate ratios
  NR <- lapply(regCurves, natural, FUN = function(reg, nat) {
    ratio <- nat[, 2] / reg[, 2]

    ## avoid infinities and NaNs
    ratio[is.infinite(ratio)] <- NA
    ratio[nat[, 2] == 0] <- 0
    ratio
  })

  # smooth spline
  if (smooth[1] == "spline") {
    NR <- lapply(NR, function(nr) { stats::smooth.spline(nr)$y })
  }
  if (smooth[1] == "rmean") {
    ## here we'd like to use the smoothed values with no fill: as .smoothing()
    ## relies on data.table::frollmean(), the only way to remove the fill
    ## is by setting `fill = NA` (which is already the default) and then
    ## omitting the NA values
    NR <- lapply(NR, function(nr) stats::na.omit(.smoothing(nr, k)))
    time <- stats::na.omit(.smoothing(time, k))
  }

  # normalise data
  NRnorm <- lapply(NR, FUN = function(nr) { nr / nr[1] })


  ## EXTRA ARGUMENTS -----

  # default values
  settings <- list(
    xlim = if (log == "x" || log ==  "xy") c(0.1, max(time)) else c(0, max(time)),
    ylim = range(pretty(c(sapply(NRnorm, range, na.rm = TRUE)))),
    xlab = "Time [s]",
    ylab = "Natural signal / Regenerated signal",
    cex = 1L,
    main = "NR(t) Plot")

  # override defaults with user settings
  settings <- modifyList(settings, list(...))

  ## PLOTTING ----------
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  # set graphical parameter
  par(cex = settings$cex)

  # empty plot
  if (is.na(pmatch(log, c("x", "y", "xy"))))
    log <- ""

  do.call(plot, modifyList(list(x = NA, y = NA, log = log, xaxs = "i", yaxs = "i"),
                           settings))

  # horizontal line
  abline(h = 1, lty = 3, col = "grey")

  col <- 1:length(NRnorm)

  # add N/R lines
  mapply(FUN = function(curve, col) {
    points(time, curve, type = "l", col = col)
  }, NRnorm, col)

  # add legend
  if (legend) {
    labels <- paste0("N/R", 1:length(NRnorm))
    ncol <- ifelse(length(NRnorm) > 4, ceiling(length(NRnorm) / 4) , 1)
    legend(legend.pos, legend = labels, col = col, lty = 1, ncol = ncol, cex = 0.8, bty = "n")
  }

  ## RETURN VALUES ----
  obj <- set_RLum("RLum.Analysis", protocol = "UNKNOWN",
                  records = mapply(FUN = function(curve, id) {
                    set_RLum("RLum.Data.Curve",
                             recordType = paste0("N/R", id),
                             curveType = "NRt",
                             data = matrix(c(time, curve), ncol = 2),
                             info = list(
                               data = curves,
                               call = sys.call(-6L),
                               args = as.list(sys.call(-6L)[-1])
                             ))
                  }, NRnorm, seq_len(length(NRnorm)))
  )
  invisible(obj)
}

test_that("test pure success of the plotting without warning or error", {
  testthat::skip_on_cran()
  local_edition(3)

  ##distribution plots
  data(ExampleData.DeValues, envir = environment())
  ExampleData.DeValues <- ExampleData.DeValues$CA1

  expect_silent(plot_RadialPlot(ExampleData.DeValues))
  expect_silent(plot_KDE(ExampleData.DeValues))
  expect_silent(plot_Histogram(ExampleData.DeValues))
  expect_silent(plot_ViolinPlot(ExampleData.DeValues))


  ##plot NRT
  data("ExampleData.BINfileData", envir = environment())
  data <- Risoe.BINfileData2RLum.Analysis(object = CWOSL.SAR.Data, pos = 8, ltype = "OSL")
  allCurves <- get_RLum(data)
  pos <- seq(1, 9, 2)
  curves <- allCurves[pos]
  expect_silent(plot_NRt(curves))

  ##filter combinations
  filter1 <- density(rnorm(100, mean = 450, sd = 20))
  filter1 <- matrix(c(filter1$x, filter1$y/max(filter1$y)), ncol = 2)
  filter2 <- matrix(c(200:799,rep(c(0,0.8,0),each = 200)), ncol = 2)
  expect_silent(plot_FilterCombinations(filters = list(filter1, filter2)))

   ##plot_Det
  data(ExampleData.BINfileData, envir = environment())
  object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
  expect_s4_class(
    plot_DetPlot(
      object,
      signal.integral.min = 1,
      signal.integral.max = 3,
      background.integral.min = 900,
      background.integral.max = 1000,
      n.channels = 5,
    ),
    "RLum.Results"
  )

  ##plot DRT
  data(ExampleData.DeValues, envir = environment())
  expect_silent(plot_DRTResults(values = ExampleData.DeValues$BT998[7:11,],
                  given.dose = 2800, mtext = "Example data"))


  ##plot RisoeBINFileData
  data(ExampleData.BINfileData, envir = environment())
  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data,position = 1))

  ##various RLum plots

    ##RLum.Data.Curve
    data(ExampleData.CW_OSL_Curve, envir = environment())
    temp <- as(ExampleData.CW_OSL_Curve, "RLum.Data.Curve")
    expect_silent(plot(temp))

    ##RLum.Data.Spectrum -------
    data(ExampleData.XSYG, envir = environment())
    expect_silent(plot(TL.Spectrum,
                            plot.type="contour",
                            xlim = c(310,750),
                            ylim = c(0,300)))

    expect_silent(suppressWarnings(plot_RLum.Data.Spectrum(TL.Spectrum,
                            plot.type="persp",
                            xlim = c(310,750),
                            ylim = c(0,100),
                            bin.rows=10,
                            bin.cols = 1)))

   expect_silent(suppressWarnings(plot_RLum.Data.Spectrum(TL.Spectrum,
                            plot.type="multiple.lines",
                            xlim = c(310,750),
                            ylim = c(0,100),
                            bin.rows=10,
                            bin.cols = 1)))

   expect_silent(suppressWarnings(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
                           xlim = c(310,750), ylim = c(0,300), bin.rows=10,
                           bin.cols = 1)))


   expect_silent(suppressWarnings(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
                           xlim = c(310,750), ylim = c(0,300), bin.rows=10,
                           bin.cols = 1,
                           type = "heatmap",
                           showscale = TRUE)))

   expect_silent(suppressWarnings(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
                                                          xlim = c(310,750), ylim = c(0,300), bin.rows=10,
                                                          bin.cols = 1,
                                                          type = "contour",
                                                          showscale = TRUE)))

   expect_error(plot(TL.Spectrum,
                      plot.type="contour",
                      xlim = c(310,750),
                      ylim = c(0,300), bin.cols = 0))


    ##RLum.Analysis
    data(ExampleData.BINfileData, envir = environment())
    temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
    expect_silent(plot(
      temp,
      subset = list(recordType = "TL"),
      combine = TRUE,
      norm = TRUE,
      abline = list(v = c(110))
    ))

    ##RLum.Results
    grains<- calc_AliquotSize(grain.size = c(100,150), sample.diameter = 1, plot = FALSE, MC.iter = 100)
    expect_silent(plot_RLum.Results(grains))

    ##special plot RLum.Reuslts
    data(ExampleData.DeValues, envir = environment())
    mam <- calc_MinDose(data = ExampleData.DeValues$CA1, sigmab = 0.2, log = TRUE, plot = FALSE)
    expect_silent(plot_RLum(mam))
    cdm <- calc_CentralDose(ExampleData.DeValues$CA1)
    expect_silent(plot_RLum(cdm))
    FMM <- calc_FiniteMixture(ExampleData.DeValues$CA1,
                             sigmab = 0.2, n.components = c(2:4),
                             pdf.weight = TRUE, dose.scale = c(0, 100))
    plot_RLum(FMM)




})

test_that("test for return values, if any", {
  testthat::skip_on_cran()
  local_edition(3)

  data(ExampleData.DeValues, envir = environment())
  output <- plot_AbanicoPlot(ExampleData.DeValues, output = TRUE)
    expect_type(output, "list")
    expect_length(output, 10)
})

## load data
data("ExampleData.BINfileData", envir = environment())
obj <- Risoe.BINfileData2RLum.Analysis(object = CWOSL.SAR.Data,
                                       pos = 8, ltype = "OSL")
curves <- get_RLum(obj)[seq(1, 9, 2)]

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_NRt("error"),
               "'data' should be of class 'list', 'data.frame', 'matrix' or")
  expect_error(plot_NRt(obj[[2]]),
               "'data' should be of class 'list', 'data.frame', 'matrix' or")
  expect_error(plot_NRt(iris[0, ]),
               "'data' cannot be an empty data.frame")
  expect_error(plot_NRt(curves[1]),
                "'data' contains only curve data for the natural signal")
  expect_error(plot_NRt(curves[[1]]@data),
                "'data' contains only curve data for the natural signal")
  expect_error(plot_NRt(iris),
                "'data' contains non-numerical columns")
  expect_error(plot_NRt(data.frame(a = NA, b = 1:5, c = 1:5)),
                "'data' contains non-numerical columns")
  expect_error(plot_NRt(list(a = 1, b = iris)),
                "'data' contains elements of different types: 'numeric', 'data.frame'")
  expect_error(plot_NRt(list(a = 1, b = 2)),
                "'data' doesn't contain the expected type of elements")
  expect_error(plot_NRt(data.frame(a = c(1:4, NA), b = 1:5, c = 1:5)),
                "'data' contains missing values in the time column")
  expect_error(plot_NRt(data.frame(a = 1:5, b = c(1:4, NA), c = 1:5)),
                "'data' contains missing values in the natural signal")
  expect_error(plot_NRt(data.frame(a = 1:5, b = 1:5, c = c(NA, 2:5))),
                "'data' contains missing values in the regenerated signal")
  expect_error(plot_NRt(curves, log = FALSE),
               "'log' should be of class 'character' and have length 1")
  expect_error(plot_NRt(curves, smooth = "error"),
               "'smooth' should be one of 'none', 'spline' or 'rmean'")

  data(ExampleData.XSYG, envir = environment())
  obj.mixed <- merge_RLum.Analysis(list(obj, TL.Spectrum))
  expect_error(plot_NRt(obj.mixed),
               "The provided 'RLum.Analysis' object must exclusively contain")
  expect_error(plot_NRt(set_RLum("RLum.Analysis",
                                 records = list(curves[[1]]))),
               "'data' contains only curve data for the natural signal")

  data("ExampleData.RLum.Analysis", envir = environment())
  expect_error(plot_NRt(IRSAR.RF.Data),
               "The time values for the natural signal don't match those for")
  expect_error(plot_NRt(merge_RLum.Analysis(list(obj, IRSAR.RF.Data))),
               "The time values for the natural signal don't match those for")

  data(ExampleData.CW_OSL_Curve, envir = environment())
  expect_error(plot_NRt(ExampleData.CW_OSL_Curve),
               "'data' contains only curve data for the natural signal")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## list
  expect_silent(plot_NRt(curves))
  expect_silent(plot_NRt(list(get_RLum(curves[[1]]),
                              get_RLum(curves[[2]]))))
  expect_silent(plot_NRt(curves, smooth = "spline", log = "x"))

  small <- curves[1:3]
  for (idx in 1:length(small))
    small[[idx]]@data <- small[[idx]]@data[1:175, ]
  expect_snapshot_RLum(plot_NRt(small, smooth = "rmean", k = 10))

  ## matrix
  plot_NRt(cbind(curves[[1]]@data, curves[[1]]@data))

  ## RLum.Analysis
  expect_silent(plot_NRt(obj))

  ## list of RLum.Analysis
  bin.v8 <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  expect_output(plot_NRt(read_BIN2R(bin.v8, fastForward = TRUE, verbose = FALSE)),
                "BIN/BINX-file non-conform. TL curve may be wrong")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              plot_NRt(curves))
  vdiffr::expect_doppelganger("spline-log",
                              plot_NRt(curves, smooth = "spline", log = "x"))
  })
})

#' @title Fit and plot a dose-response curve for luminescence data (Lx/Tx against dose)
#'
#' @description
#'
#' A dose-response curve is produced for luminescence measurements using a
#' regenerative or additive protocol as implemented in [Luminescence::fit_DoseResponseCurve]
#' and [Luminescence::plot_DoseResponseCurve]
#'
#' @inheritParams fit_DoseResponseCurve
#'
#' @param output.plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param output.plotExtended [logical] (*with default*):
#' If `TRUE`, 3 plots on one plot area are provided:
#' 1. growth curve,
#' 2. histogram from Monte Carlo error simulation and
#' 3. a test dose response plot.
#'
#' If `FALSE`, just the growth curve will be plotted.
#'
#' @param plot_singlePanels [logical] (*with default*):
#' single plot output (`TRUE/FALSE`) to allow for plotting the results in
#' single plot windows. Requires `output.plotExtended = TRUE`.
#'
#' @param n.MC [integer] (*with default*):
#' number of MC runs for error calculation.
#'
#' @param ... Further arguments to [Luminescence::fit_DoseResponseCurve] (`fit_weights`,
#' `fit_bounds`, `fit.force_through_origin`, `fit.includingRepeatedRegPoints`,
#' `fit.NumberRegPoints`, `fit.NumberRegPointsReal`, `n.MC`,
#' `txtProgressBar`) and graphical parameters to be passed (supported:
#' `xlim`, `ylim`, `main`, `xlab`, `ylab`).
#'
#' @return
#' Along with a plot (if wanted) the [Luminescence::RLum.Results-class] object produced by
#' [Luminescence::fit_DoseResponseCurve] is returned.
#'
#' @section Function version: 1.2.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @references
#'
#' Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46.
#'
#' Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
#' Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
#' from feldspar. Radiation Measurements 81, 224-231.
#'
#' Pagonis, V., Kitis, G., Chen, R., 2020. A new analytical equation for the dose response of dosimetric materials,
#' based on the Lambert W function. Journal of Luminescence 225, 117333. \doi{10.1016/j.jlumin.2020.117333}
#'
#' @seealso [Luminescence::fit_DoseResponseCurve], [Luminescence::plot_DoseResponseCurve]
#'
#' @examples
#'
#' ##(1) plot growth curve for a dummy dataset
#' data(ExampleData.LxTxData, envir = environment())
#' plot_GrowthCurve(LxTxData)
#'
#' ##(1b) horizontal plot arrangement
#' layout(mat = matrix(c(1,1,2,3), ncol = 2))
#' plot_GrowthCurve(LxTxData, plot_singlePanels = TRUE)
#'
#' ##(2) plot the growth curve with pdf output - uncomment to use
#' ##pdf(file = "~/Desktop/Growth_Curve_Dummy.pdf", paper = "special")
#' plot_GrowthCurve(LxTxData)
#' ##dev.off()
#'
#' ##(3) plot the growth curve with pdf output - uncomment to use, single output
#' ##pdf(file = "~/Desktop/Growth_Curve_Dummy.pdf", paper = "special")
#' temp <- plot_GrowthCurve(LxTxData, plot_singlePanels = TRUE)
#' ##dev.off()
#'
#' ##(4) plot resulting function for given interval x
#' x <- seq(1,10000, by = 100)
#' plot(
#'  x = x,
#'  y = eval(temp$Formula),
#'  type = "l"
#' )
#'
#' ##(5) plot using the 'extrapolation' mode
#' LxTxData[1,2:3] <- c(0.5, 0.001)
#' print(plot_GrowthCurve(LxTxData, mode = "extrapolation"))
#'
#' ##(6) plot using the 'alternate' mode
#' LxTxData[1,2:3] <- c(0.5, 0.001)
#' print(plot_GrowthCurve(LxTxData, mode = "alternate"))
#'
#' @export
plot_GrowthCurve <- function(
  object,
  mode = "interpolation",
  fit.method = "EXP",
  output.plot = TRUE,
  output.plotExtended = TRUE,
  plot_singlePanels = FALSE,
  verbose = TRUE,
  n.MC = 100,
  ...
) {
  .set_function_name("plot_GrowthCurve")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  extraArgs <- list(...)
  if ("output.plotExtended.single" %in% names(extraArgs)) {
    plot_singlePanels <- extraArgs$output.plotExtended.single
    .deprecated("output.plotExtended.single", "plot_singlePanels", since = "1.0.0")
  }
  if ("NumberIterations.MC" %in% names(extraArgs)) {
    n.MC <- extraArgs$NumberIterations.MC
    .deprecated("NumberIterations.MC", "n.MC", since = "1.0.0")
  }
  if ("sample" %in% names(extraArgs)) {
    object <- extraArgs$sample
    .deprecated(old = "sample", new = "object", since = "1.2.0")
  }

  ## input validation
  .validate_logical_scalar(output.plot)
  .validate_logical_scalar(output.plotExtended)
  .validate_logical_scalar(plot_singlePanels)
  .validate_logical_scalar(verbose)

  ## remaining input validation occurs inside the fitting function
  fit <- fit_DoseResponseCurve(object, mode, fit.method,
                               verbose = verbose, n.MC = n.MC, ...)

  if (is.null(fit)) {
    if (verbose)
      .throw_message("Fitting failed, NULL returned")
    return(NULL)
  }

  if (output.plot) {
    plot_DoseResponseCurve(fit, plot_extended = output.plotExtended,
                           plot_singlePanels = plot_singlePanels,
                           verbose = verbose, ...)
  }

  invisible(fit)
}

## load data
data(ExampleData.LxTxData, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(
      plot_GrowthCurve("error"),
      "[fit_DoseResponseCurve()] 'object' should be of class 'data.frame'",
      fixed = TRUE)
  expect_error(
      plot_GrowthCurve(LxTxData, mode = "error"),
      "[fit_DoseResponseCurve()] 'mode' should be one of 'interpolation'",
      fixed = TRUE)
  expect_error(
      plot_GrowthCurve(LxTxData, fit.method = "error"),
      "'fit.method' should be one of 'LIN', 'QDR', 'EXP', 'EXP OR LIN'")
  expect_error(
      plot_GrowthCurve(LxTxData, output.plotExtended = "error"),
      "'output.plotExtended' should be a single logical value")
  expect_error(
      plot_GrowthCurve(LxTxData, plot_singlePanels = "error"),
      "'plot_singlePanels' should be a single logical value")
  expect_error(
      plot_GrowthCurve(LxTxData, verbose = "error"),
      "'verbose' should be a single logical value")

  ## Weird LxTx values
  LxTx <- structure(list(
    Dose = c(0, 250, 500, 750, 1000, 1500, 0, 500, 500),
    LxTx = c(1, Inf, 0, -Inf, Inf, 0, Inf, -0.25, 2),
    LxTx.Error = c(1.5813365, Inf, 0, Inf, Inf, 0, Inf, 1.4114626, 3.1626729)),
    class = "data.frame", row.names = c(NA, -9L))
  SW({
  expect_warning(
      plot_GrowthCurve(LxTx),
      "Inf values found, replaced by NA")
  })

  ## same dose for all points
  tmp_LxTx <- LxTxData
  tmp_LxTx$Dose <- 10
  SW({
  expect_message(expect_null(
      plot_GrowthCurve(tmp_LxTx)),
      "All points have the same dose, NULL returned")
  })

  ## only two columns
  expect_warning(
      plot_GrowthCurve(LxTxData[, 1:2], verbose = FALSE),
      "Error column invalid or 0, 'fit.weights' ignored")

  ## test case with all NA
  tmp_LxTx <- LxTxData
  tmp_LxTx$LxTx <- NA
  expect_message(expect_warning(expect_null(
      plot_GrowthCurve(tmp_LxTx, verbose = FALSE)),
      "7 NA values removed"),
      "Error: After NA removal, nothing is left from the data set")

  ## test case without TnTx column
  tmp_LxTx <- LxTxData
  tmp_LxTx$TnTx <- NULL
  expect_s4_class(
    plot_GrowthCurve(tmp_LxTx, verbose = FALSE),
    "RLum.Results")

  ## do not include reg point
  expect_s4_class(
    plot_GrowthCurve(
      LxTxData,
      verbose = FALSE,
      fit.includingRepeatedRegPoints = FALSE),
    class = "RLum.Results")

  ## deprecated option
  expect_warning(
      plot_GrowthCurve(LxTxData, verbose = FALSE,
                       output.plotExtended.single = TRUE),
      "'output.plotExtended.single' was deprecated in v1.0.0, use 'plot_singlePanels'")
  expect_warning(
      plot_GrowthCurve(sample = LxTxData, verbose = FALSE),
      "'sample' was deprecated in v1.2.0, use 'object' instead")
})

test_that("main tests", {
  testthat::skip_on_cran()

  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "LIN",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP+LIN",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP+EXP",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "QDR",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "GOK",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "OTOR",
                                 n.MC = 10))

  ## force through the origin
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP+LIN",
                                 fit.bounds = FALSE,
                                 fit.force_through_origin = TRUE,
                                 n.MC = 10))
  temp_LxTx <- LxTxData
  temp_LxTx$LxTx[[7]] <- 1
  expect_output(plot_GrowthCurve(temp_LxTx,
                                 fit.method = "GOK",
                                 fit.force_through_origin = TRUE,
                                 n.MC = 10))
})

test_that("additional tests", {
  testthat::skip_on_cran()

  LxTxData[1, 2:3] <- c(0.5, 0.001)

  ## Check extrapolation ----------------------------------------------------
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "LIN",
                                 mode = "extrapolation",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP",
                                 mode = "extrapolation",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP+LIN",
                                 mode = "extrapolation",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "GOK",
                                 mode = "extrapolation",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "OTOR",
                                 mode = "extrapolation",
                                 n.MC = 10))

  ## force through the origin
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "QDR",
                                 mode = "extrapolation",
                                 fit.force_through_origin = TRUE,
                                 n.MC = 10))

  ## Check alternate --------------------------------------------------------
  expect_silent(plot_GrowthCurve(LxTxData,
                                 fit.method = "QDR",
                                 mode = "alternate",
                                 verbose = FALSE,
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "LIN",
                                 mode = "alternate",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP",
                                 mode = "alternate",
                                 n.MC = 10))
  expect_silent(plot_GrowthCurve(LxTxData,
                                 fit.method = "EXP+LIN",
                                 mode = "alternate",
                                 verbose = FALSE,
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "GOK",
                                 mode = "alternate",
                                 n.MC = 10))
  expect_output(plot_GrowthCurve(LxTxData,
                                 fit.method = "OTOR",
                                 mode = "alternate",
                                 n.MC = 10))

  ## only two valid points provided
  SW({
  warnings <- capture_warnings(expect_message(plot_GrowthCurve(
    data.frame(
        dose = c(0, 1388.88888888889, NA),
        LxTx = c(1.54252220145258, 4.43951568403849, NA),
        LxTx_X = c(0.130074482379272, 2.59694106608, NA)),
    verbose = TRUE, NumberIterations.MC = 10),
    "'fit.method' changed to 'LIN'"))
  })
  expect_match(warnings, "1 NA values removed",
               all = FALSE, fixed = TRUE)
  expect_match(warnings, "Fitting a non-linear least-squares model requires",
               all = FALSE, fixed = TRUE)
  expect_match(warnings, "'NumberIterations.MC' was deprecated in v1.0.0, use 'n.MC'",
               all = FALSE, fixed = TRUE)
})

#' Channel binning - method dispatcher
#'
#' Function calls the object-specific bin functions for RLum.Data S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum.Data-class] objects. \cr
#' Depending on the input object, the corresponding function will be selected.
#' Allowed arguments can be found in the documentations of the corresponding
#' [RLum.Data-class] class.
#'
#' @param object [RLum.Data-class] (**required**):
#' S4 object of class `RLum.Data`
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return An object of the same type as the input object is provided
#'
#' @section Function version: 0.2.0
#'
#' @author Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#'
#' @note Currently only `RLum.Data` objects of class [RLum.Data.Curve-class] and [RLum.Data.Spectrum-class] are supported!
#'
#' @seealso [RLum.Data.Curve-class], [RLum.Data.Spectrum-class]
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ##create RLum.Data.Curve object from this example
#' curve <-
#'   set_RLum(
#'       class = "RLum.Data.Curve",
#'       recordType = "OSL",
#'       data = as.matrix(ExampleData.CW_OSL_Curve)
#'   )
#'
#' ##plot data without and with 2 and 4 channel binning
#' plot_RLum(curve)
#' plot_RLum(bin_RLum.Data(curve, bin_size = 2))
#' plot_RLum(bin_RLum.Data(curve, bin_size = 4))
#'
#' @keywords utilities
#'
#' @md
#' @export
setGeneric("bin_RLum.Data", function(object, ...) {
  standardGeneric("bin_RLum.Data")
})

#' @title Analyse IRSAR RF measurements
#'
#' @description
#' The function analyses IRSAR RF measurements on K-feldspar samples performed
#' using the protocol according to Erfurt et al. (2003) and beyond.
#'
#' @details The function performs an IRSAR analysis described for K-feldspar samples by
#' Erfurt et al. (2003) assuming a negligible sensitivity change of the RF
#' signal.
#'
#' **General Sequence Structure** (according to Erfurt et al., 2003)
#'
#' 1. Measuring IR-RF intensity of the natural dose for a few seconds (\eqn{RF_{nat}})
#' 2. Bleach the samples under solar conditions for at least 30 min without changing the geometry
#' 3. Waiting for at least one hour
#' 4. Regeneration of the IR-RF signal to at least the natural level (measuring (\eqn{RF_{reg}})
#' 5. Fitting data with a stretched exponential function
#' 6. Calculate the palaeodose \eqn{D_{e}} using the parameters from the fitting
#'
#' Three methods are supported to obtain the \eqn{D_{e}}:
#' `method = "FIT"`, `method = "SLIDE"` and `method = "VSLIDE"`:
#'
#' **`method = "FIT"`**
#'
#' The principle is described above and follows the original suggestions by
#' Erfurt et al., 2003. For the fitting, the mean count value of the `RF_nat`
#' curve is used.
#'
#' Function used for the fitting (according to Erfurt et al. (2003)):
#'
#' \deqn{\phi(D) = \phi_{0} - \Delta\phi(1 - \exp(-\lambda D))^\beta}
#'
#' with
#' \eqn{\phi(D)} the dose dependent IR-RF flux,
#' \eqn{\phi_{0}} the initial IR-RF flux,
#' \eqn{\Delta\phi} the dose dependent change of the IR-RF flux,
#' \eqn{\lambda} the exponential parameter, \eqn{D} the dose and
#' \eqn{\beta} the dispersive factor.
#'
#' To obtain the palaeodose
#' \eqn{D_{e}} the function is changed to:
#'
#' \deqn{D_{e} = \ln(-(\phi(D) - \phi_{0})/(-\lambda\phi)^{1/\beta}+1)/-\lambda}
#'
#' The fitting is done using the `port` algorithm of the [nls] function.
#'
#'
#' **`method = "SLIDE"`**
#'
#' For this method, the natural curve is slid along the x-axis until
#' congruence with the regenerated curve is reached. Unlike fitting, this
#' allows for working with the original data without the need for any physical
#' model. This approach was introduced for RF curves by Buylaert et al., 2012
#' and Lapp et al., 2012.
#'
#' Here the sliding is done by searching for the minimum of the sum of
#' squared residuals.
#' For the mathematical details of the implementation see Frouin et al., 2017
#'
#' **`method = "VSLIDE"`**
#'
#' Same as `"SLIDE"` but searching also vertically for the best match (i.e. in xy-direction.)
#' See Kreutzer et al. (2017) and Murari et al. (2021). By default, the vertical sliding
#' range is set automatically, but can be set manually by changing the
#' `vslide_range` parameter (see `method_control`).
#'
#' **`method_control`**
#'
#' To keep the generic argument list as clear as possible, parameters to control
#' the methods for De estimation are preset with meaningful default values,
#' which can however be modified using the `method_control` argument, e.g.,
#' `method_control = list(trace = TRUE)`. Supported parameters are:
#'
#' *For FIT*
#' - `trace` ([logical], default: `FALSE`): as in [nls]; shows sum of squared
#' residuals.
#' - `maxiter` ([integer], default: 500): as in [nls].
#' - `warnOnly` ([logical], default: `FALSE`): as in [nls].
#' - `minFactor` ([numeric], default: `1 / 4096`): as in [nls].
#'
#' *For SLIDE or VSLIDE*
#' - `trace` ([logical], default: `FALSE`): as in [nls]; shows sum of squared
#' residuals.
#' - `trace_vslide` ([logical], default: `FALSE`): enable/disable the tracing
#' of the vertical sliding.
#' - `correct_onset` ([logical], default: `TRUE`): whether the curves should
#' be shifted along the x-axis by the first channel, as light is expected in
#' the first channel.
#' - `show_density` ([logical], default: `TRUE`): enable/disable KDE plots for
#' MC run results. Nothing is shown if the distribution is too narrow.
#' - `show_fit` ([logical], default: `FALSE`): enable/disable the plot of the
#' fitted curve routinely obtained during the evaluation.
#' - `n.MC` ([integer], default: 1000): number of Monte Carlo runs within the
#' sliding (assessing the possible minimum values). **Note:** This parameter
#' is not the same as the function argument `n.MC`.
#' - `vslide_range` ([numeric] or [character], default: "auto"): boundaries
#' for the vertical curve sliding. The argument expects a vector with absolute
#' minimum and maximum (e.g., `c(-1000,1000)`). The default "auto" mode
#' detects the reasonable vertical sliding range (*recommended*). `NULL`
#' disables the vertical sliding.
#' - `num_slide_windows` ([integer], default: 3): number of  differently-sized
#' windows tested when sliding: the higher the value (up to a maximum of 10),
#' the more time is spent in searching the global optimum. The default setting
#' attempts to strike a balance between quality of the fit and computation
#' speed.
#' - `cores` ([numeric] or [character], default: `NULL`): number of cores
#' allocated for a parallel processing of the Monte-Carlo runs. The default
#' value corresponds to single-threaded computation; the recommended values is
#' `"auto"`, which assigns all but two of the available cores.
#'
#' **Error estimation**
#'
#' For **`method = "FIT"`**, the \eqn{D_{e}} error range is obtained by using
#' the 2.5 % (lower) and the 97.5 % (upper) quantiles of the \eqn{RF_{nat}}
#' curve.
#'
#' For **`method = "SLIDE"`** and **`method = "VSLIDE"`**, the error is obtained
#' by bootstrapping the residuals of the slid curve to construct new natural
#' curves for a Monte Carlo simulation. The error is returned in two ways:
#' (a) the standard deviation of the \eqn{D_{e}} obtained from the MC runs and
#' (b) the confidence interval using the 2.5 % (lower) and the 97.5 % (upper)
#' quantiles. The results of the MC runs are returned with the function output.
#'
#' **Test parameters**
#'
#' The argument `test_parameters` allows to pass thresholds for several test
#' parameters, which will be evaluated during the function run. If a threshold
#' is set and it is exceeded, the test parameter status will be set to
#' `"FAILED"`. This argument is intentionally not termed `'rejection criteria'`
#' as not all test parameters are evaluated for both methods and some parameters
#' are calculated but not evaluated by default.
#'
#' `NA` and `NULL` are the allowed values for all parameters. If the parameter
#' is set to `NA`, the value is calculated but the result will not be evaluated,
#' therefore it will have no effect on the status (`"OK"` or `"FAILED"`) of
#' the parameter. Setting the parameter to `NULL` disables the parameter
#' entirely and the parameter will also be removed from the function output.
#' This might be useful in cases where a particular parameter requires a long
#' computation time. Currently supported parameters are:
#'
#' - `curves_ratio` ([numeric], default: `1.001`):
#' the ratio of \eqn{RF_{nat}} to \eqn{RF_{reg}} is calculated over the range
#' spanned by \eqn{RF_{nat}}, and should not exceed the threshold value.
#'
#' - `intersection_ratio` ([numeric], default: `NA`):
#' calculated as absolute difference from 1 of the ratio of the integral of
#' the normalised RF-curves. This value indicates intersection of the RF-curves
#' and should be close to 0 if the curves have a similar shape. For this
#' calculation first the corresponding time-count pair value on the RF_reg
#' curve is obtained using the maximum count value of the `RF_nat` curve and
#' only this segment (fitting to the `RF_nat` curve) on the RF_reg curve is
#' taken for further calculating this ratio. If nothing is found at all, `Inf`
#' is returned.
#'
#' - `residuals_slope` ([numeric], default: `NA`; only for `method = "SLIDE"`
#' and `"VSLIDE`"):
#' a linear function is fitted on the residuals after sliding.
#' The corresponding slope can be used to discard values as a high (positive,
#' negative) slope may indicate that both curves are fundamentally different
#' and the method cannot be applied at all.
#' By default, the value of this parameter is calculated but not evaluated.
#'
#' - `curves_bounds` ([numeric], default: \eqn{\max(RF_{reg_counts})}):
#' this measure uses the maximum time (x) value of the regenerated curve.
#' The maximum time (x) value of the natural curve cannot be larger than this
#' value. However, although this is not recommended the value can be changed
#' or disabled.
#'
#' - `dynamic_ratio` ([numeric], default: `NA`):
#' the dynamic ratio of the regenerated curve is calculated as ratio of the
#' minimum and maximum count values.
#'
#' - `lambda`, `beta` and `delta.phi` [numeric] (default: `NA`):
#' the stretched exponential function suggested by Erfurt et al. (2003)
#' describing the decay of the RF signal, comprises several parameters that
#' might be useful to evaluate the shape of the curves. For `method = "FIT"`,
#' this parameter is obtained during the fitting; for `method = "SLIDE"` a
#' rather rough estimation is made using the function [minpack.lm::nlsLM] and
#' the equation given above.
#' **Note:** As this procedure requests more computation time, it is performed
#' only if all three parameters are set.
#'
#' @param object [Luminescence::RLum.Analysis-class] or a [list] of [Luminescence::RLum.Analysis-class]-objects (**required**):
#' input object containing data for protocol analysis. The function expects to
#' find at least two curves in the [Luminescence::RLum.Analysis-class] object: (1) `RF_nat`, (2) `RF_reg`.
#' If a `list` is provided as input all other parameters can be provided as
#' `list` as well to gain full control.
#'
#' @param sequence_structure [vector] [character] (*with default*):
#' specifies the general sequence structure. Allowed steps are `NATURAL` and
#' `REGENERATED`, and at least one of each must appear. In addition, any other
#' character is allowed in the sequence structure; such curves will be ignored
#' during the analysis. If `sequence_structure` is shorter than the number of
#' curves in `object`, it is recycled. If multiple `NATURAL` and `REGENERATED`
#' steps are specified, the corresponding measurements are stacked.
#'
#' @param RF_nat.lim [vector] (*with default*):
#' set minimum and maximum channel range for natural signal fitting and sliding.
#' If only one value is provided this will be treated as minimum value and the
#' maximum limit will be added automatically.
#'
#' @param RF_reg.lim [vector] (*with default*):
#' set minimum and maximum channel range for regenerated signal fitting and sliding.
#' If only one value is provided this will be treated as minimum value and the
#' maximum limit will be added automatically.
#'
#' @param method [character] (*with default*): select method applied for the
#' data analysis. Possible options are `"FIT"`, `"SLIDE"`, `"VSLIDE"`;
#' `"NONE"` can be used to disable the analysis and plot the natural points
#' at their original position.
#'
#' @param method_control [list] (*optional*):
#' parameters to control the method, that can be passed to the chosen method.
#' These are for (1) `method = "FIT"`: `'trace'`, `'maxiter'`, `'warnOnly'`, `'minFactor'` and for
#' (2) `method = "SLIDE"`: `'correct_onset'`, `'show_density'`,  `'show_fit'`, `'trace'`.
#' See details.
#'
#' @param test_parameters [list] (*with default*):
#' set test parameters. Supported parameters are: `curves_ratio`,
#' `residuals_slope` (only for `method = "SLIDE"` and `"VSLIDE"`),
#' `curves_bounds`, `dynamic_ratio`, `lambda`, `beta` and `delta.phi`.
#' All input: [numeric] values, `NA` and `NULL` (see Details for further
#' information).
#'
#' @param n.MC [numeric] (*with default*):
#' number of Monte Carlo runs for the estimation of the start parameter
#' (`method = "FIT"`) or of the error (`method = "SLIDE"` and `"VSLIDE"`).
#' This value can be set to `NULL` to skip the MC runs.
#' Note: Large values will significantly increase the computation time.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar during MC runs.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param plot_reduced [logical] (*optional*):
#' provides a reduced plot output if enabled to allow common R plot combinations,
#' e.g., `par(mfrow(...))`. If `TRUE` no residual plot
#' is returned; it has no effect if `plot = FALSE`
#'
#' @param ... further arguments that will be passed to the plot output.
#' Currently supported arguments are `main`, `mtext`, `xlab`, `ylab`,
#' `xlim`, `ylim`, `log`, `legend` (`TRUE/FALSE`),
#' `legend.pos`, `legend.text` (passes argument to x,y in
#' [graphics::legend]), `xaxt`, `verbose` (`TRUE/FALSE`).
#'
#' @return
#' The function returns numerical output and an (*optional*) plot.
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' `[.. $data : data.frame]`
#'
#' \tabular{lll}{
#'  **Column** \tab **Type** \tab **Description**\cr
#'  `DE` \tab `numeric` \tab the obtained equivalent dose\cr
#'  `DE.ERROR` \tab `numeric` \tab (only `method = "SLIDE"`) standard deviation obtained from MC runs \cr
#'  `DE.LOWER` \tab `numeric`\tab 2.5% quantile for De values obtained by MC runs \cr
#'  `DE.UPPER` \tab `numeric`\tab 97.5% quantile for De values obtained by MC runs  \cr
#'  `DE.STATUS`  \tab `character`\tab test parameter status\cr
#'  `RF_NAT.LIM`  \tab `character`\tab used `RF_nat` curve limits \cr
#'  `RF_REG.LIM` \tab `character`\tab used `RF_reg` curve limits\cr
#'  `POSITION` \tab `integer`\tab (*optional*) position of the curves\cr
#'  `DATE` \tab `character`\tab (*optional*) measurement date\cr
#'  `SEQUENCE_NAME` \tab `character`\tab (*optional*) sequence name\cr
#'  `UID` \tab `character`\tab unique data set ID
#' }
#'
#' `[.. $De.MC : numeric]`
#'
#' A `numeric` vector with all the De values obtained by the MC runs.
#'
#' `[.. $test_parameters : data.frame]`
#'
#' \tabular{lll}{
#'  **Column** \tab **Type** \tab **Description**\cr
#'  `POSITION` \tab `numeric` \tab aliquot position \cr
#'  `PARAMETER` \tab `character` \tab test parameter name \cr
#'  `THRESHOLD` \tab `numeric` \tab set test parameter threshold value \cr
#'  `VALUE` \tab `numeric` \tab the calculated test parameter value (to be compared with the threshold)\cr
#'  `STATUS` \tab `character` \tab test parameter status either `"OK"` or `"FAILED"` \cr
#'  `SEQUENCE_NAME` \tab `character` \tab name of the sequence, so far available \cr
#'  `UID` \tab `character`\tab unique data set ID
#' }
#'
#' `[.. $fit : data.frame]`
#'
#' An [nls] object produced by the fitting.
#'
#' `[.. $slide : list]`
#'
#' A [list] with data produced during the sliding. Some elements are previously
#' reported with the summary object data. List elements are:
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `De` \tab `numeric` \tab the final De obtained with the sliding approach \cr
#'  `De.MC` \tab `numeric` \tab all De values obtained by the MC runs \cr
#'  `residuals` \tab `numeric` \tab the obtained residuals for each channel of the curve \cr
#'  `trend.fit` \tab `lm` \tab fitting results produced by the fitting of the residuals \cr
#'  `RF_nat.slid` \tab `matrix` \tab the slid `RF_nat` curve \cr
#'  `t_n.id` \tab `numeric` \tab the index of the t_n offset \cr
#'  `I_n` \tab `numeric` \tab the vertical intensity offset if a vertical slide was applied \cr
#'  `algorithm_error` \tab `numeric` \tab the vertical sliding suffers from a systematic effect induced by the used
#'  algorithm. The returned value is the standard deviation of all obtained De values while expanding the
#'  vertical sliding range. It can be added as systematic error to the final De error; so far wanted.\cr
#'  `vslide_range` \tab `numeric` \tab the range used for the vertical sliding \cr
#'  `num_slide_windows` \tab `integer` \tab the number of windows used for the vertical sliding \cr
#'  `squared_residuals` \tab `numeric` \tab the squared residuals (horizontal sliding)
#' }
#'
#'
#' **slot:** **`@info`**
#'
#' The original function call ([methods::language-class]-object)
#'
#' The output (`data`) should be accessed using the function [Luminescence::get_RLum].
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' The slid IR-RF curves with the finally obtained De
#'
#' @note
#' This function assumes that there is no sensitivity change during the
#' measurements (natural vs. regenerated signal), which is in contrast to the
#' findings by Buylaert et al. (2012).
#'
#' @section Function version: 0.7.10
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Results-class], [Luminescence::get_RLum],
#' [nls], [minpack.lm::nlsLM], [parallel::mclapply], [Luminescence::ExampleData.RF70Curves]
#'
#'
#' @references
#' Buylaert, J.P., Jain, M., Murray, A.S., Thomsen, K.J., Lapp, T.,
#' 2012. IR-RF dating of sand-sized K-feldspar extracts: A test of accuracy.
#' Radiation Measurements 44 (5-6), 560-565. \doi{10.1016/j.radmeas.2012.06.021}
#'
#' Erfurt, G., Krbetschek, M.R., 2003. IRSAR - A single-aliquot
#' regenerative-dose dating protocol applied to the infrared radiofluorescence
#' (IR-RF) of coarse- grain K-feldspar. Ancient TL 21, 35-42.
#' \doi{10.26034/la.atl.2003.358}
#'
#' Erfurt, G., 2003. Infrared luminescence of Pb+ centres in potassium-rich
#' feldspars. physica status solidi (a) 200, 429-438. \doi{10.1002/pssa.200306700}
#'
#' Erfurt, G., Krbetschek, M.R., 2003. Studies on the physics of the infrared
#' radioluminescence of potassium feldspar and on the methodology of its
#' application to sediment dating. Radiation Measurements 37, 505-510.
#' \doi{10.1016/s1350-4487(03)00058-1}
#'
#' Erfurt, G., Krbetschek, M.R., Bortolot, V.J., Preusser, F., 2003. A fully
#' automated multi-spectral radioluminescence reading system for geochronometry
#' and dosimetry. Nuclear Instruments and Methods in Physics Research Section
#' B: Beam Interactions with Materials and Atoms 207, 487-499.
#' \doi{10.1016/s0168-583x(03)01121-2}
#'
#' Frouin, M., Huot, S., Kreutzer, S., Lahaye, C., Lamothe, M., Philippe, A., Mercier, N., 2017.
#' An improved radiofluorescence single-aliquot regenerative dose protocol for K-feldspars.
#' Quaternary Geochronology 38, 13-24. \doi{10.1016/j.quageo.2016.11.004}
#'
#' Kreutzer, S., Murari, M.K., Frouin, M., Fuchs, M., Mercier, N., 2017.
#' Always remain suspicious: a case study on tracking down a technical artefact while measuring IR-RF.
#' Ancient TL 35, 20–30. \doi{10.26034/la.atl.2017.510}
#'
#' Murari, M.K., Kreutzer, S., Fuchs, M., 2018. Further investigations on IR-RF:
#' Dose recovery and correction. Radiation Measurements 120, 110–119.
#' doi: 10.1016/j.radmeas.2018.04.017 \doi{10.1016/j.radmeas.2018.04.017}
#'
#' Lapp, T., Jain, M., Thomsen, K.J., Murray, A.S., Buylaert, J.P., 2012. New
#' luminescence measurement facilities in retrospective dosimetry. Radiation
#' Measurements 47, 803-808.
#' \doi{10.1016/j.radmeas.2012.02.006}
#'
#' Trautmann, T., 2000. A study of radioluminescence kinetics of natural
#' feldspar dosimeters: experiments and simulations. Journal of Physics D:
#' Applied Physics 33, 2304-2310.
#' \doi{10.1088/0022-3727/33/18/315}
#'
#' Trautmann, T., Krbetschek, M.R., Dietrich, A., Stolz, W., 1998.
#' Investigations of feldspar radioluminescence: potential for a new dating
#' technique. Radiation Measurements 29, 421-425.
#' \doi{10.1016/s1350-4487(98)00012-2}
#'
#' Trautmann, T., Krbetschek, M.R., Dietrich, A., Stolz, W., 1999. Feldspar
#' radioluminescence: a new dating method and its physical background. Journal
#' of Luminescence 85, 45-58.
#' \doi{10.1016/s0022-2313(99)00152-0}
#'
#' Trautmann, T., Krbetschek, M.R., Stolz, W., 2000. A systematic study of the
#' radioluminescence properties of single feldspar grains. Radiation
#' Measurements 32, 685-690.
#' \doi{10.1016/s1350-4487(00)00077-9}
#'
#' ** Further reading**
#'
#' Murari, M.K., Kreutzer, S., King, G.E., Frouin, M., Tsukamoto, S., Schmidt, C., Lauer, T.,
#' Klasen, N., Richter, D., Friedrich, J., Mercier, N., Fuchs, M., 2021.
#' Infrared radiofluorescence (IR-RF) dating: A review. Quaternary Geochronology 64,
#' 101155. \doi{10.1016/j.quageo.2021.101155}
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.RLum.Analysis, envir = environment())
#'
#' ##(1) perform analysis using the method 'FIT'
#' results <- analyse_IRSAR.RF(object = IRSAR.RF.Data)
#'
#' ##show De results and test parameter results
#' get_RLum(results, data.object = "data")
#' get_RLum(results, data.object = "test_parameters")
#'
#' ##(2) perform analysis using the method 'SLIDE'
#' data(ExampleData.RF70Curves, envir = environment())
#' results <- analyse_IRSAR.RF(
#' object = RF70Curves,
#' method = "SLIDE",
#'  n.MC = 1)
#'
#' \dontrun{
#' ##(3) perform analysis using the method 'VSLIDE' and method control option
#' ## 'trace
#' results <- analyse_IRSAR.RF(
#'  object = RF70Curves,
#'  method = "VSLIDE",
#'  method_control = list(trace = TRUE))
#' }
#'
#' @export
analyse_IRSAR.RF<- function(
  object,
  sequence_structure = c("NATURAL", "REGENERATED"),
  RF_nat.lim = NULL,
  RF_reg.lim = NULL,
  method = c("FIT", "SLIDE", "VSLIDE", "NONE"),
  method_control = NULL,
  test_parameters = NULL,
  n.MC = 10,
  txtProgressBar = TRUE,
  plot = TRUE,
  plot_reduced = FALSE,
  ...
) {
  .set_function_name("analyse_IRSAR.RF")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO
  ## - if a file path is given, the function should try to find out whether an XSYG-file or
  ##   a BIN-file is provided
  ##  - update documentary ... if it works as expected.

  extraArgs <- list(...)

  # SELF CALL ---------------------------------------------------------------
  if (inherits(object, "list")) {
    ## expand input arguments
    rep.length <- length(object)

    sequence_structure <- .listify(sequence_structure, rep.length)
    RF_nat.lim <- .listify(RF_nat.lim, rep.length)
    RF_reg.lim <- .listify(RF_reg.lim, rep.length)
    method <- .listify(method, rep.length)
    test_parameters <- .listify(test_parameters, rep.length)
    n.MC <- .listify(n.MC, rep.length)

    ##main
    if ("main" %in% names(extraArgs)) {
      temp_main <- .listify(extraArgs$main, rep.length)
    }else{
      originator <- if (length(object) > 0) {
                      .validate_class(object[[1]], "RLum.Analysis",
                                      name = "All elements of 'object'")
                      object[[1]]@originator } else NA
      if (!is.na(originator) && originator == "read_RF2R") {
        temp_main <- lapply(object, function(x) x@info$ROI)
      } else {
        temp_main <- as.list(paste0("ALQ #",1:length(object)))
      }
    }

    ##run analysis
    temp <- lapply(seq_along(object), function(x) {
      analyse_IRSAR.RF(
        object = object[[x]],
        sequence_structure = sequence_structure[[x]],
        RF_nat.lim = RF_nat.lim[[x]],
        RF_reg.lim = RF_reg.lim[[x]],
        method = method[[x]],
        method_control = method_control,
        test_parameters = test_parameters[[x]],
        n.MC = n.MC[[x]],
        txtProgressBar = txtProgressBar,
        plot = plot,
        plot_reduced = plot_reduced,
        main = temp_main[[x]],
        ...)
    })

    ##combine everything to one RLum.Results object as this as what was written ... only
    ##one object

    ##merge results and check if the output became NULL
    results <- merge_RLum(temp)

    if (length(results) == 0)
      return(NULL)
    return(results)
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis", extra = "a 'list' of such objects")
  .validate_class(sequence_structure, "character")
  .validate_not_empty(object)
  if (length(sequence_structure) < 2) {
    .throw_error("'sequence_structure' should contain at least two elements")
  }
  if (!all(c("NATURAL", "REGENERATED") %in% sequence_structure)) {
    .throw_error("'sequence_structure' must contain one each of 'NATURAL' ",
                 "and 'REGENERATED'")
  }
  .validate_class(RF_nat.lim, c("numeric", "integer"), null.ok = TRUE)
  .validate_class(RF_reg.lim, c("numeric", "integer"), null.ok = TRUE)
  method <- .validate_args(toupper(method), c("FIT", "SLIDE", "VSLIDE", "NONE"))
  .validate_class(method_control, "list", null.ok = TRUE)
  .validate_positive_scalar(n.MC, int = TRUE, null.ok = TRUE)

  ##SELECT ONLY MEASURED CURVES
  ## (this is not really necessary but rather user friendly)
  if (length(suppressWarnings(get_RLum(object, curveType = "measured"))) > 0) {
    object <- get_RLum(object, curveType= "measured", drop = FALSE)
  }

  ##INVESTIGATE SEQUENCE OBJECT STRUCTURE

  ##grep object structure
  temp.sequence_structure <- structure_RLum(object)

  ##check whether both curve have the same length, in this case we cannot proceed (sliding
  ##is not allowed)
  if(length(unique(temp.sequence_structure[["x.max"]])) == 1 &&
     grepl("SLIDE", method) &&
     (is.null(RF_nat.lim) & is.null(RF_reg.lim))) {
    .throw_error("There is no further sliding space left. All curves have ",
                 "the same length and no limitation was set")
  }

  ##grep name of the sequence and the position this will be useful later on
  ##name
  record1 <- get_RLum(object, record.id = 1)
  aliquot.sequence_name <- suppressWarnings(get_RLum(record1,
                                                     info.object = "name")) %||% NA

  ##position
  aliquot.position <- suppressWarnings(get_RLum(record1,
                                                info.object = "position")) %||% NA

  ##date
  aliquot.date <- suppressWarnings(get_RLum(record1,
                                            info.object = "startDate"))
  if (!is.null(aliquot.date)) {
    ##transform so far the format can be identified
    if (nchar(aliquot.date) == 14) {
      aliquot.date <- paste(c(substr(aliquot.date, 1, 4),
                              substr(aliquot.date, 5, 6),
                              substr(aliquot.date, 7, 8)), collapse = "-")
    }
  }else{
    aliquot.date <- NA
  }

  ##set structure values
  temp.sequence_structure$protocol.step <-
    rep(sequence_structure, length_RLum(object))[1:length_RLum(object)]

  ## check that we both natural and regenerated are still there
  nat.idx <- which(temp.sequence_structure$protocol.step == "NATURAL")
  reg.idx <- which(temp.sequence_structure$protocol.step == "REGENERATED")
  if (length(nat.idx) == 0 || length(reg.idx) == 0) {
    .throw_error("'sequence_structure' is missing one of 'NATURAL' or ",
                 "'REGENERATED'")
  }

  ##=========================================================================
  ## SET CURVE LIMITS
  ##=========================================================================

  ##01
  ## get the allowed curve limits
  max.channels.nat <- sum(temp.sequence_structure$n.channels[nat.idx])
  max.channels.reg <- sum(temp.sequence_structure$n.channels[reg.idx])
  if (max.channels.nat > max.channels.reg) {
    .throw_error("The number of data channels for the natural (", max.channels.nat,
                 ") cannot exceed that of the regenerated (", max.channels.reg,
                 "), check your 'sequence_structure'")
  }

  ## 02 - check boundaries
  ##RF_nat.lim
  if (is.null(RF_nat.lim) || anyNA(RF_nat.lim)) {
    RF_nat.lim <- c(1, max.channels.nat)

  }else {
    ##this allows to provide only one boundary and the 2nd will be added automatically
    if (length(RF_nat.lim) == 1) {
      RF_nat.lim <- c(RF_nat.lim, max.channels.nat)
    }

    if (min(RF_nat.lim) < 1 || max(RF_nat.lim) > max.channels.nat) {
      RF_nat.lim <- c(1, max.channels.nat)
      .throw_warning("'RF_nat.lim' out of bounds, reset to c(",
                     .format_range(RF_nat.lim, sep = ", "), ")")
    }
  }

  ##RF_reg.lim
  if (is.null(RF_reg.lim)) {
    RF_reg.lim <- c(1, max.channels.reg)

  }else {
    ##this allows to provide only one boundary and the 2nd will be added automatically
    if (length(RF_reg.lim) == 1) {
      RF_reg.lim <- c(RF_reg.lim, max.channels.reg)
    }

    if (min(RF_reg.lim) < 1 || max(RF_reg.lim) > max.channels.reg) {
      RF_reg.lim <- c(1, max.channels.reg)
      .throw_warning("'RF_reg.lim' out of bounds, reset to c(",
                     .format_range(RF_reg.lim, sep = ", "), ")")
    }
  }

  ## check if intervals make sense at all
  len.RF_reg.lim <- length(RF_reg.lim[1]:RF_reg.lim[2])
  if (len.RF_reg.lim < RF_nat.lim[2]) {
    RF_reg.lim[2] <- RF_reg.lim[2] + abs(len.RF_reg.lim - RF_nat.lim[2]) + 1

    ## check that now we haven't gone beyond the size of RF_reg.lim
    if (RF_reg.lim[2] > max.channels.reg) {
      .throw_error("'RF_reg.lim' defines too short an interval and it's not ",
                   "possible to extend it")
    }

    .throw_warning("'RF_reg.lim' defines too short an interval, reset to c(",
                   .format_range(RF_reg.lim, sep = ", "), ")")
  }

  ## check again that we can actually slide the curve
  if (diff(RF_reg.lim) <= diff(RF_nat.lim)) {
    .throw_error("The range of regenerated channels should be larger than ",
                 "the range of natural channels")
  }

  # Method Control Settings -------------------------------------------------
  ##=========================================================================
  ## SET METHOD CONTROL PARAMETER - FOR BOTH METHODS
  ##=========================================================================
  ##
  ##set supported values with default
  method_control.settings <- list(
    trace = FALSE,
    trace_vslide = FALSE,
    maxiter = 500,
    warnOnly = FALSE,
    minFactor = 1 / 4096,
    correct_onset = TRUE,
    show_density = TRUE,
    show_fit = FALSE,
    n.MC = if(is.null(n.MC)) NULL else 1000,
    vslide_range = if(method[1] == "VSLIDE") "auto" else NULL,
    num_slide_windows = 3,
    cores = NULL
  )

  ## deprecated argument
  if ("method.control" %in% names(extraArgs)) {
    method_control <- extraArgs$method.control
    .deprecated("method.control", "method_control", since = "1.0.0")
  }

  ##modify list if necessary
  if (!is.null(method_control)) {

    ##check whether this arguments are supported at all
    unsupported.idx <- which(!names(method_control) %in%
                             names(method_control.settings))
    if (length(unsupported.idx) > 0) {
      .throw_warning(.collapse(names(method_control)[unsupported.idx]),
                     " not supported for 'method_control'. Supported arguments are: ",
                     .collapse(names(method_control.settings)))
    }

    ## check for odd user input
    temp.vslide_range <- method_control$vslide_range
    if (length(temp.vslide_range) > 0) {
      .validate_class(temp.vslide_range, c("character", "integer", "numeric"),
                      name = "'vslide_range' in 'method_control'")
    }
    if (is.character(temp.vslide_range) && temp.vslide_range[1] != "auto") {
      .throw_error("'vslide_range' in 'method_control' should be either ",
                   "'auto' or a 2-element numeric vector")
    }
    if (length(temp.vslide_range) > 2) {
      method_control$vslide_range <- method_control$vslide_range[1:2]
      .throw_warning("'vslide_range' in 'method_control' has more ",
                     "than 2 elements, only the first two were used")
    }

    temp.num_slide_windows <- method_control$num_slide_windows
    if (length(temp.num_slide_windows) > 0) {
      .validate_positive_scalar(temp.num_slide_windows, int = TRUE,
                                name = "'num_slide_windows' in 'method_control'")
      if (temp.num_slide_windows > 10) {
        method_control$num_slide_windows <- 10
        .throw_warning("'num_slide_windows' in 'method_control' should be ",
                       "between 1 and 10, reset to 10 ")
      }
    }

    ##modify list
    method_control.settings <- modifyList(
      x = method_control.settings,
      val = method_control,
      keep.null = TRUE)
  }

  ## control terminal output
  verbose <- extraArgs$verbose %||% TRUE

  ## don't show the progress bar if not verbose
  if (!verbose)
    txtProgressBar <- FALSE


  ##=========================================================================
  ## SET PLOT PARAMETERS
  ##=========================================================================
  ##get channel resolution (should be equal for all curves, but if not the mean is taken)
  resolution.RF <- round(mean((temp.sequence_structure$x.max/temp.sequence_structure$n.channels)),digits=1)

  plot.settings <- list(
    main = "IR-RF",
    xlab = "Time [s]",
    ylab = paste0("IR-RF [cts/", resolution.RF," s]"),
    log = "",
    cex = 1,
    legend = TRUE,
    legend.text = c("RF_nat","RF_reg"),
    legend.pos = "top",
    xaxt = "s"
    ##xlim and ylim see below as they has to be modified differently
  )

  ##modify list if something was set
  plot.settings <- modifyList(plot.settings, extraArgs)
  plot.settings$mtext.cex <- 0.7 * plot.settings$cex
  xlog <- grepl("x", plot.settings$log, fixed = TRUE)
  ylog <- grepl("y", plot.settings$log, fixed = TRUE)

  ##=============================================================================#
  ## ANALYSIS
  ##=============================================================================#

  ##grep first regenerated curve
  RF_reg <- as.data.frame(rbindlist(lapply(object@records[reg.idx],
                                           function(x) as.data.frame(x@data))))
  ## correct of the onset of detection by using the first time value
  if (grepl("SLIDE", method) && method_control.settings$correct_onset) {
      RF_reg[,1] <- RF_reg[,1] - RF_reg[1,1]
    }

  RF_reg.x <- RF_reg[RF_reg.lim[1]:RF_reg.lim[2],1]
  RF_reg.y <- RF_reg[RF_reg.lim[1]:RF_reg.lim[2],2]

  ##grep values from natural signal
  RF_nat <- as.data.frame(rbindlist(lapply(object@records[nat.idx],
                                           function(x) as.data.frame(x@data))))

  ## correct the onset of detection by using the first time value
  if (grepl("SLIDE", method) && method_control.settings$correct_onset) {
    RF_nat[,1] <- RF_nat[,1] - RF_nat[1,1]
  }

  ##limit values to fit range (at least to the minimum)
  RF_nat.limited<- RF_nat[min(RF_nat.lim):max(RF_nat.lim),]

  ##calculate some useful parameters
  RF_nat.mean <- mean(RF_nat.limited[,2])
  RF_nat.error.lower <- quantile(RF_nat.limited[,2], 0.975, na.rm = TRUE)
  RF_nat.error.upper <- quantile(RF_nat.limited[,2], 0.025, na.rm = TRUE)

  De <- De.error <- De.lower <- De.upper <- De.MC <- De.diff <- NA

  ## parameter boundaries
  lower <- c(
      phi.0 = .Machine$double.xmin,
      delta.phi = .Machine$double.xmin,
      lambda = .Machine$double.xmin,
      beta = .Machine$double.xmin
  )
  upper <- c(
      phi.0 = max(RF_reg.y),
      delta.phi = max(RF_reg.y),
      lambda = 1,
      beta = 100
  )

  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ##METHOD FIT
  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ## REGENERATED SIGNAL
    # set function for fitting ------------------------------------------------
    fit.function <- y ~ phi.0 - (delta.phi * ((1 - exp(
        -lambda * x
      )) ^ beta))

    ## stretched exponential function according to Erfurt et al. (2003)
    ## + phi.0 >> initial IR-RF flux
    ## + delta.phi >> dose dependent change of the IR-RF flux
    ## + lambda >> exponential parameter
    ## + beta >> dispersive factor

    # set start parameter estimation ------------------------------------------

    fit.parameters.start <- c(
      phi.0 = max(RF_reg.y),
      delta.phi = 1.5 * (max(RF_reg.y) - min(RF_reg.y)),
      lambda = 0.0001,
      beta = 1
    )

  if(method == "FIT"){

    # start nls fitting -------------------------------------------------------

    ##Monte Carlo approach for fitting
    fit.MC.results <- data.frame()

    ##produce set of start paramters
    n.MC <- n.MC %||% 0
    lambda.MC <- seq(0.0001, 0.001, length = n.MC)
    start.MC <- fit.parameters.start

    ##start fitting loop for MC runs
    for (i in seq_len(n.MC)) {
      start.MC["lambda"] <- lambda.MC[i]
      fit.MC <- try(nls(
        fit.function,
        trace = FALSE,
        data = list(x = RF_reg.x, y = RF_reg.y),
        algorithm = "port",
        stats::nls.control(
          maxiter = 100,
          warnOnly = FALSE,
          minFactor = 1 / 1024
        ),
        start = start.MC,
        lower = lower,
        upper = upper),
      silent = TRUE
      )

      if (!inherits(fit.MC, "try-error")) {
        fit.MC.coefs <- coef(fit.MC)
        for (val in c("phi.0", "delta.phi", "lambda", "beta"))
          fit.MC.results[i, val] <- fit.MC.coefs[val]
      }
    }

    ##FINAL fitting after successful MC
    if (length(stats::na.omit(fit.MC.results)) != 0) {

      ##choose median as final fit version
      fit.MC.results <- sapply(stats::na.omit(fit.MC.results), median)

      ##try final fitting
      fit <- try(nls(
        fit.function,
        trace = method_control.settings$trace,
        data = data.frame(x = RF_reg.x, y = RF_reg.y),
        algorithm = "port",
        stats::nls.control(
          maxiter = method_control.settings$maxiter,
          warnOnly = method_control.settings$warnOnly,
          minFactor = method_control.settings$minFactor
        ),
        start = fit.MC.results,
        lower = lower,
        upper = upper),
      silent = FALSE
      )
    }else{
      fit <- NA
      class(fit) <- "try-error"
    }

    # get parameters ----------------------------------------------------------
    # and with that the final De
    fit.parameters.results <- NA
    if (!inherits(fit,"try-error")) {
      fit.parameters.results <- coef(fit)
      residuals <- residuals(fit)
    }

    ##calculate De value
    if (!is.na(fit.parameters.results[1])) {
      RF_nat.vals <- c(RF_nat.mean, RF_nat.error.lower, RF_nat.error.upper)
      De.vals <- suppressWarnings(round(log(
          -((RF_nat.vals - fit.parameters.results["phi.0"]) /
            -fit.parameters.results["delta.phi"]
        ) ^ (1 / fit.parameters.results["beta"]) + 1
      ) / -fit.parameters.results["lambda"], digits = 2))

      De <- De.vals[1]
      De.lower <- De.vals[2]
      De.upper <- De.vals[3]

      ##This could be solved with a MC simulation, but for this the code has to be adjusted
      ##The question is: Where the parameters are coming from?
      ##TODO
      De.error <- NA
    }
  }

  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ## METHOD SLIDE - ANALYSIS
  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  else if(method == "SLIDE" || method == "VSLIDE"){
    ##convert to matrix (in fact above the matrix data were first transferred to
    ##data.frames ... here
    ##we correct this ... again)
    RF_nat.limited <- as.matrix(RF_nat.limited)
    RF_reg.limited <- matrix(c(RF_reg.x, RF_reg.y), ncol = 2)
    RF_nat <- as.matrix(RF_nat)

    ##DEFINE FUNCTION FOR SLIDING
    ##FIND MINIMUM - this is done in a function so that it can be further used for MC simulations
    # sliding() -----------------------------------------------------------------------------------
    sliding <- function(RF_nat,
                        RF_nat.limited,
                        RF_reg.limited,
                        n.MC = method_control.settings$n.MC,
                        vslide_range = method_control.settings$vslide_range,
                        num_slide_windows = method_control.settings$num_slide_windows,
                        trace = method_control.settings$trace_vslide,
                        numerical.only = FALSE){

      ##(0) set objects ... nomenclature as used in Frouin et al., please note that here the index
      ##is used instead the real time values
      t_max.id <- nrow(RF_reg.limited)
      t_max_nat.id <- nrow(RF_nat.limited)
      t_min <- RF_nat.limited[1,1]

      ##(1) calculate sum of residual squares using internal Rcpp function

      ##initialise slide range for specific conditions, namely NULL and "auto"
      compute_algorithm_error <- TRUE
      if (is.null(vslide_range)) {
        vslide_range <- 0

      } else if (vslide_range[1] == "auto") {
        range <- max(RF_reg.limited[, 2]) - min(RF_reg.limited[, 2])
        vslide_range <- -range:range
      } else{
        vslide_range <- vslide_range[1]:vslide_range[2]
        compute_algorithm_error <- FALSE
      }

      ##problem: the optimisation routine slightly depends on the chosen input sliding vector
      ##and it might get trapped in a local minimum
      ##therefore we run the algorithm by expanding the sliding vector
      algorithm_error <- NA
      if(!is.null(vslide_range) && any(vslide_range != 0)){

        ##construct list of vector ranges we want to check for, this should avoid that we
        ##got trapped in a local minimum
        mid.idx <- floor((length(vslide_range) + 1) / 2)
        steps <- as.integer(seq(min(t_max_nat.id / 2, mid.idx / 16), mid.idx,
                                length.out = num_slide_windows))
        vslide_range.list <- lapply(steps, function(x) mid.idx + c(-x, x))

        ##correct for out of bounds problem; it might occur
        vslide_range.list[[num_slide_windows]] <- c(0, length(vslide_range))

        ##TODO ... this is not really optimal, but ok for the moment, better would be
        ##the algorithm finds sufficiently the global minimum.
        ##now run it in a loop and expand the range from the inner to the outer part
        ##at least this is considered for the final error range ...
        temp_minimum_list <- lapply(vslide_range.list, function(range) {
          src_analyse_IRSARRF_SRS(
            values_regenerated_limited =  RF_reg.limited[,2],
            values_natural_limited = RF_nat.limited[,2],
            vslide_range = vslide_range[range[1]:range[2]],
            n_MC = 0, #we don't need MC runs here, so make it quick
            trace = trace)[c("sliding_vector_min_index","vslide_minimum", "vslide_index")]
        })

        ##get all horizontal index value for the local minimum (corresponding to the vslide)
        temp_hslide_indices <- vapply(temp_minimum_list, function(x) {
          x$sliding_vector_min_index}, FUN.VALUE = numeric(length = 1))

        ##get all the minimum values
        temp_minimum <- vapply(temp_minimum_list, function(x) {
          x$vslide_minimum}, FUN.VALUE = numeric(length = 1))

        ##get minimum and set it to the final range
        opt.range <- vslide_range.list[[which.min(temp_minimum)]]
        vslide_range <- vslide_range[opt.range[1]:opt.range[2]]

        ##get all possible t_n values for the range expansion ... this can be considered
        ##as somehow systematic uncertainty, but it will be only calculated if the full range
        ##is considered, otherwise it is too biased by the user's choice
        ##ToDo: So far the algorithm error is not sufficiently documented
        if (compute_algorithm_error) {
          algorithm_error <- sd(vapply(temp_hslide_indices, function(k) {
            temp.sliding.step <- RF_reg.limited[k] - t_min
            ## return the offset of the t_n values
            RF_nat[1, 1] + temp.sliding.step
          }, FUN.VALUE = numeric(1)))
        }
      }

      ##now run the final sliding with the identified range that corresponds to the minimum value
      temp.sum.residuals <-
        src_analyse_IRSARRF_SRS(
          values_regenerated_limited =  RF_reg.limited[,2],
          values_natural_limited = RF_nat.limited[,2],
          vslide_range = vslide_range,
          n_MC = n.MC %||% 0,
          trace = trace
      )

      #(2) get minimum value (index and time value)
      index_min <- which.min(temp.sum.residuals$sliding_vector)
      t_n.id <- if (length(index_min) == 0) 1 else index_min

      I_n <- vslide_range[temp.sum.residuals$vslide_index] %||% 0
      temp.sliding.step <- RF_reg.limited[t_n.id] - t_min

      ##(3) slide curve graphically ... full data set we need this for the plotting later
      RF_nat.slid <- matrix(data = c(RF_nat[,1] + temp.sliding.step, RF_nat[,2] + I_n), ncol = 2)
      t_n <- RF_nat.slid[1,1]

      ##the same for the MC runs of the minimum values
      t_n.MC <- NA_integer_
      if(!is.null(n.MC)) {
        t_n.MC <-
          vapply(
            X = temp.sum.residuals$sliding_vector_min_MC,
            FUN = function(x) {
              ##get minimum for MC
              t_n.id.MC <- which(temp.sum.residuals$sliding_vector == x)

              ## there is a non-zero chance that we have got two indices for
              ## the minimun, so we take the mean
              RF_nat[1, 1] + mean(RF_reg.limited[t_n.id.MC]) - t_min
            },
            FUN.VALUE = numeric(1))
      }

      ##(4) get residuals (needed to be plotted later)
      ## they cannot be longer than the RF_reg.limited curve
      reg.limited.idx <- t_n.id:t_max.id
      len.shorter <- min(nrow(RF_nat.limited), length(reg.limited.idx))
      residuals <- (RF_nat.limited[1:len.shorter, 2] + I_n) -
        RF_reg.limited[reg.limited.idx[1:len.shorter], 2]

      ##(4.1) calculate De from the first channel ... which is t_n here
      De <- round(t_n, digits = 2)
      De.MC <- round(t_n.MC, digits = 2)

      ##(5) calculate trend fit
      max.rows <- min(nrow(RF_nat.limited), length(residuals))
      temp.fit <- stats::lm(y ~ x, data.frame(x = RF_nat.limited[1:max.rows, 1],
                                              y = residuals))
      temp.trend.fit <- coef(temp.fit)

      ##return values and limited if they are not needed
      if (numerical.only) {
        return(list(De = De, De.MC = De.MC))
      }
      list(De = De,
            De.MC = De.MC,
            residuals = residuals,
            trend.fit = temp.trend.fit,
            RF_nat.slid = RF_nat.slid,
            t_n.id = t_n.id,
            I_n = I_n,
            algorithm_error = algorithm_error,
            vslide_range = if(is.null(vslide_range)){NA}else{range(vslide_range)},
            num_slide_windows = num_slide_windows,
            squared_residuals = temp.sum.residuals$sliding_vector
          )
    }##end of function sliding()

    ##PERFORM sliding and overwrite values
    slide <-  sliding(
      RF_nat = RF_nat,
      RF_nat.limited = RF_nat.limited,
      RF_reg.limited = RF_reg.limited
    )

    ##write results in variables
    De <- slide$De
    residuals <- slide$residuals
    RF_nat.slid <-  slide$RF_nat.slid
    I_n <- slide$I_n

    # ERROR ESTIMATION
    # MC runs for error calculation -----------------------------------------

    ##set residual matrix for MC runs, i.e. set up list of pseudo RF_nat curves as function
    ##(i.e., bootstrap from the natural curve distribution)

    if(!is.null(n.MC)){
      slide.MC.list <- lapply(1:n.MC, function(dummy) {
        reg.limited.idx <- slide$t_n.id:nrow(RF_reg.limited)
        len.shorter <- min(nrow(RF_nat.limited), length(reg.limited.idx))
        cbind(RF_nat.limited[1:len.shorter, 1],
              RF_reg.limited[reg.limited.idx[1:len.shorter], 2] +
              sample(residuals, len.shorter, replace = TRUE))
      })

      ##set parallel calculation if wanted
      if (is.null(method_control.settings$cores)) {
        cores <- 1

      } else {
        available.cores <- parallel::detectCores()
        requested.cores <- method_control.settings$cores[1]

        ##case 'auto'
        if (requested.cores == "auto") {
          cores <- max(available.cores - 2, 1) # nocov

        } else if (is.numeric(requested.cores)) {
          .validate_positive_scalar(requested.cores, int = TRUE,
                                    name = "method_control.settings$cores")
          if (requested.cores > available.cores) {
            ##assign all they have, it is not our problem
            # nocov start
            .throw_warning("Number of cores limited to the maximum ",
                           "available (", available.cores, ")")
            # nocov end
          }
          cores <- min(requested.cores, available.cores)

        }else{
          .throw_message("Invalid value for control argument 'cores', ",
                         "value set to 1")
          cores <- 1
        }

        if (verbose)
          .throw_message("Using ", cores, ifelse(cores == 1, " core", " cores"),
                         " ...", error = FALSE)
      }

      ## SINGLE CORE -----
      if (cores[1] == 1) {
        if(txtProgressBar){
          ##progress bar
          cat("\n\t Run Monte Carlo loops for error estimation\n")
          pb <- txtProgressBar(min = 0, max = n.MC, initial = 0, char = "=", style = 3)
        }

        De.MC <- sapply(1:n.MC, function(i) {
          # update progress bar
          if (txtProgressBar) setTxtProgressBar(pb, i)

          sliding(
            RF_nat = RF_nat,
            RF_reg.limited = RF_reg.limited,
            RF_nat.limited = slide.MC.list[[i]],
            numerical.only = TRUE
          )[[2]]
        })

        ## close progress bar
        if (txtProgressBar) close(pb)

      ## MULTICORE -----
      } else {
        ## Create the determined number of R copies
        cl <- parallel::makeCluster(cores)

        ##run MC runs
        De.MC <- parallel::parSapply(cl, X = slide.MC.list,
                                     FUN = function(x){
                                       sliding(
                                         RF_nat = RF_nat,
                                         RF_reg.limited = RF_reg.limited,
                                         RF_nat.limited = x,
                                         numerical.only = TRUE
                                       )[[2]]
                                     })
        ##destroy multicore cluster
        parallel::stopCluster(cl)
      }
      ##calculate absolute deviation between De and the here newly calculated De.MC
      ##this is, e.g. ^t_n.1* - ^t_n in Frouin et al.
      De.diff <- diff(x = c(De, De.MC))
      De.error <- round(sd(De.MC), digits = 2)
      De.lower <- De - quantile(De.diff, 0.975, na.rm = TRUE)
      De.upper <- De - quantile(De.diff, 0.025, na.rm = TRUE)
    }
  }

  ##=========================================================================
  ## TEST PARAMETER
  ##=========================================================================
  ## Test parameter are evaluated after all the calculations have been done as
  ## it should be up to the user to decide whether a value should be taken into account or not.

  ##(0)
  ##set default values and overwrite them if there was something new
  ##set defaults
  TP <- list(
    curves_ratio = 1.001,
    intersection_ratio = NA,
    residuals_slope = NA,
    curves_bounds = ceiling(max(RF_reg.x)),
    dynamic_ratio = NA,
    lambda = NA,
    beta = NA,
    delta.phi = NA
  )

    ##modify default values by given input
    if(!is.null(test_parameters)){TP <- modifyList(TP, test_parameters)}

    ##remove NULL elements from list
    TP <- .rm_NULL_elements(TP)

    ##set list with values we want to evaluate
    TP <- lapply(TP, function(x){
      data.frame(THRESHOLD = as.numeric(x), VALUE = NA, STATUS = "OK", stringsAsFactors = TRUE)
    })


  ##(1) check if RF_nat > RF_reg, considering the fit range
  ##TP$curves_ratio
    if ("curves_ratio" %in% names(TP)) {
      TP$curves_ratio$VALUE <-
        sum(RF_nat.limited[,2]) / sum(RF_reg[RF_nat.lim[1]:RF_nat.lim[2], 2])

      if (!is.na(TP$curves_ratio$THRESHOLD)) {
        TP$curves_ratio$STATUS <-
          ifelse(TP$curves_ratio$VALUE > TP$curves_ratio$THRESHOLD, "FAILED", "OK")
      }
    }

   ##(1.1) check if RF_nat > RF_reg, considering the fit range
   ##TP$intersection_ratio
    if ("intersection_ratio" %in% names(TP)) {

      ##It is, as always, a little bit more complicated ...
      ##We cannot just normalise both curves and compare ratios. With increasing De the curve
      ##shape of the RF_nat curve cannot be the same as the RF_reg curve at t = 0. Therefore we
      ##have to find the segment in the RF_reg curve that fits to the RF_nat curve
      ##
      ##(1) get maximum count value for RF_nat
      IR_RF_nat.max <- max(RF_nat.limited[,2])

      ##(2) find corresponding time value for RF_reg (here no limited)
      IR_RF_reg.corresponding_id <- which.min(abs(RF_reg[,2] - IR_RF_nat.max))

      ##(3) calculate ratio, but just starting from the point where both curves correspond
      ##in terms of intensity, otherwise the ratio cannot be correct

      ##the boundary check is necessary to avoid errors
      len.RF_nat.lim <- length(RF_nat.lim[1]:RF_nat.lim[2])
      if (IR_RF_reg.corresponding_id + len.RF_nat.lim > length(RF_reg[, 2])) {
        TP$intersection_ratio$VALUE <- Inf

      }else{
      this.idx <- IR_RF_reg.corresponding_id + (1:len.RF_nat.lim) - 1
      TP$intersection_ratio$VALUE <-
        abs(1 - sum(RF_nat.limited[, 2] / IR_RF_nat.max) /
            sum(RF_reg[this.idx, 2] / max(RF_reg[this.idx, 2])))

      if (!is.na(TP$intersection_ratio$THRESHOLD)) {
        TP$intersection_ratio$STATUS <-
          ifelse(TP$intersection_ratio$VALUE > TP$intersection_ratio$THRESHOLD, "FAILED", "OK")
      }
      }
    }

  ##(2) check slop of the residuals using a linear fit
  ##TP$residuals_slope
  if ("residuals_slope" %in% names(TP) && exists("slide")) {
        TP$residuals_slope$VALUE <- abs(slide$trend.fit[2])

        if (!is.na(TP$residuals_slope$THRESHOLD)) {
          TP$residuals_slope$STATUS <- ifelse(
            TP$residuals_slope$VALUE > TP$residuals_slope$THRESHOLD, "FAILED", "OK")
        }
  }

  ##(3) calculate dynamic range of regenrated curve
  ##TP$dynamic_ratio
  if ("dynamic_ratio" %in% names(TP)) {
    TP.dynamic_ratio <- subset(temp.sequence_structure,
                               temp.sequence_structure$protocol.step == "REGENERATED")
    TP$dynamic_ratio$VALUE <- min(TP.dynamic_ratio$y.max / TP.dynamic_ratio$y.min)

    if (!is.na(TP$dynamic_ratio$THRESHOLD)){
      TP$dynamic_ratio$STATUS  <- ifelse(
        TP$dynamic_ratio$VALUE < TP$dynamic_ratio$THRESHOLD , "FAILED", "OK")
    }
  }

  ##(4) decay parameter
  ##TP$lambda
  if (all(c("lambda", "beta", "delta.phi") %in% names(TP))) {

    fit.lambda <- try(minpack.lm::nlsLM(
        fit.function,
        data = data.frame(x = RF_reg.x, y = RF_reg.y),
        algorithm = "LM",
        start = fit.parameters.start,
        lower = lower,
        upper = upper),
    silent = TRUE
    )

    if(!inherits(fit.lambda, "try-error")){
       temp.coef <- coef(fit.lambda)

       TP$lambda$VALUE <- temp.coef["lambda"]
       TP$beta$VALUE <- temp.coef["beta"]
       TP$delta.phi$VALUE <- temp.coef["delta.phi"]

       if (!is.na( TP$lambda$THRESHOLD)){
        TP$lambda$STATUS <- ifelse(TP$lambda$VALUE <= TP$lambda$THRESHOLD, "FAILED", "OK")
       }

       if (!is.na( TP$beta$THRESHOLD)){
         TP$beta$STATUS <- ifelse(TP$beta$VALUE <= TP$beta$THRESHOLD, "FAILED", "OK")
       }

       if (!is.na( TP$delta.phi$THRESHOLD)){
         TP$delta.phi$STATUS <- ifelse(TP$delta.phi$VALUE <= TP$delta.phi$THRESHOLD, "FAILED", "OK")
       }
    }
  }

  ##(99) check whether after sliding the
  ##TP$curves_bounds
  if (!is.null(TP$curves_bounds)) {
    if(exists("slide")){
      ## add one channel on the top to make sure that it works
      TP$curves_bounds$VALUE <- max(RF_nat.slid[RF_nat.lim,1]) + (RF_nat[2,1] - RF_nat[1,1])

       if (!is.na(TP$curves_bounds$THRESHOLD)){
        TP$curves_bounds$STATUS <- ifelse(TP$curves_bounds$VALUE >= floor(max(RF_reg.x)), "FAILED", "OK")
       }

    }else if(exists("fit")){
      TP$curves_bounds$VALUE <- De.upper

      if (!is.na(TP$curves_bounds$THRESHOLD)){
        TP$curves_bounds$STATUS <- ifelse(TP$curves_bounds$VALUE  >= max(RF_reg.x), "FAILED", "OK")
      }
    }
  }

  ##Combine everything in a data.frame
  De.status <- "OK"
  TP.data.frame <- NULL
  if (length(TP) != 0) {
      TP.data.frame <- as.data.frame(
        cbind(
          POSITION =  as.integer(aliquot.position),
          PARAMETER = names(TP),
          do.call(data.table::rbindlist, args = list(l = TP)),
          SEQUENCE_NAME = aliquot.sequence_name,
          UID = NA
        )
      )

      ##set De.status to indicate whether there is any problem with the De according to the test parameter
      if ("FAILED" %in% TP.data.frame$STATUS) {
        De.status <- "FAILED"
      }
  }

  ##=========================================================================
  ## Plotting
  ##=========================================================================
  if (plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    ##get internal colour definition
    col <- get("col", pos = .LuminescenceEnv)

    if (!plot_reduced && method != "NONE") {
        graphics::layout(matrix(c(1, 2), 2, 1, byrow = TRUE), 2, c(1.3, 0.4), TRUE)
        par(mar = c(0, 4, 3, 1))
    }
    par(cex = plot.settings[["cex"]])

    ##here control xlim and ylim behaviour
    xlim <- extraArgs$xlim %||% c(if (xlog)
                                    min(temp.sequence_structure$x.min) else 0,
                                  max(temp.sequence_structure$x.max))
    ylim <- extraArgs$ylim %||% c(min(temp.sequence_structure$y.min),
                                  max(temp.sequence_structure$y.max))

    ##open plot area
    plot(
      NA,NA,
      xlim = xlim,
      ylim = ylim,
      xlab = ifelse(method == "NONE" || plot_reduced, plot.settings$xlab, " "),
      xaxt = ifelse(method == "NONE" || plot_reduced, plot.settings$xaxt, "n"),
      yaxt = "n",
      ylab = plot.settings$ylab,
      main = plot.settings$main,
      log = plot.settings$log
    )

    if(De.status == "FAILED"){

      ##build list of failed TP
      mtext.message <- paste0(
        "Threshold exceeded for: ",
        .collapse(TP.data.frame$PARAMETER[TP.data.frame$STATUS == "FAILED"]),
                  ". For details see manual.")

      ##print mtext
      mtext(text = mtext.message,
            side = 3, outer = TRUE, col = "red",
            cex = plot.settings$mtext.cex)
      .throw_warning(mtext.message)
    }

    ##use scientific format for y-axis
    labels <- axis(2, labels = FALSE)
    axis(side = 2, at = labels, labels = format(labels, scientific = TRUE))

    ##(1) plot points that have been not selected
    points(RF_reg[-(min(RF_reg.lim):max(RF_reg.lim)),1:2], pch=3, col=col[19])

    ##(2) plot points that has been used for the fitting
    points(RF_reg.x,RF_reg.y, pch=3, col=col[10])

    ##show natural points if no analysis was done
    if (method == "NONE") {
      ##add points
      points(RF_nat, pch = 20, col = "grey")
      points(RF_nat.limited, pch = 20, col = "red")

      ## subtitle
      if ("mtext" %in% names(extraArgs)) {
        mtext(side = 3, extraArgs$mtext, cex = plot.settings$mtext.cex)
      }

      ##legend
      if (plot.settings$legend) {
        legend(
          plot.settings$legend.pos,
          legend = plot.settings$legend.text,
          pch = c(19, 3),
          col = c("red", col[10]),
          horiz = TRUE,
          bty = "n",
          cex = 0.9
        )
      }
    }

    ##Add fitted curve, if possible. This is a graphical control that might be considered
    ##as useful before further analysis will be applied
    if (method_control.settings$show_fit) {
      if (!inherits(fit.lambda, "try-error")) {
        fit.lambda_coef <- coef(fit.lambda)
        curve(fit.lambda_coef[[1]]-
                (fit.lambda_coef[[2]]*
                   ((1-exp(-fit.lambda_coef[[3]]*x))^fit.lambda_coef[[4]])),
              add=TRUE,
              lty = 2,
              col="red")

        rm(fit.lambda_coef)
      }else{
        .throw_warning("No fit possible, no fit shown.")
      }
    }

    ## ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ## PLOT - METHOD FIT
    ## ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if(method == "FIT"){
      ## silence notes raised by R CMD check
      x <- NULL

      ##plot fitted curve
      curve(fit.parameters.results["phi.0"]-
              (fit.parameters.results["delta.phi"]*
                 ((1-exp(-fit.parameters.results["lambda"]*x))^fit.parameters.results["beta"])),
            add=TRUE,
            from = RF_reg[min(RF_reg.lim), 1],
            to = RF_reg[max(RF_reg.lim), 1],
            col="red")

      ##plotting to show the limitations if RF_reg.lim was chosen
      ## show fitted curve GREY (before red curve)
      curve(fit.parameters.results["phi.0"]-
              (fit.parameters.results["delta.phi"]*
                 ((1-exp(-fit.parameters.results["lambda"]*x))^fit.parameters.results["beta"])),
            add=TRUE,
            from = min(RF_reg[, 1]),
            to = RF_reg[min(RF_reg.lim), 1],
            col="grey")

      ##show fitted curve GREY (after red curve)
      curve(fit.parameters.results["phi.0"]-
              (fit.parameters.results["delta.phi"]*
                 ((1-exp(-fit.parameters.results["lambda"]*x))^fit.parameters.results["beta"])),
            add=TRUE,
            from = RF_reg[max(RF_reg.lim), 1],
            to = max(RF_reg[, 1]),
            col="grey")

      ##add points
      points(RF_nat, pch = 20, col = col[19])
      points(RF_nat.limited, pch = 20, col = col[2])

      ##legend
      if (plot.settings$legend) {
        legend(
          plot.settings$legend.pos,
          legend = plot.settings$legend.text,
          pch = c(19, 3),
          col = c("red", col[10]),
          horiz = TRUE,
          bty = "n",
          cex = 0.9
        )
      }

      ## plot range chosen for fitting
      abline(v=RF_reg[min(RF_reg.lim), 1], lty=2)
      abline(v=RF_reg[max(RF_reg.lim), 1], lty=2)

      ##plot De if De was calculated
      if (!is.na(De)) {
        lines(c(0,De.lower), c(RF_nat.error.lower,RF_nat.error.lower), lty=2, col="grey")
        lines(c(0,De), c(RF_nat.mean,RF_nat.mean), lty=2, col="red")
        lines(c(0,De.upper), c(RF_nat.error.upper,RF_nat.error.upper), lty=2, col="grey")

        lines(c(De.lower, De.lower),
              c(0,RF_nat.error.lower), lty=2, col="grey")
        lines(c(De,De), c(0, RF_nat.mean), lty=2, col="red")
        lines(c(De.upper, De.upper),
              c(0,RF_nat.error.upper), lty=2, col="grey")
      }

      ##Insert fit and result
      mtext.txt <-  substitute(D[e] == De,
                               list(De = paste0(De, " [", De.lower,
                                                " ; ", De.upper,"]")))
      if (!is.na(De) && max(De, De.upper) > max(RF_reg.x)) {
        try(mtext(side=3, mtext.txt,
                  line = 0, cex = plot.settings$mtext.cex, col = "red"),
            silent = TRUE)
        De.status <- "VALUE OUT OF BOUNDS"

      } else{
        if ("mtext" %in% names(extraArgs)) {
          mtext(side = 3, extraArgs$mtext, cex = plot.settings$mtext.cex)
        }else{
          try(mtext(side = 3, mtext.txt,
                    line = 0, cex = plot.settings$mtext.cex), silent = TRUE)
        }

        De.status <- "OK"
      }
    }

    ## ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ## PLOT - METHOD SLIDE
    ## ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    else if (method == "SLIDE" || method == "VSLIDE") {
      ##(0) density plot
      if (method_control.settings$show_density && !is.null(n.MC)) {
        ##showing the density makes only sense when we see at least 10 data points
        if (!anyNA(De.MC) && length(unique(De.MC)) >= 15) {

          ##calculate density De.MC
          density.De.MC <- density(De.MC)

          ##calculate transformation function
          x.1 <- max(density.De.MC$y)
          x.2 <- min(density.De.MC$y)

          ## limit the scaling a little bit
          threshold.value <- max(RF_reg.limited[, 2]) -
            (max(RF_reg.limited[, 2]) - min(RF_reg.limited[, 2])) * 0.5
          y.1 <- min(RF_nat.limited[1, 2], threshold.value)
          y.2 <- par("usr")[3]

          m <- (y.1 - y.2) / (x.1 + x.2)
          n <- y.1 - m * x.1

          density.De.MC$y <- m * density.De.MC$y + n
          rm(x.1,x.2,y.1,y.2,m,n)

          polygon(density.De.MC$x,
                  density.De.MC$y,
                  col = rgb(0,0.4,0.8,0.5))
        }else{
          .throw_warning("Narrow density distribution, ",
                         "no density distribution plotted")
        }
      }

      ##(1) plot unused points in grey ... unused points are points outside of the set limit
      points(
        matrix(RF_nat.slid[-(min(RF_nat.lim):max(RF_nat.lim)),1:2], ncol = 2),
        pch = 21, col = col[19]
      )

      ##(2) add used points
      points(RF_nat.slid[min(RF_nat.lim):max(RF_nat.lim),], pch = 21, col = col[2],
             bg = col[2])

      ##(3) add line to show the connection between the first point and the De
      lines(x = c(RF_nat.slid[1,1], RF_nat.slid[1,1]),
            y = c(.Machine$double.xmin,RF_nat.slid[1,2]),
            lty = 2,
            col = col[2]
      )

      ##(4) add arrow at the lowest y-coordinate possible to show the sliding
      if (!ylog) {
        shape::Arrows(
          x0 = 0,
          y0 = ylim[1],
          y1 = ylim[1],
          x1 = RF_nat.slid[1,1],
          arr.type = "triangle",
          arr.length = 0.3 * par()[["cex"]],
          code = 2,
          col = col[2],
          arr.adj = 1,
          arr.lwd = 1
        )
      }

      ##(5) add vertical shift as arrow; show nothing if nothing was shifted
      if (!ylog && I_n != 0) {
        shape::Arrows(
          x0 = (0 + par()$usr[1])/2,
          y0 = RF_nat[1,2],
          y1 = RF_nat[1,2] + I_n,
          x1 = (0 + par()$usr[1])/2,
          arr.type = "triangle",
          arr.length = 0.3 * par()[["cex"]],
          code = 2,
          col = col[2],
          arr.adj = 1,
          arr.lwd = 1
        )
      }

      ##TODO
      ##uncomment here to see all the RF_nat curves produced by the MC runs
      ##could become a polygone for future versions
      #lapply(1:n.MC, function(x){lines(slide.MC.list[[x]], col = rgb(0,0,0, alpha = 0.2))})

      ## plot range chosen for fitting
      abline(v=RF_reg[min(RF_reg.lim), 1], lty=2)
      abline(v=RF_reg[max(RF_reg.lim), 1], lty=2)

      if (plot.settings$legend) {
        legend(
          plot.settings$legend.pos,
          legend = plot.settings$legend.text,
          pch = c(19, 3),
          col = c("red", col[10]),
          horiz = TRUE,
          bty = "n",
          cex = 0.9
        )
      }

      ##write information on the De in the plot
      if("mtext" %in% names(extraArgs)) {
        mtext(side = 3, extraArgs$mtext, cex = plot.settings$mtext.cex)
      }else{
        try(mtext(side=3,
                  substitute(D[e] == De, list(De=paste0(De," [", De.lower, " ; ", De.upper, "]"))),
                  line = 0, cex = plot.settings$mtext.cex),
            silent=TRUE)
      }
    }

    if (!plot_reduced && method != "NONE") {
      ## lower plot of residuals
      par(mar = c(4.2, 4, 0, 1))

      fit.error <- method == "FIT" && is.na(fit.parameters.results[1])
      plot(
          NA,
          NA,
          xlim = xlim,
          ylim = if (fit.error) c(-1, 1) else range(residuals),
          xlab = plot.settings$xlab,
          ylab = "E",
          xaxt = plot.settings$xaxt,
          yaxt = "n",
          type = "p",
          log = gsub("y", "", plot.settings$log)
      )

      if (!fit.error) {
        ## add axis for 0 ... means if the 0 is not visible there is labelling
        axis(side = 4, at = 0, labels = 0)

        ## add residual points
        if (method == "FIT") {
          points(RF_reg.x, residuals, pch = 20, col = "grey")
        } else {
          temp.points.diff <- max(length(min(RF_nat.lim):max(RF_nat.lim)) -
                                  length(residuals), 0)
          points(RF_nat.slid[c(min(RF_nat.lim):(max(RF_nat.lim) - temp.points.diff)), 1],
                 residuals, pch = 20, col = rgb(0, 0, 0, 0.4))

          ## add residual indicator (should circle around 0)
          col.ramp <- grDevices::colorRampPalette(c(col[19], "white", col[19]))
          col.polygon <- col.ramp(100)

          xright <- if (xlog) 10^par("usr")[2] else par("usr")[2]
          shape::filledrectangle(
                     mid = c((xlim[2] + xright) / 2,
                             max(residuals) - diff(range(residuals)) / 2),
                     wx = xright - xlim[2],
                     wy = diff(range(residuals)),
                     col = col.polygon
                 )

          ## 0-line indicator and arrows if this is not visible
          ## red colouring here only if the 0 point is not visible to avoid too much colouring
          if (max(residuals) < 0) {
            ## draw arrow pointing down
            angle <- 270
            y0 <- max(residuals)
          } else if (min(residuals) > 0) {
            ## draw arrow pointing up
            angle <- 90
            y0 <- min(residuals)
          } else { # min < 0, max > 0
            ## draw no arrow, only a + sign
            angle <- NA
            points(xlim[2], 0, pch = 3)
          }
          if (!is.na(angle)) {
            shape::Arrowhead(
                       x0 = xlim[2] + (par("usr")[2] - xlim[2]) / 2,
                       y0 = y0,
                       angle = angle,
                       lcol = col[2],
                       arr.length = 0.4,
                       arr.type = "triangle",
                       arr.col = col[2]
                   )
          }

          ## add 0 line
          abline(h = 0, lty = 3)

          ## add vertical line to mark De (t_n)
          abline(v = De, lty = 2, col = col[2])

          ## add numeric value of De ... t_n
          axis(side = 1,
               at = De,
               labels = De,
               cex.axis = 0.8,
               col = "blue",
               padj = -1.55)
        }
      } else {
        text(x = (par()$usr[2] - par()$usr[1]) / 2,
             y = 0, "Fitting Error!")
      }
    }

    ## TODO: CONTROL PLOT! can be implemented in appropriate form in a later version
    if (method %in% c("SLIDE", "VSLIDE") && method_control.settings$trace) {
        par(new = TRUE)
        plot(
            RF_reg.limited[1:length(slide$squared_residuals),1],
            slide$squared_residuals,
            ylab = "",
            type = "l",
            xlab = "",
            xaxt = plot.settings$xaxt,
            axes = FALSE,
            xlim = xlim,
            ylim = ylim,
            log = "y"
        )
    }
  }#endif::plot

  # Return ------------------------------------------------------------------
  ##catch up worst case scenarios ... means something went wrong
  if (!exists("fit")) {
    fit  <- list()
    if (exists("fit.lambda")) {
      fit <- fit.lambda
    }
  }
  if(!exists("slide")){slide <- list()}

  ##generate unique identifier
  UID <- create_UID()

  ##combine values for De into a data frame
  De.values <- data.frame(
      DE = De,
      DE.ERROR = De.error,
      DE.LOWER = De.lower,
      DE.UPPER = De.upper,
      DE.STATUS = De.status,
      RF_NAT.LIM = .format_range(RF_nat.lim),
      RF_REG.LIM = .format_range(RF_reg.lim),
      POSITION =  as.integer(aliquot.position),
      DATE = aliquot.date,
      SEQUENCE_NAME = aliquot.sequence_name,
      UID = UID,
    row.names = NULL,
    stringsAsFactors = FALSE
  )

    if(!is.null(TP.data.frame)){
      TP.data.frame$UID <- UID
    }

  ##produce results object
  set_RLum(
      class = "RLum.Results",
      data = list(
        data = De.values,
        De.MC = De.MC,
        test_parameters = TP.data.frame,
        fit = fit,
        slide = slide
      ),
      info = list(call = sys.call())
  )
}

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RLum.Results("error"),
               "'object' should be of class 'RLum.Results'")
  expect_error(plot_RLum.Results(set_RLum("RLum.Results",
                                          originator = NA_character_)),
               "'object' has an unsupported originator")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## calc_MinDose
  d1 <- calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                     bootstrap = TRUE, bs.M = 20, bs.N = 10,
                     plot = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d1, main = "Title"))
  expect_silent(plot_RLum.Results(d1, single = FALSE, log = "", lty = 1,
                                  type = "l", col = 2))

  ## calc_CentralDose
  d2 <-calc_CentralDose(ExampleData.DeValues$CA1,
                        plot = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d2))

  ## calc_FuchsLang2001
  d3 <- calc_FuchsLang2001(ExampleData.DeValues$BT998, cvThreshold = 5,
                           plot = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d3))
  expect_silent(plot_RLum.Results(d3, main = "Title", sub = "Subtitle",
                                  xlab = "x", xlim = c(2500, 4000),
                                  ylab = "y", ylim = c(0, 25), mtext = "",
                                  cex = 1, lwd = 1, pch = 1))

  ## calc_FiniteMixture
  d4 <- calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                           n.components = c(2:4),
                           dose.scale = c(0, 100),
                           plot = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d4, pdf.colors = "colors"))
  expect_silent(plot_RLum.Results(d4, main = "Title", plot.proportions = FALSE,
                                  pdf.weight = FALSE, pdf.sigma = "sigmab"))
  expect_silent(plot_RLum.Results(d4, main = "Title", plot.proportions = FALSE,
                                  dose.scale = c(0, 100),
                                  pdf.weight = TRUE, pdf.sigma = "se",
                                  pdf.scale = 1))

  ## calc_AliquotSize
  d5 <- calc_AliquotSize(grain.size = c(100, 150), sample.diameter = 1,
                         MC.iter = 100, plot = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d5))
  expect_silent(plot_RLum.Results(d5, main = "MC simulation", xlab = "Grains"))

  ## calc_SourceDoseRate
  d6 <- calc_SourceDoseRate(measurement = "2012-01-27", calib = "2014-12-19",
                            calib.dose.rate = 0.0438, calib.error = 0.0019)
  expect_silent(plot_RLum.Results(d6))

  ## calc_FastRatio
  data(ExampleData.CW_OSL_Curve, envir = environment())
  d7 <- calc_FastRatio(ExampleData.CW_OSL_Curve, plot = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d7))
  d7 <- calc_FastRatio(ExampleData.CW_OSL_Curve, dead.channels = c(1, 1),
                       plot = FALSE, fitCW.curve = FALSE, verbose = FALSE)
  expect_silent(plot_RLum.Results(d7))
  d7_fit <- calc_FastRatio(ExampleData.CW_OSL_Curve, dead.channels = c(1, 1),
                       plot = FALSE, fitCW.curve = TRUE, verbose = FALSE)
  expect_s4_class(d7_fit@data$fit, "RLum.Results")
  expect_silent(plot_RLum.Results(d7_fit))

  ## analyse_IRSAR.RF
  data(ExampleData.RLum.Analysis, envir = environment())
  d8 <- analyse_IRSAR.RF(IRSAR.RF.Data, method = "VSLIDE", n.MC = 10,
                         plot = FALSE, txtProgressBar = FALSE)
  SW({
  expect_warning(plot_RLum.Results(d8))
  })

  ## no valid originator
  expect_silent(plot_RLum.Results(set_RLum("RLum.Results",
                                           originator = "error")))
})

test_that("check class and length of output", {
  testthat::skip_on_cran()
  local_edition(3)

  set.seed(1)
  data(ExampleData.RLum.Analysis, envir = environment())
  results_fit <- analyse_IRSAR.RF(object = IRSAR.RF.Data, plot = TRUE, method = "FIT")
  results_slide <- suppressWarnings(
    analyse_IRSAR.RF(object = IRSAR.RF.Data, plot = TRUE, method = "SLIDE", n.MC = NULL))
  results_slide_alt <-
    analyse_IRSAR.RF(
      object = IRSAR.RF.Data,
      plot = FALSE,
      method = "SLIDE",
      n.MC = 10,
      method.control = list(vslide_range = 'auto', trace_vslide = TRUE),
      txtProgressBar = FALSE
    )


  expect_equal(is(results_fit), c("RLum.Results", "RLum"))
  expect_equal(length(results_fit), 5)
  expect_equal(length(results_slide), 5)
  expect_s3_class(results_fit$fit, class = "nls")
  expect_s3_class(results_slide$fit, class = "nls")
  expect_length(results_slide$slide, 10)

  expect_equal(results_fit$data$DE, 623.25)
  expect_equal(results_fit$data$DE.LOWER, 600.63)
  expect_equal(results_slide$data$DE, 610.17)
  expect_equal(round(results_slide_alt$data$DE, digits = 0), 384)


})

test_that("test controlled chrash conditions", {
  testthat::skip_on_cran()
  local_edition(3)

  ##the sliding range should not exceed a certain value ... test it
  data(ExampleData.RLum.Analysis, envir = environment())
  expect_error(
    analyse_IRSAR.RF(
      object = IRSAR.RF.Data,
      plot = FALSE,
      method = "SLIDE",
      n.MC = 2,
      method.control = list(vslide_range = c(0,1e+07)),
      txtProgressBar = FALSE
    ), regexp = "[:::src_analyse_IRSAR_SRS()] 'vslide_range' exceeded maximum size (1e+07)!", fixed = TRUE)


})

test_that("test support for IR-RF data", {
  testthat::skip_on_cran()
  local_edition(3)

  ## get needed data
  file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
  temp <- read_RF2R(file)

  expect_s4_class(
    suppressWarnings(
      analyse_IRSAR.RF(object = temp[1:3], method = "SLIDE", plot_reduced = TRUE, n.MC = 1)),
    "RLum.Results")

})

test_that("test edge cases", {
  testthat::skip_on_cran()
  local_edition(3)

  data(ExampleData.RLum.Analysis, envir = environment())
  RF_nat <- RF_reg <- IRSAR.RF.Data[[2]]
  RF_reg@data[,2] <- runif(length(RF_reg@data[,2]), 0.007557956, 0.05377426 )
  RF_nat@data[,2] <- runif(length(RF_nat@data[,2]), 65.4, 76.7)
  RF_nat@data <- RF_nat@data[1:50,]

  expect_s4_class(analyse_IRSAR.RF(
    set_RLum("RLum.Analysis", records = list(RF_nat, RF_reg)),
    method = "SLIDE",
    method.control = list(vslide_range = 'auto', correct_onset = FALSE),
    RF_nat.lim = 2,
    RF_reg.lim = 2,
    plot = FALSE,
    txtProgressBar = FALSE
  ), "RLum.Results")

})

## ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
## Variables and functions loaded when package is attached
## ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Set namespace .LuminescenceEnv ------------------------------------------
.LuminescenceEnv <- new.env(parent = emptyenv())

# Assign variables to Namespace -------------------------------------------
##variable col to define colours in the functions for output
assign("col",
       c("#000000", "#FF3000", "#6495ED", "#458B00", "#FFB90F", "#8B6914",
         "#FF00FF", "#00FA9A", "#EEEE00", "#9ACD32", "#8B0000", "#EEE8AA",
         "#8B7D6B", "#00EE76", "#1E90FF", "#0000FF", "#551A8B", "#7CFC00",
         "#A9A9A9", "#FFFF00", "#CD0000"),
       pos = ".LuminescenceEnv",
       envir = .LuminescenceEnv)

## `fn_stack` is used to keep a stack of function names, managed by
## `.[set|unset]_function_name()`, that reflects the call stack for correct
## error/warning reporting from `.throw_error()` and `.throw_warning()`
assign("fn_stack", list("<>"),
       pos = ".LuminescenceEnv",
       envir = .LuminescenceEnv)

## Numeric constants --------------------------------------------------------
## Values were taken from https://physics.nist.gov/cuu/Constants/
.const <- list(

    ## Speed of light (m/s)
    c = 299792458,

    ## Conversion factor from degree Celsius to Kelvin
    C2K = 273.15,

    ## Planck constant (J*s)
    h = 6.62607015e-34,

    ## Planck constant (eV*s)
    h_eVs = 4.135667696e-15,

    ## Boltzmann constant (eV/K)
    kB = 8.617333262e-05,

    ## Atomic mass constant (kg)
    ma = 1.66053906892e-27,

    ## Avogadro constant (1/mol)
    Na = 6.02214076e23,

    ## year (d)
    year_d = 365.2425,

    ## year (s)
    year_s = 365.2425 * 24 * 60 * 60
)

## binding to allow us to mock readline() during tests (see test_use_DRAC.R)
readline <- NULL

##==============================================================================
##on Attach
.onAttach <- function(libname,pkgname){

  ##set startup message
  try(packageStartupMessage("Welcome to the R package Luminescence version ",
                              packageDescription(pkg="Luminescence")$Version,
                              " [Built: ",
                              trimws(strsplit(packageDescription(pkg="Luminescence")$Built, ";")[[1]][3]),
                            "]\n",
                            get_Quote()), silent=TRUE)
}

##==============================================================================
# DO NOT TOUCH! -----------------------------------------------------------
#' sTeve - sophisticated tool for efficient data validation and evaluation
#'
#' This function provides a sophisticated routine for comprehensive
#' luminescence dating data analysis.
#'
#' This amazing sophisticated function validates your data seriously.
#'
#' @param n_frames [integer] (*with default*):
#' n frames
#'
#' @param t_animation [integer] (*with default*):
#' t animation
#'
#' @param n.tree [integer] (*with default*):
#' how many trees do you want to cut?
#'
#' @param type [integer] (*optional*):
#' Make a decision: 1, 2 or 3
#'
#' @return Validates your data.
#'
#' @note This function should not be taken too seriously.
#'
#' @author R Luminescence Team, 2012-2046
#'
#' @seealso [Luminescence::plot_KDE]
#'
#' @keywords manip
#' @examples
#'
#' ##no example available
#'
#' @export
sTeve <- function(n_frames = 10, t_animation = 2, n.tree = 7, type = NULL) {
  .set_function_name("sTeve")
  on.exit(.unset_function_name(), add = TRUE)

  .validate_positive_scalar(n_frames, int = TRUE)
  .validate_positive_scalar(t_animation, int = TRUE)
  .validate_positive_scalar(n.tree, int = TRUE)

  ## allow new overlay plot
  par(new = TRUE)

  ## infer month of year
  month <- as.numeric(strsplit(x = as.character(Sys.Date()), split = "-")[[1]][2])

  ## select showtime item based on month or user-defined type
  .validate_positive_scalar(type, int = TRUE, null.ok = TRUE)
  if (is.null(type)) {
    type <- as.integer(cut(month, c(0, 3, 11, Inf)))
  }

  if(type == 1) {
    ## SHOWTIME OPTION 1
    Sys.sleep(5)
    shape::emptyplot()
    shape::filledrectangle(wx = 0.9, wy = 0.4,
                    mid = c(0.5, 0.5),
                    lcol ="red",
                    lwd=1,
                    col=0,
                    angle = 45)

    text(x=0.5, y=0.5,
         labels="NOT FUNNY",
         cex=2,
         col="red",
         font=2,
         srt=45)
  } else if(type == 2) {

    ## SHOWTIME OPTION 2
    plot(NA, xlim = c(0, 10),
         ylim = c(0, 10),
         main = "",
         xlab = "",
         ylab = "",
         axes = FALSE,
         frame.plot = FALSE)

    n_frames <- n_frames
    t_animation <- t_animation

    dt <- t_animation / n_frames
    x1 <- seq(0, 10, length.out = n_frames)
    y1 <- rep(1.5, n_frames)
    r1 <- 0.5

    x2 <- seq(0, 16, length.out = n_frames)
    y2 <- rep(8.5, n_frames)
    r2 <- 0.5

    x4 <- seq(11, 0, length.out = n_frames)
    y4 <- rep(5, n_frames)
    r4 <- 0.5

    # set angles for each step of mouth opening
    angles_mouth <- rep_len(c(0.01, 0.25, 0.5, 0.25), n_frames)

    for(i in 1:n_frames){
      # define pacman circles
      shape::filledcircle(r1 = r1,
                   r2 = 0.00001,
                   mid = c(x1[i], y1[i]),
                   from = angles_mouth[i],
                   to = 2 * pi - angles_mouth[i],
                   col = "yellow")
      shape::filledcircle(r1 = r2,
                   r2 = 0.00001,
                   mid = c(x2[i], y2[i]),
                   from = angles_mouth[i],
                   to = 2 * pi - angles_mouth[i],
                   col = "yellow")
      shape::filledcircle(r1 = r4,
                   r2 = 0.00001,
                   mid = c(x4[i], y4[i]),
                   from = angles_mouth[i] + 3,
                   to = 2 * pi - angles_mouth[i] + 3,
                   col = "yellow")

      # define eyes for pacman
      points(x1[i] + 0.2, y1[i] + 0.75, pch = 21, bg = 1, cex = 0.7)
      points(x2[i] + 0.2, y2[i] + 0.75, pch = 21, bg = 1, cex = 0.7)
      points(x4[i] - 0.05, y4[i] + 0.75, pch = 21, bg = 1, cex = 0.7)

      Sys.sleep(dt)

      shape::plotcircle(r = 1.1 * r1,
                 mid = c(x1[i], y1[i]),
                 col = "white",
                 lcol = "white")
      shape::plotcircle(r = 1.1 * r2,
                 mid = c(x2[i], y2[i]),
                 col = "white",
                 lcol = "white")
      shape::plotcircle(r = 1.1 * r4,
                 mid = c(x4[i], y4[i]),
                 col = "white",
                 lcol = "white")
    }
  } else if(type == 3) {
    ## calculate display ratio
    f <- par()$pin[2] / par()$pin[1]

    ## create new overlay plot
    plot(NA,
         xlim = c(0, 100),
         ylim = c(0, 100),
         axes = FALSE,
         frame.plot = FALSE,
         xlab = "",
         ylab = "")

    ## create semi-transparent layer
    polygon(x = c(-100, -100, 200, 200),
            y = c(-100, 200, 200, -100),
            col = rgb(1,1,1, 0.8),
            lty = 0)

    ## draw christmas trees
    n <- n.tree
    tree.x <- runif(n, 10, 90)
    tree.y <- runif(n, 10, 90)
    tree.size <- runif(n, 0.3, 1.5)

    for(i in 1:n) {
      ## stem
      polygon(x = c(tree.x[i] - 1.5 * tree.size[i],
                    tree.x[i] - 1.5 * tree.size[i],
                    tree.x[i] + 1.5 * tree.size[i],
                    tree.x[i] + 1.5 * tree.size[i]) ,
              y = c(tree.y[i] - 12 * tree.size[i],
                    tree.y[i] - 1 * tree.size[i],
                    tree.y[i] - 1 * tree.size[i],
                    tree.y[i] - 12* tree.size[i]),
              col = "rosybrown4",
              lty = 0)

      ## branch one
      shape::filledellipse(rx1 = 10 * tree.size[i],
                    rx2 = 0.00001,
                    mid = c(tree.x[i], tree.y[i] + 3 * tree.size[i]),
                    col = "darkgreen",
                    from = 4.0143,
                    to = 5.41052)

      ## branch two
      shape::filledellipse(rx1 = 8 * tree.size[i],
                    rx2 = 0.00001,
                    mid = c(tree.x[i], tree.y[i] + 7 * tree.size[i]),
                    col = "darkgreen",
                    from = 4.0143,
                    to = 5.41052)

      ## branch three
      shape::filledellipse(rx1 = 6 * tree.size[i],
                    rx2 = 0.00001,
                    mid = c(tree.x[i], tree.y[i] + 9 * tree.size[i]),
                    col = "darkgreen",
                    from = 4.0143,
                    to = 5.41052)

      ## branch four
      shape::filledellipse(rx1 = 4 * tree.size[i],
                    rx2 = 0.00001,
                    mid = c(tree.x[i], tree.y[i] + 11 * tree.size[i]),
                    col = "darkgreen",
                    from = 4.0143,
                    to = 5.41052)

      ## sphere one
      shape::filledellipse(rx1 = 1 * f * tree.size[i],
                    ry1 = 1 * tree.size[i],
                    mid = c(tree.x[i] + 2 * tree.size[i],
                            tree.y[i] + 5 * tree.size[i]),
                    col = shape::shadepalette(n = 20, endcol = "darkred"))

      ## sphere two
      shape::filledellipse(rx1 = 0.8 * f * tree.size[i],
                    ry1 = 0.8 * tree.size[i],
                    mid = c(tree.x[i] - 1 * tree.size[i],
                            tree.y[i] + -3 * tree.size[i]),
                    col = shape::shadepalette(n = 20, endcol = "orange"))

      ## sphere three
      shape::filledellipse(rx1 = 1.2 * f * tree.size[i],
                    ry1 = 1.2 * tree.size[i],
                    mid = c(tree.x[i] - 1.7 * tree.size[i],
                            tree.y[i] + 2 * tree.size[i]),
                    col = shape::shadepalette(n = 20, endcol = "yellow3"))

      ## sphere four
      shape::filledellipse(rx1 = 1 * f * tree.size[i],
                    ry1 = 1 * tree.size[i],
                    mid = c(tree.x[i] + 3 * tree.size[i],
                            tree.y[i] - 4 * tree.size[i]),
                    col = shape::shadepalette(n = 20, endcol = "darkblue"))

      Sys.sleep(0.1)
    }

    ## add snow
    points(runif(300, 0, 100), runif(300, 0, 100), pch = 8, col = "lightgrey")
  }
}#end function

#' Transform a CW-OSL curve into a pLM-OSL curve
#'
#' Transforms a conventionally measured continuous-wave (CW) curve into a
#' pseudo linearly modulated (pLM) curve using the equations given in Bulur
#' (2000).
#'
#' According to Bulur (2000) the curve data are transformed by introducing two
#' new parameters `P` (stimulation period) and `u` (transformed time):
#'
#' \deqn{P=2*max(t)} \deqn{u=\sqrt{(2*t*P)}}
#'
#' The new count values are then calculated by
#' \deqn{ctsNEW = cts(u/P)}
#'
#' and the returned `data.frame` is produced by: `data.frame(u,ctsNEW)`
#'
#' The output of the function can be further used for LM-OSL fitting.
#'
#' @param values [RLum.Data.Curve-class] or [data.frame] (**required**):
#' `RLum.Data.Curve` data object. Alternatively, a `data.frame` of the measured
#' curve data of type stimulation time (t) (`values[,1]`) and measured counts (cts)
#' (`values[,2]`) can be provided.
#'
#' @return
#' The function returns the same data type as the input data type with
#' the transformed curve values ([data.frame] or [RLum.Data.Curve-class]).
#'
#' @note
#' The transformation is recommended for curves recorded with a channel
#' resolution of at least 0.05 s/channel.
#'
#' @section Function version: 0.4.1
#'
#' @author
#' Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [CW2pHMi], [CW2pLMi], [CW2pPMi], [fit_LMCurve], [lm],
#' [RLum.Data.Curve-class]
#'
#'
#' @references
#' Bulur, E., 2000. A simple transformation for converting CW-OSL
#' curves to LM-OSL curves. Radiation Measurements, 32, 141-145.
#'
#' **Further Reading**
#'
#' Bulur, E., 1996. An Alternative Technique For Optically Stimulated
#' Luminescence (OSL) Experiment. Radiation Measurements, 26, 701-709.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##read curve from CWOSL.SAR.Data transform curve and plot values
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##read id for the 1st OSL curve
#' id.OSL <- CWOSL.SAR.Data@@METADATA[CWOSL.SAR.Data@@METADATA[,"LTYPE"] == "OSL","ID"]
#'
#' ##produce x and y (time and count data for the data set)
#' x<-seq(CWOSL.SAR.Data@@METADATA[id.OSL[1],"HIGH"]/CWOSL.SAR.Data@@METADATA[id.OSL[1],"NPOINTS"],
#'        CWOSL.SAR.Data@@METADATA[id.OSL[1],"HIGH"],
#'        by = CWOSL.SAR.Data@@METADATA[id.OSL[1],"HIGH"]/CWOSL.SAR.Data@@METADATA[id.OSL[1],"NPOINTS"])
#' y <- unlist(CWOSL.SAR.Data@@DATA[id.OSL[1]])
#' values <- data.frame(x,y)
#'
#' ##transform values
#' values.transformed <- CW2pLM(values)
#'
#' ##plot
#' plot(values.transformed)
#'
#' @md
#' @export
CW2pLM <- function(
  values
){

  # Integrity Checks --------------------------------------------------------

  ##(1) data.frame or RLum.Data.Curve object?
  if(is(values, "data.frame") == FALSE & is(values, "RLum.Data.Curve") == FALSE){
    stop("[CW2pLM()] 'values' object has to be of type 'data.frame' or 'RLum.Data.Curve'!", call. = FALSE)

  }

  ##(2) if the input object is an 'RLum.Data.Curve' object check for allowed curves
  if(is(values, "RLum.Data.Curve") == TRUE){

    if(!grepl("OSL", values@recordType) & !grepl("IRSL", values@recordType)){

      stop(paste("[CW2pLM()] recordType ",values@recordType, " is not allowed for the transformation!",
                 sep=""), call. = FALSE)

    }else{

      temp.values <- as(values, "data.frame")

    }

  }else{

    temp.values <- values


  }


  # Calculation -------------------------------------------------------------


  ##curve transformation
  P<-2*max(temp.values[,1])
  u<-((2*temp.values[,1]*P)^0.5)

  ##cw >> plm conversion, according Bulur, 2000
  temp.values[,2]<-temp.values[,2]*(u/P)
  temp.values<-data.frame(u,temp.values[,2])


  # Return values -----------------------------------------------------------

  ##returns the same data type as the input

  if(is(values, "data.frame") == TRUE){

    values <- temp.values
    return(values)

  }else{

    newRLumDataCurves.CW2pLM <- set_RLum(
      class = "RLum.Data.Curve",
      recordType = values@recordType,
                                                    data = as.matrix(temp.values),
                                                    info = values@info)
    return(newRLumDataCurves.CW2pLM)

  }

}

##load data
data(ExampleData.CW_OSL_Curve, envir = environment())
values <- CW_Curve.BosWallinga2012

test_that("Check the example and the numerical values", {
  testthat::skip_on_cran()
  local_edition(3)

  values_pLM <- CW2pLM(values)
  values_pLMi <- CW2pLMi(values, P = 1/20)
  values_pLMi_alt <- CW2pLMi(values)
  values_pHMi <- suppressWarnings(CW2pHMi(values, delta = 40))
  values_pHMi_alt <- suppressWarnings(CW2pHMi(values))
  values_pHMi_alt1 <- suppressWarnings(CW2pHMi(values, delta = 2))
  values_pPMi <- suppressWarnings(CW2pPMi(values, P = 1/10))

    ##check conversion sum values
    expect_equal(round(sum(values_pLM), digits = 0),90089)
    expect_equal(round(sum(values_pLMi[,1:2]), digits = 0),197522)
    expect_equal(round(sum(values_pLMi_alt[,1:2]), digits = 0),197522)
    expect_equal(round(sum(values_pHMi[,1:2]), digits = 0),217431)
    expect_equal(round(sum(values_pHMi_alt[,1:2]), digits = 0),217519)
    expect_equal(round(sum(values_pHMi_alt1[,1:2]), digits = 0), 221083)
    expect_equal(round(sum(values_pPMi[,1:2]), digits = 0),196150)


})

test_that("Test RLum.Types", {
  testthat::skip_on_cran()
  local_edition(3)

  ##load CW-OSL curve data
  data(ExampleData.CW_OSL_Curve, envir = environment())
  object <-
    set_RLum(
      class = "RLum.Data.Curve",
      data = as.matrix(ExampleData.CW_OSL_Curve),
      curveType = "measured",
      recordType = "OSL"
    )


  ##transform values
  expect_s4_class(CW2pLM(object), class = "RLum.Data.Curve")
  expect_s4_class(CW2pLMi(object), class = "RLum.Data.Curve")
  expect_s4_class(CW2pHMi(object), class = "RLum.Data.Curve")
  expect_s4_class(suppressWarnings(CW2pPMi(object)), class = "RLum.Data.Curve")

  ##test error handling
  expect_error(CW2pLMi(values, P = 0), regexp = "[CW2pLMi] P has to be > 0!", fixed = TRUE)
  expect_warning(CW2pLMi(values, P = 10))
  expect_error(object = CW2pLM(values = matrix(0, 2)))
  expect_error(object = CW2pLMi(values = matrix(0, 2)))
  expect_error(object = CW2pHMi(values = matrix(0, 2)))
  expect_error(object = CW2pPMi(values = matrix(0, 2)))

  object@recordType <- "RF"
  expect_error(CW2pLM(values = object),
               "recordType RF is not allowed for the transformation")
  expect_error(object = CW2pLMi(values = object),
               regexp = "[CW2pLMi()] recordType RF is not allowed for the transformation!",
               fixed = TRUE)
  expect_error(object = CW2pHMi(values = object),
               regexp = "[CW2pHMi()] recordType RF is not allowed for the transformation!",
               fixed = TRUE)
  expect_error(object = CW2pPMi(values = object),
               regexp = "[CW2pPMi()] recordType RF is not allowed for the transformation!",
               fixed = TRUE)


})

#'@title Extract Pixel Values through Circular Region-of-Interests (ROI) from an Image
#'
#'@description Light-weighted function to extract pixel values from pre-defined regions-of-interest (ROI) from
#'[Luminescence::RLum.Data.Image-class], [array] or [matrix] objects and provide simple image processing
#'capacity. The function is limited to circular ROIs.
#'
#'@details The function uses a cheap approach to decide whether a pixel lies within
#'a circle or not. It assumes that pixel coordinates are integer values and
#'that a pixel centring within the circle is satisfied by:
#'
#'\deqn{x^2 + y^2 <= (d/2)^2}
#'
#'where \eqn{x} and \eqn{y} are integer pixel coordinates and \eqn{d} is the integer
#'diameter of the circle in pixel.
#'
#'@param object [Luminescence::RLum.Data.Image-class], [array] or [matrix] (**required**): input image data
#'
#'@param roi [matrix] (**required**): matrix with three columns containing the centre coordinates
#'of the ROI (first two columns) and the diameter of the circular ROI. All numbers must by of type [integer]
#'and will forcefully coerced into such numbers using `as.integer()` regardless.
#'
#'@param roi_summary (*with default*): defines what is returned in the
#'`roi_summary` element; it can be `"mean"` (default), `"median"`, `"sd"` or
#'`"sum"`.
#'Pixel values are conveniently summarised using the above defined keyword.
#'
#' @param plot [logical] (*optional*): enable/disable the plot output. Only
#' the first image frame is shown.
#'
#'@return [Luminescence::RLum.Results-class] object with the following elements:
#'`..$roi_signals`: a named [list] with all ROI values and their coordinates
#'`..$roi_summary`: a [matrix] where rows are frames from the image, and columns are different ROI
#'The element has two attributes: `summary` (the method used to summarise pixels) and `area` (the pixel area)
#'`..$roi_coord`: a [matrix] that can be passed to [Luminescence::plot_ROI]
#'
#'If `plot = TRUE` a control plot is returned.
#'
#'@section Function version: 0.1.0
#'
#'@author
#'Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@seealso [Luminescence::RLum.Data.Image-class]
#'
#'@keywords manip
#'
#'@examples
#'
#' m <- matrix(runif(100,0,255), ncol = 10, nrow = 10)
#' roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
#' extract_ROI(object = m, roi = roi, plot = TRUE)
#'
#'@export
extract_ROI <- function(
  object,
  roi,
  roi_summary = c("mean", "median", "sd", "sum"),
  plot = FALSE
) {
  .set_function_name("extract_ROI")
  on.exit(.unset_function_name(), add = TRUE)

# Self call ---------------------------------------------------------------
  if (inherits(object, "list"))
    return(merge_RLum(lapply(object, extract_ROI, roi = roi, plot = plot)))

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Data.Image", "matrix", "array"),
                  extra = "a 'list' of such objects")

  ## check input for ROIs
  .validate_class(roi, "matrix")
  if (nrow(roi) < 1 || ncol(roi) < 3)
    .throw_error("'roi' does not have the expected format")

  ## calculate the radius
  roi <- roi[,1:3]
  roi[,3] <- ceiling(roi[,3]/2)

  ## make sure that we have integer values only in the matrix
  roi[] <- as.integer(roi)

  ## copy object (to not work on the input data)
  a <- object

  ## try to convert into something meaningful
  if (inherits(object, "RLum.Data.Image"))
    a <- object@data

  if (is.matrix(object))
    a <- array(data = object, dim = c(nrow(object), ncol(object), 1))

# Helper function ---------------------------------------------------------
  .extract_pixel <- function(m, r, mid) {
    ## get row - column combinations
    grid <- as.matrix(expand.grid(x = 1:nrow(m), y = 1:ncol(m)))

    ## adjust values for mid point
    ## + get pixel coordinates if within the circle
    px_id <- grid[(grid[,"x"] - mid[1])^2 + (grid[,"y"]  - mid[2])^2 <= r[1]^2,]

    ## extract values from matrix
    px_extract <- NA
    if (nrow(px_id) > 0) {
      px_extract <- vapply(1:nrow(px_id), function(x) {
        m[px_id[x,1],px_id[x,2]]

      }, numeric(1))
    }

    attr(px_extract, "coord") <- px_id
    return(px_extract)
  }

# Extract ROIs ------------------------------------------------------------
   roi_signals <- lapply(1:nrow(roi), function(x){
    ## iterate through a stack if needed
    temp <- lapply(1:(dim(a)[3]), function(z){
      .extract_pixel(a[,,z], roi[x,3], mid = roi[x,1:2])
    })

    ## compile into matrix
    m <- matrix(unlist(temp), nrow = length(temp[[1]]))

    ## add attributes ... including coordinates; but only one time
    colnames(m) <- paste0("frame_", 1:ncol(m))
    attr(m, "px_coord") <- attr(temp[[1]], "coord")
    return(m)
  })

  ## add names
  names(roi_signals) <- paste0("ROI_", 1:nrow(roi))

# Plot check --------------------------------------------------------------
  if (plot) {
    ## this is a control plot, so we plot only the first image; nothing more

    ## image
    graphics::image(
      x = 1:nrow(a[, , 1]),
      y = 1:ncol(a[, , 1]),
      a[, , 1],
      ylab = "y-dim [px]",
      xlab = "x-dim [px]",
      useRaster = TRUE,
      main = "extract_ROIs() - control plot")
    graphics::box()

    ## visualise ROIs
    overlay <- a[,,1]
    overlay[] <- 0
    for (i in 1:length(roi_signals))
      overlay[attr(roi_signals[[i]], "px_coord")[,1], attr(roi_signals[[i]], "px_coord")[,2]] <- 1

    ## marked ROIs
    graphics::image(
      x = 1:nrow(a[, , 1]),
      y = 1:ncol(a[, , 1]),
      overlay,
      axes = FALSE,
      add = TRUE,
      useRaster = TRUE,
      col = c(rgb(1, 1, 1, 0), rgb(0, 1, 0, 0.5)))

    ## add circles and points
    for (i in 1:nrow(roi)) {
      lines(shape::getellipse(rx = roi[i, 3], mid = c(roi[i, 1:2], dr = 0.1)), col = "red", lwd = 1.5)
      text(x = roi[i,1], y = roi[i,2], i, col = "black", cex = 1.2)
    }
  }

# ROI summary -------------------------------------------------------------
  ## set roi fun and avoid add input
  .validate_class(roi_summary, "character")
  roi_fun <- .validate_args(roi_summary, c("mean", "median", "sd", "sum"))

  ## create summary using matrixStats
  roi_summary <- matrix(unlist(
    switch(roi_fun,
      mean = lapply(roi_signals, matrixStats::colMeans2),
      median = lapply(roi_signals, matrixStats::colMedians),
      sd = lapply(roi_signals, matrixStats::colSds),
      sum = lapply(roi_signals, matrixStats::colSums2))),
    ncol = length(roi_signals))

  ## set names to make it easier
  colnames(roi_summary) <- names(roi_signals)
  rownames(roi_summary) <- paste0("frame_", 1:nrow(roi_summary))
  attr(roi_summary, "summary") <- roi_fun
  attr(roi_summary, "area") <- vapply(roi_signals, nrow, numeric(1))

  ## add more roi information to the output for further processing
  roi <- cbind(
    ROI = 1:nrow(roi),
    x = roi[,1],
    y = roi[,2],
    area = attr(roi_summary, "area"),
    width = vapply(roi_signals, function(x) diff(range(attr(x, "px_coord")[,"x"])), numeric(1)),
    height = vapply(roi_signals, function(x) diff(range(attr(x, "px_coord")[,"y"])), numeric(1)),
    img_width = nrow(a[, , 1]),
    img_height = ncol(a[, , 1]),
    grain_d = roi[,3])

# Return ------------------------------------------------------------------
  set_RLum(
      class = "RLum.Results",
      data = list(
        roi_signals = roi_signals,
        roi_summary = roi_summary,
        roi_coord = roi),
      info = list(
          call = sys.call()))
}

## load dataset
temp <- read_RF2R(system.file("extdata", "RF_file.rf", package = "Luminescence"),
                  verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_ROI(object = "stop"),
               "[plot_ROI()] 'object' should be of class 'RLum.Analysis'",
               fixed = TRUE)

  empty <- set_RLum("RLum.Results", originator = NA_character_)
  expect_error(plot_ROI(empty),
               "Object originator 'NA' not supported")
  expect_error(plot_ROI(list()),
               "'object' cannot be an empty list")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##test standard cases
  expect_silent(plot_ROI(temp))
  expect_silent(plot_ROI(temp, grid = TRUE))
  expect_silent(plot_ROI(temp, dim.CCD = c(8192,8192)))
  expect_silent(plot_ROI(temp, dist_thre = 20))
  expect_silent(plot_ROI(temp, exclude_ROI = NULL))

  ##test non-list case
  expect_silent(plot_ROI(temp[[1]]))
  expect_silent(plot_ROI(temp[[1]], exclude_ROI = NULL))

  ##output only case
  expect_s4_class(plot_ROI(temp, plot = FALSE), class = "RLum.Results")

  ## test combination with extract_ROI()
  m <- matrix(runif(100,0,255), ncol = 10, nrow = 10)
  roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
  t <- extract_ROI(object = m, roi = roi)
  expect_s4_class(plot_ROI(t, bg_image = m, exclude_ROI = NULL), "RLum.Results")

  ## trigger warning
  expect_warning(plot_ROI(t, bg_image = "stop", exclude_ROI = NULL),
                 "[plot_ROI()] 'bg_image' is not of class 'RLum.Data.Image'",
                 fixed = TRUE)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("ROI defaults",
                              plot_ROI(temp))
  vdiffr::expect_doppelganger("ROI grid",
                              plot_ROI(temp, grid = TRUE))

  set.seed(1)
  m <- matrix(runif(100, 0, 255), ncol = 10, nrow = 10)
  roi <- matrix(c(2, 4, 2, 5, 6, 7, 3, 1, 1), ncol = 3)
  t <- extract_ROI(object = m, roi = roi)
  vdiffr::expect_doppelganger("ROI bg_image",
                              plot_ROI(t, bg_image = m, exclude_ROI = NULL))
  })
})

test_that("Merge RLum.Results", {
  testthat::skip_on_cran()
  local_edition(3)

  ## check whether arguments are retained
  a <- array(runif(300, 0,255), c(10,10,3))
  roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
  t <-
    expect_s4_class(merge_RLum.Results(lapply(list(roi, roi, roi), function(x)
      extract_ROI(a, x))), "RLum.Results")

  expect_length(names(attributes(t@data$roi_summary)), 4)

  ## check standard without attributes to make sure that
  ## standard attributes become not removed
  ##load example data
  data(ExampleData.DeValues, envir = environment())

  ##apply the central dose model
  c <- calc_CentralDose(ExampleData.DeValues$CA1, plot = FALSE)
  a <- merge_RLum.Results(list(c,c))
  expect_s3_class(a@data$summary, "data.frame")


})

## load data
data(ExampleData.DeValues, envir = environment())
res <- calc_CentralDose(ExampleData.DeValues$CA1,
                        plot = FALSE, verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(merge_RLum.Results("error"),
               "'objects' should be of class 'list'")
  expect_error(merge_RLum.Results(list(res, "error")),
               "All elements of 'object' should be of class 'RLum.Results'")
  expect_message(expect_null(merge_RLum.Results(list())),
                 "'objects' contains no data, NULL returned")

  res2 <- res
  res2@originator <- "unknown"
  expect_error(merge_RLum.Results(list(res, res2)),
               "Objects cannot be merged, different originators found: 'calc_CentralDose', 'unknown'")

  res2 <- res
  res2@data[[1]][, 2] <- NULL
    expect_error(merge_RLum.Results(list(res, res2)),
               "Objects cannot be merged, different number of columns")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  set.seed(1)
  a <- array(runif(300, 0,255), c(10,10,3))
  roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
  expect_snapshot_RLum(merge_RLum.Results(lapply(list(roi, roi, roi),
                                                 function(x) extract_ROI(a, x))))

  expect_snapshot_RLum(merge_RLum.Results(list(res, res)))

  empty <- set_RLum("RLum.Results")
  expect_snapshot_RLum(merge_RLum.Results(list(empty)))
  expect_s4_class(merge_RLum.Results(list(empty, empty)),
                  "RLum.Results")
})

##################################################################################
##                     Luminescence - RStudio Add Ins                           ##
##################################################################################
##<<Prerequisite>>
##
## - Add-ins should support more experienced users. For all others we have the package 'RLumShiny'
##
## - Add-ins should be provided as non-exported function only, having the a name with a leading dot, e.g., .addin.
##   This prevents further chaos in the manuel.
##
## - Interactive add-ins are not desired, except they are implemented in the package 'RLumShiny' or they
##   are only available if the package 'RLumShiny' is available.
##
##
##<<FAQ>>
##
## Q. Why are the add-ins non-interactive ... had you been too lazy?
## A. No, but interactivity would require the installation of 'shiny" by default, which is not
##    desired.
##
## Q. The add-ins are not shown in the 'Addin' menu?
## A. Well, if you read this information you are an advanced used, so please install the
##    package 'rstudioapi', 'devtools' and get happy.


#'Search for TODOs in the source code and list them in the terminal
#'
#'This add-in is a tool developers may want to use to briefly list all open
#'issues in the terminal, instead of using search and stepping through the results.
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@noRd
.listTODO <- function(){
  # nocov start
  ##check if package is installed
  if(!requireNamespace("rstudioapi", quietly = TRUE)){
    message("Package 'rstudioapi' is not installed but needed to search for TODOs, do you want to install it?\n\n",
            " [n/N]: No (default)\n",
            " [y/Y]: Yes\n")

    ##parse answer
    answer <- readline()
    if (tolower(answer) != "y")
      return(invisible())

    utils::install.packages("rstudioapi", dependencies = TRUE)
  }

  ##parse code
  code <- rstudioapi::getActiveDocumentContext()$contents

  ##get lines with ##TODO
  id <- grep(pattern = "#\\s*TODO", x = code, fixed = FALSE)

  ##list lines
  cat("\n", "[", length(id), " issue(s)]\n", sep = "")
  for (i in id) {
    cat(" line ", i, ": ->", code[i], "\n", sep = "")
  }
  # nocov end
}

#' @title Convert an element from a Risoe.BINfileData object to an RLum.Data.Curve
#' object
#'
#' @description
#' The function converts one specified single record from a Risoe.BINfileData
#' object to an RLum.Data.Curve object.
#'
#' @details
#' The function extracts all `METADATA` from the `Risoe.BINfileData`
#' object and stores them in the [Luminescence::RLum.Data.Curve-class] object. This function
#' can be used stand-alone, but is the base function for [Luminescence::Risoe.BINfileData2RLum.Analysis].
#'
#' @param object [Luminescence::Risoe.BINfileData-class] (**required**):
#' `Risoe.BINfileData` object
#'
#' @param id [integer] (**required**):
#' record id in the `Risoe.BINfileData` object of the curve that is to be
#' stored in the [Luminescence::RLum.Data.Curve-class] object. If no value for id is provided,
#' the record has to be specified by `pos`, `set` and `run`.
#'
#' @param pos [integer] (*optional*):
#' record position number in the `Risoe.BINfileData` object of the curve that
#' is to be stored in the [Luminescence::RLum.Data.Curve-class] object. If a value for `id` is
#' provided, this argument is ignored.
#'
#' @param run [integer] (*optional*):
#' record run number in the `Risoe.BINfileData` object of the curve that is
#' to be stored in the [Luminescence::RLum.Data.Curve-class] object. If a value for `id` is
#' provided, this argument is ignored.
#'
#' @param set [integer] (*optional*):
#' record set number in the `Risoe.BINfileData` object of the curve that is
#' to be stored in the [Luminescence::RLum.Data.Curve-class] object. If a value for `id` is
#' provided, this argument is ignored.
#'
#' @return Returns an [Luminescence::RLum.Data.Curve-class] object.
#'
#' @note
#' Due to changes in the BIN-file (version 3 to version 4) format the recalculation of TL-curves might be not
#' overall correct for cases where the TL measurement is combined with a preheat.
#'
#' @section Function version: 0.5.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Christoph Burow, Universtiy of Cologne (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData2RLum.Analysis], [Luminescence::set_RLum],
#' [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::Risoe.BINfileData-class],[Luminescence::plot_RLum]
#'
#' @keywords manip
#'
#' @examples
#'
#' ##get package example data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##convert one record
#' Risoe.BINfileData2RLum.Data.Curve(CWOSL.SAR.Data, id = 1)
#'
#' @noRd
.Risoe.BINfileData2RLum.Data.Curve <- function(
  object,
  id,
  pos,
  run,
  set
){

  ##disaggregate object ... this makes it much faster below
  ##we could also access via index, not number, but this is far to risky, as
  ##every update in the BIN-file version will break the code here
  METADATA <- as.list(object@METADATA)
  DATA <- object@DATA

  # grep id of record -------------------------------------------------------
  ##if id is set, no input for pos and run is necessary
  if (missing(id)) {
    id <- METADATA$ID[METADATA[["POSITION"]] == pos &
                      METADATA[["SET"]] == set &
                      METADATA[["RUN"]] == run]
  }

  ##grep info elements
  info <- lapply(METADATA, function(x) x[id])
  names(info) <- names(METADATA)

  # Build object ------------------------------------------------------------
  set_RLum(
    class = "RLum.Data.Curve",
    recordType = paste(METADATA[["LTYPE"]][id], "(PMT)"),
    data =  src_create_RLumDataCurve_matrix(
      DATA = DATA[[id]],
      NPOINTS = METADATA[["NPOINTS"]][id],
      VERSION = METADATA[["VERSION"]][id],
      LTYPE = METADATA[["LTYPE"]][id],
      LOW =  METADATA[["LOW"]][id],
      HIGH =  METADATA[["HIGH"]][id],
      AN_TEMP = METADATA[["AN_TEMP"]][id],
      TOLDELAY = METADATA[["TOLDELAY"]][id],
      TOLON = METADATA[["TOLON"]][id],
      TOLOFF = METADATA[["TOLOFF"]][id]
    ),
    info = info
  )
}

## load data
data(ExampleData.BINfileData, envir = environment())

test_that("check functionality", {
  testthat::skip_on_cran()

  res <- .Risoe.BINfileData2RLum.Data.Curve(CWOSL.SAR.Data, id = 1)
  expect_s4_class(res, "RLum.Data.Curve")
  expect_length(res@data, 500)
  expect_length(res@info, 44)
  expect_equal(res@originator, ".Risoe.BINfileData2RLum.Data.Curve")

  res1 <- .Risoe.BINfileData2RLum.Data.Curve(CWOSL.SAR.Data,
                                             pos = 1, set = 2, run = 1)
  expect_s4_class(res1, "RLum.Data.Curve")
  expect_equal(res1@data, res@data)
  expect_equal(res1@info, res@info)
  expect_equal(res1@originator, res@originator)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  SW({
  expect_snapshot_RLum(.Risoe.BINfileData2RLum.Data.Curve(CWOSL.SAR.Data,
                                                          pos = 1, set = 2, run = 1))
  })
})

#' @title Estimate the amount of grains on an aliquot
#'
#' @description
#' This function can be used to either estimate the number of grains on an
#' aliquot or to compute the packing density, depending on the arguments
#' provided.
#'
#' The following formula is used to estimate the number of grains `n`:
#'
#' \deqn{n = (\pi*x^2)/(\pi*y^2)*d}
#'
#' where `x` is the radius of the aliquot size (microns), `y` is the mean
#' radius of the mineral grains (mm), and `d` is the packing density
#' (a value between 0 and 1).
#'
#' **Packing density**
#'
#' The default value for `packing.density` is 0.65, which is the mean of
#' empirical values determined by Heer et al. (2012) and unpublished data from
#' the Cologne luminescence laboratory. If `packing.density = Inf`, a maximum
#' density of \eqn{\pi / \sqrt{12} \approx 0.9069} is used. However, note that
#' this value is not appropriate as the standard preparation procedure of
#' aliquots resembles a PECC (*Packing Equal Circles in a Circle*) problem,
#' where the maximum packing density is asymptotic to about 0.87.
#'
#' **Monte Carlo simulation**
#'
#' The number of grains on an aliquot can be estimated by Monte Carlo simulation
#' when setting `MC = TRUE`. All parameters necessary to calculate
#' `n` (`x`, `y`, `d`) are assumed to be normally distributed with means
#' \eqn{\mu_x, \mu_y, \mu_d} and standard deviations \eqn{\sigma_x, \sigma_y, \sigma_d}.
#'
#' For the mean grain size, random samples are taken first from
#' \eqn{N(\mu_y, \sigma_y)}, where \eqn{\mu_y = mean.grain.size} and
#' \eqn{\sigma_y = (max.grain.size-min.grain.size)/4}, so that 95% of all
#' grains are within the provided the grain size range. This effectively takes
#' into account that after sieving the sample there is still a small chance of
#' having grains smaller or larger than the used mesh sizes. For each random
#' sample the mean grain size is calculated, from which random subsamples are
#' drawn for the Monte Carlo simulation.
#'
#' The packing density is assumed
#' to be normally distributed with an empirically determined \eqn{\mu = 0.65}
#' (or provided value) and \eqn{\sigma = 0.18}. The normal distribution is
#' truncated at `d = 0.87` as this is approximately the maximum packing
#' density that can be achieved in a PECC problem.
#'
#' The sample diameter has
#' \eqn{\mu = sample.diameter} and \eqn{\sigma = 0.2} to take into account
#' variations in sample disc preparation (i.e. applying silicon spray to the
#' disc). A lower truncation point at `x = 0.5` is used, which assumes
#' that aliquots with sample diameter smaller than 0.5 mm are discarded.
#' Likewise, the normal distribution is truncated at the diameter of the
#' sample carrier (9.8 mm by default, but controllable via the
#' `sample_carrier.diameter` argument).
#'
#' For each random sample drawn from the
#' normal distribution, the amount of grains on the aliquot is calculated. By
#' default, `10^4` iterations are used, but the can be controlled with
#' option `MC.iter` (see `...`). Results are visualised in a bar- and
#' boxplot together with a statistical summary.
#'
#' @param grain.size [numeric] (**required**):
#' mean grain size (microns) or a range of grain sizes from which the
#' mean grain size is computed (e.g. `c(100,200)`).
#'
#' @param sample.diameter [numeric] (**required**):
#' diameter (mm) of the targeted area on the sample carrier.
#'
#' @param packing.density [numeric] (*with default*):
#' empirical value for the mean packing density. \cr
#' If `packing.density = Inf`, a hexagonal structure on an infinite
#' plane with a packing density of \eqn{\pi / \sqrt{12} \approx 0.9069}
#' is assumed.
#'
#' @param MC [logical] (*optional*):
#' if `TRUE` the function performs a Monte Carlo simulation for estimating the
#' amount of grains on the sample carrier and assumes random errors in grain
#' size distribution and packing density. Requires `grain.size` to be specified
#' as a 2-element vector with the range (min and max) of grain sizes.
#' For more information see details.
#'
#' @param grains.counted [numeric] (*optional*):
#' grains counted on a sample carrier. If a non-zero positive integer is
#' provided, this function will calculate the packing density of the aliquot.
#' If more than one value is provided, the mean packing density and its
#' standard deviation are calculated.
#' Note that this overrides `packing.density`.
#'
#' @param sample_carrier.diameter [numeric] (*with default*):
#' diameter (mm) of the sample carrier.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments and graphical parameters to be passed. In
#' particular, `MC.iter` to control the number of Monte Carlo iterations,
#' `main`, `xlab`, `col`, `line_col`, `line_lwd`, `cex`, `rug` (`TRUE/FALSE`),
#' `boxplot` (`TRUE/FALSE`), `summary` (`TRUE/FALSE`), `legend` (`TRUE/FALSE`).
#'
#' @return
#' Returns a terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following element:
#'
#' \item{.$summary}{[data.frame] summary of all relevant calculation results.}
#' \item{.$args}{[list] used arguments}
#' \item{.$call}{[call] the function call}
#' \item{.$MC}{[list] results of the Monte Carlo simulation}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @section Function version: 0.33
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @references
#' Duller, G.A.T., 2008. Single-grain optical dating of Quaternary
#' sediments: why aliquot size matters in luminescence dating. Boreas 37,
#' 589-612.
#'
#' Heer, A.J., Adamiec, G., Moska, P., 2012. How many grains
#' are there on a single aliquot?. Ancient TL 30, 9-16.
#'
#' **Further reading**
#'
#' Chang, H.-C., Wang, L.-C., 2010. A simple proof of Thue's
#' Theorem on Circle Packing. [https://arxiv.org/pdf/1009.4322v1](),
#' 2013-09-13.
#'
#' Graham, R.L., Lubachevsky, B.D., Nurmela, K.J.,
#' Oestergard, P.R.J., 1998.  Dense packings of congruent circles in a circle.
#' Discrete Mathematics 181, 139-154.
#'
#' Huang, W., Ye, T., 2011. Global
#' optimization method for finding dense packings of equal circles in a circle.
#' European Journal of Operational Research 210, 474-481.
#'
#' @examples
#'
#' ## Estimate the amount of grains on a small aliquot
#' calc_AliquotSize(grain.size = c(100,150), sample.diameter = 1, MC.iter = 100)
#'
#' ## Calculate the mean packing density of large aliquots
#' calc_AliquotSize(grain.size = c(100,200), sample.diameter = 8,
#'                  grains.counted = c(2525,2312,2880), MC.iter = 100)
#'
#' @export
calc_AliquotSize <- function(
  grain.size,
  sample.diameter,
  packing.density = 0.65,
  MC = TRUE,
  grains.counted = NULL,
  sample_carrier.diameter = 9.8,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_AliquotSize")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(grain.size, c("numeric", "integer"))
  if (length(grain.size) == 0 || length(grain.size) > 2) {
    .throw_error("Please provide the mean grain size or the range ",
                 "of grain sizes (in microns)")
  }
  if (anyNA(grain.size) || any(grain.size <= 0)) {
    .throw_error("'grain.size' should contain positive values")
  }

  .validate_positive_scalar(packing.density)
  if (is.infinite(packing.density)) {
    # use ~0.907... from Thue's Theorem as packing density
    packing.density <- pi / sqrt(12)
  }
  if (packing.density > 1) {
    .throw_error("'packing.density' should be a value between 0 and 1")
  }

  .validate_positive_scalar(sample.diameter)
  .validate_positive_scalar(sample_carrier.diameter)
  if (sample.diameter > sample_carrier.diameter)
    .throw_warning("A sample diameter of ", sample.diameter,  " mm was ",
                   "specified for a sample carrier of ", sample_carrier.diameter,
                   " mm, values will be capped to the sample carrier size")

  .validate_logical_scalar(MC)
  if (!is.null(grains.counted) && MC) {
    MC <- FALSE
    .throw_message("Monte Carlo simulation is only available for estimating the ",
                   "amount of grains on the sample disc, 'MC' reset to FALSE",
                   error = FALSE)
  }

  if (MC && length(grain.size) != 2) {
    .throw_error("'grain.size' must be a vector containing the min and max ",
                 "grain size when using Monte Carlo simulations")
  }

  ##==========================================================================##
  ## ... ARGUMENTS
  ##==========================================================================##

  # set default parameters
  settings <- list(MC.iter = 10^4,
                   verbose = TRUE)

  # override settings with user arguments
  settings <- modifyList(settings, list(...))
  if (MC) {
    .validate_positive_scalar(settings$MC.iter, int = TRUE, null.ok = TRUE,
                              name = "'MC.iter'")
  }

  MC.n <- MC.stats <- MC.n.kde <- MC.t.test <- MC.q <- NULL

  ##==========================================================================##
  ## CALCULATIONS
  ##==========================================================================##

  # calculate the mean grain size
  range.flag<- FALSE
  if(length(grain.size) == 2) {
    gs.range <- sort(grain.size)
    grain.size<- mean(grain.size)
    range.flag<- TRUE
  }

  # function to calculate the amount of grains
  ## original formula: n <- ((pi * (sd / 2)^2) / (pi * (gs / 2000)^2)) * d
  calc_n <- function(sd, gs, d) {
    (1000 * sd / gs)^2 * d
  }

  # calculate the amount of grains on the aliquot
  if (is.null(grains.counted)) {
    n.grains<- calc_n(sample.diameter, grain.size, packing.density)

    ##========================================================================##
    ## MONTE CARLO SIMULATION

    if (MC && range.flag) {

      # create a random set of packing densities assuming a normal
      # distribution with the empirically determined standard deviation of
      # 0.18.
      d.mc<- rnorm(settings$MC.iter, packing.density, 0.18)

      # in a PECC the packing density cannot be larger than ~0.87
      d.mc <- pmin(d.mc, 0.87)
      d.mc <- pmax(d.mc, 0.25)

      # create a random set of sample diameters assuming a normal
      # distribution with an assumed standard deviation of
      # 0.2. For a more conservative estimate this is divided by 2.
      sd.mc<- rnorm(settings$MC.iter, sample.diameter, 0.2)

      # it is assumed that sample diameters < 0.5 mm either do not
      # occur, or are discarded. Either way, any smaller sample
      # diameter is capped at 0.5.
      # Also, the sample diameter can not be larger than the sample
      # disc.
      sd.mc <- pmax(sd.mc, 0.5)
      sd.mc <- pmin(sd.mc, sample_carrier.diameter)

      # create random samples assuming a normal distribution
      # with the mean grain size as mean and half the range (min:max)
      # as standard deviation. For a more conservative estimate this
      # is further divided by 2, so half the range is regarded as
      # two sigma.
      gs.mc<- rnorm(settings$MC.iter, grain.size, diff(gs.range)/4)

      # draw random samples from the grain size spectrum (gs.mc) and calculate
      # the mean for each sample. This gives an approximation of the variation
      # in mean grain size on the sample disc
      gs.mc.sampleMean <-
        if (diff(gs.range) > 0) {
          sapply(1:length(gs.mc),
                 function(x) mean(sample(gs.mc,
                                         calc_n(
                                             sample(sd.mc, size = 1),
                                             grain.size,
                                             sample(d.mc, size = 1)
                                         ),
                                         replace = TRUE)))
        } else {
          grain.size
        }

      # calculate n for each MC data set
      MC.n <- apply(data.frame(sd.mc, gs.mc.sampleMean, d.mc), 1,
                    function(x) calc_n(x[1], x[2], x[3]))

      # summarize MC estimates
      MC.q<- quantile(MC.n, c(0.05,0.95))
      MC.n.kde<- density(MC.n, n = 10000)

      # apply student's t-test
      MC.t.test<- stats::t.test(MC.n)
      MC.t.lower<- MC.t.test["conf.int"]$conf.int[1]
      MC.t.upper<- MC.t.test["conf.int"]$conf.int[2]
      MC.t.se<- (MC.t.upper-MC.t.lower)/3.92

      # get unweighted statistics from calc_Statistics() function
      MC.stats<- calc_Statistics(as.data.frame(cbind(MC.n,0.0001)))$unweighted
    }
  }#EndOf:estimate number of grains


  ##========================================================================##
  ## CALCULATE PACKING DENSITY

  if (!is.null(grains.counted)) {
    area.container <- pi * (sample.diameter / 2)^2
    area.grains <- pi * (grain.size / 1000 / 2)^2 * grains.counted
    packing.density <- area.grains / area.container
    std.d <- sd(packing.density)
  }

  ##==========================================================================##
  ##TERMINAL OUTPUT
  ##==========================================================================##
  if (settings$verbose) {

    cat("\n [calc_AliquotSize]")
    cat("\n\n ---------------------------------------------------------")
    cat("\n mean grain size (microns)  :", grain.size)
    cat("\n sample diameter (mm)       :", sample.diameter)
    if (!is.null(grains.counted)) {
      if(length(grains.counted) == 1) {
        cat("\n counted grains             :", grains.counted)
        cat("\n packing density            :", round(packing.density, 3))
      } else {
        cat("\n mean counted grains        :", round(mean(grains.counted)))
        cat("\n mean packing density       :", round(mean(packing.density), 3))
        cat("\n standard deviation         :", round(std.d, 3))
      }
    }
    if (is.null(grains.counted)) {
      cat("\n packing density            :", round(packing.density, 3))
      cat("\n number of grains           :", round(n.grains, 0))
    }

    if (MC && range.flag) {
      cat("\n\n --------------- Monte Carlo Estimates -------------------")
      cat("\n number of iterations (n)     :", settings$MC.iter)
      cat("\n median                       :", round(MC.stats$median))
      cat("\n mean                         :", round(MC.stats$mean))
      cat("\n standard deviation (mean)    :", round(MC.stats$sd.abs))
      cat("\n standard error (mean)        :", round(MC.stats$se.abs, 1))
      cat("\n 95% CI from t-test (mean)    :", round(MC.t.lower), "-", round(MC.t.upper))
      cat("\n standard error from CI (mean):", round(MC.t.se, 1))
    }
    cat("\n ---------------------------------------------------------\n")
  }

  ##==========================================================================##
  ##RETURN VALUES
  ##==========================================================================##

  if (is.null(grains.counted))
    grains.counted <- NA
  else
    n.grains <- NA
  summary <- data.frame(grain.size = grain.size,
                        sample.diameter = sample.diameter,
                        packing.density = packing.density,
                        n.grains = round(n.grains, 0),
                        grains.counted = grains.counted)

  call<- sys.call()
  args<- as.list(sys.call())[-1]

  # create S4 object
  object <- set_RLum(
    class = "RLum.Results",
    originator = "calc_AliquotSize",
    data = list(
      summary=summary,
      MC=list(estimates=MC.n,
              statistics=MC.stats,
              kde=MC.n.kde,
              t.test=MC.t.test,
              quantile=MC.q)),
    info = list(call=call,
                args=args))

  ##=========##
  ## PLOTTING
  if (plot && !is.null(object@data$MC$estimates)) {

    ## default graphical settings
    settings <- list(
        main = "Monte Carlo Simulation",
        xlab = "Amount of grains on aliquot",
        col = "gray90",
        line_col = "black",
        line_lwd = 1,
        cex = 0.8,
        rug = TRUE,
        boxplot = TRUE,
        summary = TRUE,
        legend = TRUE,
        NULL
    )
    settings <- modifyList(settings, list(...))

    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    ## set layout of plotting device
    nrow.splits <- if (settings$boxplot) 2 else 1
    graphics::layout(matrix(c(1, 1, nrow.splits)), nrow.splits, 1)
    par(cex = settings$cex, mar = c(if (settings$boxplot) 2 else 5, 5, 5, 3))

    ## plot histogram
    hist(MC.n, freq = FALSE, col = settings$col,
         main = settings$main, xlab = settings$xlab, cex = settings$cex,
         xlim = c(min(MC.n) * 0.95, max(MC.n) * 1.05),
         ylim = c(0, max(MC.n.kde$y) * 1.1))

    ## add rugs to histogram
    if (settings$rug) {
      graphics::rug(MC.n)
    }

    ## add KDE curve
    lines(MC.n.kde, col = settings$line_col, lwd = settings$line_lwd)

    ## add mean, median and quantils (0.05,0.95)
    abline(v = c(MC.stats$mean, MC.stats$median, MC.q),
           lty = c(2, 4, 3, 3), lwd = 1)

    ## add title and subtitle
    if (settings$summary) {
      mtext(as.expression(bquote(italic(n) == .(settings$MC.iter) ~ "|" ~
                                 italic(hat(mu)) == .(round(MC.stats$mean)) ~ "|" ~
                                 italic(hat(sigma)) == .(round(MC.stats$sd.abs)) ~ "|" ~
                                 italic(frac(hat(sigma), sqrt(n))) == .(round(MC.stats$se.abs)) ~ "|" ~
                                 italic(v) == .(round(MC.stats$skewness, 2)))),
            side = 3, line = -0.5, adj = 0.5, cex = 0.9 * settings$cex)
    }

    ## add legend
    if (settings$legend) {
      legend("topright", legend = c("mean","median", "0.05 / 0.95 quantile"),
             lty = c(2, 4, 3), bg = "white", box.col = "white",
             cex = 0.9)
    }

    ## BOXPLOT
    ## set margins (bottom, left, top, right)
    if (settings$boxplot) {
      par(mar = c(5, 5, 0, 3))
      plot(NA, type = "n", xlim = c(min(MC.n) * 0.95, max(MC.n) * 1.05),
           xlab = settings$xlab, ylim = c(0.5, 1.5),
           xaxt = "n", yaxt = "n", ylab = "")
      graphics::boxplot(MC.n, horizontal = TRUE, add = TRUE, bty = "n", cex = 1)
    }
  }

  # Return values
  invisible(object)
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_AliquotSize(),
               "grain.size' should be of class 'numeric' or 'integer'")
  expect_error(calc_AliquotSize(grain.size = 1:3),
               "Please provide the mean grain size or the range of grain sizes")
  expect_error(calc_AliquotSize(grain.size = 0),
               "'grain.size' should contain positive values")
  expect_error(calc_AliquotSize(grain.size = NA_real_),
               "'grain.size' should contain positive values")
  expect_error(calc_AliquotSize(grain.size = 100, packing.density = "inf"),
               "'packing.density' should be a single positive value")
  expect_error(calc_AliquotSize(grain.size = 100, packing.density = 2),
               "'packing.density' should be a value between 0 and 1")

  expect_error(calc_AliquotSize(grain.size = 100, packing.density = 1, sample.diameter = -1),
               "'sample.diameter' should be a single positive value")
  expect_error(calc_AliquotSize(grain.size = 100, packing.density = 1,
                                sample.diameter = 1,
                                sample_carrier.diameter = -1),
               "'sample_carrier.diameter' should be a single positive value")
  expect_error(calc_AliquotSize(grain.size = 100, sample.diameter = 9.8,
                                MC = TRUE),
               "'grain.size' must be a vector containing the min and max")
  SW({
  expect_message(calc_AliquotSize(grain.size = 100, packing.density = 1,
                                  sample.diameter = 9.8, grains.counted = 30,
                                  MC = TRUE),
                 "Monte Carlo simulation is only available for estimating the")
  })
  expect_error(calc_AliquotSize(grain.size = c(100, 200),
                                sample.diameter = 3.6, MC.iter = 0),
               "'MC.iter' should be a single positive integer value or NULL")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  expect_s4_class(
    calc_AliquotSize(
      grain.size = 100, packing.density = Inf, sample.diameter = 9.8, MC = FALSE), "RLum.Results")
  expect_s4_class(
    calc_AliquotSize(
      grain.size = c(100, 150), grains.counted = 1000, sample.diameter = 9.8, MC = FALSE), "RLum.Results")
  expect_warning(calc_AliquotSize(
      grain.size = c(100, 150), grains.counted = c(1000, 1100, 900),
      sample.diameter = 10, MC = FALSE),
      "A sample diameter of 10 mm was specified for a sample carrier of 9.8 mm")
  })

  set.seed(1)
  temp <- calc_AliquotSize(
      grain.size = c(100, 150),
      sample.diameter = 1,
      MC.iter = 100,
      verbose = FALSE)

  expect_equal(is(temp), c("RLum.Results", "RLum"))
  expect_equal(length(temp), 2)
  expect_s3_class(temp$summary, "data.frame")
  expect_type(temp$MC, "list")

  ## summary output
  result <- get_RLum(temp)
  expect_equal(result$grain.size, 125)
  expect_equal(result$sample.diameter, 1)
  expect_equal(result$packing.density, 0.65)
  expect_equal(result$n.grains, 42)
  expect_equal(result$grains.counted, NA)

  ## MC run
  expect_equal(round(temp$MC$statistics$n), 100)
  expect_equal(round(temp$MC$statistics$mean), 43)
  expect_equal(round(temp$MC$statistics$median), 39)
  expect_equal(round(temp$MC$statistics$sd.abs), 19)
  expect_equal(round(temp$MC$statistics$sd.rel), 45)
  expect_equal(round(temp$MC$statistics$se.abs), 2)
  expect_equal(round(temp$MC$statistics$se.rel), 5)
  expect_length(temp$MC$kde$x, 10000)
  expect_length(temp$MC$kde$y, 10000)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.5e-4

  SW({
  expect_snapshot_RLum(calc_AliquotSize(
      grain.size = c(80, 150),
      sample.diameter = 1,
      MC.iter = 20,
      plot = FALSE,
      verbose = TRUE),
      tolerance = snapshot.tolerance)

  expect_message(
  expect_snapshot_RLum(calc_AliquotSize(
      grain.size = c(200, 100),
      sample.diameter = 8,
      grains.counted = c(2525, 2312, 2880),
      MC.iter = 20,
      plot = FALSE,
      verbose = FALSE),
      tolerance = snapshot.tolerance),
  "Monte Carlo simulation is only available for estimating the amount of")

  expect_snapshot_RLum(calc_AliquotSize(
      grain.size = c(80, 150),
      sample.diameter = 7,
      sample_carrier.diameter = 7.2,
      MC.iter = 20,
      plot = FALSE,
      verbose = TRUE),
      tolerance = snapshot.tolerance)

  ## regression test - issue 1114
  expect_snapshot_RLum(calc_AliquotSize(
      grain.size = c(0.001, 0.001),
      sample.diameter = 7,
      MC.iter = 10,
      plot = FALSE),
      tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
  vdiffr::expect_doppelganger("default",
                              calc_AliquotSize(grain.size = c(80, 150),
                                               sample.diameter = 7,
                                               sample_carrier.diameter = 7.2,
                                               MC.iter = 20,
                                               verbose = FALSE))
  vdiffr::expect_doppelganger("noboxplot-cex",
                              calc_AliquotSize(grain.size = c(100, 200),
                                               sample.diameter = 8,
                                               sample_carrier.diameter = 2,
                                               MC.iter = 20,
                                               cex = 2,
                                               boxplot = FALSE,
                                               verbose = FALSE))
  })
})

#' @title Converting equivalent dose values from seconds (s) to Gray (Gy)
#'
#' @description
#' Conversion of absorbed radiation dose in seconds (s) to the SI unit Gray
#' (Gy) including error propagation. Normally used for equivalent dose data.
#'
#' Calculation of De values from seconds (s) to Gray (Gy)
#'
#' \deqn{De [Gy] = De [s] * Dose Rate [Gy/s])}
#'
#' Provided calculation error propagation methods for error calculation
#' (with `'se'` as the standard error and `'DR'` of the dose rate of the beta-source):
#'
#' **(1) `omit`** (default)
#'
#' \deqn{se(De) [Gy] = se(De) [s] * DR [Gy/s]}
#'
#' In this case the standard error of the dose rate of the beta-source is
#' treated as systematic (i.e. non-random), and error propagation is omitted.
#' However, the error must be considered during calculation of the final age
#' (cf. Aitken, 1985, pp. 242). This approach can be seen as method (2) (gaussian)
#' for the case the (random) standard error of the beta-source calibration is
#' 0. Which particular method is requested depends on the situation and cannot
#' be prescriptive.
#'
#' **(2) `gaussian`** error propagation
#'
#' \deqn{se(De) [Gy] = \sqrt{(DR [Gy/s] * se(De) [s])^2 + (De [s] * se(DR) [Gy/s])^2}}
#'
#' Applicable under the assumption that errors of `De` and `se` are uncorrelated.
#'
#' **(3) `absolute`** error propagation
#'
#' \deqn{se(De) [Gy]= abs(DR [Gy/s] * se(De) [s]) + abs(De [s] * se(DR) [Gy/s])}
#'
#' Applicable under the assumption that errors of `De` and `se` are correlated.
#'
#'
#' @param data [data.frame] (**required**):
#' input values, structure: data (`values[,1]`) and data error (`values [,2]`)
#' are required.
#'
#' @param dose.rate [Luminescence::RLum.Results-class], [data.frame] or [numeric] (**required**):
#' `RLum.Results` needs to be originated from the function [Luminescence::calc_SourceDoseRate],
#' for `vector` dose rate in Gy/s and dose rate error in Gy/s.
#'
#' @param error.propagation [character] (*with default*):
#' error propagation method used for error calculation (`omit`, `gaussian` or
#' `absolute`), see details for further information.
#'
#' @return
#' Returns a [data.frame] with converted values.
#'
#' @note
#' If no or a wrong error propagation method is given, the execution of the function is
#' stopped. Furthermore, if a `data.frame` is provided for the dose rate values is has to
#' be of the same length as the data frame provided with the argument `data`
#'
#' @section Function version: 0.6.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Margret C. Fuchs, HZDR, Helmholtz-Institute Freiberg for Resource Technology (Germany)
#'
#' @seealso [Luminescence::calc_SourceDoseRate]
#'
#' @references
#' Aitken, M.J., 1985. Thermoluminescence dating. Academic Press.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(A) for known source dose rate at date of measurement
#' ## - load De data from the example data help file
#' data(ExampleData.DeValues, envir = environment())
#' ## - convert De(s) to De(Gy)
#' convert_Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))
#'
#'
#' ##(B) for source dose rate calibration data
#' ## - calculate source dose rate first
#' dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
#'                                   calib.date = "2014-12-19",
#'                                   calib.dose.rate = 0.0438,
#'                                   calib.error = 0.0019)
#' # read example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' # apply dose.rate to convert De(s) to De(Gy)
#' convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate)
#'
#' @export
convert_Second2Gray <- function(
  data,
  dose.rate,
  error.propagation = c("omit", "gaussian", "absolute")
) {
  .set_function_name("convert_Second2Gray")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, "data.frame")
  .validate_class(dose.rate, c("RLum.Results", "data.frame", "numeric"))

  if (is.data.frame(dose.rate)) {
    if(nrow(dose.rate)!=nrow(data)){
      .throw_error("Data frames in 'data' and 'dose.rate' must have the same length")
    }
  } else if (inherits(dose.rate, "RLum.Results")) {
    .validate_originator(dose.rate, "calc_SourceDoseRate")
    dose.rate <- get_RLum(dose.rate, data.object = "dose.rate")
  }
  .validate_not_empty(dose.rate)

  error.propagation <- .validate_args(error.propagation,
                                      c("omit", "gaussian", "absolute"))


  ## Calculation ------------------------------------------------------------

  De.seconds <- data[,1]
  De.error.seconds <- data[,2]
  De.gray <- round(De.seconds * dose.rate[[1]], digits = 2)

  De.error.gray <- switch(
      error.propagation,
      omit = dose.rate[[1]] * De.error.seconds,
      gaussian = sqrt((dose.rate[[1]] * De.error.seconds)^2 +
                      (dose.rate[[2]] * De.seconds)^2),
      absolute = (abs(dose.rate[[1]] * De.error.seconds) +
                  abs(dose.rate[[2]] * De.seconds))
  )

  ## Return -----------------------------------------------------------------
  data.frame(De = De.gray, De.error = round(De.error.gray, digits = 3))
}

## load data
data(ExampleData.DeValues, envir = environment())

results <- convert_Second2Gray(ExampleData.DeValues$BT998, c(0.2,0.01))
dose.rate <- calc_SourceDoseRate(measurement.date = "2024-10-09",
                                 calib.date = "2014-12-19",
                                 calib.dose.rate = 0.2, calib.error = 0.01)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_Second2Gray("test"),
               "'data' should be of class 'data.frame'")
  expect_error(convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate = FALSE),
               "'dose.rate' should be of class 'RLum.Results', 'data.frame' or")
  expect_error(convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate = numeric(0)),
               "'dose.rate' cannot be an empty numeric")
  expect_error(convert_Second2Gray(ExampleData.DeValues$BT998,
                           dose.rate = results[1:5, ]),
               "'data' and 'dose.rate' must have the same length")
  expect_error(convert_Second2Gray(ExampleData.DeValues$BT998,
                           dose.rate = results,
                           error.propagation = "test"),
               "'error.propagation' should be one of 'omit', 'gaussian'")

  dose.rate@originator <- "unexpected-originator"
  expect_error(convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate = dose.rate),
               "'dose.rate' has an unsupported originator")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_plain(results)
  expect_snapshot_plain(convert_Second2Gray(ExampleData.DeValues$BT998,
                                    dose.rate = c(0.2, 0.01),
                                    error.propagation = "gaussian"))
  expect_snapshot_plain(convert_Second2Gray(ExampleData.DeValues$BT998,
                                    dose.rate = c(0.2, 0.01),
                                    error.propagation = "absolute"))

  expect_snapshot_plain(convert_Second2Gray(ExampleData.DeValues$BT998,
                                    dose.rate = results,
                                    error.propagation = "gaussian"))
  expect_snapshot_plain(convert_Second2Gray(ExampleData.DeValues$BT998,
                                    dose.rate = results,
                                    error.propagation = "absolute"))

  expect_snapshot_plain(convert_Second2Gray(ExampleData.DeValues$BT998,
                                    dose.rate = dose.rate))
})

## load data
data(ExampleData.DeValues, envir = environment())
df <- convert_Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_KDE("error"),
               "'data' should be of class 'RLum.Results', 'data.frame' or 'numeric'")
  expect_error(plot_KDE(df[0, ]),
               "Input data 1 has 0 rows")
  expect_error(plot_KDE(list()),
               "'data' is an empty list")
  expect_error(expect_warning(plot_KDE(data.frame(a = Inf, b = 1)),
                              "Inf values removed in rows: 1 in data.frame 1"),
               "Your input is empty due to Inf removal")
  expect_error(plot_KDE(df, df),
               "'na.rm' should be a single logical value")
  expect_error(plot_KDE(df, summary.method = "error"),
               "'summary.method' should be one of 'MCM', 'weighted' or 'unweighted'")
  expect_error(plot_KDE(df, summary = 5),
               "'summary' should be of class 'character'")
  expect_error(plot_KDE(df, summary.pos = 5),
               "'summary.pos' should have length 2")
  expect_error(plot_KDE(df, summary.pos = list()),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_KDE(df, summary.pos = "error"),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_KDE(df, bw = df),
               "'bw' should be of class 'character' or 'numeric' and have length 1")
  expect_error(plot_KDE(df, bw = c(0, 1)),
               "'bw' should be of class 'character' or 'numeric' and have length 1")
  expect_error(plot_KDE(df, bw = -1),
               "'bw' should be a single positive value")
  expect_error(plot_KDE(df, ylim = c(0, 1)),
               "'ylim' should be of class 'numeric' and have length 4")

  expect_warning(plot_KDE(df[1, ]),
                 "Single data point found, no density calculated")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## create plot straightforward
  expect_silent(plot_KDE(data = df))

  ## more coverage
  expect_silent(plot_KDE(data = df, summary = "n", summary.pos = c(105, 0.07),
                         main = "Title", mtext = "Subtitle",
                         xlab = "x", ylab = "y", layout = "default", log = "x",
                         xlim = c(100, 180), ylim = c(0, 0.07, 0, 0.01),
                         col = 2, lty = 2, lwd = 2, cex = 1))
  expect_silent(plot_KDE(data = df, summary = "mean", summary.pos = "right"))
  expect_silent(plot_KDE(data = df, sub = "test"))

  ## specify layout
  layout <- get_Layout("default")
  layout$kde$dimension$figure.width <- 100
  layout$kde$dimension$figure.height <- 100
  plot_KDE(data = df, layout = layout)

  ## numeric vector
  expect_silent(plot_KDE(df[, 1]))

  ## single-column data.frame
  expect_silent(plot_KDE(df[, 1, drop = FALSE]))

  ## RLum.Results object
  expect_silent(plot_KDE(calc_CommonDose(df, plot = FALSE, verbose = FALSE)))

  ## infinite values
  df.inf <- df
  df.inf[9, 1] <- Inf
  expect_warning(plot_KDE(df.inf),
                 "Inf values removed in rows: 9 in data.frame 1")
  df.inf[2, 2] <- Inf
  expect_warning(plot_KDE(df.inf),
                 "Inf values removed in rows: 2, 9 in data.frame 1")

  ## missing values
  df.na <- df
  df.na[2, 1] <- NA
  expect_message(plot_KDE(df.na),
                 "1 NA value excluded from data set 1")
  df.na[3, 1] <- NA
  expect_message(plot_KDE(df.na),
                 "2 NA values excluded from data set 1")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("KDE expected",
                              plot_KDE(data = df))
  vdiffr::expect_doppelganger("cex",
                              plot_KDE(data = df,
                                       mtext = "Subtitle",
                                       cex = 2))
  vdiffr::expect_doppelganger("KDE summary sub",
                              plot_KDE(data = df, summary.pos = "sub",
                                       summary = c("n", "se.rel", "kurtosis")))
  vdiffr::expect_doppelganger("KDE summary left",
                              plot_KDE(data = df, summary.pos = "left",
                                       summary.method = "weighted",
                                       summary = c("mean", "in.2s", "skewness",
                                                   "median")))
  vdiffr::expect_doppelganger("not values.cumulative",
                              plot_KDE(data = df,
                                       values.cumulative = FALSE))
  vdiffr::expect_doppelganger("rug",
                              plot_KDE(data = list(df, data.frame(c(23, 24), c(3, 3)))))
  vdiffr::expect_doppelganger("rug many points",
                              plot_KDE(data = rbind(df, df, df, df)))
  })
})

#' @title Export Risoe BIN-file(s) to CSV-files
#'
#' @description This function is a wrapper function around the functions [Luminescence::read_BIN2R] and
#' [Luminescence::write_RLum2CSV] and it imports a Risoe BIN-file and directly exports its
#' content to CSV-files. If nothing is set for the argument `path`
#' ([Luminescence::write_RLum2CSV]) the input folder will become the output folder.
#'
#' @param file [character] (**required**):
#' name of the BIN-file to be converted to CSV-files
#'
#' @param ... further arguments that will be passed to the function
#' [Luminescence::read_BIN2R] and [Luminescence::write_RLum2CSV]
#'
#' @return
#' The function returns either a CSV-file (or many of them) or for the
#' option `export == FALSE` a list comprising objects of type [data.frame] and [matrix]
#'
#' @section Function version: 0.1.0
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data-class], [Luminescence::RLum.Results-class],
#' [utils::write.table], [Luminescence::write_RLum2CSV], [Luminescence::read_BIN2R]
#'
#' @keywords IO
#'
#' @examples
#'
#' ##transform Risoe.BINfileData values to a list
#' data(ExampleData.BINfileData, envir = environment())
#' convert_BIN2CSV(subset(CWOSL.SAR.Data, POSITION == 1), export = FALSE)
#'
#' \dontrun{
#' ##select your BIN-file
#' file <- file.choose()
#'
#' ##convert
#' convert_BIN2CSV(file)
#'
#' }
#'
#' @export
convert_BIN2CSV <- function(
  file,
  ...
) {
  .set_function_name("convert_BIN2CSV")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(file, c("character", "Risoe.BINfileData"))
  .validate_not_empty(file)

  ##set input arguments
  convert_BIN2CSV_settings.default <- list(
    path = if (!inherits(file, "Risoe.BINfileData")) dirname(file) else NULL,
    show.raw.values = FALSE,
    position = NULL,
    n.records = NULL,
    zero_data.rm = TRUE,
    duplicated.rm = FALSE,
    show.record.number = FALSE,
    txtProgressBar = TRUE,
    forced.VersionNumber = NULL,
    ignore.RECTYPE = FALSE,
    pattern = NULL,
    verbose = TRUE,
    export = TRUE
  )

  ##modify list on demand
  convert_BIN2CSV_settings <- modifyList(x = convert_BIN2CSV_settings.default, val = list(...))

  # Import file ---------------------------------------------------------------------------------
  if (!inherits(file, "Risoe.BINfileData")) {
    object <- read_BIN2R(
      file = file,
      show.raw.values = convert_BIN2CSV_settings$show.raw.values,
      position = convert_BIN2CSV_settings$position,
      n.records =  convert_BIN2CSV_settings$n.records,
      zero_data.rm = convert_BIN2CSV_settings$zero_data.rm,
      duplicated.rm = convert_BIN2CSV_settings$duplicated.rm,
      fastForward = TRUE,
      show.record.number = convert_BIN2CSV_settings$show.record.number,
      txtProgressBar = convert_BIN2CSV_settings$txtProgressBar,
      forced.VersionNumber = convert_BIN2CSV_settings$forced.VersionNumber,
      ignore.RECTYPE = convert_BIN2CSV_settings$ignore.RECTYPE,
      pattern = convert_BIN2CSV_settings$pattern,
      verbose = convert_BIN2CSV_settings$verbose
   )
  }else{
   object <- Risoe.BINfileData2RLum.Analysis(file)
  }

  # Export to CSV -------------------------------------------------------------------------------

  ##get all arguments we want to pass and remove the doubled one
  arguments <- c(list(object = object, export = convert_BIN2CSV_settings$export), list(...))
  arguments[duplicated(names(arguments))] <- NULL

  ##this if-condition prevents NULL in the terminal
  if (convert_BIN2CSV_settings$export) {
    invisible(do.call("write_RLum2CSV", arguments))
  }else{
    do.call("write_RLum2CSV", arguments)
  }
}

test_that("test convert functions", {
  testthat::skip_on_cran()
  local_edition(3)

  ##test for errors
  expect_error(convert_BIN2CSV(file = "", export = FALSE),
               regexp = "[read_BIN2R()] File does not exist!",
               fixed = TRUE)
  expect_error(convert_Daybreak2CSV(file = "", export = FALSE),
               regexp = "[read_Daybreak2R()] file name does not seem to exist.",
               fixed = TRUE)
  #expect_error(convert_PSL2CSV(file = "", export = FALSE))
  expect_error(suppressWarnings(convert_XSYG2CSV(file = "", export = FALSE)))

  ##test conversion itself
    ##BIN2CSV
    data(ExampleData.BINfileData, envir = environment())
    expect_type(convert_BIN2CSV(subset(CWOSL.SAR.Data, POSITION == 1), export = FALSE), "list")

    ##XSYG2CSV
    data(ExampleData.XSYG, envir = environment())
    expect_type(convert_XSYG2CSV(OSL.SARMeasurement$Sequence.Object[1:10], export = FALSE), "list")

})

## load datax
data(ExampleData.BINfileData, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_BIN2CSV(),
               "'file' should be of class 'character' or 'Risoe.BINfileData'")
  expect_error(convert_BIN2CSV(character(0)),
               "'file' cannot be an empty character")
  expect_error(convert_BIN2CSV(file = "error", export = FALSE),
               "File '.*error' does not exist") # windows CI needs the regexp
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_type(convert_BIN2CSV(subset(CWOSL.SAR.Data, POSITION == 1),
                              export = FALSE),
              "list")
  expect_null(convert_BIN2CSV(subset(CWOSL.SAR.Data, POSITION == 1),
                              export = TRUE, path = tempdir()))
})

#' Smoothing of data
#'
#' Function calls the object-specific smooth functions for provided RLum S4-class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum-class] objects.\cr
#' Depending on the input object, the corresponding function will be selected.
#' Allowed arguments can be found in the documentations of the corresponding
#' [RLum-class] class. The smoothing is based on an internal function
#' called `.smoothing`.
#'
#' @param object [RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @param ... further arguments passed to the specific class method
#'
#' @return
#' An object of the same type as the input object is provided
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#'
#' @note
#' Currently only `RLum` objects of class `RLum.Data.Curve` and `RLum.Analysis`
#' (with curve data) are supported!
#'
#' @seealso [RLum.Data.Curve-class], [RLum.Analysis-class]
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ##create RLum.Data.Curve object from this example
#' curve <-
#'   set_RLum(
#'       class = "RLum.Data.Curve",
#'       recordType = "OSL",
#'       data = as.matrix(ExampleData.CW_OSL_Curve)
#'   )
#'
#' ##plot data without and with smoothing
#' plot_RLum(curve)
#' plot_RLum(smooth_RLum(curve))
#'
#' @keywords utilities
#'
#' @md
#' @export
setGeneric("smooth_RLum", function(object, ...) {
  standardGeneric("smooth_RLum")

})

# Method for smooth_RLum method for RLum objects in a list for a list of objects  -------------------
#' @describeIn smooth_RLum
#' Returns a list of [RLum-class] objects that had been passed to [smooth_RLum]
#'
#'
#' @md
#' @export
setMethod("smooth_RLum",
signature = "list",
function(object, ...){

  ##apply method in the objects and return the same
  lapply(object, function(x){
    if(inherits(x, "RLum")){
      return(smooth_RLum(x,...))
    }else{
      return(x)
    }

  })

})

#' @title (Un-)logged minimum age model (MAM) after Galbraith et al. (1999)
#'
#' @description
#' Function to fit the (un-)logged three or four parameter minimum dose model
#' (MAM-3/4) to De data.
#'
#' **Parameters**
#'
#' This model has four parameters:
#' \tabular{rl}{
#' `gamma`: \tab minimum dose on the log scale \cr
#' `mu`: \tab mean of the non-truncated normal distribution \cr
#' `sigma`: \tab spread in ages above the minimum \cr
#' `p0`: \tab proportion of grains at gamma \cr }
#'
#' If `par=3` (default) the 3-parameter minimum age model is applied,
#' where `gamma=mu`. For `par=4` the 4-parameter model is applied instead.
#'
#' **(Un-)logged model**
#'
#' In the original version of the minimum dose model, which is applied if
#' `log = TRUE` (default), the basic data are the natural
#' logarithms of the De estimates and relative standard errors of the De
#' estimates. The value for `sigmab` must be provided as a ratio
#' (e.g, 0.2 for 20 %).
#'
#' If `log=FALSE`, the modified un-logged model will be applied instead. This
#' has essentially the same form as the original version, but  `gamma` and
#' `sigma` are in Gy and `gamma` becomes the minimum true dose in the
#' population.
#' **Note:** the un-logged model requires `sigmab` to be in the same
#' absolute unit as the provided De values (seconds or Gray).
#'
#' While the original (logged) version of the minimum dose
#' model may be appropriate for most samples (i.e. De distributions), the
#' modified (un-logged) version is specially designed for modern-age and young
#' samples containing negative, zero or near-zero De estimates (Arnold et al.
#' 2009, p. 323).
#'
#' **Initial values and boundaries**
#'
#' The log-likelihood calculations use the [nlminb] function for box-constrained
#' optimisation using PORT routines.  Accordingly, initial values for the four
#' parameters can be specified via `init.values`. If no values are
#' provided for `init.values`, reasonable starting values are estimated
#' from the input data.  If the final estimates of *gamma*, *mu*,
#' *sigma* and *p0* are totally off target, consider providing custom
#' starting values via `init.values`.
#'
#' The boundaries for individual model parameters need not be specified
#' explicitly. To override the default boundary values, provide arguments
#' `gamma.lower`, `gamma.upper`, `sigma.lower`, `sigma.upper`,
#' `p0.lower`, `p0.upper`, `mu.lower` and `mu.upper`.
#'
#' **Bootstrap**
#'
#' When `bootstrap=TRUE` the function applies the bootstrapping method as
#' described in Cunningham & Wallinga (2012). The minimum age model produces
#' `bs.M` first-level bootstrap replicates (1000 by default) and `bs.N`
#' second-level replicates (`3 * bs.M` by default), with an uncertainty
#' `sigma.sd` on the overdispersion parameter (0.04 by default). These values
#' can be changed by using arguments `bs.M`, `bs.N` and `sigmab.sd`,
#' respectively. The bandwidth of the kernel density estimate can be specified
#' with `bs.h`. By default, this is calculated as:
#'
#' \deqn{h = (2*\sigma_{DE})/\sqrt{n}}
#'
#' **Multicore support**
#'
#' Parallel processing can be enabled by setting `multicore = TRUE`. By default,
#' the number of logical CPU cores is detected automatically, but it can be
#' overridden with the `cores` argument. Multicore processing is supported
#' only when `bootstrap = TRUE`, and spawns one R process per core to compute
#' MAM estimates for each of the `N * M` bootstrap replicates. Note that this
#' feature is experimental and may not work on all systems. Performance gains
#' grow with the number of bootstrap replicates, but each additional core (R
#' process) increases memory usage. When memory is insufficient, overall
#' performance can degrade severely.
#'
#' **Likelihood profiles**
#'
#' The likelihood profiles are generated and plotted by the `bbmle` package.
#' The profile likelihood plots look different to ordinary profile likelihood as
#'
#' "`[...]` the plot method for likelihood profiles displays the square root of
#' the deviance difference (twice the difference in negative log-likelihood from
#' the best fit), so it will be V-shaped for cases where the quadratic approximation
#' works well `[...]`." (Bolker 2016).
#'
#' For more details on the profile likelihood
#' calculations and plots please see the vignettes of the `bbmle` package
#' (also available here: [https://CRAN.R-project.org/package=bbmle]().
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with `(data[, 1])` and De error `(data[, 2])`.
#'
#' @param sigmab [numeric] (**required**):
#' additional spread in De values, representing the expected overdispersion in
#' the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
#' **Note:** For the logged model (`log = TRUE`) this value must be
#' a fraction, e.g. 0.2 (= 20 %). If the un-logged model is used (`log = FALSE`),
#' `sigmab` must be provided in the same absolute units of the De values (seconds or Gray).
#' See details.
#'
#' @param log [logical] (*with default*):
#' whether the logged minimum dose model should be fit to De data (`TRUE` by
#' default).
#'
#' @param par [numeric] (*with default*):
#' number of parameters in the minimum age model, either 3 (default) or 4.
#'
#' @param bootstrap [logical] (*with default*):
#' apply the recycled bootstrap approach of Cunningham & Wallinga (2012).
#'
#' @param init.values [numeric] (*optional*):
#' a named list with starting values for `gamma`, `sigma`, `p0` and `mu`
#' (e.g. `list(gamma=100, sigma=1.5, p0=0.1, mu=100)`). If no values are
#' provided, reasonable values will be estimated from the data.
#' **Note:** the initial values must always be given in the absolute units.
#' If a logged model is applied (`log = TRUE`), the provided `init.values`
#' are automatically log-transformed.
#'
#' @param level [logical] (*with default*):
#' the confidence level required (defaults to 0.95).
#'
#' @param log.output [logical] (*with default*):
#' if `TRUE` and `log = TRUE`, the console output will also show the logged
#' values of the final parameter estimates and confidence intervals.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param multicore [logical] (*with default*):
#' parallelize the computation of the bootstrap by creating a multicore cluster
#' (only considered if `bootstrap = TRUE`). By default, it uses all available
#' logical CPU cores, but this can be changed with the `cores` argument. Note
#' that this feature is experimental and may not work on all systems.
#'
#' @param ... (*optional*) further arguments for bootstrapping
#' (`bs.M`, `bs.N`, `bs.h`, `sigmab.sd`). See details for their usage.
#' Further arguments are
#' - `verbose`: enable/disable output to the terminal
#' - `debug`: enable/disable extended console output
#' - `cores`: number of cores to be used when `multicore=TRUE`
#'
#' @return Returns a plot (*optional*) and terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following elements:
#'
#' \item{$summary}{[data.frame] summary of all relevant model results.}
#' \item{$data}{[data.frame] original input data}
#' \item{$args}{[list] used arguments}
#' \item{$call}{[call] the function call}
#' \item{$mle}{[bbmle::mle2] object containing the maximum log likelihood functions for all parameters}
#' \item{$BIC}{[numeric] BIC score from the non-bootstrapped model}
#' \item{$confint}{[data.frame] confidence intervals for all parameters}
#' \item{$profile}{[stats::profile] the log likelihood profiles}
#' \item{$bootstrap}{[list] bootstrap results}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @note
#' The default starting values for *gamma*, *mu*, *sigma*
#' and *p0* may only be appropriate for some De data sets and may need to
#' be changed for other data. This is especially true when the un-logged
#' version is applied. \cr
#' Also note that all R warning messages are suppressed
#' when running this function. If the results seem odd consider re-running the
#' model with `debug=TRUE` which provides extended console output and
#' forwards all internal warning messages.
#'
#' @section Function version: 0.5.0
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany) \cr
#' Based on a rewritten S script of Rex Galbraith, 2010 \cr
#' The bootstrap approach is based on a MATLAB script by Alastair Cunningham,
#' who helped with implementation and cross-checking.
#'
#' @seealso [Luminescence::calc_CentralDose], [Luminescence::calc_CommonDose], [Luminescence::calc_FiniteMixture],
#' [Luminescence::calc_FuchsLang2001], [Luminescence::calc_MaxDose]
#'
#' @references
#' Arnold, L.J., Roberts, R.G., Galbraith, R.F. & DeLong, S.B.,
#' 2009. A revised burial dose estimation procedure for optical dating of young
#' and modern-age sediments. Quaternary Geochronology 4, 306-325.
#'
#' Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for mixed fission
#' track ages. Nuclear Tracks Radiation Measurements 4, 459-470.
#'
#' Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley, J.M.,
#' 1999. Optical dating of single grains of quartz from Jinmium rock shelter,
#' northern Australia. Part I: experimental design and statistical models.
#' Archaeometry 41, 339-364.
#'
#' Galbraith, R.F., 2005. Statistics for
#' Fission Track Analysis, Chapman & Hall/CRC, Boca Raton.
#'
#' Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error
#' calculation and display in OSL dating: An overview and some recommendations.
#' Quaternary Geochronology 11, 1-27.
#'
#' Olley, J.M., Roberts, R.G., Yoshida, H., Bowler, J.M., 2006. Single-grain optical dating of grave-infill
#' associated with human burials at Lake Mungo, Australia. Quaternary Science
#' Reviews 25, 2469-2474.
#'
#' **Further reading**
#'
#' Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
#' (De) distributions: Implications for OSL dating of sediment mixtures.
#' Quaternary Geochronology 4, 204-230.
#'
#' Bolker, B., 2016. Maximum likelihood estimation analysis with the bbmle package.
#' In: Bolker, B., R Development Core Team, 2016. bbmle: Tools for General Maximum Likelihood Estimation.
#' R package version 1.0.18. [https://CRAN.R-project.org/package=bbmle]()
#'
#' Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
#' assessment of procedures for estimating burial dose. Quaternary Science
#' Reviews 25, 2475-2502.
#'
#' Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
#' Quaternary Geochronology 12, 98-106.
#'
#' Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
#' of optical dating of fluvial deposits.  Quaternary Geochronology 1, 109-120.
#'
#' Rodnight, H., 2008. How many equivalent dose values are needed to
#' obtain a reproducible distribution?. Ancient TL 26, 3-10.
#'
#' @examples
#'
#' ## Load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' # (1) Apply the minimum age model with minimum required parameters.
#' # By default, this will apply the un-logged 3-parameter MAM.
#' calc_MinDose(data = ExampleData.DeValues$CA1, sigmab = 0.1)
#'
#' \dontrun{
#' # (2) Re-run the model, but save results to a variable and turn
#' # plotting of the log-likelihood profiles off.
#' mam <- calc_MinDose(
#'  data = ExampleData.DeValues$CA1,
#'  sigmab = 0.1,
#'  plot = FALSE)
#'
#' # Show structure of the RLum.Results object
#' mam
#'
#' # Show summary table that contains the most relevant results
#' res <- get_RLum(mam, "summary")
#' res
#'
#' # Plot the log likelihood profiles retroactively, because before
#' # we set plot = FALSE
#' plot_RLum(mam)
#'
#' # Plot the dose distribution in an abanico plot and draw a line
#' # at the minimum dose estimate
#' plot_AbanicoPlot(data = mam,
#'                  main = "3-parameter Minimum Age Model",
#'                  polygon.col = "none",
#'                  hist = TRUE,
#'                  rug = TRUE,
#'                  summary = c("n", "mean", "mean.weighted", "median", "in.ci"),
#'                  centrality = res$de,
#'                  grid.col = "none",
#'                  line.col = "red",
#'                  line.label = paste0(round(res$de, 1), "\U00B1",
#'                                      round(res$de_err, 1), " Gy"),
#'                  bw = 0.1,
#'                  ylim = c(-25, 18),
#'                  summary.pos = "topleft",
#'                  mtext = bquote("Parameters: " ~
#'                                   sigma[b] == .(get_RLum(mam, "args")$sigmab) ~ ", " ~
#'                                   gamma == .(round(log(res$de), 1)) ~ ", " ~
#'                                   sigma == .(round(res$sig, 1)) ~ ", " ~
#'                                   rho == .(round(res$p0, 2))))
#'
#' # (3) Run the minimum age model with bootstrap
#' # NOTE: Bootstrapping is computationally intensive
#' # (3.1) run the minimum age model with default values for bootstrapping
#' calc_MinDose(data = ExampleData.DeValues$CA1,
#'              sigmab = 0.15,
#'              bootstrap = TRUE)
#'
#' # (3.2) Bootstrap control parameters
#' mam <- calc_MinDose(data = ExampleData.DeValues$CA1,
#'                     sigmab = 0.15,
#'                     bootstrap = TRUE,
#'                     bs.M = 300,
#'                     bs.N = 500,
#'                     bs.h = 4,
#'                     sigmab.sd = 0.06,
#'                     plot = FALSE)
#'
#' # Plot the results
#' plot_RLum(mam)
#'
#' # save bootstrap results in a separate variable
#' bs <- get_RLum(mam, "bootstrap")
#'
#' # show structure of the bootstrap results
#' str(bs, max.level = 2, give.attr = FALSE)
#'
#' # print summary of minimum dose and likelihood pairs
#' summary(bs$pairs$gamma)
#'
#' # Show polynomial fits of the bootstrap pairs
#' bs$poly.fits$poly.three
#'
#' # Plot various statistics of the fit using the generic plot() function
#' par(mfcol=c(2,2))
#' plot(bs$poly.fits$poly.three, ask = FALSE)
#'
#' # Show the fitted values of the polynomials
#' summary(bs$poly.fits$poly.three$fitted.values)
#' }
#'
#' @export
calc_MinDose <- function(
  data,
  sigmab,
  log = TRUE,
  par = 3,
  bootstrap = FALSE,
  init.values = NULL,
  level = 0.95,
  log.output = FALSE,
  plot = TRUE,
  multicore = FALSE,
  ...
) {
  .set_function_name("calc_MinDose")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(data, c("data.frame", "RLum.Results"))
  .validate_not_empty(data)
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")
  }
  if (ncol(data) < 2) {
    .throw_error("'data' should have 2 columns")
  }

  if (!all(stats::complete.cases(data))) {
    .throw_message("Warning: Input data contained NA/NaN values, ",
                   "which were removed prior to calculations", error = FALSE)
    data <- data[stats::complete.cases(data), ]
    if (nrow(data) == 0)
      .throw_error("After NA removal, nothing is left from the data set")
  }

  .validate_positive_scalar(sigmab)
  if (!is.null(init.values)) {
    .validate_class(init.values, "list")
    exp.names <- c("gamma", "sigma", "p0", "mu")
    mis.names <- setdiff(exp.names, names(init.values))
    if (length(init.values) < length(exp.names) || length(mis.names) > 0) {
      .throw_error("Please provide initial values for all model parameters. ",
                   "\nMissing parameters: ",
                   toString(mis.names))
    }
  }

  ## par can only be 3 or 4
  .validate_positive_scalar(par, int = TRUE)
  if (!par %in% c(3, 4)) {
    .throw_error("'par' can only be set to 3 or 4")
  }

  .validate_positive_scalar(level)
  .validate_logical_scalar(log)
  .validate_logical_scalar(bootstrap)
  .validate_logical_scalar(log.output)
  .validate_logical_scalar(plot)
  .validate_logical_scalar(multicore)

  ##============================================================================##
  ## ... ARGUMENTS
  ##============================================================================##

  extraArgs <- list(...)

  ## check if this function is called by calc_MaxDose()
  if ("invert" %in% names(extraArgs)) {
    invert <- .validate_logical_scalar(extraArgs$invert, name = "'invert'")
    if (invert && !log) {
      log <- TRUE # overwrite user choice as max dose model currently only supports the logged version
      .throw_warning("The maximum dose model only supports the logged version, ",
                     "'log' reset to TRUE\n")
    }
  } else {
    invert <- FALSE
  }

  ## console output
  verbose <- extraArgs$verbose %||% TRUE
  .validate_logical_scalar(verbose)

  ## bootstrap replications
  # first level bootstrap
  if ("bs.M" %in% names(extraArgs)) {
    M <- .validate_positive_scalar(as.integer(extraArgs$bs.M),
                                   int = TRUE, name = "'bs.M'")
    M <- max(M, 2) # issue 900
  } else {
    M <- 1000
  }

  # second level bootstrap
  if ("bs.N" %in% names(extraArgs)) {
    N <- .validate_positive_scalar(as.integer(extraArgs$bs.N),
                                   int = TRUE, name= "'bs.N'")
    N <- max(N, 2) # issue 1355
  } else {
    N <- 3*M
  }

  # KDE bandwith
  if ("bs.h" %in% names(extraArgs)) {
    h <- .validate_positive_scalar(extraArgs$bs.h, name = "'bs.h'")
  } else {
    h <- sd(data[, 1]) / sqrt(nrow(data)) * 2
  }

  # standard deviation of sigmab
  if ("sigmab.sd" %in% names(extraArgs)) {
    sigmab.sd <- .validate_positive_scalar(extraArgs$sigmab.sd,
                                           name = "'sigmab.sd'")
  } else {
    sigmab.sd <- 0.04
  }

  if ("debug" %in% names(extraArgs)) {
    debug <- .validate_logical_scalar(extraArgs$debug, name = "'debug'")
  } else {
    debug <- FALSE
  }

  if (multicore && "cores" %in% names(extraArgs)) {
    cores <- .validate_positive_scalar(extraArgs$cores,
                                       int = TRUE, name = "'cores'")
  } else {
    cores <- ifelse(multicore, parallel::detectCores(), 1)
    if (multicore && verbose)
      message("Logical CPU cores detected: ", cores) # nocov
  }

  ## remove non-positive values if using log-transformation
  if (log) {
    nonpos <- data[, 1] <= 0
    if (sum(nonpos) > 0) {
      .throw_warning("De values must be positive with 'log = TRUE', ",
                     sum(nonpos), " values set to NA")
      data[nonpos, 1] <- NA
    }
  }


  ##============================================================================##
  ## START VALUES
  ##============================================================================##

  if (is.null(init.values)) {
    start <- list(gamma = ifelse(log, log(quantile(data[ ,1], probs = 0.25, na.rm = TRUE)),
                                 quantile(data[ ,1], probs = 0.25, na.rm = TRUE)),
                  sigma = 1.2,
                  p0 = 0.01,
                  mu = ifelse(log, log(quantile(data[ ,1], probs = 0.25, na.rm = TRUE)),
                              mean(data[ ,1])))
  } else {
    start <- list(gamma = ifelse(log, log(init.values$gamma), init.values$gamma),
                  sigma = ifelse(log, log(init.values$sigma), init.values$sigma),
                  p0 = init.values$p0,
                  mu = ifelse(log, log(init.values$mu), init.values$mu))
  }

  ##============================================================================##
  ## ESTIMATE BOUNDARY PARAMETERS
  ##============================================================================##

  boundaries <- list(
    # gamma.lower = min(data[ ,1]/10),
    # gamma.upper = max(data[ ,1]*1.1),
    # sigma.lower = 0,
    # sigma.upper = 5,
    # mu.lower = min(data[ ,1])/10,
    # mu.upper = max(data[ ,1]*1.1)
    gamma.lower = -Inf,
    gamma.upper = Inf,

    sigma.lower = 0,
    sigma.upper = Inf,

    p0.lower = 0,
    p0.upper = 1,

    mu.lower = -Inf,
    mu.upper = Inf
  )

  boundaries <- modifyList(boundaries, list(...))

  ##============================================================================##
  ## AUXILIARY FUNCTIONS
  ##============================================================================##

  # THIS FUNCTION CALCULATES THE NEGATIVE LOG LIKELIHOOD OF THE DATA
  Neglik_f <- function(gamma, sigma, p0, mu, data) {
    # this calculates the negative of the log likelihood of the
    # data (data) for a given set of parameters (gamma, sigma, p0)
    # data is a 2x2 matrix of data: De, rel_error (including sigma_b)

    # recover the data
    zi <- data[ ,1]
    si <- data[ ,2]

    # in the MAM-3 gamma and mu are assumed to be equal
    if (par == 3)
      mu <- gamma

    # calculate sigma^2 + seld^2, mu0 and sigma0
    s2 <- sigma^2 + si^2
    sigma0 <- 1/sqrt(1/sigma^2 + 1/si^2)
    mu0 <- (mu/sigma^2 + zi/si^2)/(1/sigma^2 + 1/si^2)

    # calculate the log-likelihood
    logsqrt2pi <- 0.5*log(2*pi)
    res0 <- (gamma - mu0)/sigma0
    res1 <- (gamma - mu)/sigma
    lf1i <- log(p0) - log(si) - 0.5*((zi-gamma)/si)^2   - logsqrt2pi
    lf2i <- log1p(-p0) - 0.5 * log(s2) - 0.5 * (zi - mu)^2 / s2  - logsqrt2pi
    lf2i <- lf2i + log1p(-stats::pnorm(res0)) - log1p(-stats::pnorm(res1))
    llik <- log( exp(lf1i) + exp(lf2i) )

    ## negative log-likelihood
    -sum(llik)
  }

  # THIS MAXIMIZES THE Neglik_f LIKELIHOOD FUNCTION AND RETURNS AN MLE OBJECT
  Get_mle <- function(data) {
    tryCatch({
      suppressWarnings(
        mle <- bbmle::mle2(data = list(data = data),
                           optimizer = "nlminb",
                           lower=c(gamma = boundaries$gamma.lower,
                                   sigma = boundaries$sigma.lower,
                                   p0 = boundaries$p0.lower,
                                   mu = boundaries$mu.lower),
                           upper=c(gamma = boundaries$gamma.upper,
                                   sigma = boundaries$sigma.upper,
                                   p0 = boundaries$p0.upper,
                                   mu = boundaries$mu.upper),
                           minuslogl = Neglik_f,
                           control = list(iter.max = 1000L),
                           start = start)
      )
    }, error = function(e) {
      .throw_error("Maximizing the negative log-likelihood failed: ", e$message) # nocov
    })
    mle
  }

  ## Function to extract the estimate of gamma from mle2 objects and converting
  ## it back to the 'normal' scale
  save_Gamma <- function(ests) {
    m <- bbmle::coef(ests)[["gamma"]]
    if (invert)
      m <- -1 * (m - x.offset)
    if (log)
      m <- exp(m)
    m
  }

  ## Undo the invert transformation on a confidence interval data frame
  undo_invert <- function(conf) {
    conf[1, ] <- (conf[1, ] - x.offset) * -1
    tmp <- conf[1, 1]
    conf[1, 1] <- conf[1,2]
    conf[1, 2] <- tmp
    conf
  }

  ## Collect the parameter estimates from mle2 objects, converting back
  ## log-transformed values to the normal scale
  extract_estimates <- function(ests) {
    res <- data.frame(gamma = save_Gamma(ests),
                      sigma = bbmle::coef(ests)[["sigma"]],
                      p0 = bbmle::coef(ests)[["p0"]],
                      mu = if (par == 4) bbmle::coef(ests)[["mu"]] else NA,
                      row.names = "Estimate", check.names = FALSE)
    if (log) {
      res$sigma <- exp(res$sigma)
      res$mu <- exp(res$mu)
    }

    res
  }

  ##============================================================================##
  ## MAIN PROGRAM
  ##============================================================================##

  ## adds the sigmab error to each individual De error
  combine_Errors <- function(data, err) {
    if (log) {
      lcd  <- log(data[, 1]) * ifelse(invert, -1, 1)
      if (invert) {
        lcd <- lcd + x.offset
      }
      lse <- sqrt((data[, 2] / data[, 1])^2 + err^2)
    } else {
      lcd <- data[, 1]
      lse <- sqrt(data[, 2]^2 + err^2)
    }
    cbind(lcd, lse)
  }

  ## create new data frame with DE and combined relative error
  x.offset <- 0
  dat <- combine_Errors(data, sigmab)

  ## recompute the offset if necessary
  if (log && invert) {
    x.offset <- abs(min(dat[, 1], na.rm = TRUE))
    dat[, 1] <- dat[, 1] + x.offset
  }

  ## remove NAs which may be present if we had non-positive De values with
  ## log = TRUE
  dat <- dat[!is.na(dat[, 1]), ]

  # get the maximum likelihood estimate
  ests <- Get_mle(dat)

  # check if any standard errors are NA or NaN
  coef_err <- suppressWarnings(
    t(as.data.frame(bbmle::summary(ests)@coef[ ,2]))
  )

  if (debug)
    print(suppressWarnings(bbmle::summary(ests)))

  ## deal with NA/NaN values
  if (anyNA(coef_err)) {
    na.idx <- is.na(coef_err)
    coef_err[na.idx] <- t(as.data.frame(ests@coef))[na.idx] / 100
  }

  which <- c("gamma", "sigma", "p0", if (par == 4) "mu")
  prof.lower <- c(gamma = -Inf, sigma = 0, p0 = 0, mu = -Inf)
  prof.upper <- c(gamma = Inf, sigma = Inf, p0 = 1, mu = Inf)

  # calculate profile log likelihoods
  prof <- suppressWarnings(
    bbmle::profile(ests,
                   which = which,
                   std.err = as.vector(coef_err),
                   #try_harder = TRUE,
                   quietly = TRUE,
                   tol.newmin = Inf,
                   skiperrs = TRUE,
                   prof.lower = prof.lower,
                   prof.upper = prof.upper)
  )
  # Fallback when profile() returns a 'better' fit
  maxsteps <- 100
  cnt <- 1
  while (!inherits(prof, "profile.mle2")) {
    if (maxsteps == 0L)
      .throw_error("Couldn't find a converging fit for the profile log-likelihood")
    if (verbose)
      message("## Trying to find a better fit (", cnt, "/10) ##")
    prof <- suppressWarnings(
      bbmle::profile(ests,
                     which = which,
                     std.err = as.vector(coef_err),
                     # try_harder = TRUE,
                     quietly = TRUE,
                     maxsteps = maxsteps,
                     tol.newmin = Inf,
                     skiperrs = TRUE,
                     prof.lower = prof.lower,
                     prof.upper = prof.upper)
    )
    maxsteps <- maxsteps - 10
    cnt <- cnt + 1
  }

  ## delete rows where z = -Inf/Inf or NaN
  prof@profile$gamma <- prof@profile$gamma[is.finite(prof@profile$gamma$z), ]
  prof@profile$sigma <- prof@profile$sigma[is.finite(prof@profile$sigma$z), ]
  prof@profile$p0 <- prof@profile$p0[is.finite(prof@profile$p0$z), ]
  if (par == 4) {
    prof@profile$mu <- prof@profile$mu[is.finite(prof@profile$mu$z), ]
  }

  ## calculate Bayesian Information Criterion (BIC)
  BIC <- BIC(ests)
  Lmax <- -ests@min

  ##============================================================================##
  ## ERROR CALCULATION

  #### METHOD 1: follow the instructions of Galbraith & Roberts (2012) ####
  # "If the likelihood profile is symmetrical about the parameter, an approximate standard error
  #  can be calculated by dividing the length of this interval by 3.92"
  conf <- suppressWarnings(
    as.data.frame(bbmle::confint(prof, tol.newmin = Inf, quietly = TRUE, level = level))
  )
  class(conf[,1]) <- class(conf[,2]) <- "numeric"

  if (invert) {
    conf <- undo_invert(conf)
  }

  ## keep a copy of the original conf object to be returned at the end
  conf.orig <- conf
  if (log) {
    logged.rows <- row.names(conf) != "p0"
    conf[logged.rows, ] <- exp(conf[logged.rows, ])
  }

  gamma_err <- (conf["gamma", 2] - conf["gamma", 1]) / 3.92

  ## retrieve results from mle2 object
  res <- extract_estimates(ests)

  ##============================================================================##
  ## BOOTSTRAP
  ##============================================================================##
  if (bootstrap) {

    # Function that produces N+M replicates from the original data set using
    # randomly sampled indices with replacement and adding a randomly drawn
    # sigmab error
    create_Replicates <- function(R, e) {
      d <- apply(R, 1, function(x) data[x, ])
      mapply(function(x, y) combine_Errors(x, y), d, e, SIMPLIFY = FALSE)
    }

    # Function that takes each of the N replicates and produces a kernel density
    # estimate of length n.
    get_KDE <- function(d) {
      f <- approx(density(x = d[, 1], kernel = "gaussian", bw = h, na.rm = TRUE),
                  xout = d[, 1])
      pStarTheta <- as.vector(f$y / sum(f$y))
      pStarTheta / (1 / n)
    }

    # Function that calculates the product term of the recycled bootstrap
    get_ProductTerm <- function(Pmat, freq) {
      apply(Pmat^freq, 1, prod)
    }

    # Function that calculates the pseudo likelihoods for M replicates and
    # returns the dose-likelihood pairs
    make_Pairs <- function(theta, b2mam, prodterm) {
      kdthis <- vapply(theta, function(x) (x - b2mam) / h, numeric(N))
      kd1 <- dnorm(kdthis)
      kd2 <- kd1 * prodterm
      kd <- colSums(kd2, na.rm = TRUE)
      likelihood <- kd / (N * h)
      cbind(theta, likelihood)
    }

    ## Function to produce smoothed polynomial fits
    ## USS: unshared systematic error (3.5% in the paper)
    # nocov start
    get_polyfit <- function(pairs, degree, USS = 0.035) {
      x <- pairs[, 1]
      pairs[, 1] <- suppressWarnings(rnorm(length(x), mean = x, sd = USS * x))
      pairs_sorted <- pairs[order(pairs[, 1]), ]
      x <- pairs_sorted[, 1]
      y <- pairs_sorted[, 2]

      ## remove non-positive values as we are applying the log
      ypos <- y > 0
      x <- x[ypos]
      y <- y[ypos]

      ## fit polynomial to log(y), then exponentiate back
      fit <- stats::lm(log(y) ~ poly(x, degree, raw = TRUE), na.action = na.exclude)
      f <- exp(predict(fit, newdata = data.frame(x)))

      ## vector of points to interpolate fit
      xi <- seq(min(x, na.rm = TRUE), max(x, na.rm = TRUE), length.out = 100)
      yi <- approx(x, f, xout = xi)$y

      ind1 <- which(yi > 0)
      yi[!is.finite(yi)] <- 0

      list(xycurve = cbind(xi[min(ind1):max(ind1)],
                           yi[min(ind1):max(ind1)]),
           pairsUSS = pairs,
           fit = fit)
    }
    # nocov end

    ## START BOOTSTRAP ----
    msg <- sprintf(paste("\n [calc_MinDose] \n\nRecycled Bootstrap",
                         "\n\nParameters:",
                         "\n M = %d",
                         "\n N = %d",
                         "\n sigmab = %.2f \U00B1 %.2f",
                         "\n h = %.2f",
                         "\n\nCreating %d bootstrap replicates..."),
                   M, N, sigmab, sigmab.sd, h, N+M)
    if (verbose)
      message(msg)

    n <- length(data[ ,1])
    # Draw N+M samples from a normally distributed sigmab
    sigmab.bs <- rnorm(N + M, sigmab, sigmab.sd)

    ## Draw N+M random indices of size n
    indices <- matrix(sample(n, size = n * (N + M), replace = TRUE),
                      nrow = N + M, ncol = n, byrow = TRUE)

    # Finally draw N+M bootstrap replicates
    replicates <- create_Replicates(indices, sigmab.bs)

    # MULTICORE: The call to 'Get_mle' is the bottleneck of the function.
    # Using multiple CPU cores can reduce the computation cost, but may
    # not work for all machines.
    if (verbose) {
      message("Applying the model to all replicates using ", cores,
              " cores, this may take a while...")
    }

    if (multicore) {
      cl <- parallel::makeCluster(cores)
      mle <- parallel::parLapply(cl, replicates, Get_mle)
      parallel::stopCluster(cl)
    } else {
      mle <- lapply(replicates, Get_mle)
    }

    # Final bootstrap calculations
    if (verbose)
      message("Calculating likelihoods...")
    # Save 2nd- and 1st-level bootstrap results (i.e. estimates of gamma)
    b2mam <- sapply(mle[1:N], save_Gamma)
    theta <- sapply(mle[c(N+1):c(N+M)], save_Gamma)

    # Calculate the probability/pseudo-likelihood
    Pmat <- lapply(replicates[c(N+1):c(N+M)], get_KDE)
    freq <- t(apply(indices[1:N, ], 1, tabulate, n))
    prodterm <- vapply(Pmat, get_ProductTerm, freq, FUN.VALUE = numeric(N))

    # Save the bootstrap results as dose-likelihood pairs
    pairs <- make_Pairs(theta, b2mam, prodterm)

    ## --------- FIT POLYNOMIALS -------------- ##
    if (verbose)
      message("Fitting curves to dose-likelihood pairs...")
    # polynomial fits of increasing degrees

    ## if the input values are too close to zero, we may get
    ## Inf values >>> we remove them here with a warning
    is.inf <- is.infinite(pairs[, 2])
    if (any(is.inf)) {
      # nocov start
      .throw_warning("Inf values produced by bootstrapping removed for loess ",
                     "fitting (", round(sum(is.inf) / nrow(pairs) * 100, 2),
                     "% of the total dataset). This message usually indicates ",
                     "that your values are close to 0.")
      pairs <- pairs[!is.inf, ]
      # nocov end
    }

    poly.three <- stats::lm(pairs[, 2] ~ poly(pairs[, 1], degree = 3, raw = TRUE))
    poly.four <- stats::lm(pairs[, 2] ~ poly(pairs[, 1], degree = 4, raw = TRUE))
    poly.five <- stats::lm(pairs[, 2] ~ poly(pairs[, 1], degree = 5, raw = TRUE))
    poly.six <- stats::lm(pairs[, 2] ~ poly(pairs[, 1], degree = 6, raw = TRUE))

    ## --------- FIT LOESS -------------- ##
    # Polynomials are probably not reasonable and often suffer badly from
    # overfitting, especially towards the margins of the fitted data. In this
    # particular use case polynomials may suggest a multimodal likelihood
    # distribution where actually none is given. The non-parametric
    # LOESS (LOcal polynomial regrESSion) often yields better results than
    # standard polynomials.
    if (length(unique(pairs[, 1])) >= 15) {
      loess <- loess(pairs[, 2] ~ pairs[, 1])
    } else {
      loess <- NA
      .throw_warning("Not enough bootstrap replicates for loess fitting, try ",
                     "increasing `bs.M`")
    }

    ## extract the parameters from the bootstrap replicates
    mle <- rbindlist(lapply(mle[N + 1:M], extract_estimates))

    ## compute the mean of each parameter
    res <- data.frame(apply(mle, 2, mean, simplify = FALSE))
    rownames(res) <- "Estimate"

    ## quantiles from the bootstrap replicates
    prob <- (1 + c(-level, level)) / 2
    conf <- data.frame(t(apply(mle[, 1:par], 2, quantile,
                               na.rm = TRUE, probs = prob)))
    colnames(conf) <- paste(prob * 100, "%")

    ## keep a copy of the conf object in the scale in which the models were
    ## fitted
    conf.orig <- conf
    if (log) {
      logged.rows <- row.names(conf.orig) != "p0"
      conf.orig[logged.rows, ] <- log(conf.orig[logged.rows, ])
    }

    ## standard deviation over the De values from the bootstrap replicates
    gamma_err <- sd(pairs[, "theta"])

  }#EndOf::Bootstrap

  ## ========================================================================
  ## AGGREGATE RESULTS

  summary <- data.frame(de = res$gamma,
                        de_err = gamma_err,
                        ci_level = level,
                        ci_lower = conf["gamma", 1],
                        ci_upper = conf["gamma", 2],
                        par = par,
                        sig = res$sigma,
                        p0 = res$p0,
                        mu = res$mu,
                        Lmax = Lmax,
                        BIC = BIC)
  call <- sys.call()
  args <- list(log = log, sigmab = sigmab, par = par, bootstrap = bootstrap,
               init.values = start, log.output = log.output,
               bs.M = M, bs.N = N, bs.h = h, sigmab.sd = sigmab.sd)

  if (!is.na(summary$mu) && !is.na(summary$de) &&
    ## equivalent to log(summary$de) > summary$mu, but also valid if de < 0
    summary$de > exp(summary$mu)) {
      .throw_warning("Gamma is larger than mu, consider running the model ",
                     "with new boundary values (see details '?calc_MinDose')")
  }

  ##============================================================================##
  ## CONSOLE PRINT
  ##============================================================================##

  if (verbose) {
    cat("\n----------- Meta data -----------\n")
    print(data.frame(n = length(data[, 1]),
                       par=par,
                       sigmab=sigmab,
                       logged=log,
                       Lmax = Lmax,
                       BIC=BIC,
                       row.names = ""))

    cat(sprintf("\n--- Final parameter estimates%s ---\n",
                ifelse(bootstrap, " (bootstrap)", "")))
    if (log && log.output) {
      res$`log(gamma)` <- log(res$gamma)
      res$`log(sigma)` <- log(res$sigma)
      res$`log(mu)` <- log(res$mu)
    }
    print(round(res[, 1:par], 2))

    cat(sprintf("\n------ Confidence intervals%s -----\n",
                ifelse(bootstrap, " (bootstrap)", "")))
    conf_print <- conf
    conf_print <- round(conf_print, 2)

    if (log && log.output) {
      logged.rows <- row.names(conf) != "p0"
      conf_log <- conf
      conf_log[logged.rows, ] <- log(conf_log[logged.rows, ])

      conf_log <- round(conf_log, 2)
      conf_log[which(rownames(conf_log) == "p0"), ] <- "-"
      conf_print <- cbind(conf_print, conf_log)
      ncols <- ncol(conf_print) / 2
      conf_print <- rbind(setNames(data.frame(as.list(rep("", ncols)),
                                              as.list(rep("(logged)", ncols)),
                                              row.names = ""),
                                   colnames(conf_print)),
                          conf_print)
    }

    print(conf_print)

    cat(sprintf("\n------ De (asymmetric error)%s -----\n",
                ifelse(bootstrap, " (bootstrap)", "")))
    print(round(data.frame(De = summary$de,
                           lower = conf["gamma", 1],
                           upper = conf["gamma", 2],
                           row.names = ""), 2))

    cat(sprintf("\n------ De (symmetric error)%s -----\n",
                ifelse(bootstrap, " (bootstrap)", "")))
    print(round(data.frame(De = summary$de,
                           error = gamma_err,
                           row.names = ""), 2))
  }

  ##============================================================================##
  ## RETURN VALUES
  ##============================================================================##

  if (invert)
    prof@profile$gamma$par.vals[ ,"gamma"] <- rev((prof@profile$gamma$par.vals[ ,"gamma"] - x.offset)*-1)

  if (!bootstrap)
    pairs <- poly.three <- poly.four <- poly.five <- poly.six <- loess <- NULL

  results <- set_RLum(
    class = "RLum.Results",
    originator = "calc_MinDose",
    data = list(summary = summary,
                data = data,
                args = args,
                call = call,
                mle = ests,
                BIC = BIC,
                confint = conf.orig,
                profile = prof,
                bootstrap = list(
                  pairs = list(gamma=pairs),
                  poly.fits = list(poly.three = poly.three,
                                   poly.four = poly.four,
                                   poly.five = poly.five,
                                   poly.six = poly.six),
                  loess.fit = loess)))

  ##=========##
  ## PLOTTING
  if (plot) {
    try(plot_RLum.Results(results, ...),
        outFile = stdout()) # redirect error messages so they can be silenced
  }

  results
}

## load data
data(ExampleData.DeValues, envir = environment())
temp <- calc_MinDose(data = ExampleData.DeValues$CA1,
                     sigmab = 0.1,
                     verbose = FALSE,
                     plot = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_MinDose(),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_MinDose("test"),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_MinDose(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(calc_MinDose(iris[, 1, drop = FALSE]),
               "'data' should have 2 columns")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1),
               "'sigmab' should be a single positive value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 1, init.values = 1:4),
               "'init.values' should be of class 'list'")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            init.values = list(1, 2, 3)),
               "Please provide initial values for all model parameters")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            init.values = list(p0 = 0, p1 = 1, p2 = 2, mu = 3)),
               "Missing parameters: gamma, sigma")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1, par = "error"),
               "'par' should be a single positive integer value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1, par = 2),
               "'par' can only be set to 3 or 4")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1, level = 0),
               "'level' should be a single positive value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            invert = "error"),
               "'invert' should be a single logical value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            verbose = "error"),
               "'verbose' should be a single logical value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            debug = "error"),
               "'debug' should be a single logical value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            sigmab.sd = c(0.01, 0.02)),
               "'sigmab.sd' should be a single positive value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            bootstrap = TRUE, bs.M = -1),
               "'bs.M' should be a single positive integer value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            bootstrap = TRUE, bs.N = -1),
               "'bs.N' should be a single positive integer value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            bootstrap = TRUE, bs.h = -1),
               "'bs.h' should be a single positive value")
  expect_error(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                            cores = -1, multicore = TRUE),
               "'cores' should be a single positive integer value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  ## bootstrap
  expect_message(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                              bootstrap = TRUE, bs.M = 10, bs.N = 5, bs.h = 5,
                              sigmab.sd = 0.04, debug = TRUE, log = FALSE,
                              multicore = TRUE, cores = 2),
                 "Applying the model to all replicates using 2 cores")
  })

  ## RLum.Results object
  calc_MinDose(temp, sigmab = 0.1, verbose = FALSE, log = FALSE, par = 4,
               init.values = list(gamma = 54, sigma = 1, p0 = 0.01, mu = 70))

  ## missing values
  data.na <- ExampleData.DeValues$CA1
  data.na[1, 1] <- NA
  expect_message(calc_MinDose(data.na, sigmab = 0.1, verbose = FALSE),
                 "Input data contained NA/NaN values, which were removed")
  expect_message(expect_error(
      calc_MinDose(data.frame(data.na[, 1], NA)),
      "After NA removal, nothing is left from the data set"),
      "Input data contained NA/NaN values, which were removed")

  ## no converging fit
  skip_on_os("windows")
  set.seed(7)
  data.nofit <- data.frame(rep(4, 5), rnorm(5, 5))
  SW({
  expect_error(calc_MinDose(data.nofit, sigmab = 0.9, par=4),
               "Couldn't find a converging fit for the profile log-likelihood")
  })

  ## more coverage
  SW({
  expect_warning(calc_MinDose(ExampleData.DeValues$CA1 / 100, sigmab = 0.1,
                              par = 4, gamma.lower = 2, log.output = TRUE,
                              bootstrap = TRUE, bs.M = 10, bs.N = 5, bs.h=100),
                 "Gamma is larger than mu, consider running the model with new")
  expect_warning(calc_MinDose(ExampleData.DeValues$CA1[1:9, ], sigmab = 0.8,
                             bootstrap = TRUE, bs.M = 5, bs.N = 5, bs.h = 5,
                             debug = TRUE, log = FALSE),
                 "Not enough bootstrap replicates for loess fitting")
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(temp, tolerance = snapshot.tolerance)
  SW({
  expect_snapshot_RLum(calc_MinDose(data = ExampleData.DeValues$CA1 / 100,
                                    sigmab = 0.2, gamma.upper = 4, par = 4,
                                    log.output = TRUE, plot = FALSE),
                       tolerance = snapshot.tolerance)
  })
  expect_snapshot_RLum(calc_MinDose(data = ExampleData.DeValues$CA1,
                                    sigmab = 0.2, log = FALSE,
                                    verbose = FALSE, plot = FALSE),
                       tolerance = snapshot.tolerance)

  ## bootstrap = TRUE
  suppressWarnings( # Not enough bootstrap replicates for loess fitting
  expect_snapshot_RLum(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                                    bootstrap = TRUE, bs.M = 10, bs.N = 5,
                                    verbose = FALSE, plot = FALSE),
                       tolerance = snapshot.tolerance)
  )
  expect_snapshot_RLum(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.2,
                                    invert = TRUE, bootstrap = TRUE,
                                    bs.M = 20, bs.N = 5, bs.h = 10,
                                    verbose = FALSE, plot = FALSE),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 2.1,
                                    bootstrap = TRUE, log = FALSE, par = 4,
                                    bs.M = 20, bs.N = 5, bs.h = 10,
                                    verbose = FALSE, plot = FALSE),
                       tolerance = snapshot.tolerance)
})

test_that("output snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  SW({
  expect_snapshot_output(res <- calc_MinDose(ExampleData.DeValues$CA1,
                                             sigmab = 0.1))
  expect_snapshot_output(res <- calc_MinDose(ExampleData.DeValues$CA1,
                                             sigmab = 0.1, invert = TRUE))
  expect_snapshot_output(res <- calc_MinDose(data = ExampleData.DeValues$CA1 / 100,
                                             sigmab = 0.2, gamma.upper = 4, par = 4,
                                             log.output = TRUE, plot = FALSE))
  expect_snapshot_output(res <- calc_MinDose(data = ExampleData.DeValues$CA1,
                                             sigmab = 0.2, log = FALSE,
                                             plot = FALSE))
  suppressWarnings( # Not enough bootstrap replicates for loess fitting
  expect_snapshot_output(res <- calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                                             bootstrap = TRUE, bs.M = 10, bs.N = 5,
                                             plot = FALSE))
  )
  expect_snapshot_output(res <- calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.2,
                                             invert = TRUE, bootstrap = TRUE,
                                             bs.M = 20, bs.N = 5, bs.h = 10,
                                             plot = FALSE))
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              calc_MinDose(ExampleData.DeValues$CA1,
                                           sigmab = 0.1))
  vdiffr::expect_doppelganger("invert",
                              calc_MinDose(ExampleData.DeValues$CA1,
                                           sigmab = 0.1,
                                           invert = TRUE))
  vdiffr::expect_doppelganger("small sigmab",
                              calc_MinDose(ExampleData.DeValues$CA1,
                                           sigmab = 0.009))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 898
  expect_silent(calc_MinDose(ExampleData.DeValues$CA1,
                             sigmab = 0.009, invert = TRUE,
                             verbose = FALSE, plot = FALSE))

  ## issue 900
  expect_warning(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                              bootstrap = TRUE, bs.M = 1,
                              verbose = FALSE, plot = FALSE))

  ## issue 1332
  ## the seed was picked to get the smallest number of warnings and messages;
  ## this test relies on not using SW() to do its job
  set.seed(6)
  expect_warning(expect_warning(expect_message(
      calc_MinDose(data = data.frame(De = c(rnorm(4) + 5, -1),
                                     De_Err = rnorm(5) + 1),
                   sigmab = 1, log = TRUE, bootstrap = TRUE,
                   bs.M = 10, bs.N = 5, bs.h = 2, verbose = FALSE),
      "Unable to plot the likelihood profile for: p0"),
      "De values must be positive with 'log = TRUE', 1 values set to NA"),
      "Not enough bootstrap replicates for loess fitting, try increasing `bs.M`")

  ## issue 1355
  expect_warning(calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                              bootstrap = TRUE, bs.M = 10, bs.N = 1,
                              verbose = FALSE, plot = FALSE))
})

#' @title Convert portable OSL data to a Risoe.BINfileData object
#'
#' @description
#' Converts an [Luminescence::RLum.Analysis-class] object produced by the function [Luminescence::read_PSL2R] to
#' a [Luminescence::Risoe.BINfileData-class] object **(BETA)**.
#'
#' This function converts an [Luminescence::RLum.Analysis-class] object that was produced
#' by the [Luminescence::read_PSL2R] function to a [Luminescence::Risoe.BINfileData-class].
#' The [Luminescence::Risoe.BINfileData-class] can be used to write a Risø BIN file via
#' [Luminescence::write_R2BIN].
#'
#' @param object [Luminescence::RLum.Analysis-class] (**required**):
#' `RLum.Analysis` object produced by [Luminescence::read_PSL2R]
#'
#' @param ... currently not used.
#'
#' @return
#' Returns an S4 [Luminescence::Risoe.BINfileData-class] object that can be used to write a
#' BIN file using [Luminescence::write_R2BIN].
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::Risoe.BINfileData-class]
#'
#' @author
#' Christoph Burow, University of Cologne (Germany)
#'
#' @section Function version: 0.0.1
#'
#' @keywords IO
#'
#' @examples
#'
#' # (1) load and plot example data set
#' data("ExampleData.portableOSL", envir = environment())
#' plot_RLum(ExampleData.portableOSL)
#'
#' # (2) merge all RLum.Analysis objects into one
#' merged <- merge_RLum(ExampleData.portableOSL)
#' merged
#'
#' # (3) convert to RisoeBINfile object
#' bin <- convert_PSL2Risoe.BINfileData(merged)
#' bin
#'
#' # (4) write Risoe BIN file
#' \dontrun{
#' write_R2BIN(bin, "~/portableOSL.binx")
#' }
#'
#' @export
convert_PSL2Risoe.BINfileData <- function(object, ...) {
  .set_function_name("convert_PSL2Risoe.BINfileData")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, "RLum.Analysis")
  .validate_not_empty(object)
  lapply(object, .validate_class, "RLum.Data.Curve",
         name = "All elements of 'object'")
  lapply(object, .validate_originator, "read_PSL2R",
         name = "At least one element of 'object'")

  ## EXTRACT CURVE INFORMATION ----
  curves <- get_RLum(object)

  ## COLLECT META INFORMATION ----
  META <- do.call(rbind, lapply(curves, function(x) {

    NPOINTS <- as.integer(x@info$settings$stimulation_time)
    LTYPE <- x@info$settings$stimulation_unit
    COMMENT <- x@info$settings$measurement
    HIGH <- x@info$settings$stimulation_time
    DATE <- format(x@info$settings$Date, format = "%d%m%y")
    TIME <- x@info$settings$Time
    if (nchar(TIME) < 8)
      TIME <- paste0("0", TIME)
    SAMPLE <- x@info$settings$Sample
    FNAME <- x@info$settings$Filename
    SEQUENCE <- strtrim(paste(x@info$settings$Run_Name, x@info$settings$Sample_no), 8)


    return(data.frame(NPOINTS = NPOINTS,
                LTYPE = LTYPE,
                COMMENT = COMMENT,
                HIGH = HIGH,
                DATE = DATE,
                TIME = TIME,
                SAMPLE = SAMPLE,
                FNAME = FNAME,
                SEQUENCE = SEQUENCE))
  }))

  ## SAVE DATA ----
  DATA <- lapply(curves, function(x) {
    as.integer(x@data[ ,2])
  })

  # SAVE METADATA ----
  METADATA <- data.frame(ID = seq(1, length(curves), 1),
                         SEL = rep(TRUE, length(curves)),
                         VERSION = rep(7, length(curves)),
                         LENGTH = 447 + 4 * META$NPOINTS,
                         PREVIOUS = 447 + 4 * META$NPOINTS,
                         NPOINTS = META$NPOINTS,
                         RUN = seq(1, length(curves), 1),
                         SET = rep(1, length(curves)),
                         POSITION = rep(1, length(curves)),
                         GRAIN = rep(0, length(curves)),
                         GRAINNUMBER = rep(0, length(curves)),
                         CURVENO = rep(0, length(curves)),
                         XCOORD = rep(0, length(curves)),
                         YCOORD = rep(0, length(curves)),
                         SAMPLE = META$SAMPLE,
                         COMMENT = META$COMMENT,
                         SYSTEMID = rep(0, length(curves)),
                         FNAME = META$FNAME,
                         USER = rep("RLum", length(curves)),
                         TIME = META$TIME,
                         DATE = META$DATE,
                         DTYPE = rep("Natural", length(curves)),
                         BL_TIME = rep(0, length(curves)),
                         BL_UNIT = rep(0, length(curves)),
                         NORM1 = rep(0, length(curves)),
                         NORM2 = rep(0, length(curves)),
                         NORM3 = rep(0, length(curves)),
                         BG = rep(0, length(curves)),
                         SHIFT = rep(0, length(curves)),
                         TAG = rep(1, length(curves)),
                         LTYPE = META$LTYPE,
                         LIGHTSOURCE = rep("None", length(curves)),
                         LPOWER = rep(100, length(curves)),
                         LIGHTPOWER = rep(100, length(curves)),
                         LOW = rep(0, length(curves)),
                         HIGH = META$HIGH,
                         RATE = rep(0, length(curves)),
                         TEMPERATURE = rep(0, length(curves)),
                         MEASTEMP = rep(0, length(curves)),
                         AN_TEMP = rep(0, length(curves)),
                         AN_TIME = rep(0, length(curves)),
                         TOLDELAY = rep(0, length(curves)),
                         TOLON = rep(0, length(curves)),
                         TOLOFF = rep(0, length(curves)),
                         IRR_TIME = rep(0, length(curves)),
                         IRR_TYPE = rep(0L, length(curves)),
                         IRR_UNIT = rep(0, length(curves)),
                         IRR_DOSERATE = rep(0, length(curves)),
                         IRR_DOSERATEERR = rep(0, length(curves)),
                         TIMESINCEIRR = rep(-1, length(curves)),
                         TIMETICK = rep(1e-07, length(curves)),
                         ONTIME = rep(0, length(curves)),
                         OFFTIME = rep(NA, length(curves)),
                         STIMPERIOD = rep(0, length(curves)),
                         GATE_ENABLED = rep(0, length(curves)),
                         ENABLE_FLAGS = rep(0, length(curves)),
                         GATE_START = rep(0, length(curves)),
                         GATE_STOP = rep(0, length(curves)),
                         PTENABLED = rep(0, length(curves)),
                         DTENABLED = rep(0, length(curves)),
                         DEADTIME = rep(0, length(curves)),
                         MAXLPOWER = rep(0, length(curves)),
                         XRF_ACQTIME = rep(0, length(curves)),
                         XRF_HV = rep(0, length(curves)),
                         XRF_CURR = rep(0, length(curves)),
                         XRF_DEADTIMEF = rep(0, length(curves)),
                         SEQUENCE = META$SEQUENCE,
                         DETECTOR_ID = rep(NA, length(curves)),
                         LOWERFILTER_ID = rep(NA, length(curves)),
                         UPPERFILTER_ID = rep(NA, length(curves)),
                         ENOISEFACTOR = rep(NA, length(curves)),
                         MARKPOS_X1 = rep(0, length(curves)),
                         MARKPOS_Y1 = rep(0, length(curves)),
                         MARKPOS_X2 = rep(0, length(curves)),
                         MARKPOS_Y2 = rep(0, length(curves)),
                         MARKPOS_X3 = rep(0, length(curves)),
                         MARKPOS_Y3 = rep(0, length(curves)),
                         EXTR_START = rep(0, length(curves)),
                         EXTR_END = rep(0, length(curves)),
                         RECTYPE = rep(0, length(curves)))

  ## CREATE Risoe.BINfileData OBJECT ----
  set_Risoe.BINfileData(METADATA = METADATA,
                        DATA = DATA,
                        .RESERVED = list())
}

#' @title Export PSL-file(s) to CSV-files
#'
#' @description This function is a wrapper function around the functions [Luminescence::read_PSL2R] and
#' [Luminescence::write_RLum2CSV] and it imports an PSL-file (SUERC portable OSL reader file format)
#' and directly exports its content to CSV-files.
#' If nothing is set for the argument `path` ([Luminescence::write_RLum2CSV]) the input folder will
#' become the output folder.
#'
#' @param file [character] (**required**):
#' name of the PSL-file to be converted to CSV-files
#'
#' @param extract_raw_data [logical] (*with default*): enable/disable raw data
#' extraction. The PSL files imported into R contain an element `$raw_data`, which
#' provides a few more information (e.g., count errors), sometimes it makes
#' sense to use this data of the more compact standard values created by [Luminescence::read_PSL2R]
#'
#' @param single_table [logical] (*with default*): enable/disable the creation
#' of single table with n rows and n columns, instead of separate [data.frame]
#' objects. Each curve will be represented by two columns for time and counts
#'
#' @param ... further arguments that will be passed to the function
#' [Luminescence::read_PSL2R] and [Luminescence::write_RLum2CSV]
#'
#' @return
#' The function returns either a CSV-file (or many of them) or, when
#' `export = FALSE`, a list of [data.frame] and [matrix] objects.
#'
#' @section Function version: 0.1.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data-class], [Luminescence::RLum.Results-class],
#' [utils::write.table], [Luminescence::write_RLum2CSV], [Luminescence::read_PSL2R]
#'
#' @keywords IO
#'
#' @examples
#'
#' ## export into single data.frame
#' file <- system.file("extdata/DorNie_0016.psl", package="Luminescence")
#' convert_PSL2CSV(file, export = FALSE, single_table = TRUE)
#'
#' \dontrun{
#' ##select your BIN-file
#' file <- file.choose()
#'
#' ##convert
#' convert_PSL2CSV(file)
#' }
#'
#' @export
convert_PSL2CSV <- function(
  file,
  extract_raw_data = FALSE,
  single_table = FALSE,
  ...
) {
  .set_function_name("convert_PSL2CSV")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(file, c("character", "RLum"))
  .validate_not_empty(file)
  .validate_logical_scalar(extract_raw_data)
  .validate_logical_scalar(single_table)

  ##set input arguments
  convert_PSL2R_settings.default <- list(
    drop_bg = FALSE,
    as_decay_curve = TRUE,
    smooth = FALSE,
    merge = if(single_table) TRUE else FALSE,
    export = TRUE,
    verbose = TRUE
  )

  ##modify list on demand
  convert_PSL2R_settings <- modifyList(x = convert_PSL2R_settings.default, val = list(...))

  # Import file ---------------------------------------------------------------------------------
  if(!inherits(file, "RLum")){
    object <- read_PSL2R(
      file = file,
      drop_bg = convert_PSL2R_settings$drop_bg,
      as_decay_curve = convert_PSL2R_settings$as_decay_curve,
      smooth = convert_PSL2R_settings$smooth,
      merge = convert_PSL2R_settings$merge,
      verbose = convert_PSL2R_settings$verbose
   )
  }else{
    object <- file
  }

  ## try to extract file name from object ... this will be needed later
  filename <- try({
    rev(strsplit(object@info$Datafile_Path, "\\", fixed = TRUE)[[1]])[1]
  }, silent = TRUE)

  # raw data ----------------------------------------------------------------
  ## extract raw data instead of conventional data
  if(extract_raw_data) {
    psl_raw <- lapply(object@records, function(x) x@info$raw_data)

    names(psl_raw) <- names(object)
    object <- psl_raw
  }

  # single_table ------------------------------------------------------------
  ## generate a single table
  if(single_table) {
    ## run the conversion to CSV objects
    if(inherits(object, "RLum")) {
      l <- convert_PSL2CSV(object, export = FALSE, compact = FALSE)

    } else {
      l <- object
    }

    ## get max row number
    nrow_max <- vapply(l, nrow, numeric(1))

    ## create super matrix
    m <- matrix(NA, nrow = max(nrow_max), ncol = length(nrow_max) * ncol(l[[1]]))

    ## fill matrix
    s <- matrix(seq_len(length(l) * ncol(l[[1]])), nrow = ncol(l[[1]]))
    for(i in 1:length(l)) {
      m[1:nrow(l[[i]]), s[1,i]:(rev(s[,i])[1])] <- as.matrix(l[[i]])

    }

    ## set column names
    if(!extract_raw_data) {
      colnames(m) <- paste0(rep(names(l), each = 2), c("_t", "_cts"))

    } else {
      colnames(m) <- paste0(
        rep(seq_along(l), each = ncol(l[[1]])),
        "_" ,
        rep(names(l), each = ncol(l[[1]])),
        "_",
        rep(colnames(l[[1]]), length(l)))
    }

    ## overwrite object
    object <- as.data.frame(m)

    ## if possible, provide the file name as attribute
    if(!inherits(filename, "try-error"))
      attr(object, "filename") <- gsub(".", "_", filename, fixed = TRUE)
  }

  # Export to CSV -------------------------------------------------------------------------------
  ##get all arguments we want to pass and remove the doubled one
  arguments <- c(
    list(
      object = object,
      col.names = if(single_table[1] || extract_raw_data[1]) TRUE else FALSE,
      export = convert_PSL2R_settings$export),
    list(...))
  arguments[duplicated(names(arguments))] <- NULL

  ## now modify list again to ensure that the user input is always respected
  arguments <- modifyList(arguments, val = list(...), keep.null = TRUE)

  ##this if-condition prevents NULL in the terminal
  if(convert_PSL2R_settings$export){
    invisible(do.call("write_RLum2CSV", arguments))

  }else{
    do.call("write_RLum2CSV", arguments)
  }
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_PSL2CSV(),
               "'file' should be of class 'character' or 'RLum'")
  expect_error(convert_PSL2CSV(character(0)),
               "'file' cannot be an empty character")
  expect_error(convert_PSL2CSV("filename", numeric()),
               "'extract_raw_data' should be a single logical value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## get file
  file <- system.file("extdata/DorNie_0016.psl", package = "Luminescence")

  ##the case where we have an object of type RLum
  expect_type(convert_PSL2CSV(read_PSL2R(file, verbose = FALSE), export = FALSE),
              "list")

  ##export FALSE
  expect_type(convert_PSL2CSV(file, export = FALSE, verbose = FALSE),
              "list")

  ##write to temp
  expect_silent(convert_PSL2CSV(file, export = TRUE, path = tempdir(),
                                verbose = FALSE))

  ##test single_table argument
  expect_type(convert_PSL2CSV(file, export = FALSE, single_table = TRUE,
                              verbose = FALSE),
              "list")

  ##test raw data extraction
  ## simple raw data extraction
  t <- expect_type(convert_PSL2CSV(file, export = FALSE,
                                   extract_raw_data = TRUE, verbose = FALSE),
                   "list")
  expect_length(t, 5)

  ## raw data extraction with single_table
  t <- expect_type(convert_PSL2CSV(file, export = FALSE, single_table = TRUE,
                                   extract_raw_data = TRUE, verbose = FALSE),
                   "list")
  expect_length(t, 1)
  expect_equal(nrow(t[[1]]), 100)

  ## test with files export
  tmp_path <- tempdir()
  expect_silent(convert_PSL2CSV(file, path = tmp_path, extract_raw_data = TRUE,
                                single_table = TRUE, col.names = TRUE,
                                verbose = FALSE))

  ## test with col.names
  df <- read.table(file = rev(list.files(path = tmp_path, pattern = ".csv", full.names = TRUE))[1], sep = ";", header = TRUE)
  expect_type(colnames(df), "character")
  expect_true(grepl(pattern = "USER", colnames(df)[1]))

  ## test without column names
  expect_silent(convert_PSL2CSV(file, path = tmp_path,
                                extract_raw_data = TRUE, single_table = TRUE,
                                col.names = FALSE, verbose = FALSE))
  df <- read.table(file = list.files(path = tmp_path, pattern = ".csv", full.names = TRUE)[1], sep = ";", header = TRUE)
  expect_false(grepl(pattern = "USER", colnames(df)[1]))
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_XSYG2CSV(),
               "'file' should be of class 'character' or 'RLum'")
  expect_error(convert_XSYG2CSV(character(0)),
               "'file' cannot be an empty character")
  expect_error(convert_XSYG2CSV(letters),
               "[convert_XSYG2CSV()] 'file' should have length 1",
               fixed = TRUE)
  expect_error(convert_PSL2CSV(file = "error"),
               "File '.*error' does not exist") # windows CI needs the regexp
  SW({
  expect_error(expect_message(convert_XSYG2CSV(file = "", export = FALSE),
                              "XML file not readable, nothing imported"),
               "'object' should be of class 'RLum', 'data.frame' or a 'list'")
  })
})

test_that("test convert functions", {
  testthat::skip_on_cran()

    data(ExampleData.XSYG, envir = environment())
    expect_type(convert_XSYG2CSV(OSL.SARMeasurement$Sequence.Object[1:10],
                                 export = FALSE), "list")
    expect_null(convert_XSYG2CSV(OSL.SARMeasurement$Sequence.Object[1:10],
                                 export = TRUE, path = tempdir()))
})

#' @title Export Risoe.BINfileData into Risø BIN/BINX-file
#'
#' @description Exports a [Luminescence::Risoe.BINfileData-class] object in a `*.bin` or `*.binx` file that can be
#' opened by the Analyst software or other Risø software.
#'
#' @details
#' The structure of the exported binary data follows the data structure
#' published in the Appendices of the *Analyst* manual p. 42.
#'
#' If
#' `LTYPE`, `DTYPE` and `LIGHTSOURCE` are not of type
#' [character], no transformation into numeric values is done.
#'
#' @param object [Luminescence::Risoe.BINfileData-class] (**required**):
#' input object to be stored in a bin file.
#'
#' @param file [character] (**required**):
#' file name and path of the output file
#'
#' - `[WIN]`: `write_R2BIN(object, "C:/Desktop/test.bin")`
#' - `[MAC/LINUX]`: `write_R2BIN("/User/test/Desktop/test.bin")`
#'
#' @param version [character] (*optional*):
#' version number for the output file. If no value is provided, the highest
#' version number from the [Luminescence::Risoe.BINfileData-class] is taken automatically.
#'
#' **Note:**
#' This argument can be used to convert BIN-file versions.
#'
#' @param compatibility.mode [logical] (*with default*):
#' this option recalculates the position values if necessary and set the max.
#' value to 48. The old position number is appended as comment (e.g., 'OP: 70).
#' This option accounts for potential compatibility problems with the Analyst software.
#' It further limits the maximum number of points per curve to 9,999. If a curve contains more
#' data the curve data get binned using the smallest possible bin width.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar. Ignored if `verbose = FALSE`.
#'
#' @return Write a binary file and returns the name and path of the file as [character].
#'
#' @note
#' The function just roughly checks the data structures. The validity of
#' the output data depends on the user.
#'
#' The validity of the file path is not further checked. BIN-file conversions
#' using the argument `version` may be a lossy conversion, depending on the
#' chosen input and output data (e.g., conversion from version 08 to 07 to 06 to 05 to 04 or 03).
#'
#' **Warning**
#'
#' Although the coding was done carefully, it seems that the BIN/BINX-files
#' produced by Risø DA 15/20 TL/OSL readers slightly differ on the byte level.
#' No obvious differences are observed in the METADATA, however, the
#' BIN/BINX-file may not fully compatible, at least not similar to the ones
#' directly produced by the Risø readers!
#'
#' @section Function version: 0.5.4
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @note
#' ROI definitions (introduced in BIN-file version 8) are not supported!
#' There are furthermore ignored by the function [Luminescence::read_BIN2R].
#'
#' @seealso [Luminescence::read_BIN2R], [Luminescence::Risoe.BINfileData-class], [writeBin]
#'
#' @references
#' DTU Nutech, 2016. The Sequence Editor, Users Manual, February, 2016.
#' [https://www.fysik.dtu.dk]()
#'
#' @keywords IO
#'
#' @examples
#' ##load example dataset
#' file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
#' temp <- read_BIN2R(file)
#'
#' ##create temporary file path
#' ##(for usage replace by own path)
#' temp_file <- tempfile(pattern = "output", fileext = ".binx")
#'
#' ##export to temporary file path
#' write_R2BIN(temp, file = temp_file)
#'
#' @export
write_R2BIN <- function(
  object,
  file,
  version,
  compatibility.mode = FALSE,
  verbose = TRUE,
  txtProgressBar = TRUE
) {
  .set_function_name("write_R2BIN")
  on.exit(.unset_function_name(), add = TRUE)

  # Config ------------------------------------------------------------------
  ##set supported BIN format version
  VERSION.supported <- as.raw(c(3, 4, 5, 6, 7, 8))

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, "Risoe.BINfileData")
  .validate_class(file, "character")
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(txtProgressBar)

  ## check if it fulfills the latest definition ...
  if(ncol(object@METADATA) != ncol(set_Risoe.BINfileData()@METADATA)){
    .throw_error("Your Risoe.BINfileData object is not compatible with the ",
                 "latest specification of this S4-class object. You are ",
                 "probably trying to export a Risoe.BINfileData from your ",
                 "workspace you produced manually or with an old version. ",
                 "Please re-import the BIN-file using function read_BIN2R().")
  }

  # Check Risoe.BINfileData Struture ----------------------------------------
  ##check whether the BIN-file DATA slot contains more than 9999 records; needs to be run all the time
  temp_check <- lengths(object@DATA) > 9999L

  ##force compatibility
  if(compatibility.mode && any(temp_check)){
    .throw_warning("Compatibility mode selected: some data sets have ",
                   "more than 9,999 points and will be binned")

    ##BIN data to reduce amount of data if the BIN-file is too long
    object@DATA <- lapply(object@DATA, function(x){
      if(length(x) > 9999){
        ##we want to have a minimum binning (smallest number possible)
        bin_width <- ceiling(length(x)/9999)

        ##it should be symmetric, thus, remove values
        if((length(x)/bin_width)%%2 != 0){
          x <- x[-length(x)]
        }

        ##create matrix and return
        colSums(matrix(x, nrow = bin_width))

      }else{
        x
      }
    })

    ##reset temp_check
    temp_check <- FALSE

    ##get new number of points
    temp_NPOINTS <- lengths(object@DATA)

    ##correct LENGTH
    object@METADATA[["LENGTH"]] <- object@METADATA[["LENGTH"]] - (4 * object@METADATA[["NPOINTS"]]) + (temp_NPOINTS * 4)

    ##correct PREVIOUS
    object@METADATA[["PREVIOUS"]] <- c(0, head(object@METADATA$LENGTH, -1))

    ##correct NPOINTS
    object@METADATA[["NPOINTS"]] <- temp_NPOINTS

    ##write comment
    object@METADATA[["COMMENT"]] <- paste(object@METADATA[["COMMENT"]], " - binned")
  }

  if(any(temp_check))
    .throw_error(sum(temp_check), " out of ", length(temp_check),
                 " records contain more than 9,999 data points. ",
                 "This violates the BIN/BINX-file definition")

  ##remove
  rm(temp_check)

  ## UTF-8 conversion
  conv <- function(x) base::iconv(x, "latin1", "ASCII", sub = "_")
  object@METADATA[["SAMPLE"]] <- conv(object@METADATA[["SAMPLE"]])
  object@METADATA[["COMMENT"]] <- conv(object@METADATA[["COMMENT"]])
  object@METADATA[["FNAME"]] <- conv(object@METADATA[["FNAME"]])
  object@METADATA[["USER"]] <- conv(object@METADATA[["USER"]])
  object@METADATA[["SEQUENCE"]] <- conv(object@METADATA[["SEQUENCE"]])

  ##VERSION
  ##If missing version argument set to the highest value
  if(missing(version)){
    version <- as.raw(max(as.numeric(object@METADATA[,"VERSION"])))
    version.original <- version
  }else{
    version.original <- as.raw(max(as.numeric(object@METADATA[,"VERSION"])))
    version <- as.raw(version)
    object@METADATA[["VERSION"]] <- version

    ##Furthermore, entries length needed to be recalculated
    if(version.original != version){
      ##stepping decision
      header.stepping <- switch(
        EXPR = as.character(version),
        "08" = 507,
        "07" = 447,
        "06" = 447,
        "05" = 423,
        "04" = 272,
        "03" = 272)

      object@METADATA$LENGTH <- header.stepping + 4 * object@METADATA$NPOINTS
      object@METADATA$PREVIOUS <- c(0, head(object@METADATA$LENGTH, -1))
    }
  }

  ##Check if the BINfile object contains of unsupported versions
  if (!as.raw(object@METADATA$VERSION[1]) %in% VERSION.supported ||
      !version %in% VERSION.supported) {
    .throw_error("Writing BIN-files in format version (",
                 object@METADATA[1, "VERSION"], ") is currently not supported, ",
                 "supported version numbers are: ",
                 .collapse(VERSION.supported))
  }

  ## ensure that file name for version >= 05 has *.binx extension
  if (version >= 05 && tools::file_ext(file) == "bin") {
    file <- paste0(tools::file_path_sans_ext(file), ".binx")
  }

  ##SEQUENCE
  if (suppressWarnings(max(nchar(as.character(object@METADATA[,"SEQUENCE"]), type =
                                 "bytes"), na.rm = TRUE)) > 8) {
    .throw_error("Value in 'SEQUENCE' exceeds storage limit")
  }

  ##USER
  if (suppressWarnings(max(nchar(as.character(object@METADATA[,"USER"]), type =
                                 "bytes"), na.rm = TRUE)) > 8) {
    .throw_error("'USER' exceeds storage limit")
  }

  ##SAMPLE
  if (suppressWarnings(max(nchar(as.character(object@METADATA[,"SAMPLE"]), type =
                                 "bytes"), na.rm = TRUE)) > 20) {
    .throw_error("'SAMPLE' exceeds storage limit")
  }

  ## enables compatibility to the Analyst as the max value for POSITION becomes 48
  if (compatibility.mode && max(object@METADATA[, "POSITION"]) > 48) {
      ##grep relevant IDs
      temp.POSITION48.id <- which(object@METADATA[,"POSITION"]>48)

      ##find unique values
      temp.POSITION48.unique <- unique(object@METADATA[temp.POSITION48.id,"POSITION"])

      ##set translation vector starting from 1 and ending at 48
      temp.POSITION48.new <- rep_len(1:48, length.out = length(temp.POSITION48.unique))

      ## recalculate POSITION and update comment
      for(i in 1:length(temp.POSITION48.unique)){
        idx <- object@METADATA[, "POSITION"] == temp.POSITION48.unique[i]
        object@METADATA[idx, "COMMENT"] <- paste0(object@METADATA$COMMENT[idx],
                                                  "OP:", object@METADATA$POSITION[idx])
        object@METADATA[idx, "POSITION"] <- temp.POSITION48.new[i]
      }
  }

  ##COMMENT
  if(max(nchar(as.character(object@METADATA[,"COMMENT"]), type="bytes"))>80){
    .throw_error("'COMMENT' exceeds storage limit")
  }

  ## Convert strings to integers --------------------------------------------
  ## The ordering of these entries is important, as it allows us to convert
  ## character strings to the corresponding 0-based index (which must be an
  ## integer, hence the -1L below)

  LTYPE.Conv <- c("TL", "OSL", "IRSL", "M-IR", "M-VIS", "TOL", "TRPOSL",
                  "RIR", "RBR", "USER", "POSL", "SGOSL", "RL", "XRF")
  DTYPE.Conv <- c("Natural", "N+dose", "Bleach", "Bleach+dose",
                  "Natural (Bleach)", "N+dose (Bleach)", "Dose", "Background")
  LIGHT.Conv <- c("None", "Lamp", "IR diodes/IR Laser",
                  "Calibration LED", "Blue Diodes", "White light",
                  "Green laser (single grain)", "IR laser (single grain)")

  ##TRANSLATE VALUES IN METADATA
  ##LTYPE
  if (is.character(object@METADATA$LTYPE) || is.factor(object@METADATA$LTYPE)) {
    object@METADATA[, "LTYPE"] <- match(object@METADATA[, "LTYPE"],
                                        LTYPE.Conv) - 1L
  }

  ##DTYPE
  if (is.character(object@METADATA$DTYPE) || is.factor(object@METADATA$DTYPE)) {
    object@METADATA[, "DTYPE"] <- match(object@METADATA[, "DTYPE"],
                                        DTYPE.Conv) - 1L
  }

  ##LIGHTSOURCE
  if (is.character(object@METADATA$LIGHTSOURCE) ||
      is.factor(object@METADATA$LIGHTSOURCE)) {
    object@METADATA[,"LIGHTSOURCE"] <- match(object@METADATA[, "LIGHTSOURCE"],
                                             LIGHT.Conv) - 1L
  }

  ##TIME
  object@METADATA[, "TIME"] <- vapply(object@METADATA[["TIME"]], function(x) {
    if (is.na(x))
      return("000000")
    as.character(gsub(":", "", x))
  }, character(1))

  ##TAG and SEL
  ##in TAG information on the SEL are storred, here the values are copied to TAG
  ##before export
  object@METADATA[, "TAG"] <- as.integer(object@METADATA[, "SEL"])

  # SET FILE AND VALUES -----------------------------------------------------
  con <- file(file, "wb")
  on.exit(close(con), add = TRUE)

  ##get records
  n.records <- length(object@METADATA[,"ID"])

  ## don't show the progress bar if not verbose
  if (!verbose)
    txtProgressBar <- FALSE

  if (verbose) {
    cat("\n[write_R2BIN()] Exporting ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file)))
    cat("\n n_rec:", n.records)
    cat("\n")
  }

  ##set progressbar
  if(txtProgressBar)
    pb <- txtProgressBar(min=0,max=n.records, char="=", style=3)

  # LOOP -------------------------------------------------------------------
  ID <- 1
  if(version == 03 || version == 04){
    ## version 03 and 04
    ##start loop for export BIN data
    while(ID<=n.records) {

      ##VERSION
      writeBin(as.raw(object@METADATA[ID,"VERSION"]),
               con,
               size = 1,
               endian="little")

      ##stepping
      writeBin(raw(length=1),
               con,
               size = 1,
               endian="little")

      ##LENGTH, PREVIOUS, NPOINTS
      writeBin(c(as.integer(object@METADATA[ID,"LENGTH"]),
                 as.integer(object@METADATA[ID,"PREVIOUS"]),
                 as.integer(object@METADATA[ID,"NPOINTS"])),
               con,
               size = 2,
               endian="little")

      ##LTYPE
      writeBin(object@METADATA[ID,"LTYPE"],
               con,
               size = 1,
               endian="little")

      ##LOW, HIGH, RATE
      writeBin(c(as.double(object@METADATA[ID,"LOW"]),
                 as.double(object@METADATA[ID,"HIGH"]),
                 as.double(object@METADATA[ID,"RATE"])),
               con,
               size = 4,
               endian="little")

      ##TEMPERATURE, XCOORD, YCOORD, TOLDELAY; TOLON, TOLOFF
      writeBin(c(as.integer(object@METADATA[ID,"TEMPERATURE"]),
                 as.integer(object@METADATA[ID,"XCOORD"]),
                 as.integer(object@METADATA[ID,"YCOORD"]),
                 as.integer(object@METADATA[ID,"TOLDELAY"]),
                 as.integer(object@METADATA[ID,"TOLON"]),
                 as.integer(object@METADATA[ID,"TOLOFF"])),
               con,
               size = 2,
               endian="little")

      ##POSITION, RUN
      writeBin(c(as.integer(object@METADATA[ID,"POSITION"]),
                 as.integer(object@METADATA[ID,"RUN"])),
               con,
               size = 1,
               endian="little")

      ##TIME
      TIME_SIZE <- nchar(object@METADATA[ID,"TIME"])
      writeBin(as.integer(TIME_SIZE),
               con,
               size = 1,
               endian="little")

      writeChar(object@METADATA[ID,"TIME"],
                con,
                nchars =TIME_SIZE,
                useBytes=TRUE,
                eos = NULL)

      if(6-TIME_SIZE>0){
        writeBin(raw(length = c(6-TIME_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##DATE
      writeBin(6L,
               con,
               size = 1 ,
               endian="little")

      suppressWarnings(writeChar(as.character(object@METADATA[ID,"DATE"]),
                                 con,
                                 nchars = 6,
                                 useBytes=TRUE,
                                 eos = NULL))

      ##SEQUENCE
      ##count number of characters
      SEQUENCE_SIZE <- as.integer(
        nchar(as.character(object@METADATA[["SEQUENCE"]][ID]), type = "bytes", keepNA = FALSE))

      writeBin(SEQUENCE_SIZE,
               con,
               size = 1,
               endian="little")

      writeChar(as.character(object@METADATA[ID,"SEQUENCE"]),
                con,
                nchars = SEQUENCE_SIZE,
                useBytes=TRUE,
                eos = NULL)

      ##stepping
      if(8-SEQUENCE_SIZE>0){
        writeBin(raw(length = (8-SEQUENCE_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##USER
      USER_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"USER"]), type="bytes"))

      writeBin(USER_SIZE,
               con,
               size = 1,
               endian="little")

      writeChar(as.character(object@METADATA[ID,"USER"]),
                con,
                nchars = USER_SIZE,
                useBytes=TRUE,
                eos = NULL)

      ##stepping
      if(8-USER_SIZE>0){
        writeBin(raw(length = (8-USER_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##DTYPE
      writeBin(object@METADATA[ID,"DTYPE"],
               con,
               size = 1,
               endian="little")

      ##IRR_TIME
      writeBin(as.double(object@METADATA[ID,"IRR_TIME"]),
               con,
               size = 4,
               endian="little")


      ##IRR_TYPE, IRR_UNIT
      writeBin(c(object@METADATA[ID,"IRR_TYPE"],
                 object@METADATA[ID,"IRR_UNIT"]),
               con,
               size = 1,
               endian="little")


      ##BL_TIME
      writeBin(as.double(object@METADATA[ID,"BL_TIME"]),
               con,
               size = 4,
               endian="little")

      ##BL_UNIT
      writeBin(as.integer(object@METADATA[ID,"BL_UNIT"]),
               con,
               size = 1,
               endian="little")

      ##AN_TEMP, AN_TIME, NORM1, NORM2, NORM2, BG
      writeBin(c(as.double(object@METADATA[ID,"AN_TEMP"]),
                 as.double(object@METADATA[ID,"AN_TIME"]),
                 as.double(object@METADATA[ID,"NORM1"]),
                 as.double(object@METADATA[ID,"NORM2"]),
                 as.double(object@METADATA[ID,"NORM3"]),
                 as.double(object@METADATA[ID,"BG"])),
               con,
               size = 4,
               endian="little")

      ##SHIFT
      writeBin(as.integer(object@METADATA[ID,"SHIFT"]),
               con,
               size = 2,
               endian="little")

      ##SAMPLE
      SAMPLE_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"SAMPLE"]), type="bytes"))

      ##avoid problems with empty sample names
      if(SAMPLE_SIZE == 0){
        SAMPLE_SIZE <- 2L
        object@METADATA[ID,"SAMPLE"] <- "  "
      }

      writeBin(SAMPLE_SIZE,
               con,
               size = 1,
               endian="little")

      writeChar(as.character(object@METADATA[ID,"SAMPLE"]),
                con,
                nchars = SAMPLE_SIZE,
                useBytes=TRUE,
                eos = NULL)

      if((20-SAMPLE_SIZE)>0){
        writeBin(raw(length = (20-SAMPLE_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##COMMENT
      COMMENT_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"COMMENT"]), type="bytes"))

      ##avoid problems with empty comments
      if(COMMENT_SIZE == 0){
        COMMENT_SIZE <- 2L
        object@METADATA[ID,"COMMENT"] <- "  "
      }

      writeBin(COMMENT_SIZE,
               con,
               size = 1,
               endian="little")

      suppressWarnings(writeChar(as.character(object@METADATA[ID,"COMMENT"]),
                                 con,
                                 nchars = COMMENT_SIZE,
                                 useBytes=TRUE,
                                 eos = NULL))


      if((80-COMMENT_SIZE)>0){
        writeBin(raw(length = c(80-COMMENT_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##LIGHTSOURCE, SET, TAG
      writeBin(c(as.integer(object@METADATA[ID,"LIGHTSOURCE"]),
                 as.integer(object@METADATA[ID,"SET"]),
                 as.integer(object@METADATA[ID,"TAG"])),
               con,
               size = 1,
               endian="little")

      ##GRAIN
      writeBin(as.integer(object@METADATA[ID,"GRAIN"]),
               con,
               size = 2,
               endian="little")

      ##LPOWER
      writeBin(as.double(object@METADATA[ID,"LPOWER"]),
               con,
               size = 4,
               endian="little")

      ##SYSTEMID
      writeBin(as.integer(object@METADATA[ID,"SYSTEMID"]),
               con,
               size = 2,
               endian="little")

      ##Further distinction needed to fully support format version 03 and 04 separately
      if(version == 03){
        ##RESERVED 1
        if(length(object@.RESERVED) == 0 || version.original != version){
          writeBin(raw(length=36),
                   con,
                   size = 1,
                   endian="little")
        }else{
          writeBin(object = object@.RESERVED[[ID]][[1]],
                   con,
                   size = 1,
                   endian="little")
        }

        ##ONTIME, OFFTIME
        writeBin(c(as.integer(object@METADATA[ID,"ONTIME"]),
                   as.integer(object@METADATA[ID,"OFFTIME"])),
                 con,
                 size = 4,
                 endian="little")

        ##GATE_ENABLED
        writeBin(as.integer(object@METADATA[ID,"GATE_ENABLED"]),
                 con,
                 size = 1,
                 endian="little")


        ##GATE_START, GATE_STOP
        writeBin(c(as.integer(object@METADATA[ID,"GATE_START"]),
                   as.integer(object@METADATA[ID,"GATE_STOP"])),
                 con,
                 size = 4,
                 endian="little")


        ##RESERVED 2
        if(length(object@.RESERVED) == 0 || version.original != version){
          writeBin(raw(length=1),
                   con,
                   size = 1,
                   endian="little")

        }else{
          writeBin(object@.RESERVED[[ID]][[2]],
                   con,
                   size = 1,
                   endian="little")
        }

      } else {
        ##version 04
        ##RESERVED 1
        if(length(object@.RESERVED) == 0 || version.original != version){
          writeBin(raw(length=20),
                   con,
                   size = 1,
                   endian="little")
        } else{
          writeBin(object@.RESERVED[[ID]][[1]],
                   con,
                   size = 1,
                   endian="little")
        }

        ##CURVENO
        writeBin(as.integer(object@METADATA[ID,"CURVENO"]),
                 con,
                 size = 2,
                 endian="little")

        ##TIMETICK
        writeBin(c(as.double(object@METADATA[ID,"TIMETICK"])),
                 con,
                 size = 4,
                 endian="little")

        ##ONTIME, STIMPERIOD
        writeBin(c(as.integer(object@METADATA[ID,"ONTIME"]),
                   as.integer(object@METADATA[ID,"STIMPERIOD"])),
                 con,
                 size = 4,
                 endian="little")

        ##GATE_ENABLED
        writeBin(as.integer(object@METADATA[ID,"GATE_ENABLED"]),
                 con,
                 size = 1,
                 endian="little")

        ##GATE_START, GATE_STOP
        writeBin(c(as.integer(object@METADATA[ID,"GATE_START"]),
                   as.integer(object@METADATA[ID,"GATE_STOP"])),
                 con,
                 size = 4,
                 endian="little")

        ##PTENABLED
        writeBin(as.integer(object@METADATA[ID,"PTENABLED"]),
                 con,
                 size = 1,
                 endian="little")

        ##RESERVED 2
        if(length(object@.RESERVED) == 0 || version.original != version){
          writeBin(raw(length=10),
                   con,
                   size = 1,
                   endian="little")
        } else {
          writeBin(object@.RESERVED[[ID]][[2]],
                   con,
                   size = 1,
                   endian="little")
        }
      }
      ##DPOINTS
      writeBin(as.integer(unlist(object@DATA[ID])),
               con,
               size = 4,
               endian="little")

      #SET UNIQUE ID
      ID <- ID + 1
      ##update progress bar
      if(txtProgressBar) setTxtProgressBar(pb, ID)
    }
  }
  ## ====================================================
  ## version > 06
  if (version == 05 || version == 06 || version == 07 || version == 08) {
    ##start loop for export BIN data
    while(ID<=n.records) {
      ##VERSION
      writeBin(as.raw(object@METADATA[ID,"VERSION"]),
               con,
               size = 1,
               endian="little")

      ##stepping
      writeBin(raw(length=1),
               con,
               size = 1,
               endian="little")

      ##LENGTH, PREVIOUS, NPOINTS
      writeBin(c(as.integer(object@METADATA[ID,"LENGTH"]),
                 as.integer(object@METADATA[ID,"PREVIOUS"]),
                 as.integer(object@METADATA[ID,"NPOINTS"])),
               con,
               size = 4,
               endian="little")

      if(version == 08){
        writeBin(as.integer(object@METADATA[ID,"RECTYPE"]),
                 con,
                 size = 1,
                 endian="little")
      }

      ##RUN, SET, POSITION, GRAINNUMBER, CURVENO, XCOORD, YCOORD
      writeBin(c(as.integer(object@METADATA[ID,"RUN"]),
                 as.integer(object@METADATA[ID,"SET"]),
                 as.integer(object@METADATA[ID,"POSITION"]),
                 as.integer(object@METADATA[ID,"GRAINNUMBER"]),
                 as.integer(object@METADATA[ID,"CURVENO"]),
                 as.integer(object@METADATA[ID,"XCOORD"]),
                 as.integer(object@METADATA[ID,"YCOORD"])),
               con,
               size = 2,
               endian="little")

      ##SAMPLE
      SAMPLE_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"SAMPLE"]), type="bytes"))

      ##avoid problems with empty sample names
      if(SAMPLE_SIZE == 0){
        SAMPLE_SIZE <- 2L
        object@METADATA[ID,"SAMPLE"] <- "  "
      }

      writeBin(SAMPLE_SIZE,
               con,
               size = 1,
               endian="little")

      writeChar(as.character(object@METADATA[ID,"SAMPLE"]),
                con,
                nchars = SAMPLE_SIZE,
                useBytes=TRUE,
                eos = NULL)

      if((20-SAMPLE_SIZE)>0){
        writeBin(raw(length = (20-SAMPLE_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##COMMENT
      COMMENT_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"COMMENT"]), type="bytes"))

      ##avoid problems with empty comments
      if(COMMENT_SIZE == 0){
        COMMENT_SIZE <- 2L
        object@METADATA[ID,"COMMENT"] <- "  "
      }

      writeBin(COMMENT_SIZE,
               con,
               size = 1,
               endian="little")

      writeChar(as.character(object@METADATA[ID,"COMMENT"]),
                con,
                nchars = COMMENT_SIZE,
                useBytes=TRUE,
                eos = NULL)

      if((80-COMMENT_SIZE)>0){
        writeBin(raw(length = c(80-COMMENT_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##Instrument and sequence characteristics
      ##SYSTEMID
      writeBin(as.integer(object@METADATA[ID,"SYSTEMID"]),
               con,
               size = 2,
               endian="little")

      ##FNAME
      FNAME_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"FNAME"]), type="bytes"))

      ## correct for case that this is of 0 length
      if (is.na(FNAME_SIZE) || length(FNAME_SIZE) == 0) FNAME_SIZE <- 0L

      writeBin(FNAME_SIZE,
               con,
               size = 1,
               endian="little")

      if(FNAME_SIZE>0) {
        writeChar(
          as.character(object@METADATA[ID,"FNAME"]),
          con,
          nchars = FNAME_SIZE,
          useBytes = TRUE,
          eos = NULL
        )
      }

      if((100-FNAME_SIZE)>0){
        writeBin(raw(length = c(100-FNAME_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##USER
      USER_SIZE <- as.integer(nchar(as.character(object@METADATA[ID,"USER"]), type="bytes"))

      writeBin(USER_SIZE,
               con,
               size = 1,
               endian="little")

      writeChar(as.character(object@METADATA[ID,"USER"]),
                con,
                nchars = USER_SIZE,
                useBytes=TRUE,
                eos = NULL)

      if((30-USER_SIZE)>0){
        writeBin(raw(length = c(30-USER_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##TIME
      TIME_SIZE <- nchar(object@METADATA[ID,"TIME"])

      writeBin(as.integer(TIME_SIZE),
               con,
               size = 1,
               endian="little")

      writeChar(object@METADATA[ID,"TIME"],
                con,
                nchars =TIME_SIZE,
                useBytes=TRUE,
                eos = NULL)

      if(6-TIME_SIZE>0){
        writeBin(raw(length = c(6-TIME_SIZE)),
                 con,
                 size = 1,
                 endian="little")
      }

      ##DATE
      writeBin(6L,
               con,
               size = 1 ,
               endian="little")


      suppressWarnings(writeChar(as.character(object@METADATA[ID,"DATE"]),
                                 con,
                                 nchars = 6,
                                 useBytes=TRUE,
                                 eos = NULL))

      ##Analysis
      ##DTYPE
      writeBin(object@METADATA[ID,"DTYPE"],
               con,
               size = 1,
               endian="little")

      ##BL_TIME
      writeBin(as.double(object@METADATA[ID,"BL_TIME"]),
               con,
               size = 4,
               endian="little")

      ##BL_UNIT
      writeBin(as.integer(object@METADATA[ID,"BL_UNIT"]),
               con,
               size = 1,
               endian="little")

      ##NORM1, NORM2, NORM3, BG
      writeBin(c(as.double(object@METADATA[ID,"NORM1"]),
                 as.double(object@METADATA[ID,"NORM2"]),
                 as.double(object@METADATA[ID,"NORM3"]),
                 as.double(object@METADATA[ID,"BG"])),
               con,
               size = 4,
               endian="little")

      ##SHIFT
      writeBin(as.integer(object@METADATA[ID,"SHIFT"]),
               con,
               size = 2,
               endian="little")

      ##TAG
      writeBin(c(as.integer(object@METADATA[ID,"TAG"])),
               con,
               size = 1,
               endian="little")

      ##RESERVED 1
      if(length(object@.RESERVED) == 0 || version.original != version){
        writeBin(raw(length=20),
                 con,
                 size = 1,
                 endian="little")
      }else{
        writeBin(object@.RESERVED[[ID]][[1]],
                 con,
                 size = 1,
                 endian="little")
      }

      ##Measurement characteristics
      ##LTYPE
      writeBin(object@METADATA[ID,"LTYPE"],
               con,
               size = 1,
               endian="little")

      ##LIGHTSOURCE
      writeBin(c(as.integer(object@METADATA[ID,"LIGHTSOURCE"])),
               con,
               size = 1,
               endian="little")

      ##LIGHTPOWER, LOW, HIGH, RATE
      writeBin(c(as.double(object@METADATA[ID,"LIGHTPOWER"]),
                 as.double(object@METADATA[ID,"LOW"]),
                 as.double(object@METADATA[ID,"HIGH"]),
                 as.double(object@METADATA[ID,"RATE"])),
               con,
               size = 4,
               endian="little")

      ##TEMPERATURE, MEASTEMP
      writeBin(c(as.integer(object@METADATA[ID,"TEMPERATURE"]),
                 as.integer(object@METADATA[ID,"MEASTEMP"])),
               con,
               size = 2,
               endian="little")

      ##AN_TEMP, AN_TIME
      writeBin(c(as.double(object@METADATA[ID,"AN_TEMP"]),
                 as.double(object@METADATA[ID,"AN_TIME"])),
               con,
               size = 4,
               endian="little")

      ##TOLDELAY; TOLON, TOLOFF
      writeBin(c(as.integer(object@METADATA[ID,"TOLDELAY"]),
                 as.integer(object@METADATA[ID,"TOLON"]),
                 as.integer(object@METADATA[ID,"TOLOFF"])),
               con,
               size = 2,
               endian="little")

      ##IRR_TIME
      writeBin(as.double(object@METADATA[ID,"IRR_TIME"]),
               con,
               size = 4,
               endian="little")

      ##IRR_TYPE
      writeBin(c(object@METADATA[ID,"IRR_TYPE"]),
               con,
               size = 1,
               endian="little")

      ##IRR_DOSERATE, IRR_DOSERATEERR
      if(version == 05){
        writeBin(as.double(object@METADATA[ID,"IRR_DOSERATE"]),
                 con,
                 size = 4,
                 endian="little")
      }else{
        writeBin(c(as.double(object@METADATA[ID,"IRR_DOSERATE"]),
                   as.double(object@METADATA[ID,"IRR_DOSERATEERR"])),
                 con,
                 size = 4,
                 endian="little")
      }

      ##TIMESINCEIRR
      writeBin(c(as.integer(object@METADATA[ID,"TIMESINCEIRR"])),
               con,
               size = 4,
               endian="little")

      ##TIMETICK
      writeBin(c(as.double(object@METADATA[ID,"TIMETICK"])),
               con,
               size = 4,
               endian="little")

      ##ONTIME, STIMPERIOD
      writeBin(c(suppressWarnings(as.integer(object@METADATA[ID,"ONTIME"])),
                 as.integer(object@METADATA[ID,"STIMPERIOD"])),
               con,
               size = 4,
               endian="little")

      ##GATE_ENABLED
      writeBin(as.integer(object@METADATA[ID,"GATE_ENABLED"]),
               con,
               size = 1,
               endian="little")

      ##GATE_START, GATE_STOP
      writeBin(c(as.integer(object@METADATA[ID,"GATE_START"]),
                 as.integer(object@METADATA[ID,"GATE_STOP"])),
               con,
               size = 4,
               endian="little")

      ##PTENABLED, DTENABLED
      writeBin(c(as.integer(object@METADATA[ID,"PTENABLED"]),
                 as.integer(object@METADATA[ID,"DTENABLED"])),
               con,
               size = 1,
               endian="little")

      ##DEADTIME, MAXLPOWER, XRF_ACQTIME, XRF_HV
      writeBin(c(as.double(object@METADATA[ID,"DEADTIME"]),
                 as.double(object@METADATA[ID,"MAXLPOWER"]),
                 as.double(object@METADATA[ID,"XRF_ACQTIME"]),
                 as.double(object@METADATA[ID,"XRF_HV"])),
               con,
               size = 4,
               endian="little")

      ##XRF_CURR
      writeBin(c(as.integer(object@METADATA[ID,"XRF_CURR"])),
               con,
               size = 4,
               endian="little")

      ##XRF_DEADTIMEF
      writeBin(c(as.double(object@METADATA[ID,"XRF_DEADTIMEF"])),
               con,
               size = 4,
               endian="little")

      ##add version support for V7
      if(version == 05){
        ##RESERVED 2
        if(length(object@.RESERVED) == 0 || version.original != version){
          writeBin(raw(length=4),
                   con,
                   size = 1,
                   endian="little")
        }else{
          writeBin(object@.RESERVED[[ID]][[2]],
                   con,
                   size = 1,
                   endian="little")
        }

      }else if(version == 06){

        ##RESERVED 2
        if(length(object@.RESERVED) == 0 || version.original != version){
          writeBin(raw(length=24),
                   con,
                   size = 1,
                   endian="little")
        }else{
          writeBin(object@.RESERVED[[ID]][[2]],
                   con,
                   size = 1,
                   endian="little")
        }

      }else{
        ##DETECTOR_ID
        writeBin(as.integer(object@METADATA[ID,"DETECTOR_ID"]),
                 con,
                 size = 1,
                 endian="little")

        ##LOWERFILTER_ID, UPPERFILTER_ID
        writeBin(c(as.integer(object@METADATA[ID,"LOWERFILTER_ID"]),
                   as.integer(object@METADATA[ID,"UPPERFILTER_ID"])),
                 con,
                 size = 2,
                 endian="little")

        ##ENOISEFACTOR
        writeBin(as.double(object@METADATA[ID,"ENOISEFACTOR"]),
                 con,
                 size = 4,
                 endian="little")


        ##VERSION 08
        if(version == 07){

          ##RESERVED 2
          if(length(object@.RESERVED) == 0 || version.original != version){
            writeBin(raw(length=15),
                     con,
                     size = 1,
                     endian="little")
          }else{
            writeBin(object@.RESERVED[[ID]][[2]],
                     con,
                     size = 1,
                     endian="little")
          }

        }else{

          ##MARKPOS POSITION and extraction
          writeBin(
            c(
              as.double(object@METADATA[ID, "MARKPOS_X1"]),
              as.double(object@METADATA[ID, "MARKPOS_Y1"]),
              as.double(object@METADATA[ID, "MARKPOS_X2"]),
              as.double(object@METADATA[ID, "MARKPOS_Y2"]),
              as.double(object@METADATA[ID, "MARKPOS_X3"]),
              as.double(object@METADATA[ID, "MARKPOS_Y3"]),
              as.double(object@METADATA[ID, "EXTR_START"]),
              as.double(object@METADATA[ID, "EXTR_END"])
            ),
            con,
            size = 4,
            endian = "little"
          )

          ##RESERVED 2
          if(length(object@.RESERVED) == 0 || version.original != version){
            writeBin(raw(length=42),
                     con,
                     size = 1,
                     endian="little")
          }else{
            writeBin(object@.RESERVED[[ID]][[2]],
                     con,
                     size = 1,
                     endian="little")
          }
        }

      }#end if version decision
      ##DPOINTS
      writeBin(as.integer(unlist(object@DATA[ID])),
               con,
               size = 4,
               endian="little")

      #SET UNIQUE ID
      ID <- ID + 1

      ##update progress bar
      if(txtProgressBar)  setTxtProgressBar(pb, ID)
    }
  }

  # ##close
  if(txtProgressBar) close(pb)

  ##output
  if (verbose)
    message("\t >> ", ID - 1, " records written successfully\n")

  ## return path
  invisible(file)
}

## set the path for the files
IRSL <- system.file("extdata/BINX_IRSL_LOG.TXT", package = "Luminescence")
OSL <- system.file("extdata/BINX_OSL_LOG.TXT", package = "Luminescence")
IRSL_BROKEN <- system.file("extdata/BINX_IRSL_LOG_BROKEN.TXT", package = "Luminescence")

test_that("input validation", {
  testthat::skip_on_cran()

  ## nothing for file
  expect_error(read_BINXLOG2R(), regexp = "'file' should be of class 'character' or 'list'")

  ## wrong file path
  expect_error(read_BINXLOG2R(file = "error"),
               "File '.*error' does not exist") # windows CI needs the regexp

  ## verbose wrong
  expect_error(read_BINXLOG2R(file = IRSL, verbose = "error"), regexp = "'verbose' should be a single logical value")

  ## broken file
  expect_error(read_BINXLOG2R(file = IRSL_BROKEN, verbose = FALSE), regexp = "The file does not seem to be a valid debugging log-file")
})

test_that("standard run", {
  testthat::skip_on_cran()

  ## simple input
  SW ({
  t <- expect_type(read_BINXLOG2R(file = IRSL), "list")
  expect_length(t, 1)
  expect_s4_class(t[[1]], "RLum.Analysis")
  })

  ## IRSL verbose false
  expect_type(read_BINXLOG2R(file = IRSL, verbose = FALSE), "list")

  ## OSL verbose false
  expect_type(read_BINXLOG2R(file = OSL, verbose = FALSE), "list")

  ## list verbose false
  t <- expect_type(read_BINXLOG2R(file = list(IRSL, OSL), verbose = FALSE), "list")
  expect_length(t, 2)
  expect_s4_class(t[[1]], "RLum.Analysis")

  ## check whether we can export
  expect_silent(Luminescence::write_R2BIN(
    object = Luminescence::convert_RLum2Risoe.BINfileData(t),
    file = tempfile(),
    version = 4,
    verbose = FALSE))
})

## load data
data(ExampleData.BINfileData, envir = environment())

## replace the with numeric and factors with characters
CWOSL.SAR.Data@METADATA$VERSION <- as.numeric(CWOSL.SAR.Data@METADATA$VERSION)
CWOSL.SAR.Data@METADATA[] <- lapply(CWOSL.SAR.Data@METADATA,
                                    function(x) {
                                      if (is.factor(x)) as.character(x) else x
                                    })

## combine with existing BIN-file object
empty <- set_Risoe.BINfileData()
new <- data.table::rbindlist(list(empty@METADATA, CWOSL.SAR.Data@METADATA),
                             fill = TRUE)
new <- set_Risoe.BINfileData(METADATA = as.data.frame(new, stringsAsFactors = FALSE),
                             DATA = CWOSL.SAR.Data@DATA)
new@METADATA[is.na(new@METADATA)] <- 0
new@METADATA$RECTYPE <- 1
new <- subset(new, ID == 1:2)

test_that("check functionality", {
  testthat::skip_on_cran()

  ##create files
  path <- tempfile()
  SW({
  write_R2BIN(object = new, file = paste0(path, "BINfile_V3.bin"), version = "03")
  write_R2BIN(object = new, file = paste0(path, "BINfile_V4.bin"), version = "04")
  write_R2BIN(object = new, file = paste0(path, "BINfile_V5.binx"), version = "05")
  write_R2BIN(object = new, file = paste0(path, "BINfile_V6.binx"), version = "06")
  write_R2BIN(object = new, file = paste0(path, "BINfile_V7.binx"), version = "07")
  write_R2BIN(object = new, file = paste0(path, "BINfile_V8.binx"), version = "08")
  })

  temp <- new
  temp@METADATA[1, "TIME"] <- "1215"
  temp@METADATA[1, "SEQUENCE"] <- "123456"
  temp@METADATA[1, "SAMPLE"] <- ""
  temp@METADATA[1, "COMMENT"] <- ""
  temp@.RESERVED <- list(val1 = c("a", "b"), val2 = c("c", "d"))
  expect_silent(
  write_R2BIN(object = temp, file = paste0(path, "BINfile_V3.bin"),
              version = "03", verbose = FALSE))
  SW({
  temp@METADATA[, "VERSION"] <- 4
  write_R2BIN(object = temp, file = paste0(path, "BINfile_V4.bin"),
              version = "04")
  temp@METADATA[, "VERSION"] <- 5
  write_R2BIN(object = temp, file = paste0(path, "BINfile_V5.binx"),
              version = "05")
  temp@METADATA[, "VERSION"] <- 6
  write_R2BIN(object = temp, file = paste0(path, "BINfile_V6.binx"),
              version = "06")
  temp@METADATA[, "VERSION"] <- 7
  write_R2BIN(object = temp, file = paste0(path, "BINfile_V7.binx"),
              version = "07")
  temp@METADATA[1, "TIME"] <- NA
  temp@METADATA[, "VERSION"] <- 8
  write_R2BIN(object = temp, file = paste0(path, "BINfile_V8.binx"),
              version = "08")

  ## trigger edge case
  temp@METADATA[, "FNAME"] <- numeric()
  expect_silent(write_R2BIN(object = temp, file = paste0(path, "BINfile_V8.binx"),
              version = "08", verbose = FALSE))
  })

  ## silent correction of the file extension
  SW({
  skip_on_os("windows") # FIXME(mcol): see commit 26889a6
  write_R2BIN(object = new, file = paste0(path, "BINfile_V8.bin"),
              version = "08")
  })

  ## check UTF-8 characters
  new_utf8 <- new
  new_utf8@METADATA$FNAME <- c("I do not belong in here \xb5m")
  t <- expect_silent(write_R2BIN(new_utf8, file = paste0(path, "BINfile_V8.bin"),
                     version = "08", verbose = FALSE))
  expect_type(object = t, type = "character")

  ## actually compare the files produced
  bin.v8 <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  new.v8 <- paste0(path, "new.binx")
  SW({
  write_R2BIN(read_BIN2R(bin.v8), file = new.v8)
  })
  expect_length(system(paste("diff", bin.v8, new.v8), intern = TRUE),
                0) # nothing is printed out if the files don't differ
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(write_R2BIN(object = new, file = FALSE),
               "'file' should be of class 'character'")
  expect_error(write_R2BIN(object = "a", file = ""),
               "'object' should be of class 'Risoe.BINfileData'")
  expect_error(suppressWarnings(write_R2BIN(object = set_Risoe.BINfileData(), file = "")))

  temp <- new
  temp@METADATA <- temp@METADATA[, 1:79]
  expect_error(write_R2BIN(object = temp, file = "test"),
               "Your Risoe.BINfileData object is not compatible with the latest")

  temp <- new
  temp@METADATA[1, "SEQUENCE"] <- "1234567890"
  expect_error(write_R2BIN(object = temp, file = "test"),
               "Value in 'SEQUENCE' exceeds storage limit")

  temp <- new
  temp@METADATA[1, "USER"] <- "1234567890"
  expect_error(write_R2BIN(object = temp, file = "test"),
               "'USER' exceeds storage limit")

  temp <- new
  temp@METADATA[1, "SAMPLE"] <- strrep("a", 25)
  expect_error(write_R2BIN(object = temp, file = "test"),
               "'SAMPLE' exceeds storage limit")

  temp <- new
  temp@DATA[[2]] <- 1:25000
  temp@METADATA[1, "POSITION"] <- strrep("a", 50)
  expect_error(write_R2BIN(object = temp, file = "test"),
               "records contain more than 9,999 data points")
  expect_warning(
    expect_error(write_R2BIN(object = temp, compatibility.mode = TRUE,
                             file = paste0(tempfile(), "BINfile_V8.binx")),
                 "'COMMENT' exceeds storage limit"),
    "some data sets have more than 9,999 points")
})

## load data
data(ExampleData.BINfileData, envir = environment())
test_file_MG <- test_file_SG <- CWOSL.SAR.Data
test_file_SG@METADATA$GRAIN <- 1

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_SG2MG(data.frame()),
               "'object' should be of class 'character' or 'Risoe.BINfileData'")
  expect_error(convert_SG2MG(character(0)),
               "'object' cannot be an empty character")
  expect_error(convert_SG2MG(c("error1", "error2")),
               "'object' should have length 1")
  expect_error(convert_SG2MG("error"),
               "File '.*error' does not exist") # windows CI needs the regexp
})

test_that("test functionality", {
  testthat::skip_on_cran()

  ## test pass through for pure multiple grain data
  expect_equal(convert_SG2MG(test_file_MG),
               test_file_MG)

  ## test write option
  tmp <- tempfile(fileext = ".bin")
  new <- gsub(".bin", "_SG.bin", tmp)
  SW({
  write_file_test <- write_R2BIN(
      read_BIN2R(file = test_path("_data/BINfile_V4.bin")),
      file = tmp)

  expect_s4_class(convert_SG2MG(tmp, write_file = TRUE, txtProgressBar = FALSE),
                  "Risoe.BINfileData")
  expect_true(file.exists(new))
  })

  unlink(c(tmp, new))
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  small <- subset(test_file_SG, RUN == 1)
  expect_snapshot_Risoe(convert_SG2MG(small))
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1415
  small <- subset(test_file_SG, POSITION == 1)
  expect_silent(convert_SG2MG(small))
})

xsyg <- system.file("extdata/XSYG_file.xsyg", package="Luminescence")
binx <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(extract_IrradiationTimes("fail"),
               "File '.*fail' does not exist") # windows CI needs the regexp
  expect_error(extract_IrradiationTimes(character(0)),
               "'object' cannot be an empty character")
  expect_error(extract_IrradiationTimes(letters),
               "'object' should have length 1")
  expect_error(extract_IrradiationTimes(tempdir()),
               "File '.*' does not exist") # windows CI needs the regexp
  expect_error(extract_IrradiationTimes(test_path("_data/BINfile_V3.bin")),
               "File extension 'bin' is not supported, only 'xsyg' is valid")
  expect_error(extract_IrradiationTimes(FALSE),
               "'object' should be of class 'character', 'RLum.Analysis' or a")
  expect_error(extract_IrradiationTimes(xsyg, return_same_as_input = "error"),
               "return_same_as_input' should be a single logical value")
  SW({
  expect_error(extract_IrradiationTimes(xsyg, file.BINX = "fail"),
               "File '.*fail' does not exist") # windows CI needs the regexp
  expect_error(extract_IrradiationTimes(xsyg, file.BINX = tempdir()),
               "File '.*' does not exist") # windows CI needs the regexp
  })
  expect_message(extract_IrradiationTimes(xsyg, file.BINX = binx,
                                          txtProgressBar = FALSE),
                 "XSYG-file and BINX-file do not contain similar entries")
  expect_warning(extract_IrradiationTimes(list(xsyg), file.BINX = binx,
                                          txtProgressBar = FALSE),
                 "'file.BINX' is not supported in self-call mode")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  res_a <- extract_IrradiationTimes(list(xsyg), txtProgressBar = FALSE,
                                    return_same_as_input = TRUE)
  expect_type(res_a, "list")
  expect_s4_class(res_a[[1]], "RLum.Analysis")

  ## apply the function to something previously imported via read_BIN2R
  SW({
  temp <- read_BIN2R(binx, fastForward = TRUE)
  temp <- expect_s4_class(extract_IrradiationTimes(temp)[[1]], "RLum.Results")
  })
  expect_type(temp$irr.times$START, "double")

  ## write results to BINX-file
  tmp_BINX <- tempfile(fileext = ".binx")
  file.copy(
    from = system.file("extdata/BINfile_V8.binx", package = "Luminescence"),
    to = tmp_BINX, overwrite = TRUE)
  expect_message(
    extract_IrradiationTimes(xsyg, file.BINX = tmp_BINX, txtProgressBar = FALSE),
    "XSYG-file and BINX-file do not contain similar entries, BINX-file update skipped")

  tmp_XSYG <- test_path("_data/xsyg-tests/XSYG_file_TL_CASE3.xsyg")
  tmp_BINX <- tempfile(fileext = ".binx")
  write_R2BIN(read_BIN2R(binx, position = 2, verbose = FALSE),
              tmp_BINX, verbose = FALSE)
  expect_message(extract_IrradiationTimes(tmp_XSYG, file.BINX = tmp_BINX,
                                          txtProgressBar = FALSE),
                 "'Time Since Irradiation' was redefined in the exported BINX-file")

  ## empty list
  expect_null(extract_IrradiationTimes(list()))

  ## special case extraction with startDate missing
  tmp <- read_XSYG2R(xsyg, verbose = FALSE, fastForward = TRUE)
    ## remove startDate
    tmp[[1]]@records[[9]]@info$startDate <- NULL
    ## test
    expect_s4_class(
      object = extract_IrradiationTimes(tmp[[1]]), class = "RLum.Results")

   ## use TL curves with NA
   tmp <- read_XSYG2R(xsyg, verbose = FALSE, fastForward = TRUE)
   expect_s4_class(extract_IrradiationTimes(tmp[[1]]), class = "RLum.Results")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(extract_IrradiationTimes(xsyg, txtProgressBar = FALSE),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(extract_IrradiationTimes(xsyg, txtProgressBar = FALSE,
                                                return_same_as_input = TRUE),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(extract_IrradiationTimes(list(xsyg), txtProgressBar = FALSE,
                                                recordType = list("OSL (UVVIS)")),
                       tolerance = snapshot.tolerance)
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1273
  expect_error(extract_IrradiationTimes(iris, file.BINX = NA),
               "'object' should be of class 'character', 'RLum.Analysis' or")

  ## issue 1466
  tmp <- read_XSYG2R(xsyg, verbose = FALSE, fastForward = TRUE)[[1]]
  tmp@records[[1]]@info$position <- NULL
  expect_s4_class(extract_IrradiationTimes(tmp),
                  "RLum.Results")
})

#' Class `"Risoe.BINfileData"`
#'
#' S4 class object for luminescence data in R. The object is produced as output
#' of the function [Luminescence::read_BIN2R].
#'
#' @name Risoe.BINfileData-class
#'
#' @docType class
#'
#' @slot METADATA Object of class "data.frame" containing the meta information for each curve.
#'
#' @slot DATA Object of class "list" containing numeric vector with count data.
#'
#' @slot .RESERVED Object of class "list" containing list of undocumented raw values for internal use only.
#'
#' @note
#'
#' **Internal METADATA - object structure**
#'
#' This structure is compatible with BIN/BINX-files version 03-08, however, it does not follow (in its
#' sequential arrangement) the manual provided by the manufacturer,
#' but an own structure accounting for the different versions.
#'
#' \tabular{rllll}{
#' **#** \tab **Name** \tab **Data Type** \tab **V** \tab **Description** \cr
#' `[,1]` \tab `ID`  \tab `numeric` \tab RLum \tab Unique record ID (same ID as in slot `DATA`)\cr
#' `[,2]` \tab `SEL` \tab `logic` \tab RLum \tab Record selection, not part official BIN-format, triggered by TAG\cr
#' `[,3]` \tab `VERSION` \tab `raw` \tab 03-08 \tab BIN-file version number \cr
#' `[,4]` \tab `LENGTH` \tab `integer` \tab 03-08 \tab Length of this record\cr
#' `[,5]` \tab `PREVIOUS` \tab `integer` \tab 03-08 \tab Length of previous record\cr
#' `[,6]` \tab `NPOINTS` \tab `integer` \tab 03-08 \tab Number of data points in the record\cr
#' `[,7]` \tab `RECTYPE` \tab `integer` \tab 08 \tab Record type \cr
#' `[,8]` \tab `RUN` \tab `integer` \tab 03-08 \tab Run number\cr
#' `[,9]` \tab `SET` \tab `integer` \tab 03-08 \tab Set number\cr
#' `[,10]`  \tab `POSITION` \tab  `integer` \tab 03-08 \tab Position number\cr
#' `[,11]` \tab `GRAIN` \tab `integer` \tab 03-04 \tab Grain number\cr
#' `[,12]` \tab `GRAINNUMBER` \tab `integer` \tab 05-08 \tab Grain number\cr
#' `[,13]` \tab `CURVENO` \tab `integer` \tab 05-08 \tab Curve number\cr
#' `[,14]` \tab `XCOORD` \tab `integer` \tab 03-08 \tab X position of a single grain\cr
#' `[,15]` \tab `YCOORD` \tab `integer` \tab 03-08 \tab Y position of a single grain\cr
#' `[,16]` \tab `SAMPLE` \tab `factor` \tab 03-08 \tab Sample name\cr
#' `[,17]` \tab `COMMENT` \tab `factor` \tab 03-08 \tab Comment name\cr
#' `[,18]` \tab `SYSTEMID` \tab `integer` \tab 03-08 \tab Risø system id\cr
#' `[,19]` \tab `FNAME` \tab `factor` \tab 05-08 \tab File name (*.bin/*.binx)\cr
#' `[,20]` \tab `USER` \tab `factor` \tab 03-08 \tab User name\cr
#' `[,21]` \tab `TIME` \tab `character` \tab 03-08 \tab Data collection time (`hh-mm-ss`)\cr
#' `[,22]` \tab `DATE` \tab `factor` \tab 03-08 \tab Data collection date (`ddmmyy`)\cr
#' `[,23]` \tab `DTYPE` \tab `character` \tab 03-08 \tab Data type\cr
#' `[,24]` \tab `BL_TIME` \tab `numeric` \tab 03-08 \tab Bleaching time\cr
#' `[,25]` \tab `BL_UNIT` \tab `integer` \tab 03-08 \tab Bleaching unit (mJ, J, s, min, h)\cr
#' `[,26]` \tab `NORM1` \tab `numeric` \tab 03-08 \tab Normalisation factor (1)\cr
#' `[,27]` \tab `NORM2` \tab `numeric` \tab 03-08 \tab Normalisation factor (2)\cr
#' `[,28]` \tab `NORM3` \tab `numeric` \tab 03-08 \tab Normalisation factor (3)\cr
#' `[,29]` \tab `BG` \tab `numeric` \tab 03-08 \tab Background level\cr
#' `[,30]` \tab `SHIFT` \tab `integer` \tab 03-08 \tab Number of channels to shift data\cr
#' `[,31]` \tab `TAG` \tab `integer` \tab 03-08 \tab Tag, triggers `SEL`\cr
#' `[,32]` \tab `LTYPE` \tab `character` \tab 03-08 \tab Luminescence type\cr
#' `[,33]` \tab `LIGHTSOURCE` \tab `character` \tab 03-08 \tab Light source\cr
#' `[,34]` \tab `LPOWER` \tab `numeric` \tab 03-08 \tab Optical stimulation power\cr
#' `[,35]` \tab `LIGHTPOWER` \tab `numeric` \tab 05-08 \tab Optical stimulation power\cr
#' `[,36]` \tab `LOW` \tab `numeric` \tab 03-08 \tab Low (temperature, time, wavelength)\cr
#' `[,37]` \tab `HIGH` \tab `numeric` \tab 03-08 \tab High (temperature, time, wavelength)\cr
#' `[,38]` \tab `RATE` \tab `numeric` \tab 03-08 \tab Rate (heating rate, scan rate)\cr
#' `[,39]` \tab `TEMPERATURE` \tab `integer` \tab 03-08 \tab Sample temperature\cr
#' `[,40]` \tab `MEASTEMP` \tab `integer` \tab 05-08 \tab Measured temperature\cr
#' `[,41]` \tab `AN_TEMP` \tab `numeric` \tab 03-08 \tab Annealing temperature\cr
#' `[,42]` \tab `AN_TIME` \tab `numeric` \tab 03-08 \tab Annealing time\cr
#' `[,43]` \tab `TOLDELAY` \tab `integer` \tab 03-08 \tab TOL 'delay' channels\cr
#' `[,44]` \tab `TOLON` \tab `integer` \tab 03-08 \tab TOL 'on' channels\cr
#' `[,45]` \tab `TOLOFF` \tab `integer` \tab 03-08 \tab TOL 'off' channels\cr
#' `[,46]` \tab `IRR_TIME` \tab `numeric` \tab 03-08 \tab Irradiation time\cr
#' `[,47]` \tab `IRR_TYPE` \tab `integer` \tab 03-08 \tab Irradiation type (alpha, beta or gamma)\cr
#' `[,48]` \tab `IRR_UNIT` \tab `integer` \tab 03-04 \tab Irradiation unit (Gy, rad, s, min, h)\cr
#' `[,49]` \tab `IRR_DOSERATE` \tab `numeric` \tab 05-08 \tab Irradiation dose rate (Gy/s)\cr
#' `[,50]` \tab `IRR_DOSERATEERR` \tab `numeric` \tab 06-08 \tab Irradiation dose rate error (Gy/s)\cr
#' `[,51]` \tab `TIMESINCEIRR` \tab `integer` \tab 05-08 \tab Time since irradiation (s)\cr
#' `[,52]` \tab `TIMETICK` \tab `numeric` \tab 05-08 \tab Time tick for pulsing (s)\cr
#' `[,53]` \tab `ONTIME` \tab `integer` \tab 05-08 \tab On-time for pulsing (in time ticks)\cr
#' `[,54]` \tab `OFFTIME` \tab `integer` \tab 03 \tab Off-time for pulsed stimulation (in s) \cr
#' `[,55]` \tab `STIMPERIOD` \tab `integer` \tab 05-08 \tab Stimulation period (on+off in time ticks)\cr
#' `[,56]` \tab `GATE_ENABLED` \tab `raw` \tab 05-08 \tab PMT signal gating enabled\cr
#' `[,57]` \tab `ENABLE_FLAGS` \tab `raw` \tab 05-08 \tab PMT signal gating  enabled\cr
#' `[,58]` \tab `GATE_START` \tab `integer` \tab 05-08 \tab Start gating (in time ticks)\cr
#' `[,59]` \tab `GATE_STOP` \tab `integer` \tab 05-08 \tab Stop gating (in time ticks), `'Gateend'` for version 04, here only GATE_STOP is used\cr
#' `[,60]` \tab `PTENABLED` \tab `raw` \tab 05-08 \tab Photon time enabled\cr
#' `[,61]` \tab `DTENABLED` \tab `raw` \tab 05-08 \tab PMT dead time correction enabled\cr
#' `[,62]` \tab `DEADTIME` \tab `numeric` \tab 05-08 \tab PMT dead time (s)\cr
#' `[,63]` \tab `MAXLPOWER` \tab `numeric` \tab 05-08 \tab Stimulation power to 100 percent (mW/cm^2)\cr
#' `[,64]` \tab `XRF_ACQTIME` \tab `numeric` \tab 05-08 \tab XRF acquisition time (s)\cr
#' `[,65]` \tab `XRF_HV` \tab `numeric` \tab 05-08 \tab XRF X-ray high voltage (V)\cr
#' `[,66]` \tab `XRF_CURR` \tab `integer` \tab 05-08 \tab XRF X-ray current (µA)\cr
#' `[,67]` \tab `XRF_DEADTIMEF` \tab `numeric` \tab 05-08 \tab XRF dead time fraction\cr
#' `[,68]` \tab `DETECTOR_ID` \tab `raw` \tab 07-08 \tab Detector ID\cr
#' `[,69]` \tab `LOWERFILTER_ID` \tab `integer` \tab 07-08 \tab Lower filter ID in reader\cr
#' `[,70]` \tab `UPPERFILTER_ID` \tab `integer` \tab 07-08 \tab Upper filter ID in reader\cr
#' `[,71]` \tab `ENOISEFACTOR` \tab `numeric` \tab 07-08 \tab Excess noise filter, usage unknown \cr
#' `[,72]` \tab `MARKPOS_X1` \tab `numeric` \tab 08 \tab Coordinates marker position 1 \cr
#' `[,73]` \tab `MARKPOS_Y1` \tab `numeric` \tab 08 \tab Coordinates marker position 1 \cr
#' `[,74]` \tab `MARKPOS_X2` \tab `numeric` \tab 08 \tab Coordinates marker position 2 \cr
#' `[,75]` \tab `MARKPOS_Y2` \tab `numeric` \tab 08 \tab Coordinates marker position 2 \cr
#' `[,76]` \tab `MARKPOS_X3` \tab `numeric` \tab 08 \tab Coordinates marker position 3 \cr
#' `[,77]` \tab `MARKPOS_Y3` \tab `numeric` \tab 08 \tab Coordinates marker position 3 \cr
#' `[,78]` \tab `EXTR_START` \tab `numeric` \tab 08 \tab usage unknown \cr
#' `[,79]` \tab `EXTR_END` \tab `numeric` \tab 08 \tab usage unknown\cr
#' `[,80]` \tab `SEQUENCE` \tab `character` \tab 03-04 \tab Sequence name
#' }
#' V = BIN-file version (RLum means that it does not depend on a specific BIN-file version)
#'
#' Note that the `Risoe.BINfileData` object combines all values from
#' different versions from the BIN/BINX-file, reserved bits are skipped, however,
#' the function [Luminescence::write_R2BIN] reset arbitrary reserved bits. Invalid
#' values for a specific version are set to `NA`. Furthermore, the
#' internal R data types do not necessarily match the required data types for
#' the BIN-file data import! Data types are converted during data import.\cr
#'
#' **LTYPE** values
#'
#' \tabular{rll}{
#'  VALUE \tab TYPE \tab DESCRIPTION \cr
#' `[0]` \tab `TL` \tab: Thermoluminescence \cr
#' `[1]` \tab `OSL` \tab: Optically stimulated luminescence \cr
#' `[2]` \tab `IRSL` \tab: Infrared stimulated luminescence \cr
#' `[3]` \tab `M-IR` \tab: Infrared monochromator scan\cr
#' `[4]` \tab `M-VIS` \tab: Visible monochromator scan\cr
#' `[5]` \tab `TOL` \tab: Thermo-optical luminescence \cr
#' `[6]` \tab `TRPOSL` \tab: Time Resolved Pulsed OSL\cr
#' `[7]` \tab `RIR` \tab: Ramped IRSL\cr
#' `[8]` \tab `RBR` \tab: Ramped (Blue) LEDs\cr
#' `[9]` \tab `USER` \tab: User defined\cr
#' `[10]` \tab `POSL` \tab: Pulsed OSL \cr
#' `[11]` \tab `SGOSL` \tab: Single Grain OSL\cr
#' `[12]` \tab `RL` \tab: Radio Luminescence \cr
#' `[13]` \tab `XRF` \tab: X-ray Fluorescence
#' }
#'
#' **DTYPE** values
#'
#' \tabular{rl}{
#' VALUE \tab DESCRIPTION \cr
#' `[0]` \tab Natural \cr
#' `[1]` \tab N+dose \cr
#' `[2]` \tab Bleach \cr
#' `[3]` \tab Bleach+dose \cr
#' `[4]` \tab Natural (Bleach) \cr
#' `[5]` \tab N+dose (Bleach) \cr
#' `[6]` \tab Dose \cr
#' `[7]` \tab Background
#' }
#'
#' **LIGHTSOURCE** values
#'
#' \tabular{rl}{
#'  VALUE \tab DESCRIPTION \cr
#' `[0]` \tab None \cr
#' `[1]` \tab Lamp \cr
#' `[2]` \tab IR diodes/IR Laser \cr
#' `[3]` \tab Calibration LED \cr
#' `[4]` \tab Blue Diodes \cr
#' `[5]` \tab White light \cr
#' `[6]` \tab Green laser (single grain) \cr
#' `[7]` \tab IR laser (single grain) }
#'
#' **Internal DATA - object structure**
#'
#' With version 8 of the BIN/BINX file format, slot `@DATA` (byte array `DPOINTS`) can
#' contain two different values:
#'
#' 1. `DPOINTS` (standard for `RECTYPE` := (0,1)): is a vector with the length defined
#' through `NPOINTS`. This is the standard for xy-curves since version 03. However,
#' recorded are only y-values and x-values are calculated during import using
#' information from, amongst others, `LOW` and `HIGH`. For instance, a simple TL curve
#' would store `LOW = 0` and `HIGH = 400`. This is then becomes the range for the
#' temperature values. All values in between are interpolated with a length
#' matching the number of corresponding values stored in `DPOINTS`.
#'
#' *Note: The Sequence Editor uses 0 deg. C as the lowest temperature, although
#' this temperature is usually not available in a laboratory. This, however, does not
#' mean that the calculated TL curves are invalid, as the heater would only
#' start the temperature ramp if the ambient temperature is lower than
#' the target temperature.*
#'
#' 2. `DPOINTS` (`RECTYPE` := 128) is contains no count values but information about
#' the definition of the regions of interest (ROI). Each definition is 504 bytes long.
#' The number of definitions is defined by `NPOINTS` in `@METADATA`. The record
#' describes basically the geometric features of the regions of interest.
#' The representation in R is a nested [list].
#'
#' \tabular{rllll}{
#' **#** \tab **Name** \tab **Data Type** \tab **V** \tab **Description** \cr
#'`[,1]` \tab `NOFPOINTS`  \tab `numeric` \tab 08 \tab number of points in the definition (e.g., if the ROI is a rectangle: 4)\cr
#'`[,2]` \tab `USEDFOR`  \tab `logical` \tab 08 \tab samples for which the ROI is used for; a maximum of 48 samples are allowed.\cr
#'`[,3]` \tab `SHOWNFOR`  \tab `logical` \tab 08 \tab samples for which the ROI is shown for; a maximum of 48 samples are allowed.\cr
#'`[,4]` \tab `COLOR`  \tab `numeric` \tab 08 \tab The colour values of the ROI.\cr
#'`[,5]` \tab `X`  \tab `numeric` \tab 08 \tab The x coordinates used to draw the ROI geometry (up to 50 points are allowed).\cr
#'`[,6]` \tab `Y`  \tab `numeric` \tab 08 \tab The y coordinates used to draw the ROI geometry (up to 50 points are allowed).\cr
#' }
#'
#' (information on the BIN/BINX file format are kindly provided by Risø, DTU Nutech)
#'
#' @section Objects from the Class: Objects can be created by calls of the form
#' `new("Risoe.BINfileData", ...)`.
#'
#' @section Function version: 0.4.1
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' based on information provided by Torben Lapp and Karsten Bracht Nielsen (Risø DTU, Denmark)
#'
#' @seealso [Luminescence::plot_Risoe.BINfileData], [Luminescence::read_BIN2R],
#' [Luminescence::write_R2BIN], [Luminescence::merge_Risoe.BINfileData],
#' [Luminescence::Risoe.BINfileData2RLum.Analysis]
#'
#' @references
#' Risø DTU, 2013. The Sequence Editor User Manual - Feb 2013 and Risø DTU, 2016.
#'
#' The Sequence Editor User Manual - February 2016
#'
#' [https://www.fysik.dtu.dk]()
#'
#' @keywords classes
#'
#' @examples
#'
#' showClass("Risoe.BINfileData")
#'
#' @export
setClass("Risoe.BINfileData",
         slots = list(
           METADATA = "data.frame",
           DATA = "list",
           .RESERVED = "list"
           ),
         prototype = prototype(
           METADATA = data.frame(
             ID = integer(),
             SEL = logical(),
             VERSION = integer(),
             LENGTH = integer(),
             PREVIOUS = integer(),
             NPOINTS = integer(),
             RECTYPE = integer(),
             RUN = integer(),
             SET = integer(),
             POSITION = integer(),
             GRAIN = integer(),
             GRAINNUMBER = integer(),
             CURVENO = integer(),
             XCOORD = integer(),
             YCOORD = integer(),
             SAMPLE = character(),
             COMMENT = character(),
             SYSTEMID = integer(),
             FNAME = character(),
             USER = character(),
             TIME = character(),
             DATE = character(),
             DTYPE = character(),
             BL_TIME = numeric(),
             BL_UNIT = integer(),
             NORM1 = numeric(),
             NORM2 = numeric(),
             NORM3 = numeric(),
             BG = numeric(),
             SHIFT = integer(),
             TAG = integer(),
             LTYPE = character(),
             LIGHTSOURCE = character(),
             LPOWER = numeric(),
             LIGHTPOWER = numeric(),
             LOW = numeric(),
             HIGH = numeric(),
             RATE = numeric(),
             TEMPERATURE = numeric(),
             MEASTEMP = numeric(),
             AN_TEMP = numeric(),
             AN_TIME = numeric(),
             TOLDELAY = integer(),
             TOLON = integer(),
             TOLOFF = integer(),
             IRR_TIME = numeric(),
             IRR_TYPE = integer(),
             IRR_UNIT = integer(),
             IRR_DOSERATE = numeric(),
             IRR_DOSERATEERR = numeric(),
             TIMESINCEIRR = numeric(),
             TIMETICK = numeric(),
             ONTIME = numeric(),
             OFFTIME = numeric(),
             STIMPERIOD = integer(),
             GATE_ENABLED = numeric(),
             ENABLE_FLAGS = numeric(),
             GATE_START = numeric(),
             GATE_STOP = numeric(),
             PTENABLED = numeric(),
             DTENABLED = numeric(),
             DEADTIME = numeric(),
             MAXLPOWER = numeric(),
             XRF_ACQTIME = numeric(),
             XRF_HV = numeric(),
             XRF_CURR = numeric(),
             XRF_DEADTIMEF = numeric(),
             DETECTOR_ID = integer(),
             LOWERFILTER_ID = integer(),
             UPPERFILTER_ID = integer(),
             ENOISEFACTOR = numeric(),
             MARKPOS_X1 = numeric(),
             MARKPOS_Y1 = numeric(),
             MARKPOS_X2 = numeric(),
             MARKPOS_Y2 = numeric(),
             MARKPOS_X3 = numeric(),
             MARKPOS_Y3 = numeric(),
             EXTR_START = numeric(),
             EXTR_END = numeric(),
             SEQUENCE = character(),
             stringsAsFactors=FALSE
           ),
           DATA = list(),
           .RESERVED = list()
          )
         )


## show method --------------------------------------------------------------
#' @describeIn show
#' Show the structure of `Risoe.BINfileData` objects.
#'
#' @export
setMethod(f = "show",
          signature = signature(object = "Risoe.BINfileData"),
          definition = function(object){

            if(nrow(object@METADATA) != 0){
              ## check if image/ROI data are present; get ID and remove information
              if(!is.null(object@METADATA[["RECTYPE"]]))
                id_128 <- object@METADATA[["RECTYPE"]] != 128
              else
                id_128 <- rep(TRUE, nrow(object@METADATA))

              version <- suppressWarnings(
                  .collapse(unique(object@METADATA[id_128, "VERSION"]),
                            quote = FALSE))
              systemID <- suppressWarnings(
                  .collapse(unique(object@METADATA[id_128, "SYSTEMID"]),
                            quote = FALSE))
              filename <- as.character(object@METADATA[1,"FNAME"])
              records.overall <- length(object@DATA)
              records.type <- table(object@METADATA[id_128,"LTYPE"])
              user <- .collapse(unique(object@METADATA[id_128, "USER"]),
                                quote = FALSE)
              date <- .collapse(unique(object@METADATA[id_128, "DATE"]),
                                quote = FALSE)
              grain.range <- suppressWarnings(range(object@METADATA[id_128,"GRAIN"]))
              records.type.count <- vapply(seq_along(records.type), function(x){
                paste0(names(records.type[x]),"\t(n = ", records.type[x],")")
               }, character(1))

              records.type.count <- paste(records.type.count,
                                          collapse="\n\t                      ")

              ##print
              cat("\n[Risoe.BINfileData object]")
              cat("\n\n\tBIN/BINX version:    ", version)
              if(version >= 6){
                cat("\n\tFile name:           ", filename)
              }
              cat("\n\tObject date:         ", date)
              cat("\n\tUser:                ", user)
              cat("\n\tSystem ID:           ", ifelse(systemID == 0,"0 (unknown)", systemID))
              cat("\n\tOverall records:     ", records.overall)
              cat("\n\tRecords type:        ", records.type.count)
              cat("\n\tPosition range:      ",
                  .format_range(object@METADATA[id_128, "POSITION"], sep = " : "))
              if (max(grain.range) > 0)
                cat("\n\tGrain range:         ",
                    .format_range(object@METADATA[id_128, "GRAIN"], sep = " : "))
              cat("\n\tRun range:           ",
                  .format_range(object@METADATA[id_128, "RUN"], sep = " : "))
              cat("\n\tSet range:           ",
                  .format_range(object@METADATA[id_128, "SET"], sep = " : "))

              ## if id_128
              if(any(!id_128))
                cat("\n\t + additional ROI data found in record(s):",
                    .collapse(which(!id_128), quote = FALSE))

            }else{
              cat("\n[Risoe.BINfileData object]")
              cat("\n\n >> This object is empty!<<")
            }
            cat("\n")
          }#end function
          )#end setMethod


## set method for object class ----------------------------------------------
#' @describeIn Risoe.BINfileData
#' A [Luminescence::Risoe.BINfileData-class] object is normally produced as output of the
#' function [Luminescence::read_BIN2R]. This construction method is intended for internal
#' usage only.
#'
#' @param METADATA Object of class "data.frame" containing the meta information
#' for each curve.
#'
#' @param DATA Object of class "list" containing numeric vector with count data.
#'
#' @param .RESERVED Object of class "list" containing list of undocumented raw
#' values for internal use only.
#'
#' @export
setMethod(f = "set_Risoe.BINfileData",
          signature = signature("ANY"),
          definition = function(METADATA, DATA, .RESERVED) {
            .set_function_name("set_Risoe.BINfileData")
            on.exit(.unset_function_name(), add = TRUE)

            if(length(METADATA) == 0){
              new("Risoe.BINfileData")

            }else{
              .validate_class(METADATA, "data.frame")
              .validate_class(DATA, "list")
              new(
                "Risoe.BINfileData",
                METADATA = METADATA,
                DATA = DATA,
                .RESERVED = .RESERVED
              )
            }
          })


## add_metadata() -----------------------------------------------------------
#' @describeIn metadata
#' Adds metadata to [Luminescence::Risoe.BINfileData-class] objects.
#'
#' @export
setMethod("add_metadata<-",
          signature= "Risoe.BINfileData",
          definition = function(object, info_element, value) {
            .set_function_name("add_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            .validate_class(info_element, "character", length = 1)
            valid.names <- colnames(object@METADATA)
            if (info_element %in% valid.names) {
              .throw_error("'info_element' already present, to modify it ",
                           "you should use `replace_metadata()`")
            }
            if (is.null(value)) {
              .throw_error("Cannot store a metadata entry with NULL value")
            }

            ## add the metadata element
            object@METADATA[[info_element]] <- value
            assign(x = deparse(substitute(object))[1], object)
          })

## rename_metadata() --------------------------------------------------------
#' @describeIn metadata
#' Renames a metadata entry of [Luminescence::Risoe.BINfileData-class] objects.
#'
#' @export
setMethod("rename_metadata<-",
          signature= "Risoe.BINfileData",
          definition = function(object, info_element, value) {
            .set_function_name("rename_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            .validate_class(info_element, "character", length = 1)
            valid.names <- colnames(object@METADATA)
            if (!info_element %in% valid.names) {
              .throw_error("'info_element' not recognised (",
                           .collapse(info_element), "), valid terms are: ",
                           .collapse(valid.names, quote = FALSE))
            }
            if (is.null(value)) {
              .throw_error("Cannot rename a metadata entry to NULL, ",
                           "to remove it you should use `replace_metadata()`")
            }

            ## rename the metadata element
            name.idx <- grep(info_element, valid.names)
            colnames(object@METADATA)[name.idx] <- value
            assign(x = deparse(substitute(object))[1], object)
          })

## replace_metadata() -------------------------------------------------------
#' @describeIn metadata
#' Replaces or removes metadata of [Luminescence::Risoe.BINfileData-class] objects.
#'
#' @export
setMethod("replace_metadata<-",
          signature= "Risoe.BINfileData",
          definition = function(object, info_element, subset = NULL, value = NULL) {
            .set_function_name("replace_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            .validate_class(info_element, "character")
            valid.names <- colnames(object@METADATA)
            not.found <- setdiff(info_element, valid.names)
            if (length(not.found) > 0) {
              .throw_error("'info_element' not recognised (",
                           .collapse(not.found), "), valid terms are: ",
                           .collapse(valid.names, quote = FALSE))
            }

            ## select relevant rows
            sel <- TRUE
            if (!is.null(substitute(subset))) {

              if (is.null(value)) {
                ## assigning `NULL` indicates that we want to remove a field,
                ## but that is incompatible with choosing a subset of rows
                .throw_error("'subset' is incompatible with assigning NULL")
              }

              ## evaluate the subset expression to produce a selection
              sel <- tryCatch(eval(
                  expr = substitute(subset),
                  envir = object@METADATA,
                  enclos = parent.frame()
              ), error = function(e) {
                .throw_error("Invalid 'subset' expression, valid terms are: ",
                             .collapse(valid.names, quote = FALSE))
              })
              if (!is.logical(sel)) {
                .throw_error("'subset' should contain a logical expression")
              }
              if (all(is.na(sel))) {
                sel <- FALSE
              }
              if (!any(sel)) {
                .throw_message("'subset' expression produced an ",
                               "empty selection, nothing done")
                return(object)
              }
            }

            if (!is.null(substitute(subset))) {
              ## replace the metadata elements
              for (field in info_element)
                object@METADATA[[field]][sel] <- value
            } else {
              ## remove the metadata elements
              object@METADATA[info_element] <- value
            }
            assign(x = deparse(substitute(object))[1], object)
          })


## sort_RLum() --------------------------------------------------------------
#' @describeIn sort_RLum
#' Sort method for [Luminescence::Risoe.BINfileData-class] objects.
#'
#' @export
setMethod(
  f = "sort_RLum",
  signature = "Risoe.BINfileData",
  function(object, info_element,
           decreasing = FALSE, ...) {
    .set_function_name("sort_RLum")
    on.exit(.unset_function_name(), add = TRUE)

    ## input validation
    .validate_class(info_element, "character")
    valid.names <- colnames(object@METADATA)
    if (!all(info_element %in% valid.names)) {
      .throw_error("Invalid 'info_element' name, valid names are: ",
                   .collapse(valid.names))
    }
    if (length(info_element) > 1) {
      message("[sort_RLum()]: Only the first field will be used in sorting")
      info_element <- info_element[1]
    }
    .validate_logical_scalar(decreasing)

    ## determine the new ordering
    ord <- order(object@METADATA[[info_element]], decreasing = decreasing)

    ## reorder the object
    object@METADATA <- object@METADATA[ord, ]
    object@DATA <- object@DATA[ord]

    return(object)
  }
)


## view() -------------------------------------------------------------------
#' @describeIn view
#' View method for [Luminescence::Risoe.BINfileData-class] objects.
#'
#'@export
setMethod("view",
          signature= "Risoe.BINfileData",
          definition = function(object, ...) {

    ## set title
    name <- list(...)$title
    if (is.null(name))
      name <- paste("METADATA for", deparse(substitute(object)))

    ## run view
    .view(x = object@METADATA, title = name)
})

#' @title Bayesian models (baSAR) applied on luminescence data
#'
#' @description
#' This function allows the application of Bayesian models on luminescence data
#' measured with the single-aliquot regenerative-dose (SAR, Murray and Wintle,
#' 2000) protocol. In particular, it follows the idea proposed by Combès et al.,
#' 2015 of using an hierarchical model for estimating a central equivalent dose
#' from a set of luminescence measurements. This function (I) implements this
#' approach for the R environment and (II) provides an extension and a technical
#' refinement of the published code.
#'
#' @details
#' Internally the function consists of two parts: (I) a Bayesian core that
#' performs the hierarchical modelling and calculations, and (II) a data
#' pre-processing stage. The Bayesian core can be run independently if sufficient
#' input data are provided (see below). The data pre-processing stage exists
#' to simplify the user workflow by preparing all required inputs; in principle
#' it is enough to provide a BIN/BINX-file with the SAR measurement data. For
#' the Bayesian analysis for each aliquot the following information are needed
#' from the SAR analysis: `LxTx`, the `LxTx` error and the dose values for all
#' regeneration points.
#'
#' **How is the systematic error contribution calculated?**
#'
#' Standard errors (so far) provided with the source dose rate are considered
#' as systematic uncertainties and added to final central dose by:
#'
#' \deqn{systematic.error = 1/n \sum SE(source.doserate)}
#'
#' \deqn{SE(central.dose.final) = \sqrt{SE(central.dose)^2 + systematic.error^2}}
#'
#' Please note that this approach is rather rough and can only be valid if the
#' source dose rate errors, in case different readers had been used, are
#' similar. In cases where more than one source dose rate is provided a
#' warning is given.
#'
#' **Input / output scenarios**
#'
#' Various inputs are allowed for this function. Unfortunately this makes the
#' function handling rather complex, but at the same time very powerful.
#' Available scenarios:
#'
#' **(1) - `object` is BIN-file or link to a BIN-file**
#'
#' It does not matter how the information of the BIN/BINX file are provided.
#' The function supports **(a)** either a path to a file or directory or a
#' `list` of file names or paths or  **(b)** a [Luminescence::Risoe.BINfileData-class] object
#' or a list of these objects. The latter one can be produced by using function
#' [Luminescence::read_BIN2R], but this function is called automatically if only a file name
#' and/or a path is provided. In both cases it will become the data that can be
#' used for the analysis.
#'
#' `[CSV_file = NULL]`
#'
#' If no CSV file (or data frame with the same format) is provided, the
#' function runs an automatic process that consists of the following steps:
#'
#'  1. Select all valid aliquots using the function [Luminescence::verify_SingleGrainData]
#'  2. Calculate `Lx/Tx` values using the function [Luminescence::calc_OSLLxTxRatio]
#'  3. Calculate De values using the function [Luminescence::fit_DoseResponseCurve]
#'
#' These proceeded data are subsequently used in for the Bayesian analysis
#'
#' `[CSV_file != NULL]`
#'
#' If a CSV file is provided (or a `data.frame` containing similar information)
#' the pre-processing phase consists of the following steps:
#'
#'  1. Calculate `Lx/Tx` values using the function [Luminescence::calc_OSLLxTxRatio]
#'  2. Calculate De values using the function [Luminescence::fit_DoseResponseCurve]
#'
#' The CSV file should contain the BIN-file names and the aliquots selected
#' for the further analysis. This allows a manual selection of input data, as
#' the automatic selection by [Luminescence::verify_SingleGrainData] might not be sufficient.
#'
#' **(2) - `object` `RLum.Results object`**
#'
#' If an [Luminescence::RLum.Results-class] object is provided as input and(!) this object was
#' previously created by the function `analyse_baSAR()` itself, the pre-processing part
#' is skipped and the function starts directly with the Bayesian analysis. This option is very powerful
#' as it allows to change parameters for the Bayesian analysis without the need to repeat
#' the data pre-processing. If furthermore the argument `aliquot_range` is set, aliquots
#' can be manually excluded based on previous runs.
#'
#' **`method_control`**
#'
#' These are arguments that can be passed directly to the Bayesian calculation core, supported arguments
#' are:
#'
#' \tabular{lll}{
#' **Parameter** \tab **Type** \tab **Description**\cr
#' `lower_centralD` \tab [numeric] \tab sets the lower bound for the expected De range. Change it only if you know what you are doing!\cr
#' `upper_centralD` \tab [numeric] \tab sets the upper bound for the expected De range. Change it only if you know what you are doing!\cr
#' `n.chains` \tab [integer] \tab sets number of parallel chains for the model (default = 3) (cf. [rjags::jags.model])\cr
#' `inits` \tab [list] \tab option to set initialisation values (cf. [rjags::jags.model]) \cr
#' `thin` \tab [numeric] \tab thinning interval for monitoring the Bayesian process (cf. [rjags::jags.model])\cr
#' `variable.names` \tab [character] \tab set the variables to be monitored during the MCMC run, default:
#' `'central_D'`, `'sigma_D'`, `'D'`, `'Q'`, `'a'`, `'b'`, `'c'`, `'g'`.
#' Note: only variables present in the model can be monitored.
#' }
#'
#' **User defined models**\cr
#'
#' The function provides the option to modify and to define own models that can be used for
#' the Bayesian calculation. In the case the user wants to modify a model, a new model
#' can be piped into the function via the argument `baSAR_model` as `character`.
#' The model has to be provided in the JAGS dialect of the BUGS language (cf. [rjags::jags.model])
#' and parameter names given with the pre-defined names have to be respected, otherwise the function
#' will break.
#'
#' **FAQ**
#'
#' Q: How can I set the seed for the random number generator (RNG)?
#'
#' A: Use the argument `method_control`, e.g., for three MCMC chains
#' (as it is the default):
#'
#' ```
#' method_control = list(
#' inits = list(
#'  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 1),
#'  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 2),
#'  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 3)
#' ))
#' ```
#'
#' This sets a reproducible set for every chain separately.\cr
#'
#' Q: How can I modify the output plots?
#'
#' A: You can't, but you can use the function output to create own, modified plots.
#'
#'
#' Q: Can I change the boundaries for the central_D?
#'
#' A: Yes, we made it possible, but we DO NOT recommend it, except you know what you are doing!\cr
#' Example: `method_control = list(lower_centralD = 10))`
#'
#' Q: The lines in the baSAR-model appear to be in a wrong logical order?\cr
#'
#' A: This is correct and allowed (cf. JAGS manual)
#'
#'
#' **Additional arguments support via the `...` argument**
#'
#' This list summarizes the additional arguments that can be passed to the internally used
#' functions.
#'
#' \tabular{llll}{
#' **Supported argument** \tab **Corresponding function** \tab **Default** \tab **Short description**\cr
#' `threshold` \tab [Luminescence::verify_SingleGrainData] \tab `30` \tab change rejection threshold for curve selection \cr
#' `skip` \tab [data.table::fread] \tab `0` \tab number of rows to be skipped during import\cr
#' `n.records` \tab [Luminescence::read_BIN2R] \tab `NULL` \tab limit records during BIN-file import\cr
#' `duplicated.rm` \tab [Luminescence::read_BIN2R] \tab `TRUE` \tab remove duplicated records in the BIN-file\cr
#' `pattern` \tab [Luminescence::read_BIN2R] \tab `TRUE` \tab select BIN-file by name pattern\cr
#' `position` \tab [Luminescence::read_BIN2R] \tab `NULL` \tab limit import to a specific position\cr
#' `background.count.distribution` \tab [Luminescence::calc_OSLLxTxRatio] \tab `"non-poisson"` \tab set assumed count distribution\cr
#' `fit.weights` \tab [Luminescence::fit_DoseResponseCurve] \tab `TRUE` \tab enable/disable fit weights\cr
#' `fit.bounds` \tab [Luminescence::fit_DoseResponseCurve] \tab `TRUE` \tab enable/disable fit bounds\cr
#' `n.MC` \tab [Luminescence::fit_DoseResponseCurve] \tab `100` \tab number of MC runs for error calculation\cr
#' `plot_drc` \tab [Luminescence::plot_DoseResponseCurve] \tab `TRUE` \tab enable/disable dose response curve plot\cr
#' `plot_extended` \tab [Luminescence::plot_DoseResponseCurve] \tab `TRUE` \tab enable/disable extended dose response curve plot\cr
#' `recordType` \tab [Luminescence::get_RLum] \tab `c(OSL (UVVIS), irradiation (NA)` \tab helps for the curve selection\cr
#' }
#'
#' @param object [Luminescence::Risoe.BINfileData-class], [Luminescence::RLum.Results-class], [list] of [Luminescence::RLum.Analysis-class],
#' [character] or [list] (**required**):
#' input object used for the Bayesian analysis. If a `character` is provided the function
#' assumes a file connection and tries to import a BIN/BINX-file using the provided path. If a `list` is
#' provided the list can only contain either `Risoe.BINfileData` objects or `character`s
#' providing a file connection. Mixing of both types is not allowed. If an [Luminescence::RLum.Results-class]
#' is provided the function directly starts with the Bayesian Analysis (see details)
#'
#' @param CSV_file [character] or [data.frame] (*optional*):
#' if a `character`, it must be the path to a CSV file with data for the
#' analysis. Either way, data should contain 3 columns:
#' the name of the file, the disc position and the grain position
#' (the last being 0 for multi-grain measurements).
#'
#' @param aliquot_range [numeric] (*optional*):
#' allows to limit the range of the aliquots used for the analysis.
#' This argument has only an effect if the argument `CSV_file` is used or
#' `object` is a previous output (i.e. is [Luminescence::RLum.Results-class]). In this case,
#' the new selection will add the aliquots to the removed aliquots table.
#'
#' @param source_doserate [numeric] (**required**):
#' source dose rate of beta-source used for the measurement and its uncertainty
#' in Gy/s, e.g., `source_doserate = c(0.12, 0.04)`. Parameter can be provided
#' as a list, for the case that more than one BIN-file is provided, e.g.,
#' `source_doserate = list(c(0.04, 0.004), c(0.05, 0.004))`.
#'
#' @param signal_integral [vector] (**required**):
#' vector with the limits for the signal integral used for the calculation,
#' e.g., `signal_integral = 1:5`. It is ignored if `object` is an
#' [Luminescence::RLum.Results-class] object.
#' The parameter can be provided as a list, see `source_doserate`.
#'
#' @param background_integral [vector] (**required**):
#' vector with the limits for the background integral. It is ignored if
#' `object` is an [Luminescence::RLum.Results-class] object.
#' The parameter can be provided as a list, see `source_doserate`.
#'
#' @param signal_integral_Tx [vector] (*optional*):
#' vector with the limits for the signal integral for the Tx curve. If nothing
#' is provided, the value from `signal_integral` is used. It is ignored
#' if `object` is an [Luminescence::RLum.Results-class] object.
#' The parameter can be provided as a list, see `source_doserate`.
#'
#' @param background_integral_Tx [vector] (*optional*):
#' vector with the limits for the background integral for the Tx curve.
#' If nothing is provided, the value from `background_integral` is used.
#' It is ignored if `object` is an [Luminescence::RLum.Results-class] object.
#' The parameter can be provided as a list, see `source_doserate`.
#'
#' @param irradiation_times [numeric] (*optional*): if set this vector replaces all irradiation
#' times for one aliquot and one cycle (Lx and Tx curves) and recycles it for all others cycles and aliquots.
#' Please note that if this argument is used, for every(!) single curve
#' in the dataset an irradiation time needs to be set.
#'
#' @param sigmab [numeric] (*with default*):
#' option to set a manual value for the overdispersion (for `LnTx` and `TnTx`),
#' used for the `Lx`/`Tx` error calculation. The value should be provided as
#' absolute squared count values, cf. [Luminescence::calc_OSLLxTxRatio].
#' The parameter can be provided as `list`, see `source_doserate`.
#'
#' @param sig0 [numeric] (*with default*):
#' allow adding an extra component of error to the final Lx/Tx error value
#' (e.g., instrumental error, see details is [Luminescence::calc_OSLLxTxRatio]).
#' The parameter can be provided as `list`, see `source_doserate`.
#'
#' @param distribution [character] (*with default*):
#' type of distribution used during Bayesian calculations to determine the
#' Central dose and overdispersion values (one of `"cauchy"`, `"normal"`,
#' `"log_normal"` and `"user_defined"`). If set to `"user_defined"`, argument
#' `baSAR_model` must be provided.
#'
#' @param baSAR_model [character] (*optional*):
#' option to provide an own modified or new model for the Bayesian calculation
#' (see details). If provided, argument `distribution` is automatically set to
#' `"user_defined"`.
#'
#' @param n.MCMC [integer] (*with default*):
#' number of iterations for the Markov chain Monte Carlo (MCMC) simulations
#'
#' @param fit.method [character] (*with default*):
#' equation used both for the fitting of the dose-response curve using function
#' [Luminescence::fit_DoseResponseCurve] and then for the Bayesian modelling.
#' Supported methods: `EXP`, `EXP+LIN` and `LIN`.
#'
#' @param fit.force_through_origin [logical] (*with default*):
#' force fitting through origin
#'
#' @param fit.includingRepeatedRegPoints [logical] (*with default*):
#' includes the recycling point (assumed to be measured during the last cycle)
#'
#' @param method_control [list] (*optional*):
#' named list of control parameters that can be directly
#' passed to the Bayesian analysis, e.g., `method_control = list(n.chains = 4)`.
#' See details for further information
#'
#' @param digits [integer] (*with default*):
#' round output to the number of given digits
#'
#' @param distribution_plot [character] (*with default*): sets the final distribution plot that
#' shows equivalent doses obtained using the frequentist approach and sets in the central dose
#' as comparison obtained using baSAR. Allowed input is `'abanico'` or `'kde'`.
#' If set to `NULL` nothing is plotted.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param plot_reduced [logical] (*with default*):
#' enable/disable the advanced plot output.
#'
#' @param plot_singlePanels [logical] (*with default*):
#' enable/disable single plots or plots arranged by `analyse_baSAR`.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... parameters that can be passed to the function [Luminescence::calc_OSLLxTxRatio]
#' (almost full support), [data.table::fread] (`skip`), [Luminescence::read_BIN2R] (`n.records`,
#' `position`, `duplicated.rm`), see details.
#'
#'
#' @return Function returns results numerically and graphically:
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$summary` \tab `data.frame` \tab statistical summary, including the central dose \cr
#'  `$mcmc` \tab `mcmc` \tab [coda::mcmc.list] object including raw output \cr
#'  `$models` \tab `character` \tab implemented models used in the baSAR-model core \cr
#'  `$input_object` \tab `data.frame` \tab summarising table (same format as the XLS-file) including, e.g., Lx/Tx values\cr
#'  `$removed_aliquots` \tab `data.frame` \tab table with removed aliquots (e.g., `NaN`, or `Inf` `Lx`/`Tx` values). If nothing was removed `NULL` is returned
#' }
#'
#'**slot:** **`@info`**
#'
#' The original function call
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#'  - (A) Ln/Tn curves with set integration limits,
#'  - (B) trace plots are returned by the baSAR-model, showing the convergence of the parameters (trace)
#'  and the resulting kernel density plots. If `plot_reduced = FALSE` for every(!) dose a trace and
#'  a density plot is returned (this may take a long time),
#'  - (C) dose plots showing the dose for every aliquot as boxplots and the marked
#'  HPD in within. If boxes are coloured 'orange' or 'red' the aliquot itself should be checked,
#'  - (D) the dose response curve resulting from the monitoring of the Bayesian modelling are
#'  provided along with the Lx/Tx values and the HPD. Note: The amount for curves displayed
#'  is limited to 1000 (random choice) for performance reasons,
#'  - (E) the final plot is the De distribution as calculated using the conventional (frequentist) approach
#'  and the central dose with the HPDs marked within. This figure is only provided for a comparison,
#'  no further statistical conclusion should be drawn from it.
#'
#' **Please note: If distribution was set to `log_normal` the central dose is given as geometric mean!**
#'
#' @section Function version: 0.1.41
#'
#' @author
#' Norbert Mercier, Archéosciences Bordeaux, CNRS-Université Bordeaux Montaigne (France) \cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' The underlying Bayesian model based on a contribution by Combès et al., 2015.
#'
#' @seealso [Luminescence::read_BIN2R], [Luminescence::calc_OSLLxTxRatio],
#' [Luminescence::fit_DoseResponseCurve], [Luminescence::plot_DoseResponseCurve],
#' [data.table::fread], [Luminescence::verify_SingleGrainData],
#' [rjags::jags.model], [rjags::coda.samples], [boxplot.default]
#'
#' @references
#'
#' Combès, B., Philippe, A., Lanos, P., Mercier, N., Tribolo, C., Guérin, G., Guibert, P., Lahaye, C., 2015.
#' A Bayesian central equivalent dose model for optically stimulated luminescence dating.
#' Quaternary Geochronology 28, 62-70. doi:10.1016/j.quageo.2015.04.001
#'
#' Mercier, N., Kreutzer, S., Christophe, C., Guérin, G., Guibert, P., Lahaye, C., Lanos, P., Philippe, A.,
#' Tribolo, C., 2016. Bayesian statistics in luminescence dating: The 'baSAR'-model and its implementation
#' in the R package 'Luminescence'. Ancient TL 34, 14-21.
#'
#' **Further reading**
#'
#' Gelman, A., Carlin, J.B., Stern, H.S., Dunson, D.B., Vehtari, A., Rubin, D.B., 2013.
#' Bayesian Data Analysis, Third Edition. CRC Press.
#'
#' Murray, A.S., Wintle, A.G., 2000. Luminescence dating of quartz using an improved single-aliquot
#' regenerative-dose protocol. Radiation Measurements 32, 57-73. doi:10.1016/S1350-4487(99)00253-X
#'
#' Plummer, M., 2017. JAGS Version 4.3.0 user manual. `https://sourceforge.net/projects/mcmc-jags/files/Manuals/4.x/jags_user_manual.pdf/download`
#'
#' @note
#' **If you provide more than one BIN-file**, it is **strongly** recommended to provide
#' a `list` with the same number of elements for the following parameters:
#'
#' `source_doserate`, `signal_integral`, `signal_integral_Tx`, `background_integral`,
#' `background_integral_Tx`, `sigmab`, `sig0`.
#'
#' Example for two BIN-files: `source_doserate = list(c(0.04, 0.006), c(0.05, 0.006))`
#'
#' **The function is currently limited to work with standard Risoe BIN-files only!**
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##(1) load package test data set
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##(2) selecting relevant curves, and limit dataset
#' CWOSL.SAR.Data <- subset(
#'   CWOSL.SAR.Data,
#'   subset = POSITION%in%c(1:3) & LTYPE == "OSL")
#'
#' \dontrun{
#' ##(3) run analysis
#' ## please note that the parameters used here are
#' ## chosen for performance, not for reliability
#' results <- analyse_baSAR(
#'   object = CWOSL.SAR.Data,
#'   source_doserate = c(0.04, 0.001),
#'   signal_integral = 1:2,
#'   background_integral = 80:100,
#'   fit.method = "LIN",
#'   plot = FALSE,
#'   n.MCMC = 200
#' )
#'
#' print(results)
#'
#' ##CSV_file template
#' ##copy and paste this the code below in the terminal
#' ##you can further use the function write.csv() to export the example
#'
#' CSV_file <-
#' structure(
#' list(
#'  BIN_FILE = NA_character_,
#'  DISC = NA_real_,
#'  GRAIN = NA_real_),
#'    .Names = c("BIN_FILE", "DISC", "GRAIN"),
#'    class = "data.frame",
#'    row.names = 1L)
#' }
#'
#' @export
analyse_baSAR <- function(
  object,
  CSV_file = NULL,
  aliquot_range = NULL,
  source_doserate = NULL,
  signal_integral = NULL,
  background_integral = NULL,
  signal_integral_Tx = NULL,
  background_integral_Tx = NULL,
  irradiation_times = NULL,
  sigmab = 0,
  sig0 = 0.025,
  distribution = "cauchy",
  baSAR_model = NULL,
  n.MCMC = 100000,
  fit.method = c("EXP", "EXP+LIN", "LIN"),
  fit.force_through_origin = TRUE,
  fit.includingRepeatedRegPoints = TRUE,
  method_control = list(),
  digits = 3L,
  distribution_plot = c("kde", "abanico"),
  plot = TRUE,
  plot_reduced = TRUE,
  plot_singlePanels = FALSE,
  verbose = TRUE,
  ...
) {
  .set_function_name("analyse_baSAR")
  on.exit(.unset_function_name(), add = TRUE)

  ##////////////////////////////////////////////////////////////////////////////////////////////////
  ##FUNCTION TO BE CALLED to RUN the Bayesian Model
  ##////////////////////////////////////////////////////////////////////////////////////////////////
  ##START
  .baSAR_function <-
    function(Nb_aliquots,
             distribution,
             data.Dose,
             data.Lum,
             data.sLum,
             fit.method,
             n.MCMC,
             fit.force_through_origin,
             fit.includingRepeatedRegPoints,
             method_control,
             lower_centralD,
             upper_centralD,
             baSAR_model,
             verbose)
    {
      n.chains <- method_control[["n.chains"]] %||% 3
      inits <- method_control[["inits"]]
      thin <-  if (is.null(method_control[["thin"]])) {
        if(n.MCMC >= 1e+05){
          n.MCMC / 1e+05 * 250
        }else{
          min(10, n.MCMC / 2)
        }
      } else{
        .validate_positive_scalar(method_control[["thin"]], int = TRUE,
                                  name = "'thin' in 'method_control'")
      }

      ## jags reports ugly errors if thin exceeds n.MCMC / 2, as that
      ## would correspond to producing just one posterior sample, see #407
      if (!is.null(method_control[["thin"]]) && thin > n.MCMC / 2) {
        thin <- n.MCMC / 2
        .throw_warning("'thin = ", method_control[["thin"]],
                       "' is too high for 'n.MCMC = ", n.MCMC,
                       "', reset to ", thin)
      }

      ExpoGC <- as.numeric(grepl("EXP", fit.method))
      LinGC <- as.numeric(grepl("LIN", fit.method))
      GC_Origin <- as.numeric(fit.force_through_origin)

      ## exclude repeated dose points
      if (!fit.includingRepeatedRegPoints) {
        for (i in 1:Nb_aliquots) {
          temp.logic <- !duplicated(data.Dose[, i], incomparables = 0) # logical excluding 0
          m <- length(which(!temp.logic))

          data.Dose[,i] <-  c(data.Dose[,i][temp.logic], rep(NA, m))
          data.Lum[,i] <-  c(data.Lum[,i][temp.logic], rep(NA, m))
          data.sLum[,i]  <-  c(data.sLum[,i][temp.logic], rep(NA, m))
        }
      }

      Limited_cycles <- vector()
      for (i in 1:Nb_aliquots) {
        Limited_cycles[i] <- length(stats::na.exclude(data.Dose[, i]))
      }

      ##check and correct for distribution name
      if (!is.null(baSAR_model) && distribution != "user_defined") {
        distribution <- "user_defined"
        .throw_message("'baSAR_model' provided, setting distribution to ",
                       "'user_defined'", error = FALSE)
      }

      # Bayesian Models ----------------------------------------------------------------------------
      # INFO: >
      #  > sometimes lines appear to be in a wrong logical order, however, this is allowed in the
      #  > model definition since:
      #  > "The data block is not limited to logical relations, but may also include stochastic relations."
      #  > (Plummer, 2017. JAGS Version 4.3.0 user manual, p. 9)
      baSAR_models <- list(
        cauchy = "model {
            central_D ~  dunif(lower_centralD,upper_centralD)

            precision_D ~ dt(0, pow(0.16*central_D, -2), 1)T(0, )
            sigma_D <-  1/sqrt(precision_D)

            for (i in 1:Nb_aliquots) {
              a[i] ~  dnorm(6.5 , 1/(9.2^2) ) T(0, )
              b[i] ~  dnorm(50 , 1/(1000^2) )  T(0, )
              c[i] ~  dnorm(1.002 , 1/(0.9^2) ) T(0, )
              g[i] ~  dnorm(0.5 , 1/(2.5^2) ) I(-a[i], )
              sigma_f[i]  ~  dexp (20)

              D[i] ~ dt ( central_D , precision_D, 1)    #  Cauchy distribution

              S_y[1,i] <-  1/(sLum[1,i]^2 + sigma_f[i]^2)
              Lum[1,i] ~ dnorm ( Q[1,i] , S_y[1,i])
              Q[1,i]  <-  GC_Origin * g[i] + LinGC * (c[i] * D[i] ) + ExpoGC * (a[i] * (1 - exp (-D[i] /b[i])) )

              for (m in 2:Limited_cycles[i]) {
                S_y[m,i] <-  1/(sLum[m,i]^2 + sigma_f[i]^2)
                Lum[m,i] ~ dnorm( Q[m,i] , S_y[m,i] )
                Q[m,i]  <-  GC_Origin * g[i] + LinGC * (c[i] * Dose[m,i]) + ExpoGC * (a[i] * (1 - exp (-Dose[m,i]/b[i])) )
              }
            }
          }",

       normal = "model {
            central_D ~  dunif(lower_centralD,upper_centralD)

            sigma_D ~ dunif(0.01, 1 * central_D)

            for (i in 1:Nb_aliquots) {
              a[i] ~  dnorm(6.5 , 1/(9.2^2) ) T(0, )
              b[i] ~  dnorm(50 , 1/(1000^2) )  T(0, )
              c[i] ~  dnorm(1.002 , 1/(0.9^2) ) T(0, )
              g[i] ~  dnorm(0.5 , 1/(2.5^2) ) I(-a[i], )
              sigma_f[i]  ~  dexp (20)

              D[i] ~ dnorm ( central_D , 1/(sigma_D^2) )   #   Normal distribution

              S_y[1,i] <-  1/(sLum[1,i]^2 + sigma_f[i]^2)
              Lum[1,i] ~ dnorm ( Q[1,i] , S_y[1,i])
              Q[1,i]  <-  GC_Origin * g[i] + LinGC * (c[i] * D[i] ) + ExpoGC * (a[i] * (1 - exp (-D[i] /b[i])) )

              for (m in 2:Limited_cycles[i]) {
                S_y[m,i] <-  1/(sLum[m,i]^2 + sigma_f[i]^2)
                Lum[m,i] ~ dnorm( Q[m,i] , S_y[m,i] )
                Q[m,i]  <-  GC_Origin * g[i] + LinGC * (c[i] * Dose[m,i]) + ExpoGC * (a[i] * (1 - exp (-Dose[m,i]/b[i])) )
              }
            }
            }",

       log_normal = "model {
            central_D ~  dunif(lower_centralD,upper_centralD)

            log_central_D <-  log(central_D) - 0.5 * l_sigma_D^2
            l_sigma_D ~ dunif(0.01, 1 * log(central_D))
            sigma_D <-  sqrt((exp(l_sigma_D^2) -1) * exp( 2*log_central_D + l_sigma_D^2) )

            for (i in 1:Nb_aliquots) {
              a[i] ~  dnorm(6.5 , 1/(9.2^2) ) T(0, )
              b[i] ~  dnorm(50 , 1/(1000^2) )  T(0, )
              c[i] ~  dnorm(1.002 , 1/(0.9^2) ) T(0, )
              g[i] ~  dnorm(0.5 , 1/(2.5^2) ) I(-a[i], )
              sigma_f[i]  ~  dexp (20)

              log_D[i] ~ dnorm ( log_central_D , 1/(l_sigma_D^2) )  #   Log-Normal distribution
              D[i] <-  exp(log_D[i])

              S_y[1,i] <-  1/(sLum[1,i]^2 + sigma_f[i]^2)
              Lum[1,i] ~ dnorm ( Q[1,i] , S_y[1,i])
              Q[1,i]  <-  GC_Origin * g[i] + LinGC * (c[i] * D[i] ) + ExpoGC * (a[i] * (1 - exp (-D[i] /b[i])) )

            for (m in 2:Limited_cycles[i]) {
                S_y[m,i] <-  1/(sLum[m,i]^2 + sigma_f[i]^2)
                Lum[m,i] ~ dnorm( Q[m,i] , S_y[m,i] )
                Q[m,i]  <-  GC_Origin * g[i] + LinGC * (c[i] * Dose[m,i]) + ExpoGC * (a[i] * (1 - exp (-Dose[m,i]/b[i])) )
              }
            }
        }",

        user_defined = baSAR_model
       )

      ##check whether the input for distribution was sufficient
      if (distribution == "user_defined" && is.null(baSAR_model)) {
        .throw_error("You specified a 'user_defined' distribution, ",
                     "but did not provide a model via 'baSAR_model'")
      }

      ### Bayesian inputs
      data_list  <- list(
        Dose = data.Dose,
        Lum = data.Lum,
        sLum = data.sLum,
        LinGC = LinGC,
        ExpoGC = ExpoGC,
        GC_Origin = GC_Origin,
        Limited_cycles = Limited_cycles,
        lower_centralD = lower_centralD,
        upper_centralD = upper_centralD,
        Nb_aliquots = Nb_aliquots)

      if(verbose){
        cat("\n[analyse_baSAR()] ---- baSAR-model ---- \n")
        cat("\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n")
        cat("[analyse_baSAR()] Bayesian analysis in progress ...\n")
        message(".. >> bounds set to: lower_centralD = ", lower_centralD,
                " | upper_centralD = ", upper_centralD)
        message(".. >> calculation will be done assuming a '",
                distribution, "' distribution\n")
      }

      ##set model
      jagsfit <- rjags::jags.model(
          file = textConnection(baSAR_models[[distribution]]),
          data = data_list,
          inits = inits,
          n.chains = n.chains,
          n.adapt = n.MCMC,
          quiet = !verbose
       )

      ##update jags model (it is a S3-method)
      stats::update(
        object = jagsfit,
        n.iter = n.MCMC,
        progress.bar = if(verbose){"text"}else{NULL}
        )

      ##get data ... full and reduced, the reduced one to limit the plot output
      sampling <- rjags::coda.samples(
        model = jagsfit,
        variable.names = variable.names,
        n.iter = n.MCMC,
        thin = thin
      )

      ##this we need for the output of the terminal
      ##Why sampling reduced? Because the summary() method produces a considerable overhead while
      ##running over all the variables
      sampling_reduced <- rjags::coda.samples(
        model = jagsfit,
        variable.names = c('central_D', 'sigma_D'),
        n.iter = n.MCMC,
        thin = thin
      )

      ##standard error and mean
      output.mean <-
        round(summary(sampling_reduced)[[1]][c("central_D", "sigma_D"), 1:2], digits)

      ## calculate the geometric mean if the distribution is log-normal
      ## (in other cases gm is NULL)
      gm <- if (distribution == "log_normal")
              round(exp(mean(log(as.matrix(sampling_reduced)[, 1]))), digits)

      ##quantiles
      ##68% + 95%
      output.quantiles <-
        round(summary(sampling_reduced, quantiles = c(0.025, 0.16, 0.84, 0.975))[[2]][c("central_D", "sigma_D"), 1:4], digits)

      #### output data.frame with results
      baSAR.output <- data.frame(
        DISTRIBUTION = distribution,
        NB_ALIQUOTS = Nb_aliquots,
        N.CHAINS = n.chains,
        N.MCMC = n.MCMC,
        FIT_METHOD = fit.method,
        CENTRAL = gm %||% output.mean[1, 1],
        CENTRAL.SD = output.mean[1,2],
        SIGMA = output.mean[2,1],
        SIGMA.SD = output.mean[2,2],
        CENTRAL_Q_.16 = output.quantiles[1,2],
        CENTRAL_Q_.84 = output.quantiles[1,3],
        SIGMA_Q_.16 = output.quantiles[2,2],
        SIGMA_Q_.84 = output.quantiles[2,3],
        CENTRAL_Q_.025 = output.quantiles[1,1],
        CENTRAL_Q_.975 = output.quantiles[1,4],
        SIGMA_Q_.025 = output.quantiles[2,1],
        SIGMA_Q_.975 = output.quantiles[2,4]
      )

      return(
        baSAR.output = list(
          baSAR.output_summary = baSAR.output,
          baSAR.output_mcmc = sampling,
          models = list(
            cauchy = baSAR_models[["cauchy"]],
            normal = baSAR_models[["normal"]],
            log_normal = baSAR_models[["log_normal"]],
            user_defined = baSAR_models[["user_defined"]]
          )
        )
      )
    }
  ##END
  ##////////////////////////////////////////////////////////////////////////////////////////////////

  ## deprecated arguments
  extraArgs <- list(...)
  if (any(grepl("[signal|background]\\.integral\\.?[.Tx]?", names(extraArgs))) &&
      !is.null(c(extraArgs$signal.integral, extraArgs$signal.integral.Tx,
                 extraArgs$background.integral, extraArgs$background.integral.Tx))) {
    .deprecated(old = c("signal.integral", "background.integral",
                        "signal.integral.Tx", "background.integral.Tx"),
                new = c("signal_integral", "background_integral",
                        "signal_integral_Tx", "background_integral_Tx"),
                since = "1.2.0")
    signal_integral <- extraArgs$signal.integral
    background_integral <- extraArgs$background.integral
    signal_integral_Tx <- extraArgs$signal.integral.Tx
    background_integral_Tx <- extraArgs$background.integral.Tx
  }

  ## Integrity checks -------------------------------------------------------
  .require_suggested_package("rjags")
  .require_suggested_package("coda")
  .validate_class(object, c("Risoe.BINfileData", "RLum.Results", "character", "list"))
  .validate_not_empty(object)
  .validate_class(CSV_file, c("data.frame", "character"), null.ok = TRUE)
  if (!inherits(object, "RLum.Results")) {
    signal_integral <- .validate_integral(signal_integral, list.ok = TRUE)
    background_integral <- .validate_integral(background_integral, list.ok = TRUE)
    signal_integral_Tx <- .validate_integral(signal_integral_Tx,
                                             null.ok = TRUE, list.ok = TRUE)
    background_integral_Tx <- .validate_integral(background_integral_Tx,
                                                 null.ok = TRUE, list.ok = TRUE)
  }
  distribution <- .validate_args(distribution, c("cauchy", "normal",
                                                 "log_normal", "user_defined"))
  .validate_class(baSAR_model, "character", length = 1, null.ok = TRUE)
  .validate_positive_scalar(n.MCMC, int = TRUE)
  fit.method <- .validate_args(fit.method, c("EXP", "EXP+LIN", "LIN"))
  .validate_nonnegative_scalar(digits, int = TRUE)
  distribution_plot <- .validate_args(distribution_plot, c("kde", "abanico"),
                                      null.ok = TRUE) %||% ""

  #capture additional piped arguments
  additional_arguments <- list(

    ##verify_SingleGrainData
    threshold = 30,

    ##calc_OSLLxTxRatio()
    background.count.distribution = "non-poisson",

    ## data.table::fread()
    skip = 0,

    ##read_BIN2R()
    n.records = NULL,
    duplicated.rm = TRUE,
    position = NULL,
    pattern = NULL,

    ## fit_DoseResponseCurve()
    fit.weights = TRUE,
    fit.bounds = TRUE,
    n.MC = 100,

    ## plot_DoseResponseCurve()
    plot_drc = plot,
    plot_extended = plot,

    ## get_RLum
    recordType = c("OSL (UVVIS)", "irradiation (NA)")
  )

  #modify this list on purpose
  additional_arguments <- modifyList(x = additional_arguments,
                                     val = list(...))

  ## deprecated arguments
  if ("output.plot" %in% ...names()) {
    additional_arguments$plot_drc <- list(...)$output.plot
    .deprecated("output.plot", "plot_drc", since = "1.2.0")
  }
  if ("output.plotExtended" %in% ...names()) {
    additional_arguments$plot_extended <- list(...)$output.plotExtended
    .deprecated("output.plotExtended", "plot_extended", since = "1.2.0")
  }

  ##set function arguments
  function_arguments <- NULL

  ## variable names to monitor
  variable.names <- c('central_D', 'sigma_D', 'D', 'Q', 'a', 'b', 'c', 'g')
  if (!is.null(method_control[["variable.names"]])) {
    ## take variable from the user but remove nonsense values
    variable.names <- setdiff(method_control[["variable.names"]],
                              c("", NA))
  }

  ## always monitor the D variable
  variable.names <- unique(c(variable.names, "D"))

  ## check the central_D bounds and set defaults according to Combès et al., 2015
  ## "We set the bounds for the prior on the central dose D, Dmin = 0 Gy and
  ##  Dmax = 1000 Gy, to cover the likely range of possible values for D."
  msg <- paste("You have modified the %s central_D boundary while applying",
               "a predefined model. This is possible but not recommended")

  ## check lower_centralD and upper_centralD
  lower_centralD <- method_control$lower_centralD
  if (is.null(lower_centralD)) {
    lower_centralD <- 0
  } else if (distribution != "user_defined") {
    .validate_nonnegative_scalar(lower_centralD, null.ok = TRUE,
                                 name = "'lower_centralD' in 'method_control'")
    .throw_warning(sprintf(msg, "lower"))
  }
  upper_centralD <- method_control$upper_centralD
  if (is.null(upper_centralD)) {
    upper_centralD <- 1000
  } else if (distribution != "user_defined") {
    .validate_nonnegative_scalar(upper_centralD, null.ok = TRUE,
                                 name = "'upper_centralD' in 'method_control'")
    .throw_warning(sprintf(msg, "upper"))
  }
  if (upper_centralD <= lower_centralD) {
    .throw_error("'upper_centralD' in 'method_control' must be greater than ",
                 "'lower_centralD'")
  }

  # Set input -----------------------------------------------------------------------------------

  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ##if the input is already of type RLum.Results, use the input and do not run
  ##all pre-calculations again
  if (inherits(object, "RLum.Results")) {
    .validate_originator(object, "analyse_baSAR")

    ## start with arguments used in the RLum.Results object
    function_arguments <- modifyList(formals(),
                                     as.list(object@info$call))

    ## overwrite with currently provided arguments
    arguments.new <- modifyList(function_arguments,
                                as.list(match.call()))

    ## evaluate all arguments, as match.call() may leave some of them as
    ## unevaluated symbols (such as 1:3 instead of c(1, 2, 3))
    arguments.new$`...` <- NULL
    arguments.new <- lapply(arguments.new, eval, envir = parent.frame())

     ## return NULL if not at least three aliquots are used for the calculation
     Nb_aliquots <- nrow(object$input_object)
     if (Nb_aliquots < 3) {
       .throw_message("Number of aliquots < 3, NULL returned")
       return(NULL)
     }

    ## source_doserate
    if (length(as.list(match.call())$source_doserate) > 0) {
      .throw_warning("'source_doserate' is ignored in this mode as ",
                     "it was already set")
    }

    ## set variables
    distribution <- arguments.new$distribution
    n.MCMC <- arguments.new$n.MCMC
    fit.method <- arguments.new$fit.method
    fit.force_through_origin <- arguments.new$fit.force_through_origin
    fit.includingRepeatedRegPoints <- arguments.new$fit.includingRepeatedRegPoints
    aliquot_range <- arguments.new$aliquot_range
    method_control <- arguments.new$method_control
    baSAR_model <- arguments.new$baSAR_model
    plot <- arguments.new$plot
    verbose <- arguments.new$verbose

    input_object <- object$input_object
    removed_aliquots <- object$removed_aliquots

    ## limit according to aliquot_range
    if (!is.null(aliquot_range)) {
      .validate_class(aliquot_range, c("integer", "numeric"), null.ok = TRUE)
      if (min(aliquot_range) < 1 ||  max(aliquot_range) > nrow(input_object)) {
        .throw_message("'aliquot_range' out of bounds, input ignored")
        aliquot_range <- NULL
      } else {
        input_object <- input_object[aliquot_range, ]

         ##update list of removed aliquots
         removed_aliquots <-rbind(object$removed_aliquots, object$input_object[-aliquot_range,])

         ##correct Nb_aliquots
         Nb_aliquots <- nrow(input_object)
       }
     }

     ##set non function arguments
     max_cycles <- max(object$input_object[["CYCLES_NB"]])
     Doses <- t(input_object[,9:(8 + max_cycles)])
     LxTx <- t(input_object[,(9 + max_cycles):(8 + 2 * max_cycles)])
     LxTx.error <-  t(input_object[,(9 + 2 * max_cycles):(8 + 3 * max_cycles)])

  }else{
    if(verbose){
      cat("\n[analyse_baSAR()] ---- PRE-PROCESSING ----\n")
    }

    ##Supported input types are:
    ##  (1) BIN-file
    ##      .. list
    ##      .. character
    ##  (2) RisoeBINfileData object
    ##      .. list
    ##      .. S4
    ##  (3) RLum.Analysis objects
    ##      .. list
    ##      .. S4

    ##In case an RLum.Analysis object is provided we try an ugly conversion only
    if (inherits(object, "list") &&
        all(vapply(object, inherits, "RLum.Analysis", FUN.VALUE = logical(1)))) {
      if (verbose)
        cat("[analyse_baSAR()] List of RLum.Analysis-objects detected ..\n")

      ##stop for only one element
      if (length(object) < 2)
        .throw_error("At least two aliquots are needed for the calculation")

      ##extract wanted curves
      if(verbose)
        cat("\t\t  .. extract '", additional_arguments$recordType , "'\n", sep = "")
      object <- get_RLum(object, recordType = additional_arguments$recordType, drop = FALSE)

      ## check that we are not left with empty records
      if (length(object[[1]]@records) == 0) {
        .throw_error("No records of the appropriate type found")
      }

      ##extract irradiation times
      if(is.null(irradiation_times)){
        if(verbose)
          cat("\t\t  .. extract irradiation times\n")
        irradiation_times <- extract_IrradiationTimes(object[[1]])$irr.times$IRR_TIME
      }

      ##run conversion
      if(verbose)
        cat("\t\t  .. run conversion\n")
      object <- try(convert_RLum2Risoe.BINfileData(object),
                    outFile = stdout()) # redirect error messages so they can be silenced

      ##create fallback
      if (inherits(object, "try-error")) {
         .throw_message("Object conversion failed, NULL returned")
         return(NULL)
      }

      ## remove non-OSL curves
      rm_id <- which(object@METADATA[["LTYPE"]] != "OSL")
      if (length(rm_id) > 0) {
        if(verbose)
          cat("\t\t  .. remove non-OSL curves\n")
        object@METADATA <- object@METADATA[-rm_id,]
        object@DATA[rm_id] <- NULL

        ##reset index
        object@METADATA[["ID"]] <- seq_along(object@METADATA[["ID"]])
      }
    }

    if (inherits(object, "list")) {
      ## check what the list contains ...
      object_type <- unique(sapply(object, function(x) {
        .validate_class(x, c("Risoe.BINfileData", "character"),
                        name = "All elements of 'object'")
        class(x)[1]
      }))

      if (length(object_type) > 1) {
        .throw_error("'object' only accepts a list of objects of the same type")
      }
        if (object_type == "Risoe.BINfileData") {
          fileBIN.list <- object

        } else if (object_type == "character") {
          fileBIN.list <- read_BIN2R(
            file = object,
            position = additional_arguments$position,
            duplicated.rm = additional_arguments$duplicated.rm,
            n.records = additional_arguments$n.records,
            pattern = additional_arguments$pattern,
            verbose = verbose
          )
          fileBIN.list <- .rm_NULL_elements(fileBIN.list)
          if (length(fileBIN.list) == 0)
            return(NULL)
        }

    } else if (is.character(object)) {
      fileBIN.list <- list(read_BIN2R(
          file = object,
          position = additional_arguments$position,
          duplicated.rm = additional_arguments$duplicated.rm,
          n.records = additional_arguments$n.records,
          pattern = additional_arguments$pattern,
          verbose = verbose))

    } else if (inherits(object, "Risoe.BINfileData")) {
      fileBIN.list <- list(object)
    }

    ##Problem ... the user might have made a pre-selection in the Analyst software, if this the
    ##we respect this selection
    record.selected <- unlist(
      lapply(fileBIN.list,
      FUN = function(x) x@METADATA[["SEL"]] ))

    if (!all(record.selected)) {
      if (verbose) {
        .throw_message("Record pre-selection in BIN-file detected, ",
                       "record reduced to selection\n", error = FALSE)
      }
      if (sum(record.selected) == 0) {
        .throw_warning("No records selected, NULL returned")
        return(NULL)
      }

      fileBIN.list <- lapply(fileBIN.list, function(x){
            ##reduce data
            x@DATA <- x@DATA[x@METADATA[["SEL"]]]
            x@METADATA <- x@METADATA[x@METADATA[["SEL"]], ]

            ##reset index
            x@METADATA[["ID"]] <- seq_len(nrow(x@METADATA))
            return(x)
      })
    }

    # Declare variables ---------------------------------------------------------------------------

    Disc <-  list()
    Grain <- list()
    Disc_Grain.list <- list()

    Nb_aliquots <-  0
    previous.Nb_aliquots <- 0
    object.file_name <- list()

    Mono_grain <-  TRUE

    Limited_cycles <- vector()

    ##set information
    for (i in seq_along(fileBIN.list)) {
      Disc[[i]] <-  list()
      Grain[[i]] <-  list()

      ##get BIN-file name
      object.file_name[[i]] <- unique(fileBIN.list[[i]]@METADATA[["FNAME"]])
    }

    ## remove duplicated entries
    object.filenames <- unlist(object.file_name)
    is.duplicated <- duplicated(object.filenames)
    if (any(is.duplicated)) {
      msg <- paste(.collapse(object.filenames[is.duplicated]),
                   "is a duplicate and therefore removed from the input")
      if(verbose){
        .throw_message(msg, error = FALSE)
      }
      .throw_warning(msg)

      ##remove entry
      Disc[is.duplicated] <- NULL
      Grain[is.duplicated] <- NULL
      fileBIN.list[is.duplicated] <- NULL
      object.file_name[is.duplicated] <- NULL
    }

  ## Expand input arguments -------------------------------------------------
  rep.length <- length(fileBIN.list)

  if (is.null(source_doserate)) {
    .throw_error("'source_doserate' is missing, but the current ",
                 "implementation expects dose values in Gy")
  }
  source_doserate <- .listify(source_doserate, rep.length)
  sigmab <- .listify(sigmab, rep.length)
  sig0 <- .listify(sig0, rep.length)
  signal_integral <- .listify(signal_integral, rep.length)
  background_integral <- .listify(background_integral, rep.length)

  if (!is.null(signal_integral_Tx)) {
    signal_integral_Tx <- .listify(signal_integral_Tx, rep.length)
  }
  if (!is.null(background_integral_Tx)) {
    background_integral_Tx <- .listify(background_integral_Tx, rep.length)
  }

  # Read CSV file -----------------------------------------------------------
  if (is.null(CSV_file)) {
    ##select aliquots giving light only, this function accepts also a list as input
    if(verbose){
      cat("[analyse_baSAR()] 'CSV_file' not provided, running automatic grain selection ...\n")
    }

    for (k in seq_along(fileBIN.list)) {
      ##if the uses provides only multiple grain data (GRAIN == 0), the verification
      ##here makes not really sense and should be skipped
      if(length(unique(fileBIN.list[[k]]@METADATA[["GRAIN"]])) > 1){
        aliquot_selection <-
          verify_SingleGrainData(
            object = fileBIN.list[[k]],
            cleanup_level = "aliquot",
            threshold = additional_arguments$threshold,
            cleanup = FALSE
          )

        ## remove grain position 0 (this are usually TL measurements
        ## on the cup or we are talking about multiple aliquots)
        idx.pos0 <- aliquot_selection$unique_pairs[["GRAIN"]] == 0
        datalu <- aliquot_selection$unique_pairs[!idx.pos0, ]
        num.grain.pos0 <- sum(idx.pos0, na.rm = TRUE)
        if (sum(num.grain.pos0) > 0) {
          .throw_warning("Automatic grain selection: ", num.grain.pos0,
                         " curves with grain index 0 have been removed ",
                         "from the dataset")
        }
      }else{
          .throw_warning("Only multiple grain data provided, ",
                         "automatic selection skipped")
          datalu <- unique(fileBIN.list[[k]]@METADATA[, c("POSITION", "GRAIN")])

          ##set mono grain to FALSE
          Mono_grain <- FALSE
          aliquot_selection <- NA
      }

      ##get number of aliquots (one aliquot has a position and a grain number)
      Nb_aliquots <- nrow(datalu)

      ##write information in variables
      Disc[[k]] <- datalu[["POSITION"]]
      Grain[[k]] <- datalu[["GRAIN"]]

      ##free memory
      rm(datalu, aliquot_selection)
    }
  } else {
    ## error message used multiple times
    err.msg <- paste("'CSV_file' should have at least 3 columns for the name",
                     "of the file, the disc position and the grain position")

    ##load file if we have a filename
    if (is.character(CSV_file)) {
      ## import CSV file
      datalu <- tryCatch(data.table::fread(CSV_file, data.table = FALSE,
                                           skip = additional_arguments$skip),
                         error = function(e) .throw_error(e$message))

      ###check whether data format is somehow odd, check only the first three columns
      if (ncol(datalu) < 3) {
        .throw_error(err.msg)
      }

      ##get rid of empty rows if the BIN_FILE name column is empty
      datalu <- datalu[!is.na(datalu[[1]]), ]

    } else{

      datalu <- CSV_file
      if(ncol(datalu) < 3){
        .throw_error(err.msg)
      }

      ##problem: the first column should be of type character, the others are
      ##of type numeric, unfortunately it is too risky to rely on the user, we do the
      ##proper conversion by ourself ...
      datalu[[1]] <- as.character(datalu[[1]])
      datalu[[2]] <- as.numeric(datalu[[2]])
      datalu[[3]] <- as.numeric(datalu[[3]])
    }

    ##limit aliquot range
    if (!is.null(aliquot_range)) {
      datalu <- datalu[aliquot_range,]
    }

    Nb_ali <-  0
    k <- NULL

    for (nn in seq_len(nrow(datalu))) {
      if (!is.na(datalu[nn, 1]))  {

        ##check whether one file fits
        file.basename <- tools::file_path_sans_ext(basename(datalu[nn, 1]))
        matches <- grep(pattern = file.basename, x = unlist(object.file_name))
        if (length(matches) > 0) {
          k <- matches[1]
          nj <- length(Disc[[k]]) + 1

          Disc[[k]][nj] <-  as.numeric(datalu[nn, 2])
          Grain[[k]][nj] <-  as.numeric(datalu[nn, 3])
          Nb_ali <-  Nb_ali + 1
          if (is.na(Grain[[k]][nj]) || Grain[[k]][nj] == 0) {
            Mono_grain <- FALSE
          }

        }else{
          .throw_warning("File '", datalu[nn, 1], "' doesn't match, skipped")
        }

      } else{
        if (Nb_ali == 0) {
          .throw_error("Number of discs/grains = 0")
        }
        break()
      }
    }

    ##if k is NULL it means it was not set so far, so there was
    ##no corresponding BIN-file found
    if(is.null(k)){
      .throw_error("The BIN-file names provided via 'CSV_file' do not match ",
                   "the loaded BIN-files")
    }
  }

  ###################################### loops on files_number
  for (k in seq_along(fileBIN.list)) {
    Disc_Grain.list[[k]] <- list()   # data.file number
    n_aliquots_k <- length(Disc[[k]])

      if(n_aliquots_k == 0){
        fileBIN.list[[k]] <- NULL
        .throw_warning("No data selected from BIN-file ", k,
                       ", BIN-file removed from input")
        next()
      }

    for (d in 1:n_aliquots_k) {
      dd <-  as.integer(unlist(Disc[[k]][d]))
      gg <- if (Mono_grain) as.integer(unlist(Grain[[k]][d])) else 1
      Disc_Grain.list[[k]][[dd]] <- list()  # data.file number ,  disc_number

        Disc_Grain.list[[k]][[dd]][[gg]] <- list()  # data.file number ,  disc_number, grain_number
        for (z in 1:6) {
          Disc_Grain.list[[k]][[dd]][[gg]][[z]] <- list()
          # 1 = index numbers, 2 = irradiation doses,  3 = LxTx , 4 = sLxTx,  5 = N d'aliquot, 6 = De +- D0 +- (4 values)
        }
    }
  }

    if (length(fileBIN.list) == 0) {
      .throw_error("All provided objects were removed")
    }

  if(verbose){
    cat("[analyse_baSAR()] Preliminary analysis in progress, this may take a while ... \n")
  }

  for (k in seq_along(fileBIN.list)) {
    ## check that the data available is consistent
    length.data <- nrow(fileBIN.list[[k]]@METADATA)
    length.disc <- length(Disc[[k]])
    if (length.data %% length.disc != 0) {
      ## this can happen if the input data was subset incorrectly (#517)
      .throw_error("In input ", k, " the number of data points (",
                   length.data, ") is not a multiple of the number of ",
                   "positions (", length.disc, ")")
    }

    ### METADATA
    n_index.vector <- fileBIN.list[[k]]@METADATA[["ID"]]
    measured_discs.vector <- fileBIN.list[[k]]@METADATA[["POSITION"]]
    measured_grains.vector <- fileBIN.list[[k]]@METADATA[["GRAIN"]]
    fname <- fileBIN.list[[k]]@METADATA[["FNAME"]]

    ## always get irradiation times
    irrad_time.vector <- fileBIN.list[[k]]@METADATA[["IRR_TIME"]]

    ## now we override, keep in mind that we do not care about the pattern
    if(!is.null(irradiation_times))
      irrad_time.vector <- rep_len(irradiation_times, length(irrad_time.vector))

    ##if all irradiation times are 0 we should stop here
    if (length(unique(irrad_time.vector)) == 1) {
      .throw_message("All irradiation times are identical, NULL returned")
      return(NULL)
    }

    dose.rate <- unlist(source_doserate[[k]][1]) %||% 1

    ### Automatic Filling - Disc_Grain.list
    for (i in seq_along(Disc[[k]])) {
      disc_selected <-  as.integer(Disc[[k]][i])
      grain_selected <- if (Mono_grain) as.integer(Grain[[k]][i]) else 0

      ## hard break if the disc number or grain number does not fit
      msg <- "In BIN-file '%s' %s number %d does not exist, NULL returned"

      ## disc (position)
      disc_logic <- disc_selected == measured_discs.vector
      if (!any(disc_logic)) {
        .throw_message(sprintf(msg, unique(fname), "position", disc_selected))
        return(NULL)
      }

      ## grain
      grain_logic <- grain_selected == measured_grains.vector
      if (!any(grain_logic)) {
        .throw_message(sprintf(msg, unique(fname), "grain", grain_selected))
        return(NULL)
      }

      ## if the test passed, compile index list
      index_list <- n_index.vector[disc_logic & grain_logic]

      if (!Mono_grain)
        grain_selected <- 1

      for (t in index_list) {
        dose.value <- irrad_time.vector[t] * dose.rate
              s <- 1 + length( Disc_Grain.list[[k]][[disc_selected]][[grain_selected]][[1]] )
              Disc_Grain.list[[k]][[disc_selected]][[grain_selected]][[1]][s] <- n_index.vector[t]  # indexes
              if ( s%%2 == 1) { Disc_Grain.list[[k]][[disc_selected]][[grain_selected]][[2]][as.integer(1+s/2)] <- dose.value  }      # irradiation doses
          }
    }
  }


  ######################  Data associated with a single Disc/Grain
  max_cycles <-  0
  count <- 1

  for (k in seq_along(fileBIN.list)) {

    ##plot Ln and Tn curves if wanted
    ##we want to plot the Ln and Tn curves to get a better feeling
    ##The approach here is rather rough coded, but it works
    if (plot) {
      curve_index <- vapply(seq_along(Disc[[k]]), function(i) {
        dd <- as.integer(Disc[[k]][i])
        gg <- if (Mono_grain) as.integer(Grain[[k]][i]) else 1

        Ln_index <- as.numeric(Disc_Grain.list[[k]][[dd]][[gg]][[1]][1])
        Tn_index <- as.numeric(Disc_Grain.list[[k]][[dd]][[gg]][[1]][2])
        return(c(Ln_index, Tn_index))
      }, FUN.VALUE = numeric(2))


      ## data.tables for Ln and Tn values
      Ln_dt <- rbindlist(list(fileBIN.list[[k]]@DATA[curve_index[1, ]]))
      Tn_dt <- rbindlist(list(fileBIN.list[[k]]@DATA[curve_index[2, ]]))

      ##open plot are
      if (!plot_singlePanels) {
        par(mfrow = c(1, 2))
      }

      ##get natural curve and combine them in matrix
      graphics::matplot(
        x = seq_len(nrow(Ln_dt)),
        y = Ln_dt,
        col = rgb(0, 0, 0, 0.3),
        ylab = "Luminescence [a.u.]",
        xlab = "Channel",
        main = expression(paste(L[n], " - curves")),
        type = "l"
      )

      ##add integration limits
      abline(v = range(signal_integral[[k]]), lty = 2, col = "green")
      abline(v = range(background_integral[[k]]), lty = 2, col = "red")
      mtext(paste0("ALQ: ",count, ":", count + ncol(curve_index)))

      graphics::matplot(
        x = seq_len(nrow(Tn_dt)),
        y = Tn_dt,
        col = rgb(0, 0, 0, 0.3),
        ylab = "Luminescence [a.u.]",
        xlab = "Channel",
        main = expression(paste(T[n], " - curves")),
        type = "l"
      )

      ## add integration limits depending on the chosen value
      abline(v = range(signal_integral_Tx[[k]] %||% signal_integral[[k]]),
             lty = 2, col = "green")
      abline(v = range(background_integral_Tx[[k]] %||% background_integral[[k]]),
             lty = 2, col = "red")

      mtext(paste0("ALQ: ",count, ":", count + ncol(curve_index)))

      ##reset par
      if (!plot_singlePanels) {
        par(mfrow = par.default$mfrow)
      }

      ##remove some variables
      rm(curve_index, Ln_dt, Tn_dt)
    }


    for (i in seq_along(Disc[[k]])) {
      dd <- as.integer(Disc[[k]][i])
      gg <- if (Mono_grain) as.integer(Grain[[k]][i]) else 1
      sel.disc.grain <- Disc_Grain.list[[k]][[dd]][[gg]]

      # Data for the selected Disc-Grain
      for (nb_index in 1:(length(sel.disc.grain[[1]]) / 2)) {

        index1 <- as.numeric(sel.disc.grain[[1]][2 * nb_index - 1])
        index2 <- as.numeric(sel.disc.grain[[1]][2 * nb_index])
        this.data <- fileBIN.list[[k]]@DATA
        Lx.data <- data.frame(seq_along(this.data[[index1]]),
                              this.data[[index1]])
        Tx.data <- data.frame(seq_along(this.data[[index2]]),
                              this.data[[index2]])

        ## call calc_OSLLxTxRatio()
        ## we run this function with a warnings catcher to reduce the load of warnings for the user
        LxTx.table <- .warningCatcher(
          calc_OSLLxTxRatio(
            Lx.data = Lx.data,
            Tx.data = Tx.data,
            signal_integral = signal_integral[[k]],
            signal_integral_Tx = signal_integral_Tx[[k]],
            background_integral = background_integral[[k]],
            background_integral_Tx = background_integral_Tx[[k]],
            background.count.distribution = additional_arguments$background.count.distribution,
            sigmab = sigmab[[k]],
            sig0 = sig0[[k]])$LxTx.table
        )

        Disc_Grain.list[[k]][[dd]][[gg]][[3]][nb_index] <- LxTx.table$LxTx
        Disc_Grain.list[[k]][[dd]][[gg]][[4]][nb_index] <- LxTx.table$LxTx.Error
        Disc_Grain.list[[k]][[dd]][[gg]][[5]][nb_index] <- LxTx.table$Net_TnTx

        ##free memory
        rm(LxTx.table)
      }

      ## reset `sel.disc.grain` because the data it pointed to has changed
      sel.disc.grain <- Disc_Grain.list[[k]][[dd]][[gg]]

      ## fit dose response curve and plot
      sample_dose <- unlist(sel.disc.grain[[2]])
      sample_LxTx <- unlist(sel.disc.grain[[3]])
      sample_sLxTx <- unlist(sel.disc.grain[[4]])
      TnTx <- unlist(sel.disc.grain[[5]])

      ##create needed data.frame (this way to make sure that rows are doubled if something is missing)
      selected_sample <- as.data.frame(cbind(sample_dose, sample_LxTx, sample_sLxTx, TnTx))

      ## get De and De value
      fitcurve <-
        suppressWarnings(fit_DoseResponseCurve(
          selected_sample,
          na.rm = TRUE,
          fit.method = fit.method,
          fit.force_through_origin = fit.force_through_origin,
          fit.weights = additional_arguments$fit.weights,
          fit.includingRepeatedRegPoints = fit.includingRepeatedRegPoints,
          fit.bounds = additional_arguments$fit.bounds,
          n.MC = additional_arguments$n.MC,
          verbose = verbose,
          txtProgressBar = FALSE
        ))
      if (!is.null(fitcurve) && additional_arguments$plot_drc) {
        plot_DoseResponseCurve(
            fitcurve,
            plot_extended = additional_arguments$plot_extended,
            verbose = verbose,
            main = paste0("ALQ: ", count, " | POS: ", Disc[[k]][i],
                          " | GRAIN: ", Grain[[k]][i]))
      }

        ##get data.frame with De values
        Disc_Grain.list[[k]][[dd]][[gg]][[6]][1:4] <- NA
        if(!is.null(fitcurve)){
          fitcurve_De <- get_RLum(fitcurve, data.object = "De")
          Disc_Grain.list[[k]][[dd]][[gg]][[6]][1:4] <-
            fitcurve_De[, c("De", "De.Error", "D01", "D01.ERROR")]
        }

      ## reset `sel.disc.grain` because the data it pointed to has changed
      sel.disc.grain <- Disc_Grain.list[[k]][[dd]][[gg]]

      Limited_cycles[previous.Nb_aliquots + i] <- length(sel.disc.grain[[2]])

      max_cycles <- max(length(sel.disc.grain[[2]]), max_cycles)

        previous.Nb_aliquots <-
            length(stats::na.exclude(Limited_cycles)) # Total count of aliquots

      count <- count + 1
    }
  }   ##  END of loop on BIN files

  Nb_aliquots <- previous.Nb_aliquots

  ##create results matrix
  OUTPUT_results <-
    matrix(nrow = Nb_aliquots,
           ncol = (8 + 3 * max_cycles),
           byrow = TRUE)

  ## set column name (this makes it much easier to debug)
  colnames(OUTPUT_results) <- c(
    "INDEX_BINfile",
    "DISC",
    "GRAIN",
    "DE",
    "DE.SD",
    "D0",
    "D0.SD",
    "CYCLES_NB",
    paste0("DOSE_", 1:max_cycles),
    paste0("LxTx_", 1:max_cycles),
    paste0("LxTx_", 1:max_cycles, ".SD")
  )

  comptage <- 0
  for (k in seq_along(fileBIN.list)) {
    for (i in seq_along(Disc[[k]])) {
      dd <- as.numeric(Disc[[k]][i])
      gg <- if (Mono_grain) as.numeric(Grain[[k]][i]) else 1

      comptage <- comptage + 1
      OUTPUT_results[comptage, 1] <- k
      OUTPUT_results[comptage, 2] <- dd
      OUTPUT_results[comptage, 3] <- if (Mono_grain) gg else 0

      sel.disc.grain <- Disc_Grain.list[[k]][[dd]][[gg]]
      if (length(sel.disc.grain[[6]]) != 0) {
        ## DE, DE.SD, D0, D0.SD
        OUTPUT_results[comptage, 4:7] <- as.numeric(sel.disc.grain[[6]][1:4])

        ##CYCLES_NB
        OUTPUT_results[comptage, 8] <- length(sel.disc.grain[[2]])

        ## auxiliary variable
        llong <- length(sel.disc.grain[[2]])

        ##Dose
        OUTPUT_results[comptage, 9:(8 + llong)] <- as.numeric(sel.disc.grain[[2]])

        ##LxTx values
        OUTPUT_results[comptage, (9 + max_cycles):(8 + max_cycles + llong)] <-
          as.numeric(sel.disc.grain[[3]])

        ##LxTx SD values
         OUTPUT_results[comptage, (9 + 2*max_cycles):(8 + 2*max_cycles + llong)] <-
           as.numeric(sel.disc.grain[[4]])
      }
    }
  }

  ##Clean matrix and remove all unwanted entries

    ##remove all NA columns, means all NA columns in POSITION and DISC
    ##this NA values are no calculation artefacts, but coming from the data processing and have
    ##no further value
    OUTPUT_results <- OUTPUT_results[!is.na(OUTPUT_results[, 2]), , drop = FALSE]

    ##clean up NaN values in the LxTx and corresponding error values
    ##the transposition of the matrix may increase the performance for very large matrices
    OUTPUT_results_reduced <- t(OUTPUT_results)
    selection <- vapply(seq_len(ncol(OUTPUT_results_reduced)), function(x) {
      col <- OUTPUT_results_reduced[9:(8 + 3 * max_cycles), x]
      !any(is.nan(col) | is.infinite(col))
    }, FUN.VALUE = logical(1))

    removed_aliquots <- t(OUTPUT_results_reduced[,!selection])
    OUTPUT_results_reduced <- t(OUTPUT_results_reduced[,selection])

    ## check for difference in the number of dose points, they should be the same
    if (length(unique(OUTPUT_results_reduced[, "CYCLES_NB"])) > 1) {
      .throw_warning("The number of dose points differs across your data set")
    }

    ## correct number of aliquots if necessary
    if (Nb_aliquots > nrow(OUTPUT_results_reduced)) {
      Nb_aliquots <- nrow(OUTPUT_results_reduced)
      .throw_warning("'Nb_aliquots' corrected due to NaN or Inf values ",
                     "in Lx and/or Tx to ", Nb_aliquots, ", you might want ",
                     "to check 'removed_aliquots' in the function output")
    }

    ## prepare for Bayesian analysis
    Doses <- t(OUTPUT_results_reduced[,9:(8 + max_cycles)])
    LxTx <- t(OUTPUT_results_reduced[, (9 + max_cycles):(8 + 2 * max_cycles)])
    LxTx.error <- t(OUTPUT_results_reduced[, (9 + 2 * max_cycles):(8 + 3 * max_cycles)])

    ## prepare data frame for output that can used as input
    BIN_FILE <- character(0)
    if (length(OUTPUT_results_reduced) > 0)
      BIN_FILE <- unlist(object.file_name)[OUTPUT_results_reduced[[1]]]
    input_object <- data.frame(
        BIN_FILE = BIN_FILE,
        OUTPUT_results_reduced[, -1, drop = FALSE],
        stringsAsFactors = FALSE
    )

    ## prepare data frame for output that shows rejected aliquots
    if (NROW(removed_aliquots) > 0) {
      removed_aliquots <- cbind(BIN_FILE = BIN_FILE, removed_aliquots[, -1])
    } else {
      removed_aliquots <- NULL
    }
  }

  # Call baSAR-function -------------------------------------------------------------------------

  temp.DE <- input_object$DE
  if (all(is.na(temp.DE)) ||
      min(temp.DE[temp.DE > 0], na.rm = TRUE) < lower_centralD ||
      max(temp.DE, na.rm = TRUE) > upper_centralD) {
      .throw_warning("Your lower_centralD and/or upper_centralD values ",
                     "seem not to fit to your input data, which may indicate ",
                     "an incorrect 'source_doserate'")
    }

  ##>> try here is much better, as the user might run a very long preprocessing and do not
  ##want to fail here
  results <-
    try(.baSAR_function(
      Nb_aliquots = Nb_aliquots,
      distribution = distribution,
      data.Dose = Doses,
      data.Lum = LxTx,
      data.sLum = LxTx.error,
      fit.method = fit.method,
      n.MCMC = n.MCMC,
      fit.force_through_origin = fit.force_through_origin,
      fit.includingRepeatedRegPoints = fit.includingRepeatedRegPoints,
      method_control = method_control,
      lower_centralD = lower_centralD,
      upper_centralD = upper_centralD,
      baSAR_model = baSAR_model,
      verbose = verbose
    ), outFile = stdout()) # redirect error messages so they can be silenced

  ##check whether this became NULL
  if (!inherits(results, "try-error")) {
    ##how do we add the systematic error?
    ##(1) source_doserate is a list, not a vector, but the user can
    ##provide many source dose rates and he can provide only a single vector (no error)

    systematic_error <- 0
    if(!is.null(unlist(source_doserate)) || !is.null(function_arguments$source_doserate)){
      ##if it comes from the previous call, it is, unfortunately not that simple
      if(!is.null(function_arguments$source_doserate)){
        source_doserate <- eval(function_arguments$source_doserate)
        if (!inherits(source_doserate, "list")) {
          source_doserate <- list(source_doserate)
        }
      }

      systematic_error <- unlist(lapply(source_doserate, function(x) {
        if (length(x) == 2 && !is.na(x[2])) x[2] else 0
      }))

      ## check if different errors were provided
      if (length(unique(systematic_error)) > 1) {
        .throw_warning("Provided source dose rate errors differ: the mean was ",
                       "taken, but the calculated systematic error may not be valid")
      }
    }

    ##add to the final de
    DE_FINAL.ERROR <- sqrt(results[[1]][["CENTRAL.SD"]]^2 + mean(systematic_error)^2)

    ##combine
    results[[1]] <- cbind(results[[1]], DE_FINAL = results[[1]][["CENTRAL"]], DE_FINAL.ERROR = DE_FINAL.ERROR)

  }else{
    results <- NULL
    verbose <- FALSE
    plot <- FALSE
  }

  # Terminal output -----------------------------------------------------------------------------
  if(verbose){
    cat("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n")
    cat("\n[analyse_baSAR()] ---- RESULTS ---- \n")
    cat("------------------------------------------------------------------\n")
    cat("Used distribution:\t\t", results[[1]][["DISTRIBUTION"]], "\n")
    num.aliquots <- results[[1]][["NB_ALIQUOTS"]]
    tot.aliquots <- num.aliquots
    if (!is.null(removed_aliquots))
      tot.aliquots <- tot.aliquots + nrow(removed_aliquots)
    cat(paste0("Number of aliquots used:\t", num.aliquots, "/", tot.aliquots))
    if (!is.null(aliquot_range)) {
      cat(" (manually removed: ", length(aliquot_range), ")\n", sep = "")
    } else {
      cat("\n")
    }

    cat("Considered fitting method:\t", results[[1]][["FIT_METHOD"]],
        if (!is.null(baSAR_model)) " (user defined)", "\n")
    cat("Number of independent chains:\t", results[[1]][["N.CHAINS"]], "\n")
    cat("Number MCMC iterations/chain:\t", results[[1]][["N.MCMC"]], "\n")
    cat("------------------------------------------------------------------\n")
    cat(sprintf("\t\t\t\tmean%s\tsd\tHPD\n",
                ifelse(distribution == "log_normal", "*", "")))

    cat(paste0(">> Central dose:\t\t", results[[1]][["CENTRAL"]],"\t",
               results[[1]][["CENTRAL.SD"]],"\t",
               "[", results[[1]][["CENTRAL_Q_.16"]]," ; ", results[[1]][["CENTRAL_Q_.84"]], "]**\t"))
    cat(paste0("\n\t\t\t\t\t\t[", results[[1]][["CENTRAL_Q_.025"]]," ; ", results[[1]][["CENTRAL_Q_.975"]],"]***"))

    cat(paste0("\n>> sigma_D:\t\t\t", results[[1]][["SIGMA"]],"\t", results[[1]][["SIGMA.SD"]], "\t",
               "[",results[[1]][["SIGMA_Q_.16"]]," ; ", results[[1]][["SIGMA_Q_.84"]], "]**\t"))
    cat(paste0("\n\t\t\t\t\t\t[",results[[1]][["SIGMA_Q_.025"]]," ; ", results[[1]][["SIGMA_Q_.975"]], "]***"))
    cat(paste0("\n>> Final central De:\t\t", results[[1]][["DE_FINAL"]],"\t", round(results[[1]][["DE_FINAL.ERROR"]], digits = digits), "\t",
               " - \t -"))
    cat("\n------------------------------------------------------------------\n")
    cat("(systematic error contribution to final De:",
        format((1 - results[[1]][["CENTRAL.SD"]] / results[[1]][["DE_FINAL.ERROR"]]) * 100,
               scientific = TRUE), "%)\n")
    if(distribution == "log_normal"){
     cat("* mean of the central dose is the geometric mean\n")
    }
    cat("** 68 % level | *** 95 % level\n")
  }


  # Plotting ------------------------------------------------------------------------------------
  if(plot){

    ## deprecated argument
    if ("plot.single" %in% ...names()) {
      plot_singlePanels <- list(...)$plot.single
      .deprecated("plot.single", "plot_singlePanels", since = "1.0.0")
    }

    ##get colours from the package Luminescence
    col <- get("col", pos = .LuminescenceEnv)

    ##get list of variable names (we need them later)
    varnames <- coda::varnames(results[[2]])

    ##////////////////////////////////////////////////////////////////////////////////////////////
    ##TRACE AND DENSITY PLOT
    ####//////////////////////////////////////////////////////////////////////////////////////////
    if(plot_reduced){
      if (!all(c("central_D", "sigma_D") %in% variable.names)) {
        var.missing <- setdiff(c("central_D", "sigma_D"), variable.names)
        .throw_message("Plots for 'central_D' and 'sigma_D' could not be ",
                       "produced as 'variable.names' does not include ",
                       .collapse(var.missing))
      } else {
        try(plot(results[[2]][, c("central_D", "sigma_D"), drop = FALSE]),
            silent = TRUE)
      }
    }else{
      try(plot(results[[2]]))
    }


    ##////////////////////////////////////////////////////////////////////////////////////////////
    ##TRUE DOSE PLOT AND DECISION MAKER
    ####//////////////////////////////////////////////////////////////////////////////////////////
    if (!plot_singlePanels) {
      par(mfrow = c(2, 2))
    }

    ##get list with D values
    ##get list out of it
    plot_matrix <- as.matrix(results[[2]][, grep(varnames, pattern = "^D")])
    aliquot_quantiles <- apply(plot_matrix, 2, quantile, c(0.25, 0.75))

    ##define boxplot colours ... we have red and orange
    box.col <- vapply(seq_len(ncol(aliquot_quantiles)), function(x) {
      idx <- 1
      if (aliquot_quantiles[2, x] < results[[1]][, "CENTRAL_Q_.16"] ||
          aliquot_quantiles[1, x] > results[[1]][, "CENTRAL_Q_.84"])
        idx <- idx + 1
      if (aliquot_quantiles[2, x] < results[[1]][, "CENTRAL_Q_.025"] ||
          aliquot_quantiles[1, x] > results[[1]][, "CENTRAL_Q_.975"])
        idx <- idx + 1
      c("white", "orange", col[2])[idx]
    }, FUN.VALUE = character(1))

    ## make selection according to the model for the curve plotting
    ExpoGC <- as.numeric(grepl("EXP", fit.method))
    LinGC <- as.numeric(grepl("LIN", fit.method))
    GC_Origin <- as.numeric(fit.force_through_origin)

    ## to ensure a minimum of quality no more than 15 boxes are plotted in each plot
    i <- 1
    while(i < ncol(plot_matrix)){
      step <- min(ncol(plot_matrix), i + 14)
      plot_check <- try(graphics::boxplot(
        x = plot_matrix[,i:step],
        use.cols = TRUE,
        horizontal = TRUE,
        outline = TRUE,
        col = box.col[i:step],
        xlab = ifelse(is.null(unlist(source_doserate)), "Dose [s]", "Dose [Gy]"),
        ylab = "Aliquot index",
        yaxt = "n",
        xlim = c(1,19),
        main = paste0("Individual Doses | ALQ: ", i,":",step)
      ))

      if (!inherits(plot_check, "try-error")) {
        ## pad with spaces then take first 15
        labels <- c(i:step, rep(" ", 15))[1:15]
        tick.pos <- which(labels != " ")
        axis(side = 2, at = tick.pos, labels = labels[tick.pos], cex.axis = 0.8)

      ##add HPD with text
      ##HPD - 68%
      lines(
        x = c(
          results[[1]][, "CENTRAL_Q_.16"], results[[1]][, "CENTRAL_Q_.16"],
          results[[1]][, "CENTRAL_Q_.84"], results[[1]][, "CENTRAL_Q_.84"]),
        y = c(par()$usr[3], 16, 16, par()$usr[3]),
        lty = 3,
        col = col[3],
        lwd = 1.5
      )
      text(
        x = results[[1]][, "CENTRAL"],
        y = 16,
        labels = "68 %",
        pos = 3,
        col = col[3],
        cex = 0.9 * par()$cex
      )

      ##HPD - 98 %%
      lines(
        x = c(
          results[[1]][, "CENTRAL_Q_.025"], results[[1]][, "CENTRAL_Q_.025"],
          results[[1]][, "CENTRAL_Q_.975"], results[[1]][, "CENTRAL_Q_.975"]),
        y = c(par()$usr[3], 17.5, 17.5, par()$usr[3]),
        lty = 3,
        col = col[2],
        lwd = 1.5
      )

      text(
        x = results[[1]][, "CENTRAL"],
        y = 17.5,
        labels = "95 %",
        pos = 3,
        col = col[2],
        cex = 0.9 * par()$cex)
      }
      ##update counter
      i <- i + 15
    }

    if (!plot_singlePanels) {
      par(mfrow = c(1,2))
    }
    ##////////////////////////////////////////////////////////////////////////////////////////////
    ##DOSE RESPONSE CURVES AND Lx/Tx VALUES
    ####//////////////////////////////////////////////////////////////////////////////////////////

      ##define selection vector
      selection <- c("a[", "b[", "c[", "g[", "Q[1,")

      ##get list out of it
      list_selection <- lapply(X = selection, FUN = function(x){
        unlist(results[[2]][,grep(x = varnames, pattern = x, fixed = TRUE)])
      })

      ## assign only the first letter to avoid `[` in the names
      names(list_selection) <- strtrim(selection, 1)

      ##create matrix
      plot_matrix <- do.call(what = "cbind", args = list_selection)

      ##free memory
      rm(list_selection)

      ##add choise for own provided model
      fit.method_plot <- fit.method
      if(!is.null(baSAR_model)){
        fit.method_plot <- paste(fit.method_plot, "(user defined)")
      }

      ## set xlim and ylim based on ranges in the input
      xlim <- c(0, max(input_object[, grep("DOSE", colnames(input_object))],
                       na.rm = TRUE) * 1.1)
      ylim <- range(input_object[, grep("LxTx", colnames(input_object))],
                    na.rm = TRUE) * c(1, 1.1)

      ## check for position of the legend ... we can do better
      legend_pos <- if (results[[1]][["CENTRAL_Q_.975"]] < max(xlim)/2)
                      "topright" else "topleft"

        ##set plot area
        plot_check <- try(plot(
          NA,
          NA,
          ylim = ylim,
          xlim = xlim,
          ylab = expression(paste(L[x] / T[x])),
          xlab = ifelse(is.null(unlist(source_doserate)), "Dose [s]", "Dose [Gy]"),
          main = "baSAR Dose Response Curves"
        ))

        if (!inherits(plot_check, "try-error")) {
          ##add mtext
          mtext(side = 3, text = paste("Fit:", fit.method_plot))

          ##check whether we have all data we need (might be not the case of the user
          ##selects own variables)
          var.required <- c("a", "b", "c", "g")
          if (nrow(plot_matrix) != 0 && all(var.required %in% variable.names)) {
            ##plot individual dose response curves
            x <- NA
            for (i in seq_len(min(nrow(plot_matrix), 1000))) {
              curve(
                  GC_Origin * plot_matrix[i, "g"] +
                  LinGC * (plot_matrix[i, "c"] * x) +
                  ExpoGC * (plot_matrix[i, "a"] *
                            (1 - exp(-x / plot_matrix[i, "b"]))),
                  add = TRUE,
                  col = rgb(0, 0, 0, 0.1)
              )
            }
          }else{
            var.missing <- setdiff(var.required, variable.names)
            .throw_message("Dose-response curves could not be plotted as ",
                           "'variable.names' does not include ",
                           .collapse(var.missing))
          }

          ##add dose points
          n.col <-
            length(input_object[, grep(x = colnames(input_object), pattern = "DOSE")])

          ##add rug with natural Lx/Tx
          graphics::rug(side = 2, x = input_object[[9 + n.col]])

          ##plot Lx/Tx values .. without errors ... this is enough here
          for (i in 2:length(input_object[, grep(x = colnames(input_object), pattern = "DOSE")])) {
            ##add error bars
            segments(
              x0 = input_object[[8 + i]],
              x1 = input_object[[8 + i]],
              y0 = input_object[[8 + n.col + i]] - input_object[[8 + 2 * n.col + i]],
              y1 = input_object[[8 + n.col + i]] + input_object[[8 + 2 * n.col + i]],
              col = "grey"
            )

            ##add points in the top of it
            points(
              x = input_object[[8 + i]],
              y = input_object[[8 + n.col + i]],
              pch = 21,
              col = col[11],
              bg = "grey"
            )
          }

          ##add ablines
          abline(
            v = results[[1]][, c("CENTRAL_Q_.16", "CENTRAL_Q_.84")],
            lty = 3,
            col = col[3],
            lwd = 1.2
          )
          abline(v = results[[1]][, c("CENTRAL_Q_.025", "CENTRAL_Q_.975")], lty = 2, col = col[2])

          ##add legend1
          legend(
            legend_pos,
            bty = "n",
            horiz = FALSE,
            lty = c(3, 2),
            col = c(col[3], col[2]),
            legend = c("HPD - 68 %", "HPD - 95 %")
          )

          ##add legend2
          legend(
            "bottomright",
            bty = "n",
            horiz = FALSE,
            pch = 21,
            col = col[11],
            bg = "grey",
            legend = "measured dose points"
          )
        }
      ##remove object, it might be rather big
      rm(plot_matrix)

      ##03 Abanico Plot
      if(distribution_plot == "abanico"){
        plot_check <- plot_AbanicoPlot(
          data = input_object[, c("DE", "DE.SD")],
          zlab = ifelse(is.null(unlist(source_doserate)),
                        expression(paste(D[e], " [s]")),
                        expression(paste(D[e], " [Gy]"))),
          log.z = distribution == "log_normal",
          z.0 = results[[1]]$CENTRAL,
          y.axis = FALSE,
          polygon.col = FALSE,
          line = results[[1]][,c(
            "CENTRAL_Q_.16", "CENTRAL_Q_.84", "CENTRAL_Q_.025", "CENTRAL_Q_.975")],
          line.col = c(col[3], col[3], col[2], col[2]),
          line.lty = c(3,3,2,2),
          output = TRUE,
          mtext = paste0(
            nrow(input_object) - length(which(is.na(input_object[, c("DE", "DE.SD")]))),
            "/",
            nrow(input_object),
            " plotted (removed are NA values)"
          )
        )

        if (!is.null(plot_check)) {
          legend(
            "topleft",
            legend = c("Central dose", "HPD - 68%", "HPD - 95 %"),
            lty = c(2, 3, 2),
            col = c("black", col[3], col[2]),
            bty = "n",
            cex = par()$cex * 0.8
          )
        }
      }else{
        plot_check <- NULL
      }

      ##In case the Abanico plot will not work because of negative values
      ##provide a KDE
      if(is.null(plot_check) && distribution_plot == "kde"){
        plot_check <- try(suppressWarnings(plot_KDE(
          data = input_object[, c("DE", "DE.SD")],
          xlab = ifelse(is.null(unlist(source_doserate)),
                        expression(paste(D[e], " [s]")),
                        expression(paste(D[e], " [Gy]"))),
          mtext =   paste0(
            nrow(input_object) - length(which(is.na(input_object[, c("DE", "DE.SD")]))),
            "/",
            nrow(input_object),
            " (removed are NA values)"
          )
        )), outFile = stdout()) # redirect error messages so they can be silenced

        if (!inherits(plot_check, "try-error")) {
          abline(v = results[[1]]$CENTRAL, lty = 2)
          abline(
            v = results[[1]][, c("CENTRAL_Q_.16", "CENTRAL_Q_.84")],
            lty = 3,
            col = col[3],
            lwd = 1.2
          )
          abline(v = results[[1]][, c("CENTRAL_Q_.025", "CENTRAL_Q_.975")], lty = 2, col = col[2])

          legend(
            legend_pos,
            legend = c("Central dose", "HPD - 68%", "HPD - 95 %"),
            lty = c(2, 3, 2),
            col = c("black", col[3], col[2]),
            bty = "n",
            cex = par()$cex * 0.8
          )
        }
      }
  }

  # Return --------------------------------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      summary = results[[1]],
      mcmc = results[[2]],
      models = results[[3]],
      input_object = input_object,
      removed_aliquots = removed_aliquots
      ),
    info = list(call = sys.call())
  )
}

# INTERNAL HELPER FUNCTIONS -----------------------------------------------
#+++++++++++++++++++++
#+ .set_pid()        +
#+++++++++++++++++++++
#' Set unique id of the RLum.Analysis object as parent id for each RLum.Data
#' object in the record list
#'
#' This function only applies on RLum.Analysis objects and was written for performance not
#' usability, means the functions runs without any checks and is for internal usage only.
#'
#' @param [Luminescence::RLum.Analysis-class] (**required**):
#' input object where the function should be applied on
#'
#' @return
#' Returns the same object as the input
#'
#' @section Function version: 0.1.0
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @examples
#'
#' ##example using self created data
#' object <- set_RLum(
#' "RLum.Analysis",
#' records = list(
#'  set_RLum("RLum.Data.Curve"),
#'  set_RLum("RLum.Data.Curve")))
#'
#' object <- .set_pid(object)
#'
#' @noRd
.set_pid <- function(object){
  object@records <-
    lapply(object@records, function(x) {
      x@.pid  <- object@.uid
      return(x)
    })

  object
}

#+++++++++++++++++++++
#+ .warningCatcher() +
#+++++++++++++++++++++

#' Catches warnings returned by a function and merges them into a single one.
#' The original return value of the function is returned. This function is
#' particularly helpful if a function returns a lot of identical warnings.
#'
#' @param expr [expression] (**required**):
#' the R expression, usually a function
#'
#' @return
#' Returns the same object as the input and throws a warning.
#'
#' @section Function version: 0.2.0
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @examples
#'
#' f <- function() {
#'  warning("warning 1")
#'  warning("warning 1")
#'  warning("warning 2")
#'  1:10
#' }
#' print(.warningCatcher(f()))
#'
#' @noRd
.warningCatcher <- function(expr) {
  ##set variables
  warning_collector <- list()

  ##run function and catch warnings
  results <- withCallingHandlers(
    expr = expr,
    warning = function(w) {
      warning_collector <<- c(warning_collector, w$message)
      tryInvokeRestart("muffleWarning")
    }
  )

  ##set new warning messages with merged results
  if (length(warning_collector) > 0) {
    w_table <- table(as.character(unlist(warning_collector)))
    w_single <- rep(length(w_table) == 1, length(w_table))
    msg <- sprintf("%s%s%s",
                   ifelse(w_single, "", sprintf("(%d) ", 1:length(w_table))),
                   names(w_table),
                   ifelse(w_table == 1, "", sprintf(" [%d times]", w_table)))
    warning(paste(msg, collapse = "\n"), call. = FALSE)
  }
  results
}

#+++++++++++++++++++++
#+ .smoothing()      +
#+++++++++++++++++++++

#' Smooth data based on rolling means or medians
#'
#' The function allows a direct and meaningful access to the functionality of
#' `data.table::frollmean()` and `data.table::frollmedian()`. It also provides
#' an implementation of the Poisson smoother of Carter et al. (2018).
#'
#' @param x [numeric] (**required**):
#' the object for which the smoothing should be applied.
#'
#' @param k [integer] (*with default*):
#' window size for the rolling mean or median (ignored if
#' `method = "Carter_etal_2018"`).
#'
#' @param fill [numeric] (*with default*):
#' value used to pad the result so to have the same length as the input.
#'
#' @param align [character] (*with default*):
#' one of `"right"`, `"center"` or `"left"`, specifying whether the index
#' of the result should be right-aligned (default), centred, or left-aligned
#' compared to the rolling window of observations (ignored if
#' `method = "Carter_etal_2018"`).
#'
#' @param p_acceptance [numeric] (*with default*):
#' probability threshold of accepting a value to be a sample from a Poisson
#' distribution (only used for `method = "Carter_etal_2018"`). Values that
#' have a Poisson probability below the threshold are replaced by the average
#' over the four neighbouring values.
#'
#' @param method [character] (*with default*):
#' smoothing method to be applied: one of `"mean"`, `"median"` or
#' `"Carter_etal_2018"`.
#'
#' @return
#' Returns a numeric vector with smoothed values.
#'
#' @section Function version: 0.4
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @examples
#'
#' v <- 1:100
#' .smoothing(v)
#'
#' @noRd
.smoothing <- function(
  x,
  k = NULL,
  fill = NA,
  align = "right",
  p_acceptance = 1e-7,
  method = "mean") {
  .set_function_name(".smoothing")
  on.exit(.unset_function_name(), add = TRUE)

  .validate_positive_scalar(k, int = TRUE, null.ok = TRUE)
  if (is.array(fill) || !(length(fill) == 1 && is.na(fill)))
    .validate_class(fill, "numeric", length = 1, extra = "NA")
  .validate_args(align, c("right", "center", "left"))
  method <- .validate_args(method, c("mean", "median", "Carter_etal_2018"))
  .validate_positive_scalar(p_acceptance)

  ##set k
  if (is.null(k))
   k <- ceiling(length(x) / 100)

  ##smooth data
  switch(method,
  mean = {
    data.table::frollmean(x, n = k, fill = fill, align = align)
  },
  median = {
    data.table::frollmedian(x, n = k, fill = fill, align = align)
  },
  Carter_etal_2018 = {
    ## Code derived with corrections and improvements from the supplementary
    ## materials of Carter et al. (2018)
    ## https://doi.org/10.1016/j.radmeas.2018.05.010

    ## Poisson probability of sample counts
    ## We operate on logged values to avoid overflow that would happen with
    ## the naive formulation `exp(-mx) * mx^x / factorial(x)`
    mx <- mean(x, na.rm = TRUE)
    prob <- exp(-mx + x * log(mx) - lfactorial(x))

    ## remove counts with probability below the acceptance threshold
    ## so they are not considered when averaging over the neighbours
    na.idx <- prob < p_acceptance
    x[na.idx] <- NA
    if (all(is.na(x)))
      .throw_error("'p_acceptance' rejects all counts, set it to a smaller value")

    ## average over four neighbours: we set "n = 5" instead of "n = 4" because
    ## we want to include the 2 neighbours to the left and the 2 on the right,
    ## but the window size also counts the element itself; this is not a
    ## problem, as we use the rolling mean only for elements that we have set
    ## to NA just above, and they don't have a value to contribute to the mean
    x[na.idx] <- data.table::frollmean(x, n = 5, align = "center",
                                       fill = fill, na.rm = TRUE)[na.idx]
    round(x)
  })
}

#' Computation of a weighted median
#'
#' @param x [numeric] (**required**):
#' Values whose weighted median is to be computed.
#'
#' @param w [numeric] (**required**):
#' Weights to use for elements of `x`.
#'
#' @param na.rm [logical] (*with default*):
#' Whether `NA` values in `x` should be removed before the computation.
#'
#' @return
#' A numeric value corresponding to the weighted median of the input vector.
#'
#' @noRd
.weighted.median <- function(x, w, na.rm = FALSE) {
  ## based on limma::weighted.median
  if (na.rm) {
    keep.idx <- !is.na(x)
    x <- x[keep.idx]
    w <- w[keep.idx]
  }
  ox <- order(x)
  x <- x[ox]
  w <- w[ox]
  p <- cumsum(w) / sum(w)
  n <- sum(p < 0.5)
  if (p[n + 1] > 0.5)
    return(x[n + 1])
  (x[n + 1] + x[n + 2]) / 2
}

#++++++++++++++++++++++++++++++
#+ Curve normalisation        +
#++++++++++++++++++++++++++++++

#' Curve normalisation
#'
#' Details on the normalisation methods are specified in
#' [Luminescence::plot_RLum.Data.Curve].
#'
#' The function assumes that `NA` or other invalid values have already been
#' removed by the caller function, and that the `norm` option has already
#' been validated.
#'
#' @param data [numeric], [matrix] (**required**):
#' the curve data to be normalised.
#'
#' @param norm [logical] [character] (**required**):
#' if logical, whether curve normalisation should occur; alternatively, one
#' of `"max"` (used with `TRUE`), `"min"`, `"first"`, `"last"`, `"huot"`,
#' or a positive number (e.g., 2.2; here it can also be a vector of the same
#' length as `data` to support the `"intensity"` option).
#'
#' @noRd
.normalise_curve <- function(data, norm) {
  ## early return if no normalisation is requested
  if (isFALSE(norm)) {
    return(data)
  }

  if (inherits(norm, "numeric")) {
    data <- data / norm
  }
  else if (norm == "max" || isTRUE(norm)) {
    data <- data / max(data, na.rm = TRUE)
  }
  else if (norm == "last") {
    data <- data / data[length(data)]
  }
  else if (norm == "first") {
    data <- data / data[1]
  }
  else if (norm == "min") {
    data <- data / min(data, na.rm = TRUE)
  }
  else if (norm == "huot") {
    bg <- median(data[floor(length(data) * 0.8):length(data)], na.rm = TRUE)
    data <- (data - bg) / max(data - bg, na.rm = TRUE)
  }

  ## check for Inf and NA
  if (any(is.infinite(data)) || anyNA(data)) {
    data[is.infinite(data) | is.na(data)] <- 0
    .throw_warning("Curve normalisation produced Inf/NaN values, ",
                   "values replaced by 0")
  }

  data
}

#++++++++++++++++++++++++++++++
#+ LxTx error calculation     +
#++++++++++++++++++++++++++++++

#' Calculation of the `Lx/Tx` error
#'
#' Calculates the `Lx/Tx` error according Galbraith (2014).
#'
#' @param LnLxTnTx [data.frame] (**required**): containing columns
#'        `Net_LnLx`, `Net_TnTx`, `Net_LnLx.Error`, `Net_TnTx.Error`.
#' @param sig0 [numeric] (**required**): extra error component to be added
#'        to the final `Lx/Tx` error value (e.g., instrumental error).
#' @param digits [integer] (**required**): round numbers to the specified
#'        digits. If set to `NULL`, no rounding occurs.
#'
#' @noRd
.calculate_LxTx_error <- function(LnLxTnTx, sig0, digits) {

  ## extract fields from the data.frame
  LnLx <- LnLxTnTx$Net_LnLx
  TnTx <- LnLxTnTx$Net_TnTx
  LnLx.Error <- LnLxTnTx$Net_LnLx.Error
  TnTx.Error <- LnLxTnTx$Net_TnTx.Error

  ## calculate Ln/Tx
  LxTx <- LnLx / TnTx
  if (is.nan(LxTx)) LxTx <- 0

  ## calculate Ln/Tx error
  LxTx.relError <- sqrt((LnLx.Error / LnLx)^2 + (TnTx.Error / TnTx)^2)
  LxTx.Error <- abs(LxTx * LxTx.relError)
  if (is.nan(LxTx.Error)) LxTx.Error <- 0

  ## add an extra component of error
  LxTx.Error <- sqrt(LxTx.Error^2 + (sig0 * LxTx)^2)

  ## combined values
  res <- cbind(LnLxTnTx, LxTx, LxTx.Error)
  if (!is.null(digits)) {
    res[1, ] <- round(res[1, ], digits = digits)
  }

  res
}

#++++++++++++++++++++++++++++++
#+ Scientific axis annotation +
#++++++++++++++++++++++++++++++

#' Bored of the 1e10 notation of large numbers in R? Already tried to force
#' R to produce more fancy labels? Worry not, fancy_scientific() (written by
#' Jack Aidley) is at your help!
#'
#' Source:
#' [http://stackoverflow.com/questions/11610377/how-do-i-change-the-formatting-of-numbers-on-an-axis-with-ggplot]()
#'
#' @param l [numeric] (**required**):
#' a numeric vector, i.e. the labels that you want to add to your plot
#'
#' @return
#' Returns an expression
#'
#' @section Function version: 0.1.0
#'
#' @author Jack Aidley
#'
#' @examples
#' plot(seq(1e10, 1e20, length.out = 10),
#'      1:10,
#'      xaxt = "n")
#'
#' axis(1, at = axTicks(1),
#'      labels = fancy_scientific(axTicks(1)))
#'
#' @noRd
fancy_scientific <- function(l) {
  # turn in to character string in scientific notation
  l <- format(l, scientific = TRUE)
  # quote the part before the exponent to keep all the digits
  l <- gsub("^(.*)e", "'\\1'e", l)
  # turn the 'e+' into plotmath format
  l <- gsub("e", "%*%10^", l)
  # remove plus sign
  l <- gsub("\\+", "", l)
  # return this as an expression
  parse(text=l)
}

#'Add fancy log axis with minor ticks the fancy axis labelling
#'
#'@param side [numeric] (**required**): the side where to plot the axis
#'
#'@param ... extra arguments to be passed to [graphics::axis], `side`, `at`and `labels`
#'are pre-defined and cannot be modified
#'
#'@return
#'Returns fancy log axis
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@examples
#'
#'y <- c(0.1, 0.001, 0.0001)
#'plot(1:length(y), y, yaxt = "n", log = "y")
#'.add_fancy_log_axis(side = 2, las = 1)
#'
#'@noRd
.add_fancy_log_axis <- function(side, ...){
  ## do just nothing if it would cause an error
  if(!(par()$xlog && any(c(1,3) %in% side[1])) && !(par()$ylog && any(c(2,4) %in% side[1])))
    return(NULL)

  ## get current axis ticks and get exponent
  ticks <- graphics::axTicks(side, log = TRUE)
  ticks <- unique(floor(log10(ticks)))
  minor_ticks <- vapply(ticks, function(x) {
      seq(10^(x-1),10^x, length.out = 10)[-10]
  }, numeric(9))

  ## add minor ticks
  graphics::axis(
    side,
    at = as.numeric(minor_ticks),
    lwd.ticks = 0.5,
    tcl = -0.35,
    labels = FALSE)

  ## add main axis
    ## remove settings we set
    args <- list(...)
    args$side <- NULL
    args$at <- NULL
    args$labels <- NULL

  ## call the axis
  do.call(what = graphics::axis, args = c(
    list(side = side,
    at = 10^ticks,
    labels = fancy_scientific(10^ticks)),
    args))
}

#' Add text with a background box to an existing plot
#'
#' @param label [character] (**required**):
#' The text to be written.
#'
#' @param bg [character], [numeric] (*with default*):
#' Background colour. If `NA`, no background box is drawn.
#'
#' @inheritParams base::text
#'
#' @noRd
.text_with_bg <- function(x, y, label, col = "black", bg = NA,
                          pos = NULL, cex = 1, ...) {
  switch(as.character(pos %||% 0),
         "1" = { xjust <- 0.5; yjust <- 1.0 },
         "2" = { xjust <- 1.0; yjust <- 0.5 },
         "3" = { xjust <- 0.5; yjust <- 0.0 },
         "4" = { xjust <- 0.0; yjust <- 0.5 },
         { xjust <- 0.5; yjust <- 0.5 })
  legend(x, y, label, col = col, bg = bg, box.col = NA, cex = cex,
         xjust = xjust, yjust = yjust, x.intersp = -0.5, y.intersp = -0.15, ...)
}

#' Retrieve graphical parameters
#'
#' @return
#' A list of graphical parameters that can be restored, with the following
#' exceptions:
#' - The `pin` parameter is removed because at times (for example, when a pdf
#' file with too small size is used) it can become negative, and restoring it
#' would throw an error.
#' - The `new` parameter is set to `FALSE` because calling `par()` with no
#' graphical device open will set `par("new")` to `TRUE`, which is not what
#' we want to restore, as otherwise the next plot may accidentally overwrite
#' the previous one.
#' - The `usr` parameter is removed so that additional plot elements can be
#' added afterwards using the axis scale of the plot. If we restore that, the
#' axis would range to the default 0-1 interval.
#'
#' @noRd
.par_defaults <- function() {
  pars <- par(no.readonly = TRUE)
  pars$pin <- NULL
  pars$new <- FALSE
  pars$usr <- NULL
  pars
}

#' Convert position keywords into summary placement coordinates
#'
#' @param pos [numeric] or [character] (**required**) Position coordinates
#'        or keyword (one of "topleft", "top", "topright", "left", "center",
#'        "right", "bottomleft", "bottom", "bottomright").
#' @param xlim [numeric] (**required**) The x-axis limits.
#' @param ylim [numeric] (**required**) The y-axis limits.
#'
#' @return
#' A list of two elements: `pos` and `adj`.
#'
#' @noRd
.get_keyword_coordinates <- function(pos, xlim, ylim) {
  adj <- NA
  if (missing(pos)) {
    pos <- c(xlim[1], ylim[2])
    adj <- c(0, 1)
  } else if (length(pos) == 2) {
    pos <- pos
    adj <- c(0, 1)
  } else if (pos[1] == "topleft") {
    pos <- c(xlim[1], ylim[2])
    adj <- c(0, 1)
  } else if (pos[1] == "top") {
    pos <- c(mean(xlim), ylim[2])
    adj <- c(0.5, 1)
  } else if (pos[1] == "topright") {
    pos <- c(xlim[2], ylim[2])
    adj <- c(1, 1)
  } else if (pos[1] == "left") {
    pos <- c(xlim[1], mean(ylim))
    adj <- c(0, 0.5)
  } else if (pos[1] == "center") {
    pos <- c(mean(xlim), mean(ylim))
    adj <- c(0.5, 0.5)
  } else if (pos[1] == "right") {
    pos <- c(xlim[2], mean(ylim))
    adj <- c(1, 0.5)
  }else if (pos[1] == "bottomleft") {
    pos <- c(xlim[1], ylim[1])
    adj <- c(0, 0)
  } else if (pos[1] == "bottom") {
    pos <- c(mean(xlim), ylim[1])
    adj <- c(0.5, 0)
  } else if (pos[1] == "bottomright") {
    pos <- c(xlim[2], ylim[1])
    adj <- c(1, 0)
  }

  list(pos = pos, adj = adj)
}


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#+ Statistical Summary for Plot functions
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#' Create Statistical Summary Character Vector for Plot functions
#'
#' This function automatically generates the statistical summary for the plot functions within
#' the package. This should unify the approach how such things are created and support, theoretically
#' all keywords for all plot functions in a similar way.
#'
#' @param summary [data.frame] (**required**):
#' output from function `calc_Statistics()`.
#'
#' @param keywords[character] (*with default*): keywords supported by function
#' `calc_Statistics()`.
#'
#' @param digits [numeric] (*with default*): modifiy the digits independently
#' for the plot output.
#'
#' @param sep [character] (*with default*): separator used for the creation of
#' the output of the plot.
#'
#'@param prefix [character] (*with default*): prefix to add to the string
#'
#'@param suffix [character] (*with default*): suffix to add to the string
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@section Version: 0.1.0
#'
#'
#'@noRd
.create_StatisticalSummaryText <- function(
  summary,
  keywords = NULL,
  digits = 2, #allow for different digits
  sep = " \n ",
  prefix = "",
  suffix = ""
) {
  ## loop over all keywords provided
  l <- lapply(keywords, function(k) {
    if (!nzchar(k))
      return(NULL)

    ## strip the prefix if necessary
    keywords_prefix <- "unweighted"
    k_strip <- unlist(strsplit(k, split = "$", fixed = TRUE))
    if (length(k_strip) > 1) {
      keywords_prefix <- k_strip[1]
      k_strip <- k_strip[2]
    }

    ## return early if the keyword cannot be found
    value <- summary[[keywords_prefix]][[k_strip]]
    if (is.null(value))
      return(NULL)

    ## construct string
    paste(if (keywords_prefix == "unweighted") k_strip else k,
          "=", round(value, digits))
  })

  ##remove NULL entries
  l <- .rm_NULL_elements(l)

  ##construct final call
  paste0(prefix, paste(unlist(l), collapse = sep), suffix)
}

#++++++++++++++++++++++++++++++
#+ Unlist RLum               +
#++++++++++++++++++++++++++++++
#'
#' Recursive unlisting of lists until the first element in the list
#' is something other than a list. This function helps to get rid of nested
#' lists. The function stops running if a single level list is reached.
#'
#' @param x [list] (**required**): list with lists
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @examples
#' a <- list(b = list(c = list("test")))
#' .unlist_RLum(a)
#'
#' @return [list] with only one level left
#' @noRd
.unlist_RLum <- function(x){
  stopifnot(is.list(x))

  if (length(x) == 0 || !inherits(x[[1]], "list"))
    return(x)

  .unlist_RLum(unlist(x, recursive = FALSE))
}

#++++++++++++++++++++++++++++++
#+ .rm_nonRLum                +
#++++++++++++++++++++++++++++++
#' @title Removes all non-RLum objects from list
#'
#' @description
#' Removes all non-RLum objects from a list supposed to consist only of
#' [Luminescence::RLum-class] objects.
#' As an internal function, the function is rather unforgiving, no further
#' checks are applied.
#'
#' @param x [list] (**required**): list
#'
#' @param class [character] (*with default*):
#' class of elements to keep, by default [Luminescence::RLum-class].
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @examples
#' x <- c(list(set_RLum("RLum.Analysis"), set_RLum("RLum.Analysis")), 2)
#' .rm_nonRLum(x)
#'
#' @return
#' A list containing only objects of the specified class.
#'
#' @noRd
.rm_nonRLum <- function(x, class = "RLum") {
  x[vapply(x, inherits, logical(1), class)]
}

#++++++++++++++++++++++++++++++
#+ .rm_NULL                   +
#++++++++++++++++++++++++++++++
#' @title Removes all NULL elements from list
#'
#' @param x [list] (**required**): list
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @examples
#' l <- list("a", NULL)
#' .rm_NULL_elements(l)
#'
#' @return [list] without `NULL` elements, can be empty
#'
#' @noRd
.rm_NULL_elements <- function(x){
  x[vapply(x, is.null, logical(1))] <- NULL
  x
}

#++++++++++++++++++++++++++++++
#+ .rm_unnamed_elements       +
#++++++++++++++++++++++++++++++
#' @title Removes all unnamed elements from a list
#'
#' @param x [list] (**required**): list
#'
#' @author Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @examples
#' l <- list(a1 = "a", a2 = NULL, 3, 4)
#' .rm_unnamed_elements(l)
#'
#' @return [list] without `NULL` elements, can be empty
#'
#' @noRd
.rm_unnamed_elements <- function(x) {
  nm <- names(x)
  if (is.null(nm)) return(NULL)
  x[nzchar(nm)]
}

#++++++++++++++++++++++++++++++
#+ .matrix_binning            +
#++++++++++++++++++++++++++++++
#' @title Efficient binning of matrices
#'
#' @description This function allows efficient binning of matrices including
#' row and column name handling. Internally, the function uses [rowsum],
#' means the binning is always applied on the rows. For column binning the function
#' internally transposes the matrix first
#'
#' @param m [matrix] (**required**): the matrix uses the base function [rowsum]
#'
#' @param bin_size [integer] (*with default*): bin size
#'
#' @param bin_col [logical] (*with default*): applies the binning on the columns instead of the
#' rows. If you want to perform binning on rows and columns, you have to call this function twice.
#'
#' @param names [character] (*with default*): the handling of the row and column names. The default
#' `NULL` removes the column and row names. Other allowed input is: `'groups'` this uses the group
#' name, e.g., the last time value of a group, `'mean'` this calculates the mean value of a group,
#' `'sum'` to sum-up groups and you can provide any other value which will then be recycled throughout.
#' For example: `c('row1', 'row2')`.
#'
#' @author Sebastian Kreutzer, IRAMAT-CRP2A, UMR 5060, CNRS - Université Bordeaux Montaigne (France)
#'
#' @section Function version: 0.1.2
#'
#' @note Row and column names are transformed to numeric and also summed up; this is not a bug
#' but a feature!
#'
#' @return [matrix]
#'
#' @examples
#'  m <- matrix(data = c(rep(1:20,each = 20)), ncol = 10, nrow = 20)
#'  rownames(m) <- 1:nrow(m)
#'  colnames(m) <- 1:ncol(m)
#'
#' .matrix_binning(m, bin_size = 4)
#'
#' @noRd
.matrix_binning <- function(
  m,
  bin_size = 1,
  bin_col = FALSE,
  names = NULL) {
  .set_function_name(".matrix_binning")
  on.exit(.unset_function_name(), add = TRUE)

  #@ The only check
  .validate_class(m, "matrix")

  ## transpose in column mode
  if(bin_col) m <- t(m)

  ## binning calculation
  ##set groups
  ##with the correction in the 2nd line we
  ##get rid potential problems
  groups <- rep(1:nrow(m), each = bin_size)[1:nrow(m)]

  ##row binning (thats all)
  temp_m <- rowsum(m, group = groups)

  ## Correct names
  if(!is.null(names[1])){
    if(names[1] == "groups"){
      ##get rownames correct (it is the end of each bin)
      row_names <- rownames(m)[which(diff(groups) != 0)]

      ##correct last value
      if(length(row_names) < nrow(m))
        row_names <- c(row_names,rownames(m)[nrow(m)])

    }else if(names[1] == "mean"){
      groups <- rep(1:nrow(m), each = bin_size)[1:nrow(m)]
      row_names <- as.numeric(rownames(m))
      row_names <- tapply(X = row_names, INDEX = groups, FUN = mean)

    }else if(names[1] == "sum"){
      row_names <- rowsum(as.numeric(rownames(m)), group = groups)

    }else{
      row_names <- names
    }

    ##reset rownames and make sure it fits the length
    rownames(temp_m) <- rep_len(row_names, nrow(temp_m))

  }else{
    rownames(temp_m) <- NULL
  }

  ## re-transpose in column mode
  if(bin_col) temp_m <- t(temp_m)

  ## return
  temp_m
}

#++++++++++++++++++++++++++++++
#+ .expand_parameters         +
#++++++++++++++++++++++++++++++
#' @title Expand function parameters of self-call
#'
#' @description For the self-call, the function parameters need to
#' be expanded, this was done, so far in a non-consistent way and
#' repeated in every function using the self-call. This functions
#' does it once and for all similar in all functions.
#'
#' **Note:** the first argument is never extended due to performance reasons,
#' it might be a very large object
#'
#' @param len [numeric] (**required**): length of the parameter expansion
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @return [list] with expanded parameters
#'
#' @noRd
.expand_parameters <- function(len){
  ##get original definition and the call of f
  f_def <- sys.function(sys.parent())
  f_call <- sys.call(sys.parent())

  ## get parent environment because we have to evaluate
  ## objects in the parent environment.
  p_env <- parent.env(environment())

  ##extract arguments (do not consider the first argument, this might be a very
  ##large object)
  args_default <- as.list(f_def)[-length(as.list(f_def))][-1]
  args_new <- as.list(match.call(f_def, f_call, FALSE))[-(1:2)]

  ##now we have to make sure that we evaluate all language objects
  ##before passing them further down
  if(length(args_new) > 0){
    for(i in 1:length(args_new)){
      if(class(args_new[[i]])[1] == "name" |
         class(args_new[[i]])[1] == "call" |
         class(args_new[[i]])[1] == "(" ) {
        args_new[[i]] <- eval(args_new[[i]], envir = p_env)
      }
    }
  }

  ##combine the two argument lists
  args <- modifyList(
    x = args_default,
    val = args_new,
    keep.null = TRUE)

  ##evaluate arguments and take care of missing values
  for(i in 1:length(args)){
    if(is.na(names(args[i])) || names(args[i]) == "...") next
    if(class(args[[i]])[1] == "name" & names(args[i]) != "...") {
      .throw_error("Argument '", names(args[i]), "' missing, with no default")
    }

    ##evaluate and cover special cases
    if(!is.null(args[[i]])) args[[i]] <- eval(args[[i]])
    if(inherits(args[i],  "list") & length(args[[i]]) == 0) args[[i]] <- list()

  }
  ##expand all arguments
  ##we have two conditions and three cases
  ##1:  the argument is a list AND the list itself is not named
  ##    ... the case when the user what to use different values for the objects
  ##2:  the argument is no list ...
  ##    ... the standard automated expansion
  ##    ... OR it is a list with names (e.g., rejection.criteria = list(recycling.ration = 10))
  for(i in 1:length(args)){
    if(inherits(args[[i]], "list") & is.null(names(args[[i]]))){
      args[[i]] <- rep(args[[i]], length = len[1])

    } else {
      args[[i]] <- rep(list(args[[i]]), length = len[1])
    }
  }

  args
}

#++++++++++++++++++++++++++++++
#+ .calc_HPDI                +
#++++++++++++++++++++++++++++++
#' @title Calculates Highest Probability Density Interval
#'
#' @description The idea of this function is to provide a convenient
#' method to calculate the highest probability density intervals for
#' sets of data. This function might be exported later
#' Currently it follows roughly the idea of what is implemented
#' in `code` and `hdrcde`. If the results has more than one peak,
#' also this is shown, therefore the output is a matrix
#'
#' @param object [numeric] (**required**): numeric object with input data
#'
#' @param prob [numeric] (*with default*): sets aimed probability interval
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param ... further arguments passed to [stats::density]
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @references
#' Hyndman, R.J., 1996. Computing and Graphing Highest Density Regions.
#' The American Statistician 50, 120–8. doi:10.2307/2684423
#'
#' @examples
#' x <- rnorm(100)
#' .calc_HPDI(x)
#'
#' @return [matrix] with HPDI
#'
#' @noRd
.calc_HPDI <- function(object, prob = 0.95, plot = FALSE, ...){
  ##estimate density
  dens <- density(object, ...)
  diff <- diff(dens$x[1:2])

  ##calculate probabilities
  m <- cbind(matrix(c(dens$x, dens$y), ncol = 2), dens$y * diff)
  o <- order(m[, 3], decreasing = TRUE)
  m_ind <- which(cumsum(m[o, 3]) <= prob)
  thres <- suppressWarnings(min(m[o, 2][m_ind]))

  ##get peaks
  peaks_id <- which(abs(diff((m[,2] - thres) > 0)) == 1)

  ##calculate HPDI
  HPDI <- matrix(NA, ncol = 2, nrow = 1)
  if (length(peaks_id) > 0)
    HPDI <- matrix(m[peaks_id,1], ncol = 2)

  colnames(HPDI) <- c("lower", "upper")
  attr(HPDI, "Probability") <- prob

  if(plot){
    plot(dens, main = "HPDI (control plot)")
    abline(h = thres, lty = 2)
    if (length(peaks_id) > 0) {
      for(i in seq(1,length(peaks_id),2)) {
        lines(x = m[peaks_id[i]:peaks_id[i + 1], 1],
              y = m[peaks_id[i]:peaks_id[i + 1], 2], col = "red")
      }
    }
  }

  HPDI
}

#++++++++++++++++++++++++++++++
#+ .download_file             +
#++++++++++++++++++++++++++++++
#' @title Internal file download handler
#'
#' @description
#' For file imports using function commencing with `read_` the file download
#' was little consistent and surprisingly error-prone. This function should
#' keep the callers more consistent.
#'
#' @param url [character] (**required**):
#' URL to download.
#'
#' @param destfile [character] (*with default*):
#' filename where the content of the URL should be saved into. By default, a
#' temporary file will be used.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @returns
#' Returns the file path of the downloaded file, or `NULL` if `url` is not
#' valid, or `NA` in case of failure during download.
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @examples
#' ## returns NA (no valid URL detected)
#' .download_file(url = "teststs")
#' .download_file(url = "https://")
#' .download_file(url = "_https://github.com/")
#'
#' ## attempts download
#' .download_file("https://raw.githubusercontent.com/R-Lum/rxylib/master/NEWS.md")
#'
#'@noRd
.download_file <- function(
    url,
    destfile = tempfile(),
    verbose = TRUE
) {
  ## detect and extract url after removing possible whitespace
  url <- trimws(url)
  pattern <- "^https?\\:\\/\\/.+"
  if (!grepl(pattern, url, perl = TRUE)) {
    return(NULL)
  }

    ## status reports
    if (verbose) {
      .throw_message("Downloading ", url, " ... ",
                     appendLF = FALSE, error = FALSE)
    }

    fail.msg <- function(w) {
      if (verbose)
        message("FAILED")
      NA
    }
    ## use internal download
    t <- tryCatch(
      expr = download.file(
        url = url,
        destfile = destfile,
        quiet = TRUE,
        mode = "wb", ## this is needed for Windows otherwise the download does not work
        cacheOK = FALSE,
        method = "auto"),
      warning = fail.msg,
      error = fail.msg)

  out_file_path <- NA
  if (!is.na(t) && t == 0) {
      if (verbose)
        message("OK ", appendLF = TRUE)
      out_file_path <- destfile
  }

  ## return file path
  out_file_path
}

#' @title Set/unset the function name for error/warning reporting
#'
#' @description
#' These utilities allow more precise error reporting from `.throw_error()`,
#' `.throw_warning()` and `.throw_message()`. They must be called exactly once
#' if the function calls one of the `.throw_*()` or `.validate_*()` functions.
#'
#' @param name [character] (**required**): the name of the function
#'
#' @details
#' The `.LuminescenceEnv` package environment stores a list (`fn_stack`)
#' that is used to store the stack of function calls currently executing.
#' The stack is at the beginning an empty list, but whenever a function is
#' called, that function add its own name to it via `.set_function_name()`.
#' Conversely, when the function returns, it clears the top of the stack
#' via `.unset_function_name()`.
#'
#' In order to maintain the stack in a consistent state, it is important that
#' each call to `.set_function_name()` is accompanied by a corresponding call
#' to `.unset_function_name()`. As the stack must be updated also when a
#' function (or any of its callees) encounters an error, the calls to
#' `.unset_function_name()` must be delegated to `on.exit(..., add = TRUE)`.
#'
#' Therefore, it is suggested to put these two lines at the very beginning
#' of each function (if any of the throwing/validating functions is used):
#'
#'   .set_function_name("name_of_the_function")
#'   on.exit(.unset_function_name(), add = TRUE)
#'
#' @noRd
.set_function_name <- function(name) {
  .LuminescenceEnv$fn_stack[length(.LuminescenceEnv$fn_stack) + 1] <- name
}

#' @rdname .set_function_name
#' @noRd
.unset_function_name <- function() {
  .LuminescenceEnv$fn_stack[length(.LuminescenceEnv$fn_stack)] <- NULL
}

#'@title Throws a Custom Tailored Error Message
#'
#'@param ... the error message to throw
#'
#'@noRd
.throw_error <- function(...) {
  top.idx <- length(.LuminescenceEnv$fn_stack)
  stop("[", .LuminescenceEnv$fn_stack[[top.idx]], "()] ", ..., call. = FALSE)
}

#'@title Throws a Custom Tailored Warning Message
#'
#'@param ... the warning message to throw
#'
#'@noRd
.throw_warning <- function(...) {
  top.idx <- length(.LuminescenceEnv$fn_stack)
  warning("[", .LuminescenceEnv$fn_stack[[top.idx]], "()] ", ..., call. = FALSE)
}

#'@title Throws a Custom Tailored Message
#'
#' @param ... the message to throw
#' @param error Whether the message should be preceded by "Error:" (`TRUE` by
#' default).
#'
#'@noRd
.throw_message <- function(..., error = TRUE) {
  top.idx <- length(.LuminescenceEnv$fn_stack)
  message("[", .LuminescenceEnv$fn_stack[[top.idx]], "()] ",
          if (error) "Error: ", ...)
}

#' @title Throws a deprecation warning
#'
#' @param old Old argument names.
#' @param new New argument names.
#' @param since Release when the deprecation happened.
#'
#' @noRd
.deprecated <- function(old, new, since) {
  .throw_warning(.collapse(old, last_sep = " and "),
                 if (length(old) > 1) " were " else " was ", "deprecated",
                 paste0(" in v", since),
                 ", use ", .collapse(new, last_sep = " and "), " instead")
}

#' @title Silence Output and Warnings during Tests
#'
#' @description
#' This is helpful so that during tests the terminal is not filled up by
#' the output from the function tested, which is often left intentionally
#' verbose to facilitate the coverage analysis.
#'
#' This was originally defined in `tests/testthat/setup.R`, but unfortunately
#' that file is not sourced by `covr::file_coverage()` (as opposed to what
#' happens with `testthat::test_file()` and `covr::package_coverage()`),
#' which makes it harder to work iteratively with it.
#'
#' @param expr [expression] an R expression (often a function, but can be
#'        any amount of code) the output of which needs to be silenced
#'
#' @examples
#' SW({
#'   template_DRAC(preset = "DRAC-example_quartz")
#' })
#'
#' @noRd
SW <- function(expr) {
  capture.output(suppressMessages(suppressWarnings(expr)))
}

#' @title Retrieve the name of the variable used as first argument
#'
#' @param idx [integer] (*with default*):
#' Index of the argument to use to retrieve the name.
#'
#' @noRd
.first_argument <- function(idx = 1) {
  sprintf("'%s'", all.vars(match.call(sys.function(sys.parent()),
                                      sys.call(sys.parent())))[idx])
}

#' @title Throw an error or a warning based on a condition
#'
#' @noRd
.error_or_warning <- function(msg, throw.error) {
  if (throw.error)
    .throw_error(msg)
  .throw_warning(msg)
}

#' @title Throw an error or a message based on a condition
#'
#' @noRd
.error_or_message <- function(msg, throw.error) {
  if (throw.error)
    .throw_error(msg)
  .throw_message(msg)
}

#' @title Validate a character argument from a list of choices
#'
#' @description
#' This is inspired by [base::match.arg], but is has a more user-friendly
#' error message as it reports the exact name of the argument that is being
#' validated. This function always requires the choices to be specified: this
#' better fits with the current state of the Luminescence package, which only
#' rarely lists all choices in the formal function arguments, so
#' [base::match.arg] would have very limited use.
#'
#' @param arg [character] (**required**): variable to validate.
#' @param choices [character] (**required**): vector of candidate values.
#' @param null.ok [logical] (*with default*): whether a `NULL` value should be
#'        considered valid (`FALSE` by default).
#' @param name [character] (*with default*): variable name to report in case
#'        of error; if specified, it's reported as is (so, it may need to be
#'        quoted by the caller); otherwise, it's inferred from the name of the
#'        variable tested and reported with quotes.
#' @param extra [character] (*with default*): additional choice to be reported
#'        after the valid choices and preceded by "or". If `null.ok = TRUE`,
#'        the text reported is automatically set to `"or NULL"`, otherwise it
#'        will take the form `"or {extra}"`.
#'
#' @return
#' The validated choice (may be `NULL` if `arg = NULL` and `null.ok = TRUE`).
#' If `arg` contains multiple elements, only the first matching one will be
#' returned.
#'
#' @noRd
.validate_args <- function(arg, choices, null.ok = FALSE,
                           name = NULL, extra = NULL) {

  if (is.null(arg) && null.ok)
    return(NULL)

  ## `arg` will have multiple values when the available choices are listed
  ## in the function's formal arguments: in that case all elements in `arg`
  ## are also in `choices` and we return the first one
  if (length(arg) > 1L && inherits(arg, "character")) {
    if (all(arg %in% choices))
      return(arg[1L])

    ## we throw an error to catch cases when the formal arguments are
    ## changed but `choices` has not been updated
    .throw_error(name %||% .first_argument(),
                 " contains multiple values but not all of them match 'choices'")
  }

  ## additional text to append after the valid choices to account for
  ## extra options that cannot be validated or for NULL
  choices.extra <- c(sQuote(choices, q = FALSE), extra)
  if (null.ok)
    choices.extra <- c(choices.extra, "NULL")

  idx.match <- tryCatch(pmatch(arg, choices, nomatch = 0L, duplicates.ok = TRUE),
                        error = function(e) 0L)
  if (all(idx.match == 0L))
    .throw_error(name %||% .first_argument(), " should be one of ",
                 .collapse(choices.extra, quote = FALSE, last_sep = " or "))
  idx <- idx.match[idx.match > 0L][1]
  choices[idx]
}

#' @title Validate an argument from a list of classes
#'
#' @param classes [character] (**required**): vector of candidate classes or
#'        types.
#' @param throw.error [logical] (*with default*): whether an error should be
#'        thrown in case of failed validation (`TRUE` by default). If `FALSE`,
#'        the function raises a warning and proceeds.
#' @param length [integer] (*with default*): if not `NULL`, validate that
#'        the length matches the one (or more) provided. This can be used only
#'        when `classes` contains only base vector types, not
#'        [Luminescence::RLum-class] objects or container types.
#' @inheritParams .validate_args
#'
#' @return
#' If `throw.error = TRUE`, the function throws an error and doesn't return
#' anything. Otherwise, it will return a boolean to indicate whether validation
#' was successful or not.
#'
#' @noRd
.validate_class <- function(arg, classes, null.ok = FALSE, throw.error = TRUE,
                            length = NULL, name = NULL, extra = NULL) {

  if (!missing(arg) && is.null(arg) && null.ok)
    return(TRUE)

  if (missing(arg) || sum(inherits(arg, classes)) == 0L ||
      !is.null(length) && !length(arg) %in% length) {
    ## additional text to append after the valid classes to account for
    ## extra options that cannot be validated but we want to report
    classes.extra <- c(sQuote(classes, q = FALSE), extra)
    if (null.ok)
      classes.extra <- c(classes.extra, "NULL")

    msg <- paste0(name %||% .first_argument(), " should be of class ",
                  .collapse(classes.extra, quote = FALSE, last_sep = " or "))
    if (!is.null(length))
      msg <- paste0(msg, " and have length ",
                    .collapse(length, quote = FALSE, last_sep = " or "))
    .error_or_warning(msg, throw.error)
    return(FALSE)
  }
  TRUE
}

#' @title Validate that a variable is not empty
#'
#' @param what [character] (**required**): the type of the variable, used
#'        only in the message reported; if not specified it's inferred from
#'        they type of the variable tested.
#' @param throw.error [logical] (*with default*): whether an error should be
#'        thrown in case of failed validation (`TRUE` by default). If `FALSE`,
#'        the function raises a warning and proceeds.
#' @inheritParams .validate_args
#'
#' @return
#' If `throw.error = TRUE`, the function throws an error and doesn't return
#' anything. Otherwise, it will return a boolean to indicate whether validation
#' was successful or not.
#'
#' @noRd
.validate_not_empty <- function(arg, what = NULL, throw.error = TRUE,
                                name = NULL) {

  ## type of the argument to report if not specified
  if (is.null(what))
    what <- class(arg)[1]

  if (NROW(arg) == 0 || NCOL(arg) == 0) {
    msg <- paste0(name %||% .first_argument(), " cannot be an empty ", what)
    .error_or_warning(msg, throw.error)
    return(FALSE)
  }
  TRUE
}

#' @title Validate the length of a variable
#'
#' @param exp.length [integer] (**required**): the expected length.
#' @param throw.error [logical] (*with default*): whether an error should be
#'        thrown in case of failed validation (`TRUE` by default). If `FALSE`,
#'        the function raises a warning and proceeds.
#' @inheritParams .validate_args
#'
#' @return
#' If `throw.error = TRUE`, the function throws an error and doesn't return
#' anything. Otherwise, it will return a boolean to indicate whether validation
#' was successful or not.
#'
#' @noRd
.validate_length <- function(arg, exp.length, throw.error = TRUE,
                            name = NULL) {
  if (length(arg) != exp.length) {
    msg <- paste0(name %||% .first_argument(), " should have length ", exp.length)
    .error_or_warning(msg, throw.error)
    return(FALSE)
  }
  TRUE
}

#' @title Validate scalar variables
#'
#' @description
#' This function is meant to validate unbounded (integer) scalar variables.
#' The `pos` and `log` arguments are not intended for direct use, they serve
#' to facilitate the implementation of `.validate_positive_scalar()` and
#' `.validate_logical_scalar()`.
#'
#' @param val [numeric] (**required**): value to validate.
#' @param int [logical] (*with default*): whether the value has to be an
#'        integer (`FALSE` by default).
#' @param pos [logical] (*with default*): whether the value has to be positive
#'        (`FALSE` by default).
#' @param nng [logical] (*with default*): whether the value has to be non-negative
#'        (`FALSE` by default).
#' @param log [logical] (*with default*): whether the value has to be logical
#'        (`FALSE` by default).
#' @inheritParams .validate_args
#'
#' @return
#' The validated value, unless the validation failed with an error thrown.
#'
#' @noRd
.validate_scalar <- function(val, int = FALSE, pos = FALSE, nng = FALSE, log = FALSE,
                             null.ok = FALSE, name = NULL, extra = NULL) {
  if (!missing(val) && is.null(val) && null.ok)
    return(NULL)
  if (missing(val) || NROW(val) != 1 || NCOL(val) != 1 || !is.null(dim(val)) || is.na(val) ||
      (!log && !is.numeric(val)) || (log && !is.logical(val)) ||
      (int && (is.infinite(val) || val != as.integer(val))) ||
      (pos && val <= 0) || (nng && val < 0)) {
    ## additional text to append for extra options that cannot be validated
    ## but we want to report
    if (null.ok)
      extra <- c(extra, "NULL")
    if (!is.null(extra))
      extra <- paste(" or", .collapse(extra, quote = FALSE, last_sep = " or "))
    .throw_error(name %||% .first_argument(), " should be a single ",
                 if (pos) "positive ", if (nng) "non-negative ",
                 if (int) "integer ", if (log) "logical ",
                 "value", extra)
  }
  val
}

#' @title Validate scalar variables expected to be positive
#'
#' @inheritParams .validate_scalar
#'
#' @noRd
.validate_positive_scalar <- function(val, int = FALSE, null.ok = FALSE,
                                      name = NULL, extra = NULL) {
  .validate_scalar(val, int = int, pos = TRUE, null.ok = null.ok,
                   name = name %||% .first_argument(), extra = extra)
}

#' @title Validate scalar variables expected to be non-negative
#'
#' @inheritParams .validate_scalar
#'
#' @noRd
.validate_nonnegative_scalar <- function(val, int = FALSE, null.ok = FALSE,
                                         name = NULL, extra = NULL) {
  .validate_scalar(val, int = int, nng = TRUE, null.ok = null.ok,
                   name = name %||% .first_argument(), extra = extra)
}

#' @title Validate logical scalar variables
#'
#' @inheritParams .validate_scalar
#'
#' @noRd
.validate_logical_scalar <- function(val, null.ok = FALSE,
                                     name = NULL, extra = NULL) {
  .validate_scalar(val, log = TRUE, null.ok = null.ok,
                   name = name %||% .first_argument(), extra = extra)
}

#' @title Validate a filename
#'
#' @param file [character], [list] (**required**):
#' Name of file to validate (can be more than one), or path to a directory to
#' scan for files, or list of file names.
#'
#' @param ext [character] (*optional*):
#' Vector of accepted extensions. If `NULL` (default), no check against the
#' file extension is performed. The check is always case-insensitive.
#'
#' @param pattern [character] (*optional*):
#' A regular expression pattern to select files from a directory. If `NULL`
#' (default), no pattern is applied, which corresponds to selecting all files
#' found. It is considered only when `file` is a path to a directory.
#'
#' @param scan.dir [logical] (*with default*):
#' Whether directories should be scanned for filed (`TRUE` by default).
#'
#' @param recursive [logical] (*with default*):
#' Whether the scan of a path for files should be done recursively (`FALSE`
#' by default).
#'
#' @inheritParams .validate_class
#'
#' @return
#' A list of file names or a single file name or `NULL`, unless the validation
#' failed with an error thrown.
#'
#' @noRd
.validate_file <- function(file, ext = NULL, pattern = NULL,
                           scan.dir = TRUE, recursive = FALSE,
                           throw.error = TRUE, verbose = TRUE) {
  .validate_class(file, c("character", "list"))
  .validate_not_empty(file)

  ## if it's a list, we only validate its elements and return it as is
  if (inherits(file, "list")) {
    lapply(file, .validate_class, classes = "character", length = 1,
           name = "All elements of 'file'")
    return(file)
  }

  ## if it's a character vector, we return it as a list
  if (length(file) > 1) {
    return(as.list(file))
  }

  ## check if it's a path: if we find multiple files (or none), we return
  if (scan.dir) {
    files_in_path <- .scan_path_for_files(file, pattern, recursive, verbose)
    if (length(files_in_path) != 1) {
      return(as.list(files_in_path))
    }

    ## this was not a path or the path contained a single file
    file <- files_in_path
  }

  ## check if it's a URL, in which case download the file locally
  ## .download_file() returns one of these:
  ## - NULL: the URL is not valid, so we continue with `file`
  ## - NA:   the download failed, so we exit
  ## - else: the download was successful, so we continue with `url_file`
  url_file <- tempfile("url_file_", fileext = paste0(".", ext[1]))
  url_file <- .download_file(file, destfile = url_file, verbose = verbose)
  if (.strict_na(url_file)) {
    return(NULL)
  }
  if (!is.null(url_file)) {
    file <- url_file
  }

  ## from now on, we are dealing with a filename
  filename <- normalizePath(file, mustWork = FALSE)
  info <- file.info(filename)
  if (is.na(info$size) || isTRUE(info$isdir)) {
    .error_or_message(paste0("File '", filename, "' does not exist"),
                      throw.error)
    return(NULL)
  }

  if (!is.null(ext)) {
    file.ext <- tools::file_ext(filename)
    if (!tolower(file.ext) %in% tolower(ext)) {
      msg <- paste0 ("File extension '", file.ext, "' is not supported, only ",
                     .collapse(ext, last_sep = " and "),
                     ifelse(length(ext) > 1, " are", " is"), " valid")
      .error_or_message(msg, throw.error)
      return(NULL)
    }
  }

  if (info$size == 0) {
    .error_or_message(paste0("File '", filename, "' is a zero-byte file"),
                      throw.error)
    return(NULL)
  }

  filename
}

#' Find the files matching a given pattern in a path
#'
#' @param path [character] (**required**):
#' The path to search.
#'
#' @inheritParams .validate_file
#'
#' @return
#' A vector of file names, or `NULL` if no files are found.
#'
#' @noRd
.scan_path_for_files <- function(path, pattern, recursive, verbose) {
  ## return immediately if this is not a path to a directory
  if (!dir.exists(path) || length(dir(path)) == 0) {
    return(path)
  }

  if (verbose) {
    .throw_message("Directory detected, looking for ",
                   if (is.null(pattern)) "any" else .collapse(pattern),
                   " files ...", error = FALSE)
  }

  files <- dir(path, pattern = pattern, recursive = recursive,
               full.names = TRUE, include.dirs = FALSE)
  if (length(files) == 0 && verbose) {
    .throw_message("No files matching the given pattern found in directory")
    files <- NULL
  }

  files
}

#' @title Validate the originator of an RLum object
#'
#' @param object [Luminescence::RLum-class] (**required**): object whose
#' originator should be
#'        checked.
#' @inheritParams .validate_args
#'
#' @return
#' The validated value, unless the validation failed with an error thrown.
#'
#' @noRd
.validate_originator <- function(object, choices, name = NULL) {
  if (.check_originator(object, choices))
    return(object@originator)
  .throw_error(paste0(name %||% .first_argument(), " has an unsupported originator ",
                      "(expected ", .collapse(choices), ", but found '",
                      object@originator, "')"))
}

#' @title Check that the originator matches a set of choices
#'
#' @inheritParams .validate_originator
#'
#' @return
#' A logical value.
#'
#' @noRd
.check_originator <- function(object, choices) {
  .hasSlot(object, "originator") && object@originator %in% choices
}

#' Validate a vector used for signal and background integrals
#'
#' @param integral [vector] [integer] (**required**):
#' the vector to validate.
#'
#' @param int [logical] (*with default*):
#' whether the integral values should be integers (`TRUE` by default).
#'
#' @param min [integer] (*with default*):
#' the minimum value for the integral. By default the minimum value is 1 if
#' `int = TRUE` (channels), and 0 otherwise (measurements).
#'
#' @param max [integer] (*with default*):
#' the maximum value for the integral.
#'
#' @param list.ok [logical] (*with default*):
#' whether the argument should be considered valid also as a list (`FALSE` by
#' default).
#'
#' @param na.ok [logical] (*with default*):
#' whether the argument should be considered valid also when it's exactly `NA`
#' (`FALSE` by default).
#'
#' @inheritParams .validate_args
#'
#' @return
#' A vector with negative elements and `NA` values removed, sorted and without
#' duplicates, unless the validation failed with an error thrown. If `max.value`
#' is not `NULL`, then the integral is capped to the value specified.
#'
#' @noRd
.validate_integral <- function(integral, int = TRUE, min = ifelse(int, 1, 0), max = Inf,
                               null.ok = FALSE, list.ok = FALSE, na.ok = FALSE,
                               name = NULL) {
  if (null.ok && is.null(integral) || na.ok && .strict_na(integral))
    return(integral)
  name <- name %||% .first_argument()
  .validate_class(integral, c("integer", "numeric", if (list.ok) "list"),
                  null.ok = null.ok, extra = if (na.ok) "NA" else NULL, name = name)
  if (list.ok && inherits(integral, "list")) {
    return(lapply(integral, .validate_integral,
                  name = paste("All elements of", name)))
  }
  orig.length <- length(integral)
  if (min > max)
    .throw_error(name, " is expected to be at least ", min,
                 ", but the maximum allowed is ", max)
  integral <- integral[!is.na(integral) & between(integral, min, max)]
  if (length(integral) == 0)
    .throw_error(name, " is of length 0 after removing values smaller than ",
                 min, if (!is.infinite(max)) paste(" and greater than", max))
  else if (length(integral) != orig.length)
    .throw_warning(name, " out of bounds, reset to be between ", min(integral),
                   " and ", max(integral))
  if (int && any(integral != as.integer(integral)))
    .throw_error(name, " should be a vector of integers")
  if (length(integral) == 2 && diff(integral) > 1 && int)
    .throw_warning(name, " was defined as c(", .format_range(integral, sep = ", "),
                   ") but in general we would expect it to be defined as ",
                   .format_range(integral), ", please check your input")
  sort(unique(integral))
}

#' Convert integral from time/temperature to channel numbers
#'
#' @param x.range [numeric] (**required**)
#' The measurement values.
#'
#' @param integral [numeric] (**required**):
#' The integral range expressed as time/temperature.
#'
#' @param unit [character] (**required**):
#' The measurement unit expected ("time" or "temperature").
#'
#' @inheritParams .validate_integral
#'
#' @return
#' The best matching channel numbers corresponding to the given measurement
#' range, unless the validation failed with an error thrown. If all integral
#' values are outside the object range, a warning is thrown and the lower or
#' upper bound (whichever is closest) of the valid values is returned.
#'
#' @noRd
.convert_to_channels <- function(x.range, integral, unit,
                                 null.ok = FALSE, na.ok = FALSE,
                                 list.ok = FALSE, name = NULL) {
  if (null.ok && is.null(integral) || na.ok && .strict_na(integral))
    return(integral)

  name <- name %||% .first_argument(idx = 2)
  .validate_class(integral, c("integer", "numeric", if (list.ok) "list"),
                  null.ok = null.ok, extra = if (na.ok) "NA" else NULL,
                  name = name)
  if (list.ok && inherits(integral, "list")) {
    return(lapply(integral, .convert_to_channels, x.range = x.range, unit = unit,
                  null.ok = null.ok, na.ok = na.ok, list.ok = FALSE,
                  name = paste("All elements of", name)))
  }

  integral <- integral[!is.na(integral)]
  if (min(integral) > max(x.range) || max(integral) < min(x.range))
    .throw_warning("Conversion of ", name, " from ", unit,
                   " to channels failed: expected values in ",
                   .format_range(x.range))

  integral.min <- which.min(abs(min(integral) - x.range))
  integral.max <- which.min(abs(max(integral) - x.range))
  integral.min:integral.max
}

#' Check that a suggested package is installed
#'
#' Report a message with installation instructions if a suggested package
#' is not available.
#'
#' @param pkg [character] (**required**): name of the package to check.
#' @param reason [character] (*with default*): subject of the sentence,
#'        helpful to clarify why the package is being required.
#' @param throw.error [logical] (*with default*): whether an error should be
#'        thrown in case of failed validation (`TRUE` by default). If `FALSE`,
#'        the function raises a warning and proceeds.
#'
#' @return
#' If `throw.error = TRUE`, the function throws an error and doesn't return
#' anything. Otherwise, it will return a boolean to indicate whether validation
#' was successful or not.
#'
#' @noRd
.require_suggested_package <- function(pkg, reason = "This function",
                                       throw.error = TRUE) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    msg <- paste("%s requires the '%s' package: to install it, run",
                 "`install.packages('%s')` in your R console")
    msg <- sprintf(msg, reason, pkg, pkg)
    .error_or_warning(msg, throw.error)
    return(FALSE)
  }
  TRUE
}

#' Create a list of objects repeated a given number of times
#'
#' @param x [vector] (**required**) An object to be repeated into a list.
#' @param length [integer] (**required**) Length of the list to be produced.
#'
#' @return
#' A list with the object repeated the specified number of times.
#'
#' @noRd
.listify <- function(x, length) {
  if (!inherits(x, "list"))
    x <- list(x)
  rep(x, length = length)
}

#' Check that a given object is exactly `NA`
#'
#' @param x (**required): The object to check.
#'
#' @return
#' Whether the object is exactly `NA`.
#'
#' @noRd
.strict_na <- function(x) {
  if (length(x) != 1 || is.recursive(x) || is.array(x))
    return(FALSE)
  is.na(x)
}

#' Comma-separated string concatenation
#'
#' Collapse the elements of a vector into a comma-separated string, with
#' the option of quoting each element.
#'
#' @param x [vector] (**required**): A vector of elements to be collapsed into
#'        a comma-separated string.
#' @param quote [logical] (*with default*): Whether each element should be
#'        surrounded by single quotes (`TRUE` by default).
#' @param last_sep [character] (*with default*): Separator to use before the
#'        last element (`", "` by default). This should contain the desired
#'        spacing around it (e.g. " or ").
#'
#' @return
#' A comma-separated string where each element of the original vector is
#' optionally surrounded by single quotes.
#'
#' @noRd
.collapse <- function(x, quote = TRUE, last_sep = ", ") {
  x <- if (quote) sQuote(x, FALSE) else as.character(x)
  if (length(x) < 2)
    return(toString(x))
  head <- head(x, -2)
  tail <- paste0(tail(x, 2)[1], last_sep, tail(x, 1))
  toString(c(head, tail))
}

#' Compute and format the range of a vector
#'
#' @param vals [numeric] (**required**): A numeric vector.
#' @param sep [character] (*with default*): Separator character.
#' @param nsmall [integer] (*with default*): Minimum number of digits to the
#'        right of the decimal point.
#'
#' @return
#' A string with the range of the vector formatted as `min:max`.
#'
#' @noRd
.format_range <- function(vals, sep = ":", nsmall = 0L) {
  rng <- tryCatch(range(vals, na.rm = TRUE),
                  warning = function(w) c(NA, NA))
  paste(format(rng, nsmall = nsmall, trim = TRUE), collapse = sep)
}

#' Shorten a filename
#'
#' Shorten a filename to the given width by cutting out characters from the
#' middle, leaving the most significant parts.
#'
#' @param filename [character] (**required**) A file name; can be a vector
#' @param max.width [integer] (*with default*) The maximum width available
#'
#' @return
#' A filename not longer than the maximum available width.
#'
#' @noRd
.shorten_filename <- function(filename, max.width = 70) {
  ## optimised for speed with ChatGPT, 2025
  ## get length
  name.len <- nchar(filename, keepNA = FALSE)

  ## logical index of names to shorten
  to_shorten <- name.len > max.width

  ## get out if nothing needs shortening
  if (!any(to_shorten))
    return(filename)

  # calculate parts
  dots <- "..."  # \u2026
  max.width <- max.width - nchar(dots)
  part1_last <- floor(max.width / 2)
  part2_first <- floor(name.len[to_shorten] - max.width / 2) + 1

  # shorten what needs to be shortened
  filename[to_shorten] <- paste0(
    substring(filename[to_shorten], 1, part1_last),
    dots,
    substring(filename[to_shorten], part2_first)
  )

  ## return
  filename
}

#'@title Affine Transformation of Values
#'
#'@description Affine (linear) transformation of values; which we use
#'a couple of times within functions
#'
#'@param x [numeric] (*required*): values to be transformed
#'
#'@param range_old [numeric] (*required*): original scale limits
#'
#'@param range_new [numeric] (*required*): new scale limits
#'
#'@returns rescaled values values
#'
#'@examples
#'
#' x <- stats::rnorm(100)
#' y_dens <- stats::density(x)
#' .rescale(
#'  x = y_dens$y,
#'  range_old = c(min(y_dens$y), max(y_dens$y)),
#'  range_new = c(0,20))
#'
#'@noRd
.rescale <- function(x, range_old, range_new) {
  range_new[1] +
    (x - range_old[1]) * (range_new[2] - range_new[1]) /
    (range_old[2] - range_old[1])
}

## path to the XSYG file on github
github.url <- file.path("https://raw.githubusercontent.com/R-Lum",
                        "rxylib/master/inst/extdata/TLSpectrum.xsyg")
xsyg.file <- Luminescence:::.download_file(github.url, tempfile("test_read_XSYG2R"),
                            verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_XSYG2R(data.frame()),
               "'file' should be of class 'character' or 'list'")
  expect_error(read_XSYG2R(character(0)),
               "'file' cannot be an empty character")
  expect_error(read_XSYG2R(list()),
               "'file' cannot be an empty list")
  expect_message(expect_null(read_XSYG2R("_error_file_")),
                "File '.*_error_file_' does not exist") # windows CI needs the regexp
  expect_message(expect_null(read_XSYG2R("/Test", fastForward = TRUE)),
                 "File '.*/Test' does not exist") # windows CI needs the regexp
  expect_message(expect_null(read_XSYG2R(test_path("_data/xsyg-tests/XSYG_broken.xsyg"), fastForward = TRUE)),
                 "Error: XML file not readable, nothing imported")
  expect_error(read_XSYG2R(xsyg.file, n_records = "error"),
               "'n_records' should be a single positive integer value or NULL")

  SW({
  expect_message(expect_null(read_XSYG2R(test_path("_data/bin-tests/"))),
                 "No files matching the given pattern found in directory")
  })
})

test_that("test import of XSYG files", {
  testthat::skip_on_cran()

  ## download from github
  expect_type(read_XSYG2R(github.url, import = FALSE, verbose = FALSE),
              "list")

  ## local file
  expect_type(read_XSYG2R(xsyg.file, verbose = FALSE),
              "list")
  expect_s3_class(read_XSYG2R(xsyg.file, fastForward = TRUE,
                              import = FALSE, verbose = FALSE),
                  "data.frame")
  expect_silent(read_XSYG2R(xsyg.file, verbose = FALSE,
                            recalculate.TL.curves = FALSE))
  expect_silent(read_XSYG2R(xsyg.file, verbose = FALSE, pattern = "xsyg",
                            recalculate.TL.curves = FALSE))
  expect_type(read_XSYG2R(xsyg.file, fastForward = FALSE,
                          verbose = FALSE),
              "list")
  results <- expect_type(read_XSYG2R(xsyg.file, fastForward = TRUE,
                                     verbose = FALSE),
                         type = "list")
  expect_type(results[[1]]@info$file, type = "character")
  expect_equal(results[[1]]@records[[1]]@info$recordName,
               "unknown")
  expect_equal(results[[1]]@records[[1]]@info$sequenceName,
               "20150601_BT706_TL_Spectra_FS_SLS365")
  expect_length(results[[1]]@records[[1]]@info, 26)
  expect_output(print(results))

  ## check n_records argument
  expect_type(read_XSYG2R(xsyg.file, verbose = FALSE, n_records = 1),
              "list")
  expect_type(read_XSYG2R(xsyg.file, verbose = FALSE, n_records = 10),
              "list")

  ## list input
  expect_type(read_XSYG2R(list(xsyg.file), fastForward = TRUE,
                          verbose = FALSE),
              "list")
  expect_s3_class(read_XSYG2R(list(xsyg.file), fastForward = TRUE,
                              import = FALSE, verbose = FALSE),
                  "data.frame")
  expect_type(read_XSYG2R(list(xsyg.file), fastForward = FALSE,
                          verbose = FALSE),
              "list")

  ## check also internal files
  SW({
  expect_type(read_XSYG2R(system.file("extdata/XSYG_file.xsyg",
                                      package = "Luminescence"),
                          fastForward = TRUE, import = TRUE,
                          verbose = TRUE),
              "list")

  ## recursive traversal
  expect_length(read_XSYG2R(test_path("_data"),
                            fastForward = TRUE, verbose = FALSE),
                7)
  })

  ## more tests for different TL curve calculation cases
  ## CASE2
  expect_s4_class(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_file_TL_CASE2.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE)[[1]],
    "RLum.Analysis")
  ## CASE3
  expect_s4_class(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_file_TL_CASE3.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE)[[1]],
    "RLum.Analysis")

  ## check case for no sequences in the file
  expect_warning(t <- read_XSYG2R(test_path("_data/xsyg-tests/XSYG_noSequence.xsyg"),
                                  verbose = FALSE),
                 "1 incomplete sequence(s) removed",
                 fixed = TRUE)
  expect_equal(t, list())

  ## check case for no record Type
  t <- expect_s4_class(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_noRecordType.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE)[[1]],
    "RLum.Analysis")
  expect_equal(t@records[[1]]@recordType, "not_set (UVVIS)")

  ## check case for OSL record but IRSL stimulator
  t <- expect_s4_class(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_corr_IRSL.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE)[[1]],
    "RLum.Analysis")
  expect_equal(t@records[[1]]@recordType, "IRSL (NA)")

  ## check case for spectrum
  t <- expect_s4_class(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_spectra.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE)[[1]],
    "RLum.Analysis")
  ## check case for spectrum ... equal heating element values
  t <- expect_s4_class(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_spectra_equal_TL.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE)[[1]],
    "RLum.Analysis")
  ## check case for spectrum ... duplicated values
  expect_warning(
    read_XSYG2R(test_path("_data/xsyg-tests/XSYG_spectra_duplicated_TL.xsyg"), fastForward = TRUE, txtProgressBar = FALSE, verbose = FALSE),
    regexp = "Temperature values are found to be duplicated and increased by 1 K.")

  ## test automated linearity correction
  file <- system.file("extdata/XSYG_file.xsyg", package = "Luminescence")
  t <- expect_s4_class(
    read_XSYG2R(
      file = file,
      fastForward = TRUE,
      auto_linearity_correction = TRUE,
      verbose = FALSE
    )[[1]],
    class = "RLum.Analysis")
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1260
  SW({
  expect_message(expect_null(do.call(read_XSYG2R, list("a test"))),
                 "File '.*a test' does not exist") # windows CI needs the regexp
  })
})

## load data
data(ExampleData.DeValues, envir = environment())
ExampleData.DeValues <- ExampleData.DeValues$CA1

CAM <- calc_CentralDose(ExampleData.DeValues,
                        plot = FALSE, verbose = FALSE)
FMM <- calc_FiniteMixture(ExampleData.DeValues, sigmab = 0.01,
                          n.components = 2:3, plot = FALSE, verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_AbanicoPlot(data = "error"),
               "All elements of 'data' should be of class 'data.frame'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues[, 1, drop = FALSE]),
               "Data set 1 has fewer than 2 columns: data without errors",
               fixed = TRUE)

  expect_message(expect_null(plot_AbanicoPlot(list())),
                 "Error: 'data' is empty, nothing plotted")
  expect_message(expect_null(plot_AbanicoPlot(ExampleData.DeValues[0, ])),
                 "Error: 'data' is empty, nothing plotted")

  expect_warning(expect_message(
      expect_null(plot_AbanicoPlot(ExampleData.DeValues[1, ])),
      "Error: After removing invalid entries, nothing is plotted"),
      "Data set 1 empty or consisting of only 1 row, removed")

  expect_error(plot_AbanicoPlot(ExampleData.DeValues, na.rm = "error"),
               "'na.rm' should be a single logical value")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, plot = FALSE),
               "'plot.ratio' should be a single positive value")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, xlab = "x"),
               "'xlab' must have length 2")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, z.0 = "error"),
               "'z.0' should be one of 'mean', 'mean.weighted', 'median' or")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, dispersion = "error"),
               "'dispersion' should be one of 'qr', 'sd', '2sd' or a percentile")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, dispersion = "p5"),
               "'dispersion' should be one of 'qr', 'sd', '2sd' or a percentile")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, dispersion = "p5a"),
               "'dispersion' should be one of 'qr', 'sd', '2sd' or a percentile")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, dispersion = "p500"),
               "'dispersion' should be one of 'qr', 'sd', '2sd' or a percentile")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, dispersion = c("p95", "p86")),
               "'dispersion' should have length 1")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, summary = 5),
               "'summary' should be of class 'character'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, summary.pos = 5),
               "'summary.pos' should have length 2")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, summary.pos = list()),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, summary.pos = "error"),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, legend = 5),
               "'legend' should be of class 'character'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, legend.pos = 5),
               "'legend.pos' should have length 2")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, frame = NULL),
               "'frame' should be one of '0', '1', '2' or '3'")

  expect_error(plot_AbanicoPlot(ExampleData.DeValues, xlim = NA),
               "'xlim' should be of class 'numeric'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, ylim = NA),
               "'ylim' should be of class 'numeric'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, zlim = NA),
               "'zlim' should be of class 'numeric'")
  expect_error(plot_AbanicoPlot(ExampleData.DeValues, zlim = c(-10, 10)),
               "'zlim' should only contain positive values when 'log.z = TRUE'")

  ## zero-error values
  data.zeros <- ExampleData.DeValues
  data.zeros[2, 2] <- 0
  expect_warning(plot_AbanicoPlot(data.zeros, grid.col = c(1, 2)),
                 "Values with zero errors cannot be displayed and were removed")
  data.zeros[, 2] <- 0
  expect_error(plot_AbanicoPlot(data.zeros),
               "Data set contains only values with zero errors")

  expect_warning(plot_AbanicoPlot(ExampleData.DeValues, xlim = c(2, 12)),
                 "Lower x-axis limit was 2, reset to zero")
})


test_that("Test examples from the example page", {
  testthat::skip_on_cran()

  ## plot the example data straightforward
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues))

  ## now with linear z-scale
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                                 xlim = NULL, ylim = NULL, zlim = NULL,
                                 log.z = FALSE))

  ## now with adjusted scale limits
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                                 zlim = c(10, 200),
                                 xlim = c(0, 20)))

  ## now with rug to indicate individual values in KDE part
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   rug = TRUE))

  ## now with a smaller bandwidth for the KDE plot
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   bw = 0.04))

  ## now with a histogram instead of the KDE plot
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   hist = TRUE,
                   kde = FALSE))

  ## now with a KDE plot and histogram with manual number of bins
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   hist = TRUE,
                   breaks = 20))

  ## now with a KDE plot and a dot plot
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   dots = TRUE))

  ## now with user-defined plot ratio
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   plot.ratio = 0.5))

  ## now with user-defined central value
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   z.0 = 70))

  ## now with median as central value
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   z.0 = "median"))

  ## now with the 17-83 percentile range as definition of scatter
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   z.0 = "median",
                   dispersion = "p17"))

  ## now with user-defined green line for minimum age model
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   line = list(CAM),
                   line.col = "darkgreen",
                   line.label = "CAM"))

  ## now create plot with legend, colour, different points and smaller scale
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   legend = "Sample 1",
                   col = "tomato4",
                   bar.col = "peachpuff",
                   pch = "R",
                   line = CAM,
                   cex = 0.8))

  ## now without 2-sigma bar, polygon, grid lines and central value line
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   bar.col = FALSE,
                   polygon.col = FALSE,
                   grid.col = FALSE,
                   y.axis = FALSE,
                   lwd = 0))

  ## now with direct display of De errors, without 2-sigma bar
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   bar.col = FALSE,
                   ylab = "",
                   y.axis = FALSE,
                   error.bars = TRUE))

  ## now with user-defined axes labels
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   xlab = c("Data error (%)",
                            "Data precision"),
                   ylab = "Scatter",
                   zlab = "Equivalent dose [Gy]"))

  ## now with minimum, maximum and median value indicated
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   stats = c("min", "max", "median")))

  ## now with another statistical summary
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   summary = c("mean.weighted", "median"),
                   summary.pos = "topleft"))

  ## now a plot with two 2-sigma bars for one data set
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   bar = c(30, 100)))

  ## now the data set is split into sub-groups, one is manipulated
  data.1 <- ExampleData.DeValues[1:30,]
  data.2 <- ExampleData.DeValues[31:62,] * 1.3
  data.3 <- list(data.1, data.2)

  ## now the two data sets are plotted in one plot
  expect_silent(plot_AbanicoPlot(data = data.3))

  ## now with some graphical modification
  expect_silent(plot_AbanicoPlot(data = data.3,
                   z.0 = "median",
                   col = c("steelblue4", "orange4"),
                   bar.col = c("steelblue3", "orange3"),
                   polygon.col = c("steelblue1", "orange1"),
                   pch = c(2, 6),
                   angle = c(30, 50),
                   summary = c("n", "in.2s", "median")))

  ## create Abanico plot with predefined layout definition
  layout <- expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                   layout = "journal"))

  ## trigger a few test cases related to layout
  layout$abanico$colour$centrality <- 1:2
  expect_silent(plot_AbanicoPlot(data = ExampleData.DeValues,
                                 rotate = TRUE,
                                 layout = "journal"))

  ## now with predefined layout definition and further modifications
  expect_silent(plot_AbanicoPlot(data = data.3,
                   z.0 = "median",
                   layout = "journal",
                   col = c("steelblue4", "orange4"),
                   bar.col = adjustcolor(c("steelblue3", "orange3"),
                                         alpha.f = 0.5),
                   polygon.col = c("steelblue3", "orange3")))

  ## for further information on layout definitions see documentation
  ## of function get_Layout()

  ## now with manually added plot content
  ## create empty plot with numeric output
  expect_type(plot_AbanicoPlot(data = ExampleData.DeValues,
                         pch = NA,
                         output = TRUE), "list")

  ## interactive mode
  expect_silent(plot_AbanicoPlot(ExampleData.DeValues, interactive = TRUE))
})

test_that("more coverage", {
  testthat::skip_on_cran()

  ## weights
  expect_silent(plot_AbanicoPlot(ExampleData.DeValues, weights = FALSE,
                                 boxplot = TRUE, frame = 2))
  suppressWarnings( # additional warning on weights not summing to 1
  expect_warning(plot_AbanicoPlot(ExampleData.DeValues, weights = TRUE,
                                  rotate = TRUE, line = 1,
                                  dispersion = "sd", log.z = FALSE,
                                  grid.col = c(1, 2)),
                 "Selecting bandwidth *not* using 'weights'",
                 fixed = TRUE)
  )

  ## negative values
  data.neg <- ExampleData.DeValues
  data.neg[1, 1] <- -1
  expect_silent(plot_AbanicoPlot(data.neg, z.0 = "mean", dispersion = "sd",
                                 boxplot = TRUE, frame = 3, zlim = c(200, 400),
                                 main = "Title", sub = "Subtitle"))
  data.neg[2, 1] <- -120
  expect_silent(plot_AbanicoPlot(data.neg, log.z = FALSE))

  ## missing values
  data.na <- ExampleData.DeValues
  data.na[1, 2] <- NA
  expect_message(plot_AbanicoPlot(data.na, rotate = TRUE, boxplot = TRUE,
                                  hist = TRUE, error.bars = TRUE, dots = TRUE,
                                  rug = TRUE, y.axis = FALSE, stats = "min",
                                  legend = "legend", legend.pos = "bottomleft",
                                  summary.pos = "bottomleft", log.z = FALSE,
                                  xlab = c("x1", "x2", "x3"), lty = 2,
                                  dispersion = "2sd",
                                  at = seq(20, 120, nrow(data.na) - 1)),
                 "Data set 1: 1 NA value excluded",
                 fixed = TRUE)
  expect_message(plot_AbanicoPlot(data.na, y.axis = TRUE,
                                  yaxt = "y",
                                  summary.pos = "bottomright",
                                  dispersion = "2sd"),
                 "Data set 1: 1 NA value excluded",
                 fixed = TRUE)

  ## further edge tests ... check for wrong bw parameter
  expect_warning(
    object = plot_AbanicoPlot(data = ExampleData.DeValues, bw = "tests"),
    regexp = "Option for 'bw' not valid, reset to 'nrd0'")

  ## negative values
  df <-  ExampleData.DeValues
  df[,1] <- -df[,1]
  expect_message(
    object = plot_AbanicoPlot(data = df),
    regexp = "Small standardised estimate scatter, toggle off y.axis?")

  ## test boundaries
  expect_warning(
    object = plot_AbanicoPlot(
    data = data.frame(x = c(0,1), y = c(0.1, 0.1))),
    "Found zero values in x-column of dataset 1, 'log.z' set to FALSE")

  ## handling of negative values; before it produced wrong plots
 expect_silent(plot_AbanicoPlot(data = data.frame(
    x = c(-1,10),
    y = c(0.1,3)
  ), log.z = TRUE, summary = c("mean", "sd.abs")))

 ## handling of negative values; but with zlim
 expect_silent(plot_AbanicoPlot(data = data.frame(
   x = c(-1,10),
   y = c(0.1,3),
   zlim = c(2,10)
 ), log.z = TRUE, summary = c("mean", "sd.abs")))

 ## test panel plots
 par.defaults <- .par_defaults()
 par(mfrow = c(4, 4))
 expect_silent(plot_AbanicoPlot(data = data.frame(
   x = c(-1,10),
   y = c(0.1,3)
 ), log.z = TRUE, summary = c("mean", "sd.abs")))
 par(par.defaults)

 ## test centrality from layout
 layout <- get_Layout("default")

 ## trigger a few test cases related to layout
 layout$abanico$colour$centrality <- 1:2
 expect_silent(
   plot_AbanicoPlot(data = list(ExampleData.DeValues, ExampleData.DeValues),
                  layout = layout))
})

test_that("Test graphical snapshot", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
    vdiffr::expect_doppelganger("default",
                                plot_AbanicoPlot(ExampleData.DeValues))
    vdiffr::expect_doppelganger("dispersion error bars",
                                plot_AbanicoPlot(ExampleData.DeValues,
                                                 dispersion = "sd",
                                                 error.bars = TRUE,
                                                 hist = TRUE,
                                                 boxplot = TRUE,
                                                 summary.pos = "topleft",
                                                 summary = c("n", "se.rel", "kurtosis")))
    vdiffr::expect_doppelganger("rotated",
                                plot_AbanicoPlot(ExampleData.DeValues,
                                                 rotate = TRUE, rug = TRUE,
                                                 hist = TRUE,
                                                 boxplot = TRUE,
                                                 grid.col = "grey",
                                                 summary.pos = "bottomleft",
                                                 summary = c("mean", "in.2s", "skewness")))
    vdiffr::expect_doppelganger("dots cex",
                                plot_AbanicoPlot(ExampleData.DeValues,
                                                 dots = TRUE, cex = 2,
                                                 boxplot = TRUE,
                                                 grid.col = "grey80",
                                                 legend = "primary data",
                                                 summary.pos = "left",
                                                 summary.method = "weighted",
                                                 summary = c("sd.abs", "se.abs",
                                                             "median")))
    vdiffr::expect_doppelganger("plot ratio",
                                plot_AbanicoPlot(ExampleData.DeValues,
                                                 plot.ratio = 0.1))

    data.list <- list(ExampleData.DeValues[1:30,],
                      ExampleData.DeValues[31:62,] * 1.3)
    vdiffr::expect_doppelganger("line frame legend",
                                plot_AbanicoPlot(data = data.list,
                                                 line = 75.7,
                                                 line.lty = 3,
                                                 line.label = "CAM",
                                                 frame = 3,
                                                 legend = c("data 1", "data 2"),
                                                 cex = 1.2,
                                                 log.z = FALSE,
                                                 z.0 = "median",
                                                 col = c("steelblue4", "orange4"),
                                                 bar.col = c("steelblue3", "orange3"),
                                                 polygon.col = c("steelblue1", "orange1"),
                                                 pch = c(2, 6),
                                                 angle = c(30, 50),
                                                 summary = c("n", "in.2s", "median")))
    vdiffr::expect_doppelganger("line frame legend rotated",
                                plot_AbanicoPlot(data = data.list,
                                                 rotate = TRUE,
                                                 line = 75.7,
                                                 line.lty = 3,
                                                 line.label = "CAM",
                                                 frame = 2,
                                                 legend = c("data 1", "data 2"),
                                                 cex = 1.2,
                                                 log.z = FALSE,
                                                 z.0 = "mean",
                                                 col = c("steelblue4", "orange4"),
                                                 bar.col = c("steelblue3", "orange3"),
                                                 polygon.col = c("steelblue1", "orange1"),
                                                 pch = c(2, 6),
                                                 angle = c(30, 50),
                                                 summary = c("n", "in.2s", "median")))

    vdiffr::expect_doppelganger("CAM",
                                plot_AbanicoPlot(data = CAM,
                                                 line.col = "darkseagreen",
                                                 summary.pos = "bottomleft"))
    vdiffr::expect_doppelganger("CAM cex",
                                plot_AbanicoPlot(data = CAM, cex = 2))

    vdiffr::expect_doppelganger("FMM",
                                plot_AbanicoPlot(data = FMM, rotate = TRUE))
    vdiffr::expect_doppelganger("FMM cex",
                                plot_AbanicoPlot(data = FMM, rotate = TRUE, cex = 2))
  })
})

##Full check
test_that("methods_DRAC", {
  testthat::skip_on_cran()

  input <- template_DRAC(notification = FALSE)

  ## print
  expect_message(
      expect_output(print(input, blueprint = TRUE)),
      "You can copy all lines above to your script and fill in the data")
  expect_output(print(input, blueprint = FALSE))

  ## as.data.frame
  expect_s3_class(as.data.frame(input), "data.frame")
  expect_s3_class(as.data.frame(input), "DRAC.data.frame")

  SW({

  ## [[<-
  expect_warning({
    input <- template_DRAC()
    input[[1]] <- 1i
  }, regexp = "'value' should be of class 'integer', 'character', 'numeric' or")
  expect_warning({
    input <- template_DRAC()
    input[[1]] <- c(1, 2)
  }, regexp = "Project ID should have length 1")
  expect_error({
    input <- template_DRAC()
    input[[4]] <- "error"
  }, regexp = "Conversion factors: Invalid option, valid options are:")
  expect_message({
    input <- template_DRAC()
    input[[5]] <- "1"
  }, regexp = "found character, expected numeric")

   expect_warning({
    input <- template_DRAC()
    input[[5]] <- "X"
    Luminescence:::.warningCatcher(input[[5]] <- "abc")
  }, regexp = "Cannot coerce 'abc' to a numeric value, input must be numeric or")

  expect_warning({
    input <- template_DRAC(nrow = 2)
    input[[5]] <- c("X", 1)
    Luminescence:::.warningCatcher(input[[5]] <- c("X", "abc"))
  }, regexp = "Cannot coerce 'abc' to a numeric value, input must be numeric or")

  input <- template_DRAC(nrow = 2)
  input[[5]] <- c("X", 1)
  expect_silent({
    input[[5]] <- c("X", 2)
  })

  input <- template_DRAC(notification = FALSE)
  expect_message({
    input[[5]] <- 1L
  }, regexp = "found integer, expected numeric")

  expect_warning({
    input[[5]] <- "abc"
  })

  expect_warning({
    input <- template_DRAC()
    input[[32]] <- 200L
    Luminescence:::.warningCatcher(input[[32]] <- "abc")
  }, regexp = "found character, expected integer -> cannot coerce, set NAs")
  expect_message({
    input <- template_DRAC()
    input[[32]] <- 200L
    Luminescence:::.warningCatcher(input[[32]] <- "100")
  }, regexp = "found character, expected integer -> coercing to integer")

   expect_warning({
    input <- template_DRAC()
    input[[13]] <- 1
  }, regexp = "Input must be of class 'character'")

  ## [<-
  expect_identical(
    object = template_DRAC(notification = FALSE),
    expected = {
      input <- template_DRAC(notification = FALSE)
      input[1] <- NA_character_
      input
    })

  ## $<-
  expect_identical(
    object = template_DRAC(),
    expected = {
      input <- template_DRAC()
      input$`Project ID` <- NA_character_
      input
    })
  })
})

## load data
data(ExampleData.CW_OSL_Curve, envir = environment())

## curve
temp <-
  set_RLum(
      class = "RLum.Data.Curve",
      recordType = "OSL",
      data = as.matrix(ExampleData.CW_OSL_Curve)
  )

##create RLum.Analysis object
temp_analysis <- set_RLum("RLum.Analysis", records = list(temp, temp))

## spectra object
spectra <- set_RLum("RLum.Data.Spectrum",
                    data = matrix(1:10, ncol = 2, dimnames = list(1:5, c(0.2, 1.3))))

## array
image <- set_RLum("RLum.Data.Image", data = array(1:12, dim = c(2,3,2)))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(normalise_RLum(temp, norm = "error"),
               "'norm' should be one of 'max', 'min', 'first', 'last', 'huot' or 'intensity'")
  expect_error(normalise_RLum(temp, norm = c(1,1)), "'norm' should be of class 'logical', 'character' or 'numeric' and have length 1")
  expect_error(normalise_RLum(temp, norm = -3),
               "'norm' should be a single positive value or one of 'max', 'min'")
  expect_error(normalise_RLum(temp, norm = NA),
               "'norm' should be a single logical value or one of 'max', 'min'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##standard tests
  expect_s4_class(temp, class = "RLum.Data.Curve")
  expect_snapshot_RLum(normalise_RLum(temp))

  ## spectrum
  expect_s4_class(spectra, class = "RLum.Data.Spectrum")
  expect_snapshot_RLum(normalise_RLum(spectra))

  ## test also an array, which indeed works differently
  expect_s4_class(image, class = "RLum.Data.Image")
  t <- expect_snapshot_RLum(normalise_RLum(image, global = TRUE))
  expect_snapshot_RLum(normalise_RLum(image, global = FALSE))
  expect_s4_class(normalise_RLum(image, global = TRUE), "RLum.Data.Image")
  expect_s4_class(normalise_RLum(image, norm = "first", global = TRUE), "RLum.Data.Image")
  expect_s4_class(normalise_RLum(image, norm = "last", global = TRUE), "RLum.Data.Image")
  expect_s4_class(normalise_RLum(image, norm = "first", global = FALSE), "RLum.Data.Image")
  expect_s4_class(normalise_RLum(image, norm = "last", global = FALSE), "RLum.Data.Image")

  ## tests as for the parameters are already covered in the for the
  ## internal function .normalise_curve() and should not be repeated
  ## here

  ##test on a list
  ##RLum list
  ## standard normalisation
  t <- expect_type(normalise_RLum(list(temp, temp)), "list")
  expect_equal(as.numeric(t[[1]]@data[1,2]), 1)

  ## list with non-RLum object
  t <- expect_type(normalise_RLum(list(temp, iris)), "list")
  expect_equal(as.numeric(t[[1]]@data[1,2]), 1)
  expect_equal(t[[2]], iris)

  ## normalise to "last"
  t <- expect_type(normalise_RLum(list(temp, temp), norm = "last"), "list")
  expect_equal(as.numeric(t[[1]]@data[1,2]), 317.5, tolerance = 0.01)

  ##test on an RLum.Analysis-object and a list of such objects
  t <- expect_s4_class(normalise_RLum(temp_analysis), "RLum.Analysis")
  expect_equal(as.numeric(t@records[[2]]@data[1,2]), 1)

  ## check the modification for an argument
  t <- expect_s4_class(normalise_RLum(temp_analysis, norm = "last"), "RLum.Analysis")
  expect_equal(as.numeric(t@records[[2]]@data[1,2]), 317.5, tolerance = 0.01)

  ## check the list of such objects
  t <- expect_type(normalise_RLum(list(temp_analysis, temp_analysis)), "list")
  expect_equal(as.numeric(t[[1]]@records[[2]]@data[1,2]), 1)

  ## check Image objects within
  t <- expect_s4_class(normalise_RLum(set_RLum("RLum.Analysis", records = list(image))), "RLum.Analysis")
  expect_equal(t@records[[1]]@data[2,3,2], 1)
  t <- expect_s4_class(normalise_RLum(set_RLum("RLum.Analysis", records = list(image)), global = FALSE), "RLum.Analysis")
  expect_equal(t@records[[1]]@data[2,3,1], t@records[[1]]@data[2,3,2])
  t <- expect_type(
    normalise_RLum(
      list(
        set_RLum("RLum.Analysis", records = list(image)),
        set_RLum("RLum.Analysis", records = list(image))), global = FALSE), "list")
  expect_equal(t[[1]]@records[[1]]@data[2,3,2], t[[1]]@records[[1]]@data[2,3,1])
  t <- expect_type(
    normalise_RLum(
      list(
        set_RLum("RLum.Analysis", records = list(image)),
        set_RLum("RLum.Analysis", records = list(image))), global = TRUE), "list")
  expect_equal(t[[1]]@records[[1]]@data[2,3,1], 0.5, tolerance = 0.001)

})

test_that("snapshot tests", {
 testthat::skip_on_cran()

 small <-set_RLum(class = "RLum.Data.Curve", recordType = "OSL",
                  data = as.matrix(ExampleData.CW_OSL_Curve[1:150, ]))
 expect_snapshot_RLum(normalise_RLum(small, norm = TRUE))
 expect_snapshot_RLum(normalise_RLum(small, norm = "min"))
 expect_snapshot_RLum(normalise_RLum(small, norm = "max"))
 expect_snapshot_RLum(normalise_RLum(small, norm = "first"))
 expect_snapshot_RLum(normalise_RLum(small, norm = "last"))
 expect_snapshot_RLum(normalise_RLum(small, norm = 2.2))
 expect_snapshot_RLum(normalise_RLum(small, norm = "huot"))
 expect_snapshot_RLum(normalise_RLum(small, norm = "intensity"))
 expect_snapshot_RLum(normalise_RLum(spectra, norm = "intensity"))
})

test_that("Test internals", {
  testthat::skip_on_cran()

  ## since below we are testing internal functions that use `.throw_error()`
  ## and `.throw_warning()`, we need to treat this block as a function
  ## definition and set the function name
  .set_function_name("test")
  on.exit(.unset_function_name(), add = TRUE)

  # .expand_parameters() ------------------------------------------------------
  ##create empty function ... reminder
  ##this is an internal function, the first object is always discarded, it
  ##might be a list of RLum.Analysis objects is might be super large
  f <- function(object, a, b = 1, c = list(), d = NULL) {
    Luminescence:::.expand_parameters(len = 3)
  }

  ##test some functions
  ##missing arguments must be identified
  expect_error(f(), "Argument 'a' missing, with no default")

  ##check whether the objects are properly recycled
  expect_type(f(object, a = 1), "list")
  expect_length(f(object, a = 1, c = list(a = 1, b = 2, c = 3))$c, 3)
  expect_length(f(object, a = (1), c = list(a = 1, b = 2, c = 3))$c, 3)
  expect_equal(f(object, a = (1 + 10), c = list(a = 1, b = 2, c = 3))$a[[1]], 11)

  # .calc_HPDI() ------------------------------------------------------------
  set.seed(1234)
  test <- expect_type(Luminescence:::.calc_HPDI(rnorm(100), plot = TRUE), "double")
  expect_equal(round(sum(test),2), 0.20, tolerance = 1)

  ##create a case where the value cannot be calculated
  expect_type(.calc_HPDI(rlnorm(n = 100, meanlog = 10, sdlog = 100)), type = "logical")

  ## .warningCatcher() ------------------------------------------------------
  expect_warning(Luminescence:::.warningCatcher(for(i in 1:5) {
                                                  warning("message 1")
                                                  if (i %% 2 == 0)
                                                    warning("message 2")
                                                  if (i %% 3 == 0)
                                                    warning("message 3")
                                                }),
                 paste(c("(1) message 1 [5 times]",
                         "(2) message 2 [2 times]",
                         "(3) message 3"), collapse = "\n"),
                 fixed = TRUE)
  expect_warning(Luminescence:::.warningCatcher(warning("single message")),
                 "single message")

  # .smoothing ----------------------------------------------------------------------------------
  expect_silent(Luminescence:::.smoothing(runif(100), k = 5, method = "median"))
  expect_silent(.smoothing(runif(200), method = "median"))
  expect_silent(.smoothing(runif(100), k = 4, method = "mean"))
  expect_silent(.smoothing(runif(100), k = 4, method = "median"))
  expect_equal(.smoothing(c(1, 1, 2, 50, 0, 2, 1, 2, 0, 1, 50),
                          method = "Carter"),
               c(1, 1, 2, 1, 0, 2, 1, 2, 0, 1, NA))
  expect_equal(.smoothing(c(1, 1, 2, 50, 0, 2, 1, 2, 0, 1, 50),
                          method = "Carter", fill = 0),
               c(1, 1, 2, 1, 0, 2, 1, 2, 0, 1, 0))
  expect_equal(.smoothing(c(1, 1, 2, 50, 0, 2, 1, 2, 0, 1, 50) * 10,
                          method = "Carter", p_acceptance = 1e-30),
               c(10, 10, 20, 17, 17, 20, 10, 20, 13, 10, NA))
  expect_error(.smoothing(c(1, 1, 2, 50, 0, 2, 1, 2, 0, 1, 50),
                          method = "Carter", p_acceptance = 0.5),
               "'p_acceptance' rejects all counts, set it to a smaller value")
  expect_error(.smoothing(runif(100), k = integer()),
               "'k' should be a single positive integer value or NULL")
  expect_error(.smoothing(runif(100), fill = numeric()),
               "'fill' should be of class 'numeric' or NA and have length 1")
  expect_error(.smoothing(runif(100), fill = matrix()),
               "'fill' should be of class 'numeric' or NA and have length 1")
  expect_error(.smoothing(runif(100), method = "error"),
               "'method' should be one of 'mean', 'median' or")
  expect_error(.smoothing(runif(100), align = "error"),
               "'align' should be one of 'right', 'center' or 'left'")
  expect_error(.smoothing(runif(100), p_acceptance = "error"),
               "'p_acceptance' should be a single positive value")

  ## .weighted.median() -----------------------------------------------------
  expect_equal(.weighted.median(1:10, w = rep(1, 10)),
               median(1:10))
  expect_equal(.weighted.median(1:5, w = c(0.15, 0.1, 0.2, 0.3, 0.25)),
               4)
  expect_equal(.weighted.median(c(1:5, NA), w = c(0.15, 0.1, 0.2, 0.3, 0.25, 1)),
               NA_real_)
  expect_equal(.weighted.median(c(1:5, NA), w = c(0.15, 0.1, 0.2, 0.3, 0.25, 1),
                                na.rm = TRUE),
               4)

  ## .normalise_curve() -----------------------------------------------------
  ## vector
  data <- runif(100)
  expect_equal(data, .normalise_curve(data, FALSE))
  expect_equal(.normalise_curve(data, TRUE), .normalise_curve(data, "max"))
  expect_silent(.normalise_curve(data, "min"))
  expect_silent(.normalise_curve(data, "first"))
  expect_silent(.normalise_curve(data, "last"))
  expect_silent(.normalise_curve(data, "huot"))
  expect_silent(.normalise_curve(data, "intensity"))
  expect_silent(.normalise_curve(data, 2.2))

  ## check for matrix
  m <- matrix(runif(10), ncol = 2)
  expect_equal(m, .normalise_curve(m, FALSE))
  expect_equal(.normalise_curve(m, TRUE), .normalise_curve(m, "max"))
  expect_silent(.normalise_curve(m, "min"))
  expect_silent(.normalise_curve(m, "first"))
  expect_true(is.matrix(.normalise_curve(m, "last")))
  expect_silent(.normalise_curve(m, "huot"))
  expect_silent(.normalise_curve(m, "intensity"))
  expect_silent(.normalise_curve(m, 2.2))

  data[100] <- 0
  expect_warning(.normalise_curve(data, "last"),
                 "Curve normalisation produced Inf/NaN values, values replaced")
  expect_warning(.normalise_curve(data, "min"),
                 "Curve normalisation produced Inf/NaN values, values replaced")

  data[99] <- NA_real_
  ## this will produce warnings, but still normalise, because
  ## we use na.rm internally for max(), min() etc.
  expect_equal(
    suppressWarnings(.normalise_curve(data, TRUE)),
    suppressWarnings(.normalise_curve(data, "max")))
  expect_warning(.normalise_curve(data, "last"),
                 "Curve normalisation produced Inf/NaN values, values replaced")
  expect_warning(.normalise_curve(data, "huot"),
                 "Curve normalisation produced Inf/NaN values, values replaced")
  expect_warning(.normalise_curve(data, "min"),
                 "Curve normalisation produced Inf/NaN values, values replaced")
  expect_warning(.normalise_curve(data, 2),
                 "Curve normalisation produced Inf/NaN values, values replaced")

  # fancy_scientific ()--------------------------------------------------------------------------
  plot(seq(1e10, 1e20, length.out = 10),1:10, xaxt = "n")
  expect_silent(axis(1, at = axTicks(1),labels = Luminescence:::fancy_scientific(axTicks(1))))

  # .add_fancy_log_axis() -----------------------------------------------------
  y <- c(0.1, 0.001, 0.0001)
  plot(1:length(y), y, yaxt = "n", log = "y")
  expect_silent(Luminescence:::.add_fancy_log_axis(side = 2, las = 1))
  expect_null(.add_fancy_log_axis(side = 1, las = 1))

  ## .text_with_bg() --------------------------------------------------------
  expect_silent(.text_with_bg(x = 2, y = 0.1, label = "pos = NULL", pos = NULL))
  expect_silent(.text_with_bg(x = 2, y = 0.01, label = "pos = 1", pos = 1))
  expect_silent(.text_with_bg(x = 2, y = 0.01, label = "pos = 2", pos = 2))
  expect_silent(.text_with_bg(x = 2, y = 0.01, label = "pos = 3", pos = 3))
  expect_silent(.text_with_bg(x = 2, y = 0.01, label = "pos = 4", pos = 4))

  # .get_keyword_coordinates() ----------------------------------------------
  xlim <- c(0, 5)
  ylim <- c(2, 10)
  coords <- .get_keyword_coordinates(xlim = xlim, ylim = ylim)
  expect_type(coords, "list")
  expect_named(coords, c("pos", "adj"))
  expect_length(coords, 2)
  expect_equal(coords$pos,
               c(0, 10))
  expect_equal(.get_keyword_coordinates(pos = c(1, 2), xlim, ylim)$pos,
               c(1, 2))
  expect_equal(.get_keyword_coordinates(pos = "topleft", xlim, ylim)$pos,
               c(0, 10))
  expect_equal(.get_keyword_coordinates(pos = "top", xlim, ylim)$pos,
               c(2.5, 10))
  expect_equal(.get_keyword_coordinates(pos = "topright", xlim, ylim)$pos,
               c(5, 10))
  expect_equal(.get_keyword_coordinates(pos = "left", xlim, ylim)$pos,
               c(0, 6))
  expect_equal(.get_keyword_coordinates(pos = "center", xlim, ylim)$pos,
               c(2.5, 6))
  expect_equal(.get_keyword_coordinates(pos = "right", xlim, ylim)$pos,
               c(5, 6))
  expect_equal(.get_keyword_coordinates(pos = "bottomleft", xlim, ylim)$pos,
               c(0, 2))
  expect_equal(.get_keyword_coordinates(pos = "bottom", xlim, ylim)$pos,
               c(2.5, 2))
  expect_equal(.get_keyword_coordinates(pos = "bottomright", xlim, ylim)$pos,
               c(5, 2))

  # .create_StatisticalSummaryText() ------------------------------------------------------------
  stats <- calc_Statistics(data.frame(1:10,1:10))
  expect_equal(Luminescence:::.create_StatisticalSummaryText(),
               "")
  expect_equal(.create_StatisticalSummaryText(stats, keywords = ""),
               "")
  expect_equal(.create_StatisticalSummaryText(stats, keywords = "nonexisting"),
               "")
  expect_equal(.create_StatisticalSummaryText(stats,
                                              keywords = "mean"),
               "mean = 5.5")
  expect_equal(.create_StatisticalSummaryText(stats,
                                              keywords = "unweighted$mean"),
               "mean = 5.5")
  expect_equal(.create_StatisticalSummaryText(stats,
                                              keywords = "weighted$mean"),
               "weighted$mean = 1.89")


  # .unlist_RLum() ------------------------------------------------------------------------------
  expect_length(Luminescence:::.unlist_RLum(list(a = list(b = list(c = list(d = 1, e = 2))))), 2)

  # .rm_nonRLum() -----------------------------------------------------------
  expect_type(
    Luminescence:::.rm_nonRLum(c(list(set_RLum("RLum.Analysis"), set_RLum("RLum.Analysis")), 2)),
    "list")
  expect_type(
    Luminescence:::.rm_nonRLum(
      c(list(set_RLum("RLum.Analysis"), set_RLum("RLum.Analysis")), 2), class = "RLum.Analysis"),
    "list")

  ## regression test for issue 907
  expect_length(Luminescence:::.rm_nonRLum(list(matrix(iris)), "RLum.Analysis"),
                0)

  # .rm_NULL_elements() -----------------------------------------------------------
  expect_type(.rm_NULL_elements(list("a", NULL)),
    "list")
  t <- expect_type(.rm_NULL_elements(list(NULL, NULL)),
              "list")
  expect_length(t, 0)

  # .rm_unnamed_elements() --------------------------------------------------
  expect_equal(.rm_unnamed_elements(list(a = "a", b = "b")),
               list(a = "a", b = "b"))
  expect_equal(.rm_unnamed_elements(list(a = "a", b = NULL, 3, 4, c = 5)),
               list(a = "a", b = NULL, c = 5))
  expect_null(.rm_unnamed_elements(list("a", NULL)))
  expect_null(.rm_unnamed_elements(list()))

  # .matrix_binning() ---------------------------------------------------------------------------
  m <- matrix(data = c(rep(1:20, each = 20)), ncol = 20, nrow = 20)
  rownames(m) <- 1:nrow(m)
  colnames(m) <- 1:ncol(m)

    ##crash the function
    expect_error(Luminescence:::.matrix_binning("none matrix"),
                 "'m' should be of class 'matrix'")

    ##test operation modes and arguments
    expect_type(Luminescence:::.matrix_binning(m, bin_size = 4, bin_col = FALSE), "integer")
    expect_type(Luminescence:::.matrix_binning(m, bin_size = 4, bin_col = TRUE), "integer")

    ##test row / column renaming options
    expect_type(Luminescence:::.matrix_binning(m, bin_size = 2, bin_col = FALSE, names = "groups"),
              "integer")
    expect_type(Luminescence:::.matrix_binning(m, bin_size = 2, bin_col = FALSE, names = "mean"),
             "integer")
    expect_type(Luminescence:::.matrix_binning(m, bin_size = 2, bin_col = FALSE, names = "sum"),
             "integer")
    expect_type(Luminescence:::.matrix_binning(m, bin_size = 2, bin_col = FALSE, names = c("test1", "test2")),
              "integer")

    ##clean-up
    rm(m)

  # .download_file() --------------------------------------------------------

  ## returns just NULL (no URL detected)
  expect_null(.download_file(url = "_url"))
  expect_null(.download_file(url = "https://"))
  expect_null(.download_file(url = "_https://github.com"))

  ## attempts download but fails
  url.404 <- "https://raw.githubusercontent.com/R-Lum/rxylib/master/inst/extg"
  expect_message(
          expect_message(expect_equal(.download_file(url = url.404), NA),
                         "Downloading"),
      "FAILED")

  ## attempts download and succeeds
  url.ok <- "https://raw.githubusercontent.com/R-Lum/rxylib/master/codecov.yml"
  destfile <- tempfile()
  SW({
  expect_message(expect_equal(.download_file(url = url.ok, destfile),
                              destfile),
                 "OK")
  expect_message(.download_file(url = paste("  ", url.ok, "\n")),
                 "OK")
  })

  ## .throw_error() ---------------------------------------------------------
  fun.int <- function() {
    .set_function_name("fun.int")
    on.exit(.unset_function_name(), add = TRUE)
    .throw_error("Error message")
  }
  fun.ext <- function() fun.int()
  fun.docall <- function() do.call(fun.ext, args = list())
  fun.docall_do <- function() fun.docall()
  expect_error(fun.int(),
               "[fun.int()] Error message", fixed = TRUE)
  expect_error(fun.ext(),
               "[fun.int()] Error message", fixed = TRUE)
  expect_error(fun.docall(),
               "[fun.int()] Error message", fixed = TRUE)
  expect_error(fun.docall_do(),
               "[fun.int()] Error message", fixed = TRUE)

  ## .throw_warning() -------------------------------------------------------
  fun.int <- function() {
    .set_function_name("fun.int")
    on.exit(.unset_function_name(), add = TRUE)
    .throw_warning("Warning message")
  }
  fun.ext <- function() fun.int()
  fun.docall <- function() do.call(fun.ext, args = list())
  fun.docall_do <- function() fun.docall()
  expect_warning(fun.int(),
                 "[fun.int()] Warning message", fixed = TRUE)
  expect_warning(fun.ext(),
                 "[fun.int()] Warning message", fixed = TRUE)
  expect_warning(fun.docall(),
                 "[fun.int()] Warning message", fixed = TRUE)
  expect_warning(fun.docall_do(),
                 "[fun.int()] Warning message", fixed = TRUE)

  ## .throw_message() -------------------------------------------------------
  fun.int <- function(error = TRUE) {
    .set_function_name("fun.int")
    on.exit(.unset_function_name(), add = TRUE)
    .throw_message("Simple message", error = error)
  }
  fun.ext <- function(error = TRUE) fun.int(error)
  fun.docall <- function(error = TRUE) do.call(fun.ext, args = list(error = error))
  fun.docall_do <- function(error = TRUE) fun.docall(error)
  expect_message(fun.int(),
                 "[fun.int()] Error: Simple message", fixed = TRUE)
  expect_message(fun.ext(),
                 "[fun.int()] Error: Simple message", fixed = TRUE)
  expect_message(fun.docall(),
                 "[fun.int()] Error: Simple message", fixed = TRUE)
  expect_message(fun.docall_do(),
                 "[fun.int()] Error: Simple message", fixed = TRUE)
  expect_message(fun.int(error = FALSE),
                 "[fun.int()] Simple message", fixed = TRUE)
  expect_message(fun.ext(error = FALSE),
                 "[fun.int()] Simple message", fixed = TRUE)
  expect_message(fun.docall(error = FALSE),
                 "[fun.int()] Simple message", fixed = TRUE)
  expect_message(fun.docall_do(error = FALSE),
                 "[fun.int()] Simple message", fixed = TRUE)

  ## .deprecated() ----------------------------------------------------------
  expect_warning(.deprecated("old", "new", "1.0"),
                 "'old' was deprecated in v1.0, use 'new' instead")
  expect_warning(.deprecated(c("old1", "old2"), c("new1", "new2"), "1.0"),
                 "'old1' and 'old2' were deprecated in v1.0, use 'new1' and 'new2' instead")

  ## SW() ------------------------------------------------------------------
  expect_silent(SW(cat("silenced message")))
  expect_silent(SW(message("silenced message")))
  expect_silent(SW(warning("silenced message")))
  expect_silent(SW(.throw_warning("silenced message")))
  expect_error(SW(stop("error message")),
               "error message")
  expect_error(SW(.throw_error("error message")),
               "error message")

  ## .validate_args() -------------------------------------------------------
  fun1 <- function(arg) {
    .validate_args(arg, c("val1", "val2", "val3"), null.ok = TRUE)
  }
  expect_silent(fun1(NULL))
  expect_equal(fun1(arg = "val1"), "val1")
  expect_equal(fun1(arg = c("val1", "val2")), "val1")
  expect_equal(fun1(arg = c("val3", "val2")), "val3")
  expect_equal(.validate_args(matrix(letters), c("reciprocal", "square")),
               "reciprocal")
  expect_error(fun1(arg = c("error", "val1")),
               "[test()] 'arg' contains multiple values but not all of them",
               fixed = TRUE)
  expect_error(fun1(arg = "error"),
               "[test()] 'arg' should be one of 'val1', 'val2', 'val3' or NULL",
               fixed = TRUE)
  expect_error(fun1(set_RLum("RLum.Data.Curve")),
               "'arg' should be one of 'val1', 'val2', 'val3' or NULL")
  expect_error(fun1(set_RLum("RLum.Data.Curve", data = matrix(1:10, ncol = 2))),
               "'arg' should be one of 'val1', 'val2', 'val3' or NULL")

  fun2 <- function(arg = c("val1", "val2", "val3")) {
    .validate_args(arg, c("val1", "val2", "val3"), name = "'other_name'")
  }
  expect_equal(fun2(), "val1")
  expect_error(fun2(arg = NULL),
               "[test()] 'other_name' should be one of 'val1', 'val2' or 'val3'",
               fixed = TRUE)
  expect_error(fun2(arg = "error"),
               "[test()] 'other_name' should be one of 'val1', 'val2' or 'val3'",
               fixed = TRUE)

  fun3 <- function(arg) {
    .validate_args(arg, c("val1", "val2"),
                   extra = "'other.val'", null.ok = FALSE)
  }
  expect_error(fun3(arg = "error"),
               "[test()] 'arg' should be one of 'val1', 'val2' or 'other.val'",
               fixed = TRUE)

  fun4 <- function(arg) {
    .validate_args(arg, c("val1", "val2"),
                   extra = "'other.val'", null.ok = TRUE)
  }
  expect_error(fun4(arg = "error"),
               "[test()] 'arg' should be one of 'val1', 'val2', 'other.val' or NULL",
               fixed = TRUE)

  fun.err <- function(arg) {
    .validate_args(arg)
  }
  expect_error(fun.err("val1"),
               "is missing, with no default")

  ## .validate_class() ------------------------------------------------------
  fun1 <- function(arg) {
    .validate_class(arg, "data.frame")
  }
  fun2 <- function(arg) {
    .validate_class(arg, "data.frame", null.ok = TRUE)
  }
  fun3 <- function(arg) {
    .validate_class(arg, "data.frame", throw.error = FALSE)
  }
  expect_true(fun1(iris))
  expect_true(fun2(NULL))
  expect_true(.validate_class(iris, c("data.frame", "integer")))
  expect_true(.validate_class(iris, c("data.frame", "integer"),
                              throw.error = FALSE))
  expect_true(.validate_class(1:3, "integer", length = c(2, 3)))
  expect_warning(expect_false(.validate_class(arg <- NULL, "data.frame",
                                   throw.error = FALSE)),
      "'arg' should be of class 'data.frame'")
  expect_error(fun1(),
               "'arg' should be of class 'data.frame'")
  expect_error(fun2(),
               "'arg' should be of class 'data.frame' or NULL")
  expect_error(fun1(NULL),
               "'arg' should be of class 'data.frame'")
  expect_error(.validate_class(test <- 1:5),
               "is missing, with no default")
  expect_error(.validate_class(test <- 1:5, "data.frame"),
               "'test' should be of class 'data.frame'")
  expect_error(.validate_class(test <- 1:5, "data.frame", length = 2),
               "'test' should be of class 'data.frame' and have length 2")
  expect_error(.validate_class(test <- 1:5, "data.frame", length = 1:2),
               "'test' should be of class 'data.frame' and have length 1 or 2")
  expect_error(.validate_class(test <- 1:5, c("list", "data.frame", "numeric")),
               "'test' should be of class 'list', 'data.frame' or 'numeric'")
  expect_error(.validate_class(test <- 1:5, c("list", "data.frame")),
               "'test' should be of class 'list' or 'data.frame'")
  expect_error(.validate_class(test <- 1:5, c("list", "data.frame"),
                               extra = "another type"),
               "'test' should be of class 'list', 'data.frame' or another")
  expect_error(.validate_class(test <- 1:5, c("list", "data.frame"),
                               name = "'other_name'"),
               "'other_name' should be of class 'list' or 'data.frame'")
  expect_warning(fun3(),
                 "'arg' should be of class 'data.frame'")

  ## .validate_not_empty() --------------------------------------------------
  expect_true(.validate_not_empty(letters, "vector"))

  expect_error(.validate_not_empty(test <- c(), "vector"),
               "'test' cannot be an empty vector")
  expect_error(.validate_not_empty(test <- list()),
               "'test' cannot be an empty list")
  expect_error(.validate_not_empty(test <- numeric(0)),
               "'test' cannot be an empty numeric")
  expect_error(.validate_not_empty(test <- data.frame()),
               "'test' cannot be an empty data.frame")
  expect_error(.validate_not_empty(iris[0, ]),
               "'iris' cannot be an empty data.frame")
  expect_error(.validate_not_empty(iris[, 0]),
               "'iris' cannot be an empty data.frame")
  expect_error(.validate_not_empty(test <- matrix(NA, 0, 5)),
               "'test' cannot be an empty matrix")
  expect_error(.validate_not_empty(test <- matrix(NA, 5, 0)),
               "'test' cannot be an empty matrix")
  expect_error(.validate_not_empty(test <- set_RLum("RLum.Analysis")),
               "'test' cannot be an empty RLum.Analysis")
  expect_error(.validate_not_empty(list(), "list", name = "'other_name'"),
               "'other_name' cannot be an empty list")
  expect_warning(expect_false(.validate_not_empty(test <- list(), "list",
                                                  throw.error = FALSE)),
                 "'test' cannot be an empty list")

  ## .validate_length() -----------------------------------------------------
  expect_true(.validate_length(letters, 26))
  expect_error(.validate_length(letters),
               "is missing, with no default")
  expect_error(.validate_length(letters, 25),
               "'letters' should have length 25")
  expect_error(.validate_length(letters, 25, name = "'other_name'"),
               "'other_name' should have length 25")
  expect_warning(expect_false(.validate_length(letters, 25, throw.error = FALSE)),
                 "'letters' should have length 25")

  ## .validate_scalar() -----------------------------------------------------
  expect_equal(.validate_scalar(1.3),
               1.3)
  expect_equal(.validate_scalar(-2, int = TRUE),
               -2)
  expect_null(.validate_scalar(NULL, int = TRUE, null.ok = TRUE))

  expect_error(.validate_scalar(int = TRUE),
               "'NA' should be a single integer value")
  expect_error(.validate_scalar(test <- "a"),
               "'test' should be a single value")
  expect_error(.validate_scalar(test <- NULL),
               "'test' should be a single value")
  expect_error(.validate_scalar(iris),
               "'iris' should be a single value")
  expect_error(.validate_scalar(iris[, 1, drop = FALSE]),
               "'iris' should be a single value")
  expect_error(.validate_scalar(iris[1, 0, drop = FALSE]),
               "'iris' should be a single value")
  expect_error(.validate_scalar(iris, null.ok = TRUE),
               "'iris' should be a single value or NULL")
  expect_error(.validate_scalar(iris, extra = "something else"),
               "'iris' should be a single value or something else")
  expect_error(.validate_scalar(iris, extra = "something else", null.ok = TRUE),
               "'iris' should be a single value or something else or NULL")
  expect_error(.validate_scalar(array(1, c(1, 1, 0))),
               "'NA' should be a single value")
  expect_error(.validate_scalar(-1:2, name = "'var'"),
               "'var' should be a single value")
  expect_error(.validate_scalar(Inf, int = TRUE, name = "'var'"),
               "'var' should be a single integer value")
  expect_error(.validate_scalar(1.5, int = TRUE, name = "'var'"),
               "'var' should be a single integer value")
  expect_error(.validate_scalar(NA, int = TRUE, name = "The variable"),
               "The variable should be a single integer value")
  expect_error(.validate_scalar(-1:2, pos = TRUE, name = "'var'"),
               "'var' should be a single positive value")
  expect_error(.validate_scalar(-1:2, int = TRUE, pos = TRUE, name = "'var'"),
               "'var' should be a single positive integer value")

  ## .validate_positive_scalar() --------------------------------------------
  expect_equal(.validate_positive_scalar(1.3),
               1.3)
  expect_equal(.validate_positive_scalar(2, int = TRUE),
               2)
  expect_null(.validate_positive_scalar(NULL, int = TRUE, null.ok = TRUE))

  expect_error(.validate_positive_scalar(int = TRUE),
               "'NA' should be a single positive integer value")
  expect_error(.validate_positive_scalar(test <- "a"),
               "'test' should be a single positive value")
  expect_error(.validate_positive_scalar(test <- NULL),
               "'test' should be a single positive value")
  expect_error(.validate_positive_scalar(iris),
               "'iris' should be a single positive value")
  expect_error(.validate_positive_scalar(iris, null.ok = TRUE),
               "'iris' should be a single positive value or NULL")
  expect_error(.validate_positive_scalar(iris, extra = "something else"),
               "'iris' should be a single positive value or something else")
  expect_error(.validate_positive_scalar(iris, extra = "something else", null.ok = TRUE),
               "'iris' should be a single positive value or something else or NULL")
  expect_error(.validate_positive_scalar(1:2, name = "'var'"),
               "'var' should be a single positive value")
  expect_error(.validate_positive_scalar(0, name = "'var'"),
               "'var' should be a single positive value")
  expect_error(.validate_positive_scalar(-1, name = "'var'"),
               "'var' should be a single positive value")
  expect_error(.validate_positive_scalar(Inf, int = TRUE, name = "'var'"),
               "'var' should be a single positive integer value")
  expect_error(.validate_positive_scalar(1.5, int = TRUE, name = "'var'"),
               "'var' should be a single positive integer value")
  expect_error(.validate_positive_scalar(NA, int = TRUE, name = "The variable"),
               "The variable should be a single positive integer value")

  ## .validate_nonnegative_scalar() -----------------------------------------
  expect_equal(.validate_nonnegative_scalar(0),
               0)
  expect_equal(.validate_nonnegative_scalar(2, int = TRUE),
               2)
  expect_null(.validate_nonnegative_scalar(NULL, int = TRUE, null.ok = TRUE))

  expect_error(.validate_nonnegative_scalar(int = TRUE),
               "'NA' should be a single non-negative integer value")
  expect_error(.validate_nonnegative_scalar(test <- "a"),
               "'test' should be a single non-negative value")
  expect_error(.validate_nonnegative_scalar(test <- NULL),
               "'test' should be a single non-negative value")
  expect_error(.validate_nonnegative_scalar(iris),
               "'iris' should be a single non-negative value")
  expect_error(.validate_nonnegative_scalar(iris, null.ok = TRUE),
               "'iris' should be a single non-negative value or NULL")
  expect_error(.validate_nonnegative_scalar(iris, extra = "something else"),
               "'iris' should be a single non-negative value or something else")
  expect_error(.validate_nonnegative_scalar(iris, extra = "something else", null.ok = TRUE),
               "'iris' should be a single non-negative value or something else or NULL")
  expect_error(.validate_nonnegative_scalar(1:2, name = "'var'"),
               "'var' should be a single non-negative value")
  expect_error(.validate_nonnegative_scalar(-1, name = "'var'"),
               "'var' should be a single non-negative value")
  expect_error(.validate_nonnegative_scalar(Inf, int = TRUE, name = "'var'"),
               "'var' should be a single non-negative integer value")
  expect_error(.validate_nonnegative_scalar(1.5, int = TRUE, name = "'var'"),
               "'var' should be a single non-negative integer value")
  expect_error(.validate_nonnegative_scalar(NA, int = TRUE, name = "The variable"),
               "The variable should be a single non-negative integer value")

  ## .validate_logical_scalar() ---------------------------------------------
  expect_equal(.validate_logical_scalar(TRUE),
               TRUE)
  expect_equal(.validate_logical_scalar(FALSE),
               FALSE)
  expect_null(.validate_logical_scalar(NULL, null.ok = TRUE))

  expect_error(.validate_logical_scalar(),
               "'NA' should be a single logical value")
  expect_error(.validate_logical_scalar(test <- "a"),
               "'test' should be a single logical value")
  expect_error(.validate_logical_scalar(test <- NULL),
               "'test' should be a single logical value")
  expect_error(.validate_logical_scalar(iris),
               "'iris' should be a single logical value")
  expect_error(.validate_logical_scalar(iris, null.ok = TRUE),
               "'iris' should be a single logical value or NULL")
  expect_error(.validate_logical_scalar(iris, extra = "something else"),
               "'iris' should be a single logical value or something else")
  expect_error(.validate_logical_scalar(iris, extra = "something else", null.ok = TRUE),
               "'iris' should be a single logical value or something else or NULL")
  expect_error(.validate_logical_scalar(c(TRUE, FALSE), name = "'var'"),
               "'var' should be a single logical value")
  expect_error(.validate_logical_scalar(0, name = "'var'"),
               "'var' should be a single logical value")
  expect_error(.validate_logical_scalar(NA, name = "The variable"),
               "The variable should be a single logical value")

  ## .validate_originator() ----------------------------------------------------
  expect_error(.validate_originator(set_RLum("RLum.Analysis"), "orig"),
               "'NA' has an unsupported originator (expected 'orig', but found",
               fixed = TRUE)
  expect_equal(.validate_originator(set_RLum("RLum.Analysis", originator = "orig"),
                                    "orig"),
               "orig")

  ## .check_originator() ----------------------------------------------------
  expect_true(.check_originator(set_RLum("RLum.Analysis", originator = "orig"),
                                "orig"))
  expect_false(.check_originator(set_RLum("RLum.Analysis", originator = "orig"),
                                 c("orig1", "orig2")))
  expect_false(.check_originator(set_RLum("RLum.Analysis"), "orig"))
  expect_false(.check_originator(NULL, "orig"))
  expect_false(.check_originator(iris, "orig"))

  ## .validate_integral() ----------------------------------------------------
  expect_null(.validate_integral(NULL, null.ok = TRUE))
  expect_equal(.validate_integral(integral <- 5:1),
               1:5)
  expect_equal(.validate_integral(integral <- NA, na.ok = TRUE),
               NA)
  expect_equal(.validate_integral(c(0, 1.5), int = FALSE),
               c(0, 1.5))
  expect_error(.validate_integral(integral <- 1:5 + 0.1, int = FALSE),
               integral)
  expect_warning(expect_equal(.validate_integral(integral <- c(5:1, -3:3)),
                              1:5),
                 "'integral' out of bounds, reset to be between 1 and 5")
  expect_warning(expect_equal(.validate_integral(integral <- 1:100,
                                                 min = 5, max = 50),
                              5:50),
                 "'integral' out of bounds, reset to be between 5 and 50")
  expect_error(.validate_integral(integral <- "error"),
               "'integral' should be of class 'integer' or 'numeric'")
  expect_error(.validate_integral(integral <- list(NA), na.ok = TRUE),
               "'integral' should be of class 'integer', 'numeric' or NA")
  expect_error(.validate_integral(integral <- NA, na.ok = FALSE),
               "'integral' should be of class 'integer' or 'numeric'")
  expect_error(.validate_integral(integral <- -9:0),
               "'integral' is of length 0 after removing values smaller than 1$")
  expect_error(.validate_integral(integral <- 1:10, min = 50, max = 100),
               "after removing values smaller than 50 and greater than 100")
  expect_error(.validate_integral(integral <- 1:10, min = 150, max = 100),
               "is expected to be at least 150, but the maximum allowed is 100")
  expect_error(.validate_integral(integral <- 1:5 + 0.1),
               "'integral' should be a vector of integers")
  expect_warning(expect_error(.validate_integral(integral <- c(0, 1.5)),
                              "'integral' should be a vector of integers"),
                 "'integral' out of bounds, reset to be between 1.5 and 1.5")
  expect_error(.validate_integral(list.integral <- list(1:4)),
               "'list.integral' should be of class 'integer' or 'numeric'")
  expect_error(.validate_integral(list.integral <- list("error"), list.ok = TRUE),
               "All elements of 'list.integral' should be of class 'integer' or")
  expect_equal(.validate_integral(list.integral <- list(5:1), list.ok = TRUE),
               list(1:5))
  expect_warning(.validate_integral(integral <- c(1, 3)),
                 "'integral' was defined as c(1, 3) but in general we would",
                 fixed = TRUE)
  expect_no_warning(.validate_integral(integral <- c(1, 2)),
                    message = "'integral' was defined as c(1, 2) but in general")

  ## .convert_to_channels() -------------------------------------------------
  tl <- seq(1.8, 450, by = 1.8)
  expect_equal(.convert_to_channels(tl, NULL, null.ok = TRUE),
               NULL)
  expect_equal(.convert_to_channels(tl, NA, na.ok = TRUE),
               NA)
  expect_equal(.convert_to_channels(tl, 0:20),
               1:11)
  expect_equal(.convert_to_channels(tl, 200:220),
               111:122)
  expect_equal(.convert_to_channels(tl, c(200:210, NA, 200:220)),
               111:122)
  expect_equal(.convert_to_channels(tl, list(200:210, 200:220), list.ok = TRUE),
               list(111:117, 111:122))
  expect_warning(expect_equal(.convert_to_channels(tl, signal_integral <- c(0, 1.5),
                                                   unit = "temperature"),
                              1),
                 "Conversion of 'signal_integral' from temperature to channels failed")
  expect_warning(expect_equal(.convert_to_channels(tl, signal_integral <- c(1000, 2000),
                                                   unit = "temperature"),
                              250),
                 "Conversion of 'signal_integral' from temperature to channels failed")

  expect_error(.convert_to_channels(tl, integral <- NULL),
               "'integral' should be of class 'integer' or 'numeric'")
  expect_error(.convert_to_channels(tl, integral <- NA),
               "'integral' should be of class 'integer' or 'numeric'")
  expect_error(.convert_to_channels(tl, integral <- list(200:210, NA), list.ok = FALSE),
               "'integral' should be of class 'integer' or 'numeric'")
  expect_error(.convert_to_channels(tl, integral <- list(200:210, NA), list.ok = TRUE),
               "All elements of 'integral' should be of class 'integer' or 'numeric'")

  ## .validate_file() -------------------------------------------------------

  dir.path <- system.file("extdata", package = "Luminescence")
  expect_equal(.validate_file(file <- c("file1", "file2")),
               list("file1", "file2"))
  expect_equal(.validate_file(file <- list("file1")),
               list("file1"))
  expect_message(expect_equal(.validate_file(dir.path),
                              as.list(dir(dir.path, full.names = TRUE))),
                 "Directory detected, looking for any files")
  expect_silent(.validate_file(dir.path, verbose = FALSE))
  expect_message(expect_equal(basename(.validate_file(dir.path, pattern = "xsyg")),
                              "XSYG_file.xsyg"),
                 "Directory detected, looking for 'xsyg' files")
  expect_message(expect_message(
      expect_length(.validate_file(dir.path, pattern = "_none_"), 0),
      "Directory detected, looking for '_none_' files"),
      "No files matching the given pattern found")

  url <- "https://raw.githubusercontent.com/R-Lum/rxylib/master"
  expect_message(expect_message(.validate_file(file.path(url, "_pkgdown.yml")),
                                "Downloading"), "OK")
  expect_message(expect_message(.validate_file(file.path(url, "_error_")),
                                "Downloading"), "FAILED")

  expect_error(.validate_file(file <- TRUE),
               "'file' should be of class 'character' or 'list'")
  expect_error(.validate_file(file <- list(TRUE)),
               "All elements of 'file' should be of class 'character' and have length 1")
  expect_error(.validate_file("_error_"),
               "File '.*_error_' does not exist") # windows CI needs the regexp
  expect_message(expect_null(.validate_file("_error_", throw.error = FALSE)),
                 "File '.*_error_' does not exist") # windows CI needs the regexp
  expect_error(.validate_file(dir.path, scan.dir = FALSE),
               "File '.*' does not exist")
  file.create(zero <- tempfile(pattern = "zero", fileext = ".binx"))
  expect_error(.validate_file(zero),
               "is a zero-byte file")
  expect_message(expect_null(.validate_file(zero, throw.error = FALSE)),
                             "is a zero-byte file")
  expect_error(.validate_file(test_path("test_read_BIN2R.R"), ext = "e1"),
               "File extension 'R' is not supported, only 'e1' is valid")
  expect_error(.validate_file(test_path("test_read_BIN2R.R"), ext = c("e1", "e2")),
               "File extension 'R' is not supported, only 'e1' and 'e2' are valid")
  expect_message(expect_null(.validate_file(test_path("test_read_BIN2R.R"),
                                            ext = c("e1", "e2", "e3"), throw.error = FALSE)),
                 "File extension 'R' is not supported, only 'e1', 'e2' and 'e3'")

  ## .require_suggested_package() -------------------------------------------
  expect_true(.require_suggested_package("utils"))
  expect_error(.require_suggested_package("error"),
               "This function requires the 'error' package: to install it")
  expect_error(.require_suggested_package("error",
                                          reason = "Reporting a good error"),
               "Reporting a good error requires the 'error' package")
  expect_warning(
      expect_false(.require_suggested_package("error", throw.error = FALSE),
                   "This function requires the 'error' package: to install it"))

  ## .listify() -------------------------------------------------------------
  expect_equal(.listify(1, length = 3),
               list(1, 1, 1))
  expect_equal(.listify(letters, length = 5),
               .listify(list(letters), length = 5))


  ## .strict_na() -----------------------------------------------------------
  expect_true(.strict_na(NA))
  expect_true(.strict_na(NA_real_))
  expect_false(.strict_na(NULL))
  expect_false(.strict_na(c(1, NA)))
  expect_false(.strict_na(c(NA, NA)))
  expect_false(.strict_na(matrix()))
  expect_false(.strict_na(set_RLum("RLum.Data.Curve")))

  ## .collapse() ------------------------------------------------------------
  expect_equal(.collapse(1:3),
               "'1', '2', '3'")
  expect_equal(.collapse(1:3, quote = FALSE),
               "1, 2, 3")
  expect_equal(.collapse(1:2, quote = FALSE, last_sep = " or "),
               "1 or 2")
  expect_equal(.collapse(1:3, quote = FALSE, last_sep = " or "),
               "1, 2 or 3")
  expect_equal(.collapse(as.factor(c("060920", "070920", "080920", "090920")),
                         quote = FALSE),
               "060920, 070920, 080920, 090920")
  expect_equal(.collapse(NULL), "")

  ## .format_range() ________________________________________________________
  expect_equal(.format_range(1:10),
               "1:10")
  expect_equal(.format_range(c(1, 10, NA, 8)),
               "1:10")
  expect_equal(.format_range(c(-1, -10, NA, -8)),
               "-10:-1")
  expect_equal(.format_range(NULL),
               "NA:NA")
  expect_equal(.format_range(c(NA, NA, NA)),
               "NA:NA")
  expect_equal(.format_range(c(0, 1, -1), sep = ", "),
               "-1, 1")
  expect_equal(.format_range(c(0.53, 1.39, 1.14)),
               "0.53:1.39")
  expect_equal(.format_range(c(0.53, 1.39, 1.14), nsmall = 3),
               "0.530:1.390")

  ## .shorten_filename() ----------------------------------------------------
  expect_equal(.shorten_filename("/path/to/filename"),
               "/path/to/filename")
  max.width <- 27
  shortened <- .shorten_filename("/path/to/a_somewhat_longer_filename",
                                 max.width = max.width)
  expect_equal(shortened,
               "/path/to/a_s...ger_filename")
  expect_equal(nchar(shortened), max.width)

  ## check vector
  max.width <- 10
  shortened <- .shorten_filename(c("short", NA, "muchmuchlonger"), 10)
  expect_equal(shortened,
               c("short", NA, "muc...nger"))
  expect_equal(nchar(shortened), c(5, NA, max.width))

  ## .rescale() ----------------------------------------------------
  x <- stats::rnorm(100)
  y_dens <- stats::density(x)
  expect_equal(max(.rescale(
   x = y_dens$y,
   range_old = c(min(y_dens$y), max(y_dens$y)),
   range_new = c(0, 20))), 20)

  ## C++ code ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ##
  # src_create_RLumDataCurve_matrix -------------------------------------------------------------
  ##RLum.Data.Curve() ... test src_create_RLumDataCurve_matrix()
  expect_output(
    Luminescence:::src_create_RLumDataCurve_matrix(
      DATA = 1:100,
      VERSION = 4,
      NPOINTS = 100,
      LTYPE = "TL",
      LOW = 0,
      HIGH = 500,
      AN_TEMP = 0,
      TOLDELAY = 0,
      TOLON = 0,
      TOLOFF = 0
    )
  )
  ## cover NA case
  expect_type(
    Luminescence:::src_create_RLumDataCurve_matrix(
      DATA = 1:100,
      VERSION = 4,
      NPOINTS = 0,
      LTYPE = "OSL",
      LOW = 0,
      HIGH = 500,
      AN_TEMP = 0,
      TOLDELAY = 0,
      TOLON = 0,
      TOLOFF = 0), type = "double")

  ## case for a delayed hit ramp start
  expect_type(
    Luminescence:::src_create_RLumDataCurve_matrix(
      DATA = 1:100,
      VERSION = 4,
      NPOINTS = 0,
      LTYPE = "OSL",
      LOW = 0,
      HIGH = 500,
      AN_TEMP = 0,
      TOLDELAY = 10,
      TOLON = 0,
      TOLOFF = 0), type = "double")
  ## case for a delayed hit ramp start
  expect_type(
    Luminescence:::src_create_RLumDataCurve_matrix(
      DATA = 1:100,
      VERSION = 4,
      NPOINTS = 0,
      LTYPE = "OSL",
      LOW = 0,
      HIGH = 500,
      AN_TEMP = 100,
      TOLDELAY = 0,
      TOLON = 0,
      TOLOFF = 0), type = "double")
  ## generate strange curve and more tests
  expect_type(Luminescence:::src_create_RLumDataCurve_matrix(
    DATA = 1:100,
    VERSION = 4,
    NPOINTS = 100,
    LTYPE = "TL",
    LOW = 0,
    HIGH = 500,
    AN_TEMP = 200,
    TOLDELAY = 10,
    TOLON = 10,
    TOLOFF = 400),  type = "double")
})

## load data
data(ExampleData.BINfileData, envir = environment())
data(ExampleData.XSYG, envir = environment())

## convert values for position 1
temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)

## only one curve
c1 <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1,
                                      run = 1, set = 6)

## image
set.seed(1)
img <- set_RLum("RLum.Analysis",
                records = list(set_RLum("RLum.Data.Image",
                                        data = array(rnorm(250), c(5, 5, 10)))))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RLum.Analysis("error"),
               "'object' should be of class 'RLum.Analysis'")

  expect_error(plot_RLum.Analysis(temp, nrows = -1),
               "'nrows' should be a single positive integer value")
  expect_error(plot_RLum.Analysis(temp, ncols = -1),
               "'ncols' should be a single positive integer value")
  expect_error(plot_RLum.Analysis(temp, combine = -1),
               "'combine' should be a single logical value")
  expect_error(plot_RLum.Analysis(temp, norm = -3),
               "'norm' should be a single positive value or one of 'max', 'min'")

  expect_error(plot_RLum.Analysis(
      set_RLum("RLum.Analysis", records = list(c1@records[[1]],
                                               set_RLum("RLum.Data.Image"))),
      combine = TRUE),
      "'combine' is valid only for 'RLum.Data.Curve' objects")

  ## empty object
  expect_message(expect_null(plot_RLum.Analysis(set_RLum("RLum.Analysis"))),
                 "Nothing to plot, NULL returned")
  expect_message(
      plot_RLum.Analysis(set_RLum("RLum.Analysis"), combine = TRUE),
      "Nothing to plot, NULL returned")

  ## empty RLum.Data.Image
  expect_silent(plot_RLum.Analysis(
      set_RLum("RLum.Analysis", records = list(set_RLum("RLum.Data.Image")))))

  ## this generates multiple warnings
  warnings <- capture_warnings(plot_RLum.Analysis(c1, col = 2,
                                                  xlim = c(-1, 50),
                                                  ylim = c(-1, 3000)))
  expect_match(warnings, all = FALSE, fixed = TRUE,
               "min('xlim') < x-value range for curve #1, reset to minimum")
  expect_match(warnings, all = FALSE, fixed = TRUE,
               "max('xlim') > x-value range for curve #1, reset to maximum")
  expect_match(warnings, all = FALSE, fixed = TRUE,
               "min('ylim') < y-value range for curve #1, reset to minimum")
  expect_match(warnings, all = FALSE, fixed = TRUE,
               "max('ylim') > y-value range for curve #1, reset to maximum")

  expect_error(plot_RLum.Analysis(c1, curve.transformation = "error"),
               "'curve.transformation' should be one of 'CW2pLM', 'CW2pLMi'")
  expect_warning(.warningCatcher(
      plot_RLum.Analysis(temp, subset = list(recordType = "TL"),
                         norm = TRUE, log = "y")),
      "12 y values <= 0 omitted from logarithmic plot")

  expect_message(plot_RLum.Analysis(c1, combine = TRUE, main = "Curve"),
                 "'combine' can't be used with fewer than two curves")
  expect_warning(plot_RLum.Analysis(c1, plot.single = TRUE),
                  "'plot.single' was deprecated in v1.0.0, use 'plot_singlePanels'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## trigger warning
  expect_silent(expect_warning(plot_RLum.Analysis(
    set_RLum("RLum.Analysis", records = list(
      c1@records[[1]],
      set_RLum("RLum.Data.Curve", recordType = "OSL")
      )), norm = TRUE, combine = TRUE),
    "[normalise_RLum()] Curve normalisation produced Inf/NaN values, values replaced",
    fixed = TRUE))

  ##Basic plot
  expect_silent(plot_RLum.Analysis(
    temp,
    subset = list(recordType = "TL"),
    combine = TRUE,
    norm = TRUE,
    abline = list(v = c(110))
  ))

  ## Basic plot with spectrum
  expect_silent(
    plot_RLum.Analysis(
      set_RLum(class = "RLum.Analysis", records = list(TL.Spectrum, temp[[1]])),
      plot.type = "persp"))

  ## test norm = "max"
  expect_silent(plot_RLum.Analysis(
    temp,
    subset = list(recordType = "TL"),
    combine = TRUE,
    norm = "max",
    abline = list(v = c(110))
  ))

  ## test norm = "min"
  expect_silent(plot_RLum.Analysis(
    temp,
    subset = list(recordType = "OSL"),
    combine = TRUE,
    norm = "last",
    abline = list(v = c(110))
  ))

  ## test norm = "autoscale"
  expect_silent(plot_RLum.Analysis(
    temp[1:4],
    subset = list(recordType = "OSL"),
    combine = TRUE,
    auto_scale = TRUE,
    xlim = c(10, 20),
    legend.text = paste("Curve", 1:2),
    norm = "last",
    abline = list(v = c(110))
  ))

  ## test norm = "huot
  expect_silent(plot_RLum.Analysis(
    temp,
    subset = list(recordType = "OSL"),
    combine = TRUE,
    norm = "huot",
    abline = list(v = c(110))
  ))

  ## test records_max
  expect_silent(plot_RLum.Analysis(
    temp,
    subset = list(recordType = "TL"),
    combine = TRUE,
    norm = TRUE,
    sub_title = "(5 K/s)",
    records_max = 5,
    smooth = TRUE,
    type = "p",
    abline = list(v = c(110)),
    ## more coverage
    main = "TL curves combined",
    log = "xy",
    col = get("col", pos = .LuminescenceEnv)[1:4],
    xlab = "Temperature recorded [log \u00B0C]", ylab = "log TL [a.u.]",
    xlim = c(0, 200), ylim = c(0, 1), lty = c(1, 2),
    legend.col = get("col", pos = .LuminescenceEnv)[1:4],
    legend.pos = "outside"
  ))

  ##test arguments
  ##ylim
  expect_silent(plot_RLum.Analysis(
    temp,
    subset = list(recordType = "TL"),
    combine = FALSE,
    norm = TRUE,
    ylim = c(0, 1),
    xlim = c(1,100),
    abline = list(v = c(110))
  ))

  ## curve transformation
  plot_RLum.Analysis(temp,
                     subset = list(recordType = "IRSL"),
                     curve.transformation = "CW2pLMi")

  plot_RLum.Analysis(temp,
                     subset = list(recordType = "OSL"),
                     combine = TRUE,
                     curve.transformation = "CW2pHMi")

  plot_RLum.Analysis(temp,
                     subset = list(recordType = "TL"),
                     curve.transformation = "CW2pPMi")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("plot_RLum.Analysis",
                              plot_RLum.Analysis(
                                  temp,
                                  subset = list(recordType = "TL"),
                                  combine = TRUE,
                                  norm = TRUE,
                                  abline = list(v = 110)))
  vdiffr::expect_doppelganger("combine-cex",
                              plot_RLum.Analysis(
                                  temp,
                                  subset = list(recordType = "TL"),
                                  combine = TRUE,
                                  cex = 2))
  vdiffr::expect_doppelganger("smooth-type-cex",
                              plot_RLum.Analysis(
                                  temp[1:6],
                                  smooth = TRUE,
                                  type = "b",
                                  cex = 2))
  vdiffr::expect_doppelganger("plot_RLum.Analysis persp",
                              plot_RLum.Analysis(
                                  set_RLum(class = "RLum.Analysis",
                                           records = list(TL.Spectrum, temp[[1]])),
                                  plot.type = "persp"))
  vdiffr::expect_doppelganger("records_max",
                              plot_RLum.Analysis(
                                  temp[1:6],
                                  combine = TRUE,
                                  records_max = 10))
  vdiffr::expect_doppelganger("abline after", {
                              plot_RLum.Analysis(
                                  temp[1:6],
                                  subset = list(recordType = "TL"),
                                  combine = TRUE,
                                  records_max = 10)
                              abline(v = 25)
                              })
  vdiffr::expect_doppelganger("abline outside",
                              plot_RLum.Analysis(
                                  temp[1:8],
                                  subset = list(recordType = "TL"),
                                  combine = TRUE,
                                  abline = list(v = c(50, 150)),
                                  legend.pos = "outside"))
  vdiffr::expect_doppelganger("image",
                              plot_RLum.Analysis(
                                  img,
                                  frame = 3))
  })
})

#' @title Al2O3 Irradiation Time Correction Analysis
#'
#' @description The function provides a very particular analysis to correct the irradiation
#' time while irradiating Al2O3:C chips in a luminescence reader.
#'
#' @details Background: Due to their high dose sensitivity Al2O3:C chips are usually
#' irradiated for only a very short duration or under the closed beta-source
#' within a luminescence reader. However, due to its high dose sensitivity, during
#' the movement towards the beta-source, the pellet already receives and non-negligible
#' dose. Based on measurements following a protocol suggested by Kreutzer et al., 2018,
#' a dose response curve is constructed and the intersection (absolute value) with the time axis
#' is taken as real irradiation time.
#'
#' **`method_control`**
#'
#' To keep the generic argument list as clear as possible, arguments to allow a
#' deeper control of the method are all preset with meaningful default parameters and can be
#' handled using the argument `method_control` only, e.g.,
#' `method_control = list(fit.method = "LIN")`. Supported arguments are:
#'
#' \tabular{lll}{
#' **ARGUMENT** \tab **FUNCTION** \tab **DESCRIPTION**\cr
#' `mode` \tab `fit_DoseResponseCurve` \tab sets the mode used for fitting\cr
#' `fit.method` \tab `fit_DoseResponseCurve` \tab sets the function applied for fitting\cr
#' }
#'
#' @param object [Luminescence::RLum.Analysis-class] or [list] (**required**):
#' results obtained from the measurement.
#' Alternatively a list of [Luminescence::RLum.Analysis-class] objects can be provided to allow an automatic analysis
#'
#' @param signal_integral [numeric] (*optional*):
#' signal integral, used for the signal and the background.
#' If nothing is provided the full range is used. Argument can be provided as [list].
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param dose_points [numeric] (*with default*):
#' vector with dose points, if dose points are repeated, only the general
#' pattern needs to be provided. Default values follow the suggestions
#' made by Kreutzer et al., 2018. Argument can be provided as [list].
#'
#' @param recordType [character] (*with default*): input curve selection, which is passed to
#' function [Luminescence::get_RLum]. To deactivate the automatic selection set the argument to `NULL`
#'
#' @param method_control [list] (*optional*):
#' optional parameters to control the calculation.
#' See details for further explanations
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments that can be passed to the plot output
#'
#' @return
#' Function returns results numerically and graphically:
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'   **Element** \tab **Type** \tab **Description**\cr
#'  `$data` \tab `data.frame` \tab correction value and error \cr
#'  `$table` \tab `data.frame` \tab table used for plotting  \cr
#'  `$table_mean` \tab `data.frame` \tab table used for fitting \cr
#'  `$fit` \tab `lm` or `nls` \tab the fitting as returned by the function [Luminescence::fit_DoseResponseCurve]
#' }
#'
#'**slot:** **`@info`**
#'
#' The original function call
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' - A dose response curve with the marked correction values
#'
#' @section Function version: 0.1.2
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::fit_DoseResponseCurve]
#'
#' @references
#'
#' Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018. Environmental Dose Rate
#' Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
#' Geochronometria 45, 56-67. doi: 10.1515/geochr-2015-0086
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.Al2O3C, envir = environment())
#'
#' ##run analysis
#' analyse_Al2O3C_ITC(data_ITC)
#'
#' @export
analyse_Al2O3C_ITC <- function(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  dose_points = c(2,4,8,12,16),
  recordType = "OSL (UVVIS)",
  method_control = NULL,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("analyse_Al2O3C_ITC")
  on.exit(.unset_function_name(), add = TRUE)

  .validate_not_empty(object) # here for issue 1256

  # SELF CALL -----------------------------------------------------------------------------------
  if (inherits(object, "list")) {
    ##check whether the list contains only RLum.Analysis objects
    lapply(object, .validate_class, "RLum.Analysis",
           name = "All elements of 'object'")

    ## expand input arguments
    rep.length <- length(object)
    signal_integral <- .listify(signal_integral, rep.length)
    dose_points <- .listify(dose_points, rep.length)

    ##run analysis
    results_full <- lapply(1:length(object), function(x){
      ##run analysis
      results <- try(analyse_Al2O3C_ITC(
        object = object[[x]],
        signal_integral = signal_integral[[x]],
        integral_input = integral_input,
        dose_points = dose_points[[x]],
        method_control = method_control,
        verbose = verbose,
        plot = plot,
        main = list(...)$main %||% paste0("ALQ #", x),
        ...
      ), outFile = stdout()) # redirect error messages so they can be silenced

      ##catch error
      if (is.null(results) || inherits(results, "try-error"))
        return(NULL)
      return(results)
    })

    ##return
    return(merge_RLum(results_full))
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis", extra = "a 'list' of such objects")
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  .validate_class(dose_points, c("numeric", "list"))
  if (is.list(dose_points)) {
    lapply(dose_points, .validate_class, "numeric",
           name = "All elements of 'dose_points'")
  }
  .validate_class(recordType, "character", null.ok = TRUE)
  .validate_class(method_control, "list", null.ok = TRUE)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

  # Preparation ---------------------------------------------------------------------------------
  ##select curves based on the recordType selection; if not NULL
  if (!is.null(recordType)) {
    object <- get_RLum(object, recordType = recordType, drop = FALSE)
    if (length(object) == 0)
      .throw_error("'recordType' produced an empty object")
  }

  #set method control
  method_control_settings <- list(
    mode = "extrapolation",
    fit.method = "EXP"
  )

  ## modify on request
  if (!is.null(method_control)) {
    method_control_settings <- modifyList(x = method_control_settings,
                                          val = method_control)
  }

  ##dose points enhancement
  ##make sure that the dose_point is enhanced
  dose_points <- rep(dose_points, times = length(object)/2)

  # Calculation ---------------------------------------------------------------------------------

  ## signal integral
  x.range <- object[[1]][, 1]
  if (integral_input == "measurement") {
    signal_integral <- .convert_to_channels(x.range, signal_integral,
                                            "time", null.ok = TRUE)
  }
  signal_integral <- .validate_integral(signal_integral,
                                        max = length(x.range), null.ok = TRUE)
  if (is.null(signal_integral)) {
    signal_integral <- seq_along(x.range)
  }

  ##calculate curve sums, assuming the background
  net_SIGNAL <- vapply(seq(1, length(object), by = 2), function(x) {
    temp_signal <- sum(object[[x]][signal_integral, 2])
    temp_background <- sum(object[[x + 1]][signal_integral, 2])
    return(temp_signal - temp_background)
  }, FUN.VALUE = numeric(1))

  ## silence notes raised by R CMD check
  DOSE <- net_SIGNAL_NORM <- NULL

  ## create data.tables
  df <- data.table(
      DOSE = dose_points,
      net_SIGNAL = net_SIGNAL,
      net_SIGNAL.ERROR = 0,
      net_SIGNAL_NORM = net_SIGNAL/max(net_SIGNAL),
      net_SIGNAL_NORM.ERROR = 0
    )

  ## compute means and errors
  df_mean <- as.data.frame(df[, list(net_SIGNAL = mean(net_SIGNAL),
                                     net_SIGNAL.ERROR = sd(net_SIGNAL),
                                     net_SIGNAL_NORM = mean(net_SIGNAL_NORM),
                                     net_SIGNAL_NORM.ERROR = sd(net_SIGNAL_NORM)),
                              by = DOSE])

  ##calculate GC
  GC <- fit_DoseResponseCurve(
    object = df_mean,
    mode = method_control_settings$mode,
    fit.method = method_control_settings$fit.method,
    verbose = FALSE
  )
  if (is.null(GC))
    return(NULL)

  ##output
  if(verbose){
    cat("\n[analyse_Al2O3C_ITC()]\n")
    cat("\n Used fit method:\t\t", method_control_settings$fit.method)
    cat("\n Time correction value:\t", round(GC$De$De, 3), "\u00B1", round(GC$De$De.Error, 3))
    cat("\n\n")
  }


  # Plotting ------------------------------------------------------------------------------------
  if(plot){
    ##set plot settings
    plot_settings <- list(
      xlab = "Dose [s]",
      ylab = "Integrated net GSL [a.u.]",
      main = "Irradiation Time Correction",
      xlim = c(-5, max(df$DOSE)),
      ylim = c(0,max(df$net_SIGNAL)),
      legend.pos = "right",
      legend.text = "dose points",
      mtext = ""
    )

    ##modify list on request
    plot_settings <- modifyList(x = plot_settings, val = list(...))

    ##make plot area
    plot(NA, NA,
         xlim = plot_settings$xlim,
         ylim = plot_settings$ylim,
         xlab = plot_settings$xlab,
         ylab = plot_settings$ylab,
         main = plot_settings$main)

    ##add zero lines
    abline(v = 0)
    abline(h = 0)

    ##add dose points
    points(x = df$DOSE, y = df$net_SIGNAL)

    ##add dose response curve
    x <- seq(min(plot_settings$xlim), max(plot_settings$xlim), length.out = 100)
    lines(
      x = x,
      y = eval(GC$Formula)
    )

    ##show offset
    x <- 0
    lines(x = c(-GC$De[1], -GC$De[1]), y = c(eval(GC$Formula), 0), lty = 2, col = "red")
    shape::Arrows(
      x0 = 0,
      y0 = eval(GC$Formula),
      x1 = as.numeric(-GC$De[1]),
      y1 = eval(GC$Formula),
      arr.type = "triangle",
      arr.adj = -0.5,
      col = 'red',
      cex = par()$cex)

    ##add text
    text(
      x = -GC$De[1] / 2,
      y = eval(GC$Formula),
      pos = 3,
      labels = paste(round(GC$De[1],3), "\u00B1", round(GC$De[2], 3)),
      col = 'red',
      cex = 0.8)

    ##add 2nd x-axis
    axis(
      side = 1,
      at = axTicks(side = 1),
      labels = paste0("(",(axTicks(side = 1) + round(as.numeric(GC$De[1]),2)), ")"),
      line = 1,
      col.axis = "red",
      lwd.ticks = 0,
      lwd = 0,
      cex.axis = 0.9
    )

    ##add legend
    legend(
      plot_settings$legend.pos,
      bty = "n",
      pch = 1,
      legend = plot_settings$legend.text
    )

    ##add mtext
    mtext(side = 3, text = plot_settings$mtext)
  }

  # Output --------------------------------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      data = data.frame(
        VALUE = as.numeric(GC$De$De),
        VALUE_ERROR = as.numeric(sd(GC$De.MC))
      ),
      table = as.data.frame(df),
      table_mean = df_mean,
      fit = GC$Fit
    ),
    info = list(call = sys.call())
  )
}

## load data
data(ExampleData.Al2O3C, envir = environment())

test_that("input validation", {
  skip_on_cran()

  expect_error(analyse_Al2O3C_CrossTalk("error"),
               "'object' should be of class 'RLum.Analysis' or 'list")
  expect_error(analyse_Al2O3C_CrossTalk(list("error")),
               "All elements of 'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_Al2O3C_CrossTalk(list()),
               "'object' cannot be an empty list")
  expect_error(analyse_Al2O3C_CrossTalk(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk, signal_integral = NA),
               "'signal_integral' should be of class 'integer', 'numeric' or NULL")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk, dose_points = NA),
               "'dose_points' should be of class 'numeric' or 'integer'")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk, dose_points = numeric(0)),
               "'dose_points' cannot be an empty numeric")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk, dose_points = 1:3),
               "'dose_points' should have length 1 or divisible by 2")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk,
                                        recordType = list(NA)),
               "'recordType' should be of class 'character'")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk,
                                        recordType = "error"),
               "'object' contains no records with recordType = 'error'")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk,
                                        method_control = "EXP"),
               "'method_control' should be of class 'list'")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk,
                                        irradiation_time_correction = FALSE),
               "'irradiation_time_correction' should be of class 'numeric',")
  corr <- set_RLum("RLum.Results")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk,
                                        irradiation_time_correction = corr),
               "'irradiation_time_correction' has an unsupported originator")
  expect_error(analyse_Al2O3C_CrossTalk(data_CrossTalk,
                                        signal_integral = 0),
               "is of length 0 after removing values smaller than 1 and greater than 99")
})

test_that("check functionality", {
  skip_on_cran()

  ## integral_input
  res1 <- analyse_Al2O3C_CrossTalk(data_CrossTalk[[1]],
                                   signal_integral = c(42.1, 42.4),
                                   integral_input = "measurement")
  res2 <- analyse_Al2O3C_CrossTalk(data_CrossTalk[[1]],
                                   signal_integral = 1:4,
                                   integral_input = "channel")
  res1@info <- res2@info <- list() # remove $call
  res1@data$fit <- res2@data$fit <- NULL
  res1@.uid <- res2@.uid <- NA_character_
  expect_equal(res1, res2)

  expect_warning(analyse_Al2O3C_CrossTalk(data_CrossTalk[[1]],
                                          signal_integral = 1:4,
                                          integral_input = "measurement"),
                 "from time to channels failed: expected values in 42.1:51.9")
})

test_that("snapshot tests", {
  skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.7e-5

  ## run analysis
  expect_snapshot_RLum(
      analyse_Al2O3C_CrossTalk(data_CrossTalk),
      tolerance = snapshot.tolerance)
  expect_snapshot_RLum(
      analyse_Al2O3C_CrossTalk(data_CrossTalk,
                               method_control = list(fit.method = "LIN")),
      tolerance = snapshot.tolerance)

  ## irradiation_time_correction
  SW({
  corr <- analyse_Al2O3C_ITC(data_ITC)
  })
  expect_snapshot_RLum(
      analyse_Al2O3C_CrossTalk(data_CrossTalk,
                               irradiation_time_correction = corr),
      tolerance = snapshot.tolerance)

  corr@data$data <- rbind(corr@data$data, corr@data$data)
  expect_snapshot_RLum(
      analyse_Al2O3C_CrossTalk(data_CrossTalk,
                               irradiation_time_correction = corr),
      tolerance = snapshot.tolerance)

  ## signal integral
  expect_snapshot_RLum(
      analyse_Al2O3C_CrossTalk(data_CrossTalk,
                               signal_integral = 1:5),
      tolerance = snapshot.tolerance)
})

test_that("regression tests", {
  skip_on_cran()

  ## issue 1182
  expect_silent(analyse_Al2O3C_CrossTalk(data_ITC))
})

##load data
data(ExampleData.Al2O3C, envir = environment())

test_that("input validation", {
  skip_on_cran()

  a <- set_RLum(class = "RLum.Data.Curve", recordType = "OSL",
                data = matrix(1:20, ncol = 2))
  b <- set_RLum(class = "RLum.Data.Curve", recordType = "TL")
  object <- set_RLum(class = "RLum.Analysis", records = list(a,b))

  expect_error(analyse_Al2O3C_ITC("test"),
               "'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_Al2O3C_ITC(iris),
               "'object' should be of class 'RLum.Analysis' or a 'list' of")
  expect_error(analyse_Al2O3C_ITC(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_error(analyse_Al2O3C_ITC(list()),
               "'object' cannot be an empty list")
  expect_error(analyse_Al2O3C_ITC(list(data_ITC, "test")),
               "All elements of 'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_Al2O3C_ITC(data_ITC, recordType = NA),
               "'recordType' should be of class 'character' or NULL")
  expect_error(expect_warning(analyse_Al2O3C_ITC(data_ITC, recordType = "EXP"),
                              "This request produced an empty list of records"),
               "'recordType' produced an empty object")
  expect_error(analyse_Al2O3C_ITC(data_ITC, method_control = "EXP"),
               "'method_control' should be of class 'list'")
  expect_error(analyse_Al2O3C_ITC(data_ITC, dose_points = NA),
               "'dose_points' should be of class 'numeric' or 'list'")
  expect_error(analyse_Al2O3C_ITC(data_ITC, dose_points = list(NA)),
               "All elements of 'dose_points' should be of class 'numeric'")
  expect_error(analyse_Al2O3C_ITC(data_ITC, signal_integral = 0),
               "'signal_integral' is of length 0 after removing values smaller")
})

test_that("check functionality", {
  skip_on_cran()

  SW({
  ## fit_DoseResponseCurve() fails to fit
  expect_warning(expect_null(analyse_Al2O3C_ITC(list(data_ITC),
                                                dose_points = list(2))),
                 "Nothing was merged as the object list was found to be empty")
  })

  ## integral_input
  set.seed(1)
  res1 <- analyse_Al2O3C_ITC(data_ITC, signal_integral = c(42.1, 42.3),
                             integral_input = "measurement", verbose = FALSE)
  set.seed(1)
  res2 <- analyse_Al2O3C_ITC(data_ITC, signal_integral = 1:3,
                             integral_input = "channel", verbose = FALSE)
  res1@info <- res2@info <- list() # remove $call
  res1@data$fit <- res2@data$fit <- NULL
  res1@.uid <- res2@.uid <- NA_character_
  expect_equal(res1, res2)

  expect_warning(analyse_Al2O3C_ITC(data_ITC, signal_integral = 1:4,
                                    integral_input = "measurement", verbose = FALSE),
                 "from time to channels failed: expected values in 42.1:51.9")
})

test_that("snapshot tests", {
  skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 5.0e-6

  SW({
  expect_snapshot_RLum(analyse_Al2O3C_ITC(data_ITC),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(analyse_Al2O3C_ITC(list(data_ITC),
                                          signal_integral = 2,
                                          method_control = list(fit.method = "EXP")),
                       tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
    vdiffr::expect_doppelganger("default",
                                analyse_Al2O3C_ITC(data_ITC))
  })
})

#'@title Calculate dose rate of slices in a spherical cobble
#'
#'@description
#'Calculates the dose rate profile through the cobble based on Riedesel and Autzen (2020).
#'
#'Corrects the beta dose rate in the cobble for the grain size following results
#'of Guérin et al. (2012). Sediment beta and gamma dose rates are corrected
#'for the water content of the sediment using the correction factors of Aitken (1985).
#'Water content in the cobble is assumed to be 0.
#'
#'@details
#'
#'**The input table layout**
#'
#'\tabular{lll}{
#'COLUMN \tab DATA TYPE \tab DESCRIPTION\cr
#'`Distance` \tab `numeric` \tab distance from the surface of the cobble to the top of each rock slice in mm. The distance for each slice will be listed in this column\cr
#'`DistanceError` \tab `numeric` \tab Error on the distance in mm\cr
#'`Thickness` \tab `numeric` \tab Thickness of each slice in mm\cr
#'`ThicknessError` \tab `numeric` \tab uncertainty of the thickness in mm.\cr
#'`Mineral` \tab `character` \tab `'FS'` for feldspar, `'Q'` for quartz, depending which mineral in the cobble is used for dating\cr
#'`Cobble_K` \tab `numeric` \tab K nuclide content in % of the bulk cobble\cr
#'`Cobble_K_SE` \tab `numeric` \tab error on K nuclide content in % of the bulk cobble\cr
#'`Cobble_Th` \tab `numeric` \tab Th nuclide content in ppm of the bulk cobble\cr
#'`Cobble_Th_SE` \tab `numeric` \tab error on Th nuclide content in ppm of the bulk cobble\cr
#'`Cobble_U` \tab `numeric` \tab U nuclide content in ppm of the bulk cobble\cr
#'`Cobble_U_SE` \tab `numeric` \tab error on U nuclide content in ppm of the bulk cobble\cr
#'`GrainSize` \tab `numeric` \tab average grain size in µm of the grains used for dating\cr
#'`Density` \tab `numeric` \tab Density of the cobble. Default is 2.7 g cm^-3\cr
#'`CobbleDiameter` \tab `numeric` \tab Diameter of the cobble in cm.\cr
#'`Sed_K` \tab `numeric` \tab K nuclide content in % of the sediment matrix\cr
#'`Sed_K_SE` \tab `numeric` \tab error on K nuclide content in % of the sediment matrix\cr
#'`Sed_Th` \tab `numeric` \tab Th nuclide content in ppm of the sediment matrix\cr
#'`Sed_Th_SE` \tab `numeric` \tab error on Th nuclide content in ppm of the sediment matrix\cr
#'`Sed_U` \tab `numeric` \tab U nuclide content in ppm of the sediment matrix\cr
#'`Sed_U_SE` \tab `numeric` \tab error on U nuclide content in ppm of the sediment matrix\cr
#'`GrainSize_Sed` \tab `numeric` \tab average grain size of the sediment matrix\cr
#'`Density_Sed` \tab `numeric` \tab average density of the sediment matrix\cr
#'`WaterContent` \tab `numeric` \tab mean water content of the sediment matrix in %\cr
#'`WaterContent_SE` \tab `numeric` \tab relative error on water content
#'}
#'
#'**Water content**
#'The water content provided by the user should be calculated according to:
#'
#'\deqn{(Wet\_weight - Dry\_weight) / Dry\_weight * 100}
#'
#' @param object [data.frame] (**required**):
#' A data frame containing all relevant information for each individual layer.
#' For the table layout see details.
#'
#' @param conversion [character] (*with default*): dose rate conversion factors
#' to use, see [Luminescence::BaseDataSet.ConversionFactors] for the accepted values.
#'
#' @param ... currently not used.
#'
#'@references
#'Riedesel, S., Autzen, M., 2020. Beta and gamma dose rate attenuation in rocks and sediment.
#'Radiation Measurements 133, 106295.
#'
#' @section Function version: 0.1.1
#'
#'@author Svenja Riedesel, Aberystwyth University (United Kingdom) \cr
#'Martin Autzen, DTU NUTECH Center for Nuclear Technologies (Denmark)
#'
#' @return
#' The function returns an [Luminescence::RLum.Results-class] object for which
#' the first element is a [matrix] (`DataIndividual`) that gives the dose rate
#' results for each slice for each decay chain individually, for both, the
#' cobble dose rate and the sediment dose rate. The second element is also a
#' [matrix] (`DataComponent`) that gives the total beta and gamma-dose rates
#' for the cobble and the adjacent sediment for each slice of the cobble.
#'
#'@keywords datagen
#'
#'@seealso [Luminescence::convert_Concentration2DoseRate]
#'
#'@examples
#'## load example data
#'data("ExampleData.CobbleData", envir = environment())
#'
#'## run function
#'calc_CobbleDoseRate(ExampleData.CobbleData)
#'
#'@export
calc_CobbleDoseRate <- function(
  object,
  conversion = "Guerinetal2011",
  ...
  ) {
  .set_function_name("calc_CobbleDoseRate")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("input" %in% ...names()) {
    object <- list(...)$input
    .deprecated(old = "input", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "data.frame")
  .validate_not_empty(object)

  ## this lists only the columns that are referenced by name in the script
  exp.cols <- c("Distance", "DistanceError", "Thickness", "ThicknessError",
                "Density", "CobbleDiameter")
  input <- object
  mis.cols <- setdiff(exp.cols, colnames(input))
  if (length(mis.cols) > 0)
    .throw_error("'object' doesn't contain the following columns: ",
                 .collapse(mis.cols))

  if ((max(input[,1])>input$CobbleDiameter[1]*10) ||
      ((max(input[,1]) + input[length(input[,1]),3]) > input$CobbleDiameter[1]*10))
    .throw_error("Slices outside of cobble: please ensure your distances ",
                 "are in mm and diameter is in cm")

  ## conversion factors: we do not use BaseDataSet.ConversionFactors directly
  ## as it is in alphabetical level, but we want to have 'Guerinetal2011'
  ## in first position, as that is our default value
  BaseDataSet.ConversionFactors <- NULL
  load(system.file("data", "BaseDataSet.ConversionFactors.rda",
                   package = "Luminescence"))
  valid_conversion_factors <- c("Guerinetal2011", "Cresswelletal2018",
                                "AdamiecAitken1998", "Liritzisetal2013")
  stopifnot(names(BaseDataSet.ConversionFactors) %in% valid_conversion_factors)
  conversion <- .validate_args(conversion, valid_conversion_factors)

  # Calculate Dose Rate -----------------------------------------------------
  SedDoseData <- matrix(data = NA, nrow = 1, ncol = 10)
  CobbleDoseData <- matrix(data = 0, nrow = 1, ncol = 10)

  CobbleDoseData <- input[1,5:12]
  CobbleDoseData <- cbind(CobbleDoseData,0,0)
  SedDoseData <- cbind(input[1, 5], input[1, 15:20], GrainSize = input[1, 12],
                       input[1, 23:24])

  res <- convert_Concentration2DoseRate(CobbleDoseData, conversion = conversion)
  CobbleDoseRate <- get_RLum(res)

  res <- convert_Concentration2DoseRate(SedDoseData, conversion = conversion)
  SedDoseRate <- get_RLum(res)

  ## Distance should be from the surface of the rock to the top of the slice. Distances and thicknesses are in mm
  N <- length(input$Distance)

  Diameter <- input$CobbleDiameter[1]

  ### Calculate gamma attenuation coefficient for the cobbles internal dose rate
  if (Diameter<25){
    CobbleGammaAtt <-
      (0.55 * exp(-0.45 * Diameter) + 0.09 * exp(-0.06 * Diameter)) * 10
  }else {
    CobbleGammaAtt <- 0.02
  }

  ## Scale the density and infinite matrix gamma dose rates ----
  Scaling <- input$Density[1] / 2.7
  GammaEdge <- 0.5 * (1 - exp(-0.039 * Diameter))
  GammaCentre <- 2 * GammaEdge

  DiameterSeq <-
    seq(0, Diameter * 10, by = 0.01) #Converts diameter into integer of 10 um

  ### Create matrices for use ----
  Temp <- matrix(data = NA, nrow = length(DiameterSeq), ncol = 9)
  DistanceError <- matrix(data = NA, nrow = N, ncol = 8)
  ThicknessError <- matrix(data = NA, nrow = N, ncol = 8)
  DataIndividual <- matrix(data = NA, nrow = N, ncol = 25)
  DataComponent <- matrix(data = NA, nrow = N, ncol = 9)

  ### Calculate dose rate profiles through the rock ----
  t <- Diameter * 10 - DiameterSeq
  tGamma <- t

  #Beta and gamma functions for the cobbles own dose rate
  beta.cobble.fun <- function(a, b) {
    function(x) (1 - a * exp(b * x * Scaling)) + (1 - a * exp(b * t * Scaling)) - 1
  }
  KBetaCobble <- function(x) (1 - 0.5 * exp(-3.77 * DiameterSeq))+(1-0.5*exp(-3.77*t))-1
  ThBetaCobble_short <- beta.cobble.fun(0.5, -5.36)
  ThBetaCobble_long <- beta.cobble.fun(0.33, -2.36)
  UBetaCobble_short <- beta.cobble.fun(0.5, -4.15)
  UBetaCobble_long <- beta.cobble.fun(0.33, -2.36)

  GammaCobble <- function(x) {
    (GammaCentre - GammaEdge * exp(-CobbleGammaAtt * x * Scaling)) +
    (GammaCentre - GammaEdge * exp(-CobbleGammaAtt * tGamma * Scaling)) -
    GammaCentre
  }

  #Beta and gamma functions for the sediment dose rates into the cobble
  beta.sed.fun <- function(a, b) {
    function(x) 2 - (1 - a * exp(b * x * Scaling)) - (1 - a * exp(b * t * Scaling))
  }
  KBetaSed <- beta.sed.fun(0.5, -3.77)
  ThBetaSed_short <- beta.sed.fun(0.5, -5.36)
  ThBetaSed_long <- beta.sed.fun(0.33, -2.36)
  UBetaSed_short <- beta.sed.fun(0.5, -4.15)
  UBetaSed_long <- beta.sed.fun(0.33, -2.36)

  GammaSed <- function(x) 2 - (1 - 0.5 * exp(-0.02 * x * Scaling)) - (1 - 0.5 * exp(-0.02 * tGamma *
                                                                                      Scaling))
  Temp[, 1] <- DiameterSeq
  Temp[, 2] <- KBetaCobble(DiameterSeq)
  Temp[, 3] <- ThBetaCobble_long(DiameterSeq)
  Temp[, 4] <- UBetaCobble_long(DiameterSeq)
  Temp[, 5] <- GammaCobble(DiameterSeq)
  Temp[, 6] <- KBetaSed(DiameterSeq)
  Temp[, 7] <- ThBetaSed_long(DiameterSeq)
  Temp[, 8] <- UBetaSed_long(DiameterSeq)
  Temp[, 9] <- GammaSed(DiameterSeq)

  TempThCob <- ThBetaCobble_short(DiameterSeq)
  TempUCob <- UBetaCobble_short(DiameterSeq)
  TempThSed <- ThBetaSed_short(DiameterSeq)
  TempUSed <- UBetaSed_short(DiameterSeq)

  n <- which(DiameterSeq >= (max(DiameterSeq)-0.15))[1]
  Max <- length(DiameterSeq)

  ## Create the full matrix based on the short and long beta attenuations
  idx <- c(0:16, n:Max)
  Temp[idx, 3] <- TempThCob[idx]
  Temp[idx, 7] <- TempThSed[idx]
  Temp[idx, 4] <- TempUCob[idx]
  Temp[idx, 8] <- TempUSed[idx]

  colnames(Temp) <- c(
    "Distance",
    "KBetaCob",
    "ThBetaCob",
    "UBetaCob",
    "GammaCob",
    "KBetaSed",
    "ThBetaSed",
    "UBetaSed",
    "GammaSed"
  )

  ### Create data output matrices ----
  Distances <- input$Distance / 0.01 + 1
  Thicknesses <- input$Thickness / 0.01

  MinDistance <- (input$Distance - input$DistanceError) / 0.01 + 1
  MaxDistance <- (input$Distance + input$DistanceError) / 0.01 + 1

  MinThickness <- (input$Thickness - input$ThicknessError) / 0.01
  MaxThickness <- (input$Thickness + input$ThicknessError) / 0.01

  for (i in 1:N){
    Start <- Distances[i]
    End <- Start+Thicknesses[i]

    d_min <- MinDistance[i]
    d_max <- MaxDistance[i]

    t_min <- MinThickness[i]
    t_max <- MaxThickness[i]

    #Calculate errors ----
    #Check if minimum distance from top is less than 0
    if (MinDistance[i]<0){
      d_min <- 0
    }

    j <- d_min+Thicknesses[i]
    k <- d_max+Thicknesses[i]

    for (l in 1:8){
      m <- l + 1
      if (d_min == Start){
        DistanceError[i,l]<- abs(
          (mean(Temp[d_max:k,m])-mean(Temp[Start:End,m]))/(2*mean(Temp[Start:End,m])))
      } else if (k > Max){
        DistanceError[i,l] <- abs(
          (mean(Temp[Start:End,m])-mean(Temp[d_min:j,m]))/(2*mean(Temp[Start:End,m])))
      } else {
        DistanceError[i,l] <- abs(
          mean((mean(Temp[d_max:k,m])-mean(Temp[Start:End,m])):(mean(Temp[Start:End,m])-mean(Temp[d_min:j,m])))/(2*mean(Temp[Start:End,m])))
      }

      j2 <- Start+t_min
      k2 <- Start+t_max

      if (k2 > Max){
        ThicknessError[i,l] <- abs(
          (mean(Temp[Start:End,m])-mean(Temp[Start:j2,m]))/(2*mean(Temp[Start:End,m])))
      } else {
        ThicknessError[i,l] <- abs(
          mean((mean(Temp[Start:k2,m])-mean(Temp[Start:End,m])):(mean(Temp[Start:End,m])-mean(Temp[Start:j2,m])))/(2*mean(Temp[Start:End,m])))
      }
    }
    ### Calculate average dose rates ----

    DataIndividual[i, 1] <- input[i, 1]
    # Cobble K Beta
    DataIndividual[i, 2] <- mean(Temp[Start:End, 2]) * CobbleDoseRate[1, 1]
    DataIndividual[i, 3] <-
      DataIndividual[i, 2] * sqrt(DistanceError[i, 1] ^ 2 + ThicknessError[i, 1] ^
                                    2 + (CobbleDoseRate[1, 2] / CobbleDoseRate[1, 1]) ^ 2)
    # Cobble Th Beta
    DataIndividual[i, 4] <- mean(Temp[Start:End, 3]) * CobbleDoseRate[1, 3]
    DataIndividual[i, 5] <-
      DataIndividual[i, 4] * sqrt(DistanceError[i, 2] ^ 2 + ThicknessError[i, 2] ^
                                    2 + (CobbleDoseRate[1, 4] / CobbleDoseRate[1, 3]) ^ 2)
    # Cobble U Beta
    DataIndividual[i, 6] <- mean(Temp[Start:End, 4]) * CobbleDoseRate[1, 5]
    DataIndividual[i, 7] <- DataIndividual[i, 6] * sqrt(DistanceError[i, 3] ^ 2 + ThicknessError[i, 3] ^
                                                          2 + (CobbleDoseRate[1, 6] / CobbleDoseRate[1, 5]) ^ 2)
    # Cobble K Gamma
    DataIndividual[i, 8] <- mean(Temp[Start:End, 5]) * CobbleDoseRate[2, 1]
    DataIndividual[i, 9] <- DataIndividual[i, 8] * sqrt(DistanceError[i, 4] ^ 2 + ThicknessError[i, 4] ^
                                                          2 + (CobbleDoseRate[2, 2] / CobbleDoseRate[2, 1]) ^ 2)
    # Cobble Th Gamma
    DataIndividual[i, 10] <- mean(Temp[Start:End, 5]) * CobbleDoseRate[2, 3]
    DataIndividual[i, 11] <-
      DataIndividual[i, 10] * sqrt(DistanceError[i, 4] ^ 2 + ThicknessError[i, 4] ^
                                     2 + (CobbleDoseRate[2, 4] / CobbleDoseRate[2, 3]) ^ 2)
    # Cobble U Gamma
    DataIndividual[i, 12] <- mean(Temp[Start:End, 5]) * CobbleDoseRate[2, 5]
    DataIndividual[i, 13] <-
      DataIndividual[i, 12] * sqrt(DistanceError[i, 4] ^ 2 + ThicknessError[i, 4] ^
                                     2 + (CobbleDoseRate[2, 6] / CobbleDoseRate[2, 5]) ^ 2)

    # Sediment K Beta
    DataIndividual[i, 14] <- mean(Temp[Start:End, 6]) * SedDoseRate[1, 1]
    DataIndividual[i, 15] <-
      DataIndividual[i, 14] * sqrt(DistanceError[i, 5] ^ 2 + ThicknessError[i, 5] ^
                                     2 + (SedDoseRate[1, 2] / SedDoseRate[1, 1]) ^ 2)
    # Sediment Th Beta
    DataIndividual[i, 16] <- mean(Temp[Start:End, 7]) * SedDoseRate[1, 3]
    DataIndividual[i, 17] <-
      DataIndividual[i, 16] * sqrt(DistanceError[i, 6] ^ 2 + ThicknessError[i, 6] ^
                                     2 + (SedDoseRate[1, 4] / SedDoseRate[1, 3]) ^ 2)
    # Sediment U Beta
    DataIndividual[i, 18] <- mean(Temp[Start:End, 8]) * SedDoseRate[1, 5]
    DataIndividual[i, 19] <-
      DataIndividual[i, 18] * sqrt(DistanceError[i, 7] ^ 2 + ThicknessError[i, 7] ^
                                     2 + (SedDoseRate[1, 6] / SedDoseRate[1, 5]) ^ 2)
    # Sediment K Gamma
    DataIndividual[i, 20] <- mean(Temp[Start:End, 9]) * SedDoseRate[2, 1]
    DataIndividual[i, 21] <-
      DataIndividual[i, 20] * sqrt(DistanceError[i, 8] ^ 2 + ThicknessError[i, 8] ^
                                     2 + (SedDoseRate[2, 2] / SedDoseRate[2, 1]) ^ 2)
    # Sediment Th Gamma
    DataIndividual[i, 22] <- mean(Temp[Start:End, 9]) * SedDoseRate[2, 3]
    DataIndividual[i, 23] <-
      DataIndividual[i, 22] * sqrt(DistanceError[i, 8] ^ 2 + ThicknessError[i, 8] ^
                                     2 + (SedDoseRate[2, 4] / SedDoseRate[2, 3]) ^ 2)
    # Sediment U Gamma
    DataIndividual[i, 24] <- mean(Temp[Start:End, 9]) * SedDoseRate[2, 5]
    DataIndividual[i, 25] <-
      DataIndividual[i, 24] * sqrt(DistanceError[i, 8] ^ 2 + ThicknessError[i, 8] ^
                                     2 + (SedDoseRate[2, 6] / SedDoseRate[2, 5]) ^ 2)

    ### Sum data into beta and gamma dose rates from cobble and sediment ----
    DataComponent[i, 1] <- input[i, 1]
    DataComponent[i, 2] <- DataIndividual[i, 2] + DataIndividual[i, 4] + DataIndividual[i, 6]
    DataComponent[i, 3] <- DataComponent[i,2]*sqrt((DataIndividual[i,3]/DataIndividual[i,2])^2+(DataIndividual[i,5]/DataIndividual[i,4])^2+(DataIndividual[i,7]/DataIndividual[i,6])^2)
    DataComponent[i, 4] <- DataIndividual[i, 8] + DataIndividual[i, 10] + DataIndividual[i, 12]
    DataComponent[i, 5] <- DataComponent[i,4]*sqrt((DataIndividual[i,9]/DataIndividual[i,8])^2+(DataIndividual[i,11]/DataIndividual[i,10])^2+(DataIndividual[i,13]/DataIndividual[i,12])^2)

    DataComponent[i, 6] <- DataIndividual[i, 14] + DataIndividual[i, 16] + DataIndividual[i, 18]
    DataComponent[i, 7] <- DataComponent[i,6]*sqrt((DataIndividual[i,15]/DataIndividual[i,14])^2+(DataIndividual[i,17]/DataIndividual[i,16])^2+(DataIndividual[i,19]/DataIndividual[i,18])^2)
    DataComponent[i, 8] <- DataIndividual[i, 20] + DataIndividual[i, 22] + DataIndividual[i, 24]
    DataComponent[i, 9] <- DataComponent[i,8]*sqrt((DataIndividual[i,21]/DataIndividual[i,20])^2+(DataIndividual[i,23]/DataIndividual[i,22])^2 + (DataIndividual[i,25]/DataIndividual[i,24])^2)
  }

  colnames(DataIndividual) <-
    c(
      "Distance.",
      "K Beta cobble",
      "SE",
      "Th Beta cobble",
      "SE",
      "U Beta cobble",
      "SE",
      "K Gamma cobble",
      "SE",
      "Th Gamma cobble",
      "SE",
      "U Gamma cobble",
      "SE",
      "K Beta sed.",
      "SE",
      "Th Beta sed.",
      "SE",
      "U Beta sed.",
      "SE",
      "K Gamma sed.",
      "SE",
      "Th Gamma sed.",
      "SE",
      "U Gamma sed.",
      "SE"
    )

  colnames(DataComponent) <-
    c(
      "Distance",
      "Total Cobble Beta",
      "SE",
      "Total Cobble Gamma",
      "SE",
      "Total Beta Sed.",
      "SE",
      "Total Gamma Sed.",
      "SE"
    )

  DataIndividual[is.na(DataIndividual)] <- 0
  DataComponent[is.na(DataComponent)] <- 0

  # Return ------------------------------------------------------------------
  set_RLum(
      class = "RLum.Results",
      data = list(
        DataIndividual = DataIndividual,
        DataComponent = DataComponent,
        input = input
      ),
      info = list(
        call = sys.call()
      ))
}

test_that("basic checks", {
  testthat::skip_on_cran()
  local_edition(3)

  ## simple run with example data
  data("ExampleData.CobbleData", envir = environment())
  expect_s4_class(calc_CobbleDoseRate(ExampleData.CobbleData), "RLum.Results")

  ## break the function
  df <- ExampleData.CobbleData
  df$Distance[[14]] <- 50000
  expect_error(calc_CobbleDoseRate(df),
               "Slices outside of cobble. Please check your distances and make sure they are in mm and diameter is in cm!")
})

## load data
data(ExampleData.CobbleData, envir = environment())
df <- ExampleData.CobbleData

test_that("input validation", {
  testthat::skip_on_cran()

  df$Distance[[14]] <- 50000
  expect_error(calc_CobbleDoseRate("error"),
               "'object' should be of class 'data.frame'")
  expect_error(calc_CobbleDoseRate(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(calc_CobbleDoseRate(iris),
               "'object' doesn't contain the following columns:")
  expect_error(calc_CobbleDoseRate(df),
               "Slices outside of cobble: please ensure your distances are in mm")

  expect_error(calc_CobbleDoseRate(ExampleData.CobbleData, conversion = "error"),
               "'conversion' should be one of 'Guerinetal2011', 'Cresswelletal2018'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(calc_CobbleDoseRate(df),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(calc_CobbleDoseRate(df, conv = "Cresswelletal2018"),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(calc_CobbleDoseRate(df, conv = "Liritzisetal2013"),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(calc_CobbleDoseRate(df, conv = "AdamiecAitken1998"),
                       tolerance = snapshot.tolerance)
})

test_that("more coverage", {
  testthat::skip_on_cran()

  df$CobbleDiameter[1] <- 25
  expect_s4_class(calc_CobbleDoseRate(df),
                  "RLum.Results")

  df$DistanceError[1] <- 0
  df$CobbleDiameter[1] <- 2
  df$Thickness <- 10
  df <- df[1:9, ]
  expect_s4_class(calc_CobbleDoseRate(df),
                  "RLum.Results")

  ## deprecated argument
  expect_warning(calc_CobbleDoseRate(input = df),
                 "'input' was deprecated in v1.2.0, use 'object' instead")
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1424
  df <- data.frame(Distance = 0, DistanceError = 0.01, Thickness = 0.6,
                   ThicknessError = 0.01, Mineral = "Q", Cobble_K = 0.008,
                   Cobble_K_SE = 0.0008, Cobble_Th = 0.62, Cobble_Th_SE = 0.06,
                   Cobble_U = 0.2, Cobble_U_SE = 0.02, GrainSize = 9000,
                   Density = 2.7, CobbleDiameter = 2.6, Sed_K = 0.09,
                   Sed_K_SE = 0.009, Sed_Th = 2.4, Sed_Th_SE = 0.2,
                   Sed_U = 0.75, Sed_U_SE = 0.07, GrainSize_Sed = 200L,
                   Density_Sed = 1.8, WaterContent = 8.2, WaterContent_SE = 4.1)
  expect_error(calc_CobbleDoseRate(df, conversion = "Cresswelletal2018"),
               "No attenuation data available for the grain size provided (9000)",
               fixed = TRUE)
})

#' @title Import SUERC portable OSL Reader PSL files into R
#'
#' @description
#' This function provides an import routine for the SUERC portable OSL Reader
#' PSL format (measurement data and sequence). PSL files are just plain text
#' and can be viewed with any text editor. Due to the formatting of PSL files,
#' this import function relies heavily on regular expression to find and
#' extract all relevant information. See **note**.
#'
#' @param file [character] (**required**):
#' name of one or multiple PSL files (URLs are supported); it can be the path
#' to a directory, in which case the function tries to detect and import all
#' PSL files found in the directory.
#'
#' @param drop_bg [logical] (*with default*):
#' `TRUE` to automatically remove all non-OSL/IRSL curves.
#'
#' @param as_decay_curve  [logical] (*with default*):
#' Portable OSL Reader curves are often given as cumulative light sum curves.
#' Use `TRUE` (default) to convert the curves to the more usual decay form.
#'
#' @param smooth [logical] (*with default*):
#' `TRUE` to apply Tukey's Running Median Smoothing for OSL and IRSL decay curves.
#' Smoothing is encouraged if you see random signal drops within the decay curves related
#' to hardware errors.
#'
#' @param merge [logical] (*with default*):
#' `TRUE` to merge all [Luminescence::RLum.Analysis-class] objects. Only applicable if multiple
#' files are imported.
#'
#' @param pattern [character] (*with default*):
#' regular expression pattern passed to [list.files] to construct a list of
#' files to read (used only when a path is provided).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... currently not used.
#'
#' @return
#' Returns an S4 [Luminescence::RLum.Analysis-class] object containing
#' [Luminescence::RLum.Data.Curve-class] objects for each curve. Results are
#' returned as a list when multiple files are processed or `file` is a list.
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Curve-class]
#'
#' @author Christoph Burow, University of Cologne (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#'
#' @section Function version: 0.1.2
#'
#' @note
#' Because this function relies heavily on regular expressions to parse
#' PSL files it is currently only in beta status. If the routine fails to import
#' a specific PSL file please report to `<christoph.burow@@gmx.net>` so the
#' function can be updated.
#'
#' @keywords IO
#'
#' @examples
#'
#' # (1) Import PSL file to R
#'
#' file <- system.file("extdata", "DorNie_0016.psl", package = "Luminescence")
#' psl <- read_PSL2R(file, drop_bg = FALSE, as_decay_curve = TRUE, smooth = TRUE, merge = FALSE)
#' print(str(psl, max.level = 3))
#' plot(psl, combine = TRUE)
#'
#' @export
read_PSL2R <- function(
  file,
  drop_bg = FALSE,
  as_decay_curve = TRUE,
  smooth = FALSE,
  merge = FALSE,
  pattern = "\\.psl$",
  verbose = TRUE,
  ...
) {
  .set_function_name("read_PSL2R")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_logical_scalar(verbose)
  .validate_class(pattern, "character", length = 1)
  file <- unlist(.validate_file(file, ext = "psl", pattern = pattern, verbose = verbose))
  if (length(file) == 0) {
    return(NULL)
  }
  if (!all(file.exists(file)))
    .throw_error("The following files do not exist, please check:\n",
                 paste(file[!file.exists(file)], collapse = "\n"))

  ## MAIN ----
  results <- vector("list", length(file))
  for (i in 1:length(file)) {

    if (verbose) {
      cat("\n[read_PSL2R()] Importing ...")
      cat("\n path: ", dirname(file[i]))
      cat("\n file: ", .shorten_filename(basename(file[i])))
      cat("\n")
    }

    ## Read in file ----
    doc <- readLines(file[i])

    ## Document formatting ----
    # remove lines with i) blanks only, ii) dashes, iii) equal signs
    doc <- gsub("^[ ]*$", "", doc)
    doc <- gsub("^[ -]*$", "", doc)
    doc <- gsub("^[ =]*$", "", doc)

    # the header ends with date and time with the previous line starting with a single slash
    lines_with_slashes <- doc[grepl("\\", doc, fixed = TRUE)]

    ## OFFENDING LINE: this deletes the line with sample name and time and date
    sample_and_date <- lines_with_slashes[length(lines_with_slashes)]
    sample_date_split <- strsplit(sample_and_date, "@", fixed = TRUE)[[1]]
    sample <- trimws(gsub("\\", "", sample_date_split[1], fixed = TRUE))
    date_and_time <- strsplit(sample_date_split[2], " ")[[1]]
    date_and_time_clean <- date_and_time[nzchar(date_and_time) &
                                         !date_and_time %in% c("/", "AM", "PM")]
    date <- as.Date(date_and_time_clean[1], "%m/%d/%Y")
    time <- format(date_and_time_clean[2], format = "%h:%M:%S")
    doc <- gsub(lines_with_slashes[length(lines_with_slashes)],
                "", fixed = TRUE, doc)

    # last delimiting line before measurements are only apostrophes and dashes
    lines_with_apostrophes <-doc[grepl("'", doc, fixed = TRUE)]
    doc <- gsub(lines_with_apostrophes[length(lines_with_apostrophes)],
                "", fixed = TRUE, doc)

    # finally remove all empty lines
    doc <- doc[nzchar(doc)]

    ## Split document ----
    begin_of_measurements <- grep("Measurement :", doc, fixed = TRUE)
    number_of_measurements <- length(begin_of_measurements)

    # Parse and format header
    header <- doc[1:(begin_of_measurements[1]-1)]
    header <- format_Header(header)

    # add sample name, date and time to header list
    header$Date <- date
    header$Time <- time
    header$Sample <- sample

    # Parse and format the measurement values
    measurements_split <- vector("list", number_of_measurements)

    # save lines of each measurement to individual list elements
    for (j in seq_len(number_of_measurements)) {
      last <- min(begin_of_measurements[j + 1] - 1, length(doc), na.rm = TRUE)
      measurements_split[[j]] <- doc[begin_of_measurements[j]:last]
    }

    # format each measurement; this will return a list of RLum.Data.Curve objects
    measurements_formatted <- lapply(measurements_split, format_Measurements,
                                     convert = as_decay_curve, header = header)

    # drop dark count measurements if needed
    if (drop_bg) {
      measurements_formatted <- lapply(measurements_formatted, function(x) {
        if (x@recordType != "USER")
          return(x)
      })
      measurements_formatted <- .rm_NULL_elements(measurements_formatted)
    }

    # decay curve smoothing using Tukey's Running Median Smoothing (?smooth)
    if (smooth) {
      measurements_formatted <- lapply(measurements_formatted, function(x) {
        if (x@recordType != "USER")
          x@data[, 2] <- stats::smooth(x@data[, 2])
        return(x)
      })
    }

    ## get measurement sequence
    measurement_sequence <- data.table::rbindlist(
      lapply(seq_along(measurements_split), function(x) {
      ## remove measurement
      tmp <- gsub(
        pattern = "Measurement : ",
        replacement = "",
        x = measurements_split[[x]][1],
        fixed = TRUE)

    ## split entries
     tmp <- strsplit(x = tmp, split = " | ", fixed = TRUE)[[1]]

     ## data.frame
     data.frame(
       RUN = x,
       NAME = trimws(tmp[1]),
       STIM = strsplit(tmp[2], split = " ", fixed = TRUE)[[1]][2],
       ON_OFF = strsplit(tmp[3], split = "(us)", fixed = TRUE)[[1]][2],
       CYCLE = strsplit(tmp[4], split = "(ms),", fixed = TRUE)[[1]][2])
    }))

    ## RETURN ----
    results[[i]] <- set_RLum(
      "RLum.Analysis",
       protocol = "portable OSL",
       info = c(
         header,
         list(Sequence = as.data.frame(measurement_sequence))),
       records = measurements_formatted)
  }#Eof::Loop

  ## MERGE ----
  if (length(results) > 1 && merge)
    results <- merge_RLum(results)

  ## RETURN ----
  if (length(results) == 1)
    results <- results[[1]]

  return(results)
}

################################################################################
## HELPER FUNCTIONS
################################################################################


## ------------------------- FORMAT MEASUREMENT ----------------------------- ##
format_Measurements <- function(x, convert, header) {
  ## measurement parameters are given in the first line
  settings <- x[1]
  settings_split <- trimws(unlist(strsplit(settings, "|", fixed = TRUE)))

  # welcome to regex/strsplit hell
  grepvf <- function(pattern, x) grep(pattern, x, value = TRUE, fixed = TRUE)
  settings_measurement <- gsub(".*: ", "", grepvf("Measure", settings_split))
  settings_stim_unit <- gsub("[^0-9]", "", grepvf("Stim", settings_split))
  recordType <- switch(settings_stim_unit,
                       "0" = "USER", "1" = "IRSL", "2" = "OSL")
  on_off_times <- strsplit(gsub("[^0-9,]", "", grepvf("Off", settings_split)), ",")[[1]]
  settings_on_time <- as.integer(on_off_times[1])
  settings_off_time <- as.integer(on_off_times[2])
  vals <- stats::na.omit(as.integer(strsplit(
                     gsub("[^0-9,]", "", grepvf("No", settings_split)), ",")[[1]]))
  settings_cycle <- vals[1]
  settings_stimulation_time <- vals[2]

  settings_list <- list(measurement = settings_measurement,
                        stimulation_unit = recordType,
                        on_time = settings_on_time,
                        off_time = settings_off_time,
                        cycle = settings_cycle,
                        stimulation_time = settings_stimulation_time)

  ## terminal counts are given in the last line as count / count_error
  # terminal_count_text_formatted <- gsub("[^0-9]", "",
  #                                       unlist(strsplit(x[length(x)], "/")))

  ## parse values and create a data frame
  x_stripped <- x[-c(1, 2, length(x))]
  df <- data.frame(matrix(NA, ncol = 5, nrow = length(x_stripped)))
  for (i in 1:length(x_stripped)) {
    x_split <- unlist(strsplit(x_stripped[i], " ", fixed = TRUE))
    x_split <- x_split[nzchar(x_split)]
    x_split_clean <- gsub("[^0-9\\-]", "", x_split)
    x_split_cleaner <- x_split_clean[x_split_clean != "-"]

    df[i, ] <- as.numeric(x_split_cleaner)
  }

  names(df) <- c("time", "counts", "counts_error",
                 "counts_per_cycle", "counts_per_cycle_error")

  # shape of the curve: decay or cumulative
  if (convert)
    data <- matrix(c(df$time, df$counts_per_cycle), ncol = 2)
  else
    data <- matrix(c(df$time, df$counts), ncol = 2)

  set_RLum(
    class = "RLum.Data.Curve",
    originator = "read_PSL2R",
    recordType = paste(recordType, "(PMT)"),
    curveType = "measured",
    data = data,
    info = list(settings = c(settings_list, header),
    raw_data = df))
}

## ---------------------------- FORMAT HEADER ------------------------------- ##
format_Header <- function(x) {
  # split by double blanks
  header_split <- strsplit(x, "  ", fixed = TRUE)

  # check whether there are twice as many values
  # as colons; if there is an equal amount, the previous split was not sufficient
  # and we need to further split by a colon (that is followed by a blank)
  header_split_clean <- lapply(header_split, function(x) {
    x <- x[nzchar(x)]
    n_elements <- length(x)
    n_properties <- length(grep(":", x, fixed = TRUE))

    if (n_elements == n_properties)
      x <- unlist(strsplit(x, ": ", fixed = TRUE))

    return(x)
  })

  # format parameter/settings names and corresponding values
  dt <- rbindlist(lapply(header_split_clean, function(header) {
    idx <- seq(1, length(header), 2)
    data.frame(names = header[idx], values = header[idx + 1])
  }))

  # some RegExing for nice reading
  names <- gsub(":$", "", dt$names, perl = TRUE) |> trimws()
  # for some weird reason "offset subtract" starts with '256 '
  names <- gsub("256 ", "", names, fixed = TRUE)
  # finally, replace all blanks with underscores
  names <- gsub(" ", "_", names, fixed = TRUE)

  values <- gsub(":$", "", dt$values, perl = TRUE) |> trimws()

  # return header as list
  header <- as.list(values)
  names(header) <- names

  return(header)
}

## input file
psl.file <- system.file("extdata/DorNie_0016.psl", package = "Luminescence")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_PSL2R(data.frame()),
               "'file' should be of class 'character'")
  expect_error(read_PSL2R(character(0)),
               "'file' cannot be an empty character")

  ## directory given (assumes that we have a .psl file under inst/extdata)
  SW({
  expect_message(
    read_PSL2R(file = system.file("extdata", package = "Luminescence"),
               verbose = TRUE),
    "Directory detected, looking for")
  })
  expect_silent(
    read_PSL2R(file = system.file("extdata", package = "Luminescence"),
               verbose = FALSE))
  expect_error(read_PSL2R(system.file("extdata", package = "Luminescence"),
                          pattern = NA),
               "'pattern' should be of class 'character' and have length 1")

  ## single file with no .psl extension
  expect_error(
    read_PSL2R(file = system.file("extdata/RF_file.rf",
                                  package = "Luminescence")),
    "File extension 'rf' is not supported, only 'psl' is valid")
  expect_message(expect_message(
      expect_null(read_PSL2R(system.file("extdata", package = "Luminescence"),
                             pattern = "error")),
      "Directory detected, looking for 'error' files"),
      "No files matching the given pattern found in directory")

  ## list of files with a non-existent file
  expect_error(
    read_PSL2R(file = c(psl.file, "non-existent")),
    "The following files do not exist"
  )
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  SW({
  expect_snapshot_RLum(read_PSL2R(psl.file))

  ## custom values (all inverted), multiple files given to test merge = TRUE
  expect_snapshot_RLum(read_PSL2R(c(psl.file, psl.file), drop_bg = TRUE,
                                  as_decay_curve = FALSE, smooth = TRUE,
                                  merge = TRUE, verbose = FALSE))
  })
})

#' @title Merge function for RLum.Results S4-class objects
#'
#' @description Function merges objects of class [Luminescence::RLum.Results-class]. The slots in the objects
#' are combined depending on the object type, e.g., for [data.frame] and [matrix]
#' rows are appended.
#'
#' @details Elements are appended where possible and attributes are preserved if
#' not of similar name as the default attributes of, e.g., a [data.frame]
#'
#' @note The `originator` is taken from the first element and not reset to [Luminescence::merge_RLum]
#'
#' @param objects [list] (**required**):
#' a list of [Luminescence::RLum.Results-class] objects
#'
#' @section Function version: 0.2.1
#'
#' @keywords internal
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @export
merge_RLum.Results <- function(
  objects
) {
  .set_function_name("merge_RLum.Results")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(objects, "list")

  if (length(objects) == 0) {
    .throw_message("'objects' contains no data, NULL returned")
    return(NULL)
  }

  ## check if objects in the list are of type RLum.Results
  temp.originator <- sapply(objects, function(x) {
    .validate_class(x, "RLum.Results", name = "All elements of 'object'")
    x@originator
  })

  ## check if there are multiple originators
  if (length(unique(temp.originator)) > 1) {
    .throw_error("Objects cannot be merged, different originators found: ",
                 .collapse(temp.originator))
  }

            ##-------------------------------------------------------------
            ##merge objects depending on the data structure
            for (i in seq_along(objects[[1]]@data)) {

              ## shelf list of attributes
              attr_list <- unlist(
                lapply(objects, function(x) attributes(x@data[[i]])),
                recursive = FALSE)

              ##++++++++++++++++++++++++++++++++++++++++++++++++++++++++
              ##numeric vector or data.frame or matrix
              if (inherits(objects[[1]]@data[[i]],
                           c("data.frame", "numeric", "matrix"))) {

                ##grep elements and combine them into a list
                temp.list <- lapply(objects, function(x) x@data[[i]])

                ##check whether the objects can be combined by rbind
                if(length(unique(unlist(lapply(temp.list, FUN = ncol)))) > 1)
                  .throw_error("Objects cannot be merged, different number of columns")

                ##combine them using rbind or data.table::rbindList (depends on the data type)
                if (inherits(objects[[1]]@data[[i]], "numeric")) {
                  objects[[1]]@data[[i]] <- unlist(temp.list)

                } else if (inherits(objects[[1]]@data[[i]], "matrix")) {
                  objects[[1]]@data[[i]] <- do.call("rbind", temp.list)

                }else{
                  objects[[1]]@data[[i]] <- as.data.frame(data.table::rbindlist(temp.list))
                }

                ## continue attribute preservation
                ## remove attributes that stem from the object itself
                attr_list[names(attr_list) %in% names(attributes(objects[[1]]@data[[i]]))] <- NULL

                ## just to avoid working the code if not needed
                if(length(attr_list) > 0) {
                  ## merge attributes with similar name
                  attrs <- lapply(unique(names(attr_list)), function(x){
                    tmp <- unlist(attr_list[names(attr_list)%in%x], recursive = FALSE)
                    names(tmp) <- NULL
                    tmp
                  })
                  names(attrs) <- unique(names(attr_list))

                  # set attributes ... we try because some attributes
                  for(n in names(attrs))
                    attr(objects[[1]]@data[[i]], n) <- attrs[[n]]
                }

              }else{
                ##all other elements
                ##grep elements and write them into a list
                objects[[1]]@data[[i]] <- lapply(objects, function(x) x@data[[i]])

                ##unlist to flatten list if necessary for the elements
                if (inherits(objects[[1]]@data[[i]][[1]], "list")) {
                  objects[[1]]@data[[i]] <- unlist(objects[[1]]@data[[i]],
                                                       recursive = FALSE)
                }
              }

            }##end loop

  ## return by setting a new RLum.Results (for the .uid)
  ## the originator is not reset
  set_RLum(
      class = "RLum.Results",
      originator = objects[[1]]@originator,
      data = objects[[1]]@data,
      info = unlist(lapply(objects, function(x) x@info), recursive = FALSE),
      .pid = unlist(lapply(objects, function(x) x@.uid))
  )
}

#'@title Calculate Global Standardised Growth Curve (gSGC) for Feldspar MET-pIRIR
#'
#'@description Implementation of the gSGC approach for feldspar MET-pIRIR by Li et al. (2015)
#'
#'@details ##TODO
#'
#'@param data [data.frame] (**required**): data frame with five columns per sample
#'`c("LnTn", "LnTn.error", "Lr1Tr1", "Lr1Tr1.error","Dr1")`
#'
#'@param gSGC.type [character] (*with default*): growth curve type to be selected
#'according to Table 3 in Li et al. (2015). Allowed options are
#'`"50LxTx"`, `"50Lx"`, `"50Tx"`, `"100LxTx"`, `"100Lx"`, `"100Tx"`, `"150LxTx"`,
#' `"150Lx"`, `"150Tx"`, `"200LxTx"`, `"200Lx"`, `"200Tx"`, `"250LxTx"`, `"250Lx"`,
#' `"250Tx"`
#'
#'@param gSGC.parameters [data.frame] (*optional*): an own parameter set for the
#'gSGC with the following columns `y1`, `y1_err`, `D1`
#'`D1_err`, `y2`, `y2_err`, `y0`, `y0_err`.
#'
#'@param n.MC [numeric] (*with default*): number of Monte-Carlo runs for the
#'error calculation
#'
#'@param plot [logical] (*with default*): enable/disable the plot output.
#'
#'@return Returns an S4 object of type [Luminescence::RLum.Results-class].
#'
#' **`@data`**\cr
#' `$ df` ([data.frame]) \cr
#'  `.. $DE` the calculated equivalent dose\cr
#'  `.. $DE.ERROR` error on the equivalent dose, which is the standard deviation of the MC runs\cr
#'  `.. $HPD95_LOWER` lower boundary of the highest probability density (95%)\cr
#'  `.. $HPD95_UPPER` upper boundary of the highest probability density (95%)\cr
#' `$ m.MC` ([list]) numeric vector with results from the MC runs.\cr
#'
#' **`@info`**\cr
#' `$ call`` ([call]) the original function call
#'
#' @section Function version: 0.1.1
#'
#' @author Harrison Gray, USGS (United States),
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Results-class], [Luminescence::get_RLum], [uniroot], [Luminescence::calc_gSGC]
#'
#' @references Li, B., Roberts, R.G., Jacobs, Z., Li, S.-H., Guo, Y.-J., 2015.
#' Construction of a “global standardised growth curve” (gSGC) for infrared
#' stimulated luminescence dating of K-feldspar 27, 119–130. \doi{10.1016/j.quageo.2015.02.010}
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##test on a generated random sample
#' n_samples <- 10
#' data <- data.frame(
#'   LnTn = rnorm(n=n_samples, mean=1.0, sd=0.02),
#'   LnTn.error = rnorm(n=n_samples, mean=0.05, sd=0.002),
#'   Lr1Tr1 = rnorm(n=n_samples, mean=1.0, sd=0.02),
#'   Lr1Tr1.error = rnorm(n=n_samples, mean=0.05, sd=0.002),
#'   Dr1 = rep(100,n_samples))
#'
#' results <- calc_gSGC_feldspar(
#'   data = data, gSGC.type = "50LxTx",
#'   plot = FALSE)
#'
#' plot_AbanicoPlot(results)
#'
#'@export
calc_gSGC_feldspar <- function(
  data,
  gSGC.type = "50LxTx",
  gSGC.parameters = NULL,
  n.MC = 100,
  plot = FALSE
) {
  .set_function_name("calc_gSGC_feldspar")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(data, "data.frame")
  .validate_not_empty(data)
  if (ncol(data) != 5) {
    .throw_error("'data' should have 5 columns")
  }
  colnames(data) <- c("LnTn", "LnTn.error", "Lr1Tr1", "Lr1Tr1.error",
                      "Dr1")
  .validate_positive_scalar(n.MC)
  .validate_logical_scalar(plot)

# Parametrize -------------------------------------------------------------
  params <- data.frame( # this is the data from Table 3 of Li et al., 2015
    Type = c("50LxTx", "50Lx", "50Tx", "100LxTx", "100Lx", "100Tx", "150LxTx", "150Lx", "150Tx", "200LxTx", "200Lx", "200Tx", "250LxTx", "250Lx", "250Tx"),
    y1 = c( 0.57, 0.36, 0.2, 0.39, 0.41, 0.28, 0.43, 0.4, 0.31, 0.3, 0.34, 0.37, 0.37, 0.17, 0.48),
    y1_err = c( 0.19, 0.25, 0.24, 0.12, 0.28, 0.22, 0.11, 0.27, 0.33, 0.06, 0.28, 0.28, 0.1, 0.12, 0.37),
    D1 = c( 241, 276, 259, 159, 304, 310, 177, 327, 372, 119, 316, 372, 142, 197, 410),
    D1_err = c( 66, 137, 279, 48, 131, 220, 41, 132, 300, 32, 145, 218, 35, 116, 210),
    y2 = c( 0.88, 1.37, 0.34, 0.91, 1.22, 0.42, 0.88, 1.26, 0.45, 0.95, 1.24, 0.43, 0.74, 1.32, 0.45),
    y2_err = c( 0.15, 0.19, 0.15, 0.1, 0.23, 0.26, 0.09, 0.23, 0.18, 0.05, 0.25, 0.24, 0.09, 0.1, 0.15),
    D2 = c( 1115, 1187, 1462, 741, 1146, 2715, 801, 1157, 2533, 661, 1023, 2792, 545, 830, 2175),
    D2_err = c( 344, 287, 191, 105, 288, 639, 109, 263, 608, 49, 205, 709, 62, 79, 420),
    y0 = c( 0.008, 0.003, 0.685, 0.018, 0.01, 0.64, 0.026, 0.015, 0.61, 0.034, 0.02, 0.573, 0.062, 0.028, 0.455),
    y0_err = c( 0.009, 0.009, 0.014, 0.008, 0.008, 0.015, 0.006, 0.007, 0.014, 0.006, 0.006, 0.013, 0.005, 0.005, 0.011),
    D0_2.3 = c( 2000, 2450, 1420, 1420, 2300, 2900, 1500, 2340, 2880, 1320, 2080, 2980, 1000, 1780, 2500),
    D0_3 = c( 2780, 3280, 2520, 1950, 3100, 4960, 2060, 3130, 4760, 1780, 2800, 5120, 1380, 2360, 4060)
  )
  .validate_args(gSGC.type, params$Type)
  .validate_class(gSGC.parameters, "data.frame", null.ok = TRUE)

  # these are user specified parameters if they so desire
  if (!is.null(gSGC.parameters)) {
    y1 <- gSGC.parameters$y1
    y1_err <- gSGC.parameters$y1_err
    D1 <- gSGC.parameters$D1
    D1_err <- gSGC.parameters$D1_err
    y2 <- gSGC.parameters$y2
    y2_err <- gSGC.parameters$y2_err
    y0 <- gSGC.parameters$y0
    y0_err <- gSGC.parameters$y0_err

  } else {
   # take the user input pIRSL temperature and assign the correct parameters
   index <- match(gSGC.type,params$Type)

      y1 <- params$y1[index]
      y1_err <- params$y1_err[index]

      D1 <- params$D1[index]
      D1_err <- params$D1_err[index]

      y2 <- params$y2[index]
      y2_err <- params$y2_err[index]

      D2 <- params$D2[index]
      D2_err <- params$D2_err[index]

      y0 <- params$y0[index]
      y0_err <- params$y0_err[index]
  }

  ##set function for uniroot
  ## function from Li et al., 2015 eq: 3
  ## function that equals zero when the correct De is found.
  ## This is so uniroot can find the correct value or 'root'
  f <- function(De, Dr1, Lr1Tr1, LnTn, y1, D1, y2, D2, y0){
    f_D <-  y1 * (1 - exp(-De / D1)) + y2 * (1 - exp(-De / D2)) + y0
    f_Dr <-  y1 * (1 - exp(-Dr1 / D1)) + y2 * (1 - exp(-Dr1 / D2)) + y0
    ##return(f_D/Lr1Tr1 - f_Dr/LnTn) ##TODO double check seems to be wrong
    return(f_Dr/Lr1Tr1 - f_D/LnTn)
  }

# Run calculation ---------------------------------------------------------
  l <- lapply(1:nrow(data), function(i) {
    Lr1Tr1 <- data[i, "Lr1Tr1"] #assign user's input data
    Lr1Tr1.error <- data[i, "Lr1Tr1.error"]
    Dr1 <- data[i, "Dr1"]
    LnTn <- data[i, "LnTn"]
    LnTn.error <- data[i, "LnTn.error"]

    ## uniroot solution
    temp <- try({
        uniroot(
          f,
          interval = c(0.1, 3000),
          tol = 0.001,
          Dr1 = Dr1,
          Lr1Tr1 = Lr1Tr1,
          LnTn = LnTn,
          y1 = y1,
          D1 = D1,
          y2 = y2,
          D2 = D2,
          y0 = y0,
          extendInt = "yes",
          check.conv = TRUE,
          maxiter = 1000)
       }, silent = TRUE) # solve for the correct De

      ## in case the initial uniroot solve does not work
      if(inherits(temp, "try-error")) {
        .throw_message("No solution found for dataset #", i, ", NA returned")
        return(NA)
      }

      De <- temp$root
      temp.MC.matrix <- matrix(nrow = n.MC, ncol = 8)

        # to estimate the error, use a monte carlo simulation. assume error in input data is gaussian
        # create a matrix
        colnames(temp.MC.matrix) <- c("LnTn", "Lr1Tr1","y1", "D1", "y2", "D2", "y0", "De")

        # simulate random values for each parameter
        temp.MC.matrix[, 1:7] <- matrix(
          rnorm(n.MC * 7,
            mean = c(LnTn, Lr1Tr1, y1, D1, y2, D2, y0),
            sd = c(LnTn.error, Lr1Tr1.error, y1_err, D1_err, y2_err, D2_err, y0_err)),
          ncol = 7,
          byrow = TRUE)

        # now use the randomly generated parameters to calculate De's with uniroot
        for (j in 1:n.MC){
          temp2 <- try({
              uniroot(
                f,
                interval = c(0.1, 3000),
                tol = 0.001,
                LnTn = temp.MC.matrix[j, 1],
                Lr1Tr1 =  temp.MC.matrix[j, 2],
                y1 = temp.MC.matrix[j, 3],
                D1 = temp.MC.matrix[j, 4],
                y2 = temp.MC.matrix[j, 5],
                D2 = temp.MC.matrix[j, 6],
                y0 = temp.MC.matrix[j, 7],
                Dr1 = Dr1,
                extendInt = "yes",
                check.conv = TRUE,
                maxiter = 1000
              )
            }, silent = TRUE)

          if (!inherits(temp2, "try-error")){
            temp.MC.matrix[j,8] <- temp2$root
          }
        }

        # set the De uncertainty as the standard deviations of the randomly generated des
        De.error <- sd(temp.MC.matrix[, 8], na.rm = TRUE)

        return(list(
          DE = De,
          DE.ERROR = De.error,
          m.MC = temp.MC.matrix))
    })

# Plotting ----------------------------------------------------------------
  if(plot){
    par.defaults <- .par_defaults()
    on.exit(par(par.defaults), add = TRUE)

    par(mfrow = c(mfrow = c(3,3)))
    for (i in 1:length(l)) {
      if (is.na(l[[i]][1])) next

      y_max <- max(l[[i]]$m.MC[, 1:2])
      plot(NA, NA,
           xlab = "Dose [a.u.]",
           ylab = "Norm. Signal",
           xlim = c(0, 3000),
           main = paste0("Dataset #", i),
           ylim = c(0, y_max)
           )

      for(j in 1:nrow(l[[i]]$m.MC)){
        #y1 * (1 - exp(-De / D1)) + y2 * (1 - exp(-De / D2)) + y0
        x <- NA
        curve(
           l[[i]]$m.MC[j, 3] * (1 - exp(-x  / l[[i]]$m.MC[j, 4])) +
           l[[i]]$m.MC[j, 5] * (1 - exp(-x  / l[[i]]$m.MC[j, 6])) +
           l[[i]]$m.MC[j, 7],
        col = rgb(0,0,0,0.4),
        add = TRUE)
      }

       par(new = TRUE)
       hist <- hist(na.exclude(l[[i]]$m.MC[, 8]),
           plot = FALSE
         )

       hist$counts <- ((y_max/max(hist$counts)) * hist$counts) / 2
       plot(
         hist,
         xlab = "",
         ylab = "",
         axes = FALSE,
         xlim = c(0, 3000),
         ylim = c(0, y_max),
         main = ""
       )
    }
  }


# Return ------------------------------------------------------------------

  ##output matrix
  m <- matrix(ncol = 4, nrow = nrow(data))

  ##calculate a few useful parameters
  for(i in 1:nrow(m)){
    if (is.na(l[[i]][1])) next

    m[i,1] <- l[[i]]$DE
    m[i,2] <- l[[i]]$DE.ERROR

    HPD <- .calc_HPDI(na.exclude(l[[i]]$m.MC[,8]))
    m[i,3] <- HPD[1,1]
    m[i,4] <- HPD[1,2]
  }

  df <- data.frame(
    DE = m[, 1],
    DE.ERROR = m[, 2],
    HPD95_LOWER = m[, 3],
    HPD95_UPPER = m[, 4]
  )

  set_RLum("RLum.Results",
      data = list(
        data = df,
        m.MC = lapply(l, function(x) {
          if (is.na(x[[1]])) return(x)
          x$m.MC
        })
      ),
      info = list(
        call = sys.call()
      ))
}

test_that("input validation", {
  testthat::skip_on_cran()

  ##no data.frame
  expect_error(calc_gSGC_feldspar(
    data = "data",
    gSGC.type = "50LxTx",
    plot = FALSE),
    "[calc_gSGC_feldspar()] 'data' should be of class 'data.frame'",
    fixed = TRUE)
  expect_error(calc_gSGC_feldspar(iris[0, ]),
               "'data' cannot be an empty data.frame")

  ## input is somewhat not what we expect for gSGC
  expect_error(
    calc_gSGC_feldspar(
      data = data.frame(a  = 1, b = 1, c = 1, d = 1, e = 1),
      gSGC.type = "wrong",
      plot = FALSE
    ),
    "'gSGC.type' should be one of '50LxTx', '50Lx', '50Tx', '100LxTx', '100Lx'"
  )

  ## incorrect number of columns
  expect_error(
    calc_gSGC_feldspar(
      data = data.frame(a  = 1, b = 1, c = 1, d = 1),
      gSGC.type = "50LxTx",
      plot = FALSE
    ),
    "'data' should have 5 columns"
  )

  expect_error(calc_gSGC_feldspar(iris, n.MC = -1))
  expect_error(calc_gSGC_feldspar(iris, plot = NA))
})

test_that("test functionality", {
  testthat::skip_on_cran()

 ##finally run with plot output
 #test on a generated random sample
 set.seed(1234)
 n_samples <- 2
  data <- data.frame(
    LnTn = rnorm(n=n_samples, mean=1.0, sd=0.02),
    LnTn.error = rnorm(n=n_samples, mean=0.05, sd=0.002),
    Lr1Tr1 = rnorm(n=n_samples, mean=1.0, sd=0.02),
    Lr1Tr1.error = rnorm(n=n_samples, mean=0.05, sd=0.002),
    Dr1 = rep(100,n_samples))

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(calc_gSGC_feldspar(
    data = data,
    gSGC.type = "50LxTx",
    n.MC = 20,
    plot = TRUE),
    tolerance = snapshot.tolerance)

  ## more coverage
  set.seed(733)
  data$LnTn <- rnorm(n = n_samples, mean = 1.0, sd = 0.02)
  expect_snapshot_RLum(calc_gSGC_feldspar(
      data = data, gSGC.type = "50LxTx", n.MC = 10, plot = TRUE),
      tolerance = snapshot.tolerance)

  ## test own curve parameters
  SW({
  expect_message(results <- calc_gSGC_feldspar(
    data = data,
    gSGC.parameters = data.frame(
      y1 = 0.6,
      y1_err = 0.2,
      D1 = 250,
      D1_err = 50,
      y2 = 0.90,
      y2_err = 0.10,
      y0 = 0.001,
      y0_err = 0.0001
    ),
    plot = TRUE),
    "No solution found for dataset")
  })

  ##regression tests
  expect_s4_class(results, "RLum.Results")
  expect_true(all(is.na(unlist(results$m.MC))))
})

#' @title Plot a dose-response curve for luminescence data (Lx/Tx against dose)
#'
#' @description
#'
#' A dose-response curve is produced for luminescence measurements using a
#' regenerative or additive protocol as implemented in [Luminescence::fit_DoseResponseCurve].
#'
#' @param object [Luminescence::RLum.Results-class] (**required**):
#' An object produced by [Luminescence::fit_DoseResponseCurve].
#'
#' @param plot_extended [logical] (*with default*):
#' If `TRUE`, 3 plots on one plot area are provided:
#' 1. growth curve,
#' 2. histogram from Monte Carlo error simulation and
#' 3. a test dose response plot.
#'
#' If `FALSE`, just the growth curve will be plotted.
#'
#' @param plot_singlePanels [logical] (*with default*):
#' single plot output (`TRUE/FALSE`) to allow for plotting the results in
#' single plot windows. Ignored if `plot_extended = FALSE`.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... Further graphical parameters to be passed (supported:
#' `main`, `mtext`, `xlim`, `ylim`, `xlab`, `ylab`, `log`
#' (not valid for objects fitted with `mode = "extrapolation"`), `legend` (`TRUE/FALSE`),
#' `legend.pos`, `reg_points_pch`, `density_polygon` (`TRUE/FALSE`),
#' `density_polygon_col`, `density_rug` (`TRUE`/`FALSE`),
#' `box` (`TRUE`/`FALSE`).
#'
#' @return
#' A plot (or a series of plots) is produced.
#'
#' @section Function version: 1.0.8
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @references
#'
#' Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46.
#'
#' Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
#' Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
#' from feldspar. Radiation Measurements 81, 224-231.
#'
#' Pagonis, V., Kitis, G., Chen, R., 2020. A new analytical equation for the dose response of dosimetric materials,
#' based on the Lambert W function. Journal of Luminescence 225, 117333. \doi{10.1016/j.jlumin.2020.117333}
#'
#' @seealso [Luminescence::fit_DoseResponseCurve]
#'
#' @examples
#'
#' ##(1) plot dose-response curve for a dummy dataset
#' data(ExampleData.LxTxData, envir = environment())
#' fit <- fit_DoseResponseCurve(LxTxData)
#' plot_DoseResponseCurve(fit)
#'
#' ##(1b) horizontal plot arrangement
#' layout(mat = matrix(c(1,1,2,3), ncol = 2))
#' plot_DoseResponseCurve(fit, plot_singlePanels = TRUE)
#'
#' ##(2) plot the dose-response curve with pdf output - uncomment to use
#' ##pdf(file = "~/Dose_Response_Curve_Dummy.pdf", paper = "special")
#' plot_DoseResponseCurve(fit)
#' ##dev.off()
#'
#' ##(3) plot the growth curve with pdf output - uncomment to use, single output
#' ##pdf(file = "~/Dose_Response_Curve_Dummy.pdf", paper = "special")
#' plot_DoseResponseCurve(fit, plot_singlePanels = TRUE)
#' ##dev.off()
#'
#' ##(4) plot resulting function for given interval x
#' x <- seq(1,10000, by = 100)
#' plot(
#'  x = x,
#'  y = eval(fit$Formula),
#'  type = "l"
#' )
#'
#' @export
plot_DoseResponseCurve <- function(
  object,
  plot_extended = TRUE,
  plot_singlePanels = FALSE,
  verbose = TRUE,
  ...
) {
  .set_function_name("plot_DoseResponseCurve")
  on.exit(.unset_function_name(), add = TRUE)

  ## get Luminescence colours
  col <- get("col", pos = .LuminescenceEnv)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Results")
  .validate_logical_scalar(plot_extended)
  .validate_logical_scalar(plot_singlePanels)
  .validate_logical_scalar(verbose)

  ## Fitting arguments ------------------------------------------------------
  fit.args <- object$Fit.Args
  mode <- fit.args$mode
  sample <- fit.args$object

  ## for interpolation the first point is considered as natural dose
  first.idx <- ifelse(mode == "interpolation", 2, 1)
  last.idx <- fit.args$fit.NumberRegPoints + 1
  xy <- sample[first.idx:last.idx, 1:2]
  colnames(xy) <- c("x", "y")
  y.Error <- sample[first.idx:last.idx, 3]

  De <- object@data$De$.De.plot
  x.natural <- na.exclude(object@data$De.MC)
  De.MonteCarlo <- mean(na.exclude(x.natural))
  De.Error <- sd(na.exclude(x.natural))

  ## Graphical arguments ----------------------------------------------------
  ymax <- max(xy$y) + if (max(xy$y) * 0.1 > 1.5) 1.5 else max(xy$y) * 0.2
  ylim <- if (mode == "extrapolation" || fit.args$fit.force_through_origin) {
            c(0 - max(y.Error), ymax)
          } else {
            c(0, ymax)
          }

  xmin <- if (!is.na(De)) min(De * 2, 0) else -min(xy$x) * 2
  xmax <- max(xy$x) + if (max(xy$x) * 0.4 > 50) 50 else max(xy$x) * 0.4
  xlim <- if (mode == "extrapolation") {
            c(xmin, xmax)
          } else {
            c(0, xmax)
          }

  ## set plot settings
  plot_settings <- modifyList(
    x = list(
      main = "Dose-response curve",
      xlab = "Dose [s]",
      ylab = if (mode == "interpolation") expression(L[x]/T[x]) else "Luminescence [a.u.]",
      ylim = ylim,
      xlim = xlim,
      mar = c(4, 3, 3, 1),
      mgp = c(2, 0.7, 0),
      tcl = -0.4,
      cex = 1,
      mtext = if (mode != "alternate")
                substitute(D[e] == De,
                           list(De = sprintf("%.2f \uB1 %.1e  |  fit: %s",
                                             abs(De), De.Error, fit.args$fit.method)))
              else "",
      log = "",
      legend = TRUE,
      legend.pos = if (mode == "interpolation") "topleft" else "bottomright",
      reg_points_pch = c(19, 1, 2), # point, point 0, point repeated
      density_polygon = TRUE,
      density_polygon_col = rgb(1,0,0,0.2),
      density_rug = TRUE,
      box = TRUE),
    val = list(...),
    keep.null = TRUE
  )

  if (plot_settings$log != "") {
    if (mode == "extrapolation") {
      .throw_message("Logarithmic transformation not allowed on an object ",
                     "fitted with mode = 'extrapolation', 'log' reset to ''")
      plot_settings$log <- ""
    } else {
      ## if we want to apply a log-transform on x and the first time point
      ## is 0, we shift the curves by one channel
      if (grepl("x", plot_settings$log) && min(xy$x) == 0) {
        xy$x[xy$x == 0] <- 1
        plot_settings$xlim[1] <- min(xy$x)
      }
      if (grepl("y", plot_settings$log)) {
        plot_settings$ylim[1] <- min(xy$y)
      }
    }
  }

  if (length(plot_settings$reg_points_pch) < 3) {
    plot_settings$reg_points_pch <- c(19, 1, 2)
    .throw_message("'reg_points_pch' should have length 3 (for point, point 0, ",
                   "point repeated), 'reg_points_pch' reset to c(19, 1, 2)")
  }

  ## Main plots -------------------------------------------------------------

  ## open plot area
  if (plot_extended && !plot_singlePanels) {
    par.default <- .par_defaults()
    graphics::layout(matrix(c(1, 1, 1, 1, 2, 3), 3, 2, byrow = TRUE), respect = TRUE)
    par(cex = 0.8 * plot_settings$cex, mar = c(3, 3, 3, 1))
  } else {
    ## only restore those we are changing, to avoid resetting all graphical
    ## parameters if we were restoring also mfrow/mfcol, as that would affect
    ## the plots generated by analyse_SAR.CWOSL() and analyse_pIRIRSequence()
    par.default <- par(c("cex", "mar", "mgp"))
    par(mar = plot_settings$mar)
  }
  on.exit(par(par.default), add = TRUE)
  par(mgp = plot_settings$mgp, tcl = plot_settings$tcl)

  #PLOT		#Plot input values
  ##Make selection to support manual number of reg points input
  plot_check <- try(plot(
      xy[1:fit.args$fit.NumberRegPointsReal, ],
      ylim = plot_settings$ylim,
      xlim = plot_settings$xlim,
      log = plot_settings$log,
      pch = plot_settings$reg_points_pch[1],
      xlab = plot_settings$xlab,
      ylab = plot_settings$ylab,
      frame.plot = plot_settings$box[1]
  ),
  silent = TRUE)

  ## now that we have opened the plot, we can work out the coordinates of
  ## the extremes, applying a log-transformation if necessary
  par.usr <- par("usr")
  if (grepl("x", plot_settings$log)) par.usr[1:2] <- 10^par.usr[1:2]
  if (grepl("y", plot_settings$log)) par.usr[3:4] <- 10^par.usr[3:4]

  if (!inherits(plot_check, "try-error")) {
    if (mode == "extrapolation") {
      abline(v = 0, lty = 1, col = "grey")
      abline(h = 0, lty = 1, col = "grey")
    }

    ### add header
    title(main = plot_settings$main, line = NA)

    ## add curve
    if (inherits(object$Formula, "expression")) {
      ## make sure that we always have a zero: here we operate with the
      ## original par("usr") values, so that in case of log-transformation
      ## the points are still uniformly spaced (#845)
      x <- sort(c(0, seq(par("usr")[1], par("usr")[2], length.out = 100)))
      if (grepl("x", plot_settings$log))
        x <- 10^x

      lines(x, eval(object$Formula))
    }

    ## natural value
    if (mode == "interpolation") {
      if(!plot_settings$density_polygon[1]) {
        points(
          x = sample[1, 1:2],
          col = 2,
          pch = plot_settings$reg_points_pch[1])
        segments(sample[1, 1], sample[1, 2] - sample[1, 3],
                 sample[1, 1], sample[1, 2] + sample[1, 3], col = col[2])
      }
    } else if (mode == "extrapolation"){
      points(
        x = De,
        y = 0,
        bg = col[2],
        pch = 21,
        col = "black",
        cex = 1.1)
    }

    ## repeated Point
    idx.rep <- which(duplicated(xy[, 1]))
    points(
        x = xy[idx.rep, 1],
        y = xy[idx.rep, 2],
        pch = plot_settings$reg_points_pch[3])

       ## LINES	#Insert Ln/Tn
    if (mode == "interpolation") {
      xmax <- if (is.na(De)) max(sample[, 1]) * 2 else De
      try(lines(
          c(par.usr[1], xmax),
          c(sample[1, 2], sample[1, 2]),
          col = col[2],
          lty = 2,
          lwd = 1.25
      ), silent = TRUE)
      try(lines(
          c(De, De),
          c(par.usr[3], sample[1, 2]),
          col = col[2],
          lty = 2,
          lwd = 1.25), silent = TRUE)
      try({
        points(
          x = De,
          y = sample[1, 2],
          col = "black",
          pch = 21,
          bg = col[2],
          cex = 1.1)
        },
        silent = TRUE)

      if (plot_settings$density_polygon[1] & length(x.natural) > 1 &&
          !all(is.na(x.natural))) {
          ##calculate density De.MC
          density_De <- stats::density(x.natural, na.rm = TRUE)

          ##calculate transformation function
          density_De$y <- .rescale(
            x = density_De$y,
            range_old = c(max(density_De$y), min(density_De$y)),
            range_new = c(sample[1, 2] / 2, par.usr[3]))

          ## for De
          ## we add two extra points to ensure that the base is not wonky (#847)
          polygon(
            x = c(density_De$x, tail(density_De$x, 1), density_De$x[1]),
            y = c(density_De$y, min(density_De$y), min(density_De$y)),
            col = plot_settings$density_polygon_col)

          ## for LxTx
          tmp_y <- seq(
            sample[[2]][1] - 5 * sample[[3]][1],
            sample[[2]][1] + 5 *  sample[[3]][1],
            length.out = 100)

          tmp_x <- dnorm(tmp_y, mean = sample[1, 2], sd = sample[1, 3])

          tmp_x <-  .rescale(
            x = tmp_x,
            range_old = c(max(tmp_x), min(tmp_x)),
            range_new = c(sample[[2]][1], par.usr[1]))

          # draw polygon
          polygon(
            x = tmp_x,
            y = tmp_y,
            col = plot_settings$density_polygon_col)

          rm(tmp_x, tmp_y, density_De)
      }

      ## reg Point 0
      idx.0 <- which(xy == 0)
      points(
        x = xy[idx.0, 1],
        y = xy[idx.0, 2],
        pch = plot_settings$reg_points_pch[2],
        cex = 1.5)

      ## y-error bar
      segments(xy$x, xy$y - y.Error, xy$x, xy$y + y.Error)

      if(plot_settings$density_rug[1])
        suppressWarnings(graphics::rug(x = x.natural, side = 3))

    } else if (mode == "extrapolation" && !is.na(De)) {
        abline(v = De, lty = 2, col = col[2])
        lines(x = c(0,De), y = c(0,0), lty = 2, col = col[2])
    }

    ## insert fit and result
    try(mtext(side = 3,
              plot_settings$mtext,
              cex = 0.8 * par("cex")),
        silent = TRUE)

    ## write error message in plot if De is NaN or NA
    try(if (is.na(De) & mode != "alternate") {
          text(sample[2, 1],
              0,
              "Error: De could not be calculated!",
              adj = c(0, 0),
              cex = 0.8,
              col = col[2]
          )
        }, silent = TRUE)

    ## plot legend
    if(plot_settings$legend) {
      legend(
          plot_settings$legend.pos,
          legend = paste(ifelse(mode == "interpolation", "REG", "Dose"),
                         c("point", "point 0", "point repeated")),
          pch = plot_settings$reg_points_pch,
          cex = 0.7,
          bty = "n")
    }

    if (plot_extended) {
      ## decrease spacing between axis labels and plots
      par(mar = if (plot_singlePanels) c(4, 2, 1.5, 1) else c(4, 3.5, 2.5, 1),
          mgp = c(1.2, 0.4, 0), tcl = -0.3)

      ## Histogram ----------------------------------------------------------
      if (!plot_singlePanels)
        par(cex = 0.7 * plot_settings$cex)

      ## calculate histogram data
      try(histogram <- hist(x.natural, plot = FALSE), silent = TRUE)

      ## to avoid errors plot only if histogram exists
      if (exists("histogram") && length(histogram$counts) > 2) {

        ## plot histogram
        histogram <- try(hist(
            x.natural,
            xlab = plot_settings$xlab,
            ylab = "Freq.",
            main = "MC runs",
            freq = FALSE,
            border = "white",
            axes = FALSE,
            sub = paste0("valid fits = ", length(na.exclude(x.natural)),
                         "/", fit.args$n.MC),
            cex.sub = 0.8,
            cex.lab = 0.8,
            cex.main = 0.8,
            col = "grey"
        ), silent = TRUE)

        ## add axes
        if (!inherits(histogram, "try-error")) {
          axis(side = 1, cex.axis = 0.8)
          axis(side = 2, cex.axis = 0.8,
               at = seq(min(histogram$density),
                        max(histogram$density), length = 5),
               labels = round(
                 seq(
                   min(histogram$counts),
                   max(histogram$counts),
                   length = 5),
                 digits = 0))

          ## add norm curve
          x <- seq(par("usr")[1], par("usr")[2], length.out = 100)
          y_curve <- stats::dnorm(x,
            mean = mean(x.natural, na.rm = TRUE),
            sd = sd(x.natural, na.rm = TRUE))

            ## rescale
            y_curve <- .rescale(
              x = y_curve,
              range_old = range(y_curve),
              range_new = c(0,par()$usr[4]))

          ## plot lines
          lines(x, y_curve, col = col[2], lty = 2)

          ## add rug
          suppressWarnings(graphics::rug(x.natural))

          ## add difference text
          legend('topright',
                 legend = paste0(
                  "diff. ",
                  round(
                    x = abs((abs(De) - De.MonteCarlo) / abs(De) * 100),
                    digits = 1),
                 "%"),
                 cex = 0.7,
                 bty = "n")

          ## De + Error from MC simulation + quality of error estimation
          # De label
          De.label <- paste(
            abs(round(De.MonteCarlo, 2)),
            "\u00B1",
            format(De.Error, scientific = De.Error < 0.01, digits = 2))

          # set expression
          De.expr <- substitute(D[e[MC]] == val, list(val = De.label))

          # Safely add the text to the plot
          try(
            mtext(
              side = 3,
              line = ifelse(plot_singlePanels, -0.45, -0.2),
              text = De.expr,
              cex = 0.7 * par("cex")
            ),
            silent = TRUE)
        }

      } else {
        plot_check <- try(plot(NA, NA, xlim = c(0, 10), ylim = c(0, 10),
                               axes = FALSE, xlab = "", ylab = "",
                               main = expression(paste(D[e], " from MC runs"))),
                          silent = TRUE)
        if (!inherits(plot_check,"try-error"))
          text(5, 5, "Not available", xpd = NA)
      }

      ## Test dose response curve if available ------------------------------
      ## plot Tx/Tn value for sensitivity change
      if (!inherits(plot_check, "try-error")) {
        if ("TnTx" %in% colnames(sample)) {
          plot(
              1:length(sample[, "TnTx"]),
              sample[, "TnTx"] / sample[1, "TnTx"],
              xlab = "#SAR-cycle",
              ylab = expression(paste(T[x] / T[n])),
              main = "Sensitivity",
              cex.lab = 0.8,
              cex.main = 0.8,
              cex.axis = 0.8,
              type = "o",
              xpd = NA,
              pch = 20)
          lines(c(0, nrow(sample) + 1), c(1, 1), lty = 2, col = "gray")
        } else {
          plot(NA, NA, xlim = c(0, 10), ylim = c(0, 10),
               axes = FALSE, xlab = "", ylab = "",
               main = "Sensitivity", cex.main = 0.8)
          text(5, 5, "Not available\nNo TnTx column", xpd = NA)
        }
      }
    }
  }

  ##reset graphic device if the plotting failed!
  if (inherits(plot_check, "try-error")) {
    # nocov start
    .throw_message("Figure margins too large, nothing plotted")
    grDevices::dev.off()
    # nocov end
  }

  ## return
  invisible(object)
}

## load data
data(ExampleData.LxTxData, envir = environment())
set.seed(1)
fit <- fit_DoseResponseCurve(LxTxData, verbose = FALSE)
fit.extra.gok <- fit_DoseResponseCurve(LxTxData, mode = "extrapolation",
                                       fit.method = "GOK", verbose = FALSE)
fit.alt <- fit_DoseResponseCurve(LxTxData, mode = "alternate",
                                 verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(
      plot_DoseResponseCurve("error"),
      "[plot_DoseResponseCurve()] 'object' should be of class 'RLum.Results'",
      fixed = TRUE)
  expect_error(
      plot_DoseResponseCurve(fit, plot_extended = "error"),
      "'plot_extended' should be a single logical value")
  expect_error(
      plot_DoseResponseCurve(fit, plot_singlePanels = "error"),
      "'plot_singlePanels' should be a single logical value")
  expect_error(
      plot_DoseResponseCurve(fit, verbose = "error"),
      "'verbose' should be a single logical value")
  expect_message(
      plot_DoseResponseCurve(fit, reg_points_pch = 1),
      "'reg_points_pch' should have length 3")

  ## there's nothing we can do for this, apart adding a useless 'plot' argument
  expect_error(plot_DoseResponseCurve(fit, plot = NA),
               "argument 2 matches multiple formal arguments")
  })

test_that("plot output", {
  testthat::skip_on_cran()

  ## standard return
  expect_s4_class(plot_DoseResponseCurve(fit), "RLum.Results")

  ## check plot settings
  expect_s4_class(plot_DoseResponseCurve(fit, legend = FALSE,
                                         reg_points_pch = c(19, 1, 2),
                                         density_polygon = FALSE,
                                         box = FALSE), "RLum.Results")
  expect_message(plot_DoseResponseCurve(fit.extra.gok, log = "x"),
                 "Logarithmic transformation not allowed on an object fitted with")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              plot_DoseResponseCurve(fit))
  vdiffr::expect_doppelganger("log-xy",
                              plot_DoseResponseCurve(fit, log = "xy"))
  vdiffr::expect_doppelganger("extrapolation-gok",
                              plot_DoseResponseCurve(fit.extra.gok))
  vdiffr::expect_doppelganger("alternate",
                              plot_DoseResponseCurve(fit.alt))
  vdiffr::expect_doppelganger("cex.global",
                              plot_DoseResponseCurve(fit, legend = FALSE,
                                                     density_polygon_col = "azure",
                                                     cex = 2))

  ## De is NA
  df <- data.frame(DOSE = c(0, 5, 10, 20, 30),
                   LxTx = c(10, 5, -20, -30, -40),
                   LxTx_X = c(1, 1, 1, 1, 1))
  vdiffr::expect_doppelganger("De.NA",
                              plot_DoseResponseCurve(fit_DoseResponseCurve(df)))
  })
})

#' @title Simple Subsetting of Single Grain Data from Risø BIN/BINX files
#'
#' @description
#' Most measured single grains do not exhibit light and it usually makes sense
#' to subset single grain datasets using a table of position and grain pairs.
#'
#' @param object [Luminescence::Risoe.BINfileData-class] (**required**):
#' input object with the data to subset.
#'
#' @param selection [data.frame], [matrix] (**required**):
#' selection table with two columns for position and grain, respectively.
#'
#' @return A subset [Luminescence::Risoe.BINfileData-class] object
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@keywords manip datagen
#'
#'@seealso [Luminescence::Risoe.BINfileData-class], [Luminescence::read_BIN2R],
#'[Luminescence::verify_SingleGrainData]
#'
#'@examples
#'
#'## load example data
#'data(ExampleData.BINfileData, envir = environment())
#'
#'## set POSITION/GRAIN pair dataset
#'selection <- data.frame(POSITION = c(1,5,7), GRAIN = c(0,0,0))
#'
#'##subset
#'subset_SingleGrainData(object = CWOSL.SAR.Data, selection = selection)
#'
#'@export
subset_SingleGrainData <- function(
    object,
    selection
) {
  .set_function_name("subset_SingleGrainData")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "Risoe.BINfileData")
  .validate_class(selection, c("data.frame", "matrix"))
  if (ncol(selection) < 2)
    .throw_error("'selection' should have 2 columns")

  ## try to work with selection
  selection <- as.data.frame(selection)[,1:2]
  colnames(selection) <- c("POSITION", "GRAIN")

# Subset ------------------------------------------------------------------
  ## select ids for subsetting
  sel_id <-sort(merge(object@METADATA[,c("POSITION", "GRAIN", "ID")], selection)[["ID"]])
  if (length(sel_id) == 0)
    .throw_error("No matching records, check POSITION and GRAIN in 'selection'")

  ## pick data
  object@METADATA <- object@METADATA[sel_id,]
  object@DATA <- object@DATA[sel_id]
  object@METADATA[["ID"]] <- 1:nrow(object@METADATA)

# Return ------------------------------------------------------------------
  return(object)
}

## load data
data(ExampleData.BINfileData, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(subset_SingleGrainData("error"),
               "'object' should be of class 'Risoe.BINfileData'")
  expect_error(subset_SingleGrainData(CWOSL.SAR.Data, NA),
               "'selection' should be of class 'data.frame' or 'matrix'")
  expect_error(subset_SingleGrainData(CWOSL.SAR.Data, data.frame(1:10, 1)),
               "No matching records, check POSITION and GRAIN in 'selection'")
  expect_error(subset_SingleGrainData(CWOSL.SAR.Data, data.frame(1:10)),
               "'selection' should have 2 columns")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## set POSITION/GRAIN pair dataset
  selection <- data.frame(POSITION = c(1,5,7), GRAIN = c(0,0,0))

  ## standard run
  expect_s4_class(subset_SingleGrainData(object = CWOSL.SAR.Data, selection = selection), "Risoe.BINfileData")

  ## try something different for the input
  expect_s4_class(subset_SingleGrainData(object = CWOSL.SAR.Data, selection = as.matrix(selection)), "Risoe.BINfileData")
})

#' @title Calculate the gamma dose deposited within a sample taking layer-to-layer
#' variations in radioactivity into account (according to Aitken, 1985)
#'
#' @description This function calculates the gamma dose deposited in a luminescence sample
#' taking into account layer-to-layer variations in sediment radioactivity.
#' The function scales user inputs of uranium, thorium and potassium based on
#' input parameters for sediment density, water content and given layer
#' thicknesses and distances to the sample.
#'
#' **User Input**
#'
#' To calculate the gamma dose which is deposited in a sample, the user needs
#' to provide information on those samples influencing the luminescence sample.
#' As a rule of thumb, all sediment layers within at least 30 cm radius from
#' the luminescence sample taken should be taken into account when calculating
#' the gamma dose rate. However, the actual range of gamma radiation might be
#' different, depending on the emitting radioelement, the water content and the
#' sediment density of each layer (Aitken, 1985). Therefore the user is
#' advised to provide as much detail as possible and physically sensible.
#'
#' The function requires a [data.frame] that is to be structured
#' in columns and rows, with samples listed in rows. The first column contains
#' information on the layer/sample ID, the second on the thickness (in cm) of
#' each layer, whilst column 3 should contain `NA` for all layers that are not
#' sampled for OSL/TL. For the layer the OSL/TL sample was taken from a numerical
#' value must be provided, which is the distance (in cm) measured from **bottom**
#' of the layer of interest. If the whole layer was sampled insert `0`. If the
#' sample was taken from *within* the layer, insert a numerical value `>0`,
#' which describes the distance from the middle of the sample to the bottom of
#' the layer in cm. Columns 4 to 9 should contain radionuclide concentrations
#' and their standard errors for
#' potassium (in %), thorium (in ppm) and uranium (in ppm). Columns 10 and 11
#' give information on the water content and its uncertainty (standard error)
#' in %. The layer density (in g/cm3) should be given in column 12. No cell
#' should be left blank. Please ensure to keep the column titles as given in
#' the example dataset (`data('ExampleData.ScaleGammaDose')`, see examples).
#'
#' The user can decide which dose rate
#' conversion factors should be used to calculate the gamma dose rates.
#' The options are:
#' - `"Cresswelletal2018"` (Cresswell et al., 2018)
#' - `"Liritzisetal2013"` (Liritzis et al., 2013)
#' - `"Guerinetal2011"` (Guérin et al., 2011)
#' - `"AdamiecAitken1998"` (Adamiec and Aitken, 1998)
#'
#' **Water content**
#'
#' The water content provided by the user should be calculated according to:
#'
#' \deqn{ ( Wet weight [g] - Dry weight [g] ) / Dry weight [g] * 100 }
#'
#'
#' **Calculations**
#'
#' After converting the radionuclide concentrations into dose rates, the
#' function will scale the dose rates based on the thickness of the layers,
#' the distances to the sample, the water content and the density of the sediment.
#' The calculations are based on Aitken (1985, Appendix H). As an example
#' (equivalent to Aitken, 1985), assuming three layers of sediment, where **L** is
#' inert and positioned in between the infinite thick and equally active
#' layers **A** and **B**, the dose in **L** and **B** due to **A** is given by
#'
#' \deqn{ {1-f(x)}D_A }
#'
#' Where `x` is the distance into the inert medium, so `f(x)` is the weighted
#' average fractional dose at `x` and `D_A` denotes that the dose is delivered by **A**.
#' `f(x)` is derived from table H1 (Aitken, 1985), when setting `z = x`.
#' Consequently, the dose in **A** and **L** due to **B** is given by
#'
#' \deqn{ {1 - f(t-x)}D_B }
#'
#' Here `t` is the thickness of **L** and the other parameters are denoted as above,
#' just for the dose being delivered by B. `f(t-x)` is derived from table H1
#' (Aitken, 1985), when setting `z` equal to `t-x`. Following this, the dose in **L**
#' delivered by **A** and **B** is given by
#'
#' \deqn{ {2 - f(x) - f(t-x)}D_{AB} }
#'
#' Since **A** and **B** are equally active `D_{AB} = D_A = D_B`.
#'
#' The function uses the value of the fractional dose rate at the layer
#' boundary to start the calculation for the next layer. This way, the function
#' is able to scale the gamma dose rate accurately for distant layers when the
#' density and water content is not constant for the entire section.
#'
#' @param data [data.frame] (**required**):
#' A table containing all relevant information for each individual layer. The
#' table must have the following named columns:
#'
#' - `id` ([character]): an arbitrary id or name of each layer
#' - `thickness` ([numeric]): vertical extent of each layer in cm
#' - `sample_offset` ([logical]): distance of the sample in cm,
#'  **measured from the bottom of the target layer**. Except for the target layer
#'  all values must be `NA`.
#' - `K` ([numeric]): K nuclide content in %
#' - `K_se` ([numeric]): error on the K content
#' - `Th` ([numeric]): Th nuclide content in ppm
#' - `Th_se` ([numeric]): error on the Th content
#' - `U` ([numeric]): U nuclide content in ppm
#' - `U_se` ([numeric]): error on the U content
#' - `water_content` ([numeric]): water content of each layer in %
#' - `water_content_se` ([numeric]):  error on the water content
#' - `density` ([numeric]): bulk density of each layer in g/cm^-3
#'
#' @param conversion_factors [character] (*optional*):
#' The conversion factors used to calculate the dose rate from sediment
#' nuclide contents. Valid options are:
#'
#' - `"Cresswelletal2018"` (default)
#' - `"Liritzisetal2013"`
#' - `"Guerinetal2011"`
#' - `"AdamiecAitken1998"`
#'
#' @param fractional_gamma_dose [character] (*optional*):
#' Factors to scale gamma dose rate values. Valid options are:
#'
#' - `"Aitken1985"` (default): Table H1 in the appendix
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param plot_singlePanels [logical] (*with default*):
#' enable/disable single plot mode, i.e. one plot window per plot.
#'
#' @param ... Further parameters passed to [barplot].
#'
#' @return
#'
#' After performing the calculations the user is provided with different outputs.
#' 1. The total gamma dose rate received by the sample (+/- uncertainties) as a
#' print in the console.
#' 2. A plot showing the sediment sequence, the user input sample information
#' and the contribution to total gamma dose rate.
#' 3. RLum Results. If the user wishes to save these results, writing a script
#' to run the function and to save the results would look like this:
#'
#' ```
#' mydata <- read.table("c:/path/to/input/file.txt")
#' results <- scale_GammaDose(mydata)
#' table <- get_RLum(results)
#' write.csv(table, "c:/path/to/results.csv")
#' ```
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$summary` \tab `data.frame` \tab summary of the model results \cr
#'  `$data` \tab `data.frame` \tab the original input data \cr
#'  `$dose_rates` \tab `list` \tab two `data.frames` for the scaled and infinite matrix dose rates \cr
#'  `$tables` \tab `list` \tab several `data.frames` containing intermediate results \cr
#'  `$args` \tab `character` \tab arguments of the call \cr
#'  `$call` \tab `call` \tab the original function call \cr
#' }
#'
#' **slot:** **`@info`**
#'
#' Currently unused.
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' Three plots are produced:
#'
#' - A visualisation of the provided sediment layer structure to quickly
#' assess whether the data was provided and interpreted correctly.
#' - A scatter plot of the nuclide contents per layer (K, Th, U) as well as the
#' water content. This may help to correlate the dose rate contribution of
#' specific layers to the layer of interest.
#' - A barplot visualising the contribution of each layer to the total dose rate
#' received by the sample in the target layer.
#'
#' @section Function version: 0.1.4
#'
#' @keywords datagen
#'
#' @note
#' **This function has BETA status. If possible, results should be**
#' **cross-checked.**
#'
#' @author Svenja Riedesel, Aberystwyth University (United Kingdom) \cr
#' Martin Autzen, DTU NUTECH Center for Nuclear Technologies (Denmark) \cr
#' Christoph Burow, University of Cologne (Germany) \cr
#' Based on an excel spreadsheet and accompanying macro written by Ian Bailiff.
#'
#' @seealso [Luminescence::ExampleData.ScaleGammaDose],
#' [Luminescence::BaseDataSet.ConversionFactors], [approx], [barplot]
#'
#' @references
#'
#' Aitken, M.J., 1985. Thermoluminescence Dating. Academic Press, London.
#'
#' Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update.
#' Ancient TL 16, 37-46. \doi{10.26034/la.atl.1998.292}
#'
#' Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018.
#' Dose rate conversion parameters: Assessment of nuclear data.
#' Radiation Measurements 120, 195-201. \doi{10.1016/j.radmeas.2018.02.007}
#'
#' Guérin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion
#' factors: update. Ancient TL 29, 5-8. \doi{10.26034/la.atl.2011.443}
#'
#' Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
#' A re-evaluation of radiation dose-rate conversion factors. Mediterranean
#' Archaeology and Archaeometry 13, 1-15.
#'
#' @section Acknowledgements:
#'
#' We thank Dr Ian Bailiff for the provision of an excel spreadsheet, which has
#' been very helpful when writing this function.
#'
#' @examples
#'
#' # Load example data
#' data("ExampleData.ScaleGammaDose", envir = environment())
#' x <- ExampleData.ScaleGammaDose
#'
#' # Scale gamma dose rate
#' results <- scale_GammaDose(data = x,
#'                            conversion_factors = "Cresswelletal2018",
#'                            fractional_gamma_dose = "Aitken1985",
#'                            verbose = TRUE,
#'                            plot = TRUE)
#'
#' get_RLum(results)
#'
#' @export
scale_GammaDose <- function(
  data,
  conversion_factors = c("Cresswelletal2018", "Guerinetal2011", "AdamiecAitken1998", "Liritzisetal2013")[1],
  fractional_gamma_dose = "Aitken1985",
  verbose = TRUE,
  plot = TRUE,
  plot_singlePanels = FALSE,
  ...
) {
  .set_function_name("scale_GammaDose")
  on.exit(.unset_function_name(), add = TRUE)

  ## HELPER FUNCTION ----
  # Wrapper for formatC to enforce precise digit printing
  f <- function(x, d = 3) formatC(x, digits = d, format = "f")

  ## ------------------------------------------------------------------------ ##
  ## LOAD TABLES
  ## ------------------------------------------------------------------------ ##

  # To satisfy CRAN check ('no visible global binding')
  BaseDataSet.ConversionFactors <- BaseDataSet.FractionalGammaDose <- NA

  load(system.file("data", "BaseDataSet.ConversionFactors.rda",
                   package = "Luminescence"))
  load(system.file("data", "BaseDataSet.FractionalGammaDose.rda",
                   package = "Luminescence"))


  ## ------------------------------------------------------------------------ ##
  ## DEFAULT SETTINGS
  ## ------------------------------------------------------------------------ ##
  settings <- list(
    main = "Contributions of each layer to the total \n gamma dose rate received by the sample",
    xlab = "Contribution to total gamma dose rate (%) received by the sample",
    cex = 1.0,
    col = "grey",
    info = list()
  )
  # overwrite and append default values
  settings <- modifyList(settings, list(...))


  ## ------------------------------------------------------------------------ ##
  ## CHECK INPUT
  ## ------------------------------------------------------------------------ ##

  ## Input data
  # basic class and length check
  .validate_class(data, "data.frame")
  if (ncol(data) != 12)
    .throw_error("'data' should have 12 columns (currently ", ncol(data), ")")

  # make sure that it has the correct column names
  colnames_expected <- c("id","thickness","sample_offset","K","K_se","Th","Th_se","U","U_se",
                         "water_content","water_content_se", "density")
  if (is.null(names(data)) || any(names(data) != colnames_expected)) {
    if (verbose)
      .throw_warning("Unexpected column names for 'data', new names were ",
                     "automatically assigned. See documentation to ensure ",
                     "that columns are in proper order")
    colnames(data) <- colnames_expected
  }
  # check if there is only one target layer
  if (sum(!is.na(data$sample_offset)) != 1)
    .throw_error("Only one layer must be contain a numeric value in column ",
                 "'sample_offset', all other rows must be NA")
  if (!is.numeric(data$sample_offset[which(!is.na(data$sample_offset))]))
    .throw_error("Non-numeric value in the row of the target layer")
  if (data$sample_offset[which(!is.na(data$sample_offset))] < 0)
    .throw_error("The numeric value in 'sample_offset' must be positive")
  if (data$sample_offset[which(!is.na(data$sample_offset))] > data$thickness[which(!is.na(data$sample_offset))])
    .throw_error("Sample offset larger than the target-layer's thickness")

  ## conversion factors: we do not use BaseDataSet.ConversionFactors directly
  ## as it is in alphabetical level, but we want to have 'Cresswelletal2018'
  ## in first position, as that is our default value
  valid_conversion_factors <- c("Cresswelletal2018", "Guerinetal2011",
                                "AdamiecAitken1998", "Liritzisetal2013")
  stopifnot(names(BaseDataSet.ConversionFactors) %in% valid_conversion_factors)
  conversion_factors <- .validate_args(conversion_factors,
                                       valid_conversion_factors)
  fractional_gamma_dose <- .validate_args(fractional_gamma_dose,
                                          names(BaseDataSet.FractionalGammaDose))
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)
  .validate_logical_scalar(plot_singlePanels)

  ## ------------------------------------------------------------------------ ##
  ## Select tables
  ## ------------------------------------------------------------------------ ##
  conv_fac <- BaseDataSet.ConversionFactors[[conversion_factors]]$gamma
  frac_dose <- BaseDataSet.FractionalGammaDose[[fractional_gamma_dose]]

  ## ------------------------------------------------------------------------ ##
  ## CALCULATION
  ## ------------------------------------------------------------------------ ##
  dose_rate <- data.frame(
    K = data$K * conv_fac$K[1],
    K_re = sqrt( (data$K_se / data$K)^2 + conv_fac$K[2]^2 ),
    Th = data$Th * conv_fac$Th[1],
    Th_re = sqrt( (data$Th_se / data$Th)^2 + conv_fac$Th[2]^2 ),
    U = data$U * conv_fac$U[1],
    U_re = sqrt( (data$U_se / data$U)^2 + conv_fac$U[2]^2 )
  )

  dose_rate$sum <- dose_rate$K + dose_rate$Th + dose_rate$U
  dose_rate$sum_re <- sqrt(dose_rate$K_re^2 + dose_rate$Th_re^2 + dose_rate$U_re^2)
  dose_rate$K_frac <- dose_rate$K / dose_rate$sum
  dose_rate$K_frac_re <- sqrt(dose_rate$K_re^2 + dose_rate$sum_re^2 )
  dose_rate$Th_frac <- dose_rate$Th / dose_rate$sum
  dose_rate$Th_frac_re <- sqrt(dose_rate$Th_re^2 + dose_rate$sum_re^2 )
  dose_rate$U_frac <- dose_rate$U / dose_rate$sum
  dose_rate$U_frac_re <- sqrt(dose_rate$U_re^2 + dose_rate$sum_re^2 )

  ## weighted fractional dose
  z_scale <- do.call(cbind, Map(function(d, wc) {
    (frac_dose$z * 2) / (d + ( (wc / 100) * d))
  }, data$density, data$water_content))

  layer_fracDoseRate <- do.call(cbind, Map(function(K, Th, U, K_re, Th_re, U_re) {
    data.frame(
      val = frac_dose$K * K + frac_dose$Th * Th + frac_dose$U * U,
      err = sqrt( K_re^2 + Th_re^2 + U_re^2 )
    )
  }, dose_rate$K_frac, dose_rate$Th_frac, dose_rate$U_frac,
  dose_rate$K_frac_re, dose_rate$Th_frac_re, dose_rate$U_frac_re))

  ## TODO: LEGACY CODE
  target <- which(!is.na(data$sample_offset))
  distance <- data.frame(upper = c(rev(cumsum(data$thickness[target:1])[-1]) - data$sample_offset[target],
                                   abs(data$sample_offset[target]),
                                   cumsum(data$thickness[(target+1):nrow(data)]) + data$sample_offset[target]))
  distance$lower <- abs(distance$upper - data$thickness)

  ## Calculate infitite dose rate and dose received by the sample
  ## MAP: iterate over LAYERS
  Inf_frac <- as.data.frame(do.call(rbind, Map(function(z, n) {

    interpol <- Map(function(x) {
      approx(z, x, n = 1000, method = "linear")
    }, frac_dose[, c("K", "Th", "U")])

    x1 <- data$thickness[n]
    x2 <- 0
    C1 <- which.min(abs(interpol$K$x - x1))
    C2 <- which.min(abs(interpol$K$x - x2))

    ## MAP: iterate over NUCLIDE
    do.call(cbind, Map(function(x) {
      y1 <- interpol[[x]]$y[C1]
      y2 <- interpol[[x]]$y[C2]

      ### ----
      if (n != target) {
        if (n < target) {
          k <-  n + 1
          seq <- k:target
        } else if (n > target) {
          k <- n - 1
          seq <- target:k
        }

        for (j in seq) {
          fit <- approx(z_scale[ ,j], frac_dose[ , x], n = 1000, method = "linear")
          x1_temp <- which.min(abs(fit$y - y1))
          x2_temp <- which.min(abs(fit$y - y2))

          if (j != target) {
            x1 <- fit$x[x1_temp] + data$thickness[j]
            x2 <- fit$x[x2_temp] + data$thickness[j]
          }
          if (j == target) {

            if (n < target) {
              x1 <- fit$x[x1_temp] + data$thickness[target] - data$sample_offset[target]
              x2 <- fit$x[x2_temp] + data$thickness[target] - data$sample_offset[target]
            }

            if (n > target) {
              x1 <- fit$x[x1_temp] + data$sample_offset[target]
              x2 <- fit$x[x2_temp] + data$sample_offset[target]
            }
          }

          C1_temp <- which.min(abs(fit$x - x1))
          C2_temp <- which.min(abs(fit$x - x2))

          y1 <- fit$y[C1_temp]
          y2 <- fit$y[C2_temp]
        }
        r <- y1 - y2
      }

      ### ----
      if (n == target) {
        x1 <- data$sample_offset[target]
        x2 <- abs(data$thickness[target] - data$sample_offset[target])

        C1_temp <- which.min(abs(interpol[[x]]$x - x1))
        C2_temp <- which.min(abs(interpol[[x]]$x - x2))

        r <- interpol[[x]]$y[C1_temp] + interpol[[x]]$y[C2_temp] - 1
      }

      return(r)
    }, c("K", "Th", "U")))

  }, as.data.frame(z_scale), 1:nrow(data))))


  ## Generate output object
  op <- setNames(vector(mode = "list", length = 17),
                 nm = c("K","K_se","Th","Th_se","U","U_se","sum","sum_se",
                        "K_inf","K_inf_se","Th_inf","Th_inf_se","U_inf","U_inf_se","sum_inf","sum_inf_se",
                        "contrib"))

  # fractional dose rate
  op$K <- Inf_frac$K * dose_rate$K / (1 + 1.14 * data$water_content / 100)
  op$K_se <- op$K * sqrt(dose_rate$K_re^2 + (data$water_content_se / data$water_content)^2)
  op$Th <- Inf_frac$Th * dose_rate$Th / (1 + 1.14 * data$water_content / 100)
  op$Th_se <- op$Th * sqrt(dose_rate$Th_re^2 + (data$water_content_se / data$water_content)^2)
  op$U <- Inf_frac$U * dose_rate$U / (1 + 1.14 * data$water_content / 100)
  op$U_se <- op$U * sqrt(dose_rate$U_re^2 + (data$water_content_se / data$water_content)^2)
  op$sum <- op$K + op$Th + op$U
  op$sum_se <- sqrt(op$K_se^2 + op$Th_se^2 + op$U_se^2)

  # infinite matrix dose rate
  op$K_inf <- op$K / Inf_frac$K
  op$K_inf_se <- op$K_inf * sqrt(dose_rate$K_re^2 + (data$water_content_se / data$water_content)^2)
  op$Th_inf <- op$Th / Inf_frac$Th
  op$Th_inf_se <- op$Th_inf * sqrt(dose_rate$Th_re^2 + (data$water_content_se / data$water_content)^2)
  op$U_inf <- op$U / Inf_frac$U
  op$U_inf_se <- op$U_inf * sqrt(dose_rate$U_re^2 + (data$water_content_se / data$water_content)^2)
  op$sum_inf <- op$K_inf + op$Th_inf + op$U_inf
  op$sum_inf_se <- sqrt(op$K_inf_se^2 + op$Th_inf_se^2 + op$U_inf_se^2)

  ## Calculate the total dose rates
  for(i in seq(1, length(op)-1, 2)) {
    op[[i]] <- c(op[[i]], sum(op[[i]]))
    op[[i+1]] <- c(op[[i+1]], sqrt(sum(op[[i+1]]^2)))
  }

  ## Calculate contributions for each layer
  op$contrib <- op$sum[1:nrow(data)] / op$sum[nrow(data)+1] * 100
  op$contrib <- c(op$contrib, sum(op$contrib))


  # Cast to data.frame
  op <- as.data.frame(do.call(cbind, op))

  ## ------------------------------------------------------------------------ ##
  ## CONSOLE OUTPUT
  ## ------------------------------------------------------------------------ ##
  if (verbose) {
    cat(paste0("\n [scale_GammaDose()]\n\n"))
    cat(" ----\n")
    cat(" Conversion factors:", conversion_factors, "\n")
    cat(" Gamma dose fractions:", fractional_gamma_dose, "\n")
    cat(" Target layer:", data$id[target], "\n\n")

    cat(" ---- Infinite matrix gamma dose rate per layer ----\n\n")
    print(data.frame(ID = data$id,
                     `K (Gy/ka)` = paste0(f(op$K_inf[-(nrow(data)+1)]), "\u00b1", f(op$K_inf_se[-(nrow(data)+1)])),
                     `Th (Gy/ka)` = paste0(f(op$Th_inf[-(nrow(data)+1)]), "\u00b1", f(op$Th_inf_se[-(nrow(data)+1)])),
                     `U (Gy/ka)` = paste0(f(op$U_inf[-(nrow(data)+1)]), "\u00b1", f(op$U_inf_se[-(nrow(data)+1)])),
                     `Total (Gy/ka)` = f(op$sum_inf[-(nrow(data)+1)]),
                     check.names = FALSE
    ))
    cat("\n")
    cat(sprintf(" ---- Scaled gamma dose rate for target layer: %s  ----\n\n", data$id[target]))
    print(data.frame(ID = c(data$id, "TOTAL"),
                     `K (Gy/ka)` = paste0(f(op$K), "\u00b1", f(op$K_se)),
                     `Th (Gy/ka)` = paste0(f(op$Th), "\u00b1", f(op$Th_se)),
                     `U (Gy/ka)` = paste0(f(op$U), "\u00b1", f(op$U_se)),
                     `Contribution (%)` = round(op$contrib, 1),
                     check.names = FALSE
    ))
    cat("\n ----\n")
    cat(" Infinite matrix gamma dose rate:\t",
        f(op$sum_inf[target]), "\u00b1",
        f(op$sum_inf_se[target]), "Gy/ka \n")
    cat(" Scaled gamma dose rate:\t\t",
        f(op$sum[length(op$sum)]), "\u00b1",
        f(op$sum_se[length(op$sum_se)]), "Gy/ka")
    cat("\n\n")
  }


  ## ------------------------------------------------------------------------ ##
  ## PLOT
  ## ------------------------------------------------------------------------ ##
  if (plot) {

    # save and recover plot parameters
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    if (!plot_singlePanels)
      graphics::layout(matrix(
                    c(1,1, 2, 3, 4, 5,
                      1,1, 2, 3, 4, 5,
                      1,1, 6, 6, 6, 6,
                      1,1, 6, 6, 6, 6), ncol = 6, byrow = TRUE))

    ## Plot 1 - Layer setup
    ## --------------------------------------------------------------

    ## Global plot settings
    par(mar = c(2, 5, ifelse(plot_singlePanels, 4, 1), 4) + 0.1)

    plot(NA, NA,
         main = ifelse(plot_singlePanels, "Profile structure", ""),
         xlim = c(0, 1),
         ylim = rev(range(pretty(c(sum(data$thickness), 0)))),
         xaxt = "n",
         xlab = "",
         ylab = "Depth below surface of uppermost layer (cm)",
         bty = "n",
         xaxs = "i")

    # x-axis label
    title(xlab = "Horizontal extent (a.u.)", line = 0)

    # horizontal layer lines
    abline(h = c(0, cumsum(data$thickness), sum(data$thickness)), lty = 1, col = "grey50", xpd = FALSE)

    # layer names
    mtext(side = 2, at = c(0, cumsum(data$thickness) - data$thickness / 2, sum(data$thickness)),
          text = c("", data$id, ""), las = 1, line = -5, cex = 0.75, padj = 0.3,
          col = "#428bca")

    # right y-axis
    axis(side = 4, at = c(0, cumsum(data$thickness), sum(data$thickness)),
         labels = FALSE, tck = -0.01)

    # right y-axis labels
    mtext(side = 4, at = c(0, cumsum(data$thickness) - data$thickness / 2, sum(data$thickness)),
          text = c("", paste(data$thickness, "cm"), ""), las = 1,
          line = ifelse(plot_singlePanels, 0.5, -4),
          cex = 0.8,
          col = "#b22222")

    # fill gap between lowermost layer and max range of pretty xlim
    polygon(x = c(0, 1, 1, 0),
            y = c(sum(data$thickness), sum(data$thickness),
                  max(range(pretty(c(sum(data$thickness), 0)))), max(range(pretty(c(sum(data$thickness), 0))))),
            density = 10, angle = 45
    )

    # add sample
    points(x = 0.5, y = sum(data$thickness[1:target]) - data$sample_offset[target],
           pch = 13, col = "#b22222", cex = 3, lwd = 2)


    ## PLOT 2 - Nuclide contents
    ## --------------------------------------------------------------

    # global plot settings
    if (!plot_singlePanels) {
      par(
        mar = c(4, 2, 3, 0.5) + 0.1,
        cex = 0.6,
        # oma = c(1, 1, 1, 1) + 0.1,
        pch = 16)
    } else {
      par(par.default)
      par(
        mfrow = c(1, 4),
        mar = c(4, 5, 0, 0) + 0.1,
        oma = c(1, 1, 1, 1) + 0.1,
        pch = 16)
    }

    # calculate fancy x-axis limits by considering the error bars
    calc_xlim <- function(x, se) {
      range(pretty(c(x - se, x + se)))
    }

    # horizontal error bars
    plot_error <- function(x, se) {
      segments(x - se, length(x):1, x + se, length(x):1)
      epsilon <- 0.1
      segments(x - se, length(x):1 - epsilon, x - se, length(x):1 + epsilon)
      segments(x + se, length(x):1 - epsilon, x + se, length(x):1 + epsilon)
    }

    # plot labels
    cols <- c("K", "Th", "U", "water_content")
    xlabs <- c("K content (%)", "Th content (ppm)", "U content (ppm)", "Water content (%)")

    # main plot
    for (i in 1:length(cols)) {

      # main
      plot(NA, NA,
           xlim = calc_xlim(data[[cols[i]]], data[[paste0(cols[i], "_se")]]),
           ylim = c(1, nrow(data)),
           ylab = "",
           xlab = xlabs[i],
           yaxt = "n")

      # vertical orientation lines
      abline(v = axTicks(side = 1), col = "grey", lty = 3, xpd = FALSE)

      # data points
      points(data[[cols[i]]], nrow(data):1, cex = 1.5)

      # errors
      plot_error(data[[cols[i]]], data[[paste0(cols[i], "_se")]])

      # y-axis label for the first plot
      if (i == 1)
        axis(2, at = nrow(data):1, labels = data$id, las = 1)
    }


    ## PLOT 3 - Contribution
    ## --------------------------------------------------------------

    ## Global plot settings
    # recover standard plot settings first
    if (!plot_singlePanels) {
      par(mar = c(5, 5, 1, 6) + 0.1,
          cex = 0.7)
    } else {
      par(par.default)
      par(mar = c(5, 8, 4, 4) + 0.1,
          cex = settings$cex)
    }

    ## Set colors: target layer is blue, all other grey
    cols <- c("grey", "#428bca")
    pos <- rev(as.numeric(!is.na(data$sample_offset)) + 1)

    ## Contributions of each layer
    bp <- graphics::barplot(height = op$contrib[(nrow(op)-1):1],
                  horiz = TRUE,
                  main = ifelse(plot_singlePanels, settings$main, ""),
                  xlab = settings$xlab,
                  xlim = range(pretty(op$contrib[1:(nrow(op)-1)])),
                  col = cols[pos])

    # layer names
    mtext(side = 2, at = bp,
          line = ifelse(plot_singlePanels, 3, 3.5),
          las = 1,
          cex = ifelse(plot_singlePanels, 0.8, 0.7),
          text = rev(data$id))

    # contribution percentage
    mtext(side = 2, at = rev(bp),
          text = paste(signif(op$contrib[1:(nrow(op) - 1)], 2), "%"),
          col = "#b22222",
          las = 1, line = 0.5, cex = 0.7)

    # absolute dose rate values (right side)
    mtext(side = 4, at = rev(bp),
          text = paste(c("  ", rep("+", nrow(op) - 2)), f(op$sum[1:(nrow(op) - 1)]), "Gy/ka"),
          col = "black",
          las = 1, line = -0.5, cex = 0.7)

    # sum of absolute dose rate values (-> scaled total gamma dose rate)
    mtext(side = 4,
          at = min(bp) - diff(bp)[1] / 2,
          text = paste("=", f(op$sum[nrow(op)]), "Gy/ka"),
          col = "#b22222", las = 1,
          line = ifelse(plot_singlePanels, -1, -0.5),
          cex = ifelse(plot_singlePanels, 0.8, 0.7))
  }

  ## ------------------------------------------------------------------------ ##
  ## RETURN VALUE
  ## ------------------------------------------------------------------------ ##

  ## Infinity matrix dose rate table
  infinite_matrix <- data.frame(
    ID = data$id,
    K = op$K_inf[-(nrow(data)+1)],
    K_err = op$K_inf_se[-(nrow(data)+1)],
    Th = op$Th_inf[-(nrow(data)+1)],
    Th_err = op$Th_inf_se[-(nrow(data)+1)],
    U = op$U_inf[-(nrow(data)+1)],
    U_err = op$U_inf_se[-(nrow(data)+1)],
    Total = op$sum_inf[-(nrow(data)+1)]
  )

  ## Scaled dose rate table
  scaled_dose_rate <- data.frame(
    ID = c(data$id, "TOTAL"),
    K = op$K,
    K_err = op$K_se,
    Th = op$Th,
    Th_err = op$Th_se,
    U = op$U,
    U_err = op$U_se,
    Contribution = op$contrib
  )

  ## Summary table with the most important results
  summary <- data.frame(
    id = data$id[target],
    dose_rate_K = op$K[nrow(op)],
    dose_rate_K_err = op$K_se[nrow(op)],
    dose_rate_Th = op$Th[nrow(op)],
    dose_rate_Th_err = op$Th_se[nrow(op)],
    dose_rate_U = op$U[nrow(op)],
    dose_rate_U_err = op$U_se[nrow(op)],
    dose_rate_total = op$sum[length(op$sum)],
    dose_rate_total_err = op$sum_se[length(op$sum_se)]
  )

  ## Create RLum.Results object (return object)
  set_RLum(class = "RLum.Results",
                      originator = "scale_GammaDose",
                      data = list(summary = summary,
                                  data = data,
                                  dose_rates = list(
                                    infinite_matrix = infinite_matrix,
                                    scaled_dose_rate = scaled_dose_rate
                                  ),
                                  tables = list(
                                    conversion_factors = conv_fac,
                                    distances = distance,
                                    layer_fractional_dose_rate = layer_fracDoseRate,
                                    dose_rates = dose_rate,
                                    infnite_matrix_dose_fractions = Inf_frac,
                                    z_scale = z_scale
                                  ),
                                  args = as.list(sys.call()[-1]),
                                  call = sys.call()),
                      info = settings$info
  )
}

## load data
data(ExampleData.ScaleGammaDose, envir = environment())
d <- ExampleData.ScaleGammaDose

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(
    scale_GammaDose(NA, plot = FALSE, verbose = TRUE),
    "'data' should be of class 'data.frame'"
  )
  expect_error(
    scale_GammaDose(d[ ,1:10], plot = FALSE, verbose = TRUE),
    "'data' should have 12 columns"
  )
  SW({
  expect_warning({
    tmp <- d
    colnames(tmp) <- letters[1:ncol(tmp)]
    scale_GammaDose(tmp, plot = FALSE, verbose = TRUE)
    },
    "Unexpected column names"
    )
  })
  expect_error({
    tmp <- d
    tmp$sample_offset <- NA
    scale_GammaDose(tmp, plot = FALSE, verbose = TRUE)
    },
    "Only one layer must be contain a numeric value in column 'sample_offset'"
  )
  expect_error({
    tmp <- d
    tmp$sample_offset[5] <- "target"
    scale_GammaDose(tmp, plot = FALSE, verbose = TRUE)
    },
    "Non-numeric value in the row of the target layer"
  )
  expect_error({
    tmp <- d
    tmp$sample_offset[5] <- -1
    scale_GammaDose(tmp, plot = FALSE, verbose = TRUE)
  },
  "The numeric value in 'sample_offset' must be positive"
  )
  expect_error({
    tmp <- d
    tmp$sample_offset[5] <- 20
    scale_GammaDose(tmp, plot = FALSE, verbose = TRUE)
  },
  "Sample offset larger than the target-layer's thickness"
  )

  expect_error(scale_GammaDose(d, conversion_factors = c("a", "b")),
               "'conversion_factors' contains multiple values but not all of")
  expect_error(scale_GammaDose(d, conversion_factors = 1),
               "'conversion_factors' should be one of 'Cresswelletal2018'")
  expect_error(scale_GammaDose(d, conversion_factors = "HansGuenter2020"),
               "'conversion_factors' should be one of 'Cresswelletal2018'")

  expect_error(scale_GammaDose(d, fractional_gamma_dose = c("a", "b")),
               "'fractional_gamma_dose' contains multiple values but not all")
  expect_error(scale_GammaDose(d, fractional_gamma_dose = 1),
               "'fractional_gamma_dose' should be one of 'Aitken1985'")
  expect_error(scale_GammaDose(d, fractional_gamma_dose = "error"),
               "'fractional_gamma_dose' should be one of 'Aitken1985'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  ## Conversion factors: Liritzisetal2013
  SW({
  expect_snapshot_RLum(scale_GammaDose(data = ExampleData.ScaleGammaDose,
                                       conversion_factors = "Liritzisetal2013",
                                       fractional_gamma_dose = "Aitken1985",
                                       plot = TRUE, plot_singlePanels = TRUE,
                                       verbose = TRUE),
                       tolerance = snapshot.tolerance)
  })

  ## Conversion factors: Guerinetal2011
  expect_snapshot_RLum(scale_GammaDose(data = ExampleData.ScaleGammaDose,
                                       conversion_factors = "Guerinetal2011",
                                       fractional_gamma_dose = "Aitken1985",
                                       plot = TRUE, verbose = FALSE),
                       tolerance = snapshot.tolerance)

  ## Conversion factors: AdamiecAitken1998
  expect_snapshot_RLum(scale_GammaDose(data = ExampleData.ScaleGammaDose,
                                       conversion_factors = "AdamiecAitken1998",
                                       fractional_gamma_dose = "Aitken1985",
                                       plot = FALSE, verbose = FALSE),
                       tolerance = snapshot.tolerance)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("defaults",
                              scale_GammaDose(ExampleData.ScaleGammaDose))
  })
})

#' @title Plot kernel density estimate with statistics
#'
#' @description
#' Plot a kernel density estimate of measurement values in combination with the
#' actual values and associated error bars in ascending order. If enabled, the
#' boxplot will show the usual distribution parameters (median as
#' bold line, box delimited by the first and third quartile, whiskers defined
#' by the extremes and outliers shown as points) and also the mean and
#' standard deviation as pale bold line and pale polygon, respectively.
#'
#' The function allows passing several plot arguments, such as `main`,
#' `xlab`, `cex`. However, as the figure is an overlay of two
#' separate plots, `ylim` must be specified in the order: c(ymin_axis1,
#' ymax_axis1, ymin_axis2, ymax_axis2) when using the cumulative values plot
#' option. See examples for some further explanations. For details on the
#' calculation of the bin-width (parameter `bw`) see
#' [density].
#'
#' A statistic summary, i.e. a collection of statistic measures of
#' centrality and dispersion (and further measures) can be added by specifying
#' one or more of the following keywords:
#' - `"n"` (number of samples)
#' - `"mean"` (mean De value)
#' - `"median"` (median of the De values)
#' - `"sd.rel"` (relative standard deviation in percent)
#' - `"sd.abs"` (absolute standard deviation)
#' - `"se.rel"` (relative standard error)
#' - `"se.abs"` (absolute standard error)
#' - `"in.2s"` (percent of samples in 2-sigma range)
#' - `"kurtosis"` (kurtosis)
#' - `"skewness"` (skewness)
#'
#' **Note:** the input data for the statistic summary is sent to function
#' [Luminescence::calc_Statistics] depending on the log-option for the z-scale. If
#' `"log.z = TRUE"`, the summary is based on the logarithms of the input
#' data. If `"log.z = FALSE"` the linearly-scaled data is used.
#'
#' **Note:** [Luminescence::calc_Statistics] calculates these statistic
#' measures in three different ways: `unweighted`, `weighted` and
#' `MCM-based` (i.e., based on Monte Carlo Methods). By default, the
#' MCM-based version is used. This can be controlled via the `summary.method`
#' argument.
#'
#' @param data [data.frame], [vector] or [Luminescence::RLum.Results-class] object (**required**):
#' for `data.frame`: either two columns: De (`values[,1]`) and De error
#' (`values[,2]`), or one: De (`values[,1]`). If a numeric vector or a
#' single-column data frame is provided, De error is assumed to be 10^-9
#' for all measurements and error bars are not drawn.
#' For plotting multiple data sets, these must be provided as
#' `list` (e.g. `list(dataset1, dataset2)`).
#'
#' @param na.rm [logical] (*with default*):
#' exclude `NA` values from the data set prior to any further operation.
#'
#' @param values.cumulative [logical] (*with default*):
#' show cumulative individual data.
#'
#' @param order [logical] (*with default*):
#' Order data in ascending order.
#'
#' @param boxplot [logical] (*with default*):
#' optionally show a boxplot (depicting median as thick central line,
#' first and third quartile as box limits, whiskers denoting +/- 1.5
#' interquartile ranges and dots further outliers).
#'
#' @param rug [logical] (*with default*):
#' optionally add rug.
#'
#' @param summary [character] (*with default*):
#' add statistic measures of centrality and dispersion to the plot. Can be one
#' or more of several keywords. See details for available keywords.
#'
#' @param summary.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`)
#' for the statistical summary. Alternatively, the keyword `"sub"` may be
#' specified to place the summary below the plot header. However, this latter
#' option in only possible if `mtext` is not used. In case of coordinate
#' specification, y-coordinate refers to the right y-axis.
#'
#' @param summary.method [character] (*with default*):
#' keyword indicating the method used to calculate the statistic summary.
#' One of `"MCM"` (default), `"weighted"` or `"unweighted"`.
#' See [Luminescence::calc_Statistics] for details.
#'
#' @param bw [character], [numeric] (*with default*):
#' bin-width, chose a numeric value for manual setting.
#'
#' @param ... further arguments and graphical parameters passed to [plot].
#'
#' @note
#' The plot output is no 'probability density' plot (cf. the discussion
#' of Berger and Galbraith in Ancient TL; see references)!
#'
#' @section Function version: 3.6.1
#'
#' @author
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [density], [plot]
#'
#' @examples
#'
#' ## read example data set
#' data(ExampleData.DeValues, envir = environment())
#' ExampleData.DeValues <-
#'   convert_Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))
#'
#' ## create plot straightforward
#' plot_KDE(data = ExampleData.DeValues)
#'
#' ## create plot with logarithmic x-axis
#' plot_KDE(data = ExampleData.DeValues,
#'          log = "x")
#'
#' ## create plot with user-defined labels and axes limits
#' plot_KDE(data = ExampleData.DeValues,
#'          main = "Dose distribution",
#'          xlab = "Dose (s)",
#'          ylab = c("KDE estimate", "Cumulative dose value"),
#'          xlim = c(100, 250),
#'          ylim = c(0, 0.08, 0, 30))
#'
#' ## create plot with boxplot option
#' plot_KDE(data = ExampleData.DeValues,
#'          boxplot = TRUE)
#'
#' ## create plot with statistical summary below header
#' plot_KDE(data = ExampleData.DeValues,
#'          summary = c("n", "median", "skewness", "in.2s"))
#'
#' ## create plot with statistical summary as legend
#' plot_KDE(data = ExampleData.DeValues,
#'          summary = c("n", "mean", "sd.rel", "se.abs"),
#'          summary.pos = "topleft")
#'
#' ## split data set into sub-groups, one is manipulated, and merge again
#' data.1 <- ExampleData.DeValues[1:15,]
#' data.2 <- ExampleData.DeValues[16:25,] * 1.3
#' data.3 <- list(data.1, data.2)
#'
#' ## create plot with two subsets straightforward
#' plot_KDE(data = data.3)
#'
#' ## create plot with two subsets and summary legend at user coordinates
#' plot_KDE(data = data.3,
#'          summary = c("n", "median", "skewness"),
#'          summary.pos = c(110, 0.07),
#'          col = c("blue", "orange"))
#'
#' ## example of how to use the numerical output of the function
#' ## return plot output to draw a thicker KDE line
#' KDE_out <- plot_KDE(data = ExampleData.DeValues)
#'
#' @export
plot_KDE <- function(
  data,
  na.rm = TRUE,
  values.cumulative = TRUE,
  order = TRUE,
  boxplot = TRUE,
  rug = TRUE,
  summary = "",
  summary.pos = "sub",
  summary.method = c("MCM", "weighted", "unweighted"),
  bw = "nrd0",
  ...
) {
  .set_function_name("plot_KDE")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  if (inherits(data, "list") && length(data) == 0) {
    .throw_error("'data' is an empty list")
  }

  ## Homogenise input data format
  if (!inherits(data, "list")) {
    data <- list(data)
  }

  ## check/adjust input data structure
  for(i in 1:length(data)) {
    .validate_class(data[[i]], c("RLum.Results", "data.frame", "numeric"),
                    name = "'data'")

    ## extract RLum.Results
    if (inherits(data[[i]], "RLum.Results")) {
      data[[i]] <- get_RLum(data[[i]], "data")[,1:2]
    }

      ## ensure that the dataset it not degenerate
      if (NROW(data[[i]]) == 0) {
       .throw_error("Input data ", i, " has 0 rows")
      }

      ## if `data[[i]]` is a numeric vector or a single-column data frame,
      ## append a second column with a small non-zero value (10^-9 for
      ## consistency with what `calc_Statistics() does)
      if (NCOL(data[[i]]) < 2) {
        data[[i]] <- data.frame(data[[i]], 10^-9)
        attr(data[[i]], "De.errors.available") <- FALSE
      } else {
        ## keep only the first two columns
        data[[i]] <- data[[i]][, 1:2]
        attr(data[[i]], "De.errors.available") <- TRUE
      }

    ## find the index Inf values in each of the two columns and remove the
    ## corresponding rows if needed
    inf.idx <- unlist(lapply(data[[i]], function(x) which(is.infinite(x))))
    if (length(inf.idx) > 0) {
      inf.row <- sort(unique(inf.idx))
      .throw_warning("Inf values removed in rows: ",
                     .collapse(inf.row, quote = FALSE), " in data.frame ", i)
      data[[i]] <- data[[i]][-inf.row, ]
      rm(inf.idx, inf.row)

      ##check if empty
      if(nrow(data[[i]]) == 0){
        data[i] <- NULL
      }
    }
  }

  ##check if list is empty
  if(length(data) == 0)
    .throw_error("Your input is empty due to Inf removal")

  .validate_logical_scalar(na.rm)
  .validate_logical_scalar(values.cumulative)
  .validate_logical_scalar(order)
  .validate_logical_scalar(boxplot)
  .validate_logical_scalar(rug)
  summary.method <- .validate_args(summary.method, c("MCM", "weighted", "unweighted"))
  .validate_class(summary, "character")
  if (is.numeric(summary.pos)) {
    .validate_length(summary.pos, 2)
  }
  else {
    summary.pos <- .validate_args(summary.pos,
                                  c("sub", "left", "center", "right",
                                    "topleft", "top", "topright",
                                    "bottomleft", "bottom", "bottomright"))
  }
  .validate_class(bw, c("character", "numeric"), length = 1)
  if (is.numeric(bw))
    .validate_positive_scalar(bw)

  ## set mtext output
  mtext <- list(...)$mtext %||% ""

  ## check/set layout definitions
  layout <- get_Layout(layout = list(...)$layout %||% "default")

  ## data preparation steps ---------------------------------------------------

  ## optionally, count and exclude NA values and print result
  if (na.rm) {
    for(i in 1:length(data)) {
      na.idx <- which(is.na(data[[i]][, 1]))
      n.NA <- length(na.idx)
      if (n.NA > 0) {
        message(sprintf("%d NA value%s excluded from data set %d\n",
                        n.NA, ifelse(n.NA > 1, "s", ""), i))
        data[[i]] <- data[[i]][-na.idx, ]
      }
    }
  }

  ## optionally, order data set
  if (order) {
    for(i in 1:length(data)) {
      data[[i]] <- data[[i]][order(data[[i]][,1]),]
    }
  }

  ## calculate and paste statistical summary
  De.stats <- matrix(nrow = length(data), ncol = 12)
  colnames(De.stats) <- c("n",
                          "mean",
                          "median",
                          "kdemax",
                          "sd.abs",
                          "sd.rel",
                          "se.abs",
                          "se.rel",
                          "q.25",
                          "q.75",
                          "skewness",
                          "kurtosis")
  De.density <- list(NA)

  ## loop through all data sets
  for(i in 1:length(data)) {
    statistics <- calc_Statistics(data[[i]], na.rm = na.rm)[[summary.method]]

    De.stats[i,1] <- statistics$n
    De.stats[i,2] <- statistics$mean
    De.stats[i,3] <- statistics$median
    De.stats[i,5] <- statistics$sd.abs
    De.stats[i,6] <- statistics$sd.rel
    De.stats[i,7] <- statistics$se.abs
    De.stats[i,8] <- statistics$se.rel
    De.stats[i,9] <- quantile(data[[i]][,1], 0.25)
    De.stats[i,10] <- quantile(data[[i]][,1], 0.75)
    De.stats[i,11] <- statistics$skewness
    De.stats[i,12] <- statistics$kurtosis

    if(nrow(data[[i]]) >= 2){
      De.density[[length(De.density) + 1]] <- density(data[[i]][,1],
                                                      kernel = "gaussian",
                                                      bw = bw)
    }else{
      De.density[[length(De.density) + 1]] <- NA
      .throw_warning("Single data point found, no density calculated")
    }
  }

  ## remove dummy list element
  De.density[[1]] <- NULL

  ## create global data set
  De.global <- data[[1]][,1]
  De.error.global <- data[[1]][,2]
  De.density.range <- matrix(nrow = length(data),
                             ncol = 4)

  for(i in 1:length(data)) {
    ##global De and De.error vector
    De.global <- c(De.global, data[[i]][,1])
    De.error.global <- c(De.error.global, data[[i]][,2])

    ## density range
    if(!all(is.na(De.density[[i]]))){
      De.density.range[i,1] <- min(De.density[[i]]$x)
      De.density.range[i,2] <- max(De.density[[i]]$x)
      De.density.range[i,3] <- min(De.density[[i]]$y)
      De.density.range[i,4] <- max(De.density[[i]]$y)

      ## position of maximum KDE value
      De.stats[i,4] <- De.density[[i]]$x[which.max(De.density[[i]]$y)]
    }
  }

  ## Get global range of densities
  De.density.range <- c(min(De.density.range[,1]),
                        max(De.density.range[,2]),
                        min(De.density.range[,3]),
                        max(De.density.range[,4]))

  ## helper to generate an element of the statistical summary
  .summary_line <- function(keyword, summary, val, label = keyword,
                            percent = FALSE, sep = FALSE, digits = 2) {
    ifelse(keyword %in% summary,
           paste0(label, " = ", round(val, digits),
                  if (percent) " %" else NULL, if (sep) " | " else  "\n"),
           "")
  }

  ## initialize list
  label.text <- NULL
  is.sub <- summary.pos[1] == "sub"
  stops <- NULL
  for (i in 1:length(data)) {
    if (!is.sub)
      stops <- strrep("\n", (i - 1) * length(summary))

    summary.text <- character(0)
    for (j in 1:length(summary)) {
      summary.text <-
        c(summary.text,
          .summary_line("n", summary[j], De.stats[i, 1], sep = is.sub),
          .summary_line("mean", summary[j], De.stats[i, 2], sep = is.sub),
          .summary_line("median", summary[j], De.stats[i, 3], sep = is.sub),
          .summary_line("kdemax", summary[j], De.stats[i, 4], sep = is.sub),
          .summary_line("sd.abs", summary[j], De.stats[i, 5], sep = is.sub,
                        label = "sd"),
          .summary_line("sd.rel", summary[j], De.stats[i, 6], sep = is.sub,
                        label = "rel. sd", percent = TRUE),
          .summary_line("se.abs", summary[j], De.stats[i, 7], sep = is.sub,
                        label = "se"),
          .summary_line("se.rel", summary[j], De.stats[i, 8], sep = is.sub,
                        label = "rel. se", percent = TRUE),
          .summary_line("skewness", summary[j], De.stats[i, 11], sep = is.sub),
          .summary_line("kurtosis", summary[j], De.stats[i, 12], sep = is.sub),
          .summary_line("in.2s", summary[j],
                        sum(data[[i]][, 1] > (De.stats[i, 2] - 2 *
                                              De.stats[i, 5]) &
                            data[[i]][, 1] < (De.stats[i, 2] + 2 *
                                              De.stats[i, 5])) /
                        nrow(data[[i]]) * 100, sep = is.sub,
                        label = "in 2 sigma", percent = TRUE, digits = 1))
    }

    label.text[[i]] <- paste0(
        if (is.sub) "" else stops,
        paste(summary.text, collapse = ""),
        stops)
  }

  ## remove outer vertical lines from string
  if (is.sub) {
    for (i in seq_along(label.text)) {
      label.text[[i]] <- substr(x = label.text[[i]],
                                start = 1,
                                stop = nchar(label.text[[i]]) - 3)
    }
  }

  ## read out additional parameters -------------------------------------------
  main <- list(...)$main %||% expression(bold(paste(D[e], " distribution")))
  sub <- list(...)$sub
  xlab <- list(...)$xlab %||% expression(paste(D[e], " [Gy]"))
  ylab <- list(...)$ylab %||% c("Density", "Cumulative frequency")
  xlim.plot <- list(...)$xlim %||% c(min(c(De.global - De.error.global),
                                         De.density.range[1],
                                         na.rm = TRUE),
                                     max(c(De.global + De.error.global),
                                         De.density.range[2],
                                         na.rm = TRUE))

  if ("ylim" %in% ...names()) {
    ylim.plot <- list(...)$ylim
    .validate_class(ylim.plot, "numeric", length = 4, name = "'ylim'")
  } else if (!is.na(De.density.range[1])) {
      ylim.plot <- c(De.density.range[3],
                     De.density.range[4],
                     0,
                     max(De.stats[,1]))
  } else {
      ylim.plot <- c(0,
                     max(De.stats[,1]),
                     0,
                     max(De.stats[,1]))
  }

  log.option <- list(...)$log %||% ""
  lty <- list(...)$lty %||% rep(1, length(data))
  lwd <- list(...)$lwd %||% rep(1, length(data))
  cex <- list(...)$cex %||% 1
  fun <- isTRUE(list(...)$fun)

  if ("col" %in% ...names()) {
    col.stats <- list(...)$col
    col.kde.line <- list(...)$col
    col.kde.fill <- NA
    col.value.dot <- list(...)$col
    col.value.bar <- list(...)$col
    col.value.rug <- list(...)$col
    col.boxplot.line <- list(...)$col
    col.boxplot.fill <- NA
  } else {
    .set_colour_value <- function(layout_value) {
      if (length(layout_value) == 1)
        c(layout_value, 2:length(data))
      else
        layout_value
    }
    col.stats <- .set_colour_value(layout$kde$colour$stats)
    col.kde.line <- .set_colour_value(layout$kde$colour$kde.line)
    col.kde.fill <- .set_colour_value(layout$kde$colour$kde.fill)
    col.value.dot <- .set_colour_value(layout$kde$colour$value.dot)
    col.value.bar <- .set_colour_value(layout$kde$colour$value.bar)
    col.value.rug <- .set_colour_value(layout$kde$colour$value.rug)
    col.boxplot.line <- .set_colour_value(layout$kde$colour$boxplot.line)
    col.boxplot.fill <- .set_colour_value(layout$kde$colour$boxplot.fill)
  }

  ## convert keywords into summary placement coordinates
  coords <- .get_keyword_coordinates(summary.pos, xlim.plot, ylim.plot[1:2])
  summary.pos <- coords$pos
  summary.adj <- coords$adj

  ## plot data sets -----------------------------------------------------------
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ## setup plot area
  toplines <- 1
  if (length(summary) >= 1 && is.sub) {
    toplines <- length(data)
  }

  ## extract original plot parameters
  par(mar = c(5, 5.5, 2.5 + toplines, 4.5),
      bg = layout$kde$colour$background,
      xpd = FALSE,
      cex = cex)

  dim <- layout$kde$dimension
  if (dim$figure.width != "auto" || dim$figure.height != "auto") {
    par(mai = dim$margin / 25.4,
        pin = c(dim$figure.width - dim$margin[2] - dim$margin[4],
                dim$figure.height - dim$margin[1] - dim$margin[3]) / 25.4)
  }

  ## create empty plot to get plot dimensions
  plot(NA,
       xlim = xlim.plot,
       ylim = ylim.plot[1:2],
       sub = sub,
       log = log.option,
       axes = FALSE,
       ann = FALSE)

  ## get line height in xy coordinates
  l_height <- par()$cxy[2]

  ## optionally update ylim
  if (boxplot && values.cumulative) {
    ylim.plot[1] <- ylim.plot[1] - 1.4 * l_height
  }

  ## create empty plot to set adjusted plot dimensions
  par(new = TRUE)
  plot(NA,
       xlim     = xlim.plot,
       ylim     = ylim.plot[1:2],
       log      = log.option,
       cex      = cex,
       axes = FALSE,
       ann = FALSE)

  ## add box
  graphics::box(which = "plot",
                col = layout$kde$colour$box)

  ## add x-axis
  axis(side = 1,
       col = layout$kde$colour$xtck,
       col.axis = layout$kde$colour$xtck,
       labels = NA,
       tcl = -layout$kde$dimension$xtcl / 200,
       cex = cex)

  axis(side = 1,
       line = 2 * layout$kde$dimension$xtck.line / 100 - 2,
       lwd = 0,
       col = layout$kde$colour$xtck,
       family = layout$kde$font.type$xtck,
       font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                      layout$kde$font.deco$xtck],
       col.axis = layout$kde$colour$xtck,
       cex.axis = layout$kde$font.size$xlab/12)

  mtext(text = xlab,
        side = 1,
        line = 3 * layout$kde$dimension$xlab.line / 100,
        col = layout$kde$colour$xlab,
        family = layout$kde$font.type$xlab,
        font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                       layout$kde$font.deco$xlab],
        cex = cex * layout$kde$font.size$xlab/12)

  ## add left y-axis
  axis(side = 2,
       at = pretty(x = range(De.density.range[3:4])),
       col = layout$kde$colour$ytck1,
       col.axis = layout$kde$colour$ytck1,
       labels = NA,
       tcl = -layout$kde$dimension$ytck1 / 200,
       cex = cex)

  axis(side = 2,
       at = pretty(x = range(De.density.range[3:4])),
       line = 2 * layout$kde$dimension$ytck1.line / 100 - 2,
       lwd = 0,
       col = layout$kde$colour$ytck1,
       family = layout$kde$font.type$ytck1,
       font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                      layout$kde$font.deco$ytck1],
       col.axis = layout$kde$colour$ytck1,
       cex.axis = layout$kde$font.size$ylab1/12)

  mtext(text = ylab[1],
        side = 2,
        line = 3 * layout$kde$dimension$ylab1.line / 100,
        col = layout$kde$colour$ylab1,
        family = layout$kde$font.type$ylab1,
        font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                       layout$kde$font.deco$ylab1],
        cex = cex * layout$kde$font.size$ylab1/12)

  for(i in 1:length(data)) {
    if(!all(is.na(De.density[[i]]))){
      polygon(x = c(par()$usr[1], De.density[[i]]$x, par()$usr[2]),
              y = c(min(De.density[[i]]$y),De.density[[i]]$y, min(De.density[[i]]$y)),
              border = col.kde.line[i],
              col = col.kde.fill[i],
              lty = lty[i],
              lwd = lwd[i])
    }
  }

  ## add plot title
  title(main = main,
        family = layout$kde$font.type$main,
        font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                       layout$kde$font.deco$main],
        col.main = layout$kde$colour$main,
        cex = layout$kde$font.size$main / 12,
        line = (toplines + 1.2) * layout$kde$dimension$main / 100)

  ## optionally add mtext line
  if(mtext != "") {
    mtext(text = mtext,
          side = 3,
          line = 0.5,
          family = layout$kde$font.type$mtext,
          font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                         layout$kde$font.deco$mtext],
          col.main = layout$kde$colour$mtext,
          cex = cex * layout$kde$font.size$mtext / 12)
  }

  ## add summary content
  for(i in 1:length(data)) {
    if (!is.sub) {
      text(x = summary.pos[1],
           y = summary.pos[2],
           adj = summary.adj,
           labels = label.text[[i]],
           col = col.stats[i],
           cex = layout$kde$font.size$stats / 12)
    } else if (mtext == "") {
        mtext(side = 3,
              line = (toplines + 0.3 - i) * layout$kde$dimension$stats.line / 100,
              text = label.text[[i]],
              col = col.stats[i],
              cex = cex * layout$kde$font.size$stats / 12)
    }
  }

  if (values.cumulative) {

    ## create empty overlay plot
    par(new = TRUE) # adjust plot options

    ## add empty plot, scaled to preliminary secondary plot content
    plot(x = NA,
         xlim = xlim.plot,
         ylim = ylim.plot[3:4],
         log  = log.option,
         ann = FALSE,
         axes = FALSE
         )

    ## get line height in xy coordinates
    l_height <- par()$cxy[2]

    ## optionally update ylim
    if (boxplot) {
      ylim.plot[3] <- ylim.plot[3] - 1.4 * l_height
    }

    ## create correctly scaled empty overlay plot
    par(new = TRUE) # adjust plot options

    ## add empty plot, scaled to secondary plot content
    plot(NA,
         xlim = xlim.plot,
         ylim = ylim.plot[3:4],
         log  = log.option,
         ann = FALSE,
         axes = FALSE)

    ## optionally add boxplot
    if (boxplot) {
      ## get extended boxplot data
      boxplot.data <- NULL
      for(i in 1:length(data)) {
        boxplot.i <- graphics::boxplot(x = data[[i]][,1],
                                       plot = FALSE)
        boxplot.i$group <- mean(data[[i]][, 1], na.rm = TRUE)
        boxplot.i$names <- sd(data[[i]][, 1], na.rm = TRUE)
        boxplot.data[[i]] <- boxplot.i
      }

      ## get new line heights
      l_height <- par()$cxy[2]

      for(i in 1:length(data)) {

        # ## draw sd line
        # lines(x = c(boxplot.data[[i]]$group[1] - boxplot.data[[i]]$names[1],
        #             boxplot.data[[i]]$group[1] + boxplot.data[[i]]$names[1]),
        #       y = c(-5/8 * l_height,
        #             -5/8 * l_height),
        #       col = col.mean.line[i])
        #
        # ## draw mean line
        # points(x = boxplot.data[[i]]$group[1],
        #       y = -5/8 * l_height,
        #       pch = 18,
        #       col = col.mean.line[i])

        ## draw median line
        lines(x = c(boxplot.data[[i]]$stats[3,1],
                    boxplot.data[[i]]$stats[3,1]),
              y = c(-11/8, -7/8) * l_height,
              lwd = 2,
              col = col.boxplot.line[i])

        ## draw q25-q75-polygon
        polygon(x = c(boxplot.data[[i]]$stats[2,1],
                      boxplot.data[[i]]$stats[2,1],
                      boxplot.data[[i]]$stats[4,1],
                      boxplot.data[[i]]$stats[4,1]),
                y = c(-11/8, -7/8, -7/8, -11/8) * l_height,
                col = col.boxplot.fill[i],
                border = col.boxplot.line[i])

        ## draw whiskers
        lines(x = c(boxplot.data[[i]]$stats[2,1],
                    boxplot.data[[i]]$stats[1,1]),
              y = c(-9/8, -9/8) * l_height,
              col = col.boxplot.line[i])

        lines(x = c(boxplot.data[[i]]$stats[1,1],
                    boxplot.data[[i]]$stats[1,1]),
              y = c(-10/8, -8/8) * l_height,
              col = col.boxplot.line[i])

        lines(x = c(boxplot.data[[i]]$stats[4,1],
                    boxplot.data[[i]]$stats[5,1]),
              y = c(-9/8, -9/8) * l_height,
              col = col.boxplot.line[i])

        lines(x = c(boxplot.data[[i]]$stats[5,1],
                    boxplot.data[[i]]$stats[5,1]),
              y = c(-10/8, -8/8) * l_height,
              col = col.boxplot.line[i])

        ## draw outliers
        points(x = boxplot.data[[i]]$out,
               y = rep(-9/8 * l_height,
                       length(boxplot.data[[i]]$out)),
               col = col.boxplot.line[i],
               cex = 0.8)
      }
    }

    ## optionally add rug
    if (rug) {
      ticksize <- l_height / max(sapply(data, nrow) * 4)
      for(i in 1:length(data)) {
          suppressWarnings( # some values will be clipped
          rug(data[[i]][, 1],
              ticksize = ticksize,
              side = 3,
              pos = 0,
              lwd = 1,
              col = col.value.rug[i])
          )
      }
    }

    ## add secondary y-axis
    ticks_axis <- pretty(x = c(1, ylim.plot[4]))
    ticks_axis <- ifelse(test = ticks_axis == 0,
                         yes = NA,
                         no = ticks_axis)

    ## add right y-axis
    axis(side = 4,
         at = ticks_axis,
         col = layout$kde$colour$ytck2,
         col.axis = layout$kde$colour$ytck2,
         labels = NA,
         tcl = -layout$kde$dimension$ytck2 / 200,
         cex = cex)

    axis(side = 4,
         at = ticks_axis,
         line = 2 * layout$kde$dimension$ytck2.line / 100 - 2,
         lwd = 0,
         col = layout$kde$colour$ytck2,
         family = layout$kde$font.type$ytck2,
         font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                        layout$kde$font.deco$ytck2],
         col.axis = layout$kde$colour$ytck2,
         cex.axis = layout$kde$font.size$ylab2/12)

    mtext(text = ylab[2],
          side = 4,
          line = 3 * layout$kde$dimension$ylab2.line / 100,
          col = layout$kde$colour$ylab2,
          family = layout$kde$font.type$ylab2,
          font = (1:4)[c("plain", "bold", "italic", "bold italic") ==
                         layout$kde$font.deco$ylab2],
          cex = cex * layout$kde$font.size$ylab2/12)

    ## add De error bars
    for(i in 1:length(data)) {
      if (attr(data[[i]], "De.errors.available")) {
        suppressWarnings( # zero-length arrow is of indeterminate angle and so skipped
        graphics::arrows(data[[i]][, 1] - data[[i]][, 2],
               1:length(data[[i]][,1]),
               data[[i]][, 1] + data[[i]][, 2],
               1:length(data[[i]][, 1]),
               code = 3,
               angle = 90,
               length = 0.05,
               col = col.value.bar[i])
        )
      }
      ## add De measurements
      points(data[[i]][,1], 1:De.stats[i,1],
             col = col.value.dot[i],
             pch = 20)
    }
  }

  ## FUN by R Luminescence Team
  if (fun) sTeve() # nocov

  invisible(list(De.stats = De.stats,
                 summary.pos = summary.pos,
                 De.density = De.density))
}

#' @title Use DRAC to calculate dose rate data
#'
#' @description The function provides an interface from R to DRAC. An R-object or a
#' CSV file is passed to the DRAC website and results are re-imported into R.
#'
#' @param file [character] (**required**):
#' name of a CSV file (formatted according to the DRAC v1.3 CSV template) to
#' be sent to the DRAC website for calculation. It can also be a DRAC template
#' object obtained from [Luminescence::template_DRAC()], which supports also import from
#' CSV-files.
#'
#' @param name [character] (*with default*):
#' Optional user name submitted to DRAC. If `NULL`, a random name will be
#' generated.
#'
#' @param print_references (*with default*):
#' Print all references used in the input data table to the console.
#'
#' @param citation_style (*with default*):
#' If `print_references = TRUE` this argument determines the output style of the
#' used references. Valid options are `"Bibtex"`, `"citation"`, `"html"`, `"latex"`
#' or `"R"`. Default is `"text"`.
#'
#' @param ... further arguments:
#'
#' - `url` [character]: provide an alternative URL to DRAC
#' - `ignore_version` [logical]: ignores the version check, this might come in handy
#' if the version has changed, but not the column order
#' - `user` [character]: option to provide username for secured site
#' - `password` [character]: password for secured site, only works jointly with `user`
#' - `verbose` [logical]: enable/disable output to the terminal
#'
#' @return Returns an [Luminescence::RLum.Results-class] object containing the following elements:
#'
#' \item{DRAC}{[list]: a named list containing the following elements in slot `@@data`:
#'
#' \tabular{lll}{
#'    `$highlights` \tab [data.frame] \tab summary of 25 most important input/output fields \cr
#'    `$header` \tab [character] \tab HTTP header from the DRAC server response \cr
#'    `$labels` \tab [data.frame] \tab descriptive headers of all input/output fields \cr
#'    `$content` \tab [data.frame] \tab complete DRAC input/output table \cr
#'    `$input` \tab [data.frame] \tab DRAC input table \cr
#'    `$output` \tab [data.frame] \tab DRAC output table \cr
#'    `references`\tab [list] \tab A list of bib entries of used references \cr
#' }
#'
#' }
#' \item{data}{[character] or [list] path to the input spreadsheet or a DRAC template}
#' \item{call}{[call] the function call}
#' \item{args}{[list] used arguments}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @seealso [Luminescence::template_DRAC], [Luminescence::.as.latex.table]
#'
#' @section Function version: 0.18
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Christoph Burow, University of Cologne (Germany)
#'
#' @references
#' Durcan, J.A., King, G.E., Duller, G.A.T., 2015. DRAC: Dose Rate and Age Calculator for trapped charge dating.
#' Quaternary Geochronology 28, 54-61. \doi{10.1016/j.quageo.2015.03.012}
#'
#' @examples
#'
#' ## (1) Method using the DRAC spreadsheet
#'
#' file <-  "/PATH/TO/DRAC_Input_Template.csv"
#'
#' # send the actual IO template spreadsheet to DRAC
#' \dontrun{
#' use_DRAC(file = file)
#' }
#'
#' ## (2) Method using an R template object
#' # Create a template
#' input <- template_DRAC(preset = "DRAC-example_quartz")
#'
#' # Fill the template with values
#' input$`Project ID` <- "DRAC-Example"
#' input$`Sample ID` <- "Quartz"
#' input$`Conversion factors` <- "AdamiecAitken1998"
#' input$`External U (ppm)` <- 3.4
#' input$`errExternal U (ppm)` <- 0.51
#' input$`External Th (ppm)` <- 14.47
#' input$`errExternal Th (ppm)` <- 1.69
#' input$`External K (%)` <- 1.2
#' input$`errExternal K (%)` <- 0.14
#' input$`Calculate external Rb from K conc?` <- "N"
#' input$`Calculate internal Rb from K conc?` <- "N"
#' input$`Scale gammadoserate at shallow depths?` <- "N"
#' input$`Grain size min (microns)` <- 90
#' input$`Grain size max (microns)` <- 125
#' input$`Water content ((wet weight - dry weight)/dry weight) %` <- 5
#' input$`errWater content %` <- 2
#' input$`Depth (m)` <- 2.2
#' input$`errDepth (m)` <- 0.22
#' input$`Overburden density (g cm-3)` <- 1.8
#' input$`errOverburden density (g cm-3)` <- 0.1
#' input$`Latitude (decimal degrees)` <- 30.0000
#' input$`Longitude (decimal degrees)` <- 70.0000
#' input$`Altitude (m)` <- 150
#' input$`De (Gy)` <- 20
#' input$`errDe (Gy)` <- 0.2
#'
#' # use DRAC
#' \dontrun{
#' output <- use_DRAC(input)
#'
#' ## export as LaTeX table
#' .as.latex.table(output)
#' }
#'
#' @export
use_DRAC <- function(
  file,
  name = NULL,
  print_references = TRUE,
  citation_style = "text",
  ...
) {
  .set_function_name("use_DRAC")
  on.exit(.unset_function_name(), add = TRUE)

  ## TODO:
  ## (1) Keep the data set as unmodified as possible. Check structure and order of parameters
  ## for meaningful combination.
  ##
  ## (2)
  ## Leave it to the user where the calculations made in our package should be used
  ##
 # Settings ------------------------------------------------------------------------------------
  settings <- modifyList(list(
    name = name %||% paste(sample(c(LETTERS, letters), runif(1, 2, 4)),
                           collapse = ""),
    verbose = TRUE,
    url = "https://www.aber.ac.uk/en/dges/research/quaternary/luminescence-research-laboratory/dose-rate-calculator/?show=calculator",
    ignore_version = FALSE,
    user = NULL,
    password = NULL),
    list(...),
    keep.null = TRUE)

  ## double check URL for validity
  if(!grepl(pattern = "?show=calculator", settings$url, fixed = TRUE))
     .throw_warning("URL is missing '?show=calculator', please check validity!")

  ## Integrity checks -------------------------------------------------------
  .validate_class(file, c("character", "DRAC.list", "DRAC.data.frame"))
  .validate_not_empty(file)

  if (inherits(file, "character")) {
    .validate_file(file, ext = "csv", scan.dir = FALSE,
                   verbose = settings$verbose)
    .validate_length(file, 1)

    ## Import data ----------------------------------------------------------

    ## Import and skip the first rows and remove NA lines and the 2 row, as this row contains
    ## only meta data

    ## DRAC v1.2 - CSV sheet
    if (!read.csv(file, nrows = 1, header = FALSE)[1] %in%
        c("DRAC v.1.2 Inputs", "DRAC v.1.3 Inputs") &
        !settings$ignore_version)
      .throw_error("It looks like that you are not using the original ",
                   "DRAC v1.2 or v1.3 CSV template, this is currently not supported")

    input.raw <- read.csv(file, skip = 8, check.names = FALSE, header = TRUE,
                          stringsAsFactors = FALSE)[-1, ]

  } else if (inherits(file, "DRAC.list")) {
    input.raw <- as.data.frame(file)

  } else if (inherits(file, "DRAC.data.frame")) {
    input.raw <- file
  }

  ## NOTE: if this limit is ever raised above 9999, the generation of
  ## random identifiers below at (4) must be adjusted accordingly to
  ## guarantee uniqueness of identifiers
  if (nrow(input.raw) > 5000)
    .throw_error("The limit of allowed datasets is 5000!")

  .validate_class(name, "character", null.ok = TRUE)
  .validate_logical_scalar(print_references)
  citation_style <- .validate_args(citation_style,
                                   c("text", "Bibtex", "citation", "html",
                                     "latex", "R"))


  ## Helper function --------------------------------------------------------
  ## The data are transferred without encryption, so we mask the original
  .masking <- function(mean, sd, n) {
    temp <- rnorm(n = 30 * n, mean = mean, sd = sd)
    t(vapply(seq(1, length(temp), 30), function(x) {
        c(format(mean(temp[x:(x + 29)]), digits = 2),
          format(sd(temp[x:(x + 29)]), digits = 2))
      }, character(2)))
  }


  # Process data --------------------------------------------------------------------------------
  if (settings$verbose) message("\n\t Preparing data...")
  ##(1) expand the rows in the data.frame a little bit
  mask.df <-  input.raw[rep(1:nrow(input.raw), each = 3), ]

  ##(2) generate some meaningful random variables
  mask.df <- lapply(seq(1, nrow(input.raw), 3), function(x) {
    if (!is.na(mask.df[x,"TI:52"]) && mask.df[x,"TI:52"] != "X") {
      ##replace some values - the De value
      mask.df[x:(x + 2), c("TI:52","TI:53")] <- .masking(
        mean = as.numeric(mask.df[x,"TI:52"]),
        sd = as.numeric(mask.df[x,"TI:53"]),
        n = 3)
      return(mask.df)
    }
  })

  ##(3) bind actual and masked values
  DRAC_submission.df <- rbind(input.raw, mask.df[[1]])
  num.rows.df <- nrow(DRAC_submission.df)

  ##(4) replace ID values
  DRAC_submission.df$`TI:1` <- paste0(
      sample(c(LETTERS, letters), max(4, num.rows.df), replace = TRUE),
      sample(c("", "-", "_"), num.rows.df, replace = TRUE),
      sprintf("%04d", sample(0:9999, num.rows.df)))[1:num.rows.df]

  ##(5) store the real IDs in a separate object
  DRAC_results.id <-  DRAC_submission.df[1:nrow(input.raw), "TI:1"]

  ##(6) create DRAC submission string
  if (settings$verbose) message("\t Creating submission string...")

  ## shuffle the rows of the submission
  DRAC_submission.df <- DRAC_submission.df[sample(num.rows.df), ]

  ## convert all columns of the data.frame to character
  DRAC_submission.df[] <- lapply(DRAC_submission.df, as.character)

  ##paste everything together to get the format we want
  DRAC_input <- paste0(apply(DRAC_submission.df, 1, paste, collapse = " "),
                       "\n", collapse = "")

  # Send data to DRAC ---------------------------------------------------------------------------
  if (settings$verbose)
    message("\t Establishing connection to ", settings$url)

  ## create config list
  conf_l <- list()
  # nocov start
  if(!is.null(settings$user) & !is.null(settings$password))
    conf_l <- httr::authenticate(settings$user, settings$password, type = "basic")
  # nocov end

  ## send data set to DRAC website and receive response
  DRAC.response <- try(httr::POST(
    url = settings$url,
    config = conf_l,
    body = list("drac_data[name]"  = settings$name,
                "drac_data[table]" = DRAC_input)),
    silent = TRUE)

  ## check for correct response
  if (inherits(DRAC.response, "try-error")) {
    .throw_error("Transmission failed with error: ",
                 gsub(".*\n", "", attr(DRAC.response, "condition")$message))
  } else if (DRAC.response$status_code != 200) {
    .throw_error("Transmission failed with HTTP status code: ",
                 DRAC.response$status_code)
  }

  if (settings$verbose)
    message("\t The request was successful, processing the reply...")

  ## assign DRAC response data to variables
  DRAC.content <- httr::content(x = DRAC.response, as = "text")

  ## if the input was valid from a technical standpoint, but not with regard
  ## contents, we indeed get a valid response, but no DRAC output
  if (!grepl("DRAC Outputs", DRAC.content)) {
    on.exit({
      reply <- readline("Do you want to see the DRAC error message (y/N)? ")
      if (reply %in% c("Y", "y", "1")) {
        ## convert to html and extract the error fields
        html <- XML::htmlParse(DRAC.content, asText = TRUE)
        node <- XML::getNodeSet(html, "//div[@class='drac_field_error']")
        error <- sapply(node, function(x) {
          line <- paste0(sapply(XML::xmlChildren(x), XML::saveXML), collapse = "")
          trimws(gsub("\\s*<br/>", "\n", line))
        })

        ## extract the fragment we are given if no drac_field_error is found
        if (length(error) == 0)
          error <- XML::xpathSApply(html, "//body", XML::xmlValue) # nocov

        cat("\n", paste(error, collapse = "\n"), "\n", sep = "")
      }
    }, add = TRUE)

    .throw_error("The response from the server did not contain DRAC output, ",
                 "please check your data and verify its validity.")
  }

  if (settings$verbose) message("\t Finalising the results...")

  ## split header and content
  DRAC.content.split <- strsplit(x = DRAC.content,
                                 split = "DRAC Outputs\n\n")

  ## assign DRAC header part
  DRAC.header <- as.character(DRAC.content.split[[1]][1])

  ## assign DRAC content part
  DRAC.raw <- read.table(text = as.character(DRAC.content.split[[1]][2]),
                         sep = ",",
                         stringsAsFactors = FALSE)

  ## remove first two lines
  DRAC.content <- data.table::fread(as.character(DRAC.content.split[[1]][2]),
                                    sep = ",", skip = 2,
                                    stringsAsFactors = FALSE,
                                    colClasses = c(V3 = "character"),
                                    data.table = FALSE)

  ##Get rid of all the value we do not need anymore
  DRAC.content <-  subset(DRAC.content, DRAC.content$V1 %in% DRAC_results.id)
  DRAC.content <- DRAC.content[match(DRAC_results.id, DRAC.content$V1),]

  ##replace by original names
  DRAC.content[[1]] <- input.raw[[1]]

  ## assign column names
  colnames(DRAC.content) <- DRAC.raw[1, ]

  ## reset rownames
  rownames(DRAC.content) <- seq_len(nrow(DRAC.content))

  ## save column labels and use them as attributes for the I/O table columns
  DRAC.labels <- DRAC.raw[2, ]
  for (i in 1:length(DRAC.content)) {
    attr(DRAC.content[ ,i], "description") <- DRAC.labels[1,i]
  }

  ## DRAC also returns the input, so we need to split input and output
  DRAC.content.input <- DRAC.content[ ,grep("TI:", names(DRAC.content))]
  DRAC.content.output <- DRAC.content[ ,grep("TO:", names(DRAC.content))]

  ## The DRAC ouput also contains a highlight table, which results in
  ## duplicate columns. When creating the data.frame duplicate columns
  ## are automatically appended '.1' in their names, so we can identify
  ## and remove them easily
  DRAC.content.input <- DRAC.content.input[ ,-grep("\\.1", names(DRAC.content.input))]
  DRAC.content.output <- DRAC.content.output[ ,-grep("\\.1", names(DRAC.content.output))]

  ## for some reason the returned input table is unsorted, so we resort it in increasing order
  DRAC.content.input <- DRAC.content.input[ , paste0("TI:", 1:ncol(DRAC.content.input))]

  ## The output table (v1.2) has 198 columns, making it unreasonable complex
  ## for standard data evaluation. We reproduce the DRAC highlight table
  ## and use the descriptions (saved as attributes) as column names.
  highlight.keys <- c("TI:1","TI:2","TI:3","TO:FQ","TO:FR",
                      "TO:FS", "TO:FT", "TO:FU", "TO:FV", "TO:FW",
                      "TO:FX", "TO:FY", "TO:FZ", "TO:GG", "TO:GH",
                      "TO:GI", "TO:GJ", "TO:GK", "TO:GL", "TO:GM",
                      "TO:GN", "TI:52", "TI:53", "TO:GO", "TO:GP")
  DRAC.highlights <- subset(DRAC.content, select = highlight.keys)
  DRAC.highlights.labels <- as.character(DRAC.labels[1, which(unique(names(DRAC.content)) %in% highlight.keys)])
  colnames(DRAC.highlights) <- DRAC.highlights.labels
  for (i in 1:length(DRAC.highlights)) {
    attr(DRAC.highlights[ ,i], "key") <- highlight.keys[i]
  }

  ## finally, we add the 'DRAC.highlights' class so that we can use a custom print method
  class(DRAC.highlights) <- c("DRAC.highlights", "data.frame")

  ## Final Disclaimer
  messages <- list("\t Done! \n",
                   "\t We, the authors of the R package 'Luminescence', do not take any responsibility and we are not liable for any ",
                   "\t mistakes or unforeseen misbehaviour. All calculations are done by DRAC and it is outside our reference to",
                   "\t verify the input and output. \n",
                   "\t Note that this function is only compatible with DRAC version 1.3. Before using this function make sure that",
                   "\t this is the correct version, otherwise expect unspecified errors.\n",
                   "\t Please ensure you cite the use of DRAC in your work, published or otherwise. Please cite the website name and",
                   "\t version (e.g. DRAC v1.3) and the accompanying journal article:",
                   "\t Durcan, J.A., King, G.E., Duller, G.A.T., 2015. DRAC: Dose rate and age calculation for trapped charge",
                   "\t dating. Quaternary Geochronology 28, 54-61. \n",
                   "\t Use 'verbose = FALSE' to hide this message. \n")

  if (settings$verbose) lapply(messages, message)

  ## Get and print used references
  references <- .get_DRAC_references(DRAC.content.input)

  if (print_references && settings$verbose) {
    for (i in 1:length(references$refs)) {
      message("\nReference for: ", references$desc[i])
      print(references$refs[[i]], style = citation_style)
    }
  }

  ## return output
  invisible(set_RLum(
    "RLum.Results",
    data = list(
    DRAC = list(highlights = DRAC.highlights,
          header = DRAC.header,
          labels = DRAC.labels,
          content = DRAC.content,
          input = DRAC.content.input,
          output = DRAC.content.output,
          references = references),
        data = file,
        call = sys.call(),
        args = as.list(sys.call()[-1])))
    )
}

template_v1.2 <- test_path("_data/DRAC_Input_Template_v1.2.csv")
template_v1.3 <- test_path("_data/DRAC_Input_Template_v1.3.csv")

test_that("input validation", {
  testthat::skip_on_cran()

  ## wrong file name
  expect_error(use_DRAC("error"),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_error(use_DRAC(NA),
               "'file' should be of class 'character', 'DRAC.list' or")
  expect_error(use_DRAC(character(0)),
               "'file' cannot be an empty character")
  expect_error(use_DRAC(letters),
               "'file' should have length 1")

  ## XLS input
  file.create(fake.xls <- tempfile(fileext = ".xls"))
  expect_error(use_DRAC(fake.xls),
               "File extension 'xls' is not supported, only 'csv' is valid")

  ## CSV file with the wrong header
  fake <- data.table::fread(template_v1.2)
  fake[1, 1] <- "error"
  fake.csv <- tempfile(fileext = ".csv")
  data.table::fwrite(fake, file = fake.csv)
  expect_error(use_DRAC(fake.csv),
               "you are not using the original DRAC v1.2 or v1.3 CSV template")

  expect_error(use_DRAC(template_v1.2, NA),
               "'name' should be of class 'character' or NULL")

  ## exceed allowed limit
  SW({
  expect_warning(input <- template_DRAC(preset = "DRAC-example_quartz",
                                        nrow = 5001),
                 "More than 5000 datasets might not be supported")
  })
  expect_error(use_DRAC(input), "The limit of allowed datasets is 5000")

  ## citation style
  expect_error(use_DRAC(template_DRAC(preset = "DRAC-example_quartz",
                                      notification = FALSE),
                        citation_style = "error"),
               "[use_DRAC()] 'citation_style' should be one of",
               fixed = TRUE)
})

test_that("check functionality", {
  testthat::skip_on_cran()

 ##use manual example
 ##create template
 SW({
 input <- template_DRAC(preset = "DRAC-example_quartz")
 })

 ##test
 expect_s3_class(input, "DRAC.list")

 ##fill (which also tests the methods)
 input$`Project ID` <- "DRAC-Example"
 input$`Sample ID` <- "Quartz"
 input$`Conversion factors` <- "AdamiecAitken1998"
 input$`External U (ppm)` <- 3.4
 input$`errExternal U (ppm)` <- 0.51
 input$`External Th (ppm)` <- 14.47
 input$`errExternal Th (ppm)` <- 1.69
 input$`External K (%)` <- 1.2
 input$`errExternal K (%)` <- 0.14
 input$`Calculate external Rb from K conc?` <- "Y"
 input$`Calculate internal Rb from K conc?` <- "Y"
 input$`Scale gammadoserate at shallow depths?` <- "Y"
 input$`Grain size min (microns)` <- 90L
 input$`Grain size max (microns)` <- 125L
 input$`Water content ((wet weight - dry weight)/dry weight) %` <- 5
 input$`errWater content %` <- 2
 input$`Depth (m)` <- 2.2
 input$`errDepth (m)` <- 0.22
 input$`Overburden density (g cm-3)` <- 1.8
 input$`errOverburden density (g cm-3)` <- 0.1
 input$`Latitude (decimal degrees)` <- 30.0000
 input$`Longitude (decimal degrees)` <- 70.0000
 input$`Altitude (m)` <- 150
 input$`De (Gy)` <- 20
 input$`errDe (Gy)` <- 0.2

 ##run DRAC
 SW({
 output <- expect_s4_class(use_DRAC(input), "RLum.Results")
 })

 ## print method for DRAC.highlights
 expect_output(print(output$DRAC$highlights), regexp = "TO:GP = errAge")

 ## DRAC.data.frame
 input.df <- as.data.frame(input)
 class(input.df) <- c("data.frame", "DRAC.data.frame")
 expect_s4_class(use_DRAC(input.df, verbose = FALSE),
                 "RLum.Results")

  SW({
  ## CSV input
  expect_message(expect_s4_class(
      use_DRAC(template_v1.2,
               verbose = TRUE),
      "RLum.Results"),
      "Reference for")

  ## communicate with insufficient input
  t <- template_DRAC(notification = FALSE)
  local_mocked_bindings(readline = function(prompt) "Y")
  expect_output(expect_error(
      use_DRAC(t),
      "The response from the server did not contain DRAC output"),
      "ERROR: Some fields were invalid")

  ## communicate with fake url
  expect_error(expect_warning(
      use_DRAC(t, url = "iamnotvali8793270942hd.valid"),
      "URL is missing '\\?show=calculator', please check validity!"),
      "Transmission failed with error: Could not resolve host: iamnotvali8793270942hd.valid")

  ## mock error status code
  local_mocked_bindings(POST = function(...) list(status_code = 313),
                        .package = "httr")
  expect_error(use_DRAC(t),
               "Transmission failed with HTTP status code: 313")
 })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)

  SW({
  expect_snapshot_output(use_DRAC(template_v1.3))
  })

  ## test print.DRAC.highlight()
  t <- template_DRAC(preset = "DRAC-example_quartz", notification = FALSE)
  res <- use_DRAC(t, verbose = FALSE)
  expect_snapshot_output(print(res$DRAC$highlights))

  ## communicate with insufficient input
  t[52] <- 0.0
  local_mocked_bindings(readline = function(prompt) "Y")
  set.seed(1)
  SW({
  expect_snapshot_output(expect_error(
      use_DRAC(t),
      "The response from the server did not contain DRAC output"))
  })
})

#' @title Merge function for RLum.Analysis S4 class objects
#'
#' @description
#' This function simply allows to merge [Luminescence::RLum.Analysis-class]
#' objects. Moreover, other [Luminescence::RLum.Data-class] objects can be added
#' to an existing [Luminescence::RLum.Analysis-class] object. Supported objects
#' to be added are: [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Spectrum-class] and
#' [Luminescence::RLum.Data.Image-class].
#'
#' @param objects [list] of [Luminescence::RLum.Analysis-class] (**required**):
#' list of S4 objects of class `RLum.Analysis`. Furthermore other objects of
#' class [Luminescence::RLum-class] can be added, see details.
#'
#' @return
#' Returns an [Luminescence::RLum.Analysis-class] object ordered according to the order
#' provided with the input list.
#'
#' @note
#' The information for the slot 'protocol' is taken from the first
#' [Luminescence::RLum.Analysis-class] object in the input list. Therefore at
#' least one object of type [Luminescence::RLum.Analysis-class] has to be provided.
#'
#' @section Function version: 0.2.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::merge_RLum], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Spectrum-class],
#' [Luminescence::RLum.Data.Image-class], [Luminescence::RLum-class]
#'
#' @keywords utilities internal
#'
#' @examples
#'
#' ##merge different RLum objects from the example data
#' data(ExampleData.RLum.Analysis, envir = environment())
#' data(ExampleData.BINfileData, envir = environment())
#'
#' object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#' curve <- get_RLum(object)[[2]]
#'
#' temp.merged <- merge_RLum.Analysis(list(curve, IRSAR.RF.Data, IRSAR.RF.Data))
#'
#' @export
merge_RLum.Analysis<- function(
  objects
) {
  .set_function_name("merge_RLum.Analysis")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(objects, "list")
  .validate_not_empty(objects)

  ##check if object is of class RLum
  temp.class.test <- sapply(objects, function(x) {
    .validate_class(x, c("RLum.Analysis", "RLum.Data"),
                    name = "All elements of 'object'")
    class(x)[1]
  })

  ##check if at least one object of RLum.Analysis is provided
  if(!"RLum.Analysis"%in%temp.class.test){
    .throw_error("At least one input object in the list ",
                 "has to be of class 'RLum.Analysis'")
  }


  ## Merge objects ----------------------------------------------------------

  ##(0) get recent environment to later set variable temp.meta.data.first
  temp.environment  <- environment()
  temp.meta.data.first <- NULL

  ##(1) collect all elements in a list
  temp.element.list <- unlist(lapply(objects, function(x) {
    if (inherits(x, "RLum.Data"))
      return(x)

    ## x is an RLum.Analysis object
    ## extract meta data from the first RLum.Analysis object
    if (is.null(temp.meta.data.first)) {
      assign("temp.meta.data.first", x@protocol, envir = temp.environment)
    }

    get_RLum(x)
  }))

  ## return new RLum.Analysis object
  set_RLum(
    class = "RLum.Analysis",
    originator = "merge_RLum.Analysis",
    records = temp.element.list,
    protocol = temp.meta.data.first,
    info = unlist(lapply(objects, function(x) {
      x@info
    }), recursive = FALSE),
    .pid = unlist(lapply(objects, function(x) {
      x@.uid
    }))
  )
}

## load data
data(ExampleData.RLum.Analysis, envir = environment())
o1 <- IRSAR.RF.Data
c1 <- as(object = list(1:10), Class = "RLum.Data.Curve")
r1 <- as(object = list(1:10), Class = "RLum.Results")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(merge_RLum.Analysis(),
               "'objects' should be of class 'list'")
  expect_error(merge_RLum.Analysis(o1),
               "'objects' should be of class 'list'")
  expect_error(merge_RLum.Analysis(list()),
               "'objects' cannot be an empty list")
  expect_error(merge_RLum.Analysis(list(c1)),
               "At least one input object in the list has to be of class")
  expect_error(merge_RLum.Analysis(list(NULL)),
               "All elements of 'object' should be of class 'RLum.Analysis' or")
  expect_error(merge_RLum.Analysis(list(o1, "test")),
               "All elements of 'object' should be of class 'RLum.Analysis' or")
  expect_error(merge_RLum.Analysis(list(o1, r1)),
               "All elements of 'object' should be of class 'RLum.Analysis' or")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_RLum(merge_RLum.Analysis(list(o1)))
  expect_snapshot_RLum(merge_RLum.Analysis(list(o1@records[[2]], o1, c1)))
})

#' @title Attempts to install the development version of the 'Luminescence' package
#'
#' @description
#' This function provides a convenient method for installing the development
#' version of the R package 'Luminescence' directly from GitHub.
#'
#' @details
#' This function checks whether the 'devtools' package is currently installed
#' on the system. If `force_install = TRUE` the functions attempts to install
#' the chosen development branch via [devtools::install_github].
#'
#' @param force_install [logical] (*optional*):
#' If `FALSE` (the default) the function produces and prints the required
#' code to the console for the user to run manually afterwards. When `TRUE`
#' and all requirements are fulfilled (see details) this function attempts to install
#' the package itself.
#'
#' @param branch [character] (*with default*):
#' Name of the branch to install. The default value ("master") corresponds to
#' the main development branch.
#'
#' @return
#' This function requires user input at the command prompt to choose the
#' desired development branch to be installed. The required R code to install
#' the package is then printed to the console.
#'
#' @examples
#' \dontrun{
#' install_DevelopmentVersion()
#' }
#'
#' @export
install_DevelopmentVersion <- function(force_install = FALSE, branch = "master") {
  .set_function_name("install_DevelopmentVersion")
  on.exit(.unset_function_name(), add = TRUE)

  ## input validation
  .validate_logical_scalar(force_install)
  .validate_class(branch, "character", length = 1)

  message("\n[install_DevelopmentVersion]\n")
  message("----\n",
          "For package prerequisites, make sure to have read the following:\n",
          "https://github.com/R-Lum/Luminescence/blob/master/README.md\n",
          "----\n")

  ## check if 'devtools' is available and install if not
  if (!requireNamespace("devtools", quietly = TRUE)) {
    # nocov start
    message("Please install the 'devtools' package first by running ",
            "the following command:\n\n install.packages('devtools')\n")
    return(invisible())
    # nocov end
  }

  if (!force_install) {
    message("Please copy and run the following code in your R terminal:\n",
            "devtools::install_github('R-Lum/luminescence@", branch, "')\n")
    return(invisible())
  }

  # nocov start
  reply <- NULL
  while (is.null(reply)) {
    message("Proceed with the installation?\n",
              " [n/N]: No\n",
              " [y/Y]: Yes\n")
    reply <- tolower(readline())

    if (reply == "n") {
      message("Nothing done")
      return(invisible())
    }
    if (reply != "y")
        reply <- NULL
  }

    # detach the 'Luminescence' package
    try(detach(name = "package:Luminescence", unload = TRUE, force = TRUE),
        silent = TRUE)

    # try to unload the dynamic library
    dynLibs <- sapply(.dynLibs(), function(x) x[["path"]] )
    try(dyn.unload(grep("Luminescence", dynLibs, value = TRUE, fixed = TRUE)),
        silent = TRUE)

    # install the development version
    devtools::install_github(paste0("r-lum/luminescence@", branch))
  # nocov end
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(install_DevelopmentVersion("error"),
               "'force_install' should be a single logical value")
  expect_error(install_DevelopmentVersion(TRUE, FALSE),
               "'branch' should be of class 'character' and have length 1")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  expect_message(install_DevelopmentVersion(FALSE, "master"),
                 "Please copy and run the following code in your R terminal")
  })
})

#' @title Import XSYG files into R
#'
#' @description Imports XSYG-files produced by a Freiberg Instruments lexsyg reader into R.
#'
#' @details
#' **How does the import function work?**
#'
#' The function uses the `'XML'` package to parse the file structure. Each
#' sequence is subsequently translated into an [Luminescence::RLum.Analysis-class] object.
#'
#' **General structure XSYG format**
#'
#' ```
#' <?xml?>
#' <Sample>
#'   <Sequence>
#'     <Record>
#'       <Curve name="first curve" />
#'       <Curve name="curve with data">x0 , y0 ; x1 , y1 ; x2 , y2 ; x3 , y3</Curve>
#'     </Record>
#'   </Sequence>
#' </Sample>
#' ```
#'
#' So far, each
#' XSYG file can only contain one `<Sample></Sample>`, but multiple
#' sequences.
#'
#' Each record may comprise several curves.
#'
#' **TL curve recalculation**
#'
#' On the FI lexsyg device TL curves are recorded as time against count values.
#' Temperature values are monitored on the heating plate and stored in a
#' separate curve (time vs. temperature). If the option
#' `recalculate.TL.curves = TRUE` is chosen, the time values for each TL
#' curve are replaced by temperature values.
#'
#' Practically, this means combining two matrices (Time vs. Counts and Time vs.
#' Temperature) with different row numbers by their time values. Three cases
#' are considered:
#'
#' 1. HE: Heating element
#' 2. PMT: Photomultiplier tube
#' 3. Interpolation is done using the function [approx]
#'
#' CASE (1): `nrow(matrix(PMT))` > `nrow(matrix(HE))`
#'
#' Missing temperature values from the heating element are calculated using
#' time values from the PMT measurement.
#'
#' CASE (2): `nrow(matrix(PMT))` < `nrow(matrix(HE))`
#'
#' Missing count values from the PMT are calculated using time values from the
#' heating element measurement.
#'
#' CASE (3): `nrow(matrix(PMT))` == `nrow(matrix(HE))`
#'
#' A new matrix is produced using temperature values from the heating element
#' and count values from the PMT.
#'
#' **Note:**
#' Please note that due to the recalculation of the temperature
#' values based on values delivered by the heating element, it may happen that
#' multiple count values exists for each temperature value and temperature
#' values may also decrease during heating, not only increase.
#'
#' **Count linearity correction**
#'
#' If the argument `auto_linearity_correction = TRUE`, the function offers a,
#' theoretically, automated linearity correction of the PMT signal using
#' the internal function [Luminescence::correct_PMTLinearity]. The critical parameter
#' is the count-pair resolution, which is provided to the function
#' automatically depending on the detector. Currently, the following
#' settings are used:
#'
#' \tabular{ll}{
#' DETECTOR \tab COUNT-PAIR-RESOLUTION\cr
#' UVVIS \tab 18 ns\cr
#' NIR50 \tab 70 ns\cr
#' NIR40 \tab 70 ns \cr
#' ETPMT \tab 25 ns
#' }
#'
#' Unfortunately, not all XSYG files provide correct information on
#' the used detector, depending on software version. Hence, in case
#' of doubt, please verify the settings and conduct a manual correction
#' if required.
#'
#' **Advanced file import**
#'
#' To allow for a more efficient usage of the function, instead of single path
#' to a file just a directory can be passed as input. In this particular case
#' the function tries to extract all XSYG-files found in the directory and import
#' them all. Using this option internally the function constructs as list of
#' the XSYG-files found in the directory. Please note no recursive detection
#' is supported as this may lead to endless loops.
#'
#' @param file [character] or [list] (**required**):
#' name of one or multiple XSYG files (URLs are supported); it can be the path
#' to a directory, in which case the function tries to detect and import all
#' XSYG files found recursively from the given directory.
#'
#' @param recalculate.TL.curves [logical] (*with default*):
#' if set to `TRUE`, TL curves are returned as temperature against count values
#' (see details for more information) Note: The option overwrites the time vs.
#' count TL curve. Select `FALSE` to import the raw data delivered by the
#' lexsyg. Works for TL curves and spectra.
#'
#' @param n_records [numeric] (*with default*):
#' number of records to be imported; by default the function attempts to import
#' all records.
#'
#' @param fastForward [logical] (*with default*):
#' if `TRUE` for a more efficient data processing only a list of
#' [Luminescence::RLum.Analysis-class] objects is returned.
#'
#' @param import [logical] (*with default*):
#' if set to `FALSE`, only the XSYG file structure is shown.
#'
#' @param pattern [character] (*with default*):
#' regular expression pattern passed to [list.files] to construct a list of
#' files to read (used only when a path is provided).
#'
#' @param auto_linearity_correction [logical] (*with default*): enable/disable
#' automatic count linearity correction. Because the information on the detectors
#' are not consistent and are not consistently stored, this option should be used
#' with caution.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar during import. Ignored if `verbose = FALSE`.
#'
#' @return
#' **Using the option `import = FALSE`**
#'
#' A list consisting of two elements is shown:
#' - [data.frame] with information on file.
#' - [data.frame] with information on the sequences stored in the XSYG file.
#'
#' **Using the option `import = TRUE` (default)**
#'
#' A list is provided, the list elements
#' contain: \item{Sequence.Header}{[data.frame] with information on the
#' sequence.} \item{Sequence.Object}{[Luminescence::RLum.Analysis-class]
#' containing the curves.}
#'
#' @note
#' This function is a beta version as the XSYG file format is not yet
#' fully specified. Thus, further file operations (merge, export, write) should
#' be done using the functions provided with the package `'XML'`.
#'
#' **So far, no image data import is provided!** \cr
#' Corresponding values in the XSXG file are skipped.
#'
#' @section Function version: 0.8.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso `'XML'`, [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class],
#' [approx], [Luminescence::correct_PMTLinearity]
#'
#' @references
#' Grehl, S., Kreutzer, S., Hoehne, M., 2013. Documentation of the
#' XSYG file format. Unpublished Technical Note. Freiberg, Germany
#'
#' **Further reading**
#'
#' XML: [https://en.wikipedia.org/wiki/XML]()
#'
#' @keywords IO
#'
#' @examples
#'
#' ##(1) import XSYG file to R (uncomment for usage)
#'
#' #FILE <- file.choose()
#' #temp <- read_XSYG2R(FILE)
#'
#' ##(2) additional examples for pure XML import using the package XML
#' ##    (uncomment for usage)
#'
#'   ##import entire XML file
#'   #FILE <- file.choose()
#'   #temp <- XML::xmlRoot(XML::xmlTreeParse(FILE))
#'
#'   ##search for specific subnodes with curves containing 'OSL'
#'   #getNodeSet(temp, "//Sample/Sequence/Record[@@recordType = 'OSL']/Curve")
#'
#' ##(2) How to extract single curves ... after import
#' data(ExampleData.XSYG, envir = environment())
#'
#' ##grep one OSL curves and plot the first curve
#' OSLcurve <- get_RLum(OSL.SARMeasurement$Sequence.Object, recordType="OSL")[[1]]
#'
#' ##(3) How to see the structure of an object?
#' structure_RLum(OSL.SARMeasurement$Sequence.Object)
#'
#' @export
read_XSYG2R <- function(
  file,
  recalculate.TL.curves = TRUE,
  n_records = NULL,
  fastForward = FALSE,
  import = TRUE,
  pattern = "\\.xsyg",
  auto_linearity_correction = FALSE,
  verbose = TRUE,
  txtProgressBar = TRUE
) {
  .set_function_name("read_XSYG2R")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO: this function should be reshaped:
  ##  - metadata from the sequence should go into the info slot of the RLum.Analysis object
  ##   >> however, the question is whether this works with subsequent functions
  ##  - currently not all metadata are supported, it should be extended
  ##  - the should be a mode importing ALL metadata
  ##  - xlum should be general, xsyg should take care about subsequent details

  .validate_logical_scalar(verbose)
  .validate_class(pattern, "character")
  file <- .validate_file(file, pattern = pattern, recursive = TRUE,
                         throw.error = FALSE, verbose = verbose)
  if (length(file) == 0)
    return(NULL)
  .validate_positive_scalar(n_records, int = TRUE, null.ok = TRUE)

  # Self Call -----------------------------------------------------------------------------------
  # Option (a): Input is a list, every element in the list will be treated as file connection
  # with that many file can be read in at the same time
  # Option (b): The input is just a path, the function tries to grep ALL xsyg/XSYG files in the
  # directory and import them, if this is detected, we proceed as list
  if (inherits(file, "list")) {
    temp.return <- lapply(seq_along(file), function(x) {
      read_XSYG2R(
        file = file[[x]],
        recalculate.TL.curves = recalculate.TL.curves,
        fastForward = fastForward,
        import = import,
        verbose = verbose,
        txtProgressBar = txtProgressBar
      )
    })

    if (!fastForward)
      return(temp.return)
    if (import)
      return(unlist(temp.return, recursive = FALSE))

    return(as.data.frame(data.table::rbindlist(temp.return)))
  }

  ## Integrity checks -------------------------------------------------------
  .validate_logical_scalar(fastForward)
  .validate_logical_scalar(import)
  .validate_logical_scalar(auto_linearity_correction)
  .validate_logical_scalar(txtProgressBar)

  ## don't show the progress bar if not verbose
  if (!verbose)
    txtProgressBar <- FALSE

  # (0) config --------------------------------------------------------------
  #version.supported <- c("1.0")

  #additional functions
  #get spectrum values
  # TODO: This function could be written also in C++, however, not necessary due to a low demand
  get_XSYG.spectrum.values <- function(curve.node){
    ##1st grep wavelength table
    wavelength <- XML::xmlAttrs(curve.node)["wavelengthTable"]

    ##string split (scan is considerably faster by a factor of 2)
    wavelength <- na.exclude(scan(text = wavelength, what = numeric(), sep = ";", quiet = TRUE))

    ##2nd grep time values
    raw_curve_node <- XML::xmlValue(curve.node)
    curve.node <- unlist(strsplit(raw_curve_node, split = "];", fixed = TRUE))
    curve.node <- strsplit(curve.node, split = ",[", fixed = TRUE)
    curve.node.time <- as.numeric(vapply(curve.node, function(x) x[1], character(1)))

    ##3rd grep count values
    curve.node.count <- vapply(curve.node, function(x) {
      x[length(x)]
    }, character(1))

    ## remove last bracket
    curve.node.count[length(curve.node.count)] <- gsub(
      pattern = "]", replacement = "", x = curve.node.count[length(curve.node.count)], fixed = TRUE)

    ##4th combine to spectrum matrix
    spectrum.matrix <- matrix(NA, nrow = length(wavelength), ncol = length(curve.node.time))
    for(i in 1:length(curve.node.time)) {
      tmp <- scan(text = curve.node.count[i], what = numeric(), sep = "|", quiet = TRUE)
      if(length(tmp) == length(wavelength))
        spectrum.matrix[,i] <- tmp
    }

    ## remove NA values from matrix
    id_NA <- colSums(is.na(spectrum.matrix)) != nrow(spectrum.matrix)
    spectrum.matrix <- spectrum.matrix[, id_NA]

    ##change row names (rows are wavelength)
    rownames(spectrum.matrix) <- round(wavelength, digits = 3)

    ##change column names (columns are time/temp values)
    colnames(spectrum.matrix) <- round(curve.node.time[id_NA], digits=3)

    return(spectrum.matrix)
  }

  # (1) Integrity checks ----------------------------------------------------
  ##set HUGE for larger nodes
  HUGE <- 524288

  ##parse XML tree using the package XML
  temp <- try(
    XML::xmlRoot(XML::xmlTreeParse(file, useInternalNodes = TRUE, options = HUGE, error = NULL)),
    silent = TRUE)

  ##show error
  if(inherits(temp, "try-error")){
    if(verbose)
      .throw_message("XML file not readable, nothing imported, NULL returned")
    return(NULL)
  }

  # (2) Further file processing ---------------------------------------------
  ##==========================================================================##
  ##SHOW STRUCTURE
  if(!import){
    ##sample information
    temp.sample <- as.data.frame(XML::xmlAttrs(temp), stringsAsFactors = FALSE)

    ##grep sequences files

    ##set data.frame
    names.sequence.header <- names(XML::xmlAttrs(temp[[1]]))
    temp.sequence.header <- data.frame(t(1:length(names.sequence.header)),
                                       stringsAsFactors = FALSE)
    colnames(temp.sequence.header) <- names.sequence.header

    ##fill information in data.frame
    for(i in 1:XML::xmlSize(temp)){
      temp.sequence.header[i,] <- t(XML::xmlAttrs(temp[[i]]))
    }

      ##additional option for fastForward == TRUE
      if(fastForward){
        ##change column header
        temp.sample <- t(temp.sample)
        colnames(temp.sample) <- paste0("sample::", colnames(temp.sample))
        output <- cbind(temp.sequence.header, temp.sample)

      }else{
        output <-  list(Sample = temp.sample, Sequences = temp.sequence.header)
      }

    return(output)
  }

  ## ========================================================================
  ## IMPORT XSYG FILE

  if (verbose) {
      cat("\n[read_XSYG2R()] Importing ...")
      cat("\n path: ", dirname(file))
      cat("\n file: ", .shorten_filename(basename(file)))
      cat("\n")
    }

  ## set n_records
  n_records <- min(XML::xmlSize(temp), n_records)

  ## initialize the progress bar
  if (txtProgressBar) {
    pb <- txtProgressBar(min = 0, max = n_records, char = "=", style = 3)
  }

  ## loop over the entire sequence up to the number of records to be read
  output <- lapply(1:n_records, function(x) {
    sequence <- temp[[x]]
    sequence.header <- as.data.frame(XML::xmlAttrs(sequence),
                                     stringsAsFactors = FALSE)

    ## account for non set value
    if (length(sequence.header) != 0)
        colnames(sequence.header) <- ""

    ###----------------------------------------------------------------------
    ##LOOP
    ##read records >> records are combined to one RLum.Analysis object
    sequence.object <- unlist(lapply(1:XML::xmlSize(sequence), function(i) {
      record <- sequence[[i]]

      ## the XSYG file might be broken due to a machine error during the measurement
      recordType <- try(XML::xmlAttrs(record)["recordType"], silent = TRUE)
      if (inherits(recordType, "try-error"))
          return(NULL) # nocov

      ## create a fallback, the function should not fail
      if (is.null(recordType) || is.na(recordType)) {
        recordType <- "not_set"
      }

      ## correct record type in depending on the stimulator
      xml.size <- XML::xmlSize(record)
      if (recordType == "OSL" &&
          XML::xmlAttrs(record[[xml.size]])["stimulator"] %in%
              c("ir_LED_850", "ir_LD_850")) {
          recordType  <- "IRSL"
      }

      ## get all record attributes
      attrs_record <- XML::xmlAttrs(record)
      names(attrs_record) <- gsub("^name$", "recordName", names(attrs_record))

      header.position <- as.integer(as.character(sequence.header["position", ]))
      header.name <- as.character(sequence.header["name", ])

      ## loop 3rd level
      lapply(1:xml.size, function(j) {
        curve <- record[[j]]
        attrs <- XML::xmlAttrs(curve)

        ## all curves after the first in a record are marked with a leading
        ## underscore: this should make it easier to identify the curve to
        ## analyze (the first) from the others
        recordType <- paste0(ifelse(j == 1, "", "_"), recordType)

          ##get curveType
          temp.sequence.object.curveType <- as.character(attrs["curveType"])

          ##get detector
          temp.sequence.object.detector <- as.character(attrs["detector"])

          ## combine attributes
          attrs_comb <- as.list(c(attrs, attrs_record))
          attrs_comb <- attrs_comb[!duplicated(names(attrs_comb))]

        ## get additional information
        temp.sequence.object.info <- modifyList(attrs_comb,
                                                list(position = header.position,
                                                     sequenceName = header.name))

        ## TL curve recalculation ===========================================
        if (recalculate.TL.curves) {
            ##TL curve heating values is stored in the 3rd curve of every set
            if (recordType == "TL" && j == 1) {
              #grep values from PMT measurement or spectrometer
              if(!"Spectrometer" %in% temp.sequence.object.detector){
                temp.sequence.object.curveValue.PMT <- src_get_XSYG_curve_values(XML::xmlValue(
                  record[[j]]))

                time.values <- temp.sequence.object.curveValue.PMT[, 1]
              }else{
                temp.sequence.object.curveValue.spectrum <-
                  get_XSYG.spectrum.values(curve)

                ##get time values which are stored in the row labels
                time.values <- as.numeric(
                  colnames(temp.sequence.object.curveValue.spectrum))
              }

              ## round values (1 digit is technical resolution of the heating element)
              time.values <- round(time.values, digits = 1)

              #grep values from heating element
              temp.sequence.object.curveValue.heating.element <-
                src_get_XSYG_curve_values(XML::xmlValue(record[[3]]))

              temp.element <- temp.sequence.object.curveValue.heating.element[, 1]

              ## reduce matrix values to values of the detection
              temp.sequence.object.curveValue.heating.element <-
                  temp.sequence.object.curveValue.heating.element[
                      temp.element >= min(time.values) &
                      temp.element <= max(time.values), ,
                      drop = FALSE]

              ## calculate corresponding heating rate, this makes only sense
              ## for linear heating, therefore it has to be the maximum value

              ##remove 0 values (not measured) and limit to peak
              heating.rate.values <- temp.sequence.object.curveValue.heating.element[
                temp.sequence.object.curveValue.heating.element[,2] > 0 &
                  temp.sequence.object.curveValue.heating.element[,2] <=
                  max(temp.sequence.object.curveValue.heating.element[,2]),,drop = FALSE]

              heating.rate <- (heating.rate.values[length(heating.rate.values[,2]), 2] -
                                 heating.rate.values[1,2])/
                (heating.rate.values[length(heating.rate.values[,1]), 1] -
                   heating.rate.values[1,1])

              ##round values
              heating.rate <- round(heating.rate, digits=1)

              ##add to info element
              temp.sequence.object.info <- c(temp.sequence.object.info,
                                             RATE = heating.rate)

              ##PERFORM RECALCULATION
              ##check which object contains more data
              if(!"Spectrometer" %in% temp.sequence.object.detector){
                ##CASE (1)
                if(nrow(temp.sequence.object.curveValue.PMT) >
                   nrow(temp.sequence.object.curveValue.heating.element)){
                  temp.sequence.object.curveValue.heating.element.i <- approx(
                    x = temp.sequence.object.curveValue.heating.element[,1],
                    y = temp.sequence.object.curveValue.heating.element[,2],
                    xout = time.values,
                    rule = 2)

                  temperature.values <-
                    temp.sequence.object.curveValue.heating.element.i$y

                  count.values <-
                    temp.sequence.object.curveValue.PMT[,2]

                  ##CASE (2)
                }else if((nrow(temp.sequence.object.curveValue.PMT) <
                          nrow(temp.sequence.object.curveValue.heating.element))){
                  temp.sequence.object.curveValue.PMT.i <- approx(
                    x = time.values,
                    y = temp.sequence.object.curveValue.PMT[,2],
                    xout = temp.sequence.object.curveValue.heating.element[,1],
                    rule = 2)

                  temperature.values <-
                    temp.sequence.object.curveValue.heating.element[,2]

                  count.values <- temp.sequence.object.curveValue.PMT.i$y

                  ##CASE (3)
                }else{
                  temperature.values <-
                    temp.sequence.object.curveValue.heating.element[,2]

                  count.values <- temp.sequence.object.curveValue.PMT[,2]
                }

                ##combine as matrix
                curve <- cbind(temperature.values, count.values)

                ##set curve identifier
                temp.sequence.object.info$curveDescripter <- "Temperature [\u00B0C]; Counts [a.u.]"

              }else{
                ##CASE (1) here different approach. in contrast to the PMT measurements, as
                ##         usually the resolution should be much, much lower for such measurements
                ##         Otherwise we would introduce some pseudo signals, as we have to
                ##         take care of noise later one

                if (length(time.values) !=
                    nrow(temp.sequence.object.curveValue.heating.element)) {
                  temp.sequence.object.curveValue.heating.element.i <- approx(
                    x = temp.sequence.object.curveValue.heating.element[,1],
                    y = temp.sequence.object.curveValue.heating.element[,2],
                    xout = time.values,
                    rule = 2,
                    ties = mean,
                    na.rm = FALSE)

                  temperature.values <-
                    temp.sequence.object.curveValue.heating.element.i$y

                  ##check for duplicated values and if so, increase this
                  idx.dup <- which(duplicated(temperature.values))
                  if (length(idx.dup) > 0) {
                    temperature.values[idx.dup] <- temperature.values[idx.dup] + 1
                    .throw_warning("Temperature values are found to be ",
                                   "duplicated and increased by 1 K.")
                  }
                  ##CASE (2)  (equal)
                }else{
                  ##CASE (2)  (equal)
                  temperature.values <-
                    temp.sequence.object.curveValue.heating.element[,2]
                }

                ##reset values of the matrix
                colnames(temp.sequence.object.curveValue.spectrum) <- temperature.values
                curve <- temp.sequence.object.curveValue.spectrum

                ##change curve descriptor
                temp.sequence.object.info$curveDescripter <- "Temperature [\u00B0C]; Wavelength [nm]; Counts [1/ch]"
              }
            }##endif
          }##endif recalculate.TL.curves == TRUE


          ## Cleanup info objects -------------------------------------------
          if ("curveType" %in% names(temp.sequence.object.info))
            temp.sequence.object.info[["curveType"]] <- NULL

          ## Set RLum.Data-objects ------------------------------------------
          if (!"Spectrometer" %in% temp.sequence.object.detector) {
            out.class <- "RLum.Data.Curve"
            if (!inherits(curve, "matrix")) {
              curve <- src_get_XSYG_curve_values(XML::xmlValue(curve))
            }
          } else {
            out.class <- "RLum.Data.Spectrum"
            if (!inherits(curve, "matrix")) {
              curve <- get_XSYG.spectrum.values(curve)
            }
          }

          set_RLum(
              class = out.class,
              originator = "read_XSYG2R",
              recordType = paste0(recordType,
                                  " (", temp.sequence.object.detector,")"),
              curveType = temp.sequence.object.curveType,
              data = curve,
              info = temp.sequence.object.info)
        })
    }), use.names = FALSE)

      ##if the XSYG file is broken we get NULL as list element
      if (!is.null(sequence.object)) {
        ##set RLum.Analysis object
        sequence.object <- set_RLum(
          originator = "read_XSYG2R",
          class = "RLum.Analysis",
          records = sequence.object,
          protocol = as.character(sequence.header["protocol", 1]),
          info = list(file = file)
        )

        ##set parent uid of RLum.Anlaysis as parent ID of the records
        sequence.object <- .set_pid(sequence.object)

        ##update progress bar
        if (txtProgressBar) {
          setTxtProgressBar(pb, x)
        }

        ##merge output and return values
        if(fastForward){
          return(sequence.object)
        }
        return(list(Sequence.Header = sequence.header,
                    Sequence.Object = sequence.object))
      }
  }) ##end loop for sequence list

  ## close ProgressBar
  if (txtProgressBar)
    close(pb)

  ## show output information
  num.removed <- length(output[sapply(output, is.null)])
  if (num.removed > 0)
    .throw_warning(num.removed, " incomplete sequence(s) removed")

  if (verbose) {
    cat(sprintf("\t >> %d of %d sequence(s) loaded successfully\n",
                n_records - num.removed, n_records))
  }

  ##get rid of the NULL elements (as stated before ... invalid files)
  output <- .rm_NULL_elements(output)

  ## account for linearity
  if (auto_linearity_correction) {
    ## set look-up table for common FI PMTs
    count_pair_res <- c(
      UVVIS = 18,
      NIR50 = 70,
      NIR40 = 70,
      ETPMT = 25)

    ## correct only curves that can be corrected
    output <- lapply(output, \(x) {
      x@records <- lapply(x@records, \(y) {
        detector <- y@info$detector
        if (!is.null(detector) && any(detector %in% names(count_pair_res))) {
          y <- correct_PMTLinearity(y, PMT_pulse_pair_resolution = count_pair_res[detector])
        }
        return(y)
      })
      return(x)
    })
  }

  ## return object
  return(output)
}

## load data
xsyg <- read_XSYG2R(
  system.file("extdata/XSYG_file.xsyg", package = "Luminescence"),
  fastForward = TRUE,
  verbose = FALSE
)

xsyg_v1 <- set_RLum("RLum.Analysis", records = list(
  xsyg[[1]]@records[[1]],
  xsyg[[1]]@records[[4]],
  xsyg[[1]]@records[[4]],
  xsyg[[1]]@records[[10]],
  xsyg[[1]]@records[[4]],
  xsyg[[1]]@records[[4]]
  ))
xsyg_v2 <- xsyg_v1
xsyg_v2@records[c(3, 6)] <- NULL

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(
    object = remove_SignalBackground(object = "error"),
    regexp = "'object' should be of class 'RLum.Analysis'")
  expect_error(remove_SignalBackground(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")

  ## test input object failure for object_bg
  expect_error(
    object = remove_SignalBackground(object = xsyg[[1]], object_bg = "error"),
    "'object_bg' should be of class 'RLum.Data.Curve', 'list', 'matrix', 'numeric', 'integer'")

  ## test input object failure for recordType
  expect_error(
    object = remove_SignalBackground(object = xsyg[[1]], recordType = 1),
    regexp = "'recordType' should be of class 'character'")

  ## test input object failure for clean_up
  expect_error(
    object = remove_SignalBackground(object = xsyg[[1]], clean_up = "error"),
    regexp = "'clean_up' should be a single logical value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## with clean_up
  t <- expect_s4_class(remove_SignalBackground(xsyg_v1), "RLum.Analysis")
  expect_equal(length(t@records), expected = 4)

  ## without clean_up
  t <- expect_s4_class(remove_SignalBackground(xsyg_v1, clean_up = FALSE), "RLum.Analysis")
  expect_equal(length(t@records), expected = 6)

  ## with invalid recordType set
  expect_warning(remove_SignalBackground(xsyg_v1, recordType = "error"),
                 regexp = "'recordType' setting invalid, nothing removed.")

  ## test the list treatment
  t <- expect_type(remove_SignalBackground(list(xsyg_v1, xsyg_v1)), "list")
  expect_equal(length(t), expected = 2)

  ## pass non wanted argument
  t <- expect_type(remove_SignalBackground(list(xsyg_v1, "error")), "list")
  expect_equal(length(t), expected = 1)

  ## test with different inputs for object_bg
  ## list
  expect_s4_class(remove_SignalBackground(
    object = xsyg_v1, recordType = "OSL (UVVIS)",
    object_bg = c(xsyg_v1[[2]], xsyg_v1[[2]])), "RLum.Analysis")
  ## matrix
  expect_s4_class(remove_SignalBackground(
    object = xsyg_v1, recordType = "OSL (UVVIS)",
    object_bg = matrix(runif(100,1,100), ncol = 2)), "RLum.Analysis")
  ## numeric
  expect_s4_class(remove_SignalBackground(
    object = xsyg_v1, recordType = "OSL (UVVIS)",
    object_bg = c(runif(100,1,100))), "RLum.Analysis")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(remove_SignalBackground(xsyg_v2))
  expect_snapshot_RLum(remove_SignalBackground(xsyg_v2,
                                               recordType = "OSL (UVVIS)",
                                               object_bg = xsyg_v1[[2]]))
})

#' Analyse SAR CW-OSL measurements.
#'
#' The function analyses SAR CW-OSL curve data and provides a summary of the
#' measured data for every position. The output of the function is optimised
#' for SAR OSL measurements on quartz.
#'
#' The function works only for standard SAR protocol measurements introduced by
#' Murray and Wintle (2000) with CW-OSL curves. For the calculation of the
#' Lx/Tx value the function [calc_OSLLxTxRatio] is used.
#'
#' **Provided rejection criteria**
#'
#' `[recyling ratio]`: calculated for every repeated regeneration dose point.
#'
#' `[recuperation]`: recuperation rate calculated by comparing the `Lx/Tx` values of the zero
#' regeneration point with the `Ln/Tn` value (the `Lx/Tx` ratio of the natural
#' signal). For methodological background see Aitken and Smith (1988)
#'
#' `[IRSL/BOSL]`: the integrated counts (`signal.integral`) of an
#' IRSL curve are compared to the integrated counts of the first regenerated
#' dose point. It is assumed that IRSL curves got the same dose as the first
#' regenerated dose point. **Note:** This is not the IR depletion ratio
#' described by Duller (2003).
#'
#' @param input.data [Risoe.BINfileData-class] (**required**):
#' input data from a Risø BIN file, produced by the function [read_BIN2R].
#'
#' @param signal.integral [vector] (**required**):
#' channels used for the signal integral, e.g. `signal.integral=c(1:2)`
#'
#' @param background.integral [vector] (**required**):
#' channels used for the background integral, e.g. `background.integral=c(85:100)`
#'
#' @param position [vector] (*optional*):
#' reader positions that want to be analysed (e.g. `position=c(1:48)`.
#' Empty positions are automatically omitted. If no value is given all
#' positions are analysed by default.
#'
#' @param run [vector] (*optional*):
#' range of runs used for the analysis. If no value is given the range of the
#' runs in the sequence is deduced from the `Risoe.BINfileData` object.
#'
#' @param set [vector] (*optional*):
#' range of sets used for the analysis. If no value is given the range of the
#' sets in the sequence is deduced from the `Risoe.BINfileData` object.
#'
#' @param dtype [character] (*optional*):
#' allows to further limit the curves by their data type (`DTYPE`),
#' e.g., `dtype = c("Natural", "Dose")` limits the curves to this two data types.
#' By default all values are allowed.
#' See [Risoe.BINfileData-class] for allowed data types.
#'
#' @param keep.SEL [logical] (default):
#' option allowing to use the `SEL` element of the [Risoe.BINfileData-class] manually.
#' **NOTE:** In this case any limitation provided by `run`, `set` and `dtype`
#' are ignored!
#'
#' @param info.measurement [character] (*with default*):
#' option to provide information about the measurement on the plot
#' output (e.g. name of the BIN or BINX file).
#'
#' @param output.plot [logical] (*with default*):
#' plot output (`TRUE/FALSE`)
#'
#' @param output.plot.single [logical] (*with default*):
#' single plot output (`TRUE/FALSE`) to allow for plotting the results in
#' single plot windows. Requires `output.plot = TRUE`.
#'
#' @param cex.global [numeric] (*with default*):
#' global scaling factor.
#'
#' @param ... further arguments that will be passed to the function
#' [calc_OSLLxTxRatio] (supported: `background.count.distribution`, `sigmab`,
#' `sig0`; e.g., for instrumental error) and can be used to adjust the plot.
#' Supported" `mtext`, `log`
#'
#' @return
#' A plot (*optional*) and [list] is returned containing the
#' following elements:
#'
#' \item{LnLxTnTx}{[data.frame] of all calculated Lx/Tx values including signal, background counts and the dose points.}
#' \item{RejectionCriteria}{[data.frame] with values that might by used as rejection criteria. NA is produced if no R0 dose point exists.}
#' \item{SARParameters}{[data.frame] of additional measurement parameters obtained from the BIN file, e.g. preheat or read temperature
#' (not valid for all types of measurements).}
#'
#'
#' @note
#' Rejection criteria are calculated but not considered during the
#' analysis to discard values.
#'
#' **The analysis of IRSL data is not directly supported**. You may want to
#' consider using the functions [analyse_SAR.CWOSL] or
#' [analyse_pIRIRSequence] instead.
#'
#' **The development of this function will not be continued. We recommend to use the function [analyse_SAR.CWOSL] or instead.**
#'
#'
#' @section Function version: 0.2.17
#'
#'
#' @author
#' Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)\cr
#' Margret C. Fuchs, HZDR, Freiberg (Germany)
#'
#' @seealso [calc_OSLLxTxRatio], [Risoe.BINfileData-class], [read_BIN2R], [plot_GrowthCurve]
#'
#' @references
#' Aitken, M.J. and Smith, B.W., 1988. Optical dating: recuperation
#' after bleaching. Quaternary Science Reviews 7, 387-393.
#'
#' Duller, G., 2003. Distinguishing quartz and feldspar in single grain
#' luminescence measurements. Radiation Measurements, 37 (2), 161-165.
#'
#' Murray, A.S. and Wintle, A.G., 2000. Luminescence dating of quartz using an
#' improved single-aliquot regenerative-dose protocol. Radiation Measurements
#' 32, 57-73.
#'
#' @keywords datagen dplot
#'
#' @examples
#' ##load data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##analyse data
#' output <- Analyse_SAR.OSLdata(input.data = CWOSL.SAR.Data,
#'                               signal.integral = c(1:5),
#'                               background.integral = c(900:1000),
#'                               position = c(1:1),
#'                               output.plot = TRUE)
#'
#' ##combine results relevant for further analysis
#' output.SAR <- data.frame(Dose = output$LnLxTnTx[[1]]$Dose,
#'                          LxTx = output$LnLxTnTx[[1]]$LxTx,
#'                          LxTx.Error = output$LnLxTnTx[[1]]$LxTx.Error)
#' output.SAR
#'
#' @md
#' @export
Analyse_SAR.OSLdata <- function(
  input.data,
  signal.integral,
  background.integral,
  position,
  run,
  set,
  dtype,
  keep.SEL = FALSE,
  info.measurement = "unkown measurement",
  output.plot = FALSE,
  output.plot.single = FALSE,
  cex.global = 1,
  ...
){

  ##============================================================================##
  ##CONFIG
  ##============================================================================##

  ##set colors gallery to provide more colors

  col <- get("col", pos = .LuminescenceEnv)

  ##============================================================================##
  ##ERROR HANDLING
  ##============================================================================##

  if(missing(input.data)==TRUE){stop("[Analyse_SAR.OSLdata] No input data given!")
  }else{sample.data<-input.data}

  if(missing(signal.integral)==TRUE){stop("[Analyse_SAR.OSLdata] No signal integral is given!")}
  if(missing(background.integral)==TRUE){stop("[Analyse_SAR.OSLdata] No background integral is given!")}

  ##set values for run and set if they are not defined by the user
  if(missing(position)==TRUE){position<-min(sample.data@METADATA[,"POSITION"]):max(sample.data@METADATA[,"POSITION"])}

  if(missing(run)==TRUE){run<-min(sample.data@METADATA[,"RUN"]):max(sample.data@METADATA[,"RUN"])}

  if(missing(set)==TRUE){set<-min(sample.data@METADATA[,"SET"]):max(sample.data@METADATA[,"SET"])}

  if(missing(dtype)){dtype <- c("Natural",
                                "N+dose",
                                "Bleach",
                                "Bleach+dose",
                                "Natural (Bleach)",
                                "N+dose (Bleach)",
                                "Dose",
                                "Background")}


  # Deal with extra arguments ----------------------------------------------------

  ##deal with addition arguments
  extraArgs <- list(...)

  background.count.distribution <-
    if ("background.count.distribution" %in% names(extraArgs)) {
      extraArgs$background.count.distribution
    } else
    {
      "non-poisson"
    }

  sigmab <- if("sigmab" %in% names(extraArgs)) {extraArgs$sigmab} else
  {NULL}

  ##============================================================================##
  ##CALCULATIONS
  ##============================================================================##


  ##loop over all positions
  for (i in position){

    ##checking if position is valid
    if(length(which(sample.data@METADATA["POSITION"]==i))>0){

      ##check if OSL curves are part of the data set
      if(nrow(sample.data@METADATA[sample.data@METADATA[,"LTYPE"]=="OSL",]) == 0){

        stop("[Analyse_SAR.OSLdata()] No 'OSL' curves found!")

      }

      if(!keep.SEL){
        ##select all OSL data depending on the run and set
        sample.data@METADATA[,"SEL"]<-FALSE
        sample.data@METADATA[sample.data@METADATA[,"LTYPE"]=="OSL" &
                               sample.data@METADATA[,"RUN"]%in%run==TRUE &
                               sample.data@METADATA[,"SET"]%in%set==TRUE &
                               sample.data@METADATA[,"DTYPE"]%in%dtype==TRUE, "SEL"] <- TRUE
      }

      ##grep all OSL curve IDs
      OSL.curveID<-sample.data@METADATA[sample.data@METADATA["SEL"]==TRUE &
                                          sample.data@METADATA["POSITION"]==i,"ID"]

      ##estimate LnLx.curveID and TnTx.curveID from records
      LnLx.curveID<-OSL.curveID[seq(1,length(OSL.curveID),by=2)]
      TnTx.curveID<-OSL.curveID[seq(2,length(OSL.curveID),by=2)]


      ##Provide Values For Growth Curve Fitting

      ##(1) get dose information
      Dose<-sapply(1:length(LnLx.curveID),function(x){
        Dose<-sample.data@METADATA[sample.data@METADATA["ID"]==LnLx.curveID[x],"IRR_TIME"]
      })

      ##(2) set LxTx curves
      LnLxTnTx.curves<-(sapply(1:length(LnLx.curveID),function(x){

        ##produce data.frames for Lx/Tx calculations
        Lx.HIGH<-sample.data@METADATA[sample.data@METADATA[,"ID"]==LnLx.curveID[x],"HIGH"]
        Lx.NPOINTS<-sample.data@METADATA[sample.data@METADATA[,"ID"]==LnLx.curveID[x],"NPOINTS"]
        Tx.HIGH<-sample.data@METADATA[sample.data@METADATA[,"ID"]==TnTx.curveID[x],"HIGH"]
        Tx.NPOINTS<-sample.data@METADATA[sample.data@METADATA[,"ID"]==TnTx.curveID[x],"NPOINTS"]

        Lx.curve<-data.frame(x=seq(Lx.HIGH/Lx.NPOINTS,Lx.HIGH,by=Lx.HIGH/Lx.NPOINTS),
                             y=unlist(sample.data@DATA[LnLx.curveID[x]]))
        Tx.curve<-data.frame(x=seq(Tx.HIGH/Tx.NPOINTS,Tx.HIGH,by=Tx.HIGH/Tx.NPOINTS),
                             y=unlist(sample.data@DATA[TnTx.curveID[x]]))

        return(list(Lx.curve,Tx.curve))
      }))

      ##(3) calculate Lx/Tx ratio
      LnLxTnTx <- get_RLum(
        merge_RLum(lapply(1:length(LnLxTnTx.curves[1, ]), function(k) {
          calc_OSLLxTxRatio(
            Lx.data = as.data.frame(LnLxTnTx.curves[1, k]),
            Tx.data = as.data.frame(LnLxTnTx.curves[2, k]),
            signal.integral = signal.integral,
            background.integral = background.integral,
            background.count.distribution = background.count.distribution,
            sigmab = sigmab
          )
        })))


      ##finally combine to data.frame including the record ID for further analysis
      LnLxTnTx <- cbind(LnLxTnTx,LnLx.curveID,TnTx.curveID)

      ##(4.1) set info concerning the kind of regeneration points

      ##generate unique dose id - this are also the # for the generated points
      temp.DoseID<-c(0:(length(Dose)-1))
      temp.DoseName<-paste("R",temp.DoseID,sep="")
      temp.DoseName<-cbind(Name=temp.DoseName,Dose)

      ##set natural
      temp.DoseName[temp.DoseName[,"Name"]=="R0","Name"]<-"Natural"

      ##set R0
      temp.DoseName[temp.DoseName[,"Name"]!="Natural" & temp.DoseName[,"Dose"]==0,"Name"]<-"R0"

      ##find duplicated doses (including 0 dose - which means the Natural)
      temp.DoseDuplicated<-duplicated(temp.DoseName[,"Dose"])

      ##combine temp.DoseName
      temp.DoseName<-cbind(temp.DoseName,Repeated=temp.DoseDuplicated)

      ##correct value for R0 (it is not really repeated)
      temp.DoseName[temp.DoseName[,"Dose"]==0,"Repeated"]<-FALSE


      ##(5) Combine all values in a data.frame
      temp.LnLxTnTx<-data.frame(Name=temp.DoseName[,"Name"],
                                Dose=Dose,
                                Repeated=as.logical(temp.DoseName[,"Repeated"]))
      LnLxTnTx<-cbind(temp.LnLxTnTx,LnLxTnTx)
      LnLxTnTx[,"Name"]<-as.character(LnLxTnTx[,"Name"])

      ##(6) Calculate Recyling Ratio and Recuperation Rate

      ##(6.1)
      ##Calculate Recycling Ratio

      if(length(LnLxTnTx[LnLxTnTx[,"Repeated"]==TRUE,"Repeated"])>0){

        ##identify repeated doses
        temp.Repeated<-LnLxTnTx[LnLxTnTx[,"Repeated"]==TRUE,c("Name","Dose","LxTx")]

        ##find corresponding previous dose for the repeated dose
        temp.Previous<-t(sapply(1:length(temp.Repeated[,1]),function(x){
          LnLxTnTx[LnLxTnTx[,"Dose"]==temp.Repeated[x,"Dose"] &
                     LnLxTnTx[,"Repeated"]==FALSE,c("Name","Dose","LxTx")]
        }))

        ##convert to data.frame
        temp.Previous<-as.data.frame(temp.Previous)

        ##set column names
        temp.ColNames<-sapply(1:length(temp.Repeated[,1]),function(x){
          paste(temp.Repeated[x,"Name"],"/",
                temp.Previous[temp.Previous[,"Dose"]==temp.Repeated[x,"Dose"],"Name"]
                ,sep="")
        })

        ##Calculate Recycling Ratio
        RecyclingRatio<-as.numeric(temp.Repeated[,"LxTx"])/as.numeric(temp.Previous[,"LxTx"])

        ##Just transform the matrix and add column names
        RecyclingRatio<-t(RecyclingRatio)
        colnames(RecyclingRatio) <- unique(temp.ColNames)

      }else{RecyclingRatio<-NA}

      ##(6.2)
      ##Recuperation Rate

      if("R0" %in% LnLxTnTx[,"Name"]==TRUE){
        Recuperation<-round(LnLxTnTx[LnLxTnTx[,"Name"]=="R0","LxTx"]/LnLxTnTx[LnLxTnTx[,"Name"]=="Natural","LxTx"],digits=4)
      }else{Recuperation<-NA}


      ##(6.3) IRSL
      ##Print IRSL Curves if IRSL curve is set
      sample.data@METADATA[,"SEL"]<-FALSE
      sample.data@METADATA[sample.data@METADATA["LTYPE"]=="IRSL" &
                             sample.data@METADATA[,"RUN"]%in%run==TRUE &
                             sample.data@METADATA[,"SET"]%in%set==TRUE,"SEL"]<-TRUE


      ##get IRSL curve ID & ID for Reg1 again
      IRSL.curveID<-sample.data@METADATA[sample.data@METADATA["SEL"]==TRUE & sample.data@METADATA["POSITION"]==i,"ID"]

      ##if no IRSL curve the length of the object is 0
      if(length(IRSL.curveID)>0){

        ##chose an IRSL curve with a dose of the first regeneration point
        Reg1again.curveID<-LnLxTnTx[LnLxTnTx[,"Repeated"]==TRUE & LnLxTnTx[,"Dose"]==LnLxTnTx[2,"Dose"],"LnLx.curveID"]

        if(length(Reg1again.curveID)>0){

          ##BOSL/IRSL
          IRSL_BOSL<-round(sum(unlist(sample.data@DATA[IRSL.curveID])[signal.integral])
                           /sum(unlist(sample.data@DATA[Reg1again.curveID])[signal.integral]),digits=4)
        }else{IRSL_BOSL<-NA}
      }else{IRSL_BOSL<-NA}

      ##Combine the two values
      if(exists("RejectionCriteria")==FALSE){
        RejectionCriteria<-cbind(RecyclingRatio,Recuperation,IRSL_BOSL)
      }else{
        RejectionCriteria.temp<-cbind(RecyclingRatio,Recuperation,IRSL_BOSL)
        RejectionCriteria<-rbind(RejectionCriteria,RejectionCriteria.temp)
      }

      ##============================================================================##
      ##PLOTTING
      ##============================================================================##

      if(output.plot){

        ##set plot settings
        plot.settings <- list(
          mtext = sample.data@METADATA[sample.data@METADATA[,"ID"]==LnLx.curveID[1],"SAMPLE"],
          log = ""

        )

        ##modify arguments
        plot.settings <- modifyList(plot.settings, list(...))



        if(output.plot.single==FALSE){
          layout(matrix(c(1,2,1,2,3,4,3,5),4,2,byrow=TRUE))
        }
        ##warning if number of curves exceed colour values
        if(length(col)<length(LnLx.curveID)){
          cat("\n[Analyse_SAR.OSLdata()] Warning: To many curves! Only the first",
              length(col),"curves are plotted!")
        }

        ##==========================================================================
        ##plot Ln,Lx Curves


        ##get maximum value of LnLx curves
        LnLx.curveMax<-max(unlist(sample.data@DATA[LnLx.curveID]))

        ##get channel resolution (it should be the same for all values)
        HIGH<-sample.data@METADATA[sample.data@METADATA[,"ID"]==LnLx.curveID[1],"HIGH"]
        NPOINTS<-sample.data@METADATA[sample.data@METADATA[,"ID"]==LnLx.curveID[1],"NPOINTS"]

        xaxt.values<-seq(HIGH/NPOINTS,HIGH,by=HIGH/NPOINTS)

        ##open plot area LnLx
        plot(NA,NA,
             xlab="Time [s]",
             ylab=paste("OSL [cts/",HIGH/NPOINTS," s]",sep=""),
             xlim=c(HIGH/NPOINTS,HIGH),
             ylim=c(1,max(unlist(sample.data@DATA[LnLx.curveID]))),
             main=expression(paste(L[n],",",L[x]," curves",sep="")),
             log=plot.settings$log
        )
        ##plot curves and get legend values
        sapply(1:length(LnLx.curveID),function(x){
          yaxt.values<-unlist(sample.data@DATA[LnLx.curveID[x]])
          lines(xaxt.values,yaxt.values,col=col[x])
        })

        ##mark integration limits
        abline(v=xaxt.values[min(signal.integral)], lty=2, col="gray")
        abline(v=xaxt.values[max(signal.integral)], lty=2, col="gray")
        abline(v=xaxt.values[min(background.integral)], lty=2, col="gray")
        abline(v=xaxt.values[max(background.integral)], lty=2, col="gray")


        ##plot legend
        legend("topright",as.character(LnLxTnTx$Name),lty=c(rep(1,length(LnLx.curveID))),
               cex=0.8*cex.global,col=col, bg="gray")

        ##sample name
        mtext(side=3,plot.settings$mtext,cex=0.7*cex.global)

        ##========================================================================
        ##open plot area TnTx
        plot(NA,NA,
             xlab="Time [s]",
             ylab=paste("OSL [cts/",HIGH/NPOINTS," s]",sep=""),
             xlim=c(HIGH/NPOINTS,HIGH),
             ylim=c(1,max(unlist(sample.data@DATA[TnTx.curveID]))),
             main=expression(paste(T[n],",",T[x]," curves",sep="")),
             log=plot.settings$log
        )
        ##plot curves and get legend values
        sapply(1:length(TnTx.curveID),function(x){
          yaxt.values<-unlist(sample.data@DATA[TnTx.curveID[x]])
          lines(xaxt.values,yaxt.values,col=col[x])
        })

        ##mark integration limits
        abline(v=xaxt.values[min(signal.integral)], lty=2, col="gray")
        abline(v=xaxt.values[max(signal.integral)], lty=2, col="gray")
        abline(v=xaxt.values[min(background.integral)], lty=2, col="gray")
        abline(v=xaxt.values[max(background.integral)], lty=2, col="gray")


        ##plot legend
        legend("topright",as.character(LnLxTnTx$Name),lty=c(rep(1,length(TnTx.curveID))),
               cex=0.8*cex.global,col=col, bg="gray")

        ##sample name
        mtext(side=3,plot.settings$mtext,cex=0.7*cex.global)

        ##========================================================================
        ##Print TL curves for TnTx -
        sample.data@METADATA[,"SEL"]<-FALSE
        sample.data@METADATA[sample.data@METADATA["LTYPE"]=="TL","SEL"]<-TRUE

        ##check if TL any curves is measured within the sequence
        if(length(sample.data@METADATA[sample.data@METADATA[,"SEL"]==TRUE,1])>0){


          ##to ensure that the right TL curves are used the run and set number of the LnLx and TnTx curves are used
          LnLx.SET<-sapply(LnLx.curveID,function(x){sample.data@METADATA[sample.data@METADATA["ID"]==x,"SET"]})
          LnLx.RUN<-sapply(LnLx.curveID,function(x){sample.data@METADATA[sample.data@METADATA["ID"]==x,"RUN"]})
          TnTx.SET<-sapply(TnTx.curveID,function(x){sample.data@METADATA[sample.data@METADATA["ID"]==x,"SET"]})
          TnTx.RUN<-sapply(TnTx.curveID,function(x){sample.data@METADATA[sample.data@METADATA["ID"]==x,"RUN"]})

          ##get TL curve IDs in general considering the constraints
          TL.curveID<-sapply(1:length(TnTx.curveID),function(x){results<-
                                                                  sample.data@METADATA[sample.data@METADATA["SEL"]==TRUE & sample.data@METADATA["POSITION"]==i &
                                                                                         sample.data@METADATA["SET"]>=LnLx.SET[x] & sample.data@METADATA["RUN"]>=LnLx.RUN[x] &
                                                                                         sample.data@METADATA["SET"]<=TnTx.SET[x] & sample.data@METADATA["RUN"]<=TnTx.RUN[x],"ID"]})

          ##get maximum value of  TL curves
          TL.curveMax<-max(unlist(sample.data@DATA[TL.curveID]))

          ##get channel resolution (it should be the same for all values)
          HIGH<-unique(sample.data@METADATA[sample.data@METADATA["ID"]==TL.curveID[1],"HIGH"])
          NPOINTS<-unique(sample.data@METADATA[sample.data@METADATA["ID"]==TL.curveID[1],"NPOINTS"])
          xaxt.values<-seq(HIGH/NPOINTS,HIGH,by=HIGH/NPOINTS)

          ##get heating rate
          RATE<-unique(sample.data@METADATA[sample.data@METADATA["ID"]==TL.curveID[1],"RATE"])

          ##open plot area for TL curves
          plot(NA,NA,
               xlab="T [\u00B0C]",
               ylab=paste("TL [cts/",HIGH/NPOINTS," \u00B0C]",sep=""),
               xlim=c(HIGH/NPOINTS,HIGH),
               ylim=c(1,TL.curveMax),
               main="Cutheat - TL curves",
               sub=paste("(",RATE," K/s)",sep=""),
               log=if(plot.settings$log=="y" | plot.settings$log=="xy"){"y"}else{""}
          )

          ##plot curves and get legend values
          sapply(1:length(TL.curveID),function(x){
            yaxt.values<-unlist(sample.data@DATA[TL.curveID[x]])
            lines(xaxt.values,yaxt.values,col=col[x])
          })

          ##plot legend
          legend("topleft",as.character(LnLxTnTx$Name),lty=c(rep(1,length(TL.curveID))),
                 cex=0.8*cex.global,col=col, bg="white", bty="n")

          ##sample name
          mtext(side=3,plot.settings$mtext,cex=0.7*cex.global)


        }else{
          plot(NA,NA,xlim=c(0,100),ylim=c(0,100), main="Cutheat - TL curves")
          text(50,50,"no cutheat as TL curve detected")
        }

        ##======================================================================##
        ##Print IRSL Curves if IRSL curve is set

        if(is.na(IRSL_BOSL) == FALSE){
          ##get channel resolution (it should be the same for all values)
          HIGH<-unique(sample.data@METADATA[sample.data@METADATA["ID"]==IRSL.curveID ,"HIGH"])
          NPOINTS<-unique(sample.data@METADATA[sample.data@METADATA["ID"]==IRSL.curveID ,"NPOINTS"])
          xaxt.values<-seq(HIGH/NPOINTS,HIGH,by=HIGH/NPOINTS)

          ##open plot IRSL curve
          plot(NA,NA,
               xlab="Time [s]",
               ylab=paste("OSL and IRSL [cts/",HIGH/NPOINTS," s]",sep=""),
               xlim=c(0,HIGH),
               ylim=c(0,max(unlist(sample.data@DATA[Reg1again.curveID]))),
               main="IRSLT"
          )

          ##show integral limits
          abline(v=xaxt.values[min(signal.integral)], lty=2, col="gray")
          abline(v=xaxt.values[max(signal.integral)], lty=2, col="gray")

          ##print(length(sample.data@DATA[IRSL.curveID]))
          lines(xaxt.values,unlist(sample.data@DATA[IRSL.curveID]),col="red")
          lines(xaxt.values,unlist(sample.data@DATA[Reg1again.curveID[1]]),col="blue")

          ##legend
          legend("topright",c("R1 again","IRSL"),lty=c(1,1),col=c("blue","red"), bty="n")


          mtext(side=3,paste("IRSL/BOSL = ",IRSL_BOSL*100,"%",sep=""),
                cex=.8*cex.global
          )
        }

        if(((is.na(IRSL_BOSL)==TRUE) & length(IRSL.curveID)>0) |
             ((is.na(IRSL_BOSL)==FALSE) & length(IRSL.curveID)>0)){

          ##plot only IRSL curve
          plot(xaxt.values,unlist(sample.data@DATA[IRSL.curveID]),
               xlab="Time [s]",
               ylab=paste("IRSL [cts/",HIGH/NPOINTS," s]",sep=""),
               xlim=c(0,10),
               ylim=c(0,max(unlist(sample.data@DATA[IRSL.curveID]))),
               main="IRSL curve (10 s)",
               col="red",
               type="l"
          )
        }else{
          plot(NA,NA,xlim=c(0,10), ylim=c(0,10), main="IRSL curve")
          text(5,5,"no IRSL curve detected")
        }
        ##=========================================================================
        ##Plot header
        if(output.plot.single==TRUE){
          mtext(side=3,paste("ALQ Pos. ",i,sep=""))
        }else{
          mtext(side=3,paste("ALQ Pos. ",i,sep=""),outer=TRUE,line=-2.5)
        }

        ##Plot footer
        mtext(side=4,info.measurement,outer=TRUE,line=-1.5,cex=0.6*cex.global, col="blue")

        ##output on terminal for plot
        writeLines(paste("\n[Analyse_SAR.OSLdata()] >> Figure for position ",i," produced.",sep=""))

        ##reset mfrow
        par(mfrow=c(1,1))



      }#endif for output.plot
      ##preprate output of values
      ##==============================================================================

      ##Add LnLxTnTx values to the list
      if(exists("LnLxTnTx_List")==FALSE){LnLxTnTx_List<-list()}
      LnLxTnTx_List[[i]]<-LnLxTnTx
      rm(LnLxTnTx)


    }else{writeLines(paste("[Analyse_SAR.OSLdata()] >> Position ",i," is not valid and has been omitted!",sep=""))} #end if position checking

  }#end for loop

  ##============================================================================##
  ##OUTPUT OF FUNCTION
  ##============================================================================##

  ##get further information from the position used

  ##this is what you get from the Risoe file
  readTemp<-unique(sample.data@METADATA[sample.data@METADATA[,"POSITION"]==min(position) & sample.data@METADATA[,"LTYPE"]!="TL","TEMPERATURE"])

  cutheat<-unique(sample.data@METADATA[sample.data@METADATA[,"POSITION"]==min(position)  &
                                         sample.data@METADATA[,"LTYPE"]=="TL","HIGH"])
  if(length(cutheat)==0){cutheat=NA}

  systemID<-unique(sample.data@METADATA[sample.data@METADATA[,"POSITION"]==min(position),"SYSTEMID"])

  SARParameters<-data.frame(readTemp=readTemp,cutheat=cutheat,systemID=systemID)

  return(list(LnLxTnTx=LnLxTnTx_List,
              RejectionCriteria=RejectionCriteria,
              SARParameters=SARParameters))


}

test_that("full example test", {
  testthat::skip_on_cran()
  local_edition(3)

  data(ExampleData.BINfileData, envir = environment())
  output <- Analyse_SAR.OSLdata(input.data = CWOSL.SAR.Data,
                                signal.integral = c(1:5),
                                background.integral = c(900:1000),
                                position = c(1:1),
                                output.plot = FALSE)


  ##checks
  expect_type(output, "list")
  expect_length(output, 3)

  ## errors
  expect_error({ Analyse_SAR.OSLdata() },
               regexp = "No input data given")
  expect_error({ Analyse_SAR.OSLdata(input.data = CWOSL.SAR.Data) },
               regexp = "No signal integral is given")
  expect_error({ Analyse_SAR.OSLdata(input.data = CWOSL.SAR.Data, signal.integral = 1:3) },
               regexp = "No background integral is given")
  expect_error({ Analyse_SAR.OSLdata(input.data = subset(CWOSL.SAR.Data, LTYPE == "IRSL"),
                                     signal.integral = 1:3,
                                     background.integral = 200:250) },
               regexp = "No 'OSL' curves found")

  ## should work
  expect_type(Analyse_SAR.OSLdata(input.data = CWOSL.SAR.Data, signal.integral = 1:3,
                                  background.integral = 200:250, position = 1,
                                  background.count.distribution = "non-poisson",
                                  sigmab = 0.1, output.plot = TRUE), "list")
  tmp <- subset(CWOSL.SAR.Data, LTYPE == "OSL" & POSITION == 1 & ID <= 457)
  expect_type(
    Analyse_SAR.OSLdata(tmp, 1:3, 200:250, output.plot = TRUE, output.plot.single = TRUE),
    "list")

})

#'@title Import Luminescence Data into R
#'
#'@description
#'Convenience wrapper function to provide a quicker and more standardised way of
#'reading data into R by looping through all in the package available data import
#'functions starting with `read_`. Import data types can be mixed.
#'
#'@param file [character] (**required**): file to be imported, can be a [list] or a [character] vector
#'
#'@param ... arguments to be further passed down to supported functions (please check the functions
#'to determine the correct arguments)
#'
#'@param fastForward [logical] (*with default*): option to create [Luminescence::RLum-class] objects
#'during import or a [list] of such objects
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @section Function version: 0.1.6
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@returns Always returns a [list]; empty or filled with [Luminescence::RLum.Analysis-class] objects
#'
#' @seealso `r toString(paste0("[Luminescence::", sort(grep("^read_", getNamespaceExports("Luminescence"), value = TRUE)), "]"))`
#'
#'@keywords datagen
#'
#'@examples
#'
#' ## import BINX/BIN
#' file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
#' temp <- import_Data(file)
#'
#' ## RF data
#' file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
#' temp <- import_Data(file)
#'
#'@export
import_Data <- function(
  file,
  ...,
  fastForward = TRUE,
  verbose = FALSE
) {
  .set_function_name("import_Data")
  on.exit(.unset_function_name(), add = TRUE)

  .validate_class(file, c("character", "list"))

  ## supported functions are extracted automatically from the package
  ## namespace so that we don't have to maintain this list manually
  fun <- grep("^read_", getNamespaceExports("Luminescence"), value = TRUE)

  ## put the most common formats first, the rest alphabetically
  fun <- unique(c("read_BIN2R", "read_XSYG2R", sort(fun)))

  ## make the file handling just a little bit easier
  if(length(file) > 1)
    file <- as.list(file)

  ## get arguments of functions
  args <- c(list(file = file, fastForward = fastForward, verbose = verbose), list(...))

  ## set empty output list
  out <- list()

  ## just try all functions and all files
  for (i in fun) {
    SW({
      ## get arguments and remove non-supported arguments
      t <- try(do.call(what = i, args = args), silent = TRUE)
    })

      if (!is.null(t) && !inherits(t, "try-error"))
        out <- c(out, t)
  }

  ## return if output is not empty
  if (length(out) == 0)
   .throw_message("Unknown file format, nothing imported")

  return(out)
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(import_Data(data.frame()),
               "'file' should be of class 'character' or 'list'")
  expect_message(import_Data(system.file("extdata/QNL84_2_bleached.txt")),
                 "Unknown file format, nothing imported")
})

test_that("Test general import", {
  testthat::skip_on_cran()

  ## BINX
  SW({
  expect_type(
    object = import_Data(system.file("extdata/BINfile_V8.binx", package = "Luminescence")),
    type = "list")
  })

  ## XSYG
  expect_type(
    object = import_Data(system.file("extdata/XSYG_file.xsyg", package = "Luminescence")),
    type = "list")

  ## PSL
  expect_s4_class(
    object = import_Data(system.file("extdata/DorNie_0016.psl", package = "Luminescence"))[[1]],
    class = "RLum.Analysis")

  ## DAT (Daybreak)
  expect_type(
    object = import_Data(system.file("extdata/Daybreak_TestFile.DAT", package = "Luminescence")),
    type = "list")

  ## TXT (Daybreak)
  expect_type(
    object = import_Data(system.file("extdata/Daybreak_TestFile.txt", package = "Luminescence")),
    type = "list")

  ## RF
  expect_type(
    object = import_Data(system.file("extdata/RF_file.rf", package = "Luminescence")),
    type = "list")

  ## TIFF
  expect_s4_class(
    object = import_Data(system.file("extdata/TIFFfile.tif", package = "Luminescence"))[[1]],
    class = "RLum.Data.Image")

  ## OSL
  expect_s4_class(
    object = import_Data(system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence"))[[1]],
    class = "RLum.Analysis")

  ## use character and list as input
  files <- c(
    system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence"),
    system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence")
  )
  expect_type(
    object = import_Data(files, verbose = FALSE),
    type = "list")
  expect_type(
    object = import_Data(as.list(files), verbose = FALSE),
    type = "list")

  ## mix input
  files <- c(
    system.file("extdata/BINfile_V8.binx", package = "Luminescence"),
    system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence"),
    system.file("extdata/RF_file.rf", package = "Luminescence"),
    system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence"),
    system.file("extdata/TIFFfile.tif", package = "Luminescence"),
    system.file("extdata/XSYG_file.xsyg", package = "Luminescence"))
  SW({
  t <- expect_type(
    object = import_Data(files, verbose = FALSE),
    type = "list")
  })
  expect_equal(length(t), 16)

})

#' @title Create a Dose-Response Curve Summary Plot
#'
#' @description
#' While analysing OSL SAR or pIRIR-data the view on the data is usually
#' limited to one dose-response curve (DRC) at the time for one aliquot. This
#' function overcomes this limitation by plotting all DRCs from an
#' [Luminescence::RLum.Results-class] object created by [Luminescence::analyse_SAR.CWOSL] in one single
#' plot.
#'
#' If you want plot your DRC on an energy scale (dose in Gy), you can either
#' use option `source_dose_rate` or perform your SAR analysis with the dose
#' points in Gy (better axis scaling).
#'
#' @param object [Luminescence::RLum.Results-class] (**required**): input object created by
#' [Luminescence::analyse_SAR.CWOSL]. The input object can be provided as [list].
#'
#' @param source_dose_rate [numeric] (*optional*): allows to modify the axis
#' and show values in Gy, instead seconds. Only a single numerical value is
#' allowed.
#'
#' @param sel_curves [numeric] (*optional*): id of the curves to be plotted in
#' its occurring order. A sequence can be provided for selecting, e.g., only
#' every 2nd curve from the input object.
#'
#' @param show_dose_points [logical] (*with default*): enable/disable plotting
#' of dose points in the graph.
#'
#' @param show_natural [logical] (*with default*): enable/disable the plot
#' of the natural `Lx/Tx` values.
#'
#' @param n [integer] (*with default*): number of x-values used to evaluate
#' one curve object. Large numbers slow down the plotting process and are
#' usually not needed.
#'
#'@param ... Further arguments and graphical parameters to be passed.
#'In particular: `main`, `xlab`, `ylab`, `xlim`, `ylim`, `lty`, `lwd`, `pch`,
#'`col.pch`, `col.lty`, `mtext`
#'
#'@section Function version: 0.2.4
#'
#'@return An [Luminescence::RLum.Results-class] object is returned:
#'
#' Slot: **@data**\cr
#'
#' \tabular{lll}{
#' **OBJECT** \tab **TYPE** \tab **COMMENT**\cr
#' `results` \tab [data.frame] \tab with dose and LxTx values \cr
#' `data` \tab [Luminescence::RLum.Results-class] \tab original input data \cr
#' }
#'
#' Slot: **@info**\cr
#'
#' \tabular{lll}{
#' **OBJECT** \tab **TYPE** \tab **COMMENT** \cr
#' `call` \tab `call` \tab the original function call \cr
#' `args` \tab `list` \tab arguments of the original function call \cr
#' }
#'
#'*Note: If the input object is a [list] a list of [Luminescence::RLum.Results-class] objects is returned.*
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Christoph Burow, University of Cologne (Germany)
#'
#'@seealso [Luminescence::RLum.Results-class], [Luminescence::analyse_SAR.CWOSL]
#'
#'@examples
#'
#'#load data example data
#'data(ExampleData.BINfileData, envir = environment())
#'
#'#transform the values from the first position in a RLum.Analysis object
#'object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#'
#' ## perform SAR analysis
#' results <- analyse_SAR.CWOSL(
#'     object,
#'     signal_integral = 1:2,
#'     background_integral = 900:1000,
#'     plot = FALSE
#' )
#'
#'##plot only DRC
#'plot_DRCSummary(results)
#'
#'@md
#'@export
plot_DRCSummary <- function(
  object,
  source_dose_rate = NULL,
  sel_curves = NULL,
  show_dose_points = FALSE,
  show_natural = FALSE,
  n = 51L,
  ...
) {
  .set_function_name("plot_DRCSummary")
  on.exit(.unset_function_name(), add = TRUE)

# Self-call -----------------------------------------------------------------------------------
if(inherits(object, "list")){

  ##catch ... arguments
  plot_settings <- list(...)

  ## expand input arguments
  if ("main" %in% ...names()) {
    main <- .listify(list(...)[["main"]], length(object))

    ##filter main from the ... argument list otherwise we will have a collusion
    plot_settings["main" %in% names(plot_settings)] <- NULL

  }else{
    main <- .listify("DRC", length(object))
  }

  results <- lapply(seq_along(object), function(o) {
    plot_DRCSummary(
      object = object[[o]],
      source_dose_rate = source_dose_rate,
      sel_curves = sel_curves,
      show_dose_points = show_dose_points,
      show_natural = show_natural,
      n = n,
      main = main[[o]],
      ... = plot_settings
      )
  })

  ##return merged object
  return(results)
}

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Results")
  .validate_originator(object, c("analyse_SAR.CWOSL", "analyse_pIRIRSequence"))

# Extract data from object --------------------------------------------------------------------

  ## set limit
  if (is.null(sel_curves)) {
    sel_curves <- 1:length(object@data$Formula)
  } else if (min(sel_curves) < 1 ||
         max(sel_curves) > length(object@data$Formula) ||
         length(sel_curves) > length(object@data$Formula)){
        .throw_warning("'sel_curves' out of bounds, reset to full dataset")
        sel_curves <- 1:length(object@data$Formula)
  }

  ## check whether the same fit method was used
  fit.method <- unique(object@data[["data"]][["Fit"]])
  if (length(fit.method) > 1) {
    .throw_error("Dose-response curves fitted using different equations: ",
                 .collapse(fit.method))
  }

    ##get DRC
    DRC <- object@data$Formula[sel_curves]

  ## get limits for each set
  idx.natural <- which(object@data$LnLxTnTx.table[["Name"]] == "Natural")
  dataset_limits <- cbind(idx.natural,
                          idx.natural[-1] - 1,
                          nrow(object@data$LnLxTnTx.table))

   ##create list
   LxTx <- lapply(1:nrow(dataset_limits), function(x){
     object@data$LnLxTnTx.table[dataset_limits[x,1]:dataset_limits[x,2],]
   })[sel_curves]

# Plotting ------------------------------------------------------------------------------------
  ##set default
  plot_settings <-  modifyList(x = list(
    xlab = if(is.null(source_dose_rate)) {"Dose [s]"} else {"Dose [Gy]"},
    ylab = expression(L[x]/T[x]),
    xlim = c(0,max(vapply(LxTx, function(x){max(x[["Dose"]])}, numeric(1)))),
    ylim = if(show_dose_points){
      c(0,max(vapply(LxTx, function(x){max(x[["LxTx"]] + x[["LxTx.Error"]])}, numeric(1)), na.rm = TRUE))
    }else{
      c(0,max(vapply(1:length(LxTx), function(y){
        x <- max(LxTx[[y]][["Dose"]], na.rm = TRUE)
        eval(DRC[[y]])
       },numeric(1)), na.rm = TRUE))
    },
    main = "DRC Summary",
    mtext = paste0("n_curves: ",length(sel_curves)),
    lty = 1,
    lwd = 1,
    pch = 20,
    col.lty = rgb(0,0,0,0.5),
    col.pch = rgb(0,0,0,0.5)
  ), val = list(...), keep.null = TRUE)

  ## expand parameters
  plot_settings$col.lty <- rep(plot_settings$col.lty, length(sel_curves))
  plot_settings$col.pch <- rep(plot_settings$col.pch, length(sel_curves))
  plot_settings$pch <- rep(plot_settings$pch, length(sel_curves))
  plot_settings$lty <- rep(plot_settings$lty, length(sel_curves))

  ##create empty plot window
  plot(
    x = NA,
    y = NA,
    xlab = plot_settings$xlab,
    ylab = plot_settings$ylab,
    xlim = plot_settings$xlim,
    ylim = plot_settings$ylim,
    main = plot_settings$main,
    xaxt = "n"
  )

  if(!is.null(plot_settings$mtext))
    mtext(side = 3, text = plot_settings$mtext, cex = 0.8)

  #exchange x-axis if source dose rate is set
  if(!is.null(source_dose_rate)){
    axis(side = 1, at = axTicks(side = 1), labels = round(axTicks(side = 1) * source_dose_rate[1],0))
  }else{
    axis(side = 1)
  }

  for(i in 1:length(sel_curves)){
    ##plot natural
    if(show_natural){
      segments(x0 = LxTx[[i]]$Dose[1], x1 = LxTx[[i]]$Dose[1],
               y0 = LxTx[[i]]$LxTx[1] - LxTx[[i]]$LxTx.Error[1],
               y1 = LxTx[[i]]$LxTx[1] + LxTx[[i]]$LxTx.Error[1],
               col = plot_settings$col.pch[[i]])
      points(
        x = LxTx[[i]]$Dose[1],
        y = LxTx[[i]]$LxTx[1],
        col = plot_settings$col.pch[[i]],
        pch = plot_settings$pch[[i]]
      )
    }

    ##plot dose points
    if(show_dose_points){
      segments(x0 = LxTx[[i]]$Dose[-1], x1 = LxTx[[i]]$Dose[-1],
               y0 = LxTx[[i]]$LxTx[-1] - LxTx[[i]]$LxTx.Error[-1],
               y1 = LxTx[[i]]$LxTx[-1] + LxTx[[i]]$LxTx.Error[-1],
               col = plot_settings$col.pch[[i]])
      points(
        x = LxTx[[i]]$Dose[-1],
        y = LxTx[[i]]$LxTx[-1],
        col = plot_settings$col.pch[[i]],
        pch = plot_settings$pch[[i]]
      )
    }

    ##plot lines
    x <- seq(min(plot_settings$xlim),max(plot_settings$xlim), length.out = n)
    y <- eval(DRC[[i]])

    if (anyNA(y) || any(is.nan(y))) {
      .throw_warning("Dose response curve ", i, " contains NA/NaN values, ",
                     "curve removed before plotting")
      next
    }

    lines(
      x = x,
      y = eval(DRC[[i]]),
      col = plot_settings$col.lty[[i]],
      lwd = plot_settings$lwd,
      lty = plot_settings$lty[[i]]
    )
  }

  ## Results -------------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      results = data.frame(
        dose = x,
        sapply(DRC, function(d, n) { eval(d) }, n)
        ),
      data = object
    ),
    info = list(
      call = sys.call(),
      args = as.list(sys.call())[-1])
  )
}

#' @title  Plot function for an `RLum.Data.Image` S4 class object
#'
#' @description The function provides very basic plot functionality for image data of an
#' [Luminescence::RLum.Data.Image-class] object. For more sophisticated plotting it is recommended
#' to use other very powerful packages for image processing.
#'
#'
#' **Details on the plot functions**
#'
#' Supported plot types:
#'
#' **`plot.type = "plot.raster"`**
#'
#' Uses the standard plot function of R [graphics::image]. If wanted, the image
#' is enhanced, using the argument `stretch`. Possible values are `hist`, `lin`, and
#' `NULL`. The latter does nothing. The argument `useRaster = TRUE` is used by default, but
#' can be set to `FALSE`.
#'
#' **`plot.type = "contour"`**
#'
#' This uses the function [graphics::contour]
#'
#' @param object [Luminescence::RLum.Data.Image-class] (**required**): S4
#' object of class [Luminescence::RLum.Data.Image-class]
#'
#' @param par.local [logical] (*with default*): use local graphical
#' parameters for plotting, e.g. the plot is shown in one column and one row.
#' If `par.local = FALSE` global parameters are inherited.
#'
#' @param frames [numeric] (*optional*): sets the frames to be set, by default all
#' frames are plotted. Can be sequence of numbers, as long as the frame number is valid.
#'
#' @param plot.type [character] (*with default*): plot types.
#' Supported types are `plot.raster`, `contour`
#'
#' @param ... further arguments and graphical parameters that will be passed
#' to the specific plot functions. Standard supported parameters are `xlim`,
#' `ylim`, `zlim`, `xlab`, `ylab`, `main`, `mtext`, `legend` (`TRUE` or `FALSE`),
#' `col`, `cex`, `axes` (`TRUE` or `FALSE`), `zlim_image` (adjust the z-scale
#' over different images), `stretch`, `digits`, scientific (`TRUE` or `FALSE`).
#'
#' @return Returns a plot
#'
#' @note The axes limitations (`xlim`, `zlim`, `zlim`) work directly on the object,
#' so that regardless of the chosen limits the image parameters can be adjusted for
#' best visibility. However, in particular for z-scale limitations this is not always
#' wanted, please use `zlim_image` to maintain a particular value range over a
#' series of images.
#'
#' @section Function version: 0.2.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Image-class], [plot], [Luminescence::plot_RLum],
#' [graphics::image], [graphics::contour]
#'
#' @keywords aplot
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.RLum.Data.Image, envir = environment())
#'
#' ##plot data
#' plot_RLum.Data.Image(ExampleData.RLum.Data.Image)
#'
#' @export
plot_RLum.Data.Image <- function(
  object,
  frames = NULL,
  par.local = TRUE,
  plot.type = c("plot.raster", "contour"),
  ...
) {
  .set_function_name("plot_RLum.Data.Image")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Data.Image")
  plot.type <- .validate_args(plot.type, c("plot.raster", "contour"))

  ## extract object
  object <- object@data
  if (length(dim(object)) != 3) {
    ## the object is empty or malformed
    return(NULL)
  }

# Define additional functions ---------------------------------------------
.stretch <- function(x, type = "lin"){
  if(is.null(type[1])) return(x[,,1])

  if(type[1] == "lin") {
    x <- x[,,1]
    r <- range(x)
    q <- stats::quantile(x, c(0.05, 0.95), na.rm = TRUE)

    ## consider special case for q == 0
    if(sum(q) > 0) {
      x <- (r[2] * (x - q[1])) / (q[2] - q[1])
      x[x < 0] <- r[1]
      x[x > r[2]] <- r[2]
    }
  }

  if(type[1] == "hist")
    x <- matrix(stats::ecdf(x)(x) * 255, ncol = ncol(x))

  return(x)
}

  .ticks.lab.at <- function(max.size) {
    if (max.size < 13) {
      lab <- 1:max.size
    } else {
      lab <- pretty(1:max.size, n = min(5, max.size))
      lab[lab == 0] <- 1
      lab <- lab[lab <= max.size]
      if (max.size - max(lab) > 2)
        lab <- c(lab, max.size)
    }
    ## the centre of the first and last squares are respectively at 0
    ## and 1, all others are spread linearly between these two
    at <- lab / max.size
    list(lab = lab, at = (at - at[1]) / (1 - at[1]))
  }

# Plot settings -----------------------------------------------------------
plot_settings <- modifyList(x = list(
    main = "RLum.Data.Image",
    axes = TRUE,
    xlab = "Length [px]",
    ylab = "Height [px]",
    xlim = c(1,dim(object)[1]),
    ylim = c(1,dim(object)[2]),
    zlim = range(object),
    zlim_image = NULL,
    mtext = "",
    legend = TRUE,
    useRaster = TRUE,
    stretch = "hist",
    col = c(grDevices::hcl.colors(50, palette = "Inferno")),
    cex = 1,
    digits = 1,
    scientific = TRUE
  ), val = list(...), keep.null = TRUE)

  ## set frames
  if(!is.null(frames)) {
    frames[1] <- max(1,min(frames))
    frames[length(frames)] <- min(dim(object)[3],max(frames))
    object <- object[,,frames,drop = FALSE]
  }

  ## enforce xlim, ylim and zlim directly here
  ## xlim, ylim
  object[] <- object[
    max(plot_settings$xlim[1], 1):min(plot_settings$xlim[2], dim(object)[1]),
    max(plot_settings$ylim[1], 1):min(plot_settings$ylim[2], dim(object)[2]),,
    drop = FALSE]

  ## zlim
  object[object <= plot_settings$zlim[1]] <- max(0,plot_settings$zlim[1])
  object[object >= plot_settings$zlim[2]] <- min(max(object),plot_settings$zlim[2])

  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ##par setting for possible combination with plot method for RLum.Analysis objects
  if(par.local) par(mfrow=c(1,1), cex = plot_settings$cex)

  for (i in 1:dim(object)[3]) {
    x <- object[, , i, drop = FALSE]
    range_x <- range(x)
    main <- ifelse(length(plot_settings$main) == 1,
                   paste0(plot_settings$main, " #", i),
                   paste0(plot_settings$main[i]))

    if (plot.type == "plot.raster") {
      par(mar = c(4.5, 4.5, 4, 3))
      plot.fun <- graphics::image
      x <-.stretch(x, type = plot_settings$stretch)
      useRaster <- plot_settings$useRaster
    } else if (plot.type == "contour") {
      plot.fun <- graphics::contour
      x <- x[, , 1]
      useRaster <- NULL
    }
    do.call(plot.fun, list(
        x = x,
        useRaster = useRaster,
        main = main,
        axes = FALSE,
        zlim = plot_settings$zlim_image %||% range(x),
        xlab = plot_settings$xlab,
        ylab = plot_settings$ylab,
        col = plot_settings$col))
    graphics::box()

    ## axes
    if (plot_settings$axes) {
        xticks <- .ticks.lab.at(nrow(x))
        graphics::axis(side = 1, at = xticks$at, labels = xticks$lab)
        yticks <- .ticks.lab.at(ncol(x))
        graphics::axis(side = 2, at = yticks$at, labels = yticks$lab)
    }

    ## add legend
    if (plot.type == "plot.raster" && plot_settings$legend) {
        col_grad <- plot_settings$col[seq(1, length(plot_settings$col), length.out = 14)]
        slices <- seq(0,1,length.out = 15)
        par.usr <- par("usr")
        for(s in 1:(length(slices) - 1)){
          graphics::rect(
            xleft = par.usr[4] * 1.01,
            xright = par.usr[4] * 1.03,
            ybottom = slices[s],
            ytop =  slices[s + 1],
            col = col_grad[s],
            xpd = TRUE,
            border = TRUE)
        }

        text(
          x = c(par.usr[4], par.usr[4]) * 1.04,
          y = c(par.usr[3], par.usr[2]),
          labels = format(plot_settings$zlim_image %||% range_x,
                          digits = plot_settings$digits,
                          scientific = plot_settings$scientific),
          xpd = TRUE,
          cex = 0.7,
          srt = 270,
          pos = 3)
    }

    ## add mtext
    mtext(side = 3, plot_settings$mtext)
  }
}

## create dataset to test
set.seed(1)
image <- as(array(rnorm(1000), dim = c(10,10,10)), "RLum.Data.Image")
image18 <- as(array(rnorm(18 * 18), dim = c(18, 18, 1)), "RLum.Data.Image")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RLum.Data.Image("image"),
               "[plot_RLum.Data.Image()] 'object' should be of class 'RLum.Data.Image'",
               fixed = TRUE)
  expect_error(plot_RLum.Data.Image(image, plot.type = "error"),
               "[plot_RLum.Data.Image()] 'plot.type' should be one of 'plot.raster'",
               fixed = TRUE)
})

test_that("test functionality", {
  testthat::skip_on_cran()

  expect_silent(plot_RLum.Data.Image(image, plot.type = "plot.raster",
                                     frames = c(2, 4)))

  ## empty image
  expect_null(plot_RLum.Data.Image(set_RLum("RLum.Data.Image")))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("raster",
                              plot_RLum.Data.Image(image,
                                                   plot.type = "plot.raster"))
  vdiffr::expect_doppelganger("raster lin stretch",
                              plot_RLum.Data.Image(image,
                                                   plot.type = "plot.raster",
                                                   frames = 10,
                                                   mtext = "Test",
                                                   stretch = "lin"))
  vdiffr::expect_doppelganger("raster zlim_image",
                              plot_RLum.Data.Image(image,
                                                   plot.type = "plot.raster",
                                                   frames = 1,
                                                   stretch = NULL,
                                                   scientific = FALSE,
                                                   zlim_image = c(0, 1)))
  vdiffr::expect_doppelganger("contour",
                              plot_RLum.Data.Image(image,
                                                   plot.type = "contour",
                                                   mtext = "Test",
                                                   frames = 5))
  vdiffr::expect_doppelganger("contour 18",
                              plot_RLum.Data.Image(image18,
                                                   plot.type = "contour"))
  })
})

#' @title Cosmic-ray removal and spectrum smoothing for RLum.Data.Spectrum objects
#'
#' @description
#' The function provides several methods for cosmic-ray removal and spectrum
#' smoothing for [Luminescence::RLum.Data.Spectrum-class] objects, and those embedded in
#' [list] or [Luminescence::RLum.Analysis-class] objects.
#'
#' @details
#'
#' **`method = "Pych"`**
#'
#' This method applies the cosmic-ray removal algorithm described by Pych
#' (2003). There are some differences with respect to the publication:
#'
#' - For interpolation between neighbouring values, the median is used instead
#' of the mean.
#' - The number of breaks in the histogram is set to half the number of the
#' input values.
#'
#' For further details see references below.
#'
#' **Other methods**
#' Arguments supported through `...`
#' \tabular{llll}{
#' METHOD \tab ARGUMENT \tab TYPE \tab REMARKS\cr
#' `"smooth"` \tab `kind` \tab  [character] \tab see [stats::smooth] \cr
#'   \tab `twiceit` \tab  [logical] \tab see [stats::smooth] \cr
#' `"smooth.spline"` \tab `spar` \tab  [numeric] \tab see [stats::smooth.spline] \cr
#' `"smooth_RLum"` \tab `k` \tab [numeric] \tab see [Luminescence::smooth_RLum]\cr
#'  \tab `fill` \tab [numeric] \tab see [Luminescence::smooth_RLum]\cr
#'  \tab `align` \tab [character] \tab see [Luminescence::smooth_RLum]\cr
#'  \tab `method_smooth_RLum` \tab [character] \tab see [Luminescence::smooth_RLum]\cr
#'}
#'
#' **Best practice**
#'
#' There is no single silver-bullet strategy for cosmic-ray removal, as it depends
#' on the characteristic of the detector and the chosen settings. For instance,
#' high values for pixel binning will improve the light output, but also cause
#' multiple pixels to be affected by a single cosmic ray. The same is valid for
#' longer integration times. The best strategy is to combine methods and ensure
#' that the spectrum is not distorted on a case-to-case basis.
#'
#' **How to combine methods?**
#'
#' Different methods can be combined by applying the method repeatedly to the
#' dataset (see example).
#'
#' @param object [Luminescence::RLum.Data.Spectrum-class] or
#' [Luminescence::RLum.Analysis-class] (**required**):
#' input object to be treated, which can be also provided as [list]. If an
#' [Luminescence::RLum.Analysis-class] object is provided, only its
#' [Luminescence::RLum.Data.Spectrum-class]
#' objects are treated. Please note: this mixing of objects does not work for
#' a list of `RLum.Data` objects.
#'
#' @param method [character] (*with default*):
#' Defines method that is applied for cosmic ray removal. Allowed methods are
#' `smooth`, the default, ([stats::smooth]), `smooth.spline` ([stats::smooth.spline]),
#' `smooth_RLum` ([Luminescence::smooth_RLum]) and `Pych`. See details for further information.
#'
#' @param method.Pych.smoothing [integer] (*with default*):
#' Smoothing parameter for cosmic ray removal according to Pych (2003).
#' The value defines how many neighbouring values in each frame are used for smoothing
#' (e.g., `2` means that the two previous and two following values are used).
#'
#' @param method.Pych.threshold_factor [numeric] (*with default*):
#' Threshold for zero-bins in the histogram. Small values mean that more peaks
#' are removed, but signal might be also affected by this removal.
#'
#' @param MARGIN [integer] (*with default*):
#' on which part the function cosmic ray removal should be applied on:
#'
#' - 1 = along the time axis (line by line),
#' - 2 = along the wavelength axis (column by column).
#'
#' **Note:** This argument only affects methods `smooth`, `smooth.spline` and
#' `smooth_RLum`.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' If `TRUE` the histograms used for the cosmic-ray removal are returned as plot
#' including the used threshold. Note: A separate plot is returned for each frame!
#' Currently only for `method = "Pych"` a graphical output is provided.
#'
#' @param ... further arguments and graphical parameters that will be passed
#' to the [stats::smooth], [stats::smooth.spline] or [Luminescence::smooth_RLum].
#' See details for more information.
#'
#' @return Returns same object as input.
#'
#' @section Function version: 0.4.1
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Analysis-class],
#' [stats::smooth], [stats::smooth.spline], [Luminescence::smooth_RLum]
#'
#' @references
#' Pych, W., 2004. A Fast Algorithm for Cosmic-Ray Removal from
#' Single Images. The Astronomical Society of the Pacific 116 (816), 148-153.
#' \doi{10.1086/381786}
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(1) - use with your own data and combine (uncomment for usage)
#' ## run two times the default method and smooth with another method
#' ## your.spectrum <- apply_CosmicRayRemoval(your.spectrum, method = "Pych")
#' ## your.spectrum <- apply_CosmicRayRemoval(your.spectrum, method = "Pych")
#' ## your.spectrum <- apply_CosmicRayRemoval(your.spectrum, method = "smooth")
#'
#' @export
apply_CosmicRayRemoval <- function(
  object,
  method = c("smooth", "smooth.spline", "smooth_RLum", "Pych"),
  method.Pych.smoothing = 2,
  method.Pych.threshold_factor = 3,
  MARGIN = 2,
  verbose = FALSE,
  plot = FALSE,
  ...
) {
  .set_function_name("apply_CosmicRayRemoval")
  on.exit(.unset_function_name(), add = TRUE)

  # Self-call ----------------------------------------------------------------------------------
  ##Black magic: The function recalls itself until all RLum.Data.Spectrum objects have been treated
  ##If you want to test the basics of the function please only use a single RLum.Data.Spectrum-object
  ##if it comes in as an RLum.Analysis object ... make a list out of it
  class_original <- NULL
  if(inherits(object, "RLum.Analysis")){
    object <- list(object)
    class_original <- "RLum.Analysis"
  }

  ##handle the list and recall
  if(inherits(object, "list")){
    results_list <- lapply(object, function(o){
      ##RLum.Analysis
      if(inherits(o, "RLum.Analysis")){
         ##get id of RLum.Data.Spectrum objects in this object
        record_id.spectra <- vapply(o@records, inherits, "RLum.Data.Spectrum",
                                    FUN.VALUE = logical(1))

         ##rewrite o
         temp_o <- o@records[record_id.spectra]

      }else{
        record_id.spectra <- NULL
        temp_o <- o
      }

      ##call function
      results <- do.call(
        what = apply_CosmicRayRemoval,
        args = c(list(
          object = temp_o,
          method = method,
          method.Pych.smoothing = method.Pych.smoothing,
          method.Pych.threshold_factor = method.Pych.threshold_factor,
          MARGIN = MARGIN,
          verbose = verbose,
          plot = plot),
          list(...)))

      ##combine in RLum.Analysis object if needed
      if(!is.null(record_id.spectra)){
        o@records[record_id.spectra] <- results
        return(o)
      }else{
        return(results)
      }
    })

    ##final return, make sure that we return what we had as input
    if(!is.null(class_original)){
      return(results_list[[1]])
    }else{
      return(results_list)
    }
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Data.Spectrum")
  if (length(object@data) < 2)
    .throw_error("'object' contains no data")

  method <- .validate_args(method, c("smooth", "smooth.spline", "smooth_RLum", "Pych"))
  .validate_positive_scalar(method.Pych.smoothing, int = TRUE)
  .validate_positive_scalar(method.Pych.threshold_factor)
  .validate_args(MARGIN, c(1, 2))
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

  ##deal with additional arguments
  extraArgs <- modifyList(
    x = list(
      kind = "3RS3R",
      twiceit = TRUE,
      spar = NULL,
      method_smooth_RLum = "median",
      k = NULL,
      fill = NA,
      align = "right"),
    val = list(...),
    keep.null = TRUE)

  # Apply methods -0------------------------------------------------------------
  object.data.temp.smooth <- switch(method,

  ## +++++++++++++++++++++++++++++++++++ (smooth) ++++++++++++++++++++++++++++##
  smooth = {
    ##apply smooth
    apply(
      X = object@data,
      MARGIN = MARGIN,
      FUN = stats::smooth,
      kind = extraArgs$kind[1],
      twiceit = extraArgs$twiceit[1]
    )
  },

  ## +++++++++++++++++++++++++++++++++++ (smooth.spline) +++++++++++++++++++++##
  smooth.spline = {
    ## wrap the function in a new function to access the data more easily
    temp_smooth.spline <- function(x, spar){
      stats::smooth.spline(x, spar = spar)$y
    }

    ##apply smooth.spline
    apply(
        X = object@data,
        MARGIN = MARGIN,
        FUN = temp_smooth.spline,
        spar = extraArgs$spar[1])
  },

  ## +++++++++++++++++++++++++++++++++++ (smooth_RLum) +++++++++++++++++++++##
  smooth_RLum = {
    object.data.temp.smooth <- apply(
        X = object@data,
        MARGIN = MARGIN,
        FUN = .smoothing,
        method = extraArgs$method_smooth_RLum[1],
        k = max(c(1, min(c(
          if(MARGIN == 2)
           floor(ncol(object@data)/(ncol(object@data)/2))
          else
           floor(nrow(object@data)/(nrow(object@data)/2)),
          extraArgs$k[1])))),
        fill = extraArgs$fill[1],
        align = extraArgs$align[1])

    ## remove NA values
    if (is.na(extraArgs$fill[1])) {
      id_NA <- which(matrixStats::rowAnyNAs(object.data.temp.smooth))
      if (length(id_NA) > 0) {
        object.data.temp.smooth <- object.data.temp.smooth[-id_NA, , drop = FALSE]
        if (MARGIN == 1)
          object@data <- object@data[,-id_NA, drop = FALSE]
        else
          object@data <- object@data[-id_NA,, drop = FALSE]
      }
    }
    object.data.temp.smooth
  },

  ## +++++++++++++++++++++++++++++++++++ (Pych) ++++++++++++++++++++++++++++++##
  Pych = {
    ## apply smoothing
    frame.idx <- 0
    apply(object@data, 2, function(data.col) {
      frame.idx <<- frame.idx + 1
      ##(1) - calculate sd for each subframe
      temp.sd <- sd(data.col)

      ##(2) - correct estimation of sd by 1-sigma clipping
      temp.sd.corr <- sd(data.col[data.col >= (mean(data.col) - temp.sd) &
                                  data.col <= (mean(data.col) + temp.sd)])

      ##(3) - construct histogram of count distribution
      temp.hist <- hist(data.col,
                        breaks = length(data.col)/2, plot = FALSE)

      ##(4) - find mode of the histogram (e.g. peak)
      temp.hist.max <- which.max(temp.hist$counts)

      ##(5) - find gaps in the histogram (bins with zero value)
      temp.hist.zerobin <- which(temp.hist$counts == 0)

      ##(5.1)
      ##select just values right from the peak
      temp.hist.zerobin <- temp.hist.zerobin[
        (temp.hist.max[1] + 1):length(temp.hist.zerobin)]

      ##(5.2)
      ##select non-zerobins
      temp.hist.nonzerobin <- which(temp.hist$counts != 0)
      temp.hist.nonzerobin <- temp.hist.nonzerobin[
        temp.hist.nonzerobin >=  (temp.hist.zerobin[1]-1)]

      ##(6) - find the first gap which is wider than the threshold
      temp.hist.nonzerobin.diff <- diff(
        temp.hist$breaks[temp.hist.nonzerobin])

      ## select the first value where the threshold is reached
      ## factor 3 is defined by Pych (2003)
      temp.hist.thres <- which(
        temp.hist.nonzerobin.diff >= method.Pych.threshold_factor * temp.sd.corr)[1]

      ##(7) - use counts above the threshold and recalculate values
      ## on all further values
      if(!is.na(temp.hist.thres)){
        data.col <- sapply(1:length(data.col), function(n) {

          if((n + method.Pych.smoothing) <= length(data.col) &
             (n - method.Pych.smoothing) >= 0){
            ifelse(
              data.col[n] >= temp.hist$breaks[temp.hist.thres],
              median(data.col[(n-method.Pych.smoothing):(n+method.Pych.smoothing)]),
              data.col[n])
          }else{
            data.col[n]
          }
        })
      }

      ##(8) - return histogram used for the removal as plot
      if(plot){
        plot(temp.hist,
             xlab = "Signal intensity [a.u.]",
             main = "Cosmic-ray removal histogram")

        abline(v = temp.hist$breaks[temp.hist.thres],
               col = "red")

        mtext.text <- sprintf("Frame: %d (%s)", frame.idx, colnames(object@data)[frame.idx])
        if(!is.na(temp.hist$breaks[temp.hist.thres])){
          legend("topright", "threshold" ,lty = 1, lwd = 1, col = "red", bty = "n")
        }else{
          mtext.text <- paste(mtext.text, "- no threshold applied")
        }
        mtext(side = 3, mtext.text)
      }

      ##(9) - return information on the amount of removed cosmic-rays
      if(verbose){
        #sum up removed counts values above the threshold
        sum.corrected.channels <- try(
          sum(temp.hist$counts[temp.hist.thres:length(temp.hist$counts)]),
          silent = TRUE)

        if (inherits(sum.corrected.channels, "try-error"))
          sum.corrected.channels <- 0

        .throw_message(">> ", sum.corrected.channels,
                       " channels corrected in frame ", frame.idx, "\n", error = FALSE)
      }

      return(data.col)
    })#end loop
  })

  ## Rotate matrix if required
  if (MARGIN[1] == 1 && method != "Pych")
    object.data.temp.smooth <- t(object.data.temp.smooth)

  # Correct row and column names --------------------------------------------
  colnames(object.data.temp.smooth) <- colnames(object@data)
  rownames(object.data.temp.smooth) <- rownames(object@data)

  # Return Output------------------------------------------------------------
  set_RLum(
    class = "RLum.Data.Spectrum",
    recordType = object@recordType,
    curveType = object@curveType,
    data = object.data.temp.smooth,
    info = object@info)
}

## load data
data(ExampleData.XSYG, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(apply_CosmicRayRemoval("error"),
               "'object' should be of class 'RLum.Data.Spectrum'")
  expect_error(apply_CosmicRayRemoval(TL.Spectrum, method = "error"),
               "'method' should be one of 'smooth', 'smooth.spline', 'smooth_RLum' or 'Pych'",
               fixed = TRUE)
  expect_error(apply_CosmicRayRemoval(set_RLum("RLum.Data.Spectrum"),
                                      "'object' contains no data"))

  expect_error(apply_CosmicRayRemoval(TL.Spectrum,
                                      method.Pych.smoothing = iris),
               "'method.Pych.smoothing' should be a single positive integer value")
  expect_error(apply_CosmicRayRemoval(TL.Spectrum,
                                      method.Pych.threshold_factor = -2),
               "'method.Pych.threshold_factor' should be a single positive value")
  expect_error(apply_CosmicRayRemoval(TL.Spectrum, MARGIN = 3),
               "'MARGIN' should be one of '1' or '2'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "Pych"),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth"),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum"),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 1, k = 0),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 1, k = 4),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 1, k = 10000),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 2, k = 0),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 2, k = 4),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 2, k = 4,
                                              method_smooth_RLum = "mean"),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              MARGIN = 2, k = 10000),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth_RLum",
                                              k = 10),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth",
                                              MARGIN = 1),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum, method = "Pych",
                                              method.Pych.smoothing = 2,
                                              method.Pych.threshold_factor = 2),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(apply_CosmicRayRemoval(TL.Spectrum,
                                              method = "smooth.spline",
                                              MARGIN = 1,
                                              kind = "3RS3R",
                                              twiceit = TRUE,
                                              spar = NULL),
                       tolerance = snapshot.tolerance)
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  expect_message(expect_message(
      apply_CosmicRayRemoval(TL.Spectrum, method = "Pych", MARGIN = 2,
                             verbose = TRUE, plot = TRUE),
      "1024 channels corrected in frame 4"),
      "0 channels corrected in frame 24")
  })

  ##construct objects for different tests
  RLum_list <- list(TL.Spectrum)
  RLum.Analysis <- set_RLum("RLum.Analysis", records = RLum_list)
  RLum.Analysis_list <- list(RLum.Analysis)
  RLum_list_mixed <- list(TL.Spectrum, set_RLum("RLum.Data.Curve"))
  RLum.Analysis_mixed <-  set_RLum("RLum.Analysis", records = RLum_list_mixed )
  RLum.Analysis_mixed_list <- list(RLum.Analysis_mixed)

  ##run tests
  expect_type(apply_CosmicRayRemoval(RLum_list),"list")
  expect_s4_class(apply_CosmicRayRemoval(RLum.Analysis), class = "RLum.Analysis")
  expect_type(apply_CosmicRayRemoval(RLum.Analysis_list), "list")
  expect_error(apply_CosmicRayRemoval(RLum_list_mixed))
  expect_s4_class(apply_CosmicRayRemoval(RLum.Analysis_mixed), class = "RLum.Analysis")
  expect_type(apply_CosmicRayRemoval(RLum.Analysis_mixed_list), "list")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  frame <- trim_RLum.Data(TL.Spectrum, trim_range = c(10, 10))

  SW({
  vdiffr::expect_doppelganger("default",
                              apply_CosmicRayRemoval(frame,
                                                     method = "Pych",
                                                     plot = TRUE))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 985
  expect_silent(apply_CosmicRayRemoval(TL.Spectrum, method = "smooth_RLum",
                                       method_smooth_RLum = "Carter_etal_2018"))

  ## issue 987
  expect_silent(apply_CosmicRayRemoval(TL.Spectrum, method = "Pych",
                                       MARGIN = 1))
})

#' @title Plot function for an RLum.Data.Spectrum S4 class object
#'
#' @description
#' The function provides a standardised plot output for spectrum data of an
#' [Luminescence::RLum.Data.Spectrum-class] class object. The purpose of this function is to
#' provide easy and straightforward spectra plotting, not a fully customised
#' access to all plot parameters. If this is wanted, standard R plot
#' functionality should be used instead.
#'
#' **Matrix structure** \cr (cf. [Luminescence::RLum.Data.Spectrum-class])
#'
#' - `rows` (x-values): wavelengths/channels (`xlim`, `xlab`)
#' - `columns` (y-values): time/temperature (`ylim`, `ylab`)
#' - `cells` (z-values): count values (`zlim`, `zlab`)
#'
#' *Note: This nomenclature is valid for all plot types of this function!*
#'
#' **Nomenclature for value limiting**
#'
#' - `xlim`: Limits values along the wavelength axis
#' - `ylim`: Limits values along the time/temperature axis
#' - `zlim`: Limits values along the count value axis
#'
#' **Details on the plot functions**
#'
#' Spectrum is visualised as 3D or 2D plot. Both plot types are based on
#' internal R plot functions.
#'
#'**`plot.type = "persp"`**
#'
#' Arguments that will be passed to [graphics::persp]:
#'
#' - `shade`: default is `0.4`
#' - `phi`: default is `15`
#' - `theta`: default is `-30`
#' - `lphi`: default is `15`
#' - `ltheta`: default is `-30`
#' - `expand`: default is `1`
#' - `axes`: default is `TRUE`
#' - `box`: default is `TRUE`; accepts `"alternate"` for a custom plot design
#' - `ticktype`: default is `detailed`, `r`: default is `10`
#'
#' **Note:** Further parameters can be adjusted via `par`. For example
#' to set the background transparent and reduce the thickness of the lines use:
#' `par(bg = NA, lwd = 0.7)` before the function call.
#'
#'**`plot.type = "single"`**
#'
#' Per frame a single curve is returned. Frames are time or temperature
#' steps.
#'
#' -`frames`: pick the frames to be plotted (depends on the binning!). Check without
#' this setting before plotting.
#'
#'**`plot.type = "multiple.lines"`**
#'
#' All frames plotted in one frame.
#'
#'-`frames`: pick the frames to be plotted (depends on the binning!). Check without
#' this setting before plotting.
#'
#'**`plot.type = "image"` or `plot.type = "contour"`**
#'
#' These plot types use the R functions [graphics::image] or [graphics::contour].
#' The advantage is that many plots can be arranged conveniently using standard
#' R plot functionality. If `plot.type = "image"` a contour is added by default,
#' which can be disabled using the argument `contour = FALSE` to add own contour
#' lines of choice.
#'
#'**`plot.type = "transect"`**
#'
#' Depending on the selected wavelength/channel range a transect over the
#' time/temperature (y-axis) will be plotted along the wavelength/channels
#' (x-axis). If the range contains more than one channel, values (z-values) are
#' summed up. To select a transect use the `xlim` argument, e.g.
#' `xlim = c(300,310)` plot along the summed up count values of channel
#' 300 to 310.
#'
#' **Further arguments that will be passed (depending on the plot type)**
#'
#' `xlab`, `ylab`, `zlab`, `xlim`, `ylim`, `zlim`, `main`, `mtext`,
#' `box`, `pch`, `type` (`"single"`, `"multiple.lines"`, `"interactive"`),
#' `col`, `border`, `lwd`, `bty`, `showscale` (`"interactive"`, `"image"`)
#' `contour`, `contour.col` (`"image"`), `labcex` (`"image"`, `"contour"`),
#' `n_breaks` (`"image"`), `legend` (`TRUE`/`FALSE`),
#' `legend.pos` (`"image"`), `legend.horiz` (`TRUE`/`FALSE` | `"image"`)
#'
#' @param object [Luminescence::RLum.Data.Spectrum-class] or [matrix] (**required**):
#' S4 object of class [Luminescence::RLum.Data.Spectrum-class] or a `matrix` containing count
#' values of the spectrum.\cr
#' Please note that in case of a matrix row names and col names are set
#' automatically if not provided.
#'
#' @param par.local [logical] (*with default*):
#' use local graphical parameters for plotting, e.g. the plot is shown in one
#' column and one row. If `par.local = FALSE` global parameters are inherited.
#'
#' @param plot.type [character] (*with default*):
#' for a 3D-plot use `"persp"` or `"interactive"`; for a 2D-plot you can use
#' `"image"`, `"contour"`, `"single"` or `"multiple.lines"` (along the time or
#' temperature axis) or `"transect"` (along the wavelength axis).
#'
#' @param optical.wavelength.colours [logical] (*with default*):
#' use optical wavelength colour palette. Note: For this, the spectrum range is
#' limited: `c(350,750)`. Own colours can be set with the argument `col`. If
#' you provide already binned spectra, the colour assignment is likely to be
#' wrong, since the colour gradients are calculated using the bin number.
#'
#' @param bg.spectrum [Luminescence::RLum.Data.Spectrum-class] or [matrix] (*optional*):
#' spectrum used for the background subtraction. The background spectrum should
#' be measured using the same setting as the signal spectrum. The argument
#' `bg.channels` controls how the subtraction is performed: if `bg.channels`
#' is not specified or the number of channels is identical between the signal
#' and background spectra, a channel-wise subtraction is performed; otherwise,
#' the *arithmetic mean* is calculated and subtracted from the signal.
#'
#' @param bg.channels [vector] (*optional*):
#' channels used for background subtraction. If the number of channels is
#' identical between the signal and background spectra, a channel-wise
#' subtraction is performed; otherwise this number is used to select channels
#' for calculating the *arithmetic mean*  If a spectrum is provided via
#' `bg.spectrum`, this argument only works on the background spectrum.
#'
#' **Note:** Background subtraction is applied prior to channel binning!
#'
#' @param bin.rows [integer] (*with default*):
#' allow summing-up wavelength channels (horizontal binning),
#' e.g. `bin.rows = 2` two channels are summed up.
#' Binning is applied after the background subtraction.
#'
#' @param bin.cols [integer] (*with default*):
#' allow summing-up channel counts (vertical binning) for plotting,
#' e.g. `bin.cols = 2` two channels are summed up.
#' Binning is applied after the background subtraction.
#'
#' @param norm [logical], [character] (*optional*):
#' if logical, whether curve normalisation should occur (`FALSE` by default);
#' alternatively, one of the values detailed in [Luminescence::plot_RLum.Data.Curve].
#' The normalisation is applied after binning.
#'
#' @param rug [logical] (*with default*):
#' enable/disable colour rug. Currently only implemented for plot
#' type `multiple.lines` and `single`.
#'
#' @param limit_counts [numeric] (*optional*):
#' value to limit all count values to this value, i.e. all count values above
#' this threshold will be replaced by this threshold. This is helpful
#' especially in case of TL-spectra.
#'
#' @param xaxis.energy [logical] (*with default*): enable/disable using energy
#' instead of wavelength on the x-axis. Function [Luminescence::convert_Wavelength2Energy]
#' is used to perform the conversion.
#'
#' **Note:** Besides being used in setting the axis, with this option the
#' the spectrum is recalculated in terms of intensity, see details.
#'
#' @param legend.text [character] (*with default*):
#' possibility to provide own legend text. This argument is only considered for
#' plot types providing a legend, e.g. `plot.type = "transect"`.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output. If
#' the plot output is disabled, the [matrix] used for the plotting and the
#' calculated colour values (as attributes) are returned. This way, the
#' (binned, transformed etc.) output can be used in other functions and
#' packages, such as plotting with the package `'plot3D'`.
#'
#' @param ... further arguments and graphical parameters that will be passed
#' to the `plot` function.
#'
#' @return
#' Returns a plot and the transformed `matrix` used for plotting with some
#' useful attributes such as `colour` and `pmat` (the transpose matrix from
#' [graphics::persp]).
#'
#' @note Not all additional arguments (`...`) will be passed similarly!
#'
#' @section Function version: 0.6.13
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Spectrum-class],
#' [Luminescence::convert_Wavelength2Energy], [Luminescence::plot_RLum],
#' [graphics::persp], [plotly::plot_ly], [graphics::contour], [graphics::image]
#'
#' @keywords aplot
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.XSYG, envir = environment())
#'
#' ##(1)plot simple spectrum (2D) - image
#' plot_RLum.Data.Spectrum(
#'  TL.Spectrum,
#'  plot.type="image",
#'  xlim = c(310,750),
#'  ylim = c(0,300),
#'  bin.rows=10,
#'  bin.cols = 1)
#'
#' ##(2) plot spectrum (3D)
#' plot_RLum.Data.Spectrum(
#'   TL.Spectrum,
#'   plot.type="persp",
#'   xlim = c(310,750),
#'   ylim = c(0,100),
#'   bin.rows=10,
#'   bin.cols = 1)
#'
#'##(3) plot spectrum on energy axis
#'##please note the background subtraction
#'plot_RLum.Data.Spectrum(TL.Spectrum,
#' plot.type="persp",
#' ylim = c(0,200),
#' bin.rows=10,
#' bg.channels = 10,
#' bin.cols = 1,
#' xaxis.energy = TRUE)
#'
#' ##(4) plot multiple lines (2D) - multiple.lines (with ylim)
#' plot_RLum.Data.Spectrum(
#'  TL.Spectrum,
#'  plot.type="multiple.lines",
#'  xlim = c(310,750),
#'  ylim = c(0,100),
#'  bin.rows=10,
#'  bin.cols = 1)
#'
#' \dontrun{
#'  ##(4) interactive plot using the package plotly ("surface")
#'  plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
#'  xlim = c(310,750), ylim = c(0,300), bin.rows=10,
#'  bin.cols = 1)
#'
#'  ##(5) interactive plot using the package plotly ("contour")
#'  plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
#'  xlim = c(310,750), ylim = c(0,300), bin.rows=10,
#'  bin.cols = 1,
#'  type = "contour",
#'  showscale = TRUE)
#'
#'  ##(6) interactive plot using the package plotly ("heatmap")
#'  plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
#'  xlim = c(310,750), ylim = c(0,300), bin.rows=10,
#'  bin.cols = 1,
#'  type = "heatmap",
#'  showscale = TRUE)
#' }
#'
#' @export
plot_RLum.Data.Spectrum <- function(
  object,
  par.local = TRUE,
  plot.type = c("contour", "persp", "single", "multiple.lines",
                "image", "transect", "interactive"),
  optical.wavelength.colours = TRUE,
  bg.spectrum = NULL,
  bg.channels = NULL,
  bin.rows = 1,
  bin.cols = 1,
  norm = FALSE,
  rug = TRUE,
  limit_counts = NULL,
  xaxis.energy = FALSE,
  legend.text = NULL,
  plot = TRUE,
  ...
) {
  .set_function_name("plot_RLum.Data.Spectrum")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("RLum.Data.Spectrum", "matrix"))

  if (inherits(object, "matrix")) {
    if (is.null(colnames(object))) {
      colnames(object) <- seq_len(ncol(object))
    }
    if (is.null(rownames(object))) {
      rownames(object) <- seq_len(nrow(object))
    }

    object <- set_RLum(class = "RLum.Data.Spectrum", data = object)
    .throw_message("Input has been converted to an 'RLum.Data.Spectrum' ",
                   "object using set_RLum()", error = FALSE)
  }

  if (length(object@data) < 2) {
    .throw_error("'object' contains no data")
  }
  ## `norm` is not validated here but will be validated by normalise_RLum()
  .validate_class(bg.spectrum, c("RLum.Data.Spectrum", "matrix"), null.ok = TRUE)
  plot.type <- .validate_args(plot.type, c("contour", "persp", "single",
                                           "multiple.lines", "image",
                                           "transect", "interactive"))
  .validate_positive_scalar(bin.rows, int = TRUE)
  .validate_positive_scalar(bin.cols, int = TRUE)

  ##XSYG
  ##check for curveDescripter
  if ("curveDescripter" %in% names(object@info)) {
    temp.lab <- strsplit(object@info$curveDescripter, split = ";")[[1]]
    xlab <- temp.lab[2]
    ylab <- temp.lab[1]
    zlab <- temp.lab[3]
  }else{
    xlab <- "Row values [a.u.]"
    ylab <- "Column values [a.u.]"
    zlab <- "Cell values [a.u.]"
  }

  # Do energy axis conversion -------------------------------------------------------------------
  if (xaxis.energy){
    ##conversion
    object <- convert_Wavelength2Energy(object, digits = 5)

    ##modify row order (otherwise subsequent functions, like persp, have a problem)
    object@data[] <- object@data[order(as.numeric(rownames(object@data))),]
    rownames(object@data) <- sort(as.numeric(rownames(object@data)))

    ## update axis label
    xlab <- "Energy [eV]"
  }

  ## check for duplicated column names (e.g., temperature not increasing)
  if (anyDuplicated(colnames(object@data)) > 0) {
    .throw_warning("Duplicated column names found, replaced by index")
    colnames(object@data) <- 1:ncol(object@data[])
  }

  ##deal with addition arguments
  extraArgs <- list(...)

  main <- extraArgs$main %||% "RLum.Data.Spectrum"
  zlab <- extraArgs$zlab %||% ifelse(plot.type == "multiple.lines", ylab, zlab)
  xlab <- extraArgs$xlab %||% xlab
  ylab <- extraArgs$ylab %||% ifelse(plot.type %in% c("single", "multiple.lines"),
                                     "Luminescence [cts/channel]", ylab)
  xlim <- extraArgs$xlim %||% range(as.numeric(rownames(object@data)))
  .validate_class(xlim, "numeric", length = 2)
  ylim <- extraArgs$ylim %||% range(as.numeric(colnames(object@data)))
  .validate_class(ylim, "numeric", length = 2)
  #for zlim see below

  mtext <- extraArgs$mtext %||% ""
  cex <- extraArgs$cex %||% 1
  phi <- extraArgs$phi %||% 15
  theta <- extraArgs$theta %||% -30
  lphi <- extraArgs$lphi %||% 15
  ltheta <- extraArgs$ltheta %||% -30
  r <- extraArgs$r %||% 10
  shade <- extraArgs$shade %||% 0.4
  expand <- extraArgs$expand %||% 0.6
  border <- extraArgs$border
  box <- extraArgs$box %||% TRUE
  axes <- extraArgs$axes %||% TRUE
  ticktype <- extraArgs$ticktype %||% "detailed"
  log <- extraArgs$log %||% ""
  labcex <- extraArgs$labcex %||% 0.6
  type <- extraArgs$type %||% ifelse(plot.type == "interactive", "surface", "l")
  pch <- extraArgs$pch %||% 1
  lwd <- extraArgs$lwd %||% 1
  bty <- extraArgs$bty
  sub <- extraArgs$sub %||% ""
  #for plotly::plot_ly
  showscale <- extraArgs$showscale %||% FALSE

  ## further plot settings
  plot_settings <- modifyList(
    x = list(
      legend = TRUE,
      legend.pos = "topright",
      legend.horiz = FALSE,
      n_breaks = 50
    ),
    val = extraArgs)

  # prepare values for plot ---------------------------------------------------
  ##copy data
  temp.xyz <- object@data

  ##check for NULL column names
  if(is.null(colnames(temp.xyz)))
    colnames(temp.xyz) <- 1:ncol(temp.xyz)

  if(is.null(rownames(temp.xyz)))
    rownames(temp.xyz) <- 1:nrow(temp.xyz)

  ##check for the case of a single column matrix
  if(ncol(temp.xyz)>1){
    x.vals <- as.numeric(rownames(temp.xyz))
    y.vals <- as.numeric(colnames(temp.xyz))

    ## limit the data according to xlim and ylim
    temp.xyz <- temp.xyz[x.vals >= xlim[1] & x.vals <= xlim[2],
                         y.vals >= ylim[1] & y.vals <= ylim[2],
                         drop = FALSE]
    if (nrow(temp.xyz) == 0 || ncol(temp.xyz) == 0 || all(is.na(temp.xyz))) {
      .throw_error("No data left after applying 'xlim' and 'ylim'")
    }
  }

  ## wavelength
  x <- as.numeric(rownames(temp.xyz))

  ## time/temp
  y <- as.numeric(colnames(temp.xyz))

  # Background spectrum -------------------------------------------------------------------------
  if(!is.null(bg.spectrum)){
    bg.xyz <- if (inherits(bg.spectrum, "RLum.Data.Spectrum"))
                bg.spectrum@data
              else
                bg.spectrum

    ## set row and column names
    if (is.null(rownames(bg.xyz)))
      rownames(bg.xyz) <- 1:nrow(bg.xyz)
    if (is.null(colnames(bg.xyz)))
      colnames(bg.xyz) <- 1:ncol(bg.xyz)

    ## convert to energy scale if needed
    if (xaxis.energy) {
        #conversion
        bg.xyz <- convert_Wavelength2Energy(cbind(as.numeric(rownames(bg.xyz)), bg.xyz), digits = 5)
        rownames(bg.xyz) <- bg.xyz[,1]
        bg.xyz <- bg.xyz[,-1, drop = FALSE]

        ##modify row order (otherwise subsequent functions, like persp, have a problem)
        bg.xyz <- bg.xyz[order(as.numeric(rownames(bg.xyz))),,drop = FALSE]
    }

    ## reduce for xlim
    bg.xyz <- bg.xyz[as.numeric(rownames(bg.xyz)) >= xlim[1] &
                     as.numeric(rownames(bg.xyz)) <= xlim[2], , drop = FALSE]

    ## reduce for ylim by only if channels is NULL
    if (is.null(bg.channels)) {
      bg.xyz <- bg.xyz[, as.numeric(colnames(bg.xyz)) >= ylim[1] &
                         as.numeric(colnames(bg.xyz)) <= ylim[2], drop = FALSE]
      if (ncol(bg.xyz) == 0) {
        .throw_error("No background channels left after applying 'ylim'")
      }

      bg.channels <- 1:ncol(bg.xyz)
    }
  }

  # Background subtraction ---------------------------------------------------
  if(!is.null(bg.channels)){
    ##set background object if not available
    if(is.null(bg.spectrum)) bg.xyz <- temp.xyz

    if(max(bg.channels) > ncol(bg.xyz) || any(bg.channels <= 0)){
      ##correct the mess
      bg.channels <- sort(unique(bg.channels))
      bg.channels[bg.channels <= 0] <- 1
      bg.channels[bg.channels >= ncol(bg.xyz)] <- ncol(bg.xyz)

      .throw_warning("'bg.channels' out of range, corrected to: ",
                     .format_range(bg.channels))
    }

    ## the challenge we have here is that we want to maintain
    ## a channel-wise subtraction if the spectra are identical.
    if(length(bg.channels) > 1 &&
       (length(bg.channels) != ncol(bg.xyz) ||
       length(bg.channels) != ncol(temp.xyz)))
      temp.xyz <- temp.xyz - rowMeans(bg.xyz[,bg.channels])
    else
      temp.xyz <- temp.xyz - bg.xyz[,bg.channels]

    ##set values < 0 to 0
    temp.xyz[temp.xyz < 0] <- 0

    ##check worst case
    if(sum(temp.xyz) == 0){
      .throw_message("After background subtraction all counts are negative: ",
                     "nothing plotted, NULL returned")
      return(NULL)
    }
  }

  ## Channel binning --------------------------------------------------------
  if(bin.rows > 1){
    temp.xyz <- .matrix_binning(temp.xyz, bin_size = bin.rows, bin_col = FALSE, names = "mean")
    x <- as.numeric(rownames(temp.xyz))

    ##remove last channel (this is the channel that included less data)
    if (length(x) %% bin.rows != 0 && length(x) > bin.rows) {
      .throw_warning(length(x) %% bin.rows,
                     " channels removed due to row (wavelength) binning")
      temp.xyz <- temp.xyz[-length(x),]
      x <- x[-length(x)]
    }
  }

  if(bin.cols > 1){
    temp.xyz <- .matrix_binning(temp.xyz, bin_size = bin.cols, bin_col = TRUE, names = "groups")
    y <- as.numeric(colnames(temp.xyz))

    ##remove last channel (this is the channel that included less data)
    if (length(y) %% bin.cols != 0 && length(y) > bin.cols) {
      .throw_warning(length(y) %% bin.cols,
                     " channels removed due to column (frame) binning")
      temp.xyz <- temp.xyz[,-length(y)]
      y <- y[-length(y)]
    }
  }

  ##limit z-values if requested, this idea was taken from the Diss. by Thomas Schilles, 2002
  if(!is.null(limit_counts[1])) {
    if(min(temp.xyz) > limit_counts[1]) {
      limit_counts <- floor(limit_counts[1] + min(temp.xyz))
      .throw_warning("'limit_counts' is smaller than the lowest count value, ",
                     "'limit_counts' reset to ", limit_counts)
    }

    temp.xyz[temp.xyz[] > max(min(temp.xyz), limit_counts[1])] <- limit_counts[1]
  }

  ## data normalisation -----------------------------------------------------
  if (!isFALSE(norm)) {
    ## this is a bit cumbersome, but we cannot call `normalise_RLum()` directly
    ## on a matrix (and calling `.normalise_curve()` ourselves means that we
    ## would have to duplicate the input validation steps here), see #1303
    temp.spectrum <- set_RLum("RLum.Data.Spectrum", data = temp.xyz)
    temp.xyz <- normalise_RLum(temp.spectrum, norm)@data

    ## check if all values were replaced by 0 because normalisation produced Inf/Nan
    if (sum(range(temp.xyz)) == 0) {
      .throw_message("Insufficient data for plotting, NULL returned")
      return(NULL)
    }
  }

  ## don't apply a logarithmic transformation if non-positive values are present
  if (grepl("z", log) && any(temp.xyz <= 0)) {
    .throw_warning("Data contains non-positive values, set to NA")
    temp.xyz[temp.xyz <= 0] <- NA
  }

  ##check for zlim
  zlim <- extraArgs$zlim %||% range(temp.xyz, na.rm = TRUE)
  .validate_class(zlim, "numeric", length = 2)

  # set colour values --------------------------------------------------------
  if (is.null(extraArgs$col) || plot.type %in% c("single", "multiple.lines")) {
    if (optical.wavelength.colours ||
        (rug && (!plot.type %in% c("persp", "interactive")))) {

      col.labels <- c(violet = "#EE82EE",
                      blue   = "#0000FF",
                      green  = "#00FF00",
                      yellow = "#FFFF00",
                      orange = "#FFA500",
                      red    = "#FF0000",
                      infra  = "#BEBEBE")

      ##make different colour palette for energy values
      if (xaxis.energy) {
        ## as thresholds are descending, we must reverse the labels, as cut()
        ## will sort the cutpoints in increasing order
        col.breaks <- c(min(max(xlim), 4.13), 2.76, 2.52, 2.18, 2.10, 2.00, 1.57, 0)
        col.labels <- rev(col.labels)
      }else{
        ## colour thresholds for wavelength axis
        col.breaks <- c(min(xlim, 300), 450, 495, 570, 590, 620, 790, Inf)
      }

      ## assign colours according to the defined thresholds
      col <- cut(x, col.breaks, labels = col.labels, right = FALSE)

      ## when using labels, cut generates a factor variable
      col <- as.character(col)

      ##if only one colour value, then skip gradient calculation as it causes
      ##an error
      col.unique <- unique(col)
      if(length(col.unique) > 1){
        ##set colour function for replacement
        colfunc <- grDevices::colorRampPalette(col.unique)

        ## index of the first occurrence of each colour besides the first,
        ## as we are trying to find where transition between colours occur
        col.first.idx <- match(col.unique, col)[-1]

        ## size of the colour transition band
        band.size <- 50 / bin.rows

        ##set borders for colour gradient recalculation
        grad.min <- pmax(floor(col.first.idx - band.size), 1)
        grad.max <- ceiling(col.first.idx + band.size)

        ##build up new index sequence (might be better)
        gradient.idx <- unlist(lapply(seq_along(grad.min),
                                      function(j) grad.min[j]:grad.max[j]))

        ##generate colour ramp and replace values
        col[gradient.idx] <- colfunc(length(gradient.idx))

        ##correct for overcharged colour values (causes zebra colour pattern)
        size.diff <- length(col) - nrow(temp.xyz)
        if (size.diff != 0) {
          col <- col[1:(length(col) + size.diff)]
        }
      }

    }else{
      col <- "black"
    }

  }else{
    col <- extraArgs$col
  }

  # Do log scaling if needed -------------------------------------------------
  ##x
  if(grepl("x", log)[1]) x <- log10(x)

  ##y
  if(grepl("y", log)[1]) y <- log10(y)

  ##z
  if(grepl("z", log)[1]) temp.xyz <- log10(temp.xyz)

# PLOT --------------------------------------------------------------------

## set variables we need later
pmat <- NA

if(plot){

  ##par setting for possible combination with plot method for RLum.Analysis objects
  if(par.local) {
    par.default <- par()[c("mfrow", "mar", "mgp", "xpd")]
    on.exit(par(par.default), add = TRUE)
    add.bottom <- if (plot.type %in% c("single", "transect")) 1 else 0
    par(mfrow = c(1, 1), mar = c(3.1 + add.bottom, 3.1, 2, 1),
        mgp = c(2, 0.5, 0), cex = cex)
  }

  ##rest plot type for 1 column matrix
  if(ncol(temp.xyz) == 1 && plot.type != "single"){
    plot.type <- "single"
    .throw_warning("Single column matrix, 'plot.type' reset to 'single'")
  }

  if (nrow(temp.xyz) == 1 && plot.type != "single") {
    .throw_message("Insufficient data for plotting, NULL returned")
    return(NULL)
  }

  if (plot.type == "persp") {

    ## Plot: perspective plot ----
    ## ==========================================================================#
    pmat <- graphics::persp(
      x, y, temp.xyz,
      shade = shade,
      axes = if(box[1] == "alternate") FALSE else axes,
      phi = phi,
      theta = theta,
      ltheta = ltheta,
      lphi = lphi,
      xlab = xlab,
      ylab = ylab,
      zlab = zlab,
      zlim = zlim,
      scale = TRUE,
      col = col[1:(length(x)-1)], ##needed due to recycling of the colours
      main = main,
      expand = expand,
      border = border,
      box = if(box[1] == "alternate") FALSE else box,
      r = r,
      ticktype = ticktype)

    ## this is custom plot output that might come in handy from time to time
    if(axes & box[1] == "alternate") {
      ## add axes manually
      x_axis <- seq(min(x), max(x), length.out = 20)
      y_axis <- seq(min(y), max(y), length.out = 20)
      z_axis <- seq(min(temp.xyz), max(temp.xyz), length.out = 20)

      lines(grDevices::trans3d(x_axis,min(y) - 5, min(temp.xyz),pmat))
      lines(grDevices::trans3d(min(x) - 5,y_axis, min(temp.xyz),pmat))
      lines(grDevices::trans3d(min(x) - 5,max(y), z_axis,pmat))

      ## x-axis
      px_axis <- pretty(x_axis)
      px_axis <- px_axis[(px_axis) > min(x_axis) & px_axis < max(x_axis)]

      tick_start <- grDevices::trans3d(px_axis, min(y_axis), min(z_axis), pmat)
      tick_end <- grDevices::trans3d(
        px_axis, min(y_axis) - max(y_axis) * 0.05, min(z_axis), pmat)

      ## calculate slope angle for xlab and ticks
      m <- (tick_start$y[2] - tick_start$y[1]) / (tick_start$x[2]  - tick_start$x[1])
      m <- atan(m) * 360 / (2 * pi)

      segments(tick_start$x, tick_start$y, tick_end$x, tick_end$y)
      text(
        tick_end$x,
        tick_end$y,
        adj = c(0.5,1.2),
        px_axis,
        xpd = TRUE,
        cex = 0.85,
        srt = m)

      ## x-axis label
      text(
        mean(tick_end$x),
        min(tick_end$y),
        adj = c(0.5, 1),
        xlab,
        srt = m,
        xpd = TRUE)

      ## y-axis
      py_axis <- pretty(y_axis)
      py_axis <- py_axis[(py_axis) > min(y_axis) & py_axis < max(y_axis)]

      tick_start <- grDevices::trans3d(min(x_axis), py_axis, min(z_axis), pmat)
      tick_end <- grDevices::trans3d(
        min(x_axis) - max(x_axis) * 0.025, py_axis, min(z_axis), pmat)
      segments(tick_start$x, tick_start$y, tick_end$x, tick_end$y)

      ## calculate slope angle for xlab and ticks
      m <- (tick_start$y[2] - tick_start$y[1]) / (tick_start$x[2]  - tick_start$x[1])
      m <- atan(m) * 360 / (2 * pi)

      text(
        tick_end$x,
        tick_end$y,
        py_axis,
        adj = c(0.6,1.2),
        srt = m,
        cex = 0.85,
        xpd = TRUE)

      ## y-axis label
      text(
        min(tick_end$x),
        mean(tick_end$y),
        adj = c(0.5, 1),
        ylab,
        srt = m,
        xpd = TRUE)

      ## z-axis
      pz_axis <- pretty(z_axis)
      pz_axis <- pz_axis[(pz_axis) > min(z_axis) & pz_axis < max(z_axis)]

      tick_start <- grDevices::trans3d(min(x_axis), max(y_axis), pz_axis, pmat)
      tick_end <- grDevices::trans3d(
        min(x_axis) - max(x_axis) * 0.015, max(y_axis), pz_axis, pmat)
      segments(tick_start$x, tick_start$y, tick_end$x, tick_end$y)

      ## calculate slope angle for xlab and ticks
      m <- (tick_start$y[2] - tick_start$y[1]) / (tick_start$x[2]  - tick_start$x[1])
      m <- atan(m) * 360 / (2 * pi)

      text(
        tick_end$x,
        tick_end$y,
        format(pz_axis, scientific = TRUE, digits = 1),
        adj = c(0.5,1.2),
        srt = m,
        xpd = TRUE,
        cex = 0.85)

      ## z-axis label
      text(
        min(tick_end$x),
        mean(tick_end$y),
        adj = c(0.5, 2.5),
        zlab,
        srt = m,
        xpd = TRUE)
    }

  } else if (plot.type == "interactive") {
    ## ==========================================================================#
    ##interactive plot and former persp3d
    ## ==========================================================================#

    ## Plot: interactive ----
    .require_suggested_package("plotly", "Displaying interactive plots")

       ##set up plot
        p <- plotly::plot_ly(
          z = temp.xyz,
          x = as.numeric(colnames(temp.xyz)),
          y = as.numeric(rownames(temp.xyz)),
          type = type,
          showscale = showscale
          #colors = col[1:(length(col)-1)],
        )

       ##change graphical parameters
       p <-  plotly::layout(
         p = p,
         scene = list(
           xaxis = list(title = ylab),
           yaxis = list(title = xlab),
           zaxis = list(title = zlab)
         ),
         title = main
       )

       print(p)
       on.exit(return(p), add = TRUE)

  } else if (plot.type == "contour") {
    ## Plot: contour plot ----
    ## ==========================================================================#
    graphics::contour(x,y,temp.xyz,
            xlab = xlab,
            ylab = ylab,
            main = main,
            labcex = labcex * cex,
            col = "black"
    )

  } else if (plot.type == "image") {
    ## Plot: image plot ----
    ## ==========================================================================#

    ## set breaks
    n_breaks <- plot_settings$n_breaks

    ## get colours
    col <- extraArgs$col %||% grDevices::hcl.colors(n_breaks, palette = "Inferno")

    ## set break vector
    breaks <- seq(min(temp.xyz), max(temp.xyz), length.out = length(col) + 1)

    ## render graphic
    graphics::image(
      x,y,temp.xyz,
      xlab = xlab,
      ylab = ylab,
      main = main,
      breaks = breaks,
      col = col
    )

    if (is.null(extraArgs$contour) || !isFALSE(extraArgs$contour)) {
      graphics::contour(x, y, temp.xyz,
              col = extraArgs$contour.col %||% rgb(1, 1, 1, 0.8),
              labcex = labcex * cex,
              add = TRUE)
    }

    if(plot_settings$legend[1]) {
      ## add legend
      legend_scale_id <- seq(1,length(breaks),length.out = c(min(c(length(breaks), 6))))

      legend(
          plot_settings$legend.pos,
          legend = ceiling(breaks[legend_scale_id]/10) * 10,
          fill = col[legend_scale_id],
          bg = grDevices::rgb(1,1,1,0.7),
          title = "Intensity",
          cex = 0.9,
          horiz = plot_settings$legend.horiz)
    }

  } else if(plot.type == "single") {
    ## Plot: single plot ----
    ## ==========================================================================#

    ## set colour rug
    col.rug <- col
    col <- extraArgs$col %||% "black"
    box <- extraArgs$box[1] %||% TRUE
    frames <- extraArgs$frames %||% 1:length(y)
    for(i in frames) {
      if (!"zlim" %in% names(extraArgs))
        zlim <- suppressWarnings(range(temp.xyz[, i], na.rm = TRUE))

      ## we get infinities if all elements in the current frame are NA
      if (any(is.infinite(zlim)))
        next

      plot(x, temp.xyz[,i],
           xlab = xlab,
           ylab = ylab,
           main = main,
           xlim = xlim,
           ylim = zlim,
           frame = box,
           xaxt = "n",
           col = col,
           sub = paste0(
             "(frame ",i, " | ",
             paste(ifelse(i == 1, "0.0", round(y[i - 1], digits = 1)),
                    ":", round(y[i], digits = 1)), ")"),
           type = type,
           pch = pch)

      ## add colour rug
      if(rug){
        par.usr <- par("usr")
          ##rug as continuous rectangle
          i <- floor(seq(1,length(x), length.out = 300))
          graphics::rect(
            xleft = x[i[-length(i)]],
            xright = x[i[-1]],
            ytop = par.usr[3] + diff(c(par.usr[3], min(zlim))) * 0.9,
            ybottom = par.usr[3],
            col = col.rug[i],
            border = NA,
            lwd = 1)

          ## add rectangle from zero to first value
          graphics::rect(
            xleft = par.usr[1],
            xright = x[i[1]],
            ytop = par.usr[3] + diff(c(par.usr[3], min(zlim))) * 0.9,
            ybottom = par.usr[3],
            col = col.rug[1],
            density = 50,
            border = NA,
            lwd = 1)

          ## add rectangle from the last value to end of plot
          graphics::rect(
            xleft = x[i[length(i)]],
            xright = par.usr[2],
            ytop = par.usr[3] + diff(c(par.usr[3], min(zlim))) * 0.9,
            ybottom = par.usr[3],
            col = col.rug[length(col.rug)],
            density = 50,
            border = NA,
            lwd = 1)
      }

      ## add y axis to prevent overplotting
      graphics::axis(side = 1)

      ## add box if needed
      if(box) graphics::box()
    }

  } else if (plot.type == "multiple.lines") {
    ## Plot: multiple.lines ----
    ## ========================================================================#
    col.rug <- col
    col <- extraArgs$col %||% "black"
    box <- extraArgs$box[1] %||% TRUE
    frames <- extraArgs$frames %||% 1:length(y)

    max.frames <- ncol(temp.xyz)
    if (any(frames > max.frames)) {
      .throw_message("Skipped frames exceeding the maximum (", max.frames, ")",
                     error = FALSE)
      frames <- pmin(frames, max.frames)
    }

    ##change graphic settings
    par(mar = c(3.1, 3.1, 2, 7), xpd = TRUE)

    ##grep zlim
    if (!"zlim" %in% names(extraArgs))
      zlim <- range(temp.xyz)

    ##open plot area
    plot(NA, NA,
         xlab = xlab,
         ylab = ylab,
         main = main,
         xlim = xlim,
         ylim = zlim,
         frame = box,
         xaxt = "n",
         sub = sub,
         bty = bty)

    ## add colour rug
    if(rug){
      par.usr <- par("usr")
      ##rug as continuous rectangle
      i <- floor(seq(1,length(x), length.out = 300))
      graphics::rect(
        xleft = x[i[-length(i)]],
        xright = x[i[-1]],
        ytop = par.usr[3] + diff(c(par.usr[3], min(zlim))) * 0.9,
        ybottom = par.usr[3],
        col = col.rug[i],
        border = NA,
        lwd = NA)

      ## add rectangle from zero to first value
      graphics::rect(
        xleft = par.usr[1],
        xright = x[i[1]],
        ytop = par.usr[3] + diff(c(par.usr[3], min(zlim))) * 0.9,
        ybottom = par.usr[3],
        col = col.rug[1],
        density = 50,
        border = NA,
        lwd = 1)

      ## add rectangle from the last value to end of plot
      graphics::rect(
        xleft = x[i[length(i)]],
        xright = par.usr[2],
        ytop = par.usr[3] + diff(c(par.usr[3], min(zlim))) * 0.9,
        ybottom = par.usr[3],
        col = col.rug[length(col.rug)],
        density = 50,
        border = NA,
        lwd = 1)
    }

    ##add lines
    for(i in frames){
      lines(x,
            temp.xyz[,i],
            lty = i,
            lwd = lwd,
            type = type,
            col = col)
    }

    ## add y axis to prevent overplotting
    graphics::axis(side = 1)

    ## add box if needed
    if(box) graphics::box()

    ##for missing values - legend.text
    if (is.null(legend.text))
      legend.text <- as.character(paste(round(y[frames],digits=1), zlab))

    ##legend
    if(plot_settings$legend) {
      legend(x = par()$usr[2],
             y = par()$usr[4],
             legend = legend.text,
             lwd= lwd,
             lty = frames,
             bty = "n",
             cex = 0.6 * cex)
    }

  } else if (plot.type == "transect") {
    ## Plot: transect plot ----
    ## ========================================================================#

    ##sum up rows (column sum)
    temp.xyz <- colSums(temp.xyz, na.rm = TRUE)

    ##consider differences within the arguments
    zlim <- extraArgs$zlim %||% c(0, max(temp.xyz))
    zlab <- extraArgs$ylab %||% "Counts [1/summed channels]"

    plot(y, temp.xyz,
         xlab = ylab,
         ylab = zlab,
         main = main,
         xlim = ylim,
         ylim = zlim,
         col = col,
         sub = paste0("(channel range: ", .format_range(xlim, sep = " : "), ")"),
         type = type,
         pch = pch)
  }

  ## plot additional mtext
  mtext(mtext, side = 3, cex = cex * 0.8)
}

# Return ------------------------------------------------------------------
## add some attributes
attr(temp.xyz, "colour") <- col
attr(temp.xyz, "pmat") <- pmat

  ## return visible or not
  if (plot)
    return(invisible(temp.xyz))
  return(temp.xyz)
}

## load data
data(ExampleData.XSYG, envir = environment())
bg.spectrum <- set_RLum("RLum.Data.Spectrum",
                        data = TL.Spectrum@data[, 15:16, drop = FALSE])

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RLum.Data.Spectrum("error"),
               "'object' should be of class 'RLum.Data.Spectrum' or 'matrix'")
  expect_error(plot_RLum.Data.Spectrum(set_RLum("RLum.Data.Spectrum")),
               "'object' contains no data")
  expect_error(expect_message(plot_RLum.Data.Spectrum(matrix(0, 0, 0)),
                              "Input has been converted"),
               "'object' contains no data")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type = "error"),
               "'plot.type' should be one of 'contour', 'persp', 'single'")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, norm = "error"),
               "'norm' should be one of 'max', 'min', 'first', 'last', 'huot' or")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, bg.spectrum = "error"),
               "'bg.spectrum' should be of class 'RLum.Data.Spectrum', 'matrix' or")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, bin.rows = 1.7),
               "'bin.rows' should be a single positive integer value")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, bin.cols = 0),
               "'bin.cols' should be a single positive integer value")

  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, xlim = 1.2),
               "'xlim' should be of class 'numeric' and have length 2")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, xlim = c(0, 100)),
      "No data left after applying 'xlim' and 'ylim'")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, xlim = c(NA, 1.2)),
               "No data left after applying 'xlim' and 'ylim'")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, ylim = 1.2),
               "'ylim' should be of class 'numeric' and have length 2")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, ylim = c(5, 10)),
      "No data left after applying 'xlim' and 'ylim'")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, ylim = c(NA_real_, NA_real_)),
               "No data left after applying 'xlim' and 'ylim'")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, zlim = 1.2),
               "'zlim' should be of class 'numeric' and have length 2")
  expect_error(plot_RLum.Data.Spectrum(TL.Spectrum, bg.spectrum = bg.spectrum,
                                       ylim = c(0, 100)),
               "No background channels left after applying 'ylim'")

  expect_warning(plot_RLum.Data.Spectrum(TL.Spectrum, bg.channels = -2),
                 "'bg.channels' out of range")
})

test_that("check functionality", {
  testthat::skip_on_cran()

    ##RLum.Data.Spectrum -------
    m <- TL.Spectrum@data

    ##try a matrix as input
    expect_message(plot_RLum.Data.Spectrum(object = m, xaxis.energy = TRUE),
                   "Input has been converted to an 'RLum.Data.Spectrum' object")

    ##remove rownames and column names
    rownames(m) <- NULL
    colnames(m) <- NULL
    expect_message(plot_RLum.Data.Spectrum(object = m),
                   "Input has been converted to an 'RLum.Data.Spectrum' object")

    ## test duplicated column names
    t <- TL.Spectrum
    colnames(t@data) <- rep(50, ncol(t@data))
    expect_warning(plot_RLum.Data.Spectrum(t),
                   "Duplicated column names found")

    ## no plot
    expect_type(plot_RLum.Data.Spectrum(TL.Spectrum, plot = FALSE),
                "double")
    expect_null(plot(TL.Spectrum, plot = FALSE))

    ##test background subtraction ... with bgchannel
    expect_warning(plot_RLum.Data.Spectrum(
      TL.Spectrum,
      plot.type = "persp",
      xlim = c(310, 750),
      ylim = c(0, 300),
      bg.spectrum = bg.spectrum,
      bg.channels = 1,
      plot = FALSE,
      bin.rows = 10,
      bin.cols = 1
    ), "6 channels removed due to row \\(wavelength\\) binning")

    ## plot: multiple.lines ---------
    expect_message(
      plot_RLum.Data.Spectrum(
        TL.Spectrum,
        plot.type = "multiple.lines",
        xlim = c(310, 750),
        frames = c(1, 3, 100),
        ylim = c(0, 50)),
      "Skipped frames exceeding the maximum (3)",
      fixed = TRUE)

    ## plot: transect ------------
    expect_silent(suppressWarnings(
      plot_RLum.Data.Spectrum(
        TL.Spectrum,
        plot.type = "transect",
        xlim = c(310, 750),
        ylim = c(0, 350),
        zlim = c(0, 1e6),
        ylab = "Counts [1 / summed channels]")))

    ## plot: interactive ------------
    expect_silent(suppressWarnings(
      plot_RLum.Data.Spectrum(
        TL.Spectrum,
        plot.type = "interactive",
        xlim = c(310, 750),
        ylim = c(0, 300),
        bin.rows = 10,
        bin.cols = 1
      )
    ))

    ## plot: interactive heatmap --------
    expect_silent(suppressWarnings(
      plot_RLum.Data.Spectrum(
        TL.Spectrum,
        plot.type = "interactive",
        xlim = c(310, 750),
        ylim = c(0, 300),
        bin.rows = 10,
        bin.cols = 1,
        type = "heatmap",
        showscale = TRUE
      )
    ))

    ##interactive contour
    expect_silent(suppressWarnings(
      plot_RLum.Data.Spectrum(
        TL.Spectrum,
        plot.type = "interactive",
        xlim = c(310, 750),
        ylim = c(0, 300),
        bin.rows = 10,
        bin.cols = 1,
        type = "contour",
        norm = "max",
        showscale = TRUE
      )
    ))

  ## normalisation replaces all values with 0
  expect_warning(expect_message(expect_null(
      plot_RLum.Data.Spectrum(TL.Spectrum, plot.type = "persp",
                              bg.spectrum = bg.spectrum, norm = "min")),
      "Error: Insufficient data for plotting, NULL returned"),
      "Curve normalisation produced Inf/NaN values, values replaced by 0")

  ## log transformation with negative data
  suppressWarnings( # surface extends beyond the box
  expect_warning(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type = "persp",
                                         log = "z", norm = "huot"),
                 "Data contains non-positive values, set to NA")
  )

  ## more coverage
  plot_RLum.Data.Spectrum(TL.Spectrum, plot.type = "multiple.lines",
                          phi = 15, theta = -30, r = 10, log = "xyz",
                          shade = 0.4, expand = 0.5, border = 1,
                          optical.wavelength.colours = FALSE, rug = FALSE,
                          axes = FALSE, norm = "min", col = 2, zlim = c(0, 2))

  expect_message(expect_null(
      plot_RLum.Data.Spectrum(TL.Spectrum,
                              plot.type = "multiple.lines",
                              xlim = c(1.4, 4), ylim = c(0, 300),
                              bg.spectrum = bg.spectrum@data * 2,
                              xaxis.energy = TRUE)),
      "Error: After background subtraction all counts are negative")

  spec <- TL.Spectrum
  rownames(spec@data) <- colnames(spec@data) <- NULL
  rownames(bg.spectrum@data) <- NULL
  expect_silent(plot_RLum.Data.Spectrum(spec, bg.spectrum = bg.spectrum,
                                        xlim = c(0, 300), ylim = c(0, 500)))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("contour",
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "contour",
                                                      ylim = c(0, 200),
                                                      cex = 2))
  vdiffr::expect_doppelganger("persp",
                              expect_warning(
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "persp",
                                                      xlim = c(310, 750),
                                                      limit_counts = 10000,
                                                      bg.spectrum = bg.spectrum,
                                                      bin.rows = 10),
                              "6 channels removed due to row \\(wavelength\\) binning"))
  vdiffr::expect_doppelganger("persp alternate",
                              expect_warning(
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "persp",
                                                      xlim = c(410, 700),
                                                      ylim = c(200, 300),
                                                      bg.spectrum = bg.spectrum,
                                                      box = "alternate",
                                                      zlab = "test",
                                                      bin.cols = 2),
                              "1 channels removed due to column \\(frame\\) binning"))
  vdiffr::expect_doppelganger("persp theta phi ticktype",
                              expect_warning(
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "persp",
                                                      xlim = c(410, 750),
                                                      ylim = c(0, 200),
                                                      limit_counts = 20,
                                                      theta = -25,
                                                      phi = 45,
                                                      ticktype = "simple",
                                                      bin.rows = 5),
                              "'limit_counts' is smaller than the lowest count value"))
  vdiffr::expect_doppelganger("image",
                              expect_warning(
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "image",
                                                      xlim = c(310, 750),
                                                      ylim = c(50, 300),
                                                      bin.rows = 20,
                                                      col = grDevices::hcl.colors(20),
                                                      contour.col = "yellow"),
                              "3 channels removed due to row \\(wavelength\\) binning"))
  vdiffr::expect_doppelganger("image no contour",
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "image",
                                                      ylim = c(50, 250),
                                                      contour = FALSE))
  vdiffr::expect_doppelganger("image labcex bottom",
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "image",
                                                      xlim = c(310, 750),
                                                      ylim = c(0, 300),
                                                      bin.rows = 10,
                                                      cex = 2,
                                                      labcex = 1.2,
                                                      n_breaks = 10,
                                                      legend.pos = "bottom",
                                                      legend.horiz = TRUE))
  vdiffr::expect_doppelganger("single",
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "single",
                                                      xlim = c(310, 750),
                                                      ylim = c(0, 300),
                                                      frames = 2,
                                                      bin.cols = 10))
  vdiffr::expect_doppelganger("multiple",
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "multiple.lines",
                                                      xlim = c(1.4, 4),
                                                      ylim = c(0, 300),
                                                      bg.spectrum = bg.spectrum,
                                                      bg.channels = 2,
                                                      bin.cols = 1,
                                                      xaxis.energy = TRUE))
  vdiffr::expect_doppelganger("transect",
                              plot_RLum.Data.Spectrum(TL.Spectrum,
                                                      plot.type = "transect",
                                                      xlim = c(310, 750),
                                                      ylim = c(0, 300),
                                                      bin.rows = 10))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 415
  expect_silent(plot_RLum.Data.Spectrum(
      TL.Spectrum,
      ylim = c(0, 100),
      bin.cols = 3))
  expect_warning(plot_RLum.Data.Spectrum(
      TL.Spectrum,
      ylim = c(0, 100),
      bin.cols = 8),
      "Single column matrix, 'plot.type' reset to 'single'")
  expect_silent(plot_RLum.Data.Spectrum(
      TL.Spectrum,
      bin.rows = 600))
  expect_message(expect_null(plot_RLum.Data.Spectrum(
      TL.Spectrum,
      bin.rows = 2000)),
      "Insufficient data for plotting, NULL returned")

  ## issue 726
  spec <- TL.Spectrum
  rownames(spec@data) <- colnames(spec@data) <- NULL
  bg.spectrum <- set_RLum(class = "RLum.Data.Spectrum",
                          data = spec@data[, 15:16, drop = FALSE])
  expect_silent(plot_RLum.Data.Spectrum(spec, bg.spectrum = bg.spectrum,
                                        xlim = c(0, 100), ylim = c(0, 10)))

  ## issue 1307
  expect_warning(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type = "single",
                                         log = "z", norm = "huot"),
                 "Data contains non-positive values, set to NA")
  expect_warning(plot_RLum.Data.Spectrum(TL.Spectrum, plot.type = "transect",
                                         log = "z", norm = "huot"),
                 "Data contains non-positive values, set to NA")
})

#' @title Analyse SAR CW-OSL Measurements
#'
#' @description
#' The function performs a SAR CW-OSL analysis on an [Luminescence::RLum.Analysis-class]
#' object, including growth curve fitting.
#'
#' @details
#' The function performs an analysis for a standard SAR protocol measurements
#' introduced by Murray and Wintle (2000) with CW-OSL curves. For the
#' calculation of the `Lx/Tx` value the function [Luminescence::calc_OSLLxTxRatio] is
#' used. To **change the way the Lx/Tx error is calculated** use arguments
#' `background.count.distribution` and `sigmab`, which will be passed to
#' [Luminescence::calc_OSLLxTxRatio].
#'
#' **What is part of a SAR sequence?**
#'
#' The function is rather picky when it comes down to accepted curve input
#' (OSL, IRSL,...) and structure. A SAR sequence is basically a set of
#' \eqn{L_{x}/T_{x}} curves. Hence, every second curve is considered a
#' shine-down curve related to the test dose. It also means that the number of
#' curves for \eqn{L_{x}} has to be equal to the number of \eqn{T_{x}} curves,
#' and that hot-bleach curves **do not** belong into a SAR sequence; at least
#' not for the analysis. Other curves allowed and processed are preheat curves,
#' or preheat curves measured as TL, and irradiation curves. The later one
#' indicates the duration of the irradiation, the dose and test dose points,
#' e.g., as part of XSYG files.
#'
#' **Argument `object` is of type `list`**
#'
#' If the argument `object` is of type [list] containing **only**
#' [Luminescence::RLum.Analysis-class] objects, the function re-calls itself on each element
#' in the list. This is useful if to analyse an entire measurement without
#' writing separate for-loops. To gain in full control of the parameters (e.g., `dose.points`) for
#' every aliquot (corresponding to one [Luminescence::RLum.Analysis-class] object in the list), in
#' this case the arguments can be provided as [list]. This `list` should
#' be of similar length as the `list` provided with the argument `object`,
#' otherwise the function will create an own list of the requested length.
#' Function output will be just one single [Luminescence::RLum.Results-class] object.
#'
#' Please be careful when using this option. While it may allow for a fast and
#' efficient data analysis, the function may break with an unclear error
#' message if the input data is misspecified.
#'
#' **Working with IRSL data**
#'
#' The function was originally designed to work just for 'OSL' curves,
#' following the principles of the SAR protocol. An IRSL measurement protocol
#' may follow this procedure, e.g., post-IR IRSL protocol (Thomsen et al.,
#' 2008). Therefore this functions has been enhanced to work with IRSL data,
#' however, the function is only capable of analysing curves that follow the
#' SAR protocol structure, i.e., to analyse a post-IR IRSL protocol, curve data
#' have to be pre-selected by the user to fit the standards of the SAR
#' protocol, i.e., Lx,Tx,Lx,Tx and so on.
#'
#' Example: Imagine the measurement contains `pIRIR50` and `pIRIR225` IRSL
#' curves. Only one curve type can be analysed at the same time: either the
#' `pIRIR50` curves or the `pIRIR225` curves.
#'
#' **Supported rejection criteria**
#'
#' `[recycling.ratio]`: calculated for every repeated regeneration dose point.
#'
#' `[recuperation.rate]`: recuperation rate calculated by comparing the
#' `Lx/Tx` values of the zero regeneration point with the `Ln/Tn` value (the
#' `Lx/Tx` ratio of the natural signal). For methodological background see
#' Aitken and Smith (1988). As a variant, `recuperation_reference` can be
#' specified to select another dose point as reference instead of `Ln/Tn`.
#'
#' `[testdose.error]`: set the allowed error for the test dose, which by
#' default should not exceed 10%. The test dose error is calculated as
#' `Tx_net.error/Tx_net`. The calculation of the \eqn{T_{n}} error is detailed
#' in [Luminescence::calc_OSLLxTxRatio].
#'
#' `[palaeodose.error]`: set the allowed error for the De value, which per
#' default should not exceed 10%.
#'
#' `[sn.ratio]`: set the allowed signal/noise ratio, which by default should
#' be at least 50. By default it uses the value from the natural curve, but
#' this can be changed by specifying the `sn_reference` option.
#'
#' By default, the computed values are compared directly to the corresponding
#' thresholds to establish their result status ("OK" or "FAILED"). By setting
#' the option `consider.uncertainties = TRUE` in the `rejection.criteria`
#' list, quantified uncertainties are considered into the computation of the
#' test value prior before comparing it to the threshold(currently supported
#' only for `recycling.ratio`, `recuperation.rate` and `exceed.max.regpoint`).
#' This reduces tests being marked as "FAILED" when the deviation from the
#' threshold is smaller than the uncertainty margin.
#'
#' **Irradiation times**
#'
#' The function makes two attempts to extra irradiation data (dose points)
#' automatically from the input object, if the argument `dose.points` is not
#' set (aka set to `NULL`).
#'
#' 1. It searches in every curve for an info object called `IRR_TIME`. If this
#' is found, any value set there is taken as dose point.
#'
#' 2. If the object contains curves of type `irradiation`, the function tries
#' to use this information to assign these values to the curves. However, the
#' function does **not** overwrite values preset in `IRR_TIME`.
#'
#' @param object [Luminescence::RLum.Analysis-class] (**required**):
#' input object containing data for analysis, alternatively a [list] of
#' [Luminescence::RLum.Analysis-class] objects can be provided. The object
#' should **only** contain curves considered part of the SAR protocol (see
#' Details). When `OSL.component` is set, the input object must have been
#' processed by `OSLdecomposition::RLum.OSL_decomposition()`.
#'
#' @param signal_integral [integer] (**required**):
#' vector of channels for the signal integral. It can be a [list] of integers,
#' if `object` is a list. If set to `NA`, no integrals are taken into account
#' and their settings are ignored.
#'
#' @param background_integral [integer] (**required**):
#' vector of channels for the background integral. It can be a [list] of
#' integers, if `object` is a list. If set to `NA`, no background integral is
#' subtracted.
#'
#' @param signal_integral_Tx [integer] (*optional*):
#' vector of channels for the signal integral for the `Tx` curve. It can be a
#' [list] of integers, if `object` is a list. If `NULL`, the `signal_integral`
#' vector is used also for the `Tx` curve.
#'
#' @param background_integral_Tx [integer] (*optional*):
#' vector of channels for the background integral for the `Tx` curve. It can
#' be a [list] of integers, if `object` is a list. If `NULL`, the
#' `background_integral` vector is used. If set to `NA`, no background integral
#' for the `Tx` curve is subtracted.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param OSL.component [character] or [integer] (*optional*):
#' single index or a [character] defining the signal component to be evaluated.
#' It requires that the object was processed by `OSLdecomposition::RLum.OSL_decomposition`.
#' This argument can either be the name of the OSL component assigned by
#' `OSLdecomposition::RLum.OSL_global_fitting` or the index in the descending
#' order of decay rates. Then `"1"` selects the fastest decaying component, `"2"`
#' the second fastest and so on. Can be a [list] of [integer]s or strings (or mixed)
#' If object is a [list] and this parameter is provided as [list] it alternates over
#' the elements (aliquots) of the object list, e.g., `list(1,2)` processes the first
#' aliquot with component `1` and the second aliquot with component `2`.
#' `NULL` does not process any component.
#'
#' @param rejection.criteria [list] (*with default*):
#' provide a *named* list and set rejection criteria in **percentage**. It can
#' be a nested [list], if `object` is of type [list].
#' Note: *unnamed* list elements are ignored.
#'
#' Allowed options:
#' * `recycling.ratio` [numeric] (default: `10`)
#' * `recuperation.rate` [numeric] (default: `10`)
#' * `palaeodose.error` [numeric] (default: `10`)
#' * `testdose.error` [numeric] (default: `10`)
#' * `sn.ratio` [numeric] (default: `NA`)
#' * `exceed.max.regpoint` [logical] (default: `FALSE`)
#' * `consider.uncertainties` [logical] (default: `FALSE`)
#' * `recuperation_reference` [character] (default: `"Natural"`; set to, e.g.,
#'   `"R1"` for other point)
#' * `sn_reference` [character] (default: `"Natural"`).
#'
#' Example: `rejection.criteria = list(recycling.ratio = 10)`.
#'
#' All numerical criteria can be set to `NA`, in which case values are
#' calculated, but they are not considered, and their corresponding RC.Status
#' is always `"OK"`. In the "Checks" plot, they are shown with a grey circle
#' and only their value is reported (without showing `<= NA` or `>= NA`).
#'
#' If `onlyLxTxTable = TRUE`, the `palaeodose.error` and `exceed.max.regpoint`
#' criteria are not computed.
#'
#' @param dose.points [numeric] (*optional*):
#' a numeric vector containing the dose points values. Using this argument
#' overwrites dose point values extracted from other data. Can be a [list] of
#' [numeric] vectors, if `object` is of type [list].
#'
#' @param dose.points.test [numeric] (*optional*):
#' a numeric vector containing the test dose in the same units as `dose.points`.
#' If length = 1, the values will be recycled. It has only an effect for
#' `fit.method = 'OTORX'`.
#'
#' @param dose_rate_source [numeric] (*optional*): a numerical value for the source dose rate,
#' typically Gy/s. If set, the x-axis default for the dose-response curve changes to `Dose [Gy]`.
#'
#' @param trim_channels [logical] (*with default*):
#' trim channels per record category to the lowest number of channels in the
#' category by using [Luminescence::trim_RLum.Data]. Applies only to `OSL` and `IRSL` curves.
#' For a more granular control use [Luminescence::trim_RLum.Data] before calling this
#' function.
#'
#' @param mtext.outer [character] (*optional*):
#' option to provide an outer margin `mtext`. Can be a [list] of [character]s,
#' if `object` is of type [list]
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param plot_onePage [logical] (*with default*): enable/disable one page
#' plot output.
#'
#' @param plot_singlePanels [logical] (*with default*) or [numeric] (*optional*):
#' single plot output (`TRUE/FALSE`) to allow for plotting the results in single plot windows.
#' If a [numeric] vector is provided the plots can be selected individually, i.e.
#' `plot_singlePanels = c(1,2,3,4)` will plot the TL and Lx, Tx curves but
#' not the legend (5) or the
#' growth curve (6), (7) and (8) belong to rejection criteria plots. Requires
#' `plot = TRUE`.
#'
#' @param onlyLxTxTable [logical] (*with default*):
#' If `TRUE` the dose response curve fitting and plotting is skipped, and the
#' `palaeodose.error` and `exceed.max.regpoint` criteria are not computed.
#' This allows to get hands on the `Lx/Tx` table for large datasets
#' without the need for a curve fitting.
#'
#' @param method_control [list] (*optional*):
#' options to control the function behaviour. Currently only the
#' 'auto_curve_removal' (logical) option is supported, which controls whether
#' curves with `recordType` starting with `_` should be automatically removed
#' (`TRUE` by default).
#'
#' @param ... further arguments that will be passed to
#' [Luminescence::fit_DoseResponseCurve], [Luminescence::plot_DoseResponseCurve]
#' or [Luminescence::calc_OSLLxTxRatio] (supported:
#' `background.count.distribution`, `sigmab`, `sig0`).

#' **Note:** if you consider to use the early light subtraction method,
#' `sigmab` should be provided.
#'
#' Additionally, supported are `legend.cex` and `legend.pch` to modify the
#' the legend symbols.
#
#' @return
#' A plot (*optional*) and an [Luminescence::RLum.Results-class] object is
#' returned containing the following elements:
#'
#' \item{data}{[data.frame] containing De-values, De-error and further parameters}
#' \item{LnLxTnTx.values}{[data.frame] of all calculated Lx/Tx values including signal,
#' background counts and the dose points}
#' \item{rejection.criteria}{[data.frame] with values that might by used as rejection criteria.
#' `NA` is produced if no R0 dose point exists.}
#' \item{Formula}{[formula] formula that have been used for the growth curve fitting}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' **The function currently does support only 'OSL', 'IRSL' and 'POSL' data!**
#'
#' @section Function version: 0.13.8
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [Luminescence::calc_OSLLxTxRatio], [Luminescence::fit_DoseResponseCurve],
#' [Luminescence::plot_DoseResponseCurve], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class]
#'
#' @references
#' Aitken, M.J. and Smith, B.W., 1988. Optical dating: recuperation
#' after bleaching. Quaternary Science Reviews 7, 387-393.
#'
#' Duller, G., 2003. Distinguishing quartz and feldspar in single grain
#' luminescence measurements. Radiation Measurements 37 (2), 161-165.
#'
#' Murray, A.S. and Wintle, A.G., 2000. Luminescence dating of quartz using an
#' improved single-aliquot regenerative-dose protocol. Radiation Measurements
#' 32, 57-73.
#'
#' Thomsen, K.J., Murray, A.S., Jain, M., Boetter-Jensen, L., 2008. Laboratory
#' fading rates of various luminescence signals from feldspar-rich sediment
#' extracts. Radiation Measurements 43, 1474-1486.
#' doi:10.1016/j.radmeas.2008.06.002
#'
#' @keywords datagen plot
#'
#' @examples
#'
#' ##load data
#' ##ExampleData.BINfileData contains two BINfileData objects
#' ##CWOSL.SAR.Data and TL.SAR.Data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##transform the values from the first position in a RLum.Analysis object
#' object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#'
#' ##perform SAR analysis and set rejection criteria
#' results <- analyse_SAR.CWOSL(
#' object = object,
#' signal_integral = 1:2,
#' background_integral = 900:1000,
#' log = "x",
#' fit.method = "EXP",
#' plot_onePage = TRUE,
#' rejection.criteria = list(
#'   recycling.ratio = 10,
#'   recuperation.rate = 10,
#'   testdose.error = 10,
#'   palaeodose.error = 10,
#'   recuperation_reference = "Natural",
#'   sn.ratio = 50,
#'   sn_reference = "Natural",
#'   exceed.max.regpoint = TRUE)
#')
#'
#' ##show De results
#' get_RLum(results)
#'
#' ##show LnTnLxTx table
#' get_RLum(results, data.object = "LnLxTnTx.table")
#'
#' ## Run example with special case for
#' ## the OTORX fit
#' \dontrun{
#' results <- analyse_SAR.CWOSL(
#'  object = object,
#'  signal_integral = 1:2,
#'  background_integral = 900:1000,
#'  dose.points.test = 15,
#'  n.MC = 10,
#'  fit.method = "OTORX")
#' }
#'
#' @export
analyse_SAR.CWOSL<- function(
  object,
  signal_integral = NA,
  background_integral = NA,
  signal_integral_Tx = NULL,
  background_integral_Tx = NULL,
  integral_input = c("channel", "measurement"),
  OSL.component = NULL,
  rejection.criteria = list(),
  dose.points = NULL,
  dose.points.test = NULL,
  dose_rate_source = NULL,
  trim_channels = FALSE,
  mtext.outer = "",
  plot = TRUE,
  plot_onePage = FALSE,
  plot_singlePanels = FALSE,
  onlyLxTxTable = FALSE,
  method_control = list(),
  ...
) {
  .set_function_name("analyse_SAR.CWOSL")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated arguments
  extraArgs <- list(...)
  if (any(grepl("[signal|background]\\.integral\\.[min|max]", names(extraArgs))) &&
      !is.null(c(extraArgs$signal.integral.min, extraArgs$signal.integral.max,
                 extraArgs$background.integral.min, extraArgs$background.integral.max)) &&
      (!anyNA(signal_integral) || !anyNA(background_integral))) {
    .throw_error("Convert all integral arguments to the new names, ",
                 "'signal_integral' and 'background_integral'")
  }

  ## Self-call --------------------------------------------------------------
  if (inherits(object, "list")) {
  ##clean object input and expand parameters
  object <- .rm_nonRLum(object)
  parm <- .expand_parameters(length(object))

  ##handle main separately
  if ("main" %in% ...names()) {
    main <- .listify(list(...)$main, length = length(object))
  }else{
    main <- as.list(paste0("ALQ #",1:length(object)))
  }

    ## deprecated arguments
    if (any(grepl("[signal|background]\\.integral\\.[min|max]", names(extraArgs)))) {
        extraArgs$signal.integral.min <- .listify(extraArgs$signal.integral.min,
                                                  length(object))
        extraArgs$signal.integral.max <- .listify(extraArgs$signal.integral.max,
                                                  length(object))
        extraArgs$background.integral.min <- .listify(extraArgs$background.integral.min,
                                                      length(object))
        extraArgs$background.integral.max <- .listify(extraArgs$background.integral.max,
                                                      length(object))
    }

    ## remove unnamed rejection criteria
    if (!is.null(parm$rejection.criteria)) {
      parm$rejection.criteria <- lapply(parm$rejection.criteria,
                                        .rm_unnamed_elements)
    }
  results <- .warningCatcher(merge_RLum(lapply(seq_along(object), function(x){
    analyse_SAR.CWOSL(
      object = object[[x]],
      signal_integral = parm$signal_integral[[x]],
      background_integral = parm$background_integral[[x]],
      integral_input = parm$integral_input[[x]],
      OSL.component = parm$OSL.component[[x]],
      dose.points = parm$dose.points[[x]],
      dose.points.test = parm$dose.points.test[[x]],
      dose_rate_source = parm$dose_rate_source[[x]],
      trim_channels = parm$trim_channels[[x]],
      mtext.outer = parm$mtext.outer[[x]],
      plot = parm$plot[[x]],
      rejection.criteria = parm$rejection.criteria[[x]],
      plot_singlePanels = parm$plot_singlePanels[[x]],
      plot_onePage = parm$plot_onePage[[x]],
      onlyLxTxTable = parm$onlyLxTxTable[[x]],
      main = main[[x]],
      method_control = method_control,

      ## deprecated arguments
      signal.integral.min = extraArgs$signal.integral.min[[x]],
      signal.integral.max = extraArgs$signal.integral.max[[x]],
      background.integral.min = extraArgs$background.integral.min[[x]],
      background.integral.max = extraArgs$background.integral.max[[x]],

      ## internal information
      .aliquot_number = sprintf("ALQ: #%d | ", x),

      ...)
  })))

  ##return
  if(length(results) == 0) return(NULL)

  ## add aliquot number
  results@data$data$ALQ <- seq_along(object)[1:nrow(results@data$data)]

  return(results)
}

# CONFIG  -----------------------------------------------------------------

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis")
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  .validate_class(plot_singlePanels, c("logical", "integer", "numeric"))
  .validate_logical_scalar(trim_channels)
  .validate_logical_scalar(plot)
  .validate_logical_scalar(plot_onePage)
  .validate_logical_scalar(onlyLxTxTable)
  .validate_class(rejection.criteria, "list", null.ok = TRUE)
  .validate_scalar(dose_rate_source, null.ok = TRUE)
  .validate_class(method_control, "list")
  method_control <- modifyList(x = list(auto_curve_removal = TRUE),
                               val = method_control)

  ## Protocol integrity checks ----------------------------------------------

  ## trim OSL or IRSL channels
  record.types <- vapply(object@records, function(x) x@recordType, character(1))
  if (trim_channels) {
    ## fetch names with OSL and IRSL
    tmp_names <- unique(record.types)
    tmp_names <- grep("OSL|IRSL", tmp_names, value = TRUE, perl = TRUE)

    ## trim
    object <- trim_RLum.Data(object, recordType = tmp_names)
  }

  ## remove curves starting with _
  if (isTRUE(method_control[["auto_curve_removal"]])) {
    object <- remove_RLum(object, recordType = "^_")
  }

  ## extract the correct curves for the sequence based on allowed curve types
  ## and the most common curve type (after stripping extra specifiers in the
  ## curve names)
  stripped.curve.types <- regmatches(
      x = names(object),
      m = regexpr("(P?OSL[a-zA-Z]*|IRSL[a-zA-Z]*)", names(object), perl = TRUE))

  if (length(stripped.curve.types) == 0) {
    .throw_message("No record of type 'OSL', 'IRSL', 'POSL' detected, ",
                   "NULL returned")
    return(NULL)
  }

  ## now get the type which is used most
  CWcurve.type <- names(which.max(table(stripped.curve.types)))

  ##check overall structure of the object
  ##every SAR protocol has to have equal number of curves

  ## grep curve types from analysis value and remove unwanted information
  temp.ltype <- sapply(1:length(object@records), function(x) {
    ## export as global variable
    object@records[[x]]@recordType <<- gsub(" .*", "", object@records[[x]]@recordType)
    object@records[[x]]@recordType
  })

  ## FI lexsyg devices provide irradiation information in a separate curve
  if (any("irradiation" %in% temp.ltype)) {
    temp.irradiation <- extract_IrradiationTimes(object)@data$irr.times[["IRR_TIME"]]

    ## add this to the records
    for (i in 1:length(object@records)) {
      if (is.null(object@records[[i]]@info$IRR_TIME))
        object@records[[i]]@info <- c(object@records[[i]]@info,
                                      IRR_TIME = temp.irradiation[i])
    }

    ## remove irradiation curves
    object <- get_RLum(object, record.id = !temp.ltype %in% "irradiation",
                       drop = FALSE)
  }

  ## collect error messages to be reported together
  error.list <- list()

  ## check if the wanted curves are a multiple of two
  if (table(temp.ltype)[CWcurve.type] %% 2 != 0) {
    error.list[[1]] <- "Input OSL/IRSL curves are not a multiple of two"
  }

  ## check if the curve lengths differ
  temp.matrix.length <- unlist(lapply(object@records,
                                      function(x) {
                                        if (x@recordType %in% CWcurve.type)
                                          nrow(x@data)
                                      }))

  if (length(unique(temp.matrix.length)) != 1) {
    ## check if the selected curve type (stripped of specifiers) corresponds to
    ## multiple record types, as that may be problematic if they have different
    ## numbers of channels
    matched.types <- unique(grep(CWcurve.type, record.types,
                                 fixed = TRUE, value = TRUE))
    if (length(matched.types) > 1) {
      .throw_warning("Curve type '", CWcurve.type, "' matches multiple record types: ",
                     .collapse(matched.types), ", please ensure that your curve ",
                     "selection is correct")
    }

    hint <- if (trim_channels) "" else ", consider setting 'trim_channels = TRUE'"
    error.list[[2]] <- paste0("Input curves have different lengths (",
                              .collapse(sort(unique(temp.matrix.length)),
                                        quote = FALSE), ")", hint)
  }

  ## return early in case of errors
  if (length(error.list) > 0) {
    .throw_warning(paste(unlist(error.list), collapse = "\n"),
                   "\n... >> nothing was done here!")
    return(NULL)
  }

  ## maximum number of channels available
  channel.length <- temp.matrix.length[1]

  ## Integrals checks -------------------------------------------------------

  ## deprecated arguments
  if (any(grepl("[signal|background]\\.integral\\.[min|max]", names(extraArgs))) &&
      !is.null(c(extraArgs$signal.integral.min, extraArgs$signal.integral.max,
                 extraArgs$background.integral.min, extraArgs$background.integral.max))) {
    .deprecated(old = c("signal.integral.min", "signal.integral.max",
                        "background.integral.min", "background.integral.max"),
                new = c("signal_integral", "background_integral"),
                since = "1.2.0")
    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal.integral.min <- extraArgs$signal.integral.min %||% NA
    signal.integral.max <- extraArgs$signal.integral.max %||% NA
    background.integral.min <- extraArgs$background.integral.min %||% NA
    background.integral.max <- extraArgs$background.integral.max %||% NA
    signal_integral_Tx <- background_integral_Tx <- NULL
    if (anyNA(c(signal.integral.min, signal.integral.max,
                 background.integral.min, background.integral.max))) {
      signal_integral <- background_integral <- NA
    } else {
      signal_integral <- signal.integral.min[1]:signal.integral.max[1]
      background_integral <- background.integral.min[1]:background.integral.max[1]
      if (length(signal.integral.min) == 2 && length(signal.integral.max) == 2)
        signal_integral_Tx <- signal.integral.min[2]:signal.integral.max[2]
      if (length(background.integral.min) == 2 && length(background.integral.max) == 2)
        background_integral_Tx <- background.integral.min[2]:background.integral.max[2]
    }
  }

  if (.strict_na(signal_integral)) {
    signal_integral <- background_integral <- NA
    signal_integral_Tx <- background_integral_Tx <- NULL
    if (is.null(OSL.component)) {
      .throw_warning("No signal or background integral applied as ",
                     "'signal_integral = NA' (and 'OSL.component' was not specified)")
    }
  } else {
    ## convert integrals to channels
    if (integral_input == "measurement") {
      x.range <- get_RLum(object, recordType = CWcurve.type)[[1]][, 1]
      unit <- "time"
      signal_integral <- .convert_to_channels(x.range, signal_integral, unit,
                                              na.ok = TRUE)
      background_integral <- .convert_to_channels(x.range, background_integral, unit,
                                                  na.ok = TRUE)
      signal_integral_Tx <- .convert_to_channels(x.range, signal_integral_Tx, unit,
                                                 null.ok = TRUE)
      background_integral_Tx <- .convert_to_channels(x.range, background_integral_Tx, unit,
                                                     na.ok = TRUE, null.ok = TRUE)
    }

    signal_integral <- .validate_integral(signal_integral, max = channel.length,
                                          na.ok = TRUE)
    background_integral <- .validate_integral(background_integral, na.ok = TRUE,
                                              min = max(signal_integral) + 1,
                                              max = channel.length)

    if (length(background_integral) == 1 && !.strict_na(background_integral)) {
      ## we subtract 25 to avoid warnings from calc_OSLLxTxRatio()
      background_integral <- background_integral - 25:0
      .throw_warning("Background integral should contain at least two values, reset to ",
                     .format_range(background_integral))
    }

    ## signal integrals for the Tx curve
    signal_integral_Tx <- .validate_integral(signal_integral_Tx,
                                             max = channel.length,
                                             null.ok = TRUE)
    if (is.null(signal_integral_Tx) && !is.null(background_integral_Tx)) {
      signal_integral_Tx <- signal_integral
      .throw_warning("'signal_integral_Tx' set automatically to ",
                     .format_range(signal_integral_Tx))
    }

    ## background integrals for the Tx curve
    background_integral_Tx <- .validate_integral(background_integral_Tx,
                                                 min = max(signal_integral_Tx) + 1,
                                                 max = channel.length,
                                                 na.ok = TRUE, null.ok = TRUE)
    if (!is.null(signal_integral_Tx) && is.null(background_integral_Tx)) {
      background_integral_Tx <- background_integral
      .throw_warning("'background_integral_Tx' set automatically to ",
                     if (.strict_na(background_integral_Tx)) NA
                     else .format_range(background_integral_Tx))
    }
    if (length(background_integral_Tx) == 1 && !.strict_na(background_integral_Tx)) {
      background_integral_Tx <- background_integral_Tx - 25:0
      .throw_warning("Background integral limits for Tx curves cannot be equal, reset to ",
                     .format_range(background_integral_Tx))
    }
  }

# Rejection criteria ------------------------------------------------------

  ##set list
  rejection.criteria <- modifyList(x = list(
      recycling.ratio = 10,
      recuperation.rate = 10,
      palaeodose.error = 10,
      testdose.error = 10,
      sn.ratio = NA,
      exceed.max.regpoint = TRUE,
      consider.uncertainties = FALSE,
      sn_reference = "Natural",
      recuperation_reference = "Natural"
    ),
    val = rejection.criteria %||% list(),
    keep.null = TRUE)

  consider.uncertainties <- rejection.criteria$consider.uncertainties
  .validate_logical_scalar(consider.uncertainties,
                           name = "'consider.uncertainties' in 'rejection.criteria'")
  recuperation_reference <- rejection.criteria$recuperation_reference
  .validate_class(recuperation_reference, "character", length = 1,
                  name = "'recuperation_reference' in 'rejection.criteria'")
  sn_reference <- rejection.criteria$sn_reference
  .validate_class(sn_reference, "character", length = 1,
                  name = "'sn_reference' in 'rejection.criteria'")

# Deal with extra arguments ----------------------------------------------------
  ##deal with addition arguments

  verbose <- extraArgs$verbose %||% TRUE
  main <- extraArgs$main %||% ""
  log <- extraArgs$log %||% ""
  cex <- extraArgs$cex %||% 1
  legend.cex <- extraArgs$legend.cex %||% 4
  legend.pch <- extraArgs$legend.pch %||% 20

  background.count.distribution <-
      extraArgs$background.count.distribution %||% "non-poisson"
  sigmab <- extraArgs$sigmab
  sig0 <- extraArgs$sig0 %||% 0

  ## deprecated argument
  if ("plot.single" %in% names(extraArgs)) {
    plot_singlePanels <- extraArgs$plot.single
    .deprecated("plot.single", "plot_singlePanels", since = "1.0.0")
  }

  # Grep Curves -------------------------------------------------------------
  ## extract relevant curves from RLum.Analysis object
  OSL.Curves.ID <- get_RLum(object, recordType = CWcurve.type, get.index = TRUE)

    ##separate curves by Lx and Tx (it makes it much easier)
    OSL.Curves.ID.Lx <-
      OSL.Curves.ID[seq(1,length(OSL.Curves.ID), by = 2)]
    OSL.Curves.ID.Tx <-
      OSL.Curves.ID[seq(2,length(OSL.Curves.ID), by = 2)]

    ##get index of TL curves
    TL.Curves.ID <-
      suppressWarnings(get_RLum(object, recordType = "TL$", get.index = TRUE))

    ##separate TL curves which is always coming before the OSL curve
    ##Note: we do not check anymore whether the sequence makes sense.
    TL.Curves.ID.Lx <- TL.Curves.ID[TL.Curves.ID%in%(OSL.Curves.ID.Lx - 1)]
    TL.Curves.ID.Tx <- TL.Curves.ID[TL.Curves.ID%in%(OSL.Curves.ID.Tx - 1)]

# Calculate LnLxTnTx values  --------------------------------------------------
  ##calculate LxTx values using external function
  if (length(OSL.component) > 0) {
    if (is.null(object@records[[OSL.Curves.ID[1]]]@info$COMPONENTS)) {
      .throw_warning("'object' does not appear to have been processed by ",
                     "OSLdecomposition::RLum.OSL_decomposition()")
    }
    LnLxTnTx <- try(lapply(seq(1, length(OSL.Curves.ID), by = 2), function(x) {
      temp.LnLxTnTx <- get_RLum(
          calc_OSLLxTxDecomposed(
              Lx.data = object@records[[OSL.Curves.ID[x]]]@info$COMPONENTS,
              Tx.data = object@records[[OSL.Curves.ID[x + 1]]]@info$COMPONENTS,
              OSL.component = OSL.component,
              digits = 4,
              sig0 = sig0))
    }), silent = TRUE)
  } else {
    LnLxTnTx <- try(get_RLum(
        calc_OSLLxTxRatio(
            Lx.data = object@records[OSL.Curves.ID.Lx],
            Tx.data = object@records[OSL.Curves.ID.Tx],
            signal_integral = signal_integral,
            signal_integral_Tx = signal_integral_Tx,
            background_integral = background_integral,
            background_integral_Tx = background_integral_Tx,
            background.count.distribution = background.count.distribution,
            sigmab = sigmab,
            sig0 = sig0)
    ), silent = TRUE)
  }

  ## catch errors generated in calc_OSLLxTxDecomposed() or calc_OSLLxTxRatio()
  if (inherits(LnLxTnTx, "try-error")) {
    .throw_message("Failed to generate the LxTx table, NULL returned\n",
                   "The original error was: ",
                   ## return the first part of message coming from get_RLum(),
                   ## as it makes the error too long and confusing
                   gsub("^.*\\[", "[", attr(LnLxTnTx, "condition")$message))
    return(NULL)
  }

  ## extract the dose
  temp.Dose <- lapply(OSL.Curves.ID.Lx, function(x) {
    object@records[[x]]@info$IRR_TIME %||% NA
  })
  LnLxTnTx <- lapply(seq_along(LnLxTnTx), function(x) {
    cbind(Dose = temp.Dose[[x]], LnLxTnTx[[x]])
  })

  ## combine all tables
  LnLxTnTx <- data.table::rbindlist(LnLxTnTx)

  ## Set regeneration points ------------------------------------------------
  ## overwrite dose point manually
  if (length(dose.points) > 0) {
      if (length(dose.points) != length(LnLxTnTx$Dose))
        .throw_error("Length of 'dose.points' (", length(dose.points),
                     ") differs from number of curves (", length(LnLxTnTx$Dose), ")")

      LnLxTnTx$Dose <- dose.points
  }

  ## set test dose points; we set it to -1 if nothing is available
  LnLxTnTx$Test_Dose <- rep_len(dose.points.test %||% -1, nrow(LnLxTnTx))

  ## use source dose rate
  if(!is.null(dose_rate_source)) {
    LnLxTnTx$Dose <- LnLxTnTx$Dose * dose_rate_source
    LnLxTnTx$Test_Dose <- ifelse(LnLxTnTx$Test_Dose < 1, -1, LnLxTnTx$Test_Dose  * dose_rate_source)
  }

  ##check whether we have dose points at all
  if (is.null(dose.points) && anyNA(LnLxTnTx$Dose)) {
    .throw_error("'dose.points' contains NA values or was not set")
  }

  ## check whether the first OSL/IRSL curve (i.e., the Natural) has 0 dose. If
  ## not, it is probably a dose recovery test with the given dose being treated
  ## as the unknown dose. We overwrite this value and warn the user.
  if (LnLxTnTx$Dose[1] != 0 && (list(...)$mode %||% "") != "alternate") {
      .throw_warning("The natural signal has a dose of ", LnLxTnTx$Dose[1],
                     " s, which is indicative of a dose recovery test. ",
                     "The natural dose was set to 0.")
      LnLxTnTx$Dose[1] <- 0
    }

  ## Label dose points ------------------------------------------------------
  dose <- LnLxTnTx$Dose
    ## preset names
    dose_names <- paste0("R", seq_along(dose) - 1)

    ## identify 0 dose point
    zero_id <- which(dose == 0)
    dose_names[zero_id] <- "R0"
    if (length(zero_id)) dose_names[zero_id[1]] <- "Natural"

    ## check for repeated
    is_repeated <- duplicated(dose)
    is_repeated[dose == 0] <- FALSE

    ## add to data.frame
    LnLxTnTx <- cbind(
      data.frame(
        Name = dose_names,
        Repeated = is_repeated,
        stringsAsFactors = FALSE),
      LnLxTnTx)

  ## Calculate rejection criteria -------------------------------------------

  ## compute the standard error of a ratio
  .se.ratio <- function(num, num.err, den, den.err) {
    num / den * sqrt((num.err / num)^2 + (den.err / den)^2)
  }

  ## compare a single value with a threshold (either can be NA)
  .status_from_threshold <- function(value, threshold, comparator = `<=`) {
    if (is.na(threshold) || isTRUE(comparator(value, threshold)))
      return("OK")
    "FAILED"
  }

  ## Calculate Recycling Ratio ----------------------------------------------
  RecyclingRatio <- NA
  if (any(LnLxTnTx$Repeated)) {
      ## get repeated and previous dose points
      repeated <- LnLxTnTx[LnLxTnTx$Repeated, ]
      previous <- data.table::rbindlist(
        lapply(repeated$Dose, \(x) LnLxTnTx[LnLxTnTx$Dose == x & !LnLxTnTx$Repeated, ][1, ]))

    ratio <- repeated$LxTx / previous$LxTx
    if (consider.uncertainties) {
      uncertainty <- .se.ratio(repeated$LxTx, repeated$LxTx.Error,
                               previous$LxTx, previous$LxTx.Error)
      ## add the uncertainty in the most favourable direction
      ratio <- ifelse(ratio > 1, ratio - uncertainty, ratio + uncertainty)
    }

    ## calculate value and set names
    RecyclingRatio <- t(
        setNames(round(ratio, 4),
          nm = paste0("Recycling ratio (", repeated$Name, "/", previous$Name, ")")))
  }

  ## Recycling Ratio
  recycling.threshold <- rep(rejection.criteria$recycling.ratio / 100,
                             length(RecyclingRatio))
  status.RecyclingRatio <- vapply(abs(1 - RecyclingRatio), .status_from_threshold,
                                  threshold = recycling.threshold[1],
                                  FUN.VALUE = character(1))

  if (!is.na(rejection.criteria$recycling.ratio)) {
    ## set better ratio by given the absolute margin depending
    ## on whether we have values larger or smaller than 1
    idx.gt1 <- which(RecyclingRatio > 1)
    recycling.threshold[idx.gt1] <- 1 + recycling.threshold[idx.gt1]
    idx.lt1 <- which(RecyclingRatio < 1)
    recycling.threshold[idx.lt1] <- 1 - recycling.threshold[idx.lt1]
  }

  ## Calculate Recuperation Rate --------------------------------------------
  Recuperation <- NA
  if (!recuperation_reference %in% LnLxTnTx$Name) {
      .throw_error("Recuperation reference invalid, valid values are: ",
                   .collapse(LnLxTnTx[, "Name"]))
  }

  ## Recuperation Rate (capable of handling multiple type of recuperation values)
  if ("R0" %in% LnLxTnTx$Name) {
    idx.R0 <- LnLxTnTx$Name == "R0"
    idx.Rref <- LnLxTnTx$Name == recuperation_reference
    R0 <- LnLxTnTx$LxTx[idx.R0]
    Rref <- LnLxTnTx$LxTx[idx.Rref]
    ratio <- R0 / Rref
    if (consider.uncertainties) {
      uncertainty <- .se.ratio(R0, LnLxTnTx$LxTx.Error[idx.R0],
                               Rref, LnLxTnTx$LxTx.Error[idx.Rref])
      ratio <- ratio - uncertainty
    }
    labels <- paste0("Recuperation rate (", recuperation_reference, ") ",
                     seq_along(R0))
    Recuperation <- t(setNames(ratio, labels))
  }

  recuperation.threshold <- rep(rejection.criteria$recuperation.rate / 100,
                                length(Recuperation))
  status.Recuperation <- sapply(Recuperation, function(value) {
    if (is.na(value))
      return("OK")
    .status_from_threshold(value, rejection.criteria$recuperation.rate / 100)
  })

  ## Calculate Testdose error -----------------------------------------------
  Testdose.error <- (LnLxTnTx$Net_TnTx.Error/LnLxTnTx$Net_TnTx)[1]
  testdose.threshold <- rejection.criteria$testdose.error / 100
  status.Testdose <- .status_from_threshold(Testdose.error, testdose.threshold)

  ## Calculate Signal-to-noise ratio ----------------------------------------
  sn.idx <- match(sn_reference, LnLxTnTx$Name)
  if (is.na(sn.idx)) {
    .throw_error("Signal-to-noise reference invalid, valid values are: ",
                 .collapse(LnLxTnTx[, "Name"]))
  }
  SN.ratio <- LnLxTnTx$SN_RATIO_LnLx[sn.idx]
  SN.threshold <- rejection.criteria$sn.ratio
  status.SN.ratio <- .status_from_threshold(SN.ratio, SN.threshold,
                                            comparator = `>=`)

  RejectionCriteria <- data.frame(
      Criteria = c(colnames(RecyclingRatio) %||% NA_character_,
                   colnames(Recuperation) %||% NA_character_,
                   "Testdose error",
                   "Signal-to-noise ratio"),
      Value = c(RecyclingRatio, Recuperation, Testdose.error, SN.ratio),
      Threshold = c(recycling.threshold, recuperation.threshold,
                    testdose.threshold, SN.threshold),
      Status = c(status.RecyclingRatio, status.Recuperation,
                 status.Testdose, status.SN.ratio),
      stringsAsFactors = FALSE
  )

  ## remove rejection criteria that have a NA value (it can happen for sn.ratio
  ## if OSL.component is used)
  RejectionCriteria <- RejectionCriteria[!is.na(RejectionCriteria$Value), ]

  ## Plotting ---------------------------------------------------------------
  if (plot) {
      ## the graphical parameters cannot be restored unconditionally, as that
      ## may affect the analyse_pIRIRSequence() plots
      if (plot_onePage || !plot_singlePanels[1]) {
        par.default <- .par_defaults()
        on.exit(par(par.default), add = TRUE)
      }

      ##colours and double for plotting
      col <- get("col", pos = .LuminescenceEnv)

      ## get record list
      record_list <- object@records

    layout.matrix <- matrix(c(1, 1, 3, 3, 6, 6, 7,
                              1, 1, 3, 3, 6, 6, 8,
                              2, 2, 4, 4, 9, 9, 10,
                              2, 2, 4, 4, 9, 9, 10,
                              5, 5, 5, 5, 5, 5, 5),
                            nrow = 5, ncol = 7, byrow = TRUE)

    ## Plotting - one Page config -------------------------------------------
    if (plot_onePage) {
      plot_singlePanels <- TRUE
      graphics::layout(layout.matrix)
      par(oma = c(0, 0, 0, 0),
          mar = c(4, 3, 3, 1),
          cex = cex * 0.6)
      }

    ## Plotting - old way config --------------------------------------------
    if (!plot_singlePanels[1]) {
        graphics::layout(layout.matrix[, 1:4])
        par(
          oma = c(0,0,0,0), mar = c(4,4,3,3), cex = cex * 0.6
        )

        ## 1 -> TL previous LnLx
        ## 2 -> LnLx
        ## 3 -> TL previous TnTx
        ## 4 -> TnTx
        ## 5 -> Legend

        ## set selected curves to allow plotting of all curves
        plot.single.sel <- c(1,2,3,4,5,6,7,8)

      } else {
        plot.single.sel <- 1:8

        ## check for values in the single output of the function and convert
        if (!is.logical(plot_singlePanels)) {
          ## this is used when called from analyse_pIRIRSequence()
          par(mar = c(4, 3, 3, 1))
          plot.single.sel <- plot_singlePanels
        }
      }

      ##warning if number of curves exceed colour values
      if (length(col) < length(OSL.Curves.ID) / 2) {
        .throw_warning("Too many curves, only the first ",
                       length(col), " curves are plotted")
      }

      ##get channel resolution (should be equal for all curves)
      resolution.OSLCurves <- round(object@records[[OSL.Curves.ID[1]]]@data[2,1] -
                                      object@records[[OSL.Curves.ID[1]]]@data[1,1],
                                    digits = 2)

      ## (1) Plotting TL Curves previous LnLx ----------------------------------------
      ##overall plot option selection for plot.single.sel
      if (1 %in% plot.single.sel) {
        ##check if TL curves are available
        if (length(TL.Curves.ID.Lx) > 0) {
          ##It is just an approximation taken from the data
          resolution.TLCurves <-  round(mean(diff(
            round(record_list[[TL.Curves.ID.Lx[[1]]]]@data[,1], digits = 1)
          )), digits = 1)

          ## get value ranges of the curves
          xy_xlim <- matrixStats::rowRanges(vapply(
            X = TL.Curves.ID.Lx,
            FUN = \(x) apply(record_list[[x]]@data, 2, range, na.rm = TRUE),
            FUN.VALUE = numeric(4)))

          xlim_range <- c(min(xy_xlim[1,]), max(xy_xlim[2,]))
          ylim_range <- c(1, max(xy_xlim[4,]))

          plot(
            NA,NA,
            xlab = "T [\u00B0C]",
            ylab = paste0("TL [cts/",resolution.TLCurves," \u00B0C]"),
            xlim = xlim_range,
            ylim = ylim_range,
            main = main,
            mgp = c(2, 0.7, 0),
            tcl = -0.4,
            log = gsub("x", "", log))

          #provide curve information as mtext, to keep the space for the header
          mtext(
            side = 3,
            text = expression(paste("TL previous ", L[n],",",L[x]," curves")),
            cex = cex * 0.7)

          ##plot TL curves
          for (i in seq_along(TL.Curves.ID.Lx)) {
            lines(record_list[[TL.Curves.ID.Lx[i]]]@data, col = col[i])
          }

        }else{
          plot(
            NA,NA,xlim = c(0,1), ylim = c(0,1), main = "",
            axes = FALSE,
            ylab = "",
            xlab = "")
          text(0.5,0.5, "No TL curve detected")
        }
      }#plot.single.sel

      ## (2) Plotting LnLx Curves ----------------------------------------------------
      ##overall plot option selection for plot.single.sel
      if (2 %in% plot.single.sel) {
        .plot_ShineDownCurves(
          record_list,
          curve_ids = OSL.Curves.ID.Lx,
          signal_integral = signal_integral,
          background_integral = background_integral,
          set_main = main,
          set_log = log,
          set_cex = cex,
          set_col = col,
          set_mtext = expression(paste(L[n], ", ", L[x], " curves")),
          set_curveType = CWcurve.type,
          set_curveRes = resolution.OSLCurves)

        ##mtext, implemented here, as a plot window has to be called first
        mtext(
          mtext.outer,
          side = 4,
          outer = TRUE,
          line = -1.7,
          cex = cex,
          col = "blue")

      }# plot.single.sel

      ## (3) Plotting TL Curves previous TnTx ----------------------------------------
      ##overall plot option selection for plot.single.sel
      if (3 %in% plot.single.sel) {
        ##check if TL curves are available
        if (length(TL.Curves.ID.Tx) > 0) {
          ##It is just an approximation taken from the data
          resolution.TLCurves <-  round(mean(diff(
            round(record_list[[TL.Curves.ID.Tx[[1]]]]@data[,1], digits = 1)
          )), digits = 1)

          ## get value ranges of the curves
          xy_xlim <- matrixStats::rowRanges(vapply(
            X = TL.Curves.ID.Tx,
            FUN = \(x) apply(record_list[[x]]@data, 2, range, na.rm = TRUE),
            FUN.VALUE = numeric(4)))

          xlim_range <- c(min(xy_xlim[1,]), max(xy_xlim[2,]))
          ylim_range <- c(1, max(xy_xlim[4,]))

          plot(
            NA,NA,
            xlab = "T [\u00B0C]",
            ylab = paste0("TL [cts/",resolution.TLCurves," \u00B0C]"),
            xlim = xlim_range,
            ylim = ylim_range,
            main = main,
            mgp = c(2, 0.7, 0),
            tcl = -0.4,
            log = gsub("x", "", log))

          #provide curve information as mtext, to keep the space for the header
          mtext(
            side = 3,
            text = expression(paste("TL previous ", T[n],",",T[x]," curves")),
            cex = cex * 0.7)

          ##plot TL curves
          for (i in seq_along(TL.Curves.ID.Tx)) {
            lines(record_list[[TL.Curves.ID.Tx[i]]]@data, col = col[i])
          }

        }else{
          plot(
            NA,NA,xlim = c(0,1), ylim = c(0,1), main = "",
            axes = FALSE,
            ylab = "",
            xlab = ""
          )
          text(0.5,0.5, "No TL curve detected")
        }

      }#plot.single.sel

      ## (4) Plotting TnTx Curves ----------------------------------------------------
      ##overall plot option selection for plot.single.sel
      if (4 %in% plot.single.sel) {
        .plot_ShineDownCurves(
          record_list,
          curve_ids = OSL.Curves.ID.Tx,
          signal_integral = signal_integral,
          background_integral = background_integral,
          set_main = main,
          set_log = log,
          set_cex = cex,
          set_col = col,
          set_mtext = expression(paste(T[n], ", ", T[x], " curves")),
          set_curveType = CWcurve.type,
          set_curveRes = resolution.OSLCurves)

      }# plot.single.sel

      ## (5) Plotting Legend ----------------------------------------
      ##overall plot option selection for plot.single.sel
      if (5 %in% plot.single.sel) {
        ## par.old must be assigned before changing the par() values
        ## because `mai` is affected by `mar`, and doing it in one line like
        ## it's done elsewhere would store a modified `mai` value
        par.old <- par("mar", "mai")
        par(mar = c(1,1,1,1), mai = c(0,0,0,0))

        n <- length(OSL.Curves.ID) / 2
        x <- seq_len(n)
        y <- rep(7, n)

        plot(
          x, y,
          axes = FALSE,
          xlab = "", ylab = "",
          pch = legend.pch,
          type = "p",
          col = unique(col[1:length(OSL.Curves.ID)]),
          cex = legend.cex,
          ylim = c(0,10))

        ## add legend text
        text(x, y, paste0(LnLxTnTx$Name, "\n(", round(LnLxTnTx$Dose, 2), ")"),
             offset = 1, pos = 1, xpd = NA)

        ##add line
        abline(h = 10,lwd = 0.5)

        #reset margin
        par(par.old)
      }#plot.single.sel
    }##end plot

  ## (6) Plot Dose-Response Curve --------------------------------------------
  ## overall plot option selection for plot.single.sel
  plot <- plot && 6 %in% plot.single.sel

  ## if we don't compute the dose-response curve, we'll insert empty subplots
  insert.emptyDRCPlots <- onlyLxTxTable

    ## this must be kept in sync with fit_DoseResponseCurve()
    temp.GC.all.na <- data.frame(
        De = NA,
        De.Error = NA,
        D01 = NA,
        D01.ERROR = NA,
        D02 = NA,
        D02.ERROR = NA,
        R = NA,
        R.ERROR = NA,
        Dc = NA,
        D63 = NA,
        n_N = NA,
        De.MC = NA,
        Fit = NA,
        Mode = NA,
        HPDI68_L = NA,
        HPDI68_U = NA,
        HPDI95_L = NA,
        HPDI95_U = NA,
        RC.Status = NA,
        .De.plot = NA,
        .De.raw = NA,
        stringsAsFactors = FALSE)

    ##Fit and plot growth curve
    temp.GC <- temp.GC.all.na
    temp.GC.fit.Formula <- NULL

  ## Calculate Dose-response curve ------------------------------------------
  if (!onlyLxTxTable) {

    ## create data.frame
    temp.sample <- data.frame(
        Dose = LnLxTnTx$Dose,
        LxTx = LnLxTnTx$LxTx,
        LxTx.Error = LnLxTnTx$LxTx.Error,
        TnTx = LnLxTnTx$Net_TnTx,
        Test_Dose = LnLxTnTx$Test_Dose
    )

    ## we want to force fit_DoseResponseCurve to run with verbose = FALSE so
    ## that we can print out the fit message ourselves
    extraArgs$verbose <- NULL
    temp.GC <- do.call(fit_DoseResponseCurve,
                       modifyList(list(object = temp.sample, verbose = FALSE),
                                  extraArgs))
    if (verbose && !is.null(temp.GC)) {
      .throw_message(extraArgs$.aliquot_number, temp.GC@info$fit_message,
                     error = FALSE)
    }

    if (is.null(temp.GC)) {
      temp.GC <- temp.GC.all.na
      temp.GC.fit.Formula <- NA
      insert.emptyDRCPlots <- TRUE
    } else {
          if(plot) {
            do.call(plot_DoseResponseCurve, args = modifyList(
              list(
                object = temp.GC,
                xlab = if(is.null(dose_rate_source)) "Dose [s]" else "Dose [Gy]",
                plot_singlePanels = plot_onePage || length(plot_singlePanels) > 1,
                cex = ifelse(plot_onePage, 0.6, 1)
              ),
              list(...)
            ))
          }

          ##grep information on the fit object
          temp.GC.fit.Formula  <- get_RLum(temp.GC, "Formula")

          ##grep results
          temp.GC <- get_RLum(temp.GC)
          De <- temp.GC$De
          De.err <- temp.GC$De.Error

          # Provide Rejection Criteria for Palaeodose error --------------------------
          palaeodose.error.calculated <- round(De.err / De, digits = 5)
          palaeodose.error.threshold <-
            rejection.criteria$palaeodose.error / 100

          palaeodose.error.data.frame <- data.frame(
            Criteria = "Palaeodose error",
            Value = palaeodose.error.calculated,
            Threshold = palaeodose.error.threshold,
            Status = .status_from_threshold(palaeodose.error.calculated,
                                            palaeodose.error.threshold),
            stringsAsFactors = FALSE
          )

          exceed.max.regpoint.data.frame <- data.frame(
            Criteria = "De > max. dose point",
            Value = De - ifelse(consider.uncertainties, De.err, 0),
            Threshold = if(is.na(rejection.criteria$exceed.max.regpoint)){
                NA
              }else if(!rejection.criteria$exceed.max.regpoint){
                Inf
              }else{
                as.numeric(max(LnLxTnTx$Dose))
              },
            Status = NA_character_)
      exceed.max.regpoint.data.frame$Status <-
        .status_from_threshold(De, exceed.max.regpoint.data.frame$Threshold)

          ##add to RejectionCriteria data.frame
          RejectionCriteria <- rbind(RejectionCriteria,
                                     palaeodose.error.data.frame,
                                     exceed.max.regpoint.data.frame)

          ## add rejection status
          status <- if ("FAILED" %in% RejectionCriteria$Status) "FAILED" else "OK"
          temp.GC <- data.frame(temp.GC,
                                RC.Status = status,
                                stringsAsFactors = FALSE)
      }
  }

  ## insert empty plots, otherwise the ordering may get messed up
  if (plot && insert.emptyDRCPlots) {
    shape::emptyplot()
    if (extraArgs$plot_extended %||% TRUE) {
      shape::emptyplot()
      shape::emptyplot()
    }
  }

  ## add information on the integration limits
  temp.GC.extended <- data.frame(
      signal.range = .format_range(signal_integral),
      background.range = .format_range(background_integral),
      signal.range.Tx = .format_range(signal_integral_Tx %||% NA),
      background.range.Tx = .format_range(background_integral_Tx %||% NA),
      stringsAsFactors = FALSE)

# Set return Values -----------------------------------------------------------
    ##generate unique identifier
    UID <- create_UID()

  ## get position numbers
  POSITION <- unique(unlist(lapply(object@records,
                                   function(x) x@info$POSITION %||% NA)))[1]

  ## get grain numbers
  GRAIN <- unique(unlist(lapply(object@records,
                                function(x) x@info$GRAIN %||% NA)))[1]

    temp.results.final <- set_RLum(
      class = "RLum.Results",
      data = list(
        data = as.data.frame(
          c(temp.GC, temp.GC.extended, ALQ = 1,
            POS = POSITION, GRAIN = GRAIN, UID = UID),
          stringsAsFactors = FALSE),
        LnLxTnTx.table = cbind(LnLxTnTx, UID = UID, stringsAsFactors = FALSE),
        rejection.criteria = cbind(UID, RejectionCriteria),
        Formula = temp.GC.fit.Formula
      ),
      info = list(call = sys.call())
    )

  ## (7) Plot IRSL curve/Single Grain ---------------------------------------
  if (plot[1] && 8 %in% plot.single.sel) {
    ## split the device area in two so that the IRSL curve can be plotted
    ## alongside the rejection criteria
    if (!plot_singlePanels[1]) par(mfrow = c(1,2))

      ## check grain an pos and plot single grain disc marker
      ## if we don't have single grain, we can safely use the other
      ## plot option because this kind of test is almost never
      ## for single grains but on all grains
      if(!is.na(POSITION) & !is.na(GRAIN) && GRAIN > 0) {
        .plot_SGMarker(this_grain = GRAIN, this_pos = POSITION)

      } else {
        ##graphical representation of IR-curve
        temp.IRSL <- suppressWarnings(get_RLum(object, recordType = "IRSL"))
        if(length(temp.IRSL) != 0){
          .validate_class(temp.IRSL, c("RLum.Data.Curve", "list"))
          if (inherits(temp.IRSL, "list")) {
            temp.IRSL <- temp.IRSL[[length(temp.IRSL)]]
            .throw_warning("Multiple IRSL curves detected (IRSL test), only the last one shown")
          }
          plot_RLum.Data.Curve(temp.IRSL, par.local = FALSE,
                               mgp = c(2, 0.7, 0), tcl = -0.4)
        }else{
          plot(1, type = "n", axes = FALSE, xlab = "", ylab = "")
          text(x = c(1,1), y = c(1, 1), labels = "No IRSL curve detected!")
        }
      }
    }

  ## (8) Plot rejection criteria --------------------------------------------
  if (plot && 7 %in% plot.single.sel) {
    .plot_RCCriteria(RejectionCriteria)
  }

  ## Return -----------------------------------------------------------------
  invisible(temp.results.final)
}


# Helper functions -------------------------------------------------------------
## create single grain discs with measured grain labelled
.plot_SGMarker <- function(this_grain = 1, this_pos = 1) {
  ## calculate coordinate matrix
  xy_coord <- matrix(
    data = c(
      rep(seq(0.25,0.75, length.out = 10), 10),
      rep(rev(seq(0.25,0.75, length.out = 10)), each = 10)),
    ncol = 2)

  ##set par
  par.old <- par(mar = c(3, 3, 3, 3))
  on.exit(par(par.old), add = TRUE)

  ## draw disc
  shape::emptyplot(main = "Grain location")

  ## draw super circle
  shape::plotellipse(rx = 0.4, ry = 0.4, mid = c(0.5,0.5), lwd = 1)

  ## draw positing holes
  shape::plotellipse(
    rx = 0.015,
    ry = 0.015,
    mid = c(0.15,0.5),
    lwd = 1)
  shape::plotellipse(
    rx = 0.015,
    ry = 0.015,
    mid = c(0.5,0.85),
    lwd = 1)
  shape::plotellipse(
    rx = 0.015,
    ry = 0.015,
    mid = c(0.85,0.5),
    lwd = 1)

  ## add points
  points(xy_coord, bg = "grey", pch = 21, cex = 1.2)

  ## add the one point
  points(
    x = xy_coord[this_grain,1],
    y = xy_coord[this_grain,2],
    bg = "red", pch = 21, cex = 1.2, col = "darkgreen", lwd = 2)

  ## add text
  mtext(
    side = 3,
    line = -1,
    paste0("pos: #", this_pos, " | ", "grain: #", this_grain),
    cex = 0.7)
}

# create rejection criteria plot
.plot_RCCriteria <- function(x) {
  ##set par
  par.old <- par(mar = c(1, 0, 3.2, 0.35))
  on.exit(par(par.old), add = TRUE)

  ## determine number of criteria
  n <- nrow(x)

  ## calculate middle points for the lines
  y_coord <- seq(0.1,1,length.out = n * 2)

  # set colours
  pch_set <- vapply(1:nrow(x), function(y) {
    c <- if(x[y,"Status"] == "OK") 3 else 2
    c[is.na(x[y,"Threshold"])] <- 24
    s <- if(x[y,"Status"] == "FAILED") 4 else 21

   c(c,s)
  }, numeric(2))

  ## open plot
  plot(NA, NA,
    xlim = c(0,1),
    ylim = c(0,1),
    frame = FALSE,
    main = "Checks",
    yaxt = "n",
    xaxt = "n",
    ylab = "",
    xlab = "")

  ## find how many characters can be fitted in the available space
  usr <- par("usr")
  avail.width <- par("pin")[1] / (usr[2] - usr[1])
  longest.label <- x$Criteria[which.max(nchar(x$Criteria))]
  label.length <- nchar(longest.label)
  repeat {
    if (graphics::strwidth(.shorten_filename(longest.label, label.length),
                           cex = 0.9, units = "in") < avail.width)
      break
    label.length <- label.length - 1
  }

  ## plot names
  text(
    x = 0.88,
    y = y_coord[seq(1,length(y_coord),2)],
    labels = .shorten_filename(x$Criteria, label.length),
    cex = 0.9,
    xpd = NA,
    adj = c(1, 0.5))

  ## add lines with criteria
  y_coord_l <- y_coord[seq(2,length(y_coord),2)]
  for(i in 1:nrow(x)) {
    lines(x = c(0.1,1), y = rep(y_coord_l[i],2), lwd = 0.25)
  }

  ## round to the minimum number of digits so that a difference between value
  ## and threshold can be seen
  digits <- pmax(ceiling(-log10(abs(x$Value - x$Threshold))), 0, na.rm = TRUE)

  ## set labels
  x$Value <- mapply(function(x, d) round(x, digits = d), x$Value, digits)
  x$Threshold <- mapply(function(x, d) round(x, digits = d), x$Threshold, digits)
  x$sign <- ifelse(x$Value > x$Threshold, ">=", "<=")
  x[is.na(x$sign), c("Threshold", "sign")] <- ""

  text(
    x = 0.8,
    y = y_coord_l,
    labels = paste(x$Value, x$sign, x$Threshold),
    cex = 0.7,
    adj = c(1, 1.5))

  ## add final points
  points(
    x = rep(0.95, n),
    pch = pch_set[2,],
    y = y_coord[seq(1,length(y_coord),2)],
    bg = pch_set[1,],
    col = pch_set[1,],
    cex = 1.3)
}

# plot the shine-down curves more consistently
.plot_ShineDownCurves <- function(
    record_list,
    curve_ids,
    signal_integral,
    background_integral,
    set_main,
    set_log,
    set_cex,
    set_col,
    set_mtext,
    set_curveType,
    set_curveRes
) {

  ## if we want to apply a log-transform on x and the first time point
  ## is 0, we shift the curves by one channel
  if (set_log == "x" || set_log == "xy") {
    for(i in curve_ids) {
      x.vals <- record_list[[i]]@data[, 1]
      if (x.vals[1] == 0) {
        record_list[[i]]@data[, 1] <- x.vals + x.vals[2] - x.vals[1]
        .throw_warning("Curves shifted by one channel for log-plot")
      }
    }
  }

  ## get value ranges of the curves
  xy_xlim <- matrixStats::rowRanges(vapply(
    X = curve_ids,
    FUN = \(x) apply(record_list[[x]]@data, 2, range, na.rm = TRUE),
    FUN.VALUE = numeric(4)))

  xlim_range <- c(min(xy_xlim[1,]), max(xy_xlim[2,]))
  ylim_range <- c(min(xy_xlim[3,]), max(xy_xlim[4,]))

  #open plot area LnLx
  plot(
    NA,NA,
    xlab = "Time [s]",
    ylab = paste0(set_curveType," [cts/",set_curveRes," s]"),
    xlim = xlim_range,
    ylim = ylim_range,
    main = set_main,
    mgp = c(2, 0.7, 0),
    tcl = -0.4,
    log = set_log)

  #provide curve information as mtext, to keep the space for the header
  mtext(
    side = 3,
    text = set_mtext,
    cex = set_cex * 0.7)

  for (idx in seq_along(curve_ids)) {
    lines(record_list[[curve_ids[idx]]]@data, col = set_col[idx])
  }

  ##mark integration limit Lx curves
  rec <- record_list[[curve_ids[1]]]@data[, 1]
  abline(v = c(rec[range(signal_integral)],
               rec[range(background_integral)]),
    lty = 2,
    col = "gray")
}

## load data
obj <- set_RLum(class = "RLum.Analysis",
                records = list(
                    set_RLum("RLum.Data.Curve", data = matrix(1:10, ncol = 2)),
                    set_RLum("RLum.Data.Curve", data = matrix(1:20, ncol = 2))))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RLum("error"),
               "[plot_RLum()] 'object' should be of class 'RLum'",
               fixed = TRUE)
})

test_that("empty objects", {
  testthat::skip_on_cran()

  expect_null(plot_RLum(list()))
  expect_message(plot_RLum(set_RLum("RLum.Analysis")),
                 "Nothing to plot, NULL returned")
  expect_silent(plot_RLum(set_RLum("RLum.Data.Curve")))
  expect_silent(plot_RLum(set_RLum("RLum.Data.Image")))
  expect_error(plot_RLum(set_RLum("RLum.Data.Spectrum")),
               "'object' contains no data")
  expect_silent(plot_RLum(set_RLum("RLum.Results")))
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## create dataset to test
  image <- as(array(rnorm(1000), dim = c(10,10,10)), "RLum.Data.Image")
  expect_silent(plot_RLum(image))

  ## check list with different dispatched arguments
  image_short <- as(array(rnorm(100), dim = c(10, 10, 1)), "RLum.Data.Image")
  expect_silent(plot_RLum(list(image_short, image_short), main = list("test1", "test2"), mtext = "test"))

  ## plot results objects
  data(ExampleData.BINfileData, envir = environment())
  object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1:3)
  results <- analyse_SAR.CWOSL(
    object = object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    plot = FALSE,
    verbose = FALSE,
    fit.method = "LIN")
  expect_null(plot_RLum(results))

  ## subset
  expect_silent(plot_RLum(list(obj), main = list("test", "test2"),
                          mtext = "test", subset = NA))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("RLum.Analysis main",
                              plot_RLum(obj, main = list("a", "b")))
  vdiffr::expect_doppelganger("RLum.Analysis list main",
                              plot_RLum(list(obj, obj), main = list("a", "b")))
  vdiffr::expect_doppelganger("RLum.Analysis list mtext sub",
                              plot_RLum(list(obj, obj), mtext = "test",
                                        sub = "sub"))
  })
})

## load data
set.seed(1)
data(ExampleData.BINfileData, envir = environment())
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1:2)
decomposed <- readRDS(test_path("_data/OSLdecomposed_1.2.0.rds"))

##generate different datasets removing TL curves
object_CH_TL <- get_RLum(object, record.id = -seq(1,30,4), drop = FALSE)
object_NO_TL <- get_RLum(object, record.id = -seq(1,30,2), drop = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(analyse_SAR.CWOSL("fail"),
               "'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_SAR.CWOSL(iris),
               "'object' should be of class 'RLum.Analysis'")

  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 integral_input = list()),
               "'integral_input' should be one of 'channel' or 'measurement'")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 plot_singlePanels = list()),
               "'plot_singlePanels' should be of class 'logical', 'integer' or 'numeric'")

  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = NULL,
                                 background_integral = 900:1000),
               "'signal_integral' should be of class 'integer', 'numeric' or NA")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = -9:0,
                                 background_integral = 900:1000),
               "'signal_integral' is of length 0 after removing values smaller than 1")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1.2:3.5,
                                 background_integral = 900:1000),
               "'signal_integral' should be a vector of integers")

  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = NULL),
               "'background_integral' should be of class 'integer', 'numeric' or NA")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 signal_integral_Tx = NA),
               "'signal_integral_Tx' should be of class 'integer', 'numeric' or NULL")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 signal_integral_Tx = 1:2,
                                 background_integral_Tx = "error"),
               "'background_integral_Tx' should be of class 'integer', 'numeric', NA or NULL")

  expect_warning(expect_error(analyse_SAR.CWOSL(object[[1]],
                                                signal_integral = 1:1500,
                                                background_integral = 1900:2000),
                              "expected to be at least 1001, but the maximum allowed is 1000"),
                 "'signal_integral' out of bounds, reset to be between 1 and 1000")

  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 rejection.criteria = "error"),
               "'rejection.criteria' should be of class 'list'")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 rejection.criteria = list(recuperation_reference = letters)),
               "'recuperation_reference' in 'rejection.criteria' should be of class 'character'")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 rejection.criteria = list(sn_reference = letters)),
               "'sn_reference' in 'rejection.criteria' should be of class 'character'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(
    results <- analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      plot = FALSE,
      verbose = FALSE
    ), tolerance = snapshot.tolerance
  )

  ## fit.method
  expect_snapshot_RLum(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
    ), tolerance = snapshot.tolerance
  )

  ## source dose rate
  expect_snapshot_RLum(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      dose_rate_source = 0.2,
      plot = FALSE,
      verbose = FALSE
    ), tolerance = snapshot.tolerance
  )
  ## with test dose
  expect_snapshot_RLum(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      dose.points.test = 1,
      dose_rate_source = 0.2,
      plot = FALSE,
      verbose = FALSE
    ), tolerance = snapshot.tolerance
  )

  ## check rejection criteria and recuperation point selection
  expect_snapshot_RLum(
    t <- analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      rejection.criteria = list(
        recycling.ratio = NA,
        recuperation.rate = 1,
        palaeodose.error = NA,
        testdose.error = 1,
        recuperation_reference = "R1",
        exceed.max.regpoint = FALSE),
      plot = FALSE,
      verbose = FALSE
    ), tolerance = snapshot.tolerance
  )

  ## check if a different point was selected
  expect_equal(round(t$rejection.criteria$Value[2],2), expected = 0.01)

  ## object from OSLdecomposition
  expect_snapshot_RLum(
    analyse_SAR.CWOSL(
        decomposed,
        OSL.component = 1,
        plot = FALSE,
        verbose = FALSE
    ), tolerance = snapshot.tolerance
  )
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## remove position and grain information from the curve data
  object_f <- object[[1]]
  object_f@records <- lapply(object_f@records, function(x){
    x@info$POSITION <- NULL
    x@info$GRAIN <- NULL
    x
  })
  t <- expect_s4_class(
    analyse_SAR.CWOSL(
      object = object_f,
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )

  expect_type(t@data$data$POS, "logical")

  ##signal integral set to NA
  expect_warning(
    analyse_SAR.CWOSL(
      object = object[1],
      signal.integral.min = NA,
      signal.integral.max = NA,
      background.integral.min = NA,
      background.integral.max = NA,
      fit.method = "EXP",
      plot = FALSE,
      verbose = FALSE,
      fit.weights = FALSE
    ),
    "[analyse_SAR.CWOSL()] No signal or background integral applied",
    fixed = TRUE)

  expect_warning(
    analyse_SAR.CWOSL(
      object = object[1],
      signal_integral = NA,
      background_integral = 900:1000,
      fit.method = "EXP",
      plot = FALSE,
      verbose = FALSE,
      fit.weights = FALSE
    ),
    "No signal or background integral applied as 'signal_integral = NA'")

  expect_error(
    analyse_SAR.CWOSL(
      object = object,
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "OTORX",
      plot = FALSE,
      verbose = FALSE,
      fit.weights = FALSE
    ),
    "Column 'Test_Dose' missing but mandatory for 'OTORX' fitting!",
    fixed = TRUE)

  expect_s4_class(
    suppressWarnings(analyse_SAR.CWOSL(
      object = object[1],
      signal.integral.min = NA,
      signal.integral.max = NA,
      background.integral.min = NA,
      background.integral.max = NA,
      fit.method = "EXP",
      plot = FALSE,
      verbose = FALSE
    )),
    class = "RLum.Results"
  )

  SW({
  ##no mix TL and OSL (the only TL will be sorted out automatically)
  only_TL <- set_RLum("RLum.Analysis", records = rep(object_CH_TL[[1]]@records[[2]], length(object_NO_TL[[1]]@records)))
  object_mixed <- c(object_NO_TL, only_TL)

  t <- expect_s4_class(
    analyse_SAR.CWOSL(
      object = object_mixed[2:3],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      log = "x",
      plot = FALSE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )
  expect_equal(nrow(t@data$data), 1)

  ## plot_singlePanels
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "EXP",
      plot = TRUE,
      plot_singlePanels = TRUE
    ),
    class = "RLum.Results"
  )
  })

  ## trigger stop of recuperation reference point
  expect_error(
    analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      rejection.criteria= list(
        recycling.ratio = NA,
        recuperation.rate = 1,
        palaeodose.error = 1,
        testdose.error = 1,
        recuperation_reference = "stop",
        test = "new",
        exceed.max.regpoint = FALSE),
      plot = TRUE
    ),
    "Recuperation reference invalid, valid values are: 'Natural', 'R1', 'R2'")

  expect_error(
    analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      rejection.criteria = list(sn_reference = "error"),
      plot = TRUE
    ),
    "Signal-to-noise reference invalid, valid values are: 'Natural', 'R1', 'R2'")

  ## check stop for OSL.components ... failing
  SW({
  expect_warning(expect_message(expect_null(
      analyse_SAR.CWOSL(
       object = object[[1]],
       signal_integral = 1:2,
       background_integral = 900:1000,
       OSL.component = 1,
       verbose = FALSE
      )),
      "Failed to generate the LxTx table, NULL returned"),
      "'object' does not appear to have been processed by OSLdecomposition")
  })

   expect_error(analyse_SAR.CWOSL(
     object = object[[1]],
     signal_integral = 1:2,
     background_integral = 900:1000,
     dose.points = c(0,1,2),
     fit.method = "LIN",
     plot = FALSE,
     verbose = FALSE),
     "Length of 'dose.points' (3) differs from number of curves (7)",
     fixed = TRUE)

  expect_message(
   expect_null(analyse_SAR.CWOSL(
     object = set_RLum("RLum.Analysis",records = list(set_RLum("RLum.Data.Curve", recordType = "false"))),
     signal_integral = 1:2,
     background_integral = 800:900,
     fit.method = "LIN",
     plot = FALSE,
     verbose = FALSE
   )),
   "No record of type 'OSL', 'IRSL', 'POSL' detected")

  ## check background integral
  expect_warning(analyse_SAR.CWOSL(
      object[[1]],
      signal_integral = 1:500,
      background_integral = 500:1000,
      verbose = FALSE),
      "'background_integral' out of bounds, reset to be between 501 and 1000")

   expect_warning(analyse_SAR.CWOSL(
     object = object[[1]],
     signal_integral = 1:2,
     background_integral = 800:9900,
     fit.method = "LIN",
     plot = FALSE,
     verbose = FALSE
   ), regexp = "'background_integral' out of bounds, reset to be between 800 and 1000")

  expect_warning(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:1,
      background_integral = 1000:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
  ), "Background integral should contain at least two values, reset to 975:1000")

  expect_warning(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      signal_integral_Tx = 1:1,
      background_integral = 900:1000,
      background_integral_Tx = 1000:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
  ), "Background integral limits for Tx curves cannot be equal, reset to 975:1000")

  expect_warning(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 600:900,
      background_integral_Tx = 800:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
  ), "'signal_integral_Tx' set automatically to 1:2")

  expect_warning(expect_warning(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 600:900,
      background_integral_Tx = 1:1200,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
  ), "'signal_integral_Tx' set automatically to 1:2"),
  "'background_integral_Tx' out of bounds, reset to be between 3 and 1000")

  expect_warning(expect_warning(expect_message(
      analyse_SAR.CWOSL(
          object = object[[1]],
          signal_integral = 1:2,
          signal_integral_Tx = 500:2000,
          background_integral = 800:1000,
          fit.method = "LIN",
          plot = FALSE,
          verbose = FALSE
      ), "Failed to generate the LxTx table, NULL returned"),
      "'signal_integral_Tx' out of bounds, reset to be between 500 and 1000"),
      "'background_integral_Tx' set automatically to 800:1000")

  expect_warning(analyse_SAR.CWOSL(
      object[[1]],
      signal_integral = 1:4,
      background_integral = NA,
      signal_integral_Tx = 1:3,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
  ), "'background_integral_Tx' set automatically to NA$")

  ## this generates multiple warnings
  warnings <- capture_warnings(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      signal_integral_Tx = 70:80,
      background_integral = 800:1200,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE))
  expect_match(warnings, all = FALSE,
               "'background_integral' out of bounds, reset to be between 800 and 1000")
  expect_match(warnings, all = FALSE,
               "'background_integral_Tx' set automatically to 800:1000")

  warnings <- capture_warnings(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 100:400,
      signal_integral_Tx = 500:605,
      background_integral = 800:1200,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE))
  expect_match(warnings, all = FALSE,
               "'background_integral' out of bounds, reset to be between 800 and 1000")
  expect_match(warnings, all = FALSE,
               "'background_integral_Tx' set automatically to 800:1000")

  ## add one OSL curve
  expect_warning(expect_null(
      analyse_SAR.CWOSL(
          object = merge(object[[1]], object[[1]][[2]]),
          signal_integral = 1:2,
          background_integral = 800:1200)),
      "Input OSL/IRSL curves are not a multiple of two")

   ## check different curve numbers by shorten one OSL curve
   object_short <- object
   object_short[[1]]@records[[2]]@data <- object_short[[1]]@records[[2]]@data[-nrow(object_short[[1]]@records[[2]]@data),]

   ## without fix
   expect_warning(expect_null(
     analyse_SAR.CWOSL(
       object = object_short[[1]],
       signal_integral = 1:2,
       background_integral = 800:9900,
       fit.method = "LIN",
       plot = FALSE,
       verbose = FALSE)),
     "Input curves have different lengths (999, 1000), consider setting",
     fixed = TRUE)

   ## with new parameter
   testthat::expect_s4_class(
     object = analyse_SAR.CWOSL(
       object = object_short[[1]],
       signal_integral = 1:2,
       background_integral = 800:999,
       fit.method = "LIN",
       trim_channels = TRUE,
       plot = FALSE,
       verbose = FALSE),
   class = "RLum.Results")

  ## different curve types matching
  object_short[[1]]@records[[4]]@recordType <- "OSL (NA)"
  suppressWarnings( # Input curves have different lengths (999, 1000)
  expect_warning(expect_null(
     analyse_SAR.CWOSL(
       object = object_short[[1]],
       signal_integral = 1:2,
       background_integral = 800:9900,
       fit.method = "LIN",
       trim_channels = TRUE,
       plot = FALSE,
       verbose = FALSE)),
     "Curve type 'OSL' matches multiple record types: 'OSL (PMT)', 'OSL (NA)'",
     fixed = TRUE)
  )

  ## automatic curve removal
  data(ExampleData.XSYG, envir = environment())
  expect_s4_class(analyse_SAR.CWOSL(OSL.SARMeasurement$Sequence.Object,
                                    signal_integral = 1:4,
                                    background_integral = 400:500,
                                    plot = FALSE, verbose = FALSE),
                  "RLum.Results")
  expect_message(analyse_SAR.CWOSL(OSL.SARMeasurement$Sequence.Object,
                                   signal_integral = 1:4,
                                   background_integral = 400:500,
                                   method_control = list(auto_curve_removal = FALSE),
                                   plot = FALSE, verbose = FALSE),
                 "Error: Failed to generate the LxTx table, NULL returned")

  ## consider.uncertainties
  res <- analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      rejection.criteria = list(
          recuperation.rate = 1,
          recuperation_reference = "R1",
          recycling.ratio = 4,
          consider.uncertainties = FALSE),
      plot = FALSE, verbose = FALSE)
  expect_equal(res$rejection.criteria$Status[1:2],
               c("FAILED", "FAILED"))
  res <- analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:1000,
      fit.method = "LIN",
      rejection.criteria = list(
          recuperation.rate = 1,
          recuperation_reference = "R1",
          recycling.ratio = 4,
          consider.uncertainties = TRUE),
      plot = FALSE, verbose = FALSE)
  expect_equal(res$rejection.criteria$Status[1:2],
               c("OK", "OK"))
})

test_that("advance tests run", {
  testthat::skip_on_cran()

  ## irradiation information in a separate curve
  no_irr_object <- object
  replace_metadata(no_irr_object[[1]], info_element = "IRR_TIME") <- NULL
  replace_metadata(no_irr_object[[2]], info_element = "IRR_TIME") <- NULL
  obj_irr <- set_RLum(
      "RLum.Analysis",
      protocol = "testthat",
      records = list(
          no_irr_object[[1]][[1]], no_irr_object[[1]][[2]],
          set_RLum(
          "RLum.Data.Curve",
          recordType = "irradiation",
          data = no_irr_object[[1]]@records[[1]]@data,
          info = no_irr_object[[1]]@records[[1]]@info
          ),
          no_irr_object[[2]][[1]], no_irr_object[[2]][[2]]),
      originator = "read_XSYG2R"
  )
  SW({ # repeated message
  expect_message(
      analyse_SAR.CWOSL(list(obj_irr, obj_irr),
                        signal_integral = 1:2,
                        background_integral = 900:1000,
                        fit.method = "LIN", verbose = FALSE),
      "All points have the same dose, NULL returned"
  )
  })

  ## more coverage
  SW({ # Number of background channels for Lx < 25
  expect_warning(expect_warning(
      analyse_SAR.CWOSL(merge(object[[1]], object[[2]],
                              object[[1]], object[[2]]),
                        signal_integral = 1:2,
                        background_integral = 900:900,
                        background.count.distribution = "poisson",
                        log = "xy", verbose = FALSE),
      "Too many curves, only the first 21 curves are plotted"),
      "Multiple IRSL curves detected")

  expect_warning(
      analyse_SAR.CWOSL(object = object[[1]],
                        signal_integral = 1:2,
                        background_integral = 900:1000,
                        plot.single = TRUE,
                        verbose = FALSE),
      "'plot.single' was deprecated in v1.0.0, use 'plot_singlePanels' instead")
  })

  ##test failed recycling ratio
  test_failed <-
    analyse_SAR.CWOSL(
      object = object[1],
      signal_integral = 1:2,
      background_integral = 200:1000,
      fit.method = "LIN",
      rejection.criteria = list(recycling.ratio = 0),
      plot = FALSE,
      verbose = FALSE)
  expect_equal(object = test_failed$data$RC.Status, "FAILED")

  ##test no irradiation times available
  no_irr_object <- object[1]
  replace_metadata(no_irr_object[[1]], info_element = "IRR_TIME") <- NULL
   expect_error(
     analyse_SAR.CWOSL(
      object = no_irr_object,
      signal_integral = 1:2,
      background_integral = 200:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE),
     regexp = "[analyse_SAR.CWOSL()] 'dose.points' contains NA values or was not set",
     fixed = TRUE)

   ##get null for single list element
   unsuitable_type <- object[1]
   unsuitable_type[[1]]@records <- lapply(
     unsuitable_type[[1]]@records, function(x) {
      x@recordType = "Error"
      x
     })

  SW({
  expect_null(
    analyse_SAR.CWOSL(
       object = unsuitable_type,
       signal_integral = 1:2,
       background_integral = 200:1000,
       fit.method = "OTOR",
       n.MC = 10,
       plot = FALSE,
       verbose = FALSE))
  })

  ##set all rejection criteria to NA
  test_failed <-
    analyse_SAR.CWOSL(
      object = object[1],
      signal_integral = 1:2,
      background_integral = 200:1000,
      fit.method = "LIN",
      rejection.criteria = list(
        recycling.ratio = NA,
        recuperation.rate = NA,
        palaeodose.error = NA,
        testdose.error = NA,
        exceed.max.regpoint = NA),
      plot = FALSE,
      verbose = FALSE)
  expect_equal(object = test_failed$data$RC.Status, "OK")

  ##the same test but without a named list >>> OK
  test_ok <-
    analyse_SAR.CWOSL(
      object = object[1],
      signal_integral = 1:2,
      background_integral = 200:1000,
      fit.method = "LIN",
      rejection.criteria = list(1),
      plot = FALSE,
      verbose = FALSE)
  expect_equal(object = test_ok$data$RC.Status, "OK")

  ##test multi parameter settings
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal_integral = list(1:10, 1:20),
      background_integral = 900:1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )

  ##test rejection criteria list in list + test unknown argument
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal_integral = list(1:10, 1:20),
      background_integral = 900:1000,
      rejection.criteria = list(list(recycling.ratio = 0)),
      fit.method = "LIN",
      unknown_argument = "hallo",
      main = "Title",
      plot = TRUE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )

  ## more coverage
  SW({
  expect_warning(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:900,
      background_integral_Tx = 975:975,
      dose.points = rep(2, 7),
      plot = FALSE,
      verbose = FALSE),
      "The natural signal has a dose of 2 s, which is indicative of")

    ## there should be no warning if we are in the alternate mode
    ## however we have to reset the R1 (this is OK!)
    expect_silent(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:975,
      dose.points = seq_len(7),
      mode = "alternate",
      rejection.criteria = list(
        sn_reference = "R0",
        recuperation_reference = "R1"),
      plot = FALSE,
      verbose = FALSE))

  ## OTORX
  expect_s4_class(suppressWarnings(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:900,
      background_integral_Tx = 975:975,
      fit.method = "OTORX",
      dose.points.test = 5,
      plot = FALSE,
      n.MC = 10,
      verbose = FALSE)), class = "RLum.Results")

  object[[1]]@records[[2]][1, 1] <- 0
  expect_warning(analyse_SAR.CWOSL(
      object = object[[1]],
      signal_integral = 1:2,
      background_integral = 900:975,
      log = "x",
      verbose = FALSE),
      "Curves shifted by one channel for log-plot")
  })

  ## simulate single grain
  sg <- get_RLum(object, recordType = "OSL", drop = FALSE)
  replace_metadata(sg[[1]], info_element = "GRAIN") <- 1
  replace_metadata(sg[[2]], info_element = "GRAIN") <- 2
  expect_s4_class(analyse_SAR.CWOSL(
    object = sg,
    signal_integral = 1:2,
    background_integral = 900:975,
    plot_onePage = TRUE,
    verbose = FALSE), "RLum.Results")
})

test_that("integral_input = 'measurement'", {
  testthat::skip_on_cran()

  expect_equal(analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = c(0, 0.08),
                                 background_integral = 36:40,
                                 integral_input = "measurement",
                                 plot = FALSE,
                                 verbose = FALSE)$LnLxTnTx.table[, 1:16], # remove UID
               analyse_SAR.CWOSL(object[[1]],
                                 signal_integral = 1:2,
                                 background_integral = 900:1000,
                                 integral_input = "channel",
                                 plot = FALSE,
                                 verbose = FALSE)$LnLxTnTx.table[, 1:16])

  expect_warning(expect_warning(
      analyse_SAR.CWOSL(
          object[[1]],
          signal_integral = c(0, 1.5),
          background_integral = 900:1000,
          integral_input = "measurement",
          plot = FALSE,
          verbose = FALSE),
      "'background_integral' from time to channels failed: expected values in 0.04:40.00"),
      "Background integral should contain at least two values, reset to 975:1000")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
  vdiffr::expect_doppelganger("default",
                              analyse_SAR.CWOSL(
                                  object = object[[1]],
                                  signal_integral = 1:2,
                                  background_integral = 900:1000,
                                  plot_onePage = TRUE))

    vdiffr::expect_doppelganger("legend_mod",
                                analyse_SAR.CWOSL(
                                  object = object[[1]],
                                  signal_integral = 1:2,
                                  legend.cex = 2,
                                  legend.pch = "R",
                                  background_integral = 900:1000,
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("source_dose_rate",
                                analyse_SAR.CWOSL(
                                  object = object[[1]],
                                  signal_integral = 1:2,
                                  background_integral = 900:1000,
                                  dose_rate_source = 0.22,
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("list-cex",
                              analyse_SAR.CWOSL(
                                  object = list(object[[1]]),
                                  signal_integral = 1:5,
                                  background_integral = 800:1000,
                                  plot_onePage = TRUE, cex = 1.9))

  vdiffr::expect_doppelganger("CH_TL log x",
                              analyse_SAR.CWOSL(
                                  object = object_CH_TL[[1]],
                                  signal_integral = 4:10,
                                  background_integral = 900:1000,
                                  fit.method = "LIN",
                                  log = "x",
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("NO_TL log xy",
                              analyse_SAR.CWOSL(
                                  object = object_NO_TL[[1]],
                                  signal_integral = 1:5,
                                  background_integral = 600:900,
                                  log = "xy",
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("onlyLxTxTable",
                              analyse_SAR.CWOSL(
                                  object = object,
                                  signal_integral = 1:2,
                                  background_integral = 900:1000,
                                  onlyLxTxTable = TRUE,
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("NA rejection criteria",
                              analyse_SAR.CWOSL(
                                  object = object,
                                  signal_integral = 1:2,
                                  background_integral = 900:1000,
                                  rejection.criteria = list(recycling.ratio = NA,
                                                            sn.ratio = NA,
                                                            consider.uncertainties = TRUE),
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("multiple recuperation rates",
                              analyse_SAR.CWOSL(
                                  object = merge(object[[1]], object[[2]]),
                                  signal_integral = 1:2,
                                  background_integral = 900:1000,
                                  plot_onePage = TRUE))

  vdiffr::expect_doppelganger("background_integral NA",
                              analyse_SAR.CWOSL(
                                  object = object,
                                  signal_integral = 1:5,
                                  background_integral = NA,
                                  plot_onePage = TRUE))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 868
  data <- object[[1]]
  data@records[[1]]@recordType <- NA_character_
  expect_s4_class(analyse_SAR.CWOSL(data, verbose = FALSE,
                                    signal_integral = 1:2,
                                    background_integral = 900:1000),
                  "RLum.Results")

  ## issue 1370
  SW({
  expect_message(analyse_SAR.CWOSL(object, signal_integral = 1:4,
                                   background_integral = 100:200,
                                   dose.points = 0:6),
                 "[analyse_SAR.CWOSL()] ALQ: #1 | Fit failed for EXP (interpolation)",
                 fixed = TRUE)
  })
})

test_that("deprecated arguments", {
  testthat::skip_on_cran()

  SW({
  expect_warning(analyse_SAR.CWOSL(object[[1]],
                                   signal.integral.min = 1,
                                   signal.integral.max = 2,
                                   background.integral.min = 900,
                                   background.integral.max = 1000),
                 "were deprecated in v1.2.0, use 'signal_integral'")
  expect_warning(analyse_SAR.CWOSL(object[[1]],
                                   signal.integral.min = 1:2,
                                   signal.integral.max = 2:3,
                                   background.integral.min = c(800, 900),
                                   background.integral.max = c(900, 1000)),
                 "were deprecated in v1.2.0, use 'signal_integral'")
  expect_warning(analyse_SAR.CWOSL(object[1:2],
                                   signal.integral.min = list(1, 1),
                                   signal.integral.max = list(10, 20),
                                   background.integral.min = list(900, 950),
                                   background.integral.max = list(1000, 1000)),
                 "were deprecated in v1.2.0, use 'signal_integral'")
  expect_warning(analyse_SAR.CWOSL(object,
                                   signal.integral.min = 1,
                                   signal.integral.max = 187,
                                   background.integral.min = 900,
                                   background.integral.max = 1000),
                 "were deprecated in v1.2.0, use 'signal_integral'")

  ## mix of deprecated and newly-named arguments is not supported
  expect_error(analyse_SAR.CWOSL(object[1:2],
                                 signal.integral.min = list(1, 1),
                                 signal.integral.max = list(10, 20),
                                 background_integral = 900:1000),
               "Convert all integral arguments to the new names")
  expect_error(analyse_SAR.CWOSL(object[[1]],
                                 signal.integral.min = 1,
                                 signal.integral.max = 2,
                                 background.integral.min = 900,
                                 background.integral.max = 1000,
                                 integral_input = "measurement"),
               "'integral_input' is not supported with old argument names")
  })
})

## load data
data(ExampleData.BINfileData, envir = environment())

## transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

## perform SAR analysis
results <- analyse_SAR.CWOSL(
  object = object,
  signal_integral = 1:2,
  background_integral = 900:1000,
  verbose = FALSE,
  plot = FALSE,
  onlyLxTxTable = TRUE
)

test_that("input validation", {
  testthat::skip_on_cran()

  df <- data.frame(x = c(0, 10, 20), y = c(1.4, 0.7, 2.3),
                   z = c(0.01, 0.02, 0.03))

  expect_error(calc_Lamothe2003(),
               "'object' should be of class 'data.frame' or 'RLum.Results'")

  ##dose_rate.envir
  expect_error(calc_Lamothe2003(df),
               "'dose_rate.envir' should be of class 'numeric'")
  expect_error(calc_Lamothe2003(df, dose_rate.envir = 1),
               "'dose_rate.envir' should contain 2 elements")

  ##dose_rate.source
  expect_error(calc_Lamothe2003(df, dose_rate.envir = c(1, 1)),
               "'dose_rate.source' should be of class 'numeric'")
  expect_error(calc_Lamothe2003(df, dose_rate.envir = c(1, 1),
                                dose_rate.source = 1),
               "'dose_rate.source' should contain 2 elements")

  ##g_value
  expect_error(calc_Lamothe2003(object = df, dose_rate.envir = c(1, 1),
                                dose_rate.source = c(1, 2)),
               "'g_value' should be of class 'numeric'")
  expect_error(calc_Lamothe2003(object = df, dose_rate.envir = c(1, 2),
                                dose_rate.source = c(1, 2), g_value = 1),
               "'g_value' should contain 2 elements")

  ##check warnings
  SW({
  expect_warning(
      expect_s4_class(calc_Lamothe2003(object = df,
                                       dose_rate.envir = c(1, 2, 3),
                                       dose_rate.source = c(1, 2, 3),
                                       g_value = c(1, 1, 1)),
                      "RLum.Results"),
      "taking only the first two entries")
  })

  expect_error(suppressWarnings(
    calc_Lamothe2003(
      object = set_RLum("RLum.Results"),
      dose_rate.envir = c(1, 2, 3),
      dose_rate.source = c(1, 2, 3),
      g_value = c(1, 2)
    )),
    "'object' has an unsupported originator"
  )

  ##tc
  expect_error(
    suppressWarnings(calc_Lamothe2003(
      object = set_RLum("RLum.Results"),
      dose_rate.envir = c(1, 2, 3),
      dose_rate.source = c(1, 2, 3),
      g_value = c(1, 1),
      tc.g_value = 1000
    )),
    "If you set 'tc.g_value' you have to provide a value for 'tc' too"
  )
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.5e-6

  SW({
  ##run fading correction
  expect_snapshot_RLum(calc_Lamothe2003(
    object = results,
    dose_rate.envir =  c(1.676 , 0.180),
    dose_rate.source = c(0.184, 0.003),
    g_value =  c(2.36, 0.6),
    plot = TRUE,
    fit.method = "EXP"),
    tolerance = snapshot.tolerance)

  ##run fading correction
  expect_snapshot_RLum(calc_Lamothe2003(
    object = results,
    dose_rate.envir =  c(1.676 , 0.180),
    dose_rate.source = c(0.184, 0.003),
    g_value =  c(2.36, 0.6),
    tc = 1000,
    tc.g_value = 1200,
    plot = TRUE,
    fit.method = "EXP"),
    tolerance = snapshot.tolerance)
  })
})

test_that("more coverage", {
  ## pretend to have an analyse_pIRIRSequence originator to increase coverage
  results.mod <- results
  results.mod@originator <- "analyse_pIRIRSequence"
  results.mod@data$LnLxTnTx.table$Signal <- 1
  SW({
  expect_s4_class(calc_Lamothe2003(
    object = results.mod,
    dose_rate.envir =  c(1.676 , 0.180),
    dose_rate.source = c(0.184, 0.003),
    g_value =  c(2.36, 0.6),
    plot = FALSE,
    fit.method = "EXP"), class = "RLum.Results")

  ## signal information present
  res <- suppressWarnings(
    calc_Lamothe2003(
      object = data.frame(x = c(0,10,20), y = c(1.4,0.7,2.3),
                          z = c(0.01,0.02, 0.03), Signal=1:3),
      dose_rate.envir = c(1,2), dose_rate.source = c(1,2),
      g_value = c(1,1)))
  })
  expect_equal(res$data$SIGNAL, 1:3)
})

##prepare test file for regression test
set.seed(1)
data(ExampleData.BINfileData, envir = environment())
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1:2)
results <- analyse_SAR.CWOSL(
  object = object[[1]],
  signal.integral.min = 1,
  signal.integral.max = 2,
  background.integral.min = 900,
  background.integral.max = 1000,
  plot = FALSE,
  verbose = FALSE
)


##generate different datasets removing TL curves
object_CH_TL <- get_RLum(object, record.id = -seq(1,30,4), drop = FALSE)
object_NO_TL <- get_RLum(object, record.id = -seq(1,30,2), drop = FALSE)

test_that("tests class elements", {
  testthat::skip_on_cran()
  local_edition(3)

  expect_s4_class(results, "RLum.Results")
  expect_equal(length(results), 4)
  expect_s3_class(results$data, "data.frame")
  expect_s3_class(results$LnLxTnTx.table, "data.frame")
  expect_s3_class(results$rejection.criteria, "data.frame")
  expect_type(results$Formula, "expression")
})

test_that("regression tests De values", {
  testthat::skip_on_cran()
  local_edition(3)

  ##fix for different R versions
  if(R.version$major == "3" && as.numeric(R.version$minor) < 6){
    expect_equal(object = round(sum(results$data[1:2]), digits = 0), 1716)

  }else{
    expect_equal(object = round(sum(results$data[1:2]), digits = 0), 1716)

  }

})

test_that("regression test LxTx table", {
  testthat::skip_on_cran()
  local_edition(3)

  expect_equal(object = round(sum(results$LnLxTnTx.table$LxTx), digits = 5),  20.92051)
  expect_equal(object = round(sum(results$LnLxTnTx.table$LxTx.Error), digits = 2), 0.34)

})

test_that("regression test - check rejection criteria", {
  testthat::skip_on_cran()
  local_edition(3)

  ##fix for different R versions
  if(R.version$major == "3" && as.numeric(R.version$minor) < 6){
    expect_equal(object = round(sum(results$rejection.criteria$Value), digits = 0),  1669)

  }else{
    expect_equal(object = round(sum(results$rejection.criteria$Value), digits = 0),  1669)

  }

})

test_that("simple run", {
  testthat::skip_on_cran()
  local_edition(3)

  ##verbose and plot off
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )

  local_edition(3)
  ##signal integral set to NA
  expect_warning(
    analyse_SAR.CWOSL(
      object = object[1],
      signal.integral.min = NA,
      signal.integral.max = NA,
      background.integral.min = NA,
      background.integral.max = NA,
      fit.method = "EXP",
      plot = FALSE,
      verbose = FALSE,
      fit.weights = FALSE
    ), "\\[analyse_SAR.CWOSL\\(\\)\\] No signal or background integral applied, because they were set to NA\\!")


  expect_s4_class(
    suppressWarnings(analyse_SAR.CWOSL(
      object = object[1],
      signal.integral.min = NA,
      signal.integral.max = NA,
      background.integral.min = NA,
      background.integral.max = NA,
      fit.method = "EXP",
      plot = FALSE,
      verbose = FALSE
    )),
    class = "RLum.Results"
  )

  ##verbose and plot on
  ##full dataset
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[[1]],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "LIN",
      log = "x",
    ),
    class = "RLum.Results"
  )

  ##only CH TL
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object_CH_TL[[1]],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "LIN",
      log = "x",
      plot_onePage = TRUE
    ),
    class = "RLum.Results"
  )

  ##no TL
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object_NO_TL[[1]],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "LIN",
      log = "x",
      plot_onePage = TRUE
    ),
    class = "RLum.Results"
  )

  ##plot single
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[[1]],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "EXP",
      plot = TRUE,
      plot.single = TRUE
    ),
    class = "RLum.Results"
  )

  ##check rejection criteria
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[[1]],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "LIN",
      rejection.criteria= list(
        recycling.ratio = NA,
        recuperation.rate = 1,
        palaeodose.error = 1,
        testdose.error = 1,
        test = "new",
        exceed.max.regpoint = FALSE),
      plot = TRUE,
    ),
    class = "RLum.Results"
  )

   # Trigger stops -----------------------------------------------------------
   ##trigger stops for parameters
   ##object
   expect_error(analyse_SAR.CWOSL(
      object = "fail",
      background.integral.min = 900,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
    ), regexp = "Input object is not of type 'RLum.Analysis'!")

    ##check stop for OSL.components ... failing
    expect_null(analyse_SAR.CWOSL(
       object = object[[1]],
       signal.integral.min = 1,
       signal.integral.max = 2,
       background.integral.min = 900,
       background.integral.max = 1000,
       dose.points = c(0,1,2),
       fit.method = "LIN",
       OSL.component = 1,
       plot = FALSE,
       verbose = FALSE
     ))

   expect_error(analyse_SAR.CWOSL(
     object = object[[1]],
     signal.integral.min = 1,
     signal.integral.max = 2,
     background.integral.min = 900,
     background.integral.max = 1000,
     dose.points = c(0,1,2),
     fit.method = "LIN",
     plot = FALSE,
     verbose = FALSE
   ), regexp = "length 'dose.points' differs from number of curves")


   expect_null(analyse_SAR.CWOSL(
     object = set_RLum("RLum.Analysis", records = list(set_RLum("RLum.Data.Curve", recordType = "false"))),
     signal.integral.min = 1,
     signal.integral.max = 2,
     background.integral.min = 800,
     background.integral.max = 900,
     fit.method = "LIN",
     plot = FALSE,
     verbose = FALSE
   ))

   ##check background integral
   expect_warning(analyse_SAR.CWOSL(
     object = object[[1]],
     signal.integral.min = 1,
     signal.integral.max = 2,
     background.integral.min = 800,
     background.integral.max = 9900,
     fit.method = "LIN",
     plot = FALSE,
     verbose = FALSE
   ), regexp = "Background integral out of bounds")

  local_edition(3)
})

test_that("advance tests run", {
  testthat::skip_on_cran()
  local_edition(3)

  ##this tests basically checks the parameter expansion and make
  ##sure everything is evaluated properly
  # signal.integral.min <- 1
  # signal.integral.max <- 2

  ##test with variables for signal integral
  # expect_s4_class(
  #   analyse_SAR.CWOSL(
  #     object = object[1:2],
  #     signal.integral.min = signal.integral.min,
  #     signal.integral.max = signal.integral.max,
  #     background.integral.min = 900,
  #     background.integral.max = 1000,
  #     fit.method = "LIN",
  #     rejection.criteria = list(
  #       recycling.ratio = NA,
  #       recuperation.rate = 1,
  #       palaeodose.error = 1,
  #       testdose.error = 1,
  #       test = "new",
  #       exceed.max.regpoint = FALSE),
  #     plot = FALSE,
  #     verbose = FALSE
  #   ),
  #   class = "RLum.Results"
  # )

  ##test rejection criteria is a list without(!) names,
  ##this should basically lead to no fail
  test_failed <-
    analyse_SAR.CWOSL(
      object = object[1],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 200,
      background.integral.max = 1000,
      fit.method = "LIN",
      rejection.criteria = list(recycling.ratio = 0),
      plot = FALSE,
      verbose = FALSE)
  expect_equal(object = test_failed$data$RC.Status, "FAILED")

  ##the same test but without a named list >>> OK
  test_ok <-
    analyse_SAR.CWOSL(
      object = object[1],
      signal.integral.min = 1,
      signal.integral.max = 2,
      background.integral.min = 200,
      background.integral.max = 1000,
      fit.method = "LIN",
      rejection.criteria = list(1),
      plot = FALSE,
      verbose = FALSE)

  expect_equal(object = test_ok$data$RC.Status, "OK")

  ##test multi parameter settings
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal.integral.min = 1,
      signal.integral.max = list(10,20),
      background.integral.min = 900,
      background.integral.max = 1000,
      fit.method = "LIN",
      plot = FALSE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )

  ##test rejection criteria list in list + test unknown argument
  expect_s4_class(
    analyse_SAR.CWOSL(
      object = object[1:2],
      signal.integral.min = 1,
      signal.integral.max = list(10,20),
      background.integral.min = 900,
      background.integral.max = 1000,
      rejection.criteria = list(list(recycling.ratio = 0)),
      fit.method = "LIN",
      unknown_argument = "hallo",
      plot = TRUE,
      verbose = FALSE
    ),
    class = "RLum.Results"
  )

})

#' @title Convert Risoe.BINfileData object to an RLum.Analysis object
#'
#' @description
#' Converts values from one specific position of a [Luminescence::Risoe.BINfileData-class]
#' object to an [Luminescence::RLum.Analysis-class] object.
#'
#' The [Luminescence::RLum.Analysis-class] object requires a set of curves for
#' specific further protocol analyses. However, the [Luminescence::Risoe.BINfileData-class]
#' usually contains a set of curves for different aliquots and different
#' protocol types that may be mixed up. Therefore, a conversion is needed.
#'
#' @param object [Luminescence::Risoe.BINfileData-class] (**required**):
#' object to convert.
#'
#' @param pos [numeric] (*optional*): position number of the `Risoe.BINfileData`
#' object for which the curves are stored in the [Luminescence::RLum.Analysis-class] object.
#' If `length(pos) > 1`, a list of `RLum.Analysis` objects is returned.
#' If nothing is provided every position will be converted.
#' If the position is not valid `NULL` is returned.
#'
#' @param grain [vector], [numeric] (*optional*):
#' grain number from the measurement to limit the converted data set
#' (e.g., `grain = c(1:48)`). Please be aware that this option may lead to
#' unwanted effects, as the output is strictly limited to the chosen grain
#' number for all position numbers.
#'
#' @param run [vector], [numeric] (*optional*):
#' run number from the measurement to limit the converted data set
#' (e.g., `run = c(1:48)`).
#'
#' @param set [vector], [numeric] (*optional*):
#' set number from the measurement to limit the converted data set
#' (e.g., `set = c(1:48)`).
#'
#' @param ltype [vector], [character] (*optional*):
#' curve type to limit the converted data. Commonly allowed values are:
#' `IRSL`, `OSL`, `TL`, `RIR`, `RBR` and `USER`
#' (see also [Luminescence::Risoe.BINfileData-class]).
#'
#' @param dtype [vector], [character] (*optional*):
#' data type to limit the converted data. Commonly allowed values are
#' listed in [Luminescence::Risoe.BINfileData-class].
#'
#' @param protocol [character] (*optional*):
#' sets protocol type for analysis object. Value may be used by subsequent
#' analysis functions.
#'
#' @param keep.empty [logical] (*with default*):
#' If `TRUE` (default) an `RLum.Analysis` object is returned even if it does
#' not contain any records. Set to `FALSE` to discard all empty objects.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar.
#'
#' @return Returns an [Luminescence::RLum.Analysis-class] object.
#'
#' @note
#' The `protocol` argument of the [Luminescence::RLum.Analysis-class]
#' object is set to 'unknown' if not stated otherwise.
#'
#' @section Function version: 0.4.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::read_BIN2R]
#'
#' @keywords manip
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##convert values for position 1
#' Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
#'
#' @export
Risoe.BINfileData2RLum.Analysis<- function(
  object,
  pos = NULL,
  grain = NULL,
  run = NULL,
  set = NULL,
  ltype = NULL,
  dtype = NULL,
  protocol = "unknown",
  keep.empty = TRUE,
  txtProgressBar = FALSE
) {
  .set_function_name("Risoe.BINfileData2RLum.Analysis")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "Risoe.BINfileData")
  .validate_class(pos, c("numeric", "integer"), null.ok = TRUE)

  ## check that all required columns are available
  sel.cols <- c("ID", "POSITION", "GRAIN", "RUN", "SET", "LTYPE", "DTYPE")
  if (!all(sel.cols %in% colnames(object@METADATA))) {
    .throw_error("'object' has missing columns in METADATA: ",
                 .collapse(setdiff(sel.cols, colnames(object@METADATA))))
  }

  positions.valid <- unique(object@METADATA[["POSITION"]])
  if (is.null(pos)) {
    pos <- positions.valid
  } else if (length(setdiff(pos, positions.valid)) > 0) {
    ## remove invalid positions from the input
      .throw_warning("Invalid position number skipped: ",
                     .collapse(setdiff(pos, positions.valid), quote = FALSE))
      pos <- intersect(pos, positions.valid)
  }

  # Grep run and set data ---------------------------------------------------

  ## grain
  grain.valid <- unique(object@METADATA[["GRAIN"]])
  if (is.null(grain)) {
    grain <- grain.valid
  } else if (length(setdiff(grain, grain.valid)) > 0) {
        .throw_warning("Invalid grain number skipped: ",
                       .collapse(setdiff(grain, grain.valid), quote = FALSE))
        grain <- intersect(grain, grain.valid)
  }

  ## run
  run.valid <- unique(object@METADATA[["RUN"]])
  if (is.null(run)) {
    run <- run.valid
  } else if (length(setdiff(run, run.valid)) > 0) {
      .throw_error("'run' contains invalid runs, valid runs are: ",
                   .collapse(run.valid, quote = FALSE))
  }

  ## set
  set.valid <- unique(object@METADATA[["SET"]])
  if (is.null(set)) {
    set <- set.valid
  } else if (length(setdiff(set, set.valid)) > 0) {
      .throw_error("'set' contains invalid sets, valid sets are: ",
                   .collapse(set.valid, quote = FALSE))
  }

  ## ltype
  ltype.valid <- unique(object@METADATA[["LTYPE"]])
  if (is.null(ltype)) {
    ltype <- ltype.valid
  } else if (length(setdiff(ltype, ltype.valid)) > 0) {
      .throw_error("'ltype' contains invalid ltypes, valid ltypes are: ",
                   .collapse(ltype.valid, quote = TRUE))
  }

  ## dtype
  dtype.valid <- unique(object@METADATA[["DTYPE"]])
  if (is.null(dtype)) {
    dtype <- dtype.valid
  } else if (length(setdiff(dtype, dtype.valid)) > 0) {
      .throw_error("'dtype' contains invalid dtypes, valid dtypes are: ",
                   .collapse(dtype.valid, quote = TRUE))
  }

    # Select values and convert them-----------------------------------------------------------
    ##set progressbar to false if only one position is provided
    if(txtProgressBar & length(pos)<2){
      txtProgressBar <- FALSE
    }

    ##This loop does:
    ## (a) iterating over all possible positions
    ## (b) consider grains in all possible positions
    ## (c) consider other selections
    ## (d) create the RLum.Analysis objects

    ##set progress bar
    if(txtProgressBar){
      pb <- txtProgressBar(min=min(pos),max=max(pos), char="=", style=3)
    }

  ## convert metadata object to data.table
  metadata.dt <- as.data.table(object@METADATA[, sel.cols])

  ## silence notes raised by R CMD check
  POSITION <- GRAIN <- RUN <- SET <- LTYPE <- DTYPE <- ID <- NULL

  object <- lapply(pos, function(pos) {

      ##update progress bar
      if(txtProgressBar){
        setTxtProgressBar(pb, value = pos)
      }

    ## loop over the grains and produce RLum.Analysis objects
    object <- lapply(grain, function(grain) {

        ## select data
        ## the NA check for grain is necessary as FI readers like to report
        ## NA instead of 0 in that column, and this causes some trouble
        temp_id <- metadata.dt[POSITION == pos &
                               (is.na(GRAIN) | GRAIN == grain) &
                               RUN %in% run & LTYPE %in% ltype &
                               SET %in% set & DTYPE %in% dtype,
                               ID]

        ## if the input object is empty, bypass the creation of curve objects
        if (length(object@DATA) == 0) {
          message("Empty Risoe.BINfileData object detected")
          records <- list()
        } else {
          ## create curve object
          records <- lapply(temp_id, function(x) {
            ## skip ROI information
            if (!is.null(object@METADATA[["RECTYPE"]]) &&
                object@METADATA[["RECTYPE"]][x] == 128)
              set_RLum(class = "RLum.Data.Curve")
            else
              .Risoe.BINfileData2RLum.Data.Curve(object, id = x)
          })
        }

        if (!keep.empty && length(records) == 0)
          return(NULL)

        ## create the RLum.Analysis object
        object <- set_RLum(
          class = "RLum.Analysis",
          records = records,
          protocol = protocol,
          originator = "Risoe.BINfileData2RLum.Analysis"
        )

        ## add unique id of RLum.Analysis object to each curve
        .set_pid(object)
        return(object)
      })

      return(object)
    })

  ## if only one element is included, output an RLum.Analysis object, not a list
  if (length(object) == 1 && length(object[[1]]) == 1)
    return(invisible(object[[1]][[1]]))

  return(invisible(unlist(object)))
}

##Full check
test_that("Test examples", {
  skip_on_cran()
  local_edition(3)

  ##load data
  data(ExampleData.BINfileData, envir = environment())

  ##transform the values from the first position in a RLum.Analysis object
  object <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos=3)

  ##perform analysis
  expect_s4_class(
    analyse_SAR.TL(
      object,
      signal.integral.min = 210,
      signal.integral.max = 220,
      fit.method = "EXP OR LIN",
      sequence.structure = c("SIGNAL", "BACKGROUND")
    ),
    "RLum.Results"
  )

})

## load data
data(ExampleData.BINfileData, envir = environment())
temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(trim_RLum.Data("error"),
               "[trim_RLum.Data()] 'object' should be of class 'RLum.Data' or",
               fixed = TRUE)
  expect_error(trim_RLum.Data(list(list())),
               "All elements of 'object' should be of class 'RLum.Data' or")
  expect_error(trim_RLum.Data(list(list(3))),
               "All elements of 'object' should be of class 'RLum.Data' or")
  expect_error(trim_RLum.Data(temp, recordType = c(1, 20)),
               "'recordType' should be of class 'character'")
  expect_error(trim_RLum.Data(temp, trim_range = "error"),
               "'trim_range' should be of class 'integer' or 'numeric'")
})

test_that("RLum.Data.Curve", {
  testthat::skip_on_cran()

  ## trim with range
  t <- testthat::expect_type(
    object = trim_RLum.Data(temp$TL, trim_range = c(20,50)),
    type = "list")
  testthat::expect_length(
    object = t[[1]]@data[,1], n = 31)

  ## trim maximum only
  t <- testthat::expect_type(
    object = trim_RLum.Data(temp$TL, trim_range = c(50)),
    type = "list")
  testthat::expect_length(
    object = t[[1]]@data[,1], n = 50)

  ## trim with nothing
  t <- testthat::expect_type(
    object = trim_RLum.Data(temp$TL, trim_range = NULL),
    type = "list")
  testthat::expect_length(
    object = t[[1]]@data[,1], n = 250)

  ## trim with wrong curve type ... this should do nothing
  ## because the object is just parsed through
  ## on top we use a single object instead of list only
  t <- testthat::expect_s4_class(
    object = trim_RLum.Data(temp@records[[1]], recordType = "OSL", trim_range = NULL),
    class = "RLum.Data.Curve")
  testthat::expect_length(
    object = t@data[,1], n = 250)

  ## test some edge cases from trim
  ## 1
  testthat::expect_s4_class(
    object = trim_RLum.Data(temp@records[[1]], trim_range = c(1)),
    class = "RLum.Data.Curve")
  ## 0
  testthat::expect_s4_class(
    object = trim_RLum.Data(temp@records[[1]], trim_range = c(0)),
    class = "RLum.Data.Curve")
  ## -1
  testthat::expect_s4_class(
    object = trim_RLum.Data(temp@records[[1]], trim_range = c(-1)),
    class = "RLum.Data.Curve")

  ## c(0, 1)
  t <- trim_RLum.Data(temp@records[[1]], trim_range = c(0, 1))
  expect_equal(nrow(t@data), 1)

  ## c(-10, -20)
  t <- trim_RLum.Data(temp@records[[1]], trim_range = c(-10, -20))
  expect_equal(nrow(t@data), 11)

  ## c(1025, 2)
  t <- trim_RLum.Data(temp@records[[1]], trim_range = c(1025, 2))
  expect_equal(nrow(t@data), 249)

  ## c(1,2,3)
  testthat::expect_s4_class(
    object = trim_RLum.Data(temp@records[[1]], trim_range = c(1:3)),
    class = "RLum.Data.Curve")

  ## c(265, 270)
  t <- trim_RLum.Data(temp@records[[1]], trim_range = c(265, 270))
  expect_equal(nrow(t@data), 1)
})

test_that("RLum.Data.Spectrum", {
  testthat::skip_on_cran()

  ## simple test for RLum.Data.Spectrum ... this can be kept
  ## simple because everything else was tested already
  data(ExampleData.XSYG, envir = environment())

  t <- testthat::expect_s4_class(
    object = trim_RLum.Data(TL.Spectrum, trim_range = c(2, 4)),
    class = "RLum.Data.Spectrum")
  testthat::expect_length(object = t@data[1,], n = 3)

  ## both values in the range exceeding the number of channels
  t <- expect_s4_class(trim_RLum.Data(TL.Spectrum, trim_range = c(265, 270)),
                       "RLum.Data.Spectrum")
  expect_length(ncol(t@data), 1)

  ## RLum.Analysis object with RLum.Data.Spectrum data
  obj <- set_RLum("RLum.Analysis", records = list(TL.Spectrum))
  t <- expect_s4_class(trim_RLum.Data(obj, trim_range = 10),
                       "RLum.Analysis")
  expect_equal(ncol(t@records[[1]]@data), 10)
})

test_that("RLum.Data.Image", {
  testthat::skip_on_cran()

  ## simple test for RLum.Data.Spectrum ... this can be kept
  ## simple because everything else was tested already
  data(ExampleData.RLum.Data.Image, envir = environment())

  testthat::expect_s4_class(
    trim_RLum.Data(ExampleData.RLum.Data.Image, trim_range = c(10, 100)),
    class = "RLum.Data.Image")

  empty <- set_RLum("RLum.Data.Image")
  expect_equal(empty, trim_RLum.Data(empty))
})

test_that("RLum.Analysis", {
  testthat::skip_on_cran()

  ## generate case where one OSL curve has one channel less
  temp@records[[2]]@data <- temp@records[[2]]@data[-nrow(temp[[2]]@data),]

  ## now all OSL curves should be shortened to 999
  t <- testthat::expect_s4_class(
    object = trim_RLum.Data(temp),
    class = "RLum.Analysis")

 ## check for OSL, the TL must remain untouched
 testthat::expect_length(
   object = t@records[[4]]@data[,1], n = 999)
 testthat::expect_length(
   object = t@records[[1]]@data[,1], n = 250)

 ## apply a trimming to all curves
 t <- testthat::expect_s4_class(
   object = trim_RLum.Data(temp, trim_range = c(10,20)),
   class = "RLum.Analysis")
 testthat::expect_length(
   object = t@records[[4]]@data[,1], n = 11)
 testthat::expect_length(
   object = t@records[[1]]@data[,1], n = 11)

 ## apply a trimming to OSL curves only
 t <- testthat::expect_s4_class(
   object = trim_RLum.Data(temp, recordType = "OSL (PMT)", trim_range = c(10,20)),
   class = "RLum.Analysis")
 testthat::expect_length(
   object = t@records[[4]]@data[,1], n = 11)
 testthat::expect_length(
   object = t@records[[1]]@data[,1], n = 250)
})

## load data
data(ExampleData.RLum.Analysis, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_RLum2Risoe.BINfileData(object = NA),
               "'object' should be of class 'RLum.Analysis', 'RLum.Data.Curve' or")
  expect_error(convert_RLum2Risoe.BINfileData(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
})

test_that("test functionality", {
  testthat::skip_on_cran()

  ##provide list
  expect_s4_class(convert_RLum2Risoe.BINfileData(list(IRSAR.RF.Data,IRSAR.RF.Data)), "Risoe.BINfileData")
  expect_s4_class(convert_RLum2Risoe.BINfileData(list(IRSAR.RF.Data)),
                  "Risoe.BINfileData")

  ## additional metadata
  obj <- IRSAR.RF.Data@records[[1]]
  obj@info <- list(version="12", name="test",
                   startDate="20210101150845")
  res <- convert_RLum2Risoe.BINfileData(obj)
  expect_equal(res@METADATA$VERSION, "12")
  expect_equal(res@METADATA$SAMPLE, "test")
  expect_equal(res@METADATA$DATE, "20210101")
  expect_equal(res@METADATA$TIME, "150845")

  ## regression check for "raw" conversion error from double conversion
  data(ExampleData.BINfileData, envir = environment())
  expect_s4_class(
    convert_RLum2Risoe.BINfileData(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)),
    class = "Risoe.BINfileData")

  ## regression for combination of data where we mixed datasets that should not be mixed
  tmp_object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)
  tmp_curves <- tmp_object[[3]]@records[[1]]
  tmp_curves@recordType <- "NONE"
  tmp_object[[3]]@records <- c(tmp_object[[3]]@records, rep(tmp_curves,2))
  expect_s4_class(
    convert_RLum2Risoe.BINfileData(tmp_object),
    class = "Risoe.BINfileData")
})

test_that("test functionality", {
  testthat::skip_on_cran()

  ## mock current time
  local_mocked_bindings(Sys.time = function() as.POSIXct("2026-02-16 10:22:03"),
                        .package = "base")

  ## RLum.Analysis
  expect_snapshot_Risoe(convert_RLum2Risoe.BINfileData(IRSAR.RF.Data))

  ## RLum.Data.Curve
  expect_snapshot_Risoe(convert_RLum2Risoe.BINfileData(IRSAR.RF.Data@records[[1]]))
})

## load data
data(ExampleData.BINfileData, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(Risoe.BINfileData2RLum.Analysis("test"),
               "'object' should be of class 'Risoe.BINfileData")
  expect_error(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = "test"),
               "'pos' should be of class 'numeric', 'integer' or NULL")
  expect_error(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, run = 10:12),
               "'run' contains invalid runs, valid runs are: 1, 2, 3, 4, 5, 6")
  expect_error(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, set = 10:12),
               "'set' contains invalid sets, valid sets are: 2, 3, 5, 6")
  expect_error(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, ltype = 10:12),
               "'ltype' contains invalid ltypes, valid ltypes are: 'TL', 'OSL'")
  expect_error(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, dtype = 10:12),
               "'dtype' contains invalid dtypes, valid dtypes are: 'Natural'")

  risoe <- set_Risoe.BINfileData(METADATA = data.frame(ID = 1, POSITION = 0))
  expect_error(Risoe.BINfileData2RLum.Analysis(risoe),
               "'object' has missing columns in METADATA: 'GRAIN', 'RUN', 'SET'")

  expect_warning(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1:30),
                 "Invalid position number skipped")
  expect_warning(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, grain = 1:10),
                 "Invalid grain number skipped")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  res <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data,
                                         txtProgressBar = TRUE)
  })
  expect_type(res, "list")
  expect_length(res, 24)

  SW({
  res <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1:3,
                                         txtProgressBar = TRUE)
  })
  expect_type(res, "list")
  expect_length(res, 3)

  res <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1,
                                         txtProgressBar = TRUE)
  expect_s4_class(res, "RLum.Analysis")

  ## FI readers like to write a NA instead of 0 in the grain column
  obj <- CWOSL.SAR.Data
  obj@METADATA[["GRAIN"]] <- rep(NA, length(obj@METADATA[["GRAIN"]]))
  res <- Risoe.BINfileData2RLum.Analysis(obj, pos = 1)
  expect_s4_class(res, "RLum.Analysis")

  ## reading an empty object
  empty <- read_BIN2R(test_path("_data/BINfile_V3.bin"), n.records = 999,
                      verbose = FALSE)
  expect_null(Risoe.BINfileData2RLum.Analysis(empty))
  expect_null(Risoe.BINfileData2RLum.Analysis(empty, keep.empty = FALSE))
  expect_warning(expect_null(Risoe.BINfileData2RLum.Analysis(empty, pos = 0)),
                 "Invalid position number skipped: 0")

  ## reading an object with fields set to zero
  zero <- set_Risoe.BINfileData(METADATA = data.frame(ID = 1, POSITION = 0,
                                                      GRAIN = 0, RUN = 0, SET = 0,
                                                      LTYPE = 0, DTYPE = 0))
  expect_message(res <- Risoe.BINfileData2RLum.Analysis(zero),
                 "Empty Risoe.BINfileData object detected")
  expect_s4_class(res, "RLum.Analysis")
  expect_length(res, 0)
  expect_message(res <- Risoe.BINfileData2RLum.Analysis(zero, pos = 0),
                 "Empty Risoe.BINfileData object detected")
  expect_s4_class(res, "RLum.Analysis")
  expect_length(res, 0)
  expect_message(expect_null(
      Risoe.BINfileData2RLum.Analysis(zero, keep.empty = FALSE)),
      "Empty Risoe.BINfileData object detected")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  expect_snapshot_RLum(Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data,
                                                       pos = 1, run = 1,
                                                       txtProgressBar = FALSE))
  })
})

## load data
data(ExampleData.BINfileData, envir = environment())

## transform the values from the third position in an RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos = 3)

test_that("input validation", {
  skip_on_cran()

  expect_error(analyse_SAR.TL(),
               "is missing, with no default")
  expect_error(analyse_SAR.TL("test"),
               "[analyse_SAR.TL()] 'object' should be of class 'RLum.Analysis'",
               fixed = TRUE)
  expect_error(analyse_SAR.TL(object),
               "'signal_integral' should be of class 'integer' or 'numeric'")
  expect_error(analyse_SAR.TL(object, signal_integral = NA),
               "'signal_integral' should be of class 'integer' or 'numeric'")
  expect_error(analyse_SAR.TL(list(object, "test")),
               "All elements of 'object' should be of class 'RLum.Analysis'")
  expect_error(analyse_SAR.TL(object, signal_integral = 1:2),
               "Input TL curves are not a multiple of the sequence structure")
  expect_error(analyse_SAR.TL(object, signal_integral = 1:2,
                              sequence.structure = "EXCLUDE"),
               "'sequence.structure' contains no 'SIGNAL' entry")
  expect_error(analyse_SAR.TL(object, dose.points = c(2, 2),
                              signal_integral = 210:220,
                              sequence.structure = c("SIGNAL", "BACKGROUND")),
               "Length of 'dose.points' not compatible with number of signals")
  expect_error(analyse_SAR.TL(object, signal_integral = 1:2,
                              sequence.structure = c("SIGNAL", "BACKGROUND"),
                              integral_input = "error"),
               "[analyse_SAR.TL()] 'integral_input' should be one of ",
               fixed = TRUE)

  SW({
  expect_warning(analyse_SAR.TL(object, signal_integral = c(0, 1.5),
                                sequence.structure = c("SIGNAL", "BACKGROUND"),
                                integral_input = "measurement"),
                 "Conversion of 'signal_integral' from temperature to channels failed")
  expect_warning(analyse_SAR.TL(object, signal_integral = 2000:2100,
                                sequence.structure = c("SIGNAL", "BACKGROUND"),
                                integral_input = "measurement"),
                 "Conversion of 'signal_integral' from temperature to channels failed")
  })

  obj.rm <- object
  obj.rm@records[[1]]@data <- obj.rm@records[[1]]@data[1:225, ]
  expect_error(analyse_SAR.TL(obj.rm, signal_integral = 210:220,
                              sequence.structure = c("SIGNAL", "BACKGROUND")),
               "Signal ranges differ (225, 250), check sequence structure",
               fixed = TRUE)

  expect_output(expect_null(
      analyse_SAR.TL(object[1:4], fit.method = "ERROR",
                     signal_integral = 2:3,
                     sequence.structure = "SIGNAL"),
      "'fit.method' should be one of 'LIN', 'QDR', 'EXP', 'EXP OR LIN'"))

  ## deprecated argument
  expect_warning(analyse_SAR.TL(object, signal.integral = 2:3,
                                sequence.structure = c("SIGNAL", "BACKGROUND"),
                                verbose = FALSE),
                 "was deprecated in v1.2.0, use 'signal_integral'")
  expect_warning(analyse_SAR.TL(object,
                                signal_integral = 210:220,
                                integral_input = "temperature",
                                sequence.structure = c("SIGNAL", "BACKGROUND"),
                                verbose = FALSE),
                 "was deprecated in v1.2.0, use 'integral_input = \"measurement\"")
})

test_that("snapshot tests", {
  skip_on_cran()

  SW({
  expect_snapshot_RLum(
    analyse_SAR.TL(
        list(object, object),
        signal_integral = 210:220,
        dose.points = 1:7,
        integral_input = "measurement",
        sequence.structure = c("SIGNAL", "BACKGROUND"))
  )

  expect_warning(
  expect_snapshot_RLum(
    analyse_SAR.TL(
        list(object),
        signal_integral = 210:220,
        dose.points = 1:7,
        log = "x",
        sequence.structure = c("SIGNAL", "BACKGROUND"))
    ),
    "Non-positive values detected, log-scale disabled"
  )

  SW({
  expect_message(analyse_SAR.TL(object, signal_integral = 2:3,
                                sequence.structure = "SIGNAL"),
                 "Too many curves, only the first 21 curves are plotted")
  })

  expect_warning(
  expect_snapshot_RLum(
    analyse_SAR.TL(object, signal_integral = 2:3,
                   sequence.structure = c("SIGNAL", "EXCLUDE"))
    ),
  "Error column invalid or 0, 'fit.weights' ignored")
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              analyse_SAR.TL(object,
                                             signal_integral = 210:220,
                                             sequence.structure = c("SIGNAL", "BACKGROUND")))
  })
})

test_that("regression tests", {
  skip_on_cran()

  ## issue 147 --------------------------------------------------------------

  SW({
  set.seed(1)
  expect_snapshot_RLum(
    analyse_SAR.TL(object, sequence.structure = c("SIGNAL", "BACKGROUND"),
                   signal_integral = 2:3),
    tolerance = 1.5e-4
  )

  seq.structure <- c("SIGNAL", "EXCLUDE", "BACKGROUND", "EXCLUDE", "PREHEAT",
                     "EXCLUDE", "BACKGROUND", "SIGNAL", "EXCLUDE", "EXCLUDE",
                     "EXCLUDE", "EXCLUDE")
  expect_error(analyse_SAR.TL(object, signal_integral = 2,
                              sequence.structure = seq.structure),
               "[calc_TLLxTxRatio()] 'Tx.data.signal' should be of class 'data.frame'",
               fixed = TRUE)
  })

  expect_message(
  expect_snapshot_RLum(
    analyse_SAR.TL(object, dose.points = 2,
                   signal_integral = 210:220,
                   sequence.structure = c("SIGNAL", "BACKGROUND"))
  ),
  "[fit_DoseResponseCurve()] Error: All points have the same dose, NULL returned",
  fixed = TRUE)

  ## issue 519
  bin.v8 <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  SW({
  expect_error(
      analyse_SAR.TL(read_BIN2R(bin.v8, fastForward = TRUE, verbose = FALSE),
                     signal_integral = 1:20),
      "Input TL curves are not a multiple of the sequence structure")
  })
})

## load data
data(ExampleData.BINfileData, envir = environment())
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_DetPlot("error"),
               "'object' should be of class 'RLum.Analysis'")
  expect_error(plot_DetPlot(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_warning(
  expect_error(plot_DetPlot(object, signal.integral.min = "error"),
               "'signal.integral.min' should be of class 'integer' or 'numeric'"),
               "were deprecated in v1.2.0, use 'signal_integral'")

  expect_error(plot_DetPlot(object,
                            signal_integral = 1:2,
                            background_integral = 900:1000,
                            method = "error"),
               "'method' should be one of 'shift' or 'expansion'")
  expect_error(plot_DetPlot(object,
                            signal_integral = 1:2,
                            background_integral = 900:1000,
                            analyse_function = "error",
                            verbose = FALSE),
               "'analyse_function' should be one of 'analyse_SAR.CWOSL'")
})

test_that("plot_DetPlot", {
  testthat::skip_on_cran()

  ## simple run with default but no plot
  results <- expect_s4_class(plot_DetPlot(
    object,
    method = "shift",
    signal_integral = 1:3,
    background_integral = 900:1000,
    analyse_function.control = list(
      fit.method = "LIN"),
    n.channels = 2,
    verbose = FALSE,
    plot = FALSE),
    "RLum.Results")

  ## test self call with multi core
  SW({
  expect_message(expect_s4_class(plot_DetPlot(
    object = list(x = object, y = object),
    method = "shift",
    signal_integral = 1:3,
    background_integral = 900:1000,
    analyse_function.control = list(
      fit.method = "LIN",
      trim_channels = TRUE
    ),
    multicore = 2,
    n.channels = 2,
    verbose = TRUE,
    plot = FALSE),
    "RLum.Results"),
    "Running multicore session using 2 cores")

  expect_s4_class(plot_DetPlot(
    object = list(x = object, y = object),
    method = "shift",
    signal_integral = 1:3,
    background_integral = 900:1000,
    analyse_function.control = list(
      fit.method = "LIN",
      trim_channels = TRUE
    ),
    n.channels = 2,
    verbose = TRUE,
    plot = FALSE),
    "RLum.Results")

  ## serial
  expect_s4_class(plot_DetPlot(
    object = list(x = object, y = object),
    method = "shift",
    signal_integral = 1:3,
    background_integral = 900:1000,
    analyse_function.control = list(
      fit.method = "LIN",
      trim_channels = TRUE
    ),
    multicore = FALSE,
    n.channels = 2,
    verbose = TRUE,
    plot = FALSE),
    "RLum.Results")
  })

  ## try with NA values
  object@records[[2]][,2] <- 1
  object@records[[4]][,2] <- 1
  object@records[[6]][,2] <- 1
  object@records[[8]][,2] <- 1
  results <- expect_s4_class(suppressWarnings(plot_DetPlot(
    object,
    method = "expansion",
    signal_integral = 1:3,
    background_integral = 900:1000,
    analyse_function.control = list(
      fit.method = "EXP"),
    verbose = FALSE,
    n.channels = 1)),
    "RLum.Results")

  SW({
  ## n.channels not set
  expect_message(plot_DetPlot(object,
                              method = "shift",
                              signal_integral = 5:6,
                              background_integral = 10:50,
                              analyse_function.control = list(
                                  fit.method = "LIN"),
                              verbose = TRUE),
                 "'n.channels' not specified, set to 3")

  ## analyse_pIRIRSequence
  tmp <- subset(object, recordType != "IRSL (PMT)" & ID != 1)
  plot_DetPlot(
      tmp,
      method = "expansion",
      signal_integral = 1:2,
      background_integral = 900:1000,
      analyse_function = "analyse_pIRIRSequence",
      analyse_function.control = list(
          sequence.structure = c("TL", "IR50"),
          fit.method = "LIN"),
      respect_RC.Status = TRUE,
      n.channels = 2)

  expect_warning(plot_DetPlot(
      tmp,
      method = "expansion",
      signal_integral = 1:2,
      background_integral = 900:1000,
      plot.single = TRUE,
      n.channels = 2),
      "'plot.single' was deprecated in v1.0.0, use 'plot_singlePanels' instead")

  ## analyse_pIRIRSequence on an inconsistent object
  suppressWarnings( # ignore additional warnings from fit_DoseResponseCurve()
  expect_error(
      expect_warning(plot_DetPlot(
          object,
          signal_integral = 1:2,
          background_integral = 900:1000,
          analyse_function = "analyse_pIRIRSequence",
          analyse_function.control = list(
              fit.method = "LIN"),
          verbose = FALSE,
          n.channels = 1),
          "An error occurred, analysis skipped"),
      "No valid results produced")
  )

  ## integral_input
  set.seed(1)
  res1 <- plot_DetPlot(object, method = "shift",
                       signal_integral = c(0.04, 0.12),
                       background_integral = 34:40,
                       integral_input = "measurement",
                       analyse_function.control = list(
                           fit.method = "LIN",
                           trim_channels = TRUE),
                       n.channels = 3,
                       plot = FALSE)
  set.seed(1)
  res2 <- plot_DetPlot(object, method = "shift",
                       signal_integral = 1:3,
                       background_integral = 850:1000,
                       integral_input = "channel",
                       analyse_function.control = list(
                           fit.method = "LIN",
                           trim_channels = TRUE),
                       n.channels = 3,
                       plot = FALSE)
  res1@info <- res2@info <- list() # remove $call
  res1@.uid <- res2@.uid <- NA_character_
  res1@.pid <- res2@.pid <- NA_character_
  res1@data$De.values$UID <- res2@data$De.values$UID <- NULL
  expect_equal(res1, res2)

  ## deprecated argument
  expect_warning(plot_DetPlot(
      object,
      method = "shift",
      signal.integral.min = 1,
      signal.integral.max = 3,
      background.integral.min = 900,
      background.integral.max = 1000,
      n.channels = 1),
      "were deprecated in v1.2.0, use 'signal_integral' and 'background_integral'")
  expect_warning(expect_error(plot_DetPlot(
      object,
      method = "shift",
      signal.integral.min = 1,
      signal.integral.max = 3,
      background.integral.min = 900,
      background.integral.max = 1000,
      integral_input = "measurement",
      n.channels = 1),
      "'integral_input' is not supported with old argument names"),
      "were deprecated in v1.2.0, use 'signal_integral' and 'background_integral'")
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
  vdiffr::expect_doppelganger("shift trim",
                              plot_DetPlot(object,
                                           method = "shift",
                                           signal_integral = 1:3,
                                           background_integral = 900:1000,
                                           analyse_function.control = list(
                                               fit.method = "LIN",
                                               trim_channels = TRUE
                                           ),
                                           n.channels = 3))
  vdiffr::expect_doppelganger("expansion",
                              plot_DetPlot(object,
                                           method = "expansion",
                                           signal_integral = 1:3,
                                           background_integral = 900:1000,
                                           analyse_function.control = list(
                                               fit.method = "LIN"),
                                           n.channels = 2))
  })
})

## load data
data(ExampleData.XSYG, envir = environment())
object <- get_RLum(OSL.SARMeasurement$Sequence.Object,
                   recordType = "OSL (UVVIS)", drop = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(verify_SingleGrainData("test"),
               "'object' should be of class 'Risoe.BINfileData', 'RLum.Analysis'")
  expect_error(verify_SingleGrainData(object, threshold = "error"),
               "'threshold' should be a single positive value")
  expect_error(verify_SingleGrainData(object, threshold = integer(0)),
               "'threshold' should be a single positive value")
  expect_error(verify_SingleGrainData(object, use_fft = "error"),
               "'use_fft' should be a single logical value")
  expect_error(verify_SingleGrainData(object, cleanup = "error"),
               "'cleanup' should be a single logical value")
  expect_error(verify_SingleGrainData(object, cleanup_level = "error"),
               "'cleanup_level' should be one of 'aliquot' or 'curve'")
  expect_error(verify_SingleGrainData(object, verbose = "error"),
               "'verbose' should be a single logical value")

  object@originator <- "error"
  expect_error(verify_SingleGrainData(object),
               "Object originator 'error' not supported")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## RLum.Analysis object
  expect_warning(output <- verify_SingleGrainData(object),
                 "'selection_id' is NA, everything tagged for removal")
  expect_s4_class(output, "RLum.Results")
  expect_s3_class(output$selection_full, "data.frame")
  expect_equal(sum(output@data$selection_full$VALID), 11)
  expect_equal(output@originator, "verify_SingleGrainData")

  expect_message(res <- verify_SingleGrainData(object, cleanup = TRUE,
                                              cleanup_level = "curve",
                                              threshold = 100),
                "RLum.Analysis object reduced to records")
  expect_s4_class(res, "RLum.Analysis")
  expect_equal(res@originator, "verify_SingleGrainData")
  expect_length(res@records, 5)

  ## check for empty object in a list
  object_empty <- list(
    set_RLum(class = "RLum.Analysis", originator = "Risoe.BINfileData2RLum.Analysis"),
    set_RLum(
      class = "RLum.Analysis",
      originator = "Risoe.BINfileData2RLum.Analysis",
      records = list(set_RLum(
        "RLum.Data.Curve",
        data = matrix(1:100, ncol = 2),
        info = list(POSITION = 1, GRAIN = 1)
      ))
    )
  )
  expect_warning(verify_SingleGrainData(object_empty), regexp = "Cannot process empty RLum.Analysis objects. NULL returned!")

  ## threshold too high, empty object generated
  expect_message(res <- verify_SingleGrainData(object, cleanup = TRUE,
                                              cleanup_level = "curve",
                                              threshold = 2000),
                "RLum.Analysis object reduced to records")
  expect_s4_class(res, "RLum.Analysis")
  expect_equal(res@originator, "read_XSYG2R")
  expect_length(res@records, 0)

  ## threshold too high on a list
  expect_message(res <- verify_SingleGrainData(list(object), cleanup = TRUE,
                                               cleanup_level = "curve",
                                               threshold = 2000),
                 "RLum.Analysis object reduced to records")
  expect_type(res, "list")
  expect_s4_class(res[[1]], "RLum.Analysis")
  expect_equal(res[[1]]@originator, "read_XSYG2R")
  expect_length(res[[1]]@records, 0)

  ## check for cleanup
  data(ExampleData.BINfileData, envir = environment())
  t <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)
  expect_warning(
    object = verify_SingleGrainData(t, cleanup = TRUE, threshold = 20000),
    regexp = "Verification and cleanup removed all records, NULL returned")
  expect_null(suppressWarnings(verify_SingleGrainData(t, cleanup = TRUE, threshold = 20000)))

  ## Risoe.BINfileData
  res <- expect_silent(verify_SingleGrainData(CWOSL.SAR.Data, plot = TRUE))
  expect_s4_class(res, "RLum.Results")

  expect_output(res <- verify_SingleGrainData(CWOSL.SAR.Data, cleanup = TRUE,
                                              cleanup_level = "curve"),
                       "Risoe.BINfileData object reduced to records")
  expect_s4_class(res, "Risoe.BINfileData")

  ## use_fft
  expect_s4_class(verify_SingleGrainData(CWOSL.SAR.Data, cleanup = TRUE, use_fft = TRUE, verbose = FALSE),
                "Risoe.BINfileData")

  obj.risoe <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
  res <- expect_silent(verify_SingleGrainData(obj.risoe))
  expect_s4_class(verify_SingleGrainData(obj.risoe, use_fft = TRUE), "RLum.Results")

  ## remove all and cleanup
  expect_warning(
    object = verify_SingleGrainData(CWOSL.SAR.Data, cleanup = TRUE, threshold = 20000),
    regexp = "Verification and cleanup removed all records, NULL returned")
  expect_null(suppressWarnings(verify_SingleGrainData(CWOSL.SAR.Data, cleanup = TRUE, threshold = 20000)))

  ## empty list
  expect_s4_class(res <- verify_SingleGrainData(list()),
                  "RLum.Results")
  expect_length(res@data, 0)
  expect_equal(res@originator, "verify_SingleGrainData")

  expect_s4_class(res <- verify_SingleGrainData(list(), cleanup = TRUE),
                  "RLum.Analysis")
  expect_length(res@records, 0)
  expect_equal(res@originator, "verify_SingleGrainData")

  expect_s4_class(res <- verify_SingleGrainData(list(), cleanup = NA),
                  "RLum.Results")
  expect_length(res@data, 0)
  expect_equal(res@originator, "verify_SingleGrainData")

  ## list
  expect_silent(suppressWarnings(verify_SingleGrainData(list(object))))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              verify_SingleGrainData(object,
                                                     plot = TRUE))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 740
  object@records[[1]]@info$position <- "123"
  expect_s4_class(verify_SingleGrainData(object),
                  "RLum.Results")
})

## load data
data(ExampleData.RLum.Analysis, envir = environment())
obj <- IRSAR.RF.Data
data(ExampleData.XSYG, envir = environment())
sar <- OSL.SARMeasurement$Sequence.Object[1:9]

## construct test object
tmp <- set_RLum(
    "RLum.Analysis",
    protocol = "testthat",
    records = lapply(1:20, function(x)  {
      set_RLum(
          "RLum.Data.Curve",
          recordType = "test",
          data = matrix(1:10, ncol = 2),
          info = list(el = as.character(x))
      )
    }),
    info = list(el = "test")
)

## missing info slot to test very old objects (before commit 5226b88, Jan 2016)
old <- tmp
attr(old, "info") <- NULL

test_that("Check the example and the numerical values", {
  testthat::skip_on_cran()

  ## set_RLum()
  expect_s4_class(tmp, "RLum.Analysis")
  expect_s4_class(set_RLum("RLum.Analysis", records = tmp, .pid = "test"), "RLum.Analysis")

  ##overwrite object
  expect_s4_class(set_RLum("RLum.Analysis", records = tmp), "RLum.Analysis")

  ## as()
  expect_type(as(tmp, "list"), "list")
  expect_s4_class(as(list(), "RLum.Analysis"), "RLum.Analysis")

  ## output
  expect_output(print(as(list(), "RLum.Analysis")), regexp = "This is an empty object")
  expect_s4_class(
    set_RLum(
      "RLum.Analysis",
      protocol = "testthat",
      records = set_RLum(
        "RLum.Analysis",
        protocol = "nest",
        records = list(matrix(1:10, ncol = 2))
      ),
      info = list(el = "test")
    ),
    "RLum.Analysis"
  )

  ## show()
  expect_output(print(tmp))
  expect_output(print(old))

    ## variant of show where we have NULL elements
    o_NULL <- set_RLum("RLum.Analysis", records = list(set_RLum("RLum.Data.Curve"), NULL))
    expect_output(print(o_NULL))

  ## names()
  expect_type(names(tmp), "character")
})

test_that("get_RLum", {
  testthat::skip_on_cran()

  ## input validation
  expect_error(get_RLum(obj, subset = 1),
               "[get_RLum()] 'subset' must contain a logical expression",
               fixed = TRUE)
  expect_error(get_RLum(obj, subset = (error == "OSL")),
               "[get_RLum()] Invalid subset expression, valid terms are:",
               fixed = TRUE)
  expect_error(get_RLum(tmp, record.id = "character"),
               "'record.id' should be of class 'integer', 'numeric', 'logical' or NULL")
  expect_error(get_RLum(tmp, recordType = 1L),
               "'recordType' should be of class 'character'")
  expect_error(get_RLum(tmp, curveType = 1L),
               "'curveType' should be of class 'character'")
  expect_error(get_RLum(tmp, RLum.type = 1L),
               "'RLum.type' should be of class 'character'")
  expect_error(get_RLum(tmp, get.index = "a"),
               "'get.index' should be a single logical value")
  expect_error(get_RLum(obj, get.index = NULL),
               "'get.index' should be a single logical value")
  expect_error(get_RLum(obj, drop = NA),
               "'drop' should be a single logical value")
  expect_error(get_RLum(obj, info.object = list()),
               "'info.object' should be of class 'character' or NULL and have length 1")
  expect_error(get_RLum(obj, subset = "recordType == 'RF (NA)'", get.index = NA),
               "'get.index' should be a single logical value")
  SW({
  expect_message(expect_null(get_RLum(obj, subset = (recordType == "RF"))),
                 "Error: 'subset' expression produced an empty selection")
  expect_message(expect_null(get_RLum(obj, subset = "recordType == 'RF'")),
                 "Error: 'subset' expression produced an empty selection")
  })

  ## check functionality
  expect_length(get_RLum(obj, subset = (recordType == "RF (NA)")), 2)
  expect_length(get_RLum(obj, subset = "recordType == 'RF (NA)'"), 2)
  expect_length(get_RLum(tmp, subset = (el == "2")), 1)
  expect_s4_class(get_RLum(tmp, subset = (el == "2")), "RLum.Analysis")
  expect_type(get_RLum(tmp, info.object = "el"), "character")

  ## check subset for info elements
  t_subset <- IRSAR.RF.Data
  t_subset@records[[1]]@info <- list(TEST = "SUBSET")
  expect_length(get_RLum(t_subset, subset = c(TEST == "SUBSET")), 1)

  expect_type(get_RLum(obj, get.index = FALSE), "list")
  expect_type(get_RLum(obj, get.index = TRUE), "integer")
  expect_s4_class(get_RLum(obj, get.index = FALSE, drop = FALSE),
                  "RLum.Analysis")
  expect_type(get_RLum(obj, get.index = TRUE, drop = FALSE),
              "integer")
  expect_type(get_RLum(tmp, record.id = c(3, 5), get.index = FALSE),
              "list")

  expect_s4_class(get_RLum(obj, record.id = 1),
                  "RLum.Data.Curve")
  expect_s4_class(get_RLum(obj, record.id = 1, drop = FALSE),
                  "RLum.Analysis")
  expect_equal(get_RLum(tmp, record.id = 1, get.index = TRUE),
               1)
  expect_equal(get_RLum(tmp, record.id = -c(1:15), get.index = TRUE),
               1:5)
  expect_equal(get_RLum(tmp, record.id = c(1, 10, 20), get.index = TRUE),
               1:3)
  expect_message(expect_null(get_RLum(obj, record.id = 99)),
                 "[get_RLum()] Error: At least one 'record.id' is invalid",
                 fixed = TRUE)
  expect_message(expect_null(get_RLum(obj, record.id = 99, get.index = TRUE)),
                 "[get_RLum()] Error: At least one 'record.id' is invalid",
                 fixed = TRUE)

  expect_warning(res <- get_RLum(obj, RLum.type = "error"),
                 "This request produced an empty list of records")
  expect_type(res, "list")
  expect_length(res, 0)
  expect_warning(res <- get_RLum(obj, RLum.type = "error", drop = FALSE),
                 "This request produced an empty list of records")
  expect_s4_class(res, "RLum.Analysis")
  expect_length(res, 0)
  expect_warning(expect_null(get_RLum(obj, RLum.type = "error", get.index = TRUE)),
                 "This request produced an empty list of records")
  expect_warning(get_RLum(tmp, info.object = "missing"),
                 "[get_RLum()] Invalid 'info.object' name, valid names are:",
                 fixed = TRUE)
  expect_warning(expect_null(get_RLum(set_RLum("RLum.Analysis"),
                                      info = "test")),
                 "[get_RLum()] This 'RLum.Analysis' object has no info objects",
                 fixed = TRUE)
  expect_warning(get_RLum(obj, record.id = -c(1:2)),
                 "This request produced an empty list of records")
  expect_warning(expect_null(get_RLum(obj, record.id = -c(1:2), get.index = TRUE)),
                 "This request produced an empty list of records")
  SW({
  expect_message(expect_null(get_RLum(obj, subset = (recordType == "error"))),
                 "'subset' expression produced an empty selection, NULL returned")
  expect_message(expect_null(get_RLum(obj, subset = NA)),
                 "'subset' expression produced an empty selection, NULL returned")
  })
})

test_that("sort_RLum", {
  testthat::skip_on_cran()

  ## input validation
  expect_error(sort_RLum(sar, slot = NA),
               "'slot' should be of class 'character' or NULL")
  expect_error(sort_RLum(sar, slot = c("curveType", "error")),
               "Invalid 'slot' name, valid names are:")
  expect_error(sort_RLum(sar, info_element = NA),
               "'info_element' should be of class 'character' or NULL")
  expect_error(sort_RLum(sar, info_element = c("position", "error")),
               "Invalid 'info_element' name, valid names are:")
  expect_error(sort_RLum(sar, slot = NULL, info_element = NULL),
               "At least one of 'slot' and 'info_element' should not be NULL")
  expect_error(sort_RLum(sar, slot = "recordType", decreasing = "error"),
               "'decreasing' should be of class 'logical'")
  expect_error(sort_RLum(sar, slot = "recordType", decreasing = NA),
               "'decreasing' should be of class 'logical'")

  ## check empty object
  expect_s4_class(sort_RLum(set_RLum("RLum.Analysis")), class = "RLum.Analysis")

  ## check one curve object
  expect_s4_class(
    sort_RLum(set_RLum("RLum.Analysis", records = list(set_RLum("RLum.Data.Curve"))),
              slot = "recordType"),
                  class = "RLum.Analysis")

  ## sort only using the first field until #605 is done
  expect_message(sort_RLum(sar, slot = c("curveType", "recordType")),
                 "Only the first 'slot' field will be used in sorting")

  ## check functionality
  expect_snapshot_RLum(sort_RLum(sar, slot = "recordType"))
  expect_snapshot_RLum(sort_RLum(sar, info_element = "curveDescripter"))

  ## present a list of those objects
  expect_type(sort_RLum(list(sar, sar), info_element = "X_MIN"), "list")
  expect_type(sort_RLum(list(iris, mtcars), info_element = "X_MIN"), "list")
  expect_snapshot_RLum(sort_RLum(list(sar, sar), info_element = "X_MIN"))

  ## sort after three columns
  expect_s4_class(sort_RLum(sar, info_element = c("XY_LENGTH", "NCOL", "X_MIN")), "RLum.Analysis")

  ## now add spectra
  sar_a <- sar
  sar_a@records <- c(sar_a@records, set_RLum("RLum.Data.Spectrum"))
  expect_s4_class(sort_RLum(sar_a, info_element = "X_MIN"), "RLum.Analysis")

  ## try with image
  sar_a <- sar
  sar_a@records <- c(sar_a@records, set_RLum("RLum.Data.Image"))
  expect_s4_class(sort_RLum(sar_a, info_element = "X_MIN"), "RLum.Analysis")

  ## use slot sorting with more than one element in the slot
  ## it should not break
  sar_a <- sar
  sar_a@records[[1]]@.pid <- c("a", "b")
  sort_RLum(sar_a, slot = ".pid")

  empty <- as(list(), "RLum.Analysis")
  expect_equal(sort_RLum(empty, slot = "curveType"),
               empty)

  ## check a special case where individual info elements have a length > 1
  sar@records[[1]]@info <- c(sar@records[[1]]@info, test = list(x = 1:10))
  expect_s4_class(sort_RLum(sar, info_element = "startDate"), class = "RLum.Analysis")
})

test_that("structure_RLum", {
  testthat::skip_on_cran()

  ## input validation
  expect_error(structure_RLum(
      set_RLum("RLum.Analysis",
               records = list(set_RLum("RLum.Data.Image")))),
      "All elements of 'object' should be of class 'RLum.Data.Curve'")

  ## full functionality

  ## object with empty info
  expect_s3_class(res <- structure_RLum(obj),
                  "data.frame")
  expect_equal(nrow(res), length(obj@records))
  expect_equal(ncol(res), 13)
  expect_equal(res$n.channels, c(5, 524))
  expect_equal(res$recordType, c("RF (NA)", "RF (NA)"))
  expect_equal(unlist(res$info), c(NA, NA))

  expect_s3_class(res2 <- structure_RLum(obj, fullExtent = TRUE),
                  "data.frame")
  expect_equal(names(res2), names(res))
  expect_equal(res2$info, c(NA, NA))

  ## special case with longer .pid
  obj_a <- obj
  obj_a@records[[1]]@.pid <- c(obj_a@records[[1]]@.pid, obj_a@records[[1]]@.pid)
  expect_s3_class(structure_RLum(obj_a), "data.frame")

  ## object with some info
  data(ExampleData.BINfileData, envir = environment())
  d1 <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
  res <- structure_RLum(d1)
  expect_equal(nrow(res), length(d1@records))
  expect_equal(ncol(res), 13)

  res2 <- structure_RLum(d1, fullExtent = TRUE)
  expect_equal(nrow(res2), length(d1@records))
  expect_equal(ncol(res2), 12 + length(d1@records[[1]]@info))
  expect_null(res2$info)
  expect_equal(names(res2)[-c(1:12)],
               paste0("info.", names(d1@records[[1]]@info)))

  ## on an empty object with some info
  res <- structure_RLum(tmp)
  expect_equal(nrow(res), length(tmp@records))
  expect_equal(ncol(res), 13)
  expect_equal(res$n.channels, rep(5, 20))

  res2 <- structure_RLum(tmp, fullExtent = TRUE)
  expect_equal(nrow(res2), length(tmp@records))
  expect_equal(ncol(res2), 12 + length(tmp@records[[1]]@info))
  expect_null(res2$info)                     ## since @info contains only one
  expect_equal(names(res2)[-c(1:12)],        ##  element named `el`, the last
               names(tmp@records[[1]]@info)) ##  column in res2 is also named
  expect_equal(res2$el, as.character(1:20))  ##  `el` rather than `info.el`

  ## on an even emptier object empty info
  empty <- set_RLum("RLum.Analysis",
                    records = list(set_RLum("RLum.Data.Curve")))

  expect_s3_class(res <- structure_RLum(empty),
                  "data.frame")
  expect_equal(nrow(res), length(empty@records))
  expect_equal(ncol(res), 13)
  expect_equal(res$n.channels, 1)
  expect_equal(unlist(res$info), NA)

  res2 <- structure_RLum(empty, fullExtent = TRUE)
  expect_equal(nrow(res2), length(empty@records))
  expect_equal(ncol(res2), 13)
  expect_equal(names(res2), names(res))
  expect_equal(res2$info, NA)

  ## melt
  t <- melt_RLum(tmp)
  expect_type(t, "list")
})

test_that("remove_RLum", {
  testthat::skip_on_cran()

  ## remove all OSL curves
  t <- expect_s4_class(remove_RLum(sar, recordType = "OSL"), "RLum.Analysis")
  expect_length(t@records, n = 4)

  ## provide arguments that are overwritten
  t <- expect_s4_class(remove_RLum(sar, recordType = "OSL", drop = TRUE), "RLum.Analysis")
  expect_length(t@records, n = 4)

  t <- expect_s4_class(remove_RLum(sar, record.id = 8:20), "RLum.Analysis")
  expect_length(t@records, n = 7)

  ## provide as list
  t <- expect_type(remove_RLum(list(sar, sar), recordType = "OSL"), "list")
  expect_length(t, n = 2)

  ## odd wrong element
  t <- expect_type(remove_RLum(list(sar, "error"), recordType = "OSL"), "list")
  expect_length(t, n = 2)

  ## use subset
  ## this produces and error because of the logical expression
  expect_error(remove_RLum(list(sar, sar), subset = recordType == "OSL"))

  ## simple check for info element
  expect_s4_class(remove_RLum(sar, subset = "duration== 118"), "RLum.Analysis")

  ## this works with terminal output
  SW({
  t <- expect_type(remove_RLum(list(sar, sar), subset = "recordType == 'TL'"), "list")
  })
  expect_length(t, n = 2)

  ## invalid indices
  expect_equal(remove_RLum(sar, record.id = 70),
               sar)
})

## load data
data(ExampleData.BINfileData, envir = environment())

## convert Risoe.BINfileData into a curve object
temp <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos = 3)
Lx.data.signal <- get_RLum(temp, record.id = 1)
Lx.data.background <- get_RLum(temp, record.id = 2)
Tx.data.signal <- get_RLum(temp, record.id = 3)
Tx.data.background <- get_RLum(temp, record.id = 4)
signal_integral <- 210:230

test_that("input validation", {
  testthat::skip_on_cran()

    ## different data types
    expect_error(calc_TLLxTxRatio(
      Lx.data.signal,
      Lx.data.background,
      Tx.data.signal = as.data.frame(Tx.data.signal),
      Tx.data.background,
      signal_integral),
      "[calc_TLLxTxRatio()] Data types of Lx and Tx data differ",
      fixed = TRUE)

    ## different data types
    expect_error(calc_TLLxTxRatio(
      Lx.data.signal,
      Lx.data.background,
      Tx.data.signal = as.data.frame(Tx.data.signal),
      Tx.data.background,
      signal_integral),
      "[calc_TLLxTxRatio()] Data types of Lx and Tx data differ",
      fixed = TRUE)

    ## check for allowed data types
    expect_error(calc_TLLxTxRatio(
      Lx.data.signal = as.matrix(Tx.data.signal),
      Lx.data.background,
      Tx.data.signal = as.matrix(Tx.data.signal),
      Tx.data.background,
      signal_integral),
      "'Lx.data.signal' should be of class 'data.frame' or 'RLum.Data.Curve'")
    expect_error(calc_TLLxTxRatio(
      Lx.data.signal = Lx.data.signal,
      Lx.data.background,
      Tx.data.signal = Tx.data.signal,
      Tx.data.background,
      signal_integral = "error"),
      "'signal_integral' should be of class 'integer' or 'numeric'")

    ## check for different channel numbers
    expect_error(calc_TLLxTxRatio(
      Lx.data.signal = as.data.frame(Tx.data.signal)[1:10,],
      Lx.data.background,
      Tx.data.signal = as.data.frame(Tx.data.signal),
      Tx.data.background,
      signal_integral),
      "[calc_TLLxTxRatio()] Channel numbers differ for Lx and Tx data",
      fixed = TRUE)

  ## input sizes
  expect_error(calc_TLLxTxRatio(Lx.data.signal = iris[, 1, drop = FALSE]),
               "'Lx.data.signal' should have 2 columns")
  expect_error(calc_TLLxTxRatio(Lx.data.signal = iris,
                                Tx.data.signal = iris[, 1, drop = FALSE]),
               "'Tx.data.signal' should have 2 columns")
  expect_error(calc_TLLxTxRatio(Lx.data.signal = iris,
                                Tx.data.signal = iris,
                                Lx.data.background = iris[, 1, drop = FALSE],
                                signal_integral = 1:5),
               "'Lx.data.background' should have 2 columns")
  expect_error(calc_TLLxTxRatio(Lx.data.signal = iris,
                                Tx.data.signal = iris,
                                Lx.data.background = iris,
                                Tx.data.background = iris[, 1, drop = FALSE],
                                signal_integral = 1:5),
               "'Tx.data.background' should have 2 columns")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## trigger warning
  expect_warning(calc_TLLxTxRatio(
      Lx.data.signal,
      Lx.data.background,
      Tx.data.signal,
      Tx.data.background = Lx.data.background,
      signal_integral),
      regexp = "\\[calc\\_TLLxTxRatio\\(\\)\\] The background signals for Lx and Tx appear to be similar.+")

  ## use invalid signal integral
  expect_warning(calc_TLLxTxRatio(
      Lx.data.signal,
      Lx.data.background,
      Tx.data.signal,
      Tx.data.background,
      signal_integral = 10:1000),
      "'signal_integral' out of bounds, reset to be between 10 and 250")

  ## integral_input
  expect_equal(calc_TLLxTxRatio(Lx.data.signal,
                                Lx.data.background,
                                Tx.data.signal,
                                Tx.data.background,
                                signal_integral = 378:414,
                                integral_input = "measurement")$LxTx.table,
               calc_TLLxTxRatio(Lx.data.signal,
                                Lx.data.background,
                                Tx.data.signal,
                                Tx.data.background,
                                signal_integral = 210:230,
                                integral_input = "channel")$LxTx.table)

  ## deprecated arguments
  expect_warning(calc_TLLxTxRatio(
      Lx.data.signal,
      Lx.data.background,
      Tx.data.signal,
      Tx.data.background,
      signal.integral = signal_integral),
      "was deprecated in v1.2.0, use 'signal_integral'")
  expect_warning(expect_error(calc_TLLxTxRatio(
      Lx.data.signal,
      Lx.data.background,
      Tx.data.signal,
      Tx.data.background,
      signal.integral = signal_integral,
      integral_input = "measurement"),
      "'integral_input' is not supported with old argument names"),
      "was deprecated in v1.2.0, use 'signal_integral'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  ## run function without error
  expect_snapshot_RLum(calc_TLLxTxRatio(
    Lx.data.signal,
    Lx.data.background,
    Tx.data.signal,
    Tx.data.background,
    signal_integral))

  expect_snapshot_RLum(calc_TLLxTxRatio(
    Lx.data.signal,
    Lx.data.background=NULL,
    Tx.data.signal,
    Tx.data.background=NULL,
    signal_integral))
})

#' @title Apply a simple homogeneity test after Galbraith (2003)
#'
#' @description
#' A simple homogeneity test for De estimates.
#' For details see Galbraith (2003).
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with De `(data[,1])` and De error `(values[,2])`
#'
#' @param log [logical] (*with default*):
#' perform the homogeneity test with (un-)logged data
#'
#' @param ... further arguments (`verbose`).
#'
#' @return
#' Returns a terminal output. In addition an
#' [Luminescence::RLum.Results-class]-object is returned containing the
#' following elements:
#'
#' \item{summary}{[data.frame] summary of all relevant model results.}
#' \item{data}{[data.frame] original input data}
#' \item{args}{[list] used arguments}
#' \item{call}{[call] the function call}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @section Function version: 0.3.0
#'
#' @author
#' Christoph Burow, University of Cologne (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#'
#' @seealso [stats::pchisq]
#'
#' @references
#' Galbraith, R.F., 2003. A simple homogeneity test for estimates
#' of dose obtained using OSL. Ancient TL 21, 75-77.
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## apply the homogeneity test
#' calc_HomogeneityTest(ExampleData.DeValues$BT998)
#'
#' ## using the data presented by Galbraith (2003)
#' df <-
#'  data.frame(
#'    x = c(30.1, 53.8, 54.3, 29.0, 47.6, 44.2, 43.1),
#'    y = c(4.8, 7.1, 6.8, 4.3, 5.2, 5.9, 3.0))
#'
#' calc_HomogeneityTest(df)
#'
#'
#' @export
calc_HomogeneityTest <- function(
  data,
  log = TRUE,
  ...
) {
  .set_function_name("calc_HomogeneityTest")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("data.frame", "RLum.Results"))
  .validate_not_empty(data)
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")
  }
  if (ncol(data) < 2) {
    .throw_error("'data' should have 2 columns")
  }
  data <- data[, 1:2]
  lapply(data, .validate_class, classes = c("numeric", "integer"),
         name = "All columns of 'data'")
  .validate_logical_scalar(log)

  ##==========================================================================##
  ## ... ARGUMENTS
  ##==========================================================================##
  extraArgs <- list(...)

  ## set plot main title
  verbose<- extraArgs$verbose %||% TRUE
  .validate_logical_scalar(verbose)

  ##============================================================================##
  ## CALCULATIONS
  ##============================================================================##
  if(log) {
    if (any(data < 0, na.rm = TRUE))
      .throw_warning("'data' contains negative values and 'log = TRUE', ",
                     "check your input")
    dat <- suppressWarnings(log(data))
    dat[[2]] <- data[[2]]/data[[1]]
  } else {
    dat <- data
  }

  wi <- 1 / dat[[2]] ^ 2
  wizi <- wi * dat[[1]]
  mu <- sum(wizi) / sum(wi)
  gi <- wi * (dat[[1]] - mu) ^ 2

  G <- sum(gi)
  df <- length(wi) - 1
  n <- length(wi)

  P <- stats::pchisq(G, df, lower.tail = FALSE)


  ##============================================================================##
  ## OUTPUT
  ##============================================================================##

  if(verbose) {
    cat("\n [calc_HomogeneityTest()]")
    cat("\n\n ---------------------------------")
    cat("\n n:                 ", n)
    cat("\n ---------------------------------")
    cat("\n mu:                ", round(mu,4))
    cat("\n G-value:           ", round(G,4))
    cat("\n Degrees of freedom:", df)
    cat("\n P-value:           ", round(P,4))
    cat("\n ---------------------------------\n\n")
  }

  ##============================================================================##
  ## RETURN VALUES
  ##============================================================================##
  summary <- data.frame(
    n = n,
    g.value = G,
    df = df,
    P.value = P
  )

  args <- list(log = log)

  set_RLum(
    class = "RLum.Results",
    data = list(
      summary = summary,
      data = data,
      args = args
    ),
    info = list(call = sys.call())
  )
}

##use the data example given by Galbraith (2003)
df <- data.frame(
    x = c(30.1, 53.8, 54.3, 29.0, 47.6, 44.2, 43.1),
    y = c(4.8, 7.1, 6.8, 4.3, 5.2, 5.9, 3.0))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_HomogeneityTest(TRUE),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_HomogeneityTest(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(calc_HomogeneityTest(iris[, 1, drop = FALSE]),
               "'data' should have 2 columns")
  expect_error(calc_HomogeneityTest(data.frame(1:4, letters[1:4])),
               "All columns of 'data' should be of class 'numeric' or 'integer'")
  expect_error(calc_HomogeneityTest(iris, verbose = NA),
               "'verbose' should be a single logical value")
})

test_that("check values from output example", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(temp <- calc_HomogeneityTest(df, verbose = FALSE),
                       tolerance = snapshot.tolerance)

  ## using an RLum.Results object as input
  SW({
  expect_s4_class(calc_HomogeneityTest(temp),
                  "RLum.Results")
  })

  ##test the unlogged version
  SW({
  expect_snapshot_RLum(calc_HomogeneityTest(df, log = FALSE),
                       tolerance = snapshot.tolerance)
  })

  ## negative values in data
  ## we suppress the NaN warning from log() because we throw our own warning
  expect_no_warning(expect_warning(
      calc_HomogeneityTest(data.frame(1:5, -1:3), verbose = FALSE),
      "'data' contains negative values and 'log = TRUE'"),
      message = "NaNs produced")
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 924
  expect_s4_class(
      calc_HomogeneityTest(data.frame(1:4, NA_real_), verbose = FALSE),
      "RLum.Results")

  ## issue 1224
  expect_silent(calc_HomogeneityTest(data.frame(1:4, 2:5, letters[1:4]),
                                     verbose = FALSE))
})

#' @title Create a DRAC input data template (v1.2)
#'
#' @description
#' This function returns a DRAC input template (v1.2) to be used in conjunction
#' with the [Luminescence::use_DRAC] function
#'
#' @param nrow [integer] (*with default*):
#' specifies the number of rows of the template (i.e., the number of data
#' sets you want to submit).
#'
#' @param preset [character] (*optional*):
#' By default, all values of the template are set to `NA`, which means that
#' the user needs to fill in **all** data first before submitting to DRAC
#' using [Luminescence::use_DRAC]. To reduce the number of values that need to
#' be provided, `preset` can be used to create a template with at least
#' a minimum of reasonable preset values.
#'
#' `preset` can be one of the following:
#' - `quartz_coarse`
#' - `quartz_fine`
#' - `feldspar_coarse`
#' - `polymineral_fine`
#' - `DRAC-example_quartz`
#' - `DRAC-example_feldspar`
#' - `DRAC-example_polymineral`
#'
#' Note that the last three options can be used to produce a template
#' with values directly taken from the official DRAC input `.csv` file.
#'
#' @param file_input [character] (*optional*):
#' file connection to a DRAC `.csv` file. The file will be imported and
#' translated to a template usable by [Luminescence::use_DRAC].
#' Please note that there is no check on the validity of the `.csv` file.
#'
#' @param notification [logical] (*with default*):
#' show or hide the notification.
#'
#' @return A  list of class `DRAC.list`.
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Christoph Burow, University of Cologne (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#'
#' @references
#' Durcan, J.A., King, G.E., Duller, G.A.T., 2015. DRAC: Dose Rate and Age Calculator for trapped charge dating.
#' Quaternary Geochronology 28, 54-61. doi:10.1016/j.quageo.2015.03.012
#'
#' @seealso [as.data.frame], [list]
#'
#' @examples
#'
#' # create a new DRAC input input
#' input <- template_DRAC(preset = "DRAC-example_quartz")
#'
#' # show content of the input
#' print(input)
#' print(input$`Project ID`)
#' print(input[[4]])
#'
#'
#' ## Example: DRAC Quartz example
#' # note that you only have to assign new values where they
#' # are different to the default values
#' input$`Project ID` <- "DRAC-Example"
#' input$`Sample ID` <- "Quartz"
#' input$`Conversion factors` <- "AdamiecAitken1998"
#' input$`External U (ppm)` <- 3.4
#' input$`errExternal U (ppm)` <- 0.51
#' input$`External Th (ppm)` <- 14.47
#' input$`errExternal Th (ppm)` <- 1.69
#' input$`External K (%)` <- 1.2
#' input$`errExternal K (%)` <- 0.14
#' input$`Calculate external Rb from K conc?` <- "N"
#' input$`Calculate internal Rb from K conc?` <- "N"
#' input$`Scale gammadoserate at shallow depths?` <- "N"
#' input$`Grain size min (microns)` <- 90
#' input$`Grain size max (microns)` <- 125
#' input$`Water content ((wet weight - dry weight)/dry weight) %` <- 5
#' input$`errWater content %` <- 2
#' input$`Depth (m)` <- 2.2
#' input$`errDepth (m)` <- 0.22
#' input$`Overburden density (g cm-3)` <- 1.8
#' input$`errOverburden density (g cm-3)` <- 0.1
#' input$`Latitude (decimal degrees)` <- 30.0000
#' input$`Longitude (decimal degrees)` <- 70.0000
#' input$`Altitude (m)` <- 150
#' input$`De (Gy)` <- 20
#' input$`errDe (Gy)` <- 0.2
#'
#' # use DRAC
#' \dontrun{
#' output <- use_DRAC(input)
#' }
#'
#' @export
template_DRAC <- function(
  nrow = 1L,
  preset = NULL,
  file_input = NULL,
  notification = TRUE
) {
  .set_function_name("template_DRAC")
  on.exit(.unset_function_name(), add = TRUE)

  ## TODO:
  # 1 - allow mineral specific presets; new argument 'mineral'
  # 2 - add option to return the DRAC example data set

  .validate_positive_scalar(nrow, int = TRUE)

  ## PRESETS ----
  valid_presets <- c("quartz_coarse", "quartz_fine", "feldspar_coarse", "polymineral_fine",
                     "DRAC-example_quartz", "DRAC-example_feldspar", "DRAC-example_polymineral")
  preset <- .validate_args(preset, valid_presets, null.ok = TRUE)

  ## FILE ----------
  if(!is.null(file_input)) {
    ## read csv file from DRAC
    file_input <- read.csv(
      file = file_input,
      skip = 8,
      check.names = FALSE,
      header = TRUE,
      stringsAsFactors = FALSE)[-1, ]

    ## cut columns to current template
    n_col <- length(template_DRAC(notification = FALSE))
    file_input <- file_input[,1:n_col]

    ## reset rows
    nrow <- nrow(file_input)
  }

  ## throw warning
  if (nrow > 5000)
    .throw_warning("More than 5000 datasets might not be supported")

  ## LEGAL NOTICE ----
  messages <- list("\n",
                   "\t-------------------- IMPORTANT NOTE ------------------------\n",
                   "\t This function returns a DRAC input template to be used in ",
                   "\t conjunction with the use_DRAC() function.  \n",
                   "\t The template was reproduced with great care, but we do not",
                   "\t take any responsibility and we are not liable for any ",
                   "\t mistakes or unforeseen misbehaviour.",
                   "\t Note that this template is only compatible with DRAC",
                   "\t version 1.3. Before using this template make sure that",
                   "\t this is the correct version, otherwise expect unspecified",
                   "\t errors.\n",
                   "\t Please ensure you cite the use of DRAC in your work,",
                   "\t published or otherwise. Please cite the website name and",
                   "\t version (e.g. DRAC v1.3) and the accompanying journal",
                   "\t article:",
                   "\t Durcan, J.A., King, G.E., Duller, G.A.T., 2015.",
                   "\t DRAC: Dose rate and age calculation for trapped charge",
                   "\t dating. Quaternary Geochronology 28, 54-61. \n",
                   "\t Set 'notification = FALSE' to hide this message. \n",
                   "\t-------------------- IMPORTANT NOTE ------------------------",
                   "\n")

  if (notification) lapply(messages, message)

  # CREATE TEMPLATE ----
  template <- list(
    `Project ID` =
      structure(rep(NA_character_, nrow), required = TRUE, allowsX = FALSE, default_class = "character", key = "TI:1",
                description = "Inputs can be alphabetic, numeric or selected symbols (/ - () [] _). Spaces are not permitted."), #

    `Sample ID` =
      structure(rep(NA_character_, nrow), required = TRUE, allowsX = FALSE,  default_class = "character", key = "TI:2",
                description = "Inputs can be alphabetic, numeric or selected symbols (/ - () [] _). Spaces are not permitted."), #

    Mineral =
      structure(factor(rep(NA_character_, nrow), c("Q", "F", "PM")), required = TRUE,  default_class = "factor", allowsX = FALSE, key = "TI:3",
                description = "The mineral used for dating: quartz, feldspar or polymineral. Input must be 'Q', 'F' or 'PM'."), #

    `Conversion factors` =
      structure(factor(rep(NA_character_, nrow), c("AdamiecAitken1998", "Guerinetal2011", "Liritzisetal2013", "Cresswelletal2018")), required = FALSE, allowsX = TRUE,  default_class = "factor", key = "TI:4",
                description = "The conversion factors required to calculate dose rates from radionuclide concentrations. Users have the option of datasets from Adamiec and Aitken (1998), Guerin et al. (2011), Liritzis et al. (2013) or Cresswell et al. (2018). Input must be 'AdamiecAitken1998', 'Guerinetal2011', 'Liritzisetal2013', 'Cresswelletal2018', or 'X' if conversion factors are not required."), #

    `External U (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE,  default_class = "numeric", key = "TI:5",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errExternal U (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:6",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `External Th (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:7",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errExternal Th (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:8",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `External K (%)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:9",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errExternal K (%)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:10",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `External Rb (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:11",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errExternal Rb (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:12",
                description = "Radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `Calculate external Rb from K conc?` =
      structure(factor(rep(NA_character_, nrow), c("Y", "N")), required = FALSE, default_class = "factor", allowsX = FALSE, key = "TI:13",
                description = "Option to calculate a Rubidium concentration from Potassium, using the 270:1 ratio suggested by Mejdahl (1987). Input should be yes 'Y' or no 'N'."), #

    `Internal U (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE,  default_class = "numeric", key = "TI:14",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errInternal U (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric",  key = "TI:15",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `Internal Th (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:16",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errInternal Th (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:17",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `Internal K (%)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:18",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errInternal K (%)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:19",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `Rb (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:20",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `errRb (ppm)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:21",
                description = "Internal radionuclide concentrations in parts per million for Uranium, Thorium and Rubidium and % for Potassium. Inputs must be 0 or positive and should not be left blank."), #

    `Calculate internal Rb from K conc?` =
      structure(factor(rep(NA_character_, nrow), c("Y", "N")), required = FALSE, allowsX = FALSE, default_class = "factor", key = "TI:22",
                description = "Option to calculate an internal Rubidium concentration from Potassium, using the 270:1 ratio suggested by Mejdahl (1987). Input should be yes 'Y' or no 'N'."), #

    `User external alphadoserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE,  default_class = "numeric", key = "TI:23",
                description = "Users may input directly measured values for external alpha, beta and gamma dose rates (in Gy.ka-1). Any positive inputs in these fields will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and should not be left blank."), #

    `errUser external alphadoserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:24",
                description = "Users may input directly measured values for external alpha, beta and gamma dose rates (in Gy.ka-1). Any positive inputs in these fields will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and should not be left blank."), #

    `User external betadoserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:25",
                description = "Users may input directly measured values for external alpha, beta and gamma dose rates (in Gy.ka-1). Any positive inputs in these fields will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and should not be left blank."), #

    `errUser external betadoserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:26",
                description = "Users may input directly measured values for external alpha, beta and gamma dose rates (in Gy.ka-1). Any positive inputs in these fields will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and should not be left blank."), #

    `User external gamma doserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:27",
                description = "Users may input directly measured values for external alpha, beta and gamma dose rates (in Gy.ka-1). Any positive inputs in these fields will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and should not be left blank."), #

    `errUser external gammadoserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:28",
                description = "Users may input directly measured values for external alpha, beta and gamma dose rates (in Gy.ka-1). Any positive inputs in these fields will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and should not be left blank."), #

    `User internal doserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:29",
                description = "Users may input an internal dose rate (either alpha, beta or the sum of the two; in Gy.ka-1). DRAC will assume that this value has already been corrected for attenuation. Inputs in this field will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and not left blank."), #

    `errUser internal doserate (Gy.ka-1)` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:30",
                description = "Users may input an internal dose rate (either alpha, beta or the sum of the two; in Gy.ka-1). DRAC will assume that this value has already been corrected for attenuation. Inputs in this field will override dose rates calculated from radionuclide concentrations. Inputs should be 0 or positive and not left blank."), #

    `Scale gammadoserate at shallow depths?` =
      structure(factor(rep(NA_character_, nrow), c("Y", "N")), required = FALSE, allowsX = FALSE, default_class = "factor", key = "TI:31",
                description = "Users may choose to scale gamma dose rates for samples taken within 0.3 m of the ground surface. The scaling factors of Aitken (1985) are used. Input should be yes 'Y' or no 'N'."), #

    `Grain size min (microns)` =
      structure(rep(NA_integer_, nrow), required = TRUE, allowsX = FALSE, default_class = "integer", key = "TI:32",
                description = "The grain size range analysed. DRAC can be used for the grain size ranges between 1 and 1000 microns. Inputs should range between 1 and 1000 and not be left blank."), #

    `Grain size max (microns)` =
      structure(rep(NA_integer_, nrow), required = TRUE, allowsX = FALSE, default_class = "integer", key = "TI:33",
                description = "The grain size range analysed. DRAC can be used for the grain size ranges between 1 and 1000 microns. Inputs should range between 1 and 1000 and not be left blank."), #

    `alpha-Grain size attenuation` =
      structure(factor(rep(NA_character_, nrow), c("Bell1980", "Brennanetal1991", "Martinetal2014")), required = TRUE, allowsX = FALSE,  default_class = "factor", key = "TI:34",
                description = "The grain size attenuation factors for the alpha dose rate. Users have the option of datasets from Bell (1980), Brennan et al. (1991), or Martin et al. (2014). Valid inputs are 'Bell1980' (between 1 and 1000 microns), 'Brennanetal1991' (between 1 and 1000 microns) or 'Martinetal2014' (between 1 and 10000 microns)."), #

    `beta-Grain size attenuation ` =
      structure(factor(rep(NA_character_, nrow), c("Mejdahl1979", "Brennan2003", "Guerinetal2012-Q", "Guerinetal2012-F")), required = TRUE, allowsX = FALSE,  default_class = "factor", key = "TI:35",
                description = "The grain size attenuation factors for the beta dose rate. Users have the option of datasets from Mejdahl (1979), Brennan (2003) and Guerin et al. (2012) for quartz or feldspar. Valid inputs are 'Mejdahl1979' (between 1 and 10000 microns), 'Brennan2003' (between 1 and 10000 microns), 'Guerinetal2012-Q' (between 1 and 1000 microns) or 'Guerinetal2012-F' (between 1 and 1000 microns)."), #

    `Etch depth min (microns)` =
      structure(rep(NA_integer_, nrow), required = TRUE, allowsX = FALSE,  default_class = "integer", key = "TI:36",
                description = "The user defined etch depth range (microns). Inputs should range between 0 and 30 and not be left blank."), #

    `Etch depth max (microns)` =
      structure(rep(NA_integer_, nrow), required = TRUE, allowsX = FALSE, default_class = "integer", key = "TI:37",
                description = "The user defined etch depth range (microns). Inputs should range between 0 and 30 and not be left blank."), #

    `beta-Etch depth attenuation factor` =
      structure(factor(rep(NA_character_, nrow), c("Bell1979", "Brennan2003")), required = FALSE, allowsX = TRUE, default_class = "factor", key = "TI:38",
                description = "The etch depth attenuation factors for the beta dose rate. Users have the option of datasets from Bell (1979) and Brennan (2003). Input must be 'Bell1979' or 'Brennan2003'. Note: only the dataset of Bell (1980) is provided for attenuation of the alpha dose rate by etching."), #

    `a-value` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:39",
                description = "Alpha track efficiency value and uncertainty defined by the user. Inputs should be 0 or positive and not left blank."), #

    `erra-value` =
      structure(rep(NA_real_, nrow), required = TRUE, allowsX = TRUE, default_class = "numeric", key = "TI:40",
                description = "Alpha track efficiency value and uncertainty defined by the user. Inputs should be 0 or positive and not left blank."), #

    `Water content ((wet weight - dry weight)/dry weight) %` =
      structure(rep(NA_real_, nrow), required = TRUE, allowsX = FALSE, default_class = "numeric", key = "TI:41",
                description = "Sediment water content (%) over the burial period. Inputs should be 0 or positive and not be left blank."), #

    `errWater content %` =
      structure(rep(NA_real_, nrow), required = FALSE, allowsX = FALSE, default_class = "numeric", key = "TI:42",
                description = "Sediment water content (%) over the burial period. Inputs should be 0 or positive and not be left blank."), #

    `Depth (m)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:43",
                description = "Depth and uncertainty from which sample was extracted beneath the ground surface. Inputs should be 0 or positive and not left blank."), #

    `errDepth (m)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:44",
                description = "Depth and uncertainty from which sample was extracted beneath the ground surface. Inputs should be 0 or positive and not left blank."), #

    `Overburden density (g cm-3)` =
      structure(rep(NA_real_, nrow), required = TRUE, allowsX = FALSE, default_class = "numeric", key = "TI:45",
                description = "Density of the overlying sediment matrix from which the sample was taken. Inputs should be 0 or positive and not be left blank. The scaling calculation will use the overburden density and uncertainty provided."), #

    `errOverburden density (g cm-3)` =
      structure(rep(NA_real_, nrow), required = TRUE, allowsX = FALSE, default_class = "numeric", key = "TI:46",
                description = "Density of the overlying sediment matrix from which the sample was taken. Inputs should be 0 or positive and not be left blank. The scaling calculation will use the overburden density and uncertainty provided."), #

    `Latitude (decimal degrees)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:47",
                description = "Latitude and longitude of sample location (in degree decimals). Positive values should be used for northern latitudes and eastern longitudes and negative values for southern latitudes and western longitudes. Inputs should range from -90 to 90 degrees for latitudes and -180 to 180 degrees for longitude."), #

    `Longitude (decimal degrees)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:48",
                description = "Latitude and longitude of sample location (in degree decimals). Positive values should be used for northern latitudes and eastern longitudes and negative values for southern latitudes and western longitudes. Inputs should range from -90 to 90 degrees for latitudes and -180 to 180 degrees for longitude."), #

    `Altitude (m)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:49",
                description = "Altitude of sample location in metres above sea level. Input should be less than 5000 and not left blank."), #

    `User cosmicdoserate (Gy.ka-1)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:50",
                description = "Users may input a cosmic dose rate (in Gy.ka-1). Inputs in these fields will override the DRAC calculated cosmic dose rate. Inputs should be positive or 'X' if not required, and not left blank."), #

    `errUser cosmicdoserate (Gy.ka-1)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:51",
                description = "Users may input a cosmic dose rate (in Gy.ka-1). Inputs in these fields will override the DRAC calculated cosmic dose rate. Inputs should be positive or 'X' if not required, and not left blank."), #

    `De (Gy)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:52",
                description = "Sample De and uncertainty (in Gy). Inputs should be positive or 'X' if not required, and not left blank."), #

    `errDe (Gy)` =
      structure(rep(NA_character_, nrow), required = FALSE, allowsX = TRUE, default_class = "numeric", key = "TI:53",
                description = "Sample De and uncertainty (in Gy). Inputs should be positive or 'X' if not required, and not left blank.") #
  )

  ## RETURN VALUE ----
  # add an additional DRAC class so we can define our own S3 method for as.data.frame
  class(template) <- c("DRAC.list", "list")

  # set preset
  if (!is.null(preset) && is.null(file_input))
    template <- .preset_DRAC(template, preset)

  ## just fill the template
  if(!is.null(file_input)) {
    for(i in seq_along(file_input))
      template[[i]] <- file_input[[i]]
  }

  ## return
  invisible(template)
}

.preset_DRAC <- function(x, preset) {
  preset_list <- list(
    ## DRAC COLUMNS (TI:xx) ---       TI:1               2             3         4              5    6      7     8     9     10  11 12  13   14   15   16   17    18   19   20   21   22   23   24   25    26   27   28   29   30   31    32   33                 34                 35  36  37         38      39      40  41 42    43    44   45   46  47   48   49   50   51      52    53
      quartz_coarse = list("RLum_preset", "quartz_coarse", "Q", "Guerinetal2011", "X", "X", "X",
                           "X", "X", "X", "X", "X", "N", "X", "X", "X", "X", "X", "X", "X", "X",
                           "N", "X", "X", "X", "X", "X", "X", "X", "X", "Y", 100L, 200L,
                           "Brennanetal1991", "Guerinetal2012-Q", 20L, 5L, "Bell1979",
                           0.035, 0.01, 0.0, 0.0, 0.0, 0.0, 1.8, 0.1, "X", "X", 0, "X", "X", "X", "X"),
      quartz_fine = list("RLum_preset", "quartz_fine", "Q", "Guerinetal2011", "X", "X", "X", "X",
                         "X", "X", "X", "X", "N", "X", "X", "X", "X", "X", "X", "X", "X", "N", "X",
                         "X", "X", "X", "X", "X", "X", "X", "Y", 4L, 11L, "Brennanetal1991",
                         "Guerinetal2012-Q", 0L, 0L, "Bell1979", 0.035, 0.01, 0, 0, 0, 0, 1.8, 0.1,
                         "X", "X", 0, "X", "X", "X", "X"),
      feldspar_coarse = list("RLum_preset", "feldspar_coarse", "F", "Guerinetal2011", "X", "X", "X",
                             "X", "X", "X", "X", "X", "Y", "X", "X", "X", "X", 12.5, 0.5, "X", "X",
                             "N", "X", "X", "X", "X", "X", "X", "X", "X", "Y", 100L, 200L,
                             "Brennanetal1991", "Guerinetal2012-F", 0L, 0L, "Bell1979", 0.08, 0.01,
                             0, 0, 0, 0, 1.8, 0.1, "X", "X", 0, "X", "X", "X", "X"),
      polymineral_fine = list("RLum_preset", "polymineral_fine", "PM", "Guerinetal2011", "X", "X",
                              "X", "X", "X", "X", "X", "X", "Y", "X", "X", "X", "X", 12.5, 0.5,
                              "X", "X", "N", "X", "X", "X", "X", "X", "X", "X", "X", "Y", 4L, 11L,
                              "Brennanetal1991", "Guerinetal2012-F", 0L, 0L, "Bell1979", 0.08, 0.01,
                              0, 0, 0, 0, 1.8, 0.1, "X", "X", 0, "X", "X", "X", "X"),
      `DRAC-example_quartz` = list("DRAC-example", "Quartz", "Q", "Cresswelletal2018", 3.4, 0.51, 14.47, 1.69, 1.2, 0.14, 0, 0, "N", "X", "X", "X", "X", "X", "X", "X", "X", "N", "X", "X", "X", "X", "X", "X", "X", "X", "N", 90L, 125L, "Brennanetal1991", "Guerinetal2012-Q", 8L, 10L, "Bell1979", 0, 0, 5, 2, 2.22, 0.05, 1.8, 0.1, 30, 70, 150, "X", "X", 20, 0.2),
      `DRAC-example_feldspar` = list("DRAC-example", "Feldspar", "F", "AdamiecAitken1998", 2, 0.2, 8,  0.4, 1.75, 0.05, 0, 0, "Y", "X", "X", "X", "X", 12.5, 0.5, "X", "X", "N", "X", "X", "X", "X", "X", "X", "X", "X", "Y", 180L, 212L, "Martinetal2014", "Mejdahl1979", 0L, 0L, "Bell1979", 0.15, 0.05, 10, 3, 0.15, 0.02, 1.8, 0.1, 60, 100, 200, "X", "X", 15, 1.5),
      `DRAC-example_polymineral` = list("DRAC-example", "Polymineral", "PM", "AdamiecAitken1998", 4, 0.4, 12, 0.12, 0.83, 0.08, 0, 0, "Y", "X", "X", "X", "X", 12.5, 0.5, "X", "X", "N", "X", "X", 2.5, 0.15, "X", "X", "X", "X", "Y", 4L, 11L, "Bell1980", "Mejdahl1979", 0L, 0L, "Bell1979", 0.086, 0.0038, 10, 5, 0.2, 0.02, 1.8, 0.1, 46, 118, 200, 0.2, 0.1, 204.47, 2.69)
  )

  n <- length(x[[1]])
  for (i in 1:length(x))
    x[[i]] <- rep(preset_list[[preset]][[i]], n)

  return(x)
}

##Full check
test_that("Check template creation ", {
  testthat::skip_on_cran()

  ## test output class
  SW({
  expect_message(res <- template_DRAC(),
                 "IMPORTANT NOTE")
  })
  expect_s3_class(res, "DRAC.list")
  expect_s3_class(template_DRAC(notification = FALSE), "DRAC.list")
  expect_s3_class(template_DRAC(nrow = 10, notification = FALSE), "DRAC.list")

  ## test presets
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "quartz_coarse")$`a-value`), 0.035)
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "quartz_fine")$`a-value`), 0.035)
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "feldspar_coarse")$`a-value`), 0.08)
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "polymineral_fine")$`a-value`), 0.08)
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "DRAC-example_quartz")$`De (Gy)`), 20)
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "DRAC-example_feldspar")$`De (Gy)`), 15)
  expect_identical(as.numeric(template_DRAC(notification = FALSE, preset = "DRAC-example_polymineral")$`De (Gy)`), 204.47)

  expect_true(
    do.call(all.equal, as.list(template_DRAC(nrow = 2, notification = FALSE, preset = "DRAC-example_quartz")$`De (Gy)`))
  )

  ## misc tests
  expect_true(all(is.na(template_DRAC(notification = FALSE))))
  expect_true(!all(is.na(template_DRAC(preset = "DRAC-example_polymineral", notification = FALSE))))
  expect_equal(length(template_DRAC(notification = FALSE)), 53)
  expect_equal(length(template_DRAC(nrow = 10, notification = FALSE)[[1]]), 10)

  ## use the file_input option
  SW({
    tmp_file <- tempfile(fileext = ".csv")
    write.csv(x = as.data.frame(template_DRAC(nrow = 12, preset = 'quartz_coarse')), file = tmp_file, row.names = FALSE)
    t <- expect_s3_class(suppressWarnings(template_DRAC(file_input = tmp_file, notification = FALSE)), "DRAC.list")
    expect_length(t[[1]], 3)
  })

  ## expect failure
  expect_error(template_DRAC(nrow = -1),
               "'nrow' should be a single positive integer value")
  expect_error(template_DRAC("preset"),
               "'nrow' should be a single positive integer value")
  expect_warning(template_DRAC(nrow = 5001, notification = FALSE),
                 "[template_DRAC()] More than 5000 datasets might not be supported",
                 fixed = TRUE)
  expect_error(template_DRAC(preset = "does_not_exist"),
               "'preset' should be one of 'quartz_coarse', 'quartz_fine'")
  expect_error(template_DRAC(preset = c("does_not_exist", "neither_this_one")),
               "'preset' contains multiple values but not all of them match 'choices'")
  expect_error(template_DRAC(preset = 999),
               "'preset' should be one of 'quartz_coarse', 'quartz_fine'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## generate test dataset with DRAC
  DRAC <- template_DRAC(preset = "quartz_coarse", notification = FALSE)
  t <- expect_type(.get_DRAC_references(DRAC), "list")
  expect_length(t, 2)
  expect_length(t$refs, 7)

  ## check table
  DRAC <- as.data.frame(DRAC)
  t <- expect_type(.get_DRAC_references(DRAC), "list")
  expect_length(t, 2)
  expect_length(t$refs, 7)

  ## more coverage
  DRAC[["TI:4"]] <- DRAC[["TI:34"]] <- DRAC[["TI:35"]] <- DRAC[["TI:38"]] <- "X"
  DRAC[["TI:22"]] <- "Y"
  t <- expect_type(.get_DRAC_references(DRAC), "list")
  expect_length(t$refs, 4)
})

#' @title Merge Risoe.BINfileData objects or Risoe BIN-files
#'
#' @description
#' The function allows merging Risoe BIN/BINX files or [Luminescence::Risoe.BINfileData-class]
#' objects.
#'
#' @details
#' The function allows merging different measurements to one file or one
#' object. The record IDs are recalculated for the new object. Other values
#' are kept for each object. The number of input objects is not limited.
#'
#' `position.number.append.gap` option
#'
#' If the option `keep.position.number = FALSE` is used, the position
#' numbers of the new data set are recalculated by adding the highest position
#' number of the previous data set to the each position number of the next data
#' set. For example: The highest position number is 48, then this number will
#' be added to all other position numbers of the next data set (e.g. 1 + 48 =
#' 49)
#'
#' However, there might be cases where an additional addend (summand) is needed
#' before the next position starts. Example:
#'
#' - Position number set (A): `1,3,5,7`
#' - Position number set (B): `1,3,5,7`
#'
#' With no additional summand the new position numbers would be:
#' `1,3,5,7,8,9,10,11`. That might be unwanted. Using the argument
#' `position.number.append.gap = 1` it will become:
#' `1,3,5,7,9,11,13,15,17`.
#'
#' @param objects [character] or [Luminescence::Risoe.BINfileData-class] (**required**):
#' Character vector with path and files names with ".bin" or ".binx" extension
#' (e.g. `input.objects = c("path/file1.bin", "path/file2.bin")` or a list of
#' [Luminescence::Risoe.BINfileData-class] objects (e.g. `input.objects = c(object1, object2)`).
#'
#' @param output.file [character] (*optional*):
#' File output path and name. If no value is given, a [Luminescence::Risoe.BINfileData-class]
#' object returned instead of a file.
#'
#' @param keep.position.number [logical] (*with default*):
#' Allows keeping the original position numbers of the input objects.
#' Otherwise the position numbers are recalculated.
#'
#' @param position.number.append.gap [integer] (*with default*):
#' Set the position number gap between merged BIN-file sets, if the option
#' `keep.position.number = FALSE` is used. See details for further
#' information.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... currently not used.
#'
#' @return
#' Returns a [Luminescence::Risoe.BINfileData-class] object or writes to the BIN-file
#' specified by `output.file`.
#'
#' @note
#' The validity of the output objects is not further checked.
#'
#' @section Function version: 0.2.11
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData-class], [Luminescence::read_BIN2R], [Luminescence::write_R2BIN]
#'
#' @references
#' Duller, G.A.T., 2007. Analyst (Version 3.24) (manual). Aberystwyth University, Aberystwyth.
#'
#' @keywords IO manip
#'
#' @examples
#'
#' ##merge two objects
#' data(ExampleData.BINfileData, envir = environment())
#'
#' object1 <- CWOSL.SAR.Data
#' object2 <- CWOSL.SAR.Data
#'
#' object.new <- merge_Risoe.BINfileData(c(object1, object2))
#'
#' @export
merge_Risoe.BINfileData <- function(
  objects,
  output.file,
  keep.position.number = FALSE,
  position.number.append.gap = 0,
  verbose = TRUE,
  ...
) {
  .set_function_name("merge_Risoe.BINfileData")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("input.objects" %in% ...names()) {
    objects <- list(...)$input.objects
    .deprecated(old = "input.objects", new = "objects", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(objects, c("character", "list"))
  .validate_logical_scalar(keep.position.number)
  .validate_logical_scalar(verbose)
  if (length(objects) < 2) {
    .throw_message("At least two input objects are needed, nothing done",
                   error = FALSE)
    return(objects)
  }

  ## Import files -----------------------------------------------------------
  if (is.character(objects)) {
    for (i in 1:length(objects)) {
      .validate_file(objects[[i]], ext = c("bin", "binx"),
                     scan.dir = FALSE, verbose = verbose)
    }
    temp <- lapply(objects, read_BIN2R, verbose = verbose)

  }else{
    for (i in 1:length(objects)) {
      .validate_class(objects[[i]], "Risoe.BINfileData",
                      name = "All elements of 'objects'")
    }
    temp <- objects
  }

  # Get POSITION values -------------------------------------------------------

  ##grep maximum position value from the first file
  temp.position.max <- max(temp[[1]]@METADATA[["POSITION"]])

  ##grep all position values except from the first file
  temp.position.values <- unlist(sapply(2:length(temp), function(x){
    temp <- temp[[x]]@METADATA[["POSITION"]] +
      temp.position.max +
      position.number.append.gap

    assign(x = "temp.position.max", value = max(temp), envir = parent.env(environment()))

    return(temp)
  }))

  temp.position.values <- c(temp[[1]]@METADATA[["POSITION"]], temp.position.values)

  # Get overall record length -----------------------------------------------
  temp.record.length <- sum(sapply(temp, function(x) nrow(x@METADATA)))

  # Merge Files -------------------------------------------------------------
  temp.new.METADATA <- temp[[1]]@METADATA
  temp.new.DATA <- temp[[1]]@DATA
  temp.new.RESERVED <- if (".RESERVED" %in% slotNames(temp[[1]]))
                         temp[[1]]@.RESERVED else list()

  ## loop over the remaining input objects
  for (i in 2:length(objects)) {
    temp.new.METADATA <- rbind(temp.new.METADATA, temp[[i]]@METADATA)
    temp.new.DATA <- c(temp.new.DATA, temp[[i]]@DATA)
    if (".RESERVED" %in% slotNames(temp[[i]])) {
      temp.new.RESERVED <- c(temp.new.RESERVED, temp[[i]]@.RESERVED)
    }
  }

  ##SET RECORD ID in METADATA
  temp.new.METADATA$ID <- 1:temp.record.length

  ##SET POSITION VALUES
  if (!keep.position.number) {
    temp.new.METADATA$POSITION <- temp.position.values
  }

  ##TODO version number?
  # Produce BIN file object -------------------------------------------------
  temp.new <- set_Risoe.BINfileData(
    METADATA = temp.new.METADATA,
    DATA = temp.new.DATA,
    .RESERVED = temp.new.RESERVED
  )

  # OUTPUT ------------------------------------------------------------------
  if (missing(output.file))
    return(temp.new)

  write_R2BIN(temp.new, output.file, verbose = verbose)
}

##Full check
test_that("Test merging", {
  testthat::skip_on_cran()
  local_edition(3)

  ##expect error
  expect_error(merge_Risoe.BINfileData(input.objects = "data"))
  expect_error(merge_Risoe.BINfileData(input.objects = c("data", "data2")))
  expect_error(merge_Risoe.BINfileData(input.objects = list("data", "data2")), regexp = "[merge_Risoe.BINfileData()] Input list does not contain Risoe.BINfileData objects!", fixed = TRUE)

  ##expect success
  data(ExampleData.BINfileData, envir = environment())
  object1 <- CWOSL.SAR.Data
  object2 <- CWOSL.SAR.Data
  expect_s4_class(merge_Risoe.BINfileData(c(object1, object2)), "Risoe.BINfileData")

})

## load data
data(ExampleData.BINfileData, envir = environment())
o1 <- subset(CWOSL.SAR.Data, LTYPE == c("IRSL", "TL"))
o2 <- subset(CWOSL.SAR.Data, LTYPE == "OSL" & POSITION < 3)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(merge_Risoe.BINfileData(c("data", "data2")),
               "File 'data' does not exist")
  expect_error(merge_Risoe.BINfileData(list("data", "data2")),
               "All elements of 'objects' should be of class 'Risoe.BINfileData'")
  expect_error(merge_Risoe.BINfileData(c(FALSE, FALSE)),
               "'objects' should be of class 'character' or 'list'")
  expect_error(merge_Risoe.BINfileData(c(o1, o2), verbose = NA),
               "'verbose' should be a single logical value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## nothing done
  input <- "data"
  expect_message(res <- merge_Risoe.BINfileData(objects = input),
                 "At least two input objects are needed, nothing done")
  expect_equal(res, input)

  ## write to an output file and read it back in
  binx <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  output.file <- tempfile(fileext = ".binx")
  merge_Risoe.BINfileData(c(binx, binx), output.file, verbose = FALSE)
  expect_true(file.exists(output.file))
  expect_s4_class(new <- read_BIN2R(output.file, verbose = FALSE),
                  "Risoe.BINfileData")
  expect_length(new, 4)

  ## deprecated argument
  expect_warning(merge_Risoe.BINfileData(input.objects = c(o1, o2)),
                 "'input.objects' was deprecated in v1.2.0, use 'objects'")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_Risoe(merge_Risoe.BINfileData(c(o1, o2)))
})

#' @title Dead-time (linearity) Correction for Photomultiplier tubes (PMT)
#'
#' @description
#' Correct dead-time (also linearity) of PMT counts to avoid saturation effects, depending
#' on pulse-pair-resolution of individual PMTs.
#'
#' @details
#' We correct for count linearity using a well-known formula that can be
#' found for example in the Hamamatsu Photomultiplier handbook
#' (Hamamatsu Photonics K.K., 2017):
#'
#' \deqn{N = \frac{M}{1 - Mt}}
#'
#' where \eqn{N} (in s\eqn{^{-1}}) is the true count rate, \eqn{M} (in s\eqn{^{-1}}) the measured count rate,
#' and \eqn{t} (in s) the pulse pair resolution.
#'
#' @param object [Luminescence::RLum.Analysis-class] [Luminescence::RLum.Data.Curve-class] (**required**):
#' object with records to correct; can be a [list] of such objects
#'
#' @param PMT_pulse_pair_resolution [numeric] (*with default*): pulse-pair resolution
#' in ns. Values can be found on the PMT datasheets. If `NULL` nothing is done.
#'
#' @section Function version: 0.1.1
#'
#' @returns
#' Returns the same type of object type as `object`.
#'
#' @note
#' This function is an adaptation of core from the R package
#' 'OSLdecomposition'.
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Dirk Mittelstrass, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @references
#'
#' Hamamatsu Photonics K.K., 2017. Photomultiplier Tubes: Basics and Applications, 4th edition. ed. Hamamatsu.
#'
#' Mittelstraß, D., Kreutzer, S., Schmidt, C., 2022. OSLdecomposition: Signal
#' component analysis for optically stimulated luminescence. \doi{10.32614/CRAN.package.OSLdecomposition}
#'
#' @examples
#' o <- set_RLum("RLum.Data.Curve")
#' correct_PMTLinearity(o, PMT_pulse_pair_resolution = 10)
#'
#' @keywords manip
#'
#' @md
#' @export
correct_PMTLinearity <- function(
    object,
    PMT_pulse_pair_resolution = NULL
) {
  ## set function name
  .set_function_name("correct_PMTLinearity")
  on.exit(.unset_function_name(), add = TRUE)

# Self-call ---------------------------------------------------------------
  if(inherits(object, "list")) {
    ## remove non-RLum objects and run
    return(lapply(
      X = .rm_nonRLum(object),
      FUN = correct_PMTLinearity,
      PMT_pulse_pair_resolution = PMT_pulse_pair_resolution))
  }

  ## input validation
  .validate_class(object, c("RLum.Analysis", "RLum.Data.Curve"))
  .validate_positive_scalar(PMT_pulse_pair_resolution, null.ok = TRUE)

  ## run only of parameter was set
  if (is.null(PMT_pulse_pair_resolution))
    return(object)

  ## extract records
  if (inherits(object, "RLum.Analysis"))
    records <- object@records
  else
    records <- list(object)

  ## convert to seconds
  res <- PMT_pulse_pair_resolution[1] * 1e-9

  ## set threshold for correction (taken from OSLdecomposition)
  ## function OSLdecomposition::RLum.OSL_correction() (v1.0.0)
  thresh <- -0.5 * 0.5 + 0.5 * sqrt(0.5^2 + 2 / res)

  ## loop over records
  records <- lapply(records, \(x) {
    ## skip non-supported records
    ## no further test to keep the speed up
    if (!inherits(x, "RLum.Data.Curve"))
      return(x)

    ## access the slot
    rec <- x@data

    ## get channel width (we assume constant channel width)
    ## if only one channel set width to 1
    if (nrow(rec) >= 2)
      w <- rec[2, 1] - rec[1, 1]
    else
      w <- 1

    ## fall back for channel resolution that is "0"
    if(w == 0)
      return(x)

    ## get cps
    cps <- rec[, 2, drop = FALSE] / w

    ## define threshold and tag
    idx <- cps > thresh

    ## correct if required
    if (any(idx)) {
      new_vals <- round(cps[idx] / (1 - cps[idx] * res) * w)
      rec[idx, 2] <- new_vals
      x@data <- rec
    }
    return(x)
  })

  ## extract records
  if (inherits(object, "RLum.Analysis"))
    object@records <- records
  else
    object <- records[[1]]

  return(object)
}

## load data
tmp <- matrix(data = c(seq(0, 1, 0.1), seq(10000, 1000, length.out = 11)),
              ncol = 2)
o0 <- set_RLum("RLum.Analysis",
               records = list(set_RLum("RLum.Data.Curve")))
o1 <- set_RLum("RLum.Analysis",
              records = list(set_RLum("RLum.Data.Curve", data = tmp)))
o2 <- set_RLum("RLum.Analysis",
               records = list(set_RLum("RLum.Data.Curve", data = tmp),
                              set_RLum("RLum.Data.Image")))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(
      object = correct_PMTLinearity("error", PMT_pulse_pair_resolution = 10),
      regexp = "'object' should be of class 'RLum.Analysis'")
  expect_error(correct_PMTLinearity(set_RLum("RLum.Data.Curve"),
                                    PMT_pulse_pair_resolution = NA),
               "'PMT_pulse_pair_resolution' should be a single positive value or NULL")
})

test_that("Test internals", {
  testthat::skip_on_cran()

  ## try RLum.Data.Curve
  o <- set_RLum("RLum.Data.Curve")
  expect_s4_class(
      object = correct_PMTLinearity(o, PMT_pulse_pair_resolution = 10),
      class = "RLum.Data.Curve")

  ## try special case with zero channel resolution that would create NA values
  data <- structure(c(24.24, 24.24, 24.25, 24.23, 24.24, 24.23, 24.24,
                      24.29, 24.29, 24.29, 24.31, 99, 101.9, 104.8, 107.7, 110.6, 113.5,
                      116.4, 119.3, 122.2, 125.1, 128), dim = c(11L, 2L), dimnames = list(
                        NULL, c("temperature.values", "count.values")))

  object <- set_RLum("RLum.Data.Curve", data = data)
  t <- expect_s4_class(
    correct_PMTLinearity(object, PMT_pulse_pair_resolution = 18),
    "RLum.Data.Curve")
  expect_type(t@data[,2], "double")

  ## nothing done
  expect_equal(correct_PMTLinearity(o),
               o)

  ## run with only one row
  t <- expect_s4_class(
      object = correct_PMTLinearity(o0, PMT_pulse_pair_resolution = 10),
      class = "RLum.Analysis")

  ## intermix different records; the non-RLum.Data.Curve() should be skipped
  t <- expect_s4_class(
      object = correct_PMTLinearity(o2, PMT_pulse_pair_resolution = 10),
      class = "RLum.Analysis")
  expect_length(t, 2)
  expect_equal(t@records[[1]][1,2], 10010, tolerance = 0.001)

  ## test list case
  t <- expect_type(
    object = correct_PMTLinearity(list(o2, o2, "not correct"),
                                  PMT_pulse_pair_resolution = 10),
    type = "list")
  expect_length(t, 2)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  res <- correct_PMTLinearity(o1, PMT_pulse_pair_resolution = 10)
  ## reset originator to avoid failures during the test_coverage workflow in CI
  res@records[[1]]@originator <- NA_character_
  expect_snapshot_RLum(res, tolerance = snapshot.tolerance)
})

#' @title Merge function for RLum.Data.Spectrum S4 class objects
#'
#' @description
#' This function allows to merge [Luminescence::RLum.Data.Spectrum-class] objects in
#' different ways without modifying the original objects.
#'
#' @details
#'
#' **Supported merge operations are:**
#'
#' `"mean"` (default)
#'
#' The mean over the cell values is calculated using the function
#' [rowMeans].
#'
#' `"median"`
#'
#' The median over the cell values is calculated using the function
#' [matrixStats::rowMedians].
#'
#' `"sum"`
#'
#' All cell values will be summed up using the function [rowSums].
#'
#' `"sd"`
#'
#' The standard deviation over the cell values is calculated using the function
#' [matrixStats::rowSds].
#'
#' `"var"`
#'
#' The variance over the cell values is calculated using the function
#' [matrixStats::rowVars].
#'
#' `"min"`
#'
#' The min values from the cell values is chosen using the function
#' [matrixStats::rowMins][matrixStats::rowRanges].
#'
#' `"max"`
#'
#' The max values from the cell values is chosen using the function
#' [matrixStats::rowMins][matrixStats::rowRanges].
#'
#' `"append"` (only for [Luminescence::RLum.Data.Curve-class])
#'
#' Appends cell values of all curves to one combined data curve. The channel width
#' is automatically re-calculated, but requires a constant channel width of the
#' original data.
#'
#' `"-"`
#'
#' The cell sums of the last objects are subtracted from the first object.
#'
#' `"*"`
#'
#' The cell sums of the last objects are multiplied with the first object.
#'
#' `"/"`
#'
#' Values of the first object are divided by cell sums of the last objects.
#'
#' @param object [list] of [Luminescence::RLum.Data.Spectrum-class] (**required**):
#' list of objects to be merged.
#'
#' @param merge.method [character] (**required**):
#' method for combining of the objects, e.g. `'mean'` (default), `'median'`,
#' `'sum'`, see details for
#' further information and allowed methods.  Note: Elements in slot info will
#' be taken from the first object in the list.
#'
#' @param method.info [numeric] (*optional*):
#' allows to specify how info elements of the input objects are combined,
#' e.g. `1` means that just the elements from the first object are kept,
#' `2` keeps only the info elements from the 2 object etc.
#' If set to `NULL`, all elements are combined.
#'
#' @param max.temp.diff [numeric] (*with default*):
#' maximum difference in the time/temperature values between the spectra to
#' be merged: when differences exceed this threshold value, the merging
#' occurs but a warning is raised.
#'
#' @return Returns an [Luminescence::RLum.Data.Spectrum-class] object.
#'
#' @note
#' The information from the slot `recordType` is taken from the first
#' object in the input list. The slot
#' 'curveType' is filled with the name `merged`.
#'
#' @section S3-generic support:
#'
#' This function is fully operational via S3-generics:
#' ``+``, ``-``, ``/``, ``*``, `merge`
#'
#' @section Function version: 0.1.2
#'
#' @author
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::merge_RLum], [Luminescence::RLum.Data.Spectrum-class]
#'
#' @keywords utilities internal
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.XSYG, envir = environment())
#'
#' ## plot single curve
#' plot_RLum(TL.Spectrum)
#'
#' ## sum two copies of the same curve
#' merged <- merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
#'                                    merge.method = "sum")
#' plot_RLum(merged)
#'
#' @export
merge_RLum.Data.Spectrum <- function(
  object,
  merge.method = c("mean", "median", "sum", "sd", "var", "min", "max",
                   "append", "-", "*", "/"),
  method.info = NULL,
  max.temp.diff = 0.1
) {
  .set_function_name("merge_RLum.Data.Spectrum")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, "list")

  ## check if object is of a supported RLum.Data class
  num.objects <- length(object)
  temp.recordType.test <- sapply(object, function(x) {
    .validate_class(x, "RLum.Data.Spectrum",
                    name = "All elements of 'object'")
    return(x@recordType)
  })

  ## check for similar record types
  record.types <- unique(temp.recordType.test)
  if (length(record.types) > 1) {
    .throw_error("Objects cannot be merged, different record types found: ",
                 .collapse(record.types))
  }

  merge.method <- .validate_args(merge.method,
                                 c("mean", "median", "sum", "sd", "var",
                                   "min", "max", "append", "-", "*", "/"))
  .validate_positive_scalar(method.info, int = TRUE, null.ok = TRUE)
  .validate_positive_scalar(max.temp.diff)

  ## Merge objects ----------------------------------------------------------

  ## perform additional checks
  check.rows <- vapply(object, function(x) nrow(x@data), numeric(1))
  check.cols <- vapply(object, function(x) ncol(x@data), numeric(1))
  if (length(check.rows) == 0 || length(check.cols) == 0) {
    .throw_error("'object' contains no data")
  }
  if (length(unique(check.rows)) > 1 || length(unique(check.cols)) > 1) {
    .throw_error("'RLum.Data.Spectrum' objects of different size ",
                 "cannot be merged")
  }

  ## collect the spectrum data from all objects
  x.vals <- rownames(object[[1]]@data)
  y.vals <- as.numeric(colnames(object[[1]]@data))
  cameraType <- object[[1]]@info$cameraType
  temp.matrix <- sapply(object, function(x) {
    ## row names must match exactly
    if (!identical(rownames(x@data), x.vals))
      .throw_error("'RLum.Data.Spectrum' objects with different channels ",
                   "cannot be merged")

    ## check the camera type
    if (!identical(x@info$cameraType, cameraType))
      .throw_error("'RLum.Data.Spectrum' objects from different camera types",
                   "cannot be merged")

    ## for time/temperature data we allow some small differences: we report
    ## a warning if they are too high, but continue anyway
    if (!is.null(colnames(x@data))) {
      if (max(abs(as.numeric(colnames(x@data)) - y.vals)) > max.temp.diff) {
        .throw_warning("The time/temperatures recorded are too different, ",
                       "proceed with caution")
      }
    }

    x@data
  })

  ## reshape all spectrum data into a 3D array
  num.rows <- check.rows[1]
  num.cols <- check.cols[1]
  temp.matrix <- array(temp.matrix, c(num.rows, num.cols, num.objects))

  temp.matrix <- switch(merge.method,
                        sum = apply(temp.matrix, 2, rowSums),
                        mean = apply(temp.matrix, 2, rowMeans),
                        median = apply(temp.matrix, 2, matrixStats::rowMedians),
                        sd = apply(temp.matrix, 2, matrixStats::rowSds),
                        var = apply(temp.matrix, 2, matrixStats::rowVars),
                        max = apply(temp.matrix, 2, matrixStats::rowMaxs),
                        min = apply(temp.matrix, 2, matrixStats::rowMins),
                        append = array(temp.matrix, c(num.rows, num.cols * num.objects)),
                        "-" = {
                          if (num.objects > 2) {
                            temp.matrix[, , 1] - rowSums(temp.matrix[, , -1])
                          } else {
                            temp.matrix[, , 1] - temp.matrix[, , 2]
                          }
                        },
                        "*" = {
                          if (num.objects > 2) {
                            temp.matrix[, , 1] * rowSums(temp.matrix[, , -1])
                          } else {
                            temp.matrix[, , 1] * temp.matrix[, , 2]
                          }
                        },
                        "/" = {
                          temp <- if (num.objects > 2) {
                                    temp.matrix[, , 1] / rowSums(temp.matrix[, , -1])
                                  } else {
                                    temp.matrix[, , 1] / temp.matrix[, , 2]
                                  }

                          ## replace infinities with 0 and throw warning
                          idx.inf <- which(is.infinite(temp))
                          if (length(idx.inf) > 0) {
                            temp[idx.inf]  <- 0
                            .throw_warning(length(idx.inf),
                                           " 'inf' values replaced by 0 in the matrix")
                          }
                          temp
                        })

  ## restore row and column names from the first object
  rownames(temp.matrix) <- rownames(object[[1]]@data)
  colnames(temp.matrix) <- rep(colnames(object[[1]]@data),
                               if (merge.method == "append") num.objects else 1)

  ## add the info slot
  temp.info <- if (is.null(method.info)) {
                 unlist(lapply(object, function(x) x@info), recursive = FALSE)
               } else {
                 object[[method.info]]@info
               }

  ## Build new RLum.Data.Spectrum object ------------------------------------
  set_RLum(
    class = as.character(class(object[[1]])),
    originator = "merge_RLum.Data.Spectrum",
    recordType = object[[1]]@recordType,
    curveType =  "merged",
    data = temp.matrix,
    info = temp.info,
    .pid = unlist(lapply(object, function(x) {
      x@.uid
    }))
  )
}

#' @title Create LaTeX tables from data.frames and RLum objects
#'
#' @description
#' This function takes a [data.frame] and returns a table in LaTeX code that
#' can be copied into any TeX document.
#'
#' @param x [data.frame] or `RLum` object (**required**)
#'
#' @param row.names currently unused
#'
#' @param col.names currently unused
#'
#' @param comments [logical] (*with default*):
#' insert LaTeX comments
#'
#' @param pos [character] (*with default*): `character` of length one specifying the
#' alignment of each column, e.g., `pos = 'clr'` for a three column data frame and
#' center, left and right alignment
#'
#' @param digits [numeric] (*with default*): number of digits to be displayed (numeric fields)
#'
#' @param rm.zero [logical] (*with default*): remove columns containing
#' only zeros, however, this might not be wanted in all cases
#'
#' @param select [character] (*optional*): a [character] vector passed to [subset]
#'
#' @param split [integer] (*optional*): an [integer] specifying the number of individual tables
#' the [data.frame] is split into. Useful for wide tables. Currently unused.
#'
#' @param tabular_only [logical] (*with default*): if `TRUE` only the tabular but not the
#' table environment is returned. This gives a lot of additional flexibility at hand
#'
#' @param ... options: `verbose`
#'
#' @section Function version: 0.2.1
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @keywords internal
#'
#' @return
#' Returns LaTeX code
#'
#' @examples
#'  df <- data.frame(x = 1:10, y = letters[1:10])
#' .as.latex.table(df)
#' .as.latex.table(df, pos = "lr")
#' .as.latex.table(df, select = "y", pos = "r")
#'
#' @export
.as.latex.table <- function(
    x,
    row.names = NULL,
    col.names = NULL,
    comments = TRUE,
    pos = "c",
    digits = 3,
    rm.zero = TRUE,
    select = NULL,
    split = NULL,
    tabular_only = FALSE,
...) {
  .set_function_name(".as.latex.table")
  on.exit(.unset_function_name(), add = TRUE)

  ## TODO:
  # - Improve by using RegEx to dynamically find error fields, e.g. ( "([ ]err)|(^err)" )
  # -

  .validate_logical_scalar(comments)
  .validate_logical_scalar(rm.zero)
  .validate_class(select, "character", null.ok = TRUE)
  .validate_logical_scalar(tabular_only)

  args <- list(
    x = x,
    row.names = row.names,
    col.names = col.names,
    comments = comments,
    pos = pos,
    digits = digits,
    rm.zero = rm.zero,
    select = select,
    split = split,
    tabular_only = tabular_only,
    ... = ...)

  switch(class(x)[1],
         data.frame = do.call(".as.latex.table.data.frame", args),
         DRAC.highlights = do.call(".as.latex.table.data.frame", args),
         RLum.Results = do.call(".as.latex.table.RLum.Results", args))
}

################################################################################
## "Method"                  RLum.Results                                     ##
##----------------------------------------------------------------------------##
.as.latex.table.RLum.Results <- function(
    x,
    row.names = NULL,
    col.names = NULL,
    comments = TRUE,
    pos = "c",
    digits = 3,
    rm.zero = TRUE,
    select = NULL,
    split = NULL,
    ...) {

  ## Object: DRAC.highlights
  if (.check_originator(x, "use_DRAC")) {
    x <- get_RLum(x)$highlights
    x <- .digits(x, digits)

    ##remove columns containing zero values ... they add no information
    if(rm.zero){
      x <- x[sapply(x, function(y){
        y <- suppressWarnings(as.numeric(y))
        if(anyNA(y) || sum(y, na.rm = TRUE) != 0) TRUE else FALSE

      })]
    }

    ##add +/- symbol and remove the columns we don't need
    fields.w.error <- (grep(names(x), pattern = "err", fixed = TRUE) - 1)

    for(i in fields.w.error)
      x[ ,i] <- paste0(x[ ,i], "\\,$\\pm$\\,", trimws(x[ ,i + 1]))
    x <- x[-c(fields.w.error + 1)]

    ##create latex table
    text <- .as.latex.table(x, comments = comments, pos = pos, split = split, ...)

    ##split table
    text <- strsplit(text[[1]], split = "\n", fixed = TRUE)

    ##exchange columns ... or delete them at all (2nd step)
      ##Mineral ID
      for(i in 1:length(text)){
        text[[i]][grepl(pattern = "Mineral", x = text[[i]], fixed = TRUE)] <-
          "\t\\multicolumn{1}{p{0.5cm}}{\\centering \\textbf{M.}} & "
      }

    ##put things again together (single character)
    text <- paste(text[[1]], collapse = "\n")

    ##replace some latex stuff
    text <- gsub(pattern = "p{2cm}", replacement = "p{1.5cm}", x = text, fixed = TRUE)
    text <- gsub(pattern = "Gy.ka-1", replacement = "Gy~ka$^{-1}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "De", replacement = "$D_{E}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "alphadoserate", replacement = "$\\dot{D}_{\\alpha}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "betadoserate", replacement = "$\\dot{D}_{\\beta}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "gammadoserate", replacement = "$\\dot{D}_{\\gamma}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "Cosmicdoserate", replacement = "$\\dot{D}_{cosm.}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "External \\\\ doserate", replacement = "$\\dot{D}_{ext.}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "Internal \\\\ doserate", replacement = "$\\dot{D}_{int.}$", x = text, fixed = TRUE)
    text <- gsub(pattern = "Environmental \\\\ Dose \\\\ Rate", replacement = "$\\dot{D}_{env.}$", x = text, fixed = TRUE)

    ## return result
    return(text)

  }# EndOf::use_DRAC
}

################################################################################
## "Method"                     data.frame                                    ##
##----------------------------------------------------------------------------##
.as.latex.table.data.frame <- function(
    x,
    row.names = NULL,
    col.names = NULL,
    comments = TRUE,
    pos = "c",
    digits = 3,
    select = NULL,
    split = NULL,
    tabular_only = FALSE,
    ...) {
  .set_function_name(".as.latex.table.data.frame")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(x, "data.frame")
  .validate_not_empty(x)
  if (!is.null(col.names) && length(col.names) != ncol(x))
    .throw_error("Length of 'col.names' does not match the number of columns")
  if (!is.null(row.names) && length(row.names) != nrow(x))
    .throw_error("Length of 'row.names' does not match the number of rows")
  if (length(pos) != 1)
    .throw_error("Length of 'pos' does not match the number of columns")

  ## Default settings ----
  options <- modifyList(list(verbose = TRUE),
                        list(...))
  .validate_logical_scalar(options$verbose, name = "'verbose'")

  ## Subset data frame ----
  if (!is.null(select)) {
    .validate_not_empty(select)
    is.name <- select %in% names(x)
    if (!all(is.name))
      .throw_error("Invalid 'select' column name, valid column names are: ",
                   .collapse(names(x)))
    x <- subset(x, select = select)
  }

  ## Format numeric fields ----
  x <- .digits(x, digits)

  ## Split the table
  if (is.null(split))
    split <- 1
  chunks <- ceiling(ncol(x) / split)
  chunks.start <- seq(1, ncol(x), chunks)
  chunks.end <- chunks.start + chunks - 1
  chunks.end[length(chunks.end)] <- ncol(x)

  tex.table.list <- vector("list", split)

  for (i in 1:length(tex.table.list)) {
    x.chunk <- x[ ,chunks.start[i]:chunks.end[i]]

    if (ncol(x) == 1) {
      x.chunk <- as.data.frame(x.chunk)
      colnames(x.chunk) <- names(x[i])
    }

    ## Comments ----
    tex.comment.usePackage <- ifelse(comments,
                                     "% add usepackage{adjustbox} to latex preamble \n",
                                     "")

    ## Header ----
    col.names <- tex.table.header <- gsub(pattern = " ",
                                          x = names(x.chunk),
                                          replacement = " \\\\\\\\ ")
    tex.table.header <- paste0("\t",
                               paste("\\multicolumn{1}{p{2cm}}{\\centering",
                                     col.names,
                                     "}",
                                     collapse = " & \n\t"),
                               "\\\\ \n")

    ## Rows ----
    tex.table.rows <- ""
    for (j in 1:nrow(x.chunk)) {
      tex.table.rows <- paste0(tex.table.rows,
                               paste(paste(x.chunk[j, ], collapse = " & "),
                                     "\\\\ \n"))
    }

    ## catch potential latex problems with underscores - after all are numbers, in can be only
    ## on the ID
    tex.table.rows <- gsub("_", "\\_", tex.table.rows, fixed = TRUE)

    ## Tex table ----
    pos.chars <- unlist(strsplit(pos, split = ""))
    if (!all(pos.chars %in% c("l", "c", "r")) ||
        (length(pos.chars) != 1 && length(pos.chars) != ncol(x))) {
      pos <- "c"
    }
    if (nchar(pos) == 1)
      pos <- strrep(pos, ncol(x))

    if(tabular_only){
      tex.table.begin <- paste0(
          "  \\begin{tabular}{", pos, "}\n",
          "     \\hline \n")

      tex.table.end <-  paste0(
        "     \\hline \n",
        "   \\end{tabular}")

    }else{
      tex.table.begin <- paste0(
        "\\begin{table}[ht] \n",
        "  \\centering \n",
        "  \\begin{adjustbox}{max width=\\textwidth} \n",
        "  \\begin{tabular}{", pos, "}\n",
        "     \\hline \n")

      tex.table.end <-  paste0(
        "     \\hline \n",
        "   \\end{tabular} \n",
        "   \\end{adjustbox} \n",
        "\\end{table}")
    }

    tex.table <- paste0(
      tex.comment.usePackage,
      tex.table.begin,
      tex.table.header,
      "\\hline \n",
      tex.table.rows,
      tex.table.end)

    if (options$verbose)
      cat(tex.table)

    tex.table.list[[i]] <- tex.table
  }

  invisible(tex.table.list)
}

# This function takes a data.frame, checks each column and tries to
# force the specified amount of digits if numeric or coercible to numeric
.digits <- function(x, digits) {
  for (i in 1:ncol(x)) {
    if (is.factor(x[ ,i]))
      x[ ,i] <- as.character(x[ ,i])
    test.numeric <- suppressWarnings(as.numeric(x[ ,i]))
    if (!is.na(test.numeric[1]))
      x[ ,i] <- format(round(test.numeric, digits), nsmall = digits, digits = digits)

  }
  return(x)
}

test_that("Check .as.latex.table()", {
  testthat::skip_on_cran()

  ## data.frame
  df <- data.frame(x = "test", y = 1:10, z = as.factor(letters[1:10]))
  expect_error(.as.latex.table(data.frame()),
               "'x' cannot be an empty data.frame")
  expect_error(.as.latex.table(df, comments = iris),
               "'comments' should be a single logical value")
  expect_error(.as.latex.table(df, rm.zero = NA),
               "'rm.zero' should be a single logical value")
  expect_error(.as.latex.table(df, select = 2),
               "'select' should be of class 'character' or NULL")
  expect_error(.as.latex.table(df, select = character()),
               "'select' cannot be an empty character")
  expect_error(.as.latex.table(df, select = "error"),
               "Invalid 'select' column name, valid column names are: 'x', 'y', 'z'")
  expect_error(.as.latex.table(df, tabular_only = 2:3),
               "'tabular_only' should be a single logical value")

  expect_output(.as.latex.table(df))
  expect_output(.as.latex.table(df, select = "x"))

  ## RLum.Results
  res <- as(object = list(1:10), Class = "RLum.Results")
  expect_error(.as.latex.table(res, comments = iris),
               "'comments' should be a single logical value")
  expect_error(.as.latex.table(res, rm.zero = NA),
               "'rm.zero' should be a single logical value")
  expect_null(.as.latex.table(set_RLum("RLum.Results")))
  expect_null(.as.latex.table(res))

  ## DRAC highlights
  dh <- use_DRAC(template_DRAC(preset = "DRAC-example_quartz",
                               notification = FALSE), verbose = FALSE)
  expect_output(.as.latex.table(dh))
})

test_that("Check .as.latex.table.data.frame()", {
  testthat::skip_on_cran()

  expect_error(.as.latex.table.data.frame("error"),
               "'x' should be of class 'data.frame'")

  df <- data.frame(x = "test", y = 1:10)
  expect_error(.as.latex.table(df, col.names = "col1"),
               "Length of 'col.names' does not match the number of columns")
  expect_error(.as.latex.table(df, row.names = "row1"),
               "Length of 'row.names' does not match the number of rows")
  expect_error(.as.latex.table(df, pos = c("pos1", "pos2")),
               "Length of 'pos' does not match the number of columns")
  expect_error(.as.latex.table(df, comments = iris),
               "'comments' should be a single logical value")
  expect_error(.as.latex.table(df, verbose = NA),
               "'verbose' should be a single logical value")

  expect_output(.as.latex.table(df, tabular_only = TRUE))
  expect_output(.as.latex.table(df[, 1, drop = FALSE], pos = "error"))
})

#' @title Modelling Exponential Exposure Distribution
#'
#' @description Modelling incomplete and heterogeneous bleaching of mobile grains partially
#' exposed to the light, an implementation of the EED model proposed by Guibert et al. (2019).
#'
#' @details The function is an implementation and enhancement of the scripts used in
#' Guibert et al. (2019). The implementation supports a semi-automated estimation
#' of the parameters `kappa` and `sigma_distr`. If set to `NULL`, a surface interpolation
#' is used to estimated those values.
#'
#' **Method control parameters**
#'
#' \tabular{llll}{
#'  **ARGUMENT** \tab **FUNCTION** \tab **DEFAULT** \tab **DESCRIPTION**\cr
#'  `lower`  \tab  - \tab `c(0.1,0,0)` \tab set lower bounds for kappa, sigma, and the expected De in auto mode \cr
#'  `upper` \tab  - \tab `c(1000,2)` \tab set upper bounds for kappa, sigma, and the expected De in auto mode \cr
#'  `iter_max` \tab - \tab `1000` \tab maximum number for iterations for used to find kappa and sigma \cr
#'  `trace` \tab - \tab `FALSE` \tab enable/disable terminal trace mode; overwritten by global argument `verbose`\cr
#'  `trace_plot` \tab - \tab `FALSE` \tab enable/disable additional trace plot output; overwritten by global argument `verbose` \cr
#' }
#'
#' @param data [data.frame] (**required**): input data consisting of two columns, the De and the
#' SE(De). Values are expected in Gy
#'
#' @param D0 [integer] (*with default*): D0 value (in Gy), defining the
#' characterisation behaviour of the quartz.
#'
#' @param expected_dose [numeric] (**required**):
#' expected equivalent dose.
#'
#' @param MinIndivDose [numeric] (*with default*): value specifying the minimum dose taken into
#' account for the plateau. `NULL` applies all values.
#'
#' @param MaxIndivDose [numeric] (*with default*): value specifying the maximum dose taken into
#' account for the plateau. `NULL` applies all values.
#'
#' @param kappa [numeric] (*optional*): positive dimensionless exposure parameter
#' characterising the bleaching state of the grains. Low values (< 10) indicate
#' poor bleaching
#'
#' @param sigma_distr [numeric] (*optional*): positive dose rate parameter, representing the
#' dose variability to which the grains were exposed ##TODO perhaps it should be renamed
#'
#' @param n.simul [integer] (*with default*): number of simulations
#'
#' @param n.minSimExp [integer] (*with default*): number of MC runs for calculating the uncertainty
#' contribution from the sampling
#'
#' @param sample_name [character] (*with default*): name of the sample
#'
#' @param method_control [list] (*with default*): additional deep control parameters, parameters
#' need to be provided as named list, see details
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param ... further parameters that can be passed to better control the plot output. Support arguments
#' are `xlab`, `xlim`.
#'
#' @author
#' Pierre Guibert, IRAMAT-CRP2A, UMR 5060, Université Bordeaux Montaigne (France)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#'
#' @section Function version: 0.1.1
#'
#' @references Guibert, P., Christophe, C., Urbanova, P., Guérin, G., Blain, S., 2017.
#' Modelling incomplete and heterogeneous bleaching of mobile grains partially exposed to the
#' light - Towards a new tool for single grain OSL dating of poorly bleached mortars.
#' Radiation Measurements 107, 48–57. \doi{10.1016/j.radmeas.2017.10.003}
#'
#' @seealso [Luminescence::RLum.Results-class], [Luminescence::calc_MinDose], [Luminescence::calc_FuchsLang2001],
#' [Luminescence::calc_IEU], [Luminescence::calc_FiniteMixture]
#'
#' @keywords datagen
#'
#' @examples
#'
#' data(ExampleData.MortarData, envir = environment())
#' calc_EED_Model(
#'  data = MortarData,
#'  kappa = 14,
#'  sigma_distr = 0.37,
#'  expected_dose = 11.7)
#'
#' ## automated estimation of
#' ## sigma_distribution and
#' ## kappa
#' \dontrun{
#'  calc_EED_Model(
#'  data = MortarData,
#'  kappa = NULL,
#'  sigma_distr = NULL,
#'  expected_dose = 11.7)
#' }
#'
#' @export
calc_EED_Model <- function(
  data,
  D0 = 120L,
  expected_dose = NULL,
  MinIndivDose = NULL,
  MaxIndivDose = NULL,
  kappa = NULL,
  sigma_distr = NULL,
  n.simul = 5000L,
  n.minSimExp = 50L,
  sample_name = "",
  method_control = list(),
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_EED_Model")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO
  ## add docu examples
  ## add docu details
  ##
  ## TODO Questions to Pierre
  ##  - what the user is doing of the real dose/age is not known, can it be assessed?
  ##  - the underlying assumption is obviously an exponential distribution, what
  ##  - so far the parameter uncertainty estimation does not work nicely.x
  ##  happens to data where this is not observed? I cannot work?

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, "data.frame")
  .validate_positive_scalar(D0, int = TRUE)
  .validate_positive_scalar(expected_dose)
  .validate_positive_scalar(MinIndivDose, null.ok = TRUE)
  .validate_positive_scalar(MaxIndivDose, null.ok = TRUE)

  ##store and restore par settings
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

# Helper functions ----------------------------------------------------------------------------

  ##the helper functions base on ode by Pierre, each helper was only a little bit
  ## optimised and then tested separately

  # Col_Index = 3:
  # Calcul de la variance du plateau sur la base des doses arch?o
  # corrig?es de la dose r?siduelle uniquement (modif 31.5.2018)
  # Col_Index = 4 (.calc_Plateau_Variance_uncorr):
  # Calcul de la variance du plateau      modifi? le 31.5.2018
  # sur la base des rapport observ?/simul? des doses totales brutes
  # Col_Index = 6 (.calc_Plateau_Variance_AD):
  # Calcul de la variance du plateau      ajout le 20.8.2018
  # sur la base des doses archeologiques
  # ##TODO not yet included
  .calc_Plateau_Variance <- function(M_Data, MinDose_Index, MaxDose_Index,
                                     Col_Index) {
    var_ratio <- stats::var(M_Data[MinDose_Index:MaxDose_Index, Col_Index])
    mean_ratio <- mean(M_Data[MinDose_Index:MaxDose_Index, Col_Index])
    return(var_ratio / mean_ratio^2)
  }

  .EED_Simul_Matrix <- function(M_Simul, expected_dose, sigma_distr,D0, kappa, Iinit, Nsimul) {

    ## génére une liste de Nsimul valeurs distribu?es selon une loi log normale de moyenne expected_dose ##
    M_Simul[,1] <- expected_dose * exp(sigma_distr * rnorm(Nsimul)) *
      exp(-0.5 * (sigma_distr ^ 2))

    ## génére une liste de Nsimul valeurs de dose r?siduelle selon une distribution expoentielle de l'exposition ##
    M_Simul[, 2] <-
      (-D0 * log(1 - Iinit * exp(-stats::rexp(Nsimul, rate = 1 / kappa))))

    ## génére la liste des doses individuelles ##
    M_Simul[,3] <- M_Simul[,1] + M_Simul[,2]

    ## génére la liste class?e des doses individuelles ##
    M_Simul[,4] <- sort(M_Simul[,3], decreasing = FALSE)

    ## génére une liste d'index and sort it
    index <- sort.list(rank(M_Simul[,3], ties.method = "first"))

    ## la liste d'index est utilis?e pour remettre les colonnes 1 et 2 dans l'ordre du classement des doses totales ##
    M_Simul[,5:6] <- M_Simul[index,1:2]

    ##calculate cumulative mean values
    M_Simul[,7:9] <- matrixStats::colCumsums(M_Simul[,4:6]) / seq_len(Nsimul)

    ##return value
    return(M_Simul)
  }


  #  Calcul de la matrice de M_Data : donnees experimentales traitees
  .EED_Data_Matrix <- function(M_Data, Dosedata, M_Simul, expected_dose, Ndata, Nsimul){
    ##set index
    index_SimRes <- as.integer((1:Ndata) * Nsimul / Ndata)

    # recopie Dosedata dans les colonnes 9 et 10
    M_Data[, 9] <- Dosedata[, 1]
    M_Data[, 10] <- Dosedata[, 2]

    #colonne 1 : moyenne cumulative des ED individuels
    M_Data[, 1] <- cumsum(Dosedata[, 1]) / (1:Ndata)

    # colonne 2 : incertitude statistique sur les valeurs de la moyenne cumulative
    M_Data[, 2] <- (1 / (1:Ndata)) * sqrt(cumsum(Dosedata[, 2] ^ 2))

    # colonne 3 : dose nette : dose moyenne corrigee de la moyenne des doses residuelles
    M_Data[, 3] <- M_Data[, 1] - M_Simul[index_SimRes, 9]

    # colonne 4 : ratio observed / simulated of uncorrected doses
    M_Data[, 4] <- M_Data[, 1] / M_Simul[index_SimRes, 7]

    # colonne 5 :erreur stat sur les rapports
    M_Data[, 5] <- M_Data[, 2] / M_Simul[index_SimRes, 7]

    # colonne 6 : dose nette corrigee des r?siduels ET de l'int?gration partielle de la
    # distribution log-normale des burial doses
    M_Data[, 6] <- M_Data[, 3] * (expected_dose / M_Simul[index_SimRes, 8])

    # colonne 7 : erreur sur dose nette corrigee (premier calcul :
    # incertitude uniquement bas?e sur erreur exp?riementale)
    M_Data[, 7] <- abs((M_Data[, 6] / M_Data[, 3]) * M_Data[, 2])

    # colonne 8 : ratio dose residuelle / dose nette corrigee
    M_Data[M_Data[, 6] != 0, 8] <- M_Simul[index_SimRes, 9] / M_Data[, 6]
    M_Data[M_Data[, 6] == 0, 8] <- 0

    return(M_Data)
  }

  # Calcul des incertitudes statistiques liees au tirage des Ndata grains              #
  # calcul incertitude statistique li?e au nombre limit? de grains
  #
  # On se base sur un tirage de Ndata donn?es ? partir de la matrice de simlation
  # on choisit les Ndata premi?res donn?es, on calcule les moyennes cumulatives de dose totale
  # on prend les Ndata suivantes etc de telle sorte ? ce qu'on base le calcul d'?cart-type sur au moins 50 tirages
  # Pour cela on va calculer le nb d'exp?riences simul?es possibles ? partir du rapport Nsimul/Ndata
  # on teste ce rapport et s'il est inf?rieur ? une valeur limite inf?rieure, on demande de relancer le calcul
  # avec davantage de simulations.

  .EED_Calc_Overall_StatUncertainty <- function(M_Data, M_Simul, Ndata, Nsimul, MinNbSimExp) {
    M_SimExpResults <- src_EED_Calc_Overall_StatUncertainty(M_Simul, Ndata, Nsimul, MinNbSimExp)

    # colonne 7 : calcule l'erreur totale sur la dose nette corrig?e M_Data[i,7]
    M_Data[,7] <- abs((M_Data[,6]/M_Data[,3])*sqrt((M_SimExpResults[,2]^2)+(M_Data[,2]^2)))
    return(M_Data)
  }

  # fonction d'initialisation de l'etat initial
  # si absence de valeur, par defaut prend comme etat initial la valeur 1 (saturation)
  # MaJ le 9 novembre 2018
  .Initial_State_of_OSL <- function(Dosedata, D0, method) {
    SaturationState <- (1 - exp(-max(Dosedata[, 1]) / D0))
    if (method == "max") {
      return(SaturationState)
    }
    # nocov start
    if (!is.numeric(method) || method > 1 || method < 0) {
      return(1)
    }
    return(max(method, SaturationState))
    # nocov end
  }

  # fonction permettant de retourner l'indice de la valeur minimale de la  dose individuelle
  # prise en compte pour le calcul du plateau
  # here combined to one single function returning the minium or the maxixumx
  .Get_Plateau_MinDoseIndex <- function(M_Data, Ndata, MinIndivDose) {
      if (is.null(MinIndivDose))
        return(1)
    # nocov start
      current_index <- 1
      while ((MinIndivDose > M_Data[current_index, 9]) &
             (current_index < Ndata)) {
        current_index <- current_index + 1
      }

      return(current_index)
    # nocov end
  }

  # fonction permettant de retourner l'indice de la valeur maximale de la  dose individuelle
  # prise en compte pour le calcul du plateau
  .Get_Plateau_MaxDoseIndex <- function(M_Data, Ndata, MaxIndivDose) {
    if (is.null(MaxIndivDose))
      return(Ndata)
    # nocov start
        current_index <- 1
        while ((MaxIndivDose > M_Data[current_index, 9]) &
               (current_index < Ndata)) {
          current_index <- current_index + 1
        }

      return(current_index)
    # nocov end
  }

  ## allow automated kapp and sigma_distr parameter estimation
  .guess_EED_parameters <- function(set_kappa, set_sigma_distr,
    set_expected_dose, D0, Iinit, Nsimul, Dosedata, M_Data, M_Simul, Ndata,
    set_MinDose_Index, set_MaxDose_Index, method_control_intern = method_control){

    ##settings to control the what needs to be controlled
    method_control <- modifyList(
      list(
        lower = c(0.1, 0, 1),
        upper = c(100, 1, 100),
        iter_max = 1000,
        trace = FALSE,
        trace_plot = FALSE
      ),
      method_control_intern)

    ##define function for the parameter estimation which will be part of
    ##the interpolation
    fn <- function(
      set_kappa, set_sigma_distr,
      M_Simul, set_expected_dose,
      D0, Iinit, Nsimul, Dosedata, M_Data, Ndata){

      M_Simul <- .EED_Simul_Matrix(M_Simul, set_expected_dose, set_sigma_distr, D0, set_kappa, Iinit, Nsimul)
      M_Data <- .EED_Data_Matrix(M_Data, Dosedata, M_Simul, set_expected_dose, Ndata, Nsimul)

      M_Data <- .EED_Calc_Overall_StatUncertainty(
          M_Data = M_Data,
          M_Simul = M_Simul,
          Ndata = Ndata,
          Nsimul = Nsimul,
          MinNbSimExp = MinNbSimExp
        )

      ##return variance and the mean DE
      return(
        c(
        VAR = .calc_Plateau_Variance(M_Data, MinDose_Index = set_MinDose_Index,
                                     MaxDose_Index = set_MaxDose_Index,
                                     Col_Index = 4),
        RESIDUAL = sum((M_Data[,6] - rep(set_expected_dose, nrow(M_Data)))^2)
        ))
      }

    par(mfrow = c(3,3))
    test_var <- Inf
    n_iter <- 0
    while(test_var > 1e-04 && n_iter < method_control$iter_max){
      ##define parameter matrix
      m <- matrix(NA, nrow = 16, ncol = 5)

      ##fill matrix if parameter is not NULL
      #kappa
      m[,1] <- rep(
        exp(seq(log(method_control$lower[1]), log(method_control$upper[1]), length.out = 4)), 4)

      ##sigma_distr
      m[,2] <- rep(seq(method_control$lower[2],method_control$upper[2], length.out = 4), each = 4)

      ##expected dose
      m[,3] <- rep(expected_dose[1], 16)

      ##calculate the variance
      for(i in 1:nrow(m)){
        m[i, 4:5] <-
          fn(
            set_kappa = m[i, 1],
            set_sigma_distr = m[i, 2],
            M_Simul,
            set_expected_dose = m[i,3],
            D0,
            Iinit,
            Nsimul,
            Dosedata,
            M_Data,
            Ndata
          )
      }

      ##surface interpolation
      s <-
        try(interp::interp(
          x = m[, 1],
          y = m[, 2],
          z = m[, 4],
          nx = 200,
          ny = 200,
          duplicate = "strip" #does not seem to work
        ), silent = TRUE)

      if (inherits(s, "try-error")) {
        .throw_error("Surface interpolation failed: ",
                     attr(s, "condition")$message,
                     ", you may want to try it again")
      }

      ##graphical output
      if(plot & method_control$trace_plot){
        graphics::image(
          s,
          col = grDevices::heat.colors(30, rev = TRUE),
          xlab = "kappa",
          ylab = "sigma_distr"
        )
        graphics::contour(s, add= TRUE, nlevels = 10)

        abline(h = s$y[which(s$z == min(s$z, na.rm = TRUE), arr.ind = TRUE)[,2]], lty = 2)
        abline(v = s$x[which(s$z == min(s$z, na.rm = TRUE), arr.ind = TRUE)[,1]], lty = 2)
      }

      ##our decision is the 5 % quantile
      q10 <- which(s$z<= quantile(s$z, probs = 0.05, na.rm = TRUE), arr.ind = TRUE)

      ##calculate
      optim_kappa <- c(median(s$x[unique(q10[,1])], na.rm = TRUE), sd(s$x[unique(q10[,1])]))
      optim_sigm_distr <- c(median(s$y[unique(q10[,1])]), sd(s$y[unique(q10[,1])]))

      ##write output
      if(verbose && method_control$trace){
        cat("\n\n variance threshold: ", min(s$z, na.rm = TRUE))
        cat("\n >> kappa: ",   optim_kappa[1], "\u00b1", optim_kappa[2])
        cat("\n >> sigma_distr: ", optim_sigm_distr[1] , "\u00b1", optim_sigm_distr[2])
      }

        ##reset parameters
        method_control$lower[1:2] <- c(min(s$x[unique(q10[,1])]), min(s$y[unique(q10[,2])]))
        method_control$upper[1:2] <- c(max(s$x[unique(q10[,1])]), max(s$y[unique(q10[,2])]))

      if(verbose && method_control$trace){
        cat("\n >> lower: ", toString(method_control$lower[1:2]))
        cat("\n >> upper: ", toString(method_control$upper[1:2]))
      }

      ##update threshold
      test_var <- min(s$z, na.rm = TRUE)
      n_iter <- n_iter + 1

      ##implement differential break if the search area is already smaller than the area
      if(diff(c(method_control$lower[1], method_control$upper[1])) < 0.1)
         break
    }

    return(list(
       kappa = optim_kappa ,
       sigma_distr = optim_sigm_distr,
       min_var = test_var,
       expected_dose = m[1,3]
       ))
  }

# Input data ----------------------------------------------------------------------------------

##some parameters are rewritten the leave the code by Pierre untouched

Dosedata <- data[,1:2]
Ndata <- nrow(Dosedata)

# Nsimul est le nb de tirages de la variable al?atoire qui va g?n?rer Nsimul valeurs de dose résiduelle
# et Nsimul valeurs de dose arch?ologique (burial dose), donc Nsimul valeurs de dose equivalente
# individuelle. Pour une approche finale et précise de la dose moyenne, on peut monter ce paramétre ?
# des valeurs hautes : 100000 par exemple.
# our une premiére approche rapide d'un échantillon inconnu, il faut limiter les temps de calculs
# pour explorer l'espace kappa, sigma dans lequel on va trouver les param?tres optimaux.
# On peut proposer alors une valeur minimale par défaut de 5000 simulations
# Le mieux est de laisser votre libert? de choix et d'analyse ...
Nsimul <- n.simul

# MinNbSimExp définit le nombre de tirages minimal pour calculer l'incertitude d'echantillonnage,
# 2e membre de l'incertitude statistique globale. J'ai pense que 50 valeurs pour calculer un écart-type ?
# tait le minimum. ce paramétre ajuste le nombre de simulations en l'augmentant ? une valeur
# superieure au produit du nb de donnees et du nombre de tirages minimal au cas où Nsimul est très
# faible. Si l'on souhaite vraiment faire une étude rapide on peut diminuer MinNbSimExp.
# Pensez cependant ? conserver en commentaires cette instruction por remettre ? jour
# le code une fois l'exploration faite :
# MinNbSimExp = 50
MinNbSimExp <- n.minSimExp

if(MinNbSimExp * Ndata > Nsimul)
  Nsimul <- (MinNbSimExp + 1) * Ndata

## valeur en Gy de la dose de saturation selon une loi de croissance exponentielle saturante ##
D0 <- D0 #valeur en Gy

## initialise la valeur de l'intensit? initiale (1 = saturation) ##
#si le param?tre method est "max", l'initialisation se fait selon Iinit = 1-exp(-max(Dosedata[,1])/D0)
# sinon si l'on donne une valeur comprise entre 0 et 1, la valeur introduite est prise en compte,
# sinon on affecte 1 dans tous les autres cas
# Iinit = 1 #par exemple
# ATTENTION / aucun test de coh?rence n'est fait pour savoir
# si la valeur num?rique introduite est compatible ou non avec les donn?es

Iinit <- .Initial_State_of_OSL(Dosedata, D0, "max")
# Iinit = .Initial_State_of_OSL(Dosedata, D0, method = 1)

## on définit une matrice dans laquelle on va stocker et utiliser les données de la simulation ##
M_Simul <- matrix(nrow = Nsimul, ncol = 9)

# génère une matrice à partir des donnés expérimentales #
M_Data <- matrix(nrow = Ndata, ncol = 10)
colnames(M_Data) <- c("CUM_MEAN_DE", "CUM_MEAN_DE_X", "NET_CUM_MEAN_DE_NET", "NET_CUM_MEAN_DE_X",
                      "RATIO_DE_SIM", "RATIO_DE_SIM_X", "INT_NET_DOSE", "INT_NET_DOSE_X",
                      "DE", "DE_X")

# set limits ... ##TODO M_data is still NA here ... double check, so far only NULL works
# The problem is that this call comes to early and it should be somehow integrated in the guess_EED_parameter
# function to get evaluated everytime on the other hand we just want to limit the search range, maybe this
# can be done better.
# The code in .Get_Plateau_MinDoseIndex is rather inefficient and loops over the vector, however,
# this is not out main problem. In fact we have to first understand what this function is really
# doing and whether we can achive this better and easier. So far it gets called a lot of times
Min_plateau <- .Get_Plateau_MinDoseIndex(M_Data, Ndata, MinIndivDose = NULL)
Max_plateau <- .Get_Plateau_MaxDoseIndex(M_Data, Ndata, MaxIndivDose = NULL)
if (Max_plateau <= Min_plateau) {
  Min_plateau <- 1 # nocov
} #priorité max dose

# Guess parameters if needed ------------------------------------------------------------------
if(verbose) cat("\n[calc_EED_Model()]\n")

## introduire le facteur d'eclairement kappa : kappa = 2 : très faible éclairement, kappa >100 bon
##  blanchiment ##
## TODO - this is not really what Pierre had in mind, he wanted to have the variance, not an automated
## esstimation
if(is.null(kappa) || is.null(sigma_distr)){

  if(verbose)
    cat("\n>> Running automated parameter estimation... \n")

  temp_guess <- .guess_EED_parameters(
    set_kappa = kappa,
    set_sigma_distr = sigma_distr,
    set_expected_dose = expected_dose,
    D0 = D0,
    Iinit = Iinit,
    Nsimul = Nsimul,
    Dosedata = Dosedata,
    M_Data = M_Data,
    M_Simul = M_Simul,
    Ndata = Ndata,
    set_MinDose_Index = Min_plateau,
    set_MaxDose_Index = Max_plateau
    )

  kappa <- temp_guess[[1]][1]
  sigma_distr <- temp_guess[[2]][1]
  min_var <- temp_guess[[3]]
  expected_dose <- temp_guess[[4]]

  if(verbose){
    cat(">> min. variance:", min_var)
    cat("\n>> kappa: ", kappa, " | sigma: ",sigma_distr)
  }
}

# Calculation ---------------------------------------------------------------------------------
M_Simul <- .EED_Simul_Matrix(M_Simul, expected_dose, sigma_distr,D0, kappa, Iinit, Nsimul)
M_Data <- .EED_Data_Matrix(M_Data, Dosedata, M_Simul, expected_dose, Ndata, Nsimul)

M_Data <- .EED_Calc_Overall_StatUncertainty(M_Data = M_Data, M_Simul = M_Simul, Ndata = Ndata, Nsimul = Nsimul, MinNbSimExp)

index_min_uncert <- sort.list(index_min_uncert <-
              rank(M_Data[, 7], ties.method = "first"))


# Terminal output -----------------------------------------------------------------------------
if(verbose){
cat("\n------------------------------------ \n")

cat("\n Maximal Individual Equivalent Dose integrated: ", round(M_Data[index_min_uncert[1],9],2), "Gy")
cat("\n Averaged Corrected Equivalent Dose: ",
    round(M_Data[index_min_uncert[1],6],2), "\u00b1", round(M_Data[index_min_uncert[1],7],2), "Gy")
}

#### THE ONE SHOT : calcule les valeurs de dose moyenne

##TODO discuss with Pierre

# "recherche du minimum d'incertitude statistique"
# #min(M_Data[,7])
# index_min_uncert <- rank(M_Data[,7], ties.method = "first")
# index_min_uncert<-sort.list(index_min_uncert)
# "Minimal error for maximal equivalent Dose (Gy) equal to:  "
# M_Data[index_min_uncert[1],9]
# "cumulative mean corrected equivalent dose (Gy)"
# M_Data[index_min_uncert[1],6]
#
#
# #liste_type = c(9, 10, 6, 7)
# #M_Data[,liste_type[1:4]]


# Plotting ------------------------------------------------------------------------------------
if(plot) {

  ##plot settings
  plot_settings <- modifyList(
    x = list(
      xlab = "Individual dose [Gy]",
      xlim = c(0,max(M_Simul[,2:3]))),
    val = list(...))

  ##set box width
  box_width <- 1.5

  ##set output par
  par(mfrow = c(1,2))

  ## Histograms with the data
  ##(1) first histogram showing the natural distribution
  hist(
    x = rep_len(Dosedata[, 1], length(M_Simul[, 3])),
    freq = FALSE,
    breaks = seq(0, box_width * (1 + as.integer(max(M_Simul[,3])/box_width)), box_width),
    xlim = plot_settings$xlim,
    border = rgb(0.7, 0.1, 0, 0.5),
    col = rgb(240,154,149,alpha = 255, maxColorValue = 255),
    main = paste0("Distribution De: ", sample_name),
    xlab = plot_settings$xlab
   )

  ##(2) overplotting the first histogram with simulated data
  hist(
    x = M_Simul[,3],
    breaks = seq(0, box_width*(1 + as.integer(max(M_Simul[,3])/box_width)), box_width),
    freq = FALSE,
    border = rgb(0, 0.1, 0.8, 0.5),
    col = rgb(122,206,209,alpha = 140, maxColorValue = 255),
    add = TRUE,
    xlab = plot_settings$xlab
  )

  ##add legend
  legend(
    "topright",
    bty = "n",
    legend = c("Data", "Simulation"),
    col = c(rgb(240,154,149,alpha = 255, maxColorValue = 255),rgb(122,206,209,alpha = 140, maxColorValue = 255)),
    pch = 15)

  ##add mtext
  mtext(side = 3, text = paste0("n_sim = ", length(M_Simul[,3])))

  ## Histograms with the error for the data
  ##(3) first histogram showing the natural distribution
  hist(
    x = rep_len(Dosedata[, 2], length(M_Simul[, 2])),
    freq = FALSE,
    breaks = seq(0, box_width*(1 + as.integer(max(M_Simul[,2])/box_width)), box_width),
    xlim = plot_settings$xlim,
    border = rgb(0.7, 0.1, 0, 0.5),
    col = rgb(240,154,149,alpha = 255, maxColorValue = 255),
    main = paste0("Distribution SE(De):", sample_name),
    xlab = plot_settings$xlab
  )

  ##(4) overplotting the first histogram with simulated data
  hist(
    x = M_Simul[,2],
    breaks = seq(0, box_width*(1 + as.integer(max(M_Simul[,2])/box_width)), box_width),
    freq = FALSE,
    border = rgb(0, 0.1, 0.8, 0.5),
    col = rgb(122,206,209,alpha = 140, maxColorValue = 255),
    add = TRUE,
    xlab = plot_settings$xlab
  )

  ##add legend
  legend(
    "topright",
    bty = "n",
    legend = c("Data", "Simulation"),
    col = c(rgb(240,154,149,alpha = 255, maxColorValue = 255),rgb(122,206,209,alpha = 140, maxColorValue = 255)),
    pch = 15)

  ##add mtext
  mtext(side = 3, text = paste0("n_sim = ", length(M_Simul[,2])))


  par(mfrow = c(1,3))
  liste_type <- c(4,8)

  ##1st plateau plot
  graphics::matplot(
     Dosedata[,1],
     M_Data[,liste_type],
     type = "p",
     pch = 1,
     log = "x",
     main = "Observed/simulated \n Residual Dose Fraction Ratio",
     xlab = plot_settings$xlab,
     ylab = "Ratios",
     xlim = c(min(Dosedata[,1]), plot_settings$xlim[2]),
     ylim = c(0, 1.5))
   abline(h = 1)
   for (i in 1:Ndata) {
     lines(
       x = c(Dosedata[i, 1], Dosedata[i, 1]),
       y = c((M_Data[i, 4] - M_Data[i, 5]), (M_Data[i, 4] + M_Data[i, 5])))
   }

  ##2nd plateau plot
   graphics::matplot(
     x = Dosedata[, 1],
     y = M_Data[, c(1, 6, 3)],
     type = "p",
     pch = 1,
     log = "x",
     main = paste0("sample ", sample_name),
     xlab = plot_settings$xlab,
     ylab = "Cumulative mean doses [Gy]",
     xlim = c(min(Dosedata[, 1]), plot_settings$xlim[2]),
     ylim = c(0, 2 * expected_dose)
   )

    # tracer les valeurs de la liste classee des doses archeologiques de la matrice de simulation
    # contrainte : ne tracer qu'une partie des points sinon on va passer un temps fou
    # par exemple on veut tracer un nb de points egal a :
    NbSimPtsDisplayed <- min(Nsimul, 1000)
    XY_psimul <- matrix(nrow = NbSimPtsDisplayed, ncol = 2)
    delta_index <- as.integer(Nsimul / NbSimPtsDisplayed)
    plot_index <- 1
    cur_ind <- 1

  while (plot_index < NbSimPtsDisplayed) {
     XY_psimul[cur_ind, 1] <-  M_Simul[plot_index, 4]
     XY_psimul[cur_ind, 2] <-  M_Simul[plot_index, 8]
     plot_index <- plot_index + delta_index
     cur_ind <- cur_ind + 1
   }

   abline(h = expected_dose)
   points(XY_psimul[, 1], XY_psimul[, 2], pch = 1, col = rgb(0,0,0,0.2))

   ##add error bars
   ##error bars for corrected De
   segments(
     x0 = Dosedata[, 1],
     x1 = Dosedata[, 1],
     y0 = M_Data[, 6] - M_Data[, 7],
     y1 = M_Data[, 6] + M_Data[, 7],
     col = "red"
   )

   ##error bars cum mean De
   segments(
     x0 = Dosedata[, 1],
     x1 = Dosedata[, 1],
     y0 = M_Data[, 3] - M_Data[, 2],
     y1 = M_Data[, 3] + M_Data[, 2],
     col = "green"
   )

   ##add legend
   legend(
     "topright",
     legend = c(
       "Cum Mean (uncorr.) De",
       "Cum Mean (corr.) De_sim",
       "Corr. De (resid., partial integ.)",
       "Cum mean (corr.) De"
      ),
     bty = "n",
     pch = 20,
     col = c("black", "grey", "red", "green"))

   ##add standard deviation plot
   plot(
     Dosedata[, 1],
     M_Data[, 7],
     main = "Std Dev on average corrected doses \n ",
     xlab = plot_settings$xlab,
     ylab = "Standard deviation [Gy]",
     xlim = c(min(Dosedata[,1]), plot_settings$xlim[2]),
     type = "p",
     pch = 1,
     log = "x"
   )
 }#end if plot


# Output --------------------------------------------------------------------------------------
set_RLum(
  class = "RLum.Results",
  data = list(
    M_Data = M_Data ##TODO: we should name the columns
  ),
  info = list(
    call = sys.call()
  ))
}

## load data
data(ExampleData.MortarData, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  ## data
  expect_error(calc_EED_Model(),
               "'data' should be of class 'data.frame'")
  expect_error(calc_EED_Model(1L),
               "'data' should be of class 'data.frame'")

  ## D0
  expect_error(calc_EED_Model(MortarData, NA),
               "'D0' should be a single positive integer value")

  ## expected dose
  expect_error(calc_EED_Model(MortarData),
               "'expected_dose' should be a single positive value")
  expect_error(calc_EED_Model(MortarData, expected_dose = "error"),
               "'expected_dose' should be a single positive value")
  expect_error(calc_EED_Model(MortarData, expected_dose = numeric()),
               "'expected_dose' should be a single positive value")

  ## IndivDose
  expect_error(calc_EED_Model(MortarData, expected_dose = 0.17,
                              MinIndivDose = "error"),
               "'MinIndivDose' should be a single positive value or NULL")
  expect_error(calc_EED_Model(MortarData, expected_dose = 0.17,
                              MaxIndivDose = "error"),
               "'MaxIndivDose' should be a single positive value or NULL")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  set.seed(1)
  SW({
  expect_snapshot_RLum(calc_EED_Model(
      MortarData,
      expected_dose = 11.7),
      tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("fast",
                              calc_EED_Model(MortarData,
                                             kappa = 14,
                                             sigma_distr = 0.37,
                                             verbose = FALSE,
                                             expected_dose = 11.7))
  })
})

test_that("more coverage", {
  testthat::skip_on_cran()

  SW({
  expect_s4_class(calc_EED_Model(
    data = MortarData,
    sigma_distr = 0.37,
    verbose = TRUE,
    method_control = list(trace = TRUE, trace_plot = TRUE),
    n.simul = 5,
    expected_dose = 11.7), "RLum.Results")

  set.seed(1)
  expect_error(calc_EED_Model(
    data = MortarData,
    kappa = 1,
    expected_dose = 95,
    n.simul = 300),
    "Surface interpolation failed: scales of x and y are too dissimilar")
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1051
  expect_silent(calc_EED_Model(
    data = MortarData,
    expected_dose = 11.7,
    n.simul = 10,
    verbose = FALSE,
    plot = TRUE))
})

#' @title Apply crosstalk
#'
#' @description Add crosstalk, evenly spread in rook (top-right-bottom-left)
#' directions, to all grain whole locations on one measurement discs
#' (=on position on a measurement wheel in a reader). An
#' added crosstalk value of as example 0.2 means that 0.2 of the value of
#' the central grain is added to each grain in the rook directions. This is an additive
#' action: the central grain itself is not affected by this operation (but will on its
#' turn increase because of crosstalk from its neighbours).
#' This function is used for simulations: can this added crosstalk be detected?
#'
#' @details If an element in `object` is `NA`, it is internally set to 0, so it will not
#' be added.
#'
#' @param object [Luminescence::RLum.Results-class] or [numeric] (**required**):
#' a numerical vector of length 100, representing one or more measurement
#' discs ("positions") in a reader.
#' Each element in the vector represents one grain hole location on a disc.
#'
#' @param n_crosstalk [numeric] (*with default*):
#' A single number quantifying the added
#' crosstalk. Defaults for testing purposes to 0.2. Can be any number, even negative,
#' but for realistic simulations we suggest something between 0 and 0.25.
#'
#' @keywords manip
#'
#' @return
#' A vector of size 100, with the value at each grain hole location including
#' simulated crosstalk.
#'
#' @author
#' Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
#' Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
#'
#' @references
#' de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
#' A novel method to assess crosstalk in single-grain luminescence detection.
#' Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
#'
#' @examples
#' ## Create artificial disc observation
#' observations <- set_RLum(class = "RLum.Results",
#'                          data = list(vn_values = rep(x = c(1,2), each = 50))
#'                        )
#' hist(get_RLum(object = observations))
#'
#' ## Add crosstalk (with default set to 0.2), and visualize the difference
#' ## in the resulting histogram.
#' observations_with_simulated_crosstalk <- apply_Crosstalk(observations)
#' hist(observations_with_simulated_crosstalk)
#'
#' @export
apply_Crosstalk <- function(object,
                            n_crosstalk = 0.2
) {
  .set_function_name("apply_Crosstalk")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Results", "numeric", "integer"))
  if (is.numeric(object)){
    vn_values <- object
  } else {
    vn_values <- get_RLum(object)
  }
  .validate_length(vn_values, 100, name = "'object'")
  .validate_class(n_crosstalk, "numeric", length = 1)

  vb_na_s <- is.na(vn_values)
  vn_values[vb_na_s] <- 0

  ## Prepare multiplication matrix -----------------------
  ## Note: the physical appearance of a measurement disc,
  ## 10x10 grains, is hard coded here
  df_disc_locations <- data.frame(location = 1:100,
                                  x = rep(1:10, times = 10),
                                  y = rep(1:10, each = 10))

  # Calculate matrix with euclidean distances
  mn_dist <- as.matrix(stats::dist(df_disc_locations[, c("x", "y")]))

  ## All distances equal to one are subject to crosstalk
  mn_crosstalk <- ifelse(mn_dist == 1, yes = n_crosstalk, no = 0)

  ## The diagonal -- all distances equal to zero -- are one
  diag(mn_crosstalk) <- 1

  ## Matrix calculations  -----------------------

  vn_sig_with_crosstalk <- as.numeric(mn_crosstalk %*% vn_values)

  ## Assign original NA's to output
  vn_sig_with_crosstalk[vb_na_s] <- NA

  return(vn_sig_with_crosstalk)
}

## load data
obj <- set_RLum(class = "RLum.Results",
                data = list(vn_values = rep(x = c(1,2), each = 50)))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(apply_Crosstalk("error"),
               "'object' should be of class 'RLum.Results', 'numeric' or 'integer'")
  expect_error(apply_Crosstalk(obj, n_crosstalk = "error"),
               "'n_crosstalk' should be of class 'numeric' and have length 1")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_snapshot_plain(apply_Crosstalk(obj))
  expect_snapshot_plain(apply_Crosstalk(1:100))
})

#' @title Plot Posterior OSL-Age Summary
#'
#' @description
#' The function produces a graphical summary of the statistical inference of
#' an OSL age.
#'
#' @details The function is called automatically by [Luminescence::combine_De_Dr].
#'
#' @param object [Luminescence::RLum.Results-class], [numeric] (**required**):
#' an object produced by [Luminescence::combine_De_Dr]. Alternatively, a [numeric] vector of
#' a parameter from an MCMC process.
#'
#' @param level [numeric] (*with default*):
#' probability of shown credible interval.
#'
#' @param digits [integer] (*with default*):
#' number of digits considered for the calculation.
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @param ... further arguments to modify the plot, supported: `xlim`, `ylim`,
#' `xlab`, `ylab`, `main`, `lwd`, `lty`, `col`, `rug`, `polygon_col`,
#' `polygon_density`.
#'
#'@return A posterior distribution plot and an [Luminescence::RLum.Results-class]
#' object with the credible interval.
#'
#' @author
#' Anne Philippe, Université de Nantes (France)\cr
#' Jean-Michel Galharret, Université de Nantes (France)\cr
#' Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#'
#'@section Function version: 0.1.0
#'
#'@seealso [Luminescence::combine_De_Dr], [plot.default], [rjags::rjags]
#'
#'@keywords hplot dplot
#'
#'@examples
#'##generate random data
#'set.seed(1234)
#'object <- rnorm(1000, 100, 10)
#'plot_OSLAgeSummary(object)
#'
#'@export
plot_OSLAgeSummary <- function(
  object,
  level = 0.95,
  digits = 1L,
  verbose = TRUE,
  ...
) {
  .set_function_name("plot_OSLAgeSummary")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("RLum.Results", "numeric"))
  .validate_positive_scalar(level)
  .validate_positive_scalar(digits, int = TRUE)

  if (inherits(object, "RLum.Results")) {
    if (.check_originator(object, c(".calc_BayesianCentralAgeModel",
                                    ".calc_IndividualAgeModel"))) {
      data.object <- "A"
    } else if (.check_originator(object, "combine_De_Dr")) {
      data.object <- "Ages"
    } else {
      .throw_error("'object' was created by an unsupported function ",
                   "(originator is '", object@originator, "')")
    }
    object <- get_RLum(object, data.object = data.object)
  }

  if (any(is.infinite(object))) {
    .throw_warning("Inf values found in 'object', removed")
    object <- object[!is.infinite(object)]
  }

  ## A should be a matrix
  A <- as.matrix(object, ncol = 1)

# Run calculations --------------------------------------------------------
  ## use our internal function instead of Archaeophase to avoid dependency hell
  CI <- tryCatch(round(.calc_HPDI(A, prob = level), digits),
                 error = function(e) {
                   .throw_error(gsub("'x'", "'object'", e$message))
                 })
  Bayes_est_mean <- round(mean(A), digits = digits)
  Bayes_est_sd <- round(sd(A), digits = digits)

# Terminal output ---------------------------------------------------------
  if(verbose){
    cat("\n[plot_OSLAgeSummary()]\n")
    cat(paste0(" Credible Interval (", level * 100 ),"%): ",paste(CI[1,], collapse = " : "), "\n")
    cat(paste0(" Bayes estimate (posterior mean \u00b1 sd): ", Bayes_est_mean[1], " \u00b1 ", Bayes_est_sd[1]),"\n")
  }

# Plot output -------------------------------------------------------------
  density_A <- density(A)

  plot_settings <- modifyList(x = list(
    xlim = range(A),
    ylim = range(density_A$y),
    main = "Posterior distr. of A",
    xlab = "Age [ka]",
    ylab = "Density",
    lwd = 1,
    lty = 1,
    col = "black",
    polygon_col = rgb(1,0,0,0.3),
    polygon_density = 20,
    rug = FALSE
  ), val = list(...))

  plot(
    x = density_A$x,
    y = density_A$y,
    xlim = plot_settings$xlim,
    ylim = plot_settings$ylim * 1.07,
    xlab = plot_settings$xlab,
    ylab = plot_settings$ylab,
    main = plot_settings$main,
    type = "l",
    lwd = plot_settings$lwd,
    lty = plot_settings$lty,
    col = plot_settings$col
  )

  ## add lines on the top for the CI
  lines(x = c(CI[1,]), y = rep(par()$usr[4] * 0.92, 2))
  lines(x = rep(CI[1,1], 2), y = c(par()$usr[4] * 0.91, par()$usr[4] * 0.92))
  lines(x = rep(CI[1,2], 2), y = c(par()$usr[4] * 0.91, par()$usr[4] * 0.92))

  ## add polygon fill
  polygon(
    x = c(density_A$x, rev(density_A$x)),
    y = c(density_A$y, rep(0, length(density_A$y))),
    col = plot_settings$polygon_col,
    lty = 0,
    density = NULL
  )

  ## add CI
  xy_id <- density_A$x >= CI[1,1] & density_A$x <= CI[1,2]
  polygon(
    x = c(density_A$x[xy_id], rev(density_A$x[xy_id])),
    y = c(density_A$y[xy_id], rep(0, length(density_A$y[xy_id]))),
    col = "black",
    lwd = 0.5,
    border = TRUE,
    density = plot_settings$polygon_density
  )

  ##add rug
  if (plot_settings$rug) graphics::rug(A)

  ## add text
  text(x = density_A$x[xy_id][1], y = density_A$y[xy_id][2], CI[1,1], pos = 2, cex = 0.6)
  text(x = max(density_A$x[xy_id]), y = rev(density_A$y[xy_id])[1], CI[1,2], pos = 4, cex = 0.6)
  text(
    x = median(density_A$x[xy_id]),
    y = par()$usr[4] * 0.91,
    labels = paste0("CI: ", level[1] * 100, "%"),
    pos = 3,
    cex = 0.6
  )

# Return ------------------------------------------------------------------
  set_RLum("RLum.Results",
    data = list(
      Estimate = Bayes_est_mean,
      Credible_Interval = CI,
      level = level),
    info = list(call = sys.call()))
}

## load data
set.seed(1276)
Dr <- stats::rlnorm(1000, 0, 0.3)
De <- 50*sample(Dr, 50, replace = TRUE)
s <- stats::rnorm(50, 10, 2)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(combine_De_Dr("error"),
               "'De' should be of class 'numeric'")
  expect_error(combine_De_Dr(De, "error"),
               "'s' should be of class 'numeric'")
  expect_error(combine_De_Dr(De, s, "error"),
               "'Dr' should be of class 'numeric'")
  expect_error(combine_De_Dr(De, s[-1], Dr, int_OD = 0.1),
               "'De' and 's' should have the same length")
  expect_error(combine_De_Dr(De, s, Dr, int_OD = -1),
               "'int_OD' should be a single non-negative value")
  expect_error(combine_De_Dr(De, s, Dr, int_OD = 0.1, Age_range = 1:4),
               "'Age_range' should be of class 'numeric' and have length 2")
  expect_error(combine_De_Dr(De, s, Dr, int_OD = 0.1, Age_range = c(1, NA)),
               "'Age_range' cannot contain missing values")
  expect_error(combine_De_Dr(De, s, Dr, int_OD = 0.1, outlier_threshold = NA),
               "'outlier_threshold' should be a single positive value")
  expect_error(combine_De_Dr(De, s, Dr, int_OD = 0.1, outlier_method = NA),
               "'outlier_method' should be of class 'character' and have length 1")
  SW({
  expect_error(combine_De_Dr(De, s, Dr, int_OD = 0.1,
                             method_control = list(n.chains = 1, diag = TRUE,
                                                   n.iter = 20)),
               "You need at least two chains")
  expect_error(combine_De_Dr(De, s, Dr, int_OD = 0.1,
                             method_control = list(n.chains = 2,
                                                   inits = list(
                                                       list(
                                                           .RNG.name = "base::Wichmann-Hill",
                                                           .RNG.seed = 1)
                                                   ))),
               "Length mismatch between inits and n.chains")
  })
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  ## simple run with standard settings
  results <- expect_s4_class(combine_De_Dr(
    Dr = Dr,
    int_OD = 0.1,
    De,
    s,
    outlier_analysis_plot = TRUE,
    Age_range = c(0, 100),
    cdf_ADr_quantiles = FALSE,
    legend.pos = "topright",
    legend = TRUE,
    method_control = list(n.iter = 100,
                          n.chains = 1)), "RLum.Results")
  })

  ## check whether mcmc is NULL
  expect_null(results$mcmc_IAM)
  expect_null(results$mcmc_BCAM)

  ## run the same with different par settings
  oldpar <- par(mfrow = c(2,2))
  SW({
  results <- expect_s4_class(combine_De_Dr(
    Dr = Dr,
    int_OD = 0.1,
    De,
    s,
    outlier_analysis_plot = TRUE,
    par_local = FALSE,
    Age_range = c(0, 100),
    method_control = list(
      n.iter = 100,
      n.chains = 1,
      return_mcmc = TRUE
      )), "RLum.Results")
  })

  ## check the length of the output
  expect_length(results, 9)

  ## check whether we have the MCMC plots
  expect_s3_class(results$mcmc_IAM, "mcmc.list")
  expect_s3_class(results$mcmc_BCAM, "mcmc.list")

  ## try to plot the results again
  SW({
  plot_OSLAgeSummary(results)
  })

  ## diag = TRUE
  SW({
  expect_s4_class(combine_De_Dr(
    Dr = Dr, int_OD = 0.1, De, s, Age_range = c(0, 100),
    method_control = list(n.iter = 100, n.chains = 2, diag = TRUE)),
    "RLum.Results")
  })

  ## cdf_ADr_quantiles = TRUE and outlier_method = "RousseeuwCroux1993"
  SW({
  expect_s4_class(combine_De_Dr(
    Dr = Dr, int_OD = 0.1, De, s, Age_range = c(0, 100),
    cdf_ADr_quantiles = TRUE,
    outlier_method = "RousseeuwCroux1993",
    method_control = list(n.iter = 100, n.chains = 1)),
    "RLum.Results")
  })

  ## meaningless data for coverage
  set.seed(1)
  Dr.short <- stats::rlnorm(10, 0, 0.3)
  De.short <- 50 * sample(Dr, 2, replace = TRUE)
  s.short <- stats::rnorm(2, 10, 2)

  SW({
  combine_De_Dr(
    Dr = Dr.short,
    int_OD = 0.1,
    De.short,
    s.short,
    outlier_analysis_plot = TRUE,
    Age_range = c(0, 100),
    cdf_ADr_quantiles = FALSE)
  })

  ## reset the graphical parameters to the original values
  par(oldpar)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
  vdiffr::expect_doppelganger("defaults",
                              combine_De_Dr(De, s, Dr, int_OD = 0.1,
                                            method_control =
                                              list(n.iter = 100,
                                                   n.chains = 1,
                                                   inits = list(
                                                       list(.RNG.name = "base::Wichmann-Hill",
                                                            .RNG.seed = 1)))))
  })
})

## example data
set.seed(1234)
object <- rnorm(1000, 100, 10)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_OSLAgeSummary("error"),
               "[plot_OSLAgeSummary()] 'object' should be of class 'RLum.Results'",
               fixed = TRUE)
  expect_error(plot_OSLAgeSummary(set_RLum("RLum.Results", data = list(),
                                           originator = "error")),
               "'object' was created by an unsupported function")

  empty <- set_RLum("RLum.Results", originator = NA_character_)
  expect_error(plot_OSLAgeSummary(empty),
               "'object' was created by an unsupported function")
  expect_error(plot_OSLAgeSummary(object, level = 0),
               "'level' should be a single positive value")
  expect_error(plot_OSLAgeSummary(object, digits = 1.2),
               "'digits' should be a single positive integer value")
  expect_error(plot_OSLAgeSummary(c(-1, 0, NaN, 10000)),
               "'object' contains missing values")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_warning(plot_OSLAgeSummary(c(object, Inf), verbose = FALSE),
                 "Inf values found in 'object', removed")

  ##run from S4-class
  object1 <- set_RLum("RLum.Results",
                     data = list(Ages = object), originator = "combine_De_Dr")
  object2 <- set_RLum("RLum.Results",
                      data = list(A = object), originator = ".calc_IndividualAgeModel")

  SW({
  expect_s4_class(plot_OSLAgeSummary(object1, polygon_density = 5),
                  "RLum.Results")
  expect_s4_class(plot_OSLAgeSummary(object2, polygon_density = 5),
                  "RLum.Results")
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_RLum(plot_OSLAgeSummary(object, verbose = FALSE,
                                          polygon_density = 5),
                       tolerance = 1.5e-6)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
  vdiffr::expect_doppelganger("default",
                              plot_OSLAgeSummary(object))
  vdiffr::expect_doppelganger("level",
                              plot_OSLAgeSummary(object, level = 0.3))
  vdiffr::expect_doppelganger("rug",
                              plot_OSLAgeSummary(object, rug = TRUE))
  vdiffr::expect_doppelganger("polygon density",
                              plot_OSLAgeSummary(object, polygon_density = 5))
  })
})

#' @title Analyse fading measurements and returns the fading rate per decade (g-value)
#'
#' @description
#' The function analyses fading measurements and returns a fading rate
#' including an error estimation. The function is not limited to standard
#' fading measurements, as can be seen, e.g., Huntley and Lamothe (2001).
#' Additionally, the density of recombination centres (rho') is estimated after
#' Kars et al. (2008).
#'
#' @details
#' All provided output corresponds to the \eqn{tc} value obtained by this
#' analysis. Additionally, the g-value normalised to 2-days is provided in the
#' output object. The output of this function can be passed to the function
#' [Luminescence::calc_FadingCorr].
#'
#' **Fitting and error estimation**
#'
#' For the fitting the function [stats::lm] is used without applying weights. For the
#' error estimation all input values, except `tc`, as the precision can be considered as sufficiently
#' high enough with regard to the underlying problem, are sampled assuming a normal distribution
#' for each value with the value as the mean and the provided uncertainty as standard deviation.
#'
#' **The options for `t_star`**
#'
#' \itemize{
#'  \item `t_star = "half"` (the default) The calculation follows the simplified
#'  version in Auclair et al. (2003), which reads
#'  \deqn{t_{star} := t_1 + (t_2 - t_1)/2}
#'  \item `t_star = "half_complex"` This option applies the complex function shown in Auclair et al. (2003),
#'  which is derived from Aitken (1985) appendix F, equations 9 and 11.
#'  It reads \deqn{t_{star} = t0 * 10^[(t_2 \log(t_2/t_0) - t_1 \log(t_1/t_0) - 0.43(t_2 - t_1))/(t_2 - t_1)]}
#'  where 0.43 = \eqn{1/\ln(10)}. t0, which is an arbitrary constant, is set to 1.
#'  Please note that the equation in Auclair et al. (2003) is incorrect
#'  insofar that it reads \eqn{10\exp(...)}, where the base should be 10 and not the Euler's number.
#'  Here we use the correct version (base 10).
#'  \item `t_star = "end"` This option uses the simplest possible form for `t_star` which is the time since
#'  irradiation without taking into account any addition parameter and it equals t1 in Auclair et al. (2003)
#'  \item `t_star = <function>` This last option allows you to provide an R function object that works on t1 and
#'  gives you all possible freedom. For instance, you may want to define the following
#'  function `fun <- function(x) {x^2}`, this would square all values of t1, because internally
#'  it calls `fun(t1)`. The name of the function does not matter.
#' }
#'
#' **Density of recombination centres**
#'
#' The density of recombination centres, expressed by the dimensionless variable rho', is estimated
#' by fitting equation 5 in Kars et al. (2008) to the data. For the fitting the function
#' [stats::nls] is used without applying weights. For the error estimation the same
#' procedure as for the g-value is applied (see above).
#'
#' **Multiple aliquots & Lx/Tx normalisation**
#'
#' Be aware that this function will always normalise all \eqn{\frac{L_x}{T_x}}
#' values by the \eqn{\frac{L_x}{T_x}} value of the prompt measurement of the
#' first aliquot. This implicitly assumes that there are no systematic
#' inter-aliquot variations in the \eqn{\frac{L_x}{T_x}} values.
#' If deemed necessary to normalise the \eqn{\frac{L_x}{T_x}} values  of each
#' aliquot by its individual prompt measurement please do so **before** running
#' [Luminescence::analyse_FadingMeasurement] and provide the already normalised values for
#' `object` instead.
#'
#' **Shine-down curve plots**
#' Please note that the shine-down curve plots are for information only. As such
#' a maximum of five pause steps are plotted to avoid graphically overloaded plots.
#' However, *all* pause times are taken into consideration for the analysis.
#'
#' @param object [Luminescence::RLum.Analysis-class], [data.frame] or [list] (**required**):
#' input object with the measurement data. Alternatively, a [list] containing
#' [Luminescence::RLum.Analysis-class] objects or a [data.frame] with three columns
#' (x = LxTx, y = LxTx error, z = time since irradiation) can be provided.
#' Can also be a wide table, i.e. a [data.frame] with a number of columns
#' divisible by 3 and where each triplet has the before mentioned column
#' structure.
#'
#' **Note:** The input object should solely consists of the curve needed
#' for the data analysis, i.e. only IRSL curves representing Lx (and Tx). If
#' the object originated from an XSYG file, also the irradiation steps must
#' be preserved in the input object.
#'
#' If data from multiple aliquots are provided, please see the details below
#' with regard to Lx/Tx normalisation. **Note:** The function assumes that all your
#' measurements are related to one (comparable) sample. If you have to treat
#' independent samples, you have use this function in a loop.
#'
#' @param structure [character] (*with default*):
#' the structure of the measurement data, one of `'Lx'` or `c('Lx','Tx')`.
#'
#' @param signal_integral [vector] (**required**):
#' vector with channels for the signal integral (e.g., `1:10`). It is not
#' required if a `data.frame` with `LxTx` values is provided.
#'
#' @param background_integral [vector] (**required**):
#' vector with channels for the background integral (e.g., `90:100`). It is not
#' required if a `data.frame` with `LxTx` values is provided.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param t_star [character], [function] (*with default*):
#' method for calculating the time elapsed since irradiation if input is
#' **not** a `data.frame`. Options are: `'half'` (the default), `'half_complex`,
#' which uses the long equation in Auclair et al. 2003, and `'end'`, which
#' takes the time between irradiation and the measurement step.
#' Alternatively, `t_star` can be a function with one parameter which works on `t1`.
#' For more information see details. \cr
#'
#' *`t_star` has no effect if the input is a [data.frame], because this input comes
#' without irradiation times.*
#'
#' @param n.MC [integer] (*with default*):
#' number for Monte Carlo runs for the error estimation.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param plot_singlePanels [logical] or [numeric] (*with default*):
#' enable/disable single plot mode, i.e. one plot window per plot.
#' Alternatively a vector specifying the plot to be drawn, e.g.,
#' `plot_singlePanels = c(3,4)` draws only the last two plots in separate
#' windows.
#'
#' @param ... further arguments that can be passed to internally used functions.
#' Supported arguments: `xlab`, `log`, `mtext`, `plot.trend` (enable/disable
#' trend blue line), and `xlim` for the two first curve plots, and `ylim` for
#' the fading curve plot. For further plot customization please use the
#' numerical output of the functions for own plots.
#'
#' @return
#' An [Luminescence::RLum.Results-class] object is returned:
#'
#' Slot: **@data**
#'
#' \tabular{lll}{
#'  **OBJECT** \tab **TYPE** \tab **COMMENT**\cr
#' `fading_results` \tab `data.frame` \tab results of the fading measurement in a table \cr
#' `fit` \tab `lm` \tab object returned by the used linear fitting function [stats::lm]\cr
#' `rho_prime` \tab `data.frame` \tab results of rho' estimation after Kars et al. (2008) \cr
#' `LxTx_table` \tab `data.frame` \tab Lx/Tx table, if curve data had been provided \cr
#' `irr.times` \tab `integer` \tab vector with the irradiation times in seconds \cr
#' }
#'
#' Slot: **@info**
#'
#' \tabular{lll}{
#' **OBJECT** \tab `TYPE` \tab `COMMENT`\cr
#' `call` \tab `call` \tab the original function call\cr
#' }
#'
#' @section Function version: 0.1.27
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Christoph Burow, University of Cologne (Germany)
#'
#' @keywords datagen
#'
#' @references
#'
#' Aitken, M.J., 1985. Thermoluminescence dating, Studies in archaeological science.
#' Academic Press, London, Orlando.
#'
#' Auclair, M., Lamothe, M., Huot, S., 2003. Measurement of anomalous fading for feldspar IRSL using
#' SAR. Radiation Measurements 37, 487-492. \doi{10.1016/S1350-4487(03)00018-0}
#'
#' Huntley, D.J., Lamothe, M., 2001. Ubiquity of anomalous fading in K-feldspars and the measurement
#' and correction for it in optical dating. Canadian Journal of Earth Sciences 38,
#' 1093-1106. doi: `10.1139/cjes-38-7-1093`
#'
#' Kars, R.H., Wallinga, J., Cohen, K.M., 2008. A new approach towards anomalous
#' fading correction for feldspar  IRSL dating-tests on samples in field saturation.
#' Radiation Measurements 43, 786-790. \doi{10.1016/j.radmeas.2008.01.021}
#'
#' @seealso [Luminescence::calc_OSLLxTxRatio], [Luminescence::read_BIN2R],
#' [Luminescence::read_XSYG2R],
#' [Luminescence::extract_IrradiationTimes], [Luminescence::calc_FadingCorr]
#'
#' @examples
#'
#' ## load example data (sample UNIL/NB123, see ?ExampleData.Fading)
#' data("ExampleData.Fading", envir = environment())
#'
#' ##(1) get fading measurement data (here a three column data.frame)
#' fading_data <- ExampleData.Fading$fading.data$IR50
#'
#' ##(2) run analysis
#' g_value <- analyse_FadingMeasurement(
#' fading_data,
#' plot = TRUE,
#' verbose = TRUE,
#' n.MC = 10)
#'
#' ##(3) this can be further used in the function
#' ## to correct the age according to Huntley & Lamothe, 2001
#' results <- calc_FadingCorr(
#' age.faded = c(100,2),
#' g_value = g_value,
#' n.MC = 10)
#'
#' @export
analyse_FadingMeasurement <- function(
  object,
  structure = c("Lx", "Tx"),
  signal_integral = NULL,
  background_integral = NULL,
  integral_input = c("channel", "measurement"),
  t_star = c("half", "half_complex", "end"),
  n.MC = 100,
  verbose = TRUE,
  plot = TRUE,
  plot_singlePanels = FALSE,
  ...
) {
  .set_function_name("analyse_FadingMeasurement")
  on.exit(.unset_function_name(), add = TRUE)

  integral_input <- .validate_args(integral_input, c("channel", "measurement"))

  ## deprecated arguments
  extraArgs <- list(...)
  if (any(grepl("[signal|background]\\.integral", names(extraArgs)))) {
    .deprecated(old = c("signal.integral", "background.integral"),
                new = c("signal_integral", "background_integral"),
                since = "1.2.0")
    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal_integral <- extraArgs$signal.integral
    background_integral <- extraArgs$background.integral
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("RLum.Analysis", "data.frame", "list"))
  .validate_class(structure, "character")
  ## signal_integral and background_integral are validated in calc_OSLLxTxRatio()
  .validate_class(plot_singlePanels, c("logical", "integer", "numeric"))
  .validate_positive_scalar(n.MC, int = TRUE)

  if (inherits(object, "list")) {
    wrong.class <- !sapply(object, inherits, "RLum.Analysis")
    if (any(wrong.class)) {
      .throw_warning(sum(wrong.class), " unsupported records removed")

      ## remove unwanted stuff
      object <- object[!wrong.class]
      if (length(object) == 0) {
        .throw_error("No valid records in 'object' left")
      }
    }
  } else if (inherits(object, "RLum.Analysis")) {
    object <- list(object)

  } else if(inherits(object,"data.frame")){
    .validate_not_empty(object)
    if (ncol(object) %% 3 != 0) {
      .throw_error("When 'object' is a data.frame, the number of columns ",
                   "must be a multiple of 3")
    }
    colnames(object) <- NULL
    object <- do.call(rbind,
                      lapply(seq(1, ncol(object), 3), function(col) {
                        object[, col + 0:2]
                      }))
      object <- object[stats::complete.cases(object), ]
      if (nrow(object) == 0)
        .throw_error("After NA removal, nothing is left from the data set")

    ##set table and object
    LxTx_table <- data.frame(LxTx = object[[1]], LxTx.Error = object[[2]])
    irradiation_times <- TIMESINCEIRR <- object[[3]]
    if (any(irradiation_times < 0)) {
      .throw_error("'object' cannot contain negative irradiation times")
    }
    object <- NULL
  }

  if (!inherits(t_star, "function")) {
    t_star <- .validate_args(t_star, c("half", "half_complex", "end"),
                             extra = "a function")
  }

  # Prepare data --------------------------------------------------------------------------------
  if(!is.null(object)){
    if (!(length(structure) == 1 && structure == "Lx" ||
          length(structure) == 2 && all(structure == c("Lx", "Tx")))) {
      .throw_message("'structure' can only be 'Lx' or c('Lx', 'Tx'), ",
                     "NULL returned")
      return(NULL)
    }

    ## support read_XSYG2R()
    originator <- unique(unlist(lapply(object, slot, name = "originator")))
    if (length(originator) != 1 ||
        !originator %in% c("read_XSYG2R", "read_BIN2R",
                           "Risoe.BINfileData2RLum.Analysis"))  {
      .throw_message("Unknown or unsupported originator, NULL returned")
      return(NULL)
    }

    if (originator == "read_XSYG2R") {
      ## extract irradiation times
      irradiation_times <- extract_IrradiationTimes(object)

      ## get TIMESINCEIRR
      TIMESINCEIRR <- unlist(lapply(irradiation_times, function(x) {
        x@data$irr.times[["TIMESINCEIRR"]][!grepl(pattern = "irradiation",
                                             x = x@data$irr.times[["STEP"]],
                                             fixed = TRUE)]
      }))
      if (all(TIMESINCEIRR < 0)) {
        .throw_error("No irradiation times could be retrieved, check that ",
                     "'object' contains irradiation steps")
      }

      ## get irradiation times
      irradiation_times <- unlist(lapply(irradiation_times, function(x) {
          x@data$irr.times[["IRR_TIME"]][!grepl(pattern = "irradiation",
                                           x = x@data$irr.times[["STEP"]],
                                           fixed = TRUE)]
      }))

      ##clean object by removing the irradiation step ... and yes, we drop!
      object_clean <- unlist(get_RLum(object, curveType = "measured"))

      ## if we are left with fewer curves than expected we cannot proceed
      if (length(object_clean) < length(structure)) {
        .throw_message("After irradiation step removal not enough curves ",
                       "are left from the data set, NULL returned")
        return(NULL)
      }

      ##support read_BIN2R()
    } else if (originator %in% c("read_BIN2R", "Risoe.BINfileData2RLum.Analysis")) {

      ##assign object, unlist and drop it
      object_clean <- unlist(get_RLum(object))
      bin.version <- object_clean[[1]]@info$VERSION
      if (as.integer(bin.version) < 5) {
        .throw_error("BIN-file has version ", bin.version,
                     ", but only versions from 05 on are supported")
      }

      ##set TIMESINCEIRR vector
      TIMESINCEIRR <- vapply(object_clean, function(o){
        o@info$TIMESINCEIRR
      }, numeric(1))

      ##check whether we have negative irradiation times, sort out such values
      if(any(TIMESINCEIRR < 0)){
        ##now we have a problem and we first have to make sure that we understand
        ##the data structure and remove also the corresponding values
        if(all(structure == c("Lx", "Tx"))){
          rm_id <- matrix(TIMESINCEIRR, ncol = 2, byrow = TRUE)
          rm_id[apply(rm_id < 0, MARGIN = 1, any),] <- NA
          rm_id <- is.na(as.numeric(t(rm_id)))
        }else{
          rm_id <- TIMESINCEIRR < 0
        }
        object_clean[rm_id] <- NULL
        TIMESINCEIRR <- TIMESINCEIRR[!rm_id]
        rm_records <- sum(rm_id)

        .throw_warning("removed ", rm_records, " records with negative ",
                       "'time since irradiation'")

        ## check if we have removed everything
        if (length(object_clean) == 0) {
          .throw_message("After record removal nothing is left from ",
                         "the data set, NULL returned")
          return(NULL)
        }
      }

      ##set irradiation times
      irradiation_times <- vapply(object_clean, function(o){
        o@info$IRR_TIME
      }, numeric(1))
    }

    ##correct irradiation time for t_star
    ##in accordance with Auclair et al., 2003, p. 488
    ##but here we have no t1 ... this needs to be calculated
    ##set variables
    t1 <- TIMESINCEIRR
    t2 <- TIMESINCEIRR + irradiation_times

    ## set t_star ----
    if (inherits(t_star, "function")) {
      t_star <- t_star(t1)
    } else {
      t_star <- switch(
          t_star,

          ## use the simplified equation in Auclair et al. (2003)
          half = t1 + (t2 - t1) / 2,

          ## use full equation Auclair et al. (2003)
          ## t0 is an arbitrary constant, we are setting that to 1
          half_complex = {
            t0 <- 1
            t_star <- t0 * 10^((t2 * log10(t2/t0) -
                                t1 * log10(t1/t0) -
                                (t2 - t1) * log10(exp(1))) / (t2 - t1))
          },

          ## set t_star to t_1 (so after the end of irradiation)
          end = t1
      )
    }

    ##overwrite TIMESINCEIRR
    TIMESINCEIRR <- pmax(t_star, 1e-6)
    rm(t_star)
    # Calculation ---------------------------------------------------------------------------------
    ##calculate Lx/Tx or ... just Lx, it depends on the pattern ... set IRR_TIME
    if(length(structure) == 2){
      Lx_data <- object_clean[seq(1,length(object_clean), by = 2)]
      Tx_data <- object_clean[seq(2,length(object_clean), by = 2)]

      ## check whether the length of Lx is the length of Tx
      len.Lx <- length(Lx_data)
      len.Tx <- length(Tx_data)
      if (len.Lx != len.Tx) {
        .throw_error("The number of Lx curves (", len.Lx, ") differs from ",
                     "the number of Tx curves (", len.Tx, "), ",
                     "check your data or consider setting `structure = 'Lx'`")
      }

      ##we need only every 2nd irradiation time, the one from the Tx should be the same ... all the time
      TIMESINCEIRR <- TIMESINCEIRR[seq(1,length(TIMESINCEIRR), by = 2)]

      ## check that all Lx/Tx pairs have the same size
      size.mismatch <- vapply(1:len.Lx, function(i) {
        floor(length(Lx_data[[i]])) != floor(length(Tx_data[[i]]))
      }, logical(1))

      ## skip samples with mismatching sizes
      if (any(size.mismatch)) {

        ## terminate if all pairs have mismatching sizes
        if (all(size.mismatch)) {
          .throw_error("No curves left after removing those with different ",
                       "Lx and Tx sizes")
        }

        rm.idx <- which(size.mismatch)
        .throw_warning("Skipped the following samples because Lx and Tx have ",
                       "different sizes: ", .collapse(rm.idx, quote = FALSE))
        Lx_data <- Lx_data[-rm.idx]
        Tx_data <- Tx_data[-rm.idx]
        TIMESINCEIRR <- TIMESINCEIRR[-rm.idx]
      }

    }else if(length(structure) == 1){
      Lx_data <- object_clean
      Tx_data <- NULL
    }

    ##calculate Lx/Tx table
    LxTx_table <- merge_RLum(.warningCatcher(
      calc_OSLLxTxRatio(
        Lx.data = Lx_data,
        Tx.data = Tx_data,
        signal_integral = signal_integral,
        background_integral = background_integral,
        signal_integral_Tx = list(...)$signal_integral_Tx,
        background_integral_Tx = list(...)$background_integral_Tx,
        integral_input = integral_input,
        sigmab = list(...)$sigmab,
        sig0 = list(...)$sig0 %||% formals(calc_OSLLxTxRatio)$sig0,
        background.count.distribution =
          list(...)$background.count.distribution %||%
                   formals(calc_OSLLxTxRatio)$background.count.distribution
      )
    ))$LxTx.table
  }

  ##create unique identifier
  uid <- create_UID()

  ##normalise data to prompt measurement
  tc <- min(TIMESINCEIRR)[1]

  ## remove Inf values in LxTx table
  if(any(is.infinite(LxTx_table[["LxTx"]]))){
    rm_id <- which(is.infinite(LxTx_table[["LxTx"]]))
    LxTx_table <- LxTx_table[-rm_id,]
    TIMESINCEIRR <- TIMESINCEIRR[-rm_id]
    rm(rm_id)
  }

  ##normalise
  idx.tc <- which(TIMESINCEIRR == tc)[1]
  if (length(structure) == 2 || is.null(object)){
    LxTx <- LxTx_table[["LxTx"]]
    if (LxTx[idx.tc] <= 0) {
      .throw_error("Normalisation term is not positive, check your input")
    }
    LxTx_NORM <- LxTx / LxTx[idx.tc]
    LxTx_NORM.ERROR <- LxTx_table[["LxTx.Error"]] / LxTx[idx.tc]
  } else {
    LnLx <- LxTx_table[["Net_LnLx"]]
    LxTx_NORM <- LnLx / LnLx[idx.tc]
    LxTx_NORM.ERROR <- LxTx_table[["Net_LnLx.Error"]] / LnLx[idx.tc]
  }

  ## normalise time since irradiation
  TIMESINCEIRR_NORM <- TIMESINCEIRR/tc

  ##add dose and time since irradiation
  LxTx_table <-
    cbind(
      LxTx_table,
      TIMESINCEIRR = TIMESINCEIRR,
      TIMESINCEIRR_NORM = TIMESINCEIRR_NORM,
      TIMESINCEIRR_NORM.LOG = log10(TIMESINCEIRR_NORM),
      LxTx_NORM = LxTx_NORM,
      LxTx_NORM.ERROR = LxTx_NORM.ERROR,
      UID = uid
    )

  # Fitting -------------------------------------------------------------------------------------
  ## prevent that n.MC can become smaller than 2
  n.MC <- max(n.MC, 2)

  ##we need to fit the data to get the g_value

  ##sample for monte carlo runs
  MC_matrix <- suppressWarnings(cbind(LxTx_table[["TIMESINCEIRR_NORM.LOG"]],
                     matrix(rnorm(
                       n = n.MC * nrow(LxTx_table),
                       mean = LxTx_table[["LxTx_NORM"]],
                       sd = abs(LxTx_table[["LxTx_NORM.ERROR"]])
                     ),
                     ncol = n.MC)))

  ##apply the fit
  fit_matrix <- vapply(X = 1:n.MC, FUN = function(x) {
    ##fit
    fit <- try(stats::lm(y ~ x, data = data.frame(
      x = MC_matrix[,1],
      y = MC_matrix[, x + 1]))$coefficients, silent = TRUE)

    if (inherits(fit, "try-error"))
      return(c(NA_real_, NA_real_))
    return(fit)
  }, FUN.VALUE = numeric(2))

  ##calculate g-values from matrix
  g_value.MC <- -fit_matrix[2, ] * 1 / fit_matrix[1, ] * 100

  ##calculate rho prime (Kars et al. 2008; proposed by Georgina E. King)

  ##s value after Huntley (2006) J. Phys. D.
  Hs <- 3e15

  ##sample for monte carlo runs
  MC_matrix_rhop <-  suppressWarnings(matrix(rnorm(
    n = n.MC * nrow(LxTx_table),
    mean = LxTx_table[["LxTx_NORM"]],
    sd = abs(LxTx_table[["LxTx_NORM.ERROR"]])
  ), ncol = n.MC))

  ## calculate rho prime for all MC samples
  fit_vector_rhop <- suppressWarnings(apply(MC_matrix_rhop, MARGIN = 2, FUN = function(x) {
    tryCatch({
      coef(minpack.lm::nlsLM(x ~ c * exp(-rhop * (log(1.8 * Hs * TIMESINCEIRR))^3),
                             start = list(c = x[1], rhop = 10^-5.5),
                             data = LxTx_table))[["rhop"]]
    },
    error = function(e) {
      return(NA)
    })
  }))

  ## discard all NA values produced in MC runs
  fit_vector_rhop <- fit_vector_rhop[!is.na(fit_vector_rhop)]

  .collect_results <- function(val, vals.MC) {
    data.frame(FIT = val,
               MEAN = mean(vals.MC),
               SD = sd(vals.MC),
               Q_0.025 = quantile(vals.MC, probs = 0.025, na.rm = TRUE),
               Q_0.16  = quantile(vals.MC, probs = 0.16,  na.rm = TRUE),
               Q_0.84  = quantile(vals.MC, probs = 0.84,  na.rm = TRUE),
               Q_0.975 = quantile(vals.MC, probs = 0.975, na.rm = TRUE),
               row.names = NULL)
  }

  ## calculate mean and standard deviation of rho prime (in log10 space)
  rhoPrime <- .collect_results(mean(fit_vector_rhop), fit_vector_rhop)

  ## calc g-value -----
  df <- data.frame(x = LxTx_table[["TIMESINCEIRR_NORM.LOG"]],
                   y = LxTx_table[["LxTx_NORM"]])
  fit <- try(stats::lm(y ~ x, data = df), silent = TRUE)
  fit_power <- try(stats::lm(y ~ I(x^3) + I(x^2) + I(x),
                             data = df), silent = TRUE)

  ##for predicting
  fit_predict <-
    try(stats::lm(y ~ x, data = data.frame(y = LxTx_table[["TIMESINCEIRR_NORM.LOG"]],
                                       x = LxTx_table[["LxTx_NORM"]])), silent = TRUE)

  ##calculate final g_value
  ##the 2nd term corrects for the (potential) offset from one
  g_value_fit <- NA
  if (!inherits(fit, "try-error")) {
    g_value_fit <- -fit$coefficient[2] * 1 / fit$coefficient[1] * 100
  }

  ##construct output data.frame
  g_value <- .collect_results(g_value_fit, g_value.MC)

  ##normalise the g-value to 2-days using the equation provided by Sébastien Huot via e-mail
  ##this means the data is extended
  ## calc g2-value days ----
  k0 <- g_value[,c("FIT", "SD")] / 100 / log(10)
  k1 <- k0 / (1 - k0 * log(172800/tc))
  g_value_2days <-  as.numeric(100 * k1 * log(10))
  names(g_value_2days) <- c("G_VALUE_2DAYS", "G_VALUE_2DAYS.ERROR")

  # Approximation -------------------------------------------------------------------------------
  T_0.5.interpolated <- try(approx(x = LxTx_table[["LxTx_NORM"]],
                               y = LxTx_table[["TIMESINCEIRR_NORM"]],
                               ties = mean,
                               xout = 0.5), silent = TRUE)

  if(inherits(T_0.5.interpolated, 'try-error')){
    T_0.5 <- data.frame(
        T_0.5_INTERPOLATED = NA,
        T_0.5_PREDICTED = NA,
        T_0.5_PREDICTED.LOWER = NA,
        T_0.5_PREDICTED.UPPER = NA
    )
  }else{
    ## if the number of observations is less than 3, the number of residual
    ## degrees of freedom is 0, and we would get this warning when computing
    ## the prediction interval:
    ##   In qt((1 - level)/2, df) : NaNs produced
    ## hence, we suppress the warning (#616)
    T_0.5.predict <- suppressWarnings(
        stats::predict.lm(fit_predict, newdata = data.frame(x = 0.5),
                          interval = "predict"))
    T_0.5 <- data.frame(
        T_0.5_INTERPOLATED = T_0.5.interpolated$y,
        T_0.5_PREDICTED = (10 ^ T_0.5.predict[, 1]) * tc,
        T_0.5_PREDICTED.LOWER = (10 ^ T_0.5.predict[, 2]) * tc,
        T_0.5_PREDICTED.UPPER = (10 ^ T_0.5.predict[, 2]) * tc
    )
  }

  ## Plotting ---------------------------------------------------------------
  if(plot) {
    ## deprecated argument
    if ("plot.single" %in% ...names()) {
      plot_singlePanels <- list(...)$plot.single
      .deprecated("plot.single", "plot_singlePanels", since = "1.0.0")
    }

    ## split the plot area into 4 regions if plot_singlePanels = FALSE (default)
    if (is.logical(plot_singlePanels) && !plot_singlePanels[1]) {
      par.default <- .par_defaults()
      on.exit(par(par.default), add = TRUE)
      par(mfrow = c(2, 2))
    }

    ##get package
    col <- get("col", pos = .LuminescenceEnv)

    ##set some plot settings
    plot_settings <- list(
      xlab = "Stimulation time [s]",
      ylim = NULL,
      xlim = NULL,
      log = "",
      mtext = "",
      plot.trend = TRUE
    )

    ##modify on request
    plot_settings <- modifyList(x = plot_settings, val = list(...))

    ##get unique irradiation times ... for plotting
    irradiation_times.unique <- unique(TIMESINCEIRR)

    ##limit to max 5
    if(length(irradiation_times.unique) >= 5){
      irradiation_times.unique <-
        irradiation_times.unique[seq(1, length(irradiation_times.unique),
                                     length.out = 5)]
    }

    ## convert plot_singlePanels to numeric if it was given as logical: this
    ## helps simplifying the checks below
    if (is.logical(plot_singlePanels))
      plot_singlePanels <- 1:4

    ## compute integration limits for plots
    if (!is.null(object)) {
      if (length(structure) == 2) {
        int.limits.lx <- c(object_clean[[1]][range(signal_integral), 1],
                           object_clean[[1]][range(background_integral), 1])
      } else {
        int.limits.lx <- c(range(signal_integral), range(background_integral)) *
          max(as.matrix(object_clean[[1]][, 1])) /
          nrow(as.matrix(object_clean[[1]]))
      }
      if (is.null(list(...)$signal_integral_Tx)) {
        int.limits.tx <- int.limits.lx
      } else {
        int.limits.tx <- c(range(list(...)$signal_integral_Tx),
                           range(list(...)$background_integral_Tx)) *
          max(as.matrix(object_clean[[1]][, 1])) /
          nrow(as.matrix(object_clean[[1]]))
      }
    }

    ## plot Lx-curves -----
    if (!is.null(object) && 1 %in% plot_singlePanels) {
      records <- object_clean[seq(1, length(object_clean), by = length(structure))]
      plot_RLum(
          set_RLum(class = "RLum.Analysis",
                   records = records),
          combine = length(records) > 1,
          records_max = 10,
          plot_singlePanels = TRUE,
          legend.text = c(paste(round(irradiation_times.unique, 1), "s")),
          xlab = plot_settings$xlab,
          xlim = plot_settings$xlim,
          log = plot_settings$log,
          legend.pos = "outside",
          main = bquote(L[x] ~ "- curve"),
          mtext = plot_settings$mtext
      )
      abline(v = int.limits.lx, lty = 2,
             col = c("green", "green", "red", "red"))
    } else if (1 %in% plot_singlePanels) {
      ## empty Lx plot
      shape::emptyplot()
      text(x = 0.5, y = 0.5,
           labels = expression(paste("No ", L[x], " curves detected")))
    }

    ## plot Tx-curves ----
    if (!is.null(object) && length(structure) == 2 && 2 %in% plot_singlePanels) {
      records <- object_clean[seq(2, length(object_clean), by = 2)]
      plot_RLum(
          set_RLum(class = "RLum.Analysis",
                   records = records),
          combine = length(records) > 1,
          records_max = 10,
          plot_singlePanels = TRUE,
          legend.text = paste(round(irradiation_times.unique, 1), "s"),
          xlab = plot_settings$xlab,
          log = plot_settings$log,
          legend.pos = "outside",
          main = bquote(T[x] ~ "- curve"),
          mtext = plot_settings$mtext
      )
      abline(v = int.limits.tx, lty = 2,
             col = c("green", "green", "red", "red"))
    } else if (2 %in% plot_singlePanels) {
      ## empty Tx plot
      shape::emptyplot()
      text(x = 0.5, y = 0.5,
           labels = expression(paste("No ", T[x], " curves detected")))
    }

    ## plot fading ----
    if (3 %in% plot_singlePanels) {

      if(all(is.na(LxTx_table[["LxTx_NORM"]]))){
        ## FIXME(mcol): this block seems unreachable since 5f63c1f1
        # nocov start
          shape::emptyplot()
          text(x = 0.5, y = 0.5, labels = "All NA values!")
        # nocov end
      }else{
        plot(
          NA,
          NA,
          ylab = "Norm. intensity",
          xaxt = "n",
          xlab = "Time since irradition [s]",
          sub = expression(paste("[", log[10](t / t[c]), "]")),
          ylim = plot_settings$ylim %||% (
            if (max(LxTx_table[["LxTx_NORM"]]) > 1.1) {
              c(0.1, max(LxTx_table[["LxTx_NORM"]]) + max(LxTx_table[["LxTx_NORM.ERROR"]]))
            } else {
              c(0.1, 1.1)
            }
          ),
          xlim = range(LxTx_table[["TIMESINCEIRR_NORM.LOG"]], na.rm = TRUE),
          main = "Signal Fading"
        )

        ##add axis (with an additional formatting to provide a nice log10-axis)
        ##https://stackoverflow.com/questions/6897243/labelling-logarithmic-scale-display-in-r
        x_axis_lab <- seq(0:nchar(floor(max(LxTx_table[["TIMESINCEIRR"]]))))
        x_axis_ticks <- log10((10^x_axis_lab)/tc)

        ## if we have less then two values to show, we fall back to the
        ## old data representation.
        if (length(x_axis_ticks[x_axis_ticks > 0]) > 2) {
          axis(
            side = 1,
            at = x_axis_ticks,
            labels = sapply(x_axis_lab, function(i)
              as.expression(bquote(10 ^ .(i))))
          )
          ##lower axis
          axis(
            side = 1,
            at = x_axis_ticks,
            labels = paste0("[",round(x_axis_ticks,1),"]"),
            cex.axis = 0.7,
            tick = FALSE,
            line = 0.75)

        } else {
          axis(
            side = 1,
            at = axTicks(side = 1),
            labels = suppressWarnings(format((10 ^ (axTicks(side = 1)) * tc),
                                                digits = 1,
                                                decimal.mark = "",
                                                scientific = TRUE)))

          ##lower axis
          axis(
            side = 1,
            at = axTicks(1),
            labels = axTicks(1),
            cex.axis = 0.7,
            tick = FALSE,
            line = 0.75)
        }

        mtext(
          side = 3,
          paste0(
            "g-value: ",
            round(g_value$FIT, digits = 2),
            " \u00b1 ",
            round(g_value$SD, digits = 2),
            " (%/decade) | tc = ",
            format(tc, digits = 4, scientific = TRUE)
          ),
          cex = par()$cex * 0.9
        )

        ##add MC error polygon
        x_range <- range(LxTx_table[["TIMESINCEIRR_NORM.LOG"]], na.rm = TRUE)
          x <- seq(x_range[1], x_range[2], length.out = 50)
          m <- matrixStats::rowRanges(vapply(1:n.MC, function(i){
            fit_matrix[2, i] * x + fit_matrix[1, i]

          }, numeric(length(x))))
          polygon(
            x = c(x, rev(x)),
            y = c(m[, 2], rev(m[, 1])),
            col = rgb(0, 0, 0, 0.2),
            border = NA
          )

        ##add master curve in red
        curve(
          fit$coefficient[2] * x + fit$coefficient[1],
          col = "red",
          add = TRUE,
          lwd = 1.5
        )

        ##add power law curve
        if(plot_settings$plot.trend) {
          curve(
            x ^ 3 * fit_power$coefficient[2] + x ^ 2 * fit_power$coefficient[3] + x
            * fit_power$coefficient[4] + fit_power$coefficient[1],
            add = TRUE,
            col = "blue",
            lty = 2
          )
        }

        ##add points
        points(x = LxTx_table[["TIMESINCEIRR_NORM.LOG"]],
               y = LxTx_table[["LxTx_NORM"]],
               pch = 21,
               bg = "grey")

        ##error bars
        segments(
          x0 = LxTx_table[["TIMESINCEIRR_NORM.LOG"]],
          x1 = LxTx_table[["TIMESINCEIRR_NORM.LOG"]],
          y0 = LxTx_table[["LxTx_NORM"]] + LxTx_table[["LxTx_NORM.ERROR"]],
          y1 = LxTx_table[["LxTx_NORM"]] - LxTx_table[["LxTx_NORM.ERROR"]],
          col = "grey"
        )

        ##add legend
        legend(
          "bottom",
          legend = c("fit", "fit MC", if(plot_settings$plot.trend) "trend" else NULL),
          col = c("red", "grey", if(plot_settings$plot.trend) "blue" else NULL),
          lty = c(1, 1, if(plot_settings$plot.trend) 2 else NULL),
          bty = "n",
          horiz = TRUE
        )
      }#end if a
    }#

    if (4 %in% plot_singlePanels) {

      if(all(is.na(g_value.MC))){
        shape::emptyplot()
        text(x = 0.5, y = 0.5, labels = "All NA values!")

      }else{
        plot(density(g_value.MC),
             main = "Density: g-values (%/decade)")
        graphics::rug(x = g_value.MC)
        abline(v = c(g_value[["Q_0.16"]], g_value[["Q_0.84"]]),
               lty = 2,
               col = "darkgreen")
        abline(v = c(g_value[["Q_0.025"]], g_value[["Q_0.975"]]),
               lty = 2,
               col = "red")
        legend(
          "topleft",
          legend = c("HPD - 68 %", "HPD - 95 %"),
          lty = 2,
          col = c("darkgreen", "red"),
          bty = "n"
        )
      }
    }
  }

  # Terminal ------------------------------------------------------------------------------------
  if (verbose){
    cat("\n[analyse_FadingMeasurement()]\n")
    cat("\n n.MC:\t", n.MC)
    cat("\n tc:\t", format(tc, digits = 4, scientific = TRUE), "s")
    cat("\n---------------------------------------------------")
    cat("\nT_0.5 interpolated:\t", T_0.5$T_0.5_INTERPOLATED)
    cat("\nT_0.5 predicted:\t", format(T_0.5$T_0.5_PREDICTED, digits = 2, scientific = TRUE))
    cat("\ng-value:\t\t", round(g_value$FIT, digits = 2),
        "\u00b1", round(g_value$SD, digits = 2), "(%/decade)")
    cat("\ng-value (norm. 2 days):\t", round(g_value_2days[1], digits = 2),
        "\u00b1", round(g_value_2days[2], digits = 2), "(%/decade)")
    cat("\n---------------------------------------------------")
    cat("\nrho':\t\t\t", format(rhoPrime$MEAN, digits = 3),
        "\u00b1", format(rhoPrime$SD, digits = 3))
    cat("\nlog10(rho'):\t\t", suppressWarnings(round(log10(rhoPrime$MEAN), 2)),
        "\u00b1", round(rhoPrime$SD / (rhoPrime$MEAN * log(10)), 2))
    cat("\n---------------------------------------------------\n")
  }

  # Return --------------------------------------------------------------------------------------

  ##set data.frame
  if(all(is.na(g_value))){
    fading_results <- data.frame(
      FIT = NA,
      MEAN = NA,
      SD = NA,
      Q_0.025 = NA,
      Q_0.16 = NA,
      Q_0.84 = NA,
      Q_0.975 = NA,
      TC = NA,
      G_VALUE_2DAYS = NA,
      G_VALUE_2DAYS.ERROR = NA,
      T_0.5_INTERPOLATED = NA,
      T_0.5_PREDICTED = NA,
      T_0.5_PREDICTED.LOWER = NA,
      T_0.5_PREDICTED.UPPER =  NA,
      UID = uid,
      stringsAsFactors = FALSE
    )
  }else{
    fading_results <- data.frame(
      g_value,
      TC = tc,
      G_VALUE_2DAYS = g_value_2days[1],
      G_VALUE_2DAYS.ERROR = g_value_2days[2],
      T_0.5,
      UID = uid,
      stringsAsFactors = FALSE
    )
  }

  ##return
  set_RLum(
    class = "RLum.Results",
    data = list(
      fading_results = fading_results,
      fit = fit,
      rho_prime = rhoPrime,
      LxTx_table = LxTx_table,
      irr.times = irradiation_times
    ),
    info = list(call = sys.call())
  )
}

## load data
set.seed(1)
data("ExampleData.Fading", envir = environment())
fading_data <- ExampleData.Fading$fading.data$IR50
data <- ExampleData.Fading$equivalentDose.data$IR50
ddot <- c(7.00, 0.004)
readerDdot <- c(0.134, 0.0067)

rhop <-
  analyse_FadingMeasurement(fading_data,
                            plot = FALSE,
                            verbose = FALSE,
                            n.MC = 10)
huntley <- calc_Huntley2006(
  data = data,
  rhop = rhop,
  ddot = ddot,
  readerDdot = readerDdot,
  n.MC = 50,
  plot = FALSE,
  verbose = FALSE
)

test_that("input validation", {
  testthat::skip_on_cran()

  rhop.test <- rhop
  rhop.test@originator <- "unexpected"

  expect_error(calc_Huntley2006(),
               "'data' should be of class 'data.frame'")
  expect_error(calc_Huntley2006("test"),
               "'data' should be of class 'data.frame'")
  expect_error(calc_Huntley2006(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(calc_Huntley2006(iris[, 1, drop = FALSE]),
               "'data' should have at least 2 columns")

  expect_error(calc_Huntley2006(data, fit.method = "test"),
               "'fit.method' should be one of 'EXP' or 'GOK'")
  expect_error(calc_Huntley2006(data, fit.method = "GOK", lower.bounds = 0),
               "'lower.bounds' should be of class 'numeric' and have length 4")

  expect_error(calc_Huntley2006(data, LnTn = list()),
               "'LnTn' should be of class 'data.frame'")
  expect_error(calc_Huntley2006(data, LnTn = data),
               "'LnTn' should be a data frame with 2 columns")
  expect_error(calc_Huntley2006(cbind(data, data), LnTn = data[, 1:2]),
               "When 'LnTn' is specified, 'data' should have only 2 or 3")
  expect_error(calc_Huntley2006(cbind(data, data[, 1])),
               "The number of columns in 'data' must be a multiple of 3")

  expect_error(calc_Huntley2006(data, rhop = 1),
               "'rhop' should have length 2")
  expect_error(calc_Huntley2006(data, rhop = "test"),
               "'rhop' should be of class 'numeric' or 'RLum.Results'")
  expect_error(calc_Huntley2006(data, rhop = rhop.test),
               "'rhop' has an unsupported originator")
  expect_error(calc_Huntley2006(data, rhop = c(0, 4e-7)),
               "'rhop' must be a positive number, the provided value was 0 \u00B1 4e-07")

  expect_error(calc_Huntley2006(data, rhop = rhop),
               "'ddot' should be of class 'numeric' and have length 2")
  expect_error(calc_Huntley2006(data, rhop = rhop,
                                ddot = "test", readerDdot = readerDdot),
               "'ddot' should be of class 'numeric' and have length 2")
  expect_error(calc_Huntley2006(data, rhop = rhop,
                                ddot = 0.4, readerDdot = readerDdot),
               "'ddot' should be of class 'numeric' and have length 2")
  expect_error(calc_Huntley2006(data, rhop = rhop, ddot = ddot),
               "'readerDdot' should be of class 'numeric' and have length 2")
  expect_error(calc_Huntley2006(data, rhop = rhop,
                                ddot = ddot, readerDdot = "test"),
               "'readerDdot' should be of class 'numeric' and have length 2")
  expect_error(calc_Huntley2006(data, rhop = rhop,
                                ddot = ddot, readerDdot = 0.13),
               "'readerDdot' should be of class 'numeric' and have length 2")
  expect_error(calc_Huntley2006(data, rhop = rhop,
                                ddot = ddot, readerDdot = readerDdot,
                                n.MC = 0),
               "'n.MC' should be a single positive integer value")
  expect_error(calc_Huntley2006(data, rhop = rhop,
                                ddot = ddot, readerDdot = readerDdot,
                                rprime = list()),
               "'rprime' should be of class 'numeric'")

  SW({
  expect_warning(calc_Huntley2006(data[, 1:2], rhop = rhop, n.MC = 2,
                                  ddot = ddot, readerDdot = readerDdot,
                                  fit.method = "GOK"),
                 "'data' has only two columns: we assume that the errors")
  })

  expect_warning(expect_error(
      calc_Huntley2006(data = data[1:20, ], rhop = rhop, ddot = ddot,
                       readerDdot = c(0.002, 0.003), n.MC = 2,
                       plot = FALSE, verbose = FALSE),
      "Simulated D0 is NA"),
      "Ln/Tn is smaller than the minimum computed LxTx value")

  expect_error(calc_Huntley2006(data = data[1:20, ], LnTn = data[1, c(2, 3)],
                                rhop = c(2, 2), ddot = c(7.00, 0.004),
                                readerDdot = c(0.01, 0.02), n.MC = 2,
                                verbose = FALSE),
               "Could not fit simulated curve, check suitability of model")
})

test_that("check class and length of output", {
  testthat::skip_on_cran()

  ##rhop
  expect_s4_class(rhop, class = "RLum.Results")
  expect_s3_class(rhop$fading_results, "data.frame")
  expect_s3_class(rhop$fit, "lm")
  expect_s3_class(rhop$rho_prime, "data.frame")
  expect_equal(round(sum(rhop$fading_results[,1:9]),0),415)
  expect_equal(round(sum(rhop$rho_prime),5),2e-05)
  expect_equal(round(sum(rhop$irr.times)), 2673108)

  ##kars
  expect_s4_class(huntley, class = "RLum.Results")
  expect_s3_class(huntley$results, class = "data.frame")
  expect_s3_class(huntley$data, class = "data.frame")
  expect_type(huntley$Ln, "double")
  expect_type(huntley$fits, "list")

  expect_equal(round(huntley$results$Sim_Age, 1), 34)
  expect_equal(round(huntley$results$Sim_Age_2D0, 0), 175)
  expect_equal(round(sum(huntley$Ln),2), 0.16)

  expect_equal(round(sum(huntley$data),0), 191530)
  expect_equal(round(sum(residuals(huntley$fits$simulated)),1),  0.8)
  expect_equal(round(sum(residuals(huntley$fits$measured)),4),  0.1894)
  expect_equal(round(sum(residuals(huntley$fits$unfaded)),2),  0)
})

test_that("Further tests calc_Huntley2006", {
  testthat::skip_on_cran()

  os <- tolower(Sys.info()[["sysname"]])
  snapshot.tolerance <- switch(os,
                               "linux" = 1.5e-5,
                               "darwin" = 8.0e-2,
                               "windows" = 8.0e-2)

  ## check extrapolation
  set.seed(1)
  expect_snapshot_RLum(
      calc_Huntley2006(
        data = data,
        rhop = rhop,
        ddot = ddot,
        readerDdot = readerDdot,
        n.MC = 100,
        fit.method = "GOK",
        mode = "extrapolation",
        plot = TRUE, verbose = FALSE),
      tolerance = snapshot.tolerance)

  ## check force through origin EXP with wrong mode settings
  set.seed(1)
  expect_snapshot_RLum(
      calc_Huntley2006(
        data = data,
        rhop = rhop,
        ddot = ddot,
        readerDdot = readerDdot,
        n.MC = 100,
        fit.method = "EXP",
        fit.force_through_origin = TRUE,
        mode = "extrapolation",
        plot = FALSE,
        verbose = FALSE),
      tolerance = max(snapshot.tolerance, 1.0e-2))

  ## EXP ... normal
  set.seed(1)
  expect_snapshot_RLum(
      calc_Huntley2006(
        data = data,
        rhop = rhop,
        ddot = ddot,
        readerDdot = readerDdot,
        n.MC = 100,
        fit.method = "EXP",
        fit.force_through_origin = TRUE,
        mode = "interpolation",
        plot = FALSE,
        verbose = FALSE),
      tolerance = snapshot.tolerance)

  ## GOK normal
  set.seed(1)
  expect_snapshot_RLum(
      calc_Huntley2006(
        data = data,
        rhop = rhop,
        ddot = ddot,
        readerDdot = readerDdot,
        n.MC = 100,
        fit.method = "GOK",
        fit.force_through_origin = TRUE,
        mode = "interpolation",
        plot = FALSE,
        verbose = FALSE),
      tolerance = if (os == "darwin") 1.7e-1 else max(snapshot.tolerance, 1.5e-3))

  ## check warning for failed fits
  ## dataset provided by Christine Neudorf
  df <- structure(list(V1 = c(0L, 0L, 200L, 800L, 1500L, 3000L, 8000L
  ), V2 = c(0.439, -0.046720922, 1.988131642, 7.577744961, 12.87699795,
            18.50187874, 32.72443771),
  V3 = c(0.029, 0.01269548, 0.090232208,
         0.322242141, 0.546595156, 0.739308178, 1.285979033)),
  class = "data.frame", row.names = c(NA, -7L))

  set.seed(1)
  SW({
  expect_warning(calc_Huntley2006(
    data = df,
    LnTn = NULL,
    rhop = c(0.0000121549740899913, 4.91596040125088E-07),
    ddot = c(6.96, 0.29),
    readerDdot = c(0.094, 0.01),
    normalise = FALSE,
    fit.method = "EXP",
    summary = TRUE,
    plot = TRUE,
    n.MC = 100),
    regexp = "\\[calc\\_Huntley2006\\(\\)\\] Ln\\/Tn is smaller than the minimum computed LxTx value.")
  })

  ## failing to fit unfaded EXP model
  ## test derived from data provided by Salome Oehler
  input <- data.frame(dose = c(0, 1550, 9300, 37200, 71500,
                               0, 1550, 9300, 37200, 71500),
                      LxTx = c(0.79, 4.67, 14.41, 26.59, 24.88,
                               0.95, 4.87, 14.17, 21.98, 25.12),
                      LxTx_error = c(0.002, 0.006, 0.01, 0.04, 0.02,
                                     0.002, 0.011, 0.02, 0.03, 0.03))
  set.seed(1)
  expect_error(calc_Huntley2006(input,
                                rhop = c(6.5e-06, 2.0e-08),
                                ddot = c(8.5, 1.5),
                                fit.method = "EXP",
                                mode = "extrapolation",
                                readerDdot = c(0.154, 0.1),
                                n.MC = 2),
               "Could not fit unfaded curve, check suitability of model and")

  ## more coverage
  expect_s4_class(
    calc_Huntley2006(
      data = data[1:10, ],
      LnTn = data[1:10, c(2, 3)],
      rhop = rhop, ddot = ddot, readerDdot = readerDdot,
      n.MC = 2, plot = TRUE, verbose = FALSE),
    class = "RLum.Results")
  expect_s4_class(
    calc_Huntley2006(
      data = data[1:10, ],
      LnTn = data[1, c(2, 3)],
      rhop = rhop, ddot = ddot, readerDdot = readerDdot,
      rprime = c(0.01, 2.2, length.out = 500),
      n.MC = 2, plot = FALSE, verbose = FALSE),
    class = "RLum.Results")
  ## set maxiter and trace
  expect_output(
    calc_Huntley2006(
      data = data[1:10, ],
      LnTn = data[1:10, c(2, 3)],
      rhop = rhop, ddot = ddot, readerDdot = readerDdot,
      n.MC = 2, plot = FALSE, verbose = FALSE, maxiter = 50, trace = TRUE))

  expect_error(
    calc_Huntley2006(
      data = iris[, 1:3],
      rhop = rhop, ddot = ddot, readerDdot = readerDdot,
      n.MC = 2, plot = FALSE, verbose = FALSE),
    "Unable to fit growth curve to measured data, try setting 'fit.bounds = FALSE'")
  expect_error(
    calc_Huntley2006(
      data = iris[, 1:3],
      rhop = rhop, ddot = ddot, readerDdot = readerDdot,
      n.MC = 2, plot = FALSE, verbose = FALSE, fit.bounds = FALSE),
    "Unable to fit growth curve to measured data$")
  expect_warning(expect_error(
    calc_Huntley2006(
      data = data,
      rhop = c(4e-5, 5e-7), ddot = c(8, 0.04), readerDdot = c(0.1, 0.006),
      n.MC = 2, fit.method = "GOK", plot = FALSE, verbose = FALSE),
    "Could not fit unfaded curve, check suitability of model and parameters"),
    "Ln is >10 % larger than the maximum computed LxTx value")
  expect_warning(expect_error(
    calc_Huntley2006(
      data = data,
      rhop = c(4e-5, 5e-7), ddot = c(8, 0.04), readerDdot = c(0.1, 0.006),
      n.MC = 2, mode = "extrapolation", plot = FALSE, verbose = FALSE),
    "Simulated D0 is NA: either your input values are unsuitable"),
    "Ln is >10 % larger than the maximum computed LxTx value")
  expect_error(
    calc_Huntley2006(
      data = data,
      rhop = c(1e-3, 1e-7), ddot = ddot, readerDdot = readerDdot,
      n.MC = 2, plot = FALSE, verbose = FALSE),
    "All simulated Lx/Tx values are identical and approximately zero")
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 660
  DRC <- data.frame(
      Dose = c(rep(0, 7), rep(500, 12), rep(1000, 6), rep(2000, 6)),
      Lx.Tx = c(1.79, rep(0.06, 6), rep(0.97, 12), rep(1.80, 6), rep(3.20, 6)),
      Lx.Tx.Err = c(0.038, rep(0.001, 6), rep(0.021, 12), rep(0.038, 6), rep(0.068, 6))
  )
  expect_s4_class(
      calc_Huntley2006(
          data = DRC,
          rhop = c(7.67e-7, 2.84e-7),
          fit.method = "GOK",
          readerDdot = c(0.0868, 0.005),
          ddot = c(2.372, 0.199),
      n.MC = 2, plot = FALSE, verbose = FALSE),
      "RLum.Results")

  ## issue 733
  expect_s4_class(
      calc_Huntley2006(data, rhop = c(4e-6, 5e-7), ddot = c(7, 0.004),
                       readerDdot = c(0.134, 0.0067), n.MC = 1,
                       mode = "extrapolation", verbose = FALSE),
      "RLum.Results")

  ## issue 1048
  set.seed(1)
  expect_silent(
      calc_Huntley2006(data, rhop = c(1e-7, 5e-7), ddot = c(7, 0.004),
                       readerDdot = c(0.134, 0.0067), n.MC = 1,
                       mode = "extrapolation", verbose = FALSE,
                       plot_all_DRC = FALSE))
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_FadingCorr("error"),
               "'age.faded' should be of class 'numeric' or 'integer' and have length 2")
  expect_error(calc_FadingCorr(1),
               "'age.faded' should be of class 'numeric' or 'integer' and have length 2")
  expect_error(calc_FadingCorr(c(0.1, 0), "error"),
               "'g_value' should be of class 'numeric', 'integer' or 'RLum.Results'")
  expect_error(calc_FadingCorr(age.faded = c(0.1,0), tc = 2592000,
                               g_value = set_RLum("RLum.Results")),
               "'g_value' has an unsupported originator")
  expect_error(calc_FadingCorr(c(0.1, 0), 1),
               "'g_value' should have length 2")
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0)),
               "[calc_FadingCorr()] 'tc' should be a single positive value",
               fixed = TRUE)
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0),
                               tc = c(1000, 2000)),
               "[calc_FadingCorr()] 'tc' should be a single positive value",
               fixed = TRUE)
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0),
                               tc = 2592000, n.MC = "error"),
               "'n.MC' should be a single positive integer value")
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0),
                               tc = 2592000, interval = "error"),
               "'interval' should be of class 'numeric' and have length 2")
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0),
                               tc = 2592000, interval = 1),
               "'interval' should be of class 'numeric' and have length 2")
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0),
                               tc = 2592000, txtProgressBar = "error"),
               "'txtProgressBar' should be a single logical value")
  expect_error(calc_FadingCorr(age.faded = c(0.1, 0), g_value = c(5.0, 1.0),
                               tc = 2592000, verbose = "error"),
               "'verbose' should be a single logical value")
})

test_that("check class and length of output", {
  testthat::skip_on_cran()

  ##check message
  expect_message(calc_FadingCorr(
    age.faded = c(6.404856, 0.51),
    g_value = c(17.5,1.42),
    tc = 462,
    n.MC = 100),
    "[calc_FadingCorr()] Error: No solution found, NULL returned: this usually",
    fixed = TRUE)

  ##check the verbose mode
  SW({
  expect_s4_class(calc_FadingCorr(
    age.faded = c(0.1,0),
    g_value = c(5.0, 1.0),
    tc = 2592000,
    tc.g_value = 172800,
    n.MC = 1, verbose = TRUE), class = "RLum.Results")

  ## g_value provided as RLum.Results object
  data("ExampleData.Fading", envir = environment())
  fading <- analyse_FadingMeasurement(ExampleData.Fading$fading.data$IR50,
                                      plot = FALSE)
  expect_s4_class(calc_FadingCorr(age.faded = c(0.1,0),
                                  g_value = fading, tc = 2592000),
                  "RLum.Results")
  })

  ## auto, seed (Note: this is slow!)
  SW({
  calc_FadingCorr(
    age.faded = c(0.1,0),
    g_value = c(5.0, 1.0),
    tc = 2592000,
    seed = 1,
    n.MC = "auto")
  })
})

test_that("check values from output example 1", {
  testthat::skip_on_cran()

  set.seed(1)
  temp <- calc_FadingCorr(
      age.faded = c(0.1,0),
      g_value = c(5.0, 1.0),
      tc = 2592000,
      tc.g_value = 172800,
      n.MC = 100, verbose = FALSE)

  expect_s4_class(temp, "RLum.Results")
  expect_equal(length(temp), 2)

  results <- get_RLum(temp)
  expect_equal(results$AGE, 0.1169)
  expect_equal(results$AGE.ERROR, 0.0035)
  expect_equal(results$AGE_FADED, 0.1)
  expect_equal(results$AGE_FADED.ERROR, 0)
  expect_equal(results$G_VALUE, 5.312393)
  expect_equal(round(results$G_VALUE.ERROR, 5), 1.01190)
  expect_equal(round(results$KAPPA, 3), 0.0230)
  expect_equal(round(results$KAPPA.ERROR,3), 0.004)
  expect_equal(results$TC, 8.213721e-05)
  expect_equal(results$TC.G_VALUE, 5.475814e-06)
  expect_equal(results$n.MC, 100)
  expect_equal(results$OBSERVATIONS, 100)
  expect_equal(results$SEED, NA)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-4

  set.seed(1)
  expect_snapshot_RLum(calc_FadingCorr(
      age.faded = c(1, 0),
      g_value = c(5.0, 1.0),
      tc = 25920,
      tc.g_value = 172800,
      n.MC = 20, verbose = FALSE),
      tolerance = snapshot.tolerance)

  expect_snapshot_RLum(calc_FadingCorr(
      age.faded = c(10, 0),
      g_value = c(5.0, 1.0),
      tc = 2592000,
      tc.g_value = 172800,
      n.MC = 20, verbose = FALSE),
      tolerance = snapshot.tolerance)

  expect_snapshot_RLum(calc_FadingCorr(
      age.faded = c(1, 6),
      g_value = c(5.37778156709913, 0.70382588155986),
      tc = 378000, tc.g_value = 172800,
      n.MC = 1000, seed = 11, verbose = FALSE),
      tolerance = snapshot.tolerance)
})

#' @title Calculation of the source dose rate via the date of measurement
#'
#' @description
#' Calculating the dose rate of the irradiation source via the date of
#' measurement based on: source calibration date, source dose rate, dose rate
#' error. The function returns a data.frame that provides the input argument
#' dose_rate for the function [Luminescence::convert_Second2Gray].
#'
#' Calculation of the source dose rate based on the time elapsed since the last
#' calibration of the irradiation source. Decay parameters assume a Sr-90 beta
#' source.
#' \deqn{dose.rate = D0 \exp(-\frac{\log(2)}{T.1/2} t)}
#'
#' where `D0` is the calibration dose rate, \eqn{T.1/2} is the half-life of the
#' source nuclide (in days), `t` is the time since source calibration (in days),
#' and \eqn{\log(2) / T.1/2} equals the decay constant lambda.
#'
#' Information on the date of measurements may be taken from the data's
#' original .BIN file (using e.g., `BINfile <- Luminescence::read_BIN2R()` and the slot
#' `BINfile@METADATA$DATE`).
#'
#' **Allowed source types and related values**
#'
#' \tabular{rlll}{
#'  **#** \tab **Source type** \tab **T.1/2** \tab **Reference** \cr
#'  `[1]` \tab Sr-90 \tab 28.90 y \tab NNDC, Brookhaven National Laboratory \cr
#'  `[2]` \tab Am-214 \tab 432.6 y \tab NNDC, Brookhaven National Laboratory \cr
#'  `[3]` \tab Co-60 \tab 5.274 y \tab NNDC, Brookhaven National Laboratory \cr
#'  `[4]` \tab Cs-137 \tab 30.08 y \tab NNDC, Brookhaven National Laboratory}
#'
#' @param measurement.date [character] or [Date] (*with default*):
#' Date of measurement in `"YYYY-MM-DD"` format. If no value is provided, the
#' date will be set to today. The argument can be provided as vector.
#'
#' @param calib.date [character] or [Date] (**required**):
#' date of source calibration in `"YYYY-MM-DD"` format.
#'
#' @param calib.dose.rate [numeric] (**required**):
#' dose rate at date of calibration in Gy/s or Gy/min.
#'
#' @param calib.error [numeric] (**required**):
#' error of dose rate at date of calibration Gy/s or Gy/min.
#'
#' @param source.type [character] (*with default*):
#' specify irradiation source (`Sr-90`, `Co-60`, `Cs-137`, `Am-214`),
#' see details for further information.
#'
#' @param dose.rate.unit [character] (*with default*):
#' dose rate unit for input (one of `Gy/min` or `Gy/s`). The output is given in
#' Gy/s as valid for the function [Luminescence::convert_Second2Gray].
#'
#' @param predict [integer] (*with default*):
#' option allowing to predict the dose rate of the source over time in days
#' set by the provided value. Starting date is the value set with
#' `measurement.date`, e.g., `calc_SourceDoseRate(..., predict = 100)` calculates
#' the source dose rate for the next 100 days.
#'
#' @return
#' Returns an S4 object of type [Luminescence::RLum.Results-class].
#' Slot `data` contains a [list] with the following structure:
#'
#' ```
#' $ dose.rate (data.frame)
#' .. $ dose.rate
#' .. $ dose.rate.error
#' .. $ date (corresponding measurement date)
#' $ parameters (list)
#' .. $ source.type
#' .. $ halflife
#' .. $ dose.rate.unit
#' $ call (the original function call)
#' ```
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#' A plot method of the output is provided via [Luminescence::plot_RLum].
#'
#' @note
#' Please be careful when using the option `predict`, especially when a
#' multiple set for `measurement.date` and `calib.date` is provided. For the
#' source dose rate prediction, the function takes the last `measurement.date`
#' value and predicts from that the source dose rate for the number of days
#' requested, that is: the (multiple) original input will be replaced.
#' However, the function does not change entries for the calibration dates,
#' but mixes them up. Therefore, it is not recommended to use this option
#' when multiple calibration dates (`calib.date`) are provided.
#'
#' @section Function version: 0.3.4
#'
#' @author
#' Margret C. Fuchs, HZDR, Helmholtz-Institute Freiberg for Resource Technology (Germany) \cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::convert_Second2Gray], [Luminescence::get_RLum], [Luminescence::plot_RLum]
#'
#' @references
#' NNDC, Brookhaven National Laboratory `http://www.nndc.bnl.gov/`
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(1) Simple function usage
#' ##Basic calculation of the dose rate for a specific date
#' dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
#'                                   calib.date = "2014-12-19",
#'                                   calib.dose.rate = 0.0438,
#'                                   calib.error = 0.0019)
#'
#' ##show results
#' get_RLum(dose.rate)
#'
#' ##(2) Usage in combination with another function (e.g., convert_Second2Gray() )
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## use the calculated variable dose.rate as input argument
#' ## to convert De(s) to De(Gy)
#' convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate)
#'
#' ##(3) source rate prediction and plotting
#' dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
#'                                   calib.date = "2014-12-19",
#'                                   calib.dose.rate = 0.0438,
#'                                   calib.error = 0.0019,
#'                                   predict = 1000)
#' plot_RLum(dose.rate)
#'
#'##(4) export output to a LaTeX table (example using the package 'xtable')
#'\dontrun{
#' xtable::xtable(get_RLum(dose.rate))
#'}
#'
#' @export
calc_SourceDoseRate <- function(
  measurement.date,
  calib.date,
  calib.dose.rate,
  calib.error,
  source.type = c("Sr-90", "Am-214", "Co-60", "Cs-137"),
  dose.rate.unit = c("Gy/s", "Gy/min"),
  predict = NULL
) {
  .set_function_name("calc_SourceDoseRate")
  on.exit(.unset_function_name(), add = TRUE)

  if (missing(measurement.date))
    measurement.date <- Sys.Date()
  .validate_class(measurement.date, c("Date", "character"))
  if (is.character(measurement.date)) {
    measurement.date <- tryCatch(
        as.Date(measurement.date),
        error = function(e) {
          .throw_error("'measurement.date' could not be converted to a Date, ",
                       e$message)
        })
  }
  .validate_class(calib.date, c("Date", "character"))
  if (is.character(calib.date)) {
    calib.date <- tryCatch(
        as.Date(calib.date),
        error = function(e) {
          .throw_error("'calib.date' could not be converted to a Date, ",
                       e$message)
        })
  }
  .validate_positive_scalar(calib.error)

  ## source type and dose rate unit
  source.type <- .validate_args(source.type,
                                c("Sr-90", "Am-214", "Co-60", "Cs-137"))
  dose.rate.unit <- .validate_args(dose.rate.unit, c("Gy/s", "Gy/min"))

  # --- if predict is set
  if(!is.null(predict) && predict > 1){
    measurement.date <- seq(tail(measurement.date), by = 1, length = predict)
  }

  # -- calc days since source calibration
  decay.days <- measurement.date - calib.date


  # -- calc dose rate of source at date of measurement, considering the chosen source-type

  ##set halflife
  halflife.years  <- switch(
    source.type,
    "Sr-90" = 28.90,
    "Am-241" = 432.6,
    "Co-60" = 5.274,
    "Cs-137" = 30.08
    )

  halflife.days  <- halflife.years * .const$year_d

  # N(t) = N(0)*e^((lambda * t) with lambda = log(2)/T1.2)
  measurement.dose.rate <- (calib.dose.rate) *
    exp((-log(2) / halflife.days) * as.numeric(decay.days))
  measurement.dose.rate.error <- (calib.error) *
    exp((-log(2) / halflife.days) * as.numeric(decay.days))

  # -- convert to input unit to [Gy/s]
  if(dose.rate.unit == "Gy/min"){
    source.dose.rate <- measurement.dose.rate / 60
    source.dose.rate.error <- source.dose.rate *
      (measurement.dose.rate.error / measurement.dose.rate)

  }else if(dose.rate.unit == "Gy/s"){
    source.dose.rate <- measurement.dose.rate
    source.dose.rate.error <- measurement.dose.rate.error
  }

  # Output --------------------------------------------------------------------------------------

  dose_rate <- data.frame(
    dose.rate = source.dose.rate,
    dose.rate.error = source.dose.rate.error,
    date = measurement.date,
    stringsAsFactors = TRUE
  )

  set_RLum(
    class = "RLum.Results",
    data = list(
      dose.rate = dose_rate,
      parameters = list(source.type = source.type,
                        halflife = halflife.years,
                        dose.rate.unit = dose.rate.unit),
      call = sys.call()
    ))
}

test_that("test RLum.Analysis S3 methods", {
  testthat::skip_on_cran()

  data(ExampleData.RLum.Analysis, envir = environment())
  analysis <- IRSAR.RF.Data

  expect_silent(plot(analysis))
  expect_silent(hist(analysis))
  expect_type(summary(analysis), "list")
  expect_s4_class(subset(analysis), "RLum.Analysis")
  expect_equal(length(analysis), 2)
  expect_length(rep(analysis, 2), 2)
  expect_equal(names(analysis), c("RF (NA)", "RF (NA)"))
  expect_type(as.list(analysis), "list")
  expect_equal(is(analysis), c("RLum.Analysis", "RLum"))
  expect_s4_class(merge(analysis, analysis), "RLum.Analysis")
  expect_length(unlist(analysis, recursive = TRUE), 4)
  expect_length(unlist(analysis, recursive = FALSE), 2)
  expect_length(analysis[1], 1)
  expect_length(analysis["RF"], 2)
  expect_s4_class(analysis[[1]], "RLum.Data.Curve")
  expect_type(analysis[["RF"]], "list")
  expect_length(analysis$RF, 2)
})

test_that("test RLum.Results S3 methods", {
  testthat::skip_on_cran()

  result <- calc_SourceDoseRate(
    measurement.date = "2012-01-27",
    calib.date = "2014-12-19",
    calib.dose.rate = 0.0438,
    calib.error = 0.0019)

  expect_silent(plot(result))
  expect_silent(plot(list(result, result)))
  expect_silent(hist(result))
  expect_s3_class(summary(result), "data.frame")
  expect_equal(length(result), 3)
  expect_length(rep(result, 2), 2)
  expect_equal(names(result), c("dose.rate", "parameters", "call"))
  expect_type(as.list(result), "list")
  expect_equal(is(result), c("RLum.Results", "RLum"))
  expect_s4_class(merge(result, result), "RLum.Results")
  expect_visible(result[1])
  expect_visible(result[[1]])
  expect_visible(result$dose.rate)
})

test_that("test RLum.Data.Curve S3 methods", {
  testthat::skip_on_cran()

  data(ExampleData.CW_OSL_Curve, envir = environment())
  curve <- set_RLum(
      class = "RLum.Data.Curve",
      data = as.matrix(ExampleData.CW_OSL_Curve),
      curveType = "measured",
      recordType = "OSL",
      info = list(a = "test")
  )

  expect_silent(plot(curve))
  expect_silent(plot(list(curve, curve)))
  expect_silent(hist(curve))
  expect_s3_class(summary(curve), "table")
  expect_equal(length(curve), 40)
  expect_equal(dim(curve), c(1000, 2))
  expect_s4_class(bin(curve), "RLum.Data.Curve")
  expect_length(rep(curve, 2), 2)
  expect_equal(names(curve), "a")
  expect_s3_class(as.data.frame(curve), "data.frame")
  expect_type(as.list(curve), "list")
  expect_vector(as.matrix(curve))
  expect_equal(is(curve), c("RLum.Data.Curve", "RLum.Data", "RLum"))
  expect_s4_class(merge(curve, curve), "RLum.Data.Curve")
  expect_visible(curve + curve)
  expect_visible(curve - curve)
  expect_visible(curve * curve)
  expect_visible(curve / curve)
  expect_vector(curve[1])
  expect_equal(curve$a, c(a = "test"))
  expect_error(rep(curve, numeric()),
               "'times' should be a single positive integer value")
})

test_that("test RLum.Data.Image S3 methods", {
  testthat::skip_on_cran()

  data(ExampleData.RLum.Data.Image, envir = environment())
  image <- ExampleData.RLum.Data.Image
  image3 <- set_RLum("RLum.Data.Image",
                     data = array(runif(300, 0, 255), c(10, 10, 3)),
                     info = list(NumFrames = 3))

  expect_silent(plot(image))
  expect_silent(plot(list(image, image)))
  expect_silent(plot(image3))
  expect_silent(hist(image))
  expect_silent(hist(image3))
  expect_s3_class(summary(image), "table")
  expect_s3_class(summary(image3), "table")
  expect_length(rep(image, 2), 2)
  expect_equal(names(image)[1:3],
               c("ControllerVersion", "LogicOutput", "AmpHiCapLowNoise"))
  expect_equal(names(image3), "NumFrames")
  expect_type(as.list(image), "list")
  expect_length(as.list(image), 1)
  expect_length(as.list(image3), 3)
  expect_vector(as.matrix(image))
  expect_error(as.matrix(image3),
               "No viable coercion to matrix, object contains multiple frames")
  expect_equal(is(image), c("RLum.Data.Image", "RLum.Data", "RLum"))
  expect_error(merge(image, image),
               "Merging of 'RLum.Data.Image' objects is currently not supported")
  expect_vector(image[1])
  expect_error(image3[1],
               "No viable coercion to matrix, object contains multiple frames")
})

test_that("test RLum.Data.Spectrum S3 methods", {
  testthat::skip_on_cran()

  data(ExampleData.XSYG, envir = environment())
  spectrum <- TL.Spectrum

  expect_silent(plot(spectrum))
  expect_silent(plot(list(spectrum, spectrum)))
  expect_s4_class(bin(spectrum), "RLum.Data.Spectrum")
  expect_equal(dim(spectrum), c(1024, 24))
  expect_length(rep(spectrum, 2), 2)
  expect_equal(names(spectrum)[1:3], c("state", "parentID", "startDate"))
  expect_type(row.names(spectrum), "character")
  expect_s3_class(as.data.frame(spectrum), "data.frame")
  expect_type(as.list(spectrum), "list")
  expect_length(as.list(spectrum), 24)
  expect_vector(as.matrix(spectrum))
  expect_equal(is(spectrum), c("RLum.Data.Spectrum", "RLum.Data", "RLum"))
  expect_s4_class(merge(spectrum, spectrum), "RLum.Data.Spectrum")
  expect_visible(spectrum + spectrum)
  expect_visible(spectrum - spectrum)
  expect_visible(spectrum * spectrum)
  expect_visible(spectrum / spectrum)
  expect_vector(spectrum[1])
})

test_that("test Risoe.BINfileData S3 methods", {
  testthat::skip_on_cran()

  data(ExampleData.BINfileData, envir = environment())
  risoe <- CWOSL.SAR.Data

  expect_silent(plot(risoe, pos = 1))
  expect_silent(plot(risoe, run = 1, set = 2))
  expect_error(plot(list(risoe, risoe)))
  expect_error(subset(risoe, ERROR == 1),
               "Invalid subset options, valid terms are")
  expect_warning(subset(risoe, ID == 1, error = TRUE),
                 "Argument not supported and skipped")
  sel <- subset(risoe, ID == 1)
  expect_length(sel, 1)
  expect_length(sel@.RESERVED, 0)
  sel <- subset(risoe, ID == 1, records.rm = FALSE)
  expect_length(sel, 720)
  expect_equal(sel@METADATA$SEL, c(TRUE, rep(FALSE, 719)))
  sel <- subset(read_BIN2R(system.file("extdata/BINfile_V8.binx",
                                       package = "Luminescence"),
                           verbose = FALSE),
                POSITION == 1)
  expect_equal(nrow(sel@METADATA), length(sel@.RESERVED))
  expect_null(subset(risoe, ID == -1))
  expect_equal(length(risoe), 720)
  expect_equal(names(risoe)[1:40], c(rep("TL", 24), rep("OSL", 16)))
  expect_s3_class(as.data.frame(risoe), "data.frame")
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_SourceDoseRate(data.frame()),
               "'measurement.date' should be of class 'Date' or 'character'")
  expect_error(calc_SourceDoseRate("20241001"),
               "'measurement.date' could not be converted to a Date, character")
  expect_error(calc_SourceDoseRate("2024-10-01", data.frame()),
               "'calib.date' should be of class 'Date' or 'character'")
  expect_error(calc_SourceDoseRate("2024-10-01", "01.10.2024"),
               "'calib.date' could not be converted to a Date, character string")
  expect_error(calc_SourceDoseRate("2018-01-02", "2014-12-19", 0.0438, 0.0019,
                                   source.type = "error"),
               "'source.type' should be one of 'Sr-90', 'Am-214', 'Co-60' or")
  expect_error(calc_SourceDoseRate("2018-01-02", "2014-12-19", 0.0438, 0.0019,
                                   dose.rate.unit = "error"),
               "'dose.rate.unit' should be one of 'Gy/s' or 'Gy.min'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##simple run
  expect_silent(calc_SourceDoseRate(
    calib.date = "2014-12-19",
    calib.dose.rate = 0.0438,
    calib.error = 0.0019
  ))

  ## produce more than 100 rows
  t <- expect_silent(calc_SourceDoseRate(
    measurement.date = "2018-01-02",
    calib.date = "2014-12-19",
    calib.dose.rate = 0.0438,
    calib.error = 0.0019,
    predict = 150
  ))
  expect_silent(plot_RLum.Results(t))

  ##Gy/min as unit
  expect_silent(calc_SourceDoseRate(
    measurement.date = "2018-01-02",
    calib.date = "2014-12-19",
    calib.dose.rate = 0.0438,
    calib.error = 0.0019, dose.rate.unit = "Gy/min"
  ))
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(calc_SourceDoseRate(measurement.date = "2012-01-27",
                                           calib.date = "2014-12-19",
                                           calib.dose.rate = 0.0438,
                                           calib.error = 0.0019,
                                           predict = 10),
                       tolerance = snapshot.tolerance)
})

#' @title Calculate the Lx/Tx ratio for a given set of TL curves -beta version-
#'
#' @description
#' Calculate Lx/Tx ratio for a given set of TL curves.
#'
#' @details
#' **Uncertainty estimation**
#'
#' The standard errors are calculated using the following generalised equation:
#'
#' \deqn{SE_{signal} = abs(Signal_{net} * BG_f /BG_{signal})}
#'
#' where \eqn{BG_f} is a term estimated by calculating the standard deviation of the sum of
#' the \eqn{L_x} background counts and the sum of the \eqn{T_x} background counts. However,
#' if both signals are similar the error becomes zero.
#'
#' @param Lx.data.signal [Luminescence::RLum.Data.Curve-class] or [data.frame] (**required**):
#' TL data (x = temperature, y = counts) (TL signal).
#'
#' @param Lx.data.background [Luminescence::RLum.Data.Curve-class] or [data.frame] (*optional*):
#' TL data (x = temperature, y = counts).
#' If no data are provided no background subtraction is performed.
#'
#' @param Tx.data.signal [Luminescence::RLum.Data.Curve-class] or [data.frame] (**required**):
#' TL data (x = temperature, y = counts) (TL test signal).
#'
#' @param Tx.data.background [Luminescence::RLum.Data.Curve-class] or [data.frame] (*optional*):
#' TL data (x = temperature, y = counts).
#' If no data are provided no background subtraction is performed.
#'
#' @param signal_integral [integer] (**required**):
#' vector of channels for the signal integral.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given temperature range are selected.
#'
#' @param ... currently not used.
#'
#' @return
#' Returns an S4 object of type [Luminescence::RLum.Results-class].
#' Slot `data` contains a [list] with the following structure:
#'
#' ```
#' $ LxTx.table
#' .. $ LnLx
#' .. $ LnLx.BG
#' .. $ TnTx
#' .. $ TnTx.BG
#' .. $ Net_LnLx
#' .. $ Net_LnLx.Error
#' .. $ SN_RATIO_LnLx
#' .. $ SN_RATIO_TnTx
#' .. $ LxTx
#' .. $ LxTx.Error
#' ```
#'
#' @note
#' **This function has still BETA status!** Please further note that a similar
#' background for both curves results in a zero error and is therefore set to `NA`.
#'
#' @section Function version: 0.3.7
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Christoph Schmidt, University of Bayreuth (Germany)
#'
#' @seealso [Luminescence::RLum.Results-class], [Luminescence::analyse_SAR.TL]
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##load package example data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##convert Risoe.BINfileData into a curve object
#' temp <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos = 3)
#'
#' Lx.data.signal <- get_RLum(temp, record.id=1)
#' Lx.data.background <- get_RLum(temp, record.id=2)
#' Tx.data.signal <- get_RLum(temp, record.id=3)
#' Tx.data.background <- get_RLum(temp, record.id=4)
#' signal_integral <- 210:230
#'
#' output <- calc_TLLxTxRatio(
#'  Lx.data.signal,
#'  Lx.data.background,
#'  Tx.data.signal,
#'  Tx.data.background,
#'  signal_integral)
#' get_RLum(output)
#'
#' @export
calc_TLLxTxRatio <- function(
  Lx.data.signal,
  Lx.data.background = NULL,
  Tx.data.signal = NULL,
  Tx.data.background = NULL,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  ...
) {
  .set_function_name("calc_TLLxTxRatio")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(Lx.data.signal, c("data.frame", "RLum.Data.Curve"))
  if (ncol(Lx.data.signal) < 2)
    .throw_error("'Lx.data.signal' should have 2 columns")
  .validate_class(Tx.data.signal, c("data.frame", "RLum.Data.Curve"))
  if (ncol(Tx.data.signal) < 2)
    .throw_error("'Tx.data.signal' should have 2 columns")

  ##check DATA TYPE differences
   if(is(Lx.data.signal)[1] != is(Tx.data.signal)[1])
     .throw_error("Data types of Lx and Tx data differ")

  ##--------------------------------------------------------------------------##
  ## Type conversion (assuming that all input variables are of the same type)
  if(inherits(Lx.data.signal, "RLum.Data.Curve")){
    Lx.data.signal <- as(Lx.data.signal, "matrix")
    Tx.data.signal <- as(Tx.data.signal, "matrix")

    if (!is.null(Lx.data.background))
      Lx.data.background <- as(Lx.data.background, "matrix")

    if (!is.null(Tx.data.background))
      Tx.data.background <- as(Tx.data.background, "matrix")
  }

  ##(d) - check if Lx and Tx curves have the same channel length
  if (nrow(Lx.data.signal) != nrow(Tx.data.signal)) {
    .throw_error("Channel numbers differ for Lx and Tx data")
  }

  integral_input <- .validate_args(integral_input, c("channel", "measurement"))

  ## deprecated argument
  if (any(grepl("signal.integral", ...names(), fixed = TRUE))) {
    .deprecated(old = "signal.integral",
                new = "signal_integral",
                since = "1.2.0")
    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal_integral <- list(...)$signal.integral
  }

  if (integral_input == "measurement") {
    signal_integral <- .convert_to_channels(Lx.data.signal[, 1], signal_integral,
                                            "temperature")
  }
  signal_integral <- .validate_integral(signal_integral,
                                        max = length(Lx.data.signal[, 2]))

#  Background Consideration --------------------------------------------------
  LnLx.BG <- TnTx.BG <- NA
  signal.interval <- signal_integral

  ## Lx.data
  if (!is.null(Lx.data.background)) {
    if (ncol(Lx.data.background) < 2)
      .throw_error("'Lx.data.background' should have 2 columns")
    LnLx.BG <- sum(Lx.data.background[signal.interval, 2])
  }

  ## Tx.data
  if (!is.null(Tx.data.background)) {
    if (ncol(Tx.data.background) < 2)
      .throw_error("'Tx.data.background' should have 2 columns")
    TnTx.BG <- sum(Tx.data.background[signal.interval, 2])
  }

# Calculate Lx/Tx values --------------------------------------------------
    ## preset variables
    net_LnLx <- net_LnLx.Error <- net_TnTx <- net_TnTx.Error <- NA
    BG.Error <- NA

  ## calculate values
  LnLx <- sum(Lx.data.signal[signal.interval, 2])
  TnTx <- sum(Tx.data.signal[signal.interval, 2])

  ## calculate standard deviation of background
  if (!is.na(LnLx.BG) && !is.na(TnTx.BG)) {
       BG.Error <- sd(c(LnLx.BG, TnTx.BG))

       if(BG.Error == 0) {
         .throw_warning("The background signals for Lx and Tx appear ",
                        "to be similar, no background error was calculated")
         BG.Error <- NA
       }
  }

    ## calculate net LnLx
    if(!is.na(LnLx.BG)){
      net_LnLx <-  LnLx - LnLx.BG
      net_LnLx.Error <- abs(net_LnLx * BG.Error/LnLx.BG)
    }

    ## calculate net TnTx
    if(!is.na(TnTx.BG)){
         net_TnTx <-  TnTx - TnTx.BG
         net_TnTx.Error <- abs(net_TnTx * BG.Error/TnTx.BG)
    }

    ## calculate LxTx
    if(is.na(net_TnTx)){
      LxTx <- LnLx/TnTx
      LxTx.Error <- NA
    }else{
      LxTx <- net_LnLx/net_TnTx
      LxTx.Error <- abs(LxTx*((net_LnLx.Error/net_LnLx) + (net_TnTx.Error/net_TnTx)))
    }

    ##COMBINE into a data.frame
    temp.results <- data.frame(
      LnLx,
      LnLx.BG,
      TnTx,
      TnTx.BG,
      net_LnLx,
      net_LnLx.Error,
      net_TnTx,
      net_TnTx.Error,
      SN_RATIO_LnLx = LnLx / LnLx.BG,
      SN_RATIO_TnTx = TnTx / TnTx.BG,
      LxTx,
      LxTx.Error
    )

# Return values -----------------------------------------------------------
  set_RLum(
      class = "RLum.Results",
      data = list(LxTx.table = temp.results),
      info = list(call = sys.call())
  )
}

#' @title Calculate the Fast Ratio for CW-OSL curves
#'
#' @description
#' Function to calculate the fast ratio of quartz CW-OSL single grain or single
#' aliquot curves after Durcan & Duller (2011).
#'
#' This function follows the equations of Durcan & Duller (2011). The energy
#' required to reduce the fast and medium quartz OSL components to `x` and
#' `x2` % respectively using eq. 3 to determine channels L2 and L3 (start
#' and end). The fast ratio is then calculated from: \eqn{(L1-L3)/(L2-L3)}.
#'
#' @param object [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class] or [data.frame] (**required**):
#' x, y data of measured values (time and counts).
#'
#' @param stimulation.power [numeric] (*with default*):
#' Stimulation power in mW/cm²
#'
#' @param wavelength [numeric] (*with default*):
#' Stimulation wavelength in nm
#'
#' @param sigmaF [numeric] (*with default*):
#' Photoionisation cross-section (cm²) of the fast component.
#' Default value after Durcan & Duller (2011).
#'
#' @param sigmaM [numeric] (*with default*):
#' Photoionisation cross-section (cm²) of the medium component.
#' Default value after Durcan & Duller (2011).
#'
#' @param Ch_L1 [numeric] (*with default*):
#' An integer specifying the channel for L1.
#'
#' @param Ch_L2 [numeric] (*optional*):
#' An integer specifying the channel for L2.
#'
#' @param Ch_L3 [numeric] (*optional*):
#' A vector of length 2 with integer values specifying the start and end
#' channels for L3 (e.g., `c(40, 50)`), with the second component greater
#' than or equal to the first.
#'
#' @param x [numeric] (*with default*):
#' Percentage of signal remaining from the fast component.
#' Used to define the location of L2 and L3 (start).
#'
#' @param x2 [numeric] (*with default*):
#' Percentage of signal remaining from the medium component.
#' Used to define the location of L3 (end).
#'
#' @param dead.channels [numeric] (*with default*):
#' Vector of length 2 in the form of `c(x, y)`.
#' Channels that do not contain OSL data, i.e. at the start or end of measurement.
#'
#' @param fitCW.sigma [logical] (*optional*):
#' fit CW-OSL curve using [Luminescence::fit_CWCurve] to calculate `sigmaF` and `sigmaM` (**experimental**).
#'
#' @param fitCW.curve [logical] (*optional*):
#' fit CW-OSL curve using [Luminescence::fit_CWCurve] and derive the counts of L2 and L3
#' from the fitted OSL curve (**experimental**).
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... available options: `verbose` ([logical]).
#' Further arguments passed to [Luminescence::fit_CWCurve].
#'
#' @return
#' Returns a plot (*optional*) and an S4 object of type [Luminescence::RLum.Results-class].
#' The slot `data` contains a [list] with the following elements:
#'
#' \item{summary}{[data.frame] summary of all relevant results}
#' \item{data}{the original input data}
#' \item{fit}{[Luminescence::RLum.Results-class] object if either `fitCW.sigma` or `fitCW.curve` is `TRUE`}
#' \item{args}{[list] of used arguments}
#' \item{call}{[call] the function call}
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Georgina E. King, University of Bern (Switzerland) \cr
#' Julie A. Durcan, University of Oxford (United Kingdom) \cr
#' Christoph Burow, University of Cologne (Germany)
#'
#' @references
#' Durcan, J.A. & Duller, G.A.T., 2011. The fast ratio: A rapid measure for testing
#' the dominance of the fast component in the initial OSL signal from quartz.
#' Radiation Measurements 46, 1065-1072.
#'
#' Madsen, A.T., Duller, G.A.T., Donnelly, J.P., Roberts, H.M. & Wintle, A.G., 2009.
#' A chronology of hurricane landfalls at Little Sippewissett Marsh, Massachusetts, USA,
#' using optical dating. Geomorphology 109, 36-45.
#'
#' **Further reading**
#'
#' Steffen, D., Preusser, F. & Schlunegger, 2009. OSL quartz age underestimation
#' due to unstable signal components. Quaternary Geochronology 4, 353-362.
#'
#'
#' @seealso [Luminescence::fit_CWCurve], [Luminescence::get_RLum], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class], [Luminescence::RLum.Data.Curve-class]
#'
#' @examples
#' # load example CW-OSL curve
#' data("ExampleData.CW_OSL_Curve")
#'
#' # calculate the fast ratio w/o further adjustments
#' res <- calc_FastRatio(ExampleData.CW_OSL_Curve)
#'
#' # show the summary table
#' get_RLum(res)
#'
#' @export
calc_FastRatio <- function(object,
                           stimulation.power = 30.6,
                           wavelength = 470,
                           sigmaF = 2.6E-17,
                           sigmaM = 4.28E-18,
                           Ch_L1 = 1,
                           Ch_L2 = NULL,
                           Ch_L3 = NULL,
                           x = 1,
                           x2 = 0.1,
                           dead.channels = c(0,0),
                           fitCW.sigma = FALSE,
                           fitCW.curve = FALSE,
                           plot = TRUE,
                           ...) {
  .set_function_name("calc_FastRatio")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Analysis", "RLum.Results", "RLum.Data.Curve",
                            "data.frame", "matrix"))
  .validate_not_empty(object)
  .validate_positive_scalar(stimulation.power)
  .validate_positive_scalar(wavelength)
  .validate_positive_scalar(Ch_L1, int = TRUE)
  .validate_positive_scalar(Ch_L2, int = TRUE, null.ok = TRUE)
  .validate_class(Ch_L3, c("integer", "numeric"), null.ok = TRUE, length = 2)
  if (!is.null(Ch_L3)) {
    .validate_positive_scalar(Ch_L3[1], int = TRUE, name = "'Ch_L3[1]'")
    .validate_positive_scalar(Ch_L3[2], int = TRUE, name = "'Ch_L3[2]'")
    if (Ch_L3[1] > Ch_L3[2]) {
      .throw_error("'Ch_L3[2]' must be greater than or equal to 'Ch_L3[1]'")
    }
  }
  .validate_positive_scalar(wavelength)
  .validate_positive_scalar(sigmaF)
  .validate_positive_scalar(sigmaM)
  .validate_positive_scalar(x)
  .validate_positive_scalar(x2)
  .validate_class(dead.channels, c("integer", "numeric"), length = 2)
  if (any(dead.channels < 0)) {
    .throw_error("All elements of 'dead.channels' should be non-negative")
  }
  .validate_logical_scalar(fitCW.sigma)
  .validate_logical_scalar(fitCW.curve)
  .validate_logical_scalar(plot)

  ## Input object handling -----------------------------------------------------
  if (inherits(object, "RLum.Analysis"))
    object <- get_RLum(object)

  if (inherits(object, "RLum.Results"))
    object <- get_RLum(object, "data")

  if ((is.data.frame(object) || is.matrix(object)) && ncol(object) < 2) {
    .throw_error("'object' should have at least two columns")
  }

  if (!inherits(object, "list"))
    object <-list(object)


  ## Settings ------------------------------------------------------------------
  settings <- list(verbose = TRUE,
                   n.components.max = 3,
                   fit.method = "LM",
                   output.terminal = FALSE,
                   info = list(),
                   fit = NULL)

  # override defaults with args in ...
  settings <- modifyList(settings, list(...))

  .validate_logical_scalar(settings$verbose, name = "'verbose'")
  .validate_logical_scalar(settings$output.terminal, name = "'output.terminal'")

  ## Calculations --------------------------------------------------------------
  # iterate over all user provided objects and calculate the FR
  fast.ratios <- lapply(object, function(obj) {

    if (inherits(obj, "RLum.Data.Curve"))
      A <- get_RLum(obj)
    else
      A <- obj

    ## Energy calculation
    # P = user defined stimulation power in mW
    # lambdaLED = wavelength of stimulation source in nm
    P <- stimulation.power
    lamdaLED <- wavelength

    ## c = speed of light, h = Planck's constant
    I0 <- (P / 1000) / (.const$c * .const$h / (lamdaLED * 10^-9))
    Ch_width <- max(A[ ,1]) / length(A[ ,1])

    # remove dead channels
    A <- as.data.frame(A[(dead.channels[1] + 1):(nrow(A)-dead.channels[2]), ,
                         drop = FALSE])
    A[ ,1] <- A[ ,1] - A[1,1]

    ## remove missing values
    A <- na.exclude(A[, 1:2])
    if (nrow(A) == 0)
      .throw_error("After NA removal, nothing is left from the data set")

    # estimate the photo-ionisation cross-sections of the fast and medium
    # component using the fit_CWCurve function
    if (fitCW.sigma | fitCW.curve) {
      fitCW.res <- try(fit_CWCurve(A, n.components.max = settings$n.components.max,
                                   fit.method = settings$fit.method,
                                   LED.power = stimulation.power,
                                   LED.wavelength = wavelength,
                                   verbose = settings$output.terminal,
                                   plot = plot), outFile = stdout())

      if (!inherits(fitCW.res, "try-error")) {
        settings$fit <- fitCW.res
        if (fitCW.sigma) {
          sigmaF <- get_RLum(fitCW.res)$cs1
          sigmaM <- get_RLum(fitCW.res)$cs2
          if (settings$verbose) {
            message("\n [calc_FitCWCurve()]\n")
            message("New value for sigmaF: ", format(sigmaF, digits = 3, nsmall = 2))
            message("New value for sigmaM: ", format(sigmaM, digits = 3, nsmall = 2))
          }
        }

        if (fitCW.curve) {
          nls <- get_RLum(fitCW.res, "fit")
          A[ ,2] <- predict(nls)
        }
      } else {
        settings["fit"] <- list(NULL)
        if (settings$verbose)
          .throw_message("Fitting failed, please call 'fit_CWCurve()' ",
                         "manually before calculating the fast ratio")
      }
    }

    ## The equivalent time in s of L1, L2, L3
    # Use these values to look up the channel
    t_L1 <- 0

    if (is.null(Ch_L2))
      t_L2 <- (log(x / 100)) / (-sigmaF * I0)
    else
      t_L2 <- A[Ch_L2, 1]

    if (is.null(Ch_L3)) {
      t_L3_start <- (log(x / 100)) / (-sigmaM * I0)
      t_L3_end <- (log(x2 / 100)) / (-sigmaM * I0)
    } else {
      if (any(Ch_L3 > nrow(A))) {
        .throw_error("Value in 'Ch_L3' (", .collapse(Ch_L3, quote = FALSE),
                     ") exceeds number of available channels (", nrow(A), ")")
      }
      t_L3_start <- A[Ch_L3[1], 1]
      t_L3_end <- A[Ch_L3[2], 1]
    }

    ## Channel number(s) of L2 and L3
    if (is.null(Ch_L2))
      Ch_L2 <- which.min(abs(A[,1] - t_L2))

    if (Ch_L2 <= 1) {
      msg <- sprintf("Calculated time/channel for L2 is too small (%.f, %.f), NULL returned",
                     t_L2, Ch_L2)
      settings$info <- modifyList(settings$info, list(L2 = msg))
      .throw_warning(msg)
      return(NULL)
    }

    Ch_L3st<- which.min(abs(A[,1] - t_L3_start))
    Ch_L3end <- which.min(abs(A[,1] - t_L3_end))

    ## Counts in channels L1, L2, L3
    # L1 ----
    Cts_L1 <- A[Ch_L1, 2]

    # L2 ----
    if (Ch_L2 > nrow(A)) {
      msg <- sprintf(paste("The calculated channel for L2 (%i) exceeds",
                           "the number of available channels (%i),",
                           "NULL returned"), Ch_L2, nrow(A))
      settings$info <- modifyList(settings$info, list(L2 = msg))
      .throw_warning(msg)
      return(NULL)
    }
    Cts_L2 <- A[Ch_L2, 2]

    # optional: predict the counts from the fitted curve
    if (fitCW.curve && !inherits(fitCW.res, "try-error")) {
        nls <- get_RLum(fitCW.res, "fit")
        Cts_L2 <- predict(nls, list(x = t_L2))
    }

    # L3 ----
    if (Ch_L3st >= nrow(A) | Ch_L3end > nrow(A)) {
      msg <- sprintf(paste("The calculated channels for L3 (%i, %i) exceed",
                           "the number of available channels (%i).",
                           "\nThe background has instead been estimated from the last",
                           "5 channels."), Ch_L3st, Ch_L3end, nrow(A))
      settings$info <- modifyList(settings$info, list(L3 = msg))
      .throw_warning(msg)
      Ch_L3st <- max(nrow(A) - 5, 1)
      Ch_L3end <- nrow(A)
      t_L3_start <- A[Ch_L3st,1]
      t_L3_end <- A[Ch_L3end,1]
    }
    Cts_L3 <- mean(A[Ch_L3st:Ch_L3end, 2])

    # optional: predict the counts from the fitted curve
    if (fitCW.curve && !inherits(fitCW.res, "try-error")) {
        nls <- get_RLum(fitCW.res, "fit")
        Cts_L3 <- mean(predict(nls, list(x = c(t_L3_start, t_L3_end))))
    }

    # Warn if counts are not in decreasing order
    if (Cts_L3 >= Cts_L2)
      .throw_warning(sprintf("L3 contains more counts (%.f) than L2 (%.f)",
                             Cts_L3, Cts_L2))

    ## Fast Ratio
    FR <- (Cts_L1 - Cts_L3) / (Cts_L2 - Cts_L3)

    ## Fast Ratio - Error calculation
    FR_se <- NA
    FR_rse <- NA
    if (!is.na(FR)) {

      # number of channels the background was derived from
      nBG <- abs(Ch_L3end - Ch_L3st)

      # relative standard errors
      rse_L1 <- sqrt(Cts_L1 + Cts_L3 / nBG) / (Cts_L1 - Cts_L3)
      rse_L2 <- sqrt(Cts_L2 + Cts_L3 / nBG) / (Cts_L2 - Cts_L3)

      # absolute standard errors
      se_L1 <- rse_L1 * (Cts_L1 - Cts_L3)
      se_L2 <- rse_L2 * (Cts_L2 - Cts_L3)

      # absolute standard error on fast ratio
      FR_se <- (sqrt((se_L1 / (Cts_L1 - Cts_L3))^2 + ((se_L2 / (Cts_L2 - Cts_L3))^2) )) * FR
      FR_rse <- FR_se / FR * 100
    }

    ## Return values -----------------------------------------------------------
    summary <- data.frame(fast.ratio = FR,
                          fast.ratio.se = FR_se,
                          fast.ratio.rse = FR_rse,
                          channels = nrow(A),
                          channel.width = Ch_width,
                          dead.channels.start = as.integer(dead.channels[1]),
                          dead.channels.end = as.integer(dead.channels[2]),
                          sigmaF = sigmaF,
                          sigmaM = sigmaM,
                          I0 = I0,
                          stimulation.power = stimulation.power,
                          wavelength = wavelength,
                          t_L1 = t_L1,
                          t_L2 = t_L2,
                          t_L3_start = t_L3_start,
                          t_L3_end = t_L3_end,
                          Ch_L1 = as.integer(Ch_L1),
                          Ch_L2 = as.integer(Ch_L2),
                          Ch_L3_start = as.integer(Ch_L3st),
                          Ch_L3_end = as.integer(Ch_L3end),
                          Cts_L1 = Cts_L1,
                          Cts_L2 = Cts_L2,
                          Cts_L3 = Cts_L3)

    fast.ratio <- set_RLum(class = "RLum.Results",
                           originator = "calc_FastRatio",
                           data = list(summary = summary,
                                       data = obj,
                                       fit = settings$fit,
                                       args = as.list(sys.call(-2L)[-1]),
                                       call = sys.call(-2L)),
                           info = settings$info
    )

    ## Console Output ----------------------------------------------------------
    if (settings$verbose) {

      table.names <- c(
        "Fast Ratio\t", " \U02EA Absolute error", " \U02EA Relative error (%)", "Channels\t",
        "Channel width (s)", "Dead channels start", "Dead channels end",
        "Sigma Fast\t", "Sigma Medium\t", "I0\t\t", "Stim. power (mW/cm^2)", "Wavelength (nm)",
        "-\n Time L1 (s)\t", "Time L2 (s)\t", "Time L3 start (s)", "Time L3 end (s)",
        "-\n Channel L1\t", "Channel L2\t", "Channel L3 start", "Channel L3 end\t",
        "-\n Counts L1\t", "Counts L2\t", "Counts L3\t")

      cat("\n[calc_FastRatio()]\n")
      cat("\n -------------------------------")
      for (i in 1:ncol(summary)) {
        cat(paste0("\n ", table.names[i],"\t: ",
                   format(summary[1, i], digits = 2, nsmall = 2)))
      }
      cat("\n -------------------------------\n\n")
    }

    ## Plotting -------------------------------------------------------------
    if (plot)
      try(plot_RLum.Results(fast.ratio, ...))

    # return
    return(fast.ratio)
  }) # End of lapply

  if (length(fast.ratios) == 1)
    fast.ratios <- fast.ratios[[1]]

  invisible(fast.ratios)
}

## load data
data(ExampleData.CW_OSL_Curve, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  obj <- ExampleData.CW_OSL_Curve
  expect_error(calc_FastRatio("error"),
               "'object' should be of class 'RLum.Analysis', 'RLum.Results'")
  expect_error(calc_FastRatio(iris[0, ]),
               "'object' cannot be an empty data.frame")
  expect_error(calc_FastRatio(matrix(nrow = 1, ncol = 0)),
               "'object' cannot be an empty matrix")
  expect_error(calc_FastRatio(matrix()),
               "'object' should have at least two columns")

  expect_error(calc_FastRatio(obj, stimulation.power = iris),
               "'stimulation.power' should be a single positive value")
  expect_error(calc_FastRatio(obj, wavelength = iris),
               "'wavelength' should be a single positive value")

  expect_error(calc_FastRatio(obj, Ch_L1 = NULL),
               "'Ch_L1' should be a single positive integer value")
  expect_error(calc_FastRatio(obj, Ch_L1 = 0),
               "'Ch_L1' should be a single positive integer value")
  expect_error(calc_FastRatio(obj, Ch_L1 = c(1, 2)),
               "'Ch_L1' should be a single positive integer value")

  expect_error(calc_FastRatio(obj, Ch_L2 = 0),
               "'Ch_L2' should be a single positive integer value")
  expect_error(calc_FastRatio(obj, Ch_L2 = c(1, 2)),
               "'Ch_L2' should be a single positive integer value")

  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve,
                              Ch_L3 = 50),
               "'Ch_L3' should be of class 'integer', 'numeric' or NULL and have length 2")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve,
                              Ch_L3 = c(40, 50, 60)),
               "'Ch_L3' should be of class 'integer', 'numeric' or NULL and have length 2")
  expect_error(calc_FastRatio(obj, Ch_L3 = list(4, 5)),
               "'Ch_L3' should be of class 'integer', 'numeric' or NULL")
  expect_error(calc_FastRatio(obj, Ch_L3 = c(0, 2)),
               "'Ch_L3[1]' should be a single positive integer value",
               fixed = TRUE)
  expect_error(calc_FastRatio(obj, Ch_L3 = c(5, 2)),
               "'Ch_L3[2]' must be greater than or equal to 'Ch_L3[1]'",
               fixed = TRUE)
  expect_error(calc_FastRatio(obj, Ch_L3 = c(5, 1001)),
               "Value in 'Ch_L3' (5, 1001) exceeds number of available channels",
               fixed = TRUE)

  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, wavelength=0),
               "'wavelength' should be a single positive value")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, sigmaF= 0),
               "'sigmaF' should be a single positive value")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, sigmaM= 0),
               "'sigmaM' should be a single positive value")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, x = -12),
               "'x' should be a single positive value")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, x2 = -12),
               "'x2' should be a single positive value")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, dead.channels = TRUE),
               "'dead.channels' should be of class 'integer' or 'numeric' and")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, dead.channels = 1),
               "'dead.channels' should be of class 'integer' or 'numeric' and have length 2")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, dead.channels = c(-1, 1)),
               "All elements of 'dead.channels' should be non-negative")
  expect_error(calc_FastRatio(data.frame(NA, 1:4, 1:4)),
               "After NA removal, nothing is left from the data set")
  expect_error(calc_FastRatio(data.frame(1:4, NA)),
               "After NA removal, nothing is left from the data set")
  expect_error(calc_FastRatio(ExampleData.CW_OSL_Curve, verbose = NA),
               "'verbose' should be a single logical value")

  expect_warning(expect_null(calc_FastRatio(ExampleData.CW_OSL_Curve,
                                            Ch_L2 = 1)),
                 "Calculated time/channel for L2 is too small (0, 1)",
                 fixed = TRUE)
  expect_warning(expect_null(calc_FastRatio(ExampleData.CW_OSL_Curve,
                                            Ch_L2 = 2000)),
                 "The calculated channel for L2 (2000) exceeds the number",
                 fixed = TRUE)
  SW({
  expect_warning(calc_FastRatio(ExampleData.CW_OSL_Curve,
                                Ch_L3 = c(1000, 1000)),
                 "The calculated channels for L3 (1000, 1000) exceed",
                 fixed = TRUE)
  })
})

test_that("check functionality", {
  testthat::skip_on_cran()

  temp <- calc_FastRatio(ExampleData.CW_OSL_Curve, plot = FALSE,
                         verbose = FALSE)
  expect_s4_class(temp, "RLum.Results")
  expect_equal(length(temp), 5)

  ## fitCW.sigma and fitCW.curve
  SW({
  calc_FastRatio(ExampleData.CW_OSL_Curve, plot = FALSE,
                 fitCW.sigma = TRUE, fitCW.curve = TRUE)

  ## RLum.Analysis object
  data(ExampleData.XSYG, envir = environment())
  calc_FastRatio(OSL.SARMeasurement$Sequence.Object)

  expect_warning(calc_FastRatio(get_RLum(TL.Spectrum)),
                 "L3 contains more counts (566) than L2 (562)",
                 fixed = TRUE)

  ## RLum.Results object
  calc_FastRatio(temp)
  })

  ## RLum.Data.Curve object
  curve <- set_RLum("RLum.Data.Curve", data = as.matrix(ExampleData.CW_OSL_Curve))
  expect_s4_class(calc_FastRatio(curve, plot = FALSE, verbose = FALSE),
                  "RLum.Results")

  results <- get_RLum(temp)
  expect_equal(round(results$fast.ratio, digits = 3), 405.122)
  expect_equal(round(results$fast.ratio.se, digits = 4), 119.7442)
  expect_equal(round(results$fast.ratio.rse, digits = 5), 29.55756)
  expect_equal(results$channels, 1000)
  expect_equal(round(results$channel.width, digits = 2), 0.04)
  expect_equal(results$dead.channels.start, 0)
  expect_equal(results$dead.channels.end, 0)
  expect_equal(results$sigmaF, 2.6e-17)
  expect_equal(results$sigmaM, 4.28e-18)
  expect_equal(results$stimulation.power, 30.6)
  expect_equal(results$wavelength, 470)
  expect_equal(results$t_L1, 0)
  expect_equal(round(results$t_L2, digits = 6), 2.446413)
  expect_equal(round(results$t_L3_start, digits = 5), 14.86139)
  expect_equal(round(results$t_L3_end, digits = 5), 22.29208)
  expect_equal(results$Ch_L1, 1)
  expect_equal(results$Ch_L2, 62)
  expect_equal(results$Ch_L3_start, 373)
  expect_equal(results$Ch_L3_end, 558)
  expect_equal(results$Cts_L1, 11111)
  expect_equal(results$Cts_L2, 65)
  expect_equal(round(results$Cts_L3, digits = 5), 37.66667)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("FastRatio defaults",
                              calc_FastRatio(ExampleData.CW_OSL_Curve))
  })
})

test_that("regression tests", {

  ## issue 471 --------------------------------------------------------------

  expect_s4_class(suppressWarnings(
      calc_FastRatio(ExampleData.CW_OSL_Curve[1:3, ], verbose = FALSE)),
      "RLum.Results")

  expect_s4_class(suppressWarnings(
      calc_FastRatio(ExampleData.CW_OSL_Curve[1:5, ], verbose = FALSE)),
      "RLum.Results")

  SW({
  expect_message(expect_s4_class(
      calc_FastRatio(ExampleData.CW_OSL_Curve, fitCW.sigma=TRUE,
                     n.components.max = 0),
      "RLum.Results"),
      "Error: Fitting failed, please call 'fit_CWCurve()' manually",
      fixed = TRUE)

  set.seed(1)
  expect_message(expect_s4_class(
      calc_FastRatio(ExampleData.CW_OSL_Curve[sample(50), ],
                     fitCW.sigma = TRUE),
      "RLum.Results"),
      "Error: Fitting failed, please call 'fit_CWCurve()' manually",
      fixed = TRUE)
  })

  ## issue 932
  expect_warning(expect_null(calc_FastRatio(matrix(rnorm(5), 1, 5))),
                 "Calculated time/channel for L2 is too small")
})

## load data
data(ExampleData.RLum.Analysis)
data(ExampleData.RLum.Data.Image, envir = environment())
data(ExampleData.XSYG, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(view(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_error(view(set_RLum("RLum.Data.Image")),
               "'object' cannot be an empty RLum.Data")
  expect_error(view(set_RLum("RLum.Results")),
               "'object' cannot be an empty RLum.Results")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## Risø.BINfileData
  file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  risoe <- read_BIN2R(file, verbose = FALSE)

  ## RLum.Data.Curve
  curve <- get_RLum(OSL.SARMeasurement$Sequence.Object,
                    recordType = "TL (UVVIS)")[[1]]

  ## RLum.Data.Image
  image <- ExampleData.RLum.Data.Image

  ## RLum.Data.Spectrum
  spect <- TL.Spectrum

  ## RLum.Results
  results <- calc_FastRatio(curve, verbose = FALSE, plot = FALSE)

  ## RLum.Analysis
  analysis <- OSL.SARMeasurement[[2]]

  ## we need to mock the utils::View function as it otherwise generates this
  ## error on CI:
  ##   Error in `check_screen_device("View()")`:
  ##     View() should not be used in examples etc
  with_mocked_bindings(".view" = function(x, ...) NULL, {
    expect_null(view(risoe))
    expect_null(view(curve))
    expect_null(view(image))
    expect_null(view(spect))
    expect_null(view(results))
    expect_null(view(analysis))
  })
})

#' @title Plot Single Luminescence Curves from a Risoe.BINfileData-class object
#'
#' @description
#' Plots single luminescence curves from a [Luminescence::Risoe.BINfileData-class] object
#' produced by [Luminescence::read_BIN2R].
#'
#' @details
#' **Nomenclature**
#'
#' See [Luminescence::Risoe.BINfileData-class]
#'
#' **curve.transformation**
#'
#' This argument allows transforming continuous wave (CW) curves to pseudo
#' (linear) modulated curves. For the transformation, the functions of the
#' package are used.  Currently, it is not possible to pass further arguments
#' to the transformation functions. The argument works only for `ltype`
#' `OSL` and `IRSL`.
#'
#' **Irradiation time**
#'
#' Plotting the irradiation time (s) or the given dose (Gy) requires that the
#' variable `IRR_TIME` has been set within the BIN-file.  This is normally
#' done by using the 'Run Info' option within the Sequence Editor or by editing
#' in R.
#'
#' @param data [Luminescence::Risoe.BINfileData-class] (**required**):
#' an S4 object generated by the [Luminescence::read_BIN2R] function.
#'
#' @param position [vector] (*optional*):
#' option to limit the plotted curves by position
#' (e.g. `position = 1`, `position = c(1,3,5)`).
#'
#' @param run [vector] (*optional*):
#' option to limit the plotted curves by run
#' (e.g., `run = 1`, `run = c(1,3,5)`).
#'
#' @param set [vector] (*optional*):
#' option to limit the plotted curves by set
#' (e.g., `set = 1`, `set = c(1,3,5)`).
#'
#' @param sorter [character] (*with default*):
#' ordering used in plotting the curves, one of "POSITION", "SET" or "RUN".
#' POSITION, SET and RUN are options defined in the Risoe Sequence Editor.
#'
#' @param ltype [character] (*with default*):
#' option to limit the plotted curves by the type of luminescence stimulation.
#' Allowed values: `"IRSL"`, `"OSL"`,`"TL"`, `"RIR"`, `"RBR"`
#' (corresponds to LM-OSL), `"RL"`.  All type of curves are plotted by
#' default.
#'
#' @param curve.transformation [character] (*optional*):
#' allows transforming CW-OSL and CW-IRSL curves to pseudo-LM curves via
#' transformation functions. Allowed values are: `CW2pLM`, `CW2pLMi`,
#' `CW2pHMi` and `CW2pPMi`, see details. If set to `None` (default), no
#' transformation is applied.
#'
#' @param dose_rate [numeric] (*optional*):
#' dose rate of the irradiation source at the measurement date.
#' If set, the given irradiation dose will be shown in Gy.  See details.
#'
#' @param temp.lab [character] (*with default*):
#' the temperature unit to display in the plot labels, by default deg C.
#'
#' @param cex.global [numeric] (*with default*):
#' global scaling factor.
#'
#' @param ... further undocumented plot arguments.
#'
#' @return Returns a plot.
#'
#' @note
#' The function has been successfully tested for the Sequence Editor file
#' output version 3 and 4.
#'
#' @section Function version: 0.4.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData-class], [Luminescence::read_BIN2R],
#' [Luminescence::convert_CW2pLM], [Luminescence::convert_CW2pLMi],
#' [Luminescence::convert_CW2pPMi], [Luminescence::convert_CW2pHMi]
#'
#' @references
#' Duller, G., 2007. Analyst. pp. 1-45.
#'
#' @keywords dplot
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##plot all curves from the first position to the desktop
#' #pdf(file = "~/Desktop/CurveOutput.pdf", paper = "a4", height = 11, onefile = TRUE)
#'
#' ##example - load from *.bin file
#' #BINfile<- file.choose()
#' #BINfileData<-read_BIN2R(BINfile)
#'
#' #par(mfrow = c(4,3), oma = c(0.5,1,0.5,1))
#' #plot_Risoe.BINfileData(CWOSL.SAR.Data,position = 1)
#' #mtext(side = 4, BINfile, outer = TRUE, col = "blue", cex = .7)
#' #dev.off()
#'
#' @export
plot_Risoe.BINfileData<- function(
  data,
  position = NULL,
  run = NULL,
  set = NULL,
  sorter = c("POSITION", "RUN", "SET"),
  ltype = c("IRSL","OSL","TL","RIR","RBR","RL"),
  curve.transformation = "None",
  dose_rate = NULL,
  temp.lab = "\u00B0C",
  cex.global = 1,
  ...
) {
  .set_function_name("plot_Risoe.BINfileData")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, "Risoe.BINfileData")
  .validate_class(position, c("integer", "numeric"), null.ok = TRUE)
  .validate_class(run, c("integer", "numeric"), null.ok = TRUE)
  .validate_class(set, c("integer", "numeric"), null.ok = TRUE)
  sorter <- .validate_args(sorter, c("POSITION", "RUN", "SET"))
  .validate_positive_scalar(dose_rate, null.ok = TRUE)
  curve.transformation <- .validate_args(curve.transformation,
                                         c("CW2pLM", "CW2pLMi",
                                           "CW2pHMi", "CW2pPMi", "None"))
  ## complete the function name
  if (curve.transformation != "None") {
    curve.transformation <- paste0("convert_", curve.transformation)
  }

  temp <- data@METADATA

  ## set values for missing arguments
  if (is.null(position))
    position <- sort(unique(temp[, "POSITION"]))
  if (is.null(run))
    run <- sort(unique(temp[, "RUN"]))
  if (is.null(set))
    set <- sort(unique(temp[, "SET"]))

  fun <- isTRUE(list(...)$fun)

  ## Plot -------------------------------------------------------------------

  ## keep only the curves to plot
  temp <- temp[temp[, "POSITION"] %in% position &
               temp[, "RUN"] %in% run &
               temp[, "SET"] %in% set &
               temp[, "LTYPE"] %in% ltype, ]

  ## order by RUN, SET OR BY POSITION
  temp <- temp[order(temp[, sorter]), ]

  ## plot the selected curves
  for (i in seq_len(nrow(temp))) {
      high <- temp[i, "HIGH"]
      npoints <- temp[i, "NPOINTS"]
      ltype <- temp[i, "LTYPE"]
      is.TL <- ltype == "TL"
      measured_unit <- if (is.TL) temp.lab else "s"

      ##set x and y values
      values.xy <- data.frame(x = seq(high / npoints, high, by = high / npoints),
                              y = unlist(data@DATA[temp[i, "ID"]]))

      ##set curve transformation if wanted
      if (grepl("IRSL|OSL", ltype) &&
          curve.transformation != "None") {

        ## get the actual function from the parameter value and apply it
        values.xy <- get(curve.transformation)(values.xy)[, 1:2]
      }

      ##plot graph
      plot(values.xy,
           main = paste0("pos=", temp[i, "POSITION"],
                         ", run=", temp[i, "RUN"],
                         ", set=", temp[i, "SET"]),
           type="l",
           xlab = if (is.TL) paste0("Temperature [", temp.lab, "]") else "Time [s]",
           ylab = paste0(ltype, " [cts/", round(high / npoints, 3),
                         " ", measured_unit, "]"),
           col = if (ltype %in% c("IRSL", "RIR")) "red"
                 else if (ltype %in% c("OSL", "RBR")) "blue"
                 else "black",
           sub = if (is.TL) paste0("(", temp[i, "RATE"], " K/s)"),
           lwd=1.2*cex.global,
           cex=0.9*cex.global
      )

      ##add mtext for temperature
      temp.field <- if (temp[i, "VERSION"] == "03") "AN_TEMP" else "TEMPERATURE"
      temperature <- temp[i, temp.field]
      mtext(side=3,
            if (is.TL) paste("TL to", high, temp.lab)
            else paste0(ltype, "@", temperature, " ", temp.lab),
            cex=0.9*cex.global)

      ##add mtext for irradiation
      if (temp[i, "IRR_TIME"] != 0) {
        mtext(side = 4, cex = 0.8 * cex.global, line = 0.5,
              if (is.null(dose_rate))
                paste("dose =", temp[i, "IRR_TIME"], "s")
              else
                paste("dose =", temp[i, "IRR_TIME"] * dose_rate, "Gy")
              )
      }
  }#endforloop

  if (fun) sTeve() # nocov
}

## load data
data(ExampleData.BINfileData, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_Risoe.BINfileData("error"),
               "'data' should be of class 'Risoe.BINfileData'")
  expect_error(plot_Risoe.BINfileData(CWOSL.SAR.Data, sorter = 1),
               "'sorter' should be one of 'POSITION', 'RUN' or 'SET'")
  expect_error(plot_Risoe.BINfileData(CWOSL.SAR.Data, dose_rate = c(3, 4)),
               "'dose_rate' should be a single positive value")
  expect_error(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                      curve.transformation = "error"),
               "'curve.transformation' should be one of 'CW2pLM', 'CW2pLMi'")
})

test_that("general test", {
  testthat::skip_on_cran()

  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1))
  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                       sorter = "RUN"))
  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                       sorter = "SET"))

  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                       curve.transformation = "CW2pLM"))
  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                       curve.transformation = "CW2pLMi"))

  expect_warning(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                        curve.transformation = "CW2pHMi"),
                 "132 invalid values found, replaced by the mean")
  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                        curve.transformation = "CW2pPMi"))

  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1,
                                       dose_rate = 3))
  ## simulate v4
  CWOSL.SAR.Data@METADATA$VERSION <- 04
  expect_silent(plot_Risoe.BINfileData(CWOSL.SAR.Data, position = 1))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("TL",
                              plot_Risoe.BINfileData(CWOSL.SAR.Data,
                                                     pos = 1, run = 2, set = 5))
  vdiffr::expect_doppelganger("OSL",
                              plot_Risoe.BINfileData(CWOSL.SAR.Data,
                                                     pos = 2, run = 7, set = 3))
  vdiffr::expect_doppelganger("IRSL",
                              plot_Risoe.BINfileData(CWOSL.SAR.Data,
                                                     pos = 2, run = 8, set = 3))
  vdiffr::expect_doppelganger("dose rate",
                              plot_Risoe.BINfileData(CWOSL.SAR.Data,
                                                     pos = 2, run = 5, set = 3,
                                                     dose_rate = 0.123))
  })
})

#' @title Calculates the Thermal Lifetime using the Arrhenius equation
#'
#' @description
#' The function calculates the thermal lifetime of charges for given E (in eV), s (in 1/s) and
#' T (in deg. C.) parameters. The function can be used in two operational modes:
#'
#' @details
#'
#' **Mode 1 `(profiling = FALSE)`**
#'
#' An arbitrary set of input parameters (E, s, T) can be provided and the
#' function calculates the thermal lifetimes using the Arrhenius equation for
#' all possible combinations of these input parameters. An array with 3-dimensions
#' is returned that can be used for further analyses or graphical output (see example 1)
#'
#' **Mode 2 `(profiling = TRUE)`**
#'
#' This mode tries to profile the variation of the thermal lifetime for a chosen
#' temperature by accounting for the provided E and s parameters and their corresponding
#' standard errors, e.g., `E = c(1.600, 0.001)`
#' The calculation based on a Monte Carlo simulation, where values are sampled from a normal
#' distribution (for E and s).
#'
#' **Used equation (Arrhenius equation)**
#'
#' \deqn{\tau = \frac{1}{s} \exp(E/kT)}
#' where:
#' \eqn{\tau} in s as the mean time an electron spends in the trap for a given \eqn{T},
#' \eqn{E} trap depth in eV,
#' \eqn{s} the frequency factor in 1/s,
#' \eqn{T} the temperature in K and \eqn{k} the Boltzmann constant in eV/K (cf. Furetta, 2010).
#'
#' @param E [numeric] (**required**):
#' vector of trap depths in eV,
#' if `profiling = TRUE` only the first two elements are considered
#'
#' @param s [numeric] (**required**):
#' vector of frequency factor in 1/s,
#' if `profiling = TRUE` only the first two elements are considered
#'
#' @param T [numeric] (*with default*):
#' temperature in deg. C for which the lifetime(s) will be calculated.
#' A vector can be provided.
#'
#' @param output_unit [character] (*with default*):
#' output unit of the calculated lifetimes, accepted
#' entries are: `"Ma"`, `"ka"`, `"a"`, `"d"`, `"h"`, `"min"`, `"s"`
#'
#' @param profiling [logical] (*with default*):
#' this option allows to estimate uncertainties based on
#' given E and s parameters and their corresponding standard error
#' (cf. details and examples section)
#'
#' @param profiling_config [list] (*optional*):
#' allows to set configuration parameters used for the profiling
#' (and only have an effect here). Supported parameters are:
#'
#' - `n` (number of MC runs),
#' - `E.distribution` (distribution used for the re-sampling for E) and
#' - `s.distribution` (distribution used for the re-sampling for s).
#'
#' Currently only the normal distribution is supported
#' (e.g., `profiling_config = list(E.distribution = "norm")`
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output, currently only in combination with
#' `profiling = TRUE`.
#'
#' @param ... further arguments that can be passed in combination with the plot output.
#' Standard plot parameters are supported ([plot.default])
#'
#' @return
#' A [Luminescence::RLum.Results-class] object is returned a along with a plot (for
#' `profiling = TRUE`). The output object contain the following slots:
#'
#' **`@data`**
#'
#' \tabular{lll}{
#'  **Object** \tab **Type** \tab **Description** \cr
#'  `lifetimes` \tab [array] or [numeric] \tab calculated lifetimes \cr
#'  `profiling_matrix` \tab [matrix] \tab profiling matrix used for the MC runs
#' }
#'
#' **`@info`**
#'
#' \tabular{lll}{
#' **Object** \tab **Type** \tab **Description** \cr
#' `call` \tab `call` \tab the original function call
#' }
#'
#' @note
#' The profiling is currently based on re-sampling from a normal distribution, this
#' distribution assumption might be, however, not valid for given E and s parameters.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [graphics::matplot], [stats::rnorm][stats::Normal], [Luminescence::get_RLum]
#'
#' @references
#'
#' Furetta, C., 2010. Handbook of Thermoluminescence, Second Edition. World Scientific.
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##EXAMPLE 1
#' ##calculation for two trap-depths with similar frequency factor for different temperatures
#' E <- c(1.66, 1.70)
#' s <- 1e+13
#' T <- 10:20
#' temp <- calc_ThermalLifetime(
#'   E = E,
#'   s = s,
#'   T = T,
#'   output_unit = "Ma"
#' )
#' graphics::contour(x = E, y = T, z = temp$lifetimes[1,,],
#'         ylab = "Temperature [\u00B0C]",
#'         xlab = "Trap depth [eV]",
#'         main = "Thermal Lifetime Contour Plot"
#' )
#' mtext(side = 3, "(values quoted in Ma)")
#'
#' ##EXAMPLE 2
#' ##profiling of thermal life time for E and s and their standard error
#' E <- c(1.600, 0.003)
#' s <- c(1e+13,1e+011)
#' T <- 20
#' calc_ThermalLifetime(
#'   E = E,
#'   s = s,
#'   T = T,
#'   profiling = TRUE,
#'   output_unit = "Ma"
#')
#'
#' @export
calc_ThermalLifetime <- function(
  E,
  s,
  T = 20,
  output_unit = "Ma",
  profiling = FALSE,
  profiling_config = list(),
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_ThermalLifetime")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(E, "numeric")
  .validate_not_empty(E)
  .validate_class(s, "numeric")
  .validate_not_empty(s)
  .validate_class(T, c("numeric", "integer"))
  .validate_not_empty(T)
  .validate_logical_scalar(profiling)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

# Set variables -------------------------------------------------------------------------------

  ## Boltzmann constant (ev/K)
  kB <- .const$kB

  ## recalculate temperature in Kelvin (K)
  T.K <- T + .const$C2K

  ##SETTINGS FOR PROFILING
  ##profiling settings
  profiling_settings <- modifyList(list(
    n = 1000,
    E.distribution = "norm",
    s.distribution = "norm"
  ), profiling_config)

  ##check for odd input values
  if (profiling_settings$n < 1000){
    profiling_settings$n <- 1000
    .throw_warning("Minimum MC runs are 1000, parameter 'n' ",
                   "in profiling_config reset to 1000.")
  }

# Calculation ---------------------------------------------------------------------------------

 ##set function for the calculation
 f <- function(E, s, T.K) {
    1 / s * exp(E / (kB * T.K))
 }

 ##PROFILING
  if(profiling) {
    ##set profiling matrix
    profiling_matrix <-
      matrix(NA, ncol = 4, nrow = profiling_settings$n)

    ##fill matrix

    ##E
    profiling_matrix[, 1] <-
      if( profiling_settings$E.distribution == "norm"){
        rnorm(profiling_settings$n, mean = E[1], sd = E[2])

      }else{
        .throw_error("Unknown distribution setting for E profiling")
      }

    ##s
    profiling_matrix[, 2] <-
      if (profiling_settings$s.distribution == "norm") {
        rnorm(profiling_settings$n, mean = s[1], sd = s[2])

      } else{
        .throw_error("Unknown distribution setting for s profiling")
      }

    ##T
    profiling_matrix[, 3] <-
      rep(T.K[1], each = profiling_settings$n)


    ##calulate lifetimes
    profiling_matrix[, 4] <-
      f(profiling_matrix[, 1], profiling_matrix[, 2], profiling_matrix[, 3])

    ##reduce E and s vector on the first entry
    T <- T[1]

    ##set lifetimes
    lifetimes <- profiling_matrix[, 4]

  } else{

    ##set empty profiling matrix
    profiling_matrix <- matrix()

    ##calculate lifetimes
    lifetimes <- vapply(
      X = T.K,
      FUN = function(i) {
        vapply(
          X = E,
          FUN = function(j) {
            f(E = j, s = s, T.K = i)
          },
          FUN.VALUE = numeric(length(s)))
      },
      FUN.VALUE = matrix(numeric(), ncol = length(E), nrow = length(s))
    )

    ##transform to an array in either case to have the same output
    if (!inherits(lifetimes, "array")) {
      lifetimes <-
        array(lifetimes, dim = c(length(s), length(E), length(T)))
    }

    ##set dimnames to make reading more clear
    dimnames(lifetimes) <- list(s, E, paste("T =", T, "\u00B0C"))
  }

  ## re-calculate lifetimes according to the chosen output unit
  temp.lifetimes <- switch(
    output_unit,
    s = lifetimes,
    min = lifetimes / 60,
    h = lifetimes / 60 / 60,
    d = lifetimes / 60 / 60 / 24,
    a = lifetimes / .const$year_s,
    ka = lifetimes / (1e3 * .const$year_s),
    Ma = lifetimes / (1e6 * .const$year_s)
  )

  ##check for invalid values
  if(is.null(temp.lifetimes)){
    output_unit <- "s"
    .throw_warning("'output_unit' unknown, reset to 's'")
  }else{
    lifetimes <- temp.lifetimes
    rm(temp.lifetimes)
  }

  # Terminal output -----------------------------------------------------------------------------

  if(verbose){
    cat("\n[calc_ThermalLifetime()]\n\n")

    if(profiling){
    cat("\tprofiling = TRUE")
    cat("\n\t--------------------------\n")
    }
    cat("\tmean:\t", format(mean(lifetimes), scientific = TRUE), output_unit)
    cat("\n\tsd:\t", format(sd(lifetimes), scientific = TRUE), output_unit)

    cat("\n\tmin:\t", format(min(lifetimes), scientific = TRUE), output_unit)
    if(!profiling){
      cat(" (@", T[which(lifetimes == min(lifetimes), arr.ind = TRUE)[3]], "\u00B0C)")
    }

    cat("\n\tmax:\t", format(max(lifetimes), scientific = TRUE), output_unit)
    if(!profiling){
      cat(" (@", T[which(lifetimes == max(lifetimes), arr.ind = TRUE)[3]], "\u00B0C)")
    }

    cat("\n\t--------------------------")
    cat(paste0("\n\t(", length(lifetimes), " lifetimes calculated in total)\n"))
  }


  # Plotting ------------------------------------------------------------------------------------
  if(plot & profiling){

    ##plot settings
    plot.settings <- modifyList(list(
      main = "Thermal Lifetime Density Plot",
      xlab = paste0("Thermal lifetime [",output_unit,"]"),
      ylab = "Density",
      xlim = NULL,
      ylim = NULL,
      log = "",
      lwd = 1,
      lty = 1,
      col = rgb(0, 0, 0, 0.25)
    ), list(...))

    ##split data and calculate density
    ##set seq
    id_seq <- seq(
      from = 1,
      to = length(lifetimes),
      length.out = 200)

    ##calculate lifetime of the density
    lifetimes_density <-
      lapply(1:(length(id_seq) - 1),
             function(x) {
               density(lifetimes[id_seq[x]:id_seq[x+1]])
             })

    ##get x values
    lifetimes_density.x <- matrix(unlist(lapply(1:length(lifetimes_density), function(i){
      lifetimes_density[[i]]$x

    })), nrow = length(lifetimes_density[[1]]$x))

    ##get y values
    lifetimes_density.y <- matrix(unlist(lapply(1:length(lifetimes_density), function(i){
      lifetimes_density[[i]]$y

    })), nrow = length(lifetimes_density[[1]]$y))


    ##plot density curves
    graphics::matplot(
      lifetimes_density.x,
      lifetimes_density.y,
      type = "l",
      lwd = plot.settings$lwd,
      lty = plot.settings$lty,
      col = plot.settings$col,
      main = plot.settings$main,
      xlab = plot.settings$xlab,
      ylab = plot.settings$ylab,
      xlim = plot.settings$xlim,
      ylim = plot.settings$ylim,
      log = plot.settings$log
    )
  }

  # Return values -------------------------------------------------------------------------------
  set_RLum(
   class = "RLum.Results",
   data = list(lifetimes = lifetimes,
               profiling_matrix = profiling_matrix),
   info = list(call = sys.call())
  )
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_ThermalLifetime(),
               "'E' should be of class 'numeric'")
  expect_error(calc_ThermalLifetime(numeric()),
               "'E' cannot be an empty numeric")
  expect_error(calc_ThermalLifetime(E = 1.4),
               "'s' should be of class 'numeric'")
  expect_error(calc_ThermalLifetime(E = 1.4, s = numeric()),
               "'s' cannot be an empty numeric")
  expect_error(calc_ThermalLifetime(E = 1.4, s = 2, T = "error"),
               "'T' should be of class 'numeric' or 'integer'")
  expect_error(calc_ThermalLifetime(E = 1.4, s = 2, T = numeric()),
               "'T' cannot be an empty numeric")

  ##profiling settings
  SW({
  expect_warning(
    calc_ThermalLifetime(E = 1.4, s = 1e05, profiling_config = list(n = 10)),
    "Minimum MC runs are 1000, parameter 'n' in profiling_config reset to 1000")
  })
  expect_error(calc_ThermalLifetime(
    E = 1.4,
    s = 1e05,
    profiling = TRUE,
    profiling_config = list(E.distribution = "test")
  ), "Unknown distribution setting for E profiling")
  expect_error(suppressWarnings(calc_ThermalLifetime(
    E = 1.4,
    s = 1e05,
    profiling = TRUE,
    profiling_config = list(s.distribution = "test"))
  ), "Unknown distribution setting for s profiling")

  ##output
  SW({
  expect_warning(calc_ThermalLifetime(E = 1.4, s = 1e05, output_unit = "test"),
                 "'output_unit' unknown, reset to 's'")
  })
  expect_output(calc_ThermalLifetime(E = 1.4, s = 1e05, verbose = TRUE),
                "1 lifetimes calculated in total")
  expect_output(calc_ThermalLifetime(E = c(1.4, 0.001), s = c(1e05,1e03),
                                     profiling = FALSE),
                "4 lifetimes calculated in total")
  expect_output(calc_ThermalLifetime(E = c(1.4, 0.001), s = c(1e05,1e03),
                                     plot = TRUE, profiling = TRUE),
                "1000 lifetimes calculated in total")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.5e-6

  ## Example 1: calculation for two trap-depths with similar frequency factor
  ## for different temperatures
  expect_snapshot_RLum(calc_ThermalLifetime(E = c(1.66, 1.70),
                                            s = 1e+13,
                                            T = 10:20,
                                            output_unit = "Ma",
                                            verbose = FALSE),
                       tolerance = snapshot.tolerance)

  ## Example 2: profiling of thermal life time for E and s and their standard error
  expect_snapshot_RLum(calc_ThermalLifetime(E = c(1.600, 0.003),
                                            s = c(1e+13, 1e+11),
                                            T = 20,
                                            profiling = TRUE,
                                            output_unit = "Ma",
                                            verbose = FALSE,
                                            plot = FALSE),
                       tolerance = snapshot.tolerance)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("profiling",
                              calc_ThermalLifetime(E = c(1.600, 0.003),
                                            s = c(1e+13, 1e+11),
                                            T = 20,
                                            profiling = TRUE,
                                            output_unit = "Ma"))
  })
})

#'@title Import Risø BINX-data from the BINX Log File
#'
#'@description
#' The function attempts to extract meaningful measurement data from the
#' log file (`.txt`) stored along with a BINX file if this option was chosen.
#' This is particularly helpful if the BINX file is broken and the data would
#' be lost otherwise.
#'
#'@details
#' This function is basically a hack trying to make the most out of a broken file in case
#' at least the log file is salvageable. Please do not expect black magic; the extract is
#' very basic, and the function was written to work for TL, OSL, IRSL curves only.
#' If it breaks, it breaks.
#'
#' **TL curve extraction**
#' Since the BIN/BINX do not contain x-values, they also cannot be created magically
#' here. To avoid problems with the conversion, we assume always BIN-file version `3` for
#' the conversion. This should be okay in most cases, however, it may become problematic
#' if you have data for dedicated curve analysis.
#'
#' @param file [character] (**required**):
#' name of one or multiple ASCII log files to read; it can be a path to a
#' directory, in which case the function tries to read all files it finds
#' (there is no file extension check, so it may fail ungracefully on binary
#' files).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#'@param ... further arguments that will be passed to the function (currently not used)
#'
#'@seealso [Luminescence::read_BIN2R], [Luminescence::write_R2BIN], [Luminescence::Risoe.BINfileData-class],
#' [base::readLines], [Luminescence::RLum.Analysis-class], [list.files]
#'
#'@return
#' Returns an S4 [Luminescence::RLum.Analysis-class] object. Results are
#' returned as a list when multiple files are processed or `file` is a list.
#'
#'@note
#' This function tries to extract data from a log file produced for debugging
#' purposes. There is no guarantee regarding the format stability, and the
#' function is meant as a last resort if your BIN/BINX files are unusable.
#'
#' @section Function version: 0.1.1
#'
#'@author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@examples
#'## read log file (RLum.Analysis)
#'file <-  system.file("extdata/BINX_IRSL_LOG.TXT", package = "Luminescence")
#'object <-  read_BINXLOG2R(file = file)
#'
#'## convert to RisoeBINFile-class object
#'object <- convert_RLum2Risoe.BINfileData(object)
#'
#'## export as BINX
#'\dontrun{
#'  write_R2BIN(t, file = tempfile(), version = "04")
#'}
#'
#'@keywords IO
#'@md
#'@export
read_BINXLOG2R <- function(
  file,
  verbose = TRUE,
  ...
  ){
  .set_function_name("read_BINXLOG2R")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_logical_scalar(verbose)
  file <- .validate_file(file, verbose = verbose)

  ## Self-call --------------------------------------------------------------
  if (inherits(file, "list")) {
    out <- lapply(file, read_BINXLOG2R, verbose = verbose)
    return(unlist(out, recursive = FALSE))
  }

# Function core -----------------------------------------------------------

  ## open file connection and ensure that it properly closed
  con <- file(file, "rb")
  on.exit(close(con), add = TRUE)

  if (verbose) {
    cat("\n[read_BINXLOG2R()] Importing ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file[1])))
    cat("\n size: ", round(file.size(file[1]) / 1000, 2), "kB")
    cat("\n")
  }

  ## reminder
  ## All those suppressed warnings should avoid all kind of errors
  ## when we have different encodings, we then just try.

  ## read entire file
  lines <- readLines(con, skipNul = TRUE, warn = FALSE)

  ## check first 20 lines for keywords, this should reduce the number of failures
  if(sum(grepl("Sequence Name|Datafile Name|Error trapping is active", lines[1:20])) != 3)
     .throw_error("The file does not seem to be a valid debugging log-file")

  ## extra meta header
  FNAME <- trimws(
    regmatches(
      x = lines[1:5],
      regexpr(pattern = "(?<=Datafile Name\\:).+", lines[1:5], perl = TRUE)))
  SYSTEMID <- as.numeric(
    sub(".*: *([0-9]+).*", "\\1", lines[3:8][grepl("System ID", lines[3:8])]))

  ## remove all unwanted lines
  clean_lines <- suppressWarnings(
    lines[!grepl("Wait", lines, fixed = TRUE)])

  ## get all current operation lines
  co <- suppressWarnings(grep("Current operation",
       clean_lines, fixed = TRUE, value = TRUE, useBytes = TRUE))

      ## extract the DATE, TIME, RUN, SET, SAMPLE (POSITION)
      ## we extract the rest laster in the super loop
      pattern <- "^([0-9]{2}\\.[0-9]{2}\\.[0-9]{4})\\s+([0-9]{2}:[0-9]{2}:[0-9]{2}).*Set:\\s*(\\d+)\\s+Run:\\s*(\\d+)\\s+Sample:\\s*(\\d+)"
      header <- regmatches(co,  regexec(pattern, co, useBytes = TRUE))

        ## we extract only the position here, because this we will use
        ## to sort the dataset
        POSITION <- as.integer(vapply(header, \(x) x[6], character(1)))
        POSITION_ORDER <- order(POSITION)

        COMMENT <-  vapply(header, \(x) x[1], character(1))
        DATE <-  as.Date(vapply(header, \(x) x[2], character(1)), format = "%d.%m.%Y")
        TIME <-  vapply(header, \(x) x[3], character(1))
        SET <- as.integer(vapply(header, \(x) x[4], character(1)))
        RUN <- as.integer(vapply(header, \(x) x[5], character(1)))

  ## get start boundaries for blocks and build start and stop matrix of the blocks
  start_id <- suppressWarnings(
    grep(pattern = "Current operation:", x = clean_lines, fixed = TRUE, useBytes = TRUE))
  end_id <- c(start_id[-1] - 1, length(clean_lines))
  groups  <- rep(seq_along(start_id), times = end_id - start_id + 1)

  ## split the blocks
  data_blocks <- split(clean_lines[-(1:start_id[1] - 1)], groups)

  ## now iterate over the current operation blocks; we don't
  ## extract in one go an match later because we don't know
  ## what is stored and whether there was an error
  out <- lapply(sort(unique(POSITION)), \(i) {
    records <- lapply(which(i == POSITION), \(j) {
      ## check for any data in the data, if not return empty object
      if(suppressWarnings(!any(grepl(pattern = "Data", x = data_blocks[[j]], fixed = TRUE, useBytes = TRUE))))
        return(set_RLum("RLum.Data.Curve"))

      ## get record Type and extract depending on the TYPE
      ## preset (you never know)
      recordType <- list(
        LTYPE = NA_character_,
        HIGH = 0,
        LOW = 0,
        RATE = 0,
        NPOINTS = 0,
        AN_TEMP = 0)

      ## TL
      recordType <- if(suppressWarnings(any(grepl("TL", data_blocks[[j]], fixed = TRUE, useBytes = TRUE)))) {
        match <- suppressWarnings(strsplit(unlist(regmatches(
          x = data_blocks[[j]],
          m = regexec(
            pattern = "TL\\s[0-9]+\\s[0-9]+.[0-9]+\\s[0-9]+\\s[0-9]+",
            text = data_blocks[[j]], perl = TRUE, useBytes = TRUE))), split = " ", fixed = TRUE,
          useBytes = TRUE))

        ## get elements and combine those
        LTYPE <- unique(vapply(match, \(x) x[[1]], character(1)))
        HIGH <- max(as.numeric(vapply(match, \(x) x[[2]], character(1))))
        RATE <- mean(as.numeric(vapply(match, \(x) x[[3]], character(1))))
        AN_TEMP <- sum(as.numeric(vapply(match, \(x) x[[5]], character(1))))

        ## return
        list(LTYPE = LTYPE, HIGH = HIGH, RATE = RATE, NPOINTS = 0, AN_TEMP = AN_TEMP, LOW = 0)

      ## IRSL and OSL
      } else if (suppressWarnings(any(grepl("IR SET", data_blocks[[j]], fixed = TRUE, useBytes = TRUE)))){
        ## get relevant data
        LTYPE <- "IRSL"
        match <- suppressWarnings(unlist(regmatches(
          x = data_blocks[[j]],
          m = regexec(
            pattern = "(?<=PO\\sI).+",
            text = data_blocks[[j]],
            useBytes = TRUE,
            perl = TRUE))))

        ## if it is not IRSL, it is OSL
        if(length(match) == 0) {
          LTYPE <- "OSL"
          match <- suppressWarnings(unlist(regmatches(
            x = data_blocks[[j]],
            m = regexec(
              pattern = "(?<=OS\\sB).+",
              text = data_blocks[[j]],
              useBytes = TRUE,
              perl = TRUE))))
         }

        ## get HIGH and NPOINTS
        match <- scan(text = match[nzchar(match)], what = numeric(), quiet = TRUE)
        HIGH <- match[1]

        ## get rate and temperature (the first)
        match <- suppressMessages(unlist(regmatches(
          x = data_blocks[[j]],
          m = regexec(
            pattern = "(?<=ST\\s).+",
            text = data_blocks[[j]], perl = TRUE,
            useBytes = TRUE))))

        ## get AN_TEMP and RATE
        match <- scan(text = match, what = numeric(), quiet = TRUE)
        AN_TEMP <- match[1]
        RATE <- match[2]

        ## return
        list(
          LTYPE = LTYPE,
          HIGH = HIGH,
          RATE = RATE,
          NPOINTS = 0,
          AN_TEMP = AN_TEMP,
          LOW = 0)
      }

      ## determine how many data blocks we have
      n_data <- suppressWarnings(
        grep(pattern = "Data downloaded", x = data_blocks[[j]], fixed = TRUE, useBytes = TRUE))

      ## loop over data blocks
      data <- unlist(lapply(n_data, \(x) {
          ## get data blocks boundaries
          header <- data_blocks[[j]][x]
          n_points <- suppressWarnings(as.numeric(sub(".*\\s", "", header)))
          n_lines <- floor(n_points / 25)

          ## get block
          block <- data_blocks[[j]][seq.int(x + 1, length.out = n_lines)]

          ## get everything after two spaces
          values <- suppressWarnings(
            regmatches(block, regexpr("(?<=\\s\\s).*", block, perl = TRUE, useBytes = TRUE)))

          ## parse numbers
          scan(text = values, what = numeric(), quiet = TRUE)
      }))

      ## create data using internal xy constructor
      DATA <- matrix(NA, ncol = 2)
      if (!is.na(recordType[["LTYPE"]])) {
        DATA <- src_create_RLumDataCurve_matrix(
          DATA = data,
          VERSION = 3, ## we hard code this to V3 to prevent error messages
          NPOINTS = length(data), # this helps to prevent errors if NPOINTS is wrong
          LOW = recordType[["LOW"]],
          HIGH = recordType[["HIGH"]],
          LTYPE = recordType[["LTYPE"]],
          AN_TEMP = recordType[["AN_TEMP"]],
          TOLDELAY = 0L, TOLON = 0L, TOLOFF = 0L)
      }

      ## create curve object
      set_RLum(
        class = "RLum.Data.Curve",
        originator = "read_BINXLOG2R",
        recordType = paste(recordType[["LTYPE"]], "PMT"),
        data = DATA,
        info = list(
          DATE = DATE[j],
          TIME = TIME[j],
          POSITION = POSITION[j],
          RUN = RUN[j],
          SET = SET[j],
          FNAME = FNAME,
          SYSTEMID = SYSTEMID,
          COMMENT = substr(COMMENT[j], 1, 70))
        )
    })

    ## remove recordType NA
    records[sapply(records, \(x) is.na(x@recordType))] <- NULL

    ## create RLum.Analysis
    set_RLum("RLum.Analysis", records = records, originator = "read_BINXLOG2R")
  })

  ## return
  return(out)
}

#' Converting equivalent dose values from seconds (s) to Gray (Gy)
#'
#' Conversion of absorbed radiation dose in seconds (s) to the SI unit Gray
#' (Gy) including error propagation. Normally used for equivalent dose data.
#'
#' Calculation of De values from seconds (s) to Gray (Gy)
#'
#' \deqn{De [Gy] = De [s] * Dose Rate [Gy/s])}
#'
#' Provided calculation error propagation methods for error calculation
#' (with `'se'` as the standard error and `'DR'` of the dose rate of the beta-source):
#'
#' **(1) `omit`** (default)
#'
#' \deqn{se(De) [Gy] = se(De) [s] * DR [Gy/s]}
#'
#' In this case the standard error of the dose rate of the beta-source is
#' treated as systematic (i.e. non-random), it error propagation is omitted.
#' However, the error must be considered during calculation of the final age.
#' (cf. Aitken, 1985, pp. 242). This approach can be seen as method (2) (gaussian)
#' for the case the (random) standard error of the beta-source calibration is
#' 0. Which particular method is requested depends on the situation and cannot
#' be prescriptive.
#'
#' **(2) `gaussian`** error propagation
#'
#' \deqn{se(De) [Gy] = \sqrt((DR [Gy/s] * se(De) [s])^2 + (De [s] * se(DR) [Gy/s])^2)}
#'
#' Applicable under the assumption that errors of `De` and `se` are uncorrelated.
#'
#' **(3) `absolute`** error propagation
#'
#' \deqn{se(De) [Gy]= abs(DR [Gy/s] * se(De) [s]) + abs(De [s] * se(DR) [Gy/s])}
#'
#' Applicable under the assumption that errors of `De` and `se` are correlated.
#'
#'
#' @param data [data.frame] (**required**):
#' input values, structure: data (`values[,1]`) and data error (`values [,2]`)
#' are required
#'
#' @param dose.rate [RLum.Results-class], [data.frame] or [numeric] (**required**):
#' `RLum.Results` needs to be originated from the function [calc_SourceDoseRate],
#' for `vector` dose rate in Gy/s and dose rate error in Gy/s
#'
#' @param error.propagation [character] (*with default*):
#' error propagation method used for error calculation (`omit`, `gaussian` or
#' `absolute`), see details for further information
#'
#' @return
#' Returns a [data.frame] with converted values.
#'
#' @note
#' If no or a wrong error propagation method is given, the execution of the function is
#' stopped. Furthermore, if a `data.frame` is provided for the dose rate values is has to
#' be of the same length as the data frame provided with the argument `data`
#'
#' @section Function version: 0.6.0
#'
#' @author
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)\cr
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Margret C. Fuchs, HZDR, Helmholtz-Institute Freiberg for Resource Technology (Germany)
#'
#' @seealso [calc_SourceDoseRate]
#'
#' @references
#' Aitken, M.J., 1985. Thermoluminescence dating. Academic Press.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(A) for known source dose rate at date of measurement
#' ## - load De data from the example data help file
#' data(ExampleData.DeValues, envir = environment())
#' ## - convert De(s) to De(Gy)
#' Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))
#'
#'
#'
#'
#'
#' ##(B) for source dose rate calibration data
#' ## - calculate source dose rate first
#' dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
#'                                   calib.date = "2014-12-19",
#'                                   calib.dose.rate = 0.0438,
#'                                   calib.error = 0.0019)
#' # read example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' # apply dose.rate to convert De(s) to De(Gy)
#' Second2Gray(ExampleData.DeValues$BT998, dose.rate)
#'
#' @md
#' @export
Second2Gray <- function(
  data,
  dose.rate,
  error.propagation = "omit"
){

  # Integrity tests -----------------------------------------------------------------------------

  ##(1) data.frame or RLum.Data.Curve object?
  if(!is(data, "data.frame")){

    stop("[Second2Gray()] 'data' object has to be of type 'data.frame'!")

  }

  ##(2) numeric, data.frame or RLum.Data.Curve object?
  if(!is(dose.rate, "numeric")  &  !is(dose.rate, "RLum.Results") & !is(dose.rate, "data.frame")){

    stop("[Second2Gray()] 'dose.rate' object has to be of type 'numeric', 'data.frame' or 'RLum.Results'!")

  }


  ##(3) last check to avoid problems
  if(is(dose.rate, "data.frame")){

    if(nrow(dose.rate)!=nrow(data)){

      stop("[Second2Gray()] the data frames in 'data' and 'dose.rate' need to be of similar length!")

    }

  }


  ##(4) check for right orginator
  if(is(dose.rate, "RLum.Results")){

    if(dose.rate@originator != "calc_SourceDoseRate"){

      stop("[Second2Gray()]  Wrong originator for dose.rate 'RLum.Results' object.")

    }else{

      ##check what is what
      if(!is(get_RLum(dose.rate, data.object = "dose.rate"), "data.frame")){

        dose.rate <- data.frame(
          dose.rate  <- as.numeric(get_RLum(dose.rate, data.object = "dose.rate")[1]),
          dose.rate.error <- as.numeric(get_RLum(dose.rate, data.object = "dose.rate")[2])
          )

      }else{

        dose.rate <- get_RLum(dose.rate, data.object = "dose.rate")

      }

    }

  }


  # Calculation ---------------------------------------------------------------------------------


  De.seconds <- data[,1]
  De.error.seconds <- data[,2]

  De.gray <- NA
  De.error.gray <- NA

  if(is(dose.rate,"data.frame")){
    De.gray <- round(De.seconds*dose.rate[,1], digits=2)

  }else{
    De.gray <- round(De.seconds*dose.rate[1], digits=2)

  }

  if(error.propagation == "omit"){

    if(is(dose.rate,"data.frame")){
      De.error.gray <- round(dose.rate[,1]*De.error.seconds, digits=3)

    }else{
      De.error.gray <- round(dose.rate[1]*De.error.seconds, digits=3)

    }

  }else if(error.propagation == "gaussian"){

    if(is(dose.rate,"data.frame")){
       De.error.gray <- round(sqrt((De.seconds*dose.rate[,2])^2+(dose.rate[,1]*De.error.seconds)^2), digits=3)

    }else{
      De.error.gray <- round(sqrt((De.seconds*dose.rate[2])^2+(dose.rate[1]*De.error.seconds)^2), digits=3)

    }

  }else if (error.propagation == "absolute"){

    if(is(dose.rate,"data.frame")){
      De.error.gray <- round(abs(dose.rate[,1] * De.error.seconds) + abs(De.seconds * dose.rate[,2]), digits=3)

    }else{
      De.error.gray <- round(abs(dose.rate[1] * De.error.seconds) + abs(De.seconds * dose.rate[2]), digits=3)

    }

  }else{

    stop("[Second2Gray()] unsupported error propagation method!" )

  }

  # Return --------------------------------------------------------------------------------------

  data <- data.frame(De=De.gray, De.error=De.error.gray)


  return(data)

}

data(ExampleData.DeValues, envir = environment())
results <- Second2Gray(ExampleData.DeValues$BT998, c(0.2,0.01))
results_alt1 <- Second2Gray(ExampleData.DeValues$BT998, c(0.2,0.01), error.propagation = "gaussian")
results_alt2 <- Second2Gray(ExampleData.DeValues$BT998, c(0.2,0.01), error.propagation = "absolute")

test_that("check class and length of output", {
  testthat::skip_on_cran()
  local_edition(3)
  expect_s3_class(results, class = "data.frame")

})

test_that("check values from output example", {
  testthat::skip_on_cran()
  local_edition(3)

  expect_equal(sum(results[[1]]), 14754.09)
  expect_equal(sum(results[[2]]), 507.692)
  expect_equal(sum(results_alt1[[2]]), 895.911)
  expect_equal(sum(results_alt2[[2]]), 1245.398)

})

#' @title Function to create a Radial Plot
#'
#' @description A Galbraith's radial plot is produced on a logarithmic or a linear scale.
#'
#' @details Details and the theoretical background of the radial plot are given in the
#' cited literature. This function is based on an S script of Rex Galbraith. To
#' reduce the manual adjustments, the function has been rewritten. Thanks to
#' Rex Galbraith for useful comments on this function. \cr
#' Plotting can be disabled by adding the argument `plot = "FALSE"`, e.g.
#' to return only numeric plot output.
#'
#' Earlier versions of the Radial Plot in this package had the 2-sigma-bar
#' drawn onto the z-axis. However, this might have caused misunderstanding in
#' that the 2-sigma range may also refer to the z-scale, which it does not!
#' Rather it applies only to the x-y-coordinate system (standardised error vs.
#' precision). A spread in doses or ages must be drawn as lines originating at
#' zero precision (x0) and zero standardised estimate (y0). Such a range may be
#' drawn by adding lines to the radial plot (`line`, `line.col`,
#' `line.label`, cf. examples).
#'
#' A statistic summary, i.e. a collection of statistic measures of
#' centrality and dispersion (and further measures) can be added by specifying
#' one or more of the following keywords:
#' - `"n"` (number of samples),
#' - `"mean"` (mean De value),
#' - `"mean.weighted"` (error-weighted mean),
#' - `"median"` (median of the De values),
#' - `"median.weighted"` (error-weighted median),
#' - `"sdrel"` (relative standard deviation in percent),
#' - `"sdrel.weighted"` (error-weighted relative standard deviation in percent),
#' - `"sdabs"` (absolute standard deviation),
#' - `"sdabs.weighted"` (error-weighted absolute standard deviation),
#' - `"serel"` (relative standard error),
#' - `"serel.weighted"` (error-weighted relative standard error),
#' - `"seabs"` (absolute standard error),
#' - `"seabs.weighted"` (error-weighted absolute standard error),
#' - `"in.2s"` (percent of samples in 2-sigma range),
#' - `"kurtosis"` (kurtosis) and
#' - `"skewness"` (skewness).
#'
#' @param data [data.frame] or [Luminescence::RLum.Results-class] object (**required**):
#' for `data.frame`: either two columns: De (`data[,1]`) and De error
#' (`data[,2]`), or one: De (`values[,1]`). If a single-column data frame
#' is provided, De error is assumed to be 10^-9 for all measurements.
#' To plot several data sets in one plot, the data sets must be provided as
#' `list`, e.g. `list(data.1, data.2)`.
#'
#' @param na.rm [logical] (*with default*):
#' excludes `NA` values from the data set prior to any further operations.
#'
#' @param log.z [logical] (*with default*):
#' Option to display the z-axis in logarithmic scale. Default is `TRUE`.
#'
#' @param central.value [numeric] (*optional*):
#' User-defined central value, primarily used for horizontal centring
#' of the z-axis.
#'
#' @param centrality [character] or [numeric] (*with default*):
#' measure of centrality, used for automatically centring the plot and drawing
#' the central line. Can either be one out of
#' - `"mean"`,
#' - `"median"`,
#' - `"mean.weighted"` and
#' - `"median.weighted"` or a
#' - numeric value used for the standardisation.
#'
#' @param mtext [character] (*with default*):
#' additional text below the plot title.
#'
#' @param summary [character] (*with default*):
#' add statistic measures of centrality and dispersion to the plot.
#' Can be one or more of several keywords. See details for available keywords.
#'
#' @param summary.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`)
#' for the statistical summary. Alternatively, the keyword `"sub"` may be
#' specified to place the summary below the plot header. However, this latter
#' option is only possible if `mtext` is not used.
#'
#' @param legend [character] vector (*optional*):
#' legend content to be added to the plot.
#'
#' @param legend.pos [numeric] or [character] (*with default*):
#' optional legend position coordinates or keyword (e.g. `"topright"`).
#'
#' @param stats [character] (*optional*):
#' additional labels of statistically important values in the plot. It can be
#' one or more of `"min"`, `"max"` and `"median"`; any other values will be
#' ignored.
#'
#' @param rug [logical]:
#' Option to add a rug to the z-scale, to indicate the location of individual values
#'
#' @param plot.ratio [numeric]:
#' User-defined plot area ratio (i.e. curvature of the z-axis). If omitted,
#' the default value (`4.5/5.5`) is used and modified automatically to optimise
#' the z-axis curvature. The parameter should be decreased when data points
#' are plotted outside the z-axis or when the z-axis gets too elliptic.
#'
#' @param bar.col [character] or [numeric] (*with default*):
#' colour of the bar showing the 2-sigma range around the central
#' value. To disable the bar, use `"none"`. Default is `"grey80"`.
#'
#' @param y.ticks [logical]:
#' Option to hide y-axis labels. Useful for data with small scatter.
#'
#' @param grid.col [character] or [numeric] (*with default*):
#' colour of the grid lines (originating at `[0,0]` and stretching to
#' the z-scale). To disable grid lines, use `"none"`. Default is `"grey70"`.
#'
#' @param line [numeric]:
#' numeric values of the additional lines to be added.
#'
#' @param line.col [character] or [numeric]:
#' colour of the additional lines.
#'
#' @param line.label [character]:
#' labels for the additional lines.
#'
#' @param ... Further plot arguments to pass. `xlab` must be a vector of
#' length 2, specifying the upper and lower x-axes labels.
#'
#' @return Returns a plot object.
#'
#' @section Function version: 0.5.11
#'
#' @author
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Based on a rewritten S script of Rex Galbraith, 2010
#'
#' @seealso [plot], [Luminescence::plot_KDE], [Luminescence::plot_Histogram], [Luminescence::plot_AbanicoPlot]
#'
#' @references
#' Galbraith, R.F., 1988. Graphical Display of Estimates Having
#' Differing Standard Errors. Technometrics 30 (3), 271-281.
#'
#' Galbraith, R.F., 1990. The radial plot: Graphical assessment of spread in
#' ages. International Journal of Radiation Applications and Instrumentation.
#' Part D. Nuclear Tracks and Radiation Measurements 17 (3), 207-214.
#'
#' Galbraith, R. & Green, P., 1990. Estimating the component ages in a finite
#' mixture. International Journal of Radiation Applications and
#' Instrumentation. Part D. Nuclear Tracks and Radiation Measurements 17 (3)
#' 197-206.
#'
#' Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for mixed fission
#' track ages. Nuclear Tracks And Radiation Measurements 21 (4), 459-470.
#'
#' Galbraith, R.F., 1994. Some Applications of Radial Plots. Journal of the
#' American Statistical Association 89 (428), 1232-1242.
#'
#' Galbraith, R.F., 2010. On plotting OSL equivalent doses. Ancient TL 28 (1),
#' 1-10.
#'
#' Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent
#' dose and error calculation and display in OSL dating: An overview and some
#' recommendations. Quaternary Geochronology 11, 1-27.
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#' ExampleData.DeValues <- convert_Second2Gray(
#'   ExampleData.DeValues$BT998, c(0.0438,0.0019))
#'
#' ## plot the example data straightforward
#' plot_RadialPlot(data = ExampleData.DeValues)
#'
#' ## now with linear z-scale
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   log.z = FALSE)
#'
#' ## store the the plot parameters
#' plot1 <- plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   log.z = FALSE)
#' plot1
#' plot1$zlim
#'
#' ## now with adjusted z-scale limits
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   log.z = FALSE,
#'   xlim = c(0, 5),
#'   zlim = c(100, 200))
#'
#' ## now the two plots with serious but seasonally changing fun
#' #plot_RadialPlot(data = data.3, fun = TRUE)
#'
#' ## now with user-defined central value, in log-scale again
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   central.value = 150)
#'
#' ## now with a rug, indicating individual De values at the z-scale
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   rug = TRUE)
#'
#' ## now with legend, colour, different points and smaller scale
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   legend = "Sample 1",
#'   col = "tomato4",
#'   bar.col = "peachpuff",
#'   pch = "R",
#'   cex = 0.8)
#'
#' ## now without 2-sigma bar, y-axis, grid lines and central value line
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   bar.col = "none",
#'   grid.col = "none",
#'   y.ticks = FALSE,
#'   lwd = 0)
#'
#' ## now with user-defined axes labels
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   xlab = c("Data error (%)", "Data precision"),
#'   ylab = "Scatter",
#'   zlab = "Equivalent dose [Gy]")
#'
#' ## now with minimum, maximum and median value indicated
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   central.value = 150,
#'   stats = c("min", "max", "median"))
#'
#' ## now with a brief statistical summary
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   summary = c("n", "in.2s"))
#'
#' ## now with another statistical summary as subheader
#' plot_RadialPlot(
#'   data = ExampleData.DeValues,
#'   summary = c("mean.weighted", "median"),
#'   summary.pos = "sub")
#'
#' ## now the data set is split into sub-groups, one is manipulated
#' data.1 <- ExampleData.DeValues[1:15,]
#' data.2 <- ExampleData.DeValues[16:25,] * 1.3
#'
#' ## now a common dataset is created from the two subgroups
#' data.3 <- list(data.1, data.2)
#'
#' ## now the two data sets are plotted in one plot
#' plot_RadialPlot(data = data.3)
#'
#' ## now with some graphical modification
#' plot_RadialPlot(
#'   data = data.3,
#'   col = c("darkblue", "darkgreen"),
#'   bar.col = c("lightblue", "lightgreen"),
#'   pch = c(2, 6),
#'   summary = c("n", "in.2s"),
#'   summary.pos = "sub",
#'   legend = c("Sample 1", "Sample 2"))
#'
#' @export
plot_RadialPlot <- function(
  data,
  na.rm = TRUE,
  log.z = TRUE,
  central.value = NULL,
  centrality = c("mean.weighted", "mean.weighted", "median", "median.weighted"),
  mtext = "",
  summary = c("n", "in.2s"),
  summary.pos = "sub",
  legend = NULL,
  legend.pos = "topright",
  stats = "none",
  rug = FALSE,
  plot.ratio = NULL,
  bar.col = NULL,
  y.ticks = TRUE,
  grid.col = NULL,
  line = NULL,
  line.col = NULL,
  line.label = NULL,
  ...
) {
  .set_function_name("plot_RadialPlot")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_not_empty(data)
  .validate_logical_scalar(log.z)
  .validate_class(centrality, c("character", "numeric"))
  if (is.character(centrality)) {
    centrality <- .validate_args(centrality, c("mean", "mean.weighted",
                                               "median", "median.weighted"))
  }

  ## Homogenise input data format
  if (!inherits(data, "list"))
      data <- list(data)

  ## Check input data
  for(i in 1:length(data)) {
    .validate_class(data[[i]], c("data.frame", "RLum.Results"),
                    name = "All elements of 'data'")

    if (inherits(data[[i]], "RLum.Results")) {
        data[[i]] <- get_RLum(data[[i]], "data")
    }

    ## ensure that the dataset it not degenerate
    .validate_not_empty(data[[i]],
                        name = paste0("Input 'data[[", i, "]]'"))

      ## if `data[[i]]` is a single-column data frame, append a second
      ## column with a small non-zero value (10^-9 for consistency with
      ## what `calc_Statistics() does)
      if (ncol(data[[i]]) < 2) {
        data[[i]] <- data.frame(data[[i]], 10^-9)
      } else {
        if (ncol(data[[i]]) > 2) {
          ## keep only the first two columns
          data[[i]] <- data[[i]][, 1:2]
        }

        ## if all errors are NA, we set them to 0 so we correct them in the
        ## next block
        if (all(is.na(data[[i]][, 2]))) {
          data[[i]][, 2] <- 0
        }

        ## don't let the error be NA or zero: we set it to the smallest between
        ## the smallest non-zero error and 10^-9
        is.zero <- data[[i]][, 2] == 0
        if (any(is.zero)) {
          min.value <- min(data[[i]][!is.zero, 2], 10^-9)
          data[[i]][is.zero, 2] <- min.value
          .throw_warning("Error values cannot be zero or NA, reset to ", min.value)
        }
      }
  }

  valid.pos <- c("left", "center", "right", "topleft", "top", "topright",
                 "bottomleft", "bottom", "bottomright")
  .validate_class(summary, "character")
  if (is.numeric(summary.pos)) {
    .validate_length(summary.pos, 2)
  }
  else {
    summary.pos <- .validate_args(summary.pos, c("sub", valid.pos))
  }
  .validate_class(legend, "character", null.ok = TRUE)
  if (is.numeric(legend.pos)) {
    .validate_length(legend.pos, 2)
  } else {
    legend.pos <- .validate_args(legend.pos, valid.pos)
  }

  .validate_class(stats, "character")
  .validate_logical_scalar(rug)
  .validate_positive_scalar(plot.ratio, null.ok = TRUE)
  .validate_logical_scalar(y.ticks)
  .validate_class(line, c("numeric", "integer"), null.ok = TRUE)

  if (is.null(bar.col)) {
    bar.col <- rep("grey80", length(data))
  }
  if (is.null(grid.col)) {
    grid.col <- rep("grey70", length(data))
  }

  ## check z-axis log-option for grouped data sets
  if (inherits(data, "list") && length(data) > 1 && !log.z) {
    .throw_warning("'log.z' is set to 'FALSE' altough ",
                   "more than one data set (group) is provided")
  }

  ## optionally, remove NA-values
  .validate_logical_scalar(na.rm)
  if (na.rm) {
    for(i in 1:length(data)) {
      data[[i]] <- na.exclude(data[[i]])
      if (nrow(data[[i]]) == 0)
        .throw_error("After NA removal, nothing is left from data set ", i)
    }
  }

  ## create preliminary global data set
  De.global <- unlist(lapply(data, function(x) x[, 1]))
  if (length(De.global) < 2) {
    .throw_error("At least two data points are required")
  }

  ## calculate major preliminary tick values and tick difference
  extraArgs <- list(...)

  ## calculate correction dose to shift non-positive values
  De.add <- 0
  if (log.z && min(De.global) <= 0) {
    if("zlim" %in% names(extraArgs)) {
      De.add <- abs(extraArgs$zlim[1])
    } else {
      ## exclude zeros, as they cause infinities when logged
      De.global.not0 <- De.global[De.global != 0]

      ## estimate delta De to add to all data
      De.add <- min(10^ceiling(log10(abs(De.global.not0))) * 10)

      ## optionally readjust delta De for extreme values
      if (De.add <= abs(min(De.global.not0))) {
        De.add <- De.add * 10
      }
    }

    ## add correction dose to data set and adjust error
    for(i in 1:length(data))
      data[[i]][,1] <- data[[i]][,1] + De.add

    De.global <- De.global + De.add
  }

  ## calculate and append statistical measures --------------------------------

  ## z-values and se based on log-option
  data <- lapply(data, function(x, De.add) {
    cbind(x,
          z = if (log.z) log(x[, 1]) else x[, 1],
          se = if (log.z) x[, 2] / (x[, 1] + De.add) else x[, 2])
  }, De.add = De.add)

  ## calculate central values
  data <- lapply(data, function(x) {
    cbind(x,
          z.central = switch(
            as.character(centrality[1]),
            mean = rep(mean(x[, 3], na.rm = TRUE), nrow(x)),
            median = rep(median(x[, 3], na.rm = TRUE), nrow(x)),
            mean.weighted = rep(stats::weighted.mean(x[, 3], w = 1 / x[, 4]^2), nrow(x)),
            median.weighted = rep(.weighted.median(x[, 3], w = 1 / x[, 4]^2), nrow(x)),
            if (is.numeric(centrality) && length(centrality) >= length(data)) {
             rep(median(x[, 3], na.rm = TRUE), nrow(x))
            } else NA)
          )
  })

  if (is.numeric(centrality) && length(centrality) == length(data)) {
    ## compute z.central, as this could not be done in the lapply before
    z.central.raw <- if (log.z) log(centrality + De.add) else centrality + De.add
    lapply(1:length(data), function(x) data[[x]][, 5] <<- rep(z.central.raw[x], nrow(data[[x]])))
  }

  ## calculate precision and standard estimate
  idx <- 0
  data <- lapply(data, function(x) {
    idx <<- idx + 1
    colnames(x) <- c("De", "error", "z", "se", "z.central")
    cbind(x,
          precision = 1 / x[, 4],
          std.estimate = (x[, 3] - x[, 5]) / x[, 4],
          std.estimate.plot = NA, # will be filled in further down
          .id = idx)
  })

  ## generate global data set
  data.global <- if (length(data) > 1) as.data.frame(rbindlist(data)) else data[[1]]

  ## calculate global central value
  z.central.global <- switch(as.character(centrality[1]),
                             mean = mean(data.global[, 3], na.rm = TRUE),
                             median = median(data.global[, 3], na.rm = TRUE),
                             mean.weighted = stats::weighted.mean(data.global[, 3], w = 1 / data.global[, 4]^2),
                             median.weighted = .weighted.median(data.global[, 3],
                                                                w = 1 / data.global[, 4]^2),
                             if (is.numeric(centrality) && length(centrality) >= length(data)) {
                               mean(data.global[, 3], na.rm = TRUE)
                             } else NA)

  ## optionally adjust central value by user-defined value
  if (!is.null(central.value)) {
    # ## adjust central value for De.add
    central.value <- central.value + De.add
    if (log.z)
      .validate_positive_scalar(central.value)
    z.central.global <- ifelse(log.z,
                               log(central.value),
                               central.value)
  }

  ## re-calculate standardised estimate for plotting
  for(i in 1:length(data)) {
    data[[i]][,8] <- (data[[i]][,3] - z.central.global) / data[[i]][,4]
  }
  data.global$std.estimate.plot <- unlist(lapply(data, function(x) x[, 8]))

  ## print warning for too small scatter
  if (max(abs(1 / data.global[, 6])) < 0.02) {
    .throw_message("Small standardised estimate scatter, toggle off y.ticks?",
                   error = FALSE)
  }

  ## read out additional arguments---------------------------------------------

  main <- extraArgs$main %||% expression(paste(D[e], " distribution"))
  sub <- extraArgs$sub %||% ""

  if("xlab" %in% names(extraArgs)) {
    xlab <- extraArgs$xlab
    .validate_class(xlab, "character", length = 2)
  } else {
    xlab <- c(ifelse(log.z, "Relative standard error [%]", "Standard error"),
              "Precision")
  }

  ylab <- extraArgs$ylab %||% "Standardised estimate"
  zlab <- extraArgs$zlab %||% expression(paste(D[e], " [Gy]"))

  limits.z <- extraArgs$zlim %||% {
    z.span <- (mean(data.global[,1]) * 0.5) / (sd(data.global[,1]) * 100)
    z.span <- min(z.span, 0.89)
    c((0.9 - z.span) * min(data.global[[1]]),
      (1.1 + z.span) * max(data.global[[1]]))
  }
  .validate_class(limits.z, "numeric", length = 2, name = "'zlim'")
  limits.z <- sort(limits.z)
  if (log.z && limits.z[1] <= 0) {
    limits.z[1] <- 0.01
  }

  limits.x <- extraArgs$xlim %||% c(0, max(data.global[,6]))
  if(limits.x[1] != 0) {
    limits.x[1] <- 0
    .throw_warning("Lower x-axis limit not set to zero, corrected")
  }

  limits.y <- extraArgs$ylim %||% {
    y.span <- (mean(data.global[,1]) * 10) / (sd(data.global[,1]) * 100)
    y.span <- ifelse(y.span > 1, 0.98, y.span)
    c(-(1 + y.span) * max(abs(data.global[, 7])),
      (0.8 + y.span) * max(abs(data.global[, 7])))
  }

  cex <- extraArgs$cex %||% 1
  lty <- extraArgs$lty %||% rep(2, length(data))
  lwd <- extraArgs$lwd %||% rep(1, length(data))
  pch <- extraArgs$pch %||% rep(1, length(data))
  col <- extraArgs$col %||% 1:length(data)
  .validate_positive_scalar(cex)
  .validate_length(lty, length(data))
  .validate_length(lwd, length(data))
  .validate_length(pch, length(data))
  .validate_length(col, length(data))

  tcl <- extraArgs$tcl %||% -0.5
  show <- extraArgs$show %||% TRUE
  fun <- isTRUE(extraArgs$fun)

  ## define auxiliary plot parameters -----------------------------------------

  ## optionally adjust plot ratio
  if (is.null(plot.ratio)) {
    if(log.z) {
      plot.ratio <- 1 /  (1 * ((max(data.global[,6]) - min(data.global[,6])) /
        (max(data.global[,7]) - min(data.global[,7]))))
    } else {
      plot.ratio <- 4.5 / 5.5
    }
  }

  ##limit plot ratio
  plot.ratio <- min(c(1e+06, plot.ratio))

  ## calculate conversion factor for plot coordinates
  f <- (max(data.global[,6]) - min(data.global[,6])) /
       (max(data.global[,7]) - min(data.global[,7])) * plot.ratio

  ## calculate major and minor z-tick values
  tick.values.major <- signif(pretty(limits.z, n = 5))
  tick.values.minor <- signif(pretty(limits.z, n = 25), 3)
  tick.values.major <- tick.values.major[between(tick.values.major,
                                                 min(tick.values.minor),
                                                 max(tick.values.minor))]
  tick.values.minor <- tick.values.minor[between(tick.values.minor,
                                                 limits.z[1], limits.z[2])]

  ## add ticks corresponding to the extremes of the z-axis but only if they
  ## fall outside of the major ticks
  tick.add <- signif(limits.z[!between(limits.z,
                                       min(tick.values.major),
                                       max(tick.values.major))])
  tick.values.major <- c(tick.add, tick.values.major)

  user.limits <- limits.z

  if (log.z) {
    tick.values.major[tick.values.major == 0] <- 1
    tick.values.minor[tick.values.minor == 0] <- 1
    tick.values.major <- log(tick.values.major)
    tick.values.minor <- log(tick.values.minor)
    user.limits <- log(user.limits)
  }

  ## calculate z-axis radius
  r.x <- limits.x[2] / max(data.global[,6]) + 0.05
  r <- max(sqrt((data.global[,6])^2+(data.global[,7] * f)^2)) * r.x

  ## compute the coordinates for points on the z-axis
  x.coord <- function(x) r / sqrt(1 + f^2 * (x - z.central.global)^2)
  y.coord <- function(x, x.coord) (x - z.central.global) * x.coord

  ## calculate major z-tick coordinates
  tick.x1.major <- x.coord(tick.values.major)
  tick.y1.major <- y.coord(tick.values.major, tick.x1.major)
  tick.x2.major <- (1 + 0.015 * cex) * tick.x1.major
  tick.y2.major <- y.coord(tick.values.major, tick.x2.major)
  ticks.major <- cbind(tick.x1.major, tick.x2.major,
                       tick.y1.major, tick.y2.major)

  ## calculate minor z-tick coordinates
  tick.x1.minor <- x.coord(tick.values.minor)
  tick.y1.minor <- y.coord(tick.values.minor, tick.x1.minor)
  tick.x2.minor <- (1 + 0.007 * cex) * tick.x1.minor
  tick.y2.minor <- y.coord(tick.values.minor, tick.x2.minor)
  ticks.minor <- cbind(tick.x1.minor,
                       tick.x2.minor,
                       tick.y1.minor,
                       tick.y2.minor)

  ## calculate z-label positions
  label.x <- 1.03 * x.coord(tick.values.major)
  label.y <- y.coord(tick.values.major, label.x)

  ## create z-axis labels
  label.z.text <- if (log.z)
                    signif(exp(tick.values.major), 3)
                  else
                    signif(tick.values.major, 3)

  ## to avoid overprinting we remove the z-axis labels at the extremes if
  ## they are too close to a major tick label (#1013)
  rm.idx <- NULL
  for (idx in seq_along(tick.add)) {
    dist <- sqrt((label.x[-idx] - label.x[idx])^2 +
                 (label.y[-idx] - label.y[idx])^2)
    if (any(dist < 1))
      rm.idx <- c(rm.idx, idx)
  }
  if (!is.null(rm.idx)) {
    label.x <- label.x[-rm.idx]
    label.y <- label.y[-rm.idx]
    label.z.text <- label.z.text[-rm.idx]
  }

  ## subtract De.add from label values
  label.z.text <- label.z.text - De.add

  labels <- cbind(label.x, label.y, label.z.text)

  ## calculate coordinates for 2-sigma-polygon overlay
  polygons <- matrix(nrow = length(data), ncol = 8)

  for(i in 1:length(data)) {
    polygons[i,1:4] <- c(limits.x[1],
                         limits.x[1],
                         max(data.global[,6]),
                         max(data.global[,6]))
    polygons[i,5:8] <- c(-2,
                         2,
                         (data[[i]][1,5] - z.central.global) *
                           polygons[i,3] + 2,
                         (data[[i]][1,5] - z.central.global) *
                           polygons[i,4] - 2)
  }

  ## calculate node coordinates for semi-circle
  num.values <- 500
  ellipse.values <- seq(
    from = min(c(tick.values.major, tick.values.minor, user.limits[1])),
    to = max(c(tick.values.major,tick.values.minor, user.limits[2])),
    length.out = num.values)
  ellipse.x <- x.coord(ellipse.values)
  ellipse.y <- y.coord(ellipse.values, ellipse.x)
  ellipse <- cbind(ellipse.x, ellipse.y)
  ellipse.lims <- rbind(ellipse[c(1, num.values), 1],
                        ellipse[c(1, num.values), 2])

  ## check if z-axis overlaps with 2s-polygon
  polygon_y_max <- max(polygons[,7])
  polygon_y_min <- min(polygons[,7])

  z_2s_upper <- ellipse.x[abs(ellipse.y - polygon_y_max) ==
                            min(abs(ellipse.y - polygon_y_max))]

  z_2s_lower <- ellipse.x[abs(ellipse.y - polygon_y_min) ==
                            min(abs(ellipse.y - polygon_y_min))]

  if ((max(polygons[, 3]) >= z_2s_upper || max(polygons[, 3]) >= z_2s_lower) &&
      bar.col[1] != "none") {
    .throw_warning("z-scale touches 2s-polygon, decrease plot ratio")
  }

  ## calculate statistical labels
  stats.data <- matrix(nrow = 3, ncol = 3)
  data.stats <- as.numeric(data.global[,1])

  if ("min" %in% stats) {
    stats.data[1, 3] <- data.stats[data.stats == min(data.stats)][1]
    stats.data[1, 1] <- data.global[data.stats == stats.data[1, 3], 6][1]
    stats.data[1, 2] <- data.global[data.stats == stats.data[1, 3], 8][1]
  }

  if ("max" %in% stats) {
    stats.data[2, 3] <- data.stats[data.stats == max(data.stats)][1]
    stats.data[2, 1] <- data.global[data.stats == stats.data[2, 3], 6][1]
    stats.data[2, 2] <- data.global[data.stats == stats.data[2, 3], 8][1]
  }

  if ("median" %in% stats) {
    stats.data[3, 3] <- data.stats[data.stats == quantile(data.stats, 0.5, type = 3)]
    stats.data[3, 1] <- data.global[data.stats == stats.data[3, 3], 6][1]
    stats.data[3, 2] <- data.global[data.stats == stats.data[3, 3], 8][1]
  }

  ## recalculate axes limits if necessary
  limits.z.x <- range(ellipse[,1])
  limits.z.y <- range(ellipse[,2])
  if(!("ylim" %in% names(extraArgs))) {
    if(limits.z.y[1] < 0.66 * limits.y[1]) {
      limits.y[1] <- 1.8 * limits.z.y[1]
    }
    if(limits.z.y[2] > 0.77 * limits.y[2]) {
      limits.y[2] <- 1.3 * limits.z.y[2]
    }
  }
  if (!("xlim" %in% names(extraArgs)) && limits.z.x[2] > 1.1 * limits.x[2]) {
      limits.x[2] <- limits.z.x[2]
  }

  ## calculate and paste statistical summary
  De.stats <- matrix(nrow = length(data), ncol = 18)
  colnames(De.stats) <- c("n", "mean", "mean.weighted", "median", "median.weighted",
    "kde.max", "sd.abs", "sd.rel", "se.abs", "se.rel", "q25", "q75", "skewness",
    "kurtosis", "sd.abs.weighted", "sd.rel.weighted", "se.abs.weighted",
    "se.rel.weighted")

  for(i in 1:length(data)) {
    data_to_stats <- data[[i]][,1:2]

    ## remove added De
    if(log.z) data_to_stats$De <- data_to_stats$De - De.add

    statistics <- calc_Statistics(data = data_to_stats)
    De.stats[i,1] <- statistics$weighted$n
    De.stats[i,2] <- statistics$unweighted$mean
    De.stats[i,3] <- statistics$weighted$mean
    De.stats[i,4] <- statistics$unweighted$median
    De.stats[i,5] <- statistics$weighted$median
    De.stats[i,7] <- statistics$unweighted$sd.abs
    De.stats[i,8] <- statistics$unweighted$sd.rel
    De.stats[i,9] <- statistics$unweighted$se.abs
    De.stats[i,10] <- statistics$weighted$se.rel
    De.stats[i,11] <- quantile(data[[i]][,1], 0.25)
    De.stats[i,12] <- quantile(data[[i]][,1], 0.75)
    De.stats[i,13] <- statistics$unweighted$skewness
    De.stats[i,14] <- statistics$unweighted$kurtosis
    De.stats[i,15] <- statistics$weighted$sd.abs
    De.stats[i,16] <- statistics$weighted$sd.rel
    De.stats[i,17] <- statistics$weighted$se.abs
    De.stats[i,18] <- statistics$weighted$se.rel

    ## kdemax - here a little doubled as it appears below again
    De.density <- try(density(x = data[[i]][,1],
                              kernel = "gaussian",
                              from = limits.z[1],
                              to = limits.z[2]),
                      silent = TRUE)

    De.stats[i,6] <- NA
    if(!inherits(De.density, "try-error")) {
      De.stats[i,6] <- De.density$x[which.max(De.density$y)]
    }
  }

  ## helper to generate an element of the statistical summary
  .summary_line <- function(keyword, summary, val, label = keyword,
                            percent = FALSE, sep = FALSE, digits = 2) {
    ifelse(keyword %in% summary,
           paste0(label, " = ", round(val, digits),
                  if (percent) " %" else NULL, if (sep) " | " else "\n"),
           "")
  }

  ## initialize list with a dummy element, it will be removed afterwards
  label.text <- list(NA)

  is.sub <- summary.pos[1] == "sub"
  stops <- NULL
  for (i in 1:length(data)) {
    if (!is.sub)
      stops <- strrep("\n", (i - 1) * length(summary))

    summary.text <- character(0)
    for (j in 1:length(summary)) {
      summary.text <-
        c(summary.text,
          .summary_line("n", summary[j], De.stats[i, 1], sep = is.sub),
          .summary_line("mean", summary[j], De.stats[i, 2], sep = is.sub),
          .summary_line("mean.weighted", summary[j], De.stats[i, 3], sep = is.sub,
                        label = "weighted mean"),
          .summary_line("median", summary[j], De.stats[i, 4], sep = is.sub),
          .summary_line("median.weighted", summary[j], De.stats[i, 5], sep = is.sub,
                        label = "weighted median"),
          .summary_line("kdemax", summary[j], De.stats[i, 6], sep = is.sub),
          .summary_line("sdabs", summary[j], De.stats[i, 7], sep = is.sub,
                        label = "sd"),
          .summary_line("sdrel", summary[j], De.stats[i, 8], sep = is.sub,
                        label = "rel. sd", percent = TRUE),
          .summary_line("seabs", summary[j], De.stats[i, 9], sep = is.sub,
                        label = "se"),
          .summary_line("serel", summary[j], De.stats[i, 10], sep = is.sub,
                        label = "rel. se", percent = TRUE),
          .summary_line("skewness", summary[j], De.stats[i, 13], sep = is.sub),
          .summary_line("kurtosis", summary[j], De.stats[i, 14], sep = is.sub),
          .summary_line("in.2s", summary[j],
                        sum(data[[i]][,7] > -2 & data[[i]][,7] < 2) /
                        nrow(data[[i]]) * 100, sep = is.sub,
                        label = "in 2 sigma", percent = TRUE, digits = 1),
          .summary_line("sdabs.weighted", summary[j], De.stats[i, 15], sep = is.sub,
                        label = "abs. weighted sd"),
          .summary_line("sdrel.weighted", summary[j], De.stats[i, 16], sep = is.sub,
                        label = "rel. weighted sd"),
          .summary_line("seabs.weighted", summary[j], De.stats[i, 17], sep = is.sub,
                        label = "abs. weighted se"),
          .summary_line("serel.weighted", summary[j], De.stats[i, 18], sep = is.sub,
                        label = "rel. weighted se"))
    }
    label.text[[length(label.text) + 1]] <- paste0(
        if (is.sub ) "" else stops,
        paste(summary.text, collapse = ""),
        stops)
  }

  ## remove dummy list element
  label.text[[1]] <- NULL

  ## remove outer vertical lines from string
  if (is.sub) {
    for (i in seq_along(label.text)) {
      label.text[[i]] <- substr(x = label.text[[i]],
                                start = 1,
                                stop = nchar(label.text[[i]]) - 3)
    }
  }

  ## convert keywords into summary and legend placement coordinates
  coords <- .get_keyword_coordinates(summary.pos, limits.x, limits.y)
  summary.pos <- coords$pos
  summary.adj <- coords$adj

  ## calculate line coordinates and further parameters
  if (!is.null(line)) {
    #line = line + De.add

    if (log.z) {
      if (any(line < 0)) {
        line <- line[line >= 0]
        .throw_message("Lines with negative value skipped due to 'log.z = TRUE'")
      }
      line <- log(line)
    }

    line.coords <- NULL
    for (i in seq_along(line)) {
      line.x <- c(limits.x[1], x.coord(line[i]))
      line.y <- c(0, y.coord(line[i], line.x[2]))
      line.coords[[i]] <- rbind(line.x, line.y)
    }

    if (is.null(line.col)) {
      line.col <- seq_along(line.coords)
    }

    if (is.null(line.label)) {
      line.label <- rep("", length(line.coords))
    }
  }

  ## Generate plot ------------------------------------------------------------

  ## check if plotting is enabled
  if(show) {

    ## determine number of subheader lines to shift the plot
    if(length(summary) > 0 & summary.pos[1] == "sub") {
      shift.lines <- length(data) + 1
    } else {shift.lines <- 1}

    ## setup plot area
    default <- par(mar = c(4, 4, shift.lines + 1.5, 7),
        xpd = TRUE,
        cex = cex)

    ## reset on exit
    on.exit(par(default), add = TRUE)

    ## create empty plot
    plot(NA,
         xlim = limits.x,
         ylim = limits.y,
         main = "",
         sub = sub,
         xlab = "",
         ylab = "",
         xaxs = "i",
         yaxs = "i",
         frame.plot = FALSE,
         axes = FALSE)

    ## add y-axis label
    mtext(side = 2,
          line = 2.5,
          at = 0,
          adj = 0.5,
          cex = cex,
          text = ylab)

    ## calculate upper x-axis label values
    label.x.upper <- if (log.z) {
      as.character(round(1/axTicks(side = 1)[-1] * 100, 1))
    } else {
      as.character(round(1/axTicks(side = 1)[-1], 1))
    }

    ## optionally, plot 2-sigma-bar
    if(bar.col[1] != "none") {
      for(i in 1:length(data)) {
        polygon(x = polygons[i,1:4],
                y = polygons[i,5:8],
                xpd = FALSE,
                lty = "blank",
                col = bar.col[i])
      }
    }

    ## optionally, add grid lines
    if(grid.col[1] != "none") {
      for(i in 1:length(tick.x1.major)) {
        lines(x = c(limits.x[1], tick.x1.major[i]),
              y = c(0, tick.y1.major[i]),
              col = grid.col)
      }
    }

    ## optionally, plot central value lines
    if(lwd[1] > 0 & lty[1] > 0) {
      for(i in 1:length(data)) {
        x2 <- x.coord(data[[i]][1, 5])
        y2 <- y.coord(data[[i]][1, 5], x2)
        lines(x = c(limits.x[1], x2),
              y = c(0, y2),
              lty = lty[i],
              lwd = lwd[i],
              col = col[i])
      }
    }

    ## optionally add further lines
    for (i in seq_along(line)) {
        lines(x = line.coords[[i]][1,],
              y = line.coords[[i]][2,],
              col = line.col[i])
        text(x = line.coords[[i]][1,2],
             y = line.coords[[i]][2,2] + par()$cxy[2] * 0.3,
             labels = line.label[i],
             pos = 2,
             col = line.col[i],
             cex = 0.9)
    }

    ## overplot unwanted parts
    bex <- ellipse[1, 1]                    # start of ellipse (x)
    bey <- min(ellipse[1, 2], -2)           # start of ellipse (y)
    eex <- ellipse[num.values, 1]           # end of ellipse (x)
    par.usr <- par("usr")
    x1 <- min(par.usr[1], ellipse[, 1])     # left
    x2 <- max(par.usr[2], ellipse[, 1])     # right
    y1 <- min(par.usr[3], ellipse[, 2], -2) # bottom
    y2 <- max(par.usr[4], ellipse[, 2], 2)  # top
    polygon(x = c(ellipse[, 1], eex, x2, x2, x1, x1,  bex),
            y = c(ellipse[, 2], y2,  y2, y1, y1, bey, bey),
            col = "white",
            lty = 0)

    ## add plot title
    title(main = main, line = shift.lines, font = 2)

    ## plot lower x-axis (precision)
    x.axis.ticks <- axTicks(side = 1)
    x.axis.ticks <- x.axis.ticks[c(TRUE, x.axis.ticks <= limits.x[2])]
    x.axis.ticks <- x.axis.ticks[x.axis.ticks <= limits.x[2]]

    ## axis with labels and ticks
    axis(side = 1,
         at = x.axis.ticks,
         lwd = 1,
         xlab = "")

    ## extend axis line to right side of the plot
    lines(x = c(max(x.axis.ticks, na.rm = TRUE), limits.x[2]),
          y = c(limits.y[1], limits.y[1]))

    ## draw closing tick on right hand side
    axis(side = 1, tcl = 0.5, lwd = 0, lwd.ticks = 1, at = limits.x[2],
         labels = FALSE)
    axis(side = 1, tcl = -0.5, lwd = 0, lwd.ticks = 1, at = limits.x[2],
         labels = FALSE)

    ## add upper and lower axis label
    mtext(text = xlab,
          at = (limits.x[1] + limits.x[2]) / 2,
          side = 1,
          line = c(-3.5, 2.5),
          cex = cex)

    ## plot upper x-axis
    axis(side = 1,
         tcl = 0.5,
         lwd = 0,
         lwd.ticks = 1,
         at = x.axis.ticks[-1],
         labels = FALSE)

    ## remove first tick label (infinity)
    label.x.upper <- label.x.upper[1:(length(x.axis.ticks) - 1)]

    ## add tick labels
    axis(side = 1,
         lwd = 0,
         labels = label.x.upper,
         at = x.axis.ticks[-1],
         line = -3)

    ## plot minor z-ticks
    for(i in 1:length(tick.values.minor)) {
      lines(x = c(tick.x1.minor[i], tick.x2.minor[i]),
            y = c(tick.y1.minor[i], tick.y2.minor[i]))
    }

    ## plot major z-ticks
    for(i in 1:length(tick.values.major)) {
      lines(x = c(tick.x1.major[i], tick.x2.major[i]),
            y = c(tick.y1.major[i], tick.y2.major[i]))
    }

    ## plot z-axis
    lines(ellipse)

    ## plot z-values
    text(x = label.x,
         y = label.y,
         label = label.z.text, 0)

    ## plot z-label
    mtext(side = 4,
          at = 0,
          line = 5,
          las = 3,
          adj = 0.5,
          cex = cex,
          text = zlab)

    ## optionally add rug
    if (rug) {
      rug.values <- if (log.z) log(De.global) else De.global

      for (i in 1:length(rug.values)) {
        rug.x <- x.coord(rug.values[i]) * c(0.988, 0.995)
        rug.y <- y.coord(rug.values[i], rug.x)
        lines(x = rug.x,
              y = rug.y,
              col = col[data.global[i,9]])
      }
    }

    ## plot values
    for(i in 1:length(data)) {
      points(data[[i]][,6][data[[i]][,6] <= limits.x[2]],
             data[[i]][,8][data[[i]][,6] <= limits.x[2]],
             col = col[i],
             pch = pch[i])
    }

    ## optionally add min, max, median sample text
    if(length(stats) > 0) {
      text(x = stats.data[,1],
           y = stats.data[,2],
           labels = round(stats.data[,3], 1),
           pos = 2,
           cex = 0.85)
    }

    ## optionally add legend content
    if (!is.null(legend)) {
      coords <- .get_keyword_coordinates(legend.pos, limits.x, limits.y)
      legend.pos <- coords$pos
      legend.adj <- coords$adj
      legend(x = legend.pos[1],
             y = 0.8 * legend.pos[2],
             xjust = legend.adj[1],
             yjust = legend.adj[2],
             legend = legend,
             pch = pch,
             col = col,
             text.col = col,
             cex = 0.8,
             bty = "n")
    }

    ## plot y-axis
    if (y.ticks) {
      char.height <- par()$cxy[2]
      if (char.height > 4.5 / cex) {
        axis(side = 2, las = 2, at = 0, labels = "\uB1 2", tcl = 0, hadj = 0.5)
      } else {
        axis(side = 2, las = 2, at = c(-2, 0, 2))
      }
    } else {
      axis(side = 2, at = 0)
    }

    ## optionally add subheader text
    mtext(side = 3,
          line = shift.lines - 2,
          text = mtext,
          cex = 0.8 * cex)

    ## add summary content
    for(i in 1:length(data)) {
      if(summary.pos[1] != "sub") {
        text(x = summary.pos[1],
             y = 0.8 * summary.pos[2],
             adj = summary.adj,
             labels = label.text[[i]],
             cex = 0.8,
             col = col[i])
      } else if (mtext == "") {
          mtext(side = 3,
                line = shift.lines - 1 - i,
                text = label.text[[i]],
                col = col[i],
                cex = 0.8 * cex)
      }
    }

    ##FUN by R Luminescence Team
    if (fun) sTeve() # nocov
  }

  invisible(list(data = data,
                data.global = data.global,
                xlim = limits.x,
                ylim = limits.y,
                zlim = limits.z,
                r = r,
                plot.ratio = plot.ratio,
                ticks.major = ticks.major,
                ticks.minor = ticks.minor,
                labels = labels,
                polygons = polygons,
                ellipse.lims = ellipse.lims))
}

## load data
data(ExampleData.DeValues, envir = environment())

set.seed(12310)
x <- rnorm(30, 5, 0.5)
y <- x * runif(30, 0.05, 0.10)
df <- data.frame(x, y)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RadialPlot("error"),
               "All elements of 'data' should be of class 'data.frame' or")
  expect_error(plot_RadialPlot(list()),
               "'data' cannot be an empty list")
  expect_error(plot_RadialPlot(df[, 1]),
               "All elements of 'data' should be of class 'data.frame' or")
  expect_error(plot_RadialPlot(df[0, ]),
               "'data' cannot be an empty data.frame")
  expect_error(plot_RadialPlot(list(df[0, ])),
               "Input 'data[[1]]' cannot be an empty data.frame",
               fixed = TRUE)
  expect_error(plot_RadialPlot(data.frame(NA, 1:5)),
               "After NA removal, nothing is left from data set 1")
  expect_error(plot_RadialPlot(data.frame(1, 3)),
               "At least two data points are required")
  expect_error(plot_RadialPlot(df, na.rm = -1),
               "'na.rm' should be a single logical value")
  expect_error(plot_RadialPlot(df, central.value = -1),
               "'central.value' should be a single positive value")
  expect_error(plot_RadialPlot(df, xlab = "x"),
               "'xlab' should be of class 'character' and have length 2")
  expect_error(plot_RadialPlot(df, centrality = list("error")),
               "'centrality' should be of class 'character' or 'numeric'")
  expect_error(plot_RadialPlot(df, centrality = "error"),
               "'centrality' should be one of 'mean', 'mean.weighted', 'median'")
  expect_error(plot_RadialPlot(df, summary = 5),
               "'summary' should be of class 'character'")
  expect_error(plot_RadialPlot(df, summary.pos = 5),
               "'summary.pos' should have length 2")
  expect_error(plot_RadialPlot(df, summary.pos = list()),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_RadialPlot(df, summary.pos = "error"),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_RadialPlot(df, plot.ratio = NA),
               "'plot.ratio' should be a single positive value or NULL")
  expect_error(plot_RadialPlot(df, line = c(NA, NA)),
               "'line' should be of class 'numeric', 'integer' or NULL")
  expect_error(plot_RadialPlot(df, zlim = 1),
               "'zlim' should be of class 'numeric' and have length 2")
  expect_error(plot_RadialPlot(list(df, df), lty = 1),
               "'lty' should have length 2")

  expect_warning(plot_RadialPlot(df, xlim = c(-1, 100), show = FALSE),
                 "Lower x-axis limit not set to zero, corrected")
  expect_warning(plot_RadialPlot(ExampleData.DeValues, log.z = FALSE,
                                 xlim = c(0, 5), zlim = c(100, 200),
                                 show = FALSE),
                 "'log.z' is set to 'FALSE' altough more than one data set")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## standard data
  expect_silent(
  plot_RadialPlot(
    data = df,
    centrality = 5))

  ## standard data with two datasets
  expect_silent(
    plot_RadialPlot(
      data = list(df, df),
      centrality = c(5,5)))

  ## simple test - unlogged
  expect_silent(
    plot_RadialPlot(
      data = df,
      centrality = 5,
      log.z = FALSE))

  ## simple test - unlogged with statistics
  expect_silent(
    plot_RadialPlot(
      data = df,
      summary = c(
        "n", "mean", "median", "mean.weighted", "median.weighted", "kdemax", "sdabs",
        "sdrel", "seabs", "serel", "skewness", "kurtosis", "in.2s", "sdabs.weighted",
        "sdrel.weighted", "seabs.weighted", "serel.weighted"),
      log.z = FALSE))

  ## simple test - unlogged
  expect_silent(
    plot_RadialPlot(
      data = df,
      centrality = -1,
      log.z = FALSE))

  ## single-column data frame
  expect_message(plot_RadialPlot(data.frame(x = c(-0.1, -1.2, 10))),
                 "Small standardised estimate scatter")

  ## data frame with more than 2 columns
  expect_silent(plot_RadialPlot(cbind(df, df)))

  ## data frame with negative values
  df.neg <- df
  df.neg[, 1] <- df.neg[, 1] - 5
  plot_RadialPlot(df.neg)

  ## data frame with zeros
  df.zeros <- data.frame(ED = c(rep(0, 4), 10),
                         ED_Error = rnorm(5) + 3)
  expect_silent(plot_RadialPlot(df.zeros))
  expect_silent(plot_RadialPlot(df.zeros, zlim = c(5, 10),
                                centrality = "median.weighted"))

  ## more coverage
  expect_type(plot_RadialPlot(df, main = "Title", sub = "Subtitle", rug = TRUE,
                              centrality = "mean", log.z = TRUE,
                              stats = c("min", "max", "median"),
                              summary = "mean", summary.pos = c(0, 40),
                              legend = "Data", legend.pos = c(4, 10),
                              xlab = c("x1", "x2"), xlim = c(0, 20),
                              ylab = "y", ylim = c(-10, 10),
                              zlab = "z", zlim = c(3, 7),
                              line = c(3.5, 5.5), y.ticks = FALSE,
                              cex = 0.8, lty = 2, lwd = 2, pch = 2, col = 2,
                              tck = 1, tcl = 2),
              "list")

  plot_RadialPlot(df, show = FALSE, centrality = c(1, 2, 3))
  plot_RadialPlot(df, show = FALSE, centrality = "median",
                  summary.pos = "topleft", legend.pos = "topright",
                  log.z = FALSE, rug = TRUE)

  ## RLum.Results object
  expect_silent(plot_RadialPlot(calc_CommonDose(ExampleData.DeValues$BT998,
                                                verbose = FALSE)))

  # Messages,  Warnings, and Errors -----------------------------------------
  ## trigger message
  expect_message(
    plot_RadialPlot(
      data = data.frame(x = df$x, y = rep(0.0001, nrow(df))),
      centrality = -1,
      log.z = FALSE),
    regexp = "Small standardised estimate scatter")

  expect_message(plot_RadialPlot(df, line = -1),
                 "Lines with negative value skipped due to 'log.z = TRUE'")

  ## trigger warning
  expect_warning(plot_RadialPlot(
      data = df,
      #centrality = ,
      central.value = -1,
      log.z = FALSE),
      "z-scale touches 2s-polygon, decrease plot ratio"
    )
  expect_silent(plot_RadialPlot(df, central.value = -1, log.z = FALSE,
                                bar.col = "none"))

  expect_warning(plot_RadialPlot(data.frame(c(12, 2, 7), c(0, 2, 3))),
                 "Error values cannot be zero or NA, reset to 1e-09")
  expect_message(expect_warning(plot_RadialPlot(data.frame(1:5, NA)),
                                "Error values cannot be zero or NA, reset to"),
                 "Small standardised estimate scatter, toggle off y.ticks?")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("RadialPlot defaults",
                              plot_RadialPlot(df, centrality = 6))
  vdiffr::expect_doppelganger("RadialPlot summary sub",
                              plot_RadialPlot(df, summary.pos = "sub",
                                              stat = c("min", "max"),
                                              summary = c("n", "se.rel", "kurtosis")))
  vdiffr::expect_doppelganger("RadialPlot summary left",
                              plot_RadialPlot(df, summary.pos = "left",
                                              summary = c("mean", "in.2s", "skewness")))
  vdiffr::expect_doppelganger("summary line cex",
                              plot_RadialPlot(df, summary.pos = "topleft",
                                              summary = "seabs.weighted",
                                              line = c(4.5, 6.5),
                                              line.label = c("4.5", "6.5"),
                                              cex = 2))
  vdiffr::expect_doppelganger("central value xlim zlim pch",
                              plot_RadialPlot(ExampleData.DeValues$CA1,
                                              central.value = 69.9,
                                              xlim = c(0, 16),
                                              zlim = c(15, 143),
                                              pch = 1,
                                              summary = ""))
  vdiffr::expect_doppelganger("regression 1044",
                              plot_RadialPlot(ExampleData.DeValues$CA1,
                                              xlim = c(0, 21.2),
                                              zlim = c(23, 135.1),
                                              summary = ""))
  vdiffr::expect_doppelganger("regression 1060",
                              plot_RadialPlot(ExampleData.DeValues$CA1,
                                              xlim = c(0, 30),
                                              zlim = c(5, 1000),
                                              plot.ratio = 0.0001))
  vdiffr::expect_doppelganger("regression 1194",
                              plot_RadialPlot(ExampleData.DeValues$CA1,
                                              xlim = c(0, 0.416),
                                              zlim = c(55, 87),
                                              log.z = FALSE,
                                              central.value = 13,
                                              centrality = 'mean'))
  df2 <- data.frame(ED = c(12, 13), ED_Error = c(1, 2))
  vdiffr::expect_doppelganger("regression 1194_2",
                              plot_RadialPlot(list(ExampleData.DeValues$CA1, df2),
                                              xlim = c(0, 16),
                                              zlim = c(9.6, 143),
                                              central.value = 68))
  df2 <- data.frame(x = df$x - 1, y = df$y * 0.75)
  vdiffr::expect_doppelganger("RadialPlot list",
                              plot_RadialPlot(list(df, df2),
                                              centrality = "median.weighted",
                                              summary = c("n", "in.2s", "median.weighted"),
                                              rug = TRUE, col = c(2, 3)))
  })
})

test_that("regression tests", {
    testthat::skip_on_cran()

    ## issue 1140
    expect_silent(plot_RadialPlot(df, zlim = c(0, 100)))
})

#' @title Transform a CW-OSL curve into a pHM-OSL curve via interpolation under
#' hyperbolic modulation conditions
#'
#' @description
#' This function transforms a conventionally measured continuous-wave (CW)
#' OSL-curve to a pseudo hyperbolic modulated (pHM) curve under hyperbolic
#' modulation conditions using the interpolation procedure described by Bos &
#' Wallinga (2012).
#'
#' @details
#' The complete procedure of the transformation is described in Bos & Wallinga
#' (2012). The input `data.frame` consists of two columns: time (t) and
#' count values (CW(t))
#'
#' **Internal transformation steps**
#'
#' (1) log(CW-OSL) values
#'
#' (2)
#' Calculate t' which is the transformed time:
#' \deqn{t' = t-(1/\delta) * \log(1 + \delta t)}
#'
#' (3)
#' Interpolate CW(t'), i.e. use the log(CW(t)) to obtain the count values
#' for the transformed time (t'). Values beyond `min(t)` and `max(t)`
#' produce `NA` values.
#'
#' (4)
#' Select all values for t' < `min(t)`, i.e. values beyond the time
#' resolution of t. Select the first two values of the transformed data set
#' which contain no `NA` values and use these values for a linear fit
#' using [lm].
#'
#' (5)
#' Extrapolate values for t' < `min(t)` based on the previously
#' obtained fit parameters.
#'
#' (6)
#' Transform values using
#' \deqn{pHM(t) = (\delta t / (1 + \delta t))*c*CW(t')}
#' \deqn{c = (1 + \delta P)/\delta*P}
#' \deqn{P = length(stimulation~period)}
#'
#' (7) Combine all values and truncate all values for t' > `max(t)`
#'
#' **Note:**
#' The number of values for t' < `min(t)` depends on the stimulation rate
#' parameter `delta`. To avoid the production of too many artificial data
#' at the raising tail of the determined pHM curve, it is recommended to use
#' the automatic estimation routine for `delta`, i.e. provide no value for
#' `delta`.
#'
#' @inheritParams convert_CW2pLMi
#'
#' @param delta [vector] (*optional*):
#' stimulation rate parameter, if no value is given, the optimal value is
#' estimated automatically (see details). Smaller values of delta produce more
#' points in the rising tail of
#' the curve.
#'
#' @return
#' The function returns the same data type as the input data type with
#' the transformed curve values.
#'
#' **`RLum.Data.Curve`**
#'
#' \tabular{ll}{
#' `$CW2pHMi.x.t` \tab: transformed time values \cr
#' `$CW2pHMi.method` \tab: used method for the production of the new data points
#' }
#'
#' **`data.frame`**
#'
#' \tabular{ll}{
#' `$x` \tab: time\cr
#' `$y.t` \tab: transformed count values\cr
#' `$x.t` \tab: transformed time values \cr
#' `$method` \tab: used method for the production of the new data points
#' }
#'
#' @note
#' According to Bos & Wallinga (2012), the number of extrapolated points
#' should be limited to avoid artificial intensity data. If `delta` is
#' provided manually and more than two points are extrapolated, a warning
#' message is returned.
#'
#' The function [approx] may produce some `Inf` and `NaN` data, in which case
#' the function tries to interpolate these values by calculating the mean of
#' the adjacent channels. If two invalid values are succeeding,
#' the values are removed and no further interpolation is attempted.
#' In every case a warning message is shown.
#'
#' @section Function version: 0.2.7
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#' Based on comments and suggestions from:
#' Adrie J.J. Bos, Delft University of Technology, The Netherlands
#'
#' @seealso [Luminescence::convert_CW2pLM], [Luminescence::convert_CW2pLMi], [Luminescence::convert_CW2pPMi],
#' [Luminescence::fit_LMCurve], [lm], [Luminescence::RLum.Data.Curve-class]
#'
#' @references
#' Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL
#' signal components. Radiation Measurements 47, 752-758.\cr
#'
#' **Further Reading**
#'
#' Bulur, E., 1996. An Alternative Technique For
#' Optically Stimulated Luminescence (OSL) Experiment. Radiation Measurements
#' 26, 701-709.
#'
#' Bulur, E., 2000. A simple transformation for converting CW-OSL curves to
#' LM-OSL curves. Radiation Measurements 32, 141-145.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(1) - simple transformation
#'
#' ##load CW-OSL curve data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ##transform values
#' values.transformed <- convert_CW2pHMi(ExampleData.CW_OSL_Curve)
#'
#' ##plot
#' plot(values.transformed$x, values.transformed$y.t, log = "x")
#'
#' ##(2) - load CW-OSL curve from BIN-file and plot transformed values
#'
#' ##load BINfile
#' #BINfileData<-readBIN2R("[path to BIN-file]")
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##grep first CW-OSL curve from ALQ 1
#' curve.ID<-CWOSL.SAR.Data@@METADATA[CWOSL.SAR.Data@@METADATA[,"LTYPE"]=="OSL" &
#'                                     CWOSL.SAR.Data@@METADATA[,"POSITION"]==1
#'                                   ,"ID"]
#'
#' curve.HIGH<-CWOSL.SAR.Data@@METADATA[CWOSL.SAR.Data@@METADATA[,"ID"]==curve.ID[1]
#'                                     ,"HIGH"]
#'
#' curve.NPOINTS<-CWOSL.SAR.Data@@METADATA[CWOSL.SAR.Data@@METADATA[,"ID"]==curve.ID[1]
#'                                        ,"NPOINTS"]
#'
#' ##combine curve to data set
#' curve<-data.frame(x = seq(curve.HIGH/curve.NPOINTS,curve.HIGH,
#'                           by = curve.HIGH/curve.NPOINTS),
#'                   y=unlist(CWOSL.SAR.Data@@DATA[curve.ID[1]]))
#'
#' ##transform values
#' curve.transformed <- convert_CW2pHMi(curve)
#'
#' ##plot curve
#' plot(curve.transformed$x, curve.transformed$y.t, log = "x")
#'
#' ##(3) - produce Fig. 4 from Bos & Wallinga (2012)
#'
#' ##load data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#' values <- CW_Curve.BosWallinga2012
#'
#' ##open plot area
#' plot(NA, NA,
#'      xlim=c(0.001,10),
#'      ylim=c(0,8000),
#'      ylab="pseudo OSL (cts/0.01 s)",
#'      xlab="t [s]",
#'      log="x",
#'      main="Fig. 4 - Bos & Wallinga (2012)")
#'
#' values.t <- convert_CW2pLMi(values, P = 1/20)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col="red" ,lwd=1.3)
#' text(0.03,4500,"LM", col="red" ,cex=.8)
#'
#' values.t <- convert_CW2pHMi(values, delta = 40)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col="black", lwd=1.3)
#' text(0.005,3000,"HM", cex=.8)
#'
#' values.t <- convert_CW2pPMi(values, P = 1/10)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col="blue", lwd=1.3)
#' text(0.5,6500,"PM", col="blue" ,cex=.8)
#'
#' @export
convert_CW2pHMi<- function(
  object,
  delta = NULL,
  ...
) {
  .set_function_name("convert_CW2pHMi")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  temp.values <- .prepare_CW2pX(object)
  .validate_class(delta, "numeric", null.ok = TRUE)
  if (length(delta) == 0)
    delta <- NULL
  if (anyNA(delta)) {
    .throw_error("'delta' cannot contain NA values")
  }

  # (1) Transform values ------------------------------------------------------

  ##log transformation of the CW-OSL count values
  CW_OSL.log<-log(temp.values[,2])

  ##time transformation t >> t'
  t<-temp.values[,1]
  if (anyDuplicated(t) > 0) {
    .throw_error("'object' contains duplicated time values")
  }

  ##set delta
  ##if no values for delta is set selected a delta value for a maximum of
  ##two extrapolation points
  if (is.null(delta)) {
    i<-10
    delta<-i
    t.transformed<-t-(1/delta)*log(1+delta*t)

    while(length(t.transformed[t.transformed<min(t)])>2){

      delta<-i
      t.transformed<-t-(1/delta)*log(1+delta*t)
      i<-i+10
    }
  }else{
    t.transformed<-t-(1/delta)*log(1+delta*t)
  }

  # (2) Interpolation ---------------------------------------------------------

  ##interpolate values, values beyond the range return NA values
  CW_OSL.interpolated <- approx(t,CW_OSL.log, xout=t.transformed, rule=1)

  ## In some cases the interpolation algorithm is not working properly, and
  ## Inf or NaN values are produced
  interpolated <- .fix_interpolation_inf_nan(unlist(CW_OSL.interpolated$y),
                                             warn = TRUE)

  ## combine t.transformed and CW_OSL.interpolated in a data.frame
  temp <- data.frame(x = t.transformed, y = interpolated)

  # (3) Extrapolate first values of the curve ---------------------------------

  res <- .extrapolate_first(temp, t = t[1:2], y = CW_OSL.log[1:2])
  temp <- res$df
  temp.method <- res$method

  # (4) Convert, transform and combine values ---------------------------------

  ##unlog CW-OSL count values, i.e. log(CW) >> CW
  CW_OSL<-exp(temp$y)

  ##set values for c and P

  ##P is the stimulation period
  P<-max(temp.values[,1])

  ##c is a dimensionless constant
  c<-(1+(delta*P))/(delta*P)

  ##transform CW-OSL values to pLM-OSL values
  pHM<-((delta*t)/(1+(delta*t)))*c*CW_OSL

  ##combine all values and exclude NA values
  temp.values <- data.frame(x=t,y.t=pHM,x.t=t.transformed,method=temp.method)
  temp.values <- na.exclude(temp.values)

  # (5) Return values ---------------------------------------------------------

  ##returns the same data type as the input
  if (is.data.frame(object)) {
    return(temp.values)
  }

    ##add old info elements to new info elements
    temp.info <- c(object@info,
                   CW2pHMi.x.t = list(temp.values$x.t),
                   CW2pHMi.method = list(temp.values$method))

  set_RLum(
      class = "RLum.Data.Curve",
      recordType = object@recordType,
      data = as.matrix(temp.values[,1:2]),
      info = temp.info)
}

## load data
data(ExampleData.CW_OSL_Curve, envir = environment())
values <- CW_Curve.BosWallinga2012
object <- set_RLum(class = "RLum.Data.Curve",
                   data = as.matrix(ExampleData.CW_OSL_Curve),
                   curveType = "measured",
                   recordType = "OSL")

test_that("check functionality", {
  testthat::skip_on_cran()

  tol <- 1.5e-6
  expect_snapshot_plain(convert_CW2pLM(values),
                        tolerance = tol)
  expect_snapshot_plain(convert_CW2pLMi(values, P = 1/20),
                        tolerance = tol)
  expect_snapshot_plain(convert_CW2pLMi(values),
                        tolerance = tol)
  expect_warning(
      expect_snapshot_plain(convert_CW2pHMi(values),
                            tolerance = tol),
      "56 invalid values found, replaced by the mean of the nearest values")
  expect_warning(
      expect_snapshot_plain(convert_CW2pHMi(values, delta = 40),
                            tolerance = tol),
      "56 invalid values found, replaced by the mean of the nearest values")
  SW({ # repeated warning about invalid values
  expect_warning(
      expect_snapshot_plain(convert_CW2pHMi(values, delta = 2),
                            tolerance = tol),
      "t' is beyond the time resolution and more than two data points")
  })
  expect_warning(
      expect_snapshot_plain(convert_CW2pPMi(values, P = 1/10),
                            tolerance = tol),
      "t' is beyond the time resolution and more than two data points")

  expect_error(convert_CW2pLMi(iris[1:5, ]),
               "All points are outside the interpolation range")
  expect_error(convert_CW2pPMi(iris[1:5, ]),
               "All points are outside the interpolation range")

  ## deprecated argument
  expect_warning(convert_CW2pHMi(values = object),
                 "'values' was deprecated in v1.2.0, use 'object' instead")
  expect_warning(convert_CW2pLM(values = object),
                 "'values' was deprecated in v1.2.0, use 'object' instead")
  expect_warning(convert_CW2pLMi(values = object),
                 "'values' was deprecated in v1.2.0, use 'object' instead")
  SW({
  expect_warning(convert_CW2pPMi(values = object),
                 "'values' was deprecated in v1.2.0, use 'object' instead")
  })
})

test_that("Test RLum.Types", {
  testthat::skip_on_cran()

  expect_s4_class(convert_CW2pLM(object),
                  class = "RLum.Data.Curve")
  expect_s4_class(convert_CW2pLMi(object),
                  class = "RLum.Data.Curve")
  expect_s4_class(convert_CW2pHMi(object),
                  class = "RLum.Data.Curve")
  expect_s4_class(suppressWarnings(convert_CW2pPMi(object)),
                  class = "RLum.Data.Curve")
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_CW2pLMi(values, P = 0),
               "[convert_CW2pLMi()] 'P' should be a single positive value", fixed = TRUE)
  expect_error(convert_CW2pLMi(data.frame(-1:2, 1:4)),
               "'object' cannot contain negative times")
  expect_error(convert_CW2pLMi(data.frame(1:4, -1:2)),
               "'object' cannot contain negative counts")
  expect_warning(convert_CW2pLMi(values, P = 10),
                 "t' is beyond the time resolution and more than two data points")

  expect_error(convert_CW2pHMi(matrix(0, 2)),
               "'object' should be of class 'data.frame' or 'RLum.Data.Curve'")
  expect_error(convert_CW2pHMi(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(convert_CW2pHMi(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")
  expect_error(convert_CW2pHMi(data.frame(a = c(12, -2.1), b = c(1.34, 0))),
               "'object' cannot contain negative times")
  expect_error(convert_CW2pHMi(data.frame(a = c(12, 2.1), b = c(1.34, 0))),
               "All interpolated values are Inf/NaN/NA, check your data")
  expect_error(convert_CW2pHMi(data.frame(a = 1:10, b = NA)),
               "'object' should have at least 2 non-missing values")
  expect_error(convert_CW2pHMi(iris),
               "'object' contains duplicated time values")
  expect_error(convert_CW2pHMi(values, iris),
               "'delta' should be of class 'numeric' or NULL")
  expect_error(convert_CW2pHMi(values, NA_real_),
               "'delta' cannot contain NA values")

  expect_error(convert_CW2pLMi(matrix(0, 2)),
               "'object' should be of class 'data.frame' or 'RLum.Data.Curve'")
  expect_error(convert_CW2pLMi(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(convert_CW2pLMi(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")
  expect_error(convert_CW2pLMi(data.frame(a = 1:10, b = NA)),
               "'object' should have at least 2 non-missing values")
  expect_error(convert_CW2pLMi(iris),
               "'object' contains duplicated time values")
  expect_error(convert_CW2pLMi(object, iris),
               "'P' should be a single positive value")

  expect_error(convert_CW2pLM(matrix(0, 2)),
               "'object' should be of class 'data.frame' or 'RLum.Data.Curve'")
  expect_error(convert_CW2pLM(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(convert_CW2pLM(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")

  expect_error(convert_CW2pPMi(matrix(0, 2)),
               "'object' should be of class 'data.frame' or 'RLum.Data.Curve'")
  expect_error(convert_CW2pPMi(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(convert_CW2pPMi(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")
  expect_error(convert_CW2pPMi(data.frame(a = 1:10, b = NA)),
               "'object' should have at least 2 non-missing values")
  expect_error(convert_CW2pPMi(iris),
               "'object' contains duplicated time values")
  expect_error(convert_CW2pPMi(values, iris),
               "'P' should be a single positive value or NULL")

  object@recordType <- "RF"
  expect_error(convert_CW2pLM(object),
               "recordType RF is not allowed for the transformation")
  expect_error(convert_CW2pLMi(object),
               "[convert_CW2pLMi()] recordType RF is not allowed for the transformation",
               fixed = TRUE)
  expect_error(convert_CW2pHMi(object),
               "[convert_CW2pHMi()] recordType RF is not allowed for the transformation",
               fixed = TRUE)
  expect_error(convert_CW2pPMi(object),
               "[convert_CW2pPMi()] recordType RF is not allowed for the transformation",
               fixed = TRUE)
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1483
  expect_warning(expect_s3_class(convert_CW2pHMi(values, numeric()),
                                 "data.frame"),
                 "56 invalid values found, replaced by the mean of the nearest")
})

#' @title Apply fading correction after Lamothe et al., 2003
#'
#' @description
#' This function applies the fading correction for the prediction of long-term
#' fading as suggested by Lamothe et al., 2003. The function basically adjusts
#' the $L_n/T_n$ values and fits a new dose-response curve using function
#' [Luminescence::fit_DoseResponseCurve].
#'
#' @details
#'
#' **Format of `object` if `data.frame`**
#'
#' If `object` is a [data.frame], all input values must be of type [numeric].
#' Dose values are expected in seconds (s) not Gray (Gy). No `NA` values are
#' allowed and the value for the natural dose (first row) should be `0`.
#' Example for three dose points (column names are arbitrary):
#'
#' ```
#'  object <- data.frame(
#'  dose = c(0,25,50),
#'  LxTx = c(4.2, 2.5, 5.0),
#'  LxTx_error = c(0.2, 0.1, 0.2))
#'  ```
#'
#'  **Note on the g-value and `tc`**
#'
#' Users new to R and fading measurements are often confused about what to
#' enter for `tc` and why it may differ from `tc.g_value`. By convention
#' (Huntley & Lamothe 2001), the `tc` value is the time elapsed between the
#' end of the irradiation and the prompt measurement. Usually there is no
#' reason for having a `tc` value different for the equivalent dose measurement
#' and the *g*-value measurement, except if different equipment was used.
#' However, if, for instance, the *g*-value measurement sequence was analysed
#' with the *Analyst* (Duller 2015) and `Luminescence` is used to correct for
#' fading, there is a high chance that the value returned by the *Analyst*
#' comes normalised to 2-days, even if the `tc` values of the measurement were
#' identical. In such cases, the fading correction cannot be correct until the
#' `tc.g_value` is manually set to 2-days (`172800` s) because the function
#' will internally recalculate values to an identical `tc` value.
#'
#' @param object [Luminescence::RLum.Results-class] [data.frame] (**required**):
#' Input data for applying the fading correction, can be (1) a [data.frame]
#' with three columns (`dose`, `LxTx`, `LxTx error`; see details), or (2) an
#' [Luminescence::RLum.Results-class] object created by [Luminescence::analyse_SAR.CWOSL] or
#' [Luminescence::analyse_pIRIRSequence].
#'
#' @param dose_rate.envir [numeric] vector of length 2 (**required**):
#' Environmental dose rate in mGy/a.
#'
#' @param dose_rate.source [numeric] vector of length 2 (**required**):
#' Irradiation source dose rate in Gy/s, which is, according to Lamothe et al.
#' (2003) De/t.
#'
#' @param g_value [numeric] vector of length 2 (**required**): g_value in
#' %/decade *recalculated at the moment* the equivalent dose was calculated,
#' i.e. `tc` is either similar for the *g*-value measurement **and** the
#' De measurement or needs be to recalculated (cf. [Luminescence::calc_FadingCorr]).
#' Inserting a normalised g-value, e.g., normalised to 2-days , will
#' lead to wrong results.
#'
#' @param tc [numeric] (*optional*): time in seconds between the **end** of
#' the irradiation and the prompt measurement used in the equivalent dose
#' estimation (cf. Huntley & Lamothe 2001).
#' If set to `NULL`, it is assumed that `tc` is similar for the equivalent
#' dose estimation and the *g*-value estimation.
#'
#' @param tc.g_value [numeric] (*with default*):
#' time in seconds between irradiation and the prompt measurement estimating
#' the *g*-value. If the *g*-value was normalised to, e.g., 2 days, this time
#' in seconds (i.e., `172800`) should be entered here along with the time used
#' for the equivalent dose estimation. If nothing is provided the time is set
#' to `tc`, which is the usual case for *g*-values obtained using the SAR
#' method and *g*-values that had been not normalised to 2 days.
#' Note: If this value is not `NULL` the functions expects a [numeric] value for `tc`.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @param ... further arguments passed to [Luminescence::plot_DoseResponseCurve].
#'
#' @return
#' The function returns an [Luminescence::RLum.Results-class] object and the graphical
#' output produced by [Luminescence::plot_DoseResponseCurve].
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$data` \tab `data.frame` \tab the fading corrected values \cr
#'  `$fit` \tab `nls` \tab the object returned by the dose response curve fitting \cr
#' }
#'
#' '**slot:** **`@info`**
#'
#' The original function call
#'
#' @references
#'
#' Huntley, D.J., Lamothe, M., 2001. Ubiquity of anomalous fading in K-feldspars and the measurement
#' and correction for it in optical dating. Canadian Journal of Earth Sciences 38, 1093-1106.
#'
#' Duller, G.A.T., 2015. The Analyst software package for luminescence data: overview and recent improvements.
#' Ancient TL 33, 35–42.
#'
#' Lamothe, M., Auclair, M., Hamzaoui, C., Huot, S., 2003.
#' Towards a prediction of long-term anomalous fading of feldspar IRSL. Radiation Measurements 37,
#' 493-498.
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France)
#'
#' @keywords datagen
#'
#' @seealso [Luminescence::fit_DoseResponseCurve], [Luminescence::plot_DoseResponseCurve],
#' [Luminescence::calc_FadingCorr], [Luminescence::analyse_SAR.CWOSL], [Luminescence::analyse_pIRIRSequence]
#'
#' @examples
#'
#'##load data
#'##ExampleData.BINfileData contains two BINfileData objects
#'##CWOSL.SAR.Data and TL.SAR.Data
#'data(ExampleData.BINfileData, envir = environment())
#'
#'##transform the values from the first position in a RLum.Analysis object
#'object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#'
#'##perform SAR analysis and set rejection criteria
#'results <- analyse_SAR.CWOSL(
#' object = object,
#' signal_integral = 1:2,
#' background_integral = 900:900,
#' verbose = FALSE,
#' plot = FALSE,
#' onlyLxTxTable = TRUE
#' )
#'
#' ##run fading correction
#' results_corr <- calc_Lamothe2003(
#'   object = results,
#'   dose_rate.envir =  c(1.676 , 0.180),
#'   dose_rate.source = c(0.184, 0.003),
#'   g_value =  c(2.36, 0.6),
#'   plot = TRUE,
#'   fit.method = "EXP")
#'
#'
#'@export
calc_Lamothe2003 <- function(
  object,
  dose_rate.envir,
  dose_rate.source,
  g_value,
  tc = NULL,
  tc.g_value = tc,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_Lamothe2003")
  on.exit(.unset_function_name(), add = TRUE)

  ## Input parameter test ---------------------------------------------------

  .validate_class(object, c("data.frame", "RLum.Results"))

  .validate_length_2_vector <- function(vec) {
    name <- sprintf("'%s'", all.vars(match.call())[1])
    .validate_class(vec, "numeric", name = name)
    if (length(vec) < 2)
      .throw_error(name, " should contain 2 elements")
    if (length(vec) > 2) {
      .throw_warning(name, " has length > 2, taking only the first two entries")
      vec <- vec[1:2]
    }
    return(vec)
  }
  .validate_length_2_vector(dose_rate.envir)
  .validate_length_2_vector(dose_rate.source)
  .validate_length_2_vector(g_value)

  ##tc
  if(is.null(tc) && !is.null(tc.g_value))
    .throw_error("If you set 'tc.g_value' you have to provide a value for 'tc' too")


  # Input assignment -----------------------------------------------------------------------------
  ## We allow input as data.frame() and RLum.Results objects ... the output from functions listed
  ## below .. if we allow a data.frame it should have at least Dose, Lx/Tx, Lx/Tx Error
  if(inherits(object, "data.frame")){
    data <- object[,1:3]

    ##add signal information
    if(any(grepl(pattern = "Signal", x = colnames(object), fixed = TRUE))){
      SIGNAL <- object[[grep(pattern = "Signal", colnames(object), fixed = TRUE)[1]]]
    }else{
      SIGNAL <- NA
    }

  }else if(inherits(object, "RLum.Results")){
    .validate_originator(object, c("analyse_SAR.CWOSL", "analyse_pIRIRSequence"))

    ## get number of datasets; we have to search for the word "Natural",
    ## everything else is not safe enough
    full_table <- object@data$LnLxTnTx.table
    set_start <- grep(full_table$Name, pattern = "Natural", fixed = TRUE)
    set_end <- c(set_start[-1] - 1, nrow(full_table))

    ## columns of interest
    cols <- c("Dose", "LxTx", "LxTx.Error",
              if (.check_originator(object, "analyse_pIRIRSequence")) "Signal")
    object <- full_table[, cols]

    ## we make a self-call here since this file can contain a lot of information
    results <- lapply(1:length(set_start), function(x){
          calc_Lamothe2003(
            object = object[set_start[x]:set_end[x], ],
            dose_rate.envir = dose_rate.envir,
            dose_rate.source = dose_rate.source,
            g_value = g_value,
            tc = tc,
            tc.g_value = tc.g_value,
            verbose = verbose,
            plot = plot,
            ...
          )
        })

    ## merge output
    return(merge_RLum(results))
  }

  # Apply correction----------------------------------------------------------------------------

  ##recalculate the g-value to the given tc ...
  ##re-calculation thanks to the help by Sébastien Huot, e-mail: 2016-07-19
  if(!is.null(tc)){
    k0 <- g_value / 100 / log(10)
    k1 <- k0 / (1 - k0 * log(tc[1]/tc.g_value[1]))
    g_value <-  100 * k1 * log(10)
  }

  # transform irradiation times to dose values
  data[[1]] <- data[[1]] * dose_rate.source[1]

  ## fading correction (including dose rate conversion from Gy/s to Gy/ka)
  ## and error calculation
  ## the formula in Lamothe et al. (2003) reads:
  ## I_faded = I_unfaded*(1-g*log((1/e)*DR_lab/DR_soil)))
  rr <-  31.5576e+09 * dose_rate.source[1] / (exp(1) * dose_rate.envir[1])
  s_rr <- sqrt((dose_rate.source[2]/dose_rate.source[1])^2 + (dose_rate.envir[2]/dose_rate.envir[1])^2) * rr
  Fading_C <- 1 - g_value[1] / 100 * log10(rr)
  sFading_C <- sqrt((log10(rr) * g_value[2]/100)^2 + (g_value[1]/(100 * rr) * s_rr)^2)

  # store original Lx/Tx in new object
  LnTn_BEFORE <- data[[2]][1]
  LnTn_BEFORE.ERROR <- data[[3]][1]

  # apply to input data
  data[[2]][1] <-  data[[2]][1] / Fading_C
  data[[3]][1] <-  sqrt((data[[3]][1]/data[[2]][1])^2 +
                        (sFading_C/Fading_C)^2) * data[[2]][1]


  # Fitting ---------------------------------------------------------------------------------

  fit_results <- fit_DoseResponseCurve(data, verbose = FALSE)

  # Age calculation -----------------------------------------------------------------------------
  res <- get_RLum(fit_results)
  Age <- res[["De"]] / dose_rate.envir[1]
  s_Age <- sqrt((100 * res[["De.Error"]] / res[["De"]])^2 +
                (100 * dose_rate.envir[2] / dose_rate.envir[1])^2) * Age / 100

  if (plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    argument_list <- modifyList(list(
        object = fit_results,
        verbose = FALSE,
        main = "Corrected Dose Response Curve",
        xlab = "Dose [Gy]",
        plot_extended = FALSE),
        val = list(...))

    do.call(plot_DoseResponseCurve, args = argument_list)
  }

  # Terminal output -----------------------------------------------------------------------------
  if(verbose){
    cat("\n[calc_Lamothe2003()] \n\n")
    cat(" Used g_value:\t\t", round(g_value[1],3)," \u00b1 ",round(g_value[2],3),"%/decade \n")
    if(!is.null(tc)){
      cat(" tc for g_value:\t", tc.g_value, " s\n")
    }
    cat("\n")
    cat(" Fading_C:\t\t", round(Fading_C,3), " \u00b1 ", round(sFading_C,3),"\n")
    cat(" Corrected Ln/Tn:\t", round(data[[2]][1],3), " \u00b1 ", round(data[[3]][1],3),"\n")
    cat(" Corrected De:\t\t", round(res[["De"]], 2), " \u00b1 ", round(res[["De.Error"]], 2)," Gy \n")
    cat("--------------------------------------------------------\n")
    cat(" Corrected Age:\t\t", round(Age,2), " \u00b1 ", round(s_Age,2)," ka \n")
    cat("--------------------------------------------------------\n")
  }

  # Compile output ------------------------------------------------------------------------------
  set_RLum(
      class = "RLum.Results",
      data = list(
        data = data.frame(
          g_value = g_value[1],
          g_value.ERROR = g_value[2],
          tc = tc %||% NA,
          tc.g_value = tc.g_value %||% NA,
          FADING_C = Fading_C,
          FADING_C.ERROR = sFading_C,
          LnTn_BEFORE = LnTn_BEFORE,
          LnTn_BEFORE.ERROR = LnTn_BEFORE.ERROR,
          LnTn_AFTER = data[[2]][1],
          LnTn_AFTER.ERROR = data[[3]][1],
          DE = res[["De"]],
          DE.ERROR = res[["De.Error"]],
          AGE = Age,
          AGE.ERROR = s_Age,
          SIGNAL = SIGNAL
          ),
        fit = get_RLum(fit_results, data.object = "Fit")
      ),
      info = list(
        call = sys.call()
      )
  )
}

##################################################################################
##                      METHODS FOR S3 GENERICS                                 ##
##################################################################################

## ---------------------------------------------------------------------------##
## DATA FRAME COERCION METHOD

## This is a method for the as.data.frame S3 generic. We need this to intercept the
## DRAC list object after it hast passed the actual list-method. After it was
## coerced to a data.frame we assign new column names (DRAC ID keys) and
## make sure that all columns are either of class 'character' or 'numeric'.
## Finally, we attach a further class name to identify it as a valid DRAC object
## when passed to use_DRAC

#' @export
as.data.frame.DRAC.list <- function(x, row.names = NULL, optional = FALSE, ...) {
  DF <- as.data.frame.list(x)
  colnames(DF) <- paste0("TI:", 1:ncol(DF))
  for (i in 1:ncol(DF)) {
    if (is.factor(DF[ ,i]))
      DF[ ,i] <- as.character(DF[, i])
  }
  class(DF) <- c("data.frame", "DRAC.data.frame")
  return(DF)
}


## ---------------------------------------------------------------------------##
## PRINT METHOD

#' @export
print.DRAC.highlights <- function(x, ...) {
  mapply(function(el, name) {
    cat(attributes(el)$key, " = ", name, ":\n  ",
        paste(el, collapse = ",\n  "), "\n", sep = "")
  }, x, names(x))
}

#' @export
print.DRAC.list <- function(x, blueprint = FALSE, ...) {
  ## CASE 1: Pretty print the structure of the DRAC list
  if (!blueprint) {
    limit <- 80

    for (i in 1:length(x)) {
      # for pretty printing we insert newlines and tabs at specified lengths
      ls <- attributes(x[[i]])$description
      ls.n <- nchar(ls)
      strStarts <- seq(0, ls.n, limit)
      strEnds <- seq(limit-1, ls.n + limit, limit)
      blockString <- paste(mapply(function(start, end) {
        trimmedString <- paste(substr(ls, start, end), "\n\t\t\t")
        trimws(trimmedString, "left", whitespace = " ")
      }, strStarts, strEnds), collapse="")

      msg <- paste(attributes(x[[i]])$key, "=>",names(x)[i], "\n",
                   "\t VALUES =", .collapse(x[[i]], quote = FALSE), "\n",
                   "\t ALLOWS 'X' = ", attributes(x[[i]])$allowsX, "\n",
                   "\t REQUIRED =", attributes(x[[i]])$required, "\n",
                   "\t DESCRIPTION = ", blockString, "\n"
      )
      if (!is.null(levels(x[[i]]))) {
        msg <- paste(msg,
                     "\t OPTIONS = ", .collapse(levels(x[[i]]), quote = FALSE),
                     "\n\n")
      } else {
        msg <- paste(msg, "\n")
      }
      cat(msg)
    }
  }

  ## CASE 2: Return a 'blueprint' that can be copied from the console to a
  ## script so the user does not need to write down all >50 fields by hand
  if (blueprint) {
    var <- as.list(sys.call())[[2]]
    names <- names(x)

    for (i in 1:length(x)) {

      # in case of factors also show available levels as comments so you don't
      # have to look it up
      if (is.factor(x[[i]]))
        options <- paste("# OPTIONS:", .collapse(levels(x[[i]]), quote = FALSE))
      else
        options <- ""

      ## decide if values should be quoted
      use.quotes <- is.character(x[[i]]) || is.factor(x[[i]])
      cat(paste0(var, "$`", names[i], "` <- c(",
                 .collapse(x[[i]], quote = use.quotes), ") ", options , "\n"))
    }
    message("\n\t You can copy all lines above to your script and fill in the data.")
  }
}


## ---------------------------------------------------------------------------##
## DOUBLE SQUARE BRACKETS METHOD

#' @export
`[[<-.DRAC.list` <- function(x, i, value) {
  .set_function_name("[[<-.DRAC.list]]")
  on.exit(.unset_function_name(), add = TRUE)

  ## REJECT ALL INADEQUATE CLASSES ----
  acceptedClasses <- c("integer", "character", "numeric", "factor")
  if (!.validate_class(value, acceptedClasses,
                       throw.error = FALSE)) {
    return(x)
  }

  ## CHECK INPUT LENGTH ----
  if (!.validate_length(value, length(x[[i]]), throw.error = FALSE,
                        name = names(x)[i])) {
    return(x)
  }

  ## CHECK INPUT CLASS ----
  class.old <- attr(x[[i]], "default_class")
  class.new <- class(value)

  ## CHECK INPUT FIELDS THAT ALLOW 'X' -----
  # the following checks apply to fields that are normally numeric, but also
  # accept 'X' as input. this EXCLUDES factors!
  if (class.old != "factor") {
    # some input fields allow 'X' as input, so in terms of R can be of class
    # "character" or "numeric/integer". hence, we check if input is "X" and
    # if the field allows it. If so, we change the old class to "character".
    if (any(value == "X") && attributes(x[[i]])$allowsX) {
      if (anyNA(as.numeric(value[which(value != "X")]))) {
        .throw_warning("Cannot coerce '", value[which(value != "X")],
                       "' to a numeric value, input must be numeric or 'X'")
        return(x)
      }
      class.old <- "character"
    }

    # where the input field is already "X" we have to check whether the new
    # non-character input is allowed
    if (!all(is.na(x[[i]])) &&
        any(x[[i]] == "X") && attributes(x[[i]])$allowsX) {
        if (anyNA(as.numeric(value[which(value != "X")]))) {
          .throw_warning("Cannot coerce '", value[which(value != "X")],
                         "' to a numeric value, input must be numeric or 'X'\n")
          return(x)
        }
        class.new <- "character"
        value <- as.character(value)
    }
  }

  # numeric input can be both of class 'integer' or 'numeric'. We will
  # allow any combination and reject only non-numeric/integer input
  if (class.old == "numeric" || class.old == "integer") {
    if (class.new != class.old) {
      ## check if coercion is possible
      if (anyNA(suppressWarnings(as.numeric(value)))) {
        .throw_warning(names(x)[i], ": found ", class.new, ", expected ", class.old, " -> cannot coerce, set NAs")
         value <- if (class.old == "integer") NA_integer_ else NA_real_

      } else {
      ## try coercion
      .throw_message(names(x)[i], ": found ", class.new, ", expected ", class.old, " -> coercing to ", class.old)
        value <- if (class.old == "integer") as.integer(value) else as.numeric(value)
      }
    }

    ## silently round all values to a maximum of 5 digits
    if(inherits(value, "numeric"))
      value <- round(value,5)
  }

  # for 'factor' and 'character' elements only 'character' input is allowed
  if (class.old %in% c("factor", "character") && class.new != "character") {
      .throw_warning(names(x)[i], ": Input must be of class 'character'")
      return(x)
  }

  ## CHECK IF VALID OPTION ----
  # in case of 'factor's the user is only allowed input that matches one of
  # the options specified by the factor levels. if it is a valid option,
  # the input is converted to a factor to keep the information.
  if (class.old == "factor") {
    levels <- levels(x[[i]])
    if (!all(value %in% levels)) {
      .throw_error(names(x)[i], ": Invalid option, valid options are: ",
                     .collapse(levels))
    }
    value <- factor(value, levels)
  }

  ## WRITE NEW VALUES ----
  # we strip our custom class and the attributes, pass the object to the default generic and
  # finally re-attach our class and attributes
  tmp.attributes <- attributes(x[[i]])[names(attributes(x[[i]])) != "class"]
  class(x) <- "list"
  x <- `[[<-`(x, i, value)
  attributes(x[[i]]) <- tmp.attributes
  if (class.old == "factor")
    class(x[[i]]) <- "factor"
  class(x) <- c("DRAC.list", "list")
  return(x)
}

## ---------------------------------------------------------------------------##
## SINGLE SQUARE BRACKET METHOD

#' @export
`[<-.DRAC.list` <- function(x, i, value) {
  return(`[[<-`(x, i, value))
}

## ---------------------------------------------------------------------------##
## DOLLAR SIGN METHOD

#' @export
`$<-.DRAC.list`<- function(x, name, value) {
  # this is straightforward; retrieve the index and pass the object
  # to the custom [[<- function, which does the data verification
  index <- which(names(x) == name)
  x[[index]] <- value
  return(x)
}

#' @title Create a HTML-report for (RLum) objects
#'
#' @details This function creates a HTML-report for a given object, listing its complete
#' structure and content. The object itself is saved as a serialised .Rds file.
#' The report file serves both as a convenient way of browsing through objects with
#' complex data structures as well as a mean of properly documenting and saving
#' objects.
#'
#' The HTML report is created with [rmarkdown::render] and has the
#' following structure:
#'
#' \tabular{ll}{
#'  **Section** \tab **Description** \cr
#'  `Header` \tab A summary of general characteristics of the object \cr
#'  `Object content` \tab A comprehensive list of the complete structure and content of the provided object. \cr
#'  `Object structure` \tab Summary of the objects structure given as a table \cr
#'  `File` \tab Information on the saved RDS file \cr
#'  `Session Info` \tab Captured output from `sessionInfo()` \cr
#'  `Plots` \tab (*optional*) For `RLum-class` objects a variable number of plots \cr
#' }
#'
#' The structure of the report can be controlled individually by providing one or more of the
#' following arguments (all `logical`):
#'
#' \tabular{ll}{
#' **Argument** \tab **Description** \cr
#' `header` \tab Hide or show general information on the object \cr
#' `main` \tab Hide or show the object's content \cr
#' `structure` \tab Hide or show object's structure \cr
#' `rds` \tab Hide or show information on the saved RDS file \cr
#' `session` \tab Hide or show the session info \cr
#' `plot` \tab Hide or show the plots (depending on object) \cr
#' }
#'
#' Note that these arguments have higher precedence than `compact`.
#'
#' Further options that can be provided via the `...` argument:
#'
#' \tabular{ll}{
#' **Argument** \tab **Description** \cr
#' `short_table` \tab If `TRUE` only show the first and last 5 rows of long tables. \cr
#' `theme` \tab Specifies the Bootstrap
#' theme to use for the report. Valid themes include `"default"`, `"cerulean"`, `"journal"`, `"flatly"`,
#' `"readable"`, `"spacelab"`, `"united"`, `"cosmo"`, `"lumen"`, `"paper"`, `"sandstone"`,
#' `"simplex"`, and `"yeti"`. \cr
#' `highlight` \tab Specifies the syntax highlighting style.
#'  Supported styles include `"default"`, `"tango"`, `"pygments"`, `"kate"`, `"monochrome"`,
#'  `"espresso"`, `"zenburn"`, `"haddock"`, and `"textmate"`. \cr
#' `css` \tab `TRUE` or `FALSE` to enable/disable custom CSS styling \cr
#' }
#'
#' The following arguments can be used to customise the report via CSS (Cascading Style Sheets):
#'
#' \tabular{ll}{
#' **Argument** \tab **Description** \cr
#' `font_family` \tab Define the font family of the HTML document (default: `"arial"`) \cr
#' `headings_size` \tab Size of the `<h1>` to `<h6>` tags used to define HTML headings (default: 166%). \cr
#' `content_color` \tab Colour of the object's content (default: #a72925). \cr
#' }
#'
#' Note that these arguments must all be of class [character] and follow standard CSS syntax.
#' For exhaustive CSS styling you can provide a custom CSS file for argument `css.file`.
#' CSS styling can be turned of using `css = FALSE`.
#'
#' @param object (**required**):
#' The object to be reported on, preferably of any `RLum`-class.
#'
#' @param file [character] (*with default*):
#' A character string naming the output file. If no filename is provided a
#' temporary file is created.
#'
#' @param title [character] (*with default*):
#' A character string specifying the title of the document.
#'
#' @param compact [logical] (*with default*):
#' When `TRUE` the following report components are hidden:
#' `@@.pid`, `@@.uid`, `'Object structure'`, `'Session Info'`
#' and only the first and last 5 rows of long matrices and data frames are shown.
#' See details.
#'
#' @param timestamp [logical] (*with default*):
#' `TRUE` to add a timestamp to the filename (suffix).
#'
#' @param show_report [logical] (*with default*): If set to `TRUE` the function tries to display
#' the report output in the local viewer, e.g., within *RStudio* after rendering.
#'
#' @param launch.browser [logical] (*with default*):
#' `TRUE` to open the HTML file in the system's default web browser after
#' it has been rendered.
#'
#' @param css.file [character] (*optional*):
#' Path to a CSS file to change the default styling of the HTML document.
#'
#' @param quiet [logical] (*with default*):
#' `TRUE` to suppress printing of the pandoc command line.
#'
#' @param clean [logical] (*with default*):
#' `TRUE` to clean intermediate files created during rendering.
#'
#' @param ... further arguments passed to or from other methods and to control
#' the document's structure (see details).
#'
#' @section Function version: 0.1.5
#'
#' @author
#' Christoph Burow, University of Cologne (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#'
#' @note
#' This function requires the R packages 'rmarkdown', 'pander' and 'rstudioapi'.
#'
#' @seealso [rmarkdown::render], [pander::pander_return],
#' [pander::openFileInOS], [rstudioapi::viewer],
#' [browseURL]
#'
#' @return
#' Writes a HTML and .Rds file.
#'
#' @examples
#'
#' \dontrun{
#' ## Example: RLum.Results ----
#'
#' # load example data
#' data("ExampleData.DeValues")
#'
#' # apply the MAM-3 age model and save results
#' mam <- calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.2)
#'
#' # create the HTML report
#' report_RLum(object = mam, file = "~/CA1_MAM.Rmd",
#'             timestamp = FALSE,
#'             title = "MAM-3 for sample CA1")
#'
#' # when creating a report the input file is automatically saved to a
#' # .Rds file (see saveRDS()).
#' mam_report <- readRDS("~/CA1_MAM.Rds")
#' all.equal(mam, mam_report)
#'
#'
#' ## Example: Temporary file & Viewer/Browser ----
#'
#' # (a)
#' # Specifying a filename is not necessarily required. If no filename is provided,
#' # the report is rendered in a temporary file. If you use the RStudio IDE, the
#' # temporary report is shown in the interactive Viewer pane.
#' report_RLum(object = mam)
#'
#' # (b)
#' # Additionally, you can view the HTML report in your system's default web browser.
#' report_RLum(object = mam, launch.browser = TRUE)
#'
#'
#' ## Example: RLum.Analysis ----
#'
#' data("ExampleData.RLum.Analysis")
#'
#' # create the HTML report (note that specifying a file
#' # extension is not necessary)
#' report_RLum(object = IRSAR.RF.Data, file = "~/IRSAR_RF")
#'
#'
#' ## Example: RLum.Data.Curve ----
#'
#' data.curve <- get_RLum(IRSAR.RF.Data)[[1]]
#'
#' # create the HTML report
#' report_RLum(object = data.curve, file = "~/Data_Curve")
#'
#' ## Example: Any other object ----
#' x <- list(x = 1:10,
#'           y = runif(10, -5, 5),
#'           z = data.frame(a = LETTERS[1:20], b = dnorm(0:9)),
#'           NA)
#'
#' report_RLum(object = x, file = "~/arbitray_list")
#' }
#'
#' @export
report_RLum <- function(
  object,
  file  = tempfile(),
  title = "RLum.Report",
  compact = TRUE,
  timestamp = TRUE,
  show_report = TRUE,
  launch.browser = FALSE,
  css.file = NULL,
  quiet = TRUE,
  clean = TRUE,
  ...
) {
  .set_function_name("report_RLum")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  # check if required namespace(s) are available
  .require_suggested_package("rmarkdown", "Creating object reports")
  .require_suggested_package("pander", "Creating object reports")
  # nocov start
  if (.require_suggested_package("rstudioapi", "Creating object reports",
                                 throw.error = FALSE)) {
    isRStudio <- FALSE
  } else {
    isRStudio <- rstudioapi::isAvailable()
  }
  # nocov end

  .validate_class(file, "character", length = 1)
  .validate_class(title, "character", length = 1)
  .validate_logical_scalar(compact)
  .validate_logical_scalar(timestamp)
  .validate_logical_scalar(show_report)
  .validate_logical_scalar(launch.browser)
  .validate_class(css.file, "character", length = 1, null.ok = TRUE)
  if (!is.null(css.file) && !file.exists(css.file)) {
      .throw_error("Couldn't find the specified CSS file at '", css.file, "'")
  }
  .validate_logical_scalar(quiet)
  .validate_logical_scalar(clean)

  ## ------------------------------------------------------------------------ ##
  ## STRUCTURE ----
  structure <- list(header = TRUE,
                    main = TRUE,
                    structure = !compact,
                    rds = TRUE,
                    session = !compact,
                    plot = TRUE)

  # specifying report components has higher precedence than the 'compact' arg
  structure <- modifyList(structure, list(...))


  ## OPTIONS ----
  options <- list(short_table = compact,
                  theme = "cerulean",
                  highlight = "haddock",
                  css = TRUE)

  options <- modifyList(options, list(...))

  ## CSS DEFAULTS ----
  css <- list(font_family = "arial",
              headings_size = "166%",
              content_color = "#a72925")

  css <- modifyList(css, list(...))

  ## ------------------------------------------------------------------------ ##
  ## CREATE FILE ----

  isTemp <- missing(file)

  # make sure the filename ends with .Rmd extension
  if (!grepl(".rmd$", file, ignore.case = TRUE))
    file <- paste0(file, ".Rmd")

  # Timestamp: currently added as a suffix to the filename
  # if we were to change it to a prefix, we need to first figure out the filename
  # (i.e., separate it from the possible path) using the following regular
  # expression strsplit(string, "\\\\|\\\\\\\\|\\/|\\/\\/"). This looks for
  # \, \\, /, // and the last element is the filename.
  if (timestamp)
    file <- gsub(".rmd$", paste0(format(Sys.time(), "_%Y%b%d"), ".Rmd"), file,
                 ignore.case = TRUE)

  # sanitize file name
  file <- gsub("\\\\", "\\/", file)
  file.html <- gsub(".rmd$", ".html", file, ignore.case = TRUE)
  file.rds <- gsub(".rmd$", ".Rds", file, ignore.case = TRUE)

  # Create and open the file
  file.create(file)
  tmp <- file(file, open = "wt", blocking = TRUE)

  # save RDS file
  saveRDS(object, file.rds)

  # get object
  elements <- .struct_RLum(object, root = deparse1(substitute(object)))

  ## ------------------------------------------------------------------------ ##
  ## WRITE CONTENT ----

  # HEADER ----
  writeLines("---", tmp)
  writeLines("title: RLum.Report", tmp)
  writeLines("output:", tmp)
  writeLines("  html_document:", tmp)
  writeLines("    mathjax: null", tmp)
  writeLines("    title: RLum.Report", tmp)
  writeLines(paste("    theme:", options$theme), tmp)
  writeLines(paste("    highlight:", options$highlight), tmp)
  writeLines("    toc: true", tmp)
  writeLines("    toc_float: true", tmp)
  writeLines("    toc_depth: 6", tmp)
  if (!is.null(css.file))
    writeLines(paste("    css:", css.file), tmp)
  writeLines("    md_extensions: -autolink_bare_uris", tmp)
  writeLines("---", tmp)

  # CASCADING STYLE SHEETS ----
  if (options$css) {
    writeLines(paste0(
      "<style>",
      paste0("h1, h2, h3, h4, h5, h6 { font-size:", css$headings_size," } \n"),
      paste0("#root", seq(1,nrow(elements)), " { color: ",
             css$content_color," } \n"),
      paste0("BODY { font-family:", css$font_family, " } \n"),
      "</style>"
    ),
    tmp)
  }

  # INFO ----
  # check if Luminescence package is installed and get details
  pkg <- as.data.frame(installed.packages(), row.names = FALSE)
  pkg <- pkg[which(pkg$Package == "Luminescence"), ]

  # Title
  writeLines(paste("<div align='center'><h1>", title, "</h1></div>\n\n<hr>"), tmp)

  # write information on R, Luminescence package, Object
  if (structure$header) {
    writeLines(paste("**Date:**", Sys.time(), "\n\n",
                     "**R version:**", R.version.string, "\n\n",
                     "**Luminescence package** \n\n",
                     "**&nbsp;&nbsp;&raquo; Path:**", pkg$LibPath, "\n\n",
                     "**&nbsp;&nbsp;&raquo; Version:**", pkg$Version, "\n\n",
                     "**&nbsp;&nbsp;&raquo; Built:**", pkg$Built, "\n\n",
                     "**Object** \n\n",
                     "**&nbsp;&nbsp;&raquo; Created:**",
                     tryCatch(paste(paste(strsplit(object@.uid, '-|\\.')[[1]][1:3], collapse = "-"),
                                    strsplit(object@.uid, '-|\\.')[[1]][4]),
                              error = function(e) "-"), "\n\n",
                     "**&nbsp;&nbsp;&raquo; Class:**", class(object), "\n\n",
                     "**&nbsp;&nbsp;&raquo; Originator:**",
                     tryCatch(object@originator, error = function(e) "-"), "\n\n",
                     "**&nbsp;&nbsp;&raquo; Name:**", deparse(substitute(object)), "\n\n",
                     "**&nbsp;&nbsp;&raquo; Parent ID:**",
                     tryCatch(object@.pid, error = function(e) "-"), "\n\n",
                     "**&nbsp;&nbsp;&raquo; Unique ID:**",
                     tryCatch(object@.uid, error = function(e) "-"), "\n\n",
                     "<hr>"),
               tmp)

    if (isTemp) {
      writeLines(paste("<a href=", paste0("file:///", file.html),
                       "class='btn btn-primary' download>Save report</a>"), tmp)
      writeLines(paste("<a href=", paste0("file:///", file.rds),
                       "class='btn btn-primary' download>Save data</a> \n\n"), tmp)
    }

  }#EndOf::Header


  # OBJECT ----
  if (structure$main) {
    for (i in 1:nrow(elements)) {
      # SKIP ELEMENT?
      # hide @.pid and @.uid if this is a shortened report (default)
      if (elements$bud[i] %in% c(".uid", ".pid") && compact)
        next

      # HEADER
      short.name <- elements$bud[i]
      links <- gsub("[^@$\\[]", "", as.character(elements$branch[i]))
      type <- ifelse(nzchar(links), substr(links, nchar(links), nchar(links)), "")
      if (type == "[")
        type <- ""

      # HTML header level is determined by the elements depth in the object
      # exception: first row is always the object's name and has depth zero
      if (i == 1)
        hlevel <- "#"
      else
        hlevel <- strrep("#", elements$depth[i])

      # write header; number of dots represents depth in the object. because there
      # may be duplicate header names, for each further occurrence of a name
      # Zero-width non-joiner entities are added to the name (non visible)
      writeLines(paste0(hlevel, " ",
                        "<span style='color:#74a9d8'>",
                        strrep("..", elements$depth[i]),
                        type,
                        "</span>",
                        strrep("&zwnj;", elements$bud.freq[i]),
                        short.name[length(short.name)],
                        ifelse(elements$endpoint[i], "", paste0("{#root",i,"}")),
                        ##ifelse(elements$endpoint[i], "", "{#root}"),
                        "\n\n"),
                 tmp)

      # SUBHEADER
      # contains information on Class, Length, Dimensions, Path
      writeLines(paste0("<pre style='padding:0px;border:0px'>",
                        "<span style='color:#428bca'>",
                        " Class: </span>", elements$class[i],
                        "<span style='color:#428bca'>",
                        "   Length: </span>", elements$length[i],
                        "<span style='color:#428bca'>",
                        "   Dimensions: </span>",
                        ifelse(elements$row[i] != 0, paste0(elements$row[i], ", ", elements$col[i]), "-"),
                        "<span style='color:#428bca'>",
                        "\n Path: </span>", gsub("@", "<span>@</span>", elements$branch[i]),
                        "</pre>",
                        "\n\n"),
                 tmp)

      # TABLE CONTENT
      # the content of a branch is only printed if it was determined an endpoint
      # in the objects structure
      if (elements$endpoint[i]) {
        table <- tryCatch(eval.parent(parse(text = elements$branch[i])),
                          error = function(e) {
                            return(NULL) # nocov
                          })
        # exceptions: content may be NULL; convert raw to character to stay
        # compatible with pander::pander
        if (is.null(table) | length(table) == 0)
          table <- "NULL"
        if (any(is.raw(table)))
          table <- as.character(table)

        # exception: surround objects of class "call" with <pre> tags to prevent
        # HTML auto formatting
        if (elements$class[i] == "call") {
          table <- capture.output(table)
          writeLines("<pre>", tmp)
          for (i in 1:length(table))
            writeLines(table[i], tmp)
          writeLines("</pre>", tmp)
          table <- NULL
        }

        # shorten the table if it has more than 15 rows
        if (options$short_table &&
            (is.matrix(table) || is.data.frame(table)) && nrow(table) > 15) {
              text <- pander::pander_return(
                rbind(head(table, 5), tail(table, 5)),
                caption = "shortened (only first and last five rows shown)")

              writeLines(text, tmp)
              next
        }

        # write table using pander and end each table with a horizontal line
        writeLines(suppressWarnings(pander::pander_return(table)), tmp)
        writeLines("\n\n<hr>", tmp)
      }
    }
  }#EndOf::Main

  # OBJECT STRUCTURE ----
  if (structure$structure) {
    writeLines(paste("\n\n# Object structure\n\n"), tmp)

    elements.html <- elements
    elements.html$branch <- gsub("\\$", "&#36;", elements$branch)
    writeLines(pander::pander_return(elements.html,
                                     justify = strrep("l", ncol(elements))),
               tmp)
    writeLines("\n\n", tmp)
  }#EndOf::Structure

  if (structure$rds) {
    # SAVE SERIALISED OBJECT (.rds file) ----
    writeLines(paste("<hr># File \n\n"), tmp)

    writeLines(paste0("<code>",
                      "<a href='", normalizePath(file.rds),"' download>",
                      "Click here to access the data file", "</a>",
                      "</code>"), tmp)

    writeLines(paste("\nThe R object was saved to <span style='color:#428bca'>", normalizePath(file.rds),
                     "</span>.",
                     "To import the object into your R session with the following command:",
                     paste0("<pre>",
                            "x <- readRDS('", normalizePath(file.rds), "')",
                            "</pre>"),
                     "**NOTE:** If you moved the file to another directory or",
                     "renamed the file you need to change the path/filename in the",
                     "code above accordingly!"),
               tmp)
  }#EndOf::File

  # SESSION INFO ----
  if (structure$session) {
    writeLines(paste("\n\n<hr># Session Info\n\n"), tmp)
    sessionInfo <- capture.output(sessionInfo())
    writeLines(paste(sessionInfo, collapse = "\n\n"),
               tmp)
  }

  # PLOTTING ----
  if (structure$plot) {
    isRLumList <- is.list(object) &&
      all(sapply(object, inherits, "RLum.Data.Curve"))
    if (inherits(object, "RLum") || isRLumList) {

      # mutual exclusivity: it is either a list or an RLum-Object
      if (isRLumList)
        plotCommand <- "invisible(sapply(x, plot)) \n"
      else
        plotCommand <- "plot(x) \n"

      writeLines(paste("\n\n<hr># Plots\n\n"), tmp)
      writeLines(paste0(
        "```{r}\n",
        "library(Luminescence) \n",
        "x <- readRDS('", normalizePath(file.rds),"') \n",
        plotCommand,
        "```"),
        tmp)

      if (inherits(object, "RLum.Results") &&
          .check_originator(object, c("calc_AverageDose",
                                        "calc_CommonDose",
                                        "calc_CentralDose",
                                        "calc_FiniteMixture",
                                        "calc_FuchsLang2001",
                                        "calc_IEU",
                                        "calc_MaxDose",
                                        "calc_MinDose"))) {
        writeLines(paste0(
            "```{r}\n",
            "plot_AbanicoPlot(x) \n",
            "plot_Histogram(x) \n",
            "plot_KDE(x) \n",
            "plot_ViolinPlot(x) \n",
            "```"),
            tmp)
      }
    }
  }#EndOf::Plot

  close(tmp)
  ## ------------------------------------------------------------------------ ##
  ## CLOSE & RENDER ----
  rmarkdown::render(file, clean = clean, quiet = quiet)

  ## ------------------------------------------------------------------------ ##
  ## SHOW FILE -----

  # SHOW REPORT IN RSTUDIOS VIEWER PANE ----
  # nocov start
  if (isRStudio && show_report) {
    if (isTemp) {
      try(rstudioapi::viewer(file.html))
    } else {
      # The Viewer Pane only works for files in a sessions temp directory
      # see: https://support.rstudio.com/hc/en-us/articles/202133558-Extending-RStudio-with-the-Viewer-Pane
      file.copy(file.html, file.path(tempdir(), "report.html"), overwrite = TRUE)
      try(rstudioapi::viewer(file.path(tempdir(), "report.html")))
    }
  }

  # launch browser if desired
  # browseURL() listens on localhost to show the file with the problem that
  # the download links dont work anymore. hence, we try to open the file
  # with pander::openFileInOS and use browseURL() only as fallback
  if (launch.browser) {
    opened <- tryCatch(pander::openFileInOS(file.html), error = function(e) "error")
    if (!is.null(opened))
      try(browseURL(file.html))
  }
  # nocov end

  ## ------------------------------------------------------------------------ ##
  ## CLEANUP ----

  # note that 'clean' as also passed to rmarkdown::render
  if (clean)
    file.remove(file)

  invisible()
}


################################################################################
##                                                                            ##
##                        HELPER FUNCTIONS                                    ##
##                                                                            ##
################################################################################

# ---------------------------------------------------------------------------- #
# This is a recursive function that goes the objects structure and prints
# all slots/elements along with their class, length, depth.
# ---------------------------------------------------------------------------- #
.tree_RLum <- function(x, root) {

  ## S4 object -----
  if (isS4(x)) {

    # print -----
    cat(c(root, .class(x), base::length(x), .depth(root), FALSE, .dimension(x), "\n"), sep = "|")

    for (slot in slotNames(x)) {
      s4.root <- paste0(root, "@", slot)
      .tree_RLum(slot(x, slot), root = s4.root)
    }

    ## List objects -----
  }  else if (inherits(x, "list") | typeof(x) == "list" & !inherits(x, "data.frame")) {

    ## print -----
    cat(c(root, .class(x), base::length(x), .depth(root), FALSE,
          .dimension(x), "\n"), sep = "|")

    if (length(x) > 0) {
      element <- names(x) %||% paste0("[[", seq_along(x), "]]")

      for (i in 1:length(x)) {
        if (grepl(" ", element[i]))
          element[i] <- paste0("`", element[i], "`")

        if (is.null(names(x)))
          list.root <- paste0(root, element[i])
        else if (element[i] == "")
          list.root <- paste0(root, "[[", i, "]]")
        else
          list.root <- paste0(root, "$", element[i])

        .tree_RLum(x[[i]], root = list.root)
      }
    }

    ## Data frames -----
  } else if (inherits(x, "data.frame")) {

    if (any(sapply(x, inherits, "matrix"))) {
      element <- names(x)
      for (i in 1:length(x)) {
        if (grepl(" ", element[i]))
          element[i] <- paste0("`", element[i], "`")

        list.root <- paste0(root, "$", element[[i]])
        .tree_RLum(x[[i]], root = list.root)
      }
    } else {
      # print ----
      cat(c(root, .class(x), base::length(x), .depth(root), TRUE, .dimension(x), "\n"), sep = "|")
    }

    ## Last elements -----
  }  else {

    # print ----
    cat(c(root, .class(x), base::length(x), .depth(root), TRUE, .dimension(x), "\n"), sep = "|")
  }
}

# ---------------------------------------------------------------------------- #
# a) Derive depth in the structure tree by splitting the directory by
# indicative accessors @, $, [[
# b) Wrapper for dim() to cope with NULL values
# c) Wrapper for class() that collapses the classes of an object
# ---------------------------------------------------------------------------- #
.depth <- function(x) {
  length(strsplit(x, split = "\\$|@|\\[\\[")[[1]]) - 1
}
.dimension <- function(x) {
  ## ensure `dim` is padded with zeros if its length is shorter than 2
  dim <- numeric(2)
  dim[seq_along(dim(x))] <- dim(x)
  dim
}
.class <- function(x) {
  paste(class(x), collapse = "/")
}

# ---------------------------------------------------------------------------- #
# This function captures the output of the real worker .tree_RLum and returns
# the structure of the object as a data.frame
# ---------------------------------------------------------------------------- #
.struct_RLum <- function(x, root) {
  s <- capture.output(.tree_RLum(x, root = root))
  df <- as.data.frame(do.call(rbind, strsplit(s, "|", fixed = TRUE)), stringsAsFactors = FALSE)
  names(df) <- c("branch", "class", "length", "depth", "endpoint", "row", "col")
  df$depth <- as.integer(df$depth)
  df$length <- as.numeric(df$length)
  df$endpoint <- as.logical(df$endpoint)
  df$row <- as.integer(df$row)
  df$col <- as.integer(df$col)
  df$bud <- do.call(c, lapply(strsplit(df$branch, "\\$|@|\\[\\["),
                              function(x) x[length(x)]))
  if (any(grepl("]", df$bud, fixed = TRUE)))
    df$bud[grep("]", df$bud)] <- paste0("[[", grep("]", df$bud, value = TRUE))
  df$bud.freq <- NA # 1:nrow(df)

  # reorder data.frame
  df <- df[ ,c("branch", "bud", "bud.freq", "class",
               "length", "depth", "row", "col", "endpoint")]

  # for the report we must not have the same last element names of same
  # depth (HTML cannot discriminate between #links of <h> headers)
  ## TODO: this is highly inefficient for unnamed list due to recurrent indices
  dlevel <- max(table(df$bud), 0)

  for (i in seq_len(dlevel)) {
    unique.bud <- unique(df[is.na(df$bud.freq), ]$bud)
    df[is.na(df$bud.freq), ][match(unique.bud, df[is.na(df$bud.freq), ]$bud), ]$bud.freq <- i - 1
  }

  df
}

## load data
data(ExampleData.DeValues, envir = environment())
data(ExampleData.RLum.Analysis, envir = environment())
temp <- calc_CommonDose(ExampleData.DeValues$CA1, verbose = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(report_RLum(temp, file = iris),
               "'file' should be of class 'character' and have length 1")
  expect_error(report_RLum(temp, title = iris),
               "'title' should be of class 'character' and have length 1")
  expect_error(report_RLum(temp, compact = NA),
               "'compact' should be a single logical value")
  expect_error(report_RLum(temp, timestamp = "error"),
               "'timestamp' should be a single logical value")
  expect_error(report_RLum(temp, show_report = iris),
               "'show_report' should be a single logical value")
  expect_error(report_RLum(temp, launch.browser = c(12, 13)),
               "'launch.browser' should be a single logical value")
  expect_error(report_RLum(temp, css.file = iris),
               "'css.file' should be of class 'character' or NULL and have length 1")
  expect_error(report_RLum(temp, css.file = "error"),
               "Couldn't find the specified CSS file at 'error'")
  expect_error(report_RLum(temp, quiet = letters),
               "'quiet' should be a single logical value")
  expect_error(report_RLum(temp, clean = c(FALSE, TRUE)),
               "'clean' should be a single logical value")
})

test_that("test functionality", {
  testthat::skip_on_cran()

  ## the test fails on AppVeyor for no obvious reason on the windows
  ## platform ... attempts to reproduce this failure failed. So
  ## we skip this platform for the test
  testthat::skip_on_os("windows")

  # create the standard HTML report
  testthat::expect_null(report_RLum(object = temp, timestamp = FALSE, show_report = FALSE))
  testthat::expect_null(report_RLum(object = temp, timestamp = TRUE,
                                    show_report = FALSE, compact = FALSE))

  ## compact view
  expect_null(report_RLum(list(temp, temp, NULL),
                          show_report = FALSE, compact = TRUE))

  ## data.frame
  expect_null(report_RLum(ExampleData.DeValues$CA1))

  ## list of RLum objects
  report_RLum(IRSAR.RF.Data@records)

  ## array()
  expect_null(report_RLum(array()))

  ## names with spaces or missing
  ll <- as.list(ExampleData.DeValues$CA1)
  names(ll) <- c("", "ED Error")
  expect_null(report_RLum(ll))

  ## more coverage: data frame with a column containing a matrix, raw data,
  ## and the css.file option (any existing file is enough for coverage)
  m <- matrix(c(1, 2, 3, 4), nrow = 2, ncol = 2)
  df <- data.frame(id = 1, mat = rep(0, nrow(m)), raw = raw(2))
  df$mat <- m
  df$raw[1] <- as.raw(65)
  df$raw[2] <- as.raw(66)
  colnames(df)[2] <- "mat col"
  fake.css <- system.file("CITATION", package = "Luminescence")
  expect_null(report_RLum(df, css.file = fake.css))
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1260
  expect_silent(do.call(report_RLum, list(NaN)))
  expect_silent(do.call(report_RLum, list(iris[0, ])))
})

#' @title General plot function for RLum S4 class objects
#'
#' @description
#' The function calls object specific plot functions for RLum S4 class objects.
#'
#' @details
#' The function provides a generalised access point for plotting various
#' [Luminescence::RLum-class] objects. Depending on the input object, the
#' corresponding plot function will be selected. Allowed arguments can be
#' found in the documentations of each plot function.
#'
#' \tabular{lll}{
#' **object** \tab **corresponding plot function** \cr
#' [Luminescence::RLum.Data.Curve-class] \tab [Luminescence::plot_RLum.Data.Curve] \cr
#' [Luminescence::RLum.Data.Spectrum-class] \tab [Luminescence::plot_RLum.Data.Spectrum]\cr
#' [Luminescence::RLum.Data.Image-class] \tab [Luminescence::plot_RLum.Data.Image]\cr
#' [Luminescence::RLum.Analysis-class] \tab [Luminescence::plot_RLum.Analysis]\cr
#' [Luminescence::RLum.Results-class] \tab [Luminescence::plot_RLum.Results]
#' }
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' object of class [Luminescence::RLum-class] or a list of such objects. If a
#' list is provided, the function tries to plot every object in the list
#' according to its `RLum` class, after removing non-RLum objects.
#'
#' @param ... further arguments and graphical parameters to pass to the
#' specific plot functions. The only arguments that are supported directly are
#' `main` (plot title) and `mtext` (plot subtitle). Here `main` can be provided
#' as a list and the arguments in the
#' list will dispatched to the plots if `object` is of type `list` as well.
#'
#' @return
#' Produces a plot depending on the input object.
#'
#' @section Function version: 0.5
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#'
#' @seealso [Luminescence::plot_RLum.Analysis],
#' [Luminescence::plot_RLum.Data.Curve],
#' [Luminescence::plot_RLum.Data.Spectrum],
#' [Luminescence::plot_RLum.Data.Image],
#' [Luminescence::plot_RLum.Results]
#'
#' @keywords dplot
#'
#' @examples
#' #load Example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' #transform data.frame to RLum.Data.Curve object
#' temp <- as(ExampleData.CW_OSL_Curve, "RLum.Data.Curve")
#'
#' #plot RLum object
#' plot_RLum(temp)
#'
#' @export
plot_RLum <- function(
  object,
  ...
) {
  .set_function_name("plot_RLum")
  on.exit(.unset_function_name(), add = TRUE)

  orig.list <- inherits(object, "list")
  if (orig.list) {
    ## we might have plenty of sublists before we reach the list containing
    ## only RLum-objects
    object <- .unlist_RLum(object)
    object <- .rm_nonRLum(object)

    ## return early if there is nothing to plot
    if (length(object) == 0)
      return(NULL)
  } else {
    ## transform a single object to a list, so the logic can be simplified
    object <- list(object)
  }

  extraArgs <- list(...)

  ## allow for different plot titles
  main <- NULL
  if (!is.null(extraArgs$main)) {
    main <- if (orig.list)
              .listify(extraArgs$main, length = length(object))
            else
              list(extraArgs$main)
  }

  ## allow for different subtitles
  mtext <- NULL
  if (!is.null(extraArgs$mtext)) {
    mtext <- rep_len(extraArgs$mtext, length(object))
  } else if (orig.list && inherits(object[[1]], "RLum.Analysis")) {
    mtext <- paste("Record:", 1:length(object))
  }

  for (i in seq_along(object)) {
    .validate_class(object[[i]], "RLum")
    plot_fun <- switch(
        class(object[[i]]),
        ## here we have to prevent the partial matching of 'sub' by 'subset'
        RLum.Analysis = function(object, ...) {
          if (!"subset" %in% ...names())
            plot_RLum.Analysis(object = object, subset = NULL, ...)
          else
            plot_RLum.Analysis(object = object, ...)
        },
        RLum.Data.Curve = plot_RLum.Data.Curve,
        RLum.Data.Spectrum = plot_RLum.Data.Spectrum,
        RLum.Data.Image = plot_RLum.Data.Image,
        RLum.Results = plot_RLum.Results)

    plot_fun(object = object[[i]],
             main = main[[i]],
             mtext = mtext[[i]],
             ...)
  }
}

##++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++##
##                      METHODS FOR S3 GENERICS                                 ##
##++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++##

##CAUTION NOTE:
##(1) Please DO NOT access to the S4 objects by using the slots this causes inconsistent
## behaviour, please use the corresponding RLum-methods instead!
##
##(2) Especially, please DO NOT include S3-methods for which no S4-method is implemented! Especially
##for coercing.
##
##(3) Finally, what ever you want to implement, check whether a S4-method exists, it should
##be just passed to this methods, not the opposite, otherwise this will yield in undesired behaviour
##
##TODO: For this S3 generics so far no proper documentation exists ... we should consider
##to provide an overview within a separate document, as it becomes otherwise rather
##complicated for beginners to work with the documentation.
##


##  INTRODUCED WITH 0.5.0 ##

#' methods_RLum
#'
#' Methods for S3-generics implemented for the package 'Luminescence'.
#' This document summarises all implemented S3-generics. The name of the function
#' is given before the first dot, after the dot the name of the object that is
#' supported by this method is given, e.g. `plot.RLum.Data.Curve` can be called
#' by `plot(object, ...)`, where `object` is the `RLum.Data.Curve` object.
#'
#' The term S3-generics sounds complicated, however, it just means that something
#' has been implemented in the package to increase the usability for users new
#' in R and who are not familiar with the underlying `RLum`-object structure of
#' the package. The practical outcome is that operations and functions presented
#' in standard books on R can be used without knowing the specifics of the R
#' package `'Luminescence'`. For examples see the example section.
#'
#' @param x [Luminescence::RLum-class] or [Luminescence::Risoe.BINfileData-class] (**required**):
#' input object
#'
#' @param object [Luminescence::RLum-class] (**required**):
#' input object
#'
#' @param y [integer] (*optional*):
#' the row index of the matrix, data.frame
#'
#' @param z [integer] (*optional*):
#' the column index of the matrix, data.frame
#'
#' @param i [character] (*optional*):
#' name of the wanted record type or data object or row in the
#' [Luminescence::RLum.Data.Curve-class] object
#'
#' @param j [integer] (*optional*):
#' column of the data matrix in the [Luminescence::RLum.Data.Curve-class] object
#'
#' @param value [numeric] (**required**):
#' numeric value which replace the value in the [Luminescence::RLum.Data.Curve-class] object
#'
#' @param drop [logical] (*with default*):
#' keep object structure or drop it
#'
#' @param subset `[subset]` [expression] (**required**):
#' logical expression indicating elements or rows to keep, this function works
#' in [Luminescence::Risoe.BINfileData-class] objects like [subset.data.frame], but takes care
#' of the object structure. Works also on [Luminescence::RLum.Analysis-class] objects.
#'
#' @param row.names [logical] (*with default*):
#' enable/disable row names (`as.data.frame`).
#'
#' @param recursive [logical] (*with default*):
#' enable/disable further sub-setting (`unlist`).
#'
#' @param optional [logical] (*with default*):
#' logical. If TRUE, setting row names and converting column names
#' (to syntactic names: see make.names) is optional (see [base::as.data.frame])
#'
#' @param ... further arguments that can be passed to the method
#'
#' @note
#' `methods_RLum` are not really new functions, everything given here are mostly just
#' surrogates for existing functions in the package.
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.RLum.Analysis, envir = environment())
#'
#' @keywords internal
#' @name methods_RLum
NULL

# plot()------------------------------------------------------------------------

#' @rdname methods_RLum
#' @method plot list
#' @export
plot.list <- function(x, y, ...) {
  if (all(sapply(x, inherits, "RLum"))) {
    plot_RLum(object = x, ...)
  }
  else {
    if (missing(y))
      y <- NULL
    graphics::plot.default(x, y, ...)
  }
}

#' @rdname methods_RLum
#' @method plot RLum.Results
#' @export
plot.RLum.Results <- function(x, y, ...) plot_RLum(object = x, ...)

#' @rdname methods_RLum
#' @method plot RLum.Analysis
#' @export
plot.RLum.Analysis <- function(x, y, ...) plot_RLum(object = x, ...)

#' @rdname methods_RLum
#' @method plot RLum.Data.Curve
#' @export
plot.RLum.Data.Curve <- function(x, y, ...) plot_RLum(object = x, ...)

#' @rdname methods_RLum
#' @method plot RLum.Data.Spectrum
#' @export
plot.RLum.Data.Spectrum <- function(x, y, ...) plot_RLum(object = x, ...)

#' @rdname methods_RLum
#' @method plot RLum.Data.Image
#' @export
plot.RLum.Data.Image <- function(x, y, ...) plot_RLum(object = x, ...)

#' @rdname methods_RLum
#' @method plot Risoe.BINfileData
#' @export
plot.Risoe.BINfileData <- function(x, y, ...) plot_Risoe.BINfileData(data = x, ...)

# hist() -----------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
hist.RLum.Results <- function(x, ...) plot_Histogram(data = x, ...)

#' @rdname methods_RLum
#' @export
hist.RLum.Data.Image <- function(x, ...) hist(x = get_RLum(x), ...)

#' @rdname methods_RLum
#' @export
hist.RLum.Data.Curve <- function(x, ...) hist(as(get_RLum(x),"matrix")[,2])

#' @rdname methods_RLum
#' @export
hist.RLum.Analysis <- function(x, ...) lapply(1:length_RLum(x), function(z){
  hist(as(get_RLum(x, record.id = z, ...),"matrix")[,2])})

# summary() --------------------------------------------------------------------

#' @rdname methods_RLum
#' @method summary RLum.Results
#' @export
summary.RLum.Results <- function(object, ...) get_RLum(object = object, ...)

#' @rdname methods_RLum
#' @method summary RLum.Analysis
#' @export
summary.RLum.Analysis <- function(object, ...) lapply(object@records, function(x) summary(x@data))

#' @rdname methods_RLum
#' @method summary RLum.Data.Image
#' @export
summary.RLum.Data.Image <- function(object, ...) summary(object@data)

# summary.RLum.Data.Spectrum <- function(object, ...)

#' @rdname methods_RLum
#' @method summary RLum.Data.Curve
#' @export
summary.RLum.Data.Curve <- function(object, ...) summary(object@data, ...)

# subset() ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @method subset Risoe.BINfileData
#'
#' @param records.rm [subset] [logical] (*with default*):
#' whether records should be removed from the data set. If `FALSE`, the `SEL`
#' column marks the selected records.
#'
#' @export
subset.Risoe.BINfileData <- function(x, subset, records.rm = TRUE, ...) {
  .set_function_name("subset.Risoe.BINfileData")
  on.exit(.unset_function_name(), add = TRUE)

  if (...length())
    .throw_warning("Argument not supported and skipped: ", ...names())

  ##select relevant rows
  sel <- tryCatch(eval(
    expr = substitute(subset),
    envir = x@METADATA,
    enclos = parent.frame()
  ),
  error = function(e) {
    .throw_error("Invalid subset options, valid terms are: ",
                 .collapse(names(x@METADATA), quote = FALSE))
  })

  ##probably everything is FALSE now?
  if (records.rm) {
    if (any(sel)) {
      x@METADATA <- x@METADATA[sel, ]
      x@DATA <- x@DATA[sel]
      x@METADATA[["ID"]] <- 1:length(x@METADATA[["ID"]])
      ## the .RESERVED slot may not be there, in which case we don't subset it,
      ## otherwise we generate a list of NULLs which can't be written out by
      ## write_R2BIN()
      if (length(x@.RESERVED) > 0)
        x@.RESERVED <- x@.RESERVED[sel]
      return(x)

    } else{
      return(NULL)
    }
  }else{
    x@METADATA[["SEL"]] <- sel
    return(x)
  }
}

#' @rdname methods_RLum
#' @method subset RLum.Analysis
#' @export
subset.RLum.Analysis <- function(x, subset = NULL, ...) {
  do.call(get_RLum, list(object = x, drop = FALSE, subset = substitute(subset), env = parent.frame())) }

# bin() ------------------------------------------------------------------------
#' @rdname methods_RLum
#' @export
bin <- function(x, ...) {
  UseMethod("bin")
}

#' @rdname methods_RLum
#' @export
bin.RLum.Data.Curve <- function(x, bin_size = 2, ...) bin_RLum.Data(x, bin_size = bin_size)

#' @rdname methods_RLum
#' @export
bin.RLum.Data.Spectrum <- function(x, bin_size.row = 1, bin_size.col = 1, ...){
  bin_RLum.Data(x, bin_size.row = bin_size.row, bin_size.col = bin_size.col)
}

# length() ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
length.RLum.Results <- function(x, ...) length_RLum(x)

#' @rdname methods_RLum
#' @export
length.RLum.Analysis <- function(x, ...) length_RLum(x)

#' @rdname methods_RLum
#' @export
length.RLum.Data.Curve <- function(x, ...) length_RLum(x)

#' @rdname methods_RLum
#' @export
length.Risoe.BINfileData <- function(x, ...) length(x@METADATA$ID)

# dim() ------------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
dim.RLum.Data.Curve <- function(x) dim(as(x, "matrix"))

#' @rdname methods_RLum
#' @export
dim.RLum.Data.Spectrum <- function(x) dim(as(x, "matrix"))

# rep() -------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
rep.RLum <- function(x, ...) replicate_RLum(x, ...)


# names() -----------------------------------------------------------------

#' @rdname methods_RLum
#' @export
names.RLum.Data.Curve <- function(x, ...) names_RLum(x)

#' @rdname methods_RLum
#' @export
names.RLum.Data.Spectrum <- function(x, ...) names_RLum(x)

#' @rdname methods_RLum
#' @export
names.RLum.Data.Image <- function(x, ...) names_RLum(x)

#' @rdname methods_RLum
#' @export
names.RLum.Analysis <- function(x, ...) names_RLum(x)

#' @rdname methods_RLum
#' @export
names.RLum.Results <- function(x, ...) names_RLum(x)

#' @rdname methods_RLum
#' @export
names.Risoe.BINfileData <- function(x)  as.character(x@METADATA$LTYPE)


# row.name() --------------------------------------------------------------

#' @rdname methods_RLum
#' @export
row.names.RLum.Data.Spectrum <- function(x, ...) rownames(as(x, "matrix"))


# as.data.frame() ---------------------------------------------------------

#' @rdname methods_RLum
#' @export
as.data.frame.RLum.Data.Curve <- function(x, row.names = NULL, optional = FALSE, ...) as(x, "data.frame")

#' @rdname methods_RLum
#' @export
as.data.frame.RLum.Data.Spectrum <- function(x,  row.names = NULL, optional = FALSE, ...) as(x, "data.frame")
# for RLum.Results ... makes no sense and may yield in unpredictable behaviour

#' @rdname methods_RLum
#' @export
as.data.frame.Risoe.BINfileData <- function(x,  row.names = NULL, optional = FALSE, ...) {

  ## set matrix
  m <- matrix(NA, ncol = max(lengths(x@DATA)), nrow = length(x@DATA))

  ## rename columns
  colnames(m) <- paste0("C",1:ncol(m))

  ## fill matrix
  for(i in 1:length(x@DATA)) m[i,1:length(x@DATA[[i]])] <- x@DATA[[i]]

  ##convert to data.frame and bind
  cbind(x@METADATA, as.data.frame(m))
}


# as.list() ---------------------------------------------------------------

#' @rdname methods_RLum
#' @export
as.list.RLum.Results <- function(x, ...) as(x, "list")

#' @rdname methods_RLum
#' @export
as.list.RLum.Data.Curve <- function(x, ...) as(x, "list")

#' @rdname methods_RLum
#' @export
as.list.RLum.Data.Image <- function(x, ...) as(x, "list")

#' @rdname methods_RLum
#' @export
as.list.RLum.Data.Spectrum <- function(x, ...) as(x, "list")

#' @rdname methods_RLum
#' @export
as.list.RLum.Analysis <- function(x, ...) as(x, "list")


# as.matrix() -------------------------------------------------------------

#' @rdname methods_RLum
#' @export
as.matrix.RLum.Data.Curve <- function(x, ...) as(x, "matrix")

#' @rdname methods_RLum
#' @export
as.matrix.RLum.Data.Spectrum <- function(x, ...) as(x, "matrix")

#' @rdname methods_RLum
#' @export
as.matrix.RLum.Data.Image <- function(x, ...) as(x, "matrix")
# for RLum.Results ... makes no sense and may yield in unpredictable behaviour


# merge() -----------------------------------------------------------------

#' @rdname methods_RLum
#' @export
merge.RLum <- function(x, y, ...) merge_RLum(append(list(...), values = c(x, y)))


# unlist() ----------------------------------------------------------------

#' @rdname methods_RLum
#' @method unlist RLum.Analysis
#' @export
unlist.RLum.Analysis <- function(x, recursive = TRUE, ...){
  temp <- get_RLum(object = x, recursive = recursive, ... )
  if (!recursive)
    return(temp)
  unlist(lapply(1:length(temp), function(x){
      get_RLum(temp)
    }), recursive = FALSE)
}


# `+` ---------------------------------------------------------------------

#' @rdname methods_RLum
#'
#' @examples
#'
#' ##combine curve is various ways
#' curve1 <- IRSAR.RF.Data[[1]]
#' curve2 <-  IRSAR.RF.Data[[1]]
#' curve1 + curve2
#' curve1 - curve2
#' curve1 / curve2
#' curve1 * curve2
#'
#' @export
`+.RLum.Data.Curve` <- function(x, y) merge_RLum(list(x, y), merge.method = "sum")

#' @rdname methods_RLum
#' @export
`+.RLum.Data.Spectrum` <- function(x, y) merge_RLum(list(x, y), merge.method = "sum")


# `-` ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`-.RLum.Data.Curve` <- function(x, y) merge_RLum(list(x, y), merge.method = "-")

#' @rdname methods_RLum
#' @export
`-.RLum.Data.Spectrum` <- function(x, y) merge_RLum(list(x, y), merge.method = "-")


# `*` ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`*.RLum.Data.Curve` <- function(x, y) merge_RLum(list(x, y), merge.method = "*")

#' @rdname methods_RLum
#' @export
`*.RLum.Data.Spectrum` <- function(x, y) merge_RLum(list(x, y), merge.method = "*")


# `/` ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`/.RLum.Data.Curve` <- function(x, y) merge_RLum(list(x, y), merge.method = "/")

#' @rdname methods_RLum
#' @export
`/.RLum.Data.Spectrum` <- function(x, y) merge_RLum(list(x, y), merge.method = "/")


# `[` ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`[.RLum.Data.Curve` <- function(x,y,z, drop = TRUE) {as(x, "matrix")[y,z, drop = drop]}

#' @rdname methods_RLum
#' @export
`[.RLum.Data.Spectrum` <- function(x,y,z, drop = TRUE) {as(x, "matrix")[y,z, drop = drop]}

#' @rdname methods_RLum
#' @export
`[.RLum.Data.Image` <- function(x,y,z, drop = TRUE) {as(x, "matrix")[y,z, drop = drop]}

#' @rdname methods_RLum
#' @export
`[.RLum.Analysis` <- function(x, i, drop = FALSE) {
  if (is.character(i)) {
    get_RLum(x, recordType = i, drop = drop)
  } else{
    get_RLum(x, record.id = i, drop = drop)
  }
}

#' @rdname methods_RLum
#' @export
`[.RLum.Results` <- function(x, i, drop = TRUE) {get_RLum(x, data.object = i, drop = drop)}


# `[<-` -------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`[<-.RLum.Data.Curve` <- function(x, i, j, value){
  x@data[i,j] <- value #this is without any S4-method, but otherwise the overhead it too high
  return(x)
}


# `[[` --------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`[[.RLum.Analysis` <- function(x, i) {
  if (is.character(i)) {
    get_RLum(x, recordType = i)
  } else{
    get_RLum(x, record.id = i)
  }
}

#' @rdname methods_RLum
#' @export
`[[.RLum.Results` <- function(x, i) {get_RLum(x, data.object = i)}


# `$` ---------------------------------------------------------------------

#' @rdname methods_RLum
#' @export
`$.RLum.Data.Curve` <- function(x, i) {get_RLum(x, info.object = i)}

#' @rdname methods_RLum
#'
#' @examples
#'
#' ##`$` access curves
#' IRSAR.RF.Data$RF
#'
#' @export
`$.RLum.Analysis` <- function(x, i) {get_RLum(x, recordType = i)}

#' @rdname methods_RLum
#' @export
`$.RLum.Results` <- function(x, i) {get_RLum(x, data.object = i)}

#' @title Visualise dose recovery test results
#'
#' @description
#' The function provides a standardised plot output for dose recovery test
#' measurements.
#'
#' @details
#' The procedure tests the accuracy of a measurement protocol to reliably
#' determine the dose of a specific sample. Here, the natural signal is erased
#' and a known laboratory dose administered, which is treated as unknown. Then
#' the De measurement is carried out and the degree of congruence between
#' administered and recovered dose is a measure of the protocol's accuracy for
#' this sample.\cr
#' In the plot the normalised De is shown on the y-axis, i.e. obtained De/Given Dose.
#'
#' @param object [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' input values containing at least De and De error. To plot
#' more than one data set in one figure, a `list` of the individual data
#' sets must be provided (e.g. `list(dataset.1, dataset.2)`).
#'
#' @param given.dose [numeric] (*optional*):
#' given dose used for the dose recovery test to normalise data.
#' If only one given dose is provided, this given dose is valid for all input
#' data sets (i.e., `object` is a list). Otherwise, a given dose for each input
#' data set has to be provided (e.g., `given.dose = c(100,200)`).
#' If `given.dose` is `NULL` or 0, the values are plotted without normalisation
#' (might be useful for preheat plateau tests).
#' **Note:** Unit has to be the same as from the input values (e.g., Seconds or
#' Gray).
#'
#' @param error.range [numeric] (*with default*):
#' symmetric error range in percent will be shown as dashed lines in the plot.
#' It can be set to 0 to remove the error ranges.
#'
#' @param preheat [numeric] (*optional*):
#' optional vector of preheat temperatures to be used for grouping the De values.
#' If specified, the temperatures are assigned to the x-axis.
#'
#' @param boxplot [logical] (*with default*):
#' plot values that are grouped by preheat temperature as boxplots.
#' Only possible when `preheat` vector is specified.
#'
#' @param mtext [character] (*with default*):
#' additional text below the plot title.
#'
#' @param summary [character] (*optional*):
#' adds numerical output to the plot.  Can be one or more out of:
#' - `"n"` (number of samples),
#' - `"mean"` (mean De value),
#' - `"weighted$mean"` (error-weighted mean),
#' - `"median"` (median of the De values),
#' - `"weighted$median"` (error-weighted median),
#' - `"sd.rel"` (relative standard deviation in percent),
#' - `"sd.abs"` (absolute standard deviation),
#' - `"se.rel"` (relative standard error) and
#' - `"se.abs"` (absolute standard error)
#'
#' and all other measures returned by the function [Luminescence::calc_Statistics].
#'
#' @param summary.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`)
#' for the statistical summary. Alternatively, the keyword `"sub"` may be
#' specified to place the summary below the plot header. However, this latter
#' option in only possible if `mtext` is not used.
#'
#' @param legend [character] vector (*optional*):
#' legend content to be added to the plot.
#'
#' @param legend.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`) for the
#' legend to be plotted.
#'
#' @param par.local [logical] (*with default*):
#' use local graphical parameters for plotting, e.g. the plot is shown in one
#' column and one row. If `par.local = FALSE`, global parameters are inherited,
#' i.e. parameters provided via `par()` work.
#'
#' @param na.rm [logical] (*with default*):
#' whether `NA` values should be removed from the input data before plotting.
#'
#' @param ... further arguments and graphical parameters passed to [plot],
#' supported are:
#' `xlab`, `ylab`, `xlim`, `ylim`, `main`, `cex`, `las` and `pch`.
#'
#' @return A plot is returned.
#'
#' @note
#' Further data and plot arguments can be added by using the appropriate R
#' commands.
#'
#' @section Function version: 0.1.17
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany)
#'
#' @seealso [plot]
#'
#' @references
#' Wintle, A.G., Murray, A.S., 2006. A review of quartz optically
#' stimulated luminescence characteristics and their relevance in
#' single-aliquot regeneration dating protocols. Radiation Measurements 41,
#' 369-391.
#'
#' @keywords dplot
#'
#' @examples
#'
#' ## read example data set and misapply them for this plot type
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## plot values
#' plot_DRTResults(
#'   ExampleData.DeValues$BT998[7:11,],
#'   given.dose = 2800,
#'   mtext = "Example data")
#'
#' ## plot values with legend
#' plot_DRTResults(
#'   ExampleData.DeValues$BT998[7:11,],
#'   given.dose = 2800,
#'   legend = "Test data set")
#'
#' ## create and plot two subsets with randomised values
#' x.1 <- ExampleData.DeValues$BT998[7:11,]
#' x.2 <- ExampleData.DeValues$BT998[7:11,] * c(runif(5, 0.9, 1.1), 1)
#'
#' plot_DRTResults(
#'   list(x.1, x.2),
#'   given.dose = 2800)
#'
#' ## some more user-defined plot parameters
#' plot_DRTResults(
#'   list(x.1, x.2),
#'   given.dose = 2800,
#'   pch = c(2, 5),
#'   col = c("orange", "blue"),
#'   xlim = c(0, 8),
#'   ylim = c(0.85, 1.15),
#'   xlab = "Sample aliquot")
#'
#' ## plot the data with user-defined statistical measures as legend
#' plot_DRTResults(
#'   list(x.1, x.2),
#'   given.dose = 2800,
#'   summary = c("n", "weighted$mean", "sd.abs"))
#'
#' ## plot the data with user-defined statistical measures as sub-header
#' plot_DRTResults(
#'   list(x.1, x.2),
#'   given.dose = 2800,
#'   summary = c("n", "weighted$mean", "sd.abs"),
#'   summary.pos = "sub")
#'
#' ## plot the data grouped by preheat temperatures
#' plot_DRTResults(
#'   ExampleData.DeValues$BT998[7:11,],
#'   given.dose = 2800,
#'   preheat = c(200, 200, 200, 240, 240))
#'
#' ## read example data set and misapply them for this plot type
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## plot values
#' plot_DRTResults(
#'   ExampleData.DeValues$BT998[7:11,],
#'   given.dose = 2800,
#'   mtext = "Example data")
#'
#' ## plot two data sets grouped by preheat temperatures
#' plot_DRTResults(
#'   list(x.1, x.2),
#'   given.dose = 2800,
#'   preheat = c(200, 200, 200, 240, 240))
#'
#' ## plot the data grouped by preheat temperatures as boxplots
#' plot_DRTResults(
#'   ExampleData.DeValues$BT998[7:11,],
#'   given.dose = 2800,
#'   preheat = c(200, 200, 200, 240, 240),
#'   boxplot = TRUE)
#'
#' @export
plot_DRTResults <- function(
  object,
  given.dose = NULL,
  error.range = 10,
  preheat = NULL,
  boxplot = FALSE,
  mtext = "",
  summary = "",
  summary.pos = "topleft",
  legend = NULL,
  legend.pos = "topright",
  par.local = TRUE,
  na.rm  = FALSE,
  ...
) {
  .set_function_name("plot_DRTResults")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  .validate_not_empty(object)
  .validate_class(given.dose, c("numeric", "integer"), null.ok = TRUE)
  if (anyNA(given.dose))
    .throw_error("'given.dose' cannot contain NA values")
  .validate_class(preheat, c("numeric", "integer"), null.ok = TRUE)
  .validate_logical_scalar(boxplot)
  if (boxplot && is.null(preheat)) {
    boxplot <- FALSE
    .throw_warning("'boxplot' requires a value in 'preheat', reset to FALSE")
  }

  .validate_class(mtext, "character", length = 1)
  valid.pos <- c("left", "center", "right", "topleft", "top", "topright",
                 "bottomleft", "bottom", "bottomright")
  .validate_class(summary, "character")
  if (is.numeric(summary.pos)) {
    .validate_length(summary.pos, 2)
  }
  else {
    summary.pos <- .validate_args(summary.pos, c("sub", valid.pos))
  }
  .validate_class(legend, "character", null.ok = TRUE)
  if (is.numeric(legend.pos)) {
    .validate_length(legend.pos, 2)
  }
  else {
    legend.pos <- .validate_args(legend.pos, valid.pos)
  }
  .validate_logical_scalar(par.local)
  .validate_logical_scalar(na.rm)

  ## Homogenise and check input data
  values <- object
  if (!inherits(values, "list"))
      values <- list(values)

  for (i in seq_along(values)) {
    .validate_class(values[[i]], c("data.frame", "RLum.Results"),
                    name = "'object'")
    if (inherits(values[[i]], "RLum.Results")) {
      val <- get_RLum(values[[i]])[, 1:2] %||% NA
      values[[i]] <- val
    } else if (ncol(values[[i]]) < 2) {
      .throw_error("'object' should have 2 columns")
    } else {
      ## mark for removal if all De values are missing
      if (all(is.na(values[[i]][, 1])))
        values[[i]] <- NA
    }
  }

  ## remove invalid records
  values[is.na(values)] <- NULL
  if (length(values) == 0) {
    .throw_error("No valid records in 'object'")
  }

  ## check for preheat temperature values
  num.de.values <- max(sapply(values, nrow))
  if (!is.null(preheat) && length(preheat) < num.de.values) {
    .throw_error("'preheat' should have length equal to the number of ",
                 "De values (", num.de.values, ")")
  }

  ## Check input arguments ----------------------------------------------------
  for (i in seq_along(values)) {

    ## keep only the required columns and assign names
    values[[i]] <- values[[i]][, 1:2]
    colnames(values[[i]]) <- c("De", "De.error")

    ##remove NA values; yes Micha, it is not that simple
    if (na.rm) {
      ##currently we assume that all input data sets comprise a similar of data
      if (!is.null(preheat) && i == length(values)) {
        ## remove preheat entries corresponding to NA values
        preheat <- preheat[!is.na(values[[i]][, 1]) &
                           !is.na(values[[i]][, 2])]
      }

      values[[i]] <- na.exclude(values[[i]])
      if (nrow(values[[i]]) == 0)
        .throw_error("No valid data remains after removing NA values")
    }
  }

  ## create global data set
  values.global <- NULL
  n.values <- NULL
  for (i in seq_along(values)) {
    values.global <- rbind(values.global, values[[i]])
    n.values <- c(n.values, nrow(values[[i]]))
  }

  ## Set plot format parameters -----------------------------------------------
  extraArgs <- list(...) # read out additional arguments list

  main <- extraArgs$main %||% "Dose recovery test"
  xlab <- extraArgs$xlab %||% ifelse(is.null(preheat),
                                     "# Aliquot", "Preheat temperature [\u00B0C]")

  ylab <- extraArgs$ylab %||% (
    if (!is.null(given.dose) && length(given.dose) > 0 && given.dose[1] > 0)
      expression(paste("Normalised ", D[e]))
    else expression(paste(D[e], " [s]"))
  )

  xlim <- extraArgs$xlim %||% (c(0, max(n.values)) + 0.5)
  ylim <- extraArgs$ylim %||% c(0.75, 1.25) # check below for further corrections if boundaries exceed set range
  cex <- extraArgs$cex %||% 1
  pch <- extraArgs$pch %||% abs(seq(from = 20, to = -100))
  las <- extraArgs$las %||% 0
  fun <- isTRUE(extraArgs$fun)

  ## calculations and settings-------------------------------------------------

  ## normalise data if given.dose is given
  if (!is.null(given.dose)) {
    .validate_not_empty(given.dose)
    if (length(given.dose) == 1) {
      given.dose <- rep(given.dose, length(values))
    }
    else if (length(given.dose) != length(values)) {
      .throw_error("'given.dose' should have length equal to the number ",
                   "of input data sets")
    }

    if (all(given.dose > 0)) {
      for (i in 1:length(values)) {
        values[[i]] <- values[[i]] / given.dose[i]
      }
    } else {
      given.dose <- NULL
    }
  }

  ##correct ylim for data set which exceed boundaries
  if ((max(sapply(values, function(x) max(x[, 1], na.rm = TRUE))) > 1.25 ||
       min(sapply(values, function(x) min(x[, 1], na.rm = TRUE))) < 0.75) &&
       (!"ylim" %in% names(extraArgs))) {
    ylim <- c(
        ## append a 0 element to the errors to avoid crashing if all errors are NA
        min(sapply(values, function(x) min(x[, 1], na.rm = TRUE) -
                                       max(c(x[, 2], 0), na.rm = TRUE))),
        max(sapply(values, function(x) max(x[, 1], na.rm = TRUE) +
                                       max(c(x[, 2], 0), na.rm = TRUE))))
  }

  ## optionally group data by preheat temperature
  if (!is.null(preheat)) {
    values.preheat <- list()
    modes <- unique(preheat)
    for(mode in modes) {
      for(j in 1:length(values)) {
        values.preheat[[length(values.preheat) + 1]] <-
          cbind(values[[j]][preheat == mode, ], mode)
      }
    }
    modes.plot <- rep(modes, each = length(values))
    xlim <- c(min(modes.plot) * 0.9, max(modes.plot) * 1.1)
  } else {
    modes.plot <- 1 + 0:max(sapply(values, nrow))
  }

  if (boxplot)
    xlim <- c(0.5, length(unique(preheat)) + 0.5)

  ## assign colour indices
  col <- extraArgs$col %||% (
    if (is.null(preheat)) {
      seq(from = 1, to = length(values))
    } else {
      rep(seq(from = 1, to = length(values)), length(unique(preheat)))
    }
  )

  ## calculate and paste statistical summary
  if(summary.pos[1] != "sub") {
    label.text <- lapply(1:length(values), function(i) {
      .create_StatisticalSummaryText(
        calc_Statistics(values[[i]]),
        keywords = summary,
        digits = 2,
        sep = " \n",
        prefix = strrep("\n", (i - 1) * length(summary))
      )
    })
  }else{
    label.text <- lapply(values, function(x) {
      .create_StatisticalSummaryText(
        calc_Statistics(x),
        keywords = summary,
        digits = 2,
        sep = " | "
      )
    })
  }

  ## keep track if the summary is in the bottom row as we may need to compute
  ## an adjustment further down, after the plot device has been opened
  summary.pos_is_bottom <- grepl("bottom", summary.pos[1])

  ## convert keywords into summary and legend placement coordinates
  coords <- .get_keyword_coordinates(summary.pos, xlim, ylim)
  summary.pos <- coords$pos
  summary.adj <- c(coords$adj[1], 1) # always top-aligned

  ## Plot output ------------------------------------------------------------

  ## determine number of subheader lines to shift the plot
  shift.lines <- if (mtext == "") {
                   if (summary.pos[1] == "sub") length(label.text) else 0
                 } else 1

  ## setup plot area
  if(par.local){
    par.default <- par(mfrow = c(1, 1), cex = cex,
                       mar = c(2.5, 2.5, shift.lines, 0) + 2.1)
    on.exit(par(par.default), add = TRUE)
  }

  ## optionally plot values and error bars
  if (!boxplot) {
    if (!missing(preheat))
      xlim <- range(modes.plot) * c(0.9, 1.1)

    ## create empty plot
    plot(NA, NA,
         xlim = xlim,
         ylim = ylim,
         xlab = xlab,
         ylab = ylab,
         las = las,
         main = "",
         xaxt = "n")
    axis(1, at = modes.plot, labels = modes.plot, las = las)

    .plot_elements(main, shift.lines, given.dose, error.range)

    ## allow assigning a separate colour to each point, but only if there
    ## is one input dataset
    oneinput <- length(values) == 1
    multicol <- oneinput && nrow(values[[1]]) == length(col)
    if (missing(preheat)) {
      ## add data and error bars
      for(i in 1:length(values)) {
        points(x = 1:nrow(values[[i]]),
               y = values[[i]][,1],
               pch = if (oneinput && nrow(values[[i]]) == length(pch)) pch else pch[i],
               col = if (multicol) col else col[i],
               cex = 1.2)

        suppressWarnings( # zero-length arrow is of indeterminate angle and so skipped
        graphics::arrows(1:nrow(values[[i]]),
               values[[i]][,1] + values[[i]][,2],
               1:nrow(values[[i]]),
               values[[i]][,1] - values[[i]][,2],
               angle = 90,
               length = 0.075,
               code = 3,
               col = if (multicol) col else col[i])
        )

        ## add summary content
        .add_summary(summary.pos, summary.pos_is_bottom, summary.adj,
                     label.text, mtext, i, cex, values, col)
      }
    } else {
      ## option for provided preheat data
      ## plot values
      for(i in 1:length(values.preheat)) {
        points(x = values.preheat[[i]][,3],
               y = values.preheat[[i]][,1],
               pch = pch[i],
               col = col[i],
               cex = 1.2)

        suppressWarnings( # zero-length arrow is of indeterminate angle and so skipped
        graphics::arrows(values.preheat[[i]][,3],
               values.preheat[[i]][,1] + values.preheat[[i]][,2],
               values.preheat[[i]][,3],
               values.preheat[[i]][,1] - values.preheat[[i]][,2],
               angle = 90,
               length = 0.075,
               code = 3,
               col = col[i])
        )
      }
    }
  }

  ## optionally, plot boxplot
  if(boxplot) {
    values.boxplot <- lapply(values.preheat, function(x) x[, 1])

    ## create empty plot
    graphics::boxplot(values.boxplot,
            names = modes.plot,
            ylim = ylim,
            xlab = xlab,
            ylab = ylab,
            las = las,
            xaxt = "n",
            main = "",
            border = col)

    ## add axis label, if necessary
    if (length(modes.plot) <= length(unique(modes.plot))) {
      axis(side = 1, at = 1:length(unique(modes.plot)),
           labels = unique(modes.plot), las = las)
    } else {
      ticks <- seq(from = 1 + ((length(values.boxplot)/length(unique(modes.plot)) - 1)/2),
                   to = length(values.boxplot),
                   by = length(values.boxplot)/length(unique(modes.plot)))

      axis(
        side = 1,
        at = ticks,
        las = las,
        labels = unique(modes.plot))

      ##polygon for a better graphical representation of the groups
      polygon.x <- seq(
        1,length(values.boxplot),
        by = length(values.boxplot) / length(unique(modes.plot))
      )

      polygon.step <- unique(diff(polygon.x) - 1)
      if (length(polygon.step) == 0)
        polygon.step <- 1

      for (x.plyg in polygon.x) {
        polygon(
          x = c(x.plyg,x.plyg,x.plyg + polygon.step, x.plyg + polygon.step),
          y = c(
            par()$usr[3],
            ylim[1] - (ylim[1] - par()$usr[3]) / 2,
            ylim[1] - (ylim[1] - par()$usr[3]) / 2,
            par()$usr[3]
          ),
          col = "grey",
          border = "grey")
      }
    }

    .plot_elements(main, shift.lines, given.dose, error.range)

    ## plot data and error
    for(i in 1:length(values)) {
      ## add summary content
      .add_summary(summary.pos, summary.pos_is_bottom, summary.adj,
                   label.text, mtext, i, cex, values, col)
    }
  }

  ## optionally add legend content
  if (!is.null(legend)) {
    coords <- .get_keyword_coordinates(legend.pos, xlim, ylim)
    legend.pos <- coords$pos
    legend.adj <- coords$adj
    legend(x = legend.pos[1],
           y = legend.pos[2],
           xjust = legend.adj[1],
           yjust = legend.adj[2],
           legend = legend,
           col = unique(col),
           pch = unique(pch),
           lty = 1,
           cex = 0.8)
  }

  ## optionally add subheader text
  mtext(side = 3,
        text = mtext,
        cex = 0.8 * cex)

  ##FUN by R Luminescence Team
  if (fun) sTeve() # nocov
}

.plot_elements <- function(main, shift.lines, given.dose, error.range) {
  ## add title
  title(main = main, line = shift.lines + 0.5)

  ## add additional lines
  if (!is.null(given.dose)) {
    abline(h = 1)

    if (error.range > 0) {
      error.value <- error.range
      error.range <- c(1 - error.value / 100, 1 + error.value / 100)

      ## error range lines and labels
      abline(h = error.range, lty = 2)
      text(par()$usr[2], error.range + c(-0.02, 0.02),
           paste0(c("-", "+"), error.value , "%"), pos = 2, cex = 0.8)
    }
  }
}

.add_summary <- function(summary.pos, summary.pos_is_bottom, summary.adj,
                         label.text, mtext, i, cex, values, col) {
  oneinput <- length(values) == 1
  multicol <- oneinput && nrow(values[[1]] == length(col))
  col <- if (multicol) "black" else col[i]
  if (summary.pos[1] != "sub") {
    vadj <- 0
    if (summary.pos_is_bottom) {
      ## adjust the vertical coordinate by the height of the longest label
      vadj <- graphics::strheight(tail(label.text, 1), cex = 0.8)
    }
    text(x = summary.pos[1],
         y = summary.pos[2] + vadj,
         adj = summary.adj,
         labels = label.text[[i]],
         cex = 0.8,
         col = col)
  } else {
    if (mtext == "") {
      mtext(side = 3,
            line = - i + 1,
            text = label.text[[i]],
            cex = cex * 0.8,
            col = col)
    }
  }
}

## load data
set.seed(1)
data(ExampleData.DeValues, envir = environment())
df <- ExampleData.DeValues$BT998[7:11,]
df.list <- list(df, df * c(runif(5, 0.9, 1.1), 1))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_DRTResults("error"),
               "'object' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(plot_DRTResults(list("error")),
               "'object' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(plot_DRTResults(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")
  expect_error(plot_DRTResults(data.frame(1:5, NA), na.rm = TRUE),
               "No valid data remains after removing NA values")
  expect_error(plot_DRTResults(df, preheat = c(200, 240, 240)),
               "'preheat' should have length equal to the number of De values")
  expect_error(plot_DRTResults(df, given.dose = "error"),
               "'given.dose' should be of class 'numeric', 'integer' or NULL")
  expect_error(plot_DRTResults(df, given.dose = numeric(0)),
               "'given.dose' cannot be an empty numeric")
  expect_error(plot_DRTResults(df, given.dose = NA_real_),
               "'given.dose' cannot contain NA values")
  expect_error(plot_DRTResults(df, given.dose = c(2800, 3000)),
               "'given.dose' should have length equal to the number of input")
  expect_error(plot_DRTResults(df, preheat = iris),
               "'preheat' should be of class 'numeric', 'integer' or NULL")
  expect_warning(plot_DRTResults(df, boxplot = TRUE),
                 "'boxplot' requires a value in 'preheat', reset to FALSE")
  expect_error(plot_DRTResults(df, mtext = c("a", "b")),
               "'mtext' should be of class 'character' and have length 1")
  expect_error(plot_DRTResults(df, summary = 5),
               "'summary' should be of class 'character'")
  expect_error(plot_DRTResults(df, summary.pos = 5),
               "'summary.pos' should have length 2")
  expect_error(plot_DRTResults(df, summary.pos = list()),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_DRTResults(df, summary.pos = "error"),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_DRTResults(df, par.local = NA),
               "'par.local' should be a single logical value")
  expect_error(plot_DRTResults(df, na.rm = 2),
               "'na.rm' should be a single logical value")

  empty <- set_RLum("RLum.Results")
  expect_error(plot_DRTResults(empty),
               "'object' cannot be an empty RLum.Results")
  expect_error(plot_DRTResults(list()),
               "'object' cannot be an empty list")
  expect_error(plot_DRTResults(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(plot_DRTResults(data.frame(NA, 1:4)),
                     "No valid records in 'object'")
  expect_error(plot_DRTResults(list(empty, empty)),
                     "No valid records in 'object'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_silent(plot_DRTResults(df))
  expect_silent(plot_DRTResults(df, preheat = c(200, 200, 200, 240, 240)))
  expect_silent(plot_DRTResults(df, preheat = c(200, 200, 200, 240, 240),
                                boxplot = FALSE, given.dose = 2800))
  expect_silent(plot_DRTResults(df, preheat = c(200, 200, 200, 240, 240),
                                boxplot = TRUE, given.dose = 2800,
                                summary = "mean", summary.pos = "sub"))
  expect_silent(plot_DRTResults(df.list, given.dose = c(2800, 2900)))
  expect_silent(plot_DRTResults(df.list, preheat = c(200, 200, 200, 240, 240),
                                boxplot = TRUE, summary.pos = "bottom"))

  ## more coverage
  expect_silent(plot_DRTResults(df[1, ], preheat = 200, boxplot = TRUE))
  expect_silent(plot_DRTResults(df[1, ], preheat = 200, boxplot = TRUE,
                                given.dose = 0,
                                summary.pos = "sub", col = 3))
  expect_silent(plot_DRTResults(df[1, ], boxplot = FALSE,
                                summary.pos = "sub", pch = 12, col = 1))
  expect_silent(plot_DRTResults(df, given.dose = 2800,
                                main = "Title", mtext = "Example data",
                                xlim = c(0, 6), ylim = c(0.8, 1.2),
                                xlab = "x", ylab = "y",
                                summary = "mean", summary.pos = c(0, 1.2),
                                legend = "legend", legend.pos = c(5, 1.2),
                                col = 2, cex = 1, pch = 2))
  expect_silent(plot_DRTResults(df, summary = "n", summary.pos = "sub"))
  expect_silent(plot_DRTResults(df, summary.pos = "bottomright",
                                legend.pos = "top"))
  expect_silent(plot_DRTResults(df, preheat = 1:5, na.rm = TRUE))

  ## RLum.Results object
  expect_silent(plot_DRTResults(calc_CommonDose(df, plot = FALSE,
                                                verbose = FALSE)))

  ## missing values
  df.na <- df
  df.na[2, 1] <- NA
  expect_silent(plot_DRTResults(df.na, preheat = c(200, 200, 200, 240, 240)))

  ## deprecated argument
  expect_warning(plot_DRTResults(values = df, preheat = c(200, 200, 200, 240, 240)))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("defaults",
                              plot_DRTResults(df))
  vdiffr::expect_doppelganger("cex col",
                              plot_DRTResults(df, cex = 2, col = 1:5))
  vdiffr::expect_doppelganger("summary sub",
                              plot_DRTResults(df, summary.pos = "sub",
                                              summary = c("n", "se.rel", "median",
                                                          "weighted$median")))
  vdiffr::expect_doppelganger("summary left",
                              plot_DRTResults(df, summary.pos = "left",
                                              summary = c("mean", "sd.abs")))
  vdiffr::expect_doppelganger("list bottomright",
                              plot_DRTResults(df.list,
                                              summary.pos = "bottomright",
                                              summary = c("n", "mean", "sd.abs")))
  vdiffr::expect_doppelganger("list size 2",
                              plot_DRTResults(list(df, df[1:2, ] * 1.1),
                                              summary = c("mean", "sd.rel")))
  vdiffr::expect_doppelganger("list ticks",
                              plot_DRTResults(list(df, rbind(df, df) * 1.05)))
  vdiffr::expect_doppelganger("errors NA",
                              plot_DRTResults(data.frame(1:5, NA)))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 769
  expect_silent(plot_DRTResults(list(df, df * 2), preheat = rep(200, 5),
                                boxplot = TRUE))

  ## issue 1184
  expect_silent(plot_DRTResults(list(df, df), given.dose = c(10, 10000)))
  expect_silent(plot_DRTResults(list(df, df), given.dose = c(10, 10000),
                                preheat = rep(200, 5)))

  ## issue 1522
  expect_silent(plot_DRTResults(list(data.frame(5, 1), data.frame(2, 3))))
  expect_silent(plot_DRTResults(list(data.frame(de = 5, err = 1),
                                     data.frame(de = 2, err = 3, other = 4))))
})

#' @title Fit Isothermal Holding Data
#'
#' @description ##TODO
#'
#' @details ##TODO
#'
#' @param data [character] or [data.frame] (**required**): file path or data
#' frame with 5 columns named "SAMPLE", "TEMP", "TIME", "LxTx", "LxTx_ERROR".
#'
#' @param rhop [numeric] or [Luminescence::RLum.Results-class] (*with default*):
#' a vector of rho prime values (one for each sample) or an [Luminescence::RLum.Results-class]
#' object produced by [Luminescence::analyse_FadingMeasurement].
#'
#' @param ITL_model [character] (*with default*): model to be fitted, either
#' `"GOK"` or `"BTS"`.
#'
#' @param normalise_LxTx [logical] (*with default*):
#' whether the `LxTx` data should be normalised to its maximum value (`TRUE`
#' by default).
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @param txtProgressBar [logical] (*with default*): enable/disable the
#' progress bar. Ignored if `verbose = FALSE`.
#'
#' @param trace [logical] (*with default*): enable/disable tracing during
#' the nls fitting ([minpack.lm::nlsLM]).
#'
#' @param ... further arguments and graphical parameters that will be passed
#' to the `plot` function.
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Svenja Riedesel, DTU Risø (Denmark)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @keywords datagen
#'
#' @return
#' The function returns an [Luminescence::RLum.Results-class] object and an *optional* plot.
#' The object returned contains the following elements:
#'
#' \item{fit}{[list] with the fitted models}
#' \item{coefs}{[data.frame] containing the fitted coefficients for the models}
#' \item{data}{[data.frame] containing the data used in the fitting process}
#'
#' It may return `NULL` if no model could be fitted.
#'
#' @seealso [Luminescence::analyse_ThermochronometryData], [Luminescence::analyse_FadingMeasurement]
#'
#' @references
#' Li, B., Li, S.H., The effect of band-tail states on the
#' thermal stability of the infrared stimulated luminescence
#' from K-feldspar, Journal of Luminescence 136 (2013) 5–10.
#' doi: 10.1016/j.jlumin.2012.08.043
#'
#' @examples
#' # example code ##TODO
#'
#' @noRd
fit_IsothermalHolding <- function(
    data,
    rhop,
    ITL_model = c("GOK", "BTS"),
    normalise_LxTx = TRUE,
    plot = TRUE,
    verbose = TRUE,
    txtProgressBar = TRUE,
    trace = FALSE,
    ...
) {
  .set_function_name("fit_IsothermalHolding")
  on.exit(.unset_function_name(), add = TRUE)

  ## TODOs
  ## - other functions for fitting need to be implemented
  ## - uncertainties are not yet considered for the fitting, because they are not
  ##   part of the input data.
  ## - documentation needs to be completed
  ## - so far non confidence intervals on the fit ... discussion needed
  ## - the rhop value has uncertainties, which are not yet considered

  .validate_class(data, c("character", "RLum.Results", "data.frame"))
  .validate_class(rhop, c("numeric", "RLum.Results"))
  ITL_model <- .validate_args(ITL_model, c("GOK", "BTS"))
  .validate_logical_scalar(normalise_LxTx)
  .validate_logical_scalar(plot)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(txtProgressBar)

  if (inherits(data[1], "character")) {
    records_ITL <- .import_ThermochronometryData(file = data, output_type = "RLum.Results")@data$ITL

  } else if (inherits(data, "RLum.Results")) {
    .validate_originator(data, ".import_ThermochronometryData")
    records_ITL <- data@data$ITL

  } else if (inherits(data, "data.frame")) {
    records_ITL <- data
  }

  if (!all(colnames(records_ITL) %in%
           c("SAMPLE", "TEMP", "TIME", "LxTx", "LxTx_ERROR"))) {
    .throw_error("'data' has the wrong column headers, please check the manual")
  }

  ## never show the progress bar if not verbose
  if (!verbose) {
    txtProgressBar <- FALSE
  }

  ###### --- Extract data from RLum.Results for ITL fitting --- #####
  ## get unique sample names; we will use this to filter the data
  sample_id <- unique(records_ITL[["SAMPLE"]])

  ## data normalisation
  if (normalise_LxTx) {
    records_ITL$LxTx <- .normalise_curve(records_ITL$LxTx, "max")
  }

  ## extract data.frames for each sample with all information
  df_raw_list <- lapply(sample_id, function(x) records_ITL[records_ITL$SAMPLE == x, ])

  ## allow to control how many random values for the s parameter should be
  ## generated when fitting the BTS model
  extraArgs <- list(...)
  num_s_values_bts <- if (ITL_model == "BTS")
                        .validate_positive_scalar(extraArgs$num_s_values_bts,
                                                  int = TRUE, null.ok = TRUE,
                                                  name = "'num_s_values_bts'") %||% 1000
                      else 1

  ## initialise the progress bar
  if (txtProgressBar) {
    num.models <- sum(sapply(df_raw_list, function(x) length(unique(x$TEMP))))
    pb <- txtProgressBar(min = 0, max = num.models * num_s_values_bts,
                         char = "=", style = 3)
  }

  ###### --- Perform ITL fitting --- #####
  # Define variables --------------------------------------------------------
  kB <- .const$kB  # Boltzmann constant (eV/K)

  ## get the rhop value from the fading measurement analysis if available
  if (inherits(rhop, "RLum.Results") && .check_originator(rhop, "analyse_FadingMeasurement"))
    rhop <- rhop@data$rho_prime[[1]]

  ## Define formulas to fit -------------------------------------------------
  ##
  ## We define each model as a function that describes the right-hand side
  ## of the formula. This allows us to use the `$value` term in the BTS model
  ## to extract the solution of the integral, which would otherwise be
  ## incorrectly interpreted by nlsLM().

  f_GOK <- function(A, b, Et, s10, isoT, x) {
    T_K <- isoT + .const$C2K
    A * exp(-rhop * log(1.8 * 3e15 * (250 + x))^3) *
      (1 - (1 - b) * 10^s10 * exp(-Et / (kB * T_K)) * x)^(1 / (1 - b))
  }
  f_BTS <- function(A, Eu, Et, s10, isoT, x) {
    T_K <- isoT + .const$C2K
    ## call C++ calculation part; which is a lot faster than
    ## repeated calls to stats::integrate
    ## an even faster implementation would use pure C++ for everything,
    ## however, then we would use minpack.lm::nls.lm() instead ... perhaps
    ## in the future
    f_BTS_cpp_part(
      x, A, Eu, s10, Et, kB, T_K, Et, rhop)
  }

  ## switch the models
  start <- switch(
    ITL_model,
    # https://github.com/GeorginaKing/OSLThermo/blob/5f32762e2cea87aac26d65aeabe9a434b2ce19a7/Stage2a_Fitparameters.m#L311
    GOK = list(A = 1, b = 4, Et = 1.4, s10 = 9),
    # https://github.com/GeorginaKing/OSLThermo/blob/5f32762e2cea87aac26d65aeabe9a434b2ce19a7/Stage2a_Fitparameters.m#L297
    BTS = list(A = 1, Eu = 0.1, Et = 1.4))

  lower <- switch(
    ITL_model,
    GOK = c(0, 0, 0, 0),
    BTS = c(0, 0, 1))

  upper <- switch(
    ITL_model,
    GOK = c(20, Inf, 3, 20),
    BTS = c(20, 0.5, 3))

  ## Fitting ----------------------------------------------------------------
  ## we have a double loop situation: we have a list with n samples, and
  ## each sample has n temperature steps
  num.fits <- 0
  fitted.coefs <- NULL
  fit_list <- lapply(df_raw_list, function(s) {

    ## extract temperatures
    isoT <- unique(s$TEMP)

    ## run the fitting at each temperature
    tmp <- lapply(isoT, function(isoT) {

      ## extract data to fit
      tmp_fitdata <- s[s$TEMP == isoT,]
      df <- data.frame(x = tmp_fitdata$TIME,
                       y = tmp_fitdata$LxTx)

      if (ITL_model == "GOK") {
        fit <- try({
          minpack.lm::nlsLM(
              formula = y ~ f_GOK(A, b, Et, s10, isoT, x),
              data = df,
              start = start,
              lower = lower,
              upper = upper,
              control = list(
                  maxiter = 500
              ),
              trace = trace)
        }, silent = TRUE)

        coefs <- if (!inherits(fit, "try-error")) {
                   coef(fit)
                 } else {
                   stats::setNames(rep(NA_real_, length(start)), names(start))
                 }
        fitted.coefs <<- rbind(fitted.coefs,
                               data.frame(SAMPLE = unique(s$SAMPLE),
                                          TEMP = isoT,
                                          t(coefs)))

        ## update the progress bar
        num.fits <<- num.fits + 1  # uses <<- as we are in a nested lapply()
        if (txtProgressBar) {
          setTxtProgressBar(pb, num.fits)
        }

        if (inherits(fit, "try-error"))
          fit <- NA

      } else if (ITL_model == "BTS") {
        ## run fitting with different start parameters for s10
        ## all.s10 <- rnorm(num_s_values_bts, mean = 9, sd = 1.5)
        all.s10 <- stats::qnorm(seq(0.01, 0.99, length.out = num_s_values_bts),
                                mean = 9, sd = 1.5)
        fit <- lapply(1:length(all.s10), function(idx) {
          s10 <- all.s10[idx]
          t <- try(minpack.lm::nlsLM(
                       formula = y ~ f_BTS(A, Eu, Et, s10, isoT, x),
                       data = df,
                       start = start,
                       lower = lower,
                       upper = upper,
                       control = list(
                           maxiter = 500
                       ), trace = FALSE),
                   silent = TRUE)

          ## update the progress bar
          num.fits <<- num.fits + 1  # uses <<- as we are in a nested lapply()
          if (txtProgressBar) {
            setTxtProgressBar(pb, num.fits)
          }

          if (inherits(t, "try-error"))
            return(NULL)
          return(t)
        })

        ## pick the one with the best fit after removing those that didn't fit
        fit <- .rm_NULL_elements(fit)
        if (length(fit) == 0)
          return(NA)

        fit <- fit[[which.min(vapply(fit, stats::deviance, numeric(1)))]]
        s10 <- environment(fit$m$predict)$env$s10
        fitted.coefs <<- rbind(fitted.coefs,
                               data.frame(SAMPLE = unique(s$SAMPLE),
                                          TEMP = isoT,
                                          t(coef(fit)), s10))
      }

      ## return fit
      return(fit)
    })

    ## add temperature as name to list
    names(tmp) <- isoT
    return(tmp)
  })

  ## add sample names
  names(fit_list) <- sample_id

  if (txtProgressBar) {
    close(pb)
  }

  if (is.null(fitted.coefs)) {
    .throw_warning("No ITL model could be fitted")
    return(NULL)
  }

  ## summarize the parameters of interest (Et, s10) in terms of mean, median and
  ## interquartile range
  ITL_params <- tapply(fitted.coefs, sample_id, function(coefs) {
    data.frame(t(c(Et = c(mean = mean(coefs$Et, na.rm = TRUE),
                          quantile(coefs$Et, c(0.5, 0.25, 0.75), na.rm = TRUE)),
                   s10 = c(mean = mean(coefs$s10, na.rm = TRUE),
                           quantile(coefs$s10, c(0.5, 0.25, 0.75), na.rm = TRUE)))))
  })

  ## sort the ITL_params list, as tapply() may have put the results in a
  ## different order when converting internally the sample names to a factor
  ITL_params <- ITL_params[sample_id]

  ## convert to a data table
  ITL_params <- rbindlist(ITL_params[sample_id], idcol = "SAMPLE")
  colnames(ITL_params)[-1] <- c("Et_mean", "Et_median", "Et_Q_0.25", "Et_Q_0.75",
                                "s10_mean", "s10_median", "s10_Q_0.25", "s10_Q_0.75")

  if (verbose) {
    ## silence notes raised by R CMD check
    Et <- Et_Q_0.25 <- Et_Q_0.75 <- Et_median <- Et_mean <- NULL
    s10 <- s10_Q_0.25 <- s10_Q_0.75 <- s10_median <- s10_mean <- NULL

    ## report as mean; median (IQR)
    format.out <- function(x, x1, x2, x3, n = 3)
      sprintf("%.*f;   %.*f (%.*f, %.*f)", n, x, n, x1, n, x2, n, x3)
    ITL_params[, Et := format.out(Et_mean, Et_median, Et_Q_0.25, Et_Q_0.75)]
    ITL_params[, s10 := format.out(s10_mean, s10_median, s10_Q_0.25, s10_Q_0.75)]

    fmt <- "%20s | %30s | %32s |\n"
    cat("\n---- Isothermal holding parameters [mean; median (IQR)] ----\n\n")
    cat(sprintf(fmt, "SAMPLE", "Et", "log10(s)"))
    for (i in seq(nrow(ITL_params))) {
      row <- ITL_params[i, ]
      cat(sprintf(fmt, row$SAMPLE, row$Et, row$s10))
    }
    cat("\n")

    ITL_params[, c("Et", "s10") := NULL]
  }

  ## Plotting ---------------------------------------------------------------
  if (plot) {

    ## define plot settings
    plot_settings <- modifyList(
      x = list(
        xlim = range(lapply(df_raw_list, function(x) range(x$TIME))) * c(0.5, 10),
        ylim = c(0, 1.03),
        log = "x",
        xlab = "Isothermal holding time [s]",
        ylab = expression(paste("Norm. lumin. [", L[x]/T[x], "]")),
        pch = 21,
        col = grDevices::palette.colors(),
        col.border = "black",
        main = sample_id,
        mtext = paste("Fitted with the", ITL_model, "model"),
        cex = 1.0,
        mfrow = if (length(unique(sample_id)) > 1) c(min(c(2, ceiling(length(sample_id)/2))),2) else NULL,
        legend = TRUE,
        legend.pos = "bottomleft",
        legend.cex = 0.6
      ),
      val = list(...)
    )

    ## get x vector for the prediction later; we use the same
    ## value for all provided data
    x <- c(1:1e+03,seq(1e+03,plot_settings$xlim[2], length.out = 10000))

    ## par settings
    ## the check for mfrow not being null prevents mfrow from being restored on
    ## exit when this function is called from analyse_ThermochronometryData(),
    ## because restoring mfrow (even if to the same value) causes the graphical
    ## device to start a new page
    if (!is.null(plot_settings$mfrow)) {
      par.default <- .par_defaults()
      on.exit(par(par.default), add = TRUE)
      par(cex = plot_settings$cex, mfrow = plot_settings$mfrow)
    }

    ## now we have to loop over the samples
    for (i in seq_along(sample_id)) {
      ## open plot area
      plot(NA,NA,
        xlim = plot_settings$xlim,
        ylim = plot_settings$ylim,
        log = plot_settings$log,
        xlab = plot_settings$xlab,
        ylab = plot_settings$ylab,
        yaxs = "i",
        main = rep_len(plot_settings$main, length(sample_id))[i])

      ## add plot subtitle
      mtext(side = 3, plot_settings$mtext, cex = 0.7 * plot_settings$cex)

      ## add fitted curves
      isoT <- unique(df_raw_list[[i]][["TEMP"]])
      for (c in seq_along(isoT)) {
        ## only plot the fitted lines if the fit had worked
        if (inherits(fit_list[[i]][[c]], "nls")) {
          y <- predict(fit_list[[i]][[c]], newdata = data.frame(x = x), interval = "confidence", se.fit = TRUE)
          lines(
            x = x,
            y = y,
            col = rep_len(plot_settings$col[c], length(isoT)))
        }

        ## plot the points (don't use matplot because this would assume
        ## the same length for the time vector)
        df_pts <- df_raw_list[[i]][df_raw_list[[i]][["TEMP"]] == isoT[c], ]
        points(
          x = df_pts[["TIME"]],
          y = df_pts[["LxTx"]],
          xpd = TRUE,
          pch = plot_settings$pch,
          bg = plot_settings$col[c],
          col = plot_settings$col.border)

        ## add error bars (if NA nothing is plotted by R default)
        segments(
          x0 = df_pts[["TIME"]],
          x1 = df_pts[["TIME"]],
          y0 = df_pts[["LxTx"]] - df_pts[["LxTx_ERROR"]],
          y1 = df_pts[["LxTx"]] + df_pts[["LxTx_ERROR"]],
          col = plot_settings$col[c])
      }

      ## add legend
      if (plot_settings$legend[1]) {
       legend(
         plot_settings$legend.pos,
         legend = paste0(isoT, "\u00b0C"),
         bty = "n",
         pch = 20,
         lty = 1,
         col = plot_settings$col,
         cex = plot_settings$legend.cex)
      }
    }## plot loop

  }## plot condition

  # Return results ----------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
     ITL_params = data.frame(ITL_params),
     fit = fit_list,
     coefs = fitted.coefs,
     data = records_ITL),
    info = list(
      call = sys.call())
    )
}

## load data
input.csv <- file.path(test_path("_data"),
                       paste0("CLBR_IR", c(50, 100, 150, 225), ".csv"))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fit_IsothermalHolding(list()),
               "'data' should be of class 'character', 'RLum.Results' or")
  expect_error(fit_IsothermalHolding("error", rhop = 1e-7),
               "File 'error' does not exist")
  expect_error(fit_IsothermalHolding(set_RLum("RLum.Results", data = list(1)),
                                     rhop = 1e-7),
               "'data' has an unsupported originator")
  expect_error(fit_IsothermalHolding(iris, rhop = 1e-7),
               "'data' has the wrong column headers")
  expect_error(fit_IsothermalHolding(input.csv[1], rhop = 1e-4, ITL_model = "error"),
               "'ITL_model' should be one of 'GOK' or 'BTS'")
  expect_error(fit_IsothermalHolding(input.csv[1], rhop = "error"),
               "'rhop' should be of class 'numeric' or 'RLum.Results'")
  expect_error(fit_IsothermalHolding(input.csv[1], rhop = 1e-7, plot = "error"),
               "'plot' should be a single logical value")
  expect_error(fit_IsothermalHolding(input.csv[1], rhop = 1e-7, verbose = "error"),
               "'verbose' should be a single logical value")
  expect_error(fit_IsothermalHolding(test_path("_data/CLBR.xlsx"), rhop = 1e-7),
               "File extension 'xlsx' is not supported, only 'csv' is valid")
  expect_error(fit_IsothermalHolding(input.csv[1], rhop = 1e-7, ITL_model = "BTS",
                                     num_s_values_bts = 0),
               "'num_s_values_bts' should be a single positive integer value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  expect_s4_class(fit_IsothermalHolding(input.csv, rhop = 1e-7),
                  "RLum.Results")
  expect_s4_class(fit_IsothermalHolding(input.csv[2], rhop = 1e-5,
                                        ITL_model = "BTS",
                                        num_s_values_bts = 2),
                  "RLum.Results")

  data <- .import_ThermochronometryData(input.csv[1])
  expect_s4_class(fit_IsothermalHolding(data, rhop = 1e-7, mfrow = c(2, 2)),
                  "RLum.Results")

  expect_warning(expect_null(fit_IsothermalHolding(input.csv[1], rhop = 1e-1,
                                                   ITL_model = "BTS",
                                                   num_s_values_bts = 2)),
                 "No ITL model could be fitted")
  })
})

test_that("regression tests", {

  ## issue 652
  df <- data.frame(SAMPLE = rep("S1", 6),
                   TEMP = rep(150, 6),
                   TIME = c(1, 10, 30, 100, 300, 1000),
                   LxTx = c(5.323892, 5.235491, 5.186621, 4.893767, 4.272581, 3.064905),
                   LxTx_ERROR = rep(0.05, 6))

  res <- fit_IsothermalHolding(df, rhop = 1e-7, verbose = FALSE)
  expect_equal(res@data$coefs,
               data.frame(SAMPLE = "S1", TEMP = 150, A = NA_real_,
                          b = NA_real_, Et = NA_real_, s10 = NA_real_))

  ## issue 1003
  set.seed(5)
  expect_s4_class(fit_IsothermalHolding(input.csv[2], rhop = 1e-5,
                                        ITL_model = "BTS", verbose = FALSE,
                                        num_s_values_bts = 2),
                  "RLum.Results")
})

#' @title Import RF-files to R
#'
#' @description Import files produced by the IR-RF 'ImageJ' macro (`SR-RF.ijm`;
#' Mittelstraß and Kreutzer, 2021) into R and create a list of
#' [Luminescence::RLum.Analysis-class] objects
#'
#' @details The results of spatially resolved IR-RF data are summarised in
#' so-called RF-files (Mittelstraß and Kreutzer, 2021).
#' This functions provides an easy import to process the data seamlessly with the R package 'Luminescence'.
#' The output of the function can be passed to function [Luminescence::analyse_IRSAR.RF].
#'
#' @param file [character] (**required**):
#' name of one or multiple RF files (URLs are supported); it can be the path
#' to a directory, in which case the function tries to detect and import all
#' RF files found in the directory.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... not used, only for compatible reasons
#'
#' @return
#' Returns an S4 [Luminescence::RLum.Analysis-class] object containing
#' [Luminescence::RLum.Data.Curve-class] objects for each curve. Results are
#' returned as a list when multiple files are processed or `file` is a list.
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::analyse_IRSAR.RF]
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @section Function version: 0.1.2
#'
#' @keywords IO
#'
#' @references Mittelstraß, D., Kreutzer, S., 2021. Spatially resolved infrared radiofluorescence:
#' single-grain K-feldspar dating using CCD imaging. Geochronology 3, 299–319. \doi{10.5194/gchron-3-299-2021}
#'
#' @examples
#'
#' ##Import
#' file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
#' temp <- read_RF2R(file)
#'
#' @export
read_RF2R <- function(
  file,
  verbose = TRUE,
  ...
) {
  .set_function_name("read_RF2R")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_logical_scalar(verbose)
  file <- .validate_file(file, verbose = verbose)

  ## Self-call --------------------------------------------------------------
  if (inherits(file, "list")) {
    results_list <- lapply(file, function(f){
      temp <- try(read_RF2R(file = f, verbose = verbose), silent = TRUE)

      ##check whether it worked
      if(inherits(temp, "try-error")){
        .throw_message("Import for file '", f, "' failed, NULL returned")
        return(NULL)
      }else{
        return(temp)
      }
    })

    return(unlist(results_list, recursive = FALSE))
  }

  ##read first line to ensure the format
  vers_str <-  readLines(file, 1)
  version_supported <- c("17-10-2018", "27-11-2018", "0.1.0")
  version_found <- regmatches(vers_str,
                              regexpr("(?<=macro\\_version=)[0-9-.]+", vers_str, perl = TRUE))

  if (!any(version_found %in% version_supported))
    .throw_error("File format not supported")

  ## Import -----------------------------------------------------------------

  if (verbose) {
    cat("\n[read_RF2R()] Importing ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file)))
    cat("\n")
  }

  ##import the entire file
  temp <- readLines(file, warn = FALSE)

# Extract information -------------------------------------------------------------------------

    ##extract header (here as function; that might be useful in future)
    .extract_header <- function(x){
      x <- gsub(pattern = "<", replacement = "", fixed = TRUE, x = x)
      x <- gsub(pattern = ">", replacement = "", fixed = TRUE, x = x)
      header <- strsplit(x = x, split = " ", fixed = TRUE)[[1]]
      header <- unlist(strsplit(x = header, split = "=", fixed = TRUE))

      header_names <- header[seq(1, length(header), by = 2)]
      header <-  as.list(header[seq(2, length(header), by = 2)])
      names(header) <- header_names
      return(header)
    }

    header <- try(.extract_header(temp[1]), silent = TRUE)

    ##test the header
    if(inherits(header, 'try-error')){
      .throw_message("Header extraction failed, trying to continue without ...")
      header <- NA
    }

    ##extract tag boundaries framed by tags +++++++++++++++++++
    ##the 2nd line corrects the inner boundaries
    ##(1) statistics
    id_statistics <- grep(pattern = "Statistics>", x = temp, fixed = TRUE)
    id_statistics <- c(id_statistics[1] + 1, id_statistics[2] - 1)

    ##(2) Natural (henceforth: RF_nat)
    id_RF_nat <- grep(pattern = "Natural>", x = temp, fixed = TRUE)
    id_RF_nat <- c(id_RF_nat[1] + 1, id_RF_nat[2] - 1)

    ##(3) Bleached (henceforth: RF_reg)
    id_RF_reg <- grep(pattern = "Bleached>", x = temp, fixed = TRUE)
    id_RF_reg <- c(id_RF_reg[1] + 1, id_RF_reg[2] - 1)

    ##extract content within the tags +++++++++++++++++++
    ##(1) statistics
    ##
    ####header
      statistics_header <- strsplit(x = temp[id_statistics[1]], split = "\t", fixed = TRUE)[[1]][-1]

      ##data
      m_statistics <- as.data.frame(lapply((id_statistics[1]+1):(id_statistics[2]), function(x){
          strsplit(x = temp[x], split = "\t", fixed = TRUE)[[1]]

      }), stringsAsFactors = FALSE)

      ##extract colnames
      colnames(m_statistics) <-
        unlist(strsplit(
          x = as.character(m_statistics[1, ]),
          split = ":",
          fixed = TRUE
        ))

      ##remove first
      df_statistics <-
        cbind(ROI = statistics_header, m_statistics[-1, ], stringsAsFactors = FALSE)


    ##(2) RF_nat
    ##
    ####header
      RF_nat_header <-
        strsplit(x = temp[id_RF_nat[1]], split = "\t", fixed = TRUE)[[1]]

      ##data
      m_RF_nat <- matrix(
        data = as.numeric(strsplit(
          x = paste(temp[(id_RF_nat[1] + 1):(id_RF_nat[2])], collapse = "\t"),
          split = "\t",
          fixed = TRUE
        )[[1]]),
        ncol = length(RF_nat_header),
        byrow = TRUE
      )

      ##set colnames
      colnames(m_RF_nat) <- RF_nat_header

    ##(3) RF_reg
    ##
    ####header
      RF_reg_header <-
        strsplit(x = temp[id_RF_reg[1]], split = "\t", fixed = TRUE)[[1]]

      ##data
      m_RF_reg <- matrix(
        data = as.numeric(strsplit(
          x = paste(temp[(id_RF_reg[1] + 1):(id_RF_reg[2])], collapse = "\t"),
          split = "\t",
          fixed = TRUE
        )[[1]]),
        ncol = length(RF_reg_header),
        byrow = TRUE
      )

      ##set colnames
      colnames(m_RF_reg) <- RF_reg_header


# Create RLum.Analysis objects ----------------------------------------------------------------
  lapply(1:nrow(df_statistics), function(a) {

      ##set records
      records <- lapply(1:2, function(o) {
        if(o == 1){
          temp_curve <- m_RF_nat[,c(2,2 + a)]
        }else{
          temp_curve <- m_RF_reg[,c(2,2 + a)]
        }

        ##write curve
        set_RLum(
          class = "RLum.Data.Curve",
          originator = "read_RF2R",
          curveType = "measured",
          recordType = "RF (NA)",
          data = temp_curve
        )
      })

      ##create RLum.Analysis object
      set_RLum(class = "RLum.Analysis",
               originator = "read_RF2R",
               records = records,
               info = c(
                 as.list(df_statistics[a,]),
                 header
                 ))
  })
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_RF2R("error"),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_error(read_RF2R(2),
               "'file' should be of class 'character' or 'list'")
  expect_error(read_RF2R(character(0)),
               "'file' cannot be an empty character")
  expect_error(read_RF2R(list(data.frame())),
                 "All elements of 'file' should be of class 'character'")
  expect_message(expect_null(read_RF2R(list("error"))),
                 "Import for file 'error' failed, NULL returned")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##load file path
  file <- system.file("extdata", "RF_file.rf", package = "Luminescence")

  ##simple import
  SW({
  expect_type(read_RF2R(file), type = "list")
  })

  ##import list
  expect_type(expect_message(read_RF2R(list(file, "test"), verbose = FALSE),
                             "Error: Import for file 'test' failed"),
              type = "list")

  ## import multiple files is allowed
  expect_type(read_RF2R(c(file, file), verbose = FALSE),
              type = "list")

  ## create a file with unsupported version
  file.wrong <- "RF_wrong_version.Rf"
  writeLines(gsub("17-10-2018", "wrong-version", readLines(file)),
             file.wrong)
  expect_error(read_RF2R(file.wrong),
               "File format not supported")
  file.remove(file.wrong)

  ## create a file with malformed header
  file.wrong <- "RF_wrong_header.Rf"
  writeLines(gsub("grain_d=20", "grain_d=", readLines(file)),
             file.wrong)
  expect_message(read_RF2R(file.wrong, verbose = FALSE),
                 "Error: Header extraction failed")
  file.remove(file.wrong)
})

#' General set function for RLum S4 class objects
#'
#' Function calls object-specific set functions for RLum S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum-class] objects.\cr 
#' Depending on the given class, the corresponding method to create an object 
#' from this class will be selected. Allowed additional arguments can be found 
#' in the documentations of the corresponding [RLum-class] class: 
#' - [RLum.Data.Curve-class],
#' - [RLum.Data.Image-class], 
#' - [RLum.Data.Spectrum-class],
#' - [RLum.Analysis-class], 
#' - [RLum.Results-class]
#'
#' @param class [RLum-class] (**required**): 
#' name of the S4 class to create
#'
#' @param originator [character] (*automatic*): 
#' contains the name of the calling function (the function that produces this object); 
#' can be set manually.
#'
#' @param .uid [character] (*automatic*): 
#' sets an unique ID for this object using the internal C++ function `create_UID`.
#'
#' @param .pid [character] (*with default*): 
#' option to provide a parent id for nesting at will.
#'
#' @param ... further arguments that one might want to pass to the specific set method
#'
#' @return 
#' Returns an object of the specified class.
#'
#' @section Function version: 0.3.0
#'
#' @author 
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#'
#' @seealso [RLum.Data.Curve-class], [RLum.Data.Image-class], 
#' [RLum.Data.Spectrum-class], [RLum.Analysis-class], [RLum.Results-class]
#'
#' @keywords utilities
#'
#' @examples
#'
#' ##produce empty objects from each class
#' set_RLum(class = "RLum.Data.Curve")
#' set_RLum(class = "RLum.Data.Spectrum")
#' set_RLum(class = "RLum.Data.Spectrum")
#' set_RLum(class = "RLum.Analysis")
#' set_RLum(class = "RLum.Results")
#'
#' ##produce a curve object with arbitrary curve values
#' object <- set_RLum(
#' class = "RLum.Data.Curve",
#' curveType = "arbitrary",
#' recordType = "OSL",
#' data = matrix(c(1:100,exp(-c(1:100))),ncol = 2))
#'
#' ##plot this curve object
#' plot_RLum(object)
#' 
#' @md
#' @export
setGeneric("set_RLum", function (class, originator, .uid = create_UID(), .pid = NA_character_, ... ) {
  class(class) <- as.character(class)

  if(missing(originator)) {
    if (is(sys.call(which = -1)[[1]], "language")) {
      originator <- as.character(sys.call(which = -1)[[1]])

      ##account for calls using the double colons, in this case the vector is
      ##of length 3, not only 1
      if(length(originator) == 3){
        originator <- originator[3]
      }

    } else{
      originator <- NA_character_
    }
  }

  standardGeneric("set_RLum")
})

#'@title Emission Spectra Conversion from Wavelength to Energy Scales (Jacobian Conversion)
#'
#' @description
#' The function provides a convenient and fast way to convert emission spectra
#' wavelength to energy scales. The function works on [Luminescence::RLum.Data.Spectrum-class],
#' [data.frame] and [matrix] and a [list] of such objects. The function was
#' written to smooth the workflow while analysing emission spectra data. This
#' is in particular useful if you want to further treat your data and apply,
#' e.g., a signal deconvolution.
#'
#' @details
#' The intensity of the spectrum is re-calculated using the following approach
#' to recalculate wavelength and corresponding intensity values
#' (e.g., Appendix 4 in Blasse and Grabmaier, 1994; Mooney and Kambhampati, 2013):
#'
#' \deqn{\phi_{E} = \phi_{\lambda} * \lambda^2 / (hc)}
#'
#' with
#' \eqn{\phi_{E}} the intensity per interval of energy \eqn{E} (1/eV),
#' \eqn{\phi_{\lambda}} the intensity per interval of wavelength \eqn{\lambda}
#' (1/nm) and
#' \eqn{h} (eV * s) the Planck constant and \eqn{c} (nm/s) the velocity of light.
#'
#' For transforming the wavelength axis (x-values) the equation as follow is used
#'
#' \deqn{E = hc/\lambda}
#'
#' @param object [Luminescence::RLum.Data.Spectrum-class], [data.frame], [matrix] (**required**):
#' input object to be converted. If the input is not an `RLum.Data.Spectrum`
#' object, the first column is always treated as the wavelength
#' column. The function supports a list of allowed input objects.
#'
#' @param digits [integer] (*with default*):
#' number of digits on the returned energy axis.
#'
#' @param order [logical] (*with default*): enable/disable sorting of the
#' values in ascending energy order. After the conversion, the longest
#' wavelength has the lowest energy value and the shortest wavelength the
#' highest. While this is correct, some R functions expect increasing x-values.
#'
#' @return The same object class as provided as input is returned.
#'
#' @note This conversion works solely for emission spectra. In case of absorption spectra only
#' the x-axis has to be converted.
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data.Spectrum-class], [Luminescence::plot_RLum]
#'
#' @keywords IO
#'
#' @references
#'
#' Blasse, G., Grabmaier, B.C., 1994. Luminescent Materials. Springer.
#'
#' Mooney, J., Kambhampati, P., 2013. Get the Basics Right: Jacobian Conversion of Wavelength and
#' Energy Scales for Quantitative Analysis of Emission Spectra. J. Phys. Chem. Lett. 4, 3316–3318.
#' \doi{10.1021/jz401508t}
#'
#' Mooney, J., Kambhampati, P., 2013. Correction to “Get the Basics Right: Jacobian Conversion of
#' Wavelength and Energy Scales for Quantitative Analysis of Emission Spectra.” J. Phys. Chem. Lett. 4,
#' 3316–3318. \doi{10.1021/jz401508t}
#'
#' **Further reading**
#'
#' Angulo, G., Grampp, G., Rosspeintner, A., 2006. Recalling the appropriate representation of
#' electronic spectra. Spectrochimica Acta Part A: Molecular and Biomolecular Spectroscopy 65,
#' 727–731. \doi{10.1016/j.saa.2006.01.007}
#'
#' Wang, Y., Townsend, P.D., 2013. Potential problems in collection and data processing of
#' luminescence signals. Journal of Luminescence 142, 202–211. \doi{10.1016/j.jlumin.2013.03.052}
#'
#' @examples
#'
#' ##=====================##
#' ##(1) Literature example after Mooney et al. (2013)
#' ##(1.1) create matrix
#' m <- matrix(
#'   data = c(seq(400, 800, 50), rep(1, 9)), ncol = 2)
#'
#' ##(1.2) set plot function to reproduce the literature figure
#'p <- function(m) {
#'  plot(x = m[, 1], y = m[, 2])
#'  polygon(
#'  x = c(m[, 1], rev(m[, 1])),
#'  y = c(m[, 2], rep(0, nrow(m))))
#'  for (i in 1:nrow(m)) {
#'   lines(x = rep(m[i, 1], 2), y = c(0, m[i, 2]))
#'  }
#'}
#'
#'##(1.3) plot curves
#' par(mfrow = c(1,2))
#' p(m)
#' p(convert_Wavelength2Energy(m))
#'
#'##=====================##
#'##(2) Another example using density curves
#' ##create dataset
#' xy <- density(
#'  c(rnorm(n = 100, mean = 500, sd = 20),
#'  rnorm(n = 100, mean = 800, sd = 20)))
#' xy <- data.frame(xy$x, xy$y)
#'
#' ##plot
#' par(mfrow = c(1,2))
#' plot(
#'  xy,
#'  type = "l",
#'  xlim = c(150, 1000),
#'  xlab = "Wavelength [nm]",
#'  ylab = "Luminescence [a.u.]"
#' )
#'plot(
#'  convert_Wavelength2Energy(xy),
#'  xy$y,
#'  type = "l",
#'  xlim = c(1.23, 8.3),
#'  xlab = "Energy [eV]",
#'  ylab = "Luminescence [a.u.]"
#' )
#'
#'@export
convert_Wavelength2Energy <- function(
  object,
  digits = 3L,
  order = FALSE
) {
  .set_function_name("convert_Wavelength2Energy")
  on.exit(.unset_function_name(), add = TRUE)

  # Self-call -----------------------------------------------------------------------------------
  if(inherits(object, "list")){
    return(lapply(object, convert_Wavelength2Energy))
  }

  # Conversion function -------------------------------------------------------------------------

  ##this treats the matrix; in either caes and we play safe, means, we create in either case
  ##colnames and rownames, but remove them later depending on the input
  .conv_intensity <- function(m){
      h <- .const$h_eVs   # eV*s
      c <- .const$c * 1e9 # nm/s

      ##convert count values
      m[] <- m * as.numeric(rownames(m))^2 / (h * c)

      ##modify rownames
      rownames(m) <- round((h * c) / as.numeric(rownames(m)),digits)

      ##return results
      return(m)
  }

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Data.Spectrum", "data.frame", "matrix"),
                  extra = "a 'list' of such objects")
  .validate_not_empty(object)
  .validate_positive_scalar(digits, int = TRUE)
  .validate_logical_scalar(order)

  if(inherits(object, "RLum.Data.Spectrum")){
    if (length(object@data) < 2) {
      .throw_error("'object' contains no data")
    }

     ##check whether the object might have this scale already
    ##this only works on RLum.Data.Spectrum objects and is sugar for using RLum-objects
    if(any("curveDescripter" %in% names(object@info))){
     if(any(grepl(pattern = "energy", x = tolower(object@info$curveDescripter), fixed = TRUE))){
        .throw_message("'object' has already an energy scale, nothing done",
                       error = FALSE)
         return(object)
     }

      ## correct curveDescripter (we do not attach the table, otherwise the
      ## object gets too big)
      descr <- strsplit(object@info$curveDescripter, ";", fixed = TRUE)[[1]]
      descr[grepl("wavelength", descr, fixed = TRUE)] <- "energy [eV]"
      object@info$curveDescripter <- paste(descr, collapse = ";")
    }

    ##convert data
    object@data <- .conv_intensity(object@data)

    #sort values if needed
    if(order){
      object@data <- object@data[order(as.numeric(rownames(object@data))), ,
                                 drop = FALSE]
      rownames(object@data) <- sort(as.numeric(rownames(object@data)))
    }

    ##return new object
    return(object)

  } else if (inherits(object, c("matrix", "data.frame"))) {

    if (ncol(object) < 2) {
      .throw_error("'object' should have at least two columns")
    }

    temp <- as.matrix(object[, 2:ncol(object), drop = FALSE])
    if (!is.numeric(temp)) {
      .throw_error("'object' should have only numeric fields")
    }

    ##set rownames
    rownames(temp) <- object[,1]

    ##convert values
    temp <- .conv_intensity(temp)

    ##construct new full matrix
    temp <- cbind(as.numeric(rownames(temp)), temp)
    rownames(temp) <- rownames(object)
    colnames(temp) <- colnames(object)

    ##order on request (remember, here it is the first column)
    if(order) temp <- temp[order(temp[,1]),]

    ##return
    if(inherits(object, "data.frame"))
      return(as.data.frame(temp))

    return(temp)
  }
}

#' @title Import Princeton Instruments (TM) SPE-file into R
#'
#' @description
#' The function imports Princeton Instruments (TM) SPE-files into R and
#' provides [Luminescence::RLum.Data.Image-class] objects as output. The import
#' functionality is based on the file format description provided by
#' Princeton Instruments and a MatLab script written by Carl Hall (see
#' references).
#'
#' @param file [character] (**required**):
#' name of the SPE file to read (URLs are supported).
#'
#' @param output.object [character] (*with default*):
#' set the output object type. Allowed types are [Luminescence::RLum.Data.Spectrum-class],
#' [Luminescence::RLum.Data.Image-class] or `"matrix"`.
#'
#' @param frame.range [vector], [integer] (*optional*):
#' range of frames to read. For example, `frame.range = c(1, 10)` selects only
#' the first 10 frames. If not specified, all available frames (up to a maximum
#' of 100 if `output.object = "RLum.Data.Image"`) are read.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar. Ignored if `verbose = FALSE`.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... not used, for compatibility reasons only.
#'
#' @return
#' Depending on the chosen option the functions returns three different
#' type of objects:
#'
#' `output.object`
#'
#' `RLum.Data.Spectrum`
#'
#' An object of type [Luminescence::RLum.Data.Spectrum-class] is returned.  Row
#' sums are used to integrate all counts over one channel.
#'
#' `RLum.Data.Image`
#'
#' An object of type [Luminescence::RLum.Data.Image-class] is returned.  Due to
#' performance reasons the import is aborted for files containing more than 100
#' frames. This limitation can be overwritten manually by using the argument
#' `frame.range`.
#'
#' `matrix`
#'
#' Returns a matrix of the form: Rows = Channels, columns = Frames. For the
#' transformation the function [Luminescence::get_RLum] is used,
#' meaning that the same results can be obtained by using the function
#' [Luminescence::get_RLum] on an [Luminescence::RLum.Data.Spectrum-class] or
#' [Luminescence::RLum.Data.Image-class] object.
#'
#' @note
#' **The function does not test whether the input data are spectra or pictures for spatial resolved analysis!**
#'
#' The function has been successfully tested for SPE format versions 2.x.
#'
#' *Currently not all information provided by the SPE format are supported.*
#'
#' @section Function version: 0.1.6
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [readBin], [Luminescence::RLum.Data.Spectrum-class]
#'
#' @references
#' Princeton Instruments, 2014. Princeton Instruments SPE 3.0 File
#' Format Specification, Version 1.A (for document URL please use an internet search machine)
#'
#' Hall, C., 2012: readSPE.m.
#' `https://www.mathworks.com/matlabcentral/fileexchange/35940-readspe`
#'
#' @keywords IO
#'
#' @examples
#'
#' ## to run examples uncomment lines and run the code
#'
#' ##(1) Import data as RLum.Data.Spectrum object
#' #file <- file.choose()
#' #temp <- read_SPE2R(file)
#' #temp
#'
#' ##(2) Import data as RLum.Data.Image object
#' #file <- file.choose()
#' #temp <- read_SPE2R(file, output.object = "RLum.Data.Image")
#' #temp
#'
#' ##(3) Import data as matrix object
#' #file <- file.choose()
#' #temp <- read_SPE2R(file, output.object = "matrix")
#' #temp
#'
#' ##(4) Export raw data to csv, if temp is a RLum.Data.Spectrum object
#' # write.table(x = get_RLum(temp),
#' #             file = "[your path and filename]",
#' #             sep = ";", row.names = FALSE)
#'
#' @export
read_SPE2R <- function(
  file,
  output.object = c("RLum.Data.Image", "RLum.Data.Spectrum", "matrix"),
  frame.range = NULL,
  txtProgressBar = TRUE,
  verbose = TRUE,
  ...
) {
  .set_function_name("read_SPE2R")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(file, "character")
  .validate_length(file, 1)
  output.object <- .validate_args(output.object,
                                  c("RLum.Data.Image", "RLum.Data.Spectrum", "matrix"))
  .validate_class(frame.range, c("integer", "numeric"), null.ok = TRUE)
  if (anyNA(frame.range) || any(frame.range <= 0))
    .throw_error("'frame.range' should contain positive values")

  if (!verbose)
    txtProgressBar <- FALSE

  # Open Connection ---------------------------------------------------------

  file <- .validate_file(file, ext = "SPE", pattern = "\\.SPE$",
                         verbose = verbose)
  if (length(file) == 0)
    return(NULL)

  con <- file(file, "rb")

  if (verbose) {
    cat("\n[read_SPE2R()] Importing ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file)))
    cat("\n")
  }

  # read header -------------------------------------------------------------

  temp <- readBin(con, what="int", 2, size=2, endian="little", signed = TRUE)
  ControllerVersion <- temp[1] #Hardware version
  LogicOutput <- temp[2] #Definition of Output BNC

  temp <- readBin(con, what="int", 2, size=2, endian="little", signed = FALSE)
  AmpHiCapLowNoise <- temp[1] #Amp Switching Mode
  xDimDet <- temp[2] #Detector x dimension of chip.

  #timing mode
  mode <- readBin(con, what="int", 1, size=2, endian="little", signed = TRUE)

  #alternative exposure, in sec.
  exp_sec <- readBin(con, what="double", 1, size=4, endian="little")

  temp <- readBin(con, what="int", 2, size=2, endian="little", signed = TRUE)
  VChipXdim <- temp[1] # Virtual Chip X dim
  VChipYdim <- temp[2] # Virtual Chip Y dim

  #y dimension of CCD or detector.
  yDimDet <- readBin(con, what="int", 1, size=2, endian="little", signed = TRUE)

  #Date
  Date <- suppressWarnings(readChar(con, 10, useBytes=TRUE))

  ##jump
  readBin(con, what = "raw", 4, size = 1, endian = "little", signed = TRUE)

  #Old number of scans - should always be -1
  noscan <- readBin(con, what="int", 1, size=2, endian="little", signed = TRUE)

  #Detector Temperature Set
  DetTemperature <- readBin(con, what="double", 1, size=4, endian="little")

  # CCD/DiodeArray type
  DetType <- readBin(con, what="int", 1, size=2, endian="little", signed = TRUE)

  #actual # of pixels on x axis
  xdim <- readBin(con, what="int", 1, size=2, endian="little", signed = FALSE)

  ##jump
  readBin(con, what = "raw", 64, size = 1, endian = "little", signed = TRUE)

  ##experiment data type
  ##0 = 32f (4 bytes)
  ##1 = 32s (4 bytes)
  ##3 = 16u (2 bytes)
  ##8 = 32u (4 bytes)
  datatype <- readBin(con, what="int", 1, size=2, endian="little", signed = TRUE)

  ##jump
  readBin(con, what = "raw", 546, size = 1, endian = "little")

  #y dimension of raw data.
  ydim <- readBin(con, what="int", 1, size=2, endian="little", signed = FALSE)

  ##0=scrambled,1=unscrambled
  scramble <- readBin(con, what="int", 1, size=2, endian="little", signed = FALSE)

  ##jump
  readBin(con, what = "raw", 4, size = 1, endian = "little")

  #Number of scans (Early WinX)
  lnoscan <- readBin(con, what="int", 1, size=4, endian="little", signed = TRUE)

  #Number of Accumulations
  lavgexp <- readBin(con, what="int", 1, size=4, endian="little", signed = TRUE)

  ##Experiment readout time
  ReadoutTime <- readBin(con, what="double", 1, size=4, endian="little")

  #T/F Triggered Timing Option
  TriggeredModeFlag <- readBin(con, what="int", 1, size=2, endian="little", signed = TRUE)

  ##jump
  readBin(con, what = "raw", 768, size = 1, endian = "little")

  ##number of frames in file.
  NumFrames <- readBin(con, what="int", 1, size=4, endian="little", signed = TRUE)
  if (NumFrames > 100 && is.null(frame.range) &&
      output.object == "RLum.Data.Image") {
    # nocov start
    .throw_error("This file contains > 100 frames (", NumFrames, "), ",
                 "use argument 'frame.range' to force import")
    # nocov end
  }

  ## set frame.range
  frame.range <- if (is.null(frame.range))
                   c(1, NumFrames)
                 else
                   pmin(pmax(range(frame.range), 1), NumFrames)

  ##jump
  readBin(con, what = "raw", 542, size = 1, endian = "little")

  #file_header_ver
  file_header_ver <- readBin(con, what="double", 1, size=4, endian="little")

  ##jump
  readBin(con, what = "raw", 1000, size = 1, endian = "little")

  ##WinView_id - set to 19,088,743 (or 1234567 hex) (required for legacy reasons)
  readBin(con, what = "integer", 1, size = 4, endian = "little", signed = TRUE)

  ##jump
  readBin(con, what = "raw", 1098, size = 1, endian = "little")

  ##lastvalue - set to 21,845 (or 5555 hex) (required for legacy reasons)
  readBin(con, what = "integer", 1, size = 2, endian = "little", signed = TRUE)


  ##end header
  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  ##create info element list from data
  temp.info <- list(ControllerVersion,
                    LogicOutput,
                    AmpHiCapLowNoise,
                    xDimDet, yDimDet,
                    xdim, ydim,
                    VChipXdim, VChipYdim,
                    Date,
                    noscan,
                    mode,  exp_sec,
                    DetTemperature,
                    DetType,
                    datatype,
                    scramble,
                    lnoscan,
                    lavgexp,
                    ReadoutTime,
                    TriggeredModeFlag,
                    NumFrames,
                    file_header_ver)

  ##set name for list elements
  names(temp.info) <- c("ControllerVersion", "LogicOutput", "AmpHiCapLowNoise", "xDimDet", "yDimDet",
                        "xdim", "ydim", "VChipXdim", "VChipYdim", "Date", "noscan", "mode", "exp_sec",
                        "DetTemperature", "DetType", "datatype", "scramble", "lnoscan", "lavgexp",
                        "ReadoutTime", "TriggeredModeFlag", "NumFrames", "file_header_ver")

  # read count value data ---------------------------------------------------
  ##set functions

  ## define the reading function according to the datatype
  if (!datatype %in% c(0, 1, 2, 3, 8)) {
    .throw_error("Unknown 'datatype'") # nocov
  }
  what <- if (datatype == 0) "double" else "integer"
  size <- if (datatype %in% 2:3) 2 else 4
  sign <- if (datatype %in% 0:2) TRUE else FALSE
  read.data <- function(n.counts){
    readBin(con, what = what, n = n.counts, size = size, signed = sign,
            endian = "little")
  }

  ##loop over all frames
  ##output
  if(verbose)
    cat("\n[read_SPE2R()]\n\t >>", file)

  ##set progressbar
  if (txtProgressBar) {
    pb<-txtProgressBar(min=0,max=diff(frame.range)+1, char="=", style=3)
  }

  ##stepping for frame range
  temp <- readBin(con, what = "raw", (min(frame.range)-1)*2, size = 1, endian = "little")

  data.list <- NULL
  for(i in 1:(diff(frame.range)+1)){#NumFrames
    temp.data <- matrix(read.data(n.counts = (xdim * ydim)),
                        ncol = ydim,
                        nrow = xdim)
    data.list <- c(data.list, list(temp.data))

    ##update progress bar
    if (txtProgressBar) {
      setTxtProgressBar(pb, i)
    }
  }

  ##close
  if (txtProgressBar) {
    close(pb)
    cat("\t >>", i,"records have been read successfully!\n\n")
  }

  # Output ------------------------------------------------------------------

  if (output.object %in% c("RLum.Data.Spectrum", "matrix")) {
    ##to create a spectrum object the matrix has to transposed and
    ##the row sums are needed

    data.spectrum.vector <- sapply(1:length(data.list), function(x){
      rowSums(data.list[[x]])
    })

    ##split vector to matrix
    data.spectrum.matrix <- matrix(data.spectrum.vector,
                                   nrow = xdim,
                                   ncol = length(data.list))

    ##set column and row names
    colnames(data.spectrum.matrix) <- as.character(1:ncol(data.spectrum.matrix))
    rownames(data.spectrum.matrix) <- as.character(1:nrow(data.spectrum.matrix))

    ##set output object
    object <- set_RLum(
      class = "RLum.Data.Spectrum",
      originator = "read_SPE2R",
      recordType = "Spectrum (NA)",
      curveType = "measured",
      data = data.spectrum.matrix,
      info = temp.info)

    ##optional matrix object
    if (output.object == "matrix") {
      object <- get_RLum(object)
    }

  }else if(output.object == "RLum.Data.Image"){
    object <- as(data.list, "RLum.Data.Image")
    object@originator <- "read_SPE2R"
    object@recordType <- "Image (NA)"
    object@curveType <- "measured"
    object@info <- temp.info
  }

  ##close con
  close(con)

  ##return values
  return(object)
}

#' @title Convert Nuclide Activities to Abundance and Vice Versa
#'
#' @description The function performs the conversion of the specific activities into
#' mass abundance and vice versa for the radioelements U, Th, and K to
#' harmonise the measurement unit with the required data input unit of
#' potential analytical tools for, e.g. dose rate calculation or related
#' functions such as [Luminescence::use_DRAC].
#'
#' @details The conversion from nuclide activity of a sample to nuclide concentration
#' is performed using conversion factors that are based on the mass-related
#' specific activity of the respective nuclide.
#'
#' Constants used in this function were obtained from `https://physics.nist.gov/cuu/Constants/`
#' all atomic weights and composition values from
#' `https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses`
#' and the nuclide data from `https://www.iaea.org/resources/databases/livechart-of-nuclides-advanced-version`
#'
#' The factors can be calculated using the equation:
#' \deqn{
#' A = N_A \frac{N_{abund}}{N_{mol.mass}} \ln(2) / N.half.life
#' }
#'
#' to convert in µg/g we further use:
#' \deqn{
#' f = A / 10^6
#' }
#'
#' where:
#' - `N_A` - Avogadro constant in 1/mol
#' - `A` - specific activity of the nuclide in Bq/kg
#' - `N.abund` - relative natural abundance of the isotope
#' - `N.mol.mass` molar mass in kg/mol
#' - `N.half.life` half-life of the nuclide in s
#'
#' example for calculating the activity of the radionuclide U-238:
#'  * `N_A` = 6.02214076e+23 (1/mol)
#'  * `T_0.5` = 1.41e+17 (s)
#'  * `m_U_238` = 0.23802891 (kg/mol)
#'  * `U_abund` = 0.992745 (unitless)
#'
#' \deqn{A_{U} = N_{A} * U_{abund} / m_{U_238} * \ln(2) / T_{1/2} = 2347046} (Bq/kg)
#'
#' \deqn{f.U = A_{U} / 10^6}
#'
#' @param data [data.frame] (**required**):
#' provide dose rate data (activity or concentration) in three columns.
#' The first column indicates the nuclide, the 2nd column measured value and
#' in the 3rd column its error value. Allowed nuclide data are
#' `'U-238'`, `'Th-232'` and `'K-40'`. See examples for an example.
#'
#' @param input_unit [character] (*with default*):
#' specify unit of input data given in the dose rate data frame, one of
#' `"activity"` (considered as given Bq/kg, the default) and `"abundance"`
#' (considered as given in mug/g or mass. %).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @section Function version: 0.1.2
#'
#' @author Margret C. Fuchs, Helmholtz-Institute Freiberg for Resource Technology (Germany)
#'
#' @references
#' Debertin, K., Helmer, R.G., 1988. Gamma- and X-ray Spectrometry with
#' Semiconductor Detectors, Elsevier Science Publishers, p.283
#'
#' Wiechen, A., Ruehle, H., Vogl, K., 2013. Bestimmung der massebezogenen
#' Aktivitaet von Radionukliden. AEQUIVAL/MASSAKT, ISSN 1865-8725,
#' [https://www.bundesumweltministerium.de/fileadmin/Daten_BMU/Download_PDF/Strahlenschutz/aequival-massakt_v2013-07_bf.pdf]()
#'
#' @keywords IO
#'
#' @note Although written otherwise for historical reasons. Input values must be element values.
#' For instance, if a value is provided for U-238 the function assumes that this value
#' represents the sum (activity or abundance) of U-238, U-235 and U-234.
#' In other words, 1 µg/g of U means that this is the composition of 0.992 parts of U-238,
#' 0.000054 parts of U-234, and 0.00072 parts of U-235.
#'
#' @return Returns an [Luminescence::RLum.Results-class] object with a [data.frame] containing
#' input and newly calculated values. Please not that in the column header µg/g
#' is written as `mug/g` due to the R requirement to maintain packages portable using
#' ASCII characters only.
#'
#' @examples
#'
#' ##construct data.frame
#' data <- data.frame(
#'  NUCLIDES = c("U-238", "Th-232", "K-40"),
#'  VALUE = c(40,80,100),
#'  VALUE_ERROR = c(4,8,10),
#'  stringsAsFactors = FALSE)
#'
#' ##perform analysis
#' convert_Activity2Concentration(data)
#'
#' @export
convert_Activity2Concentration <- function(
  data,
  input_unit = c("activity", "abundance"),
  verbose = TRUE
) {
  .set_function_name("convert_Activity2Concentration")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(data, "data.frame")
  .validate_logical_scalar(verbose)
  .validate_class(input_unit, "character")

  if(ncol(data)<3)
    .throw_error("'data' should have at least 3 columns")

  # Set output data.frame -----------------------------------------------------------------------
  output <- data.frame(
    NUCLIDE = rep(NA, nrow(data)),
    ACTIVIY = rep(NA, nrow(data)),
    ACTIVIY_ERROR = rep(NA, nrow(data)),
    CONC = rep(NA, nrow(data)),
    CONC_ERROR = rep(NA, nrow(data)),
    stringsAsFactors = FALSE
  )

  ##set column names
  colnames(output) <- c(
    "NUCLIDE",
    "ACTIVIY (Bq/kg)",
    "ACTIVIY ERROR (Bq/kg)",
    "ABUND. (mug/g or mass. %)",
    "ABUND. ERROR (mug/g or mass. %)")

  ##set column for output
  output$NUCLIDE <- data[[1]]

  ## check input unit
  ## we silently let the old input values unflagged for back compatibility reasons
  if (!tolower(input_unit[1]) %in% c("bq/kg", "ppm/%")) {
    input_unit <- .validate_args(tolower(input_unit), c("activity", "abundance"))
  }

  # Set conversion factors ----------------------------------------------------------------------

  #############################################################################
  ### conversion factors

  mass_constant <- .const$ma  # in kg

  ## set conversion factors ... this are the expected activity per kg of the radionuclide
  ## a = log(2) / ((unified_atomic_mass * mass_constant) / abundance) * T_0.5
  convers.factor.U238 <- log(2) / (((238.0507884 * mass_constant)/ 0.992742) * 1.409963e+17) / 1e+06
  convers.factor.Th232 <- log(2) / (((232.0380558 * mass_constant)/ 1) * 4.41797e+17) / 1e+06
  convers.factor.K40 <- log(2) / (((39.963998166 * mass_constant)/ 0.000117) * 3.9383e+16) / 1e+02

  # Run conversion ------------------------------------------------------------------------------
  U <- which(data$NUCLIDE == "U-238")
  Th <- which(data$NUCLIDE == "Th-232")
  K <- which(data$NUCLIDE == "K-40")

  ##Activity to concentration
  if(input_unit == "activity" || input_unit == "bq/kg"){
    output[U,4:5] <- data[U,2:3] / convers.factor.U238
    output[Th,4:5] <- data[Th,2:3] / convers.factor.Th232
    output[K,4:5] <- data[K,2:3] / convers.factor.K40

    output[U,2:3] <- data[U,2:3]
    output[Th,2:3] <- data[Th,2:3]
    output[K,2:3] <- data[K,2:3]
  }

  ##Concentration to activity
  if(input_unit == "abundance" || input_unit == "ppm/%"){
    output[U,2:3] <- data[U,2:3] * convers.factor.U238
    output[Th,2:3] <- data[Th,2:3] * convers.factor.Th232
    output[K,2:3] <- data[K,2:3] * convers.factor.K40

    output[U,4:5] <- data[U,2:3]
    output[Th,4:5] <- data[Th,2:3]
    output[K,4:5] <- data[K,2:3]
  }

  # Return value --------------------------------------------------------------------------------
  if(verbose) print(output)

  invisible(set_RLum(
    class = "RLum.Results",
    data = list(data = output),
    info = list(call = sys.call())))
}

## load data
data_activity <- data.frame(
    NUCLIDES = c("U-238", "Th-232", "K-40"),
    VALUE = c(40,80,100),
    VALUE_ERROR = c(4,8,10),
    stringsAsFactors = FALSE)

data_abundance <- data.frame(
    NUCLIDES = c("U-238", "Th-232", "K-40"),
    VALUE = c(4,8,1),
    VALUE_ERROR = c(0.1,0.1,0.1),
    stringsAsFactors = FALSE)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(
    object = convert_Activity2Concentration(),
    "[convert_Activity2Concentration()] 'data' should be of class 'data.frame'",
    fixed = TRUE)

  expect_error(
    object = convert_Activity2Concentration(data = data_activity[,1:2]),
    "[convert_Activity2Concentration()] 'data' should have at least 3 columns",
    fixed = TRUE)

  expect_error(
    object = convert_Activity2Concentration(data = data_activity, input_unit = "stop"),
    "'input_unit' should be one of 'activity' or 'abundance'")

  expect_error(
    object = convert_Activity2Concentration(data_activity,
                                            input_unit = set_RLum("RLum.Data.Curve")),
    "'input_unit' should be of class 'character'")

  expect_error(convert_Activity2Concentration(data_activity, verbose = NA),
    "'verbose' should be a single logical value")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  SW({
  ## check for standard input
  expect_snapshot_RLum(convert_Activity2Concentration(data_activity),
                       tolerance = snapshot.tolerance)

  ## check for concentration input
  expect_snapshot_RLum(convert_Activity2Concentration(data_abundance,
                                                      input_unit = "abundance"),
                       tolerance = snapshot.tolerance)
  })
})

test_that("more checks", {
  testthat::skip_on_cran()

  ## additional checks for input
  ## capitalized input units
  SW({
    expect_s4_class(convert_Activity2Concentration(data_activity, input_unit = "ACTIVITY"), c("RLum.Results"))

    ## check backwards compatibility
    expect_s4_class(convert_Activity2Concentration(data_activity, input_unit = "Bq/kg"), c("RLum.Results"))
  })
})

#' @title Apply the internal-external-uncertainty (IEU) model after Thomsen et al.
#' (2007) to a given De distribution
#'
#' @description Function to calculate the IEU De for a De data set.
#'
#' @details
#' This function uses the equations of Thomsen et al. (2007). The parameters a
#' and b are estimated from dose-recovery experiments.
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with De `(data[,1])` and
#' De error `(values[,2])`
#'
#' @param a [numeric] (**required**):
#' slope
#'
#' @param b [numeric] (**required**):
#' intercept
#'
#' @param interval [numeric] (**required**):
#' fixed interval (e.g. 5 Gy) used for iteration of `Dbar`, from the mean to
#' Lowest.De used to create Graph.IEU `[Dbar.Fixed vs Z]`
#'
#' @param decimal.point [numeric] (*with default*):
#' number of decimal points for rounding calculations (e.g. 2)
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments (`trace, verbose`).
#'
#' @return
#' Returns a plot (*optional*) and terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following elements:
#'
#' \item{.$summary}{[data.frame] summary of all relevant model results.}
#' \item{.$data}{[data.frame] original input data}
#' \item{.$args}{[list] used arguments}
#' \item{.$call}{[call] the function call}
#' \item{.$tables}{[list] a list of data frames containing all calculation tables}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @section Function version: 0.1.3
#'
#' @author
#' Rachel Smedley, Geography & Earth Sciences, Aberystwyth University (United Kingdom) \cr
#' Based on an excel spreadsheet and accompanying macro written by Kristina Thomsen. \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [plot], [Luminescence::calc_CommonDose], [Luminescence::calc_CentralDose],
#' [Luminescence::calc_FiniteMixture], [Luminescence::calc_FuchsLang2001], [Luminescence::calc_MinDose]
#'
#' @references
#' Smedley, R.K., 2015. A new R function for the Internal External Uncertainty (IEU) model.
#' Ancient TL 33, 16-21.
#'
#' Thomsen, K.J., Murray, A.S., Boetter-Jensen, L. & Kinahan, J.,
#' 2007. Determination of burial dose in incompletely bleached fluvial samples
#' using single grains of quartz. Radiation Measurements 42, 370-379.
#'
#' @examples
#'
#' ## load data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## apply the IEU model
#' ieu <- calc_IEU(ExampleData.DeValues$CA1, a = 0.2, b = 1.9, interval = 1)
#'
#' @export
calc_IEU <- function(
  data,
  a,
  b,
  interval,
  decimal.point = 2,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_IEU")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("data.frame", "RLum.Results"))
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data)
  }

  ## check that we actually have data
  if (length(data) == 0 || nrow(data) == 0) {
    .throw_error("'data' contains no data")
  }
  if (ncol(data) < 2) {
    .throw_error("'data' should have at least two columns")
  }
  data <- data[, 1:2]
  colnames(data) <- c("De", "De.Error")

  .validate_class(a, "numeric")
  .validate_class(b, "numeric")
  .validate_positive_scalar(interval)

  ##==========================================================================##
  ## ... ARGUMENTS
  ##==========================================================================##
  extraArgs <- list(...)

  ## console output
  verbose <- extraArgs$verbose %||% TRUE

  ## trace calculations
  trace <- extraArgs$trace %||% FALSE
  # TODO: main, xlab, ylab, xlim, ylim, pch, col


  ##============================================================================##
  ## CALCULATIONS
  ##============================================================================##

  main.calculation <- function(data, Dbar, round.z = FALSE,
                               plot = FALSE, trace = FALSE) {
    N <- nrow(data)
    De <- data$De
    De.Total.Error <- sqrt(data$De.Error^2 + (a * Dbar + b)^2)
    Inv.De.Total.Error.Sq <- 1 / De.Total.Error^2
    Z.top <- cumsum(Inv.De.Total.Error.Sq * De)
    Z.bot <- cumsum(Inv.De.Total.Error.Sq)
    Z <- Z.top / Z.bot
    if (round.z)
      Z <- round(Z, decimal.point)

    EXT.top <- NULL
    for (j in 1:N) {
      x <- (De[1:j] - Z[j])^2 * Inv.De.Total.Error.Sq[1:j]
      EXT.top <- c(EXT.top, sum(x))
    }
    EXT.bot <- (1:N - 1) * Z.bot
    EXT <- EXT.top / EXT.bot
    INT <- 1 / Z.bot
    R <- sqrt(INT / EXT)
    R.Error <- (2 * (1:N - 1))^(-0.5)
    Rank.number <- 1:N
    Table.IEU <- data.table(Rank.number = Rank.number,
                            De = De, De.Error = De.Total.Error,
                            Z, EXT.top, EXT, INT, R, R.Error)

    ## to reduce the number of plots and increase performance,
    ## intermediate calculations are plotted only when `trace = TRUE`
    if (plot && trace) {
      do.plot(Table.IEU$Z, Table.IEU$R, Table.IEU$R.Error)
    }

    Max <- Table.IEU[R >= 1, suppressWarnings(max(Rank.number, na.rm = TRUE))]
    if (is.infinite(Max)) {
      Max <- 1
    }
    Above <- Table.IEU[Max]
    Below <- Table.IEU[Max + 1]
    Slope <- (Above$R - Below$R) / (Above$Z - Below$Z)
    Intercept <- Above$R - (Slope * Above$Z)
    IEU.De <- round((1 - Intercept) / Slope, decimal.point)
    IEU.Error <- max(sqrt(Above$EXT), sqrt(Below$EXT))
    IEU.Error <- round(IEU.Error, decimal.point)
    n <- Max + 1

    if (trace) {
      message(sprintf("[Iteration of Dbar] \n Dbar: %.4f \n IEU.De: %.4f \n IEU.Error: %.4f \n n: %i \n R: %.4f \n",
                      Dbar, IEU.De, IEU.Error, n, Below$R))
    }

    Dbar.Fixed <- Dbar - interval
    return(list(Dbar.Mean = c(1, Dbar, Dbar.Fixed, IEU.De, IEU.Error,
                              n, Below$R, a, b),
                Table.IEU = Table.IEU))
  }

  do.plot <- function(x.vals, y.vals, y.errs,
                      xlab = "Z [Gy]",
                      ylab = expression(paste("R = [", alpha["in"], "/", alpha["ex"],"]")),
                      abline.vals = c(1, 0),
                      asp = NA) {
    ymin <- min((y.vals - y.errs)[-1], na.rm = TRUE)
    ymax <- max((y.vals + y.errs)[-1], na.rm = TRUE)

    plot(x.vals, y.vals, type = "b", xlab = xlab, ylab = ylab,
         ylim = c(min(ymin, 0), ymax),
         asp = asp)

    graphics::arrows(x.vals, y.vals + y.errs, x.vals, y.vals - y.errs,
                    col = 1, angle = 90, length = 0.05, code = 3)

    abline(abline.vals[1], abline.vals[2], untf = FALSE, lty = 3)
  }

  ## ------------------------------------------------------------------------
  ## Start of the algorithm

  # (a) Calculate IEU at fixed intervals of Dbar starting from the Mean and
  # subtracting the interval until Dbar is < Lowest.De; this creates a plot
  data <- data[order(data$De), ]
  Lowest.De <- round(data$De[1], decimal.point)
  Dbar <- round(mean(data$De), decimal.point)

  Dbar.Mean <- main.calculation(data, Dbar)$Dbar.Mean
  Fixed.Iterations <- rbind(data.frame(matrix(nrow = 0, ncol = 9)),
                            Dbar.Mean)
  repeat {
    Dbar.Fixed <- Dbar.Mean[3]
    if (Dbar.Fixed < Lowest.De) {
      break
    }
    Dbar <- Dbar.Fixed
    Dbar.Mean <- main.calculation(data, Dbar)$Dbar.Mean
    Fixed.Iterations <- rbind(Fixed.Iterations, Dbar.Mean)
  }

  colnames(Fixed.Iterations) <- c(FALSE, "Dbar", "Dbar.Fixed", "Zbar",
                                  "Zbar.Error", "n", "Below.R", "a", "b")

  if (plot) {
    do.plot(Fixed.Iterations$Dbar,
            Fixed.Iterations$Zbar,
            Fixed.Iterations$Zbar.Error,
            xlab = "Dbar (Gy)", ylab = "Zbar, weighted mean (Gy)",
            abline.vals = c(0, 1), asp = 1)
  }

  # (b) Calculate Dbar by iteration from [Dbar = Lowest.De] until [IEU.De = Dbar];
  # this calculates the IEU De
  Dbar <- Lowest.De
  res <- main.calculation(data, Dbar)
  Dbar.Mean <- res$Dbar.Mean

  repeat {
    IEU.De <- Dbar.Mean[4]
    if (is.na(IEU.De)) {
      .throw_warning("Numerical error, try changing your 'a' and 'b' values")
      break
    }
    if (IEU.De <= Dbar) {
      break
    }
    Dbar <- IEU.De
    res <- main.calculation(data, Dbar, round.z = TRUE, plot, trace)
    Dbar.Mean <- res$Dbar.Mean
  }

  Dbar <- Dbar.Mean[2]
  IEU.De <- Dbar.Mean[4]
  IEU.Error <- Dbar.Mean[5]
  n <- Dbar.Mean[6]
  Table.IEU <- res$Table.IEU

  # final plot
  if (plot) {
    do.plot(Table.IEU$Z, Table.IEU$R, Table.IEU$R.Error)
  }

  Table.Results <- data.frame(Dbar, IEU.De, IEU.Error, n, a, b)
  colnames(Table.Results) <- c("Dbar", "IEU.De (Gy)", "IEU.Error (Gy)",
                               "Number of De", "a", "b")

  ##==========================================================================##
  ## TERMINAL OUTPUT
  ##==========================================================================##
  if (verbose) {
    message(sprintf(
      "\n [calc_IEU] \n\n Dbar: %.2f \n IEU.De (Gy): %.2f \n IEU.Error (Gy): %.2f Number of De: %.0f \n a: %.4f \n b: %.4f",
      Table.Results[1], Table.Results[2], Table.Results[3],
      Table.Results[4], Table.Results[5], Table.Results[6]))
  }

  ##==========================================================================##
  ## RETURN VALUES
  ##==========================================================================##
  summary <- Table.Results[ ,c(-1, -5, -6)]
  colnames(summary) <- c("de", "de_err", "n")

  call <- sys.call()
  args <- list(a = a, b = b, interval = interval,
               decimal.point = decimal.point, plot = plot)

  set_RLum(
    class = "RLum.Results",
    data = list(summary = summary,
                data = data,
                args = args,
                call = call,
                tables = list(
                  Table.IEUCalculations = as.data.frame(Table.IEU),
                  Table.Fixed.Iteration = Fixed.Iterations,
                  Table.IEUResults = Table.Results
                )))
}

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  df <- ExampleData.DeValues$CA1
  expect_error(calc_IEU("error", a = 0.2, b = 1.9, interval = 1),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_IEU(df[, 1, drop = FALSE], a = 0.2, b = 1.9, interval = 1),
               "'data' should have at least two columns")
  expect_error(calc_IEU(df, a = "error", b = 1.9, interval = 1),
               "'a' should be of class 'numeric'")
  expect_error(calc_IEU(df, a = 0.2, b = "error", interval = 1),
               "'b' should be of class 'numeric'")
  expect_error(calc_IEU(df, a = 0.2, b = 1.9, interval = "error"),
               "'interval' should be a single positive value")
  expect_error(calc_IEU(df, a = 0.2, b = 1.9, interval = NA_real_),
               "'interval' should be a single positive value")
  expect_error(calc_IEU(data.frame(), a = 0.2, b = 1.9, interval = 1),
               "'data' contains no data")
  expect_error(calc_IEU(iris[0, ], a = 0.2, b = 1.9, interval = 1),
               "'data' contains no data")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##standard
  expect_silent(calc_IEU(
    ExampleData.DeValues$CA1,
    a = 0.2,
    b = 1.9,
    interval = 1,
    verbose = FALSE, plot =FALSE
  ))

  ## enable plot and verbose (using default values for coverage)
  SW({
  expect_message(calc_IEU(
    ExampleData.DeValues$CA1,
    a = 0.2,
    b = 1.9,
    interval = 1,
    trace = TRUE
  ))
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(calc_IEU(
      ExampleData.DeValues$CA1,
      a = 0.25,
      b = 1.29,
      interval = 1,
      verbose = FALSE,
      plot = FALSE),
      tolerance = snapshot.tolerance)

  ##provide RLum.Results
  expect_snapshot_RLum(calc_IEU(
      set_RLum(class = "RLum.Results",
               data = list(test = ExampleData.DeValues$CA1)),
      a = 0.12,
      b = 2.14,
      interval = 0.9,
      verbose = FALSE, plot = FALSE),
      tolerance = snapshot.tolerance)
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 424
  expect_warning(calc_IEU(
      ExampleData.DeValues$CA1,
      a = 0.45,
      b = 1.29,
      interval = 1,
      verbose = FALSE,
      plot = FALSE),
      "Numerical error, try changing your 'a' and 'b' values")
  expect_warning(calc_IEU(
      ExampleData.DeValues$CA1,
      a = 0.12,
      b = 1.29,
      interval = 10,
      verbose = FALSE,
      plot = FALSE),
      "Numerical error, try changing your 'a' and 'b' values")
})

test_that("test additional C++ functionality", {
  testthat::skip_on_cran()

  SW({
  expect_type(Luminescence:::src_EED_Calc_Overall_StatUncertainty(
    M_Simul = matrix(1:100, ncol = 20),
    Ndata = 20,
    Nsimul = 2,
    MinNbSimExp = 20), "double")
  })
})

#' General accessor function for RLum S4 class objects
#'
#' Function calls object-specific get functions for RisoeBINfileData S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [Risoe.BINfileData-class] objects. \cr
#' Depending on the input object, the corresponding get function will be selected.
#' Allowed arguments can be found in the documentations of the corresponding 
#' [Risoe.BINfileData-class] class.
#'
#' @param object [Risoe.BINfileData-class] (**required**): 
#' S4 object of class `RLum`
#' 
#' @param ... further arguments that one might want to pass to the specific
#' get function
#' 
#' @return Return is the same as input objects as provided in the list
#' 
#' @section Function version: 0.1.0
#' 
#' @author 
#' Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
#' 
#' @seealso [Risoe.BINfileData-class]
#' 
#' @keywords utilities
#'
#' @md
#' @export
setGeneric(
  name = "get_Risoe.BINfileData",
  def = function(object, ...) {
    standardGeneric("get_Risoe.BINfileData")
  },
  package = "Luminescence"
)

#' General replication function for RLum S4 class objects
#'
#' Function replicates RLum S4 class objects and returns a list for this objects
#'
#' @param object [RLum-class] (**required**):
#' an [RLum-class] object
#'
#' @param times [integer] (*optional*): 
#' number for times each element is repeated element
#'
#' @return Returns a [list] of the object to be repeated
#'
#' @section Function version: 0.1.0
#'
#' @author 
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#'
#' @seealso [RLum-class]
#'
#' @keywords utilities
#'
#' @md
#' @export
setGeneric("replicate_RLum", function (object, times = NULL) {
   standardGeneric("replicate_RLum")
})

#' @title Transform a CW-OSL curve into a pPM-OSL curve via interpolation under
#' parabolic modulation conditions
#'
#' @description
#' Transforms a conventionally measured continuous-wave (CW) OSL-curve into a
#' pseudo parabolic modulated (pPM) curve under parabolic modulation conditions
#' using the interpolation procedure described by Bos & Wallinga (2012).
#'
#' @details
#' The complete procedure of the transformation is given in Bos & Wallinga
#' (2012). The input `data.frame` consists of two columns: time (t) and
#' count values (CW(t))
#'
#' **Nomenclature**
#'
#' - P = stimulation time (s)
#' - 1/P = stimulation rate (1/s)
#'
#' **Internal transformation steps**
#'
#' (1)
#' log(CW-OSL) values
#'
#' (2)
#' Calculate t' which is the transformed time:
#' \deqn{t' = (1/3)*(1/P^2)t^3}
#'
#' (3)
#' Interpolate CW(t'), i.e. use the log(CW(t)) to obtain the count values for
#' the transformed time (t'). Values beyond `min(t)` and `max(t)`
#' produce `NA` values.
#'
#' (4)
#' Select all values for t' < `min(t)`, i.e. values beyond the time resolution
#' of t. Select the first two values of the transformed data set which contain
#' no `NA` values and use these values for a linear fit using [lm].
#'
#' (5)
#' Extrapolate values for t' < `min(t)` based on the previously obtained
#' fit parameters. The extrapolation is limited to two values. Other values at
#' the beginning of the transformed curve are set to 0.
#'
#' (6)
#' Transform values using
#' \deqn{pLM(t) = t^2/P^2*CW(t')}
#'
#' (7)
#' Combine all values and truncate all values for t' > `max(t)`
#'
#' **Note:**
#' The number of values for t' < `min(t)` depends on the stimulation
#' period `P`. To avoid the production of too many artificial data at the
#' raising tail of the determined pPM curve, it is recommended to use the
#' automatic estimation routine for `P`, i.e. provide no value for
#' `P`.
#'
#' @inheritParams convert_CW2pLMi
#'
#' @return
#' The function returns the same data type as the input data type with
#' the transformed curve values.
#'
#' `RLum.Data.Curve`
#'
#' \tabular{rl}{
#' `$CW2pPMi.x.t` \tab: transformed time values \cr
#' `$CW2pPMi.method` \tab: used method for the production of the new data points
#' }
#'
#' `data.frame`
#'
#' \tabular{rl}{
#' `$x` \tab: time\cr
#' `$y.t` \tab: transformed count values\cr
#' `$x.t` \tab: transformed time values \cr
#' `$method` \tab: used method for the production of the new data points
#' }
#'
#' @note
#' According to Bos & Wallinga (2012), the number of extrapolated points
#' should be limited to avoid artificial intensity data. If `P` is
#' provided manually, not more than two points are extrapolated.
#'
#' @section Function version: 0.2.5
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#' Based on comments and suggestions from:
#' Adrie J.J. Bos, Delft University of Technology, The Netherlands
#'
#' @seealso [Luminescence::convert_CW2pLM], [Luminescence::convert_CW2pLMi], [Luminescence::convert_CW2pHMi],
#' [Luminescence::fit_LMCurve], [Luminescence::RLum.Data.Curve-class]
#'
#' @references
#' Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL
#' signal components. Radiation Measurements 47, 752-758.
#'
#' **Further Reading**
#'
#' Bulur, E., 1996. An Alternative Technique For
#' Optically Stimulated Luminescence (OSL) Experiment. Radiation Measurements
#' 26, 701-709.
#'
#' Bulur, E., 2000. A simple transformation for converting CW-OSL curves to
#' LM-OSL curves. Radiation Measurements 32, 141-145.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(1)
#' ##load CW-OSL curve data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ##transform values
#' values.transformed <- convert_CW2pPMi(ExampleData.CW_OSL_Curve)
#'
#' ##plot
#' plot(values.transformed$x,values.transformed$y.t, log = "x")
#'
#' ##(2) - produce Fig. 4 from Bos & Wallinga (2012)
#'
#' ##load data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#' values <- CW_Curve.BosWallinga2012
#'
#' ##open plot area
#' plot(NA, NA,
#'      xlim = c(0.001,10),
#'      ylim = c(0,8000),
#'      ylab = "pseudo OSL (cts/0.01 s)",
#'      xlab = "t [s]",
#'      log = "x",
#'      main = "Fig. 4 - Bos & Wallinga (2012)")
#'
#' values.t <- convert_CW2pLMi(values, P = 1/20)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col = "red",lwd = 1.3)
#' text(0.03,4500,"LM", col = "red", cex = .8)
#'
#' values.t <- convert_CW2pHMi(values, delta = 40)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col = "black", lwd = 1.3)
#' text(0.005,3000,"HM", cex = .8)
#'
#' values.t <- convert_CW2pPMi(values, P = 1/10)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col = "blue", lwd = 1.3)
#' text(0.5,6500,"PM", col = "blue", cex = .8)
#'
#' @export
convert_CW2pPMi<- function(
  object,
  P = NULL,
  ...
) {
  .set_function_name("convert_CW2pPMi")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  temp.values <- .prepare_CW2pX(object)
  .validate_positive_scalar(P, null.ok = TRUE)

  # (3) Transform values ------------------------------------------------------

  ##log transformation of the CW-OSL count values
  CW_OSL.log<-log(temp.values[,2])

  ##time transformation t >> t'
  t<-temp.values[,1]
  if (anyDuplicated(t) > 0) {
    .throw_error("'object' contains duplicated time values")
  }

  ##set P
  ##if no values for P is set selected a P value for a maximum of
  ##two extrapolation points
  if (is.null(P)) {
    i<-1
    P<-1/i
    t.transformed<-(1/3)*(1/P^2)*t^3

    while(length(t.transformed[t.transformed<min(t)])>2){
      P<-1/i
      t.transformed<-(1/3)*(1/P^2)*t^3
      i<-i+1
    }
  }else{
    t.transformed<-(1/3)*(1/P^2)*t^3
  }

  # (4) Interpolation ---------------------------------------------------------

  ##interpolate values, values beyond the range return NA values
  CW_OSL.interpolated <- approx(t, CW_OSL.log, xout=t.transformed, rule=1 )
  if (all(is.na(CW_OSL.interpolated$y))) {
    .throw_error("All points are outside the interpolation range")
  }

  ##combine t.transformed and CW_OSL.interpolated in a data.frame
  temp<-data.frame(x=t.transformed, y = unlist(CW_OSL.interpolated$y))


  # (5) Extrapolate first values of the curve ---------------------------------
  res <- .extrapolate_first(temp, t = t[1:2], y = CW_OSL.log[1:2])
  temp <- res$df
  temp.method <- res$method

  # (6) Convert, transform and combine values ---------------------------------

  ##unlog CW-OSL count values, i.e. log(CW) >> CW
  CW_OSL<-exp(temp$y)

  ##transform CW-OSL values to pPM-OSL values

  pPM<-(t^2/P^2)*CW_OSL

  ##combine all values and exclude NA values
  temp.values <- data.frame(x=t, y.t=pPM, x.t=t.transformed, method=temp.method)
  temp.values <- na.exclude(temp.values)

  # (7) Return values ---------------------------------------------------------

  ##returns the same data type as the input
  if (is.data.frame(object)) {
    return(temp.values)
  }

    ##add old info elements to new info elements
    temp.info <- c(object@info,
                   CW2pPMi.x.t = list(temp.values$x.t),
                   CW2pPMi.method = list(temp.values$method))

  set_RLum(
      class = "RLum.Data.Curve",
      recordType = object@recordType,
      data = as.matrix(temp.values[,1:2]),
      info = temp.info)
}

#' S4-names function for RLum S4 class objects
#'
#' Function calls object-specific names functions for RLum S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum-class] objects.\cr
#' Depending on the input object, the corresponding 'names' function will be
#' selected. Allowed arguments can be found in the documentations of the
#' corresponding [RLum-class] class.
#'
#' @param object [RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @return Returns a [character]
#'
#' @section Function version: 0.1.0
#'
#' @author
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#'
#' @seealso [RLum.Data.Curve-class], [RLum.Data.Image-class],
#' [RLum.Data.Spectrum-class], [RLum.Analysis-class], [RLum.Results-class]
#'
#' @keywords utilities
#'
#' @aliases names_RLum
#'
#' @md
#' @export
setGeneric("names_RLum", function(object) {
  standardGeneric("names_RLum")
})

# Method for names_RLum method for RLum objects in a list for a list of objects  -------------
#' @describeIn names_RLum
#' Returns a list of [RLum-class] objects that had been passed to [names_RLum]
#'
#'
#' @md
#' @export
setMethod("names_RLum",
          signature = "list",
          function(object) {
            ##apply method in the objects and return the same
            lapply(object, function(x) {
              if (inherits(x, "RLum")) {
                return(names_RLum(x))
              } else{
                return(x)
              }

            })

          })

#' @title Class `"RLum.Data"`
#'
#' @description Generalized virtual data class for luminescence data.
#'
#' @name RLum.Data-class
#'
#' @docType class
#'
#' @section Objects from the Class:
#' A virtual class: no objects can be created from it.
#'
#' @section Class version: 0.2.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum-class], [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Data.Image-class]
#'
#' @keywords classes internal
#'
#' @examples
#'
#' showClass("RLum.Data")
#'
#' @export
setClass("RLum.Data",
         contains = c("RLum", "VIRTUAL")
)

## add_metadata() -----------------------------------------------------------
#' @describeIn metadata
#' Add metadata entries to [Luminescence::RLum.Data-class] objects.
#'
#' @export
setMethod("add_metadata<-",
          signature = "RLum.Data",
          definition = function(object, info_element, value) {
            .set_function_name("add_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            .validate_class(info_element, "character", length = 1)
            valid.names <- names(object@info)
            if (info_element %in% valid.names) {
              .throw_error("'info_element' already present, to modify it ",
                           "you should use `replace_metadata()`")
            }
            if (is.null(value)) {
              .throw_error("Cannot store a metadata entry with NULL value")
            }

            ## add the metadata element
            object@info[[info_element]] <- value
            assign(x = deparse(substitute(object))[1], object)
          })

## rename_metadata() --------------------------------------------------------
#' @describeIn metadata
#' Rename a metadata entry of [Luminescence::RLum.Data-class] objects.
#'
#' @export
setMethod("rename_metadata<-",
          signature = "RLum.Data",
          definition = function(object, info_element, value) {
            .set_function_name("rename_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            .validate_class(info_element, "character", length = 1)
            valid.names <- names(object@info)
            if (!info_element %in% valid.names) {
              .throw_error("'info_element' not recognised (",
                           .collapse(info_element), "), valid terms are: ",
                           .collapse(valid.names, quote = FALSE))
            }
            if (is.null(value)) {
              .throw_error("Cannot rename a metadata entry to NULL, ",
                           "to remove it you should use `replace_metadata()`")
            }

            ## rename the metadata element
            name.idx <- grep(info_element, valid.names)
            names(object@info)[name.idx] <- value
            assign(x = deparse(substitute(object))[1], object)
          })

## replace_metadata() -------------------------------------------------------
#' @describeIn metadata
#' Replaces or removes metadata of [Luminescence::RLum.Data-class] objects.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @export
setMethod("replace_metadata<-",
          signature = "RLum.Data",
          definition = function(object, info_element, subset = NULL,
                                verbose = TRUE, value = NULL) {
            .set_function_name("replace_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## Integrity checks ---------------------------------------------
            .validate_class(info_element, "character")
            valid.names <- names(object@info)
            not.found <- setdiff(info_element, valid.names)
            if (length(not.found) > 0) {
              .throw_error("'info_element' not recognised (",
                           .collapse(not.found), "), valid terms are: ",
                           .collapse(valid.names, quote = FALSE))
            }

            ## select relevant rows
            sel <- TRUE
            if (!is.null(substitute(subset))) {

              ## assigning `NULL` indicates that we want to remove a field,
              ## but that is incompatible with choosing a subset of rows
              if (is.null(value)) {
                .throw_error("'subset' is incompatible with assigning NULL")
              }

              ## evaluate the expression to produce a selection
              sel <- tryCatch(eval(
                  expr = substitute(subset),
                  envir = object@info,
                  enclos = parent.frame()
              ), error = function(e) {
                .throw_error("Invalid 'subset' expression, valid terms are: ",
                             .collapse(valid.names, quote = FALSE))
              })
              if (!is.logical(sel)) {
                .throw_error("'subset' should contain a logical expression")
              }
              if (all(is.na(sel))) {
                sel <- FALSE
              }
              if (!any(sel)) {
                if (verbose)
                  .throw_message("'subset' expression produced an ",
                                 "empty selection, nothing done")
                return(object)
              }
            }

            if (!is.null(substitute(subset))) {
              ## replace the metadata elements
              for (field in info_element)
                object@info[[field]][sel] <- value
            } else {
              ## remove the metadata elements
              object@info[info_element] <- value
            }
            assign(x = deparse(substitute(object))[1], object)
          })

## view() -------------------------------------------------------------------
#' @describeIn view
#' View method for [Luminescence::RLum.Data-class] objects.
#'
#' @export
setMethod("view",
          signature = "RLum.Data",
          definition = function(object, ...) {
  .set_function_name("view")
  on.exit(.unset_function_name(), add = TRUE)

  .validate_not_empty(object@info, what = "RLum.Data")

    ## set title
    name <- list(...)$title
    if (is.null(name))
      name <- deparse(substitute(object))

    ## run view
    .view(x = object@info, title = name)
})

#' @title Non-linear Least Squares Fit for LM-OSL curves
#'
#' @description
#' The function determines weighted non-linear least-squares estimates of the
#' component parameters of an LM-OSL curve (Bulur 1996) for a given number of
#' components and returns various component parameters. The fitting procedure
#' uses the Levenberg-Marquardt algorithm as implemented in function `nlsLM`
#' from package `minpack.lm`.
#'
#' @details
#' **Fitting function**
#'
#' The function for the fitting has the general
#' form:
#'
#' \deqn{y = (\exp(0.5) * Im_1 * x / xm_1) * \exp(-x^2 / (2 * xm_1^2)) + \ldots +
#'            \exp(0.5) * Im_i * x / xm_i) * \exp(-x^2 / (2 * xm_i^2))}
#'
#' where \eqn{1 < i < 8}
#'
#' This function and the equations for the conversion to b (detrapping probability)
#' and n0 (proportional to initially trapped charge) have been taken from Kitis
#' et al. (2008):
#'
#' \deqn{xm_i=\sqrt{\max(t)/b_i}}
#' \deqn{Im_i = \exp(-0.5) n0/xm_i}
#'
#' **Background subtraction**
#'
#' When a background signal is provided with the `object.bg` argument, the
#' user can choose among three methods for background subtraction by setting
#' the `bg.subtraction` argument to one of these:
#'
#' - `"polynomial"` (default): a polynomial function is fitted using [glm]
#' and the resulting function is used for background subtraction:
#' \deqn{y = a x^4 + b x^3 + c x^2 + d x + e}
#'
#' - `"linear"`: a linear function is fitted using [glm] and the resulting
#' function is used for background subtraction:
#' \deqn{y = a x + b}
#'
#' - `"channel"`: the measured background signal is subtracted channel-wise
#' from the measured signal.
#'
#' - `"none"`: this disables background subtraction even if `object.bg` is
#' provided.
#'
#' **Start values**
#'
#' The choice of the initial parameters for the `nls`-fitting is a crucial
#' point and the fitting procedure may mainly fail due to ill chosen start
#' parameters. Here, three options are provided:
#'
#' **(a)**
#' If `start_values` is not provided by the user, a cheap guess is made
#' by using the detrapping values found by Jain et al. (2003) for quartz for a
#' maximum of 7 components. Based on these values, the pseudo start parameters
#' `xm` and `Im` are recalculated for the given data set. In all cases, fitting
#' starts with the ultra-fast component and (depending on `n.components`)
#' steps through the following values. If no fit could be achieved, an error
#' plot (for `plot = TRUE`) with the pseudo curve (based on the
#' pseudo start parameters) is provided. This may give the opportunity to
#' identify appropriate start parameters visually.
#'
#' **(b)**
#' If start values are provided, the function works like a simple [nls]
#' fitting approach.
#'
#' **Goodness of fit**
#'
#' The goodness of the fit is given by a pseudo-R² value (pseudo coefficient of
#' determination). According to Lave (1970), the value is calculated as:
#'
#' \deqn{pseudoR^2 = 1 - RSS/TSS}
#'
#' where \eqn{RSS} is the Residual Sum of Squares, and \eqn{TSS} is the Total
#' Sum of Squares.
#'
#' **Error of fitted component parameters**
#'
#' The 1-sigma error for the components is calculated using
#' the function [stats::confint]. Due to considerable calculation time, this
#' option is disabled by default. In addition, the error for the components
#' can be estimated by using internal R functions like [summary]. See the
#' [nls] help page for more information.
#'
#' *For more details on the nonlinear regression in R, see Ritz & Streibig (2008).*
#'
#' @param object [Luminescence::RLum.Data.Curve-class] or [data.frame] (**required**):
#' x,y data of measured values (time and counts).
#'
#' @param object.bg [Luminescence::RLum.Data.Curve-class] or [data.frame] (*optional*):
#' x,y data of measured values (time and counts) for background subtraction.
#'
#' @param n.components [integer] (*with default*):
#' fixed number of components that are to be recognised during fitting
#' (min = 1, max = 7).
#'
#' @param start_values [data.frame] (*optional*):
#' starting values for `Im` and `xm` parameters in the fit. If set to `NULL`,
#' an automatic start value estimation is attempted (see details).
#'
#' @param input.dataType [character] (*with default*):
#' alter the plot output depending on the input data: `"LM"` or `"pLM"` (pseudo-LM).
#' See: [Luminescence::convert_CW2pLM]
#'
#' @param sample_code [character] (*optional*):
#' sample code used for the plot and the optional output table (mtext).
#'
#' @param sample_ID [character] (*optional*):
#' additional identifier used as column header for the table output.
#'
#' @param LED.power [numeric] (*with default*):
#' LED power (max.) used for intensity ramping in mW/cm².
#' **Note:** This value is used for the calculation of the absolute
#' photoionisation cross section.
#'
#' @param LED.wavelength [numeric] (*with default*):
#' LED wavelength in nm used for stimulation.
#' **Note:** This value is used for the calculation of the absolute
#' photoionisation cross section.
#'
#' @param fit.trace [logical] (*with default*):
#' traces the fitting process on the terminal.
#'
#' @param fit.calcError [logical] (*with default*):
#' calculate 1-sigma error range of components using [stats::confint].
#'
#' @param bg.subtraction [character] (*with default*):
#' specifies method for background subtraction (one of `"polynomial"`,
#' `"linear"`, `"channel"`, or `"none"`, see Details). Only considered if
#' `object.bg` is specified.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param plot.BG [logical] (*with default*):
#' enable/disable a plot of the background values with the fit used for the
#' background subtraction.
#'
#' @param plot.residuals [logical] (*with default*):
#' enable/disable the plot of the residuals.
#'
#' @param plot.contribution [logical] (*with default*):
#' enable/disable the plot of the component contribution.
#'
#' @param legend [logical] (*with default*):
#' enable/disable the plot legend.
#'
#' @param legend.pos [character] (*with default*):
#' keyword specifying the position of the legend.
#'
#' @param method_control [list] (*optional*): options to control the output
#' produced. Currently only the 'export.comp.contrib.matrix' (logical) option
#' is supported, to enable/disable export of the component contribution
#' matrix.
#'
#' @param ... Further arguments that may be passed to the plot output, e.g.
#' `main`, `xlab`, `xlab`, `xlim`, `ylim`, `cex`, `log`.
#'
#' @return
#' Various types of plots are returned. For details see above. Furthermore an
#' `RLum.Results` object is returned with the following structure:
#'
#' **`@data:`**
#'
#' `.. $data` : [data.frame] with fitting results\cr
#' `.. $fit` : nls ([nls] object)\cr
#' `.. $component_matrix` : [matrix] with numerical xy-values of the single fitted components with the resolution of the input data
#' `.. $component.contribution.matrix` : [list] component distribution matrix
#'  (produced only if `method_control$export.comp.contrib.matrix = TRUE`)
#'
#' **`info:`**
#'
#' `.. $call` : [call] the original function call
#'
#' Matrix structure for the distribution matrix:
#'
#' Column 1 and 2: time and `rev(time)` values\cr
#' Additional columns are used for the components, two for each component,
#' containing I0 and n0. The last columns `cont.` provide information on
#' the relative component contribution for each time interval including the row
#' sum for this values.
#'
#' @note
#' The pseudo-R² may not be the best parameter to describe the goodness
#' of the fit. The trade off between the `n.components` and the pseudo-R²
#' value currently remains unconsidered.
#'
#' The function **does not** ensure that the fitting procedure has reached a
#' global minimum rather than a local minimum! In any case of doubt, the use of
#' manual start values is highly recommended.
#'
#' @section Function version: 0.3.8
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::fit_CWCurve], [plot], [nls], [minpack.lm::nlsLM], [Luminescence::get_RLum]
#'
#' @references
#' Bulur, E., 1996. An Alternative Technique For Optically
#' Stimulated Luminescence (OSL) Experiment. Radiation Measurements 26, 5,
#' 701-709.
#'
#' Jain, M., Murray, A.S., Boetter-Jensen, L., 2003. Characterisation of
#' blue-light stimulated luminescence components in different quartz samples:
#' implications for dose measurement. Radiation Measurements 37 (4-5),
#' 441-449.
#'
#' Kitis, G. & Pagonis, V., 2008. Computerized curve deconvolution analysis for
#' LM-OSL. Radiation Measurements 43, 737-741.
#'
#' Lave, C.A.T., 1970. The Demand for Urban Mass Transportation. The Review of
#' Economics and Statistics 52 (3), 320-323.
#'
#' Ritz, C. & Streibig, J.C., 2008. Nonlinear Regression with R. R. Gentleman,
#' K. Hornik, & G. Parmigiani, eds., Springer, p. 150.
#'
#' @keywords dplot models
#'
#' @examples
#'
#' ##(1) fit LM data without background subtraction
#' data(ExampleData.FittingLM, envir = environment())
#' fit_LMCurve(values.curve, n.components = 3, log = "x")
#'
#' ##(2) fit LM data with background subtraction and export as JPEG
#' ## -alter file path for your preferred system
#' ##jpeg(file = "~/Desktop/Fit_Output\%03d.jpg", quality = 100,
#' ## height = 3000, width = 3000, res = 300)
#' data(ExampleData.FittingLM, envir = environment())
#' fit_LMCurve(values.curve, object.bg = values.curveBG,
#'             n.components = 2, log = "x", plot.BG = TRUE)
#' ##dev.off()
#'
#' ##(3) fit LM data with manual start parameters
#' data(ExampleData.FittingLM, envir = environment())
#' fit_LMCurve(values.curve,
#'             object.bg = values.curveBG,
#'             n.components = 3,
#'             log = "x",
#'             start_values = data.frame(Im = c(170,25,400), xm = c(56,200,1500)))
#'
#' @export
fit_LMCurve<- function(
  object,
  object.bg,
  n.components = 3,
  start_values = NULL,
  input.dataType = c("LM", "pLM"),
  sample_code = "",
  sample_ID = "",
  LED.power = 36,
  LED.wavelength = 470,
  fit.trace = FALSE,
  fit.calcError = FALSE,
  bg.subtraction = c("polynomial", "linear", "channel", "none"),
  verbose = TRUE,
  plot = TRUE,
  plot.BG = FALSE,
  plot.residuals = TRUE,
  plot.contribution = TRUE,
  legend = TRUE,
  legend.pos = "topright",
  method_control = list(),
  ...
) {
  .set_function_name("fit_LMCurve")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated arguments
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }
  if ("values.bg" %in% ...names()) {
    object.bg <- list(...)$values.bg
    .deprecated(old = "values.bg", new = "object.bg", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------

  ##(1) data.frame or RLum.Data.Curve object?
  .validate_class(object, c("data.frame", "RLum.Data.Curve"))
  .validate_not_empty(object)

  if (inherits(object, "RLum.Data.Curve")) {
    if (!object@recordType %in% c("RBR", "LM-OSL")) {
      .throw_error("recordType should be 'RBR' or 'LM-OSL'. ",
                   "Consider using `as(object, 'data.frame')` if you ",
                   "have used a pseudo transformation function")
    }

    object <- as(object,"data.frame")
  }
  if (ncol(object) < 2) {
    .throw_error("'object' should have 2 columns")
  }

  ##(2) data.frame or RLum.Data.Curve object?
  if (!missing(object.bg)) {
    .validate_class(object.bg, c("data.frame", "RLum.Data.Curve"))

    if (inherits(object.bg, "RLum.Data.Curve") &&
        (is.na(object.bg@recordType) || object.bg@recordType != "RBR"))
      .throw_error("'recordType' for 'object.bg' should be 'RBR'")

    if (inherits(object.bg, "RLum.Data.Curve"))
      object.bg <- as(object.bg, "data.frame")

    ## check if length of bg and signal is consistent
    if (nrow(object) != nrow(object.bg))
      .throw_error("'object' and 'object.bg' have different lengths")

    ## silently limit columns of object.bg to prevent problems
    if (ncol(object.bg) < 2) {
      .throw_error("'object.bg' should have 2 columns")
    }
    object.bg <- object.bg[, 1:2]
  }

  .validate_positive_scalar(n.components, int = TRUE)
  input.dataType <- .validate_args(input.dataType, c("LM", "pLM"))
  bg.subtraction <- .validate_args(bg.subtraction,
                                   c("polynomial", "linear", "channel", "none"))
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)
  .validate_logical_scalar(plot.BG)
  .validate_logical_scalar(plot.residuals)
  .validate_logical_scalar(plot.contribution)
  .validate_logical_scalar(legend)
  .validate_class(method_control, "list")

  ## remove missing values
  values <- object
  if (!missing(object.bg))
    values.bg <- object.bg
  if (anyNA(values)) {
    na.idx <- unique(which(is.na(values), arr.ind = TRUE)[, 1])
    values <- values[-na.idx, ]
    if (!missing(object.bg))
      values.bg <- values.bg[-na.idx, ]
  }
  if (!missing(object.bg) && anyNA(object.bg)) {
    na.idx <- unique(which(is.na(values.bg), arr.ind = TRUE)[, 1])
    values <- values[-na.idx, ]
    values.bg <- values.bg[-na.idx, ]
  }

  if (nrow(values) == 0)
    .throw_error("After NA removal, nothing is left from the data set")

  ## the models use two variables for each component
  if (nrow(values) < n.components * 2) {
    .throw_error("At least ", n.components * 2, " data points are required to ",
                 "fit ", n.components, " components, but only ", nrow(values),
                 " provided")
  }

  ## Set plot format parameters -----------------------------------------------
  extraArgs <- list(...) # read out additional arguments list

  settings <- modifyList(list(
      main = "Default",
      xlim = range(values[, 1]),
      ylim = c(if (input.dataType == "LM") min(values[, 2]) else 0,
               max(values[, 2] * 1.1)),
      xlab = if (input.dataType == "LM") "Time [s]" else "u [s]",
      ylab = if (input.dataType == "LM") {
               paste0("LM-OSL [cts/",
                      round(max(values[, 1]) / nrow(values), digits = 2), " s]")
             } else "pLM-OSL [a.u.]",
      log = "",
      cex = 0.8,
      fun = FALSE
  ), extraArgs)

  method_control <- modifyList(x = list(export.comp.contrib.matrix = FALSE),
                               val = method_control)

  ## deprecated argument
  if ("fit.method" %in% names(extraArgs) && extraArgs$fit.method == "port") {
    .throw_warning("`fit.method = 'port'` is deprecated, fitting always ",
                   "occurs with the 'LM' method")
  }

  # layout safety settings
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ##============================================================================##
  ##  BACKGROUND SUBTRACTION
  ##============================================================================##
  if (!missing(object.bg) && bg.subtraction != "none") {
    if (bg.subtraction %in% c("polynomial", "linear")) {
      if (bg.subtraction == "polynomial") {
        ## fit polynomial function to background
        glm.fit <- stats::glm(values.bg[, 2] ~ values.bg[, 1]
                                         + I(values.bg[, 1]^2)
                                         + I(values.bg[, 1]^3))
      } else {
        ## fit linear function to background
        glm.fit <- stats::glm(values.bg[, 2] ~ values.bg[, 1])
      }

      ## subtract background with fitted function
      values[, 2] <- values[, 2] - glm.fit$fitted.values

    }else if(bg.subtraction=="channel"){
      values[,2]<-values[,2]-values.bg[,2]
    }

    if (plot.BG) {
      par(mfrow = c(1, 1), cex = 1.5 * settings$cex)
      plot(values.bg, main = "Background",
           ylab = "LM-OSL [a.u.]", xlab = "Time [s]")
      mtext(side = 3, sample_code, cex = 0.8 * settings$cex)

      if (bg.subtraction %in% c("polynomial", "linear")) {
        lines(values.bg[, 1], glm.fit$fitted.values, col = "red", lwd = 1.5)

        ## extract and round coefficients to produce the fitted equation
        glm.coef <- round(coef(glm.fit), 2)
        text(0,max(values.bg[,2]),
             paste0("y = ",
                    if (bg.subtraction == "polynomial")
                      paste0(glm.coef[4], "*x^3+", glm.coef[3], "*x^2+"),
                    glm.coef[2], "*x+", glm.coef[1]), pos = 4)
      }
    }

    if (verbose) {
      .throw_message(">> Background subtracted (method = '",
                     bg.subtraction, "')", error = FALSE)
    }
  }


  ##============================================================================##
  ##  FITTING
  ##============================================================================##

  ##------------------------------------------------------------------------##
  ##set function for fit equation (according Kitis and Pagonis, 2008)
  ##////equation used for fitting////(start)
  fit.equation<-function(Im.i,xm.i){
    parse(text = paste0("exp(0.5) * Im[", Im.i, "] * (values[, 1] / xm[",
                        xm.i, "]) * exp(-values[, 1]^2 / (2 * xm[", xm.i, "]^2))",
                        collapse = "+"))
  }
  ##////equation used for fitting///(end)
  ##------------------------------------------------------------------------##

  ##set formula elements for fitting functions
  ## the upper two functions should be removed ... but chances are needed ... TODO
  ##////equation used for fitting////(start)
  fit.formula <- function(n.components){
    Im <- paste0("Im.",1:n.components)
    xm <- paste0("xm.",1:n.components)

    stats::as.formula(paste0("y ~ ", paste("(exp(0.5) * ", Im, "* x /", xm,
                                           ") * exp(-x^2 / (2 *", xm, "^2))",
                                           collapse = " + ")))
  }
  ##////equation used for fitting///(end)

  ##------------------------------------------------------------------------##
  ##automatic start parameter estimation

  ##set fit function
  fit.function <- fit.equation(Im.i = 1:n.components, xm.i = 1:n.components)

  if (is.null(start_values)) {

    ##set b (detrapping) values for a 7-component function taken from Jain et al. (2003)
    b.pseudo<-c(32,2.5,0.65,0.15,0.025,0.0025,0.00030)

    ##calculate xm parameters from values set based on the pseudo curves
    xm.pseudo<-sqrt(max(values[,1])/b.pseudo)

    ## the Im values obtained by calculating residuals
    Im.pseudo <- sapply(xm.pseudo, function(x) {
      xm.residual <- abs(values[, 1] - x)
      values[which.min(xm.residual), 1] # time value of minimum residual
    })

    ##set additional variables
    b.pseudo_start<-1
    b.pseudo_end<-0
    fit.found <- FALSE

    while(!fit.found) {
      xm <- xm.pseudo[b.pseudo_start:(n.components + b.pseudo_end)]
      Im <- Im.pseudo[b.pseudo_start:(n.components + b.pseudo_end)]

          ##re-name for method == "LM"
          names(Im) <- paste0("Im.", 1:n.components)
          names(xm) <- paste0("xm.", 1:n.components)
          start.list <- c(as.list(Im), as.list(xm))

          fit <- try(minpack.lm::nlsLM(
            fit.formula(n.components),
            data = data.frame(x = values[,1],
                              y = values[,2]),
            start = start.list,
            lower = rep(0, length(start.list)),
            trace = fit.trace,
            control = minpack.lm::nls.lm.control(maxiter = 500)
          ), silent = TRUE)

      if (!inherits(fit, "try-error") || n.components + b.pseudo_end == 7) {
        fit.found <- TRUE
      }
      b.pseudo_start <- b.pseudo_start + 1
      b.pseudo_end <- b.pseudo_end + 1
    }#end:while loop fit found

  }else{#endif::missing start values
    ##------------------------------------------------------------------------##
    Im <- start_values[, 1]
    names(Im) <- paste0("Im.", 1:n.components)
    xm <- start_values[, 2]
    names(xm) <- paste0("xm.", 1:n.components)
    start.list <- c(as.list(Im), as.list(xm))
    fit <- try(minpack.lm::nlsLM(
                       fit.formula(n.components),
                       data = data.frame(x = values[, 1],
                                         y = values[, 2]),
                       start = start.list,
                       lower = rep(0, length(start.list)),
                       trace = fit.trace,
                       control = minpack.lm::nls.lm.control(maxiter = 500)),
               outFile = stdout()) # redirect error messages so they can be silenced
  }#endif::startparameter

  ##------------------------------------------------------------------------##

  ##grep parameters
  if (!inherits(fit, "try-error")) {
    parameters<-coef(fit)

    ##write parameters in vectors and order parameters
    Im<-parameters[1:(length(parameters)/2)]
    Im.names <- names(Im)
    xm<-parameters[(1+(length(parameters)/2)):length(parameters)]
    xm.names <- names(xm)

    ##order parameters
    o <- order(xm)
    xm <- xm[o]
    names(xm) <- xm.names
    Im <- Im[o]
    names(Im) <- Im.names

    if (verbose){
      ##print rough fitting information - use the nls() control for more information
      writeLines("\n[fit_LMCurve()]")
      writeLines(paste0("\nFitting was done using a ", n.components,
                        "-component function:\n"))

      ##print parameters
      print(c(xm, Im))

      #print some additional information
      writeLines("\n(equation used for fitting according to Kitis & Pagonis, 2008)")
    }#end if

    ##============================================================================##
    ##  Additional Calculations
    ##============================================================================##

    ##calculate stimulation intensity Schmidt (2008)

    ##Energy - E = h*v
    h <- .const$h # Planck constant (W*s^2)
    ny <- .const$c / (LED.wavelength / 10^9) # frequency of the light
    E<-h*ny

    ##transform LED.power in W/cm^2
    LED.power<-LED.power/1000

    stimulation_intensity<-LED.power/E

    ##calculate b and n from the equation of Bulur(1996) to compare results
    ##Using Equation 5 and 6 from Kitis (2008)
    b<-as.vector(max(values[,1])/xm^2) #detrapping probability
    n0<-as.vector((Im/exp(-0.5))*xm)


    ##CALCULATE 1- sigma CONFIDENCE INTERVAL
    ##------------------------------------------------------------------------##
    b.error <- rep(NA, n.components)
    n0.error <- rep(NA, n.components)

    if(fit.calcError){
      ##option for confidence interval
      values.confint <- try(stats::confint(fit, level = 0.68), silent = TRUE)

      if(!inherits(values.confint, "try-error")) {
        Im.confint <- values.confint[1:(length(values.confint[, 1]) / 2), ]
        xm.confint <- values.confint[((length(values.confint[,1])/2)+1):length(values.confint[,1]),]

        ##error calculation
        b.error <- as.vector(abs((max(values[, 1]) / xm.confint[, 1]^2) -
                                 (max(values[, 1]) / xm.confint[, 2]^2)))
        n0.error <- as.vector(abs(((Im.confint[,1]/exp(-0.5))*xm.confint[,1]) - ((Im.confint[,2]/exp(-0.5))*xm.confint[,2])))

      } else {
        .throw_warning("The computation of the parameter confidence intervals ",
                       "failed. Please try to run stats::confint() manually ",
                       "on the $fit output object")
      }
    }
    ##------------------------------------------------------------------------##


    ##calculate photoionisation cross section and print on terminal
    ##using EQ (5) in Kitis
    cs<-as.vector((max(values[,1])/xm^2)/stimulation_intensity)
    rel_cs<-round(cs/cs[1],digits=4)

    ##coefficient of determination after law
    RSS <- sum(residuals(fit)^2) #residual sum of squares
    TSS <- sum((values[,2] - mean(values[,2]))^2) #total sum of squares
    pR<-round(1-RSS/TSS,digits=4)

    ##============================================================================##
    ## COMPONENT TO SUM CONTRIBUTION MATRIX
    ##============================================================================##

    ##+++++++++++++++++++++++++++++++
    ##set matrix
    ##set polygon matrix for optional plot output
    component.contribution.matrix <- matrix(NA,
                                            nrow = length(values[,1]),
                                            ncol = (2*length(xm)) + 2)

    ##set x-values
    component.contribution.matrix[,1] <- values[,1]
    component.contribution.matrix[,2] <- rev(values[,1])

    ##+++++++++++++++++++++++++++++++
    ##set 1st polygon
    ##1st polygon (calculation)
    y.contribution_first <- (exp(0.5)*Im[1]*values[,1]/
                               xm[1]*exp(-values[,1]^2/(2*xm[1]^2))/
                               (eval(fit.function))*100)

    ##avoid NaN values (might happen with synthetic curves)
    y.contribution_first[is.nan(y.contribution_first)] <- 0

    ##set values in matrix
    component.contribution.matrix[,3] <- 100
    component.contribution.matrix[,4] <- 100-rev(y.contribution_first)

    ##+++++++++++++++++++++++++++++++
    ##set polygons in between
    ##polygons in between (calculate and plot)
    if (length(xm)>2){
      y.contribution_prev <- y.contribution_first
      i<-2

      ##matrix stepping
      k <- seq(3, ncol(component.contribution.matrix), by=2)

      while (i<=length(xm)-1) {
        y.contribution_next<-(exp(0.5)*Im[i]*values[,1]/
                                xm[i]*exp(-values[,1]^2/(2*xm[i]^2))/
                                (eval(fit.function))*100)

        ##avoid NaN values
        y.contribution_next[is.nan(y.contribution_next)] <- 0

        ##set values in matrix
        component.contribution.matrix[, k[i]] <- 100-y.contribution_prev
        component.contribution.matrix[, k[i]+1] <- rev(100-y.contribution_prev-
                                                         y.contribution_next)

        y.contribution_prev <- y.contribution_prev + y.contribution_next

        i<-i+1
      }#end while loop
    }#end if

    ##+++++++++++++++++++++++++++++++
    ##set last polygon

    ##last polygon (calculation)
    y.contribution_last<-(exp(0.5)*Im[length(xm)]*values[,1]/
                            xm[length(xm)]*exp(-values[,1]^2/
                                                 (2*xm[length(xm)]^2))/
                            (eval(fit.function))*100)

    ##avoid NaN values
    y.contribution_last[is.nan(y.contribution_last)] <- 0

    component.contribution.matrix[,((2*length(xm))+1)] <- y.contribution_last
    component.contribution.matrix[,((2*length(xm))+2)] <- 0

    ##change names of matrix to make more easy to understand
    component.contribution.matrix.names <- c("x", "rev.x",
                                             paste0(c("y.c", "rev.y.c"), rep(1:n.components, each = 2)))

    ##calculate area for each component, for each time interval
    component.contribution.matrix.area <- sapply(
      seq(3,ncol(component.contribution.matrix),by=2),
      function(x){
        matrixStats::rowDiffs(cbind(rev(component.contribution.matrix[,(x+1)]),
                       component.contribution.matrix[,x]))
      })

    ##append to existing matrix
    component.contribution.matrix <- cbind(
      component.contribution.matrix,
      component.contribution.matrix.area,
      rowSums(component.contribution.matrix.area)
    )

    ##set final column names
    colnames(component.contribution.matrix) <- c(
      component.contribution.matrix.names,
      paste0("cont.c", 1:n.components),
      "cont.sum")

    ##============================================================================##
    ##  Terminal Output (advanced)
    ##============================================================================##
    if (verbose){
      ##write fill lines
      writeLines("------------------------------------------------------------------------------")
      writeLines("(1) Corresponding values according to the equation in Bulur, 1996 for b and n0:\n")
      for (i in 1:length(b)){
        cat(paste0("b", i), "=", format(b[i], scientific = TRUE),
            "\u00B1", format(b.error[i], scientific = TRUE), "\n")
        cat(paste0("n0", i), "=", format(n0[i], scientific = TRUE),
            "\u00B1", format(n0.error[i], scientific = TRUE), "\n\n")
      }#end for loop

      ##write photoionisation cross section on terminal
      for (i in 1:length(cs)){
        cat(paste0("cs from component.", i), "=",
            format(cs[i], scientific = TRUE, digits = 4),
            "cm^2\t >> relative: ", round(cs[i] / cs[1], 4), "\n")
      }#end for loop

      cat("\n(stimulation intensity value used for calculation: ",
          format(stimulation_intensity, scientific = TRUE), " 1/s 1/cm^2)\n")
      writeLines("(errors quoted as 1-sigma uncertainties)")
      writeLines("------------------------------------------------------------------------------\n")

      #sum of squares
      cat("pseudo-R^2 =", pR, "\n")
    }#end if

    ##============================================================================##
    ##  COMPOSE RETURN VALUES (data.frame)
    ##============================================================================##

    ##write output table if values exists
    if (exists("fit")){
      ## build column names according to a pattern
      pat <- c("Im%d", "xm%d", "b%d", "b%d.error",
               "n%02d", "n%02d.error", "cs%d", "rel_cs%d")
      output.tableColNames <- unlist(lapply(1:7, function(x) sprintf(pat, x)))

      ## set data.frame for a max value of 7 components
      output.table <- rep(NA_real_, length(pat) * 7)
      for (k in 1:n.components) {
        output.table[(k - 1) * 8 + 1:8] <- c(Im[k], xm[k], b[k], b.error[k],
                                             n0[k], n0.error[k], cs[k], rel_cs[k])
      }

      ##add pR and n.components
      output.table <- cbind(sample_ID, sample_code, n.components,
                            data.frame(as.list(output.table)), pR)
      colnames(output.table)<-c("ID","sample_code","n.components",output.tableColNames,"pseudo-R^2")

      ##----------------------------------------------------------------------------##
    }#endif::exists fit

  }else{
    output.table <- NA
    component.contribution.matrix <- NA
    .throw_message("Fitting failed, plot without fit produced")
  }

  # Calculate component curves ----------------------------------------------
  component_matrix <- NA
  if(!inherits(fit,"try-error")){
    component_matrix <- matrix(NA, nrow = nrow(values), ncol = 2 + length(Im))
    colnames(component_matrix) <- c("TIME", "SUM", paste("COMP_", 1:length(Im)))
    component_matrix[, 1] <- values[, 1]
    component_matrix[, 2] <- eval(fit.function)

    ## add single components
    for(i in 1:length(Im)){
      component_matrix[, 2 + i] <-
        exp(0.5) * Im[i] * values[, 1] /
        xm[i] * exp(-values[, 1] ^ 2 / (2 * xm[i] ^ 2))
    }
  }

  # Plotting ----------------------------------------------------------------
  if(plot){
    ##cheat the R check routine
    x <- NULL; rm(x)

    ##grep package colour gallery
    col <- get("col", pos = .LuminescenceEnv)

    ## change xlim/ylim values in case of log plot to avoid problems
    if (settings$log %in% c("x", "xy") && settings$xlim[1] == 0) {
      .throw_warning("'xlim' changed to avoid 0 values for log-scale")
      settings$xlim <- c(2^0.5 / 2 * max(values[, 1]) / nrow(values), settings$xlim[2])

    }
    if (settings$log %in% c("y", "xy") && settings$ylim[1] == 0) {
      settings$ylim[1] <- 0.01
    }

    ##set plot frame
    graphics::layout(matrix(seq(1 + plot.residuals + plot.contribution), ncol = 1),
                     heights = c(10, if (plot.residuals) 3,
                                 if (plot.contribution) 4))
    par(oma = c(1, 1, 1, 1), mar = c(0.5, 2.5, 2, 0), cex = settings$cex)
    mgp <- c(1.5, 0.5, 0)

    ##==upper plot==##
    ##open plot area
    plot_check <- try(plot(
      NA,
      NA,
      xlim = settings$xlim,
      ylim = settings$ylim,
      xlab = "",
      xaxt = "n",
      main = settings$main,
      mgp = mgp,
      log = settings$log,
      cex.axis = 0.8,
      cex.lab = 0.9,
      ylab = settings$ylab
    ), silent = TRUE)

    if (inherits(plot_check, "try-error")) {
      ## reset the graphic device if plotting failed
      .throw_message("Figure margins too large or plot area too small, ",
                     "nothing plotted")
      grDevices::dev.off()
    } else {

    mtext(side = 3, sample_code, cex = 0.8 * settings$cex)

    ##plotting measured signal
    points(values[, 1],
           values[, 2],
           pch = 20,
           col = rgb(0.4, 0.4, 0.4, 0.5))

    ##==pseudo curve==##------------------------------------------------------#

    ##curve for used pseudo values
    if (inherits(fit, "try-error") && is.null(start_values)) {
      fit.function<-fit.equation(Im.i=1:n.components,xm.i=1:n.components)
      Im<-Im.pseudo[1:n.components]
      xm<-xm.pseudo[1:n.components]

      ##draw pseudo curve
      lines(values[,1],eval(fit.function), lwd=2, col="red", lty=2)

      axis(side=1)
      mtext(side = 1, settings$xlab, cex = 0.9 * settings$cex, line = 2)

      mtext(side = 4, paste(n.components, "component pseudo function is shown"),
            cex = 0.7, col = "blue")

      ##draw information text on plot
      text(min(values[,1]),max(values[,2]),"FITTING ERROR!",pos=4)
      legend("topright", "pseudo sum function",
             lty = 2, lwd = 2, col = "red", bty = "n")
    }
    ##==pseudo curve==##------------------------------------------------------##

    ##plot sum function
    if (!inherits(fit, "try-error")) {
      lines(values[,1],eval(fit.function), lwd=2, col="black")
      legend.caption <- c("Sum curve", paste("Component", 1:n.components))
      curve.col <- seq(n.components + 1)

      ##plot signal curves

      ##plot curve for additional parameters
      for (i in 1:length(xm)) {
        curve(exp(0.5) * Im[i] * x / xm[i] * exp(-x^2 / (2 * xm[i]^2)),
              col = col[i + 1], lwd = 2, add = TRUE)
      }

      ## plot legend
      if (legend)
        legend(legend.pos, legend.caption, lty = 1, lwd = 2, bty = "n",
               col = col[curve.col])

      ##==lower plot==##
      ##plot residuals
      if (plot.residuals) {
      par(mar = c(0.5, 2.5, 0, 0))
      plot(values[,1],residuals(fit),
           xlim = settings$xlim,
           xlab = "",
           xaxt = "n",
           type="l",
           col="grey",
           ylab="Residual",
           lwd=2,
           xpd = NA,
           mgp = mgp,
           cex.axis = 0.8,
           cex.lab = 0.9,
           log = gsub("y", "", settings$log))

      ##ad 0 line
      abline(h=0)
      }

      ##------------------------------------------------------------------------#
      ##++component to sum contribution plot ++##
      ##------------------------------------------------------------------------#

      ##plot component contribution to the whole signal
      #open plot area
      if (plot.contribution) {
      par(mar = c(2, 2.5, 0, 0))
      plot(NA,NA,
           xlim = settings$xlim,
           ylim=c(0,100),
           ylab = "Component\ncontrib. [%]",
           xlab = settings$xlab,
           main = "",
           xpd = NA,
           mgp = mgp,
           cex.axis = 0.8,
           cex.lab = 0.9,
           log = gsub("y", "", settings$log))

      stepping <- seq(3,length(component.contribution.matrix),2)

      for(i in 1:length(xm)){

        polygon(c(component.contribution.matrix[,1],
                  component.contribution.matrix[,2]),
                c(component.contribution.matrix[,stepping[i]],
                  component.contribution.matrix[,stepping[i]+1]),
                col = col[i+1])
      }
      }
      ##------------------------------------------------------------------------##
    }#end if try-error for fit

    if (settings$fun) sTeve() # nocov

    } # end if (plot_check)
  }
  ##-----------------------------------------------------------------------------
  ##remove objects
  try(unlist("parameters"))

  ##============================================================================#
  ## Return Values
  ##============================================================================#

  if (!method_control$export.comp.contrib.matrix) {
    component.contribution.matrix <- NA
  }
  set_RLum(
    class = "RLum.Results",
    data = list(
      data = output.table,
      fit = fit,
      component_matrix = component_matrix,
      component.contribution.matrix = list(component.contribution.matrix)
    ),
    info = list(call = sys.call())
  )
}

## load data
data(ExampleData.FittingLM, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fit_LMCurve("error"),
               "'object' should be of class 'data.frame' or 'RLum.Data.Curve'")
  expect_error(fit_LMCurve(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(fit_LMCurve(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")
  expect_error(fit_LMCurve(set_RLum("RLum.Data.Curve")),
               "recordType should be 'RBR' or 'LM-OSL'")
  expect_error(fit_LMCurve(set_RLum("RLum.Data.Curve", recordType = "OSL")),
               "recordType should be 'RBR' or 'LM-OSL'")
  expect_error(fit_LMCurve(values.curve, object.bg = "error"),
               "'object.bg' should be of class 'data.frame' or 'RLum.Data.Curve'")
  expect_error(fit_LMCurve(set_RLum("RLum.Data.Curve", recordType = "LM-OSL"),
                           object.bg = values.curveBG),
               "'object' and 'object.bg' have different lengths")
  expect_error(fit_LMCurve(values.curve, object.bg = values.curveBG[, 1, drop = FALSE]),
               "'object.bg' should have 2 columns")
  expect_error(fit_LMCurve(values.curve, object.bg = values.curveBG,
                           bg.subtraction = "error"),
               "'bg.subtraction' should be one of 'polynomial', 'linear', 'channel'")
  expect_error(fit_LMCurve(values.curve, n.components = "error"),
               "'n.components' should be a single positive integer value")
  expect_error(fit_LMCurve(values.curve, input.dataType = "error"),
               "'input.dataType' should be one of 'LM' or 'pLM'")
  expect_error(fit_LMCurve(
    values.curve,
    object.bg = set_RLum("RLum.Data.Curve", data = as.matrix(values.curve), recordType = "OSL"),
    verbose = FALSE),
               "'recordType' for 'object.bg' should be 'RBR'")
  expect_error(fit_LMCurve(data.frame(NA, 1:5)),
               "After NA removal, nothing is left from the data set")
  expect_error(fit_LMCurve(values.curve[1:5, ]),
               "At least 6 data points are required to fit 3 components")

  ## warning for failed confint ...skip on windows because with R >= 4.2 is does not fail anymore
  # SW({
  # if (!grepl(pattern = "mingw", sessionInfo()$platform) && !grepl(pattern = "linux", sessionInfo()$platform))
  #   fit_LMCurve(values.curve, fit.calcError = TRUE)
  # })

  ## deprecated argument
  SW({
  expect_warning(fit_LMCurve(values = values.curve, fit.calcError = TRUE),
                 "'values' was deprecated in v1.2.0, use 'object' instead")
  expect_warning(fit_LMCurve(values.curve, values.bg = values.curveBG),
                 "'values.bg' was deprecated in v1.2.0, use 'object.bg' instead")
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-4

  SW({
  set.seed(1)
  fit <- fit_LMCurve(values.curve, object.bg = values.curveBG,
                     n.components = 3, log = "x",
                     method_control = list(
                         export.comp.contrib.matrix = TRUE),
                     start_values = data.frame(Im = c(170,25,400),
                                               xm = c(56,200,1500)))
  })
  expect_snapshot_RLum(fit, tolerance = snapshot.tolerance)

  SW({
  fit2 <- fit_LMCurve(values.curve,
                      object.bg = values.curveBG,
                      n.components = 3,
                      log = "x",
                      method_control = list(
                          export.comp.contrib.matrix = TRUE),
                      plot = FALSE)
  })
  expect_snapshot_RLum(fit2, tolerance = snapshot.tolerance)

  SW({
  expect_message(expect_warning(
      fit <- fit_LMCurve(values.curve, object.bg = values.curveBG,
                         n.components = 6, fit.method = "port"),
      "`fit.method = 'port'` is deprecated, fitting always occurs with the 'LM'"),
      "Error: Fitting failed, plot without fit produced")
  expect_equal(fit@data$component_matrix, NA)

  set.seed(1)
  expect_snapshot_RLum(fit_LMCurve(values.curve, object.bg = values.curveBG,
                                   method_control = list(
                                       export.comp.contrib.matrix = TRUE),
                                   plot.BG = TRUE, bg.subtraction = "linear"),
                       tolerance = snapshot.tolerance)

  suppressWarnings(
      expect_warning(fit_LMCurve(values.curve, object.bg = values.curveBG,
                                 n.components = 4, fit.calcError = TRUE),
                 "The computation of the parameter confidence intervals failed")
  )

  ## more coverage
  curveBG <- set_RLum("RLum.Data.Curve", data = as.matrix(values.curveBG), recordType = "RBR")
  expect_warning(fit_LMCurve(values.curve, curveBG, xlab = "s", ylab = "a.u.",
                             xlim = c(0, 4000), log = "xy"),
                 "'xlim' changed to avoid 0 values for log-scale")
  })
  expect_message(fit_LMCurve(values.curve[1:15, ], main = ""),
                 "Fitting failed, plot without fit produced")
  pdf(tempfile(), width = 1, height = 1)
  expect_message(fit_LMCurve(values.curve, verbose = FALSE),
                 "Figure margins too large or plot area too small")

  SW({
  skip_on_os("mac")
  expect_snapshot_RLum(fit_LMCurve(values.curve, object.bg = values.curveBG,
                                   xlim = c(0, 4000), ylim = c(0, 600), cex = 0.9,
                                   method_control = list(
                                       export.comp.contrib.matrix = TRUE),
                                   fit.calcError = TRUE),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(fit_LMCurve(values.curve, object.bg = values.curveBG,
                                   plot.BG = TRUE, input.dataType = "pLM",
                                   method_control = list(
                                       export.comp.contrib.matrix = TRUE),
                                   bg.subtraction = "channel"),
                       tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              fit_LMCurve(values.curve))
  vdiffr::expect_doppelganger("logy",
                              fit_LMCurve(values.curve, log = "y",
                                          legend.pos = "topleft",
                                          plot.contribution = FALSE))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 755
  expect_silent(fit_LMCurve(values.curve, log = "y", verbose = FALSE))
  expect_silent(fit_LMCurve(values.curve, log = "xy", verbose = FALSE,
                            input.dataType = "pLM"))

  ## issue 757
  values.na <- values.curve
  values.na[c(5, 25), ] <- NA
  expect_silent(fit_LMCurve(values.na, verbose = FALSE))
  expect_silent(fit_LMCurve(values.na, object.bg = values.curveBG, verbose = FALSE))

  ## issue 763
  values.na <- values.curveBG
  values.na[c(5, 25), ] <- NA
  expect_silent(fit_LMCurve(values.curve, object.bg = values.na, verbose = FALSE))
})

#' @title Export XSYG-file(s) to CSV-files
#'
#' @description
#' This function is a wrapper function around the functions [Luminescence::read_XSYG2R] and
#' [Luminescence::write_RLum2CSV] and it imports an XSYG-file and directly exports its content
#' to CSV-files. If nothing is set for the argument `path` ([Luminescence::write_RLum2CSV])
#' the input folder will become the output folder.
#'
#' @param file [character] (**required**):
#' name of the XSYG-file to be converted to CSV-files
#'
#' @param ... further arguments that will be passed to the function
#' [Luminescence::read_XSYG2R] and [Luminescence::write_RLum2CSV]
#'
#' @return
#' The function returns either a CSV-file (or many of them) or for the option `export = FALSE`
#' a list comprising objects of type [data.frame] and [matrix]
#'
#' @section Function version: 0.1.1
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data-class], [Luminescence::RLum.Results-class],
#' [utils::write.table], [Luminescence::write_RLum2CSV], [Luminescence::read_XSYG2R]
#'
#' @keywords IO
#'
#' @examples
#'
#' ##transform XSYG-file values to a list
#' data(ExampleData.XSYG, envir = environment())
#' convert_XSYG2CSV(OSL.SARMeasurement$Sequence.Object[1:10], export = FALSE)
#'
#' \dontrun{
#' ##select your BIN-file
#' file <- file.choose()
#'
#' ##convert
#' convert_XSYG2CSV(file)
#'
#' }
#'
#' @export
convert_XSYG2CSV <- function(
  file,
  ...
) {
  .set_function_name("convert_XSYG2CSV")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(file, c("character", "RLum"))
  .validate_not_empty(file)

  ##set input arguments
  convert_XSYG2R_settings.default <- list(
    recalculate.TL.curves = TRUE,
    pattern = ".xsyg",
    txtProgressBar = TRUE,
    export = TRUE
  )

  ##modify list on demand
  convert_XSYG2R_settings <- modifyList(x = convert_XSYG2R_settings.default, val = list(...))

  # Import file ---------------------------------------------------------------------------------
  if(!inherits(file, "RLum")){
    .validate_length(file, 1)
    object <- read_XSYG2R(
      file = file,
      fastForward = TRUE,
      recalculate.TL.curves = convert_XSYG2R_settings$recalculate.TL.curves,
      pattern = convert_XSYG2R_settings$pattern,
      txtProgressBar = convert_XSYG2R_settings$txtProgressBar
   )
  }else{
    object <- file
  }

  # Export to CSV -------------------------------------------------------------------------------

  ##get all arguments we want to pass and remove the doubled one
  arguments <- c(list(object = object, export = convert_XSYG2R_settings$export), list(...))
  arguments[duplicated(names(arguments))] <- NULL

  ##this if-condition prevents NULL in the terminal
  if (convert_XSYG2R_settings$export) {
    invisible(do.call("write_RLum2CSV", arguments))
  }else{
    do.call("write_RLum2CSV", arguments)
  }
}

#' Export measurement data produced by a Daybreak luminescence reader to CSV-files
#'
#' This function is a wrapper function around the functions [read_Daybreak2R] and
#' [write_RLum2CSV] and it imports a Daybreak-file (TXT-file, DAT-file)
#' and directly exports its content to CSV-files.  If nothing is set for the
#' argument `path` ([write_RLum2CSV]) the input folder will become the output folder.
#'
#' @param file [character] (**required**):
#' name of the Daybreak-file (TXT-file, DAT-file) to be converted to CSV-files
#'
#' @param ... further arguments that will be passed to the function
#' [read_Daybreak2R] and [write_RLum2CSV]
#'
#' @return
#' The function returns either a CSV-file (or many of them) or for the option `export = FALSE`
#' a list comprising objects of type [data.frame] and [matrix]
#'
#' @section Function version: 0.1.0
#'
#' @author Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
#'
#' @seealso [RLum.Analysis-class], [RLum.Data-class], [RLum.Results-class],
#' [utils::write.table], [write_RLum2CSV], [read_Daybreak2R]
#'
#' @keywords IO
#'
#' @examples
#'
#' \dontrun{
#' ##select your BIN-file
#' file <- file.choose()
#'
#' ##convert
#' convert_Daybreak2CSV(file)
#'
#' }
#'
#' @export
convert_Daybreak2CSV <- function(
  file,
  ...
) {
  .set_function_name("convert_Daybreak2CSV")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(file, c("character", "RLum"))
  .validate_not_empty(file)

  ##set input arguments
  convert_Daybreak2R_settings.default <- list(
    raw = FALSE,
    verbose = TRUE,
    txtProgressBar = TRUE,
    export = TRUE
  )

  ##modify list on demand
  convert_Daybreak2R_settings <- modifyList(x = convert_Daybreak2R_settings.default, val = list(...))

  # Import file ---------------------------------------------------------------------------------
  if(!inherits(file, "RLum")){
    object <- read_Daybreak2R(
      file = file,
      raw = convert_Daybreak2R_settings$raw,
      verbose = convert_Daybreak2R_settings$raw,
      txtProgressBar = convert_Daybreak2R_settings$raw
   )
  }else{
    object <- file
  }

  # Export to CSV -------------------------------------------------------------------------------

  ##get all arguments we want to pass and remove the doubled one
  arguments <- c(list(object = object, export = convert_Daybreak2R_settings$export), list(...))
  arguments[duplicated(names(arguments))] <- NULL

  ##this if-condition prevents NULL in the terminal
  if (convert_Daybreak2R_settings$export)
    return(invisible(do.call("write_RLum2CSV", arguments)))
  do.call("write_RLum2CSV", arguments)
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_Daybreak2CSV(),
               "'file' should be of class 'character' or 'RLum'")
  expect_error(convert_Daybreak2CSV(character(0)),
               "'file' cannot be an empty character")
  expect_error(convert_Daybreak2CSV(letters),
               "[convert_Daybreak2CSV()] 'file' should have length 1",
               fixed = TRUE)
  expect_error(convert_Daybreak2CSV(""),
               "File '' does not exist")
})

test_that("check class and length of output", {
  testthat::skip_on_cran()

  SW({
  res <- read_Daybreak2R(file = system.file("extdata/Daybreak_TestFile.txt",
                                            package = "Luminescence"))[[1]]
  })
  expect_null(convert_Daybreak2CSV(res, path = tempdir()))
  expect_type(convert_Daybreak2CSV(res, path = tempdir(), export = FALSE),
              "list")
})

#' Class `"RLum.Analysis"`
#'
#' Object class to represent analysis data for protocol analysis, i.e. all curves,
#' spectra etc. from one measurements. Objects from this class are produced,
#' by e.g. [Luminescence::read_XSYG2R], [Luminescence::read_Daybreak2R]
#'
#' @name RLum.Analysis-class
#'
#' @docType class
#'
#' @slot protocol
#' Object of class [character] describing the applied measurement protocol
#'
#' @slot records
#' Object of class [list] containing objects of class [Luminescence::RLum.Data-class]
#'
#' @note
#' The method [Luminescence::structure_RLum] is currently just available for objects
#' containing [Luminescence::RLum.Data.Curve-class].
#'
#' @section Objects from the Class:
#' Objects can be created by calls of the form `set_RLum("RLum.Analysis", ...)`.
#'
#' @section Class version: 0.4.19
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData2RLum.Analysis],
#' [Luminescence::Risoe.BINfileData-class], [Luminescence::RLum-class]
#'
#' @keywords classes methods
#'
#' @examples
#'
#' ## show method
#' showClass("RLum.Analysis")
#'
#' ##set an empty object
#' set_RLum(class = "RLum.Analysis")
#'
#' ## use example data
#' ##load data
#' data(ExampleData.RLum.Analysis, envir = environment())
#'
#' ##show curves in object
#' get_RLum(IRSAR.RF.Data)
#'
#' ##show only the first object, but by keeping the object
#' get_RLum(IRSAR.RF.Data, record.id = 1, drop = FALSE)
#'
#' ## subsetting with SAR sample data
#' data(ExampleData.BINfileData, envir = environment())
#' sar <- object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)
#'
#' ## get
#' get_RLum(sar, subset = "NPOINTS == 250")
#'
#' ## remove
#' remove_RLum(sar, subset = "NPOINTS == 250")
#'
#' @export
setClass("RLum.Analysis",
         slots = list(
           protocol = "character",
           records = "list"
         ),
         contains = "RLum",
         prototype = list(
           protocol = NA_character_,
           records = list()
         )
)


## as() ---------------------------------------------------------------------
##LIST
##COERCE RLum.Analyse >> list AND list >> RLum.Analysis
#' as() - RLum-object coercion
#'
#' for `[RLum.Analysis-class]`
#'
#' **[Luminescence::RLum.Analysis-class]**
#'
#' \tabular{ll}{
#'  **from** \tab **to**\cr
#'   `list` \tab `list`\cr
#' }
#'
#' Given that the [list] consists of [Luminescence::RLum.Analysis-class] objects.
#'
#' @name as
setAs("list", "RLum.Analysis",
      function(from,to){
        new(to,
            protocol = NA_character_,
            records = from)
      })

setAs("RLum.Analysis", "list",
      function(from){
        lapply(from@records, function(x) x)
      })


## show() -------------------------------------------------------------------
#' @describeIn show
#' Show the structure of `RLum.Analysis` objects.
#'
#' @export
setMethod("show",
          signature(object = "RLum.Analysis"),
          function(object){

            ##print
            cat("\n [RLum.Analysis-class]")
            cat("\n\t originator:", paste0(object@originator, "()"))

            cat("\n\t protocol:", object@protocol)
            cat("\n\t additional info elements: ", if(.hasSlot(object, "info")){length(object@info)}else{0})
            cat("\n\t number of records:", length(object@records))

            #skip this part if nothing is included in the object
            if(length(object@records) > 0){
              ##get object class types
              temp <- vapply(object@records, \(x) class(x)[1], character(1))

              ##print object class types
              table.temp <- table(temp)
              lapply(names(table.temp), function(x) {

                ##show RLum class type
                cat("\n\t .. :", x, ":", table.temp[x])

                ##show structure
                ##set width option ... just an implementation for the tutorial output
                temp.width <- if (getOption("width") <= 50) 4 else 7

                ##set line break variable
                linebreak <- FALSE
                env <- environment()

                ##create terminal output
                terminal_output <-
                  vapply(seq_along(object@records),  function(i) {
                    ## take care of NULL objects and keep this output the rest
                    o <- object@records[[i]] %||% return("<NULL object>")
                    
                    if (inherits(object@records[[i]], x)) {
                      if (i %% temp.width == 0 & i != length(object@records)) 
                        assign(x = "linebreak", value = TRUE, envir = env)
              
                      ##FIRST
                      first <-  paste0("#", i, " ", object@records[[i]]@recordType)

                      ##LAST
                      if (i < length(object@records) &&
                          !is.null(object@records[[i]]@info[["parentID"]]) &&
                          !is.null(object@records[[i + 1]]@info[["parentID"]]) &&
                          (object@records[[i]]@info[["parentID"]] ==
                           object@records[[i+1]]@info[["parentID"]])) {
                        last <- " <> "

                      }else {
                        last <- " | "
                        if (i == length(object@records)) {
                          last <- ""

                        } else if (linebreak) {
                          last <- "\n\t .. .. : "
                          assign(x = "linebreak", value = FALSE, envir = env)
                        }
                      }
                      return(paste0(first,last))
                    }
                    return("")
                  }, FUN.VALUE = character(1))

                 ##print on screen, differentiate between records with many
                 ##curves or just one
                 if(any(grepl(terminal_output, pattern = "<>", fixed = TRUE))){
                   cat("\n\t .. .. : ",
                       gsub(pattern = "|", replacement = "\n\t .. .. :",
                            x = terminal_output, fixed = TRUE), sep = "")

                 } else{
                   cat("\n\t .. .. : ", terminal_output, sep = "")
                 }
              })

            }else{
              cat("\n\t >> This is an empty object, which cannot be used for further analysis! <<")
            }
            cat("\n")
          }
)##end show method

## set_RLum() ---------------------------------------------------------------
#' @describeIn set_RLum
#' Construction method for [Luminescence::RLum.Analysis-class] objects.
#'
#' @param protocol [character] (*optional*):
#' sets protocol type for analysis object. Value may be used by subsequent analysis functions.
#'
#' @param records [list] (*optional*):
#' list of [Luminescence::RLum.Analysis-class] objects
#'
#' @param info [list] (*optional*):
#' a list containing additional info data for the object.
#'
#' @export
setMethod(
  "set_RLum",
  signature = "RLum.Analysis",
  definition = function(
    class,
    originator,
    .uid,
    .pid,
    protocol = NA_character_,
    records = list(),
    info = list()) {

    ##produce empty class object
    newRLumAnalysis <- new(Class = "RLum.Analysis")

    ##allow self set to reset an RLum.Analysis object
    if(inherits(records, "RLum.Analysis")){
      #fill slots (this is much faster than the old code!)
      newRLumAnalysis@protocol <- if(missing(protocol)) records@protocol else protocol
      newRLumAnalysis@originator <- originator
      newRLumAnalysis@records <- records@records
      newRLumAnalysis@info <- if(missing(info)) records@info else c(records@info, info)
      newRLumAnalysis@.uid <- .uid
      newRLumAnalysis@.pid <- if(missing(.pid)) records@.uid else .pid

    }else{
      #fill slots (this is much faster than the old code!)
      newRLumAnalysis@protocol <- protocol
      newRLumAnalysis@originator <- originator
      newRLumAnalysis@records <- records
      newRLumAnalysis@info <- info
      newRLumAnalysis@.uid <- .uid
      newRLumAnalysis@.pid <- .pid
    }

    return(newRLumAnalysis)
  }
)

## get_RLum() ---------------------------------------------------------------
#' @describeIn get_RLum
#' Accessor method for RLum.Analysis objects.
#' The optional arguments `record.id`, `recordType`, `curveType` and `RLum.type`
#' allow to limit records by their id (list index number), their record type
#' (e.g. `recordType = "OSL"`), their curve type (e.g. `curveType = "predefined"`
#' or `curveType ="measured"`), or object type.
#'
#' The selection of a specific RLum.type object superimposes the default selection.
#' Currently supported objects are: RLum.Data.Curve and RLum.Data.Spectrum
#'
#' Returns:
#'
#' 1. [list] of [Luminescence::RLum.Data-class] objects or
#' 2. Single [Luminescence::RLum.Data-class] object, if only one object
#' is contained and `recursive = FALSE` or
#' 3. [Luminescence::RLum.Analysis-class] objects for `drop = FALSE`
#'
#' @param record.id [numeric] or [logical] (*optional*):
#' IDs of specific records. If of type `logical` the entire id range is assumed
#' and `TRUE` and `FALSE` indicates the selection.
#'
#' @param recordType [character] (*optional*):
#' record type (e.g., "OSL"). Can be also a vector, for multiple matching,
#' e.g., `recordType = c("OSL", "IRSL")`
#'
#' @param curveType [character] (*optional*):
#' curve type (e.g. "predefined" or "measured")
#'
#' @param RLum.type [character] (*optional*):
#' RLum object type. Defaults to "RLum.Data.Curve" and "RLum.Data.Spectrum".
#'
#' @param protocol [character] (*optional*):
#' currently ignored.
#'
#' @param get.index [logical] (*optional*):
#' return a numeric vector with the index of each element in the RLum.Analysis
#' object (`FALSE` by default).
#'
#' @param drop [logical] (*with default*):
#' coerce to the next possible layer (which are [Luminescence::RLum.Data-class] objects if
#' `object` is an [Luminescence::RLum.Analysis-class] object). If `drop = FALSE`, an object
#' of the same type as the input is returned.
#'
#' @param recursive [logical] (*with default*):
#' if `TRUE` (default) when the result of the `get_RLum()` request is a single
#' object, the object itself will be returned directly, rather than being
#' wrapped in a list. Mostly this makes things easier, but this might be
#' undesired if this method is used within a loop.
#'
#' @param subset [expression] (*optional*):
#' logical or character masking a logical expression indicating elements or rows to keep:
#' missing values are taken as false. This argument takes precedence over all other arguments,
#' meaning they are not considered when subsetting the object. `subset` works slots and
#' info elements.
#'
#' @param env [environment] (*with default*):
#' An environment passed to [eval] as the enclosure. This argument is only
#' relevant when subsetting the object and should not be used manually.
#'
#' @export
setMethod("get_RLum",
          signature = ("RLum.Analysis"),
          function(object, record.id = NULL, recordType = NULL, curveType = NULL, RLum.type = NULL,
                   protocol = "UNKNOWN", get.index = FALSE, drop = TRUE, recursive = TRUE,
                   info.object = NULL, subset = NULL, env = parent.frame(2)) {
            .set_function_name("get_RLum")
            on.exit(.unset_function_name(), add = TRUE)

            if (!is.null(substitute(subset))) {
              # To account for different lengths and elements in the @info slot we first
              # check all unique elements (in all records)
              info_el <- unique(unlist(lapply(object@records, function(el) names(el@info))))

              envir <- as.data.frame(do.call(rbind,
                lapply(object@records, function(el) {
                  val <- c(curveType = el@curveType, recordType = el@recordType, unlist(el@info))
                  # add missing info elements and set NA
                  if (!all(info_el %in% names(val))) {
                    new <- info_el[!info_el %in% names(val)]
                    val <- c(val, setNames(rep("", length(new)), new))
                  }

                 # order the named char vector by its names so we don't mix up the columns
                 val <- val[order(names(val))]
                 return(val)
               })), stringAsFactors = FALSE)

              ## coerce subset to logical if character to make it compatible
              ## with remove_RLum()
              if(inherits(substitute(subset), "character"))
                subset <- parse(text = subset[1])[[1]]

              ##select relevant rows
              sel <- tryCatch(eval(
                expr = substitute(subset),
                envir = envir,
                enclos = env
              ),
              error = function(e) {
                .throw_error("Invalid subset expression, valid terms are: ",
                             .collapse(names(envir)))
              })

              if (!is.logical(sel)) {
                .throw_error("'subset' must contain a logical expression")
              }

              if (all(is.na(sel)))
                sel <- FALSE

              if (any(sel)) {
                .validate_logical_scalar(get.index)
                if (get.index)
                  return(which(sel))

                object@records <- object@records[sel]
                return(object)
              } else {
                mapply(function(name, op) {
                  message("  ", name, ": ", .collapse(unique(op), quote = FALSE))
                }, names(envir), envir)
                .throw_message("'subset' expression produced an ",
                               "empty selection, NULL returned")
                return(NULL)
              }
            }

            ##if info.object is set, only the info objects are returned
            else if(!is.null(info.object)) {
              .validate_class(info.object, "character", null.ok = TRUE, length = 1)
              if(info.object %in% names(object@info)){
                return(unlist(object@info[info.object]))
              }

              ## check for entries
              if (length(object@info) == 0) {
                .throw_warning("This 'RLum.Analysis' object has no info ",
                               "objects, NULL returned")
              } else {
                .throw_warning("Invalid 'info.object' name, valid names are: ",
                               .collapse(names(object@info)))
              }
              return(NULL)

            } else {

              ##check for records
              if (length(object@records) == 0) {
                .throw_warning("This 'RLum.Analysis' object has no records, ",
                               "NULL returned")
                return(NULL)
              }

              ##record.id
              .validate_class(record.id, c("integer", "numeric", "logical"),
                              null.ok = TRUE)
              if (is.null(record.id)) {
                record.id <- 1:length(object@records)
              }
              ##logical needs a slightly different treatment
              ##Why do we need this? Because a lot of standard R functions work with logical
              ##values instead of numerical indices
              if (is.logical(record.id)) {
                record.id <- (1:length(object@records))[record.id]
              }

              ##check if record.id exists
              if (!all(abs(record.id) %in% (1:length(object@records)))) {
                .throw_message("At least one 'record.id' is invalid, ",
                               "NULL returned")
                return(NULL)
              }

              ##recordType
              .validate_class(recordType, "character", null.ok = TRUE)
              if (is.null(recordType)) {
                recordType <- unique(names(object))
              }

              ##curveType
              .validate_class(curveType, "character", null.ok = TRUE)
              if (is.null(curveType)) {
                curveType <- unique(vapply(object@records, function(x)
                  x@curveType, character(1)))
              }

              ##RLum.type
              .validate_class(RLum.type, "character", null.ok = TRUE)
              if (is.null(RLum.type)) {
                RLum.type <- c("RLum.Data.Curve", "RLum.Data.Spectrum", "RLum.Data.Image")
              }

              ##get.index
              .validate_logical_scalar(get.index)

              ##-----------------------------------------------------------------##
              ##a pre-selection is necessary to support negative index selection
              object@records <- object@records[record.id]
              record.id <- seq_along(object@records)

              ## translate input to regular expression and remove ^ $
              recordType <- glob2rx(recordType)
              recordType <- substr(recordType,
                                   start = 2, stop = nchar(recordType) - 1)

              ##select curves according to the chosen parameter
              temp <- lapply(record.id, function(x) {
                  if (inherits(object@records[[x]], RLum.type) &&
                      object@records[[x]]@curveType %in% curveType) {
                    ## use format() to handle NA so that it gets turned into
                    ## the "NA" string (as.character() would leave it as NA)
                    recordType_comp <- format(object@records[[x]]@recordType)

                    temp <- lapply(recordType, function(type) {
                      ## get the results object and if requested, get the index
                      if (grepl(type, recordType_comp)) {
                        if (!get.index) object@records[[x]] else x
                      }
                    })

                    ##remove empty entries and select just one to unlist
                    temp <- .rm_NULL_elements(temp)

                    ##if list has length 0 skip entry
                    if (length(temp) != 0) {
                      temp[[1]]
                    }
                  }
                })

                ##remove empty list element
                temp <- .rm_NULL_elements(temp)

                ##check if the produced object is empty and show warning message
                if (length(temp) == 0)
                  .throw_warning("This request produced an empty list of records")

                ##remove list for get.index
                if (get.index) {
                  return(unlist(temp))
                }
                .validate_logical_scalar(drop)
                if (!drop) {
                    temp <- set_RLum(
                      class = "RLum.Analysis",
                      originator = object@originator,
                      records = temp,
                      protocol = object@protocol,
                      .pid = object@.pid
                    )
                    return(temp)
                }
                if (length(temp) == 1 && recursive)
                  return(temp[[1]])
                return(temp)
            }
          })


## remove_RLum() ------------------------------------------------------------
#' @describeIn remove_RLum
#' Method to remove records from an [Luminescence::RLum.Analysis-class] object.
#'
#' @param ... parameters to be passed to [Luminescence::get_RLum]. The arguments `get.index` and
#' `drop` are preset and have no effect when provided
#'
#' @export
setMethod("remove_RLum",
      signature= "RLum.Analysis",
      definition = function(object, ...) {
  .set_function_name("remove_RLum")
  on.exit(.unset_function_name(), add = TRUE)

# Treatment of ... --------------------------------------------------------
  ## make settings with preset
  args_set <- list(
    get.index = TRUE,
    drop = FALSE
  )
  ## we do not support all arguments; therefore we make a positive list
  args <- list(...)
  args[!names(args) %in% c(
    "record.id",
    "recordType",
    "curveType",
    "RLum.type",
    "protocol",
    "info.object",
    "subset",
    "recursive"
  )] <- NULL

  ## construct call ... record.id always takes priority
  if(!is.null(args$record.id))
    rm_id <- args$record.id
  else
    rm_id <- suppressWarnings(do.call(get_RLum, args = c(object, args_set, args)))

  ## ensure that we don't specify invalid record indices, or we end up adding
  ## NULL records
  rm_id <- rm_id[rm_id <= length(object@records)]

  ## remove objects
  object@records[rm_id] <- NULL
  return(object)
})

## structure_RLum() ---------------------------------------------------------
#' @describeIn structure_RLum
#' Returns the structure of an [Luminescence::RLum.Analysis-class] object.
#'
#' @param fullExtent [logical] (*with default*):
#' extends the returned `data.frame` to its full extent, i.e. all info elements
#' are part of the return as well. The default value is `FALSE` as the data
#' frame might become rather big.
#'
#' @export
setMethod("structure_RLum",
          signature = "RLum.Analysis",
          definition = function(object, fullExtent = FALSE) {
            .set_function_name("structure_RLum")
            on.exit(.unset_function_name(), add = TRUE)

            ##check if the object containing other elements than allowed
            lapply(object@records, .validate_class, "RLum.Data.Curve",
                   name = "All elements of 'object'")

            ##get length object
            temp.object.length <- length(object@records)

            ##ID
            temp.id <- seq_len(temp.object.length)

            ##recordType
            temp.recordType <- names(object)

            ##PROTOCOL STEP
            temp.protocol.step <- rep_len(NA_character_, temp.object.length)

            ## GET LIMITS
            temp.limits <- t(vapply(object@records, function(x) {
              c(nrow(x@data), range(x@data[, 1]), range(x@data[, 2]))
            }, numeric(5)))

            temp.n.channels <- temp.limits[, 1]
            temp.x.min <- temp.limits[, 2]
            temp.x.max <- temp.limits[, 3]
            temp.y.min <- temp.limits[, 4]
            temp.y.max <- temp.limits[, 5]

            ##.uid
            temp.uid <- vapply(object@records, function(x) x@.uid, character(1))

            ##.pid
            temp.pid <- lapply(object@records, function(x) x@.pid )

            ##originator
            temp.originator <- vapply(object@records, function(x) x@originator,
                                      character(1))

            ##curveType
            temp.curveType <- vapply(object@records, function(x) x@curveType,
                                     character(1))

            ##info elements as character value
            if (fullExtent) {
              temp.info.elements <- as.data.frame(
                data.table::rbindlist(lapply(object@records, function(x) {
                x@info
              }), fill = TRUE))

              if (nrow(temp.info.elements) == 0) {
                ## if we are here temp.info.elements 0 rows and 0 columns:
                ## to avoid crashing further down in the data.frame() call,
                ## we create a data frame with the expected number of rows
                temp.info.elements <- data.frame(info = rep(NA, temp.object.length))
              }
            } else {
               temp.info.elements <- lapply(object@records, function(x) {
                 if(is.null(names(x@info)))
                    return(NA)
                  names(x@info)
                 })
            }

            ##combine output to a data.frame
            data.frame(
                id = temp.id,
                recordType = temp.recordType,
                curveType = temp.curveType,
                protocol.step = temp.protocol.step,
                n.channels = temp.n.channels,
                x.min = temp.x.min,
                x.max = temp.x.max,
                y.min = temp.y.min,
                y.max = temp.y.max,
                originator = temp.originator,
                .uid = temp.uid,
                .pid = I(as.list(temp.pid)),
                info = if(fullExtent) temp.info.elements else I(temp.info.elements),
                stringsAsFactors = FALSE
            )
          })

## length_RLum() ------------------------------------------------------------
#' @describeIn length_RLum
#' Returns the number of records stored in the object.
#'
#' @export
setMethod("length_RLum",
          "RLum.Analysis",
          function(object){
            length(object@records)
          })

## names_RLum() -------------------------------------------------------------
#' @describeIn names_RLum
#' Returns the names of the [Luminescence::RLum.Data-class] objects stored in the object.
#'
#' @export
setMethod("names_RLum",
          "RLum.Analysis",
          function(object){
            vapply(object@records, function(x) x@recordType, character(1))
          })


## add_metadata() -----------------------------------------------------------
#' @describeIn metadata
#' Adds metadata to [Luminescence::RLum.Analysis-class] objects.
#'
#' @param info_element [character] (**required**):
#' name of the metadata entry to manipulate.
#'
#' @export
setMethod("add_metadata<-",
          signature= "RLum.Analysis",
          definition = function(object, info_element, value) {
            .set_function_name("add_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## add the metadata to all records
            records <- lapply(object@records, function(x) {
              do.call(`add_metadata<-`,
                      list(x, info_element = info_element, value = value))
            })

            object@records <- records
            assign(x = deparse(substitute(object))[1], object)
          })

## rename_metadata() --------------------------------------------------------
#' @describeIn metadata
#' Renames a metadata entry of [Luminescence::RLum.Analysis-class] objects.
#'
#' @export
setMethod("rename_metadata<-",
          signature= "RLum.Analysis",
          definition = function(object, info_element, value) {
            .set_function_name("rename_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## rename the metadata entry in all records
            records <- lapply(object@records, function(x) {
              do.call(`rename_metadata<-`,
                      list(x, info_element = info_element, value = value))
            })

            object@records <- records
            assign(x = deparse(substitute(object))[1], object)
          })

## replace_metadata() -------------------------------------------------------
#' @describeIn metadata
#' Replaces or removes metadata of [Luminescence::RLum.Analysis-class] objects.
#'
#' @param subset [expression] (*optional*):
#' logical expression to limit the substitution only to the selected subset
#' of elements.
#'
#' @export
setMethod("replace_metadata<-",
          signature= "RLum.Analysis",
          definition = function(object, info_element, subset = NULL, value = NULL) {
            .set_function_name("replace_metadata")
            on.exit(.unset_function_name(), add = TRUE)

            ## this must be evaluated now, as inside the lapply() inner
            ## function it would be evaluated differently
            subset.expr <- substitute(subset)

            ## replace the metadata in all records
            records <- lapply(object@records, function(x) {
              do.call(`replace_metadata<-`,
                      list(x, info_element = info_element,
                           subset = subset.expr,  verbose = FALSE,
                           value = value))
            })

            object@records <- records
            assign(x = deparse(substitute(object))[1], object)
          })


## smooth_RLum() ------------------------------------------------------------
#' @describeIn smooth_RLum
#' Smoothing of `RLum.Data` records contained in the input object.
#'
#' @export
setMethod(
  f = "smooth_RLum",
  signature = "RLum.Analysis",
  function(object, ...) {
    object@records <- lapply(object@records, smooth_RLum, ...)
    return(object)
  }
)

## normalise_RLum() ------------------------------------------------------------
#' @describeIn normalise_RLum
#' Normalisation of `RLum.Data` records contained in the input object.
#'
#' @export
setMethod(
  f = "normalise_RLum",
  signature = "RLum.Analysis",
  function(object, norm, ...) {
    object@records <- lapply(
      X = object@records,
      FUN = normalise_RLum, norm = norm, ...)
    return(object)
  }
)

## sort_RLum() --------------------------------------------------------------
#' @describeIn sort_RLum
#' Sorting of `RLum.Data` objects contained in this `RLum.Analysis` object.
#' At least one of `slot` and `info_element` must be provided. If both are
#' given, ordering by `slot` always takes priority over `info_element`.
#' Only the first element in each `slot` and each `info_element` is used
#' for sorting. Example: `.pid` can contain multiple values, however, only the
#' first is taken.
#' Please note that the `show()` method does some structuring, which may
#' lead to the impression that the sorting did not work.
#'
#' @param slot [character] (*optional*): slot name to use in sorting.
#'
#' @param info_element [character] (*optional*): names of the `info` field
#' to use in sorting. The order of the names sets the sorting priority.
#' Regardless of available info elements, the following
#' elements always exist because they are calculated from the record
#' `XY_LENGTH`, `NCOL`, `X_MIN`, `X_MAX`, `Y_MIN`, `Y_MAX`.
#'
#' @param decreasing [logical] (*with default*): whether the sort order should
#' be decreasing (`FALSE` by default). It can be provided as a vector to control
#' the ordering sequence for each sorting element.
#'
#' @examples
#' data(ExampleData.XSYG, envir = environment())
#' sar <- OSL.SARMeasurement$Sequence.Object[1:5]
#' sort_RLum(sar, solt = "recordType", info_element = c("startDate"))
#'
#' @export
setMethod(
  f = "sort_RLum",
  signature = "RLum.Analysis",
  function(object, slot = NULL, info_element = NULL,
           decreasing = FALSE, ...) {
    .set_function_name("sort_RLum")
    on.exit(.unset_function_name(), add = TRUE)

    ## an empty object has nothing to sort
    if (length(object@records) == 0)
      return(object)

    ## input validation
    .validate_class(slot, "character", null.ok = TRUE)
    if (!is.null(slot)) {
      valid.names <- slotNames(object@records[[1]])
      if (!all(slot %in% valid.names)) {
        .throw_error("Invalid 'slot' name, valid names are: ",
                     .collapse(valid.names))
      }
      if (length(slot) > 1) {
        message("[sort_RLum()]: Only the first 'slot' field will be used in sorting")
        slot <- slot[1]
      }
    }

    .validate_class(info_element, "character", null.ok = TRUE)
    if (!is.null(info_element)) {
      valid.names <- c(
        "XY_LENGTH", "NCOL", "X_MIN", "X_MAX", "Y_MIN", "Y_MAX",
        names(object@records[[1]]@info))
      if (!all(info_element %in% valid.names)) {
        .throw_error("Invalid 'info_element' name, valid names are: ",
                     .collapse(valid.names))
      }
    }

    if (is.null(slot) && is.null(info_element)) {
      .throw_error("At least one of 'slot' and 'info_element' should not be NULL")
    }

    .validate_class(decreasing, classes = "logical")
    if (anyNA(decreasing))
      .throw_error("'decreasing' should be of class 'logical'")

    ## recycle decreasing to match selection
    decreasing <- rep_len(decreasing, length(info_element) + 1)

    ## translate to -1 and 1 to match data.table requirements
    decreasing[decreasing] <- -1
    decreasing[!decreasing] <- 1

    ## extract the values from the records
    ## (1) extract slot values (should be a character; take only the first)
    SLOT <- if(!is.null(slot)) {
      data.table::as.data.table(
       unlist(lapply(object@records, function(x) slot(x, slot)[[1]])))

    } else {
      data.table::data.table(V1 = NA)
    }

    ## (2) extract info elements; ensure to take only the first element
    ## of the info element is a vector
    INFO <- if (!is.null(info_element)) {
      data.table::rbindlist(
        lapply(object@records, function(x) {
          data.table::as.data.table(lapply(x@info, function(l) l[[1]]))
        }),
        fill = TRUE)
      }

    ## (3) calculate general data parameters we always want to have
    EXTRA <- t(vapply(object@records, function(x) {
      ## ncol > 2 happens for RLum.Data.Spectrum and RLum.Data.Image
      n_col <- ncol(x@data)

      ## NA happens for RLum.Data.Image
      if(is.na(n_col))
        c(rep(NA_real_, 6))
      else
       c(nrow(x@data), n_col, min(x@data[,1]), max(x@data[,1]), min(x@data[,n_col]), max(x@data[,n_col]))

    }, numeric(6)))

    ## add UID and combine information
    ## the UID is required for the ordering index
    vals <- suppressWarnings(cbind(
      UID = seq_len(max(c(nrow(SLOT), nrow(INFO), nrow(EXTRA)))),
      SLOT = SLOT,
      XY_LENGTH = EXTRA[,1],
      NCOL = EXTRA[,2],
      X_MIN = EXTRA[,3],
      X_MAX = EXTRA[,4],
      Y_MIN = EXTRA[,5],
      Y_MAX = EXTRA[,6],
      INFO))

    ## determine the new ordering if possible
    ord <- data.table::setorderv(
      x = vals,
      cols = c("SLOT.V1", info_element),
      order = decreasing)[["UID"]]

    ## return reordered object
    return(object[ord])
  }
)

## melt_RLum() --------------------------------------------------------------
#' @describeIn melt_RLum
#' Melts [Luminescence::RLum.Analysis-class] objects into a flat data.frame with columns
#' `X`, `Y`, `TYPE`, `UID`, to be used in combination with other packages
#' such as `ggplot2`.
#'
#' @export
setMethod(
  f = "melt_RLum",
  signature = "RLum.Analysis",
  function(object) {
    melt_RLum(object@records)
  }
)

## view() -------------------------------------------------------------------
#' @describeIn view
#' View method for [Luminescence::RLum.Analysis-class] objects.
#'
#' @export
setMethod("view",
          signature = "RLum.Analysis",
          definition = function(object, ...) {
    .set_function_name("view")
    on.exit(.unset_function_name(), add = TRUE)

    .validate_not_empty(object)

    ## set title
    name <- list(...)$title
    if (is.null(name))
      name <- deparse(substitute(object))

    ## collect info lists from all records
    info <- lapply(seq_along(object@records),
                   function(x) c(aliquot = x, object@records[[x]]@info))
    info <- rbindlist(info, fill = TRUE)

    ## run view
    .view(x = info, title = name)
})

#'@title Dose-rate conversion function
#'
#' @description
#' This function converts radionuclide concentrations (K in %, Th and U in ppm)
#' into dose rates (Gy/ka). Beta-dose rates are also attenuated for the grain
#' size. Beta and gamma-dose rates are corrected for the water content. This
#' function converts concentrations into dose rates (Gy/ka) and corrects for
#' grain size attenuation and water content.
#'
#'Dose rate conversion factors can be chosen from Adamiec and Aitken (1998),
#'Guérin et al. (2011), Liritzis et al. (201) and Cresswell et al. (2018).
#'Default is Guérin et al. (2011).
#'
#'Grain size correction for beta dose rates is achieved using the correction
#'factors published by Guérin et al. (2012).
#'
#'Water content correction is based on factors provided by Aitken (1985),
#'with the factor for beta dose rate being 1.25 and for gamma 1.14.
#'
#'@details
#'
#'**The input data**
#'
#'\tabular{lll}{
#'COLUMN \tab DATA TYPE \tab DESCRIPTION\cr
#'`Mineral` \tab `character` \tab `'FS'` for feldspar, `'Q'` for quartz\cr
#'`K` \tab `numeric` \tab K nuclide content in %\cr
#'`K_SE` \tab `numeric` \tab error on K nuclide content in %\cr
#'`Th` \tab `numeric` \tab Th nuclide content in ppm\cr
#'`Th_SE` \tab `numeric` \tab error on Th nuclide content in ppm\cr
#'`U` \tab `numeric` \tab U nuclide content in ppm\cr
#'`U_SE` \tab `numeric` \tab error on U nuclide content in ppm\cr
#'`GrainSize` \tab `numeric` \tab average grain size in µm\cr
#'`WaterContent` \tab `numeric` \tab mean water content in %\cr
#'`WaterContent_SE` \tab `numeric` \tab relative error on water content
#'}
#'
#'**Water content**
#'The water content provided by the user should be calculated according to:
#'
#'\deqn{(Wet_weight - Dry_weight) / Dry_weight * 100}
#'
#'The unit for the weight is gram (g).
#'
#' @param object [data.frame] (*optional*):
#' a data frame containing all relevant information for each individual layer.
#' If nothing is provided, the function returns a template data frame, the
#' values of which should be filled in by the user. Please note that only one
#' dataset per input is supported.
#'
#' @param conversion [character] (*with default*):
#' dose rate conversion factors to use, by default those by Guérin et al. (2011).
#' For accepted values see [Luminescence::BaseDataSet.ConversionFactors].
#'
#' @param ... currently not used.
#'
#' @return
#' The function returns an [Luminescence::RLum.Results-class] object for which
#' the first element is a [matrix] with the converted values. If no input is
#' provided, the function returns a template [data.frame] that can be used as
#' input.
#'
#' @section Function version: 0.1.1
#'
#'@author Svenja Riedesel, Aberystwyth University (United Kingdom) \cr
#'Martin Autzen, DTU NUTECH Center for Nuclear Technologies (Denmark)
#'
#'@references
#'Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update. Ancient TL 16, 37-46.
#'
#'Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018. Dose rate conversion parameters:
#'Assessment of nuclear data. Radiation Measurements 120, 195-201.
#'
#'Guérin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion factors: update.
#'Ancient TL 29, 5-8.
#'
#'Guérin, G., Mercier, N., Nathan, R., Adamiec, G., Lefrais, Y., 2012. On the use
#'of the infinite matrix assumption and associated concepts: A critical review.
#'Radiation Measurements 47, 778-785.
#'
#'Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
#'A re-evaluation of radiation dose-rate conversion factors. Mediterranean
#'Archaeology and Archaeometry 13, 1-15.
#'
#'@keywords datagen
#'
#'@examples
#'
#'## create input template
#'input <- convert_Concentration2DoseRate()
#'
#'## fill input
#'input$Mineral <- "FS"
#'input$K <- 2.13
#'input$K_SE <- 0.07
#'input$Th <- 9.76
#'input$Th_SE <- 0.32
#'input$U <- 2.24
#'input$U_SE <- 0.12
#'input$GrainSize <- 200
#'input$WaterContent <- 30
#'input$WaterContent_SE <- 5
#'
#'## convert
#'convert_Concentration2DoseRate(input)
#'
#'@export
convert_Concentration2DoseRate <- function(
  object,
  conversion = "Guerinetal2011",
  ...
) {
  .set_function_name("convert_Concentration2DoseRate")
  on.exit(.unset_function_name(), add = TRUE)

  template <- data.frame(
      Mineral = NA_character_,
      K = NA_integer_,
      K_SE = NA_integer_,
      Th = NA_integer_,
      Th_SE = NA_integer_,
      U = NA_integer_,
      U_SE = NA_integer_,
      GrainSize = NA_integer_,
      WaterContent = NA_integer_,
      WaterContent_SE = NA_integer_)

  ## deprecated argument
  if ("input" %in% ...names()) {
    object <- list(...)$input
    .deprecated(old = "input", new = "object", since = "1.2.0")
  }

  ## return a template if no input is given
  if (missing(object)) {
    .throw_message("Input template returned, please fill this data frame ",
                   "and use it as input to the function", error = FALSE)
    return(template)
  }

  ## Load datasets ----------------------------------------------------------

  ## silence notes raised by R CMD check
  BaseDataSet.ConversionFactors <- BaseDataSet.GrainSizeAttenuation <- NA

  load(system.file("data", "BaseDataSet.ConversionFactors.rda",
                   package = "Luminescence"))
  load(system.file("data", "BaseDataSet.GrainSizeAttenuation.rda",
                   package = "Luminescence"))

  ## we do this to be consistent with the code written by Svenja and Martin
  GSA <- BaseDataSet.GrainSizeAttenuation

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("data.frame", "matrix"))

  if (ncol(object) != ncol(template))
    .throw_error("'object' should have ", ncol(template), " columns")
  if (nrow(object) > 1)
    .throw_error("'object' should have only one row")
  if (anyNA(object))
    .throw_error("'object' should not contain NA values")
  for (idx in 2:ncol(object)) {
    .validate_class(object[, idx], c("numeric", "integer"),
                    name = "Each element of 'object' other than the first")
  }
  if (!"GrainSize" %in% colnames(object))
    .throw_error("'object' should contain a 'GrainSize' column")

  ## conversion factors: we do not use BaseDataSet.ConversionFactors directly
  ## as it is in alphabetical level, but we want to have 'Guerinetal2011'
  ## in first position, as that is our default value
  valid_conversion_factors <- c("Guerinetal2011", "Cresswelletal2018",
                                "AdamiecAitken1998", "Liritzisetal2013")
  stopifnot(names(BaseDataSet.ConversionFactors) %in% valid_conversion_factors)
  conversion <- .validate_args(conversion, valid_conversion_factors)

  if (!any(object[, 1] %in% c("FS","Q")))
    .throw_error("As mineral only 'FS' or 'Q' is supported")

  ## Convert ----------------------------------------------------------------

  input <- object
    InfDR <- matrix(data = NA, nrow = 2, ncol = 6)
    colnames(InfDR) <- c("K","SE","Th","SE","U","SE")
    rownames(InfDR) <- c("Beta","Gamma")

    ### --- BETA DOSE RATES
    for (i in 1:3){
      if (i == 1){
        Col <- "K"
      } else if (i == 2){
        Col <- "Th"
      } else {
        Col <- "U"
      }

      for (j in 1:2){
        Temp <- if (j == 1) "beta" else "gamma"

        ConvFactor <- BaseDataSet.ConversionFactors[[conversion]][[Temp]][[Col]]
        Nuclide <- i * 2
        N <- 2 * i - 1
        Error <- Nuclide + 1

        ## calculate dose rate
        InfDR[j, N] <- input[1, Nuclide] * ConvFactor[1]

        # calculate error
        InfDR[j, Nuclide] <- sqrt((input[1, Error] / input[1, Nuclide])^2 +
                                  (ConvFactor[2] / ConvFactor[1])^2)
      }
    }

  ##### --- dose rate for grain size --- #####
  grain.size <- input$GrainSize[1]
  if (grain.size < min(GSA$GrainSize) || grain.size > max(GSA$GrainSize)) {
    .throw_error("No attenuation data available for the grain size provided (",
                 grain.size, ")")
  }

  if (input[1, 1] == "FS") {
    ## feldspar
    KFit <- approx(GSA$GrainSize, GSA$FS_K, xout = grain.size, method = "linear")
    ThFit <- approx(GSA$GrainSize, GSA$FS_Th, xout = grain.size, method = "linear")
    UFit <- approx(GSA$GrainSize, GSA$FS_U, xout = grain.size, method = "linear")

  } else if (input[1, 1] == "Q")  {
    ## quartz
    KFit <- approx(GSA$GrainSize, GSA$Q_K, xout = grain.size, method = "linear")
    ThFit <- approx(GSA$GrainSize, GSA$Q_Th, xout = grain.size, method = "linear")
    UFit <- approx(GSA$GrainSize, GSA$Q_U, xout = grain.size, method = "linear")
  }

  InfDR[1, 1] <- InfDR[1, 1] * (1 - KFit$y)   # K
  InfDR[1, 3] <- InfDR[1, 3] * (1 - ThFit$y)  # Th
  InfDR[1, 5] <- InfDR[1, 5] * (1 - UFit$y)   # U

    ##### --- Correct beta sediment dose rate for water content --- #####
    InfDRG <- matrix(data = NA, nrow = 2, ncol = 6)
    colnames(InfDRG) <- c("K", "SE", "Th", "SE", "U", "SE")
    rownames(InfDRG) <- c("Beta", "Gamma")

    WC <- input[1, 9] / 100
    WCerr <- input[1, 10] / 100

    for (i in 1:6){
      for (j in 1:2){
        if (j == 1){
          k <- 1.25 #Water content correction for beta
        } else {
          k <- 1.14 #Water content correction for gamma
        }

        Remain <- i %% 2

        if (Remain == 1){
          InfDRG[j,i] <- InfDR[j,i]/(1 + k*WC)
        } else if (Remain == 0){
          Temp <- (InfDR[j,i]/(1 + k*WC)) - (InfDR[j,i]/(1+k*(WC+WCerr)))
          InfDRG[j,i] <- InfDRG[j,i-1]*sqrt(InfDR[j,i]^2+(Temp/InfDR[j,i-1])^2)
        }
      }
    }
    InfDRG <- round(InfDRG, digits = 3)

# Return ------------------------------------------------------------------
  set_RLum(
      class = "RLum.Results",
      data = list(
        InfDRG = InfDRG,
        input_data = input
      ),
      info = list(
        call = sys.call()
      ))
}

## load data
df <- data.frame(Mineral = "FS",
                 K = 2.13, K_SE = 0.07,
                 Th = 9.76, Th_SE = 0.32,
                 U = 2.24, U_SE = 0.12,
                 GrainSize = 200, WaterContent = 30, WaterContent_SE = 5)

test_that("input validation", {
  testthat::skip_on_cran()

  ## template
  SW({
  template <- expect_s3_class(convert_Concentration2DoseRate(), "data.frame")
  })

  expect_error(convert_Concentration2DoseRate("fail"),
               "'object' should be of class 'data.frame' or 'matrix'")
  expect_error(convert_Concentration2DoseRate(data.frame(x = 1, y = 2)),
               "'object' should have 10 columns")
  expect_error(convert_Concentration2DoseRate(rbind(df, df)),
               "'object' should have only one row")
  expect_error(convert_Concentration2DoseRate(template),
               "'object' should not contain NA values")
  expect_error(convert_Concentration2DoseRate(df, conversion = "error"),
               "'conversion' should be one of 'Guerinetal2011', 'Cresswelletal2018'")

  df[[8]] <- "error"
  expect_error(convert_Concentration2DoseRate(df),
               "Each element of 'object' other than the first should be of class")
  df[[8]] <- 9999
  expect_error(convert_Concentration2DoseRate(df),
               "No attenuation data available for the grain size provided")
  df[[1]] <- "fail"
  expect_error(convert_Concentration2DoseRate(df),
               "As mineral only 'FS' or 'Q' is supported")
  colnames(df)[8] <- "error"
  expect_error(convert_Concentration2DoseRate(df),
               "'object' should contain a 'GrainSize' column")
})

test_that("deprecated argument", {
  testthat::skip_on_cran()

  ## deprecated argument
  expect_warning(convert_Concentration2DoseRate(input = df),
                 "'input' was deprecated in v1.2.0, use 'object' instead")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  df$Mineral <- "FS"
  expect_snapshot_RLum(convert_Concentration2DoseRate(df))

  df$Mineral <- "Q"
  expect_snapshot_RLum(convert_Concentration2DoseRate(df))

  ## issue 1424
  df <- data.frame(Mineral = "Q", K = 0.09, K_SE = 0.009,
                   Th = 2.4, Th_SE = 0.2, U = 0.75, U_SE = 0.07,
                   GrainSize = 617.5, WaterContent = 8, WaterContent_SE = 4)
  expect_snapshot_RLum(convert_Concentration2DoseRate(df, "Cresswelletal2018"))
})

#' General accessor function for RLum S4 class objects
#'
#' Function calls object-specific get functions for RisoeBINfileData S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [Risoe.BINfileData-class] objects.\cr 
#' Depending on the input object, the corresponding get function will be selected. 
#' Allowed arguments can be found in the documentations of the corresponding 
#' [Risoe.BINfileData-class] class.
#'
#' @param METADATA x
#' 
#' @param DATA x
#' 
#' @param .RESERVED x
#'
#' @return Return is the same as input objects as provided in the list.
#' 
#' @section Function version: 0.1
#' 
#' @author 
#' Sebastian Kreutzer, Geography & Earth Sciences, Aberystwyth University (United Kingdom)
#' 
#' @seealso [Risoe.BINfileData-class]
#' 
#' @keywords utilities
#'
#' @md
#' @export
setGeneric("set_Risoe.BINfileData",
  function(METADATA = data.frame(), DATA = list(), .RESERVED = list()) {
    standardGeneric("set_Risoe.BINfileData")
  },
  package = "Luminescence"
)

#' @title Apply the model after Fuchs & Lang (2001) to a given De distribution
#'
#' @description
#' This function applies the method according to Fuchs & Lang (2001) for
#' heterogeneously bleached samples with a given coefficient of variation
#' threshold.
#'
#' @details
#'
#' **Used values**
#'
#' If the coefficient of variation (`c[v]`) of the first
#' two values is larger than the threshold `c[v_threshold]`, the first value is
#' skipped.  Use the `startDeValue` argument to define a start value for
#' calculation (e.g. 2nd or 3rd value).
#'
#' **Basic steps of the approach**
#'
#' 1. Estimate natural relative variation of the sample using a dose recovery test
#' 2. Sort the input values in ascending order
#' 3. Calculate a running mean, starting with the lowermost two values and add values iteratively.
#' 4. Stop if the calculated `c[v]` exceeds the specified `cvThreshold`
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with De `(data[,1])` and De error `(values[,2])`
#'
#' @param cvThreshold [numeric] (*with default*):
#' coefficient of variation in percent, as threshold for the method,
#' e.g. `cvThreshold = 3`. See details
#' .
#' @param startDeValue [numeric] (*with default*):
#' number of the first aliquot that is used for the calculations
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments and graphical parameters passed to [plot]
#'
#' @return
#' Returns a plot (*optional*) and terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following elements:
#'
#' \item{summary}{[data.frame] summary of all relevant model results.}
#' \item{data}{[data.frame] original input data}
#' \item{args}{[list] used arguments}
#' \item{call}{[call] the function call}
#' \item{usedDeValues}{[data.frame] containing the used values for the calculation}
#'
#' @note Please consider the requirements and the constraints of this method
#' (see Fuchs & Lang, 2001)
#'
#' @section Function version: 0.4.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Christoph Burow, University of Cologne (Germany)
#'
#' @seealso [plot], [Luminescence::calc_MinDose], [Luminescence::calc_FiniteMixture],
#' [Luminescence::calc_CentralDose], [Luminescence::calc_CommonDose], [Luminescence::RLum.Results-class]
#'
#' @references
#' Fuchs, M. & Lang, A., 2001. OSL dating of coarse-grain fluvial
#' quartz using single-aliquot protocols on sediments from NE Peloponnese,
#' Greece. In: Quaternary Science Reviews 20, 783-787.
#'
#' Fuchs, M. & Wagner, G.A., 2003. Recognition of insufficient bleaching by
#' small aliquots of quartz for reconstructing soil erosion in Greece.
#' Quaternary Science Reviews 22, 1161-1167.
#'
#' @keywords dplot
#'
#'
#' @examples
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## calculate De according to Fuchs & Lang (2001)
#' temp<- calc_FuchsLang2001(ExampleData.DeValues$BT998, cvThreshold = 5)
#'
#' @export
calc_FuchsLang2001 <- function(
  data,
  cvThreshold = 5,
  startDeValue = 1,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_FuchsLang2001")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("data.frame", "RLum.Results"))
  .validate_not_empty(data)
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")
  }
  if (ncol(data) < 2) {
    .throw_error("'data' should have 2 columns")
  }
  .validate_positive_scalar(cvThreshold)
  .validate_positive_scalar(startDeValue, int = TRUE)
  .validate_logical_scalar(plot)

  # Deal with extra arguments -----------------------------------------------
  ##deal with addition arguments
  extraArgs <- list(...)
  verbose <- extraArgs$verbose %||% TRUE
  .validate_logical_scalar(verbose)

  ##============================================================================##
  ##PREPARE DATA
  ##============================================================================##

  ##1. order values in ascending order write used D[e] values in data.frame
  o <- order(data[[1]]) # o is only an order parameter
  data_ordered <- data[o,] # sort values after o and write them into a new variable

  ##2. estimate D[e]
  # set variables
  usedDeValues <- data.frame(De = NA, De_Error = NA, cv = NA)
  endDeValue <- startDeValue[1]

  # if the first D[e] values are not used write this information in the data.frame
  if (startDeValue[1] != 1) {
    n <- abs(1 - startDeValue[1])

    #  write used D[e] values in data.frame
    usedDeValues[1:n, 1] <- data_ordered[1:n, 1]
    usedDeValues[1:n, 2] <- data_ordered[1:n, 2]
    usedDeValues[1:n, 3] <- "skipped"
  }

  ##=================================================================================================##
  ##LOOP FOR MODEL
  ##=================================================================================================##
  # repeat loop (run at least one time)
  repeat {
    #calculate mean, sd and cv
    mean<-round(mean(data_ordered[startDeValue:endDeValue,1]),digits=2) #calculate mean from ordered D[e] values
    sd<-round(sd(data_ordered[startDeValue:endDeValue,1]),digits=2)		#calculate sd from ordered D[e] values
    cv <- round(sd / mean * 100, digits = 2) #calculate coefficient of variation

    ## avoid crashes if the both mean and sd are zero
    if (is.na(cv))
      cv <- 0

    # break if cv > cvThreshold
    if (cv > cvThreshold[1] & endDeValue > startDeValue) {
      # if the first two D[e] values give a cv > cvThreshold, than skip the first D[e] value
      if (endDeValue-startDeValue<2) {
        #  write used D[e] values in data.frame
        usedDeValues[endDeValue, 1] <- data_ordered[endDeValue, 1]
        usedDeValues[endDeValue, 2] <- data_ordered[endDeValue, 2]
        usedDeValues[endDeValue - 1, 3] <- "not used"

        # go to the next D[e] value
        startDeValue <- startDeValue + 1

      } else {
        usedDeValues[endDeValue, 1] <- data_ordered[endDeValue, 1]
        usedDeValues[endDeValue, 2] <- data_ordered[endDeValue, 2]
        usedDeValues[endDeValue, 3] <- paste("#", cv, "%")

        break #break loop
      }

    }#EndIf
    else {

      # write used D[e] values in data.frame
      usedDeValues[endDeValue,1]<-data_ordered[endDeValue,1]
      usedDeValues[endDeValue,2]<-data_ordered[endDeValue,2]
      usedDeValues[endDeValue,3] <- paste(cv, "%")

    }#EndElse

    # go the next D[e] value until the maximum number is reached
    if (endDeValue<length(data_ordered[,1])) {
      endDeValue<-endDeValue+1
    } else {break}

  }#EndRepeat

  ##=================================================================================================##
  ##ADDITIONAL CALCULATIONS and TERMINAL OUTPUT
  ##=================================================================================================##

  # additional calculate weighted mean
  w <- 1 / (data_ordered[startDeValue:endDeValue, 2]) ^ 2 #weights for weighted mean
  weighted_mean <- round(stats::weighted.mean(data_ordered[startDeValue:endDeValue, 1], w),
                         digits = 2)
  weighted_sd <- round(sqrt(1 / sum(w)), digits = 2)
  n.usedDeValues <- endDeValue - startDeValue + 1

  # standard error
  se <- round(sd / sqrt(endDeValue - startDeValue + 1), digits = 2)

  if(verbose){
    cat("\n[calc_FuchsLang2001]")
    cat("\n\n----------- meta data --------------")
    cat("\n cvThreshold:            ", cvThreshold[1], "%")
    cat("\n used values:            ", n.usedDeValues)
    cat("\n----------- dose estimate ----------")
    cat("\n mean:                   ", mean)
    cat("\n sd:                     ", sd)
    cat("\n weighted mean:          ", weighted_mean)
    cat("\n weighted sd:            ", weighted_sd)
    cat("\n se:                     ", se)
    cat("\n------------------------------------\n\n")
  }

  ##===========================================================================#
  ##RETURN  VALUES
  ##==========================================================================##
  summary <- data.frame(
    de = mean,
    de_err = sd,
    de_weighted = weighted_mean,
    de_weighted_err = weighted_sd,
    se = se,
    n.usedDeValues = n.usedDeValues
  )

  args <- list(cvThreshold = cvThreshold, startDeValue = startDeValue)
  newRLumResults.calc_FuchsLang2001 <- set_RLum(
    class = "RLum.Results",
    data = list(
      summary = summary,
      data = data,
      args = args,
      usedDeValues = usedDeValues
    ),
    info = list(call = sys.call())
  )

  ##=========##
  ## PLOTTING
  if(plot) {
    try(plot_RLum.Results(newRLumResults.calc_FuchsLang2001, ...),
        outFile = stdout()) # redirect error messages so they can be silenced
  }#endif::plot

  invisible(newRLumResults.calc_FuchsLang2001)
}

test_that("check class", {
  testthat::skip_on_cran()
  local_edition(3)

  object <- set_RLum(class = "RLum.Data.Curve")
  expect_length(rep(object, 10), 10)


  ## <RLum.Results>
  data(ExampleData.DeValues, envir = environment())
  temp <- calc_FuchsLang2001(ExampleData.DeValues$BT998, cvThreshold = 5, plot = FALSE)
  expect_output(show(temp))

  ## test coercion
  expect_s4_class(as(as(object = temp, Class = "list"), "RLum.Results"), "RLum.Results")
})

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_FuchsLang2001("error"),
               "[calc_FuchsLang2001()] 'data' should be of class 'data.frame'",
               fixed = TRUE)
  expect_error(calc_FuchsLang2001(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(calc_FuchsLang2001(iris[, 1, drop = FALSE]),
               "'data' should have 2 columns")
  expect_error(calc_FuchsLang2001(ExampleData.DeValues$BT998, "error"),
               "'cvThreshold' should be a single positive value")
  expect_error(calc_FuchsLang2001(ExampleData.DeValues$BT998, startDeValue = numeric()),
               "'startDeValue' should be a single positive integer value")
  expect_error(calc_FuchsLang2001(ExampleData.DeValues$BT998, plot = NA),
               "'plot' should be a single logical value")
  expect_error(calc_FuchsLang2001(ExampleData.DeValues$BT998, verbose = NA),
               "'verbose' should be a single logical value")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(temp <- calc_FuchsLang2001(data = ExampleData.DeValues$BT998,
                                          cvThreshold = 5,
                                          plot = FALSE,
                                          verbose = FALSE),
                       tolerance = snapshot.tolerance)

  ## using an RLum.Results object as input
  SW({
  expect_snapshot_RLum(calc_FuchsLang2001(data = temp, startDeValue = 24,
                                          plot = FALSE),
                       tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              calc_FuchsLang2001(ExampleData.DeValues$BT998))
  vdiffr::expect_doppelganger("cvDefault startDeValue",
                              calc_FuchsLang2001(ExampleData.DeValues$BT998,
                                                 cvDefault = 3,
                                                 startDeValue = 5))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 818
  expect_silent(calc_FuchsLang2001(data.frame(ED = c(rep(0, 4), 10),
                                              ED_Error = rnorm(5) + 1),
                                   verbose = FALSE))

  ## issue 917
  expect_output(calc_FuchsLang2001(data.frame(ED = NA,
                                              ED_Error = rnorm(5) + 1),
                                   verbose = FALSE),
                "need finite 'xlim' values")
})

data(ExampleData.DeValues, envir = environment())
obj <- calc_FuchsLang2001(ExampleData.DeValues$BT998, cvThreshold = 5,
                          plot = FALSE, verbose = FALSE)
empty <- set_RLum("RLum.Results")

test_that("check class", {
  testthat::skip_on_cran()

  ## set_RLum()
  expect_s4_class(obj, "RLum.Results")

  ## as()
  expect_type(as(obj, "list"),
              "list")
  expect_s4_class(as(list(), "RLum.Results"),
                  "RLum.Results")

  ## names()
  expect_equal(names_RLum(obj),
               c("summary", "data", "args", "usedDeValues"))
})

test_that("set_RLum", {
  testthat::skip_on_cran()

  ## input validation
  expect_error(set_RLum(obj, TRUE),
               "'class' should be of class 'character'")
})

test_that("get_RLum", {
  testthat::skip_on_cran()

  ## input validation
  expect_error(get_RLum(obj, "error"),
               "Unknown 'data.object', valid names are:")
  expect_error(get_RLum(obj, FALSE),
               "'data.object' should be of class 'character' or 'numeric'")
  expect_error(get_RLum(obj, 100),
               "'data.object' index out of bounds")
  expect_error(get_RLum(obj, info.object = 1:2),
               "'info.object' should be of class 'character' or NULL and have length 1")
  expect_warning(expect_null(get_RLum(obj, info.object = "error")),
                 "[get_RLum()] Invalid 'info.object' name, valid names are:",
                 fixed = TRUE)
  expect_warning(expect_null(get_RLum(empty, info.object = "error")),
                 "[get_RLum()] This 'RLum.Results' object has no info objects",
                 fixed = TRUE)
  expect_error(get_RLum(obj, drop = NA),
               "'drop' should be a single logical value")

  ## full functionality
  expect_s3_class(get_RLum(obj),
                  "data.frame")
  expect_s3_class(get_RLum(obj, data.object = 2),
                  "data.frame")
  expect_s3_class(get_RLum(obj, data.object = c(1, 2)),
                  "data.frame")
  expect_s3_class(get_RLum(obj, data.object = "summary"),
                  "data.frame")
  expect_s3_class(get_RLum(obj, data.object = c("summary", "data")),
                  "data.frame")
  expect_s4_class(get_RLum(obj, data.object = c("summary", "data"), drop = FALSE),
                  "RLum.Results")
  expect_type(get_RLum(obj, info.object = "call"),
              "list")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  ## show()
  SW({
  expect_snapshot_output(show(obj))
  expect_snapshot_output(show(empty))
  })
})

#' Class `"RLum"`
#'
#' Abstract class for data in the package Luminescence
#' Subclasses are:
#'
#' **RLum-class**\cr
#' |\cr
#' |----[Luminescence::RLum.Data-class]\cr
#' |----|-- [Luminescence::RLum.Data.Curve-class]\cr
#' |----|-- [Luminescence::RLum.Data.Spectrum-class]\cr
#' |----|-- [Luminescence::RLum.Data.Image-class]\cr
#' |----[Luminescence::RLum.Analysis-class]\cr
#' |----[Luminescence::RLum.Results-class]
#'
#' @name RLum-class
#'
#' @docType class
#'
#' @slot originator
#' Object of class [character] containing the name of the producing
#' function for the object. Set automatically by using the function [Luminescence::set_RLum].
#'
#' @slot info
#' Object of class [list] for additional information on the object itself
#'
#' @slot .uid
#' Object of class [character] for a unique object identifier. This id is
#' usually calculated using the internal function `create_UID` if the function [Luminescence::set_RLum]
#' is called.
#'
#' @slot .pid
#' Object of class [character] for a parent id. This allows nesting RLum-objects
#' at will. The parent id can be the uid of another object.
#'
#' @note `RLum` is a virtual class.
#'
#' @section Objects from the Class:
#' A virtual Class: No objects can be created from it.
#'
#' @section Class version: 0.4.0
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Data-class], [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Data.Image-class],
#' [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Results-class], [Luminescence::methods_RLum]
#'
#' @keywords classes
#'
#' @examples
#'
#' showClass("RLum")
#'
#' @export
setClass("RLum",
           slots = list(
             originator = "character",
             info = "list",
             .uid = "character",
             .pid = "character"
             ),
           contains = "VIRTUAL",
           prototype = prototype(
             originator = NA_character_,
             info = list(),
             .uid = NA_character_,
             .pid = NA_character_
           )
         )


## replicate_RLum() ---------------------------------------------------------
#' @describeIn replicate_RLum
#' Replication method for [Luminescence::RLum-class] objects.
#'
#' @export
setMethod(
  "replicate_RLum",
  "RLum",
  definition = function(object, times = 1) {
    .set_function_name("replicate_RLum")
    on.exit(.unset_function_name(), add = TRUE)

    .validate_positive_scalar(times, int = TRUE)
    lapply(1:times, function(x) object)
  }
)

#' @title Import Thermochronometry Data
#'
#' @description
#' Import data from thermochronometry experiments into R.
#' This function is an adaption of the script `STAGE1, ExcelToStructure` by
#' Benny Guralnik, 2014, modified to accept CSV files with the same structure
#' as the original Excel files.
#'
#' @param file [character] (**required**):
#' path to a CSV file; alternatively a [vector] of paths.
#'
#' @param output_type [character] (*with default*):
#' return type for the function, either [Luminescence::RLum.Results-class] (default) or `"list"`
#' (for a plain R list).
#'
#' @returns
#' Depending on the setting of `output_type` it will be either a plain R [list]
#' or an [Luminescence::RLum.Results-class] object with the following structure data elements:
#'
#' `$data:`
#' `.. $ITL`: data frame with columns `SAMPLE`, `TEMP`, `TIME`, `LxTx`, `LxTx_ERROR`
#' `.. $DRC`: data frame with columns `SAMPLE`, `ALQ`, `TIME`, `LxTx`, `LxTx_ERROR`
#' `.. $FAD`: data frame with columns `SAMPLE`, `ALQ`, `TIME`, `LxTx`, `LxTx_ERROR`
#'
#' This refers to `$ITL`: Isothermal curves, `$DRC`: Dose-response curve,
#' `$FAD`: Fading.
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @noRd
.import_ThermochronometryData <- function(
  file,
  output_type = "RLum.Results"
) {
# Helper functions -------------------------------------------------------
  ## consistently extract numerical data
  .extract_numerics <- function(x) {
    tmp <- suppressWarnings(as.numeric(na.exclude(as.numeric(x))))
    if(length(tmp) == 0)
      return(NA)
    tmp
  }

  ## Integrity checks -------------------------------------------------------
  file <- .validate_file(file, ext = "csv", scan.dir = FALSE)
  .validate_args(output_type, c("RLum.Results", "list"))

  ## define variable
  ka <- 1e+3 * .const$year_s # ka in seconds

  ## Import -----------------------------------------------------------------
  records <- lapply(file, function(x) {
    ## read the files separating header and body
    .validate_file(x, ext = "csv", scan.dir = FALSE)
    head <- data.table::fread(x, nrows = 3, select = 1:5)
    body <- data.table::fread(x, skip = 3, header = TRUE)

    list(
        id = grep("\\.\\.\\.[0-9]+", colnames(head)[-1], invert = TRUE,
                  value = TRUE),
        natT = .extract_numerics(head[1, -1]),         # natural temperature
        natDdot = .extract_numerics(head[2, -1]) / ka, # natural dose rate
        T = body[seq(1, nrow(body), 2)][[2]],          # temperature
        Ddot = body[seq(2, nrow(body), 2)][[2]],       # instrument dose rate
        rawdata = lapply(seq(1, nrow(body), 2), function(y) {
          data.table(SAMPLE = basename(tools::file_path_sans_ext(x)),
                     ALQ = ceiling(y / 2),
                     TEMP = body[y][[2]],
                     ## measurement time (irradiation or delay time)
                     TIME = .extract_numerics(body[y, -(1:4)]) * 1e+3,
                     LxTx = .extract_numerics(body[y + 1, -(1:4)]))
        })
    )
  })
  names(records) <- basename(tools::file_path_sans_ext(file))

  # Create output -----------------------------------------------------------
  if (output_type == "RLum.Results") {
    ## create data frame for each data type

    ## we will use the temperature to discriminate the records; everything
    ## with temperature < 15 is either for DRC or FAD, the rest ITL.
    ## here we save the list index of each record type so that we can access
    ## them later
    res <- lapply(seq_along(records), function(idx) {
      x <- records[[idx]]
      tmp <- cumsum(x$T > 15)
      idx_DRC <- which(tmp == 0)
      idx_ITL <- setdiff(which(!duplicated(tmp)), idx_DRC)
      idx_FAD <- setdiff(which(tmp == max(tmp)), idx_ITL)

      ## extract variables
      DRC <- data.table::rbindlist(x$rawdata[idx_DRC])
      ITL <- data.table::rbindlist(x$rawdata[idx_ITL])
      FAD <- data.table::rbindlist(x$rawdata[idx_FAD])

      ## adjust aliquot numbers to start from 1 in each block
      DRC$ALQ <- DRC$ALQ - DRC$ALQ[1] + 1
      FAD$ALQ <- FAD$ALQ - FAD$ALQ[1] + 1

      list(DRC = data.frame(DRC[, -3], # drop TEMP
                            LxTx_ERROR = NA),
           ITL = data.frame(ITL[, -2], # drop ALQ
                            LxTx_ERROR = NA),
           FAD = data.frame(FAD[, -3], # drop TEMP
                            LxTx_ERROR = NA),
           ## Ddot is only relevant for DRC data
           Ddot_DRC = x$Ddot[idx_DRC]
           )
    })

    DRC <- as.data.frame(data.table::rbindlist(lapply(res, function(x) x$DRC)))
    ITL <- as.data.frame(data.table::rbindlist(lapply(res, function(x) x$ITL)))
    FAD <- as.data.frame(data.table::rbindlist(lapply(res, function(x) x$FAD)))
    Ddot_DRC <- lapply(res, function(x) x$Ddot_DRC)
    nat_Ddot <- lapply(records, function(x) x$natDdot)

    ## create RLum.Results object
    records <- set_RLum(
      class = "RLum.Results",
      data = list(
        DRC = DRC,
        ITL = ITL,
        FAD = FAD),
      info = list(
        call = sys.call(),
        sample_names = unique(names(records)),
        Ddot_DRC = Ddot_DRC,
        nat_Ddot = nat_Ddot)
      )
  }

  ## always return records
  return(records)
}

input.xls <- test_path("_data/CLBR.xlsx")
input.csv <- file.path(test_path("_data"),
                       paste0("CLBR_IR", c(50, 100, 150, 225), ".csv"))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(.import_ThermochronometryData(list()),
               "'file' cannot be an empty list")
  expect_error(.import_ThermochronometryData("filename", output_type = "error"),
               "File 'filename' does not exist")
  expect_error(.import_ThermochronometryData(input.xls),
               "File extension 'xlsx' is not supported, only 'csv' is valid")
  expect_error(.import_ThermochronometryData(c(input.csv, input.xls)),
               "File extension 'xlsx' is not supported, only 'csv' is valid")
  expect_error(.import_ThermochronometryData("error"),
               "File 'error' does not exist")
  expect_error(.import_ThermochronometryData(c(input.csv, "error")),
               "File 'error' does not exist")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_snapshot_RLum(.import_ThermochronometryData(input.csv))
  expect_type(.import_ThermochronometryData(input.csv, output_type = "list"),
              "list")
})

#' @title Merge function for RLum.Data.Curve S4 class objects
#'
#' @description This function allows to merge [Luminescence::RLum.Data.Curve-class] objects in different
#' ways without modifying the original objects. Merging is always applied on
#' the 2nd column of the object's data matrix.
#'
#' **Supported merge operations are:**
#'
#' `"mean"` (default)
#'
#' The mean over the count values is calculated using the function
#' [rowMeans].
#'
#' `"median"`
#'
#' The median over the count values is calculated using the function
#' [matrixStats::rowMedians].
#'
#' `"sum"`
#'
#' All count values will be summed up using the function [rowSums].
#'
#' `"sd"`
#'
#' The standard deviation over the count values is calculated using the function
#' [matrixStats::rowSds].
#'
#' `"var"`
#'
#' The variance over the count values is calculated using the function
#' [matrixStats::rowVars].
#'
#' `"min"`
#'
#' The min values from the count values is chosen using the function
#' [matrixStats::rowMins][matrixStats::rowRanges].
#'
#' `"max"`
#'
#' The max values from the count values is chosen using the function
#' [matrixStats::rowMins][matrixStats::rowRanges].
#'
#' `"append"`
#'
#' Appends count values of all curves to one combined data curve. The channel width
#' is automatically re-calculated, but requires a constant channel width of the
#' original data.
#'
#' `"-"`
#'
#' The row sums of the last objects are subtracted from the first object.
#'
#' `"*"`
#'
#' The row sums of the last objects are multiplied with the first object.
#'
#' `"/"`
#'
#' Values of the first object are divided by row sums of the last objects.
#'
#' @param object [list] of [Luminescence::RLum.Data.Curve-class] (**required**):
#' list of objects to be merged.
#'
#' @param merge.method [character] (**required**):
#' method for combining of the objects, e.g. `'mean'` (default), `'median'`,
#' `'sum'`, see details for
#' further information and allowed methods.  Note: Elements in slot info will
#' be taken from the first curve in the list.
#'
#' @param method.info [numeric] (*optional*):
#' allows to specify how info elements of the input objects are combined,
#' e.g. `1` means that just the elements from the first object are kept,
#' `2` keeps only the info elements from the 2 object etc.
#' If set to `NULL`, all elements are combined.
#'
#' @return Returns an [Luminescence::RLum.Data.Curve-class] object.
#'
#' @note
#' The information from the slot `recordType` is taken from the first
#' object in the input list. The slot
#' 'curveType' is filled with the name `merged`.
#'
#' @section S3-generic support:
#'
#' This function is fully operational via S3-generics:
#' ``+``, ``-``, ``/``, ``*``, `merge`
#'
#' @section Function version: 0.2.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::merge_RLum], [Luminescence::RLum.Data.Curve-class]
#'
#' @keywords utilities internal
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.XSYG, envir = environment())
#'
#' ##grep first and 3d TL curves
#' TL.curves  <- get_RLum(OSL.SARMeasurement$Sequence.Object, recordType = "TL (UVVIS)")
#' TL.curve.1 <- TL.curves[[1]]
#' TL.curve.3 <- TL.curves[[3]]
#'
#' ##plot single curves
#' plot_RLum(TL.curve.1)
#' plot_RLum(TL.curve.3)
#'
#' ##subtract the 1st curve from the 2nd and plot
#' TL.curve.merged <- merge_RLum.Data.Curve(list(TL.curve.3, TL.curve.1), merge.method = "/")
#' plot_RLum(TL.curve.merged)
#'
#' @export
merge_RLum.Data.Curve<- function(
  object,
  merge.method = c("mean", "median", "sum", "sd", "var", "max", "min",
                   "append", "-", "*", "/"),
  method.info = NULL
) {
  .set_function_name("merge_RLum.Data.Curve")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "list")
  .validate_positive_scalar(method.info, int = TRUE, null.ok = TRUE)

  ##(1) check if object is of class RLum.Data.Curve
  temp.recordType.test <- sapply(object, function(x) {
    .validate_class(x, "RLum.Data.Curve",
                    name = "All elements of 'object'")
    return(x@recordType)
  })

  ##(2) Check for similar record types
  record.types <- unique(temp.recordType.test)
  if (length(record.types) > 1) {
    .throw_error("Objects cannot be merged, different record types found: ",
                 .collapse(record.types))
  }

  merge.method <- .validate_args(merge.method,
                                 c("mean", "median", "sum", "sd", "var", "max",
                                   "min", "append", "-", "*", "/"))

  ## Merge objects ----------------------------------------------------------
  ##merge data objects
  ##problem ... how to handle data with different resolution or length?

  ##(1) build new data matrix
  ## first find the shortest object
  check.rows <- vapply(object, function(x) nrow(x@data), numeric(1))
  if (length(check.rows) < 1 || min(check.rows) < 2) {
    .throw_error("'object' contains no data")
  }
  num.rows <- min(check.rows)

  ## channel resolution of the first object: we need to round as there may
  ## otherwise be numerical artefacts that would make the step not unique
  step <- round(diff(object[[1]]@data[, 1]), 1)[1]

  ## extract the curve values from each object
  temp.matrix <- sapply(object, function(x) {
    ## check the resolution (roughly)
    if (round(diff(x@data[, 1]), 1)[1] != step)
      .throw_warning("The curves do not seem to have the same channel resolution")
    ## limit all objects to the shortest one
    x@data[1:num.rows, 2]
  })

  ## throw the warning only now to avoid printing it in case of error
  if (length(unique(check.rows)) != 1) {
    .throw_warning("The number of channels differs between the curves, the ",
                   "merged curve will have the length of the shortest object")
  }

  ##(2) apply selected method for merging
  temp.matrix <- switch(merge.method,
                        sum = rowSums(temp.matrix),
                        mean = rowMeans(temp.matrix),
                        median = matrixStats::rowMedians(temp.matrix),
                        sd = matrixStats::rowSds(temp.matrix),
                        var = matrixStats::rowVars(temp.matrix),
                        max = matrixStats::rowMaxs(temp.matrix),
                        min = matrixStats::rowMins(temp.matrix),
                        append = sapply(temp.matrix, c),
                        "-" = {
                          temp.matrix[, 1] - rowSums(temp.matrix[, -1, drop = FALSE])
                        },
                        "*" = {
                          temp.matrix[, 1] * rowSums(temp.matrix[, -1, drop = FALSE])
                        },
                        "/" = {
                          temp <- temp.matrix[, 1] / rowSums(temp.matrix[, -1, drop = FALSE])

                          ## replace infinities with 0 and throw warning
                          id.inf <- which(is.infinite(temp))
                          if (length(id.inf) > 0) {
                            temp[id.inf]  <- 0
                            .throw_warning(length(id.inf),
                                           " 'Inf' values replaced by 0 in the matrix")
                          }
                          temp
                        })

  ## add back the first column to RLum.Data.Curve objects
  #If we append the data of the second to the first curve we have to recalculate
  #the x-values (probably time/channel). The difference should always be the
  #same, so we just expand the sequence if this is true. If this is not true,
  #we revert to the default behaviour (i.e., append the x values)
  if (merge.method == "append") {
    newx <- seq(from = min(object[[1]]@data[, 1]), by = step,
                length.out = sum(check.rows))
    temp.matrix <- cbind(newx, temp.matrix)
  } else {
    temp.matrix <- cbind(object[[1]]@data[1:num.rows, 1], temp.matrix)
  }

  ## remove spurious column names added by cbind()
  temp.matrix <- unname(temp.matrix)

  ##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ##merge info objects as simple as possible ... just keep them all ... other possibility
  ##would be to choose on the input objects

  ##unlist is needed here, as otherwise it would cause unexpected behaviour further using
  ##the RLum.object
  if (is.null(method.info)) {
    temp.info <- unlist(lapply(object, function(x) x@info), recursive = FALSE)
  }else{
    temp.info <- object[[method.info]]@info
  }

  ## Build new RLum.Data.Curve object ---------------------------------------
  set_RLum(
    class = "RLum.Data.Curve",
    originator = "merge_RLum.Data.Curve",
    recordType = object[[1]]@recordType,
    curveType =  "merged",
    data = temp.matrix,
    info = temp.info,
    .pid = unlist(lapply(object, function(x) x@.uid))
  )
}

test_that("Merge tests", {
  testthat::skip_on_cran()
  local_edition(3)

  ##load example data
  data(ExampleData.XSYG, envir = environment())
  TL.curves  <- get_RLum(OSL.SARMeasurement$Sequence.Object, recordType = "TL (UVVIS)")
  TL.curve.1 <- TL.curves[[1]]
  TL.curve.3 <- TL.curves[[3]]
  TL.curve.3_short <- TL.curves[[3]]
  TL.curve.3_short@data <-  TL.curve.3_short@data[1:(nrow(TL.curve.3_short@data) - 1),]

  ##check for error
  expect_error(merge_RLum.Data.Curve("", merge.method = "/"))

  ## check warning for different curve lengths
  expect_warning(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3_short), merge.method = "mean"))

  ##check error for different resolution
  TL.curve.3_short@data <-  TL.curve.3_short@data[-2,]
  expect_error(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3_short), merge.method = "mean"),
          "\\[merge\\_RLum.Data.Curve\\(\\)\\] The objects do not seem to have the same channel resolution!")

  ##check various operations
  expect_s4_class(TL.curve.1 + TL.curve.3, "RLum.Data.Curve")
  expect_s4_class(TL.curve.1 - TL.curve.3, "RLum.Data.Curve")
  expect_s4_class(suppressWarnings(TL.curve.3 / TL.curve.1), "RLum.Data.Curve")
  expect_warning(TL.curve.3 / TL.curve.1)
  expect_s4_class(TL.curve.1 * TL.curve.3, "RLum.Data.Curve")

})

## load example data
data(ExampleData.XSYG, envir = environment())

## RLum.Data.Curve
TL.curves  <- get_RLum(OSL.SARMeasurement$Sequence.Object,
                       recordType = "TL (UVVIS)")
TL.curve.1 <- TL.curves[[1]]
TL.curve.3 <- TL.curves[[3]]
TL.curve.3_short <- TL.curves[[3]]
TL.curve.3_short@data <- TL.curve.3_short@data[1:(nrow(TL.curve.3@data) - 1), ]
TL.curve.3_resol <- TL.curve.3_short
TL.curve.3_resol@data <- TL.curve.3_resol@data[-2, ]
TL.curve.3_types <- TL.curves[[3]]
TL.curve.3_types@recordType <- "IRSL"
TL.curve.3_zeros <- TL.curves[[3]]
TL.curve.3_zeros@data[10:12, 2] <- 0

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(merge_RLum.Data.Curve("", merge.method = "/"),
               "'object' should be of class 'list'")
  expect_error(merge_RLum.Data.Curve(list("")),
               "All elements of 'object' should be of class 'RLum.Data.Curve'")
  expect_error(merge_RLum.Data.Curve(list(), merge.method = "/"),
               "'object' contains no data")
  expect_error(merge_RLum.Data.Curve(list(set_RLum("RLum.Data.Curve"))),
               "'object' contains no data")
  expect_error(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                     merge.method = "error"),
               "'merge.method' should be one of 'mean', 'median', 'sum', 'sd'")
  expect_error(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3_types)),
               "Objects cannot be merged, different record types found")

  ## different curve lengths
  expect_warning(res <- merge_RLum.Data.Curve(list(TL.curve.1,
                                                   TL.curve.3_short)),
                 "The number of channels differs between the curves")
  expect_equal(nrow(res@data), nrow(TL.curve.3_short@data))

  ## different resolution
  expect_warning(expect_warning(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3_resol)),
               "The curves do not seem to have the same channel resolution"))

})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_s4_class(TL.curve.1 + TL.curve.3, "RLum.Data.Curve")
  expect_s4_class(TL.curve.1 - TL.curve.3, "RLum.Data.Curve")
  suppressWarnings( # silence repeated warning
  expect_warning(expect_s4_class(TL.curve.3 / TL.curve.1,
                                 "RLum.Data.Curve"),
                 "8 'Inf' values replaced by 0 in the matrix"))
  expect_s4_class(TL.curve.1 * TL.curve.3, "RLum.Data.Curve")

  ## only one curve
  expect_s4_class(new <- merge_RLum.Data.Curve(list(TL.curve.1)),
                  "RLum.Data.Curve")
  expect_equal(TL.curve.1@data,
               new@data)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             method.info = 1))
  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "sum"))
  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "median"))

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "sd"))

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "var"))

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "max"))

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "min"))

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "-"))

  expect_s4_class(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3, TL.curve.3),
                                             merge.method = "-"), "RLum.Data.Curve")

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "*"))

  expect_s4_class(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3, TL.curve.3),
                                        merge.method = "*"), "RLum.Data.Curve")

  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "/"))

  expect_s4_class(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3, TL.curve.3),
                                        merge.method = "/"), "RLum.Data.Curve")

  expect_warning(
      expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3_zeros),
                                                 merge.method = "/")),
      "3 'Inf' values replaced by 0 in the matrix")
  expect_snapshot_RLum(merge_RLum.Data.Curve(list(TL.curve.1, TL.curve.3),
                                             merge.method = "append"))
})

#' @title Import measurement data produced by a Daybreak TL/OSL reader into R
#'
#' @description
#' Import a TXT-file (ASCII file) or a DAT-file (binary file) produced by a
#' Daybreak reader into R. The import of the DAT-files is limited to the file
#' format described for the software TLAPLLIC v.3.2 used for Daybreak models
#' 1100 and 1150, and for FLConsole (any release) for Daybreak models 11xx and
#' 2200.
#'
#' @param file [character] or [list] (**required**):
#' name of one or multiple `.DAT` or `.txt` files (URLs are supported); it can
#' be the path to a directory, in which case the function tries to detect and
#' import all files with `.txt` extension.
#'
#' @param raw [logical] (*with default*):
#' if the input is a DAT-file (binary) a [data.table::data.table] instead of
#' the [Luminescence::RLum.Analysis-class] object can be returned for debugging purposes.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable [txtProgressBar].
#'
#' @param ... not in use, for compatibility reasons only
#'
#' @return
#' A list of [Luminescence::RLum.Analysis-class] objects (each per position)
#' is returned.
#'
#' @note
#' **`[BETA VERSION]`**
#' This function still needs to be tested properly. In particular
#' the function has underwent only very rough tests using a few files.
#'
#' @section Function version: 0.4.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Andrzej Bluszcz, Silesian University of Technology, Gliwice (Poland)\cr
#' Antoine Zink, C2RMF, Palais du Louvre, Paris (France)
#'
#' The ASCII-file import is based on a suggestion by William Amidon and
#' Andrew Louis Gorin.
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class], [data.table::data.table]
#'
#' @keywords IO
#'
#' @examples
#'
#' \dontrun{
#' file <- system.file("extdata/Daybreak_TestFile.txt", package = "Luminescence")
#' temp <- read_Daybreak2R(file)
#' }
#'
#' @export
read_Daybreak2R <- function(
  file,
  raw = FALSE,
  verbose = TRUE,
  txtProgressBar = TRUE,
  ...
) {
  .set_function_name("read_Daybreak2R")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO
  ## - run tests
  ## - check where the warning messages are coming from
  ## - implement further integrity tests  (ASCII import)

  # Self Call -----------------------------------------------------------------------------------
  # Option (a): Input is a list, every element in the list will be treated as file connection
  # with that many file can be read in at the same time
  # Option (b): The input is just a path, the function tries to grep ALL Daybreaks-txt files in the
  # directory and import them, if this is detected, we proceed as list

  .validate_class(file, c("character", "list"))
  .validate_logical_scalar(raw)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(txtProgressBar)
  if (!verbose)
    txtProgressBar <- FALSE

  ## Integrity checks -------------------------------------------------------
  file <- .validate_file(file, ext = c("DAT", "txt"), pattern = ".txt$",
                         verbose = verbose)

  ##if the input is already a list
  if (inherits(file, "list")) {
    temp.return <- lapply(seq_along(file), function(x) {
      read_Daybreak2R(
        file = file[[x]],
        txtProgressBar = txtProgressBar,
        verbose = verbose
      )
    })

    return(temp.return)
  }

  ##check for file extension ... distinguish between TXT and DAT
  if (endsWith(toupper(file), ".DAT")) {
    ## Read DAT-file --------------------------------------------------------
    on.exit(close(con), add = TRUE)
      ##screen file to get information on the number of stored records
      con <-file(file, "rb")
      file.data <- file.info(file)
      max.pt <- readBin(con,what="int",6,size=2,endian="little")[6]
      file.size <- file.data$size
      n.length <- file.size/(190+8*(max.pt+1)) ##190 is is size of the header for each data set
      close(con)

      ##import data
      con <- file(file, "rb")
      ##pre-define data.table
      results.DATA <-
        data.table::data.table(
          ID = integer(length = n.length),
          MAXPT = integer(length = n.length),
          SPACING = integer(length = n.length),
          NDISK = integer(length = n.length),
          NRUN = integer(length = n.length),
          D1 = integer(length = n.length),
          DEVN = integer(length = n.length), # DEVN <- D1 %/% 100 codes a stimulation device
                       # TL dev is 0; Bleach dev is 10+; OSL/MOSL dev is 20+; 30+ is isoTL; Irrad dev is 100+; missing pnt is 200
          NPT = integer(length = n.length),
          NATL = logical(length = n.length),
          TLRUN = logical(length = n.length),
          BEFORE_IRRAD = logical(length = n.length),
          NORMALIZATION = logical(length = n.length),
          SINGLE_ALIQUOT = logical(length = n.length),
          SEKOND = logical(length = n.length),
          SHIFT = double(length = n.length),
                       # tb <- as.integer(writeBin(SHIFT, con = raw()))
                       # NTB[0:3] <- tb[0:3] number of channels with each time base 10, 100, 1000, 10000 ms
                       # TEMPERATURE <- tb[4]*256 + tb[7]
                       # OSLBG <- (tb[5]*256 + tb[6])/10.0
          NTB = integer(length = n.length),
          TEMPERATURE = integer(length = n.length),
          OSLBG = double(length = n.length),
          RAMPRATE = double(length = n.length),
          LPOWER = double(length = n.length), # LPOWER <- RAMPRATE  % of max light power
          GRATE = double(length = n.length),
          BRATE = double(length = n.length),
          ARATE = double(length = n.length),
          GAMMADOSE = double(length = n.length),
          BETADOSE = double(length = n.length),
          ALPHADOSE = double(length = n.length),
          BLEACHINGTIME = double(length = n.length),
          GRUNIT = character(length = n.length),
          BRUNIT = character(length = n.length),
          ARUNIT = character(length = n.length),
          BFILTER = character(length = n.length),
          GSOURCE = character(length = n.length),
          BSOURCE = character(length = n.length),
          ASOURCE = character(length = n.length),
          IRRAD_DATE = character(length = n.length),
          RUNREMARK = character(length = n.length),
          DATA = list()
        )

      ##TERMINAL FEEDBACK
      if(verbose){
        cat("\n[read_Daybreak()] Importing ...")
        cat("\n path: ", dirname(file[1]))
        cat("\n file: ", .shorten_filename(basename(file[1])))
        cat("\n")
      }

      ##PROGRESS BAR
      if (txtProgressBar)
        pb <- txtProgressBar(min=0,max=n.length, char = "=", style=3)

      ##LOOP over file
      i <- 1
      while (i <= n.length) {
        #integer
        ligne1<-readBin(con,what="int",6,size=2,endian="little")
        i_NPT<-ligne1[1]
        i_SPACING<-ligne1[2]
        i_NRUN<-ligne1[3]
        i_NDISK<-ligne1[4]
        i_D1<-ligne1[5]
        i_MAXPT<-ligne1[6]
        i_DEVN <- i_D1 %/% 100

        #boolean
        ligne2 <- readBin(con, what = "logical", 6, size = 1, endian = "little")
        i_BEFORE_IRRAD<-ligne2[1]
        i_NORMALIZATION <- ligne2[2]
        i_TLRUN <- ligne2[3]
        i_SINGLE_ALIQUOT <- ligne2[4]
        i_NATL <- ligne2[5]
        i_SEKOND <- ligne2[6]

        #double (real)
        ligne3<-readBin(con,what="double",9,size=8,endian="little")
        i_BLEACHINGTIME<-ligne3[1]#0
        i_RAMPRATE<-ligne3[2] #5
        i_GRATE<-ligne3[3]#0.00000
        i_BRATE<-ligne3[4]#0.074800
        i_ARATE<-ligne3[5]#0.114740
        i_GAMMADOSE<-ligne3[6]#0
        i_BETADOSE<-ligne3[7]#0
        i_ALPHADOSE<-ligne3[8]#0
        i_SHIFT<-ligne3[9] #0
        if (i_DEVN == 0) {
          ## nocov start
          i_LPOWER <- NA
          i_TEMPERATURE <- NA
          i_OSLBG <- NA
          i_NTB <- NA
          ## nocov end
        } else {
          i_LPOWER <- i_RAMPRATE
          tb <- writeBin(i_SHIFT, con = raw())
          i_NTB <- readBin(tb, what = "integer", 1, size = 4)
          # i_RAMPRATE <- NA
          # i_SHIFT <- NA
          i_TEMPERATURE <- as.integer(tb[5]) * 256 + as.integer(tb[8])
          i_OSLBG <- (as.integer(tb[6]) * 256 + as.integer(tb[7])) / 10.0
        }

        #string[7]
        i_GRUNIT <- i_BRUNIT <- i_ARUNIT <- ""
        temp <- readBin(con, "raw", 24, 1)
        l <- as.numeric(temp[1])
        if (l > 0) i_GRUNIT <- rawToChar(temp[2:(1+l)])
        l <- as.numeric(temp[9])
        if (l > 0) i_BRUNIT <- rawToChar(temp[10:(9+l)])
        l <- as.numeric(temp[17])
        if (l > 0) i_ARUNIT <- rawToChar(temp[18:(17+l)])

        #string[7]
        i_BFILTER <- i_GSOURCE <- i_BSOURCE <- i_ASOURCE <- ""
        temp <- readBin(con, "raw", 32, 1)
        l <- as.numeric(temp[1])
        if (l > 0) i_BFILTER <- rawToChar(temp[2:(1+l)])
        l <- as.numeric(temp[9])
        if (l > 0) i_GSOURCE <- rawToChar(temp[10:(9+l)])
        l <- as.numeric(temp[17])
        if (l > 0) i_BSOURCE <- rawToChar(temp[18:(17+l)])
        l <- as.numeric(temp[25])
        if (l > 0) i_ASOURCE <- rawToChar(temp[26:(25+l)])

        #date record
        raw_IRRAD_DATE <- readBin(con, what = "raw", 2, size = 1, endian = "little") # 27-Nov-2006
        bitDATE<-as.integer(rawToBits(raw_IRRAD_DATE))
        DATE.AAAA <- sum(bitDATE[10:16] * c(1, 2, 4, 8, 16, 32, 64)) + 1900
        if (DATE.AAAA < 1950) DATE.AAAA <- DATE.AAAA + 128 # nocov
        DATE.MM<-sum(bitDATE[seq(1,4)]*c(1,2,4,8))
        DATE.JJ<-sum(bitDATE[seq(5,9)]*c(1,2,4,8,16))
        i_IRRAD_DATE<-paste0(DATE.AAAA,"-",MM=DATE.MM,"-",JJ=DATE.JJ)

        #string[41]
        i_RUNREMARK <- readBin(con, "raw", n = 42)
        l <- as.numeric(i_RUNREMARK[1])
        i_RUNREMARK <- if (l > 0) rawToChar(i_RUNREMARK[2:(1 + l)]) else ""

        i_DATA <- readBin(con,what="double",i_MAXPT+1,size=8,endian="little")

        results.DATA[i,':='(ID=i,MAXPT=i_MAXPT,SPACING=i_SPACING,NDISK=i_NDISK,NRUN=i_NRUN,D1=i_D1,
                            DEVN = i_DEVN,
                            NPT = i_NPT,
                            NATL=i_NATL,TLRUN=i_TLRUN,BEFORE_IRRAD=i_BEFORE_IRRAD,
                            NORMALIZATION = i_NORMALIZATION,
                            SINGLE_ALIQUOT = i_SINGLE_ALIQUOT,
                            SEKOND = i_SEKOND,
                            SHIFT = i_SHIFT,
                            NTB = i_NTB,
                            TEMPERATURE = i_TEMPERATURE,
                            OSLBG = i_OSLBG,
                            RAMPRATE = i_RAMPRATE,
                            LPOWER = i_LPOWER,
                            GRATE=i_GRATE,BRATE=i_BRATE,ARATE=i_ARATE,
                            GAMMADOSE=i_GAMMADOSE,BETADOSE=i_BETADOSE,ALPHADOSE=i_ALPHADOSE,
                            BLEACHINGTIME=i_BLEACHINGTIME,
                            GRUNIT=i_GRUNIT,BRUNIT=i_BRUNIT, ARUNIT=i_ARUNIT,
                            BFILTER=i_BFILTER,
                            GSOURCE=i_GSOURCE,BSOURCE=i_BSOURCE,ASOURCE=i_ASOURCE,
                            IRRAD_DATE=i_IRRAD_DATE,
                            RUNREMARK=i_RUNREMARK,
                            DATA=list(i_DATA))]

        ##update
        i <- i + 1

        ##update progress bar
        if (txtProgressBar) {
          setTxtProgressBar(pb, i)
        }
      }

    if (txtProgressBar)
      close(pb)

    ## Output a data.table
    if (raw)
      return(results.DATA)

    ## Output an RLum.Analysis object
    ## remove NULL entries, otherwise we have to deal with them later
    results.DATA <- results.DATA[!sapply(results.DATA[["DATA"]], is.null), ]

    ## we need a double loop, as one aliquot defines one object ...
    output <- lapply(unique(results.DATA[["NDISK"]]), function(i) {

          ##subset
          DT <- results.DATA[results.DATA[["NDISK"]] == i,]

          ##create list of records
          records <- lapply(1:nrow(DT), function(j){
            if (DT[["DEVN"]][j] == 0) {
              xx <- seq(from = DT[["SPACING"]][j], by = DT[["SPACING"]][j], length.out = DT[["NPT"]][j]) # nocov
            } else {
              tb <- c(0.01, 0.1, 1.0, 10.0)
              last_t <- 0
              BT <- as.integer(writeBin(DT[["NTB"]][j], con = raw()))
              xx <- numeric()
              for (k in 1:4) {
                if (BT[k] > 0) {
                  xx <- c(xx, seq(from = last_t + tb[k], by = tb[k], length.out = BT[k]))
                  last_t <- xx[length(xx)]
                }
              }
            }
            dev <- as.integer(DT[["DEVN"]][j])
            set_RLum(
              class = "RLum.Data.Curve",
              originator = "read_Daybreak2R",
              recordType = ifelse(dev == 0, "TL (PMT)",
                           ifelse((dev >= 10) && (dev < 20), "bleaching (PMT)",
                           ifelse((dev >= 20) && (dev < 30), "OSL (PMT)",
                           ifelse((dev >= 30) && (dev < 40), "isoTL (PMT)",
                           ifelse((dev == 200), "missing_point (NA)", NA_character_))))),
              curveType = "measured",
              data = matrix(
                data = c(
                    xx, DT[["DATA"]][j][[1]][2:(DT[["NPT"]][j] + 1)]),
                ncol = 2),
              info = as.list(DT[j,1:(ncol(DT) - 1)])
            )
          })

          ##combine in RLum.Analysis object
          temp <- set_RLum(
            class = "RLum.Analysis",
            originator = "read_Daybreak2R",
            protocol = "Custom",
            records =  records
            )

      .set_pid(temp)
    })

  }else{

    ## Read ASCII file ------------------------------------------------------
    if(verbose){
      cat("\n[read_Daybreak()] File extension not of type '.DAT', trying to import ASCII-file ... \n")
    }

    ##read file
    file2read <- suppressWarnings(readLines(file))

    ## remove all the empty lines
    file2read <- file2read[nzchar(file2read)]

    ## check whether the file contains non-ASCII characters: the [^ -~]
    ## regexp matches all ASCII characters from space to tilde
    if (any(grepl("[^ -~]", file2read[1])))
      .throw_error("The provided file is not ASCII and cannot be imported")

    ##(1)
    ##get all rows with the term "[NewRecord]" - that's what we are interested in and it defines
    ##the number of elements we need
    records.row_number <- grep(pattern = "\\[NewRecord\\]", x = file2read)
    if (length(records.row_number) == 0) {
      .throw_error("File '", file, "' doesn't appear to be in Daybreak format")
    }

    ##(1)
    ##make a list ... this is not essentially needed but it makes things easier
    data.list <- lapply(1:length(records.row_number), function(x) {
      last <- records.row_number[x + 1]
      last <- if (!is.na(last)) last - 1 else length(file2read)
      file2read[records.row_number[x]:last]
    })

      ##clear memory
      rm(file2read)

    ##TERMINAL FEEDBACK
    if(verbose){
      cat("\n[read_Daybreak()] Importing ...")
      cat("\n path: ", dirname(file[1]))
      cat("\n file: ", .shorten_filename(basename(file[1])))
      cat("\n")
    }

    ##PROGRESS BAR
    if(txtProgressBar & verbose)
      pb <- txtProgressBar(min = 0,max=length(data.list), char = "=", style = 3)

    ##(2)
    ##Loop over the list to create RLum.Data.Curve objects
    RLum.Data.Curve.list <- lapply(seq_along(data.list), function(x) {
      record <- data.list[[x]]
      record.length <- length(record)

      ##get header length until the argument 'Points'
      header.length <- grep("Points", x = record, fixed = TRUE)

      last.idx <- if (length(header.length) > 0) header.length else record.length
      temp.meta_data <- unlist(strsplit(record[2:last.idx], split = "=", fixed = TRUE))

      ##get list names for the info element list
      info.names <- temp.meta_data[seq(1,length(temp.meta_data), by = 2)]

      ##info elements
      info <- as.list(temp.meta_data[seq(2,length(temp.meta_data), by = 2)])
      names(info) <- info.names

      ##add position, which is 'Disk'
      info <- c(info, position = as.integer(info$Disk))

      if(length(header.length)>0){
        ## get measurement data ... this construction makes no assumption on
        ## the number of columns
        temp.data <- data.table::tstrsplit(
          record[header.length + 1:as.numeric(info[["Points"]])], ";", fixed = TRUE,
          type.convert = TRUE)

        ## construct data matrix
        data <- matrix(unlist(temp.data[2:3]), ncol = 2)

      } else if ("IrradTime" %in% names(info)) {
          point.x <- c(1,as.numeric(info$IrradTime))
          point.y <- rep(1, 2)
          data <- matrix(c(point.x,point.y), ncol = 2)
      } else if (info[["DataType"]] == "Command") {
        ## Note from Andrzej Bluszcz:
        # The regular [NewRecord] DataType=Command was used 2009 through 2023 and
        # later it was replaced by a single line Commands=abc belonging formally to the previous record.
        # It was probably not an official Daybreak release, but my modification used locally in our lab.
        # Nevertheless, it is wiser to use record[header.length + 1:as.numeric(info[["Points"]])]
        # to extract data points
        data <- matrix(0, ncol = 2)
      }

      ##update progress bar
      if (txtProgressBar)
        setTxtProgressBar(pb, x)

      ##return RLum object
      set_RLum(
          class = "RLum.Data.Curve",
          originator = "read_Daybreak2R",
          recordType = sub(" ", replacement = "_", x = info$DataType),
          curveType = "measured",
          data = data,
          info = info
      )
    })

    ##close ProgressBar
    if (txtProgressBar)
      close(pb)

    ##(3)
    ##Now we have to find out how many aliquots we do have
    positions.id <- vapply(RLum.Data.Curve.list, get_RLum, info.object = "position", integer(1))

    ##(4)
    ##now combine everything in an RLum.Analysis object in accordance to the position number
    output <- lapply(unique(positions.id), function(x) {
      ##get list ids for position number
      temp.list <- RLum.Data.Curve.list[which(positions.id == x)]

      ##put in RLum.Analysis object
      object <- set_RLum(
        class = "RLum.Analysis",
        originator = "read_Daybreak2R",
        protocol = "Custom",
        records = temp.list
      )

      ##set parent id of records
      .set_pid(object)
    })
  }

  ## terminal feedback
  if (verbose) {
    cat("\n ", length(unlist(get_RLum(output))),
        "records have been read successfully\n")
  }

  output
}

test_that("Test functionality", {
  testthat::skip_on_cran()

  txt.file <- system.file("extdata/Daybreak_TestFile.txt",
                          package = "Luminescence")
  dat.file <- system.file("extdata/Daybreak_TestFile.DAT",
                          package = "Luminescence")

  ## TXT
  SW({
  expect_type(read_Daybreak2R(txt.file), "list")
  expect_type(read_Daybreak2R(txt.file, txtProgressBar = FALSE),
              "list")
  expect_type(read_Daybreak2R(test_path("_data/daybreak-tests/morethan4cols.txt")),
              "list")
  })
  expect_silent(read_Daybreak2R(txt.file, verbose = FALSE))

  ## DAT
  SW({
  expect_type(read_Daybreak2R(dat.file), "list")
  expect_s3_class(read_Daybreak2R(dat.file, raw = TRUE),
                  "data.table")
  })
  expect_silent(read_Daybreak2R(dat.file, verbose = FALSE))

  ## lowercase extension
  expect_s3_class(read_Daybreak2R(test_path("_data/daybreak-tests/lowercase.dat"),
                                  raw = TRUE, verbose = FALSE),
                  "data.table")

  ## list
  SW({
  expect_type(read_Daybreak2R(list(dat.file)), "list")
  })
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_Daybreak2R(data.frame()),
               "'file' should be of class 'character' or 'list'")
  expect_error(read_Daybreak2R(character(0)),
               "'file' cannot be an empty character")
  expect_error(read_Daybreak2R(list()),
               "'file' cannot be an empty list")
  expect_error(read_Daybreak2R("not-existing"),
               "File '.*not-existing' does not exist") # windows CI needs the regexp

  ## directory
  SW({
  expect_error(
    expect_message(read_Daybreak2R(
      file = system.file("extdata", package = "Luminescence")),
      "Directory detected, looking for '.txt$' files"),
    "File '.*.txt' doesn't appear to be in Daybreak format")
  })

  expect_error(read_Daybreak2R(
    file = system.file("extdata/BINfile_V8.binx", package = "Luminescence"),
    verbose = FALSE),
    "File extension 'binx' is not supported, only 'DAT' and 'txt' are valid")

  ## test presence of non-ascii characters
  file.nonascii <- tempfile(fileext = ".txt")
  writeLines(gsub("ScriptFile", "ScriptFile \uf6", # ö
                  readLines(system.file("extdata/Daybreak_TestFile.txt",
                                        package = "Luminescence"))),
             file.nonascii)
  expect_error(read_Daybreak2R(file = file.nonascii, verbose = FALSE),
    "The provided file is not ASCII and cannot be imported")
})

test_that("command records", {
  testthat::skip_on_cran()

  res <- read_Daybreak2R(test_path("_data/daybreak-tests/command-1.txt"),
                         verbose = FALSE)
  expect_type(res, "list")
  expect_length(res, 1)
  expect_length(res[[1]], 2)
  expect_output(print(res[[1]]@records[[2]]))

  res <- read_Daybreak2R(test_path("_data/daybreak-tests/command-2.txt"),
                         verbose = FALSE)
  expect_type(res, "list")
  expect_length(res, 1)
  expect_length(res[[1]], 11)
  expect_true(all(!is.na(res[[1]]@records[[11]]@data)))
})

#' @title Plot filter combinations along with the (optional) net transmission window
#'
#' @description
#' The function allows to plot transmission windows for different filters.
#' Missing data for specific wavelengths are automatically interpolated for
#' the given filter data using function [approx]. With that, a standardised
#' output is reached and a net transmission window can be shown.
#'
#' **Calculations**
#'
#' **Net transmission window**
#'
#' The net transmission window of two filters is approximated by
#'
#' \deqn{T_{final} = T_{1} * T_{2}}
#'
#' **Optical density**
#'
#' \deqn{OD = -\log10(T)}
#'
#' **Total optical density**
#'
#' \deqn{OD_{total} = OD_{1} +  OD_{2}}
#'
#' Please consider using your own calculations for more precise values.
#'
#' **How to provide input data?**
#'
#' *CASE 1*
#'
#' The function expects that all filter values are either of type `matrix` or
#' `data.frame` with two columns. The first columns contains the wavelength,
#' the second the relative transmission (but not in percentage, i.e. the
#' maximum transmission can be only become 1).
#'
#' In this case only the transmission window is show as provided. Changes in
#' filter thickness and reflection factor are not considered.
#'
#' *CASE 2*
#'
#' The filter data itself are provided as list element containing a `matrix` or
#' `data.frame` and additional information on the thickness of the filter, e.g.,
#' `list(filter1 = list(filter_matrix, d = 2))`.
#' The given filter data are always considered as standard input and the filter
#' thickness value is taken into account by
#'
#' \deqn{Transmission = Transmission^d}
#'
#' with d given in the same dimension as the original filter data.
#'
#' *CASE 3*
#'
#' Same as CASE 2 but additionally a reflection factor P is provided, e.g.,
#' `list(filter1 = list(filter_matrix, d = 2, P = 0.9))`.
#' The final transmission becomes:
#'
#' \deqn{Transmission = Transmission^d * P}
#'
#' **Advanced plotting parameters**
#'
#' The following further non-common plotting parameters can be passed to the function:
#'
#' \tabular{lll}{
#' **`Argument`** \tab **`Datatype`** \tab **`Description`**\cr
#' `legend` \tab `logical` \tab enable/disable legend \cr
#' `legend.pos` \tab `character` \tab change legend position ([graphics::legend]) \cr
#' `legend.text` \tab `character` \tab same as the argument `legend` in ([graphics::legend]) \cr
#' `net_transmission.col` \tab `col` \tab colour of net transmission window polygon \cr
#' `net_transmission.col_lines` \tab `col` \tab colour of net transmission window polygon lines \cr
#' `net_transmission.density` \tab  `numeric` \tab specify line density in the transmission polygon \cr
#' `grid` \tab `list` \tab full list of arguments that can be passed to the function [graphics::grid]
#' }
#'
#' For further modifications standard additional R plot functions are recommend,
#' e.g., the legend can be fully customised by disabling the standard legend
#' and using the function [graphics::legend] instead.
#'
#' @param filters [list] (**required**):
#' a named list of filter data for each filter to be shown.
#' The filter data itself should be either provided as [data.frame] or [matrix]
#' (see details for more options).
#'
#' @param wavelength_range [numeric] (*with default*):
#' wavelength range used for the interpolation
#'
#' @param show_net_transmission [logical] (*with default*):
#' show net transmission window as polygon.
#'
#' @param interactive [logical] (*with default*):
#' enable/disable interactive plots.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments that can be passed to control the plot output.
#' Supported are `main`, `xlab`, `ylab`, `xlim`, `ylim`, `type`, `lty`, `lwd`.
#' For non-common plotting parameters, see the details section.
#'
#' @return Returns an S4 object of type [Luminescence::RLum.Results-class].
#'
#' **@data**
#'
#' \tabular{lll}{
#'  **`Object`** \tab **`Type`** \tab **`Description`** \cr
#'  `net_transmission_window` \tab `matrix` \tab the resulting net transmission window \cr
#'  `OD_total` \tab `matrix` \tab the total optical density\cr
#'  `filter_matrix` \tab `matrix` \tab the filter matrix used for plotting
#' }
#'
#' **@info**
#'
#' \tabular{lll}{
#' **Object** \tab **Type**\tab  **Description** \cr
#' `call` \tab [call] \tab the original function call
#' }
#'
#' @section Function version: 0.3.3
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Results-class], [approx]
#'
#' @keywords datagen aplot
#'
#' @examples
#'
#' ## (For legal reasons no real filter data are provided)
#'
#' ## Create filter sets
#' filter1 <- density(rnorm(100, mean = 450, sd = 20))
#' filter1 <- matrix(c(filter1$x, filter1$y/max(filter1$y)), ncol = 2)
#' filter2 <- matrix(c(200:799,rep(c(0,0.8,0),each = 200)), ncol = 2)
#'
#' ## Example 1 (standard)
#' plot_FilterCombinations(filters = list(filter1, filter2))
#'
#' ## Example 2 (with d and P value and name for filter 2)
#' results <- plot_FilterCombinations(
#' filters = list(filter_1 = filter1, Rectangle = list(filter2, d = 2, P = 0.6)))
#' results
#'
#' ## Example 3 show optical density
#' plot(results$OD_total)
#'
#' \dontrun{
#' ##Example 4
#' ##show the filters using the interactive mode
#' plot_FilterCombinations(filters = list(filter1, filter2), interactive = TRUE)
#' }
#'
#' @export
plot_FilterCombinations <- function(
  filters,
  wavelength_range = 200:1000,
  show_net_transmission = TRUE,
  interactive = FALSE,
  plot = TRUE,
  ...
) {
  .set_function_name("plot_FilterCombinations")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(filters, "list")
  .validate_not_empty(filters, "list")

  #input should either data.frame or matrix
  lapply(filters, function(x) {
    .validate_class(x, c("data.frame", "matrix", "list"),
                    name = "All elements of 'filters'")
    .validate_not_empty(x, name = "Each element of 'filters'")
  })

  #check for named list, if not set names
  if (is.null(names(filters))) {
    names(filters) <- paste("Filter ", seq_along(filters))
  }

  ## Data preparation -------------------------------------------------------

  ## check if filters are provided with their thickness, if so correct
  ## transmission for this ... relevant for glass filters
  filters <- lapply(filters, function(x) {
    if (inherits(x, "list")) {

      ## correction for the transmission accounting for filter thickness,
      ## the provided thickness is always assumed to be 1
      if (length(x) > 1) {
        x[[1]][, 2] <- x[[1]][, 2] ^ x[[2]]
      }

      ## account for potentially provided transmission reflection factor
      if (length(x) > 2) {
        x[[1]][, 2] <- x[[1]][, 2] * x[[3]]
      }

      return(x[[1]])
    }

    return(x)
  })

  #check if there are transmission values greater than one, this is not possible
  lapply(filters, function(x) {
    if (NCOL(x) < 2)
      .throw_error("All data frames in 'filters' should have 2 columns")
    if (max(x[, 2], na.rm = TRUE) > 1.01) {
      .throw_error("Transmission values > 1 found, check your data")
    }
  })

  ##combine everything in a matrix using approx for interpolation
  filter_matrix <- vapply(filters, function(x) {
    approx(x = x[, 1], y = x[, 2], xout = wavelength_range)$y
  }, FUN.VALUE = numeric(length(wavelength_range)))

  ##calculate transmission window
  filter_matrix <- cbind(filter_matrix)
  net_transmission_window <- matrix(
    c(wavelength_range, matrixStats::rowProds(filter_matrix)),
    ncol = 2)

  ##add optical density to filter matrix

  ##calculate OD
  OD <- -log10(filter_matrix)

  ##calculate  total OD
  OD_total <- cbind(wavelength_range, matrixStats::rowSums2(OD))

  ##add to matrix
  filter_matrix <- cbind(filter_matrix, OD)

  ##set rownames of filter matrix
  rownames(filter_matrix) <- wavelength_range

  ##set column names for filter matrix
  colnames(filter_matrix) <- c(names(filters), paste0(names(filters), "_OD"))

  # Plotting ------------------------------------------------------------------------------------
  if (plot) {
    ##(1) ... select transmission values
    filter_matrix_transmission <- filter_matrix[, !grepl(pattern = "OD",
                                                         colnames(filter_matrix)),
                                                drop = FALSE]

    ##set plot settings
    plot_settings <- list(
      main = "Filter Combination",
      xlab = "Wavelength [nm]",
      ylab = "Transmission [a.u.]",
      xlim = range(wavelength_range),
      ylim = c(0, 1),
      legend.pos = "topleft",
      lty = 1,
      lwd = 1,
      col = 1:length(filters),
      grid = expression(nx = 10, ny = 10),
      legend = TRUE,
      legend.text = colnames(filter_matrix_transmission),
      net_transmission.col = rgb(0, 0.7, 0, 0.2),
      net_transmission.col_lines = "grey",
      net_transmission.density = 20
    )

    ##modify settings on request
    plot_settings <- modifyList(plot_settings, list(...))

    if(interactive){
      .require_suggested_package("plotly", "Producing interactive plots")

      ##create basic plot
      p <-
        plotly::plot_ly(
          x = wavelength_range,
          y = filter_matrix[,1],
          type = "scatter",
          name = plot_settings$legend.text[1],
          mode = "lines")

        ##add further filters
        if (ncol(filter_matrix_transmission) > 1) {
          for (i in 2:ncol(filter_matrix_transmission)) {
            p <- plotly::add_trace(p,
                        y = filter_matrix[, i],
                        name = plot_settings$legend.text[i],
                        mode = 'lines')
          }
        }

      ##add polygon
      ##replace all NA values with 0, otherwise it looks odd
      net_transmission_window[is.na(net_transmission_window)] <- 0

      if(show_net_transmission) {
        p <-  plotly::add_polygons(
          p,
          x = c(wavelength_range, rev(wavelength_range)),
          y = c(net_transmission_window[, 2], rep(0, length(wavelength_range))),
          name = "net transmission"
        )
      }

      ##change graphical parameters
      p <-  plotly::layout(
        p = p,
        xaxis = list(
          title = plot_settings$xlab,
          range = plot_settings$xlim
        ),
        yaxis = list(
          title = plot_settings$ylab,
          range = plot_settings$ylim
        ),
        title = plot_settings$main,
        showlegend = plot_settings$legend
      )

      ## use hidden option for shiny to make it
      ## work in the filter app
      if(is.null(list(...)$.shiny_server))
        print(p)

      on.exit(return(p), add = TRUE)

    }else{
      ##plot individual filters
      graphics::matplot(
        x = wavelength_range,
        y = filter_matrix_transmission,
        type = "l",
        main = plot_settings$main,
        xlab = plot_settings$xlab,
        ylab = plot_settings$ylab,
        xlim = plot_settings$xlim,
        ylim = plot_settings$ylim,
        lty = plot_settings$lty,
        lwd = plot_settings$lwd,
        col = plot_settings$col
      )

      if (!is.null(plot_settings$grid)) {
        graphics::grid(eval(plot_settings$grid))
      }

      ##show effective transmission, which is the minimum for each row
      if (show_net_transmission) {
        ##replace all NA values with 0, otherwise it looks odd
        net_transmission_window[is.na(net_transmission_window)] <- 0

        polygon(
          x = c(wavelength_range, rev(wavelength_range)),
          y = c(net_transmission_window[, 2],
                rep(0, length(wavelength_range))),
          col = plot_settings$net_transmission.col,
          border = NA
        )
        polygon(
          x = c(wavelength_range, rev(wavelength_range)),
          y = c(net_transmission_window[, 2],
                rep(0, length(wavelength_range))),
          col = plot_settings$net_transmission.col_lines,
          border = NA,
          density = plot_settings$net_transmission.density
        )
      }

      #legend
      if (plot_settings$legend) {
        legend(
          plot_settings$legend.pos,
          legend = plot_settings$legend.text,
          col = plot_settings$col,
          lty = plot_settings$lty,
          bty = "n"
        )
      }
    }
  }

  # Produce output object -----------------------------------------------------------------------
  invisible(set_RLum(
    class = "RLum.Results",
    data = list(
      net_transmission_window = net_transmission_window,
      OD_total = OD_total,
      filter_matrix = filter_matrix

    ),
    info = list(call = sys.call())
  ))
}

## load data
set.seed(1)
density <- density(rnorm(100, mean = 450, sd = 20))
filter1 <- matrix(c(density$x, density$y / max(density$y)), ncol = 2)
filter2 <- matrix(c(200:799, rep(c(0, 0.8, 0), each = 200)), ncol = 2)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_FilterCombinations("error"),
               "'filters' should be of class 'list'")
  expect_error(plot_FilterCombinations(list()),
               "'filters' cannot be an empty list")
  expect_error(plot_FilterCombinations(list(list())),
               "Each element of 'filters' cannot be an empty list")
  expect_error(plot_FilterCombinations(list("error")),
               "All elements of 'filters' should be of class 'data.frame'")
  expect_error(plot_FilterCombinations(list(iris[, 1, drop = FALSE])),
               "All data frames in 'filters' should have 2 columns")
  expect_error(plot_FilterCombinations(list(list(3))),
               "All data frames in 'filters' should have 2 columns")
  expect_error(plot_FilterCombinations(list(data.frame(a = c(100, 200),
                                                       b = c(0.2, 1.1)))),
               "Transmission values > 1 found, check your data")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_silent(plot_FilterCombinations(filters = list(filter1, filter2)))
  expect_silent(plot_FilterCombinations(filters = list(filter1, filter2),
                                        interactive = TRUE))
  expect_silent(plot_FilterCombinations(list(list(filter1))))

  ## filter thickness and reflection factor provided
  expect_silent(plot_FilterCombinations(list(list(filter1, d = 2),
                                             list(filter2, d = 2, P = 0.9))))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("FilterCombinations defaults",
                              plot_FilterCombinations(list(filter1, filter2)))
  vdiffr::expect_doppelganger("FilterCombinations values",
                              plot_FilterCombinations(list(
                                  list(filter1, d = 2, P = 0.8),
                                  Rectangle = list(filter2, d = 2, P = 0.6))))
  })
})

#' General accessor function for RLum S4 class objects
#'
#' Function calls object-specific get functions for RLum S4 class objects.
#'
#' The function provides a generalised access point for specific
#' [RLum-class] objects.\cr
#' Depending on the input object, the corresponding get function will be selected.
#' Allowed arguments can be found in the documentations of the corresponding
#' [RLum-class] class.
#'
#' @param object [RLum-class] (**required**):
#' S4 object of class `RLum`
#'
#' @return Return is the same as input objects as provided in the list.
#'
#' @section Function version: 0.1.0
#'
#' @author Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
#' (France)
#' @seealso
#' [RLum.Data.Curve-class],
#' [RLum.Data.Image-class],
#' [RLum.Data.Spectrum-class],
#' [RLum.Analysis-class],
#' [RLum.Results-class]
#' @keywords utilities
#'
#'
#' @md
#' @export
setGeneric("length_RLum", function(object) {
  standardGeneric("length_RLum")
})

#' @title  Create a violin plot
#'
#' @description
#' Draws a kernel density plot in combination with a boxplot in its middle.
#' The shape of the violin is constructed using a mirrored density curve.
#' This plot is especially designed for cases where the individual errors are
#' zero or too small to be visualised. The idea for this plot is based on the
#' 'violin plot' in the `ggplot2` package by Hadley Wickham and Winston Chang.
#' The general idea for the violin plot seems to have been introduced by
#' Hintze and Nelson (1998).
#'
#' @details
#' The function is passing several arguments to the functions [plot],
#' [stats::density], [graphics::boxplot]:
#'
#' Supported arguments are:
#' `xlim`, `main`, `xlab`, `ylab`, `col.violin`, `col.boxplot`, `mtext`, `cex`, `mtext`
#'
#' **Valid summary keywords**
#'
#' `'n'`, `'mean'`, `'median'`, `'sd.abs'`, `'sd.rel'`, `'se.abs'`, `'se.rel'`.
#' `'skewness'`, `'kurtosis'`
#'
#' @param data [numeric] or [Luminescence::RLum.Results-class] (**required**):
#' input data for plotting. Alternatively a [data.frame] or a [matrix] can
#' be provided, but only the first column will be considered by the
#' function.
#'
#' @param boxplot [logical] (*with default*):
#' enable/disable the boxplot.
#'
#' @param rug [logical] (*with default*):
#' enable/disable the rug.
#'
#' @param summary [character] (*with default*):
#' add statistic measures of centrality and dispersion to the plot.
#' Can be one or more of several keywords. See details for available keywords.
#'
#' @param summary.pos [numeric] or [character] (*with default*):
#' optional position keywords (cf. [legend]) for the statistical summary.
#' Alternatively, the keyword `"sub"` may be specified to place the summary
#' below the plot header. However, this latter option in only possible if
#' `mtext` is not used.
#'
#' @param na.rm [logical] (*with default*):
#' exclude `NA` values from the data set prior to any further operations.
#'
#' @param ... further arguments and graphical parameters passed to
#' [plot.default], [stats::density] and [boxplot]. See details for further
#' information.
#'
#' @note
#' Although the code for this function was developed independently and just
#' the idea for the plot was based on the 'ggplot2' plot type 'violin', it
#' should be mentioned that, beyond this, at least another R package produces
#' this kind of plot, namely `'vioplot'` (see references for details).
#'
#' @section Function version: 0.1.4
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @references
#' Daniel Adler (2025). vioplot: violin plot.
#' R package version 0.5.1 http://CRAN.R-project.org/package=vioplot
#'
#' Hintze, J.L., Nelson, R.D. (1998). A Box Plot-Density Trace Synergism.
#' The American Statistician 52, 181-184.
#'
#' Wickham. H (2009). ggplot2: elegant graphics for data analysis. Springer New York.
#'
#' @seealso [stats::density], [plot], [boxplot], [rug], [Luminescence::calc_Statistics]
#'
#' @examples
#'
#' ## read example data set
#' data(ExampleData.DeValues, envir = environment())
#' ExampleData.DeValues <- convert_Second2Gray(ExampleData.DeValues$BT998,
#'                                             c(0.0438,0.0019))
#'
#' ## create plot straightforward
#' plot_ViolinPlot(data = ExampleData.DeValues)
#'
#' @export
plot_ViolinPlot <- function(
  data,
  boxplot = TRUE,
  rug = TRUE,
  summary = c("n", "median"),
  summary.pos = "sub",
  na.rm = TRUE,
  ...
) {
  .set_function_name("plot_ViolinPlot")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("RLum.Results", "data.frame", "matrix"))
  .validate_not_empty(data)
  .validate_logical_scalar(boxplot)
  .validate_logical_scalar(rug)
  .validate_class(summary, "character")
  if (is.numeric(summary.pos)) {
    .validate_length(summary.pos, 2)
  }
  else {
    summary.pos <- .validate_args(summary.pos,
                                  c("sub", "left", "center", "right",
                                    "topleft", "top", "topright",
                                    "bottomleft", "bottom", "bottomright"))
  }

  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")
  }
  if (is.data.frame(data) || is.matrix(data)) {
    data <- data[, 1]
  }
  .validate_class(data, "numeric", name = "All elements of 'data'")

    ##Remove NA values
    if(na.rm){
      data <- na.exclude(data)

      if(length(attr(data, "na.action")) > 0){
        .throw_warning(length(attr(data, "na.action")), " NA values removed")
      }
    }

  ##stop if only one or 0 values are left in data
  if(length(data) == 0){
    .throw_warning("After NA removal, nothing is left from the data set")
    return()
  }


  # Pre-calculations ----------------------------------------------------------------------------

  ##density for the violin
  if(length(data)>1){
    density <-
      density(x = data,
              bw = list(...)$bw %||% "nrd0")
  }else{
    density <- NULL
    .throw_warning("Single data point found, no density calculated")
  }


  ##some statistical parameter, get rid of the weighted statistics
  stat.summary <- list(suppressWarnings(calc_Statistics(as.data.frame(data), digits = 2)[["unweighted"]]))
  names(stat.summary) <- "unweighted"

    ##at least show a warning for invalid keywords
    if(!all(summary %in% names(stat.summary[[1]]))){
      .throw_warning("Only keywords for unweighted statistical measures ",
                     "are supported. Valid keywords are: ",
                     .collapse(names(stat.summary[[1]])))
    }

    ##make sure that only valid keywords make it
    summary <- summary[(summary %in% names(stat.summary[[1]]))]

    stat.text <- .create_StatisticalSummaryText(stat.summary, keywords = summary, sep = " \n ")
    stat.mtext <- .create_StatisticalSummaryText(stat.summary, keywords = summary, sep = " | ")

  # Plot settings -------------------------------------------------------------------------------

  ##set default values
  plot.settings <- list(
    xlim = if(!is.null(density)){range(density$x)}else{c(data[1]*0.9, data[1]*1.1)},
    main = "Violin Plot",
    xlab = expression(paste(D[e], " [a.u.]")),
    ylab = if(!is.null(density)){"Density"}else{" "},
    col.violin = rgb(0,0,0,0.2),
    col.boxplot = NULL,
    mtext = ifelse(summary.pos != 'sub', "", stat.mtext),
    cex = 1
  )

  ##modify list accordingly
  plot.settings <- modifyList(plot.settings, val = list(...))


  # Plot ----------------------------------------------------------------------------------------

  ##open empty plot area
  plot(
    NA,NA,
    xlim = plot.settings$xlim,
    ylim = c(0.2,1.8),
    xlab = plot.settings$xlab,
    ylab = plot.settings$ylab,
    yaxt = "n",
    main = plot.settings$main,
    cex = plot.settings$cex
  )

  ##add polygon ... the violin
  if(!is.null(density)){
    polygon(
      x = c(density$x, rev(density$x)),
      y = c(1 + density$y / max(density$y) * 0.5,
            rev(1 - density$y / max(density$y) * 0.5)),
      col = plot.settings$col.violin,
      border = plot.settings$col.violin
    )
  }

  ##add the boxplot
  if(boxplot){
    graphics::boxplot(
      data,
      outline = TRUE,
      boxwex = 0.4,
      horizontal = TRUE,
      axes = FALSE,
      add = TRUE,
      col = plot.settings$col.boxplot
    )
  }

  ##add rug
  if(rug){
    graphics::rug(x = data)
  }

  ##add mtext
  if(!is.null(plot.settings$mtext)){
    mtext(side = 3, text = plot.settings$mtext)
  }

  ##add stat.text
  if (summary.pos != "sub") {
      legend(summary.pos, legend = stat.text, bty = "n")
  }
}

## load data
data(ExampleData.DeValues, envir = environment())
df <- ExampleData.DeValues$CA1

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_ViolinPlot(),
               "'data' should be of class 'RLum.Results', 'data.frame' or 'matrix'")
  expect_error(plot_ViolinPlot("error"),
               "'data' should be of class 'RLum.Results', 'data.frame' or 'matrix'")
  expect_error(plot_ViolinPlot(df, boxplot = 5),
               "'boxplot' should be a single logical value")
  expect_error(plot_ViolinPlot(df, rug = iris),
               "'rug' should be a single logical value")
  expect_error(plot_ViolinPlot(df, summary = 5),
               "'summary' should be of class 'character'")
  expect_error(plot_ViolinPlot(df, summary.pos = 5),
               "'summary.pos' should have length 2")
  expect_error(plot_ViolinPlot(df, summary.pos = list()),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_ViolinPlot(df, summary.pos = "error"),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_ViolinPlot(data.frame(a = letters)),
               "All elements of 'data' should be of class 'numeric'")
  expect_error(plot_ViolinPlot(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(plot_ViolinPlot(df[0, ]),
               "'data' cannot be an empty data.frame")
  expect_warning(plot_ViolinPlot(df[1, ]),
                 "Single data point found, no density calculated")
  expect_warning(plot_ViolinPlot(df, summary = "error"),
                 "Only keywords for unweighted statistical measures are supported")

  ## missing values
  df.na <- df
  df.na[, 1] <- NA_real_
  expect_warning(expect_warning(
      plot_ViolinPlot(df.na),
      "62 NA values removed"),
      "After NA removal, nothing is left from the data set")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## data.frame
  expect_silent(plot_ViolinPlot(df, summary.pos = "topleft"))

  ## matrix
  expect_silent(plot_ViolinPlot(as.matrix(df)))

  ## RLum.Results object
  expect_silent(plot_ViolinPlot(calc_CommonDose(df, plot = FALSE,
                                                verbose = FALSE)))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("ViolinPlot defaults",
                              plot_ViolinPlot(df))
  vdiffr::expect_doppelganger("ViolinPlot summary sub",
                              plot_ViolinPlot(df, summary.pos = "sub",
                                              summary = c("n", "serel", "kurtosis")))
  vdiffr::expect_doppelganger("ViolinPlot summary left",
                              plot_ViolinPlot(df, summary.pos = "left",
                                              summary = c("mean", "skewness")))
  })
})

#'@title Remove Signal Background from `RLum.Data.Curve` Objects
#'
#'@description Convenient (background) of luminescence curves using [Luminescence::merge_RLum]
#'
#'@details
#'
#'The function aims to simplify a frequently encountered task: subtracting curves
#'or backgrounds from luminescence curves, such as OSL, TL, or RF. The function
#'presumes that if curves are presented in pairs, for instance, TL - TL,
#'the second curve represents a background signal that needs to be removed
#'from the first curve. Following the removal, the background curve is discarded from
#'the dataset. Alternatively, custom background curves can be provided,
#'which are then utilised for the subtraction. In essence, the function
#'utilises the [Luminescence::merge_RLum] function but simplifies the selection of pairs and curves.
#'
#'@param object [Luminescence::RLum.Analysis-class] (**required**): A non-empty [Luminescence::RLum.Analysis-class] object
#'with, e.g., OSL/IRSL/TL curves or a [list] of such object. If a list is provided
#'non-conform list elements are silently removed from the [list]
#'
#'@param object_bg [Luminescence::RLum.Data.Curve-class], a [list] of such objects, a [matrix] or [numeric] (*optional*):
#'Sets the background as a curve that is subtracted from the record types set with `recordType`.
#'If you provide a [matrix] or [numeric] internally, everything is coerced to a [Luminescence::RLum.Data.Curve-class] object.
#'If you desire full freedom, you can construct a [list] of [Luminescence::RLum.Data.Curve-class] objects.
#'
#'@param recordType [character] (*optional*): provide the `recordType` subject to the
#'background subtraction. Subsequent curve selection uses [Luminescence::get_RLum].
#'If set to `NULL` the record type of highest occurrence will be used. Example: `recordType = "TL (UVVIS)"`
#'
#'@param clean_up [logical] (*with default*): enable/disable background curve removal
#'after background subtraction. If `object_bg` is set, nothing is removed from
#'the input object as the background is already stored separately.
#'
#'@returns Returns an [Luminescence::RLum.Analysis-class] object or a [list] of such objects.
#'
#'@section Function version: 0.1.0
#'
#'@author
#'Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@seealso [Luminescence::get_RLum], [Luminescence::merge_RLum], [Luminescence::RLum.Analysis-class],
#'[Luminescence::RLum.Data.Curve-class]
#'
#'@keywords datagen
#'
#'@examples
#'
#'## load example dataset
#'xsyg <- read_XSYG2R(
#'system.file("extdata/XSYG_file.xsyg", package = "Luminescence"),
#'fastForward = TRUE,
#'verbose = FALSE)
#'
#'## remove constant background from OSL curves
#'remove_SignalBackground(
#' object = xsyg,
#' object_bg = 100,
#' recordType = "OSL (UVVIS)")
#'
#' ## use a more elaborate examples
#' ## with two TL curves (2nd is background)
#' xsyg_v1 <- set_RLum("RLum.Analysis", records = c(
#'  rep(xsyg[[1]]@records[[1]], 2),
#'  xsyg[[1]]@records[[4]],
#'  xsyg[[1]]@records[[4]],
#'  rep(xsyg[[1]]@records[[10]], 2),
#'  xsyg[[1]]@records[[4]],
#'  xsyg[[1]]@records[[4]]))
#'
#' ## remove background and strip background
#' ## curves from the object
#' o <- remove_SignalBackground(
#'  object = xsyg_v1,
#'  recordType = "TL (UVVIS)")
#'
#'@export
remove_SignalBackground <- function(
  object,
  object_bg = NULL,
  recordType = NULL,
  clean_up = TRUE
) {
  .set_function_name("remove_SignalBackground")
  on.exit(.unset_function_name(), add = TRUE)

  # Self-call ---------------------------------------------------------------
  if(inherits(object, "list")) {
    ## no expansion not special treatment except for silent object removal
    return(
      lapply(
        X = .rm_nonRLum(object, "RLum.Analysis"),
        FUN = remove_SignalBackground,
        object_bg = object_bg,
        recordType = recordType,
        clean_up = clean_up)
    )
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis")
  .validate_not_empty(object)
  .validate_class(object_bg, null.ok = TRUE,
                  c("RLum.Data.Curve","list", "matrix", "numeric", "integer"))
  .validate_class(recordType, "character", null.ok = TRUE)
  .validate_logical_scalar(clean_up)

  # Find curves for removal -------------------------------------------------
  ## if nothing is set, we do quick and dirty recordType guess based on the
  ## occurrence
  if(is.null(recordType))
    recordType <- names_RLum(object)|> table() |> which.max() |> names() |> unique()

  ## get index of relevant curves; these are the only record types of concern
  id_pairs <- suppressWarnings(
    get_RLum(object, recordType = recordType[1], get.index = TRUE))

    ## be gentle if the recordType does not exist, this behaviour
    ## should make it easier in case we process a list
    if(is.null(id_pairs)) {
      .throw_warning("'recordType' setting invalid, nothing removed.")
      return(object)
    }

  if(!is.null(object_bg)) {
    ## if it is a vector recycle if required using the first curve
    if(inherits(object_bg, c("matrix", "numeric", "integer"))) {
      m_ref <- object@records[[id_pairs[1]]]
      object_bg <- as.matrix(object_bg)

      object_bg <- set_RLum(
        "RLum.Data.Curve",
        recordType = recordType,
        data = matrix(
          c(m_ref[,1],
            rep_len(object_bg[, max(ncol(object_bg))], nrow(m_ref))),
          ncol = 2))
    }

    ## check for list
    if(inherits(object_bg, "RLum.Data.Curve"))
      object_bg <- list(object_bg)

    ## if we have an object for bg, we take this an recycle it to the length
    object_bg <- rep_len(object_bg, length(id_pairs))

    ## set id signal
    id_signal <- id_bg <- id_pairs
    object_bg[id_bg] <- object_bg

  } else {
    ## create matrix with pairs; we recycle on purpose
    m_pairs <- suppressWarnings(matrix(id_pairs, ncol = 2, byrow = TRUE))

    ## calculate difference between pairs ... if 1 we have a pair,
    ## if negative it is invalid because we recycle
    m_pairs <- cbind(m_pairs, matrixStats::rowDiffs(m_pairs))

    ## get signal id and background id
    id_signal <- m_pairs[m_pairs[,3] == 1,1]
    id_bg <- m_pairs[m_pairs[,3] == 1,2]

    ## set object_bg to have it consistent
    ## this creates an empty list elements, which we will keep for the index
    object_bg[id_bg] <- object@records[id_bg]
  }

  # Subtract ----------------------------------------------------------------
  ## now we perform the background subtraction element by element using our
  ## merge function, this makes things a lot easier
  object@records[id_signal] <- lapply(seq_along(id_signal), function(x) {
    merge_RLum.Data.Curve(
      object = list(object@records[id_signal[x]][[1]],object_bg[id_bg[x]][[1]]),
      merge.method = "-")
  })

  # Return ------------------------------------------------------------------
  if (clean_up && any(id_signal != id_bg))
    object@records[id_bg] <- NULL

  ## return whatever is left
  return(object)
}

#' @title Apply the (un-)logged common age model after Galbraith et al. (1999) to a
#' given De distribution
#'
#' @description
#'
#' Function to calculate the common dose of a De distribution.
#'
#' @details
#' **(Un-)logged model**
#'
#' When `log = TRUE` this function
#' calculates the weighted mean of logarithmic De values. Each of the estimates
#' is weighted by the inverse square of its relative standard error. The
#' weighted mean is then transformed back to the dose scale (Galbraith &
#' Roberts 2012, p. 14).
#'
#' The log transformation is not applicable if the
#' De estimates are close to zero or negative. In this case the un-logged model
#' can be applied instead (`log = FALSE`). The weighted mean is then
#' calculated using the un-logged estimates of De and their absolute standard
#' error (Galbraith & Roberts 2012, p. 14).
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with De `(data[,1])` and De error `(data[,2])`
#'
#' @param sigmab [numeric] (*with default*):
#' additional spread in De values, representing the expected overdispersion in
#' the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
#' **Note:** For the logged model (`log = TRUE`) this value must be
#' a fraction, e.g. 0.2 (= 20%). If the un-logged model is used (`log = FALSE`),
#' `sigmab` must be provided in the same absolute units of the De values
#' (seconds or Gray).
#'
#' @param log [logical] (*with default*):
#' fit the (un-)logged central age model to De data
#'
#' @param ... currently not used.
#'
#' @return
#' Returns a terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following element:
#'
#' \item{$summary}{[data.frame] summary of all relevant model results.}
#' \item{$data}{[data.frame] original input data}
#' \item{$args}{[list] used arguments}
#' \item{$call}{[call] the function call}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Christoph Burow, University of Cologne (Germany)
#'
#' @seealso [Luminescence::calc_CentralDose], [Luminescence::calc_FiniteMixture],
#' [Luminescence::calc_FuchsLang2001], [Luminescence::calc_MinDose]
#'
#' @references
#' Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for
#' mixed fission track ages. Nuclear Tracks Radiation Measurements 4, 459-470.
#'
#' Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley,
#' J.M., 1999. Optical dating of single grains of quartz from Jinmium rock
#' shelter, northern Australia. Part I: experimental design and statistical
#' models.  Archaeometry 41, 339-364.
#'
#' Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error calculation and
#' display in OSL dating: An overview and some recommendations. Quaternary
#' Geochronology 11, 1-27.
#'
#' **Further reading**
#'
#' Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
#' (De) distributions: Implications for OSL dating of sediment mixtures.
#' Quaternary Geochronology 4, 204-230.
#'
#' Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
#' assessment of procedures for estimating burial dose. Quaternary Science
#' Reviews 25, 2475-2502.
#'
#' Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
#' Quaternary Geochronology 12, 98-106.
#'
#' Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
#' of optical dating of fluvial deposits.  Quaternary Geochronology 1 109-120.
#'
#' Rodnight, H., 2008. How many equivalent dose values are needed to
#' obtain a reproducible distribution?. Ancient TL 26, 3-10.
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## apply the common dose model
#' calc_CommonDose(ExampleData.DeValues$CA1)
#'
#' @export
calc_CommonDose <- function(
  data,
  sigmab = 0,
  log=TRUE,
  ...
) {
  .set_function_name("calc_CommonDose")
  on.exit(.unset_function_name(), add = TRUE)

  ##============================================================================##
  ## CONSISTENCY CHECK OF INPUT DATA
  ##============================================================================##

  .validate_class(data, c("data.frame", "RLum.Results"))
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")
  }
  if (ncol(data) < 2) {
    .throw_error("'data' object must have two columns")
  }
  if (any(data[, 1] < 0)) {
    .throw_error("'data' cannot contain negative times")
  }
  .validate_nonnegative_scalar(sigmab)
  .validate_logical_scalar(log)
  if (log && (sigmab < 0 || sigmab > 1)) {
    .throw_error("'sigmab' must be a value between 0 and 1 if 'log = TRUE'")
  }

  ## set expected column names
  colnames(data)[1:2] <- c("ED", "ED_Error")


  ##============================================================================##
  ## ADDITIONAL ARGUMENTS
  ##============================================================================##

  settings <- modifyList(list(verbose = TRUE),
                         list(...))
  .validate_logical_scalar(settings$verbose, name = "'verbose'")


  ##============================================================================##
  ## CALCULATIONS
  ##============================================================================##

  # calculate  yu = log(ED) and su = se(logED)
  if (log) {
    yu<- log(data$ED)
    su<- sqrt( (data$ED_Error/data$ED)^2 + sigmab^2 )
  }
  else {
    yu<- data$ED
    su<- sqrt((data$ED_Error)^2 + sigmab^2)
  }

  # calculate weights
  wu<- 1/su^2
  delta<- sum(wu*yu)/sum(wu)
  n<- length(yu)

  #standard error
  sedelta<- 1/sqrt(sum(wu))
  if (!log) {
    sedelta<- sedelta/delta
  }

  if (log){
    delta<- exp(delta)
  }

  ##============================================================================##
  ## TERMINAL OUTPUT
  ##============================================================================##

  if (settings$verbose) {
    cat("\n [calc_CommonDose]")
    cat("\n\n----------- meta data --------------")
    cat("\n n:                      ", n)
    cat("\n log:                    ", log)
    cat("\n----------- dose estimate ----------")
    cat("\n common dose:            ", round(delta, 2))
    cat("\n SE:                     ", round(delta * sedelta, 2))
    cat("\n rel. SE [%]:            ", round(sedelta * 100, 2))
    cat("\n------------------------------------\n\n")
  }

  ##============================================================================##
  ## RETURN VALUES
  ##============================================================================##

  summary<- data.frame(de=delta,
                       de_err=delta*sedelta)

  call<- sys.call()
  args<- list(log=log, sigmab=sigmab)

  set_RLum(
    class = "RLum.Results",
    data = list(summary = summary,
                data = data,
                args = args,
                call = call))
}

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_CommonDose(),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_CommonDose("test"),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_CommonDose(data.frame(col = 1:10)),
               "'data' object must have two columns")
  expect_error(calc_CommonDose(data.frame(col = 1:10)),
               "'data' object must have two columns")
  expect_error(calc_CommonDose(data.frame(a = c(12, -2.1), b = 1:2)),
               "'data' cannot contain negative times")
  expect_error(calc_CommonDose(ExampleData.DeValues$CA1, sigmab = iris),
               "'sigmab' should be a single non-negative value")
  expect_error(calc_CommonDose(ExampleData.DeValues$CA1, sigmab = 2),
               "'sigmab' must be a value between 0 and 1 if 'log = TRUE'")
  expect_error(calc_CommonDose(iris, log = numeric()),
               "'log' should be a single logical value")
  expect_error(calc_CommonDose(ExampleData.DeValues$CA1, verbose = NA),
               "'verbose' should be a single logical value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
    temp <- calc_CommonDose(ExampleData.DeValues$CA1, plot = FALSE,
                            verbose = TRUE)
    temp.nolog <- calc_CommonDose(ExampleData.DeValues$CA1, log = FALSE,
                                  plot = FALSE, verbose = TRUE)
  })
  expect_s4_class(temp, "RLum.Results")
  expect_equal(length(temp), 4)

  all.equal(calc_CommonDose(temp, verbose = FALSE),
            temp)

  results <- get_RLum(temp)
  expect_equal(round(results$de, digits = 5), 62.15999)
  expect_equal(round(results$de_err, digits = 7), 0.7815117)
  expect_true(temp@data$args$log)
  expect_equal(temp@data$args$sigmab, 0)

  results <- get_RLum(temp.nolog)
  expect_false(temp.nolog@data$args$log)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(calc_CommonDose(
      ExampleData.DeValues$CA1, sigmab = 0.75,
      plot = FALSE, verbose = FALSE),
      tolerance = snapshot.tolerance)
  expect_snapshot_RLum(calc_CommonDose(
      ExampleData.DeValues$CA1, sigmab = 0.25, log = FALSE,
      plot = FALSE, verbose = FALSE),
      tolerance = snapshot.tolerance)
  expect_snapshot_RLum(calc_CommonDose(
      ExampleData.DeValues$CA1, sigmab = 1.25, log = FALSE,
      plot = FALSE, verbose = FALSE),
      tolerance = snapshot.tolerance)
})

#' @title Moran Scatter Plot: Visualizing Spatial Dependency
#'
#' @description
#' The function produces a scatter plot, with on the x axis the original grain
#' signal and on the y axis the weighted mean of the neighbour grain signals.
#' The plot area is divided into four quadrants, and also a least square line
#' (which slopes indicates, but not exactly represents, Moran's I) and
#' an 1:1 line (which indicates a Moran's I of around 1).
#'
#' @param object [Luminescence::RLum.Results-class] or [numeric] (**required**):
#' containing a numerical vector of length 100,
#' representing one or more measurement discs ("positions") in a reader.
#' Each element in the vector represents one grain hole location on a disc.
#'
#' @param df_neighbours [data.frame] (*with default*):
#' Data frame indicating
#' which borders to consider, and their respective weights (see the description
#' provided for [Luminescence::calc_MoransI]). If `NULL` (default), this is constructed
#' automatically by the internal function `.get_Neighbours`.
#'
#' @param str_y_def [character] (*with default*):
#' Calculation of y position. Defaults to `"mean_neighbours"`
#' which is the plain mean of all neighbour values and the most illustrative. The other option is `"weighted_sum"`,
#' which means the sum of the border weights times the neighbour values, which is actually closer
#' to the way Moran's I is by default calculated in this package.
#'
#' @param ... Other parameters to be forwarded to the base R plot functions.
#' `legend` (`TRUE/FALSE`) to enable/disable the legend.
#' Note that `xlab` (x axis label), `ylab` (y axis label) and `cex` (scaling
#' value) are given default values. Because of sometimes large value
#' differences, `log = "x"`, `log = "y"` and `log = "xy"`are supported.
#' In case of negative values and logarithmic plotting, values are increased
#' so the smallest value to plot is 1. Summary elements such as means, least
#' square line etc. will still be based on the linear case.
#' `pch` accepts options `"show_location_ids"` (plots grain location id's),
#' `"show_n_neighbours"` (indicates numbers of neighbours) and the normal
#' base plot `pch` options.
#'
#' @details Note that this function plots on the y-axis the mean of the neighbours, while the function
#' [Luminescence::calc_MoransI] by default will for its global calculation weight every border the same. So, grain locations
#' with 1, 2 or 3 neighbours will appear higher on the y-axis than their influence on Moran's I justify -- apart
#' from scaling, this explains a part of the differences of Moran's scatter plots between different packages.
#' Also note that island' grain locations (=those not bordering other grains) are left out of these plots but
#' might still influence Moran's I calculations.
#'
#' @return
#' Returns (invisibly) a data frame with plotting coordinates and
#' grain location id's, for creating user-defined plots.
#'
#' @examples
#' plot_MoranScatterplot(1:100)
#'
#' @author
#' Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
#' Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
#'
#' @references
#' de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
#' A novel method to assess crosstalk in single-grain luminescence detection.
#' Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
#'
#' @export
plot_MoranScatterplot <- function(
    object,
    df_neighbours = NULL,
    str_y_def = c("mean_neighbours", "weighted_sum"),
    ...
) {
  .set_function_name("plot_MoranScatterplot")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Results", "numeric", "integer"))
  if (is.numeric(object)) {
    vn_values <- object
  } else {
    vn_values <- get_RLum(object)
  }
  .validate_length(vn_values, 100, name = "'object'")

  .validate_class(df_neighbours, "data.frame", null.ok = TRUE)
  if (is.null(df_neighbours)) {
    df_neighbours <- .get_Neighbours(object)
  } else if (ncol(df_neighbours) != 3) {
    .throw_error("'df_neighbours' should be a data frame with 3 columns")
  }

  str_y_def <- .validate_args(str_y_def, c("mean_neighbours", "weighted_sum"))

  ## get ... arguments
  plot_settings <- modifyList(
      x = list(
          legend = TRUE,
          pch = 20,
          cex = 0.8,
          xlab = "Value central grain",
          ylab = ifelse(str_y_def == "weighted_sum",
                        "Weighted sum neighbouring grains",
                        "Plain mean neighbour grains"),
          log = ""),
      val = list(...))

  vs_points_appearance <- plot_settings$pch

  show_location_ids <- FALSE
  show_n_neighbours <- FALSE
  if(plot_settings$pch == "show_location_ids") {
    show_location_ids <- TRUE

  } else if (plot_settings$pch == "show_n_neighbours") {
    show_n_neighbours <- TRUE
 }

  vs_log <- plot_settings$log  # because we need the function "log" later

  df_moran_plot <- data.frame(
    x = vn_values,
    y = NA,
    grain_id = NA,
    n_neighbours = NA)

  vn_locations <- unique(c(df_neighbours$location, df_neighbours$neighbour))
  for(i in 1:length(vn_locations) ) {
    n_location <- vn_locations[i]
    ## Find the neighbouring locations according to df_neighbour
    ## For a given n_location, in which row of df_neighbour are neighbouring locations provided?
    vi_neighbour_row <- which(df_neighbours$location == n_location)
    ## And thus, which locations are provided in these rows?
    vn_neighbour <- df_neighbours$neighbour[vi_neighbour_row]
    ## And add weights
    vn_weights <- df_neighbours$weight[vi_neighbour_row]
    df_neighbour_and_weights <- data.frame(adj_pos = vn_neighbour,
                                           weight = vn_weights)

    ## Find the mirror neighbouring locations ("locations") according to df_neighbour
    ## For a given n_location, in which row of df_neighbour are locations provided?
    vi_neighbour_mirror_row <- which(df_neighbours$neighbour == n_location)
    ## And thus, which locations are provided in these rows?
    vn_neighbour_mirror <- df_neighbours$location[vi_neighbour_mirror_row]
    ## And add weights
    vn_weights <- df_neighbours$weight[vi_neighbour_mirror_row]

    df_neighbour_and_weights <- rbind(df_neighbour_and_weights,
                                      data.frame(adj_pos = vn_neighbour_mirror,
                                                 weight = vn_weights)
    )


    if(str_y_def == "mean_neighbours") {
      n_mean_of_neighbours <- mean(
        df_moran_plot$x[df_neighbour_and_weights$adj_pos]*
          df_neighbour_and_weights$weight, na.rm = TRUE)

      df_moran_plot$y[n_location]  <- n_mean_of_neighbours

    } else if(str_y_def == "weighted_sum") {
      n_weighted_sum_neighbours <- sum(
        df_moran_plot$x[df_neighbour_and_weights$adj_pos]*
          df_neighbour_and_weights$weight, na.rm = TRUE) ## NOG AANPASSEN

      df_moran_plot$y[n_location]  <- n_weighted_sum_neighbours
    }

    ## For plotting: how many neighbours were involved?
    df_moran_plot$n_neighbours[n_location] <-
      length(vi_neighbour_mirror_row) + length(vi_neighbour_row)

    df_moran_plot$grain_id[n_location] <- n_location
  }

  n_grains_without_neighbour <- nrow(df_moran_plot) -
    sum(stats::complete.cases(df_moran_plot$y))

  if(n_grains_without_neighbour>0)
    .throw_warning("Grain observations removed from plot because no neighbours")

  df_moran_plot <- df_moran_plot[stats::complete.cases(df_moran_plot), ]

  ## If we want a log plot, make sure that all relevant values are 1 or more (per axis)
  if (grepl("x", vs_log, fixed = TRUE) && min(df_moran_plot$x) < 1) {
    if (min(df_moran_plot$x) < 0) {
      df_moran_plot$x <- df_moran_plot$x - min(df_moran_plot$x) + 1
    } else {
      df_moran_plot$x <- df_moran_plot$x + (1 - min(df_moran_plot$x))
    }
    .throw_warning("x-axis values rescaled because of log transform ",
                   "(also in return df)")
  }

  if (grepl("y", vs_log, fixed = TRUE) && min(df_moran_plot$y) < 1) {
    if (min(df_moran_plot$y) < 0) {
      df_moran_plot$y <- df_moran_plot$y - min(df_moran_plot$y) + 1

    } else {
      df_moran_plot$y <- df_moran_plot$y + (1 - min(df_moran_plot$y))
    }
    .throw_warning("y-axis values rescaled because of log transform ",
                   "(also in return df)")
  }

  ## Set point appearance
  vs_pch <- if (vs_points_appearance == "show_location_ids" ||
                vs_points_appearance == "show_n_neighbours") {
    NA # Numbers will be added later
  } else {
    plot_settings$pch
  }

  ## plot
  ## remove used arguments to avoid collusion
  args <- list(...)
  args <- args[!...names() %in% c("pch", "log", "xlab", "ylab", "cex", "legend")]

  do.call(
    what = plot,
    args = c(
     args,
     list(
      df_moran_plot[,1:2],
      pch = vs_pch,
      log = vs_log,
      xlab = plot_settings$xlab,
      ylab = plot_settings$ylab,
      cex = plot_settings$cex)))

  if(vs_points_appearance == "show_location_ids") {
    text(x = df_moran_plot[,1:2],
         labels = df_moran_plot$grain_id,
         cex = plot_settings$cex)
  }

  if(vs_points_appearance == "show_n_neighbours") {
    text(x = df_moran_plot[,1:2],
         labels = df_moran_plot$n_neighbours,
         cex = plot_settings$cex)
  }


  ## Mean lines
  abline(v = mean(df_moran_plot$x), h = mean(df_moran_plot$y), lty = "dotted", untf = TRUE)

  ## 1:1 one, only makes only sense for the plain mean of neighbours
  if(str_y_def == "mean_neighbours")
  {
    abline(a = mean(df_moran_plot$y) - mean(df_moran_plot$x), b = 1, lty = "dotdash", untf = TRUE)
  }

  ## Calculate least square line
  lm_moransI <- stats::lm(y ~ x, data = df_moran_plot)
  abline(a = coef(lm_moransI)[1], b = coef(lm_moransI)[2], col = "orange", untf=TRUE)

  if (plot_settings$legend[1]) {
    vs_legend <- c("means",  "least square")
    vs_cols   <- c("black",  "orange")
    vs_lty    <- c("dotted", "solid")
    vs_pch    <- c(NA,       NA)

    if(str_y_def == "mean_neighbours") {
      vs_legend <- c(vs_legend, "1:1 line")
      vs_cols   <- c(vs_cols,   "black")
      vs_lty    <- c(vs_lty,    "dotdash")
      vs_pch    <- c(vs_pch,    NA)
    }

    if(vs_points_appearance == "show_location_ids") {
      vs_legend <- c(vs_legend, "location ids")
      vs_cols   <- c(vs_cols,   "black")
      vs_lty    <- c(vs_lty,    NA)
      vs_pch    <- c(vs_pch,    "#")
    }

    if(vs_points_appearance == "show_n_neighbours") {
      vs_legend <- c(vs_legend, "numb. of neighb.")
      vs_cols   <- c(vs_cols,   "black")
      vs_lty    <- c(vs_lty,    NA)
      vs_pch    <- c(vs_pch,    "#")
    }

    legend(x = "topleft",
           legend = vs_legend,
           col    = vs_cols,
           lty    = vs_lty,
           pch = vs_pch,
           cex    = 0.8)
  }

  invisible(df_moran_plot)
}

## load data
obj <- set_RLum(class = "RLum.Results",
                data = list(vn_values = rep(x = c(1,2), each = 50)))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_MoranScatterplot("error"),
               "'object' should be of class 'RLum.Results', 'numeric' or")
  expect_error(plot_MoranScatterplot(obj, df_neighbours = "error"),
               "'df_neighbours' should be of class 'data.frame'")
  expect_error(plot_MoranScatterplot(obj, df_neighbours = data.frame()),
               "'df_neighbours' should be a data frame with 3 columns")
  expect_error(plot_MoranScatterplot(obj, str_y_def = "error"),
               "'str_y_def' should be one of 'mean_neighbours' or")
  expect_error(plot_MoranScatterplot(1:10),
               "'object' should have length 100")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_silent(plot_MoranScatterplot(obj))
  expect_silent(plot_MoranScatterplot(1:100,
                                      pch = "show_location_ids"))
  expect_silent(plot_MoranScatterplot(1:100,
                                      pch = "show_n_neighbours"))
  expect_silent(plot_MoranScatterplot(obj, str_y_def = "weighted_sum",
                                      pch = 16))

  ## problematic values for the log transform
  expect_warning(plot_MoranScatterplot(1:100 - 1, log = "xy",
                                       legend = FALSE),
                 "x-axis values rescaled because of log transform")
  expect_warning(plot_MoranScatterplot(1:100 - 2, log = "xy",
                                       legend = FALSE),
                 "x-axis values rescaled because of log transform")
  expect_warning(plot_MoranScatterplot((1:100) - 20, log = "y",
                                       legend = FALSE),
                 "y-axis values rescaled because of log transform \\(also in return df\\)")
  expect_warning(plot_MoranScatterplot((1:100) - 4.5, log = "y",
                                       legend = FALSE),
                 "y-axis values rescaled because of log transform \\(also in return df\\)")

  expect_silent(plot_MoranScatterplot((1:100) - 2, log = "y",
                                       legend = FALSE))

  obj.na <- obj
  obj.na@data$vn_values[c(24, 73)] <- NA
  expect_warning(plot_MoranScatterplot(obj.na, legend = TRUE),
                 "Grain observations removed from plot because no neighbours")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("MoranScatterplot defaults",
                              plot_MoranScatterplot(obj))
  vdiffr::expect_doppelganger("MoranScatterplot show_location_ids",
                              plot_MoranScatterplot(1:100,
                                                    pch = "show_location_ids"))
  vdiffr::expect_doppelganger("MoranScatterplot show_n_neighbours",
                              plot_MoranScatterplot(1:100,
                                                    pch = "show_n_neighbours"))
  })
})

#' @title Converts Single-Grain Data to Multiple-Grain Data
#'
#' @description Conversion of single-grain data to multiple-grain data by adding signals
#' from grains belonging to one disc (unique pairs of position, set and run).
#'
#' @param object [Luminescence::Risoe.BINfileData-class], [character] (**required**):
#' [Luminescence::Risoe.BINfileData-class] object or BIN/BINX-file name.
#'
#' @param write_file [logical] (*with default*):
#' whether the output should be written to a file (only considered if `object`
#' is of type character). The multiple grain file will be written to a file
#' of named after the original one with `-SG` appended to it.
#'
#' @param ... further arguments passed down to [Luminescence::read_BIN2R] and
#' [Luminescence::write_R2BIN] if `object` is a path to a file.
#'
#' @return
#' A [Luminescence::Risoe.BINfileData-class] object. If `write_file = TRUE` an
#' the input was a file path, a file is also written out to origin folder.
#'
#' @section Function version: 0.1.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Norbert Mercier, IRAMAT-CRP2A, UMR 5060, CNRS-Université Bordeaux Montaigne (France)
#'
#' @seealso [Luminescence::Risoe.BINfileData-class], [Luminescence::read_BIN2R],
#' [Luminescence::write_R2BIN]
#'
#' @keywords IO
#'
#' @examples
#' ## simple run
#' ## (please not that the example is not using SG data)
#' data(ExampleData.BINfileData, envir = environment())
#' convert_SG2MG(CWOSL.SAR.Data)
#'
#' @export
convert_SG2MG <- function(
  object,
  write_file = FALSE,
  ...
) {
  .set_function_name("convert_SG2MG")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("character", "Risoe.BINfileData"))
  .validate_not_empty(object)
  .validate_logical_scalar(write_file)

  file_name <- NULL
  if (!inherits(object, "Risoe.BINfileData")) {
    .validate_length(object, 1)
    file_name <- normalizePath(object, mustWork = FALSE)
    object <- read_BIN2R(object, ...)
  }

  ## Transform --------------------------------------------------------------
  ## reset the rownames because they may not correspond to the row indices
  ## (see #1415)
  rownames(object@METADATA) <- NULL

  ## ids of single grain records
  sg_id <- as.numeric(rownames(
      unique(object@METADATA[object@METADATA[["GRAIN"]] != 0, "ID", drop = FALSE])))

  for (i in sg_id) {
    ##get IDs of all relevant records
    records_id <- object@METADATA[
      object@METADATA[["POSITION"]] == object@METADATA[["POSITION"]][[i]] &
      object@METADATA[["RUN"]] == object@METADATA[["RUN"]][[i]] &
      object@METADATA[["SET"]] == object@METADATA[["SET"]][[i]],
      "ID"]

    # sum up count values and write it into the first grain record
    object@DATA[[i]] <- matrixStats::rowSums2(
      matrix(
        unlist(object@DATA[records_id]),
        ncol = length(records_id)))
  }

  ## clean dataset and remove all irrelevant data
  upairs_id <- as.numeric(rownames(
    unique(object@METADATA[, c("POSITION", "RUN", "SET")])))

  object@METADATA <- object@METADATA[upairs_id, ]
  object@DATA <- object@DATA[upairs_id]

  ##recalculate IDs and reset GRAIN
  object@METADATA[["ID"]] <- 1:length(object@DATA)
  object@METADATA[["GRAIN"]] <- 0

# Write file --------------------------------------------------------------
  if (write_file && !is.null(file_name)) {
    write_R2BIN(object,
                file = paste0(tools::file_path_sans_ext(file_name), "_SG.",
                              tools::file_ext(file_name)),
                ...)
  }

  object
}

#' @title Al2O3:C Reader Cross-Talk Analysis
#'
#' @description The function provides the analysis of cross-talk measurements
#' on a FI lexsyg SMART reader using Al2O3:C chips.
#'
#' @param object [Luminescence::RLum.Analysis-class] or [list] (**required**):
#' measurement input
#'
#' @param signal_integral [numeric] (*optional*):
#' signal integral, used for the signal and the background.
#' If nothing is provided, the full range is used.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param dose_points [numeric] (*with default*):
#' vector with dose points, if dose points are repeated, only the general
#' pattern needs to be provided. Default values follow the suggestions
#' made by Kreutzer et al., 2018.
#'
#' @param recordType [character] (*with default*):
#' input curve selection, which is passed to [Luminescence::get_RLum]. To deactivate the
#' automatic selection set the argument to `NULL`.
#'
#' @param irradiation_time_correction [numeric] or [Luminescence::RLum.Results-class] (*optional*):
#' information on the used irradiation time correction obtained by another
#' experiment.
#'
#' @param method_control [list] (*optional*):
#' optional parameters to control the calculation.
#' See details for further explanations.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments that can be passed to the plot output
#'
#' @return
#' Function returns results numerically and graphically:
#'
#'  -----------------------------------\cr
#'  `[ NUMERICAL OUTPUT ]`\cr
#'  -----------------------------------\cr
#'
#'  **`RLum.Results`**-object
#'
#'  **slot:** **`@data`**
#'
#'  \tabular{lll}{
#'   **Element** \tab **Type** \tab **Description**\cr
#'   `$data` \tab `data.frame` \tab summed apparent dose table \cr
#'   `$data_full` \tab `data.frame` \tab full apparent dose table \cr
#'   `$fit` \tab `lm` \tab the linear model obtained from fitting \cr
#'   `$col.seq` \tab `numeric` \tab the used colour vector \cr
#'  }
#'
#' **slot:** **`@info`**
#'
#' The original function call
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' - An overview of the obtained apparent dose values
#'
#' @section Function version: 0.1.4
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::analyse_Al2O3C_ITC]
#'
#' @references
#'
#' Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018. Environmental Dose Rate
#' Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
#' Geochronometria 45, 56-67. doi: 10.1515/geochr-2015-0086
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.Al2O3C, envir = environment())
#'
#' ##run analysis
#' analyse_Al2O3C_CrossTalk(data_CrossTalk)
#'
#' @export
analyse_Al2O3C_CrossTalk <- function(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  dose_points = c(0,4),
  recordType = "OSL (UVVIS)",
  irradiation_time_correction = NULL,
  method_control = NULL,
  plot = TRUE,
  ...
) {
  .set_function_name("analyse_Al2O3C_CrossTalk")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Analysis", "list"))
  .validate_not_empty(object, class(object)[1])
  if (is.list(object)) {
    lapply(object, .validate_class, class = "RLum.Analysis",
           name = "All elements of 'object'")
  } else {
    object <- list(object)
  }
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  .validate_class(dose_points, c("numeric", "integer"))
  .validate_not_empty(dose_points)
  if (length(dose_points) != 1 && length(dose_points) %% 2 != 0) {
    .throw_error("'dose_points' should have length 1 or divisible by 2")
  }
  .validate_class(recordType, "character", null.ok = TRUE)
  .validate_class(irradiation_time_correction, c("numeric", "RLum.Results"),
                  null.ok = TRUE)
  .validate_class(method_control, "list", null.ok = TRUE)

  ##TODO ... do more, push harder
  ##Accept the entire sequence ... including TL and extract
  ##Add sufficient unit tests

  ## Preparation ------------------------------------------------------------

  ##select curves based on the recordType selection; if not NULL
  if(!is.null(recordType)){
    object <- suppressWarnings(get_RLum(object, recordType = recordType,
                                        drop = FALSE))
  }
  if (is.null(object) || all(lengths(object) == 0)) {
    .throw_error("'object' contains no records with recordType = '", recordType, "'")
  }

  #set method control
  method_control_settings <- list(
    fit.method = "EXP"
  )

  ##modify on request
  if(!is.null(method_control)){
    method_control_settings <- modifyList(x = method_control_settings, val = method_control)
  }

  ## signal integral
  x.range <- object[[1]]@records[[1]][, 1]
  if (integral_input == "measurement") {
    signal_integral <- .convert_to_channels(x.range, signal_integral,
                                            "time", null.ok = TRUE)
  }
  signal_integral <- .validate_integral(signal_integral,
                                        max = length(x.range), null.ok = TRUE)
  if (is.null(signal_integral)) {
    signal_integral <- seq_along(x.range)
  }

  ##check irradiation time correction
    if (inherits(irradiation_time_correction, "RLum.Results")) {
      .validate_originator(irradiation_time_correction, "analyse_Al2O3C_ITC")
      irradiation_time_correction <- get_RLum(irradiation_time_correction)

        ##insert case for more than one observation ...
        if(nrow(irradiation_time_correction)>1){
          irradiation_time_correction <- c(mean(irradiation_time_correction[[1]]), sd(irradiation_time_correction[[1]]))

        }else{
          irradiation_time_correction <- c(irradiation_time_correction[[1]], irradiation_time_correction[[2]])
        }
    }

  # Calculation ---------------------------------------------------------------------------------
  ##we have two dose points, and one background curve, we do know only the 2nd dose

  ##create signal table list
  signal_table_list <- lapply(object, function(x) {
    ##calculate all the three signals needed
    NATURAL <- sum(x[[1]][signal_integral, 2])
    REGENERATED <- sum(x[[2]][signal_integral, 2])
    BACKGROUND <- sum(x[[3]][signal_integral, 2])

    temp_df <- data.frame(
      POSITION = get_RLum(x[[1]], info.object = "position"),
      DOSE = if(!is.null(irradiation_time_correction)){
        dose_points + irradiation_time_correction[1]
      }else{
        dose_points
      },
      DOSE_ERROR = if(!is.null(irradiation_time_correction)){
        dose_points * irradiation_time_correction[2]/irradiation_time_correction[1]
      }else{
        0
      },
      STEP = c("NATURAL", "REGENERATED"),
      INTEGRAL = c(NATURAL, REGENERATED),
      BACKGROUND = c(BACKGROUND, BACKGROUND),
      NET_INTEGRAL = c(NATURAL, REGENERATED) - BACKGROUND,
      row.names = NULL
    )

    ##0 dose points should not be biased by the correction ..
    id_zero <- which(dose_points == 0)
    temp_df$DOSE[id_zero] <- 0
    temp_df$DOSE_ERROR[id_zero] <- 0

    return(temp_df)
  })

  APPARENT_DOSE <- data.table::rbindlist(lapply(signal_table_list, function(x) {
    ##run in MC run
    DOSE <- if (!is.null(irradiation_time_correction)) {
              rnorm(1000, mean = x$DOSE[2], sd = x$DOSE_ERROR[2])
            } else {
              x$DOSE[2]
            }

    ##calculation
    temp <- (DOSE * x$NET_INTEGRAL[1]) / x$NET_INTEGRAL[2]

    data.frame(
      POSITION = x$POSITION[1],
      AD = mean(temp),
      AD_ERROR = sd(temp))
  }))

  ##combine
  data_full <- as.data.frame(cbind(data.table::rbindlist(signal_table_list),
                                   APPARENT_DOSE[rep(1:.N, each = 2), 2:3]))

  # Plotting ------------------------------------------------------------------------------------
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

    ## set colours
    col_pal <- grDevices::hcl.colors(100, palette = "RdYlGn", rev = TRUE)

    ##settings
    plot_settings <- list(
      main = "Sample Carousel Crosstalk",
      mtext = ""
    )

      ##modify on request
      plot_settings <- modifyList(x = plot_settings, list(...))


    ##pre-calculations for graphical parameters
    n.positions <- length(unique(APPARENT_DOSE$POSITION))
    arc.step <- (2 * pi) / n.positions
    step <- 0

  ## calculate mean and standard deviation for similar positions
  AD <- POSITION <- NULL  # silence notes raised by R CMD check
  AD_matrix <- APPARENT_DOSE[, list(AD = mean(AD), AD_ERROR = sd(AD)),
                             by = POSITION]

    ##create colour ramp
    col.seq <- data.frame(
      POSITION = AD_matrix[order(AD_matrix[,2]),1],
      COLOUR = col_pal[seq(1,100, length.out = nrow(AD_matrix))],
      stringsAsFactors = FALSE)

    col.seq <- col.seq[["COLOUR"]][order(col.seq[["POSITION"]])]

    ##calculate model
    fit <- stats::lm(
      formula = y ~ poly(x, 2, raw=TRUE),
      data = data.frame(y = APPARENT_DOSE$AD[order(APPARENT_DOSE$POSITION)], x = sort(APPARENT_DOSE$POSITION)))

    ##enable or disable plot ... we cannot put the condition higher, because we here
    ##calculate something we are going to need later
    if (plot) {

      ##set layout matrix
      graphics::layout(mat = matrix(
        c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 3),
        5,
        5,
        byrow = TRUE
      ))

      ##create empty plot
      par(
        mar = c(1, 1, 1, 1),
        omi = c(1, 1, 1, 1),
        oma = c(0.2, 0.2, 0.2, 0.2),
        cex = 1.1
      )
      shape::emptyplot(c(-1.15, 1.15), main = plot_settings$main, frame.plot = FALSE)

      ##add outher circle
      shape::plotcircle(r = 1.1, col = rgb(0.9, 0.9, 0.9, 1))

      ##add inner octagon
      shape::filledcircle(
        r1 = 0.6,
        mid = c(0, 0),
        lwd = 1,
        lcol = "black",
        col = "white"
      )

      ##add circles
      for (i in 1:n.positions) {
        shape::plotcircle(
          r = 0.05,
          mid = c(cos(step), sin(step)),
          cex = 6,
          pch = 20,
          col = col.seq[i]
        )
        text(x = cos(step) * 0.85,
             y = sin(step) * 0.85,
             labels = i)
        step <- step + arc.step
      }

      ##add center plot with position
      plot(NA, NA,
           xlim = range(AD_matrix[,1]),
           ylim = range(APPARENT_DOSE[,2]),
           frame.plot = FALSE,
           type = "l")

        ##add points
        points(x = APPARENT_DOSE, pch = 20, col = rgb(0,0,0,0.3))

        ##add linear model
        lines(sort(APPARENT_DOSE$POSITION), predict(fit), col = "red")

      ##add colour legend
      shape::emptyplot(c(-1.2, 1.2), frame.plot = FALSE)
      graphics::rect(
        xleft = rep(-0.6, 100),
        ybottom = seq(-1.2,1.1,length.out = 100),
        xright = rep(0, 100),
        ytop = seq(-1.1,1.2,length.out = 100),
        col = col_pal,
        lwd = 0,
        border = FALSE
      )

      ##add scale text
      text(
        x = -0.3,
        y = 1.2,
        label = "[s]",
        pos = 3,
        cex = 1.1
      )
      text(
        x = 0.4,
        y = 1,
        label = round(max(AD_matrix[, 2]),2),
        pos = 3,
        cex = 1.1
      )
      text(
        x = 0.4,
        y = -1.5,
        label = 0,
        pos = 3,
        cex = 1.1
      )
    }

  # Output --------------------------------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      data = as.data.frame(AD_matrix),
      data_full = data_full,
      fit = fit,
      col.seq = col.seq
      ),
    info = list(
      call = sys.call()
    )
  )
}

#' Plot function for an RLum.Results S4 class object
#'
#' The function provides a standardised plot output for data of an RLum.Results
#' S4 class object
#'
#' The function produces a multiple plot output.  A file output is recommended
#' (e.g., [pdf]).
#'
#' @param object [Luminescence::RLum.Results-class] (**required**):
#' S4 object of class `RLum.Results`
#'
#' @param single [logical] (*with default*):
#' single plot output (`TRUE/FALSE`) to allow for plotting the results in as
#' few plot windows as possible.
#'
#' @param ... further arguments and graphical parameters will be passed to
#' the `plot` function.
#'
#' @return Returns multiple plots.
#'
#' @note
#' Not all arguments available for [plot] will be passed!
#' Only plotting of `RLum.Results` objects are supported.
#'
#' @section Function version: 0.2.1
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [plot], [Luminescence::plot_RLum]
#'
#' @keywords aplot
#'
#' @examples
#'
#'
#' ###load data
#' data(ExampleData.DeValues, envir = environment())
#'
#' # apply the un-logged minimum age model
#' mam <- calc_MinDose(data = ExampleData.DeValues$CA1, sigmab = 0.2, log = TRUE, plot = FALSE)
#'
#' ##plot
#' plot_RLum.Results(mam)
#'
#' # estimate the number of grains on an aliquot
#' grains<- calc_AliquotSize(grain.size = c(100,150), sample.diameter = 1, plot = FALSE, MC.iter = 100)
#'
#' ##plot
#' plot_RLum.Results(grains)
#'
#'
#' @export
plot_RLum.Results<- function(
  object,
  single = TRUE,
  ...
) {
  .set_function_name("plot_RLum.Results")
  on.exit(.unset_function_name(), add = TRUE)

  ##============================================================================##
  ## CONSISTENCY CHECK OF INPUT DATA
  ##============================================================================##

  .validate_class(object, "RLum.Results")
  if (is.na(object@originator)) {
    .validate_originator(object, c("analyse_SAR.CWOSL",
                                   "analyse_IRSAR.RF",
                                   "analyse_pIRIRSequence",
                                   "calc_AliquotSize",
                                   "calc_CentralDose",
                                   "calc_FastRatio",
                                   "calc_FiniteMixture",
                                   "calc_FuchsLang2001",
                                   "calc_MinDose",
                                   "calc_MaxDose",
                                   "calc_SourceDoseRate"))
  }

  ##============================================================================##
  ## SAFE AND RESTORE PLOT PARAMETERS ON EXIT
  ##============================================================================##
  par.default <- .par_defaults()
  on.exit(suppressWarnings(par(par.default)), add = TRUE)

  ##============================================================================##
  ## ... ARGUMENTS
  ##============================================================================##

  ##deal with addition arguments
  extraArgs <- list(...)

  main <- extraArgs$main %||% ""
  mtext <- extraArgs$mtext %||% ""
  log <- extraArgs$log %||% ""
  lwd <- extraArgs$lwd %||% 1
  lty <- extraArgs$lty %||% 1
  type <- extraArgs$type %||% "l"
  pch <- extraArgs$pch %||% 1
  col <- extraArgs$col %||% "black"

  ##============================================================================##
  ## PLOTTING
  ##============================================================================##

  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
  ## CASE 0: General plot dispatcher ----------
  switch(object@originator,
      analyse_SAR.CWOSL = plot_AbanicoPlot(object),
      analyse_pIRIRSequence = plot_AbanicoPlot(object),
      analyse_IRSAR.RF = plot_AbanicoPlot(object),
      calc_FiniteMixture = do.call(calc_FiniteMixture, c(object, extraArgs)),
    NULL
    )

  ## CASE 1: Minimum Age Model / Maximum Age Model -------
  if (.check_originator(object, c("calc_MinDose", "calc_MaxDose"))) {
    ## single MAM estimate
    # plot profile log likelihood

    profiles <- object@data$profile
    if (object@data$args$log) {
      profiles@profile$gamma$par.vals[ ,"gamma"] <- exp(profiles@profile$gamma$par.vals[ ,"gamma"])
      profiles@profile$sigma$par.vals[ ,"sigma"] <- exp(profiles@profile$sigma$par.vals[ ,"sigma"])

      if (object@data$args$par == 4)
        profiles@profile$mu$par.vals[ ,"mu"] <- exp(profiles@profile$mu$par.vals[ ,"mu"])
    }

    if (single)
      par(mfrow=c(2, 2))

    param <- c("gamma", "sigma", "p0", "mu")

    for (i in param) {
      if (object@data$summary$par == 3 && i == "mu")
        break

      tryCatch({

        xvals <- as.data.frame(profiles@profile[[i]]$par.vals)[[i]]
        xlim <- range(xvals[xvals > 0])
        suppressWarnings(
          bbmle::plot(profiles, which = i, xlab = "", xaxt = "n", xlim = xlim)
        )

        axis(1, mgp = c(3, 0.5, 0))
        title(xlab = i, line = 1.2)

        if (i %in% c("gamma", "sigma", "mu") && object@data$args$log && object@data$args$log.output) {
          axis(1, at = axTicks(1),
               labels = format(round(log(axTicks(1)), 2), nsmall = 2),
               line = 2.5, mgp = c(3, 0.5, 0))
          title(xlab = paste0("log(", i, ")"), line = 4)
        }

      }, error = function(e)  {
        message("Unable to plot the likelihood profile for: ", i,
                " (likelihood probably infinite)")
      })
    }
    par(mfrow=c(1,1))

    # })

    ## bootstrap MAM estimates
    if (object@data$args$bootstrap) {

      # save previous plot parameter and set new ones
      .pardefault <- .par_defaults()

      # get De-llik pairs
      pairs<- object@data$bootstrap$pairs$gamma

      # get polynomial fit objects
      poly.fits <- list(three = object@data$bootstrap$poly.fits$poly.three,
                        four = object@data$bootstrap$poly.fits$poly.four,
                        five = object@data$bootstrap$poly.fits$poly.five,
                        six = object@data$bootstrap$poly.fits$poly.six)

      # define polynomial curve functions for plotting
      ## these are equivalent to
      ## coefficient[1] + coefficient[2]*x + coefficient[3]*x^2 +  ...
      poly.curves <- list(
          three = function(x) colSums(poly.fits$three$coefficient * t(outer(x, 0:3, "^"))),
          four = function(x) colSums(poly.fits$four$coefficient * t(outer(x, 0:4, "^"))),
          five = function(x) colSums(poly.fits$five$coefficient * t(outer(x, 0:5, "^"))),
          six = function(x) colSums(poly.fits$six$coefficient * t(outer(x, 0:6, "^"))))

      ## --------- PLOT "RECYCLE" BOOTSTRAP RESULTS ------------ ##

      if (single) {
        graphics::layout(cbind(c(1, 1, 2, 5, 5, 6), c(3, 3, 4, 7, 7, 8)))
        par(cex = 0.6)
      } else {
        graphics::layout(matrix(c(1, 1, 2)), 2, 1)
        par(cex = 0.8)
      }

      for (i in 1:length(poly.fits)) {
        ## ----- LIKELIHOODS

        # set margins (bottom, left, top, right)
        par(mar=c(0,5,5,3))

        # sort De and likelihoods by De (increasing)
        pairs<- pairs[order(pairs[,1]),]

        # remove invalid NA values
        pairs <- stats::na.omit(pairs)

        plot(x=pairs[,1],
             y=pairs[,2],
             xlab="Equivalent Dose [Gy]",
             ylab="Likelihood",
             xlim=range(pretty(pairs[,1])),
             ylim=range(pretty(c(0, as.double(quantile(pairs[,2],probs=0.98))))),
             xaxt = "n",
             xaxs = "i",
             yaxs = "i",
             bty = "l",
             main="Recycled bootstrap MAM-3")

        axis(side = 1, labels = FALSE, tick = FALSE)

        # add subtitle
        mtext(as.expression(bquote(italic(M) == .(object@data$args$bs.M) ~ "|" ~
                                     italic(N) == .(object@data$args$bs.N) ~ "|" ~
                                     italic(sigma[b])  == .(object@data$args$sigmab) ~
                                     "\u00B1" ~ .(object@data$args$sigmab.sd) ~ "|" ~
                                     italic(h) == .(round(object@data$args$bs.h,1))
        )
        ),
        side = 3, line = 0.3, adj = 0.5,
        cex = if(single){0.5}else{0.8})

        # add points
        points(x=pairs[,1], y=pairs[,2], pch=1, col = "grey80")

        # get polynomial function
        poly.curve<- poly.curves[[i]]

        # add curve to plot
        curve(poly.curve, from = min(pairs[,1]), to = (max(pairs[,1])),
              col = "black", add = TRUE, type = "l")

        # add legend
        legend<- c("Third degree", "Fourth degree", "Fifth degree", "Sixth degree")
        legend("topright",  xjust = 0,
               legend = legend[i],
               y.intersp = 1.2,
               bty = "n",
               title = "Polynomial Fit",
               lty = 1,
               lwd= 1.5)

        ## ----- RESIDUALS

        # set margins (bottom, left, top, right)
        par(mar=c(5,5,0,3))

        plot(x = pairs[,1],
             y = residuals(poly.fits[[i]]),
             ylim = c(min(residuals(poly.fits[[i]])) * 1.2,
                      as.double(quantile(residuals(poly.fits[[i]]), probs = 0.99))),
             xlim=range(pretty(pairs[,1])),
             xaxt = "n",
             bty = "l",
             xaxs = "i",
             col = "grey80",
             ylab = "Fit residual",
             xlab = "Equivalent dose [Gy]")

        axis(side = 1, labels = TRUE, tick = TRUE)

        # add horizontal line
        abline(h = 0, lty=2)

        # calculate residual sum of squares (RSS) and add to plot
        rss<- sum(residuals(poly.fits[[i]])^2)
        mtext(text = paste("RSS =",round(rss,3)), adj = 1,
              side = 3, line = -2,
              cex = if(single){0.6}else{0.8})

        ## ----- PROPORTIONS

      }##EndOf::Plot_loop

      # restore previous plot parameters
      par(.pardefault)

      ### TODO: plotting of the LOESS fit needs to be fleshed out
      ### possibly integrate this in the prior polynomial plot loop

      ### LOESS PLOT
      if (!anyNA(object@data$bootstrap$loess.fit)) {
      pairs<- object@data$bootstrap$pairs$gamma
      pred<- predict(object@data$bootstrap$loess.fit)
      loess<- cbind(pairs[,1], pred)
      loess<- loess[order(loess[,1]),]

      # plot gamma-llik pairs
      plot(pairs,
           ylim = c(0, as.double(quantile( pairs[,2],probs=0.99))),
           ylab = "Likelihood",
           xlab = "Equivalent dose [Gy]",
           col = "gray80")

      # add LOESS line
      lines(loess, type = "l", col = "black")
      }

      ### ------ PLOT BOOTSTRAP LIKELIHOOD FIT

      par(mar=c(5,4,4,4))

      xlim<- range(pretty(object@data$data[,1]))
      xlim[1]<- xlim[1]-object@data$data[which.min(object@data$data[,1]),2]
      xlim[2]<- xlim[2]+object@data$data[which.max(object@data$data[,1]),2]
      xlim<- range(pretty(xlim))

      # empty plot
      plot(NA,NA,
           xlim=xlim,
           ylim=c(0,2),
           xlab="Equivalent dose [Gy]",
           ylab="",
           bty="l",
           axes=FALSE,
           xaxs="i",
           yaxs="i",
           yaxt="n")

      axis(side = 1)
      axis(side = 2, at = c(0,0.5,1))

      mtext(text = "Normalised likelihood / density", side = 2, line = 2.5, adj = 0)

      # set the polynomial to plot
      poly.curve<- poly.curves[[1]] # three degree poly

      # plot a nice grey polygon as in the publication
      step<- 0.1
      x<- seq(min(pairs[,1]), max(pairs[,1]), step)
      y<- poly.curve(x)
      # normalise y-values
      y<- y/max(y)

      x<- c(min(pairs[,1]), x, max(pairs[,1]))
      y<- c(0, y, 0)

      # cutoff negative y values
      y.idx <- !is.na(y) & y >= 0
      x <- x[y.idx]
      y <- y[y.idx]

      # add bootstrap likelihood polygon to plot
      polygon(x, y, col = "grey80", border = NA)

      if (all(x > max(xlim)) || all(x < min(xlim)))
        .throw_warning("Bootstrap estimates out of x-axis range")


      ### ----- PLOT MAM SINGLE ESTIMATE

      # symmetric errors, might not be appropriate
      mean<- object@data$summary$de
      sd<- object@data$summary$de_err


      if (anyNA(c(mean, sd))) {
        ## no longer reachable since #1353
        # nocov start
        .throw_warning("Unable to plot the MAM single estimate (NA value)")
        # nocov end
      } else {

        x<- seq(mean-5*sd, mean+5*sd, 0.001)
        y<- dnorm(seq(mean-5*sd, mean+5*sd, 0.001), mean, sd)
        # normalise y-values
        y<- y/max(y)

        points(x, y,
               type="l",
               col="red")

        ## asymmetric errors
        x<- unlist(object@data$profile@profile$gamma$par.vals[,1])
        y<- abs(unlist(object@data$profile@profile$gamma$z))

        if (object@data$args$log) {
          x<- exp(x)
        }

        # now invert the data by shifting
        y<- -y
        y<- y-min(y)
        y<- y/max(y)

        # fit a smoothing spline
        l<- stats::spline(x = x, y = y, method = "n", n = 1000)

        # make the endpoints zero
        l$y[1]<- l$y[length(l$y)]<- 0

        # add profile log likelihood curve to plot
        lines(l, col="blue", lwd=1)

        # add vertical lines of the mean values
        #points(x = 80, y = 100,type = "l")
      }

      #### ------ PLOT DE
      par(new = TRUE)

      # sort the data in ascending order
      dat<- object@data$data[order(object@data$data[,1]),]

      x<- dat[,1]
      y<- 1:length(object@data$data[,1])

      plot(x = x, y = y,
           xlim=xlim,
           ylim=c(0, max(y)+1),
           axes = FALSE,
           pch = 16,
           xlab = "",
           ylab="",
           xaxs="i",
           yaxs="i")

      axis(side = 4)
      mtext(text = "# Grain / aliquot", side = 4, line = 2.5)

      # get sorted errors
      err<- object@data$data[order(object@data$data[,1]),2]

      # fancy error bars
      graphics::arrows(x0 = x-err, y0 = y,
             x1 =  x+err, y1 = y,
             code = 3, angle = 90, length = 0.05)

      ### ---- AUXILLARY

      # add legend
      legend("bottomright",
             bty = "n",
             col = c("grey80", "red", "blue", "black"),
             pch = c(NA,NA,NA,16),
             lty = c(1,1,1,1),
             lwd=c(10,2,2,2),
             legend = c("Bootstrap likelihood", "Profile likelihood (gaussian fit)",
                        "Profile likelihood", "Grain / aliquot")
      )
    }##EndOf::Bootstrap_plotting
  }#EndOf::CASE1_MinimumAgeModel-3


  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
  ## CASE 2: Central Age Model ---------
  if (.check_originator(object, "calc_CentralDose")) {

    # get profile log likelihood data
    sig<- object@data$profile$sig*100
    llik<- object@data$profile$llik

    # save previous plot parameter and set new ones
    par.default <- .par_defaults()
    par(mar = c(4, 4, 3, 1) + 0.1, las = 1, cex.axis = 1.2, cex.lab = 1.2)

    ## don't display log-likelihoods more negative than -100
    y.min <- max(-100, min(llik))

    ## plot the profile log likelihood
    plot(sig, llik,
         ylim = c(y.min, 0),
         xlab = as.expression(bquote(sigma[OD] ~ "[%]")),
         ylab = "Log likelihood",
         type = "l",
         lwd = 1.5)
    abline(h=0,lty=3)
    abline(h=-1.92,lty=3)
    title(as.expression(bquote("Profile log likelihood for" ~ sigma[OD])))

    # find upper and lower confidence limits for sigma
    sigmax<- sig[which.max(llik)]
    tf<- abs(llik+1.92) < 0.05
    sig95<- sig[tf]
    ntf<- length(sig95)
    sigL <- ifelse(sig95[1] < sigmax, sig95[1], NA)
    sigU <- ifelse(sig95[ntf] > sigmax, sig95[ntf], NA)

    # put them on the graph
    res <- c(sigL, sigmax, sigU)
    abline(v = res)
    text(res, y.min, labels = round(res, 2), adj = 0, pos = 4, offset = 0.3)

    # restore previous plot parameters
    par(par.default)
  }##EndOf::Case 2 - calc_CentralDose()


  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
  ## CASE 3: Fuchs & Lang 2001 --------
  if (.check_originator(object, "calc_FuchsLang2001")) {

    ##deal with addition arguments
    extraArgs <- list(...)

    main <- extraArgs$main %||% "Fuchs & Lang (2001)"
    xlab <- extraArgs$xlab %||% expression(paste(D[e]," [s]"))
    ylab <- extraArgs$ylab %||% "# Aliquots"
    sub <-  extraArgs$sub %||% ""
    cex <- extraArgs$cex %||% 1
    lwd <- extraArgs$lwd %||% 1
    pch <- extraArgs$pch %||% 19
    ylim <- extraArgs$ylim %||% c(1, length(object@data$data[, 1]) + 3)
    xlim <- extraArgs$xlim %||% c(min(object@data$data[, 1]) - max(object@data$data[, 2]),
                                  max(object@data$data[, 1]) + max(object@data$data[, 2]))
    mtext <- extraArgs$mtext %||% "unknown sample"

    # extract relevant plotting parameters
    o<- order(object@data$data[[1]])
    data_ordered<- object@data$data[o,]
    usedDeValues<- object@data$usedDeValues
    n.usedDeValues<- object@data$summary$n.usedDeValues

    par(cex = cex, mfrow=c(1,1))

    ##PLOT
    counter<-seq(1,max(o))

    plot(NA,NA,
         ylim = ylim,
         xlim = xlim,
         xlab = xlab,
         ylab = ylab,
         main = main,
         sub = sub)

    ##SEGMENTS
    segments(data_ordered[,1]-data_ordered[,2],1:length(data_ordered[,1]),
             data_ordered[,1]+data_ordered[,2],1:length(data_ordered[,1]),
             col="gray")

    ##POINTS
    points(data_ordered[,1], counter,pch=pch)

    ##LINES
    ##BOUNDARY INFORMATION
    ##lower boundary
    lines(c(
      usedDeValues[length(usedDeValues[,1])-n.usedDeValues+1,1], #boundary_counter for incorporate skipped values
      usedDeValues[length(usedDeValues[,1])-n.usedDeValues+1,1]),
      c(min(o)-0.5,max(o)+0.5),
      col="red",
      lty="dashed", lwd = lwd)

    #upper boundary
    lines(c(max(usedDeValues[,1]),max(usedDeValues[,1])),c(min(o)-0.5,max(o)+0.5),
          col="red",lty="dashed", lwd = lwd)

    #plot some further informations into the grafik
    graphics::arrows(
      usedDeValues[length(usedDeValues[,1])-n.usedDeValues+1,1]+usedDeValues[length(usedDeValues[,1])-n.usedDeValues+1,1]*0.02, #x1
      max(o)+0.5, #y1
      max(usedDeValues[,1]-usedDeValues[,1]*0.02), #x2
      max(o)+0.5, #y2,
      code=3,
      length=0.03)

    text(
      c(
        usedDeValues[length(usedDeValues[,1])-n.usedDeValues+1,1],
        usedDeValues[length(usedDeValues[,1])-n.usedDeValues+1,1]),
      c(max(o)+2,max(o)+2),
      labels=paste("used values = ", n.usedDeValues),
      cex=0.6*cex,
      adj=0)

    ##MTEXT
    mtext(side=3,mtext,cex=cex)
  }

  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
  ## CASE 5: Aliquot Size ---------
  if (.check_originator(object, "calc_AliquotSize") &&
      !is.null(object@data$MC$estimates)) {
    ## FIXME(mcol): The following code has been ported to calc_AliquotSize(),
    ## where it has been further updated. This version is preserved to support
    ## the plotting of an RLum.Results object directly, which calc_AliquotSize()
    ## doesn't yet support.
      extraArgs <- list(...)

      main <- extraArgs$main %||% "Monte Carlo Simulation"
      xlab <- extraArgs$xlab %||% "Amount of grains on aliquot"

      # extract relevant data
      MC.n<- object@data$MC$estimates
      MC.n.kde<- object@data$MC$kde
      MC.stats<- object@data$MC$statistics
      MC.q<- object@data$MC$quantile
      MC.iter<- object@data$args$MC.iter

      # set layout of plotting device
      graphics::layout(matrix(c(1, 1, 2)), 2, 1)
      par(cex = 0.8)

      ## plot MC estimate distribution

      # set margins (bottom, left, top, right)
      par(mar=c(2,5,5,3))

      # plot histogram
      hist(MC.n, freq=FALSE, col = "gray90",
           main="", xlab=xlab,
           xlim = c(min(MC.n)*0.95, max(MC.n)*1.05),
           ylim = c(0, max(MC.n.kde$y)*1.1))

      # add rugs to histogram
      graphics::rug(MC.n)

      # add KDE curve
      lines(MC.n.kde, col = "black", lwd = 1)

      # add mean, median and quantils (0.05,0.95)
      abline(v=c(MC.stats$mean, MC.stats$median, MC.q),
             lty=c(2, 4, 3,3), lwd = 1)

      # add main- and subtitle
      mtext(main, side = 3, adj = 0.5,
            line = 3, cex = 1)
      mtext(as.expression(bquote(italic(n) == .(MC.iter) ~ "|" ~
                                   italic(hat(mu)) == .(round(MC.stats$mean)) ~ "|" ~
                                   italic(hat(sigma))  == .(round(MC.stats$sd.abs)) ~ "|" ~
                                   italic(frac(hat(sigma),sqrt(n))) == .(round(MC.stats$se.abs))  ~ "|" ~
                                   italic(v) == .(round(MC.stats$skewness, 2))
      )
      ),
      side = 3, line = 0.3, adj = 0.5,
      cex = 0.9)

      # add legend
      legend("topright", legend = c("mean","median", "0.05 / 0.95 quantile"),
             lty = c(2, 4, 3), bg = "white", box.col = "white", cex = 0.9)

      ## BOXPLOT
      # set margins (bottom, left, top, right)
      par(mar=c(5,5,0,3))

      plot(NA, type="n", xlim=c(min(MC.n)*0.95, max(MC.n)*1.05),
           xlab=xlab,  ylim=c(0.5,1.5),
           xaxt="n", yaxt="n", ylab="")
      par(bty="n")
      graphics::boxplot(MC.n, horizontal = TRUE, add = TRUE, bty = "n", cex = 1)
  }#EndOf::Case 5 - calc_AliquotSize()

  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
  ## CASE 6: calc_SourceDoseRate() ----------
  if (.check_originator(object, "calc_SourceDoseRate")) {
    ##prepare data
    ##get data
    df <- get_RLum(object = object, data.object = "dose.rate")

    ##reduce the size for plotting, more than 100 points make no sense
    if(nrow(df)>100)
      df <- df[seq(1,nrow(df), length = 100),]

    ##plot settings
    plot.settings <- list(
      main = "Source Dose Rate Prediction",
      xlab = "Date",
      ylab = paste0(
        "Dose rate/(",get_RLum(object = object, data.object = "parameters")$dose.rate.unit,")"),
      log = "",
      cex = 1,
      xlim = NULL,
      ylim = c(min(df[,1]) - max(df[,2]), max(df[,1]) + max(df[,2])),
      pch = 1,
      mtext = paste0(
        "source type: ", get_RLum(object = object, data.object = "parameters")$source.type,
        " | ",
        "half-life: ", get_RLum(object = object, data.object = "parameters")$halflife,
        " a"
      ),
      grid = expression(nx = 10, ny = 10),
      col = 1,
      type = "b",
      lty = 1,
      lwd = 1,
      segments = ""
    )

    ##modify list if something was set
    plot.settings <- modifyList(plot.settings, list(...))

    ##plot
    plot(
      df[,3], df[,1],
      main = plot.settings$main,
      xlab = plot.settings$xlab,
      ylab = plot.settings$ylab,
      xlim = plot.settings$xlim,
      ylim = plot.settings$ylim,
      log = plot.settings$log,
      pch = plot.settings$pch,
      col = plot.settings$pch,
      type = plot.settings$type,
      lty = plot.settings$lty,
      lwd = plot.settings$lwd
    )

    if(!is.null(plot.settings$segments)){
      segments(
        x0 = df[,3], y0 = df[,1] + df[,2],
        x1 = df[,3], y1 = df[,1] - df[,2]
      )
    }

    mtext(side = 3, plot.settings$mtext)

    if(!is.null(plot.settings$grid)){
      graphics::grid(eval(plot.settings$grid))
    }

  }#EndOf::Case 6 - calc_SourceDoseRate()

  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
  ## CASE 7: Fast Ratio ----------
  if (.check_originator(object, "calc_FastRatio")) {
    # graphical settings
    settings <- list(main = "Fast Ratio",
                     xlab = "t/s",
                     ylab = "Signal/cts",
                     type = "b",
                     log = "",
                     pch = 16,
                     cex = 1.0,
                     col = "black")
    settings <- modifyList(settings, list(...))

    par(cex = settings$cex)

    # fetch data from RLum.Results object
    curve <- get_RLum(object, "data")
    if (inherits(curve, "RLum.Data.Curve"))
      curve <- get_RLum(curve)
    res <- get_RLum(object, "summary")
    fit <- get_RLum(object, "fit")

    # calculate the dead channel time offset
    offset <- res$dead.channels.start * res$channel.width

    # plot the OSL curve
    plot(curve, type = "n", main = settings$main,
         xlab = settings$xlab, ylab = settings$ylab, log = settings$log)

    # plot points to show measured data points (i.e., the channels)
    if (settings$type == "p" || settings$type == "b")
      points(curve[(res$dead.channels.start + 1):(nrow(curve) - res$dead.channels.end), ],
             pch = settings$pch, col = settings$col)

    # plot dead channels as empty circles
    if (res$dead.channels.start > 0)
      points(curve[1:res$dead.channels.start,])
    if (res$dead.channels.end > 0)
      points(curve[(nrow(curve) - res$dead.channels.end):nrow(curve), ])

    if (settings$type == "l" || settings$type == "b")
      lines(curve, col = settings$col)

    # optional: plot fitted CW curve
    if (!is.null(fit)) {
      nls.fit <- get_RLum(fit, "fit")
      if (!inherits(fit, "try-error") && isTRUE(object@data$args$fitCW.curve)) {
          lines(curve[(res$dead.channels.start + 1):(nrow(curve) - res$dead.channels.end), 1],
                predict(nls.fit), col = "red", lty = 1)

          ##plot curve for additional parameters
          col_components <- c("red", "green", "blue")
          for (i in 1:3) {
            if (!is.na(fit@data$data[[paste0("I0", i)]]))
              curve(fit@data$data[[paste0("I0", i)]] * fit@data$data[[paste0("lambda", i)]] * exp(-fit@data$data[[paste0("lambda", i)]] * x),
                    lwd = 1, lty = 4, add = TRUE, col = col_components[i])
          }
      }
    }

    # add vertical lines and labels for L1, L2, L3
    L_times <- c(curve[res$Ch_L1, 1],
                 curve[res$Ch_L2, 1],
                 curve[res$Ch_L3_start, 1],
                 curve[res$Ch_L3_end, 1]) + offset
    abline(v = L_times,
           lty = 2)
    text(L_times, max(curve[ ,2]) * 0.95, pos = c(4,4,2,2),
         labels = expression('L'[1], 'L'[2], 'L'[3['start']], 'L'[3['end']]))

  }#EndOf::Case7 - calc_FastRatio()
}

#'@title Trim Channels of RLum.Data-class Objects
#'
#'@description Trim off the number of channels of [Luminescence::RLum.Data-class] objects of similar record type
#' on the time domain. This function is useful in cases where objects have different lengths (short/longer
#'measurement time) but should be analysed jointly by other functions.
#'
#'@details
#'The function has two modes of operation:
#'
#' 1. Single [Luminescence::RLum.Data-class] objects or a [list] of such objects:
#' the function is applied separately over each object.
#'
#' 2. Multiple curves via [Luminescence::RLum.Analysis-class] or a [list] of such objects:
#' in this mode, the function first determines the minimum number of channels
#' for each category of records and then jointly processes them. For instance,
#' if the object contains one TL curve with 100 channels and two OSL curves
#' with 100 and 99 channels, respectively, then the minimum would be set to
#' 100 channels for the TL curve and to 99 for the OSL curves. If no further
#' parameters are applied, the function will shorten all OSL curves to 99
#' channels, but leave the TL curve untouched.
#'
#'@param object [Luminescence::RLum.Data-class] [Luminescence::RLum.Analysis-class] (**required**): input object,
#'can be a [list] of objects. Please note that in the latter case the function works
#'only isolated on each element of the [list].
#'
#'@param recordType [character] (*optional*): type of the record where the trim
#'should be applied. If not set, the types are determined automatically and applied
#'for each record type classes. Can be provided as [list].
#'
#' @param trim_range [numeric] (*optional*): sets the range of indices to
#' keep. If only one value is given, this is taken to be the minimum; if two
#' values are given, then the range is defined between the two values
#' (inclusive). Any value beyond the second is silently ignored. If nothing
#' is set (default), then all curves are trimmed to the same maximum length.
#'
#'@returns A trimmed object or [list] of such objects similar to the input objects
#'
#'@section Function version: 0.2
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@seealso [Luminescence::RLum.Data-class], [Luminescence::RLum.Analysis-class]
#'
#'@keywords manip
#'
#'@examples
#'## trim all TL curves in the object to channels 10 to 20
#'data(ExampleData.BINfileData, envir = environment())
#'temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)
#'
#'c <- trim_RLum.Data(
#' object = temp,
#' recordType = "TL",
#' trim_range = c(10,20))
#'
#'plot_RLum.Analysis(
#'object = c,
#'combine = TRUE,
#'subset = list(recordType = "TL"))
#'
#'## simulate a situation where one OSL curve
#'## in the dataset has only 999 channels instead of 1000
#'## all curves should be limited to 999
#'temp@records[[2]]@data <- temp@records[[2]]@data[-nrow(temp[[2]]@data),]
#'
#'c <- trim_RLum.Data(object = temp)
#'nrow(c@records[[4]]@data)
#'
#'
#'@export
trim_RLum.Data <- function(
    object,
    recordType = NULL,
    trim_range = NULL
) {
  .set_function_name("trim_RLum.Data")
  on.exit(.unset_function_name(), add = TRUE)

# Self-call ---------------------------------------------------------------
  if(inherits(object, "list")) {
    ## expand parameters
    parm <- .expand_parameters(length(object))

    l <- lapply(seq_along(object), function(x){
      .validate_class(object[[x]], c("RLum.Data", "RLum.Analysis"),
                      name = "All elements of 'object'")
      trim_RLum.Data(
        object = object[[x]],
        recordType = parm$recordType[[x]],
        trim_range = parm$trim_range[[x]])
    })

    return(l)
  }

# Work horse functions ----------------------------------------------------
  ## RLum.Data.Curve
  .trim_RLum.Data.Curve <- function(object, type, range){
    ## only if type is matched
    if  (any(object@recordType[1] %in% type)) {
      range <- pmin(nrow(object@data), range)
      object@data <- object@data[range[1]:range[2], , drop = FALSE]
    }
    object
  }

  ## RLum.Data.Spectrum
  .trim_RLum.Data.Spectrum <- function(object, type, range){
    ## only if type is matched
    if (any(object@recordType[1] %in% type)) {
      range <- pmin(ncol(object@data), range)
      object@data <- object@data[, range[1]:range[2], drop = FALSE]
    }
    object
  }

  ## RLum.Data.Image
  .trim_RLum.Data.Image <- function(object, type, range){
    ## only if type is matched and data is not empty
    if (any(object@recordType[1] %in% type && length(dim(object@data)) == 3)) {
      range <- pmin(range, dim(object@data)[3])
      object@data <- object@data[, , range[1]:range[2], drop = FALSE]
    }
    object
  }

  ## RLum.Analysis (which calls the functions above)
  .trim_RLum.Analysis <- function(object, type, range) {
    ## determine lengths of objects
    ln <- unlist(lapply(object@records, function(x) {
      ln <- switch(
        class(x)[1],
        RLum.Data.Curve = dim(x@data)[1],
        RLum.Data.Spectrum = dim(x@data)[2],
        RLum.Data.Image = dim(x@data)[3]
      )
      names(ln) <- x@recordType
      ln
    }))

    ## run over single objects
    object@records <- lapply(object@records, function(x){
      ## determine max and min of the particular record Types compared to the global information
      tmp_max <- min(c(ln[names(ln) == x@recordType], range[2]))
      tmp_min <- range[1] ## cannot be smaller than 1

      ## call sub-function to process and return
      .trim_object(x, recordType, c(tmp_min, tmp_max))
    })

    return(object)
  }

  ## function dispatcher
  .trim_object <- function(obj, type, range) {
    trim.fun <- switch(class(obj)[1],
                       RLum.Analysis      = .trim_RLum.Analysis,
                       RLum.Data.Curve    = .trim_RLum.Data.Curve,
                       RLum.Data.Image    = .trim_RLum.Data.Image,
                       RLum.Data.Spectrum = .trim_RLum.Data.Spectrum)
    trim.fun(obj, type, range)
  }


  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Data", "RLum.Analysis"))

  ## determine classes for record Types
  if(is.null(recordType)) {
    recordType <- switch(
      class(object)[1],
      RLum.Analysis = unique(vapply(object@records, function(x) x@recordType, character(1))),
      object@recordType
    )
  }
  .validate_class(recordType, "character")

  ## set trim_range if not provided
  if (is.null(trim_range))
    trim_range <- c(1,Inf)
  .validate_class(trim_range, c("integer", "numeric"))

  ## silently sanitize `trim_range` to ensure that it has length 2,
  ## it contains no elements smaller than 1, and is sorted
  if (length(trim_range) == 1)
    trim_range <- c(1, trim_range)
  trim_range <- sort(pmax(abs(trim_range[1:2]), 1))

  ## Dispatch and return ----------------------------------------------------
  .trim_object(object, type = recordType, range = trim_range)
}

#' @title Import Risø BIN/BINX-files into R
#'
#' @description Import a `*.bin` or a `*.binx` file produced by a Risø DA15 and DA20 TL/OSL
#' reader into R.
#'
#' @details
#' The binary data file is parsed byte by byte following the data structure
#' published in the Appendices of the Analyst manual p. 42.
#'
#' For the general BIN/BINX-file structure, the reader is referred to the
#' Risø website: [https://www.fysik.dtu.dk]()
#'
#' @param file [character] or [list] (**required**):
#' name of one or multiple BIN/BINX files (URLs are supported); it can be the
#' path to a directory, in which case the function tries to detect and import
#' all BIN/BINX files found recursively from the given directory.
#'
#' @param show.raw.values [logical] (*with default*):
#' shows raw values from BIN-file for `LTYPE`, `DTYPE` and `LIGHTSOURCE` without
#' translation in characters. Can be provided as `list` if `file` is a `list`.
#'
#' @param n.records [numeric] (*optional*): limits the number of imported records
#' to the provided record id (e.g., `n.records = 1:10` imports the first ten records,
#' while `n.records = 3` imports only record number 3. Can be used in combination with
#' `show.record.number` for debugging purposes, e.g. corrupt BIN-files.
#' Can be provided as `list` if `file` is a `list`.
#'
#' @param zero_data.rm [logical] (*with default*):
#' remove erroneous data with no count values. As such data are usually not
#' needed for the subsequent data analysis they will be removed by default.
#' Can be provided as `list` if `file` is a `list`.
#'
#' @param duplicated.rm [logical] (*with default*):
#' remove duplicated entries if `TRUE`. This may happen due to an erroneous
#' produced BIN/BINX-file. This option compares only predecessor and successor.
#' Can be provided as `list` if `file` is a `list`.
#'
#' @param position [numeric] (*optional*):
#' imports only the selected position. Note: the import performance will not
#' benefit by any selection made here.
#' Can be provided as `list` if `file` is a `list`.
#'
#' @param fastForward [logical] (*with default*):
#' if `TRUE` for a more efficient data processing only a list of
#' [Luminescence::RLum.Analysis-class]
#' objects is returned instead of a [Luminescence::Risoe.BINfileData-class] object.
#' Can be provided as `list` if `file` is a `list`.
#'
#' @param show.record.number [logical] (*with default*):
#' shows record number of the imported record, for debugging usage only.
#' Can be provided as `list` if `file` is a `list`.
#' Ignored if `verbose = FALSE`.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar. Ignored if `verbose = FALSE`.
#'
#' @param forced.VersionNumber [integer] (*optional*):
#' allows to cheat the version number check in the function by own values for
#' cases where the BIN-file version is not supported.
#' Can be provided as `list` if `file` is a `list`.
#'
#' **Note:** The usage is at own risk, only supported BIN-file versions have been tested.
#'
#' @param ignore.RECTYPE [logical] or [numeric] (*with default*):
#' this argument allows to ignore values in the byte 'RECTYPE' (BIN-file version 08),
#' in case there are not documented or faulty set. In this case the corrupted records are skipped.
#' If the setting is [numeric] (e.g., `ignore.RECTYPE = 128`), records of those type are ignored
#' for import.
#'
#' @param pattern [character] (*optional*):
#' regular expression pattern passed to [list.files] to construct a list of
#' files to read (used only when `file` specifies a path).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... further arguments that will be passed to the function
#' [Luminescence::Risoe.BINfileData2RLum.Analysis]. Please note that any matching argument
#' automatically sets `fastForward = TRUE`
#'
#' @return
#' Returns an S4 [Luminescence::Risoe.BINfileData-class] object containing two
#' slots:
#'
#' \item{METADATA}{A [data.frame] containing all variables stored in the BIN-file.}
#' \item{DATA}{A [list] containing a numeric [vector] of the measured data.
#' The ID corresponds to the record ID in METADATA.}
#'
#' If `fastForward = TRUE` an [Luminescence::RLum.Analysis-class] object is
#' returned (coerced by [Luminescence::Risoe.BINfileData2RLum.Analysis]).
#'
#' Results are returned as a list when multiple files are processed or `file`
#' is a list.
#'
#' @note
#' The function works for BIN/BINX-format versions 03, 04, 05, 06, 07 and 08. The
#' version number depends on the used Sequence Editor.
#'
#' @section Function version: 0.19
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Margret C. Fuchs, HZDR Freiberg, (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#' based on information provided by Torben Lapp and Karsten Bracht Nielsen (Risø DTU, Denmark)
#'
#'
#' @seealso [Luminescence::write_R2BIN], [Luminescence::Risoe.BINfileData-class],
#' [base::readBin], [Luminescence::merge_Risoe.BINfileData], [Luminescence::RLum.Analysis-class]
#' [utils::txtProgressBar], [list.files]
#'
#'
#'@references
#'DTU Nutech, 2016. The Sequence Editor, Users Manual, February, 2016.
#'[https://www.fysik.dtu.dk]()
#'
#'
#'@keywords IO
#'
#'@examples
#'
#'file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
#'temp <- read_BIN2R(file)
#'temp
#'
#' @export
read_BIN2R <- function(
  file,
  show.raw.values = FALSE,
  position = NULL,
  n.records = NULL,
  zero_data.rm = TRUE,
  duplicated.rm = FALSE,
  fastForward = FALSE,
  show.record.number = FALSE,
  txtProgressBar = TRUE,
  forced.VersionNumber = NULL,
  ignore.RECTYPE = FALSE,
  pattern = NULL,
  verbose = TRUE,
  ...
) {
  .set_function_name("read_BIN2R")
  on.exit(.unset_function_name(), add = TRUE)

  # Self Call -----------------------------------------------------------------------------------
  # Option (a): Input is a list, every element in the list will be treated as file connection
  # with that many file can be read in at the same time
  # Option (b): The input is just a path, the function tries to grep ALL BIN/BINX files in the
  # directory and import them, if this is detected, we proceed as list
  file <- .validate_file(file, ext = c("bin", "binx"),
                         pattern = pattern %||% "\\.bin.*$",
                         recursive = TRUE,
                         verbose = verbose)

  if (is.list(file)) {
    ## expand input arguments
    rep.length <- length(file)

    position <- .listify(position, rep.length)
    n.records <- .listify(n.records, rep.length)
    zero_data.rm <- .listify(zero_data.rm, rep.length)
    duplicated.rm <- .listify(duplicated.rm, rep.length)
    show.raw.values <- .listify(show.raw.values, rep.length)
    show.record.number <- .listify(show.record.number, rep.length)
    forced.VersionNumber <- .listify(forced.VersionNumber, rep.length)

    temp.return <- lapply(seq_along(file), function(x) {
      temp <- try(read_BIN2R(
        file = file[[x]],
        fastForward = fastForward,
        position = position[[x]],
        n.records = n.records[[x]],
        duplicated.rm = duplicated.rm[[x]],
        zero_data.rm = zero_data.rm[[x]],
        show.raw.values =  show.raw.values[[x]],
        show.record.number = show.record.number[[x]],
        txtProgressBar = txtProgressBar,
        forced.VersionNumber = forced.VersionNumber[[x]],
        ignore.RECTYPE = ignore.RECTYPE,
        verbose = verbose,
        ...
      ), outFile = stdout()) # redirect error messages so they can be silenced
      if (inherits(temp, "try-error"))
        return(NULL)
      temp
    })

    ##return
    if (fastForward) {
      return(unlist(temp.return, recursive = FALSE))

    }else{
      return(temp.return)
    }
  }

  ## Config -----------------------------------------------------------------

  .validate_length(file, 1)
  .validate_logical_scalar(show.raw.values)
  .validate_logical_scalar(zero_data.rm)
  .validate_logical_scalar(duplicated.rm)
  .validate_logical_scalar(fastForward)
  .validate_logical_scalar(show.record.number)
  .validate_logical_scalar(txtProgressBar)
  .validate_class(ignore.RECTYPE, c("logical", "numeric"), length = 1)

  ##set file_link for internet downloads
  on_exit <- function(){
    ##close connection
    if(exists("con") && !is.null(con)){
      close(con)
    }
  }
  on.exit(expr = on_exit(), add = TRUE)

  ## never show the progress bar if not verbose
  if (!verbose) {
    txtProgressBar <- FALSE
  }

  ##set supported BIN format version
  VERSIONS.supported <- as.raw(c(03, 04, 05, 06, 07, 08))

  ## Short file parsing to get number of records ----------------------------

  #open connection
  con <- file(file, "rb")

  ##read data up to the end of con
  ##set ID
  temp.ID <- 0

  ##start for BIN-file check up
  while(length(temp.VERSION <- readBin(con, what="raw", 1, size=1, endian="little"))>0) {
    ## force version number
    if(!is.null(forced.VersionNumber)){
      temp.VERSION <- as.raw(forced.VersionNumber)
      if (verbose)
        .throw_message("'forced.VersionNumber' set to ", temp.VERSION,
                       ", but this version may not match your input file", error = FALSE)
    }

    ##stop input if wrong VERSION
    if (!temp.VERSION %in% VERSIONS.supported) {
      if(temp.ID > 0){
        if(is.null(n.records)){
          .throw_warning("BIN-file appears to be corrupt, import limited ",
                         "to the first ", temp.ID, " records")
        }else{
          .throw_warning("BIN-file appears to be corrupt, 'n.records' ",
                         "reset to ", temp.ID)
        }

        ##set or reset n.records
        n.records <- seq_len(temp.ID)
        break()

      }else{
        .throw_error("BIN/BINX format version (", temp.VERSION, ") ",
                     "is not supported or file is broken. ",
                     "Supported version numbers are: ",
                     .collapse(VERSIONS.supported))
      }
    }

    #empty byte position
    seek.connection(con, 1, origin = "current")

    ## get record LENGTH
    int.size <- if (temp.VERSION >= 05) 4 else 2
    temp.LENGTH  <- readBin(con, what = "integer", 1, size = int.size,
                            endian = "little")

    num.toread <- max(0, temp.LENGTH - int.size - 2)
    if (num.toread > 0) {
      seek.connection(con, num.toread, origin = "current")
    } else {
      if (verbose) {
        message("") # add a newline
        .throw_message("Record #", temp.ID + 1,
                       " skipped due to wrong record length", error = FALSE)
      }
      next()
    }
    temp.ID <- temp.ID + 1
  }

  ##set n.length we will need it later
  n.length <- temp.ID
  if (n.length == 0) {
    .throw_warning("0 records read, NULL returned")
    return(NULL)
  }

  rm(temp.ID)
  close(con) ##we have to close the connection here

# Set Lookup tables  --------------------------------------------------------------------------

  ##LTYPE
  LTYPE.lookup <- c(
    "0" = "TL",
    "1" = "OSL",
    "2" = "IRSL",
    "3" = "M-IR",
    "4" = "M-VIS",
    "5" = "TOL",
    "6" = "TRPOSL",
    "7" = "RIR",
    "8" = "RBR",
    "9" = "USER",
    "10" = "POSL",
    "11" = "SGOSL",
    "12" = "RL",
    "13" = "XRF"
  )

  ##DTYPE
  DTYPE.lookup <-
    c(
      "0" = "Natural",
      "1" = "N+dose",
      "2" = "Bleach",
      "3" = "Bleach+dose",
      "4" = "Natural (Bleach)",
      "5" = "N+dose (Bleach)",
      "6" = "Dose",
      "7" = "Background"
    )

  ##LIGHTSOURCE
  LIGHTSOURCE.lookup <- c(
    "0" = "None",
    "1" = "Lamp",
    "2" = "IR diodes/IR Laser",
    "3" = "Calibration LED",
    "4" = "Blue Diodes",
    "5" = "White light",
    "6" = "Green laser (single grain)",
    "7" = "IR laser (single grain)"
  )

  ## helper to import string from pascal format
  .read_string <- function(con, field.length, force.size = NULL) {
    raw <- readBin(con, what = "raw", field.length, size = 1, endian = "little")
    strlen <- max(as.integer(raw[1]), 0)
    if (strlen > 0) {
      if (!is.null(force.size))
        strlen <- force.size
      return(suppressWarnings(readChar(raw[-1], strlen, useBytes = TRUE)))
    }
    return("") # nocov
  }

  ##PRESET VALUES
  temp.RECTYPE <- 0

  ##overwrite length if required
  if(!is.null(n.records))
    n.length <- length(n.records)

  ## set index for entry row in table
  id_row <- 1

  ## 1 to 7
  ID       <- integer(length = n.length)
  SEL      <- NULL # derived from TAG
  VERSION  <- rep_len(NA_integer_, n.length)
  LENGTH   <- integer(length = n.length)
  PREVIOUS <- integer(length = n.length)
  NPOINTS  <- integer(length = n.length)
  RECTYPE  <- integer(length = n.length)

  ## 8 to 17
  RUN         <- rep_len(NA_integer_, n.length)
  SET         <- rep_len(NA_integer_, n.length)
  POSITION    <- integer(n.length) # default value 0
  GRAIN       <- NULL # derived from GRAINNUMBER
  GRAINNUMBER <- integer(n.length) # default value 0
  CURVENO     <- rep_len(NA_integer_, n.length)
  XCOORD      <- rep_len(NA_integer_, n.length)
  YCOORD      <- rep_len(NA_integer_, n.length)
  SAMPLE      <- character(length = n.length)
  COMMENT     <- character(length = n.length)

  ## 18 to 22
  SYSTEMID <- rep_len(NA_integer_, n.length)
  FNAME    <- character(length = n.length)
  USER     <- character(length = n.length)
  TIME     <- character(length = n.length)
  DATE     <- character(length = n.length)

  ## 23 to 31
  DTYPE   <- character(length = n.length)
  BL_TIME <- rep_len(NA_real_, n.length)
  BL_UNIT <- rep_len(NA_integer_, n.length)
  NORM1   <- rep_len(NA_real_, n.length)
  NORM2   <- rep_len(NA_real_, n.length)
  NORM3   <- rep_len(NA_real_, n.length)
  BG      <- rep_len(NA_real_, n.length)
  SHIFT   <- rep_len(NA_integer_, n.length)
  TAG     <- rep_len(NA_integer_, n.length)

  ## 32 to 67
  LTYPE        <- character(length = n.length)
  LIGHTSOURCE  <- character(length = n.length)
  LPOWER       <- NULL # derived from LIGHTPOWER
  LIGHTPOWER   <- rep_len(NA_real_, n.length)
  LOW          <- rep_len(NA_real_, n.length)
  HIGH         <- rep_len(NA_real_, n.length)
  RATE         <- rep_len(NA_real_, n.length)
  TEMPERATURE  <- rep_len(NA_real_, n.length)
  MEASTEMP     <- rep_len(NA_real_, n.length)
  AN_TEMP      <- rep_len(NA_real_, n.length)
  AN_TIME      <- rep_len(NA_real_, n.length)
  TOLDELAY     <- rep_len(NA_integer_, n.length)
  TOLON        <- rep_len(NA_integer_, n.length)
  TOLOFF       <- rep_len(NA_integer_, n.length)
  IRR_TIME     <- rep_len(NA_real_, n.length)
  IRR_TYPE     <- rep_len(NA_integer_, n.length)
  IRR_UNIT     <- rep_len(NA_integer_, n.length)
  IRR_DOSERATE <- rep_len(NA_real_, n.length)
  IRR_DOSERATEERR <- rep_len(NA_real_, n.length)
  TIMESINCEIRR  <- rep_len(NA_real_, n.length)
  TIMETICK      <- rep_len(NA_real_, n.length)
  ONTIME        <- rep_len(NA_real_, n.length)
  OFFTIME       <- rep_len(NA_real_, n.length)
  STIMPERIOD    <- rep_len(NA_integer_, n.length)
  GATE_ENABLED  <- rep_len(NA_real_, n.length)
  ENABLE_FLAGS  <- NULL # derived from GATE_ENABLED
  GATE_START    <- rep_len(NA_real_, n.length)
  GATE_STOP     <- rep_len(NA_real_, n.length)
  PTENABLED     <- rep_len(NA_real_, n.length)
  DTENABLED     <- rep_len(NA_real_, n.length)
  DEADTIME      <- rep_len(NA_real_, n.length)
  MAXLPOWER     <- rep_len(NA_real_, n.length)
  XRF_ACQTIME   <- rep_len(NA_real_, n.length)
  XRF_HV        <- rep_len(NA_real_, n.length)
  XRF_CURR      <- rep_len(NA_real_, n.length)
  XRF_DEADTIMEF <- rep_len(NA_real_, n.length)

  ## 68 to 79
  DETECTOR_ID    <- rep_len(NA_integer_, n.length)
  LOWERFILTER_ID <- rep_len(NA_integer_, n.length)
  UPPERFILTER_ID <- rep_len(NA_integer_, n.length)
  ENOISEFACTOR   <- rep_len(NA_real_, n.length)
  MARKPOS_X1 <- MARKPOS_Y1 <- rep_len(NA_real_, n.length)
  MARKPOS_X2 <- MARKPOS_Y2 <- rep_len(NA_real_, n.length)
  MARKPOS_X3 <- MARKPOS_Y3 <- rep_len(NA_real_, n.length)
  EXTR_START <- rep_len(NA_real_, n.length)
  EXTR_END   <- rep_len(NA_real_, n.length)

  ## 80
  SEQUENCE <- character(length = n.length)

  #set variable for DPOINTS handling
  results.DATA <- list()

  ##set list for RESERVED values
  results.RESERVED <- rep(list(list()), n.length)

  # Open Connection ---------------------------------------------------------

  #open connection
  con <- file(file, "rb")

  if (verbose) {
    cat("\n[read_BIN2R()] Importing ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file)))
    cat("\n n_rec:", n.length)
    cat("\n")
  }

  ##set progress bar
  if (txtProgressBar) {
    info <- file.info(file)
    pb <- txtProgressBar(min = 0, max = info$size, char = "=", style = 3)
  }

  ##read data up to the end of con

  ##set ID
  temp.ID <- 0

  # LOOP --------------------------------------------------------------------
  ##start loop for import BIN data
  while(length(temp.VERSION <- readBin(con, what="raw", 1, size=1, endian="little"))>0) {

    ##force version number
    if(!is.null(forced.VersionNumber)){
      temp.VERSION <- as.raw(forced.VersionNumber)
    }

    ##print record ID for debugging purposes
    if (show.record.number && verbose) {
        cat(temp.ID,",", sep = "")
        if(temp.ID%%10==0){
          cat("\n")
        }
    }

    #empty byte position
    seek.connection(con, 1, origin = "current")

    ## (1) Header size and structure
    ## LENGTH, PREVIOUS, NPOINTS
    int.size <- if (temp.VERSION >= 05) 4 else 2
    temp <- readBin(con, what = "int", 3, size = int.size, endian = "little")
    temp.LENGTH   <- temp[1]
    temp.PREVIOUS <- temp[2]
    temp.NPOINTS  <- temp[3]

    ## skip record if not selected in n.records
    ## the first condition boosts the speed of reading if n.records is not used
    if (!is.null(n.records) && !(temp.ID + 1) %in% n.records) {
      temp.ID <- temp.ID + 1
      seek.connection(con, temp.LENGTH - 3 * int.size - 2, origin = "current")
      next()
    }

    ## these must be set only after the n.records check
    ## we don't set VERSION now, as we must leave it set to NA in case we
    ## decide to skip the current record
    LENGTH[id_row]   <- temp.LENGTH
    PREVIOUS[id_row] <- temp.PREVIOUS
    NPOINTS[id_row]  <- temp.NPOINTS

    ## ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ## BINX FORMAT SUPPORT
    if (temp.VERSION == 05 || temp.VERSION == 06 ||
        temp.VERSION == 07 || temp.VERSION == 08) {

      #for temp.VERSION == 08
      #RECTYPE
      if(temp.VERSION == 08){
        temp.RECTYPE <- readBin(con, what = "int", 1, size = 1, endian = "little", signed = FALSE)
        RECTYPE[id_row] <- temp.RECTYPE

        ## we can check for a specific value for temp.RECTYPE
        if (inherits(ignore.RECTYPE, "numeric") && temp.RECTYPE == ignore.RECTYPE) {
          seek.connection(con, temp.LENGTH - 15, origin = "current")
          if(verbose) {
            message("") # add a newline
            .throw_message("Record #", temp.ID + 1,
                           " skipped due to 'ignore.RECTYPE = ", ignore.RECTYPE,
                           "'", error = FALSE)
          }
          next()
        }

        if(temp.RECTYPE != 0 & temp.RECTYPE != 1 & temp.RECTYPE != 128) {
          ##jump to the next record by stepping the record length minus the already read bytes
          seek.connection(con, temp.LENGTH - 15, origin = "current")
          msg <- paste0("Byte RECTYPE = ", temp.RECTYPE,
                        " is not supported in record #", temp.ID + 1)
          if (!ignore.RECTYPE) {
            .throw_error(msg, ", set 'ignore.RECTYPE = TRUE' to skip this record")
          }

          ## skip to next record
          if (verbose) {
            message("") # add a newline
            .throw_message(msg, ", record skipped", error = FALSE)
          }
          temp.ID <- temp.ID + 1
          next()
        }
      }

      ## RECTYPE == 128
      ## If the RECTYPE is 128, only the header bytes until here make any sense,
      ## the rest are just random bytes (e-mail K.B., 2024-07-04)
      ## the header length is 507, hence we have to jump 507 - 15 to get
      ## the data
      ## This is a very ugly construction and the function should be refactored
      if (temp.RECTYPE == 128){
        seek.connection(con, 492, origin = "current")

      } else {
        ##(2) Sample characteristics
        ##RUN, SET, POSITION, GRAINNUMBER, CURVENO, XCOORD, YCOORD
        temp <- readBin(con, what = "int", 7, size = 2, endian = "little")
        RUN[id_row] <- temp[1]
        SET[id_row] <- temp[2]
        POSITION[id_row] <- temp[3]
        GRAINNUMBER[id_row] <- temp[4]
        CURVENO[id_row] <- temp[5]
        XCOORD[id_row] <- temp[6]
        YCOORD[id_row] <- temp[7]

        ##SAMPLE, COMMENT
        SAMPLE[id_row]  <- .read_string(con, 21)
        COMMENT[id_row] <- .read_string(con, 81)

        ##(3) Instrument and sequence characteristic
        ##SYSTEMID
        SYSTEMID[id_row] <- readBin(con, what = "integer", 1, size = 2, endian = "little")

        ## FNAME, USER, TIME, DATE
        FNAME[id_row] <- .read_string(con, 101)
        USER[id_row]  <- .read_string(con, 31)
        TIME[id_row]  <- .read_string(con, 7)
        DATE[id_row]  <- .read_string(con, 7, force.size = 6)

        ##(4) Analysis
        ##DTYPE
        DTYPE[id_row] <- readBin(con, what="int", 1, size=1, endian="little")

        ##BL_TIME
        BL_TIME[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ##BL_UNIT
        BL_UNIT[id_row] <- readBin(con, what="int", 1, size=1, endian="little")

        ##NORM1, NORM2, NORM3, BG
        temp <- readBin(con, what="double", 4, size=4, endian="little")
        NORM1[id_row] <- temp[1]
        NORM2[id_row] <- temp[2]
        NORM3[id_row] <- temp[3]
        BG[id_row] <- temp[4]

        ##SHIFT
        SHIFT[id_row] <- readBin(con, what = "int", 1, size = 2, endian = "little")

        ##TAG
        TAG[id_row] <- readBin(con, what = "int", 1, size = 1, endian = "little")

        ##RESERVED
        temp.RESERVED1 <-readBin(con, what="raw", 20, size=1, endian="little")

        ##(5) Measurement characteristics

        ##LTYPE
        ##LTYPESOURCE
        temp <- readBin(con, what = "integer", 2, size = 1, endian = "little")
        LTYPE[id_row] <- temp[1]
        LIGHTSOURCE[id_row] <- temp[2]

        ##LIGHTPOWER, LOW, HIGH, RATE
        temp <- readBin(con, what="double", 4, size=4, endian="little")
        LIGHTPOWER[id_row] <- temp[1]
        LOW[id_row] <- temp[2]
        HIGH[id_row] <- temp[3]
        RATE[id_row] <- temp[4]

        ##TEMPERATURE
        ##MEASTEMP
        temp <- readBin(con, what = "integer", 2, size = 2, endian = "little")
        TEMPERATURE[id_row] <- temp[1]
        MEASTEMP[id_row] <- temp[2]

        ##AN_TEMP
        ##AN_TIME
        temp <- readBin(con, what = "double", 2, size = 4, endian = "little")
        AN_TEMP[id_row] <- temp[1]
        AN_TIME[id_row] <- temp[2]

        ##DELAY, ON, OFF
        temp <- readBin(con, what="int", 3, size=2, endian="little")
        TOLDELAY[id_row] <- temp[1]
        TOLON[id_row] <- temp[2]
        TOLOFF[id_row] <- temp[3]

        ##IRR_TIME
        IRR_TIME[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ##IRR_TYPE
        IRR_TYPE[id_row] <- readBin(con, what="int", 1, size=1, endian="little")

        ##IRR_DOSERATE
        IRR_DOSERATE[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ##IRR_DOSERATEERR
        if(temp.VERSION != 05)
          IRR_DOSERATEERR[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ##TIMESINCEIRR
        TIMESINCEIRR[id_row] <- readBin(con, what="integer", 1, size=4, endian="little")

        ##TIMETICK
        TIMETICK[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ##ONTIME
        ##STIMPERIOD
        temp <- readBin(con, what = "integer", 2, size = 4, endian = "little")
        ONTIME[id_row] <- temp[1]
        STIMPERIOD[id_row] <- temp[2]

        ##GATE_ENABLED
        GATE_ENABLED[id_row] <- as.numeric(readBin(con, what="raw", 1, size=1, endian="little"))

        ##GATE_START
        ##GATE_STOP
        temp <- readBin(con, what = "integer", 2, size = 4, endian = "little")
        GATE_START[id_row] <- temp[1]
        GATE_STOP[id_row] <- temp[2]

        ##PTENABLED
        ##DTENABLED
        temp <- as.numeric(readBin(con, what = "raw", 2, size = 1, endian = "little"))
        PTENABLED[id_row] <- temp[1]
        DTENABLED[id_row] <- temp[2]

        ##DEADTIME, MAXLPOWER, XRF_ACQTIME, XRF_HV
        temp <- readBin(con, what="double", 4, size=4, endian="little")
        DEADTIME[id_row] <- temp[1]
        MAXLPOWER[id_row] <- temp[2]
        XRF_ACQTIME[id_row] <- temp[3]
        XRF_HV[id_row] <- temp[4]

        ##XRF_CURR
        XRF_CURR[id_row] <- readBin(con, what="integer", 1, size=4, endian="little")

        ##XRF_DEADTIMEF
        XRF_DEADTIMEF[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ###Account for differences between V5, V6 and V7
        if(temp.VERSION == 06){
          reserved2.len <- 24

        }else if(temp.VERSION == 05){
          reserved2.len <- 4

        }else{

          ##DETECTOR_ID
          DETECTOR_ID[id_row] <- readBin(con, what="int", 1, size=1, endian="little")

          ##LOWERFILTER_ID, UPPERFILTER_ID
          temp <- readBin(con, what="int", 2, size=2, endian="little")
          LOWERFILTER_ID[id_row] <- temp[1]
          UPPERFILTER_ID[id_row] <- temp[2]

          ##ENOISEFACTOR
          ENOISEFACTOR[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

          ##CHECK FOR VERSION 07
          if(temp.VERSION == 07){
            reserved2.len <- 15

          }else {
            ##MARKER_POSITION
            ###EXTR_START, EXTR_END
            temp <- readBin(con, what = "double", 8, size = 4, endian = "little")
            MARKPOS_X1[id_row] <- temp[1]
            MARKPOS_Y1[id_row] <- temp[2]
            MARKPOS_X2[id_row] <- temp[3]
            MARKPOS_Y2[id_row] <- temp[4]
            MARKPOS_X3[id_row] <- temp[5]
            MARKPOS_Y3[id_row] <- temp[6]
            EXTR_START[id_row] <- temp[7]
            EXTR_END[id_row]   <- temp[8]
            reserved2.len <- 42
          }
        }# end RECTYPE 128
        temp.RESERVED2 <- readBin(con, what = "raw", reserved2.len, size = 1,
                                  endian = "little")
      }
    }

    ## ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ## BIN FILE FORMAT SUPPORT
    else if (temp.VERSION == 03 || temp.VERSION == 04) {

      ##LTYPE
      LTYPE[id_row]<-readBin(con, what="int", 1, size=1, endian="little")

      ##LOW, HIGH, RATE
      temp <- readBin(con, what="double", 3, size=4, endian="little")
      LOW[id_row] <- temp[1]
      HIGH[id_row] <- temp[2]
      RATE[id_row] <- temp[3]

      ##XCOORD, YCOORD, TOLDELAY, TOLON, TOLOFF
      temp <- readBin(con, what = "integer", 6, size = 2, endian = "little")
      TEMPERATURE[id_row] <- temp[1]
      XCOORD[id_row]      <- temp[2]
      YCOORD[id_row]      <- temp[3]
      TOLDELAY[id_row]    <- temp[4]
      TOLON[id_row]       <- temp[5]
      TOLOFF[id_row]      <- temp[6]

      ##POSITION
      ##RUN
      temp <- readBin(con, what = "integer", 2, size = 1, endian = "little",
                      signed = FALSE)
      POSITION[id_row] <- temp[1]
      RUN[id_row] <- temp[2]

      ## TIME, DATE, SEQUENCE, USER
      TIME[id_row] <- .read_string(con, 7, force.size = 6)
      DATE[id_row] <- .read_string(con, 7, force.size = 6)
      SEQUENCE[id_row] <- .read_string(con, 9)
      USER[id_row] <- .read_string(con, 9)

      ##DTYPE
      DTYPE[id_row] <- readBin(con, what="int", 1, size=1, endian="little")

      ##IRR_TIME
      IRR_TIME[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

      ##IRR_TYPE
      ##IRR_UNIT
      temp <- readBin(con, what = "integer", 2, size = 1, endian = "little")
      IRR_TYPE[id_row] <- temp[1]
      IRR_UNIT[id_row] <- temp[2]

      ##BL_TIME
      BL_TIME[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

      ##BL_UNIT
      BL_UNIT[id_row] <- readBin(con, what="int", 1, size=1, endian="little")

      ##AN_TEMP, AN_TIME, NORM1, NORM2, NORM3, BG
      temp <- readBin(con, what="double", 6, size=4, endian="little")
      AN_TEMP[id_row] <- temp[1]
      AN_TIME[id_row] <- temp[2]
      NORM1[id_row] <- temp[3]
      NORM2[id_row] <- temp[4]
      NORM3[id_row] <- temp[5]
      BG[id_row] <- temp[6]

      ##SHIFT
      SHIFT[id_row] <- readBin(con, what="integer", 1, size=2, endian="little")

      ## SAMPLE, COMMENT
      SAMPLE[id_row]  <- .read_string(con, 21)
      COMMENT[id_row] <- .read_string(con, 81)

      ##LIGHTSOURCE, SET, TAG
      temp <- readBin(con, what="int", 3, size=1, endian="little")
      LIGHTSOURCE[id_row] <- temp[1]
      SET[id_row] <- temp[2]
      TAG[id_row] <- temp[3]

      ##GRAIN
      GRAINNUMBER[id_row] <- readBin(con, what="int", 1, size=2, endian="little")

      ##LPOWER
      LIGHTPOWER[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

      ##SYSTEMID
      SYSTEMID[id_row] <- readBin(con, what="integer", 1, size=2, endian="little")

      ##Unfortunately an inconsitent BIN-file structure forces a differenciation ...
      if(temp.VERSION == 03){
        ##RESERVED
        temp.RESERVED1 <- readBin(con, what="raw", 36, size=1, endian="little")

        ##ONTIME, OFFTIME
        temp <- readBin(con, what="double", 2, size=4, endian="little")
        ONTIME[id_row] <- temp[1]
        OFFTIME[id_row] <- temp[2]

        ##Enable flags  #GateEnabled for v 06
        GATE_ENABLED[id_row] <- as.numeric(readBin(con, what="raw", 1, size=1, endian="little"))

        ##ONGATEDELAY, OFFGATEDELAY
        temp <- readBin(con, what="double", 2, size=4, endian="little")
        GATE_START[id_row] <- temp[1]
        GATE_STOP[id_row] <- temp[2]

        ##RESERVED
        temp.RESERVED2<-readBin(con, what="raw", 1, size=1, endian="little")

      }else{
        ##RESERVED
        temp.RESERVED1<-readBin(con, what="raw", 20, size=1, endian="little")

        ##CURVENO
        CURVENO[id_row] <- readBin(con, what="integer", 1, size=2, endian="little")

        ##TIMETICK
        TIMETICK[id_row] <- readBin(con, what="double", 1, size=4, endian="little")

        ##ONTIME, STIMPERIOD
        temp <- readBin(con, what="integer", 2, size=4, endian="little")
        ONTIME[id_row] <- temp[1]
        STIMPERIOD[id_row] <- temp[2]

        ##GATE_ENABLED
        GATE_ENABLED[id_row] <- as.numeric(readBin(con, what="raw", 1, size=1, endian="little"))

        ##ONGATEDELAY, OFFGATEDELAY
        temp <- readBin(con, what="double", 2, size=4, endian="little")
        GATE_START[id_row] <- temp[1]
        GATE_STOP[id_row] <- temp[2]

        ##PTENABLED
        ##RESERVED
        temp <- readBin(con, what = "raw", 11, size = 1, endian = "little")
        PTENABLED[id_row] <- as.numeric(temp[1])
        temp.RESERVED2 <- temp[2:11]
      }
    }

    ## ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ## Unrecognised version
    ##
    ## We should already have raised a warning that the file is corrupt
    ## during the first scan of the BIN/BINX file: at that point we have
    ## set `n.records` so that we would stop reading before encountering
    ## again the record with unrecognised version number, so here we just
    ## assert that that is the case
    stopifnot(temp.VERSION %in% VERSIONS.supported)

     #DPOINTS
     if(temp.RECTYPE != 128) {
       temp.DPOINTS <- readBin(con, what = "integer", temp.NPOINTS, size = 4, endian = "little")

     } else {
       temp.DPOINTS <- lapply(1:temp.NPOINTS, function(x) {
         list(
           NOFPOINTS = readBin(con, what = "int", 1, size = 4, endian = "little"),
           USEDFOR = as.logical(readBin(con, what = "raw", 48, size = 1, endian = "little")),
           SHOWFOR = as.logical(readBin(con, what = "raw", 48, size = 1, endian = "little")),
           ROICOLOR = readBin(con, what = "integer", 1, size = 4, endian = "little"),
           X = readBin(con, what = "double", 50, size = 4, endian = "little"),
           Y = readBin(con, what = "double", 50, size = 4, endian = "little"))
       })
     }

    ## ----------------------------------------------------------------------
    ## close the current record

    temp.ID <- temp.ID + 1
    ID[id_row] <- temp.ID

    ## we set VERSION only now to its real value: as we've got here, we have
    ## read the entire record and decided to keep it
    VERSION[id_row]  <- as.numeric(temp.VERSION)

    ## update the progress bar
    if (txtProgressBar) {
      setTxtProgressBar(pb, seek.connection(con, origin = "current"))
    }

    results.DATA[[id_row]] <- temp.DPOINTS
    results.RESERVED[[id_row]][[1]] <- temp.RESERVED1
    results.RESERVED[[id_row]][[2]] <- temp.RESERVED2

    ## update id row
    id_row <- id_row + 1

  }#endwhile::end loop

  ##close
  if (txtProgressBar) {
    close(pb)
  }

  ## generate the final data.table
  results.METADATA <- data.table(
      ID,
      SEL = as.logical(TAG),
      VERSION,
      LENGTH,
      PREVIOUS,
      NPOINTS,
      RECTYPE,
      RUN,
      SET,
      POSITION,
      GRAIN = GRAINNUMBER,
      GRAINNUMBER,
      CURVENO,
      XCOORD,
      YCOORD,
      SAMPLE,
      COMMENT,
      SYSTEMID,
      FNAME,
      USER,
      TIME,
      DATE,
      DTYPE,
      BL_TIME,
      BL_UNIT,
      NORM1,
      NORM2,
      NORM3,
      BG,
      SHIFT,
      TAG,
      LTYPE,
      LIGHTSOURCE,
      LPOWER = LIGHTPOWER,
      LIGHTPOWER,
      LOW,
      HIGH,
      RATE,
      TEMPERATURE,
      MEASTEMP,
      AN_TEMP,
      AN_TIME,
      TOLDELAY,
      TOLON,
      TOLOFF,
      IRR_TIME,
      IRR_TYPE,
      IRR_UNIT,
      IRR_DOSERATE,
      IRR_DOSERATEERR,
      TIMESINCEIRR,
      TIMETICK,
      ONTIME,
      OFFTIME,
      STIMPERIOD,
      GATE_ENABLED,
      ENABLE_FLAGS = GATE_ENABLED,
      GATE_START,
      GATE_STOP,
      PTENABLED,
      DTENABLED,
      DEADTIME,
      MAXLPOWER,
      XRF_ACQTIME,
      XRF_HV,
      XRF_CURR,
      XRF_DEADTIMEF,
      DETECTOR_ID,
      LOWERFILTER_ID,
      UPPERFILTER_ID,
      ENOISEFACTOR,
      MARKPOS_X1, MARKPOS_Y1,
      MARKPOS_X2, MARKPOS_Y2,
      MARKPOS_X3, MARKPOS_Y3,
      EXTR_START,
      EXTR_END,
      SEQUENCE
    )

  ## remove NA values created by skipping records
  results.METADATA <- stats::na.omit(results.METADATA, cols = "VERSION")

  ##output
  if(verbose)
    message("\t >> ", length(results.DATA), " records read successfully\n")

  ## Record removals --------------------------------------------------------

  ## check if only the specified positions should be returned
  if(!is.null(position)){

    ##check whether the position is valid at all
    if (results.METADATA[, all(position %in% POSITION)]) {
      keep.positions <- results.METADATA[, POSITION %in% position]
      results.METADATA <- results.METADATA[(keep.positions), ]
      results.DATA <- results.DATA[keep.positions]
      results.RESERVED <- results.RESERVED[keep.positions]

      if (verbose) {
        .throw_message("Kept records matching 'position': ",
                       .collapse(position, quote = FALSE), error = FALSE)
      }
    }else{
      .throw_warning("At least one position number is not valid, ",
                     "valid position numbers are: ",
                     .collapse(results.METADATA[, unique(POSITION)],
                               quote = FALSE))
    }
  }

  ##check for position that have no data at all (error during the measurement)
  if(zero_data.rm){
    zero_data.check <- which(lengths(results.DATA) == 0)

    ##remove records if there is something to remove
    if (length(zero_data.check) > 0) {
      results.METADATA <- results.METADATA[-zero_data.check, ]
      results.DATA[zero_data.check] <- NULL
      results.RESERVED[zero_data.check] <- NULL

      .throw_warning("Zero-data records detected and removed: ",
                     .collapse(zero_data.check, quote = FALSE))
    }
  }

  ## if nothing is left, return an empty object
  if (nrow(results.METADATA) == 0) {
    if (verbose)
      .throw_message("Empty object returned", error = FALSE)
    return(set_Risoe.BINfileData())
  }

  ##check for duplicated entries and remove them if wanted, but only if we have more than 2 records
  ##this check is skipped for results with a RECTYPE 128, which stems from camera measurements
  if (n.length >= 2 && length(results.DATA) >= 2 && all(results.METADATA[["RECTYPE"]] != 128)) {
    duplication.check <- suppressWarnings(which(c(
      0, vapply(
        2:length(results.DATA),
        FUN = function(x) {
          length(results.DATA[[x - 1]]) == length(results.DATA[[x]]) &&
          all(results.DATA[[x - 1]] == results.DATA[[x]])
        },
        FUN.VALUE = 1
      )
    ) == 1))
    if (length(duplication.check) != 0) {
      if (duplicated.rm) {
        ##remove records
        results.METADATA <- results.METADATA[-duplication.check, ]
        results.DATA[duplication.check] <- NULL
        results.RESERVED[duplication.check] <- NULL

        ##message
        if(verbose) {
          .throw_message("Duplicated records detected and removed: ",
                         .collapse(duplication.check, quote = FALSE), error = FALSE)
        }

      } else{
        .throw_warning("Duplicated records detected: ",
                       .collapse(duplication.check, quote = FALSE),
                       "\n >> You should consider using 'duplicated.rm = TRUE'.")
      }
    }
  }

  ## recalculate ID as some records may not have been read if n.records
  ## was set or they were dropped in one of the previous blocks
  if (results.METADATA[, .N != n.length || max(ID) > n.length]) {
    results.METADATA[, ID := 1:.N]
    if (verbose)
      .throw_message("The record index has been recalculated", error = FALSE)
  }

  # Convert Translation Matrix Values ---------------------------------------

  if (!show.raw.values) {
    ##LIGHTSOURCE CONVERSION
    results.METADATA[, LIGHTSOURCE := unname(LIGHTSOURCE.lookup[LIGHTSOURCE])]

    ##LTYPE CONVERSION
    results.METADATA[, LTYPE := unname(LTYPE.lookup[LTYPE])]

    ##DTYPE CONVERSION
    results.METADATA[, DTYPE := unname(DTYPE.lookup[DTYPE])]

    ## check for oddly set LTYPES, this may happen in old BIN-file versions
    if (results.METADATA$VERSION[1] == 3) {
      results.METADATA[LTYPE == "OSL" & LIGHTSOURCE == "IR diodes/IR Laser",
                       LTYPE := "IRSL"]
    }

    ##TIME CONVERSION, do not do for odd time formats as this could cause problems during export
    results.METADATA[nchar(TIME) == 5,
                     TIME := paste0("0", TIME)]
    results.METADATA[nchar(TIME) == 6,
                     TIME := format(as.POSIXct(TIME, format = "%H%M%S"),
                                    "%H:%M:%S")]
  }

  ## check for empty BIN-files names ... if so, set the name of the file as BIN-file name
  ## This can happen if the user uses different equipment
  if (results.METADATA[, !any(nzchar(FNAME))]) {
    results.METADATA[, FNAME := tools::file_path_sans_ext(basename(file))]
  }

  ## produce S4 object for output
  object <- set_Risoe.BINfileData(
    METADATA = results.METADATA,
    DATA = results.DATA,
    .RESERVED = results.RESERVED)

  ## Fast Forward -----------------------------------------------------------

  ## set fastForward to TRUE if arguments to Risoe.BINfileData2RLum.Analysis
  ## were specified
  if (!fastForward) {
    dots <- ...names()
    args <- dots[dots %in% names(formals(Risoe.BINfileData2RLum.Analysis))[-1]]
    if (length(args) > 0) {
      fastForward <- TRUE
      .throw_warning("Additional arguments specified: ", .collapse(args),
                     ", setting 'fastForward = TRUE'")
    }
  }

  ##return values
  ##with fast fastForward they will be converted directly to a list of RLum.Analysis objects
  if(fastForward){
     object <- Risoe.BINfileData2RLum.Analysis(object, ...)

     ##because we expect a list
     if(!inherits(object, "list"))
       object <- list(object)
  }

   return(object)
}

test_that("input validation", {
  testthat::skip_on_cran()

  bin.v3 <- test_path("_data/BINfile_V3.bin")

  expect_error(read_BIN2R(TRUE),
               "'file' should be of class 'character' or 'list'")
  expect_error(read_BIN2R(character(0)),
               "'file' cannot be an empty character")
  expect_error(read_BIN2R(list()),
               "'file' cannot be an empty list")
  expect_error(read_BIN2R(file = "error"),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_error(read_BIN2R(test_path("test_read_BIN2R.R")),
               "extension 'R' is not supported, only 'bin' and 'binx' are valid")
  expect_error(read_BIN2R(test_path("_data/BINfile_V3.bin"), verbose = FALSE,
                          forced.VersionNumber = 1),
               "BIN/BINX format version (01) is not supported or file is broken",
               fixed = TRUE)
  SW({
  expect_warning(read_BIN2R(bin.v3, position = 99),
                 "At least one position number is not valid")
  expect_message(read_BIN2R(bin.v3, forced.VersionNumber = 3),
                 "'forced.VersionNumber' set to 03, but this version")
  })

  ## arguments
  expect_error(read_BIN2R(bin.v3, show.raw.values = "error"),
               "'show.raw.values' should be a single logical value")
  expect_error(read_BIN2R(bin.v3, zero_data.rm = list(FALSE)),
               "'zero_data.rm' should be a single logical value")
  expect_error(read_BIN2R(bin.v3, duplicated.rm = "error"),
               "'duplicated.rm' should be a single logical value")
  expect_error(read_BIN2R(bin.v3, fastForward = "error"),
               "'fastForward' should be a single logical value")
  expect_error(read_BIN2R(bin.v3, show.record.number = NULL),
               "'show.record.number' should be a single logical value")
  expect_error(read_BIN2R(bin.v3, ignore.RECTYPE = "error"),
               "'ignore.RECTYPE' should be of class 'logical' or 'numeric'")
  expect_error(read_BIN2R(bin.v3, ignore.RECTYPE = 1:2),
               "'ignore.RECTYPE' should be of class 'logical' or 'numeric' and have length 1")

  ## check for broken files
  zero <- tempfile(pattern = "zero", fileext = ".binx")
  file.create(zero)
  expect_error(read_BIN2R(zero, verbose = FALSE),
               "is a zero-byte file")
  write(raw(), zero)
  expect_error(read_BIN2R(zero, verbose = FALSE),
               "BIN/BINX format version \\(..\\) is not supported or file is")
  SW({
  expect_warning(
      expect_message(expect_null(read_BIN2R(zero, verbose = TRUE,
                                            forced.VersionNumber = 8)),
                     "Record #1 skipped due to wrong record length"),
      "0 records read, NULL returned")
  expect_warning(
      expect_message(expect_null(read_BIN2R(zero, verbose = TRUE,
                                            forced.VersionNumber = 3)),
                     "Record #1 skipped due to wrong record length"),
      "0 records read, NULL returned")
  })

  fake <- tempfile(pattern = "fake", fileext = ".binx")
  writeBin(as.raw(c(  8, 0,       # version
                    227, 5, 0, 0, # length
                      0, 0, 0, 0, # previous
                    250, 0, 0, 0, # npoints
                     99)          # rectype set to 99 to trigger errors
                  ), fake)
  expect_error(read_BIN2R(fake, verbose = FALSE),
               "Byte RECTYPE = 99 is not supported in record #1")
  SW({
  expect_message(res <- read_BIN2R(fake, verbose = TRUE, ignore.RECTYPE = TRUE),
                 "Byte RECTYPE = 99 is not supported in record #1")
  })
  expect_length(res, 0)
})

test_that("test the import of various BIN-file versions", {
  testthat::skip_on_cran()

  ## this test needs an internet connection ... test for it
  github.url <- file.path("https://github.com/R-Lum/Luminescence",
                          "raw/archive/dev_0.9.x/tests/testthat/_data")

    ## V3
    bin <- read_BIN2R(file.path(github.url, "BINfile_V3.bin"),
                      verbose = FALSE)
    ## remove randomness in FNAME due to the use of a temporary file
  bin@METADATA$FNAME <- substr(bin@METADATA$FNAME, 1, 9)
    expect_snapshot_plain(bin)

  ## V4
  expect_snapshot_plain(read_BIN2R(test_path("_data/BINfile_V4.bin"),
                                   verbose = FALSE))

  ## V5
  expect_snapshot_plain(read_BIN2R(test_path("_data/BINfile_V5.binx"),
                                   verbose = FALSE))

  ## V6
  expect_snapshot_plain(read_BIN2R(test_path("_data/BINfile_V6.binx"),
                                   verbose = FALSE))

  ## V7
  expect_snapshot_plain(read_BIN2R(test_path("_data/BINfile_V7.binx"),
                                   verbose = FALSE))

  SW({
  ## V8 - as part of the package
  bin.v8 <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  expect_snapshot_plain(read_BIN2R(bin.v8, verbose = FALSE))

  expect_message(res <- read_BIN2R(bin.v8, txtProgressBar = FALSE,
                                   position = 2),
                 "Kept records matching 'position': 2")
  expect_equal(length(res@.RESERVED), nrow(res@METADATA))
  expect_message(res <- read_BIN2R(bin.v8, txtProgressBar = FALSE,
                                   position = 1, fastForward = TRUE),
                 "Kept records matching 'position': 1")
  expect_type(res, "list")

    ## test n.records argument
    t_n.records_1 <- expect_s4_class(
      read_BIN2R(
        file = bin.v8,
        txtProgressBar = FALSE,
        n.records = 1), class = "Risoe.BINfileData")
    expect_length(t_n.records_1, n = 1)

    t_n.records_0 <- expect_s4_class(
      read_BIN2R(
        file = bin.v8,
        txtProgressBar = FALSE,
        n.records = 0), class = "Risoe.BINfileData")
    expect_length(t_n.records_0, n = 0)

    t_n.records_1_2 <- expect_s4_class(
      read_BIN2R(
        file = bin.v8,
        txtProgressBar = FALSE,
        n.records = 1:2), class = "Risoe.BINfileData")
    expect_length(t_n.records_1_2, n = 2)

  ## directory
  res <- read_BIN2R(test_path("_data"), show_record_number = TRUE)
  expect_type(res, "list")
  expect_length(res, 11)

  res <- read_BIN2R(test_path("_data"), pattern = "_V[34]")
  expect_type(res, "list")
  expect_length(res, 2)

  ## test further options
  ## n.records and fastForward
  expect_type(read_BIN2R(test_path("_data/BINfile_V4.bin"),
                         txtProgressBar = FALSE, n.records = 1,
                         fastForward = TRUE, verbose = FALSE),
              "list")

  ## options to Risoe.BINfileData2RLum.Analysis
  expect_warning(read_BIN2R(bin.v8, fastForward = FALSE, protocol = "test"),
                 "Additional arguments specified: 'protocol'")

  ## check ignore RECTYPE settings
  expect_message(t <- expect_s4_class(read_BIN2R(bin.v8, verbose = TRUE,
                                                 ignore.RECTYPE = 1),
                                      class = "Risoe.BINfileData"),
                 "Record #1 skipped due to 'ignore.RECTYPE = 1'")
  })

    ## should be zero now
    expect_length(t@DATA, n = 0)

  SW({
  res <- read_BIN2R(list(bin.v8, bin.v8), verbose = TRUE,
             position = list(1, 2), n.records = list(1, 2),
             show.raw.values = list(TRUE), zero_data.rm = list(FALSE),
             duplicated.rm = list(TRUE), show.record.number = list(TRUE),
             forced.VersionNumber = list(8), fastForward = TRUE)
  expect_length(res[[2]], 1)

  res <- read_BIN2R(c(bin.v8, bin.v8), verbose = FALSE, n.records = 2,
                    fastForward = TRUE)
  expect_length(res, 2)

  res <- read_BIN2R(bin.v8, verbose = FALSE, n.records = 2, fastForward = TRUE)
  expect_length(res, 1)

  res <- read_BIN2R(test_path("_data/BINfile_V3.bin"), n.records = 2)
  expect_length(res, 1)
  })
})

test_that("test hand-crafted files", {
  testthat::skip_on_cran()

  ## corrupted
  corrupted.bin <- test_path("_data/bin-tests/corrupted.bin")
  expect_warning(res <- read_BIN2R(corrupted.bin, verbose = FALSE),
                 "BIN-file appears to be corrupt, import limited to the first")
  expect_equal(nrow(res@METADATA), 1)
  expect_warning(res <- read_BIN2R(corrupted.bin, verbose = FALSE,
                                   n.records = 2),
                 "BIN-file appears to be corrupt, 'n.records' reset to 1")
  expect_equal(nrow(res@METADATA), 1)

  ## two versions
  res <- read_BIN2R(test_path("_data/bin-tests/two-versions.binx"),
                    verbose = FALSE)
  expect_equal(nrow(res@METADATA), 4)

  ## rectype 128
  rectype.128.bin <- test_path("_data/bin-tests/rectype-128.binx")
  res <- read_BIN2R(rectype.128.bin, verbose = FALSE)
  expect_equal(nrow(res@METADATA), 3)

  expect_output(res <- read_BIN2R(rectype.128.bin, verbose = FALSE,
                                  fastForward = TRUE),
                "BIN/BINX-file non-conform. TL curve may be wrong")
  expect_type(res, "list")
  expect_s4_class(res[[1]]@records[[1]], "RLum.Data.Curve")

  ## duplicate
  duplicate.bin <- test_path("_data/bin-tests/duplicated-records.binx")
  expect_warning(res <- read_BIN2R(duplicate.bin, verbose = FALSE),
                 "Duplicated records detected: 2")
  expect_equal(nrow(res@METADATA), 3)

  SW({
  expect_message(res <- read_BIN2R(duplicate.bin,
                                   duplicated.rm = TRUE, verbose = TRUE),
                 "Duplicated records detected and removed: 2")
  expect_equal(nrow(res@METADATA), 2)
  expect_equal(length(res@.RESERVED), nrow(res@METADATA))

  ## zero-data
  zero.data.bin <- test_path("_data/bin-tests/zero-data-record.binx")
  expect_warning(res <- read_BIN2R(zero.data.bin, verbose = TRUE),
                 "Zero-data records detected and removed: 2")
  expect_equal(nrow(res@METADATA), 1)
  expect_equal(length(res@.RESERVED), nrow(res@METADATA))

  expect_silent(res <- read_BIN2R(list(zero.data.bin), verbose = FALSE,
                                  zero_data.rm = FALSE))
  expect_equal(nrow(res[[1]]@METADATA), 2)

  expect_silent(res <- read_BIN2R(zero.data.bin, verbose = FALSE,
                                  zero_data.rm = FALSE))

  zero.data.all <- test_path("_data/bin-tests/zero-data-all.binx")
  expect_message(expect_warning(read_BIN2R(zero.data.all, txtProgressBar = FALSE),
                                "Zero-data records detected and removed: 1, 2"),
                 "Empty object returned")
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1260
  expect_error(do.call(read_BIN2R, list("a test")),
               "File '.*a test' does not exist") # windows CI needs the regexp
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

create_UID <- function() {
    .Call(`_Luminescence_create_UID`)
}

src_EED_Calc_Overall_StatUncertainty <- function(M_Simul, Ndata, Nsimul, MinNbSimExp) {
    .Call(`_Luminescence_src_EED_Calc_Overall_StatUncertainty`, M_Simul, Ndata, Nsimul, MinNbSimExp)
}

f_BTS_cpp_part <- function(x, A, Eu, s10, Et, kB, T_K, DeltaE, rhop) {
    .Call(`_Luminescence_f_BTS_cpp_part`, x, A, Eu, s10, Et, kB, T_K, DeltaE, rhop)
}

src_analyse_IRSARRF_SRS <- function(values_regenerated_limited, values_natural_limited, vslide_range, n_MC, trace = FALSE) {
    .Call(`_Luminescence_analyse_IRSARRF_SRS`, values_regenerated_limited, values_natural_limited, vslide_range, n_MC, trace)
}

src_create_RLumDataCurve_matrix <- function(DATA, VERSION, NPOINTS, LTYPE, LOW, HIGH, AN_TEMP, TOLDELAY, TOLON, TOLOFF) {
    .Call(`_Luminescence_create_RLumDataCurve_matrix`, DATA, VERSION, NPOINTS, LTYPE, LOW, HIGH, AN_TEMP, TOLDELAY, TOLON, TOLOFF)
}

fit_functionEXP_cpp <- function(a, b, c, x) {
    .Call(`_Luminescence_fit_functionEXP_cpp`, a, b, c, x)
}

fit_functionEXPLIN_cpp <- function(a, b, c, g, x) {
    .Call(`_Luminescence_fit_functionEXPLIN_cpp`, a, b, c, g, x)
}

fit_functionEXPEXP_cpp <- function(a1, a2, b1, b2, x) {
    .Call(`_Luminescence_fit_functionEXPEXP_cpp`, a1, a2, b1, b2, x)
}

fit_functionGOK_cpp <- function(a, b, c, d, x) {
    .Call(`_Luminescence_fit_functionGOK_cpp`, a, b, c, d, x)
}

src_get_XSYG_curve_values <- function(s) {
    .Call(`_Luminescence_src_get_XSYG_curve_values`, s)
}

#' Apply the maximum age model to a given De distribution
#'
#' Function to fit the maximum age model to De data. This is a wrapper function
#' that calls [Luminescence::calc_MinDose] and applies a similar approach as described in
#' Olley et al. (2006).
#'
#' **Data transformation**
#'
#' To estimate the maximum dose population
#' and its standard error, the three parameter minimum age model of Galbraith
#' et al. (1999) is adapted. The measured De values are transformed as follows:
#'
#' 1. convert De values to natural logs
#' 2. multiply the logged data to create a mirror image of the De distribution
#' 3. shift De values along x-axis by the smallest x-value found to obtain only positive values
#' 4. combine in quadrature the measurement error associated with each De value
#' with a relative error specified by `sigmab`
#' 5. apply the MAM to these data
#'
#' When all calculations are done the results are then converted as follows
#' 1. subtract the x-offset
#' 2. multiply the natural logs by -1
#' 3. take the exponent to obtain the maximum dose estimate in Gy
#'
#' **Further documentation**
#'
#' Please see [Luminescence::calc_MinDose].
#'
#' @inheritParams calc_MinDose
#'
#' @param ... further arguments for bootstrapping (`bs.M, bs.N, bs.h, sigmab.sd`).
#' See details for their usage.
#'
#' @return Please see [Luminescence::calc_MinDose].
#'
#' @section Function version: 0.3.2
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Based on a rewritten S script of Rex Galbraith, 2010
#'
#' @seealso [Luminescence::calc_CentralDose], [Luminescence::calc_CommonDose], [Luminescence::calc_FiniteMixture],
#' [Luminescence::calc_FuchsLang2001], [Luminescence::calc_MinDose]
#'
#' @references
#' Arnold, L.J., Roberts, R.G., Galbraith, R.F. & DeLong, S.B.,
#' 2009. A revised burial dose estimation procedure for optical dating of young
#' and modern-age sediments. Quaternary Geochronology 4, 306-325.
#'
#' Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for mixed fission
#' track ages. Nuclear Tracks Radiation Measurements 4, 459-470.
#'
#' Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley, J.M.,
#' 1999. Optical dating of single grains of quartz from Jinmium rock shelter,
#' northern Australia. Part I: experimental design and statistical models.
#' Archaeometry 41, 339-364.
#'
#' Galbraith, R.F., 2005. Statistics for
#' Fission Track Analysis, Chapman & Hall/CRC, Boca Raton.
#'
#' Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error
#' calculation and display in OSL dating: An overview and some recommendations.
#' Quaternary Geochronology 11, 1-27.
#'
#' Olley, J.M., Roberts, R.G., Yoshida, H., Bowler, J.M., 2006. Single-grain optical dating of grave-infill
#' associated with human burials at Lake Mungo, Australia. Quaternary Science
#' Reviews 25, 2469-2474
#'
#' **Further reading**
#'
#' Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
#' (De) distributions: Implications for OSL dating of sediment mixtures.
#' Quaternary Geochronology 4, 204-230.
#'
#' Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
#' assessment of procedures for estimating burial dose. Quaternary Science
#' Reviews 25, 2475-2502.
#'
#' Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
#' Quaternary Geochronology 12, 98-106.
#'
#' Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
#' of optical dating of fluvial deposits.  Quaternary Geochronology 1, 109-120.
#'
#' Rodnight, H., 2008. How many equivalent dose values are needed to
#' obtain a reproducible distribution?. Ancient TL 26, 3-10.
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' # apply the maximum dose model
#' calc_MaxDose(ExampleData.DeValues$CA1, sigmab = 0.2, par = 3)
#'
#' @export
calc_MaxDose<- function(
  data,
  sigmab,
  log=TRUE,
  par=3,
  bootstrap=FALSE,
  init.values = NULL,
  plot=TRUE,
  ...
){
  res<- calc_MinDose(data, sigmab, log, par, bootstrap, init.values, plot=FALSE, invert=TRUE, ...)
  res@originator<- "calc_MaxDose"
  if (plot) {
    try(plot_RLum.Results(res, ...),
        outFile = stdout()) # redirect error messages so they can be silenced
  }

  invisible(res)
}

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  SW({
  expect_warning(calc_MaxDose(ExampleData.DeValues$CA1, sigmab = 0.1,
                              log = FALSE, log.output = TRUE,
                              verbose = TRUE, plot = FALSE),
                 "'log' reset to TRUE")
  })
})

test_that("check functionality", {
  testthat::skip_on_cran()

  temp <- calc_MaxDose(ExampleData.DeValues$CA1,
                       sigmab = 0.2,
                       par = 3,
                       plot = TRUE,
                       verbose = FALSE)

  expect_s4_class(temp, "RLum.Results")
  expect_equal(length(temp), 9)

  results <- get_RLum(temp)

  expect_equal(round(results$de, digits = 2), 76.58)
  expect_equal(round(results$de_err, digits = 2), 7.57)
  expect_equal(results$ci_level, 0.95)
  expect_equal(round(results$ci_lower, digits = 2), 69.65)
  expect_equal(round(results$ci_upper, digits = 2), 99.33)
  expect_equal(results$par, 3)
  expect_equal(round(results$sig, digits = 2), 1.71)
  expect_equal(round(results$p0, digits = 2), 0.65)
  expect_equal(results$mu, NA_real_)
  expect_equal(round(results$Lmax, digits = 2), -19.79)
  expect_equal(round(results$BIC, digits = 2), 58.87)
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1320
  set.seed(9)
  SW({
  expect_warning(calc_MaxDose(data = data.frame(De = c(rnorm(4) + 5, -1),
                                                De_Err = rnorm(5) + 1),
                              sigmab = 1, log = TRUE, bootstrap = TRUE,
                              bs.M = 10, bs.N = 5, bs.h = 2, plot = FALSE),
                 "De values must be positive with 'log = TRUE', 1 values set to NA")
  })
})

#' @title Apply the central age model (CAM) after Galbraith et al. (1999) to a given
#' De distribution
#'
#' @description
#' This function calculates the central dose and dispersion of the De
#' distribution, their standard errors and the profile log likelihood function
#' for `sigma`.
#'
#' This function uses the equations of Galbraith & Roberts (2012). The
#' parameters `delta` and `sigma` are estimated by numerically solving
#' eq. 15 and 16. Their standard errors are approximated using eq. 17.
#' In addition, the profile log-likelihood function for `sigma` is
#' calculated using eq. 18 and presented as a plot. Numerical values of the
#' maximum likelihood approach are **only** presented in the plot and **not**
#' in the console. A detailed explanation on maximum likelihood estimation can
#' be found in the appendix of Galbraith & Laslett (1993, 468-470) and
#' Galbraith & Roberts (2012, 15)
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with De `(data[,1])` and De error `(data[,2])`.
#' Records containing missing values will be removed.
#'
#' @param sigmab [numeric] (*with default*):
#' additional spread in De values, representing the expected overdispersion in
#' the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
#' **Note:** For the logged model (`log = TRUE`) this value must be
#' a fraction, e.g. 0.2 (= 20 %). If the un-logged model is used (`log = FALSE`),
#' sigmab must be provided in the same absolute units of the De values (seconds or Gray).
#'
#' @param log [logical] (*with default*):
#' fit the (un-)logged central age model to De data. Log transformation is
#' allowed only if the De values are positive.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further arguments (`trace`, `verbose`).
#'
#' @return Returns a plot (*optional*) and terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the following elements:
#'
#' \item{.$summary}{[data.frame] summary of all relevant model results.}
#' \item{.$data}{[data.frame] original input data}
#' \item{.$args}{[list] used arguments}
#' \item{.$call}{[call] the function call}
#' \item{.$profile}{[data.frame] the log likelihood profile for sigma}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @section Function version: 1.5
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Based on a rewritten S script of Rex Galbraith, 2010
#'
#' @seealso [plot], [Luminescence::calc_CommonDose], [Luminescence::calc_FiniteMixture],
#' [Luminescence::calc_FuchsLang2001], [Luminescence::calc_MinDose]
#'
#' @references
#' Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for
#' mixed fission track ages. Nuclear Tracks Radiation Measurements 4, 459-470.
#'
#' Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley,
#' J.M., 1999. Optical dating of single grains of quartz from Jinmium rock
#' shelter, northern Australia. Part I: experimental design and statistical
#' models.  Archaeometry 41, 339-364.
#'
#' Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error calculation and
#' display in OSL dating: An overview and some recommendations. Quaternary
#' Geochronology 11, 1-27.
#'
#' **Further reading**
#'
#' Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
#' (De) distributions: Implications for OSL dating of sediment mixtures.
#' Quaternary Geochronology 4, 204-230.
#'
#' Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
#' assessment of procedures for estimating burial dose. Quaternary Science
#' Reviews 25, 2475-2502.
#'
#' Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
#' Quaternary Geochronology 12, 98-106.
#'
#' Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
#' of optical dating of fluvial deposits.  Quaternary Geochronology 1 109-120.
#'
#' Rodnight, H., 2008. How many equivalent dose values are needed to
#' obtain a reproducible distribution?. Ancient TL 26, 3-10.
#'
#' @examples
#'
#' ##load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ##apply the central dose model
#' calc_CentralDose(ExampleData.DeValues$CA1)
#'
#' @export
calc_CentralDose <- function(
  data,
  sigmab = 0,
  log = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_CentralDose")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("data.frame", "RLum.Results"))
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data, "data")
  }
  .validate_nonnegative_scalar(sigmab)

  ##remove NA values
  if (anyNA(data)) {
    .throw_message(length(which(is.na(data))), " NA values removed from dataset",
                   error = FALSE)
    data <- na.exclude(data)
  }

  ## check that we have enough data available
  if(ncol(data) < 2 || nrow(data) < 2)
    .throw_error("'data' should have at least two columns and two rows")

  ##extract only the first two columns and set column names
  data <- data[,1:2]
  colnames(data) <- c("ED", "ED_Error")

  ## don't allow log transformation if there are non-positive values
  .validate_logical_scalar(log)
  if (log && any(data[, 1] <= 0)) {
    log <- FALSE
    .throw_warning("'data' contains non-positive De values, 'log' set to FALSE")
  }
  .validate_logical_scalar(plot)

  ## don't allow negative errors, silently make them positive
  if (any(data[, 2] < 0)) {
    data[, 2] <- abs(data[, 2])
  }

  if (log && sigmab > 1) {
    .throw_error("'sigmab' should be a value between 0 and 1 if log = TRUE")
  }

  ## ============================================================================##
  ## ... ARGUMENTS
  ## ============================================================================##

  options <- modifyList(list(verbose = TRUE,
                             trace = FALSE),
                        list(...))
  .validate_logical_scalar(options$verbose, name = "'verbose'")
  .validate_logical_scalar(options$trace, name = "'trace'")

  ## deprecated argument
  if ("na.rm" %in% ...names()) {
    .throw_warning("'na.rm' is deprecated, missing values are always removed ",
                   "by default")
  }

  ## ============================================================================##
  ## CALCULATIONS
  ## ============================================================================##

  # calculate yu = log(ED) and su = se(logED)
  if (log) {
    yu <- log(data$ED)
    su <- sqrt((data$ED_Error / data$ED)^2 + sigmab^2)
  } else {
    yu <- data$ED
    su <- sqrt((data$ED_Error)^2 + sigmab^2)
  }

  # What does the code do?
  # >> email conversation with Rex Galbraith 2019-06-29
  # >> "fixed point iteration" method to estimate sigma
  # >> starting with a fixed value
  # >> once sqrt(sum((wu^2) * (yu - delta)^2 / sum(wu))) gets equal to 1
  # >> the iteration is complete
  # >> if everything has converged against those fixed values
  # >> this is the maximum likelihood estimate for
  # >> sigma and delta

  # calculate starting values and weights
  sigma <- 0.15 # keep in mind that this is a relative value
  wu <- 1 / (sigma^2 + su^2)

  # compute mle's
  for (j in 1:200) {
    sum_wu <- sum(wu)
    delta <- sum(wu * yu) / sum_wu
    sigma <- sigma * sqrt(sum((wu^2) * (yu - delta)^2) / sum_wu)
    wu <- 1 / (sigma^2 + su^2)

    # print iterations
    if (options$trace)
      print(round(c(delta, sigma), 4))

    ## don't let sigma become zero
    if (sigma < 1e-16)
      break
  }

  # log likelihood
  llik <- 0.5 * sum(log(wu)) - 0.5 * sum(wu * (yu - delta)^2)

  # standard errors
  sedelta <- 1 / sqrt(sum(wu))
  sesigma <- 1 / sqrt(2 * sigma^2 * sum(wu^2))

  # save parameters for terminal output
  if (log) {
    out.delta <- exp(delta)
    out.sigma <- sigma * 100
    out.sedelta <- sedelta * 100
    out.sesigma <- sesigma
  } else {
    out.delta <- delta
    out.sigma <- sigma / out.delta * 100
    out.sedelta <- sedelta / out.delta * 100
    out.sesigma <- sqrt((sedelta / delta)^2 +
                          (sesigma / out.delta * 100 / out.sigma)^2) * out.sigma / 100
  }
  if (is.infinite(out.sesigma))
    out.sesigma <- NA

  # profile log likelihood
  Lmax <- llik
  llik <- 0
  sig0 <- max(0, sigma - 8 * sesigma)
  sig1 <- sigma + 9.5 * sesigma
  sig <- try(seq(sig0, sig1, sig1 / 1000), silent = TRUE)

  if (!inherits(sig, "try-error")) {
    # TODO: rewrite this loop as a function and maximise with mle2 ll is the actual
    # log likelihood, llik is a vector of all ll
    for (s in sig) {
      wu <- 1 / (s^2 + su^2)
      mu <- sum(wu * yu)/sum(wu)
      ll <- 0.5 * sum(log(wu)) - 0.5 * sum(wu * (yu - mu)^2)
      llik <- c(llik, ll)
    }
    llik <- llik[-1] - Lmax
  }

  ## ============================================================================##
  ## TERMINAL OUTPUT
  ## ============================================================================##

  if (options$verbose) {
    cat("\n [calc_CentralDose]")
    cat("\n\n----------- meta data ----------------")
    cat("\n n:                      ", length(yu))
    cat("\n log:                    ", log)
    cat("\n----------- dose estimate ------------")
    cat("\n abs. central dose:      ", format(out.delta, digits = 2, nsmall = 2))
    cat("\n abs. SE:                ", format(out.delta * out.sedelta / 100,
                                              digits = 2, nsmall = 2))
    cat("\n rel. SE [%]:            ", format(out.sedelta, digits = 2, nsmall = 2))
    cat("\n----------- overdispersion -----------")
    cat("\n abs. OD:                ", format(ifelse(log, sigma * out.delta, sigma),
                                              digits = 2, nsmall = 2))
    cat("\n abs. SE:                ", format(ifelse(log, sesigma * out.delta, sesigma),
                                              digits = 2, nsmall = 2))
    cat("\n OD [%]:                 ", format(out.sigma, digits = 2, nsmall = 2))
    cat("\n SE [%]:                 ", if (!is.na(out.sesigma)) {
      format(out.sesigma * 100, digits = 2, nsmall = 2)
    } else {
      "-"
    })
    cat("\n-------------------------------------\n\n")
  }

  ## ============================================================================##
  ## RETURN VALUES
  ## ============================================================================##

  if(!log)  sig <- sig / delta

  summary <- data.frame(
    de = out.delta,
    de_err = out.delta * out.sedelta / 100,
    OD = ifelse(log, sigma * out.delta, sigma),
    OD_err = ifelse(log, sesigma * out.delta, sesigma),
    rel_OD = out.sigma,
    rel_OD_err = out.sesigma * 100,
    Lmax = Lmax)

  args <- list(log = log, sigmab = sigmab)

  newRLumResults.calc_CentralDose <- set_RLum(
    class = "RLum.Results",
    data = list(
      summary = summary,
      data = data,
      args = args,
      profile = data.frame(
        sig = if(!inherits(sig, "try-error")) sig else NA,
        llik = llik)
    ),
    info = list(
      call = sys.call()
    )
  )

  ## =========## PLOTTING
  if (plot && !inherits(sig, "try-error"))
    try(plot_RLum.Results(newRLumResults.calc_CentralDose, ...))

  invisible(newRLumResults.calc_CentralDose)
}

## load data
data(ExampleData.DeValues, envir = environment())

temp <- calc_CentralDose(
  ExampleData.DeValues$CA1,
  plot = FALSE,
  verbose = FALSE)

set.seed(1)
temp_NA <- data.frame(rnorm(10)+5, rnorm(10)+5)
temp_NA[1,1] <- NA

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_CentralDose(data = "error"),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_CentralDose(temp, sigmab = -10),
               "'sigmab' should be a single non-negative value")
  expect_error(calc_CentralDose(temp, sigmab = 10),
               "'sigmab' should be a value between 0 and 1 if log = TRUE")
  expect_error(calc_CentralDose(data.frame()),
               "should have at least two columns and two rows")
  expect_error(calc_CentralDose(iris, iris),
               "'sigmab' should be a single non-negative value")
  expect_error(calc_CentralDose(iris, log = NA),
               "'log' should be a single logical value")
  expect_error(calc_CentralDose(iris, plot = NA),
               "'plot' should be a single logical value")
  expect_error(calc_CentralDose(iris, verbose = NA),
               "'verbose' should be a single logical value")

  SW({
  expect_s4_class(calc_CentralDose(temp_NA), "RLum.Results")
  expect_warning(calc_CentralDose(temp, na.rm = FALSE),
                 "'na.rm' is deprecated, missing values are always")
  expect_message(calc_CentralDose(temp_NA),
                 "NA values removed from dataset")
  })
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  ## negative De values
  neg_vals <- data.frame(De = c(-0.56, -0.16, 0.0, 0.04),
                         De.err = c(0.15, 0.1, 0.12, 0.1))
  expect_warning(calc_CentralDose(neg_vals, log = TRUE),
                 "'data' contains non-positive De values, 'log' set to FALSE")

  ## negative De errors
  neg_errs <- data.frame(De = c(0.56, 0.16, 0.1, 0.04),
                         De.err = c(-0.15, -0.1, 0.12, -0.1))
  res1 <- calc_CentralDose(neg_errs)
  res2 <- calc_CentralDose(abs(neg_errs))
  res1@info$call <- res2@info$call <- NULL
  res1@.uid <- res2@.uid <- NA_character_
  expect_equal(res1, res2)
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6
  expect_snapshot_RLum(temp,
                       tolerance = snapshot.tolerance)
  SW({
  expect_snapshot_RLum(calc_CentralDose(ExampleData.DeValues$CA1,
                                        log = FALSE, trace = TRUE),
                       tolerance = snapshot.tolerance)

  expect_snapshot_RLum(calc_CentralDose(temp_NA, log = FALSE),
                       tolerance = snapshot.tolerance)

  ## more coverage
  df <- data.frame(De = c(1e-160, 1e-156, 1e-120, 4e-22),
                   De.err = c(1e5, 1e40, 1e12, 1e28))
  expect_snapshot_RLum(calc_CentralDose(df),
                       tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              calc_CentralDose(ExampleData.DeValues$CA1))
  vdiffr::expect_doppelganger("log = FALSE",
                              calc_CentralDose(ExampleData.DeValues$CA1,
                                               log = FALSE))
  vdiffr::expect_doppelganger("sigmab 0.2",
                              calc_CentralDose(ExampleData.DeValues$CA1,
                                               sigmab = 0.2))
  vdiffr::expect_doppelganger("sigmab 0.3",
                              calc_CentralDose(ExampleData.DeValues$CA1,
                                               sigmab = 0.3))
  })
})

## load data
data(ExampleData.DeValues, envir = environment())
temp <- calc_CentralDose(ExampleData.DeValues$CA1, plot = FALSE, verbose = FALSE)

temp_RLumDataCurve <- set_RLum(class = "RLum.Data.Curve")
temp_RLumDataImage <- set_RLum(class = "RLum.Data.Image")
temp_RLumDataSpectrum <- set_RLum(class = "RLum.Data.Spectrum")
temp_RLumAnalysis <- set_RLum(class = "RLum.Analysis")
temp_RLumResults <- set_RLum(class = "RLum.Results")

test_that("check class and length of output", {
  testthat::skip_on_cran()

  expect_s3_class(get_RLum(temp), class = "data.frame")
  expect_type(get_RLum(temp, data.object = "args"), "list")

  ##test objects
  expect_type(get_RLum(temp_RLumDataCurve), "double")
  expect_type(get_RLum(temp_RLumDataImage), "logical")
  expect_type(get_RLum(temp_RLumDataSpectrum), "logical")
  expect_null(suppressWarnings(get_RLum(temp_RLumAnalysis)))
  expect_null(get_RLum(temp_RLumResults))
})

test_that("check get_RLum on a list and NULL", {
  testthat::skip_on_cran()

  object <- set_RLum(class = "RLum.Analysis", records = rep(set_RLum(class = "RLum.Data.Curve"), 10))
  expect_warning(get_RLum(object, recordType = "test"),
                 "This request produced an empty list of records")
  expect_null(get_RLum(NULL), "NULL")

  expect_warning(res <- get_RLum(list(temp, "test")),
                 "object #2 in the list is not of class 'RLum'")
  expect_length(res, 2)
  res <- get_RLum(list(temp, temp_RLumAnalysis), null.rm = TRUE)
  expect_length(res, 1)

  ##check list of such objects
  l <- list(object, object)
  l[[1]]@records[[1]]@recordType <- "IRSL"
  expect_length(suppressWarnings(get_RLum(l, recordType = "IRSL", null.rm = FALSE)), 2)
  expect_length(suppressWarnings(get_RLum(l, recordType = "IRSL", null.rm = TRUE)), 0)

  ##check class argument
  a <- list(set_RLum("RLum.Results"),
            set_RLum("RLum.Analysis", records = list(set_RLum("RLum.Data.Curve"))))
  expect_type(get_RLum(a, class = "test", drop = FALSE), "list")
  expect_type(get_RLum(a, class = "test", drop = TRUE), "list")
  expect_type(get_RLum(a, class = "RLum.Results", drop = FALSE), "list")
  expect_type(get_RLum(a, class = "RLum.Analysis", drop = TRUE), "list")
  expect_type(get_RLum(list(temp_RLumResults, temp_RLumAnalysis)), "list")

  ## regression test - issue 1120
  extra <- iris[1:3, 1:3]
  class(extra) <- c("test", "class2")
  expect_warning(expect_length(get_RLum(list(extra, 2, 3), class = "test"), 1),
                 "object #1 in the list is not of class 'RLum'")
})

#' @title Collection of layout definitions
#'
#' @description
#' This helper function returns a list with layout definitions for homogeneous
#' plotting.
#'
#' The easiest way to create a user-specific layout definition is perhaps to
#' create either an empty or a default layout object and fill/modify the
#' definitions (`user.layout <- get_Layout(data = "empty")`).
#'
#' @param layout [character] or [list] object (**required**):
#' name of the layout definition to be returned. If name is provided the
#' respective definition is returned. One of the following
#' supported layout definitions is possible: `"default"`,
#' `"journal.1"`, `"small"`, `"empty"`.
#'
#' User-specific layout definitions must be provided as a list object of
#' predefined structure, see details.
#'
#' @return A list object with layout definitions for plot functions.
#'
#' @section Function version: 0.1
#'
#' @author Michael Dietze, GFZ Potsdam (Germany)
#'
#' @examples
#'
#' ## read example data set
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## show structure of the default layout definition
#' layout.default <- get_Layout(layout = "default")
#' str(layout.default)
#'
#' ## show colour definitions for Abanico plot, only
#' layout.default$abanico$colour
#'
#' ## set Abanico plot title colour to orange
#' layout.default$abanico$colour$main <- "orange"
#'
#' ## create Abanico plot with modofied layout definition
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  layout = layout.default)
#'
#' ## create Abanico plot with predefined layout "journal"
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  layout = "journal")
#'
#' @export
get_Layout <- function(
  layout
) {
  .set_function_name("get_Layout")
  on.exit(.unset_function_name(), add = TRUE)

  ## pre-defined layout selections
  if (is.character(layout) && length(layout) == 1) {

    if (!layout %in% c("empty", "default", "journal")) {
      .throw_warning("Layout definition not supported, ",
                     "default layout is used.")
      return(get_Layout(layout = "default"))
    }

    switch(layout,
    empty = {
      layout <- list(

        ## empty Abanico plot -------------------------------------------------
        abanico = list(
          font.type = list(
            main    = character(1),
            xlab1   = character(1),
            xlab2   = character(1),
            ylab    = character(1),
            zlab    = character(1),
            xtck1   = character(1),
            xtck2   = character(1),
            xtck3   = character(1),
            ytck    = character(1),
            ztck    = character(1),
            mtext   = character(1),
            summary = character(1), # optionally vector
            stats   = character(1), # optionally vector
            legend  = character(1) # optionally vector
          ),

          font.size = list(
            main    = numeric(1),
            xlab1   = numeric(1),
            xlab2   = numeric(1),
            xlab3   = numeric(1),
            ylab    = numeric(1),
            zlab    = numeric(1),
            xtck1   = numeric(1),
            xtck2   = numeric(1),
            xtck3   = numeric(1),
            ytck    = numeric(1),
            ztck    = numeric(1),
            mtext   = numeric(1),
            summary = numeric(1), # optionally vector
            stats   = numeric(1), # optionally vector
            legend  = numeric(1)  # optionally vector
          ),

          font.deco = list(
            main    = character(1),
            xlab1   = character(1),
            xlab2   = character(1),
            xlab3   = character(1),
            ylab    = character(1),
            zlab    = character(1),
            xtck1   = character(1),
            xtck2   = character(1),
            xtck3   = character(1),
            ytck    = character(1),
            ztck    = character(1),
            mtext   = character(1),
            summary = character(1), # optionally vector
            stats   = character(1), # optionally vector
            legend  = character(1) # optionally vector
          ),

          colour = list(
            main    = numeric(1), # plot title colour
            xlab1   = numeric(1), # left x-axis label colour
            xlab2   = numeric(1), # right x-axis label colour
            xlab3   = numeric(1), # right x-axis label colour
            ylab    = numeric(1), # y-axis label colour
            zlab    = numeric(1), # z-axis label colour
            xtck1   = numeric(1), # left x-axis tick colour
            xtck2   = numeric(1), # right x-axis tick colour
            xtck3   = numeric(1), # right x-axis tick colour
            ytck    = numeric(1), # y-axis tick colour
            ztck    = numeric(1), # z-axis tick colour
            mtext   = numeric(1), # subheader text colour
            summary = numeric(1), # statistic summary colour
            stats   = numeric(1), # value statistics colour
            legend  = numeric(1), # legend colour
            centrality = numeric(1), # Centrality line colour
            value.dot  = numeric(1), # De value dot colour
            value.bar  = numeric(1), # De value error bar colour
            value.rug  = numeric(1), # De value rug colour
            poly.line  = numeric(1), # polygon line colour
            poly.fill  = numeric(1), # polygon fill colour
            bar.line   = numeric(1), # polygon line colour
            bar.fill   = numeric(1), # polygon fill colour
            kde.line   = numeric(1),
            kde.fill   = numeric(1),
            grid.major = numeric(1),
            grid.minor = numeric(1),
            border     = numeric(1),
            background = numeric(1)),

          dimension = list(
            figure.width    = numeric(1), # figure width in mm
            figure.height   = numeric(1), # figure height in mm
            margin          = numeric(4), # margin sizes in mm
            main.line       = numeric(1), # line height in %
            xlab1.line      = numeric(1), # line height in %
            xlab2.line      = numeric(1), # line height in %
            xlab3.line      = numeric(1), # line height in %
            ylab.line       = numeric(1), # line height in %
            zlab.line       = numeric(1), # line height in %
            xtck1.line      = numeric(1), # line height in %
            xtck2.line      = numeric(1), # line height in %
            xtck3.line      = numeric(1), # line height in %
            ytck.line       = numeric(1), # line height in %
            ztck.line       = numeric(1), # line height in %
            xtcl1           = numeric(1), # tick length in %
            xtcl2           = numeric(1), # tick length in %
            xtcl3           = numeric(1), # tick length in %
            ytcl            = numeric(1), # tick length in %
            ztcl            = numeric(1), # tick length in %
            rugl            = numeric(1), # rug length in %
            mtext           = numeric(1), # line height in %
            summary.line    = numeric(1) # line height in %
          )),

        ## empty KDE plot -----------------------------------------------------
        kde = list(
          font.type = list(
            main   = character(1),
            xlab   = character(1),
            ylab1  = character(1),
            ylab2  = character(1),
            xtck   = character(1),
            ytck1  = character(1),
            ytck2  = character(1),
            stats  = character(1), # optionally vector
            legend = character(1) # optionally vector
          ),

          font.size = list(
            main   = numeric(1),
            xlab   = numeric(1),
            ylab1  = numeric(1),
            ylab2  = numeric(1),
            xtck   = numeric(1),
            ytck1  = numeric(1),
            ytck2  = numeric(1),
            stats  = numeric(1), # optionally vector
            legend = numeric(1) # optionally vector
          ),

          font.deco = list(
            main   = character(1),
            xlab   = character(1),
            ylab1  = character(1),
            ylab2  = character(1),
            xtck   = character(1),
            ytck1  = character(1),
            ytck2  = character(1),
            stats  = character(1), # optionally vector
            legend = character(1) # optionally vector
          ),

          colour = list(
            main   = numeric(1), # plot title colour
            xlab   = numeric(1), # x-axis label colour
            ylab1  = numeric(1), # primary y-axis label colour
            ylab2  = numeric(1), # secondary y-axis label colour
            xtck   = numeric(1), # x-axis tick colour
            ytck1  = numeric(1), # primary y-axis tick colour
            ytck2  = numeric(1), # secondary y-axis tick colour
            box    = numeric(1), # plot frame box line colour
            mtext  = numeric(1), # subheader text colour
            stats  = numeric(1), # statistic summary colour
            kde.line        = numeric(1),  # KDE line colour
            kde.fill        = numeric(1),  # KDE fill colour
            value.dot       = numeric(1),  # De value dot colour
            value.bar       = numeric(1),  # De value error bar colour
            value.rug       = numeric(1),  # De value rug colour
            boxplot.line    = numeric(1), # boxplot line colour
            boxplot.fill    = numeric(1), # boxplot fill colour
            mean.line       = numeric(1),  # mean line colour
            sd.bar          = numeric(1),  # sd-line colour
            background      = numeric(1)), # background colour

          dimension = list(
            figure.width    = numeric(1), # figure width in mm
            figure.height   = numeric(1), # figure height in mm
            margin          = numeric(4), # margin sizes in mm
            main.line       = numeric(1), # line height in %
            xlab.line       = numeric(1), # line height in %
            ylab1.line      = numeric(1), # line height in %
            ylab2.line      = numeric(1), # line height in %
            xtck.line       = numeric(1), # line height in %
            ytck1.line      = numeric(1), # line height in %
            ytck2.line      = numeric(1), # line height in %
            xtcl            = numeric(1), # tick length in %
            ytcl1           = numeric(1), # tick length in %
            ytcl2           = numeric(1), # tick length in %
            stats.line      = numeric(1) # line height in %
          )
        )
      )
    },

    default = {
      layout <- list(

        ## default Abanico plot -----------------------------------------------
        abanico = list(
          font.type = list(
            main    = "",
            xlab1   = "",
            xlab2   = "",
            ylab    = "",
            zlab    = "",
            xtck1   = "",
            xtck2   = "",
            xtck3   = "",
            ytck    = "",
            ztck    = "",
            mtext   = "",
            summary = "", # optionally vector
            stats   = "", # optionally vector
            legend  = "" # optionally vector
          ),

          font.size = list(
            main    = 12,
            xlab1   = 12,
            xlab2   = 12,
            xlab3   = 12,
            ylab    = 12,
            zlab    = 12,
            xtck1   = 12,
            xtck2   = 12,
            xtck3   = 12,
            ytck    = 12,
            ztck    = 12,
            mtext   = 10,
            summary = 10, # optionally vector
            stats   = 10, # optionally vector
            legend  = 10 # optionally vector
          ),

          font.deco = list(
            main    = "bold",
            xlab1   = "normal",
            xlab2   = "normal",
            xlab3   = "normal",
            ylab    = "normal",
            zlab    = "normal",
            xtck1   = "normal",
            xtck2   = "normal",
            xtck3   = "normal",
            ytck    = "normal",
            ztck    = "normal",
            mtext   = "normal",
            summary = "normal", # optionally vector
            stats   = "normal", # optionally vector
            legend  = "normal" # optionally vector
          ),

          colour = list(
            main    = 1, # plot title colour
            xlab1   = 1, # left x-axis label colour
            xlab2   = 1, # right x-axis label colour
            xlab3   = 1, # right x-axis label colour
            ylab    = 1, # y-axis label colour
            zlab    = 1, # z-axis label colour
            xtck1   = 1, # left x-axis tick colour
            xtck2   = 1, # right x-axis tick colour
            xtck3   = 1, # right x-axis tick colour
            ytck    = 1, # y-axis tick colour
            ztck    = 1, # z-axis tick colour
            mtext   = 1, # subheader text colour
            summary = 1, # statistic summary colour
            stats   = 1, # value statistics colour
            legend  = 1, # legend colour
            centrality = 1, # Centrality line colour
            value.dot  = 1, # De value dot colour
            value.bar  = 1, # De value error bar colour
            value.rug = 1, # De value rug colour
            poly.line  = NA, # polygon line colour
            poly.fill  = adjustcolor("grey75", alpha.f = 0.6), # polygon fill colour
            bar.line   = NA, # polygon line colour
            bar.fill   = "grey60", # bar fill colour
            kde.line   = 1,
            kde.fill   = NA,
            grid.major = "grey80",
            grid.minor = "none",
            border     = 1,
            background = NA),

          dimension = list(
            figure.width    = "auto", # figure width in mm
            figure.height   = "auto", # figure height in mm
            margin = c(10, 10, 10, 10), # margin sizes in mm
            main.line       = 100, # line height in %
            xlab1.line      = 90, # line height in %
            xlab2.line      = 90, # line height in %
            xlab3.line      = 90, # line height in %
            ylab.line       = 100, # line height in %
            zlab.line       = 70, # line height in %
            xtck1.line      = 100, # line height in %
            xtck2.line      = 100, # line height in %
            xtck3.line      = 100, # line height in %
            ytck.line       = 100, # line height in %
            ztck.line       = 100, # line height in %
            xtcl1           = 100, # tick length in %
            xtcl2           = 100, # tick length in %
            xtcl3           = 100, # tick length in %
            ytcl            = 100, # tick length in %
            ztcl            = 100, # tick length in %
            rugl            = 100, # rug length in %
            mtext           = 100, # line height in %
            summary.line    = 100 # line height in %
          )),

        ## default KDE plot ---------------------------------------------------
        kde = list(
          font.type = list(
            main   = "",
            xlab   = "",
            ylab1  = "",
            ylab2  = "",
            xtck   = "",
            ytck1  = "",
            ytck2  = "",
            stats  = "", # optionally vector
            legend = "" # optionally vector
          ),

          font.size = list(
            main   = 14,
            xlab   = 12,
            ylab1  = 12,
            ylab2  = 12,
            xtck   = 12,
            ytck1  = 12,
            ytck2  = 12,
            mtext  = 10,
            stats  = 12, # optionally vector
            legend = 12 # optionally vector
          ),

          font.deco = list(
            main   = "bold",
            xlab   = "normal",
            ylab1  = "normal",
            ylab2  = "normal",
            xtck   = "normal",
            ytck1  = "normal",
            ytck2  = "normal",
            stats  = "normal", # optionally vector
            legend = "normal" # optionally vector
          ),

          colour = list(
            main   = 1, # plot title colour
            xlab   = 1, # x-axis label colour
            ylab1  = 1, # primary y-axis label colour
            ylab2  = 1, # secondary y-axis label colour
            xtck   = 1, # x-axis tick colour
            ytck1  = 1, # primary y-axis tick colour
            ytck2  = 1, # secondary y-axis tick colour
            box    = 1, # plot frame box line colour
            mtext  = 2, # subheader text colour
            stats  = 1, # statistic summary colour
            kde.line        = 1, # KDE line colour
            kde.fill        = NULL, # KDE fill colour
            value.dot       = 1, # De value dot colour
            value.bar       = 1, # De value error bar colour
            value.rug       = 1, # De value rug colour
            boxplot.line    = 1, # boxplot line colour
            boxplot.fill    = NULL, # boxplot fill colour
            mean.point       = 1, # mean line colour
            sd.line          = 1, # sd bar colour
            background      = NA), # background colour

          dimension = list(
            figure.width    = "auto", # figure width in mm
            figure.height   = "auto", # figure height in mm
            margin = c(10, 10, 10, 10), # margin sizes in mm
            main.line       = 100, # line height in %
            xlab.line       = 100, # line height in %
            ylab1.line      = 100, # line height in %
            ylab2.line      = 100, # line height in %
            xtck.line       = 100, # line height in %
            ytck1.line      = 100, # line height in %
            ytck2.line      = 100, # line height in %
            xtcl            = 100, # tick length in %
            ytcl1           = 100, # tick length in %
            ytcl2           = 100, # tick length in %
            stats.line      = 100 # line height in %
          )
        )
      )
    },

    journal = {
      layout <- list(

        ## journal Abanico plot -----------------------------------------------
        abanico = list(
          font.type = list(
            main    = "",
            xlab1   = "",
            xlab2   = "",
            ylab    = "",
            zlab    = "",
            xtck1   = "",
            xtck2   = "",
            xtck3   = "",
            ytck    = "",
            ztck    = "",
            mtext   = "",
            summary = "", # optionally vector
            stats   = "", # optionally vector
            legend  = "" # optionally vector
          ),

          font.size = list(
            main    = 8,
            xlab1   = 7,
            xlab2   = 7,
            xlab3   = 7,
            ylab    = 7,
            zlab    = 7,
            xtck1   = 7,
            xtck2   = 7,
            xtck3   = 7,
            ytck    = 7,
            ztck    = 7,
            mtext   = 6,
            summary = 6, # optionally vector
            stats   = 6, # optionally vector
            legend  = 6 # optionally vector
          ),

          font.deco = list(
            main    = "bold",
            xlab1   = "normal",
            xlab2   = "normal",
            xlab3   = "normal",
            ylab    = "normal",
            zlab    = "normal",
            xtck1   = "normal",
            xtck2   = "normal",
            xtck3   = "normal",
            ytck    = "normal",
            ztck    = "normal",
            mtext   = "normal",
            summary = "normal", # optionally vector
            stats   = "normal", # optionally vector
            legend  = "normal" # optionally vector
          ),

          colour = list(
            main    = 1, # plot title colour
            xlab1   = 1, # left x-axis label colour
            xlab2   = 1, # right x-axis label colour
            xlab3   = 1, # right x-axis label colour
            ylab    = 1, # y-axis label colour
            zlab    = 1, # z-axis label colour
            xtck1   = 1, # left x-axis tick colour
            xtck2   = 1, # right x-axis tick colour
            xtck3   = 1, # right x-axis tick colour
            ytck    = 1, # y-axis tick colour
            ztck    = 1, # z-axis tick colour
            mtext   = 1, # subheader text colour
            summary = 1, # statistic summary colour
            stats   = 1, # value statistics colour
            legend  = 1, # legend colour
            centrality = 1, # Centrality line colour
            value.dot  = 1, # De value dot colour
            value.bar  = 1, # De value error bar colour
            value.rug  = 1, # De value rug colour
            poly.line  = NA, # polygon line colour
            poly.fill  = adjustcolor("grey75", alpha.f = 0.6), # polygon fill colour
            bar.line   = NA, # polygon line colour
            bar.fill   = "grey60", # bar fill colour
            kde.line   = 1,
            kde.fill   = NA,
            grid.major = "grey80",
            grid.minor = "none",
            border     = 1,
            background = NA),

          dimension = list(
            figure.width    = 100, # figure width in mm
            figure.height   = 100, # figure height in mm
            margin = c(10, 10, 10, 10), # margin sizes in mm
            main.line       = 70, # line height in %
            xlab1.line      = 30, # line height in %
            xlab2.line      = 65, # line height in %
            xlab3.line      = 30, # line height in %
            ylab.line       = 30, # line height in %
            zlab.line       = 40, # line height in %
            xtck1.line      = 50, # line height in %
            xtck2.line      = 50, # line height in %
            xtck3.line      = 50, # line height in %
            ytck.line       = 70, # line height in %
            ztck.line       = 70, # line height in %
            xtcl1           = 50, # tick length in %
            xtcl2           = 50, # tick length in %
            xtcl3           = 50, # tick length in %
            ytcl            = 50, # tick length in %
            ztcl            = 70, # tick length in %
            rugl            = 70, # rug length in %
            mtext           = 100, # line height in %
            summary.line    = 70, # line height in %
            pch             = 50  # point size in %
          )),

        ## journal KDE plot ---------------------------------------------------
        kde = list(
          font.type = list(
            main   = "",
            xlab   = "",
            ylab1  = "",
            ylab2  = "",
            xtck   = "",
            ytck1  = "",
            ytck2  = "",
            stats  = "", # optionally vector
            legend = "" # optionally vector
          ),

          font.size = list(
            main   = 8,
            xlab   = 7,
            ylab1  = 7,
            ylab2  = 7,
            xtck   = 7,
            ytck1  = 7,
            ytck2  = 7,
            stats  = 7,
            legend = 7
          ),

          font.deco = list(
            main   = "bold",
            xlab   = "normal",
            ylab1  = "normal",
            ylab2  = "normal",
            xtck   = "normal",
            ytck1  = "normal",
            ytck2  = "normal",
            stats  = "normal", # optionally vector
            legend = "normal" # optionally vector
          ),

          colour = list(
            main   = 1, # plot title colour
            xlab   = 1, # x-axis label colour
            ylab1  = 1, # primary y-axis label colour
            ylab2  = 1, # secondary y-axis label colour
            xtck   = 1, # x-axis tick colour
            ytck1  = 1, # primary y-axis tick colour
            ytck2  = 1, # secondary y-axis tick colour
            box    = 1, # plot frame box line colour
            mtext  = 1, # subheader text colour
            stats  = "#2062B3", # statistic summary colour
            kde.line        = "#2062B3", # KDE line colour
            kde.fill        = NULL, # KDE fill colour
            value.dot       = 1, # De value dot colour
            value.bar       = 1, # De value error bar colour
            value.rug       = 1, # De value rug colour
            boxplot.line    = 1, # boxplot line colour
            boxplot.fill    = NULL, # boxplot fill colour
            mean.line       = adjustcolor(col = 1,
                                          alpha.f = 0.4), # mean line colour
            sd.bar          = adjustcolor(col = 1,
                                          alpha.f = 0.4), # sd bar colour
            background      = NULL),

          dimension = list(
            figure.width    = 80, # figure width in mm
            figure.height   = 80, # figure height in mm
            margin = c(10, 10, 10, 10), # margin sizes in mm
            main.line       = 70, # line height in %
            xlab.line       = 30, # line height in %
            ylab1.line      = 40, # line height in %
            ylab2.line      = 30, # line height in %
            xtck.line       = 50, # line height in %
            ytck1.line      = 65, # line height in %
            ytck2.line      = 50, # line height in %
            xtcl            = 50, # tick length in %
            ytcl1           = 20, # tick length in %
            ytcl2           = 50, # tick length in %
            stats.line      = 70 # line height in %
          )
        )
      )
    })
  }

  ## return layout parameters
  return(layout)
}

test_that("input validation", {
  skip_on_cran()

  expect_type(get_Layout("empty"),
              "list")
  expect_type(get_Layout("journal"),
              "list")
  expect_warning(expect_type(res <- get_Layout(NA_character_),
                             "list"),
                 "Layout definition not supported, default layout is used")
  expect_warning(res <- get_Layout("error"),
                 "Layout definition not supported, default layout is used")
  expect_equal(res,
               get_Layout("default"))

  input <- list(element = "value")
  expect_equal(get_Layout(input),
               input)
})

#' @title Analyse SAR TL measurements
#'
#' @description The function performs a SAR TL analysis on a
#' [Luminescence::RLum.Analysis-class] object including growth curve fitting.
#'
#' @details This function performs a SAR TL analysis on a set of curves. The SAR
#' procedure in general is given by Murray and Wintle (2000). For the
#' calculation of the `Lx/Tx` value the function [Luminescence::calc_TLLxTxRatio] is
#' used.
#'
#' **Provided rejection criteria**
#'
#' `[recycling.ratio]`: calculated for every repeated regeneration dose point.
#'
#' `[recuperation.rate]`: recuperation rate calculated by
#' comparing the `Lx/Tx` values of the zero regeneration point with the `Ln/Tn`
#' value (the `Lx/Tx` ratio of the natural signal).  For methodological
#' background see Aitken and Smith (1988)
#'
#' @param object [Luminescence::RLum.Analysis-class] or a [list] of such objects (**required**) :
#' input object containing data for analysis
#'
#' @param object.background currently not used
#'
#' @param signal_integral [integer] (**required**):
#' vector of inputs (as defined by `integral_input`) for the signal integral.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given temperature range are selected.
#'
#' @param sequence.structure [vector] [character] (*with default*):
#' specifies the general sequence structure. Three steps are allowed
#' (`"PREHEAT"`, `"SIGNAL"`, `"BACKGROUND"`), in addition a
#' parameter `"EXCLUDE"`. This allows excluding TL curves which are not
#' relevant for the protocol analysis.  (**Note:** No TL are removed by default)
#'
#' @param rejection.criteria [list] (*with default*):
#' list containing rejection criteria in percentage for the calculation.
#'
#' @param dose.points [numeric] (*optional*):
#' option set dose points manually
#'
#' @param log [character] (*with default*):
#' a character string which contains `"x"` if the x-axis is to be logarithmic,
#' `"y"` if the y axis is to be logarithmic and `"xy"` or `"yx"` if both axes
#' are to be logarithmic. See
#' [plot.default]).
#'
#' @param ... further arguments that will be passed to the function [Luminescence::fit_DoseResponseCurve]
#'
#' @return
#' A plot (*optional*) and an [Luminescence::RLum.Results-class] object is
#' returned containing the following elements:
#'
#' \item{De.values}{[data.frame] containing De-values and further parameters}
#' \item{LnLxTnTx.values}{[data.frame] of all calculated `Lx/Tx` values including signal, background counts and the dose points.}
#' \item{rejection.criteria}{[data.frame] with values that might by used as rejection criteria. `NA` is produced if no R0 dose point exists.}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @note
#' **THIS IS A BETA VERSION**
#'
#' No TL curves will be removed from the input object without further warning.
#'
#' @section Function version: 0.3.4
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::calc_TLLxTxRatio], [Luminescence::fit_DoseResponseCurve], [Luminescence::RLum.Analysis-class],
#' [Luminescence::RLum.Results-class], [Luminescence::get_RLum]
#'
#' @references
#' Aitken, M.J. and Smith, B.W., 1988. Optical dating: recuperation
#' after bleaching.  Quaternary Science Reviews 7, 387-393.
#'
#' Murray, A.S. and Wintle, A.G., 2000. Luminescence dating of quartz using an
#' improved single-aliquot regenerative-dose protocol. Radiation Measurements
#' 32, 57-73.
#'
#' @keywords datagen plot
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##transform the values from the first position in a RLum.Analysis object
#' object <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos=3)
#'
#' ##perform analysis
#' analyse_SAR.TL(
#'  object = object,
#'  signal_integral = 210:220,
#'  fit.method = "EXP OR LIN",
#'  sequence.structure = c("SIGNAL", "BACKGROUND"))
#'
#' @export
analyse_SAR.TL <- function(
  object,
  object.background,
  signal_integral,
  integral_input = "channel",
  sequence.structure = c("PREHEAT", "SIGNAL", "BACKGROUND"),
  rejection.criteria = list(recycling.ratio = 10, recuperation.rate = 10),
  dose.points = NULL,
  log = "",
  ...
) {
  .set_function_name("analyse_SAR.TL")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if (any(grepl("signal.integral", ...names(), fixed = TRUE))) {
    .deprecated(old = "signal.integral",
                new = "signal_integral",
                since = "1.2.0")
    signal_integral <- list(...)$signal.integral
  }
  integral_input <- .validate_args(integral_input, c("channel", "temperature",
                                                     "measurement"))
  if (integral_input == "temperature") {
    .deprecated(old = "integral_input = \"temperature\"",
                new = "integral_input = \"measurement\"",
                since = "1.2.0")
    integral_input <- "measurement"
  }

  # Self-call -----------------------------------------------------------------------------------
  if(inherits(object, "list")){
    lapply(object, .validate_class, "RLum.Analysis",
           name = "All elements of 'object'")

    ##run sequence
    results <- lapply(object, function(o){
      analyse_SAR.TL(
        object = o,
        object.background = object.background,
        signal_integral = signal_integral,
        integral_input = integral_input,
        sequence.structure =  sequence.structure,
        rejection.criteria = rejection.criteria,
        dose.points = dose.points,
        log = log,
        ...
      )
    })

    ##combine results
    results <- merge_RLum(results)

    ##return
    return(results)
  }

  # CONFIG  -----------------------------------------------------------------

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, "RLum.Analysis")
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  signal_integral <- .validate_integral(signal_integral,
                                        int = integral_input == "channel")

  # Protocol Integrity Checks --------------------------------------------------

  ## silence notes raised by R CMD check
  id <- n.channels <- protocol.step <- NULL
  prev.idx <- criterion <- value <- threshold <- status <- NULL
  Name <- Dose <- LxTx <- Repeated <- NULL

  ## we must have a signal to analyse
  if (!"SIGNAL" %in% sequence.structure) {
    .throw_error("'sequence.structure' contains no 'SIGNAL' entry")
  }

  ##Remove non TL-curves from object by selecting TL curves
  object@records <- get_RLum(object, recordType = "TL",
                             recursive = FALSE)

  ##ANALYSE SEQUENCE OBJECT STRUCTURE

  ##set vector for sequence structure
  temp.protocol.step <- rep(sequence.structure,length(object@records))[1:length(object@records)]

  ## grep object structure
  temp.sequence.structure <- as.data.table(structure_RLum(object))

  ##set values for step
  temp.sequence.structure[, protocol.step := temp.protocol.step]

  ##remove TL curves which are excluded
  temp.sequence.structure <- temp.sequence.structure[protocol.step != "EXCLUDE"]

  ##check integrity; signal and bg range should be equal
  signal.channels <- temp.sequence.structure[protocol.step == "SIGNAL",
                                             unique(n.channels)]
  if (length(signal.channels) > 1) {
    .throw_error("Signal ranges differ (", .collapse(signal.channels, quote = FALSE),
                 "), check sequence structure")
  }

  ##check if the wanted curves are a multiple of the structure
  if (nrow(temp.sequence.structure) %% length(sequence.structure) != 0)
    .throw_error("Input TL curves are not a multiple of the sequence structure")

  # # Calculate LnLxTnTx values  --------------------------------------------------
  ##grep IDs for signal and background curves
  TL.signal.ID <- temp.sequence.structure[protocol.step == "SIGNAL", id]
  TL.background.ID <- temp.sequence.structure[protocol.step == "BACKGROUND", id]

  ## check that `dose.points` is compatible with our signals:
  ## as we expect each signal to have an Lx and a Tx components (see calls
  ## to calc_TLLxTxRatio()), `dose.points` must divide `length(TL.signal.ID)`
  ## in order for vector recycling to work when further down we do
  ## `LnLxTnTx$Dose <- dose.points`
  if (!is.null(dose.points) &&
      (length(TL.signal.ID) / 2) %% length(dose.points) != 0) {
      .throw_error("Length of 'dose.points' not compatible with number ",
                   "of signals")
  }

  ## convert integral limits from temperature to channel
  if (integral_input == "measurement") {
    x.values <- get_RLum(object, record.id = TL.signal.ID[1])[, 1]
    signal_integral <- .convert_to_channels(x.values, signal_integral,
                                            unit = "temperature")
  }

  ##calculate LxTx values using external function
  LnLxTnTx <- NULL
  for(i in seq(1,length(TL.signal.ID),by=2)){
    Lx.data.background <- Tx.data.background <- NULL
    if (length(TL.background.ID) > 0) {
      Lx.data.background <- get_RLum(object, record.id = TL.background.ID[i])
      Tx.data.background <- get_RLum(object, record.id = TL.background.ID[i + 1])
    }
    LxTxRatio <- calc_TLLxTxRatio(
        Lx.data.signal = get_RLum(object, record.id = TL.signal.ID[i]),
        Tx.data.signal = get_RLum(object, record.id = TL.signal.ID[i + 1]),
        Lx.data.background = Lx.data.background,
        Tx.data.background = Tx.data.background,
        signal_integral
    )
    temp.LnLxTnTx <- get_RLum(LxTxRatio)
    rm(LxTxRatio)

    ##grep dose
    temp.Dose <- object@records[[TL.signal.ID[i]]]@info$IRR_TIME %||% NA

    ## append row to the data.frame
    LnLxTnTx <- rbind(LnLxTnTx, cbind(Dose = temp.Dose, temp.LnLxTnTx))
  }

  ##set dose.points manually if argument was set
  if (!is.null(dose.points)) {
    temp.Dose <- dose.points
    LnLxTnTx$Dose <- dose.points
  }

  # Set regeneration points -------------------------------------------------
  #generate unique dose id - this are also the # for the generated points
  temp.DoseName <- data.frame(Name = paste0("R", seq_len(nrow(LnLxTnTx)) - 1),
                              Dose = LnLxTnTx[["Dose"]])

  ##set natural
  temp.DoseName[temp.DoseName[, "Name"] == "R0", "Name"] <- "Natural"

  ##set R0
  temp.DoseName[temp.DoseName[, "Name"] != "Natural" &
                temp.DoseName[, "Dose"] == 0, "Name"] <- "R0"

  ##find duplicated doses (including 0 dose - which means the Natural)
  temp.DoseName <- cbind(temp.DoseName,
                         Repeated = duplicated(temp.DoseName[, "Dose"]))

  ##correct value for R0 (it is not really repeated)
  temp.DoseName[temp.DoseName[,"Dose"]==0,"Repeated"]<-FALSE

  ##combine in the data frame
  LnLxTnTx <- cbind(temp.DoseName[, c("Name", "Repeated")],
                    LnLxTnTx)

  ## convert to data.table for more convenient column manipulation
  temp <- data.table(LnLxTnTx[, c("Name", "Dose", "Repeated", "LxTx")])

  # Calculate Recycling Ratio -----------------------------------------------
  ## we first create a dummy object to use in case there are no repeated doses,
  ## but replace it in the `if` block if there are any
  rej.thresh <- rejection.criteria$recycling.ratio / 100
  rej.thresh.text <- paste("\u00b1", rej.thresh) # \u00b1 is ±
  RecyclingRatio <- data.table(criterion = "recycling ratio",
                               value = NA,
                               threshold = rej.thresh.text,
                               status = NA_character_)
  if (any(temp$Repeated)) {

    ## find the index of the previous dose of each repeated dose
    temp[, prev.idx := match(Dose, Dose)]

    ## calculate the recycling ratio
    temp[, criterion := paste0(Name, "/", Name[prev.idx])]
    temp[, value := LxTx / LxTx[prev.idx]]

    ## set status according to the given rejection threshold
    temp[, threshold := rej.thresh.text]
    temp[, status := fifelse(abs(1 - value) > rej.thresh, "FAILED", "OK")]

    ## keep only the repeated doses
    RecyclingRatio <- temp[(Repeated),
                           list(criterion, value, threshold, status)]
  }

  # Calculate Recuperation Rate ---------------------------------------------
  ## we first create a dummy object to use in case there is no R0 dose,
  ## but replace it in the `if` block if there is one
  Recuperation <- data.table(criterion = "recuperation rate",
                             value = NA_real_,
                             threshold = rejection.criteria$recuperation.rate / 100,
                             status = NA_character_)
  if ("R0" %in% temp$Name) {
    Recuperation[, value := round(temp[Name == "R0", LxTx] /
                                  temp[Name == "Natural", LxTx], digits = 4)]
    Recuperation[, status := fifelse(value > threshold, "FAILED", "OK")]
  }

  ## join the two tables and convert back to data.frame
  RejectionCriteria <- as.data.frame(rbind(RecyclingRatio, Recuperation))
  rm(temp)

  ##============================================================================##
  ##PLOTTING
  ##============================================================================##

  # Plotting - Config -------------------------------------------------------
  ##grep plot parameter
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ##grep colours
  col <- get("col", pos = .LuminescenceEnv)

  ##set layout matrix
  graphics::layout(matrix(c(
    1, 1, 2, 2,
    1, 1, 2, 2,
    3, 3, 4, 4,
    3, 3, 4, 4,
    5, 5, 5, 5
  ), 5, 4, byrow = TRUE))

  par(oma = c(0, 0, 0, 0), mar = c(4, 4, 3, 3))

  ## 1 -> TL Lx
  ## 2 -> TL Tx
  ## 3 -> TL Lx Plateau
  ## 4 -> TL Tx Plateau
  ## 5 -> Legend

  ## recalculate signal_integral from channels to temperature
  signal.integral.temperature <- object@records[[TL.signal.ID[1]]]@data[signal_integral, 1]

  ## warning if number of curves exceeds colour values
  if(length(col)<length(TL.signal.ID/2)){
    .throw_message("Too many curves, only the first ", length(col),
                   " curves are plotted", error = FALSE)
  }

  # # Plotting TL LnLx Curves ----------------------------------------------------
  ##matrix with LnLx curves
  LnLx_matrix <- vapply(seq(1, length(TL.signal.ID), by = 2), function(x){
    if(length(TL.background.ID) != 0){
      object@records[[TL.signal.ID[x]]]@data[,2] -
        object@records[[TL.background.ID[x]]]@data[,2]
    }else{
      object@records[[TL.signal.ID[x]]]@data[,2]
    }
  }, numeric(nrow(object@records[[TL.signal.ID[1]]]@data)))

  ##add time axis
  LnLx_matrix <- cbind(object@records[[TL.signal.ID[1]]]@data[,1], LnLx_matrix)

  ##matrix with TnTx curves
  TnTx_matrix <- vapply(seq(2, length(TL.signal.ID), by = 2), function(x){
    if(length(TL.background.ID) != 0){
      object@records[[TL.signal.ID[x]]]@data[,2] -
        object@records[[TL.background.ID[x]]]@data[,2]
    }else{
      object@records[[TL.signal.ID[x]]]@data[,2]
    }
  }, numeric(nrow(object@records[[TL.signal.ID[1]]]@data)))

  ##add time axis
  TnTx_matrix <- cbind(object@records[[TL.signal.ID[1]]]@data[,1], TnTx_matrix)

  ##catch log-scale problem
  if (log != "" && (min(LnLx_matrix) <= 0 || min(TnTx_matrix) <= 0)) {
    .throw_warning("Non-positive values detected, log-scale disabled")
    log <- ""
  }

  #open plot area LnLx
  plot(NA,NA,
       xlab = "Temp. [\u00B0C]",
       ylab = "TL [a.u.]",
       xlim = range(LnLx_matrix[, 1]),
       ylim = range(LnLx_matrix[, -1]),
       main = expression(paste(L[n], ",", L[x], " curves")),
       log = log
  )

  ##plot curves
  for (i in 2:ncol(LnLx_matrix)) {
    lines(x = LnLx_matrix[, 1], y = LnLx_matrix[, i],
          col = col[i - 1])
  }

  ##mark integration limits
  abline(v = range(signal.integral.temperature), lty = 2, col = "gray")

  #open plot area TnTx
  plot(NA,NA,
    xlab = "Temp. [\u00B0C]",
    ylab = paste0("TL [a.u.]"),
    xlim = range(TnTx_matrix[, 1]),
    ylim = range(TnTx_matrix[, -1]),
    main = expression(paste(T[n], ",", T[x], " curves")),
    log = log
  )

  ##plot curves
  for (i in 2:ncol(TnTx_matrix)) {
    lines(x = TnTx_matrix[, 1], y = TnTx_matrix[, i],
          col = col[i - 1])
  }

  ##mark integration limits
  abline(v = range(signal.integral.temperature), lty = 2, col = "gray")

  ##clean
  rm(LnLx_matrix, TnTx_matrix)

  # Plotting Plateau Test LnLx -------------------------------------------------
  if(length(TL.background.ID) != 0){
    NTL.net.LnLx <-
      data.frame(object@records[[TL.signal.ID[1]]]@data[, 1],
                 object@records[[TL.signal.ID[1]]]@data[, 2] -
                   object@records[[TL.background.ID[1]]]@data[, 2])

    Reg1.net.LnLx <-
      data.frame(object@records[[TL.signal.ID[3]]]@data[, 1],
                 object@records[[TL.signal.ID[3]]]@data[, 2] -
                   object@records[[TL.background.ID[3]]]@data[, 2])


    TL.Plateau.LnLx <-
      data.frame(NTL.net.LnLx[, 1], Reg1.net.LnLx[, 2] / NTL.net.LnLx[, 2])

    ##Plot Plateau Test
    plot(
      NA,
      NA,
      xlab = "Temp. [\u00B0C]",
      ylab = "TL [a.u.]",
      xlim = c(
        min(signal.integral.temperature) * 0.9,
        max(signal.integral.temperature) * 1.1
      ),
      ylim = c(0, max(NTL.net.LnLx[, 2])),
      main = expression(paste("Plateau test ", L[n], ",", L[x], " curves"))
    )

    ##plot single curves
    lines(NTL.net.LnLx, col = col[1])
    lines(Reg1.net.LnLx, col = col[2])

    ##plot
    TL.tmp <- TL.Plateau.LnLx[signal_integral, 2]
    ylim.max <- quantile(TL.tmp[!is.infinite(TL.tmp)],
                         probs = 0.90, na.rm = TRUE)
    par(new = TRUE)
    plot(
      TL.Plateau.LnLx,
      axes = FALSE,
      xlab = "",
      ylab = "",
      ylim = c(0, ylim.max + 3),
      col = "darkgreen"
    )
    axis(4)


    # Plotting Plateau Test TnTx -------------------------------------------------
    ##get NTL signal
    NTL.net.TnTx <- matrix(
      data = c(object@records[[TL.signal.ID[2]]]@data[, 1],
                 object@records[[TL.signal.ID[2]]]@data[, 2] -
                   object@records[[TL.background.ID[2]]]@data[, 2]),
      ncol = 2)

    ##get signal from the first regeneration point
    Reg1.net.TnTx <- matrix(
      data = c(object@records[[TL.signal.ID[4]]]@data[, 1],
                 object@records[[TL.signal.ID[4]]]@data[, 2] -
                   object@records[[TL.background.ID[4]]]@data[, 2]),
      ncol = 2)


    ##combine values
    TL.Plateau.TnTx <- matrix(
      data = c(NTL.net.TnTx[, 1],
               Reg1.net.TnTx[, 2] / NTL.net.TnTx[, 2]),
      ncol = 2)

    ##Plot Plateau Test
    plot(
      NA,
      NA,
      xlab = "Temp. [\u00B0C]",
      ylab = "TL [a.u.]",
      xlim = c(
        min(signal.integral.temperature) * 0.9,
        max(signal.integral.temperature) * 1.1
      ),
      ylim = c(0, max(NTL.net.TnTx[, 2])),
      main = expression(paste("plateau Test ", T[n], ",", T[x], " curves"))
    )

    ##plot single curves
    lines(NTL.net.TnTx, col = col[1])
    lines(Reg1.net.TnTx, col = col[2])

    ##plot
    par(new = TRUE)
    plot(
      TL.Plateau.TnTx,
      axes = FALSE,
      xlab = "",
      ylab = "",
      ylim = c(0,
               quantile(
                 TL.Plateau.TnTx[signal_integral, 2],
                 probs = 0.90, na.rm = TRUE
               ) + 3),
      col = "darkgreen"
    )
    axis(4)

    # Plotting Legend ----------------------------------------
    plot(
      1:(length(TL.signal.ID) / 2),
      rep(8, length(TL.signal.ID) / 2),
      type = "p",
      axes = FALSE,
      xlab = "",
      ylab = "",
      pch = 15,
      col = col[1:length(TL.signal.ID)],
      cex = 2,
      ylim = c(0, 10)
    )

    ##add text
    text(1:(length(TL.signal.ID) / 2),
         rep(4, length(TL.signal.ID) / 2),
         paste0(LnLxTnTx$Name, "\n(", LnLxTnTx$Dose, ")"))

    ##add line
    abline(h = 10, lwd = 0.5)

    ##set failed text and mark De as failed
    if (any(grepl("FAILED", RejectionCriteria$status, fixed = TRUE))) {
      mtext("[FAILED]", col = "red")
    }
  }

  # Plotting  GC  ----------------------------------------

  ##create data.frame
  temp.sample <- data.frame(
    Dose = LnLxTnTx$Dose,
    LxTx = LnLxTnTx$LxTx,
    LxTx.Error = LnLxTnTx$LxTx.Error,
    TnTx = LnLxTnTx$TnTx
  )

  ##set NA values to 0
  temp.sample[is.na(temp.sample$LxTx.Error),"LxTx.Error"] <- 0

  ##run curve fitting
  temp.GC <- try(fit_DoseResponseCurve(
    object = temp.sample,
    ...
  ), outFile = stdout()) # redirect error messages so they can be silenced

  ## fit_DoseResponseCurve() can fail in two ways:
  ## 1. either with a hard error, in which case there's nothing much we
  ##    can do and stop early by returning NULL
  if(inherits(temp.GC, "try-error")){
    return(NULL)
  }

  ## 2. or with a soft error by returning NULL, in which case we set
  ##    temp.GC to NA and continue (this can be done after the call to
  ##    get_RLum(), as it deals well with NULLs)
  temp.GC <- get_RLum(temp.GC)[, c("De", "De.Error")] %||% NA

  ##add rejection status
  RC.status <- if (any(grepl("FAILED", RejectionCriteria$status, fixed = TRUE)))
                 "FAILED" else "OK"
  temp.GC <- data.frame(temp.GC, RC.Status = RC.status)

  # Return Values -----------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      data = temp.GC,
      LnLxTnTx.table = LnLxTnTx,
      rejection.criteria = RejectionCriteria
    ),
    info = list(info = sys.call())
  )
}

#' @title Verify single grain data sets and check for invalid grains, i.e.
#' zero-light level grains
#'
#' @description
#' This function tries to identify automatically zero-light level curves (grains)
#' from single grain data measurements.
#'
#' @details
#'
#' **How does the method work?**
#'
#' The function compares the expected values (\eqn{E(X)}) and the variance
#' (\eqn{Var(X)}) of the count values for each curve. Assuming that the
#' background roughly follows a Poisson distribution, the absolute difference
#' of both values should be zero or at least around zero as
#'
#' \deqn{E(x) = Var(x) = \lambda}
#'
#' Thus the function checks for:
#'
#' \deqn{abs(E(x) - Var(x)) >= \Theta}
#'
#' With \eqn{\Theta} an arbitrary, user defined, threshold. Values above the
#' threshold indicate curves comprising a signal.
#'
#' Note: the absolute difference of \eqn{E(X)} and \eqn{Var(x)} instead of the
#' ratio was chosen as both terms can become 0 which would result in 0 or `Inf`,
#' if the ratio is calculated.
#'
#' @param object [Luminescence::Risoe.BINfileData-class] or [Luminescence::RLum.Analysis-class] (**required**):
#' input object. The function also accepts a list with objects of allowed type.
#'
#' @param threshold [numeric] (*with default*):
#' numeric threshold value for the allowed difference between the `mean` and
#' the `var` of the count values (see details).
#'
#' @param use_fft [logical] (*with default*): applies an additional approach based on [stats::fft].
#' The threshold is fixed and cannot be changed.
#'
#' @param cleanup [logical] (*with default*):
#' if set to `TRUE`, curves/aliquots identified as zero light level curves/aliquots are
#' automatically removed. Output is an object as same type as the input, i.e.
#' either [Luminescence::Risoe.BINfileData-class] or [Luminescence::RLum.Analysis-class]
#'
#' @param cleanup_level [character] (*with default*):
#' selects the level for the clean-up of the input data sets.
#' Two options are allowed: `"curve"` or `"aliquot"`:
#'
#' - If  `"curve"` is selected, every single curve marked as `invalid` is removed.
#' - If `"aliquot"` is selected, curves of one aliquot (grain or disc) can be
#' marked as invalid, but will not be removed. An aliquot will be only removed
#' if all curves of this aliquot are marked as invalid.
#'
#' @param verbose [logical] (*with default*):
#' enable/disables output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... further parameters to control the plot output; if selected.
#' Supported arguments `main`, `ylim`
#'
#' @return
#' The function returns
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** `@data`
#'
#' \tabular{lll}{
#' **Element** \tab **Type** \tab **Description**\cr
#'  `$unique_pairs` \tab `data.frame` \tab the unique position and grain pairs \cr
#'  `$selection_id` \tab `numeric` \tab the selection as record ID \cr
#'  `$selection_full` \tab `data.frame` \tab implemented models used in the baSAR-model core \cr
#' }
#'
#' **slot:** `@info`
#'
#' The original function call
#'
#' **Output variation**
#'
#' For `cleanup = TRUE` the same object as the input is returned, but cleaned up
#' (invalid curves were removed). This means: Either a [Luminescence::Risoe.BINfileData-class]
#' or an [Luminescence::RLum.Analysis-class] object is returned in such cases.
#' A [Luminescence::Risoe.BINfileData-class] object can be exported to a BINX-file by
#' using the function [Luminescence::write_R2BIN].
#'
#' @note
#' This function can work with [Luminescence::Risoe.BINfileData-class] objects or
#' [Luminescence::RLum.Analysis-class] objects (or a list of it). However, the function is
#' highly optimised for [Luminescence::Risoe.BINfileData-class] objects as it make sense to
#' remove identify invalid grains before the conversion to an
#' [Luminescence::RLum.Analysis-class] object.
#'
#' The function checking for invalid curves works rather robust and it is likely
#' that Reg0 curves within a SAR cycle are removed as well. Therefore it is
#' strongly recommended to use the argument `cleanup = TRUE` carefully if
#' the cleanup works only on curves.
#'
#' @section Function version: 0.2.7
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::Risoe.BINfileData-class], [Luminescence::RLum.Analysis-class],
#' [Luminescence::write_R2BIN], [Luminescence::read_BIN2R]
#'
#' @keywords manip datagen
#'
#' @examples
#'
#' ##01 - basic example I
#' ##just show how to apply the function
#' data(ExampleData.XSYG, envir = environment())
#'
#' ##verify and get data.frame out of it
#' verify_SingleGrainData(OSL.SARMeasurement$Sequence.Object)$selection_full
#'
#' ##02 - basic example II
#' data(ExampleData.BINfileData, envir = environment())
#' id <- verify_SingleGrainData(object = CWOSL.SAR.Data,
#' cleanup_level = "aliquot")$selection_id
#'
#' \dontrun{
#' ##03 - advanced example I
#' ##importing and exporting a BIN-file
#'
#' ##select and import file
#' file <- file.choose()
#' object <- read_BIN2R(file)
#'
#' ##remove invalid aliquots(!)
#' object <- verify_SingleGrainData(object, cleanup = TRUE)
#'
#' ##export to new BIN-file
#' write_R2BIN(object, paste0(dirname(file),"/", basename(file), "_CLEANED.BIN"))
#' }
#'
#' @export
verify_SingleGrainData <- function(
    object,
    threshold = 10,
    use_fft = FALSE,
    cleanup = FALSE,
    cleanup_level = c("aliquot", "curve"),
    verbose = TRUE,
    plot = FALSE,
    ...
) {
  .set_function_name("verify_SingleGrainData")
  on.exit(.unset_function_name(), add = TRUE)

  ##three types of input are allowed:
  ##(1) RisoeBINfileData
  ##(2) RLum.Analysis
  ##(3) List of RLum.Analysis

  # Self Call -----------------------------------------------------------------------------------
  if (inherits(object, "list")) {
    if (length(object) == 0)
      return(set_RLum(class = if (isTRUE(cleanup)) "RLum.Analysis" else "RLum.Results"))

    results <- .warningCatcher(lapply(seq_along(object), function(x) {
      verify_SingleGrainData(
        object = object[[x]],
        threshold = threshold,
        use_fft = use_fft,
        cleanup = cleanup,
        cleanup_level = cleanup_level,
        verbose = verbose,
        plot = plot,
        main = paste0("Record #",x)
      )
    }))

    ##account for cleanup
    if (!cleanup)
      return(merge_RLum(results))

    results <- .rm_NULL_elements(.rm_nonRLum(results))
    if (length(results) == 0)
      return(NULL)

    return(results)
  }

  ## ------------------------------------------------------------------------
  ## input validation

  .validate_class(object, c("Risoe.BINfileData", "RLum.Analysis"),
                  extra = "a 'list' of such objects")
  .validate_positive_scalar(threshold)
  .validate_logical_scalar(use_fft)
  .validate_logical_scalar(cleanup)
  cleanup_level <- .validate_args(cleanup_level, c("aliquot", "curve"))
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

  ## implement Fourier Transform for Frequency Analysis
  ## inspired by ChatGPT (OpenAI, 2024)
  .calc_FFT_selection <- function(l, tmp_threshold = threshold/2) {
    vapply(l, function(x){
      x <- x[x>0]
      tmp_power_spectrum <- Mod(stats::fft(x)^2)
      tmp_mean_power <- mean(tmp_power_spectrum[-1])
      tmp_dominant_power <- max(tmp_power_spectrum[2:(length(tmp_power_spectrum)/2)])
      tmp_dominant_power > tmp_threshold * tmp_mean_power
    }, logical(1))
  }

  ##++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ##RisoeBINfileData
  if(inherits(object, "Risoe.BINfileData")){
    ##run test on DATA slot
    ##MEAN + SD
    temp.results_matrix <- t(vapply(X = object@DATA, FUN = function(x){
      c(mean(x), stats::var(x))
    }, numeric(2)))

    ##DIFF
    temp.results_matrix_RATIO <- temp.results_matrix[,2]/temp.results_matrix[,1]

    ##SEL
    temp.results_matrix_VALID <-
      temp.results_matrix_RATIO > threshold &
      if (use_fft) .calc_FFT_selection(object@DATA) else TRUE

    ##combine everything to in a data.frame
    selection <- data.frame(
      POSITION = object@METADATA$POSITION,
      GRAIN = object@METADATA$GRAIN,
      MEAN = temp.results_matrix[, 1],
      VAR = temp.results_matrix[, 2],
      RATIO = temp.results_matrix_RATIO,
      THRESHOLD = rep_len(threshold, length(object@DATA)),
      VALID = temp.results_matrix_VALID
    )

    ##get unique pairs for POSITION and GRAIN for VALID == TRUE
    unique_pairs <- unique(
      selection[selection[["VALID"]], c("POSITION", "GRAIN")])

    if(cleanup_level == "aliquot"){
      selection_id <- sort(unlist(lapply(1:nrow(unique_pairs), function(x) {
        which(
          .subset2(selection, 1) == .subset2(unique_pairs, 1)[x] &
            .subset2(selection, 2) == .subset2(unique_pairs, 2)[x]
        )
      })))

    }else{
      ##reduce data to TRUE selection
      selection_id <- which(selection[["VALID"]])
    }

    ##select output on the chosen input
    if(cleanup){
      ##selected wanted elements
      object@DATA <- object@DATA[selection_id]
      if(length(object@DATA) > 0) {
        object@METADATA <- object@METADATA[selection_id,]
        object@METADATA$ID <- 1:length(object@DATA)

        ##print message
        selection_id <- .collapse(selection_id, quote = FALSE)
        if(verbose){
          cat("\n[verify_SingleGrainData()] Risoe.BINfileData object reduced to records:\n",
              selection_id)
          cat("\n\n[verify_SingleGrainData()] Risoe.BINfileData object record index reset.\n")
        }
      } else {
        object <- NULL
      }

      ##return
      return_object <- object

    }else{
      return_object <- set_RLum(
        class = "RLum.Results",
        data = list(
          unique_pairs =  unique_pairs,
          selection_id = selection_id,
          selection_full = selection),
        info = list(call = sys.call())
      )
    }


    ##++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ##RLum.Analysis and list with RLum.Analysis objects
    ## ... and yes it make sense not to mix that up with the code above
  } else if (inherits(object,"RLum.Analysis")) {
    ## check for empty records
    if(length(object@records) == 0) {
      .throw_warning("Cannot process empty RLum.Analysis objects. NULL returned!")
      return(NULL)

    }

    ##first extract all count values from all curves
    object_list <- lapply(object@records, function(x){
      ##yes, would work differently, but it is faster
      x@data[,2]
    })

    ##MEAN + SD
    temp.results_matrix <- lapply(X = object_list, FUN = function(x){
      c(mean(x), stats::var(x))
    })

    temp.results_matrix <- do.call(rbind,  temp.results_matrix)

    ##DIFF
    temp.results_matrix_RATIO <- temp.results_matrix[,2]/temp.results_matrix[,1]

    ##SEL
    temp.results_matrix_VALID <- temp.results_matrix_RATIO > threshold &
      if (use_fft) .calc_FFT_selection(object_list) else TRUE

    ##get structure for the RLum.Analysis object
    temp_structure <- structure_RLum(object, fullExtent = TRUE)

    ##now we have two cases, depending on where measurement is coming from
    if (.check_originator(object, "Risoe.BINfileData2RLum.Analysis")) {
      ##combine everything to in a data.frame
      selection <- data.frame(
        POSITION = temp_structure$info.POSITION,
        GRAIN = temp_structure$info.GRAIN,
        MEAN = temp.results_matrix[, 1],
        VAR = temp.results_matrix[, 2],
        RATIO = temp.results_matrix_RATIO,
        THRESHOLD = rep_len(threshold, length(object_list)),
        VALID = temp.results_matrix_VALID
      )
      sel.cols <- c("POSITION", "GRAIN")

    } else if (.check_originator(object, "read_XSYG2R")) {
      ##combine everything to in a data.frame
      selection <- data.frame(
        POSITION = if(any(grepl(pattern = "position", names(temp_structure)))){
          temp_structure$info.position}else{
            NA
          },
        GRAIN = NA,
        MEAN = temp.results_matrix[, 1],
        VAR = temp.results_matrix[, 2],
        RATIO = temp.results_matrix_RATIO,
        THRESHOLD = rep_len(threshold, length(object_list)),
        VALID = temp.results_matrix_VALID
      )
      sel.cols <- "POSITION"

    } else{
      .throw_error("Object originator '", object@originator, "' not supported")
    }

    ## get unique POSITION and GRAIN pairs where VALID == TRUE
    unique_pairs <- unique(selection[selection[["VALID"]], sel.cols, drop = FALSE])

    ##set up cleanup
    if(cleanup_level == "aliquot") {
      if (.check_originator(object, "read_XSYG2R")) {

        if (!all(is.na(unique_pairs))) {
          selection_id <-
            sort(unlist(lapply(1:nrow(unique_pairs), function(x) {
              which(.subset2(selection, 1) == .subset2(unique_pairs, 1)[x])
            })))

        }else{
          selection_id <- NA
        }

      } else if (.check_originator(object, "Risoe.BINfileData2RLum.Analysis")) {
        selection_id <-
          sort(unlist(lapply(1:nrow(unique_pairs), function(x) {
            which(
              .subset2(selection, 1) == .subset2(unique_pairs, 1)[x] &
                .subset2(selection, 2) == .subset2(unique_pairs, 2)[x]
            )
          })))
      }

      ##make sure that we do not break subsequent code
      if(length(selection_id) == 0) selection_id <- NA

    } else{
      ##reduce data to TRUE selection
      selection_id <- which(selection[["VALID"]])
    }

    ##return value
    ##select output on the chosen input
    if (cleanup && !anyNA(selection_id)) {
      ##print message
      if(verbose && cleanup_level == "curve"){
        selection_id_text <- .collapse(selection_id, quote = FALSE)
        if(selection_id_text == "")
          selection_id_text <- "<none>"

        .throw_message("RLum.Analysis object reduced to records: ",
                       selection_id_text, error = FALSE)
      }

      ##selected wanted elements
      if (length(selection_id) == 0) {
        object <- set_RLum(
          class = "RLum.Analysis",
          originator = object@originator,
          protocol = object@protocol,
          records = list(),
          info = list(
            unique_pairs = unique_pairs,
            selection_id = selection_id,
            selection_full = selection)
        )

      } else{
        object <- set_RLum(
          class = "RLum.Analysis",
          records = get_RLum(object, record.id = selection_id, drop = FALSE),
          info = list(
            unique_pairs = unique_pairs,
            selection_id = selection_id,
            selection_full = selection)
        )
      }

      ##return
      return_object <- object

    }else{
      if (anyNA(selection_id))
        .throw_warning("'selection_id' is NA, everything tagged for removal")

      return_object <- set_RLum(
        class = "RLum.Results",
        data = list(
          unique_pairs = unique_pairs,
          selection_id = selection_id,
          selection_full = selection),
        info = list(call = sys.call())
      )

      ## cleanup means cleanup
      if (cleanup)
        return_object <- NULL
    }
  }

  # Plot ----------------------------------------------------------------------------------------
  if(plot){
    ##set plot settings
    plot_settings <-
      modifyList(x = list(
        main = "Record selection",
        ylim = range(c(selection[["RATIO"]], threshold * 1.1))
      ),
      val = list(...))

    ##plot area
    plot(
      NA,
      NA,
      xlim = c(1,nrow(selection)),
      ylim = plot_settings$ylim,
      log = "y",
      xlab = "Record index",
      ylab = "Calculated ratio [a.u.]",
      main = plot_settings$main
    )

    ##plot points above the threshold
    points(x = which(selection[["VALID"]]),
           y = selection[["RATIO"]][selection[["VALID"]]], pch = 20, col = "darkgreen")
    points(x = which(!selection[["VALID"]]),
           y = selection[["RATIO"]][!selection[["VALID"]]], pch = 20, col = rgb(0,0,0,0.5))

    abline(h = threshold, col = "red", lty = 1, lwd = 2)

    mtext(
      side = 3,
      text = paste0(
        "(total: ", nrow(selection),
        " | valid: ", length(which(selection[["VALID"]])),
        " | invalid: ", length(which(!selection[["VALID"]])), ")"),
      cex = 0.9 * par()$cex)
  }

  # Return --------------------------------------------------------------------------------------
  if(is.null(return_object))
    .throw_warning("Verification and cleanup removed all records, NULL returned")

  return(return_object)
}

#' Tune data for experimental purpose
#'
#' The error can be reduced and sample size increased for specific purpose.
#'
#' @param data [data.frame] (**required**):
#' input values, structure: data (`values[,1]`) and data error (`values [,2]`)
#' are required
#'
#' @param decrease.error [numeric]:
#' factor by which the error is decreased, ranges between 0 and 1.
#'
#' @param increase.data [numeric]:
#' factor by which the error is decreased, ranges between 0 and `Inf`.
#'
#' @return Returns a [data.frame] with tuned values.
#'
#' @note
#' You should not use this function to improve your poor data set!
#'
#' @section Function version: 0.5.0
#'
#' @author
#' Michael Dietze, GFZ Potsdam (Germany)
#'
#' @keywords manip
#'
#' @examples
#'
#' ## load example data set
#' data(ExampleData.DeValues, envir = environment())
#' x <- ExampleData.DeValues$CA1
#'
#' ## plot original data
#' plot_AbanicoPlot(data = x,
#'                  summary = c("n", "mean"))
#'
#' ## decrease error by 10 %
#' plot_AbanicoPlot(data = tune_Data(x, decrease.error = 0.1),
#'                  summary = c("n", "mean"))
#'
#' ## increase sample size by 200 %
#' #plot_AbanicoPlot(data = tune_Data(x, increase.data = 2),
#' #                summary = c("n", "mean"))
#'
#' @export
tune_Data <- function(
  data,
  decrease.error = 0,
  increase.data = 0
) {
  .set_function_name("tune_Data")
  on.exit(.unset_function_name(), add = TRUE)

  if (!missing(decrease.error)) {
    error.rel <- data[,2] / data[,1]
    data[,2] <- error.rel * (1 - decrease.error) * data[,1]
  }

  if (!missing(increase.data)) {
    n <- round(x = increase.data * 100,
               digits = 0)

    i.new <- sample(x = 1:nrow(data),
                    size = n,
                    replace = TRUE)

    x.new <- rnorm(n = n,
                   mean = data[i.new, 1],
                   sd = data[i.new, 2])

    e.new <- rnorm(n = n,
                   mean = data[i.new, 2],
                   sd = data[i.new, 2] * 0.05)

    x.merge <- c(data[,1], x.new)
    e.merge <- c(data[,2], e.new)

    ord <- order(x.merge)
    e.merge <- e.merge[ord]
    x.merge <- x.merge[ord]

    data.out <- data.frame(x.merge, e.merge)

    names(data.out) <- names(data)

    data <- data.out
  }

  info <- Sys.info()
  user <- info[length(info)]
  os <- info[1]

  .throw_warning("Dear ",
                user,
                ", these activities on your ",
                os,
                " machine have been tracked and will be submitted to ",
                "the R.Lum data base. Cheating does not pay off! [",
                Sys.time(),
                "]",
                sep = "")

  return(data)
}

#' @title Create De(t) plot
#'
#' @description Plots the equivalent dose (\eqn{D_e}) in dependency of the chosen signal integral
#' (cf. Bailey et al., 2003). The function is simply passing several arguments
#' to the function [plot] and the used analysis functions and runs it in a loop.
#' Example: `legend.pos` for legend position, `legend` for legend text.
#'
#' @details
#'
#' **method**
#'
#' The original method presented by Bailey et al., 2003 shifted the signal integrals and slightly
#' extended them accounting for changes in the counting statistics. Example: `c(1:3, 3:5, 5:7)`.
#' However, here also another method is provided allowing to expand the signal integral by
#' consecutively expanding the integral by its chosen length. Example: `c(1:3, 1:5, 1:7)`
#'
#' Note that in both cases the integral limits are overlap. The finally applied limits are part
#' of the function output.
#'
#' **analyse_function.control**
#'
#' The argument `analyse_function.control` currently supports the following arguments:
#' `sequence.structure`, `dose.points`, `mtext.outer`, `fit.method`,
#' `fit.force_through_origin`, `plot`, `plot_singlePanels`
#'
#' @param object [Luminescence::RLum.Analysis-class] (**required**): input object containing data for analysis
#' Can be provided as a [list] of such objects.
#'
#' @param signal_integral [integer] (**required**):
#' vector of channels for the signal integral.
#'
#' @param background_integral [integer] (**required**):
#' vector of channels for the background integral.
#'
#' @param method [character] (*with default*):
#' method applied for constructing the De(t) plot.
#' - `shift` (*the default*): the chosen signal integral is shifted the shine down curve,
#' - `expansion`: the chosen signal integral is expanded each time by its length
#'
#' @param signal_integral.seq [numeric] (*optional*):
#' argument to provide an own signal integral sequence for constructing the De(t) plot
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param analyse_function [character] (*with default*):
#' name of the analyse function to be called. Supported functions are:
#' [Luminescence::analyse_SAR.CWOSL], [Luminescence::analyse_pIRIRSequence]
#'
#' @param analyse_function.control [list] (*optional*):
#' selected arguments to be passed to the supported analyse functions
#' ([Luminescence::analyse_SAR.CWOSL], [Luminescence::analyse_pIRIRSequence]). The arguments must be provided
#' as named [list], e.g., `list(dose.points = c(0,10,20,30,0,10)` will set the
#' regeneration dose points.
#'
#' @param n.channels [integer] (*optional*):
#' number of channels used for the De(t) plot. If nothing is provided all
#' De-values are calculated and plotted until the start of the background
#' integral.
#'
#' @param show_ShineDownCurve [logical] (*with default*):
#' enable/disable shine down curve in the plot output.
#'
#' @param respect_RC.Status [logical] (*with default*):
#' remove De values with 'FAILED' RC.Status from the plot (cf. [Luminescence::analyse_SAR.CWOSL]
#' and [Luminescence::analyse_pIRIRSequence]).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param multicore [logical] (*with default*) : enable/disable multi core
#' calculation if `object` is a [list] of [Luminescence::RLum.Analysis-class] objects. Can be an
#' [integer] specifying the number of cores
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#' Disabling the plot is useful in cases where the output need to be processed
#' differently.
#'
#' @param ... further arguments and graphical parameters passed to
#' [plot.default], [Luminescence::analyse_SAR.CWOSL] and [Luminescence::analyse_pIRIRSequence] (see details for further information).
#' Plot control parameters are: `ylim`, `xlim`, `ylab`, `xlab`, `main`, `pch`, `mtext`, `cex`, `legend`,
#' `legend.text`, `legend.pos`
#'
#' @return
#' A plot and an [Luminescence::RLum.Results-class] object with the produced \eqn{D_e} values
#'
#' `@data`:
#'
#' \tabular{lll}{
#' **Object** \tab **Type** \tab **Description**\cr
#' `De.values` \tab `data.frame` \tab table with De values \cr
#' `signal_integral.seq` \tab `numeric` \tab integral sequence used for the calculation
#' }
#'
#' `@info`:
#'
#' \tabular{lll}{
#' **Object** \tab **Type** \tab **Description**\cr
#' call \tab `call` \tab the original function call
#' }
#'
#' @note
#' The entire analysis is based on the used analysis functions, namely
#' [Luminescence::analyse_SAR.CWOSL] and [Luminescence::analyse_pIRIRSequence]. However, the integrity
#' checks of this function are not that thoughtful as in these functions itself.
#' It means, that every sequence should be checked carefully before running long
#' calculations using several hundreds of channels.
#'
#' @section Function version: 0.1.10
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @references
#' Bailey, R.M., Singarayer, J.S., Ward, S., Stokes, S., 2003. Identification of partial resetting
#' using De as a function of illumination time. Radiation Measurements 37, 511-518.
#' doi:10.1016/S1350-4487(03)00063-5
#'
#' @seealso [plot], [Luminescence::analyse_SAR.CWOSL], [Luminescence::analyse_pIRIRSequence]
#'
#' @examples
#'
#' \dontrun{
#' ##load data
#' ##ExampleData.BINfileData contains two BINfileData objects
#' ##CWOSL.SAR.Data and TL.SAR.Data
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##transform the values from the first position in a RLum.Analysis object
#' object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#'
#' plot_DetPlot(
#'   object,
#'   signal_integral = 1:3,
#'   background_integral = 900:1000,
#'   n.channels = 5)
#' }
#'
#' @export
plot_DetPlot <- function(
  object,
  signal_integral,
  background_integral,
  method = "shift",
  signal_integral.seq = NULL,
  integral_input = c("channel", "measurement"),
  analyse_function = c("analyse_SAR.CWOSL", "analyse_pIRIRSequence"),
  analyse_function.control = list(),
  n.channels = NULL,
  show_ShineDownCurve = TRUE,
  respect_RC.Status = FALSE,
  multicore = TRUE,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("plot_DetPlot")
  on.exit(.unset_function_name(), add = TRUE)

# SELF CALL ---------------------------------------------------------------
  if(inherits(object, "list")) {
   ## remove all RLum.Analysis objects
   object <- .rm_nonRLum(x = object, class = "RLum.Analysis")

   ## get parameters to be passed on
   f_def <- sys.call(sys.parent(n = -1))
   args_default <- as.list(f_def)[-(1:2)]

    ## detect cores
    .validate_class(multicore, c("logical", "numeric", "integer"), length = 1)
    cores <- if (isTRUE(multicore)) {
               min(parallel::detectCores(), length(object))
             } else {
               max(multicore, 1)
             }

    ## function that calls plot_DetPlot() on each element of the list
    nested.fun <- function(x) {
      do.call(plot_DetPlot, c(list(object = x), args_default))
    }

    if (cores > 1) {
      ## run in parallel
      cl <- parallel::makeCluster(cores)
      on.exit(parallel::stopCluster(cl), add = TRUE)

      if (verbose)
        .throw_message("Running multicore session using ", cores,
                       " cores ...", error = FALSE)
      res_list <- parallel::parLapply(cl = cl, X = object, fun = nested.fun)
    } else {
      ## run in serial
      res_list <- lapply(object, nested.fun)
    }

    return(merge_RLum(res_list))
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis")
  .validate_not_empty(object)
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))
  method <- .validate_args(method, c("shift", "expansion"))
  analyse_function <- .validate_args(analyse_function,
                                     c("analyse_SAR.CWOSL", "analyse_pIRIRSequence"))

  ## deprecated arguments
  if ("plot.single" %in% ...names()) {
    plot_singlePanels <- list(...)$plot.single
    .deprecated("plot.single", "plot_singlePanels", since = "1.0.0")
  }
  if (any(grepl("[signal|background]\\.integral\\.[min|max]", ...names()))) {
    extraArgs <- list(...)
    .deprecated(old = c("signal.integral.min", "signal.integral.max",
                        "background.integral.min", "background.integral.max"),
                new = c("signal_integral", "background_integral"),
                since = "1.2.0")

    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal.integral.min <- extraArgs$signal.integral.min
    signal.integral.max <- extraArgs$signal.integral.max
    background.integral.min <- extraArgs$background.integral.min
    background.integral.max <- extraArgs$background.integral.max
    .validate_class(signal.integral.min, c("integer", "numeric"))
    .validate_class(signal.integral.max, c("integer", "numeric"))
    .validate_class(background.integral.min, c("integer", "numeric"))
    .validate_class(background.integral.max, c("integer", "numeric"))
    signal_integral <- signal.integral.min:signal.integral.max
    background_integral <- background.integral.min:background.integral.max
  }

  if (integral_input == "measurement") {
    x.range <- get_RLum(object, recordType = c("OSL", "IRSL"))[[1]][, 1]
    signal_integral <- .convert_to_channels(x.range, signal_integral, "time")
    background_integral <- .convert_to_channels(x.range, background_integral, "time")
  }
  signal_integral <- .validate_integral(signal_integral)
  background_integral <- .validate_integral(background_integral,
                                            min = max(signal_integral) + 1)

# Set parameters ------------------------------------------------------------------------------
  ##set n.channels
  if(is.null(n.channels)){
    n.channels <- ceiling(
      (min(background_integral) - 1 - max(signal_integral)) / (max(signal_integral) - min(signal_integral))
    )
    if (verbose) {
      .throw_message("'n.channels' not specified, set to ", n.channels,
                     error = FALSE)
    }
  }

  analyse_function.settings <- list(
     sequence.structure = c("TL", "IR50", "pIRIR225"),
     dose.points = NULL,
     mtext.outer = "",
     fit.method = "EXP",
     fit.force_through_origin = FALSE,
     trim_channels = FALSE,
     plot = FALSE,
     plot_singlePanels = FALSE
  )

  analyse_function.settings <- modifyList(analyse_function.settings, analyse_function.control)

# Analyse -------------------------------------------------------------------------------------
  ##set integral sequence
  if (is.null(signal_integral.seq)) {
    signal_integral.seq <- seq(min(signal_integral),
                               min(background_integral) - 1,
                               by = max(signal_integral) - min(signal_integral))
  }

  if(analyse_function  == "analyse_SAR.CWOSL"){
    results <- merge_RLum(lapply(1:n.channels, function(x){
      integral <- signal_integral.seq[if (method == "shift") x else 1]:signal_integral.seq[x+1]
      analyse_SAR.CWOSL(
        object = object,
        signal_integral = integral,
        background_integral = background_integral,
        dose.points = analyse_function.settings$dose.points,
        mtext.outer = analyse_function.settings$mtext.outer,
        fit.force_through_origin = analyse_function.settings$fit.force_through_origin,
        fit.method = analyse_function.settings$fit.method,
        trim_channels = analyse_function.settings$trim_channels,
        plot = analyse_function.settings$plot,
        plot_singlePanels = analyse_function.settings$plot_singlePanels,
        verbose = verbose
      )
    }))
  } else if (analyse_function == "analyse_pIRIRSequence") {
    results <- lapply(1:n.channels, function(x) {
      integral <- signal_integral.seq[if (method == "shift") x else 1]:signal_integral.seq[x+1]
      analyse_pIRIRSequence(
        object = object,
        signal_integral = integral,
        background_integral = background_integral,
        dose.points = analyse_function.settings$dose.points,
        mtext.outer = analyse_function.settings$mtext.outer,
        plot = analyse_function.settings$plot,
        plot_singlePanels = analyse_function.settings$plot_singlePanels,
        sequence.structure = analyse_function.settings$sequence.structure,
        verbose = verbose
      )
    })

    ## as the analyse_pIRIRSequence() may fail, we see how many results
    ## we've actually managed to produce
    num.valid.results <- sum(!sapply(results, is.null))
    if (num.valid.results == 0) {
      .throw_error("No valid results produced")
    }
    if (num.valid.results > 1) {
      results <- merge_RLum(results)
    }
  }

  ## Plot -------------------------------------------------------------------

  pIRIR_signals <- if (analyse_function == "analyse_pIRIRSequence") {
                     unique(get_RLum(results)$Signal)
                   } else NA

  ## run this in a loop to account for pIRIR data
  df_final <- lapply(1:length(pIRIR_signals), function(i) {
      ##get data.frame
      df <- get_RLum(results)

      ##further limit
      if(!is.na(pIRIR_signals[1]))
        df <- df[df$Signal == pIRIR_signals[i],]

      ##add shine down curve, which is by definition the first IRSL/OSL curve
      ##and normalise on the highest De value
      OSL_curve <-
        as(get_RLum(object, recordType = "SL")[[i]], "matrix")

      ##limit to what we see
      OSL_curve <- OSL_curve[1:signal_integral.seq[n.channels + 1],]

      min.de <- min(df$De - df$De.Error, na.rm = TRUE)
      max.de <- max(df$De, na.rm = TRUE) + max(df$De.Error, na.rm = TRUE)
      m <- (min.de - max.de) /
        (min(OSL_curve[, 2], na.rm = TRUE) -
           max(OSL_curve[, 2], na.rm = TRUE))
      n <- max.de - m * max(OSL_curve[, 2])

      OSL_curve[, 2] <- m * OSL_curve[, 2] + n
      rm(n, m)

        ##set plot setting
        plot.settings <- modifyList(list(
          ylim = c(min.de, max.de),
          xlim = c(min(OSL_curve[, 1]), max(OSL_curve[, 1])),
          ylab = if(show_ShineDownCurve[1])
                  expression(paste(D[e], " [s] and ", L[n], " [a.u.]"))
                else
                  expression(paste(D[e], " [s]")),
          xlab = "Stimulation time [s]",
          main = "De(t) plot",
          pch = 1,
          mtext = ifelse(is.na(pIRIR_signals[1]), "", paste0("Signal: ",pIRIR_signals[i])),
          cex = 1,
          legend = if(show_ShineDownCurve[1]) TRUE else FALSE,
          legend.text = if(show_ShineDownCurve[1])
                          c(expression(D[e]), expression(L[n]-signal))
                        else
                          expression(D[e]),
          legend.pos = "bottomleft"
        ), list(...))

        ##set x-axis
        df_x <-
          OSL_curve[seq(max(signal_integral), signal_integral.seq[n.channels+1], length.out = nrow(df)),1]

        #combine everything to allow excluding unwanted values
        df_final <- cbind(df, df_x)

        if (respect_RC.Status)
          df_final <- df_final[df_final$RC.Status != "FAILED", ]

       if(plot[1]) {
        ##general settings
        par.default <- .par_defaults()
        on.exit(par(par.default), add = TRUE)
        par(cex = plot.settings$cex)

        ##open plot area
        plot(
          NA,
          NA,
          xlim = plot.settings$xlim,
          ylim = if(any(is.infinite(plot.settings$ylim))) c(-1,1) else plot.settings$ylim ,
          xlab = plot.settings$xlab,
          ylab = plot.settings$ylab,
          main = plot.settings$main
        )

        if (show_ShineDownCurve)
          lines(OSL_curve, type = "b", pch = 20)

        ##ToDo:color failed points red
        ##plot points and error bars
        points(df_final[, c("df_x", "De")], pch = plot.settings$pch)
        segments(
          x0 = df_final$df_x,
          y0 = df_final$De + df_final$De.Error,
          x1 = df_final$df_x,
          y1 = df_final$De - df_final$De.Error
        )

        ##set mtext
        mtext(side = 3, plot.settings$mtext)

        ##legend
        if(plot.settings$legend){
          legend(
            plot.settings$legend.pos,
            legend = plot.settings$legend.text,
            pch = c(plot.settings$pch, 20),
            bty = "n"
          )
        }
      } ## end plot
      ##set return
      return(df_final)
  })

# Return ------------------------------------------------------------------
  ##merge results
  set_RLum(
    class = "RLum.Results",
    data = list(
      De.values = as.data.frame(data.table::rbindlist(df_final)),
      signal_integral.seq = signal_integral.seq
      ),
    info = list(
      call = sys.call())
  )
}

#' Class `"RLum.Data.Curve"`
#'
#' Class for representing luminescence curve data.
#'
#' @name RLum.Data.Curve-class
#'
#' @docType class
#'
#' @slot recordType
#' Object of class "character" containing the type of the curve (e.g. "TL" or
#' "OSL").
#'
#' @slot curveType
#' Object of class "character" containing curve type, allowed values are
#' "measured" or "predefined".
#'
#' @slot data
#' Object of class [matrix] containing curve x and y data.
#' 'data' can also be of type `RLum.Data.Curve` to change object values without
#' de-constructing the object. For example:
#' ```
#' set_RLum(class = 'RLum.Data.Curve',
#'          data = Your.RLum.Data.Curve,
#'          recordType = 'never seen before')
#' ```
#' would just change the `recordType`. Missing arguments  the value is taken
#' from the input object in 'data' (which is already an `RLum.Data.Curve`
#' object in this example).
#'
#' @note
#' The class should only contain data for a single curve. For additional
#' elements the slot `info` can be used (e.g. providing additional heating
#' ramp curve). Objects from the class `RLum.Data.Curve` are produced by other
#' functions (partly within [Luminescence::RLum.Analysis-class] objects),
#' namely: [Luminescence::Risoe.BINfileData2RLum.Analysis], [Luminescence::read_XSYG2R]
#'
#' @section Create objects from this Class:
#' Objects can be created by calls of the form
#' `set_RLum(class = "RLum.Data.Curve", ...)`.
#'
#' @section Class version: 0.5.2
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum-class], [Luminescence::RLum.Data-class],
#' [Luminescence::plot_RLum], [Luminescence::merge_RLum]
#'
#' @keywords classes
#'
#' @examples
#'
#' showClass("RLum.Data.Curve")
#'
#' ##set empty curve object
#' set_RLum(class = "RLum.Data.Curve")
#'
#' @export
setClass("RLum.Data.Curve",
         slots = list(
           recordType = "character",
           curveType = "character",
           data = "matrix"
           ),
         contains = "RLum.Data",
         prototype = list(
           recordType = NA_character_,
           curveType = NA_character_,
           data = matrix(data = 0, ncol = 2)
           )
         )

## as() ---------------------------------------------------------------------
##LIST
##COERCE RLum.Data.Curve >> list AND list >> RLum.Data.Curve
#' as() - RLum-object coercion
#'
#' for `[RLum.Data.Curve-class]`
#'
#' **[Luminescence::RLum.Data.Curve-class]**
#'
#' \tabular{ll}{
#'  **from** \tab **to**\cr
#'   `list` \tab `list` \cr
#'   `data.frame` \tab `data.frame`\cr
#'   `matrix` \tab `matrix`
#' }
#'
#' @param from [Luminescence::RLum-class], [list], [data.frame], [matrix] (**required**):
#'  object to be coerced from
#'
#' @param to [character] (**required**):
#' class name to be coerced to
#'
#' @seealso [methods::as]
#'
#' @note
#' Due to the complex structure of the `RLum` objects itself a coercing to standard
#' R data structures will be always loosely!
#'
#' @name as
setAs("list", "RLum.Data.Curve",
      function(from,to){
        if (length(from) == 0)
          return(set_RLum("RLum.Data.Curve"))
        new(to,
            recordType = "unknown curve type",
            curveType = NA_character_,
            data = matrix(unlist(from), ncol = 2),
            info = list())
      })


setAs("RLum.Data.Curve", "list",
      function(from){
          list(x = from@data[,1], y = from@data[,2])
      })

##DATA.FRAME
##COERCE RLum.Data.Curve >> data.frame AND data.frame >> RLum.Data.Curve
setAs("data.frame", "RLum.Data.Curve",
      function(from,to){
              new(to,
                  recordType = "unknown curve type",
                  curveType = NA_character_,
                  data = as.matrix(from),
                  info = list())
            })

setAs("RLum.Data.Curve", "data.frame",
      function(from){
        data.frame(x = from@data[,1],
                   y = from@data[,2])
      })


##MATRIX
##COERCE RLum.Data.Curve >> matrix AND matrix >> RLum.Data.Curve
setAs("matrix", "RLum.Data.Curve",
      function(from,to){
        new(to,
            recordType = "unknown curve type",
            curveType = NA_character_,
            data = from,
            info = list())
      })

setAs("RLum.Data.Curve", "matrix",
      function(from){
        from@data
      })

## show() -------------------------------------------------------------------
#' @describeIn show
#' Show the structure of `RLum.Data.Curve` objects.
#'
#' @export
setMethod("show",
          signature(object = "RLum.Data.Curve"),
          function(object){

            ##print information
            cat("\n [RLum.Data.Curve-class]")
            cat("\n\t recordType:", object@recordType)
            cat("\n\t curveType:",  object@curveType)
            cat("\n\t measured values:", length(object@data[,1]))
            cat("\n\t .. range of x-values:", suppressWarnings(range(object@data[,1])))
            cat("\n\t .. range of y-values:",
                suppressWarnings(min(object@data[,2], na.rm = TRUE)),
                suppressWarnings(max(object@data[,2], na.rm = TRUE)),
                ifelse(anyNA(object@data[, 2]), "(contains NA values)", "")
               )
            cat("\n\t additional info elements:", length(object@info), "\n")
            #cat("\n\t\t >> names:", names(object@info))
          }
)

## set_RLum() ---------------------------------------------------------------
#' @describeIn set_RLum
#' Construction method for [Luminescence::RLum.Data.Curve-class] objects.
#'
#' @param recordType [character] (*optional*):
#' record type (e.g., "OSL")
#'
#' @param curveType [character] (*optional*):
#' curve type (e.g., "predefined" or "measured")
#'
#' @param data [matrix] or [list] (*with default*):
#' a matrix containing raw curve data or a list containing the data to be
#' stored in the object (for [Luminescence::RLum.Results-class] objects) . If `data` itself
#' is a `RLum.Data.Curve`-object this can be used to re-construct the object,
#' i.e. modified parameters except `.uid`, `.pid` and `originator`. The rest
#' will be subject to copy and paste unless provided.
#'
#' @param info [list] (*optional*):
#' a list containing additional info data for the object.
#'
#' @export
setMethod(
  "set_RLum",
  signature = signature("RLum.Data.Curve"),
  definition = function(
    class,
    originator,
    .uid,
    .pid,
    recordType = NA_character_,
    curveType = NA_character_,
    data = matrix(0, ncol = 2),
    info = list()) {
    .set_function_name("set_RLum")
    on.exit(.unset_function_name(), add = TRUE)

    ##The case where an RLum.Data.Curve object can be provided
    ##with this RLum.Data.Curve objects can be provided to be reconstructed
    if (inherits(data, "RLum.Data.Curve")) {
      ##check for missing curveType
      if (missing(curveType))
        curveType <- data@curveType

      ##check for missing recordType
      if(missing(recordType))
        recordType <- data@recordType

      ##check for missing info
      if(missing(info))
        info <- data@info

      originator <- data@originator
      .uid <- data@.uid
      .pid <- data@.pid
      data <- data@data
    }

    if (NCOL(data) != 2) {
      .throw_error("'data' should be a matrix with 2 columns")
    }

    ## set empty class form object
    newRLumDataCurve <- new("RLum.Data.Curve")

    ## fill - this is the faster way, filling in new() costs ...
    newRLumDataCurve@originator <- originator
    newRLumDataCurve@recordType <- recordType
    newRLumDataCurve@curveType <- curveType
    newRLumDataCurve@data <- data
    newRLumDataCurve@info <- info
    newRLumDataCurve@.uid <- .uid
    newRLumDataCurve@.pid <- .pid

    return(newRLumDataCurve)
  }
)

## get_RLum() ---------------------------------------------------------------
#' @describeIn get_RLum
#' Accessor method for [Luminescence::RLum.Data.Curve-class] object.
#' The argument `info.object` is optional to directly access the info elements.
#' If no info element name is provided, the raw curve data (matrix) will be
#' returned.
#'
#' @export
setMethod("get_RLum",
          signature("RLum.Data.Curve"),
          definition = function(object, info.object = NULL) {
          .set_function_name("get_RLum")
          on.exit(.unset_function_name(), add = TRUE)

    if (is.null(info.object)) {
      return(object@data)
    }
    .validate_class(info.object, "character", null.ok = TRUE, length = 1)
    if (length(object@info) == 0) {
      .throw_warning("'object' has no info objects, NULL returned")
      return(NULL)
    }
    if (!info.object %in% names(object@info)) {
      .throw_warning("Invalid 'info.object' name, valid names are: ",
                     .collapse(names(object@info)))
      return(NULL)
    }

    unlist(object@info[info.object])
  })


## length_RLum() ------------------------------------------------------------
#' @describeIn length_RLum
#' Returns the length of the curve object, which is the maximum of the value
#' time/temperature of the curve (corresponding to the stimulation length).
#'
#' @export
setMethod("length_RLum",
          "RLum.Data.Curve",
          function(object){
            max(object@data[,1])
          })


## names_RLum() -------------------------------------------------------------
#' @describeIn names_RLum
#' Returns the names info elements stored in the object.
#'
#' @export
setMethod("names_RLum",
          "RLum.Data.Curve",
          function(object){
            names(object@info)
          })


## bin_RLum.Data() ----------------------------------------------------------
#' @describeIn bin_RLum.Data
#' Allows binning of RLum.Data.Curve data.
#'
#' @param bin_size [integer] (*with default*):
#' number of channels used for each bin, e.g. `bin_size = 2` means that
#' two channels are binned.
#'
#' @export
setMethod(f = "bin_RLum.Data",
          signature = "RLum.Data.Curve",
          function(object, bin_size = 2) {
            .set_function_name("bin_RLum.Data.Curve")
            on.exit(.unset_function_name(), add = TRUE)

            .validate_positive_scalar(bin_size, int = TRUE)

              ##set stepping vector
              stepping <- seq(1, nrow(object@data), by = bin_size)

              ##get bin vector
              bin_vector <- object@data[, 2]

              ##set desired length of the vector
              ##to avoid add effects later
              length(bin_vector) <-
                suppressWarnings(prod(dim(matrix(
                  bin_vector, ncol = length(stepping)
                ))))

              ##define new matrix for binning
              bin_matrix <-
                matrix(bin_vector, ncol = length(stepping))

              ##calcuate column sums and replace matrix
              ##this is much faster than anly apply loop
              object@data <-
                matrix(c(object@data[stepping], colSums(bin_matrix, na.rm = TRUE)), ncol = 2)

              ##set matrix
              return(set_RLum(class = "RLum.Data.Curve",
                              data = object))
          })


## smooth_RLum() ------------------------------------------------------------
#' @describeIn smooth_RLum
#' Smoothing of [Luminescence::RLum.Data.Curve-class] objects using a rolling mean or median.
#' For methods `"mean"` and `"median"`, smoothing is performed by rolling
#' mean and rolling median with window of size `k`. Method `"Carter_etal_2018"`
#' implements a Poisson smoother for dark-background signals measured by a
#' photomultiplier tube.
#'
#' @param k [Luminescence::smooth_RLum]; [integer] (*with default*):
#' window for the rolling mean or median. If `NULL`, this set automatically
#' (ignored if `method = "Carter_etal_2018"`).
#'
#' @param fill [Luminescence::smooth_RLum]; [numeric] (*with default*):
#' value used to pad the result so to have the same length as the input.
#'
#' @param align [Luminescence::smooth_RLum]; [character] (*with default*):
#' one of `"right"`, `"center"` or `"left"`, specifying whether the index
#' of the result should be right-aligned (default), centred, or left-aligned
#' compared to the rolling window of observations (ignored if
#' `method = "Carter_etal_2018"`).
#'
#' @param method [Luminescence::smooth_RLum]; [character] (*with default*):
#' smoothing method to be applied: one of `"mean"`, `"median"` or
#' `"Carter_etal_2018"`.
#'
#' @param p_acceptance [Luminescence::smooth_RLum]; [numeric] (*with default*):
#' probability threshold of accepting a value to be a sample from a Poisson
#' distribution (only used for `method = "Carter_etal_2018"`). Values that
#' have a Poisson probability below the threshold are replaced by the average
#' over the four neighbouring values.
#'
#' @references
#' Carter, J., Cresswell, A.J., Kinnaird, T.C., Carmichael, L.A., Murphy, S. &
#' Sanderson, D.C.W., 2018. Non-Poisson variations in photomultipliers and
#' implications for luminescence dating. Radiation Measurements 120, 267-273.
#' \doi{10.1016/j.radmeas.2018.05.010}
#'
#' @export
setMethod(
  f = "smooth_RLum",
  signature = "RLum.Data.Curve",
  function(object, k = NULL, fill = NA, align = "right", method = "mean",
           p_acceptance = 1e-7) {

        object@data[,2] <- .smoothing(
          x = object@data[,2],
          k = k,
          fill = fill,
          align = align,
          p_acceptance = p_acceptance,
          method = method)

        ##return via set function to get a new id
        set_RLum(class = "RLum.Data.Curve",
                 originator = "smooth_RLum",
                 data = object)
    }
 )
## normalise_RLum() --------------------------------------------------------------
#' @describeIn normalise_RLum
#' Normalise [Luminescence::RLum.Data.Curve-class] objects to value set via
#' the argument `norm`
#'
#' @export
setMethod(
  f = "normalise_RLum",
  signature = "RLum.Data.Curve",
  function(object, norm) {
    if (norm == "intensity") {
      ## compute the channel length and assign it to `norm`, so that we can
      ## piggy-back on the code for normalisation by a scalar
      norm <- diff(c(0, object@data[, 1]))
    }
    object@data[, 2] <- .normalise_curve(object@data[, 2], norm)
    object
  }
)
## melt_RLum() --------------------------------------------------------------
#' @describeIn melt_RLum
#' Melts [Luminescence::RLum.Data.Curve-class] objects into a flat data.frame with columns
#' `X`, `Y`, `TYPE`, `UID`, to be used in combination with other packages
#' such as `ggplot2`.
#'
#' @export
setMethod(
  f = "melt_RLum",
  signature = "RLum.Data.Curve",
  function(object) {
    data.frame(
      X = object@data[,1],
      Y = object@data[,2],
      TYPE = object@recordType,
      UID = object@.uid)
  }
)

#' @title Fit a dose-response curve for luminescence data (Lx/Tx against dose)
#'
#' @description
#' A dose-response curve is produced for luminescence measurements using a
#' regenerative or additive protocol. The function supports interpolation and
#' extrapolation to calculate the equivalent dose.
#'
#' @details
#'
#' ## Implemented fitting methods
#'
#' For all options (except for the `LIN`, `QDR` and the `EXP OR LIN`),
#' the [minpack.lm::nlsLM] function with the `LM` (Levenberg-Marquardt algorithm)
#' algorithm is used. Note: For historical reasons for the Monte Carlo
#' simulations partly the function [nls] using the `port` algorithm.
#'
#' The solution is found by transforming the function or using [stats::uniroot].
#'
#' **Keyword: `LIN`**
#'
#' Fits a linear function to the data using [lm]:
#' \deqn{y = mx + n}
#'
#' **Keyword: `QDR`**
#'
#' Fits a linear function with a quadratic term to the data using  [lm]:
#' \deqn{y = a + bx + cx^2}
#'
#' **Keyword: `EXP`**
#'
#' Adapts a function of the form
#' \deqn{y = a(1 - \exp(-\frac{(x+c)}{b}))}
#'
#' Parameters b and c are approximated by a linear fit using [lm]. Note: \eqn{b = D0}
#'
#' **Keyword: `EXP OR LIN`**
#'
#' Works for some cases where an `EXP` fit fails. If the `EXP` fit fails,
#' a `LIN` fit is done instead, which always works.
#'
#' **Keyword: `EXP+LIN`**
#'
#' Tries to fit an exponential plus linear function of the form:
#'
#' \deqn{y = a(1 - \exp(-\frac{x + c}{b}) + (gx))}
#' The \eqn{D_e} is calculated by iteration.
#'
#' **Note:** In the context of luminescence dating, this function has no physical meaning.
#' Therefore, no \eqn{D_0} value is returned.
#'
#' **Keyword: `EXP+EXP`**
#'
#' Tries to fit a double exponential function of the form
#'
#' \deqn{y = (a_1 (1 - \exp(-\frac{x}{b_1}))) + (a_2 (1 - \exp(-\frac{x}{b_2})))}
#'
#' *This fitting procedure is not really robust against wrong start parameters.*
#'
#' **Keyword: `GOK`**
#'
#' Tries to fit the general-order kinetics function following Guralnik et al. (2015)
#' of the form
#'
#' \deqn{y = a (d - (1 + (\frac{1}{b}) x c)^{(-1/c)})}
#'
#' where \eqn{c > 0} is a kinetic order modifier
#' (not to be confused with **c** in `EXP` or `EXP+LIN`!).
#'
#' **Keyword: `OTOR`** (former `LambertW`)
#'
#' This tries to fit a dose-response curve based on the Lambert W function
#' and the one trap one recombination centre (OTOR) model according to Pagonis
#' et al. (2020). The function has the form:
#'
#' \deqn{y = (1 + (\mathcal{W}((R - 1) * \exp(R - 1 - ((x + D_{int}) / D_{c}))) / (1 - R))) * N}
#'
#' with \eqn{W} the Lambert W function (calculated using [lamW::lambertW0]),
#' \eqn{R} the dimensionless retrapping ratio, \eqn{N} the total concentration
#' of trappings states in cm\eqn{^{-3}}, \eqn{D_{c} = N/R} a constant, and
#' \eqn{D_{int}} is the offset on the x-axis. Note that \eqn{R} and \eqn{D_{c}}
#' have a valid physical interpretation only when saturation is reached.
#' Please note that finding the root in `mode = "extrapolation"`
#' is a non-easy task due to the shape of the function and the results might be
#' unexpected.
#'
#' **Keyword: `OTORX`**
#'
#' This adapts extended OTOR (therefore: OTORX) model proposed by Lawless and
#' Timar-Gabor (2024) accounting for retrapping (the equation implemented here
#' is written slightly differently than in the original manuscript):
#'
#' \deqn{F_{OTORX} = 1 + \left[\mathcal{W}\left(-Q * \exp\left(-Q-(1-Q(1-\frac{1}{\exp(1)})) \frac{(D + a)}{D_{63}}\right)\right)\right] / Q}
#'
#' with
#'
#' \deqn{Q = \frac{A_m - A_n}{A_m}\frac{N}{N+N_D}}
#'
#' where \eqn{A_m} and \eqn{A_n} are rate constants for the recombination and
#' the trapping of electrons (\eqn{N}), respectively. \eqn{D_{63}} corresponds to
#' the value at which the trap occupation corresponds to the 63% of the saturation value.
#' \eqn{a} is in an offset. If set to zero, the curve will be forced through the origin
#' as in the original publication.
#'
#' For the implementation the calculation reads further
#'
#' \deqn{y = \frac{F_{OTORX}(((D + a)/D_{63}), Q)}{F_{OTORX}((D_{test} + a)/D_{63}, Q)}}
#'
#' with \eqn{D_{test}} being the test dose in the same unit (usually s or Gy) as
#' the regeneration dose points. This value is essential and needs to provided
#' along with the usual dose and \eqn{\frac{L_x}{T_x}} values (see `object` parameter input
#' and the example section). For more details see Lawless and Timar-Gabor (2024).
#'
#' The fit also returns the parameter \eqn{R} know from `OTOR`, which is derived
#' as \eqn{R = 1 - Q}.
#'
#' *Note: The offset adder \eqn{a} is not part of the formula in Timar-Gabor (2024) and can
#' be set to zero with the option `fit.force_through_origin = TRUE`*
#'
#' **Fit weighting**
#'
#' If the option `fit.weights =  TRUE` is chosen, weights are calculated using
#' provided signal errors (\eqn{\frac{L_x}{T_x}} error):
#' \deqn{fit.weights = \frac{\frac{1}{error}}{\Sigma{\frac{1}{error}}}}
#'
#' **Error estimation using Monte Carlo simulation**
#'
#' Error estimation is done using a parametric bootstrapping approach. A set of
#' \eqn{\frac{L_x}{T_x}} values is constructed by randomly drawing curve data sampled from normal
#' distributions. The normal distribution is defined by the input values (`mean
#' = value`, `sd = value.error`). Then, a dose-response curve fit is attempted for each
#' dataset resulting in a new distribution of single `De` values. The standard
#' deviation of this distribution becomes then the error of the `De`. With increasing
#' iterations, the error value becomes more stable. However, naturally the error
#' will not decrease with more MC runs.
#'
#' Alternatively, the function returns highest probability density interval
#' estimates as output, users may find more useful under certain circumstances.
#'
#' **Note:** It may take some calculation time with increasing MC runs,
#' especially for the composed functions (`EXP+LIN` and `EXP+EXP`).\cr
#' Each error estimation is done with the function of the chosen fitting method.
#'
#' @param object [data.frame] or a [list] of such objects (**required**):
#' data frame with columns for `Dose`, `LxTx`, `LxTx.Error` and `TnTx`.
#' The column for the test dose response is optional, but requires `'TnTx'` as
#' column name if used. For exponential fits at least three dose points
#' (including the natural) should be provided. If `object` is a list,
#' the function is called on each of its elements.
#' If `fit.method = "OTORX"` you have  to provide the test dose in the same unit
#' as the dose in a column called `Test_Dose`. The function searches explicitly
#' for this column name. Only the first value will be used assuming a constant
#' test dose over the measurement cycle.
#'
#' @param mode [character] (*with default*):
#' selects calculation mode of the function.
#' - `"interpolation"` (default) calculates the De by interpolation,
#' - `"extrapolation"` calculates the equivalent dose by extrapolation
#'    (useful for MAAD measurements) and
#' - `"alternate"` calculates no equivalent dose and just fits the data points.
#'
#' Please note that for option `"interpolation"` the first point is considered
#' as natural dose
#'
#' @param fit.method [character] (*with default*):
#' function used for fitting. Possible options are: `LIN`, `QDR`, `EXP`,
#' `EXP OR LIN`, `EXP+LIN`, `EXP+EXP` (not defined for extrapolation), `GOK`,
#' `OTOR` and `OTORX`. See details.
#'
#' @param fit.force_through_origin [logical] (*with default*)
#' allow to force the fitted function through the origin.
#' For `method = "EXP+EXP"` the function will be fixed through
#' the origin in either case, so this option will have no effect.
#'
#' @param fit.weights [logical] (*with default*):
#' option whether the fitting is done with or without weights. See details.
#'
#' @param fit.includingRepeatedRegPoints [logical] (*with default*):
#' includes repeated points for fitting (`TRUE`/`FALSE`).
#'
#' @param fit.NumberRegPoints [integer] (*optional*):
#' set number of regeneration points manually. By default the number of all (!)
#' regeneration points is used automatically.
#'
#' @param fit.NumberRegPointsReal [integer] (*optional*):
#' if the number of regeneration points is provided manually, the value of the
#' real, regeneration points = all points (repeated points) including reg 0,
#' has to be inserted.
#'
#' @param fit.bounds [logical] (*with default*):
#' set lower fit bounds for all fitting parameters to 0. Limited for the use
#' with the fit methods `EXP`, `EXP+LIN`, `EXP OR LIN`, `GOK`, `OTOR`, `OTORX`
#' Argument to be inserted for experimental application only!
#'
#' @param n.MC [integer] (*with default*):
#' number of Monte Carlo simulations for error estimation, see details.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar. If `verbose = FALSE` also no
#' `txtProgressBar` is shown.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param ... Further arguments to be passed (currently ignored).
#'
#' @return
#' An [Luminescence::RLum.Results-class] object is returned
#' containing the slot `data` with the
#' following elements:
#'
#' **Overview elements**
#' \tabular{lll}{
#' **DATA.OBJECT** \tab **TYPE** \tab **DESCRIPTION** \cr
#' `..$De` : \tab  `data.frame` \tab Table with De values \cr
#' `..$De.MC` : \tab `numeric` \tab Table with De values from MC runs \cr
#' `..$Fit` : \tab [nls] or [lm] \tab object from the fitting for `EXP`, `EXP+LIN` and `EXP+EXP`.
#' In case of a resulting  linear fit when using `LIN`, `QDR` or `EXP OR LIN` \cr
#' `..Fit.Args` : \tab `list` \tab Arguments to the function \cr
#' `..$Formula` : \tab [expression] \tab Fitting formula as R expression \cr
#' }
#'
#' The `@info` slot contains the following elements:
#' \tabular{lll}{
#' **DATA.OBJECT** \tab **TYPE** \tab **DESCRIPTION** \cr
#' `..$fit_messag`: \tab `character` \tab The fit message reported \cr
#' `..$call` : \tab `call` \tab The original function call \cr
#' }
#'
#' If `object` is a list, then the function returns a list of
#' [Luminescence::RLum.Results-class]
#' objects as defined above.
#'
#' **Details - `DATA.OBJECT$De`**
#' This object is a [data.frame] with the following columns
#' \tabular{lll}{
#' `De` \tab [numeric] \tab equivalent dose \cr
#' `De.Error` \tab [numeric] \tab standard error the equivalent dose \cr
#' `D01` \tab [numeric] \tab \eqn{D_0} value, curvature parameter of the exponential \cr
#' `D01.ERROR` \tab [numeric] \tab standard error of the \eqn{D_0} value\cr
#' `D02` \tab [numeric] \tab 2nd \eqn{D_0} value, only for `EXP+EXP`\cr
#' `D02.ERROR` \tab [numeric] \tab standard error for 2nd \eqn{D_0}; only for `EXP+EXP`\cr
#' `R` \tab [numeric] \tab the material specific parameter \eqn{R}\cr
#' `R.ERROR` \tab [numeric] \tab the uncertainty of R \eqn{R}\cr
#' `Dc` \tab [numeric] \tab value indicating saturation level; only for `OTOR` \cr
#' `D63` \tab [numeric] \tab the specific saturation level; only for `OTORX` \cr
#' `n_N` \tab [numeric] \tab saturation level of dose-response curve derived via integration from the used function; it compares the full integral of the curves (`N`) to the integral until `De` (`n`) (e.g.,  Guralnik et al., 2015)\cr
#' `De.MC` \tab [numeric] \tab equivalent dose derived by Monte-Carlo simulation; ideally identical to `De`\cr
#' `Fit` \tab [character] \tab applied fit function \cr
#' `Mode` \tab [character] \tab mode used in fitting \cr
#' `HPDI68_L` \tab [numeric] \tab highest probability density of approximated equivalent dose probability curve representing the lower boundary of 68% probability \cr
#' `HPDI68_U` \tab [numeric] \tab same as `HPDI68_L` for the upper bound \cr
#' `HPDI95_L` \tab [numeric] \tab same as `HPDI68_L` but for 95% probability \cr
#' `HPDI95_U` \tab [numeric] \tab same as `HPDI95_L` but for the upper bound \cr
#' `.De.plot` \tab [numeric] \tab equivalent dose used internally for plotting \cr
#' `.De.raw` \tab [numeric] \tab equivalent dose reported 'as is', that is containing infinities and negative values if they could be calculated. Bear in mind that negative values are meaningless and may be arbitrary.\cr
#' }
#'
#' @section Function version: 1.4.5
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Michael Dietze, GFZ Potsdam (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @references
#'
#' Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46. \doi{10.26034/la.atl.1989.150}
#'
#' Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
#' Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
#' from feldspar. Radiation Measurements 81, 224-231. \doi{10.1016/j.radmeas.2015.02.011}
#'
#' Lawless, J.L., Timar-Gabor, A., 2024. A new analytical model to fit both fine and coarse grained quartz luminescence dose response curves. Radiation Measurements 170, 107045. \doi{10.1016/j.radmeas.2023.107045}
#'
#' Pagonis, V., Kitis, G., Chen, R., 2020. A new analytical equation for the dose response of dosimetric materials,
#' based on the Lambert W function. Journal of Luminescence 225, 117333. \doi{10.1016/j.jlumin.2020.117333}
#'
#' @seealso [Luminescence::plot_DoseResponseCurve], [nls],
#' [Luminescence::RLum.Results-class], [Luminescence::get_RLum],
#' [minpack.lm::nlsLM], [lm], [uniroot], [lamW::lambertW0]
#'
#' @examples
#'
#' ##(1) fit growth curve for a dummy data.set and show De value
#' data(ExampleData.LxTxData, envir = environment())
#' temp <- fit_DoseResponseCurve(LxTxData)
#' get_RLum(temp)
#'
#' ##(1b) to access the fitting value try
#' get_RLum(temp, data.object = "Fit")
#'
#' ##(2) fit using the 'extrapolation' mode
#' LxTxData[1,2:3] <- c(0.5, 0.001)
#' print(fit_DoseResponseCurve(LxTxData, mode = "extrapolation"))
#'
#' ##(3) fit using the 'alternate' mode
#' LxTxData[1,2:3] <- c(0.5, 0.001)
#' print(fit_DoseResponseCurve(LxTxData, mode = "alternate"))
#'
#' ##(4) import and fit test data set by Berger & Huntley 1989
#' QNL84_2_unbleached <-
#' read.table(system.file("extdata/QNL84_2_unbleached.txt", package = "Luminescence"))
#'
#' results <- fit_DoseResponseCurve(
#'  QNL84_2_unbleached,
#'  mode = "extrapolation",
#'  verbose = FALSE)
#'
#' #calculate confidence interval for the parameters
#' #as alternative error estimation
#' confint(results$Fit, level = 0.68)
#'
#' \dontrun{
#' ##(5) special case the OTORX model with test dose column
#' df <- cbind(LxTxData, Test_Dose = 15)
#' fit_DoseResponseCurve(object = df, fit.method = "OTORX", n.MC = 10) |>
#'  plot_DoseResponseCurve()
#'
#' QNL84_2_bleached <-
#' read.table(system.file("extdata/QNL84_2_bleached.txt", package = "Luminescence"))
#' STRB87_1_unbleached <-
#' read.table(system.file("extdata/STRB87_1_unbleached.txt", package = "Luminescence"))
#' STRB87_1_bleached <-
#' read.table(system.file("extdata/STRB87_1_bleached.txt", package = "Luminescence"))
#'
#' print(
#'  fit_DoseResponseCurve(
#'  QNL84_2_bleached,
#'  mode = "alternate",
#'  verbose = FALSE)$Fit)
#'
#' print(
#'  fit_DoseResponseCurve(
#'  STRB87_1_unbleached,
#'  mode = "alternate",
#'  verbose = FALSE)$Fit)
#'
#' print(
#'  fit_DoseResponseCurve(
#'  STRB87_1_bleached,
#'  mode = "alternate",
#'  verbose = FALSE)$Fit)
#'  }
#'
#' @export
fit_DoseResponseCurve <- function(
  object,
  mode = c("interpolation", "extrapolation", "alternate"),
  fit.method = c("EXP", "LIN", "QDR", "EXP OR LIN", "EXP+LIN", "EXP+EXP",
                 "GOK", "OTOR", "OTORX"),
  fit.force_through_origin = FALSE,
  fit.weights = TRUE,
  fit.includingRepeatedRegPoints = TRUE,
  fit.NumberRegPoints = NULL,
  fit.NumberRegPointsReal = NULL,
  fit.bounds = TRUE,
  n.MC = 100,
  txtProgressBar = TRUE,
  verbose = TRUE,
  ...
) {
  .set_function_name("fit_DoseResponseCurve")
  on.exit(.unset_function_name(), add = TRUE)

  ## Self-call --------------------------------------------------------------
  if (inherits(object, "list")) {
    lapply(object,
           function(x) .validate_class(x, c("data.frame", "matrix"),
                                       name = "All elements of 'object'"))

    results <- lapply(object, function(x) {
      fit_DoseResponseCurve(
          object = x,
          mode = mode,
          fit.method = fit.method,
          fit.force_through_origin = fit.force_through_origin,
          fit.weights = fit.weights,
          fit.includingRepeatedRegPoints = fit.includingRepeatedRegPoints,
          fit.NumberRegPoints = fit.NumberRegPoints,
          fit.NumberRegPointsReal = fit.NumberRegPointsReal,
          fit.bounds = fit.bounds,
          n.MC = n.MC,
          txtProgressBar = txtProgressBar,
          verbose = verbose,
          ...
      )
    })

    return(results)
  }
  ## Self-call end ----------------------------------------------------------

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("data.frame", "matrix", "list"))
  .validate_not_empty(object)
  mode <- .validate_args(mode, c("interpolation", "extrapolation", "alternate"))
  fit.method_supported <- c("LIN", "QDR", "EXP", "EXP OR LIN",
                            "EXP+LIN", "EXP+EXP", "GOK", "OTOR", "OTORX")
  fit.method <- .validate_args(fit.method, fit.method_supported)
  .validate_logical_scalar(fit.force_through_origin)
  .validate_logical_scalar(fit.weights)
  .validate_logical_scalar(fit.includingRepeatedRegPoints)
  .validate_logical_scalar(fit.bounds)
  .validate_positive_scalar(fit.NumberRegPoints, int = TRUE, null.ok = TRUE)
  .validate_positive_scalar(fit.NumberRegPointsReal, int = TRUE, null.ok = TRUE)
  .validate_positive_scalar(n.MC, int = TRUE)
  .validate_logical_scalar(txtProgressBar)
  .validate_logical_scalar(verbose)

  ## convert input to data.frame
  switch(
    class(object)[1],
    data.frame = object,
    matrix = object <- as.data.frame(object),
  )

  ##2.1 check column numbers; we assume that in this particular case no error value
  ##was provided, e.g., set all errors to 0
  if (ncol(object) < 2) {
    .throw_error("'object' should have at least 2 columns")
  }
  if (ncol(object) == 2)
    object <- cbind(object, 0)

  ##2.2 check for inf data in the data.frame
  if (any(is.infinite(unlist(object)))) {
    ## https://stackoverflow.com/questions/12188509/cleaning-inf-values-from-an-r-dataframe
    ## This is slow, but it does not break with previous code
    object <- do.call(data.frame,
                      lapply(object, function(x) replace(x, is.infinite(x), NA)))
      .throw_warning("Inf values found, replaced by NA")
  }

  ##2.3 check whether the dose value is equal all the time
  if (sum(abs(diff(object[[1]])), na.rm = TRUE) == 0) {
    .throw_message("All points have the same dose, NULL returned")
    return(NULL)
  }

  ## count and exclude NA values and print result
  if (sum(!stats::complete.cases(object)) > 0)
    .throw_warning(sum(!stats::complete.cases(object)),
                   " NA values removed")

  ## exclude NA
  object <- na.exclude(object)

  ## Check if anything is left after removal
  if (nrow(object) == 0) {
    .throw_message("After NA removal, nothing is left from the data set, ",
                   "NULL returned")
    return(NULL)
  }

  ##3. verbose mode
  if(!verbose)
    txtProgressBar <- FALSE

  ##remove rownames from data.frame, as this could causes errors for the reg point calculation
  rownames(object) <- NULL

  ## zero values in the data.frame are not allowed for the y-column
  y.zero <- object[, 2] == 0
  if (sum(y.zero) > 0) {
    .throw_warning(sum(y.zero), " values with 0 for Lx/Tx detected, ",
                   "replaced by ", .Machine$double.eps)
    object[y.zero, 2] <- .Machine$double.eps
  }

  ##1. INPUT
  #1.0.1 calculate number of reg points if not set
  if(is.null(fit.NumberRegPoints))
    fit.NumberRegPoints <- length(object[-1,1])

  if(is.null(fit.NumberRegPointsReal)){
    fit.RegPointsReal <- which(!duplicated(object[,1]) | object[,1] != 0)
    fit.NumberRegPointsReal <- length(fit.RegPointsReal)
  }

  ## 1.1 Produce data.frame from input values

  ## for interpolation the first point is considered as natural dose
  first.idx <- ifelse(mode == "interpolation", 2, 1)
  last.idx <- fit.NumberRegPoints + 1

  xy <- object[first.idx:last.idx, 1:2]
  colnames(xy) <- c("x", "y")
  y.Error <- object[first.idx:last.idx, 3]

  ##1.1.1 produce weights for weighted fitting
  if(fit.weights){
    fit.weights <- 1 / abs(y.Error) / sum(1 / abs(y.Error))

    if (anyNA(fit.weights)) { # FIXME(mcol): infinities?
      fit.weights <- rep(1, length(y.Error))
      .throw_warning("Error column invalid or 0, 'fit.weights' ignored")
    }
  }else{
    fit.weights <- rep(1, length(y.Error))
  }

  #1.2 Prepare data sets regeneration points for MC Simulation
  ## for interpolation the first point is considered as natural dose
  first.idx <- ifelse(mode == "interpolation", 2, 1)
  last.idx <- fit.NumberRegPoints + 1

  data.MC <- t(vapply(
      X = first.idx:last.idx,
      FUN = function(x) {
        sample(rnorm(
          n = 10000,
          mean = object[[2]][x],
          sd = abs(object[[3]][x])
        ),
        size = n.MC,
        replace = TRUE)
      },
      FUN.VALUE = numeric(n.MC)
    ))

  if (mode == "interpolation") {
    #1.3 Do the same for the natural signal
    data.MC.De <-
      sample(rnorm(10000, mean = object[1, 2], sd = abs(object[1, 3])),
             n.MC,
             replace = TRUE)
  } else if (mode == "extrapolation") {
    data.MC.De <- rep(0, n.MC)
  }

  #1.3 set x.natural
  x.natural <- rep(NA_real_, n.MC)

  ##1.4 set initialise variables
  De <- De.Error <- D01 <- R <- R.ERROR <- Dc <- D63 <- N <- TEST_DOSE <- NA_real_

  ##1.5 create bindings (we generate this with an internal function klate)
  var.g <- d <- Dint <- Q <- NA_real_

  ## FITTING ----------------------------------------------------------------
  ##3. Fitting values with nonlinear least-squares estimation of the parameters
  ## set functions for fitting
  ## REMINDER: DO NOT ADD {} brackets, otherwise the formula construction will not
  ## work

  ## get current environment, we need that later
  currn_env <- environment()

  ## Define functions ---------
  ### EXP ------- (C++ version available)
  fit.functionEXP <- function(a,b,c,x) a*(1-exp(-(x+c)/b))

  ### EXP+LIN ----------- (C++ version available)
  fit.functionEXPLIN <- function(a,b,c,g,x) a*(1-exp(-(x+c)/b)+(g*x))

  ### EXP+EXP ---------- (C++ version available)
  fit.functionEXPEXP <- function(a1,a2,b1,b2,x) (a1*(1-exp(-(x)/b1)))+(a2*(1-exp(-(x)/b2)))

  ### GOK ---------------- (C++ version available)
  fit.functionGOK <- function(a,b,c,d,x) a*(d-(1+(1/b)*x*c)^(-1/c))

  ### OTOR -------------
  fit.functionOTOR <- function(R, Dc, N, Dint, x) (1 + (lamW::lambertW0((R - 1) * exp(R - 1 - ((x + Dint) / Dc ))) / (1 - R))) * N

  ### OTORX -------------
  fit.functionOTORX <- function(x, Q, D63, c, a) .D2nN(x, Q, D63, a) * c / .D2nN(TEST_DOSE, Q, D63, a)

  ## input data for fitting; exclude repeated RegPoints
  if (!fit.includingRepeatedRegPoints[1]) {
    is.dup <- duplicated(xy$x)
    fit.weights <- fit.weights[!is.dup]
    data.MC <- data.MC[!is.dup, , drop = FALSE]
    y.Error <- y.Error[!is.dup]
    xy <- xy[!is.dup, , drop = FALSE]
  }
  data <- xy

  ## number of parameters in the non-linear models
  num.params <- 4
  if (fit.method %in% c("EXP", "EXP OR LIN"))
    num.params <- 3

  ## if the number of data points is smaller than the number of parameters
  ## to fit, the nls() function gets trapped in an infinite loop
  if (!fit.method %in% c("LIN", "QDR") && nrow(data) < num.params) {
    fit.method <- "LIN"
    msg <- paste("Fitting a non-linear least-squares model requires at least",
                 num.params, "dose points, 'fit.method' changed to 'LIN'")
    .throw_warning(msg)
    if (verbose)
      .throw_message(msg, error = FALSE)
  }

  ## helper to report the fit: this assigns the
  fit_message <- ""
  .report_fit <- function(De, ...) {
      fit_message <<- paste0(sprintf("Fit: %s (%s) | De = %.2f",
                                     fit.method, mode, abs(De)), ...)
      if (verbose)
        writeLines(paste("[fit_DoseResponseCurve()]", fit_message))
  }

  ## helper to report a failure in the fit
  .report_fit_failure <- function(method, mode, ...) {
    fit_message <<- sprintf("Fit failed for %s (%s)", fit.method, mode)
    if (verbose)
      writeLines(paste("[fit_DoseResponseCurve()]", fit_message))
  }

  ##START PARAMETER ESTIMATION
  ##general setting of start parameters for fitting

  ## a - estimation for the maximum of the y-values (Lx/Tx)
  a <- max(data[,2])

  ##b - get start parameters from a linear fit of the log(y) data
  ##    (don't even try fitting if no y value is positive)
  b <- 1
  if (any(data$y > 0)) {
    ## this may cause NaN values so we have to handle those later
    fit.lm <- try(stats::lm(suppressWarnings(log(data$y)) ~ data$x),
                  silent = TRUE)

    if (!inherits(fit.lm, "try-error") && !is.na(fit.lm$coefficients[2]))
      b <- as.numeric(1 / fit.lm$coefficients[2])
  }

  ##c - get start parameters from a linear fit - offset on x-axis
  fit.lm <- stats::lm(data$y ~ data$x)
  c <- as.numeric(abs(fit.lm$coefficients[1]/fit.lm$coefficients[2]))

  #take slope from x - y scaling
  g <- max(data[,2]/max(data[,1]))

  #set D01 and D02 (in case of EXP+EXP)
  D01 <- D01.ERROR <- D02 <- D02.ERROR <- NA

  ## Let start parameter vary -------------------------------------------------
  ## to be a little bit more flexible, the start parameters varies within
  ## a normal distribution

  ## draw 50 start values from a normal distribution
  if (fit.method != "LIN") {
    a.MC <- suppressWarnings(rnorm(50, mean = a, sd = a / 100))

    if (!is.na(b)) {
      b.MC <- suppressWarnings(rnorm(50, mean = b, sd = b / 100))
    }

    if(fit.force_through_origin)
      c.MC <- rep(0, 50)
    else
      c.MC <- suppressWarnings(rnorm(50, mean = c, sd = c / 100))
    g.MC <- suppressWarnings(rnorm(50, mean = g, sd = g / 1))

    ##set start vector (to avoid errors within the loop)
    a.start <-   b.start <- c.start <- g.start <- NA
  }

  ## QDR --------------------------------------------------------------------
  if (fit.method == "QDR") {
    ## establish models without and with intercept term
    model.qdr <- stats::update(
      y ~ I(x) + I(x^2),
      stats::reformulate(".", intercept = !fit.force_through_origin))

    if (mode == "interpolation") {
      y <- object[1, 2]
      lower <- 0
    } else if (mode == "extrapolation") {
      y <- 0
      lower <- -1e06
    }
    upper <- max(object[, 1]) * 1.5

    .fit_qdr_model <- function(model, data, y) {
      fit <- stats::lm(model, data = data, weights = fit.weights)

      ## solve and get De
      success <- TRUE
      if (mode != "alternate") {
        De.fs <- function(fit, x, y) {
          predict(fit, newdata = data.frame(x)) - y
        }
        De.uniroot <- try(uniroot(De.fs, fit = fit, y = y,
                                  lower = lower, upper = upper),
                          silent = TRUE)

        success <- !inherits(De.uniroot, "try-error")
        if (success) {
          De <- De.uniroot$root
        }
      }
      return(list(fit = fit, De = De, success = success))
    }

    res <- .fit_qdr_model(model.qdr, data, y)
    fit <- res$fit
    De <- res$De
    if (res$success)
      .report_fit(De)
    else
      .report_fit_failure(fit.method, mode)

    ##set progressbar
    if(txtProgressBar){
      cat("\n\t Run Monte Carlo loops for error estimation of the QDR fit\n")
      pb <- txtProgressBar(min=0,max=n.MC, char="=", style=3)
    }

    ## Monte Carlo Error estimation
    x.natural <- vapply(1:n.MC, function(i) {
      if (txtProgressBar) setTxtProgressBar(pb, i)
      .fit_qdr_model(
        model = model.qdr,
        data = list(x = xy$x, y = data.MC[, i]),
        y = data.MC.De[i])$De
    }, numeric(1))

    if(txtProgressBar) close(pb)
  }

  ## EXP --------------------------------------------------------------------
  if (fit.method %in% c("EXP", "EXP OR LIN", "LIN")) {
    if(fit.method != "LIN"){
      if (anyNA(c(a, b, c))) {
        .throw_message("Fit ", fit.method, " (", mode,
                       ") could not be applied to this data set, NULL returned")
        return(NULL)
      }

      ##FITTING on GIVEN VALUES##
      ##try to create some start parameters from the input values to make
      ## the fitting more stable

      ## prepare what we can outside the loop
      a.start <-  b.start <- c.start <- numeric(length(a.MC))
      lower_bounds <- c(a = 0, b = 1e-6, c = 0)
      control_settings <-  minpack.lm::nls.lm.control(
        maxiter = 500)

      ## loop for better attempt
      for (i in seq_along(a.MC)) {
        ## get start list
        start_list <- list(a = a.MC[i], b = b.MC[i], c = c.MC[i])

        ## run fit
        fit.initial <- suppressWarnings(try(minpack.lm::nlsLM(
          formula = y ~ fit_functionEXP_cpp(a, b, c, x),
          data = data,
          start = start_list,
          trace = FALSE,
          algorithm = "LM",
          lower = lower_bounds,
          control = control_settings)
        , silent = TRUE))

        if(!inherits(fit.initial, "try-error")){
          #get parameters out of it
          parameters <- coef(fit.initial)
          a.start[i] <- as.vector(parameters["a"])
          b.start[i] <- as.vector(parameters["b"])
          c.start[i] <- as.vector(parameters["c"])
        }
      }

      ##used median as start parameters for the final fitting
      a <- median(a.start, na.rm = TRUE)
      b <- median(b.start, na.rm = TRUE)
      c <- median(c.start, na.rm = TRUE)
      ## exception: if b is 1 it is likely to be wrong and should be reset
      if(!is.na(b) && b == 1)
        b <- mean(b.MC) # nocov

      ## set boundaries
      lower <- if (fit.bounds) c(0, 0, 0) else c(-Inf, -Inf, -Inf)
      upper <- if (fit.force_through_origin) c(Inf, Inf, 0) else c(Inf, Inf, Inf)

      #FINAL Fit curve on given values
      fit <- try(minpack.lm::nlsLM(
        formula = y ~ fit_functionEXP_cpp(a, b, c, x),
        data = data,
        start = list(a = a, b = b, c = 0),
        weights = fit.weights,
        trace = FALSE,
        algorithm = "LM",
        lower = lower,
        upper = upper,
        control = minpack.lm::nls.lm.control(maxiter = 500)
      ), silent = TRUE)

      if (inherits(fit, "try-error") & inherits(fit.initial, "try-error")){
        .report_fit_failure(fit.method, mode)

      }else{
        ##this is to avoid the singular convergence failure due to a perfect fit at the beginning
        ##this may happen especially for simulated data
        if(inherits(fit, "try-error") & !inherits(fit.initial, "try-error")){
          fit <- fit.initial
          rm(fit.initial)
        }

        ## replace with formula so that we can have the C++ version
        f <- function(x) .toFormula(fit.functionEXP, env = currn_env)
        fit$m$formula <- f

        #get parameters out of it
        .get_coef(fit)

        #calculate De
        De <- NA
        if(mode == "interpolation"){
          De <- suppressWarnings(-c - b * log(1 - object[1, 2] / a))
        }else if (mode == "extrapolation"){
          De <- suppressWarnings(-c-b*log(1-0/a))
        }

        #print D01 value
        D01 <- b

        .report_fit(De, " | D01 = ", round(D01, 2))

        #EXP MC -----
        ##Monte Carlo Simulation
        #	--Fit many curves and calculate a new De +/- De_Error
        #	--take De_Error

        ## preallocate variable
        var.b <- vector(mode="numeric", length=n.MC)

        #start loop
        for (i in 1:n.MC) {
          fit.MC <- try(minpack.lm::nlsLM(
            formula = y ~ fit_functionEXP_cpp(a, b, c, x),
            data = list(x = xy$x,y = data.MC[,i]),
            start = list(a = a, b = b, c = c),
            weights = fit.weights,
            trace = FALSE,
            algorithm = "LM",
            lower = lower,
            upper = upper,
            control = minpack.lm::nls.lm.control(maxiter = 500)
          ), silent = TRUE
          )

          #get parameters out of it including error handling
          if (!inherits(fit.MC, "try-error") & mode != "alternate") {
            #get parameters out
            parameters <- coef(fit.MC)
            var.a <- as.numeric(parameters["a"]) # Imax
            var.b[i] <- as.numeric(parameters["b"]) # D0
            var.c <- as.numeric(parameters["c"])

            #calculate x.natural for error calculation
            x.natural[i] <- suppressWarnings(
                -var.c - var.b[i] * log(1 - data.MC.De[i] / var.a))
          }

        }#end for loop

        ##write D01.ERROR
        D01.ERROR <- sd(var.b, na.rm = TRUE)

        ##remove values
        rm(var.b)

      }#endif::try-error fit
    }#endif:fit.method!="LIN"

    ## LIN ------------------------------------------------------------------
    ##two options: just linear fit or LIN fit after the EXP fit failed

    #set fit object, if fit object was not set before
    if (!exists("fit")) fit <- NA

    if ((fit.method=="EXP OR LIN" & inherits(fit, "try-error")) |
        fit.method == "LIN") {

      ## establish models without and with intercept term
      model.lin <- stats::update(y ~ x,
                          stats::reformulate(".", intercept = !fit.force_through_origin))

      if (fit.force_through_origin)
        De.fs <- function(fit, y) y / coef(fit)[1]
      else
        De.fs <- function(fit, y) (y - coef(fit)[1]) / coef(fit)[2]

      y <- object[1, 2]
      if (mode == "extrapolation")
        y <- 0

      .fit_lin_model <- function(model, data, y) {
        fit <- stats::lm(model, data = data, weights = fit.weights)

        ## solve and get De
        De <- NA
        if (mode != "alternate")
          De <- De.fs(fit, y)

        return(list(fit = fit, De = unname(De)))
      }

      res <- .fit_lin_model(model.lin, data, y)
      fit.lm <- res$fit
      De <- res$De
      .report_fit(De)

      ## Monte Carlo Error estimation
      x.natural <- vapply(1:n.MC, function(i) {
        .fit_lin_model(
          model = model.lin,
          data = list(x = xy$x, y = data.MC[, i]),
          y = data.MC.De[i])$De
      }, numeric(1))

      #correct for fit.method
      fit.method <- "LIN"

      ##set fit object
      if(fit.method == "LIN") fit <- fit.lm

    }else{fit.method<-"EXP"}#endif::LIN
  }#end if EXP (this includes the LIN fit option)

  ## EXP+LIN ----------------------------------------------------------------
  else if (fit.method=="EXP+LIN") {
    ## set boundaries
    lower <- if (fit.bounds) c(0, 10, 0, 0) else rep(-Inf, 4)
    upper <- if (fit.force_through_origin) c(Inf, Inf, 0, Inf) else rep(Inf, 4)

    ##try some start parameters from the input values to makes the fitting more stable
    for (i in seq_along(a.MC)) {
      a <- a.MC[i]
      b <- b.MC[i]
      c <- c.MC[i]
      g <- max(0, g.MC[i])

      ##---------------------------------------------------------##
      ##start: with EXP function
      fit.EXP <- try({
        suppressWarnings(minpack.lm::nlsLM(
        formula = y ~ fit_functionEXP_cpp(a, b, c, x),
        data = data,
        start = c(a=a,b=b,c=c),
        trace = FALSE,
        algorithm = "LM",
        lower = c(a = 0, b = 10, c = 0),
        control = minpack.lm::nls.lm.control(
          maxiter=100)
      ))},
      silent=TRUE)

      if(!inherits(fit.EXP, "try-error")){
        #get parameters out of it
        .get_coef(fit.EXP)

        ##end: with EXP function
        ##---------------------------------------------------------##
      }

      fit <- try({
        suppressWarnings(minpack.lm::nlsLM(
          formula = y ~ fit_functionEXPLIN_cpp(a, b, c, g, x),
          data = data,
          start = c(a=a,b=b,c=c,g=g),
          trace = FALSE,
          algorithm = "LM",
          lower = lower,
          control = minpack.lm::nls.lm.control(
            maxiter = 500) #increase max. iterations
          ))
        }, silent=TRUE)

      if(!inherits(fit, "try-error")){
        #get parameters out of it
        parameters <- coef(fit)
        a.start[i] <- parameters[["a"]]
        b.start[i] <- parameters[["b"]]
        c.start[i] <- parameters[["c"]]
        g.start[i] <- parameters[["g"]]
      }
    }##end for loop

    ## used mean as start parameters for the final fitting
    a <- median(a.start, na.rm = TRUE)
    b <- median(b.start, na.rm = TRUE)
    c <- median(c.start, na.rm = TRUE)
    g <- median(g.start, na.rm = TRUE)

    ##perform final fitting
    fit <- try(suppressWarnings(minpack.lm::nlsLM(
      formula = y ~ fit_functionEXPLIN_cpp(a, b, c, g, x),
      data = data,
      start = list(a = a, b = b,c = c, g = g),
      weights = fit.weights,
      trace = FALSE,
      algorithm = "LM",
      lower = lower,
      upper = upper,
      control = minpack.lm::nls.lm.control(maxiter = 500)
    )), silent = TRUE)

    #if try error stop calculation
    if(!inherits(fit, "try-error")){
      ## replace with formula so that we can have the C++ version
      f <- function(x) .toFormula(fit.functionEXPLIN, env = currn_env)
      fit$m$formula <- f

      #get parameters out of it
      .get_coef(fit)

      #problem: analytically it is not easy to calculate x,
      #use uniroot to solve that problem ... readjust function first
      f.unirootEXPLIN <- function(a, b, c, g, x, LnTn) {
        fit_functionEXPLIN_cpp(a, b, c, g, x) - LnTn
      }

      if (mode == "interpolation") {
        LnTn <- object[1, 2]
        min.val <- 0
      } else if (mode == "extrapolation") {
        LnTn <- 0
        min.val <- -1e6
      }

      De <- NA
      if (mode != "alternate") {
        temp.De <-  try(uniroot(
          f = f.unirootEXPLIN,
          interval = c(min.val, max(xy$x) * 1.5),
          tol = 0.001,
          a = a,
          b = b,
          c = c,
          g = g,
          LnTn = LnTn,
          extendInt = "yes",
          maxiter = 3000
        ),
        silent = TRUE)

        if (!inherits(temp.De, "try-error"))
          De <- temp.De$root

        .report_fit(De)
      }

      ##Monte Carlo Simulation for error estimation
      #	--Fit many curves and calculate a new De +/- De_Error
      #	--take De_Error

      ##set progressbar
      if(txtProgressBar){
        cat("\n\t Run Monte Carlo loops for error estimation of the EXP+LIN fit\n")
        pb <- txtProgressBar(min=0,max=n.MC, char="=", style=3)
      }

      ## start Monte Carlo loops
      for(i in  1:n.MC){
        ##perform MC fitting
        fit.MC <- try(suppressWarnings(minpack.lm::nlsLM(
          formula = y ~ fit_functionEXPLIN_cpp(a, b, c, g, x),
          data = list(x=xy$x,y=data.MC[,i]),
          start = list(a = a, b = b,c = c, g = g),
          weights = fit.weights,
          trace = FALSE,
          algorithm = "LM",
          lower = lower,
          control = minpack.lm::nls.lm.control(maxiter = 500)
        )), silent = TRUE)

        #get parameters out of it including error handling
        if (!inherits(fit.MC, "try-error")) {
          .get_coef(fit.MC, pre = "var.")

          min.val <- 0
          if (mode == "extrapolation")
            min.val <- -1e6

          #problem: analytically it is not easy to calculate x,
          #use uniroot to solve this problem
          temp.De.MC <- try(uniroot(
              f = f.unirootEXPLIN,
              interval = c(min.val, max(xy$x) * 1.5),
              tol = 0.001,
              a = var.a,
              b = var.b,
              c = var.c,
              g = var.g,
              LnTn = data.MC.De[i]
            ),
            silent = TRUE)

          if (!inherits(temp.De.MC, "try-error")) {
            x.natural[i] <- temp.De.MC$root
          }
        }
        ##update progress bar
        if(txtProgressBar) setTxtProgressBar(pb, i)

      }#end for loop

      ##close
      if(txtProgressBar) close(pb)

    }else{
      .report_fit_failure(fit.method, mode)
    } #end if "try-error" Fit Method
  } #End if EXP+LIN

  ## EXP+EXP ----------------------------------------------------------------
  else if (fit.method == "EXP+EXP") {
    ## initialise objects
    a1.start <-  a2.start <- b1.start <- b2.start <- NA

    ## try to create some start parameters from the input values to make the fitting more stable
    for (i in seq_along(a.MC)) {
      a1 <- a.MC[i]; a2 <- a1 / 2
      b1 <- b.MC[i]; b2 <- b1 / 2

      fit.start <- try({
        minpack.lm::nlsLM(
        formula = y ~ fit_functionEXPEXP_cpp(a1, a2, b1, b2, x),
        data = data,
        start = list(a1 = a1,a2 = a2,b1 = b1,b2 = b2),
        trace = FALSE,
        algorithm = "LM",
        lower = c(a1 = 1e-6, a2 = 1e-6, b1 = 1e-6, b2 = 1e-6),
        control = minpack.lm::nls.lm.control(maxiter = 500))
      }, silent = TRUE)

      if (!inherits(fit.start, "try-error")) {
        #get parameters out of it
        parameters <- coef(fit.start)
        a1.start[i] <- as.vector((parameters["a1"]))
        b1.start[i] <- as.vector((parameters["b1"]))
        a2.start[i] <- as.vector((parameters["a2"]))
        b2.start[i] <- as.vector((parameters["b2"]))
      }
    }

    ##use obtained parameters for fit input
    a1.start <- median(a1.start, na.rm = TRUE)
    b1.start <- median(b1.start, na.rm = TRUE)
    a2.start <- median(a2.start, na.rm = TRUE)
    b2.start <- median(b2.start, na.rm = TRUE)

    ## set fit bounds
    lower <- if (fit.bounds) rep(0, 4) else rep(-Inf, 4)

    ##perform final fitting
    fit <- try(minpack.lm::nlsLM(
      formula = .toFormula(fit.functionEXPEXP, env = currn_env),
      data = data,
      start = list(a1 = a1, b1 = b1, a2 = a2, b2 = b2),
      weights = fit.weights,
      trace = FALSE,
      algorithm = "LM",
      lower = lower,
      control = minpack.lm::nls.lm.control(maxiter = 500)
    ), silent = TRUE)

    ##insert if for try-error
    if (!inherits(fit, "try-error")) {
      #get parameters out of it
      .get_coef(fit)

      ##set D0 values
      D01 <- round(b1,digits = 2)
      D02 <- round(b2,digits = 2)

      #problem: analytically it is not easy to calculate x, use uniroot
      De <- NA
      if (mode == "interpolation") {
        f.unirootEXPEXP <-
          function(a1, a2, b1, b2, x, LnTn) {
            fit_functionEXPEXP_cpp(a1, a2, b1, b2, x) - LnTn
          }

        temp.De <-  try(uniroot(
          f = f.unirootEXPEXP,
          interval = c(0, max(xy$x) * 1.5),
          tol = 0.001,
          a1 = a1,
          a2 = a2,
          b1 = b1,
          b2 = b2,
          LnTn = object[1, 2],
          extendInt = "yes",
          maxiter = 3000
        ),
        silent = TRUE)

        if (!inherits(temp.De, "try-error")) {
          De <- temp.De$root
        }

        ##remove object
        rm(temp.De)

      }else if (mode == "extrapolation"){
        .throw_error("Mode 'extrapolation' for fitting method 'EXP+EXP' ",
                     "not supported")
      }

      #print D0 and De value values
      .report_fit(De, " | D01 = ", D01, " | D02 = ", D02)

      ##Monte Carlo Simulation for error estimation
      #	--Fit many curves and calculate a new De +/- De_Error
      #	--take De_Error from the simulation
      # --comparison of De from the MC and original fitted De gives a value for quality

      ##progress bar
      if(txtProgressBar){
        cat("\n\t Run Monte Carlo loops for error estimation of the EXP+EXP fit\n")
        pb <- txtProgressBar(min=0,max=n.MC, initial=0, char="=", style=3)
      }

      #set variables
      var.b1 <- var.b2 <- vector(mode="numeric", length=n.MC)

      ## start Monte Carlo loops
      for (i in 1:n.MC) {
        #update progress bar
        if(txtProgressBar) setTxtProgressBar(pb,i)

        ##perform final fitting
        fit.MC <- try(minpack.lm::nlsLM(
          formula = y ~ fit_functionEXPEXP_cpp(a1, a2, b1, b2, x),
          data = list(x=xy$x,y=data.MC[,i]),
          start = list(a1 = a1, b1 = b1, a2 = a2, b2 = b2),
          weights = fit.weights,
          trace = FALSE,
          algorithm = "LM",
          lower = lower,
          control = minpack.lm::nls.lm.control(maxiter = 500)
        ), silent = TRUE)

        #get parameters out of it including error handling
        if (!inherits(fit.MC, "try-error")) {
          parameters <- coef(fit.MC)
          var.a1 <- as.numeric(parameters["a1"])
          var.a2 <- as.numeric(parameters["a2"])
          var.b1[i] <- as.vector((parameters["b1"]))
          var.b2[i] <- as.vector((parameters["b2"]))

          #problem: analytically it is not easy to calculate x, here an simple approximation is made
          temp.De.MC <-  try(uniroot(
            f = f.unirootEXPEXP,
            interval = c(0,max(xy$x) * 1.5),
            tol = 0.001,
            a1 = var.a1,
            a2 = var.a2,
            b1 = var.b1[i],
            b2 = var.b2[i],
            LnTn = data.MC.De[i]
          ), silent = TRUE)

          if (!inherits(temp.De.MC, "try-error"))
            x.natural[i] <- temp.De.MC$root

        } #end if "try-error" MC simulation
      } #end for loop

      ##write D01.ERROR
      D01.ERROR <- sd(var.b1, na.rm = TRUE)
      D02.ERROR <- sd(var.b2, na.rm = TRUE)

      ##remove values
      rm(var.b1, var.b2)

    }else{
      .report_fit_failure(fit.method, mode)
    } #end if "try-error" Fit Method

    ##close
    if (txtProgressBar && exists("pb")) close(pb)
  }

  ## GOK --------------------------------------------------------------------
  else if (fit.method[1] == "GOK") {
    ## set bounds
    lower <- if (fit.bounds) rep(0, 4) else rep(-Inf, 4)
    upper <- if (fit.force_through_origin) c(Inf, Inf, Inf, 1) else rep(Inf, 4)

    fit <- try(minpack.lm::nlsLM(
      formula = .toFormula(fit.functionGOK, env = currn_env),
      data = data,
      start = list(a = a, b = b, c = 1, d = 1),
      weights = fit.weights,
      trace = FALSE,
      algorithm = "LM",
      lower = lower,
      upper = upper,
      control = minpack.lm::nls.lm.control(maxiter = 500)
    ), silent = TRUE)

    if (inherits(fit, "try-error")){
      .report_fit_failure(fit.method, mode)

    }else{
      #get parameters out of it
      .get_coef(fit)

      #calculate De
      y <- object[1, 2]
      De <- switch(
        mode,
        interpolation = suppressWarnings(
          -(b * (( (a * d - y)/a)^c - 1) * ( ((a * d - y)/a)^-c  )) / c),
        extrapolation = suppressWarnings(
          -(b * (( (a * d - 0)/a)^c - 1) * ( ((a * d - 0)/a)^-c  )) / c),
        NA)

      #print D01 value
      D01 <- b

      .report_fit(De, " | D01 = ", round(D01, 2), " | c = ", round(c, 2))

      #EXP MC -----
      ##Monte Carlo Simulation
      #	--Fit many curves and calculate a new De +/- De_Error
      #	--take De_Error

      ## preallocate variable
      var.b <- vector(mode = "numeric", length = n.MC)

      #start loop
      for (i in 1:n.MC) {
        ##set data set
        fit.MC <- try({
          minpack.lm::nlsLM(
          formula = y ~ fit_functionGOK_cpp(a, b, c, d, x),
          data = list(x = xy$x,y = data.MC[,i]),
          start = list(a = a, b = b, c = 1, d = 1),
          weights = fit.weights,
          trace = FALSE,
          algorithm = "LM",
          lower = lower,
          upper = upper,
          control = minpack.lm::nls.lm.control(maxiter = 500)
        )}, silent = TRUE)

        # get parameters out of it including error handling
        if (!inherits(fit.MC, "try-error") && mode != "alternate") {
          # get parameters out
          parameters<-coef(fit.MC)
          var.a <- as.numeric(parameters["a"]) #Imax
          var.b[i] <- as.numeric(parameters["b"]) #D0
          var.c <- as.numeric(parameters["c"]) #kinetic order modifier
          var.d <- as.numeric(parameters["d"]) #origin

          # calculate x.natural for error calculation
          ## note that data.MC.De contains only 0s for extrapolation
          temp <- (var.a * var.d - data.MC.De[i]) / var.a
          x.natural[i] <- suppressWarnings(-var.b[i] * (1 - temp^-var.c) / var.c)
        }

      }#end for loop

      ##write D01.ERROR
      D01.ERROR <- sd(var.b, na.rm = TRUE)

      ##remove values
      rm(var.b)
    }
  }

  ## OTOR ---------------------------------------------------------------
  else if (fit.method == "OTOR") {
    Dint_lower <- 0.01
    if(mode == "extrapolation")
      Dint_lower <- 50 ##TODO - fragile ... however it is only used by a few

    ## set bounds
    lower <- if (fit.bounds) c(0, 0, 0, Dint_lower) else rep(-Inf, 4)
    upper <- if (fit.force_through_origin) c(10, Inf, Inf, 0) else c(10, Inf, Inf, Inf)

    fit <- try(minpack.lm::nlsLM(
          formula = .toFormula(fit.functionOTOR, env = currn_env),
          data = data,
          start = list(R = 0, Dc = b, N = b, Dint = 0),
          weights = fit.weights,
          trace = FALSE,
          algorithm = "LM",
          lower = lower,
          upper = upper,
          control = minpack.lm::nls.lm.control(maxiter = 500)
        ), silent = TRUE)

    if (inherits(fit, "try-error")) {
      .report_fit_failure(fit.method, mode)

    } else {
          #get parameters out of it
         .get_coef(fit)

          #calculate De
          De <- NA
          if(mode == "interpolation"){
             De <- try(suppressWarnings(stats::uniroot(
               f = function(x, R, Dc, N, Dint, LnTn) {
                 fit.functionOTOR(R, Dc, N, Dint, x) - LnTn},
               interval = c(0, max(object[[1]]) * 1.2),
               R = R,
               Dc = Dc,
               N = N,
               Dint = Dint,
               LnTn = object[1, 2])$root), silent = TRUE)

          }else if (mode == "extrapolation"){
            De <- try(suppressWarnings(stats::uniroot(
              f = function(x, R, Dc, N, Dint) {
                fit.functionOTOR(R, Dc, N, Dint, x)},
              interval = c(-max(object[[1]]), 0),
              R = R,
              Dc = Dc,
              N = N,
              Dint = Dint)$root), silent = TRUE)

            ## there are cases where the function cannot calculate the root
            ## due to its shape, here we have to use the minimum
            if(inherits(De, "try-error")){
              .throw_warning(
                  "Standard root estimation using stats::uniroot() failed. ",
                  "Using stats::optimize() instead, which may lead, however, ",
                  "to unexpected and inconclusive results for fit.method = 'OTOR'")

              De <- try(suppressWarnings(stats::optimize(
                f = function(x, R, Dc, N, Dint) {
                  fit.functionOTOR(R, Dc, N, Dint, x)},
                interval = c(-max(object[[1]]), 0),
                R = R,
                Dc = Dc,
                N = N,
                Dint = Dint)$minimum), silent = TRUE)
            }
          }

          if (inherits(De, "try-error")) De <- NA # nocov

          .report_fit(De, " | R = ", round(R, 2), " | Dc = ", round(Dc, 2))

          #OTOR MC -----
          ##Monte Carlo Simulation
          #	--Fit many curves and calculate a new De +/- De_Error
          #	--take De_Error
          #set variables
          var.Dc <- var.R <- vector(mode = "numeric", length = n.MC)

          #start loop
          for (i in 1:n.MC) {
            ##set data set
            fit.MC <- try(minpack.lm::nlsLM(
              formula = .toFormula(fit.functionOTOR, env = currn_env),
              data = list(x = xy$x,y = data.MC[,i]),
              start = list(R = 0, Dc = b, N = 0, Dint = 0),
              weights = fit.weights,
              trace = FALSE,
              algorithm = "LM",
              lower = if (fit.bounds) c(0, 0, 0, Dint*runif(1,0,2)) else c(-Inf,-Inf,-Inf, -Inf),
              upper = upper,
              control = minpack.lm::nls.lm.control(maxiter = 500)
            ), silent = TRUE)

            # get parameters out of it including error handling
            if (!inherits(fit.MC, "try-error")) {
              # get parameters out
              parameters <- coef(fit.MC)
              var.R[i] <- as.numeric(parameters["R"])
              var.Dc[i] <- as.numeric(parameters["Dc"])
              var.N <- as.numeric(parameters["N"])
              var.Dint <- as.numeric(parameters["Dint"])

              # calculate x.natural for error calculation
              if(mode == "interpolation"){
                try <- try({
                  suppressWarnings(stats::uniroot(
                  f = function(x, R, Dc, N, Dint, LnTn) {
                    fit.functionOTOR(R, Dc, N, Dint, x) - LnTn},
                  interval = c(0, max(object[[1]]) * 1.2),
                  R = var.R[i],
                  Dc = var.Dc[i],
                  N = var.N,
                  Dint = var.Dint,
                  LnTn = data.MC.De[i])$root)
                }, silent = TRUE)

              } else if (mode == "extrapolation"){
                try <- try(
                  suppressWarnings(stats::uniroot(
                    f = function(x, R, Dc, N, Dint) {
                      fit.functionOTOR(R, Dc, N, Dint, x)},
                    interval = c(-max(object[[1]]), 0),
                    R = var.R[i],
                    Dc = var.Dc[i],
                    N = var.N,
                    Dint = var.Dint)$root),
                  silent = TRUE)

                if(inherits(try, "try-error")){
                  try <- try(suppressWarnings(stats::optimize(
                    f = function(x, R, Dc, N, Dint) {
                      fit.functionOTOR(R, Dc, N, Dint, x)},
                    interval = c(-max(object[[1]]), 0),
                    R = var.R[i],
                    Dc = var.Dc[i],
                    N = var.N,
                    Dint = var.Dint)$minimum),
                    silent = TRUE)
                }
              }##endif extrapolation
              if (!inherits(try, c("try-error", "function")))
                x.natural[i] <- try
            }

          }#end for loop

          ##write Dc.ERROR
          Dc.ERROR <- sd(var.Dc, na.rm = TRUE)
          R.ERROR <- sd(var.R, na.rm = TRUE)

          ##remove values
          rm(var.Dc)
          rm(var.R)

    }#endif::try-error fit

  }  ## OTORX ---------------------------------------------------------------
  else if (fit.method == "OTORX") {
    if(is.null(object$Test_Dose) || all(object$Test_Dose == -1))
      .throw_error("Column 'Test_Dose' missing but mandatory for 'OTORX' fitting!")

    ## we need a test dose; the default value is -1 because an NA will cause
    ## additional problems
      TEST_DOSE <- object$Test_Dose[[1]]

      ## here we replace TEST_DOSE by an evaluated value
      ## in the function body; this makes things ALOT easier below
      body(fit.functionOTORX) <- do.call(
        substitute, list(body(fit.functionOTORX), list(TEST_DOSE = TEST_DOSE)))

    ## set boundaries
    lower <- if (fit.bounds) c(0, 0, 0, 0) else rep(-Inf, 4)
    upper <- c(Inf, Inf, Inf, Inf)

      ## correct boundaries for origin forced through zero
      if (fit.force_through_origin[1] & mode == "interpolation")
        lower[4] <- upper[4] <- 0

    fit <- try(minpack.lm::nlsLM(
      formula = .toFormula(fit.functionOTORX, env = currn_env),
      data = data,
      start = list(Q = 1, D63 = b, c = 1, a = 1),
      weights = fit.weights,
      trace = FALSE,
      algorithm = "LM",
      lower = lower,
      upper = upper,
      control = minpack.lm::nls.lm.control(maxiter = 500)
    ), silent = TRUE)

    if (inherits(fit, "try-error")) {
      .report_fit_failure(fit.method, mode)

    } else {
      #get parameters out of it
      .get_coef(fit)

      ## get also R, this is not part of the fit
      R <- 1-Q

      #calculate De
      De <- NA
      if(mode == "interpolation"){
        De <- try(suppressWarnings(stats::uniroot(
          f = function(x, Q, D63, c, a, LnTn) {
            fit.functionOTORX(x, Q, D63, c, a) - LnTn},
          interval = c(0, max(object[[1]]) * 1.2),
          Q = Q,
          D63 = D63,
          c = c,
          a = a,
          LnTn = object[1, 2])$root), silent = TRUE)

      }else if (mode == "extrapolation"){
        De <- try(suppressWarnings(stats::uniroot(
          f = function(x, Q, D63, c, a) {
            fit.functionOTORX(x, Q, D63, c, a)},
          interval = c(-max(object[[1]]), 0),
          Q = Q,
          D63 = D63,
          c = c,
          a = a)$root), silent = TRUE)

        ## there are cases where the function cannot calculate the root
        ## due to its shape, here we have to use the minimum
        if(inherits(De, "try-error")){
          .throw_warning(
            "Standard root estimation using stats::uniroot() failed. ",
            "Using stats::optimize() instead, which may lead, however, ",
            "to unexpected and inconclusive results for fit.method = 'OTORX'")

          De <- try(suppressWarnings(stats::optimize(
            f = function(x, Q, D63, c, a) {
              fit.functionOTORX(x, Q, D63, c, a)},
            interval = c(-max(object[[1]]), 0),
            Q = Q,
            D63 = D63,
            c = c,
            a = a)$minimum), silent = TRUE)
        }
      }

      if (inherits(De, "try-error")) De <- NA # nocov

      .report_fit(De, " | R = ", round(1-Q, 2), " | D63 = ", round(D63, 2))

      #OTORX MC -----
      ##Monte Carlo Simulation
      #	--Fit many curves and calculate a new De +/- De_Error
      #	--take De_Error
      #set variables
      var.D63 <- var.Q <- vector(mode = "numeric", length = n.MC)

      #start loop
      for (i in 1:n.MC) {
        ##set data set
        fit.MC <- try(minpack.lm::nlsLM(
          formula = .toFormula(fit.functionOTORX, env = currn_env),
          data = list(x = xy$x,y = data.MC[,i]),
          start = list(Q = 1, D63 = b, c = 1, a = 1),
          weights = fit.weights,
          trace = FALSE,
          algorithm = "LM",
          lower = lower,
          upper = upper,
          control = minpack.lm::nls.lm.control(maxiter = 500)
        ), silent = TRUE)

        # get parameters out of it including error handling
        if (!inherits(fit.MC, "try-error")) {
          # get parameters out
          parameters<-coef(fit.MC)
          var.Q[i] <- as.numeric(parameters["Q"])
          var.D63[i] <- as.numeric(parameters["D63"])
          var.c <- as.numeric(parameters["c"])
          var.a <- as.numeric(parameters["a"])

          # calculate x.natural for error calculation
          if(mode == "interpolation"){
            try <- try(
              {suppressWarnings(stats::uniroot(
                f = function(x, Q, D63, c, a, LnTn) {
                  fit.functionOTORX(x, Q, D63, c, a) - LnTn},
                interval = c(0, max(object[[1]]) * 1.2),
                Q = var.Q[i],
                D63 = var.D63[i],
                c = var.c,
                a = var.a,
                LnTn = data.MC.De[i])$root)
              }, silent = TRUE)

          }else if(mode == "extrapolation"){
            try <- try(
              suppressWarnings(stats::uniroot(
                f = function(x, Q, D63, c, a, LnTn) {
                  fit.functionOTORX(x, Q, D63, c, a, x)},
                interval = c(-max(object[[1]]), 0),
                Q = var.Q[i],
                D63 = var.D63[i],
                c = var.c,
                a = var.a)$root),
              silent = TRUE)

            if(inherits(try, "try-error")){
              try <- try(suppressWarnings(stats::optimize(
                f = function(x, Q, D63, c, a) {
                  fit.functionOTOR(x, Q, D63, c, a)},
                interval = c(-max(object[[1]]), 0),
                Q = var.Q[i],
                D63 = var.D63[i],
                c = var.c,
                a = var.a)$minimum),
                silent = TRUE)
            }
          }##endif extrapolation
          if (!inherits(try, c("try-error", "function")))
            x.natural[i] <- try
        }
      }#end for loop

      ##write Dc.ERROR
      D63.ERROR <- sd(var.D63, na.rm = TRUE)
      R.ERROR <- sd(1-var.Q, na.rm = TRUE)

      ##remove values
      rm(var.D63)
      rm(var.Q)

    }#endif::try-error fit
  }#End if fit.method selection (for all)

  ## get De values from Monte Carlo simulation
  De.MC <- De.MC.NA <- x.natural
  if (mode == "interpolation") {
    ## censor negative values
    De.MC <- pmax(x.natural, 0)
    De.MC.NA[x.natural < 0] <- NA
  } else if (mode == "extrapolation") {
    ## always return positive values
    De.MC <- De.MC.NA <- x.natural <- abs(x.natural)
  }

  ## calculate mean and sd (ignore NaN values)
  De.MonteCarlo <- mean(De.MC, na.rm = TRUE)

  #De.Error is Error of the whole De (ignore NaN values)
  De.Error <- sd(De.MC.NA, na.rm = TRUE)

  # Formula creation --------------------------------------------------------
  ## This information is part of the fit object output anyway, but
  ## we keep it here for legacy reasons
  fit_formula <- NA
  if(!inherits(fit, "try-error") && !is.na(fit[1]))
    fit_formula <- .replace_coef(fit)

# Output ------------------------------------------------------------------
  ##calculate HPDI
  HPDI <- matrix(c(NA,NA,NA,NA), ncol = 4)
  ## here we use the original x.natural because we need the entire
  ## distribution of De values, not a censored one
  if (sum(!is.na(x.natural)) >= 5) {
    HPDI <- cbind(
        .calc_HPDI(x.natural, prob = 0.68, na.rm = TRUE)[1, , drop = FALSE],
        .calc_HPDI(x.natural, prob = 0.95, na.rm = TRUE)[1, , drop = FALSE])
  }

  ## calculate the n/N value (the relative saturation level)
  ## the absolute intensity is the integral of curve
      ## define the function
      f_int <- function(x) eval(fit_formula)

      ## run integrations (they may fail; so we have to check)
      N <- try({
        suppressWarnings(
          stats::integrate(f_int, lower = 0, upper = max(xy$x, na.rm = TRUE))$value)
      }, silent = TRUE)
      n <- try({
        suppressWarnings(
          stats::integrate(f_int, lower = 0, upper = max(De, na.rm = TRUE))$value)
      }, silent = TRUE)

      if(inherits(N, "try-error") || inherits(n, "try-error"))
        n_N <- NA
      else
        n_N <- n/N

  ## account for the fact that we can still calculate a De that is negative
  ## even it does not make sense for interpolation
  De.raw <- De
  if (mode == "interpolation" && !is.na(De) && De < 0) {
    De <- NA
  }

  ## if fields in this objects are changed, update also `temp.GC.all.na`
  ## in analyse_SAR.CWOSL()
  output <- try(data.frame(
    De = abs(De),
    De.Error = De.Error,
    D01 = D01,
    D01.ERROR = D01.ERROR,
    D02 = D02,
    D02.ERROR = D02.ERROR,
    R = R,
    R.ERROR = R.ERROR,
    Dc = Dc,
    D63 = D63,
    n_N = n_N,
    De.MC = De.MonteCarlo,
    Fit = fit.method,
    Mode = mode,
    HPDI68_L = HPDI[1,1],
    HPDI68_U = HPDI[1,2],
    HPDI95_L = HPDI[1,3],
    HPDI95_U = HPDI[1,4],
    .De.plot = De,    # no absolute value, needed for plot_DoseResposeCurve()
    .De.raw = De.raw, # negative values not set to NA for interpolation
    row.names = NULL
  ), silent = TRUE)

  ##make RLum.Results object
  set_RLum(
    class = "RLum.Results",
    data = list(
      De = output,
      De.MC = De.MC,
      Fit = fit,
      Fit.Args = list(
          object = object,
          fit.method = fit.method,
          mode = mode,
          fit.force_through_origin = fit.force_through_origin,
          fit.includingRepeatedRegPoints = fit.includingRepeatedRegPoints,
          fit.NumberRegPoints = fit.NumberRegPoints,
          fit.NumberRegPointsReal = fit.NumberRegPointsReal,
          fit.weights = fit.weights,
          fit.bounds = fit.bounds,
          n.MC = n.MC
      ),
      Formula = fit_formula
    ),
    info = list(
        fit_message = fit_message,
        call = sys.call()
    )
  )
}

# Helper functions in fit_DoseResponseCurve() -------------------------------------

#'@title Returns coefficient into parent environment
#'
#'@description Write fitting coefficients into parent environment
#'
#'@param x [stats::nls] (**required**): the fitting output
#'
#'@param pre [character] (*with default*): names prefix
#'
#'@param sufx [character] (*with default*): names suffix
#'
#'@returns New objects into the parent environment
#'
#'@noRd
.get_coef <- function(x, pre = "", sufx = "") {
  ## get coefficients and set their names
  tmp <- stats::coef(x)
  names(tmp) <- paste0(pre, names(tmp), sufx)

  ## assign to parent frame
  for (name in names(tmp))
    assign(name, as.vector(tmp[name]), pos = parent.frame())
}

#'@title Replace coefficients in formula
#'
#'@description
#'
#'Replace the parameters in a fitting function by the true, fitted values.
#'This way the results can be easily used by the other functions
#'
#'@param f [nls] or [lm] (**required**): the output object of the fitting
#'
#'@returns Returns an [expression]
#'
#'@noRd
.replace_coef <- function(f) {
  ## get formula as character string
  if(inherits(f, "nls")) {
    str <- as.character(f$m$formula())[3]
    param <- coef(f)
  } else {
    str <- "a * x + b * x^2 + n"
    param <- c(n = 0, a = 0, b = 0)
     if(!"(Intercept)" %in% names(coef(f)))
      param[2:(length(coef(f))+1)] <- coef(f)
    else
      param[1:length(coef(f))] <- coef(f)
  }

  ## if the following assertion is triggered, it means that we have used a C++
  ## function to implement the model but forgot to replace the formula in the
  ## fit object, which can be done with these lines:
  ##   f <- function(x) .toFormula(fit.functionXXX, env = currn_env)
  ##   fit$m$formula <- f
  stopifnot(!startsWith("fit_function", str))

  ## replace parameters with fitted coefficients
  for (i in 1:length(param)) {
    str <- gsub(
      pattern = names(param)[i],
      replacement = format(param[i], digits = 3, scientific = TRUE),
      x = str,
      fixed = TRUE)
  }

  ## return
  return(parse(text = str))
}

#'@title Convert function to formula
#'
#'@description The fitting functions are provided as functions, however, later is
#'easer to work with them as expressions, this functions converts to formula
#'
#'@param f [function] (**required**): function to be converted
#'
#'@param env [environment] (*with default*): environment for the formula
#'creation. This argument is required otherwise it can cause all kind of
#'very complicated to-track-down errors when R tries to access the function
#'stack
#'
#'@noRd
.toFormula <- function(f, env) {
  ## deparse
  tmp <- deparse(f)

  ## set formula
  ## this is very fragile and works only if the functions are constructed
  ## without {} brackets, otherwise it will not work in combination
  ## of covr and testthat
  tmp_formula <- stats::as.formula(paste0("y ~ ", paste(tmp[-1], collapse = "")),
                                   env = env)
  return(tmp_formula)
}

#'@title Convert n/N ratio to Dose
#'
#'@description Helper function for OTORX model fit according to
#'Lawless & Timar-Gabor (2024) Eq. 9
#'
#'@param nN [numeric] (**required**): n/N ratio value
#'
#'@param Q [numeric] (**required**): product of relative production rates and
#'hole pairs (see Lawless & Timar-Gabor, 2024)
#'
#'@param D63 [numeric] (**required**): characteristic dose
#'
#'@references https://github.com/jll2/LumDRC/blob/main/otorx.py
#'
#'@note Not used here, however, part of the reference implementation.
#'
#'@noRd
.nN2D <- function(nN, Q, D63) D63 * ((-log(1-nN) - Q*nN)/(1 - Q*(1-exp(-1)))) # nocov

#'@title Convert Dose back to n/N ratio
#'
#'@description Return n/N for a given dose D and parameters Q & D63.
#'see Lawless & Timar-Gabor (2024)
#'
#'@param D [numeric] (**required**): dose
#'
#'@param Q [numeric] (**required**): product of relative production rates and
#'hole pairs (see Lawless & Timar-Gabor, 2024)
#'
#'@param D63 [numeric] (**required**): characteristic dose
#'
#'@param a [numeric] (**required**): offset parameter
#'
#'@references https://github.com/jll2/LumDRC/blob/main/otorx.py
#'
#'@noRd
.D2nN <- function(D, Q, D63, a) {
  if(all(abs(Q) < 1e-06))
    r <- 1 - exp(-D/D63)
  else if (any(abs(Q) < 1e-06))
    .throw_error("Unsupported zero and non-zero Q in .D2nN()")
  else
    r <- 1 + (lamW::lambertW0(-Q * exp(-Q-(1-Q*(1-1/exp(1))) * (D + a) /D63))) / Q

  return(r)
}

## load data
data(ExampleData.LxTxData, envir = environment())

## odd data set where the calculated De is negative for EXP
df_odd <- data.frame(
    dose = c(
      0, 0, 2.71828182845905, 2.74202785430992,
      2.76598131771852, 2.79014403079188, 2.814517821467, 2.83910453364916,
      2.86390602735137, 2.88892417883514, 2.91416088075237, 2.93961804228855,
      2.96529758930721, 2.99120146449558, 3.01733162751159, 3.04369005513209,
      3.07027874140241, 3.09709969778723, 3.1241549533227, 3.15144655477),
    LxTx = c(0.439, 0.851865307456887, 0.881530377359027, 0.881788046363334,
      0.882047940677405, 0.882310079298982, 0.882574481384203, 0.882841166248859,
      0.883110153369655, 0.883381462385488, 0.883655113098727, 0.883931125476509,
      0.884209519652028, 0.88449031592586, 0.884773534767266, 0.885059196815536,
      0.885347322881313, 0.885637933947948, 0.885931051172847, 0.886226695888845),
    LxTx.error = c(0.029, 1e-04, 0.0393190034426231, 0.0393304962836274,
      0.0393420883803856, 0.0393537805802564, 0.0393655737376631, 0.0393774687141504,
      0.03938946637844, 0.0394015676064878, 0.0394137732815415, 0.0394260842941973,
      0.0394385015424588, 0.0394510259317946, 0.0394636583751979, 0.0394763997932453,
      0.0394892511141562, 0.0395022132738537, 0.0395152872160239, 0.0395284738921779))

test_that("input validation", {
  testthat::skip_on_cran()

  ## object
  expect_error(
      fit_DoseResponseCurve("error"),
      "[fit_DoseResponseCurve()] 'object' should be of class 'data.frame'",
      fixed = TRUE)
  expect_error(fit_DoseResponseCurve(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(fit_DoseResponseCurve(iris[, 1, drop = FALSE]),
               "'object' should have at least 2 columns")
  expect_error(fit_DoseResponseCurve(as.list(LxTxData)),
               "All elements of 'object' should be of class 'data.frame'")

  ## mode
  expect_error(
    fit_DoseResponseCurve(LxTxData, mode = "error"),
    "[fit_DoseResponseCurve()] 'mode' should be one of 'interpolation'",
    fixed = TRUE)
  expect_error(fit_DoseResponseCurve(LxTxData, mode = set_RLum("RLum.Results")),
               "'mode' should be one of 'interpolation', 'extrapolation' or")

  ## fit.method
  expect_error(
    fit_DoseResponseCurve(LxTxData, fit.method = "error"),
    "[fit_DoseResponseCurve()] 'fit.method' should be one of 'LIN', 'QDR'",
    fixed = TRUE)

  ## other arguments
  expect_error(fit_DoseResponseCurve(LxTxData,
                                     fit.force_through_origin = "error"),
               "'fit.force_through_origin' should be a single logical value")
  expect_error(fit_DoseResponseCurve(LxTxData, fit.weights = "error"),
               "'fit.weights' should be a single logical value")
  expect_error(fit_DoseResponseCurve(LxTxData,
                                     fit.includingRepeatedRegPoints = "error"),
               "'fit.includingRepeatedRegPoints' should be a single logical")
  expect_error(fit_DoseResponseCurve(LxTxData,
                                     fit.NumberRegPoints = "error"),
               "'fit.NumberRegPoints' should be a single positive integer value")
  expect_error(fit_DoseResponseCurve(LxTxData, fit.NumberRegPointsReal = "error"),
               "'fit.NumberRegPointsReal' should be a single positive integer value")
  expect_error(fit_DoseResponseCurve(LxTxData,
                                     fit.bounds = "error"),
               "'fit.bounds' should be a single logical value")
  expect_error(fit_DoseResponseCurve(LxTxData,
                                     n.MC = "error"),
               "'n.MC' should be a single positive integer value")
  expect_error(fit_DoseResponseCurve(LxTxData,
                                     verbose = "error"),
               "'verbose' should be a single logical value")

  ## shorten dataframe
  expect_warning(fit_DoseResponseCurve(LxTxData[1:2, ], verbose = FALSE),
                 "Fitting a non-linear least-squares model requires at least 3")
  expect_warning(fit_DoseResponseCurve(LxTxData[1:3, ], fit.method = "GOK",
                                       verbose = FALSE),
                 "Fitting a non-linear least-squares model requires at least 4")

  ## wrong combination of fit.method and mode
  expect_error(
    fit_DoseResponseCurve(LxTxData, fit.method = "EXP+EXP",
                     mode = "extrapolation"),
    "Mode 'extrapolation' for fitting method 'EXP+EXP' not supported",
    fixed = TRUE)

})

test_that("weird LxTx values", {
  testthat::skip_on_cran()

  LxTx <- structure(list(
    Dose = c(0, 250, 500, 750, 1000, 1500, 0, 500, 500),
    LxTx = c(1, Inf, 0, -Inf, Inf, 0, Inf, -0.25, 2),
    LxTx.Error = c(1.58133646008685, Inf, 0, Inf, Inf, 0, Inf, 1.41146256149428, 3.16267292017369)),
    class = "data.frame", row.names = c(NA, -9L))

  ##fit
  SW({
  expect_warning(
    fit_DoseResponseCurve(LxTx[, c("Dose", "LxTx", "LxTx.Error")]),
    "Inf values found, replaced by NA")
  })

  ##all points have the same dose ... error but NULL
  data(ExampleData.LxTxData, envir = environment())
  tmp_LxTx <- LxTxData
  tmp_LxTx$Dose <- 10

  expect_message(expect_null(
      fit_DoseResponseCurve(tmp_LxTx)),
      "Error: All points have the same dose, NULL returned")

  ## check input objects ... matrix
  SW({
  expect_s4_class(
    fit_DoseResponseCurve(as.matrix(LxTxData)),
    class = "RLum.Results")
  })

  ## test case for only two columns
  expect_s4_class(
    suppressWarnings(fit_DoseResponseCurve(LxTxData[,1:2], verbose = FALSE)),
    class = "RLum.Results")

  ## test case with all NA
  tmp_LxTx <- LxTxData
  tmp_LxTx$LxTx <- NA
  expect_message(expect_null(
      suppressWarnings(fit_DoseResponseCurve(tmp_LxTx))),
      "Error: After NA removal, nothing is left from the data set")

  ## test case without TnTx column
  tmp_LxTx <- LxTxData
  tmp_LxTx$TnTx <- NULL
  SW({
  expect_s4_class(
    fit_DoseResponseCurve(tmp_LxTx, verbose = FALSE),
    "RLum.Results")
  })

  ## do not include reg point
  expect_s4_class(
    fit_DoseResponseCurve(
      LxTxData,
      verbose = FALSE,
      fit.includingRepeatedRegPoints = FALSE),
    class = "RLum.Results")

  ## issue 961
  expect_output(fit_DoseResponseCurve(df_odd, fit.method = "QDR"),
                "Fit failed for QDR (interpolation)", fixed = TRUE)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  ## see https://github.com/R-Lum/Luminescence/pull/308
  skip_on_os("windows")
  skip_on_os("mac")

  snapshot.tolerance <- 1.5e-6

  set.seed(1)
  SW({
  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "EXP",
      verbose = FALSE,
      n.MC = 10
    ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "LIN",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "LIN",
      mode = "extrapolation",
      fit.force_through_origin = TRUE,
      verbose = FALSE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "EXP+LIN",
      fit.bounds = FALSE,
      fit.force_through_origin = TRUE,
      verbose = FALSE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "EXP+EXP",
      verbose = TRUE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "QDR",
      verbose = TRUE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "GOK",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "OTOR",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = snapshot.tolerance)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "QDR",
      mode = "extrapolation",
      verbose = TRUE,
      n.MC = 10
  ), tolerance = 2.0e-5)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "QDR",
      mode = "extrapolation",
      fit.force_through_origin = TRUE,
      verbose = TRUE,
      n.MC = 10
  ), tolerance = 5.0e-5)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "LIN",
      mode = "extrapolation",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = 5.0e-5)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      LxTxData,
      fit.method = "GOK",
      mode = "extrapolation",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = 5e-5)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      cbind(LxTxData, Test_Dose = 17),
      fit.method = "OTORX",
      fit.force_through_origin = TRUE,
      mode = "interpolation",
      verbose = TRUE,
      n.MC = 10
  ), tolerance = 5.0e-5)

  ## negative De
  expect_snapshot_RLum(fit_DoseResponseCurve(
      df_odd,
      mode = "interpolation",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = 5.0e-3)

  expect_snapshot_RLum(fit_DoseResponseCurve(
      df_odd,
      mode = "extrapolation",
      verbose = FALSE,
      n.MC = 10
  ), tolerance = 4.0e-2)
  })
})

test_that("additional tests", {
  testthat::skip_on_cran()

  ## self-call
  res <- fit_DoseResponseCurve(
      list(LxTxData, LxTxData),
      fit.method = "LIN",
      verbose = FALSE,
      n.MC = 10)
  expect_type(res, "list")
  expect_length(res, 2)

  set.seed(1)
  SW({
  expect_output(temp_EXP <-
    fit_DoseResponseCurve(
      LxTxData,
      fit.method = "EXP",
      verbose = TRUE,
      n.MC = 10
    ), " | D01 = ", fixed = TRUE)
  temp_LIN <-
    fit_DoseResponseCurve(
      LxTxData,
      fit.method = "LIN",
      verbose = FALSE,
      n.MC = 10
    )
  temp_EXPLIN <-
    fit_DoseResponseCurve(
      LxTxData,
      fit.method = "EXP+LIN",
      verbose = FALSE,
      n.MC = 10
    )
  temp_EXPEXP <-
    fit_DoseResponseCurve(
      LxTxData,
      fit.method = "EXP+EXP",
      verbose = FALSE,
      n.MC = 10
    )
  temp_QDR <-
    fit_DoseResponseCurve(
      LxTxData,
      fit.method = "QDR",
      verbose = FALSE,
      n.MC = 10
    )
  temp_GOK <-
    fit_DoseResponseCurve(
      LxTxData,
      fit.method = "GOK",
      verbose = FALSE,
      n.MC = 10)

  ## force through the origin
  temp_LxTx <-LxTxData
  temp_LxTx$LxTx[[7]] <- 1
  expect_s4_class(fit_DoseResponseCurve(
    temp_LxTx,
    fit.method = "GOK",
    verbose = FALSE,
    n.MC = 10,
    fit.force_through_origin = TRUE
  ), "RLum.Results")
temp_OTOR <-
  fit_DoseResponseCurve(
    LxTxData,
    fit.method = "OTOR",
    verbose = FALSE,
    n.MC = 10
  )
temp_OTORX <-
  fit_DoseResponseCurve(
    cbind(LxTxData, Test_Dose = 17),## we have to set the TEST_DOSE
    fit.method = "OTORX",
    fit.force_through_origin = TRUE,
    verbose = FALSE,
    n.MC = 10
  )
## test reference dataset from
## (https://raw.githubusercontent.com/jll2/LumDRC/refs/heads/main/otorx.py)
LxTxData_alt <- data.frame(
  Dose = c(0, 17, 47, 94, 201, 402, 804, 1.07e+03, 1.68e+03, 3.35e+03, 5.36e+03,
           6.7e+03, 8.04e+03, 1e+04, 47, 94, 3.35e+03, 5.36e+03),
  LxTx = c(10, 1.09, 2.3, 3.67, 5.63, 8.09, 10.3, 11.4, 13, 14.8, 15.8, 16.9,
           17, 17.1, 2.26, 3.65, 15.6, 15.9),
  LxTx.Error = 0.1,
  TEST_DOSE = 17)
temp_OTORX_alt <-
  fit_DoseResponseCurve(
    cbind(LxTxData_alt, Test_Dose = 17),## we have to set the TEST_DOSE
    fit.method = "OTORX",
    fit.force_through_origin = TRUE,
    verbose = FALSE,
    n.MC = 10)

  temp_OTORX_alt2 <-
    fit_DoseResponseCurve(
      cbind(LxTxData_alt, Test_Dose = 17),## we have to set the TEST_DOSE
      fit.method = "OTORX",
      fit.force_through_origin = FALSE,
      verbose = FALSE,
      n.MC = 10)
  })

  expect_s3_class(temp_EXP$Fit, class = "nls")
  expect_s3_class(temp_LIN$Fit, class = "lm")
  expect_s3_class(temp_EXPLIN$Fit, class = "nls")
  expect_s3_class(temp_EXPEXP$Fit, class = "nls")
  expect_s3_class(temp_QDR$Fit, class = "lm")
  expect_s3_class(temp_GOK$Fit, class = "nls")
  expect_s3_class(temp_OTOR$Fit, class = "nls")
  expect_s3_class(temp_OTORX$Fit, class = "nls")
  expect_s3_class(temp_OTORX_alt$Fit, class = "nls")
  expect_s3_class(temp_OTORX_alt2$Fit, class = "nls")

   expect_equal(round(temp_EXP$De[[1]], digits = 2), 1737.88)
   expect_equal(round(sum(temp_EXP$De.MC, na.rm = TRUE), digits = 0), 17562)
   expect_equal(round(temp_LIN$De[[1]], digits = 2), 1811.33)
   expect_equal(round(sum(temp_LIN$De.MC, na.rm = TRUE), digits = 0),18398)
   expect_equal(round(temp_EXPLIN$De[[1]], digits = 2), 1791.53)
   expect_equal(round(sum(temp_EXPLIN$De.MC, na.rm = TRUE), digits = 0),18045)
   expect_equal(round(temp_EXPEXP$De[[1]], digits = 2), 1787.15)
   expect_equal(round(sum(temp_EXPEXP$De.MC, na.rm = TRUE), digits = 0), 7303,
                tolerance = 10)
   expect_equal(round(temp_QDR$De[[1]], digits = 2), 1666.2)
   expect_equal(round(sum(temp_QDR$De.MC, na.rm = TRUE), digits = 0), 16476)
   expect_equal(round(temp_GOK$De[[1]], digits = 0), 1786)
   ##fix for different R versions
   if (R.version$major > "3"){
     if(any(grepl("aarch64", sessionInfo()$platform))) {
       expect_equal(round(sum(temp_GOK$De.MC, na.rm = TRUE), digits = 1), 17796,
                    tolerance = 0.001)

     } else {
       expect_equal(round(sum(temp_GOK$De.MC, na.rm = TRUE), digits = 1), 17828.9,
                    tolerance = 0.1)
     }
   }

   expect_equal(round(temp_OTOR$De[[1]], digits = 2),  1784.78)
   expect_equal(round(temp_OTORX$De[[1]], digits = 2),  1785.43)
   expect_equal(round(temp_OTORX_alt$De[[1]], digits = 2),  758.280)
   expect_equal(round(temp_OTORX_alt2$De[[1]], digits = 2),  793.21, tolerance = 0.2)
   expect_equal(round(sum(temp_OTOR$De.MC, na.rm = TRUE), digits = 0), 17719)
   expect_equal(round(sum(temp_OTORX$De.MC, na.rm = TRUE), digits = 0), 17851, tolerance = 0.2)

# Check extrapolation -----------------------------------------------------
  ## load data
  data(ExampleData.LxTxData, envir = environment())

  set.seed(1)
  LxTxData[1,2:3] <- c(0.5, 0.001)
  SW({
  LIN <- expect_s4_class(
    fit_DoseResponseCurve(LxTxData,mode = "extrapolation", fit.method = "LIN"),
    "RLum.Results")
  EXP <- expect_s4_class(
    fit_DoseResponseCurve(LxTxData,mode = "extrapolation", fit.method = "EXP"),
    "RLum.Results")
  EXPLIN <- expect_s4_class(
    suppressWarnings(
      fit_DoseResponseCurve(
        LxTxData,mode = "extrapolation", fit.method = "EXP+LIN")),
    "RLum.Results")

  GOK <- expect_s4_class(
    fit_DoseResponseCurve(LxTxData,mode = "interpolation", fit.method = "GOK"),
    "RLum.Results")

  OTOR <- expect_s4_class(
    fit_DoseResponseCurve(LxTxData,mode = "extrapolation", fit.method = "OTOR"), "RLum.Results")

  ##OTORX
  OTORX <- expect_s4_class(
    fit_DoseResponseCurve(
      object = cbind(LxTxData, Test_Dose = 17),
      mode = "extrapolation", fit.method = "OTORX"), "RLum.Results")

  ##OTORX ... trigger uniroot warning
  LxTxData[1,2:3] <- c(5, 0.001)
  expect_warning(
    fit_DoseResponseCurve(
      object = cbind(LxTxData, Test_Dose = 17),
      mode = "extrapolation", fit.method = "OTORX"))

  })

  expect_equal(round(LIN$De$De,0), 165)
  expect_equal(round(EXP$De$De,0),  110)
  expect_equal(round(OTOR$De$De,0),  114)
  expect_equal(round(OTORX$De$De,0),  110)

  #it fails on some unix platforms for unknown reason.
  #expect_equivalent(round(EXPLIN$De$De,0), 110)

# Check alternate ---------------------------------------------------------
  ## load data
  data(ExampleData.LxTxData, envir = environment())

  set.seed(1)
  LxTxData[1,2:3] <- c(0.5, 0.001)

  SW({
  ##LIN
  expect_s4_class(
    fit_DoseResponseCurve(LxTxData, mode = "alternate", fit.method = "LIN"),
    "RLum.Results")

  ## EXP
  EXP <- expect_s4_class(
    fit_DoseResponseCurve(LxTxData, mode = "alternate", fit.method = "EXP"),
    "RLum.Results")
  })

  ## EXP+LIN
  EXPLIN <- expect_s4_class(
    suppressWarnings(
      fit_DoseResponseCurve(LxTxData, mode = "alternate", fit.method = "EXP+LIN", verbose = FALSE)),
    "RLum.Results")

  ## GOK
  expect_s4_class(
    fit_DoseResponseCurve(
      LxTxData,
      mode = "alternate",
      fit.method = "GOK",
      verbose = FALSE
    ),
    "RLum.Results"
  )

  ## OTOR
  expect_s4_class(
    fit_DoseResponseCurve(
      LxTxData,
      mode = "alternate",
      fit.method = "OTOR",
      verbose = FALSE
    ),
    "RLum.Results"
  )

  ## OTORX
  expect_s4_class(
    fit_DoseResponseCurve(
      cbind(LxTxData, Test_Dose = 17),
      mode = "alternate",
      fit.method = "OTORX",
      verbose = FALSE
    ),
    "RLum.Results"
  )

  ## trigger OTOR related warning for
  ## extrapolation mode
  tmp <- structure(list(
    dose = c(
      0,
      1388.88888888889,
      2777.77777777778,
      4166.66666666667,
      8333.33333333333,
      16666.6666666667,
      33333.3333333333,
      0,
      1388.88888888889,
      2777.77777777778,
      4166.66666666667,
      8333.33333333333,
      16666.6666666667,
      33333.3333333333,
      0,
      1388.88888888889,
      2777.77777777778,
      4166.66666666667,
      8333.33333333333,
      16666.6666666667,
      33333.3333333333
    ),
    LxTx = c(
      1.54252220145258,
      4.43951568403849,
      6.23268064543138,
      7.84372723139206,
      12.1816246695694,
      16.220421545207,
      19.9805214420208,
      1.5693958789807,
      4.01446969642433,
      6.50442121919275,
      8.13912565845306,
      11.2791435536017,
      14.2739718127927,
      17.7646886436743,
      1.55083317135234,
      4.10327222363961,
      6.1705969614814,
      8.30005789933367,
      12.7612004529065,
      14.807776070804,
      17.1563663039162
    ),
    LxTx_X = c(
      0.130074482379272,
      2.59694106608334,
      4.46970034588506,
      3.0630786645803,
      0.744512263874143,
      6.0383153231303,
      0.785060450424326,
      3.16210365279,
      0.0425273193228004,
      2.9667194222907,
      0.187174353876429,
      4.29989597009486,
      4.19802308979151,
      2.77791088935002,
      0.248412040945932,
      0.626745230335262,
      3.80396486752602,
      16.1846310553925,
      4.14921514089229,
      1.40190110413806,
      7.74406545663656
    )
  ),
  class = "data.frame",
  row.names = c(NA, -21L))

  expect_warning(fit_DoseResponseCurve(
    tmp,
    mode = "extrapolation",
    fit.method = "OTOR",
    verbose = FALSE),
    "[fit_DoseResponseCurve()] Standard root estimation using stats::uniroot()",
    fixed = TRUE)

  ## only two valid points provided: this generates two warnings, hence
  ## we cannot use expect_warning(), which can only capture one at a time
  SW({
  warnings <- capture_warnings(expect_message(fit_DoseResponseCurve(
    data.frame(
        dose = c(0, 1388.88888888889, NA),
        LxTx = c(1.54252220145258, 4.43951568403849, NA),
        LxTx_X = c(0.130074482379272, 2.59694106608, NA)),
    verbose = TRUE),
    "'fit.method' changed to 'LIN'"))
  })
  expect_match(warnings, "1 NA values removed",
               all = FALSE, fixed = TRUE)
  expect_match(warnings, "Fitting a non-linear least-squares model requires",
               all = FALSE, fixed = TRUE)

  ## more coverage
  tmp$dose <- c(0:6, 1000 + 0:6, 20000 + 0:6)
  expect_output(fit_DoseResponseCurve(
      tmp[4:8, ],
      fit.method = "GOK",
      verbose = TRUE,
      n.MC = 10),
      "Fit failed for GOK")
  expect_output(fit_DoseResponseCurve(
      cbind(tmp[4:8, ], Test_Dose = 17),
      fit.method = "OTORX",
      verbose = TRUE,
      n.MC = 10),
      "Fit failed for OTORX")
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 374 --------------------------------------------------------------

  ## odd data that cause NaN but must not fail
  df <- data.frame(DOSE = c(0,5,10,20,30), LxTx = c(10,5,-20,-30,-40), LxTx_X = c(1, 1,1,1,1))
  SW({
  expect_s4_class(
    fit_DoseResponseCurve(df, fit.method = "EXP"), "RLum.Results")
  expect_s4_class(
    fit_DoseResponseCurve(df, fit.method = "EXP+LIN"), "RLum.Results")
  expect_s4_class(
    fit_DoseResponseCurve(df, fit.method = "EXP+EXP"), "RLum.Results")
  expect_s4_class(
    fit_DoseResponseCurve(df, fit.method = "OTOR"), "RLum.Results")
  })

  ## issue 723
  set.seed(1)
  df <- data.frame(DOSE = c(0, 5, 10, 20, 25),
                   LxTx = c(40, -10, 30, -5, -20), LxTx_X = c(2, 2, 1, 2, 1))
  SW({
  expect_s4_class(fit_DoseResponseCurve(df, fit.method = "EXP"),
                  "RLum.Results")
  })
})

test_that("test internal functions", {
  testthat::skip_on_cran()

  ## reference tests from
  ## https://github.com/jll2/LumDRC/blob/main/otorx.py
  expect_equal(sum(Luminescence:::.nN2D(
    nN = 1-exp(-1),
    Q = c(0.5,0.6,0.7,0.8),
    D63 = c(1,10,100,1000))), expected = 1111)

  expect_equal(sum(Luminescence:::.nN2D(
    nN = 0.5,
    Q = c(0.5,0.6,0.7,0.8),
    D63 = c(1,10,100,1000))),
    expected = 661, tolerance = 2)

  expect_equal(sum(Luminescence:::.D2nN(
    D = 1,
    Q = c(-10,-3,0.1,1),
    a = 0,
    D63 = 1)), expected = 2.5, tolerance = 1)

  expect_error(Luminescence:::.D2nN(D = 1, Q = c(-10, 0.1, 0, 0), D63 = 1),
               "Unsupported zero and non-zero Q")
})

#' @title Extract Irradiation Times from an XSYG-file or `RLum.Analysis` object
#'
#' @description
#' Extracts irradiation times, dose and times since last irradiation, from a
#' Freiberg Instruments XSYG-file. These information can be further used to
#' update an existing BINX-file.
#'
#' @details
#' The function was written to compensate missing information in the BINX-file
#' output of Freiberg Instruments lexsyg readers. As all information are
#' available within the XSYG-file anyway, these information can be extracted
#' and used for further analysis or/and to stored in a new BINX-file, which can
#' be further used by other software, e.g., *Analyst* (Geoff Duller).
#'
#' Typical application example: *g*-value estimation from fading measurements
#' using the Analyst or any other self-written script.
#'
#' Beside some simple data transformation steps, the function relies on
#' [Luminescence::read_XSYG2R], [Luminescence::read_BIN2R], [Luminescence::write_R2BIN] for data import and export.
#'
#' **Calculation details**
#'
#' - The value `DURATION.STEP` is calculated as `START` + the end of the time axis for all curves
#' except for `TL`, where the function tries to extract meta information about the duration.
#'
#' - The value `END` is calculated as `START` + `DURACTION.STEP`
#'
#' - All curves for which no prior irradiation was detected receive an `-1` (*Analyst* convention)
#'
#' - Irradiation steps have always `IRR_TIME = 0`
#'
#' @param object [character], [Luminescence::RLum.Analysis-class] or [list] (**required**):
#' path and file name of the XSYG file or an [Luminescence::RLum.Analysis-class]
#' produced by the function [Luminescence::read_XSYG2R];
#' alternatively, a `list` of [Luminescence::RLum.Analysis-class] can be provided.
#'
#' **Note:** If an [Luminescence::RLum.Analysis-class] is used, any input for
#' the arguments `file.BINX` and `recordType` will be ignored!
#'
#' @param file.BINX [character] (*optional*):
#' path and file name of an existing BINX-file. If a file name is provided the
#' file will be updated with the information from the XSYG file in the same
#' folder as the original BINX-file.
#'
#' **Note:** The XSYG and the BINX-file must originate from the
#' same measurement!
#'
#' @param recordType [character] (*with default*):
#' select relevant curves types from the XSYG file or [Luminescence::RLum.Analysis-class]
#' object. As the XSYG-file format comprises much more information than usually
#' needed for routine data analysis and allowed in the BINX-file format, only
#' the relevant curves are selected by using the function
#' [Luminescence::get_RLum]. The argument `recordType` works as
#' described for this function.
#'
#' **Note:** A wrong selection will causes a function error. Please change this
#' argument only if you have reasons to do so.
#'
#' @param return_same_as_input [logical] (*with default*):
#' if set to `TRUE`, an updated [Luminescence::RLum.Analysis-class] object (or a [list] of
#' it) is returned, with each record having gained two new info element fields:
#' `IRR_TIME` and `TIMESCINCEIRR`. This makes the [Luminescence::RLum.Analysis-class] object
#' compatible with external functions that search explicitly for `IRR_TIME`
#' and `TIMESCINCEIRR`.
#'
#' @param compatibility.mode [logical] (*with default*):
#' whether all position values should be reset to a maximum value of 48 (see
#' [Luminescence::write_R2BIN]). Only used if `file.BINX` is specified.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar during import and export.
#'
#' @note The function can be also used to extract irradiation times from
#' [Luminescence::RLum.Analysis-class] objects imported via [Luminescence::read_BIN2R] with option
#' `fastForward = TRUE`, or in combination with
#' [Luminescence::Risoe.BINfileData2RLum.Analysis].
#' Unfortunately the timestamp might not be very precise (or even invalid),
#' but it allows to essentially treat different formats in a similar manner.
#'
#' @return
#' An [Luminescence::RLum.Results-class] object is returned with the
#' following structure:
#'
#' ```
#' .. $irr.times (data.frame)
#' ```
#'
#' If `return_same_as_input = TRUE` an [Luminescence::RLum.Analysis-class] or a [list] of it, but
#' we updated info elements including irradiation times.
#'
#' If a BINX-file path and name is set, the output will be additionally
#' transferred into a new BINX-file (with the function name as suffix) located
#' in the folder of the input BINX-file. Note that this will
#' not work if the input object is a file path to an XSYG-file, instead of a
#' link to only one file. In this case the argument input for `file.BINX` is ignored.
#'
#' In the self call mode (input is a `list` of [Luminescence::RLum.Analysis-class] objects
#' a list of [Luminescence::RLum.Results-class] is returned.
#'
#' @note
#' The produced output object contains still the irradiation steps to
#' keep the output transparent. However, for the BINX-file export this steps
#' are removed as the BINX-file format description does not allow irradiations
#' as separate sequences steps.
#'
#' **BINX-file 'Time Since Irradiation' value differs from the table output?**
#'
#' The way the value 'Time Since Irradiation' is defined differs. In the BINX-file the
#' 'Time Since Irradiation' is calculated as the 'Time Since Irradiation' plus the 'Irradiation
#' Time'. The table output returns only the real 'Time Since Irradiation', i.e. time between the
#' end of the irradiation and the next step.
#'
#' **Negative values for `TIMESINCELAST.STEP`?**
#'
#' Yes, this is possible and not a bug, as in the XSYG-file multiple curves are stored for one step.
#' Example: TL step may comprise three curves:
#'
#' - (a) counts vs. time,
#' - (b) measured temperature vs. time and
#' - (c) predefined temperature vs. time.
#'
#' Three curves, but they are all belonging to one TL measurement step, but with regard to
#' the time stamps this could produce negative values as the important function
#' ([Luminescence::read_XSYG2R]) do not change the order of entries for one step
#' towards a correct time order.
#'
#' **`TIMESINCELAST.STEP` is odd if TL curves are involved?**
#'
#' Yes, this is possible! The end time is calculated as start + duration,
#' and the duration is deduced from the time values of each step.
#' A typical TL curve, however, does not have a time but a temperature axis.
#' Hence, the function tries to extract the information from metadata.
#' If that information is missing, the x-values are presumed time values,
#' which might still be wrong.
#'
#' @section Function version: 0.4.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Results-class],
#' [Luminescence::Risoe.BINfileData-class],
#' [Luminescence::read_XSYG2R], [Luminescence::read_BIN2R], [Luminescence::write_R2BIN]
#'
#' @references
#' Duller, G.A.T., 2015. The Analyst software package for luminescence data: overview and
#' recent improvements. Ancient TL 33, 35-42. \doi{10.26034/la.atl.2015.489}
#'
#' @keywords IO manip
#'
#' @examples
#' ## take system file
#' xsyg <- system.file("extdata/XSYG_file.xsyg", package="Luminescence")
#'
#' ## the import is automatically
#' ## but you can import it before
#' irr_times <- extract_IrradiationTimes(xsyg)
#' irr_times$irr.times
#'
#' \dontrun{
#' # (1) - example for your own data
#'
#' # set files and run function
#' file.XSYG <- file.choose()
#' file.BINX <- file.choose()
#'
#' extract_IrradiationTimes(file.XSYG = file.XSYG, file.BINX = file.BINX)
#'
#' # export results additionally to a CSV-file in the same directory as the XSYG-file
#' write.table(x = get_RLum(output),
#'  file = paste0(file.BINX,"_extract_IrradiationTimes.csv"),
#'  sep = ";",
#'  row.names = FALSE)
#' }
#'
#' @export
extract_IrradiationTimes <- function(
  object,
  file.BINX,
  recordType = c("irradiation (NA)", "IRSL (UVVIS)", "OSL (UVVIS)", "TL (UVVIS)"),
  return_same_as_input = FALSE,
  compatibility.mode = TRUE,
  txtProgressBar = TRUE
) {
  .set_function_name("extract_IrradiationTimes")
  on.exit(.unset_function_name(), add = TRUE)

  ## Self-call --------------------------------------------------------------
  if (inherits(object, "list")) {
    if(!missing(file.BINX))
      .throw_warning("argument 'file.BINX' is not supported in self-call mode.")

    ## expand input arguments
    recordType <- .listify(recordType, length(object))

    ## run function
    results <- lapply(seq_along(object), function(x) {
        extract_IrradiationTimes(
          object = object[[x]],
          recordType = recordType[[x]],
          return_same_as_input = return_same_as_input,
          txtProgressBar = txtProgressBar
        )
      })

    if (length(results) == 0)
      return(NULL)
    return(results)
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("character", "RLum.Analysis"),
                  extra = "a 'list' of such objects")
  .validate_not_empty(object)
  .validate_logical_scalar(return_same_as_input)

  if (is.character(object[1])) {
    .validate_length(object, 1)

    ## Import XSYG and BINX -------------------------------------------------
    file.XSYG <- .validate_file(object, ext = "xsyg", scan.dir = FALSE,
                                verbose = txtProgressBar)
    temp.XSYG <- read_XSYG2R(
      file = file.XSYG,
      txtProgressBar = txtProgressBar,
      verbose = txtProgressBar)

    if(!missing(file.BINX)){
      file.BINX <- .validate_file(file.BINX, ext = "binx", scan.dir = FALSE,
                                  verbose = txtProgressBar)
      temp.BINX <- read_BIN2R(
        file = file.BINX,
        txtProgressBar = txtProgressBar,
        verbose = txtProgressBar)
    }

    # Some data preparation -----------------------------------------------------------------------
    ##set list
    temp.sequence.list <- list()

    ##select all analysis objects and combine them
    for(i in 1:length(temp.XSYG)){
      ##select sequence and reduce the data set to really wanted values
      temp.sequence.list[[i]] <- get_RLum(
        object = temp.XSYG[[i]]$Sequence.Object,
        recordType = recordType,
        drop = FALSE)

      ##get corresponding position number, this will be needed later on
      temp.sequence.position <- as.numeric(as.character(temp.XSYG[[i]]$Sequence.Header["position",]))
    }

  }else{
    ##now we assume a single RLum.Analysis object
    ##select sequence and reduce the data set to really wanted values, note that no
    ##record selection was made!
    temp.sequence.list <- list(object)
  }

  ## merge objects and restore originator
  temp.sequence <- merge_RLum(temp.sequence.list)
  temp.sequence@originator <- temp.sequence.list[[1]]@originator

# Grep relevant information -------------------------------------------------------------------
  ##Sequence STEP
  STEP <- names_RLum(temp.sequence)
  bin.originators <- c("Risoe.BINfileData2RLum.Analysis", "read_BIN2R")

  #START time of each step
  ## we try also to support BIN/BINX files with this function if imported
  ## accordingly
  if (.check_originator(temp.sequence, bin.originators)) {
    temp.START <- vapply(temp.sequence, function(x) {
       paste0(get_RLum(x, info.object = "DATE"), get_RLum(x, info.object = "TIME"))
    }, character(1))

    fmt <- if (grepl(":", temp.START[1])) "%y%m%d%H:%M:%S" else "%y%m%d%H%M%S"

  } else {
    temp.START <- vapply(temp.sequence, function(x) {
      suppressWarnings(get_RLum(x, info.object = "startDate")) %||%
        as.character(Sys.Date())
    }, character(1))

    fmt <- "%Y%m%d%H%M%S"
  }

  ## reformat start date
  START <- strptime(temp.START, format = fmt, tz = "GMT")

  ##DURATION of each STEP
  DURATION.STEP <- vapply(temp.sequence, function(x) {
    ## we treat TL curves differently **if** they have temperature axis
    ## in such case we search for 'duration' that should be there
    if(grepl(pattern = "TL", x@recordType, fixed = TRUE)) {
      t <- as.numeric(x@info$duration)
      if(length(t) != 0)
        return(t)
    }

    ## general fall back
    max(x@data[,1])

  }, numeric(1))

  ##Calculate END time of each STEP
  END <- START + DURATION.STEP

  ##add position number so far an XSYG file was the input
  POSITION <- NA
  if(exists("file.XSYG")){
    POSITION <- rep(temp.sequence.position, each = length_RLum(temp.sequence))

  } else {
    ## check that @records[[1]]@info$position is present
    check <- suppressWarnings(get_RLum(get_RLum(temp.sequence, record.id = 1),
                                       info.object = "position"))
    if (!is.null(check)) {
      ## POSITION of each STEP
      POSITION <- vapply(temp.sequence, function(x) {
        suppressWarnings(get_RLum(x, info.object = "position")) %||%
                         get_RLum(x, info.object = "POSITION")}, numeric(1))
    }
  }

  ##Combine the results
  temp.results <- data.frame(POSITION,STEP,START,DURATION.STEP,END)

  # Calculate irradiation duration ------------------------------------------------------------
  if (.check_originator(temp.sequence, bin.originators)) {
    IRR_TIME <- vapply(temp.sequence, get_RLum, info.object = "IRR_TIME",
                       FUN.VALUE = numeric(1))
  } else {
    IRR_TIME <- numeric(length = nrow(temp.results))
    temp_last <- 0
    for(i in 1:nrow(temp.results)){
      if(grepl("irradiation", temp.results[["STEP"]][i])) {
        temp_last <- temp.results[["DURATION.STEP"]][i]
        next()
      }

      IRR_TIME[i] <- temp_last
    }
  }

  ## Calculate time since irradiation ---------------------------------------
  time.irr.end <- NA
  TIMESINCEIRR <- vapply(1:nrow(temp.results), function(x) {
    if(grepl("irradiation", temp.results[x,"STEP"])){
      time.irr.end<<-temp.results[x,"END"]
      return(-1)
    }
    if (is.na(time.irr.end))
      return(-1)

    difftime(temp.results[x, "START"], time.irr.end, units = "secs")
  }, numeric(1))

  # Calculate time since last step --------------------------------------------------------------
  TIMESINCELAST.STEP <- vapply(1:nrow(temp.results), function(x){
    if(x == 1)
      return(0)

    difftime(temp.results[x,"START"],temp.results[x-1, "END"], units = "secs")
  }, numeric(1))

  # Combine final results -----------------------------------------------------------------------
  ##results table, export as CSV
  results <- cbind(temp.results,IRR_TIME, TIMESINCEIRR,TIMESINCELAST.STEP)

  # Write BINX-file if wanted -------------------------------------------------------------------
  if(!missing(file.BINX)){
    ##(1) remove all irradiation steps as there is no record in the BINX file and update information
    results.BINX <- results[results[, "STEP"] != "irradiation (NA)", ]

    ## (2) compare entries in the BINX-file with the entries in the table
    ## to make sure that both have the same length
    if(nrow(results.BINX) == nrow(temp.BINX@METADATA)){

      ## (1a) update information on the irradiation time
      temp.BINX@METADATA[["IRR_TIME"]] <- results.BINX[["IRR_TIME"]]

      ## (1b) update information on the time since irradiation by using the
      ## Risoe definition of the parameter, to make the file compatible to
      ## the Analyst
      temp.BINX@METADATA[["TIMESINCEIRR"]] <- results.BINX[["IRR_TIME"]] + results.BINX[["TIMESINCEIRR"]]

      ## update BINX-file
      try <- write_R2BIN(temp.BINX, version = "06",
                         file = paste0(file.BINX,"_extract_IrradiationTimes.BINX"),
                         compatibility.mode = compatibility.mode,
                         verbose = txtProgressBar,
                         txtProgressBar = txtProgressBar)

      ##set message on the format definition
      if(!inherits(x = try, 'try-error')){
        .throw_message("'Time Since Irradiation' was redefined in the ",
                       "exported BINX-file to: 'Time Since Irradiation' plus ",
                       "'Irradiation Time' to be compatible with the Analyst",
                       error = FALSE)
      }
    } else {
      .throw_message("XSYG-file and BINX-file do not contain similar entries, ",
                     "BINX-file update skipped")
    }
  }

  # Output --------------------------------------------------------------------------------------
   if(return_same_as_input[1]) {
     ## This odd mode we need to ensure that the function supports
     ## correctly all the odd modes with the automated import of XSYG files
     if(exists(x = "temp.XSYG", envir = environment())) {
       object <- lapply(temp.XSYG, function(x) x$Sequence.Object)

        ## make sure that the output is compatible
        if(length(object) == 1)
          object <- object[[1]]

        ## call the function again
        return(
          extract_IrradiationTimes(object, return_same_as_input = TRUE, recordType = recordType))

     }

    object@records <- lapply(seq_along(results$IRR_TIME), function(x){
      add_metadata(object@records[[x]], info_element = "IRR_TIME") <- results$IRR_TIME[[x]]
      add_metadata(object@records[[x]], info_element = "TIMESINCEIRR") <- results$TIMESINCEIRR[[x]]
      object@records[[x]]
    })

    return(object)
   }

  ## regular return
  set_RLum("RLum.Results",
           data = list(irr.times = results))
}

#' @title Calculate Moran's I
#'
#' @details
#'
#' **Case of no spatial autocorrelation**
#'
#' Perhaps a bit counter-intuitive, the expected value of Moran's I under the
#' null hypothesis of no spatial correlation is a value slightly smaller than
#' zero. When setting `spatial_autocorrelation = FALSE`, this function
#' calculates the expected value based on the number of observations or the
#' length of the observation vector (while taking out `NA` values). Note that
#' the expected value only depends on the number of observed separate grain
#' values. This can be useful for plotting.
#'
#' The expected Moran's I for the null hypothesis of no spatial correlation
#' corresponds to `-1 / (n - 1)`, with `n` being the number of non-missing
#' observations.
#'
#' @param object [Luminescence::RLum.Results-class] or [numeric] (**required**):
#' values of the grains of one disc. Should have length 100; can contain `NA`
#' values.
#'
#' @param df_neighbours [data.frame] (*with default*): a data frame with 3
#' columns (`location`, `neighbour` and `weight`, respectively), with each row
#' indicating the index of one location and one of its neighbours (note that
#' the concept of "location" versus "neighbour" is symmetric, so each
#' neighbouring pair needs to be specified only once), alongside their relative
#' weight (generally set to 1). If `NULL` (default), this is constructed
#' automatically by the internal function `.get_Neighbours`.
#'
#' @param spatial_autocorrelation [logical] (*with default*): whether spatial
#' autocorrelation should be considered in the computation of Moran's I
#' (`TRUE` by default). If `FALSE`, the function computes Moran's I expected
#' value in case of no autocorrelation (H_0). See details for further
#' information.
#'
#' @param compute_pseudo_p [logical] (*with default*): whether a pseudo p-value
#' should be computed (`FALSE` by default).
#'
#' @param tested_moransI [numeric] (*with default*): The value of Moran's I to
#' be tested against when computing the pseudo p-value. If `NULL` (default), the
#' value calculated by the function will be used. Ignored if `compute_pseudo_p`
#' is `FALSE`.
#'
#' @param n_permutations [integer] (*with default*): number of random
#' permutations tested to calculate the fraction which is the pseudo p
#' (defaults to 999). Influences the calculation speed, which will have
#' impact in case of large scale simulation loops. Ignored if `compute_pseudo_p`
#' is `FALSE`.
#'
#' @param ignore_borders [logical] (*with default*): whether only grain
#' locations that do not lie on the border of the disc should be considered
#' (`FALSE` by default). Thus if `TRUE`, only the inner 8x8 grain locations
#' rather than the full 10x10 are considered. Ignored if `df_neighbours` is
#' not `NULL` or if `spatial_autocorrelation = FALSE`.
#'
#' @param return_intermediate_values [logical] (*with default*): whether the
#' function should return a list with several intermediate calculation results
#' (defaults to `FALSE`). Ignored if `spatial_autocorrelation` is `FALSE`.
#'
#' @return By default one numerical value, roughly between -1 and 1, where
#' close to zero means no spatial correlation, and value close to 1 a positive
#' spatial correlation given the pattern we interested in (by default all rook
#' neighbours). A value closer to -1 has no meaning within the context of
#' luminescence crosstalk. If `compute_pseudo_p = TRUE`, then the computed
#' pseudo p-value is returned. If `return_intermediate_values` is set to `TRUE`,
#' a list with several values used for calculation is returned instead of a
#' single outcome.
#'
#' @author
#' Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
#' Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
#'
#' @references
#' de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
#' A novel method to assess crosstalk in single-grain luminescence detection.
#' Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
#'
#' @examples
#'
#' ## Test a fictional sample with spatial correlation
#' calc_MoransI(object = c(1:100))
#'
#' ## Test some fictional samples without spatial correlation;
#' ## note the randomness with each repetition
#' calc_MoransI(object = rnorm(n = 100))
#' calc_MoransI(object = rnorm(n = 100))
#' calc_MoransI(object = rnorm(n = 100))
#'
#' @export
calc_MoransI <- function(object,
                         df_neighbours = NULL,
                         spatial_autocorrelation = TRUE,
                         compute_pseudo_p = FALSE,
                         tested_moransI = NULL,
                         n_permutations = 999,
                         ignore_borders = FALSE,
                         return_intermediate_values = FALSE
) {
  .set_function_name("calc_MoransI")
  on.exit(.unset_function_name(), add = TRUE)

  ## Validate input arguments; set variables  -----------------------

  .validate_class(object, c("RLum.Results", "numeric", "integer"))
  if (is.numeric(object)) {
    vn_values <- object
  } else {
    vn_values <- get_RLum(object)
  }
  .validate_length(vn_values, 100, name = "'object'")
  .validate_logical_scalar(spatial_autocorrelation)
  .validate_logical_scalar(compute_pseudo_p)
  .validate_class(df_neighbours, "data.frame", null.ok = TRUE)
  .validate_class(tested_moransI, "numeric", null.ok = TRUE, length = 1)
  .validate_positive_scalar(n_permutations, int = TRUE)
  .validate_logical_scalar(ignore_borders, null.ok = TRUE)
  .validate_logical_scalar(return_intermediate_values)

  if (is.null(df_neighbours)) {
    df_neighbours <- .get_Neighbours(object, ignore_borders)
  } else if (ncol(df_neighbours) != 3) {
    .throw_error("'df_neighbours' should be a data frame with 3 columns")
  }

  ## No spatial autocorrelation ---------------------------------------------

  if (!spatial_autocorrelation) {
    n.valid <- sum(!is.na(vn_values))
    if (n.valid < 2) {
      .throw_error("There should be at least 2 non-missing values")
    }
    n_MoransI_expt_no_cor <- -1 / (n.valid - 1)
    return(n_MoransI_expt_no_cor)
  }

  ## Core calculation -------------------------------------------------------

  if (nrow(df_neighbours) == 0) {
    .throw_warning("No bordering grain locations given in 'df_neighbours', ",
                   "returning NaN")
    return(NaN)
  }

  n <- sum(!is.na(vn_values))
  n_mean <- mean(vn_values, na.rm = TRUE)
  n_spatial_auto_correlation <- sum((vn_values-n_mean)^2, na.rm = TRUE)/n # Population variance
  n_sum_similarities <-
    sum((vn_values[df_neighbours$location] - n_mean) *
        (vn_values[df_neighbours$neighbour] - n_mean) *
         df_neighbours$weight,
         na.rm = TRUE
    )
  n_sum_weights <- sum(df_neighbours$weight)

  n_average_auto_correlation <- n_sum_similarities/n_sum_weights

  n_moransI <- n_average_auto_correlation / n_spatial_auto_correlation

  if (compute_pseudo_p) {
    if (is.null(tested_moransI))
      tested_moransI <- n_moransI
    pseudo_p <- .compute_pseudo_p(vn_values, tested_moransI,
                                  n_permutations, df_neighbours)
  }

  if (return_intermediate_values) {
    li_return <- list(n = n,
                      n_mean = n_mean,
                      n_population_variance = n_spatial_auto_correlation,
                      n_sum_similarities = n_sum_similarities,
                      n_sum_weights = n_sum_weights,
                      n_average_auto_correlation = n_average_auto_correlation,
                      n_moransI = n_moransI)
    if (compute_pseudo_p) {
      li_return$pseudo_p <- pseudo_p
    }

    return(li_return)
  }

  if (compute_pseudo_p) {
    return(pseudo_p$pseudo_p)
  }
  return(n_moransI)
}


#' @title Calculate Moran's I pseudo p
#'
#' @param values [numeric] (**required**): numerical vector of length 100
#' representing one discs ("position") in a reader.
#'
#' @param moransI [numeric] (**required**): The value of Moran's I to be
#' tested against.
#'
#' @param n_permutations [integer] (*with default*) Number of random
#' permutations tested to calculate the fraction which is the pseudo p.
#'
#' @param df_neighbours [data.frame] (**required**): Data frame indicating
#' which borders to consider, and their respective weights.
#'
#' @return Pseudo p-value
#'
#' @seealso https://geodacenter.github.io/workbook/5a_global_auto/lab5a.html#permutation-inference
#'
#' @author
#' Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
#' Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
#'
#' @references
#' de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
#' A novel method to assess crosstalk in single-grain luminescence detection.
#' Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
#'
#' @keywords internal
#' @noRd
.compute_pseudo_p <- function(values, moransI, n_permutations, df_neighbours) {

    ## Local helper function: reshuffle (=permutate) observations, an test if Moran's I
    ## from reshuffled observations if equal or above the calculated Moran's I
    reshuffle_and_test <- function(vn_values, n_moransI, df_neighbours) {
      vn_values_reshuffled <- rep_len(NA_real_, length(vn_values))
      vn_values_reshuffled[vb_contains_observation] <- sample(vn_values[vb_contains_observation])
      (calc_MoransI(object = vn_values_reshuffled,
                    df_neighbours = df_neighbours) >= n_moransI)
    }
    vb_contains_observation <- !is.na(values)

    ## How many different Moran's Is from reshuffled observations are equal or
    ## above the provided Moran's I?
    n_test_pos <- sum(replicate(n = n_permutations,
                                expr = reshuffle_and_test(values,
                                                          moransI,
                                                          df_neighbours),
                                 simplify = TRUE)
    )

  n_pseudo_p <- (n_test_pos + 1) / (n_permutations + 1)

  if (n_test_pos == 0 && n_pseudo_p > 0)
    .throw_warning("Pseudo-p might be overestimated: the real p-value ",
                   "is closer to zero, try increasing 'n_permutations'")

  return(list(pseudo_p = n_pseudo_p,
              tested_moransI = moransI,
              n_permutations = n_permutations))
}


#' @title Get neighbour positions
#'
#' @description  Determines which locations on one or more
#' 10x10 discs are neighbouring horizontally and/or vertically ("rook"), in
#' order to calculate Moran's I.
#' For the general case, we assume symmetry, and all adjacency get
#' an equal relative weight of one.
#' In case there are holes in the image, indicated by `NA` (no value observed),
#' the related border or borders are removed.
#'
#' @details Will mostly be used as an helper function. If called without any
#' arguments, it will return the neighbouring positions data frame for one
#' disc (POS) with on every grain location an observation.
#'
#' @param object [Luminescence::RLum.Results-class] or [numeric] (**required**): numerical
#' vector containing the values of one or more discs, or an `RLum.Results`
#' object containing such values. Can contain `NA` values.
#'
#' @param ignore_borders [logical] (*with default*) Defaults to `FALSE`.
#' If `TRUE`, only on-disc borders to grain locations who do not border the
#' disc are considered; thus only the inner 8x8 grain locations rather than
#' the full 10x10.
#'
#' @return Data frame with columns `location`, `neighbour` and `weight`,
#' respectively. The integers in the first two columns indicate the index
#' of the vector with observations; the relative weight is standard set to one.
#' Note that the concept of "location" versus "neighbour" is symmetric.
#'
#' @author
#' Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
#' Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
#'
#' @references
#' de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
#' A novel method to assess crosstalk in single-grain luminescence detection.
#' Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
#'
#' @keywords internal
#'
#' @noRd
.get_Neighbours <- function(object,
                            ignore_borders = FALSE
) {
  .set_function_name(".get_Neighbours")
  on.exit(.unset_function_name(), add = TRUE)

  ## Validate input arguments -----------------------

  .validate_class(object, c("RLum.Results", "numeric", "integer"))
  if (is.numeric(object)) {
      vn_values <- object
  } else{
      vn_values <- get_RLum(object)
  }
  vb_contains_observation <- !is.na(vn_values)

  ## Construct the dataframe with borders and weights ---------------------------------

  df_neighbour <- data.frame(location = 1:100)

  ## for each neighbour, calculate two neighbouring locations,
  ## and remove when invalid

  df_neighbour$south <- df_neighbour$location - 10
  df_neighbour$west <- df_neighbour$location - 1

  ## Remove neighbouring locations at southside of disc
  df_neighbour$south <-
    ifelse(df_neighbour$south < 1,
           yes = NA,
           no = df_neighbour$south)

  ## Remove neighbouring locations at westside of disc
  df_neighbour$west <-
    ifelse(!(df_neighbour$west %% 10),
           yes = NA,
           no = df_neighbour$west)

  df_neighbour <- rbind(df_neighbour[,1:2],
                        setNames(df_neighbour[,c(1,3)],
                                 nm = names(df_neighbour)[1:2])
  )

  names(df_neighbour) <- c("location", "neighbour")

  df_neighbour <- stats::na.omit(df_neighbour)

  if (ignore_borders) {
    ## Remove the disc borders
    vn_disc_border_locations <- c(1:10,
                                  91:100,
                                  seq(from = 11, to = 81, by = 10),
                                  seq(from = 20, to = 90, by = 10)
    )

    contains_discborders <- df_neighbour$location %in% vn_disc_border_locations |
      df_neighbour$neighbour %in% vn_disc_border_locations

    df_neighbour <- df_neighbour[!contains_discborders, ]
  }

  ## Define relative weight factor ---------------------------------
  df_neighbour$weight <- 1

  ## If there are missing observations (NA's), we need to take those out ------------------
  missing.idx <- which(!vb_contains_observation)
  if (length(missing.idx) > 0) {
    ## Create vector of indices of rows to throw away
    vi_observation_missing <- (df_neighbour$location  %in% missing.idx |
                               df_neighbour$neighbour %in% missing.idx)
    df_neighbour <- df_neighbour[!vi_observation_missing, ]
  }

  return(df_neighbour)
}

## load data
obj <- set_RLum(class = "RLum.Results",
                data = list(vn_values = rep(x = c(1,2), each = 50)))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_MoransI("error"),
               "'object' should be of class 'RLum.Results', 'numeric' or 'integer'")
  expect_error(calc_MoransI(1:50),
               "'object' should have length 100")
  expect_error(calc_MoransI(obj, df_neighbours = "error"),
               "'df_neighbours' should be of class 'data.frame'")
  expect_error(calc_MoransI(obj, df_neighbours = iris),
               "'df_neighbours' should be a data frame with 3 columns")
  expect_error(calc_MoransI(obj, spatial_autocorrelation = "error"),
               "'spatial_autocorrelation' should be a single logical value")
  expect_error(calc_MoransI(obj, compute_pseudo_p = "error"),
               "'compute_pseudo_p' should be a single logical value")
  expect_error(calc_MoransI(obj, tested_moransI = "error"),
               "'tested_moransI' should be of class 'numeric' or NULL and have")
  expect_error(calc_MoransI(obj, n_permutations = "error"),
               "'n_permutations' should be a single positive integer value")
  expect_error(calc_MoransI(obj, ignore_borders = "error"),
               "'ignore_borders' should be a single logical value")
  expect_error(calc_MoransI(obj, return_intermediate_values = "error"),
               "'return_intermediate_values' should be a single logical value")

  expect_error(calc_MoransI(c(1, rep(NA, 99)),
                            spatial_autocorrelation = FALSE),
               "There should be at least 2 non-missing values")
  expect_warning(calc_MoransI(c(1, rep(NA, 99)),
                              spatial_autocorrelation = TRUE),
                 "No bordering grain locations given in 'df_neighbours'") # FIXME(mcol)
  expect_warning(res <- calc_MoransI(obj, df_neighbours = iris[0, 1:3]),
                 "No bordering grain locations given in 'df_neighbours'")
  expect_equal(res, NaN)
})

test_that("check functionality", {
  testthat::skip_on_cran()

  set.seed(1)

  expect_snapshot_plain(calc_MoransI(obj))
  expect_snapshot_plain(calc_MoransI(1:100, ignore_borders = TRUE))
  expect_snapshot_plain(calc_MoransI(obj, return_intermediate_values = TRUE))
  expect_snapshot_plain(calc_MoransI(obj,
                                     compute_pseudo_p = TRUE,
                                     tested_moransI = 0.125,
                                     return_intermediate_values = TRUE))

  expect_equal(calc_MoransI(obj, spatial_autocorrelation = FALSE),
               -0.010101010)
  expect_equal(calc_MoransI(1:100, spatial_autocorrelation = FALSE),
               -0.010101010)
  expect_equal(calc_MoransI(c(1:20, rep(NA, 80)), spatial_autocorrelation = FALSE),
               -0.052631579)

  expect_warning(res <- calc_MoransI(obj, compute_pseudo_p = TRUE),
                 "Pseudo-p might be overestimated: the real p-value is closer")
  expect_equal(res, 0.0010)
  expect_equal(calc_MoransI(1:100, compute_pseudo_p = TRUE,
                            tested_moransI = 0.05),
               0.188)
})

test_that("check .get_Neighbours", {
  testthat::skip_on_cran()

  expect_error(.get_Neighbours(NULL),
               "'object' should be of class 'RLum.Results', 'numeric' or")
  expect_snapshot_plain(.get_Neighbours(obj))
  expect_snapshot_plain(.get_Neighbours(obj, ignore_borders = TRUE))
  expect_snapshot_plain(.get_Neighbours(c(1:99, NA)))
})

#' @title Transform a CW-OSL curve into a pLM-OSL curve via interpolation under linear
#' modulation conditions
#'
#' @description
#' Transforms a conventionally measured continuous-wave (CW) OSL-curve into a
#' pseudo linearly modulated (pLM) curve under linear modulation conditions
#' using the interpolation procedure described by Bos & Wallinga (2012).
#'
#' @details
#' The complete procedure of the transformation is given in Bos & Wallinga
#' (2012). The input `data.frame` consists of two columns: time (t) and
#' count values (CW(t))
#'
#' **Nomenclature**
#'
#' - P = stimulation time (s)
#' - 1/P = stimulation rate (1/s)
#'
#' **Internal transformation steps**
#'
#' (1)
#' log(CW-OSL) values
#'
#' (2)
#' Calculate t' which is the transformed time:
#' \deqn{t' = 1/2*1/P*t^2}
#'
#' (3)
#' Interpolate CW(t'), i.e. use the log(CW(t)) to obtain the count values
#' for the transformed time (t'). Values beyond `min(t)` and `max(t)`
#' produce `NA` values.
#'
#' (4)
#' Select all values for t' < `min(t)`, i.e. values beyond the time resolution
#' of t. Select the first two values of the transformed data set which contain
#' no `NA` values and use these values for a linear fit using [lm].
#'
#' (5)
#' Extrapolate values for t' < `min(t)` based on the previously obtained
#' fit parameters.
#'
#' (6)
#' Transform values using
#' \deqn{pLM(t) = t/P*CW(t')}
#'
#' (7)
#' Combine values and truncate all values for t' > `max(t)`
#'
#' **Note:**
#' The number of values for t' < `min(t)` depends on the stimulation
#' period (P) and therefore on the stimulation rate 1/P. To avoid the
#' production of too many artificial data at the raising tail of the determined
#' pLM curves it is recommended to use the automatic estimation routine for
#' `P`, i.e. provide no own value for `P`.
#'
#' @param object [Luminescence::RLum.Data.Curve-class] or [data.frame] (**required**):
#' [Luminescence::RLum.Data.Curve-class] object or a `data.frame` with measured
#' curve data of type stimulation time (t) (`object[, 1]`) and measured counts
#' (cts) (`object[, 2]`).
#'
#' @param P [numeric] (*optional*):
#' stimulation time in seconds. If set to `NULL`, the optimal value is
#' estimated automatically (see details). Greater values of P produce more
#' points in the rising tail of the curve.
#'
#' @param ... currently not used.
#'
#' @return
#' The function returns the same data type as the input data type with
#' the transformed curve values.
#'
#' **`RLum.Data.Curve`**
#'
#' \tabular{rl}{
#' `$CW2pLMi.x.t` \tab: transformed time values \cr
#' `$CW2pLMi.method` \tab: used method for the production of the new data points
#' }
#'
#' @note
#' According to Bos & Wallinga (2012) the number of extrapolated points
#' should be limited to avoid artificial intensity data. If `P` is
#' provided manually and more than two points are extrapolated, a warning
#' message is returned.
#'
#' @section Function version: 0.3.5
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#' Based on comments and suggestions from:
#' Adrie J.J. Bos, Delft University of Technology, The Netherlands
#'
#' @seealso [Luminescence::convert_CW2pLM], [Luminescence::convert_CW2pHMi], [Luminescence::convert_CW2pPMi],
#' [Luminescence::fit_LMCurve], [Luminescence::RLum.Data.Curve-class]
#'
#' @references
#' Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL
#' signal components. Radiation Measurements 47, 752-758.
#'
#' **Further Reading**
#'
#' Bulur, E., 1996. An Alternative Technique For
#' Optically Stimulated Luminescence (OSL) Experiment. Radiation Measurements
#' 26, 701-709.
#'
#' Bulur, E., 2000. A simple transformation for converting CW-OSL curves to
#' LM-OSL curves. Radiation Measurements 32, 141-145.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##(1)
#' ##load CW-OSL curve data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ##transform values
#' values.transformed <- convert_CW2pLMi(ExampleData.CW_OSL_Curve)
#'
#' ##plot
#' plot(values.transformed$x, values.transformed$y.t, log = "x")
#'
#' ##(2) - produce Fig. 4 from Bos & Wallinga (2012)
#' ##load data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#' values <- CW_Curve.BosWallinga2012
#'
#' ##open plot area
#' plot(NA, NA,
#'      xlim = c(0.001,10),
#'      ylim = c(0,8000),
#'      ylab = "pseudo OSL (cts/0.01 s)",
#'      xlab = "t [s]",
#'      log = "x",
#'      main = "Fig. 4 - Bos & Wallinga (2012)")
#'
#'
#' values.t <- convert_CW2pLMi(values, P = 1/20)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col = "red", lwd = 1.3)
#' text(0.03,4500,"LM", col = "red", cex = .8)
#'
#' values.t <- convert_CW2pHMi(values, delta = 40)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col = "black", lwd = 1.3)
#' text(0.005,3000,"HM", cex =.8)
#'
#' values.t <- convert_CW2pPMi(values, P = 1/10)
#' lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
#'       col = "blue", lwd = 1.3)
#' text(0.5,6500,"PM", col = "blue", cex = .8)
#'
#' @export
convert_CW2pLMi<- function(
  object,
  P = NULL,
  ...
) {
  .set_function_name("convert_CW2pLMi")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  temp.values <- .prepare_CW2pX(object)
  .validate_positive_scalar(P, null.ok = TRUE)

  # (1) Transform values ------------------------------------------------------------------------

  ##(a) log transformation of the CW-OSL count values
  CW_OSL.log<-log(temp.values[,2])

  ##(b) time transformation t >> t'
  t<-temp.values[,1]
  if (anyDuplicated(t) > 0) {
    .throw_error("'object' contains duplicated time values")
  }

  ##set P
  ##if no values for P is set selected a P value for a maximum of
  ##two extrapolation points
  if (is.null(P)) {
    i<-10
    P<-1/i
    t.transformed<-0.5*1/P*t^2

    while(length(t.transformed[t.transformed<min(t)])>2){
      P<-1/i
      t.transformed<-0.5*1/P*t^2
      i<-i+10

    }#end::while
  }else{
    t.transformed<-0.5*1/P*t^2
  }
  #endif

  # (2) Interpolation ---------------------------------------------------------------------------

  ##interpolate values, values beyond the range return NA values
  CW_OSL.interpolated<-approx(t,CW_OSL.log, xout=t.transformed, rule=1 )
  if (all(is.na(CW_OSL.interpolated$y))) {
    .throw_error("All points are outside the interpolation range")
  }

  ## In some cases the interpolation algorithm is not working properly, and
  ## Inf or NaN values are produced
  interpolated <- .fix_interpolation_inf_nan(unlist(CW_OSL.interpolated$y),
                                             warn = FALSE)

  ## combine t.transformed and CW_OSL.interpolated in a data.frame
  temp <- data.frame(x = t.transformed, y = interpolated)

  # (3) Extrapolate first values of the curve ---------------------------------------------------

  res <- .extrapolate_first(temp, t = t[1:2], y = CW_OSL.log[1:2])
  temp <- res$df
  temp.method <- res$method

  # (4) Convert, transform and combine values ---------------------------------------------------

  ##unlog CW-OSL count values, i.e. log(CW) >> CW
  CW_OSL<-exp(temp$y)

  ##transform CW-OSL values to pLM-OSL values
  pLM<-1/P*t*CW_OSL

  ##combine all values and exclude NA values
  temp.values <- data.frame(x=t,y.t=pLM,x.t=t.transformed, method=temp.method)
  temp.values <- na.exclude(temp.values)

  # (5) Return values ---------------------------------------------------------------------------

  ##returns the same data type as the input
  if (is.data.frame(object)) {
    return(temp.values)
  }

    ##add old info elements to new info elements
    temp.info <- c(object@info,
                   CW2pLMi.x.t = list(temp.values$x.t),
                   CW2pLMi.method = list(temp.values$method))

  set_RLum(
      class = "RLum.Data.Curve",
      recordType = object@recordType,
      data = as.matrix(temp.values[,1:2]),
      info = temp.info)
}

.prepare_CW2pX <- function(object) {
  .validate_class(object, c("data.frame", "RLum.Data.Curve"))
  .validate_not_empty(object)
  if (ncol(object) < 2) {
    .throw_error("'object' should have 2 columns")
  }

  ##(2) if the input object is an 'RLum.Data.Curve' object check for allowed curves
  if (inherits(object, "RLum.Data.Curve")) {
    if (!grepl("OSL", object@recordType) && !grepl("IRSL", object@recordType)) {
      .throw_error("recordType ", object@recordType,
                   " is not allowed for the transformation")
    }

    object <- as(object, "data.frame")
  }
  if (isTRUE(any(object[, 1] < 0))) {
    .throw_error("'object' cannot contain negative times")
  }
  if (isTRUE(any(object[, 2] < 0))) {
    .throw_error("'object' cannot contain negative counts")
  }

  ## remove NAs
  object <- na.exclude(object)
  if (nrow(object) < 2) {
    .throw_error("'object' should have at least 2 non-missing values")
  }

  object
}

.fix_interpolation_inf_nan <- function(values, warn) {
  invalid.idx <- which(is.infinite(values) | is.nan(values))
  if (length(invalid.idx) == 0)
    return(values)
  if (all(is.infinite(values) | is.na(values)))
    .throw_error("All interpolated values are Inf/NaN/NA, check your data")

  ## replace invalid values with mean of the value before and the value after
  values[invalid.idx] <- sapply(invalid.idx,
                                function(x) mean(values[c(x - 1, x + 1)]))

  if (warn) {
    .throw_warning(length(invalid.idx), " invalid values found, ",
                   "replaced by the mean of the nearest values")
  }

  values
}

.extrapolate_first <- function(df, t, y) {
  ##(a) - find index of first rows which contain NA values (needed for extrapolation)
  temp.sel.id <- min(which(!is.na(df[, 2])))

  ##(b) - fit linear function
  fit.lm <- stats::lm(y ~ x, data.frame(x = t, y = y))

  ## select values to extrapolate and predict (extrapolate) values based on
  ## the fitted function
  x.i <- data.frame(x = df[1:(min(temp.sel.id) - 1), 1])
  y.i <- predict(fit.lm, x.i)

  ## replace NA values by extrapolated values
  df[1:length(y.i), 2] <- y.i

  ##set method values
  temp.method <- c(rep("extrapolation", length(y.i)),
                   rep("interpolation", length(df[, 2]) - length(y.i)))

  ## print a warning message for more than two extrapolation points
  if (length(y.i) > 2) {
    .throw_warning("t' is beyond the time resolution and more than two ",
                   "data points have been extrapolated")
  }

  list(df = df, method = temp.method)
}

#' @title Non-linear Least Squares Fit for CW-OSL curves -beta version-
#'
#' @description
#' The function determines the weighted least-squares estimates of the
#' component parameters of a CW-OSL signal for a given maximum number of
#' components and returns various component parameters.
#'
#' **Fitting function**
#'
#' The function for the CW-OSL fitting has the general form:
#'
#' \deqn{y = I0_{1}\lambda_{1}\exp(-\lambda_1 x) + \ldots + I0_{i}\lambda_{i}\exp(-\lambda_i x) }
#'
#' where \eqn{0 < i < 8}, \eqn{\lambda} is the decay constant,
#' and \eqn{I0} the initial number of trapped electrons.
#'
#' *(for the used equation cf. Boetter-Jensen et al., 2003, Eq. 2.31)*
#'
#' **Start values**
#'
#' Start values are estimated automatically by fitting a linear function to the
#' logarithmized input data set. Currently, there is no option to manually
#' provide start parameters.
#'
#' **Goodness of fit**
#'
#' The goodness of the fit is given as pseudoR² value (pseudo coefficient of
#' determination). According to Lave (1970), the value is calculated as:
#'
#' \deqn{pseudoR^2 = 1 - RSS/TSS}
#'
#' where \eqn{RSS} is the Residual Sum of Squares, and \eqn{TSS} is the Total
#' Sum of Squares.
#'
#' **Error of fitted component parameters**
#'
#' The 1-sigma error for the
#' components is calculated using the function [stats::confint]. Due to
#' considerable calculation time, this option is deactivated by default. In
#' addition, the error for the components can be estimated by using internal R
#' functions like [summary]. See the [nls] help page
#' for more information.
#'
#' *For details on the nonlinear regression in R, see Ritz & Streibig (2008).*
#'
#' @param object [Luminescence::RLum.Data.Curve-class] or [data.frame] (**required**):
#' x, y data of measured values (time and counts).
#'
#' @param n.components.max [vector] (*optional*):
#' maximum number of components that are to be used for fitting.
#' The upper limit is 7.
#'
#' @param fit.failure_threshold [integer] (*with default*):
#' limits the failed fitting attempts.
#'
#' @param fit.method [character] (*with default*):
#' select the fit method, either `"port"` to use the 'port' routine from
#' function [nls], or `"LM"` to use the Levenberg-Marquardt algorithm as
#' implemented in function [minpack.lm::nlsLM].
#'
#' @param fit.trace [logical] (*with default*):
#' traces the fitting process on the terminal.
#'
#' @param fit.calcError [logical] (*with default*):
#' calculate 1-sigma error range of components using [stats::confint]
#'
#' @param LED.power [numeric] (*with default*):
#' LED power (max.) used for intensity ramping in mW/cm².
#' **Note:** The value is used for the calculation of the absolute
#' photoionisation cross section.
#'
#' @param LED.wavelength [numeric] (*with default*):
#' LED wavelength used for stimulation in nm.
#' **Note:** The value is used for the calculation of the absolute
#' photoionisation cross section.
#'
#' @param cex.global [numeric] (*with default*):
#' global scaling factor.
#'
#' @param sample_code [character] (*optional*):
#' sample code used for the plot and the optional output table (`mtext`).
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param output.terminalAdvanced [logical] (*with default*):
#' enhanced terminal output. Only valid if `verbose = TRUE`.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param method_control [list] (*optional*): options to control the output
#' produced. Currently only the 'export.comp.contrib.matrix' (logical) option
#' is supported, to enable/disable export of the component contribution
#' matrix.
#'
#' @param ... further arguments and graphical parameters passed to [plot].
#'
#' @return
#' **plot (*optional*)**
#'
#' the fitted CW-OSL curves are returned as plot.
#'
#' **RLum.Results object**
#'
#' Beside the plot and table output options, an [Luminescence::RLum.Results-class] object is
#' returned.
#'
#' `fit`:
#' an `nls` object (`$fit`) for which generic R functions are
#' provided, e.g. [summary], [stats::confint], [profile]. For more
#' details, see [nls].
#'
#' `output.table`:
#' a [data.frame] containing the summarised parameters including the error
#'
#' `component.contribution.matrix`:
#' [matrix] containing the values for the component to sum contribution plot
#' (`$component.contribution.matrix`).
#' Produced only if `method_control$export.comp.contrib.matrix = TRUE`).
#'
#' Matrix structure:\cr
#' Column 1 and 2: time and `rev(time)` values \cr
#' Additional columns are used for the components, two for each component,
#' containing I0 and n0. The last columns `cont.` provide information on
#' the relative component contribution for each time interval including the row
#' sum for this values.
#'
#' @note
#'
#' **Beta version - This function has not been properly tested yet and**
#' **should therefore not be used for publication purposes!**
#'
#' The pseudo-R² may not be the best parameter to describe the goodness of the
#' fit. The trade off between the `n.components` and the pseudo-R² value
#' is currently not considered.
#'
#' The function **does not** ensure that the fitting procedure has reached a
#' global minimum rather than a local minimum!
#'
#' @section Function version: 0.5.6
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::fit_LMCurve], [plot],[nls], [Luminescence::RLum.Data.Curve-class],
#' [Luminescence::RLum.Results-class], [Luminescence::get_RLum], [minpack.lm::nlsLM]
#'
#' @references
#' Boetter-Jensen, L., McKeever, S.W.S., Wintle, A.G., 2003.
#' Optically Stimulated Luminescence Dosimetry. Elsevier Science B.V.
#'
#' Lave, C.A.T., 1970. The Demand for Urban Mass Transportation. The Review of
#' Economics and Statistics, 52 (3), 320-323.
#'
#' Ritz, C. & Streibig, J.C., 2008. Nonlinear Regression with R. In: R.
#' Gentleman, K. Hornik, G. Parmigiani, eds., Springer, p. 150.
#'
#' @keywords dplot models
#'
#' @examples
#'
#' ##load data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' ##fit data
#' fit <- fit_CWCurve(ExampleData.CW_OSL_Curve,
#'                    main = "CW Curve Fit",
#'                    n.components.max = 4,
#'                    log = "x")
#'
#' @export
fit_CWCurve<- function(
  object,
  n.components.max = 7,
  fit.failure_threshold = 5,
  fit.method = c("port", "LM"),
  fit.trace = FALSE,
  fit.calcError = FALSE,
  LED.power = 36,
  LED.wavelength = 470,
  cex.global = 0.6,
  sample_code = "Default",
  verbose = TRUE,
  output.terminalAdvanced = TRUE,
  plot = TRUE,
  method_control = list(),
  ...
) {
  .set_function_name("fit_CWCurve")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, c("RLum.Data.Curve", "data.frame"))
  .validate_not_empty(object)
  if (inherits(object, "RLum.Data.Curve")) {
    object <- as.data.frame(object@data[, 1:2, drop = FALSE])
  }
  if (ncol(object) < 2) {
    .throw_error("'object' should have 2 columns")
  }

  ## set x and y values
  values <- object
  x <- values[, 1]
  y <- values[, 2]

  if (all(y <= 0, na.rm = TRUE)) {
    .throw_error("'object' contains no positive counts")
  }
  if (any(order(x) != seq_along(x))) {
    .throw_error("Time values are not ordered")
  }
  fit.method <- .validate_args(fit.method, c("port", "LM"))
  .validate_positive_scalar(n.components.max, int = TRUE)
  .validate_positive_scalar(fit.failure_threshold, int = TRUE)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(output.terminalAdvanced)
  .validate_logical_scalar(plot)
  .validate_class(method_control, "list")

  # Deal with extra arguments -----------------------------------------------

  ##deal with addition arguments
  extraArgs <- list(...)

  main <- extraArgs$main %||% "CW-OSL Curve Fit"
  log <- extraArgs$log %||% ""
  xlab <- extraArgs$xlab %||% "Time [s]"
  ylab <- extraArgs$ylab %||% paste0("OSL [cts/", round(max(x) / length(x), digits = 2), " s]")
  method_control <- modifyList(x = list(export.comp.contrib.matrix = FALSE),
                               val = method_control)

  ##============================================================================##
  ## FITTING
  ##============================================================================##
  ##
  ##////equation used for fitting////(start)
  fit.equation <- function(I0.i,lambda.i){
    parse(text = paste0("I0[", I0.i, "] * lambda[", lambda.i,
                        "] * exp(-lambda[", lambda.i ,"] * x)", collapse = "+"))
  }
  ##////equation used for fitting///(end)

  ##set formula elements for fitting functions
  ## the upper two funtions should be removed ... but chances are needed ... TODO
  ##////equation used for fitting////(start)
  fit.formula <- function(n.components){
    I0 <- paste0("I0.",1:n.components)
    lambda <- paste0("lambda.",1:n.components)
    stats::as.formula(paste0("y ~ ", paste(I0," * ", lambda,
                                           "* exp(-",lambda," * x)",
                                           collapse = " + ")))
  }
  ##////equation used for fitting///(end)

  ##////equation used for fitting////(start)
  fit.formula.simple <- function(n.components){
    I0 <- paste0("I0.",1:n.components)
    lambda <- paste0("lambda.",1:n.components)
    stats::as.formula(paste0("y ~ ", paste(I0," * exp(-",lambda," * x)",
                             collapse = " + ")))
  }
  ##////equation used for fitting///(end)

  ##set variables
  fit <- NULL
  keep.fitting <- TRUE # set to FALSE if the fitting should be stopped early
  n.components <- 1 #number of components used for fitting - start with 1
  fit.failure_counter <- 0 #counts the failed fitting attempts

  ## set n.components.max it to the maximum value that can be
  ## fitted given the data size (issue #953), up to a maximum of 7 components
  n.fittable <- max(nrow(values) - 3, 1)
  n.components.max <- min(n.components.max, n.fittable, 7)

  ##++++Fitting loop++++(start)
  while(keep.fitting && n.components <= n.components.max) {
    ##(0) START PARAMETER ESTIMATION
    ##rough automatic start parameter estimation

    if (fit.trace)
      cat("n.component:", n.components, "/", n.components.max, "\n")

    ##I0
    I0<-rep(values[1,2]/3,n.components)
    names(I0) <- paste0("I0.",1:n.components)

    ##lambda
    ##ensure that no values <=0 are included remove them for start parameter
    ##estimation and fit an linear function a first guess
    temp.values <- data.frame(x[y > 0], log(y[y > 0]))

    temp <- tryCatch(stats::lm(temp.values),
                     error = function(e) .throw_error(e$message))
    lambda<-abs(temp$coefficient[2])/nrow(values)

    k<-2
    while(k<=n.components){
      lambda[k]<-lambda[k-1]/100
      k<-k+1
    }
    names(lambda) <- paste0("lambda.",1:n.components)

    ##(1) FIRST FIT WITH A SIMPLE FUNCTION
    if(fit.method == "LM"){
      ##try fit simple
      fit.try<-suppressWarnings(try(minpack.lm::nlsLM(fit.formula.simple(n.components),
                                          data=values,
                                          start=c(I0,lambda),
                                          na.action = "na.exclude",
                                          trace = fit.trace,
                                          control = minpack.lm::nls.lm.control(
                                            maxiter = 500
                                          )),
                                    silent = TRUE
      ))#end try

    }else if(fit.method == "port"){
      ##try fit simple
      fit.try<-suppressWarnings(try(nls(fit.formula.simple(n.components),
                                        data=values,
                                        trace = fit.trace,
                                        algorithm="port",
                                        na.action = "na.exclude",
                                        start=c(I0,lambda),
                                        stats::nls.control(
                                          tol = 1,
                                          maxiter=100,
                                          warnOnly=FALSE,
                                          minFactor=1/1024
                                        ),
                                        lower=rep(0,n.components * 2)# set lower boundaries for components
      ), silent=TRUE# nls
      ))#end try
    }

    ##(3) FIT WITH THE FULL FUNCTION
    if (!inherits(fit.try, "try-error")) {

      ##grep parameters from simple fit to further work with them
      parameters <- coef(fit.try)

      ##grep parameters an set new starting parameters, here just lambda is choosen as
      ##it seems to be the most valuable parameter
      lambda <- parameters[(n.components+1):length(parameters)]

      if(fit.method == "LM"){

        ##try fit simple
        fit.try<-suppressWarnings(try(minpack.lm::nlsLM(fit.formula(n.components),
                                            data=values,
                                            start=c(I0,lambda),
                                            trace = fit.trace,
                                            na.action = "na.exclude",
                                            lower = rep(0,n.components * 2),
                                            control = minpack.lm::nls.lm.control(
                                              maxiter = 500
                                            )),
                                      silent = TRUE))

        ## HACK:
        # minpack.lm::nlsLM() stores the 'lower' argument as class "call" rather
        # than "numeric" as nls() does. Before running confint() on this object
        # we overwrite the "lower" slot with the numeric values again.
        if (!inherits(fit.try, "try-error")) {
          fit.try$call$lower <- rep(0,n.components * 2)
        }

      }else{
        ##try fit
        fit.try<-suppressWarnings(try(nls(fit.formula(n.components),
                                          trace=fit.trace,
                                          data=values,
                                          algorithm="port",
                                          na.action = "na.exclude",
                                          start=c(I0,lambda),
                                          stats::nls.control(
                                            maxiter = 500,
                                            warnOnly = FALSE,
                                            minFactor = 1/4096
                                          ),
                                          lower=rep(0,n.components * 2)# set lower boundaries for components
        ), silent=TRUE# nls
        ))#end try

      }#fit.method
    }
    n.components <- n.components + 1

    if (!inherits(fit.try, "try-error") || is.null(fit)) {
      fit <- fit.try
    }

    ## count failed fitting attempts and stop after a given number of attempts
    if (inherits(fit.try, "try-error")) {
      fit.failure_counter <- fit.failure_counter + 1
      if (fit.failure_counter >= fit.failure_threshold)
        keep.fitting <- FALSE
    }

  }##end while
  ##++++Fitting loop++++(end)

  ##============================================================================##
  ## FITTING OUTPUT
  ##============================================================================##

  ##grep parameters
  output.table <- component.contribution.matrix <- NA
  if (!inherits(fit, "try-error")) {

    parameters <- coef(fit)

    ##correct fit equation for the de facto used number of components
    I0.i<-1:(length(parameters)/2)
    lambda.i<-1:(length(parameters)/2)
    fit.function<-fit.equation(I0.i=I0.i,lambda.i=lambda.i)
    n.components<-length(I0.i)

    ##write parameters in vectors and order by decreasing lambda value
    I0<-parameters[1:(length(parameters)/2)]
    lambda<-parameters[(1+(length(parameters)/2)):length(parameters)]

    o<-order(lambda,decreasing=TRUE)
    I0<-I0[o]
    lambda<-lambda[o]

    ##============================================================================##
    ## Additional Calculation
    ##============================================================================##

    ## ---------------------------------------------
    ##calculate stimulation intensity Schmidt (2008)

    ##Energy - E = h*v
    h <- .const$h # Planck constant (W*s^2)
    ny <- .const$c / (LED.wavelength / 10^9) # frequency of light
    E<-h*ny

    ## transform LED.power in W/cm²
    LED.power<-LED.power/1000

    ##gets stimulation intensity
    stimulation_intensity<-LED.power/E

    ## ---------------------------------------------
    ##calculate photoionisation cross section and print on terminal

    ##using EQ (5) in Kitis
    cs<-as.vector(lambda/stimulation_intensity)
    cs.rel<-round(cs/cs[1],digits=4)

    ## ---------------------------------------------
    ##coefficient of determination after law

    RSS <- sum(residuals(fit)^2) #residual sum of squares
    TSS <- sum((y - mean(y))^2) #total sum of squares
    pR<-round(1-RSS/TSS,digits=4)

    if(pR<0){
      .throw_warning("pseudo-R^2 < 0!") # nocov
    }

    ## ---------------------------------------------
    ##calculate 1- sigma CONFIDENCE INTERVALL

    lambda.error<-rep(NA, n.components)
    I0.error<-rep(NA, n.components)

    ## option for confidence interval
    if (fit.calcError) {
      tryCatch({
        values.confint <- stats::confint(fit, level = 0.68)
        half <- nrow(values.confint) / 2
        I0.confint <- values.confint[1:half, ]
        lambda.confint <- values.confint[half + 1:half, ]

        ## error calculation
        I0.error <- abs(I0.confint[, 1] - I0.confint[, 2])
        lambda.error <- abs(lambda.confint[, 1] - lambda.confint[, 2])
      }, error = function(e) {
        ## report the error from confint()
        .throw_message("Computation of confidence interval failed: ",
                       e$message)
      })
    }#endif::fit.calcError

    ##============================================================================##
    ## Terminal Output
    ##============================================================================##

    if (verbose) {
      ##print rough fitting information - use the nls() control for more information
      writeLines("\n[fit_CWCurve()]")
      cat(paste0("\nFitting was finally done using a ", n.components,
                 "-component function (max=", n.components.max, "):\n"))
      writeLines("------------------------------------------------------------------------------")
      writeLines(paste0("y ~ ", as.character(fit.formula(n.components))[3], "\n"))

      ##combine values and change rows names
      fit.results<-cbind(I0,I0.error,lambda,lambda.error,cs, cs.rel)
      row.names(fit.results) <- paste0("c", 1:(length(parameters) / 2))

      ##print parameters
      print(fit.results)

      #print some additional information
      if (fit.calcError)
        cat("(errors quoted as 1-sigma values)\n")
      writeLines("------------------------------------------------------------------------------")
    }#end if

    ##============================================================================##
    ## Terminal Output (advanced)
    ##============================================================================##
    if (verbose && output.terminalAdvanced) {
      ##sum of squares
      cat("pseudo-R^2 = ", pR, "\n")
    }#end if

    ##============================================================================##
    ## Table Output
    ##============================================================================##

    ##write output table if values exists
    ## set data.frame for a max value of 7 components
    output.table <- data.frame(matrix(rbind(I0, I0.error, lambda, lambda.error,
                                            cs, cs.rel),
                                      nrow = 1))
    ncols.table <- ncol(output.table) / 6
    colnames(output.table) <-  c(rbind(paste0("I0", 1:ncols.table),
                                       paste0("I0", 1:ncols.table, ".error"),
                                       paste0("lambda", 1:ncols.table),
                                       paste0("lambda", 1:ncols.table, ".error"),
                                       paste0("cs", 1:ncols.table),
                                       paste0("cs", 1:ncols.table, ".rel")))
    output.table <- cbind(sample_code, n.components, output.table,
                          "pseudo-R^2"=pR)

      ##============================================================================##
      ## COMPONENT TO SUM CONTRIBUTION PLOT
      ##============================================================================##

      ##+++++++++++++++++++++++++++++++
      ##set matrix
      ##set polygon matrix for optional plot output
      component.contribution.matrix <- matrix(NA,
                                              nrow = length(values[,1]),
                                              ncol = (2*length(I0)) + 2)

      ##set x-values
      component.contribution.matrix[,1] <- values[,1]
      component.contribution.matrix[,2] <- rev(values[,1])

      ##+++++++++++++++++++++++++++++++
      ##set 1st polygon
      ##1st polygon (calculation)
      y.contribution_first<-(I0[1]*lambda[1]*exp(-lambda[1]*x))/(eval(fit.function))*100

      ##avoid NaN values (might happen with synthetic curves)
      y.contribution_first[is.nan(y.contribution_first)] <- 0

      ##set values in matrix
      component.contribution.matrix[,3] <- 100
      component.contribution.matrix[,4] <- 100 - rev(y.contribution_first)

      ##+++++++++++++++++++++++++++++++
      ##set polygons in between
      ##polygons in between (calculate and plot)
      if (length(I0)>2){

        y.contribution_prev <- y.contribution_first
        i<-2

        ##matrix stepping
        k <- seq(3, ncol(component.contribution.matrix), by=2)

        while (i<=length(I0)-1) {
          y.contribution_next<-I0[i]*lambda[i]*exp(-lambda[i]*x)/(eval(fit.function))*100

          ##avoid NaN values
          y.contribution_next[is.nan(y.contribution_next)] <- 0

          ##set values in matrix
          component.contribution.matrix[,k[i]] <- 100 - y.contribution_prev
          component.contribution.matrix[, k[i]+1] <- rev(100-y.contribution_prev-
                                                           y.contribution_next)

          y.contribution_prev <- y.contribution_prev + y.contribution_next

          i <- i+1
        }#end while loop
      }#end if

      ##+++++++++++++++++++++++++++++++
      ##set last polygon

      ##last polygon (calculation)
      y.contribution_last <- I0[length(I0)]*lambda[length(lambda)]*exp(-lambda[length(lambda)]*x)/
        (eval(fit.function))*100

      ##avoid NaN values
      y.contribution_last[is.nan(y.contribution_last)] <- 0

      component.contribution.matrix[,((2*length(I0))+1)] <- y.contribution_last
      component.contribution.matrix[,((2*length(I0))+2)] <- 0

      ##change names of matrix to make more easy to understand
      component.contribution.matrix.names <- c(
        "x", "rev.x",
        paste0(c("y.c", "rev.y.c"), rep(1:n.components, each = 2)))

      ##calculate area for each component, for each time interval
      component.contribution.matrix.area <- sapply(
        seq(3,ncol(component.contribution.matrix),by=2),
        function(x){
          matrixStats::rowDiffs(cbind(rev(component.contribution.matrix[,(x+1)]),
                         component.contribution.matrix[,x]))
        })

      ##append to existing matrix
      component.contribution.matrix <- cbind(
        component.contribution.matrix,
        component.contribution.matrix.area,
        rowSums(component.contribution.matrix.area)
      )

      ##set final column names
      colnames(component.contribution.matrix) <- c(
        component.contribution.matrix.names,
        paste0("cont.c", 1:n.components),
        "cont.sum")
  }

  ##============================================================================##
  ## PLOTTING
  ##============================================================================##
  if (plot) {
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    ##set colors gallery to provide more colors
    col <- get("col", pos = .LuminescenceEnv)

    ##set plot frame
    par(cex = cex.global)
    if(!inherits(fit, "try-error")){
      graphics::layout(matrix(c(1, 2, 3), 3, 1, byrow = TRUE),
                       c(1.6, 1, 1), c(1, 0.3, 0.4), TRUE)
      par(oma = c(1, 1, 1, 1), mar = c(0, 4, 3, 0))
    }
    ##== upper plot ==##
    ##open plot area
    plot_check <- try(plot(NA, NA,
         xlim=c(min(x),max(x)),
         ylim = c(as.integer(log == "xy"), max(y)),
         xlab = ifelse(inherits(fit, "try-error"), xlab, ""),
         xaxt = ifelse(inherits(fit, "try-error"), "s", "n"),
         ylab=ylab,
         main=main,
         log = log), silent = TRUE)

    if (inherits(plot_check, "try-error")) {
      ## reset the graphic device if plotting failed
      .throw_message("Figure margins too large or plot area too small, ",
                     "nothing plotted")
      grDevices::dev.off()
    } else {

    ##plotting measured signal
    points(x,y,pch=20, col="grey")

    ##add additional labeling (fitted function)
    mtext(side=3, sample_code, cex=0.7*cex.global)

    ##plot sum function
    if (!inherits(fit, "try-error")) {
      lines(x,eval(fit.function), lwd=2, col="black")
      legend.caption<-"sum curve"
      curve.col <- 1

      ##plot signal curves

      ##plot curve for additional parameters
      for (i in seq_along(I0)) {
          curve(I0[i]*lambda[i]*exp(-lambda[i]*x),col=col[i+1],
                lwd = 2,
                add = TRUE)
          legend.caption<- c(legend.caption, paste("component", i))
          curve.col<-c(curve.col,i+1)
      }
      ##plot legend
      #legend(y=max(y)*1,"measured values",pch=20, col="gray", bty="n")
      legend("topright", legend.caption, lty = rep_len(1, n.components + 1),
             lwd = 2, col = col[curve.col], bty = "n")

      ##==lower plot==##
      ##plot residuals
      par(mar=c(4.2,4,0,0))
      plot_check2 <- try(plot(x,residuals(fit),
           xlim=c(min(x),max(x)),
           xlab=xlab,
           type="l",
           col="grey",
           ylab="Residual [a.u.]",
           lwd=2,
           log = gsub("y", "", log)
      ), silent = TRUE)

      if (inherits(plot_check2, "try-error")) {
        ## reset the graphic device if plotting failed
        .throw_message("Figure margins too large or plot area too small, ",
                       "nothing plotted")
        grDevices::dev.off()
      } else {

      ##add 0 line
      abline(h=0)

      ##------------------------------------------------------------------------##
      ##++component to sum contribution plot ++##
      ##------------------------------------------------------------------------##

      ##plot component contribution to the whole signal
      #open plot area
      par(mar=c(4,4,3.2,0))
      plot(NA,NA,
           xlim=c(min(x),max(x)),
           ylim=c(0,100),
           ylab="Contribution [%]",
           xlab=xlab,
           main="Component contribution to sum curve",
           log = gsub("y", "", log))

      stepping <- seq(3,length(component.contribution.matrix[1,]),2)

      for(i in 1:length(I0)){
        polygon(c(component.contribution.matrix[,1],
                  component.contribution.matrix[,2]),
                c(component.contribution.matrix[,stepping[i]],
                  component.contribution.matrix[,stepping[i]+1]),
                col = col[i+1])
      }
      rm(stepping)
      } # end if (plot_check2)

    } else if (verbose) {
        .throw_message("Fitting failed, plot without fit produced")
    }#end if try-error for fit
    } # end if (plot_check)
  }

  ##============================================================================##
  ## Return Values
  ##============================================================================##

  if (!method_control$export.comp.contrib.matrix) {
    component.contribution.matrix <- NA
  }
  set_RLum(
    class = "RLum.Results",
    data = list(
      data = output.table,
      fit = fit,
      component.contribution.matrix = list(component.contribution.matrix)
    ),
    info = list(call = sys.call())
  )
}

## load data
data(ExampleData.CW_OSL_Curve, envir = environment())
data(ExampleData.FittingLM, envir = environment())
curve <- set_RLum("RLum.Data.Curve",
                  data = as.matrix(ExampleData.CW_OSL_Curve),
                  curveType = "measured",
                  recordType = "OSL")
curve@data <- curve@data[1:90, ]

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fit_CWCurve("error"),
               "'object' should be of class 'RLum.Data.Curve' or 'data.frame'")
  expect_error(fit_CWCurve(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(fit_CWCurve(iris[, 1, drop = FALSE]),
               "'object' should have 2 columns")
  expect_error(fit_CWCurve(data.frame(a = 1:10, b = NA)),
               "'object' contains no positive counts")
  expect_error(fit_CWCurve(set_RLum("RLum.Data.Curve")),
               "'object' contains no positive counts")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve, fit.method = "error"),
               "'fit.method' should be one of 'port' or 'LM'")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve, n.components.max = 0),
               "'n.components.max' should be a single positive integer value")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve, fit.failure_threshold = -1),
               "'fit.failure_threshold' should be a single positive integer value")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve, verbose = "error"),
               "'verbose' should be a single logical value")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve, output.terminalAdvanced = "error"),
               "'output.terminalAdvanced' should be a single logical value")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve, plot = "error"),
               "'plot' should be a single logical value")
  expect_error(fit_CWCurve(ExampleData.CW_OSL_Curve[5:1, ]),
               "Time values are not ordered")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  SW({
  ## data.frame
  expect_snapshot_RLum(fit_CWCurve(ExampleData.CW_OSL_Curve,
                                   n.components.max = 3,
                                   fit.method = "LM",
                                   plot = FALSE),
                       tolerance = snapshot.tolerance)

  ## RLum.Data.Curve object
  expect_snapshot_RLum(fit_CWCurve(curve,
                                   n.components.max = 2,
                                   fit.calcError = TRUE,
                                   method_control = list(export.comp.contrib.matrix = TRUE),
                                   verbose = FALSE,
                                   plot = FALSE),
                       tolerance = snapshot.tolerance)
  })
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  ## FIXME(mcol): n.components.max is set to 2 due to failures in CI that
  ## are not reproducible on Ubuntu 22.04 with R 4.5.1 or R-devel: the Ubuntu
  ## 24.04 with R-devel on CI tend to pick more than 2 components, while
  ## locally only 2 are chosen even when the maximum is higher
  SW({
  vdiffr::expect_doppelganger("default",
                              fit_CWCurve(ExampleData.CW_OSL_Curve,
                                          n.components.max = 2))
  vdiffr::expect_doppelganger("logx cex",
                              fit_CWCurve(ExampleData.CW_OSL_Curve,
                                          main = "CW Curve Fit",
                                          n.components.max = 2,
                                          cex.global = 2,
                                          log = "x"))
  })
})

test_that("more coverage", {
  testthat::skip_on_cran()

  expect_message(fit_CWCurve(ExampleData.CW_OSL_Curve[1, ]),
                 "Error: Fitting failed, plot without fit produced")

  pdf(tempfile(), width = 1, height = 1)
  expect_message(fit_CWCurve(values.curve, n.components.max = 3,
                             verbose = FALSE),
                 "Figure margins too large or plot area too small")
  pdf(tempfile(), width = 1, height = 1)
  expect_message(fit_CWCurve(ExampleData.CW_OSL_Curve, verbose = FALSE),
                 "Figure margins too large or plot area too small")
  expect_error(fit_CWCurve(data.frame(NA, 1:5)),
               "0 (non-NA) cases", fixed = TRUE)

  ## deprecated argument
  SW({
  expect_warning(fit_CWCurve(values = ExampleData.CW_OSL_Curve,
                             n.components.max = 2),
                 "'values' was deprecated in v1.2.0, use 'object' instead")
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 509
  SW({
  expect_message(fit_CWCurve(ExampleData.CW_OSL_Curve[1:20, ],
                             fit.method = "LM", fit.calcError = TRUE),
                 "Error: Computation of confidence interval failed")
  })

  ## issue 953
  SW({
  fit_CWCurve(ExampleData.CW_OSL_Curve[1:2, ], fit.trace = TRUE)
  })
})

#'@title Estimate Individual Age using Bayesian Inference
#'
#'@description A Bayesian robust estimation of central age from equivalent dose
#' measurements under the assumption that the dose rate is modelled by a
#' finite Gaussian mixture model.
#'
#'@param theta [numeric] (**required**): the weight vector of the Gaussian mixture
#'
#'@param mu [numeric] (**required**): is the mean vector of the Gaussian mixture
#'
#'@param sigma [numeric] (**required**): is the standard deviation vector of the Gaussian mixture
#'
#'@param De [numeric] (**required**): the equivalent dose sample
#'
#'@param s [numeric] (**required**): the vector of measurement errors on De.
#'
#'@param sig0 [numeric] (**required**): the prior shrinkage parameter
#'
#'@param Age_range [numeric] (*with default*): the age range to investigate
#'
#'@param method_control [list] (*with default*): parameters passed down
#' to the jags process
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#'@return An [Luminescence::RLum.Results-class] object to be used in [Luminescence::combine_De_Dr]
#'
#'@section Function version: 0.1.0
#'
#'@note The function is intended to be called by [Luminescence::combine_De_Dr], however, for
#' reasons of transparency
#'
#'@author Anne Philippe, Université de Nantes (France),
#' Jean-Michel Galharret, Université de Nantes (France),
#' Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France),
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@examples
#' n <- 1000
#' sdt <- 0.3
#' Dr <- stats::rlnorm (n, 0, sdt)
#' int_OD <-  0.1
#' tildeDr = Dr * (1 + rnorm(length(Dr), 0, int_OD))
#' De <-  c(50 * sample(Dr, 50, replace = TRUE), 10, 12, 200, 250)
#' k <- length(De)
#' s <- stats::rnorm(k, 10, 2)
#' a <- De / mean(tildeDr)
#' sig_a2 <- a ^ 2 * (s / De) ^ 2
#' sig0 <- sqrt(1 / mean(1 / sig_a2))
#' fit <- mclust::Mclust(tildeDr, model = "V")
#' theta <- fit$parameters$pro
#' mu <- fit$parameters$mean
#' sigma <- sqrt(fit$parameters$variance$sigmasq)
#' Age_range <- c(0, Dr * (1 + rnorm(length(Dr), 0, int_OD[1])))
#' res <- .calc_IndividualAgeModel(theta, mu, sigma, De, s, sig0, Age_range = Age_range)
#'@noRd
.calc_IndividualAgeModel <- function(
  theta,
  mu,
  sigma,
  De,
  s,
  sig0,
  Age_range,
  method_control = list(),
  verbose = TRUE
){

  # Set parameters and models -----------------------------------------------
  nobs <- length(De)
  event1 <- "model{
            for( i in 1 : N ) {
              D_e[i] ~ dnorm(a[i] * mu, tau[i])
              tau[i] <- 1 / (a[i] * sigma) ^ 2
              De[i] ~ dnorm(D_e[i], prec2[i])
              a[i] ~ dnorm(A, prec_a[i])
              u[i] ~ dunif(0,1)
              prec_a[i] <- 1 / s02 * u[i] / (1 - u[i])
              prec2[i] <- 1 / (s2[i])
              sig_a[i] <- 1 / sqrt(prec_a[i])
            }
          A ~ dunif(Amin, Amax)
        }"

  event2 <- "model{
            for( i in 1 : N ) {
              D_e[i] ~ dnorm(a[i] * mu[z[i]], tau[i])
              tau[i] <- 1 / (a[i] * sigma[z[i]]) ^ 2
              z[i] ~ dcat(theta)
              De[i] ~ dnorm(D_e[i], prec2[i])
              a[i] ~ dnorm(A, prec_a[i])
              u[i] ~ dunif(0, 1)
              prec_a[i] <- 1 / s02 * u[i] / (1 - u[i])
              prec2[i] <- 1 / (s2[i])
              sig_a[i] <- 1 / sqrt(prec_a[i])
            }
            A ~ dunif(Amin, Amax)
          }"

  data1 <- list(
    theta = theta,
    mu = mu,
    sigma = sigma,
    N = nobs,
    De = De,
    s2 = s^2,
    s02 = sig0[1]^2,
    Amin = Age_range[1],
    Amax = Age_range[2])

  # Run Bayesian model ------------------------------------------------------
  method_control <- modifyList(
    x = list(
      variable.names = c('A', 'a', 'sig_a'),
      n.chains = 4,
      n.adapt = 1000,
      n.iter = 5000,
      thin = 1,
      progress.bar = if(verbose) "text" else "none",
      quiet = !verbose,
      diag = verbose,
      return_mcmc = FALSE
    ),
    val = method_control)

  on.exit(close(model), add = TRUE)
  ## select model
  if(length(theta) == 1) {
    data1$theta <- NULL
    model <- textConnection(event1)
  } else {
    model <- textConnection(event2)
  }

  ## run model
  if(verbose) cat("(1) Running Bayesian modelling 'Individual Age Model' ... ")

  jags <- rjags::jags.model(
    file = model,
    data = data1,
    inits = method_control$inits,
    n.chains = method_control$n.chains,
    n.adapt = method_control$n.adapt,
    quiet = method_control$quiet
  )

  stats::update(
    jags,
    n.iter = method_control$n.iter,
    progress.bar = method_control$progress.bar,
    quiet = method_control$quiet
  )

  samp <-
    rjags::coda.samples(
      model = jags,
      variable.names = method_control$variable.names,
      n.iter = method_control$n.iter,
      thin = method_control$thin,
      progress.bar = method_control$progress.bar
    )

  if(verbose & method_control$quiet) cat("DONE")
  if(method_control$diag) {
    cat("\n[.calc_IndividualAgeModel()]\n")
    print(coda::gelman.diag(samp))
  }

  # Return ------------------------------------------------------------------
  return(set_RLum(
    "RLum.Results",
    data = list(
      A = unlist(samp[, "A"]),
      a = do.call(rbind, samp[, 2:(nobs + 1)]),
      sig_a = do.call(rbind, samp[, (2 + nobs):(2 * nobs + 1)]),
      model = paste(jags$model(), ""),
      mcmc_IAM = if(method_control$return_mcmc) samp else NULL),
    info = list(call = sys.call())
  ))
}


#'@title Central Bayesian Central Age Model
#'
#'@description A Bayesian estimation of central age from equivalent dose measurements
#'under the assumption that the dose rate is modelled by finite Gaussian mixture model.
#'MCMC outputs provide to JAGS program.
#'
#'@param theta [numeric] (**required**): the weight vector of the Gaussian mixture
#'
#'@param mu [numeric] (**required**): is the mean vector of the Gaussian mixture
#'
#'@param sigma [numeric] (**required**): is the standard deviation vector of the Gaussian mixture
#'
#'@param De [numeric] (**required**): the equivalent dose sample
#'
#'@param s [numeric] (**required**): the vector of measurement errors on De.
#'
#'@param Age_range [numeric] (*with default*): the age range to investigate
#'
#'@param method_control [list] (*with default*): parameters passed down to the jags process
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#'@return An [Luminescence::RLum.Results-class] object
#'
#'@section Function version: 0.1.0
#'
#'@author Anne Philippe, Université de Nantes (France),
#'Jean-Michel Galharret, Université de Nantes (France),
#'Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France),
#'Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@noRd
.calc_BayesianCentralAgeModel <- function(
  theta,
  mu,
  sigma,
  De,
  s,
  Age_range,
  method_control = list(),
  verbose = TRUE
){

  # Set models --------------------------------------------------------------
  central_age_model1 <- "
    model{
        for( i in 1:J) {
        D_e[i] ~ dnorm(A*mu,1/(A*sigma)^2)
        De[i] ~ dnorm(D_e[i], prec2[i])
        prec2[i] <- 1/(s2[i])
        }

        A ~ dunif(Amin,Amax)
    }"

  central_age_model2 <- "
    model{
        for( i in 1:J) {
        D_e[i] ~ dnorm(A*mu[z[i]],tau[i])
        tau[i]<-1/(A*sigma[z[i]])^2
        z[i] ~ dcat(theta)
        De[i] ~ dnorm(D_e[i], prec2[i])
        prec2[i] <- 1/(s2[i])
        }

        A ~ dunif(Amin,Amax)
    }"

  # Run Bayesian modelling --------------------------------------------------
  method_control <- modifyList(
    x = list(
      variable.names = c('A', 'D_e'),
      n.chains = 4,
      n.adapt = 1000,
      n.iter = 5000,
      thin = 1,
      progress.bar = if(verbose) "text" else "none",
      quiet = !verbose,
      diag = verbose,
      return_mcmc = FALSE
    ),
    val = method_control)

  on.exit(close(model), add = TRUE)
  data <- list(theta = theta,
               mu = mu,
               sigma = sigma,
               De = De,
               J =  length(De),
               s2 = s^2,
               Amin = Age_range[1],
               Amax = Age_range[2])

  ## select model
  if(length(theta) == 1) {
    data$theta <- NULL
    model <- textConnection(central_age_model1)
  } else {
    model <- textConnection(central_age_model2)
  }

  ## run modelling
  if(verbose) cat("\n(2) Running Bayesian modelling 'Bayesian Central Age Model' ... ")

  jags2 <- rjags::jags.model(
    file = model,
    data = data,
    inits = method_control$inits,
    n.chains = method_control$n.chains,
    n.adapt = method_control$n.adapt,
    quiet = method_control$quiet
  )

  stats::update(
    object = jags2,
    n.iter = method_control$n.iter,
    progress.bar = method_control$progress.bar,
    quiet = method_control$quiet
  )

  samp2 <- rjags::coda.samples(
    model = jags2,
    variable.names = method_control$variable.names,
    n.iter = method_control$n.iter,
    thin = method_control$thin,
    progress.bar = method_control$progress.bar
  )

  if(verbose & method_control$quiet) cat("DONE\n")

  if(method_control$diag) {
    cat("\n[.calc_BayesianCentralAgeModel()]\n")
    print(coda::gelman.diag(samp2))
  }

  # Return ------------------------------------------------------------------
  return(set_RLum(
    "RLum.Results",
    data = list(
      A = unlist(samp2[, "A"]),
      D_e = do.call(rbind, samp2[, -1]),
      model = paste(jags2$model(), ""),
      mcmc_BCAM = if(method_control$return_mcmc) samp2 else NULL),
    info = list(call = sys.call())
  ))
}

#'@title Combine Dose Rate and Equivalent Dose Distribution
#'
#'@description A Bayesian statistical analysis of OSL age requiring dose rate sample.
#'Estimation contains a preliminary step for detecting outliers in the equivalent
#'dose sample.
#'
#'@details
#'
#'**Outlier detection**
#'
#'Two different outlier detection methods are implemented (full details are given
#'in the cited literature).
#'
#'1. The *default* and recommend method, uses quantiles to compare prior and
#'posterior distributions of the individual variances of the equivalent doses.
#'If the corresponding quantile in the corresponding posterior distribution is larger
#'than the quantile in the prior distribution, the value is marked
#'as outlier (cf. Galharret et al., preprint)
#'
#'2. The alternative method employs the method suggested by Rousseeuw and Croux (1993)
#'using the absolute median distance.
#'
#'**Parameters available for `method_control`**
#'
#'The parameters listed below are used to granular control Bayesian modelling using
#'[rjags::rjags]. Internally the functions `.calc_IndividualAgeModel()` and
#'`.calc_BayesianCentraAgelModel()`. The parameter settings affect both models.
#'Note: `method_control` expects a **named** list of parameters
#'
#'\tabular{llll}{
#' **PARAMETER** \tab **TYPE** \tab **DEFAULT** \tab **REMARKS** \cr
#' `variable.names_IAM` \tab [character] \tab `c('A', 'a', 'sig_a')` \tab variables names to be monitored in the modelling process using the internal function `.calc_IndividualAgeModel()`\cr
#' `variable.names_BCAM` \tab [character] \tab `c('A', 'D_e')` \tab variables names to be monitored in the modelling process using the internal function `.calc_BayesianCentraAgelModel()`\cr
#' `n.chains` \tab [integer] \tab `4` \tab number of MCMC chains\cr
#' `n.adapt` \tab [integer] \tab `1000` \tab number of iterations for the adaptation\cr
#' `n.iter` \tab [integer] \tab `5000` \tab number of iterations to monitor cf. [rjags::coda.samples]\cr
#' `thin` \tab [numeric] \tab `1` \tab thinning interval for the monitoring cf. [rjags::coda.samples]\cr
#' `diag` \tab [logical] \tab `FALSE` \tab additional terminal convergence diagnostic.
#' `FALSE` if `verbose = FALSE`\cr
#' `progress.bar` \tab [logical] \tab `FALSE` \tab enable/disable progress bar. `FALSE` if `verbose = FALSE`\cr
#' `quiet` \tab [logical] \tab `TRUE` \tab silence terminal output. Set to `TRUE` if `verbose = FALSE`\cr
#' `return_mcmc`\tab [logical] \tab `FALSE` \tab return additional MCMC diagnostic information\cr
#'}
#'
#' @param De [numeric] (**required**):
#' an equivalent dose sample.
#'
#'@param s [numeric] (**required**): a vector of measurement errors on the equivalent dose
#'
#'@param Dr [numeric] (**required**): a dose rate sample
#'
#'@param int_OD [numeric] (**required**): the intrinsic overdispersion, typically the standard deviation
#'characterizing a dose-recovery test distribution
#'
#' @param Age_range [numeric] (*with default*):
#' age range to be investigated by the algorithm. The larger the value, the
#' more iterations are needed and the longer it takes. It should not be set
#' too narrow to cut the algorithm some slack.
#'
#'@param outlier_threshold [numeric] (*with default*): the required significance level used
#'for the outlier detection. If set to `1`, no outliers are removed. If
#'`outlier_method = "RousseeuwCroux1993"`, the median distance is used as outlier threshold.
#'Please see details for further information.
#'
#' @param outlier_method [character] (*with default*):
#' outlier detection method, either `"default"` or `"RousseeuwCroux1993"`.
#' See details.
#'
#' @param outlier_analysis_plot [logical] (*with default*): enable/disable the
#' outlier analysis plot. Note: the outlier analysis will happen independently
#' of the plot output.
#'
#' @param method_control [list] (*with default*):
#' named [list] of parameters to control [rjags::jags.model]. This can be
#' used to set the random seed for the MCMC chains (four by default):
#' ```
#' method_control = list(inits = list(
#'   list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 1),
#'   list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 2),
#'   list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 3),
#'   list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 4))
#' )
#' ```
#'
#'@param par_local [logical] (*with default*): if set to `TRUE` the function uses its
#'own [graphics::par] settings (which will end in two plots next to each other)
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#'@param ... a few further arguments to fine-tune the plot output such as
#'`cdf_ADr_quantiles` (`TRUE`/`FALSE`), `legend.pos`, `legend` (`TRUE`/`FALSE`)
#'
#'@return The function returns a plot if `plot = TRUE` and an [Luminescence::RLum.Results-class]
#'object with the following slots:
#'
#' `@data`\cr
#' `.. $Ages`: a [numeric] vector with the modelled ages to be further analysed or visualised\cr
#' `.. $Ages_stats`: a [data.frame] with sum HPD, CI 68% and CI 95% for the ages \cr
#' `.. $outliers_index`: the index with the detected outliers\cr
#' `.. $cdf_ADr_mean` : empirical cumulative density distribution A * Dr (mean)\cr
#' `.. $cdf_ADr_quantiles` : empirical cumulative density distribution A * Dr (quantiles .025,.975)\cr
#' `.. $cdf_De_no_outlier` : empirical cumulative density distribution of the De with no outliers\cr
#' `.. $cdf_De_initial` : empirical cumulative density distribution of the initial De\cr
#' `.. $mcmc_IAM` : the MCMC list of the Individual Age Model, only of `method_control = list(return_mcmc = TRUE)` otherwise `NULL`\cr
#' `.. $mcmc_BCAM` : the MCMC list of the Bayesian Central Age Model, only of `method_control = list(return_mcmc = TRUE)` otherwise `NULL`\cr
#'
#' `@info`\cr
#' `.. $call`: the original function call\cr
#' `.. $model_IAM`: the BUGS model used to derive the individual age\cr
#' `.. $model_BCAM`: the BUGS model used to calculate the Bayesian Central Age\cr
#'
#' @references
#'
#' Mercier, N., Galharret, J.-M., Tribolo, C., Kreutzer, S., Philippe, A., (2022).
#' Luminescence age calculation through Bayesian convolution of equivalent dose
#' and dose-rate distributions: the De_Dr model. Geochronology 4, 297-310.
#' \doi{10.5194/gchron-4-297-2022}
#'
#' Galharret, J-M., Philippe, A., Mercier, N. (2021). Detection of outliers
#' with a Bayesian hierarchical model: application to the single-grain
#' luminescence dating method. Electronic Journal of Applied Statistics 14 (2),
#' 318-338. \doi{10.1285/i20705948v14n2p318}
#'
#'**Further reading**
#'
#'Rousseeuw, P.J., Croux, C., 1993. Alternatives to the median absolute deviation.
#'Journal of the American Statistical Association 88, 1273–1283. \doi{10.2307/2291267}
#'
#'Rousseeuw, P.J., Debruyne, M., Engelen, S., Hubert, M., 2006. Robustness and outlier detection in chemometrics.
#'Critical Reviews in Analytical Chemistry 36, 221–242. \doi{10.1080/10408340600969403}
#'
#'@author Anne Philippe, Université de Nantes (France),
#'Jean-Michel Galharret, Université de Nantes (France),
#'Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France),
#'Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@seealso [Luminescence::plot_OSLAgeSummary], [rjags::rjags], [mclust::mclust-package]
#'
#'@section Function version: 0.1.0
#'
#'@keywords dplot distribution datagen
#'
#'@examples
#'## set parameters
#' Dr <- stats::rlnorm (1000, 0, 0.3)
#' De <- 50*sample(Dr, 50, replace = TRUE)
#' s <- stats::rnorm(50, 10, 2)
#'
#'## run modelling
#'## note: modify parameters for more realistic results
#'\dontrun{
#'results <- combine_De_Dr(
#'  Dr = Dr,
#'  int_OD = 0.1,
#'  De,
#'  s,
#'  Age_range = c(0,100),
#'  method_control = list(
#'   n.iter = 100,
#'   n.chains = 1))
#'
#'## show models used
#'writeLines(results@info$model_IAM)
#'writeLines(results@info$model_BCAM)
#'}
#'
#'@export
combine_De_Dr <- function(
  De,
  s,
  Dr,
  int_OD,
  Age_range = c(1,300),
  outlier_threshold = 0.05,
  outlier_method = "default",
  outlier_analysis_plot = FALSE,
  method_control = list(),
  par_local = TRUE,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("combine_De_Dr")
  on.exit(.unset_function_name(), add = TRUE)

  ## check package availability ---------------------------------------------
  .require_suggested_package("rjags")
  .require_suggested_package("coda")
  .require_suggested_package("mclust")

  ## Integrity checks -------------------------------------------------------
  .validate_class(De, "numeric")
  .validate_class(s, "numeric")
  .validate_class(Dr, "numeric")
  if (length(De) != length(s))
    .throw_error("'De' and 's' should have the same length")
  .validate_nonnegative_scalar(int_OD)
  .validate_class(Age_range, "numeric", length = 2)
  if (anyNA(Age_range))
    .throw_error("'Age_range' cannot contain missing values")
  .validate_positive_scalar(outlier_threshold)
  .validate_class(outlier_method, "character", length = 1)
  .validate_logical_scalar(outlier_analysis_plot)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

# Prepare data ------------------------------------------------------------
  ## we have to fetch the function otherwise
  ## we would need it in import instead of suggests
  mclustBIC <- mclust::mclustBIC

  ##  Estimation of the rate dose Dr1 by a Gaussian Mixture Model
  tildeDr <- Dr * (1 + rnorm(length(Dr), 0, int_OD[1]))
  fit <-
    mclust::Mclust(data = tildeDr,
                   modelNames = "V",
                   verbose = FALSE)

  theta <- fit$parameters$pro
  mu <- fit$parameters$mean
  sigma <- sqrt(fit$parameters$variance$sigmasq)
  a <- De / mean(tildeDr)
  sig_a2 <- a ^ 2 * (s / De) ^ 2
  sig0 <- sqrt(1 / mean(1 / sig_a2))

# Set parameters ----------------------------------------------------------
  method_control <- modifyList(
    x = list(
      variable.names_IAM = c('A', 'a', 'sig_a'),
      variable.names_BCAM = c('A', 'D_e'),
      n.chains = 4,
      n.adapt = 1000,
      n.iter = 5000,
      thin = 1,
      progress.bar = "none",
      quiet = TRUE,
      diag = FALSE,
      return_mcmc = FALSE
    ),
    val = method_control)

# Bayesian Modelling IAM --------------------------------------------------
if(verbose)  cat("\n[combine_De_Dr()]\n")

fit_IAM <- .calc_IndividualAgeModel(
      theta = theta,
      mu = mu,
      sigma = sigma,
      De = De,
      s = s,
      sig0 = sig0,
      Age_range = Age_range[1:2],
      verbose = verbose,
      method_control = list(
          variable.names = method_control$variable.names_IAM,
          n.chains = method_control$n.chains,
          n.adapt = method_control$n.adapt,
          n.iter = method_control$n.iter,
          inits = method_control$inits,
          thin = method_control$thin,
          progress.bar = method_control$progress.bar,
          quiet = method_control$quiet,
          diag = method_control$diag,
          return_mcmc = method_control$return_mcmc)
      )

# Outlier detection -------------------------------------------------------
  ## set threshold for outliers
  alpha <- outlier_threshold[1]

  ## apply method ... default is method develop by Jean-Michel and Anne
  if(outlier_method == "RousseeuwCroux1993") {
    ## calculate the median of the sig_a
    xj <- log(matrixStats::colMedians(fit_IAM$sig_a))
    MAD <- 1.483 * median(abs(xj - median(xj)))
    test <- (xj - median(xj)) / MAD
    out <- sort(which(test > alpha))

  } else {
    sig_max <- sig0 * ((1 - alpha) / alpha) ^ 0.5
    test <- vapply(1:length(De), function(j){
      mean(fit_IAM$sig_a[, j] >= sig_max)
    }, numeric(1))

    out <- sort(which(test > alpha))
  }

  ##some terminal output
  if (verbose && length(out) > 0) {
    cat("\n    >> Outliers detected: ", length(out), "/", length(De),
        " (", round(length(out) / length(De) * 100, 1), "%)", sep = "")
  }

  ## apply the removal
  if (length(out) == 0) {
      De1 <- De
      s1 <- s
   } else {
      De1 <- De[-out]
      s1 <- s[-out]
   }

# Bayesian modelling BCAM -------------------------------------------------
  fit_BCAM <- .calc_BayesianCentralAgeModel(
      theta,
      mu,
      sigma,
      De = De1,
      s = s1,
      Age_range = Age_range,
      verbose = verbose,
      method_control = list(
        variable.names = method_control$variable.names_BCAM,
        n.chains = method_control$n.chains,
        n.adapt = method_control$n.adapt,
        n.iter = method_control$n.iter,
        inits = method_control$inits,
        thin = method_control$thin,
        progress.bar = method_control$progress.bar,
        quiet = method_control$quiet,
        diag = method_control$diag,
        return_mcmc = method_control$return_mcmc)
    )

# Calculate EDFC -------------------------------------------------
  ## calculate various parameters
  D_e <- fit_BCAM$D_e
  A2 <- fit_BCAM$A

  ## calculate bandwidths
  h <- density(De)$bw
  h1 <- density(De1)$bw

  t <- seq(min(D_e), max(D_e), length.out = min(1000, round(max(D_e) - min(D_e), 0)))

  ind <- min(5000, length(A2))
  subsamp <- sample(1:length(A2), ind, replace = FALSE)
  cdf_ADr <- matrix(0, nrow = ind, ncol = length(t))

  ## De distribution re-sampled without outliers -> De2
  ## De distribution re-sampled initial -> De3
  De2 <-
    rnorm(length(subsamp), sample(De1, size = length(subsamp), replace = TRUE), h1)
  De3 <-
    rnorm(length(subsamp), sample(De, size = length(subsamp), replace = TRUE), h)

  ## calculate ecdf
  cdf_De_no_outlier<- stats::ecdf(De2)(t)
  cdf_De_initial <- stats::ecdf(De3)(t)

  for (i in 1:ind)
    cdf_ADr[i, ] <- stats::ecdf(A2[subsamp[i]] * tildeDr)(t)

  ## calculate mean value and quantiles for the ecdf A * Dr
  cdf_ADr_mean <- matrixStats::colMeans2(cdf_ADr)
  cdf_ADr_quantiles <- matrixStats::colQuantiles(cdf_ADr, probs = c(0.025, 0.975))

  ## further values to ease the interpretation
  d <- density(fit_BCAM$A)
  HPD <- d$x[which.max(d$y)[1]]
  CI_68 <- .calc_HPDI(fit_BCAM$A, prob = 0.68)
  CI_95 <- .calc_HPDI(fit_BCAM$A, prob = 0.95)

# Additional terminal output ----------------------------------------------
if(verbose){
  cat("(3) Age results (presumably in ka) \n")
  cat("    -----------------------------------\n")
  cat("    Age (HPD)   :\t", format(round(HPD,2), nsmall = 2), "\n")
  cat("    Age (CI 68%):\t", .format_range(round(CI_68, 2), sep = " : ", nsmall = 2), "\n")
  cat("    Age (CI 95%):\t", .format_range(round(CI_95, 2), sep = " : ", nsmall = 2), "\n")
  cat("    -----------------------------------\n")
}

# Plotting ----------------------------------------------------------------
if(plot){
  ##check incoming plot settings
  plot_settings <- modifyList(x = list(
    cdf_ADr_quantiles = FALSE,
    legend = TRUE,
    legend.pos = "bottomright"
  ), list(...))

  ##make sure we reset plots
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  if(outlier_analysis_plot){
    N <- length(De)

    ##plot with outliers
    graphics::boxplot(fit_IAM$sig_a, outline = FALSE,
            col = (abs(as.numeric(
              1:length(De) %in% out
            ) - 1) + 2),
    main = "Outlier detection",
    xaxt = "n",
    xlab = expression(paste("Index of ", sigma[a])))

    ## add axis
    axis(side = 1, at = 1:length(De))
    mtext(
      text = paste0(length(out), "/", N, " (", round(length(out) / N * 100, 1), "%)"),
      side = 3,
      cex = 0.8
    )
    abline(h = sig0, col = "violet")

    ##plot sd of outliers
    if(length(out) > 0){
      graphics::boxplot(fit_IAM$sig_a[, out],
              outline = FALSE,
              names = out,
              ylab = "Individual sd [a.u.]",
              main = "Outliers: posterior distr.")

    abline(h = sig0, col = "violet")

    } else {
      shape::emptyplot()
      text(0.5, 0.5, "No outlier detected!")
    }
  }

  ##plot age summary
  plot_OSLAgeSummary(
    object = fit_BCAM,
    level = 0.68,
    rug = FALSE,
    polygon_col = rgb(100, 149, 237, 75, maxColorValue = 255),
    verbose = FALSE
  )

  ## open plot area
  plot(NA,
    xlim = range(t),
    ylim = c(0, 1),
    ylab = "ecdf (mean)",
    xlab = "Dose [Gy]",
    main= "ECDF")

  ## add quantile range (only for A * Dr)
  if(plot_settings$cdf_ADr_quantiles){
    polygon(
     x = c(t, rev(t)),
     y = c(cdf_ADr_quantiles[,1], rev(cdf_ADr_quantiles[,2])),
     col = rgb(1,0,0,0.2), lty = 0)
  }

  ##add mean lines for the ecdfs
  lines(t, cdf_ADr_mean, col = 2, lty = 1, lwd = 2)
  lines(t, cdf_De_no_outlier, type = "l", col = 3, lty = 2, lwd = 2)
  lines(t, cdf_De_initial, type = "l", col = 4, lty = 3, lwd = 2)

  if(plot_settings$legend){
    legend(
      plot_settings$legend.pos,
      legend = c(
        expression(A %*% Dr),
        expression(paste(D[e], " no outliers")),
        expression(paste(D[e], " initial"))),
      lty = c(1,2,3),
      bty = "n",
      col = c(2,3,4),
      cex = 0.8)
  }
}

# Return results ----------------------------------------------------------
  set_RLum(
    "RLum.Results",
    data = list(
      Ages = fit_BCAM$A,
      Ages_stats = data.frame(
        HPD = HPD,
        CI_68_lower = CI_68[1],
        CI_68_upper = CI_68[2],
        CI_95_lower = CI_95[1],
        CI_95_upper = CI_95[2]),
      outliers_index = out,
      cdf_ADr_mean = cdf_ADr_mean,
      cdf_ADr_quantiles = cdf_ADr_quantiles,
      cdf_De_no_outlier = cdf_De_no_outlier,
      cdf_De_initial = cdf_De_initial,
      mcmc_IAM = fit_IAM$mcmc_IAM,
      mcmc_BCAM = fit_BCAM$mcmc_BCAM
      ),
    info = list(
      call = sys.call(),
      model_IAM = fit_IAM$model,
      model_BCAM = fit_BCAM$model)
  )
}

## --------------------------------------------------------------------------
## helper functions for snapshotting
##

## the ... can be used to set the tolerance
expect_snapshot_RLum <- function(object, ...) {
  if (inherits(object, "list")) {
    for (idx in seq_along(object))
      expect_snapshot_RLum(object[[idx]])
    return()
  }
  object@.uid <- NA_character_
  object@.pid <- NA_character_
  object@info[names(object@info) == "call"] <- NULL # may be multiple "call"
  if ("data" %in% slotNames(object)) {
    if ("fit" %in% names(object@data))
      object@data$fit <- NULL
    if ("call" %in% names(object@data))
      object@data$call <- NULL
    if ("info" %in% names(object@data)) {
      if ("call" %in% names(object@data$info))
        object@data$info$call <- NULL
    }
    if ("fits" %in% names(object@data)) { # calc_Huntley2006()
      if ("simulated" %in% names(object@data$fits)) {
        object@data$fits$simulated$m <- NULL
        object@data$fits$simulated$call <- NULL
        object@data$fits$simulated$weights <- NULL # for macos/windows CI
        object@data$fits$simulated$convInfo$finIter <- NULL
      }
      if ("measured" %in% names(object@data$fits)) {
        object@data$fits$measured$m <- NULL
        object@data$fits$measured$call <- NULL
      }
      if ("unfaded" %in% names(object@data$fits)) {
        object@data$fits$unfaded$m <- NULL
        object@data$fits$unfaded$call <- NULL
        object@data$fits$unfaded$convInfo$finIter <- NULL # for macos
      }
    }
    if ("data" %in% names(object@data)) {
      if ("UID" %in% names(object@data$data))
        object@data$data$UID <- NULL
    }
    if ("Fit" %in% names(object@data))
      object@data$Fit <- NULL
    if ("Formula" %in% names(object@data))
      object@data$Formula <- NULL
    if ("LnLxTnTx.table" %in% names(object@data))
      object@data$LnLxTnTx.table$UID <- NULL
    if ("rejection.criteria" %in% names(object@data))
      object@data$rejection.criteria$UID <- NULL
    if ("test_parameters" %in% names(object@data))
      object@data$test_parameters$UID <- NULL
    if (object@originator %in% c("calc_MinDose", "calc_MaxDose")) {
      object@data$call <- NULL
      object@data$mle <- NULL
      object@data$profile <- NULL
      object@data$bootstrap$poly.fits <- NULL
      object@data$bootstrap$loess.fit <- NULL
    }
  }
  if ("info" %in% slotNames(object)) {
    if ("call" %in% names(object@info)) {
      object@info$call <- NULL
    }
    if ("file" %in% names(object@info)) {
      object@info$file <- NULL
    }
    if ("info" %in% names(object@info) && is.call(object@info$info)) {
      object@info[names(object@info) == "info"] <- NULL # may be multiple "info"
    }
  }
  if ("records" %in% slotNames(object)) {
    for (idx in seq_along(object@records)) {
      object@records[[idx]]@info$args <- NULL
      object@records[[idx]]@info$call <- NULL
      object@records[[idx]]@.uid <- NA_character_
      object@records[[idx]]@.pid <- NA_character_
    }
  }
  expect_snapshot_value(object, style = "json2", ...)
}

## wrapper for Risoe.BINfileData objects
expect_snapshot_Risoe <- function(object, ...) {
  attr(object, ".S3Class") <- NULL
  expect_snapshot_value(object, style = "json2", ...)
}

## wrapper for plain R objects, such as lists, data.frames, etc
expect_snapshot_plain <- function(object, ...) {
  expect_snapshot_value(object, style = "json2", ...)
}

df <- data.frame(LnTn = 2.361, LnTn.error = 0.087,
                 Lr1Tr1 = 2.744, Lr1Tr1.error = 0.091,
                 Dr1 = 34.4)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_gSGC(data = NA),
               "'data' should be of class 'data.frame'")
  expect_error(calc_gSGC(iris[0, ]),
               "'data' cannot be an empty data.frame")
  expect_error(calc_gSGC(data.frame(a = 1, b = 1, c = 1, d = 1, e = 1, f = 1)),
               "'data' is expected to have 5 columns")
  expect_error(calc_gSGC(df, gSGC.type = 3),
               "'gSGC.type' should be one of '0-250' or '0-450'")
  expect_error(calc_gSGC(df, gSGC.type = "error"),
               "'gSGC.type' should be one of '0-250' or '0-450'")
  expect_error(calc_gSGC(df, gSGC.type = set_RLum("RLum.Analysis")),
               "'gSGC.type' should be one of '0-250' or '0-450'")
  expect_error(calc_gSGC(df, n.MC = 0),
               "'n.MC' should be a single positive integer value")
  expect_error(calc_gSGC(df, verbose = NA),
               "'verbose' should be a single logical value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## standard test
  SW({
  expect_s4_class(calc_gSGC(data = df,
    gSGC.type = "0-450",
    plot = TRUE,
    verbose = TRUE
  ), "RLum.Results")

  pars <- list(A = 0, A.error = 0, D0 = 0, D0.error = 0,
               c = 0, c.error = 0, Y0 = 0, Y0.error = 0)
  expect_s4_class(calc_gSGC(data = df, gSGC.parameters = pars),
                  "RLum.Results")
  })

  ## break plot (no solution)
  df_break <- data.frame(LnTn = 0, LnTn.error = 0.087,
                   Lr1Tr1 = 0, Lr1Tr1.error = 0.091,
                   Dr1 = 34.4)
  SW({
    expect_warning(calc_gSGC(data = df_break), regexp = "No solution was found")
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  set.seed(seed = 1)
  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(calc_gSGC(df, plot = FALSE, verbose = FALSE),
                       tolerance = snapshot.tolerance)

  ## apply some random values for more coverage
  df1 <- data.frame(LnTn = 0.361, LnTn.error = 2.087,
                    Lr1Tr1 = 0.744, Lr1Tr1.error = 10.091,
                    Dr1 = 0.4)
  expect_snapshot_RLum(calc_gSGC(df1, plot = TRUE, verbose = FALSE),
                       tolerance = snapshot.tolerance)

  df2 <- data.frame(LnTn = 10.361, LnTn.error = 0.087,
                    Lr1Tr1 = 0.044, Lr1Tr1.error = 0.091,
                    Dr1 = 0.04)
  expect_snapshot_RLum(calc_gSGC(df2, plot = TRUE, verbose = FALSE),
                       tolerance = snapshot.tolerance)

  df3 <- data.frame(LnTn = 521440.0361, LnTn.error = 0.087,
                    Lr1Tr1 = 10.044, Lr1Tr1.error = -2.091,
                    Dr1 = 10.04)
  expect_snapshot_RLum(calc_gSGC(df3, plot = TRUE, verbose = FALSE),
                       tolerance = snapshot.tolerance)

  ## graphical snapshot tests
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("calc_gSGC expected",
                              fig = calc_gSGC(df))
  vdiffr::expect_doppelganger("calc_gSGC expected df1",
                              fig = calc_gSGC(df1))
  vdiffr::expect_doppelganger("calc_gSGC expected df2",
                              fig = calc_gSGC(df2))
  })
})

## load data
data(ExampleData.DeValues, envir = environment())

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_FiniteMixture(),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_FiniteMixture("test"),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(calc_FiniteMixture(data.frame(col = 1:10)),
               "'data' object must have two columns")
  expect_error(calc_FiniteMixture(data.frame(col = 0:10, 1:11)),
               "'data' must have only positive values in its De column")
  expect_error(calc_FiniteMixture(data.frame(col = c(1:10, NA), 1:11)),
               "'data' must have only positive values in its De column")
  expect_error(calc_FiniteMixture(ExampleData.DeValues$CA1),
               "'sigmab' should be a single positive value")
  expect_error(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 2),
               "'sigmab' must be a value between 0 and 1")
  expect_error(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2),
               "'n.components' should be of class 'integer' or 'numeric'")
  expect_error(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                                  n.components = 1),
               "'n.components' should be at least 2")
  expect_error(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                                  n.components = 2, pdf.sigma = "error"),
               "'pdf.sigma' should be one of 'sigmab' or 'se'")
  expect_error(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                                  n.components = 2, pdf.colors = "error"),
               "'pdf.colors' should be one of 'gray', 'colors' or 'none'")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  ## simple run
  SW({
  expect_snapshot_RLum(calc_FiniteMixture(
    ExampleData.DeValues$CA1,
    sigmab = 0.2,
    n.components = 2,
    grain.probability = TRUE,
    verbose = TRUE),
    tolerance = snapshot.tolerance)
  })

  expect_snapshot_RLum(calc_FiniteMixture(
    ExampleData.DeValues$CA1,
    sigmab = 0.2,
    n.components = 3:5,
    grain.probability = TRUE,
    verbose = FALSE),
    tolerance = snapshot.tolerance)

  ## more coverage
  SW({
  expect_warning(expect_message(calc_FiniteMixture(
    ExampleData.DeValues$CA1[2:9, ],
    sigmab = 0.1,
    n.components = 3,
    verbose = TRUE),
    "'n.components' specified only one component, nothing plotted"),
    "The model produced NA values: either the input data are inapplicable")
  })

  expect_output(calc_FiniteMixture(
    set_RLum("RLum.Results",
             data = list(data = ExampleData.DeValues$CA1)),
    sigmab = 0.62,
    n.components = 2:3,
    pdf.colors = "colors",
    verbose = TRUE),
    "No significant increase in maximum log-likelihood estimates")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              calc_FiniteMixture(
                                  ExampleData.DeValues$CA1,
                                  sigmab = 0.2,
                                  n.components = 2:3,
                                  grain.probability = TRUE,
                                  trace = TRUE,
                                  main = "Plot title",
                                  verbose = TRUE))
  vdiffr::expect_doppelganger("cex pdf.weights",
                              calc_FiniteMixture(
                                  ExampleData.DeValues$CA1,
                                  sigmab = 0.2,
                                  n.components = 2:3,
                                  grain.probability = TRUE,
                                  pdf.weights = FALSE,
                                  pdf.colors = "none",
                                  cex = 2))
  vdiffr::expect_doppelganger("NA values only pdf",
                              calc_FiniteMixture(
                                  ExampleData.DeValues$CA1,
                                  sigmab = 0.25,
                                  n.components = 2:4,
                                  pdf.colors = "colors",
                                  plot.proportions = FALSE,
                                  plot.criteria = FALSE))

  ## plot with plot_RLum.Results
  res <- calc_FiniteMixture(ExampleData.DeValues$CA1,
                            sigmab = 0.2,
                            n.components = 3:4,
                            grain.probability = FALSE,
                            trace = FALSE,
                            plot = FALSE,
                            main = "From RLum.Results",
                            verbose = FALSE)
  vdiffr::expect_doppelganger("from RLum.Results",
                              calc_FiniteMixture(res,
                                                 pdf.colors = "colors"))
  vdiffr::expect_doppelganger("from RLum.Results none",
                              calc_FiniteMixture(res,
                                                 pdf.colors = "none"))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  SW({
  ## issue 691
  expect_s4_class(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                                     n.components = c(2, 4), plot = FALSE),
                  "RLum.Results")
  expect_s4_class(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                                     n.components = 3:2, plot = FALSE),
                  "RLum.Results")

  ## issue 704
  expect_silent(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.2,
                                   n.components = 2:9, verbose = FALSE))

  ## issue 708
  expect_silent(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.57,
                                   n.components = 2:6, verbose = FALSE))

  ## issue 710
  expect_silent(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.54,
                                   n.components = 2:5, pdf.weight = FALSE,
                                   verbose = FALSE))

  ## issue 1458
  expect_silent(calc_FiniteMixture(ExampleData.DeValues$CA1, sigmab = 0.1,
                                   n.components = 1:3, verbose = FALSE))
  })
})

## load data
data(ExampleData.XSYG, envir = environment())

## subtract background
TL.Spectrum@data <- TL.Spectrum@data[] - TL.Spectrum@data[, 15]

## matrix input
mat <- get_RLum(TL.Spectrum)[, 1:4]

## spectrum with only 1 frame
TL1 <- set_RLum("RLum.Data.Spectrum", data = TL.Spectrum[, 1, drop = FALSE])

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fit_EmissionSpectra("fail"),
               "'object' should be of class 'RLum.Data.Spectrum', 'matrix' or")
  expect_error(fit_EmissionSpectra(list(TL.Spectrum, "fail")),
               "[fit_EmissionSpectra()] List elements of different class detected",
               fixed = TRUE)
  expect_error(fit_EmissionSpectra(matrix()),
               "'object' should have at least two columns")
  expect_error(fit_EmissionSpectra(TL.Spectrum, n_components = 1.4),
               "'n_components' should be a single positive integer value")
  expect_error(fit_EmissionSpectra(TL.Spectrum, input_scale = "error"),
               "'input_scale' should be one of 'wavelength', 'energy' or NULL")
  expect_error(fit_EmissionSpectra(TL.Spectrum, method_control = "error"),
               "'method_control' should be of class 'list'")
  expect_error(fit_EmissionSpectra(TL.Spectrum, verbose = "error"),
               "'verbose' should be a single logical value")
  expect_error(fit_EmissionSpectra(TL.Spectrum, plot = "error"),
               "'plot' should be a single logical value")

  ## wrong frame range -------
  expect_error(fit_EmissionSpectra(TL.Spectrum, frame = "error"),
               "'frame' should be of class 'integer', 'numeric' or NULL")
  expect_error(fit_EmissionSpectra(TL.Spectrum, frame = 1000),
               "Invalid 'frame', allowed values range from 1 to 24")

  ## wrong graining argument -------
  SW({
  expect_error(fit_EmissionSpectra(TL.Spectrum, frame = 5,
      method_control = list(graining = 10000)),
      "method_control$graining cannot exceed the available channels (1024)",
      fixed = TRUE)
  })

  ## for matrix input -------
  expect_error(fit_EmissionSpectra(object = mat, frame = "error"),
               "'frame' should be of class 'integer', 'numeric' or NULL")
  expect_error(fit_EmissionSpectra(object = mat, frame = 1000),
               "Invalid 'frame', allowed values range from 1 to 3")

  ## empty object
  expect_error(fit_EmissionSpectra(set_RLum("RLum.Data.Spectrum")),
               "'object' contains no data")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## no frame specified
  results <- fit_EmissionSpectra(
      object = TL1,
      plot = FALSE,
      verbose = FALSE,
      method_control = list(max.runs = 5))
  expect_equal(results@data$df_plot[[1]], NA)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 5.0e-1

  SW({
  set.seed(1)
  expect_snapshot_RLum(fit_EmissionSpectra(
    object = TL.Spectrum,
    frame = 5,
    n_components = 3,
    plot = FALSE,
    start_parameters = c(2.17),
    method_control = list(max.runs = 20,
                          export.plot.data = TRUE)),
    tolerance = snapshot.tolerance)

  ## input_scale
  set.seed(1)
  expect_snapshot_RLum(
      fit_EmissionSpectra(object = TL.Spectrum, frame = 5,
                          input_scale = "wavelength", plot = FALSE,
                          method_control = list(max.runs = 5)),
      tolerance = snapshot.tolerance)
  })

  ## for matrix input -------
  set.seed(17)
  expect_snapshot_RLum(
      fit_EmissionSpectra(object = mat, plot = FALSE, verbose = FALSE,
                          method_control = list(max.runs = 5)),
      tolerance = snapshot.tolerance)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  set.seed(1)
  vdiffr::expect_doppelganger("3 comps",
                              fit_EmissionSpectra(TL.Spectrum,
                                                  frame = 5,
                                                  n_components = 3,
                                                  main = "TL spectrum",
                                                  mtext = "Subtitle",
                                                  verbose = FALSE,
                                                  method_control = list(max.runs = 5)))
  set.seed(1)
  vdiffr::expect_doppelganger("wavelength",
                              fit_EmissionSpectra(TL.Spectrum,
                                                  frame = 5,
                                                  n_components = 2,
                                                  input_scale = "wavelength",
                                                  verbose = FALSE,
                                                  method_control = list(max.runs = 5)))
   vdiffr::expect_doppelganger("control plot",
                              fit_EmissionSpectra(TL.Spectrum,
                                                  frame = 4,
                                                  n_components = 2,
                                                  verbose = FALSE,
                                                  method_control = list(max.runs = 5)))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 761
  expect_silent(fit_EmissionSpectra(TL.Spectrum, frame = 1, plot = TRUE,
                                    n_components = 2, verbose = FALSE,
                                    method_control = list(max.runs = 5)))

  ## issue 932
  expect_error(fit_EmissionSpectra(matrix(rnorm(5), 1, 5), verbose = FALSE),
               "method_control$graining cannot exceed the available channels",
               fixed = TRUE)
})

## load data
data(ExampleData.RLum.Data.Image, envir = environment())

## generate random data
set.seed(1)
m <- matrix(runif(100,0,255), ncol = 10, nrow = 10)
a <- array(runif(300, 0,255), c(10,10,3))
RLum <- set_RLum("RLum.Data.Image", data = a)
RLum_list <- list(RLum, RLum)
roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(extract_ROI(object = "error", roi),
               "[extract_ROI()] 'object' should be of class 'RLum.Data.Image'",
               fixed = TRUE)
  expect_error(extract_ROI(object = m, "error"),
               "[extract_ROI()] 'roi' should be of class 'matrix'",
               fixed = TRUE)
  expect_error(extract_ROI(object = m, matrix()),
               "[extract_ROI()] 'roi' does not have the expected format",
               fixed = TRUE)
  expect_error(extract_ROI(object = m, matrix(ncol = 3, nrow = 0)),
               "[extract_ROI()] 'roi' does not have the expected format",
               fixed = TRUE)
  expect_error(extract_ROI(object = RLum, roi, roi_summary = "error"),
               "[extract_ROI()] 'roi_summary' should be one of 'mean', 'median'",
               fixed = TRUE)
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  ## matrix
  expect_snapshot_RLum(extract_ROI(object = m, roi),
                       tolerance = snapshot.tolerance)

  ## list
  expect_snapshot_RLum(extract_ROI(object = RLum_list, roi),
                       tolerance = snapshot.tolerance)

  ## test ROI summary options
  expect_snapshot_plain(extract_ROI(m, roi,
                                    roi_summary = "mean")@data$roi_summary)
  expect_snapshot_plain(extract_ROI(RLum, roi,
                                    roi_summary = "median")@data$roi_summary)
  expect_snapshot_plain(extract_ROI(RLum, roi,
                                    roi_summary = "sd")@data$roi_summary)
  expect_snapshot_plain(extract_ROI(RLum, roi,
                                    roi_summary = "sum")@data$roi_summary)
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              extract_ROI(RLum,
                                          roi,
                                          plot = TRUE))
  vdiffr::expect_doppelganger("list",
                              extract_ROI(RLum_list,
                                          roi + 2,
                                          plot = TRUE))
  vdiffr::expect_doppelganger("image",
                              extract_ROI(ExampleData.RLum.Data.Image,
                                          matrix(c(200, 400, 200, 40, 60, 80,
                                                   10, 10, 10), ncol = 3),
                                          plot = TRUE))
  })
})

## load data
data(ExampleData.TR_OSL, envir = environment())
temp_mat <- get_RLum(ExampleData.TR_OSL)[1:200, ]
temp_list <- list(ExampleData.TR_OSL, ExampleData.TR_OSL)
temp_analysis <- set_RLum("RLum.Analysis", records = temp_list)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_warning(expect_null(fit_OSLLifeTimes("error")),
                 "'object' should be of class 'RLum.Data.Curve', 'data.frame'")
  expect_error(fit_OSLLifeTimes(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(fit_OSLLifeTimes(set_RLum("RLum.Analysis")),
               "'object' cannot be an empty RLum.Analysis")
  expect_error(fit_OSLLifeTimes(matrix(NA, 0, 3)),
               "'object' cannot be an empty matrix")
  expect_error(fit_OSLLifeTimes(matrix()),
               "'object' should have at least two columns")
  expect_error(fit_OSLLifeTimes(ExampleData.TR_OSL, n.components = -1),
               "'n.components' should be a single positive integer value")
  expect_error(fit_OSLLifeTimes(ExampleData.TR_OSL, signal_range = FALSE),
               "'signal_range' should be of class 'integer' or 'numeric'")
  expect_warning(expect_error(fit_OSLLifeTimes(ExampleData.TR_OSL, signal_range = NaN),
                              "'signal_range' cannot be an empty numeric"),
                 "'signal_range' contains missing values, removed")
  expect_error(fit_OSLLifeTimes(ExampleData.TR_OSL, signal_range = numeric()),
               "'signal_range' should be of class 'integer' or 'numeric' and")
  expect_error(fit_OSLLifeTimes(temp_mat, signal_range = 150:200),
               "'signal_range' should be of class 'integer' or 'numeric' and have length 1 or 2")
  expect_error(fit_OSLLifeTimes(ExampleData.TR_OSL, plot = NA),
               "'plot' should be a single logical value")
  expect_error(fit_OSLLifeTimes(ExampleData.TR_OSL, verbose = NA),
               "'verbose' should be a single logical value")

  empty <- set_RLum(class = "RLum.Data.Curve")
  expect_error(fit_OSLLifeTimes(empty),
               "recordType 'NA' not supported for input object")
  expect_warning(expect_output(fit_OSLLifeTimes(list(empty)),
                               "recordType 'NA' not supported for input object"),
                 "Nothing was merged as the object list was found to be empty")

  expect_message(expect_null(fit_OSLLifeTimes(temp_mat[1:3, ])),
                 "For 1 components the dataset must have at least 5 signal points")
  expect_message(fit_OSLLifeTimes(temp_mat, n.components = 1,
                                  signal_range = c(1, 3), verbose = FALSE),
                 "For 1 components the dataset must have at least 5 signal points")
  expect_message(fit_OSLLifeTimes(temp_mat, n.components = 2,
                                  signal_range = c(1, 6), verbose = FALSE),
                 "For 2 components the dataset must have at least 7 signal points")

  expect_warning(fit_OSLLifeTimes(temp_mat, n.components = 1,
                                  signal_range = -1, verbose = FALSE),
                 "'signal_range' accepts only positive values")
  expect_warning(fit_OSLLifeTimes(temp_mat, n.components = 1,
                                  signal_range = c(1, 300), verbose = FALSE),
                 "The last element of 'signal_range' exceeds the number of channels")
  expect_warning(fit_OSLLifeTimes(temp_mat, n.components = 1,
                                  signal_range = 300, verbose = FALSE),
                 "The first element of 'signal_range' exceeds the last element"
                )

  temp <- temp_mat
  temp[100, 2] <- NA
  expect_warning(fit_OSLLifeTimes(temp, n.components = 1, verbose = FALSE),
                 "NA values detected and removed from dataset")

  temp[100:110, 2] <- 0
  expect_warning(fit_OSLLifeTimes(temp, n.components = 1, verbose = FALSE),
                 "The dataset contains 0, a value of 0.1 has been added")

  expect_warning(fit_OSLLifeTimes(ExampleData.TR_OSL,
                                  method_control = list(seed = 1,
                                                        DEoptim.itermax = 15,
                                                        nlsLM.lower = FALSE),
                                  plot = FALSE,
                                  verbose = FALSE,
                                  n.components = 1),
                 "At least one parameter is negative")

  expect_error(fit_OSLLifeTimes(data.frame(ED = c(rep(0, 4), 10),
                                           ED_Error = rnorm(5) + 1),
                                verbose = FALSE),
               "Failed to optimize the function, check the input data")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  ## Test different inputs
  ##simple run
  set.seed(1)
  SW({
  expect_snapshot_RLum(fit_OSLLifeTimes(
    object = ExampleData.TR_OSL,
    plot = FALSE,
    method_control = list(DEoptim.itermax = 15),
    n.components = 1),
    tolerance = snapshot.tolerance)

  ##simple list
  expect_snapshot_RLum(fit_OSLLifeTimes(
    object = temp_list,
    log = "x",
    method_control = list(DEoptim.itermax = 25),
    n.components = 1),
    tolerance = snapshot.tolerance)
  })

  ## RLum.Analysis
  expect_s4_class(object = fit_OSLLifeTimes(
    object = temp_analysis,
    verbose = FALSE,
    plot = FALSE,
    method_control = list(DEoptim.itermax = 25),
    n.components = 1), class = "RLum.Results")
  expect_s4_class(fit_OSLLifeTimes(
    object = list(temp_analysis),
    verbose = FALSE,
    plot = FALSE,
    method_control = list(seed = 1, weights = FALSE,
                          DEoptim.itermax = 25,
                          nlsLM.upper = FALSE, nlsLM.lower = FALSE),
    n.components = 1), class = "RLum.Results")

  ## simple data.frame
  SW({
  expect_snapshot_RLum(fit_OSLLifeTimes(
    object = as.data.frame(temp_mat),
    method_control = list(seed = 1, weights = FALSE,
                          DEoptim.itermax = 50),
    signal.range = 3,
    verbose = TRUE,
    plot = FALSE,
    n.components = NULL),
    tolerance = snapshot.tolerance)
  })

  SW({
  expect_message(fit_OSLLifeTimes(temp_mat[1:10, ]),
                 "The fitting was not successful, consider trying again")
  })

  ## more coverage
  expect_snapshot_RLum(fit_OSLLifeTimes(
      object = temp_mat[125:130, ],
      verbose = FALSE,
      plot = FALSE),
      tolerance = 1.5e-5)
  expect_snapshot_RLum(fit_OSLLifeTimes(
      temp_mat,
      n.components = 1,
      signal_range = c(1, 150),
      verbose = FALSE,
      plot = TRUE),
      tolerance = 1.5e-5)

  expect_warning(fit_OSLLifeTimes(
    object = ExampleData.TR_OSL,
    plot = FALSE,
    verbose = FALSE,
    method_control = list(seed = 100, DEoptim.itermax = 15),
    n.components = 1),
    "Fitting failed: singular gradient matrix at initial parameter estimates")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  set.seed(1)

  SW({
  vdiffr::expect_doppelganger("1 comp",
                              fit_OSLLifeTimes(
                                  object = ExampleData.TR_OSL,
                                  method_control = list(DEoptim.itermax = 15),
                                  n.components = 1))

  expect_warning(expect_warning(
      vdiffr::expect_doppelganger("log xy",
                                  fit_OSLLifeTimes(
                                      object = temp_mat,
                                      verbose = FALSE,
                                      plot = TRUE,
                                      plot_simple = TRUE,
                                      log = list("xy"),
                                      lty = 1,
                                      col = 1,
                                      n.components = 1)),
      "log-scale requires x-values > 0, set min xlim to 0.01"),
      "log-scale requires y-values > 0, set min ylim to 1.69e+10",
      fixed = TRUE)
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 1514
  expect_s4_class(fit_OSLLifeTimes(temp_analysis, select = NULL,
                                   n.components = 1,
                                   verbose = FALSE, plot = FALSE),
                  "RLum.Results")
  expect_s4_class(fit_OSLLifeTimes(temp_analysis, select = iris[0, ],
                                   n.components = 1,
                                   verbose = FALSE, plot = FALSE),
                  "RLum.Results")
})

## path to the SPE file on github, as it's not included in the package
github.url <- file.path("https://github.com/R-Lum/Luminescence",
                        "raw/master/tests/testthat/_data")
local.file <- test_path("_data/SPEfile.SPE")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_SPE2R(data.frame()),
               "'file' should be of class 'character'")
  expect_error(read_SPE2R(character(0)),
               "'file' should have length 1")
  expect_error(read_SPE2R("error"),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_error(read_SPE2R(local.file, output.object = "error"),
               "'output.object' should be one of 'RLum.Data.Image'")
  expect_error(read_SPE2R(local.file, frame.range = NA),
               "'frame.range' should be of class 'integer', 'numeric' or NULL")
  expect_error(read_SPE2R(local.file, frame.range = c(2, NA)),
               "'frame.range' should contain positive values")
  expect_error(read_SPE2R(local.file, frame.range = c(0, 3)),
               "'frame.range' should contain positive values")
  expect_message(expect_message(expect_null(read_SPE2R("http://httpbingo.org/status/404")),
                                "Downloading"), "FAILED")

  wrong <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
  expect_error(read_SPE2R(wrong),
               "File extension 'binx' is not supported, only 'SPE' is valid")
  expect_message(expect_message(expect_null(read_SPE2R(dirname(wrong))),
                                "Directory detected, looking for"),
                 "Error: No files matching the given pattern found in directory")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## default values
  SW({
  expect_message(
      expect_s4_class(read_SPE2R(file.path(github.url, "SPEfile.SPE")),
                      "RLum.Data.Image"),
      "Downloading")
  })
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(read_SPE2R(local.file,
                                  output.object = "RLum.Data.Spectrum",
                                  frame.range = 20:40,
                                  verbose = FALSE),
                       tolerance = snapshot.tolerance)
  expect_snapshot_plain(read_SPE2R(local.file,
                                   output.object = "matrix",
                                   verbose = FALSE),
                        tolerance = snapshot.tolerance)
})

## artifical dataset according to Mooney et al. (2013)
lambda <- seq(400, 800, 50)
data <- cbind(lambda, 1)
rownames(data) <- lambda
colnames(data) <- 1:ncol(data)

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(convert_Wavelength2Energy("test"),
               "'object' should be of class 'RLum.Data.Spectrum', 'data.frame'")
  expect_error(convert_Wavelength2Energy(data.frame()),
               "'object' cannot be an empty data.frame")
  expect_error(convert_Wavelength2Energy(matrix()),
               "'object' should have at least two columns")
  expect_error(convert_Wavelength2Energy(list(iris)),
               "'object' should have only numeric fields")
  expect_error(convert_Wavelength2Energy(set_RLum("RLum.Data.Spectrum")),
               "'object' contains no data")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##set plot function
  p <- function(m) {
    plot(x = m[, 1], y = m[, 2])
    polygon(x = c(m[, 1], rev(m[, 1])), y = c(m[, 2], rep(0, nrow(m))))
    for (i in 1:nrow(m)) {
      lines(x = rep(m[i, 1], 2), y = c(0, m[i, 2]))
    }
  }

  snapshot.tolerance <- 1.5e-6

  ##test all three allowed input objects
  expect_snapshot_plain(convert_Wavelength2Energy(data),
                        tolerance = snapshot.tolerance)
  expect_s3_class(convert_Wavelength2Energy(as.data.frame(data)), class = "data.frame")
  object <- set_RLum(class = "RLum.Data.Spectrum", data = data[,1,drop = FALSE])
  expect_s4_class(convert_Wavelength2Energy(object), class = "RLum.Data.Spectrum")
  expect_type(convert_Wavelength2Energy(matrix(rnorm(5), 1, 5)),
              "double")

  ##test the list option
  expect_type(convert_Wavelength2Energy(list(data, as.data.frame(data), object)), "list")

  ##test order argument
  expect_snapshot_plain(convert_Wavelength2Energy(data, order = TRUE),
                        tolerance = snapshot.tolerance)
  expect_snapshot_RLum(convert_Wavelength2Energy(object, order = TRUE),
                        tolerance = snapshot.tolerance)

  ##test special treatment of RLum.Data.Spectrum objects
  object@info[["curveDescripter"]] <- "energy"
  expect_message(convert_Wavelength2Energy(object),
                 "'object' has already an energy scale, nothing done")
  object@info[["curveDescripter"]] <- "wavelength"
  res <- convert_Wavelength2Energy(object)
  expect_equal(res@info[["curveDescripter"]],
               "energy [eV]")

  ##Code below just a cross check if wanted
  ##matrix
  # m <- cbind(as.numeric(rownames(data)), data)
  # par(mfrow = c(1,2))
  # p(m)
  # p(convert_Wavelength2Energy(m))
  #
  # ##df
  # df <- as.data.frame(cbind(as.numeric(rownames(data)), data))
  # p(df)
  # p(convert_Wavelength2Energy(df))
  #
  # ##RLum.Data.Spectrum
  # object <- set_RLum(class = "RLum.Data.Spectrum", data = data[,1,drop = FALSE])
  # par(mfrow = c(1,2))
  # plot_RLum.Data.Spectrum(object, plot.type = "single", par.local = FALSE)
  # plot_RLum.Data.Spectrum(convert_Wavelength2Energy(object), plot.type = "single", par.local = FALSE)
})

## load data
data(ExampleData.RLum.Analysis, envir = environment())

test_that("snapshot tests", {
  testthat::skip_on_cran()

  expect_snapshot_plain(structure_RLum(IRSAR.RF.Data))
  expect_snapshot_plain(structure_RLum(list(IRSAR.RF.Data,
                                            IRSAR.RF.Data,
                                            "a")))

  ## empty object
  res <- structure_RLum(set_RLum("RLum.Analysis"))
  expect_snapshot_plain(res)
  expect_equal(nrow(res), 0)
})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(calc_CosmicDoseRate(depth = "error"),
               "'depth' should be of class 'numeric'")
  expect_error(calc_CosmicDoseRate(depth = numeric(0), density = 1.7,
                                   latitude = 38.1, longitude = 1.4),
               "'depth' cannot be an empty numeric")
  expect_error(calc_CosmicDoseRate(depth = 10, density = "error"),
               "'density' should be of class 'numeric'")
  expect_error(calc_CosmicDoseRate(depth = 10, density = numeric(0),
                                   latitude = 38.1, longitude = 1.4),
               "'density' cannot be an empty numeric")
  expect_error(calc_CosmicDoseRate(depth = -2, density = 1.7, altitude = 364,
                                   latitude = 38.1, longitude = 1.4),
               "No negative or missing values allowed for 'depth' and 'density'")
  expect_error(calc_CosmicDoseRate(depth = c(1.78, NA), density = c(0.7, 1.0),
                                   latitude = 10, longitude = 20, altitude = 1),
               "No negative or missing values allowed for 'depth' and 'density'")
  expect_error(calc_CosmicDoseRate(depth = c(1.78, 0.12), density = c(0.7, NA),
                                   latitude = 10, longitude = 20, altitude = 1),
               "No negative or missing values allowed for 'depth' and 'density'")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7, altitude = 364,
                                   latitude = 38.1, longitude = 1.4,
                                   corr.fieldChanges = TRUE),
               "requires an age estimate")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7, altitude = 364,
                                   latitude = 38.1, longitude = 1.4,
                                   corr.fieldChanges = TRUE, est.age = 0),
               "'est.age' should be a single positive value")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7),
               "'latitude' should be of class 'numeric' and have length 1")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                                   latitude = 38.06451),
               "'longitude' should be of class 'numeric' and have length 1")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                                   latitude = 38.1, longitude = numeric(0)),
               "'longitude' should be of class 'numeric' and have length 1")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                                   latitude = 38.06451, longitude = 1.49646),
               "'altitude' should be of class 'numeric' and have length 1")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                                   latitude = 38.06451, longitude = 1.49646,
                                   altitude = numeric(0)),
               "'altitude' should be of class 'numeric' and have length 1")
  expect_error(calc_CosmicDoseRate(depth = 2.78, density = c(1.7, 2.9),
                                   corr.fieldChanges = TRUE, est.age = 20,
                                   latitude = 38.06451, longitude = 1.49646,
                                   altitude = 364),
               "The number of values for 'density' should either be 1 or")
  expect_error(calc_CosmicDoseRate(depth = rep(2.78, 3), density = c(1.7, 2.9),
                                   latitude = 38.06451, longitude = 1.49646,
                                   altitude = 364),
               "The number of values for 'density' should either be 1 or")
  SW({
  expect_message(calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                                    corr.fieldChanges = TRUE, est.age = 100,
                                    latitude = 38.0645, longitude = 1.4964,
                                    altitude = 364),
                 "No geomagnetic field change correction possible for samples")
  })
  expect_output(calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                                    corr.fieldChanges = TRUE, est.age = 20,
                                    latitude = 38.0645, longitude = 1.4964,
                                    altitude = 364, half.depth = TRUE),
                "No geomagnetic field change correction necessary for geomagnetic latitude >35 degrees")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  SW({
  expect_snapshot_RLum(
      calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                          latitude = 38.06451, longitude = 1.49646,
                          altitude = 364, error = 10),
      tolerance = snapshot.tolerance)

  ## length(depth) > length(density), half.depth
  expect_snapshot_RLum(
      calc_CosmicDoseRate(depth = c(2.78, 3.12), density = 1.7,
                          corr.fieldChanges = TRUE, est.age = 20,
                          latitude = 28.06451, longitude = 1.49646,
                          altitude = 364, half.depth = TRUE),
      tolerance = snapshot.tolerance)

  expect_snapshot_RLum(
      calc_CosmicDoseRate(depth = c(1.78, 0.12), density = c(0.7, 0.2),
                          latitude = 120, longitude = 30, altitude = 1200),
      tolerance = snapshot.tolerance)

  expect_snapshot_RLum(
      calc_CosmicDoseRate(depth = c(1.78, 0.12), density = c(0.2, 0.2),
                          latitude = 30, longitude = 120, altitude = 120),
      tolerance = snapshot.tolerance)
  })
})

test_that("check values from output example 2b", {
  testthat::skip_on_cran()

  SW({
  temp <- calc_CosmicDoseRate(depth = c(5.0, 2.78), density = c(2.65, 1.7),
                              latitude = 12.04332, longitude = 4.43243,
                              altitude = 364, corr.fieldChanges = TRUE,
                              est.age = 67, error = 15)
  })
  results <- get_RLum(temp)

  expect_equal(results$depth.1, 5)
  expect_equal(results$depth.2, 2.78)
  expect_equal(results$density.1, 2.65)
  expect_equal(results$density.2, 1.7)
  expect_equal(results$latitude, 12.04332)
  expect_equal(results$longitude, 4.43243)
  expect_equal(results$altitude, 364)
  expect_equal(round(results$total_absorber.gcm2, digits = 0), 1798)
  expect_equal(round(results$d0, digits = 4), 0.0705)
  expect_equal(round(results$geom_lat, digits =  1), 15.1)
  expect_equal(round(results$dc, digits = 3), 0.072)
})

##create example data
data <- data.frame(
  T = c(25, 40, 50, 60, 70, 80, 90, 100, 110),
  V = c(0.06, 0.058, 0.052, 0.051, 0.041, 0.034, 0.035, 0.033, 0.032),
  V_X = c(0.012, 0.009, 0.008, 0.008, 0.007, 0.006, 0.005, 0.005, 0.004))
data_NA <- data
data_NA[1,] <- NA

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(fit_ThermalQuenching(data = "test"),
               "'data' should be of class 'data.frame' or a 'list' of such objects")
  expect_error(fit_ThermalQuenching(
    data = data[,1:2],
    n.MC = NULL),
    "'data' is empty or has fewer than three columns")
  expect_error(fit_ThermalQuenching(data = data[1:3, ]),
               "Insufficient number of data points to fit the model")
  expect_warning(expect_error(fit_ThermalQuenching(data = data_NA[1:4, ]),
                              "Insufficient number of data points to fit the model"),
                 "NA values in 'data' automatically removed")
  expect_warning(expect_error(
      fit_ThermalQuenching(data.frame(1:10, NA, 1:10)),
      "After NA removal, nothing is left from the data set"),
      "NA values in 'data' automatically removed")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ##simple run with warning
  SW({
  expect_warning(fit_ThermalQuenching(
    data = cbind(data,data),
    n.MC = NULL),
    "data' has more than 3 columns, taking only the first three")

  ##simple run with warning NA
  expect_warning(fit_ThermalQuenching(
    data = data_NA,
    n.MC = NULL),
    "NA values in 'data' automatically removed")

  ##simple run with fitting error
  expect_message(expect_null(fit_ThermalQuenching(
    data = data.frame(T = 1:10, V = 1:10, V_X = 1:10),
    n.MC = NULL)),
    "Error: Fitting failed, NULL returned")

  ##simple list run
  expect_s4_class(fit_ThermalQuenching(
    data = list(data, data),
    n.MC = NULL), class = "RLum.Results")
  })
})

test_that("add snapshot tests", {
  testthat::skip_on_cran()

  set.seed(1)
  snapshot.tolerance <- 1.5e-6

  SW({
  expect_snapshot_RLum(fit_ThermalQuenching(data = data, n.MC = NULL),
                       tolerance = snapshot.tolerance)

  ## switch off weights
  expect_snapshot_RLum(fit_ThermalQuenching(data = data, n.MC = NULL,
                                            method_control = list(weights = NULL)),
                       tolerance = snapshot.tolerance)
  })

  ## with Monte Carlo
  expect_snapshot_RLum(fit_ThermalQuenching(data = data, n.MC = 10,
                                            verbose = FALSE, plot = TRUE),
                       tolerance = snapshot.tolerance)
})

## load data
data(ExampleData.CW_OSL_Curve, envir = environment())
curve <- set_RLum(class = "RLum.Data.Curve",
                  recordType = "OSL",
                  data = as.matrix(ExampleData.CW_OSL_Curve))

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(bin_RLum.Data(set_RLum("RLum.Data.Spectrum")),
               "'object' contains no data")
})

test_that("snapshot tests", {
  testthat::skip_on_cran()

  snapshot.tolerance <- 1.5e-6

  expect_snapshot_RLum(bin_RLum.Data(curve),
                       tolerance = snapshot.tolerance)
  expect_snapshot_RLum(bin_RLum.Data(curve, bin = 5),
                       tolerance = snapshot.tolerance)
})

## load data
data(ExampleData.BINfileData, envir = environment())
obj <- subset(CWOSL.SAR.Data, ID <= 50)

test_that("Check the example and the numerical values", {
  testthat::skip_on_cran()

  expect_error(set_Risoe.BINfileData(METADATA = NA),
               "'METADATA' should be of class 'data.frame'")

  ##construct empty object
  temp <-
    set_Risoe.BINfileData(METADATA = data.frame(), DATA = list(), .RESERVED = list())

  ##check object
  expect_s4_class(temp, class = "Risoe.BINfileData")
  expect_output(show(temp),
                "This object is empty!")

  ## also empty object
  temp <- set_Risoe.BINfileData(METADATA = NULL)
  expect_s4_class(temp, class = "Risoe.BINfileData")
  expect_output(show(temp),
                "This object is empty!")

  ##show method
  expect_output(show(CWOSL.SAR.Data))

  ##as.data.frame
  expect_s3_class(as.data.frame(CWOSL.SAR.Data), "data.frame")

  ## metadata provided
  meta <- data.frame(ID = 1, VERSION = 7, RECTYPE = 128)
  temp <- set_Risoe.BINfileData(METADATA = meta, DATA = list(),
                                .RESERVED = list())
  expect_equal(temp@METADATA, meta)
  expect_output(show(temp))

  meta$RECTYPE <- NULL
  temp <- set_Risoe.BINfileData(METADATA = meta, DATA = list(),
                                .RESERVED = list())
  expect_equal(temp@METADATA, meta)
  expect_output(show(temp))
})

test_that("sort_RLum", {
  testthat::skip_on_cran()

  ## input validation
  expect_error(sort_RLum(obj),
               "'info_element' should be of class 'character'")
  expect_error(sort_RLum(obj, info_element = c("ID", "error")),
               "Invalid 'info_element' name, valid names are:")
  expect_error(sort_RLum(obj, info_element = "ID", decreasing = "error"),
               "'decreasing' should be a single logical value")

  ## sort only using the first field until we get #605 done
  expect_message(sort_RLum(obj, info_element = c("ID", "POSITION")),
                 "Only the first field will be used in sorting")

  ## check functionality
  expect_snapshot_Risoe(sort_RLum(obj, info_element = "POSITION"))
})

test_that("subset.Risoe.BINfileData", {
  testthat::skip_on_cran()

  small <- subset(CWOSL.SAR.Data, POSITION == 1)
  expect_length(small, 30)
  expect_equal(rownames(small@METADATA),
               c("1", "25", "49", "73", "97", "121", "145", "169", "193", "217",
                 "241", "265", "289", "313", "337", "361", "385", "409", "433",
                 "457", "481", "505", "529", "553", "577", "601", "625", "649",
                 "673", "697"))
})

data(ExampleData.CW_OSL_Curve, envir = environment())
temp <-
  set_RLum(
      class = "RLum.Data.Curve",
      recordType = "OSL",
      data = as.matrix(ExampleData.CW_OSL_Curve)
  )

##create RLum.Analysis object
temp_analysis <- set_RLum("RLum.Analysis", records = list(temp, temp))

test_that("check class and length of output", {
  testthat::skip_on_cran()

  ##standard tests
  expect_s4_class(temp, class = "RLum.Data.Curve")
  expect_snapshot_RLum(smooth_RLum(temp))

  ##test on a list
    ##RLum list
    expect_type(smooth_RLum(list(temp, temp)), "list")

    ##normal list
    expect_type(smooth_RLum(list(a = 1, b = 2)), "list")

  ##test on an RLum.Analysis-object
  expect_s4_class(smooth_RLum(temp_analysis), "RLum.Analysis")

})

test_that("snapshot tests", {
 testthat::skip_on_cran()

 small <-set_RLum(class = "RLum.Data.Curve", recordType = "OSL",
                  data = as.matrix(ExampleData.CW_OSL_Curve[1:150, ]))
 expect_snapshot_RLum(smooth_RLum(small, k = 5))
 expect_snapshot_RLum(smooth_RLum(small, k = 10))
 expect_snapshot_RLum(smooth_RLum(small, k = 11, method = "median"))
})

#' @title Plot a histogram with separate error plot
#'
#' @description
#' Function plots a predefined histogram with an accompanying error plot as
#' suggested by Rex Galbraith at the UK LED in Oxford 2010.
#'
#' If the normal curve is added, the y-axis in the histogram will show the
#' probability density.
#'
#' A statistic summary, i.e. a collection of statistic measures of
#' centrality and dispersion (and further measures) can be added by specifying
#' one or more of the following keywords:
#' - `"n"` (number of samples),
#' - `"mean"` (mean De value),
#' - `"mean.weighted"` (error-weighted mean),
#' - `"median"` (median of the De values),
#' - `"median.weighted"` (error-weighted median),
#' - `"sdrel"` (relative standard deviation in percent),
#' - `"sdrel.weighted"` (error-weighted relative standard deviation in percent),
#' - `"sdabs"` (absolute standard deviation),
#' - `"sdabs.weighted"` (error-weighted absolute standard deviation),
#' - `"serel"` (relative standard error),
#' - `"serel.weighted"` (error-weighted relative standard error),
#' - `"seabs"` (absolute standard error),
#' - `"seabs.weighted"` (error-weighted absolute standard error),
#' - `"kurtosis"` (kurtosis) and
#' - `"skewness"` (skewness).
#'
#' @param data [data.frame] or [Luminescence::RLum.Results-class] object (**required**):
#' for `data.frame`: two columns: De (`data[,1]`) and De error (`data[,2]`).
#' If the error column is missing or only contains `NA` values, then the error
#' at each measurement is assumed to be 10^-9.
#'
#' @param na.rm [logical] (*with default*):
#' excludes `NA` values from the data set prior to any further operations.
#'
#' @param mtext [character] (*optional*):
#' further sample information ([mtext]).
#'
#' @param cex.global [numeric] (*with default*):
#' global scaling factor.
#'
#' @param se [logical] (*with default*):
#' plots standard error points over the histogram, default is `TRUE`.
#'
#' @param rug [logical] (*with default*):
#' adds rugs to the histogram, default is `TRUE`.
#'
#' @param normal_curve [logical] (*with default*):
#' adds a normal curve to the histogram. Mean and standard deviation are
#' calculated from the input data. If `TRUE`, the y-axis in the histogram
#' will show the probability density.
#'
#' @param summary [character] (*with default*):
#' add statistic measures of centrality and dispersion to the plot.
#' Can be one or more of several keywords. See details for available keywords.
#'
#' @param summary.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`)
#' for the statistical summary. Alternatively, the keyword `"sub"` may be
#' specified to place the summary below the plot header. However, this latter
#' option in only possible if `mtext` is not used. In case of coordinate
#' specification, y-coordinate refers to the right y-axis.
#'
#' @param colour [numeric] or [character] (*with default*):
#' optional vector of length 4 which specifies the colours of the following
#' plot items in exactly this order: histogram bars, rug lines and summary
#' text, normal distribution curve, standard error points
#' (e.g., `c("grey", "black", "red", "grey")`).
#'
#' @param interactive [logical] (*with default*):
#' create an interactive histogram plot (requires the 'plotly' package)
#'
#' @param ... further arguments and graphical parameters passed to [plot] or
#' [hist]. If y-axis labels are provided, these must be specified as a vector
#' of length 2 since the plot features two axes
#' (e.g. `ylab = c("axis label 1", "axis label 2")`). Y-axes limits
#' (`ylim`) must be provided as vector of length four, with the first two
#' elements specifying the left axes limits and the latter two elements giving
#' the right axis limits.
#'
#' @note The input data is not restricted to a special type.
#'
#' @section Function version: 0.4.5
#'
#' @author
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [hist], [plot]
#'
#' @examples
#'
#' ## load data
#' data(ExampleData.DeValues, envir = environment())
#' ExampleData.DeValues <- convert_Second2Gray(ExampleData.DeValues$BT998,
#'                                             dose.rate = c(0.0438,0.0019))
#'
#' ## plot histogram the easiest way
#' plot_Histogram(ExampleData.DeValues)
#'
#' ## plot histogram with some more modifications
#' plot_Histogram(ExampleData.DeValues,
#'                rug = TRUE,
#'                normal_curve = TRUE,
#'                cex.global = 0.9,
#'                pch = 2,
#'                colour = c("grey", "black", "blue", "green"),
#'                summary = c("n", "mean", "sdrel"),
#'                summary.pos = "topleft",
#'                main = "Histogram of De-values",
#'                mtext = "Example data set",
#'                ylab = c(expression(paste(D[e], " distribution")),
#'                         "Standard error"),
#'                xlim = c(100, 250),
#'                ylim = c(0, 0.1, 5, 20))
#'
#' @md
#' @export
plot_Histogram <- function(
  data,
  na.rm = TRUE,
  mtext = "",
  cex.global = 1,
  se = TRUE,
  rug = TRUE,
  normal_curve = FALSE,
  summary = "",
  summary.pos = "sub",
  colour = c("white", "black", "red", "black"),
  interactive = FALSE,
  ...
) {
  .set_function_name("plot_Histogram")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("data.frame", "RLum.Results"))
  .validate_not_empty(data)
  if (inherits(data, "RLum.Results")) {
    data <- get_RLum(data)[,1:2]
  }

  ## we don't check the second column, as that can be NA
  .validate_class(data[, 1], c("numeric", "integer"),
                  name = "All columns of 'data'")

  .validate_class(mtext, "character", length = 1)
  .validate_class(cex.global, "numeric")
  .validate_class(summary, "character")
  if (is.numeric(summary.pos)) {
    .validate_length(summary.pos, 2)
  }
  else {
    summary.pos <- .validate_args(summary.pos,
                                  c("sub", "left", "center", "right",
                                    "topleft", "top", "topright",
                                    "bottomleft", "bottom", "bottomright"))
  }
  .validate_logical_scalar(na.rm)
  .validate_logical_scalar(se)
  .validate_logical_scalar(rug)
  .validate_logical_scalar(normal_curve)
  .validate_class(colour, "character", length = 4)

  ## handle error-free data sets
  if (length(data) < 2 || all(is.na(data[, 2]))) {
    data[, 2] <- 1e-9
    se <- FALSE
  }

  ## Set general parameters ---------------------------------------------------

  ## read out additional arguments list
  extraArgs <- list(...)

  ## optionally, count and exclude NA values and print result
  if (na.rm) {
    n.NA <- sum(is.na(data[,1]))
    if (n.NA > 0) {
      .throw_message(sprintf("%d NA value%s excluded\n",
                             n.NA, ifelse(n.NA > 1, "s", "")),
                     error = FALSE)
      data <- data[!is.na(data[, 1]), ]
    }
  }

  if("breaks" %in% names(extraArgs)) {
    breaks.plot <- extraArgs$breaks
    breaks_calc <- hist(x = data[,1],
                        breaks = breaks.plot,
                        plot = FALSE)$breaks
  } else {
    breaks.plot <- hist(x = data[,1],
                        plot = FALSE)$breaks
    breaks_calc <- breaks.plot
  }

  if("ylim" %in% names(extraArgs)) {
    ylim.plot <- extraArgs$ylim
    .validate_class(ylim.plot, "numeric", length = 4, name = "'ylim'")
  } else {
    H.lim <- hist(data[,1],
                  breaks = breaks.plot,
                  plot = FALSE)
    left.ylim <- c(0, if (normal_curve) max(H.lim$density) else max(H.lim$counts))
    range.error <- try(expr = range(data[,2], na.rm = TRUE),
                       silent = TRUE)
    range.error[1] <- ifelse(is.infinite(range.error[1]), 0, range.error[1])
    range.error[2] <- ifelse(is.infinite(range.error[2]), 0, range.error[2])
    ylim.plot <- c(left.ylim, range.error)
  }

  settings <- modifyList(list(
      main = "Histogram",
      xlab = expression(paste(D[e], " [Gy]")),
      ylab = c("Frequency", "Standard error"),
      xlim = range(breaks_calc),
      pch = 1,
      fun = FALSE
  ), extraArgs)

  ## Set plot area format
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)
  par(mar = c(4.5, 4.5, 4.5, if (se) 4.5 else 1),
      cex = cex.global)

  ## Plot histogram -----------------------------------------------------------
  HIST <- hist(data[,1],
               main = settings$main,
               xlab = settings$xlab,
               ylab = settings$ylab[1],
               xlim = settings$xlim,
               ylim = settings$ylim[1:2],
               breaks = breaks.plot,
               freq = !normal_curve,
               col = colour[1]
  )

  ## Optionally, add rug ------------------------------------------------------
  if (rug) {
    graphics::rug(data[, 1], col = colour[2], quiet = TRUE)
  }

  ## Optionally, add a normal curve based on the data -------------------------
  if (normal_curve) {
    ## silence notes raised by R CMD check
    x <- NULL

    ## add normal distribution curve
    curve(dnorm(x,
                mean = mean(na.exclude(data[,1])),
                sd = sd(na.exclude(data[,1]))),
          col = colour[3],
          add = TRUE,
          lwd = 1.2 * cex.global)
  }

  ## calculate and paste statistical summary
  data.stats <- list(data = data)

  ## calculate and paste statistical summary
  De.stats <- matrix(nrow = length(data), ncol = 18)
  colnames(De.stats) <- c("n",
                          "mean",
                          "mean.weighted",
                          "median",
                          "median.weighted",
                          "kde.max",
                          "sd.abs",
                          "sd.rel",
                          "se.abs",
                          "se.rel",
                          "q25",
                          "q75",
                          "skewness",
                          "kurtosis",
                          "sd.abs.weighted",
                          "sd.rel.weighted",
                          "se.abs.weighted",
                          "se.rel.weighted")

  for(i in 1:length(data)) {
    statistics <- calc_Statistics(data)
    De.stats[i,1] <- statistics$weighted$n
    De.stats[i,2] <- statistics$unweighted$mean
    De.stats[i,3] <- statistics$weighted$mean
    De.stats[i,4] <- statistics$unweighted$median
    De.stats[i,5] <- statistics$weighted$median
    De.stats[i,7] <- statistics$unweighted$sd.abs
    De.stats[i,8] <- statistics$unweighted$sd.rel
    De.stats[i,9] <- statistics$unweighted$se.abs
    De.stats[i,10] <- statistics$weighted$se.rel
    De.stats[i,11] <- quantile(data[,1], 0.25)
    De.stats[i,12] <- quantile(data[,1], 0.75)
    De.stats[i,13] <- statistics$unweighted$skewness
    De.stats[i,14] <- statistics$unweighted$kurtosis
    De.stats[i,15] <- statistics$weighted$sd.abs
    De.stats[i,16] <- statistics$weighted$sd.rel
    De.stats[i,17] <- statistics$weighted$se.abs
    De.stats[i,18] <- statistics$weighted$se.rel

    ##kdemax - here a little doubled as it appears below again
    De.density <- NA
    De.stats[i,6] <- NA
    if(nrow(data) >= 2){
      De.density <-density(x = data[,1],
                           kernel = "gaussian",
                           from = settings$xlim[1],
                           to = settings$xlim[2])

      De.stats[i,6] <- De.density$x[which.max(De.density$y)]
    }
  }

  ## helper to generate an element of the statistical summary
  .summary_line <- function(keyword, summary, val, label = keyword,
                            percent = FALSE, sep = FALSE, digits = 2) {
    ifelse(keyword %in% summary,
           paste0(label, " = ", round(val, digits),
                  if (percent) " %" else NULL, if (sep) " | " else "\n"),
           "")
  }

  ## initialize list
  label.text <- NULL
  is.sub <- summary.pos[1] == "sub"
  stops <- NULL
  for (i in 1:length(data)) {
    if (!is.sub)
      stops <- strrep("\n", (i - 1) * length(summary))

    summary.text <- character(0)
    for (j in 1:length(summary)) {
      summary.text <-
        c(summary.text,
          .summary_line("n", summary[j], De.stats[i, 1], sep = is.sub),
          .summary_line("mean", summary[j], De.stats[i, 2], sep = is.sub),
          .summary_line("mean.weighted", summary[j], De.stats[i, 3], sep = is.sub,
                        label = "weighted mean"),
          .summary_line("median", summary[j], De.stats[i, 4], sep = is.sub),
          .summary_line("median.weighted", summary[j], De.stats[i, 5], sep = is.sub,
                        label = "weighted median"),
          .summary_line("kdemax", summary[j], De.stats[i, 6], sep = is.sub),
          .summary_line("sdabs", summary[j], De.stats[i, 7], sep = is.sub,
                        label = "sd"),
          .summary_line("sdrel", summary[j], De.stats[i, 8], sep = is.sub,
                        label = "rel. sd", percent = TRUE),
          .summary_line("seabs", summary[j], De.stats[i, 9], sep = is.sub,
                        label = "se"),
          .summary_line("serel", summary[j], De.stats[i, 10], sep = is.sub,
                        label = "rel. se", percent = TRUE),
          .summary_line("skewness", summary[j], De.stats[i, 13], sep = is.sub),
          .summary_line("kurtosis", summary[j], De.stats[i, 14], sep = is.sub),
          .summary_line("sdabs.weighted", summary[j], De.stats[i, 15], sep = is.sub,
                        label = "abs. weighted sd"),
          .summary_line("sdrel.weighted", summary[j], De.stats[i, 16], sep = is.sub,
                        label = "rel. weighted sd"),
          .summary_line("seabs.weighted", summary[j], De.stats[i, 17], sep = is.sub,
                        label = "abs. weighted se"),
          .summary_line("serel.weighted", summary[j], De.stats[i, 18], sep = is.sub,
                        label = "rel. weighted se"))
    }
    label.text[[i]] <- paste0(
        if (is.sub) "" else stops,
        paste(summary.text, collapse = ""),
        stops)
  }

  ## remove outer vertical lines from string
  if (is.sub) {
    for (i in seq_along(label.text)) {
      label.text[[i]] <- substr(x = label.text[[i]],
                                start = 1,
                                stop = nchar(label.text[[i]]) - 3)
    }
  }

  ## convert keywords into summary placement coordinates
  coords <- .get_keyword_coordinates(summary.pos, settings$xlim, ylim.plot[1:2])
  summary.pos <- coords$pos
  summary.adj <- coords$adj

  ## add summary content
  for(i in 1:length(data.stats)) {
    if(summary.pos[1] != "sub") {
      text(x = summary.pos[1],
           y = summary.pos[2],
           adj = summary.adj,
           labels = label.text[[i]],
           col = colour[2],
           cex = 0.8)
    } else if (mtext == "") {
        mtext(side = 3,
              line = 1 - i,
              text = label.text[[i]],
              col = colour[2],
              cex = cex.global * 0.8)
    }
  }

  ## Optionally, add standard error plot --------------------------------------

  if (se) {
    par(new = TRUE)
    plot.data <- data[!is.na(data[,2]),]

    plot(x = plot.data[,1],
         y = plot.data[,2],
         xlim = settings$xlim,
         ylim = ylim.plot[3:4],
         pch = settings$pch,
         col = colour[4],
         main = "",
         xlab = "",
         ylab = "",
         axes = FALSE,
         frame.plot = FALSE
    )
    axis(side = 4,
         labels = TRUE,
         cex = cex.global
    )
    mtext(settings$ylab[2],
          side = 4,
          line = 3,
          cex = cex.global)
  }

  ## Optionally add user-defined mtext
  mtext(side = 3,
        line = 0.5,
        text = mtext,
        cex = 0.8 * cex.global)

  ## FUN by R Luminescence Team
  if (settings$fun && !interactive) sTeve() # nocov

  ## Optionally: Interactive Plot ----------------------------------------------
  if (interactive) {
    .require_suggested_package("plotly", "The interactive histogram")

    ## tidy data ----
    data <- as.data.frame(data)
    colnames(data) <- c("x", "y")
    x <- y <- NULL # suffice CRAN check for no visible binding
    if (any(grepl("paste", as.character(settings$xlab), fixed = TRUE)))
      settings$xlab <- "Equivalent dose [Gy]"

    ## create plots ----

    # histogram
    hist <- plotly::plot_ly(data = data, x = ~x,
                            type = "histogram",
                            showlegend = FALSE,
                            name = "Bin", opacity = 0.75,
                            marker = list(color = "#428BCA",
                                          line = list(width = 1.0,
                                                      color = "white")),
                            histnorm = ifelse(normal_curve, "probability density", ""),
                            yaxis = "y"
    )

    # normal curve ----
    if (normal_curve) {
      density.curve <- density(data$x)
      normal.curve <- data.frame(x = density.curve$x, y = density.curve$y)
      hist <- plotly::add_trace(hist, data = normal.curve, x = ~x, y = ~y,
                                inherit = FALSE,
                                type = "scatter", mode = "lines",
                                line = list(color = "red"),
                                name = "Normal curve",
                                yaxis = "y")
    }

    # scatter plot of individual errors
    if (se) {
      yaxis2 <- list(overlaying = "y", side = "right",
                     showgrid = FALSE, title = settings$ylab[2],
                     ticks = "", showline = FALSE)

      se.text <- paste0("Measured value:</br>",
                        data$x, " &plusmn; ", data$y,"</br>")

      hist <- plotly::add_trace(hist, data = data, x = ~x, y = ~y,
                                inherit = FALSE,
                                type = "scatter", mode = "markers",
                                name = "Error", hoverinfo = "text",
                                text = se.text,
                                marker = list(color = "black"),
                                yaxis = "y2")

      hist <- plotly::layout(hist, yaxis2 = yaxis2)
    }

    # set layout ----
    hist <- plotly::layout(hist, hovermode = "closest",
                           title = paste("<b>", settings$main, "</b>"),
                           margin = list(r = 90),
                           xaxis = list(title = settings$xlab,
                                        ticks = ""),
                           yaxis = list(title = settings$ylab[1],
                                        ticks = "",
                                        showline = FALSE,
                                        showgrid = FALSE)
    )

    ## show and return plot ----
    print(hist)
    return(hist)
  }
}

## load data
data(ExampleData.DeValues, envir = environment())
df <- ExampleData.DeValues$CA1

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_Histogram("error"),
               "'data' should be of class 'data.frame' or 'RLum.Results'")
  expect_error(plot_Histogram(iris[0, ]),
               "'data' cannot be an empty data.frame")
  expect_error(plot_Histogram(data.frame()),
               "'data' cannot be an empty data.frame")
  expect_error(plot_Histogram(data.frame(a = letters)),
               "All columns of 'data' should be of class 'numeric'")
  expect_error(plot_Histogram(set_RLum("RLum.Results")),
               "'data' cannot be an empty RLum.Results")
  expect_error(plot_Histogram(df, mtext = numeric()),
               "'mtext' should be of class 'character' and have length 1")
  expect_error(plot_Histogram(df, summary = 5),
               "'summary' should be of class 'character'")
  expect_error(plot_Histogram(df, summary.pos = 5),
               "'summary.pos' should have length 2")
  expect_error(plot_Histogram(df, summary.pos = list()),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_Histogram(df, summary.pos = "error"),
               "'summary.pos' should be one of 'sub', 'left', 'center', 'right'")
  expect_error(plot_Histogram(df, colour = "black"),
               "'colour' should be of class 'character' and have length 4")
  expect_error(plot_Histogram(df, ylim = c(0, 1)),
               "'ylim' should be of class 'numeric' and have length 4")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  expect_silent(plot_Histogram(df))
  expect_silent(plot_Histogram(df, normal_curve = TRUE))

  ## more coverage
  expect_silent(plot_Histogram(df, main = "Title", xlab = "x", ylab = "y",
                               xlim = c(0, 120), ylim = c(0, 0.1, 0, 0.1),
                               pch = 1, breaks = "Sturges",
                               normal_curve = TRUE,
                               summary.pos = c(20, 0.017),
                               summary = c("n", "mean", "mean.weighted",
                                           "median", "sdrel")))

  ## interactive
  expect_silent(plot_Histogram(df, interactive = TRUE,
                               normal_curve = TRUE, se = TRUE))

  ## missing values
  df.na <- df
  df.na[10, 1] <- NA
  expect_message(plot_Histogram(set_RLum("RLum.Results", data = list(df.na)),
                                summary.pos = "bottom"),
                 "1 NA value excluded")
  df.na[20, 1] <- NA
  expect_message(plot_Histogram(df.na),
                 "2 NA values excluded")
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("Histogram normal",
                              plot_Histogram(df, normal_curve = TRUE))
  vdiffr::expect_doppelganger("Histogram summary sub",
                              plot_Histogram(df, summary.pos = "sub",
                                             summary = c("n", "serel", "kurtosis")))
  vdiffr::expect_doppelganger("Histogram summary left",
                              plot_Histogram(df, summary.pos = "left",
                                             summary = c("mean", "skewness",
                                                         "median.weighted")))
  vdiffr::expect_doppelganger("normal_curve summary center",
                              plot_Histogram(df, summary.pos = "center",
                                             normal_curve = TRUE,
                                             summary = c("mean", "skewness",
                                                         "median.weighted")))
  })
})

test_that("regression tests", {
  testthat::skip_on_cran()

  ## issue 744
  expect_silent(plot_Histogram(df[, 1, drop = FALSE]))
  expect_silent(plot_Histogram(cbind(df[, 1, drop = FALSE], NA)))
})

#' @title Create Regions of Interest (ROI) Graphic
#'
#' @description
#' The function creates ROI graphic with data extracted from the data imported
#' via [Luminescence::read_RF2R]. This function might be of use to work with
#' reduced data from spatially resolved measurements.
#' The plot dimensions mimic the original image dimensions.
#'
#' @param object [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Results-class] or a [list] of such objects (**required**):
#' input data created either by [Luminescence::read_RF2R] or [Luminescence::extract_ROI].
#'
#' @param exclude_ROI [numeric] (*with default*):
#' option to remove particular ROIs from the analysis. Those ROIs are plotted
#' but not coloured and not taken into account in distance analysis. `NULL`
#' excludes nothing.
#'
#' @param dist_thre [numeric] (*optional*):
#' euclidean distance threshold in pixel distance. All ROI for which the
#' euclidean distance is smaller are marked. This helps to identify ROIs that
#' might be affected by signal cross-talk. Note: the distance is calculated
#' from the centre of an ROI, e.g., the threshold should include consider the
#' ROIs or grain radius.
#'
#' @param dim.CCD [numeric] (*optional*):
#' metric x and y for the recorded (chip) surface in µm. For instance
#' `c(8192, 8192)`, if set additional x and y-axes are shown.
#'
#' @param bg_image [Luminescence::RLum.Data.Image-class] (*optional*):
#' background image object (please note that dimensions are not checked).
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param ... further parameters to manipulate the plot. Besides the standard
#' arguments of [graphics::plot.default], the following arguments are
#' supported: `lwd.ROI`, `lty.ROI`, `col.ROI`, `col.pixel`, `text.labels`,
#' `text.offset`, `grid` (`TRUE/FALSE`), `legend` (`TRUE/FALSE`),
#' `legend.text`, `legend.pos`
#'
#' @return
#' An ROI plot and an [Luminescence::RLum.Results-class] object with a matrix
#' containing the extracted ROI data and a object produced by [stats::dist]
#' containing the euclidean distance between the ROIs.
#'
#'@section Function version: 0.2.0
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::read_RF2R], [Luminescence::extract_ROI]
#'
#'@keywords datagen plot
#'
#'@examples
#'
#'## simple example
#'file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
#'temp <- read_RF2R(file)
#'plot_ROI(temp)
#'
#'## in combination with extract_ROI()
#'m <- matrix(runif(100,0,255), ncol = 10, nrow = 10)
#'roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
#'t <- extract_ROI(object = m, roi = roi)
#'plot_ROI(t, bg_image = m)
#'
#'@export
plot_ROI <- function(
  object,
  exclude_ROI = 1,
  dist_thre = -Inf,
  dim.CCD = NULL,
  bg_image = NULL,
  plot = TRUE,
  ...
) {
  .set_function_name("plot_ROI")
  on.exit(.unset_function_name(), add = TRUE)

  ##helper function to extract content
  .spatial_data <- function(x) {
    .validate_class(x, c("RLum.Analysis", "RLum.Results"),
                    extra = "a 'list' of such objects",
                    name = "'object'")

    ##ignore all none RLum.Analysis
    if (!inherits(x, "RLum.Analysis") || !.check_originator(x, "read_RF2R"))
      .throw_error("Object originator '", x@originator, "' not supported")

    ##extract some of the elements
    info <- x@info
    info$ROI <- strsplit(split = " ", info$ROI, fixed = TRUE)[[1]][2]
    info <- info[c("ROI", "x", "y", "area", "width", "height",
                    "image_width", "image_height", "grain_d")]
    names(info) <- gsub("image_", "img_", names(info))
    unlist(lapply(info, as.numeric))
  }

  if (inherits(object, "RLum.Results") &&
      .check_originator(object, "extract_ROI")) {
    m <- object@data$roi_coord

  } else {
    ## make sure the object is a list
    if(!is.list(object)) object <- list(object)
    .validate_not_empty(object, "list")

    ##extract values and convert to numeric matrix
    m <- t(vapply(object, .spatial_data, numeric(length = 9)))

    ## correct coordinates (they come in odd from the file)
    m[,"x"] <- m[,"x"] + m[,"width"] / 2
    m[,"y"] <- m[,"y"] + m[,"height"] / 2
  }

  .validate_positive_scalar(exclude_ROI, int = TRUE, null.ok = TRUE)

  ##make sure the ROI selection works
  if (is.null(exclude_ROI))
    exclude_ROI <- nrow(m) + 1

  ## distance calculation
  euc_dist <- sel_euc_dist <- stats::dist(m[-exclude_ROI])

  ## add half_width and half_height
  m <- cbind(m, half_width = m[, "width"] / 2, half_height = m[, "height"] / 2)
  rownames(m) <- m[, "ROI"]

  ## distance threshold selector
  sel_euc_dist[sel_euc_dist < dist_thre[1]] <- NA
  sel_euc_dist <- suppressWarnings(as.numeric(rownames(na.exclude(as.matrix(sel_euc_dist)))))

  ## add information to matrix
  m <- cbind(m, dist_sel = FALSE)
  m[m[,"ROI"]%in%sel_euc_dist,"dist_sel"] <- TRUE

  ## --- Plotting ---
  if(plot) {
    plot_settings <- modifyList(x = list(
      xlim = c(0, max(m[, "img_width"])),
      ylim = c(max(m[, "img_height"]), 0),
      xlab = "width [px]",
      ylab = "height [px]",
      main = "Spatial ROI Distribution",
      frame.plot = FALSE,
      lwd.ROI = 0.75,
      lty.ROI = 2,
      col.ROI = "black",
      col.pixel = rgb(0,1,0,0.6),
      text.labels = m[,"ROI"],
      text.offset = 0.3,
      grid = FALSE,
      legend = TRUE,
      legend.text = c("ROI", "sel. pixel", "> dist_thre"),
      legend.pos = "topright"
    ), val = list(...))

    ## set plot area
    do.call(
      what = graphics::plot.default,
      args = c(x = NA, y = NA,
               plot_settings[names(plot_settings) %in%
                             methods::formalArgs(graphics::plot.default)])
    )

    ## add background image if available
    if(!is.null(bg_image)){
      a <- try({
        as(bg_image, "RLum.Data.Image")
      }, silent = TRUE)
      if(inherits(a, "try-error")) {
        .throw_warning("'bg_image' is not of class 'RLum.Data.Image' and ",
                       "cannot be converted into it, background image plot skipped")
      } else {
        a <- a@data
        graphics::image(
          x = 1:nrow(a[, , 1]),
          y = 1:ncol(a[, , 1]),
          z = a[,order(1:dim(a)[2], decreasing = TRUE),1],
          add = TRUE,
          col = grDevices::hcl.colors(20, "inferno", rev = FALSE),
          useRaster = TRUE)
      }
   }

    if (plot_settings$grid) graphics::grid(nx = max(m[, "img_width"]),
                                           ny = max(m[, "img_height"]))

    ## plot metric scale
    if (!is.null(dim.CCD)) {
      axis(
        side = 1,
        at = axTicks(1),
        labels = paste(floor(dim.CCD[1] / max(m[,"img_width"]) * axTicks(1)), "\u00b5m"),
        lwd = -1,
        lwd.ticks = -1,
        line = -2.2,
        cex.axis = 0.8
      )
      axis(
        side = 2,
        at = axTicks(2)[-1],
        labels = paste(floor(dim.CCD[2] / max(m[,"img_height"]) * axTicks(2)), "\u00b5m")[-1],
        lwd = -1,
        lwd.ticks = -1,
        line = -2.2,
        cex.axis = 0.8
      )
    }

    ## add circles and rectangles
    for (i in 1:nrow(m)) {
      if (!i%in%exclude_ROI) {
        ## mark selected pixels
        x0 <- m[i, "x"] - m[i, "half_width"]
        x1 <- m[i, "x"] + m[i, "half_width"]
        y0 <- m[i, "y"] - m[i, "half_height"]
        y1 <- m[i, "y"] + m[i, "half_height"]
        polygon(
          x = c(x0, x0, x1, x1),
          y = c(y0, y1, y1, y0),
          col = plot_settings$col.pixel
        )
      }

      ## add ROIs
      shape::plotellipse(
        rx = m[i, "half_width"],
        ry = m[i, "half_width"],
        mid = c(m[i, "x"], m[i, "y"]),
        lcol = plot_settings$col.ROI,
        lty = plot_settings$lty.ROI,
        lwd = plot_settings$lwd.ROI)
    }

    ## add distance marker
    points(
      x = m[!m[,"ROI"]%in%sel_euc_dist & !m[,"ROI"]%in%exclude_ROI, "x"],
      y = m[!m[,"ROI"]%in%sel_euc_dist & !m[,"ROI"]%in%exclude_ROI, "y"],
      pch = 4,
      col = "red")

    ## add text
    if(length(m[-exclude_ROI,"x"]) > 0) {
      graphics::text(
         x = m[-exclude_ROI, "x"],
         y = m[-exclude_ROI, "y"],
         labels = plot_settings$text.labels[-exclude_ROI],
         cex = 0.6,
         col = if(!is.null(bg_image)) "white" else "black",
         pos = 1,
         offset = plot_settings$text.offset
       )
    }

    ##add legend
    if(plot_settings$legend) {
      legend(
        plot_settings$legend.pos,
        bty = "",
        pch  = c(1, 15, 4),
        box.lty = 0,
        bg = rgb(1,1,1,0.7),
        legend = plot_settings$legend.text,
        col = c(plot_settings$col.ROI, plot_settings$col.pixel, "red")
      )
    }

  }##end if plot

  ## return results
  invisible(set_RLum(
    class = "RLum.Results",
    data = list(
      ROI = m,
      euc_dist = euc_dist),
    info = list(
      call = sys.call()
    )))
}

#'@title Import Luminescence Data from Helios Luminescence Reader
#'
#' @description
#' Import of files with `.osl` extension produced by the zero rad Helios
#' luminescence reader and conversion to [Luminescence::RLum.Analysis-class]
#' objects.
#'
#' @param file [character], [list] (**required**):
#' name of one or multiple `.osl` files (URLs are supported); it can be the
#' path to a directory, in which case the function tries to detect and import
#' all `.osl` files found in the directory.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#'@param ... not in use, for compatibility reasons only
#'
#'@note Thanks to Krzysztof Maternicki for providing example data.
#'
#' @return
#' A [Luminescence::RLum.Analysis-class] object. Results are returned as a
#' list when multiple files are processed or `file` is a list.
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @section Function version: 0.1.1
#'
#'@seealso [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Analysis-class]
#'
#'@keywords IO
#'
#'@examples
#'file <- system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence")
#'read_HeliosOSL2R(file)
#'
#'@export
read_HeliosOSL2R <- function(
  file,
  verbose = TRUE,
  ...
) {
  .set_function_name("read_HeliosOSL2R")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_logical_scalar(verbose)
  file <- .validate_file(file, ext = "osl", pattern = "\\.osl$",
                         verbose = verbose)

  ## Self-call --------------------------------------------------------------
  if (inherits(file, "list")) {
    out <- lapply(file, function(x) {
      tmp <- try(read_HeliosOSL2R(x, verbose = verbose), silent = TRUE)
      if(inherits(tmp, "try-error")) {
        .throw_message(strsplit(tmp, ":", fixed = TRUE)[[1]][2], "->... record skipped!")
        return(NULL)
      }
      tmp
    })

    ## remove NULL
    out <- .rm_NULL_elements(out)

    if(length(out) == 0)
      out <- NULL

    return(out)
  }

  ## Import -----------------------------------------------------------------

  if (verbose) {
    cat("\n[read_HeliosOSL2R()] Importing ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file)))
    cat("\n")
  }

  ## read entire file
  lines <- readLines(file)

    ## get footer lines id, which is printed in quotes but not the first
    footer_id <- grep(pattern = '\\"', x = lines)[2]

    ## import data measurement data
    df <- read.table(
      file = textConnection(lines[1:(footer_id - 1)]),
      header = TRUE,
      sep = ",")

    ## extract metadata
    ## remove quotes and split at : + whitespace
    meta <- strsplit(x = gsub(
      pattern = '\\"',
      replacement = "",
      x = lines[footer_id:length(lines)]),
      split = ": ",
      fixed = TRUE)

    ## drop lines object
    rm(lines)

    ## create info object
    ## extract names
    info_names <- vapply(seq_along(meta), function(x){
      if(length(meta[[x]]) == 2)
        meta[[x]][1]
      else
        paste0("unnamed_", x)

    }, character(1))
    info_names <- gsub(pattern = " ", "_", x = info_names, fixed = TRUE)

    ## extra elements
    info <- lapply(meta, function(x){
      if(length(x) == 2)
        x[2]
      else
        x[1]
    })
    names(info) <- info_names

# Create object -----------------------------------------------------------
  ## this needs to be checked for the moment
  ## we extract only three curves
  pid <- create_UID()

  ## get records
  records <- lapply(3:ncol(df), function(x){
    set_RLum(
      class = "RLum.Data.Curve",
      originator = "read_HeliosOSL2R",
      curveType = "measured",
      recordType = paste0("OSL (", colnames(df)[x], ")"),
      data = as.matrix(df[,c(2,x)]),
      info = c(
        xlab = colnames(df)[2],
        ylab = colnames(df)[x],
        parentID = pid,
        info))
  })

  ## create RLum.Analysis as output
  set_RLum(
    class = "RLum.Analysis",
    originator = "read_HeliosOSL2R",
    records = records)
}

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_HeliosOSL2R(data.frame()),
               "'file' should be of class 'character' or 'list'")
  expect_error(read_HeliosOSL2R(character(0)),
               "'file' cannot be an empty character")
  expect_error(read_HeliosOSL2R(NA_character_),
               "File 'NA' does not exist")
  expect_error(read_HeliosOSL2R("error"),
               "File '.*error' does not exist") # windows CI needs the regexp
  expect_error(object = read_HeliosOSL2R(test_path("_data/BINfile_V3.bin")),
               "[read_HeliosOSL2R()] File extension 'bin' is not supported, only",
               fixed = TRUE)
})

test_that("Test functionality", {
  testthat::skip_on_cran()

  ## standard input
  file <- system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence")
  SW({
  expect_s4_class(
    object = read_HeliosOSL2R(file),
    class = "RLum.Analysis")
  })

  ## no verbose
  expect_silent(
    object = read_HeliosOSL2R(file, verbose = FALSE))

  ## list input
  files <- list(file, file)
  SW({
  expect_type(
    object = read_HeliosOSL2R(files),
    type = "list")
  })

  ## list input
  files <- list(file, file, "wrong.xs")
  SW({
    t <- expect_type(
      object = read_HeliosOSL2R(files, verbose = FALSE),
      type = "list")
  })
  expect_length(t,2)

  ## all list elements skipped
  expect_message(expect_null(read_HeliosOSL2R(list("error"), verbose = FALSE)),
                 "record skipped")
})

#' @title Plot function for an RLum.Data.Curve S4 class object
#'
#' @description The function provides a standardised plot output for curve data of an
#' [Luminescence::RLum.Data.Curve-class] S4-class object.
#'
#' @details Only single curve data can be plotted with this function. Arguments
#' according to [plot].
#'
#' **Curve normalisation**
#'
#' The argument `norm` normalises all count values. To date the following
#' options are supported:
#'
#' `norm = TRUE` or `norm = "max"`: Curve values are normalised to the highest
#' count value in the curve
#'
#' `norm = "min"`: Curve values are normalised to the smallest count value
#' in the curve
#'
#' `norm = "first"`: Curve values are normalised to the very first count value
#'
#' `norm = "last"`: Curve values are normalised to the last count value
#' (this can be useful in particular for radiofluorescence curves)
#'
#' `norm = "huot"`: Curve values are normalised as suggested by Sébastien Huot
#'  via GitHub:
#' \deqn{
#' y = (observed - median(background)) / (\max(observed) - median(background))
#' }
#'
#' The background of the curve is defined as the last 20% of the count values
#' of a curve.
#'
#' `norm = "intensity"`: Curve values are normalised to the channel length.
#'
#' `norm = 2.2`: Curve values are normalised to a positive number (e.g., 2.2).
#'
#' @param object [Luminescence::RLum.Data.Curve-class] (**required**):
#' S4 object of class [Luminescence::RLum.Data.Curve-class]
#'
#' @param par.local [logical] (*with default*):
#' use local graphical parameters for plotting, e.g. the plot is shown in one
#' column and one row. If `par.local = FALSE`, global parameters are inherited.
#'
#' @param norm [logical] [character] (*with default*): whether curve
#' normalisation should occur (`FALSE` by default). Alternatively, the function
#' offers modes `"max"` (used with `TRUE`), `"last"` and `"huot"`, see details.
#'
#' @param smooth [logical] (*with default*):
#' provides automatic curve smoothing based on the internal function `.smoothing`
#'
#' @param auto_scale [logical] (*with default*): if activated, auto scales `xlim` or `ylim`
#' to the extent of the other. If both are set, the auto-scaling is skipped.
#'
#' @param interactive [logical] (*with default*): enables/disables interactive
#' plotting mode using [plotly::plot_ly]
#'
#' @param ... further arguments and graphical parameters that will be passed
#' to [graphics::plot.default] and [graphics::par]
#'
#' @return Returns a plot.
#'
#' @note Not all arguments of [plot] will be passed!
#'
#' @section Function version: 0.4.0
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [plot], [Luminescence::plot_RLum]
#'
#' @keywords aplot
#'
#' @examples
#'
#' ##plot curve data
#'
#' #load Example data
#' data(ExampleData.CW_OSL_Curve, envir = environment())
#'
#' #transform data.frame to RLum.Data.Curve object
#' temp <- as(ExampleData.CW_OSL_Curve, "RLum.Data.Curve")
#'
#' #plot RLum.Data.Curve object
#' plot_RLum.Data.Curve(temp)
#'
#' @export
plot_RLum.Data.Curve<- function(
  object,
  par.local = TRUE,
  norm = FALSE,
  smooth = FALSE,
  auto_scale = FALSE,
  interactive = FALSE,
  ...
) {
  .set_function_name("plot_RLum.Data.Curve")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks  ------------------------------------------------------
  .validate_class(object, "RLum.Data.Curve")

  ## check for NA values
  if(all(is.na(object@data))){
    .throw_warning("Curve contains only NA-values, nothing plotted")
    return(NULL)
  }

  ## `norm` is not validated here but will be validated by normalise_RLum()
  .validate_logical_scalar(par.local)
  .validate_logical_scalar(smooth)
  .validate_logical_scalar(auto_scale)
  .validate_logical_scalar(interactive)

# Preset plot -------------------------------------------------------------

  lab.unit <- "unknown"
  lab.xlab <- "Independent"
  xlab.xsyg <- ylab.xsyg <- NULL

  ## set labelling unit
  if (!is.na(object@recordType)) {
    recordType.stripped <- gsub(" \\(.*)", "", object@recordType[1])
    if (recordType.stripped %in% c("OSL", "IRSL", "RL", "RF", "LM-OSL", "RBR")) {
        lab.unit <- "s"
        lab.xlab <- "Stimulation time"
    } else if (recordType.stripped == "TL") {
        lab.unit <- "\u00B0C"
        lab.xlab <- "Temperature"
    }
  }

    ##XSYG
    ##check for curveDescripter
    if ("curveDescripter" %in% names(object@info)) {
      temp.lab <-
        strsplit(object@info$curveDescripter,
                 split = ";",
                 fixed = TRUE)[[1]]
      xlab.xsyg <- temp.lab[1]
      ylab.xsyg <- temp.lab[2]
    }

  xlab <- xlab.xsyg %||% paste0(lab.xlab, " [", lab.unit, "]")
  ylab <- ylab.xsyg %||% (
    if (lab.xlab == "Independent") {
      "Dependent [unknown]"
    } else {
      vals <- object@data[, 1]
      sprintf("%s [cts/%.2f %s]", object@recordType,
              round(max(vals) / length(vals), digits = 2), lab.unit)
    }
  )

  sub <- if (grepl("TL", object@recordType) && "RATE" %in% names(object@info)) {
           paste0("(", object@info$RATE, " K/s)")
         } else if (grepl("OSL|IRSL", object@recordType) &&
                    "interval" %in% names(object@info)) {
           paste("(resolution:", object@info$interval, "s)")
         }

  ## curve normalisation
  if (!isFALSE(norm)) {
    object <- normalise_RLum(object, norm)
  }

  extraArgs <- list(...)

  ## set auto scale if required and only one of xlim or ylim is specified
  if (auto_scale && xor(is.null(extraArgs$xlim), is.null(extraArgs$ylim))) {
    if (!is.null(extraArgs$xlim))  {
        x_range <- range(object@data[,1])
        x_min <- max(c(x_range[1], min(c(extraArgs$xlim[1], x_range[2]))), na.rm = TRUE)
        x_max <- min(c(x_range[2], max(c(extraArgs$xlim[2], x_range[1]))), na.rm = TRUE)

        object@data <- object[object@data[,1] >= x_min & object@data[,1] <= x_max, ,drop = FALSE]
      }
    if (!is.null(extraArgs$ylim)) {
        y_range <- range(object@data[,2])
        y_min <- max(c(y_range[1], min(c(extraArgs$ylim[1], y_range[2]))), na.rm = TRUE)
        y_max <- min(c(y_range[2], max(c(extraArgs$ylim[2], y_range[1]))), na.rm = TRUE)

        object@data <- object[object@data[,2] >= y_min & object@data[,2] <= y_max, ,drop = FALSE]
      }
  }

  ## deal with additional arguments
  plot_settings <- modifyList(x = list(
      main = object@recordType[1],
      xlab = xlab,
      ylab = ylab,
      sub = sub,
      cex = 1,
      type = if(interactive) "scatter" else "l",
      las = NULL,
      lwd = 1,
      lty = 1,
      pch = NULL,
      col = 1,
      axes = TRUE,
      log = "",
      mtext = ""
    ), val = extraArgs, keep.null = TRUE)

    ##par setting for possible combination with plot method for RLum.Analysis objects
    if (par.local)
      par(mfrow = c(1,1), cex = plot_settings$cex)

    ##smooth
    if(smooth){
      k <- ceiling(length(object@data[, 2])/100)
      object@data[, 2] <- .smoothing(object@data[, 2], k = k, align = "center")
    }

  ## remove parameters already set in plot_settings
  extraArgs[names(extraArgs) %in% names(plot_settings)] <- NULL
  extraArgs <- c(extraArgs, plot_settings)

  ## remove parameters not supported by plot.default and par
  extraArgs[!names(extraArgs) %in% c(names(formals(graphics::plot.default)), names(par()))] <- NULL

  ## call the plot
  if (interactive) {
        .require_suggested_package("plotly", "Displaying interactive plots")
        p <- do.call(
          plotly::plot_ly,
          args = c(
            list(
              x = object@data[,1],
              y = object@data[,2],
              type = "scatter",
              line = list(
                width = plot_settings$lwd * 1.3,
                dash = switch(as.character(plot_settings$lty), `2` = "dash", `3` = "dot", "solid"),
                color = if(is.numeric(plot_settings$col))
                  grDevices::palette()[plot_settings$col]
                else
                  plot_settings$col)),
              symbol = plot_settings$pch,
              mode = "lines"))

        ## add more scene information
        p <-  plotly::layout(
          p = p,
          xaxis = list(
            title = plot_settings$xlab,
            type = if (grepl(pattern = "x", plot_settings$log[1])) "log" else "linear"
            ),
          yaxis = list(
            title = plot_settings$ylab,
            type = if (grepl(pattern = "y", plot_settings$log[1])) "log" else "linear"
            ),
          title = plot_settings$main,
          annotations = list(plot_settings$mtext)
        )

        ## print and return
        suppressMessages(print(p))
        on.exit(return(p), add = TRUE)

  } else {
    do.call(graphics::plot.default,
            c(list(x = object@data[, 1], y = object@data[, 2]), extraArgs))

    ## plot additional mtext
    mtext(plot_settings$mtext, side = 3, cex = 0.8 * ifelse(par.local, extraArgs$cex, 1))
  }
}

## create dataset
data(ExampleData.CW_OSL_Curve, envir = environment())
temp <- as(ExampleData.CW_OSL_Curve, "RLum.Data.Curve")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(plot_RLum.Data.Curve("error"),
               "'object' should be of class 'RLum.Data.Curve'")
  expect_error(plot_RLum.Data.Curve(temp, norm = "error"),
               "'norm' should be one of 'max', 'min', 'first', 'last', 'huot' or 'intensity'")
  expect_error(plot_RLum.Data.Curve(temp, norm = -2),
               "'norm' should be a single positive value or one of 'max', 'min'")

  temp_NA <- temp
  temp_NA@data[] <- suppressWarnings(NA_real_)
  expect_warning(expect_null(plot_RLum.Data.Curve(temp_NA)),
                 "Curve contains only NA-values, nothing plotted")
  expect_warning(plot_RLum.Data.Curve(set_RLum("RLum.Data.Curve"), norm = TRUE),
                 "Curve normalisation produced Inf/NaN values, values replaced by 0")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ## run function with various conditions
  expect_silent(plot_RLum.Data.Curve(temp))
  expect_silent(plot_RLum.Data.Curve(temp, norm = TRUE))
  expect_silent(plot_RLum.Data.Curve(temp, norm = "max"))
  expect_silent(plot_RLum.Data.Curve(temp, norm = "last"))
  expect_silent(plot_RLum.Data.Curve(temp, norm = "huot"))
  expect_silent(plot_RLum.Data.Curve(temp, norm = "intensity"))
  expect_silent(plot_RLum.Data.Curve(temp, main = "title", col = "red"))
  expect_silent(plot_RLum.Data.Curve(temp, auto_scale = TRUE, xlim = c(10,20)))
  expect_silent(plot_RLum.Data.Curve(temp, auto_scale = TRUE, ylim = c(1,200)))
  expect_silent(plot_RLum.Data.Curve(temp, smooth = TRUE))
  expect_silent(plot_RLum.Data.Curve(temp, par.local = FALSE))

  temp@recordType <- "OSL"
  temp@info <- list(interval = 1)
  expect_silent(plot_RLum.Data.Curve(temp))
  temp@recordType <- "TL"
  temp@info <- list(curveDescripter = "xlab;ylab", RATE = 2)
  expect_silent(plot_RLum.Data.Curve(temp))
})

test_that("check interactive mode", {
  testthat::skip_on_cran()

  ## simple plot
  expect_silent(plot_RLum.Data.Curve(temp, interactive = TRUE))

  ## with arguments
  expect_silent(plot_RLum.Data.Curve(
    object = temp,
    interactive = TRUE,
    lty = 2,
    col = "red",
    log = "xy"))
})

test_that("graphical snapshot tests", {
  testthat::skip_on_cran()
  testthat::skip_if_not_installed("vdiffr")

  SW({
  vdiffr::expect_doppelganger("default",
                              plot_RLum.Data.Curve(temp))
  vdiffr::expect_doppelganger("autoscale-xlim-smooth",
                              plot_RLum.Data.Curve(temp, auto_scale = TRUE,
                                                   xlim = c(10, 20),
                                                   smooth = TRUE))
  vdiffr::expect_doppelganger("autoscale-ylim-mtext-cex",
                              plot_RLum.Data.Curve(temp, auto_scale = TRUE,
                                                   ylim = c(1, 200),
                                                   mtext = "Test", cex = 2))
  })
})

#'@title Fading Correction after Huntley & Lamothe (2001)
#'
#'@description Apply a fading correction according to Huntley & Lamothe (2001) for a given
#'\eqn{g}-value and a given \eqn{t_{c}}
#'
#'@details
#'This function solves the equation used for correcting the fading affected age
#'including the error for a given \eqn{g}-value according to Huntley & Lamothe (2001):
#'
#'\deqn{
#' \frac{A_{f}}{A} = 1 - \kappa * \Big[\ln(\frac{A}{t_c}) - 1\Big]
#'}
#'
#'with \eqn{\kappa} defined as
#'
#'\deqn{
#' \kappa = \frac{\frac{\mathrm{g\_value}}{\ln(10)}}{100}
#'}
#'
#' \eqn{A} and \eqn{A_{f}} are given in ka. \eqn{t_c} is given in s, however, it
#' is internally recalculated to ka.
#'
#' As the \eqn{g}-value slightly depends on the time \eqn{t_{c}} between
#' irradiation and the prompt measurement, a value for `tc` must always be
#' provided.
#' If the \eqn{g}-value was normalised to a distinct
#' time or evaluated with a different tc value (e.g., external irradiation), also
#' the \eqn{t_{c}} value for the \eqn{g}-value needs to be provided (argument `tc.g_value`
#' and then the \eqn{g}-value is recalculated
#' to \eqn{t_{c}} of the measurement used for estimating the age applying the
#' following equation:
#'
#' \deqn{\kappa_{tc} = \kappa_{tc.g} / (1 - \kappa_{tc.g} * \ln(tc/tc.g))}
#'
#' where
#'
#' \deqn{\kappa_{tc.g} = g / 100 / \ln(10)}
#'
#' The error of the fading-corrected age is determined using a Monte Carlo
#' simulation approach. Solving of the equation is performed using
#' [uniroot]. Large values for `n.MC` will significantly
#' increase the computation time.\cr
#'
#' **`n.MC = 'auto'`**
#'
#' The error estimation based on a stochastic process, i.e. for a small number of
#' MC runs the calculated error varies considerably every time the function is called,
#' even with the same input values.
#' The argument option `n.MC = 'auto'` tries to find a stable value for the standard error, i.e.
#' the standard deviation of values calculated during the MC runs (`age.corr.MC`),
#' within a given precision (2 digits) by increasing the number of MC runs stepwise and
#' calculating the corresponding error.
#'
#' If the determined error does not differ from the 9 values calculated previously
#' within a precision of (here) 3 digits the calculation is stopped as it is assumed
#' that the error is stable. Please note that (a) the duration depends on the input
#' values as well as on the provided computation resources and it may take a while,
#' (b) the length (size) of the output
#' vector `age.corr.MC`, where all the single values produced during the MC runs
#' are stored, equals the number of MC runs (here termed observations).
#'
#' To avoid an endless loop the calculation is stopped if the number of observations
#' exceeds 10^7.
#' This limitation can be overwritten by setting the number of MC runs manually,
#' e.g. `n.MC = 10000001`. Note: For this case the function is not checking whether the calculated
#' error is stable.\cr
#'
#' **`seed`**
#'
#' This option allows to recreate previously calculated results by setting the seed
#' for the R random number generator (see [set.seed] for details). This option
#' should not be mixed up with the option **`n.MC = 'auto'`**. The results may
#' appear similar, but they are not comparable!\cr
#'
#' **FAQ**\cr
#'
#' **Q**: Which \eqn{t_{c}} value is expected?\cr
#'
#' **A**: \eqn{t_{c}} is the time in seconds between irradiation and the prompt measurement
#' applied during your \eqn{D_{e}} measurement. However, this \eqn{t_{c}} might
#' differ from the \eqn{t_{c}} used for estimating the \eqn{g}-value. In the
#' case of an SAR measurement \eqn{t_{c}} should be similar, however,
#' if it differs, you have to provide this
#' \eqn{t_{c}}  value (the one used for estimating the \eqn{g}-value) using
#' the argument `tc.g_value`.\cr
#'
#' **Q**: The function could not find a solution, what should I do?\cr
#'
#' **A**: This usually happens for model parameters exceeding the boundaries of the
#' fading correction model (e.g., very high \eqn{g}-value). Please check
#' whether another fading correction model might be more appropriate.
#'
#' @param age.faded [numeric] [vector] (**required**):
#' vector of length 2 containing the uncorrected age and the error in ka
#' (see example).
#'
#' @param g_value [vector] or [Luminescence::RLum.Results-class] (**required**):
#' either a vector of length 2 containing the g-value and error obtained from
#' separate fading measurements (see example), or an [Luminescence::RLum.Results-class]
#' object produced by [Luminescence::analyse_FadingMeasurement]. If the latter, the `tc`
#' argument is set automatically.
#'
#' @param tc [numeric] (**required**):
#' time in seconds between irradiation and the prompt measurement (cf. Huntley & Lamothe 2001).
#' The argument is ignored when `g_value` is an [Luminescence::RLum.Results-class] object.
#'
#' @param tc.g_value [numeric] (*with default*):
#' time in seconds between irradiation and the prompt measurement used in the
#' estimation of the g-value. If the g-value was normalised, the normalisation
#' time (in seconds) should be given, e.g., for a g-value normalised to 2 days,
#' the value 172800 should be used. If nothing is provided the time is set to
#' `tc`, which is usual case for g-values obtained using the
#' SAR method and \eqn{g}-values that have been not normalised to 2 days.
#'
#' @param n.MC [integer] (*with default*):
#' number of Monte Carlo simulation runs for error estimation.
#' If `n.MC = 'auto'` is used the function tries to find a 'stable' error for
#' the age. See details for further information.
#' **Note:** This may take a while!
#'
#' @param seed [integer] (*optional*):
#' sets the seed for the random number generator in R using [set.seed]
#'
#' @param interval [numeric] (*with default*):
#' a vector containing the end-points (age interval) of the interval to be searched for the root in 'ka'.
#' This argument is passed to the function [stats::uniroot] used for solving the equation.
#'
#' @param txtProgressBar [logical] (*with default*):
#' enable/disable the progress bar.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @return Returns an S4 object of type [Luminescence::RLum.Results-class].\cr
#'
#' Slot: **`@data`**\cr
#' \tabular{lll}{
#' **Object** \tab **Type** \tab **Comment** \cr
#'  `age.corr` \tab [data.frame] \tab Corrected age \cr
#'  `age.corr.MC` \tab [numeric] \tab MC simulation results with all possible ages from that simulation \cr
#' }
#'
#' Slot: **`@info`**\cr
#'
#' \tabular{lll}{
#' **Object** \tab **Type** \tab **Comment** \cr
#'  `info` \tab [character] \tab the original function call
#' }
#'
#'
#' @note Special thanks to Sébastien Huot for his support and clarification via e-mail.
#'
#'
#' @section Function version: 0.4.4
#'
#' @author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Results-class], [Luminescence::analyse_FadingMeasurement],
#' [Luminescence::get_RLum], [uniroot]
#'
#' @references
#' Huntley, D.J., Lamothe, M., 2001. Ubiquity of anomalous fading
#' in K-feldspars and the measurement and correction for it in optical dating.
#' Canadian Journal of Earth Sciences, 38, 1093-1106.
#'
#' @keywords datagen
#'
#' @examples
#'
#' ##run the examples given in the appendix of Huntley and Lamothe, 2001
#'
#' ##(1) faded age: 100 a
#' results <- calc_FadingCorr(
#'    age.faded = c(0.1,0),
#'    g_value = c(5.0, 1.0),
#'    tc = 2592000,
#'    tc.g_value = 172800,
#'    n.MC = 100)
#'
#' ##(2) faded age: 1 ka
#' results <- calc_FadingCorr(
#'    age.faded = c(1,0),
#'    g_value = c(5.0, 1.0),
#'    tc = 2592000,
#'    tc.g_value = 172800,
#'    n.MC = 100)
#'
#' ##(3) faded age: 10.0 ka
#' results <- calc_FadingCorr(
#'    age.faded = c(10,0),
#'    g_value = c(5.0, 1.0),
#'    tc = 2592000,
#'    tc.g_value = 172800,
#'    n.MC = 100)
#'
#' ##access the last output
#' get_RLum(results)
#'
#' @export
calc_FadingCorr <- function(
  age.faded,
  g_value,
  tc,
  tc.g_value = tc,
  n.MC = 10000,
  seed = NULL,
  interval = c(0.01,500),
  txtProgressBar = TRUE,
  verbose = TRUE
) {
  .set_function_name("calc_FadingCorr")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(age.faded, c("numeric", "integer"), length = 2)
  .validate_class(g_value, c("numeric", "integer", "RLum.Results"))
  if(inherits(g_value, "RLum.Results")){
    .validate_originator(g_value, "analyse_FadingMeasurement")
    tc <- get_RLum(g_value)[["TC"]]
    g_value <- as.numeric(get_RLum(g_value)[, c("FIT", "SD")])
  } else {
    .validate_length(g_value, 2)
  }

  ## tc is validated only now, as it may be set in the previous block
  .validate_positive_scalar(tc)
  .validate_positive_scalar(tc.g_value)
  .validate_class(interval, "numeric", length = 2)
  .validate_logical_scalar(txtProgressBar)
  .validate_logical_scalar(verbose)


  ##============================================================================##
  ##DEFINE FUNCTION
  ##============================================================================##
  f <- function(x, af, kappa, tc) {
    1 - kappa * (log(x / tc) - 1) - (af / x)
  }

  ##============================================================================##
  ##CALCULATION
  ##============================================================================##

  ##recalculate the g-value to the given tc ... should be similar
  ##of tc = tc.g_value
  ##re-calculation thanks to the help by Sébastien Huot, e-mail: 2016-07-19
  ##Please note that we take the vector for the g_value here
  k0 <- g_value / 100 / log(10)
  k1 <- k0 / (1 - k0 * log(tc[1]/tc.g_value[1]))
  g_value <- 100 * k1 * log(10)

  ##calculate kappa (equation [5] in Huntley and Lamothe, 2001)
  kappa <- g_value / log(10) / 100

  ## transform tc from s to ka years
  tc <- tc / (1000 * .const$year_s)
  tc.g_value <- tc.g_value / (1000 * .const$year_s)

  ##calculate mean value
  temp <-
    try(suppressWarnings(uniroot(
      f,
      interval = interval,
      tol = 0.0001,
      tc = tc,
      extendInt = "yes",
      af = age.faded[1],
      kappa = kappa[1],
      check.conv = TRUE
    )), silent = TRUE)

  if(inherits(temp, "try-error")){
    .throw_message("No solution found, NULL returned: this usually happens ",
                   "for very large, unrealistic g-values, please consider ",
                   "another model for the fading correction")
    return(NULL)
  }

  ##--------------------------------------------------------------------------##
  ##Monte Carlo simulation for error estimation
  tempMC.sd.recent <- NA
  tempMC.sd.count <- 1:10
  counter <- 1

  ## show a progress bar of the process
  if (n.MC == 'auto') {
    n.MC.i <- 10000

    cat("\n[calc_FadingCorr()] ... trying to find stable error value ...")
    if (txtProgressBar) {
      cat("\n -------------------------------------------------------------\n")
      cat(paste0("   ",paste0("(",0:9,")", collapse = "   "), "\n"))
    }
  }else{
    n.MC.i <- .validate_positive_scalar(n.MC, int = TRUE)
  }

  # Start loop  ---------------------------------------------------------------------------------

  ##set object and preallocate memory
  tempMC <- vector("numeric", length = if (n.MC == "auto") 1e+07 else n.MC)
  tempMC[] <- NA
  i <- 1
  j <- n.MC.i

  while (length(unique(tempMC.sd.count)) > 1 || j > length(tempMC)) {

    ##set previous
    if(!is.na(tempMC.sd.recent)){
      tempMC.sd.count[counter] <- tempMC.sd.recent
    }

    ##set seed
    if (!is.null(seed)) set.seed(seed)

    ##set-values
    g_valueMC <- rnorm(n.MC.i,mean = g_value[1],sd = g_value[2])
    age.fadedMC <- rnorm(n.MC.i,mean = age.faded[1],sd = age.faded[2])
    kappaMC <- g_valueMC / log(10) / 100

    ##calculate for all values
    tempMC[i:j] <- suppressWarnings(vapply(X = 1:length(age.fadedMC), FUN = function(x) {
      temp <- try(uniroot(
        f,
        interval = interval,
        tol = 0.001,
        tc = tc,
        af = age.fadedMC[[x]],
        kappa = kappaMC[[x]],
        check.conv = TRUE,
        maxiter = 1000,
        extendInt = "yes"
      ), silent = TRUE)

      ##otherwise the automatic error value finding
      ##will never work
      res <- NA
      if (!inherits(temp, "try-error") && temp$root < 1e8) {
        res <- temp$root
      }
      return(res)

    }, FUN.VALUE = 1))

    i <- j + 1
    j <- j + n.MC.i

    ##stop here if a fixed value is set
    if(n.MC != 'auto'){
      break
    }

    ##set recent
    tempMC.sd.recent <- round(sd(tempMC, na.rm = TRUE), digits = 3)

    if (counter %% 10 == 0) {
      counter <- 1

    }else{
      counter <- counter + 1
    }

    ##show progress in terminal
    if (txtProgressBar) {
      text <- rep("CHECK",10)
      if (counter %% 2 == 0) {
        text[1:length(unique(tempMC.sd.count))] <- "-----"
      }else{
        text[1:length(unique(tempMC.sd.count))] <- " CAL "
      }

      cat(paste("\r ",paste(rev(text), collapse = " ")))
    }
  }

  ##--------------------------------------------------------------------------##

  ##remove all NA values from tempMC
  tempMC <- tempMC[!is.na(tempMC)]

  ## discard wild outliers, as they will bias the error if present
  tempMC <- tempMC[tempMC < 100 * stats::IQR(tempMC)]

  ##obtain corrected age
  age.corr <- data.frame(
    AGE = round(temp$root, digits = 4),
    AGE.ERROR = round(sd(tempMC), digits = 4),
    AGE_FADED = age.faded[1],
    AGE_FADED.ERROR = age.faded[2],
    G_VALUE = g_value[1],
    G_VALUE.ERROR = g_value[2],
    KAPPA = kappa[1],
    KAPPA.ERROR = kappa[2],
    TC = tc,
    TC.G_VALUE = tc.g_value,
    n.MC = n.MC,
    OBSERVATIONS = length(tempMC),
    SEED = seed %||% NA
  )

  ##============================================================================##
  ##OUTPUT VISUAL
  ##============================================================================##
  if(verbose) {
    cat("\n\n[calc_FadingCorr()]\n")
    cat("\n >> Fading correction according to Huntley & Lamothe (2001)")

    if (tc != tc.g_value) {
      cat("\n >> g-value re-calculated for the given tc")
    }

    cat(paste0(
      "\n\n .. used g-value:\t",
      round(g_value[1], digits = 3),
      " \u00b1 ",
      round(g_value[2], digits = 3),
      " %/decade"
    ))
    cat(paste0(
      "\n .. used tc:\t\t",
      format(tc, digits = 4, scientific = TRUE),
      " ka"
    ))
    cat(paste0(
      "\n .. used kappa:\t\t",
      round(kappa[1], digits = 4),
      " \u00b1 ",
      round(kappa[2], digits = 4)
    ))
    cat("\n ----------------------------------------------")
    cat("\n seed: \t\t\t", seed %||% NA)
    cat(paste0("\n n.MC: \t\t\t", n.MC))
    cat(paste0(
      "\n observations: \t\t",
      format(length(tempMC), digits = 2, scientific = TRUE)
    ))
    cat("\n ----------------------------------------------")
    cat(paste0(
      "\n Age (faded):\t\t",
      round(age.faded[1], digits = 4),
      " ka \u00b1 ",
      round(age.faded[2], digits = 4),
      " ka"
    ))
    cat(paste0(
      "\n Age (corr.):\t\t",
      round(age.corr[1], digits = 4),
      " ka \u00b1 ",
      round(age.corr[2], digits = 4),
      " ka"
    ))
    cat("\n ---------------------------------------------- \n")
  }

  ##============================================================================##
  ##OUTPUT RLUM
  ##============================================================================##
  set_RLum(
    class = "RLum.Results",
    data = list(age.corr = age.corr,
                age.corr.MC = tempMC),
    info = list(call = sys.call())
  )
}

#'@title Converts RLum.Analysis and RLum.Data.Curve objects to RLum2Risoe.BINfileData objects
#'
#' @description The functions converts [Luminescence::RLum.Analysis-class] and
#' [Luminescence::RLum.Data.Curve-class] objects (or a [list] of such objects) to
#' [Luminescence::Risoe.BINfileData-class] objects. The function intends to provide a
#' minimum of compatibility between both formats. The created
#' [Luminescence::RLum.Analysis-class] object can be later exported to a BIN-file using
#' function [Luminescence::write_R2BIN].
#'
#'@param object [Luminescence::RLum.Analysis-class] or [Luminescence::RLum.Data.Curve-class] (**required**): input object to
#'be converted
#'
#'@param keep.position.number [logical] (*with default*): keeps the original
#' position number or re-calculate the numbers to avoid doubling
#'
#'@section Function version: 0.1.4
#'
#'@author  Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#'@seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class], [Luminescence::write_R2BIN]
#'
#'@note The conversion can be never perfect. The `RLum` objects may contain information which are
#'not part of the [Luminescence::Risoe.BINfileData-class] definition.
#'
#'@keywords IO
#'
#'@examples
#'
#'##simple conversion using the example dataset
#'data(ExampleData.RLum.Analysis, envir = environment())
#'convert_RLum2Risoe.BINfileData(IRSAR.RF.Data)
#'
#'@return The function returns a [Luminescence::Risoe.BINfileData-class] object.
#'
#'@export
convert_RLum2Risoe.BINfileData <- function(
  object,
  keep.position.number = FALSE
) {
  .set_function_name("convert_RLum2Risoe.BINfileData")
  on.exit(.unset_function_name(), add = TRUE)

  # Self call ---------------------------------------------------------------
  if (inherits(object, "list")) {
    object_list <- lapply(object, convert_RLum2Risoe.BINfileData)

    ##merge objects
    if (length(object_list) == 1)
      return(object_list[[1]])

    return(merge_Risoe.BINfileData(object_list,
                                   keep.position.number = keep.position.number))
  }

  ## Integrity checks -------------------------------------------------------

  .validate_class(object, c("RLum.Analysis", "RLum.Data.Curve"),
                  extra = "a 'list' of such objects")
  .validate_not_empty(object)

  ##RLum.Data.Curve
  if(inherits(object, "RLum.Data.Curve"))
    object <- set_RLum(class = "RLum.Analysis", records = list(object))

  # Set PROTOTYPE & DATA --------------------------------------------------------------------------

  ##set Risoe.BINfiledata prototype
  prototype <- set_Risoe.BINfileData()

  ##grep records (this will avoid further the subsetting)
  records <- object@records

  ##write DATA
  prototype@DATA <- lapply(records, function(x) {x@data[,2]})

  # Create METADATA -----------------------------------------------------------------------------

  ## allowed names
  allowed_names <- names(prototype@METADATA)
  ##create METADATA list
  METADATA_list <- lapply(records, function(x){
    ##grep matching arguments only
    temp <- x@info[toupper(names(x@info)) %in% allowed_names]

    ## correct names if a match was found
    if (length(temp) > 0) {
      names(temp) <- toupper(names(temp))
      return(temp)
    }

    return(list(ID = NA))
  })

  ##make data.frame out of it
  METADATA_df <- as.data.frame(data.table::rbindlist(METADATA_list, fill = TRUE))

  ## sanitize version for change if it comes in wrongly because this
  ## cannot be handled by merge
  if(!is.null(METADATA_df$VERSION))
    METADATA_df$VERSION <- as.character(METADATA_df$VERSION)

  ##write METADATA
  prototype@METADATA <- merge(prototype@METADATA, METADATA_df, all = TRUE)

  ## +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#
  ##fill various missing values

  ##helper function ...otherwise the code gets too nasty ... only used for NA values!
  .replace <- function(field, value){
    ## get class
    tmp_class <- class(prototype@METADATA[[field]])[1]

    ## convert
    value <- switch(
      tmp_class,
      factor = as.factor(prototype@METADATA[[field]]),
      as(prototype@METADATA[[field]], tmp_class)
    )

    prototype@METADATA[[field]][is.na(prototype@METADATA[[field]])] <<- value
  }

  ## >> ID << ##
  prototype@METADATA[["ID"]] <- seq_along(records)

  ## >> SEL << ##
  prototype@METADATA[["SEL"]] <- TRUE

  ## >> VERSION << ##
  .replace("VERSION", "08")

  ## >> RECTYPE << ##
  .replace("RECTYPE", 0)

  ## >> NPOINTS << ##
  if (anyNA(prototype@METADATA[["NPOINTS"]])) {
    prototype@METADATA[["NPOINTS"]] <- vapply(records, function(x){
      length(x@data)/2
    }, FUN.VALUE = numeric(1))
  }

  ## >> LENGTH <<  + PREVIOUS
  if (anyNA(prototype@METADATA[["LENGTH"]])) {
    ##even we have values here before, it will make no difference
    prototype@METADATA[["LENGTH"]] <- (prototype@METADATA[["NPOINTS"]] * 4) + 507
    prototype@METADATA[["PREVIOUS"]] <- c(0,prototype@METADATA[["LENGTH"]][1:length(records) - 1])
  }

  ## >> RUN << ##
  ##if there is only one NA, we should overwrite it, to be consistent
  if (anyNA(prototype@METADATA[["RUN"]]))
    prototype@METADATA[["RUN"]] <- 1:length(records)

  ## >> SET << ##
  .replace("SET", 1)

  ## >> GRAIN << ##
  .replace("GRAIN", 0)

  ## >> GRAINNUMBER << ##
  .replace("GRAINNUMBER", 0)

  ## >> USER << ##
  .replace("USER", "RLum.Data")

  ## >> DTYPE << ##
  .replace("DTYPE", "Natural")

  ## >> LIGHTSOURCE << ##
  .replace("LIGHTSOURCE", "None")

  ## >> SAMPLE << ##
  if (anyNA(prototype@METADATA[["SAMPLE"]])) {
    ##get only the id's to change
    temp_id <- which(is.na(prototype@METADATA[["SAMPLE"]]))

    ##set name
    prototype@METADATA[["SAMPLE"]] <- vapply(temp_id, function(x){
      if(any(names(records[[x]]@info) == "name")){
          records[[x]]@info$name
      }else{
        "unknown"
      }
    }, character(length = 1))
  }

  ## >> COMMENT << ##
  .replace("COMMENT", "convert_RLum2Risoe.BINfileData()")

  ## >> FNAME << ##
  .replace("FNAME", " ")

  ## >> DATE << ## + TIME
  if (anyNA(prototype@METADATA[["DATE"]])) {
    ##get only the id's to change
    temp_id <- which(is.na(prototype@METADATA[["DATE"]]))

    ##set date
    prototype@METADATA[["DATE"]] <- vapply(temp_id, function(x){
      if(any(names(records[[x]]@info) == "startDate")){
        strtrim(records[[x]]@info[["startDate"]], width = 8)
      }else{
        as.character(format(Sys.Date(),"%Y%m%d"))
      }
    }, character(length = 1))

    ##set time
    prototype@METADATA[["TIME"]] <- vapply(temp_id, function(x){
      if(any(names(records[[x]]@info) == "startDate")){
        substr(records[[x]]@info[["startDate"]], start = 9, stop = 14)
      }else{
        as.character(format(Sys.time(),"%H%m%S"))
      }
    }, character(length = 1))
  }

  ## >> LOW << ##
  if (anyNA(prototype@METADATA[["LOW"]])) {
    ##get only the id's to change
    temp_id <- which(is.na(prototype@METADATA[["LOW"]]))

    ##set date
    prototype@METADATA[["LOW"]] <- vapply(temp_id, function(x){
       min(records[[x]]@data[,1])
    }, numeric(length = 1))
  }

  ## >> HIGH << ##
  if (anyNA(prototype@METADATA[["HIGH"]])) {
    ##get only the id's to change
    temp_id <- which(is.na(prototype@METADATA[["HIGH"]]))

    ##set date
    prototype@METADATA[["HIGH"]] <- vapply(temp_id, function(x){
      max(records[[x]]@data[,1])
    }, numeric(length = 1))
  }

  ## >> SEQUENCE << ##
  .replace("SEQUENCE", "")


  ## METADATA >> correct information ----------------------------------------

  ##we have to correct the LTYPE, the format is rather strict
    ##(a) create LTYPE from names of objects
  LTYPE <- vapply(names(object), function(x) {
    strsplit(x, split = " (", fixed = TRUE)[[1]][1]
    }, FUN.VALUE = character(1))

    ##(b) replace characters
    ##(b.1) irradiation
    LTYPE <-  gsub(pattern = "irradiation", replacement = "USER", fixed = TRUE, x = LTYPE)

    ##(b.2 RF
    LTYPE <-  gsub(pattern = "RF", replacement = "RL", fixed = TRUE, x = LTYPE)

    ##set value
    prototype@METADATA[["LTYPE"]] <- LTYPE


  ##correct USER
  ##limit user to 8 characters
  prototype@METADATA[["USER"]] <- strtrim(prototype@METADATA[["USER"]], 8)

  ##correct SAMPLE
  ##limit user to 21 characters
  prototype@METADATA[["SAMPLE"]] <- strtrim(prototype@METADATA[["SAMPLE"]], 20)

  ##replace all remaining NA values by 0
  ##all remaining values are numbers
  prototype@METADATA <- replace(prototype@METADATA, is.na(prototype@METADATA), 0L)

  # Return --------------------------------------------------------------------------------------
  return(prototype)
}

#' @title Transform a CW-OSL curve into a pLM-OSL curve
#'
#' @description Transforms a conventionally measured continuous-wave (CW) curve into a
#' pseudo linearly modulated (pLM) curve using the equations given in Bulur
#' (2000).
#'
#' @details
#' According to Bulur (2000) the curve data are transformed by introducing two
#' new parameters `P` (stimulation period) and `u` (transformed time):
#'
#' \deqn{P=2\max(t)} \deqn{u=\sqrt{(2t*P)}}
#'
#' The new count values are then calculated by
#' \deqn{ctsNEW = cts(u/P)}
#'
#' and the returned `data.frame` is produced by: `data.frame(u,ctsNEW)`
#'
#' The output of the function can be further used for LM-OSL fitting.
#'
#' @inheritParams convert_CW2pLMi
#'
#' @return
#' The function returns the same data type as the input data type with
#' the transformed curve values ([data.frame] or [Luminescence::RLum.Data.Curve-class]).
#'
#' @note
#' The transformation is recommended for curves recorded with a channel
#' resolution of at least 0.05 s/channel.
#'
#' @section Function version: 0.4.4
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#'
#' @seealso [Luminescence::convert_CW2pHMi], [Luminescence::convert_CW2pLMi], [Luminescence::convert_CW2pPMi],
#' [Luminescence::fit_LMCurve], [lm], [Luminescence::RLum.Data.Curve-class]
#'
#' @references
#' Bulur, E., 2000. A simple transformation for converting CW-OSL
#' curves to LM-OSL curves. Radiation Measurements 32, 141-145.
#'
#' **Further Reading**
#'
#' Bulur, E., 1996. An Alternative Technique For Optically Stimulated
#' Luminescence (OSL) Experiment. Radiation Measurements 26, 701-709.
#'
#' @keywords manip
#'
#' @examples
#'
#' ##read curve from CWOSL.SAR.Data transform curve and plot values
#' data(ExampleData.BINfileData, envir = environment())
#'
#' ##read id for the 1st OSL curve
#' id.OSL <- CWOSL.SAR.Data@@METADATA[CWOSL.SAR.Data@@METADATA[,"LTYPE"] == "OSL","ID"]
#'
#' ##produce x and y (time and count data for the data set)
#' x<-seq(CWOSL.SAR.Data@@METADATA[id.OSL[1],"HIGH"]/CWOSL.SAR.Data@@METADATA[id.OSL[1],"NPOINTS"],
#'        CWOSL.SAR.Data@@METADATA[id.OSL[1],"HIGH"],
#'        by = CWOSL.SAR.Data@@METADATA[id.OSL[1],"HIGH"]/CWOSL.SAR.Data@@METADATA[id.OSL[1],"NPOINTS"])
#' y <- unlist(CWOSL.SAR.Data@@DATA[id.OSL[1]])
#' values <- data.frame(x,y)
#'
#' ##transform values
#' values.transformed <- convert_CW2pLM(values)
#'
#' ##plot
#' plot(values.transformed)
#'
#' @export
convert_CW2pLM <- function(
  object,
  ...
) {
  .set_function_name("convert_CW2pLM")
  on.exit(.unset_function_name(), add = TRUE)

  ## deprecated argument
  if ("values" %in% ...names()) {
    object <- list(...)$values
    .deprecated(old = "values", new = "object", since = "1.2.0")
  }

  ## Integrity checks -------------------------------------------------------
  temp.values <- .prepare_CW2pX(object)

  # Calculation -------------------------------------------------------------

  ##curve transformation
  P<-2*max(temp.values[,1])
  u<-((2*temp.values[,1]*P)^0.5)

  ##cw >> plm conversion, according Bulur, 2000
  temp.values[,2]<-temp.values[,2]*(u/P)
  temp.values<-data.frame(u,temp.values[,2])


  # Return values -----------------------------------------------------------

  ##returns the same data type as the input
  if (is.data.frame(object)) {
    return(temp.values)
  }

  set_RLum(
      class = "RLum.Data.Curve",
      recordType = object@recordType,
      data = as.matrix(temp.values),
      info = object@info)
}

#' @title Calculate the cosmic dose rate
#'
#' @description
#' This function calculates the cosmic dose rate taking into account the soft-
#' and hard-component of the cosmic ray flux and allows corrections for
#' geomagnetic latitude, altitude above sea-level and geomagnetic field
#' changes.
#'
#' This function calculates the total cosmic dose rate considering both the
#' soft- and hard-component of the cosmic ray flux.
#'
#' **Internal calculation steps**
#'
#' (1)
#' Calculate total depth of all absorber in hg/cm² (1 hg/cm² = 100 g/cm²)
#'
#' \deqn{absorber = depth_1*density_1 + depth_2*density_2 + ... + depth_n*density_n}
#'
#' (2)
#' If `half.depth = TRUE`
#'
#' \deqn{absorber = absorber/2}
#'
#' (3)
#' Calculate cosmic dose rate at sea-level and 55 deg. latitude
#'
#' a) If absorber is > 167 g/cm² (only hard-component; Allkofer et al.  1975):
#' apply equation given by Prescott & Hutton (1994) (c.f. Barbouti & Rastin
#' 1983)
#'
#' \deqn{D0 = C/(((absorber+d)^\alpha+a)*(absober+H)) * \exp(-B*absorber)}
#'
#' b) If absorber is < 167 g/cm² (soft- and hard-component): derive D0 from
#' Fig. 1 in Prescott & Hutton (1988).
#'
#' (4)
#' Calculate geomagnetic latitude (Prescott & Stephan 1982, Prescott &
#' Hutton 1994)
#'
#' \deqn{\lambda = \arcsin(0.203\cos(latitude)\cos(longitude-291)
#'               + 0.979\sin(latitude))}
#'
#' (5)
#' Apply correction for geomagnetic latitude and altitude above sea-level.
#' Values for F, J and H were read from Fig. 3 shown in Prescott & Stephan
#' (1982) and fitted with 3-degree polynomials for lambda < 35 degree and a
#' linear fit for lambda > 35 degree.
#'
#' \deqn{Dc = D0*(F + J \exp((altitude / 1000) / H))}
#'
#' (6)
#' Optional: Apply correction for geomagnetic field changes in the last
#' 0-80 ka (Prescott & Hutton 1994). Correction and altitude factors are given
#' in Table 1 and Fig. 1 in Prescott & Hutton (1994). Values for altitude
#' factor were fitted with a 2-degree polynomial. The altitude factor is
#' operated on the decimal part of the correction factor.
#'
#' \deqn{Dc' = Dc*correctionFactor}
#'
#'
#' **Usage of `depth` and `density`**
#'
#' (1) If only one value for depth and density is provided, the cosmic dose
#' rate is calculated for exactly one sample and one absorber as overburden
#' (i.e. `depth*density`).
#'
#' (2) In some cases it might be useful to calculate the cosmic dose rate for a
#' sample that is overlain by more than one absorber, e.g. in a profile with
#' soil layers of different thickness and a distinct difference in density.
#' This can be calculated by providing a matching number of values for
#' `depth` and `density` (e.g. `depth = c(1, 2), density = c(1.7, 2.4)`)
#'
#' (3) Another possibility is to calculate the cosmic dose rate for more than
#' one sample of the same profile. This is done by providing more than one
#' values for `depth` and only one for `density`. For example,
#' `depth = c(1, 2, 3)` and `density = 1.7` will calculate the cosmic dose rate
#' for three samples in 1, 2 and 3 m depth in a sediment of density 1.7 g/cm³.
#'
#' @param depth [numeric] (**required**):
#' depth of overburden (m). For more than one absorber use \cr
#' `c(depth_1, depth_2, ..., depth_n)`
#'
#' @param density [numeric] (**required**):
#' average overburden density (g/cm³). For more than one absorber use \cr
#' `c(density_1, density_2, ..., density_n)`
#'
#' @param latitude [numeric] (**required**):
#' latitude (decimal degree), N positive
#'
#' @param longitude [numeric] (**required**):
#' longitude (decimal degree), E positive
#'
#' @param altitude [numeric] (**required**):
#' altitude (m above sea-level)
#'
#' @param corr.fieldChanges [logical] (*with default*):
#' correct for geomagnetic field changes after Prescott & Hutton (1994).
#' Apply only when justified by the data.
#'
#' @param est.age [numeric] (*with default*):
#' estimated age range (ka) for geomagnetic field change correction (0-80 ka allowed)
#'
#' @param half.depth [logical] (*with default*):
#' How to overcome with varying overburden thickness. If `TRUE` only half the
#' depth is used for calculation. Apply only when justified, i.e. when a constant
#' sedimentation rate can safely be assumed.
#'
#' @param error [numeric] (*with default*):
#' general error (percentage) to be implemented on corrected cosmic dose rate estimate
#'
#' @param ... further arguments (`verbose` to disable/enable output to the
#' terminal).
#'
#' @return
#' Returns a terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following element:
#'
#' \item{summary}{[data.frame] summary of all relevant calculation results.}
#' \item{args}{[list] used arguments}
#' \item{call}{[call] the function call}
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @note
#' Despite its universal use, the equation to calculate the cosmic dose
#' rate provided by Prescott & Hutton (1994) is falsely stated to be valid from
#' the surface to 10^4 hg/cm² of standard rock. The original expression by
#' Barbouti & Rastin (1983) only considers the muon flux (i.e. hard-component)
#' and is, by their own definition, only valid for depths between 10-10^4
#' hg/cm².
#'
#' Thus, for near-surface samples (i.e. for depths < 167 g/cm²) the equation
#' of Prescott & Hutton (1994) underestimates the total cosmic dose rate, as it
#' neglects the influence of the soft-component of the cosmic ray flux. For
#' samples at zero depth and at sea-level the underestimation can be as large
#' as ~0.1 Gy/ka. In a previous article, Prescott & Hutton (1988) give another
#' approximation of Barbouti & Rastin's equation in the form of
#'
#' \deqn{D = 0.21 * \exp(-0.070 * absorber + 0.0005 * absorber^2)}
#'
#' which is valid for depths between 150-5000 g/cm². For shallower depths (<
#' 150 g/cm²) they provided a graph (Fig. 1) from which the dose rate can be
#' read.
#'
#' As a result, this function employs the equation of Prescott & Hutton (1994)
#' only for depths > 167 g/cm², i.e. only for the hard-component of the cosmic
#' ray flux. Cosmic dose rate values for depths < 167 g/cm² were obtained from
#' the "AGE" program (Gruen 2009) and fitted with a 6-degree polynomial curve
#' (and hence reproduces the graph shown in Prescott & Hutton 1988). However,
#' these values assume an average overburden density of 2 g/cm³.
#'
#' It is currently not possible to obtain more precise cosmic dose rate values
#' for near-surface samples as there is no equation known to the author of this
#' function at the time of writing.
#'
#'
#' @section Function version: 0.5.3
#'
#' @author
#' Christoph Burow, University of Cologne (Germany)
#'
#' @seealso [Luminescence::BaseDataSet.CosmicDoseRate]
#'
#' @references
#' Allkofer, O.C., Carstensen, K., Dau, W.D., Jokisch, H., 1975.
#' Letter to the editor. The absolute cosmic ray flux at sea level. Journal of
#' Physics G: Nuclear and Particle Physics 1, L51-L52.
#'
#' Barbouti, A.I., Rastin, B.C., 1983. A study of the absolute intensity of muons at sea level
#' and under various thicknesses of absorber. Journal of Physics G: Nuclear and
#' Particle Physics 9, 1577-1595.
#'
#' Crookes, J.N., Rastin, B.C., 1972. An
#' investigation of the absolute intensity of muons at sea-level. Nuclear
#' Physics B 39, 493-508.
#'
#' Gruen, R., 2009. The "AGE" program for the
#' calculation of luminescence age estimates. Ancient TL 27, 45-46.
#'
#' Prescott, J.R., Hutton, J.T., 1988. Cosmic ray and gamma ray dosimetry for
#' TL and ESR. Nuclear Tracks and Radiation Measurements 14, 223-227.
#'
#' Prescott, J.R., Hutton, J.T., 1994. Cosmic ray contributions to dose rates
#' for luminescence and ESR dating: large depths and long-term time variations.
#' Radiation Measurements 23, 497-500.
#'
#' Prescott, J.R., Stephan, L.G., 1982. The contribution of cosmic radiation to the environmental dose for
#' thermoluminescence dating. Latitude, altitude and depth dependences. PACT 6, 17-25.
#'
#' @examples
#'
#' ##(1) calculate cosmic dose rate (one absorber)
#' calc_CosmicDoseRate(depth = 2.78, density = 1.7,
#'                     latitude = 38.06451, longitude = 1.49646,
#'                     altitude = 364, error = 10)
#'
#' ##(2a) calculate cosmic dose rate (two absorber)
#' calc_CosmicDoseRate(depth = c(5.0, 2.78), density = c(2.65, 1.7),
#'                     latitude = 38.06451, longitude = 1.49646,
#'                     altitude = 364, error = 10)
#'
#' ##(2b) calculate cosmic dose rate (two absorber) and
#' ##correct for geomagnetic field changes
#' calc_CosmicDoseRate(depth = c(5.0, 2.78), density = c(2.65, 1.7),
#'                     latitude = 12.04332, longitude = 4.43243,
#'                     altitude = 364, corr.fieldChanges = TRUE,
#'                     est.age = 67, error = 15)
#'
#'
#' ##(3) calculate cosmic dose rate and export results to .csv file
#' #calculate cosmic dose rate and save to variable
#' results<- calc_CosmicDoseRate(depth = 2.78, density = 1.7,
#'                               latitude = 38.06451, longitude = 1.49646,
#'                               altitude = 364, error = 10)
#'
#' # the results can be accessed by
#' get_RLum(results, "summary")
#'
#' #export results to .csv file - uncomment for usage
#' #write.csv(results, file = "c:/users/public/results.csv")
#'
#' ##(4) calculate cosmic dose rate for 6 samples from the same profile
#' ##    and save to .csv file
#' #calculate cosmic dose rate and save to variable
#' results<- calc_CosmicDoseRate(depth = c(0.1, 0.5 , 2.1, 2.7, 4.2, 6.3),
#'                               density = 1.7, latitude = 38.06451,
#'                               longitude = 1.49646, altitude = 364,
#'                               error = 10)
#'
#' #export results to .csv file - uncomment for usage
#' #write.csv(results, file = "c:/users/public/results_profile.csv")
#'
#' @export
calc_CosmicDoseRate<- function(
  depth,
  density,
  latitude,
  longitude,
  altitude,
  corr.fieldChanges = FALSE,
  est.age = NA,
  half.depth = FALSE,
  error = 10,
  ...
) {
  .set_function_name("calc_CosmicDoseRate")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(depth, "numeric")
  .validate_not_empty(depth)
  .validate_class(density, "numeric")
  .validate_not_empty(density)
  .validate_class(latitude, "numeric", length = 1)
  .validate_class(longitude, "numeric", length = 1)
  .validate_class(altitude, "numeric", length = 1)

  if (any(depth < 0) || anyNA(depth) || any(density < 0) || anyNA(density)) {
    .throw_error("No negative or missing values allowed for 'depth' and 'density'")
  }

  if (corr.fieldChanges) {
    if (is.na(est.age)) {
      .throw_error("Correction for geomagnetic field ",
                   "changes requires an age estimate")
    }
    .validate_positive_scalar(est.age)
    if(est.age > 80) {
      .throw_message("No geomagnetic field change correction possible for ",
                     "samples older than 80 ka, 'corr.fieldChanges' set to FALSE")
      corr.fieldChanges<- FALSE
    }
  }

  if (length(depth) < length(density) ||
      (length(depth) > length(density) && length(density) > 1)) {
    .throw_error("The number of values for 'density' should either be 1 ",
                 "or correspond to the number of values for 'depth'")
  }
  .validate_logical_scalar(half.depth)

  settings <- modifyList(list(verbose = TRUE), list(...))

  ##============================================================================##
  ## CALCULATIONS
  ##============================================================================##

  # initialize parameter for Prescott & Hutton (1994) equation
  C<- 6072
  B<- 0.00055
  d<- 11.6
  alpha<- 1.68
  a<- 75
  H<- 212

  #variable needed to check if cosmic dose rate is calculated for more
  #than one sample
  profile.mode<- FALSE

  ## calculate total absorber of n depths and n densities [single sample]
  ## the calculation is still valid if there is only one depth and one density
  if(length(depth)==length(density)){
    hgcm <- sum(depth * density)
  }

  #if there are >1 depths and only one density, calculate
  #absorber for each sample [multi sample]
  else if (length(depth) > length(density) && length(density) == 1) {
    profile.mode<- TRUE
    hgcm <- depth * density
    profile.results <- data.frame(rbind(1:3), cbind(1:length(depth)))
    colnames(profile.results)<- c("depth (m)", "d0 (Gy/ka)",
                                  "dc (Gy/ka)","dc_error (Gy/ka)")
  }

  if (half.depth) {
    hgcm <- hgcm / 2
  }

  pi.radians <- pi / 180
  for(i in 1:length(hgcm)) {
    # calculate cosmic dose rate at sea-level for geomagnetic latitude 55 degrees
    d0 <- (C / ((((hgcm[i] + d)^alpha) + a) * (hgcm[i] + H))) * exp(-B * hgcm[i])

    temp.hgcm <- hgcm[i] * 100
    if (temp.hgcm < 167) {
      d0.ph <- d0

      if (temp.hgcm < 40) {
        d0<- -6*10^-8*temp.hgcm^3+2*10^-5*temp.hgcm^2-0.0025*temp.hgcm+0.2969
      }
      else {
        d0<- 2*10^-6*temp.hgcm^2-0.0008*temp.hgcm+0.2535
      }
      d0 <- max(d0, d0.ph)
    }
    # Calculate geomagnetic latitude
    gml.temp <- 0.203 * cos(pi.radians * latitude) *
      cos(pi.radians * (longitude - 291)) + 0.979 * sin(pi.radians * latitude)
    true.gml <- asin(gml.temp) / pi.radians
    gml <- abs(true.gml)

    # Find values for F, J and H from graph shown in Prescott & Hutton (1994)
    # values were read from the graph and fitted with 3 degree polynomials and a
    # linear part

    if(gml < 36.5) { # Polynomial fit
      F_ph<- -7*10^-7*gml^3-8*10^-5*gml^2-0.0009*gml+0.3988
    }
    else { # Linear fit
      F_ph<- -0.0001*gml + 0.2347
    }

    if(gml < 34) { # Polynomial fit
      J_ph<- 5*10^-6*gml^3-5*10^-5*gml^2+0.0026*gml+0.5177
    }
    else { # Linear fit
      J_ph<- 0.0005*gml + 0.7388
    }

    if(gml < 36) { # Polynomial fit
      H_ph<- -3*10^-6*gml^3-5*10^-5*gml^2-0.0031*gml+4.398
    }
    else { # Linear fit
      H_ph<- 0.0002*gml + 4.0914
    }

    # Apply correction for geomagnetic latitude and altitude according to
    # Prescott & Hutton (1994)
    dc<- d0*(F_ph + J_ph*exp((altitude/1000)/H_ph))


    ## Additional correction for geomagnetic field change
    if (corr.fieldChanges) {

      if(gml <= 35) {

        # Correction matrix for geomagnetic field changes at
        # sea-level (Prescott & Hutton (1994), Table 1)

        corr.matrix<- data.frame(rbind(1:5),1:7)
        colnames(corr.matrix)<- c(0, 10, 20, 30, 35, ">35")
        rownames(corr.matrix)<- c("0-5","5-10","10-15","15-20","20-35","35-50",
                                  "50-80")

        corr.matrix[1,]<- c(0.97, 0.97, 0.98, 0.98, 0.98, 1.00)
        corr.matrix[2,]<- c(0.99, 0.99, 0.99, 0.99, 0.99, 1.00)
        corr.matrix[3,]<- c(1.00, 1.00, 1.00, 1.00, 1.00, 1.00)
        corr.matrix[4,]<- c(1.01, 1.01, 1.01, 1.00, 1.00, 1.00)
        corr.matrix[5,]<- c(1.02, 1.02, 1.02, 1.01, 1.00, 1.00)
        corr.matrix[6,]<- c(1.03, 1.03, 1.02, 1.01, 1.00, 1.00)
        corr.matrix[7,]<- c(1.02, 1.02, 1.02, 1.01, 1.00, 1.00)

        ## Find the correction factor for the given geomagnetic latitude
        row.idx <- cut(est.age, c(0, 5, 10, 15, 20, 35, 50, 80),
                       labels = FALSE)
        col.idx <- cut(gml, c(0, 5, 15, 25, 32.5, 35, Inf),
                       labels = FALSE)
        corr.fac <- corr.matrix[row.idx, col.idx]

        # Find altitude factor via fitted function 2-degree polynomial
        # This factor is only available for positive altitudes
        if(altitude > 0) {
          alt.fac<- -0.026*(altitude/1000)^2 + 0.6628*altitude/1000 + 1.0435

          # Combine geomagnetic latitude correction with altitude
          # correction (figure caption of Fig. 1 in Precott and Hutton (1994))
          corr.fac <- corr.fac + (corr.fac - 1) * alt.fac
        }

        # Final correction of cosmic dose rate
        dc<- dc * corr.fac

        if (settings$verbose)
          cat("corr.fac:", corr.fac, " diff.one:", corr.fac - 1,
              " alt.fac:", alt.fac, "\n")

      } else if (settings$verbose) {
          cat("\n No geomagnetic field change correction necessary for geomagnetic latitude >35 degrees!")
      }
    }

    # calculate error
    dc.err<- dc*error/100

    # save intermediate results before next sample is calculated
    if (profile.mode) {
      profile.results[i,1]<- round(depth[i],2)
      profile.results[i,2]<- round(d0,4)
      profile.results[i,3]<- round(dc,4)
      profile.results[i,4]<- round(dc.err,4)
    }

  }#END.OF.LOOP

  call<- sys.call()
  args<- list(depth = depth, density = density, latitude = latitude, longitude = longitude,
              altitude = altitude, corr.fieldChanges = corr.fieldChanges, est.age = est.age,
              half.depth = half.depth, error = error)

  if(length(hgcm)==1) {

    ##============================================================================##
    ##TERMINAL OUTPUT
    ##============================================================================##
    if (settings$verbose) {
      cat("\n\n [calc_CosmicDoseRate]")
      cat("\n\n ---------------------------------------------------------")
      cat("\n depth (m)              :", depth)
      cat("\n density (g cm^-3)      :", density)
      cat("\n latitude (N deg.)      :", latitude)
      cat("\n longitude (E deg.)     :", longitude)
      cat("\n altitude (m)           :", altitude)
      cat("\n ---------------------------------------------------------")
      cat("\n total absorber (g cm^-2)       :", round(hgcm[i] * 100, 3))
      cat("\n")
      cat("\n cosmic dose rate (Gy ka^-1)    :", round(d0, 4))
      cat("\n  [@sea-level & 55 deg. N G.lat]")
      cat("\n")
      cat("\n geomagnetic latitude (deg.)    :", round(true.gml, 1))
      cat("\n")
      cat("\n cosmic dose rate (Gy ka^-1)    :", round(dc, 4), "+-", round(dc.err, 4))
      cat("\n  [corrected]                 ")
      cat("\n ---------------------------------------------------------\n\n")
    }
    ##============================================================================##
    ##RETURN VALUES
    ##============================================================================##

    if(length(depth)==1) {
      temp1<- data.frame(depth=depth,density=density)
    } else {
      temp1 <- cbind(data.frame(as.list(depth)),
                     data.frame(as.list(density)))
      colnames(temp1) <- c(paste("depth", seq_along(depth)),
                           paste("density", seq_along(density)))
    }

    temp2<- data.frame(latitude=latitude,longitude=longitude,
                       altitude=altitude,total_absorber.gcm2=hgcm*100,
                       d0=d0,geom_lat=true.gml,dc=dc)

    summary<- data.frame(cbind(temp1,temp2))

  } else {

    #terminal output
    if (settings$verbose) {
      cat("\n\n [calc_CosmicDoseRate]")
      cat("\n\n Calculating cosmic dose rate for", length(depth), "samples.\n\n")
      print(profile.results)
    }

    #return value
    add.info<- data.frame(latitude=latitude,longitude=longitude,
                          altitude=altitude,total_absorber.gcm2=hgcm*100,
                          geom_lat=true.gml)
    add.info<- rbind(add.info*length(i))
    colnames(profile.results)<- c("depth","d0","dc","dc_err")

    summary<- data.frame(cbind(profile.results,add.info))
  }

  set_RLum(
      class = "RLum.Results",
      data = list(summary=summary,
                  args=args,
                  call=call))
}

#' @title Analyse portable CW-OSL measurements
#'
#' @description
#' The function analyses CW-OSL curve data produced by a SUERC portable OSL reader and
#' produces a combined plot of OSL/IRSL signal intensities, OSL/IRSL depletion ratios
#' and the IRSL/OSL ratio.
#'
#' @details
#' This function only works with [Luminescence::RLum.Analysis-class] objects produced by [Luminescence::read_PSL2R].
#' It further assumes (or rather requires) an equal amount of OSL and IRSL curves that
#' are pairwise combined for calculating the IRSL/OSL ratio.
#' For calculating the depletion ratios, the cumulative signal of the last *n*
#' channels (same number of channels as specified by `signal_integral`) is
#' divided by cumulative signal of the first *n* channels (`signal_integral`).
#'
#' **Note:** The function assumes the following sequence pattern:
#' `DARK COUNT`, `IRSL`, `DARK COUNT`, `BSL`, `DARK COUNT`. Therefore, the
#' total number of curves in the input object must be a multiple of 5, and
#' there must be 3 `DARK_COUNT` records for each IRSL/BSL pair. If a different
#' sequence was used, the function will produce an error.
#'
#' **Signal processing**
#' The function processes the signals as follows: `BSL` and `IRSL` signals are
#' extracted using the chosen signal integral, dark counts are taken in full.
#'
#' **Working with coordinates**
#' Usually samples are taken from a profile with a certain stratigraphy. In
#' the past, the function calculated an index. Currently, you have two ways
#' of passing on xy-coordinates to the function:
#'
#' * (1) Add coordinates to the sample name during measurement. The form is rather
#' strict and has to follow the scheme `_x:<number>|y:<number>`. Example:
#' `sample_x:0.2|y:0.4`.
#'
#' * (2) Alternatively, you can provide a [list] or [matrix] with the *(x, y)*
#' coordinates of each sample in meters (m) using the `coord` argument:
#' Example: `coord = list(c(0.2, 1), c(0.3,1.2))`
#'
#' If in your profile the x-coordinates were not measured, *x* should be set
#' to 0. Note that, in such case, a surface plot cannot be produced.
#'
#' @param object [Luminescence::RLum.Analysis-class] (**required**):
#' object produced by [Luminescence::read_PSL2R]. The input can be a [list] of such objects,
#' in which case each input is treated as a separate sample and the results
#' are merged.
#'
#' @param signal_integral [numeric] (**required**):
#' vector specifying the range of channels used to calculate the OSL/IRSL
#' signal. It can be provided as a vector of length 2 such as `c(1, 5)`, or
#' as a sequence such as `1:5`, in which case the lowest and highest values
#' define the range.
#'
#' @param integral_input [character] (*with default*):
#' input type for `signal_integral`, one of `"channel"` (default) or
#' `"measurement"`. If set to `"measurement"`, the best matching channels
#' corresponding to the given time range (in seconds) are selected.
#'
#' @param invert [logical] (*with default*):
#' whether the data should be plotted in reverse order (`FALSE` by default).
#'
#' @param normalise [logical] (*with default*):
#' whether the OSL/IRSL signals should be normalised to the *mean* of all
#' corresponding data curves.
#'
#' @param mode [character] (*with default*):
#' analysis mode, one of `"profile"` (the default) or `"surface"` for surface
#' interpolation.
#'
#' @param coord [list] [matrix] (*optional*): a list or a 2-column matrix
#' with the *x* and *y* coordinates for the sampling positions in meters (m),
#' of the same length as the number of samples measured. For example, the
#' coordinates for one sample could be `coord = list(samp1 = c(0.1, 0.2)`.
#' If the *x* coordinates were not measured, *x* should be set to 0.
#' Note that, in such case, a surface plot cannot be produced.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param ... other parameters to be passed to modify the plot output.
#' Supported are `run` to provide the run name (if the input is a `list`, this
#' is set automatically). Further plot parameters accepted are `main`,
#' `col`, `xlim` (a named [list] for profile mode), `ylim`, `ylab`, `xlab`.
#' Additional parameters for `mode = "profile"` are  `type`, `pch`, `grid`
#' (`TRUE`/`FALSE`), `bg_img` (a raster object for the background image,
#' usually a profile photo), `bg_img_positions` (a vector with the four corner
#' positions, see [graphics::rasterImage]), `zlab` (here x-axis labelling).
#' Additional parameters for `mode = "surface"` are `surface_value`
#' ([character] with names of the surfaces to plot), `col_ramp`, `legend`
#' (`TRUE`/`FALSE`), `contour` (`TRUE`/`FALSE`), `contour_nlevels`,
#' `contour_col`, `nx` and `ny` (size of the interpolation grid),
#' `labcex` (scaling of the contour labels), `zlim`.
#'
#' @return
#' Returns an S4 [Luminescence::RLum.Results-class] object with the following elements:
#'
#' `$data`\cr
#' `.. $summary`: [data.frame] with the results\cr
#' `.. $data`: [list] with the [Luminescence::RLum.Analysis-class] objects\cr
#' `.. $args`: [list] the input arguments
#'
#' @seealso [Luminescence::RLum.Analysis-class], [Luminescence::RLum.Data.Curve-class], [Luminescence::read_PSL2R]
#'
#' @author Christoph Burow, University of Cologne (Germany) \cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @section Function version: 0.1.5
#'
#' @keywords datagen plot
#'
#' @examples
#'
#' ## example profile plot
#' # (1) load example data set
#' data("ExampleData.portableOSL", envir = environment())
#'
#' # (2) merge and plot all RLum.Analysis objects
#' merged <- merge_RLum(ExampleData.portableOSL)
#' plot_RLum(
#'  object = merged,
#'  combine = TRUE,
#'  records_max = 5,
#'  legend.pos = "outside")
#' merged
#'
#' # (3) analyse and plot
#' results <- analyse_portableOSL(
#'   merged,
#'   signal_integral = 1:5,
#'   invert = FALSE,
#'   normalise = TRUE)
#' get_RLum(results)
#'
#' @export
analyse_portableOSL <- function(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  invert = FALSE,
  normalise = FALSE,
  mode = c("profile", "surface"),
  coord = NULL,
  plot = TRUE,
  ...
) {
  .set_function_name("analyse_portableOSL")
  on.exit(.unset_function_name(), add = TRUE)

  ## Self-call --------------------------------------------------------------
  if (inherits(object, "list")) {
      temp <- .warningCatcher(lapply(seq_along(object), function(x) {
        analyse_portableOSL(
          object = object[[x]],
          signal_integral = signal_integral,
          integral_input = integral_input,
          invert = invert,
          normalise = normalise,
          mode = mode,
          plot = plot,
          run = paste0("RUN #", x))
      }))

      return(merge_RLum(temp))
  }

  ## Start function ---------------------------------------------------------

  ## only RLum.Analysis objects
  .validate_class(object, "RLum.Analysis")
  .validate_not_empty(object)
  integral_input <- .validate_args(integral_input, c("channel", "measurement"))

  ## only curve objects
  if (!all(sapply(object, class) == "RLum.Data.Curve"))
    .throw_error("The 'RLum.Analysis' object must contain only objects ",
                 "of class 'RLum.Data.Curve'")

  ## check originator
  lapply(object, .validate_originator, "read_PSL2R",
         name = "At least one element of 'object'")

  ## check length and start of the sequence pattern, we check it further below
  if (length(object) %% 5 != 0 ||
      !all(gsub(" (PMT)", "", names(object)[1:5], fixed = TRUE) ==
           c("USER", "IRSL", "USER", "OSL", "USER")))
    .throw_error("Sequence pattern not supported: see the manual for details")

  ## deprecated argument
  if (any(grepl("signal.integral", ...names(), fixed = TRUE))) {
    .deprecated(old = "signal.integral",
                new = "signal_integral",
               since = "1.2.0")
    if (integral_input != "channel") {
      .throw_error("'integral_input' is not supported with old argument names")
    }
    signal_integral <- list(...)$signal.integral
  }

  ## set the maximum signal_integral allowed: as this must be valid across all
  ## records, we cap it to the minimum number of points
  num.points <- min(lengths(get_RLum(object, recordType = c("OSL", "IRSL"))))

  if (integral_input == "measurement") {
    x.range <- get_RLum(object, recordType = c("OSL", "IRSL"))[[1]][, 1]
    signal_integral <- .convert_to_channels(x.range, signal_integral, "time",
                                            null.ok = TRUE)
  }
  signal_integral <- .validate_integral(signal_integral, max = num.points,
                                        null.ok = TRUE)

  ## convert the signal integral to a range
  if (is.null(signal_integral)) {
    signal_integral <- c(1, 1)
    .throw_warning("No value for 'signal_integral' provided. Only the ",
                   "first data point of each curve was used")
  } else {
    signal_integral <- range(signal_integral)
  }

  mode <- .validate_args(mode, c("profile", "surface"))
  .validate_logical_scalar(invert)
  .validate_logical_scalar(normalise)
  .validate_logical_scalar(plot)

  ## set SAMPLE --------
  run <- if ("run" %in% ...names()) list(...)$run
         else object@info$Run_Name %||% "Run #1"

  ## CALCULATIONS ----
  ## Note: the list ... unlist construction is used make sure that get_RLum() always
  ## returns a list
  ### get OSL -------
  OSL <- .unlist_RLum(list(get_RLum(object, recordType = "OSL")))
  OSL <- do.call(rbind, lapply(OSL, .posl_get_signal, signal_integral))

  ### get IRSL -------
  IRSL <- .unlist_RLum(list(get_RLum(object, recordType = "IRSL")))
  IRSL <- do.call(rbind, lapply(IRSL, .posl_get_signal, signal_integral))

  if (nrow(OSL) != nrow(IRSL)) {
    .throw_error("Sequence pattern not supported: the number of OSL records ",
                 "does not match the number of IRSL records")
  }

  ### get DARK counts ----------
  ### we assume that USER contains the dark count measurements
  DARK_COUNT <- .unlist_RLum(list(get_RLum(object, recordType = "USER")))

  ## we expect a sequence pattern with 3 DARK_COUNT records for each OSL/IRSL pair
  num.dark.count <- length(DARK_COUNT)
  if (num.dark.count %% 3 != 0) {
    .throw_error("Sequence pattern not supported: expected ", nrow(OSL) * 3,
                 " DARK_COUNT records, but found ", num.dark.count)
  }

  DARK_COUNT <- lapply(seq(1, num.dark.count, 3), function(x) DARK_COUNT[x:(x+2)])
  DARK_COUNT <- do.call(rbind, lapply(DARK_COUNT, .posl_get_dark_count))

  ### NORMALISE ----
  if (normalise) {
    OSL <- .posl_normalise(OSL)
    IRSL <- .posl_normalise(IRSL)
  }

  ### OSL/IRSL Ratio -------
  RATIO <- IRSL$sum_signal / OSL$sum_signal

  ### extract  coordinates -------
  if(is.null(coord)) {
    settings_sample <- unique(
        vapply(object, function(x) x@info$settings$Sample, character(1)))
    num.names <- length(settings_sample)
    if (num.names != length(RATIO)) {
      .throw_error("'object' references ", num.names, " sample names, but ",
                   length(RATIO), " IRSL/OSL pairs found")
    }
    coord <- .extract_PSL_coord(settings_sample)

  } else {
    .validate_class(coord, c("matrix", "list"))
    if(inherits(coord, "list"))
      coord <- do.call(rbind, coord)

    ## check length
    if(nrow(coord) != length(OSL$sum_signal))
      .throw_error("The number of coordinates in 'coord' should match the ",
                   "number of samples (", length(OSL$sum_signal), ")")
    if (ncol(coord) != 2)
      .throw_error("'coord' should specify two coordinates per sample")
  }

  ### GENERATE SUMMARY data.frame -----
  summary <- data.frame(
    ID = seq_along(OSL$sum_signal),
    RUN = run,
    BSL = OSL$sum_signal,
    BSL_error = OSL$sum_signal_err,
    IRSL = IRSL$sum_signal,
    IRSL_error = IRSL$sum_signal_err,
    BSL_depletion = OSL$sum_signal_depletion,
    IRSL_depletion = IRSL$sum_signal_depletion,
    IRSL_BSL_RATIO = RATIO,
    DARK =  DARK_COUNT$mean_dark_count,
    DARK_error = DARK_COUNT$sd_dark_count,
    COORD_X = coord[,1],
    COORD_Y = coord[,2]
  )

    ## if coordinates exist, sort by depth
    if (!anyNA(coord[, 2]))
      summary <- summary[order(coord[,2]),]

   ### INVERT ----------
   if(invert)
     summary <- summary[nrow(summary):1,]

  if (mode == "surface" && plot && all(range(summary$COORD_X) == c(0, 0))) {
    plot <- FALSE
    .throw_message("Surface plot is not available when all x-coordinates ",
                   "are 0, plot reset to FALSE", error = FALSE)
  }

  # PLOTTING -------------------------------------------------------------------
  ## generate list of plot matrices
  ## this done to have consistent settings for all plot types
  parm <- grep("^(BSL|IRSL|DARK)", colnames(summary), value = TRUE)
  m_list <- lapply(parm, function(x){
     cbind(x = summary[["COORD_X"]], y = summary[["COORD_Y"]], value = summary[[x]])
  })

    ## correct names of the list
    names(m_list) <- parm

  if (plot) {

    ## add a few attributes to be used later
    attr(m_list, "xlim") <- lapply(m_list, function(x) range(x[,1]))
    attr(m_list, "ylim") <- if(invert) rev(range(m_list[[1]][,2])) else range(m_list[[1]][,2])
    attr(m_list, "zlim") <- lapply(m_list, function(x) range(x[,3]))

    ## account for surface case
    if (mode == "surface") {
      attr(m_list, "ylim") <- if (invert) rev(range(summary$COORD_Y)) else range(summary$COORD_Y)
      attr(m_list, "xlim") <- range(summary$COORD_X)
    }

   ## preset plot settings
   ## plot settings -------
   plot_settings <- modifyList(
     x = list(
       col_ramp = grDevices::heat.colors(30, rev = TRUE, alpha = 0.5),
       bg_img = NULL,
       bg_img_positions = NULL,
       surface_value = c("BSL", "IRSL", "IRSL_BSL_RATIO"),
       legend = TRUE,
       type = "b",
       cex = 1,
       labcex = 0.6,
       col = c("blue", "red", "blue", "red", "black", "grey"),
       pch = rep(16, length(m_list)),
       xlim = attr(m_list, "xlim"),
       ylim = attr(m_list, "ylim"),
       zlim = if(mode == "surface") NA else attr(m_list, "zlim"),
       ylab = if (!anyNA(summary$COORD_Y)) "Depth [m]" else "Index",
       xlab = "x [m]",
       nx = 200,
       ny = 200,
       grid = TRUE,
       contour = FALSE,
       contour_nlevels = 10,
       contour_col = "grey",
       zlab = c("BSL", "IRSL", "BSL depl.", "IRSL depl.", "IRSL/BSL", "mean DARK"),
       main = summary$RUN[1]
     ),
     val = list(...), keep.null = TRUE)

    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    ## mode == "surface" ---------
    if (mode == "surface") {
     ### check for validity of surface value -------
     if(!all(plot_settings$surface_value %in% names(m_list)))
       .throw_error("Unknown value to plot, valid values are: ",
                    .collapse(names(m_list)))

     ## set par -------
     if(length(plot_settings$surface_value) > 1) {
       par(mfrow = c(2, 2))
     }

     ## loop over surface values -------
     for(i in plot_settings$surface_value) {
       ## set matrix for the plot
       m <- m_list[[i]]

       ## respect xlim and ylim range
       m <- m[m[, 2] >= min(plot_settings$ylim) &
              m[, 2] <= max(plot_settings$ylim), , drop = FALSE]
       m <- m[m[, 1] >= min(plot_settings$xlim) &
              m[, 1] <= max(plot_settings$xlim), , drop = FALSE]

       ## respect z_values
       if(!all(is.na(plot_settings$zlim)))
         m <- m[m[, 3] >= min(plot_settings$zlim) &
                m[, 3] <= max(plot_settings$zlim), , drop = FALSE]

       ## interpolate ------
       s <-
         try(interp::interp(
           x = m[, 1],
           y = m[, 2],
           z = m[, 3],
           nx = plot_settings$nx,
           ny = plot_settings$ny
         ), silent = TRUE)

       ## show only warning
       if(inherits(s, "try-error"))
         .throw_warning("Surface interpolation failed: this happens when ",
                        "all points are arranged in one line or xlim/ylim/zlim ",
                        "are too tight. Nothing plotted!")

       ## show error
       if(!inherits(s, "try-error")) {
         par(mar = c(4.5, 4.5, 4, 3), xpd = FALSE)

         ## open empty plot
         plot(
           x = NA,
           y = NA,
           ylim = plot_settings$ylim,
           xlim = plot_settings$xlim,
           xlab = plot_settings$xlab,
           ylab = plot_settings$ylab,
           cex.lab = plot_settings$cex,
           cex.axis = plot_settings$cex,
           main = plot_settings$main)

         ## add background image if available -------
         if (!is.null(plot_settings$bg_img)) {
           ## get corner positions
           positions <- plot_settings$bg_img_positions[1:4] %||% par()$usr
           graphics::rasterImage(
             image = plot_settings$bg_img,
             xleft = positions[1],
             ybottom = positions[4],
             xright = positions[2],
             ytop = positions[3],
             interpolate = TRUE)
         }

         ## plot image -------
         graphics::image(
           s,
           col = plot_settings$col_ramp,
           add = TRUE
         )

         ## add contour
         if (plot_settings$contour)
           graphics::contour(
             x = s$x,
             y = s$y,
             z = s$z,
             add = TRUE,
             labcex = plot_settings$labcex * plot_settings$cex,
             nlevels = plot_settings$contour_nlevels,
             col = plot_settings$contour_col)

         ## add points
         points(m[, 1:2], pch = 20, cex = plot_settings$cex)

         ## add what is shown in the plot
         mtext(side = 3, text = i, cex = plot_settings$cex * 0.7)

         ## add legend
         if(plot_settings$legend) {
           par(xpd = NA)

           col_grad <- plot_settings$col_ramp[
             seq(1, length(plot_settings$col_ramp), length.out = 14)]

           slices <- seq(par()$usr[3],par()$usr[4],length.out = 15)

           for(s in 1:(length(slices) - 1)){
             graphics::rect(
               xleft = par()$usr[2] * 1,
               xright = par()$usr[2] * 1.02,
               ybottom = slices[s],
               ytop =  slices[s + 1],
               col = col_grad[s],
               border = TRUE)
           }

           ## add legend text
           text(
             x = par()$usr[2] * 1.03,
             y = par()$usr[4],
             labels = if(is.null(plot_settings$zlim_image)) {
               format(max(m[,3]), digits = 1, scientific = TRUE)
             } else {
               format(plot_settings$zlim_image[2], digits = 1, scientific = TRUE)
             },
             cex = plot_settings$cex * 0.6,
             srt = 270,
             pos = 3)

           text(
             x = par()$usr[2] * 1.03,
             y = par()$usr[3],
             labels = if(is.null(plot_settings$zlim_image)) {
               format(min(m[,3]), digits = 1, scientific = TRUE)
             } else {
               format(plot_settings$zlim_image[1], digits = 1, scientific = TRUE)
             },
             cex = plot_settings$cex * 0.6,
             pos = 3,
             srt = 270)

           ## add legend labelling (central)
           text(
             x = par()$usr[2] * 1.035,
             y = (par()$usr[4] - par()$usr[3])/2 + par()$usr[3],
             labels = "Intensity [a.u.]",
             cex = plot_settings$cex * 0.7,
             pos = 3,
             srt = 270)
         }
       }
     }# end for loop
    }

    ## mode == "profile" ---------
    if (mode == "profile") {

    # default: par(mar = c(5, 4, 4, 2) + 0.1) // bottom, left, top, right
    par(mfrow = c(1, 7))
    par(mar = c(5, 4, 4, 1) + 0.1)

    graphics::frame()

      mtext(side = 3, plot_settings$main, line = 2,
            cex = plot_settings$cex * 0.7)

    par(mar = c(5, 0, 4, 1) + 0.1)

    ## make sure that wrong zlim settings do not screw up the function
    if(!inherits(plot_settings$zlim, "list")) {
      .throw_warning("In profile mode, zlim needs to be provided as a named ",
                     "list, example: list(BSL = c(0,1)). Reset to default")
      plot_settings$zlim <- attr(m_list, "zlim")
    }

      ## plot the profile for each measure
      profile.parts <- c("BSL", "IRSL", "BSL_depletion", "IRSL_depletion",
                         "IRSL_BSL_RATIO", "DARK")
      for (prof in profile.parts) {
        idx <- match(prof, profile.parts)

        xlim <- plot_settings$zlim[[prof]]
        if (prof == "DARK") {
          xlim <- range(c(
              plot_settings$zlim[["DARK"]] - plot_settings$zlim[["DARK_error"]],
              plot_settings$zlim[["DARK"]] + plot_settings$zlim[["DARK_error"]]))
        }
        plot(
            NA,
            NA,
            ylim = plot_settings$ylim,
            xlim = xlim,
            xlab = plot_settings$zlab[idx],
            ylab = "",
            cex.lab = plot_settings$cex,
            cex.axis = plot_settings$cex,
            bty = "n",
            yaxt = "n"
        )
        if (plot_settings$grid)
          graphics::grid()

        x.val <- m_list[[prof]][, "value"]
        y.val <- m_list[[prof]][, "y"]
        lines(
            x = x.val,
            y = y.val,
            type = plot_settings$type,
            cex = plot_settings$cex,
            lty = ifelse(prof %in% c("BSL_depletion", "IRSL_depletion"), 2, 1),
            pch = plot_settings$pch[idx],
            col = plot_settings$col[idx]
        )

        ## add error bars
        if (prof %in% c("BSL", "IRSL", "DARK")) {
          x.err <- m_list[[paste0(prof, "_error")]][, "value"]
          segments(
              x0 = x.val - x.err,
              x1 = x.val + x.err,
              y0 = y.val,
              y1 = y.val,
              col = plot_settings$col[idx])
        }

        axis(3, cex.axis = plot_settings$cex)

        ## add general axis labels
        if (idx == 1) {
          labs <- pretty(y.val, n = 10)
          axis(2, line = 3, at = labs, labels = labs,
               cex.axis = plot_settings$cex)
          mtext(plot_settings$ylab[1], side = 2, line = 6,
                cex = plot_settings$cex * 0.8)
        }
      }

    } ## end mode == "profile"
  }

  ## RETURN VALUE ----
  call<- sys.call()
  args <- as.list(call)[2:length(call)]

  set_RLum(
    class = "RLum.Results",
    data = list(
      summary=summary,
      data = object,
      args=args
    ),
    info = list(call = call))
}

# HELPER FUNCTIONS ----------
## This extracts the relevant curve data information of the RLum.Data.Curve
## objects
.posl_get_signal <- function(x, sigint) {
    raw_signal <- get_RLum(x)[,2]
    sum_signal <- sum(raw_signal[sigint[1]:sigint[2]])
    sum_signal_err <- sqrt(sum(x@info$raw_data$counts_per_cycle_error[sigint[1]:sigint[2]]^2))
    sum_signal_depletion <- sum(raw_signal[(length(raw_signal)-length(sigint[1]:sigint[2])):length(raw_signal)]) / sum_signal
    return(data.frame(sum_signal, sum_signal_err, sum_signal_depletion))
}

.posl_get_dark_count <- function(x) {
  ## we do assume a fixed sequence pattern, hence, we know what to
  ## expect that anything that comes in here, can be merged
  counts <- unlist(lapply(x, function(x) as.matrix(x)[,2]))

  return(data.frame(mean_dark_count = mean(counts), sd_dark_count = sd(counts)))
}

## This function normalises the data curve by the mean signal
.posl_normalise <- function(x) {
  rel.error <- x$sum_signal_err / x$sum_signal
  x$sum_signal <- x$sum_signal / mean(x$sum_signal)
  x$sum_signal_err <- x$sum_signal * rel.error
  x$sum_signal_depletion <- x$sum_signal_depletion / mean(x$sum_signal_depletion)
  return(x)
}

## This function extracts the coordinates from the sample names
.extract_PSL_coord <- function(settings_sample) {
  ## set character vector
  tmp_coord <- character(length(settings_sample))

  ## search for pattern match ... why?
  ## because otherwise the dataset becomes inconsistent
  pattern <- "_x:[0-9].*\\|y:[0-9].*"
  pattern_match <- grepl(
    pattern = pattern,
    x = settings_sample, perl = TRUE)

  ## extract coordinates
  tmp_coord[pattern_match] <- regexpr(
    pattern = pattern,
    text = settings_sample[pattern_match ], perl = TRUE) |>
    regmatches(x = settings_sample[pattern_match], m = _)

  ## extract x and y
  coord_split <- strsplit(tmp_coord, split = "|y:", fixed = TRUE)

  ## assign values
  coord <- vapply(coord_split, function(x) {
    if(length(x) == 0)
      return(c(x = NA_real_, y = NA_real_))

    c(x = as.numeric(strsplit(x, "_x:", fixed = TRUE)[[1]][[2]]),
      y = as.numeric(x[2]))},
    numeric(2)) |> t()

  ## if NA, assign index
  if (anyNA(coord[, 1])) coord[, 1] <- 0
  if (anyNA(coord[, 2])) coord[, 2] <- 1:nrow(coord)

  return(coord)
}

#' @title Fitting and Deconvolution of OSL Lifetime Components
#'
#' @details
#' The function intends to provide an easy access to pulsed optically stimulated luminescence (POSL) data,
#' in order determine signal lifetimes. The fitting is currently optimised to work with the off-time flank of POSL measurements
#' only. For the signal deconvolution, a differential evolution optimisation is combined with nonlinear least-square fitting
#' following the approach by Bluszcz & Adamiec (2006).
#'
#' **Component deconvolution algorithm**
#'
#' The component deconvolution consists of two steps:
#'
#' (1) Adaptation phase
#'
#' In the adaptation phase the function tries to figure out the optimal and statistically justified
#' number of signal components following roughly the approach suggested by Bluszcz & Adamiec (2006). In
#' contrast to their work, for the optimisation by differential evolution here the package `'DEoptim'` is used.
#'
#' The function to be optimized has the form:
#'
#' \deqn{\chi^2 = \sum(w * (n_i/c - \sum(A_i * \exp(-x/(tau_i + t_p))))^2)}
#'
#' with \eqn{w = 1} for unweighted regression analysis (`method_control = list(weights = FALSE)`) or
#' \eqn{w = c^2/n_i} for weighted regression analysis. The default values is `TRUE`.
#'
#' \deqn{F = (\Delta\chi^2 / 2) / (\chi^2/(N - 2*m - 2))}
#'
#' (2) Final fitting
#'
#' **`method_control`**
#'
#' \tabular{lll}{
#' **Parameter** \tab **Type** \tab **Description**\cr
#' `p` \tab [numeric] \tab controls the probability for the F statistic reference values. For a significance level of 5 % a value of 0.95 (the default) should be added, for 1 %, a value of 0.99 is sufficient: 1 > p > 0 (cf. [stats::FDist])\cr
#' `seed` \tab [numeric] \tab set the seed for the random number generator, provide a value here to get reproducible results \cr
#' `DEoptim.trace` \tab [logical] \tab enable/disable the tracing of the differential evolution (cf. [DEoptim::DEoptim.control]) \cr
#' `DEoptim.itermax` \tab [logical] \tab control the number of the allowed generations (cf. [DEoptim::DEoptim.control]) \cr
#' `weights` \tab [logical] \tab enable/disable the weighting for the start parameter estimation and fitting (see equations above).
#' The default values is `TRUE` \cr
#' `nlsLM.trace` \tab [logical] \tab enable/disable trace mode for the nls fitting ([minpack.lm::nlsLM]), can be used to identify convergence problems, default is `FALSE` \cr
#' `nlsLM.upper` \tab [logical] \tab enable/disable upper parameter boundary, default is `TRUE` \cr
#' `nlsLM.lower` \tab [logical] \tab enable/disable lower parameter boundary, default is `TRUE`
#' }
#'
#' @param object [Luminescence::RLum.Data.Curve-class], [Luminescence::RLum.Analysis-class], [data.frame] or [matrix] (**required**):
#' Input object containing the data to be analysed. All objects can be provided also as list for an automated
#' processing. Please note: `NA` values are automatically removed and the dataset should comprise at least 5 data points (possibly more if `n.components` is
#' set to a value greater than 1)
#'
#' @param tp [numeric] (*with default*): option to account for the stimulation pulse width. For off-time measurements
#' the default value is 0. `tp` has the same unit as the measurement data, e.g., µs. Please set this parameter
#' carefully, if it all, otherwise you may heavily bias your fit results.
#'
#' @param signal_range [numeric] (*optional*):
#' set a channel range, either by specifying both extremes, e.g.
#' `signal_range = c(2, 100)` considers only channels 2 to 100, or only the
#' start of the range, e.g. `signal_range = 2` considers only channels from
#' channel 2 onwards. If `NULL` (default) all channels are used.
#'
#' @param n.components [numeric] (*optional*): Fix the number of components. If set the algorithm will try
#' to fit the number of predefined components. If nothing is set, the algorithm will try to find the best number
#' of components.
#'
#' @param method_control [list] (*optional*): Named to allow a more fine control of the fitting process. See details
#' for allowed options.
#'
#' @param plot [logical] (*with default*): Enable/disable the plot output.
#'
#' @param plot_simple [logical] (*with default*): Enable/disable the reduced
#' plot output. If `TRUE`, no
#' residual plot is shown, however, plot output can be combined using the standard R layout options,
#' such as `par(mfrow = c(2,2))`.
#'
#' @param verbose [logical] (*with default*): Enable/disable output to the
#' terminal.
#'
#' @param ... parameters passed to [plot.default] to control the plot output, supported are:
#' `main`, `xlab`, `ylab`, `log`, `xlim`, `ylim`, `col`, `lty`, `legend.pos`, `legend.text`. If the input
#' object is of type [Luminescence::RLum.Analysis-class] this arguments can be provided as a [list].
#'
#' @return
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$data` \tab `matrix` \tab the final fit matrix \cr
#'  `$start_matrix` \tab `matrix` \tab the start matrix used for the fitting \cr
#'  `$total_counts` \tab `integer` \tab Photon count sum \cr
#'  `$fit` \tab `nls` \tab the fit object returned by [minpack.lm::nls.lm] \cr
#' }
#'
#'**slot:** **`@info`**
#'
#' The original function call
#'
#' ------------------------\cr
#' `[ TERMINAL OUTPUT ]`\cr
#' ------------------------\cr
#'
#' Terminal output is only shown of the argument `verbose = TRUE`.
#'
#' *(1) Start parameter and component adaption*\cr
#' Trave of the parameter adaptation process
#'
#' *(2) Fitting results (sorted by ascending tau)*\cr
#' The fitting results sorted by ascending tau value. Please note
#' that if you access the `nls` fitting object, the values are not sorted.
#'
#' *(3) Further information*\cr
#' - The photon count sum
#' - Durbin-Watson residual statistic to assess whether the residuals are correlated, ideally
#' the residuals should be not correlated at all. Rough measures are: \cr
#' D = 0: the residuals are systematically correlated \cr
#' D = 2: the residuals are randomly distributed \cr
#' D = 4: the residuals are systematically anti-correlated\cr
#'
#' You should be suspicious if D differs largely from 2.
#'
#'
#' ------------------------\cr
#' `[ PLOT OUTPUT ]`\cr
#' ------------------------\cr
#'
#' A plot showing the original data and the fit so far possible. The lower plot shows the
#' residuals of the fit.
#'
#' @section Function version: 0.1.6
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Christoph Schmidt, University of Bayreuth (Germany)
#'
#' @seealso [minpack.lm::nls.lm], [DEoptim::DEoptim]
#'
#' @references
#' Bluszcz, A., Adamiec, G., 2006. Application of differential evolution to fitting OSL decay curves.
#' Radiation Measurements 41, 886-891. \doi{10.1016/j.radmeas.2006.05.016}\cr
#'
#' Durbin, J., Watson, G.S., 1950. Testing for Serial Correlation in Least Squares Regression: I.
#' Biometrika 37, 409-21. doi:10.2307/2332391
#'
#' **Further reading**
#'
#' Hughes, I., Hase, T., 2010. Measurements and Their Uncertainties. Oxford University Press.
#'
#' Storn, R., Price, K., 1997. Differential Evolution –
#' A Simple and Efficient Heuristic for Global Optimization over Continuous Spaces.
#' Journal of Global Optimization 11, 341–359.
#'
#'@examples
#'
#'##load example data
#'data(ExampleData.TR_OSL, envir = environment())
#'
#'##fit lifetimes (short run)
#'fit_OSLLifeTimes(
#'  object = ExampleData.TR_OSL,
#'  n.components = 1)
#'
#'##long example
#'\dontrun{
#'fit_OSLLifeTimes(
#' object = ExampleData.TR_OSL)
#' }
#'
#'@export
fit_OSLLifeTimes <- function(
  object,
  tp = 0,
  signal_range = NULL,
  n.components = NULL,
  method_control = list(),
  plot = TRUE,
  plot_simple = FALSE,
  verbose = TRUE,
  ...
) {
  .set_function_name("fit_OSLLifeTimes")
  on.exit(.unset_function_name(), add = TRUE)

  ## Self-call --------------------------------------------------------------
  if (inherits(object, c("list", "RLum.Analysis"))) {
    .validate_not_empty(object)

  ##allow RLum.Analysis objects
  if (all(vapply(object, inherits, "RLum.Analysis", FUN.VALUE = logical(1)))) {
    object <- lapply(object, function(x){x@records})
    object <- .unlist_RLum(object)
  }

  ## expand input arguments
  rep.length <- length(object)

  if (!is.null(n.components))
    n.components <- .listify(n.components, rep.length)
  tp <- .listify(tp, rep.length)

  ## names of extra arguments
  arg_names <- ...names()

  ##pretreat some of the ... settings to avoid
  ## expand all arguments
  arg_list <- NULL
  if(!is.null(arg_names)){
    arg_list <- lapply(arg_names , function(x){
      .listify(list(...)[[x]], length(object))
    })

    ## make sure we organise this list (not nice but it works)
    arg_list <- lapply(1:length(object), function(x){
      args <- lapply(arg_list, function(y) y[[x]])
      names(args) <- arg_names
      args
    })
  }

  ##run function
  temp_results <- lapply(1:length(object), function(x){
    temp <- try(do.call(what = fit_OSLLifeTimes,
        c(list(
         object = object[[x]],
         tp = tp[[x]],
         signal_range = signal_range,
         n.components = n.components[[x]],
         method_control = method_control,
         plot = plot,
         plot_simple = plot_simple,
         verbose = verbose
         ),
         arg_list[[x]])
     ), outFile = stdout()) # redirect error messages so they can be silenced

     if(inherits(temp, "try-error")){
       return(NULL)
     }
    return(temp)
  })

  ##combine results and return
  results <- merge_RLum(temp_results)

  if(!is.null(results))
    results@originator <- "fit_OSLLifeTimes"

  ##return
  return(results)
}

  ## Integrity checks -------------------------------------------------------

  is.valid <- .validate_class(object,
                              c("RLum.Data.Curve", "data.frame", "matrix"),
                              extra = "a 'list' of such objects",
                              throw.error = FALSE)
  if (!is.valid)
    return(NULL)

  .validate_not_empty(object)
  if(inherits(object, "RLum.Data.Curve")){
   if(!grepl(pattern = "POSL", x = object@recordType, fixed = TRUE))
     .throw_error("recordType '", object@recordType,
                  "' not supported for input object")

    object <- as.data.frame(object@data)

  } else if(inherits(object, "matrix")){
    object <- as.data.frame(object)
  }

  if (ncol(object) < 2) {
    .throw_error("'object' should have at least two columns")
  }
  df <- object[, 1:2]

  ##remove NA values, whatever it is worth for
  if (anyNA(df)) {
    df <- na.exclude(df)
    .throw_warning("NA values detected and removed from dataset")
  }

  ##rename columns for data.frame
  colnames(df) <- c("x","y")

  #check for 0 data in dataset ... we opt for hard stop
  if(any(df[[2]] == 0)){
    .throw_warning("The dataset contains 0, a value of 0.1 ",
                   "has been added to your count values")
    df[[2]] <- df[[2]] + 0.1
  }

  ##save original data for later
  df_raw <- df

  ##signal_range
  if(!is.null(signal_range)){
    .validate_class(signal_range, c("integer", "numeric"), length = 1:2)
    if (anyNA(signal_range)) {
      .throw_warning("'signal_range' contains missing values, removed")
      signal_range <- signal_range[!is.na(signal_range)]
    }
    .validate_not_empty(signal_range)

    ## format the extremes of the signal range
    reset_msg <- function(sr) {
      sprintf("'signal_range' reset to c(%d, %d)", sr[1], sr[2])
    }

    ##check lengths
    if(length(signal_range) == 1)
      signal_range <- c(signal_range, nrow(df))

    if (any(signal_range < 1)) {
      signal_range <- pmax(signal_range, 1)
      .throw_warning("'signal_range' accepts only positive values, ",
                     reset_msg(signal_range))
    }

    if(signal_range[2] > nrow(df)){
      signal_range[2] <- nrow(df)
      .throw_warning("The last element of 'signal_range' exceeds the number ",
                     "of channels, ", reset_msg(signal_range))
    }

    if(signal_range[1] > signal_range[2]){
      signal_range <- c(1, nrow(df))
      .throw_warning("The first element of 'signal_range' exceeds the last ",
                     "element, ", reset_msg(signal_range))
    }

    ##set range
    df <- df[signal_range[1]:signal_range[2],]
  }

  ## number of components requested
  m <- max(.validate_positive_scalar(n.components, int = TRUE, null.ok = TRUE), 1)

  ## ensure that we have a minimum of data points available: the minimum
  ## is computed so that the degrees of freedom for the F distribution is
  ## positive (see `qf()` in the `while` loop further down at (B))
  min.num.signals <- 2 * m + 2 + 1
  if (nrow(df) < min.num.signals) {
    .throw_message("For ", m, " components the dataset must have at least ",
                   min.num.signals, " signal points, NULL returned")
    return(NULL)
  }

  .validate_logical_scalar(plot)
  .validate_logical_scalar(plot_simple)
  .validate_logical_scalar(verbose)

# Fitting -------------------------------------------------------------------------------------

  ##(0) CONTROL +++++++++++++++++++++++++++++++++++++++++++++++++++++
  method_control_setting <- list(
    p = 0.95,
    seed = NULL,
    DEoptim.trace = FALSE,
    DEoptim.itermax = 1000,
    weights = TRUE,
    nlsLM.trace = FALSE,
    nlsLM.upper = TRUE,
    nlsLM.lower = TRUE
  )

  ##udpate list if the user did something
  method_control_setting <- modifyList(x = method_control_setting, val = method_control)

  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ##(A) SETTINGS
  ##
  ##
  ##(1) >> set fitting function for minpack.lm
  x <- 0 #cheat R check routine
  fit_formula <- function(n.components, tp) {
    comps <- 1:n.components
    stats::reformulate(paste0("A.", comps, " * exp(-x / (tau.", comps,
                              " + ", tp, "))", collapse = " + "),
                       response = "y")
  }
  ##
  ##
  ##(2) create formula for differential evolution run
  fn_constructor <- function(m) {
    ##generate term
    idx <- seq(1, 2 * m, by = 2)
    term <- paste0("(x[", idx, "] * exp(-t / (x[", idx + 1, "] + tp)))",
                   collapse = " + ")

    ##set weight (should be given as character)
    w <- if (method_control_setting$weights) "c^2 / n" else "1"

    ##combine
    term <- paste0("sum(",w," * ((n/c) - (",term,"))^2)")

    ##parse ... if we do it here, we boost the speed of the evaluation
    parse(text = eval(term))
  }
  ##
  ##
  ##(3) initialise objects
  chi_squared <- c(NA, NA)
  F <- c(Inf, Inf)
  start <- NULL

  ##
  ##
  ##(4) set seed
  if(!is.null(method_control_setting$seed))
    set.seed(method_control_setting$seed)

  ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ##(B) RUN DIFFERENTIAL EVOLUTION TO DETERMINE NUMBER OF COMPONENTS
  ##prevent collateral damage, so we want a data.frame that has at least 10 rows
  if(verbose){
      cat("\n[fit_OSLLifeTime()]\n")
      cat("\n(1) Start parameter and component adapation")
      cat("\n---------------------(start adaption)---------------------\n")
    }

    while(!is.na(suppressWarnings(stats::qf(method_control_setting$p, df1 = 2, df2 = length(df[[2]]) - 2 * m - 2))) && (
          F[2] > stats::qf(method_control_setting$p, df1 = 2, df2 = length(df[[2]]) - 2 * m - 2) & F[1] >= F[2])){

      ##set F
      F[1] <- F[2]

      ##construct formula outside of the loop; needs to be done here, otherwise the performance
      ##goes down
      formula_string <- fn_constructor(m)

      ##set fn
      set_tp <- tp
      set_c <- diff(c(0,df[[1]]))
      set_t <- df[[1]]
      set_n <- df[[2]]

      ##set function
      ##Personal reminder:
      ##Why this function is not written in C++ ... because it adds basically nothing
      ##in terms of speed ~ 10 µs faster, but needed to be compiled and thus cannot be changed
      ##directly in the code
      fn <- function(x, tp = set_tp, n = set_n, c = set_c, t = set_t, term = formula_string){
         eval(formula_string)
      }

      ##set start parameters
      if(!is.null(start))
        start_parameters <- start$optim$bestmem

      ##run differential evolution
      start <- try(DEoptim::DEoptim(
        fn = fn,
        lower = rep(0, 2 * m),
        upper = rep(c(10 * sum(df[[2]]), 10000), m),
        control = DEoptim::DEoptim.control(
           trace = method_control_setting$DEoptim.trace,
           itermax = method_control_setting$DEoptim.itermax,
           c = 0.5,
           strategy = 2,
           parallelType = 0 #Does it make sense to use parallel processing here: no, it does not scale well
         )
      ), silent = TRUE)

      if (inherits(start, "try-error")) {
        .throw_error("Failed to optimize the function, check the input data")
      }

      ##set chi^2 value and calculate F for the 2nd run
      chi_squared[2] <- start$optim$bestval
      if(!is.na(chi_squared[1])){
        F[2] <- (abs(diff(chi_squared))/2) /
          (chi_squared[2]/(nrow(df) - 2 * m  - 2))
      }

      ##terminal feedback
      if(verbose){
        cat("\n>> + adaption for",m, "comp.", ": ", round(F[2],2), "(calc.) <> ",
            round(stats::qf(method_control_setting$p, df1 = 2, df2 = length(df[[2]]) - 2 * m - 2), 2), "(ref.)")

        if(F[2] > stats::qf(method_control_setting$p, df1 = 2, df2 = length(df[[2]]) - 2 * m - 2) & F[1] >= F[2]){
          cat(" >> [add comp.]")

        }else{
          cat(" >> [stop]\n")
          cat("----------------------(end adaption)----------------------\n\n")
        }
      }

      ##break here if n.components was set others than NULL, in such case we force the number
      if(!is.null(n.components)){
        if(verbose){
         cat(" >> [forced stop]\n")
         cat("----------------------(end adaption)----------------------\n\n")
        }

        start_parameters <- start$optim$bestmem
        break()
      }

      ##update objects
      chi_squared[1] <- chi_squared[2]
      m <- m + 1
    }

    ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ##(C) RUN LM-FITTING
    ##
    ##reduce m by 2, why 2?
    ##  - the last component violated the F statistic, so was obviously not the best call
    ##  - the loop adds every time another component
    if(is.null(n.components)){
      ## this covers the extreme case that the process stops after the first run
      if(m == 2){
        start_parameters <- start$optim$bestmem
      }
      m <- max(m - 2, 1)
    }

    A <- start_parameters[seq(1,length(start_parameters), by = 2)]
    tau <- start_parameters[seq(2,length(start_parameters), by = 2)]
    names(A) <- paste0("A.", 1:(m))
    names(tau) <- paste0("tau.", 1:(m))

    ##create start_matrix
    start_matrix <- matrix(data = c(A,tau), ncol = 2)
    colnames(start_matrix) <- c("A", "tau")
    rownames(start_matrix) <- paste0("Comp.", 1:(m))

    ##add terminal feedback
    if(verbose){
      cat("\n>> Applied component matrix\n")
      print(start_matrix)
      cat("\n\n")
    }

    ##run fitting using the Levenberg-Marquardt algorithm
    fit <- try(minpack.lm::nlsLM(
      formula = fit_formula(n.components = m, tp = tp),
      data = df,
      start = c(A, tau),
      upper = if(method_control_setting$nlsLM.upper){
        c(rep(sum(df[[2]]), length(A)), rep(Inf,length(tau)))
       }else{
         NULL
       },
      lower = if(method_control_setting$nlsLM.lower){
         rep(0, 2 * length(A))
       }else{
         NULL
       },
      na.action = "na.exclude",
      weights = if(method_control_setting$weights){
        set_c^2/df[,2]
      }else{
        rep(1,nrow(df))
      },
      trace = method_control_setting$nlsLM.trace,
      control = minpack.lm::nls.lm.control(maxiter = 500)
    ), silent = TRUE)

# Post-processing -----------------------------------------------------------------------------

  A <- NA
  tau <- NA
  summary_matrix <- NA
  D <- NA
  if (!inherits(fit, 'try-error')) {
    ##extract coefficients
    A <- coef(fit)[1:(m)]
    tau <- coef(fit)[(m + 1):(2 * m)]

    ##order coef
    o <- order(tau)
    tau <- tau[o]
    A <- A[o]

    ##summary matrix
    summary_matrix <- summary(fit)$coefficients

    ##return warning if one parameter is negative, this can happen if the user let the boundaries
    ##free float
    if(any(summary_matrix[,1]<0))
      .throw_warning("At least one parameter is negative, ",
                     "please check carefully your results")

    ##order matrix by tau, but keep the rownames
    temp_rownames <- rownames(summary_matrix)
    summary_matrix <- summary_matrix[c(o,o + length(A)),]
    rownames(summary_matrix) <- temp_rownames
    rm(temp_rownames)

    ##calculate Durbin-Watson statistic
    R <- residuals(fit)
    D <- round(sum((R - c(0,R[-length(R)]))^2) / sum(R^2),2)
    rm(R)
  } else {
    .throw_warning("Fitting failed: ", attr(fit, "condition")$message)
    m <- 1
  }

# Terminal output -----------------------------------------------------------------------------
if(verbose){

  if (!inherits(fit, 'try-error')) {
    cat("(2) Fitting results (sorted by ascending tau)\n")
    cat("-------------------------------------------------------------------------\n")
    print(summary_matrix)
    cat("-------------------------------------------------------------------------\n")

  }else{
    .throw_message("The fitting was not successful, consider trying again")
  }

  cat("\n(3) Further information\n")
  cat("-------------------------------------------------------------------------\n")
  cat("Photon count sum: ", sum(df[[2]]),"\n")
  cat("Durbin-Watson residual statistic: ", D,"")

  string <- NA
  if(!is.na(D)){
    string <- c("\u005b",rep(" ",(D * 10)/4),"\u003c\u003e",rep(" ",10 - (D * 10)/4),"\u005d\n")
  }
  cat(paste(string, collapse = ""))
  cat("\n")
}

# Plotting ------------------------------------------------------------------------------------
if(plot) {

  ##set plot default settings
  plot_settings <- list(
    main = "OSL Lifetimes",
    xlab = "Time [a.u.]",
    ylab = "POSL [a.u.]",
    log = "",
    xlim = c(0,max(df_raw[[1]])),
    ylim = c(0,max(df_raw[[2]])),
    col = get("col", pos = .LuminescenceEnv)[-1],
    lty = rep(1, (m + 1)),
    legend.pos = "topright",
    legend.text = c("sum", paste0("comp. ", 1:m))
  )

    ##modify settings on request
    plot_settings <- modifyList(x = plot_settings, val = list(...))

    ##catch log scale
    if (is.list(plot_settings$log))
      plot_settings$log <- unlist(plot_settings$log)
    if (grepl(pattern = "x", plot_settings$log, fixed = TRUE) &&
        plot_settings$xlim[1] == 0) {
        plot_settings$xlim[1] <- max(min(df_raw[[1]]), 1e-4)
        .throw_warning("log-scale requires x-values > 0, set min xlim to ",
                       round(plot_settings$xlim[1], 4))
    }

    if (grepl(pattern = "y", plot_settings$log, fixed = TRUE) &&
        plot_settings$ylim[1] == 0) {
        plot_settings$ylim[1] <- max(min(df_raw[[2]]), 1e-04)
        .throw_warning("log-scale requires y-values > 0, set min ylim to ",
                       round(plot_settings$ylim[1], 4))
    }

  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ##plot if the fitting was a success
  if (!inherits(fit, 'try-error')) {

    if(!plot_simple){
      graphics::layout(matrix(c(1, 2), 2, 1), height = c(0.7, 0.3))
      par(mar = c(0, 4.5, 3, 2))
    }

    plot(NA,NA,
         xaxt = if(plot_simple) "s" else "n",
         xlab = if(plot_simple) plot_settings$xlab else "",
         ylab = plot_settings$ylab,
         ylim = plot_settings$ylim,
         xlim = plot_settings$xlim,
         log = plot_settings$log,
         main = plot_settings$main
         )

    ##add used points
    points(df, col = rgb(0,0,0,0.8))

    ##add not used points df_raw (this solution avoids overplotting)
    if(nrow(df) != nrow(df_raw))
      points(df_raw[!df_raw[[1]]%in%df[[1]],], col = "grey")

    ##+ add some curve
    lines(
      df$x,
      stats::fitted(fit),
      col = plot_settings$col[1],
      lwd = 1.3,
      lty = plot_settings$lty[1]
     )

    ##+ add components
    for(i in 1:m) {
      if (length(plot_settings$lty) < 2)
        plot_settings$lty <- rep(plot_settings$lty, 1 + m)

      if (length(plot_settings$col) < 2)
        plot_settings$col <- rep(plot_settings$col, 1 + m)

      curve(
        A[i] * exp(-x / (tau[i] + tp)),
        add = TRUE,
        col = plot_settings$col[i + 1],
        lty = plot_settings$lty[i + 1]
      )
    }

    ##+ add legend
    legend(
      plot_settings$legend.pos,
      legend = plot_settings$legend.text,
      lty = plot_settings$lty,
      col = plot_settings$col[c(1, 2:(m + 2))],
      bty = "n"
    )

    if(!plot_simple){
      par(mar = c(5, 4.5, 0, 2))
      plot(
        x = df[[1]],
        y = residuals(fit),
        xlab = plot_settings$xlab,
        type = "b",
        pch = 20,
        xlim = plot_settings$xlim,
        log = gsub("y", "", plot_settings$log),
        ylab = "Resid."
      )
    }

  }else{
      ## fitting failed
      plot(
        df,
        xlab = plot_settings$xlab,
        ylab = plot_settings$ylab,
        col = rgb(0, 0, 0, 0.8),
        main = plot_settings$main,
        xlim = plot_settings$xlim,
        ylim = plot_settings$ylim,
        log = plot_settings$log
      )
  }
}#if plot

# Return --------------------------------------------------------------------------------------

  ##create return object
  set_RLum(
      class = "RLum.Results",
      data = list(
        data = summary_matrix,
        start_matrix = start_matrix,
        total_counts = sum(df[[2]]),
        fit = fit
      ),
      info = list(
        call = sys.call()
      )
  )
}

#'@title Luminescence Emission Spectra Deconvolution
#'
#' @description
#' This function performs a luminescence spectra deconvolution on
#' [Luminescence::RLum.Data.Spectrum-class] and [matrix] objects on an **energy scale**.
#' The function is optimised for emission spectra typically obtained in the
#' context of TL, OSL and RF  measurements detected between 200 and 1000 nm.
#' The function is not designed to deconvolve TL curves (counts against
#' temperature; no wavelength scale). If you are interested in such analysis,
#' please check, e.g., package `'tgcd'`.
#'
#'@details
#'
#'**Used equation**
#'
#'The emission spectra (on an energy scale) can be best described as the sum of multiple
#'Gaussian components:
#'
#'\deqn{
#' y = \sum_i \frac{C_i}{\sigma_{i} \sqrt{2 \pi}} \exp(-1/2 ((x - \mu_{i})/\sigma_{i})^2)
#'}
#'
#'with the parameters \eqn{\sigma} (peak width) and \eqn{\mu} (peak centre) and \eqn{C}
#'(scaling factor).
#'
#'**Start parameter estimation and fitting algorithm**
#'
#'The spectrum deconvolution consists of the following steps:
#'
#'1. Peak finding \cr
#'2. Start parameter estimation \cr
#'3. Fitting via [minpack.lm::nls.lm]\cr
#'
#' The peak finding is based on an approach (re-)suggested by Petr Pikal in
#' November 2005 in the R-help mailing list (https://stat.ethz.ch/pipermail/r-help/2005-November/thread.html).
#'This goes back to even earlier discussion in 2001 based on Prof Brian Ripley's idea.
#'It smartly uses the functions [stats::embed] and [max.col] to identify peaks positions.
#'For the use in this context, the algorithm has been further modified to scale on the
#'input data resolution (cf. source code).\cr
#'
#'The start parameter estimation uses random sampling from a range of meaningful parameters
#'and repeats the fitting until 1000 successful fits have been produced or the set `max.runs` value
#'is exceeded.
#'
#'Currently the best fit is the one with the lowest number for squared residuals, but
#'other parameters are returned as well. If a series of curves needs to be analysed,
#'it is recommended to make few trial runs, then fix the number of components and
#'run  at least 10,000 iterations (parameter `method_control = list(max.runs = 10000)`).
#'
#'**Supported `method_control` settings**
#'
#'\tabular{llll}{
#' **Parameter** \tab **Type** \tab **Default** \tab **Description**\cr
#' `max.runs` \tab [integer] \tab `10000` \tab maximum allowed search iterations, if exceed
#' the searching stops \cr
#' `graining` \tab [numeric] \tab `15` \tab control over how coarse or fine the spectrum is split into search intervals for the peak finding algorithm \cr
#' `norm` \tab [logical] \tab `TRUE` \tab normalise data to the highest count value before fitting \cr
#' `export.plot.data` \tab [logical] \tab `FALSE` \tab enable/disable export
#'  of the values of all curves in the plot for each frame analysed \cr
#' `trace` \tab [logical] \tab `FALSE` \tab enable/disable the tracing of the minimisation routine
#'}
#'
#'@param object [Luminescence::RLum.Data.Spectrum-class], [matrix] (**required**): input
#'object. Please note that an energy spectrum is expected
#'
#' @param frame [integer] (*optional*): number of the frame to be analysed. If
#' `NULL`, all available frames are analysed.
#'
#'@param start_parameters [numeric] (*optional*): allows to provide own start parameters for a
#'semi-automated procedure. Parameters need to be provided in eV. Every value provided replaces a
#'value from the automated peak finding algorithm (in ascending order).
#'
#' @param n_components [integer] (*optional*): maximum number of components
#' desired: the number of component actually fitted may be smaller than this.
#' Can be combined with other parameters.
#'
#' @param input_scale [character] (*optional*): defines whether your x-values
#' are expressed as wavelength or energy values. Allowed values are `"wavelength"`,
#' `"energy"` or `NULL`, in which case the function tries to guess the input
#' automatically.
#'
#'@param sub_negative [numeric] (*with default*): substitute negative values in the input object
#'by the number provided here (default: `0`). Can be set to `NULL`, i.e. negative values are kept.
#'
#' @param method_control [list] (*optional*): options to control the fit method
#' and the output produced, see details.
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#'@param ... further arguments to be passed to control the plot output
#'(supported: `main`, `xlab`, `ylab`, `xlim`, `ylim`, `log`, `mtext`, `legend` (`TRUE` or `FALSE`),
#'`legend.text`, `legend.pos`)
#'
#'@return
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' \tabular{lll}{
#'  **Element** \tab **Type** \tab **Description**\cr
#'  `$data` \tab `matrix` \tab the final fit matrix \cr
#'  `$fit` \tab `nls` \tab the fit object returned by [minpack.lm::nls.lm] \cr
#'  `$fit_info` \tab `list` \tab a few additional parameters that can be used
#'   to assess the quality of the fit \cr
#'  `$df_plot` \tab `list` \tab values of all curves in the plot for each
#'   frame analysed (only if `method_control$export.plot.data = TRUE`)
#' }
#'
#'
#'**slot:** **`@info`**
#'
#' The original function call
#'
#' ---------------------------------\cr
#' `[ TERMINAL OUTPUT ]`   \cr
#' ---------------------------------\cr
#'
#' The terminal output provides brief information on the
#' deconvolution process and the obtained results.
#' Terminal output is only shown when `verbose = TRUE`.
#'
#' ---------------------------\cr
#' `[ PLOT OUTPUT ]`      \cr
#' ---------------------------\cr
#'
#' The function returns a plot showing the raw signal with the
#' detected components. If the fitting failed, a basic plot is returned
#' showing the raw data and indicating the peaks detected for the start
#' parameter estimation. The grey band in the residual plot indicates the
#' 10% deviation from 0 (means no residual).
#'
#' @section Function version: 0.1.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#'@seealso [Luminescence::RLum.Data.Spectrum-class], [Luminescence::RLum.Results-class], [Luminescence::plot_RLum],
#'[Luminescence::convert_Wavelength2Energy], [minpack.lm::nls.lm]
#'
#'@keywords datagen
#'
#'@examples
#'
#'##load example data
#'data(ExampleData.XSYG, envir = environment())
#'
#'##subtract background
#'TL.Spectrum@@data <- TL.Spectrum@@data[] - TL.Spectrum@@data[,15]
#'
#'results <- fit_EmissionSpectra(
#'  object = TL.Spectrum,
#'  frame = 5,
#'  method_control = list(max.runs = 10)
#' )
#'
#' ##deconvolution of a TL spectrum
#' \dontrun{
#'
#' ##load example data
#'
#' ##replace 0 values
#' results <- fit_EmissionSpectra(
#'  object = TL.Spectrum,
#'  frame = 5, main = "TL spectrum"
#' )
#'
#'}
#'
#'@export
fit_EmissionSpectra <- function(
  object,
  frame = NULL,
  n_components = NULL,
  start_parameters = NULL,
  sub_negative = 0,
  input_scale = NULL,
  method_control = list(),
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("fit_EmissionSpectra")
  on.exit(.unset_function_name(), add = TRUE)

  ##TODO: Find a way to get a significant number of components
  ## This function works only on a list of matrices, so what ever we do here, we have to
  ## create a list of data treat, frame controls the number of frames analysed

  .validate_class(object, c("RLum.Data.Spectrum", "matrix", "list"))
  .validate_class(frame, c("integer", "numeric"), null.ok = TRUE)
  .validate_positive_scalar(n_components, int = TRUE, null.ok = TRUE)
  input_scale <- .validate_args(input_scale, c("wavelength", "energy"),
                                null.ok = TRUE)
  .validate_class(method_control, "list")
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

  ##input RLum.Data.Spectrum ... make list either way
  if(inherits(object, "RLum.Data.Spectrum"))
    object <- list(object)

  ##stop, mixed input is not allowed
  if(inherits(object, "list") && length(unique(sapply(object, class))) != 1)
    .throw_error("List elements of different class detected")

  ##deal with RLum.Data.Spectrum lists
  if(inherits(object, "list") && all(sapply(object, class) == "RLum.Data.Spectrum")){
    temp <- lapply(object, function(o){
      if (length(o@data) < 2) {
        .throw_error("'object' contains no data")
      }

      ##get x-axis
      x <- as.numeric(rownames(o@data))
      rownames(o@data) <- NULL

      ##set frame
      if(is.null(frame)){
        frame <- 1:ncol(o@data)
      } else if (max(frame) > ncol(o@data)|| min(frame) < 1) {
          .throw_error("Invalid 'frame', allowed values range from 1 to ",
                       ncol(o@data))
      }

      ##get frame
      temp_frame <- lapply(frame, function(f) cbind(x, o@data[,f]))
      names(temp_frame) <- paste0("Frame: ", frame)
      return(temp_frame)
    })

    ##set object name
    names(temp) <- paste0("ALQ: ", 1:length(temp), " ")

    ##unlist, now we have what we want
    object <- unlist(temp, use.names = TRUE, recursive = FALSE)
    names(object) <- gsub(" .", names(object), replacement = " | ", fixed = TRUE)
    rm(temp)
  }

  ##handle a single matrix that may have different columns
  if(inherits(object, "matrix") && ncol(object) > 2){
    rownames(object) <- NULL

    ##set frame
    if(is.null(frame)){
      frame <- 1:(ncol(object) - 1)
    } else if(max(frame) > (ncol(object)-1) || min(frame) < 1) {
        .throw_error("Invalid 'frame', allowed values range from 1 to ",
                     ncol(object) - 1)
    }

    temp <- lapply(frame +1 , function(x) cbind(object[,1],object[,x]))
    names(temp) <- paste0("Frame: ",frame)
    object <- temp
    rm(temp)
  }

  ##now treat different lists, the aim is to have a list of 2-column matrices
  ##we have two types of lists,
  # Self-call -----------------------------------------------------------------------------------
  if(inherits(object, "list")){
    ##get argument list
    args_list <- list(...)

    ##recycle arguments
    if(!"mtext" %in% names(args_list)){
      mtext <- names(object)

    }else{
      mtext <- .listify(args_list$mtext, length(object))
      args_list$mtext <- NULL
    }

    ##run over the list
    results <- lapply(1:length(object), function(o){
      do.call(fit_EmissionSpectra, args = c(
        list(
          object = object[[o]],
          start_parameters = start_parameters,
          n_components = n_components,
          sub_negative = sub_negative,
          input_scale = input_scale,
          method_control = method_control,
          frame = frame,
          mtext = mtext[[o]]),
        verbose = verbose,
        plot = plot,
        args_list)
      )
    })

    ##merge output and return
    return(merge_RLum(results))
  }


  # Start main core -----------------------------------------------------------------------------

  if (ncol(object) < 2) {
    .throw_error("'object' should have at least two columns")
  }

  ##extract matrix for everything below
  m <- object[, 1:2, drop = FALSE]

  ##replace all negative values
  if(!is.null(sub_negative))
    m[m[,2] < 0,2] <- sub_negative

  ##output
  if(verbose){
    cat("\n[fit_EmissionSpectra()]\n")
    cat("\n>> Treating dataset >>",frame,"<<\n")
  }

  ##check the scale
  if(is.null(input_scale)){
    ##values above 30 are unlikely, means its likely that we have a wavelength scale
    if(max(m[,1]) > 30){
      if(verbose) cat(">> Wavelength scale detected ...\n")
      m <- convert_Wavelength2Energy(m, order = TRUE)
      if(verbose) cat(">> Wavelength to energy scale conversion ... \t[OK]\n")
    }

  }else if(input_scale == "wavelength"){
    m <- convert_Wavelength2Energy(m, order = TRUE)
    if(verbose) cat(">> Wavelength to energy scale conversion ... \t[OK]\n")
  }


  # Settings ------------------------------------------------------------------------------------
  ##create peak finding function ... this helps to get good start parameters
  ##https://grokbase.com/t/r/r-help/05bqza71c4/r-finding-peaks-in-a-simple-dataset-with-r
  ##https://stat.ethz.ch/pipermail/r-help/2005-November/thread.html
  ##author: Petr Pikal in 2004; with modifications by Sebastian Kreutzer
  .peaks <- function(x, span, size = nrow(m)) {
    z <- stats::embed(x, span)
    s <- span %/% sample(1:4, size = 1)
    ##the part `ceiling(...)` scales the entire algorithm
    v <- max.col(z, ties.method = "first") == ceiling(10^(3 - log10(nrow(m)))) + s
    result <- c(rep(FALSE, s), v)
    which(result[1:(length(result) - s)])
  }

  ##set fit function
  x <- 0 #cheat R check routine
  fit_formula <- function(n.components) {
    sigma <- paste0("sigma.",1:n.components)
    mu <- paste0("mu.",1:n.components)
    C <- paste0("C.",1:n.components)
    stats::as.formula(
      paste0("y ~ ",
             paste(C," * 1/(",sigma," * sqrt(2 * pi)) * exp(-0.5 * ((x - ",mu,")/",sigma,")^2)",
                   collapse = " + ")))
  }

# Fitting -------------------------------------------------------------------------------------
  ## method_control --------
  method_control <- modifyList(x = list(
    max.runs = 10000,
    graining = 15,
    norm = TRUE,
    export.plot.data = FALSE,
    trace = FALSE
  ), val = method_control)

  # set data.frame ------------------------------------------------------------------------------
  df <- data.frame(x = m[,1], y = m[,2]) ##normalise values, it is just easier

  if (method_control$norm[1] && max(m[, 2]) > 0)
    df[["y"]] <- df[["y"]]/max(m[,2]) ##normalise values, it is just easier

  ## check graining parameter
  if (method_control$graining >= nrow(m))
    .throw_error("method_control$graining cannot exceed the ",
                 "available channels (", nrow(m) ,")")

  ##initialise objects
  success_counter <- 0
  run <- 0
  fit <- list()
  mu <- C <- sigma <- NA
  R2 <- SSR <- SST <- R2adj <- NA

  ## (WHILE LOOP START) -------
  ##run iterations
  while(success_counter < 1000 && run < method_control$max.runs){
    ##try to find start parameters

    ##identify peaks
    id_peaks <- .peaks(m[,2], sample(method_control$graining[1]:(nrow(m) - 1), 1))

      ##make sure that we do not end up in an endless loop
      if(length(id_peaks) == 0){
        if (verbose) cat("\r>> Searching components ... \t\t\t[-]")
        run <- run + 1
        next()
      }

    ## set start parameters for fitting --------
    mu <- m[id_peaks,1]

    if(!is.null(start_parameters))
      mu <- c(sort(start_parameters), mu[-(1:length(start_parameters))])

    ## limit the number of components
    if(!is.null(n_components[1]))
      mu <- mu[seq(1,length(mu), length.out = n_components[1])]

    sigma <- rep(sample(0.01:10,1),length(mu))
    C <- rep(max(df[[2]])/2, length(mu))

    names(mu) <- paste0("mu.", 1:length(mu))
    names(sigma) <- paste0("sigma.", 1:length(mu))
    names(C) <- paste0("C.", 1:length(mu))

    ##run fitting using the Levenberg-Marquardt algorithm
    fit_try <- try(minpack.lm::nlsLM(
      formula = fit_formula(n.components = length(mu)),
      data = df,
      start = c(sigma, mu, C),
      trace = method_control$trace,
      lower = rep(0, 3 * length(mu)),
      upper = c(
        rep(1000, length(mu)),
        rep(max(df[[1]]), length(mu)),
        rep(max(df[[2]]), length(mu))),
      control = minpack.lm::nls.lm.control(maxiter = 500)
    ), silent = TRUE)

    ##handle output
    if (class(fit_try)[1] != "try-error") {
      success_counter <- success_counter + 1
      fit[[success_counter]] <- fit_try
      if (verbose) cat("\r>> Searching components ... \t\t\t[/]")
    } else if (verbose) {
      cat("\r>> Searching components ... \t\t\t[\\]")
    }

    ##update run counter
    run <- run + 1
  }


  ## handle the output
  if (verbose) {
    cat("\r>> Searching components ... \t\t\t",
        if (length(fit) > 0) "[OK]" else "[FAILED]")
  }

  ## store the values that will be used when plotting
  h <- 4.135667662e-15 # eV * s
  c <- 299792458e+09 # nm/s
  df_plot <- data.frame(ENERGY = m[, 1],
                        WAVELENGTH = h * c / m[, 1])

  ##Extract best fit values
  ##TODO ... should be improved, its works, but maybe there are better solutions
  if (length(fit) != 0) {
    ##obtain the fit with the best fit
    best_fit <- vapply(fit, function(x) sum(residuals(x) ^ 2), numeric(1))
    fit <- fit[[which.min(best_fit)]]
    df_plot$SUM <- predict(fit)

    ## more parameters
    SSR <- min(best_fit)
    SST <- sum((df[[2]] - mean(df[[2]]))^2)
    R2 <- 1 - SSR/SST
    R2adj <- ((1 - R2) * (nrow(df) - 1)) /
      (nrow(df) - length(coef(fit)) - 1)

  }else{
    fit <- NA
  }

  ## Extract values of components -------------------------------------------
  m_coef <- NA
  if (!is.na(fit[1]) && inherits(fit, "nls")) {
    ##extract values we need only
    m_coef <- summary(fit)$coefficients
    m_coef <- matrix(
      data = c(
        as.numeric(m_coef[grepl(pattern = "mu", x = rownames(m_coef), fixed = TRUE),1:2]),
        as.numeric(m_coef[grepl(pattern = "sigma", x = rownames(m_coef), fixed = TRUE),1:2]),
        as.numeric(m_coef[grepl(pattern = "C", x = rownames(m_coef), fixed = TRUE),1:2])
      ),
      ncol = 6
    )

    ##set colnames
    colnames(m_coef) <- c("mu", "SE(mu)", "sigma", "SE(sigma)", "C", "SE(C)")

    ##order by sigma
    m_coef <- m_coef[order(m_coef[,1]),, drop = FALSE]

    ##extract single values, we need this later
    mu <- m_coef[,"mu"]
    sigma <- m_coef[,"sigma"]
    C <- m_coef[,"C"]

    ## evaluate the components at the given points
    for (i in 1:length(mu)) {
      df_plot[[paste0("COMP_", i)]] <- C[i] * dnorm(df_plot$ENERGY, mu[i], sigma[i])
    }
  }

  # Terminal output -----------------------------------------------------------------------------
  if(verbose && !is.na(m_coef[1])){
    cat(paste0("\n\n>> Fitting results (",length(mu), " component model):\n"))
    cat("-------------------------------------------------------------------------\n")
    print(m_coef)
    cat("-------------------------------------------------------------------------")
    cat(paste0("\nSE: standard error | SSR: ", format(min(best_fit), scientific=TRUE, digits = 4),
               "| R^2: ", round(R2,3), " | R^2_adj: ", round(R2adj,4)))
    cat("\n(use the output in $fit for a more detailed analysis)\n\n")
  }

  ## Plotting ---------------------------------------------------------------
  if(plot){
    ##get colour values
    col <- get("col", pos = .LuminescenceEnv)[-1]
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    ##plot settings
    plot_settings <- modifyList(x = list(
      xlab = "Energy [eV]",
      ylab = "Luminescence [a.u.]",
      main = "Emission Spectrum Deconvolution",
      xlim = range(df[[1]]),
      ylim = range(df[[2]]),
      log = "",
      mtext = "",
      legend = TRUE,
      legend.pos = "topright",
      legend.text = c("sum", paste0("c",1:length(mu),": ", round(mu,2), " eV"))
    ), val = list(...))

    if (!is.na(fit[1]) && !inherits(fit, "try-error")) {
      graphics::layout(matrix(c(1, 2), 2, 1), height = c(0.7, 0.3))

    ##SCREEN 1 ----------
    par(mar = c(0, 4.5, 3, 2))
    plot(
      df,
      pch = 20,
      xlab = plot_settings$xlab,
      ylab = plot_settings$ylab,
      xlim = plot_settings$xlim,
      ylim = plot_settings$ylim,
      main = plot_settings$main,
      col = rgb(0, 0, 0, 0.6),
      xaxt = "n",
      yaxt = "n",
      log = plot_settings$log
    )

    ## add axis normalised
    axis(side = 2,
           at = axTicks(side = 2),
           labels = c(axTicks(2)))

    ## add axis with real count vales
    if(method_control$norm[1]) {
      axis(
        side = 2,
        axTicks(side = 2)[-1],
        labels = format(
          max(m[, 2]) * axTicks(side = 2)[-1],
          digit = 1,
          scientific = TRUE
        ),
        line = 0.8,
        cex.axis = 0.7,
        tick = FALSE
      )
    }

    ##plot sum curve
    lines(x = df_plot$ENERGY, y = df_plot$SUM, col = col[1], lwd = 1.5)

    ##add mtext
    mtext(side = 3, text = plot_settings$mtext)

    ##add components
    first.comp.idx <- grep("COMP_", colnames(df_plot))[1] - 1
    for(i in 1:length(mu)){
      lines(df_plot$ENERGY, df_plot[[first.comp.idx + i]], col = col[i + 1])
    }

    ##add legend
    if(plot_settings$legend){
      legend(
        plot_settings$legend.pos,
        legend = plot_settings$legend.text,
        lwd = 1,
        col = col[1:(length(mu) + 2)],
        bty = "n"
      )
    }

    ## SCREEN 2 -----
    par(mar = c(5, 4.5, 0, 2))
    plot(NA, NA,
      ylim = range(residuals(fit)),
      xlab = plot_settings$xlab,
      type = "b",
      pch = 20,
      yaxt = "n",
      xlim = plot_settings$xlim,
      ylab = "",
      xpd = NA,
      col = rgb(0, 0, 0, 0.6),
      log = ifelse(grepl(plot_settings$log[1], pattern = "x", fixed = TRUE), "x", "")
    )

    ## add one axis label
    axis(side = 2, at = 0, labels = 0)

    ## add ± 5 line
    polygon(x = c(df[[1]], rev(df[[1]])),
            y = c(df[[2]] * 1.1 - df[[2]], rev(df[[2]] * 0.9 - df[[2]])),
            border = FALSE, col = rgb(0.8,0.8,0.8))

    ## add points
    points(df[[1]],residuals(fit), pch = 20, col = rgb(0,0,0,0.3))

    ## add zero line
    abline(h = 0, lty = 2)

    ##add wavelength axis
    h <- 4.135667662e-15 #eV * s
    c <- 299792458e+09 #nm/s
    axis(
      side = 1,
      labels = paste0("(",round((h * c) / axTicks(side = 3), 0), " nm)"),
      at = axTicks(side = 3),
      cex.axis = 0.7,
      line = 0.8,
      tick = FALSE
    )

   } else { # the fit failed

    ##provide control plot
    plot(df, main = "fit_EmissionSpectra() - control plot")

    ##abline
    abline(v = mu, lty = 2)

    ##add information
    mtext(side = 3, text = "(dashed lines indicate identified peaks)")

    ##add components
    for(i in 1:length(mu)){
      curve(
        (C[i] * 1 / (sigma[i] * sqrt(2 * pi)) * exp(-0.5 * ((x - mu[i])/sigma[i])^2)),
        add = TRUE,
        col = i
      )
    }
   }
  }##if plot

  ## Output -----------------------------------------------------------------
  if (!method_control$export.plot.data) {
    df_plot <- NA
  }
  set_RLum(
    class = "RLum.Results",
    data = list(data = m_coef,
                fit = fit,
                fit_info = list(
                  SSR = SSR,
                  SST = SST,
                  R2 = R2,
                  R2adj = R2adj),
                df_plot = list(df_plot)
                ),
    info = list(call = sys.call())
  )
}

#' @title Function to create an Abanico Plot.
#'
#' @description A plot is produced which allows comprehensive presentation of data precision
#' and its dispersion around a central value as well as illustration of a
#' kernel density estimate, histogram and/or dot plot of the dose values.
#'
#' @details
#' The Abanico Plot is a combination of the classic Radial Plot
#' [Luminescence::plot_RadialPlot] and a kernel density estimate plot (e.g
#' [Luminescence::plot_KDE]. It allows straightforward visualisation of data precision,
#' error scatter around a user-defined central value and the combined
#' distribution of the values, on the actual scale of the measured data (e.g.
#' seconds, equivalent dose, years). The principle of the plot is shown in
#' Galbraith & Green (1990). The function authors are thankful for the
#' thought-provoking figure in this article.
#'
#' The semi circle (z-axis) of the classic Radial Plot is bent to a straight
#' line here, which actually is the basis for combining this polar (radial)
#' part of the plot with any other Cartesian visualisation method
#' (KDE, histogram, PDF and so on). Note that the plot allows displaying
#' two measures of distribution. One is the 2-sigma
#' bar, which illustrates the spread in value errors, and the other is the
#' polygon, which stretches over both parts of the Abanico Plot (polar and
#' Cartesian) and illustrates the actual spread in the values themselves.
#'
#' Since the 2-sigma-bar is a polygon, it can be (and is) filled with shaded
#' lines. To change density (lines per inch, default is 15) and angle (default
#' is 45 degrees) of the shading lines, specify these parameters. See
#' `?polygon()` for further help.
#'
#' The Abanico Plot supports other than the weighted mean as measure of
#' centrality. When it is obvious that the data
#' is not (log-)normally distributed, the mean (weighted or not) cannot be a
#' valid measure of centrality and hence central dose. Accordingly, the median
#' and the weighted median can be chosen as well to represent a proper measure
#' of centrality (e.g. `centrality = "median.weighted"`). Also
#' user-defined numeric values (e.g. from the central age model) can be used if
#' this appears appropriate.
#'
#' The proportion of the polar part and the cartesian part of the Abanico Plot
#' can be modified for display reasons (`plot.ratio = 0.75`). By default,
#' the polar part spreads over 75 % and leaves 25 % for the part that
#' shows the KDE graph.
#'
#'
#' A statistic summary, i.e. a collection of statistic measures of
#' centrality and dispersion (and further measures) can be added by specifying
#' one or more of the following keywords:
#'
#' - `"n"` (number of samples)
#' - `"mean"` (mean De value)
#' - `"median"` (median of the De values)
#' - `"sd.rel"` (relative standard deviation in percent)
#' - `"sd.abs"` (absolute standard deviation)
#' - `"se.rel"` (relative standard error)
#' - `"se.abs"` (absolute standard error)
#' - `"in.2s"` (percent of samples in 2-sigma range)
#' - `"kurtosis"` (kurtosis)
#' - `"skewness"` (skewness)
#'
#' **Note:** the input data for the statistic summary is sent to the function
#' [Luminescence::calc_Statistics] depending on the log-option for the z-scale. If
#' `"log.z = TRUE"`, the summary is based on the logarithms of the input
#' data. If `"log.z = FALSE"` the linearly scaled data is used.
#'
#' **Note:** [Luminescence::calc_Statistics] calculates these statistic
#' measures in three different ways: `unweighted`, `weighted` and
#' `MCM-based` (i.e., based on Monte Carlo Methods). By default, the
#' MCM-based version is used. If you wish to use another method, indicate this
#' with the appropriate keyword using the argument `summary.method`.
#'
#' The optional parameter `layout` allows more sophisticated ways to modify
#' the entire plot. Each element of the plot can be addressed and its properties
#' can be defined. This includes font type, size and decoration, colours and
#' sizes of all plot items. To infer the definition of a specific layout style
#' cf. [Luminescence::get_Layout] or type, e.g. for the layout type `"journal"`.
#' A layout type can be modified by the user by
#' assigning new values to the list object.
#'
#' It is possible for the z-scale to specify where ticks are to be drawn
#'  by using the parameter `at`, e.g. `at = seq(80, 200, 20)`, cf. function
#'  documentation of `axis`. Specifying tick positions manually overrides a
#' `zlim`-definition.
#'
#' @param data [data.frame] or [Luminescence::RLum.Results-class] object (**required**):
#' for `data.frame` two columns: De (`data[,1]`) and De error (`data[,2]`).
#'  To plot several data sets in one plot the data sets must be provided as
#'  `list`, e.g. `list(data.1, data.2)`.\cr
#' For some [Luminescence::RLum.Results-class] objects, one or more lines (and
#' corresponding labels) are drawn automatically:
#' - [Luminescence::calc_AverageDose] (ADM)
#' - [Luminescence::calc_CentralDose] (CDM)
#' - [Luminescence::calc_MaxDose] (MDM)
#' - [Luminescence::calc_MinDose] (MAM)
#' - [Luminescence::calc_FiniteMixture]
#' Alternative labels can be set via the `line.label` option. This behaviour
#' can be suppressed altogether by setting `line = NA`.
#'
#' @param na.rm [logical] (*with default*):
#' exclude `NA` values from the data set prior to any further operations.
#'
#' @param log.z [logical] (*with default*):
#' Option to display the z-axis in logarithmic scale. Default is `TRUE`.
#'
#' @param z.0 [character] or [numeric] (*with default*):
#' User-defined central value, used for centring of data. One out of `"mean"`,
#' `"mean.weighted"` and `"median"` or a numeric value (not its logarithm).
#' Default is `"mean.weighted"`.
#'
#' @param dispersion [character] (*with default*):
#' measure of dispersion, used for drawing the scatter polygon. One out of
#' - `"qr"` (quartile range, default),
#' - `"pnn"` (symmetric percentile range with `nn` the lower percentile, e.g.
#' `"p05"` indicating the range between 5 and 95 %, or `"p10"` indicating
#' the range between 10 and 90 %), or
#' - `"sd"` (standard deviation) and
#' - `"2sd"` (2 standard deviations),
#'
#' The default is `"qr"`. Note that `"sd"` and `"2sd"` are only meaningful in
#' combination with `"z.0 = 'mean'"` because the unweighted mean is used to
#' centre the polygon.
#'
#' @param plot.ratio [numeric] (*with default*):
#' Relative space, given to the radial versus the cartesian plot part,
#' default is 0.75.
#'
#' @param rotate [logical] (*with default*):
#' Option to turn the plot by 90 degrees.
#'
#' @param mtext [character] (*with default*):
#' additional text below the plot title.
#'
#' @param summary [character] (*with default*):
#' add statistic measures of centrality and dispersion to the plot.
#' Can be one or more of several keywords. See details for available keywords.
#' Results differ depending on the log-option for the z-scale (see details).
#'
#' @param summary.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`) for the
#' statistical summary. Alternatively, the keyword `"sub"` may be
#' specified to place the summary below the plot header. However, this latter
#' option in only possible if `mtext` is not used.
#'
#' @param summary.method [character] (*with default*):
#' keyword indicating the method used to calculate the statistic summary.
#' One of `"MCM"` (default), `"weighted"` or `"unweighted"`.
#' See [Luminescence::calc_Statistics] for details.
#'
#' @param legend [character] vector (*optional*):
#' legend content to be added to the plot.
#'
#' @param legend.pos [numeric] or [character] (*with default*):
#' optional position coordinates or keyword (e.g. `"topright"`)
#' for the legend to be plotted.
#'
#' @param stats [character]:
#' additional labels of statistically important values in the plot.
#' One or more out of the following:
#' - `"min"`,
#' - `"max"`,
#' - `"median"`.
#'
#' @param rug [logical] (*with default*):
#' Option to add a rug to the KDE part, to indicate the location of individual values.
#'
#' @param kde [logical] (*with default*):
#' Option to add a KDE plot to the dispersion part, default is `TRUE`.
#'
#' @param hist [logical] (*with default*):
#' Option to add a histogram to the dispersion part. Only meaningful when not
#' more than one data set is plotted.
#'
#' @param dots [logical] (*with default*):
#' Option to add a dot plot to the dispersion part. If number of dots exceeds
#' space in the dispersion part, a square indicates this.
#'
#' @param boxplot [logical] (*with default*):
#' Option to add a boxplot to the dispersion part, default is `FALSE`.
#'
#' @param y.axis [logical] (*with default*): Option to hide standard y-axis
#' labels and show 0 only.
#' Useful for data with small scatter. If you want to suppress the y-axis entirely
#' please use `yaxt == 'n'` (the standard [graphics::par] setting) instead.
#'
#' @param error.bars [logical] (*with default*):
#' Option to show De-errors as error bars on De-points. Useful in combination
#' with `y.axis = FALSE, bar.col = "none"`.
#'
#' @param bar [numeric] (*with default*):
#' option to add one or more dispersion bars (i.e., bar showing the 2-sigma range)
#' centred at the defined values. By default a bar is drawn according to `"z.0"`.
#' To omit the bar set `"bar = FALSE"`.
#'
#' @param bar.col [character] or [numeric] (*with default*):
#' colour of the dispersion bar. Default is `"grey60"`.
#'
#' @param polygon.col [character] or [numeric] (*with default*):
#' colour of the polygon showing the data scatter. Sometimes this
#' polygon may be omitted for clarity. To disable it use `FALSE` or
#' `polygon = FALSE`. Default is `"grey80"`.
#'
#' @param line [numeric] or [Luminescence::RLum.Results-class]:
#' numeric values of the additional lines to be added. This can be set to `NA`
#' to suppress the line added automatically by some `RLum.Results` objects.
#'
#' @param line.col [character] or [numeric]:
#' colour of the additional lines.
#'
#' @param line.lty [integer]:
#' line type of additional lines.
#'
#' @param line.label [character]:
#' labels for the additional lines.
#'
#' @param grid.col [character] or [numeric] (*with default*):
#' colour of the grid lines (originating at `[0,0]` and stretching to
#' the z-scale). To disable grid lines use `FALSE`. Default is `"grey"`.
#'
#' @param frame [numeric] (*with default*):
#' option to modify the plot frame type. Can be one out of
#' - `0` (no frame),
#' - `1` (frame originates at 0,0 and runs along min/max isochrons),
#' - `2` (frame embraces the 2-sigma bar),
#' - `3` (frame embraces the entire plot as a rectangle).
#'
#' Default is `1`.
#'
#' @param bw [character] (*with default*):
#' bin-width for KDE, choose a numeric value for manual setting.
#'
#' @param interactive [logical] (*with default*):
#' create an interactive abanico plot (requires the `'plotly'` package)
#'
#' @param ... Further plot arguments to pass (see [graphics::plot.default]).
#' Supported are: `main`, `sub`, `ylab`, `xlab`, `zlab`, `zlim`, `ylim`, `cex`,
#' `lty`, `lwd`, `pch`, `col`, `at`, `breaks`. `xlab` must be
#' a vector of length two, specifying the upper and lower x-axis labels.
#'
#' @return
#' Returns a plot object and, optionally, a list with plot calculus data.
#'
#' @section Function version: 0.1.21
#'
#' @author
#' Michael Dietze, GFZ Potsdam (Germany)\cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
#' Inspired by a plot introduced by Galbraith & Green (1990)
#'
#' @seealso [Luminescence::plot_RadialPlot], [Luminescence::plot_KDE],
#' [Luminescence::plot_Histogram], [Luminescence::plot_ViolinPlot]
#'
#' @references
#' Galbraith, R. & Green, P., 1990. Estimating the component ages
#' in a finite mixture. International Journal of Radiation Applications and
#' Instrumentation. Part D. Nuclear Tracks and Radiation Measurements 17 (3),
#' 197-206.
#'
#' Dietze, M., Kreutzer, S., Burow, C., Fuchs, M.C., Fischer, M., Schmidt, C., 2015.
#' The abanico plot: visualising chronometric data with individual standard errors.
#' Quaternary Geochronology. doi:10.1016/j.quageo.2015.09.003
#'
#' @examples
#'
#' ## load example data and recalculate to Gray
#' data(ExampleData.DeValues, envir = environment())
#' ExampleData.DeValues <- ExampleData.DeValues$CA1
#'
#' ## plot the example data straightforward
#' plot_AbanicoPlot(data = ExampleData.DeValues)
#'
#' ## now with linear z-scale
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  log.z = FALSE)
#'
#' ## now with output of the plot parameters
#' plot1 <- plot_AbanicoPlot(data = ExampleData.DeValues)
#' str(plot1)
#' plot1$zlim
#'
#' ## now with adjusted z-scale limits
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  zlim = c(10, 200))
#'
#' ## now with adjusted x-scale limits
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  xlim = c(0, 20))
#'
#' ## now with rug to indicate individual values in KDE part
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  rug = TRUE)
#'
#' ## now with a smaller bandwidth for the KDE plot
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  bw = 0.04)
#'
#' ## now with a histogram instead of the KDE plot
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  hist = TRUE,
#'                  kde = FALSE)
#'
#' ## now with a KDE plot and histogram with manual number of bins
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  hist = TRUE,
#'                  breaks = 20)
#'
#' ## now with a KDE plot and a dot plot
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  dots = TRUE)
#'
#' ## now with user-defined plot ratio
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  plot.ratio = 0.5)

#' ## now with user-defined central value
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  z.0 = 70)
#'
#' ## now with median as central value
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  z.0 = "median")
#'
#' ## now with the 17-83 percentile range as definition of scatter
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  z.0 = "median",
#'                  dispersion = "p17")
#'
#' ## now with user-defined green line for minimum age model
#' CAM <- calc_CentralDose(ExampleData.DeValues,
#'                         plot = FALSE)
#' plot_AbanicoPlot(data = CAM,
#'                  line.col = "darkseagreen")
#'
#' ## now create plot with legend, colour, different points and smaller scale
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  legend = "Sample 1",
#'                  col = "tomato4",
#'                  bar.col = "peachpuff",
#'                  pch = "R",
#'                  cex = 0.8)
#'
#' ## now without 2-sigma bar, polygon, grid lines and central value line
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  bar.col = FALSE,
#'                  polygon.col = FALSE,
#'                  grid.col = FALSE,
#'                  y.axis = FALSE,
#'                  lwd = 0)
#'
#' ## now with direct display of De errors, without 2-sigma bar
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  bar.col = FALSE,
#'                  ylab = "",
#'                  y.axis = FALSE,
#'                  error.bars = TRUE)
#'
#' ## now with user-defined axes labels
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  xlab = c("Data error (%)",
#'                           "Data precision"),
#'                  ylab = "Scatter",
#'                  zlab = "Equivalent dose [Gy]")
#'
#' ## now with minimum, maximum and median value indicated
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  stats = c("min", "max", "median"))
#'
#' ## now with a brief statistical summary as subheader
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  summary = c("n", "in.2s"))
#'
#' ## now with another statistical summary
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  summary = c("mean.weighted", "median"),
#'                  summary.pos = "topleft")
#'
#' ## now a plot with two 2-sigma bars for one data set
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  bar = c(30, 100))
#'
#' ## now the data set is split into sub-groups, one is manipulated
#' data.1 <- ExampleData.DeValues[1:30,]
#' data.2 <- ExampleData.DeValues[31:62,] * 1.3
#'
#' ## now a common dataset is created from the two subgroups
#' data.3 <- list(data.1, data.2)
#'
#' ## now the two data sets are plotted in one plot
#' plot_AbanicoPlot(data = data.3)
#'
#' ## now with some graphical modification
#' plot_AbanicoPlot(data = data.3,
#'                  z.0 = "median",
#'                  col = c("steelblue4", "orange4"),
#'                  bar.col = c("steelblue3", "orange3"),
#'                  polygon.col = c("steelblue1", "orange1"),
#'                  pch = c(2, 6),
#'                  angle = c(30, 50),
#'                  summary = c("n", "in.2s", "median"))
#'
#' ## create Abanico plot with predefined layout definition
#' plot_AbanicoPlot(data = ExampleData.DeValues,
#'                  layout = "journal")
#'
#' ## now with predefined layout definition and further modifications
#' plot_AbanicoPlot(
#'  data = data.3,
#'  z.0 = "median",
#'  layout = "journal",
#'  col = c("steelblue4", "orange4"),
#'  bar.col = adjustcolor(c("steelblue3", "orange3"),
#'                          alpha.f = 0.5),
#'  polygon.col = c("steelblue3", "orange3"))
#'
#' ## for further information on layout definitions see documentation
#' ## of function get_Layout()
#'
#' ## now with manually added plot content
#' ## create empty plot with numeric output
#' AP <- plot_AbanicoPlot(data = ExampleData.DeValues,
#'                        pch = NA)
#'
#' ## identify data in 2 sigma range
#' in_2sigma <- AP$data[[1]]$data.in.2s
#'
#' ## restore function-internal plot parameters
#' par(AP$par)
#'
#' ## add points inside 2-sigma range
#' points(x = AP$data[[1]]$precision[in_2sigma],
#'        y = AP$data[[1]]$std.estimate.plot[in_2sigma],
#'        pch = 16)
#'
#' ## add points outside 2-sigma range
#' points(x = AP$data[[1]]$precision[!in_2sigma],
#'        y = AP$data[[1]]$std.estimate.plot[!in_2sigma],
#'        pch = 1)
#'
#' @export
plot_AbanicoPlot <- function(
  data,
  na.rm = TRUE,
  log.z = TRUE,
  z.0 = c("mean.weighted", "mean.weighted", "median"),
  dispersion = c("qr", "sd", "2sd"),
  plot.ratio = 0.75,
  rotate = FALSE,
  mtext = "",
  summary = c("n", "in.2s"),
  summary.pos = "sub",
  summary.method = c("MCM", "weighted", "unweighted"),
  legend = NULL,
  legend.pos = "topleft",
  stats = NULL,
  rug = FALSE,
  kde = TRUE,
  hist = FALSE,
  dots = FALSE,
  boxplot = FALSE,
  y.axis = TRUE,
  error.bars = FALSE,
  bar = NULL,
  bar.col = NULL,
  polygon.col = NULL,
  line = NULL,
  line.col = NULL,
  line.lty = NULL,
  line.label = NULL,
  grid.col = NULL,
  frame = 1,
  bw = "SJ",
  interactive = FALSE,
  ...
) {
  .set_function_name("plot_AbanicoPlot")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  ## Homogenise input data format
  if (!inherits(data, "list")) {
    data <- list(data)
  }

  ## whether the user has set the line argument or not
  line.is.null <- is.null(line)
  line.mtext <- NULL

  ## Check input data
  for (i in seq_along(data)) {
    .validate_class(data[[i]], c("data.frame", "RLum.Results"),
                    name = "All elements of 'data'")
    if (inherits(data[[i]], "RLum.Results")) {
      if (line.is.null &&
          .check_originator(data[[i]], c("calc_AverageDose", "calc_CentralDose",
                                         "calc_MaxDose", "calc_MinDose",
                                         "calc_FiniteMixture"))) {
        ## set lines automatically based on originator
        de <- data[[i]]$summary$de
        de.pm.err <- sprintf("%.2f \u00b1 %.2f", de, data[[i]]$summary$de_err)
        lab <- switch(data[[i]]@originator,
                      "calc_AverageDose" = "ADM",
                      "calc_CentralDose" = "CDM",
                      "calc_MaxDose" = "MDM",
                      "calc_MinDose" = "MAM",
                      "calc_FiniteMixture" = de.pm.err)
        line <- c(line, de)
        line.label <- c(line.label, lab)
        if (data[[i]]@originator == "calc_FiniteMixture")
          line.mtext <- c(line.mtext, paste("FMM:", length(de), "components"))
        else
          line.mtext <- c(line.mtext, paste0(lab, ": ", de.pm.err))
      }
      data[[i]] <- get_RLum(data[[i]], "data")
    }

      if (ncol(data[[i]]) < 2) {
        .throw_error("Data set ", i, " has fewer than 2 columns: data ",
                     "without errors cannot be displayed")
      }

      data[[i]] <- data[[i]][, 1:2]
  }
  if (!is.null(line.mtext) && mtext == "" && summary.pos != "sub")
    mtext <- line.mtext

  ## optionally, remove NA-values
  .validate_logical_scalar(na.rm)
  if (na.rm) {
    for(i in seq_along(data)) {
      n.NA <- sum(!stats::complete.cases(data[[i]]))
      if (n.NA > 0) {
        .throw_message("Data set ", i, ": ", n.NA, " NA value",
                       ifelse(n.NA > 1, "s", ""), " excluded", error = FALSE)
        data[[i]] <- na.exclude(data[[i]])
      }
    }
  }

  ##AFTER NA removal, we should check the data set carefully again ...
  ##(1)
  ##check if there is still data left in the entire set
  if(all(sapply(data, nrow) == 0)){
    .throw_message("'data' is empty, nothing plotted")
    return(NULL)
  }
  ##(2)
  ##check for sets with only 1 row or 0 rows at all
  if (any(sapply(data, nrow) < 2)) {
    ##select problematic sets and remove the entries from the list
    NArm.id <- which(sapply(data, nrow) <= 1)
    data[NArm.id] <- NULL
    .throw_warning("Data set ", toString(NArm.id),
                   " empty or consisting of only 1 row, removed")

    ##unfortunately, the data set might become now empty at all
    if(length(data) == 0){
      .throw_message("After removing invalid entries, nothing is plotted")
      return(NULL)
    }
  }

  ## check for zero-error values
  for(i in 1:length(data)) {
    if (any(data[[i]][, 2] == 0)) {
      data[[i]] <- data[[i]][data[[i]][,2] > 0,]
      if (nrow(data[[i]]) < 1){
        .throw_error("Data set contains only values with zero errors")
      }
      .throw_warning("Values with zero errors cannot be displayed and were removed")
    }
  }

  ## check for 0 values in dataset for log
  .validate_logical_scalar(log.z)
  if (log.z) {
    for(i in 1:length(data)) {
      if(any(data[[i]][[1]] == 0)) {
        .throw_warning("Found zero values in x-column of dataset ", i,
                       ", 'log.z' set to FALSE")
        log.z <- FALSE
      }
    }
  }

  ## plot.ratio must be numeric and positive
  .validate_positive_scalar(plot.ratio)
  .validate_logical_scalar(rotate)
  .validate_logical_scalar(rug)
  .validate_logical_scalar(kde)
  .validate_logical_scalar(hist)
  .validate_logical_scalar(dots)
  .validate_logical_scalar(boxplot)
  .validate_logical_scalar(error.bars)

  if (!is.numeric(z.0)) {
    .validate_class(z.0, "character")
    z.0 <- .validate_args(z.0, c("mean", "mean.weighted", "median"),
                          extra = "a numerical value")
  }

  ## the 'pnn' option need some special treatment
  main.choices <- c("qr", "sd", "2sd")
  extra.choice <-"a percentile of the form 'pnn' (eg. 'p05')"
  if (!any(grepl("^p[0-9][0-9]$", dispersion)))
    dispersion <- .validate_args(dispersion, main.choices, extra = extra.choice)
  .validate_length(dispersion, 1)

  valid.pos <- c("left", "center", "right", "topleft", "top", "topright",
                 "bottomleft", "bottom", "bottomright")
  .validate_class(summary, "character")
  if (is.numeric(summary.pos)) {
    .validate_length(summary.pos, 2)
  }
  else {
    summary.pos <- .validate_args(summary.pos, c("sub", valid.pos))
  }
  .validate_class(legend, "character", null.ok = TRUE)
  if (is.numeric(legend.pos)) {
    .validate_length(legend.pos, 2)
  } else {
    legend.pos <- .validate_args(legend.pos, valid.pos)
  }

  summary.method <- .validate_args(summary.method, c("MCM", "weighted", "unweighted"))
  frame <- .validate_args(frame, c(0, 1, 2, 3))

  ## check/set layout definitions
  layout <- get_Layout(layout = list(...)$layout %||% "default")

  if (is.null(bar))
    bar <- rep(TRUE, length(data))

  if (is.null(bar.col)) {
    bar.fill <- rep(rep_len(layout$abanico$colour$bar.fill, length(data)),
                    length(bar))
    bar.line <- rep(rep_len(layout$abanico$colour$bar.line, length(data)),
                    length(bar))
  } else {
    bar.fill <- bar.col
    bar.line <- NA
  }

  if (is.null(polygon.col)) {
    polygon.fill <- rep_len(layout$abanico$colour$poly.fill, length(data))
    polygon.line <- rep_len(layout$abanico$colour$poly.line, length(data))
  } else {
    polygon.fill <- polygon.col
    polygon.line <- NA
  }

  if (is.null(grid.col)) {
    grid.major <- layout$abanico$colour$grid.major
    grid.minor <- layout$abanico$colour$grid.minor
  } else {
    grid.major <- grid.col[1]
    grid.minor <- grid.col[min(length(grid.col), 2)]
  }

  ## create preliminary global data set
  De.global <- unlist(lapply(data, function(x) x[, 1]))

  ## additional arguments
  extraArgs <- list(...)
  breaks <- extraArgs$breaks %||% "Sturges"
  fun <- isTRUE(list(...)$fun)

  ## check/set bw-parameter
  for(i in 1:length(data)) {
    bw.test <- try(density(x = data[[i]][,1],
                           bw = bw),
                   silent = TRUE)
    if (inherits(bw.test, "try-error")) {
      bw <- "nrd0"
      .throw_warning("Option for 'bw' not valid, reset to 'nrd0'")
    }
  }

  ## check for negative values, stop function, but do not stop
  De.add <- 0
  if(min(De.global) < 0) {
    if("zlim" %in% names(extraArgs)) {
      De.add <- abs(extraArgs$zlim[1])
    } else {
      ## estimate delta De to add to all data
      De.add <-  min(10^ceiling(log10(abs(De.global))) * 10)

      ## optionally readjust delta De for extreme values
      if(De.add <= abs(min(De.global))) {
        De.add <- De.add * 10
      }
    }
  }

  ## optionally add correction dose to data set and adjust error
  if (log.z) {
    for(i in 1:length(data))
      data[[i]][,1] <- data[[i]][,1] + De.add

    De.global <- De.global + De.add
  }

  ## calculate and append statistical measures --------------------------------

  ## z-values and se based on log-option
  data <- lapply(data, function(x, De.add) {
    cbind(x,
          z = if (log.z) log(x[, 1]) else x[, 1],
          se = if (log.z) x[, 2] / (x[, 1] + De.add) else x[, 2])
  }, De.add = De.add)

  ## calculate initial data statistics
  stats.init <- lapply(data, function(x) calc_Statistics(data = x[, 3:4]))

  ## calculate central values
  if (z.0 %in% c("mean", "median")) {
    z.central <- lapply(1:length(data), function(x){
      rep(stats.init[[x]]$unweighted[[z.0]],
          length(data[[x]][,3]))})

  } else if(z.0 == "mean.weighted") {
    z.central <- lapply(1:length(data), function(x){
      rep(stats.init[[x]]$weighted$mean,
          length(data[[x]][,3]))})

  } else {
    ## z.0 is numeric
    z.central <- lapply(1:length(data), function(x){
      rep(ifelse(log.z,
                 log(z.0),
                 z.0),
          length(data[[x]][,3]))})
  }

  data <- lapply(1:length(data), function(x) {
    cbind(data[[x]], z.central[[x]])})
  rm(z.central)

  ## calculate precision and standard estimate
  data <- lapply(data, function(x) {
    cbind(x,
          precision = 1 / x[, 4],
          std.estimate = (x[, 3] - x[, 5]) / x[, 4],
          std.estimate.plot = NA)
  })

  ## append optional weights for KDE curve
  use.weights <- "weights" %in% names(extraArgs) && extraArgs$weights
  data <- lapply(data, function(x) {
    cbind(x,
          weights = if (use.weights) (1 / x[, 2]) / sum(1 / x[, 2]^2)
                    else 1 / nrow(x))
  })

  ## generate global data set
  data.global <- cbind(data[[1]], 1)
  colnames(data.global) <- rep("", 10)

  if(length(data) > 1) {
    for(i in 2:length(data)) {
      data.add <- cbind(data[[i]], i)
      colnames(data.add) <- rep("", 10)
      data.global <- rbind(data.global,
                           data.add)
    }
  }

  ## create column names
  colnames(data.global) <- c("De",
                             "error",
                             "z",
                             "se",
                             "z.central",
                             "precision",
                             "std.estimate",
                             "std.estimate.plot",
                             "weights",
                             "data set")

  ## calculate global data statistics
  stats.global <- calc_Statistics(data = data.global[,3:4])

  ## calculate global central value
  if (z.0 %in% c("mean", "median")) {
    z.central.global <- stats.global$unweighted[[z.0]]
  } else  if(z.0 == "mean.weighted") {
    z.central.global <- stats.global$weighted$mean
  } else if(is.numeric(z.0)) {
    z.central.global <- ifelse(log.z,
                               log(z.0),
                               z.0)
  }

  ## create column names
  for(i in 1:length(data)) {
    colnames(data[[i]]) <- c("De",
                             "error",
                             "z",
                             "se",
                             "z.central",
                             "precision",
                             "std.estimate",
                             "std.estimate.plot",
                             "weights")
  }

  ## re-calculate standardised estimate for plotting
  for(i in 1:length(data)) {
    data[[i]][,8] <- (data[[i]][,3] - z.central.global) / data[[i]][,4]
  }

  data.global.plot <- data[[1]][,8]
  if(length(data) > 1) {
    for(i in 2:length(data)) {
      data.global.plot <- c(data.global.plot, data[[i]][,8])
    }
  }
  data.global[,8] <- data.global.plot

  ## print message for too small scatter
  if(max(abs(1 / data.global[6])) < 0.02) {
    .throw_message("Small standardised estimate scatter, toggle off y.axis?",
                   error = FALSE)
  }

  ## read out additional arguments---------------------------------------------
  extraArgs <- list(...)

  main <- extraArgs$main %||% expression(D[e] * " " * "distribution")
  sub <- extraArgs$sub %||% ""

  xlab <- if ("xlab" %in% names(extraArgs)) {
            if (!length(extraArgs$xlab) %in% c(2, 3))
              .throw_error("'xlab' must have length 2")
            c(extraArgs$xlab[1:2], "Density")
          } else {
            c(if (log.z) "Relative standard error [%]" else "Standard error",
              "Precision",
              "Density")
          }
  ylab <- extraArgs$ylab %||% "Standardised estimate"
  zlab <- extraArgs$zlab %||% expression(D[e] * " " * "[Gy]")

  limits.z <- extraArgs$zlim
  if (!is.null(limits.z)) {
    .validate_class(limits.z, "numeric", name = "'zlim'")
    if (log.z && any(limits.z <= 0)) {
      .throw_error("'zlim' should only contain positive values when 'log.z = TRUE'")
    }
  } else {
    z.span <- (mean(data.global[,1]) * 0.5) / (sd(data.global[,1]) * 100)
    z.span <- ifelse(z.span > 1, 0.9, z.span)
    limits.z <- c((0.9 - z.span) * min(data.global[[1]]),
                  (1.1 + z.span) * max(data.global[[1]]))
  }

  limits.x <- extraArgs$xlim
  if (!is.null(limits.x)) {
    .validate_class(limits.x, "numeric", name = "'xlim'")
    if (limits.x[1] != 0) {
      .throw_warning("Lower x-axis limit was ", limits.x[1], ", reset to zero")
      limits.x[1] <- 0
    }
  } else {
    limits.x <- c(0, max(data.global[,6]) * 1.05)
  }

  limits.y <- extraArgs$ylim
  if (!is.null(limits.y)) {
    .validate_class(limits.y, "numeric", name = "'ylim'")
  } else {
    y.span <- (mean(data.global[,1]) * 10) / (sd(data.global[,1]) * 100)
    y.span <- ifelse(y.span > 1, 0.98, y.span)
    limits.y <- c(-(1 + y.span) * max(abs(data.global[,7])),
                  (1 + y.span) * max(abs(data.global[,7])))
  }

  cex <- extraArgs$cex %||% 1
  lty <- extraArgs$lty %||% rep(rep(2, length(data)), length(bar))
  lwd <- extraArgs$lwd %||% rep(rep(1, length(data)), length(bar))
  pch <- extraArgs$pch %||% rep(20, length(data))

  if("col" %in% names(extraArgs)) {
    bar.col <- extraArgs$col
    kde.line <- extraArgs$col
    kde.fill <- NA
    value.dot <- extraArgs$col
    value.bar <- extraArgs$col
    value.rug <- extraArgs$col
    summary.col <- extraArgs$col
    centrality.col <- extraArgs$col
  } else {
    bar.col <- layout$abanico$colour$bar.fill
    if(length(layout$abanico$colour$bar.fill) == 1) {
      bar.col <- 1:length(data)
    }

    kde.line <- layout$abanico$colour$kde.line
    if(length(layout$abanico$colour$kde.line) == 1) {
      kde.line <- 1:length(data)
    }

    kde.fill <- layout$abanico$colour$kde.fill
    if(length(layout$abanico$colour$kde.fill) == 1) {
      kde.fill <- rep(layout$abanico$colour$kde.fill, length(data))
    }

    value.dot <- layout$abanico$colour$value.dot
    if(length(layout$abanico$colour$value.dot) == 1) {
      value.dot <- 1:length(data)
    }

    value.bar <- layout$abanico$colour$value.bar
    if(length(layout$abanico$colour$value.bar) == 1) {
      value.bar <- 1:length(data)
    }

    value.rug <- layout$abanico$colour$value.rug
    if(length(layout$abanico$colour$value.rug) == 1) {
      value.rug <- 1:length(data)
    }

    summary.col <- layout$abanico$colour$summary
    if(length(layout$abanico$colour$summary) == 1) {
      summary.col <- 1:length(data)
    }

    if(length(layout$abanico$colour$centrality) == 1) {
      centrality.col <- rep(x = 1:length(data), times = length(bar))
    } else {
      centrality.col <- rep(x = layout$abanico$colour$centrality,
                            times = length(bar))
    }
  }

  ## update central line colour
  centrality.col <- rep(centrality.col, length(bar))

  ## define auxiliary plot parameters -----------------------------------------
  ## set space between z-axis and baseline of cartesian part
  lostintranslation <- 1.03
  if (!boxplot) {
    plot.ratio <- plot.ratio * 1.05
  }

  ## save the original plot parameters and restore them when exiting
  ## this must be done after all validations have completed, otherwise a
  ## warning may be generated (#1001) if any validation step fails.
  if (sum(par()$mfrow) == 2 && sum(par()$mfcol) == 2) {
    par.default <- .par_defaults()
  } else {
    ## this ensures that mfrow/mfcol are not reset when we want to draw
    ## several plots on one page
    par.default <- par(c("mar", "mai", "xpd", "cex"))
  }
  on.exit(par(par.default), add = TRUE)

  ## wrapper functions to deal with rotation
  plot.rot <- function(xlim, ylim, ...) {
    if (!rotate) plot(xlim = xlim, ylim = ylim, ...) else plot(xlim = ylim, ylim = xlim, ...)
  }
  polygon.rot <- function(x, y, ...) {
    if (!rotate) polygon(x, y, ...) else polygon(y, x, ...)
  }
  points.rot <- function(x, y, ...) {
    if (!rotate) points(x, y, ...) else points(y, x, ...)
  }
  lines.rot <- function(x, y, ...) {
    if (!rotate) lines(x, y, ...) else lines(y, x, ...)
  }
  text.rot <- function(x, y, ...) {
    if (!rotate) text(x, y, ...) else text(y, x, ...)
  }
  text_with_bg.rot <- function(x, y, ...) {
    if (!rotate) .text_with_bg(x, y, ...) else .text_with_bg(y, x, ...)
  }

  ## create empty plot to update plot parameters
  plot.rot(NA,
       xlim = c(limits.x[1], limits.x[2] * (1 / plot.ratio)),
       ylim = limits.y,
       main = "",
       sub = "",
       xlab = "",
       ylab = "",
       xaxs = "i",
       yaxs = "i",
       frame.plot = FALSE,
       axes = FALSE)

  ## calculate conversion factor for plot coordinates
  f <- 0

  ## calculate major and minor z-tick values
  if("at" %in% names(extraArgs)) {
    tick.values.major <- extraArgs$at
    tick.values.minor <- extraArgs$at
  } else {
    tick.values.major <- signif(pretty(limits.z, n = 5), 3)
    tick.values.minor <- signif(pretty(limits.z, n = 25), 3)
  }

  tick.values.major <- tick.values.major[
      between(tick.values.major, limits.z[1], limits.z[2]) &
      between(tick.values.major, min(tick.values.minor), max(tick.values.minor))]
  tick.values.minor <- tick.values.minor[
      between(tick.values.minor, limits.z[1], limits.z[2])]
  label.z.text <- signif(tick.values.major, 3)

  if (log.z) {
    tick.values.major[tick.values.major == 0] <- 1
    tick.values.minor[tick.values.minor == 0] <- 1

    tick.values.major <- log(tick.values.major)
    tick.values.minor <- log(tick.values.minor)
    label.z.text <- signif(exp(tick.values.major) - De.add, 3)
  }

  ## calculate z-axis radius
  r <- max(sqrt((limits.x[2])^2 + (data.global[,7] * f)^2))

  ## calculate node coordinates for semi-circle
  ellipse.values <- c(min(ifelse(log.z,
                                 log(limits.z[1]),
                                 limits.z[1]),
                          tick.values.major,
                          tick.values.minor),
                      max(ifelse(log.z,
                                 log(limits.z[2]),
                                 limits.z[2]),
                          tick.values.major,
                          tick.values.minor))

  ## correct for unpleasant value
  ellipse.values[ellipse.values == -Inf] <- 0

  ellipse.x <- r / sqrt(1 + f^2 * (ellipse.values - z.central.global)^2)
  ellipse.y <- (ellipse.values - z.central.global) * ellipse.x
  ellipse <- cbind(ellipse.x, ellipse.y)
  if (rotate)
    ellipse <- ellipse[, 2:1]

  ## calculate statistical labels
  stats.data <- matrix(nrow = 3, ncol = 3)
  data.stats <- as.numeric(data.global[,1])

  if ("min" %in% stats) {
    stats.data[1, 3] <- data.stats[data.stats == min(data.stats)][1]
    stats.data[1, 1] <- data.global[data.stats == stats.data[1, 3], 6][1]
    stats.data[1, 2] <- data.global[data.stats == stats.data[1, 3], 8][1]
  }

  if ("max" %in% stats) {
    stats.data[2, 3] <- data.stats[data.stats == max(data.stats)][1]
    stats.data[2, 1] <- data.global[data.stats == stats.data[2, 3], 6][1]
    stats.data[2, 2] <- data.global[data.stats == stats.data[2, 3], 8][1]
  }

  if ("median" %in% stats) {
    stats.data[3, 3] <- data.stats[data.stats == quantile(data.stats, 0.5, type = 3)]
    stats.data[3, 1] <- data.global[data.stats == stats.data[3, 3], 6][1]
    stats.data[3, 2] <- data.global[data.stats == stats.data[3, 3], 8][1]
  }

  ## index to pick according to the value of the rotate argument
  rotate.idx <- if (!rotate) 1 else 2
  min.ellipse <- min(ellipse[, rotate.idx])
  max.ellipse <- max(ellipse[, rotate.idx])

  ## re-calculate axes limits if necessary
  if(!("ylim" %in% names(extraArgs))) {
    limits.z.y <- range(ellipse[, 3 - rotate.idx])
    if(limits.z.y[1] < 0.66 * limits.y[1]) {
      limits.y[1] <- 1.8 * limits.z.y[1]
    }
    if(limits.z.y[2] > 0.77 * limits.y[2]) {
      limits.y[2] <- 1.3 * limits.z.y[2]
    }

    if (rotate) {
      limits.y <- c(-max(abs(limits.y)), max(abs(limits.y)))
    }
  }
  if(!("xlim" %in% names(extraArgs))) {
    limits.z.x <- range(ellipse[, rotate.idx])
    limits.x[2] <- max(limits.z.x[2], limits.z.x)
  }

  ## calculate and paste statistical summary
  De.stats <- matrix(nrow = length(data), ncol = 12)
  colnames(De.stats) <- c("n",
                          "mean",
                          "median",
                          "kdemax",
                          "sd.abs",
                          "sd.rel",
                          "se.abs",
                          "se.rel",
                          "q.25",
                          "q.75",
                          "skewness",
                          "kurtosis")

  for(i in 1:length(data)) {
    statistics <- calc_Statistics(data[[i]])[[summary.method]]
    statistics.2 <- calc_Statistics(data[[i]][,3:4])[[summary.method]]

    De.stats[i,1] <- statistics$n
    De.stats[i,2] <- statistics.2$mean
    De.stats[i,3] <- statistics.2$median
    De.stats[i,5] <- statistics$sd.abs
    De.stats[i,6] <- statistics$sd.rel
    De.stats[i,7] <- statistics$se.abs
    De.stats[i,8] <- statistics$se.rel
    De.stats[i,9] <- quantile(data[[i]][,1], 0.25)
    De.stats[i,10] <- quantile(data[[i]][,1], 0.75)
    De.stats[i,11] <- statistics$skewness
    De.stats[i,12] <- statistics$kurtosis

    ## account for log.z-option
    if (log.z) {
      De.stats[i,2:4] <- exp(De.stats[i,2:4]) - De.add
    }

    ## kdemax - here a little doubled as it appears below again
    De.density <- try(density(x = data[[i]][,1],
                              kernel = "gaussian",
                              bw = bw,
                              from = limits.z[1],
                              to = limits.z[2]),
                      silent = TRUE)

    De.stats[i, 4] <- NA
    if (!inherits(De.density, "try-error")) {
      De.stats[i, 4] <- De.density$x[which.max(De.density$y)]
    }
  }

  ## helper to generate an element of the statistical summary
  .summary_line <- function(keyword, summary, val, label = keyword,
                            percent = FALSE, sep = FALSE, digits = 2) {
    ifelse(keyword %in% summary,
           paste0(label, " = ", round(val, digits),
                  if (percent) " %" else NULL, if (sep) " | " else "\n"),
           "")
  }

  is.sub <- summary.pos[1] == "sub"
  stops <- NULL
  label.text <- list()
  for (i in 1:length(data)) {
    if (!is.sub)
      stops <- strrep("\n", (i - 1) * length(summary))

    summary.text <- character(0)
    for (j in 1:length(summary)) {
      summary.text <-
        c(summary.text,
          .summary_line("n", summary[j], De.stats[i, 1], sep = is.sub),
          .summary_line("mean", summary[j], De.stats[i, 2], sep = is.sub),
          .summary_line("median", summary[j], De.stats[i, 3], sep = is.sub),
          .summary_line("kdemax", summary[j], De.stats[i, 4], sep = is.sub),
          .summary_line("sd.abs", summary[j], De.stats[i, 5], sep = is.sub,
                        label = "abs. sd"),
          .summary_line("sd.rel", summary[j], De.stats[i, 6], sep = is.sub,
                        label = "rel. sd"),
          .summary_line("se.abs", summary[j], De.stats[i, 7], sep = is.sub,
                        label = "se"),
          .summary_line("se.rel", summary[j], De.stats[i, 8], sep = is.sub,
                        label = "rel. se", percent = TRUE),
          .summary_line("skewness", summary[j], De.stats[i, 11], sep = is.sub),
          .summary_line("kurtosis", summary[j], De.stats[i, 12], sep = is.sub),
          .summary_line("in.2s", summary[j],
                        sum(data[[i]][,7] > -2 & data[[i]][,7] < 2) /
                        nrow(data[[i]]) * 100, sep = is.sub,
                        label = "in 2 sigma", percent = TRUE, digits = 1))
    }
    label.text[[i]] <- paste0(
        if (is.sub) "" else stops,
        paste(summary.text, collapse = ""),
        stops)
  }

  ## remove outer vertical lines from string1
  if (is.sub) {
    for (i in seq_along(label.text)) {
      label.text[[i]] <- substr(x = label.text[[i]],
                                start = 1,
                                stop = nchar(label.text[[i]]) - 3)
      if (!is.null(line.mtext))
        label.text[[i]] <- paste(label.text[[i]], "|", line.mtext[i])
    }
  }

  if (!rotate) {
    ## convert keywords into summary placement coordinates
    coords <- .get_keyword_coordinates(summary.pos, limits.x, limits.y)

    ## apply some adjustments to the y positioning
    if (summary.pos[1] %in% c("topleft", "top", "topright")) {
        coords$pos[2] <- coords$pos[2] - par()$cxy[2] * 1.0
      } else if (summary.pos[1] %in% c("bottomleft", "bottom", "bottomright")) {
        coords$pos[2] <- coords$pos[2] + par()$cxy[2] * 3.5
    }
    summary.pos <- coords$pos
    summary.adj <- coords$adj

    ## convert keywords into legend placement coordinates
    coords <- .get_keyword_coordinates(legend.pos, limits.x, limits.y)
    legend.pos <- coords$pos
    legend.adj <- coords$adj

  } else {
    ## convert keywords into summary placement coordinates
    ## this time we swap x and y limits as we are rotated, then apply some
    ## adjustments to the x positioning
    coords <- .get_keyword_coordinates(summary.pos, limits.y, limits.x)
    if (summary.pos[1] %in% c("topleft", "left", "bottomleft")) {
      coords$pos[1] <- coords$pos[1] + par()$cxy[1] * 7.5
    }
    summary.pos <- coords$pos
    summary.adj <- coords$adj

    ## convert keywords into legend placement coordinates
    ## this time we swap x and y limits as we are rotated, then apply some
    ## adjustments to the x positioning
    coords <- .get_keyword_coordinates(legend.pos, limits.y, limits.x)
    if (!is.null(legend.pos) &&
        legend.pos[1] %in% c("topleft", "left", "bottomleft")) {
      coords$pos[1] <- coords$pos[1] + par()$cxy[1] * 7.5
    }
    legend.pos <- coords$pos
    legend.adj <- coords$adj
  }

  ## define cartesian plot origins
  if (!rotate) {
    xy.0 <- c(min(ellipse[,1]) * lostintranslation, min(ellipse[,2]))
  } else {
    xy.0 <- c(min(ellipse[,1]), min(ellipse[,2]) * lostintranslation)
  }

  ## calculate coordinates for dispersion polygon overlay
  y.max <- if (!rotate) par()$usr[2] else par()$usr[4]

  polygons.x <- c(limits.x[1], limits.x[2], xy.0[rotate.idx], y.max, y.max,
                  xy.0[rotate.idx], limits.x[2])
  polygons.y <- matrix(nrow = length(data), ncol = 7)
  for(i in 1:length(data)) {
    if(dispersion == "qr") {
      ci.lo_up <- quantile(data[[i]][, 1], c(0.25, 0.75))
    } else if (grepl("p", dispersion)) {
      ci.plot <- as.numeric(strsplit(x = dispersion,
                                     split = "p")[[1]][2])
      ci.plot <- (100 - ci.plot) / 100
      ci.lo_up <- quantile(data[[i]][, 1], c(ci.plot, 1 - ci.plot))
    } else if(dispersion == "sd") {
      if (log.z) {
        ci.lo_up <- exp(mean(log(data[[i]][, 1])) + c(-1, 1) * sd(log(data[[i]][, 1])))
      } else {
        ci.lo_up <- mean(data[[i]][, 1]) + c(-1, 1) * sd(data[[i]][, 1])
      }
    } else if(dispersion == "2sd") {
      if (log.z) {
        ci.lo_up <- exp(mean(log(data[[i]][, 1])) + c(-2, 2) * sd(log(data[[i]][, 1])))
      } else {
        ci.lo_up <- mean(data[[i]][, 1]) + c(-2, 2) * sd(data[[i]][, 1])
      }
    }

    if (log.z) {
      ci.lo_up[which(ci.lo_up < 0)] <- 1
      ci.lo_up <- log(ci.lo_up)
    }
    y.lower <- ci.lo_up[1] - z.central.global
    y.upper <- ci.lo_up[2] - z.central.global

    polygons.y[i, ] <- c(0,
                         y.upper * limits.x[2],
                         y.upper * xy.0[rotate.idx],
                         y.upper * xy.0[rotate.idx],
                         y.lower * xy.0[rotate.idx],
                         y.lower * xy.0[rotate.idx],
                         y.lower * limits.x[2])
  }

  ## append information about data in confidence interval
  for(i in 1:length(data)) {
    data.in.2s <- rep(x = FALSE, times = nrow(data[[i]]))
    data.in.2s[data[[i]][,8] > -2 & data[[i]][,8] < 2] <- TRUE
    data[[i]] <- cbind(data[[i]], data.in.2s)
  }

  ## calculate KDE
  KDE <- list()
  KDE.bw <- numeric(length(data))

  for(i in 1:length(data)) {
    KDE.i <- density(x = data[[i]][,3],
                     kernel = "gaussian",
                     bw = bw,
                     from = ellipse.values[1],
                     to = ellipse.values[2],
                     weights = data[[i]]$weights)
    KDE.bw[i] <- KDE.i$bw
    KDE[[i]] <- rbind(c(min(KDE.i$x), 0),
                      cbind(KDE.i$x, KDE.i$y),
                      c(max(KDE.i$x), 0))
  }

  ## calculate mean KDE bandwidth
  KDE.bw <- mean(KDE.bw, na.rm = TRUE)

  ## calculate KDE width
  KDE.max <- max(vapply(KDE, function(x) max(x[, 2]), numeric(1)))

  ## optionally adjust KDE width for boxplot option
  if (boxplot) {
    KDE.max <- 1.3 * KDE.max
  }

  ## Generate plot ------------------------------------------------------------
  ##
  ## determine number of subheader lines to shift the plot
  if(length(summary) > 0 & summary.pos[1] == "sub") {
    shift.lines <- (length(data) + 1) * layout$abanico$dimension$summary.line/100
  } else {
    shift.lines <- 1
  }

  ## extract original plot parameters
  bg.original <- par()$bg
  par(bg = layout$abanico$colour$background)

  ## setup plot area
  par(mar = if (!rotate) c(4.5, 4.5, shift.lines + 1.5, 7) else c(4, 4, shift.lines + 5, 4),
      xpd = TRUE,
      cex = cex)

    dim <- layout$abanico$dimension
    if (dim$figure.width != "auto" || dim$figure.height != "auto") {
      par(mai = dim$margin / 25.4,
          pin = c(dim$figure.width - dim$margin[2] - dim$margin[4],
                  dim$figure.height - dim$margin[1] - dim$margin[3]) / 25.4)
    }

    ## create empty plot
    par(new = TRUE)
    plot.rot(NA,
         xlim = c(limits.x[1], limits.x[2] * (1 / plot.ratio)),
         ylim = limits.y,
         main = "",
         sub = sub,
         xlab = "",
         ylab = "",
         xaxs = "i",
         yaxs = "i",
         frame.plot = FALSE,
         axes = FALSE)

    ## add y-axis label
    mtext(text = ylab,
          at = 0,
          adj = 0.5,
          side = 3 - rotate.idx,
          line = 3 * layout$abanico$dimension$ylab.line / 100,
          col = layout$abanico$colour$ylab,
          family = layout$abanico$font.type$ylab,
          font = which(c("normal", "bold", "italic", "bold italic") ==
                         layout$abanico$font.deco$ylab)[1],
          cex = cex * layout$abanico$font.size$ylab / 12)

    ## calculate upper x-axis label values
    label.x.upper <- as.character(round(1 / axTicks(side = rotate.idx)[-1] *
                                        if (log.z) 100 else 1, 1))

    ## optionally, plot 2-sigma-bar
    if (!isFALSE(bar[1])) {
      if (is.logical(bar)) {
        bar <- sapply(data, function(x) x[1, 5])
      } else if (log.z) {
        bar <- log(bar)
      }
      bars.xmax <- ifelse("xlim" %in% names(extraArgs),
                          extraArgs$xlim[2] * 0.95,
                          max(data.global$precision))
      bars.ymax <- (bar - z.central.global) * bars.xmax
      bars.x <- c(limits.x[1], limits.x[1], bars.xmax, bars.xmax)
      bars.y <- cbind(-2, 2, bars.ymax + 2, bars.ymax - 2)

      for (i in 1:length(bar)) {
        polygon.rot(x = bars.x,
                    y = bars.y[i, ],
                    col = bar.fill[i],
                    border = bar.line[i])
      }
    }

    ## remove unwanted parts
    polygon.rot(x = par()$usr[2] * c(1, 1, 2, 2),
                y = c(min(ellipse[, 2]), max(ellipse[, 2]),
                      max(ellipse[, 2]), min(ellipse[, 2])) * 2,
            col = bg.original,
            lty = 0)

    ## optionally, plot dispersion polygon
    if (polygon.fill[1] != "none") {
      for (i in 1:length(data)) {
        polygon.rot(x = polygons.x,
                    y = polygons.y[i, ],
                col = polygon.fill[i],
                border = polygon.line[i])
      }
    }

    ## optionally, add minor grid lines
    if (grid.minor != "none") {
      for (i in 1:length(tick.values.minor)) {
        lines.rot(x = c(limits.x[1], min.ellipse),
              y = c(0, tick.values.minor[i] - z.central.global) *
                min.ellipse,
              col = grid.minor,
              lwd = 1)
        lines.rot(x = c(xy.0[rotate.idx], y.max),
              y = c(tick.values.minor[i] - z.central.global,
                    tick.values.minor[i] - z.central.global) * min.ellipse,
              col = grid.minor,
              lwd = 1)
      }
    }

    ## optionally, add major grid lines
    if (grid.major != "none") {
      for (i in 1:length(tick.values.major)) {
        lines.rot(x = c(limits.x[1], min.ellipse),
              y = c(0, tick.values.major[i] - z.central.global) *
                min.ellipse,
              col = grid.major,
              lwd = 1)
        lines.rot(x = c(xy.0[rotate.idx], y.max),
              y = c(tick.values.major[i] - z.central.global,
                    tick.values.major[i] - z.central.global) * min.ellipse,
              col = grid.major,
              lwd = 1)
      }
    }

    ## optionally, plot lines for each bar
    if (lwd[1] > 0 && lty[1] > 0 && !isFALSE(bar[1])) {
      for (i in 1:length(data)) {
        z.line <- if (length(bar) == 1) bar[1] else bar[i]
        x2 <- r / sqrt(1 + f^2 * (z.line - z.central.global)^2)
        y2 <- (z.line - z.central.global) * x2
        lines.rot(x = c(limits.x[1], x2, xy.0[rotate.idx], y.max),
              y = c(0, y2, y2, y2),
              lty = lty[i],
              lwd = lwd[i],
              col = centrality.col[i])
      }
    }

  ## optionally add KDE plot
  if (kde) {
    ## calculate max KDE value for axis label
    KDE.max.plot <- 0
    for (x in data) {
      KDE.plot <- density(x[, 1],
                          kernel = "gaussian",
                          bw = bw,
                          from = limits.z[1],
                          to = limits.z[2])
      KDE.max.plot <- max(KDE.plot$y, KDE.max.plot)
    }
    KDE.scale <- (y.max - xy.0[rotate.idx]) / (KDE.max * 1.05)

    ## plot KDE lines
    for (i in 1:length(data)) {
      polygon.rot(x = xy.0[rotate.idx] + KDE[[i]][, 2] * KDE.scale,
                  y = (KDE[[i]][, 1] - z.central.global) * min.ellipse,
                  col = kde.fill[i],
                  border = kde.line[i],
                  lwd = 1.7)
    }

    ## plot KDE x-axis
    axis(side = rotate.idx,
         at = c(xy.0[rotate.idx], y.max),
         col = layout$abanico$colour$xtck3,
         col.axis = layout$abanico$colour$xtck3,
         labels = NA,
         tcl = -layout$abanico$dimension$xtcl3 / 200,
         cex = cex)

    axis(side = rotate.idx,
         at = c(xy.0[rotate.idx], y.max),
         labels = as.character(round(c(0, KDE.max.plot), 3)),
         line = 2 * layout$abanico$dimension$xtck3.line / 100 - 2,
         lwd = 0,
         col = layout$abanico$colour$xtck3,
         family = layout$abanico$font.type$xtck3,
         font = which(c("normal", "bold", "italic", "bold italic") ==
                      layout$abanico$font.deco$xtck3)[1],
         col.axis = layout$abanico$colour$xtck3,
         cex.axis = layout$abanico$font.size$xtck3 / 12)

    ## KDE x-axis label including bandwidth size
    mtext(text = paste0(xlab[3],
                        " (bw ",
                        round(x = KDE.bw,
                              digits = 3),
                        ")"),
          at = (xy.0[rotate.idx] + y.max) / 2,
          side = rotate.idx,
          line = 2.5 * layout$abanico$dimension$xlab3.line / 100,
          col = layout$abanico$colour$xlab3,
          family = layout$abanico$font.type$xlab3,
          font = which(c("normal", "bold", "italic", "bold italic") ==
                       layout$abanico$font.deco$xlab3)[1],
          cex = cex * layout$abanico$font.size$xlab3 / 12)
  }

  ## optionally add further lines
  if (length(line) > 0) {

    ## check if line parameters are RLum.Results objects
    if (is.list(line)) {
      for (i in seq_along(line)) {
        if (inherits(line[[i]], "RLum.Results")) {
          line[[i]] <- as.numeric(get_RLum(line[[i]], data.object = "summary")$de)
        }
      }
    } else if (inherits(line, "RLum.Results")) {
      line <- as.numeric(get_RLum(line, data.object = "summary")$de)
    }

    ## convert list to vector
    if (is.list(line))
      line <- unlist(line)
    if (log.z)
      line <- log(line)
    if (is.null(line.col))
      line.col <- seq_along(line) + ifelse(line.is.null, 1, 0)

    ## calculate line coordinates and further parameters
    line.x <- c(limits.x[1], min.ellipse, y.max)
    line.y <- (line - z.central.global) * min.ellipse

    for (i in 1:length(line)) {
        lines.rot(x = line.x,
                  y = c(0, line.y[i], line.y[i]),
                  col = line.col[i],
                  lty = line.lty[i] %||% 1
                  )
        text_with_bg.rot(x = line.x[3] - par()$cxy[rotate.idx] * 0.5,
                         y = line.y[i] + par()$cxy[3 - rotate.idx] * 0.5,
                         label = line.label[i] %||% "",
                         pos = 3 - rotate.idx,
                         bg = ifelse(line.is.null, line.col[i], NA),
                         cex = 0.8)
      }
    }

    ## add plot title
    add.shift <- if (!rotate) 0 else 3.5
    title(main = main,
          family = layout$abanico$font.type$main,
          font = which(c("normal", "bold", "italic", "bold italic") ==
                         layout$abanico$font.deco$main)[1],
          col.main = layout$abanico$colour$main,
          cex = layout$abanico$font.size$main / 12,
          line = (shift.lines + add.shift) * layout$abanico$dimension$main / 100)

    ## calculate lower x-axis (precision)
    x.axis.ticks <- axTicks(side = rotate.idx)
    x.axis.ticks <- x.axis.ticks[c(TRUE, x.axis.ticks <= limits.x[2])]
    x.axis.ticks <- x.axis.ticks[x.axis.ticks <= max.ellipse]

    ## x-axis with labels and ticks
    axis(side = rotate.idx,
         at = x.axis.ticks,
         col = layout$abanico$colour$xtck1,
         col.axis = layout$abanico$colour$xtck1,
         labels = NA,
         tcl = -layout$abanico$dimension$xtcl1 / 200,
         cex = cex)
    axis(side = rotate.idx,
         at = x.axis.ticks,
         line = 2 * layout$abanico$dimension$xtck1.line / 100 - 2,
         lwd = 0,
         col = layout$abanico$colour$xtck1,
         family = layout$abanico$font.type$xtck1,
         font = which(c("normal", "bold", "italic", "bold italic") ==
                        layout$abanico$font.deco$xtck1)[1],
         col.axis = layout$abanico$colour$xtck1,
         cex.axis = layout$abanico$font.size$xlab1 / 12)

    ## extend axis line to right side of the plot
    lines.rot(x = c(max(x.axis.ticks), max.ellipse),
          y = c(limits.y[1], limits.y[1]),
          col = layout$abanico$colour$xtck1)

    ## draw closing tick on right hand side
    axis(side = rotate.idx,
         tcl = -layout$abanico$dimension$xtcl1 / 200,
         lwd = 0,
         lwd.ticks = 1,
         at = limits.x[2],
         labels = FALSE,
         col = layout$abanico$colour$xtck1)

    axis(side = rotate.idx,
         tcl = layout$abanico$dimension$xtcl2 / 200,
         lwd = 0,
         lwd.ticks = 1,
         at = limits.x[2],
         labels = FALSE,
         col = layout$abanico$colour$xtck2)

    ## add lower axis label
    mtext(xlab[2],
          at = (limits.x[1] + max.ellipse) / 2,
          side = rotate.idx,
          line = 2.5 * layout$abanico$dimension$xlab1.line / 100,
          col = layout$abanico$colour$xlab1,
          family = layout$abanico$font.type$xlab1,
          font = which(c("normal", "bold", "italic", "bold italic") ==
                         layout$abanico$font.deco$xlab1)[1],
          cex = cex * layout$abanico$font.size$xlab1 / 12)

    ## add upper axis label
    mtext(xlab[1],
          at = (limits.x[1] + max.ellipse) / 2,
          side = rotate.idx,
          line = -3.5 * layout$abanico$dimension$xlab2.line / 100,
          col = layout$abanico$colour$xlab2,
          family = layout$abanico$font.type$xlab2,
          font = which(c("normal", "bold", "italic", "bold italic") ==
                         layout$abanico$font.deco$xlab2)[1],
          cex = cex * layout$abanico$font.size$xlab2 / 12)

    ## plot upper x-axis
    axis(side = rotate.idx,
         at = x.axis.ticks[-1],
         col = layout$abanico$colour$xtck2,
         col.axis = layout$abanico$colour$xtck2,
         labels = NA,
         tcl = layout$abanico$dimension$xtcl2 / 200,
         cex = cex)

    ## remove first tick label (infinity)
    label.x.upper <- label.x.upper[1:(length(x.axis.ticks) - 1)]

  if (length(x.axis.ticks) > 1) {
    axis(side = rotate.idx,
         at = x.axis.ticks[-1],
         labels = label.x.upper,
         line = -1 * layout$abanico$dimension$xtck2.line / 100 - 2,
         lwd = 0,
         col = layout$abanico$colour$xtck2,
         family = layout$abanico$font.type$xtck2,
         font = which(c("normal", "bold", "italic", "bold italic") ==
                        layout$abanico$font.deco$xtck2)[1],
         col.axis = layout$abanico$colour$xtck2,
         cex.axis = layout$abanico$font.size$xlab2 / 12)
  }

  ## plot y-axis
  if (is.null(extraArgs$yaxt) || extraArgs$yaxt != "n") {
    line <- 2 * layout$abanico$dimension$ytck.line / 100 - 2
    family <- layout$abanico$font.type$ytck
    font <- which(c("normal", "bold", "italic", "bold italic") ==
                  layout$abanico$font.deco$ytck)[1]
    col.axis <- layout$abanico$colour$ytck
    cex.axis <- layout$abanico$font.size$ylab / 12
    if (y.axis) {
      char.height <- par()$cxy[2]

      ## this comes into play for panel plots, e.g., par(mfrow = c(4,4))
      if (char.height > 4.5 / cex) {
        axis(side = 3 - rotate.idx,
             at = c(-2, 2),
             tcl = -layout$abanico$dimension$ytcl / 200,
             lwd = 1,
             lwd.ticks = 1,
             labels = NA,
             las = 1,
             col = col.axis)
        axis(side = 3 - rotate.idx,
             at = 0,
             tcl = 0,
             labels = "\u00B1 2",
             line = line, las = 1,
             family = family, font = font,
             col.axis = col.axis, cex.axis = cex.axis)
      } else {
        axis(side = 3 - rotate.idx,
             at = seq(-2, 2, by = 2),
             line = line, las = 1,
             tcl = -layout$abanico$dimension$ytcl / 200,
             family = family, font = font,
             col.axis = col.axis, cex.axis = cex.axis)
      }
    } else {
      axis(side = 3 - rotate.idx,
           at = 0,
           line = line, las = 1,
           tcl = -layout$abanico$dimension$ytcl / 200,
           family = family, font = font,
           col.axis = col.axis, cex.axis = cex.axis)
    }
  }

  ## plot minor z-ticks
  for (i in 1:length(tick.values.minor)) {
    lines.rot(x = y.max * c(1, 1 + 0.007 * layout$abanico$dimension$ztcl / 100),
            y = c(tick.values.minor[i] - z.central.global,
                  tick.values.minor[i] - z.central.global) *
              min(ellipse[, rotate.idx]),
            col = layout$abanico$colour$ztck)
  }

  ## plot major z-ticks
  for (i in 1:length(tick.values.major)) {
    lines.rot(x = y.max * c(1, 1 + 0.015 * layout$abanico$dimension$ztcl / 100),
              y = c(tick.values.major[i] - z.central.global,
                    tick.values.major[i] - z.central.global) *
                min(ellipse[, rotate.idx]),
              col = layout$abanico$colour$ztck)
  }

  ## plot z-axes
  lines(ellipse, col = layout$abanico$colour$border)
  lines.rot(x = rep(y.max, nrow(ellipse)),
            y = ellipse[, 2],
            col = layout$abanico$colour$ztck)

  ## plot z-axis text
  text.rot(x = y.max * (1 + 0.02 * layout$abanico$dimension$ztcl / 100),
           y = (tick.values.major - z.central.global) * min(ellipse[, rotate.idx]),
           labels = label.z.text,
           adj = if (rotate) c(0.5, 0) else c(0, 0.5),
           family = layout$abanico$font.type$ztck,
           font = which(c("normal", "bold", "italic", "bold italic") ==
                        layout$abanico$font.deco$ztck)[1],
           cex = layout$abanico$font.size$ztck / 12)

  ## plot z-label
  mtext(text = zlab,
        at = 0,
        side = 5 - rotate.idx,
        las = ifelse(rotate, 1, 3),
        adj = 0.5,
        line = (ifelse(rotate, 1.5, 4) + cex) * layout$abanico$dimension$zlab.line / 100,
        col = layout$abanico$colour$zlab,
        family = layout$abanico$font.type$zlab,
        font = which(c("normal", "bold", "italic", "bold italic") ==
                     layout$abanico$font.deco$zlab)[1],
        cex = cex * layout$abanico$font.size$zlab / 12)

  ## plot values and optionally error bars
  if (error.bars) {
    for (i in 1:length(data)) {
      arrow.x <- data[[i]][, 6]
      arrow.y1 <- data[[i]][, 1] - data[[i]][, 2]
      arrow.y2 <- data[[i]][, 1] + data[[i]][, 2]
      if (log.z) {
        arrow.y1 <- log(arrow.y1)
        arrow.y2 <- log(arrow.y2)
      }

      arrow.coords <- cbind(
          arrow.x,
          arrow.x,
          (arrow.y1 - z.central.global) * arrow.x,
          (arrow.y2 - z.central.global) * arrow.x)

      graphics::arrows(
               x0 = arrow.coords[, 2 * rotate.idx - 1],
               x1 = arrow.coords[, 2 * rotate.idx],
               y0 = arrow.coords[, 2 * (3 - rotate.idx) - 1],
               y1 = arrow.coords[, 2 * (3 - rotate.idx)],
               length = 0,
               angle = 90,
               code = 3,
               col = value.bar[i])
    }
  }

  for (i in 1:length(data)) {
    points.rot(x = data[[i]][, 6][data[[i]][, 6] <= limits.x[2]],
               y = data[[i]][, 8][data[[i]][, 6] <= limits.x[2]],
             col = value.dot[i],
             pch = pch[i],
             cex = layout$abanico$dimension$pch / 100)
  }

  ## compute data for histogram and dot plot
  if (hist || dots) {
    ## calculate histogram data without plotting
    hist.data <- lapply(data, function(x) {
      hist(x[, 3], plot = FALSE, breaks = breaks)
    })

    ## calculate maximum histogram bar height for normalisation
    hist.max <- max(vapply(hist.data, function(x) max(x$counts, na.rm = TRUE),
                           numeric(1)), na.rm = TRUE)

    ## calculate scaling factor for histogram bar heights
    hist.scale <- (y.max - xy.0[rotate.idx]) / (hist.max * 1.05)

    ## normalise histogram bar height to KDE dimensions
    for (i in 1:length(data)) {
      hist.data[[i]]$density <- hist.data[[i]]$counts * hist.scale
      hist.data[[i]]$breaks <- hist.data[[i]]$breaks - z.central.global
    }
  }

  ## optionally add histogram
  if (hist) {
      axis(side = rotate.idx,
           at = c(xy.0[rotate.idx], y.max),
           labels = as.character(c(0, hist.max)),
           line = -1 * layout$abanico$dimension$xtck3.line / 100 - 2,
           lwd = 0,
           col = layout$abanico$colour$xtck3,
           family = layout$abanico$font.type$xtck3,
           font = which(c("normal", "bold", "italic", "bold italic") ==
                          layout$abanico$font.deco$xtck3)[1],
           col.axis = layout$abanico$colour$xtck3,
           cex.axis = layout$abanico$font.size$xtck3 / 12)

      ## add label
      mtext(text = "n",
            at = (xy.0[rotate.idx] + y.max) / 2,
            side = rotate.idx,
            line = -3.5 * layout$abanico$dimension$xlab2.line / 100,
            col = layout$abanico$colour$xlab2,
            family = layout$abanico$font.type$xlab2,
            font = which(c("normal", "bold", "italic", "bold italic") ==
                           layout$abanico$font.deco$xlab2)[1],
            cex = cex * layout$abanico$font.size$xlab2 / 12)

      ## plot ticks
      axis(side = rotate.idx,
           at = c(xy.0[rotate.idx], y.max),
           col = layout$abanico$colour$xtck2,
           col.axis = layout$abanico$colour$xtck2,
           labels = NA,
           tcl = layout$abanico$dimension$xtcl2 / 200,
           cex = cex)

    ## draw each bar for each data set
    for (i in 1:length(data)) {
      for (j in 1:length(hist.data[[i]]$density)) {
          ## calculate x-coordinates
          hist.x.i <- xy.0[rotate.idx] + c(0, 0, rep(hist.data[[i]]$density[j], 2))

          ## calculate y-coordinates
          hist.y.i <- c(hist.data[[i]]$breaks[j],
                        hist.data[[i]]$breaks[j + 1],
                        hist.data[[i]]$breaks[j + 1],
                        hist.data[[i]]$breaks[j]) * min.ellipse

          ## remove data out of z-axis range
          hist.y.i <- pmax(hist.y.i, min(ellipse[, 3 - rotate.idx]))
          hist.y.i <- pmin(hist.y.i, max(ellipse[, 3 - rotate.idx]))

          ## draw the bars
          polygon.rot(x = hist.x.i,
                      y = hist.y.i,
                      col = kde.fill[i],
                      border = kde.line[i])
        }
    }
  }

  ## optionally add box plot
  if (boxplot) {

    box.x <- c(min.ellipse + KDE.max * 0.85, xy.0[rotate.idx] + KDE.max * 0.95)
    for (i in 1:length(data)) {
      ## calculate boxplot data without plotting
      boxplot.data <- graphics::boxplot(data[[i]][, 3], plot = FALSE)
      stats <- (boxplot.data$stats[, 1] - z.central.global) * min.ellipse

      ## draw median line
      lines.rot(x = box.x,
                y = c(stats[3], stats[3]),
                lwd = 2,
                col = kde.line[i])

      ## draw p25-p75-polygon
      polygon.rot(x = rep(box.x, each = 2),
                  y = c(stats[2], stats[4], stats[4], stats[2]),
                  border = kde.line[i])

      ## draw lower whisker
      lines.rot(x = c(rep(mean(box.x), 2), box.x),
                y = c(stats[2], stats[1], stats[1], stats[1]),
                col = kde.line[i])

      ## draw upper whisker
      lines.rot(x = c(rep(mean(box.x), 2), box.x),
                y = c(stats[4], stats[5], stats[5], stats[5]),
                col = kde.line[i])

      ## draw outlier points
      points.rot(x = rep(mean(box.x), length(boxplot.data$out)),
                 y = (boxplot.data$out - z.central.global) * min.ellipse,
                 cex = 0.8,
                 col = kde.line[i])
    }
  }

  ## optionally add dot plot
  if (dots) {
    ## calculate distance between dots
    dots.distance <- (y.max - (xy.0[rotate.idx] + par()$cxy[rotate.idx] * 0.4)) / hist.max

    for (i in 1:length(data)) {
      for (j in 1:length(hist.data[[i]]$counts)) {
        dots.x.i <- seq(from = xy.0[rotate.idx] + par()$cxy[rotate.idx] * 0.4,
                        by = dots.distance,
                        length.out = hist.data[[i]]$counts[j])

        dots.y.i <- rep((hist.data[[i]]$mids[j] - z.central.global) *
                        min.ellipse, length(dots.x.i))

        ## remove data out of z-axis range
        keep.idx <- between(dots.y.i,
                            min(ellipse[, 3 - rotate.idx]),
                            max(ellipse[, 3 - rotate.idx]))
        dots.x.i <- dots.x.i[keep.idx]
        dots.y.i <- dots.y.i[keep.idx]

        max.val <- y.max - par()$cxy[rotate.idx] * 0.4
        if (max(c(0, dots.x.i), na.rm = TRUE) >= max.val) {
          dots.y.i <- dots.y.i[dots.x.i < max.val]
          dots.x.i <- dots.x.i[dots.x.i < max.val]
        }

        ## plot points
        points.rot(x = dots.x.i,
                   y = dots.y.i,
                   pch = if (!rotate) "|" else "-",
                   cex = 0.7,
                   col = kde.line[i])
      }
    }
  }

  ## optionally add stats, i.e. min, max, median sample text
  if (length(stats) > 0) {
    text.rot(x = stats.data[, 1],
             y = stats.data[, 2],
             pos = 2,
             labels = round(stats.data[, 3], 1),
             family = layout$abanico$font.type$stats,
             font = which(c("normal", "bold", "italic", "bold italic") ==
                          layout$abanico$font.deco$stats)[1],
             cex = layout$abanico$font.size$stats / 12,
             col = layout$abanico$colour$stats)
  }

  ## optionally add rug
  if (rug) {
    rug.x <- c(1 - 0.013 * (layout$abanico$dimension$rugl / 100), 1) * xy.0[rotate.idx]
    rug.y <- ((if (log.z) log(De.global) else De.global) - z.central.global) * min.ellipse
    for (i in 1:length(rug.y)) {
      lines.rot(x = rug.x,
                y = rep(rug.y[i], 2),
                col = value.rug[data.global[i, 10]])
    }
  }

  ## plot KDE base line
  lines.rot(x = c(xy.0[rotate.idx], xy.0[rotate.idx]),
            y = c(min(ellipse[, 3 - rotate.idx]), max(ellipse[, 3 - rotate.idx])),
            col = layout$abanico$colour$border)

  ## draw border around plot
  if (frame == 1) {
    polygon.rot(x = c(limits.x[1], min.ellipse, y.max,
                      y.max, min.ellipse),
                y = c(0, max(ellipse[, 3 - rotate.idx]), max(ellipse[, 3 - rotate.idx]),
                      min(ellipse[, 3 - rotate.idx]), min(ellipse[, 3 - rotate.idx])),
                border = layout$abanico$colour$border,
                lwd = 0.8)
  } else if (frame == 2) {
    polygon.rot(x = c(limits.x[1], min.ellipse, y.max,
                      y.max, min.ellipse, limits.x[1]),
                y = c(2, max(ellipse[, 3 - rotate.idx]), max(ellipse[, 3 - rotate.idx]),
                      min(ellipse[, 3 - rotate.idx]), min(ellipse[, 3 - rotate.idx]), -2),
                border = layout$abanico$colour$border,
                lwd = 0.8)
  } else if (frame == 3) {
    polygon.rot(x = c(limits.x[1], y.max, y.max, limits.x[1]),
                y = c(max(ellipse[, 3 - rotate.idx]), max(ellipse[, 3 - rotate.idx]),
                      min(ellipse[, 3 - rotate.idx]), min(ellipse[, 3 - rotate.idx])),
                border = layout$abanico$colour$border,
                lwd = 0.8)
  }

  ## optionally add legend content
  if (!is.null(legend)) {
    ## store and change font familiy
    par.family <- par()$family
    par(family = layout$abanico$font.type$legend)

    scale.rot <- if (!rotate) c(1, 0.8) else c(0.8, 1)
    if (rotate)
      legend.adj <- rev(legend.adj)
    legend(x = legend.pos[1] * scale.rot[1],
           y = legend.pos[2] * scale.rot[2],
           xjust = legend.adj[1],
           yjust = legend.adj[2],
           legend = legend,
           pch = pch,
           col = value.dot,
           text.col = value.dot,
           text.font = which(c("normal", "bold", "italic", "bold italic") ==
                             layout$abanico$font.deco$legend)[1],
           cex = layout$abanico$font.size$legend / 12,
           bty = "n")

    ## restore font family
    par(family = par.family)
  }

  ## optionally add subheader text
  add.shift <- if (!rotate) 0 else 3.5
  mtext(text = mtext,
        side = 3,
        line = (shift.lines - 2 + add.shift) * layout$abanico$dimension$mtext / 100,
        col = layout$abanico$colour$mtext,
        family = layout$abanico$font.type$mtext,
        font = which(c("normal", "bold", "italic", "bold italic") ==
                     layout$abanico$font.deco$mtext)[1],
        cex = cex * layout$abanico$font.size$mtext / 12)

  ## add summary content
  for (i in 1:length(data)) {
    if (summary.pos[1] != "sub") {
        text(x = summary.pos[1],
             y = summary.pos[2],
             adj = summary.adj,
             labels = label.text[[i]],
             col = summary.col[i],
             family = layout$abanico$font.type$summary,
             font = which(c("normal", "bold", "italic", "bold italic") ==
                            layout$abanico$font.deco$summary)[1],
             cex = layout$abanico$font.size$summary / 12)
    } else if (mtext == "") {
          mtext(side = 3,
                line = (shift.lines - 1 + add.shift - i) *
                  layout$abanico$dimension$summary / 100 ,
                text = label.text[[i]],
                col = summary.col[i],
                family = layout$abanico$font.type$summary,
                font = which(c("normal", "bold", "italic", "bold italic") ==
                               layout$abanico$font.deco$summary)[1],
                cex = cex * layout$abanico$font.size$summary / 12)
    }
  }

  ##sTeve
  if (fun && !interactive) sTeve() # nocov

  ## create numeric output
  plot.output <- list(xlim = limits.x,
                      ylim = limits.y,
                      zlim = limits.z,
                      polar.box = c(limits.x[1],
                                    limits.x[2],
                                    min(ellipse[,2]),
                                    max(ellipse[,2])),
                      cartesian.box = c(xy.0[1],
                                        par()$usr[2],
                                        xy.0[2],
                                        max(ellipse[,2])),
                      plot.ratio = plot.ratio,
                      data = data,
                      data.global = data.global,
                      KDE = KDE,
                      par = par(no.readonly = TRUE))

  ## INTERACTIVE PLOT ----------------------------------------------------------
  if (interactive) {
    .require_suggested_package("plotly", "The interactive abanico plot")

    ## tidy data ----
    data <- plot.output
    kde <- data.frame(x = data$KDE[[1]][ ,2], y = data$KDE[[1]][ ,1])

    # radial scatter plot ----
    point.text <- paste0("Measured value:<br />",
                         data$data.global$De, " &plusmn; ",
                         data$data.global$error, "<br />",
                         "P(",format(data$data.global$precision,  digits = 2, nsmall = 1),", ",
                         format(data$data.global$std.estimate,  digits = 2, nsmall = 1),")")
    IAP <- plotly::plot_ly(data = data$data.global,
                           x = data$data.global$precision,
                           y = data$data.global$std.estimate,
                           type = "scatter", mode = "markers",
                           hoverinfo = "text", text = point.text,
                           name = "Points",
                           yaxis = "y")

    ellipse <- as.data.frame(ellipse)
    IAP <- plotly::add_trace(IAP, data = ellipse,
                             x = ~ellipse.x, y = ~ellipse.y,
                             type = "scatter", mode = "lines",
                             hoverinfo = "none", text = "",
                             name = "z-axis (left)",
                             line = list(color = "black",
                                         width = 1),
                             yaxis = "y")

    ellipse.right <- ellipse
    ellipse.right$ellipse.x <- ellipse.right$ellipse.x * 1/0.75

    IAP <- plotly::add_trace(IAP, data = ellipse.right,
                             x = ~ellipse.x, y = ~ellipse.y,
                             type = "scatter", mode = "lines",
                             hoverinfo = "none", text = "",
                             name = "z-axis (right)",
                             line = list(color = "black",
                                         width = 1),
                             yaxis = "y")

    # z-axis ticks
    major.ticks.x <- c(data$xlim[2] * 1/0.75,
                       (1 + 0.015 * layout$abanico$dimension$ztcl / 100) *
                         data$xlim[2] * 1/0.75)
    minor.ticks.x <- c(data$xlim[2] * 1/0.75,
                       (1 + 0.01 * layout$abanico$dimension$ztcl / 100) *
                         data$xlim[2] * 1/0.75)
    major.ticks.y <- (tick.values.major - z.central.global) *  min(ellipse[ ,1])
    minor.ticks.y <- (tick.values.minor - z.central.global) *  min(ellipse[ ,1])

    # major z-tick lines
    for (i in 1:length(major.ticks.y)) {
      major.tick <- data.frame(x = major.ticks.x, y = rep(major.ticks.y[i], 2))
      IAP <- plotly::add_trace(IAP, data = major.tick,
                               x = ~x, y = ~y, showlegend = FALSE,
                               type = "scatter", mode = "lines",
                               hoverinfo = "none", text = "",
                               line = list(color = "black",
                                           width = 1),
                               yaxis = "y")
    }

    # minor z-tick lines
    for (i in 1:length(minor.ticks.y)) {
      minor.tick <- data.frame(x = minor.ticks.x, y = rep(minor.ticks.y[i], 2))
      IAP <- plotly::add_trace(IAP, data = minor.tick,
                               x = ~x, y = ~y, showlegend = FALSE,
                               type = "scatter", mode = "lines",
                               hoverinfo = "none", text = "",
                               line = list(color = "black",
                                           width = 1),
                               yaxis = "y")
    }

    # z-tick label
    tick.text <- paste(" ", exp(tick.values.major))
    tick.pos <- data.frame(x = major.ticks.x[2],
                           y = major.ticks.y)

    IAP <- plotly::add_trace(IAP, data = tick.pos,
                             x = ~x, y = ~y, showlegend = FALSE,
                             hoverinfo = "none",
                             text = tick.text, textposition = "right",
                             type = "scatter", mode = "text",
                             yaxis = "y")

    # Central Line ----
    central.line <- data.frame(x = c(-100, data$xlim[2]*1/0.75), y = c(0, 0))
    central.line.text <- paste0("Central value: ",
                                format(exp(z.central.global), digits = 2, nsmall = 1))

    IAP <- plotly::add_trace(IAP, data = central.line,
                             x = ~x, y = ~y, name = "Central line",
                             type = "scatter", mode = "lines",
                             hoverinfo = "text", text = central.line.text,
                             yaxis = "y",
                             line = list(color = "black",
                                         width = 0.5,
                                         dash = 2))

    # KDE plot ----
    KDE.x <- xy.0[1] + KDE[[1]][ ,2] * KDE.scale
    KDE.y <- (KDE[[1]][ ,1] - z.central.global) * min(ellipse[,1])
    KDE.curve <- data.frame(x = KDE.x, y = KDE.y)
    KDE.curve <- KDE.curve[KDE.curve$x != xy.0[1], ]
    KDE.text <- paste0("Value:",
                       format(exp(KDE.curve$x), digits = 2, nsmall = 1), "<br />",
                       "Density:",
                       format(KDE.curve$y, digits = 2, nsmall = 1))

    IAP <- plotly::add_trace(IAP, data = KDE.curve,
                             x = ~x, y = ~y, name = "KDE",
                             type = "scatter", mode = "lines",
                             hoverinfo = "text",
                             text = KDE.text,
                             line = list(color = "red"),
                             yaxis = "y")

    # set layout ----
    IAP <- plotly::layout(IAP,
                          hovermode = "closest",
                          dragmode = "pan",
                          xaxis = list(range = c(data$xlim[1], data$xlim[2] * 1/0.65),
                                       zeroline = FALSE,
                                       showgrid = FALSE,
                                       tickmode = "array",
                                       tickvals = x.axis.ticks),
                          yaxis = list(range = data$ylim,
                                       zeroline = FALSE,
                                       showline = FALSE,
                                       showgrid = FALSE,
                                       tickmode = "array",
                                       tickvals = c(-2, 0, 2)),
                          shapes = list(list(type = "rect", # 2 sigma bar
                                             x0 = 0, y0 = -2,
                                             x1 = bars.x[3], y1 = 2,
                                             xref = "x", yref = "y",
                                             fillcolor = "grey",
                                             opacity = 0.2))
    )

    # show and return interactive plot ----
    #print(plotly::subplot(IAP, IAP.kde))
    print(IAP)
    return(IAP)
  }

  ## create and return numeric output
  invisible(plot.output)
}

#'@title Import TIFF Image Data into R
#'
#' @description
#' Simple wrapper around [tiff::readTIFF] to import TIFF images and TIFF image
#' stacks to be further processed within the 'Luminescence' package.
#'
#' @param file [character] (**required**):
#' name of the TIFF file to read (URLs are supported).
#'
#'@param merge2stack [logical] (*with default*): if `file` is a [list] it merges
#'the individual images into one image stack. Please note that the smallest image
#'dimension determines pixel dimension of the output stack.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#'@param ... not in use, for compatibility reasons only
#'
#' @return
#' Returns an [Luminescence::RLum.Data.Image-class] object. Results are
#' returned as a list when multiple files are processed or `file` is a list.
#'
#'@author Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @section Function version: 0.2.1
#'
#'@seealso [tiff::readTIFF], [Luminescence::RLum.Data.Image-class]
#'
#'@keywords IO
#'
#'@examples
#'
#'## use system file
#'file <- system.file("extdata", "TIFFfile.tif", package = "Luminescence")
#'
#'## import image
#'image <- read_TIFF2R(file)
#'
#'@export
read_TIFF2R <- function(
  file,
  merge2stack = FALSE,
  verbose = TRUE,
  ...
) {
  .set_function_name("read_TIFF2R")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .require_suggested_package("tiff", "Importing TIFF files")
  .validate_logical_scalar(verbose)
  file <- .validate_file(file, ext = c("tiff", "tif"), pattern = "\\.tiff?$")

  # Self call ---------------------------------------------------------------
  if(inherits(file, "list") || length(file) > 1) {
    .validate_logical_scalar(merge2stack)

    ## read list
    tmp <- lapply(file, read_TIFF2R, verbose = verbose)

    if (merge2stack) {
      ## because we don't know what we get, we determine the minimal dimensions
      t_dim <- vapply(tmp, function(x) dim(x@data), numeric(3))
      t_range <- matrixStats::rowMins(t_dim)

      ## the image stack cannot be bigger than the smallest image
      tmp <- lapply(tmp, function(x) x@data[1:t_range[1],1:t_range[2],])
      tmp <- array(unlist(tmp), dim = c(t_range[1], t_range[2], length(tmp)))
      tmp <- set_RLum(class = "RLum.Data.Image", data = tmp)
    }

    ## return the results
    return(tmp)
  }

  ## Import -----------------------------------------------------------------

  if (verbose) {
    cat("\n[read_TIFF2R()] Importing ...")
    cat("\n path: ", dirname(file))
    cat("\n file: ", .shorten_filename(basename(file)))
    cat("\n")
  }

  ## import
  temp <- tiff::readTIFF(file, all = TRUE, as.is = TRUE)

  if (inherits(temp, "list"))
    temp <- as(temp, "RLum.Data.Image")

# Return ------------------------------------------------------------------
  set_RLum(class = "RLum.Data.Image", data = temp@data)
}

## load data
file <- system.file("extdata", "TIFFfile.tif", package = "Luminescence")

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(read_TIFF2R(data.frame()),
               "'file' should be of class 'character' or 'list'")
  expect_error(read_TIFF2R(character(0)),
               "'file' cannot be an empty character")
  expect_error(object = read_TIFF2R(file = "text"),
               "File '.*text' does not exist") # windows CI needs the regexp
  expect_error(read_TIFF2R(list(file), merge2stack = NA),
               "'merge2stack' should be a single logical value")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  SW({
  expect_s4_class(read_TIFF2R(file), "RLum.Data.Image")

  ## check image stack
  t <- expect_s4_class(read_TIFF2R(list(file, file), merge2stack = TRUE), "RLum.Data.Image")
  expect_equal(dim(t@data)[3], 2)

  t <- expect_type(read_TIFF2R(list(file, file), merge2stack = FALSE), "list")
  expect_length(t, 2)

  t <- expect_type(read_TIFF2R(c(file, file), merge2stack = FALSE), "list")
  expect_length(t, 2)
  })
})

#' Class `"RLum.Data.Image"`
#'
#' Class for representing luminescence image data (TL/OSL/RF). Such data are
#' for example produced by function [Luminescence::read_SPE2R].
#'
#' @name RLum.Data.Image-class
#'
#' @docType class
#'
#' @slot recordType
#' Object of class [character] containing the type of the curve (e.g.
#' "OSL image", "TL image").
#'
#' @slot curveType
#' Object of class [character] containing curve type, allowed values
#' are measured or predefined
#'
#' @slot data
#' Object of class [array] containing image data.
#'
#' @slot info
#' Object of class [list] containing further meta information objects
#'
#' @note
#' The class should only contain data for a set of images. For additional
#' elements the slot `info` can be used.
#'
#' @section Objects from the class:
#' Objects can be created by calls of the form `set_RLum("RLum.Data.Image", ...)`.
#'
#' @section Class version: 0.5.1
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum-class], [Luminescence::RLum.Data-class], [Luminescence::plot_RLum],
#' [Luminescence::read_SPE2R], [Luminescence::read_TIFF2R]
#'
#' @keywords classes
#'
#' @examples
#'
#' showClass("RLum.Data.Image")
#'
#' ##create empty RLum.Data.Image object
#' set_RLum(class = "RLum.Data.Image")
#'
#' @export
setClass(
  "RLum.Data.Image",
  slots = list(
    recordType = "character",
    curveType = "character",
    data = "array",
    info = "list"
  ),
  contains = "RLum.Data",
  prototype = list(
    recordType = character(),
    curveType = character(),
    data = array(),
    info = list()
  )
)

## as() ---------------------------------------------------------------------
##DATA.FRAME
##COERCE RLum.Data.Image >> data.frame AND data.frame >> RLum.Data.Image
#' as()
#'
#' for `[RLum.Data.Image-class]`
#'
#' **[Luminescence::RLum.Data.Image-class]**
#'
#' \tabular{ll}{
#'  **from** \tab **to**\cr
#'   `data.frame` \tab `data.frame`\cr
#'   `matrix` \tab `matrix`
#' }
#'
#' @name as
## from data.frame ----
setAs("data.frame", "RLum.Data.Image",
      function(from,to){
        new(to,
            recordType = "unknown curve type",
            curveType = "NA",
            data = array(unlist(from), dim = c(nrow(from),ncol(from),1)),
            info = list())
      })

## to data.frame ----
setAs("RLum.Data.Image", "data.frame",
      function(from) {
        if (dim(from@data)[3] != 1) {
          stop("No viable coercion to data.frame, object contains multiple frames",
               call. = FALSE)
        }
        as.data.frame(from@data[, , 1])
      })


## from matrix   ----
setAs("matrix", "RLum.Data.Image",
      function(from,to){
        new(to,
            recordType = "unknown curve type",
            curveType = "NA",
            data = array(from, c(nrow(from), ncol(from), 1)),
            info = list())
      })

## to matrix ----
setAs("RLum.Data.Image", "matrix",
      function(from){
        if (dim(from@data)[3] != 1) {
          stop("No viable coercion to matrix, object contains multiple frames, ",
               "please convert to array instead", call. = FALSE)
        }
        from@data[, , 1, drop = TRUE]
      })

## from array ----
setAs("array", "RLum.Data.Image",
      function(from, to){
        new(to,
            recordType = "unknown curve type",
            curveType = "NA",
            data = from,
            info = list())
      })

## to array ----
setAs("RLum.Data.Image", "array",
      function(from) from@data)


## from list ----
setAs("list", "RLum.Data.Image",
      function(from, to){
        if (length(from) == 0)
          return(set_RLum("RLum.Data.Image"))

        array_list <- lapply(from, function(x) array(unlist(as.vector(x)), c(nrow(x), ncol(x), 1)))
        new(to,
            recordType = "unknown curve type",
            curveType = "NA",
            data = array(unlist(array_list),
                         c(nrow(array_list[[1]]), ncol(array_list[[1]]), length(array_list))),
            info = list())
      })

## to list ----
setAs("RLum.Data.Image", "list",
      function(from){
        num.images <- dim(from@data)[3]
        if (is.na(num.images))
          return(list())
        lapply(1:num.images, function(x) from@data[,,x])
      })

## show() -------------------------------------------------------------------
#' @describeIn show
#' Show the structure of `RLum.Data.Image` objects.
#'
#' @export
setMethod("show",
          signature(object = "RLum.Data.Image"),
          function(object){

            ## get dimension
            dim <- dim(object@data)

            ##print information
            cat("\n [RLum.Data.Image-class]")
            cat("\n\t recordType:", object@recordType)
            cat("\n\t curveType:",  object@curveType)
            cat("\n\t .. recorded frames:", max(1,dim[3], na.rm = TRUE))
            cat("\n\t .. .. pixel per frame:", dim[1]*dim[2])
            cat("\n\t .. .. x dimension [px]:", dim[1])
            cat("\n\t .. .. y dimension [px]:", dim[2])
            cat("\n\t .. .. full pixel value range:", paste(format(range(object@data), scientific = TRUE, digits = 2), collapse=" : "))
            cat("\n\t additional info elements:", length(object@info))
            #cat("\n\t\t >> names:", names(object@info))
            cat("\n")
          }
)


## set_RLum() ---------------------------------------------------------------
#' @describeIn set_RLum
#' Construction method for [Luminescence::RLum.Data.Image-class] objects.
#'
#' @export
setMethod(
  "set_RLum",
  signature = signature("RLum.Data.Image"),
  definition = function(
    class,
    originator,
    .uid,
    .pid,
    recordType = "Image",
    curveType = NA_character_,
    data = array(),
    info = list()) {

    ##The case where an RLum.Data.Image object can be provided
    ##with this RLum.Data.Image objects can be provided to be reconstructed
    if (inherits(data, "RLum.Data.Image")) {
      ##check for missing curveType
      if (missing(curveType))
        curveType <- data@curveType

      ##check for missing recordType
      if (missing(recordType))
        recordType <- data@recordType

      ##check for missing data ... not possible as data is the object itself

      ##check for missing info
      if (missing(info))
        info <- data@info

      ##check for modified .uid & .pid
      ## >> this cannot be changed here, since both would be reset, by
      ## the arguments passed down from set_RLum() ... the generic function

      ##set empty class form object
      newRLumDataImage <- new("RLum.Data.Image")

      ##fill - this is the faster way, filling in new() costs ...
      newRLumDataImage@originator <- data@originator
      newRLumDataImage@recordType <- recordType
      newRLumDataImage@curveType <- curveType
      newRLumDataImage@data <- data@data
      newRLumDataImage@info <- info
      newRLumDataImage@.uid <- data@.uid
      newRLumDataImage@.pid <- data@.pid

    } else {
      ##set empty class from object
      newRLumDataImage <- new("RLum.Data.Image")

      ##fill - this is the faster way, filling in new() costs ...
      newRLumDataImage@originator <- originator
      newRLumDataImage@recordType <- recordType
      newRLumDataImage@curveType <- curveType
      newRLumDataImage@data <- data
      newRLumDataImage@info <- info
      newRLumDataImage@.uid <- .uid
      newRLumDataImage@.pid <- .pid
    }
    return(newRLumDataImage)
  }
)

## get_RLum() ---------------------------------------------------------------
#' @describeIn get_RLum
#' Accessor method for [Luminescence::RLum.Data.Image-class] objects.
#' The argument `info.object` is optional to directly access the info elements.
#' If no info element name is provided, the raw image data (`array`) will be
#' returned.
#'
#' @export
setMethod("get_RLum",
          signature("RLum.Data.Image"),
          definition = function(object, info.object = NULL) {
            .set_function_name("get_RLum")
            on.exit(.unset_function_name(), add = TRUE)

            .validate_class(info.object, "character", null.ok = TRUE, length = 1)
            if (!is.null(info.object)) {
              if(info.object %in% names(object@info)){
                unlist(object@info[info.object])
              } else {
                .throw_error("Invalid element name, valid names are: ",
                             .collapse(names(object@info)))
             }
            } else {
              object@data
            }
          })


# names_RLum() --------------------------------------------------------------------------------
#' @describeIn names_RLum
#' Returns the names of the info elements stored in the object.
#'
#' @export
setMethod(
  "names_RLum",
  "RLum.Data.Image",
  function(object) names(object@info))

## normalise_RLum() --------------------------------------------------------------
#' @describeIn normalise_RLum
#' Normalise [Luminescence::RLum.Data.Image-class] objects to value set via
#' the argument `norm`.
#'
#' @param global [logical] (*with default): this defines whether the normalisation
#' is applied globally (same to all) or locally, in which case each frame has its
#' own normalisation. If `global = TRUE` the arguments for `norm  = 'first'` and
#' `norm = 'last'` work as expected and consider either the first or the last
#' frame for the normalisation.
#'
#'  Normalise [Luminescence::RLum.Data.Image-class] objects to value set via
#' the argument `norm`
#'
#' @export
setMethod(
  f = "normalise_RLum",
  signature = "RLum.Data.Image",
  function(object, norm = TRUE, global = TRUE) {
    ## for frames we use the last frame
    if(global) {
       if(norm == "last") {
         object@data[] <- object@data / array(
           object@data[,,dim(object@data)[3]], dim = dim(object@data))

       } else if (norm == "first") {
         object@data[] <- object@data / array(
           object@data[,,1], dim = dim(object@data))

       } else {
         object@data[] <- .normalise_curve(object@data[], norm = norm)

       }

    } else {
      object@data[] <- apply(
        object@data,
        MARGIN = 3,
        FUN = .normalise_curve,
        norm = norm)
    } ## end global

    return(object)
  })

#' @title Fitting Thermal Quenching Data
#'
#' @description Applying a nls-fitting to thermal quenching data.
#'
#' @details
#'
#' **Used equation**\cr
#'
#' The equation used for the fitting is
#'
#' \deqn{y = A / (1 + C * \exp(-W / (k_B x))) + c}
#'
#' *W* is the energy depth in eV, *C* is a dimensionless constant, *A* and *c*
#' are used to adjust the curve for the given signal, \eqn{k_B} is the Boltzmann
#' constant in eV/K and *x* is the absolute temperature in K.
#'
#' **Error estimation**\cr
#'
#' The error estimation is done by varying the input parameters using the
#' given uncertainties in a Monte Carlo simulation. Errors are assumed to
#' follow a normal distribution.
#'
#' **`start_param`** \cr
#'
#' The function allows the injection of starting values for the parameters to
#' be optimised via the `start_param` argument. The parameters must be provided
#' as a named list.
#' Examples: `start_param = list(A = 1, C = 1e+5, W = 0.5, c = 0)`
#'
#' **`method_control`** \cr
#'
#' The following arguments can be provided via `method_control`. Please note that arguments provided
#' via `method_control` are not further tested, i.e., if the function crashes your input was probably
#' wrong.
#'
#' \tabular{lll}{
#' **ARGUMENT** \tab **TYPE** \tab **DESCRIPTION**\cr
#' `upper` \tab named [vector] \tab sets upper fitting boundaries, if provided boundaries for all arguments
#' are required, e.g., `c(A = 0, C = 0, W = 0, c = 0)` \cr
#' `lower` \tab names [vector] \tab set lower fitting boundaries (see `upper` for details) \cr
#' `trace`   \tab [logical] \tab enable/disable progression trace for [minpack.lm::nlsLM]\cr
#'  `weights` \tab [numeric] \tab option to provide own weights for the fitting, the length of this
#'  vector needs to be equal to the number for rows of the input `data.frame`. If set to `NULL` no weights
#'  are applied. The weights are defined by the third column of the input `data.frame`.
#' }
#'
#' @param data [data.frame] (**required**): input data with three columns, the first column contains
#' temperature values in deg. C, columns 2 and 3 the dependent values with its error
#'
#' @param start_param [list] (*optional*): option to provide the start
#' parameters for the fitting, see details
#'
#' @param method_control [list] (*optional*): further options to fine tune
#' the fitting, see details for
#' further information
#'
#' @param n.MC [integer] (*with default*): number of Monte Carlo runs for the
#' error estimation. If `n.MC` is `NULL` or 1, the error estimation is skipped.
#'
#' @param verbose [logical] (*with default*): enable/disable output to the
#' terminal.
#'
#' @param plot [logical] (*with default*): enable/disable the plot output.
#'
#' @param ... further arguments that can be passed to control the plotting, support are `main`, `pch`,
#' `col_fit`, `col_points`, `lty`, `lwd`, `xlab`, `ylab`, `xlim`, `ylim`, `xaxt`
#'
#' @return
#'
#' The function returns numerical output and an (*optional*) plot.
#'
#' -----------------------------------\cr
#' `[ NUMERICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' **`RLum.Results`**-object
#'
#' **slot:** **`@data`**
#'
#' `[.. $data : data.frame]`\cr
#'
#' A table with all fitting parameters and the number of Monte Carlo runs used
#' for the error estimation (this may be smaller that `n.MC`).
#'
#' `[.. $fit : nls object]` \cr
#'
#'  The nls [stats::nls] object returned by the function [minpack.lm::nlsLM]. This object
#'  can be further passed to other functions supporting an nls object (cf. details section
#'  in [stats::nls])
#'
#' **slot:** **`@info`**
#'
#' `[.. $call : call]`\cr
#'
#' The original function call.
#'
#' -----------------------------------\cr
#' `[ GAPHICAL OUTPUT ]`\cr
#' -----------------------------------\cr
#'
#' Plotted are temperature against the signal and their uncertainties.
#' The fit is shown as dashed-line (can be modified). Please note that for the fitting the absolute
#' temperature values are used but are re-calculated to deg. C for the plot.
#'
#' @section Function version: 0.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @references
#'
#' Wintle, A.G., 1975. Thermal Quenching of Thermoluminescence in Quartz. Geophys. J. R. astr. Soc. 41, 107–113.
#'
#' @seealso [minpack.lm::nlsLM]
#'
#' @examples
#'
#' ##create short example dataset
#' data <- data.frame(
#'   T = c(25, 40, 50, 60, 70, 80, 90, 100, 110),
#'   V = c(0.06, 0.058, 0.052, 0.051, 0.041, 0.034, 0.035, 0.033, 0.032),
#'   V_X = c(0.012, 0.009, 0.008, 0.008, 0.007, 0.006, 0.005, 0.005, 0.004))
#'
#' ##fit
#' fit_ThermalQuenching(
#'  data = data,
#'  n.MC = NULL)
#'
#' @export
fit_ThermalQuenching <- function(
  data,
  start_param = list(),
  method_control = list(),
  n.MC = 100,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("fit_ThermalQuenching")
  on.exit(.unset_function_name(), add = TRUE)

  # Self-call -----------------------------------------------------------------------------------
  if(inherits(data, "list")){
    args <- as.list(match.call())
    args[[1]] <- NULL
    args$data <- NULL

    ##run function
    results_list <- lapply(data, function(x){
       do.call(fit_ThermalQuenching, c(list(data = x),args))
    })

    ##combine and return
    return(merge_RLum(results_list))
  }

  ## Integrity checks -------------------------------------------------------
  .validate_class(data, "data.frame",
                  extra = "a 'list' of such objects")

  if(nrow(data) < 1 || ncol(data) < 3)
    .throw_error("'data' is empty or has fewer than three columns")

  if (ncol(data) > 3) {
    .throw_warning("'data' has more than 3 columns, taking only the first three")
    data <- data[, 1:3]
  }

  if (anyNA(data)) {
    .throw_warning("NA values in 'data' automatically removed")
    data <- na.exclude(data)
    if (nrow(data) == 0)
      .throw_error("After NA removal, nothing is left from the data set")
  }
  if (nrow(data) < 4) {
    .throw_error("Insufficient number of data points to fit the model")
  }

  .validate_class(start_param, "list")
  .validate_class(method_control, "list")
  n.MC <- max(.validate_positive_scalar(n.MC, int = TRUE, null.ok = TRUE), 1)

  ## Prepare data -----------------------------------------------------------
  ##
  ## Set formula for quenching according to Wintle 1973.
  ## Here we add a constant, otherwise the fit will not really work.
  ##
  ## The actual implementation adopts a slight reformulation, which is
  ## mathematically equivalent to the one presented in the documentation
  ## above, but is numerically more stable:
  ##
  ## y = A / (1 + C * exp(-W / (kB * x))) + c
  ##   = A / (1 + exp(log(C)) * exp(-W / (kB * x))) + c
  ##   = A / (1 + exp(log(C) - W / (kB * x))) + c
  ##   = A / (1 + exp(C' - W / (kB * x))) + c
  ##
  ## For backward compatibility, we keep using the name C for the logged
  ## parameter, so that the user will keep seeing values in the original scale,
  ## although internally we operate in log-space for C. When we show the
  ## final results, we'll have to exponentiate the fitted parameter.
  ##
  kB <- .const$kB
  f <- y ~ A / (1 + exp(C - W / (kB * x))) + c

  ##set translate values in data.frame to absolute temperature
  data[[1]] <- data[[1]] + .const$C2K

  ##start parameter
  start_param <- modifyList(x = list(
    A = max(data[[2]]),
    C = log(max(data[[1]]) * 1e5),
    W = 0.5,
    c = 0),
    val = start_param)

  ##method control
  method_control <- modifyList(
    x = list(
     lower = c(A = 0, C = 0, W = 0, c = -Inf),
     upper = c(A = 10 * start_param$A,
               C = log(max(data[[1]]) * 1e12),
               W = 10, c = 10 * start_param$A),
     trace = FALSE,
     weights = data[[3]]
    ),
    val = method_control)

  weights <- method_control$weights %||% rep(1, nrow(data))

  # Fitting -------------------------------------------------------------------------------------
  ##guine fitting
  fit <- try(minpack.lm::nlsLM(
    formula = f,
    data = data.frame(x = data[[1]], y = data[[2]]),
    weights = weights,
    control = list(
      maxiter = 500,
      maxfev = 1000,
      trace = method_control$trace),
    start = start_param,
    lower = method_control$lower,
    upper = method_control$upper
  ), silent = TRUE)

  ##only continue if the first fitting worked out
  if (inherits(fit, "try-error")) {
    .throw_message("Fitting failed, NULL returned")
    return(NULL)
  }

    ##Prepare MC runs for the fitting
    x_MC <- data[[1]]
    y_MC <- matrix(
      data = data[[2]] + rnorm(n.MC * length(x_MC)) * data[[3]],
      nrow = length(x_MC),
      ncol = n.MC)
    y_MC[y_MC < 0] <- 0

  ## run fitting
  fit_MC <- lapply(1:n.MC, function(x) {
      temp <- try(minpack.lm::nlsLM(
        formula = f,
        data = data.frame(x = x_MC, y = y_MC[,x]),
        weights = weights,
        control = list(
          maxiter = 500,
          maxfev = 1000
        ),
        start = start_param,
        lower = method_control$lower,
        upper = method_control$upper
      ), silent = TRUE)

      ##return value
      if (inherits(temp, "try-error"))
        return(NULL) # nocov
      temp
  })

  ## remove NULL (the fit was not successful)
  fit_MC <- .rm_NULL_elements(fit_MC)
  n.MC <- length(fit_MC)

  ## Extract values ---------------------------------------------------------

   ##(1) - extract parameters from main fit
   fit_coef <- coef(fit)
   A <- fit_coef[["A"]]
   C <- fit_coef[["C"]]
   W <- fit_coef[["W"]]
   c <- fit_coef[["c"]]

   ##(2) - extract values from MC run
   fit_coef_MC_full <- vapply(X = fit_MC, FUN = coef, FUN.VALUE = numeric(4))
   fit_coef_MC <- round(matrixStats::rowSds(fit_coef_MC_full),3)

  ## (3) - unlog the C parameter before computing the standard deviation
  fit_coef_MC[2] <- signif(sd(exp(fit_coef_MC_full[2, ])), 4)

   A_MC_X <- fit_coef_MC[1]
   C_MC_X <- fit_coef_MC[2]
   W_MC_X <- fit_coef_MC[3]
   c_MC_X <- fit_coef_MC[4]

  ## Terminal output --------------------------------------------------------
  if (verbose) {

    cat("\n[fit_ThermalQuenching()]\n\n")
    cat(" A =", A, "\u00b1", A_MC_X, "\n")
    cat(" C =", exp(C), "\u00b1", C_MC_X, "\n")
    cat(" W =", W, "\u00b1", W_MC_X, "eV\n")
    cat(" c =", c, "\u00b1", c_MC_X, "\n")
    cat(" --------------------------------\n")
  }

  ## Plotting ---------------------------------------------------------------
  if(plot) {
    ##plot settings
    plot_settings <- list(
      xlim = range(data[[1]]),
      ylim = c(min(data[[2]]) - data[[3]][which.min(data[[2]])],
               max(data[[2]]) + data[[3]][which.max(data[[2]])]),
      pch = 1,
      xaxt = "n",
      xlab = "Temperature [\u00b0C]",
      ylab = "Dependent [a.u.]",
      main = "Thermal quenching",
      lty = 2,
      col_points = "black",
      col_fit = "red",
      lwd = 1.3,
      mtext = if(n.MC == 1) "" else paste0("n.MC = ", n.MC)
    )

    ##overwrite settings
    plot_settings <- modifyList(x = plot_settings, val = list(...))

    ##create plot window
    plot(
      x = NA,
      y = NA,
      xlim = plot_settings$xlim,
      ylim = plot_settings$ylim,
      xaxt = plot_settings$xaxt,
      xlab = plot_settings$xlab,
      ylab = plot_settings$ylab,
      main = plot_settings$main
    )

    ##add axis with correct temperature
    if(!is.null(plot_settings$xaxt) && plot_settings$xaxt == "n"){
      at <- pretty(round(axTicks(side = 1) - .const$C2K))
      axis(side = 1, at = at + .const$C2K, labels = at)
    }

    ## add MC curves
    if (n.MC > 1) {
      for(i in 1:n.MC){
        A <- fit_coef_MC_full[1,i]
        C <- fit_coef_MC_full[2,i]
        W <- fit_coef_MC_full[3,i]
        c <- fit_coef_MC_full[4,i]
        x <- 0
        curve(A / (1 + exp(C - W / (kB * x))) + c,
              col = rgb(0, 0, 0, 0.1), add = TRUE)
      }
    }

    ##add points and uncertainties
    points(data[, 1:2],
           pch = plot_settings$pch,
           lwd = 2,
           col = plot_settings$col_points)
    segments(x0 = data[[1]], x1 = data[[1]],
             y0 = data[[2]] + data[[3]],
             y1 = data[[2]] - data[[3]],
             col = plot_settings$col_points
             )

    ##add central fit
    A <- fit_coef[["A"]]
    C <- fit_coef[["C"]]
    W <- fit_coef[["W"]]
    c <- fit_coef[["c"]]

    x <- 0
    curve(A / (1 + exp(C - W / (kB * x))) + c,
          lty = plot_settings$lty,
          lwd = plot_settings$lwd,
          col = plot_settings$col_fit,
          add = TRUE)

    ##add mtext
    mtext(side = 3, text = plot_settings$mtext)
  }

  ## Return -----------------------------------------------------------------
  set_RLum(
    class = "RLum.Results",
    data = list(
      data = data.frame(
          A = A,
          A_X = A_MC_X,
          C = exp(C),
          C_X = C_MC_X,
          W = W,
          W_X = W_MC_X,
          c = c,
          c_X = c_MC_X,
          n.MC = n.MC
      ),
      fit = fit
    ),
    info = list(
      call = sys.call()
    )
  )
}

#' @title Apply the finite mixture model (FMM) after Galbraith (2005) to a given De
#' distribution
#'
#' @description
#' This function fits a k-component mixture to a De distribution with differing
#' known standard errors. Parameters (doses and mixing proportions) are
#' estimated by maximum likelihood assuming that the log dose estimates come
#' from a mixture of normal distributions.
#'
#' @details
#' This model uses the maximum likelihood and Bayesian Information Criterion
#' (BIC) approaches.
#'
#' Indications of overfitting are:
#'
#' - increasing BIC
#' - repeated dose estimates
#' - covariance matrix not positive definite
#' - covariance matrix produces `NaN`
#' - convergence problems
#'
#' **Plot**
#'
#' If `n.components` is a vector (`c(k.min:k.max)`), a plot is generated
#' showing the *k* components equivalent doses as normal distributions.
#' By default `pdf.weight` is set to `TRUE`, so that the probability density
#' functions are weighted by the components proportion for each iteration of
#' *k* components, so the sum of areas of each component equals 1.
#' If `pdf.weight` is set to `FALSE`, the area under each normal distribution
#' is always 1. While the density values are on the same scale when no weights
#' are used, the y-axis are individually scaled if the probability density are
#' weighted by the components proportion.\cr
#' The standard deviation (sigma) of the normal distributions is by default
#' determined by a common `sigmab` (see `pdf.sigma`). For
#' `pdf.sigma = "se"` the standard error of each component is taken
#' instead.\cr
#' The stacked [graphics::barplot] shows the proportion of each component (in
#' per cent) calculated by the FMM. The last plot shows the achieved BIC scores
#' and maximum log-likelihood estimates for each value of *k*.
#'
#' @param data [Luminescence::RLum.Results-class] or [data.frame] (**required**):
#' for [data.frame]: two columns with De `(data[,1])` and De error `(values[,2])`
#'
#' @param sigmab [numeric] (**required**):
#' spread in De values (given as a fraction, e.g. 0.2), representing the
#' expected overdispersion in the data should the sample be well-bleached
#' (Cunningham & Wallinga 2012, p. 100).
#'
#' @param n.components [numeric] (**required**):
#' number of components to be fitted. If a vector is provided (e.g. `c(2:8)`) the
#' finite mixtures for 2, 3 ... 8 components are calculated and a plot and a
#' statistical evaluation of the model performance (BIC score and maximum
#' log-likelihood) is provided.
#'
#' @param grain.probability [logical] (*with default*):
#' prints the estimated probabilities of which component each grain is in
#'
#' @param pdf.weight [logical] (*with default*):
#' weight the probability density functions by the components proportion.
#' Ignored if `n.components` has length 1.
#'
#' @param pdf.sigma [character] (*with default*):
#' if `"sigmab"` the components normal distributions are plotted with a common standard
#' deviation (i.e. `sigmab`) as assumed by the FFM. Alternatively,
#' `"se"` takes the standard error of each component for the sigma
#' parameter of the normal distribution
#'
#' @param pdf.colors [character] (*with default*):
#' colour coding of the components in the plot.
#' Possible options are `"gray"`, `"colors"` and `"none"`.
#'
#' @param plot.proportions [logical] (*with default*):
#' plot a [graphics::barplot] showing the proportions of components.
#' Ignored if `n.components` has length 1.
#'
#' @param plot.criteria [logical] (*with default*):
#' plot the statistical criteria (BIC and log-likelihood).
#' Ignored if `n.components` has length 1.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output. This is ignored and no plot is produced if
#' `n.components` has length 1.
#'
#' @param ... other parameters to control the plot output. Supported are
#' `cex`, `main`, `main.densities`, `main.proportions`, `main.criteria`,
#' `pdf.scale`, `dose.scale`.
#'
#' @return
#' Returns a plot (*optional*) and terminal output. In addition an
#' [Luminescence::RLum.Results-class] object is returned containing the
#' following elements:
#'
#' \item{.$summary}{[data.frame] summary of all relevant model results.}
#' \item{.$data}{[data.frame] original input data}
#' \item{.$args}{[list] used arguments}
#' \item{.$call}{[call] the function call}
#' \item{.$mle}{ covariance matrices of the log likelihoods}
#' \item{.$BIC}{ BIC score}
#' \item{.$llik}{ maximum log likelihood}
#' \item{.$grain.probability}{ probabilities of a grain belonging to a component}
#' \item{.$components}{[matrix] estimates of the de, de error and proportion for each component}
#' \item{.$single.comp}{[data.frame] single component FFM estimate}
#'
#' If a vector for `n.components` is provided (e.g.  `c(2:8)`),
#' `mle` and `grain.probability` are lists containing matrices of the
#' results for each iteration of the model.
#'
#' The output should be accessed using the function [Luminescence::get_RLum].
#'
#' @section Function version: 0.4.4
#'
#' @author
#' Christoph Burow, University of Cologne (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany) \cr
#' Based on a rewritten S script of Rex Galbraith, 2006.
#'
#' @seealso [Luminescence::calc_CentralDose], [Luminescence::calc_CommonDose],
#' [Luminescence::calc_FuchsLang2001], [Luminescence::calc_MinDose]
#'
#' @references
#' Galbraith, R.F. & Green, P.F., 1990. Estimating the component
#' ages in a finite mixture. Nuclear Tracks and Radiation Measurements 17,
#' 197-206.
#'
#' Galbraith, R.F. & Laslett, G.M., 1993. Statistical models
#' for mixed fission track ages. Nuclear Tracks Radiation Measurements 4,
#' 459-470.
#'
#' Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of
#' equivalent dose and error calculation and display in OSL dating: An overview
#' and some recommendations. Quaternary Geochronology 11, 1-27.
#'
#' Roberts, R.G., Galbraith, R.F., Yoshida, H., Laslett, G.M. & Olley, J.M., 2000.
#' Distinguishing dose populations in sediment mixtures: a test of single-grain
#' optical dating procedures using mixtures of laboratory-dosed quartz.
#' Radiation Measurements 32, 459-465.
#'
#' Galbraith, R.F., 2005. Statistics for Fission Track Analysis, Chapman & Hall/CRC, Boca Raton.
#'
#' **Further reading**
#'
#' Arnold, L.J. & Roberts, R.G., 2009. Stochastic
#' modelling of multi-grain equivalent dose (De) distributions: Implications
#' for OSL dating of sediment mixtures. Quaternary Geochronology 4,
#' 204-230.
#'
#' Cunningham, A.C. & Wallinga, J., 2012. Realizing the
#' potential of fluvial archives using robust OSL chronologies. Quaternary
#' Geochronology 12, 98-106.
#'
#' Rodnight, H., Duller, G.A.T., Wintle, A.G. &
#' Tooth, S., 2006. Assessing the reproducibility and accuracy of optical
#' dating of fluvial deposits.  Quaternary Geochronology 1, 109-120.
#'
#' Rodnight, H. 2008. How many equivalent dose values are needed to obtain a
#' reproducible distribution?. Ancient TL 26, 3-10.
#'
#'
#' @examples
#'
#' ## load example data
#' data(ExampleData.DeValues, envir = environment())
#'
#' ## (1) apply the finite mixture model
#' ## NOTE: the data set is not suitable for the finite mixture model,
#' ## which is why a very small sigmab is necessary
#' calc_FiniteMixture(ExampleData.DeValues$CA1,
#'                    sigmab = 0.2, n.components = 2,
#'                    grain.probability = TRUE)
#'
#' ## (2) repeat the finite mixture model for 2, 3 and 4 maximum number of fitted
#' ## components and save results
#' ## NOTE: The following example is computationally intensive. Please un-comment
#' ## the following lines to make the example work.
#' FMM<- calc_FiniteMixture(ExampleData.DeValues$CA1,
#'                          sigmab = 0.2, n.components = c(2:4),
#'                          pdf.weight = TRUE)
#'
#' ## show structure of the results
#' FMM
#'
#' ## show the results on equivalent dose, standard error and proportion of
#' ## fitted components
#' get_RLum(object = FMM, data.object = "components")
#'
#' @export
calc_FiniteMixture <- function(
  data,
  sigmab,
  n.components,
  grain.probability = FALSE,
  pdf.weight = TRUE,
  pdf.sigma = c("sigmab", "se"),
  pdf.colors = c("gray", "colors", "none"),
  plot.proportions = TRUE,
  plot.criteria = TRUE,
  plot=TRUE,
  ...
) {
  .set_function_name("calc_FiniteMixture")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, c("data.frame", "RLum.Results"))
  if (inherits(data, "RLum.Results")) {
    if (.check_originator(data, "calc_FiniteMixture")) {
      do.call(.plot_FiniteMixture, c(data, as.list(sys.call())[-(1:2)]))
      return(invisible(data))
    }
    data <- get_RLum(data, "data")
  }
  if (ncol(data) < 2) {
    .throw_error("'data' object must have two columns")
  }
  if (anyNA(data[, 1]) || any(data[, 1] == 0)) {
    .throw_error("'data' must have only positive values in its De column")
  }
  .validate_positive_scalar(sigmab)
  if (sigmab < 0 || sigmab > 1) {
    .throw_error("'sigmab' must be a value between 0 and 1")
  }
  .validate_class(n.components, c("integer", "numeric"))
  if (length(n.components) == 1 && n.components < 2) {
    .throw_error("'n.components' should be at least 2")
  }
  .validate_logical_scalar(grain.probability)
  .validate_logical_scalar(pdf.weight)
  pdf.sigma <- .validate_args(pdf.sigma, c("sigmab", "se"))
  pdf.colors <- .validate_args(pdf.colors, c("gray", "colors", "none"))
  .validate_logical_scalar(plot.proportions)
  .validate_logical_scalar(plot.criteria)
  .validate_logical_scalar(plot)

  ## ensure that the chosen components are sorted
  n.components <- setdiff(n.components, 1)
  n.components <- sort(n.components)
  multiple.components <- length(n.components) > 1

  ## set expected column names
  colnames(data)[1:2] <- c("ED", "ED_Error")

  ## ... ARGUMENTS ------------
  ##============================================================================##

  extraArgs <- list(...)

  ## console output
  verbose <- TRUE
  if("verbose" %in% names(extraArgs)) {
    verbose <- .validate_logical_scalar(extraArgs$verbose, name = "'verbose'")
  }

  ##============================================================================##
  ## CALCULATIONS
  ##============================================================================##

  ## create storage variables if more than one k is provided
  if (multiple.components) {

    # counter needed for various purposes
    cnt<- 1

    # create summary matrix containing DE, standard error (se) and proportion
    # for each component
    comp.n<- matrix(data = NA, ncol = length(n.components),
                    nrow = n.components[length(n.components)] * 3,
                    byrow = TRUE)

    # create empty vector as storage for BIC and LLIK scores
    BIC.n<- vector(mode = "double")
    LLIK.n<- vector(mode = "double")

    # create empty vectors of type "lists" as storage for mle matrices and
    # grain probabilities
    vmat.n<- vector(mode = "list", length = length(n.components))
    grain.probability.n<- vector(mode = "list", length = length(n.components))
  }

  ## calculate yu = log(ED), su = se(logED), n = number of grains
  yu <- log(data$ED)
  su <- data$ED_Error / data$ED
  n <- length(yu)

  num.iters <- 499L
  wu <- 1/(sigmab^2 + su^2)
  rwu <- sqrt(wu)

  ## start actual calculation (loop) for each provided maximum components to
  ## be fitted.
  for(i in 1:length(n.components)) {
    k<- n.components[i]

    # compute starting values
    pui<- matrix(0,n,k)
    nui<- matrix(0,n,k)
    pii<- rep(1/k,k)
    mu<- min(yu) + (max(yu)-min(yu))*(1:k)/(k+1)

    # remove the # in the line below to get alternative starting values
    # (useful to check that the algorithm converges to the same values)
    #	mu<- quantile(yu,(1:k)/(k+1))

    # compute maximum log likelihood estimates
    for (j in 1:num.iters) {
      for(i in 1:k)
      {
        fui <- rwu * exp(-0.5 * wu * (yu - mu[i])^2)
        nui[, i] <- pii[i] * fui
      }
      pui <- nui / rowSums(nui)
      mu <- colSums(wu * yu * pui) / colSums(wu * pui)
      pii <- colMeans(pui)
    }

    # calculate the log likelihood and BIC
    llik <- sum(log( 1 / sqrt(2 * pi) * rowSums(nui) ))
    bic<- -2*llik + (2*k - 1)*log(n)

    # calculate the covariance matrix and standard errors of the estimates
    # i.e., the dose estimates in Gy and relative standard errors, and
    # the mixing proportions and standard errors.
    aui<- matrix(0,n,k)
    bui<- matrix(0,n,k)
    for(i in 1:k)
    {
      aui[,i]<- wu*(yu-mu[i])
      bui[, i] <- -wu + aui[, i]^2
    }
    delta<- diag(rep(1,k))

    Au<- matrix(0,k-1,k-1)
    Bu<- matrix(0,k-1,k)
    Cu<- matrix(0,k,k)

    for(i in 1:(k-1)){ for(j in 1:(k-1)){
      Au[i,j]<- sum( (pui[,i]/pii[i] - pui[,k]/pii[k])*(pui[,j]/pii[j] -
                                                          pui[,k]/pii[k]) )}}

    for(i in 1:(k-1)){ for(j in 1:k){
      Bu[i,j]<- sum( pui[,j]*aui[,j]*(pui[,i]/pii[i] - pui[,k]/pii[k] -
                                        delta[i,j]/pii[i] + delta[k,j]/pii[k] ) )}}

    for(i in 1:k){ for(j in 1:k){
      Cu[i,j]<- sum( pui[,i]*pui[,j]*aui[,i]*aui[,j] - delta[i,j]*bui[,i]*
                       pui[,i] ) }}

    invvmat<- rbind(cbind(Au,Bu),cbind(t(Bu),Cu))
    vmat<- solve(invvmat, tol=.Machine$double.xmin)

    # calculate DE, relative standard error, standard error
    dose<- exp(mu)
    re <- suppressWarnings(sqrt(diag(vmat)))[-(1:(k-1))]
    re[is.nan(re)] <- NA

    sed<- dose*re
    estd<- rbind(dose,re,sed)

    # this calculates the proportional standard error of the proportion of grains
    # in the fitted components. However, the calculation is most likely erroneous.
    # rek<- sqrt(sum(vmat[1:(k-1),1:(k-1)]))
    # sep<-  c(sqrt(diag(vmat))[c(1:(k-1))],rek)

    # rename proportion
    estp <- pii

    # merge results to a data frame
    blk<- rep("    ",k)
    comp<- rbind(blk,round(estd,4),blk,round(estp,4))
    comp<- data.frame(comp,row.names=c("","dose (Gy)    ","rse(dose)    ",
                                       "se(dose)(Gy)"," ","proportion   "))
    names(comp) <- paste0("comp", 1:k)

    # calculate the log likelihood and BIC for a single component -- can
    # be useful to see if there is evidence of more than one component
    mu0<- sum(wu*yu)/sum(wu)
    fu0<-  rwu*exp(-0.5*wu*(yu-mu0)^2)
    L0<- sum( log((1/sqrt(2*pi))*fu0 ) )
    bic0<- -2*L0 + log(n)
    comp0<- round(c(exp(mu0),sigmab,L0,bic0),4)

    ## save results for k components in storage variables
    if (multiple.components) {

      # vector of indices needed for finding the dose rows of the summary
      # matrix - position 1,4,7...n
      pos.n<- seq(from = 1, to = n.components[cnt]*3, by = 3)

      # save results of each iteration to summary matrix
      for(i in 1:n.components[cnt]) {
        ## store De, SE, Proportion
        comp.n[pos.n[i] + 0:2, cnt] <- round(c(dose[i], sed[i], estp[i]), 4)
      }

      # save BIC and llik of each iteration to corresponding vector
      BIC.n[cnt]<- bic
      LLIK.n[cnt]<- llik

      # save mle matrix and grain probabilities to corresponding vector
      vmat.n[[cnt]]<- vmat
      grain.probability.n[[cnt]]<- as.data.frame(pui)

      # increase counter by one for next iteration
      cnt<- cnt+1
    }#EndOf::save intermediate results
  }##EndOf::calculation loop

  ##============================================================================##
  ## STATISTICAL CHECK
  ##============================================================================##

  if (multiple.components) {

    ## Likelihood ratio test: the difference in deviance is compared to a
    ## chi-squared distribution with degrees of freedom corresponding to the
    ## difference in number of parameters (number of components) to obtain a
    ## p-value.
    ## We also consider the likelihood of the single component model, so we
    ## can determine the significance of the 2-component model.
    LRT <- -2 * (c(L0, LLIK.n) - c(LLIK.n, 0))[1:length(n.components)]
    dfs <- diff(c(1, n.components))
    p.vals <- 1 - stats::pchisq(LRT, dfs)

    ## Add significance stars
    stars <- as.character(cut(p.vals,
                              breaks = c(-Inf, 0.001, 0.01, 0.05, 0.1, Inf),
                              labels = c("***", "**", "*", ".", " ")))

    ## merge BIC and llik scores to a single data frame, including those
    ## for the single component model
    results.n <- rbind(BIC = round(c(bic0, BIC.n), 3),
                       loglik = round(c(L0, LLIK.n), 3),
                       signif = c("", stars))
    colnames(results.n) <- c(1, n.components)
  }

  ## OUTPUT ---------
  ##============================================================================##
  if(verbose) {

    ## HEADER (always printed)
    cat("\n [calc_FiniteMixture]")

    # general information on sample and model performance
    cat("\n\n----------- meta data ------------")
    cat("\n n:                    ", n)
    cat("\n sigmab:               ", sigmab)
    cat("\n number of components: ", .collapse(n.components, quote = FALSE))

    ## OUTPUT WHEN ONLY ONE VALUE FOR n.components IS PROVIDED
    if (!multiple.components) {
      cat("\n llik:                 ", round(llik,4))
      cat("\n BIC:                  ", round(bic,3))

      # covariance matrix
      cat("\n\n--- covariance matrix of mle's ---\n\n")
      print(round(vmat,6))

      # fitted components
      cat("\n\n----------- components -----------\n\n")
      print(comp)

      # print (to 2 decimal places) the estimated probabilities of which component
      # each grain is in -- sometimes useful for diagnostic purposes
      if (grain.probability) {
        cat("\n-------- grain probability -------\n\n")
        print(round(pui,2))
      }
    }#EndOf::Output for length(n.components) == 1

    ## output for single component
    cat("\n\n-------- single component --------")
    cat("\n mu:                    ", comp0[1])
    cat("\n sigmab:                ", comp0[2])
    cat("\n llik:                  ", comp0[3])
    cat("\n BIC:                   ", comp0[4])
    cat("\n\n")

    ##----------------------------------------------------------------------------
    ## OUTPUT WHEN ONLY >1 VALUE FOR n.components IS PROVIDED
    if (multiple.components) {

      ## final labeling of component and BIC/llik matrices

      ## create row labels in correct order (dose, se, prop)
      prefix <- paste0("c", 1:n.components[length(n.components)])
      n.lab <- unlist(lapply(prefix,
                             function(x) paste0(x, c("_dose", "_se", "_prop"))))

      # label columns and rows of summary matrix and BIC/LLIK data frame
      colnames(comp.n) <- n.components
      rownames(comp.n) <- n.lab

      ## CONSOLE OUTPUT

      # print component matrix
      cat("\n---------- k components ----------\n")
      print(round(comp.n, 2), na.print = "<NA>")

      # print BIC scores and LLIK estimates
      cat("\n------ statistical criteria ------\n")
      print(results.n, quote = FALSE, right = TRUE)

      ## print evaluation of statistical criteria
      # lowest BIC score
      cat("\n Lowest BIC score for k =", n.components[which.min(BIC.n)])

      # first significant increase in LLIK estimates
      if (!any(p.vals < 0.05)) {
        cat("\n No significant increase in maximum log-likelihood estimates.\n")
      } else {
        cat("\n First significant increase in maximum log-likelihood for",
            "k =", n.components[(p.vals < 0.05)][1], "\n")
      }

      cat("\n")
    }#EndOf::Output for length(n.components) > 1
  }

  ## RETURN VALUES --------
  ##============================================================================##

  # .@data$meta
  BIC <- data.frame(n.components = n.components,
                    BIC = if (multiple.components) BIC.n else bic)
  llik <- data.frame(n.components = n.components,
                     llik = if (multiple.components) LLIK.n else llik)

  # .@data$components
  comp.re<- t(rbind(round(estd,4),round(estp,4)))
  colnames(comp.re)<- c("de","rel_de_err","de_err","proportion")
  comp.re<- comp.re[,-2] # remove the relative error column

  # .@data$grain.probability
  grain.probability<- round(pui, 2)

  summary<- data.frame(comp.re)
  call<- sys.call()
  args<- list(sigmab = sigmab, n.components = n.components)

  # create S4 object
  results <- set_RLum(
    class = "RLum.Results",
    originator = "calc_FiniteMixture",
    data = list(
      summary=summary,
      data=data,
      args=args,
      call=call,
      mle = if (multiple.components) vmat.n else vmat,
      BIC = BIC,
      llik = llik,
      grain.probability = if (multiple.components) grain.probability.n else grain.probability,
      components = if (multiple.components) comp.n else comp.re,
      single.comp = data.frame(mu = comp0[1], sigmab = comp0[2],
                               llik = comp0[3], BIC = comp0[4])))

  if (anyNA(unlist(summary)) && verbose)
    .throw_warning("The model produced NA values: either the input data are ",
                   "inapplicable for the model, or the model parameters ",
                   "need to be adjusted (e.g. 'sigmab')")

  ##=========##
  ## PLOTTING -----------
  if (plot)
    try(do.call(.plot_FiniteMixture, c(results, as.list(sys.call())[-(1:2)])))

  # Return values
  invisible(results)
}

## Function to plot an RLum.Results object generated by calc_FiniteMixture().
## This is a separate function to allows calling plot_RLum.Results() directly
## on the same object.
.plot_FiniteMixture <- function(object, ...) {
  if (length(object@data$args$n.components) == 1) {
    .throw_message("'n.components' specified only one component, nothing plotted",
                   error = FALSE)
    return()
  }

  ## deal with additional arguments
  extraArgs <- list(...)
  settings <- list(
      main = "Finite Mixture Model",
      pdf.weight = TRUE,
      pdf.sigma = "sigmab",
      pdf.colors = "gray",
      cex = 1,
      main.densities = "Normal distributions",
      main.proportions = "Proportion of components",
      main.criteria = "Statistical criteria",
      plot.proportions = TRUE,
      plot.criteria = TRUE
  )
  settings <- modifyList(settings, extraArgs)

  ## extract relevant data from object
  n.components <- object@data$args$n.components
  comp.n <- object@data$components
  sigmab <- object@data$args$sigmab
  BIC.n <- object@data$BIC$BIC
  LLIK.n <- object@data$llik$llik

  ## save previous plot parameter and set new ones
  par.default <- .par_defaults()
  on.exit(par(par.default), add = TRUE)

  ## DEVICE AND PLOT LAYOUT
  n.plots <- length(n.components) #number of PDF plots in plot area #1
  seq.matrix <- rbind(1:n.plots, 1:n.plots)
  if (settings$plot.proportions)
    seq.matrix <- rbind(seq.matrix, rep(max(seq.matrix) + 1))
  if (settings$plot.criteria)
    seq.matrix <- rbind(seq.matrix, rep(max(seq.matrix) + 1))

  ## create device layout
  graphics::layout(seq.matrix)

  ## general plot parameters (global scaling, allow overplotting)
  par(cex = 0.8 * settings$cex, xpd = NA)

  ## set margins (bottom, left, top, right)
  par(mar = c(2, 0, 2, 0), oma = c(0.6, 5, 1.3, 5))

  ## define colour palette for prettier output
  col.n <- switch(settings$pdf.colors,
                  colors = c("red3", "slateblue3", "seagreen", "tan3", "yellow3",
                             "burlywood4", "magenta4", "mediumpurple3", "brown4",
                             "grey", "aquamarine"),
                  gray = grDevices::gray.colors(length(n.components) * 2),
                  none = rgb(0, 0, 0, 0))
  poly.border <- settings$pdf.colors == "none"

  ##--------------------------------------------------------------------------
  ## PLOT 1: EQUIVALENT DOSES OF COMPONENTS

  ## calculate weighted density values
  .calc.density <- function(x, mu, sd, weights) {
    dnorm(x, mean = mu, sd = sd) * weights
  }

  max.dose <- max(object@data$data[, 1]) + sd(object@data$data[, 1]) / 2
  min.dose <- min(object@data$data[, 1]) - sd(object@data$data[, 1]) / 2

  ## determine y-axis scaling
  y.scale <- extraArgs$dose.scale %||% c(min.dose, max.dose)

  ## create empty plot without x-axis
  for (i in 1:n.plots) {
    pos.n <- seq(from = 1, to = n.components[i] * 3, by = 3)

    ## empty plot area
    plot(NA, NA,
         xlim = c(min(n.components)-0.2, max(n.components)+0.2),
         ylim = c(min(comp.n[pos.n, ] - comp.n[pos.n + 1, ], na.rm = TRUE),
                  max((comp.n[pos.n, ] + comp.n[pos.n + 1, ]) * 1.1, na.rm = TRUE)),
         ylab = "",
         xaxt = "n",
         yaxt = "n",
         xlab = "")

    ## add y-axis label (only for the first plot)
    if (i == 1) {
      mtext(expression(paste("D"[e]," [Gy]")), side = 2, line = 2.7,
            cex = settings$cex)
    }

    ## NORMAL DISTR. OF EACH COMPONENT

    if (settings$pdf.weight) {
      dens <- 0
      for (b in 1:max(n.components)) {
        comp.mu <- comp.n[pos.n[b], i]
        comp.sd <- if (settings$pdf.sigma == "se") comp.n[pos.n[b] + 1, i]
                   else comp.mu * sigmab
        if (!is.na(comp.mu) && !is.na(comp.sd)) {
          wi <- comp.n[pos.n[b] + 2, i]
          dens <- dens + sapply(0:max.dose, .calc.density,
                                comp.mu, comp.sd, wi)
        }
      }
    } else {
      ## estimate the y-scaling if no weights are used
      dens <- sapply(0:max.dose, .calc.density,
                     na.exclude(c(comp.n[pos.n, ])),
                     na.exclude(c(comp.n[pos.n + 1, ])), 1)
    }

    ## x-axis (density)
    x.scale <- extraArgs$pdf.scale %||% max(dens) * 1.1

    ## LOOP - iterate over number of components
    dens.sum <- 0
    for (j in 1:max(n.components)) {
      comp.mu <- comp.n[pos.n[j], i]
      comp.sd <- if (settings$pdf.sigma == "se") comp.n[pos.n[j] + 1, i]
                 else comp.mu * sigmab
      if (!is.na(comp.mu) && !is.na(comp.sd)) {
        ## calculate density values for 0 to maximum dose
        wi <- if (settings$pdf.weight) comp.n[pos.n[j] + 2, i] else 1
        dens <- sapply(0:max.dose, .calc.density,
                       comp.mu, comp.sd, wi)

        ## save density values in list for sum curve of gaussians
        dens.sum <- dens.sum + dens

        ## PLOT Normal Distributions
        par(new = TRUE)
        plot(dens, 1:length(dens) - 1,
             type = "l", yaxt = "n", xaxt = "n", col = col.n[j], lwd = 1,
             ylim = y.scale,
             xlim = c(0, x.scale),
             xaxs = "i", yaxs = "i",
             ann = FALSE, xpd = FALSE)

        ## add x-axis with corrected tick positions
        if (j == 1)
          axis(side = 1, labels = paste("k =", n.components[i]),
               at = x.scale / 2, tick = FALSE, line = -1)

        # draw coloured polygons under curve
        polygon(x = c(0, min(dens), dens, 0),
                y = c(0, 0, 0:max.dose, max.dose),
                col = adjustcolor(col.n[j], alpha.f = 0.66),
                yaxt = "n", border = poly.border, xpd = FALSE, lty = 2, lwd = 1.5)
      }
    }##EndOf::Component loop

    ## Add sum of Gaussian curve
    par(new = TRUE)
    plot(dens.sum, 1:length(dens) - 1,
         type = "l", yaxt = "n", xaxt = "n", col = "black",
         lwd = 1.5, lty = 1,
         ylim = y.scale,
         xlim = c(0, x.scale),
         xaxs = "i", yaxs = "i", ann = FALSE, xpd = FALSE)

    ## draw additional info during first k-cycle
    if (i == 1) {

      ## main title
      mtext(settings$main,
            side = 3, font = 2, line = 1.8, adj = 0.5, cex = settings$cex,
            at = graphics::grconvertX(0.5, from = "ndc", to = "user"))

      ## subtitle
      has.nas <- anyNA(unlist(object$summary))
      subtitle <- as.expression(bquote(italic(sigma[b]) == .(sigmab) ~
                                         "|" ~ n == .(length(object@data$data[, 1])) ~
                                         .(if (has.nas) "| The model produced NA values"
                                           else "")
                                       ))
      mtext(subtitle,
            side = 3, font = 1, line = 0.5,
            at = graphics::grconvertX(0.5, from = "ndc", to = "user"),
            col = ifelse(has.nas, 2, 1),
            cex = 0.9 * settings$cex)

      ## plot title
      mtext(settings$main.densities,
            side = 3, font = 2, line = 0, adj = 0, cex = 0.8 * settings$cex)

      ## x-axis label
      mtext("Density [a.u.]",
            side = 1, line = 1.5, adj = 0.5, cex = settings$cex,
            at = graphics::grconvertX(0.5, from = "ndc", to = "user"))

      ## draw y-axis with proper labels
      axis(side = 2, labels = TRUE)
    }

    if (settings$pdf.colors != "none" && i == n.plots) {
      legend("topright",
             legend = paste("Comp.", 1:max(n.components)),
             col = col.n[1:max(n.components)],
             pch = 19,
             pt.cex = 1.3,
             bty = "n")
    }

  } ## EndOf::k-loop and Plot 1

  ## this value (found by trial and error) should be summed to xlim to remove
  ## the empty gaps before the first bar and after the last in the proportions
  ## plot, and for better centring of points in the statistical criteria plot
  xlim.adj <- length(n.components) / (length(n.components) + 18) * c(1, -1)

  ##--------------------------------------------------------------------------
  ## PLOT 2: PROPORTION OF COMPONENTS
  if (settings$plot.proportions) {

    ## create matrix with proportions from a subset of the summary matrix
    prop.matrix <- comp.n[pos.n + 2, ] * 100

    ## stacked barplot of proportions without x-axis
    graphics::barplot(prop.matrix,
                      width = 1,
                      xlim = c(0, length(n.components)) + xlim.adj,
                      ylim = c(0, 100),
                      axes = TRUE,
                      space = 0,
                      col = col.n,
                      xpd = FALSE,
                      xaxt = "n")

    ## y-axis label
    mtext("Proportion [%]",
          side = 2,line = 3, cex = settings$cex)

    ## add x-axis with corrected tick positions
    axis(side = 1, labels = paste("k =", n.components),
         at = 1:length(n.components) - 0.5, tick = FALSE, line = -1)

    ## draw a box (not possible with barplot())
    graphics::box(lty = 1, col = "black")

    ## add subtitle
    mtext(settings$main.proportions,
          side = 3, font = 2, line = 0, adj = 0, cex = 0.8 * settings$cex)
  }

  ##--------------------------------------------------------------------------
  ## PLOT 3: BIC & LLIK

  if (settings$plot.criteria) {
  ## prepare scaling for both y-axes
  BIC.scale <- c(min(BIC.n) * if (min(BIC.n) < 0) 1.2 else 0.8,
                max(BIC.n) * if (max(BIC.n) < 0) 0.8 else 1.2)
  LLIK.scale <- c(min(LLIK.n) * if (min(LLIK.n) < 0) 1.2 else 0.8,
                  max(LLIK.n) * if (max(LLIK.n) < 0) 0.8 else 1.2)

  ## plot BIC scores
  plot(1:length(n.components) - 0.5, BIC.n,
       main = "",
       type = "b",
       pch = 22,
       cex = 1.5,
       xlim = c(0, length(n.components)) + xlim.adj,
       ylim = BIC.scale,
       xlab = "", xaxt = "n",
       ylab = expression(paste("BIC")),
       cex.lab = 1.25)

  ## following plot should be added to previous
  par(new = TRUE)

  ## plot LLIK estimates
  plot(1:length(n.components) - 0.5, LLIK.n,
       xlim = c(0, length(n.components)) + xlim.adj,
       ylim = LLIK.scale,
       axes = FALSE, type = "b", pch = 16, ylab = "", xlab = "", lty = 2, cex = 1.5)

  ## add x-axis with corrected tick positions
  axis(side = 1, labels = paste("k =", n.components),
       at = 1:length(n.components) - 0.5, tick = FALSE, line = -1)

  ## subtitle
  mtext(settings$main.criteria,
        side = 3, font = 2, line = 0, adj = 0, cex = 0.8 * settings$cex)

  ## second y-axis (LLIK) with label
  axis(side = 4)
  mtext(bquote(italic(L)[max]),
        side = 4,line = 3, cex = 1.3 * settings$cex)

  ## legend
  legend("topleft",
         legend = c("BIC", as.expression(bquote(italic(L)[max]))),
         pch = c(22, 16), pt.bg = c("white", "black"), pt.cex = 1.3,
         bty = "n", horiz = FALSE, inset = 0.01)
  }
}

#' @title Plot function for an `RLum.Analysis` S4 class object
#'
#' @description
#' The function provides a standardised plot output for curve data of an
#' [Luminescence::RLum.Analysis-class] object.
#'
#' The function produces a multiple plot output. A file output is recommended
#' (e.g., [pdf]).
#'
#' **curve.transformation**
#'
#' This argument allows transforming continuous wave (CW) curves to pseudo
#' (linear) modulated curves. For the transformation, the functions of the
#' package are used. Currently, it is not possible to pass further arguments to
#' the transformation functions. The argument works only for `ltype`
#' `OSL` and `IRSL`.
#'
#' Please note: The curve transformation within this functions works roughly,
#' i.e. every IRSL or OSL curve is transformed, without considering whether it
#' is measured with the PMT or not! However, for a fast look it might be
#' helpful.
#'
#' @param object [Luminescence::RLum.Analysis-class] (**required**):
#' S4 object of class [Luminescence::RLum.Analysis-class]
#'
#' @param subset named [list] (*optional*):
#' subsets elements for plotting. The arguments in the named [list] will be
#' directly passed to the function [Luminescence::get_RLum]
#' (e.g., `subset = list(curveType = "measured")`)
#'
#' @param nrows [integer] (*optional*):
#' number of rows in the plot output. If set to `NULL` the function tries to
#' find a reasonable value.
#'
#' @param ncols [integer] (*optional*):
#' number of columns in the plot output. If set to `NULL` the function tries to
#' find a reasonable value.
#'
#' @param abline [list] (*optional*):
#' allows to add ab-lines to the plot. Argument are provided
#' in a list and will be forward to the function [abline],
#' e.g., `list(v = c(10, 100))` adds two vertical lines add 10 and 100 to all
#' plots. In contrast `list(v = c(10), v = c(100)` adds a vertical at 10 to
#' the first and a vertical line at 100 to the 2nd plot.
#'
#' @param combine [logical] (*with default*):
#' allows to combine all [Luminescence::RLum.Data.Curve-class] objects in one single plot.
#'
#' @param records_max [integer] (*optional*):
#' limits number of records shown when `combine = TRUE` is used. The first and
#' last curves are always shown, the other curves shown are distributed evenly;
#' this may result in fewer curves plotted than specified. This argument must
#' be at least 2 to have an effect.
#'
#' @param curve.transformation [character] (*with default*):
#' allows transforming CW-OSL and CW-IRSL curves to pseudo-LM curves via
#' transformation functions. Allowed values are: `CW2pLM`, `CW2pLMi`,
#' `CW2pHMi` and `CW2pPMi`, see details. If set to `None` (default), no
#' transformation is applied.
#'
#' @param plot_singlePanels [logical] (*with default*):
#' global par settings are considered, normally this should end in one plot per page
#'
#' @param ... further arguments and graphical parameters will be passed to
#' the `plot` function.
#'
#' Supported arguments: `main`, `mtext`, `log`, `lwd`, `lty` `type`, `pch`, `col`,
#' `norm` (see [Luminescence::plot_RLum.Data.Curve]), `xlim`,`ylim`, `xlab`, `ylab`, ...
#'
#' and for `combine = TRUE` also: `sub_title`, `legend`, `legend.text`, `legend.pos`
#' (typical plus 'outside'), `legend.col`, `smooth`.
#'
#' All arguments can be provided as `vector` or `list` to gain in full control
#' of all plot settings.
#'
#' @return Returns multiple plots.
#'
#' @note
#' Not all arguments available for [plot] will be passed and they partly do not behave in the
#' way you might expect them to work. This function was designed to serve as an overview
#' plot, if you want to have more control, extract the objects and plot them individually.
#'
#' @section Function version: 0.3.16
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [plot], [Luminescence::plot_RLum], [Luminescence::plot_RLum.Data.Curve]
#'
#' @keywords aplot
#'
#' @examples
#'
#'##load data
#'data(ExampleData.BINfileData, envir = environment())
#'
#'##convert values for position 1
#'temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
#'
#'##(1) plot (combine) TL curves in one plot
#'plot_RLum.Analysis(
#' temp,
#' subset = list(recordType = "TL"),
#' combine = TRUE,
#' norm = TRUE,
#' abline = list(v = c(110))
#' )
#'
#'##(2) same as example (1) but using
#'## the argument smooth = TRUE
#'plot_RLum.Analysis(
#' temp,
#' subset = list(recordType = "TL"),
#' combine = TRUE,
#' norm = TRUE,
#' smooth = TRUE,
#' abline = list(v = c(110))
#' )
#'
#' @export
plot_RLum.Analysis <- function(
  object,
  subset = NULL,
  nrows = NULL,
  ncols = NULL,
  abline = NULL,
  combine = FALSE,
  records_max = NULL,
  curve.transformation = "None",
  plot_singlePanels = FALSE,
  ...
) {
  .set_function_name("plot_RLum.Analysis")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------
  .validate_class(object, "RLum.Analysis")
  if (length(object) == 0) {
    .throw_message("Nothing to plot, NULL returned", error = FALSE)
    return(NULL)
  }
  .validate_positive_scalar(records_max, null.ok = TRUE)

  if(!is.null(subset)){
    ##check whether the user set the drop option and remove it, as we cannot work with it
    subset <- subset[!sapply(names(subset), function(x){"drop" %in% x})]
    object <- do.call(get_RLum, c(object = object, subset, drop = FALSE))
  }

  .validate_logical_scalar(combine)
  if (combine) {
    sapply(object@records, function(x) {
      if (!inherits(x, "RLum.Data.Curve")) {
        .throw_error("'combine' is valid only for 'RLum.Data.Curve' objects")
      }
    })

    if (length(object@records) < 2) {
      combine <- FALSE
      .throw_message("'combine' can't be used with fewer than two curves, ",
                     "reset to FALSE", error = FALSE)
    }
  }

  .validate_positive_scalar(nrows, int = TRUE, null.ok = TRUE)
  .validate_positive_scalar(ncols, int = TRUE, null.ok = TRUE)

  ## Deal with additional arguments -----------------------------------------
  extraArgs <- list(...)

  ##create plot settings list
  plot.settings <- list(
    main = NULL,
    mtext = NULL,
    log = "",
    lwd = 1,
    lty = 1,
    type = "l",
    xlab = NULL,
    ylab = NULL,
    xlim = NULL,
    ylim = NULL,
    pch = NULL,
    col = "auto",
    norm = FALSE,
    sub_title = NULL,
    cex = 1,
    legend = TRUE,
    legend.text = NULL,
    legend.pos = NULL,
    legend.col = NULL,
    smooth = FALSE
  )

  plot.settings <- modifyList(x = plot.settings, val = extraArgs, keep.null = TRUE)

  ## deprecated argument
  if ("plot.single" %in% names(extraArgs)) {
    plot_singlePanels <- extraArgs$plot.single
    .deprecated("plot.single", "plot_singlePanels", since = "1.0.0")
  }

  ##try to find optimal parameters, this is however, a little bit stupid, but
  ##better than without any presetting
  if (combine) {
    n.plots <- length(unique(structure_RLum(object)$recordType))
  }
  else
    n.plots <- max(length_RLum(object), 1)

  ## set appropriate values for nrows and ncols if not specified
  if (is.null(nrows)) {
    nrows <- if (n.plots <= 4) ceiling(n.plots / 2) else 3
  }
  if (is.null(ncols)) {
    ncols <- min(n.plots, 2)
  }

  curve.transformation <- .validate_args(curve.transformation,
                                         c("CW2pLM", "CW2pLMi",
                                           "CW2pHMi", "CW2pPMi", "None"))
  ## complete the function name
  if (curve.transformation != "None") {
    curve.transformation <- paste0("convert_", curve.transformation)
  }

  # Plotting ------------------------------------------------------------------

  ## (1) NORMAL (combine == FALSE) -------------------------------------------
  if (!combine) {
    temp <- object@records

    ##set par
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)
    if (!plot_singlePanels) {
      par(mfrow = c(nrows, ncols))
    }

    ##expand plot settings list
    plot.settings <- lapply(plot.settings, function(setting) {
      if (is.null(setting))
        return(NULL)
      if (length(setting) == 1 || inherits(setting, "list"))
        return(rep_len(setting, length(temp)))
      rep_len(list(setting), length(temp))
    })

    ##expand abline
    if(!is.null(abline)){
      abline.names <- rep_len(names(abline), length.out = length(temp))
      abline <- rep_len(abline, length.out = length(temp))
      names(abline) <- abline.names
    }

    ##apply curve transformation
    for (i in seq_along(temp)) {
      if (inherits(temp[[i]], "RLum.Data.Curve")) {
        ##set curve transformation if wanted
        if (grepl("IRSL|OSL", temp[[i]]@recordType) &&
            curve.transformation != "None") {

          ## get the actual function from the parameter value and apply it
          temp[[i]] <- get(curve.transformation)(temp[[i]])
        }

        ##check plot settings and adjust
        ##xlim
        xlim.set <- plot.settings$xlim[[i]]
        if (!is.null(plot.settings$xlim)) {
          if (plot.settings$xlim[[i]][1] < min(temp[[i]]@data[,1])) {
            .throw_warning("min('xlim') < x-value range for curve #", i,
                           ", reset to minimum")
            xlim.set[1] <- min(temp[[i]]@data[,1])
          }
          if (plot.settings$xlim[[i]][2] > max(temp[[i]]@data[,1])) {
            .throw_warning("max('xlim') > x-value range for curve #", i,
                           ", reset to maximum")
            xlim.set[2] <- max(temp[[i]]@data[,1])
          }
        }

        ##ylim
        ylim.set <- plot.settings$ylim[[i]]
        if (!is.null(plot.settings$ylim)) {
          if (plot.settings$ylim[[i]][1] < min(temp[[i]]@data[,2])) {
            .throw_warning("min('ylim') < y-value range for curve #", i,
                           ", reset to minimum")
            ylim.set[1] <- min(temp[[i]]@data[,2])
          }
          if (plot.settings$ylim[[i]][2] > max(temp[[i]]@data[,2])) {
            .throw_warning("max('ylim') > y-value range for curve #", i,
                           ", reset to maximum")
            ylim.set[2] <- max(temp[[i]]@data[,2])
          }
        }

        ##col
        if (unique(plot.settings$col) != "auto") {
          col <- plot.settings$col[i]

        } else{
          col <- "black"
          if (grepl("IRSL", temp[[i]]@recordType, fixed = TRUE)) {
            col <- "red"
          } else if (grepl("OSL", temp[[i]]@recordType, fixed = TRUE)) {
            col <- "blue"
          }
        }

        ##main
        main <- plot.settings$main[[i]] %||% temp[[i]]@recordType

        ##++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        ##PLOT
        ##++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        ##plot RLum.Data.Curve curve
          ##we have to do this via this way, otherwise we run into a duplicated arguments
          ##problem
          ##check and remove duplicated arguments
          arguments <- c(
            list(
              object = temp[[i]],
              col = col,
              mtext = plot.settings$mtext[[i]] %||% paste0("#", i),
              par.local = FALSE,
              main = main,
              log = plot.settings$log[[i]],
              lwd = plot.settings$lwd[[i]],
              type = plot.settings$type[[i]],
              lty = plot.settings$lty[[i]],
              xlim = xlim.set,
              ylim = ylim.set,
              norm = plot.settings$norm[[i]],
              pch = plot.settings$pch[[i]],
              cex = plot.settings$cex[[i]],
              legend.col = plot.settings$legend.col[[i]],
              smooth = plot.settings$smooth[[i]]
            ),
            extraArgs
          )

          arguments[duplicated(names(arguments))] <- NULL

        ##call the function plot_RLum.Data.Curve
        do.call(what = "plot_RLum.Data.Curve", args = arguments)
        rm(arguments)

        ##add abline
        if(!is.null(abline[[i]])){
          do.call(what = "abline", args = abline[i])
        }

      } else {
        ## RLum.Data.Spectrum and RLum.Data.Image objects
        args <- extraArgs
        if (inherits(temp[[i]], "RLum.Data.Spectrum")) {
          ## remove already provided arguments
          args <- args[!names(args) %in% c("object", "mtext", "par.local", "main")]
        }

        do.call(what = paste0("plot_", class(temp[[i]])), args = c(list(
            object = temp[[i]],
            mtext = plot.settings$mtext[[i]] %||% paste0("#", i),
            par.local = FALSE,
            main = plot.settings$main %||% temp[[i]]@recordType
        ), args))
      }
    }#end for loop

  }else{

    ## (2) NORMAL (combine == TRUE)----------------------------------------------
    ##(1) check RLum objects in the set

    ##account for different curve types, combine similar
    temp.object.structure  <- structure_RLum(object)
    temp.recordType <- as.character(unique(temp.object.structure$recordType))

    ##change graphic settings
    if (!plot_singlePanels && !is.null(ncols) && !is.null(nrows)) {
        par(mfrow = c(nrows, ncols))
    }
    ## this 2nd par request is needed as setting mfrow resets the par
    ## settings ... this might not be wanted
    par(cex = plot.settings$cex)

    ##expand plot settings list
    plot.settings <- lapply(plot.settings, function(setting) {
      if (is.null(setting))
        return(NULL)
      if (is.list(setting))
        return(rep_len(setting, length(temp.recordType)))
      rep_len(list(setting), length(temp.recordType))
    })

    ##expand abline
    if(!is.null(abline)){
      abline.names <- rep_len(names(abline), length.out = length(temp.recordType))
      abline <- rep_len(abline, length.out = length(temp.recordType))
      names(abline) <- abline.names
    }

    ##(2) PLOT values
    for(k in 1:length(temp.recordType)) {

      ###get type of curves
      temp.object <-
        get_RLum(object, recordType = temp.recordType[k], drop = FALSE)

      ##get structure
      object.structure  <- structure_RLum(temp.object)

      ##now get the real list object (note the argument recursive = FALSE)
      object.list <-
        get_RLum(object, recordType = temp.recordType[k], recursive = FALSE)

      ## limit number of records shown ... show always first and last;
      ## distribute the rest
      records_show <- NULL
      if (!is.null(records_max) && records_max > 2) {
        num.objects <- length(object.list)
        records_show <- ceiling(seq(1, num.objects,
                                    length.out = min(records_max, num.objects)))
        object.list[(1:num.objects)[-records_show]] <- NULL
      }

      ## recompute num.objects, as it may have changed due to records_max
      num.objects <- length(object.list)

      ##transform values to data.frame and norm values
      temp.data.list <- lapply(1:num.objects, function(x) {

        ## set curve transformation if wanted
        if (grepl("IRSL|OSL", object.list[[x]]@recordType) &&
            curve.transformation != "None") {

          ## get the actual function from the parameter value and apply it
          object.list[[x]] <- get(curve.transformation)(object.list[[x]])
        }

        ## curve normalisation
        object.list[[x]] <- normalise_RLum(object.list[[x]],
                                           plot.settings$norm[[k]][1])
        as(object.list[[x]], "data.frame")
      })

      ##set plot parameters
      ##main
      main <- plot.settings$main[[k]] %||% paste(temp.recordType[[k]], "combined")

      ##xlab
      xlab <- plot.settings$xlab[[k]] %||% (
        if (grepl("TL", temp.recordType[[k]], fixed = TRUE))
          "Temperature [\u00B0C]"
        else
          "Time [s]"
      )

      ##ylab
      ylab <- plot.settings$ylab[[k]] %||% paste(temp.recordType[[k]], "[a.u.]")

      ##xlim
      xlim <- if (length(plot.settings$xlim[[k]]) > 1) {
        plot.settings$xlim[[k]]
      } else {
        c(min(object.structure$x.min), max(object.structure$x.max))
      }
      if (grepl("x", plot.settings$log[[k]], ignore.case = TRUE))
        xlim[which(xlim == 0)] <- 1

      ##ylim
      ylim <- if (length(plot.settings$ylim[[k]]) > 1) {
        plot.settings$ylim[[k]]
      } else {
        range(unlist(lapply(X = temp.data.list, FUN = function(x){
          range(x[,2])
        })))
      }
      if (grepl("y", plot.settings$log[[k]], ignore.case = TRUE))
        ylim[which(ylim == 0)] <- 1

      ##col (again)
      col <- if(length(plot.settings$col[[k]]) > 1 || plot.settings$col[[k]][1] != "auto"){
        plot.settings$col[[k]]
      }else{
        get("col", pos = .LuminescenceEnv)
      }

      ##if length of provided colours is < the number of objects, just one colour is supported
      if (length(col) < num.objects) {
        col <- rep_len(col, num.objects)
      }

      ##lty
      lty <- plot.settings$lty[[k]]
      if (length(plot.settings$lty[[k]]) < num.objects) {
        lty <- rep(plot.settings$lty[[k]], times = num.objects)
      }

      ##pch
      pch <- plot.settings$pch[[k]]
      if (length(plot.settings$pch[[k]]) < num.objects) {
        pch <- rep(plot.settings$pch[[k]], times = num.objects)
      }

      ## legend
      legend.text <- plot.settings$legend.text[[k]] %||%
        paste("Curve", records_show %||% 1:num.objects)
      legend.col <- plot.settings$legend.col[[k]]
      legend.pos <- plot.settings$legend.pos[[k]] %||% "topright"

      if (legend.pos == "outside") {
        par(mar = c(5.1, 4.1, 4.1, 8.1))
      }

      ##open plot area
      plot(
        NA,NA,
        xlim = xlim,
        ylim = ylim,
        main = main,
        xlab = xlab,
        ylab = ylab,
        log = plot.settings$log[[k]],
        sub = plot.settings$sub_title[[k]]
      )

      ##plot single curve values
      ## ...?Why using matplot is a bad idea: The channel resolution might be different
      for (n in 1:length(temp.data.list)) {

        ##smooth
        ##Why here again ... because the call differs from the one before, where the argument
        ##is passed to plot_RLum.Data.Curve()
        if(plot.settings$smooth[[k]]){
          k_factor <- ceiling(length(temp.data.list[[n]][, 2])/100)
          temp.data.list[[n]][, 2] <- .smoothing(temp.data.list[[n]][, 2],
                                                 k = k_factor, fill = NA,
                                                 align = "center")
        }

        ##remove 0 values if plotted on a log-scale
        # y-Axis
        if (grepl("y", plot.settings$log[[k]], ignore.case = TRUE))
          temp.data.list[[n]] <- temp.data.list[[n]][which(temp.data.list[[n]]$y > 0), ]
        # x-Axis
        if (grepl("x", plot.settings$log[[k]], ignore.case = TRUE))
          temp.data.list[[n]] <- temp.data.list[[n]][which(temp.data.list[[n]]$x > 0), ]

        ##print lines
        if (plot.settings$type[[k]] %in% c("l", "b")) {
          lines(
            temp.data.list[[n]],
            col = col[n],
            lty = lty[n],
            lwd = plot.settings$lwd[[k]]
          )
        }

        ##add points if requested
        if (plot.settings$type[[k]] %in% c("p", "b")) {
          points(
            temp.data.list[[n]],
            col = col[n],
            pch = pch[n]
          )
        }
      }

      ##add abline
      if(!is.null(abline[[k]])){
        do.call(what = "abline", args = abline[k])
      }

      ##mtext
      mtext(plot.settings$mtext[[k]], side = 3, cex = 0.8 * plot.settings$cex[[k]])

      ##if legend is outside of the plotting area we need to allow overplotting
      ##AFTER all lines have been drawn
      if (legend.pos == "outside") {
        # determine legend position on log(y) scale
        ypos <- par()$usr[4]
        if (grepl("y", plot.settings$log[[k]], ignore.case = TRUE))
          ypos <- 10^ypos

        # determine position on log(x) scale
        xpos <- par()$usr[2]
        if (grepl("x", plot.settings$log[[k]], ignore.case = TRUE))
          xpos <- 10^xpos
      }

      ##legend
      if (plot.settings$legend[[k]]) {
        legend(
          x = ifelse(legend.pos == "outside", xpos, legend.pos),
          y = ifelse(legend.pos == "outside", ypos, NULL),
          legend = legend.text,
          lwd = plot.settings$lwd[[k]],
          lty = plot.settings$lty[[k]],
          col = legend.col %||% col[1:num.objects],
          bty = "n",
          xpd = legend.pos == "outside",
          cex = 0.8
        )
      }
    }
  }
}

#' Class `"RLum.Results"`
#'
#' Object class contains results data from functions (e.g., [Luminescence::analyse_SAR.CWOSL]).
#'
#' @name RLum.Results-class
#'
#' @docType class
#'
#' @slot data
#' Object of class [list] containing output data
#'
#' @note
#' The class is intended to store results from functions to be used by
#' other functions. The data in the object should always be accessed by the
#' method [Luminescence::get_RLum].
#'
#' @section Objects from the Class:
#' Objects can be created by calls of the form `new("RLum.Results", ...)`.
#'
#' @section Class version: 0.5.2
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum-class], [Luminescence::plot_RLum], [Luminescence::merge_RLum]
#'
#' @keywords classes methods
#'
#' @examples
#'
#' showClass("RLum.Results")
#'
#' ##create an empty object from this class
#' set_RLum(class = "RLum.Results")
#'
#' ##use another function to show how it works
#'
#' ##Basic calculation of the dose rate for a specific date
#'  dose.rate <-  calc_SourceDoseRate(
#'    measurement.date = "2012-01-27",
#'    calib.date = "2014-12-19",
#'    calib.dose.rate = 0.0438,
#'    calib.error = 0.0019)
#'
#' ##show object
#' dose.rate
#'
#' ##get results
#' get_RLum(dose.rate)
#'
#' ##get parameters used for the calcualtion from the same object
#' get_RLum(dose.rate, data.object = "parameters")
#'
#' ##alternatively objects can be accessed using S3 generics, such as
#' dose.rate$parameters
#'
#' @export
setClass(
  Class = "RLum.Results",
  slots = list(data = "list"),
  contains = "RLum",
  prototype = list(data = list())
)


## as() ---------------------------------------------------------------------
##LIST
##COERCE RLum.Results >> list AND list >> RLum.Results
#' as() - RLum-object coercion
#'
#' for `[RLum.Results-class]`
#'
#' **[Luminescence::RLum.Results-class]**
#'
#' \tabular{ll}{
#'  **from** \tab **to**\cr
#'   `list` \tab `list`\cr
#' }
#'
#' Given that the [list] consists of [Luminescence::RLum.Results-class] objects.
#'
#' @name as
setAs("list", "RLum.Results",
      function(from,to){
        new(to,
            originator = "coercion",
            data = from)
      })

setAs("RLum.Results", "list",
      function(from){
        from@data
      })

## show() -------------------------------------------------------------------
#' @describeIn show
#' Show the structure of `RLum.Results` objects.
#'
#' @export
setMethod("show",
          signature(object = "RLum.Results"),
          function(object) {
            ##data elements
            temp.names <- names(object@data)

            data_list <- if (length(object) > 0) object@data else list(object@data)
            temp.type <- paste0("\t .. $", temp.names, " : ",
                                vapply(data_list, function(x) is(x)[1], character(1)))
            temp.type <- paste(temp.type, collapse = "\n")

            ##print information
            cat("\n [RLum.Results-class]")
            cat("\n\t originator: ", object@originator, "()", sep = "")
            cat("\n\t data:", length(object@data))
            cat("\n", temp.type)
            cat("\n\t additional info elements: ", length(object@info),"\n")
          })


## set_RLum() ---------------------------------------------------------------
#' @describeIn set_RLum
#' Construction method for [Luminescence::RLum.Results-class] objects.
#'
#' @export
setMethod("set_RLum",
          signature = signature("RLum.Results"),
          function(class,
                   originator,
                   .uid,
                   .pid,
                   data = list(),
                   info = list()) {

            ##create new class
            newRLumReuslts <- new("RLum.Results")

            ##fill object
            newRLumReuslts@originator <- originator
            newRLumReuslts@data <- data
            newRLumReuslts@info <- info
            newRLumReuslts@.uid <- .uid
            newRLumReuslts@.pid <- .pid

            return(newRLumReuslts)
          })


## get_RLum() ---------------------------------------------------------------
#' @describeIn get_RLum
#' Accessor method for [Luminescence::RLum.Results-class] object.
#' The argument `data.object` allows to access directly objects stored
#' within the slot data. The default return object depends on the object
#' originator (e.g., `fit_LMCurve`). If nothing is specified always the first
#' `data.object` will be returned.
#'
#' Note: Detailed specification should be made in combination with the
#' originator slot in the receiving function if results are piped.
#'
#' Returns:
#'
#' 1. Data object from the specified slot
#' 2. [list] of data objects from the slots if 'data.object' is vector or
#' 3. an [Luminescence::RLum.Results-class] for `drop = FALSE`.
#'
#' @param data.object [character] or [numeric]:
#' name or index of the data slot to be returned.
#'
#' @param info.object [character] (*optional*):
#' name of the wanted info element.
#'
#' @export
setMethod(
  "get_RLum",
  signature = signature("RLum.Results"),
  definition = function(object, data.object, info.object = NULL, drop = TRUE) {
    .set_function_name("get_RLum")
    on.exit(.unset_function_name(), add = TRUE)

    ## if info.object is set, only the info objects are returned if present
    if (!is.null(info.object)) {
      .validate_class(info.object, "character", null.ok = TRUE, length = 1)
      if (!info.object %in% names(object@info)) {
        if (length(object@info) == 0) {
          .throw_warning("This 'RLum.Results' object has no info objects, ",
                         "NULL returned")
        } else {
          .throw_warning("Invalid 'info.object' name, valid names are: ",
                         .collapse(names(object@info)))
        }
        return(NULL)
      }
      return(unlist(object@info[info.object]))
    }

    if (!missing(data.object)) {
      ##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      .validate_class(data.object, c("character", "numeric"))

      ## CASE 1: data.object is of type 'character'
      if (is.character(data.object) &&
          !all(data.object %in% names(object@data))) {
            .throw_error("Unknown 'data.object', valid names are: ",
                         .collapse(names(object@data)))
      }

      ## CASE 2: data.object is of type 'numeric'
      if (is.numeric(data.object) && max(data.object) > length(object@data)) {
            .throw_error("'data.object' index out of bounds")
      }

      ## extract the chosen subset of objects
      temp.return <- lapply(data.object, function(x) object@data[[x]])

      ## restore names as they get lost if data.object is numeric
      names(temp.return) <- names(object@data)[data.object]

    } else {
      ## CASE 3: data.object is missing
        ##return always the first object if nothing is specified
        temp.return <- object@data[1]
    }

    ## check whether an RLum.Results object needs to be produced
    .validate_logical_scalar(drop)
    if (drop) {
        ##we need to access the list here, otherwise we get unexpected behaviour as drop = TRUE
        ##should always return the lowest possible element here
        return(temp.return[[1]])
    }
    set_RLum(
          "RLum.Results",
          originator = object@originator,
          data = temp.return
        )
  }
)

## length_RLum() ------------------------------------------------------------
#' @describeIn length_RLum
#' Returns the number of stored data elements.
#'
#' @export
setMethod("length_RLum",
          "RLum.Results",
          function(object){
            length(object@data)
          })

## names_RLum() -------------------------------------------------------------
#' @describeIn names_RLum
#' Returns the names of the `data` field stored in the object.
#'
#' @export
setMethod("names_RLum",
          "RLum.Results",
          function(object){
             names(object@data)
          })

## view() -------------------------------------------------------------------
#' @describeIn view
#' View method for [Luminescence::RLum.Results-class] objects.
#'
#' @param element [integer] (*with default*):
#' index of the element to display.
#'
#' @export
setMethod("view",
          signature = "RLum.Results",
          definition = function(object, element = 1, ...) {
    .set_function_name("view")
    on.exit(.unset_function_name(), add = TRUE)

    .validate_not_empty(object)

    ## set title
    name <- list(...)$title
    if (is.null(name))
      name <- deparse(substitute(object))

    ## run view
    data <- get_RLum(object, data.object = element)
    .view(x = data, title = name)
})

## FUNCTIONS -------------------------------------------------------------------

#' @title Subset the DRAC reference
#'
#' @param x [data.frame] or `DRAC.list`
#'
#' @noRd
.get_DRAC_references <- function(x) {
  refs <- DRAC_refs()
  refs_names <- names(refs)

  used <- list(refs = NULL, desc = NULL)

  # TI:4 - Conversion factors
  ref_tmp <- unique(x[[.search_key(x, "TI:4")]])
  for (i in 1:length(ref_tmp)) {
    if (ref_tmp[i] == "X")
      next

    used$refs <- c(used$refs, refs[refs_names %in% ref_tmp[i]])
    used$desc <- c(used$desc, "Conversion factors")
  }

  # TI:13 - External Rubidium
  ref_tmp <- unique(x[[.search_key(x, "TI:13")]])
  if (any(ref_tmp == "Y")) {
    used$refs <- c(used$refs, refs["Mejdahl1987"])
    used$desc <- c(used$desc, "External rubidium")
  }

  # TI:22 - Internal Rubidium
  ref_tmp <- unique(x[[.search_key(x, "TI:22")]])
  if (any(ref_tmp == "Y")) {
    used$refs <- c(used$refs, refs["Mejdahl1987"])
    used$desc <- c(used$desc, "Internal rubidium")
  }

  # TI:31 - Gamma dose rate scaling
  ref_tmp <- unique(x[[.search_key(x, "TI:31")]])
  if (any(ref_tmp == "Y")) {
    used$refs <- c(used$refs, refs["Aitken1985"])
    used$desc <- c(used$desc, "Gamma dose rate scaling")
  }

  # TI:34 - alpha grain size attenuation
  ref_tmp <- unique(x[[.search_key(x, "TI:34")]])
  for (i in 1:length(ref_tmp)) {
    if (ref_tmp[i] == "X")
      next
    used$refs <- c(used$refs, refs[refs_names %in% ref_tmp[i]])
    used$desc <- c(used$desc, "Alpha grain size attenuation factors")
  }

  # TI:35 - Beta grain size attenuation
  ref_tmp <- unique(x[[.search_key(x, "TI:35")]])
  for (i in 1:length(ref_tmp)) {
    if (ref_tmp[i] == "X")
      next
    used$refs <- c(used$refs, refs[refs_names %in% ref_tmp[i]])
    used$desc <- c(used$desc, "Beta grain size attenuation factors")
  }

  # TI:38 - beta etch attenuation factor
  ref_tmp <- unique(x[[.search_key(x, "TI:38")]])
  for (i in 1:length(ref_tmp)) {
    if (ref_tmp[i] == "X")
      next
    used$refs <- c(used$refs, refs[refs_names %in% ref_tmp[i]])
    used$desc <- c(used$desc, "Beta etch attenuation factor")
  }

  # TI:50 - Cosmic dose rate
  ref_tmp <- unique(x[[.search_key(x, "TI:50")]])
  if (any(ref_tmp == "X")) {
    used$refs <- c(used$refs, refs[c("PrescottHutton1994", "PrescottStephan1982")])
    used$desc <- c(used$desc, c("Cosmic dose rate", "Cosmic dose rate"))
  }

  return(used)
}

## REFERENCE LIST --------------------------------------------------------------
DRAC_refs <- function() {

  list(
    Aitken1985 = bibentry(
      bibtype = "Book",
      author = person("M.J.", "Aitken"),
      title = "Thermoluminescence Dating",
      year = "1985",
      publisher = "Academic Press",
      adress = "London"
    ),

    AitkenXie1990 = bibentry(
      bibtype = "Article",
      author = c(
        person("M.J.", "Aitken"),
        person("J.", "Xie")
      ),
      title = "Moisture correction for annual gamma dose",
      year = "1990",
      journal = "Ancient TL",
      volume = "8",
      pages = "6-9"
    ),

    AdamiecAitken1998 = bibentry(
      bibtype = "Article",
      author = c(
        person("G.", "Adamiec"),
        person("M.J.", "Aitken")
      ),
      title = "Dose-rate conversion factors: update",
      year = "1998",
      journal = "Ancient TL",
      volume = "16",
      pages = "37-46"
    ),

    Guerinetal2011 = bibentry(
      bibtype = "Article",
      author = c(
        person("G.", "Guerin"),
        person("N.", "Mercier"),
        person("G.", "Adamiec")
      ),
      title = "Dose-rate conversion factors: update",
      year = "2011",
      journal = "Ancient TL",
      volume = "29",
      pages = "5-8",
      url = "http://ancienttl.org/ATL_29-1_2011/ATL_29-1_Guerin_p5-8.pdf"
    ),

    Liritzisetal2013 = bibentry(
      bibtype = "Article",
      author = c(
        person("I.", "Liritzis"),
        person("K.", "Stamoulis"),
        person("C.", "Papachristodoulou"),
        person("K.", "Ioannides")
      ),
      title = "A re-evaluation of radiation dose-rate conversion factors. ",
      year = "2013",
      journal = "Mediterranean Archaeology and Archaeometry",
      volume = "13",
      pages = "1-15",
      url = "http://maajournal.com/Issues/2012/pdf/FullTextLiritzis.pdf"
    ),

    Bell1979 = bibentry(
      bibtype = "Article",
      author = c(
        person("W.T.", "Bell")
      ),
      title = "Attenuation factors for the absorbed radiation dose in quartz inclusions for thermoluminescence dating",
      year = "1979",
      journal = "Ancient TL",
      volume = "8",
      pages = "1-12"
    ),

    Bell1980 = bibentry(
      bibtype = "Article",
      author = c(
        person("W.T.", "Bell")
      ),
      title = "Alpha attenuation in Quartz grains for Thermoluminescence Dating",
      year = "1980",
      journal = "Ancient TL",
      volume = "12",
      pages = "4-8"
    ),

    Brennanetal1991 = bibentry(
      bibtype = "Article",
      author = c(
        person("B.J.", "Brennan"),
        person("R.G.", "Lyons"),
        person("S.W.", "Phillips")
      ),
      title = "Attenuation of alpha particle track dose for spherical grains",
      year = "1991",
      journal = "International Journal of Radiation Applications and Instrumentation. Part D. Nuclear Tracks and Radiation Measurements",
      volume = "18",
      pages = "249-253"
    ),

    Cresswelletal2018 = bibentry(
      bibtype = "Article",
      author = c(
        person("A.J.", "Cresswell"),
        person("J.", "Carter"),
        person("D.C.W.", "Sanderson")
      ),
      title = "Dose rate conversion parameters: Assessment of nuclear data",
      year = "2008",
      journal = "Radiation Measurements",
      volume = "120",
      pages = "195-201",
      doi = "10.1016/j.radmeas.2018.02.007"
    ),

    Mejdahl1979 = bibentry(
      bibtype = "Article",
      author = c(
        person("V.", "Mejdahl")
      ),
      title = "Thermoluminescence Dating: Beta-Dose Attenuation in Quartz Grains",
      year = "1979",
      journal = "Archaeometry",
      volume = "21",
      pages = "61-72"
    ),

    Mejdahl1987 = bibentry(
      bibtype = "Article",
      author = c(
        person("V.", "Mejdahl")
      ),
      title = "Internal radioactivity in quartz and feldspar grains",
      year = "1987",
      journal = "Ancient TL",
      volume = "5",
      pages = "10-17"
    ),

    Brennan2003 = bibentry(
      bibtype = "Article",
      author = c(
        person("B.J.", "Brennan")
      ),
      title = "Beta doses to spherical grains",
      year = "2003",
      journal = "Radiation Measurements",
      volume = "37",
      pages = "299-303"
    ),

    `Guerinetal2012-Q` = bibentry(
      bibtype = "Article",
      author = c(
        person("G.", "Guerin"),
        person("N.", "Mercier"),
        person("R.", "Nathan"),
        person("G.", "Adamiec"),
        person("Y.", "Lefrais")
      ),
      title = "On the use of the infinite matrix assumption and associated concepts: A critical review",
      year = "2012",
      journal = "Radiation Measurements",
      volume = "47",
      pages = "778-785"
    ),

    `Guerinetal2012-F` = bibentry(
      bibtype = "Article",
      author = c(
        person("G.", "Guerin"),
        person("N.", "Mercier"),
        person("R.", "Nathan"),
        person("G.", "Adamiec"),
        person("Y.", "Lefrais")
      ),
      title = "On the use of the infinite matrix assumption and associated concepts: A critical review",
      year = "2012",
      journal = "Radiation Measurements",
      volume = "47",
      pages = "778-785"
    ),

    PrescottHutton1994 = bibentry(
      bibtype = "Article",
      author = c(
        person("J.R.", "Prescott"),
        person("J.T.", "Hutton")
      ),
      title = "Cosmic ray contributions to dose rates for luminescence and ESR dating: Large depths and long-term time variations",
      year = "1994",
      journal = "Radiation Measurements",
      volume = "23",
      pages = "497-500"
    ),

    PrescottStephan1982 = bibentry(
      bibtype = "Article",
      author = c(
        person("J.R.", "Prescott"),
        person("L.G.", "Stephan")
      ),
      title = "The contribution of cosmic radiation to the environmental dose for thermoluminescence dating",
      year = "1982",
      journal = "PACT",
      volume = "6",
      pages = "17-25"
    ),

    Readhead2002 = bibentry(
      bibtype = "Article",
      author = c(
        person("M.L.", "ReadHead")
      ),
      title = "Absorbed dose fraction for 87Rb beta particles",
      year = "2002",
      journal = "Ancient TL",
      volume = "20",
      pages = "25-29"
    )
  )
}

## internal helper to search key from the DRAC table
.search_key <- function(x, key) {
  if(inherits(x, "DRAC.list"))
    which(vapply(x, function(i) attr(i, "key") == key, logical(1)))
  else
    which(colnames(x) == key)
}

#' @title Calculate De value based on the gSGC by Li et al., 2015
#'
#' @description The function computes De value and De value error using the
#' global standardised growth curve (gSGC) assumption proposed by Li et al.,
#' 2015 for OSL dating of sedimentary quartz.
#'
#' @details
#' The error of the De value is determined using a Monte Carlo simulation approach.
#' Solving of the equation is realised using [uniroot].
#' Large values for `n.MC` will significantly increase the computation time.
#'
#' @param data [data.frame] (**required**):
#' input data  the following columns five columns in the given order:
#' `LnTn`, `LnTn.error`, `Lr1Tr1`, `Lr1Tr1.error`, `Dr1`. Column names are
#' not required.
#'
#' @param gSGC.type [character] (*with default*):
#' function parameters to use for the iteration procedure, either `"0-450"` or
#' `"0-250"`, as presented in Li et al., 2015 (Table 2). This is ignored if
#' `gSGC.parameters` is set.
#'
#' @param gSGC.parameters [list] (*optional*):
#' option to provide own function parameters used for fitting as named list.
#' Nomenclature follows Li et al., 2015, i.e. `list(A, A.error, D0, D0.error,
#' c, c.error, Y0, Y0.error, range)`, where `range` is defines the interval
#' where the function is considered as valid, e.g. `range = c(0,250)`.\cr
#' If set, option `gSGC.type` will be ignored.
#'
#' @param n.MC [integer] (*with default*):
#' number of Monte Carlo simulation runs for error estimation, see details.
#'
#' @param verbose [logical] (*with default*):
#' enable/disable output to the terminal.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ... parameters will be passed to the plot output
#'
#' @return Returns an S4 object of type [Luminescence::RLum.Results-class].
#'
#' **`@data`**\cr
#' `$ De.value` ([data.frame]) \cr
#'  `.. $ De`  \cr
#'  `.. $ De.error` \cr
#'  `.. $ Eta` \cr
#' `$ De.MC` ([list]) contains the matrices from the error estimation.\cr
#' `$ uniroot` ([list]) contains the [uniroot] outputs of the De estimations\cr
#'
#' **`@info`**\cr
#' `$ call`` ([call]) the original function call
#'
#' @section Function version: 0.1.3
#'
#' @author
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
#'
#' @seealso [Luminescence::RLum.Results-class], [Luminescence::get_RLum], [uniroot]
#'
#' @references
#' Li, B., Roberts, R.G., Jacobs, Z., Li, S.-H., 2015. Potential of establishing
#' a 'global standardised growth curve' (gSGC) for optical dating of quartz from sediments.
#' Quaternary Geochronology 27, 94-104. doi:10.1016/j.quageo.2015.02.011
#'
#' @keywords datagen
#'
#' @examples
#'
#' results <- calc_gSGC(data = data.frame(
#' LnTn =  2.361, LnTn.error = 0.087,
#' Lr1Tr1 = 2.744, Lr1Tr1.error = 0.091,
#' Dr1 = 34.4))
#'
#' get_RLum(results, data.object = "De")
#'
#' @export
calc_gSGC<- function(
  data,
  gSGC.type = c("0-250", "0-450"),
  gSGC.parameters = NULL,
  n.MC = 100,
  verbose = TRUE,
  plot = TRUE,
  ...
) {
  .set_function_name("calc_gSGC")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, "data.frame")
  .validate_not_empty(data)
  if (ncol(data) != 5)
    .throw_error("'data' is expected to have 5 columns")
  gSGC.type <- .validate_args(gSGC.type, c("0-250", "0-450"))
  .validate_class(gSGC.parameters, "list", null.ok = TRUE)
  .validate_positive_scalar(n.MC, int = TRUE)
  .validate_logical_scalar(verbose)
  .validate_logical_scalar(plot)

  ##rename columns for consistency reasons
  colnames(data) <- c('LnTn', 'LnTn.error', 'Lr1Tr1', 'Lr1Tr1.error', 'Dr1')

  ## ensure errors are not negative
  data$LnTn.error <- abs(data$LnTn.error)
  data$Lr1Tr1.error <- abs(data$Lr1Tr1.error)


##============================================================================##
##DEFINE FUNCTION
##============================================================================##

    ##define function, nomenclature according to publication that should be solved
    f <- function(x,A,D0,c,Y0,Dr1,Lr1Tr1,LnTn) {
      (((A * (1 - exp( - Dr1 / D0))) + c * Dr1 + Y0)/Lr1Tr1) -
      (((A * (1 - exp( - x/D0))) + c * x + Y0)/LnTn)
    }

    ##set general parameters
    if (!is.null(gSGC.parameters)) {
      A <- gSGC.parameters$A
      A.error <- gSGC.parameters$A.error
      D0 <- gSGC.parameters$D0
      D0.error <- gSGC.parameters$D0.error
      c <- gSGC.parameters$c
      c.error <- gSGC.parameters$c.error
      Y0 <- gSGC.parameters$Y0
      Y0.error <- gSGC.parameters$Y0.error
      range <- gSGC.parameters$range

    }else{
      if (gSGC.type == "0-450") {
        A <- 0.723
        A.error <- 0.014
        D0 <- 65.1
        D0.error <- 0.9
        c <- 0.001784
        c.error <- 0.000016
        Y0 <- 0.009159
        Y0.error <- 0.004795

        range <- c(0.1,250)

      }else if (gSGC.type == "0-250") {
        A <- 0.787
        A.error <- 0.051
        D0 <- 73.9
        D0.error <- 2.2
        c <- 0.001539
        c.error <- 0.000068
        Y0 <- 0.01791
        Y0.error <- 0.00490

        range <- c(0.1,250)
      }
    }

    ##Define size of output objects
    output.data <- data.table::data.table(
      DE = numeric(length = nrow(data)),
      DE.ERROR =  numeric(length = nrow(data)),
      ETA =  numeric(length = nrow(data))
    )

    ##set list for De.MC
    output.De.MC <- vector("list",  nrow(data))

    ##set list for uniroot
    output.uniroot <-  vector("list",  nrow(data))


##============================================================================##
##CALCULATION
##============================================================================##
for(i in 1:nrow(data)){
    Lr1Tr1 <- data[i, "Lr1Tr1"]
    Lr1Tr1.error <- data[i,"Lr1Tr1.error"]
    Dr1 <- data[i,"Dr1"]
    LnTn <- data[i,"LnTn"]
    LnTn.error <- data[i,"LnTn.error"]

  ##calculate mean value
  temp <- try(uniroot(
      f,
      interval = c(0.1,450),
      tol = 0.001,
      A = A,
      D0 = D0,
      c = c,
      Y0 = Y0,
      Dr1 = Dr1,
      Lr1Tr1 = Lr1Tr1,
      LnTn = LnTn,
      extendInt = 'yes',
      check.conv = TRUE,
      maxiter = 1000
    ), silent = TRUE)

  ## allocate matrix
  temp.MC.matrix <- matrix(nrow = n.MC, ncol = 8)
  colnames(temp.MC.matrix) <- c("LnTn", "Lr1Tr1", "A", "D0", "c", "Y0", "De", "Eta")

  ## fill the first 6 columns of the matrix
  temp.MC.matrix[, 1:6] <- matrix(rnorm(
      n.MC * 6,
      mean = c(LnTn, Lr1Tr1, A, D0, c, Y0),
      sd = c(LnTn.error, Lr1Tr1.error, A.error, D0.error, c.error, Y0.error)
  ), ncol = 6, byrow = TRUE)

  if(!inherits(temp, "try-error")){

    ##get De
    De <- temp$root

    ##calculate Eta, which is the normalisation factor
    Eta <- ((A * (1 - exp( - Dr1 / D0))) + c * Dr1 + Y0)/Lr1Tr1

    ##--------------------------------------------------------------------------##
    ##Monte Carlo simulation for error estimation

      ##run uniroot to get the De
      temp.MC.matrix[,7] <- vapply(X = 1:n.MC, FUN = function(x){

        uniroot(f,
                interval = c(0.1,450),
                tol = 0.001,
                A = temp.MC.matrix[x,3],
                D0 = temp.MC.matrix[x,4],
                c = temp.MC.matrix[x,5],
                Y0 = temp.MC.matrix[x,6],
                Dr1 = Dr1,
                Lr1Tr1 =temp.MC.matrix[x,2],
                LnTn = temp.MC.matrix[x,1],
                check.conv = TRUE,
                extendInt = 'yes',
                maxiter = 1000
                )$root

      }, FUN.VALUE = numeric(1))

      ##calculate also the normalisation factor
      temp.MC.matrix[,8] <- (temp.MC.matrix[,3] * (1 - exp( - Dr1 / temp.MC.matrix[,4])) +
        temp.MC.matrix[,5] * Dr1 + temp.MC.matrix[,6])/temp.MC.matrix[,2]

      ##get De error as SD
      De.error <- sd(temp.MC.matrix[,7])

  }else{
    .throw_warning("No solution was found")
    De <- NA
    Eta <- NA
    De.error <- NA
  }


# Plot output -------------------------------------------------------------
  if (plot) {

    ##set plot settings
    plot.settings <- list(
      main = "gSGC and resulting De",
      xlab = "Dose [a.u.]",
      ylab = expression(paste("Re-norm. ", L[x]/T[x])),
      xlim = NULL,
      ylim = NULL,
      lwd = 1,
      lty = 1,
      pch = 21,
      col = "red",
      grid = expression(nx = 10, ny = 10),
      mtext = ""
    )

    plot.settings <-  modifyList(plot.settings, list(...))

    ##graphical feedback
    x <- NA
    curve(
      A * (1 - exp(-x / D0)) + c * x + Y0, from = 0, to = 500,
      xlab = plot.settings$xlab,
      ylab = plot.settings$ylab,
      main = plot.settings$main,
      xlim = plot.settings$xlim,
      ylim = plot.settings$ylim,
      lwd = plot.settings$lwd,
      lty = plot.settings$lty
    )

    mtext(side = 3, plot.settings$mtext)

    if(!is.null(plot.settings$grid)){
      graphics::grid(eval(plot.settings$grid))
    }

    if(!inherits(temp, "try-error")){

      if(temp$root < 450 & temp$root > 0){
        points(temp$root,Eta*LnTn, col = plot.settings$col, pch = plot.settings$pch)
        segments(De - De.error,Eta * LnTn,
                 De + De.error,Eta * LnTn)

        hist <-
          hist(
            temp.MC.matrix[, 7],
            freq = FALSE,
            add = TRUE,
            col = rgb(0, 0, 0, 0.2),
            border = rgb(0, 0, 0, 0.5)
          )
        lines(hist$mids,hist$density)

      }else{

        if(temp$root < 450){
          x0 <- 450
          x1 <- 500
        } else {
          x0 <- 50
          x1 <- 0
        }
        shape::Arrows(
            x0 = x0,
            y0 = par()$usr[4] - 0.2,
            x1 = x1,
            y1 = par()$usr[4] - 0.2,
            arr.type = "triangle",
            col = "red"
        )

        mtext(side = 1, text = "Out of bounds!", col = "red")
      }

    }else{
      mtext(side = 1, text = "No solution found!", col = "red")
    }
  }


# Terminal output ---------------------------------------------------------
    if (verbose) {
      cat("\n[calc_gSGC()]")
      cat("\n Corresponding De based on the gSGC\n")

      cat(paste0("\n"," Ln/Tn:\t\t ",LnTn," \u00B1 ", LnTn.error,"\n"))
      cat(paste0(""," Lr1/Tr1:\t ",Lr1Tr1," \u00B1 ", Lr1Tr1.error,"\n"))
      cat(paste0(""," Dr1:\t\t ",Dr1,"\n"))
      cat(paste0(""," f(D):\t\t ",A," * (1 - exp(-D /",D0,")) + c * D + ",Y0,"\n"))
      cat(paste0(""," n.MC:\t\t ",n.MC,"\n"))
      cat(paste0(" ------------------------------ \n"))
      cat(paste0(" De:\t\t",round(De,digits = 2)," \u00B1 ",round(De.error,digits = 2),"\n"))
      cat(paste0(" ------------------------------ \n"))
    }


##============================================================================##
##CREATE OUTPUT OBJECTS
##============================================================================##

    ##replace values in the data.table with values
    output.data[i, `:=`(DE = De,
                         DE.ERROR = De.error,
                         ETA = Eta)]

    ##matrix - to prevent memory overload limit output
    if(n.MC * nrow(data) > 1e6){
      # nocov start
      if(i == 1){
        output.De.MC[[i]] <- temp.MC.matrix
      }else{
        output.De.MC[[i]] <- NA
      }

      .throw_warning("Only the first MC matrix is returned to prevent ",
                     "memory overload")
      # nocov end
    }else{
      output.De.MC[[i]] <- temp.MC.matrix
    }

    output.uniroot[[i]] <- temp

}##end for loop

##============================================================================##
##OUTPUT RLUM
##============================================================================##

  set_RLum(
      class = "RLum.Results",
      data = list(
        De = as.data.frame(output.data),
        De.MC =  output.De.MC,
        uniroot = output.uniroot
      ),
      info = list( call = sys.call())
    )
}

#' @title Apply the Huntley (2006) model
#'
#' @description
#' The function calculates the expected sample specific fraction of saturation
#' based on the model of Huntley (2006), using the approach as implemented
#' in Kars et al. (2008) or Guralnik et al. (2015).
#'
#' @details
#'
#' This function applies the approach described in Kars et al. (2008) or Guralnik et al. (2015),
#' which are both developed from the model of Huntley (2006), to calculate the expected sample
#' specific fraction of saturation of a feldspar and also to calculate fading
#' corrected age using this model. \eqn{\rho}' (`rhop`), the density of recombination
#' centres, is a crucial parameter of this model and must be determined
#' separately from a fading measurement. The function [Luminescence::analyse_FadingMeasurement]
#' can be used to calculate the sample specific \eqn{\rho}' value.
#'
#' **Kars et al. (2008) -- Single saturating exponential**
#'
#' To apply the approach after Kars et al. (2008), use `fit.method = "EXP"`.
#'
#' Firstly, the unfaded \eqn{D_0} value is determined through applying equation 5 of
#' Kars et al. (2008) to the measured \eqn{\frac{L_x}{T_x}} data as a function of irradiation
#' time, and fitting the data with a single saturating exponential of the form:
#'
#' \deqn{\frac{L_x}{T_x}(t^*) = A  \phi(t^*) \{1 - \exp(-\frac{t^*}{D_0}))\}}
#'
#' where
#'
#' \deqn{\phi(t^*) = \exp(-\rho' \ln(1.8  \tilde{s} t^*)^3)}
#'
#' after King et al. (2016) where \eqn{A} is a pre-exponential factor,
#' \eqn{t^*} (s) is the irradiation time, starting at the mid-point of
#' irradiation (Auclair et al. 2003) and \eqn{\tilde{s}} (\eqn{3\times10^{15}} s\eqn{^{-1}})
#' is the athermal frequency factor after Huntley (2006).
#'
#' Using fit parameters \eqn{A} and \eqn{D_0}, the function then computes a natural dose
#' response curve using the environmental dose rate, \eqn{\dot{D}} (Gy/s) and equations
#' `[1]` and `[2]`. Computed \eqn{\frac{L_x}{T_x}} values are then fitted using the
#' [Luminescence::fit_DoseResponseCurve] function and the laboratory measured LnTn can then
#' be interpolated onto this curve to determine the fading corrected
#' \eqn{D_e} value, from which the fading corrected age is calculated.
#'
#' **Guralnik et al. (2015) -- General-order kinetics**
#'
#' To apply the approach after Guralnik et al. (2015) use `fit.method = "GOK"`.
#'
#' The approach of Guralnik et al. (2015) is very similar to that of
#' Kars et al. (2008) but, instead of using a single saturating exponential,
#' the model fits a general-order kinetics function of the form:
#'
#' \deqn{\frac{L_x}{T_x}(t^*) = A \phi (t^*)(1 - (1 + (\frac{1}{D_0}) t^* c)^{-1/c})}
#'
#' where \eqn{A}, \eqn{\phi}, \eqn{t^*} and \eqn{D_0} are the same as above and \eqn{c} is a
#' dimensionless kinetic order modifier (cf. equation 10 in
#' Guralnik et al., 2015).
#'
#' **Level of saturation**
#'
#' The [Luminescence::calc_Huntley2006] function also calculates the level of
#' saturation (\eqn{\frac{n}{N}})
#' and the field saturation (i.e. athermal steady state, (n/N)_SS) value for
#' the sample under investigation using the sample specific \eqn{\rho}',
#' unfaded \eqn{D_0} and \eqn{\dot{D}} values, following the approach of Kars et al. (2008).
#'
#' The computation is done using 1000 equally-spaced points in the interval
#' \[0.01, 3\]. This can be controlled by setting option `rprime`, such as
#' in `rprime = seq(0.01, 3, length.out = 1000)` (the default).
#'
#' **Uncertainties**
#'
#' Uncertainties are reported at \eqn{1\sigma} and are assumed to be normally
#' distributed and are estimated using Monte-Carlo re-sampling (`n.MC = 10000`
#' by default) of \eqn{\rho}' and \eqn{\frac{L_x}{T_x}} during dose response
#' curve fitting, and of \eqn{\rho}' #' in the derivation of (\eqn{n/N}) and
#' (n/N)_SS.
#'
#' **Age calculated from 2D0 of the simulated natural DRC**
#'
#' In addition to the age calculated from the equivalent dose derived from
#' \eqn{\frac{L_n}{T_n}} projected on the simulated natural dose response curve (DRC), this function
#' also calculates an age from twice the characteristic saturation dose (`D0`)
#' of the simulated natural DRC. This can be a useful information for
#' (over)saturated samples (i.e., no intersect of \eqn{\frac{L_n}{T_n}} on the natural DRC)
#' to obtain at least a "minimum age" estimate of the sample. In the console
#' output this value is denoted by *"Age @2D0 (ka):"*.
#'
#' @param data [data.frame] (**required**):
#' A `data.frame` with one of the following structures:
#' - **three columns** with numeric values for dose (s), `LxTx` and `LxTx`
#' error, in this order.
#' - **two columns** with numeric values for dose (s) and `LxTx`, in this order.
#' This assumes that errors on `LxTx` are missing, and a third column will be
#' automatically attached with an arbitrary 5 % error on the provided `LxTx`
#' values.
#' - **wide table**, i.e. a [data.frame] with a number of columns divisible by 3
#' and where each triplet has the aforementioned column structure.
#'
#' ```
#'                         (optional)
#'      | dose (s)| LxTx | LxTx error |
#'      |  [ ,1]  | [ ,2]|    [ ,3]   |
#'      |---------|------|------------|
#' [1, ]|  0      | LnTn | LnTn error | (optional, see arg 'LnTn')
#' [2, ]|  R1     | L1T1 | L1T1 error |
#'  ... |    ...  |  ... |     ...    |
#' [x, ]|  Rx     | LxTx | LxTx error |
#'
#' ```
#' **Note:** The function assumes the first row of the data to be the
#' `Ln/Tn`-value. If you want to provide more than one `Ln/Tn`-values, consider
#' using argument `LnTn`.
#'
#' @param LnTn [data.frame] (*optional*):
#' A two column data frame with the following structure:
#'
#' ```
#'      |  LnTn  |  LnTn error  |
#'      |  [ ,1] |      [ ,2]   |
#'      |--------|--------------|
#' [1, ]| LnTn_1 | LnTn_1 error |
#' [2, ]| LnTn_2 | LnTn_2 error |
#'  ... |   ...  |      ...     |
#' [x, ]| LnTn_x | LnTn_x error |
#' ```
#'
#' The function will calculate a **mean** `Ln/Tn`-value and uses either the
#' standard deviation or the highest individual error, whichever is larger. If
#' another mean value (e.g. a weighted mean or median) or error is preferred,
#' this value must be calculated beforehand and used in the first row in the
#' data frame for argument `data`.
#'
#' **Note:** This argument should **only** be used to provide more than one
#' `Ln/Tn`-value. If you provide `LnTn`-values with this argument, the data
#' frame for the `data`-argument **must not** contain any `LnTn`-values.
#'
#' @param rhop [numeric] (**required**):
#' A vector of length 2 for the density of recombination centres (\eqn{\rho}')
#' and its error (see Huntley 2006). Note that \eqn{\rho}' must **not** be
#' provided as the common logarithm. Example: `rhop = c(2.92e-06, 4.93e-07)`.
#'
#' @param ddot [numeric] (**required**):
#' A vector of length 2 for the environmental dose rate and its error.
#' Expected unit: Gy/ka. Example: `ddot = c(3.7, 0.4)`.
#'
#' @param readerDdot [numeric] (**required**):
#' A vector of length 2 for the dose rate of the irradiation source of the OSL
#' reader and its error.
#' Expected unit: Gy/s. Example: `readerDdot = c(0.08, 0.01)`.
#'
#' @param normalise [logical] (*with default*): If `TRUE` (the default) all
#' measured and computed \eqn{\frac{L_x}{T_x}} values are normalised by the
#' pre-exponential factor `A` (see details).
#'
#' @param fit.method [character] (*with default*):
#' Fit function of the dose response curve. Can either be `"EXP"` (default) or
#' `"GOK"`. Note that `"EXP"` (single saturating exponential) is the original
#' function the model after Huntley (2006) and Kars et al. (2008) was
#' designed to use. The use of a general-order kinetics function (`"GOK"`)
#' is an experimental adaptation of the model and should be used
#' with great care.
#'
#' @param lower.bounds [numeric] (*with default*):
#' A vector of length 4 for the values of the lower bounds to be applied
#' when fitting the models with [minpack.lm::nlsLM]. In most cases, the
#' default values (`c(-Inf, -Inf, -Inf, -Inf)`) are appropriate for finding
#' a best fit, but sometimes it may be useful to restrict the lower bounds to
#' e.g. `c(0, 0, 0, 0)`. The values of the vectors are, respectively, for
#' parameters `a`, `D0`, `c` and `d` in that order (parameter `d` is ignored
#' when `fit.method = "EXP"`). More details can be found in
#' [Luminescence::fit_DoseResponseCurve].
#'
#' @param cores [integer] (*with default*):
#' The number of cores to use. This will be capped to the number of available
#' cores if set to too high.
#'
#' @param summary [logical] (*with default*):
#' If `TRUE` (the default) various parameters provided by the user
#' and calculated by the model are added as text on the right-hand side of the
#' plot.
#'
#' @param plot [logical] (*with default*):
#' enable/disable the plot output.
#'
#' @param ...
#' Further parameters:
#' - `verbose` [logical]: Enable/disable output to the terminal (default = `TRUE`)
#' - `n.MC` [numeric]: Number of Monte Carlo iterations (default = 10000)
#' - `cex` [numeric]: Scaling of the plot (default = 1)
#' - `maxiter` [numeric]: Number of iteration limits for nls fitting
#' - `trace` [logical]: Enable/disable value tracing the terminal during fitting
#' **Note:** it is generally advised to have a large number of Monte Carlo
#' iterations for the results to converge. Decreasing the number of iterations
#' will often result in unstable estimates.
#'
#' All other arguments are passed to [plot] and [Luminescence::fit_DoseResponseCurve] (in
#' particular `mode` for the De calculation mode, `fit.force_through_origin`,
#' and `fit.bounds`).
#'
#' @return An [Luminescence::RLum.Results-class] object is returned:
#'
#' Slot: **@data**\cr
#'
#' \tabular{lll}{
#' **OBJECT** \tab **TYPE** \tab **COMMENT**\cr
#' `results` \tab [data.frame] \tab results of the of Kars et al. 2008 model \cr
#' `data` \tab [data.frame] \tab original input data \cr
#' `Ln` \tab [numeric] \tab Ln and its error \cr
#' `LxTx_tables` \tab `list` \tab A `list` of `data.frames` containing data on dose,
#'  LxTx and LxTx error for each of the dose response curves.
#'  Note that these **do not** contain the natural `Ln` signal, which is provided separately. \cr
#' `fits` \tab `list` \tab A `list` of `nls` objects produced by [minpack.lm::nlsLM] when fitting the dose response curves \cr
#' }
#'
#' Slot: **@info**\cr
#'
#' \tabular{lll}{
#' **OBJECT** \tab **TYPE** \tab **COMMENT** \cr
#' `call` \tab `call` \tab the original function call \cr
#' `args` \tab `list` \tab arguments of the original function call \cr
#' }
#'
#' @section Function version: 0.4.7
#'
#' @author
#' Georgina E. King, University of Lausanne (Switzerland) \cr
#' Christoph Burow, University of Cologne (Germany) \cr
#' Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
#' Marco Colombo, Institute of Geography, Heidelberg University (Germany)
#'
#' @keywords datagen
#'
#' @note This function has BETA status, in particular for the GOK implementation. Please verify
#' your results carefully.
#'
#' @references
#'
#' Kars, R.H., Wallinga, J., Cohen, K.M., 2008. A new approach towards anomalous fading correction for feldspar
#' IRSL dating-tests on samples in field saturation. Radiation Measurements 43, 786-790. doi:10.1016/j.radmeas.2008.01.021
#'
#' Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
#' Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
#' from feldspar. Radiation Measurements 81, 224-231.
#'
#' Huntley, D.J., 2006. An explanation of the power-law decay of luminescence.
#' Journal of Physics: Condensed Matter 18, 1359-1365. doi:10.1088/0953-8984/18/4/020
#'
#' King, G.E., Herman, F., Lambert, R., Valla, P.G., Guralnik, B., 2016.
#' Multi-OSL-thermochronometry of feldspar. Quaternary Geochronology 33, 76-87. doi:10.1016/j.quageo.2016.01.004
#'
#'
#' **Further reading**
#'
#' Morthekai, P., Jain, M., Cunha, P.P., Azevedo, J.M., Singhvi, A.K., 2011. An attempt to correct
#' for the fading in million year old basaltic rocks. Geochronometria 38(3), 223-230.
#'
#' @examples
#'
#' ## Load example data (sample UNIL/NB123, see ?ExampleData.Fading)
#' data("ExampleData.Fading", envir = environment())
#'
#' ## (1) Set all relevant parameters
#' # a. fading measurement data (IR50)
#' fading_data <- ExampleData.Fading$fading.data$IR50
#'
#' # b. Dose response curve data
#' data <- ExampleData.Fading$equivalentDose.data$IR50
#'
#' ## (2) Define required function parameters
#' ddot <- c(7.00, 0.004)
#' readerDdot <- c(0.134, 0.0067)
#'
#' # Analyse fading measurement and get an estimate of rho'.
#' # Note that the RLum.Results object can be directly used for further processing.
#' # The number of MC runs is reduced for this example
#' rhop <- analyse_FadingMeasurement(fading_data, plot = TRUE, verbose = FALSE, n.MC = 10)
#'
#' ## (3) Apply the Kars et al. (2008) model to the data
#' kars <- calc_Huntley2006(
#'  data = data,
#'  rhop = rhop,
#'  ddot = ddot,
#'  readerDdot = readerDdot,
#'  n.MC = 25)
#'
#' \dontrun{
#' # You can also provide LnTn values separately via the 'LnTn' argument.
#' # Note, however, that the data frame for 'data' must then NOT contain
#' # a LnTn value. See argument descriptions!
#' LnTn <- data.frame(
#'  LnTn = c(1.84833, 2.24833),
#'  nTn.error = c(0.17, 0.22))
#'
#' LxTx <- data[2:nrow(data), ]
#'
#' kars <- calc_Huntley2006(
#'  data = LxTx,
#'  LnTn = LnTn,
#'  rhop = rhop,
#'  ddot = ddot,
#'  readerDdot = readerDdot,
#'  n.MC = 25)
#' }
#' @export
calc_Huntley2006 <- function(
    data,
    LnTn = NULL,
    rhop = NULL,
    ddot = NULL,
    readerDdot = NULL,
    normalise = TRUE,
    fit.method = c("EXP", "GOK"),
    lower.bounds = c(-Inf, -Inf, -Inf, -Inf),
    cores = 1,
    summary = TRUE,
    plot = TRUE,
    ...
) {
  .set_function_name("calc_Huntley2006")
  on.exit(.unset_function_name(), add = TRUE)

  ## Integrity checks -------------------------------------------------------

  .validate_class(data, "data.frame")
  .validate_not_empty(data)
  .validate_class(LnTn, "data.frame", null.ok = TRUE)
  fit.method <- .validate_args(fit.method, c("EXP", "GOK"))
  .validate_class(lower.bounds, "numeric", length = 4)
  .validate_logical_scalar(summary)
  .validate_logical_scalar(plot)

  ## Check 'data'
  if (ncol(data) < 2)
    .throw_error("'data' should have at least 2 columns")
  if (ncol(data) == 2) {
      .throw_warning("'data' has only two columns: we assume that the errors ",
                     "on LxTx are missing and automatically add a 5% error.\n",
                     "Please provide a data frame with three columns ",
                     "if you wish to use actually measured LxTx errors.")
      data[ ,3] <- data[ ,2] * 0.05
  }

  ## Check if 'LnTn' is used and overwrite 'data'
  if (!is.null(LnTn)) {
    if (ncol(LnTn) != 2)
      .throw_error("'LnTn' should be a data frame with 2 columns")
    if (ncol(data) > 3)
      .throw_error("When 'LnTn' is specified, 'data' should have only ",
                   "2 or 3 columns")

      # case 1: only one LnTn value
      if (nrow(LnTn) == 1) {
        LnTn <- setNames(cbind(0, LnTn), names(data))

        # case 2: >1 LnTn value
      } else {
        LnTn_mean <- mean(LnTn[ ,1])
        LnTn_sd <- sd(LnTn[ ,1])
        LnTn_error <- max(LnTn_sd, LnTn[ ,2])
        LnTn <- setNames(data.frame(0, LnTn_mean, LnTn_error), names(data))
      }
    data <- rbind(LnTn, data)
  }

    # check number of columns
    if (ncol(data) %% 3 != 0) {
      .throw_error("The number of columns in 'data' must be a multiple of 3.")
    } else {
      # extract all LxTx values
      data_tmp <- do.call(rbind,
                          lapply(seq(1, ncol(data), 3), function(col) {
                            setNames(data[2:nrow(data), col:c(col+2)], c("dose", "LxTx", "LxTxError"))
                          })
      )

      # extract the LnTn values (assumed to be the first row) and calculate the column mean
      LnTn_tmp <- do.call(rbind,
                          lapply(seq(1, ncol(data), 3), function(col) {
                            setNames(data[1, col:c(col+2)], c("dose", "LxTx", "LxTxError"))
                          })
      )

      # check whether the standard deviation of LnTn estimates or the largest
      # individual error is highest, and take the larger one
      LnTn_error_tmp <- max(c(sd(LnTn_tmp[ ,2]), mean(LnTn_tmp[ ,3])), na.rm = TRUE)
      LnTn_tmp <- colMeans(LnTn_tmp)

      # re-bind the data frame
      data <- rbind(LnTn_tmp, data_tmp)
      data[1, 3] <- LnTn_error_tmp
      data <- data[stats::complete.cases(data), ]
  }

  ## Check 'rhop'
  .validate_class(rhop, c("numeric", "RLum.Results"))

  # check if numeric
  if (is.numeric(rhop)) {
    .validate_length(rhop, 2)
  } else {
    ## alternatively, an RLum.Results object produced by
    ## analyse_FadingMeasurement() can be provided
    .validate_originator(rhop, "analyse_FadingMeasurement")
    rhop <- c(rhop@data$rho_prime$MEAN, rhop@data$rho_prime$SD)
  }

  # check if 'rhop' is actually a positive value
  if (anyNA(rhop) || rhop[1] <= 0 || any(is.infinite(rhop))) {
    .throw_error("'rhop' must be a positive number, the provided value ",
                 "was ", signif(rhop[1], 3), " \u00B1 ", signif(rhop[2], 3))
  }

  ## Check ddot & readerDdot
  .validate_class(ddot, "numeric", length = 2)
  .validate_class(readerDdot, "numeric", length = 2)

  ## set up the parallel cluster
  .validate_positive_scalar(cores, int = TRUE)
  available.cores <- parallel::detectCores()
  cores <- min(cores, available.cores)
  cl <- parallel::makeCluster(cores)
  on.exit(parallel::stopCluster(cl), add = TRUE)

  ## Settings ------------------------------------------------------------------
  extraArgs <- list(...)
  settings <- modifyList(
    list(
      verbose = TRUE,
      n.MC = 10000,
      plot_all_DRC = plot,
      maxiter = 500,
      trace = FALSE),
    extraArgs)

  .validate_positive_scalar(settings$n.MC, int = TRUE, name = "'n.MC'")
  if (settings$n.MC == 1) {
    settings$n.MC <- max(settings$n.MC, 2)
    extraArgs$n.MC <- settings$n.MC
  }

  ## Define Constants ----------------------------------------------------------

  Hs <- 3e15 # s value after Huntley (2006)
  ka <- 1e3 * .const$year_s # in seconds

  ## Define Functions ----------------------------------------------------------
  # fit data using using Eq 5. from Kars et al (2008) employing
  # theta after King et al. (2016)
  theta <- function(t, r) {
    res <- exp(-r * log(1.8 * Hs * (0.5 * t))^3)
    res[!is.finite(res)] <- 0
    return(res)
  }

  ## Preprocessing -------------------------------------------------------------
  readerDdot.error <- readerDdot[2]
  readerDdot <- readerDdot[1]
  ddot.error <- ddot[2]
  ddot <- ddot[1]

  colnames(data) <- c("dose", "LxTx", "LxTx.Error")
  dosetime <- data[["dose"]][2:nrow(data)]
  LxTx.measured <- data[["LxTx"]][2:nrow(data)]
  LxTx.measured.error <- data[["LxTx.Error"]][2:nrow(data)]

  #Keep LnTn separate for derivation of measured fraction of saturation
  Ln <- data[["LxTx"]][1]
  Ln.error <- data[["LxTx.Error"]][1]

  ## set a sensible default for rprime: in most papers the upper boundary is
  ## around 2.2, so setting it to 3 should be enough in general, and 1000
  ## points seem also enough; in any case, we let the user override it
  rprime <- seq(0.01, 3, length.out = 1000)
  if ("rprime" %in% names(extraArgs)) {
    rprime <- extraArgs$rprime
    .validate_class(rprime, "numeric")
  }

  ## (1) MEASURED ----------------------------------------------------
  data.tmp <- data
  data.tmp[ ,1] <- data.tmp[ ,1] * readerDdot

  GC.settings <- list(
    mode = "interpolation",
    fit.method = fit.method[1],
    fit.bounds = TRUE,
    fit.force_through_origin = FALSE,
    verbose = FALSE)

  GC.settings <- modifyList(GC.settings, extraArgs)
  GC.settings$object <- data.tmp
  GC.settings$verbose <- FALSE

  ## unset the n.MC argument so that fit_DoseResponseCurve() won't use it,
  ## which would result in a large performance slowdown (#867)
  GC.settings$n.MC <- NULL

  fit.bounds <- GC.settings$fit.bounds
  force_through_origin <- GC.settings$fit.force_through_origin
  mode_is_extrapolation <- GC.settings$mode == "extrapolation"

  ## call the fitting
  GC.measured <- try(do.call(fit_DoseResponseCurve, GC.settings))

  if (inherits(GC.measured$Fit, "try-error")) {
    .throw_error("Unable to fit growth curve to measured data",
                 ifelse(fit.bounds, ", try setting 'fit.bounds = FALSE'", ""))
  }
  if (settings$plot_all_DRC) {
    plot_DoseResponseCurve(GC.measured, main = "Measured dose response curve",
                           xlab = "Dose [Gy]", verbose = FALSE)
  }

  # extract results and calculate age
  GC.results <- get_RLum(GC.measured)
  fit_measured <- GC.measured@data$Fit
  De.measured <- GC.results$De
  De.measured.error <- GC.results$De.Error
  D0.measured <- GC.results$D01
  D0.measured.error <- GC.results$D01.ERROR
  Age.measured <- De.measured/ ddot
  Age.measured.error <- Age.measured * sqrt( (De.measured.error / De.measured)^2 +
                                               (readerDdot.error / readerDdot)^2 +
                                               (ddot.error / ddot)^2)


  ## (2) SIMULATED -----------------------------------------------------
  # create MC samples
  rhop_MC <- rnorm(n = settings$n.MC, mean = rhop[1], sd = rhop[2])

  if (fit.method == "EXP") {
    model <- LxTx.measured ~ a * theta(dosetime, rhop_i) *
      (1 - exp(-(dosetime + c) / D0))
    start <- list(a = coef(fit_measured)[["a"]],
                  D0 = D0.measured / readerDdot,
                  c = coef(fit_measured)[["c"]])
    lower.bounds <- lower.bounds[1:3]

    ## c = 0 if force_through_origin
    upper.bounds <- c(rep(Inf, 2), if (force_through_origin) 0 else Inf)
  }
  if (fit.method == "GOK") {
    model <- LxTx.measured ~ a * theta(dosetime, rhop_i) *
      (d - (1 + (1 / D0) * dosetime * c)^(-1 / c))
    start <- list(a = coef(fit_measured)[["a"]],
                  D0 = D0.measured / readerDdot,
                  c = coef(fit_measured)[["c"]] * ddot,
                  d = coef(fit_measured)[["d"]])

    ## d = 1 if force_through_origin
    upper.bounds <- c(rep(Inf, 3), if (force_through_origin) 1 else Inf)
  }

  ## do the fitting
  fitcoef <- do.call(rbind, parallel::parLapply(cl, rhop_MC, function(rhop_i) {
    fit_sim <- try({
      minpack.lm::nlsLM(
       formula = model,
       start = start,
       lower = lower.bounds,
       upper = upper.bounds,
       trace = settings$trace,
       control = list(maxiter = settings$maxiter))
    }, silent = TRUE)

    if (!inherits(fit_sim, "try-error"))
      coefs <- coef(fit_sim)
    else {
      ## As the fit from the given starting values failed, try again by fixing
      ## D0 at different values to make the fitting a bit easier. Values are
      ## spaced logarithmically so that the ratio between consecutive values
      ## is constant.
      # all.D0 <- c(start$D0, exp(seq(log(start$D0 / 10), log(start$D0 * 10),
      #                               length.out = 99)))
      ## ALTERNATIVE: sample from a gamma distribution; with only 10, this should
      ## be enough based on tests RLumSK (2025-03-29)
      all.D0 <- stats::rgamma(10, shape = start$D0)
      fit.D0 <- lapply(1:length(all.D0), function(idx) {
        D0 <- all.D0[idx]
        t <- try(minpack.lm::nlsLM(
                  formula = model,
                  start = start[-2],
                  lower = lower.bounds[-2],
                  upper = upper.bounds[-2],
                  control = list(maxiter = settings$maxiter)),
                 silent = TRUE)

        if (inherits(t, "try-error"))
          return(NULL)
        return(t)
      })

      ## pick the one with the best fit after removing those that didn't fit
      fit.D0 <- .rm_NULL_elements(fit.D0)

      ## if also this fails, we should throw an error, but as we are inside
      ## a parallel region, we cannot do that cleanly, so we return NA and
      ## only afterwards we'll throw the error
      if (length(fit.D0) == 0)
        return(NA)

      ## extract the coefficients from the one with the best fit
      fit.D0 <- fit.D0[[which.min(vapply(fit.D0, stats::deviance, numeric(1)))]]
      coefs <- coef(fit.D0)

      ## add back the coefficient for D0
      D0 <- environment(fit.D0$m$predict)$env$D0
      coefs <- c(coefs[1], D0 = D0, coefs[2:length(coefs)])
    }

    return(coefs)
  }))

  ## check if errors had occurred during model fitting
  if (anyNA(fitcoef)) {
    .throw_error("Could not fit simulated curve, check suitability of ",
                 "model and parameters")
  }

  # final fit for export
  # fit_simulated <- minpack.lm::nlsLM(LxTx.measured ~ a * theta(dosetime, rhop[1]) * (1 - exp(-dosetime / D0)),
  #                      start = list(a = max(LxTx.measured), D0 = D0.measured / readerDdot))
  # scaling factor
  A <- mean(fitcoef[, "a"], na.rm = TRUE)
  A.error <- sd(fitcoef[ ,"a"], na.rm = TRUE)

  # calculate measured fraction of saturation
  nN <- Ln / A
  nN.error <- nN * sqrt( (Ln.error / Ln)^2 + (A.error / A)^2)

  # compute a natural dose response curve following the assumptions of
  # Morthekai et al. 2011, Geochronometria
  # natdosetime <- seq(0, 1e14, length.out = settings$n.MC)
  # natdosetimeGray <- natdosetime * ddot / ka

  # Legacy code:
  # calculate D0 dose in seconds
  # computedD0 <- (fitcoef[ ,"D0"] * readerDdot) / (ddot / ka)
  #
  # This is an older approximation to calculate the natural dose response curve,
  # which sometimes tended to slightly underestimate nN_ss. This is now replaced
  # with the newer approach below.
  # compute natural dose response curve
  # LxTx.sim <- A * theta(natdosetime, rhop[1]) * (1 - exp(-natdosetime / mean(computedD0, na.rm = TRUE) ))
  # warning("LxTx Curve: ", round(max(LxTx.sim) / A, 3), call. = FALSE)

  # compute natural dose response curve
  ddots <- ddot / ka
  natdosetimeGray <- c(0, exp(seq(1, log(max(data[ ,1]) * 2), length.out = 999)))
  natdosetime <- natdosetimeGray
  pr <- 3 * rprime^2 * exp(-rprime^3) # Huntley 2006, eq. 3
  K <- Hs * exp(-rhop[1]^-(1/3) * rprime)
  UFD0 <- mean(fitcoef[ ,"D0"], na.rm = TRUE) * readerDdot

  c_val <- mean(fitcoef[, "c"], na.rm = TRUE)
  if (fit.method[1] == "GOK") {
    ## prevent negative c values, which will cause NaN values
    if (c_val < 0) c_val <- 1

    d_gok <- mean(fitcoef[ ,"d"], na.rm = TRUE)
  }

  ## the original formulation used:
  ##  (1) (ddots / UFD0) / (ddots / UFD0 + K[k])
  ##  (2) 1 / UFD0 + K[k] / ddots
  ## which are algebraically equivalent to:
  ##  (1) ddots / (ddots + UFD0 * K[k]) -> scaled.ddots
  ##  (2) 1 / scaled.dots / UFD0
  scaled.ddots <- ddots / (ddots + UFD0 * K)
  A.pr.ddots <- A * pr * scaled.ddots
  inv.UFD0.K <- 1 / scaled.ddots / UFD0
  if (fit.method == "EXP") {
    fun <- function(k) A.pr.ddots[k] *
        (1 - exp(-(natdosetime + c_val) * inv.UFD0.K[k]))
  } else if (fit.method == "GOK") {
    fun <- function(k) A.pr.ddots[k] *
        (d_gok - (1 + inv.UFD0.K[k] * natdosetime * c_val)^(-1 / c_val))
  }
  TermA <- t(vapply(seq_along(rprime), fun, USE.NAMES = FALSE,
                    FUN.VALUE = numeric(length(natdosetime))))

  LxTx.sim <- colSums(TermA) / sum(pr)
  # warning("LxTx Curve (new): ", round(max(LxTx.sim) / A, 3), call. = FALSE)

  # calculate Age
  positive <- which(diff(LxTx.sim) > 0)
  if (length(positive) == 0) {
    .throw_error("All simulated Lx/Tx values are identical and approximately ",
                 "zero. Please verify the accuracy of your rho' value, as this ",
                 "is likely too large and may not be realistic")
  }

  data.unfaded <- data.frame(
    dose = c(0, natdosetimeGray[positive]),
    LxTx = c(Ln, LxTx.sim[positive]),
    LxTx.error = c(Ln.error, LxTx.sim[positive] * A.error/A))
  data.unfaded$LxTx.error[2] <- 0.0001

  ## update the parameter list for fit_DoseResponseCurve()
  GC.settings$object <- data.unfaded

  ## calculate simulated DE
  GC.simulated <- suppressWarnings(try(do.call(fit_DoseResponseCurve, GC.settings)))

  fit_simulated <- NA
  De.sim <- De.error.sim <- D0.sim.Gy <- D0.sim.Gy.error <- NA
  Age.sim <- Age.sim.error <- Age.sim.2D0 <- Age.sim.2D0.error <- NA
  if (!inherits(GC.simulated, "try-error")) {
    if (settings$plot_all_DRC) {
      plot_DoseResponseCurve(GC.simulated, main = "Simulated dose response curve",
                             xlab = "Dose (Gy)", verbose = FALSE)
    }
    GC.simulated.results <- get_RLum(GC.simulated)
    fit_simulated <- get_RLum(GC.simulated, "Fit")
    De.sim <- GC.simulated.results$De

    De.error.sim <- GC.simulated.results$De.Error

    # derive simulated D0
    D0.sim.Gy <- GC.simulated.results$D01
    D0.sim.Gy.error <- GC.simulated.results$D01.ERROR

    Age.sim <- De.sim / ddot
    Age.sim.error <- Age.sim * sqrt( ( De.error.sim/ De.sim)^2 +
                                       (readerDdot.error / readerDdot)^2 +
                                       (ddot.error / ddot)^2)

    Age.sim.2D0 <- 2 * D0.sim.Gy / ddot
    Age.sim.2D0.error <- Age.sim.2D0 * sqrt( ( D0.sim.Gy.error / D0.sim.Gy)^2 +
                                               (readerDdot.error / readerDdot)^2 +
                                               (ddot.error / ddot)^2)
  }

  if (Ln > max(LxTx.sim) * 1.1)
    .throw_warning("Ln is >10 % larger than the maximum computed LxTx value, ",
                   "the De and age should be regarded as infinite estimates")

  if (Ln < min(LxTx.sim) * 0.95 && !mode_is_extrapolation)
    .throw_warning("Ln/Tn is smaller than the minimum computed LxTx value: ",
                   "if, in consequence, your age result is NA, either your ",
                   "input values are unsuitable, or you should consider using ",
                   "a different model for your data")

  if (is.na(D0.sim.Gy)) {
    .throw_error("Simulated D0 is NA: either your input values are unsuitable, ",
                 "or you should consider using a different model for your data")
  }

  # Estimate nN_(steady state) by Monte Carlo Simulation
  ddot_MC <- rnorm(n = settings$n.MC, mean = ddot, sd = ddot.error)
  UFD0_MC <- rnorm(n = settings$n.MC, mean = D0.sim.Gy, sd = D0.sim.Gy.error)

  ## The original formulation was:
  ##
  ##  (1) rho_i <- 3 * alpha^3 * rhop_MC[i] / (4 * pi)
  ##  (2) r <- rprime / (4 * pi * rho_i / 3)^(1 / 3)
  ##  (3) tau <- ((1 / Hs) * exp(1)^(alpha * r)) / ka
  ##
  ## Substituting the expression for `rho_i` into `r`, many simplifications
  ## can be made, so (2) becomes:
  ##
  ##  (2') r <- rprime / (alpha * (rhop_MC[i])^(1 / 3))
  ##
  ## Now, substituting (2') into (3) we get:
  ##
  ##  (3') tau <- (1 / Hs) * exp(rprime / (rhop_MC[i]^(1 / 3))) / ka
  ##
  ## The current formulation then follows:
  ##
  ##  rho_i <- rhop_MC[i]^(1 / 3)
  ##  tau <- (1 / Hs) * exp(rprime / rho_i) / ka

  rho_MC <- rhop_MC^(1 / 3)
  nN_SS_MC <- mapply(function(rho_i, ddot_i, UFD0_i) {
    tau <- (1 / Hs) * exp(rprime / rho_i) / ka
    Ls <- 1 / (1 + UFD0_i / (ddot_i * tau))
    Lstrap <- (pr * Ls) / sum(pr)

    # field saturation
    nN_SS_i <- sum(Lstrap)
    return(nN_SS_i)

  }, rho_MC, ddot_MC, UFD0_MC, SIMPLIFY = TRUE)

  nN_SS <- suppressWarnings(exp(mean(log(nN_SS_MC), na.rm = TRUE)))
  nN_SS.error <- suppressWarnings(nN_SS * abs(sd(log(nN_SS_MC), na.rm = TRUE) / mean(log(nN_SS_MC), na.rm = TRUE)))

  ## legacy code for debugging purposes
  ## nN_SS is often lognormally distributed, so we now take the mean and sd
  ## of the log values.
  # warning(mean(nN_SS_MC, na.rm = TRUE))
  # warning(sd(nN_SS_MC, na.rm = TRUE))

  ## (3) UNFADED ---------------------------------------------------------------
  LxTx.unfaded <- LxTx.measured / theta(dosetime, rhop[1])

  ## set Inf and NaN values to 0
  LxTx.unfaded[!is.finite(LxTx.unfaded)] <- 0
  dosetimeGray <- dosetime * readerDdot

  ## run this first model also for GOK as in general it provides more
  ## stable estimates that can be used as starting point for GOK
  fit_unfaded <- try(minpack.lm::nlsLM(
      LxTx.unfaded ~ a * (1 - exp(-(dosetimeGray + c) / D0)),
      start = list(
        a = coef(fit_simulated)[["a"]],
        D0 = D0.measured / readerDdot,
        c = coef(fit_simulated)[["c"]]),
        upper = if(force_through_origin) {
           c(a = Inf, D0 = max(dosetimeGray), c = 0)
          } else {
           c(Inf, max(dosetimeGray), Inf)
          },
        lower = lower.bounds[1:3],
        trace = settings$trace,
      control = list(maxiter = settings$maxiter)), silent = TRUE)

  ## if this fit has failed, what we do depends on fit.method:
  ## - for EXP, this error is irrecoverable
  if (inherits(fit_unfaded, "try-error") && fit.method == "EXP") {
    .throw_error("Could not fit unfaded curve, check suitability of ",
                 "model and parameters")
  }

  ## - for GOK, we use the simulated fit to set the starting point
  if (fit.method == "GOK") {
    fit_start <- if (inherits(fit_unfaded, "try-error"))
                   fit_simulated else fit_unfaded

    fit_unfaded <- try(minpack.lm::nlsLM(
      LxTx.unfaded ~ a * (d-(1+(1/D0)*dosetimeGray*c)^(-1/c)),
      start = list(
        a = coef(fit_start)[["a"]],
        D0 = coef(fit_start)[["D0"]],
        c = max(coef(fit_start)[["c"]], 1),
        d = coef(fit_simulated)[["d"]]),
      upper = if(force_through_origin) {
        c(a = Inf, D0 = max(dosetimeGray), c = Inf, d = 1)
       } else {
        c(Inf, max(dosetimeGray), Inf, Inf)},
      lower = lower.bounds[1:4],
      control = list(maxiter = settings$maxiter)), silent = TRUE)

    if(inherits(fit_unfaded, "try-error"))
      .throw_error("Could not fit unfaded curve, check suitability of ",
                   "model and parameters")
  }

  D0.unfaded <- coef(fit_unfaded)[["D0"]]
  D0.error.unfaded <- summary(fit_unfaded)$coefficients["D0", "Std. Error"]

  ## Create LxTx tables --------------------------------------------------------
  # normalise by A (saturation point of the un-faded curve)
  if (normalise) {
    LxTx.measured <- LxTx.measured / A
    LxTx.measured.error <- LxTx.measured.error / A

    LxTx.sim <- LxTx.sim / A
    LxTx.unfaded <- LxTx.unfaded / A

    Ln <- Ln / A
    Ln.error <- Ln.error / A
  }

  # combine all computed LxTx values
  LxTx_measured <- data.frame(
    dose = dosetimeGray,
    LxTx = LxTx.measured,
    LxTx.Error = LxTx.measured.error)

  LxTx_simulated <- data.frame(
    dose = natdosetimeGray,
    LxTx = LxTx.sim,
    LxTx.Error = LxTx.sim * A.error / A)

  LxTx_unfaded <- data.frame(
    dose = dosetimeGray,
    LxTx = LxTx.unfaded,
    LxTx.Error = LxTx.unfaded * A.error / A)

  ## Plot settings -------------------------------------------------------------
  plot.settings <- modifyList(list(
    main = "Dose response curves",
    cex = 1,
    xlab = "Dose [Gy]",
    ylab = ifelse(normalise, "normalised LxTx [a.u.]", "LxTx [a.u.]")
  ), extraArgs)

  ## Plotting ------------------------------------------------------------------
  if (plot) {
    ### par settings ---------
    par.default <- .par_defaults()
    on.exit(par(par.default), add = TRUE)

    # set graphical parameters
    par(mfrow = c(1,1), mar = c(4.5, 4, 4, 4), cex = 0.8 * plot.settings$cex,
        oma = c(0, 0, 0, if (summary) 12 / plot.settings$cex else 0))

    # Find a good estimate of the x-axis limits
    if (mode_is_extrapolation && !force_through_origin) {
      dosetimeGray <- c(-De.measured - De.measured.error, dosetimeGray)
      De.measured <- -De.measured
    }

    xlim <- range(pretty(c(dosetimeGray, De.sim), n = 15))

    # Create figure after Kars et al. (2008) contrasting the dose response curves
    ## open plot window ------------
    plot(
      x = dosetimeGray[dosetimeGray >= 0],
      y = LxTx_measured$LxTx,
      main = plot.settings$main,
      xlab = plot.settings$xlab,
      ylab = plot.settings$ylab,
      pch = 16,
      ylim = c(0, max(do.call(rbind, list(LxTx_measured, LxTx_unfaded))[["LxTx"]])),
      xlim = xlim
    )

    ## add horizontal line at zero
    abline(v = 0, h = 0, col = "gray")

    # LxTx error bars
    segments(x0 = dosetimeGray[dosetimeGray >= 0],
             y0 = LxTx_measured$LxTx + LxTx_measured$LxTx.Error,
             x1 = dosetimeGray[dosetimeGray >= 0],
             y1 = LxTx_measured$LxTx - LxTx_measured$LxTx.Error,
             col = "black")

    # re-calculate the measured dose response curve in Gray
    xNew <- seq(par()$usr[1],par()$usr[2], length.out = 200)
    yNew <- predict(GC.measured@data$Fit, list(x = xNew))
    if (normalise)
      yNew <- yNew / A

    ## add measured curve -------
    lines(xNew, yNew, col  = "black")

    # add error polygon
    polygon(x = c(natdosetimeGray, rev(natdosetimeGray)),
            y = c(LxTx_simulated$LxTx + LxTx_simulated$LxTx.Error,
                  rev(LxTx_simulated$LxTx - LxTx_simulated$LxTx.Error)),
            col = adjustcolor("grey", alpha.f = 0.5), border = NA)

    ## add simulated curve -------
    xNew <- seq(if (mode_is_extrapolation) par()$usr[1] else 0,
                par()$usr[2], length.out = 200)
    yNew <- predict(GC.simulated@data$Fit, list(x = xNew))
    if (normalise)
      yNew <- yNew / A
    points(
      x = xNew,
      y = yNew,
      type = "l",
      lty = 3)

    # Ln and DE as points
    points(x = if (mode_is_extrapolation)
                 rep(De.measured, 2)
               else
                 c(0, De.measured),
           y = if (mode_is_extrapolation)
                c(0,0)
               else
                c(Ln, Ln),
           col = "red",
           pch = c(2, 16))

    # Ln error bar
    segments(x0 = 0, y0 = Ln - Ln.error,
             x1 = 0, y1 = Ln + Ln.error,
             col = "red")

    # Ln as a horizontal line
    lines(x = if (mode_is_extrapolation)
                c(0, min(c(De.measured, De.sim), na.rm = TRUE))
              else
                c(par()$usr[1], max(c(De.measured, De.sim, 0), na.rm = TRUE)),
          y = c(Ln, Ln),
          col = "red ", lty = 3)

    #vertical line of measured DE
    lines(x = c(De.measured, De.measured),
          y = c(par()$usr[3], Ln),
          col = "red",
          lty = 3)

    # add legends
    legend("bottomright",
           legend = c(
             "Unfaded DRC",
             "Measured DRC",
             "Simulated natural DRC"),
           inset = c(0, 0.04),
           lty = c(5, 1, 3),
           bty = "n",
           cex = 0.8)

    # add vertical line of simulated De
    if (!is.na(De.sim)) {
      lines(x = if (mode_is_extrapolation)
                  c(-De.sim, -De.sim)
                else
                  c(De.sim, De.sim),
            y = c(par()$usr[3], Ln),
            col = "red", lty = 3)

      points(x = if (mode_is_extrapolation) -De.sim else De.sim,
             y = if (mode_is_extrapolation) 0 else Ln,
             col = "red" , pch = 16)
    } else {
      lines(x = c(De.measured, xlim[2]),
            y = c(Ln, Ln), col = "black", lty = 3)
    }

    # add unfaded DRC --------
    yNew <- predict(fit_unfaded, list(dosetimeGray = xNew))
    if (normalise)
      yNew <- yNew / A

    lines(xNew, yNew, col  = "black", lty = 5)

    points(x = dosetimeGray[dosetimeGray >= 0],
           y = LxTx_unfaded$LxTx,
           col = "black")

    # LxTx error bars
    segments(
      x0 = dosetimeGray[dosetimeGray >= 0],
      y0 = LxTx_unfaded$LxTx + LxTx_unfaded$LxTx.Error,
      x1 = dosetimeGray[dosetimeGray >= 0],
      y1 = LxTx_unfaded$LxTx - LxTx_unfaded$LxTx.Error,
      col = "black")

    # add text
    if (summary) {
      # define labels as expressions
      labels.text <- list(
        bquote(dot(D) == .(format(ddot, digits = 2, nsmall = 2)) %+-% .(round(as.numeric(format(ddot.error, digits = 3, nsmall = 3)), 3)) ~ frac(Gy, ka)),
        bquote(dot(D)["Reader"] == .(format(readerDdot, digits = 2, nsmall = 2)) %+-% .(round(as.numeric(format(readerDdot.error, digits = 3, nsmall = 3)), 3)) ~ frac(Gy, s)),
        bquote(log[10]~(rho~"'") == .(format(log10(rhop[1]), digits = 2, nsmall = 2)) %+-% .(round(as.numeric(format(rhop[2] / (rhop[1] * log(10, base = exp(1))), digits = 2, nsmall = 2)), 2)) ),
        bquote(bgroup("(", frac(n, N), ")") == .(format(nN, digits = 2, nsmall = 2)) %+-% .(round(as.numeric(format(nN.error, digits = 2, nsmall = 2)), 2)) ),
        bquote(bgroup("(", frac(n, N), ")")[SS] == .(format(nN_SS, digits = 2, nsmall = 2)) %+-% .(round(nN_SS.error, 2))),
        bquote(D["E,sim"] == .(format(De.sim, digits = 1, nsmall = 0)) %+-% .(format(De.error.sim, digits = 1, nsmall = 0)) ~ Gy),
        bquote(D["0,sim"] == .(format(D0.sim.Gy, digits = 1, nsmall = 0)) %+-% .(format(D0.sim.Gy.error, digits = 1, nsmall = 0)) ~ Gy),
        bquote(Age["sim"] == .(format(Age.sim, digits = 1, nsmall = 0)) %+-% .(format(Age.sim.error, digits = 1, nsmall = 0)) ~ ka)
      )

      # each of the labels is positioned at 1/10 of the available y-axis space
      ypos <- seq(range(axTicks(2))[2], range(axTicks(2))[1], length.out = 10)
      for (i in seq_along(labels.text))
        mtext(labels.text[[i]], at = ypos[i],
              side = 4, line = 1, las = 1, padj = 1)
    }
  }

  ## Results -------------------------------------------------------------------
  results <- set_RLum(
    class = "RLum.Results",
    data = list(
      results = data.frame(
        nN = nN,
        nN.error = nN.error,
        nN_SS = nN_SS,
        nN_SS.error = nN_SS.error,
        Meas_De = abs(De.measured),
        Meas_De.error = De.measured.error,
        Meas_D0 =  D0.measured,
        Meas_D0.error = D0.measured.error,
        Meas_Age = Age.measured,
        Meas_Age.error = Age.measured.error,
        Sim_De = De.sim,
        Sim_De.error = De.error.sim,
        Sim_D0 = D0.sim.Gy,
        Sim_D0.error = D0.sim.Gy.error,
        Sim_Age = Age.sim,
        Sim_Age.error = Age.sim.error,
        Sim_Age_2D0 = Age.sim.2D0,
        Sim_Age_2D0.error = Age.sim.2D0.error,
        Unfaded_D0 = D0.unfaded,
        Unfaded_D0.error = D0.error.unfaded,
        row.names = NULL),
      data = data,
      Ln = c(Ln, Ln.error),
      LxTx_tables = list(
        simulated = LxTx_simulated,
        measured = LxTx_measured,
        unfaded = LxTx_unfaded),
      fits = list(
        simulated = fit_simulated,
        measured = fit_measured,
        unfaded = fit_unfaded
      )
    ),
    info = list(
      call = sys.call(),
      args = as.list(sys.call())[-1])
  )

  ## Console output ------------------------------------------------------------
  if (settings$verbose) {
    cat("\n\n[calc_Huntley2006()]\n")
    cat("\n -------------------------------")
    cat("\n (n/N) [-]:\t",
        round(results@data$results$nN, 2), "\u00b1",
        round(results@data$results$nN.error, 2))
    cat("\n (n/N)_SS [-]:\t",
        round(results@data$results$nN_SS, 2),"\u00b1",
        round(results@data$results$nN_SS.error, 2))
    cat("\n\n ---------- Measured -----------")
    cat("\n DE [Gy]:\t",
        round(results@data$results$Meas_De, 2), "\u00b1",
        round(results@data$results$Meas_De.error, 2))
    cat("\n D0 [Gy]:\t",
        round(results@data$results$Meas_D0, 2), "\u00b1",
        round(results@data$results$Meas_D0.error, 2))
    if (fit.method[1] == "GOK") {
      cat("\n c [-]:\t\t",
          round(summary(fit_measured)$coefficients["c", "Estimate"], 2), "\u00b1",
          round(summary(fit_measured)$coefficients["c", "Std. Error"], 2))
    }
    cat("\n Age [ka]:\t",
        round(results@data$results$Meas_Age, 2), "\u00b1",
        round(results@data$results$Meas_Age.error, 2))
    cat("\n\n ---------- Un-faded -----------")
    cat("\n D0 [Gy]:\t",
        round(results@data$results$Unfaded_D0, 2), "\u00b1",
        round(results@data$results$Unfaded_D0.error, 2))
    if (fit.method[1] == "GOK") {
      cat("\n c [-]:\t\t",
          round(summary(fit_unfaded)$coefficients["c", "Estimate"], 2), "\u00b1",
          round(summary(fit_unfaded)$coefficients["c", "Std. Error"], 2))
    }
    cat("\n\n ---------- Simulated ----------")
    cat("\n DE [Gy]:\t",
        round(results@data$results$Sim_De, 2), "\u00b1",
        round(results@data$results$Sim_De.error, 2))
    cat("\n D0 [Gy]:\t",
        round(results@data$results$Sim_D0, 2), "\u00b1",
        round(results@data$results$Sim_D0.error, 2))
    if (fit.method[1] == "GOK") {
      cat("\n c [-]:\t\t",
          round(summary(fit_simulated)$coefficients["c", "Estimate"], 2), "\u00b1",
          round(summary(fit_simulated)$coefficients["c", "Std. Error"], 2))
    }
    cat("\n Age [ka]:\t",
        round(results@data$results$Sim_Age, 2), "\u00b1",
        round(results@data$results$Sim_Age.error, 2))
    cat("\n Age @2D0 [ka]:\t",
        round(results@data$results$Sim_Age_2D0, 2), "\u00b1",
        round(results@data$results$Sim_Age_2D0.error, 2))
    cat("\n -------------------------------\n\n")
  }

  ## Return value --------------------------------------------------------------
  return(results)
}

test_that("Check the example and the numerical values", {
  testthat::skip_on_cran()
  local_edition(3)

  ##construct empty object
  temp <-
    set_Risoe.BINfileData(METADATA = data.frame(), DATA = list(), .RESERVED = list())

  ##get function and check whether we get NULL
  expect_null(get_Risoe.BINfileData(temp))

  ##check object
  expect_s4_class(temp, class = "Risoe.BINfileData")
  expect_output(show(temp))

  ##show method
  data(ExampleData.BINfileData, envir = environment())
  expect_output(show(CWOSL.SAR.Data))

  ##as.data.frame
  expect_s3_class(as.data.frame(CWOSL.SAR.Data), "data.frame")

})

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(set_RLum(1),
               "'class' should be of class 'character' and have length 1")
  expect_error(set_RLum(character()),
               "'class' should be of class 'character' and have length 1")
  expect_error(set_RLum(letters),
               "'class' should be of class 'character' and have length 1")
  expect_error(set_RLum(""),
               "'class' cannot be an empty character")
})

## load data
data(ExampleData.XSYG, envir = environment())
object <- set_RLum(class = "RLum.Data.Curve")

test_that("check class", {
  testthat::skip_on_cran()

  ##set empty curve object and show it
  expect_output(show(set_RLum(class = "RLum.Data.Curve")))

  expect_error(set_RLum("RLum.Data.Curve", data = matrix(1:5, ncol = 1)))

  ##check replacements fo
  expect_s4_class(set_RLum(class = "RLum.Data.Curve", data = object), class = "RLum.Data.Curve")

  ##check get_RLum
  expect_error(get_RLum(object, info.object = list()),
               "'info.object' should be of class 'character' or NULL and have length 1")
  expect_warning(expect_null(get_RLum(object, info.object = "test")),
                 "'object' has no info objects, NULL returned")
  object <- set_RLum(class = "RLum.Data.Curve", data = object, info = list(a = "test"))
  expect_warning(get_RLum(object, info.object = "error"),
                 "Invalid 'info.object' name, valid names are:")

  ##test names
  expect_type(names(object), "character")

  ##test bin
  expect_error(bin_RLum.Data(object, bin_size = -2),
               "'bin_size' should be a single positive integer value")

  ##check conversions
  expect_s4_class(as(object = list(1:10), Class = "RLum.Data.Curve"), "RLum.Data.Curve")
  expect_s4_class(as(list(), "RLum.Data.Curve"),
                  "RLum.Data.Curve")
  expect_type(as(object = object, Class = "list"), "list")
  expect_s4_class(as(object = matrix(1:10,ncol = 2), Class = "RLum.Data.Curve"), "RLum.Data.Curve")

  ## show
  expect_output(print(object),
                "range of y-values: 0 0")
  object@data[, 2] <- NA
  expect_output(print(object),
                "range of y-values: Inf -Inf (contains NA values)",
                fixed = TRUE)
})

test_that("melt_RLum", {
  testthat::skip_on_cran()

  ## test melt simple
  expect_type(melt_RLum(object = object), "list")

  ## test melt more complicated
  t <- melt_RLum(OSL.SARMeasurement[[2]][[1]])
  expect_length(t, n = 4)

  ## provide a list
  expect_length(melt_RLum(OSL.SARMeasurement[[2]]@records[1:5]), n = 4)

  ## test problematic records
  l <- c(OSL.SARMeasurement[[2]]@records[1:5], "non-RLum")
  expect_length(melt_RLum(l), n = 4)

  l <- list("non-RLum")
  expect_null(melt_RLum(l))

  l <- c(list("non-RLum"), set_RLum("RLum.Results"))
  expect_type(melt_RLum(l), "list")
})

## load data
data(ExampleData.RLum.Analysis, envir = environment())
temp <- IRSAR.RF.Data

test_that("input validation", {
  testthat::skip_on_cran()

  expect_error(subset(temp, LTYPE == "RF"),
               "[get_RLum()] Invalid subset expression, valid terms are",
               fixed = TRUE)
  SW({
  expect_message(expect_null(subset(temp, recordType == "xx")),
                 "'subset' expression produced an empty selection, NULL returned")
  expect_message(expect_null(subset(temp, recordType == "RF")),
                 "'subset' expression produced an empty selection, NULL returned")
  })
})

test_that("check functionality", {
  testthat::skip_on_cran()

  ### empty call
  expect_s4_class(subset(temp), "RLum.Analysis")
  expect_length(subset(temp), length(temp))
  expect_identical(subset(temp)[[1]], temp[[1]])

  ### valid
  expect_s4_class(subset(temp, recordType == "RF (NA)"), class = "RLum.Analysis")
  expect_s4_class(subset(temp, recordType == "RF (NA)")[[1]], class = "RLum.Data.Curve")
  expect_length(subset(temp, recordType == "RF (NA)"), n = length(temp))

  ## get_RLum(<obj>, subset = (<condition>))
  expect_s4_class(get_RLum(temp, subset = recordType == "RF (NA)"),
                  "RLum.Analysis")
  expect_s4_class(get_RLum(temp, subset = recordType == "RF (NA)")[[1]],
                  "RLum.Data.Curve")
  expect_length(get_RLum(temp, subset = recordType == "RF (NA)"),
                length(temp))
})

test_that("Test replication of RLum-objects", {
  testthat::skip_on_cran()

  data(ExampleData.RLum.Analysis, envir = environment())
  expect_silent(results <- rep(IRSAR.RF.Data[[1]], 5))
  expect_silent(rep(IRSAR.RF.Data[[1]]))

  ##check
  expect_equal(length(results),5)

})

test_that("check class", {
  testthat::skip_on_cran()

  ##set empty spectrum object and show it
  expect_output(show(set_RLum(class = "RLum.Data.Spectrum")))

  ##check replacements
  object <- set_RLum(class = "RLum.Data.Spectrum")
  expect_s4_class(set_RLum(class = "RLum.Data.Spectrum", data = object), class = "RLum.Data.Spectrum")

  ##check get_RLum
  object <- set_RLum(class = "RLum.Data.Spectrum", data = object, info = list(a = "test"))
  expect_error(get_RLum(object, info.object = "test"),
               "Invalid element name, valid names are: 'a'")
  expect_error(get_RLum(object, info.object = 1L),
               "'info.object' should be of class 'character'")
  expect_error(get_RLum(object, info.object = list()),
               "'info.object' should be of class 'character' or NULL and have length 1")
  expect_type(get_RLum(object, info.object = "a"), "character")

  ##test method names
  expect_type(names(object), "character")

  ##test bin_RLum()
  expect_error(bin_RLum.Data(object),
               "'object' contains no data")
  object@data <- matrix(data = rep(1:20, each = 10), ncol = 20)
  rownames(object@data) <- 1:10
  colnames(object@data) <- 1:20
  expect_error(bin_RLum.Data(object, bin_size.row = "test"),
               "'bin_size.row' should be a single positive integer value")
  expect_error(bin_RLum.Data(object, bin_size.row = 12, bin_size.col = "test"),
               "'bin_size.col' should be a single positive integer value")
  expect_s4_class(object = bin_RLum.Data(object, bin_size.row = 2), "RLum.Data.Spectrum")

  expect_s4_class(object = bin_RLum.Data(object, bin_size.row = 1, bin_size.col = 2), "RLum.Data.Spectrum")

  ##check conversions
  expect_s4_class(as(object = data.frame(x = 1:10), Class = "RLum.Data.Spectrum"), "RLum.Data.Spectrum")
  expect_s3_class(as(set_RLum("RLum.Data.Spectrum"), "data.frame"), "data.frame")
  expect_s4_class(as(object = matrix(1:10,ncol = 2), Class = "RLum.Data.Spectrum"), "RLum.Data.Spectrum")
  expect_s4_class(as(list(1:10), "RLum.Data.Spectrum"),
                  "RLum.Data.Spectrum")
  expect_s4_class(as(list(), "RLum.Data.Spectrum"),
                  "RLum.Data.Spectrum")
  expect_type(as(set_RLum("RLum.Data.Spectrum"), "list"),
              "list")
})

test_that("check functionality", {
  testthat::skip_on_cran()

  data(ExampleData.RLum.Analysis, envir = environment())
  expect_silent(names_RLum(IRSAR.RF.Data))
  expect_type(names_RLum(IRSAR.RF.Data), "character")

  ##test a list of such elements
  expect_type(names_RLum(list(IRSAR.RF.Data,IRSAR.RF.Data, "a")), "list")

  ## empty object
  res <- names_RLum(set_RLum("RLum.Analysis"))
  expect_type(res, "character")
  expect_length(res, 0)
})

test_that("check class ", {
  testthat::skip_on_cran()

  ##load example data
  data(ExampleData.RLum.Data.Image, envir = environment())

  ##set-method
  ##set empty class
  expect_s4_class(set_RLum(class = "RLum.Data.Image"), "RLum.Data.Image")

  ##overwrite only data
  expect_s4_class(set_RLum(class = "RLum.Data.Image", data = set_RLum("RLum.Data.Image")), "RLum.Data.Image")

  ##show-method
  ##show example data
  expect_output(show(ExampleData.RLum.Data.Image))

  ##get-method
  expect_error(get_RLum(ExampleData.RLum.Data.Image, info.object = 1),
               "'info.object' should be of class 'character'")
  expect_error(get_RLum(ExampleData.RLum.Data.Image, info.object = list()),
               "'info.object' should be of class 'character' or NULL and have length 1")
  expect_error(get_RLum(ExampleData.RLum.Data.Image, info.object = "unknown"))
  expect_type(get_RLum(ExampleData.RLum.Data.Image, info.object = "NumFrames"), "integer")

  ##names
  expect_type(names_RLum(ExampleData.RLum.Data.Image), "character")

  ##conversions
  ##from matrix and to matrix
  expect_s4_class(as(matrix(1:10, ncol = 2), "RLum.Data.Image"), "RLum.Data.Image")
  expect_type(as(ExampleData.RLum.Data.Image, "matrix"), "integer")

  ##from data.frame and to data.frame
  df <- as.data.frame(as(ExampleData.RLum.Data.Image, "matrix"))
  expect_s4_class(as(df, "RLum.Data.Image"), "RLum.Data.Image")
  expect_s3_class(as(ExampleData.RLum.Data.Image, "data.frame"), "data.frame")

  ## to and from array
  expect_type(as(ExampleData.RLum.Data.Image, "array"), "integer")
  from_array <- expect_s4_class(as(array(1,dim = c(10,10,2)), "RLum.Data.Image"), "RLum.Data.Image")

  ## to and from list
  expect_s4_class(as(list(matrix(1, nrow = 10, ncol = 5), matrix(1, nrow = 10, ncol = 5)), "RLum.Data.Image"),
                  "RLum.Data.Image")
  expect_s4_class(as(list(), "RLum.Data.Image"),
                  "RLum.Data.Image")
  expect_type(as(ExampleData.RLum.Data.Image, "list"), "list")
  res <- as(set_RLum("RLum.Data.Image"), "list")
  expect_type(res, "list")
  expect_length(res, 0)

  ## check edge cases
  expect_error(as(from_array, "matrix"),
               "object contains multiple frames, please convert to array instead")
  expect_error(as(from_array, "data.frame"),
               "No viable coercion to data.frame, object contains multiple frames")
})

---
title: "Assessing crosstalk in single-grain luminescence detection"
output:
    rmarkdown::html_vignette:
     toc: true
author: "Luc Steinbuch, Anna-Maartje de Boer, Wageningen University & Research"
date: "February 2025"
vignette: >
  %\VignetteIndexEntry{Assessing crosstalk in single-grain luminescence detection}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Crosstalk

This vignette explores the crosstalk-related functions in the "Luminescence"
package, related to  research paper "A novel tool to assess crosstalk in
single-grain luminescence detection", by Anna-Maartje de Boer, Luc Steinbuch,
Gerard Heuvelink and Jakob Wallinga.
Crosstalk in single-grain luminescence imaging (EMCCD) is the overlapping of luminescence signals from adjacent grains on a single-grain disc. The actual signal on one grain location influences the observed signal on a neighboring grain location which happens to be on the same measurement disc (one "position" in the reader). In this research and the shown functions, we define "neighboring" as rook-wise (horizontally and vertically) only, and we assume that all measurement discs have a regular grid of 10x10 grain locations.


```{r init}
library(Luminescence)
```


# Explore imaged luminescence signals from a single-grain disc (one "position" in the reader)

All grain location observations on one single-grain measurement disc are represented by a vector of 100 numbers. Here we simulate such a disc by randomly selecting from two normal distributions. The prefix "vn_" means: a vector of numbers:

```{r simulate_signal, fig.dim = c(8, 6), out.width="800px"}

vn_simulated <- sample(x = c(rnorm(n = 30, mean = 2000, sd = 500),
                             rnorm(n = 70, mean = 20, sd = 1)),
                       size = 100)

vn_simulated <- round(vn_simulated) # Because photons are discrete

head(vn_simulated, n = 25)

hist(vn_simulated,
     main = "Simulated signal (histogram)",
     xlab = "Photon counts",
     ylab = "Frequency",
     breaks = 30
     )
```

## plot_SingleGrainDisc

Let's visualize the disc/position:

```{r show_disc_basic, fig.dim = c(7, 7), out.width="800px"}

par(mar = c(1, 4, 6, 4))

plot_SingleGrainDisc(object = vn_simulated,
          main = "Simulated signal (measurement disc)"
          )

```


Let's calculate Moran's I, and the associated pseudo-p, of our randomly ordered simulated disc:

```{r calculate_moransI}

calc_MoransI(object = vn_simulated)

calc_MoransI(object = vn_simulated, compute_pseudo_p = TRUE)


```
what changes if we add a serious amount, say 10%, of crosstalk?

```{r add_crosstalk, fig.dim = c(8, 6), out.width="800px"}

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                              n_crosstalk = 0.10)

vn_simulated_with_crosstalk <- round(vn_simulated_with_crosstalk)

hist(vn_simulated_with_crosstalk,
     main = "Simulated signal with crosstalk (histogram)",
     xlab = "Photon counts",
     ylab = "Frequency",
     breaks = 30
     )

plot_SingleGrainDisc(object = vn_simulated_with_crosstalk,
          main = "Simulated signal with crosstalk (measurement disc)")

calc_MoransI(object = vn_simulated_with_crosstalk)

calc_MoransI(object = vn_simulated_with_crosstalk, compute_pseudo_p = TRUE)


```

Let's try several amounts of simulated crosstalk:


```{r test_different_amounts, fig.dim = c(8, 6), out.width="800px"}


df_MoransI <- data.frame(crosstalk = seq(from = 0,
                    to = 0.30,
                    length.out=50))

df_MoransI$MoransI <- NA
df_MoransI$pseudo_p <- NA

old.opts <- options(warn = -1) # silence warnings from compute_pseudo_p
for (i in 1:nrow(df_MoransI))
{

  vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                              n_crosstalk = df_MoransI$crosstalk[i])

  df_MoransI$MoransI[i]  <- calc_MoransI(object = vn_simulated_with_crosstalk)
  df_MoransI$pseudo_p[i] <-
    calc_MoransI(object = vn_simulated_with_crosstalk, compute_pseudo_p = TRUE)
}
options(old.opts) # restore the default options

n_expected_I_no_spatial_autocorr <- calc_MoransI(1:100,
                                                 spatial_autocorrelation = FALSE)

##

plot(x = df_MoransI$crosstalk,
     y = df_MoransI$MoransI,
     ylim = range(
          pretty(x = c(df_MoransI$MoransI, n_expected_I_no_spatial_autocorr))
          ),
     ## Set ylim manually to make sure the value for I for no crosstalk is visible
     xlab = "Amount of added crosstalk",
     ylab = "Calculated Moran's I"
     )
graphics::grid()
abline(h = n_expected_I_no_spatial_autocorr,
       col = "purple")

legend(x = "topleft",
       legend = "Expected I if no spatial autocorrelation",
       lty = "solid",
       col = "purple",
       cex = 0.8)

plot(x = df_MoransI$crosstalk,
     y = df_MoransI$pseudo_p,
     xlab = "Amount of added crosstalk",
     ylab = "Generated pseudo-p of related Moran's I")
graphics::grid()
```

Please note that above two plots are subject to randomness; for a good assessment many simulations have to be performed.


# Moran scatterplot


A way to visualise spatial auto-correlation is the Moran scatterplot:
```{r show_`disc_extended_moran_scatterplot, fig.dim = c(8, 5), out.width="800px"}


plot_MoranScatterplot(object = vn_simulated,
           main = "Moran scatterplot, simulated signal without crosstalk")

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                              n_crosstalk = 0.25)
vn_simulated_with_crosstalk <- round(vn_simulated_with_crosstalk)

plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
           main = "Moran scatterplot, simulated signal with added crosstalk")
```

The plot area is divided into four quadrants using the mean in each dimension; the South-west and North-east quadrant represent a contribution to a positive spatial autocorrelation, while the North-west and South-east quadrants indicate a negative spatial correlation. Between the point, a least square line (which slopes indicates, but not exactly represents, Moran's I) is added, as well as an 1:1 line (which indicates a Moran's I of around 1, suggesting a perfect positive spatial correlation).

The internal function `.get_Neighbours()` was until now used on the
background, but we can explicitly call it to generate a dataframe with all
positions which are rook connected to each other (note that we need to use
the `Luminescence:::` prefix as this function is not exported by the package):

```{r adjacent_grain_locations, fig.dim = c(8, 5), out.width="800px"}

    vn_simulated_with_holes <- c(rnorm(30, mean = 10, sd = 5),
                              rnorm(30, mean = 500, sd = 50),
                              rep(NA, times = 40)
                              )

df_Neighbours <- Luminescence:::.get_Neighbours(object = vn_simulated_with_holes)

head(df_Neighbours)
```

And we can plot the first disc, while indicating which borders are taken into account. The corresponding `adjacent_grain_locations` is calculated in the background:

```{r adjacent_grain_locations_show_disc, fig.dim = c(8, 5), out.width="800px"}

    plot_SingleGrainDisc(object = vn_simulated_with_holes,
              show_neighbours = TRUE,
              show_location_ids = TRUE)
```
Also other functions, such as `calc_MoransI()`, work as long as the remaining grid does not become too sparse. Note that in this context all observations, even "islands" who do not border any other grains, are used for the calculation of Moran's I.

# Add borders with certain weight

One can manually change `df_Neighbours`, for example add the diagonal borders and attain a certain weight to it, and add it as argument to almost all functions. Assume that we want to add a few diagonal borders with weight 1/sqrt(2) to a full disc (note that the standard relative weight for rook borders is set to one):


```{r adjacent_grain_locations_add, fig.dim = c(8, 5), out.width="800px"}

df_Neighbours_with_diag <- Luminescence:::.get_Neighbours(object = vn_simulated)


for (i in c(1:9, 11:19, 21:29) )
{
  df_Neighbours_with_diag <- rbind(df_Neighbours_with_diag,
                                            c(i, i+11, 1/sqrt(2))
                                          )
}

tail(df_Neighbours_with_diag)


plot_SingleGrainDisc(object = vn_simulated,
          df_neighbours = df_Neighbours_with_diag,
          show_neighbours = TRUE,
          show_location_ids = TRUE)
```

To exclude all border effects, we can use the `ignore_borders` option
available in `calc_MoransI()` and `plot_SingleGrainDisc()`, which can
take out all border rows and columns when computing the data frame of
neighbours:


``` {r restrict_to_inner_8x8, fig.dim = c(8, 5), out.width="800px"}

vn_values_to_show <-
      sample(x = c(rnorm(n = 30, mean = 2000, sd = 500),
                   rnorm(n = 70, mean = 20, sd = 1)),
             size = 100)

## Set the outer rows to NA before adding crosstalk
vn_disc_border_locations <- c(1:10,
                              91:100,
                              seq(from = 11, to = 81, by = 10),
                              seq(from = 20, to = 90, by = 10)
                             )
vn_values_to_show[vn_disc_border_locations] <- NA

vn_values_to_show <- apply_Crosstalk(object = vn_values_to_show,
                                     n_crosstalk = 0.15)

calc_MoransI(object = vn_values_to_show)

plot_SingleGrainDisc(object = vn_values_to_show,
                     show_neighbours = TRUE,
                     ignore_borders = TRUE)

calc_MoransI(object = vn_values_to_show,
             ignore_borders = TRUE)

```


# Plot disc options

```{r show_disc_extended, fig.dim = c(8, 5), out.width="800px"}

plot_SingleGrainDisc(object = vn_simulated,
          main = "",
          legend = TRUE,
          show_coordinates = TRUE,
          show_location_ids = TRUE,
          show_positioning_holes = FALSE)

```

When there is a wide range in values, it can be helpful to apply a logarithmic scale in plotting (note that the default is "square root"):

```{r show_disc_scales, fig.dim = c(8, 5), out.width="800px"}

plot_SingleGrainDisc(object = vn_simulated,
          main = "Linear scale",
          legend = TRUE,
          show_coordinates = FALSE,
          show_location_ids = FALSE,
          show_positioning_holes = TRUE,
          str_transform = "lin")


plot_SingleGrainDisc(object = vn_simulated,
          main = "Logarithmic scale",
          legend = TRUE,
          show_coordinates = FALSE,
          show_location_ids = FALSE,
          show_positioning_holes = TRUE,
          str_transform = "log")

```

# Moran scatterplot options
```{r show_scatterplot_options, fig.dim = c(8, 5), out.width="800px"}


vn_simulated <- c(rnorm(75, mean = 10, sd = 5),
                  rnorm(25, mean = 500, sd = 50)  )
vn_simulated <- sample(size = 100, vn_simulated)

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                               n_crosstalk = 0.15)

## Base use
plot_MoranScatterplot(object = vn_simulated,
                      main = "Without crosstalk")


plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
                      main = "With crosstalk")


## Layout options
plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
           pch = "show_location_ids",
           legend = FALSE,
           log = "xy",
           main = "With location ID's; and with log scales"
           )

plot_MoranScatterplot(object = vn_simulated_with_crosstalk,
           pch = "show_n_neighbours",
           legend = FALSE,
           str_y_def = "weighted_sum",
           main = "With number of neighbours, and other y calculation"
           )
```

# Moran's I function options

The function `calc_MoransI()` can return many intermediate calculation numbers:

```{r moransI_intermediate_numbers}


calc_MoransI(object = 1:100,
             return_intermediate_values = TRUE)

```

If the weights (and thus the spatial pattern under investigation) and the
number of observations remain the same, this can be useful to understand what
is happening. For example, if we add crosstalk, we can see that the population
variance in most cases slightly increases (the values are indeed spatially
smoothed, but the average increases) but the spatial autocorrelation strongly
increases:

```{r moransI_intermediate_numbers_extended}

vn_simulated <- sample(x = c(rnorm(n = 30, mean = 2000, sd = 500),
                             rnorm(n = 70, mean = 20, sd = 1)),
                       size = 100)
vn_simulated <- round(vn_simulated)

vn_simulated_with_crosstalk <- apply_Crosstalk(object = vn_simulated,
                                               n_crosstalk = 0.20)
vn_simulated_with_crosstalk <- round(vn_simulated_with_crosstalk)


df_compare <-
  data.frame(Case = c("Without crosstalk", "With crosstalk"),
             MoransI =  c(calc_MoransI(object = vn_simulated),
                          calc_MoransI(object = vn_simulated_with_crosstalk)
                          ),
             PopulationVar =
               c(calc_MoransI(object = vn_simulated,
                 return_intermediate_values = TRUE)$n_population_variance,
                 calc_MoransI(object = vn_simulated_with_crosstalk,
                 return_intermediate_values = TRUE)$n_population_variance),
             SpatialAutoCor =
               c(calc_MoransI(object = vn_simulated,
                 return_intermediate_values = TRUE)$n_average_auto_correlation,
                 calc_MoransI(object = vn_simulated_with_crosstalk,
                 return_intermediate_values = TRUE)$n_average_auto_correlation)
      )

df_compare[,2] <-  round(df_compare[,2],2)

(df_compare)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_Risoe.BINfileData.R
\name{merge_Risoe.BINfileData}
\alias{merge_Risoe.BINfileData}
\title{Merge Risoe.BINfileData objects or Risoe BIN-files}
\usage{
merge_Risoe.BINfileData(
  objects,
  output.file,
  keep.position.number = FALSE,
  position.number.append.gap = 0,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{objects}{\link{character} or \linkS4class{Risoe.BINfileData} (\strong{required}):
Character vector with path and files names with ".bin" or ".binx" extension
(e.g. \code{input.objects = c("path/file1.bin", "path/file2.bin")} or a list of
\linkS4class{Risoe.BINfileData} objects (e.g. \code{input.objects = c(object1, object2)}).}

\item{output.file}{\link{character} (\emph{optional}):
File output path and name. If no value is given, a \linkS4class{Risoe.BINfileData}
object returned instead of a file.}

\item{keep.position.number}{\link{logical} (\emph{with default}):
Allows keeping the original position numbers of the input objects.
Otherwise the position numbers are recalculated.}

\item{position.number.append.gap}{\link{integer} (\emph{with default}):
Set the position number gap between merged BIN-file sets, if the option
\code{keep.position.number = FALSE} is used. See details for further
information.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{currently not used.}
}
\value{
Returns a \linkS4class{Risoe.BINfileData} object or writes to the BIN-file
specified by \code{output.file}.
}
\description{
The function allows merging Risoe BIN/BINX files or \linkS4class{Risoe.BINfileData}
objects.
}
\details{
The function allows merging different measurements to one file or one
object. The record IDs are recalculated for the new object. Other values
are kept for each object. The number of input objects is not limited.

\code{position.number.append.gap} option

If the option \code{keep.position.number = FALSE} is used, the position
numbers of the new data set are recalculated by adding the highest position
number of the previous data set to the each position number of the next data
set. For example: The highest position number is 48, then this number will
be added to all other position numbers of the next data set (e.g. 1 + 48 =
49)

However, there might be cases where an additional addend (summand) is needed
before the next position starts. Example:
\itemize{
\item Position number set (A): \verb{1,3,5,7}
\item Position number set (B): \verb{1,3,5,7}
}

With no additional summand the new position numbers would be:
\verb{1,3,5,7,8,9,10,11}. That might be unwanted. Using the argument
\code{position.number.append.gap = 1} it will become:
\verb{1,3,5,7,9,11,13,15,17}.
}
\note{
The validity of the output objects is not further checked.
}
\section{Function version}{
 0.2.11
}

\examples{

##merge two objects
data(ExampleData.BINfileData, envir = environment())

object1 <- CWOSL.SAR.Data
object2 <- CWOSL.SAR.Data

object.new <- merge_Risoe.BINfileData(c(object1, object2))

} 

\section{How to cite}{
Kreutzer, S., 2026. merge_Risoe.BINfileData(): Merge Risoe.BINfileData objects or Risoe BIN-files. Function version 0.2.11. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Duller, G.A.T., 2007. Analyst (Version 3.24) (manual). Aberystwyth University, Aberystwyth.
}
\seealso{
\linkS4class{Risoe.BINfileData}, \link{read_BIN2R}, \link{write_R2BIN}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{IO}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_ViolinPlot.R
\name{plot_ViolinPlot}
\alias{plot_ViolinPlot}
\title{Create a violin plot}
\usage{
plot_ViolinPlot(
  data,
  boxplot = TRUE,
  rug = TRUE,
  summary = c("n", "median"),
  summary.pos = "sub",
  na.rm = TRUE,
  ...
)
}
\arguments{
\item{data}{\link{numeric} or \linkS4class{RLum.Results} (\strong{required}):
input data for plotting. Alternatively a \link{data.frame} or a \link{matrix} can
be provided, but only the first column will be considered by the
function.}

\item{boxplot}{\link{logical} (\emph{with default}):
enable/disable the boxplot.}

\item{rug}{\link{logical} (\emph{with default}):
enable/disable the rug.}

\item{summary}{\link{character} (\emph{with default}):
add statistic measures of centrality and dispersion to the plot.
Can be one or more of several keywords. See details for available keywords.}

\item{summary.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position keywords (cf. \link{legend}) for the statistical summary.
Alternatively, the keyword \code{"sub"} may be specified to place the summary
below the plot header. However, this latter option in only possible if
\code{mtext} is not used.}

\item{na.rm}{\link{logical} (\emph{with default}):
exclude \code{NA} values from the data set prior to any further operations.}

\item{...}{further arguments and graphical parameters passed to
\link{plot.default}, \link[stats:density]{stats::density} and \link{boxplot}. See details for further
information.}
}
\description{
Draws a kernel density plot in combination with a boxplot in its middle.
The shape of the violin is constructed using a mirrored density curve.
This plot is especially designed for cases where the individual errors are
zero or too small to be visualised. The idea for this plot is based on the
'violin plot' in the \code{ggplot2} package by Hadley Wickham and Winston Chang.
The general idea for the violin plot seems to have been introduced by
Hintze and Nelson (1998).
}
\details{
The function is passing several arguments to the functions \link{plot},
\link[stats:density]{stats::density}, \link[graphics:boxplot]{graphics::boxplot}:

Supported arguments are:
\code{xlim}, \code{main}, \code{xlab}, \code{ylab}, \code{col.violin}, \code{col.boxplot}, \code{mtext}, \code{cex}, \code{mtext}

\strong{Valid summary keywords}

\code{'n'}, \code{'mean'}, \code{'median'}, \code{'sd.abs'}, \code{'sd.rel'}, \code{'se.abs'}, \code{'se.rel'}.
\code{'skewness'}, \code{'kurtosis'}
}
\note{
Although the code for this function was developed independently and just
the idea for the plot was based on the 'ggplot2' plot type 'violin', it
should be mentioned that, beyond this, at least another R package produces
this kind of plot, namely \code{'vioplot'} (see references for details).
}
\section{Function version}{
 0.1.4
}

\examples{

## read example data set
data(ExampleData.DeValues, envir = environment())
ExampleData.DeValues <- convert_Second2Gray(ExampleData.DeValues$BT998,
                                            c(0.0438,0.0019))

## create plot straightforward
plot_ViolinPlot(data = ExampleData.DeValues)

} 

\section{How to cite}{
Kreutzer, S., 2026. plot_ViolinPlot(): Create a violin plot. Function version 0.1.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Daniel Adler (2025). vioplot: violin plot.
R package version 0.5.1 http://CRAN.R-project.org/package=vioplot

Hintze, J.L., Nelson, R.D. (1998). A Box Plot-Density Trace Synergism.
The American Statistician 52, 181-184.

Wickham. H (2009). ggplot2: elegant graphics for data analysis. Springer New York.
}
\seealso{
\link[stats:density]{stats::density}, \link{plot}, \link{boxplot}, \link{rug}, \link{calc_Statistics}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.ScaleGammaDose}
\alias{ExampleData.ScaleGammaDose}
\title{Example data for scale_GammaDose()}
\format{
A \code{\link{data.frame}}. Please see \code{?scale_GammaDose()} for a detailed description
of its structure.
}
\description{
An example data set for the function \code{scale_GammaDose()} containing layer
specific information to scale the gamma dose rate considering variations in
soil radioactivity.
}
\section{Version}{
 0.1
}

\examples{

## Load data
data("ExampleData.ScaleGammaDose", envir = environment())

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_RLum.Data.Spectrum.R
\name{merge_RLum.Data.Spectrum}
\alias{merge_RLum.Data.Spectrum}
\title{Merge function for RLum.Data.Spectrum S4 class objects}
\usage{
merge_RLum.Data.Spectrum(
  object,
  merge.method = c("mean", "median", "sum", "sd", "var", "min", "max", "append", "-",
    "*", "/"),
  method.info = NULL,
  max.temp.diff = 0.1
)
}
\arguments{
\item{object}{\link{list} of \linkS4class{RLum.Data.Spectrum} (\strong{required}):
list of objects to be merged.}

\item{merge.method}{\link{character} (\strong{required}):
method for combining of the objects, e.g. \code{'mean'} (default), \code{'median'},
\code{'sum'}, see details for
further information and allowed methods.  Note: Elements in slot info will
be taken from the first object in the list.}

\item{method.info}{\link{numeric} (\emph{optional}):
allows to specify how info elements of the input objects are combined,
e.g. \code{1} means that just the elements from the first object are kept,
\code{2} keeps only the info elements from the 2 object etc.
If set to \code{NULL}, all elements are combined.}

\item{max.temp.diff}{\link{numeric} (\emph{with default}):
maximum difference in the time/temperature values between the spectra to
be merged: when differences exceed this threshold value, the merging
occurs but a warning is raised.}
}
\value{
Returns an \linkS4class{RLum.Data.Spectrum} object.
}
\description{
This function allows to merge \linkS4class{RLum.Data.Spectrum} objects in
different ways without modifying the original objects.
}
\details{
\strong{Supported merge operations are:}

\code{"mean"} (default)

The mean over the cell values is calculated using the function
\link{rowMeans}.

\code{"median"}

The median over the cell values is calculated using the function
\link[matrixStats:rowMedians]{matrixStats::rowMedians}.

\code{"sum"}

All cell values will be summed up using the function \link{rowSums}.

\code{"sd"}

The standard deviation over the cell values is calculated using the function
\link[matrixStats:rowSds]{matrixStats::rowSds}.

\code{"var"}

The variance over the cell values is calculated using the function
\link[matrixStats:rowVars]{matrixStats::rowVars}.

\code{"min"}

The min values from the cell values is chosen using the function
\link[matrixStats:rowRanges]{matrixStats::rowMins}.

\code{"max"}

The max values from the cell values is chosen using the function
\link[matrixStats:rowRanges]{matrixStats::rowMins}.

\code{"append"} (only for \linkS4class{RLum.Data.Curve})

Appends cell values of all curves to one combined data curve. The channel width
is automatically re-calculated, but requires a constant channel width of the
original data.

\code{"-"}

The cell sums of the last objects are subtracted from the first object.

\code{"*"}

The cell sums of the last objects are multiplied with the first object.

\code{"/"}

Values of the first object are divided by cell sums of the last objects.
}
\note{
The information from the slot \code{recordType} is taken from the first
object in the input list. The slot
'curveType' is filled with the name \code{merged}.
}
\section{S3-generic support}{


This function is fully operational via S3-generics:
\code{+}, \code{-}, \code{/}, \code{*}, \code{merge}
}

\section{Function version}{
 0.1.2
}

\examples{

## load example data
data(ExampleData.XSYG, envir = environment())

## plot single curve
plot_RLum(TL.Spectrum)

## sum two copies of the same curve
merged <- merge_RLum.Data.Spectrum(list(TL.Spectrum, TL.Spectrum),
                                   merge.method = "sum")
plot_RLum(merged)

}
\seealso{
\link{merge_RLum}, \linkS4class{RLum.Data.Spectrum}
}
\author{
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Colombo, M., Kreutzer, S., 2026. merge_RLum.Data.Spectrum(): Merge function for RLum.Data.Spectrum S4 class objects. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{internal}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_CosmicDoseRate.R
\name{calc_CosmicDoseRate}
\alias{calc_CosmicDoseRate}
\title{Calculate the cosmic dose rate}
\usage{
calc_CosmicDoseRate(
  depth,
  density,
  latitude,
  longitude,
  altitude,
  corr.fieldChanges = FALSE,
  est.age = NA,
  half.depth = FALSE,
  error = 10,
  ...
)
}
\arguments{
\item{depth}{\link{numeric} (\strong{required}):
depth of overburden (m). For more than one absorber use \cr
\code{c(depth_1, depth_2, ..., depth_n)}}

\item{density}{\link{numeric} (\strong{required}):
average overburden density (g/cm³). For more than one absorber use \cr
\code{c(density_1, density_2, ..., density_n)}}

\item{latitude}{\link{numeric} (\strong{required}):
latitude (decimal degree), N positive}

\item{longitude}{\link{numeric} (\strong{required}):
longitude (decimal degree), E positive}

\item{altitude}{\link{numeric} (\strong{required}):
altitude (m above sea-level)}

\item{corr.fieldChanges}{\link{logical} (\emph{with default}):
correct for geomagnetic field changes after Prescott & Hutton (1994).
Apply only when justified by the data.}

\item{est.age}{\link{numeric} (\emph{with default}):
estimated age range (ka) for geomagnetic field change correction (0-80 ka allowed)}

\item{half.depth}{\link{logical} (\emph{with default}):
How to overcome with varying overburden thickness. If \code{TRUE} only half the
depth is used for calculation. Apply only when justified, i.e. when a constant
sedimentation rate can safely be assumed.}

\item{error}{\link{numeric} (\emph{with default}):
general error (percentage) to be implemented on corrected cosmic dose rate estimate}

\item{...}{further arguments (\code{verbose} to disable/enable output to the
terminal).}
}
\value{
Returns a terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following element:

\item{summary}{\link{data.frame} summary of all relevant calculation results.}
\item{args}{\link{list} used arguments}
\item{call}{\link{call} the function call}

The output should be accessed using the function \link{get_RLum}.
}
\description{
This function calculates the cosmic dose rate taking into account the soft-
and hard-component of the cosmic ray flux and allows corrections for
geomagnetic latitude, altitude above sea-level and geomagnetic field
changes.

This function calculates the total cosmic dose rate considering both the
soft- and hard-component of the cosmic ray flux.

\strong{Internal calculation steps}

(1)
Calculate total depth of all absorber in hg/cm² (1 hg/cm² = 100 g/cm²)

\deqn{absorber = depth_1*density_1 + depth_2*density_2 + ... + depth_n*density_n}

(2)
If \code{half.depth = TRUE}

\deqn{absorber = absorber/2}

(3)
Calculate cosmic dose rate at sea-level and 55 deg. latitude

a) If absorber is > 167 g/cm² (only hard-component; Allkofer et al.  1975):
apply equation given by Prescott & Hutton (1994) (c.f. Barbouti & Rastin
1983)

\deqn{D0 = C/(((absorber+d)^\alpha+a)*(absober+H)) * \exp(-B*absorber)}

b) If absorber is < 167 g/cm² (soft- and hard-component): derive D0 from
Fig. 1 in Prescott & Hutton (1988).

(4)
Calculate geomagnetic latitude (Prescott & Stephan 1982, Prescott &
Hutton 1994)

\deqn{\lambda = \arcsin(0.203\cos(latitude)\cos(longitude-291)
              + 0.979\sin(latitude))}

(5)
Apply correction for geomagnetic latitude and altitude above sea-level.
Values for F, J and H were read from Fig. 3 shown in Prescott & Stephan
(1982) and fitted with 3-degree polynomials for lambda < 35 degree and a
linear fit for lambda > 35 degree.

\deqn{Dc = D0*(F + J \exp((altitude / 1000) / H))}

(6)
Optional: Apply correction for geomagnetic field changes in the last
0-80 ka (Prescott & Hutton 1994). Correction and altitude factors are given
in Table 1 and Fig. 1 in Prescott & Hutton (1994). Values for altitude
factor were fitted with a 2-degree polynomial. The altitude factor is
operated on the decimal part of the correction factor.

\deqn{Dc' = Dc*correctionFactor}

\strong{Usage of \code{depth} and \code{density}}

(1) If only one value for depth and density is provided, the cosmic dose
rate is calculated for exactly one sample and one absorber as overburden
(i.e. \code{depth*density}).

(2) In some cases it might be useful to calculate the cosmic dose rate for a
sample that is overlain by more than one absorber, e.g. in a profile with
soil layers of different thickness and a distinct difference in density.
This can be calculated by providing a matching number of values for
\code{depth} and \code{density} (e.g. \verb{depth = c(1, 2), density = c(1.7, 2.4)})

(3) Another possibility is to calculate the cosmic dose rate for more than
one sample of the same profile. This is done by providing more than one
values for \code{depth} and only one for \code{density}. For example,
\code{depth = c(1, 2, 3)} and \code{density = 1.7} will calculate the cosmic dose rate
for three samples in 1, 2 and 3 m depth in a sediment of density 1.7 g/cm³.
}
\note{
Despite its universal use, the equation to calculate the cosmic dose
rate provided by Prescott & Hutton (1994) is falsely stated to be valid from
the surface to 10^4 hg/cm² of standard rock. The original expression by
Barbouti & Rastin (1983) only considers the muon flux (i.e. hard-component)
and is, by their own definition, only valid for depths between 10-10^4
hg/cm².

Thus, for near-surface samples (i.e. for depths < 167 g/cm²) the equation
of Prescott & Hutton (1994) underestimates the total cosmic dose rate, as it
neglects the influence of the soft-component of the cosmic ray flux. For
samples at zero depth and at sea-level the underestimation can be as large
as ~0.1 Gy/ka. In a previous article, Prescott & Hutton (1988) give another
approximation of Barbouti & Rastin's equation in the form of

\deqn{D = 0.21 * \exp(-0.070 * absorber + 0.0005 * absorber^2)}

which is valid for depths between 150-5000 g/cm². For shallower depths (<
150 g/cm²) they provided a graph (Fig. 1) from which the dose rate can be
read.

As a result, this function employs the equation of Prescott & Hutton (1994)
only for depths > 167 g/cm², i.e. only for the hard-component of the cosmic
ray flux. Cosmic dose rate values for depths < 167 g/cm² were obtained from
the "AGE" program (Gruen 2009) and fitted with a 6-degree polynomial curve
(and hence reproduces the graph shown in Prescott & Hutton 1988). However,
these values assume an average overburden density of 2 g/cm³.

It is currently not possible to obtain more precise cosmic dose rate values
for near-surface samples as there is no equation known to the author of this
function at the time of writing.
}
\section{Function version}{
 0.5.3
}

\examples{

##(1) calculate cosmic dose rate (one absorber)
calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                    latitude = 38.06451, longitude = 1.49646,
                    altitude = 364, error = 10)

##(2a) calculate cosmic dose rate (two absorber)
calc_CosmicDoseRate(depth = c(5.0, 2.78), density = c(2.65, 1.7),
                    latitude = 38.06451, longitude = 1.49646,
                    altitude = 364, error = 10)

##(2b) calculate cosmic dose rate (two absorber) and
##correct for geomagnetic field changes
calc_CosmicDoseRate(depth = c(5.0, 2.78), density = c(2.65, 1.7),
                    latitude = 12.04332, longitude = 4.43243,
                    altitude = 364, corr.fieldChanges = TRUE,
                    est.age = 67, error = 15)


##(3) calculate cosmic dose rate and export results to .csv file
#calculate cosmic dose rate and save to variable
results<- calc_CosmicDoseRate(depth = 2.78, density = 1.7,
                              latitude = 38.06451, longitude = 1.49646,
                              altitude = 364, error = 10)

# the results can be accessed by
get_RLum(results, "summary")

#export results to .csv file - uncomment for usage
#write.csv(results, file = "c:/users/public/results.csv")

##(4) calculate cosmic dose rate for 6 samples from the same profile
##    and save to .csv file
#calculate cosmic dose rate and save to variable
results<- calc_CosmicDoseRate(depth = c(0.1, 0.5 , 2.1, 2.7, 4.2, 6.3),
                              density = 1.7, latitude = 38.06451,
                              longitude = 1.49646, altitude = 364,
                              error = 10)

#export results to .csv file - uncomment for usage
#write.csv(results, file = "c:/users/public/results_profile.csv")

} 

\section{How to cite}{
Burow, C., 2026. calc_CosmicDoseRate(): Calculate the cosmic dose rate. Function version 0.5.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Allkofer, O.C., Carstensen, K., Dau, W.D., Jokisch, H., 1975.
Letter to the editor. The absolute cosmic ray flux at sea level. Journal of
Physics G: Nuclear and Particle Physics 1, L51-L52.

Barbouti, A.I., Rastin, B.C., 1983. A study of the absolute intensity of muons at sea level
and under various thicknesses of absorber. Journal of Physics G: Nuclear and
Particle Physics 9, 1577-1595.

Crookes, J.N., Rastin, B.C., 1972. An
investigation of the absolute intensity of muons at sea-level. Nuclear
Physics B 39, 493-508.

Gruen, R., 2009. The "AGE" program for the
calculation of luminescence age estimates. Ancient TL 27, 45-46.

Prescott, J.R., Hutton, J.T., 1988. Cosmic ray and gamma ray dosimetry for
TL and ESR. Nuclear Tracks and Radiation Measurements 14, 223-227.

Prescott, J.R., Hutton, J.T., 1994. Cosmic ray contributions to dose rates
for luminescence and ESR dating: large depths and long-term time variations.
Radiation Measurements 23, 497-500.

Prescott, J.R., Stephan, L.G., 1982. The contribution of cosmic radiation to the environmental dose for
thermoluminescence dating. Latitude, altitude and depth dependences. PACT 6, 17-25.
}
\seealso{
\link{BaseDataSet.CosmicDoseRate}
}
\author{
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_PSL2CSV.R
\name{convert_PSL2CSV}
\alias{convert_PSL2CSV}
\title{Export PSL-file(s) to CSV-files}
\usage{
convert_PSL2CSV(file, extract_raw_data = FALSE, single_table = FALSE, ...)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of the PSL-file to be converted to CSV-files}

\item{extract_raw_data}{\link{logical} (\emph{with default}): enable/disable raw data
extraction. The PSL files imported into R contain an element \verb{$raw_data}, which
provides a few more information (e.g., count errors), sometimes it makes
sense to use this data of the more compact standard values created by \link{read_PSL2R}}

\item{single_table}{\link{logical} (\emph{with default}): enable/disable the creation
of single table with n rows and n columns, instead of separate \link{data.frame}
objects. Each curve will be represented by two columns for time and counts}

\item{...}{further arguments that will be passed to the function
\link{read_PSL2R} and \link{write_RLum2CSV}}
}
\value{
The function returns either a CSV-file (or many of them) or, when
\code{export = FALSE}, a list of \link{data.frame} and \link{matrix} objects.
}
\description{
This function is a wrapper function around the functions \link{read_PSL2R} and
\link{write_RLum2CSV} and it imports an PSL-file (SUERC portable OSL reader file format)
and directly exports its content to CSV-files.
If nothing is set for the argument \code{path} (\link{write_RLum2CSV}) the input folder will
become the output folder.
}
\section{Function version}{
 0.1.2
}

\examples{

## export into single data.frame
file <- system.file("extdata/DorNie_0016.psl", package="Luminescence")
convert_PSL2CSV(file, export = FALSE, single_table = TRUE)

\dontrun{
##select your BIN-file
file <- file.choose()

##convert
convert_PSL2CSV(file)
}

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data}, \linkS4class{RLum.Results},
\link[utils:write.table]{utils::write.table}, \link{write_RLum2CSV}, \link{read_PSL2R}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. convert_PSL2CSV(): Export PSL-file(s) to CSV-files. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_ROI.R
\name{extract_ROI}
\alias{extract_ROI}
\title{Extract Pixel Values through Circular Region-of-Interests (ROI) from an Image}
\usage{
extract_ROI(
  object,
  roi,
  roi_summary = c("mean", "median", "sd", "sum"),
  plot = FALSE
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Image}, \link{array} or \link{matrix} (\strong{required}): input image data}

\item{roi}{\link{matrix} (\strong{required}): matrix with three columns containing the centre coordinates
of the ROI (first two columns) and the diameter of the circular ROI. All numbers must by of type \link{integer}
and will forcefully coerced into such numbers using \code{as.integer()} regardless.}

\item{roi_summary}{(\emph{with default}): defines what is returned in the
\code{roi_summary} element; it can be \code{"mean"} (default), \code{"median"}, \code{"sd"} or
\code{"sum"}.
Pixel values are conveniently summarised using the above defined keyword.}

\item{plot}{\link{logical} (\emph{optional}): enable/disable the plot output. Only
the first image frame is shown.}
}
\value{
\linkS4class{RLum.Results} object with the following elements:
\code{..$roi_signals}: a named \link{list} with all ROI values and their coordinates
\code{..$roi_summary}: a \link{matrix} where rows are frames from the image, and columns are different ROI
The element has two attributes: \code{summary} (the method used to summarise pixels) and \code{area} (the pixel area)
\code{..$roi_coord}: a \link{matrix} that can be passed to \link{plot_ROI}

If \code{plot = TRUE} a control plot is returned.
}
\description{
Light-weighted function to extract pixel values from pre-defined regions-of-interest (ROI) from
\linkS4class{RLum.Data.Image}, \link{array} or \link{matrix} objects and provide simple image processing
capacity. The function is limited to circular ROIs.
}
\details{
The function uses a cheap approach to decide whether a pixel lies within
a circle or not. It assumes that pixel coordinates are integer values and
that a pixel centring within the circle is satisfied by:

\deqn{x^2 + y^2 <= (d/2)^2}

where \eqn{x} and \eqn{y} are integer pixel coordinates and \eqn{d} is the integer
diameter of the circle in pixel.
}
\section{Function version}{
 0.1.0
}

\examples{

m <- matrix(runif(100,0,255), ncol = 10, nrow = 10)
roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
extract_ROI(object = m, roi = roi, plot = TRUE)

}
\seealso{
\linkS4class{RLum.Data.Image}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. extract_ROI(): Extract Pixel Values through Circular Region-of-Interests (ROI) from an Image. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_DetPlot.R
\name{plot_DetPlot}
\alias{plot_DetPlot}
\title{Create De(t) plot}
\usage{
plot_DetPlot(
  object,
  signal_integral,
  background_integral,
  method = "shift",
  signal_integral.seq = NULL,
  integral_input = c("channel", "measurement"),
  analyse_function = c("analyse_SAR.CWOSL", "analyse_pIRIRSequence"),
  analyse_function.control = list(),
  n.channels = NULL,
  show_ShineDownCurve = TRUE,
  respect_RC.Status = FALSE,
  multicore = TRUE,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}): input object containing data for analysis
Can be provided as a \link{list} of such objects.}

\item{signal_integral}{\link{integer} (\strong{required}):
vector of channels for the signal integral.}

\item{background_integral}{\link{integer} (\strong{required}):
vector of channels for the background integral.}

\item{method}{\link{character} (\emph{with default}):
method applied for constructing the De(t) plot.
\itemize{
\item \code{shift} (\emph{the default}): the chosen signal integral is shifted the shine down curve,
\item \code{expansion}: the chosen signal integral is expanded each time by its length
}}

\item{signal_integral.seq}{\link{numeric} (\emph{optional}):
argument to provide an own signal integral sequence for constructing the De(t) plot}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{analyse_function}{\link{character} (\emph{with default}):
name of the analyse function to be called. Supported functions are:
\link{analyse_SAR.CWOSL}, \link{analyse_pIRIRSequence}}

\item{analyse_function.control}{\link{list} (\emph{optional}):
selected arguments to be passed to the supported analyse functions
(\link{analyse_SAR.CWOSL}, \link{analyse_pIRIRSequence}). The arguments must be provided
as named \link{list}, e.g., \verb{list(dose.points = c(0,10,20,30,0,10)} will set the
regeneration dose points.}

\item{n.channels}{\link{integer} (\emph{optional}):
number of channels used for the De(t) plot. If nothing is provided all
De-values are calculated and plotted until the start of the background
integral.}

\item{show_ShineDownCurve}{\link{logical} (\emph{with default}):
enable/disable shine down curve in the plot output.}

\item{respect_RC.Status}{\link{logical} (\emph{with default}):
remove De values with 'FAILED' RC.Status from the plot (cf. \link{analyse_SAR.CWOSL}
and \link{analyse_pIRIRSequence}).}

\item{multicore}{\link{logical} (\emph{with default}) : enable/disable multi core
calculation if \code{object} is a \link{list} of \linkS4class{RLum.Analysis} objects. Can be an
\link{integer} specifying the number of cores}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.
Disabling the plot is useful in cases where the output need to be processed
differently.}

\item{...}{further arguments and graphical parameters passed to
\link{plot.default}, \link{analyse_SAR.CWOSL} and \link{analyse_pIRIRSequence} (see details for further information).
Plot control parameters are: \code{ylim}, \code{xlim}, \code{ylab}, \code{xlab}, \code{main}, \code{pch}, \code{mtext}, \code{cex}, \code{legend},
\code{legend.text}, \code{legend.pos}}
}
\value{
A plot and an \linkS4class{RLum.Results} object with the produced \eqn{D_e} values

\verb{@data}:

\tabular{lll}{
\strong{Object} \tab \strong{Type} \tab \strong{Description}\cr
\code{De.values} \tab \code{data.frame} \tab table with De values \cr
\code{signal_integral.seq} \tab \code{numeric} \tab integral sequence used for the calculation
}

\verb{@info}:

\tabular{lll}{
\strong{Object} \tab \strong{Type} \tab \strong{Description}\cr
call \tab \code{call} \tab the original function call
}
}
\description{
Plots the equivalent dose (\eqn{D_e}) in dependency of the chosen signal integral
(cf. Bailey et al., 2003). The function is simply passing several arguments
to the function \link{plot} and the used analysis functions and runs it in a loop.
Example: \code{legend.pos} for legend position, \code{legend} for legend text.
}
\details{
\strong{method}

The original method presented by Bailey et al., 2003 shifted the signal integrals and slightly
extended them accounting for changes in the counting statistics. Example: \code{c(1:3, 3:5, 5:7)}.
However, here also another method is provided allowing to expand the signal integral by
consecutively expanding the integral by its chosen length. Example: \code{c(1:3, 1:5, 1:7)}

Note that in both cases the integral limits are overlap. The finally applied limits are part
of the function output.

\strong{analyse_function.control}

The argument \code{analyse_function.control} currently supports the following arguments:
\code{sequence.structure}, \code{dose.points}, \code{mtext.outer}, \code{fit.method},
\code{fit.force_through_origin}, \code{plot}, \code{plot_singlePanels}
}
\note{
The entire analysis is based on the used analysis functions, namely
\link{analyse_SAR.CWOSL} and \link{analyse_pIRIRSequence}. However, the integrity
checks of this function are not that thoughtful as in these functions itself.
It means, that every sequence should be checked carefully before running long
calculations using several hundreds of channels.
}
\section{Function version}{
 0.1.10
}

\examples{

\dontrun{
##load data
##ExampleData.BINfileData contains two BINfileData objects
##CWOSL.SAR.Data and TL.SAR.Data
data(ExampleData.BINfileData, envir = environment())

##transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

plot_DetPlot(
  object,
  signal_integral = 1:3,
  background_integral = 900:1000,
  n.channels = 5)
}

} 

\section{How to cite}{
Kreutzer, S., 2026. plot_DetPlot(): Create De(t) plot. Function version 0.1.10. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bailey, R.M., Singarayer, J.S., Ward, S., Stokes, S., 2003. Identification of partial resetting
using De as a function of illumination time. Radiation Measurements 37, 511-518.
doi:10.1016/S1350-4487(03)00063-5
}
\seealso{
\link{plot}, \link{analyse_SAR.CWOSL}, \link{analyse_pIRIRSequence}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_portableOSL.R
\name{analyse_portableOSL}
\alias{analyse_portableOSL}
\title{Analyse portable CW-OSL measurements}
\usage{
analyse_portableOSL(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  invert = FALSE,
  normalise = FALSE,
  mode = c("profile", "surface"),
  coord = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}):
object produced by \link{read_PSL2R}. The input can be a \link{list} of such objects,
in which case each input is treated as a separate sample and the results
are merged.}

\item{signal_integral}{\link{numeric} (\strong{required}):
vector specifying the range of channels used to calculate the OSL/IRSL
signal. It can be provided as a vector of length 2 such as \code{c(1, 5)}, or
as a sequence such as \code{1:5}, in which case the lowest and highest values
define the range.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{invert}{\link{logical} (\emph{with default}):
whether the data should be plotted in reverse order (\code{FALSE} by default).}

\item{normalise}{\link{logical} (\emph{with default}):
whether the OSL/IRSL signals should be normalised to the \emph{mean} of all
corresponding data curves.}

\item{mode}{\link{character} (\emph{with default}):
analysis mode, one of \code{"profile"} (the default) or \code{"surface"} for surface
interpolation.}

\item{coord}{\link{list} \link{matrix} (\emph{optional}): a list or a 2-column matrix
with the \emph{x} and \emph{y} coordinates for the sampling positions in meters (m),
of the same length as the number of samples measured. For example, the
coordinates for one sample could be \verb{coord = list(samp1 = c(0.1, 0.2)}.
If the \emph{x} coordinates were not measured, \emph{x} should be set to 0.
Note that, in such case, a surface plot cannot be produced.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{...}{other parameters to be passed to modify the plot output.
Supported are \code{run} to provide the run name (if the input is a \code{list}, this
is set automatically). Further plot parameters accepted are \code{main},
\code{col}, \code{xlim} (a named \link{list} for profile mode), \code{ylim}, \code{ylab}, \code{xlab}.
Additional parameters for \code{mode = "profile"} are  \code{type}, \code{pch}, \code{grid}
(\code{TRUE}/\code{FALSE}), \code{bg_img} (a raster object for the background image,
usually a profile photo), \code{bg_img_positions} (a vector with the four corner
positions, see \link[graphics:rasterImage]{graphics::rasterImage}), \code{zlab} (here x-axis labelling).
Additional parameters for \code{mode = "surface"} are \code{surface_value}
(\link{character} with names of the surfaces to plot), \code{col_ramp}, \code{legend}
(\code{TRUE}/\code{FALSE}), \code{contour} (\code{TRUE}/\code{FALSE}), \code{contour_nlevels},
\code{contour_col}, \code{nx} and \code{ny} (size of the interpolation grid),
\code{labcex} (scaling of the contour labels), \code{zlim}.}
}
\value{
Returns an S4 \linkS4class{RLum.Results} object with the following elements:

\verb{$data}\cr
\code{.. $summary}: \link{data.frame} with the results\cr
\code{.. $data}: \link{list} with the \linkS4class{RLum.Analysis} objects\cr
\code{.. $args}: \link{list} the input arguments
}
\description{
The function analyses CW-OSL curve data produced by a SUERC portable OSL reader and
produces a combined plot of OSL/IRSL signal intensities, OSL/IRSL depletion ratios
and the IRSL/OSL ratio.
}
\details{
This function only works with \linkS4class{RLum.Analysis} objects produced by \link{read_PSL2R}.
It further assumes (or rather requires) an equal amount of OSL and IRSL curves that
are pairwise combined for calculating the IRSL/OSL ratio.
For calculating the depletion ratios, the cumulative signal of the last \emph{n}
channels (same number of channels as specified by \code{signal_integral}) is
divided by cumulative signal of the first \emph{n} channels (\code{signal_integral}).

\strong{Note:} The function assumes the following sequence pattern:
\verb{DARK COUNT}, \code{IRSL}, \verb{DARK COUNT}, \code{BSL}, \verb{DARK COUNT}. Therefore, the
total number of curves in the input object must be a multiple of 5, and
there must be 3 \code{DARK_COUNT} records for each IRSL/BSL pair. If a different
sequence was used, the function will produce an error.

\strong{Signal processing}
The function processes the signals as follows: \code{BSL} and \code{IRSL} signals are
extracted using the chosen signal integral, dark counts are taken in full.

\strong{Working with coordinates}
Usually samples are taken from a profile with a certain stratigraphy. In
the past, the function calculated an index. Currently, you have two ways
of passing on xy-coordinates to the function:
\itemize{
\item (1) Add coordinates to the sample name during measurement. The form is rather
strict and has to follow the scheme \verb{_x:<number>|y:<number>}. Example:
\code{sample_x:0.2|y:0.4}.
\item (2) Alternatively, you can provide a \link{list} or \link{matrix} with the \emph{(x, y)}
coordinates of each sample in meters (m) using the \code{coord} argument:
Example: \code{coord = list(c(0.2, 1), c(0.3,1.2))}
}

If in your profile the x-coordinates were not measured, \emph{x} should be set
to 0. Note that, in such case, a surface plot cannot be produced.
}
\section{Function version}{
 0.1.5
}

\examples{

## example profile plot
# (1) load example data set
data("ExampleData.portableOSL", envir = environment())

# (2) merge and plot all RLum.Analysis objects
merged <- merge_RLum(ExampleData.portableOSL)
plot_RLum(
 object = merged,
 combine = TRUE,
 records_max = 5,
 legend.pos = "outside")
merged

# (3) analyse and plot
results <- analyse_portableOSL(
  merged,
  signal_integral = 1:5,
  invert = FALSE,
  normalise = TRUE)
get_RLum(results)

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve}, \link{read_PSL2R}
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Burow, C., Kreutzer, S., Colombo, M., 2026. analyse_portableOSL(): Analyse portable CW-OSL measurements. Function version 0.1.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply_EfficiencyCorrection.R
\name{apply_EfficiencyCorrection}
\alias{apply_EfficiencyCorrection}
\title{Apply spectral efficiency correction to RLum.Data.Spectrum objects}
\usage{
apply_EfficiencyCorrection(object, spectral.efficiency)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Spectrum} or \linkS4class{RLum.Analysis} (\strong{required}):
object of class \code{RLum.Data.Spectrum}, \code{RLum.Analysis}or a \link{list} of such
objects. Other objects in the list are skipped.}

\item{spectral.efficiency}{\link{data.frame} (\strong{required}):
data frame with 2 columns containing wavelengths and relative spectral
response values (values between 0 and 1). The provided data will be used to
correct all spectra if \code{object} is a \link{list}.}
}
\value{
Returns same object as provided as input
}
\description{
The function applies spectral efficiency correction to
\linkS4class{RLum.Data.Spectrum} objects.
}
\details{
The efficiency correction is based on a spectral response dataset provided
by the user. Usually the data set for the quantum efficiency is of lower
resolution and values are interpolated for the required spectral resolution using
function \link[stats:approxfun]{stats::approx}.

If the energy calibration differs for both data set \code{NA} values are produces that
will be removed from the matrix.
}
\note{
Please note that the spectral efficiency data from the camera alone may not
sufficiently correct for spectral efficiency of the entire optical system
(e.g., spectrometer, camera ...).
}
\section{Function version}{
 0.2.1
}

\examples{

##(1) - use with your own data (uncomment for usage)
## spectral.efficiency <- read.csv("your data")
##
## your.spectrum <- apply_EfficiencyCorrection(your.spectrum, )

}
\seealso{
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Johannes Friedrich, University of Bayreuth (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Friedrich, J., 2026. apply_EfficiencyCorrection(): Apply spectral efficiency correction to RLum.Data.Spectrum objects. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Analyse_SAR.OSLdata.R
\name{Analyse_SAR.OSLdata}
\alias{Analyse_SAR.OSLdata}
\title{Analyse SAR CW-OSL measurements.}
\usage{
Analyse_SAR.OSLdata(
  input.data,
  signal.integral,
  background.integral,
  position,
  run,
  set,
  dtype,
  keep.SEL = FALSE,
  info.measurement = "unkown measurement",
  output.plot = FALSE,
  output.plot.single = FALSE,
  cex.global = 1,
  ...
)
}
\arguments{
\item{input.data}{\linkS4class{Risoe.BINfileData} (\strong{required}):
input data from a Risø BIN file, produced by the function \link{read_BIN2R}.}

\item{signal.integral}{\link{vector} (\strong{required}):
channels used for the signal integral, e.g. \code{signal.integral=c(1:2)}}

\item{background.integral}{\link{vector} (\strong{required}):
channels used for the background integral, e.g. \code{background.integral=c(85:100)}}

\item{position}{\link{vector} (\emph{optional}):
reader positions that want to be analysed (e.g. \code{position=c(1:48)}.
Empty positions are automatically omitted. If no value is given all
positions are analysed by default.}

\item{run}{\link{vector} (\emph{optional}):
range of runs used for the analysis. If no value is given the range of the
runs in the sequence is deduced from the \code{Risoe.BINfileData} object.}

\item{set}{\link{vector} (\emph{optional}):
range of sets used for the analysis. If no value is given the range of the
sets in the sequence is deduced from the \code{Risoe.BINfileData} object.}

\item{dtype}{\link{character} (\emph{optional}):
allows to further limit the curves by their data type (\code{DTYPE}),
e.g., \code{dtype = c("Natural", "Dose")} limits the curves to this two data types.
By default all values are allowed.
See \linkS4class{Risoe.BINfileData} for allowed data types.}

\item{keep.SEL}{\link{logical} (default):
option allowing to use the \code{SEL} element of the \linkS4class{Risoe.BINfileData} manually.
\strong{NOTE:} In this case any limitation provided by \code{run}, \code{set} and \code{dtype}
are ignored!}

\item{info.measurement}{\link{character} (\emph{with default}):
option to provide information about the measurement on the plot
output (e.g. name of the BIN or BINX file).}

\item{output.plot}{\link{logical} (\emph{with default}):
plot output (\code{TRUE/FALSE})}

\item{output.plot.single}{\link{logical} (\emph{with default}):
single plot output (\code{TRUE/FALSE}) to allow for plotting the results in
single plot windows. Requires \code{output.plot = TRUE}.}

\item{cex.global}{\link{numeric} (\emph{with default}):
global scaling factor.}

\item{...}{further arguments that will be passed to the function
\link{calc_OSLLxTxRatio} (supported: \code{background.count.distribution}, \code{sigmab},
\code{sig0}; e.g., for instrumental error) and can be used to adjust the plot.
Supported" \code{mtext}, \code{log}}
}
\value{
A plot (\emph{optional}) and \link{list} is returned containing the
following elements:

\item{LnLxTnTx}{\link{data.frame} of all calculated Lx/Tx values including signal, background counts and the dose points.}
\item{RejectionCriteria}{\link{data.frame} with values that might by used as rejection criteria. NA is produced if no R0 dose point exists.}
\item{SARParameters}{\link{data.frame} of additional measurement parameters obtained from the BIN file, e.g. preheat or read temperature
(not valid for all types of measurements).}
}
\description{
The function analyses SAR CW-OSL curve data and provides a summary of the
measured data for every position. The output of the function is optimised
for SAR OSL measurements on quartz.
}
\details{
The function works only for standard SAR protocol measurements introduced by
Murray and Wintle (2000) with CW-OSL curves. For the calculation of the
Lx/Tx value the function \link{calc_OSLLxTxRatio} is used.

\strong{Provided rejection criteria}

\verb{[recyling ratio]}: calculated for every repeated regeneration dose point.

\verb{[recuperation]}: recuperation rate calculated by comparing the \code{Lx/Tx} values of the zero
regeneration point with the \code{Ln/Tn} value (the \code{Lx/Tx} ratio of the natural
signal). For methodological background see Aitken and Smith (1988)

\verb{[IRSL/BOSL]}: the integrated counts (\code{signal.integral}) of an
IRSL curve are compared to the integrated counts of the first regenerated
dose point. It is assumed that IRSL curves got the same dose as the first
regenerated dose point. \strong{Note:} This is not the IR depletion ratio
described by Duller (2003).
}
\note{
Rejection criteria are calculated but not considered during the
analysis to discard values.

\strong{The analysis of IRSL data is not directly supported}. You may want to
consider using the functions \link{analyse_SAR.CWOSL} or
\link{analyse_pIRIRSequence} instead.

\strong{The development of this function will not be continued. We recommend to use the function \link{analyse_SAR.CWOSL} or instead.}
}
\section{Function version}{
 0.2.17
}

\examples{
##load data
data(ExampleData.BINfileData, envir = environment())

##analyse data
output <- Analyse_SAR.OSLdata(input.data = CWOSL.SAR.Data,
                              signal.integral = c(1:5),
                              background.integral = c(900:1000),
                              position = c(1:1),
                              output.plot = TRUE)

##combine results relevant for further analysis
output.SAR <- data.frame(Dose = output$LnLxTnTx[[1]]$Dose,
                         LxTx = output$LnLxTnTx[[1]]$LxTx,
                         LxTx.Error = output$LnLxTnTx[[1]]$LxTx.Error)
output.SAR

} 

\section{How to cite}{
Kreutzer, S., Fuchs, M.C., 2023. Analyse_SAR.OSLdata(): Analyse SAR CW-OSL measurements.. Function version 0.2.17. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., 2023. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 0.9.21. https://CRAN.R-project.org/package=Luminescence
}

\references{
Aitken, M.J. and Smith, B.W., 1988. Optical dating: recuperation
after bleaching. Quaternary Science Reviews 7, 387-393.

Duller, G., 2003. Distinguishing quartz and feldspar in single grain
luminescence measurements. Radiation Measurements, 37 (2), 161-165.

Murray, A.S. and Wintle, A.G., 2000. Luminescence dating of quartz using an
improved single-aliquot regenerative-dose protocol. Radiation Measurements
32, 57-73.
}
\seealso{
\link{calc_OSLLxTxRatio}, \linkS4class{Risoe.BINfileData}, \link{read_BIN2R}, \link{plot_GrowthCurve}
}
\author{
Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)\cr
Margret C. Fuchs, HZDR, Freiberg (Germany)
, RLum Developer Team}
\keyword{datagen}
\keyword{dplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_XSYG2R.R
\name{read_XSYG2R}
\alias{read_XSYG2R}
\title{Import XSYG files into R}
\usage{
read_XSYG2R(
  file,
  recalculate.TL.curves = TRUE,
  n_records = NULL,
  fastForward = FALSE,
  import = TRUE,
  pattern = "\\\\.xsyg",
  auto_linearity_correction = FALSE,
  verbose = TRUE,
  txtProgressBar = TRUE
)
}
\arguments{
\item{file}{\link{character} or \link{list} (\strong{required}):
name of one or multiple XSYG files (URLs are supported); it can be the path
to a directory, in which case the function tries to detect and import all
XSYG files found recursively from the given directory.}

\item{recalculate.TL.curves}{\link{logical} (\emph{with default}):
if set to \code{TRUE}, TL curves are returned as temperature against count values
(see details for more information) Note: The option overwrites the time vs.
count TL curve. Select \code{FALSE} to import the raw data delivered by the
lexsyg. Works for TL curves and spectra.}

\item{n_records}{\link{numeric} (\emph{with default}):
number of records to be imported; by default the function attempts to import
all records.}

\item{fastForward}{\link{logical} (\emph{with default}):
if \code{TRUE} for a more efficient data processing only a list of
\linkS4class{RLum.Analysis} objects is returned.}

\item{import}{\link{logical} (\emph{with default}):
if set to \code{FALSE}, only the XSYG file structure is shown.}

\item{pattern}{\link{character} (\emph{with default}):
regular expression pattern passed to \link{list.files} to construct a list of
files to read (used only when a path is provided).}

\item{auto_linearity_correction}{\link{logical} (\emph{with default}): enable/disable
automatic count linearity correction. Because the information on the detectors
are not consistent and are not consistently stored, this option should be used
with caution.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar during import. Ignored if \code{verbose = FALSE}.}
}
\value{
\strong{Using the option \code{import = FALSE}}

A list consisting of two elements is shown:
\itemize{
\item \link{data.frame} with information on file.
\item \link{data.frame} with information on the sequences stored in the XSYG file.
}

\strong{Using the option \code{import = TRUE} (default)}

A list is provided, the list elements
contain: \item{Sequence.Header}{\link{data.frame} with information on the
sequence.} \item{Sequence.Object}{\linkS4class{RLum.Analysis}
containing the curves.}
}
\description{
Imports XSYG-files produced by a Freiberg Instruments lexsyg reader into R.
}
\details{
\strong{How does the import function work?}

The function uses the \code{'XML'} package to parse the file structure. Each
sequence is subsequently translated into an \linkS4class{RLum.Analysis} object.

\strong{General structure XSYG format}

\if{html}{\out{<div class="sourceCode">}}\preformatted{<?xml?>
<Sample>
  <Sequence>
    <Record>
      <Curve name="first curve" />
      <Curve name="curve with data">x0 , y0 ; x1 , y1 ; x2 , y2 ; x3 , y3</Curve>
    </Record>
  </Sequence>
</Sample>
}\if{html}{\out{</div>}}

So far, each
XSYG file can only contain one \verb{<Sample></Sample>}, but multiple
sequences.

Each record may comprise several curves.

\strong{TL curve recalculation}

On the FI lexsyg device TL curves are recorded as time against count values.
Temperature values are monitored on the heating plate and stored in a
separate curve (time vs. temperature). If the option
\code{recalculate.TL.curves = TRUE} is chosen, the time values for each TL
curve are replaced by temperature values.

Practically, this means combining two matrices (Time vs. Counts and Time vs.
Temperature) with different row numbers by their time values. Three cases
are considered:
\enumerate{
\item HE: Heating element
\item PMT: Photomultiplier tube
\item Interpolation is done using the function \link{approx}
}

CASE (1): \code{nrow(matrix(PMT))} > \code{nrow(matrix(HE))}

Missing temperature values from the heating element are calculated using
time values from the PMT measurement.

CASE (2): \code{nrow(matrix(PMT))} < \code{nrow(matrix(HE))}

Missing count values from the PMT are calculated using time values from the
heating element measurement.

CASE (3): \code{nrow(matrix(PMT))} == \code{nrow(matrix(HE))}

A new matrix is produced using temperature values from the heating element
and count values from the PMT.

\strong{Note:}
Please note that due to the recalculation of the temperature
values based on values delivered by the heating element, it may happen that
multiple count values exists for each temperature value and temperature
values may also decrease during heating, not only increase.

\strong{Count linearity correction}

If the argument \code{auto_linearity_correction = TRUE}, the function offers a,
theoretically, automated linearity correction of the PMT signal using
the internal function \link{correct_PMTLinearity}. The critical parameter
is the count-pair resolution, which is provided to the function
automatically depending on the detector. Currently, the following
settings are used:

\tabular{ll}{
DETECTOR \tab COUNT-PAIR-RESOLUTION\cr
UVVIS \tab 18 ns\cr
NIR50 \tab 70 ns\cr
NIR40 \tab 70 ns \cr
ETPMT \tab 25 ns
}

Unfortunately, not all XSYG files provide correct information on
the used detector, depending on software version. Hence, in case
of doubt, please verify the settings and conduct a manual correction
if required.

\strong{Advanced file import}

To allow for a more efficient usage of the function, instead of single path
to a file just a directory can be passed as input. In this particular case
the function tries to extract all XSYG-files found in the directory and import
them all. Using this option internally the function constructs as list of
the XSYG-files found in the directory. Please note no recursive detection
is supported as this may lead to endless loops.
}
\note{
This function is a beta version as the XSYG file format is not yet
fully specified. Thus, further file operations (merge, export, write) should
be done using the functions provided with the package \code{'XML'}.

\strong{So far, no image data import is provided!} \cr
Corresponding values in the XSXG file are skipped.
}
\section{Function version}{
 0.8.2
}

\examples{

##(1) import XSYG file to R (uncomment for usage)

#FILE <- file.choose()
#temp <- read_XSYG2R(FILE)

##(2) additional examples for pure XML import using the package XML
##    (uncomment for usage)

  ##import entire XML file
  #FILE <- file.choose()
  #temp <- XML::xmlRoot(XML::xmlTreeParse(FILE))

  ##search for specific subnodes with curves containing 'OSL'
  #getNodeSet(temp, "//Sample/Sequence/Record[@recordType = 'OSL']/Curve")

##(2) How to extract single curves ... after import
data(ExampleData.XSYG, envir = environment())

##grep one OSL curves and plot the first curve
OSLcurve <- get_RLum(OSL.SARMeasurement$Sequence.Object, recordType="OSL")[[1]]

##(3) How to see the structure of an object?
structure_RLum(OSL.SARMeasurement$Sequence.Object)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. read_XSYG2R(): Import XSYG files into R. Function version 0.8.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Grehl, S., Kreutzer, S., Hoehne, M., 2013. Documentation of the
XSYG file format. Unpublished Technical Note. Freiberg, Germany

\strong{Further reading}

XML: \url{https://en.wikipedia.org/wiki/XML}
}
\seealso{
\code{'XML'}, \linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve},
\link{approx}, \link{correct_PMTLinearity}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_baSAR.R
\name{analyse_baSAR}
\alias{analyse_baSAR}
\title{Bayesian models (baSAR) applied on luminescence data}
\usage{
analyse_baSAR(
  object,
  CSV_file = NULL,
  aliquot_range = NULL,
  source_doserate = NULL,
  signal_integral = NULL,
  background_integral = NULL,
  signal_integral_Tx = NULL,
  background_integral_Tx = NULL,
  irradiation_times = NULL,
  sigmab = 0,
  sig0 = 0.025,
  distribution = "cauchy",
  baSAR_model = NULL,
  n.MCMC = 1e+05,
  fit.method = c("EXP", "EXP+LIN", "LIN"),
  fit.force_through_origin = TRUE,
  fit.includingRepeatedRegPoints = TRUE,
  method_control = list(),
  digits = 3L,
  distribution_plot = c("kde", "abanico"),
  plot = TRUE,
  plot_reduced = TRUE,
  plot_singlePanels = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData}, \linkS4class{RLum.Results}, \link{list} of \linkS4class{RLum.Analysis},
\link{character} or \link{list} (\strong{required}):
input object used for the Bayesian analysis. If a \code{character} is provided the function
assumes a file connection and tries to import a BIN/BINX-file using the provided path. If a \code{list} is
provided the list can only contain either \code{Risoe.BINfileData} objects or \code{character}s
providing a file connection. Mixing of both types is not allowed. If an \linkS4class{RLum.Results}
is provided the function directly starts with the Bayesian Analysis (see details)}

\item{CSV_file}{\link{character} or \link{data.frame} (\emph{optional}):
if a \code{character}, it must be the path to a CSV file with data for the
analysis. Either way, data should contain 3 columns:
the name of the file, the disc position and the grain position
(the last being 0 for multi-grain measurements).}

\item{aliquot_range}{\link{numeric} (\emph{optional}):
allows to limit the range of the aliquots used for the analysis.
This argument has only an effect if the argument \code{CSV_file} is used or
\code{object} is a previous output (i.e. is \linkS4class{RLum.Results}). In this case,
the new selection will add the aliquots to the removed aliquots table.}

\item{source_doserate}{\link{numeric} (\strong{required}):
source dose rate of beta-source used for the measurement and its uncertainty
in Gy/s, e.g., \code{source_doserate = c(0.12, 0.04)}. Parameter can be provided
as a list, for the case that more than one BIN-file is provided, e.g.,
\code{source_doserate = list(c(0.04, 0.004), c(0.05, 0.004))}.}

\item{signal_integral}{\link{vector} (\strong{required}):
vector with the limits for the signal integral used for the calculation,
e.g., \code{signal_integral = 1:5}. It is ignored if \code{object} is an
\linkS4class{RLum.Results} object.
The parameter can be provided as a list, see \code{source_doserate}.}

\item{background_integral}{\link{vector} (\strong{required}):
vector with the limits for the background integral. It is ignored if
\code{object} is an \linkS4class{RLum.Results} object.
The parameter can be provided as a list, see \code{source_doserate}.}

\item{signal_integral_Tx}{\link{vector} (\emph{optional}):
vector with the limits for the signal integral for the Tx curve. If nothing
is provided, the value from \code{signal_integral} is used. It is ignored
if \code{object} is an \linkS4class{RLum.Results} object.
The parameter can be provided as a list, see \code{source_doserate}.}

\item{background_integral_Tx}{\link{vector} (\emph{optional}):
vector with the limits for the background integral for the Tx curve.
If nothing is provided, the value from \code{background_integral} is used.
It is ignored if \code{object} is an \linkS4class{RLum.Results} object.
The parameter can be provided as a list, see \code{source_doserate}.}

\item{irradiation_times}{\link{numeric} (\emph{optional}): if set this vector replaces all irradiation
times for one aliquot and one cycle (Lx and Tx curves) and recycles it for all others cycles and aliquots.
Please note that if this argument is used, for every(!) single curve
in the dataset an irradiation time needs to be set.}

\item{sigmab}{\link{numeric} (\emph{with default}):
option to set a manual value for the overdispersion (for \code{LnTx} and \code{TnTx}),
used for the \code{Lx}/\code{Tx} error calculation. The value should be provided as
absolute squared count values, cf. \link{calc_OSLLxTxRatio}.
The parameter can be provided as \code{list}, see \code{source_doserate}.}

\item{sig0}{\link{numeric} (\emph{with default}):
allow adding an extra component of error to the final Lx/Tx error value
(e.g., instrumental error, see details is \link{calc_OSLLxTxRatio}).
The parameter can be provided as \code{list}, see \code{source_doserate}.}

\item{distribution}{\link{character} (\emph{with default}):
type of distribution used during Bayesian calculations to determine the
Central dose and overdispersion values (one of \code{"cauchy"}, \code{"normal"},
\code{"log_normal"} and \code{"user_defined"}). If set to \code{"user_defined"}, argument
\code{baSAR_model} must be provided.}

\item{baSAR_model}{\link{character} (\emph{optional}):
option to provide an own modified or new model for the Bayesian calculation
(see details). If provided, argument \code{distribution} is automatically set to
\code{"user_defined"}.}

\item{n.MCMC}{\link{integer} (\emph{with default}):
number of iterations for the Markov chain Monte Carlo (MCMC) simulations}

\item{fit.method}{\link{character} (\emph{with default}):
equation used both for the fitting of the dose-response curve using function
\link{fit_DoseResponseCurve} and then for the Bayesian modelling.
Supported methods: \code{EXP}, \code{EXP+LIN} and \code{LIN}.}

\item{fit.force_through_origin}{\link{logical} (\emph{with default}):
force fitting through origin}

\item{fit.includingRepeatedRegPoints}{\link{logical} (\emph{with default}):
includes the recycling point (assumed to be measured during the last cycle)}

\item{method_control}{\link{list} (\emph{optional}):
named list of control parameters that can be directly
passed to the Bayesian analysis, e.g., \code{method_control = list(n.chains = 4)}.
See details for further information}

\item{digits}{\link{integer} (\emph{with default}):
round output to the number of given digits}

\item{distribution_plot}{\link{character} (\emph{with default}): sets the final distribution plot that
shows equivalent doses obtained using the frequentist approach and sets in the central dose
as comparison obtained using baSAR. Allowed input is \code{'abanico'} or \code{'kde'}.
If set to \code{NULL} nothing is plotted.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{plot_reduced}{\link{logical} (\emph{with default}):
enable/disable the advanced plot output.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}):
enable/disable single plots or plots arranged by \code{analyse_baSAR}.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{parameters that can be passed to the function \link{calc_OSLLxTxRatio}
(almost full support), \link[data.table:fread]{data.table::fread} (\code{skip}), \link{read_BIN2R} (\code{n.records},
\code{position}, \code{duplicated.rm}), see details.}
}
\value{
Function returns results numerically and graphically:

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$summary} \tab \code{data.frame} \tab statistical summary, including the central dose \cr
\verb{$mcmc} \tab \code{mcmc} \tab \link[coda:mcmc.list]{coda::mcmc.list} object including raw output \cr
\verb{$models} \tab \code{character} \tab implemented models used in the baSAR-model core \cr
\verb{$input_object} \tab \code{data.frame} \tab summarising table (same format as the XLS-file) including, e.g., Lx/Tx values\cr
\verb{$removed_aliquots} \tab \code{data.frame} \tab table with removed aliquots (e.g., \code{NaN}, or \code{Inf} \code{Lx}/\code{Tx} values). If nothing was removed \code{NULL} is returned
}

\strong{slot:} \strong{\verb{@info}}

The original function call

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr
\itemize{
\item (A) Ln/Tn curves with set integration limits,
\item (B) trace plots are returned by the baSAR-model, showing the convergence of the parameters (trace)
and the resulting kernel density plots. If \code{plot_reduced = FALSE} for every(!) dose a trace and
a density plot is returned (this may take a long time),
\item (C) dose plots showing the dose for every aliquot as boxplots and the marked
HPD in within. If boxes are coloured 'orange' or 'red' the aliquot itself should be checked,
\item (D) the dose response curve resulting from the monitoring of the Bayesian modelling are
provided along with the Lx/Tx values and the HPD. Note: The amount for curves displayed
is limited to 1000 (random choice) for performance reasons,
\item (E) the final plot is the De distribution as calculated using the conventional (frequentist) approach
and the central dose with the HPDs marked within. This figure is only provided for a comparison,
no further statistical conclusion should be drawn from it.
}

\strong{Please note: If distribution was set to \code{log_normal} the central dose is given as geometric mean!}
}
\description{
This function allows the application of Bayesian models on luminescence data
measured with the single-aliquot regenerative-dose (SAR, Murray and Wintle,
2000) protocol. In particular, it follows the idea proposed by Combès et al.,
2015 of using an hierarchical model for estimating a central equivalent dose
from a set of luminescence measurements. This function (I) implements this
approach for the R environment and (II) provides an extension and a technical
refinement of the published code.
}
\details{
Internally the function consists of two parts: (I) a Bayesian core that
performs the hierarchical modelling and calculations, and (II) a data
pre-processing stage. The Bayesian core can be run independently if sufficient
input data are provided (see below). The data pre-processing stage exists
to simplify the user workflow by preparing all required inputs; in principle
it is enough to provide a BIN/BINX-file with the SAR measurement data. For
the Bayesian analysis for each aliquot the following information are needed
from the SAR analysis: \code{LxTx}, the \code{LxTx} error and the dose values for all
regeneration points.

\strong{How is the systematic error contribution calculated?}

Standard errors (so far) provided with the source dose rate are considered
as systematic uncertainties and added to final central dose by:

\deqn{systematic.error = 1/n \sum SE(source.doserate)}

\deqn{SE(central.dose.final) = \sqrt{SE(central.dose)^2 + systematic.error^2}}

Please note that this approach is rather rough and can only be valid if the
source dose rate errors, in case different readers had been used, are
similar. In cases where more than one source dose rate is provided a
warning is given.

\strong{Input / output scenarios}

Various inputs are allowed for this function. Unfortunately this makes the
function handling rather complex, but at the same time very powerful.
Available scenarios:

\strong{(1) - \code{object} is BIN-file or link to a BIN-file}

It does not matter how the information of the BIN/BINX file are provided.
The function supports \strong{(a)} either a path to a file or directory or a
\code{list} of file names or paths or  \strong{(b)} a \linkS4class{Risoe.BINfileData} object
or a list of these objects. The latter one can be produced by using function
\link{read_BIN2R}, but this function is called automatically if only a file name
and/or a path is provided. In both cases it will become the data that can be
used for the analysis.

\verb{[CSV_file = NULL]}

If no CSV file (or data frame with the same format) is provided, the
function runs an automatic process that consists of the following steps:
\enumerate{
\item Select all valid aliquots using the function \link{verify_SingleGrainData}
\item Calculate \code{Lx/Tx} values using the function \link{calc_OSLLxTxRatio}
\item Calculate De values using the function \link{fit_DoseResponseCurve}
}

These proceeded data are subsequently used in for the Bayesian analysis

\verb{[CSV_file != NULL]}

If a CSV file is provided (or a \code{data.frame} containing similar information)
the pre-processing phase consists of the following steps:
\enumerate{
\item Calculate \code{Lx/Tx} values using the function \link{calc_OSLLxTxRatio}
\item Calculate De values using the function \link{fit_DoseResponseCurve}
}

The CSV file should contain the BIN-file names and the aliquots selected
for the further analysis. This allows a manual selection of input data, as
the automatic selection by \link{verify_SingleGrainData} might not be sufficient.

\strong{(2) - \code{object} \verb{RLum.Results object}}

If an \linkS4class{RLum.Results} object is provided as input and(!) this object was
previously created by the function \code{analyse_baSAR()} itself, the pre-processing part
is skipped and the function starts directly with the Bayesian analysis. This option is very powerful
as it allows to change parameters for the Bayesian analysis without the need to repeat
the data pre-processing. If furthermore the argument \code{aliquot_range} is set, aliquots
can be manually excluded based on previous runs.

\strong{\code{method_control}}

These are arguments that can be passed directly to the Bayesian calculation core, supported arguments
are:

\tabular{lll}{
\strong{Parameter} \tab \strong{Type} \tab \strong{Description}\cr
\code{lower_centralD} \tab \link{numeric} \tab sets the lower bound for the expected De range. Change it only if you know what you are doing!\cr
\code{upper_centralD} \tab \link{numeric} \tab sets the upper bound for the expected De range. Change it only if you know what you are doing!\cr
\code{n.chains} \tab \link{integer} \tab sets number of parallel chains for the model (default = 3) (cf. \link[rjags:jags.model]{rjags::jags.model})\cr
\code{inits} \tab \link{list} \tab option to set initialisation values (cf. \link[rjags:jags.model]{rjags::jags.model}) \cr
\code{thin} \tab \link{numeric} \tab thinning interval for monitoring the Bayesian process (cf. \link[rjags:jags.model]{rjags::jags.model})\cr
\code{variable.names} \tab \link{character} \tab set the variables to be monitored during the MCMC run, default:
\code{'central_D'}, \code{'sigma_D'}, \code{'D'}, \code{'Q'}, \code{'a'}, \code{'b'}, \code{'c'}, \code{'g'}.
Note: only variables present in the model can be monitored.
}

\strong{User defined models}\cr

The function provides the option to modify and to define own models that can be used for
the Bayesian calculation. In the case the user wants to modify a model, a new model
can be piped into the function via the argument \code{baSAR_model} as \code{character}.
The model has to be provided in the JAGS dialect of the BUGS language (cf. \link[rjags:jags.model]{rjags::jags.model})
and parameter names given with the pre-defined names have to be respected, otherwise the function
will break.

\strong{FAQ}

Q: How can I set the seed for the random number generator (RNG)?

A: Use the argument \code{method_control}, e.g., for three MCMC chains
(as it is the default):

\if{html}{\out{<div class="sourceCode">}}\preformatted{method_control = list(
inits = list(
 list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 1),
 list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 2),
 list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 3)
))
}\if{html}{\out{</div>}}

This sets a reproducible set for every chain separately.\cr

Q: How can I modify the output plots?

A: You can't, but you can use the function output to create own, modified plots.

Q: Can I change the boundaries for the central_D?

A: Yes, we made it possible, but we DO NOT recommend it, except you know what you are doing!\cr
Example: \verb{method_control = list(lower_centralD = 10))}

Q: The lines in the baSAR-model appear to be in a wrong logical order?\cr

A: This is correct and allowed (cf. JAGS manual)

\strong{Additional arguments support via the \code{...} argument}

This list summarizes the additional arguments that can be passed to the internally used
functions.

\tabular{llll}{
\strong{Supported argument} \tab \strong{Corresponding function} \tab \strong{Default} \tab \strong{Short description}\cr
\code{threshold} \tab \link{verify_SingleGrainData} \tab \code{30} \tab change rejection threshold for curve selection \cr
\code{skip} \tab \link[data.table:fread]{data.table::fread} \tab \code{0} \tab number of rows to be skipped during import\cr
\code{n.records} \tab \link{read_BIN2R} \tab \code{NULL} \tab limit records during BIN-file import\cr
\code{duplicated.rm} \tab \link{read_BIN2R} \tab \code{TRUE} \tab remove duplicated records in the BIN-file\cr
\code{pattern} \tab \link{read_BIN2R} \tab \code{TRUE} \tab select BIN-file by name pattern\cr
\code{position} \tab \link{read_BIN2R} \tab \code{NULL} \tab limit import to a specific position\cr
\code{background.count.distribution} \tab \link{calc_OSLLxTxRatio} \tab \code{"non-poisson"} \tab set assumed count distribution\cr
\code{fit.weights} \tab \link{fit_DoseResponseCurve} \tab \code{TRUE} \tab enable/disable fit weights\cr
\code{fit.bounds} \tab \link{fit_DoseResponseCurve} \tab \code{TRUE} \tab enable/disable fit bounds\cr
\code{n.MC} \tab \link{fit_DoseResponseCurve} \tab \code{100} \tab number of MC runs for error calculation\cr
\code{plot_drc} \tab \link{plot_DoseResponseCurve} \tab \code{TRUE} \tab enable/disable dose response curve plot\cr
\code{plot_extended} \tab \link{plot_DoseResponseCurve} \tab \code{TRUE} \tab enable/disable extended dose response curve plot\cr
\code{recordType} \tab \link{get_RLum} \tab \verb{c(OSL (UVVIS), irradiation (NA)} \tab helps for the curve selection\cr
}
}
\note{
\strong{If you provide more than one BIN-file}, it is \strong{strongly} recommended to provide
a \code{list} with the same number of elements for the following parameters:

\code{source_doserate}, \code{signal_integral}, \code{signal_integral_Tx}, \code{background_integral},
\code{background_integral_Tx}, \code{sigmab}, \code{sig0}.

Example for two BIN-files: \code{source_doserate = list(c(0.04, 0.006), c(0.05, 0.006))}

\strong{The function is currently limited to work with standard Risoe BIN-files only!}
}
\section{Function version}{
 0.1.41
}

\examples{

##(1) load package test data set
data(ExampleData.BINfileData, envir = environment())

##(2) selecting relevant curves, and limit dataset
CWOSL.SAR.Data <- subset(
  CWOSL.SAR.Data,
  subset = POSITION\%in\%c(1:3) & LTYPE == "OSL")

\dontrun{
##(3) run analysis
## please note that the parameters used here are
## chosen for performance, not for reliability
results <- analyse_baSAR(
  object = CWOSL.SAR.Data,
  source_doserate = c(0.04, 0.001),
  signal_integral = 1:2,
  background_integral = 80:100,
  fit.method = "LIN",
  plot = FALSE,
  n.MCMC = 200
)

print(results)

##CSV_file template
##copy and paste this the code below in the terminal
##you can further use the function write.csv() to export the example

CSV_file <-
structure(
list(
 BIN_FILE = NA_character_,
 DISC = NA_real_,
 GRAIN = NA_real_),
   .Names = c("BIN_FILE", "DISC", "GRAIN"),
   class = "data.frame",
   row.names = 1L)
}

} 

\section{How to cite}{
Mercier, N., Kreutzer, S., 2026. analyse_baSAR(): Bayesian models (baSAR) applied on luminescence data. Function version 0.1.41. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Combès, B., Philippe, A., Lanos, P., Mercier, N., Tribolo, C., Guérin, G., Guibert, P., Lahaye, C., 2015.
A Bayesian central equivalent dose model for optically stimulated luminescence dating.
Quaternary Geochronology 28, 62-70. doi:10.1016/j.quageo.2015.04.001

Mercier, N., Kreutzer, S., Christophe, C., Guérin, G., Guibert, P., Lahaye, C., Lanos, P., Philippe, A.,
Tribolo, C., 2016. Bayesian statistics in luminescence dating: The 'baSAR'-model and its implementation
in the R package 'Luminescence'. Ancient TL 34, 14-21.

\strong{Further reading}

Gelman, A., Carlin, J.B., Stern, H.S., Dunson, D.B., Vehtari, A., Rubin, D.B., 2013.
Bayesian Data Analysis, Third Edition. CRC Press.

Murray, A.S., Wintle, A.G., 2000. Luminescence dating of quartz using an improved single-aliquot
regenerative-dose protocol. Radiation Measurements 32, 57-73. doi:10.1016/S1350-4487(99)00253-X

Plummer, M., 2017. JAGS Version 4.3.0 user manual. \verb{https://sourceforge.net/projects/mcmc-jags/files/Manuals/4.x/jags_user_manual.pdf/download}
}
\seealso{
\link{read_BIN2R}, \link{calc_OSLLxTxRatio},
\link{fit_DoseResponseCurve}, \link{plot_DoseResponseCurve},
\link[data.table:fread]{data.table::fread}, \link{verify_SingleGrainData},
\link[rjags:jags.model]{rjags::jags.model}, \link[rjags:coda.samples]{rjags::coda.samples}, \link{boxplot.default}
}
\author{
Norbert Mercier, Archéosciences Bordeaux, CNRS-Université Bordeaux Montaigne (France) \cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
The underlying Bayesian model based on a contribution by Combès et al., 2015.
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_RF2R.R
\name{read_RF2R}
\alias{read_RF2R}
\title{Import RF-files to R}
\usage{
read_RF2R(file, verbose = TRUE, ...)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of one or multiple RF files (URLs are supported); it can be the path
to a directory, in which case the function tries to detect and import all
RF files found in the directory.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{not used, only for compatible reasons}
}
\value{
Returns an S4 \linkS4class{RLum.Analysis} object containing
\linkS4class{RLum.Data.Curve} objects for each curve. Results are
returned as a list when multiple files are processed or \code{file} is a list.
}
\description{
Import files produced by the IR-RF 'ImageJ' macro (\code{SR-RF.ijm};
Mittelstraß and Kreutzer, 2021) into R and create a list of
\linkS4class{RLum.Analysis} objects
}
\details{
The results of spatially resolved IR-RF data are summarised in
so-called RF-files (Mittelstraß and Kreutzer, 2021).
This functions provides an easy import to process the data seamlessly with the R package 'Luminescence'.
The output of the function can be passed to function \link{analyse_IRSAR.RF}.
}
\section{Function version}{
 0.1.2
}

\examples{

##Import
file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
temp <- read_RF2R(file)

} 

\section{How to cite}{
Kreutzer, S., 2026. read_RF2R(): Import RF-files to R. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Mittelstraß, D., Kreutzer, S., 2021. Spatially resolved infrared radiofluorescence:
single-grain K-feldspar dating using CCD imaging. Geochronology 3, 299–319. \doi{10.5194/gchron-3-299-2021}
}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve},
\link{analyse_IRSAR.RF}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.SurfaceExposure}
\alias{ExampleData.SurfaceExposure}
\title{Example OSL surface exposure dating data}
\format{
A \link{list} with 4 elements:

\tabular{ll}{
\strong{Element} \tab \strong{Content} \cr
\verb{$sample_1} \tab A \link{data.frame} with 3 columns (depth, intensity, error) \cr
\verb{$sample_2} \tab A \link{data.frame} with 3 columns (depth, intensity, error) \cr
\verb{$set_1} \tab A \link{list} of 4 \link{data.frame}s, each representing a sample with different ages \cr
\verb{$set_2} \tab A \link{list} of 5 \link{data.frame}s, each representing a sample with different ages \cr
}
}
\source{
See examples for the code used to create the data sets.
}
\description{
A set of synthetic OSL surface exposure dating data to demonstrate the
\link{fit_SurfaceExposure} functionality. See examples to reproduce the data
interactively.
}
\details{
\strong{\verb{$sample_1}}

\tabular{ccc}{
\strong{mu} \tab \strong{\code{sigmaphi}} \tab \strong{age} \cr
0.9 \tab 5e-10 \tab 10000 \cr
}

\strong{\verb{$sample_2}}

\tabular{ccccc}{
\strong{mu} \tab \strong{\code{sigmaphi}} \tab \strong{age} \tab \strong{Dose rate} \tab \strong{D0} \cr
0.9 \tab 5e-10 \tab 10000 \tab 2.5 \tab 40 \cr
}

\strong{\verb{$set_1}}

\tabular{ccc}{
\strong{mu} \tab \strong{\code{sigmaphi}} \tab \strong{ages} \cr
0.9 \tab 5e-10 \tab 1e3, 1e4, 1e5, 1e6 \cr
}

\strong{\verb{$set_2}}

\tabular{ccccc}{
\strong{mu} \tab \strong{\code{sigmaphi}} \tab \strong{ages} \tab \strong{Dose rate} \tab \strong{D0} \cr
0.9 \tab 5e-10 \tab 1e2, 1e3, 1e4, 1e5, 1e6 \tab 1.0 \tab 40 \cr
}
}
\examples{

## ExampleData.SurfaceExposure$sample_1
sigmaphi <- 5e-10
age <- 10000
mu <- 0.9
x <- seq(0, 10, 0.1)
fun <- exp(-sigmaphi * age * 365.25*24*3600 * exp(-mu * x))

set.seed(666)
synth_1 <- data.frame(depth = x,
                      intensity = jitter(fun, 1, 0.1),
                      error = runif(length(x), 0.01, 0.2))

## VALIDATE sample_1
fit_SurfaceExposure(synth_1, mu = mu, sigmaphi = sigmaphi)


## ExampleData.SurfaceExposure$sample_2
sigmaphi <- 5e-10
age <- 10000
mu <- 0.9
x <- seq(0, 10, 0.1)
Ddot <- 2.5   / 1000 / 365.25 / 24 / 60 / 60 # 2.5 Gy/ka in Seconds
D0 <- 40
fun <- (sigmaphi * exp(-mu * x) *
          exp(-(age * 365.25*24*3600) *
                (sigmaphi * exp(-mu * x) + Ddot/D0)) + Ddot/D0) /
  (sigmaphi * exp(-mu * x) + Ddot/D0)

set.seed(666)
synth_2 <- data.frame(depth = x,
                      intensity = jitter(fun, 1, 0.1),
                      error = runif(length(x), 0.01, 0.2))

## VALIDATE sample_2
fit_SurfaceExposure(synth_2, mu = mu, sigmaphi = sigmaphi, Ddot = 2.5, D0 = D0)


## ExampleData.SurfaceExposure$set_1
sigmaphi <- 5e-10
mu <- 0.9
x <- seq(0, 15, 0.2)
age <- c(1e3, 1e4, 1e5, 1e6)
set.seed(666)

synth_3 <- vector("list", length = length(age))

for (i in 1:length(age)) {
  fun <- exp(-sigmaphi * age[i] * 365.25*24*3600 * exp(-mu * x))
  synth_3[[i]] <- data.frame(depth = x,
                             intensity = jitter(fun, 1, 0.05))
}


## VALIDATE set_1
fit_SurfaceExposure(synth_3, age = age, sigmaphi = sigmaphi)


## ExampleData.SurfaceExposure$set_2
sigmaphi <- 5e-10
mu <- 0.9
x <- seq(0, 15, 0.2)
age <- c(1e2, 1e3, 1e4, 1e5, 1e6)
Ddot <- 1.0 / 1000 / 365.25 / 24 / 60 / 60 # 2.0 Gy/ka in Seconds
D0 <- 40
set.seed(666)

synth_4 <- vector("list", length = length(age))

for (i in 1:length(age)) {
  fun <- (sigmaphi * exp(-mu * x) *
            exp(-(age[i] * 365.25*24*3600) *
                  (sigmaphi * exp(-mu * x) + Ddot/D0)) + Ddot/D0) /
    (sigmaphi * exp(-mu * x) + Ddot/D0)

  synth_4[[i]] <- data.frame(depth = x,
                             intensity = jitter(fun, 1, 0.05))
}


## VALIDATE set_2
fit_SurfaceExposure(synth_4, age = age, sigmaphi = sigmaphi, D0 = D0, Ddot = 1.0)

\dontrun{
ExampleData.SurfaceExposure <- list(
  sample_1 = synth_1,
  sample_2 = synth_2,
  set_1 = synth_3,
  set_2 = synth_4
)
}

}
\references{
Unpublished synthetic data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_DRAC.R
\name{use_DRAC}
\alias{use_DRAC}
\title{Use DRAC to calculate dose rate data}
\usage{
use_DRAC(
  file,
  name = NULL,
  print_references = TRUE,
  citation_style = "text",
  ...
)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of a CSV file (formatted according to the DRAC v1.3 CSV template) to
be sent to the DRAC website for calculation. It can also be a DRAC template
object obtained from \code{\link[=template_DRAC]{template_DRAC()}}, which supports also import from
CSV-files.}

\item{name}{\link{character} (\emph{with default}):
Optional user name submitted to DRAC. If \code{NULL}, a random name will be
generated.}

\item{print_references}{(\emph{with default}):
Print all references used in the input data table to the console.}

\item{citation_style}{(\emph{with default}):
If \code{print_references = TRUE} this argument determines the output style of the
used references. Valid options are \code{"Bibtex"}, \code{"citation"}, \code{"html"}, \code{"latex"}
or \code{"R"}. Default is \code{"text"}.}

\item{...}{further arguments:
\itemize{
\item \code{url} \link{character}: provide an alternative URL to DRAC
\item \code{ignore_version} \link{logical}: ignores the version check, this might come in handy
if the version has changed, but not the column order
\item \code{user} \link{character}: option to provide username for secured site
\item \code{password} \link{character}: password for secured site, only works jointly with \code{user}
\item \code{verbose} \link{logical}: enable/disable output to the terminal
}}
}
\value{
Returns an \linkS4class{RLum.Results} object containing the following elements:

\item{DRAC}{\link{list}: a named list containing the following elements in slot \verb{@data}:

\tabular{lll}{
\verb{$highlights} \tab \link{data.frame} \tab summary of 25 most important input/output fields \cr
\verb{$header} \tab \link{character} \tab HTTP header from the DRAC server response \cr
\verb{$labels} \tab \link{data.frame} \tab descriptive headers of all input/output fields \cr
\verb{$content} \tab \link{data.frame} \tab complete DRAC input/output table \cr
\verb{$input} \tab \link{data.frame} \tab DRAC input table \cr
\verb{$output} \tab \link{data.frame} \tab DRAC output table \cr
\code{references}\tab \link{list} \tab A list of bib entries of used references \cr
}

}
\item{data}{\link{character} or \link{list} path to the input spreadsheet or a DRAC template}
\item{call}{\link{call} the function call}
\item{args}{\link{list} used arguments}

The output should be accessed using the function \link{get_RLum}.
}
\description{
The function provides an interface from R to DRAC. An R-object or a
CSV file is passed to the DRAC website and results are re-imported into R.
}
\section{Function version}{
 0.18
}

\examples{

## (1) Method using the DRAC spreadsheet

file <-  "/PATH/TO/DRAC_Input_Template.csv"

# send the actual IO template spreadsheet to DRAC
\dontrun{
use_DRAC(file = file)
}

## (2) Method using an R template object
# Create a template
input <- template_DRAC(preset = "DRAC-example_quartz")

# Fill the template with values
input$`Project ID` <- "DRAC-Example"
input$`Sample ID` <- "Quartz"
input$`Conversion factors` <- "AdamiecAitken1998"
input$`External U (ppm)` <- 3.4
input$`errExternal U (ppm)` <- 0.51
input$`External Th (ppm)` <- 14.47
input$`errExternal Th (ppm)` <- 1.69
input$`External K (\%)` <- 1.2
input$`errExternal K (\%)` <- 0.14
input$`Calculate external Rb from K conc?` <- "N"
input$`Calculate internal Rb from K conc?` <- "N"
input$`Scale gammadoserate at shallow depths?` <- "N"
input$`Grain size min (microns)` <- 90
input$`Grain size max (microns)` <- 125
input$`Water content ((wet weight - dry weight)/dry weight) \%` <- 5
input$`errWater content \%` <- 2
input$`Depth (m)` <- 2.2
input$`errDepth (m)` <- 0.22
input$`Overburden density (g cm-3)` <- 1.8
input$`errOverburden density (g cm-3)` <- 0.1
input$`Latitude (decimal degrees)` <- 30.0000
input$`Longitude (decimal degrees)` <- 70.0000
input$`Altitude (m)` <- 150
input$`De (Gy)` <- 20
input$`errDe (Gy)` <- 0.2

# use DRAC
\dontrun{
output <- use_DRAC(input)

## export as LaTeX table
.as.latex.table(output)
}

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., Burow, C., 2026. use_DRAC(): Use DRAC to calculate dose rate data. Function version 0.18. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Durcan, J.A., King, G.E., Duller, G.A.T., 2015. DRAC: Dose Rate and Age Calculator for trapped charge dating.
Quaternary Geochronology 28, 54-61. \doi{10.1016/j.quageo.2015.03.012}
}
\seealso{
\link{template_DRAC}, \link{.as.latex.table}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany)\cr
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_MoranScatterplot.R
\name{plot_MoranScatterplot}
\alias{plot_MoranScatterplot}
\title{Moran Scatter Plot: Visualizing Spatial Dependency}
\usage{
plot_MoranScatterplot(
  object,
  df_neighbours = NULL,
  str_y_def = c("mean_neighbours", "weighted_sum"),
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} or \link{numeric} (\strong{required}):
containing a numerical vector of length 100,
representing one or more measurement discs ("positions") in a reader.
Each element in the vector represents one grain hole location on a disc.}

\item{df_neighbours}{\link{data.frame} (\emph{with default}):
Data frame indicating
which borders to consider, and their respective weights (see the description
provided for \link{calc_MoransI}). If \code{NULL} (default), this is constructed
automatically by the internal function \code{.get_Neighbours}.}

\item{str_y_def}{\link{character} (\emph{with default}):
Calculation of y position. Defaults to \code{"mean_neighbours"}
which is the plain mean of all neighbour values and the most illustrative. The other option is \code{"weighted_sum"},
which means the sum of the border weights times the neighbour values, which is actually closer
to the way Moran's I is by default calculated in this package.}

\item{...}{Other parameters to be forwarded to the base R plot functions.
\code{legend} (\code{TRUE/FALSE}) to enable/disable the legend.
Note that \code{xlab} (x axis label), \code{ylab} (y axis label) and \code{cex} (scaling
value) are given default values. Because of sometimes large value
differences, \code{log = "x"}, \code{log = "y"} and \code{log = "xy"}are supported.
In case of negative values and logarithmic plotting, values are increased
so the smallest value to plot is 1. Summary elements such as means, least
square line etc. will still be based on the linear case.
\code{pch} accepts options \code{"show_location_ids"} (plots grain location id's),
\code{"show_n_neighbours"} (indicates numbers of neighbours) and the normal
base plot \code{pch} options.}
}
\value{
Returns (invisibly) a data frame with plotting coordinates and
grain location id's, for creating user-defined plots.
}
\description{
The function produces a scatter plot, with on the x axis the original grain
signal and on the y axis the weighted mean of the neighbour grain signals.
The plot area is divided into four quadrants, and also a least square line
(which slopes indicates, but not exactly represents, Moran's I) and
an 1:1 line (which indicates a Moran's I of around 1).
}
\details{
Note that this function plots on the y-axis the mean of the neighbours, while the function
\link{calc_MoransI} by default will for its global calculation weight every border the same. So, grain locations
with 1, 2 or 3 neighbours will appear higher on the y-axis than their influence on Moran's I justify -- apart
from scaling, this explains a part of the differences of Moran's scatter plots between different packages.
Also note that island' grain locations (=those not bordering other grains) are left out of these plots but
might still influence Moran's I calculations.
}
\examples{
plot_MoranScatterplot(1:100)

} 

\section{How to cite}{
Boer, A.d., Steinbuch, L., 2026. plot_MoranScatterplot(): Moran Scatter Plot: Visualizing Spatial Dependency. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
A novel method to assess crosstalk in single-grain luminescence detection.
Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
}
\author{
Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_BIN2R.R
\name{read_BIN2R}
\alias{read_BIN2R}
\title{Import Risø BIN/BINX-files into R}
\usage{
read_BIN2R(
  file,
  show.raw.values = FALSE,
  position = NULL,
  n.records = NULL,
  zero_data.rm = TRUE,
  duplicated.rm = FALSE,
  fastForward = FALSE,
  show.record.number = FALSE,
  txtProgressBar = TRUE,
  forced.VersionNumber = NULL,
  ignore.RECTYPE = FALSE,
  pattern = NULL,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{file}{\link{character} or \link{list} (\strong{required}):
name of one or multiple BIN/BINX files (URLs are supported); it can be the
path to a directory, in which case the function tries to detect and import
all BIN/BINX files found recursively from the given directory.}

\item{show.raw.values}{\link{logical} (\emph{with default}):
shows raw values from BIN-file for \code{LTYPE}, \code{DTYPE} and \code{LIGHTSOURCE} without
translation in characters. Can be provided as \code{list} if \code{file} is a \code{list}.}

\item{position}{\link{numeric} (\emph{optional}):
imports only the selected position. Note: the import performance will not
benefit by any selection made here.
Can be provided as \code{list} if \code{file} is a \code{list}.}

\item{n.records}{\link{numeric} (\emph{optional}): limits the number of imported records
to the provided record id (e.g., \code{n.records = 1:10} imports the first ten records,
while \code{n.records = 3} imports only record number 3. Can be used in combination with
\code{show.record.number} for debugging purposes, e.g. corrupt BIN-files.
Can be provided as \code{list} if \code{file} is a \code{list}.}

\item{zero_data.rm}{\link{logical} (\emph{with default}):
remove erroneous data with no count values. As such data are usually not
needed for the subsequent data analysis they will be removed by default.
Can be provided as \code{list} if \code{file} is a \code{list}.}

\item{duplicated.rm}{\link{logical} (\emph{with default}):
remove duplicated entries if \code{TRUE}. This may happen due to an erroneous
produced BIN/BINX-file. This option compares only predecessor and successor.
Can be provided as \code{list} if \code{file} is a \code{list}.}

\item{fastForward}{\link{logical} (\emph{with default}):
if \code{TRUE} for a more efficient data processing only a list of
\linkS4class{RLum.Analysis}
objects is returned instead of a \linkS4class{Risoe.BINfileData} object.
Can be provided as \code{list} if \code{file} is a \code{list}.}

\item{show.record.number}{\link{logical} (\emph{with default}):
shows record number of the imported record, for debugging usage only.
Can be provided as \code{list} if \code{file} is a \code{list}.
Ignored if \code{verbose = FALSE}.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar. Ignored if \code{verbose = FALSE}.}

\item{forced.VersionNumber}{\link{integer} (\emph{optional}):
allows to cheat the version number check in the function by own values for
cases where the BIN-file version is not supported.
Can be provided as \code{list} if \code{file} is a \code{list}.

\strong{Note:} The usage is at own risk, only supported BIN-file versions have been tested.}

\item{ignore.RECTYPE}{\link{logical} or \link{numeric} (\emph{with default}):
this argument allows to ignore values in the byte 'RECTYPE' (BIN-file version 08),
in case there are not documented or faulty set. In this case the corrupted records are skipped.
If the setting is \link{numeric} (e.g., \code{ignore.RECTYPE = 128}), records of those type are ignored
for import.}

\item{pattern}{\link{character} (\emph{optional}):
regular expression pattern passed to \link{list.files} to construct a list of
files to read (used only when \code{file} specifies a path).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{further arguments that will be passed to the function
\link{Risoe.BINfileData2RLum.Analysis}. Please note that any matching argument
automatically sets \code{fastForward = TRUE}}
}
\value{
Returns an S4 \linkS4class{Risoe.BINfileData} object containing two
slots:

\item{METADATA}{A \link{data.frame} containing all variables stored in the BIN-file.}
\item{DATA}{A \link{list} containing a numeric \link{vector} of the measured data.
The ID corresponds to the record ID in METADATA.}

If \code{fastForward = TRUE} an \linkS4class{RLum.Analysis} object is
returned (coerced by \link{Risoe.BINfileData2RLum.Analysis}).

Results are returned as a list when multiple files are processed or \code{file}
is a list.
}
\description{
Import a \verb{*.bin} or a \verb{*.binx} file produced by a Risø DA15 and DA20 TL/OSL
reader into R.
}
\details{
The binary data file is parsed byte by byte following the data structure
published in the Appendices of the Analyst manual p. 42.

For the general BIN/BINX-file structure, the reader is referred to the
Risø website: \url{https://www.fysik.dtu.dk}
}
\note{
The function works for BIN/BINX-format versions 03, 04, 05, 06, 07 and 08. The
version number depends on the used Sequence Editor.
}
\section{Function version}{
 0.19
}

\examples{

file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
temp <- read_BIN2R(file)
temp

} 

\section{How to cite}{
Kreutzer, S., Fuchs, M.C., Colombo, M., 2026. read_BIN2R(): Import Risø BIN/BINX-files into R. Function version 0.19. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
DTU Nutech, 2016. The Sequence Editor, Users Manual, February, 2016.
\url{https://www.fysik.dtu.dk}
}
\seealso{
\link{write_R2BIN}, \linkS4class{Risoe.BINfileData},
\link[base:readBin]{base::readBin}, \link{merge_Risoe.BINfileData}, \linkS4class{RLum.Analysis}
\link[utils:txtProgressBar]{utils::txtProgressBar}, \link{list.files}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Margret C. Fuchs, HZDR Freiberg, (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
based on information provided by Torben Lapp and Karsten Bracht Nielsen (Risø DTU, Denmark)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.CobbleData}
\alias{ExampleData.CobbleData}
\title{Example data for calc_CobbleDoseRate()}
\format{
A \code{\link{data.frame}}. Please see \link{calc_CobbleDoseRate} for detailed information
on the structure of the \link{data.frame}.
}
\description{
An example data set for the function \link{calc_CobbleDoseRate} containing layer
specific information for the cobble to be used in the function.
}
\section{Version}{
 0.1.0
}

\examples{

## Load data
data("ExampleData.CobbleData", envir = environment())

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_MoransI.R
\name{calc_MoransI}
\alias{calc_MoransI}
\title{Calculate Moran's I}
\usage{
calc_MoransI(
  object,
  df_neighbours = NULL,
  spatial_autocorrelation = TRUE,
  compute_pseudo_p = FALSE,
  tested_moransI = NULL,
  n_permutations = 999,
  ignore_borders = FALSE,
  return_intermediate_values = FALSE
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} or \link{numeric} (\strong{required}):
values of the grains of one disc. Should have length 100; can contain \code{NA}
values.}

\item{df_neighbours}{\link{data.frame} (\emph{with default}): a data frame with 3
columns (\code{location}, \code{neighbour} and \code{weight}, respectively), with each row
indicating the index of one location and one of its neighbours (note that
the concept of "location" versus "neighbour" is symmetric, so each
neighbouring pair needs to be specified only once), alongside their relative
weight (generally set to 1). If \code{NULL} (default), this is constructed
automatically by the internal function \code{.get_Neighbours}.}

\item{spatial_autocorrelation}{\link{logical} (\emph{with default}): whether spatial
autocorrelation should be considered in the computation of Moran's I
(\code{TRUE} by default). If \code{FALSE}, the function computes Moran's I expected
value in case of no autocorrelation (H_0). See details for further
information.}

\item{compute_pseudo_p}{\link{logical} (\emph{with default}): whether a pseudo p-value
should be computed (\code{FALSE} by default).}

\item{tested_moransI}{\link{numeric} (\emph{with default}): The value of Moran's I to
be tested against when computing the pseudo p-value. If \code{NULL} (default), the
value calculated by the function will be used. Ignored if \code{compute_pseudo_p}
is \code{FALSE}.}

\item{n_permutations}{\link{integer} (\emph{with default}): number of random
permutations tested to calculate the fraction which is the pseudo p
(defaults to 999). Influences the calculation speed, which will have
impact in case of large scale simulation loops. Ignored if \code{compute_pseudo_p}
is \code{FALSE}.}

\item{ignore_borders}{\link{logical} (\emph{with default}): whether only grain
locations that do not lie on the border of the disc should be considered
(\code{FALSE} by default). Thus if \code{TRUE}, only the inner 8x8 grain locations
rather than the full 10x10 are considered. Ignored if \code{df_neighbours} is
not \code{NULL} or if \code{spatial_autocorrelation = FALSE}.}

\item{return_intermediate_values}{\link{logical} (\emph{with default}): whether the
function should return a list with several intermediate calculation results
(defaults to \code{FALSE}). Ignored if \code{spatial_autocorrelation} is \code{FALSE}.}
}
\value{
By default one numerical value, roughly between -1 and 1, where
close to zero means no spatial correlation, and value close to 1 a positive
spatial correlation given the pattern we interested in (by default all rook
neighbours). A value closer to -1 has no meaning within the context of
luminescence crosstalk. If \code{compute_pseudo_p = TRUE}, then the computed
pseudo p-value is returned. If \code{return_intermediate_values} is set to \code{TRUE},
a list with several values used for calculation is returned instead of a
single outcome.
}
\description{
Calculate Moran's I
}
\details{
\strong{Case of no spatial autocorrelation}

Perhaps a bit counter-intuitive, the expected value of Moran's I under the
null hypothesis of no spatial correlation is a value slightly smaller than
zero. When setting \code{spatial_autocorrelation = FALSE}, this function
calculates the expected value based on the number of observations or the
length of the observation vector (while taking out \code{NA} values). Note that
the expected value only depends on the number of observed separate grain
values. This can be useful for plotting.

The expected Moran's I for the null hypothesis of no spatial correlation
corresponds to \code{-1 / (n - 1)}, with \code{n} being the number of non-missing
observations.
}
\examples{

## Test a fictional sample with spatial correlation
calc_MoransI(object = c(1:100))

## Test some fictional samples without spatial correlation;
## note the randomness with each repetition
calc_MoransI(object = rnorm(n = 100))
calc_MoransI(object = rnorm(n = 100))
calc_MoransI(object = rnorm(n = 100))

} 

\section{How to cite}{
Boer, A.d., Steinbuch, L., 2026. calc_MoransI(): Calculate Moran's I. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
A novel method to assess crosstalk in single-grain luminescence detection.
Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
}
\author{
Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_BINXLOG2R.R
\name{read_BINXLOG2R}
\alias{read_BINXLOG2R}
\title{Import Risø BINX-data from the BINX Log File}
\usage{
read_BINXLOG2R(file, verbose = TRUE, ...)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of one or multiple ASCII log files to read; it can be a path to a
directory, in which case the function tries to read all files it finds
(there is no file extension check, so it may fail ungracefully on binary
files).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{further arguments that will be passed to the function (currently not used)}
}
\value{
Returns an S4 \linkS4class{RLum.Analysis} object. Results are
returned as a list when multiple files are processed or \code{file} is a list.
}
\description{
The function attempts to extract meaningful measurement data from the
log file (\code{.txt}) stored along with a BINX file if this option was chosen.
This is particularly helpful if the BINX file is broken and the data would
be lost otherwise.
}
\details{
This function is basically a hack trying to make the most out of a broken file in case
at least the log file is salvageable. Please do not expect black magic; the extract is
very basic, and the function was written to work for TL, OSL, IRSL curves only.
If it breaks, it breaks.

\strong{TL curve extraction}
Since the BIN/BINX do not contain x-values, they also cannot be created magically
here. To avoid problems with the conversion, we assume always BIN-file version \code{3} for
the conversion. This should be okay in most cases, however, it may become problematic
if you have data for dedicated curve analysis.
}
\note{
This function tries to extract data from a log file produced for debugging
purposes. There is no guarantee regarding the format stability, and the
function is meant as a last resort if your BIN/BINX files are unusable.
}
\section{Function version}{
 0.1.1
}

\examples{
## read log file (RLum.Analysis)
file <-  system.file("extdata/BINX_IRSL_LOG.TXT", package = "Luminescence")
object <-  read_BINXLOG2R(file = file)

## convert to RisoeBINFile-class object
object <- convert_RLum2Risoe.BINfileData(object)

## export as BINX
\dontrun{
 write_R2BIN(t, file = tempfile(), version = "04")
}

}
\seealso{
\link{read_BIN2R}, \link{write_R2BIN}, \linkS4class{Risoe.BINfileData},
\link[base:readLines]{base::readLines}, \linkS4class{RLum.Analysis}, \link{list.files}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. read_BINXLOG2R(): Import Risø BINX-data from the BINX Log File. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_OSLAgeSummary.R
\name{plot_OSLAgeSummary}
\alias{plot_OSLAgeSummary}
\title{Plot Posterior OSL-Age Summary}
\usage{
plot_OSLAgeSummary(object, level = 0.95, digits = 1L, verbose = TRUE, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Results}, \link{numeric} (\strong{required}):
an object produced by \link{combine_De_Dr}. Alternatively, a \link{numeric} vector of
a parameter from an MCMC process.}

\item{level}{\link{numeric} (\emph{with default}):
probability of shown credible interval.}

\item{digits}{\link{integer} (\emph{with default}):
number of digits considered for the calculation.}

\item{verbose}{\link{logical} (\emph{with default}): enable/disable output to the
terminal.}

\item{...}{further arguments to modify the plot, supported: \code{xlim}, \code{ylim},
\code{xlab}, \code{ylab}, \code{main}, \code{lwd}, \code{lty}, \code{col}, \code{rug}, \code{polygon_col},
\code{polygon_density}.}
}
\value{
A posterior distribution plot and an \linkS4class{RLum.Results}
object with the credible interval.
}
\description{
The function produces a graphical summary of the statistical inference of
an OSL age.
}
\details{
The function is called automatically by \link{combine_De_Dr}.
}
\section{Function version}{
 0.1.0
}

\examples{
##generate random data
set.seed(1234)
object <- rnorm(1000, 100, 10)
plot_OSLAgeSummary(object)

}
\seealso{
\link{combine_De_Dr}, \link{plot.default}, \link[rjags:rjags-package]{rjags::rjags}
}
\author{
Anne Philippe, Université de Nantes (France)\cr
Jean-Michel Galharret, Université de Nantes (France)\cr
Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
, RLum Developer Team} 

\section{How to cite}{
Philippe, A., Galharret, J., Mercier, N., Kreutzer, S., 2026. plot_OSLAgeSummary(): Plot Posterior OSL-Age Summary. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{dplot}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/Risoe.BINfileData-class.R
\name{sort_RLum}
\alias{sort_RLum}
\alias{sort_RLum,list-method}
\alias{sort_RLum,RLum.Analysis-method}
\alias{sort_RLum,Risoe.BINfileData-method}
\title{Sort data for RLum-class and Risoe.BINfileData-class objects}
\usage{
sort_RLum(object, ...)

\S4method{sort_RLum}{list}(object, ...)

\S4method{sort_RLum}{RLum.Analysis}(object, slot = NULL, info_element = NULL, decreasing = FALSE, ...)

\S4method{sort_RLum}{Risoe.BINfileData}(object, info_element, decreasing = FALSE, ...)
}
\arguments{
\item{object}{\linkS4class{RLum} or
\linkS4class{Risoe.BINfileData} (\strong{required}):
S4 object of class \code{RLum.Analysis} or \code{Risoe.BINfileData}.}

\item{...}{further arguments passed to the specific class method}

\item{slot}{\link{character} (\emph{optional}): slot name to use in sorting.}

\item{info_element}{\link{character} (\emph{optional}): names of the \code{info} field
to use in sorting. The order of the names sets the sorting priority.
Regardless of available info elements, the following
elements always exist because they are calculated from the record
\code{XY_LENGTH}, \code{NCOL}, \code{X_MIN}, \code{X_MAX}, \code{Y_MIN}, \code{Y_MAX}.}

\item{decreasing}{\link{logical} (\emph{with default}): whether the sort order should
be decreasing (\code{FALSE} by default). It can be provided as a vector to control
the ordering sequence for each sorting element.}
}
\value{
An object of the same type as the input object provided.
}
\description{
The function provides a generalised access point for specific \linkS4class{RLum}
and \linkS4class{Risoe.BINfileData} objects. Depending on the input object,
the corresponding function will be selected.
}
\section{Functions}{
\itemize{
\item \code{sort_RLum(list)}: Returns a list of sorted \linkS4class{RLum} objects.

\item \code{sort_RLum(RLum.Analysis)}: Sorting of \code{RLum.Data} objects contained in this \code{RLum.Analysis} object.
At least one of \code{slot} and \code{info_element} must be provided. If both are
given, ordering by \code{slot} always takes priority over \code{info_element}.
Only the first element in each \code{slot} and each \code{info_element} is used
for sorting. Example: \code{.pid} can contain multiple values, however, only the
first is taken.
Please note that the \code{show()} method does some structuring, which may
lead to the impression that the sorting did not work.

\item \code{sort_RLum(Risoe.BINfileData)}: Sort method for \linkS4class{Risoe.BINfileData} objects.

}}
\section{Function version}{
 0.1.0
}

\examples{

## load example data
data(ExampleData.XSYG, envir = environment())
obj <- OSL.SARMeasurement$Sequence.Object[1:9]

sort_RLum(obj, slot = "recordType")
sort_RLum(obj, info_element = "curveDescripter")

data(ExampleData.XSYG, envir = environment())
sar <- OSL.SARMeasurement$Sequence.Object[1:5]
sort_RLum(sar, solt = "recordType", info_element = c("startDate"))

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{Risoe.BINfileData}
}
\author{
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Colombo, M., 2026. sort_RLum(): Sort data for RLum-class and Risoe.BINfileData-class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_Al2O3C_ITC.R
\name{analyse_Al2O3C_ITC}
\alias{analyse_Al2O3C_ITC}
\title{Al2O3 Irradiation Time Correction Analysis}
\usage{
analyse_Al2O3C_ITC(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  dose_points = c(2, 4, 8, 12, 16),
  recordType = "OSL (UVVIS)",
  method_control = NULL,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} or \link{list} (\strong{required}):
results obtained from the measurement.
Alternatively a list of \linkS4class{RLum.Analysis} objects can be provided to allow an automatic analysis}

\item{signal_integral}{\link{numeric} (\emph{optional}):
signal integral, used for the signal and the background.
If nothing is provided the full range is used. Argument can be provided as \link{list}.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{dose_points}{\link{numeric} (\emph{with default}):
vector with dose points, if dose points are repeated, only the general
pattern needs to be provided. Default values follow the suggestions
made by Kreutzer et al., 2018. Argument can be provided as \link{list}.}

\item{recordType}{\link{character} (\emph{with default}): input curve selection, which is passed to
function \link{get_RLum}. To deactivate the automatic selection set the argument to \code{NULL}}

\item{method_control}{\link{list} (\emph{optional}):
optional parameters to control the calculation.
See details for further explanations}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments that can be passed to the plot output}
}
\value{
Function returns results numerically and graphically:

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$data} \tab \code{data.frame} \tab correction value and error \cr
\verb{$table} \tab \code{data.frame} \tab table used for plotting  \cr
\verb{$table_mean} \tab \code{data.frame} \tab table used for fitting \cr
\verb{$fit} \tab \code{lm} or \code{nls} \tab the fitting as returned by the function \link{fit_DoseResponseCurve}
}

\strong{slot:} \strong{\verb{@info}}

The original function call

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr
\itemize{
\item A dose response curve with the marked correction values
}
}
\description{
The function provides a very particular analysis to correct the irradiation
time while irradiating Al2O3:C chips in a luminescence reader.
}
\details{
Background: Due to their high dose sensitivity Al2O3:C chips are usually
irradiated for only a very short duration or under the closed beta-source
within a luminescence reader. However, due to its high dose sensitivity, during
the movement towards the beta-source, the pellet already receives and non-negligible
dose. Based on measurements following a protocol suggested by Kreutzer et al., 2018,
a dose response curve is constructed and the intersection (absolute value) with the time axis
is taken as real irradiation time.

\strong{\code{method_control}}

To keep the generic argument list as clear as possible, arguments to allow a
deeper control of the method are all preset with meaningful default parameters and can be
handled using the argument \code{method_control} only, e.g.,
\code{method_control = list(fit.method = "LIN")}. Supported arguments are:

\tabular{lll}{
\strong{ARGUMENT} \tab \strong{FUNCTION} \tab \strong{DESCRIPTION}\cr
\code{mode} \tab \code{fit_DoseResponseCurve} \tab sets the mode used for fitting\cr
\code{fit.method} \tab \code{fit_DoseResponseCurve} \tab sets the function applied for fitting\cr
}
}
\section{Function version}{
 0.1.2
}

\examples{

##load data
data(ExampleData.Al2O3C, envir = environment())

##run analysis
analyse_Al2O3C_ITC(data_ITC)

} 

\section{How to cite}{
Kreutzer, S., 2026. analyse_Al2O3C_ITC(): Al2O3 Irradiation Time Correction Analysis. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018. Environmental Dose Rate
Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
Geochronometria 45, 56-67. doi: 10.1515/geochr-2015-0086
}
\seealso{
\link{fit_DoseResponseCurve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R,
%   R/RLum.Data.Image-class.R, R/RLum.Data.Spectrum-class.R,
%   R/RLum.Results-class.R
\name{get_RLum}
\alias{get_RLum}
\alias{get_RLum,list-method}
\alias{get_RLum,NULL-method}
\alias{get_RLum,RLum.Analysis-method}
\alias{get_RLum,RLum.Data.Curve-method}
\alias{get_RLum,RLum.Data.Image-method}
\alias{get_RLum,RLum.Data.Spectrum-method}
\alias{get_RLum,RLum.Results-method}
\title{General accessor function for RLum-class objects}
\usage{
get_RLum(object, ...)

\S4method{get_RLum}{list}(object, class = NULL, null.rm = FALSE, ...)

\S4method{get_RLum}{NULL}(object, ...)

\S4method{get_RLum}{RLum.Analysis}(
  object,
  record.id = NULL,
  recordType = NULL,
  curveType = NULL,
  RLum.type = NULL,
  protocol = "UNKNOWN",
  get.index = FALSE,
  drop = TRUE,
  recursive = TRUE,
  info.object = NULL,
  subset = NULL,
  env = parent.frame(2)
)

\S4method{get_RLum}{RLum.Data.Curve}(object, info.object = NULL)

\S4method{get_RLum}{RLum.Data.Image}(object, info.object = NULL)

\S4method{get_RLum}{RLum.Data.Spectrum}(object, info.object = NULL)

\S4method{get_RLum}{RLum.Results}(object, data.object, info.object = NULL, drop = TRUE)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum} or an object of type \link{list} containing only objects
of type \linkS4class{RLum}}

\item{...}{further arguments passed to the specific class method.}

\item{class}{\link{character} (\emph{optional}):
define which class gets selected if applied to a list, e.g., if a list
consists of different type of \linkS4class{RLum} objects, this arguments allows
to make a selection. If nothing is provided, all \linkS4class{RLum} objects are
treated.}

\item{null.rm}{\link{logical} (\emph{with default}):
whether empty and \code{NULL} objects should be removed.}

\item{record.id}{\link{numeric} or \link{logical} (\emph{optional}):
IDs of specific records. If of type \code{logical} the entire id range is assumed
and \code{TRUE} and \code{FALSE} indicates the selection.}

\item{recordType}{\link{character} (\emph{optional}):
record type (e.g., "OSL"). Can be also a vector, for multiple matching,
e.g., \code{recordType = c("OSL", "IRSL")}}

\item{curveType}{\link{character} (\emph{optional}):
curve type (e.g. "predefined" or "measured")}

\item{RLum.type}{\link{character} (\emph{optional}):
RLum object type. Defaults to "RLum.Data.Curve" and "RLum.Data.Spectrum".}

\item{protocol}{\link{character} (\emph{optional}):
currently ignored.}

\item{get.index}{\link{logical} (\emph{optional}):
return a numeric vector with the index of each element in the RLum.Analysis
object (\code{FALSE} by default).}

\item{drop}{\link{logical} (\emph{with default}):
coerce to the next possible layer (which are \linkS4class{RLum.Data} objects if
\code{object} is an \linkS4class{RLum.Analysis} object). If \code{drop = FALSE}, an object
of the same type as the input is returned.}

\item{recursive}{\link{logical} (\emph{with default}):
if \code{TRUE} (default) when the result of the \code{get_RLum()} request is a single
object, the object itself will be returned directly, rather than being
wrapped in a list. Mostly this makes things easier, but this might be
undesired if this method is used within a loop.}

\item{info.object}{\link{character} (\emph{optional}):
name of the wanted info element.}

\item{subset}{\link{expression} (\emph{optional}):
logical or character masking a logical expression indicating elements or rows to keep:
missing values are taken as false. This argument takes precedence over all other arguments,
meaning they are not considered when subsetting the object. \code{subset} works slots and
info elements.}

\item{env}{\link{environment} (\emph{with default}):
An environment passed to \link{eval} as the enclosure. This argument is only
relevant when subsetting the object and should not be used manually.}

\item{data.object}{\link{character} or \link{numeric}:
name or index of the data slot to be returned.}
}
\value{
An object of the same type as the input object provided.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the input object, the corresponding
function will be selected.
}
\section{Functions}{
\itemize{
\item \code{get_RLum(list)}: Returns a list of \linkS4class{RLum} objects that had been passed to
\link{get_RLum}

\item \code{get_RLum(`NULL`)}: Returns \code{NULL}.

\item \code{get_RLum(RLum.Analysis)}: Accessor method for RLum.Analysis objects.
The optional arguments \code{record.id}, \code{recordType}, \code{curveType} and \code{RLum.type}
allow to limit records by their id (list index number), their record type
(e.g. \code{recordType = "OSL"}), their curve type (e.g. \code{curveType = "predefined"}
or \code{curveType ="measured"}), or object type.

The selection of a specific RLum.type object superimposes the default selection.
Currently supported objects are: RLum.Data.Curve and RLum.Data.Spectrum

Returns:
\enumerate{
\item \link{list} of \linkS4class{RLum.Data} objects or
\item Single \linkS4class{RLum.Data} object, if only one object
is contained and \code{recursive = FALSE} or
\item \linkS4class{RLum.Analysis} objects for \code{drop = FALSE}
}

\item \code{get_RLum(RLum.Data.Curve)}: Accessor method for \linkS4class{RLum.Data.Curve} object.
The argument \code{info.object} is optional to directly access the info elements.
If no info element name is provided, the raw curve data (matrix) will be
returned.

\item \code{get_RLum(RLum.Data.Image)}: Accessor method for \linkS4class{RLum.Data.Image} objects.
The argument \code{info.object} is optional to directly access the info elements.
If no info element name is provided, the raw image data (\code{array}) will be
returned.

\item \code{get_RLum(RLum.Data.Spectrum)}: Accessor method for \linkS4class{RLum.Data.Spectrum} objects.
The argument \code{info.object} is optional to directly access the info elements.
If no info element name is provided, the raw curve data (matrix) will be
returned.

\item \code{get_RLum(RLum.Results)}: Accessor method for \linkS4class{RLum.Results} object.
The argument \code{data.object} allows to access directly objects stored
within the slot data. The default return object depends on the object
originator (e.g., \code{fit_LMCurve}). If nothing is specified always the first
\code{data.object} will be returned.

Note: Detailed specification should be made in combination with the
originator slot in the receiving function if results are piped.

Returns:
\enumerate{
\item Data object from the specified slot
\item \link{list} of data objects from the slots if 'data.object' is vector or
\item an \linkS4class{RLum.Results} for \code{drop = FALSE}.
}

}}
\section{Function version}{
 0.3.3
}

\examples{

## Example based using data and from the calc_CentralDose() function

## load example data
data(ExampleData.DeValues, envir = environment())

## apply the central dose model 1st time
temp1 <- calc_CentralDose(ExampleData.DeValues$CA1)

## get results and store them in a new object
temp.get <- get_RLum(object = temp1)

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Data.Image},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. get_RLum(): General accessor function for RLum-class objects. Function version 0.3.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{BaseDataSet.ConversionFactors}
\alias{BaseDataSet.ConversionFactors}
\title{Base data set of dose-rate conversion factors}
\format{
A \code{\link{list}} with three elements with dose-rate conversion factors
sorted by article and radiation type (alpha, beta, gamma):

\tabular{ll}{

\code{AdamiecAitken1998}: \tab
Conversion factors from Tables 5 and 6 \cr

\code{Cresswelletal2018}: \tab
Conversion factors from Tables 5 and 6 \cr

\code{Guerinetal2011}: \tab
Conversion factors from Tables 1, 2 and 3 \cr

\code{Liritzisetal2013}: \tab
Conversion factors from Tables 1, 2 and 3 \cr
}
}
\source{
All gamma conversion factors were carefully read from the tables given in the
references above.
}
\description{
Collection of published dose-rate conversion factors to convert concentrations
of radioactive isotopes to dose rate values.
}
\section{Version}{
 0.2.0
}

\examples{

## Load data
data("BaseDataSet.ConversionFactors", envir = environment())

}
\references{
Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update.
Ancient TL 16, 37-46.

Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018.
Dose rate conversion parameters: Assessment of nuclear data.
Radiation Measurements 120, 195-201.

Guerin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion
factors: update. Ancient TL, 29, 5-8.

Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
A re-evaluation of radiation dose-rate conversion factors. Mediterranean
Archaeology and Archaeometry 13, 1-15.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_XSYG2CSV.R
\name{convert_XSYG2CSV}
\alias{convert_XSYG2CSV}
\title{Export XSYG-file(s) to CSV-files}
\usage{
convert_XSYG2CSV(file, ...)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of the XSYG-file to be converted to CSV-files}

\item{...}{further arguments that will be passed to the function
\link{read_XSYG2R} and \link{write_RLum2CSV}}
}
\value{
The function returns either a CSV-file (or many of them) or for the option \code{export = FALSE}
a list comprising objects of type \link{data.frame} and \link{matrix}
}
\description{
This function is a wrapper function around the functions \link{read_XSYG2R} and
\link{write_RLum2CSV} and it imports an XSYG-file and directly exports its content
to CSV-files. If nothing is set for the argument \code{path} (\link{write_RLum2CSV})
the input folder will become the output folder.
}
\section{Function version}{
 0.1.1
}

\examples{

##transform XSYG-file values to a list
data(ExampleData.XSYG, envir = environment())
convert_XSYG2CSV(OSL.SARMeasurement$Sequence.Object[1:10], export = FALSE)

\dontrun{
##select your BIN-file
file <- file.choose()

##convert
convert_XSYG2CSV(file)

}

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data}, \linkS4class{RLum.Results},
\link[utils:write.table]{utils::write.table}, \link{write_RLum2CSV}, \link{read_XSYG2R}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. convert_XSYG2CSV(): Export XSYG-file(s) to CSV-files. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_SPE2R.R
\name{read_SPE2R}
\alias{read_SPE2R}
\title{Import Princeton Instruments (TM) SPE-file into R}
\usage{
read_SPE2R(
  file,
  output.object = c("RLum.Data.Image", "RLum.Data.Spectrum", "matrix"),
  frame.range = NULL,
  txtProgressBar = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of the SPE file to read (URLs are supported).}

\item{output.object}{\link{character} (\emph{with default}):
set the output object type. Allowed types are \linkS4class{RLum.Data.Spectrum},
\linkS4class{RLum.Data.Image} or \code{"matrix"}.}

\item{frame.range}{\link{vector}, \link{integer} (\emph{optional}):
range of frames to read. For example, \code{frame.range = c(1, 10)} selects only
the first 10 frames. If not specified, all available frames (up to a maximum
of 100 if \code{output.object = "RLum.Data.Image"}) are read.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar. Ignored if \code{verbose = FALSE}.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{not used, for compatibility reasons only.}
}
\value{
Depending on the chosen option the functions returns three different
type of objects:

\code{output.object}

\code{RLum.Data.Spectrum}

An object of type \linkS4class{RLum.Data.Spectrum} is returned.  Row
sums are used to integrate all counts over one channel.

\code{RLum.Data.Image}

An object of type \linkS4class{RLum.Data.Image} is returned.  Due to
performance reasons the import is aborted for files containing more than 100
frames. This limitation can be overwritten manually by using the argument
\code{frame.range}.

\code{matrix}

Returns a matrix of the form: Rows = Channels, columns = Frames. For the
transformation the function \link{get_RLum} is used,
meaning that the same results can be obtained by using the function
\link{get_RLum} on an \linkS4class{RLum.Data.Spectrum} or
\linkS4class{RLum.Data.Image} object.
}
\description{
The function imports Princeton Instruments (TM) SPE-files into R and
provides \linkS4class{RLum.Data.Image} objects as output. The import
functionality is based on the file format description provided by
Princeton Instruments and a MatLab script written by Carl Hall (see
references).
}
\note{
\strong{The function does not test whether the input data are spectra or pictures for spatial resolved analysis!}

The function has been successfully tested for SPE format versions 2.x.

\emph{Currently not all information provided by the SPE format are supported.}
}
\section{Function version}{
 0.1.6
}

\examples{

## to run examples uncomment lines and run the code

##(1) Import data as RLum.Data.Spectrum object
#file <- file.choose()
#temp <- read_SPE2R(file)
#temp

##(2) Import data as RLum.Data.Image object
#file <- file.choose()
#temp <- read_SPE2R(file, output.object = "RLum.Data.Image")
#temp

##(3) Import data as matrix object
#file <- file.choose()
#temp <- read_SPE2R(file, output.object = "matrix")
#temp

##(4) Export raw data to csv, if temp is a RLum.Data.Spectrum object
# write.table(x = get_RLum(temp),
#             file = "[your path and filename]",
#             sep = ";", row.names = FALSE)

} 

\section{How to cite}{
Kreutzer, S., 2026. read_SPE2R(): Import Princeton Instruments (TM) SPE-file into R. Function version 0.1.6. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Princeton Instruments, 2014. Princeton Instruments SPE 3.0 File
Format Specification, Version 1.A (for document URL please use an internet search machine)

Hall, C., 2012: readSPE.m.
\verb{https://www.mathworks.com/matlabcentral/fileexchange/35940-readspe}
}
\seealso{
\link{readBin}, \linkS4class{RLum.Data.Spectrum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_SAR.CWOSL.R
\name{analyse_SAR.CWOSL}
\alias{analyse_SAR.CWOSL}
\title{Analyse SAR CW-OSL Measurements}
\usage{
analyse_SAR.CWOSL(
  object,
  signal_integral = NA,
  background_integral = NA,
  signal_integral_Tx = NULL,
  background_integral_Tx = NULL,
  integral_input = c("channel", "measurement"),
  OSL.component = NULL,
  rejection.criteria = list(),
  dose.points = NULL,
  dose.points.test = NULL,
  dose_rate_source = NULL,
  trim_channels = FALSE,
  mtext.outer = "",
  plot = TRUE,
  plot_onePage = FALSE,
  plot_singlePanels = FALSE,
  onlyLxTxTable = FALSE,
  method_control = list(),
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}):
input object containing data for analysis, alternatively a \link{list} of
\linkS4class{RLum.Analysis} objects can be provided. The object
should \strong{only} contain curves considered part of the SAR protocol (see
Details). When \code{OSL.component} is set, the input object must have been
processed by \code{OSLdecomposition::RLum.OSL_decomposition()}.}

\item{signal_integral}{\link{integer} (\strong{required}):
vector of channels for the signal integral. It can be a \link{list} of integers,
if \code{object} is a list. If set to \code{NA}, no integrals are taken into account
and their settings are ignored.}

\item{background_integral}{\link{integer} (\strong{required}):
vector of channels for the background integral. It can be a \link{list} of
integers, if \code{object} is a list. If set to \code{NA}, no background integral is
subtracted.}

\item{signal_integral_Tx}{\link{integer} (\emph{optional}):
vector of channels for the signal integral for the \code{Tx} curve. It can be a
\link{list} of integers, if \code{object} is a list. If \code{NULL}, the \code{signal_integral}
vector is used also for the \code{Tx} curve.}

\item{background_integral_Tx}{\link{integer} (\emph{optional}):
vector of channels for the background integral for the \code{Tx} curve. It can
be a \link{list} of integers, if \code{object} is a list. If \code{NULL}, the
\code{background_integral} vector is used. If set to \code{NA}, no background integral
for the \code{Tx} curve is subtracted.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{OSL.component}{\link{character} or \link{integer} (\emph{optional}):
single index or a \link{character} defining the signal component to be evaluated.
It requires that the object was processed by \code{OSLdecomposition::RLum.OSL_decomposition}.
This argument can either be the name of the OSL component assigned by
\code{OSLdecomposition::RLum.OSL_global_fitting} or the index in the descending
order of decay rates. Then \code{"1"} selects the fastest decaying component, \code{"2"}
the second fastest and so on. Can be a \link{list} of \link{integer}s or strings (or mixed)
If object is a \link{list} and this parameter is provided as \link{list} it alternates over
the elements (aliquots) of the object list, e.g., \code{list(1,2)} processes the first
aliquot with component \code{1} and the second aliquot with component \code{2}.
\code{NULL} does not process any component.}

\item{rejection.criteria}{\link{list} (\emph{with default}):
provide a \emph{named} list and set rejection criteria in \strong{percentage}. It can
be a nested \link{list}, if \code{object} is of type \link{list}.
Note: \emph{unnamed} list elements are ignored.

Allowed options:
\itemize{
\item \code{recycling.ratio} \link{numeric} (default: \code{10})
\item \code{recuperation.rate} \link{numeric} (default: \code{10})
\item \code{palaeodose.error} \link{numeric} (default: \code{10})
\item \code{testdose.error} \link{numeric} (default: \code{10})
\item \code{sn.ratio} \link{numeric} (default: \code{NA})
\item \code{exceed.max.regpoint} \link{logical} (default: \code{FALSE})
\item \code{consider.uncertainties} \link{logical} (default: \code{FALSE})
\item \code{recuperation_reference} \link{character} (default: \code{"Natural"}; set to, e.g.,
\code{"R1"} for other point)
\item \code{sn_reference} \link{character} (default: \code{"Natural"}).
}

Example: \code{rejection.criteria = list(recycling.ratio = 10)}.

All numerical criteria can be set to \code{NA}, in which case values are
calculated, but they are not considered, and their corresponding RC.Status
is always \code{"OK"}. In the "Checks" plot, they are shown with a grey circle
and only their value is reported (without showing \verb{<= NA} or \verb{>= NA}).

If \code{onlyLxTxTable = TRUE}, the \code{palaeodose.error} and \code{exceed.max.regpoint}
criteria are not computed.}

\item{dose.points}{\link{numeric} (\emph{optional}):
a numeric vector containing the dose points values. Using this argument
overwrites dose point values extracted from other data. Can be a \link{list} of
\link{numeric} vectors, if \code{object} is of type \link{list}.}

\item{dose.points.test}{\link{numeric} (\emph{optional}):
a numeric vector containing the test dose in the same units as \code{dose.points}.
If length = 1, the values will be recycled. It has only an effect for
\code{fit.method = 'OTORX'}.}

\item{dose_rate_source}{\link{numeric} (\emph{optional}): a numerical value for the source dose rate,
typically Gy/s. If set, the x-axis default for the dose-response curve changes to \code{Dose [Gy]}.}

\item{trim_channels}{\link{logical} (\emph{with default}):
trim channels per record category to the lowest number of channels in the
category by using \link{trim_RLum.Data}. Applies only to \code{OSL} and \code{IRSL} curves.
For a more granular control use \link{trim_RLum.Data} before calling this
function.}

\item{mtext.outer}{\link{character} (\emph{optional}):
option to provide an outer margin \code{mtext}. Can be a \link{list} of \link{character}s,
if \code{object} is of type \link{list}}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{plot_onePage}{\link{logical} (\emph{with default}): enable/disable one page
plot output.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}) or \link{numeric} (\emph{optional}):
single plot output (\code{TRUE/FALSE}) to allow for plotting the results in single plot windows.
If a \link{numeric} vector is provided the plots can be selected individually, i.e.
\code{plot_singlePanels = c(1,2,3,4)} will plot the TL and Lx, Tx curves but
not the legend (5) or the
growth curve (6), (7) and (8) belong to rejection criteria plots. Requires
\code{plot = TRUE}.}

\item{onlyLxTxTable}{\link{logical} (\emph{with default}):
If \code{TRUE} the dose response curve fitting and plotting is skipped, and the
\code{palaeodose.error} and \code{exceed.max.regpoint} criteria are not computed.
This allows to get hands on the \code{Lx/Tx} table for large datasets
without the need for a curve fitting.}

\item{method_control}{\link{list} (\emph{optional}):
options to control the function behaviour. Currently only the
'auto_curve_removal' (logical) option is supported, which controls whether
curves with \code{recordType} starting with \verb{_} should be automatically removed
(\code{TRUE} by default).}

\item{...}{further arguments that will be passed to
\link{fit_DoseResponseCurve}, \link{plot_DoseResponseCurve}
or \link{calc_OSLLxTxRatio} (supported:
\code{background.count.distribution}, \code{sigmab}, \code{sig0}).
\strong{Note:} if you consider to use the early light subtraction method,
\code{sigmab} should be provided.

Additionally, supported are \code{legend.cex} and \code{legend.pch} to modify the
the legend symbols.}
}
\value{
A plot (\emph{optional}) and an \linkS4class{RLum.Results} object is
returned containing the following elements:

\item{data}{\link{data.frame} containing De-values, De-error and further parameters}
\item{LnLxTnTx.values}{\link{data.frame} of all calculated Lx/Tx values including signal,
background counts and the dose points}
\item{rejection.criteria}{\link{data.frame} with values that might by used as rejection criteria.
\code{NA} is produced if no R0 dose point exists.}
\item{Formula}{\link{formula} formula that have been used for the growth curve fitting}

The output should be accessed using the function \link{get_RLum}.

\strong{The function currently does support only 'OSL', 'IRSL' and 'POSL' data!}
}
\description{
The function performs a SAR CW-OSL analysis on an \linkS4class{RLum.Analysis}
object, including growth curve fitting.
}
\details{
The function performs an analysis for a standard SAR protocol measurements
introduced by Murray and Wintle (2000) with CW-OSL curves. For the
calculation of the \code{Lx/Tx} value the function \link{calc_OSLLxTxRatio} is
used. To \strong{change the way the Lx/Tx error is calculated} use arguments
\code{background.count.distribution} and \code{sigmab}, which will be passed to
\link{calc_OSLLxTxRatio}.

\strong{What is part of a SAR sequence?}

The function is rather picky when it comes down to accepted curve input
(OSL, IRSL,...) and structure. A SAR sequence is basically a set of
\eqn{L_{x}/T_{x}} curves. Hence, every second curve is considered a
shine-down curve related to the test dose. It also means that the number of
curves for \eqn{L_{x}} has to be equal to the number of \eqn{T_{x}} curves,
and that hot-bleach curves \strong{do not} belong into a SAR sequence; at least
not for the analysis. Other curves allowed and processed are preheat curves,
or preheat curves measured as TL, and irradiation curves. The later one
indicates the duration of the irradiation, the dose and test dose points,
e.g., as part of XSYG files.

\strong{Argument \code{object} is of type \code{list}}

If the argument \code{object} is of type \link{list} containing \strong{only}
\linkS4class{RLum.Analysis} objects, the function re-calls itself on each element
in the list. This is useful if to analyse an entire measurement without
writing separate for-loops. To gain in full control of the parameters (e.g., \code{dose.points}) for
every aliquot (corresponding to one \linkS4class{RLum.Analysis} object in the list), in
this case the arguments can be provided as \link{list}. This \code{list} should
be of similar length as the \code{list} provided with the argument \code{object},
otherwise the function will create an own list of the requested length.
Function output will be just one single \linkS4class{RLum.Results} object.

Please be careful when using this option. While it may allow for a fast and
efficient data analysis, the function may break with an unclear error
message if the input data is misspecified.

\strong{Working with IRSL data}

The function was originally designed to work just for 'OSL' curves,
following the principles of the SAR protocol. An IRSL measurement protocol
may follow this procedure, e.g., post-IR IRSL protocol (Thomsen et al.,
2008). Therefore this functions has been enhanced to work with IRSL data,
however, the function is only capable of analysing curves that follow the
SAR protocol structure, i.e., to analyse a post-IR IRSL protocol, curve data
have to be pre-selected by the user to fit the standards of the SAR
protocol, i.e., Lx,Tx,Lx,Tx and so on.

Example: Imagine the measurement contains \code{pIRIR50} and \code{pIRIR225} IRSL
curves. Only one curve type can be analysed at the same time: either the
\code{pIRIR50} curves or the \code{pIRIR225} curves.

\strong{Supported rejection criteria}

\verb{[recycling.ratio]}: calculated for every repeated regeneration dose point.

\verb{[recuperation.rate]}: recuperation rate calculated by comparing the
\code{Lx/Tx} values of the zero regeneration point with the \code{Ln/Tn} value (the
\code{Lx/Tx} ratio of the natural signal). For methodological background see
Aitken and Smith (1988). As a variant, \code{recuperation_reference} can be
specified to select another dose point as reference instead of \code{Ln/Tn}.

\verb{[testdose.error]}: set the allowed error for the test dose, which by
default should not exceed 10\%. The test dose error is calculated as
\code{Tx_net.error/Tx_net}. The calculation of the \eqn{T_{n}} error is detailed
in \link{calc_OSLLxTxRatio}.

\verb{[palaeodose.error]}: set the allowed error for the De value, which per
default should not exceed 10\%.

\verb{[sn.ratio]}: set the allowed signal/noise ratio, which by default should
be at least 50. By default it uses the value from the natural curve, but
this can be changed by specifying the \code{sn_reference} option.

By default, the computed values are compared directly to the corresponding
thresholds to establish their result status ("OK" or "FAILED"). By setting
the option \code{consider.uncertainties = TRUE} in the \code{rejection.criteria}
list, quantified uncertainties are considered into the computation of the
test value prior before comparing it to the threshold(currently supported
only for \code{recycling.ratio}, \code{recuperation.rate} and \code{exceed.max.regpoint}).
This reduces tests being marked as "FAILED" when the deviation from the
threshold is smaller than the uncertainty margin.

\strong{Irradiation times}

The function makes two attempts to extra irradiation data (dose points)
automatically from the input object, if the argument \code{dose.points} is not
set (aka set to \code{NULL}).
\enumerate{
\item It searches in every curve for an info object called \code{IRR_TIME}. If this
is found, any value set there is taken as dose point.
\item If the object contains curves of type \code{irradiation}, the function tries
to use this information to assign these values to the curves. However, the
function does \strong{not} overwrite values preset in \code{IRR_TIME}.
}
}
\section{Function version}{
 0.13.8
}

\examples{

##load data
##ExampleData.BINfileData contains two BINfileData objects
##CWOSL.SAR.Data and TL.SAR.Data
data(ExampleData.BINfileData, envir = environment())

##transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

##perform SAR analysis and set rejection criteria
results <- analyse_SAR.CWOSL(
object = object,
signal_integral = 1:2,
background_integral = 900:1000,
log = "x",
fit.method = "EXP",
plot_onePage = TRUE,
rejection.criteria = list(
  recycling.ratio = 10,
  recuperation.rate = 10,
  testdose.error = 10,
  palaeodose.error = 10,
  recuperation_reference = "Natural",
  sn.ratio = 50,
  sn_reference = "Natural",
  exceed.max.regpoint = TRUE)
)

##show De results
get_RLum(results)

##show LnTnLxTx table
get_RLum(results, data.object = "LnLxTnTx.table")

## Run example with special case for
## the OTORX fit
\dontrun{
results <- analyse_SAR.CWOSL(
 object = object,
 signal_integral = 1:2,
 background_integral = 900:1000,
 dose.points.test = 15,
 n.MC = 10,
 fit.method = "OTORX")
}

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. analyse_SAR.CWOSL(): Analyse SAR CW-OSL Measurements. Function version 0.13.8. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Aitken, M.J. and Smith, B.W., 1988. Optical dating: recuperation
after bleaching. Quaternary Science Reviews 7, 387-393.

Duller, G., 2003. Distinguishing quartz and feldspar in single grain
luminescence measurements. Radiation Measurements 37 (2), 161-165.

Murray, A.S. and Wintle, A.G., 2000. Luminescence dating of quartz using an
improved single-aliquot regenerative-dose protocol. Radiation Measurements
32, 57-73.

Thomsen, K.J., Murray, A.S., Jain, M., Boetter-Jensen, L., 2008. Laboratory
fading rates of various luminescence signals from feldspar-rich sediment
extracts. Radiation Measurements 43, 1474-1486.
doi:10.1016/j.radmeas.2008.06.002
}
\seealso{
\link{calc_OSLLxTxRatio}, \link{fit_DoseResponseCurve},
\link{plot_DoseResponseCurve}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}
\keyword{datagen}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_CommonDose.R
\name{calc_CommonDose}
\alias{calc_CommonDose}
\title{Apply the (un-)logged common age model after Galbraith et al. (1999) to a
given De distribution}
\usage{
calc_CommonDose(data, sigmab = 0, log = TRUE, ...)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with De \code{(data[,1])} and De error \code{(data[,2])}}

\item{sigmab}{\link{numeric} (\emph{with default}):
additional spread in De values, representing the expected overdispersion in
the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
\strong{Note:} For the logged model (\code{log = TRUE}) this value must be
a fraction, e.g. 0.2 (= 20\%). If the un-logged model is used (\code{log = FALSE}),
\code{sigmab} must be provided in the same absolute units of the De values
(seconds or Gray).}

\item{log}{\link{logical} (\emph{with default}):
fit the (un-)logged central age model to De data}

\item{...}{currently not used.}
}
\value{
Returns a terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following element:

\item{$summary}{\link{data.frame} summary of all relevant model results.}
\item{$data}{\link{data.frame} original input data}
\item{$args}{\link{list} used arguments}
\item{$call}{\link{call} the function call}

The output should be accessed using the function \link{get_RLum}.
}
\description{
Function to calculate the common dose of a De distribution.
}
\details{
\strong{(Un-)logged model}

When \code{log = TRUE} this function
calculates the weighted mean of logarithmic De values. Each of the estimates
is weighted by the inverse square of its relative standard error. The
weighted mean is then transformed back to the dose scale (Galbraith &
Roberts 2012, p. 14).

The log transformation is not applicable if the
De estimates are close to zero or negative. In this case the un-logged model
can be applied instead (\code{log = FALSE}). The weighted mean is then
calculated using the un-logged estimates of De and their absolute standard
error (Galbraith & Roberts 2012, p. 14).
}
\section{Function version}{
 0.1.1
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())

## apply the common dose model
calc_CommonDose(ExampleData.DeValues$CA1)

} 

\section{How to cite}{
Burow, C., 2026. calc_CommonDose(): Apply the (un-)logged common age model after Galbraith et al. (1999) to a given De distribution. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for
mixed fission track ages. Nuclear Tracks Radiation Measurements 4, 459-470.

Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley,
J.M., 1999. Optical dating of single grains of quartz from Jinmium rock
shelter, northern Australia. Part I: experimental design and statistical
models.  Archaeometry 41, 339-364.

Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error calculation and
display in OSL dating: An overview and some recommendations. Quaternary
Geochronology 11, 1-27.

\strong{Further reading}

Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
(De) distributions: Implications for OSL dating of sediment mixtures.
Quaternary Geochronology 4, 204-230.

Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
assessment of procedures for estimating burial dose. Quaternary Science
Reviews 25, 2475-2502.

Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
Quaternary Geochronology 12, 98-106.

Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
of optical dating of fluvial deposits.  Quaternary Geochronology 1 109-120.

Rodnight, H., 2008. How many equivalent dose values are needed to
obtain a reproducible distribution?. Ancient TL 26, 3-10.
}
\seealso{
\link{calc_CentralDose}, \link{calc_FiniteMixture},
\link{calc_FuchsLang2001}, \link{calc_MinDose}
}
\author{
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_BIN2CSV.R
\name{convert_BIN2CSV}
\alias{convert_BIN2CSV}
\title{Export Risoe BIN-file(s) to CSV-files}
\usage{
convert_BIN2CSV(file, ...)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of the BIN-file to be converted to CSV-files}

\item{...}{further arguments that will be passed to the function
\link{read_BIN2R} and \link{write_RLum2CSV}}
}
\value{
The function returns either a CSV-file (or many of them) or for the
option \code{export == FALSE} a list comprising objects of type \link{data.frame} and \link{matrix}
}
\description{
This function is a wrapper function around the functions \link{read_BIN2R} and
\link{write_RLum2CSV} and it imports a Risoe BIN-file and directly exports its
content to CSV-files. If nothing is set for the argument \code{path}
(\link{write_RLum2CSV}) the input folder will become the output folder.
}
\section{Function version}{
 0.1.0
}

\examples{

##transform Risoe.BINfileData values to a list
data(ExampleData.BINfileData, envir = environment())
convert_BIN2CSV(subset(CWOSL.SAR.Data, POSITION == 1), export = FALSE)

\dontrun{
##select your BIN-file
file <- file.choose()

##convert
convert_BIN2CSV(file)

}

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data}, \linkS4class{RLum.Results},
\link[utils:write.table]{utils::write.table}, \link{write_RLum2CSV}, \link{read_BIN2R}
}
\author{
Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2023. convert_BIN2CSV(): Export Risoe BIN-file(s) to CSV-files. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., 2023. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 0.9.23. https://CRAN.R-project.org/package=Luminescence
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Data.Spectrum-class.R
\docType{class}
\name{RLum.Data.Spectrum-class}
\alias{RLum.Data.Spectrum-class}
\title{Class \code{"RLum.Data.Spectrum"}}
\description{
Class for representing luminescence spectra data (TL/OSL/RF).
}
\section{Slots}{

\describe{
\item{\code{recordType}}{Object of class \link{character} containing the type of the curve (e.g. "TL" or "OSL")}

\item{\code{curveType}}{Object of class \link{character} containing curve type, allowed values are
"measured" or "predefined"}

\item{\code{data}}{Object of class \link{matrix} containing spectrum (count) values.
Row labels indicate wavelength/pixel values, column labels are temperature or time values.}

\item{\code{info}}{Object of class \link{list} containing further meta information objects}
}}

\note{
The class should only contain data for a single spectra data set. For
additional elements the slot \code{info} can be used. Objects from this class are automatically
created by, e.g., \link{read_XSYG2R}
}
\section{Objects from the Class}{

Objects can be created by calls of the form \code{set_RLum("RLum.Data.Spectrum", ...)}.
}

\section{Class version}{
 0.5.2
}

\examples{

showClass("RLum.Data.Spectrum")

##show example data
data(ExampleData.XSYG, envir = environment())
TL.Spectrum

##show data matrix
get_RLum(TL.Spectrum)

##plot spectrum
\dontrun{
plot_RLum(TL.Spectrum)
}

}
\seealso{
\linkS4class{RLum}, \linkS4class{RLum.Data}, \link{plot_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_CW2pHMi.R
\name{convert_CW2pHMi}
\alias{convert_CW2pHMi}
\title{Transform a CW-OSL curve into a pHM-OSL curve via interpolation under
hyperbolic modulation conditions}
\usage{
convert_CW2pHMi(object, delta = NULL, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
\linkS4class{RLum.Data.Curve} object or a \code{data.frame} with measured
curve data of type stimulation time (t) (\code{object[, 1]}) and measured counts
(cts) (\code{object[, 2]}).}

\item{delta}{\link{vector} (\emph{optional}):
stimulation rate parameter, if no value is given, the optimal value is
estimated automatically (see details). Smaller values of delta produce more
points in the rising tail of
the curve.}

\item{...}{currently not used.}
}
\value{
The function returns the same data type as the input data type with
the transformed curve values.

\strong{\code{RLum.Data.Curve}}

\tabular{ll}{
\verb{$CW2pHMi.x.t} \tab: transformed time values \cr
\verb{$CW2pHMi.method} \tab: used method for the production of the new data points
}

\strong{\code{data.frame}}

\tabular{ll}{
\verb{$x} \tab: time\cr
\verb{$y.t} \tab: transformed count values\cr
\verb{$x.t} \tab: transformed time values \cr
\verb{$method} \tab: used method for the production of the new data points
}
}
\description{
This function transforms a conventionally measured continuous-wave (CW)
OSL-curve to a pseudo hyperbolic modulated (pHM) curve under hyperbolic
modulation conditions using the interpolation procedure described by Bos &
Wallinga (2012).
}
\details{
The complete procedure of the transformation is described in Bos & Wallinga
(2012). The input \code{data.frame} consists of two columns: time (t) and
count values (CW(t))

\strong{Internal transformation steps}

(1) log(CW-OSL) values

(2)
Calculate t' which is the transformed time:
\deqn{t' = t-(1/\delta) * \log(1 + \delta t)}

(3)
Interpolate CW(t'), i.e. use the log(CW(t)) to obtain the count values
for the transformed time (t'). Values beyond \code{min(t)} and \code{max(t)}
produce \code{NA} values.

(4)
Select all values for t' < \code{min(t)}, i.e. values beyond the time
resolution of t. Select the first two values of the transformed data set
which contain no \code{NA} values and use these values for a linear fit
using \link{lm}.

(5)
Extrapolate values for t' < \code{min(t)} based on the previously
obtained fit parameters.

(6)
Transform values using
\deqn{pHM(t) = (\delta t / (1 + \delta t))*c*CW(t')}
\deqn{c = (1 + \delta P)/\delta*P}
\deqn{P = length(stimulation~period)}

(7) Combine all values and truncate all values for t' > \code{max(t)}

\strong{Note:}
The number of values for t' < \code{min(t)} depends on the stimulation rate
parameter \code{delta}. To avoid the production of too many artificial data
at the raising tail of the determined pHM curve, it is recommended to use
the automatic estimation routine for \code{delta}, i.e. provide no value for
\code{delta}.
}
\note{
According to Bos & Wallinga (2012), the number of extrapolated points
should be limited to avoid artificial intensity data. If \code{delta} is
provided manually and more than two points are extrapolated, a warning
message is returned.

The function \link{approx} may produce some \code{Inf} and \code{NaN} data, in which case
the function tries to interpolate these values by calculating the mean of
the adjacent channels. If two invalid values are succeeding,
the values are removed and no further interpolation is attempted.
In every case a warning message is shown.
}
\section{Function version}{
 0.2.7
}

\examples{

##(1) - simple transformation

##load CW-OSL curve data
data(ExampleData.CW_OSL_Curve, envir = environment())

##transform values
values.transformed <- convert_CW2pHMi(ExampleData.CW_OSL_Curve)

##plot
plot(values.transformed$x, values.transformed$y.t, log = "x")

##(2) - load CW-OSL curve from BIN-file and plot transformed values

##load BINfile
#BINfileData<-readBIN2R("[path to BIN-file]")
data(ExampleData.BINfileData, envir = environment())

##grep first CW-OSL curve from ALQ 1
curve.ID<-CWOSL.SAR.Data@METADATA[CWOSL.SAR.Data@METADATA[,"LTYPE"]=="OSL" &
                                    CWOSL.SAR.Data@METADATA[,"POSITION"]==1
                                  ,"ID"]

curve.HIGH<-CWOSL.SAR.Data@METADATA[CWOSL.SAR.Data@METADATA[,"ID"]==curve.ID[1]
                                    ,"HIGH"]

curve.NPOINTS<-CWOSL.SAR.Data@METADATA[CWOSL.SAR.Data@METADATA[,"ID"]==curve.ID[1]
                                       ,"NPOINTS"]

##combine curve to data set
curve<-data.frame(x = seq(curve.HIGH/curve.NPOINTS,curve.HIGH,
                          by = curve.HIGH/curve.NPOINTS),
                  y=unlist(CWOSL.SAR.Data@DATA[curve.ID[1]]))

##transform values
curve.transformed <- convert_CW2pHMi(curve)

##plot curve
plot(curve.transformed$x, curve.transformed$y.t, log = "x")

##(3) - produce Fig. 4 from Bos & Wallinga (2012)

##load data
data(ExampleData.CW_OSL_Curve, envir = environment())
values <- CW_Curve.BosWallinga2012

##open plot area
plot(NA, NA,
     xlim=c(0.001,10),
     ylim=c(0,8000),
     ylab="pseudo OSL (cts/0.01 s)",
     xlab="t [s]",
     log="x",
     main="Fig. 4 - Bos & Wallinga (2012)")

values.t <- convert_CW2pLMi(values, P = 1/20)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col="red" ,lwd=1.3)
text(0.03,4500,"LM", col="red" ,cex=.8)

values.t <- convert_CW2pHMi(values, delta = 40)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col="black", lwd=1.3)
text(0.005,3000,"HM", cex=.8)

values.t <- convert_CW2pPMi(values, P = 1/10)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col="blue", lwd=1.3)
text(0.5,6500,"PM", col="blue" ,cex=.8)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. convert_CW2pHMi(): Transform a CW-OSL curve into a pHM-OSL curve via interpolation under hyperbolic modulation conditions. Function version 0.2.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL
signal components. Radiation Measurements 47, 752-758.\cr

\strong{Further Reading}

Bulur, E., 1996. An Alternative Technique For
Optically Stimulated Luminescence (OSL) Experiment. Radiation Measurements
26, 701-709.

Bulur, E., 2000. A simple transformation for converting CW-OSL curves to
LM-OSL curves. Radiation Measurements 32, 141-145.
}
\seealso{
\link{convert_CW2pLM}, \link{convert_CW2pLMi}, \link{convert_CW2pPMi},
\link{fit_LMCurve}, \link{lm}, \linkS4class{RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
Based on comments and suggestions from:
Adrie J.J. Bos, Delft University of Technology, The Netherlands
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_CW2pLM.R
\name{convert_CW2pLM}
\alias{convert_CW2pLM}
\title{Transform a CW-OSL curve into a pLM-OSL curve}
\usage{
convert_CW2pLM(object, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
\linkS4class{RLum.Data.Curve} object or a \code{data.frame} with measured
curve data of type stimulation time (t) (\code{object[, 1]}) and measured counts
(cts) (\code{object[, 2]}).}

\item{...}{currently not used.}
}
\value{
The function returns the same data type as the input data type with
the transformed curve values (\link{data.frame} or \linkS4class{RLum.Data.Curve}).
}
\description{
Transforms a conventionally measured continuous-wave (CW) curve into a
pseudo linearly modulated (pLM) curve using the equations given in Bulur
(2000).
}
\details{
According to Bulur (2000) the curve data are transformed by introducing two
new parameters \code{P} (stimulation period) and \code{u} (transformed time):

\deqn{P=2\max(t)} \deqn{u=\sqrt{(2t*P)}}

The new count values are then calculated by
\deqn{ctsNEW = cts(u/P)}

and the returned \code{data.frame} is produced by: \code{data.frame(u,ctsNEW)}

The output of the function can be further used for LM-OSL fitting.
}
\note{
The transformation is recommended for curves recorded with a channel
resolution of at least 0.05 s/channel.
}
\section{Function version}{
 0.4.4
}

\examples{

##read curve from CWOSL.SAR.Data transform curve and plot values
data(ExampleData.BINfileData, envir = environment())

##read id for the 1st OSL curve
id.OSL <- CWOSL.SAR.Data@METADATA[CWOSL.SAR.Data@METADATA[,"LTYPE"] == "OSL","ID"]

##produce x and y (time and count data for the data set)
x<-seq(CWOSL.SAR.Data@METADATA[id.OSL[1],"HIGH"]/CWOSL.SAR.Data@METADATA[id.OSL[1],"NPOINTS"],
       CWOSL.SAR.Data@METADATA[id.OSL[1],"HIGH"],
       by = CWOSL.SAR.Data@METADATA[id.OSL[1],"HIGH"]/CWOSL.SAR.Data@METADATA[id.OSL[1],"NPOINTS"])
y <- unlist(CWOSL.SAR.Data@DATA[id.OSL[1]])
values <- data.frame(x,y)

##transform values
values.transformed <- convert_CW2pLM(values)

##plot
plot(values.transformed)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. convert_CW2pLM(): Transform a CW-OSL curve into a pLM-OSL curve. Function version 0.4.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bulur, E., 2000. A simple transformation for converting CW-OSL
curves to LM-OSL curves. Radiation Measurements 32, 141-145.

\strong{Further Reading}

Bulur, E., 1996. An Alternative Technique For Optically Stimulated
Luminescence (OSL) Experiment. Radiation Measurements 26, 701-709.
}
\seealso{
\link{convert_CW2pHMi}, \link{convert_CW2pLMi}, \link{convert_CW2pPMi},
\link{fit_LMCurve}, \link{lm}, \linkS4class{RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_DRTResults.R
\name{plot_DRTResults}
\alias{plot_DRTResults}
\title{Visualise dose recovery test results}
\usage{
plot_DRTResults(
  object,
  given.dose = NULL,
  error.range = 10,
  preheat = NULL,
  boxplot = FALSE,
  mtext = "",
  summary = "",
  summary.pos = "topleft",
  legend = NULL,
  legend.pos = "topright",
  par.local = TRUE,
  na.rm = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
input values containing at least De and De error. To plot
more than one data set in one figure, a \code{list} of the individual data
sets must be provided (e.g. \code{list(dataset.1, dataset.2)}).}

\item{given.dose}{\link{numeric} (\emph{optional}):
given dose used for the dose recovery test to normalise data.
If only one given dose is provided, this given dose is valid for all input
data sets (i.e., \code{object} is a list). Otherwise, a given dose for each input
data set has to be provided (e.g., \code{given.dose = c(100,200)}).
If \code{given.dose} is \code{NULL} or 0, the values are plotted without normalisation
(might be useful for preheat plateau tests).
\strong{Note:} Unit has to be the same as from the input values (e.g., Seconds or
Gray).}

\item{error.range}{\link{numeric} (\emph{with default}):
symmetric error range in percent will be shown as dashed lines in the plot.
It can be set to 0 to remove the error ranges.}

\item{preheat}{\link{numeric} (\emph{optional}):
optional vector of preheat temperatures to be used for grouping the De values.
If specified, the temperatures are assigned to the x-axis.}

\item{boxplot}{\link{logical} (\emph{with default}):
plot values that are grouped by preheat temperature as boxplots.
Only possible when \code{preheat} vector is specified.}

\item{mtext}{\link{character} (\emph{with default}):
additional text below the plot title.}

\item{summary}{\link{character} (\emph{optional}):
adds numerical output to the plot.  Can be one or more out of:
\itemize{
\item \code{"n"} (number of samples),
\item \code{"mean"} (mean De value),
\item \code{"weighted$mean"} (error-weighted mean),
\item \code{"median"} (median of the De values),
\item \code{"weighted$median"} (error-weighted median),
\item \code{"sd.rel"} (relative standard deviation in percent),
\item \code{"sd.abs"} (absolute standard deviation),
\item \code{"se.rel"} (relative standard error) and
\item \code{"se.abs"} (absolute standard error)
}

and all other measures returned by the function \link{calc_Statistics}.}

\item{summary.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"})
for the statistical summary. Alternatively, the keyword \code{"sub"} may be
specified to place the summary below the plot header. However, this latter
option in only possible if \code{mtext} is not used.}

\item{legend}{\link{character} vector (\emph{optional}):
legend content to be added to the plot.}

\item{legend.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"}) for the
legend to be plotted.}

\item{par.local}{\link{logical} (\emph{with default}):
use local graphical parameters for plotting, e.g. the plot is shown in one
column and one row. If \code{par.local = FALSE}, global parameters are inherited,
i.e. parameters provided via \code{par()} work.}

\item{na.rm}{\link{logical} (\emph{with default}):
whether \code{NA} values should be removed from the input data before plotting.}

\item{...}{further arguments and graphical parameters passed to \link{plot},
supported are:
\code{xlab}, \code{ylab}, \code{xlim}, \code{ylim}, \code{main}, \code{cex}, \code{las} and \code{pch}.}
}
\value{
A plot is returned.
}
\description{
The function provides a standardised plot output for dose recovery test
measurements.
}
\details{
The procedure tests the accuracy of a measurement protocol to reliably
determine the dose of a specific sample. Here, the natural signal is erased
and a known laboratory dose administered, which is treated as unknown. Then
the De measurement is carried out and the degree of congruence between
administered and recovered dose is a measure of the protocol's accuracy for
this sample.\cr
In the plot the normalised De is shown on the y-axis, i.e. obtained De/Given Dose.
}
\note{
Further data and plot arguments can be added by using the appropriate R
commands.
}
\section{Function version}{
 0.1.17
}

\examples{

## read example data set and misapply them for this plot type
data(ExampleData.DeValues, envir = environment())

## plot values
plot_DRTResults(
  ExampleData.DeValues$BT998[7:11,],
  given.dose = 2800,
  mtext = "Example data")

## plot values with legend
plot_DRTResults(
  ExampleData.DeValues$BT998[7:11,],
  given.dose = 2800,
  legend = "Test data set")

## create and plot two subsets with randomised values
x.1 <- ExampleData.DeValues$BT998[7:11,]
x.2 <- ExampleData.DeValues$BT998[7:11,] * c(runif(5, 0.9, 1.1), 1)

plot_DRTResults(
  list(x.1, x.2),
  given.dose = 2800)

## some more user-defined plot parameters
plot_DRTResults(
  list(x.1, x.2),
  given.dose = 2800,
  pch = c(2, 5),
  col = c("orange", "blue"),
  xlim = c(0, 8),
  ylim = c(0.85, 1.15),
  xlab = "Sample aliquot")

## plot the data with user-defined statistical measures as legend
plot_DRTResults(
  list(x.1, x.2),
  given.dose = 2800,
  summary = c("n", "weighted$mean", "sd.abs"))

## plot the data with user-defined statistical measures as sub-header
plot_DRTResults(
  list(x.1, x.2),
  given.dose = 2800,
  summary = c("n", "weighted$mean", "sd.abs"),
  summary.pos = "sub")

## plot the data grouped by preheat temperatures
plot_DRTResults(
  ExampleData.DeValues$BT998[7:11,],
  given.dose = 2800,
  preheat = c(200, 200, 200, 240, 240))

## read example data set and misapply them for this plot type
data(ExampleData.DeValues, envir = environment())

## plot values
plot_DRTResults(
  ExampleData.DeValues$BT998[7:11,],
  given.dose = 2800,
  mtext = "Example data")

## plot two data sets grouped by preheat temperatures
plot_DRTResults(
  list(x.1, x.2),
  given.dose = 2800,
  preheat = c(200, 200, 200, 240, 240))

## plot the data grouped by preheat temperatures as boxplots
plot_DRTResults(
  ExampleData.DeValues$BT998[7:11,],
  given.dose = 2800,
  preheat = c(200, 200, 200, 240, 240),
  boxplot = TRUE)

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., 2026. plot_DRTResults(): Visualise dose recovery test results. Function version 0.1.17. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Wintle, A.G., Murray, A.S., 2006. A review of quartz optically
stimulated luminescence characteristics and their relevance in
single-aliquot regeneration dating protocols. Radiation Measurements 41,
369-391.
}
\seealso{
\link{plot}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany)
, RLum Developer Team}
\keyword{dplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_GrowthCurve.R
\name{plot_GrowthCurve}
\alias{plot_GrowthCurve}
\title{Fit and plot a dose-response curve for luminescence data (Lx/Tx against dose)}
\usage{
plot_GrowthCurve(
  object,
  mode = "interpolation",
  fit.method = "EXP",
  output.plot = TRUE,
  output.plotExtended = TRUE,
  plot_singlePanels = FALSE,
  verbose = TRUE,
  n.MC = 100,
  ...
)
}
\arguments{
\item{object}{\link{data.frame} or a \link{list} of such objects (\strong{required}):
data frame with columns for \code{Dose}, \code{LxTx}, \code{LxTx.Error} and \code{TnTx}.
The column for the test dose response is optional, but requires \code{'TnTx'} as
column name if used. For exponential fits at least three dose points
(including the natural) should be provided. If \code{object} is a list,
the function is called on each of its elements.
If \code{fit.method = "OTORX"} you have  to provide the test dose in the same unit
as the dose in a column called \code{Test_Dose}. The function searches explicitly
for this column name. Only the first value will be used assuming a constant
test dose over the measurement cycle.}

\item{mode}{\link{character} (\emph{with default}):
selects calculation mode of the function.
\itemize{
\item \code{"interpolation"} (default) calculates the De by interpolation,
\item \code{"extrapolation"} calculates the equivalent dose by extrapolation
(useful for MAAD measurements) and
\item \code{"alternate"} calculates no equivalent dose and just fits the data points.
}

Please note that for option \code{"interpolation"} the first point is considered
as natural dose}

\item{fit.method}{\link{character} (\emph{with default}):
function used for fitting. Possible options are: \code{LIN}, \code{QDR}, \code{EXP},
\verb{EXP OR LIN}, \code{EXP+LIN}, \code{EXP+EXP} (not defined for extrapolation), \code{GOK},
\code{OTOR} and \code{OTORX}. See details.}

\item{output.plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{output.plotExtended}{\link{logical} (\emph{with default}):
If \code{TRUE}, 3 plots on one plot area are provided:
\enumerate{
\item growth curve,
\item histogram from Monte Carlo error simulation and
\item a test dose response plot.
}

If \code{FALSE}, just the growth curve will be plotted.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}):
single plot output (\code{TRUE/FALSE}) to allow for plotting the results in
single plot windows. Requires \code{output.plotExtended = TRUE}.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{n.MC}{\link{integer} (\emph{with default}):
number of MC runs for error calculation.}

\item{...}{Further arguments to \link{fit_DoseResponseCurve} (\code{fit_weights},
\code{fit_bounds}, \code{fit.force_through_origin}, \code{fit.includingRepeatedRegPoints},
\code{fit.NumberRegPoints}, \code{fit.NumberRegPointsReal}, \code{n.MC},
\code{txtProgressBar}) and graphical parameters to be passed (supported:
\code{xlim}, \code{ylim}, \code{main}, \code{xlab}, \code{ylab}).}
}
\value{
Along with a plot (if wanted) the \linkS4class{RLum.Results} object produced by
\link{fit_DoseResponseCurve} is returned.
}
\description{
A dose-response curve is produced for luminescence measurements using a
regenerative or additive protocol as implemented in \link{fit_DoseResponseCurve}
and \link{plot_DoseResponseCurve}
}
\section{Function version}{
 1.2.3
}

\examples{

##(1) plot growth curve for a dummy dataset
data(ExampleData.LxTxData, envir = environment())
plot_GrowthCurve(LxTxData)

##(1b) horizontal plot arrangement
layout(mat = matrix(c(1,1,2,3), ncol = 2))
plot_GrowthCurve(LxTxData, plot_singlePanels = TRUE)

##(2) plot the growth curve with pdf output - uncomment to use
##pdf(file = "~/Desktop/Growth_Curve_Dummy.pdf", paper = "special")
plot_GrowthCurve(LxTxData)
##dev.off()

##(3) plot the growth curve with pdf output - uncomment to use, single output
##pdf(file = "~/Desktop/Growth_Curve_Dummy.pdf", paper = "special")
temp <- plot_GrowthCurve(LxTxData, plot_singlePanels = TRUE)
##dev.off()

##(4) plot resulting function for given interval x
x <- seq(1,10000, by = 100)
plot(
 x = x,
 y = eval(temp$Formula),
 type = "l"
)

##(5) plot using the 'extrapolation' mode
LxTxData[1,2:3] <- c(0.5, 0.001)
print(plot_GrowthCurve(LxTxData, mode = "extrapolation"))

##(6) plot using the 'alternate' mode
LxTxData[1,2:3] <- c(0.5, 0.001)
print(plot_GrowthCurve(LxTxData, mode = "alternate"))

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., Colombo, M., 2026. plot_GrowthCurve(): Fit and plot a dose-response curve for luminescence data (Lx/Tx against dose). Function version 1.2.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46.

Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
from feldspar. Radiation Measurements 81, 224-231.

Pagonis, V., Kitis, G., Chen, R., 2020. A new analytical equation for the dose response of dosimetric materials,
based on the Lambert W function. Journal of Luminescence 225, 117333. \doi{10.1016/j.jlumin.2020.117333}
}
\seealso{
\link{fit_DoseResponseCurve}, \link{plot_DoseResponseCurve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_RLum.Data.Curve.R
\name{merge_RLum.Data.Curve}
\alias{merge_RLum.Data.Curve}
\title{Merge function for RLum.Data.Curve S4 class objects}
\usage{
merge_RLum.Data.Curve(object, merge.method = "mean", method.info)
}
\arguments{
\item{object}{\link{list} of \linkS4class{RLum.Data.Curve} (\strong{required}):
list of S4 objects of class \code{RLum.Curve}.}

\item{merge.method}{\link{character} (\strong{required}):
method for combining of the objects, e.g.  \code{'mean'}, \code{'sum'}, see details for
further information and allowed methods.  Note: Elements in slot info will
be taken from the first curve in the list.}

\item{method.info}{\link{numeric} (\emph{optional}):
allows to specify how info elements of the input objects are combined,
e.g. \code{1} means that just the elements from the first object are kept,
\code{2} keeps only the info elements from the 2 object etc.
If nothing is provided all elements are combined.}
}
\value{
Returns an \linkS4class{RLum.Data.Curve} object.
}
\description{
Function allows merging of RLum.Data.Curve objects in different ways
}
\details{
This function simply allowing to merge \linkS4class{RLum.Data.Curve}
objects without touching the objects itself. Merging is always applied on
the 2nd column of the data matrix of the object.

\strong{Supported merge operations are \linkS4class{RLum.Data.Curve}}

\code{"sum"}

All count values will be summed up using the function \link{rowSums}.

\code{"mean"}

The mean over the count values is calculated using the function
\link{rowMeans}.

\code{"median"}

The median over the count values is calculated using the function
\link[matrixStats:rowMedians]{matrixStats::rowMedians}.

\code{"sd"}

The standard deviation over the count values is calculated using the function
\link[matrixStats:rowSds]{matrixStats::rowSds}.

\code{"var"}

The variance over the count values is calculated using the function
\link[matrixStats:rowVars]{matrixStats::rowVars}.

\code{"min"}

The min values from the count values is chosen using the function
\link[matrixStats:rowRanges]{matrixStats::rowMins}.

\code{"max"}

The max values from the count values is chosen using the function
\link[matrixStats:rowRanges]{matrixStats::rowMins}.

\code{"append"}

Appends count values of all curves to one combined data curve. The channel width
is automatically re-calculated, but requires a constant channel width of the
original data.

\code{"-"}

The row sums of the last objects are subtracted from the first object.

\code{"*"}

The row sums of the last objects are multiplied with the first object.

\code{"/"}

Values of the first object are divided by row sums of the last objects.
}
\note{
The information from the slot \code{recordType} is taken from the first
\linkS4class{RLum.Data.Curve} object in the input list. The slot
'curveType' is filled with the name \code{merged}.
}
\section{S3-generic support}{


This function is fully operational via S3-generics:
\code{+}, \code{-}, \code{/}, \code{*}, \code{merge}
}

\section{Function version}{
 0.2.1
}

\examples{


##load example data
data(ExampleData.XSYG, envir = environment())

##grep first and 3d TL curves
TL.curves  <- get_RLum(OSL.SARMeasurement$Sequence.Object, recordType = "TL (UVVIS)")
TL.curve.1 <- TL.curves[[1]]
TL.curve.3 <- TL.curves[[3]]

##plot single curves
plot_RLum(TL.curve.1)
plot_RLum(TL.curve.3)

##subtract the 1st curve from the 2nd and plot
TL.curve.merged <- merge_RLum.Data.Curve(list(TL.curve.3, TL.curve.1), merge.method = "/")
plot_RLum(TL.curve.merged)

}
\seealso{
\link{merge_RLum}, \linkS4class{RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2023. merge_RLum.Data.Curve(): Merge function for RLum.Data.Curve S4 class objects. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., 2023. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 0.9.23. https://CRAN.R-project.org/package=Luminescence
}

\keyword{internal}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_IrradiationTimes.R
\name{extract_IrradiationTimes}
\alias{extract_IrradiationTimes}
\title{Extract Irradiation Times from an XSYG-file or \code{RLum.Analysis} object}
\usage{
extract_IrradiationTimes(
  object,
  file.BINX,
  recordType = c("irradiation (NA)", "IRSL (UVVIS)", "OSL (UVVIS)", "TL (UVVIS)"),
  return_same_as_input = FALSE,
  compatibility.mode = TRUE,
  txtProgressBar = TRUE
)
}
\arguments{
\item{object}{\link{character}, \linkS4class{RLum.Analysis} or \link{list} (\strong{required}):
path and file name of the XSYG file or an \linkS4class{RLum.Analysis}
produced by the function \link{read_XSYG2R};
alternatively, a \code{list} of \linkS4class{RLum.Analysis} can be provided.

\strong{Note:} If an \linkS4class{RLum.Analysis} is used, any input for
the arguments \code{file.BINX} and \code{recordType} will be ignored!}

\item{file.BINX}{\link{character} (\emph{optional}):
path and file name of an existing BINX-file. If a file name is provided the
file will be updated with the information from the XSYG file in the same
folder as the original BINX-file.

\strong{Note:} The XSYG and the BINX-file must originate from the
same measurement!}

\item{recordType}{\link{character} (\emph{with default}):
select relevant curves types from the XSYG file or \linkS4class{RLum.Analysis}
object. As the XSYG-file format comprises much more information than usually
needed for routine data analysis and allowed in the BINX-file format, only
the relevant curves are selected by using the function
\link{get_RLum}. The argument \code{recordType} works as
described for this function.

\strong{Note:} A wrong selection will causes a function error. Please change this
argument only if you have reasons to do so.}

\item{return_same_as_input}{\link{logical} (\emph{with default}):
if set to \code{TRUE}, an updated \linkS4class{RLum.Analysis} object (or a \link{list} of
it) is returned, with each record having gained two new info element fields:
\code{IRR_TIME} and \code{TIMESCINCEIRR}. This makes the \linkS4class{RLum.Analysis} object
compatible with external functions that search explicitly for \code{IRR_TIME}
and \code{TIMESCINCEIRR}.}

\item{compatibility.mode}{\link{logical} (\emph{with default}):
whether all position values should be reset to a maximum value of 48 (see
\link{write_R2BIN}). Only used if \code{file.BINX} is specified.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar during import and export.}
}
\value{
An \linkS4class{RLum.Results} object is returned with the
following structure:

\if{html}{\out{<div class="sourceCode">}}\preformatted{.. $irr.times (data.frame)
}\if{html}{\out{</div>}}

If \code{return_same_as_input = TRUE} an \linkS4class{RLum.Analysis} or a \link{list} of it, but
we updated info elements including irradiation times.

If a BINX-file path and name is set, the output will be additionally
transferred into a new BINX-file (with the function name as suffix) located
in the folder of the input BINX-file. Note that this will
not work if the input object is a file path to an XSYG-file, instead of a
link to only one file. In this case the argument input for \code{file.BINX} is ignored.

In the self call mode (input is a \code{list} of \linkS4class{RLum.Analysis} objects
a list of \linkS4class{RLum.Results} is returned.
}
\description{
Extracts irradiation times, dose and times since last irradiation, from a
Freiberg Instruments XSYG-file. These information can be further used to
update an existing BINX-file.
}
\details{
The function was written to compensate missing information in the BINX-file
output of Freiberg Instruments lexsyg readers. As all information are
available within the XSYG-file anyway, these information can be extracted
and used for further analysis or/and to stored in a new BINX-file, which can
be further used by other software, e.g., \emph{Analyst} (Geoff Duller).

Typical application example: \emph{g}-value estimation from fading measurements
using the Analyst or any other self-written script.

Beside some simple data transformation steps, the function relies on
\link{read_XSYG2R}, \link{read_BIN2R}, \link{write_R2BIN} for data import and export.

\strong{Calculation details}
\itemize{
\item The value \code{DURATION.STEP} is calculated as \code{START} + the end of the time axis for all curves
except for \code{TL}, where the function tries to extract meta information about the duration.
\item The value \code{END} is calculated as \code{START} + \code{DURACTION.STEP}
\item All curves for which no prior irradiation was detected receive an \code{-1} (\emph{Analyst} convention)
\item Irradiation steps have always \code{IRR_TIME = 0}
}
}
\note{
The function can be also used to extract irradiation times from
\linkS4class{RLum.Analysis} objects imported via \link{read_BIN2R} with option
\code{fastForward = TRUE}, or in combination with
\link{Risoe.BINfileData2RLum.Analysis}.
Unfortunately the timestamp might not be very precise (or even invalid),
but it allows to essentially treat different formats in a similar manner.

The produced output object contains still the irradiation steps to
keep the output transparent. However, for the BINX-file export this steps
are removed as the BINX-file format description does not allow irradiations
as separate sequences steps.

\strong{BINX-file 'Time Since Irradiation' value differs from the table output?}

The way the value 'Time Since Irradiation' is defined differs. In the BINX-file the
'Time Since Irradiation' is calculated as the 'Time Since Irradiation' plus the 'Irradiation
Time'. The table output returns only the real 'Time Since Irradiation', i.e. time between the
end of the irradiation and the next step.

\strong{Negative values for \code{TIMESINCELAST.STEP}?}

Yes, this is possible and not a bug, as in the XSYG-file multiple curves are stored for one step.
Example: TL step may comprise three curves:
\itemize{
\item (a) counts vs. time,
\item (b) measured temperature vs. time and
\item (c) predefined temperature vs. time.
}

Three curves, but they are all belonging to one TL measurement step, but with regard to
the time stamps this could produce negative values as the important function
(\link{read_XSYG2R}) do not change the order of entries for one step
towards a correct time order.

\strong{\code{TIMESINCELAST.STEP} is odd if TL curves are involved?}

Yes, this is possible! The end time is calculated as start + duration,
and the duration is deduced from the time values of each step.
A typical TL curve, however, does not have a time but a temperature axis.
Hence, the function tries to extract the information from metadata.
If that information is missing, the x-values are presumed time values,
which might still be wrong.
}
\section{Function version}{
 0.4.0
}

\examples{
## take system file
xsyg <- system.file("extdata/XSYG_file.xsyg", package="Luminescence")

## the import is automatically
## but you can import it before
irr_times <- extract_IrradiationTimes(xsyg)
irr_times$irr.times

\dontrun{
# (1) - example for your own data

# set files and run function
file.XSYG <- file.choose()
file.BINX <- file.choose()

extract_IrradiationTimes(file.XSYG = file.XSYG, file.BINX = file.BINX)

# export results additionally to a CSV-file in the same directory as the XSYG-file
write.table(x = get_RLum(output),
 file = paste0(file.BINX,"_extract_IrradiationTimes.csv"),
 sep = ";",
 row.names = FALSE)
}

} 

\section{How to cite}{
Kreutzer, S., 2026. extract_IrradiationTimes(): Extract Irradiation Times from an XSYG-file or RLum.Analysis object. Function version 0.4.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Duller, G.A.T., 2015. The Analyst software package for luminescence data: overview and
recent improvements. Ancient TL 33, 35-42. \doi{10.26034/la.atl.2015.489}
}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Results},
\linkS4class{Risoe.BINfileData},
\link{read_XSYG2R}, \link{read_BIN2R}, \link{write_R2BIN}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{IO}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_DevelopmentVersion.R
\name{install_DevelopmentVersion}
\alias{install_DevelopmentVersion}
\title{Attempts to install the development version of the 'Luminescence' package}
\usage{
install_DevelopmentVersion(force_install = FALSE, branch = "master")
}
\arguments{
\item{force_install}{\link{logical} (\emph{optional}):
If \code{FALSE} (the default) the function produces and prints the required
code to the console for the user to run manually afterwards. When \code{TRUE}
and all requirements are fulfilled (see details) this function attempts to install
the package itself.}

\item{branch}{\link{character} (\emph{with default}):
Name of the branch to install. The default value ("master") corresponds to
the main development branch.}
}
\value{
This function requires user input at the command prompt to choose the
desired development branch to be installed. The required R code to install
the package is then printed to the console.
}
\description{
This function provides a convenient method for installing the development
version of the R package 'Luminescence' directly from GitHub.
}
\details{
This function checks whether the 'devtools' package is currently installed
on the system. If \code{force_install = TRUE} the functions attempts to install
the chosen development branch via \link[devtools:remote-reexports]{devtools::install_github}.
}
\examples{
\dontrun{
install_DevelopmentVersion()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_HeliosOSL2R.R
\name{read_HeliosOSL2R}
\alias{read_HeliosOSL2R}
\title{Import Luminescence Data from Helios Luminescence Reader}
\usage{
read_HeliosOSL2R(file, verbose = TRUE, ...)
}
\arguments{
\item{file}{\link{character}, \link{list} (\strong{required}):
name of one or multiple \code{.osl} files (URLs are supported); it can be the
path to a directory, in which case the function tries to detect and import
all \code{.osl} files found in the directory.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{not in use, for compatibility reasons only}
}
\value{
A \linkS4class{RLum.Analysis} object. Results are returned as a
list when multiple files are processed or \code{file} is a list.
}
\description{
Import of files with \code{.osl} extension produced by the zero rad Helios
luminescence reader and conversion to \linkS4class{RLum.Analysis}
objects.
}
\note{
Thanks to Krzysztof Maternicki for providing example data.
}
\section{Function version}{
 0.1.1
}

\examples{
file <- system.file("extdata/HeliosOSL_Example.osl", package = "Luminescence")
read_HeliosOSL2R(file)

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. read_HeliosOSL2R(): Import Luminescence Data from Helios Luminescence Reader. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verify_SingleGrainData.R
\name{verify_SingleGrainData}
\alias{verify_SingleGrainData}
\title{Verify single grain data sets and check for invalid grains, i.e.
zero-light level grains}
\usage{
verify_SingleGrainData(
  object,
  threshold = 10,
  use_fft = FALSE,
  cleanup = FALSE,
  cleanup_level = c("aliquot", "curve"),
  verbose = TRUE,
  plot = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData} or \linkS4class{RLum.Analysis} (\strong{required}):
input object. The function also accepts a list with objects of allowed type.}

\item{threshold}{\link{numeric} (\emph{with default}):
numeric threshold value for the allowed difference between the \code{mean} and
the \code{var} of the count values (see details).}

\item{use_fft}{\link{logical} (\emph{with default}): applies an additional approach based on \link[stats:fft]{stats::fft}.
The threshold is fixed and cannot be changed.}

\item{cleanup}{\link{logical} (\emph{with default}):
if set to \code{TRUE}, curves/aliquots identified as zero light level curves/aliquots are
automatically removed. Output is an object as same type as the input, i.e.
either \linkS4class{Risoe.BINfileData} or \linkS4class{RLum.Analysis}}

\item{cleanup_level}{\link{character} (\emph{with default}):
selects the level for the clean-up of the input data sets.
Two options are allowed: \code{"curve"} or \code{"aliquot"}:
\itemize{
\item If  \code{"curve"} is selected, every single curve marked as \code{invalid} is removed.
\item If \code{"aliquot"} is selected, curves of one aliquot (grain or disc) can be
marked as invalid, but will not be removed. An aliquot will be only removed
if all curves of this aliquot are marked as invalid.
}}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disables output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further parameters to control the plot output; if selected.
Supported arguments \code{main}, \code{ylim}}
}
\value{
The function returns

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \verb{@data}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$unique_pairs} \tab \code{data.frame} \tab the unique position and grain pairs \cr
\verb{$selection_id} \tab \code{numeric} \tab the selection as record ID \cr
\verb{$selection_full} \tab \code{data.frame} \tab implemented models used in the baSAR-model core \cr
}

\strong{slot:} \verb{@info}

The original function call

\strong{Output variation}

For \code{cleanup = TRUE} the same object as the input is returned, but cleaned up
(invalid curves were removed). This means: Either a \linkS4class{Risoe.BINfileData}
or an \linkS4class{RLum.Analysis} object is returned in such cases.
A \linkS4class{Risoe.BINfileData} object can be exported to a BINX-file by
using the function \link{write_R2BIN}.
}
\description{
This function tries to identify automatically zero-light level curves (grains)
from single grain data measurements.
}
\details{
\strong{How does the method work?}

The function compares the expected values (\eqn{E(X)}) and the variance
(\eqn{Var(X)}) of the count values for each curve. Assuming that the
background roughly follows a Poisson distribution, the absolute difference
of both values should be zero or at least around zero as

\deqn{E(x) = Var(x) = \lambda}

Thus the function checks for:

\deqn{abs(E(x) - Var(x)) >= \Theta}

With \eqn{\Theta} an arbitrary, user defined, threshold. Values above the
threshold indicate curves comprising a signal.

Note: the absolute difference of \eqn{E(X)} and \eqn{Var(x)} instead of the
ratio was chosen as both terms can become 0 which would result in 0 or \code{Inf},
if the ratio is calculated.
}
\note{
This function can work with \linkS4class{Risoe.BINfileData} objects or
\linkS4class{RLum.Analysis} objects (or a list of it). However, the function is
highly optimised for \linkS4class{Risoe.BINfileData} objects as it make sense to
remove identify invalid grains before the conversion to an
\linkS4class{RLum.Analysis} object.

The function checking for invalid curves works rather robust and it is likely
that Reg0 curves within a SAR cycle are removed as well. Therefore it is
strongly recommended to use the argument \code{cleanup = TRUE} carefully if
the cleanup works only on curves.
}
\section{Function version}{
 0.2.7
}

\examples{

##01 - basic example I
##just show how to apply the function
data(ExampleData.XSYG, envir = environment())

##verify and get data.frame out of it
verify_SingleGrainData(OSL.SARMeasurement$Sequence.Object)$selection_full

##02 - basic example II
data(ExampleData.BINfileData, envir = environment())
id <- verify_SingleGrainData(object = CWOSL.SAR.Data,
cleanup_level = "aliquot")$selection_id

\dontrun{
##03 - advanced example I
##importing and exporting a BIN-file

##select and import file
file <- file.choose()
object <- read_BIN2R(file)

##remove invalid aliquots(!)
object <- verify_SingleGrainData(object, cleanup = TRUE)

##export to new BIN-file
write_R2BIN(object, paste0(dirname(file),"/", basename(file), "_CLEANED.BIN"))
}

}
\seealso{
\linkS4class{Risoe.BINfileData}, \linkS4class{RLum.Analysis},
\link{write_R2BIN}, \link{read_BIN2R}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. verify_SingleGrainData(): Verify single grain data sets and check for invalid grains, i.e. zero-light level grains. Function version 0.2.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_AliquotSize.R
\name{calc_AliquotSize}
\alias{calc_AliquotSize}
\title{Estimate the amount of grains on an aliquot}
\usage{
calc_AliquotSize(
  grain.size,
  sample.diameter,
  packing.density = 0.65,
  MC = TRUE,
  grains.counted = NULL,
  sample_carrier.diameter = 9.8,
  plot = TRUE,
  ...
)
}
\arguments{
\item{grain.size}{\link{numeric} (\strong{required}):
mean grain size (microns) or a range of grain sizes from which the
mean grain size is computed (e.g. \code{c(100,200)}).}

\item{sample.diameter}{\link{numeric} (\strong{required}):
diameter (mm) of the targeted area on the sample carrier.}

\item{packing.density}{\link{numeric} (\emph{with default}):
empirical value for the mean packing density. \cr
If \code{packing.density = Inf}, a hexagonal structure on an infinite
plane with a packing density of \eqn{\pi / \sqrt{12} \approx 0.9069}
is assumed.}

\item{MC}{\link{logical} (\emph{optional}):
if \code{TRUE} the function performs a Monte Carlo simulation for estimating the
amount of grains on the sample carrier and assumes random errors in grain
size distribution and packing density. Requires \code{grain.size} to be specified
as a 2-element vector with the range (min and max) of grain sizes.
For more information see details.}

\item{grains.counted}{\link{numeric} (\emph{optional}):
grains counted on a sample carrier. If a non-zero positive integer is
provided, this function will calculate the packing density of the aliquot.
If more than one value is provided, the mean packing density and its
standard deviation are calculated.
Note that this overrides \code{packing.density}.}

\item{sample_carrier.diameter}{\link{numeric} (\emph{with default}):
diameter (mm) of the sample carrier.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments and graphical parameters to be passed. In
particular, \code{MC.iter} to control the number of Monte Carlo iterations,
\code{main}, \code{xlab}, \code{col}, \code{line_col}, \code{line_lwd}, \code{cex}, \code{rug} (\code{TRUE/FALSE}),
\code{boxplot} (\code{TRUE/FALSE}), \code{summary} (\code{TRUE/FALSE}), \code{legend} (\code{TRUE/FALSE}).}
}
\value{
Returns a terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following element:

\item{.$summary}{\link{data.frame} summary of all relevant calculation results.}
\item{.$args}{\link{list} used arguments}
\item{.$call}{\link{call} the function call}
\item{.$MC}{\link{list} results of the Monte Carlo simulation}

The output should be accessed using the function \link{get_RLum}.
}
\description{
This function can be used to either estimate the number of grains on an
aliquot or to compute the packing density, depending on the arguments
provided.

The following formula is used to estimate the number of grains \code{n}:

\deqn{n = (\pi*x^2)/(\pi*y^2)*d}

where \code{x} is the radius of the aliquot size (microns), \code{y} is the mean
radius of the mineral grains (mm), and \code{d} is the packing density
(a value between 0 and 1).

\strong{Packing density}

The default value for \code{packing.density} is 0.65, which is the mean of
empirical values determined by Heer et al. (2012) and unpublished data from
the Cologne luminescence laboratory. If \code{packing.density = Inf}, a maximum
density of \eqn{\pi / \sqrt{12} \approx 0.9069} is used. However, note that
this value is not appropriate as the standard preparation procedure of
aliquots resembles a PECC (\emph{Packing Equal Circles in a Circle}) problem,
where the maximum packing density is asymptotic to about 0.87.

\strong{Monte Carlo simulation}

The number of grains on an aliquot can be estimated by Monte Carlo simulation
when setting \code{MC = TRUE}. All parameters necessary to calculate
\code{n} (\code{x}, \code{y}, \code{d}) are assumed to be normally distributed with means
\eqn{\mu_x, \mu_y, \mu_d} and standard deviations \eqn{\sigma_x, \sigma_y, \sigma_d}.

For the mean grain size, random samples are taken first from
\eqn{N(\mu_y, \sigma_y)}, where \eqn{\mu_y = mean.grain.size} and
\eqn{\sigma_y = (max.grain.size-min.grain.size)/4}, so that 95\% of all
grains are within the provided the grain size range. This effectively takes
into account that after sieving the sample there is still a small chance of
having grains smaller or larger than the used mesh sizes. For each random
sample the mean grain size is calculated, from which random subsamples are
drawn for the Monte Carlo simulation.

The packing density is assumed
to be normally distributed with an empirically determined \eqn{\mu = 0.65}
(or provided value) and \eqn{\sigma = 0.18}. The normal distribution is
truncated at \code{d = 0.87} as this is approximately the maximum packing
density that can be achieved in a PECC problem.

The sample diameter has
\eqn{\mu = sample.diameter} and \eqn{\sigma = 0.2} to take into account
variations in sample disc preparation (i.e. applying silicon spray to the
disc). A lower truncation point at \code{x = 0.5} is used, which assumes
that aliquots with sample diameter smaller than 0.5 mm are discarded.
Likewise, the normal distribution is truncated at the diameter of the
sample carrier (9.8 mm by default, but controllable via the
\code{sample_carrier.diameter} argument).

For each random sample drawn from the
normal distribution, the amount of grains on the aliquot is calculated. By
default, \code{10^4} iterations are used, but the can be controlled with
option \code{MC.iter} (see \code{...}). Results are visualised in a bar- and
boxplot together with a statistical summary.
}
\section{Function version}{
 0.33
}

\examples{

## Estimate the amount of grains on a small aliquot
calc_AliquotSize(grain.size = c(100,150), sample.diameter = 1, MC.iter = 100)

## Calculate the mean packing density of large aliquots
calc_AliquotSize(grain.size = c(100,200), sample.diameter = 8,
                 grains.counted = c(2525,2312,2880), MC.iter = 100)

} 

\section{How to cite}{
Burow, C., Colombo, M., 2026. calc_AliquotSize(): Estimate the amount of grains on an aliquot. Function version 0.33. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Duller, G.A.T., 2008. Single-grain optical dating of Quaternary
sediments: why aliquot size matters in luminescence dating. Boreas 37,
589-612.

Heer, A.J., Adamiec, G., Moska, P., 2012. How many grains
are there on a single aliquot?. Ancient TL 30, 9-16.

\strong{Further reading}

Chang, H.-C., Wang, L.-C., 2010. A simple proof of Thue's
Theorem on Circle Packing. \url{https://arxiv.org/pdf/1009.4322v1},
2013-09-13.

Graham, R.L., Lubachevsky, B.D., Nurmela, K.J.,
Oestergard, P.R.J., 1998.  Dense packings of congruent circles in a circle.
Discrete Mathematics 181, 139-154.

Huang, W., Ye, T., 2011. Global
optimization method for finding dense packings of equal circles in a circle.
European Journal of Operational Research 210, 474-481.
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R,
%   R/RLum.Data.Image-class.R, R/RLum.Data.Spectrum-class.R,
%   R/RLum.Results-class.R
\name{names_RLum}
\alias{names_RLum}
\alias{names_RLum,list-method}
\alias{names_RLum,RLum.Analysis-method}
\alias{names_RLum,RLum.Data.Curve-method}
\alias{names_RLum,RLum.Data.Image-method}
\alias{names_RLum,RLum.Data.Spectrum-method}
\alias{names_RLum,RLum.Results-method}
\title{Name retrieval function for RLum-class objects}
\usage{
names_RLum(object)

\S4method{names_RLum}{list}(object)

\S4method{names_RLum}{RLum.Analysis}(object)

\S4method{names_RLum}{RLum.Data.Curve}(object)

\S4method{names_RLum}{RLum.Data.Image}(object)

\S4method{names_RLum}{RLum.Data.Spectrum}(object)

\S4method{names_RLum}{RLum.Results}(object)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum}}
}
\value{
A \link{character} vector.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the input object, the corresponding
function will be selected.
}
\section{Functions}{
\itemize{
\item \code{names_RLum(list)}: Returns a list of names of the \linkS4class{RLum} objects that had been passed to
it.

\item \code{names_RLum(RLum.Analysis)}: Returns the names of the \linkS4class{RLum.Data} objects stored in the object.

\item \code{names_RLum(RLum.Data.Curve)}: Returns the names info elements stored in the object.

\item \code{names_RLum(RLum.Data.Image)}: Returns the names of the info elements stored in the object.

\item \code{names_RLum(RLum.Data.Spectrum)}: Returns the names of the info elements stored in the object.

\item \code{names_RLum(RLum.Results)}: Returns the names of the \code{data} field stored in the object.

}}
\section{Function version}{
 0.1.0
}

\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Data.Image},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. names_RLum(): Name retrieval function for RLum-class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_OSLLxTxRatio.R
\name{calc_OSLLxTxRatio}
\alias{calc_OSLLxTxRatio}
\title{Calculate \code{Lx/Tx} ratio for CW-OSL curves}
\usage{
calc_OSLLxTxRatio(
  Lx.data,
  Tx.data = NULL,
  signal_integral = NULL,
  background_integral = NULL,
  signal_integral_Tx = NULL,
  background_integral_Tx = NULL,
  integral_input = c("channel", "measurement"),
  background.count.distribution = "non-poisson",
  use_previousBG = FALSE,
  sigmab = NULL,
  sig0 = 0,
  digits = NULL,
  ...
)
}
\arguments{
\item{Lx.data}{\linkS4class{RLum.Data.Curve}, \link{data.frame}, \link{list} (\strong{required}):
requires a CW-OSL shine down curve (x = time, y = counts). Data can also be
provided as a list.}

\item{Tx.data}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\emph{optional}):
requires a CW-OSL shine down curve (x = time, y = counts). If no
input is given the \code{Tx.data} will be treated as \code{NA} and no \code{Lx/Tx} ratio
is calculated. When \code{Lx.data} is a list, it must be provided as a list of
the same length.}

\item{signal_integral}{\link{integer} (\strong{required}):
vector of channels for the signal integral. If set to \code{NA} (alternate mode),
no integrals are taken into account and their settings are ignored.}

\item{background_integral}{\link{integer} (\strong{required}):
vector of channels for the background integral. If set to \code{NA}, no
background integral is subtracted; in this case, \code{sigmab.LnLx} (unless
manually set) and the signal-to-noise ratio for Ln/Lx\verb{will be}NA`.}

\item{signal_integral_Tx}{\link{integer} (\emph{optional}):
vector of channels for the signal integral for the \code{Tx} curve.
If \code{NULL}, the \code{signal_integral} vector is used.}

\item{background_integral_Tx}{\link{integer} (\emph{optional}):
vector of channels for the background integral for the \code{Tx} curve.
If \code{NULL}, the \code{background_integral} vector is used. If set to \code{NA}, no
background integral for the \code{Tx} curve is subtracted; in this case,
\code{sigmab.TxTx} (unless manually set) and the signal-to-noise ratio for
\code{Tn/Tx} will be \code{NA}.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time/temperature range are selected.}

\item{background.count.distribution}{\link{character} (\emph{with default}):
sets the count distribution assumed for the error calculation.
Possible arguments are \code{"poisson"} or \code{"non-poisson"} (default). See
details for further information.}

\item{use_previousBG}{\link{logical} (\emph{with default}):
If set to \code{TRUE} the background of the \code{Lx}-signal is subtracted also
from the \code{Tx}-signal. Please note that in this case separate
signal integral limits for the \code{Tx}-signal are not allowed and will be reset.}

\item{sigmab}{\link{numeric} (\emph{optional}):
option to set a manual value for the overdispersion (for \code{LnTx} and \code{TnTx}),
used for the \code{Lx/Tx} error calculation. The value should be provided as
absolute squared count values, e.g. \code{sigmab = c(300,300)}.
\strong{Note:} If only one value is provided this value is taken for both (\code{LnTx} and \code{TnTx}) signals.}

\item{sig0}{\link{numeric} (\emph{with default}):
allow adding an extra component of error to the final \code{Lx/Tx} error value
(e.g., instrumental error, see details).}

\item{digits}{\link{integer} (\emph{with default}):
round numbers to the specified digits. If set to \code{NULL} no rounding occurs.}

\item{...}{currently not used.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.

Slot \code{data} contains a \link{list} with the following structure:

\strong{@data}

\if{html}{\out{<div class="sourceCode">}}\preformatted{$LxTx.table (data.frame)
.. $ LnLx
.. $ LnLx.BG
.. $ TnTx
.. $ TnTx.BG
.. $ Net_LnLx
.. $ Net_LnLx.Error
.. $ Net_TnTx
.. $ Net_TnTx.Error
.. $ SN_RATIO_LnLx,
.. $ SN_RATIO_TnTx,
.. $ LxTx
.. $ LxTx.Error
$calc.parameters (list)
.. $ sigmab.LnTx
.. $ sigmab.TnTx
.. $ k
}\if{html}{\out{</div>}}

\strong{@info}

\if{html}{\out{<div class="sourceCode">}}\preformatted{$ call (original function call)
}\if{html}{\out{</div>}}
}
\description{
Calculate \code{Lx/Tx} ratios from a given set of CW-OSL curves assuming late light
background subtraction.
}
\details{
The function checks the integrity of the values chosen for the signal and
background integrals; the signal integral limits have to be lower than
the background integral limits. If a \link{vector} is given as input instead
of a \link{data.frame}, an artificial \link{data.frame} is produced. The
error calculation is done according to Galbraith (2002).

\strong{Please note:} In cases where the calculation results in \code{NaN} values (for
example due to zero-signal, and therefore a division of 0 by 0), these \code{NaN}
values are replaced by 0.

\strong{\code{sigmab}}

The default value of \code{sigmab} is calculated assuming the background is
constant and \strong{would not} applicable when the background varies as,
e.g., as observed for the early light subtraction method.

\strong{\code{sig0}}

This argument allows to add an extra component of error to the final \code{Lx/Tx}
error value. The input will be treated as factor that is multiplied with
the already calculated \code{LxTx} and the result is add up by:

\deqn{se(LxTx) = \sqrt{se(LxTx)^2 + (LxTx * sig0)^2}}

\strong{\code{SN_RATIO_LnLx} and \code{SN_RATIO_TnTx}}

For convenience, the function returns the signal-to-noise ratio (\code{SN_RATIO})
for the \code{LnLx} and the \code{TnTx} curves. This is simply the signal divided
by the background signal counts normalised to the \code{k} value (see below).

\strong{\code{background.count.distribution}}

This argument allows selecting the distribution assumption that is used for
the error calculation. According to Galbraith (2002, 2014) the background
counts may be overdispersed (i.e. do not follow a Poisson distribution,
which is assumed for the photomultiplier counts). In that case (might be the
normal case) it has to be accounted for the overdispersion by estimating
\eqn{\sigma^2} (i.e. the overdispersion value). Therefore the relative
standard error is calculated as:
\itemize{
\item \code{poisson}
\deqn{rse(\mu_{S}) \approx \sqrt{Y_{0} + Y_{1}/k^2) / (Y_{0} - Y_{1}/k)} }
\item \code{non-poisson}
\deqn{rse(\mu_{S}) \approx \sqrt{Y_{0} + Y_{1}/k^2 + \sigma^2(1+1/k)) / (Y_{0} - Y_{1}/k)} }
}

If \code{background_integral = NA}, then in both cases the relative standard
error simplifies to:

\deqn{rse(\mu_{S}) \approx \sqrt{Y_{0}} / Y_{0}}

\strong{Please note} that when using the early background subtraction method in
combination with the 'non-poisson' distribution argument, the corresponding \code{Lx/Tx} error
may considerably increase due to a high \code{sigmab} value.
Please check whether this is valid for your data set and if necessary
consider to provide an own \code{sigmab} value using the corresponding argument \code{sigmab}.
}
\note{
The results of this function have been cross-checked with the Analyst
(version 3.24b). Access to the results object via \link{get_RLum}.

\strong{Caution:} If you are using early light subtraction (EBG), please either provide your
own \code{sigmab} value or use \code{background.count.distribution = "poisson"}.
}
\section{Function version}{
 0.9.8
}

\examples{

##load data
data(ExampleData.LxTxOSLData, envir = environment())

##calculate Lx/Tx ratio
results <- calc_OSLLxTxRatio(
 Lx.data = Lx.data,
 Tx.data = Tx.data,
 signal_integral = 1:2,
 background_integral = 85:100)

##get results object
get_RLum(results)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. calc_OSLLxTxRatio(): Calculate Lx/Tx ratio for CW-OSL curves. Function version 0.9.8. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Duller, G., 2018. Analyst v4.57 - User Manual.
\verb{https://users.aber.ac.uk/ggd}\cr

Galbraith, R.F., 2002. A note on the variance of a background-corrected OSL
count. Ancient TL, 20 (2), 49-51. \doi{10.26034/la.atl.2002.348}

Galbraith, R.F., 2014. A further note on the variance of a
background-corrected OSL count. Ancient TL, 31 (2), 1-3. \doi{10.26034/la.atl.2014.477}
}
\seealso{
\linkS4class{RLum.Data.Curve}, \link{fit_DoseResponseCurve},
\link{analyse_SAR.CWOSL}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Risoe.BINfileData-class.R
\docType{class}
\name{Risoe.BINfileData-class}
\alias{Risoe.BINfileData-class}
\alias{set_Risoe.BINfileData,ANY-method}
\title{Class \code{"Risoe.BINfileData"}}
\usage{
\S4method{set_Risoe.BINfileData}{ANY}(
  METADATA = data.frame(),
  DATA = list(),
  .RESERVED = list()
)
}
\arguments{
\item{METADATA}{Object of class "data.frame" containing the meta information
for each curve.}

\item{DATA}{Object of class "list" containing numeric vector with count data.}

\item{.RESERVED}{Object of class "list" containing list of undocumented raw
values for internal use only.}
}
\description{
S4 class object for luminescence data in R. The object is produced as output
of the function \link{read_BIN2R}.
}
\section{Methods (by generic)}{
\itemize{
\item \code{set_Risoe.BINfileData(ANY)}: A \linkS4class{Risoe.BINfileData} object is normally produced as output of the
function \link{read_BIN2R}. This construction method is intended for internal
usage only.

}}
\section{Slots}{

\describe{
\item{\code{METADATA}}{Object of class "data.frame" containing the meta information for each curve.}

\item{\code{DATA}}{Object of class "list" containing numeric vector with count data.}

\item{\code{.RESERVED}}{Object of class "list" containing list of undocumented raw values for internal use only.}
}}

\note{
\strong{Internal METADATA - object structure}

This structure is compatible with BIN/BINX-files version 03-08, however, it does not follow (in its
sequential arrangement) the manual provided by the manufacturer,
but an own structure accounting for the different versions.

\tabular{rllll}{
\strong{#} \tab \strong{Name} \tab \strong{Data Type} \tab \strong{V} \tab \strong{Description} \cr
\verb{[,1]} \tab \code{ID}  \tab \code{numeric} \tab RLum \tab Unique record ID (same ID as in slot \code{DATA})\cr
\verb{[,2]} \tab \code{SEL} \tab \code{logic} \tab RLum \tab Record selection, not part official BIN-format, triggered by TAG\cr
\verb{[,3]} \tab \code{VERSION} \tab \code{raw} \tab 03-08 \tab BIN-file version number \cr
\verb{[,4]} \tab \code{LENGTH} \tab \code{integer} \tab 03-08 \tab Length of this record\cr
\verb{[,5]} \tab \code{PREVIOUS} \tab \code{integer} \tab 03-08 \tab Length of previous record\cr
\verb{[,6]} \tab \code{NPOINTS} \tab \code{integer} \tab 03-08 \tab Number of data points in the record\cr
\verb{[,7]} \tab \code{RECTYPE} \tab \code{integer} \tab 08 \tab Record type \cr
\verb{[,8]} \tab \code{RUN} \tab \code{integer} \tab 03-08 \tab Run number\cr
\verb{[,9]} \tab \code{SET} \tab \code{integer} \tab 03-08 \tab Set number\cr
\verb{[,10]}  \tab \code{POSITION} \tab  \code{integer} \tab 03-08 \tab Position number\cr
\verb{[,11]} \tab \code{GRAIN} \tab \code{integer} \tab 03-04 \tab Grain number\cr
\verb{[,12]} \tab \code{GRAINNUMBER} \tab \code{integer} \tab 05-08 \tab Grain number\cr
\verb{[,13]} \tab \code{CURVENO} \tab \code{integer} \tab 05-08 \tab Curve number\cr
\verb{[,14]} \tab \code{XCOORD} \tab \code{integer} \tab 03-08 \tab X position of a single grain\cr
\verb{[,15]} \tab \code{YCOORD} \tab \code{integer} \tab 03-08 \tab Y position of a single grain\cr
\verb{[,16]} \tab \code{SAMPLE} \tab \code{factor} \tab 03-08 \tab Sample name\cr
\verb{[,17]} \tab \code{COMMENT} \tab \code{factor} \tab 03-08 \tab Comment name\cr
\verb{[,18]} \tab \code{SYSTEMID} \tab \code{integer} \tab 03-08 \tab Risø system id\cr
\verb{[,19]} \tab \code{FNAME} \tab \code{factor} \tab 05-08 \tab File name (\emph{.bin/}.binx)\cr
\verb{[,20]} \tab \code{USER} \tab \code{factor} \tab 03-08 \tab User name\cr
\verb{[,21]} \tab \code{TIME} \tab \code{character} \tab 03-08 \tab Data collection time (\code{hh-mm-ss})\cr
\verb{[,22]} \tab \code{DATE} \tab \code{factor} \tab 03-08 \tab Data collection date (\code{ddmmyy})\cr
\verb{[,23]} \tab \code{DTYPE} \tab \code{character} \tab 03-08 \tab Data type\cr
\verb{[,24]} \tab \code{BL_TIME} \tab \code{numeric} \tab 03-08 \tab Bleaching time\cr
\verb{[,25]} \tab \code{BL_UNIT} \tab \code{integer} \tab 03-08 \tab Bleaching unit (mJ, J, s, min, h)\cr
\verb{[,26]} \tab \code{NORM1} \tab \code{numeric} \tab 03-08 \tab Normalisation factor (1)\cr
\verb{[,27]} \tab \code{NORM2} \tab \code{numeric} \tab 03-08 \tab Normalisation factor (2)\cr
\verb{[,28]} \tab \code{NORM3} \tab \code{numeric} \tab 03-08 \tab Normalisation factor (3)\cr
\verb{[,29]} \tab \code{BG} \tab \code{numeric} \tab 03-08 \tab Background level\cr
\verb{[,30]} \tab \code{SHIFT} \tab \code{integer} \tab 03-08 \tab Number of channels to shift data\cr
\verb{[,31]} \tab \code{TAG} \tab \code{integer} \tab 03-08 \tab Tag, triggers \code{SEL}\cr
\verb{[,32]} \tab \code{LTYPE} \tab \code{character} \tab 03-08 \tab Luminescence type\cr
\verb{[,33]} \tab \code{LIGHTSOURCE} \tab \code{character} \tab 03-08 \tab Light source\cr
\verb{[,34]} \tab \code{LPOWER} \tab \code{numeric} \tab 03-08 \tab Optical stimulation power\cr
\verb{[,35]} \tab \code{LIGHTPOWER} \tab \code{numeric} \tab 05-08 \tab Optical stimulation power\cr
\verb{[,36]} \tab \code{LOW} \tab \code{numeric} \tab 03-08 \tab Low (temperature, time, wavelength)\cr
\verb{[,37]} \tab \code{HIGH} \tab \code{numeric} \tab 03-08 \tab High (temperature, time, wavelength)\cr
\verb{[,38]} \tab \code{RATE} \tab \code{numeric} \tab 03-08 \tab Rate (heating rate, scan rate)\cr
\verb{[,39]} \tab \code{TEMPERATURE} \tab \code{integer} \tab 03-08 \tab Sample temperature\cr
\verb{[,40]} \tab \code{MEASTEMP} \tab \code{integer} \tab 05-08 \tab Measured temperature\cr
\verb{[,41]} \tab \code{AN_TEMP} \tab \code{numeric} \tab 03-08 \tab Annealing temperature\cr
\verb{[,42]} \tab \code{AN_TIME} \tab \code{numeric} \tab 03-08 \tab Annealing time\cr
\verb{[,43]} \tab \code{TOLDELAY} \tab \code{integer} \tab 03-08 \tab TOL 'delay' channels\cr
\verb{[,44]} \tab \code{TOLON} \tab \code{integer} \tab 03-08 \tab TOL 'on' channels\cr
\verb{[,45]} \tab \code{TOLOFF} \tab \code{integer} \tab 03-08 \tab TOL 'off' channels\cr
\verb{[,46]} \tab \code{IRR_TIME} \tab \code{numeric} \tab 03-08 \tab Irradiation time\cr
\verb{[,47]} \tab \code{IRR_TYPE} \tab \code{integer} \tab 03-08 \tab Irradiation type (alpha, beta or gamma)\cr
\verb{[,48]} \tab \code{IRR_UNIT} \tab \code{integer} \tab 03-04 \tab Irradiation unit (Gy, rad, s, min, h)\cr
\verb{[,49]} \tab \code{IRR_DOSERATE} \tab \code{numeric} \tab 05-08 \tab Irradiation dose rate (Gy/s)\cr
\verb{[,50]} \tab \code{IRR_DOSERATEERR} \tab \code{numeric} \tab 06-08 \tab Irradiation dose rate error (Gy/s)\cr
\verb{[,51]} \tab \code{TIMESINCEIRR} \tab \code{integer} \tab 05-08 \tab Time since irradiation (s)\cr
\verb{[,52]} \tab \code{TIMETICK} \tab \code{numeric} \tab 05-08 \tab Time tick for pulsing (s)\cr
\verb{[,53]} \tab \code{ONTIME} \tab \code{integer} \tab 05-08 \tab On-time for pulsing (in time ticks)\cr
\verb{[,54]} \tab \code{OFFTIME} \tab \code{integer} \tab 03 \tab Off-time for pulsed stimulation (in s) \cr
\verb{[,55]} \tab \code{STIMPERIOD} \tab \code{integer} \tab 05-08 \tab Stimulation period (on+off in time ticks)\cr
\verb{[,56]} \tab \code{GATE_ENABLED} \tab \code{raw} \tab 05-08 \tab PMT signal gating enabled\cr
\verb{[,57]} \tab \code{ENABLE_FLAGS} \tab \code{raw} \tab 05-08 \tab PMT signal gating  enabled\cr
\verb{[,58]} \tab \code{GATE_START} \tab \code{integer} \tab 05-08 \tab Start gating (in time ticks)\cr
\verb{[,59]} \tab \code{GATE_STOP} \tab \code{integer} \tab 05-08 \tab Stop gating (in time ticks), \code{'Gateend'} for version 04, here only GATE_STOP is used\cr
\verb{[,60]} \tab \code{PTENABLED} \tab \code{raw} \tab 05-08 \tab Photon time enabled\cr
\verb{[,61]} \tab \code{DTENABLED} \tab \code{raw} \tab 05-08 \tab PMT dead time correction enabled\cr
\verb{[,62]} \tab \code{DEADTIME} \tab \code{numeric} \tab 05-08 \tab PMT dead time (s)\cr
\verb{[,63]} \tab \code{MAXLPOWER} \tab \code{numeric} \tab 05-08 \tab Stimulation power to 100 percent (mW/cm^2)\cr
\verb{[,64]} \tab \code{XRF_ACQTIME} \tab \code{numeric} \tab 05-08 \tab XRF acquisition time (s)\cr
\verb{[,65]} \tab \code{XRF_HV} \tab \code{numeric} \tab 05-08 \tab XRF X-ray high voltage (V)\cr
\verb{[,66]} \tab \code{XRF_CURR} \tab \code{integer} \tab 05-08 \tab XRF X-ray current (µA)\cr
\verb{[,67]} \tab \code{XRF_DEADTIMEF} \tab \code{numeric} \tab 05-08 \tab XRF dead time fraction\cr
\verb{[,68]} \tab \code{DETECTOR_ID} \tab \code{raw} \tab 07-08 \tab Detector ID\cr
\verb{[,69]} \tab \code{LOWERFILTER_ID} \tab \code{integer} \tab 07-08 \tab Lower filter ID in reader\cr
\verb{[,70]} \tab \code{UPPERFILTER_ID} \tab \code{integer} \tab 07-08 \tab Upper filter ID in reader\cr
\verb{[,71]} \tab \code{ENOISEFACTOR} \tab \code{numeric} \tab 07-08 \tab Excess noise filter, usage unknown \cr
\verb{[,72]} \tab \code{MARKPOS_X1} \tab \code{numeric} \tab 08 \tab Coordinates marker position 1 \cr
\verb{[,73]} \tab \code{MARKPOS_Y1} \tab \code{numeric} \tab 08 \tab Coordinates marker position 1 \cr
\verb{[,74]} \tab \code{MARKPOS_X2} \tab \code{numeric} \tab 08 \tab Coordinates marker position 2 \cr
\verb{[,75]} \tab \code{MARKPOS_Y2} \tab \code{numeric} \tab 08 \tab Coordinates marker position 2 \cr
\verb{[,76]} \tab \code{MARKPOS_X3} \tab \code{numeric} \tab 08 \tab Coordinates marker position 3 \cr
\verb{[,77]} \tab \code{MARKPOS_Y3} \tab \code{numeric} \tab 08 \tab Coordinates marker position 3 \cr
\verb{[,78]} \tab \code{EXTR_START} \tab \code{numeric} \tab 08 \tab usage unknown \cr
\verb{[,79]} \tab \code{EXTR_END} \tab \code{numeric} \tab 08 \tab usage unknown\cr
\verb{[,80]} \tab \code{SEQUENCE} \tab \code{character} \tab 03-04 \tab Sequence name
}
V = BIN-file version (RLum means that it does not depend on a specific BIN-file version)

Note that the \code{Risoe.BINfileData} object combines all values from
different versions from the BIN/BINX-file, reserved bits are skipped, however,
the function \link{write_R2BIN} reset arbitrary reserved bits. Invalid
values for a specific version are set to \code{NA}. Furthermore, the
internal R data types do not necessarily match the required data types for
the BIN-file data import! Data types are converted during data import.\cr

\strong{LTYPE} values

\tabular{rll}{
VALUE \tab TYPE \tab DESCRIPTION \cr
\verb{[0]} \tab \code{TL} \tab: Thermoluminescence \cr
\verb{[1]} \tab \code{OSL} \tab: Optically stimulated luminescence \cr
\verb{[2]} \tab \code{IRSL} \tab: Infrared stimulated luminescence \cr
\verb{[3]} \tab \code{M-IR} \tab: Infrared monochromator scan\cr
\verb{[4]} \tab \code{M-VIS} \tab: Visible monochromator scan\cr
\verb{[5]} \tab \code{TOL} \tab: Thermo-optical luminescence \cr
\verb{[6]} \tab \code{TRPOSL} \tab: Time Resolved Pulsed OSL\cr
\verb{[7]} \tab \code{RIR} \tab: Ramped IRSL\cr
\verb{[8]} \tab \code{RBR} \tab: Ramped (Blue) LEDs\cr
\verb{[9]} \tab \code{USER} \tab: User defined\cr
\verb{[10]} \tab \code{POSL} \tab: Pulsed OSL \cr
\verb{[11]} \tab \code{SGOSL} \tab: Single Grain OSL\cr
\verb{[12]} \tab \code{RL} \tab: Radio Luminescence \cr
\verb{[13]} \tab \code{XRF} \tab: X-ray Fluorescence
}

\strong{DTYPE} values

\tabular{rl}{
VALUE \tab DESCRIPTION \cr
\verb{[0]} \tab Natural \cr
\verb{[1]} \tab N+dose \cr
\verb{[2]} \tab Bleach \cr
\verb{[3]} \tab Bleach+dose \cr
\verb{[4]} \tab Natural (Bleach) \cr
\verb{[5]} \tab N+dose (Bleach) \cr
\verb{[6]} \tab Dose \cr
\verb{[7]} \tab Background
}

\strong{LIGHTSOURCE} values

\tabular{rl}{
VALUE \tab DESCRIPTION \cr
\verb{[0]} \tab None \cr
\verb{[1]} \tab Lamp \cr
\verb{[2]} \tab IR diodes/IR Laser \cr
\verb{[3]} \tab Calibration LED \cr
\verb{[4]} \tab Blue Diodes \cr
\verb{[5]} \tab White light \cr
\verb{[6]} \tab Green laser (single grain) \cr
\verb{[7]} \tab IR laser (single grain) }

\strong{Internal DATA - object structure}

With version 8 of the BIN/BINX file format, slot \verb{@DATA} (byte array \code{DPOINTS}) can
contain two different values:
\enumerate{
\item \code{DPOINTS} (standard for \code{RECTYPE} := (0,1)): is a vector with the length defined
through \code{NPOINTS}. This is the standard for xy-curves since version 03. However,
recorded are only y-values and x-values are calculated during import using
information from, amongst others, \code{LOW} and \code{HIGH}. For instance, a simple TL curve
would store \code{LOW = 0} and \code{HIGH = 400}. This is then becomes the range for the
temperature values. All values in between are interpolated with a length
matching the number of corresponding values stored in \code{DPOINTS}.
}

\emph{Note: The Sequence Editor uses 0 deg. C as the lowest temperature, although
this temperature is usually not available in a laboratory. This, however, does not
mean that the calculated TL curves are invalid, as the heater would only
start the temperature ramp if the ambient temperature is lower than
the target temperature.}
\enumerate{
\item \code{DPOINTS} (\code{RECTYPE} := 128) is contains no count values but information about
the definition of the regions of interest (ROI). Each definition is 504 bytes long.
The number of definitions is defined by \code{NPOINTS} in \verb{@METADATA}. The record
describes basically the geometric features of the regions of interest.
The representation in R is a nested \link{list}.
}

\tabular{rllll}{
\strong{#} \tab \strong{Name} \tab \strong{Data Type} \tab \strong{V} \tab \strong{Description} \cr
\verb{[,1]} \tab \code{NOFPOINTS}  \tab \code{numeric} \tab 08 \tab number of points in the definition (e.g., if the ROI is a rectangle: 4)\cr
\verb{[,2]} \tab \code{USEDFOR}  \tab \code{logical} \tab 08 \tab samples for which the ROI is used for; a maximum of 48 samples are allowed.\cr
\verb{[,3]} \tab \code{SHOWNFOR}  \tab \code{logical} \tab 08 \tab samples for which the ROI is shown for; a maximum of 48 samples are allowed.\cr
\verb{[,4]} \tab \code{COLOR}  \tab \code{numeric} \tab 08 \tab The colour values of the ROI.\cr
\verb{[,5]} \tab \code{X}  \tab \code{numeric} \tab 08 \tab The x coordinates used to draw the ROI geometry (up to 50 points are allowed).\cr
\verb{[,6]} \tab \code{Y}  \tab \code{numeric} \tab 08 \tab The y coordinates used to draw the ROI geometry (up to 50 points are allowed).\cr
}

(information on the BIN/BINX file format are kindly provided by Risø, DTU Nutech)
}
\section{Objects from the Class}{
 Objects can be created by calls of the form
\code{new("Risoe.BINfileData", ...)}.
}

\section{Function version}{
 0.4.1
}

\examples{

showClass("Risoe.BINfileData")

} 

\section{How to cite}{
Kreutzer, S., 2026. Risoe.BINfileData-class(): Class 'Risoe.BINfileData'. Function version 0.4.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Risø DTU, 2013. The Sequence Editor User Manual - Feb 2013 and Risø DTU, 2016.

The Sequence Editor User Manual - February 2016

\url{https://www.fysik.dtu.dk}
}
\seealso{
\link{plot_Risoe.BINfileData}, \link{read_BIN2R},
\link{write_R2BIN}, \link{merge_Risoe.BINfileData},
\link{Risoe.BINfileData2RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
based on information provided by Torben Lapp and Karsten Bracht Nielsen (Risø DTU, Denmark)
, RLum Developer Team}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_Risoe.BINfileData.R
\name{get_Risoe.BINfileData}
\alias{get_Risoe.BINfileData}
\title{General accessor function for RLum S4 class objects}
\usage{
get_Risoe.BINfileData(object, ...)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData} (\strong{required}):
S4 object of class \code{RLum}}

\item{...}{further arguments that one might want to pass to the specific
get function}
}
\value{
Return is the same as input objects as provided in the list
}
\description{
Function calls object-specific get functions for RisoeBINfileData S4 class objects.
}
\details{
The function provides a generalised access point for specific
\linkS4class{Risoe.BINfileData} objects. \cr
Depending on the input object, the corresponding get function will be selected.
Allowed arguments can be found in the documentations of the corresponding
\linkS4class{Risoe.BINfileData} class.
}
\section{Function version}{
 0.1.0
}

\seealso{
\linkS4class{Risoe.BINfileData}
}
\author{
Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2023. get_Risoe.BINfileData(): General accessor function for RLum S4 class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., 2023. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 0.9.21. https://CRAN.R-project.org/package=Luminescence
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R,
%   R/RLum.Data.Image-class.R, R/RLum.Data.Spectrum-class.R
\name{normalise_RLum}
\alias{normalise_RLum}
\alias{normalise_RLum,list-method}
\alias{normalise_RLum,RLum.Analysis-method}
\alias{normalise_RLum,RLum.Data.Curve-method}
\alias{normalise_RLum,RLum.Data.Image-method}
\alias{normalise_RLum,RLum.Data.Spectrum-method}
\title{Normalisation of RLum-class objects}
\usage{
normalise_RLum(object, norm = TRUE, ...)

\S4method{normalise_RLum}{list}(object, norm = TRUE, ...)

\S4method{normalise_RLum}{RLum.Analysis}(object, norm = TRUE, ...)

\S4method{normalise_RLum}{RLum.Data.Curve}(object, norm)

\S4method{normalise_RLum}{RLum.Data.Image}(object, norm = TRUE, global = TRUE)

\S4method{normalise_RLum}{RLum.Data.Spectrum}(object, norm = TRUE)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum}}

\item{norm}{\link{logical} \link{character} (\strong{required}):
if logical, whether curve normalisation should occur; alternatively, one
of \code{"max"} (used with \code{TRUE}), \code{"min"}, \code{"first"}, \code{"last"}, \code{"huot"},
\code{"intensity"} or a positive number (e.g., 2.2).}

\item{...}{further arguments passed to the specific class method}

\item{global}{\link{logical} (*with default): this defines whether the normalisation
is applied globally (same to all) or locally, in which case each frame has its
own normalisation. If \code{global = TRUE} the arguments for \code{norm  = 'first'} and
\code{norm = 'last'} work as expected and consider either the first or the last
frame for the normalisation.

Normalise \linkS4class{RLum.Data.Image} objects to value set via
the argument \code{norm}}
}
\value{
An object of the same type as the input object provided.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the input object, the corresponding
function will be selected. The normalisation is performed in the internal
function \code{.normalise_curve()}.
}
\details{
The \code{norm} argument normalises all count values. The following options are
supported:

\code{norm = TRUE} or \code{norm = "max"}: Curve values are normalised to the highest
count value in the curve

\code{norm = "min"}: Curve values are normalised to the smallest count value
in the curve

\code{norm = "first"}: Curve values are normalised to the first count value.

\code{norm = "last"}: Curve values are normalised to the last count value
(this can be useful in particular for radiofluorescence curves)

\code{norm = "huot"}: Curve values are normalised as suggested by Sébastien Huot
via GitHub:
\deqn{
y = (observed - median(background)) / (\max(observed) - median(background))
}

The background of the curve is defined as the last 20\% of the count values
of a curve.

\code{norm = "intensity"}: Curve values are normalised to the channel length.

\code{norm = 2.2}: Curve values are normalised to a positive number (e.g., 2.2).
}
\section{Functions}{
\itemize{
\item \code{normalise_RLum(list)}: Returns a list of \linkS4class{RLum.Data} objects that had been passed to
\link{normalise_RLum}

\item \code{normalise_RLum(RLum.Analysis)}: Normalisation of \code{RLum.Data} records contained in the input object.

\item \code{normalise_RLum(RLum.Data.Curve)}: Normalise \linkS4class{RLum.Data.Curve} objects to value set via
the argument \code{norm}

\item \code{normalise_RLum(RLum.Data.Image)}: Normalise \linkS4class{RLum.Data.Image} objects to value set via
the argument \code{norm}.

\item \code{normalise_RLum(RLum.Data.Spectrum)}: Normalise \linkS4class{RLum.Data.Spectrum} objects to value set via
the argument \code{norm}

}}
\section{Function version}{
 0.1.3
}

\examples{

## load example data
data(ExampleData.CW_OSL_Curve, envir = environment())

## create RLum.Data.Curve object from this example
curve <-
  set_RLum(
      class = "RLum.Data.Curve",
      recordType = "OSL",
      data = as.matrix(ExampleData.CW_OSL_Curve)
  )

## plot data without and with smoothing
plot_RLum(curve)
plot_RLum(normalise_RLum(curve))

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Data.Spectrum},  \linkS4class{RLum.Data.Image}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. normalise_RLum(): Normalisation of RLum-class objects. Function version 0.1.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.Fading}
\alias{ExampleData.Fading}
\title{Example data for feldspar fading measurements}
\format{
A \link{list} with two elements, each containing a further \link{list} of
\link{data.frame}s containing the data on the fading and equivalent dose measurements:

\describe{
\verb{$fading.data}: A named \link{list} of \link{data.frame}s,
each having three named columns (\verb{LxTx, LxTx.error, timeSinceIrradiation}).\cr
\code{..$IR50}: Fading data of the IR50 signal.\cr
\code{..$IR100}: Fading data of the IR100 signal.\cr
\code{..$IR150}: Fading data of the IR150 signal.\cr
\code{..$IR225}: Fading data of the IR225 signal.\cr

\verb{$equivalentDose.data}: A named of \link{data.frame}s,
each having three named columns (\verb{dose, LxTx, LxTx.error}).\cr
\code{..$IR50}: Equivalent dose measurement data of the IR50 signal.\cr
\code{..$IR100}: Equivalent dose measurement data of the IR100 signal.\cr
\code{..$IR150}: Equivalent dose measurement data of the IR150 signal.\cr
\code{..$IR225}: Equivalent dose measurement data of the IR225 signal.\cr
}
}
\source{
These data were kindly provided by Georgina E. King. Detailed information
on the sample UNIL/NB123 can be found in the reference given below. The raw
data can be found in the accompanying supplementary information.
}
\description{
Example data set for fading measurements of the IR50, IR100, IR150 and
IR225 feldspar signals of sample UNIL/NB123. It further contains regular equivalent dose
measurement data of the same sample, which can be used to apply a
fading correction to.
}
\examples{

## Load example data
data("ExampleData.Fading", envir = environment())

## Get fading measurement data of the IR50 signal
IR50_fading <- ExampleData.Fading$fading.data$IR50
head(IR50_fading)

## Determine g-value and rho' for the IR50 signal
IR50_fading.res <- analyse_FadingMeasurement(IR50_fading)

## Show g-value and rho' results
gval <- get_RLum(IR50_fading.res)
rhop <- get_RLum(IR50_fading.res, "rho_prime")

gval
rhop

## Get LxTx values of the IR50 DE measurement
IR50_De.LxTx <- ExampleData.Fading$equivalentDose.data$IR50

## Calculate the De of the IR50 signal
IR50_De <- fit_DoseResponseCurve(IR50_De.LxTx,
                                mode = "interpolation",
                                fit.method = "EXP")

## Extract the calculated De and its error
IR50_De.res <- get_RLum(IR50_De)
De <- c(IR50_De.res$De, IR50_De.res$De.Error)

## Apply fading correction (age conversion greatly simplified)
IR50_Age <- De / 7.00
IR50_Age.corr <- calc_FadingCorr(IR50_Age, g_value = IR50_fading.res)

}
\references{
King, G.E., Herman, F., Lambert, R., Valla, P.G., Guralnik, B., 2016.
Multi-OSL-thermochronometry of feldspar. Quaternary Geochronology 33, 76-87.
doi:10.1016/j.quageo.2016.01.004

\strong{Details}

\tabular{ll}{
Lab: \tab University of Lausanne \cr
Lab-Code: \tab UNIL/NB123 \cr
Location: \tab Namche Barwa (eastern Himalayas)\cr
Material: \tab Coarse grained (180-212 microns) potassium feldspar \cr
Units: \tab Values are given in seconds \cr
Lab Dose Rate: \tab Dose rate of the beta-source at measurement ca. 0.1335 +/- 0.004 Gy/s \cr
Environmental Dose Rate: \tab 7.00 +/- 0.92 Gy/ka (includes internal dose rate)
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_CobbleDoseRate.R
\name{calc_CobbleDoseRate}
\alias{calc_CobbleDoseRate}
\title{Calculate dose rate of slices in a spherical cobble}
\usage{
calc_CobbleDoseRate(object, conversion = "Guerinetal2011", ...)
}
\arguments{
\item{object}{\link{data.frame} (\strong{required}):
A data frame containing all relevant information for each individual layer.
For the table layout see details.}

\item{conversion}{\link{character} (\emph{with default}): dose rate conversion factors
to use, see \link{BaseDataSet.ConversionFactors} for the accepted values.}

\item{...}{currently not used.}
}
\value{
The function returns an \linkS4class{RLum.Results} object for which
the first element is a \link{matrix} (\code{DataIndividual}) that gives the dose rate
results for each slice for each decay chain individually, for both, the
cobble dose rate and the sediment dose rate. The second element is also a
\link{matrix} (\code{DataComponent}) that gives the total beta and gamma-dose rates
for the cobble and the adjacent sediment for each slice of the cobble.
}
\description{
Calculates the dose rate profile through the cobble based on Riedesel and Autzen (2020).

Corrects the beta dose rate in the cobble for the grain size following results
of Guérin et al. (2012). Sediment beta and gamma dose rates are corrected
for the water content of the sediment using the correction factors of Aitken (1985).
Water content in the cobble is assumed to be 0.
}
\details{
\strong{The input table layout}

\tabular{lll}{
COLUMN \tab DATA TYPE \tab DESCRIPTION\cr
\code{Distance} \tab \code{numeric} \tab distance from the surface of the cobble to the top of each rock slice in mm. The distance for each slice will be listed in this column\cr
\code{DistanceError} \tab \code{numeric} \tab Error on the distance in mm\cr
\code{Thickness} \tab \code{numeric} \tab Thickness of each slice in mm\cr
\code{ThicknessError} \tab \code{numeric} \tab uncertainty of the thickness in mm.\cr
\code{Mineral} \tab \code{character} \tab \code{'FS'} for feldspar, \code{'Q'} for quartz, depending which mineral in the cobble is used for dating\cr
\code{Cobble_K} \tab \code{numeric} \tab K nuclide content in \% of the bulk cobble\cr
\code{Cobble_K_SE} \tab \code{numeric} \tab error on K nuclide content in \% of the bulk cobble\cr
\code{Cobble_Th} \tab \code{numeric} \tab Th nuclide content in ppm of the bulk cobble\cr
\code{Cobble_Th_SE} \tab \code{numeric} \tab error on Th nuclide content in ppm of the bulk cobble\cr
\code{Cobble_U} \tab \code{numeric} \tab U nuclide content in ppm of the bulk cobble\cr
\code{Cobble_U_SE} \tab \code{numeric} \tab error on U nuclide content in ppm of the bulk cobble\cr
\code{GrainSize} \tab \code{numeric} \tab average grain size in µm of the grains used for dating\cr
\code{Density} \tab \code{numeric} \tab Density of the cobble. Default is 2.7 g cm^-3\cr
\code{CobbleDiameter} \tab \code{numeric} \tab Diameter of the cobble in cm.\cr
\code{Sed_K} \tab \code{numeric} \tab K nuclide content in \% of the sediment matrix\cr
\code{Sed_K_SE} \tab \code{numeric} \tab error on K nuclide content in \% of the sediment matrix\cr
\code{Sed_Th} \tab \code{numeric} \tab Th nuclide content in ppm of the sediment matrix\cr
\code{Sed_Th_SE} \tab \code{numeric} \tab error on Th nuclide content in ppm of the sediment matrix\cr
\code{Sed_U} \tab \code{numeric} \tab U nuclide content in ppm of the sediment matrix\cr
\code{Sed_U_SE} \tab \code{numeric} \tab error on U nuclide content in ppm of the sediment matrix\cr
\code{GrainSize_Sed} \tab \code{numeric} \tab average grain size of the sediment matrix\cr
\code{Density_Sed} \tab \code{numeric} \tab average density of the sediment matrix\cr
\code{WaterContent} \tab \code{numeric} \tab mean water content of the sediment matrix in \%\cr
\code{WaterContent_SE} \tab \code{numeric} \tab relative error on water content
}

\strong{Water content}
The water content provided by the user should be calculated according to:

\deqn{(Wet\_weight - Dry\_weight) / Dry\_weight * 100}
}
\section{Function version}{
 0.1.1
}

\examples{
## load example data
data("ExampleData.CobbleData", envir = environment())

## run function
calc_CobbleDoseRate(ExampleData.CobbleData)

} 

\section{How to cite}{
Riedesel, S., Autzen, M., 2026. calc_CobbleDoseRate(): Calculate dose rate of slices in a spherical cobble. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Riedesel, S., Autzen, M., 2020. Beta and gamma dose rate attenuation in rocks and sediment.
Radiation Measurements 133, 106295.
}
\seealso{
\link{convert_Concentration2DoseRate}
}
\author{
Svenja Riedesel, Aberystwyth University (United Kingdom) \cr
Martin Autzen, DTU NUTECH Center for Nuclear Technologies (Denmark)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_SG2MG.R
\name{convert_SG2MG}
\alias{convert_SG2MG}
\title{Converts Single-Grain Data to Multiple-Grain Data}
\usage{
convert_SG2MG(object, write_file = FALSE, ...)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData}, \link{character} (\strong{required}):
\linkS4class{Risoe.BINfileData} object or BIN/BINX-file name.}

\item{write_file}{\link{logical} (\emph{with default}):
whether the output should be written to a file (only considered if \code{object}
is of type character). The multiple grain file will be written to a file
of named after the original one with \code{-SG} appended to it.}

\item{...}{further arguments passed down to \link{read_BIN2R} and
\link{write_R2BIN} if \code{object} is a path to a file.}
}
\value{
A \linkS4class{Risoe.BINfileData} object. If \code{write_file = TRUE} an
the input was a file path, a file is also written out to origin folder.
}
\description{
Conversion of single-grain data to multiple-grain data by adding signals
from grains belonging to one disc (unique pairs of position, set and run).
}
\section{Function version}{
 0.1.1
}

\examples{
## simple run
## (please not that the example is not using SG data)
data(ExampleData.BINfileData, envir = environment())
convert_SG2MG(CWOSL.SAR.Data)

}
\seealso{
\linkS4class{Risoe.BINfileData}, \link{read_BIN2R},
\link{write_R2BIN}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Norbert Mercier, IRAMAT-CRP2A, UMR 5060, CNRS-Université Bordeaux Montaigne (France)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Mercier, N., 2026. convert_SG2MG(): Converts Single-Grain Data to Multiple-Grain Data. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_WodaFuchs2008.R
\name{calc_WodaFuchs2008}
\alias{calc_WodaFuchs2008}
\title{Obtain the equivalent dose using the approach by Woda and Fuchs 2008}
\usage{
calc_WodaFuchs2008(data, breaks = NULL, plot = TRUE, ...)
}
\arguments{
\item{data}{\link{data.frame} \link{vector}, or \linkS4class{RLum.Results} object (\strong{required}):
for \link{data.frame}: either two columns: De (\code{values[,1]}) and De error
(\code{values[,2]}), or one: De (\code{values[,1]}). If a numeric vector or a
single-column data frame is provided, De error is set to \code{NA}.
For plotting multiple data sets, these must be provided as \code{list}
(e.g. \code{list(dataset1, dataset2)}).}

\item{breaks}{\link{numeric}:
Either number or locations of breaks. See \verb{[hist]} for details.
If missing, the number of breaks will be estimated based on the bin width
(as function of median error).}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{Further plot arguments passed to the function.}
}
\description{
The function generates a histogram-like reorganisation of the data, to
assess counts per bin. The log-transformed counts per bin are used to
calculate the second derivative of the data (i.e., the curvature of the
curve) and to find the central value of the bin hosting the distribution
maximum. A normal distribution model is fitted to the counts per bin
data to estimate the dose distribution parameters. The uncertainty of the
model is estimated based on all input equivalent doses smaller that of the
modelled central value.
}
\section{Function version}{
 0.2.0
}

\examples{

## read example data set
data(ExampleData.DeValues, envir = environment())

results <- calc_WodaFuchs2008(
  data = ExampleData.DeValues$CA1,
   xlab = expression(paste(D[e], " [Gy]"))
 )

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., 2026. calc_WodaFuchs2008(): Obtain the equivalent dose using the approach by Woda and Fuchs 2008. Function version 0.2.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Woda, C., Fuchs, M., 2008. On the applicability of the leading edge method to
obtain equivalent doses in OSL dating and dosimetry. Radiation Measurements 43, 26-37.
}
\seealso{
\link{calc_FuchsLang2001}, \link{calc_CentralDose}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_RLum2CSV.R
\name{write_RLum2CSV}
\alias{write_RLum2CSV}
\title{Export RLum-objects to CSV}
\usage{
write_RLum2CSV(
  object,
  path = NULL,
  prefix = "",
  export = TRUE,
  compact = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum} or a \link{list} of \code{RLum} objects (\strong{required}):
objects to be written. Can be a \link{data.frame} if needed internally.}

\item{path}{\link{character} (\emph{optional}):
character string naming folder for the output to be written. If nothing
is provided \code{path} will be set to the working directory.
\strong{Note:} this argument is ignored when \code{export = FALSE}.}

\item{prefix}{\link{character} (\emph{with default}):
optional prefix to name the files. This prefix is valid for all written files}

\item{export}{\link{logical} (\emph{with default}):
enable/disable the file export. If set to \code{FALSE} nothing is written to
the file connection, but a list comprising objects of type \link{data.frame} and \link{matrix}
is returned instead.}

\item{compact}{\link{logical} (\emph{with default}): if \code{TRUE} (default) the output
will be simpler but less comprehensive, that is not all elements in the
objects will be fully broken down. This is in particular useful for writing
\linkS4class{RLum.Results} objects to CSV files, as such objects can be rather complex
and not all information are needed in a CSV file or can be meaningfully
translated to CSV format.}

\item{...}{further arguments that will be passed to the function
\link[utils:write.table]{utils::write.table}. All arguments except the argument \code{file} are supported}
}
\value{
The function returns either a CSV-file (or many of them) or for the
option \code{export == FALSE} a list comprising objects of type \link{data.frame} and \link{matrix}
}
\description{
This function exports \linkS4class{RLum}-objects to CSV-files using the R function
\link[utils:write.table]{utils::write.table}. All \linkS4class{RLum}-objects are supported, but the
export is lossy, i.e. the pure numerical values are exported only. Information
that cannot be coerced to a \link{data.frame} or a \link{matrix} are discarded as well as
metadata.
}
\details{
However, in combination with the implemented import functions, nearly every
supported import data format can be exported to CSV-files, this gives a great
deal of freedom in terms of compatibility with other tools.

\strong{Input is a list of objects}

If the input is a \link{list} of objects all explicit function arguments can be provided
as \link{list}.
}
\section{Function version}{
 0.2.2
}

\examples{

## transform values to a list (and do not write)
data(ExampleData.BINfileData, envir = environment())
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)[[1]]
write_RLum2CSV(object, export = FALSE)

\dontrun{
## write to a temporary file
temp_file <- tempfile(pattern = "output", fileext = ".csv")
write_RLum2CSV(object, temp_file)

## write to the working directory
write_RLum2CSV(object)
}

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data}, \linkS4class{RLum.Results},
\link[utils:write.table]{utils::write.table}
}
\author{
Sebastian Kreutzer, Geography & Earth Science, Aberystwyth University (United Kingdom)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. write_RLum2CSV(): Export RLum-objects to CSV. Function version 0.2.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_FilterCombinations.R
\name{plot_FilterCombinations}
\alias{plot_FilterCombinations}
\title{Plot filter combinations along with the (optional) net transmission window}
\usage{
plot_FilterCombinations(
  filters,
  wavelength_range = 200:1000,
  show_net_transmission = TRUE,
  interactive = FALSE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{filters}{\link{list} (\strong{required}):
a named list of filter data for each filter to be shown.
The filter data itself should be either provided as \link{data.frame} or \link{matrix}
(see details for more options).}

\item{wavelength_range}{\link{numeric} (\emph{with default}):
wavelength range used for the interpolation}

\item{show_net_transmission}{\link{logical} (\emph{with default}):
show net transmission window as polygon.}

\item{interactive}{\link{logical} (\emph{with default}):
enable/disable interactive plots.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments that can be passed to control the plot output.
Supported are \code{main}, \code{xlab}, \code{ylab}, \code{xlim}, \code{ylim}, \code{type}, \code{lty}, \code{lwd}.
For non-common plotting parameters, see the details section.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.

\strong{@data}

\tabular{lll}{
\strong{\code{Object}} \tab \strong{\code{Type}} \tab \strong{\code{Description}} \cr
\code{net_transmission_window} \tab \code{matrix} \tab the resulting net transmission window \cr
\code{OD_total} \tab \code{matrix} \tab the total optical density\cr
\code{filter_matrix} \tab \code{matrix} \tab the filter matrix used for plotting
}

\strong{@info}

\tabular{lll}{
\strong{Object} \tab \strong{Type}\tab  \strong{Description} \cr
\code{call} \tab \link{call} \tab the original function call
}
}
\description{
The function allows to plot transmission windows for different filters.
Missing data for specific wavelengths are automatically interpolated for
the given filter data using function \link{approx}. With that, a standardised
output is reached and a net transmission window can be shown.

\strong{Calculations}

\strong{Net transmission window}

The net transmission window of two filters is approximated by

\deqn{T_{final} = T_{1} * T_{2}}

\strong{Optical density}

\deqn{OD = -\log10(T)}

\strong{Total optical density}

\deqn{OD_{total} = OD_{1} +  OD_{2}}

Please consider using your own calculations for more precise values.

\strong{How to provide input data?}

\emph{CASE 1}

The function expects that all filter values are either of type \code{matrix} or
\code{data.frame} with two columns. The first columns contains the wavelength,
the second the relative transmission (but not in percentage, i.e. the
maximum transmission can be only become 1).

In this case only the transmission window is show as provided. Changes in
filter thickness and reflection factor are not considered.

\emph{CASE 2}

The filter data itself are provided as list element containing a \code{matrix} or
\code{data.frame} and additional information on the thickness of the filter, e.g.,
\code{list(filter1 = list(filter_matrix, d = 2))}.
The given filter data are always considered as standard input and the filter
thickness value is taken into account by

\deqn{Transmission = Transmission^d}

with d given in the same dimension as the original filter data.

\emph{CASE 3}

Same as CASE 2 but additionally a reflection factor P is provided, e.g.,
\code{list(filter1 = list(filter_matrix, d = 2, P = 0.9))}.
The final transmission becomes:

\deqn{Transmission = Transmission^d * P}

\strong{Advanced plotting parameters}

The following further non-common plotting parameters can be passed to the function:

\tabular{lll}{
\strong{\code{Argument}} \tab \strong{\code{Datatype}} \tab \strong{\code{Description}}\cr
\code{legend} \tab \code{logical} \tab enable/disable legend \cr
\code{legend.pos} \tab \code{character} \tab change legend position (\link[graphics:legend]{graphics::legend}) \cr
\code{legend.text} \tab \code{character} \tab same as the argument \code{legend} in (\link[graphics:legend]{graphics::legend}) \cr
\code{net_transmission.col} \tab \code{col} \tab colour of net transmission window polygon \cr
\code{net_transmission.col_lines} \tab \code{col} \tab colour of net transmission window polygon lines \cr
\code{net_transmission.density} \tab  \code{numeric} \tab specify line density in the transmission polygon \cr
\code{grid} \tab \code{list} \tab full list of arguments that can be passed to the function \link[graphics:grid]{graphics::grid}
}

For further modifications standard additional R plot functions are recommend,
e.g., the legend can be fully customised by disabling the standard legend
and using the function \link[graphics:legend]{graphics::legend} instead.
}
\section{Function version}{
 0.3.3
}

\examples{

## (For legal reasons no real filter data are provided)

## Create filter sets
filter1 <- density(rnorm(100, mean = 450, sd = 20))
filter1 <- matrix(c(filter1$x, filter1$y/max(filter1$y)), ncol = 2)
filter2 <- matrix(c(200:799,rep(c(0,0.8,0),each = 200)), ncol = 2)

## Example 1 (standard)
plot_FilterCombinations(filters = list(filter1, filter2))

## Example 2 (with d and P value and name for filter 2)
results <- plot_FilterCombinations(
filters = list(filter_1 = filter1, Rectangle = list(filter2, d = 2, P = 0.6)))
results

## Example 3 show optical density
plot(results$OD_total)

\dontrun{
##Example 4
##show the filters using the interactive mode
plot_FilterCombinations(filters = list(filter1, filter2), interactive = TRUE)
}

}
\seealso{
\linkS4class{RLum.Results}, \link{approx}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. plot_FilterCombinations(): Plot filter combinations along with the (optional) net transmission window. Function version 0.3.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{aplot}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.LxTxData}
\alias{ExampleData.LxTxData}
\alias{LxTxData}
\title{Example Lx/Tx data from CW-OSL SAR measurement}
\format{
A \code{\link{data.frame}} with 4 columns (Dose, LxTx, LxTx.Error, TnTx).
}
\source{
\tabular{ll}{
Lab: \tab Luminescence Laboratory Bayreuth\cr
Lab-Code: \tab BT607\cr
Location: \tab Ostrau (Saxony-Anhalt/Germany)\cr
Material: \tab Middle grain (38-63 \eqn{\mu}m) quartz measured on a Risoe TL/OSL DA-15 reader.
}
}
\description{
LxTx data from a SAR measurement for the package Luminescence.
}
\examples{

## plot Lx/Tx data vs dose [s]
data(ExampleData.LxTxData, envir = environment())
plot(LxTxData$Dose,LxTxData$LxTx)

}
\references{
unpublished data
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_gSGC.R
\name{calc_gSGC}
\alias{calc_gSGC}
\title{Calculate De value based on the gSGC by Li et al., 2015}
\usage{
calc_gSGC(
  data,
  gSGC.type = c("0-250", "0-450"),
  gSGC.parameters = NULL,
  n.MC = 100,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}):
input data  the following columns five columns in the given order:
\code{LnTn}, \code{LnTn.error}, \code{Lr1Tr1}, \code{Lr1Tr1.error}, \code{Dr1}. Column names are
not required.}

\item{gSGC.type}{\link{character} (\emph{with default}):
function parameters to use for the iteration procedure, either \code{"0-450"} or
\code{"0-250"}, as presented in Li et al., 2015 (Table 2). This is ignored if
\code{gSGC.parameters} is set.}

\item{gSGC.parameters}{\link{list} (\emph{optional}):
option to provide own function parameters used for fitting as named list.
Nomenclature follows Li et al., 2015, i.e. \code{list(A, A.error, D0, D0.error, c, c.error, Y0, Y0.error, range)}, where \code{range} is defines the interval
where the function is considered as valid, e.g. \code{range = c(0,250)}.\cr
If set, option \code{gSGC.type} will be ignored.}

\item{n.MC}{\link{integer} (\emph{with default}):
number of Monte Carlo simulation runs for error estimation, see details.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{parameters will be passed to the plot output}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.

\strong{\verb{@data}}\cr
\verb{$ De.value} (\link{data.frame}) \cr
\code{.. $ De}  \cr
\code{.. $ De.error} \cr
\code{.. $ Eta} \cr
\verb{$ De.MC} (\link{list}) contains the matrices from the error estimation.\cr
\verb{$ uniroot} (\link{list}) contains the \link{uniroot} outputs of the De estimations\cr

\strong{\verb{@info}}\cr
`$ call`` (\link{call}) the original function call
}
\description{
The function computes De value and De value error using the
global standardised growth curve (gSGC) assumption proposed by Li et al.,
2015 for OSL dating of sedimentary quartz.
}
\details{
The error of the De value is determined using a Monte Carlo simulation approach.
Solving of the equation is realised using \link{uniroot}.
Large values for \code{n.MC} will significantly increase the computation time.
}
\section{Function version}{
 0.1.3
}

\examples{

results <- calc_gSGC(data = data.frame(
LnTn =  2.361, LnTn.error = 0.087,
Lr1Tr1 = 2.744, Lr1Tr1.error = 0.091,
Dr1 = 34.4))

get_RLum(results, data.object = "De")

} 

\section{How to cite}{
Kreutzer, S., 2026. calc_gSGC(): Calculate De value based on the gSGC by Li et al., 2015. Function version 0.1.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Li, B., Roberts, R.G., Jacobs, Z., Li, S.-H., 2015. Potential of establishing
a 'global standardised growth curve' (gSGC) for optical dating of quartz from sediments.
Quaternary Geochronology 27, 94-104. doi:10.1016/j.quageo.2015.02.011
}
\seealso{
\linkS4class{RLum.Results}, \link{get_RLum}, \link{uniroot}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.RLum.Analysis}
\alias{ExampleData.RLum.Analysis}
\alias{IRSAR.RF.Data}
\title{Example data as \linkS4class{RLum.Analysis} objects}
\format{
\code{IRSAR.RF.Data}: IRSAR.RF.Data on coarse grain feldspar

Each object contains data needed for the given protocol analysis.
}
\source{
\strong{IRSAR.RF.Data}

These data were kindly provided by Tobias Lauer and Matthias Krbetschek.

\tabular{ll}{
Lab: \tab Luminescence Laboratory TU Bergakademie Freiberg\cr
Lab-Code: \tab ZEU/SA1\cr
Location: \tab Zeuchfeld (Zeuchfeld Sandur; Saxony-Anhalt/Germany)\cr
Material: \tab K-feldspar (130-200 \eqn{\mu}m)\cr
Reference: \tab Kreutzer et al. (2014)
}
}
\description{
Collection of different \linkS4class{RLum.Analysis} objects for
protocol analysis.
}
\section{Version}{
 0.1
}

\examples{

##load data
data(ExampleData.RLum.Analysis, envir = environment())

##plot data
plot_RLum(IRSAR.RF.Data)

}
\references{
\strong{IRSAR.RF.Data}

Kreutzer, S., Lauer, T., Meszner, S., Krbetschek, M.R., Faust, D., Fuchs,
M., 2014. Chronology of the Quaternary profile Zeuchfeld in Saxony-Anhalt /
Germany - a preliminary luminescence dating study. Zeitschrift fuer
Geomorphologie 58, 5-26. doi: 10.1127/0372-8854/2012/S-00112
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correct_PMTLinearity.R
\name{correct_PMTLinearity}
\alias{correct_PMTLinearity}
\title{Dead-time (linearity) Correction for Photomultiplier tubes (PMT)}
\usage{
correct_PMTLinearity(object, PMT_pulse_pair_resolution = NULL)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} \linkS4class{RLum.Data.Curve} (\strong{required}):
object with records to correct; can be a \link{list} of such objects}

\item{PMT_pulse_pair_resolution}{\link{numeric} (\emph{with default}): pulse-pair resolution
in ns. Values can be found on the PMT datasheets. If \code{NULL} nothing is done.}
}
\value{
Returns the same type of object type as \code{object}.
}
\description{
Correct dead-time (also linearity) of PMT counts to avoid saturation effects, depending
on pulse-pair-resolution of individual PMTs.
}
\details{
We correct for count linearity using a well-known formula that can be
found for example in the Hamamatsu Photomultiplier handbook
(Hamamatsu Photonics K.K., 2017):

\deqn{N = \frac{M}{1 - Mt}}

where \eqn{N} (in s\eqn{^{-1}}) is the true count rate, \eqn{M} (in s\eqn{^{-1}}) the measured count rate,
and \eqn{t} (in s) the pulse pair resolution.
}
\note{
This function is an adaptation of core from the R package
'OSLdecomposition'.
}
\section{Function version}{
 0.1.1
}

\examples{
o <- set_RLum("RLum.Data.Curve")
correct_PMTLinearity(o, PMT_pulse_pair_resolution = 10)

} 

\section{How to cite}{
Kreutzer, S., Mittelstrass, D., 2026. correct_PMTLinearity(): Dead-time (linearity) Correction for Photomultiplier tubes (PMT). Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Hamamatsu Photonics K.K., 2017. Photomultiplier Tubes: Basics and Applications, 4th edition. ed. Hamamatsu.

Mittelstraß, D., Kreutzer, S., Schmidt, C., 2022. OSLdecomposition: Signal
component analysis for optically stimulated luminescence. \doi{10.32614/CRAN.package.OSLdecomposition}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Dirk Mittelstrass, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\docType{data}
\name{ExampleData.BINfileData}
\alias{ExampleData.BINfileData}
\alias{CWOSL.SAR.Data}
\alias{TL.SAR.Data}
\title{Example data from a SAR OSL and SAR TL measurement for the package
Luminescence}
\format{
\code{CWOSL.SAR.Data}: SAR OSL measurement data

\code{TL.SAR.Data}: SAR TL measurement data

Each class object contains two slots: (a) \code{METADATA} is a \link{data.frame} with
all metadata stored in the BIN file of the measurements and (b) \code{DATA}
contains a list of vectors of the measured data (usually count values).
}
\source{
\strong{CWOSL.SAR.Data}

\tabular{ll}{
Lab: \tab Luminescence Laboratory Bayreuth \cr
Lab-Code: \tab BT607 \cr
Location: \tab Saxony/Germany \cr
Material: \tab Middle grain quartz measured on aluminium cups on a Risø TL/OSL DA-15 reader\cr
Reference: \tab unpublished
}

\strong{TL.SAR.Data}

\tabular{ll}{
Lab: \tab Luminescence Laboratory of Cologne\cr
Lab-Code: \tab LP1_5\cr
Location: \tab Spain\cr
Material: \tab Flint \cr
Setup: \tab Risoe TL/OSL DA-20 reader (Filter: Semrock Brightline, HC475/50, N2, unpolished steel discs) \cr
Reference: \tab unpublished \cr
Remarks: \tab dataset limited to one position
}
}
\description{
Example data from a SAR OSL and TL measurement for package Luminescence
directly extracted from a Risoe BIN-file and provided in an object of type
\linkS4class{Risoe.BINfileData}
}
\note{
Please note that this example data cannot be exported to a BIN-file using the function
\code{writeR2BIN} as it was generated and implemented in the package long time ago. In the meantime
the BIN-file format changed.
}
\section{Version}{
 0.1
}

\examples{

## show first 5 elements of the METADATA and DATA elements in the terminal
data(ExampleData.BINfileData, envir = environment())
CWOSL.SAR.Data@METADATA[1:5,]
CWOSL.SAR.Data@DATA[1:5]

}
\references{
\strong{CWOSL.SAR.Data}: unpublished data

\strong{TL.SAR.Data}: unpublished data
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Results-class.R
\docType{class}
\name{RLum.Results-class}
\alias{RLum.Results-class}
\title{Class \code{"RLum.Results"}}
\description{
Object class contains results data from functions (e.g., \link{analyse_SAR.CWOSL}).
}
\section{Slots}{

\describe{
\item{\code{data}}{Object of class \link{list} containing output data}
}}

\note{
The class is intended to store results from functions to be used by
other functions. The data in the object should always be accessed by the
method \link{get_RLum}.
}
\section{Objects from the Class}{

Objects can be created by calls of the form \code{new("RLum.Results", ...)}.
}

\section{Class version}{
 0.5.2
}

\examples{

showClass("RLum.Results")

##create an empty object from this class
set_RLum(class = "RLum.Results")

##use another function to show how it works

##Basic calculation of the dose rate for a specific date
 dose.rate <-  calc_SourceDoseRate(
   measurement.date = "2012-01-27",
   calib.date = "2014-12-19",
   calib.dose.rate = 0.0438,
   calib.error = 0.0019)

##show object
dose.rate

##get results
get_RLum(dose.rate)

##get parameters used for the calcualtion from the same object
get_RLum(dose.rate, data.object = "parameters")

##alternatively objects can be accessed using S3 generics, such as
dose.rate$parameters

}
\seealso{
\linkS4class{RLum}, \link{plot_RLum}, \link{merge_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.TR_OSL}
\alias{ExampleData.TR_OSL}
\title{Example TR-OSL data}
\format{
One \linkS4class{RLum.Data.Curve} dataset imported using the function \link{read_XSYG2R}

\describe{
\code{ExampleData.TR_OSL}: A single \linkS4class{RLum.Data.Curve} object with the TR-OSL data
}
}
\description{
Single TR-OSL curve obtained by Schmidt et al. (2019) for quartz sample BT729
(origin: Trebgast Valley, Germany, quartz, 90-200 µm, unpublished data).
}
\examples{

##(1) curves
data(ExampleData.TR_OSL, envir = environment())
plot_RLum(ExampleData.TR_OSL)

}
\references{
Schmidt, C., Simmank, O., Kreutzer, S., 2019.
Time-Resolved Optically Stimulated Luminescence of Quartz in the Nanosecond Time Domain.
Journal of Luminescence 213, 376-387.
}
\seealso{
\link{fit_OSLLifeTimes}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_Layout.R
\name{get_Layout}
\alias{get_Layout}
\title{Collection of layout definitions}
\usage{
get_Layout(layout)
}
\arguments{
\item{layout}{\link{character} or \link{list} object (\strong{required}):
name of the layout definition to be returned. If name is provided the
respective definition is returned. One of the following
supported layout definitions is possible: \code{"default"},
\code{"journal.1"}, \code{"small"}, \code{"empty"}.

User-specific layout definitions must be provided as a list object of
predefined structure, see details.}
}
\value{
A list object with layout definitions for plot functions.
}
\description{
This helper function returns a list with layout definitions for homogeneous
plotting.

The easiest way to create a user-specific layout definition is perhaps to
create either an empty or a default layout object and fill/modify the
definitions (\code{user.layout <- get_Layout(data = "empty")}).
}
\section{Function version}{
 0.1
}

\examples{

## read example data set
data(ExampleData.DeValues, envir = environment())

## show structure of the default layout definition
layout.default <- get_Layout(layout = "default")
str(layout.default)

## show colour definitions for Abanico plot, only
layout.default$abanico$colour

## set Abanico plot title colour to orange
layout.default$abanico$colour$main <- "orange"

## create Abanico plot with modofied layout definition
plot_AbanicoPlot(data = ExampleData.DeValues,
                 layout = layout.default)

## create Abanico plot with predefined layout "journal"
plot_AbanicoPlot(data = ExampleData.DeValues,
                 layout = "journal")

}
\author{
Michael Dietze, GFZ Potsdam (Germany)
, RLum Developer Team} 

\section{How to cite}{
Dietze, M., 2026. get_Layout(): Collection of layout definitions. Function version 0.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_LMCurve.R
\name{fit_LMCurve}
\alias{fit_LMCurve}
\title{Non-linear Least Squares Fit for LM-OSL curves}
\usage{
fit_LMCurve(
  object,
  object.bg,
  n.components = 3,
  start_values = NULL,
  input.dataType = c("LM", "pLM"),
  sample_code = "",
  sample_ID = "",
  LED.power = 36,
  LED.wavelength = 470,
  fit.trace = FALSE,
  fit.calcError = FALSE,
  bg.subtraction = c("polynomial", "linear", "channel", "none"),
  verbose = TRUE,
  plot = TRUE,
  plot.BG = FALSE,
  plot.residuals = TRUE,
  plot.contribution = TRUE,
  legend = TRUE,
  legend.pos = "topright",
  method_control = list(),
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
x,y data of measured values (time and counts).}

\item{object.bg}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\emph{optional}):
x,y data of measured values (time and counts) for background subtraction.}

\item{n.components}{\link{integer} (\emph{with default}):
fixed number of components that are to be recognised during fitting
(min = 1, max = 7).}

\item{start_values}{\link{data.frame} (\emph{optional}):
starting values for \code{Im} and \code{xm} parameters in the fit. If set to \code{NULL},
an automatic start value estimation is attempted (see details).}

\item{input.dataType}{\link{character} (\emph{with default}):
alter the plot output depending on the input data: \code{"LM"} or \code{"pLM"} (pseudo-LM).
See: \link{convert_CW2pLM}}

\item{sample_code}{\link{character} (\emph{optional}):
sample code used for the plot and the optional output table (mtext).}

\item{sample_ID}{\link{character} (\emph{optional}):
additional identifier used as column header for the table output.}

\item{LED.power}{\link{numeric} (\emph{with default}):
LED power (max.) used for intensity ramping in mW/cm².
\strong{Note:} This value is used for the calculation of the absolute
photoionisation cross section.}

\item{LED.wavelength}{\link{numeric} (\emph{with default}):
LED wavelength in nm used for stimulation.
\strong{Note:} This value is used for the calculation of the absolute
photoionisation cross section.}

\item{fit.trace}{\link{logical} (\emph{with default}):
traces the fitting process on the terminal.}

\item{fit.calcError}{\link{logical} (\emph{with default}):
calculate 1-sigma error range of components using \link[stats:confint]{stats::confint}.}

\item{bg.subtraction}{\link{character} (\emph{with default}):
specifies method for background subtraction (one of \code{"polynomial"},
\code{"linear"}, \code{"channel"}, or \code{"none"}, see Details). Only considered if
\code{object.bg} is specified.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{plot.BG}{\link{logical} (\emph{with default}):
enable/disable a plot of the background values with the fit used for the
background subtraction.}

\item{plot.residuals}{\link{logical} (\emph{with default}):
enable/disable the plot of the residuals.}

\item{plot.contribution}{\link{logical} (\emph{with default}):
enable/disable the plot of the component contribution.}

\item{legend}{\link{logical} (\emph{with default}):
enable/disable the plot legend.}

\item{legend.pos}{\link{character} (\emph{with default}):
keyword specifying the position of the legend.}

\item{method_control}{\link{list} (\emph{optional}): options to control the output
produced. Currently only the 'export.comp.contrib.matrix' (logical) option
is supported, to enable/disable export of the component contribution
matrix.}

\item{...}{Further arguments that may be passed to the plot output, e.g.
\code{main}, \code{xlab}, \code{xlab}, \code{xlim}, \code{ylim}, \code{cex}, \code{log}.}
}
\value{
Various types of plots are returned. For details see above. Furthermore an
\code{RLum.Results} object is returned with the following structure:

\strong{\verb{@data:}}

\code{.. $data} : \link{data.frame} with fitting results\cr
\code{.. $fit} : nls (\link{nls} object)\cr
\code{.. $component_matrix} : \link{matrix} with numerical xy-values of the single fitted components with the resolution of the input data
\code{.. $component.contribution.matrix} : \link{list} component distribution matrix
(produced only if \code{method_control$export.comp.contrib.matrix = TRUE})

\strong{\verb{info:}}

\code{.. $call} : \link{call} the original function call

Matrix structure for the distribution matrix:

Column 1 and 2: time and \code{rev(time)} values\cr
Additional columns are used for the components, two for each component,
containing I0 and n0. The last columns \code{cont.} provide information on
the relative component contribution for each time interval including the row
sum for this values.
}
\description{
The function determines weighted non-linear least-squares estimates of the
component parameters of an LM-OSL curve (Bulur 1996) for a given number of
components and returns various component parameters. The fitting procedure
uses the Levenberg-Marquardt algorithm as implemented in function \code{nlsLM}
from package \code{minpack.lm}.
}
\details{
\strong{Fitting function}

The function for the fitting has the general
form:

\deqn{y = (\exp(0.5) * Im_1 * x / xm_1) * \exp(-x^2 / (2 * xm_1^2)) + \ldots +
           \exp(0.5) * Im_i * x / xm_i) * \exp(-x^2 / (2 * xm_i^2))}

where \eqn{1 < i < 8}

This function and the equations for the conversion to b (detrapping probability)
and n0 (proportional to initially trapped charge) have been taken from Kitis
et al. (2008):

\deqn{xm_i=\sqrt{\max(t)/b_i}}
\deqn{Im_i = \exp(-0.5) n0/xm_i}

\strong{Background subtraction}

When a background signal is provided with the \code{object.bg} argument, the
user can choose among three methods for background subtraction by setting
the \code{bg.subtraction} argument to one of these:
\itemize{
\item \code{"polynomial"} (default): a polynomial function is fitted using \link{glm}
and the resulting function is used for background subtraction:
\deqn{y = a x^4 + b x^3 + c x^2 + d x + e}
\item \code{"linear"}: a linear function is fitted using \link{glm} and the resulting
function is used for background subtraction:
\deqn{y = a x + b}
\item \code{"channel"}: the measured background signal is subtracted channel-wise
from the measured signal.
\item \code{"none"}: this disables background subtraction even if \code{object.bg} is
provided.
}

\strong{Start values}

The choice of the initial parameters for the \code{nls}-fitting is a crucial
point and the fitting procedure may mainly fail due to ill chosen start
parameters. Here, three options are provided:

\strong{(a)}
If \code{start_values} is not provided by the user, a cheap guess is made
by using the detrapping values found by Jain et al. (2003) for quartz for a
maximum of 7 components. Based on these values, the pseudo start parameters
\code{xm} and \code{Im} are recalculated for the given data set. In all cases, fitting
starts with the ultra-fast component and (depending on \code{n.components})
steps through the following values. If no fit could be achieved, an error
plot (for \code{plot = TRUE}) with the pseudo curve (based on the
pseudo start parameters) is provided. This may give the opportunity to
identify appropriate start parameters visually.

\strong{(b)}
If start values are provided, the function works like a simple \link{nls}
fitting approach.

\strong{Goodness of fit}

The goodness of the fit is given by a pseudo-R² value (pseudo coefficient of
determination). According to Lave (1970), the value is calculated as:

\deqn{pseudoR^2 = 1 - RSS/TSS}

where \eqn{RSS} is the Residual Sum of Squares, and \eqn{TSS} is the Total
Sum of Squares.

\strong{Error of fitted component parameters}

The 1-sigma error for the components is calculated using
the function \link[stats:confint]{stats::confint}. Due to considerable calculation time, this
option is disabled by default. In addition, the error for the components
can be estimated by using internal R functions like \link{summary}. See the
\link{nls} help page for more information.

\emph{For more details on the nonlinear regression in R, see Ritz & Streibig (2008).}
}
\note{
The pseudo-R² may not be the best parameter to describe the goodness
of the fit. The trade off between the \code{n.components} and the pseudo-R²
value currently remains unconsidered.

The function \strong{does not} ensure that the fitting procedure has reached a
global minimum rather than a local minimum! In any case of doubt, the use of
manual start values is highly recommended.
}
\section{Function version}{
 0.3.8
}

\examples{

##(1) fit LM data without background subtraction
data(ExampleData.FittingLM, envir = environment())
fit_LMCurve(values.curve, n.components = 3, log = "x")

##(2) fit LM data with background subtraction and export as JPEG
## -alter file path for your preferred system
##jpeg(file = "~/Desktop/Fit_Output\\%03d.jpg", quality = 100,
## height = 3000, width = 3000, res = 300)
data(ExampleData.FittingLM, envir = environment())
fit_LMCurve(values.curve, object.bg = values.curveBG,
            n.components = 2, log = "x", plot.BG = TRUE)
##dev.off()

##(3) fit LM data with manual start parameters
data(ExampleData.FittingLM, envir = environment())
fit_LMCurve(values.curve,
            object.bg = values.curveBG,
            n.components = 3,
            log = "x",
            start_values = data.frame(Im = c(170,25,400), xm = c(56,200,1500)))

} 

\section{How to cite}{
Kreutzer, S., 2026. fit_LMCurve(): Non-linear Least Squares Fit for LM-OSL curves. Function version 0.3.8. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bulur, E., 1996. An Alternative Technique For Optically
Stimulated Luminescence (OSL) Experiment. Radiation Measurements 26, 5,
701-709.

Jain, M., Murray, A.S., Boetter-Jensen, L., 2003. Characterisation of
blue-light stimulated luminescence components in different quartz samples:
implications for dose measurement. Radiation Measurements 37 (4-5),
441-449.

Kitis, G. & Pagonis, V., 2008. Computerized curve deconvolution analysis for
LM-OSL. Radiation Measurements 43, 737-741.

Lave, C.A.T., 1970. The Demand for Urban Mass Transportation. The Review of
Economics and Statistics 52 (3), 320-323.

Ritz, C. & Streibig, J.C., 2008. Nonlinear Regression with R. R. Gentleman,
K. Hornik, & G. Parmigiani, eds., Springer, p. 150.
}
\seealso{
\link{fit_CWCurve}, \link{plot}, \link{nls}, \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}, \link{get_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{dplot}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_Lamothe2003.R
\name{calc_Lamothe2003}
\alias{calc_Lamothe2003}
\title{Apply fading correction after Lamothe et al., 2003}
\usage{
calc_Lamothe2003(
  object,
  dose_rate.envir,
  dose_rate.source,
  g_value,
  tc = NULL,
  tc.g_value = tc,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} \link{data.frame} (\strong{required}):
Input data for applying the fading correction, can be (1) a \link{data.frame}
with three columns (\code{dose}, \code{LxTx}, \verb{LxTx error}; see details), or (2) an
\linkS4class{RLum.Results} object created by \link{analyse_SAR.CWOSL} or
\link{analyse_pIRIRSequence}.}

\item{dose_rate.envir}{\link{numeric} vector of length 2 (\strong{required}):
Environmental dose rate in mGy/a.}

\item{dose_rate.source}{\link{numeric} vector of length 2 (\strong{required}):
Irradiation source dose rate in Gy/s, which is, according to Lamothe et al.
(2003) De/t.}

\item{g_value}{\link{numeric} vector of length 2 (\strong{required}): g_value in
\%/decade \emph{recalculated at the moment} the equivalent dose was calculated,
i.e. \code{tc} is either similar for the \emph{g}-value measurement \strong{and} the
De measurement or needs be to recalculated (cf. \link{calc_FadingCorr}).
Inserting a normalised g-value, e.g., normalised to 2-days , will
lead to wrong results.}

\item{tc}{\link{numeric} (\emph{optional}): time in seconds between the \strong{end} of
the irradiation and the prompt measurement used in the equivalent dose
estimation (cf. Huntley & Lamothe 2001).
If set to \code{NULL}, it is assumed that \code{tc} is similar for the equivalent
dose estimation and the \emph{g}-value estimation.}

\item{tc.g_value}{\link{numeric} (\emph{with default}):
time in seconds between irradiation and the prompt measurement estimating
the \emph{g}-value. If the \emph{g}-value was normalised to, e.g., 2 days, this time
in seconds (i.e., \code{172800}) should be entered here along with the time used
for the equivalent dose estimation. If nothing is provided the time is set
to \code{tc}, which is the usual case for \emph{g}-values obtained using the SAR
method and \emph{g}-values that had been not normalised to 2 days.
Note: If this value is not \code{NULL} the functions expects a \link{numeric} value for \code{tc}.}

\item{verbose}{\link{logical} (\emph{with default}): enable/disable output to the
terminal.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{...}{further arguments passed to \link{plot_DoseResponseCurve}.}
}
\value{
The function returns an \linkS4class{RLum.Results} object and the graphical
output produced by \link{plot_DoseResponseCurve}.

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$data} \tab \code{data.frame} \tab the fading corrected values \cr
\verb{$fit} \tab \code{nls} \tab the object returned by the dose response curve fitting \cr
}

'\strong{slot:} \strong{\verb{@info}}

The original function call
}
\description{
This function applies the fading correction for the prediction of long-term
fading as suggested by Lamothe et al., 2003. The function basically adjusts
the $L_n/T_n$ values and fits a new dose-response curve using function
\link{fit_DoseResponseCurve}.
}
\details{
\strong{Format of \code{object} if \code{data.frame}}

If \code{object} is a \link{data.frame}, all input values must be of type \link{numeric}.
Dose values are expected in seconds (s) not Gray (Gy). No \code{NA} values are
allowed and the value for the natural dose (first row) should be \code{0}.
Example for three dose points (column names are arbitrary):

\if{html}{\out{<div class="sourceCode">}}\preformatted{ object <- data.frame(
 dose = c(0,25,50),
 LxTx = c(4.2, 2.5, 5.0),
 LxTx_error = c(0.2, 0.1, 0.2))
}\if{html}{\out{</div>}}

\strong{Note on the g-value and \code{tc}}

Users new to R and fading measurements are often confused about what to
enter for \code{tc} and why it may differ from \code{tc.g_value}. By convention
(Huntley & Lamothe 2001), the \code{tc} value is the time elapsed between the
end of the irradiation and the prompt measurement. Usually there is no
reason for having a \code{tc} value different for the equivalent dose measurement
and the \emph{g}-value measurement, except if different equipment was used.
However, if, for instance, the \emph{g}-value measurement sequence was analysed
with the \emph{Analyst} (Duller 2015) and \code{Luminescence} is used to correct for
fading, there is a high chance that the value returned by the \emph{Analyst}
comes normalised to 2-days, even if the \code{tc} values of the measurement were
identical. In such cases, the fading correction cannot be correct until the
\code{tc.g_value} is manually set to 2-days (\code{172800} s) because the function
will internally recalculate values to an identical \code{tc} value.
}
\section{Function version}{
 0.1.1
}

\examples{

##load data
##ExampleData.BINfileData contains two BINfileData objects
##CWOSL.SAR.Data and TL.SAR.Data
data(ExampleData.BINfileData, envir = environment())

##transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

##perform SAR analysis and set rejection criteria
results <- analyse_SAR.CWOSL(
object = object,
signal_integral = 1:2,
background_integral = 900:900,
verbose = FALSE,
plot = FALSE,
onlyLxTxTable = TRUE
)

##run fading correction
results_corr <- calc_Lamothe2003(
  object = results,
  dose_rate.envir =  c(1.676 , 0.180),
  dose_rate.source = c(0.184, 0.003),
  g_value =  c(2.36, 0.6),
  plot = TRUE,
  fit.method = "EXP")


} 

\section{How to cite}{
Kreutzer, S., Mercier, N., 2026. calc_Lamothe2003(): Apply fading correction after Lamothe et al., 2003. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Huntley, D.J., Lamothe, M., 2001. Ubiquity of anomalous fading in K-feldspars and the measurement
and correction for it in optical dating. Canadian Journal of Earth Sciences 38, 1093-1106.

Duller, G.A.T., 2015. The Analyst software package for luminescence data: overview and recent improvements.
Ancient TL 33, 35–42.

Lamothe, M., Auclair, M., Hamzaoui, C., Huot, S., 2003.
Towards a prediction of long-term anomalous fading of feldspar IRSL. Radiation Measurements 37,
493-498.
}
\seealso{
\link{fit_DoseResponseCurve}, \link{plot_DoseResponseCurve},
\link{calc_FadingCorr}, \link{analyse_SAR.CWOSL}, \link{analyse_pIRIRSequence}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_Huntley2006.R
\name{calc_Huntley2006}
\alias{calc_Huntley2006}
\title{Apply the Huntley (2006) model}
\usage{
calc_Huntley2006(
  data,
  LnTn = NULL,
  rhop = NULL,
  ddot = NULL,
  readerDdot = NULL,
  normalise = TRUE,
  fit.method = c("EXP", "GOK"),
  lower.bounds = c(-Inf, -Inf, -Inf, -Inf),
  cores = 1,
  summary = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}):
A \code{data.frame} with one of the following structures:
\itemize{
\item \strong{three columns} with numeric values for dose (s), \code{LxTx} and \code{LxTx}
error, in this order.
\item \strong{two columns} with numeric values for dose (s) and \code{LxTx}, in this order.
This assumes that errors on \code{LxTx} are missing, and a third column will be
automatically attached with an arbitrary 5 \% error on the provided \code{LxTx}
values.
\item \strong{wide table}, i.e. a \link{data.frame} with a number of columns divisible by 3
and where each triplet has the aforementioned column structure.
}

\if{html}{\out{<div class="sourceCode">}}\preformatted{                        (optional)
     | dose (s)| LxTx | LxTx error |
     |  [ ,1]  | [ ,2]|    [ ,3]   |
     |---------|------|------------|
[1, ]|  0      | LnTn | LnTn error | (optional, see arg 'LnTn')
[2, ]|  R1     | L1T1 | L1T1 error |
 ... |    ...  |  ... |     ...    |
[x, ]|  Rx     | LxTx | LxTx error |

}\if{html}{\out{</div>}}

\strong{Note:} The function assumes the first row of the data to be the
\code{Ln/Tn}-value. If you want to provide more than one \code{Ln/Tn}-values, consider
using argument \code{LnTn}.}

\item{LnTn}{\link{data.frame} (\emph{optional}):
A two column data frame with the following structure:

\if{html}{\out{<div class="sourceCode">}}\preformatted{     |  LnTn  |  LnTn error  |
     |  [ ,1] |      [ ,2]   |
     |--------|--------------|
[1, ]| LnTn_1 | LnTn_1 error |
[2, ]| LnTn_2 | LnTn_2 error |
 ... |   ...  |      ...     |
[x, ]| LnTn_x | LnTn_x error |
}\if{html}{\out{</div>}}

The function will calculate a \strong{mean} \code{Ln/Tn}-value and uses either the
standard deviation or the highest individual error, whichever is larger. If
another mean value (e.g. a weighted mean or median) or error is preferred,
this value must be calculated beforehand and used in the first row in the
data frame for argument \code{data}.

\strong{Note:} This argument should \strong{only} be used to provide more than one
\code{Ln/Tn}-value. If you provide \code{LnTn}-values with this argument, the data
frame for the \code{data}-argument \strong{must not} contain any \code{LnTn}-values.}

\item{rhop}{\link{numeric} (\strong{required}):
A vector of length 2 for the density of recombination centres (\eqn{\rho}')
and its error (see Huntley 2006). Note that \eqn{\rho}' must \strong{not} be
provided as the common logarithm. Example: \code{rhop = c(2.92e-06, 4.93e-07)}.}

\item{ddot}{\link{numeric} (\strong{required}):
A vector of length 2 for the environmental dose rate and its error.
Expected unit: Gy/ka. Example: \code{ddot = c(3.7, 0.4)}.}

\item{readerDdot}{\link{numeric} (\strong{required}):
A vector of length 2 for the dose rate of the irradiation source of the OSL
reader and its error.
Expected unit: Gy/s. Example: \code{readerDdot = c(0.08, 0.01)}.}

\item{normalise}{\link{logical} (\emph{with default}): If \code{TRUE} (the default) all
measured and computed \eqn{\frac{L_x}{T_x}} values are normalised by the
pre-exponential factor \code{A} (see details).}

\item{fit.method}{\link{character} (\emph{with default}):
Fit function of the dose response curve. Can either be \code{"EXP"} (default) or
\code{"GOK"}. Note that \code{"EXP"} (single saturating exponential) is the original
function the model after Huntley (2006) and Kars et al. (2008) was
designed to use. The use of a general-order kinetics function (\code{"GOK"})
is an experimental adaptation of the model and should be used
with great care.}

\item{lower.bounds}{\link{numeric} (\emph{with default}):
A vector of length 4 for the values of the lower bounds to be applied
when fitting the models with \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}. In most cases, the
default values (\code{c(-Inf, -Inf, -Inf, -Inf)}) are appropriate for finding
a best fit, but sometimes it may be useful to restrict the lower bounds to
e.g. \code{c(0, 0, 0, 0)}. The values of the vectors are, respectively, for
parameters \code{a}, \code{D0}, \code{c} and \code{d} in that order (parameter \code{d} is ignored
when \code{fit.method = "EXP"}). More details can be found in
\link{fit_DoseResponseCurve}.}

\item{cores}{\link{integer} (\emph{with default}):
The number of cores to use. This will be capped to the number of available
cores if set to too high.}

\item{summary}{\link{logical} (\emph{with default}):
If \code{TRUE} (the default) various parameters provided by the user
and calculated by the model are added as text on the right-hand side of the
plot.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{Further parameters:
\itemize{
\item \code{verbose} \link{logical}: Enable/disable output to the terminal (default = \code{TRUE})
\item \code{n.MC} \link{numeric}: Number of Monte Carlo iterations (default = 10000)
\item \code{cex} \link{numeric}: Scaling of the plot (default = 1)
\item \code{maxiter} \link{numeric}: Number of iteration limits for nls fitting
\item \code{trace} \link{logical}: Enable/disable value tracing the terminal during fitting
\strong{Note:} it is generally advised to have a large number of Monte Carlo
iterations for the results to converge. Decreasing the number of iterations
will often result in unstable estimates.
}

All other arguments are passed to \link{plot} and \link{fit_DoseResponseCurve} (in
particular \code{mode} for the De calculation mode, \code{fit.force_through_origin},
and \code{fit.bounds}).}
}
\value{
An \linkS4class{RLum.Results} object is returned:

Slot: \strong{@data}\cr

\tabular{lll}{
\strong{OBJECT} \tab \strong{TYPE} \tab \strong{COMMENT}\cr
\code{results} \tab \link{data.frame} \tab results of the of Kars et al. 2008 model \cr
\code{data} \tab \link{data.frame} \tab original input data \cr
\code{Ln} \tab \link{numeric} \tab Ln and its error \cr
\code{LxTx_tables} \tab \code{list} \tab A \code{list} of \code{data.frames} containing data on dose,
LxTx and LxTx error for each of the dose response curves.
Note that these \strong{do not} contain the natural \code{Ln} signal, which is provided separately. \cr
\code{fits} \tab \code{list} \tab A \code{list} of \code{nls} objects produced by \link[minpack.lm:nlsLM]{minpack.lm::nlsLM} when fitting the dose response curves \cr
}

Slot: \strong{@info}\cr

\tabular{lll}{
\strong{OBJECT} \tab \strong{TYPE} \tab \strong{COMMENT} \cr
\code{call} \tab \code{call} \tab the original function call \cr
\code{args} \tab \code{list} \tab arguments of the original function call \cr
}
}
\description{
The function calculates the expected sample specific fraction of saturation
based on the model of Huntley (2006), using the approach as implemented
in Kars et al. (2008) or Guralnik et al. (2015).
}
\details{
This function applies the approach described in Kars et al. (2008) or Guralnik et al. (2015),
which are both developed from the model of Huntley (2006), to calculate the expected sample
specific fraction of saturation of a feldspar and also to calculate fading
corrected age using this model. \eqn{\rho}' (\code{rhop}), the density of recombination
centres, is a crucial parameter of this model and must be determined
separately from a fading measurement. The function \link{analyse_FadingMeasurement}
can be used to calculate the sample specific \eqn{\rho}' value.

\strong{Kars et al. (2008) -- Single saturating exponential}

To apply the approach after Kars et al. (2008), use \code{fit.method = "EXP"}.

Firstly, the unfaded \eqn{D_0} value is determined through applying equation 5 of
Kars et al. (2008) to the measured \eqn{\frac{L_x}{T_x}} data as a function of irradiation
time, and fitting the data with a single saturating exponential of the form:

\deqn{\frac{L_x}{T_x}(t^*) = A  \phi(t^*) \{1 - \exp(-\frac{t^*}{D_0}))\}}

where

\deqn{\phi(t^*) = \exp(-\rho' \ln(1.8  \tilde{s} t^*)^3)}

after King et al. (2016) where \eqn{A} is a pre-exponential factor,
\eqn{t^*} (s) is the irradiation time, starting at the mid-point of
irradiation (Auclair et al. 2003) and \eqn{\tilde{s}} (\eqn{3\times10^{15}} s\eqn{^{-1}})
is the athermal frequency factor after Huntley (2006).

Using fit parameters \eqn{A} and \eqn{D_0}, the function then computes a natural dose
response curve using the environmental dose rate, \eqn{\dot{D}} (Gy/s) and equations
\verb{[1]} and \verb{[2]}. Computed \eqn{\frac{L_x}{T_x}} values are then fitted using the
\link{fit_DoseResponseCurve} function and the laboratory measured LnTn can then
be interpolated onto this curve to determine the fading corrected
\eqn{D_e} value, from which the fading corrected age is calculated.

\strong{Guralnik et al. (2015) -- General-order kinetics}

To apply the approach after Guralnik et al. (2015) use \code{fit.method = "GOK"}.

The approach of Guralnik et al. (2015) is very similar to that of
Kars et al. (2008) but, instead of using a single saturating exponential,
the model fits a general-order kinetics function of the form:

\deqn{\frac{L_x}{T_x}(t^*) = A \phi (t^*)(1 - (1 + (\frac{1}{D_0}) t^* c)^{-1/c})}

where \eqn{A}, \eqn{\phi}, \eqn{t^*} and \eqn{D_0} are the same as above and \eqn{c} is a
dimensionless kinetic order modifier (cf. equation 10 in
Guralnik et al., 2015).

\strong{Level of saturation}

The \link{calc_Huntley2006} function also calculates the level of
saturation (\eqn{\frac{n}{N}})
and the field saturation (i.e. athermal steady state, (n/N)_SS) value for
the sample under investigation using the sample specific \eqn{\rho}',
unfaded \eqn{D_0} and \eqn{\dot{D}} values, following the approach of Kars et al. (2008).

The computation is done using 1000 equally-spaced points in the interval
[0.01, 3]. This can be controlled by setting option \code{rprime}, such as
in \code{rprime = seq(0.01, 3, length.out = 1000)} (the default).

\strong{Uncertainties}

Uncertainties are reported at \eqn{1\sigma} and are assumed to be normally
distributed and are estimated using Monte-Carlo re-sampling (\code{n.MC = 10000}
by default) of \eqn{\rho}' and \eqn{\frac{L_x}{T_x}} during dose response
curve fitting, and of \eqn{\rho}' #' in the derivation of (\eqn{n/N}) and
(n/N)_SS.

\strong{Age calculated from 2D0 of the simulated natural DRC}

In addition to the age calculated from the equivalent dose derived from
\eqn{\frac{L_n}{T_n}} projected on the simulated natural dose response curve (DRC), this function
also calculates an age from twice the characteristic saturation dose (\code{D0})
of the simulated natural DRC. This can be a useful information for
(over)saturated samples (i.e., no intersect of \eqn{\frac{L_n}{T_n}} on the natural DRC)
to obtain at least a "minimum age" estimate of the sample. In the console
output this value is denoted by \emph{"Age @2D0 (ka):"}.
}
\note{
This function has BETA status, in particular for the GOK implementation. Please verify
your results carefully.
}
\section{Function version}{
 0.4.7
}

\examples{

## Load example data (sample UNIL/NB123, see ?ExampleData.Fading)
data("ExampleData.Fading", envir = environment())

## (1) Set all relevant parameters
# a. fading measurement data (IR50)
fading_data <- ExampleData.Fading$fading.data$IR50

# b. Dose response curve data
data <- ExampleData.Fading$equivalentDose.data$IR50

## (2) Define required function parameters
ddot <- c(7.00, 0.004)
readerDdot <- c(0.134, 0.0067)

# Analyse fading measurement and get an estimate of rho'.
# Note that the RLum.Results object can be directly used for further processing.
# The number of MC runs is reduced for this example
rhop <- analyse_FadingMeasurement(fading_data, plot = TRUE, verbose = FALSE, n.MC = 10)

## (3) Apply the Kars et al. (2008) model to the data
kars <- calc_Huntley2006(
 data = data,
 rhop = rhop,
 ddot = ddot,
 readerDdot = readerDdot,
 n.MC = 25)

\dontrun{
# You can also provide LnTn values separately via the 'LnTn' argument.
# Note, however, that the data frame for 'data' must then NOT contain
# a LnTn value. See argument descriptions!
LnTn <- data.frame(
 LnTn = c(1.84833, 2.24833),
 nTn.error = c(0.17, 0.22))

LxTx <- data[2:nrow(data), ]

kars <- calc_Huntley2006(
 data = LxTx,
 LnTn = LnTn,
 rhop = rhop,
 ddot = ddot,
 readerDdot = readerDdot,
 n.MC = 25)
}
} 

\section{How to cite}{
King, G.E., Burow, C., Kreutzer, S., Colombo, M., 2026. calc_Huntley2006(): Apply the Huntley (2006) model. Function version 0.4.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Kars, R.H., Wallinga, J., Cohen, K.M., 2008. A new approach towards anomalous fading correction for feldspar
IRSL dating-tests on samples in field saturation. Radiation Measurements 43, 786-790. doi:10.1016/j.radmeas.2008.01.021

Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
from feldspar. Radiation Measurements 81, 224-231.

Huntley, D.J., 2006. An explanation of the power-law decay of luminescence.
Journal of Physics: Condensed Matter 18, 1359-1365. doi:10.1088/0953-8984/18/4/020

King, G.E., Herman, F., Lambert, R., Valla, P.G., Guralnik, B., 2016.
Multi-OSL-thermochronometry of feldspar. Quaternary Geochronology 33, 76-87. doi:10.1016/j.quageo.2016.01.004

\strong{Further reading}

Morthekai, P., Jain, M., Cunha, P.P., Azevedo, J.M., Singhvi, A.K., 2011. An attempt to correct
for the fading in million year old basaltic rocks. Geochronometria 38(3), 223-230.
}
\author{
Georgina E. King, University of Lausanne (Switzerland) \cr
Christoph Burow, University of Cologne (Germany) \cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_IEU.R
\name{calc_IEU}
\alias{calc_IEU}
\title{Apply the internal-external-uncertainty (IEU) model after Thomsen et al.
(2007) to a given De distribution}
\usage{
calc_IEU(data, a, b, interval, decimal.point = 2, plot = TRUE, ...)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with De \code{(data[,1])} and
De error \code{(values[,2])}}

\item{a}{\link{numeric} (\strong{required}):
slope}

\item{b}{\link{numeric} (\strong{required}):
intercept}

\item{interval}{\link{numeric} (\strong{required}):
fixed interval (e.g. 5 Gy) used for iteration of \code{Dbar}, from the mean to
Lowest.De used to create Graph.IEU \verb{[Dbar.Fixed vs Z]}}

\item{decimal.point}{\link{numeric} (\emph{with default}):
number of decimal points for rounding calculations (e.g. 2)}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments (\verb{trace, verbose}).}
}
\value{
Returns a plot (\emph{optional}) and terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following elements:

\item{.$summary}{\link{data.frame} summary of all relevant model results.}
\item{.$data}{\link{data.frame} original input data}
\item{.$args}{\link{list} used arguments}
\item{.$call}{\link{call} the function call}
\item{.$tables}{\link{list} a list of data frames containing all calculation tables}

The output should be accessed using the function \link{get_RLum}.
}
\description{
Function to calculate the IEU De for a De data set.
}
\details{
This function uses the equations of Thomsen et al. (2007). The parameters a
and b are estimated from dose-recovery experiments.
}
\section{Function version}{
 0.1.3
}

\examples{

## load data
data(ExampleData.DeValues, envir = environment())

## apply the IEU model
ieu <- calc_IEU(ExampleData.DeValues$CA1, a = 0.2, b = 1.9, interval = 1)

} 

\section{How to cite}{
Smedley, R.K., Colombo, M., 2026. calc_IEU(): Apply the internal-external-uncertainty (IEU) model after Thomsen et al. (2007) to a given De distribution. Function version 0.1.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Smedley, R.K., 2015. A new R function for the Internal External Uncertainty (IEU) model.
Ancient TL 33, 16-21.

Thomsen, K.J., Murray, A.S., Boetter-Jensen, L. & Kinahan, J.,
2007. Determination of burial dose in incompletely bleached fluvial samples
using single grains of quartz. Radiation Measurements 42, 370-379.
}
\seealso{
\link{plot}, \link{calc_CommonDose}, \link{calc_CentralDose},
\link{calc_FiniteMixture}, \link{calc_FuchsLang2001}, \link{calc_MinDose}
}
\author{
Rachel Smedley, Geography & Earth Sciences, Aberystwyth University (United Kingdom) \cr
Based on an excel spreadsheet and accompanying macro written by Kristina Thomsen. \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remove_SignalBackground.R
\name{remove_SignalBackground}
\alias{remove_SignalBackground}
\title{Remove Signal Background from \code{RLum.Data.Curve} Objects}
\usage{
remove_SignalBackground(
  object,
  object_bg = NULL,
  recordType = NULL,
  clean_up = TRUE
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}): A non-empty \linkS4class{RLum.Analysis} object
with, e.g., OSL/IRSL/TL curves or a \link{list} of such object. If a list is provided
non-conform list elements are silently removed from the \link{list}}

\item{object_bg}{\linkS4class{RLum.Data.Curve}, a \link{list} of such objects, a \link{matrix} or \link{numeric} (\emph{optional}):
Sets the background as a curve that is subtracted from the record types set with \code{recordType}.
If you provide a \link{matrix} or \link{numeric} internally, everything is coerced to a \linkS4class{RLum.Data.Curve} object.
If you desire full freedom, you can construct a \link{list} of \linkS4class{RLum.Data.Curve} objects.}

\item{recordType}{\link{character} (\emph{optional}): provide the \code{recordType} subject to the
background subtraction. Subsequent curve selection uses \link{get_RLum}.
If set to \code{NULL} the record type of highest occurrence will be used. Example: \code{recordType = "TL (UVVIS)"}}

\item{clean_up}{\link{logical} (\emph{with default}): enable/disable background curve removal
after background subtraction. If \code{object_bg} is set, nothing is removed from
the input object as the background is already stored separately.}
}
\value{
Returns an \linkS4class{RLum.Analysis} object or a \link{list} of such objects.
}
\description{
Convenient (background) of luminescence curves using \link{merge_RLum}
}
\details{
The function aims to simplify a frequently encountered task: subtracting curves
or backgrounds from luminescence curves, such as OSL, TL, or RF. The function
presumes that if curves are presented in pairs, for instance, TL - TL,
the second curve represents a background signal that needs to be removed
from the first curve. Following the removal, the background curve is discarded from
the dataset. Alternatively, custom background curves can be provided,
which are then utilised for the subtraction. In essence, the function
utilises the \link{merge_RLum} function but simplifies the selection of pairs and curves.
}
\section{Function version}{
 0.1.0
}

\examples{

## load example dataset
xsyg <- read_XSYG2R(
system.file("extdata/XSYG_file.xsyg", package = "Luminescence"),
fastForward = TRUE,
verbose = FALSE)

## remove constant background from OSL curves
remove_SignalBackground(
object = xsyg,
object_bg = 100,
recordType = "OSL (UVVIS)")

## use a more elaborate examples
## with two TL curves (2nd is background)
xsyg_v1 <- set_RLum("RLum.Analysis", records = c(
 rep(xsyg[[1]]@records[[1]], 2),
 xsyg[[1]]@records[[4]],
 xsyg[[1]]@records[[4]],
 rep(xsyg[[1]]@records[[10]], 2),
 xsyg[[1]]@records[[4]],
 xsyg[[1]]@records[[4]]))

## remove background and strip background
## curves from the object
o <- remove_SignalBackground(
 object = xsyg_v1,
 recordType = "TL (UVVIS)")

}
\seealso{
\link{get_RLum}, \link{merge_RLum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. remove_SignalBackground(): Remove Signal Background from RLum.Data.Curve Objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_RLum2Risoe.BINfileData.R
\name{convert_RLum2Risoe.BINfileData}
\alias{convert_RLum2Risoe.BINfileData}
\title{Converts RLum.Analysis and RLum.Data.Curve objects to RLum2Risoe.BINfileData objects}
\usage{
convert_RLum2Risoe.BINfileData(object, keep.position.number = FALSE)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} or \linkS4class{RLum.Data.Curve} (\strong{required}): input object to
be converted}

\item{keep.position.number}{\link{logical} (\emph{with default}): keeps the original
position number or re-calculate the numbers to avoid doubling}
}
\value{
The function returns a \linkS4class{Risoe.BINfileData} object.
}
\description{
The functions converts \linkS4class{RLum.Analysis} and
\linkS4class{RLum.Data.Curve} objects (or a \link{list} of such objects) to
\linkS4class{Risoe.BINfileData} objects. The function intends to provide a
minimum of compatibility between both formats. The created
\linkS4class{RLum.Analysis} object can be later exported to a BIN-file using
function \link{write_R2BIN}.
}
\note{
The conversion can be never perfect. The \code{RLum} objects may contain information which are
not part of the \linkS4class{Risoe.BINfileData} definition.
}
\section{Function version}{
 0.1.4
}

\examples{

##simple conversion using the example dataset
data(ExampleData.RLum.Analysis, envir = environment())
convert_RLum2Risoe.BINfileData(IRSAR.RF.Data)

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve}, \link{write_R2BIN}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. convert_RLum2Risoe.BINfileData(): Converts RLum.Analysis and RLum.Data.Curve objects to RLum2Risoe.BINfileData objects. Function version 0.1.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{extdata}
\alias{extdata}
\title{Collection of External Data}
\description{
Description and listing of data provided in the folder \code{data/extdata}
}
\details{
The \strong{R} package \code{Luminescence} includes a number of raw data files, which are mostly used in
the example sections of appropriate functions. They are also used internally for testing corresponding
functions using the \code{testthat} package (see files in \verb{tests/testthat/}) to ensure their operational
reliability.

\strong{Accessibility}

If the \strong{R} package \code{Luminescence} is installed correctly the preferred way to access and use these
data from within \strong{R} is as follows:

\code{system.file("extdata/<FILENAME>", package = "Luminescence")}

\strong{Individual file descriptions}

\emph{>>Daybreak_TestFile.DAT.txt<<}

\strong{Type:} raw measurement data \cr
\strong{Device:} Daybreak OSL/TL reader\cr
\strong{Measurement date:} unknown\cr
\strong{Location:} unknown\cr
\strong{Provided by:} unknown\cr
\strong{Related R function(s):} \code{read_Daybreak2R()}\cr
\strong{Reference:} unknown

\emph{>>DorNie_0016.psl<<}

\strong{Type:} raw measurement data \cr
\strong{Device:} SUERC portable OSL reader \cr
\strong{Measurement date:} 19/05/2016 \cr
\strong{Location:} Dormagen-Nievenheim, Germany \cr
\strong{Provided by:} Christoph Burow (University of Cologne) \cr
\strong{Related R function(s):} \code{read_PSL2R()} \cr
\strong{Reference:} unpublished \cr
\strong{Additional information:} Sample measured at an archaeological site near \cr
Dormagen-Nievenheim (Germany) during a practical course on Luminescence dating in 2016. \cr

\emph{>>QNL84_2_bleached.txt}, \emph{QNL84_2_unbleached.txt<<}

\strong{Type:} Test data for exponential fits \cr
\strong{Reference:} Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46. \doi{10.26034/la.atl.1989.150}\cr

\emph{>>STRB87_1_bleached.txt}, \emph{STRB87_1_unbleached.txt<<}

\strong{Type:} Test data for exponential fits \cr
\strong{Reference:} Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46. \doi{10.26034/la.atl.1989.150}

\emph{>>XSYG_file.xsyg}

\strong{Type:} XSYG-file stump \cr
\strong{Info:} XSYG-file with some basic curves to test functions \cr
\strong{Reference:} no reference available
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{BaseDataSet.FractionalGammaDose}
\alias{BaseDataSet.FractionalGammaDose}
\title{Base data set of fractional gamma-dose values}
\format{
A \code{\link{list}} with fractional gamma dose-rate values
sorted by article:

\tabular{ll}{

\code{Aitken1985}: \tab
Fractional gamma-dose values from table H.1
}
}
\source{
Fractional gamma dose values were carefully read from the tables given in the
references above.
}
\description{
Collection of (un-)published fractional gamma dose-rate values to scale the
gamma-dose rate considering layer-to-layer variations in soil radioactivity.
}
\section{Version}{
 0.1
}

\examples{

## Load data
data("BaseDataSet.FractionalGammaDose", envir = environment())

}
\references{
Aitken, M.J., 1985. Thermoluminescence Dating. Academic Press, London.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply_Crosstalk.R
\name{apply_Crosstalk}
\alias{apply_Crosstalk}
\title{Apply crosstalk}
\usage{
apply_Crosstalk(object, n_crosstalk = 0.2)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} or \link{numeric} (\strong{required}):
a numerical vector of length 100, representing one or more measurement
discs ("positions") in a reader.
Each element in the vector represents one grain hole location on a disc.}

\item{n_crosstalk}{\link{numeric} (\emph{with default}):
A single number quantifying the added
crosstalk. Defaults for testing purposes to 0.2. Can be any number, even negative,
but for realistic simulations we suggest something between 0 and 0.25.}
}
\value{
A vector of size 100, with the value at each grain hole location including
simulated crosstalk.
}
\description{
Add crosstalk, evenly spread in rook (top-right-bottom-left)
directions, to all grain whole locations on one measurement discs
(=on position on a measurement wheel in a reader). An
added crosstalk value of as example 0.2 means that 0.2 of the value of
the central grain is added to each grain in the rook directions. This is an additive
action: the central grain itself is not affected by this operation (but will on its
turn increase because of crosstalk from its neighbours).
This function is used for simulations: can this added crosstalk be detected?
}
\details{
If an element in \code{object} is \code{NA}, it is internally set to 0, so it will not
be added.
}
\examples{
## Create artificial disc observation
observations <- set_RLum(class = "RLum.Results",
                         data = list(vn_values = rep(x = c(1,2), each = 50))
                       )
hist(get_RLum(object = observations))

## Add crosstalk (with default set to 0.2), and visualize the difference
## in the resulting histogram.
observations_with_simulated_crosstalk <- apply_Crosstalk(observations)
hist(observations_with_simulated_crosstalk)

} 

\section{How to cite}{
Boer, A.d., Steinbuch, L., 2026. apply_Crosstalk(): Apply crosstalk. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
A novel method to assess crosstalk in single-grain luminescence detection.
Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
}
\author{
Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{BaseDataSet.GrainSizeAttenuation}
\alias{BaseDataSet.GrainSizeAttenuation}
\title{Base dataset for grain size attenuation data by Guérin et al. (2012)}
\source{
Guérin, G., Mercier, N., Nathan, R., Adamiec, G., Lefrais, Y., 2012.
On the use of the infinite matrix assumption and associated concepts:
A critical review. Radiation Measurements, 47, 778-785.
}
\description{
Grain size correction data for beta-dose rates
published by Guérin et al. (2012).

#' @format

A \code{\link{data.frame}} seven columns and sixteen rows. Column headers
are \code{GrainSize}, \code{Q_K}, \code{FS_K}, \code{Q_Th}, \code{FS_Th}, \code{Q_U}, \code{FS_U}.
Grain sizes are quoted in µm (e.g., 20, 40, 60 etc.)
}
\section{Version}{
 0.1.0
}

\examples{

## load data
data("BaseDataSet.GrainSizeAttenuation", envir = environment())

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.LxTxOSLData}
\alias{ExampleData.LxTxOSLData}
\alias{Lx.data}
\alias{Tx.data}
\title{Example Lx and Tx curve data from an artificial OSL measurement}
\format{
Two \code{\link{data.frame}}s containing time and count values.
}
\source{
Arbitrary OSL measurement.
}
\description{
\code{Lx} and \code{Tx} data of continuous wave (CW-) OSL signal curves.
}
\examples{

##load data
data(ExampleData.LxTxOSLData, envir = environment())

##plot data
plot(Lx.data)
plot(Tx.data)

}
\references{
unpublished data
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{BaseDataSet}
\alias{BaseDataSet}
\alias{BaseDataSet.ConversionFactors}
\alias{BaseDataSet.GrainSizeAttenuation}
\alias{BaseDataSet.FractionalGammaDose}
\title{Base datasets}
\format{
\strong{Dose-rate conversion factors}

A \code{\link{list}} with three elements with dose-rate conversion factors
sorted by article and radiation type (alpha, beta, gamma):

\tabular{ll}{

\code{AdamiecAitken1998}: \tab
Conversion factors from Tables 5 and 6 \cr

\code{Cresswelletal2018}: \tab
Conversion factors from Tables 5 and 6 \cr

\code{Guerinetal2011}: \tab
Conversion factors from Tables 1, 2 and 3 \cr

\code{Liritzisetal2013}: \tab
Conversion factors from Tables 1, 2 and 3 \cr
}

\strong{Grain size attenuation data}

A \code{\link{data.frame}} seven columns and sixteen rows. Column headers
are \code{GrainSize}, \code{Q_K}, \code{FS_K}, \code{Q_Th}, \code{FS_Th}, \code{Q_U}, \code{FS_U}.
Grain sizes are quoted in µm (e.g., 20, 40, 60 etc.)

\strong{Fractional gamma dose-rate values}

A \code{\link{list}} with fractional gamma dose-rate values sorted by article:

\tabular{ll}{

\code{Aitken1985}: \tab
Fractional gamma-dose values from table H.1
}
}
\source{
\strong{Dose-rate conversion factors}

All gamma conversion factors were carefully read from the tables given in:

Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update.
Ancient TL 16, 37-46.

Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018.
Dose rate conversion parameters: Assessment of nuclear data.
Radiation Measurements 120, 195-201.

Guérin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion
factors: update. Ancient TL 29, 5-8.

Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
A re-evaluation of radiation dose-rate conversion factors. Mediterranean
Archaeology and Archaeometry 13, 1-15.

\strong{Grain size attenuation data}

Guérin, G., Mercier, N., Nathan, R., Adamiec, G., Lefrais, Y., 2012.
On the use of the infinite matrix assumption and associated concepts:
A critical review. Radiation Measurements 47, 778-785.

\strong{Fractional gamma dose-rate values}

Fractional gamma dose values were carefully read from the tables given in:

Aitken, M.J., 1985. Thermoluminescence Dating. Academic Press, London.
}
\description{
Collection of datasets with published and unpublished data used within the
package.

\describe{
\item{BaseData.ConversionFactors}{Collection of published dose-rate
conversion factors to convert concentrations of radioactive isotopes to
dose rate values}
\item{BaseData.GrainSizeAttenuation}{Grain size attenuation data by Guérin
et al. (2012)}
\item{BaseData.FractionalGammaDose}{Collection of (un-)published fractional
gamma dose-rate values to scale the gamma-dose rate considering layer-to-layer
variations in soil radioactivity}
}
}
\section{Version}{
 0.2.0
}

\examples{
## conversion factors
data("BaseDataSet.ConversionFactors", envir = environment())

## grain size attenuation
data("BaseDataSet.GrainSizeAttenuation", envir = environment())

## fractional gamma dose
data("BaseDataSet.FractionalGammaDose", envir = environment())

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trim_RLum.Data.R
\name{trim_RLum.Data}
\alias{trim_RLum.Data}
\title{Trim Channels of RLum.Data-class Objects}
\usage{
trim_RLum.Data(object, recordType = NULL, trim_range = NULL)
}
\arguments{
\item{object}{\linkS4class{RLum.Data} \linkS4class{RLum.Analysis} (\strong{required}): input object,
can be a \link{list} of objects. Please note that in the latter case the function works
only isolated on each element of the \link{list}.}

\item{recordType}{\link{character} (\emph{optional}): type of the record where the trim
should be applied. If not set, the types are determined automatically and applied
for each record type classes. Can be provided as \link{list}.}

\item{trim_range}{\link{numeric} (\emph{optional}): sets the range of indices to
keep. If only one value is given, this is taken to be the minimum; if two
values are given, then the range is defined between the two values
(inclusive). Any value beyond the second is silently ignored. If nothing
is set (default), then all curves are trimmed to the same maximum length.}
}
\value{
A trimmed object or \link{list} of such objects similar to the input objects
}
\description{
Trim off the number of channels of \linkS4class{RLum.Data} objects of similar record type
on the time domain. This function is useful in cases where objects have different lengths (short/longer
measurement time) but should be analysed jointly by other functions.
}
\details{
The function has two modes of operation:
\enumerate{
\item Single \linkS4class{RLum.Data} objects or a \link{list} of such objects:
the function is applied separately over each object.
\item Multiple curves via \linkS4class{RLum.Analysis} or a \link{list} of such objects:
in this mode, the function first determines the minimum number of channels
for each category of records and then jointly processes them. For instance,
if the object contains one TL curve with 100 channels and two OSL curves
with 100 and 99 channels, respectively, then the minimum would be set to
100 channels for the TL curve and to 99 for the OSL curves. If no further
parameters are applied, the function will shorten all OSL curves to 99
channels, but leave the TL curve untouched.
}
}
\section{Function version}{
 0.2
}

\examples{
## trim all TL curves in the object to channels 10 to 20
data(ExampleData.BINfileData, envir = environment())
temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)

c <- trim_RLum.Data(
object = temp,
recordType = "TL",
trim_range = c(10,20))

plot_RLum.Analysis(
object = c,
combine = TRUE,
subset = list(recordType = "TL"))

## simulate a situation where one OSL curve
## in the dataset has only 999 channels instead of 1000
## all curves should be limited to 999
temp@records[[2]]@data <- temp@records[[2]]@data[-nrow(temp[[2]]@data),]

c <- trim_RLum.Data(object = temp)
nrow(c@records[[4]]@data)


}
\seealso{
\linkS4class{RLum.Data}, \linkS4class{RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. trim_RLum.Data(): Trim Channels of RLum.Data-class Objects. Function version 0.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Data.Curve-class.R
\docType{class}
\name{RLum.Data.Curve-class}
\alias{RLum.Data.Curve-class}
\title{Class \code{"RLum.Data.Curve"}}
\description{
Class for representing luminescence curve data.
}
\section{Slots}{

\describe{
\item{\code{recordType}}{Object of class "character" containing the type of the curve (e.g. "TL" or
"OSL").}

\item{\code{curveType}}{Object of class "character" containing curve type, allowed values are
"measured" or "predefined".}

\item{\code{data}}{Object of class \link{matrix} containing curve x and y data.
'data' can also be of type \code{RLum.Data.Curve} to change object values without
de-constructing the object. For example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{set_RLum(class = 'RLum.Data.Curve',
         data = Your.RLum.Data.Curve,
         recordType = 'never seen before')
}\if{html}{\out{</div>}}

would just change the \code{recordType}. Missing arguments  the value is taken
from the input object in 'data' (which is already an \code{RLum.Data.Curve}
object in this example).}
}}

\note{
The class should only contain data for a single curve. For additional
elements the slot \code{info} can be used (e.g. providing additional heating
ramp curve). Objects from the class \code{RLum.Data.Curve} are produced by other
functions (partly within \linkS4class{RLum.Analysis} objects),
namely: \link{Risoe.BINfileData2RLum.Analysis}, \link{read_XSYG2R}
}
\section{Create objects from this Class}{

Objects can be created by calls of the form
\code{set_RLum(class = "RLum.Data.Curve", ...)}.
}

\section{Class version}{
 0.5.2
}

\examples{

showClass("RLum.Data.Curve")

##set empty curve object
set_RLum(class = "RLum.Data.Curve")

}
\seealso{
\linkS4class{RLum}, \linkS4class{RLum.Data},
\link{plot_RLum}, \link{merge_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_Kars2008.R
\name{calc_Kars2008}
\alias{calc_Kars2008}
\title{Apply the Kars et al. (2008) model (deprecated)}
\usage{
calc_Kars2008(fit.method = "EXP", ...)
}
\arguments{
\item{fit.method}{\link{character} (\emph{with default}):
Fit function of the dose response curve. Can either be \code{EXP} (the default)
or \code{GOK}. Note that \code{EXP} (single saturating exponential) is the original
function the model after Huntley (2006) and Kars et al. (2008) was
designed to use. The use of a general-order kinetics function (\code{GOK})
is an experimental adaption of the model and should only be used
with great care.}

\item{...}{Parameters passed to \link{calc_Huntley2006}.}
}
\value{
An \linkS4class{RLum.Results} object is returned:
}
\description{
A function to calculate the expected sample specific fraction of saturation
following Kars et al. (2008) and Huntley (2006). This function is deprecated
and will eventually be removed. Please use \code{calc_Huntley2006()} instead.
}
\details{
This function applies the approach described in Kars et al. (2008),
developed from the model of Huntley (2006) to calculate the expected sample
specific fraction of saturation of a feldspar and also to calculate fading
corrected age using this model. \eqn{\rho}' (\code{rhop}), the density of recombination
centres, is a crucial parameter of this model and must be determined
separately from a fading measurement. The function \link{analyse_FadingMeasurement}
can be used to calculate the sample specific \eqn{\rho}' value.
}
\note{
\strong{This function is deprecated and will eventually be removed from the package.}
\strong{Please use the function \code{\link[=calc_Huntley2006]{calc_Huntley2006()}} instead}
\strong{(use \code{fit.method = "EXP"} to apply the model after Kars et al., 2008).}
}
\section{Function version}{
 0.4.0
}

\examples{

## Load example data (sample UNIL/NB123, see ?ExampleData.Fading)
data("ExampleData.Fading", envir = environment())

## (1) Set all relevant parameters
# a. fading measurement data (IR50)
fading_data <- ExampleData.Fading$fading.data$IR50

# b. Dose response curve data
data <- ExampleData.Fading$equivalentDose.data$IR50

## (2) Define required function parameters
ddot <- c(7.00, 0.004)
readerDdot <- c(0.134, 0.0067)

# Analyse fading measurement and get an estimate of rho'.
# Note that the RLum.Results object can be directly used for further processing.
# The number of MC runs is reduced for this example
rhop <- analyse_FadingMeasurement(fading_data, plot = TRUE, verbose = FALSE, n.MC = 10)

## (3) Apply the Kars et al. (2008) model to the data
kars <- suppressWarnings(
  calc_Kars2008(data = data,
                rhop = rhop,
                ddot = ddot,
                readerDdot = readerDdot,
                n.MC = 25)
)

} 

\section{How to cite}{
King, G.E., Burow, C., 2023. calc_Kars2008(): Apply the Kars et al. (2008) model (deprecated). Function version 0.4.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., 2023. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 0.9.23. https://CRAN.R-project.org/package=Luminescence
}

\references{
Kars, R.H., Wallinga, J., Cohen, K.M., 2008. A new approach towards anomalous fading correction for feldspar
IRSL dating-tests on samples in field saturation. Radiation Measurements 43, 786-790. doi:10.1016/j.radmeas.2008.01.021

Huntley, D.J., 2006. An explanation of the power-law decay of luminescence.
Journal of Physics: Condensed Matter 18, 1359-1365. doi:10.1088/0953-8984/18/4/020

King, G.E., Herman, F., Lambert, R., Valla, P.G., Guralnik, B., 2016.
Multi-OSL-thermochronometry of feldspar. Quaternary Geochronology 33, 76-87. doi:10.1016/j.quageo.2016.01.004

\strong{Further reading}

Morthekai, P., Jain, M., Cunha, P.P., Azevedo, J.M., Singhvi, A.K., 2011. An attempt to correct
for the fading in million year old basaltic rocks. Geochronometria 38(3), 223-230.
}
\author{
Georgina E. King, University of Bern (Switzerland) \cr
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_Second2Gray.R
\name{convert_Second2Gray}
\alias{convert_Second2Gray}
\title{Converting equivalent dose values from seconds (s) to Gray (Gy)}
\usage{
convert_Second2Gray(
  data,
  dose.rate,
  error.propagation = c("omit", "gaussian", "absolute")
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}):
input values, structure: data (\code{values[,1]}) and data error (\code{values [,2]})
are required.}

\item{dose.rate}{\linkS4class{RLum.Results}, \link{data.frame} or \link{numeric} (\strong{required}):
\code{RLum.Results} needs to be originated from the function \link{calc_SourceDoseRate},
for \code{vector} dose rate in Gy/s and dose rate error in Gy/s.}

\item{error.propagation}{\link{character} (\emph{with default}):
error propagation method used for error calculation (\code{omit}, \code{gaussian} or
\code{absolute}), see details for further information.}
}
\value{
Returns a \link{data.frame} with converted values.
}
\description{
Conversion of absorbed radiation dose in seconds (s) to the SI unit Gray
(Gy) including error propagation. Normally used for equivalent dose data.

Calculation of De values from seconds (s) to Gray (Gy)

\deqn{De [Gy] = De [s] * Dose Rate [Gy/s])}

Provided calculation error propagation methods for error calculation
(with \code{'se'} as the standard error and \code{'DR'} of the dose rate of the beta-source):

\strong{(1) \code{omit}} (default)

\deqn{se(De) [Gy] = se(De) [s] * DR [Gy/s]}

In this case the standard error of the dose rate of the beta-source is
treated as systematic (i.e. non-random), and error propagation is omitted.
However, the error must be considered during calculation of the final age
(cf. Aitken, 1985, pp. 242). This approach can be seen as method (2) (gaussian)
for the case the (random) standard error of the beta-source calibration is
0. Which particular method is requested depends on the situation and cannot
be prescriptive.

\strong{(2) \code{gaussian}} error propagation

\deqn{se(De) [Gy] = \sqrt{(DR [Gy/s] * se(De) [s])^2 + (De [s] * se(DR) [Gy/s])^2}}

Applicable under the assumption that errors of \code{De} and \code{se} are uncorrelated.

\strong{(3) \code{absolute}} error propagation

\deqn{se(De) [Gy]= abs(DR [Gy/s] * se(De) [s]) + abs(De [s] * se(DR) [Gy/s])}

Applicable under the assumption that errors of \code{De} and \code{se} are correlated.
}
\note{
If no or a wrong error propagation method is given, the execution of the function is
stopped. Furthermore, if a \code{data.frame} is provided for the dose rate values is has to
be of the same length as the data frame provided with the argument \code{data}
}
\section{Function version}{
 0.6.0
}

\examples{

##(A) for known source dose rate at date of measurement
## - load De data from the example data help file
data(ExampleData.DeValues, envir = environment())
## - convert De(s) to De(Gy)
convert_Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))


##(B) for source dose rate calibration data
## - calculate source dose rate first
dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
                                  calib.date = "2014-12-19",
                                  calib.dose.rate = 0.0438,
                                  calib.error = 0.0019)
# read example data
data(ExampleData.DeValues, envir = environment())

# apply dose.rate to convert De(s) to De(Gy)
convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate)

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., Fuchs, M.C., 2026. convert_Second2Gray(): Converting equivalent dose values from seconds (s) to Gray (Gy). Function version 0.6.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Aitken, M.J., 1985. Thermoluminescence dating. Academic Press.
}
\seealso{
\link{calc_SourceDoseRate}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany)\cr
Margret C. Fuchs, HZDR, Helmholtz-Institute Freiberg for Resource Technology (Germany)
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_De_Dr.R
\name{combine_De_Dr}
\alias{combine_De_Dr}
\title{Combine Dose Rate and Equivalent Dose Distribution}
\usage{
combine_De_Dr(
  De,
  s,
  Dr,
  int_OD,
  Age_range = c(1, 300),
  outlier_threshold = 0.05,
  outlier_method = "default",
  outlier_analysis_plot = FALSE,
  method_control = list(),
  par_local = TRUE,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{De}{\link{numeric} (\strong{required}):
an equivalent dose sample.}

\item{s}{\link{numeric} (\strong{required}): a vector of measurement errors on the equivalent dose}

\item{Dr}{\link{numeric} (\strong{required}): a dose rate sample}

\item{int_OD}{\link{numeric} (\strong{required}): the intrinsic overdispersion, typically the standard deviation
characterizing a dose-recovery test distribution}

\item{Age_range}{\link{numeric} (\emph{with default}):
age range to be investigated by the algorithm. The larger the value, the
more iterations are needed and the longer it takes. It should not be set
too narrow to cut the algorithm some slack.}

\item{outlier_threshold}{\link{numeric} (\emph{with default}): the required significance level used
for the outlier detection. If set to \code{1}, no outliers are removed. If
\code{outlier_method = "RousseeuwCroux1993"}, the median distance is used as outlier threshold.
Please see details for further information.}

\item{outlier_method}{\link{character} (\emph{with default}):
outlier detection method, either \code{"default"} or \code{"RousseeuwCroux1993"}.
See details.}

\item{outlier_analysis_plot}{\link{logical} (\emph{with default}): enable/disable the
outlier analysis plot. Note: the outlier analysis will happen independently
of the plot output.}

\item{method_control}{\link{list} (\emph{with default}):
named \link{list} of parameters to control \link[rjags:jags.model]{rjags::jags.model}. This can be
used to set the random seed for the MCMC chains (four by default):

\if{html}{\out{<div class="sourceCode">}}\preformatted{method_control = list(inits = list(
  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 1),
  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 2),
  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 3),
  list(.RNG.name = "base::Wichmann-Hill", .RNG.seed = 4))
)
}\if{html}{\out{</div>}}}

\item{par_local}{\link{logical} (\emph{with default}): if set to \code{TRUE} the function uses its
own \link[graphics:par]{graphics::par} settings (which will end in two plots next to each other)}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{a few further arguments to fine-tune the plot output such as
\code{cdf_ADr_quantiles} (\code{TRUE}/\code{FALSE}), \code{legend.pos}, \code{legend} (\code{TRUE}/\code{FALSE})}
}
\value{
The function returns a plot if \code{plot = TRUE} and an \linkS4class{RLum.Results}
object with the following slots:

\verb{@data}\cr
\code{.. $Ages}: a \link{numeric} vector with the modelled ages to be further analysed or visualised\cr
\code{.. $Ages_stats}: a \link{data.frame} with sum HPD, CI 68\% and CI 95\% for the ages \cr
\code{.. $outliers_index}: the index with the detected outliers\cr
\code{.. $cdf_ADr_mean} : empirical cumulative density distribution A * Dr (mean)\cr
\code{.. $cdf_ADr_quantiles} : empirical cumulative density distribution A * Dr (quantiles .025,.975)\cr
\code{.. $cdf_De_no_outlier} : empirical cumulative density distribution of the De with no outliers\cr
\code{.. $cdf_De_initial} : empirical cumulative density distribution of the initial De\cr
\code{.. $mcmc_IAM} : the MCMC list of the Individual Age Model, only of \code{method_control = list(return_mcmc = TRUE)} otherwise \code{NULL}\cr
\code{.. $mcmc_BCAM} : the MCMC list of the Bayesian Central Age Model, only of \code{method_control = list(return_mcmc = TRUE)} otherwise \code{NULL}\cr

\verb{@info}\cr
\code{.. $call}: the original function call\cr
\code{.. $model_IAM}: the BUGS model used to derive the individual age\cr
\code{.. $model_BCAM}: the BUGS model used to calculate the Bayesian Central Age\cr
}
\description{
A Bayesian statistical analysis of OSL age requiring dose rate sample.
Estimation contains a preliminary step for detecting outliers in the equivalent
dose sample.
}
\details{
\strong{Outlier detection}

Two different outlier detection methods are implemented (full details are given
in the cited literature).
\enumerate{
\item The \emph{default} and recommend method, uses quantiles to compare prior and
posterior distributions of the individual variances of the equivalent doses.
If the corresponding quantile in the corresponding posterior distribution is larger
than the quantile in the prior distribution, the value is marked
as outlier (cf. Galharret et al., preprint)
\item The alternative method employs the method suggested by Rousseeuw and Croux (1993)
using the absolute median distance.
}

\strong{Parameters available for \code{method_control}}

The parameters listed below are used to granular control Bayesian modelling using
\link[rjags:rjags-package]{rjags::rjags}. Internally the functions \code{.calc_IndividualAgeModel()} and
\code{.calc_BayesianCentraAgelModel()}. The parameter settings affect both models.
Note: \code{method_control} expects a \strong{named} list of parameters

\tabular{llll}{
\strong{PARAMETER} \tab \strong{TYPE} \tab \strong{DEFAULT} \tab \strong{REMARKS} \cr
\code{variable.names_IAM} \tab \link{character} \tab \code{c('A', 'a', 'sig_a')} \tab variables names to be monitored in the modelling process using the internal function \code{.calc_IndividualAgeModel()}\cr
\code{variable.names_BCAM} \tab \link{character} \tab \code{c('A', 'D_e')} \tab variables names to be monitored in the modelling process using the internal function \code{.calc_BayesianCentraAgelModel()}\cr
\code{n.chains} \tab \link{integer} \tab \code{4} \tab number of MCMC chains\cr
\code{n.adapt} \tab \link{integer} \tab \code{1000} \tab number of iterations for the adaptation\cr
\code{n.iter} \tab \link{integer} \tab \code{5000} \tab number of iterations to monitor cf. \link[rjags:coda.samples]{rjags::coda.samples}\cr
\code{thin} \tab \link{numeric} \tab \code{1} \tab thinning interval for the monitoring cf. \link[rjags:coda.samples]{rjags::coda.samples}\cr
\code{diag} \tab \link{logical} \tab \code{FALSE} \tab additional terminal convergence diagnostic.
\code{FALSE} if \code{verbose = FALSE}\cr
\code{progress.bar} \tab \link{logical} \tab \code{FALSE} \tab enable/disable progress bar. \code{FALSE} if \code{verbose = FALSE}\cr
\code{quiet} \tab \link{logical} \tab \code{TRUE} \tab silence terminal output. Set to \code{TRUE} if \code{verbose = FALSE}\cr
\code{return_mcmc}\tab \link{logical} \tab \code{FALSE} \tab return additional MCMC diagnostic information\cr
}
}
\section{Function version}{
 0.1.0
}

\examples{
## set parameters
Dr <- stats::rlnorm (1000, 0, 0.3)
De <- 50*sample(Dr, 50, replace = TRUE)
s <- stats::rnorm(50, 10, 2)

## run modelling
## note: modify parameters for more realistic results
\dontrun{
results <- combine_De_Dr(
 Dr = Dr,
 int_OD = 0.1,
 De,
 s,
 Age_range = c(0,100),
 method_control = list(
  n.iter = 100,
  n.chains = 1))

## show models used
writeLines(results@info$model_IAM)
writeLines(results@info$model_BCAM)
}

} 

\section{How to cite}{
Philippe, A., Galharret, J., Mercier, N., Kreutzer, S., 2026. combine_De_Dr(): Combine Dose Rate and Equivalent Dose Distribution. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Mercier, N., Galharret, J.-M., Tribolo, C., Kreutzer, S., Philippe, A., (2022).
Luminescence age calculation through Bayesian convolution of equivalent dose
and dose-rate distributions: the De_Dr model. Geochronology 4, 297-310.
\doi{10.5194/gchron-4-297-2022}

Galharret, J-M., Philippe, A., Mercier, N. (2021). Detection of outliers
with a Bayesian hierarchical model: application to the single-grain
luminescence dating method. Electronic Journal of Applied Statistics 14 (2),
318-338. \doi{10.1285/i20705948v14n2p318}

\strong{Further reading}

Rousseeuw, P.J., Croux, C., 1993. Alternatives to the median absolute deviation.
Journal of the American Statistical Association 88, 1273–1283. \doi{10.2307/2291267}

Rousseeuw, P.J., Debruyne, M., Engelen, S., Hubert, M., 2006. Robustness and outlier detection in chemometrics.
Critical Reviews in Analytical Chemistry 36, 221–242. \doi{10.1080/10408340600969403}
}
\seealso{
\link{plot_OSLAgeSummary}, \link[rjags:rjags-package]{rjags::rjags}, \link[mclust:mclust-package]{mclust::mclust-package}
}
\author{
Anne Philippe, Université de Nantes (France),
Jean-Michel Galharret, Université de Nantes (France),
Norbert Mercier, IRAMAT-CRP2A, Université Bordeaux Montaigne (France),
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}
\keyword{distribution}
\keyword{dplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_MinDose.R
\name{calc_MinDose}
\alias{calc_MinDose}
\title{(Un-)logged minimum age model (MAM) after Galbraith et al. (1999)}
\usage{
calc_MinDose(
  data,
  sigmab,
  log = TRUE,
  par = 3,
  bootstrap = FALSE,
  init.values = NULL,
  level = 0.95,
  log.output = FALSE,
  plot = TRUE,
  multicore = FALSE,
  ...
)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with \code{(data[, 1])} and De error \code{(data[, 2])}.}

\item{sigmab}{\link{numeric} (\strong{required}):
additional spread in De values, representing the expected overdispersion in
the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
\strong{Note:} For the logged model (\code{log = TRUE}) this value must be
a fraction, e.g. 0.2 (= 20 \%). If the un-logged model is used (\code{log = FALSE}),
\code{sigmab} must be provided in the same absolute units of the De values (seconds or Gray).
See details.}

\item{log}{\link{logical} (\emph{with default}):
whether the logged minimum dose model should be fit to De data (\code{TRUE} by
default).}

\item{par}{\link{numeric} (\emph{with default}):
number of parameters in the minimum age model, either 3 (default) or 4.}

\item{bootstrap}{\link{logical} (\emph{with default}):
apply the recycled bootstrap approach of Cunningham & Wallinga (2012).}

\item{init.values}{\link{numeric} (\emph{optional}):
a named list with starting values for \code{gamma}, \code{sigma}, \code{p0} and \code{mu}
(e.g. \code{list(gamma=100, sigma=1.5, p0=0.1, mu=100)}). If no values are
provided, reasonable values will be estimated from the data.
\strong{Note:} the initial values must always be given in the absolute units.
If a logged model is applied (\code{log = TRUE}), the provided \code{init.values}
are automatically log-transformed.}

\item{level}{\link{logical} (\emph{with default}):
the confidence level required (defaults to 0.95).}

\item{log.output}{\link{logical} (\emph{with default}):
if \code{TRUE} and \code{log = TRUE}, the console output will also show the logged
values of the final parameter estimates and confidence intervals.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{multicore}{\link{logical} (\emph{with default}):
parallelize the computation of the bootstrap by creating a multicore cluster
(only considered if \code{bootstrap = TRUE}). By default, it uses all available
logical CPU cores, but this can be changed with the \code{cores} argument. Note
that this feature is experimental and may not work on all systems.}

\item{...}{(\emph{optional}) further arguments for bootstrapping
(\code{bs.M}, \code{bs.N}, \code{bs.h}, \code{sigmab.sd}). See details for their usage.
Further arguments are
\itemize{
\item \code{verbose}: enable/disable output to the terminal
\item \code{debug}: enable/disable extended console output
\item \code{cores}: number of cores to be used when \code{multicore=TRUE}
}}
}
\value{
Returns a plot (\emph{optional}) and terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following elements:

\item{$summary}{\link{data.frame} summary of all relevant model results.}
\item{$data}{\link{data.frame} original input data}
\item{$args}{\link{list} used arguments}
\item{$call}{\link{call} the function call}
\item{$mle}{\link[bbmle:mle2]{bbmle::mle2} object containing the maximum log likelihood functions for all parameters}
\item{$BIC}{\link{numeric} BIC score from the non-bootstrapped model}
\item{$confint}{\link{data.frame} confidence intervals for all parameters}
\item{$profile}{\link[stats:profile]{stats::profile} the log likelihood profiles}
\item{$bootstrap}{\link{list} bootstrap results}

The output should be accessed using the function \link{get_RLum}.
}
\description{
Function to fit the (un-)logged three or four parameter minimum dose model
(MAM-3/4) to De data.

\strong{Parameters}

This model has four parameters:
\tabular{rl}{
\code{gamma}: \tab minimum dose on the log scale \cr
\code{mu}: \tab mean of the non-truncated normal distribution \cr
\code{sigma}: \tab spread in ages above the minimum \cr
\code{p0}: \tab proportion of grains at gamma \cr }

If \code{par=3} (default) the 3-parameter minimum age model is applied,
where \code{gamma=mu}. For \code{par=4} the 4-parameter model is applied instead.

\strong{(Un-)logged model}

In the original version of the minimum dose model, which is applied if
\code{log = TRUE} (default), the basic data are the natural
logarithms of the De estimates and relative standard errors of the De
estimates. The value for \code{sigmab} must be provided as a ratio
(e.g, 0.2 for 20 \%).

If \code{log=FALSE}, the modified un-logged model will be applied instead. This
has essentially the same form as the original version, but  \code{gamma} and
\code{sigma} are in Gy and \code{gamma} becomes the minimum true dose in the
population.
\strong{Note:} the un-logged model requires \code{sigmab} to be in the same
absolute unit as the provided De values (seconds or Gray).

While the original (logged) version of the minimum dose
model may be appropriate for most samples (i.e. De distributions), the
modified (un-logged) version is specially designed for modern-age and young
samples containing negative, zero or near-zero De estimates (Arnold et al.
2009, p. 323).

\strong{Initial values and boundaries}

The log-likelihood calculations use the \link{nlminb} function for box-constrained
optimisation using PORT routines.  Accordingly, initial values for the four
parameters can be specified via \code{init.values}. If no values are
provided for \code{init.values}, reasonable starting values are estimated
from the input data.  If the final estimates of \emph{gamma}, \emph{mu},
\emph{sigma} and \emph{p0} are totally off target, consider providing custom
starting values via \code{init.values}.

The boundaries for individual model parameters need not be specified
explicitly. To override the default boundary values, provide arguments
\code{gamma.lower}, \code{gamma.upper}, \code{sigma.lower}, \code{sigma.upper},
\code{p0.lower}, \code{p0.upper}, \code{mu.lower} and \code{mu.upper}.

\strong{Bootstrap}

When \code{bootstrap=TRUE} the function applies the bootstrapping method as
described in Cunningham & Wallinga (2012). The minimum age model produces
\code{bs.M} first-level bootstrap replicates (1000 by default) and \code{bs.N}
second-level replicates (\code{3 * bs.M} by default), with an uncertainty
\code{sigma.sd} on the overdispersion parameter (0.04 by default). These values
can be changed by using arguments \code{bs.M}, \code{bs.N} and \code{sigmab.sd},
respectively. The bandwidth of the kernel density estimate can be specified
with \code{bs.h}. By default, this is calculated as:

\deqn{h = (2*\sigma_{DE})/\sqrt{n}}

\strong{Multicore support}

Parallel processing can be enabled by setting \code{multicore = TRUE}. By default,
the number of logical CPU cores is detected automatically, but it can be
overridden with the \code{cores} argument. Multicore processing is supported
only when \code{bootstrap = TRUE}, and spawns one R process per core to compute
MAM estimates for each of the \code{N * M} bootstrap replicates. Note that this
feature is experimental and may not work on all systems. Performance gains
grow with the number of bootstrap replicates, but each additional core (R
process) increases memory usage. When memory is insufficient, overall
performance can degrade severely.

\strong{Likelihood profiles}

The likelihood profiles are generated and plotted by the \code{bbmle} package.
The profile likelihood plots look different to ordinary profile likelihood as

"\verb{[...]} the plot method for likelihood profiles displays the square root of
the deviance difference (twice the difference in negative log-likelihood from
the best fit), so it will be V-shaped for cases where the quadratic approximation
works well \verb{[...]}." (Bolker 2016).

For more details on the profile likelihood
calculations and plots please see the vignettes of the \code{bbmle} package
(also available here: \url{https://CRAN.R-project.org/package=bbmle}.
}
\note{
The default starting values for \emph{gamma}, \emph{mu}, \emph{sigma}
and \emph{p0} may only be appropriate for some De data sets and may need to
be changed for other data. This is especially true when the un-logged
version is applied. \cr
Also note that all R warning messages are suppressed
when running this function. If the results seem odd consider re-running the
model with \code{debug=TRUE} which provides extended console output and
forwards all internal warning messages.
}
\section{Function version}{
 0.5.0
}

\examples{

## Load example data
data(ExampleData.DeValues, envir = environment())

# (1) Apply the minimum age model with minimum required parameters.
# By default, this will apply the un-logged 3-parameter MAM.
calc_MinDose(data = ExampleData.DeValues$CA1, sigmab = 0.1)

\dontrun{
# (2) Re-run the model, but save results to a variable and turn
# plotting of the log-likelihood profiles off.
mam <- calc_MinDose(
 data = ExampleData.DeValues$CA1,
 sigmab = 0.1,
 plot = FALSE)

# Show structure of the RLum.Results object
mam

# Show summary table that contains the most relevant results
res <- get_RLum(mam, "summary")
res

# Plot the log likelihood profiles retroactively, because before
# we set plot = FALSE
plot_RLum(mam)

# Plot the dose distribution in an abanico plot and draw a line
# at the minimum dose estimate
plot_AbanicoPlot(data = mam,
                 main = "3-parameter Minimum Age Model",
                 polygon.col = "none",
                 hist = TRUE,
                 rug = TRUE,
                 summary = c("n", "mean", "mean.weighted", "median", "in.ci"),
                 centrality = res$de,
                 grid.col = "none",
                 line.col = "red",
                 line.label = paste0(round(res$de, 1), "\U00B1",
                                     round(res$de_err, 1), " Gy"),
                 bw = 0.1,
                 ylim = c(-25, 18),
                 summary.pos = "topleft",
                 mtext = bquote("Parameters: " ~
                                  sigma[b] == .(get_RLum(mam, "args")$sigmab) ~ ", " ~
                                  gamma == .(round(log(res$de), 1)) ~ ", " ~
                                  sigma == .(round(res$sig, 1)) ~ ", " ~
                                  rho == .(round(res$p0, 2))))

# (3) Run the minimum age model with bootstrap
# NOTE: Bootstrapping is computationally intensive
# (3.1) run the minimum age model with default values for bootstrapping
calc_MinDose(data = ExampleData.DeValues$CA1,
             sigmab = 0.15,
             bootstrap = TRUE)

# (3.2) Bootstrap control parameters
mam <- calc_MinDose(data = ExampleData.DeValues$CA1,
                    sigmab = 0.15,
                    bootstrap = TRUE,
                    bs.M = 300,
                    bs.N = 500,
                    bs.h = 4,
                    sigmab.sd = 0.06,
                    plot = FALSE)

# Plot the results
plot_RLum(mam)

# save bootstrap results in a separate variable
bs <- get_RLum(mam, "bootstrap")

# show structure of the bootstrap results
str(bs, max.level = 2, give.attr = FALSE)

# print summary of minimum dose and likelihood pairs
summary(bs$pairs$gamma)

# Show polynomial fits of the bootstrap pairs
bs$poly.fits$poly.three

# Plot various statistics of the fit using the generic plot() function
par(mfcol=c(2,2))
plot(bs$poly.fits$poly.three, ask = FALSE)

# Show the fitted values of the polynomials
summary(bs$poly.fits$poly.three$fitted.values)
}

} 

\section{How to cite}{
Burow, C., Colombo, M., 2026. calc_MinDose(): (Un-)logged minimum age model (MAM) after Galbraith et al. (1999). Function version 0.5.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Arnold, L.J., Roberts, R.G., Galbraith, R.F. & DeLong, S.B.,
2009. A revised burial dose estimation procedure for optical dating of young
and modern-age sediments. Quaternary Geochronology 4, 306-325.

Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for mixed fission
track ages. Nuclear Tracks Radiation Measurements 4, 459-470.

Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley, J.M.,
1999. Optical dating of single grains of quartz from Jinmium rock shelter,
northern Australia. Part I: experimental design and statistical models.
Archaeometry 41, 339-364.

Galbraith, R.F., 2005. Statistics for
Fission Track Analysis, Chapman & Hall/CRC, Boca Raton.

Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error
calculation and display in OSL dating: An overview and some recommendations.
Quaternary Geochronology 11, 1-27.

Olley, J.M., Roberts, R.G., Yoshida, H., Bowler, J.M., 2006. Single-grain optical dating of grave-infill
associated with human burials at Lake Mungo, Australia. Quaternary Science
Reviews 25, 2469-2474.

\strong{Further reading}

Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
(De) distributions: Implications for OSL dating of sediment mixtures.
Quaternary Geochronology 4, 204-230.

Bolker, B., 2016. Maximum likelihood estimation analysis with the bbmle package.
In: Bolker, B., R Development Core Team, 2016. bbmle: Tools for General Maximum Likelihood Estimation.
R package version 1.0.18. \url{https://CRAN.R-project.org/package=bbmle}

Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
assessment of procedures for estimating burial dose. Quaternary Science
Reviews 25, 2475-2502.

Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
Quaternary Geochronology 12, 98-106.

Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
of optical dating of fluvial deposits.  Quaternary Geochronology 1, 109-120.

Rodnight, H., 2008. How many equivalent dose values are needed to
obtain a reproducible distribution?. Ancient TL 26, 3-10.
}
\seealso{
\link{calc_CentralDose}, \link{calc_CommonDose}, \link{calc_FiniteMixture},
\link{calc_FuchsLang2001}, \link{calc_MaxDose}
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany) \cr
Based on a rewritten S script of Rex Galbraith, 2010 \cr
The bootstrap approach is based on a MATLAB script by Alastair Cunningham,
who helped with implementation and cross-checking.
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/template_DRAC.R
\name{template_DRAC}
\alias{template_DRAC}
\title{Create a DRAC input data template (v1.2)}
\usage{
template_DRAC(nrow = 1L, preset = NULL, file_input = NULL, notification = TRUE)
}
\arguments{
\item{nrow}{\link{integer} (\emph{with default}):
specifies the number of rows of the template (i.e., the number of data
sets you want to submit).}

\item{preset}{\link{character} (\emph{optional}):
By default, all values of the template are set to \code{NA}, which means that
the user needs to fill in \strong{all} data first before submitting to DRAC
using \link{use_DRAC}. To reduce the number of values that need to
be provided, \code{preset} can be used to create a template with at least
a minimum of reasonable preset values.

\code{preset} can be one of the following:
\itemize{
\item \code{quartz_coarse}
\item \code{quartz_fine}
\item \code{feldspar_coarse}
\item \code{polymineral_fine}
\item \code{DRAC-example_quartz}
\item \code{DRAC-example_feldspar}
\item \code{DRAC-example_polymineral}
}

Note that the last three options can be used to produce a template
with values directly taken from the official DRAC input \code{.csv} file.}

\item{file_input}{\link{character} (\emph{optional}):
file connection to a DRAC \code{.csv} file. The file will be imported and
translated to a template usable by \link{use_DRAC}.
Please note that there is no check on the validity of the \code{.csv} file.}

\item{notification}{\link{logical} (\emph{with default}):
show or hide the notification.}
}
\value{
A  list of class \code{DRAC.list}.
}
\description{
This function returns a DRAC input template (v1.2) to be used in conjunction
with the \link{use_DRAC} function
}
\section{Function version}{
 0.1.1
}

\examples{

# create a new DRAC input input
input <- template_DRAC(preset = "DRAC-example_quartz")

# show content of the input
print(input)
print(input$`Project ID`)
print(input[[4]])


## Example: DRAC Quartz example
# note that you only have to assign new values where they
# are different to the default values
input$`Project ID` <- "DRAC-Example"
input$`Sample ID` <- "Quartz"
input$`Conversion factors` <- "AdamiecAitken1998"
input$`External U (ppm)` <- 3.4
input$`errExternal U (ppm)` <- 0.51
input$`External Th (ppm)` <- 14.47
input$`errExternal Th (ppm)` <- 1.69
input$`External K (\%)` <- 1.2
input$`errExternal K (\%)` <- 0.14
input$`Calculate external Rb from K conc?` <- "N"
input$`Calculate internal Rb from K conc?` <- "N"
input$`Scale gammadoserate at shallow depths?` <- "N"
input$`Grain size min (microns)` <- 90
input$`Grain size max (microns)` <- 125
input$`Water content ((wet weight - dry weight)/dry weight) \%` <- 5
input$`errWater content \%` <- 2
input$`Depth (m)` <- 2.2
input$`errDepth (m)` <- 0.22
input$`Overburden density (g cm-3)` <- 1.8
input$`errOverburden density (g cm-3)` <- 0.1
input$`Latitude (decimal degrees)` <- 30.0000
input$`Longitude (decimal degrees)` <- 70.0000
input$`Altitude (m)` <- 150
input$`De (Gy)` <- 20
input$`errDe (Gy)` <- 0.2

# use DRAC
\dontrun{
output <- use_DRAC(input)
}

} 

\section{How to cite}{
Burow, C., Kreutzer, S., 2026. template_DRAC(): Create a DRAC input data template (v1.2). Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Durcan, J.A., King, G.E., Duller, G.A.T., 2015. DRAC: Dose Rate and Age Calculator for trapped charge dating.
Quaternary Geochronology 28, 54-61. doi:10.1016/j.quageo.2015.03.012
}
\seealso{
\link{as.data.frame}, \link{list}
}
\author{
Christoph Burow, University of Cologne (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_FadingMeasurement.R
\name{analyse_FadingMeasurement}
\alias{analyse_FadingMeasurement}
\title{Analyse fading measurements and returns the fading rate per decade (g-value)}
\usage{
analyse_FadingMeasurement(
  object,
  structure = c("Lx", "Tx"),
  signal_integral = NULL,
  background_integral = NULL,
  integral_input = c("channel", "measurement"),
  t_star = c("half", "half_complex", "end"),
  n.MC = 100,
  verbose = TRUE,
  plot = TRUE,
  plot_singlePanels = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis}, \link{data.frame} or \link{list} (\strong{required}):
input object with the measurement data. Alternatively, a \link{list} containing
\linkS4class{RLum.Analysis} objects or a \link{data.frame} with three columns
(x = LxTx, y = LxTx error, z = time since irradiation) can be provided.
Can also be a wide table, i.e. a \link{data.frame} with a number of columns
divisible by 3 and where each triplet has the before mentioned column
structure.

\strong{Note:} The input object should solely consists of the curve needed
for the data analysis, i.e. only IRSL curves representing Lx (and Tx). If
the object originated from an XSYG file, also the irradiation steps must
be preserved in the input object.

If data from multiple aliquots are provided, please see the details below
with regard to Lx/Tx normalisation. \strong{Note:} The function assumes that all your
measurements are related to one (comparable) sample. If you have to treat
independent samples, you have use this function in a loop.}

\item{structure}{\link{character} (\emph{with default}):
the structure of the measurement data, one of \code{'Lx'} or \code{c('Lx','Tx')}.}

\item{signal_integral}{\link{vector} (\strong{required}):
vector with channels for the signal integral (e.g., \code{1:10}). It is not
required if a \code{data.frame} with \code{LxTx} values is provided.}

\item{background_integral}{\link{vector} (\strong{required}):
vector with channels for the background integral (e.g., \code{90:100}). It is not
required if a \code{data.frame} with \code{LxTx} values is provided.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{t_star}{\link{character}, \link{function} (\emph{with default}):
method for calculating the time elapsed since irradiation if input is
\strong{not} a \code{data.frame}. Options are: \code{'half'} (the default), \verb{'half_complex},
which uses the long equation in Auclair et al. 2003, and \code{'end'}, which
takes the time between irradiation and the measurement step.
Alternatively, \code{t_star} can be a function with one parameter which works on \code{t1}.
For more information see details. \cr

\emph{\code{t_star} has no effect if the input is a \link{data.frame}, because this input comes
without irradiation times.}}

\item{n.MC}{\link{integer} (\emph{with default}):
number for Monte Carlo runs for the error estimation.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{plot_singlePanels}{\link{logical} or \link{numeric} (\emph{with default}):
enable/disable single plot mode, i.e. one plot window per plot.
Alternatively a vector specifying the plot to be drawn, e.g.,
\code{plot_singlePanels = c(3,4)} draws only the last two plots in separate
windows.}

\item{...}{further arguments that can be passed to internally used functions.
Supported arguments: \code{xlab}, \code{log}, \code{mtext}, \code{plot.trend} (enable/disable
trend blue line), and \code{xlim} for the two first curve plots, and \code{ylim} for
the fading curve plot. For further plot customization please use the
numerical output of the functions for own plots.}
}
\value{
An \linkS4class{RLum.Results} object is returned:

Slot: \strong{@data}

\tabular{lll}{
\strong{OBJECT} \tab \strong{TYPE} \tab \strong{COMMENT}\cr
\code{fading_results} \tab \code{data.frame} \tab results of the fading measurement in a table \cr
\code{fit} \tab \code{lm} \tab object returned by the used linear fitting function \link[stats:lm]{stats::lm}\cr
\code{rho_prime} \tab \code{data.frame} \tab results of rho' estimation after Kars et al. (2008) \cr
\code{LxTx_table} \tab \code{data.frame} \tab Lx/Tx table, if curve data had been provided \cr
\code{irr.times} \tab \code{integer} \tab vector with the irradiation times in seconds \cr
}

Slot: \strong{@info}

\tabular{lll}{
\strong{OBJECT} \tab \code{TYPE} \tab \code{COMMENT}\cr
\code{call} \tab \code{call} \tab the original function call\cr
}
}
\description{
The function analyses fading measurements and returns a fading rate
including an error estimation. The function is not limited to standard
fading measurements, as can be seen, e.g., Huntley and Lamothe (2001).
Additionally, the density of recombination centres (rho') is estimated after
Kars et al. (2008).
}
\details{
All provided output corresponds to the \eqn{tc} value obtained by this
analysis. Additionally, the g-value normalised to 2-days is provided in the
output object. The output of this function can be passed to the function
\link{calc_FadingCorr}.

\strong{Fitting and error estimation}

For the fitting the function \link[stats:lm]{stats::lm} is used without applying weights. For the
error estimation all input values, except \code{tc}, as the precision can be considered as sufficiently
high enough with regard to the underlying problem, are sampled assuming a normal distribution
for each value with the value as the mean and the provided uncertainty as standard deviation.

\strong{The options for \code{t_star}}

\itemize{
\item \code{t_star = "half"} (the default) The calculation follows the simplified
version in Auclair et al. (2003), which reads
\deqn{t_{star} := t_1 + (t_2 - t_1)/2}
\item \code{t_star = "half_complex"} This option applies the complex function shown in Auclair et al. (2003),
which is derived from Aitken (1985) appendix F, equations 9 and 11.
It reads \deqn{t_{star} = t0 * 10^[(t_2 \log(t_2/t_0) - t_1 \log(t_1/t_0) - 0.43(t_2 - t_1))/(t_2 - t_1)]}
where 0.43 = \eqn{1/\ln(10)}. t0, which is an arbitrary constant, is set to 1.
Please note that the equation in Auclair et al. (2003) is incorrect
insofar that it reads \eqn{10\exp(...)}, where the base should be 10 and not the Euler's number.
Here we use the correct version (base 10).
\item \code{t_star = "end"} This option uses the simplest possible form for \code{t_star} which is the time since
irradiation without taking into account any addition parameter and it equals t1 in Auclair et al. (2003)
\item \verb{t_star = <function>} This last option allows you to provide an R function object that works on t1 and
gives you all possible freedom. For instance, you may want to define the following
function \code{fun <- function(x) {x^2}}, this would square all values of t1, because internally
it calls \code{fun(t1)}. The name of the function does not matter.
}

\strong{Density of recombination centres}

The density of recombination centres, expressed by the dimensionless variable rho', is estimated
by fitting equation 5 in Kars et al. (2008) to the data. For the fitting the function
\link[stats:nls]{stats::nls} is used without applying weights. For the error estimation the same
procedure as for the g-value is applied (see above).

\strong{Multiple aliquots & Lx/Tx normalisation}

Be aware that this function will always normalise all \eqn{\frac{L_x}{T_x}}
values by the \eqn{\frac{L_x}{T_x}} value of the prompt measurement of the
first aliquot. This implicitly assumes that there are no systematic
inter-aliquot variations in the \eqn{\frac{L_x}{T_x}} values.
If deemed necessary to normalise the \eqn{\frac{L_x}{T_x}} values  of each
aliquot by its individual prompt measurement please do so \strong{before} running
\link{analyse_FadingMeasurement} and provide the already normalised values for
\code{object} instead.

\strong{Shine-down curve plots}
Please note that the shine-down curve plots are for information only. As such
a maximum of five pause steps are plotted to avoid graphically overloaded plots.
However, \emph{all} pause times are taken into consideration for the analysis.
}
\section{Function version}{
 0.1.27
}

\examples{

## load example data (sample UNIL/NB123, see ?ExampleData.Fading)
data("ExampleData.Fading", envir = environment())

##(1) get fading measurement data (here a three column data.frame)
fading_data <- ExampleData.Fading$fading.data$IR50

##(2) run analysis
g_value <- analyse_FadingMeasurement(
fading_data,
plot = TRUE,
verbose = TRUE,
n.MC = 10)

##(3) this can be further used in the function
## to correct the age according to Huntley & Lamothe, 2001
results <- calc_FadingCorr(
age.faded = c(100,2),
g_value = g_value,
n.MC = 10)

} 

\section{How to cite}{
Kreutzer, S., Burow, C., 2026. analyse_FadingMeasurement(): Analyse fading measurements and returns the fading rate per decade (g-value). Function version 0.1.27. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Aitken, M.J., 1985. Thermoluminescence dating, Studies in archaeological science.
Academic Press, London, Orlando.

Auclair, M., Lamothe, M., Huot, S., 2003. Measurement of anomalous fading for feldspar IRSL using
SAR. Radiation Measurements 37, 487-492. \doi{10.1016/S1350-4487(03)00018-0}

Huntley, D.J., Lamothe, M., 2001. Ubiquity of anomalous fading in K-feldspars and the measurement
and correction for it in optical dating. Canadian Journal of Earth Sciences 38,
1093-1106. doi: \code{10.1139/cjes-38-7-1093}

Kars, R.H., Wallinga, J., Cohen, K.M., 2008. A new approach towards anomalous
fading correction for feldspar  IRSL dating-tests on samples in field saturation.
Radiation Measurements 43, 786-790. \doi{10.1016/j.radmeas.2008.01.021}
}
\seealso{
\link{calc_OSLLxTxRatio}, \link{read_BIN2R},
\link{read_XSYG2R},
\link{extract_IrradiationTimes}, \link{calc_FadingCorr}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R,
%   R/RLum.Data.Image-class.R, R/RLum.Data.Spectrum-class.R,
%   R/RLum.Results-class.R, R/Risoe.BINfileData-class.R
\name{show}
\alias{show}
\alias{show,RLum.Analysis-method}
\alias{show,RLum.Data.Curve-method}
\alias{show,RLum.Data.Image-method}
\alias{show,RLum.Data.Spectrum-method}
\alias{show,RLum.Results-method}
\alias{show,Risoe.BINfileData-method}
\title{Show the structure of RLum-class and Risoe.BINfileData-class objects}
\usage{
\S4method{show}{RLum.Analysis}(object)

\S4method{show}{RLum.Data.Curve}(object)

\S4method{show}{RLum.Data.Image}(object)

\S4method{show}{RLum.Data.Spectrum}(object)

\S4method{show}{RLum.Results}(object)

\S4method{show}{Risoe.BINfileData}(object)
}
\arguments{
\item{object}{\linkS4class{RLum}, \linkS4class{Risoe.BINfileData} (\strong{required}):
object of class \code{RLum} or \code{Risoe.BINfileData}.}
}
\description{
Show the structure of RLum-class and Risoe.BINfileData-class objects
}
\section{Methods (by class)}{
\itemize{
\item \code{show(RLum.Analysis)}: Show the structure of \code{RLum.Analysis} objects.

\item \code{show(RLum.Data.Curve)}: Show the structure of \code{RLum.Data.Curve} objects.

\item \code{show(RLum.Data.Image)}: Show the structure of \code{RLum.Data.Image} objects.

\item \code{show(RLum.Data.Spectrum)}: Show the structure of \code{RLum.Data.Spectrum} objects.

\item \code{show(RLum.Results)}: Show the structure of \code{RLum.Results} objects.

\item \code{show(Risoe.BINfileData)}: Show the structure of \code{Risoe.BINfileData} objects.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_SingleGrainDisc.R
\name{plot_SingleGrainDisc}
\alias{plot_SingleGrainDisc}
\title{Plot a disc with its values}
\usage{
plot_SingleGrainDisc(
  object,
  show_coordinates = FALSE,
  show_location_ids = FALSE,
  show_neighbours = FALSE,
  show_positioning_holes = TRUE,
  df_neighbours = NULL,
  ignore_borders = FALSE,
  str_transform = c("sqrt", "lin", "log"),
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} or \link{numeric} (\strong{required}):
the values to show, should have length 100.}

\item{show_coordinates}{\link{logical} (\emph{with default}): Show coordinates (1..10)
in x and in y direction. Defaults to \code{FALSE}.}

\item{show_location_ids}{\link{logical} (\emph{with default}): Show id with every
grain location (1..100). Defaults to \code{FALSE}.}

\item{show_neighbours}{\link{logical} (\emph{with default}): Show which
neighbour connections are taken into account if calculating Moran's I.
This makes sense when there are \code{NA} observations, or when a non-standard
neighbour setting is defined.}

\item{show_positioning_holes}{\link{logical} (\emph{with default}): Show the 3
positioning holes for orientation. Defaults to \code{TRUE}.}

\item{df_neighbours}{\link{data.frame} (\emph{with default}): only relevant if
\code{show_neighbours} is \code{TRUE}. Data frame indicating which borders to
consider, and their respective weights (see the description provided for
\link{calc_MoransI}). If \code{NULL} (default), this is constructed automatically by
the internal function \code{.get_Neighbours}.}

\item{ignore_borders}{\link{logical} (\emph{with default}): whether only grain
locations that do not lie on the border of the disc should be considered
(\code{FALSE} by default). Thus if \code{TRUE}, only the inner 8x8 grain locations
rather than the full 10x10 are considered. Ignored if \code{df_neighbours} is
not \code{NULL} or if \code{show_neighbours = FALSE}.}

\item{str_transform}{\link{character} (\emph{with default}): The observed value of each individual grain is
reflected in the size of a triangle (or other dot-like element). To account for large value differences,
the transformation from value to triangle size can be \code{"lin"} (linear), \code{"log"} (logarithmic) and \code{"sqrt"}
(square root). Defaults to \code{"sqrt"}, so that the surface is linear to the value. Note that
the log and sqrt transformations can come with an addition to avoid negative values. When the legend
is shown, the actual lower, middle and upper values are printed.}

\item{...}{other arguments to be given to the base R plot function, such
as \code{main}, \code{col} and \code{pch}. \code{legend} can be used to enable/disable the
legend (\code{FALSE} by default).}
}
\description{
Shows a schematic representation of the physical appearance
of one disc (one position in the reader)
and illustrates the measured or calculated values per grain location.
}
\details{
Depending of the available plotting space, some optional elements might have not enough room
to be displayed. As this function is wrapped around the base plot function, one can also choose to add elements
manually.
}
\examples{

plot_SingleGrainDisc(1:100)

} 

\section{How to cite}{
Boer, A.d., Steinbuch, L., 2026. plot_SingleGrainDisc(): Plot a disc with its values. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
de Boer, A-M., Steinbuch, L., Heuvelink, G.B.M., Wallinga, J., 2025.
A novel method to assess crosstalk in single-grain luminescence detection.
Radiation Measurements 186, 107459. \doi{10.1016/j.radmeas.2025.107459}
}
\author{
Anna-Maartje de Boer, Wageningen University & Research (The Netherlands)\cr
Luc Steinbuch, Wageningen University & Research (The Netherlands)\cr
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_ROI.R
\name{plot_ROI}
\alias{plot_ROI}
\title{Create Regions of Interest (ROI) Graphic}
\usage{
plot_ROI(
  object,
  exclude_ROI = 1,
  dist_thre = -Inf,
  dim.CCD = NULL,
  bg_image = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis}, \linkS4class{RLum.Results} or a \link{list} of such objects (\strong{required}):
input data created either by \link{read_RF2R} or \link{extract_ROI}.}

\item{exclude_ROI}{\link{numeric} (\emph{with default}):
option to remove particular ROIs from the analysis. Those ROIs are plotted
but not coloured and not taken into account in distance analysis. \code{NULL}
excludes nothing.}

\item{dist_thre}{\link{numeric} (\emph{optional}):
euclidean distance threshold in pixel distance. All ROI for which the
euclidean distance is smaller are marked. This helps to identify ROIs that
might be affected by signal cross-talk. Note: the distance is calculated
from the centre of an ROI, e.g., the threshold should include consider the
ROIs or grain radius.}

\item{dim.CCD}{\link{numeric} (\emph{optional}):
metric x and y for the recorded (chip) surface in µm. For instance
\code{c(8192, 8192)}, if set additional x and y-axes are shown.}

\item{bg_image}{\linkS4class{RLum.Data.Image} (\emph{optional}):
background image object (please note that dimensions are not checked).}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{...}{further parameters to manipulate the plot. Besides the standard
arguments of \link[graphics:plot.default]{graphics::plot.default}, the following arguments are
supported: \code{lwd.ROI}, \code{lty.ROI}, \code{col.ROI}, \code{col.pixel}, \code{text.labels},
\code{text.offset}, \code{grid} (\code{TRUE/FALSE}), \code{legend} (\code{TRUE/FALSE}),
\code{legend.text}, \code{legend.pos}}
}
\value{
An ROI plot and an \linkS4class{RLum.Results} object with a matrix
containing the extracted ROI data and a object produced by \link[stats:dist]{stats::dist}
containing the euclidean distance between the ROIs.
}
\description{
The function creates ROI graphic with data extracted from the data imported
via \link{read_RF2R}. This function might be of use to work with
reduced data from spatially resolved measurements.
The plot dimensions mimic the original image dimensions.
}
\section{Function version}{
 0.2.0
}

\examples{

## simple example
file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
temp <- read_RF2R(file)
plot_ROI(temp)

## in combination with extract_ROI()
m <- matrix(runif(100,0,255), ncol = 10, nrow = 10)
roi <- matrix(c(2.,4,2,5,6,7,3,1,1), ncol = 3)
t <- extract_ROI(object = m, roi = roi)
plot_ROI(t, bg_image = m)

}
\seealso{
\link{read_RF2R}, \link{extract_ROI}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. plot_ROI(): Create Regions of Interest (ROI) Graphic. Function version 0.2.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_R2TIFF.R
\name{write_R2TIFF}
\alias{write_R2TIFF}
\title{Export RLum.Data.Image and RLum.Data.Spectrum objects to TIFF Images}
\usage{
write_R2TIFF(object, file = tempfile(), norm = 65535, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Image} or \linkS4class{RLum.Data.Spectrum} object (\strong{required}):
input object, can be a \link{list} of such objects.}

\item{file}{\link{character} (\strong{required}):
name of the output file.}

\item{norm}{\link{numeric} (\emph{with default}):
normalisation value. Usually, in imaging applications the pixel values are
integer count values, but values in TIFF files must be in the 0-1 range.
Normalising to the to the highest 16-bit integer values - 1 ensures that
the numerical values are retained in the exported image. If \code{1} nothing is
normalised.}

\item{...}{further arguments to be passed to \link[tiff:writeTIFF]{tiff::writeTIFF}.}
}
\value{
A TIFF file
}
\description{
Simple wrapper around \link[tiff:writeTIFF]{tiff::writeTIFF} to export suitable \linkS4class{RLum}
objects to TIFF images. Per default 16-bit TIFF files are exported.
}
\section{Function version}{
 0.1.2
}

\examples{
data(ExampleData.RLum.Data.Image, envir = environment())
write_R2TIFF(ExampleData.RLum.Data.Image, file = tempfile(fileext = ".tiff"))

}
\seealso{
\link[tiff:writeTIFF]{tiff::writeTIFF}, \linkS4class{RLum.Data.Image},
\linkS4class{RLum.Data.Spectrum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. write_R2TIFF(): Export RLum.Data.Image and RLum.Data.Spectrum objects to TIFF Images. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{sTeve}
\alias{sTeve}
\title{sTeve - sophisticated tool for efficient data validation and evaluation}
\usage{
sTeve(n_frames = 10, t_animation = 2, n.tree = 7, type = NULL)
}
\arguments{
\item{n_frames}{\link{integer} (\emph{with default}):
n frames}

\item{t_animation}{\link{integer} (\emph{with default}):
t animation}

\item{n.tree}{\link{integer} (\emph{with default}):
how many trees do you want to cut?}

\item{type}{\link{integer} (\emph{optional}):
Make a decision: 1, 2 or 3}
}
\value{
Validates your data.
}
\description{
This function provides a sophisticated routine for comprehensive
luminescence dating data analysis.
}
\details{
This amazing sophisticated function validates your data seriously.
}
\note{
This function should not be taken too seriously.
}
\examples{

##no example available

}
\seealso{
\link{plot_KDE}
}
\author{
R Luminescence Team, 2012-2046
, RLum Developer Team} 

\section{How to cite}{
NA, NA, , , 2026. sTeve(): sTeve - sophisticated tool for efficient data validation and evaluation. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_RLum.R
\name{methods_RLum}
\alias{methods_RLum}
\alias{plot.list}
\alias{plot.RLum.Results}
\alias{plot.RLum.Analysis}
\alias{plot.RLum.Data.Curve}
\alias{plot.RLum.Data.Spectrum}
\alias{plot.RLum.Data.Image}
\alias{plot.Risoe.BINfileData}
\alias{hist.RLum.Results}
\alias{hist.RLum.Data.Image}
\alias{hist.RLum.Data.Curve}
\alias{hist.RLum.Analysis}
\alias{summary.RLum.Results}
\alias{summary.RLum.Analysis}
\alias{summary.RLum.Data.Image}
\alias{summary.RLum.Data.Curve}
\alias{subset.Risoe.BINfileData}
\alias{subset.RLum.Analysis}
\alias{bin}
\alias{bin.RLum.Data.Curve}
\alias{bin.RLum.Data.Spectrum}
\alias{length.RLum.Results}
\alias{length.RLum.Analysis}
\alias{length.RLum.Data.Curve}
\alias{length.Risoe.BINfileData}
\alias{dim.RLum.Data.Curve}
\alias{dim.RLum.Data.Spectrum}
\alias{rep.RLum}
\alias{names.RLum.Data.Curve}
\alias{names.RLum.Data.Spectrum}
\alias{names.RLum.Data.Image}
\alias{names.RLum.Analysis}
\alias{names.RLum.Results}
\alias{names.Risoe.BINfileData}
\alias{row.names.RLum.Data.Spectrum}
\alias{as.data.frame.RLum.Data.Curve}
\alias{as.data.frame.RLum.Data.Spectrum}
\alias{as.data.frame.Risoe.BINfileData}
\alias{as.list.RLum.Results}
\alias{as.list.RLum.Data.Curve}
\alias{as.list.RLum.Data.Image}
\alias{as.list.RLum.Data.Spectrum}
\alias{as.list.RLum.Analysis}
\alias{as.matrix.RLum.Data.Curve}
\alias{as.matrix.RLum.Data.Spectrum}
\alias{as.matrix.RLum.Data.Image}
\alias{merge.RLum}
\alias{unlist.RLum.Analysis}
\alias{+.RLum.Data.Curve}
\alias{+.RLum.Data.Spectrum}
\alias{-.RLum.Data.Curve}
\alias{-.RLum.Data.Spectrum}
\alias{*.RLum.Data.Curve}
\alias{*.RLum.Data.Spectrum}
\alias{/.RLum.Data.Curve}
\alias{/.RLum.Data.Spectrum}
\alias{[.RLum.Data.Curve}
\alias{[.RLum.Data.Spectrum}
\alias{[.RLum.Data.Image}
\alias{[.RLum.Analysis}
\alias{[.RLum.Results}
\alias{[<-.RLum.Data.Curve}
\alias{[[.RLum.Analysis}
\alias{[[.RLum.Results}
\alias{$.RLum.Data.Curve}
\alias{$.RLum.Analysis}
\alias{$.RLum.Results}
\title{methods_RLum}
\usage{
\method{plot}{list}(x, y, ...)

\method{plot}{RLum.Results}(x, y, ...)

\method{plot}{RLum.Analysis}(x, y, ...)

\method{plot}{RLum.Data.Curve}(x, y, ...)

\method{plot}{RLum.Data.Spectrum}(x, y, ...)

\method{plot}{RLum.Data.Image}(x, y, ...)

\method{plot}{Risoe.BINfileData}(x, y, ...)

\method{hist}{RLum.Results}(x, ...)

\method{hist}{RLum.Data.Image}(x, ...)

\method{hist}{RLum.Data.Curve}(x, ...)

\method{hist}{RLum.Analysis}(x, ...)

\method{summary}{RLum.Results}(object, ...)

\method{summary}{RLum.Analysis}(object, ...)

\method{summary}{RLum.Data.Image}(object, ...)

\method{summary}{RLum.Data.Curve}(object, ...)

\method{subset}{Risoe.BINfileData}(x, subset, records.rm = TRUE, ...)

\method{subset}{RLum.Analysis}(x, subset = NULL, ...)

bin(x, ...)

\method{bin}{RLum.Data.Curve}(x, bin_size = 2, ...)

\method{bin}{RLum.Data.Spectrum}(x, bin_size.row = 1, bin_size.col = 1, ...)

\method{length}{RLum.Results}(x, ...)

\method{length}{RLum.Analysis}(x, ...)

\method{length}{RLum.Data.Curve}(x, ...)

\method{length}{Risoe.BINfileData}(x, ...)

\method{dim}{RLum.Data.Curve}(x)

\method{dim}{RLum.Data.Spectrum}(x)

\method{rep}{RLum}(x, ...)

\method{names}{RLum.Data.Curve}(x, ...)

\method{names}{RLum.Data.Spectrum}(x, ...)

\method{names}{RLum.Data.Image}(x, ...)

\method{names}{RLum.Analysis}(x, ...)

\method{names}{RLum.Results}(x, ...)

\method{names}{Risoe.BINfileData}(x)

\method{row.names}{RLum.Data.Spectrum}(x, ...)

\method{as.data.frame}{RLum.Data.Curve}(x, row.names = NULL, optional = FALSE, ...)

\method{as.data.frame}{RLum.Data.Spectrum}(x, row.names = NULL, optional = FALSE, ...)

\method{as.data.frame}{Risoe.BINfileData}(x, row.names = NULL, optional = FALSE, ...)

\method{as.list}{RLum.Results}(x, ...)

\method{as.list}{RLum.Data.Curve}(x, ...)

\method{as.list}{RLum.Data.Image}(x, ...)

\method{as.list}{RLum.Data.Spectrum}(x, ...)

\method{as.list}{RLum.Analysis}(x, ...)

\method{as.matrix}{RLum.Data.Curve}(x, ...)

\method{as.matrix}{RLum.Data.Spectrum}(x, ...)

\method{as.matrix}{RLum.Data.Image}(x, ...)

\method{merge}{RLum}(x, y, ...)

\method{unlist}{RLum.Analysis}(x, recursive = TRUE, ...)

\method{+}{RLum.Data.Curve}(x, y)

\method{+}{RLum.Data.Spectrum}(x, y)

\method{-}{RLum.Data.Curve}(x, y)

\method{-}{RLum.Data.Spectrum}(x, y)

\method{*}{RLum.Data.Curve}(x, y)

\method{*}{RLum.Data.Spectrum}(x, y)

\method{/}{RLum.Data.Curve}(x, y)

\method{/}{RLum.Data.Spectrum}(x, y)

\method{[}{RLum.Data.Curve}(x, y, z, drop = TRUE)

\method{[}{RLum.Data.Spectrum}(x, y, z, drop = TRUE)

\method{[}{RLum.Data.Image}(x, y, z, drop = TRUE)

\method{[}{RLum.Analysis}(x, i, drop = FALSE)

\method{[}{RLum.Results}(x, i, drop = TRUE)

\method{[}{RLum.Data.Curve}(x, i, j) <- value

\method{[[}{RLum.Analysis}(x, i)

\method{[[}{RLum.Results}(x, i)

\method{$}{RLum.Data.Curve}(x, i)

\method{$}{RLum.Analysis}(x, i)

\method{$}{RLum.Results}(x, i)
}
\arguments{
\item{x}{\linkS4class{RLum} or \linkS4class{Risoe.BINfileData} (\strong{required}):
input object}

\item{y}{\link{integer} (\emph{optional}):
the row index of the matrix, data.frame}

\item{...}{further arguments that can be passed to the method}

\item{object}{\linkS4class{RLum} (\strong{required}):
input object}

\item{subset}{\verb{[subset]} \link{expression} (\strong{required}):
logical expression indicating elements or rows to keep, this function works
in \linkS4class{Risoe.BINfileData} objects like \link{subset.data.frame}, but takes care
of the object structure. Works also on \linkS4class{RLum.Analysis} objects.}

\item{records.rm}{\link{subset} \link{logical} (\emph{with default}):
whether records should be removed from the data set. If \code{FALSE}, the \code{SEL}
column marks the selected records.}

\item{row.names}{\link{logical} (\emph{with default}):
enable/disable row names (\code{as.data.frame}).}

\item{optional}{\link{logical} (\emph{with default}):
logical. If TRUE, setting row names and converting column names
(to syntactic names: see make.names) is optional (see \link[base:as.data.frame]{base::as.data.frame})}

\item{recursive}{\link{logical} (\emph{with default}):
enable/disable further sub-setting (\code{unlist}).}

\item{z}{\link{integer} (\emph{optional}):
the column index of the matrix, data.frame}

\item{drop}{\link{logical} (\emph{with default}):
keep object structure or drop it}

\item{i}{\link{character} (\emph{optional}):
name of the wanted record type or data object or row in the
\linkS4class{RLum.Data.Curve} object}

\item{j}{\link{integer} (\emph{optional}):
column of the data matrix in the \linkS4class{RLum.Data.Curve} object}

\item{value}{\link{numeric} (\strong{required}):
numeric value which replace the value in the \linkS4class{RLum.Data.Curve} object}
}
\description{
Methods for S3-generics implemented for the package 'Luminescence'.
This document summarises all implemented S3-generics. The name of the function
is given before the first dot, after the dot the name of the object that is
supported by this method is given, e.g. \code{plot.RLum.Data.Curve} can be called
by \code{plot(object, ...)}, where \code{object} is the \code{RLum.Data.Curve} object.
}
\details{
The term S3-generics sounds complicated, however, it just means that something
has been implemented in the package to increase the usability for users new
in R and who are not familiar with the underlying \code{RLum}-object structure of
the package. The practical outcome is that operations and functions presented
in standard books on R can be used without knowing the specifics of the R
package \code{'Luminescence'}. For examples see the example section.
}
\note{
\code{methods_RLum} are not really new functions, everything given here are mostly just
surrogates for existing functions in the package.
}
\examples{

##load example data
data(ExampleData.RLum.Analysis, envir = environment())


##combine curve is various ways
curve1 <- IRSAR.RF.Data[[1]]
curve2 <-  IRSAR.RF.Data[[1]]
curve1 + curve2
curve1 - curve2
curve1 / curve2
curve1 * curve2


##`$` access curves
IRSAR.RF.Data$RF

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data-class.R, R/Risoe.BINfileData-class.R
\name{add_metadata<-}
\alias{add_metadata<-}
\alias{rename_metadata<-}
\alias{replace_metadata<-}
\alias{add_metadata<-,RLum.Analysis-method}
\alias{rename_metadata<-,RLum.Analysis-method}
\alias{replace_metadata<-,RLum.Analysis-method}
\alias{add_metadata<-,RLum.Data-method}
\alias{rename_metadata<-,RLum.Data-method}
\alias{replace_metadata<-,RLum.Data-method}
\alias{add_metadata<-,Risoe.BINfileData-method}
\alias{rename_metadata<-,Risoe.BINfileData-method}
\alias{replace_metadata<-,Risoe.BINfileData-method}
\title{Safe manipulation of object metadata}
\usage{
add_metadata(object, ...) <- value

rename_metadata(object, ...) <- value

replace_metadata(object, ...) <- value

\S4method{add_metadata}{RLum.Analysis}(object, info_element) <- value

\S4method{rename_metadata}{RLum.Analysis}(object, info_element) <- value

\S4method{replace_metadata}{RLum.Analysis}(object, info_element, subset = NULL) <- value

\S4method{add_metadata}{RLum.Data}(object, info_element) <- value

\S4method{rename_metadata}{RLum.Data}(object, info_element) <- value

\S4method{replace_metadata}{RLum.Data}(object, info_element, subset = NULL, verbose = TRUE) <- value

\S4method{add_metadata}{Risoe.BINfileData}(object, info_element) <- value

\S4method{rename_metadata}{Risoe.BINfileData}(object, info_element) <- value

\S4method{replace_metadata}{Risoe.BINfileData}(object, info_element, subset = NULL) <- value
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis},
\linkS4class{Risoe.BINfileData} (\strong{required}):
object of class \code{RLum.Analysis} or \code{Risoe.BINfileData} to manipulate.}

\item{...}{further arguments passed to the specific class method}

\item{value}{(\strong{required}):
value to be assigned to the selected metadata entry. A \code{NULL} value is
acceptable only for \code{replace_metadata}, in which case the elements named
in \code{info_element} will be removed.}

\item{info_element}{\link{character} (\strong{required}):
name of the metadata entry to manipulate.}

\item{subset}{\link{expression} (\emph{optional}):
logical expression to limit the substitution only to the selected subset
of elements.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}
}
\description{
Generic functions for manipulation of metadata in \linkS4class{Risoe.BINfileData},
\linkS4class{RLum.Analysis} and \linkS4class{RLum.Data} objects.
}
\section{Functions}{
\itemize{
\item \code{add_metadata(RLum.Analysis) <- value}: Adds metadata to \linkS4class{RLum.Analysis} objects.

\item \code{rename_metadata(RLum.Analysis) <- value}: Renames a metadata entry of \linkS4class{RLum.Analysis} objects.

\item \code{replace_metadata(RLum.Analysis) <- value}: Replaces or removes metadata of \linkS4class{RLum.Analysis} objects.

\item \code{add_metadata(RLum.Data) <- value}: Add metadata entries to \linkS4class{RLum.Data} objects.

\item \code{rename_metadata(RLum.Data) <- value}: Rename a metadata entry of \linkS4class{RLum.Data} objects.

\item \code{replace_metadata(RLum.Data) <- value}: Replaces or removes metadata of \linkS4class{RLum.Data} objects.

\item \code{add_metadata(Risoe.BINfileData) <- value}: Adds metadata to \linkS4class{Risoe.BINfileData} objects.

\item \code{rename_metadata(Risoe.BINfileData) <- value}: Renames a metadata entry of \linkS4class{Risoe.BINfileData} objects.

\item \code{replace_metadata(Risoe.BINfileData) <- value}: Replaces or removes metadata of \linkS4class{Risoe.BINfileData} objects.

}}
\examples{

## load example data
data(ExampleData.BINfileData, envir = environment())

## show data
CWOSL.SAR.Data

## add a new field
add_metadata(CWOSL.SAR.Data,
             info_element = "INSTITUTE") <- "Heidelberg University"

## rename a field
rename_metadata(CWOSL.SAR.Data,
                info_element = "INSTITUTE") <- "INSTITUTION"

## replace all LTYPE to RSL
## but only for the first position
replace_metadata(
 object = CWOSL.SAR.Data,
 info_element = "LTYPE",
 subset = (POSITION == 1)) <- "RSL"

## replacing a field with NULL allows to remove that field
replace_metadata(CWOSL.SAR.Data,
                 info_element = "PREVIOUS") <- NULL

## show the modified data
CWOSL.SAR.Data

}
\seealso{
\linkS4class{RLum.Data}, \linkS4class{RLum.Analysis},
\linkS4class{Risoe.BINfileData}
}
\author{
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Colombo, M., 2026. add_metadata<-(): Safe manipulation of object metadata. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_PSL2Risoe.BINfileData.R
\name{convert_PSL2Risoe.BINfileData}
\alias{convert_PSL2Risoe.BINfileData}
\title{Convert portable OSL data to a Risoe.BINfileData object}
\usage{
convert_PSL2Risoe.BINfileData(object, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}):
\code{RLum.Analysis} object produced by \link{read_PSL2R}}

\item{...}{currently not used.}
}
\value{
Returns an S4 \linkS4class{Risoe.BINfileData} object that can be used to write a
BIN file using \link{write_R2BIN}.
}
\description{
Converts an \linkS4class{RLum.Analysis} object produced by the function \link{read_PSL2R} to
a \linkS4class{Risoe.BINfileData} object \strong{(BETA)}.

This function converts an \linkS4class{RLum.Analysis} object that was produced
by the \link{read_PSL2R} function to a \linkS4class{Risoe.BINfileData}.
The \linkS4class{Risoe.BINfileData} can be used to write a Risø BIN file via
\link{write_R2BIN}.
}
\section{Function version}{
 0.0.1
}

\examples{

# (1) load and plot example data set
data("ExampleData.portableOSL", envir = environment())
plot_RLum(ExampleData.portableOSL)

# (2) merge all RLum.Analysis objects into one
merged <- merge_RLum(ExampleData.portableOSL)
merged

# (3) convert to RisoeBINfile object
bin <- convert_PSL2Risoe.BINfileData(merged)
bin

# (4) write Risoe BIN file
\dontrun{
write_R2BIN(bin, "~/portableOSL.binx")
}

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve},
\linkS4class{Risoe.BINfileData}
}
\author{
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team} 

\section{How to cite}{
Burow, C., 2026. convert_PSL2Risoe.BINfileData(): Convert portable OSL data to a Risoe.BINfileData object. Function version 0.0.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_HomogeneityTest.R
\name{calc_HomogeneityTest}
\alias{calc_HomogeneityTest}
\title{Apply a simple homogeneity test after Galbraith (2003)}
\usage{
calc_HomogeneityTest(data, log = TRUE, ...)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with De \code{(data[,1])} and De error \code{(values[,2])}}

\item{log}{\link{logical} (\emph{with default}):
perform the homogeneity test with (un-)logged data}

\item{...}{further arguments (\code{verbose}).}
}
\value{
Returns a terminal output. In addition an
\linkS4class{RLum.Results}-object is returned containing the
following elements:

\item{summary}{\link{data.frame} summary of all relevant model results.}
\item{data}{\link{data.frame} original input data}
\item{args}{\link{list} used arguments}
\item{call}{\link{call} the function call}

The output should be accessed using the function \link{get_RLum}.
}
\description{
A simple homogeneity test for De estimates.
For details see Galbraith (2003).
}
\section{Function version}{
 0.3.0
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())

## apply the homogeneity test
calc_HomogeneityTest(ExampleData.DeValues$BT998)

## using the data presented by Galbraith (2003)
df <-
 data.frame(
   x = c(30.1, 53.8, 54.3, 29.0, 47.6, 44.2, 43.1),
   y = c(4.8, 7.1, 6.8, 4.3, 5.2, 5.9, 3.0))

calc_HomogeneityTest(df)


} 

\section{How to cite}{
Burow, C., Kreutzer, S., 2026. calc_HomogeneityTest(): Apply a simple homogeneity test after Galbraith (2003). Function version 0.3.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Galbraith, R.F., 2003. A simple homogeneity test for estimates
of dose obtained using OSL. Ancient TL 21, 75-77.
}
\seealso{
\link[stats:Chisquare]{stats::pchisq}
}
\author{
Christoph Burow, University of Cologne (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_RLum.R
\name{report_RLum}
\alias{report_RLum}
\title{Create a HTML-report for (RLum) objects}
\usage{
report_RLum(
  object,
  file = tempfile(),
  title = "RLum.Report",
  compact = TRUE,
  timestamp = TRUE,
  show_report = TRUE,
  launch.browser = FALSE,
  css.file = NULL,
  quiet = TRUE,
  clean = TRUE,
  ...
)
}
\arguments{
\item{object}{(\strong{required}):
The object to be reported on, preferably of any \code{RLum}-class.}

\item{file}{\link{character} (\emph{with default}):
A character string naming the output file. If no filename is provided a
temporary file is created.}

\item{title}{\link{character} (\emph{with default}):
A character string specifying the title of the document.}

\item{compact}{\link{logical} (\emph{with default}):
When \code{TRUE} the following report components are hidden:
\verb{@.pid}, \verb{@.uid}, \code{'Object structure'}, \code{'Session Info'}
and only the first and last 5 rows of long matrices and data frames are shown.
See details.}

\item{timestamp}{\link{logical} (\emph{with default}):
\code{TRUE} to add a timestamp to the filename (suffix).}

\item{show_report}{\link{logical} (\emph{with default}): If set to \code{TRUE} the function tries to display
the report output in the local viewer, e.g., within \emph{RStudio} after rendering.}

\item{launch.browser}{\link{logical} (\emph{with default}):
\code{TRUE} to open the HTML file in the system's default web browser after
it has been rendered.}

\item{css.file}{\link{character} (\emph{optional}):
Path to a CSS file to change the default styling of the HTML document.}

\item{quiet}{\link{logical} (\emph{with default}):
\code{TRUE} to suppress printing of the pandoc command line.}

\item{clean}{\link{logical} (\emph{with default}):
\code{TRUE} to clean intermediate files created during rendering.}

\item{...}{further arguments passed to or from other methods and to control
the document's structure (see details).}
}
\value{
Writes a HTML and .Rds file.
}
\description{
Create a HTML-report for (RLum) objects
}
\details{
This function creates a HTML-report for a given object, listing its complete
structure and content. The object itself is saved as a serialised .Rds file.
The report file serves both as a convenient way of browsing through objects with
complex data structures as well as a mean of properly documenting and saving
objects.

The HTML report is created with \link[rmarkdown:render]{rmarkdown::render} and has the
following structure:

\tabular{ll}{
\strong{Section} \tab \strong{Description} \cr
\code{Header} \tab A summary of general characteristics of the object \cr
\verb{Object content} \tab A comprehensive list of the complete structure and content of the provided object. \cr
\verb{Object structure} \tab Summary of the objects structure given as a table \cr
\code{File} \tab Information on the saved RDS file \cr
\verb{Session Info} \tab Captured output from \code{sessionInfo()} \cr
\code{Plots} \tab (\emph{optional}) For \code{RLum-class} objects a variable number of plots \cr
}

The structure of the report can be controlled individually by providing one or more of the
following arguments (all \code{logical}):

\tabular{ll}{
\strong{Argument} \tab \strong{Description} \cr
\code{header} \tab Hide or show general information on the object \cr
\code{main} \tab Hide or show the object's content \cr
\code{structure} \tab Hide or show object's structure \cr
\code{rds} \tab Hide or show information on the saved RDS file \cr
\code{session} \tab Hide or show the session info \cr
\code{plot} \tab Hide or show the plots (depending on object) \cr
}

Note that these arguments have higher precedence than \code{compact}.

Further options that can be provided via the \code{...} argument:

\tabular{ll}{
\strong{Argument} \tab \strong{Description} \cr
\code{short_table} \tab If \code{TRUE} only show the first and last 5 rows of long tables. \cr
\code{theme} \tab Specifies the Bootstrap
theme to use for the report. Valid themes include \code{"default"}, \code{"cerulean"}, \code{"journal"}, \code{"flatly"},
\code{"readable"}, \code{"spacelab"}, \code{"united"}, \code{"cosmo"}, \code{"lumen"}, \code{"paper"}, \code{"sandstone"},
\code{"simplex"}, and \code{"yeti"}. \cr
\code{highlight} \tab Specifies the syntax highlighting style.
Supported styles include \code{"default"}, \code{"tango"}, \code{"pygments"}, \code{"kate"}, \code{"monochrome"},
\code{"espresso"}, \code{"zenburn"}, \code{"haddock"}, and \code{"textmate"}. \cr
\code{css} \tab \code{TRUE} or \code{FALSE} to enable/disable custom CSS styling \cr
}

The following arguments can be used to customise the report via CSS (Cascading Style Sheets):

\tabular{ll}{
\strong{Argument} \tab \strong{Description} \cr
\code{font_family} \tab Define the font family of the HTML document (default: \code{"arial"}) \cr
\code{headings_size} \tab Size of the \verb{<h1>} to \verb{<h6>} tags used to define HTML headings (default: 166\%). \cr
\code{content_color} \tab Colour of the object's content (default: #a72925). \cr
}

Note that these arguments must all be of class \link{character} and follow standard CSS syntax.
For exhaustive CSS styling you can provide a custom CSS file for argument \code{css.file}.
CSS styling can be turned of using \code{css = FALSE}.
}
\note{
This function requires the R packages 'rmarkdown', 'pander' and 'rstudioapi'.
}
\section{Function version}{
 0.1.5
}

\examples{

\dontrun{
## Example: RLum.Results ----

# load example data
data("ExampleData.DeValues")

# apply the MAM-3 age model and save results
mam <- calc_MinDose(ExampleData.DeValues$CA1, sigmab = 0.2)

# create the HTML report
report_RLum(object = mam, file = "~/CA1_MAM.Rmd",
            timestamp = FALSE,
            title = "MAM-3 for sample CA1")

# when creating a report the input file is automatically saved to a
# .Rds file (see saveRDS()).
mam_report <- readRDS("~/CA1_MAM.Rds")
all.equal(mam, mam_report)


## Example: Temporary file & Viewer/Browser ----

# (a)
# Specifying a filename is not necessarily required. If no filename is provided,
# the report is rendered in a temporary file. If you use the RStudio IDE, the
# temporary report is shown in the interactive Viewer pane.
report_RLum(object = mam)

# (b)
# Additionally, you can view the HTML report in your system's default web browser.
report_RLum(object = mam, launch.browser = TRUE)


## Example: RLum.Analysis ----

data("ExampleData.RLum.Analysis")

# create the HTML report (note that specifying a file
# extension is not necessary)
report_RLum(object = IRSAR.RF.Data, file = "~/IRSAR_RF")


## Example: RLum.Data.Curve ----

data.curve <- get_RLum(IRSAR.RF.Data)[[1]]

# create the HTML report
report_RLum(object = data.curve, file = "~/Data_Curve")

## Example: Any other object ----
x <- list(x = 1:10,
          y = runif(10, -5, 5),
          z = data.frame(a = LETTERS[1:20], b = dnorm(0:9)),
          NA)

report_RLum(object = x, file = "~/arbitray_list")
}

}
\seealso{
\link[rmarkdown:render]{rmarkdown::render}, \link[pander:pander_return]{pander::pander_return},
\link[pander:openFileInOS]{pander::openFileInOS}, \link[rstudioapi:viewer]{rstudioapi::viewer},
\link{browseURL}
}
\author{
Christoph Burow, University of Cologne (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
, RLum Developer Team} 

\section{How to cite}{
Burow, C., Kreutzer, S., 2026. report_RLum(): Create a HTML-report for (RLum) objects. Function version 0.1.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal_as.latex.table.R
\name{.as.latex.table}
\alias{.as.latex.table}
\title{Create LaTeX tables from data.frames and RLum objects}
\usage{
.as.latex.table(
  x,
  row.names = NULL,
  col.names = NULL,
  comments = TRUE,
  pos = "c",
  digits = 3,
  rm.zero = TRUE,
  select = NULL,
  split = NULL,
  tabular_only = FALSE,
  ...
)
}
\arguments{
\item{x}{\link{data.frame} or \code{RLum} object (\strong{required})}

\item{row.names}{currently unused}

\item{col.names}{currently unused}

\item{comments}{\link{logical} (\emph{with default}):
insert LaTeX comments}

\item{pos}{\link{character} (\emph{with default}): \code{character} of length one specifying the
alignment of each column, e.g., \code{pos = 'clr'} for a three column data frame and
center, left and right alignment}

\item{digits}{\link{numeric} (\emph{with default}): number of digits to be displayed (numeric fields)}

\item{rm.zero}{\link{logical} (\emph{with default}): remove columns containing
only zeros, however, this might not be wanted in all cases}

\item{select}{\link{character} (\emph{optional}): a \link{character} vector passed to \link{subset}}

\item{split}{\link{integer} (\emph{optional}): an \link{integer} specifying the number of individual tables
the \link{data.frame} is split into. Useful for wide tables. Currently unused.}

\item{tabular_only}{\link{logical} (\emph{with default}): if \code{TRUE} only the tabular but not the
table environment is returned. This gives a lot of additional flexibility at hand}

\item{...}{options: \code{verbose}}
}
\value{
Returns LaTeX code
}
\description{
This function takes a \link{data.frame} and returns a table in LaTeX code that
can be copied into any TeX document.
}
\section{Function version}{
 0.2.1
}

\examples{
 df <- data.frame(x = 1:10, y = letters[1:10])
.as.latex.table(df)
.as.latex.table(df, pos = "lr")
.as.latex.table(df, select = "y", pos = "r")

}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Burow, C., Kreutzer, S., 2026. .as.latex.table(): Create LaTeX tables from data.frames and RLum objects. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_NRt.R
\name{plot_NRt}
\alias{plot_NRt}
\title{Visualise natural/regenerated signal ratios}
\usage{
plot_NRt(
  data,
  log = "",
  smooth = c("none", "spline", "rmean"),
  k = 3,
  legend = TRUE,
  legend.pos = "topright",
  ...
)
}
\arguments{
\item{data}{\link{list}, \link{data.frame}, \link{matrix} or \linkS4class{RLum.Analysis} (\strong{required}):
X,Y data of measured values (time and counts). See details on individual data structure.}

\item{log}{\link{character} (\emph{with default}):
logarithmic axes (\code{"x"}, \code{"y"}, \code{"xy"}) or \code{""} (default) for linear axes.}

\item{smooth}{\link{character} (\emph{with default}):
apply data smoothing. If \code{"none"} (default), no data smoothing is applied.
Use \code{"rmean"} to calculate the rolling mean, where \code{k} determines the
width of the rolling window (see \link[data.table:froll]{data.table::frollmean}). \code{"spline"}
applies a smoothing spline to each curve (see \link[stats:smooth.spline]{stats::smooth.spline})}

\item{k}{\link{integer} (\emph{with default}):
integer width of the rolling window.}

\item{legend}{\link{logical} (\emph{with default}):
enable/disable the plot legend.}

\item{legend.pos}{\link{character} (\emph{with default}):
keyword specifying the position of the legend (see \link{legend}).}

\item{...}{further parameters passed to \link{plot} (also see \link{par}).}
}
\value{
Returns a plot and \linkS4class{RLum.Analysis} object.
}
\description{
This function creates a Natural/Regenerated signal vs. time (NR(t)) plot
as shown in Steffen et al. 2009.

This function accepts the individual curve data in many different formats. If
\code{data} is a \code{list}, each element of the list must contain a two
column \code{data.frame} or \code{matrix} containing the \code{XY} data of the curves
(time and counts). Alternatively, the elements can be objects of class
\linkS4class{RLum.Data.Curve}.

Input values can also be provided as a \code{data.frame} or \code{matrix} where
the first column contains the time values and each following column contains
the counts of each curve.
}
\examples{

## load example data
data("ExampleData.BINfileData", envir = environment())

## EXAMPLE 1

## convert Risoe.BINfileData object to RLum.Analysis object
data <- Risoe.BINfileData2RLum.Analysis(object = CWOSL.SAR.Data, pos = 8, ltype = "OSL")

## extract all OSL curves
allCurves <- get_RLum(data)

## keep only the natural and regenerated signal curves
pos <- seq(1, 9, 2)
curves <- allCurves[pos]

## plot a standard NR(t) plot
plot_NRt(curves)

## re-plot with rolling mean data smoothing
plot_NRt(curves, smooth = "rmean", k = 10)

## re-plot with a logarithmic x-axis
plot_NRt(curves, log = "x", smooth = "rmean", k = 5)

## re-plot with custom axes ranges
plot_NRt(curves, smooth = "rmean", k = 5,
         xlim = c(0.1, 5), ylim = c(0.4, 1.6),
         legend.pos = "bottomleft")

## re-plot with smoothing spline on log scale
plot_NRt(curves, smooth = "spline", log = "x",
         legend.pos = "top")

## EXAMPLE 2

# you may also use this function to check whether all
# TD curves follow the same shape (making it a TnTx(t) plot).
posTD <- seq(2, 14, 2)
curves <- allCurves[posTD]

plot_NRt(curves, main = "TnTx(t) Plot",
         smooth = "rmean", k = 20,
         ylab = "TD natural / TD regenerated",
         xlim = c(0, 20), legend = FALSE)

## EXAMPLE 3

# extract data from all positions
data <- lapply(1:24, FUN = function(pos) {
   Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = pos, ltype = "OSL")
})

# get individual curve data from each aliquot
aliquot <- lapply(data, get_RLum)

# set graphical parameters
par(mfrow = c(2, 2))

# create NR(t) plots for all aliquots
for (i in 1:length(aliquot)) {
   plot_NRt(aliquot[[i]][pos],
            main = paste0("Aliquot #", i),
            smooth = "rmean", k = 20,
            xlim = c(0, 10),
            cex = 0.6, legend.pos = "bottomleft")
}

# reset graphical parameters
par(mfrow = c(1, 1))

} 

\section{How to cite}{
Burow, C., 2026. plot_NRt(): Visualise natural/regenerated signal ratios. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Steffen, D., Preusser, F., Schlunegger, F., 2009. OSL quartz underestimation due to
unstable signal components. Quaternary Geochronology, 4, 353-362.
}
\seealso{
\link{plot}
}
\author{
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Data.Curve-class.R, R/RLum.Data.Spectrum-class.R
\name{bin_RLum.Data}
\alias{bin_RLum.Data}
\alias{bin_RLum.Data,RLum.Data.Curve-method}
\alias{bin_RLum.Data,RLum.Data.Spectrum-method}
\title{Channel binning for RLum.Data-class objects}
\usage{
bin_RLum.Data(object, ...)

\S4method{bin_RLum.Data}{RLum.Data.Curve}(object, bin_size = 2)

\S4method{bin_RLum.Data}{RLum.Data.Spectrum}(object, bin_size.col = 1, bin_size.row = 1)
}
\arguments{
\item{object}{\linkS4class{RLum.Data} (\strong{required}):
S4 object of class \code{RLum.Data}}

\item{...}{further arguments passed to the specific class method}

\item{bin_size}{\link{integer} (\emph{with default}):
number of channels used for each bin, e.g. \code{bin_size = 2} means that
two channels are binned.}

\item{bin_size.col}{\link{integer} (\emph{with default}):
number of channels used for each bin, e.g. \code{bin_size.col = 2} means that
two channels are binned. Note: The function does not check the input, very large values
mean a full column binning (a single sum)}

\item{bin_size.row}{\link{integer} (\emph{with default}):
number of channels used for each bin, e.g. \code{bin_size.row = 2} means that
two channels are binned. Note: The function does not check the input, very large values
mean a full row binning (a single sum)}
}
\value{
An object of the same type as the input provided after binning is applied.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum.Data} objects. Depending on the input object, the corresponding
function will be selected.
}
\section{Functions}{
\itemize{
\item \code{bin_RLum.Data(RLum.Data.Curve)}: Allows binning of RLum.Data.Curve data.

\item \code{bin_RLum.Data(RLum.Data.Spectrum)}: Allows binning of RLum.Data.Spectrum data. Count values and values on the
x-axis are summed up; for wavelength/energy values, the mean is calculated.

}}
\note{
Currently only \code{RLum.Data} objects of class \linkS4class{RLum.Data.Curve}
and \linkS4class{RLum.Data.Spectrum} are supported.
}
\section{Function version}{
 0.2.0
}

\examples{

## load example data
data(ExampleData.CW_OSL_Curve, envir = environment())

## create RLum.Data.Curve object from this example
curve <-
  set_RLum(
      class = "RLum.Data.Curve",
      recordType = "OSL",
      data = as.matrix(ExampleData.CW_OSL_Curve)
  )

## plot data without and with 2 and 4 channel binning
plot_RLum(curve)
plot_RLum(bin_RLum.Data(curve, bin_size = 2))
plot_RLum(bin_RLum.Data(curve, bin_size = 4))

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Data.Spectrum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. bin_RLum.Data(): Channel binning for RLum.Data-class objects. Function version 0.2.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.Al2O3C}
\alias{ExampleData.Al2O3C}
\alias{data_CrossTalk}
\alias{data_ITC}
\title{Example Al2O3:C Measurement Data}
\format{
Two datasets comprising \linkS4class{RLum.Analysis} data imported using the function
\link{read_XSYG2R}

\describe{
\code{data_ITC}: Measurement data to determine the irradiation time correction, the data can
be analysed with the function \link{analyse_Al2O3C_ITC}

\code{data_CrossTalk}: Measurement data obtained while estimating the irradiation cross-talk of the
reader used for the experiments. The data can be analysed either with the function
\link{analyse_Al2O3C_CrossTalk} or \link{analyse_Al2O3C_Measurement}
}
}
\description{
Measurement data obtained from measuring Al2O3:C chips at the IRAMAT-CRP2A, Université Bordeaux
Montaigne in 2017 on a Freiberg Instruments lexsyg SMART reader.
The example data used in particular to allow test of the functions
developed in framework of the work by Kreutzer et al., 2018.
}
\note{
From both datasets unneeded curves have been removed and
the number of aliquots have been reduced to a required minimum to keep the file size small, but
still being able to run the corresponding functions.
}
\examples{

##(1) curves
data(ExampleData.Al2O3C, envir = environment())
plot_RLum(data_ITC[1:2])

}
\references{
Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018.
Environmental Dose Rate Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
Geochronometria 45, 56–67.
}
\seealso{
\link{analyse_Al2O3C_ITC}, \link{analyse_Al2O3C_CrossTalk},
\link{analyse_Al2O3C_Measurement}
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_OSLLxTxDecomposed.R
\name{calc_OSLLxTxDecomposed}
\alias{calc_OSLLxTxDecomposed}
\title{Calculate Lx/Tx ratio for decomposed CW-OSL signal components}
\usage{
calc_OSLLxTxDecomposed(
  Lx.data,
  Tx.data = NULL,
  OSL.component = 1L,
  sig0 = 0,
  digits = NULL
)
}
\arguments{
\item{Lx.data}{\link{data.frame} (\strong{required}): Component table created by
\code{OSLdecomposition::RLum.OSL_decomposition} and per default located
at \code{object@records[[...]]@info$COMPONENTS}.The value of \verb{$n[OSL.component]}
is set as \code{LnLx}. The value of \verb{$n.error[OSL.component]} is set as \code{LnLx.error}}

\item{Tx.data}{\link{data.frame} (\emph{optional}): Component table created by
\code{OSLdecomposition::RLum.OSL_decomposition} and per default located at
\code{object@records[[...]]@info$COMPONENTS}. The value of \verb{$n[OSL.component]}
is set as \code{TnTx}. The value of \verb{$n.error[OSL.component]} is set as \code{TnTx.error}}

\item{OSL.component}{\link{integer} or \link{character} (\emph{optional}):
a single index or a name describing which OSL signal component shall be evaluated.
This argument can either be the name of the OSL component assigned by
\code{OSLdecomposition::RLum.OSL_global_fitting} or the index of component.
Then \code{'1'} selects the fastest decaying component, \code{'2'} the
second fastest and so on. If not defined, the fastest decaying component is selected.}

\item{sig0}{\link{numeric} (\emph{with default}): allows adding an extra error component
to the final \code{Lx/Tx} error value (e.g., instrumental error).}

\item{digits}{\link{integer} (\emph{with default}): round numbers to the specified digits.
If digits is set to \code{NULL} nothing is rounded.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.

Slot \code{data} contains a \link{list} with the following structure:

\strong{@data}

\if{html}{\out{<div class="sourceCode">}}\preformatted{$LxTx.table (data.frame)
.. $ Net_LnLx
.. $ Net_LnLx.Error
.. $ Net_TnTx
.. $ Net_TnTx.Error
.. $ SN_RATIO_LnLx = NA,
.. $ SN_RATIO_TnTx = NA
.. $ LxTx
.. $ LxTx.Error
}\if{html}{\out{</div>}}
}
\description{
Calculate \code{Lx/Tx} ratios from a given set of decomposed
CW-OSL curves decomposed by \code{OSLdecomposition::RLum.OSL_decomposition}.
}
\section{Function version}{
 0.1.1
}
 

\section{How to cite}{
Mittelstrass, D., 2026. calc_OSLLxTxDecomposed(): Calculate Lx/Tx ratio for decomposed CW-OSL signal components. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Mittelstrass D., Schmidt C., Beyer J., Straessner A., 2019.
Automated identification and separation of quartz CW-OSL signal components with R.
talk presented at DLED 2019, Bingen, Germany
\url{http://luminescence.de/OSLdecomp_talk.pdf}\cr
}
\seealso{
\linkS4class{RLum.Data.Curve}, \link{fit_DoseResponseCurve},
\link{analyse_SAR.CWOSL}
}
\author{
Dirk Mittelstrass
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_AbanicoPlot.R
\name{plot_AbanicoPlot}
\alias{plot_AbanicoPlot}
\title{Function to create an Abanico Plot.}
\usage{
plot_AbanicoPlot(
  data,
  na.rm = TRUE,
  log.z = TRUE,
  z.0 = c("mean.weighted", "mean.weighted", "median"),
  dispersion = c("qr", "sd", "2sd"),
  plot.ratio = 0.75,
  rotate = FALSE,
  mtext = "",
  summary = c("n", "in.2s"),
  summary.pos = "sub",
  summary.method = c("MCM", "weighted", "unweighted"),
  legend = NULL,
  legend.pos = "topleft",
  stats = NULL,
  rug = FALSE,
  kde = TRUE,
  hist = FALSE,
  dots = FALSE,
  boxplot = FALSE,
  y.axis = TRUE,
  error.bars = FALSE,
  bar = NULL,
  bar.col = NULL,
  polygon.col = NULL,
  line = NULL,
  line.col = NULL,
  line.lty = NULL,
  line.label = NULL,
  grid.col = NULL,
  frame = 1,
  bw = "SJ",
  interactive = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} or \linkS4class{RLum.Results} object (\strong{required}):
for \code{data.frame} two columns: De (\code{data[,1]}) and De error (\code{data[,2]}).
To plot several data sets in one plot the data sets must be provided as
\code{list}, e.g. \code{list(data.1, data.2)}.\cr
For some \linkS4class{RLum.Results} objects, one or more lines (and
corresponding labels) are drawn automatically:
\itemize{
\item \link{calc_AverageDose} (ADM)
\item \link{calc_CentralDose} (CDM)
\item \link{calc_MaxDose} (MDM)
\item \link{calc_MinDose} (MAM)
\item \link{calc_FiniteMixture}
Alternative labels can be set via the \code{line.label} option. This behaviour
can be suppressed altogether by setting \code{line = NA}.
}}

\item{na.rm}{\link{logical} (\emph{with default}):
exclude \code{NA} values from the data set prior to any further operations.}

\item{log.z}{\link{logical} (\emph{with default}):
Option to display the z-axis in logarithmic scale. Default is \code{TRUE}.}

\item{z.0}{\link{character} or \link{numeric} (\emph{with default}):
User-defined central value, used for centring of data. One out of \code{"mean"},
\code{"mean.weighted"} and \code{"median"} or a numeric value (not its logarithm).
Default is \code{"mean.weighted"}.}

\item{dispersion}{\link{character} (\emph{with default}):
measure of dispersion, used for drawing the scatter polygon. One out of
\itemize{
\item \code{"qr"} (quartile range, default),
\item \code{"pnn"} (symmetric percentile range with \code{nn} the lower percentile, e.g.
\code{"p05"} indicating the range between 5 and 95 \%, or \code{"p10"} indicating
the range between 10 and 90 \%), or
\item \code{"sd"} (standard deviation) and
\item \code{"2sd"} (2 standard deviations),
}

The default is \code{"qr"}. Note that \code{"sd"} and \code{"2sd"} are only meaningful in
combination with \code{"z.0 = 'mean'"} because the unweighted mean is used to
centre the polygon.}

\item{plot.ratio}{\link{numeric} (\emph{with default}):
Relative space, given to the radial versus the cartesian plot part,
default is 0.75.}

\item{rotate}{\link{logical} (\emph{with default}):
Option to turn the plot by 90 degrees.}

\item{mtext}{\link{character} (\emph{with default}):
additional text below the plot title.}

\item{summary}{\link{character} (\emph{with default}):
add statistic measures of centrality and dispersion to the plot.
Can be one or more of several keywords. See details for available keywords.
Results differ depending on the log-option for the z-scale (see details).}

\item{summary.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"}) for the
statistical summary. Alternatively, the keyword \code{"sub"} may be
specified to place the summary below the plot header. However, this latter
option in only possible if \code{mtext} is not used.}

\item{summary.method}{\link{character} (\emph{with default}):
keyword indicating the method used to calculate the statistic summary.
One of \code{"MCM"} (default), \code{"weighted"} or \code{"unweighted"}.
See \link{calc_Statistics} for details.}

\item{legend}{\link{character} vector (\emph{optional}):
legend content to be added to the plot.}

\item{legend.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"})
for the legend to be plotted.}

\item{stats}{\link{character}:
additional labels of statistically important values in the plot.
One or more out of the following:
\itemize{
\item \code{"min"},
\item \code{"max"},
\item \code{"median"}.
}}

\item{rug}{\link{logical} (\emph{with default}):
Option to add a rug to the KDE part, to indicate the location of individual values.}

\item{kde}{\link{logical} (\emph{with default}):
Option to add a KDE plot to the dispersion part, default is \code{TRUE}.}

\item{hist}{\link{logical} (\emph{with default}):
Option to add a histogram to the dispersion part. Only meaningful when not
more than one data set is plotted.}

\item{dots}{\link{logical} (\emph{with default}):
Option to add a dot plot to the dispersion part. If number of dots exceeds
space in the dispersion part, a square indicates this.}

\item{boxplot}{\link{logical} (\emph{with default}):
Option to add a boxplot to the dispersion part, default is \code{FALSE}.}

\item{y.axis}{\link{logical} (\emph{with default}): Option to hide standard y-axis
labels and show 0 only.
Useful for data with small scatter. If you want to suppress the y-axis entirely
please use \code{yaxt == 'n'} (the standard \link[graphics:par]{graphics::par} setting) instead.}

\item{error.bars}{\link{logical} (\emph{with default}):
Option to show De-errors as error bars on De-points. Useful in combination
with \verb{y.axis = FALSE, bar.col = "none"}.}

\item{bar}{\link{numeric} (\emph{with default}):
option to add one or more dispersion bars (i.e., bar showing the 2-sigma range)
centred at the defined values. By default a bar is drawn according to \code{"z.0"}.
To omit the bar set \code{"bar = FALSE"}.}

\item{bar.col}{\link{character} or \link{numeric} (\emph{with default}):
colour of the dispersion bar. Default is \code{"grey60"}.}

\item{polygon.col}{\link{character} or \link{numeric} (\emph{with default}):
colour of the polygon showing the data scatter. Sometimes this
polygon may be omitted for clarity. To disable it use \code{FALSE} or
\code{polygon = FALSE}. Default is \code{"grey80"}.}

\item{line}{\link{numeric} or \linkS4class{RLum.Results}:
numeric values of the additional lines to be added. This can be set to \code{NA}
to suppress the line added automatically by some \code{RLum.Results} objects.}

\item{line.col}{\link{character} or \link{numeric}:
colour of the additional lines.}

\item{line.lty}{\link{integer}:
line type of additional lines.}

\item{line.label}{\link{character}:
labels for the additional lines.}

\item{grid.col}{\link{character} or \link{numeric} (\emph{with default}):
colour of the grid lines (originating at \verb{[0,0]} and stretching to
the z-scale). To disable grid lines use \code{FALSE}. Default is \code{"grey"}.}

\item{frame}{\link{numeric} (\emph{with default}):
option to modify the plot frame type. Can be one out of
\itemize{
\item \code{0} (no frame),
\item \code{1} (frame originates at 0,0 and runs along min/max isochrons),
\item \code{2} (frame embraces the 2-sigma bar),
\item \code{3} (frame embraces the entire plot as a rectangle).
}

Default is \code{1}.}

\item{bw}{\link{character} (\emph{with default}):
bin-width for KDE, choose a numeric value for manual setting.}

\item{interactive}{\link{logical} (\emph{with default}):
create an interactive abanico plot (requires the \code{'plotly'} package)}

\item{...}{Further plot arguments to pass (see \link[graphics:plot.default]{graphics::plot.default}).
Supported are: \code{main}, \code{sub}, \code{ylab}, \code{xlab}, \code{zlab}, \code{zlim}, \code{ylim}, \code{cex},
\code{lty}, \code{lwd}, \code{pch}, \code{col}, \code{at}, \code{breaks}. \code{xlab} must be
a vector of length two, specifying the upper and lower x-axis labels.}
}
\value{
Returns a plot object and, optionally, a list with plot calculus data.
}
\description{
A plot is produced which allows comprehensive presentation of data precision
and its dispersion around a central value as well as illustration of a
kernel density estimate, histogram and/or dot plot of the dose values.
}
\details{
The Abanico Plot is a combination of the classic Radial Plot
\link{plot_RadialPlot} and a kernel density estimate plot (e.g
\link{plot_KDE}. It allows straightforward visualisation of data precision,
error scatter around a user-defined central value and the combined
distribution of the values, on the actual scale of the measured data (e.g.
seconds, equivalent dose, years). The principle of the plot is shown in
Galbraith & Green (1990). The function authors are thankful for the
thought-provoking figure in this article.

The semi circle (z-axis) of the classic Radial Plot is bent to a straight
line here, which actually is the basis for combining this polar (radial)
part of the plot with any other Cartesian visualisation method
(KDE, histogram, PDF and so on). Note that the plot allows displaying
two measures of distribution. One is the 2-sigma
bar, which illustrates the spread in value errors, and the other is the
polygon, which stretches over both parts of the Abanico Plot (polar and
Cartesian) and illustrates the actual spread in the values themselves.

Since the 2-sigma-bar is a polygon, it can be (and is) filled with shaded
lines. To change density (lines per inch, default is 15) and angle (default
is 45 degrees) of the shading lines, specify these parameters. See
\code{?polygon()} for further help.

The Abanico Plot supports other than the weighted mean as measure of
centrality. When it is obvious that the data
is not (log-)normally distributed, the mean (weighted or not) cannot be a
valid measure of centrality and hence central dose. Accordingly, the median
and the weighted median can be chosen as well to represent a proper measure
of centrality (e.g. \code{centrality = "median.weighted"}). Also
user-defined numeric values (e.g. from the central age model) can be used if
this appears appropriate.

The proportion of the polar part and the cartesian part of the Abanico Plot
can be modified for display reasons (\code{plot.ratio = 0.75}). By default,
the polar part spreads over 75 \% and leaves 25 \% for the part that
shows the KDE graph.

A statistic summary, i.e. a collection of statistic measures of
centrality and dispersion (and further measures) can be added by specifying
one or more of the following keywords:
\itemize{
\item \code{"n"} (number of samples)
\item \code{"mean"} (mean De value)
\item \code{"median"} (median of the De values)
\item \code{"sd.rel"} (relative standard deviation in percent)
\item \code{"sd.abs"} (absolute standard deviation)
\item \code{"se.rel"} (relative standard error)
\item \code{"se.abs"} (absolute standard error)
\item \code{"in.2s"} (percent of samples in 2-sigma range)
\item \code{"kurtosis"} (kurtosis)
\item \code{"skewness"} (skewness)
}

\strong{Note:} the input data for the statistic summary is sent to the function
\link{calc_Statistics} depending on the log-option for the z-scale. If
\code{"log.z = TRUE"}, the summary is based on the logarithms of the input
data. If \code{"log.z = FALSE"} the linearly scaled data is used.

\strong{Note:} \link{calc_Statistics} calculates these statistic
measures in three different ways: \code{unweighted}, \code{weighted} and
\code{MCM-based} (i.e., based on Monte Carlo Methods). By default, the
MCM-based version is used. If you wish to use another method, indicate this
with the appropriate keyword using the argument \code{summary.method}.

The optional parameter \code{layout} allows more sophisticated ways to modify
the entire plot. Each element of the plot can be addressed and its properties
can be defined. This includes font type, size and decoration, colours and
sizes of all plot items. To infer the definition of a specific layout style
cf. \link{get_Layout} or type, e.g. for the layout type \code{"journal"}.
A layout type can be modified by the user by
assigning new values to the list object.

It is possible for the z-scale to specify where ticks are to be drawn
by using the parameter \code{at}, e.g. \code{at = seq(80, 200, 20)}, cf. function
documentation of \code{axis}. Specifying tick positions manually overrides a
\code{zlim}-definition.
}
\section{Function version}{
 0.1.21
}

\examples{

## load example data and recalculate to Gray
data(ExampleData.DeValues, envir = environment())
ExampleData.DeValues <- ExampleData.DeValues$CA1

## plot the example data straightforward
plot_AbanicoPlot(data = ExampleData.DeValues)

## now with linear z-scale
plot_AbanicoPlot(data = ExampleData.DeValues,
                 log.z = FALSE)

## now with output of the plot parameters
plot1 <- plot_AbanicoPlot(data = ExampleData.DeValues)
str(plot1)
plot1$zlim

## now with adjusted z-scale limits
plot_AbanicoPlot(data = ExampleData.DeValues,
                 zlim = c(10, 200))

## now with adjusted x-scale limits
plot_AbanicoPlot(data = ExampleData.DeValues,
                 xlim = c(0, 20))

## now with rug to indicate individual values in KDE part
plot_AbanicoPlot(data = ExampleData.DeValues,
                 rug = TRUE)

## now with a smaller bandwidth for the KDE plot
plot_AbanicoPlot(data = ExampleData.DeValues,
                 bw = 0.04)

## now with a histogram instead of the KDE plot
plot_AbanicoPlot(data = ExampleData.DeValues,
                 hist = TRUE,
                 kde = FALSE)

## now with a KDE plot and histogram with manual number of bins
plot_AbanicoPlot(data = ExampleData.DeValues,
                 hist = TRUE,
                 breaks = 20)

## now with a KDE plot and a dot plot
plot_AbanicoPlot(data = ExampleData.DeValues,
                 dots = TRUE)

## now with user-defined plot ratio
plot_AbanicoPlot(data = ExampleData.DeValues,
                 plot.ratio = 0.5)
## now with user-defined central value
plot_AbanicoPlot(data = ExampleData.DeValues,
                 z.0 = 70)

## now with median as central value
plot_AbanicoPlot(data = ExampleData.DeValues,
                 z.0 = "median")

## now with the 17-83 percentile range as definition of scatter
plot_AbanicoPlot(data = ExampleData.DeValues,
                 z.0 = "median",
                 dispersion = "p17")

## now with user-defined green line for minimum age model
CAM <- calc_CentralDose(ExampleData.DeValues,
                        plot = FALSE)
plot_AbanicoPlot(data = CAM,
                 line.col = "darkseagreen")

## now create plot with legend, colour, different points and smaller scale
plot_AbanicoPlot(data = ExampleData.DeValues,
                 legend = "Sample 1",
                 col = "tomato4",
                 bar.col = "peachpuff",
                 pch = "R",
                 cex = 0.8)

## now without 2-sigma bar, polygon, grid lines and central value line
plot_AbanicoPlot(data = ExampleData.DeValues,
                 bar.col = FALSE,
                 polygon.col = FALSE,
                 grid.col = FALSE,
                 y.axis = FALSE,
                 lwd = 0)

## now with direct display of De errors, without 2-sigma bar
plot_AbanicoPlot(data = ExampleData.DeValues,
                 bar.col = FALSE,
                 ylab = "",
                 y.axis = FALSE,
                 error.bars = TRUE)

## now with user-defined axes labels
plot_AbanicoPlot(data = ExampleData.DeValues,
                 xlab = c("Data error (\%)",
                          "Data precision"),
                 ylab = "Scatter",
                 zlab = "Equivalent dose [Gy]")

## now with minimum, maximum and median value indicated
plot_AbanicoPlot(data = ExampleData.DeValues,
                 stats = c("min", "max", "median"))

## now with a brief statistical summary as subheader
plot_AbanicoPlot(data = ExampleData.DeValues,
                 summary = c("n", "in.2s"))

## now with another statistical summary
plot_AbanicoPlot(data = ExampleData.DeValues,
                 summary = c("mean.weighted", "median"),
                 summary.pos = "topleft")

## now a plot with two 2-sigma bars for one data set
plot_AbanicoPlot(data = ExampleData.DeValues,
                 bar = c(30, 100))

## now the data set is split into sub-groups, one is manipulated
data.1 <- ExampleData.DeValues[1:30,]
data.2 <- ExampleData.DeValues[31:62,] * 1.3

## now a common dataset is created from the two subgroups
data.3 <- list(data.1, data.2)

## now the two data sets are plotted in one plot
plot_AbanicoPlot(data = data.3)

## now with some graphical modification
plot_AbanicoPlot(data = data.3,
                 z.0 = "median",
                 col = c("steelblue4", "orange4"),
                 bar.col = c("steelblue3", "orange3"),
                 polygon.col = c("steelblue1", "orange1"),
                 pch = c(2, 6),
                 angle = c(30, 50),
                 summary = c("n", "in.2s", "median"))

## create Abanico plot with predefined layout definition
plot_AbanicoPlot(data = ExampleData.DeValues,
                 layout = "journal")

## now with predefined layout definition and further modifications
plot_AbanicoPlot(
 data = data.3,
 z.0 = "median",
 layout = "journal",
 col = c("steelblue4", "orange4"),
 bar.col = adjustcolor(c("steelblue3", "orange3"),
                         alpha.f = 0.5),
 polygon.col = c("steelblue3", "orange3"))

## for further information on layout definitions see documentation
## of function get_Layout()

## now with manually added plot content
## create empty plot with numeric output
AP <- plot_AbanicoPlot(data = ExampleData.DeValues,
                       pch = NA)

## identify data in 2 sigma range
in_2sigma <- AP$data[[1]]$data.in.2s

## restore function-internal plot parameters
par(AP$par)

## add points inside 2-sigma range
points(x = AP$data[[1]]$precision[in_2sigma],
       y = AP$data[[1]]$std.estimate.plot[in_2sigma],
       pch = 16)

## add points outside 2-sigma range
points(x = AP$data[[1]]$precision[!in_2sigma],
       y = AP$data[[1]]$std.estimate.plot[!in_2sigma],
       pch = 1)

} 

\section{How to cite}{
Dietze, M., Kreutzer, S., Colombo, M., 2026. plot_AbanicoPlot(): Function to create an Abanico Plot.. Function version 0.1.21. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Galbraith, R. & Green, P., 1990. Estimating the component ages
in a finite mixture. International Journal of Radiation Applications and
Instrumentation. Part D. Nuclear Tracks and Radiation Measurements 17 (3),
197-206.

Dietze, M., Kreutzer, S., Burow, C., Fuchs, M.C., Fischer, M., Schmidt, C., 2015.
The abanico plot: visualising chronometric data with individual standard errors.
Quaternary Geochronology. doi:10.1016/j.quageo.2015.09.003
}
\seealso{
\link{plot_RadialPlot}, \link{plot_KDE},
\link{plot_Histogram}, \link{plot_ViolinPlot}
}
\author{
Michael Dietze, GFZ Potsdam (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
Inspired by a plot introduced by Galbraith & Green (1990)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_rightAnswer.R
\name{get_rightAnswer}
\alias{get_rightAnswer}
\title{Function to get the right answer}
\usage{
get_rightAnswer(...)
}
\arguments{
\item{...}{you can pass an infinite number of further arguments}
}
\value{
Returns the right answer
}
\description{
This function returns just the right answer
}
\section{Function version}{
 0.1.0
}

\examples{

## you really want to know?
get_rightAnswer()

}
\author{
inspired by R.G.
, RLum Developer Team} 

\section{How to cite}{
NA, NA, , , 2026. get_rightAnswer(): Function to get the right answer. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_Statistics.R
\name{calc_Statistics}
\alias{calc_Statistics}
\title{Function to calculate statistic measures}
\usage{
calc_Statistics(
  data,
  weight.calc = c("square", "reciprocal"),
  digits = NULL,
  n.MCM = NULL,
  na.rm = TRUE
)
}
\arguments{
\item{data}{\link{data.frame} or \linkS4class{RLum.Results} object (\strong{required}):
for \link{data.frame} two columns: De (\code{data[, 1]}) and De error (\code{data[, 2]}).}

\item{weight.calc}{\link{character} (\emph{with default}):
type of weight calculation. One out of \code{"reciprocal"} (weight is 1/error),
\code{"square"} (weight is 1/error^2). Default is \code{"square"}.}

\item{digits}{\link{integer} (\emph{with default}):
number of decimal places to be used when rounding numbers. If set to \code{NULL}
(default), no rounding occurs.}

\item{n.MCM}{\link{numeric} (\emph{with default}):
number of samples drawn for Monte Carlo-based statistics.
\code{NULL} (the default) disables MC runs.}

\item{na.rm}{\link{logical} (\emph{with default}):
indicating whether \code{NA} values should be stripped before the computation proceeds.}
}
\value{
Returns a list with weighted and unweighted statistic measures.
}
\description{
This function calculates a number of descriptive statistics for estimates
with a given standard error (SE), most fundamentally using error-weighted approaches.
}
\details{
The option to use Monte Carlo Methods (\code{n.MCM}) allows calculating
all descriptive statistics based on random values. The distribution of these
random values is based on the Normal distribution with \code{De} values as
means and \code{De_error} values as one standard deviation. Increasing the
number of MCM-samples linearly increases computation time. On a Lenovo X230
machine evaluation of 25 Aliquots with n.MCM = 1000 takes 0.01 s, with
n = 100000, ca. 1.65 s. It might be useful to work with logarithms of these
values. See Dietze et al. (2016, Quaternary Geochronology) and the function
\link{plot_AbanicoPlot} for details.
}
\section{Function version}{
 0.1.7
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())

## show a rough plot of the data to illustrate the non-normal distribution
plot_KDE(ExampleData.DeValues$BT998)

## calculate statistics and show output
str(calc_Statistics(ExampleData.DeValues$BT998))

\dontrun{
## now the same for 10000 normal distributed random numbers with equal errors
x <- as.data.frame(cbind(rnorm(n = 10^5, mean = 0, sd = 1),
                         rep(0.001, 10^5)))

## note the congruent results for weighted and unweighted measures
str(calc_Statistics(x))
}

}
\author{
Michael Dietze, GFZ Potsdam (Germany)
, RLum Developer Team} 

\section{How to cite}{
Dietze, M., 2026. calc_Statistics(): Function to calculate statistic measures. Function version 0.1.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R
\name{structure_RLum}
\alias{structure_RLum}
\alias{structure_RLum,list-method}
\alias{structure_RLum,RLum.Analysis-method}
\title{General structure function for RLum-class objects}
\usage{
structure_RLum(object, ...)

\S4method{structure_RLum}{list}(object, ...)

\S4method{structure_RLum}{RLum.Analysis}(object, fullExtent = FALSE)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum}}

\item{...}{further arguments passed to the specific class method}

\item{fullExtent}{\link{logical} (\emph{with default}):
extends the returned \code{data.frame} to its full extent, i.e. all info elements
are part of the return as well. The default value is \code{FALSE} as the data
frame might become rather big.}
}
\value{
Returns a \link{data.frame} with structure of the object.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the input object, the corresponding
function will be selected.
}
\section{Functions}{
\itemize{
\item \code{structure_RLum(list)}: Returns a list of data frames containing the structure of each \linkS4class{RLum}
object in the input list.

\item \code{structure_RLum(RLum.Analysis)}: Returns the structure of an \linkS4class{RLum.Analysis} object.

}}
\section{Function version}{
 0.2.0
}

\examples{

## load example data
data(ExampleData.XSYG, envir = environment())

## show structure
structure_RLum(OSL.SARMeasurement$Sequence.Object)

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Data.Image},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. structure_RLum(): General structure function for RLum-class objects. Function version 0.2.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R,
%   R/RLum.Data.Image-class.R, R/RLum.Data.Spectrum-class.R,
%   R/RLum.Results-class.R
\name{set_RLum}
\alias{set_RLum}
\alias{set_RLum,RLum.Analysis-method}
\alias{set_RLum,RLum.Data.Curve-method}
\alias{set_RLum,RLum.Data.Image-method}
\alias{set_RLum,RLum.Data.Spectrum-method}
\alias{set_RLum,RLum.Results-method}
\title{General setter function for RLum-class objects}
\usage{
set_RLum(class, originator, .uid = create_UID(), .pid = NA_character_, ...)

\S4method{set_RLum}{RLum.Analysis}(
  class,
  originator,
  .uid,
  .pid,
  protocol = NA_character_,
  records = list(),
  info = list()
)

\S4method{set_RLum}{RLum.Data.Curve}(
  class,
  originator,
  .uid,
  .pid,
  recordType = NA_character_,
  curveType = NA_character_,
  data = matrix(0, ncol = 2),
  info = list()
)

\S4method{set_RLum}{RLum.Data.Image}(
  class,
  originator,
  .uid,
  .pid,
  recordType = "Image",
  curveType = NA_character_,
  data = array(),
  info = list()
)

\S4method{set_RLum}{RLum.Data.Spectrum}(
  class,
  originator,
  .uid,
  .pid,
  recordType = "Spectrum",
  curveType = NA_character_,
  data = matrix(),
  info = list()
)

\S4method{set_RLum}{RLum.Results}(class, originator, .uid, .pid, data = list(), info = list())
}
\arguments{
\item{class}{\link{character} (\strong{required}):
name of the S4 class to create, must correspond to one of the \linkS4class{RLum}
classes.}

\item{originator}{\link{character} (\emph{automatic}):
contains the name of the calling function (the function that produces this object);
can be set manually.}

\item{.uid}{\link{character} (\emph{automatic}):
unique ID for this object, by default set using the internal C++ function
\code{create_UID}.}

\item{.pid}{\link{character} (\emph{with default}):
option to provide a parent id for nesting at will.}

\item{...}{further arguments passed to the specific class method}

\item{protocol}{\link{character} (\emph{optional}):
sets protocol type for analysis object. Value may be used by subsequent analysis functions.}

\item{records}{\link{list} (\emph{optional}):
list of \linkS4class{RLum.Analysis} objects}

\item{info}{\link{list} (\emph{optional}):
a list containing additional info data for the object.}

\item{recordType}{\link{character} (\emph{optional}):
record type (e.g., "OSL")}

\item{curveType}{\link{character} (\emph{optional}):
curve type (e.g., "predefined" or "measured")}

\item{data}{\link{matrix} or \link{list} (\emph{with default}):
a matrix containing raw curve data or a list containing the data to be
stored in the object (for \linkS4class{RLum.Results} objects) . If \code{data} itself
is a \code{RLum.Data.Curve}-object this can be used to re-construct the object,
i.e. modified parameters except \code{.uid}, \code{.pid} and \code{originator}. The rest
will be subject to copy and paste unless provided.}
}
\value{
An object of the specified \linkS4class{RLum} class.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the given class, the corresponding
method to create an object from this class will be selected.
}
\section{Functions}{
\itemize{
\item \code{set_RLum(RLum.Analysis)}: Construction method for \linkS4class{RLum.Analysis} objects.

\item \code{set_RLum(RLum.Data.Curve)}: Construction method for \linkS4class{RLum.Data.Curve} objects.

\item \code{set_RLum(RLum.Data.Image)}: Construction method for \linkS4class{RLum.Data.Image} objects.

\item \code{set_RLum(RLum.Data.Spectrum)}: Construction method for \linkS4class{RLum.Data.Spectrum} objects.

\item \code{set_RLum(RLum.Results)}: Construction method for \linkS4class{RLum.Results} objects.

}}
\section{Function version}{
 0.3.0
}

\examples{

## produce empty objects from each class
set_RLum(class = "RLum.Data.Curve")
set_RLum(class = "RLum.Data.Spectrum")
set_RLum(class = "RLum.Data.Spectrum")
set_RLum(class = "RLum.Analysis")
set_RLum(class = "RLum.Results")

## produce a curve object with arbitrary curve values
object <- set_RLum(
class = "RLum.Data.Curve",
curveType = "arbitrary",
recordType = "OSL",
data = matrix(c(1:100,exp(-c(1:100))),ncol = 2))

## plot this curve object
plot_RLum(object)

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Data.Image},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. set_RLum(): General setter function for RLum-class objects. Function version 0.3.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R, R/RLum.Results-class.R
\name{length_RLum}
\alias{length_RLum}
\alias{length_RLum,RLum.Analysis-method}
\alias{length_RLum,RLum.Data.Curve-method}
\alias{length_RLum,RLum.Results-method}
\title{Length retrieval function for RLum-class objects}
\usage{
length_RLum(object)

\S4method{length_RLum}{RLum.Analysis}(object)

\S4method{length_RLum}{RLum.Data.Curve}(object)

\S4method{length_RLum}{RLum.Results}(object)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum}}
}
\value{
An \link{integer} indicating the length of the object.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the input object, the corresponding
function will be selected.
}
\section{Functions}{
\itemize{
\item \code{length_RLum(RLum.Analysis)}: Returns the number of records stored in the object.

\item \code{length_RLum(RLum.Data.Curve)}: Returns the number of channels in the curve, which is the maximum of the
value time/temperature of the curve (corresponding to the stimulation
length).

\item \code{length_RLum(RLum.Results)}: Returns the number of stored data elements.

}}
\section{Function version}{
 0.1.0
}

\seealso{
\linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Image},
\linkS4class{RLum.Data.Spectrum},
\linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2025. length_RLum(): Length retrieval function for RLum-class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., 2025. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.1.2. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_RLum.R
\name{merge_RLum}
\alias{merge_RLum}
\title{General merge function for RLum-class objects}
\usage{
merge_RLum(objects, ...)
}
\arguments{
\item{objects}{\link{list} of \linkS4class{RLum} (\strong{required}):
list of S4 object of class \code{RLum}.}

\item{...}{further arguments that one might want to pass to the specific merge function}
}
\value{
Returns an \linkS4class{RLum.Analysis} object of class if any of the inputs is
of that class. Otherwise, it returns an object of the same type as the
input.
}
\description{
The function provides a generalised access point for merging specific
\linkS4class{RLum} objects. Depending on the input object, the
corresponding merge function will be selected.  Allowed arguments can be
found in the documentation of each merge function.
Empty list elements (\code{NULL}) are automatically removed from the input list.

\tabular{lll}{
\strong{object} \tab \tab \strong{corresponding merge function} \cr
\linkS4class{RLum.Data.Curve} \tab -> \tab \link{merge_RLum.Data.Curve} \cr
\linkS4class{RLum.Data.Spectrum} \tab -> \tab \link{merge_RLum.Data.Spectrum} \cr
\linkS4class{RLum.Analysis} \tab -> \tab \link{merge_RLum.Analysis} \cr
\linkS4class{RLum.Results} \tab -> \tab \link{merge_RLum.Results}
}
}
\note{
So far merging of \linkS4class{RLum.Data.Image} objects is not supported.
}
\section{Function version}{
 0.1.3
}

\examples{

##Example based using data and from the calc_CentralDose() function

##load example data
data(ExampleData.DeValues, envir = environment())

##apply the central dose model 1st time
temp1 <- calc_CentralDose(ExampleData.DeValues$CA1)

##apply the central dose model 2nd time
temp2 <- calc_CentralDose(ExampleData.DeValues$CA1)

##merge the results and store them in a new object
temp.merged <- get_RLum(merge_RLum(objects = list(temp1, temp2)))

}
\seealso{
\linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. merge_RLum(): General merge function for RLum-class objects. Function version 0.1.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_pIRIRSequence.R
\name{analyse_pIRIRSequence}
\alias{analyse_pIRIRSequence}
\title{Analyse post-IR IRSL measurement sequences}
\usage{
analyse_pIRIRSequence(
  object,
  signal_integral,
  background_integral,
  integral_input = c("channel", "measurement"),
  dose.points = NULL,
  sequence.structure = c("TL", "IR50", "pIRIR225"),
  plot = TRUE,
  plot_singlePanels = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} or \link{list} of
\linkS4class{RLum.Analysis} objects (\strong{required}):
input object containing data for analysis.
If a \link{list} is provided the functions tries to iterate over each element
in the list.}

\item{signal_integral}{\link{integer} (\strong{required}):
vector of channels for the signal integral. Provide this value as a list
for different integration limits for the different IRSL curves.}

\item{background_integral}{\link{integer} (\strong{required}):
vector of channels for the background integral. Provide this value as a list
for different integration limits for the different IRSL curves.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{dose.points}{\link{numeric} (\emph{optional}):
a numeric vector containing the dose points values. Using this argument overwrites dose point
values in the signal curves.}

\item{sequence.structure}{\link{vector} \link{character} (\emph{with default}):
specifies the general sequence structure. Allowed values are \code{"TL"} and
any \code{"IR"} combination (e.g., \code{"IR50"},\code{"pIRIR225"}).
Additionally a parameter \code{"EXCLUDE"} is allowed to exclude curves from
the analysis (Note: If a preheat without PMT measurement is used, i.e.
preheat as none TL, remove the TL step.)}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}):
enable/disable plotting of the results in a single windows for each plot.
Ignored if \code{plot = FALSE}.}

\item{...}{further arguments that will be passed to
\link{analyse_SAR.CWOSL} and \link{plot_DoseResponseCurve}. Furthermore, the
arguments \code{main} (headers), \code{log} (IRSL curves), \code{cex} (control
the size) and \code{mtext.outer} (additional text on the plot area) can be passed to influence the plotting. If the input
is a list, \code{main} can be passed as \link{vector} or \link{list}.}
}
\value{
Plots (\emph{optional}) and an \linkS4class{RLum.Results} object is
returned containing the following elements:

\tabular{lll}{
\strong{DATA.OBJECT} \tab \strong{TYPE} \tab \strong{DESCRIPTION} \cr
\code{..$data} : \tab  \code{data.frame} \tab Table with De values \cr
\code{..$LnLxTnTx.table} : \tab \code{data.frame} \tab with the \code{LnLxTnTx} values \cr
\code{..$rejection.criteria} : \tab \link{data.frame} \tab rejection criteria \cr
\code{..$Formula} : \tab \link{list} \tab Function used for fitting of the dose response curve \cr
\code{..$call} : \tab \link{call} \tab the original function call
}

The output should be accessed using the function \link{get_RLum}.
}
\description{
The function performs an analysis of post-IR IRSL sequences including curve
fitting on \linkS4class{RLum.Analysis} objects.
}
\details{
To allow post-IR IRSL protocol (Thomsen et al., 2008) measurement
analyses, this function has been written as extended wrapper for function
\link{analyse_SAR.CWOSL}, thus facilitating an entire sequence analysis in
one run. With this, its functionality is strictly limited by the
functionality provided by  \link{analyse_SAR.CWOSL}.

\strong{Defining the sequence structure}

The argument \code{sequence.structure} expects a shortened pattern of your
sequence structure and was mainly introduced to ease the use of the function.
For example: If your measurement data contains the following curves: \code{TL},
\code{IRSL}, \code{IRSL}, \code{TL}, \code{IRSL}, \code{IRSL}, the pattern in \code{sequence.structure}
becomes \code{c('TL', 'IRSL', 'IRSL')}. The second part of your sequence for one
cycle should be similar and can be discarded. If this is not the case (e.g.,
additional hotbleach) such curves must be removed before using the function.

\strong{If the input is a \code{list}}

If the input is a list of \linkS4class{RLum.Analysis} objects, every argument
can be provided as list to allow
for different sets of parameters for every single input element.
For further information see \link{analyse_SAR.CWOSL}.
}
\note{
Best graphical output can be achieved by using the function \code{pdf}
with the following options:

\code{pdf(file = "<YOUR FILENAME>", height = 18, width = 18)}
}
\section{Function version}{
 0.2.8
}

\examples{

### NOTE: For this example existing example data are used. These data are non pIRIR data.
###
##(1) Compile example data set based on existing example data (SAR quartz measurement)
##(a) Load example data
data(ExampleData.BINfileData, envir = environment())

##(b) Transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

##(c) Grep curves and exclude the last two (one TL and one IRSL)
object <- get_RLum(object, record.id = c(-29,-30))

##(d) Define new sequence structure and set new RLum.Analysis object
sequence.structure  <- c(1,2,2,3,4,4)
sequence.structure <- as.vector(sapply(seq(0,length(object)-1,by = 4),
                                       function(x){sequence.structure + x}))

object <-  sapply(1:length(sequence.structure), function(x){
  object[[sequence.structure[x]]]
})

object <- set_RLum(class = "RLum.Analysis", records = object, protocol = "pIRIR")

##(2) Perform pIRIR analysis (for this example with quartz OSL data!)
## Note: output as single plots to avoid problems with this example
results <- analyse_pIRIRSequence(object,
     signal_integral = 1:2,
     background_integral = 900:1000,
     fit.method = "EXP",
     sequence.structure = c("TL", "pseudoIRSL1", "pseudoIRSL2"),
     main = "Pseudo pIRIR data set based on quartz OSL",
     plot_singlePanels = TRUE)


##(3) Perform pIRIR analysis (for this example with quartz OSL data!)
## Alternative for PDF output, uncomment and complete for usage
\dontrun{
tempfile <- tempfile(fileext = ".pdf")
pdf(file = tempfile, height = 18, width = 18)
  results <- analyse_pIRIRSequence(object,
         signal_integral = 1:2,
         background_integral = 900:1000,
         fit.method = "EXP",
         main = "Pseudo pIRIR data set based on quartz OSL")

  dev.off()
}

} 

\section{How to cite}{
Kreutzer, S., 2026. analyse_pIRIRSequence(): Analyse post-IR IRSL measurement sequences. Function version 0.2.8. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Murray, A.S., Wintle, A.G., 2000. Luminescence dating of quartz
using an improved single-aliquot regenerative-dose protocol. Radiation
Measurements 32, 57-73. \doi{10.1016/S1350-4487(99)00253-X}

Thomsen, K.J., Murray, A.S., Jain, M., Boetter-Jensen, L., 2008. Laboratory
fading rates of various luminescence signals from feldspar-rich sediment
extracts. Radiation Measurements 43, 1474-1486.
\doi{10.1016/j.radmeas.2008.06.002}
}
\seealso{
\link{analyse_SAR.CWOSL}, \link{calc_OSLLxTxRatio},
\link{plot_DoseResponseCurve}, \link{fit_DoseResponseCurve}
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Results},
\link{get_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data-class.R, R/RLum.Results-class.R,
%   R/Risoe.BINfileData-class.R
\name{view}
\alias{view}
\alias{view,RLum.Analysis-method}
\alias{view,RLum.Data-method}
\alias{view,RLum.Results-method}
\alias{view,Risoe.BINfileData-method}
\title{Convenience data visualisation function}
\usage{
view(object, ...)

\S4method{view}{RLum.Analysis}(object, ...)

\S4method{view}{RLum.Data}(object, ...)

\S4method{view}{RLum.Results}(object, element = 1, ...)

\S4method{view}{Risoe.BINfileData}(object, ...)
}
\arguments{
\item{object}{(\strong{required}) object to view}

\item{...}{further arguments passed to the specific class method}

\item{element}{\link{integer} (\emph{with default}):
index of the element to display.}
}
\value{
\code{NULL} and opens the data viewer.
}
\description{
Invokes the \link[utils:View]{utils::View} function tailored to objects in the package.
If started from RStudio, it uses the RStudio viewer.
}
\section{Functions}{
\itemize{
\item \code{view(RLum.Analysis)}: View method for \linkS4class{RLum.Analysis} objects.

\item \code{view(RLum.Data)}: View method for \linkS4class{RLum.Data} objects.

\item \code{view(RLum.Results)}: View method for \linkS4class{RLum.Results} objects.

\item \code{view(Risoe.BINfileData)}: View method for \linkS4class{Risoe.BINfileData} objects.

}}
\seealso{
\code{\link[utils:View]{utils::View()}}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. view(): Convenience data visualisation function. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_Wavelength2Energy.R
\name{convert_Wavelength2Energy}
\alias{convert_Wavelength2Energy}
\title{Emission Spectra Conversion from Wavelength to Energy Scales (Jacobian Conversion)}
\usage{
convert_Wavelength2Energy(object, digits = 3L, order = FALSE)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Spectrum}, \link{data.frame}, \link{matrix} (\strong{required}):
input object to be converted. If the input is not an \code{RLum.Data.Spectrum}
object, the first column is always treated as the wavelength
column. The function supports a list of allowed input objects.}

\item{digits}{\link{integer} (\emph{with default}):
number of digits on the returned energy axis.}

\item{order}{\link{logical} (\emph{with default}): enable/disable sorting of the
values in ascending energy order. After the conversion, the longest
wavelength has the lowest energy value and the shortest wavelength the
highest. While this is correct, some R functions expect increasing x-values.}
}
\value{
The same object class as provided as input is returned.
}
\description{
The function provides a convenient and fast way to convert emission spectra
wavelength to energy scales. The function works on \linkS4class{RLum.Data.Spectrum},
\link{data.frame} and \link{matrix} and a \link{list} of such objects. The function was
written to smooth the workflow while analysing emission spectra data. This
is in particular useful if you want to further treat your data and apply,
e.g., a signal deconvolution.
}
\details{
The intensity of the spectrum is re-calculated using the following approach
to recalculate wavelength and corresponding intensity values
(e.g., Appendix 4 in Blasse and Grabmaier, 1994; Mooney and Kambhampati, 2013):

\deqn{\phi_{E} = \phi_{\lambda} * \lambda^2 / (hc)}

with
\eqn{\phi_{E}} the intensity per interval of energy \eqn{E} (1/eV),
\eqn{\phi_{\lambda}} the intensity per interval of wavelength \eqn{\lambda}
(1/nm) and
\eqn{h} (eV * s) the Planck constant and \eqn{c} (nm/s) the velocity of light.

For transforming the wavelength axis (x-values) the equation as follow is used

\deqn{E = hc/\lambda}
}
\note{
This conversion works solely for emission spectra. In case of absorption spectra only
the x-axis has to be converted.
}
\section{Function version}{
 0.1.1
}

\examples{

##=====================##
##(1) Literature example after Mooney et al. (2013)
##(1.1) create matrix
m <- matrix(
  data = c(seq(400, 800, 50), rep(1, 9)), ncol = 2)

##(1.2) set plot function to reproduce the literature figure
p <- function(m) {
 plot(x = m[, 1], y = m[, 2])
 polygon(
 x = c(m[, 1], rev(m[, 1])),
 y = c(m[, 2], rep(0, nrow(m))))
 for (i in 1:nrow(m)) {
  lines(x = rep(m[i, 1], 2), y = c(0, m[i, 2]))
 }
}

##(1.3) plot curves
par(mfrow = c(1,2))
p(m)
p(convert_Wavelength2Energy(m))

##=====================##
##(2) Another example using density curves
##create dataset
xy <- density(
 c(rnorm(n = 100, mean = 500, sd = 20),
 rnorm(n = 100, mean = 800, sd = 20)))
xy <- data.frame(xy$x, xy$y)

##plot
par(mfrow = c(1,2))
plot(
 xy,
 type = "l",
 xlim = c(150, 1000),
 xlab = "Wavelength [nm]",
 ylab = "Luminescence [a.u.]"
)
plot(
 convert_Wavelength2Energy(xy),
 xy$y,
 type = "l",
 xlim = c(1.23, 8.3),
 xlab = "Energy [eV]",
 ylab = "Luminescence [a.u.]"
)

} 

\section{How to cite}{
Kreutzer, S., 2026. convert_Wavelength2Energy(): Emission Spectra Conversion from Wavelength to Energy Scales (Jacobian Conversion). Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Blasse, G., Grabmaier, B.C., 1994. Luminescent Materials. Springer.

Mooney, J., Kambhampati, P., 2013. Get the Basics Right: Jacobian Conversion of Wavelength and
Energy Scales for Quantitative Analysis of Emission Spectra. J. Phys. Chem. Lett. 4, 3316–3318.
\doi{10.1021/jz401508t}

Mooney, J., Kambhampati, P., 2013. Correction to “Get the Basics Right: Jacobian Conversion of
Wavelength and Energy Scales for Quantitative Analysis of Emission Spectra.” J. Phys. Chem. Lett. 4,
3316–3318. \doi{10.1021/jz401508t}

\strong{Further reading}

Angulo, G., Grampp, G., Rosspeintner, A., 2006. Recalling the appropriate representation of
electronic spectra. Spectrochimica Acta Part A: Molecular and Biomolecular Spectroscopy 65,
727–731. \doi{10.1016/j.saa.2006.01.007}

Wang, Y., Townsend, P.D., 2013. Potential problems in collection and data processing of
luminescence signals. Journal of Luminescence 142, 202–211. \doi{10.1016/j.jlumin.2013.03.052}
}
\seealso{
\linkS4class{RLum.Data.Spectrum}, \link{plot_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_SourceDoseRate.R
\name{calc_SourceDoseRate}
\alias{calc_SourceDoseRate}
\title{Calculation of the source dose rate via the date of measurement}
\usage{
calc_SourceDoseRate(
  measurement.date,
  calib.date,
  calib.dose.rate,
  calib.error,
  source.type = c("Sr-90", "Am-214", "Co-60", "Cs-137"),
  dose.rate.unit = c("Gy/s", "Gy/min"),
  predict = NULL
)
}
\arguments{
\item{measurement.date}{\link{character} or \link{Date} (\emph{with default}):
Date of measurement in \code{"YYYY-MM-DD"} format. If no value is provided, the
date will be set to today. The argument can be provided as vector.}

\item{calib.date}{\link{character} or \link{Date} (\strong{required}):
date of source calibration in \code{"YYYY-MM-DD"} format.}

\item{calib.dose.rate}{\link{numeric} (\strong{required}):
dose rate at date of calibration in Gy/s or Gy/min.}

\item{calib.error}{\link{numeric} (\strong{required}):
error of dose rate at date of calibration Gy/s or Gy/min.}

\item{source.type}{\link{character} (\emph{with default}):
specify irradiation source (\code{Sr-90}, \code{Co-60}, \code{Cs-137}, \code{Am-214}),
see details for further information.}

\item{dose.rate.unit}{\link{character} (\emph{with default}):
dose rate unit for input (one of \code{Gy/min} or \code{Gy/s}). The output is given in
Gy/s as valid for the function \link{convert_Second2Gray}.}

\item{predict}{\link{integer} (\emph{with default}):
option allowing to predict the dose rate of the source over time in days
set by the provided value. Starting date is the value set with
\code{measurement.date}, e.g., \code{calc_SourceDoseRate(..., predict = 100)} calculates
the source dose rate for the next 100 days.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.
Slot \code{data} contains a \link{list} with the following structure:

\if{html}{\out{<div class="sourceCode">}}\preformatted{$ dose.rate (data.frame)
.. $ dose.rate
.. $ dose.rate.error
.. $ date (corresponding measurement date)
$ parameters (list)
.. $ source.type
.. $ halflife
.. $ dose.rate.unit
$ call (the original function call)
}\if{html}{\out{</div>}}

The output should be accessed using the function \link{get_RLum}.
A plot method of the output is provided via \link{plot_RLum}.
}
\description{
Calculating the dose rate of the irradiation source via the date of
measurement based on: source calibration date, source dose rate, dose rate
error. The function returns a data.frame that provides the input argument
dose_rate for the function \link{convert_Second2Gray}.

Calculation of the source dose rate based on the time elapsed since the last
calibration of the irradiation source. Decay parameters assume a Sr-90 beta
source.
\deqn{dose.rate = D0 \exp(-\frac{\log(2)}{T.1/2} t)}

where \code{D0} is the calibration dose rate, \eqn{T.1/2} is the half-life of the
source nuclide (in days), \code{t} is the time since source calibration (in days),
and \eqn{\log(2) / T.1/2} equals the decay constant lambda.

Information on the date of measurements may be taken from the data's
original .BIN file (using e.g., \code{BINfile <- Luminescence::read_BIN2R()} and the slot
\code{BINfile@METADATA$DATE}).

\strong{Allowed source types and related values}

\tabular{rlll}{
\strong{#} \tab \strong{Source type} \tab \strong{T.1/2} \tab \strong{Reference} \cr
\verb{[1]} \tab Sr-90 \tab 28.90 y \tab NNDC, Brookhaven National Laboratory \cr
\verb{[2]} \tab Am-214 \tab 432.6 y \tab NNDC, Brookhaven National Laboratory \cr
\verb{[3]} \tab Co-60 \tab 5.274 y \tab NNDC, Brookhaven National Laboratory \cr
\verb{[4]} \tab Cs-137 \tab 30.08 y \tab NNDC, Brookhaven National Laboratory}
}
\note{
Please be careful when using the option \code{predict}, especially when a
multiple set for \code{measurement.date} and \code{calib.date} is provided. For the
source dose rate prediction, the function takes the last \code{measurement.date}
value and predicts from that the source dose rate for the number of days
requested, that is: the (multiple) original input will be replaced.
However, the function does not change entries for the calibration dates,
but mixes them up. Therefore, it is not recommended to use this option
when multiple calibration dates (\code{calib.date}) are provided.
}
\section{Function version}{
 0.3.4
}

\examples{

##(1) Simple function usage
##Basic calculation of the dose rate for a specific date
dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
                                  calib.date = "2014-12-19",
                                  calib.dose.rate = 0.0438,
                                  calib.error = 0.0019)

##show results
get_RLum(dose.rate)

##(2) Usage in combination with another function (e.g., convert_Second2Gray() )
## load example data
data(ExampleData.DeValues, envir = environment())

## use the calculated variable dose.rate as input argument
## to convert De(s) to De(Gy)
convert_Second2Gray(ExampleData.DeValues$BT998, dose.rate)

##(3) source rate prediction and plotting
dose.rate <-  calc_SourceDoseRate(measurement.date = "2012-01-27",
                                  calib.date = "2014-12-19",
                                  calib.dose.rate = 0.0438,
                                  calib.error = 0.0019,
                                  predict = 1000)
plot_RLum(dose.rate)

##(4) export output to a LaTeX table (example using the package 'xtable')
\dontrun{
xtable::xtable(get_RLum(dose.rate))
}

} 

\section{How to cite}{
Fuchs, M.C., Kreutzer, S., 2026. calc_SourceDoseRate(): Calculation of the source dose rate via the date of measurement. Function version 0.3.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
NNDC, Brookhaven National Laboratory \verb{http://www.nndc.bnl.gov/}
}
\seealso{
\link{convert_Second2Gray}, \link{get_RLum}, \link{plot_RLum}
}
\author{
Margret C. Fuchs, HZDR, Helmholtz-Institute Freiberg for Resource Technology (Germany) \cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RLum.Results.R
\name{plot_RLum.Results}
\alias{plot_RLum.Results}
\title{Plot function for an RLum.Results S4 class object}
\usage{
plot_RLum.Results(object, single = TRUE, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} (\strong{required}):
S4 object of class \code{RLum.Results}}

\item{single}{\link{logical} (\emph{with default}):
single plot output (\code{TRUE/FALSE}) to allow for plotting the results in as
few plot windows as possible.}

\item{...}{further arguments and graphical parameters will be passed to
the \code{plot} function.}
}
\value{
Returns multiple plots.
}
\description{
The function provides a standardised plot output for data of an RLum.Results
S4 class object
}
\details{
The function produces a multiple plot output.  A file output is recommended
(e.g., \link{pdf}).
}
\note{
Not all arguments available for \link{plot} will be passed!
Only plotting of \code{RLum.Results} objects are supported.
}
\section{Function version}{
 0.2.1
}

\examples{


###load data
data(ExampleData.DeValues, envir = environment())

# apply the un-logged minimum age model
mam <- calc_MinDose(data = ExampleData.DeValues$CA1, sigmab = 0.2, log = TRUE, plot = FALSE)

##plot
plot_RLum.Results(mam)

# estimate the number of grains on an aliquot
grains<- calc_AliquotSize(grain.size = c(100,150), sample.diameter = 1, plot = FALSE, MC.iter = 100)

##plot
plot_RLum.Results(grains)


}
\seealso{
\link{plot}, \link{plot_RLum}
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Burow, C., Kreutzer, S., 2026. plot_RLum.Results(): Plot function for an RLum.Results S4 class object. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{aplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tune_Data.R
\name{tune_Data}
\alias{tune_Data}
\title{Tune data for experimental purpose}
\usage{
tune_Data(data, decrease.error = 0, increase.data = 0)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}):
input values, structure: data (\code{values[,1]}) and data error (\code{values [,2]})
are required}

\item{decrease.error}{\link{numeric}:
factor by which the error is decreased, ranges between 0 and 1.}

\item{increase.data}{\link{numeric}:
factor by which the error is decreased, ranges between 0 and \code{Inf}.}
}
\value{
Returns a \link{data.frame} with tuned values.
}
\description{
The error can be reduced and sample size increased for specific purpose.
}
\note{
You should not use this function to improve your poor data set!
}
\section{Function version}{
 0.5.0
}

\examples{

## load example data set
data(ExampleData.DeValues, envir = environment())
x <- ExampleData.DeValues$CA1

## plot original data
plot_AbanicoPlot(data = x,
                 summary = c("n", "mean"))

## decrease error by 10 \%
plot_AbanicoPlot(data = tune_Data(x, decrease.error = 0.1),
                 summary = c("n", "mean"))

## increase sample size by 200 \%
#plot_AbanicoPlot(data = tune_Data(x, increase.data = 2),
#                summary = c("n", "mean"))

}
\author{
Michael Dietze, GFZ Potsdam (Germany)
, RLum Developer Team} 

\section{How to cite}{
Dietze, M., 2026. tune_Data(): Tune data for experimental purpose. Function version 0.5.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_SingleGrainData.R
\name{subset_SingleGrainData}
\alias{subset_SingleGrainData}
\title{Simple Subsetting of Single Grain Data from Risø BIN/BINX files}
\usage{
subset_SingleGrainData(object, selection)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData} (\strong{required}):
input object with the data to subset.}

\item{selection}{\link{data.frame}, \link{matrix} (\strong{required}):
selection table with two columns for position and grain, respectively.}
}
\value{
A subset \linkS4class{Risoe.BINfileData} object
}
\description{
Most measured single grains do not exhibit light and it usually makes sense
to subset single grain datasets using a table of position and grain pairs.
}
\section{Function version}{
 0.1.1
}

\examples{

## load example data
data(ExampleData.BINfileData, envir = environment())

## set POSITION/GRAIN pair dataset
selection <- data.frame(POSITION = c(1,5,7), GRAIN = c(0,0,0))

##subset
subset_SingleGrainData(object = CWOSL.SAR.Data, selection = selection)

}
\seealso{
\linkS4class{Risoe.BINfileData}, \link{read_BIN2R},
\link{verify_SingleGrainData}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. subset_SingleGrainData(): Simple Subsetting of Single Grain Data from Risø BIN/BINX files. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_Concentration2DoseRate.R
\name{convert_Concentration2DoseRate}
\alias{convert_Concentration2DoseRate}
\title{Dose-rate conversion function}
\usage{
convert_Concentration2DoseRate(object, conversion = "Guerinetal2011", ...)
}
\arguments{
\item{object}{\link{data.frame} (\emph{optional}):
a data frame containing all relevant information for each individual layer.
If nothing is provided, the function returns a template data frame, the
values of which should be filled in by the user. Please note that only one
dataset per input is supported.}

\item{conversion}{\link{character} (\emph{with default}):
dose rate conversion factors to use, by default those by Guérin et al. (2011).
For accepted values see \link{BaseDataSet.ConversionFactors}.}

\item{...}{currently not used.}
}
\value{
The function returns an \linkS4class{RLum.Results} object for which
the first element is a \link{matrix} with the converted values. If no input is
provided, the function returns a template \link{data.frame} that can be used as
input.
}
\description{
This function converts radionuclide concentrations (K in \%, Th and U in ppm)
into dose rates (Gy/ka). Beta-dose rates are also attenuated for the grain
size. Beta and gamma-dose rates are corrected for the water content. This
function converts concentrations into dose rates (Gy/ka) and corrects for
grain size attenuation and water content.

Dose rate conversion factors can be chosen from Adamiec and Aitken (1998),
Guérin et al. (2011), Liritzis et al. (201) and Cresswell et al. (2018).
Default is Guérin et al. (2011).

Grain size correction for beta dose rates is achieved using the correction
factors published by Guérin et al. (2012).

Water content correction is based on factors provided by Aitken (1985),
with the factor for beta dose rate being 1.25 and for gamma 1.14.
}
\details{
\strong{The input data}

\tabular{lll}{
COLUMN \tab DATA TYPE \tab DESCRIPTION\cr
\code{Mineral} \tab \code{character} \tab \code{'FS'} for feldspar, \code{'Q'} for quartz\cr
\code{K} \tab \code{numeric} \tab K nuclide content in \%\cr
\code{K_SE} \tab \code{numeric} \tab error on K nuclide content in \%\cr
\code{Th} \tab \code{numeric} \tab Th nuclide content in ppm\cr
\code{Th_SE} \tab \code{numeric} \tab error on Th nuclide content in ppm\cr
\code{U} \tab \code{numeric} \tab U nuclide content in ppm\cr
\code{U_SE} \tab \code{numeric} \tab error on U nuclide content in ppm\cr
\code{GrainSize} \tab \code{numeric} \tab average grain size in µm\cr
\code{WaterContent} \tab \code{numeric} \tab mean water content in \%\cr
\code{WaterContent_SE} \tab \code{numeric} \tab relative error on water content
}

\strong{Water content}
The water content provided by the user should be calculated according to:

\deqn{(Wet_weight - Dry_weight) / Dry_weight * 100}

The unit for the weight is gram (g).
}
\section{Function version}{
 0.1.1
}

\examples{

## create input template
input <- convert_Concentration2DoseRate()

## fill input
input$Mineral <- "FS"
input$K <- 2.13
input$K_SE <- 0.07
input$Th <- 9.76
input$Th_SE <- 0.32
input$U <- 2.24
input$U_SE <- 0.12
input$GrainSize <- 200
input$WaterContent <- 30
input$WaterContent_SE <- 5

## convert
convert_Concentration2DoseRate(input)

} 

\section{How to cite}{
Riedesel, S., Autzen, M., 2026. convert_Concentration2DoseRate(): Dose-rate conversion function. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update. Ancient TL 16, 37-46.

Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018. Dose rate conversion parameters:
Assessment of nuclear data. Radiation Measurements 120, 195-201.

Guérin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion factors: update.
Ancient TL 29, 5-8.

Guérin, G., Mercier, N., Nathan, R., Adamiec, G., Lefrais, Y., 2012. On the use
of the infinite matrix assumption and associated concepts: A critical review.
Radiation Measurements 47, 778-785.

Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
A re-evaluation of radiation dose-rate conversion factors. Mediterranean
Archaeology and Archaeometry 13, 1-15.
}
\author{
Svenja Riedesel, Aberystwyth University (United Kingdom) \cr
Martin Autzen, DTU NUTECH Center for Nuclear Technologies (Denmark)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RadialPlot.R
\name{plot_RadialPlot}
\alias{plot_RadialPlot}
\title{Function to create a Radial Plot}
\usage{
plot_RadialPlot(
  data,
  na.rm = TRUE,
  log.z = TRUE,
  central.value = NULL,
  centrality = c("mean.weighted", "mean.weighted", "median", "median.weighted"),
  mtext = "",
  summary = c("n", "in.2s"),
  summary.pos = "sub",
  legend = NULL,
  legend.pos = "topright",
  stats = "none",
  rug = FALSE,
  plot.ratio = NULL,
  bar.col = NULL,
  y.ticks = TRUE,
  grid.col = NULL,
  line = NULL,
  line.col = NULL,
  line.label = NULL,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} or \linkS4class{RLum.Results} object (\strong{required}):
for \code{data.frame}: either two columns: De (\code{data[,1]}) and De error
(\code{data[,2]}), or one: De (\code{values[,1]}). If a single-column data frame
is provided, De error is assumed to be 10^-9 for all measurements.
To plot several data sets in one plot, the data sets must be provided as
\code{list}, e.g. \code{list(data.1, data.2)}.}

\item{na.rm}{\link{logical} (\emph{with default}):
excludes \code{NA} values from the data set prior to any further operations.}

\item{log.z}{\link{logical} (\emph{with default}):
Option to display the z-axis in logarithmic scale. Default is \code{TRUE}.}

\item{central.value}{\link{numeric} (\emph{optional}):
User-defined central value, primarily used for horizontal centring
of the z-axis.}

\item{centrality}{\link{character} or \link{numeric} (\emph{with default}):
measure of centrality, used for automatically centring the plot and drawing
the central line. Can either be one out of
\itemize{
\item \code{"mean"},
\item \code{"median"},
\item \code{"mean.weighted"} and
\item \code{"median.weighted"} or a
\item numeric value used for the standardisation.
}}

\item{mtext}{\link{character} (\emph{with default}):
additional text below the plot title.}

\item{summary}{\link{character} (\emph{with default}):
add statistic measures of centrality and dispersion to the plot.
Can be one or more of several keywords. See details for available keywords.}

\item{summary.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"})
for the statistical summary. Alternatively, the keyword \code{"sub"} may be
specified to place the summary below the plot header. However, this latter
option is only possible if \code{mtext} is not used.}

\item{legend}{\link{character} vector (\emph{optional}):
legend content to be added to the plot.}

\item{legend.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional legend position coordinates or keyword (e.g. \code{"topright"}).}

\item{stats}{\link{character} (\emph{optional}):
additional labels of statistically important values in the plot. It can be
one or more of \code{"min"}, \code{"max"} and \code{"median"}; any other values will be
ignored.}

\item{rug}{\link{logical}:
Option to add a rug to the z-scale, to indicate the location of individual values}

\item{plot.ratio}{\link{numeric}:
User-defined plot area ratio (i.e. curvature of the z-axis). If omitted,
the default value (\code{4.5/5.5}) is used and modified automatically to optimise
the z-axis curvature. The parameter should be decreased when data points
are plotted outside the z-axis or when the z-axis gets too elliptic.}

\item{bar.col}{\link{character} or \link{numeric} (\emph{with default}):
colour of the bar showing the 2-sigma range around the central
value. To disable the bar, use \code{"none"}. Default is \code{"grey80"}.}

\item{y.ticks}{\link{logical}:
Option to hide y-axis labels. Useful for data with small scatter.}

\item{grid.col}{\link{character} or \link{numeric} (\emph{with default}):
colour of the grid lines (originating at \verb{[0,0]} and stretching to
the z-scale). To disable grid lines, use \code{"none"}. Default is \code{"grey70"}.}

\item{line}{\link{numeric}:
numeric values of the additional lines to be added.}

\item{line.col}{\link{character} or \link{numeric}:
colour of the additional lines.}

\item{line.label}{\link{character}:
labels for the additional lines.}

\item{...}{Further plot arguments to pass. \code{xlab} must be a vector of
length 2, specifying the upper and lower x-axes labels.}
}
\value{
Returns a plot object.
}
\description{
A Galbraith's radial plot is produced on a logarithmic or a linear scale.
}
\details{
Details and the theoretical background of the radial plot are given in the
cited literature. This function is based on an S script of Rex Galbraith. To
reduce the manual adjustments, the function has been rewritten. Thanks to
Rex Galbraith for useful comments on this function. \cr
Plotting can be disabled by adding the argument \code{plot = "FALSE"}, e.g.
to return only numeric plot output.

Earlier versions of the Radial Plot in this package had the 2-sigma-bar
drawn onto the z-axis. However, this might have caused misunderstanding in
that the 2-sigma range may also refer to the z-scale, which it does not!
Rather it applies only to the x-y-coordinate system (standardised error vs.
precision). A spread in doses or ages must be drawn as lines originating at
zero precision (x0) and zero standardised estimate (y0). Such a range may be
drawn by adding lines to the radial plot (\code{line}, \code{line.col},
\code{line.label}, cf. examples).

A statistic summary, i.e. a collection of statistic measures of
centrality and dispersion (and further measures) can be added by specifying
one or more of the following keywords:
\itemize{
\item \code{"n"} (number of samples),
\item \code{"mean"} (mean De value),
\item \code{"mean.weighted"} (error-weighted mean),
\item \code{"median"} (median of the De values),
\item \code{"median.weighted"} (error-weighted median),
\item \code{"sdrel"} (relative standard deviation in percent),
\item \code{"sdrel.weighted"} (error-weighted relative standard deviation in percent),
\item \code{"sdabs"} (absolute standard deviation),
\item \code{"sdabs.weighted"} (error-weighted absolute standard deviation),
\item \code{"serel"} (relative standard error),
\item \code{"serel.weighted"} (error-weighted relative standard error),
\item \code{"seabs"} (absolute standard error),
\item \code{"seabs.weighted"} (error-weighted absolute standard error),
\item \code{"in.2s"} (percent of samples in 2-sigma range),
\item \code{"kurtosis"} (kurtosis) and
\item \code{"skewness"} (skewness).
}
}
\section{Function version}{
 0.5.11
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())
ExampleData.DeValues <- convert_Second2Gray(
  ExampleData.DeValues$BT998, c(0.0438,0.0019))

## plot the example data straightforward
plot_RadialPlot(data = ExampleData.DeValues)

## now with linear z-scale
plot_RadialPlot(
  data = ExampleData.DeValues,
  log.z = FALSE)

## store the the plot parameters
plot1 <- plot_RadialPlot(
  data = ExampleData.DeValues,
  log.z = FALSE)
plot1
plot1$zlim

## now with adjusted z-scale limits
plot_RadialPlot(
  data = ExampleData.DeValues,
  log.z = FALSE,
  xlim = c(0, 5),
  zlim = c(100, 200))

## now the two plots with serious but seasonally changing fun
#plot_RadialPlot(data = data.3, fun = TRUE)

## now with user-defined central value, in log-scale again
plot_RadialPlot(
  data = ExampleData.DeValues,
  central.value = 150)

## now with a rug, indicating individual De values at the z-scale
plot_RadialPlot(
  data = ExampleData.DeValues,
  rug = TRUE)

## now with legend, colour, different points and smaller scale
plot_RadialPlot(
  data = ExampleData.DeValues,
  legend = "Sample 1",
  col = "tomato4",
  bar.col = "peachpuff",
  pch = "R",
  cex = 0.8)

## now without 2-sigma bar, y-axis, grid lines and central value line
plot_RadialPlot(
  data = ExampleData.DeValues,
  bar.col = "none",
  grid.col = "none",
  y.ticks = FALSE,
  lwd = 0)

## now with user-defined axes labels
plot_RadialPlot(
  data = ExampleData.DeValues,
  xlab = c("Data error (\%)", "Data precision"),
  ylab = "Scatter",
  zlab = "Equivalent dose [Gy]")

## now with minimum, maximum and median value indicated
plot_RadialPlot(
  data = ExampleData.DeValues,
  central.value = 150,
  stats = c("min", "max", "median"))

## now with a brief statistical summary
plot_RadialPlot(
  data = ExampleData.DeValues,
  summary = c("n", "in.2s"))

## now with another statistical summary as subheader
plot_RadialPlot(
  data = ExampleData.DeValues,
  summary = c("mean.weighted", "median"),
  summary.pos = "sub")

## now the data set is split into sub-groups, one is manipulated
data.1 <- ExampleData.DeValues[1:15,]
data.2 <- ExampleData.DeValues[16:25,] * 1.3

## now a common dataset is created from the two subgroups
data.3 <- list(data.1, data.2)

## now the two data sets are plotted in one plot
plot_RadialPlot(data = data.3)

## now with some graphical modification
plot_RadialPlot(
  data = data.3,
  col = c("darkblue", "darkgreen"),
  bar.col = c("lightblue", "lightgreen"),
  pch = c(2, 6),
  summary = c("n", "in.2s"),
  summary.pos = "sub",
  legend = c("Sample 1", "Sample 2"))

} 

\section{How to cite}{
Dietze, M., Kreutzer, S., 2026. plot_RadialPlot(): Function to create a Radial Plot. Function version 0.5.11. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Galbraith, R.F., 1988. Graphical Display of Estimates Having
Differing Standard Errors. Technometrics 30 (3), 271-281.

Galbraith, R.F., 1990. The radial plot: Graphical assessment of spread in
ages. International Journal of Radiation Applications and Instrumentation.
Part D. Nuclear Tracks and Radiation Measurements 17 (3), 207-214.

Galbraith, R. & Green, P., 1990. Estimating the component ages in a finite
mixture. International Journal of Radiation Applications and
Instrumentation. Part D. Nuclear Tracks and Radiation Measurements 17 (3)
197-206.

Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for mixed fission
track ages. Nuclear Tracks And Radiation Measurements 21 (4), 459-470.

Galbraith, R.F., 1994. Some Applications of Radial Plots. Journal of the
American Statistical Association 89 (428), 1232-1242.

Galbraith, R.F., 2010. On plotting OSL equivalent doses. Ancient TL 28 (1),
1-10.

Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent
dose and error calculation and display in OSL dating: An overview and some
recommendations. Quaternary Geochronology 11, 1-27.
}
\seealso{
\link{plot}, \link{plot_KDE}, \link{plot_Histogram}, \link{plot_AbanicoPlot}
}
\author{
Michael Dietze, GFZ Potsdam (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Based on a rewritten S script of Rex Galbraith, 2010
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_Risoe.BINfileData.R
\name{plot_Risoe.BINfileData}
\alias{plot_Risoe.BINfileData}
\title{Plot Single Luminescence Curves from a Risoe.BINfileData-class object}
\usage{
plot_Risoe.BINfileData(
  data,
  position = NULL,
  run = NULL,
  set = NULL,
  sorter = c("POSITION", "RUN", "SET"),
  ltype = c("IRSL", "OSL", "TL", "RIR", "RBR", "RL"),
  curve.transformation = "None",
  dose_rate = NULL,
  temp.lab = "°C",
  cex.global = 1,
  ...
)
}
\arguments{
\item{data}{\linkS4class{Risoe.BINfileData} (\strong{required}):
an S4 object generated by the \link{read_BIN2R} function.}

\item{position}{\link{vector} (\emph{optional}):
option to limit the plotted curves by position
(e.g. \code{position = 1}, \code{position = c(1,3,5)}).}

\item{run}{\link{vector} (\emph{optional}):
option to limit the plotted curves by run
(e.g., \code{run = 1}, \code{run = c(1,3,5)}).}

\item{set}{\link{vector} (\emph{optional}):
option to limit the plotted curves by set
(e.g., \code{set = 1}, \code{set = c(1,3,5)}).}

\item{sorter}{\link{character} (\emph{with default}):
ordering used in plotting the curves, one of "POSITION", "SET" or "RUN".
POSITION, SET and RUN are options defined in the Risoe Sequence Editor.}

\item{ltype}{\link{character} (\emph{with default}):
option to limit the plotted curves by the type of luminescence stimulation.
Allowed values: \code{"IRSL"}, \code{"OSL"},\code{"TL"}, \code{"RIR"}, \code{"RBR"}
(corresponds to LM-OSL), \code{"RL"}.  All type of curves are plotted by
default.}

\item{curve.transformation}{\link{character} (\emph{optional}):
allows transforming CW-OSL and CW-IRSL curves to pseudo-LM curves via
transformation functions. Allowed values are: \code{CW2pLM}, \code{CW2pLMi},
\code{CW2pHMi} and \code{CW2pPMi}, see details. If set to \code{None} (default), no
transformation is applied.}

\item{dose_rate}{\link{numeric} (\emph{optional}):
dose rate of the irradiation source at the measurement date.
If set, the given irradiation dose will be shown in Gy.  See details.}

\item{temp.lab}{\link{character} (\emph{with default}):
the temperature unit to display in the plot labels, by default deg C.}

\item{cex.global}{\link{numeric} (\emph{with default}):
global scaling factor.}

\item{...}{further undocumented plot arguments.}
}
\value{
Returns a plot.
}
\description{
Plots single luminescence curves from a \linkS4class{Risoe.BINfileData} object
produced by \link{read_BIN2R}.
}
\details{
\strong{Nomenclature}

See \linkS4class{Risoe.BINfileData}

\strong{curve.transformation}

This argument allows transforming continuous wave (CW) curves to pseudo
(linear) modulated curves. For the transformation, the functions of the
package are used.  Currently, it is not possible to pass further arguments
to the transformation functions. The argument works only for \code{ltype}
\code{OSL} and \code{IRSL}.

\strong{Irradiation time}

Plotting the irradiation time (s) or the given dose (Gy) requires that the
variable \code{IRR_TIME} has been set within the BIN-file.  This is normally
done by using the 'Run Info' option within the Sequence Editor or by editing
in R.
}
\note{
The function has been successfully tested for the Sequence Editor file
output version 3 and 4.
}
\section{Function version}{
 0.4.3
}

\examples{

##load data
data(ExampleData.BINfileData, envir = environment())

##plot all curves from the first position to the desktop
#pdf(file = "~/Desktop/CurveOutput.pdf", paper = "a4", height = 11, onefile = TRUE)

##example - load from *.bin file
#BINfile<- file.choose()
#BINfileData<-read_BIN2R(BINfile)

#par(mfrow = c(4,3), oma = c(0.5,1,0.5,1))
#plot_Risoe.BINfileData(CWOSL.SAR.Data,position = 1)
#mtext(side = 4, BINfile, outer = TRUE, col = "blue", cex = .7)
#dev.off()

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., Colombo, M., 2026. plot_Risoe.BINfileData(): Plot Single Luminescence Curves from a Risoe.BINfileData-class object. Function version 0.4.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Duller, G., 2007. Analyst. pp. 1-45.
}
\seealso{
\linkS4class{Risoe.BINfileData}, \link{read_BIN2R},
\link{convert_CW2pLM}, \link{convert_CW2pLMi},
\link{convert_CW2pPMi}, \link{convert_CW2pHMi}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}
\keyword{dplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\docType{package}
\name{Luminescence-package}
\alias{Luminescence-package}
\alias{Luminescence}
\title{Comprehensive Luminescence Dating Data Analysis\cr
\if{html}{
\figure{Luminescence_logo.png}{options: width="75" alt="r-luminescence.org"}
}}
\description{
A collection of various R functions for the purpose of luminescence dating
data analysis. This includes, amongst others, data import, export,
application of age models, curve deconvolution, sequence analysis and
plotting of equivalent dose distributions.
}
\details{
\strong{Supervisor of the initial version in 2012}

Markus Fuchs, Justus-Liebig-University Giessen, Germany

\strong{Support contact}
\itemize{
\item \email{developers@r-luminescence.org}
\item \url{https://github.com/R-Lum/Luminescence/discussions}
}

\strong{Bug reporting}
\itemize{
\item \email{developers@r-luminescence.org} or
\item \url{https://github.com/R-Lum/Luminescence/issues}
}

\strong{Project website}
\itemize{
\item \url{https://r-luminescence.org}
}

\strong{Project source code repository}
\itemize{
\item \url{https://github.com/R-Lum/Luminescence}
}

\strong{Related package projects}
\itemize{
\item \url{https://cran.r-project.org/package=RLumShiny}
\item \url{https://cran.r-project.org/package=RLumModel}
\item \url{https://cran.r-project.org/package=RLumCarlo}
\item \url{https://cran.r-project.org/package=RCarb}
}

\strong{Funding}
\itemize{
\item 2011-2013:  The initial version of the package was developed, while Sebastian Kreutzer
was funded through the DFG programme "Rekonstruktion der Umweltbedingungen
des Spätpleistozäns in Mittelsachsen anhand von Löss-Paläobodensequenzen"
(DFG id: 46526743)
\item 2014-2018: Cooperation and personal exchange between the developers is gratefully
funded by the DFG (SCHM 3051/3-1) in the framework of the program
"Scientific Networks". Project title: "RLum.Network: Ein
Wissenschaftsnetzwerk zur Analyse von Lumineszenzdaten mit R" (2014-2018)
\item 05/2014-12/2019: The work of Sebastian Kreutzer as maintainer of the package was supported
by LabEx LaScArBx (ANR - n. ANR-10-LABX-52).
\item 01/2020-04/2022: Sebastian Kreutzer as maintainer of the package has received funding
from the European Union’s Horizon 2020 research and innovation programme under
the Marie Skłodowska-Curie grant agreement No 844457 (CREDit), and could continue
maintaining the package.
\item since 03/2023: Sebastian Kreutzer as maintainer of the package receives funding from the
DFG Heisenberg programme No 505822867.
\item since 08/2024: The future and sustainable development of 'Luminescence'
towards better reproducibility and usability is supported through the DFG
programme "REPLAY: REProducible Luminescence Data AnalYses" No 528704761,
led by Dr Sebastian Kreutzer (PI at LIAG - Institute for Applied Geophysics, Hannover, DE) and Dr Thomas
Kolb (PI at Justus-Liebig-University Giessen, DE).
\item All other authors gratefully received additional funding from various public funding bodies.
}
}
\references{
Dietze, M., Kreutzer, S., Fuchs, M.C., Burow, C., Fischer, M.,
Schmidt, C., 2013. A practical guide to the R package Luminescence.
Ancient TL 31 (1), 11-18.

Dietze, M., Kreutzer, S., Burow, C., Fuchs, M.C., Fischer, M., Schmidt, C., 2016. The abanico plot:
visualising chronometric data with individual standard errors.
Quaternary Geochronology 31, 1-7. \doi{10.1016/j.quageo.2015.09.003}

Fuchs, M.C., Kreutzer, S., Burow, C., Dietze, M., Fischer, M., Schmidt, C.,
Fuchs, M., 2015. Data processing in luminescence dating analysis: An
exemplary workflow using the R package 'Luminescence'. Quaternary
International 362, 8-13. \doi{10.1016/j.quaint.2014.06.034}

Kreutzer, S., Schmidt, C., Fuchs, M.C., Dietze, M., Fischer, M., Fuchs, M.,
2012. Introducing an R package for luminescence dating analysis. Ancient TL
30 (1), 1-8.

Mercier, N., Kreutzer, S., Christophe, C., Guérin, G., Guibert, P., Lahaye, C., Lanos, P., Philippe, A.,
Tribolo, C., 2016. Bayesian statistics in luminescence dating: The 'baSAR'-model and its
implementation in the R package 'Luminescence'. Ancient TL 34 (2), 14-21.

Mercier, N., Galharret, J.-M., Tribolo, C., Kreutzer, S., Philippe, A., 2022.
Luminescence age calculation through Bayesian convolution of equivalent dose
and dose-rate distributions: the De_Dr model.
Geochronology 4, 297–310. \doi{10.5194/gchron-4-297-2022}

Smedley, R.K., 2015. A new R function for the Internal External Uncertainty (IEU) model.
Ancient TL 33 (1), 16-21.

King, E.G., Burow, C., Roberts, H., Pearce, N.J.G., 2018. Age determination
using feldspar: evaluating fading-correction model performance.
Radiation Measurements 119, 58-73. \doi{10.1016/j.radmeas.2018.07.013}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://r-lum.github.io/Luminescence/}
  \item Report bugs at \url{https://github.com/R-Lum/Luminescence/issues}
}

}
\author{
\strong{Maintainer}: Sebastian Kreutzer \email{maintainer_luminescence@r-luminescence.org} (\href{https://orcid.org/0000-0002-0734-2199}{ORCID}) [translator, data contributor]

Authors:
\itemize{
  \item Christoph Burow (\href{https://orcid.org/0000-0002-5023-4046}{ORCID}) [translator, data contributor]
  \item Michael Dietze (\href{https://orcid.org/0000-0001-6063-1726}{ORCID})
  \item Margret C. Fuchs (\href{https://orcid.org/0000-0001-7210-1132}{ORCID})
  \item Christoph Schmidt (\href{https://orcid.org/0000-0002-2309-3209}{ORCID})
  \item Manfred Fischer [translator]
  \item Johannes Friedrich (\href{https://orcid.org/0000-0002-0805-9547}{ORCID})
  \item Norbert Mercier (\href{https://orcid.org/0000-0002-6375-9108}{ORCID})
  \item Anne Philippe (\href{https://orcid.org/0000-0002-5331-5087}{ORCID})
  \item Svenja Riedesel (\href{https://orcid.org/0000-0003-2936-8776}{ORCID})
  \item Martin Autzen (\href{https://orcid.org/0000-0001-6249-426X}{ORCID})
  \item Dirk Mittelstrass (\href{https://orcid.org/0000-0002-9567-8791}{ORCID})
  \item Harrison J. Gray (\href{https://orcid.org/0000-0002-4555-7473}{ORCID})
  \item Jean-Michel Galharret (\href{https://orcid.org/0000-0003-2219-8727}{ORCID})
  \item Marco Colombo (\href{https://orcid.org/0000-0001-6672-0623}{ORCID})
  \item Luc Steinbuch (\href{https://orcid.org/0000-0001-6484-0920}{ORCID})
  \item Anna-Maartje de Boer (\href{https://orcid.org/0000-0002-7359-6939}{ORCID})
  \item Andrzej Bluszcz (\href{https://orcid.org/0000-0002-8027-6941}{ORCID})
}

Other contributors:
\itemize{
  \item Rachel K. Smedley (\href{https://orcid.org/0000-0001-7773-5193}{ORCID}) [contributor]
  \item Claire Christophe [contributor]
  \item Antoine Zink (\href{https://orcid.org/0000-0001-7146-1101}{ORCID}) [contributor]
  \item Julie Durcan (\href{https://orcid.org/0000-0001-8724-8022}{ORCID}) [contributor]
  \item Georgina E. King (\href{https://orcid.org/0000-0003-1059-8192}{ORCID}) [contributor, data contributor]
  \item Guillaume Guérin (\href{https://orcid.org/0000-0001-6298-5579}{ORCID}) [contributor]
  \item Pierre Guibert (\href{https://orcid.org/0000-0001-8969-8684}{ORCID}) [contributor]
  \item Markus Fuchs (\href{https://orcid.org/0000-0003-4669-6528}{ORCID}) [thesis advisor]
}

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Analysis-class.R
\docType{class}
\name{RLum.Analysis-class}
\alias{RLum.Analysis-class}
\title{Class \code{"RLum.Analysis"}}
\description{
Object class to represent analysis data for protocol analysis, i.e. all curves,
spectra etc. from one measurements. Objects from this class are produced,
by e.g. \link{read_XSYG2R}, \link{read_Daybreak2R}
}
\section{Slots}{

\describe{
\item{\code{protocol}}{Object of class \link{character} describing the applied measurement protocol}

\item{\code{records}}{Object of class \link{list} containing objects of class \linkS4class{RLum.Data}}
}}

\note{
The method \link{structure_RLum} is currently just available for objects
containing \linkS4class{RLum.Data.Curve}.
}
\section{Objects from the Class}{

Objects can be created by calls of the form \code{set_RLum("RLum.Analysis", ...)}.
}

\section{Class version}{
 0.4.19
}

\examples{

## show method
showClass("RLum.Analysis")

##set an empty object
set_RLum(class = "RLum.Analysis")

## use example data
##load data
data(ExampleData.RLum.Analysis, envir = environment())

##show curves in object
get_RLum(IRSAR.RF.Data)

##show only the first object, but by keeping the object
get_RLum(IRSAR.RF.Data, record.id = 1, drop = FALSE)

## subsetting with SAR sample data
data(ExampleData.BINfileData, envir = environment())
sar <- object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data)

## get
get_RLum(sar, subset = "NPOINTS == 250")

## remove
remove_RLum(sar, subset = "NPOINTS == 250")

}
\seealso{
\link{Risoe.BINfileData2RLum.Analysis},
\linkS4class{Risoe.BINfileData}, \linkS4class{RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Data.Image-class.R
\docType{class}
\name{RLum.Data.Image-class}
\alias{RLum.Data.Image-class}
\title{Class \code{"RLum.Data.Image"}}
\description{
Class for representing luminescence image data (TL/OSL/RF). Such data are
for example produced by function \link{read_SPE2R}.
}
\section{Slots}{

\describe{
\item{\code{recordType}}{Object of class \link{character} containing the type of the curve (e.g.
"OSL image", "TL image").}

\item{\code{curveType}}{Object of class \link{character} containing curve type, allowed values
are measured or predefined}

\item{\code{data}}{Object of class \link{array} containing image data.}

\item{\code{info}}{Object of class \link{list} containing further meta information objects}
}}

\note{
The class should only contain data for a set of images. For additional
elements the slot \code{info} can be used.
}
\section{Objects from the class}{

Objects can be created by calls of the form \code{set_RLum("RLum.Data.Image", ...)}.
}

\section{Class version}{
 0.5.1
}

\examples{

showClass("RLum.Data.Image")

##create empty RLum.Data.Image object
set_RLum(class = "RLum.Data.Image")

}
\seealso{
\linkS4class{RLum}, \linkS4class{RLum.Data}, \link{plot_RLum},
\link{read_SPE2R}, \link{read_TIFF2R}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_CW2pPMi.R
\name{convert_CW2pPMi}
\alias{convert_CW2pPMi}
\title{Transform a CW-OSL curve into a pPM-OSL curve via interpolation under
parabolic modulation conditions}
\usage{
convert_CW2pPMi(object, P = NULL, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
\linkS4class{RLum.Data.Curve} object or a \code{data.frame} with measured
curve data of type stimulation time (t) (\code{object[, 1]}) and measured counts
(cts) (\code{object[, 2]}).}

\item{P}{\link{numeric} (\emph{optional}):
stimulation time in seconds. If set to \code{NULL}, the optimal value is
estimated automatically (see details). Greater values of P produce more
points in the rising tail of the curve.}

\item{...}{currently not used.}
}
\value{
The function returns the same data type as the input data type with
the transformed curve values.

\code{RLum.Data.Curve}

\tabular{rl}{
\verb{$CW2pPMi.x.t} \tab: transformed time values \cr
\verb{$CW2pPMi.method} \tab: used method for the production of the new data points
}

\code{data.frame}

\tabular{rl}{
\verb{$x} \tab: time\cr
\verb{$y.t} \tab: transformed count values\cr
\verb{$x.t} \tab: transformed time values \cr
\verb{$method} \tab: used method for the production of the new data points
}
}
\description{
Transforms a conventionally measured continuous-wave (CW) OSL-curve into a
pseudo parabolic modulated (pPM) curve under parabolic modulation conditions
using the interpolation procedure described by Bos & Wallinga (2012).
}
\details{
The complete procedure of the transformation is given in Bos & Wallinga
(2012). The input \code{data.frame} consists of two columns: time (t) and
count values (CW(t))

\strong{Nomenclature}
\itemize{
\item P = stimulation time (s)
\item 1/P = stimulation rate (1/s)
}

\strong{Internal transformation steps}

(1)
log(CW-OSL) values

(2)
Calculate t' which is the transformed time:
\deqn{t' = (1/3)*(1/P^2)t^3}

(3)
Interpolate CW(t'), i.e. use the log(CW(t)) to obtain the count values for
the transformed time (t'). Values beyond \code{min(t)} and \code{max(t)}
produce \code{NA} values.

(4)
Select all values for t' < \code{min(t)}, i.e. values beyond the time resolution
of t. Select the first two values of the transformed data set which contain
no \code{NA} values and use these values for a linear fit using \link{lm}.

(5)
Extrapolate values for t' < \code{min(t)} based on the previously obtained
fit parameters. The extrapolation is limited to two values. Other values at
the beginning of the transformed curve are set to 0.

(6)
Transform values using
\deqn{pLM(t) = t^2/P^2*CW(t')}

(7)
Combine all values and truncate all values for t' > \code{max(t)}

\strong{Note:}
The number of values for t' < \code{min(t)} depends on the stimulation
period \code{P}. To avoid the production of too many artificial data at the
raising tail of the determined pPM curve, it is recommended to use the
automatic estimation routine for \code{P}, i.e. provide no value for
\code{P}.
}
\note{
According to Bos & Wallinga (2012), the number of extrapolated points
should be limited to avoid artificial intensity data. If \code{P} is
provided manually, not more than two points are extrapolated.
}
\section{Function version}{
 0.2.5
}

\examples{

##(1)
##load CW-OSL curve data
data(ExampleData.CW_OSL_Curve, envir = environment())

##transform values
values.transformed <- convert_CW2pPMi(ExampleData.CW_OSL_Curve)

##plot
plot(values.transformed$x,values.transformed$y.t, log = "x")

##(2) - produce Fig. 4 from Bos & Wallinga (2012)

##load data
data(ExampleData.CW_OSL_Curve, envir = environment())
values <- CW_Curve.BosWallinga2012

##open plot area
plot(NA, NA,
     xlim = c(0.001,10),
     ylim = c(0,8000),
     ylab = "pseudo OSL (cts/0.01 s)",
     xlab = "t [s]",
     log = "x",
     main = "Fig. 4 - Bos & Wallinga (2012)")

values.t <- convert_CW2pLMi(values, P = 1/20)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col = "red",lwd = 1.3)
text(0.03,4500,"LM", col = "red", cex = .8)

values.t <- convert_CW2pHMi(values, delta = 40)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col = "black", lwd = 1.3)
text(0.005,3000,"HM", cex = .8)

values.t <- convert_CW2pPMi(values, P = 1/10)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col = "blue", lwd = 1.3)
text(0.5,6500,"PM", col = "blue", cex = .8)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. convert_CW2pPMi(): Transform a CW-OSL curve into a pPM-OSL curve via interpolation under parabolic modulation conditions. Function version 0.2.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL
signal components. Radiation Measurements 47, 752-758.

\strong{Further Reading}

Bulur, E., 1996. An Alternative Technique For
Optically Stimulated Luminescence (OSL) Experiment. Radiation Measurements
26, 701-709.

Bulur, E., 2000. A simple transformation for converting CW-OSL curves to
LM-OSL curves. Radiation Measurements 32, 141-145.
}
\seealso{
\link{convert_CW2pLM}, \link{convert_CW2pLMi}, \link{convert_CW2pHMi},
\link{fit_LMCurve}, \linkS4class{RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
Based on comments and suggestions from:
Adrie J.J. Bos, Delft University of Technology, The Netherlands
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_CW2pLMi.R
\name{convert_CW2pLMi}
\alias{convert_CW2pLMi}
\title{Transform a CW-OSL curve into a pLM-OSL curve via interpolation under linear
modulation conditions}
\usage{
convert_CW2pLMi(object, P = NULL, ...)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
\linkS4class{RLum.Data.Curve} object or a \code{data.frame} with measured
curve data of type stimulation time (t) (\code{object[, 1]}) and measured counts
(cts) (\code{object[, 2]}).}

\item{P}{\link{numeric} (\emph{optional}):
stimulation time in seconds. If set to \code{NULL}, the optimal value is
estimated automatically (see details). Greater values of P produce more
points in the rising tail of the curve.}

\item{...}{currently not used.}
}
\value{
The function returns the same data type as the input data type with
the transformed curve values.

\strong{\code{RLum.Data.Curve}}

\tabular{rl}{
\verb{$CW2pLMi.x.t} \tab: transformed time values \cr
\verb{$CW2pLMi.method} \tab: used method for the production of the new data points
}
}
\description{
Transforms a conventionally measured continuous-wave (CW) OSL-curve into a
pseudo linearly modulated (pLM) curve under linear modulation conditions
using the interpolation procedure described by Bos & Wallinga (2012).
}
\details{
The complete procedure of the transformation is given in Bos & Wallinga
(2012). The input \code{data.frame} consists of two columns: time (t) and
count values (CW(t))

\strong{Nomenclature}
\itemize{
\item P = stimulation time (s)
\item 1/P = stimulation rate (1/s)
}

\strong{Internal transformation steps}

(1)
log(CW-OSL) values

(2)
Calculate t' which is the transformed time:
\deqn{t' = 1/2*1/P*t^2}

(3)
Interpolate CW(t'), i.e. use the log(CW(t)) to obtain the count values
for the transformed time (t'). Values beyond \code{min(t)} and \code{max(t)}
produce \code{NA} values.

(4)
Select all values for t' < \code{min(t)}, i.e. values beyond the time resolution
of t. Select the first two values of the transformed data set which contain
no \code{NA} values and use these values for a linear fit using \link{lm}.

(5)
Extrapolate values for t' < \code{min(t)} based on the previously obtained
fit parameters.

(6)
Transform values using
\deqn{pLM(t) = t/P*CW(t')}

(7)
Combine values and truncate all values for t' > \code{max(t)}

\strong{Note:}
The number of values for t' < \code{min(t)} depends on the stimulation
period (P) and therefore on the stimulation rate 1/P. To avoid the
production of too many artificial data at the raising tail of the determined
pLM curves it is recommended to use the automatic estimation routine for
\code{P}, i.e. provide no own value for \code{P}.
}
\note{
According to Bos & Wallinga (2012) the number of extrapolated points
should be limited to avoid artificial intensity data. If \code{P} is
provided manually and more than two points are extrapolated, a warning
message is returned.
}
\section{Function version}{
 0.3.5
}

\examples{

##(1)
##load CW-OSL curve data
data(ExampleData.CW_OSL_Curve, envir = environment())

##transform values
values.transformed <- convert_CW2pLMi(ExampleData.CW_OSL_Curve)

##plot
plot(values.transformed$x, values.transformed$y.t, log = "x")

##(2) - produce Fig. 4 from Bos & Wallinga (2012)
##load data
data(ExampleData.CW_OSL_Curve, envir = environment())
values <- CW_Curve.BosWallinga2012

##open plot area
plot(NA, NA,
     xlim = c(0.001,10),
     ylim = c(0,8000),
     ylab = "pseudo OSL (cts/0.01 s)",
     xlab = "t [s]",
     log = "x",
     main = "Fig. 4 - Bos & Wallinga (2012)")


values.t <- convert_CW2pLMi(values, P = 1/20)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col = "red", lwd = 1.3)
text(0.03,4500,"LM", col = "red", cex = .8)

values.t <- convert_CW2pHMi(values, delta = 40)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col = "black", lwd = 1.3)
text(0.005,3000,"HM", cex =.8)

values.t <- convert_CW2pPMi(values, P = 1/10)
lines(values[1:length(values.t[, 1]), 1], values.t[, 2],
      col = "blue", lwd = 1.3)
text(0.5,6500,"PM", col = "blue", cex = .8)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. convert_CW2pLMi(): Transform a CW-OSL curve into a pLM-OSL curve via interpolation under linear modulation conditions. Function version 0.3.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL
signal components. Radiation Measurements 47, 752-758.

\strong{Further Reading}

Bulur, E., 1996. An Alternative Technique For
Optically Stimulated Luminescence (OSL) Experiment. Radiation Measurements
26, 701-709.

Bulur, E., 2000. A simple transformation for converting CW-OSL curves to
LM-OSL curves. Radiation Measurements 32, 141-145.
}
\seealso{
\link{convert_CW2pLM}, \link{convert_CW2pHMi}, \link{convert_CW2pPMi},
\link{fit_LMCurve}, \linkS4class{RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
Based on comments and suggestions from:
Adrie J.J. Bos, Delft University of Technology, The Netherlands
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Analysis-class.R,
%   R/RLum.Data.Curve-class.R, R/RLum.Data.Image-class.R,
%   R/RLum.Data.Spectrum-class.R, R/RLum.Results-class.R
\name{as}
\alias{as}
\title{as() - RLum-object coercion}
\arguments{
\item{from}{\linkS4class{RLum}, \link{list}, \link{data.frame}, \link{matrix} (\strong{required}):
object to be coerced from}

\item{to}{\link{character} (\strong{required}):
class name to be coerced to}
}
\description{
for \verb{[RLum.Analysis-class]}

for \verb{[RLum.Data.Curve-class]}

for \verb{[RLum.Data.Image-class]}

for \verb{[RLum.Data.Spectrum-class]}

for \verb{[RLum.Results-class]}
}
\details{
\strong{\linkS4class{RLum.Analysis}}

\tabular{ll}{
\strong{from} \tab \strong{to}\cr
\code{list} \tab \code{list}\cr
}

Given that the \link{list} consists of \linkS4class{RLum.Analysis} objects.

\strong{\linkS4class{RLum.Data.Curve}}

\tabular{ll}{
\strong{from} \tab \strong{to}\cr
\code{list} \tab \code{list} \cr
\code{data.frame} \tab \code{data.frame}\cr
\code{matrix} \tab \code{matrix}
}

\strong{\linkS4class{RLum.Data.Image}}

\tabular{ll}{
\strong{from} \tab \strong{to}\cr
\code{data.frame} \tab \code{data.frame}\cr
\code{matrix} \tab \code{matrix}
}

\strong{\linkS4class{RLum.Data.Spectrum}}

\tabular{ll}{
\strong{from} \tab \strong{to}\cr
\code{data.frame} \tab \code{data.frame}\cr
\code{matrix} \tab \code{matrix} \cr
\code{list} \tab \code{list}
}

\strong{\linkS4class{RLum.Results}}

\tabular{ll}{
\strong{from} \tab \strong{to}\cr
\code{list} \tab \code{list}\cr
}

Given that the \link{list} consists of \linkS4class{RLum.Results} objects.
}
\note{
Due to the complex structure of the \code{RLum} objects itself a coercing to standard
R data structures will be always loosely!
}
\seealso{
\link[methods:as]{methods::as}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_RLum.Results.R
\name{merge_RLum.Results}
\alias{merge_RLum.Results}
\title{Merge function for RLum.Results S4-class objects}
\usage{
merge_RLum.Results(objects)
}
\arguments{
\item{objects}{\link{list} (\strong{required}):
a list of \linkS4class{RLum.Results} objects}
}
\description{
Function merges objects of class \linkS4class{RLum.Results}. The slots in the objects
are combined depending on the object type, e.g., for \link{data.frame} and \link{matrix}
rows are appended.
}
\details{
Elements are appended where possible and attributes are preserved if
not of similar name as the default attributes of, e.g., a \link{data.frame}
}
\note{
The \code{originator} is taken from the first element and not reset to \link{merge_RLum}
}
\section{Function version}{
 0.2.1
}

\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. merge_RLum.Results(): Merge function for RLum.Results S4-class objects. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.XSYG}
\alias{ExampleData.XSYG}
\alias{OSL.SARMeasurement}
\alias{TL.Spectrum}
\title{Example data for a SAR OSL measurement and a TL spectrum using a lexsyg
reader}
\format{
\code{OSL.SARMeasurement}: SAR OSL measurement data

The data contain two elements: (a) \verb{$Sequence.Header} is a
\link{data.frame} with metadata from the measurement,(b)
\code{Sequence.Object} contains an \linkS4class{RLum.Analysis} object
for further analysis.

\code{TL.Spectrum}: TL spectrum data

\linkS4class{RLum.Data.Spectrum} object for further analysis. The
spectrum was cleaned from cosmic-rays using the function

\code{apply_CosmicRayRemoval}. Note that no quantum efficiency calibration
was performed.
}
\source{
\strong{OSL.SARMeasurement}

\tabular{ll}{
Lab: \tab Luminescence Laboratory Giessen\cr
Lab-Code: \tab no code\cr
Location: \tab not specified\cr
Material: \tab Coarse grain quartz  on steel cups on lexsyg research reader\cr
Reference: \tab unpublished
}

\strong{TL.Spectrum}

\tabular{ll}{
Lab: \tab Luminescence Laboratory Giessen\cr
Lab-Code: \tab BT753\cr
Location: \tab Dolni Vestonice/Czech Republic\cr
Material: \tab Fine grain polymineral on steel cups on lexsyg research reader\cr
Reference: \tab Fuchs et al., 2013 \cr
Spectrum: \tab Integration time 19 s, channel time 20 s\cr
Heating: \tab 1 K/s, up to 500 deg. C
}
}
\description{
Example data from a SAR OSL measurement and a TL spectrum for package
Luminescence imported from a Freiberg Instruments XSYG file using the
function \link{read_XSYG2R}.
}
\section{Version}{
 0.1
}

\examples{
##show data
data(ExampleData.XSYG, envir = environment())

## =========================================
##(1) OSL.SARMeasurement
OSL.SARMeasurement

##show $Sequence.Object
OSL.SARMeasurement$Sequence.Object

##grep OSL curves and plot the first curve
OSLcurve <- get_RLum(OSL.SARMeasurement$Sequence.Object,
recordType="OSL")[[1]]
plot_RLum(OSLcurve)

## =========================================
##(2) TL.Spectrum
TL.Spectrum

##plot simple spectrum (2D)
plot_RLum.Data.Spectrum(TL.Spectrum,
                        plot.type="contour",
                        xlim = c(310,750),
                        ylim = c(0,300),
                        bin.rows=10,
                        bin.cols = 1)

##plot 3d spectrum (uncomment for usage)
# plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="persp",
# xlim = c(310,750), ylim = c(0,300), bin.rows=10,
# bin.cols = 1)

}
\references{
Unpublished data measured to serve as example data for that
package. Location origin of sample BT753 is given here:

Fuchs, M., Kreutzer, S., Rousseau, D.D., Antoine, P., Hatte, C., Lagroix,
F., Moine, O., Gauthier, C., Svoboda, J., Lisa, L., 2013. The loess sequence
of Dolni Vestonice, Czech Republic: A new OSL-based chronology of the Last
Climatic Cycle. Boreas, 42, 664--677.
}
\seealso{
\link{read_XSYG2R}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Data.Spectrum},
\link{plot_RLum}, \link{plot_RLum.Analysis}, \link{plot_RLum.Data.Spectrum}
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\docType{data}
\name{ExampleData.CW_OSL_Curve}
\alias{ExampleData.CW_OSL_Curve}
\alias{CW_Curve.BosWallinga2012}
\title{Example CW-OSL curve data for the package Luminescence}
\format{
Data frame with 1000 observations on the following 2 variables:
\itemize{
\item \code{list("x")}: a numeric vector, time
\item \code{list("y")}: a numeric vector, counts
}
}
\source{
\strong{ExampleData.CW_OSL_Curve}

\tabular{ll}{
Lab: \tab Luminescence Laboratory Bayreuth\cr
Lab-Code: \tab BT607\cr
Location: \tab Saxony/Germany\cr
Material: \tab Middle grain quartz measured on aluminium cups on a Risø TL/OSL DA-15 reader.\cr
Reference: \tab unpublished data }

\strong{CW_Curve.BosWallinga2012}

\tabular{ll}{
Lab: \tab Netherlands Centre for Luminescence Dating (NCL)\cr
Lab-Code: \tab NCL-2108077\cr
Location: \tab Guadalentin Basin, Spain\cr
Material: \tab Coarse grain quartz\cr
Reference: \tab Bos & Wallinga (2012) and Baartman et al. (2011)
}
}
\description{
\code{data.frame} containing CW-OSL curve data (time, counts)
}
\examples{

data(ExampleData.CW_OSL_Curve, envir = environment())
plot(ExampleData.CW_OSL_Curve)

}
\references{
Baartman, J.E.M., Veldkamp, A., Schoorl, J.M., Wallinga, J.,
Cammeraat, L.H., 2011. Unravelling Late Pleistocene and Holocene landscape
dynamics: The Upper Guadalentin Basin, SE Spain. Geomorphology 125,
172-185.

Bos, A.J.J. & Wallinga, J., 2012. How to visualize quartz OSL signal
components. Radiation Measurements 47, 752-758.
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R
\name{remove_RLum}
\alias{remove_RLum}
\alias{remove_RLum,list-method}
\alias{remove_RLum,RLum.Analysis-method}
\title{Strips records from RLum-class objects}
\usage{
remove_RLum(object, ...)

\S4method{remove_RLum}{list}(object, ...)

\S4method{remove_RLum}{RLum.Analysis}(object, ...)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
object with records to be removed.}

\item{...}{parameters to be passed to \link{get_RLum}. The arguments \code{get.index} and
\code{drop} are preset and have no effect when provided}
}
\value{
An object of the same type as the input provided with records removed; it
can result in an empty object.
}
\description{
Remove records from an RLum-class object in a convenient way using
\link{get_RLum} for the selection.
}
\section{Functions}{
\itemize{
\item \code{remove_RLum(list)}: Returns a list of \linkS4class{RLum} objects where the selected records are stripped

\item \code{remove_RLum(RLum.Analysis)}: Method to remove records from an \linkS4class{RLum.Analysis} object.

}}
\section{Function version}{
 0.1.0
}

\examples{
## load example data
data(ExampleData.XSYG, envir = environment())
sar <- OSL.SARMeasurement$Sequence.Object[1:9]

## strop only OSL curves
sar <- remove_RLum(sar, recordType = "OSL")
sar

}
\seealso{
\linkS4class{RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. remove_RLum(): Strips records from RLum-class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RLum.R
\name{plot_RLum}
\alias{plot_RLum}
\title{General plot function for RLum S4 class objects}
\usage{
plot_RLum(object, ...)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
object of class \linkS4class{RLum} or a list of such objects. If a
list is provided, the function tries to plot every object in the list
according to its \code{RLum} class, after removing non-RLum objects.}

\item{...}{further arguments and graphical parameters to pass to the
specific plot functions. The only arguments that are supported directly are
\code{main} (plot title) and \code{mtext} (plot subtitle). Here \code{main} can be provided
as a list and the arguments in the
list will dispatched to the plots if \code{object} is of type \code{list} as well.}
}
\value{
Produces a plot depending on the input object.
}
\description{
The function calls object specific plot functions for RLum S4 class objects.
}
\details{
The function provides a generalised access point for plotting various
\linkS4class{RLum} objects. Depending on the input object, the
corresponding plot function will be selected. Allowed arguments can be
found in the documentations of each plot function.

\tabular{lll}{
\strong{object} \tab \strong{corresponding plot function} \cr
\linkS4class{RLum.Data.Curve} \tab \link{plot_RLum.Data.Curve} \cr
\linkS4class{RLum.Data.Spectrum} \tab \link{plot_RLum.Data.Spectrum}\cr
\linkS4class{RLum.Data.Image} \tab \link{plot_RLum.Data.Image}\cr
\linkS4class{RLum.Analysis} \tab \link{plot_RLum.Analysis}\cr
\linkS4class{RLum.Results} \tab \link{plot_RLum.Results}
}
}
\section{Function version}{
 0.5
}

\examples{
#load Example data
data(ExampleData.CW_OSL_Curve, envir = environment())

#transform data.frame to RLum.Data.Curve object
temp <- as(ExampleData.CW_OSL_Curve, "RLum.Data.Curve")

#plot RLum object
plot_RLum(temp)

}
\seealso{
\link{plot_RLum.Analysis},
\link{plot_RLum.Data.Curve},
\link{plot_RLum.Data.Spectrum},
\link{plot_RLum.Data.Image},
\link{plot_RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)\cr
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. plot_RLum(): General plot function for RLum S4 class objects. Function version 0.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{dplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_MaxDose.R
\name{calc_MaxDose}
\alias{calc_MaxDose}
\title{Apply the maximum age model to a given De distribution}
\usage{
calc_MaxDose(
  data,
  sigmab,
  log = TRUE,
  par = 3,
  bootstrap = FALSE,
  init.values = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with \code{(data[, 1])} and De error \code{(data[, 2])}.}

\item{sigmab}{\link{numeric} (\strong{required}):
additional spread in De values, representing the expected overdispersion in
the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
\strong{Note:} For the logged model (\code{log = TRUE}) this value must be
a fraction, e.g. 0.2 (= 20 \%). If the un-logged model is used (\code{log = FALSE}),
\code{sigmab} must be provided in the same absolute units of the De values (seconds or Gray).
See details.}

\item{log}{\link{logical} (\emph{with default}):
whether the logged minimum dose model should be fit to De data (\code{TRUE} by
default).}

\item{par}{\link{numeric} (\emph{with default}):
number of parameters in the minimum age model, either 3 (default) or 4.}

\item{bootstrap}{\link{logical} (\emph{with default}):
apply the recycled bootstrap approach of Cunningham & Wallinga (2012).}

\item{init.values}{\link{numeric} (\emph{optional}):
a named list with starting values for \code{gamma}, \code{sigma}, \code{p0} and \code{mu}
(e.g. \code{list(gamma=100, sigma=1.5, p0=0.1, mu=100)}). If no values are
provided, reasonable values will be estimated from the data.
\strong{Note:} the initial values must always be given in the absolute units.
If a logged model is applied (\code{log = TRUE}), the provided \code{init.values}
are automatically log-transformed.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments for bootstrapping (\verb{bs.M, bs.N, bs.h, sigmab.sd}).
See details for their usage.}
}
\value{
Please see \link{calc_MinDose}.
}
\description{
Function to fit the maximum age model to De data. This is a wrapper function
that calls \link{calc_MinDose} and applies a similar approach as described in
Olley et al. (2006).
}
\details{
\strong{Data transformation}

To estimate the maximum dose population
and its standard error, the three parameter minimum age model of Galbraith
et al. (1999) is adapted. The measured De values are transformed as follows:
\enumerate{
\item convert De values to natural logs
\item multiply the logged data to create a mirror image of the De distribution
\item shift De values along x-axis by the smallest x-value found to obtain only positive values
\item combine in quadrature the measurement error associated with each De value
with a relative error specified by \code{sigmab}
\item apply the MAM to these data
}

When all calculations are done the results are then converted as follows
\enumerate{
\item subtract the x-offset
\item multiply the natural logs by -1
\item take the exponent to obtain the maximum dose estimate in Gy
}

\strong{Further documentation}

Please see \link{calc_MinDose}.
}
\section{Function version}{
 0.3.2
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())

# apply the maximum dose model
calc_MaxDose(ExampleData.DeValues$CA1, sigmab = 0.2, par = 3)

} 

\section{How to cite}{
Burow, C., 2026. calc_MaxDose(): Apply the maximum age model to a given De distribution. Function version 0.3.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Arnold, L.J., Roberts, R.G., Galbraith, R.F. & DeLong, S.B.,
2009. A revised burial dose estimation procedure for optical dating of young
and modern-age sediments. Quaternary Geochronology 4, 306-325.

Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for mixed fission
track ages. Nuclear Tracks Radiation Measurements 4, 459-470.

Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley, J.M.,
1999. Optical dating of single grains of quartz from Jinmium rock shelter,
northern Australia. Part I: experimental design and statistical models.
Archaeometry 41, 339-364.

Galbraith, R.F., 2005. Statistics for
Fission Track Analysis, Chapman & Hall/CRC, Boca Raton.

Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error
calculation and display in OSL dating: An overview and some recommendations.
Quaternary Geochronology 11, 1-27.

Olley, J.M., Roberts, R.G., Yoshida, H., Bowler, J.M., 2006. Single-grain optical dating of grave-infill
associated with human burials at Lake Mungo, Australia. Quaternary Science
Reviews 25, 2469-2474

\strong{Further reading}

Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
(De) distributions: Implications for OSL dating of sediment mixtures.
Quaternary Geochronology 4, 204-230.

Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
assessment of procedures for estimating burial dose. Quaternary Science
Reviews 25, 2475-2502.

Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
Quaternary Geochronology 12, 98-106.

Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
of optical dating of fluvial deposits.  Quaternary Geochronology 1, 109-120.

Rodnight, H., 2008. How many equivalent dose values are needed to
obtain a reproducible distribution?. Ancient TL 26, 3-10.
}
\seealso{
\link{calc_CentralDose}, \link{calc_CommonDose}, \link{calc_FiniteMixture},
\link{calc_FuchsLang2001}, \link{calc_MinDose}
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Based on a rewritten S script of Rex Galbraith, 2010
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_Histogram.R
\name{plot_Histogram}
\alias{plot_Histogram}
\title{Plot a histogram with separate error plot}
\usage{
plot_Histogram(
  data,
  na.rm = TRUE,
  mtext = "",
  cex.global = 1,
  se = TRUE,
  rug = TRUE,
  normal_curve = FALSE,
  summary = "",
  summary.pos = "sub",
  colour = c("white", "black", "red", "black"),
  interactive = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} or \linkS4class{RLum.Results} object (\strong{required}):
for \code{data.frame}: two columns: De (\code{data[,1]}) and De error (\code{data[,2]}).
If the error column is missing or only contains \code{NA} values, then the error
at each measurement is assumed to be 10^-9.}

\item{na.rm}{\link{logical} (\emph{with default}):
excludes \code{NA} values from the data set prior to any further operations.}

\item{mtext}{\link{character} (\emph{optional}):
further sample information (\link{mtext}).}

\item{cex.global}{\link{numeric} (\emph{with default}):
global scaling factor.}

\item{se}{\link{logical} (\emph{with default}):
plots standard error points over the histogram, default is \code{TRUE}.}

\item{rug}{\link{logical} (\emph{with default}):
adds rugs to the histogram, default is \code{TRUE}.}

\item{normal_curve}{\link{logical} (\emph{with default}):
adds a normal curve to the histogram. Mean and standard deviation are
calculated from the input data. If \code{TRUE}, the y-axis in the histogram
will show the probability density.}

\item{summary}{\link{character} (\emph{with default}):
add statistic measures of centrality and dispersion to the plot.
Can be one or more of several keywords. See details for available keywords.}

\item{summary.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"})
for the statistical summary. Alternatively, the keyword \code{"sub"} may be
specified to place the summary below the plot header. However, this latter
option in only possible if \code{mtext} is not used. In case of coordinate
specification, y-coordinate refers to the right y-axis.}

\item{colour}{\link{numeric} or \link{character} (\emph{with default}):
optional vector of length 4 which specifies the colours of the following
plot items in exactly this order: histogram bars, rug lines and summary
text, normal distribution curve, standard error points
(e.g., \code{c("grey", "black", "red", "grey")}).}

\item{interactive}{\link{logical} (\emph{with default}):
create an interactive histogram plot (requires the 'plotly' package)}

\item{...}{further arguments and graphical parameters passed to \link{plot} or
\link{hist}. If y-axis labels are provided, these must be specified as a vector
of length 2 since the plot features two axes
(e.g. \code{ylab = c("axis label 1", "axis label 2")}). Y-axes limits
(\code{ylim}) must be provided as vector of length four, with the first two
elements specifying the left axes limits and the latter two elements giving
the right axis limits.}
}
\description{
Function plots a predefined histogram with an accompanying error plot as
suggested by Rex Galbraith at the UK LED in Oxford 2010.

If the normal curve is added, the y-axis in the histogram will show the
probability density.

A statistic summary, i.e. a collection of statistic measures of
centrality and dispersion (and further measures) can be added by specifying
one or more of the following keywords:
\itemize{
\item \code{"n"} (number of samples),
\item \code{"mean"} (mean De value),
\item \code{"mean.weighted"} (error-weighted mean),
\item \code{"median"} (median of the De values),
\item \code{"median.weighted"} (error-weighted median),
\item \code{"sdrel"} (relative standard deviation in percent),
\item \code{"sdrel.weighted"} (error-weighted relative standard deviation in percent),
\item \code{"sdabs"} (absolute standard deviation),
\item \code{"sdabs.weighted"} (error-weighted absolute standard deviation),
\item \code{"serel"} (relative standard error),
\item \code{"serel.weighted"} (error-weighted relative standard error),
\item \code{"seabs"} (absolute standard error),
\item \code{"seabs.weighted"} (error-weighted absolute standard error),
\item \code{"kurtosis"} (kurtosis) and
\item \code{"skewness"} (skewness).
}
}
\note{
The input data is not restricted to a special type.
}
\section{Function version}{
 0.4.5
}

\examples{

## load data
data(ExampleData.DeValues, envir = environment())
ExampleData.DeValues <- convert_Second2Gray(ExampleData.DeValues$BT998,
                                            dose.rate = c(0.0438,0.0019))

## plot histogram the easiest way
plot_Histogram(ExampleData.DeValues)

## plot histogram with some more modifications
plot_Histogram(ExampleData.DeValues,
               rug = TRUE,
               normal_curve = TRUE,
               cex.global = 0.9,
               pch = 2,
               colour = c("grey", "black", "blue", "green"),
               summary = c("n", "mean", "sdrel"),
               summary.pos = "topleft",
               main = "Histogram of De-values",
               mtext = "Example data set",
               ylab = c(expression(paste(D[e], " distribution")),
                        "Standard error"),
               xlim = c(100, 250),
               ylim = c(0, 0.1, 5, 20))

}
\seealso{
\link{hist}, \link{plot}
}
\author{
Michael Dietze, GFZ Potsdam (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Dietze, M., Kreutzer, S., 2026. plot_Histogram(): Plot a histogram with separate error plot. Function version 0.4.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_SurfaceExposure.R
\name{fit_SurfaceExposure}
\alias{fit_SurfaceExposure}
\title{Nonlinear Least Squares Fit for OSL surface exposure data}
\usage{
fit_SurfaceExposure(
  data,
  sigmaphi = NULL,
  mu = NULL,
  age = NULL,
  Ddot = NULL,
  D0 = NULL,
  weights = FALSE,
  plot = TRUE,
  legend = TRUE,
  error_bars = TRUE,
  coord_flip = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} or \link{list} (\strong{required}):
Measured OSL surface exposure data with the following structure:

\if{html}{\out{<div class="sourceCode">}}\preformatted{                               (optional)
     | depth (a.u.)| intensity | error |
     |     [ ,1]   |    [ ,2]  | [ ,3] |
     |-------------|-----------|-------|
[1, ]|    ~~~~     |    ~~~~   | ~~~~  |
[2, ]|    ~~~~     |    ~~~~   | ~~~~  |
 ... |     ...     |     ...   |  ...  |
[x, ]|    ~~~~     |    ~~~~   | ~~~~  |

}\if{html}{\out{</div>}}

Alternatively, a \link{list} of \code{data.frames} can be provided, where each
\code{data.frame} has the same structure as shown above, with the exception that
they must \strong{not} include the optional error column. Providing a \link{list} as
input automatically activates the global fitting procedure (see details).}

\item{sigmaphi}{\link{numeric} (\emph{optional}):
A numeric value for \code{sigmaphi}, i.e. the charge detrapping rate.
Example: \code{sigmaphi = 5e-10}}

\item{mu}{\link{numeric} (\emph{optional}):
A numeric value for \code{mu}, i.e. the light attenuation coefficient. If \code{data}
is a list of \emph{n} samples, then \code{mu} must be a vector of length 1 or \emph{n}.
Example: \code{mu = 0.9}, or \code{mu = c(0.9, 0.8, 0.9)}.}

\item{age}{\link{numeric} (\emph{optional}):
The age (a) of the sample, if known. If \code{data} is a \link{list} of \emph{n} samples,
then \code{age} must be a numeric vector of length \emph{n}.
Example: \code{age = 10000}, or \code{age = c(1e4, 1e5, 1e6)}.}

\item{Ddot}{\link{numeric} (\emph{optional}):
A numeric value for the environmental dose rate (Gy/ka). For this argument
to be considered a value for \code{D0} must also be provided; otherwise it will be
ignored.}

\item{D0}{\link{numeric} (\emph{optional}):
A numeric value for the characteristic saturation dose (Gy). For this argument
to be considered a value for \code{Ddot} must also be provided; otherwise it will be
ignored.}

\item{weights}{\link{logical} (\emph{with default}):
If \code{TRUE} the fit will be weighted by the inverse square of the error.
Requires \code{data} to be a \link{data.frame} with three columns.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{legend}{\link{logical} (\emph{with default}):
Show or hide the equation inside the plot.}

\item{error_bars}{\link{logical} (\emph{with default}):
Show or hide error bars (only applies if errors were provided).}

\item{coord_flip}{\link{logical} (\emph{with default}):
Flip the coordinate system.}

\item{...}{Further parameters passed to \link{plot}.
Custom parameters include:
\itemize{
\item \code{verbose} (\link{logical}): enable/disable output to the terminal
\item \code{line_col}: Colour of the fitted line
\item \code{line_lty}: Type of the fitted line (see \code{lty} in \code{?par})
\item \code{line_lwd}: Line width of the fitted line (see \code{lwd} in \code{?par})
}}
}
\value{
Function returns results numerically and graphically:

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$summary} \tab \code{data.frame} \tab summary of the fitting results \cr
\verb{$data} \tab \code{data.frame} \tab the original input data \cr
\verb{$fit} \tab \code{nls} \tab the fitting object produced by \link[minpack.lm:nlsLM]{minpack.lm::nlsLM} \cr
\verb{$args} \tab \code{character} \tab arguments of the call \cr
\verb{$call} \tab \code{call} \tab the original function call \cr
}

\strong{slot:} \strong{\verb{@info}}

Currently unused.

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr

A scatter plot of the provided depth-intensity OSL surface exposure data
with the fitted model.
}
\description{
This function determines the (weighted) least-squares estimates of the
parameters of either equation 1 in \emph{Sohbati et al. (2012a)} or equation 12 in
\emph{Sohbati et al. (2012b)} for a given OSL surface exposure data set (\strong{BETA}).
}
\details{
\strong{Weighted fitting}

If \code{weights = TRUE} the function will use the inverse square of the error (\eqn{1/\sigma^2})
as weights during fitting using \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}. Naturally, for this to
take effect individual errors must be provided in the third column of the
\code{data.frame} for \code{data}. Weighted fitting is \strong{not} supported if \code{data}
is a list of multiple \code{data.frame}s, i.e., it is not available for global
fitting.

\strong{Dose rate}
If any of the arguments \code{Ddot} or \code{D0} is at its default value (\code{NULL}),
this function will fit equation 1 in Sohbati et al. (2012a) to the data. If
the effect of dose rate (i.e., signal saturation) needs to be considered,
numeric values for the dose rate (\code{Ddot}) (in Gy/ka) and the characteristic
saturation dose (\code{D0}) (in Gy) must be provided. The function will then fit
equation 12 in Sohbati et al. (2012b) to the data.

\strong{Note:} Currently, this function does \strong{not} consider the variability
of the dose rate with sample depth (\code{x})! In the original equation the dose
rate \code{D} is an arbitrary function of \code{x} (term \code{D(x)}), but here \code{D} is assumed
constant.

\strong{Global fitting}
If \code{data} is \link{list} of multiple \code{data.frame}s, each representing a separate
sample, the function automatically performs a global fit to the data. This
may be useful to better constrain the parameters \code{sigmaphi} or \code{mu} and
\strong{requires} that a known age for each sample
(e.g., \code{age = c(100, 1000)} if \code{data} is a list with two samples).
}
\note{
\strong{This function has BETA status. If possible, results should be}
\strong{cross-checked.}
}
\section{Function version}{
 0.1.0
}

\examples{

## Load example data
data("ExampleData.SurfaceExposure")

## Example 1 - Single sample
# Known parameters: 10000 a, mu = 0.9, sigmaphi = 5e-10
sample_1 <- ExampleData.SurfaceExposure$sample_1
head(sample_1)
results <- fit_SurfaceExposure(
 data = sample_1,
 mu = 0.9,
 sigmaphi = 5e-10)
get_RLum(results)

## Example 2 - Single sample and considering dose rate
# Known parameters: 10000 a, mu = 0.9, sigmaphi = 5e-10,
# dose rate = 2.5 Gy/ka, D0 = 40 Gy
sample_2 <- ExampleData.SurfaceExposure$sample_2
head(sample_2)
results <- fit_SurfaceExposure(
 data = sample_2,
 mu = 0.9,
 sigmaphi = 5e-10,
 Ddot = 2.5,
 D0 = 40)
get_RLum(results)

## Example 3 - Multiple samples (global fit) to better constrain 'mu'
# Known parameters: ages = 1e3, 1e4, 1e5, 1e6 a, mu = 0.9, sigmaphi = 5e-10
set_1 <- ExampleData.SurfaceExposure$set_1
str(set_1, max.level = 2)
results <- fit_SurfaceExposure(
  data = set_1,
  age = c(1e3, 1e4, 1e5, 1e6),
  sigmaphi = 5e-10)
get_RLum(results)

## Example 4 - Multiple samples (global fit) and considering dose rate
# Known parameters: ages = 1e2, 1e3, 1e4, 1e5, 1e6 a, mu = 0.9, sigmaphi = 5e-10,
# dose rate = 1.0 Ga/ka, D0 = 40 Gy
set_2 <- ExampleData.SurfaceExposure$set_2
str(set_2, max.level = 2)
results <- fit_SurfaceExposure(
 data = set_2,
 age = c(1e2, 1e3, 1e4, 1e5, 1e6),
 sigmaphi = 5e-10,
 Ddot = 1,
 D0 = 40)
get_RLum(results)

} 

\section{How to cite}{
Burow, C., 2026. fit_SurfaceExposure(): Nonlinear Least Squares Fit for OSL surface exposure data. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Sohbati, R., Murray, A.S., Chapot, M.S., Jain, M., Pederson, J., 2012a.
Optically stimulated luminescence (OSL) as a chronometer for surface exposure
dating. Journal of Geophysical Research 117, B09202. doi:
\doi{10.1029/2012JB009383}

Sohbati, R., Jain, M., Murray, A.S., 2012b. Surface exposure dating of
non-terrestrial bodies using optically stimulated luminescence: A new method.
Icarus 221, 160-166.
}
\seealso{
\link{ExampleData.SurfaceExposure}, \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}
}
\author{
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github.R
\name{GitHub-API}
\alias{GitHub-API}
\alias{github_commits}
\alias{github_branches}
\alias{github_issues}
\title{GitHub API - Deprecated}
\usage{
github_commits(user = "r-lum", repo = "luminescence", branch = "master", n = 5)

github_branches(user = "r-lum", repo = "luminescence")

github_issues(user = "r-lum", repo = "luminescence", verbose = TRUE)
}
\arguments{
\item{user}{\link{character} (\emph{with default}):
GitHub user name (defaults to \code{'r-lum'}).}

\item{repo}{\link{character} (\emph{with default}):
name of a GitHub repository (defaults to \code{'luminescence'}).}

\item{branch}{\link{character} (\emph{with default}):
branch of a GitHub repository (defaults to \code{'master'}).}

\item{n}{\link{integer} (\emph{with default}):
number of commits returned (defaults to 5).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}
}
\value{
\code{github_commits}: \link{data.frame} with columns:

\tabular{ll}{
\verb{[ ,1]} \tab SHA \cr
\verb{[ ,2]} \tab AUTHOR \cr
\verb{[ ,3]} \tab DATE \cr
\verb{[ ,4]} \tab MESSAGE \cr
}

\code{github_branches}: \link{data.frame} with columns:

\tabular{ll}{
\verb{[ ,1]} \tab BRANCH \cr
\verb{[ ,2]} \tab SHA \cr
\verb{[ ,3]} \tab INSTALL \cr
}

\code{github_commits}: Nested \link{list} with \code{n} elements.
Each commit element is a list with elements:

\tabular{ll}{
\verb{[[1]]} \tab NUMBER \cr
\verb{[[2]]} \tab TITLE \cr
\verb{[[3]]} \tab BODY \cr
\verb{[[4]]} \tab CREATED \cr
\verb{[[5]]} \tab UPDATED \cr
\verb{[[6]]} \tab CREATOR \cr
\verb{[[7]]} \tab URL \cr
\verb{[[8]]} \tab STATUS \cr
}
}
\description{
R Interface to the GitHub API v3.
}
\details{
These functions can be used to query a specific repository hosted on GitHub. \cr

\code{github_commits} lists the most recent \code{n} commits of a specific branch of a repository.

\code{github_branches} can be used to list all current branches of a
repository and returns the corresponding SHA hash as well as an installation
command to install the branch in R via the 'devtools' package.

\code{github_issues} lists all open issues for a repository in valid YAML.
}
\section{Function version}{
 0.1.0
}

\examples{

\dontrun{
github_branches(user = "r-lum", repo = "luminescence")
github_issues(user = "r-lum", repo = "luminescence")
github_commits(user = "r-lum", repo = "luminescence", branch = "master", n = 10)
}

} 

\section{How to cite}{
Burow, C., 2025. GitHub-API(): GitHub API - Deprecated. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., 2025. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.1.2. https://r-lum.github.io/Luminescence/
}

\references{
GitHub Developer API v3. \url{https://docs.github.com/v3/}, last accessed: 10/01/2017.
}
\author{
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_DoseResponseCurve.R
\name{fit_DoseResponseCurve}
\alias{fit_DoseResponseCurve}
\title{Fit a dose-response curve for luminescence data (Lx/Tx against dose)}
\usage{
fit_DoseResponseCurve(
  object,
  mode = c("interpolation", "extrapolation", "alternate"),
  fit.method = c("EXP", "LIN", "QDR", "EXP OR LIN", "EXP+LIN", "EXP+EXP", "GOK", "OTOR",
    "OTORX"),
  fit.force_through_origin = FALSE,
  fit.weights = TRUE,
  fit.includingRepeatedRegPoints = TRUE,
  fit.NumberRegPoints = NULL,
  fit.NumberRegPointsReal = NULL,
  fit.bounds = TRUE,
  n.MC = 100,
  txtProgressBar = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{\link{data.frame} or a \link{list} of such objects (\strong{required}):
data frame with columns for \code{Dose}, \code{LxTx}, \code{LxTx.Error} and \code{TnTx}.
The column for the test dose response is optional, but requires \code{'TnTx'} as
column name if used. For exponential fits at least three dose points
(including the natural) should be provided. If \code{object} is a list,
the function is called on each of its elements.
If \code{fit.method = "OTORX"} you have  to provide the test dose in the same unit
as the dose in a column called \code{Test_Dose}. The function searches explicitly
for this column name. Only the first value will be used assuming a constant
test dose over the measurement cycle.}

\item{mode}{\link{character} (\emph{with default}):
selects calculation mode of the function.
\itemize{
\item \code{"interpolation"} (default) calculates the De by interpolation,
\item \code{"extrapolation"} calculates the equivalent dose by extrapolation
(useful for MAAD measurements) and
\item \code{"alternate"} calculates no equivalent dose and just fits the data points.
}

Please note that for option \code{"interpolation"} the first point is considered
as natural dose}

\item{fit.method}{\link{character} (\emph{with default}):
function used for fitting. Possible options are: \code{LIN}, \code{QDR}, \code{EXP},
\verb{EXP OR LIN}, \code{EXP+LIN}, \code{EXP+EXP} (not defined for extrapolation), \code{GOK},
\code{OTOR} and \code{OTORX}. See details.}

\item{fit.force_through_origin}{\link{logical} (\emph{with default})
allow to force the fitted function through the origin.
For \code{method = "EXP+EXP"} the function will be fixed through
the origin in either case, so this option will have no effect.}

\item{fit.weights}{\link{logical} (\emph{with default}):
option whether the fitting is done with or without weights. See details.}

\item{fit.includingRepeatedRegPoints}{\link{logical} (\emph{with default}):
includes repeated points for fitting (\code{TRUE}/\code{FALSE}).}

\item{fit.NumberRegPoints}{\link{integer} (\emph{optional}):
set number of regeneration points manually. By default the number of all (!)
regeneration points is used automatically.}

\item{fit.NumberRegPointsReal}{\link{integer} (\emph{optional}):
if the number of regeneration points is provided manually, the value of the
real, regeneration points = all points (repeated points) including reg 0,
has to be inserted.}

\item{fit.bounds}{\link{logical} (\emph{with default}):
set lower fit bounds for all fitting parameters to 0. Limited for the use
with the fit methods \code{EXP}, \code{EXP+LIN}, \verb{EXP OR LIN}, \code{GOK}, \code{OTOR}, \code{OTORX}
Argument to be inserted for experimental application only!}

\item{n.MC}{\link{integer} (\emph{with default}):
number of Monte Carlo simulations for error estimation, see details.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar. If \code{verbose = FALSE} also no
\code{txtProgressBar} is shown.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{Further arguments to be passed (currently ignored).}
}
\value{
An \linkS4class{RLum.Results} object is returned
containing the slot \code{data} with the
following elements:

\strong{Overview elements}
\tabular{lll}{
\strong{DATA.OBJECT} \tab \strong{TYPE} \tab \strong{DESCRIPTION} \cr
\code{..$De} : \tab  \code{data.frame} \tab Table with De values \cr
\code{..$De.MC} : \tab \code{numeric} \tab Table with De values from MC runs \cr
\code{..$Fit} : \tab \link{nls} or \link{lm} \tab object from the fitting for \code{EXP}, \code{EXP+LIN} and \code{EXP+EXP}.
In case of a resulting  linear fit when using \code{LIN}, \code{QDR} or \verb{EXP OR LIN} \cr
\code{..Fit.Args} : \tab \code{list} \tab Arguments to the function \cr
\code{..$Formula} : \tab \link{expression} \tab Fitting formula as R expression \cr
}

The \verb{@info} slot contains the following elements:
\tabular{lll}{
\strong{DATA.OBJECT} \tab \strong{TYPE} \tab \strong{DESCRIPTION} \cr
\code{..$fit_messag}: \tab \code{character} \tab The fit message reported \cr
\code{..$call} : \tab \code{call} \tab The original function call \cr
}

If \code{object} is a list, then the function returns a list of
\linkS4class{RLum.Results}
objects as defined above.

\strong{Details - \code{DATA.OBJECT$De}}
This object is a \link{data.frame} with the following columns
\tabular{lll}{
\code{De} \tab \link{numeric} \tab equivalent dose \cr
\code{De.Error} \tab \link{numeric} \tab standard error the equivalent dose \cr
\code{D01} \tab \link{numeric} \tab \eqn{D_0} value, curvature parameter of the exponential \cr
\code{D01.ERROR} \tab \link{numeric} \tab standard error of the \eqn{D_0} value\cr
\code{D02} \tab \link{numeric} \tab 2nd \eqn{D_0} value, only for \code{EXP+EXP}\cr
\code{D02.ERROR} \tab \link{numeric} \tab standard error for 2nd \eqn{D_0}; only for \code{EXP+EXP}\cr
\code{R} \tab \link{numeric} \tab the material specific parameter \eqn{R}\cr
\code{R.ERROR} \tab \link{numeric} \tab the uncertainty of R \eqn{R}\cr
\code{Dc} \tab \link{numeric} \tab value indicating saturation level; only for \code{OTOR} \cr
\code{D63} \tab \link{numeric} \tab the specific saturation level; only for \code{OTORX} \cr
\code{n_N} \tab \link{numeric} \tab saturation level of dose-response curve derived via integration from the used function; it compares the full integral of the curves (\code{N}) to the integral until \code{De} (\code{n}) (e.g.,  Guralnik et al., 2015)\cr
\code{De.MC} \tab \link{numeric} \tab equivalent dose derived by Monte-Carlo simulation; ideally identical to \code{De}\cr
\code{Fit} \tab \link{character} \tab applied fit function \cr
\code{Mode} \tab \link{character} \tab mode used in fitting \cr
\code{HPDI68_L} \tab \link{numeric} \tab highest probability density of approximated equivalent dose probability curve representing the lower boundary of 68\% probability \cr
\code{HPDI68_U} \tab \link{numeric} \tab same as \code{HPDI68_L} for the upper bound \cr
\code{HPDI95_L} \tab \link{numeric} \tab same as \code{HPDI68_L} but for 95\% probability \cr
\code{HPDI95_U} \tab \link{numeric} \tab same as \code{HPDI95_L} but for the upper bound \cr
\code{.De.plot} \tab \link{numeric} \tab equivalent dose used internally for plotting \cr
\code{.De.raw} \tab \link{numeric} \tab equivalent dose reported 'as is', that is containing infinities and negative values if they could be calculated. Bear in mind that negative values are meaningless and may be arbitrary.\cr
}
}
\description{
A dose-response curve is produced for luminescence measurements using a
regenerative or additive protocol. The function supports interpolation and
extrapolation to calculate the equivalent dose.
}
\details{
\subsection{Implemented fitting methods}{

For all options (except for the \code{LIN}, \code{QDR} and the \verb{EXP OR LIN}),
the \link[minpack.lm:nlsLM]{minpack.lm::nlsLM} function with the \code{LM} (Levenberg-Marquardt algorithm)
algorithm is used. Note: For historical reasons for the Monte Carlo
simulations partly the function \link{nls} using the \code{port} algorithm.

The solution is found by transforming the function or using \link[stats:uniroot]{stats::uniroot}.

\strong{Keyword: \code{LIN}}

Fits a linear function to the data using \link{lm}:
\deqn{y = mx + n}

\strong{Keyword: \code{QDR}}

Fits a linear function with a quadratic term to the data using  \link{lm}:
\deqn{y = a + bx + cx^2}

\strong{Keyword: \code{EXP}}

Adapts a function of the form
\deqn{y = a(1 - \exp(-\frac{(x+c)}{b}))}

Parameters b and c are approximated by a linear fit using \link{lm}. Note: \eqn{b = D0}

\strong{Keyword: \verb{EXP OR LIN}}

Works for some cases where an \code{EXP} fit fails. If the \code{EXP} fit fails,
a \code{LIN} fit is done instead, which always works.

\strong{Keyword: \code{EXP+LIN}}

Tries to fit an exponential plus linear function of the form:

\deqn{y = a(1 - \exp(-\frac{x + c}{b}) + (gx))}
The \eqn{D_e} is calculated by iteration.

\strong{Note:} In the context of luminescence dating, this function has no physical meaning.
Therefore, no \eqn{D_0} value is returned.

\strong{Keyword: \code{EXP+EXP}}

Tries to fit a double exponential function of the form

\deqn{y = (a_1 (1 - \exp(-\frac{x}{b_1}))) + (a_2 (1 - \exp(-\frac{x}{b_2})))}

\emph{This fitting procedure is not really robust against wrong start parameters.}

\strong{Keyword: \code{GOK}}

Tries to fit the general-order kinetics function following Guralnik et al. (2015)
of the form

\deqn{y = a (d - (1 + (\frac{1}{b}) x c)^{(-1/c)})}

where \eqn{c > 0} is a kinetic order modifier
(not to be confused with \strong{c} in \code{EXP} or \code{EXP+LIN}!).

\strong{Keyword: \code{OTOR}} (former \code{LambertW})

This tries to fit a dose-response curve based on the Lambert W function
and the one trap one recombination centre (OTOR) model according to Pagonis
et al. (2020). The function has the form:

\deqn{y = (1 + (\mathcal{W}((R - 1) * \exp(R - 1 - ((x + D_{int}) / D_{c}))) / (1 - R))) * N}

with \eqn{W} the Lambert W function (calculated using \link[lamW:lamW]{lamW::lambertW0}),
\eqn{R} the dimensionless retrapping ratio, \eqn{N} the total concentration
of trappings states in cm\eqn{^{-3}}, \eqn{D_{c} = N/R} a constant, and
\eqn{D_{int}} is the offset on the x-axis. Note that \eqn{R} and \eqn{D_{c}}
have a valid physical interpretation only when saturation is reached.
Please note that finding the root in \code{mode = "extrapolation"}
is a non-easy task due to the shape of the function and the results might be
unexpected.

\strong{Keyword: \code{OTORX}}

This adapts extended OTOR (therefore: OTORX) model proposed by Lawless and
Timar-Gabor (2024) accounting for retrapping (the equation implemented here
is written slightly differently than in the original manuscript):

\deqn{F_{OTORX} = 1 + \left[\mathcal{W}\left(-Q * \exp\left(-Q-(1-Q(1-\frac{1}{\exp(1)})) \frac{(D + a)}{D_{63}}\right)\right)\right] / Q}

with

\deqn{Q = \frac{A_m - A_n}{A_m}\frac{N}{N+N_D}}

where \eqn{A_m} and \eqn{A_n} are rate constants for the recombination and
the trapping of electrons (\eqn{N}), respectively. \eqn{D_{63}} corresponds to
the value at which the trap occupation corresponds to the 63\% of the saturation value.
\eqn{a} is in an offset. If set to zero, the curve will be forced through the origin
as in the original publication.

For the implementation the calculation reads further

\deqn{y = \frac{F_{OTORX}(((D + a)/D_{63}), Q)}{F_{OTORX}((D_{test} + a)/D_{63}, Q)}}

with \eqn{D_{test}} being the test dose in the same unit (usually s or Gy) as
the regeneration dose points. This value is essential and needs to provided
along with the usual dose and \eqn{\frac{L_x}{T_x}} values (see \code{object} parameter input
and the example section). For more details see Lawless and Timar-Gabor (2024).

The fit also returns the parameter \eqn{R} know from \code{OTOR}, which is derived
as \eqn{R = 1 - Q}.

\emph{Note: The offset adder \eqn{a} is not part of the formula in Timar-Gabor (2024) and can
be set to zero with the option \code{fit.force_through_origin = TRUE}}

\strong{Fit weighting}

If the option \code{fit.weights =  TRUE} is chosen, weights are calculated using
provided signal errors (\eqn{\frac{L_x}{T_x}} error):
\deqn{fit.weights = \frac{\frac{1}{error}}{\Sigma{\frac{1}{error}}}}

\strong{Error estimation using Monte Carlo simulation}

Error estimation is done using a parametric bootstrapping approach. A set of
\eqn{\frac{L_x}{T_x}} values is constructed by randomly drawing curve data sampled from normal
distributions. The normal distribution is defined by the input values (\code{mean = value}, \code{sd = value.error}). Then, a dose-response curve fit is attempted for each
dataset resulting in a new distribution of single \code{De} values. The standard
deviation of this distribution becomes then the error of the \code{De}. With increasing
iterations, the error value becomes more stable. However, naturally the error
will not decrease with more MC runs.

Alternatively, the function returns highest probability density interval
estimates as output, users may find more useful under certain circumstances.

\strong{Note:} It may take some calculation time with increasing MC runs,
especially for the composed functions (\code{EXP+LIN} and \code{EXP+EXP}).\cr
Each error estimation is done with the function of the chosen fitting method.
}
}
\section{Function version}{
 1.4.5
}

\examples{

##(1) fit growth curve for a dummy data.set and show De value
data(ExampleData.LxTxData, envir = environment())
temp <- fit_DoseResponseCurve(LxTxData)
get_RLum(temp)

##(1b) to access the fitting value try
get_RLum(temp, data.object = "Fit")

##(2) fit using the 'extrapolation' mode
LxTxData[1,2:3] <- c(0.5, 0.001)
print(fit_DoseResponseCurve(LxTxData, mode = "extrapolation"))

##(3) fit using the 'alternate' mode
LxTxData[1,2:3] <- c(0.5, 0.001)
print(fit_DoseResponseCurve(LxTxData, mode = "alternate"))

##(4) import and fit test data set by Berger & Huntley 1989
QNL84_2_unbleached <-
read.table(system.file("extdata/QNL84_2_unbleached.txt", package = "Luminescence"))

results <- fit_DoseResponseCurve(
 QNL84_2_unbleached,
 mode = "extrapolation",
 verbose = FALSE)

#calculate confidence interval for the parameters
#as alternative error estimation
confint(results$Fit, level = 0.68)

\dontrun{
##(5) special case the OTORX model with test dose column
df <- cbind(LxTxData, Test_Dose = 15)
fit_DoseResponseCurve(object = df, fit.method = "OTORX", n.MC = 10) |>
 plot_DoseResponseCurve()

QNL84_2_bleached <-
read.table(system.file("extdata/QNL84_2_bleached.txt", package = "Luminescence"))
STRB87_1_unbleached <-
read.table(system.file("extdata/STRB87_1_unbleached.txt", package = "Luminescence"))
STRB87_1_bleached <-
read.table(system.file("extdata/STRB87_1_bleached.txt", package = "Luminescence"))

print(
 fit_DoseResponseCurve(
 QNL84_2_bleached,
 mode = "alternate",
 verbose = FALSE)$Fit)

print(
 fit_DoseResponseCurve(
 STRB87_1_unbleached,
 mode = "alternate",
 verbose = FALSE)$Fit)

print(
 fit_DoseResponseCurve(
 STRB87_1_bleached,
 mode = "alternate",
 verbose = FALSE)$Fit)
 }

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., Colombo, M., 2026. fit_DoseResponseCurve(): Fit a dose-response curve for luminescence data (Lx/Tx against dose). Function version 1.4.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46. \doi{10.26034/la.atl.1989.150}

Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
from feldspar. Radiation Measurements 81, 224-231. \doi{10.1016/j.radmeas.2015.02.011}

Lawless, J.L., Timar-Gabor, A., 2024. A new analytical model to fit both fine and coarse grained quartz luminescence dose response curves. Radiation Measurements 170, 107045. \doi{10.1016/j.radmeas.2023.107045}

Pagonis, V., Kitis, G., Chen, R., 2020. A new analytical equation for the dose response of dosimetric materials,
based on the Lambert W function. Journal of Luminescence 225, 117333. \doi{10.1016/j.jlumin.2020.117333}
}
\seealso{
\link{plot_DoseResponseCurve}, \link{nls},
\linkS4class{RLum.Results}, \link{get_RLum},
\link[minpack.lm:nlsLM]{minpack.lm::nlsLM}, \link{lm}, \link{uniroot}, \link[lamW:lamW]{lamW::lambertW0}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\docType{data}
\name{ExampleData.portableOSL}
\alias{ExampleData.portableOSL}
\title{Example portable OSL curve data for the package Luminescence}
\source{
\strong{ExampleData.portableOSL}

\tabular{ll}{
Lab: \tab Cologne Luminescence Laboratory\cr
Lab-Code: \tab \code{none} \cr
Location: \tab Nievenheim/Germany\cr
Material: \tab Fine grain quartz \cr
Reference: \tab unpublished data
}
}
\description{
A \code{list} of \linkS4class{RLum.Analysis} objects, each containing
the same number of \linkS4class{RLum.Data.Curve} objects representing
individual OSL, IRSL and dark count measurements of a sample.
}
\examples{

data(ExampleData.portableOSL, envir = environment())
plot_RLum(ExampleData.portableOSL)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Risoe.BINfileData2RLum.Analysis.R
\name{Risoe.BINfileData2RLum.Analysis}
\alias{Risoe.BINfileData2RLum.Analysis}
\title{Convert Risoe.BINfileData object to an RLum.Analysis object}
\usage{
Risoe.BINfileData2RLum.Analysis(
  object,
  pos = NULL,
  grain = NULL,
  run = NULL,
  set = NULL,
  ltype = NULL,
  dtype = NULL,
  protocol = "unknown",
  keep.empty = TRUE,
  txtProgressBar = FALSE
)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData} (\strong{required}):
object to convert.}

\item{pos}{\link{numeric} (\emph{optional}): position number of the \code{Risoe.BINfileData}
object for which the curves are stored in the \linkS4class{RLum.Analysis} object.
If \code{length(pos) > 1}, a list of \code{RLum.Analysis} objects is returned.
If nothing is provided every position will be converted.
If the position is not valid \code{NULL} is returned.}

\item{grain}{\link{vector}, \link{numeric} (\emph{optional}):
grain number from the measurement to limit the converted data set
(e.g., \code{grain = c(1:48)}). Please be aware that this option may lead to
unwanted effects, as the output is strictly limited to the chosen grain
number for all position numbers.}

\item{run}{\link{vector}, \link{numeric} (\emph{optional}):
run number from the measurement to limit the converted data set
(e.g., \code{run = c(1:48)}).}

\item{set}{\link{vector}, \link{numeric} (\emph{optional}):
set number from the measurement to limit the converted data set
(e.g., \code{set = c(1:48)}).}

\item{ltype}{\link{vector}, \link{character} (\emph{optional}):
curve type to limit the converted data. Commonly allowed values are:
\code{IRSL}, \code{OSL}, \code{TL}, \code{RIR}, \code{RBR} and \code{USER}
(see also \linkS4class{Risoe.BINfileData}).}

\item{dtype}{\link{vector}, \link{character} (\emph{optional}):
data type to limit the converted data. Commonly allowed values are
listed in \linkS4class{Risoe.BINfileData}.}

\item{protocol}{\link{character} (\emph{optional}):
sets protocol type for analysis object. Value may be used by subsequent
analysis functions.}

\item{keep.empty}{\link{logical} (\emph{with default}):
If \code{TRUE} (default) an \code{RLum.Analysis} object is returned even if it does
not contain any records. Set to \code{FALSE} to discard all empty objects.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar.}
}
\value{
Returns an \linkS4class{RLum.Analysis} object.
}
\description{
Converts values from one specific position of a \linkS4class{Risoe.BINfileData}
object to an \linkS4class{RLum.Analysis} object.

The \linkS4class{RLum.Analysis} object requires a set of curves for
specific further protocol analyses. However, the \linkS4class{Risoe.BINfileData}
usually contains a set of curves for different aliquots and different
protocol types that may be mixed up. Therefore, a conversion is needed.
}
\note{
The \code{protocol} argument of the \linkS4class{RLum.Analysis}
object is set to 'unknown' if not stated otherwise.
}
\section{Function version}{
 0.4.3
}

\examples{

##load data
data(ExampleData.BINfileData, envir = environment())

##convert values for position 1
Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos = 1)

}
\seealso{
\linkS4class{Risoe.BINfileData}, \linkS4class{RLum.Analysis},
\link{read_BIN2R}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. Risoe.BINfileData2RLum.Analysis(): Convert Risoe.BINfileData object to an RLum.Analysis object. Function version 0.4.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R
\name{smooth_RLum}
\alias{smooth_RLum}
\alias{smooth_RLum,list-method}
\alias{smooth_RLum,RLum.Analysis-method}
\alias{smooth_RLum,RLum.Data.Curve-method}
\title{Smoothing of data for RLum-class objects}
\usage{
smooth_RLum(object, ...)

\S4method{smooth_RLum}{list}(object, ...)

\S4method{smooth_RLum}{RLum.Analysis}(object, ...)

\S4method{smooth_RLum}{RLum.Data.Curve}(
  object,
  k = NULL,
  fill = NA,
  align = "right",
  method = "mean",
  p_acceptance = 1e-07
)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum}}

\item{...}{further arguments passed to the specific class method}

\item{k}{\link{smooth_RLum}; \link{integer} (\emph{with default}):
window for the rolling mean or median. If \code{NULL}, this set automatically
(ignored if \code{method = "Carter_etal_2018"}).}

\item{fill}{\link{smooth_RLum}; \link{numeric} (\emph{with default}):
value used to pad the result so to have the same length as the input.}

\item{align}{\link{smooth_RLum}; \link{character} (\emph{with default}):
one of \code{"right"}, \code{"center"} or \code{"left"}, specifying whether the index
of the result should be right-aligned (default), centred, or left-aligned
compared to the rolling window of observations (ignored if
\code{method = "Carter_etal_2018"}).}

\item{method}{\link{smooth_RLum}; \link{character} (\emph{with default}):
smoothing method to be applied: one of \code{"mean"}, \code{"median"} or
\code{"Carter_etal_2018"}.}

\item{p_acceptance}{\link{smooth_RLum}; \link{numeric} (\emph{with default}):
probability threshold of accepting a value to be a sample from a Poisson
distribution (only used for \code{method = "Carter_etal_2018"}). Values that
have a Poisson probability below the threshold are replaced by the average
over the four neighbouring values.}
}
\value{
An object of the same type as the input object provided.
}
\description{
The function provides a generalised access point for specific
\linkS4class{RLum} objects. Depending on the input object, the corresponding
function will be selected. The smoothing is performed in the internal
function \code{.smoothing()}.
}
\section{Functions}{
\itemize{
\item \code{smooth_RLum(list)}: Returns a list of \linkS4class{RLum} objects that had been passed to
\link{smooth_RLum}

\item \code{smooth_RLum(RLum.Analysis)}: Smoothing of \code{RLum.Data} records contained in the input object.

\item \code{smooth_RLum(RLum.Data.Curve)}: Smoothing of \linkS4class{RLum.Data.Curve} objects using a rolling mean or median.
For methods \code{"mean"} and \code{"median"}, smoothing is performed by rolling
mean and rolling median with window of size \code{k}. Method \code{"Carter_etal_2018"}
implements a Poisson smoother for dark-background signals measured by a
photomultiplier tube.

}}
\note{
Currently only \code{RLum} objects of class \code{RLum.Data.Curve} and \code{RLum.Analysis}
(with curve data) are supported.
}
\section{Function version}{
 0.1.0
}

\examples{

## load example data
data(ExampleData.CW_OSL_Curve, envir = environment())

## create RLum.Data.Curve object from this example
curve <-
  set_RLum(
      class = "RLum.Data.Curve",
      recordType = "OSL",
      data = as.matrix(ExampleData.CW_OSL_Curve)
  )

## plot data without and with smoothing
plot_RLum(curve)
plot_RLum(smooth_RLum(curve))

} 

\section{How to cite}{
Kreutzer, S., 2026. smooth_RLum(): Smoothing of data for RLum-class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Carter, J., Cresswell, A.J., Kinnaird, T.C., Carmichael, L.A., Murphy, S. &
Sanderson, D.C.W., 2018. Non-Poisson variations in photomultipliers and
implications for luminescence dating. Radiation Measurements 120, 267-273.
\doi{10.1016/j.radmeas.2018.05.010}
}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_ThermalLifetime.R
\name{calc_ThermalLifetime}
\alias{calc_ThermalLifetime}
\title{Calculates the Thermal Lifetime using the Arrhenius equation}
\usage{
calc_ThermalLifetime(
  E,
  s,
  T = 20,
  output_unit = "Ma",
  profiling = FALSE,
  profiling_config = list(),
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{E}{\link{numeric} (\strong{required}):
vector of trap depths in eV,
if \code{profiling = TRUE} only the first two elements are considered}

\item{s}{\link{numeric} (\strong{required}):
vector of frequency factor in 1/s,
if \code{profiling = TRUE} only the first two elements are considered}

\item{T}{\link{numeric} (\emph{with default}):
temperature in deg. C for which the lifetime(s) will be calculated.
A vector can be provided.}

\item{output_unit}{\link{character} (\emph{with default}):
output unit of the calculated lifetimes, accepted
entries are: \code{"Ma"}, \code{"ka"}, \code{"a"}, \code{"d"}, \code{"h"}, \code{"min"}, \code{"s"}}

\item{profiling}{\link{logical} (\emph{with default}):
this option allows to estimate uncertainties based on
given E and s parameters and their corresponding standard error
(cf. details and examples section)}

\item{profiling_config}{\link{list} (\emph{optional}):
allows to set configuration parameters used for the profiling
(and only have an effect here). Supported parameters are:
\itemize{
\item \code{n} (number of MC runs),
\item \code{E.distribution} (distribution used for the re-sampling for E) and
\item \code{s.distribution} (distribution used for the re-sampling for s).
}

Currently only the normal distribution is supported
(e.g., \code{profiling_config = list(E.distribution = "norm")}}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output, currently only in combination with
\code{profiling = TRUE}.}

\item{...}{further arguments that can be passed in combination with the plot output.
Standard plot parameters are supported (\link{plot.default})}
}
\value{
A \linkS4class{RLum.Results} object is returned a along with a plot (for
\code{profiling = TRUE}). The output object contain the following slots:

\strong{\verb{@data}}

\tabular{lll}{
\strong{Object} \tab \strong{Type} \tab \strong{Description} \cr
\code{lifetimes} \tab \link{array} or \link{numeric} \tab calculated lifetimes \cr
\code{profiling_matrix} \tab \link{matrix} \tab profiling matrix used for the MC runs
}

\strong{\verb{@info}}

\tabular{lll}{
\strong{Object} \tab \strong{Type} \tab \strong{Description} \cr
\code{call} \tab \code{call} \tab the original function call
}
}
\description{
The function calculates the thermal lifetime of charges for given E (in eV), s (in 1/s) and
T (in deg. C.) parameters. The function can be used in two operational modes:
}
\details{
\strong{Mode 1 \code{(profiling = FALSE)}}

An arbitrary set of input parameters (E, s, T) can be provided and the
function calculates the thermal lifetimes using the Arrhenius equation for
all possible combinations of these input parameters. An array with 3-dimensions
is returned that can be used for further analyses or graphical output (see example 1)

\strong{Mode 2 \code{(profiling = TRUE)}}

This mode tries to profile the variation of the thermal lifetime for a chosen
temperature by accounting for the provided E and s parameters and their corresponding
standard errors, e.g., \code{E = c(1.600, 0.001)}
The calculation based on a Monte Carlo simulation, where values are sampled from a normal
distribution (for E and s).

\strong{Used equation (Arrhenius equation)}

\deqn{\tau = \frac{1}{s} \exp(E/kT)}
where:
\eqn{\tau} in s as the mean time an electron spends in the trap for a given \eqn{T},
\eqn{E} trap depth in eV,
\eqn{s} the frequency factor in 1/s,
\eqn{T} the temperature in K and \eqn{k} the Boltzmann constant in eV/K (cf. Furetta, 2010).
}
\note{
The profiling is currently based on re-sampling from a normal distribution, this
distribution assumption might be, however, not valid for given E and s parameters.
}
\section{Function version}{
 0.1.0
}

\examples{

##EXAMPLE 1
##calculation for two trap-depths with similar frequency factor for different temperatures
E <- c(1.66, 1.70)
s <- 1e+13
T <- 10:20
temp <- calc_ThermalLifetime(
  E = E,
  s = s,
  T = T,
  output_unit = "Ma"
)
graphics::contour(x = E, y = T, z = temp$lifetimes[1,,],
        ylab = "Temperature [\u00B0C]",
        xlab = "Trap depth [eV]",
        main = "Thermal Lifetime Contour Plot"
)
mtext(side = 3, "(values quoted in Ma)")

##EXAMPLE 2
##profiling of thermal life time for E and s and their standard error
E <- c(1.600, 0.003)
s <- c(1e+13,1e+011)
T <- 20
calc_ThermalLifetime(
  E = E,
  s = s,
  T = T,
  profiling = TRUE,
  output_unit = "Ma"
)

} 

\section{How to cite}{
Kreutzer, S., 2026. calc_ThermalLifetime(): Calculates the Thermal Lifetime using the Arrhenius equation. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Furetta, C., 2010. Handbook of Thermoluminescence, Second Edition. World Scientific.
}
\seealso{
\link[graphics:matplot]{graphics::matplot}, \link[stats:Normal]{stats::rnorm}, \link{get_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_CWCurve.R
\name{fit_CWCurve}
\alias{fit_CWCurve}
\title{Non-linear Least Squares Fit for CW-OSL curves -beta version-}
\usage{
fit_CWCurve(
  object,
  n.components.max = 7,
  fit.failure_threshold = 5,
  fit.method = c("port", "LM"),
  fit.trace = FALSE,
  fit.calcError = FALSE,
  LED.power = 36,
  LED.wavelength = 470,
  cex.global = 0.6,
  sample_code = "Default",
  verbose = TRUE,
  output.terminalAdvanced = TRUE,
  plot = TRUE,
  method_control = list(),
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
x, y data of measured values (time and counts).}

\item{n.components.max}{\link{vector} (\emph{optional}):
maximum number of components that are to be used for fitting.
The upper limit is 7.}

\item{fit.failure_threshold}{\link{integer} (\emph{with default}):
limits the failed fitting attempts.}

\item{fit.method}{\link{character} (\emph{with default}):
select the fit method, either \code{"port"} to use the 'port' routine from
function \link{nls}, or \code{"LM"} to use the Levenberg-Marquardt algorithm as
implemented in function \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}.}

\item{fit.trace}{\link{logical} (\emph{with default}):
traces the fitting process on the terminal.}

\item{fit.calcError}{\link{logical} (\emph{with default}):
calculate 1-sigma error range of components using \link[stats:confint]{stats::confint}}

\item{LED.power}{\link{numeric} (\emph{with default}):
LED power (max.) used for intensity ramping in mW/cm².
\strong{Note:} The value is used for the calculation of the absolute
photoionisation cross section.}

\item{LED.wavelength}{\link{numeric} (\emph{with default}):
LED wavelength used for stimulation in nm.
\strong{Note:} The value is used for the calculation of the absolute
photoionisation cross section.}

\item{cex.global}{\link{numeric} (\emph{with default}):
global scaling factor.}

\item{sample_code}{\link{character} (\emph{optional}):
sample code used for the plot and the optional output table (\code{mtext}).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{output.terminalAdvanced}{\link{logical} (\emph{with default}):
enhanced terminal output. Only valid if \code{verbose = TRUE}.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{method_control}{\link{list} (\emph{optional}): options to control the output
produced. Currently only the 'export.comp.contrib.matrix' (logical) option
is supported, to enable/disable export of the component contribution
matrix.}

\item{...}{further arguments and graphical parameters passed to \link{plot}.}
}
\value{
\strong{plot (\emph{optional})}

the fitted CW-OSL curves are returned as plot.

\strong{RLum.Results object}

Beside the plot and table output options, an \linkS4class{RLum.Results} object is
returned.

\code{fit}:
an \code{nls} object (\verb{$fit}) for which generic R functions are
provided, e.g. \link{summary}, \link[stats:confint]{stats::confint}, \link{profile}. For more
details, see \link{nls}.

\code{output.table}:
a \link{data.frame} containing the summarised parameters including the error

\code{component.contribution.matrix}:
\link{matrix} containing the values for the component to sum contribution plot
(\verb{$component.contribution.matrix}).
Produced only if \code{method_control$export.comp.contrib.matrix = TRUE}).

Matrix structure:\cr
Column 1 and 2: time and \code{rev(time)} values \cr
Additional columns are used for the components, two for each component,
containing I0 and n0. The last columns \code{cont.} provide information on
the relative component contribution for each time interval including the row
sum for this values.
}
\description{
The function determines the weighted least-squares estimates of the
component parameters of a CW-OSL signal for a given maximum number of
components and returns various component parameters.

\strong{Fitting function}

The function for the CW-OSL fitting has the general form:

\deqn{y = I0_{1}\lambda_{1}\exp(-\lambda_1 x) + \ldots + I0_{i}\lambda_{i}\exp(-\lambda_i x) }

where \eqn{0 < i < 8}, \eqn{\lambda} is the decay constant,
and \eqn{I0} the initial number of trapped electrons.

\emph{(for the used equation cf. Boetter-Jensen et al., 2003, Eq. 2.31)}

\strong{Start values}

Start values are estimated automatically by fitting a linear function to the
logarithmized input data set. Currently, there is no option to manually
provide start parameters.

\strong{Goodness of fit}

The goodness of the fit is given as pseudoR² value (pseudo coefficient of
determination). According to Lave (1970), the value is calculated as:

\deqn{pseudoR^2 = 1 - RSS/TSS}

where \eqn{RSS} is the Residual Sum of Squares, and \eqn{TSS} is the Total
Sum of Squares.

\strong{Error of fitted component parameters}

The 1-sigma error for the
components is calculated using the function \link[stats:confint]{stats::confint}. Due to
considerable calculation time, this option is deactivated by default. In
addition, the error for the components can be estimated by using internal R
functions like \link{summary}. See the \link{nls} help page
for more information.

\emph{For details on the nonlinear regression in R, see Ritz & Streibig (2008).}
}
\note{
\strong{Beta version - This function has not been properly tested yet and}
\strong{should therefore not be used for publication purposes!}

The pseudo-R² may not be the best parameter to describe the goodness of the
fit. The trade off between the \code{n.components} and the pseudo-R² value
is currently not considered.

The function \strong{does not} ensure that the fitting procedure has reached a
global minimum rather than a local minimum!
}
\section{Function version}{
 0.5.6
}

\examples{

##load data
data(ExampleData.CW_OSL_Curve, envir = environment())

##fit data
fit <- fit_CWCurve(ExampleData.CW_OSL_Curve,
                   main = "CW Curve Fit",
                   n.components.max = 4,
                   log = "x")

} 

\section{How to cite}{
Kreutzer, S., 2026. fit_CWCurve(): Non-linear Least Squares Fit for CW-OSL curves -beta version-. Function version 0.5.6. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Boetter-Jensen, L., McKeever, S.W.S., Wintle, A.G., 2003.
Optically Stimulated Luminescence Dosimetry. Elsevier Science B.V.

Lave, C.A.T., 1970. The Demand for Urban Mass Transportation. The Review of
Economics and Statistics, 52 (3), 320-323.

Ritz, C. & Streibig, J.C., 2008. Nonlinear Regression with R. In: R.
Gentleman, K. Hornik, G. Parmigiani, eds., Springer, p. 150.
}
\seealso{
\link{fit_LMCurve}, \link{plot},\link{nls}, \linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Results}, \link{get_RLum}, \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{dplot}
\keyword{models}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.DeValues}
\alias{ExampleData.DeValues}
\title{Example De data sets for the package Luminescence}
\format{
A \link{list} with two elements, each containing a two column \link{data.frame}:

\describe{
\verb{$BT998}: De and De error values for a fine grain quartz
sample from a loess section in Rottewitz.

\verb{$CA1}: Single grain De
and De error values for a coarse grain quartz sample from a fluvial deposit
in the rock shelter of Cueva Anton
}
}
\description{
Equivalent dose (De) values measured for a fine grain quartz sample from a
loess section in Rottewitz (Saxony/Germany) and for a coarse grain quartz
sample from a fluvial deposit in the rock shelter of Cueva Anton
(Murcia/Spain).
}
\examples{

##(1) plot values as histogram
data(ExampleData.DeValues, envir = environment())
plot_Histogram(ExampleData.DeValues$BT998, xlab = "De [s]")

##(2) plot values as histogram (with second to gray conversion)
data(ExampleData.DeValues, envir = environment())

De.values <- convert_Second2Gray(ExampleData.DeValues$BT998,
                         dose.rate = c(0.0438, 0.0019))

plot_Histogram(De.values, xlab = "De [Gy]")

}
\references{
\strong{BT998}

Unpublished data

\strong{CA1}

Burow, C., Kehl, M., Hilgers, A., Weniger, G.-C., Angelucci, D., Villaverde,
V., Zapata, J. and Zilhao, J.  (2015). Luminescence dating of fluvial
deposits in the rock shelter of Cueva Anton, Spain. Geochronometria 52, 107-125.

\strong{BT998}

\tabular{ll}{
Lab: \tab Luminescence Laboratory Bayreuth\cr
Lab-Code: \tab BT998\cr
Location: \tab Rottewitz (Saxony/Germany)\cr
Material: \tab Fine grain quartz measured on aluminium discs on a Risø TL/OSL DA-15 reader\cr
Units: \tab Values are given in seconds \cr
Dose Rate: \tab Dose rate of the beta-source at measurement ca. 0.0438 Gy/s +/- 0.0019 Gy/s\cr
Measurement Date: \tab 2012-01-27
}

\strong{CA1}

\tabular{ll}{
Lab: \tab Cologne Luminescence Laboratory (CLL)\cr
Lab-Code: \tab C-L2941\cr
Location: \tab Cueva Anton (Murcia/Spain)\cr
Material: \tab Coarse grain quartz (200-250 microns) measured on single grain discs on a Risoe TL/OSL DA-20 reader\cr
Units: \tab Values are given in Gray \cr
Measurement Date: \tab 2012
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply_CosmicRayRemoval.R
\name{apply_CosmicRayRemoval}
\alias{apply_CosmicRayRemoval}
\title{Cosmic-ray removal and spectrum smoothing for RLum.Data.Spectrum objects}
\usage{
apply_CosmicRayRemoval(
  object,
  method = c("smooth", "smooth.spline", "smooth_RLum", "Pych"),
  method.Pych.smoothing = 2,
  method.Pych.threshold_factor = 3,
  MARGIN = 2,
  verbose = FALSE,
  plot = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Spectrum} or
\linkS4class{RLum.Analysis} (\strong{required}):
input object to be treated, which can be also provided as \link{list}. If an
\linkS4class{RLum.Analysis} object is provided, only its
\linkS4class{RLum.Data.Spectrum}
objects are treated. Please note: this mixing of objects does not work for
a list of \code{RLum.Data} objects.}

\item{method}{\link{character} (\emph{with default}):
Defines method that is applied for cosmic ray removal. Allowed methods are
\code{smooth}, the default, (\link[stats:smooth]{stats::smooth}), \code{smooth.spline} (\link[stats:smooth.spline]{stats::smooth.spline}),
\code{smooth_RLum} (\link{smooth_RLum}) and \code{Pych}. See details for further information.}

\item{method.Pych.smoothing}{\link{integer} (\emph{with default}):
Smoothing parameter for cosmic ray removal according to Pych (2003).
The value defines how many neighbouring values in each frame are used for smoothing
(e.g., \code{2} means that the two previous and two following values are used).}

\item{method.Pych.threshold_factor}{\link{numeric} (\emph{with default}):
Threshold for zero-bins in the histogram. Small values mean that more peaks
are removed, but signal might be also affected by this removal.}

\item{MARGIN}{\link{integer} (\emph{with default}):
on which part the function cosmic ray removal should be applied on:
\itemize{
\item 1 = along the time axis (line by line),
\item 2 = along the wavelength axis (column by column).
}

\strong{Note:} This argument only affects methods \code{smooth}, \code{smooth.spline} and
\code{smooth_RLum}.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
If \code{TRUE} the histograms used for the cosmic-ray removal are returned as plot
including the used threshold. Note: A separate plot is returned for each frame!
Currently only for \code{method = "Pych"} a graphical output is provided.}

\item{...}{further arguments and graphical parameters that will be passed
to the \link[stats:smooth]{stats::smooth}, \link[stats:smooth.spline]{stats::smooth.spline} or \link{smooth_RLum}.
See details for more information.}
}
\value{
Returns same object as input.
}
\description{
The function provides several methods for cosmic-ray removal and spectrum
smoothing for \linkS4class{RLum.Data.Spectrum} objects, and those embedded in
\link{list} or \linkS4class{RLum.Analysis} objects.
}
\details{
\strong{\code{method = "Pych"}}

This method applies the cosmic-ray removal algorithm described by Pych
(2003). There are some differences with respect to the publication:
\itemize{
\item For interpolation between neighbouring values, the median is used instead
of the mean.
\item The number of breaks in the histogram is set to half the number of the
input values.
}

For further details see references below.

\strong{Other methods}
Arguments supported through \code{...}
\tabular{llll}{
METHOD \tab ARGUMENT \tab TYPE \tab REMARKS\cr
\code{"smooth"} \tab \code{kind} \tab  \link{character} \tab see \link[stats:smooth]{stats::smooth} \cr
\tab \code{twiceit} \tab  \link{logical} \tab see \link[stats:smooth]{stats::smooth} \cr
\code{"smooth.spline"} \tab \code{spar} \tab  \link{numeric} \tab see \link[stats:smooth.spline]{stats::smooth.spline} \cr
\code{"smooth_RLum"} \tab \code{k} \tab \link{numeric} \tab see \link{smooth_RLum}\cr
\tab \code{fill} \tab \link{numeric} \tab see \link{smooth_RLum}\cr
\tab \code{align} \tab \link{character} \tab see \link{smooth_RLum}\cr
\tab \code{method_smooth_RLum} \tab \link{character} \tab see \link{smooth_RLum}\cr
}

\strong{Best practice}

There is no single silver-bullet strategy for cosmic-ray removal, as it depends
on the characteristic of the detector and the chosen settings. For instance,
high values for pixel binning will improve the light output, but also cause
multiple pixels to be affected by a single cosmic ray. The same is valid for
longer integration times. The best strategy is to combine methods and ensure
that the spectrum is not distorted on a case-to-case basis.

\strong{How to combine methods?}

Different methods can be combined by applying the method repeatedly to the
dataset (see example).
}
\section{Function version}{
 0.4.1
}

\examples{

##(1) - use with your own data and combine (uncomment for usage)
## run two times the default method and smooth with another method
## your.spectrum <- apply_CosmicRayRemoval(your.spectrum, method = "Pych")
## your.spectrum <- apply_CosmicRayRemoval(your.spectrum, method = "Pych")
## your.spectrum <- apply_CosmicRayRemoval(your.spectrum, method = "smooth")

} 

\section{How to cite}{
Kreutzer, S., 2026. apply_CosmicRayRemoval(): Cosmic-ray removal and spectrum smoothing for RLum.Data.Spectrum objects. Function version 0.4.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Pych, W., 2004. A Fast Algorithm for Cosmic-Ray Removal from
Single Images. The Astronomical Society of the Pacific 116 (816), 148-153.
\doi{10.1086/381786}
}
\seealso{
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Analysis},
\link[stats:smooth]{stats::smooth}, \link[stats:smooth.spline]{stats::smooth.spline}, \link{smooth_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_DRCSummary.R
\name{plot_DRCSummary}
\alias{plot_DRCSummary}
\title{Create a Dose-Response Curve Summary Plot}
\usage{
plot_DRCSummary(
  object,
  source_dose_rate = NULL,
  sel_curves = NULL,
  show_dose_points = FALSE,
  show_natural = FALSE,
  n = 51L,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} (\strong{required}): input object created by
\link{analyse_SAR.CWOSL}. The input object can be provided as \link{list}.}

\item{source_dose_rate}{\link{numeric} (\emph{optional}): allows to modify the axis
and show values in Gy, instead seconds. Only a single numerical value is
allowed.}

\item{sel_curves}{\link{numeric} (\emph{optional}): id of the curves to be plotted in
its occurring order. A sequence can be provided for selecting, e.g., only
every 2nd curve from the input object.}

\item{show_dose_points}{\link{logical} (\emph{with default}): enable/disable plotting
of dose points in the graph.}

\item{show_natural}{\link{logical} (\emph{with default}): enable/disable the plot
of the natural \code{Lx/Tx} values.}

\item{n}{\link{integer} (\emph{with default}): number of x-values used to evaluate
one curve object. Large numbers slow down the plotting process and are
usually not needed.}

\item{...}{Further arguments and graphical parameters to be passed.
In particular: \code{main}, \code{xlab}, \code{ylab}, \code{xlim}, \code{ylim}, \code{lty}, \code{lwd}, \code{pch},
\code{col.pch}, \code{col.lty}, \code{mtext}}
}
\value{
An \linkS4class{RLum.Results} object is returned:

Slot: \strong{@data}\cr

\tabular{lll}{
\strong{OBJECT} \tab \strong{TYPE} \tab \strong{COMMENT}\cr
\code{results} \tab \link{data.frame} \tab with dose and LxTx values \cr
\code{data} \tab \linkS4class{RLum.Results} \tab original input data \cr
}

Slot: \strong{@info}\cr

\tabular{lll}{
\strong{OBJECT} \tab \strong{TYPE} \tab \strong{COMMENT} \cr
\code{call} \tab \code{call} \tab the original function call \cr
\code{args} \tab \code{list} \tab arguments of the original function call \cr
}

\emph{Note: If the input object is a \link{list} a list of \linkS4class{RLum.Results} objects is returned.}
}
\description{
While analysing OSL SAR or pIRIR-data the view on the data is usually
limited to one dose-response curve (DRC) at the time for one aliquot. This
function overcomes this limitation by plotting all DRCs from an
\linkS4class{RLum.Results} object created by \link{analyse_SAR.CWOSL} in one single
plot.

If you want plot your DRC on an energy scale (dose in Gy), you can either
use option \code{source_dose_rate} or perform your SAR analysis with the dose
points in Gy (better axis scaling).
}
\section{Function version}{
 0.2.4
}

\examples{

#load data example data
data(ExampleData.BINfileData, envir = environment())

#transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

## perform SAR analysis
results <- analyse_SAR.CWOSL(
    object,
    signal_integral = 1:2,
    background_integral = 900:1000,
    plot = FALSE
)

##plot only DRC
plot_DRCSummary(results)

}
\seealso{
\linkS4class{RLum.Results}, \link{analyse_SAR.CWOSL}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Burow, C., 2026. plot_DRCSummary(): Create a Dose-Response Curve Summary Plot. Function version 0.2.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_Activity2Concentration.R
\name{convert_Activity2Concentration}
\alias{convert_Activity2Concentration}
\title{Convert Nuclide Activities to Abundance and Vice Versa}
\usage{
convert_Activity2Concentration(
  data,
  input_unit = c("activity", "abundance"),
  verbose = TRUE
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}):
provide dose rate data (activity or concentration) in three columns.
The first column indicates the nuclide, the 2nd column measured value and
in the 3rd column its error value. Allowed nuclide data are
\code{'U-238'}, \code{'Th-232'} and \code{'K-40'}. See examples for an example.}

\item{input_unit}{\link{character} (\emph{with default}):
specify unit of input data given in the dose rate data frame, one of
\code{"activity"} (considered as given Bq/kg, the default) and \code{"abundance"}
(considered as given in mug/g or mass. \%).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}
}
\value{
Returns an \linkS4class{RLum.Results} object with a \link{data.frame} containing
input and newly calculated values. Please not that in the column header µg/g
is written as \code{mug/g} due to the R requirement to maintain packages portable using
ASCII characters only.
}
\description{
The function performs the conversion of the specific activities into
mass abundance and vice versa for the radioelements U, Th, and K to
harmonise the measurement unit with the required data input unit of
potential analytical tools for, e.g. dose rate calculation or related
functions such as \link{use_DRAC}.
}
\details{
The conversion from nuclide activity of a sample to nuclide concentration
is performed using conversion factors that are based on the mass-related
specific activity of the respective nuclide.

Constants used in this function were obtained from \verb{https://physics.nist.gov/cuu/Constants/}
all atomic weights and composition values from
\verb{https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses}
and the nuclide data from \verb{https://www.iaea.org/resources/databases/livechart-of-nuclides-advanced-version}

The factors can be calculated using the equation:
\deqn{
A = N_A \frac{N_{abund}}{N_{mol.mass}} \ln(2) / N.half.life
}

to convert in µg/g we further use:
\deqn{
f = A / 10^6
}

where:
\itemize{
\item \code{N_A} - Avogadro constant in 1/mol
\item \code{A} - specific activity of the nuclide in Bq/kg
\item \code{N.abund} - relative natural abundance of the isotope
\item \code{N.mol.mass} molar mass in kg/mol
\item \code{N.half.life} half-life of the nuclide in s
}

example for calculating the activity of the radionuclide U-238:
\itemize{
\item \code{N_A} = 6.02214076e+23 (1/mol)
\item \code{T_0.5} = 1.41e+17 (s)
\item \code{m_U_238} = 0.23802891 (kg/mol)
\item \code{U_abund} = 0.992745 (unitless)
}

\deqn{A_{U} = N_{A} * U_{abund} / m_{U_238} * \ln(2) / T_{1/2} = 2347046} (Bq/kg)

\deqn{f.U = A_{U} / 10^6}
}
\note{
Although written otherwise for historical reasons. Input values must be element values.
For instance, if a value is provided for U-238 the function assumes that this value
represents the sum (activity or abundance) of U-238, U-235 and U-234.
In other words, 1 µg/g of U means that this is the composition of 0.992 parts of U-238,
0.000054 parts of U-234, and 0.00072 parts of U-235.
}
\section{Function version}{
 0.1.2
}

\examples{

##construct data.frame
data <- data.frame(
 NUCLIDES = c("U-238", "Th-232", "K-40"),
 VALUE = c(40,80,100),
 VALUE_ERROR = c(4,8,10),
 stringsAsFactors = FALSE)

##perform analysis
convert_Activity2Concentration(data)

} 

\section{How to cite}{
Fuchs, M.C., 2026. convert_Activity2Concentration(): Convert Nuclide Activities to Abundance and Vice Versa. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Debertin, K., Helmer, R.G., 1988. Gamma- and X-ray Spectrometry with
Semiconductor Detectors, Elsevier Science Publishers, p.283

Wiechen, A., Ruehle, H., Vogl, K., 2013. Bestimmung der massebezogenen
Aktivitaet von Radionukliden. AEQUIVAL/MASSAKT, ISSN 1865-8725,
\url{https://www.bundesumweltministerium.de/fileadmin/Daten_BMU/Download_PDF/Strahlenschutz/aequival-massakt_v2013-07_bf.pdf}
}
\author{
Margret C. Fuchs, Helmholtz-Institute Freiberg for Resource Technology (Germany)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_TIFF2R.R
\name{read_TIFF2R}
\alias{read_TIFF2R}
\title{Import TIFF Image Data into R}
\usage{
read_TIFF2R(file, merge2stack = FALSE, verbose = TRUE, ...)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of the TIFF file to read (URLs are supported).}

\item{merge2stack}{\link{logical} (\emph{with default}): if \code{file} is a \link{list} it merges
the individual images into one image stack. Please note that the smallest image
dimension determines pixel dimension of the output stack.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{not in use, for compatibility reasons only}
}
\value{
Returns an \linkS4class{RLum.Data.Image} object. Results are
returned as a list when multiple files are processed or \code{file} is a list.
}
\description{
Simple wrapper around \link[tiff:readTIFF]{tiff::readTIFF} to import TIFF images and TIFF image
stacks to be further processed within the 'Luminescence' package.
}
\section{Function version}{
 0.2.1
}

\examples{

## use system file
file <- system.file("extdata", "TIFFfile.tif", package = "Luminescence")

## import image
image <- read_TIFF2R(file)

}
\seealso{
\link[tiff:readTIFF]{tiff::readTIFF}, \linkS4class{RLum.Data.Image}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. read_TIFF2R(): Import TIFF Image Data into R. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum.Data-class.R
\docType{class}
\name{RLum.Data-class}
\alias{RLum.Data-class}
\title{Class \code{"RLum.Data"}}
\description{
Generalized virtual data class for luminescence data.
}
\section{Objects from the Class}{

A virtual class: no objects can be created from it.
}

\section{Class version}{
 0.2.1
}

\examples{

showClass("RLum.Data")

}
\seealso{
\linkS4class{RLum}, \linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Data.Image}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_IRSAR.RF.R
\name{analyse_IRSAR.RF}
\alias{analyse_IRSAR.RF}
\title{Analyse IRSAR RF measurements}
\usage{
analyse_IRSAR.RF(
  object,
  sequence_structure = c("NATURAL", "REGENERATED"),
  RF_nat.lim = NULL,
  RF_reg.lim = NULL,
  method = c("FIT", "SLIDE", "VSLIDE", "NONE"),
  method_control = NULL,
  test_parameters = NULL,
  n.MC = 10,
  txtProgressBar = TRUE,
  plot = TRUE,
  plot_reduced = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} or a \link{list} of \linkS4class{RLum.Analysis}-objects (\strong{required}):
input object containing data for protocol analysis. The function expects to
find at least two curves in the \linkS4class{RLum.Analysis} object: (1) \code{RF_nat}, (2) \code{RF_reg}.
If a \code{list} is provided as input all other parameters can be provided as
\code{list} as well to gain full control.}

\item{sequence_structure}{\link{vector} \link{character} (\emph{with default}):
specifies the general sequence structure. Allowed steps are \code{NATURAL} and
\code{REGENERATED}, and at least one of each must appear. In addition, any other
character is allowed in the sequence structure; such curves will be ignored
during the analysis. If \code{sequence_structure} is shorter than the number of
curves in \code{object}, it is recycled. If multiple \code{NATURAL} and \code{REGENERATED}
steps are specified, the corresponding measurements are stacked.}

\item{RF_nat.lim}{\link{vector} (\emph{with default}):
set minimum and maximum channel range for natural signal fitting and sliding.
If only one value is provided this will be treated as minimum value and the
maximum limit will be added automatically.}

\item{RF_reg.lim}{\link{vector} (\emph{with default}):
set minimum and maximum channel range for regenerated signal fitting and sliding.
If only one value is provided this will be treated as minimum value and the
maximum limit will be added automatically.}

\item{method}{\link{character} (\emph{with default}): select method applied for the
data analysis. Possible options are \code{"FIT"}, \code{"SLIDE"}, \code{"VSLIDE"};
\code{"NONE"} can be used to disable the analysis and plot the natural points
at their original position.}

\item{method_control}{\link{list} (\emph{optional}):
parameters to control the method, that can be passed to the chosen method.
These are for (1) \code{method = "FIT"}: \code{'trace'}, \code{'maxiter'}, \code{'warnOnly'}, \code{'minFactor'} and for
(2) \code{method = "SLIDE"}: \code{'correct_onset'}, \code{'show_density'},  \code{'show_fit'}, \code{'trace'}.
See details.}

\item{test_parameters}{\link{list} (\emph{with default}):
set test parameters. Supported parameters are: \code{curves_ratio},
\code{residuals_slope} (only for \code{method = "SLIDE"} and \code{"VSLIDE"}),
\code{curves_bounds}, \code{dynamic_ratio}, \code{lambda}, \code{beta} and \code{delta.phi}.
All input: \link{numeric} values, \code{NA} and \code{NULL} (see Details for further
information).}

\item{n.MC}{\link{numeric} (\emph{with default}):
number of Monte Carlo runs for the estimation of the start parameter
(\code{method = "FIT"}) or of the error (\code{method = "SLIDE"} and \code{"VSLIDE"}).
This value can be set to \code{NULL} to skip the MC runs.
Note: Large values will significantly increase the computation time.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar during MC runs.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{plot_reduced}{\link{logical} (\emph{optional}):
provides a reduced plot output if enabled to allow common R plot combinations,
e.g., \code{par(mfrow(...))}. If \code{TRUE} no residual plot
is returned; it has no effect if \code{plot = FALSE}}

\item{...}{further arguments that will be passed to the plot output.
Currently supported arguments are \code{main}, \code{mtext}, \code{xlab}, \code{ylab},
\code{xlim}, \code{ylim}, \code{log}, \code{legend} (\code{TRUE/FALSE}),
\code{legend.pos}, \code{legend.text} (passes argument to x,y in
\link[graphics:legend]{graphics::legend}), \code{xaxt}, \code{verbose} (\code{TRUE/FALSE}).}
}
\value{
The function returns numerical output and an (\emph{optional}) plot.

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\verb{[.. $data : data.frame]}

\tabular{lll}{
\strong{Column} \tab \strong{Type} \tab \strong{Description}\cr
\code{DE} \tab \code{numeric} \tab the obtained equivalent dose\cr
\code{DE.ERROR} \tab \code{numeric} \tab (only \code{method = "SLIDE"}) standard deviation obtained from MC runs \cr
\code{DE.LOWER} \tab \code{numeric}\tab 2.5\% quantile for De values obtained by MC runs \cr
\code{DE.UPPER} \tab \code{numeric}\tab 97.5\% quantile for De values obtained by MC runs  \cr
\code{DE.STATUS}  \tab \code{character}\tab test parameter status\cr
\code{RF_NAT.LIM}  \tab \code{character}\tab used \code{RF_nat} curve limits \cr
\code{RF_REG.LIM} \tab \code{character}\tab used \code{RF_reg} curve limits\cr
\code{POSITION} \tab \code{integer}\tab (\emph{optional}) position of the curves\cr
\code{DATE} \tab \code{character}\tab (\emph{optional}) measurement date\cr
\code{SEQUENCE_NAME} \tab \code{character}\tab (\emph{optional}) sequence name\cr
\code{UID} \tab \code{character}\tab unique data set ID
}

\verb{[.. $De.MC : numeric]}

A \code{numeric} vector with all the De values obtained by the MC runs.

\verb{[.. $test_parameters : data.frame]}

\tabular{lll}{
\strong{Column} \tab \strong{Type} \tab \strong{Description}\cr
\code{POSITION} \tab \code{numeric} \tab aliquot position \cr
\code{PARAMETER} \tab \code{character} \tab test parameter name \cr
\code{THRESHOLD} \tab \code{numeric} \tab set test parameter threshold value \cr
\code{VALUE} \tab \code{numeric} \tab the calculated test parameter value (to be compared with the threshold)\cr
\code{STATUS} \tab \code{character} \tab test parameter status either \code{"OK"} or \code{"FAILED"} \cr
\code{SEQUENCE_NAME} \tab \code{character} \tab name of the sequence, so far available \cr
\code{UID} \tab \code{character}\tab unique data set ID
}

\verb{[.. $fit : data.frame]}

An \link{nls} object produced by the fitting.

\verb{[.. $slide : list]}

A \link{list} with data produced during the sliding. Some elements are previously
reported with the summary object data. List elements are:

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\code{De} \tab \code{numeric} \tab the final De obtained with the sliding approach \cr
\code{De.MC} \tab \code{numeric} \tab all De values obtained by the MC runs \cr
\code{residuals} \tab \code{numeric} \tab the obtained residuals for each channel of the curve \cr
\code{trend.fit} \tab \code{lm} \tab fitting results produced by the fitting of the residuals \cr
\code{RF_nat.slid} \tab \code{matrix} \tab the slid \code{RF_nat} curve \cr
\code{t_n.id} \tab \code{numeric} \tab the index of the t_n offset \cr
\code{I_n} \tab \code{numeric} \tab the vertical intensity offset if a vertical slide was applied \cr
\code{algorithm_error} \tab \code{numeric} \tab the vertical sliding suffers from a systematic effect induced by the used
algorithm. The returned value is the standard deviation of all obtained De values while expanding the
vertical sliding range. It can be added as systematic error to the final De error; so far wanted.\cr
\code{vslide_range} \tab \code{numeric} \tab the range used for the vertical sliding \cr
\code{num_slide_windows} \tab \code{integer} \tab the number of windows used for the vertical sliding \cr
\code{squared_residuals} \tab \code{numeric} \tab the squared residuals (horizontal sliding)
}

\strong{slot:} \strong{\verb{@info}}

The original function call (\link[methods:LanguageClasses]{methods::language}-object)

The output (\code{data}) should be accessed using the function \link{get_RLum}.

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr

The slid IR-RF curves with the finally obtained De
}
\description{
The function analyses IRSAR RF measurements on K-feldspar samples performed
using the protocol according to Erfurt et al. (2003) and beyond.
}
\details{
The function performs an IRSAR analysis described for K-feldspar samples by
Erfurt et al. (2003) assuming a negligible sensitivity change of the RF
signal.

\strong{General Sequence Structure} (according to Erfurt et al., 2003)
\enumerate{
\item Measuring IR-RF intensity of the natural dose for a few seconds (\eqn{RF_{nat}})
\item Bleach the samples under solar conditions for at least 30 min without changing the geometry
\item Waiting for at least one hour
\item Regeneration of the IR-RF signal to at least the natural level (measuring (\eqn{RF_{reg}})
\item Fitting data with a stretched exponential function
\item Calculate the palaeodose \eqn{D_{e}} using the parameters from the fitting
}

Three methods are supported to obtain the \eqn{D_{e}}:
\code{method = "FIT"}, \code{method = "SLIDE"} and \code{method = "VSLIDE"}:

\strong{\code{method = "FIT"}}

The principle is described above and follows the original suggestions by
Erfurt et al., 2003. For the fitting, the mean count value of the \code{RF_nat}
curve is used.

Function used for the fitting (according to Erfurt et al. (2003)):

\deqn{\phi(D) = \phi_{0} - \Delta\phi(1 - \exp(-\lambda D))^\beta}

with
\eqn{\phi(D)} the dose dependent IR-RF flux,
\eqn{\phi_{0}} the initial IR-RF flux,
\eqn{\Delta\phi} the dose dependent change of the IR-RF flux,
\eqn{\lambda} the exponential parameter, \eqn{D} the dose and
\eqn{\beta} the dispersive factor.

To obtain the palaeodose
\eqn{D_{e}} the function is changed to:

\deqn{D_{e} = \ln(-(\phi(D) - \phi_{0})/(-\lambda\phi)^{1/\beta}+1)/-\lambda}

The fitting is done using the \code{port} algorithm of the \link{nls} function.

\strong{\code{method = "SLIDE"}}

For this method, the natural curve is slid along the x-axis until
congruence with the regenerated curve is reached. Unlike fitting, this
allows for working with the original data without the need for any physical
model. This approach was introduced for RF curves by Buylaert et al., 2012
and Lapp et al., 2012.

Here the sliding is done by searching for the minimum of the sum of
squared residuals.
For the mathematical details of the implementation see Frouin et al., 2017

\strong{\code{method = "VSLIDE"}}

Same as \code{"SLIDE"} but searching also vertically for the best match (i.e. in xy-direction.)
See Kreutzer et al. (2017) and Murari et al. (2021). By default, the vertical sliding
range is set automatically, but can be set manually by changing the
\code{vslide_range} parameter (see \code{method_control}).

\strong{\code{method_control}}

To keep the generic argument list as clear as possible, parameters to control
the methods for De estimation are preset with meaningful default values,
which can however be modified using the \code{method_control} argument, e.g.,
\code{method_control = list(trace = TRUE)}. Supported parameters are:

\emph{For FIT}
\itemize{
\item \code{trace} (\link{logical}, default: \code{FALSE}): as in \link{nls}; shows sum of squared
residuals.
\item \code{maxiter} (\link{integer}, default: 500): as in \link{nls}.
\item \code{warnOnly} (\link{logical}, default: \code{FALSE}): as in \link{nls}.
\item \code{minFactor} (\link{numeric}, default: \code{1 / 4096}): as in \link{nls}.
}

\emph{For SLIDE or VSLIDE}
\itemize{
\item \code{trace} (\link{logical}, default: \code{FALSE}): as in \link{nls}; shows sum of squared
residuals.
\item \code{trace_vslide} (\link{logical}, default: \code{FALSE}): enable/disable the tracing
of the vertical sliding.
\item \code{correct_onset} (\link{logical}, default: \code{TRUE}): whether the curves should
be shifted along the x-axis by the first channel, as light is expected in
the first channel.
\item \code{show_density} (\link{logical}, default: \code{TRUE}): enable/disable KDE plots for
MC run results. Nothing is shown if the distribution is too narrow.
\item \code{show_fit} (\link{logical}, default: \code{FALSE}): enable/disable the plot of the
fitted curve routinely obtained during the evaluation.
\item \code{n.MC} (\link{integer}, default: 1000): number of Monte Carlo runs within the
sliding (assessing the possible minimum values). \strong{Note:} This parameter
is not the same as the function argument \code{n.MC}.
\item \code{vslide_range} (\link{numeric} or \link{character}, default: "auto"): boundaries
for the vertical curve sliding. The argument expects a vector with absolute
minimum and maximum (e.g., \code{c(-1000,1000)}). The default "auto" mode
detects the reasonable vertical sliding range (\emph{recommended}). \code{NULL}
disables the vertical sliding.
\item \code{num_slide_windows} (\link{integer}, default: 3): number of  differently-sized
windows tested when sliding: the higher the value (up to a maximum of 10),
the more time is spent in searching the global optimum. The default setting
attempts to strike a balance between quality of the fit and computation
speed.
\item \code{cores} (\link{numeric} or \link{character}, default: \code{NULL}): number of cores
allocated for a parallel processing of the Monte-Carlo runs. The default
value corresponds to single-threaded computation; the recommended values is
\code{"auto"}, which assigns all but two of the available cores.
}

\strong{Error estimation}

For \strong{\code{method = "FIT"}}, the \eqn{D_{e}} error range is obtained by using
the 2.5 \% (lower) and the 97.5 \% (upper) quantiles of the \eqn{RF_{nat}}
curve.

For \strong{\code{method = "SLIDE"}} and \strong{\code{method = "VSLIDE"}}, the error is obtained
by bootstrapping the residuals of the slid curve to construct new natural
curves for a Monte Carlo simulation. The error is returned in two ways:
(a) the standard deviation of the \eqn{D_{e}} obtained from the MC runs and
(b) the confidence interval using the 2.5 \% (lower) and the 97.5 \% (upper)
quantiles. The results of the MC runs are returned with the function output.

\strong{Test parameters}

The argument \code{test_parameters} allows to pass thresholds for several test
parameters, which will be evaluated during the function run. If a threshold
is set and it is exceeded, the test parameter status will be set to
\code{"FAILED"}. This argument is intentionally not termed \code{'rejection criteria'}
as not all test parameters are evaluated for both methods and some parameters
are calculated but not evaluated by default.

\code{NA} and \code{NULL} are the allowed values for all parameters. If the parameter
is set to \code{NA}, the value is calculated but the result will not be evaluated,
therefore it will have no effect on the status (\code{"OK"} or \code{"FAILED"}) of
the parameter. Setting the parameter to \code{NULL} disables the parameter
entirely and the parameter will also be removed from the function output.
This might be useful in cases where a particular parameter requires a long
computation time. Currently supported parameters are:
\itemize{
\item \code{curves_ratio} (\link{numeric}, default: \code{1.001}):
the ratio of \eqn{RF_{nat}} to \eqn{RF_{reg}} is calculated over the range
spanned by \eqn{RF_{nat}}, and should not exceed the threshold value.
\item \code{intersection_ratio} (\link{numeric}, default: \code{NA}):
calculated as absolute difference from 1 of the ratio of the integral of
the normalised RF-curves. This value indicates intersection of the RF-curves
and should be close to 0 if the curves have a similar shape. For this
calculation first the corresponding time-count pair value on the RF_reg
curve is obtained using the maximum count value of the \code{RF_nat} curve and
only this segment (fitting to the \code{RF_nat} curve) on the RF_reg curve is
taken for further calculating this ratio. If nothing is found at all, \code{Inf}
is returned.
\item \code{residuals_slope} (\link{numeric}, default: \code{NA}; only for \code{method = "SLIDE"}
and \verb{"VSLIDE}"):
a linear function is fitted on the residuals after sliding.
The corresponding slope can be used to discard values as a high (positive,
negative) slope may indicate that both curves are fundamentally different
and the method cannot be applied at all.
By default, the value of this parameter is calculated but not evaluated.
\item \code{curves_bounds} (\link{numeric}, default: \eqn{\max(RF_{reg_counts})}):
this measure uses the maximum time (x) value of the regenerated curve.
The maximum time (x) value of the natural curve cannot be larger than this
value. However, although this is not recommended the value can be changed
or disabled.
\item \code{dynamic_ratio} (\link{numeric}, default: \code{NA}):
the dynamic ratio of the regenerated curve is calculated as ratio of the
minimum and maximum count values.
\item \code{lambda}, \code{beta} and \code{delta.phi} \link{numeric} (default: \code{NA}):
the stretched exponential function suggested by Erfurt et al. (2003)
describing the decay of the RF signal, comprises several parameters that
might be useful to evaluate the shape of the curves. For \code{method = "FIT"},
this parameter is obtained during the fitting; for \code{method = "SLIDE"} a
rather rough estimation is made using the function \link[minpack.lm:nlsLM]{minpack.lm::nlsLM} and
the equation given above.
\strong{Note:} As this procedure requests more computation time, it is performed
only if all three parameters are set.
}
}
\note{
This function assumes that there is no sensitivity change during the
measurements (natural vs. regenerated signal), which is in contrast to the
findings by Buylaert et al. (2012).
}
\section{Function version}{
 0.7.10
}

\examples{

##load data
data(ExampleData.RLum.Analysis, envir = environment())

##(1) perform analysis using the method 'FIT'
results <- analyse_IRSAR.RF(object = IRSAR.RF.Data)

##show De results and test parameter results
get_RLum(results, data.object = "data")
get_RLum(results, data.object = "test_parameters")

##(2) perform analysis using the method 'SLIDE'
data(ExampleData.RF70Curves, envir = environment())
results <- analyse_IRSAR.RF(
object = RF70Curves,
method = "SLIDE",
 n.MC = 1)

\dontrun{
##(3) perform analysis using the method 'VSLIDE' and method control option
## 'trace
results <- analyse_IRSAR.RF(
 object = RF70Curves,
 method = "VSLIDE",
 method_control = list(trace = TRUE))
}

} 

\section{How to cite}{
Kreutzer, S., 2026. analyse_IRSAR.RF(): Analyse IRSAR RF measurements. Function version 0.7.10. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Buylaert, J.P., Jain, M., Murray, A.S., Thomsen, K.J., Lapp, T.,
2012. IR-RF dating of sand-sized K-feldspar extracts: A test of accuracy.
Radiation Measurements 44 (5-6), 560-565. \doi{10.1016/j.radmeas.2012.06.021}

Erfurt, G., Krbetschek, M.R., 2003. IRSAR - A single-aliquot
regenerative-dose dating protocol applied to the infrared radiofluorescence
(IR-RF) of coarse- grain K-feldspar. Ancient TL 21, 35-42.
\doi{10.26034/la.atl.2003.358}

Erfurt, G., 2003. Infrared luminescence of Pb+ centres in potassium-rich
feldspars. physica status solidi (a) 200, 429-438. \doi{10.1002/pssa.200306700}

Erfurt, G., Krbetschek, M.R., 2003. Studies on the physics of the infrared
radioluminescence of potassium feldspar and on the methodology of its
application to sediment dating. Radiation Measurements 37, 505-510.
\doi{10.1016/s1350-4487(03)00058-1}

Erfurt, G., Krbetschek, M.R., Bortolot, V.J., Preusser, F., 2003. A fully
automated multi-spectral radioluminescence reading system for geochronometry
and dosimetry. Nuclear Instruments and Methods in Physics Research Section
B: Beam Interactions with Materials and Atoms 207, 487-499.
\doi{10.1016/s0168-583x(03)01121-2}

Frouin, M., Huot, S., Kreutzer, S., Lahaye, C., Lamothe, M., Philippe, A., Mercier, N., 2017.
An improved radiofluorescence single-aliquot regenerative dose protocol for K-feldspars.
Quaternary Geochronology 38, 13-24. \doi{10.1016/j.quageo.2016.11.004}

Kreutzer, S., Murari, M.K., Frouin, M., Fuchs, M., Mercier, N., 2017.
Always remain suspicious: a case study on tracking down a technical artefact while measuring IR-RF.
Ancient TL 35, 20–30. \doi{10.26034/la.atl.2017.510}

Murari, M.K., Kreutzer, S., Fuchs, M., 2018. Further investigations on IR-RF:
Dose recovery and correction. Radiation Measurements 120, 110–119.
doi: 10.1016/j.radmeas.2018.04.017 \doi{10.1016/j.radmeas.2018.04.017}

Lapp, T., Jain, M., Thomsen, K.J., Murray, A.S., Buylaert, J.P., 2012. New
luminescence measurement facilities in retrospective dosimetry. Radiation
Measurements 47, 803-808.
\doi{10.1016/j.radmeas.2012.02.006}

Trautmann, T., 2000. A study of radioluminescence kinetics of natural
feldspar dosimeters: experiments and simulations. Journal of Physics D:
Applied Physics 33, 2304-2310.
\doi{10.1088/0022-3727/33/18/315}

Trautmann, T., Krbetschek, M.R., Dietrich, A., Stolz, W., 1998.
Investigations of feldspar radioluminescence: potential for a new dating
technique. Radiation Measurements 29, 421-425.
\doi{10.1016/s1350-4487(98)00012-2}

Trautmann, T., Krbetschek, M.R., Dietrich, A., Stolz, W., 1999. Feldspar
radioluminescence: a new dating method and its physical background. Journal
of Luminescence 85, 45-58.
\doi{10.1016/s0022-2313(99)00152-0}

Trautmann, T., Krbetschek, M.R., Stolz, W., 2000. A systematic study of the
radioluminescence properties of single feldspar grains. Radiation
Measurements 32, 685-690.
\doi{10.1016/s1350-4487(00)00077-9}

** Further reading**

Murari, M.K., Kreutzer, S., King, G.E., Frouin, M., Tsukamoto, S., Schmidt, C., Lauer, T.,
Klasen, N., Richter, D., Friedrich, J., Mercier, N., Fuchs, M., 2021.
Infrared radiofluorescence (IR-RF) dating: A review. Quaternary Geochronology 64,
101155. \doi{10.1016/j.quageo.2021.101155}
}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Results}, \link{get_RLum},
\link{nls}, \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}, \link[parallel:mclapply]{parallel::mclapply}, \link{ExampleData.RF70Curves}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_Al2O3C_CrossTalk.R
\name{analyse_Al2O3C_CrossTalk}
\alias{analyse_Al2O3C_CrossTalk}
\title{Al2O3:C Reader Cross-Talk Analysis}
\usage{
analyse_Al2O3C_CrossTalk(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  dose_points = c(0, 4),
  recordType = "OSL (UVVIS)",
  irradiation_time_correction = NULL,
  method_control = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} or \link{list} (\strong{required}):
measurement input}

\item{signal_integral}{\link{numeric} (\emph{optional}):
signal integral, used for the signal and the background.
If nothing is provided, the full range is used.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{dose_points}{\link{numeric} (\emph{with default}):
vector with dose points, if dose points are repeated, only the general
pattern needs to be provided. Default values follow the suggestions
made by Kreutzer et al., 2018.}

\item{recordType}{\link{character} (\emph{with default}):
input curve selection, which is passed to \link{get_RLum}. To deactivate the
automatic selection set the argument to \code{NULL}.}

\item{irradiation_time_correction}{\link{numeric} or \linkS4class{RLum.Results} (\emph{optional}):
information on the used irradiation time correction obtained by another
experiment.}

\item{method_control}{\link{list} (\emph{optional}):
optional parameters to control the calculation.
See details for further explanations.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments that can be passed to the plot output}
}
\value{
Function returns results numerically and graphically:

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$data} \tab \code{data.frame} \tab summed apparent dose table \cr
\verb{$data_full} \tab \code{data.frame} \tab full apparent dose table \cr
\verb{$fit} \tab \code{lm} \tab the linear model obtained from fitting \cr
\verb{$col.seq} \tab \code{numeric} \tab the used colour vector \cr
}

\strong{slot:} \strong{\verb{@info}}

The original function call

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr
\itemize{
\item An overview of the obtained apparent dose values
}
}
\description{
The function provides the analysis of cross-talk measurements
on a FI lexsyg SMART reader using Al2O3:C chips.
}
\section{Function version}{
 0.1.4
}

\examples{

##load data
data(ExampleData.Al2O3C, envir = environment())

##run analysis
analyse_Al2O3C_CrossTalk(data_CrossTalk)

} 

\section{How to cite}{
Kreutzer, S., 2026. analyse_Al2O3C_CrossTalk(): Al2O3:C Reader Cross-Talk Analysis. Function version 0.1.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018. Environmental Dose Rate
Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
Geochronometria 45, 56-67. doi: 10.1515/geochr-2015-0086
}
\seealso{
\link{analyse_Al2O3C_ITC}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_EED_Model.R
\name{calc_EED_Model}
\alias{calc_EED_Model}
\title{Modelling Exponential Exposure Distribution}
\usage{
calc_EED_Model(
  data,
  D0 = 120L,
  expected_dose = NULL,
  MinIndivDose = NULL,
  MaxIndivDose = NULL,
  kappa = NULL,
  sigma_distr = NULL,
  n.simul = 5000L,
  n.minSimExp = 50L,
  sample_name = "",
  method_control = list(),
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}): input data consisting of two columns, the De and the
SE(De). Values are expected in Gy}

\item{D0}{\link{integer} (\emph{with default}): D0 value (in Gy), defining the
characterisation behaviour of the quartz.}

\item{expected_dose}{\link{numeric} (\strong{required}):
expected equivalent dose.}

\item{MinIndivDose}{\link{numeric} (\emph{with default}): value specifying the minimum dose taken into
account for the plateau. \code{NULL} applies all values.}

\item{MaxIndivDose}{\link{numeric} (\emph{with default}): value specifying the maximum dose taken into
account for the plateau. \code{NULL} applies all values.}

\item{kappa}{\link{numeric} (\emph{optional}): positive dimensionless exposure parameter
characterising the bleaching state of the grains. Low values (< 10) indicate
poor bleaching}

\item{sigma_distr}{\link{numeric} (\emph{optional}): positive dose rate parameter, representing the
dose variability to which the grains were exposed ##TODO perhaps it should be renamed}

\item{n.simul}{\link{integer} (\emph{with default}): number of simulations}

\item{n.minSimExp}{\link{integer} (\emph{with default}): number of MC runs for calculating the uncertainty
contribution from the sampling}

\item{sample_name}{\link{character} (\emph{with default}): name of the sample}

\item{method_control}{\link{list} (\emph{with default}): additional deep control parameters, parameters
need to be provided as named list, see details}

\item{verbose}{\link{logical} (\emph{with default}): enable/disable output to the
terminal.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{...}{further parameters that can be passed to better control the plot output. Support arguments
are \code{xlab}, \code{xlim}.}
}
\description{
Modelling incomplete and heterogeneous bleaching of mobile grains partially
exposed to the light, an implementation of the EED model proposed by Guibert et al. (2019).
}
\details{
The function is an implementation and enhancement of the scripts used in
Guibert et al. (2019). The implementation supports a semi-automated estimation
of the parameters \code{kappa} and \code{sigma_distr}. If set to \code{NULL}, a surface interpolation
is used to estimated those values.

\strong{Method control parameters}

\tabular{llll}{
\strong{ARGUMENT} \tab \strong{FUNCTION} \tab \strong{DEFAULT} \tab \strong{DESCRIPTION}\cr
\code{lower}  \tab  - \tab \code{c(0.1,0,0)} \tab set lower bounds for kappa, sigma, and the expected De in auto mode \cr
\code{upper} \tab  - \tab \code{c(1000,2)} \tab set upper bounds for kappa, sigma, and the expected De in auto mode \cr
\code{iter_max} \tab - \tab \code{1000} \tab maximum number for iterations for used to find kappa and sigma \cr
\code{trace} \tab - \tab \code{FALSE} \tab enable/disable terminal trace mode; overwritten by global argument \code{verbose}\cr
\code{trace_plot} \tab - \tab \code{FALSE} \tab enable/disable additional trace plot output; overwritten by global argument \code{verbose} \cr
}
}
\section{Function version}{
 0.1.1
}

\examples{

data(ExampleData.MortarData, envir = environment())
calc_EED_Model(
 data = MortarData,
 kappa = 14,
 sigma_distr = 0.37,
 expected_dose = 11.7)

## automated estimation of
## sigma_distribution and
## kappa
\dontrun{
 calc_EED_Model(
 data = MortarData,
 kappa = NULL,
 sigma_distr = NULL,
 expected_dose = 11.7)
}

} 

\section{How to cite}{
Guibert, P., Kreutzer, S., 2026. calc_EED_Model(): Modelling Exponential Exposure Distribution. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Guibert, P., Christophe, C., Urbanova, P., Guérin, G., Blain, S., 2017.
Modelling incomplete and heterogeneous bleaching of mobile grains partially exposed to the
light - Towards a new tool for single grain OSL dating of poorly bleached mortars.
Radiation Measurements 107, 48–57. \doi{10.1016/j.radmeas.2017.10.003}
}
\seealso{
\linkS4class{RLum.Results}, \link{calc_MinDose}, \link{calc_FuchsLang2001},
\link{calc_IEU}, \link{calc_FiniteMixture}
}
\author{
Pierre Guibert, IRAMAT-CRP2A, UMR 5060, Université Bordeaux Montaigne (France)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RLum.Analysis.R
\name{plot_RLum.Analysis}
\alias{plot_RLum.Analysis}
\title{Plot function for an \code{RLum.Analysis} S4 class object}
\usage{
plot_RLum.Analysis(
  object,
  subset = NULL,
  nrows = NULL,
  ncols = NULL,
  abline = NULL,
  combine = FALSE,
  records_max = NULL,
  curve.transformation = "None",
  plot_singlePanels = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}):
S4 object of class \linkS4class{RLum.Analysis}}

\item{subset}{named \link{list} (\emph{optional}):
subsets elements for plotting. The arguments in the named \link{list} will be
directly passed to the function \link{get_RLum}
(e.g., \code{subset = list(curveType = "measured")})}

\item{nrows}{\link{integer} (\emph{optional}):
number of rows in the plot output. If set to \code{NULL} the function tries to
find a reasonable value.}

\item{ncols}{\link{integer} (\emph{optional}):
number of columns in the plot output. If set to \code{NULL} the function tries to
find a reasonable value.}

\item{abline}{\link{list} (\emph{optional}):
allows to add ab-lines to the plot. Argument are provided
in a list and will be forward to the function \link{abline},
e.g., \code{list(v = c(10, 100))} adds two vertical lines add 10 and 100 to all
plots. In contrast \verb{list(v = c(10), v = c(100)} adds a vertical at 10 to
the first and a vertical line at 100 to the 2nd plot.}

\item{combine}{\link{logical} (\emph{with default}):
allows to combine all \linkS4class{RLum.Data.Curve} objects in one single plot.}

\item{records_max}{\link{integer} (\emph{optional}):
limits number of records shown when \code{combine = TRUE} is used. The first and
last curves are always shown, the other curves shown are distributed evenly;
this may result in fewer curves plotted than specified. This argument must
be at least 2 to have an effect.}

\item{curve.transformation}{\link{character} (\emph{with default}):
allows transforming CW-OSL and CW-IRSL curves to pseudo-LM curves via
transformation functions. Allowed values are: \code{CW2pLM}, \code{CW2pLMi},
\code{CW2pHMi} and \code{CW2pPMi}, see details. If set to \code{None} (default), no
transformation is applied.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}):
global par settings are considered, normally this should end in one plot per page}

\item{...}{further arguments and graphical parameters will be passed to
the \code{plot} function.

Supported arguments: \code{main}, \code{mtext}, \code{log}, \code{lwd}, \code{lty} \code{type}, \code{pch}, \code{col},
\code{norm} (see \link{plot_RLum.Data.Curve}), \code{xlim},\code{ylim}, \code{xlab}, \code{ylab}, ...

and for \code{combine = TRUE} also: \code{sub_title}, \code{legend}, \code{legend.text}, \code{legend.pos}
(typical plus 'outside'), \code{legend.col}, \code{smooth}.

All arguments can be provided as \code{vector} or \code{list} to gain in full control
of all plot settings.}
}
\value{
Returns multiple plots.
}
\description{
The function provides a standardised plot output for curve data of an
\linkS4class{RLum.Analysis} object.

The function produces a multiple plot output. A file output is recommended
(e.g., \link{pdf}).

\strong{curve.transformation}

This argument allows transforming continuous wave (CW) curves to pseudo
(linear) modulated curves. For the transformation, the functions of the
package are used. Currently, it is not possible to pass further arguments to
the transformation functions. The argument works only for \code{ltype}
\code{OSL} and \code{IRSL}.

Please note: The curve transformation within this functions works roughly,
i.e. every IRSL or OSL curve is transformed, without considering whether it
is measured with the PMT or not! However, for a fast look it might be
helpful.
}
\note{
Not all arguments available for \link{plot} will be passed and they partly do not behave in the
way you might expect them to work. This function was designed to serve as an overview
plot, if you want to have more control, extract the objects and plot them individually.
}
\section{Function version}{
 0.3.16
}

\examples{

##load data
data(ExampleData.BINfileData, envir = environment())

##convert values for position 1
temp <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)

##(1) plot (combine) TL curves in one plot
plot_RLum.Analysis(
temp,
subset = list(recordType = "TL"),
combine = TRUE,
norm = TRUE,
abline = list(v = c(110))
)

##(2) same as example (1) but using
## the argument smooth = TRUE
plot_RLum.Analysis(
temp,
subset = list(recordType = "TL"),
combine = TRUE,
norm = TRUE,
smooth = TRUE,
abline = list(v = c(110))
)

}
\seealso{
\link{plot}, \link{plot_RLum}, \link{plot_RLum.Data.Curve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. plot_RLum.Analysis(): Plot function for an RLum.Analysis S4 class object. Function version 0.3.16. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{aplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_KDE.R
\name{plot_KDE}
\alias{plot_KDE}
\title{Plot kernel density estimate with statistics}
\usage{
plot_KDE(
  data,
  na.rm = TRUE,
  values.cumulative = TRUE,
  order = TRUE,
  boxplot = TRUE,
  rug = TRUE,
  summary = "",
  summary.pos = "sub",
  summary.method = c("MCM", "weighted", "unweighted"),
  bw = "nrd0",
  ...
)
}
\arguments{
\item{data}{\link{data.frame}, \link{vector} or \linkS4class{RLum.Results} object (\strong{required}):
for \code{data.frame}: either two columns: De (\code{values[,1]}) and De error
(\code{values[,2]}), or one: De (\code{values[,1]}). If a numeric vector or a
single-column data frame is provided, De error is assumed to be 10^-9
for all measurements and error bars are not drawn.
For plotting multiple data sets, these must be provided as
\code{list} (e.g. \code{list(dataset1, dataset2)}).}

\item{na.rm}{\link{logical} (\emph{with default}):
exclude \code{NA} values from the data set prior to any further operation.}

\item{values.cumulative}{\link{logical} (\emph{with default}):
show cumulative individual data.}

\item{order}{\link{logical} (\emph{with default}):
Order data in ascending order.}

\item{boxplot}{\link{logical} (\emph{with default}):
optionally show a boxplot (depicting median as thick central line,
first and third quartile as box limits, whiskers denoting +/- 1.5
interquartile ranges and dots further outliers).}

\item{rug}{\link{logical} (\emph{with default}):
optionally add rug.}

\item{summary}{\link{character} (\emph{with default}):
add statistic measures of centrality and dispersion to the plot. Can be one
or more of several keywords. See details for available keywords.}

\item{summary.pos}{\link{numeric} or \link{character} (\emph{with default}):
optional position coordinates or keyword (e.g. \code{"topright"})
for the statistical summary. Alternatively, the keyword \code{"sub"} may be
specified to place the summary below the plot header. However, this latter
option in only possible if \code{mtext} is not used. In case of coordinate
specification, y-coordinate refers to the right y-axis.}

\item{summary.method}{\link{character} (\emph{with default}):
keyword indicating the method used to calculate the statistic summary.
One of \code{"MCM"} (default), \code{"weighted"} or \code{"unweighted"}.
See \link{calc_Statistics} for details.}

\item{bw}{\link{character}, \link{numeric} (\emph{with default}):
bin-width, chose a numeric value for manual setting.}

\item{...}{further arguments and graphical parameters passed to \link{plot}.}
}
\description{
Plot a kernel density estimate of measurement values in combination with the
actual values and associated error bars in ascending order. If enabled, the
boxplot will show the usual distribution parameters (median as
bold line, box delimited by the first and third quartile, whiskers defined
by the extremes and outliers shown as points) and also the mean and
standard deviation as pale bold line and pale polygon, respectively.

The function allows passing several plot arguments, such as \code{main},
\code{xlab}, \code{cex}. However, as the figure is an overlay of two
separate plots, \code{ylim} must be specified in the order: c(ymin_axis1,
ymax_axis1, ymin_axis2, ymax_axis2) when using the cumulative values plot
option. See examples for some further explanations. For details on the
calculation of the bin-width (parameter \code{bw}) see
\link{density}.

A statistic summary, i.e. a collection of statistic measures of
centrality and dispersion (and further measures) can be added by specifying
one or more of the following keywords:
\itemize{
\item \code{"n"} (number of samples)
\item \code{"mean"} (mean De value)
\item \code{"median"} (median of the De values)
\item \code{"sd.rel"} (relative standard deviation in percent)
\item \code{"sd.abs"} (absolute standard deviation)
\item \code{"se.rel"} (relative standard error)
\item \code{"se.abs"} (absolute standard error)
\item \code{"in.2s"} (percent of samples in 2-sigma range)
\item \code{"kurtosis"} (kurtosis)
\item \code{"skewness"} (skewness)
}

\strong{Note:} the input data for the statistic summary is sent to function
\link{calc_Statistics} depending on the log-option for the z-scale. If
\code{"log.z = TRUE"}, the summary is based on the logarithms of the input
data. If \code{"log.z = FALSE"} the linearly-scaled data is used.

\strong{Note:} \link{calc_Statistics} calculates these statistic
measures in three different ways: \code{unweighted}, \code{weighted} and
\code{MCM-based} (i.e., based on Monte Carlo Methods). By default, the
MCM-based version is used. This can be controlled via the \code{summary.method}
argument.
}
\note{
The plot output is no 'probability density' plot (cf. the discussion
of Berger and Galbraith in Ancient TL; see references)!
}
\section{Function version}{
 3.6.1
}

\examples{

## read example data set
data(ExampleData.DeValues, envir = environment())
ExampleData.DeValues <-
  convert_Second2Gray(ExampleData.DeValues$BT998, c(0.0438,0.0019))

## create plot straightforward
plot_KDE(data = ExampleData.DeValues)

## create plot with logarithmic x-axis
plot_KDE(data = ExampleData.DeValues,
         log = "x")

## create plot with user-defined labels and axes limits
plot_KDE(data = ExampleData.DeValues,
         main = "Dose distribution",
         xlab = "Dose (s)",
         ylab = c("KDE estimate", "Cumulative dose value"),
         xlim = c(100, 250),
         ylim = c(0, 0.08, 0, 30))

## create plot with boxplot option
plot_KDE(data = ExampleData.DeValues,
         boxplot = TRUE)

## create plot with statistical summary below header
plot_KDE(data = ExampleData.DeValues,
         summary = c("n", "median", "skewness", "in.2s"))

## create plot with statistical summary as legend
plot_KDE(data = ExampleData.DeValues,
         summary = c("n", "mean", "sd.rel", "se.abs"),
         summary.pos = "topleft")

## split data set into sub-groups, one is manipulated, and merge again
data.1 <- ExampleData.DeValues[1:15,]
data.2 <- ExampleData.DeValues[16:25,] * 1.3
data.3 <- list(data.1, data.2)

## create plot with two subsets straightforward
plot_KDE(data = data.3)

## create plot with two subsets and summary legend at user coordinates
plot_KDE(data = data.3,
         summary = c("n", "median", "skewness"),
         summary.pos = c(110, 0.07),
         col = c("blue", "orange"))

## example of how to use the numerical output of the function
## return plot output to draw a thicker KDE line
KDE_out <- plot_KDE(data = ExampleData.DeValues)

}
\seealso{
\link{density}, \link{plot}
}
\author{
Michael Dietze, GFZ Potsdam (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Dietze, M., Kreutzer, S., 2026. plot_KDE(): Plot kernel density estimate with statistics. Function version 3.6.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_R2BIN.R
\name{write_R2BIN}
\alias{write_R2BIN}
\title{Export Risoe.BINfileData into Risø BIN/BINX-file}
\usage{
write_R2BIN(
  object,
  file,
  version,
  compatibility.mode = FALSE,
  verbose = TRUE,
  txtProgressBar = TRUE
)
}
\arguments{
\item{object}{\linkS4class{Risoe.BINfileData} (\strong{required}):
input object to be stored in a bin file.}

\item{file}{\link{character} (\strong{required}):
file name and path of the output file
\itemize{
\item \verb{[WIN]}: \code{write_R2BIN(object, "C:/Desktop/test.bin")}
\item \verb{[MAC/LINUX]}: \code{write_R2BIN("/User/test/Desktop/test.bin")}
}}

\item{version}{\link{character} (\emph{optional}):
version number for the output file. If no value is provided, the highest
version number from the \linkS4class{Risoe.BINfileData} is taken automatically.

\strong{Note:}
This argument can be used to convert BIN-file versions.}

\item{compatibility.mode}{\link{logical} (\emph{with default}):
this option recalculates the position values if necessary and set the max.
value to 48. The old position number is appended as comment (e.g., 'OP: 70).
This option accounts for potential compatibility problems with the Analyst software.
It further limits the maximum number of points per curve to 9,999. If a curve contains more
data the curve data get binned using the smallest possible bin width.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar. Ignored if \code{verbose = FALSE}.}
}
\value{
Write a binary file and returns the name and path of the file as \link{character}.
}
\description{
Exports a \linkS4class{Risoe.BINfileData} object in a \verb{*.bin} or \verb{*.binx} file that can be
opened by the Analyst software or other Risø software.
}
\details{
The structure of the exported binary data follows the data structure
published in the Appendices of the \emph{Analyst} manual p. 42.

If
\code{LTYPE}, \code{DTYPE} and \code{LIGHTSOURCE} are not of type
\link{character}, no transformation into numeric values is done.
}
\note{
The function just roughly checks the data structures. The validity of
the output data depends on the user.

The validity of the file path is not further checked. BIN-file conversions
using the argument \code{version} may be a lossy conversion, depending on the
chosen input and output data (e.g., conversion from version 08 to 07 to 06 to 05 to 04 or 03).

\strong{Warning}

Although the coding was done carefully, it seems that the BIN/BINX-files
produced by Risø DA 15/20 TL/OSL readers slightly differ on the byte level.
No obvious differences are observed in the METADATA, however, the
BIN/BINX-file may not fully compatible, at least not similar to the ones
directly produced by the Risø readers!

ROI definitions (introduced in BIN-file version 8) are not supported!
There are furthermore ignored by the function \link{read_BIN2R}.
}
\section{Function version}{
 0.5.4
}

\examples{
##load example dataset
file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
temp <- read_BIN2R(file)

##create temporary file path
##(for usage replace by own path)
temp_file <- tempfile(pattern = "output", fileext = ".binx")

##export to temporary file path
write_R2BIN(temp, file = temp_file)

} 

\section{How to cite}{
Kreutzer, S., 2026. write_R2BIN(): Export Risoe.BINfileData into Risø BIN/BINX-file. Function version 0.5.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
DTU Nutech, 2016. The Sequence Editor, Users Manual, February, 2016.
\url{https://www.fysik.dtu.dk}
}
\seealso{
\link{read_BIN2R}, \linkS4class{Risoe.BINfileData}, \link{writeBin}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_RLum.Analysis.R
\name{merge_RLum.Analysis}
\alias{merge_RLum.Analysis}
\title{Merge function for RLum.Analysis S4 class objects}
\usage{
merge_RLum.Analysis(objects)
}
\arguments{
\item{objects}{\link{list} of \linkS4class{RLum.Analysis} (\strong{required}):
list of S4 objects of class \code{RLum.Analysis}. Furthermore other objects of
class \linkS4class{RLum} can be added, see details.}
}
\value{
Returns an \linkS4class{RLum.Analysis} object ordered according to the order
provided with the input list.
}
\description{
This function simply allows to merge \linkS4class{RLum.Analysis}
objects. Moreover, other \linkS4class{RLum.Data} objects can be added
to an existing \linkS4class{RLum.Analysis} object. Supported objects
to be added are: \linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Spectrum} and
\linkS4class{RLum.Data.Image}.
}
\note{
The information for the slot 'protocol' is taken from the first
\linkS4class{RLum.Analysis} object in the input list. Therefore at
least one object of type \linkS4class{RLum.Analysis} has to be provided.
}
\section{Function version}{
 0.2.1
}

\examples{

##merge different RLum objects from the example data
data(ExampleData.RLum.Analysis, envir = environment())
data(ExampleData.BINfileData, envir = environment())

object <- Risoe.BINfileData2RLum.Analysis(CWOSL.SAR.Data, pos=1)
curve <- get_RLum(object)[[2]]

temp.merged <- merge_RLum.Analysis(list(curve, IRSAR.RF.Data, IRSAR.RF.Data))

}
\seealso{
\link{merge_RLum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Spectrum},
\linkS4class{RLum.Data.Image}, \linkS4class{RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. merge_RLum.Analysis(): Merge function for RLum.Analysis S4 class objects. Function version 0.2.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{internal}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_FuchsLang2001.R
\name{calc_FuchsLang2001}
\alias{calc_FuchsLang2001}
\title{Apply the model after Fuchs & Lang (2001) to a given De distribution}
\usage{
calc_FuchsLang2001(data, cvThreshold = 5, startDeValue = 1, plot = TRUE, ...)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with De \code{(data[,1])} and De error \code{(values[,2])}}

\item{cvThreshold}{\link{numeric} (\emph{with default}):
coefficient of variation in percent, as threshold for the method,
e.g. \code{cvThreshold = 3}. See details
.}

\item{startDeValue}{\link{numeric} (\emph{with default}):
number of the first aliquot that is used for the calculations}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments and graphical parameters passed to \link{plot}}
}
\value{
Returns a plot (\emph{optional}) and terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following elements:

\item{summary}{\link{data.frame} summary of all relevant model results.}
\item{data}{\link{data.frame} original input data}
\item{args}{\link{list} used arguments}
\item{call}{\link{call} the function call}
\item{usedDeValues}{\link{data.frame} containing the used values for the calculation}
}
\description{
This function applies the method according to Fuchs & Lang (2001) for
heterogeneously bleached samples with a given coefficient of variation
threshold.
}
\details{
\strong{Used values}

If the coefficient of variation (\code{c[v]}) of the first
two values is larger than the threshold \code{c[v_threshold]}, the first value is
skipped.  Use the \code{startDeValue} argument to define a start value for
calculation (e.g. 2nd or 3rd value).

\strong{Basic steps of the approach}
\enumerate{
\item Estimate natural relative variation of the sample using a dose recovery test
\item Sort the input values in ascending order
\item Calculate a running mean, starting with the lowermost two values and add values iteratively.
\item Stop if the calculated \code{c[v]} exceeds the specified \code{cvThreshold}
}
}
\note{
Please consider the requirements and the constraints of this method
(see Fuchs & Lang, 2001)
}
\section{Function version}{
 0.4.1
}

\examples{
## load example data
data(ExampleData.DeValues, envir = environment())

## calculate De according to Fuchs & Lang (2001)
temp<- calc_FuchsLang2001(ExampleData.DeValues$BT998, cvThreshold = 5)

} 

\section{How to cite}{
Kreutzer, S., Burow, C., 2026. calc_FuchsLang2001(): Apply the model after Fuchs & Lang (2001) to a given De distribution. Function version 0.4.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Fuchs, M. & Lang, A., 2001. OSL dating of coarse-grain fluvial
quartz using single-aliquot protocols on sediments from NE Peloponnese,
Greece. In: Quaternary Science Reviews 20, 783-787.

Fuchs, M. & Wagner, G.A., 2003. Recognition of insufficient bleaching by
small aliquots of quartz for reconstructing soil erosion in Greece.
Quaternary Science Reviews 22, 1161-1167.
}
\seealso{
\link{plot}, \link{calc_MinDose}, \link{calc_FiniteMixture},
\link{calc_CentralDose}, \link{calc_CommonDose}, \linkS4class{RLum.Results}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}
\keyword{dplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_PSL2R.R
\name{read_PSL2R}
\alias{read_PSL2R}
\title{Import SUERC portable OSL Reader PSL files into R}
\usage{
read_PSL2R(
  file,
  drop_bg = FALSE,
  as_decay_curve = TRUE,
  smooth = FALSE,
  merge = FALSE,
  pattern = "\\\\.psl$",
  verbose = TRUE,
  ...
)
}
\arguments{
\item{file}{\link{character} (\strong{required}):
name of one or multiple PSL files (URLs are supported); it can be the path
to a directory, in which case the function tries to detect and import all
PSL files found in the directory.}

\item{drop_bg}{\link{logical} (\emph{with default}):
\code{TRUE} to automatically remove all non-OSL/IRSL curves.}

\item{as_decay_curve}{\link{logical} (\emph{with default}):
Portable OSL Reader curves are often given as cumulative light sum curves.
Use \code{TRUE} (default) to convert the curves to the more usual decay form.}

\item{smooth}{\link{logical} (\emph{with default}):
\code{TRUE} to apply Tukey's Running Median Smoothing for OSL and IRSL decay curves.
Smoothing is encouraged if you see random signal drops within the decay curves related
to hardware errors.}

\item{merge}{\link{logical} (\emph{with default}):
\code{TRUE} to merge all \linkS4class{RLum.Analysis} objects. Only applicable if multiple
files are imported.}

\item{pattern}{\link{character} (\emph{with default}):
regular expression pattern passed to \link{list.files} to construct a list of
files to read (used only when a path is provided).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{currently not used.}
}
\value{
Returns an S4 \linkS4class{RLum.Analysis} object containing
\linkS4class{RLum.Data.Curve} objects for each curve. Results are
returned as a list when multiple files are processed or \code{file} is a list.
}
\description{
This function provides an import routine for the SUERC portable OSL Reader
PSL format (measurement data and sequence). PSL files are just plain text
and can be viewed with any text editor. Due to the formatting of PSL files,
this import function relies heavily on regular expression to find and
extract all relevant information. See \strong{note}.
}
\note{
Because this function relies heavily on regular expressions to parse
PSL files it is currently only in beta status. If the routine fails to import
a specific PSL file please report to \verb{<christoph.burow@gmx.net>} so the
function can be updated.
}
\section{Function version}{
 0.1.2
}

\examples{

# (1) Import PSL file to R

file <- system.file("extdata", "DorNie_0016.psl", package = "Luminescence")
psl <- read_PSL2R(file, drop_bg = FALSE, as_decay_curve = TRUE, smooth = TRUE, merge = FALSE)
print(str(psl, max.level = 3))
plot(psl, combine = TRUE)

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Curve}
}
\author{
Christoph Burow, University of Cologne (Germany)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
, RLum Developer Team} 

\section{How to cite}{
Burow, C., Kreutzer, S., 2026. read_PSL2R(): Import SUERC portable OSL Reader PSL files into R. Function version 0.1.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_Quote.R
\name{get_Quote}
\alias{get_Quote}
\title{Function to return essential quotes}
\usage{
get_Quote(ID, separated = FALSE)
}
\arguments{
\item{ID}{\link{character} (\emph{optional}): quote ID to be returned.}

\item{separated}{\link{logical} (\emph{with default}): return result in separated form.}
}
\value{
Returns a character with quote and respective (false) author.
}
\description{
This function returns one of the collected essential quotes in the
growing library. If called without any parameters, a random quote is
returned.
}
\section{Function version}{
 0.1.5
}

\examples{

## ask for an arbitrary quote
get_Quote()

}
\author{
Quote credits: Michael Dietze, GFZ Potsdam (Germany), Sebastian Kreutzer, Geography & Earth Science, Aberystwyth University (United Kingdom), Dirk Mittelstraß, TU Dresden (Germany), Jakob Wallinga (Wageningen University, Netherlands)
, RLum Developer Team} 

\section{How to cite}{
Dietze, M., Kreutzer, S., 2023. get_Quote(): Function to return essential quotes. Function version 0.1.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., 2023. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 0.9.21. https://CRAN.R-project.org/package=Luminescence
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_FastRatio.R
\name{calc_FastRatio}
\alias{calc_FastRatio}
\title{Calculate the Fast Ratio for CW-OSL curves}
\usage{
calc_FastRatio(
  object,
  stimulation.power = 30.6,
  wavelength = 470,
  sigmaF = 2.6e-17,
  sigmaM = 4.28e-18,
  Ch_L1 = 1,
  Ch_L2 = NULL,
  Ch_L3 = NULL,
  x = 1,
  x2 = 0.1,
  dead.channels = c(0, 0),
  fitCW.sigma = FALSE,
  fitCW.curve = FALSE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
x, y data of measured values (time and counts).}

\item{stimulation.power}{\link{numeric} (\emph{with default}):
Stimulation power in mW/cm²}

\item{wavelength}{\link{numeric} (\emph{with default}):
Stimulation wavelength in nm}

\item{sigmaF}{\link{numeric} (\emph{with default}):
Photoionisation cross-section (cm²) of the fast component.
Default value after Durcan & Duller (2011).}

\item{sigmaM}{\link{numeric} (\emph{with default}):
Photoionisation cross-section (cm²) of the medium component.
Default value after Durcan & Duller (2011).}

\item{Ch_L1}{\link{numeric} (\emph{with default}):
An integer specifying the channel for L1.}

\item{Ch_L2}{\link{numeric} (\emph{optional}):
An integer specifying the channel for L2.}

\item{Ch_L3}{\link{numeric} (\emph{optional}):
A vector of length 2 with integer values specifying the start and end
channels for L3 (e.g., \code{c(40, 50)}), with the second component greater
than or equal to the first.}

\item{x}{\link{numeric} (\emph{with default}):
Percentage of signal remaining from the fast component.
Used to define the location of L2 and L3 (start).}

\item{x2}{\link{numeric} (\emph{with default}):
Percentage of signal remaining from the medium component.
Used to define the location of L3 (end).}

\item{dead.channels}{\link{numeric} (\emph{with default}):
Vector of length 2 in the form of \code{c(x, y)}.
Channels that do not contain OSL data, i.e. at the start or end of measurement.}

\item{fitCW.sigma}{\link{logical} (\emph{optional}):
fit CW-OSL curve using \link{fit_CWCurve} to calculate \code{sigmaF} and \code{sigmaM} (\strong{experimental}).}

\item{fitCW.curve}{\link{logical} (\emph{optional}):
fit CW-OSL curve using \link{fit_CWCurve} and derive the counts of L2 and L3
from the fitted OSL curve (\strong{experimental}).}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{available options: \code{verbose} (\link{logical}).
Further arguments passed to \link{fit_CWCurve}.}
}
\value{
Returns a plot (\emph{optional}) and an S4 object of type \linkS4class{RLum.Results}.
The slot \code{data} contains a \link{list} with the following elements:

\item{summary}{\link{data.frame} summary of all relevant results}
\item{data}{the original input data}
\item{fit}{\linkS4class{RLum.Results} object if either \code{fitCW.sigma} or \code{fitCW.curve} is \code{TRUE}}
\item{args}{\link{list} of used arguments}
\item{call}{\link{call} the function call}
}
\description{
Function to calculate the fast ratio of quartz CW-OSL single grain or single
aliquot curves after Durcan & Duller (2011).

This function follows the equations of Durcan & Duller (2011). The energy
required to reduce the fast and medium quartz OSL components to \code{x} and
\code{x2} \% respectively using eq. 3 to determine channels L2 and L3 (start
and end). The fast ratio is then calculated from: \eqn{(L1-L3)/(L2-L3)}.
}
\section{Function version}{
 0.1.1
}

\examples{
# load example CW-OSL curve
data("ExampleData.CW_OSL_Curve")

# calculate the fast ratio w/o further adjustments
res <- calc_FastRatio(ExampleData.CW_OSL_Curve)

# show the summary table
get_RLum(res)

} 

\section{How to cite}{
King, G.E., Durcan, J., Burow, C., 2026. calc_FastRatio(): Calculate the Fast Ratio for CW-OSL curves. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Durcan, J.A. & Duller, G.A.T., 2011. The fast ratio: A rapid measure for testing
the dominance of the fast component in the initial OSL signal from quartz.
Radiation Measurements 46, 1065-1072.

Madsen, A.T., Duller, G.A.T., Donnelly, J.P., Roberts, H.M. & Wintle, A.G., 2009.
A chronology of hurricane landfalls at Little Sippewissett Marsh, Massachusetts, USA,
using optical dating. Geomorphology 109, 36-45.

\strong{Further reading}

Steffen, D., Preusser, F. & Schlunegger, 2009. OSL quartz age underestimation
due to unstable signal components. Quaternary Geochronology 4, 353-362.
}
\seealso{
\link{fit_CWCurve}, \link{get_RLum}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}, \linkS4class{RLum.Data.Curve}
}
\author{
Georgina E. King, University of Bern (Switzerland) \cr
Julie A. Durcan, University of Oxford (United Kingdom) \cr
Christoph Burow, University of Cologne (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\docType{data}
\name{ExampleData}
\alias{ExampleData}
\title{All examples}
\description{
These examples are fully documented:

\itemize{
\item \link{ExampleData.CobbleData}
\item \link{ExampleData.DeValues}
\item \link{ExampleData.Fading}
\item \link{ExampleData.portableOSL}
\item \link{ExampleData.RLum.Data.Image}
\item \link{ExampleData.ScaleGammaDose}
\item \link{ExampleData.SurfaceExposure}
\item \link{ExampleData.TR_OSL}
}

The following ones are also available, but are visible only via the
internal documentation, as they correspond to unpublished or synthetic data:

\itemize{
\item \link{ExampleData.Al2O3C}
\item \link{ExampleData.BINfileData}
\item \link{ExampleData.CW_OSL_Curve}
\item \link{ExampleData.FittingLM}
\item \link{ExampleData.LxTxData}
\item \link{ExampleData.LxTxOSLData}
\item \link{ExampleData.MortarData}
\item \link{ExampleData.RF70Curves}
\item \link{ExampleData.RLum.Analysis}
\item \link{ExampleData.XSYG}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RLum.Data.Curve.R
\name{plot_RLum.Data.Curve}
\alias{plot_RLum.Data.Curve}
\title{Plot function for an RLum.Data.Curve S4 class object}
\usage{
plot_RLum.Data.Curve(
  object,
  par.local = TRUE,
  norm = FALSE,
  smooth = FALSE,
  auto_scale = FALSE,
  interactive = FALSE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve} (\strong{required}):
S4 object of class \linkS4class{RLum.Data.Curve}}

\item{par.local}{\link{logical} (\emph{with default}):
use local graphical parameters for plotting, e.g. the plot is shown in one
column and one row. If \code{par.local = FALSE}, global parameters are inherited.}

\item{norm}{\link{logical} \link{character} (\emph{with default}): whether curve
normalisation should occur (\code{FALSE} by default). Alternatively, the function
offers modes \code{"max"} (used with \code{TRUE}), \code{"last"} and \code{"huot"}, see details.}

\item{smooth}{\link{logical} (\emph{with default}):
provides automatic curve smoothing based on the internal function \code{.smoothing}}

\item{auto_scale}{\link{logical} (\emph{with default}): if activated, auto scales \code{xlim} or \code{ylim}
to the extent of the other. If both are set, the auto-scaling is skipped.}

\item{interactive}{\link{logical} (\emph{with default}): enables/disables interactive
plotting mode using \link[plotly:plot_ly]{plotly::plot_ly}}

\item{...}{further arguments and graphical parameters that will be passed
to \link[graphics:plot.default]{graphics::plot.default} and \link[graphics:par]{graphics::par}}
}
\value{
Returns a plot.
}
\description{
The function provides a standardised plot output for curve data of an
\linkS4class{RLum.Data.Curve} S4-class object.
}
\details{
Only single curve data can be plotted with this function. Arguments
according to \link{plot}.

\strong{Curve normalisation}

The argument \code{norm} normalises all count values. To date the following
options are supported:

\code{norm = TRUE} or \code{norm = "max"}: Curve values are normalised to the highest
count value in the curve

\code{norm = "min"}: Curve values are normalised to the smallest count value
in the curve

\code{norm = "first"}: Curve values are normalised to the very first count value

\code{norm = "last"}: Curve values are normalised to the last count value
(this can be useful in particular for radiofluorescence curves)

\code{norm = "huot"}: Curve values are normalised as suggested by Sébastien Huot
via GitHub:
\deqn{
y = (observed - median(background)) / (\max(observed) - median(background))
}

The background of the curve is defined as the last 20\% of the count values
of a curve.

\code{norm = "intensity"}: Curve values are normalised to the channel length.

\code{norm = 2.2}: Curve values are normalised to a positive number (e.g., 2.2).
}
\note{
Not all arguments of \link{plot} will be passed!
}
\section{Function version}{
 0.4.0
}

\examples{

##plot curve data

#load Example data
data(ExampleData.CW_OSL_Curve, envir = environment())

#transform data.frame to RLum.Data.Curve object
temp <- as(ExampleData.CW_OSL_Curve, "RLum.Data.Curve")

#plot RLum.Data.Curve object
plot_RLum.Data.Curve(temp)

}
\seealso{
\link{plot}, \link{plot_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. plot_RLum.Data.Curve(): Plot function for an RLum.Data.Curve S4 class object. Function version 0.4.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{aplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.RF70Curves}
\alias{ExampleData.RF70Curves}
\alias{RF70Curves}
\title{RF70 Example data as \linkS4class{RLum.Analysis} objects}
\format{
\code{BDX16648}: \linkS4class{RLum.Analysis} with two IR-RF curves as \linkS4class{RLum.Data.Curve} objects
}
\source{
\strong{RF70Curves}

\tabular{ll}{
Lab: \tab IRAMAT-CRP2A, CNRS-Université Bordeaux Montaigne, Bordeaux, France\cr
Lab-Code: \tab BDX16648\cr
Location: \tab Médoc, France (details see reference)\cr
Material: \tab K-feldspar (100-200 \eqn{\mu}m)\cr
Reference: \tab Kreutzer et al. (2018)
}
}
\description{
Collection two curves from Kreutzer et al. (2018) measured using the RF70 protocol
(Frouin et al., 2017) as \linkS4class{RLum.Analysis} objects.
}
\section{Version}{
 0.1.0
}

\examples{

##load data
data(ExampleData.RF70Curves, envir = environment())

##plot data
plot_RLum(RF70Curves)

}
\references{
Frouin, M., Huot, S., Kreutzer, S., Lahaye, C., Lamothe, M., Philippe, A., Mercier, N., 2017. An improved
radiofluorescence single-aliquot regenerative dose protocol for K-feldspars. Quaternary Geochronology 38, 13–24.
\doi{10.1016/j.quageo.2016.11.004}

Kreutzer, S., Duval, M., Bartz, M., Bertran, P., Bosq, M., Eynaud, F., Verdin, F., Mercier, N., 2018.
Deciphering long-term coastal dynamics using IR-RF and ESR dating: A case study from Médoc,
south-West France. Quaternary Geochronology 48, 108–120. \doi{10.1016/j.quageo.2018.09.005}
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_EmissionSpectra.R
\name{fit_EmissionSpectra}
\alias{fit_EmissionSpectra}
\title{Luminescence Emission Spectra Deconvolution}
\usage{
fit_EmissionSpectra(
  object,
  frame = NULL,
  n_components = NULL,
  start_parameters = NULL,
  sub_negative = 0,
  input_scale = NULL,
  method_control = list(),
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Spectrum}, \link{matrix} (\strong{required}): input
object. Please note that an energy spectrum is expected}

\item{frame}{\link{integer} (\emph{optional}): number of the frame to be analysed. If
\code{NULL}, all available frames are analysed.}

\item{n_components}{\link{integer} (\emph{optional}): maximum number of components
desired: the number of component actually fitted may be smaller than this.
Can be combined with other parameters.}

\item{start_parameters}{\link{numeric} (\emph{optional}): allows to provide own start parameters for a
semi-automated procedure. Parameters need to be provided in eV. Every value provided replaces a
value from the automated peak finding algorithm (in ascending order).}

\item{sub_negative}{\link{numeric} (\emph{with default}): substitute negative values in the input object
by the number provided here (default: \code{0}). Can be set to \code{NULL}, i.e. negative values are kept.}

\item{input_scale}{\link{character} (\emph{optional}): defines whether your x-values
are expressed as wavelength or energy values. Allowed values are \code{"wavelength"},
\code{"energy"} or \code{NULL}, in which case the function tries to guess the input
automatically.}

\item{method_control}{\link{list} (\emph{optional}): options to control the fit method
and the output produced, see details.}

\item{verbose}{\link{logical} (\emph{with default}): enable/disable output to the
terminal.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{...}{further arguments to be passed to control the plot output
(supported: \code{main}, \code{xlab}, \code{ylab}, \code{xlim}, \code{ylim}, \code{log}, \code{mtext}, \code{legend} (\code{TRUE} or \code{FALSE}),
\code{legend.text}, \code{legend.pos})}
}
\value{
-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$data} \tab \code{matrix} \tab the final fit matrix \cr
\verb{$fit} \tab \code{nls} \tab the fit object returned by \link[minpack.lm:nls.lm]{minpack.lm::nls.lm} \cr
\verb{$fit_info} \tab \code{list} \tab a few additional parameters that can be used
to assess the quality of the fit \cr
\verb{$df_plot} \tab \code{list} \tab values of all curves in the plot for each
frame analysed (only if \code{method_control$export.plot.data = TRUE})
}

\strong{slot:} \strong{\verb{@info}}

The original function call

---------------------------------\cr
\verb{[ TERMINAL OUTPUT ]}   \cr
---------------------------------\cr

The terminal output provides brief information on the
deconvolution process and the obtained results.
Terminal output is only shown when \code{verbose = TRUE}.

---------------------------\cr
\verb{[ PLOT OUTPUT ]}      \cr
---------------------------\cr

The function returns a plot showing the raw signal with the
detected components. If the fitting failed, a basic plot is returned
showing the raw data and indicating the peaks detected for the start
parameter estimation. The grey band in the residual plot indicates the
10\% deviation from 0 (means no residual).
}
\description{
This function performs a luminescence spectra deconvolution on
\linkS4class{RLum.Data.Spectrum} and \link{matrix} objects on an \strong{energy scale}.
The function is optimised for emission spectra typically obtained in the
context of TL, OSL and RF  measurements detected between 200 and 1000 nm.
The function is not designed to deconvolve TL curves (counts against
temperature; no wavelength scale). If you are interested in such analysis,
please check, e.g., package \code{'tgcd'}.
}
\details{
\strong{Used equation}

The emission spectra (on an energy scale) can be best described as the sum of multiple
Gaussian components:

\deqn{
y = \sum_i \frac{C_i}{\sigma_{i} \sqrt{2 \pi}} \exp(-1/2 ((x - \mu_{i})/\sigma_{i})^2)
}

with the parameters \eqn{\sigma} (peak width) and \eqn{\mu} (peak centre) and \eqn{C}
(scaling factor).

\strong{Start parameter estimation and fitting algorithm}

The spectrum deconvolution consists of the following steps:
\enumerate{
\item Peak finding \cr
\item Start parameter estimation \cr
\item Fitting via \link[minpack.lm:nls.lm]{minpack.lm::nls.lm}\cr
}

The peak finding is based on an approach (re-)suggested by Petr Pikal in
November 2005 in the R-help mailing list (https://stat.ethz.ch/pipermail/r-help/2005-November/thread.html).
This goes back to even earlier discussion in 2001 based on Prof Brian Ripley's idea.
It smartly uses the functions \link[stats:embed]{stats::embed} and \link{max.col} to identify peaks positions.
For the use in this context, the algorithm has been further modified to scale on the
input data resolution (cf. source code).\cr

The start parameter estimation uses random sampling from a range of meaningful parameters
and repeats the fitting until 1000 successful fits have been produced or the set \code{max.runs} value
is exceeded.

Currently the best fit is the one with the lowest number for squared residuals, but
other parameters are returned as well. If a series of curves needs to be analysed,
it is recommended to make few trial runs, then fix the number of components and
run  at least 10,000 iterations (parameter \code{method_control = list(max.runs = 10000)}).

\strong{Supported \code{method_control} settings}

\tabular{llll}{
\strong{Parameter} \tab \strong{Type} \tab \strong{Default} \tab \strong{Description}\cr
\code{max.runs} \tab \link{integer} \tab \code{10000} \tab maximum allowed search iterations, if exceed
the searching stops \cr
\code{graining} \tab \link{numeric} \tab \code{15} \tab control over how coarse or fine the spectrum is split into search intervals for the peak finding algorithm \cr
\code{norm} \tab \link{logical} \tab \code{TRUE} \tab normalise data to the highest count value before fitting \cr
\code{export.plot.data} \tab \link{logical} \tab \code{FALSE} \tab enable/disable export
of the values of all curves in the plot for each frame analysed \cr
\code{trace} \tab \link{logical} \tab \code{FALSE} \tab enable/disable the tracing of the minimisation routine
}
}
\section{Function version}{
 0.1.3
}

\examples{

##load example data
data(ExampleData.XSYG, envir = environment())

##subtract background
TL.Spectrum@data <- TL.Spectrum@data[] - TL.Spectrum@data[,15]

results <- fit_EmissionSpectra(
 object = TL.Spectrum,
 frame = 5,
 method_control = list(max.runs = 10)
)

##deconvolution of a TL spectrum
\dontrun{

##load example data

##replace 0 values
results <- fit_EmissionSpectra(
 object = TL.Spectrum,
 frame = 5, main = "TL spectrum"
)

}

}
\seealso{
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Results}, \link{plot_RLum},
\link{convert_Wavelength2Energy}, \link[minpack.lm:nls.lm]{minpack.lm::nls.lm}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. fit_EmissionSpectra(): Luminescence Emission Spectra Deconvolution. Function version 0.1.3. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RLum.Data.Spectrum.R
\name{plot_RLum.Data.Spectrum}
\alias{plot_RLum.Data.Spectrum}
\title{Plot function for an RLum.Data.Spectrum S4 class object}
\usage{
plot_RLum.Data.Spectrum(
  object,
  par.local = TRUE,
  plot.type = c("contour", "persp", "single", "multiple.lines", "image", "transect",
    "interactive"),
  optical.wavelength.colours = TRUE,
  bg.spectrum = NULL,
  bg.channels = NULL,
  bin.rows = 1,
  bin.cols = 1,
  norm = FALSE,
  rug = TRUE,
  limit_counts = NULL,
  xaxis.energy = FALSE,
  legend.text = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Spectrum} or \link{matrix} (\strong{required}):
S4 object of class \linkS4class{RLum.Data.Spectrum} or a \code{matrix} containing count
values of the spectrum.\cr
Please note that in case of a matrix row names and col names are set
automatically if not provided.}

\item{par.local}{\link{logical} (\emph{with default}):
use local graphical parameters for plotting, e.g. the plot is shown in one
column and one row. If \code{par.local = FALSE} global parameters are inherited.}

\item{plot.type}{\link{character} (\emph{with default}):
for a 3D-plot use \code{"persp"} or \code{"interactive"}; for a 2D-plot you can use
\code{"image"}, \code{"contour"}, \code{"single"} or \code{"multiple.lines"} (along the time or
temperature axis) or \code{"transect"} (along the wavelength axis).}

\item{optical.wavelength.colours}{\link{logical} (\emph{with default}):
use optical wavelength colour palette. Note: For this, the spectrum range is
limited: \code{c(350,750)}. Own colours can be set with the argument \code{col}. If
you provide already binned spectra, the colour assignment is likely to be
wrong, since the colour gradients are calculated using the bin number.}

\item{bg.spectrum}{\linkS4class{RLum.Data.Spectrum} or \link{matrix} (\emph{optional}):
spectrum used for the background subtraction. The background spectrum should
be measured using the same setting as the signal spectrum. The argument
\code{bg.channels} controls how the subtraction is performed: if \code{bg.channels}
is not specified or the number of channels is identical between the signal
and background spectra, a channel-wise subtraction is performed; otherwise,
the \emph{arithmetic mean} is calculated and subtracted from the signal.}

\item{bg.channels}{\link{vector} (\emph{optional}):
channels used for background subtraction. If the number of channels is
identical between the signal and background spectra, a channel-wise
subtraction is performed; otherwise this number is used to select channels
for calculating the \emph{arithmetic mean}  If a spectrum is provided via
\code{bg.spectrum}, this argument only works on the background spectrum.

\strong{Note:} Background subtraction is applied prior to channel binning!}

\item{bin.rows}{\link{integer} (\emph{with default}):
allow summing-up wavelength channels (horizontal binning),
e.g. \code{bin.rows = 2} two channels are summed up.
Binning is applied after the background subtraction.}

\item{bin.cols}{\link{integer} (\emph{with default}):
allow summing-up channel counts (vertical binning) for plotting,
e.g. \code{bin.cols = 2} two channels are summed up.
Binning is applied after the background subtraction.}

\item{norm}{\link{logical}, \link{character} (\emph{optional}):
if logical, whether curve normalisation should occur (\code{FALSE} by default);
alternatively, one of the values detailed in \link{plot_RLum.Data.Curve}.
The normalisation is applied after binning.}

\item{rug}{\link{logical} (\emph{with default}):
enable/disable colour rug. Currently only implemented for plot
type \code{multiple.lines} and \code{single}.}

\item{limit_counts}{\link{numeric} (\emph{optional}):
value to limit all count values to this value, i.e. all count values above
this threshold will be replaced by this threshold. This is helpful
especially in case of TL-spectra.}

\item{xaxis.energy}{\link{logical} (\emph{with default}): enable/disable using energy
instead of wavelength on the x-axis. Function \link{convert_Wavelength2Energy}
is used to perform the conversion.

\strong{Note:} Besides being used in setting the axis, with this option the
the spectrum is recalculated in terms of intensity, see details.}

\item{legend.text}{\link{character} (\emph{with default}):
possibility to provide own legend text. This argument is only considered for
plot types providing a legend, e.g. \code{plot.type = "transect"}.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output. If
the plot output is disabled, the \link{matrix} used for the plotting and the
calculated colour values (as attributes) are returned. This way, the
(binned, transformed etc.) output can be used in other functions and
packages, such as plotting with the package \code{'plot3D'}.}

\item{...}{further arguments and graphical parameters that will be passed
to the \code{plot} function.}
}
\value{
Returns a plot and the transformed \code{matrix} used for plotting with some
useful attributes such as \code{colour} and \code{pmat} (the transpose matrix from
\link[graphics:persp]{graphics::persp}).
}
\description{
The function provides a standardised plot output for spectrum data of an
\linkS4class{RLum.Data.Spectrum} class object. The purpose of this function is to
provide easy and straightforward spectra plotting, not a fully customised
access to all plot parameters. If this is wanted, standard R plot
functionality should be used instead.

\strong{Matrix structure} \cr (cf. \linkS4class{RLum.Data.Spectrum})
\itemize{
\item \code{rows} (x-values): wavelengths/channels (\code{xlim}, \code{xlab})
\item \code{columns} (y-values): time/temperature (\code{ylim}, \code{ylab})
\item \code{cells} (z-values): count values (\code{zlim}, \code{zlab})
}

\emph{Note: This nomenclature is valid for all plot types of this function!}

\strong{Nomenclature for value limiting}
\itemize{
\item \code{xlim}: Limits values along the wavelength axis
\item \code{ylim}: Limits values along the time/temperature axis
\item \code{zlim}: Limits values along the count value axis
}

\strong{Details on the plot functions}

Spectrum is visualised as 3D or 2D plot. Both plot types are based on
internal R plot functions.

\strong{\code{plot.type = "persp"}}

Arguments that will be passed to \link[graphics:persp]{graphics::persp}:
\itemize{
\item \code{shade}: default is \code{0.4}
\item \code{phi}: default is \code{15}
\item \code{theta}: default is \code{-30}
\item \code{lphi}: default is \code{15}
\item \code{ltheta}: default is \code{-30}
\item \code{expand}: default is \code{1}
\item \code{axes}: default is \code{TRUE}
\item \code{box}: default is \code{TRUE}; accepts \code{"alternate"} for a custom plot design
\item \code{ticktype}: default is \code{detailed}, \code{r}: default is \code{10}
}

\strong{Note:} Further parameters can be adjusted via \code{par}. For example
to set the background transparent and reduce the thickness of the lines use:
\code{par(bg = NA, lwd = 0.7)} before the function call.

\strong{\code{plot.type = "single"}}

Per frame a single curve is returned. Frames are time or temperature
steps.

-\code{frames}: pick the frames to be plotted (depends on the binning!). Check without
this setting before plotting.

\strong{\code{plot.type = "multiple.lines"}}

All frames plotted in one frame.

-\code{frames}: pick the frames to be plotted (depends on the binning!). Check without
this setting before plotting.

\strong{\code{plot.type = "image"} or \code{plot.type = "contour"}}

These plot types use the R functions \link[graphics:image]{graphics::image} or \link[graphics:contour]{graphics::contour}.
The advantage is that many plots can be arranged conveniently using standard
R plot functionality. If \code{plot.type = "image"} a contour is added by default,
which can be disabled using the argument \code{contour = FALSE} to add own contour
lines of choice.

\strong{\code{plot.type = "transect"}}

Depending on the selected wavelength/channel range a transect over the
time/temperature (y-axis) will be plotted along the wavelength/channels
(x-axis). If the range contains more than one channel, values (z-values) are
summed up. To select a transect use the \code{xlim} argument, e.g.
\code{xlim = c(300,310)} plot along the summed up count values of channel
300 to 310.

\strong{Further arguments that will be passed (depending on the plot type)}

\code{xlab}, \code{ylab}, \code{zlab}, \code{xlim}, \code{ylim}, \code{zlim}, \code{main}, \code{mtext},
\code{box}, \code{pch}, \code{type} (\code{"single"}, \code{"multiple.lines"}, \code{"interactive"}),
\code{col}, \code{border}, \code{lwd}, \code{bty}, \code{showscale} (\code{"interactive"}, \code{"image"})
\code{contour}, \code{contour.col} (\code{"image"}), \code{labcex} (\code{"image"}, \code{"contour"}),
\code{n_breaks} (\code{"image"}), \code{legend} (\code{TRUE}/\code{FALSE}),
\code{legend.pos} (\code{"image"}), \code{legend.horiz} (\code{TRUE}/\code{FALSE} | \code{"image"})
}
\note{
Not all additional arguments (\code{...}) will be passed similarly!
}
\section{Function version}{
 0.6.13
}

\examples{

##load example data
data(ExampleData.XSYG, envir = environment())

##(1)plot simple spectrum (2D) - image
plot_RLum.Data.Spectrum(
 TL.Spectrum,
 plot.type="image",
 xlim = c(310,750),
 ylim = c(0,300),
 bin.rows=10,
 bin.cols = 1)

##(2) plot spectrum (3D)
plot_RLum.Data.Spectrum(
  TL.Spectrum,
  plot.type="persp",
  xlim = c(310,750),
  ylim = c(0,100),
  bin.rows=10,
  bin.cols = 1)

##(3) plot spectrum on energy axis
##please note the background subtraction
plot_RLum.Data.Spectrum(TL.Spectrum,
plot.type="persp",
ylim = c(0,200),
bin.rows=10,
bg.channels = 10,
bin.cols = 1,
xaxis.energy = TRUE)

##(4) plot multiple lines (2D) - multiple.lines (with ylim)
plot_RLum.Data.Spectrum(
 TL.Spectrum,
 plot.type="multiple.lines",
 xlim = c(310,750),
 ylim = c(0,100),
 bin.rows=10,
 bin.cols = 1)

\dontrun{
 ##(4) interactive plot using the package plotly ("surface")
 plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
 xlim = c(310,750), ylim = c(0,300), bin.rows=10,
 bin.cols = 1)

 ##(5) interactive plot using the package plotly ("contour")
 plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
 xlim = c(310,750), ylim = c(0,300), bin.rows=10,
 bin.cols = 1,
 type = "contour",
 showscale = TRUE)

 ##(6) interactive plot using the package plotly ("heatmap")
 plot_RLum.Data.Spectrum(TL.Spectrum, plot.type="interactive",
 xlim = c(310,750), ylim = c(0,300), bin.rows=10,
 bin.cols = 1,
 type = "heatmap",
 showscale = TRUE)
}

}
\seealso{
\linkS4class{RLum.Data.Spectrum},
\link{convert_Wavelength2Energy}, \link{plot_RLum},
\link[graphics:persp]{graphics::persp}, \link[plotly:plot_ly]{plotly::plot_ly}, \link[graphics:contour]{graphics::contour}, \link[graphics:image]{graphics::image}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. plot_RLum.Data.Spectrum(): Plot function for an RLum.Data.Spectrum S4 class object. Function version 0.6.13. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{aplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_OSLLifeTimes.R
\name{fit_OSLLifeTimes}
\alias{fit_OSLLifeTimes}
\title{Fitting and Deconvolution of OSL Lifetime Components}
\usage{
fit_OSLLifeTimes(
  object,
  tp = 0,
  signal_range = NULL,
  n.components = NULL,
  method_control = list(),
  plot = TRUE,
  plot_simple = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Analysis}, \link{data.frame} or \link{matrix} (\strong{required}):
Input object containing the data to be analysed. All objects can be provided also as list for an automated
processing. Please note: \code{NA} values are automatically removed and the dataset should comprise at least 5 data points (possibly more if \code{n.components} is
set to a value greater than 1)}

\item{tp}{\link{numeric} (\emph{with default}): option to account for the stimulation pulse width. For off-time measurements
the default value is 0. \code{tp} has the same unit as the measurement data, e.g., µs. Please set this parameter
carefully, if it all, otherwise you may heavily bias your fit results.}

\item{signal_range}{\link{numeric} (\emph{optional}):
set a channel range, either by specifying both extremes, e.g.
\code{signal_range = c(2, 100)} considers only channels 2 to 100, or only the
start of the range, e.g. \code{signal_range = 2} considers only channels from
channel 2 onwards. If \code{NULL} (default) all channels are used.}

\item{n.components}{\link{numeric} (\emph{optional}): Fix the number of components. If set the algorithm will try
to fit the number of predefined components. If nothing is set, the algorithm will try to find the best number
of components.}

\item{method_control}{\link{list} (\emph{optional}): Named to allow a more fine control of the fitting process. See details
for allowed options.}

\item{plot}{\link{logical} (\emph{with default}): Enable/disable the plot output.}

\item{plot_simple}{\link{logical} (\emph{with default}): Enable/disable the reduced
plot output. If \code{TRUE}, no
residual plot is shown, however, plot output can be combined using the standard R layout options,
such as \code{par(mfrow = c(2,2))}.}

\item{verbose}{\link{logical} (\emph{with default}): Enable/disable output to the
terminal.}

\item{...}{parameters passed to \link{plot.default} to control the plot output, supported are:
\code{main}, \code{xlab}, \code{ylab}, \code{log}, \code{xlim}, \code{ylim}, \code{col}, \code{lty}, \code{legend.pos}, \code{legend.text}. If the input
object is of type \linkS4class{RLum.Analysis} this arguments can be provided as a \link{list}.}
}
\value{
-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$data} \tab \code{matrix} \tab the final fit matrix \cr
\verb{$start_matrix} \tab \code{matrix} \tab the start matrix used for the fitting \cr
\verb{$total_counts} \tab \code{integer} \tab Photon count sum \cr
\verb{$fit} \tab \code{nls} \tab the fit object returned by \link[minpack.lm:nls.lm]{minpack.lm::nls.lm} \cr
}

\strong{slot:} \strong{\verb{@info}}

The original function call

------------------------\cr
\verb{[ TERMINAL OUTPUT ]}\cr
------------------------\cr

Terminal output is only shown of the argument \code{verbose = TRUE}.

\emph{(1) Start parameter and component adaption}\cr
Trave of the parameter adaptation process

\emph{(2) Fitting results (sorted by ascending tau)}\cr
The fitting results sorted by ascending tau value. Please note
that if you access the \code{nls} fitting object, the values are not sorted.

\emph{(3) Further information}\cr
\itemize{
\item The photon count sum
\item Durbin-Watson residual statistic to assess whether the residuals are correlated, ideally
the residuals should be not correlated at all. Rough measures are: \cr
D = 0: the residuals are systematically correlated \cr
D = 2: the residuals are randomly distributed \cr
D = 4: the residuals are systematically anti-correlated\cr
}

You should be suspicious if D differs largely from 2.

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr

A plot showing the original data and the fit so far possible. The lower plot shows the
residuals of the fit.
}
\description{
Fitting and Deconvolution of OSL Lifetime Components
}
\details{
The function intends to provide an easy access to pulsed optically stimulated luminescence (POSL) data,
in order determine signal lifetimes. The fitting is currently optimised to work with the off-time flank of POSL measurements
only. For the signal deconvolution, a differential evolution optimisation is combined with nonlinear least-square fitting
following the approach by Bluszcz & Adamiec (2006).

\strong{Component deconvolution algorithm}

The component deconvolution consists of two steps:

(1) Adaptation phase

In the adaptation phase the function tries to figure out the optimal and statistically justified
number of signal components following roughly the approach suggested by Bluszcz & Adamiec (2006). In
contrast to their work, for the optimisation by differential evolution here the package \code{'DEoptim'} is used.

The function to be optimized has the form:

\deqn{\chi^2 = \sum(w * (n_i/c - \sum(A_i * \exp(-x/(tau_i + t_p))))^2)}

with \eqn{w = 1} for unweighted regression analysis (\code{method_control = list(weights = FALSE)}) or
\eqn{w = c^2/n_i} for weighted regression analysis. The default values is \code{TRUE}.

\deqn{F = (\Delta\chi^2 / 2) / (\chi^2/(N - 2*m - 2))}

(2) Final fitting

\strong{\code{method_control}}

\tabular{lll}{
\strong{Parameter} \tab \strong{Type} \tab \strong{Description}\cr
\code{p} \tab \link{numeric} \tab controls the probability for the F statistic reference values. For a significance level of 5 \% a value of 0.95 (the default) should be added, for 1 \%, a value of 0.99 is sufficient: 1 > p > 0 (cf. \link[stats:Fdist]{stats::FDist})\cr
\code{seed} \tab \link{numeric} \tab set the seed for the random number generator, provide a value here to get reproducible results \cr
\code{DEoptim.trace} \tab \link{logical} \tab enable/disable the tracing of the differential evolution (cf. \link[DEoptim:DEoptim.control]{DEoptim::DEoptim.control}) \cr
\code{DEoptim.itermax} \tab \link{logical} \tab control the number of the allowed generations (cf. \link[DEoptim:DEoptim.control]{DEoptim::DEoptim.control}) \cr
\code{weights} \tab \link{logical} \tab enable/disable the weighting for the start parameter estimation and fitting (see equations above).
The default values is \code{TRUE} \cr
\code{nlsLM.trace} \tab \link{logical} \tab enable/disable trace mode for the nls fitting (\link[minpack.lm:nlsLM]{minpack.lm::nlsLM}), can be used to identify convergence problems, default is \code{FALSE} \cr
\code{nlsLM.upper} \tab \link{logical} \tab enable/disable upper parameter boundary, default is \code{TRUE} \cr
\code{nlsLM.lower} \tab \link{logical} \tab enable/disable lower parameter boundary, default is \code{TRUE}
}
}
\section{Function version}{
 0.1.6
}

\examples{

##load example data
data(ExampleData.TR_OSL, envir = environment())

##fit lifetimes (short run)
fit_OSLLifeTimes(
 object = ExampleData.TR_OSL,
 n.components = 1)

##long example
\dontrun{
fit_OSLLifeTimes(
object = ExampleData.TR_OSL)
}

} 

\section{How to cite}{
Kreutzer, S., Schmidt, C., 2026. fit_OSLLifeTimes(): Fitting and Deconvolution of OSL Lifetime Components. Function version 0.1.6. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Bluszcz, A., Adamiec, G., 2006. Application of differential evolution to fitting OSL decay curves.
Radiation Measurements 41, 886-891. \doi{10.1016/j.radmeas.2006.05.016}\cr

Durbin, J., Watson, G.S., 1950. Testing for Serial Correlation in Least Squares Regression: I.
Biometrika 37, 409-21. doi:10.2307/2332391

\strong{Further reading}

Hughes, I., Hase, T., 2010. Measurements and Their Uncertainties. Oxford University Press.

Storn, R., Price, K., 1997. Differential Evolution –
A Simple and Efficient Heuristic for Global Optimization over Continuous Spaces.
Journal of Global Optimization 11, 341–359.
}
\seealso{
\link[minpack.lm:nls.lm]{minpack.lm::nls.lm}, \link[DEoptim:DEoptim]{DEoptim::DEoptim}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Christoph Schmidt, University of Bayreuth (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_RLum.Data.Image.R
\name{plot_RLum.Data.Image}
\alias{plot_RLum.Data.Image}
\title{Plot function for an \code{RLum.Data.Image} S4 class object}
\usage{
plot_RLum.Data.Image(
  object,
  frames = NULL,
  par.local = TRUE,
  plot.type = c("plot.raster", "contour"),
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Data.Image} (\strong{required}): S4
object of class \linkS4class{RLum.Data.Image}}

\item{frames}{\link{numeric} (\emph{optional}): sets the frames to be set, by default all
frames are plotted. Can be sequence of numbers, as long as the frame number is valid.}

\item{par.local}{\link{logical} (\emph{with default}): use local graphical
parameters for plotting, e.g. the plot is shown in one column and one row.
If \code{par.local = FALSE} global parameters are inherited.}

\item{plot.type}{\link{character} (\emph{with default}): plot types.
Supported types are \code{plot.raster}, \code{contour}}

\item{...}{further arguments and graphical parameters that will be passed
to the specific plot functions. Standard supported parameters are \code{xlim},
\code{ylim}, \code{zlim}, \code{xlab}, \code{ylab}, \code{main}, \code{mtext}, \code{legend} (\code{TRUE} or \code{FALSE}),
\code{col}, \code{cex}, \code{axes} (\code{TRUE} or \code{FALSE}), \code{zlim_image} (adjust the z-scale
over different images), \code{stretch}, \code{digits}, scientific (\code{TRUE} or \code{FALSE}).}
}
\value{
Returns a plot
}
\description{
The function provides very basic plot functionality for image data of an
\linkS4class{RLum.Data.Image} object. For more sophisticated plotting it is recommended
to use other very powerful packages for image processing.

\strong{Details on the plot functions}

Supported plot types:

\strong{\code{plot.type = "plot.raster"}}

Uses the standard plot function of R \link[graphics:image]{graphics::image}. If wanted, the image
is enhanced, using the argument \code{stretch}. Possible values are \code{hist}, \code{lin}, and
\code{NULL}. The latter does nothing. The argument \code{useRaster = TRUE} is used by default, but
can be set to \code{FALSE}.

\strong{\code{plot.type = "contour"}}

This uses the function \link[graphics:contour]{graphics::contour}
}
\note{
The axes limitations (\code{xlim}, \code{zlim}, \code{zlim}) work directly on the object,
so that regardless of the chosen limits the image parameters can be adjusted for
best visibility. However, in particular for z-scale limitations this is not always
wanted, please use \code{zlim_image} to maintain a particular value range over a
series of images.
}
\section{Function version}{
 0.2.2
}

\examples{

##load data
data(ExampleData.RLum.Data.Image, envir = environment())

##plot data
plot_RLum.Data.Image(ExampleData.RLum.Data.Image)

}
\seealso{
\linkS4class{RLum.Data.Image}, \link{plot}, \link{plot_RLum},
\link[graphics:image]{graphics::image}, \link[graphics:contour]{graphics::contour}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. plot_RLum.Data.Image(): Plot function for an RLum.Data.Image S4 class object. Function version 0.2.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{aplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R
\name{set_Risoe.BINfileData}
\alias{set_Risoe.BINfileData}
\title{General setter function for Risoe.BINfileData objects}
\usage{
set_Risoe.BINfileData(
  METADATA = data.frame(),
  DATA = list(),
  .RESERVED = list()
)
}
\arguments{
\item{METADATA}{x}

\item{DATA}{x}

\item{.RESERVED}{x}
}
\value{
A \linkS4class{Risoe.BINfileData} object.
}
\description{
The function provides a generalised access point for specific
\linkS4class{Risoe.BINfileData} objects. Depending on the input object, the
corresponding function will be selected.
Allowed arguments can be found in the documentations of the corresponding
\linkS4class{Risoe.BINfileData} class.
}
\section{Function version}{
 0.1
}

\seealso{
\linkS4class{Risoe.BINfileData}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. set_Risoe.BINfileData(): General setter function for Risoe.BINfileData objects. Function version 0.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_TLLxTxRatio.R
\name{calc_TLLxTxRatio}
\alias{calc_TLLxTxRatio}
\title{Calculate the Lx/Tx ratio for a given set of TL curves -beta version-}
\usage{
calc_TLLxTxRatio(
  Lx.data.signal,
  Lx.data.background = NULL,
  Tx.data.signal = NULL,
  Tx.data.background = NULL,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  ...
)
}
\arguments{
\item{Lx.data.signal}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
TL data (x = temperature, y = counts) (TL signal).}

\item{Lx.data.background}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\emph{optional}):
TL data (x = temperature, y = counts).
If no data are provided no background subtraction is performed.}

\item{Tx.data.signal}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\strong{required}):
TL data (x = temperature, y = counts) (TL test signal).}

\item{Tx.data.background}{\linkS4class{RLum.Data.Curve} or \link{data.frame} (\emph{optional}):
TL data (x = temperature, y = counts).
If no data are provided no background subtraction is performed.}

\item{signal_integral}{\link{integer} (\strong{required}):
vector of channels for the signal integral.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given temperature range are selected.}

\item{...}{currently not used.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.
Slot \code{data} contains a \link{list} with the following structure:

\if{html}{\out{<div class="sourceCode">}}\preformatted{$ LxTx.table
.. $ LnLx
.. $ LnLx.BG
.. $ TnTx
.. $ TnTx.BG
.. $ Net_LnLx
.. $ Net_LnLx.Error
.. $ SN_RATIO_LnLx
.. $ SN_RATIO_TnTx
.. $ LxTx
.. $ LxTx.Error
}\if{html}{\out{</div>}}
}
\description{
Calculate Lx/Tx ratio for a given set of TL curves.
}
\details{
\strong{Uncertainty estimation}

The standard errors are calculated using the following generalised equation:

\deqn{SE_{signal} = abs(Signal_{net} * BG_f /BG_{signal})}

where \eqn{BG_f} is a term estimated by calculating the standard deviation of the sum of
the \eqn{L_x} background counts and the sum of the \eqn{T_x} background counts. However,
if both signals are similar the error becomes zero.
}
\note{
\strong{This function has still BETA status!} Please further note that a similar
background for both curves results in a zero error and is therefore set to \code{NA}.
}
\section{Function version}{
 0.3.7
}

\examples{

##load package example data
data(ExampleData.BINfileData, envir = environment())

##convert Risoe.BINfileData into a curve object
temp <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos = 3)

Lx.data.signal <- get_RLum(temp, record.id=1)
Lx.data.background <- get_RLum(temp, record.id=2)
Tx.data.signal <- get_RLum(temp, record.id=3)
Tx.data.background <- get_RLum(temp, record.id=4)
signal_integral <- 210:230

output <- calc_TLLxTxRatio(
 Lx.data.signal,
 Lx.data.background,
 Tx.data.signal,
 Tx.data.background,
 signal_integral)
get_RLum(output)

}
\seealso{
\linkS4class{RLum.Results}, \link{analyse_SAR.TL}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Christoph Schmidt, University of Bayreuth (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Schmidt, C., 2026. calc_TLLxTxRatio(): Calculate the Lx/Tx ratio for a given set of TL curves -beta version-. Function version 0.3.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_Data.R
\name{import_Data}
\alias{import_Data}
\title{Import Luminescence Data into R}
\usage{
import_Data(file, ..., fastForward = TRUE, verbose = FALSE)
}
\arguments{
\item{file}{\link{character} (\strong{required}): file to be imported, can be a \link{list} or a \link{character} vector}

\item{...}{arguments to be further passed down to supported functions (please check the functions
to determine the correct arguments)}

\item{fastForward}{\link{logical} (\emph{with default}): option to create \linkS4class{RLum} objects
during import or a \link{list} of such objects}

\item{verbose}{\link{logical} (\emph{with default}): enable/disable output to the
terminal.}
}
\value{
Always returns a \link{list}; empty or filled with \linkS4class{RLum.Analysis} objects
}
\description{
Convenience wrapper function to provide a quicker and more standardised way of
reading data into R by looping through all in the package available data import
functions starting with \code{read_}. Import data types can be mixed.
}
\section{Function version}{
 0.1.6
}

\examples{

## import BINX/BIN
file <- system.file("extdata/BINfile_V8.binx", package = "Luminescence")
temp <- import_Data(file)

## RF data
file <- system.file("extdata", "RF_file.rf", package = "Luminescence")
temp <- import_Data(file)

}
\seealso{
\link{read_BIN2R}, \link{read_BINXLOG2R}, \link{read_Daybreak2R}, \link{read_HeliosOSL2R}, \link{read_PSL2R}, \link{read_RF2R}, \link{read_SPE2R}, \link{read_TIFF2R}, \link{read_XSYG2R}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. import_Data(): Import Luminescence Data into R. Function version 0.1.6. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_FadingCorr.R
\name{calc_FadingCorr}
\alias{calc_FadingCorr}
\title{Fading Correction after Huntley & Lamothe (2001)}
\usage{
calc_FadingCorr(
  age.faded,
  g_value,
  tc,
  tc.g_value = tc,
  n.MC = 10000,
  seed = NULL,
  interval = c(0.01, 500),
  txtProgressBar = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{age.faded}{\link{numeric} \link{vector} (\strong{required}):
vector of length 2 containing the uncorrected age and the error in ka
(see example).}

\item{g_value}{\link{vector} or \linkS4class{RLum.Results} (\strong{required}):
either a vector of length 2 containing the g-value and error obtained from
separate fading measurements (see example), or an \linkS4class{RLum.Results}
object produced by \link{analyse_FadingMeasurement}. If the latter, the \code{tc}
argument is set automatically.}

\item{tc}{\link{numeric} (\strong{required}):
time in seconds between irradiation and the prompt measurement (cf. Huntley & Lamothe 2001).
The argument is ignored when \code{g_value} is an \linkS4class{RLum.Results} object.}

\item{tc.g_value}{\link{numeric} (\emph{with default}):
time in seconds between irradiation and the prompt measurement used in the
estimation of the g-value. If the g-value was normalised, the normalisation
time (in seconds) should be given, e.g., for a g-value normalised to 2 days,
the value 172800 should be used. If nothing is provided the time is set to
\code{tc}, which is usual case for g-values obtained using the
SAR method and \eqn{g}-values that have been not normalised to 2 days.}

\item{n.MC}{\link{integer} (\emph{with default}):
number of Monte Carlo simulation runs for error estimation.
If \code{n.MC = 'auto'} is used the function tries to find a 'stable' error for
the age. See details for further information.
\strong{Note:} This may take a while!}

\item{seed}{\link{integer} (\emph{optional}):
sets the seed for the random number generator in R using \link{set.seed}}

\item{interval}{\link{numeric} (\emph{with default}):
a vector containing the end-points (age interval) of the interval to be searched for the root in 'ka'.
This argument is passed to the function \link[stats:uniroot]{stats::uniroot} used for solving the equation.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable the progress bar.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.\cr

Slot: \strong{\verb{@data}}\cr
\tabular{lll}{
\strong{Object} \tab \strong{Type} \tab \strong{Comment} \cr
\code{age.corr} \tab \link{data.frame} \tab Corrected age \cr
\code{age.corr.MC} \tab \link{numeric} \tab MC simulation results with all possible ages from that simulation \cr
}

Slot: \strong{\verb{@info}}\cr

\tabular{lll}{
\strong{Object} \tab \strong{Type} \tab \strong{Comment} \cr
\code{info} \tab \link{character} \tab the original function call
}
}
\description{
Apply a fading correction according to Huntley & Lamothe (2001) for a given
\eqn{g}-value and a given \eqn{t_{c}}
}
\details{
This function solves the equation used for correcting the fading affected age
including the error for a given \eqn{g}-value according to Huntley & Lamothe (2001):

\deqn{
\frac{A_{f}}{A} = 1 - \kappa * \Big[\ln(\frac{A}{t_c}) - 1\Big]
}

with \eqn{\kappa} defined as

\deqn{
\kappa = \frac{\frac{\mathrm{g\_value}}{\ln(10)}}{100}
}

\eqn{A} and \eqn{A_{f}} are given in ka. \eqn{t_c} is given in s, however, it
is internally recalculated to ka.

As the \eqn{g}-value slightly depends on the time \eqn{t_{c}} between
irradiation and the prompt measurement, a value for \code{tc} must always be
provided.
If the \eqn{g}-value was normalised to a distinct
time or evaluated with a different tc value (e.g., external irradiation), also
the \eqn{t_{c}} value for the \eqn{g}-value needs to be provided (argument \code{tc.g_value}
and then the \eqn{g}-value is recalculated
to \eqn{t_{c}} of the measurement used for estimating the age applying the
following equation:

\deqn{\kappa_{tc} = \kappa_{tc.g} / (1 - \kappa_{tc.g} * \ln(tc/tc.g))}

where

\deqn{\kappa_{tc.g} = g / 100 / \ln(10)}

The error of the fading-corrected age is determined using a Monte Carlo
simulation approach. Solving of the equation is performed using
\link{uniroot}. Large values for \code{n.MC} will significantly
increase the computation time.\cr

\strong{\code{n.MC = 'auto'}}

The error estimation based on a stochastic process, i.e. for a small number of
MC runs the calculated error varies considerably every time the function is called,
even with the same input values.
The argument option \code{n.MC = 'auto'} tries to find a stable value for the standard error, i.e.
the standard deviation of values calculated during the MC runs (\code{age.corr.MC}),
within a given precision (2 digits) by increasing the number of MC runs stepwise and
calculating the corresponding error.

If the determined error does not differ from the 9 values calculated previously
within a precision of (here) 3 digits the calculation is stopped as it is assumed
that the error is stable. Please note that (a) the duration depends on the input
values as well as on the provided computation resources and it may take a while,
(b) the length (size) of the output
vector \code{age.corr.MC}, where all the single values produced during the MC runs
are stored, equals the number of MC runs (here termed observations).

To avoid an endless loop the calculation is stopped if the number of observations
exceeds 10^7.
This limitation can be overwritten by setting the number of MC runs manually,
e.g. \code{n.MC = 10000001}. Note: For this case the function is not checking whether the calculated
error is stable.\cr

\strong{\code{seed}}

This option allows to recreate previously calculated results by setting the seed
for the R random number generator (see \link{set.seed} for details). This option
should not be mixed up with the option \strong{\code{n.MC = 'auto'}}. The results may
appear similar, but they are not comparable!\cr

\strong{FAQ}\cr

\strong{Q}: Which \eqn{t_{c}} value is expected?\cr

\strong{A}: \eqn{t_{c}} is the time in seconds between irradiation and the prompt measurement
applied during your \eqn{D_{e}} measurement. However, this \eqn{t_{c}} might
differ from the \eqn{t_{c}} used for estimating the \eqn{g}-value. In the
case of an SAR measurement \eqn{t_{c}} should be similar, however,
if it differs, you have to provide this
\eqn{t_{c}}  value (the one used for estimating the \eqn{g}-value) using
the argument \code{tc.g_value}.\cr

\strong{Q}: The function could not find a solution, what should I do?\cr

\strong{A}: This usually happens for model parameters exceeding the boundaries of the
fading correction model (e.g., very high \eqn{g}-value). Please check
whether another fading correction model might be more appropriate.
}
\note{
Special thanks to Sébastien Huot for his support and clarification via e-mail.
}
\section{Function version}{
 0.4.4
}

\examples{

##run the examples given in the appendix of Huntley and Lamothe, 2001

##(1) faded age: 100 a
results <- calc_FadingCorr(
   age.faded = c(0.1,0),
   g_value = c(5.0, 1.0),
   tc = 2592000,
   tc.g_value = 172800,
   n.MC = 100)

##(2) faded age: 1 ka
results <- calc_FadingCorr(
   age.faded = c(1,0),
   g_value = c(5.0, 1.0),
   tc = 2592000,
   tc.g_value = 172800,
   n.MC = 100)

##(3) faded age: 10.0 ka
results <- calc_FadingCorr(
   age.faded = c(10,0),
   g_value = c(5.0, 1.0),
   tc = 2592000,
   tc.g_value = 172800,
   n.MC = 100)

##access the last output
get_RLum(results)

} 

\section{How to cite}{
Kreutzer, S., 2026. calc_FadingCorr(): Fading Correction after Huntley & Lamothe (2001). Function version 0.4.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Huntley, D.J., Lamothe, M., 2001. Ubiquity of anomalous fading
in K-feldspars and the measurement and correction for it in optical dating.
Canadian Journal of Earth Sciences, 38, 1093-1106.
}
\seealso{
\linkS4class{RLum.Results}, \link{analyse_FadingMeasurement},
\link{get_RLum}, \link{uniroot}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_DoseResponseCurve.R
\name{plot_DoseResponseCurve}
\alias{plot_DoseResponseCurve}
\title{Plot a dose-response curve for luminescence data (Lx/Tx against dose)}
\usage{
plot_DoseResponseCurve(
  object,
  plot_extended = TRUE,
  plot_singlePanels = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Results} (\strong{required}):
An object produced by \link{fit_DoseResponseCurve}.}

\item{plot_extended}{\link{logical} (\emph{with default}):
If \code{TRUE}, 3 plots on one plot area are provided:
\enumerate{
\item growth curve,
\item histogram from Monte Carlo error simulation and
\item a test dose response plot.
}

If \code{FALSE}, just the growth curve will be plotted.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}):
single plot output (\code{TRUE/FALSE}) to allow for plotting the results in
single plot windows. Ignored if \code{plot_extended = FALSE}.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{Further graphical parameters to be passed (supported:
\code{main}, \code{mtext}, \code{xlim}, \code{ylim}, \code{xlab}, \code{ylab}, \code{log}
(not valid for objects fitted with \code{mode = "extrapolation"}), \code{legend} (\code{TRUE/FALSE}),
\code{legend.pos}, \code{reg_points_pch}, \code{density_polygon} (\code{TRUE/FALSE}),
\code{density_polygon_col}, \code{density_rug} (\code{TRUE}/\code{FALSE}),
\code{box} (\code{TRUE}/\code{FALSE}).}
}
\value{
A plot (or a series of plots) is produced.
}
\description{
A dose-response curve is produced for luminescence measurements using a
regenerative or additive protocol as implemented in \link{fit_DoseResponseCurve}.
}
\section{Function version}{
 1.0.8
}

\examples{

##(1) plot dose-response curve for a dummy dataset
data(ExampleData.LxTxData, envir = environment())
fit <- fit_DoseResponseCurve(LxTxData)
plot_DoseResponseCurve(fit)

##(1b) horizontal plot arrangement
layout(mat = matrix(c(1,1,2,3), ncol = 2))
plot_DoseResponseCurve(fit, plot_singlePanels = TRUE)

##(2) plot the dose-response curve with pdf output - uncomment to use
##pdf(file = "~/Dose_Response_Curve_Dummy.pdf", paper = "special")
plot_DoseResponseCurve(fit)
##dev.off()

##(3) plot the growth curve with pdf output - uncomment to use, single output
##pdf(file = "~/Dose_Response_Curve_Dummy.pdf", paper = "special")
plot_DoseResponseCurve(fit, plot_singlePanels = TRUE)
##dev.off()

##(4) plot resulting function for given interval x
x <- seq(1,10000, by = 100)
plot(
 x = x,
 y = eval(fit$Formula),
 type = "l"
)

} 

\section{How to cite}{
Kreutzer, S., Dietze, M., Colombo, M., 2026. plot_DoseResponseCurve(): Plot a dose-response curve for luminescence data (Lx/Tx against dose). Function version 1.0.8. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Berger, G.W., Huntley, D.J., 1989. Test data for exponential fits. Ancient TL 7, 43-46.

Guralnik, B., Li, B., Jain, M., Chen, R., Paris, R.B., Murray, A.S., Li, S.-H., Pagonis, P.,
Herman, F., 2015. Radiation-induced growth and isothermal decay of infrared-stimulated luminescence
from feldspar. Radiation Measurements 81, 224-231.

Pagonis, V., Kitis, G., Chen, R., 2020. A new analytical equation for the dose response of dosimetric materials,
based on the Lambert W function. Journal of Luminescence 225, 117333. \doi{10.1016/j.jlumin.2020.117333}
}
\seealso{
\link{fit_DoseResponseCurve}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Michael Dietze, GFZ Potsdam (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_Daybreak2R.R
\name{read_Daybreak2R}
\alias{read_Daybreak2R}
\title{Import measurement data produced by a Daybreak TL/OSL reader into R}
\usage{
read_Daybreak2R(file, raw = FALSE, verbose = TRUE, txtProgressBar = TRUE, ...)
}
\arguments{
\item{file}{\link{character} or \link{list} (\strong{required}):
name of one or multiple \code{.DAT} or \code{.txt} files (URLs are supported); it can
be the path to a directory, in which case the function tries to detect and
import all files with \code{.txt} extension.}

\item{raw}{\link{logical} (\emph{with default}):
if the input is a DAT-file (binary) a \link[data.table:data.table]{data.table::data.table} instead of
the \linkS4class{RLum.Analysis} object can be returned for debugging purposes.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{txtProgressBar}{\link{logical} (\emph{with default}):
enable/disable \link{txtProgressBar}.}

\item{...}{not in use, for compatibility reasons only}
}
\value{
A list of \linkS4class{RLum.Analysis} objects (each per position)
is returned.
}
\description{
Import a TXT-file (ASCII file) or a DAT-file (binary file) produced by a
Daybreak reader into R. The import of the DAT-files is limited to the file
format described for the software TLAPLLIC v.3.2 used for Daybreak models
1100 and 1150, and for FLConsole (any release) for Daybreak models 11xx and
2200.
}
\note{
\strong{\verb{[BETA VERSION]}}
This function still needs to be tested properly. In particular
the function has underwent only very rough tests using a few files.
}
\section{Function version}{
 0.4.1
}

\examples{

\dontrun{
file <- system.file("extdata/Daybreak_TestFile.txt", package = "Luminescence")
temp <- read_Daybreak2R(file)
}

}
\seealso{
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Data.Curve}, \link[data.table:data.table]{data.table::data.table}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)\cr
Andrzej Bluszcz, Silesian University of Technology, Gliwice (Poland)\cr
Antoine Zink, C2RMF, Palais du Louvre, Paris (France)

The ASCII-file import is based on a suggestion by William Amidon and
Andrew Louis Gorin.
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., Bluszcz, A., Zink, A., 2026. read_Daybreak2R(): Import measurement data produced by a Daybreak TL/OSL reader into R. Function version 0.4.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.FittingLM}
\alias{ExampleData.FittingLM}
\alias{values.curve}
\alias{values.curveBG}
\title{Example data for fit_LMCurve() in the package Luminescence}
\format{
Two objects (data.frames) with two columns (time and counts).
}
\source{
\tabular{ll}{
Lab: \tab Luminescence Laboratory Bayreuth\cr
Lab-Code: \tab BT900\cr
Location: \tab Norway\cr
Material: \tab Beach deposit, coarse grain quartz measured on aluminium discs on a Risø TL/OSL DA-15 reader\cr
}
}
\description{
Linearly modulated (LM) measurement data from a quartz sample from Norway
including background measurement. Measurements carried out in the
luminescence laboratory at the University of Bayreuth.
}
\examples{

##show LM data
data(ExampleData.FittingLM, envir = environment())
plot(values.curve,log="x")

}
\references{
Fuchs, M., Kreutzer, S., Fischer, M., Sauer, D., Soerensen, R., 2012. OSL and IRSL
dating of raised beach sand deposits along the south-eastern coast of Norway.
Quaternary Geochronology 10, 195-200.
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\docType{data}
\name{BaseDataSet.CosmicDoseRate}
\alias{BaseDataSet.CosmicDoseRate}
\alias{values.cosmic.Softcomp}
\alias{values.factor.Altitude}
\alias{values.par.FJH}
\title{Base data set for cosmic dose rate calculation}
\format{
\tabular{ll}{

\code{values.cosmic.Softcomp}: \tab
data frame containing cosmic dose rates
for shallow depths (< 167 g cm^-2) obtained using the "AGE" program by
Rainer Gruen (cf. Gruen 2009). These data essentially reproduce the graph
shown in Fig. 1 of Prescott & Hutton (1988). \cr

\code{values.factor.Altitude}: \tab
data frame containing altitude factors
for adjusting geomagnetic field-change factors. Values were read from Fig. 1
in Prescott & Hutton (1994). \cr

\code{values.par.FJH}: \tab
data frame containing values for parameters F, J
and H (read from Fig. 2 in Prescott & Hutton 1994) used in the expression \cr
}

\deqn{Dc = D0*(F + J \exp((altitude/1000)/H))}
}
\source{
The following data were carefully read from figures in mentioned
sources and used for fitting procedures. The derived expressions are used in
the function \code{calc_CosmicDoseRate}.

\strong{values.cosmic.Softcomp}

\tabular{ll}{
Program: \tab "AGE"\cr
Reference: \tab Gruen (2009) \cr
Fit: \tab Polynomials in the form of
}

For depths between 40-167 g cm^-2:

\deqn{y = 2*10^-6*x^2-0.0008*x+0.2535}

(For depths <40 g cm^-2)

\deqn{y = -6*10^-8*x^3+2*10^-5*x^2-0.0025*x+0.2969}

\strong{\code{values.factor.Altitude}}

\tabular{ll}{
Reference: \tab Prescott & Hutton (1994) \cr
Page: \tab 499 \cr
Figure: \tab 1 \cr
Fit: \tab 2-degree polynomial in the form of
}

\deqn{y = -0.026*x^2 + 0.6628*x + 1.0435}

\strong{\code{values.par.FJH}}

\tabular{ll}{
Reference: \tab Prescott & Hutton (1994) \cr
Page: \tab 500 \cr
Figure: \tab 2 \cr
Fits: \tab 3-degree polynomials and linear fits
}

F (non-linear part, \eqn{\lambda} < 36.5 deg.):

\deqn{y = -7*10^-7*x^3-8*10^-5*x^2-0.0009*x+0.3988}

F (linear part, \eqn{\lambda} > 36.5 deg.):

\deqn{y = -0.0001*x + 0.2347}

J (non-linear part, \eqn{\lambda} < 34 deg.):

\deqn{y = 5*10^-6*x^3-5*10^-5*x^2+0.0026*x+0.5177}

J (linear part, \eqn{\lambda} > 34 deg.):

\deqn{y = 0.0005*x + 0.7388}

H (non-linear part, \eqn{\lambda} < 36 deg.):

\deqn{y = -3*10^-6*x^3-5*10^-5*x^2-0.0031*x+4.398}

H (linear part, \eqn{\lambda} > 36 deg.):

\deqn{y = 0.0002*x + 4.0914}
}
\description{
Collection of data from various sources needed for cosmic dose rate
calculation
}
\section{Version}{
 0.1
}

\examples{

##load data
data(BaseDataSet.CosmicDoseRate)

}
\references{
Gruen, R., 2009. The "AGE" program for the calculation of luminescence age estimates.
Ancient TL 27, pp. 45-46.

Prescott, J.R., Hutton, J.T., 1988. Cosmic ray and gamma ray dosimetry for
TL and ESR. Nuclear Tracks and Radiation Measurements 14, pp. 223-227.

Prescott, J.R., Hutton, J.T., 1994. Cosmic ray contributions to dose rates
for luminescence and ESR dating: large depths and long-term time variations.
Radiation Measurements 23, pp. 497-500.
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_gSGC_feldspar.R
\name{calc_gSGC_feldspar}
\alias{calc_gSGC_feldspar}
\title{Calculate Global Standardised Growth Curve (gSGC) for Feldspar MET-pIRIR}
\usage{
calc_gSGC_feldspar(
  data,
  gSGC.type = "50LxTx",
  gSGC.parameters = NULL,
  n.MC = 100,
  plot = FALSE
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}): data frame with five columns per sample
\code{c("LnTn", "LnTn.error", "Lr1Tr1", "Lr1Tr1.error","Dr1")}}

\item{gSGC.type}{\link{character} (\emph{with default}): growth curve type to be selected
according to Table 3 in Li et al. (2015). Allowed options are
\code{"50LxTx"}, \code{"50Lx"}, \code{"50Tx"}, \code{"100LxTx"}, \code{"100Lx"}, \code{"100Tx"}, \code{"150LxTx"},
\code{"150Lx"}, \code{"150Tx"}, \code{"200LxTx"}, \code{"200Lx"}, \code{"200Tx"}, \code{"250LxTx"}, \code{"250Lx"},
\code{"250Tx"}}

\item{gSGC.parameters}{\link{data.frame} (\emph{optional}): an own parameter set for the
gSGC with the following columns \code{y1}, \code{y1_err}, \code{D1}
\code{D1_err}, \code{y2}, \code{y2_err}, \code{y0}, \code{y0_err}.}

\item{n.MC}{\link{numeric} (\emph{with default}): number of Monte-Carlo runs for the
error calculation}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}
}
\value{
Returns an S4 object of type \linkS4class{RLum.Results}.

\strong{\verb{@data}}\cr
\verb{$ df} (\link{data.frame}) \cr
\code{.. $DE} the calculated equivalent dose\cr
\code{.. $DE.ERROR} error on the equivalent dose, which is the standard deviation of the MC runs\cr
\code{.. $HPD95_LOWER} lower boundary of the highest probability density (95\%)\cr
\code{.. $HPD95_UPPER} upper boundary of the highest probability density (95\%)\cr
\verb{$ m.MC} (\link{list}) numeric vector with results from the MC runs.\cr

\strong{\verb{@info}}\cr
`$ call`` (\link{call}) the original function call
}
\description{
Implementation of the gSGC approach for feldspar MET-pIRIR by Li et al. (2015)
}
\details{
##TODO
}
\section{Function version}{
 0.1.1
}

\examples{

##test on a generated random sample
n_samples <- 10
data <- data.frame(
  LnTn = rnorm(n=n_samples, mean=1.0, sd=0.02),
  LnTn.error = rnorm(n=n_samples, mean=0.05, sd=0.002),
  Lr1Tr1 = rnorm(n=n_samples, mean=1.0, sd=0.02),
  Lr1Tr1.error = rnorm(n=n_samples, mean=0.05, sd=0.002),
  Dr1 = rep(100,n_samples))

results <- calc_gSGC_feldspar(
  data = data, gSGC.type = "50LxTx",
  plot = FALSE)

plot_AbanicoPlot(results)

} 

\section{How to cite}{
Gray, H.J., Kreutzer, S., 2026. calc_gSGC_feldspar(): Calculate Global Standardised Growth Curve (gSGC) for Feldspar MET-pIRIR. Function version 0.1.1. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Li, B., Roberts, R.G., Jacobs, Z., Li, S.-H., Guo, Y.-J., 2015.
Construction of a “global standardised growth curve” (gSGC) for infrared
stimulated luminescence dating of K-feldspar 27, 119–130. \doi{10.1016/j.quageo.2015.02.010}
}
\seealso{
\linkS4class{RLum.Results}, \link{get_RLum}, \link{uniroot}, \link{calc_gSGC}
}
\author{
Harrison Gray, USGS (United States),
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_AverageDose.R
\name{calc_AverageDose}
\alias{calc_AverageDose}
\title{Calculate the Average Dose and the dose rate dispersion}
\usage{
calc_AverageDose(
  data,
  sigma_m,
  Nb_BE = 500,
  na.rm = TRUE,
  plot = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with \code{De} \code{(data[,1])} and \verb{De error} \code{(values[,2])}}

\item{sigma_m}{\link{numeric} (\strong{required}):
the overdispersion resulting from a dose recovery
experiment, i.e. when all grains have  received the same dose. Indeed in such a case, any
overdispersion (i.e. dispersion on top of analytical uncertainties) is, by definition, an
unrecognised measurement uncertainty.}

\item{Nb_BE}{\link{integer} (\emph{with default}):
sample size used for the bootstrapping}

\item{na.rm}{\link{logical} (\emph{with default}):
exclude \code{NA} values from the data set prior to any further operation.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{...}{further arguments that can be passed to \link[graphics:hist]{graphics::hist}. As three plots
are returned all arguments need to be provided as \link{list},
e.g., \code{main = list("Plot 1", "Plot 2", "Plot 3")}.
Note: not all arguments of \code{hist} are
supported, but the output of \code{hist} is returned and can be used of own plots. \cr

Further supported arguments: \code{mtext} (\link{character}), \code{rug} (\code{TRUE/FALSE}).}
}
\value{
The function produces numerical output and an optional plot. In addition,
an \linkS4class{RLum.Results} object containing the following
elements is returned:

\item{$summary}{\link{data.frame} summary of all relevant model results.}
\item{$data}{\link{data.frame} with original input data.}
\item{$dstar}{\link{matrix} with bootstrap values.}
\item{$hist}{\link{list} object as produced by the \link{hist} function.}

The \verb{$summary} data.frame contains the following columns:
\tabular{lll}{
\strong{Column} \tab \strong{Type} \tab \strong{Description}\cr
AVERAGE_DOSE \tab \link{numeric} \tab the obtained average dose\cr
AVERAGE_DOSE.SE \tab \link{numeric} \tab the average dose error \cr
SIGMA_D \tab \link{numeric}\tab sigma \cr
SIGMA_D.SE \tab \link{numeric}\tab standard error of the sigma  \cr
IC_AVERAGE_DOSE.LEVEL  \tab \link{character}\tab confidence level average dose\cr
IC_AVERAGE_DOSE.LOWER  \tab \link{character}\tab lower quantile of average dose \cr
IC_AVERAGE_DOSE.UPPER \tab \link{character}\tab upper quantile of average dose\cr
IC_SIGMA_D.LEVEL \tab \link{integer}\tab confidence level sigma\cr
IC_SIGMA_D.LOWER \tab \link{character}\tab lower sigma quantile\cr
IC_SIGMA_D.UPPER \tab \link{character}\tab upper sigma quantile\cr
L_MAX \tab \link{character}\tab maximum likelihood value \cr
de \tab \link{numeric} \tab same as \code{AVERAGE_DOSE} (internal use)\cr
de_err \tab \link{numeric} \tab same as \code{AVERAGE_DOSE.SE} (internal use)\cr
}

The function returns two different plot panels.

(1) An abanico plot with the dose values

(2) A histogram panel comprising 3 histograms with the equivalent dose and the bootstrapped average
dose and the sigma values.
}
\description{
This functions calculates the Average Dose and its extrinsic dispersion,
estimating the standard errors by bootstrapping based on the Average
Dose Model by Guérin et al., 2017.

\strong{\code{sigma_m}}\cr

The program requires the input of a known value of \code{sigma_m},
which corresponds to the intrinsic overdispersion, as determined
by a dose recovery experiment. Then the dispersion in doses (\code{sigma_d})
will be that over and above \code{sigma_m} (and individual uncertainties \code{sigma_wi}).
}
\note{
This function has beta status!
}
\section{Function version}{
 0.1.7
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())

## calculate Average dose (using only the first 56 values here)
AD <- calc_AverageDose(ExampleData.DeValues$CA1[1:56, ], sigma_m = 0.1)

## plot De and set Average dose as central value
plot_AbanicoPlot(AD, z.0 = AD$summary$de)

} 

\section{How to cite}{
Christophe, C., Philippe, A., Guérin, G., Kreutzer, S., 2026. calc_AverageDose(): Calculate the Average Dose and the dose rate dispersion. Function version 0.1.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Guérin, G., Christophe, C., Philippe, A., Murray, A.S., Thomsen, K.J., Tribolo, C., Urbanova, P.,
Jain, M., Guibert, P., Mercier, N., Kreutzer, S., Lahaye, C., 2017. Absorbed dose, equivalent dose,
measured dose rates, and implications for OSL age estimates: Introducing the Average Dose Model.
Quaternary Geochronology 1-32. doi:10.1016/j.quageo.2017.04.002

\strong{Further reading}\cr

Efron, B., Tibshirani, R., 1986. Bootstrap Methods for Standard Errors, Confidence Intervals,
and Other Measures of Statistical Accuracy. Statistical Science 1, 54-75.
}
\seealso{
\link{read.table}, \link[graphics:hist]{graphics::hist}
}
\author{
Claire Christophe, IRAMAT-CRP2A, Université de Nantes (France)\cr
Anne Philippe, Université de Nantes, (France)\cr
Guillaume Guérin, IRAMAT-CRP2A, Université Bordeaux Montaigne, (France)\cr
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_FiniteMixture.R
\name{calc_FiniteMixture}
\alias{calc_FiniteMixture}
\title{Apply the finite mixture model (FMM) after Galbraith (2005) to a given De
distribution}
\usage{
calc_FiniteMixture(
  data,
  sigmab,
  n.components,
  grain.probability = FALSE,
  pdf.weight = TRUE,
  pdf.sigma = c("sigmab", "se"),
  pdf.colors = c("gray", "colors", "none"),
  plot.proportions = TRUE,
  plot.criteria = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with De \code{(data[,1])} and De error \code{(values[,2])}}

\item{sigmab}{\link{numeric} (\strong{required}):
spread in De values (given as a fraction, e.g. 0.2), representing the
expected overdispersion in the data should the sample be well-bleached
(Cunningham & Wallinga 2012, p. 100).}

\item{n.components}{\link{numeric} (\strong{required}):
number of components to be fitted. If a vector is provided (e.g. \code{c(2:8)}) the
finite mixtures for 2, 3 ... 8 components are calculated and a plot and a
statistical evaluation of the model performance (BIC score and maximum
log-likelihood) is provided.}

\item{grain.probability}{\link{logical} (\emph{with default}):
prints the estimated probabilities of which component each grain is in}

\item{pdf.weight}{\link{logical} (\emph{with default}):
weight the probability density functions by the components proportion.
Ignored if \code{n.components} has length 1.}

\item{pdf.sigma}{\link{character} (\emph{with default}):
if \code{"sigmab"} the components normal distributions are plotted with a common standard
deviation (i.e. \code{sigmab}) as assumed by the FFM. Alternatively,
\code{"se"} takes the standard error of each component for the sigma
parameter of the normal distribution}

\item{pdf.colors}{\link{character} (\emph{with default}):
colour coding of the components in the plot.
Possible options are \code{"gray"}, \code{"colors"} and \code{"none"}.}

\item{plot.proportions}{\link{logical} (\emph{with default}):
plot a \link[graphics:barplot]{graphics::barplot} showing the proportions of components.
Ignored if \code{n.components} has length 1.}

\item{plot.criteria}{\link{logical} (\emph{with default}):
plot the statistical criteria (BIC and log-likelihood).
Ignored if \code{n.components} has length 1.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output. This is ignored and no plot is produced if
\code{n.components} has length 1.}

\item{...}{other parameters to control the plot output. Supported are
\code{cex}, \code{main}, \code{main.densities}, \code{main.proportions}, \code{main.criteria},
\code{pdf.scale}, \code{dose.scale}.}
}
\value{
Returns a plot (\emph{optional}) and terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the
following elements:

\item{.$summary}{\link{data.frame} summary of all relevant model results.}
\item{.$data}{\link{data.frame} original input data}
\item{.$args}{\link{list} used arguments}
\item{.$call}{\link{call} the function call}
\item{.$mle}{ covariance matrices of the log likelihoods}
\item{.$BIC}{ BIC score}
\item{.$llik}{ maximum log likelihood}
\item{.$grain.probability}{ probabilities of a grain belonging to a component}
\item{.$components}{\link{matrix} estimates of the de, de error and proportion for each component}
\item{.$single.comp}{\link{data.frame} single component FFM estimate}

If a vector for \code{n.components} is provided (e.g.  \code{c(2:8)}),
\code{mle} and \code{grain.probability} are lists containing matrices of the
results for each iteration of the model.

The output should be accessed using the function \link{get_RLum}.
}
\description{
This function fits a k-component mixture to a De distribution with differing
known standard errors. Parameters (doses and mixing proportions) are
estimated by maximum likelihood assuming that the log dose estimates come
from a mixture of normal distributions.
}
\details{
This model uses the maximum likelihood and Bayesian Information Criterion
(BIC) approaches.

Indications of overfitting are:
\itemize{
\item increasing BIC
\item repeated dose estimates
\item covariance matrix not positive definite
\item covariance matrix produces \code{NaN}
\item convergence problems
}

\strong{Plot}

If \code{n.components} is a vector (\code{c(k.min:k.max)}), a plot is generated
showing the \emph{k} components equivalent doses as normal distributions.
By default \code{pdf.weight} is set to \code{TRUE}, so that the probability density
functions are weighted by the components proportion for each iteration of
\emph{k} components, so the sum of areas of each component equals 1.
If \code{pdf.weight} is set to \code{FALSE}, the area under each normal distribution
is always 1. While the density values are on the same scale when no weights
are used, the y-axis are individually scaled if the probability density are
weighted by the components proportion.\cr
The standard deviation (sigma) of the normal distributions is by default
determined by a common \code{sigmab} (see \code{pdf.sigma}). For
\code{pdf.sigma = "se"} the standard error of each component is taken
instead.\cr
The stacked \link[graphics:barplot]{graphics::barplot} shows the proportion of each component (in
per cent) calculated by the FMM. The last plot shows the achieved BIC scores
and maximum log-likelihood estimates for each value of \emph{k}.
}
\section{Function version}{
 0.4.4
}

\examples{

## load example data
data(ExampleData.DeValues, envir = environment())

## (1) apply the finite mixture model
## NOTE: the data set is not suitable for the finite mixture model,
## which is why a very small sigmab is necessary
calc_FiniteMixture(ExampleData.DeValues$CA1,
                   sigmab = 0.2, n.components = 2,
                   grain.probability = TRUE)

## (2) repeat the finite mixture model for 2, 3 and 4 maximum number of fitted
## components and save results
## NOTE: The following example is computationally intensive. Please un-comment
## the following lines to make the example work.
FMM<- calc_FiniteMixture(ExampleData.DeValues$CA1,
                         sigmab = 0.2, n.components = c(2:4),
                         pdf.weight = TRUE)

## show structure of the results
FMM

## show the results on equivalent dose, standard error and proportion of
## fitted components
get_RLum(object = FMM, data.object = "components")

} 

\section{How to cite}{
Burow, C., Colombo, M., 2026. calc_FiniteMixture(): Apply the finite mixture model (FMM) after Galbraith (2005) to a given De distribution. Function version 0.4.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Galbraith, R.F. & Green, P.F., 1990. Estimating the component
ages in a finite mixture. Nuclear Tracks and Radiation Measurements 17,
197-206.

Galbraith, R.F. & Laslett, G.M., 1993. Statistical models
for mixed fission track ages. Nuclear Tracks Radiation Measurements 4,
459-470.

Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of
equivalent dose and error calculation and display in OSL dating: An overview
and some recommendations. Quaternary Geochronology 11, 1-27.

Roberts, R.G., Galbraith, R.F., Yoshida, H., Laslett, G.M. & Olley, J.M., 2000.
Distinguishing dose populations in sediment mixtures: a test of single-grain
optical dating procedures using mixtures of laboratory-dosed quartz.
Radiation Measurements 32, 459-465.

Galbraith, R.F., 2005. Statistics for Fission Track Analysis, Chapman & Hall/CRC, Boca Raton.

\strong{Further reading}

Arnold, L.J. & Roberts, R.G., 2009. Stochastic
modelling of multi-grain equivalent dose (De) distributions: Implications
for OSL dating of sediment mixtures. Quaternary Geochronology 4,
204-230.

Cunningham, A.C. & Wallinga, J., 2012. Realizing the
potential of fluvial archives using robust OSL chronologies. Quaternary
Geochronology 12, 98-106.

Rodnight, H., Duller, G.A.T., Wintle, A.G. &
Tooth, S., 2006. Assessing the reproducibility and accuracy of optical
dating of fluvial deposits.  Quaternary Geochronology 1, 109-120.

Rodnight, H. 2008. How many equivalent dose values are needed to obtain a
reproducible distribution?. Ancient TL 26, 3-10.
}
\seealso{
\link{calc_CentralDose}, \link{calc_CommonDose},
\link{calc_FuchsLang2001}, \link{calc_MinDose}
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany) \cr
Based on a rewritten S script of Rex Galbraith, 2006.
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale_GammaDose.R
\name{scale_GammaDose}
\alias{scale_GammaDose}
\title{Calculate the gamma dose deposited within a sample taking layer-to-layer
variations in radioactivity into account (according to Aitken, 1985)}
\usage{
scale_GammaDose(
  data,
  conversion_factors = c("Cresswelletal2018", "Guerinetal2011", "AdamiecAitken1998",
    "Liritzisetal2013")[1],
  fractional_gamma_dose = "Aitken1985",
  verbose = TRUE,
  plot = TRUE,
  plot_singlePanels = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}):
A table containing all relevant information for each individual layer. The
table must have the following named columns:
\itemize{
\item \code{id} (\link{character}): an arbitrary id or name of each layer
\item \code{thickness} (\link{numeric}): vertical extent of each layer in cm
\item \code{sample_offset} (\link{logical}): distance of the sample in cm,
\strong{measured from the bottom of the target layer}. Except for the target layer
all values must be \code{NA}.
\item \code{K} (\link{numeric}): K nuclide content in \%
\item \code{K_se} (\link{numeric}): error on the K content
\item \code{Th} (\link{numeric}): Th nuclide content in ppm
\item \code{Th_se} (\link{numeric}): error on the Th content
\item \code{U} (\link{numeric}): U nuclide content in ppm
\item \code{U_se} (\link{numeric}): error on the U content
\item \code{water_content} (\link{numeric}): water content of each layer in \%
\item \code{water_content_se} (\link{numeric}):  error on the water content
\item \code{density} (\link{numeric}): bulk density of each layer in g/cm^-3
}}

\item{conversion_factors}{\link{character} (\emph{optional}):
The conversion factors used to calculate the dose rate from sediment
nuclide contents. Valid options are:
\itemize{
\item \code{"Cresswelletal2018"} (default)
\item \code{"Liritzisetal2013"}
\item \code{"Guerinetal2011"}
\item \code{"AdamiecAitken1998"}
}}

\item{fractional_gamma_dose}{\link{character} (\emph{optional}):
Factors to scale gamma dose rate values. Valid options are:
\itemize{
\item \code{"Aitken1985"} (default): Table H1 in the appendix
}}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{plot_singlePanels}{\link{logical} (\emph{with default}):
enable/disable single plot mode, i.e. one plot window per plot.}

\item{...}{Further parameters passed to \link{barplot}.}
}
\value{
After performing the calculations the user is provided with different outputs.
\enumerate{
\item The total gamma dose rate received by the sample (+/- uncertainties) as a
print in the console.
\item A plot showing the sediment sequence, the user input sample information
and the contribution to total gamma dose rate.
\item RLum Results. If the user wishes to save these results, writing a script
to run the function and to save the results would look like this:
}

\if{html}{\out{<div class="sourceCode">}}\preformatted{mydata <- read.table("c:/path/to/input/file.txt")
results <- scale_GammaDose(mydata)
table <- get_RLum(results)
write.csv(table, "c:/path/to/results.csv")
}\if{html}{\out{</div>}}

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$summary} \tab \code{data.frame} \tab summary of the model results \cr
\verb{$data} \tab \code{data.frame} \tab the original input data \cr
\verb{$dose_rates} \tab \code{list} \tab two \code{data.frames} for the scaled and infinite matrix dose rates \cr
\verb{$tables} \tab \code{list} \tab several \code{data.frames} containing intermediate results \cr
\verb{$args} \tab \code{character} \tab arguments of the call \cr
\verb{$call} \tab \code{call} \tab the original function call \cr
}

\strong{slot:} \strong{\verb{@info}}

Currently unused.

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr

Three plots are produced:
\itemize{
\item A visualisation of the provided sediment layer structure to quickly
assess whether the data was provided and interpreted correctly.
\item A scatter plot of the nuclide contents per layer (K, Th, U) as well as the
water content. This may help to correlate the dose rate contribution of
specific layers to the layer of interest.
\item A barplot visualising the contribution of each layer to the total dose rate
received by the sample in the target layer.
}
}
\description{
This function calculates the gamma dose deposited in a luminescence sample
taking into account layer-to-layer variations in sediment radioactivity.
The function scales user inputs of uranium, thorium and potassium based on
input parameters for sediment density, water content and given layer
thicknesses and distances to the sample.

\strong{User Input}

To calculate the gamma dose which is deposited in a sample, the user needs
to provide information on those samples influencing the luminescence sample.
As a rule of thumb, all sediment layers within at least 30 cm radius from
the luminescence sample taken should be taken into account when calculating
the gamma dose rate. However, the actual range of gamma radiation might be
different, depending on the emitting radioelement, the water content and the
sediment density of each layer (Aitken, 1985). Therefore the user is
advised to provide as much detail as possible and physically sensible.

The function requires a \link{data.frame} that is to be structured
in columns and rows, with samples listed in rows. The first column contains
information on the layer/sample ID, the second on the thickness (in cm) of
each layer, whilst column 3 should contain \code{NA} for all layers that are not
sampled for OSL/TL. For the layer the OSL/TL sample was taken from a numerical
value must be provided, which is the distance (in cm) measured from \strong{bottom}
of the layer of interest. If the whole layer was sampled insert \code{0}. If the
sample was taken from \emph{within} the layer, insert a numerical value \verb{>0},
which describes the distance from the middle of the sample to the bottom of
the layer in cm. Columns 4 to 9 should contain radionuclide concentrations
and their standard errors for
potassium (in \%), thorium (in ppm) and uranium (in ppm). Columns 10 and 11
give information on the water content and its uncertainty (standard error)
in \%. The layer density (in g/cm3) should be given in column 12. No cell
should be left blank. Please ensure to keep the column titles as given in
the example dataset (\code{data('ExampleData.ScaleGammaDose')}, see examples).

The user can decide which dose rate
conversion factors should be used to calculate the gamma dose rates.
The options are:
\itemize{
\item \code{"Cresswelletal2018"} (Cresswell et al., 2018)
\item \code{"Liritzisetal2013"} (Liritzis et al., 2013)
\item \code{"Guerinetal2011"} (Guérin et al., 2011)
\item \code{"AdamiecAitken1998"} (Adamiec and Aitken, 1998)
}

\strong{Water content}

The water content provided by the user should be calculated according to:

\deqn{ ( Wet weight [g] - Dry weight [g] ) / Dry weight [g] * 100 }

\strong{Calculations}

After converting the radionuclide concentrations into dose rates, the
function will scale the dose rates based on the thickness of the layers,
the distances to the sample, the water content and the density of the sediment.
The calculations are based on Aitken (1985, Appendix H). As an example
(equivalent to Aitken, 1985), assuming three layers of sediment, where \strong{L} is
inert and positioned in between the infinite thick and equally active
layers \strong{A} and \strong{B}, the dose in \strong{L} and \strong{B} due to \strong{A} is given by

\deqn{ {1-f(x)}D_A }

Where \code{x} is the distance into the inert medium, so \code{f(x)} is the weighted
average fractional dose at \code{x} and \code{D_A} denotes that the dose is delivered by \strong{A}.
\code{f(x)} is derived from table H1 (Aitken, 1985), when setting \code{z = x}.
Consequently, the dose in \strong{A} and \strong{L} due to \strong{B} is given by

\deqn{ {1 - f(t-x)}D_B }

Here \code{t} is the thickness of \strong{L} and the other parameters are denoted as above,
just for the dose being delivered by B. \code{f(t-x)} is derived from table H1
(Aitken, 1985), when setting \code{z} equal to \code{t-x}. Following this, the dose in \strong{L}
delivered by \strong{A} and \strong{B} is given by

\deqn{ {2 - f(x) - f(t-x)}D_{AB} }

Since \strong{A} and \strong{B} are equally active \verb{D_\{AB\} = D_A = D_B}.

The function uses the value of the fractional dose rate at the layer
boundary to start the calculation for the next layer. This way, the function
is able to scale the gamma dose rate accurately for distant layers when the
density and water content is not constant for the entire section.
}
\note{
\strong{This function has BETA status. If possible, results should be}
\strong{cross-checked.}
}
\section{Function version}{
 0.1.4
}

\section{Acknowledgements}{


We thank Dr Ian Bailiff for the provision of an excel spreadsheet, which has
been very helpful when writing this function.
}

\examples{

# Load example data
data("ExampleData.ScaleGammaDose", envir = environment())
x <- ExampleData.ScaleGammaDose

# Scale gamma dose rate
results <- scale_GammaDose(data = x,
                           conversion_factors = "Cresswelletal2018",
                           fractional_gamma_dose = "Aitken1985",
                           verbose = TRUE,
                           plot = TRUE)

get_RLum(results)

} 

\section{How to cite}{
Riedesel, S., Autzen, M., Burow, C., 2026. scale_GammaDose(): Calculate the gamma dose deposited within a sample taking layer-to-layer variations in radioactivity into account (according to Aitken, 1985). Function version 0.1.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Aitken, M.J., 1985. Thermoluminescence Dating. Academic Press, London.

Adamiec, G., Aitken, M.J., 1998. Dose-rate conversion factors: update.
Ancient TL 16, 37-46. \doi{10.26034/la.atl.1998.292}

Cresswell., A.J., Carter, J., Sanderson, D.C.W., 2018.
Dose rate conversion parameters: Assessment of nuclear data.
Radiation Measurements 120, 195-201. \doi{10.1016/j.radmeas.2018.02.007}

Guérin, G., Mercier, N., Adamiec, G., 2011. Dose-rate conversion
factors: update. Ancient TL 29, 5-8. \doi{10.26034/la.atl.2011.443}

Liritzis, I., Stamoulis, K., Papachristodoulou, C., Ioannides, K., 2013.
A re-evaluation of radiation dose-rate conversion factors. Mediterranean
Archaeology and Archaeometry 13, 1-15.
}
\seealso{
\link{ExampleData.ScaleGammaDose},
\link{BaseDataSet.ConversionFactors}, \link{approx}, \link{barplot}
}
\author{
Svenja Riedesel, Aberystwyth University (United Kingdom) \cr
Martin Autzen, DTU NUTECH Center for Nuclear Technologies (Denmark) \cr
Christoph Burow, University of Cologne (Germany) \cr
Based on an excel spreadsheet and accompanying macro written by Ian Bailiff.
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.RLum.Data.Image}
\alias{ExampleData.RLum.Data.Image}
\title{Example data as \linkS4class{RLum.Data.Image} objects}
\format{
Object of class \linkS4class{RLum.Data.Image}
}
\source{
\strong{ExampleData.RLum.Data.Image}

These data were kindly provided by Regina DeWitt.

\tabular{ll}{
Lab.: \tab Department of Physics, East-Carolina University, NC, USA\cr
Lab-Code: \tab - \cr
Location: \tab - \cr
Material: \tab - \cr
Reference: \tab - \cr
}

Image data is a measurement of fluorescent ceiling lights with a cooled
Princeton Instruments (TM) camera fitted on Risø DA-20 TL/OSL reader.
}
\description{
Measurement of Princton Instruments camera imported with the function
\link{read_SPE2R} to R to produce an
\linkS4class{RLum.Data.Image} object.
}
\section{Version}{
 0.1
}

\examples{

##load data
data(ExampleData.RLum.Data.Image, envir = environment())

##plot data
plot_RLum(ExampleData.RLum.Data.Image)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_ThermalQuenching.R
\name{fit_ThermalQuenching}
\alias{fit_ThermalQuenching}
\title{Fitting Thermal Quenching Data}
\usage{
fit_ThermalQuenching(
  data,
  start_param = list(),
  method_control = list(),
  n.MC = 100,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{\link{data.frame} (\strong{required}): input data with three columns, the first column contains
temperature values in deg. C, columns 2 and 3 the dependent values with its error}

\item{start_param}{\link{list} (\emph{optional}): option to provide the start
parameters for the fitting, see details}

\item{method_control}{\link{list} (\emph{optional}): further options to fine tune
the fitting, see details for
further information}

\item{n.MC}{\link{integer} (\emph{with default}): number of Monte Carlo runs for the
error estimation. If \code{n.MC} is \code{NULL} or 1, the error estimation is skipped.}

\item{verbose}{\link{logical} (\emph{with default}): enable/disable output to the
terminal.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.}

\item{...}{further arguments that can be passed to control the plotting, support are \code{main}, \code{pch},
\code{col_fit}, \code{col_points}, \code{lty}, \code{lwd}, \code{xlab}, \code{ylab}, \code{xlim}, \code{ylim}, \code{xaxt}}
}
\value{
The function returns numerical output and an (\emph{optional}) plot.

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\verb{[.. $data : data.frame]}\cr

A table with all fitting parameters and the number of Monte Carlo runs used
for the error estimation (this may be smaller that \code{n.MC}).

\verb{[.. $fit : nls object]} \cr

The nls \link[stats:nls]{stats::nls} object returned by the function \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}. This object
can be further passed to other functions supporting an nls object (cf. details section
in \link[stats:nls]{stats::nls})

\strong{slot:} \strong{\verb{@info}}

\verb{[.. $call : call]}\cr

The original function call.

-----------------------------------\cr
\verb{[ GAPHICAL OUTPUT ]}\cr
-----------------------------------\cr

Plotted are temperature against the signal and their uncertainties.
The fit is shown as dashed-line (can be modified). Please note that for the fitting the absolute
temperature values are used but are re-calculated to deg. C for the plot.
}
\description{
Applying a nls-fitting to thermal quenching data.
}
\details{
\strong{Used equation}\cr

The equation used for the fitting is

\deqn{y = A / (1 + C * \exp(-W / (k_B x))) + c}

\emph{W} is the energy depth in eV, \emph{C} is a dimensionless constant, \emph{A} and \emph{c}
are used to adjust the curve for the given signal, \eqn{k_B} is the Boltzmann
constant in eV/K and \emph{x} is the absolute temperature in K.

\strong{Error estimation}\cr

The error estimation is done by varying the input parameters using the
given uncertainties in a Monte Carlo simulation. Errors are assumed to
follow a normal distribution.

\strong{\code{start_param}} \cr

The function allows the injection of starting values for the parameters to
be optimised via the \code{start_param} argument. The parameters must be provided
as a named list.
Examples: \code{start_param = list(A = 1, C = 1e+5, W = 0.5, c = 0)}

\strong{\code{method_control}} \cr

The following arguments can be provided via \code{method_control}. Please note that arguments provided
via \code{method_control} are not further tested, i.e., if the function crashes your input was probably
wrong.

\tabular{lll}{
\strong{ARGUMENT} \tab \strong{TYPE} \tab \strong{DESCRIPTION}\cr
\code{upper} \tab named \link{vector} \tab sets upper fitting boundaries, if provided boundaries for all arguments
are required, e.g., \code{c(A = 0, C = 0, W = 0, c = 0)} \cr
\code{lower} \tab names \link{vector} \tab set lower fitting boundaries (see \code{upper} for details) \cr
\code{trace}   \tab \link{logical} \tab enable/disable progression trace for \link[minpack.lm:nlsLM]{minpack.lm::nlsLM}\cr
\code{weights} \tab \link{numeric} \tab option to provide own weights for the fitting, the length of this
vector needs to be equal to the number for rows of the input \code{data.frame}. If set to \code{NULL} no weights
are applied. The weights are defined by the third column of the input \code{data.frame}.
}
}
\section{Function version}{
 0.2
}

\examples{

##create short example dataset
data <- data.frame(
  T = c(25, 40, 50, 60, 70, 80, 90, 100, 110),
  V = c(0.06, 0.058, 0.052, 0.051, 0.041, 0.034, 0.035, 0.033, 0.032),
  V_X = c(0.012, 0.009, 0.008, 0.008, 0.007, 0.006, 0.005, 0.005, 0.004))

##fit
fit_ThermalQuenching(
 data = data,
 n.MC = NULL)

} 

\section{How to cite}{
Kreutzer, S., Colombo, M., 2026. fit_ThermalQuenching(): Fitting Thermal Quenching Data. Function version 0.2. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Wintle, A.G., 1975. Thermal Quenching of Thermoluminescence in Quartz. Geophys. J. R. astr. Soc. 41, 107–113.
}
\seealso{
\link[minpack.lm:nlsLM]{minpack.lm::nlsLM}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany) \cr
Marco Colombo, Institute of Geography, Heidelberg University (Germany)
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R, R/RLum-class.R
\name{replicate_RLum}
\alias{replicate_RLum}
\alias{replicate_RLum,RLum-method}
\title{General replication function for RLum-class objects}
\usage{
replicate_RLum(object, times = 1)

\S4method{replicate_RLum}{RLum}(object, times = 1)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
an \linkS4class{RLum} object}

\item{times}{\link{integer} (\emph{optional}):
number for times each element should be repeated.}
}
\value{
A \link{list} with the object repeated.
}
\description{
The function replicates \linkS4class{RLum} objects and returns a list of such
objects.
}
\section{Functions}{
\itemize{
\item \code{replicate_RLum(RLum)}: Replication method for \linkS4class{RLum} objects.

}}
\section{Function version}{
 0.1.0
}

\seealso{
\linkS4class{RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. replicate_RLum(): General replication function for RLum-class objects. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_Al2O3C_Measurement.R
\name{analyse_Al2O3C_Measurement}
\alias{analyse_Al2O3C_Measurement}
\title{Al2O3:C Passive Dosimeter Measurement Analysis}
\usage{
analyse_Al2O3C_Measurement(
  object,
  signal_integral = NULL,
  integral_input = c("channel", "measurement"),
  dose_points = c(0, 4),
  recordType = c("OSL (UVVIS)", "TL (UVVIS)"),
  calculate_TL_dose = FALSE,
  irradiation_time_correction = NULL,
  cross_talk_correction = NULL,
  travel_dosimeter = NULL,
  test_parameters = NULL,
  verbose = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} (\strong{required}):  measurement input}

\item{signal_integral}{\link{numeric} (\emph{optional}): signal integral, used for the signal
and the background. Example: \code{c(1:10)} for the first 10 channels.
If nothing is provided the full range is used}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given time range (in seconds) are selected.}

\item{dose_points}{\link{numeric} (\emph{with default}):
vector with dose points, if dose points are repeated, only the general
pattern needs to be provided. Default values follow the suggestions
made by Kreutzer et al., 2018}

\item{recordType}{\link{character} (\emph{with default}): input curve selection, which is passed to
function \link{get_RLum}. To deactivate the automatic selection set the argument to \code{NULL}}

\item{calculate_TL_dose}{\link{logical} (\emph{with default}):
enables/disable experimental dose estimation based on the TL curves.
It is computed as the ratio of the peak sums of each curves +/- 5 channels.}

\item{irradiation_time_correction}{\link{numeric} or \linkS4class{RLum.Results} (\emph{optional}):
information on the used irradiation time correction obtained by another experiments.
If a \code{numeric} is provided it has to be of length two: mean, standard error}

\item{cross_talk_correction}{\link{numeric} or \linkS4class{RLum.Results} (\emph{optional}):
information on the used irradiation time correction obtained by another experiments.
If a \code{numeric} vector is provided it has to be of length three:
mean, 2.5 \% quantile, 97.5 \% quantile.}

\item{travel_dosimeter}{\link{numeric} (\emph{optional}): specify the position of
the travel dosimeter (so far measured at the same time). The dose of
travel dosimeter will be subtracted from all other values.}

\item{test_parameters}{\link{list} (\emph{with default}):
set test parameters. Supported parameters are: \code{TL_peak_shift} All input: \link{numeric}
values, \code{NA} and \code{NULL} (see details).}

\item{verbose}{\link{logical} (\emph{with default}):
enable/disable output to the terminal.}

\item{plot}{\link{logical} (\emph{with default}): enable/disable the plot output.
If \code{object} is of type \link{list}, a numeric vector can be provided to limit
the plot output to certain aliquots.}

\item{...}{further arguments that can be passed to the plot output, supported are \code{norm}, \code{main}, \code{mtext},
\code{title} (for self-call mode to specify, e.g., sample names)}
}
\value{
Function returns results numerically and graphically:

-----------------------------------\cr
\verb{[ NUMERICAL OUTPUT ]}\cr
-----------------------------------\cr

\strong{\code{RLum.Results}}-object

\strong{slot:} \strong{\verb{@data}}

\tabular{lll}{
\strong{Element} \tab \strong{Type} \tab \strong{Description}\cr
\verb{$data} \tab \code{data.frame} \tab the estimated equivalent dose \cr
\verb{$data_table} \tab \code{data.frame} \tab full dose and signal table \cr
\code{test_parameters} \tab \code{data.frame} \tab results with test parameters \cr
\code{data_TDcorrected} \tab \code{data.frame} \tab travel dosimeter corrected results (only if TD was provided)\cr
}

\strong{Note:} If correction the irradiation time and the cross-talk correction
method is used, the De values in the table \code{data} table are already
corrected, i.e. if you want to get an uncorrected value, you can use the
column \code{CT_CORRECTION} remove the correction.

\strong{slot:} \strong{\verb{@info}}

The original function call

------------------------\cr
\verb{[ PLOT OUTPUT ]}\cr
------------------------\cr
\itemize{
\item OSL and TL curves, combined on two plots.
}
}
\description{
The function provides the analysis routines for measurements on a
FI lexsyg SMART reader using Al2O3:C chips according to Kreutzer et al., 2018
}
\details{
\strong{Working with a travel dosimeter}

The function allows to define particular position numbers as travel
dosimeters. For example: \code{travel_dosimeter = c(1,3,5)} sets positions 1,
3 and 5 as travel dosimeters. These dose values #' of this dosimeters
are combined and automatically subtracted from the obtained dose values
of the other dosimeters.

\strong{Calculate TL dose}

The argument \code{calculate_TL_dose} provides the possibility to experimentally calculate a TL-dose,
i.e. an apparent dose value derived from the TL curve ratio. However, it should be noted that
this value is only a fall back in case something went wrong during the measurement of the optical
stimulation. The TL derived dose value is corrected for cross-talk and for the irradiation time,
but not considered if a travel dosimeter is defined.

Calculating the palaeodose is possible without \strong{any TL} curve in the sequence!

\strong{Test parameters}

\code{TL_peak_shift} \link{numeric} (default: \code{15}):

Checks whether the TL peak shift is bigger > 15 K, indicating a problem with the
thermal contact of the chip.

\code{stimulation_power} \link{numeric} (default: \code{0.05}):

So far available, information on the delivered optical stimulation are compared. Compared are
the information from the first curves with all others. If the ratio differs more from
unity than the defined by the threshold, a warning is returned.
}
\section{Function version}{
 0.2.7
}

\examples{
##load data
data(ExampleData.Al2O3C, envir = environment())

##run analysis
analyse_Al2O3C_Measurement(data_CrossTalk)

} 

\section{How to cite}{
Kreutzer, S., 2026. analyse_Al2O3C_Measurement(): Al2O3:C Passive Dosimeter Measurement Analysis. Function version 0.2.7. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Kreutzer, S., Martin, L., Guérin, G., Tribolo, C., Selva, P., Mercier, N., 2018. Environmental Dose Rate
Determination Using a Passive Dosimeter: Techniques and Workflow for alpha-Al2O3:C Chips.
Geochronometria 45, 56-67.
}
\seealso{
\link{analyse_Al2O3C_ITC}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RLum-class.R
\docType{class}
\name{RLum-class}
\alias{RLum-class}
\title{Class \code{"RLum"}}
\description{
Abstract class for data in the package Luminescence
Subclasses are:
}
\details{
\strong{RLum-class}\cr
|\cr
|----\linkS4class{RLum.Data}\cr
|----|-- \linkS4class{RLum.Data.Curve}\cr
|----|-- \linkS4class{RLum.Data.Spectrum}\cr
|----|-- \linkS4class{RLum.Data.Image}\cr
|----\linkS4class{RLum.Analysis}\cr
|----\linkS4class{RLum.Results}
}
\section{Slots}{

\describe{
\item{\code{originator}}{Object of class \link{character} containing the name of the producing
function for the object. Set automatically by using the function \link{set_RLum}.}

\item{\code{info}}{Object of class \link{list} for additional information on the object itself}

\item{\code{.uid}}{Object of class \link{character} for a unique object identifier. This id is
usually calculated using the internal function \code{create_UID} if the function \link{set_RLum}
is called.}

\item{\code{.pid}}{Object of class \link{character} for a parent id. This allows nesting RLum-objects
at will. The parent id can be the uid of another object.}
}}

\note{
\code{RLum} is a virtual class.
}
\section{Objects from the Class}{

A virtual Class: No objects can be created from it.
}

\section{Class version}{
 0.4.0
}

\examples{

showClass("RLum")

}
\seealso{
\linkS4class{RLum.Data}, \linkS4class{RLum.Data.Curve},
\linkS4class{RLum.Data.Spectrum}, \linkS4class{RLum.Data.Image},
\linkS4class{RLum.Analysis}, \linkS4class{RLum.Results}, \link{methods_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_CentralDose.R
\name{calc_CentralDose}
\alias{calc_CentralDose}
\title{Apply the central age model (CAM) after Galbraith et al. (1999) to a given
De distribution}
\usage{
calc_CentralDose(data, sigmab = 0, log = TRUE, plot = TRUE, ...)
}
\arguments{
\item{data}{\linkS4class{RLum.Results} or \link{data.frame} (\strong{required}):
for \link{data.frame}: two columns with De \code{(data[,1])} and De error \code{(data[,2])}.
Records containing missing values will be removed.}

\item{sigmab}{\link{numeric} (\emph{with default}):
additional spread in De values, representing the expected overdispersion in
the data should the sample be well-bleached (Cunningham & Wallinga 2012, p. 100).
\strong{Note:} For the logged model (\code{log = TRUE}) this value must be
a fraction, e.g. 0.2 (= 20 \%). If the un-logged model is used (\code{log = FALSE}),
sigmab must be provided in the same absolute units of the De values (seconds or Gray).}

\item{log}{\link{logical} (\emph{with default}):
fit the (un-)logged central age model to De data. Log transformation is
allowed only if the De values are positive.}

\item{plot}{\link{logical} (\emph{with default}):
enable/disable the plot output.}

\item{...}{further arguments (\code{trace}, \code{verbose}).}
}
\value{
Returns a plot (\emph{optional}) and terminal output. In addition an
\linkS4class{RLum.Results} object is returned containing the following elements:

\item{.$summary}{\link{data.frame} summary of all relevant model results.}
\item{.$data}{\link{data.frame} original input data}
\item{.$args}{\link{list} used arguments}
\item{.$call}{\link{call} the function call}
\item{.$profile}{\link{data.frame} the log likelihood profile for sigma}

The output should be accessed using the function \link{get_RLum}.
}
\description{
This function calculates the central dose and dispersion of the De
distribution, their standard errors and the profile log likelihood function
for \code{sigma}.

This function uses the equations of Galbraith & Roberts (2012). The
parameters \code{delta} and \code{sigma} are estimated by numerically solving
eq. 15 and 16. Their standard errors are approximated using eq. 17.
In addition, the profile log-likelihood function for \code{sigma} is
calculated using eq. 18 and presented as a plot. Numerical values of the
maximum likelihood approach are \strong{only} presented in the plot and \strong{not}
in the console. A detailed explanation on maximum likelihood estimation can
be found in the appendix of Galbraith & Laslett (1993, 468-470) and
Galbraith & Roberts (2012, 15)
}
\section{Function version}{
 1.5
}

\examples{

##load example data
data(ExampleData.DeValues, envir = environment())

##apply the central dose model
calc_CentralDose(ExampleData.DeValues$CA1)

} 

\section{How to cite}{
Burow, C., 2026. calc_CentralDose(): Apply the central age model (CAM) after Galbraith et al. (1999) to a given De distribution. Function version 1.5. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Galbraith, R.F. & Laslett, G.M., 1993. Statistical models for
mixed fission track ages. Nuclear Tracks Radiation Measurements 4, 459-470.

Galbraith, R.F., Roberts, R.G., Laslett, G.M., Yoshida, H. & Olley,
J.M., 1999. Optical dating of single grains of quartz from Jinmium rock
shelter, northern Australia. Part I: experimental design and statistical
models.  Archaeometry 41, 339-364.

Galbraith, R.F. & Roberts, R.G., 2012. Statistical aspects of equivalent dose and error calculation and
display in OSL dating: An overview and some recommendations. Quaternary
Geochronology 11, 1-27.

\strong{Further reading}

Arnold, L.J. & Roberts, R.G., 2009. Stochastic modelling of multi-grain equivalent dose
(De) distributions: Implications for OSL dating of sediment mixtures.
Quaternary Geochronology 4, 204-230.

Bailey, R.M. & Arnold, L.J., 2006. Statistical modelling of single grain quartz De distributions and an
assessment of procedures for estimating burial dose. Quaternary Science
Reviews 25, 2475-2502.

Cunningham, A.C. & Wallinga, J., 2012. Realizing the potential of fluvial archives using robust OSL chronologies.
Quaternary Geochronology 12, 98-106.

Rodnight, H., Duller, G.A.T., Wintle, A.G. & Tooth, S., 2006. Assessing the reproducibility and accuracy
of optical dating of fluvial deposits.  Quaternary Geochronology 1 109-120.

Rodnight, H., 2008. How many equivalent dose values are needed to
obtain a reproducible distribution?. Ancient TL 26, 3-10.
}
\seealso{
\link{plot}, \link{calc_CommonDose}, \link{calc_FiniteMixture},
\link{calc_FuchsLang2001}, \link{calc_MinDose}
}
\author{
Christoph Burow, University of Cologne (Germany) \cr
Based on a rewritten S script of Rex Galbraith, 2010
, RLum Developer Team}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-generics.R,
%   R/RLum.Analysis-class.R, R/RLum.Data.Curve-class.R
\name{melt_RLum}
\alias{melt_RLum}
\alias{melt_RLum,list-method}
\alias{melt_RLum,RLum.Analysis-method}
\alias{melt_RLum,RLum.Data.Curve-method}
\title{Melt RLum-class objects into a flat data.frame}
\usage{
melt_RLum(object, ...)

\S4method{melt_RLum}{list}(object, ...)

\S4method{melt_RLum}{RLum.Analysis}(object)

\S4method{melt_RLum}{RLum.Data.Curve}(object)
}
\arguments{
\item{object}{\linkS4class{RLum} (\strong{required}):
S4 object of class \code{RLum}}

\item{...}{further arguments passed to the specific class method}
}
\value{
A flat \link{data.frame}.
}
\description{
Melt RLum-class objects into a flat data.frame
}
\section{Functions}{
\itemize{
\item \code{melt_RLum(list)}: Returns a list of melted \linkS4class{RLum} objects; non-RLum objects are silently
removed.

\item \code{melt_RLum(RLum.Analysis)}: Melts \linkS4class{RLum.Analysis} objects into a flat data.frame with columns
\code{X}, \code{Y}, \code{TYPE}, \code{UID}, to be used in combination with other packages
such as \code{ggplot2}.

\item \code{melt_RLum(RLum.Data.Curve)}: Melts \linkS4class{RLum.Data.Curve} objects into a flat data.frame with columns
\code{X}, \code{Y}, \code{TYPE}, \code{UID}, to be used in combination with other packages
such as \code{ggplot2}.

}}
\section{Function version}{
 0.1.0
}

\examples{
data(ExampleData.XSYG, envir = environment())
melt_RLum(OSL.SARMeasurement[[2]][[1]])

}
\seealso{
\linkS4class{RLum.Data.Curve}, \linkS4class{RLum.Analysis}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team} 

\section{How to cite}{
Kreutzer, S., 2026. melt_RLum(): Melt RLum-class objects into a flat data.frame. Function version 0.1.0. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/analyse_SAR.TL.R
\name{analyse_SAR.TL}
\alias{analyse_SAR.TL}
\title{Analyse SAR TL measurements}
\usage{
analyse_SAR.TL(
  object,
  object.background,
  signal_integral,
  integral_input = "channel",
  sequence.structure = c("PREHEAT", "SIGNAL", "BACKGROUND"),
  rejection.criteria = list(recycling.ratio = 10, recuperation.rate = 10),
  dose.points = NULL,
  log = "",
  ...
)
}
\arguments{
\item{object}{\linkS4class{RLum.Analysis} or a \link{list} of such objects (\strong{required}) :
input object containing data for analysis}

\item{object.background}{currently not used}

\item{signal_integral}{\link{integer} (\strong{required}):
vector of inputs (as defined by \code{integral_input}) for the signal integral.}

\item{integral_input}{\link{character} (\emph{with default}):
input type for \code{signal_integral}, one of \code{"channel"} (default) or
\code{"measurement"}. If set to \code{"measurement"}, the best matching channels
corresponding to the given temperature range are selected.}

\item{sequence.structure}{\link{vector} \link{character} (\emph{with default}):
specifies the general sequence structure. Three steps are allowed
(\code{"PREHEAT"}, \code{"SIGNAL"}, \code{"BACKGROUND"}), in addition a
parameter \code{"EXCLUDE"}. This allows excluding TL curves which are not
relevant for the protocol analysis.  (\strong{Note:} No TL are removed by default)}

\item{rejection.criteria}{\link{list} (\emph{with default}):
list containing rejection criteria in percentage for the calculation.}

\item{dose.points}{\link{numeric} (\emph{optional}):
option set dose points manually}

\item{log}{\link{character} (\emph{with default}):
a character string which contains \code{"x"} if the x-axis is to be logarithmic,
\code{"y"} if the y axis is to be logarithmic and \code{"xy"} or \code{"yx"} if both axes
are to be logarithmic. See
\link{plot.default}).}

\item{...}{further arguments that will be passed to the function \link{fit_DoseResponseCurve}}
}
\value{
A plot (\emph{optional}) and an \linkS4class{RLum.Results} object is
returned containing the following elements:

\item{De.values}{\link{data.frame} containing De-values and further parameters}
\item{LnLxTnTx.values}{\link{data.frame} of all calculated \code{Lx/Tx} values including signal, background counts and the dose points.}
\item{rejection.criteria}{\link{data.frame} with values that might by used as rejection criteria. \code{NA} is produced if no R0 dose point exists.}

The output should be accessed using the function \link{get_RLum}.
}
\description{
The function performs a SAR TL analysis on a
\linkS4class{RLum.Analysis} object including growth curve fitting.
}
\details{
This function performs a SAR TL analysis on a set of curves. The SAR
procedure in general is given by Murray and Wintle (2000). For the
calculation of the \code{Lx/Tx} value the function \link{calc_TLLxTxRatio} is
used.

\strong{Provided rejection criteria}

\verb{[recycling.ratio]}: calculated for every repeated regeneration dose point.

\verb{[recuperation.rate]}: recuperation rate calculated by
comparing the \code{Lx/Tx} values of the zero regeneration point with the \code{Ln/Tn}
value (the \code{Lx/Tx} ratio of the natural signal).  For methodological
background see Aitken and Smith (1988)
}
\note{
\strong{THIS IS A BETA VERSION}

No TL curves will be removed from the input object without further warning.
}
\section{Function version}{
 0.3.4
}

\examples{

##load data
data(ExampleData.BINfileData, envir = environment())

##transform the values from the first position in a RLum.Analysis object
object <- Risoe.BINfileData2RLum.Analysis(TL.SAR.Data, pos=3)

##perform analysis
analyse_SAR.TL(
 object = object,
 signal_integral = 210:220,
 fit.method = "EXP OR LIN",
 sequence.structure = c("SIGNAL", "BACKGROUND"))

} 

\section{How to cite}{
Kreutzer, S., 2026. analyse_SAR.TL(): Analyse SAR TL measurements. Function version 0.3.4. In: Kreutzer, S., Burow, C., Dietze, M., Fuchs, M.C., Schmidt, C., Fischer, M., Friedrich, J., Mercier, N., Philippe, A., Riedesel, S., Autzen, M., Mittelstrass, D., Gray, H.J., Galharret, J., Colombo, M., Steinbuch, L., Boer, A.d., Bluszcz, A., 2026. Luminescence: Comprehensive Luminescence Dating Data Analysis. R package version 1.2.1. https://r-lum.github.io/Luminescence/
}

\references{
Aitken, M.J. and Smith, B.W., 1988. Optical dating: recuperation
after bleaching.  Quaternary Science Reviews 7, 387-393.

Murray, A.S. and Wintle, A.G., 2000. Luminescence dating of quartz using an
improved single-aliquot regenerative-dose protocol. Radiation Measurements
32, 57-73.
}
\seealso{
\link{calc_TLLxTxRatio}, \link{fit_DoseResponseCurve}, \linkS4class{RLum.Analysis},
\linkS4class{RLum.Results}, \link{get_RLum}
}
\author{
Sebastian Kreutzer, F2.1 Geophysical Parametrisation/Regionalisation, LIAG - Institute for Applied Geophysics (Germany)
, RLum Developer Team}
\keyword{datagen}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Luminescence-package.R
\name{ExampleData.MortarData}
\alias{ExampleData.MortarData}
\alias{MortarData}
\title{Example equivalent dose data from mortar samples}
\format{
Two \code{\link{data.frame}}s containing De and De error
}
\source{
Arbitrary measurements.
}
\description{
Arbitrary data to test the function \code{calc_EED_Model}
}
\examples{

##load data
data(ExampleData.MortarData, envir = environment())

##plot data
plot(MortarData)

}
\references{
unpublished data
}
\keyword{datasets}
\keyword{internal}
